/*
 * @(#)AutoMulti.java	1.6 98/08/26
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
import java.lang.reflect.*;
import java.util.*;
import java.io.*;

/**
 * AutoMulti - automatically generates the Multiplexing UI classes
 * for Swing.
 * <p>
 * To use, type 'java AutoMulti <plafdir> <prefix>' where <plafdir>
 * is the directory containing the source for Swing's UI classes and
 * <prefix> is the package prefix to use for the Multiplexing package.
 * For JDK1.1 builds, <prefix> should be com.sun.java.  For JDK1.2
 * builds, <prefix> should be java.awt.
 *
 * AutoMulti will scour the plaf directory for *UI.java files and
 * generate Multi*UI.java files that do the multiplexing thing.
 * <p>
 * NOTE:  This depends upon the existance of <plafdir> and the
 * compiled classes from <plafdir> being somewhere in the CLASSPATH.
 *
 * @version 1.6 08/26/98 20:34:05
 * @author Willie Walker
 */
public class AutoMulti {
    static String importLines;

    /**
     * A silly list of parameter names to use.  Skip "i" because we use
     * it as a 'for' loop counter.  If you want to get fancy, please feel
     * to change how parameter names are obtained.  This will break if
     * someone decides to create a UI method that takes more than 8
     * parameters.  Which one is a bug (this breaking or having a method
     * with more than eight parameters) is a subjective thing.
     */
    public static String[] paramNames = {"a","b","c","d","e","f","g","h"};

    /**
     * Remove the package names (e.g., javax.swing) from the name.
     */
    public static String unqualifyName(String name) {
        StringTokenizer parser = new StringTokenizer(name,".");
        String unqualifiedName = null;
        while (parser.hasMoreTokens()) {
            unqualifiedName = parser.nextToken();
        }
        return unqualifiedName;
    }

    /**
     * Strips the extension from the filename
     */
    public static String stripExtension(String name) {
        StringTokenizer parser = new StringTokenizer(name,".");
        return parser.nextToken();
    }

    /**
     * Print out some spaces
     */
    public static void indent(StringBuffer s, int i) {
        while (i > 0) {
            s.append(" ");
	    i--;
        }
    }

    /**
     * Spit out all the beginning stuff.
     */
    public static StringBuffer createPreamble(String prefixName) {
	StringBuffer s = new StringBuffer();
	s.append("/*\n");
        s.append(" * %W" + "% %" + "E%\n"); // We don't want SCCS expanding this yet
        s.append(" * \n");
        s.append(" * Copyright 1997, 1998 by Sun Microsystems, Inc.,\n");
        s.append(" * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.\n");
        s.append(" * All rights reserved.\n");
        s.append(" *\n");
        s.append(" * This software is the confidential and proprietary information\n");
        s.append(" * of Sun Microsystems, Inc. (");
	s.append('"');
	s.append("Confidential Information");
	s.append('"');
	s.append(").  You\n");
        s.append(" * shall not disclose such Confidential Information and shall use\n");
        s.append(" * it only in accordance with the terms of the license agreement\n");
        s.append(" * you entered into with Sun.\n");
        s.append(" */\n");
        s.append("package " + prefixName + ".swing.plaf.multi;\n");
        s.append("\n");
	return s;
    }

    /**
     * Add an import line to the String
     */
    public static void addImport(String s, Class theClass) {
	if (!theClass.isPrimitive() && (theClass != Object.class)) {
	    String importLine = new String("import " + theClass.getName() + ";\n");
	    if (importLines.indexOf(importLine) == -1) {
	    	importLines += importLine;
	    }
	}
    }

    /**
     * Spit out the class header information
     */
    public static void addHeader(StringBuffer s, String className) {
	s.append("/**\n");
	s.append(" * Multi" + className + " implementation\n");
        s.append(" * \n");
        s.append(" * <p>This file was automatically generated by AutoMulti.\n");
        s.append(" *\n");
        s.append(" * @version %" + "I% %" + "G% %" + "U%\n"); 	// Don't let SCCS expand this yet
        s.append(" * @author  Otto Multey\n");			// Get it?  I crack myself up.
        s.append(" */\n");
        s.append("public class Multi" + className + " extends " + className + " {\n");
        s.append("\n");
        s.append("    /**\n");
        s.append("     * The Vector containing the real UI's.  This is populated \n");
        s.append("     * in the call to createUI, and can be obtained by calling\n");
        s.append("     * getUIs.  The first element is guaranteed to the real UI \n");
        s.append("     * obtained from the default look and feel.\n");
        s.append("     */\n");
        s.append("    protected Vector uis = new Vector();\n");
        s.append("\n");
        s.append("////////////////////\n");
        s.append("// Common UI methods\n");
        s.append("////////////////////\n");
        s.append("\n");
        s.append("    /**\n");
        s.append("     * Return the list of UI's associated with this multiplexing UI.  This \n");
        s.append("     * allows processing of the UI's by an application aware of multiplexing \n");
        s.append("     * UI's on components.\n");
        s.append("     */\n");
        s.append("    public ComponentUI[] getUIs() {\n");
        s.append("        return MultiLookAndFeel.uisToArray(uis);\n");
        s.append("    }\n");
    }

    /**
     * Print out the code for a method.  This is pretty specific to the
     * Multiplexing UI code, so don't get any fancy ideas.
     */
    public static void addMethod(StringBuffer s, Method m, String origName, String className) {

        // Get the method name and the return type.  Be a little careful about arrays.
        //
        String methodName = unqualifyName(m.getName());
        String returnType;
        if (!m.getReturnType().isArray()) {
            returnType = unqualifyName(m.getReturnType().toString());
	    addImport(importLines,m.getReturnType());
        } else {
            returnType = unqualifyName(m.getReturnType().getComponentType().toString())
		 	 + "[]";
	    addImport(importLines,m.getReturnType().getComponentType());
        }

        // Print the javadoc
        //
        s.append("\n");
        if (methodName.equals("createUI")) {
	    s.append("    /**\n");
	    s.append("     * Return a multiplexing UI instance if any of the auxiliary\n");
	    s.append("     * LookAndFeels support this UI.  Otherwise, just return a \n");
	    s.append("     * UI obtained using the normal methods.\n");
	    s.append("     */\n");
        } else if (!returnType.equals("void")) {
	    s.append("    /**\n");
	    s.append("     * Call " + methodName + " on each UI handled by this MultiUI.\n");
	    s.append("     * Return only the value obtained from the first UI, which is\n");
	    s.append("     * the UI obtained from the default LookAndFeel.\n");
	    s.append("     */\n");
        } else {
	    s.append("    /**\n");
	    s.append("     * Call " + methodName
		     + " on each UI handled by this MultiUI.\n");
	    s.append("     */\n");
        }

        // Print the method signature
        //
        s.append("    public");
        if (Modifier.isStatic(m.getModifiers())) {
            s.append(" static");
        }
        s.append(" " + returnType);
        s.append(" " + methodName);
        s.append("(");

        Class[] params = m.getParameterTypes();
	Class temp;
	String braces;
        for (int i = 0; i < params.length; i++) {
            if (i > 0) {
                s.append(", ");
            }
	    temp = params[i];
	    braces = new String("");
	    while (temp.isArray()) {
	        braces += "[]";
                temp = temp.getComponentType();
            }
            s.append(unqualifyName(temp.getName()) + braces + " " + paramNames[i]);
	    addImport(importLines,temp);
        }
        s.append(")");

        // Don't forget about exceptions
        //
        Class exceptions[] = m.getExceptionTypes();
	String throwsString = new String("");

        if (exceptions.length > 0) {
	    s.append("\n");
	    indent(s,12);
            s.append("throws ");
            for (int i = 0; i < exceptions.length; i++) {
                if (i > 0) {
                    s.append(", ");
                }
	        s.append(unqualifyName(exceptions[i].getName()));
	        addImport(importLines,exceptions[i]);
	    }
        }
        s.append(throwsString + " {\n");

        // Now print out the contents of the method.  We do a special thing
        // for the createUI method, another thing if the method returns 'void'
        // and a third thing if we don't do either of the first two.  If
        // you want to squash this down, feel free.
        //
        if (methodName.equals("createUI")) {
	    indent(s,8);
            s.append("ComponentUI mui = new Multi" + origName + "();\n");
	    indent(s,8);
            s.append("return MultiLookAndFeel.createUIs(mui,\n");
	    indent(s,42);
	    s.append("((Multi" + origName +") mui).uis,\n");
	    indent(s,42);
	    for (int i = 0; i < params.length; i++) {
                if (i > 0) {
                    s.append(",");
                }
		s.append(paramNames[i]);
            }
            s.append(");\n");
        } else if (!returnType.equals("void")) {
	    indent(s,8);
            s.append(returnType + " returnValue = \n");
	    indent(s,12);
            s.append("((" + className + ") (uis.elementAt(0)))."
	  	     + methodName + "(");
	    for (int i = 0; i < params.length; i++) {
                if (i > 0) {
                    s.append(",");
                }
		s.append(paramNames[i]);
            }
            s.append(");\n");
	    indent(s,8);
	    s.append("for (int i = 1; i < uis.size(); i++) {\n");
	    indent(s,12);
	    s.append("((" + className + ") (uis.elementAt(i)))."
		     + methodName + "(");
	    for (int i = 0; i < params.length; i++) {
                if (i > 0) {
                    s.append(",");
                }
		s.append(paramNames[i]);
            }
            s.append(");\n");
	    indent(s,8);
            s.append("}\n");
	    indent(s,8);
            s.append("return returnValue;\n");
        } else {
	    indent(s,8);
	    s.append("for (int i = 0; i < uis.size(); i++) {\n");
	    indent(s,12);
	    s.append("((" + className + ") (uis.elementAt(i)))."
		     + methodName + "(");
	    for (int i = 0; i < params.length; i++) {
                if (i > 0) {
                    s.append(",");
                }
		s.append(paramNames[i]);
            }
            s.append(");\n");
	    indent(s,8);
            s.append("}\n");
        }
	indent(s,4);
        s.append("}\n");
    }

    /**
     * Take a plaf class name (e.g., "MenuUI") and generate the corresponding
     * Multiplexing UI java source code (e.g., "MultiMenuUI.java").
     */
    public static void generateFile(String prefixName, String className) {
        try {
            FileOutputStream fos;
            PrintWriter outFile;

	    importLines = new String();
	    importLines += new String("import java.util.Vector;\n");

	    StringBuffer body = new StringBuffer();
            Class wee = Class.forName(prefixName + ".swing.plaf." + className);
	    String weeName = unqualifyName(wee.getName());
	    addImport(importLines,wee);
	    while (!weeName.equals("Object")) {
                body.append("\n");
                body.append("////////////////////\n");
	        body.append("// " + weeName + " methods\n");
                body.append("////////////////////\n");
                Method[] methods = wee.getDeclaredMethods();
                for (int i=0; i < methods.length; i++) {
                    if (Modifier.isPublic(methods[i].getModifiers())) {
                        addMethod(body,methods[i],className,weeName);
                    }
                }
		wee = wee.getSuperclass();
                weeName = unqualifyName(wee.getName());
	        addImport(importLines,wee);
            }

	    fos = new FileOutputStream("Multi" + className + ".java");
            outFile = new PrintWriter(fos);
	    StringBuffer outText = createPreamble(prefixName);
	    outText.append(importLines.toString() + "\n");
	    addHeader(outText,className);
	    outText.append(body.toString());
            outText.append("}\n");
	    outFile.write(outText.toString());
	    outFile.flush();
	    outFile.close();
        } catch (Exception e) {
	    System.err.println(e);
        }
    }

    /**
     * D'Oh!  Something bad happened.
     */
    public static void usage(String s) throws IOException {
	System.err.println("Usage:  AutoMulti <plafdir> [com.sun]");
	throw new IllegalArgumentException(s);
    }

    /**
     * Take the plaf directory name and generate the multiplexing UI
     * source code.
     */
    public static void main(String[] args) throws IOException {

        if (args.length < 1) {
	    usage("");
        }

        String dirName = args[0];
        File dir = new File(dirName);
	if (!dir.isDirectory()) {
	    System.err.println("No such directory:  " + dirName);
	    usage("");
        }

        String prefixName;
	if (args.length > 1) {
	    prefixName = args[1];
	} else {
	    prefixName = "com.sun.java";
        }

        String plafUIs[] = dir.list(new UIJavaFilter());
        for (int i = 0; i < plafUIs.length; i++) {
	    generateFile(prefixName,stripExtension(plafUIs[i]));
	}
    }
}

/**
 * Only accept file names of the form *UI.java.  The one exception
 * is not accepting ComponentUI.java because we don't need to generate
 * a multiplexing class for it.
 */
class UIJavaFilter implements FilenameFilter {
    public boolean accept(File dir, String name) {
	if (name.equals("ComponentUI.java")) {
	    return false;
	} else if (name.endsWith("UI.java")) {
	    return true;
	} else {
	    return false;
        }
    }
}
