/*
 * @(#)Data.java	1.4 98/05/11
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.im.iiimp;

import java.io.IOException;
import java.io.EOFException;

class Data {
    int pos;
    int count;
    byte buf[];

    private final static int DATA_BOUNDARY = 4;

    /**
     * Creates a new instance of Data.
     */
    Data() {
	buf = new byte[32];
	count = 0;
	pos = 0;
    }

    /**
     * Creates a new instance of Data with the specified length
     * of the specified byte buffer.
     */
    Data(byte b[], int len) {
	pos = 0;
	count = len;
	if (len > 0) {
	    buf = new byte[len];
	    System.arraycopy(b, 0, buf, 0, len);
	}
    }

    /**
     * Creates a copy of the specified Data.  The only unread 
     * buffer will be copied.
     */
    Data(Data d) {
	int newVal = d.count - d.pos;
	if (newVal > 0) {
	    byte newValue[] = new byte[newVal];
	    System.arraycopy(d.buf, d.pos, newValue, 0, newVal);
	    buf = newValue;
	}
	count = newVal;
	pos = 0;
    }

    void write(int b) throws IOException {
	ensureCapacity(count + 1);
	buf[count++] = (byte) (b & 0xFF);
    }

    void write2(int b) throws IOException {
	ensureCapacity(count + 2);
	buf[count++] = (byte) ((b >>> 8) & 0xFF);
	buf[count++] = (byte) ((b >>> 0) & 0xFF);
    }

    void write4(int b) throws IOException {
	ensureCapacity(count + 4);
	buf[count++] = (byte) ((b >>> 24) & 0xFF);
	buf[count++] = (byte) ((b >>> 16) & 0xFF);
	buf[count++] = (byte) ((b >>>  8) & 0xFF);
	buf[count++] = (byte) ((b >>>  0) & 0xFF);
    }

    void writeBytes(byte b[]) throws IOException {
	ensureCapacity(count + b.length);
	System.arraycopy(b, 0, buf, count, b.length);
	count += b.length;
    }

    void writeBytes(byte b[], int length) throws IOException {
	ensureCapacity(count + length);
	System.arraycopy(b, 0, buf, count, length);
	count += length;
    }

    void writeString(String utf) throws IOException {
	int utflen = utf.length();
	int len = utflen * 2;

	write2(len);

	int pad = paddings(len + 2);

	ensureCapacity(count + len + pad);

	for (int i = 0 ; i < utflen ; i++) {
	    char c = utf.charAt(i);
	    buf[count++] = (byte) ((c >>  8) & 0xFF);
	    buf[count++] = (byte) ((c >>  0) & 0xFF);
	}

	for (int j = 0; j < pad; j++) {
	    buf[count++] = (byte) 0;
	}
    }

    int read() throws IOException {
	if (pos >= count) {
	    throw new EOFException();
	}
	int ch = (int) (buf[pos++] & 0x000000FF);
	if (ch < 0) {
	    throw new EOFException();
	}
	return ch;
    }

    int read2() throws IOException {
	int ch1 = read();
	int ch2 = read();
	return (ch1 << 8) + (ch2 << 0);
    }

    int read4() throws IOException {
	int ch1 = read();
	int ch2 = read();
	int ch3 = read();
	int ch4 = read();
	return ((ch1 << 24) + (ch2 << 16) + (ch3 << 8) + (ch4 << 0));
    }

    int read(byte b[], int offset, int size) throws IOException {
	if (size <= 0) {
	    return 0;
	}

	int c = read();
	if (c == -1) {
	    return -1;
	}
	b[offset] = (byte) c;

	int i = 1;
	try {
	    for (; i < size ; i++) {
		c = read();
		if (c == -1) {
		    break;
		}
		if (b != null) {
		    b[offset + i] = (byte) c;
		}
	    }
	} catch (IOException ee) {
	}
	return i;
    }

    String readString() throws IOException {
	int len = read2();
	if (len == 0) {
	    // need to read 2 bytes for padding
	    read2();
	    return (String)null;
	}

	int utflen = len / 2;
        char str[] = new char[utflen];
	int i = 0;
	while (i < utflen) {
	    int ch1 = read();
	    int ch2 = read();
	    str[i++] = (char) ((ch1 << 8) + (ch2 << 0));
	}
	int pad = paddings(len + 2);
	for (int j = 0; j < pad; j++) {
	    read();
	}
        return new String(str, 0, utflen);
    }

    void ensureCapacity(int minimumNum) {
	int maxNum = buf.length;
	if (minimumNum > maxNum) {
	    int newNum = (maxNum + 1) * 2;
	    if (minimumNum > newNum) {
		newNum = minimumNum;
	    }
	    byte newValue[] = new byte[newNum];
	    System.arraycopy(buf, 0, newValue, 0, count);
	    buf = newValue;
	}
    }

    final static int paddings(int len) {
	return ((DATA_BOUNDARY -((len) % DATA_BOUNDARY)) % DATA_BOUNDARY);
    }

    void pad() throws IOException {
	int pad = paddings(count);
	int total = count + pad;
	for (int i = count; i < total; i++) {
	    write(0);
	}
    }

    void skipBytes(int skip) {
	int num = (count < (pos + skip)) ? (count - pos) : skip;
	pos += num;
    }

    int available() {
	return (count > pos) ? (count - pos) : 0;
    }
}
