/*
 * @(#)IIIMPAdapter.java	1.13 98/08/25
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.im.iiimp;

import java.lang.Character.Subset;
import java.awt.*;
import java.awt.event.*;
import java.awt.font.TextAttribute;
import java.awt.font.TextHitInfo;
import java.awt.im.InputMethodHighlight;
import java.awt.im.InputMethodRequests;
import java.io.IOException;
import java.text.AttributedString;
import java.text.AttributedCharacterIterator;
import java.util.Locale;
import java.util.Vector;
import sun.awt.im.InputMethodContext;

public class IIIMPAdapter extends sun.awt.im.InputMethodAdapter {

    // LOCK is used to obtain the global lock for the IIIMP adapter
    // global resources (im, status, lookup)
    private final static Object LOCK = new Object();

    private static sun.awt.im.iiimp.InputMethod im = null;

    private static LookupWindow lookup = null;

    private static StatusWindow status = null;

    private sun.awt.im.iiimp.InputContext ic = null;

    private boolean createICFailed = false;

    private Component client = null;

    private String committedText = null;

    private class ComposedChar {
	char c;
	java.awt.im.InputMethodHighlight attr;
    }
    private Vector composedText = new Vector();

    /**
     * This method returns a list of currently available locales via an
     * IIIMP server. If the IIIMP environment is properly configured and the
     * server is not temporarily available (e.g., connection time out), null
     * is returned. If the IIIMP environment is not properly configured
     * (e.g., the server host name is not given), IOException is thrown.
     */
    public static Locale[] getAvailableLocales() throws IOException {
	synchronized (LOCK) {
	    if (im == null) {
		im = sun.awt.im.iiimp.InputMethod.getInstance();
	    }
	}

	try {
	    if (im.openIM() == false) {
		return null;
	    }
	    return im.getAvailableLocales();
	} catch (IOException e) {
	    synchronized (LOCK) {
		im.dispose();
		im = null;
	    }
	    throw e;
	}
    }

    public IIIMPAdapter() throws AWTException {
	synchronized (LOCK) {
	    if (im == null) {
		im = sun.awt.im.iiimp.InputMethod.getInstance();
	    }
	    if (lookup == null) {
		lookup = new LookupWindow();
	    }
	    if (status == null) {
		status = new StatusWindow();
	    }
	}
	boolean ret = false;
	try {
	    ret = im.openIM();
	} catch (Exception e) {
	    /*
	     * IOException will be thrown when the server location
	     * is neither specified nor a correct URL.
	     */
	    synchronized (LOCK) {
		im.dispose();
		im = null;
		lookup = null;
		status = null;
	    }
	    throw new AWTException("can't find the IIIMP server");
	}
	/* "ret" will be false when the server does not run. */
	if (ret == false) {
	    throw new AWTException("can't connect the IIIMP server");
	}
    }

    public synchronized boolean setLocale(Locale lang) {
	if (lang == null) {
	    return false;
	}
	if (ic == null) {
	    return createIC(lang);
	}
	return ic.setLocale(lang);
    }

    public synchronized void dispatchEvent(AWTEvent e) {
	if (ic == null || ((e instanceof KeyEvent) == false)) {
	    return;
	}
	ic.dispatchKeyEvent((KeyEvent) e);
    }

    public synchronized void activate() {
	if (createICFailed == true) {
	    return;
	}

	client = getClientComponent();

	if (ic == null) {
	    if (createIC(Locale.getDefault()) == false) {
		return;
	    }
	}

	/*
	 * While Lookup is up,  if the mouse is moved to another window,
	 * reset the ic that starts the Lookup.  This allows to do the same
	 * behavior as when endComposed() is called.
	 */
	if (lookup.isShowing() == true && client != lookup.getComponent()) {
	    sun.awt.im.iiimp.InputContext resetIC = lookup.getInputContext();
	    if (resetIC != null) {
		resetIC.reset(false);
	    }
	}

	ic.activate();
    }

    public synchronized void deactivate(boolean isTemporary) {
	if (ic != null) {
	    ic.deactivate();
	}
	status.hide();
    }

    public synchronized void removeNotify() {
	if (composedText != null)
	    composedText.setSize(0);
	dispose();
	client = null;
    }

    public void endComposition() {
	if (ic != null) {
	    ic.reset(false);
	}
    }

    public synchronized void dispose() {
	if (ic != null) {
	    ic.dispose();
	    ic = null;
	    // Does "im" need to close the connection?
	    if (im.getInputContext() == null) {
		// im.dispose();
	    }
	}
	status.hide();
	status.done();
	lookup.hide();
	lookup.done();
    }

    public void finalize() {
	if (ic != null) {
	    ic.dispose();
	    ic = null;
	    // Does "im" need to close the connection?
	    if (im.getInputContext() == null) {
		// im.dispose();
	    }
	}
    }

    public void setCharacterSubsets(Subset subsets[]) {}
    
    public Object getControlObject() {
        return null;
    }

    private boolean createIC(Locale loc) {
	try {
	    ic = sun.awt.im.iiimp.InputContext.getInstance(im, loc);
	    ic.addCommitStringListener(new Commit());
	    ic.addLookupListener(new Lookup());
	    ic.addStatusListener(new Status());
	    ic.addPreeditListener(new Preedit());
	} catch (Exception e) {
	    createICFailed = true;
	    return false;
	}
	return true;
    }

    private void postInputMethodEvent(int id,
				      AttributedCharacterIterator text,
				      int committedCharacterCount,
				      TextHitInfo caret,
				      TextHitInfo visiblePosition) {
	inputContext.dispatchInputMethodEvent(
	    id,
	    text,
	    committedCharacterCount,
	    caret,
	    visiblePosition);
    }

    synchronized private void dispatchCommittedText(String str) {
        if (str == null || str.length() == 0) {
            return;
	}
	if (composedText.isEmpty() == true) {
	    AttributedString attrstr = new AttributedString(str);
	    postInputMethodEvent(InputMethodEvent.INPUT_METHOD_TEXT_CHANGED,
				 attrstr.getIterator(),
				 str.length(),
				 null,
				 null);
	} else {
	    // if there is composed text, wait until the preedit
	    // callback is invoked.
	    committedText = str;
	}
    }

    private InputMethodHighlight convertVisualFeedbackToHighlight(int
								  feedback) {
	InputMethodHighlight h;
	switch (feedback) {
	case sun.awt.im.iiimp.InputMethodFeedback.UNDERLINE:
	    h = InputMethodHighlight.UNSELECTED_CONVERTED_TEXT_HIGHLIGHT;
	    break;
	case sun.awt.im.iiimp.InputMethodFeedback.REVERSE:
	    h = InputMethodHighlight.SELECTED_CONVERTED_TEXT_HIGHLIGHT;
	    break;
	case sun.awt.im.iiimp.InputMethodFeedback.HIGHLIGHT:
	    h = InputMethodHighlight.SELECTED_RAW_TEXT_HIGHLIGHT;
	    break;
	case sun.awt.im.iiimp.InputMethodFeedback.PRIMARY:
	    h = InputMethodHighlight.UNSELECTED_CONVERTED_TEXT_HIGHLIGHT;
	    break;
	case sun.awt.im.iiimp.InputMethodFeedback.SECONDARY:
	    h = InputMethodHighlight.SELECTED_CONVERTED_TEXT_HIGHLIGHT;
	    break;
	case sun.awt.im.iiimp.InputMethodFeedback.TERTIARY:
	default:
	    h = InputMethodHighlight.SELECTED_RAW_TEXT_HIGHLIGHT;
	    break;
	}
	return h;
    }

    class Commit implements CommitListener {
	public void commitPerformed(CommitEvent e) {
	    String str = null;
	    if (e.getType() == CommitEvent.TEXT_STRING) {
		str = e.getString();
	    } else if (e.getType() == CommitEvent.TEXT_INPUTMETHODTEXT) {
		InputMethodText t = e.getText();
		str = t.toString();
	    }
	    dispatchCommittedText(str);
	}
    }

    private InputMethodContext getInputContext() {
	return inputContext;
    }
    class Lookup implements LookupListener {
	public void lookupStart(LookupEvent e) {
	    InputMethodLookupStart d = e.getStart();
	    lookup.setRowCol(d.getRows(), d.getColumns());
	}
	public void lookupDraw(LookupEvent e) {
	    lookup.setComponent(client);
	    lookup.setInputContext(e.getInputContext());
	    InputMethodContext ic = getInputContext();

	    if (ic != null && !composedText.isEmpty()) {
		Rectangle rec = ic.getTextLocation(TextHitInfo.leading(0));
                
		int start_x = rec.x;
		int start_y = rec.y;
		int end_x = 0;
		int end_y = 0;
		int end_pos = composedText.size();
		if (end_pos > 0) {
		    rec = ic.getTextLocation(TextHitInfo.leading(end_pos));
		    end_x = rec.x;
		    end_y = rec.y;
		}
		if (end_y > start_y) {
		    start_x = end_x;
		    start_y = end_y;
		}
		lookup.setLocation(start_x, start_y);
	    }
	    lookup.show(e);
	}
	public void lookupProcess(LookupEvent d) {}
	public void lookupDone(LookupEvent e) {
	    lookup.hide();
	    lookup.done();
	    Component c = lookup.getComponent();
	    if (c != null) {
		c.requestFocus();
	    }
	}
    }

    class Status implements StatusListener {
	public void statusStart(StatusEvent e) {
	    status.setComponent(client);
	    status.setInputContext(e.getInputContext());
	}
	public void statusDraw(StatusEvent e) {
	    status.show(e);
	}
	public void statusDone(StatusEvent e) {
	    status.done();
	    if (im.isConnected() == false) {
		status.hide();
	    }
	}
    }

    class Preedit implements PreeditListener {
	int caretPos = 0;
	public int preeditStart(PreeditEvent e) {
	    caretPos = 0;
	    composedText.setSize(0);
	    return 1024;
	}
	public void preeditDone(PreeditEvent e) {
	    caretPos = 0;
	    composedText.setSize(0);
	}
	public void preeditDraw(PreeditEvent e) {
	    dispatchComposedText(e);
	}
	public int preeditCaret(PreeditEvent e) {
	    InputMethodCaret c = e.getCaret();
	    return dispatchCaretPosition(c.getPosition());
	}

	synchronized void dispatchComposedText(PreeditEvent e) {
	    InputMethodPreeditText p = e.getPreeditText();
	    InputMethodText text = p.getText();
	    int pos = p.getPosition();
	    int first_changed = p.getFirstChangedPosition();
	    int length_changed = p.getTextChangedLength();

	    // composedText deletion
	    if (text == null) {
		for (int i = (first_changed + length_changed - 1);
		     i >= first_changed; i--) {
		    if (composedText.isEmpty() == false) {
			composedText.removeElementAt(i);
		    }
		}
	    } else {
		// composedText replace
		if (length_changed > 0) {
		    for (int i = (first_changed + length_changed - 1);
			 i >= first_changed; i--) {
			if (composedText.isEmpty() == false) {
			    composedText.removeElementAt(i);
			}
		    }
		}
		int len = text.count;
		for (int i = first_changed, j = 0;
		     i < (first_changed+len); i++, j++) {
		    InputMethodChar c = text.value[j];
		    ComposedChar comp = new ComposedChar();
		    comp.c = c.c;
		    comp.attr =
			convertVisualFeedbackToHighlight(c.fd[0].value);
		    composedText.insertElementAt(comp, i);
		}
	    }

	    // no composedText or no composedText but committedText
	    if (composedText.isEmpty() == true) {
		if (committedText == null) {
		    postInputMethodEvent(InputMethodEvent.
					 INPUT_METHOD_TEXT_CHANGED,
					 null,
					 0,
					 null,
					 null);
		} else {
		    dispatchCommittedText(committedText);
                    committedText = null;
		}
		return;
	    }

	    int composedOffset = 0;

	    // merge committedText to composedText if any committedText
	    if (committedText != null) {
		for (int i = 0; i < committedText.length(); i++) {
		    ComposedChar comp = new ComposedChar();
		    comp.c = committedText.charAt(i);
		    comp.attr = null;
		    composedText.insertElementAt(comp, i);
		}
		composedOffset = committedText.length();
		committedText = null;
	    }

	    // post im event
	    AttributedString attrstr = getAttributedString();
	    postInputMethodEvent(InputMethodEvent.INPUT_METHOD_TEXT_CHANGED,
				 attrstr.getIterator(),
				 composedOffset,
				 TextHitInfo.leading(pos),
				 null);
	}

	int dispatchCaretPosition(int pos) {
	    caretPos = pos;
	    postInputMethodEvent(InputMethodEvent.CARET_POSITION_CHANGED,
				 null,
				 0,
				 TextHitInfo.leading(caretPos),
				 null);
	    return caretPos;
	}
    }

    private AttributedString getAttributedString() {
	if (composedText.isEmpty() == true) {
	    return new AttributedString("");
	}

	ComposedChar[] c = new ComposedChar[composedText.size()];
	composedText.copyInto(c);
        
	StringBuffer buf = new StringBuffer();
	for (int i = 0; i < c.length; i++) {
	    buf.append(c[i].c);
	}
	AttributedString attrstr = new AttributedString(buf.toString());

	int start = 0;
	int i = 0;
	InputMethodHighlight cur = c[0].attr;
	for (; i < c.length; i++) {
	    InputMethodHighlight next = c[i].attr;
	    if (next != cur) {
		if (cur != null) {
		    attrstr.
		    addAttribute(TextAttribute.INPUT_METHOD_HIGHLIGHT,
				 cur, start, i);
		}
		cur = next;
		start = i;
	    }
	}
	if (cur != null) {
	    attrstr.addAttribute(TextAttribute.INPUT_METHOD_HIGHLIGHT,
			     cur, start, i);
	}

	return attrstr;
    }
}

abstract class IIIMPWindow {
    static int screenHeight;
    static int screenWidth;

    Window win = null;
    Component source = null;
    sun.awt.im.iiimp.InputContext ic = null;

    synchronized void setComponent(Component comp) {
	this.source = comp;
    }

    Component getComponent() {
	return source;
    }

    sun.awt.im.iiimp.InputContext getInputContext() {
	return ic;
    }

    synchronized void setInputContext(sun.awt.im.iiimp.InputContext ic) {
	this.ic = ic;
    }

    synchronized void done() {
	source = null;
	ic = null;
    }

    synchronized void hide() {
	if (win != null) {
	    win.setVisible(false);
	}
    }

    static {
	Dimension d = Toolkit.getDefaultToolkit().getScreenSize();
	screenHeight = d.height;
	screenWidth = d.width;
    }
}

class LookupWindow extends IIIMPWindow {
    private static final int widthM = 20;
    private static final int heightM = 45;
    private static final String fontFace = "Monospaced";

    private int fontHeight;
    private int fontAscent;
    private int fontWidth;
    private FontMetrics fm;
    private LookupPane pane;
    private int columHeight;
    private int rows = 0;
    private int cols = 0;
    private int ic_x = 0;
    private int ic_y = 0;
    
    LookupWindow() {}

    void init() {
        win = new Dialog(new Frame(""));
        ((Dialog)win).setResizable(true);

	Font font = new Font(fontFace, Font.PLAIN, 12);
        win.setFont(font);

        fm = Toolkit.getDefaultToolkit().getFontMetrics(font);
        fontHeight = fm.getHeight();
        fontWidth = fm.getMaxAdvance();
        fontAscent = fm.getAscent();
        columHeight = fontHeight + 2;

        pane = new LookupPane();
        win.add(pane);

	win.pack();
    }

    void setRowCol(int r, int c) {
	rows = r;
	cols = c;
    }

    synchronized void show(LookupEvent e) {
	if (source == null) {
	    return;
	}

	if (win == null) {
	    init();
	}

	InputMethodLookupDraw d = e.getDraw();

	// set the title.
	InputMethodText title = d.getTitle();
	if (title != null) {
	    ((Dialog)win).setTitle(title.toString());
	} else {
	    ((Dialog)win).setTitle("");
	}

	// repaint the pane with the new data.
        pane.setData(d);
        pane.repaint();

	// show it.
        win.show();
    }

    boolean isShowing() {
	return (win == null) ? false : win.isShowing();
    }

    private boolean isSame = false;
    boolean compareText(InputMethodLookupText[] lc1, int lc1_start,
			InputMethodLookupText[] lc2, int lc2_start,
			int length) {
	return false;

	/*
	 * compareText() does compare the previous candidates and
	 * the newly sent candidates.  The content of those candidates
	 * often is the same except the index that the language
	 * engine focuses on right now.  By compareText(), the pane
	 * tries to avoid painting the same candidates twice.
	 * However, it does not work very well on Windows though it
	 * works fine on Solaris.  There must be a bug to be fixed or
	 * a trick needed in the painting routines.  Meantime,
	 * compareText() always returns false.
	 */
	/*
        if (lc1 == null || lc2 == null) {
            return false;
        }
        for (int i = 0; i < length; i++) {
            int len = lc1[lc1_start + i].value.count;

            if (len != lc2[lc2_start + i].value.count)
                return false;
            
            for (int j = 0; j < len; j++) {
                if (lc1[lc1_start + i].value.value[j].c !=
                    lc2[lc2_start + i].value.value[j].c)
                    return false;
            }
        }
        return true;
	*/
    }

    void setLocation(int x, int y) {
	ic_x = x;
	ic_y = y;
    }

    /*
     * Calculate the window position based on the given coordinate by
     * InputMethodContext's getTextLocation().
     */
    Point calculatePopupPosition() {
	Point pos = new Point(ic_x, ic_y);

	Rectangle rec = win.getBounds();
	int lookupHeight = rec.height;
	int lookupWidth = rec.width;

	pos.translate(0, widthM);
	int slide = pos.y + lookupHeight - screenHeight;
	if (slide > 0) {
	    // then position it above ic_y.
	    pos.translate(0, (widthM + lookupHeight + heightM) * -1);
	}
	if (pos.x < 0) {
	    pos.translate(pos.x * -1, 0);
	}
	slide = pos.x + lookupWidth - screenWidth;
	if (slide > 0) {
	    pos.translate(slide * -1, 0);
	}

	return pos;
    }

    private class LookupPane extends Panel {
        private final int initSize = 26;
        private final int xMargin = 10;
        private final int yMargin = 5;
        private final int cMargin = 3;

	private int width = 0;
	private int height = 0;
        private int first = 0;
        private int last = 0;
        private int length = 0;
        private int current = 0;
        private int previous = 0;
        private InputMethodLookupText[] text = null;
        private String[] textDrawn = null;
        private int maxWidth = 0;

        LookupPane() {
            super();
	    enableInputMethods(false);
	    addFocusListener(new FocusLostListener());
	    addKeyListener(new KeyTypeListener());
	    addMouseListener(new MouseFocusListener());
            textDrawn = new String[initSize];
        }

        void setData(InputMethodLookupDraw data) {
	    int first_index = data.getFirstIndex();
	    int last_index = data.getLastIndex();
	    InputMethodLookupText[] value = data.getText();

            if (length == (last_index - first_index + 1) &&
                compareText(text, first, value, first_index, length)) {
                isSame = true;
            } else {
                text = value;
                first = first_index;
                last = last_index;
                length = last - first + 1;
                if (textDrawn.length < length) {
                    textDrawn = new String[length];
                }

                String candidate; 
                width = height = 0;
                int realRow = 0;
                int temp;
                maxWidth = 0;
                for (int i = 0; i < length; i++) {
                    candidate = text[first + i].label.toString() +
                        " : " + text[first + i].value.toString() +
                        "  ";
                    if (maxWidth < (temp = fm.stringWidth(candidate))) {
                        maxWidth = temp;
                    }
                    textDrawn[i] = candidate;
                    if (i % cols == 0) {
                        realRow++;
                    }
                }
                width = maxWidth * (length > cols ? cols : length);
                height = columHeight * realRow;
            }
            previous = current;
            current = data.current_index;
        }

        public void update(Graphics g) {
            if (isSame == false) {
		super.update(g);
		return;
            }
            paint(g);
        }
        
        public void paint(Graphics g) {
            if (text == null || length == 0) {
                return;
	    }

            int x = xMargin;
            int y = yMargin;

	    if (!isSame) {
		win.setSize(pane.width + widthM, pane.height + heightM);
		Point p = calculatePopupPosition();
		if (!p.equals(win.getLocation())) {
		    win.setLocation(p.x, p.y);
		}
		win.show();
	    }
            for (int i = 0; i < length; i++) {
                if (i % cols == 0) {
                    y += columHeight;
                    x = xMargin;
                } else {
                    x += maxWidth;
                }
                if (isSame && i == previous) {
                    g.setColor(getBackground());
                    g.drawRoundRect(x - cMargin,
                                    y - columHeight + cMargin,
                                    maxWidth - cMargin, columHeight - 1,
                                    cMargin, cMargin);
                    g.setColor(getForeground());
                }
                if (i == current) {
                    g.drawRoundRect(x - cMargin,
                                    y - columHeight + cMargin,
                                    maxWidth - cMargin, columHeight - 1,
                                    cMargin, cMargin);
                }
                if (!isSame) {
                    g.drawString(textDrawn[i], x, y);
		}
            }
	    isSame = false;
        }
    }


    class FocusLostListener extends FocusAdapter {
	public void focusLost(FocusEvent e) {
	    if (win.isShowing() == true) {
		e.getComponent().requestFocus();
	    }
	}
    }

    class KeyTypeListener extends KeyAdapter {
	public void keyPressed(KeyEvent e) {
	    if (source != null) {
		source.dispatchEvent(e);
	    }
	}
    }

    class MouseFocusListener extends MouseAdapter {
	public void mouseClicked(MouseEvent e) {
	    if (source != null) {
		KeyEvent kev = new KeyEvent(source,
					    KeyEvent.KEY_PRESSED,
					    e.getWhen(), 0,
					    KeyEvent.VK_ENTER, '\n');
		source.dispatchEvent(kev);
	    }
	} 
    }
}

class StatusWindow extends IIIMPWindow { 
    private static final int xMargin = 15;
    private static final int bottomMargin = 5;
    private static final int windowW = 80;
    private static final int windowH = 20;

    private StatusPane pane = null;
    private boolean isRoot = false;

    StatusWindow() {}

    void init() {
	String style =
	    sun.awt.im.iiimp.InputMethod.getProperty("iiimp.status.style");
	if (style.equals("root") == true) {
	    isRoot = true;
	}

	win = new Window(new Frame());
	win.setForeground(Color.black);
	win.setBackground(Color.lightGray);

	pane = new StatusPane();
	win.add(pane);

	/* The size has to be set now though no reason for this at start-up. */
	win.setBounds(xMargin, screenHeight - windowH - bottomMargin,
		      windowW, windowH);
	win.setVisible(true);

	win.pack();
    }

    synchronized void show(StatusEvent d) {
	if (source == null) {
	    return;
	}

	if (win == null) {
	    init();
	}

	// repaint the pane with the new data.
	pane.setData(d);
	pane.repaint();

	// relocate the pane if necessary.
	if (isRoot == false) {
	    Point p = calculatePopupPosition();
	    if (!p.equals(win.getLocation())) {
		win.setLocation(p.x, p.y);
	    }
	}

	// show it.
	win.show();
    }

    class StatusPane extends Canvas {
	private StatusEvent event;
	
	StatusPane() {
	    super();
	    enableInputMethods(false);
	}

	void setData(StatusEvent e) {
	    event = e;
	}

	public void paint(Graphics g) {
	    if (event == null) {
		return;
	    }

	    String str = null;
	    if (event.getType() == StatusEvent.TEXT_STRING) {
		str = event.getString();
	    } else if (event.getType() == StatusEvent.TEXT_INPUTMETHODTEXT) {
		InputMethodText t = event.getText();
		str = t.toString();
	    }
	    if (str == null) {
		return;
	    }

	    int w = g.getFontMetrics().stringWidth(str);
	    Rectangle rec = win.getBounds();
	    if (w > rec.width) {
		win.setBounds(rec.x, rec.y, xMargin * 2 + w, rec.height);
	    }

	    g.setColor(Color.darkGray);
	    g.fillRect(rec.width - 2, 0, 2, rec.height);
	    g.fillRect(0, rec.height - 2, rec.width, 2);

	    g.setColor(Color.white);
	    g.fillRect(0, 0, 2, rec.height);
	    g.fillRect(0, 0, rec.width, 2);

	    g.setColor(Color.black);
	    g.drawString(str, xMargin, 14);
	}

    }

    Point calculatePopupPosition() {
	Container container;

	if (source == null) {
	    return new Point(0, 0);
	}

	if (source instanceof Container) {
	    container = (Container) source;
	} else {
	    container = source.getParent();
	}

	while (container != null
	       && !(container instanceof java.awt.Window))
	    container = container.getParent();

	if (container == null)
	    return new Point(0, 0);

	Component tc = (Component)container;
	if (!tc.isVisible())
	    return new Point(0, 0);

	int compWidth = tc.getBounds().width;
	int compHeight = tc.getBounds().height;

	Point pointOnScreen = tc.getLocationOnScreen();

	Rectangle rec = win.getBounds();
	int statusHeight = rec.height;
	int statusWidth = rec.width;

	pointOnScreen.translate(0, compHeight);
	int slide = pointOnScreen.y + statusHeight - screenHeight;
	if (slide > 0) {
	    pointOnScreen.translate(0, slide * -1);
	}
	pointOnScreen.translate(compWidth - statusWidth, 0);
	if (pointOnScreen.x < 0) {
	    pointOnScreen.translate(pointOnScreen.x * -1, 0);
	}
	slide = pointOnScreen.x + statusWidth - screenWidth;
	if (slide > 0) {
	    pointOnScreen.translate(slide * -1, 0);
	}

	return pointOnScreen;
    }
}
