/*
 * @(#)IIIMPClient.java	1.6 98/06/29
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.im.iiimp;

import java.awt.event.KeyEvent;
import java.io.IOException;
import java.net.URL;
import java.net.MalformedURLException;
import java.util.Locale;
import java.util.Properties;
import java.util.StringTokenizer;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.security.PrivilegedExceptionAction;
import java.security.PrivilegedActionException;

class IIIMPClient implements IIIMProtocol {

    private static IIIMPClient client = null;

    private boolean connected = false;

    private String userName = null;

    private URL url = null;

    private IIIMPURLConnection uc;

    private IIIMPInputStream in;

    private IIIMPOutputStream out;

    static IIIMPClient getInstance() {
	if (client == null) {
	    client = new IIIMPClient();
	}
	return client;
    }

    private IIIMPClient() {}

    boolean isConnected() {
	return connected;
    }

    synchronized void lostConnect() {
	if (connected == false) {
	    return;
	}
        InputMethod[] im = InputMethod.getInputMethod();
        if (im == null) {
            return;
        }
	for (int i = 0; i < im.length; i++) {
	    im[i].lostConnect();
	}
	connected = false;
    }

    synchronized IIIMPIMValues connectIM() throws IOException {
	// for reconnection...
	if (url != null) {
	    return connectIMImpl(url);
	}

	final String urlAddr = (String) AccessController.doPrivileged(
						new PrivilegedAction() {
	    public Object run() {
		return InputMethod.getProperty("iiimp.server", null);
	    }
	});
	if (urlAddr == null) {
	    throw new IOException("No IIIMP server specified");
	}

	URL tmp;
	try {
	    tmp = (URL) AccessController.doPrivileged(
					new PrivilegedExceptionAction() {
		public Object run() throws MalformedURLException {
		    return new URL(urlAddr);
		}
	    });
	} catch (PrivilegedActionException e) {
	    throw (IOException)e.getException();
	}
	IIIMPIMValues ret = connectIMImpl(tmp);
	if (ret != null) {
	    url = tmp;
	    return ret;
	}

	return null;
    }

    IIIMPIMValues connectIMImpl(URL url) {
	try {
	    uc = (IIIMPURLConnection) url.openConnection();
	    if (uc.isConnected() == false) {
		uc.connect();
		if (uc.isConnected() == false) {
		    throw new IOException();
		}
	    }
	    in = new IIIMPInputStream(this, uc.getInputStream());
	    out = new IIIMPOutputStream(uc.getOutputStream());
	    return openIM();
	} catch (IOException e) {}
	return null;
    }

    synchronized IIIMPIMValues openIM() throws IOException {
	IIIMPIMValues ret = new IIIMPIMValues();

	if (userName == null) {
	    userName = (String) AccessController.doPrivileged(
						    new PrivilegedAction() {
		public Object run() {
		    return System.getProperty("user.name", null);
		}
	    });
	    String hostName = getHostName();
	    if (hostName != null) {
		userName += "@" + hostName;
	    }
	}
	if (userName == null) {
	    throw new IOException();
	}

	// creates a packet for sending IM_CONNECT
	PacketData ps = new PacketData(IM_CONNECT);
	ps.write(BIGENDIAN); // endian
	ps.write(VERSION); // protocol version
	ps.writeString(userName); // user name and host name
	ps.write2(0); // client auth protocol names
	send(ps.toPacket());
	out.flush();

	Packet response = getReply(IM_REGISTER_TRIGGER_KEYS, IM_CONNECT_REPLY);

	// receives IM_REGISTER_TRIGGER_KEYS before IM_CONNECT_REPLY
	if (response.protocol == IM_REGISTER_TRIGGER_KEYS) { 
	    if (response.len == 0) {
		throw new IIIMProtocolException("IM_REGISTER_TRIGGER_KEYS");
	    }

	    // creates a packet for responding IM_REGISTER_TRIGGER_KEYS
	    ps = new PacketData(response);
	    ps.read2(); // skip the input method id
	    ps.skipBytes(2);
	    int num_on_keys = (ps.read4() / 16); // on keys
	    if (num_on_keys > 0) {
		ret.onKey = new IIIMPKeyEvent[num_on_keys];
		for (int i = 0; i < num_on_keys; i++) {
		    int keycode = ps.read4();
		    char keychar = (char) ps.read4();
		    int modifier = ps.read4();
		    int timestamp = ps.read4();
		    ret.onKey[i] =
			new IIIMPKeyEvent(keycode, keychar, modifier);
		}
	    }

	    int num_off_keys = (ps.read4() / 16); // off keys
	    if (num_off_keys > 0) {
		ret.offKey = new IIIMPKeyEvent[num_off_keys];
		for (int i = 0; i < num_off_keys; i++) {
		    int keycode = ps.read4();
		    char keychar = (char) ps.read4();
		    int modifier = ps.read4();
		    int timestamp = ps.read4();
		    ret.offKey[i] = new IIIMPKeyEvent(keycode, keychar,modifier);
		}
	    }

	    ret.dynamic_event_flow = true; 

	    response = getReply(IM_CONNECT_REPLY);
	    if (response.len == 0) {
		throw new IIIMProtocolException("IM_CONNECT_REPLY");
	    }
	} else {
	    ret.dynamic_event_flow = false; 
	}

	// creates a packet for responding IM_CONNECT_REPLY
	ps = new PacketData(response);
	ret.id = ps.read2(); // input method id
	String[] localeNameList = null;
	try {
	    StringData d = new StringData((Data) ps);
	    localeNameList = d.toStringArray();
	} catch (Exception e) {
	    e.printStackTrace();
	}

	if (localeNameList != null) {
	    int max = localeNameList.length;
	    ret.localeList = new Locale[max];
	    for (int i = 0; i < max; i++) {
		ret.localeList[i] =
		    IIIMPInputLanguage.toLocale(localeNameList[i]);
	    }
	}

	connected = true;

	return ret;
    }

    void disconnect() throws IOException {
	if (connected == false) {
	    return;
	}
	if (uc.isConnected() == true) {
	    uc.disconnect();
	}
	connected = false;
    }

    void closeIM(int imID) throws IOException {
	if (connected == false) {
	    return;
	}

	PacketData ps = new PacketData(IM_DISCONNECT);
	ps.write2(imID); // input method id
	ps.write2(0); // padding

	Packet response = request(ps.toPacket(), IM_DISCONNECT_REPLY);
	if (response.len == 0) {
	    throw new IIIMProtocolException("IM_DISCONNECT_REPLY");
	}
	ps = new PacketData(response);
	int ret_id = ps.read2();
	if (imID != ret_id) {
	    throw new IIIMProtocolException("IM_DISCONNECT_REPLY");
	}
    }

    int createIC(int imID, Locale locale) throws IOException {
	if (connected == false) {
	    return -1;
	}

	PacketData ps = new PacketData(IM_CREATEIC);
	ps.write2(imID);
	IIIMPICValues val = new IIIMPICValues(locale);
	byte[] buf = val.getListOfICAttr();
	ps.write2(buf.length);
	ps.writeBytes(buf);

	Packet response = request(ps.toPacket(), IM_CREATEIC_REPLY);
	if (response.len == 0) {
	    throw new IIIMProtocolException("IM_CREATEIC_REPLY");
	}
	ps = new PacketData(response);
	ps.read2(); // imid
	int icID = ps.read2();

	return icID;
    }

    void destroyIC(int imID, int icID) throws IOException {
	if (connected == false) {
	    return;
	}

	PacketData ps = new PacketData(IM_DESTROYIC);
	ps.write2(imID);
	ps.write2(icID);

	Packet response = request(ps.toPacket(), IM_DESTROYIC_REPLY);
	if (response.len == 0) {
	    throw new IIIMProtocolException("IM_DESTROYIC_REPLY");
	}
	ps = new PacketData(response);
	ps.read2(); // imid
	ps.read2(); // icid
    }

    boolean setICValues(int imID, int icID, IIIMPICValues val)
	throws IOException {
	if (connected == false) {
	    return false;
	}

	byte[] buf = val.getListOfICAttr();
	if (buf == null) {
	    // nothing to be set.
	    return true;
	}

	PacketData ps = new PacketData(IM_SETICVALUES);
	ps.write2(imID);
	ps.write2(icID);
	ps.write2(buf.length);
	ps.writeBytes(buf);
	ps.write2(0);

	Packet response = request(ps.toPacket(), IM_SETICVALUES_REPLY);
	if (response.len == 0) {
	    throw new IIIMProtocolException("IM_SETICVALUES_REPLY");
	}
	ps = new PacketData(response);
	ps.read2(); // imid
	ps.read2(); // icid

	return true;
    }

    IIIMPICValues getICValues(int imID, int icID, IIIMPICValues val)
	throws IOException {
	if (connected == false) {
	    return (IIIMPICValues) null;
	}

	PacketData ps = new PacketData(IM_GETICVALUES);
	ps.write2(imID);
	ps.write2(icID);
	ps.writeBytes(val.getListOfICAttrId());

	Packet response = request(ps.toPacket(), IM_GETICVALUES_REPLY);
	if (response.len == 0) {
	    throw new IIIMProtocolException("IM_GETICVALUES_REPLY");
	}

	ps = new PacketData(response);
	ps.read2(); // imid
	ps.read2(); // icid

	int len = ps.read2(); // byte length of attr

	byte[] buf = new byte[len];
	ps.read(buf, 0, len);

	ps.read2(); // pad

	return new IIIMPICValues(buf);
    }

    void resetIC(int imID, int icID) throws IOException {
	if (connected == false) {
	    return;
	}

	PacketData ps = new PacketData(IM_RESETIC);
	ps.write2(imID);
	ps.write2(icID);

	Packet response = request(ps.toPacket(), IM_RESETIC_REPLY);
	if (response.len == 0) {
	    throw new IIIMProtocolException("IM_RESETIC_REPLY");
	}
	ps = new PacketData(response);
	ps.read2(); // imid
	ps.read2(); // icid
    }

    void setFocus(int imID, int icID, boolean mode) throws IOException {
	if (connected == false) {
	    return;
	}

	PacketData ps = new PacketData(mode ? IM_SETICFOCUS : IM_UNSETICFOCUS);
	ps.write2(imID);
	ps.write2(icID);

	Packet response = request(ps.toPacket(),
				  mode ? IM_SETICFOCUS_REPLY : IM_UNSETICFOCUS_REPLY);
	if (response.len == 0) {
	    throw new IIIMProtocolException(
		mode ? "IM_SETICFOCUS_REPLY" : "IM_UNSETICFOCUS_REPLY");
	}
	ps = new PacketData(response);
	ps.read2(); // imid
	ps.read2(); // icid
    }

   void notifyTrigger(int imID, int icID, boolean mode) throws IOException {
	if (connected == false) {
	    return;
	}

	PacketData ps = new PacketData(IM_TRIGGER_NOTIFY);
	ps.write2(imID);
	ps.write2(icID);
	ps.write2((mode == true) ? 0 : 1);

	Packet response = request(ps.toPacket(), IM_TRIGGER_NOTIFY_REPLY);
	if (response.len == 0) {
	    throw new IIIMProtocolException("IM_TRIGGER_NOTIFY_REPLY");
	}
	ps = new PacketData(response);
	ps.read2(); // imid
	ps.read2(); // icid
    }

    void sendAuxData(int imID, int icID,
		     int index, String name, int[] idata, String[] sdata)
	throws IOException {
	if (connected == false) {
	    return;
	}

	PacketData ps = new PacketData(IM_AUX_SETVALUES);
	ps.write2(imID);
	ps.write2(icID);
	ps.write4(index); // auxiliary window class index
	ps.writeString(name); // engine name

	Data d = new Data();
	for (int i = 0; i < idata.length; i++) {
	    d.write4(idata[i]);
	}
	ps.write4(d.count); // byte length of integer value list
	ps.writeBytes(d.buf, d.count); // interger value list

	d = new Data();
	for (int i = 0; i < sdata.length; i++) {
	    d.writeString(sdata[i]);
	}
	ps.write4(d.count); // byte length of string value list
	ps.writeBytes(d.buf, d.count); // string value list

	Packet response = request(ps.toPacket(), IM_AUX_SETVALUES_REPLY);
	if (response.len == 0) {
	    throw new IIIMProtocolException("IM_AUX_SETVALUES_REPLY");
	}
	ps = new PacketData(response);
	ps.read2(); // imid
	ps.read2(); // icid
	int ret_index = ps.read4(); // auxiliary window class index
	String ret_name = ps.readString(); // engine name
    }

    boolean processKeyEvent(int imID, int icID, KeyEvent e) throws IOException{
	if (connected == false) {
	    return false;
	}

	PacketData ps = new PacketData(IM_FORWARD_EVENT);
	ps.write2(imID);
	ps.write2(icID);
	ps.write4(2); // keyevent type is #2
	KeyEventData kd = new KeyEventData(e);
	ps.writeBytes(kd.toByteStream());

	Packet response = request(ps.toPacket(), IM_FORWARD_EVENT_REPLY);
	if (response.len == 0) {
	    throw new IIIMProtocolException("IM_FORWARD_EVENT_REPLY");
	}
	ps = new PacketData(response);
	ps.read2(); // imid
	ps.read2(); // icid

	return true;
    }

    private synchronized Packet request(Packet packet, int expectedReply)
					throws IOException {
	out.writeBasicPacket(packet.protocol, packet.len);
	debugPrint(packet, "client->server");
	out.write(packet.buf, 0, packet.len);
	out.flush();
	return getReply(expectedReply, 0);
    }

    synchronized void send(Packet pac) throws IOException {
	out.writeBasicPacket(pac.protocol, pac.len);
	debugPrint(pac, "client->server");
	out.write(pac.buf, 0, pac.len);
    }

    private Packet getReply(int protocol) throws IOException {
	return getReply(protocol, 0);
    }
    
    private Packet getReply(int proto1, int proto2) throws IOException {
	int opcode;
	int len;
	Packet packet;

	while (true) {
	    opcode = in.readOpCode();
	    len = in.readPacketLength();
	    packet = in.readPacket(opcode, len);
	    debugPrint(packet, "server->client");

	    // If it matches one of the expected reply packet, return it.
	    if ((opcode == proto1) || (opcode == proto2)) {
		break;
	    }

	    // Otherwise, it should be a request from the server.
	    PacketData ps = new PacketData(packet);
	    int imID = ps.read2();
	    int icID = ps.read2();

	    if (IIIMPDispatchEvent.processProtocol(imID, icID, opcode, ps)
		== false) { 
		// Received an unexpected packet
		lostConnect();
		throw new IIIMProtocolException("Unexpected protocol received: "
						+ opcode + " against: " + proto1);
	    }
	}
	return packet;
    }


    private String getHostName() {
	try {
	    java.net.InetAddress hostAddr =
		java.net.InetAddress.getLocalHost();
	    return hostAddr.getHostName();
	} catch (Exception e) {}
	return null;
    }

    void debugPrint(Packet packet, String dir) {
	// IIIMPDebug.on();
	// IIIMPDebug.println(packet, dir);
    }
    
    private static final String IIIMP_PGK_PREFIX = "sun.awt.im";
    private static final String JAVA_PROTOCOL_HANDLER_PKGS = "java.protocol.handler.pkgs";

    static {
	// Add sun.awt.im to the java.protocol.handler.pkgs property
	// if it is not defined yet.
	AccessController.doPrivileged(new PrivilegedAction() {
	    public Object run() {
		Properties defaults = System.getProperties();
		synchronized (defaults) {
		    String pkgs = defaults.getProperty(JAVA_PROTOCOL_HANDLER_PKGS);
		    if (pkgs == null) {
			pkgs = IIIMP_PGK_PREFIX;
		    } else {
			StringTokenizer pkgPrefixIter = new StringTokenizer(pkgs, "|");
			while (pkgPrefixIter.hasMoreTokens()) {
			    if (pkgPrefixIter.nextToken().equals(IIIMP_PGK_PREFIX)) {
				return null;
			    }
			}
			pkgs += "|" + IIIMP_PGK_PREFIX;
		    }
		    defaults.put(JAVA_PROTOCOL_HANDLER_PKGS, pkgs);
		}
		return null;
	    }
	});
    }
}
