/*
 * @(#)InputContext.java	1.3 98/03/23
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.im.iiimp;

import java.awt.event.KeyEvent;
import java.io.IOException;
import java.util.Locale;
import java.util.Vector;

//public 
class InputContext {

    private int id = 0;

    private Locale locale = null;

    private boolean active = false;

    private boolean convMode = false;

    private boolean isForwardEvent = false;

    private boolean connected = false;

    private InputMethod im = null;
    
    private static IIIMPClient client = IIIMPClient.getInstance();

    InputContext() {}

    public static InputContext getInstance(InputMethod im)
	throws LocaleNotSupportedException, IOException {
	return getInstance(im, Locale.getDefault());
    }

    public static InputContext getInstance(InputMethod im, Locale lc)
	throws LocaleNotSupportedException, IOException {
	if (im.isConnected() == false) {
	    if (im.open(lc) == false) {
		if (im.isConnected() == true) {
		    throw new LocaleNotSupportedException();
		}
		throw new IOException();
	    }
	} else {
	    if (im.isLocaleSupported(lc) == false) {
		throw new LocaleNotSupportedException();
	    }
	}
	InputContext ic = new InputContext();
	ic.create(im, lc);
        im.addInputContextHandler(ic);
	return ic;
    }

    public InputMethod getInputMethod() {
	return im;
    }

    private void create(InputMethod im, Locale lc) {
        this.im = im;
	locale = lc;
        active = false;

	try {
	    id = client.createIC(im.getID(), locale);
	} catch (Exception e) {}
        connected = (id == -1) ? false : true;

	addForwardListener(new ForwardPerform());
	addTriggerListener(new TriggerPerform());
    }

    public void dispose() {
        done();
        try {
            client.destroyIC(im.getID(), id);
        } catch (Exception e) {}
        im.removeInputContextHandler(this);
    }

    public boolean setLocale(Locale locale) {
        if (im.isLocaleSupported(locale) == true) {
            try {
                IIIMPICValues val = new IIIMPICValues(locale);
                if ((client.setICValues(im.getID(), id, val)) == false) {
                    return false;
                }
                return true;
            } catch (Exception e) {}
        }
        return false;
    }

    public void activate() {
        if (active == true) {
            return;
        }
        active = true;
        setFocus(true);
    }

    public void deactivate() {
        if (active == false) {
            return;
        }
        active = false;
        setFocus(false);
    }

    private boolean beingProcessed = false;

    public void dispatchKeyEvent(KeyEvent kev) {
	if (beingProcessed && (kev.getID() == KeyEvent.KEY_TYPED ||
			  kev.getID() == KeyEvent.KEY_RELEASED)) {
	    kev.consume();
	    return;
	}
	beingProcessed  = dispatchKeyEventImpl(kev);
	if (beingProcessed == true) {
	    kev.consume();
	}
    }

    private boolean isUninterestingModifiers(KeyEvent kev) {
        switch (kev.getKeyCode()) {
        case KeyEvent.VK_CONTROL:
        case KeyEvent.VK_CAPS_LOCK:
        case KeyEvent.VK_ALT:
        case KeyEvent.VK_META:
        case KeyEvent.VK_SHIFT:
            return true;
        }
        return false;
    }

    private boolean dispatchKeyEventImpl(KeyEvent kev) {
        if (isConversionMode() == true) {
            if (isUninterestingModifiers(kev) == true) {
                return true;
            }
            if (kev.getID() == KeyEvent.KEY_TYPED) {
                return (isForwardEvent == true) ? false : true;
            }
            if (kev.getID() == KeyEvent.KEY_RELEASED) {
                return (isForwardEvent == true) ? false : true;
            }
        } else {
            if (isUninterestingModifiers(kev) == true) {
                return false;
            }
        }

        if (connected == false) {
            if (im.isConversionOnKey(kev) == true) {
                if (im.reopen(locale) == true) {
                    if (reconnect() == true) {
                        activate();
                        setConversionMode(true);
                    }
                } 
                return true;
            }
            return false;
        }
        if (isForwardEvent == true) {
            isForwardEvent = false;
        }

        if (im.isDynamicEventFlow() == true) {
            // conversion on
            if (isConversionMode() == true) {
                // turn off the conversion if the conversion mode is on
                if (im.isConversionOffKey(kev) == true) {
                    setConversionMode(false);
                    return true;
                }
                return deliverKeyEvent(kev);
            }
            // conversion off
            else {
                // turn on the conversion if the conversion mode is off
                if (im.isConversionOnKey(kev) == true) {
                    setConversionMode(true);
                    return true;
                }
            }
            return false;
        }

        return deliverKeyEvent(kev);
    }

    private boolean deliverKeyEvent(KeyEvent kev) {
        try {
            boolean ret = client.processKeyEvent(im.getID(), id, kev);
            if (isForwardEvent == true) {
                ret = false;
            }
            return ret;
        } catch (Exception e) {
            // e.printStackTrace();
        }
        return false;
    }

    void processEvent(IIIMPEvent e) {
	if (e instanceof CommitEvent) {
	    processCommitEvent((CommitEvent) e);
	    return;
	}
	if (e instanceof PreeditEvent) {
	    processPreeditEvent((PreeditEvent) e);
	    return;
	}
	if (e instanceof StatusEvent) {
	    processStatusEvent((StatusEvent) e);
	    return;
	}
	if (e instanceof LookupEvent) {
	    processLookupEvent((LookupEvent) e);
	    return;
	}
	if (e instanceof AuxEvent) {
	    processAuxEvent((AuxEvent) e);
	    return;
	}
	if (e instanceof ForwardEvent) {
	    processForwardEvent((ForwardEvent) e);
	    return;
	}
	if (e instanceof TriggerEvent) {
	    processTriggerEvent((TriggerEvent) e);
	    return;
	}
    }

    CommitListener commitListener = null;
    public void addCommitStringListener(CommitListener l) {
	commitListener = l;
    }
    public void removeCommitStringListener() {
	commitListener = null;
    }
    void processCommitEvent(CommitEvent e) {
	if (commitListener != null) {
	    commitListener.commitPerformed(e);
	}
    }

    int preeditBufferSize = 512;
    int preeditCaretPosition = 0;
    PreeditListener preeditListener = null;
    public void addPreeditListener(PreeditListener l) {
	preeditListener = l;
    }
    public void removePreeditListener() {
	preeditListener = null;
    }
    void processPreeditEvent(PreeditEvent e) {
	if (preeditListener != null) {
	    int id = e.getID();
	    switch (id) {
	    case PreeditEvent.PREEDIT_START:
		preeditBufferSize = preeditListener.preeditStart(e);
		break;
	    case PreeditEvent.PREEDIT_DRAW:
		preeditListener.preeditDraw(e);
		break;
	    case PreeditEvent.PREEDIT_CARET:
		preeditCaretPosition = preeditListener.preeditCaret(e);
		break;
	    case PreeditEvent.PREEDIT_DONE:
		preeditListener.preeditDone(e);
		break;
	    }
	}
    }

    StatusListener statusListener = null;
    public void addStatusListener(StatusListener l) {
	statusListener = l;
    }
    public void removeStatusListener() {
	statusListener = null;
    }
    void processStatusEvent(StatusEvent e) {
	if (statusListener != null) {
	    int id = e.getID();
	    switch (id) {
	    case StatusEvent.STATUS_START:
		statusListener.statusStart(e);
		break;
	    case StatusEvent.STATUS_DRAW:
		statusListener.statusDraw(e);
		break;
	    case StatusEvent.STATUS_DONE:
		statusListener.statusDone(e);
		break;
	    }
	}
    }

    LookupListener lookupListener = null;
    public void addLookupListener(LookupListener l) {
	lookupListener = l;
    }
    public void removeLookupListener() {
	lookupListener = null;
    }
    void processLookupEvent(LookupEvent e) {
	if (lookupListener != null) {
	    int id = e.getID();
	    switch (id) {
	    case LookupEvent.LOOKUP_START:
		lookupListener.lookupStart(e);
		break;
	    case LookupEvent.LOOKUP_DRAW:
		lookupListener.lookupDraw(e);
		break;
	    case LookupEvent.LOOKUP_DONE:
		lookupListener.lookupDone(e);
		break;
	    case LookupEvent.LOOKUP_PROCESS:
		lookupListener.lookupProcess(e);
		break;
	    }
	}
    }

    AuxListener auxListener = null;
    void addAuxListener(AuxListener l) {
	auxListener = l;
    }
    void removeAuxListener() {
	auxListener = null;
    }
    void processAuxEvent(AuxEvent e) {
	if (auxListener != null) {
	    int id = e.getID();
	    switch (id) {
	    case AuxEvent.AUX_START:
		auxListener.auxStart(e);
		break;
	    case AuxEvent.AUX_DRAW:
		auxListener.auxDraw(e);
		break;
	    case AuxEvent.AUX_DONE:
		auxListener.auxDone(e);
		break;
	    }
	}
    }

    ForwardListener forwardListener = null;
    void addForwardListener(ForwardListener l) {
	forwardListener = l;
    }
    void removeForwardListener() {
	forwardListener = null;
    }
    void processForwardEvent(ForwardEvent e) {
	if (forwardListener != null) {
	    forwardListener.forwardPerformed(e);
	}
    }
    class ForwardPerform implements ForwardListener {
	public void forwardPerformed(ForwardEvent e) {
	    isForwardEvent = true;
	}
    }

    TriggerListener triggerListener = null;
    void addTriggerListener(TriggerListener l) {
	triggerListener = l;
    }
    void removeTriggerListener(TriggerListener l) {
	triggerListener = null;
    }
    void processTriggerEvent(TriggerEvent e) {
	if (triggerListener != null) {
	    triggerListener.triggerPerformed(e);
	}
    }
    class TriggerPerform implements TriggerListener {
	public void triggerPerformed(TriggerEvent e) {
	    if (connected == false || isForwardEvent == true) {
		return;
	    }
	    if (isConversionMode() == e.status) {
		return;
	    }
	    convMode = e.status;
	}
    }

    void sendAuxData(int index, String name,
                            int[] idata, String[] sdata) {
        try {
            client.sendAuxData(im.getID(), id, index, name, idata, sdata);
        } catch (Exception e) {}
    }

    public void reset(boolean state_is_preserved) {
        try {
            client.resetIC(im.getID(), id);
        } catch (Exception e) {}
        if (state_is_preserved == false) {
            setConversionMode(false);
        }
    }

    int getID() {
        return id;
    }

    private void setConversionMode(boolean mode) {
        try {
            client.notifyTrigger(im.getID(), id, mode);
        } catch (Exception e) {}
        convMode = mode;
    }

    private void setFocus(boolean mode) {
        try {
            client.setFocus(im.getID(), id, mode);
        } catch (Exception e) {}
    }

    private boolean isConversionMode() {
        return (connected == false) ? false : convMode;
    }

    private boolean reconnect() {
        if (connected == true) {
            return true;
        }
        try {
            id = client.createIC(im.getID(), locale);
        } catch (Exception e) {
            // e.printStackTrace();
            return false;
        }
        connected = true;
        return true;
    }

    private void done() {
        if (connected == false) {
	    return;
	}

	if (lookupListener != null) {
	    LookupEvent e = new LookupEvent(this, LookupEvent.LOOKUP_DONE);
	    lookupListener.lookupDone(e);
	}

	if (preeditListener != null) {
	    PreeditEvent e = new PreeditEvent(this, PreeditEvent.PREEDIT_DONE);
	    preeditListener.preeditDone(e);
	}

	if (commitListener != null) {
	    CommitEvent e = new CommitEvent(this, new String(""));
	    commitListener.commitPerformed(e);
	}

	if (statusListener != null) {
	    StatusEvent e = new StatusEvent(this,
					    StatusEvent.STATUS_DONE,
					    new String(""));
            statusListener.statusDraw(e);
	    e = new StatusEvent(this, StatusEvent.STATUS_DONE);
            statusListener.statusDone(e);
	}

	active = false;
    }

    void lostConnect() {
        if (connected == true) {
            done();
            connected = false;
            id *= -1;
        }
    }
}
