/*
 * @(#)InitialNamingClient.java	1.10 98/09/21
 *
 * Copyright 1998-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.CORBA.idl;

import java.net.*;
import java.util.*;

import org.omg.CORBA.CompletionStatus;
import org.omg.CORBA.DATA_CONVERSION;
import org.omg.CORBA.portable.Delegate;
import org.omg.CORBA.ORBPackage.InvalidName;

import com.sun.CORBA.*;
import com.sun.CORBA.util.*;

/** This class encapsulates all the state and logic for listing and
 *  resolving services that are published in the initial naming (bootstrap)
 *  service. It is only called through list_initial_services() and
 *  resolve_initial_references() in ORB.java.
 */

public class InitialNamingClient 
{
    static private final int defaultInitialServicesPort = 900;

    // State for the initial services: a URL if specified by a URL,
    // otherwise a representation to invoke on,
    // a port and its default value for the endpoint, a list of
    // initial services (a cache), and a list of resolved services (a cache).
    private URL servicesURL = null;
    protected int initialServicesPort = defaultInitialServicesPort;
    private String[] listOfInitialServices;
    private java.util.Properties resolvedInitialReferences = null;

    com.sun.CORBA.idl.ORB orb;



    InitialNamingClient(com.sun.CORBA.idl.ORB orb)
    {
	this.orb = orb;
    }

    void setServicesURL(URL url)
    {
	servicesURL = url;
    }

    void setInitialServicesPort(int port)
    {
	initialServicesPort = port;
    }


    /**
     * Get a list of the initially available CORBA services. 
     * This does not work unless an ORBInitialHost is specified during initialization
     * (or unless there is an ORB running on the AppletHost) since the localhostname 
     * is inaccessible to applets. If a service properties URL was specified,
     * then it is used, otherwise the bootstrapping protocol is used.
     * @return A list of the initial services available.
     */
    public String[] list_initial_services () 
    {
        // Call real function
        return this.cachedServices();
    }

    private Properties getServicesFromURL()
    {
        if ( resolvedInitialReferences == null ) {
            try {
		// Open connection
		URLConnection urlc = servicesURL.openConnection();
		java.io.InputStream is = urlc.getInputStream();

		// Create and load properties
		Properties serviceProps = new Properties();
		serviceProps.load(is);
		resolvedInitialReferences = serviceProps;

		is.close();

            } catch (java.io.IOException ex) {
		throw new org.omg.CORBA.COMM_FAILURE(ex.toString(), 
	  		    MinorCodes.GET_PROPERTIES_ERROR,
				       CompletionStatus.COMPLETED_NO);
            }
        }
        return resolvedInitialReferences;
    }

    private synchronized String[] cachedServices()
    {
        if ( listOfInitialServices == null ) {
            String[] list = null;

            if (( listOfInitialServices==null) && (servicesURL != null)) {
                // A URL was given, so use it.
                Properties serviceProps = getServicesFromURL();

                // Create a list of strings (the keys) and copy entries
                list = new String[serviceProps.size()];
                Enumeration theKeys = serviceProps.keys();
                for (int index=0;theKeys.hasMoreElements();index++) {
                    list[index] = (String)theKeys.nextElement();
                }
            } 
            else {
                list = getInitialServices(orb.getORBInitialHost(),
					  initialServicesPort);
            }
            listOfInitialServices = list;
        }
        return listOfInitialServices;
    }
    
    private String[] getInitialServices(String host, int port)
    {
	// Create a delegate (subcontract) to invoke on
	GenericCORBAClientSC sc = this.getInitialRep(host, port);
	
	// Invoke
        ClientRequest os = sc.createRequest("list", false);
        ClientResponse is = sc.invoke(os);

        // Process reply
        if ( is.isSystemException() )
            throw is.getSystemException();
        else if ( is.isLocationForward() ) // Invoke again.
            return getInitialServices(host, port);
        else { 
	    // normal return: unmarshal String[]
            int length = is.read_long();
            String[] ret = new String[length];
            for (int i=0; i<length; i++) 
                ret[i] = is.read_string();
	    return ret;
	}
    }
    
    private GenericCORBAClientSC getInitialRep(String host, int port)
    {
	// Create a representation (delegate) for the initial name server.

	// Create a new IOR
	byte[] initialKey = new byte[4];
	initialKey[0] = 0x49; initialKey[1] = 0x4e;
	initialKey[2] = 0x49; initialKey[3] = 0x54;
	IOR initialIOR = new IOR(orb, "",host, port,initialKey);

        // Create a subcontract
        GenericCORBAClientSC rep = new GenericCORBAClientSC(initialIOR);
        rep.setId(SubcontractList.defaultSubcontract);
	rep.setOrb(orb);
        return rep;
    }

    /**
     * Resolve the stringified reference of one of the initially
     * available CORBA services.
     * @param identifier The stringified object reference of the
     * desired service.
     * @return An object reference for the desired service.
     * @exception InvalidName The supplied identifier is not associated
     * with a known service.
     * @exception SystemException One of a fixed set of Corba system exceptions.
     */
    public org.omg.CORBA.Object resolve_initial_references(String identifier)
        throws InvalidName
    {
        return cachedInitialReferences(identifier);
    }

    private org.omg.CORBA.Object resolve(String identifier, String host, 
							    int port)
    {
	// URL not specified: do bootstrapping
	GenericCORBAClientSC sc = this.getInitialRep(host, port);
	
	// Invoke
        ClientRequest os = sc.createRequest("get",false);
        os.write_string(identifier);
        ClientResponse is = sc.invoke(os);

        // Process reply
        if ( is.isSystemException() )
            throw is.getSystemException();
        else if ( is.isLocationForward() ) // Invoke again.
            return resolve(identifier, host, port);
        else { 
	    // normal return: unmarshal object
	    return is.read_Object();
	}
    }

    public org.omg.CORBA.Object resolve_initial_references(String identifier, 
                                                            String modifier)
    	throws InvalidName
    {
        int ind = modifier.indexOf(":"); 
        if (ind <= 0)
            throw new DATA_CONVERSION(MinorCodes.BAD_MODIFIER, 
                                    CompletionStatus.COMPLETED_NO);
        try {
            String hostName = modifier.substring(0,ind);
 
            int port = java.lang.Integer.parseInt(modifier.substring(ind+1,
                                                            modifier.length()));
            return resolve_initial_references(identifier, hostName, port);

        } catch (Exception e) {
                throw new DATA_CONVERSION(MinorCodes.BAD_MODIFIER,
                                    CompletionStatus.COMPLETED_NO);
        }        

    }

    public org.omg.CORBA.Object resolve_initial_references(String identifier, 
							String host, int port)
        throws InvalidName
    {
        org.omg.CORBA.Object result = resolve(identifier, host, port);

        if (result == null)
            throw new InvalidName();
        
        return result;
    } 
    
    private synchronized org.omg.CORBA.Object cachedInitialReferences(
							      String identifier)
        throws InvalidName
    {
        org.omg.CORBA.Object result = null;
        String sresult = null;

        // Check cache
        if (resolvedInitialReferences != null)
            sresult = resolvedInitialReferences.getProperty(identifier);
        
        // Found it?
        if (sresult == null) {
            if (servicesURL != null) {

	        // URL was specified: use it
	        if (resolvedInitialReferences == null) {

	            // First time around do a lookup; 
		    // otherwise we could not find it
	            sresult = getServicesFromURL().getProperty(identifier);

		    if (sresult == null || sresult.length() == 0)
	    		throw new InvalidName();
	  
	  	    result = orb.string_to_object(sresult);
		} else {
	  	    // Could not find it 
	  	    throw new InvalidName();
		}
            } else {
		// Use bootstrap protocol
                result = resolve(identifier, orb.getORBInitialHost(), 
				 initialServicesPort);
		if (result == null)
	  	    throw new InvalidName();
		sresult = orb.object_to_string(result);
            }

            if (resolvedInitialReferences == null)
	        resolvedInitialReferences = new java.util.Properties();

            // Add to cached properties
            if (sresult != null && sresult.length() > 0)
	        resolvedInitialReferences.put(identifier,sresult);
        } else {
            // Get from cache
            result = orb.string_to_object(sresult);
        }
        
        return result;
    }
}
