/*
 * @(#)ConnectionTable.java	1.49 98/07/09
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.CORBA.iiop;

import java.io.*;
import java.util.Hashtable;
import java.util.Enumeration;
import java.net.*;

import org.omg.CORBA.COMM_FAILURE;
import org.omg.CORBA.CompletionStatus;

import com.sun.CORBA.ServerGIOP;
import com.sun.CORBA.util.*;

/**
 * One instance of this class is created per ORB object.
 */
public class ConnectionTable {
    protected ORB orb;
    protected Hashtable connectionCache = new Hashtable();

    protected long globalCounter=0;

    private int MAX_SOCKET_RETRIES=5;

    ServerGIOP server;

    public ConnectionTable(ORB orb, ServerGIOP server)
    {
	this.orb = orb;
	this.server = server;
    }

    private void dprint(String msg) {
        Utility.dprint(this, msg);
    }

    public static String createKey(String host, int port) {
        return host + ":" + port;
    }


    /**
     * Get a connection from the connection table, or create a new one.
     * Called only on client side.
     * @param host The hostname for the connection.
     * @param port The port of the connection.
     * @return The connection, pre-existing or new.
     */
    public Connection get(String host, int port)
    {
        String key = createKey(host, port);

        // Try to get a connection from the cache.
	// note: Hashtable.get() is synchronized
        Connection c = (Connection) connectionCache.get(key);
        if (c != null) 
            return c;

        // Create a new connection object. Even though Hashtable.get/put are
        // synchronized, we need to synchronize on "this" to keep the
        // get+put atomic and prevent two threads from creating connections
        // to the same destination at the same time.
        synchronized(this) {
            // Always recheck the entry condition after getting the lock !!
            c = (Connection) connectionCache.get(key);
            if (c != null)
                return c;

            // Create an entry in the connection table
            c = new IIOPConnection(orb, server, this, host, port);
            stampTime(c);
            connectionCache.put(key, c);
        }

        // Try to open a socket to the server
        int tries = 0;
        while (true)
        {
	    try {
	        Socket socket = Connection.newSocket(orb, host, port);
                c.setConnection(socket, this);
                checkConnectionTable();
                break;
            }
	    catch (SocketException ex) {
                if ( ex instanceof BindException
		     || ex instanceof ConnectException
		     || ex instanceof NoRouteToHostException )
		{
                    c.abortConnection();
                    throw new COMM_FAILURE(MinorCodes.CONNECT_FAILURE,
         	       			   CompletionStatus.COMPLETED_NO);
		}

		// Probably the system's file descriptor limit has been reached
		// Try cleaning up some connections and retry.
                if ((tries == MAX_SOCKET_RETRIES) || (!cleanUp()))
                {
                    c.abortConnection();
                    throw new COMM_FAILURE(MinorCodes.CONNECT_FAILURE,
                		     	   CompletionStatus.COMPLETED_NO);
                }
                tries++;
	    }
	    catch ( Exception ex ) {
                c.abortConnection();
                throw new COMM_FAILURE(MinorCodes.CONNECT_FAILURE,
                		       CompletionStatus.COMPLETED_NO);
	    }
        }

        return c;
    }

    /**
     * Called only on server side when a connection has been accepted.
     * Always creates a new Connection instance, which starts the upcall.
     * @param sock The socket for the connection.
     */
    public synchronized Connection get(Socket sock)
    {
        try {
	    InputStream in = null;
	    OutputStream out = null;
	    try {
		in = sock.getInputStream();
		out = sock.getOutputStream();
	    } catch ( Exception ex ) {
		throw new COMM_FAILURE(MinorCodes.CONNECT_FAILURE,
				       CompletionStatus.COMPLETED_NO);
	    }

	    String host = sock.getInetAddress().getHostName();
	    int port = sock.getPort();

	    // Create a new Connection instance, and start the upcall.
	    // We explicitly give the input/output streams to IIOPConnection
	    // so that there is the flexibility of substituting the streams
	    // e.g. for handling multiple protocols later.
	    Connection c = new IIOPConnection(orb, server, host, port, sock, 
					      in, out, this);

	    String key = ConnectionTable.createKey(host, port);
	    connectionCache.put(key, c);

	    stampTime(c);

	    return c;

        } catch ( Exception ex ) {
	    // We should not throw an exception back to the ListenerThread.
	    // Just close the socket and return silently.
	    try {
		sock.close();
	    } catch ( Exception ex2 ) {}
	    return null;
        }
    }


    /**
     * Delete a connection from the connection table.
     */
    public void deleteConn(String host, int port)
    {
        String key = createKey(host, port);
	connectionCache.remove(key);
    }

    /**
     * CleanUp by discarding least recently used Connections that
     * are not busy
     */
    private static final int LOW_WATER_MARK = 100;
    // On Solaris there seems to be a max of 256 connections per process.
    private static final int HIGH_WATER_MARK = 240;
    // How many connections to clean at a time.
    private static final int NCLEAN = 5;

    public boolean cleanUp()
    {
	if (connectionCache.size() < LOW_WATER_MARK)
	    return false;

	for ( int i=0; i<NCLEAN; i++ ) {
	    Connection toClean = null;
	    long lru = java.lang.Long.MAX_VALUE;
	    Enumeration e = connectionCache.elements();

	    // Find least recently used and not busy connection in cache
	    while ( e.hasMoreElements() )
	    {
		Connection c = (Connection) e.nextElement();
		if ( !c.isBusy() && c.timeStamp < lru )
		{
		    toClean = c; 
		    lru = c.timeStamp;
		}
	    }

	    if ( toClean == null )
		return false;

	    // Clean connection
	    try {
		toClean.cleanUp();
	    } catch (Exception ex) {}
	}

	// XXX is necessary to do a GC to reclaim closed network connections ??
        // java.lang.System.gc();

	return true;
    }

   /**
    * Check if number of connections has exceeded high watermark
    */
    public void checkConnectionTable()
    {
	// note: Hashtable.size() is not synchronized
        if (connectionCache.size() > HIGH_WATER_MARK)
            cleanUp();
    }


    public synchronized void stampTime(Connection c)
    {
	// XXX Need to worry about wrap around some day
        c.timeStamp = globalCounter++;
    }


    public synchronized void print()
    {
        System.out.println("***ConnectionTable***");
	int siz = connectionCache.size();
        System.out.println("  SIZE=" + siz);
	if ( siz < 10 ) {
            Enumeration e = connectionCache.elements(); 
            while ( e.hasMoreElements() )
            {
                Connection c = (Connection) e.nextElement();
                c.print();
            }
	}
    }

}
