/*
 * @(#)IIOPConnection.java	1.59 98/07/09
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.CORBA.iiop;

import java.util.Hashtable;
import java.util.Enumeration;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.InetAddress;
import java.net.SocketException;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import org.omg.CORBA.SystemException;
import org.omg.CORBA.CompletionStatus;
import org.omg.CORBA.COMM_FAILURE;
import org.omg.CORBA.INTERNAL;
import org.omg.CORBA.OBJECT_NOT_EXIST;
import org.omg.CORBA.MARSHAL;
import org.omg.CORBA.Object;

import com.sun.CORBA.ServerGIOP;

import com.sun.CORBA.util.Condition;
import com.sun.CORBA.util.Lock;
import com.sun.CORBA.util.MinorCodes;



final class OutCallDesc
{
     Condition done = new Condition();
     Thread thd;
     SystemException exc;
     IIOPInputStream s;
}

/**
 * A thread class to handle reading incoming messages off the wire
 */
final class ReaderThread extends Thread
{
    protected IIOPConnection c;

    public ReaderThread(ThreadGroup g, IIOPConnection c, String host,
      int port)
    {
	super(g, "JavaIDL Reader for " + ConnectionTable.createKey(host, port));
        this.c = c;
    }

    public ReaderThread(IIOPConnection c, String host, int port)
    {
	super("JavaIDL Reader for " + ConnectionTable.createKey(host, port));
        this.c = c;
    }

    /**
     */
    public void run() {
        for (;;) {
            try {
	        // Create a new IIOPInputStream buffer to hold each incoming 
		// IIOP message, capture the entire message, and process it.

	        IIOPInputStream s = null;

                try {
		    // creates and prefills the mb from wire
	            s = c.createInputStream();
                } 
		catch (IOException ex) {
		    // Close the connection, inform all threads 
		    // and stop this thread.
                    c.purge_calls(Connection.CONN_ABORT, true, false);
                }
	        c.processInput(s);
            } 
	    catch (DeleteConn dc) {
		// Close the connection, inform all threads 
		// and stop this thread.
                c.purge_calls (dc.minorCode, true, false);
            } 
	    catch (ThreadDeath td) {
                try {
		    // Close the connection and inform all threads.
                    c.purge_calls(Connection.CONN_ABORT, false, false);
                } finally {
                    throw td;
                }   
            } 
	    catch (Exception ex) {
	        break;
            }
        }
    }
}

final class DeleteConn extends java.lang.Throwable {
    int minorCode;

    DeleteConn (int code) {
        minorCode = code;
    }
}




/** 
 * A network connection which processes IIOP messages.
 */

public final class IIOPConnection extends Connection {
    //
    // Connection status
    //
    private static final int OPENING = 1;
    private static final int ESTABLISHED = 2;
    private static final int CLOSE_SENT = 3;
    private static final int CLOSE_RECVD = 4;
    private static final int ABORT = 5;

    //
    // Table of pending invocations on this connection indexed by
    // Integer(requestId).
    //
    private Hashtable out_calls = new Hashtable();

    //
    // Remote address that we're talking to.
    //
    protected String remote_host;
    protected int remote_port;

    protected int requestCount = 0;

    private ServerGIOP server;

    //
    // Various connection state.
    //
    Thread reader;
    int state;
    private Condition event;
    private Lock lock;
    private boolean writeLocked;

    // These I/O streams are the ONLY ONES that should be used.
    // i.e. Do not directly use the input/output streams that socket gives.
    // This restriction is to allow connections to service multiple
    // protocols (e.g. http for tunneling), which may require the
    // socket's streams to be wrapped in other streams.
    InputStream inputStream;
    OutputStream outputStream;


    /**
     * Called after client creates a connection to server 
     * or after server accepts an incoming connection.
     * @param host The remote host pointed to by this connection.
     * @param port The remote port used by this connection.
     */
    public IIOPConnection(ORB orb, ServerGIOP server,
			  ConnectionTable ctab, String host, int port)
    {
	this.orb = orb;
	this.server = server;
	this.connectionTable = ctab;

        try {
            reader = new ReaderThread(orb.threadGroup, this, host, port);
        } catch (SecurityException e) {
            //
            // For some reason we're not allowed to create a new thread
            // in the same thread group that the ORB was initialized in.
            // Fall back on creating the thread in the calling thread's
            // group.
            //
            reader = new ReaderThread(this, host, port);
        }

	remote_host = host;
	remote_port = port;
	state = OPENING;
	event = new Condition();
	lock = new Lock();
    }

    /** 
     * Called only from ConnectionTable.get() after server accepts an 
     * incoming connection.
     * @param sock The socket for this connection. 
     * @param inputStream The inputstream to use. It may be different
		from a socket's inputstream.
     * @param outputStream The outputstream to use. It may be different
		from a socket's outputstream.
     */
    public IIOPConnection(ORB orb, ServerGIOP server,
			  String host, int port, Socket sock,
			  InputStream inputStream, OutputStream outputStream,
                          ConnectionTable ct) 
    {
	this(orb, server, ct, host, port);

	this.socket = sock;
	this.inputStream = inputStream;
	this.outputStream = outputStream;
        this.connectionTable = ct;

        isServer = true;

	state = ESTABLISHED;

        // Catch exceptions since setDaemon can cause a
        // security exception to be thrown under netscape
        // in the Applet mode
        try {
	    reader.setDaemon(true);
        } catch (Exception e) {}
	reader.start();
    }


    public java.io.InputStream getInputStream() {
	return inputStream;
    }

    ServerGIOP getServerGIOP() {
	return server;
    }

    String getHost() {
	return remote_host;
    }

    int getPort() {
	return remote_port;
    }

    

    /**
     * Read in the IIOP message from the network's InputStream and
     * create an IIOPInputStream object.
     * Called from ReaderThread only.
     */
    public IIOPInputStream createInputStream() throws Exception 
    {
	byte[] buf = new byte[1024];

	// Read in the message header and create the appropriate type
	// of IIOP message.
	Message msg = Message.createFromStream(buf, inputStream);

	// Create the stream object and read in the entire message.
	switch (msg.getType()) {
	    case Message.GIOPRequest:
		return new ServerRequestImpl(this, buf, (RequestMessage)msg);

	    case Message.GIOPReply:
		return new ClientResponseImpl(this, buf, (ReplyMessage)msg);

	    case Message.GIOPLocateRequest:
	    case Message.GIOPLocateReply:
	    case Message.GIOPCancelRequest:
	    case Message.GIOPCloseConnection:
	    case Message.GIOPMessageError:
		return new IIOPInputStream(this, buf, msg);

	    default:
		throw new INTERNAL(MinorCodes.BAD_GIOP_REQUEST_TYPE,
				   CompletionStatus.COMPLETED_NO);
	}
    }


    /** 
     * Process the IIOP message contained in the IIOPInputStream object.
     * Called from ReaderThread only.
     */
    public void processInput(IIOPInputStream s) 
        throws DeleteConn
    {
	int msgType = s.getMessage().getType();

	switch (msgType) {

	    case Message.GIOPLocateRequest: 
	    case Message.GIOPRequest: {
		if (debug)
		    dprint("Connection:processInput: got incoming request");

		//
		// Create a worker thread and start the upcall.
		//
		WorkerThread wt = new WorkerThread(server.getRequestHandler(), 
						   this, s);
     
		// Catch exceptions since setDaemon can cause a   
		// security exception to be thrown under netscape 
		// in the Applet mode 
		try {
		    wt.setDaemon(false);
		}
		catch (Exception e) {}

		wt.start();

		break;
	    }

	    case Message.GIOPReply: 
	    case Message.GIOPLocateReply: {
		if (debug)
		    dprint("Connection:processInput: got RESPONSE");

		// Remove this call from the list of outstanding calls
		Integer index = new Integer(s.getMessage().getRequestId());
		OutCallDesc call = (OutCallDesc) out_calls.get(index);
		out_calls.remove(index);
		
		// Set the reply IIOPInputStream and signal the client thread 
		// that the reply has been received.
		lock.lock();
		call.s = s;
		call.done.signal();
		lock.unlock();

		break;
	    }

	    case Message.GIOPCancelRequest: {
		if (debug)
		    dprint("Connection:processInput: got cancel");

		//
		// Do nothing for now (actually legal IIOP).
		//
		break;
	    }

	    case Message.GIOPCloseConnection: {
		//
		// For any outstanding requests on the connection, arrange
		// for the sending thread to receive a rebind exception.
		//
		if (debug)
		    dprint("Connection.processInput: got CloseConn, purging");

		purge_calls(Connection.CONN_REBIND, true, false);

		break;
	    }

	    case Message.GIOPMessageError:
	    default: {
		//
		// Any other message type is unexpected and is a fatal
		// error for this connection.
		//
		if (debug)
		    dprint("Connection: bad message type " 
		      				+ String.valueOf(msgType));
		purge_calls(Connection.CONN_ABORT, true, false);
	    }
	}
    }



    /**
     * Wake up the outstanding requests on the connection, and hand them
     * COMM_FAILURE exception with a given minor code. Also, delete connection
     * from connection table and stop the reader thread. Note that this should only
     * ever be called by the Reader thread for this connection.
     * Note that this may be called more than once for the same connection.
     *
     * @param minor_code The minor code for the COMM_FAILURE major code.
     * @param die Kill the reader thread (this thread) before exiting.
     */
    void purge_calls(int minor_code, boolean die, boolean lockHeld) 
    {
        OutCallDesc call;

	if (debug)
            dprint("purge_calls: starting: code = " + minor_code 
						    + " die = " + die);

	//
        // If this invocation is a result of ThreadDeath caused
        // by a previous execution of this routine, just exit.
	//
        if ((state == ABORT) || (state == CLOSE_RECVD)) {
	    if (debug)
                dprint("purge_calls: exiting duplicate invocation");
            return;
        }

        //
        // Grab the writeLock (freeze the calls)
        //
        try {
            if (!lockHeld)
	        writeLock();
        } catch (SystemException ex) {
	    if (debug)
	        dprint("purge_calls: caught exception " + ex + "; continuing");
        }

        //
        // Mark the state of the connection and determine the request status
        //
        org.omg.CORBA.CompletionStatus completion_status;

        lock.lock();
        if (minor_code ==  Connection.CONN_REBIND) {
	    state = CLOSE_RECVD;
	    completion_status = CompletionStatus.COMPLETED_NO;
        } else {
	    state = ABORT;
	    completion_status = CompletionStatus.COMPLETED_MAYBE;
        }
        lock.unlock();

        //
        // Close the socket (if its not already closed)
        //
        try {
	    // if theres no socket/connection, this is just ignored
	    inputStream.close();
	    outputStream.close();
	    socket.close();  
        } catch (Exception ex) {
        }

        SystemException comm_failure_exc = 
			new COMM_FAILURE(minor_code, completion_status);

	// Signal all threads with outstanding requests on this
	// connection and give them the COMM_FAILURE exception.
        java.util.Enumeration e = out_calls.elements();
        while(e.hasMoreElements()) {
	    call = (OutCallDesc) e.nextElement();

	    lock.lock();
	    call.exc = comm_failure_exc;
	    call.done.signal();
	    lock.unlock();
        }


        //  clear out the OutCallDesc table
        out_calls.clear ();

        //
        // delete connection from cache and stop the reader thread
        //
        connectionTable.deleteConn(remote_host, remote_port);

        //
        // Signal all the waiters of the writeLock.
	// There are 4 types of writeLock waiters:
	// 1. Send waiters: 
	// 2. SendReply waiters: 
	// 3. cleanUp waiters:
	// 4. purge_call waiters:
	//
        writeUnlock(); 

	// Stop the current thread if needed.
        if (die) {
            Thread.currentThread().stop();
        }
    }


    /** 
     * Sets up an established connection
     */
    public void setConnection(Socket _socket, ConnectionTable ctab) 
        throws Exception 
    {
   	socket = _socket;
        inputStream = socket.getInputStream();
	outputStream = socket.getOutputStream();
        connectionTable = ctab;

        lock.lock();

        state = ESTABLISHED;
        event.signalAll();

        // Catch exceptions since setDaemon can cause a
        // security exception to be thrown under netscape
        // in the Applet mode
        try {
            reader.setDaemon(true);
        }
        catch (Exception e) {}

        reader.start();

        lock.unlock();
    }

    /** 
     * Changes state of connection to aborted, notifying waiters
     */
    public void abortConnection()
    {
        lock.lock();

        state = ABORT;
        event.signalAll();

        lock.unlock();
    }

    /**
     * Sets the writeLock for this connection.
     * If the writeLock is already set by someone else, block till the
     * writeLock is released and can set by us.
     * IMPORTANT: this connection's lock must be acquired before 
     * setting the writeLock and must be unlocked after setting the writeLock.
     */
    private boolean writeLock() 
    {
	lock.lock();

	// Keep looping till we can set the writeLock.
	while ( true ) { 

	    switch ( state ) {

	        case OPENING:
		    // Wait for someone else to change the state to ESTABLISHED
		    event.wait(lock);

		    // Loop back 
		    break;

	        case ESTABLISHED:
		    if (!writeLocked) {
			writeLocked = true;
			lock.unlock();
			return true;
		    }

		    // Wait for the connection lock.
		    event.wait(lock);

		    // Loop back 
		    break;

                //
                // XXX
                // Need to distinguish between client and server roles
                // here probably.
                //
                case ABORT:
                case CLOSE_RECVD:
		    // the connection has been closed or closing
		    // ==> throw rebind exception
		    lock.unlock();
		    throw new COMM_FAILURE( MinorCodes.CONN_CLOSE_REBIND,
		      CompletionStatus.COMPLETED_NO);

	        default:
		    if (debug)
			dprint("Connection:writeLock: weird state");
		    lock.unlock();
		    delete(Connection.CONN_ABORT);
		    return false;
	    }
	}
    }

    /**
     * Release the write lock on this connection.
     */
    private void writeUnlock() 
    {
	lock.lock();
	writeLocked = false;
	event.signalAll();
	lock.unlock();
    }


    public void delete()
    {
        delete(Connection.CONN_ABORT);
    }

    void delete(int code)
    {
        DeleteConn dc = new DeleteConn(code);
        reader.stop(dc);
    }


    /** Send a two-way IIOP message to the server. 
     */
    public IIOPInputStream invoke(IIOPOutputStream s)
        throws SystemException
    {
        return send(s, false);
    }


    /** Send an IIOP message to the server. 
     *  If not oneway, wait for the reply.
     */
    public IIOPInputStream send(IIOPOutputStream s, boolean oneWay)
    {
        OutCallDesc call;
        int id;
	boolean unlock = false;

	call = new OutCallDesc();
        call.thd = Thread.currentThread();
	id = s.getMessage().getRequestId();

	try {
	    writeLock();
	    unlock = true;
            out_calls.put(new Integer(id), call);
	    try {
	        s.writeTo(outputStream);
	        outputStream.flush();
            } catch (java.io.IOException exc) {
                throw new COMM_FAILURE(MinorCodes.WRITE_ERROR_SEND,
                          CompletionStatus.COMPLETED_NO);
            }

	    writeUnlock();
	    unlock = false;

	    if (oneWay)
		return null;

	    lock.lock();
            while (call.s == null && call.exc == null) {
		// Wait for the reply from the server.
		// The ReaderThread reads in the reply IIOP message 
		// and signals us.
	        call.done.wait(lock);
            }
	    lock.unlock();

	    if (call.exc != null)
		throw call.exc;
	}
	finally {
	    if (unlock)
	        writeUnlock();
	}

	return call.s;
    }

    public void sendReply(IIOPOutputStream os)
	throws Exception
    {
	boolean unlock = false;

	try {
            writeLock();
            unlock = true;
            os.writeTo(outputStream);
            outputStream.flush();
            writeUnlock();
            unlock = false;
	} catch ( Exception e ) {
	    if (unlock)
                writeUnlock();
	}
    }



    /** public IOR locate(byte[] key) is now in Connection.java */





/***************************************************************************
    The following methods are for dealing with Connection cleaning for
    better scalability of servers in high network load conditions.
 ***************************************************************************/

    /** 
     * Sends CloseConnection packet assumes write lock already held
     */
    private void sendCloseConnection()
    {
        Message msg = new Message();
        msg.setType(Message.GIOPCloseConnection);
        IIOPOutputStream os;
        IIOPInputStream is;

        os = orb.newOutputStream(this);
        msg.write(os);
        os.setMessage(msg);

        try {
            os.writeTo(outputStream);
            outputStream.flush();
        } catch (java.io.IOException exc) {}
    }


    public boolean isBusy()
    {
	// Note: Hashtable.size() is not synchronized
        if (requestCount > 0 || out_calls.size() > 0)
            return true;
	else
	    return false;
    }


   /**
    * Cleans up this Connection.
    * This is called from ConnectionTable, from the ListenerThread.
    * Note:it is possible for this to be called more than once 
    */
    public synchronized void cleanUp() throws Exception
    {
        writeLock();

        if (requestCount > 0 || out_calls.size() > 0) { // we are busy!
	    writeUnlock();
            throw new Exception();
	}

        try {
            if (isServer) {
                sendCloseConnection();
                lock.lock();
                state = CLOSE_SENT;
                lock.unlock();
            }

	    // stop the reader without causing it to do purge_calls
            Exception ex = new Exception();
            reader.stop(ex); 

	    // this also does writeUnlock();
            purge_calls(Connection.CONN_REBIND, false, true);

        } catch (Exception ex) {}
    }

    /** It is possible for a Close Connection to have been
     ** sent here, but we will not check for this. A "lazy"
     ** Exception will be thrown in the Worker thread after the
     ** incoming request has been processed even though the connection
     ** is closed before the request is processed. This is o.k because
     ** it is a boundary condition. To prevent it we would have to add
     ** more locks which would reduce performance in the normal case.
     **/
    public synchronized void requestBegins()
    {
        requestCount++;
    }

    public synchronized void requestEnds()
    {
        requestCount--;
    }

    public void print()
    {
        System.out.println("Connection for " + remote_host + " @ " + remote_port);
        System.out.println("    Time stamp = " + timeStamp);
        boolean alive = reader.isAlive();
        if (alive)
            System.out.println(" Reader is Alive");
        else
            System.out.println(" Reader is not Alive");
    }
}
