/*
 * @(#)ClassDoc.java	1.6 98/07/07
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.javadoc;


/**
 * Represents a java class and provides access to information
 * about the class, the class' comment and tags, and the
 * members of the class.  A ClassDoc only exists if it was
 * processed in this run of javadoc.  References to classes
 * which may or may not have been processed in this run are
 * referred to using Type (which can be converted to ClassDoc,
 * if possible).
 *
 * @see Type
 *
 * @since JDK1.2
 * @author Kaiyang Liu (original)
 * @author Robert Field (rewrite)
 */
public interface ClassDoc extends ProgramElementDoc, Type {

    /**
     * Return true if this class is abstract
     */
    boolean isAbstract();

    /**
     * Return true if this class implements <code>java.io.Serializable</code>.
     *
     * Since <code>java.io.Externalizable</code> extends
     * <code>java.io.Serializable</code>,
     * Externalizable objects are also Serializable.
     */
    boolean isSerializable();

    /**
     * Return true if this class implements
     * <code>java.io.Externalizable</code>.
     */
    boolean isExternalizable();

    /**
     * Return the serialization methods for this class.
     *
     * @return an array of <code>MethodDoc</code> that represents
     * the serialization methods for this class.
     */
    MethodDoc[] serializationMethods();

    /**
     * Return the Serializable fields of class.<p>
     *
     * Return either a list of default fields documented by
     * <code>serial</code> tag<br>
     * or return a single <code>FieldDoc</code> for
     * <code>serialPersistentField</code> member.
     * There should be a <code>serialField</code> tag for
     * each Serializable field defined by an <code>ObjectStreamField</code>
     * array component of <code>serialPersistentField</code>.
     *
     * @returns an array of <code>FieldDoc</code> for the Serializable fields
     * of this class.
     *
     * @see #definesSerializableFields()
     * @see SerialFieldTag
     */
    FieldDoc[] serializableFields();

    /**
     *  Return true if Serializable fields are explicitly defined with
     *  the special class member <code>serialPersistentFields</code>.
     *
     * @see #serializableFields()
     * @see SerialFieldTag
     */
    boolean definesSerializableFields();

    /**
     * Return the superclass of this class
     *
     * @return the ClassDoc for the superclass of this class, null if
     * there is no superclass.
     */
    ClassDoc superclass();

    /**
     * Test whether this class is a subclass of the specified class.
     *
     * @param cd the candidate superclass.
     * @return true if cd is a superclass of this class.
     */
    boolean subclassOf(ClassDoc cd);

    /**
     * Return interfaces implemented by this class or interfaces extended
     * by this interface. Includes only directly declared interfaces, not
     * inherited interfaces.
     *
     * @return An array of ClassDoc representing the interfaces.
     * Return an empty array if there are no interfaces.
     */
    ClassDoc[] interfaces();

    /**
     * Return fields in class.
     *
     * @return an array of FieldDoc for representing the visible
     * fields in this class.
     */
    FieldDoc[] fields();

    /**
     * Return methods in class.
     *
     * @return an array of MethodDoc for representing the visible
     * methods in this class.  Does not include constructors.
     */
    MethodDoc[] methods();

    /**
     * Return constructors in class.
     *
     * @return an array of ConstructorDoc for representing the visible
     * constructors in this class.
     */
    ConstructorDoc[] constructors();

    /**
     * Return inner classes within this class.
     *
     * @return an array of ClassDoc for representing the visible
     * classes defined in this class. Anonymous and local classes
     * are not included.
     */
    ClassDoc[] innerClasses();

    /**
     * Find a class within the context of this class.
     * Search order: qualified name, in this class (inner),
     * in this package, in the class imports, in the package
     * imports.
     * Return the ClassDoc if found, null if not found.
     */
    ClassDoc findClass(String className);

    /**
     * Get the list of classes declared as imported.
     * These are called "single-type-import declarations" in the JLS.
     *
     * @return an array of ClassDoc representing the imported classes.
     */
    ClassDoc[] importedClasses();

    /**
     * Get the list of packages declared as imported.
     * These are called "type-import-on-demand declarations" in the JLS.
     *
     * @return an array of PackageDoc representing the imported packages.
     */
    PackageDoc[] importedPackages();
}


