/*
 * @(#)Doc.java	1.4 98/07/09
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.javadoc;

/**
 * abstract base class of all Doc classes.  Doc item's are representations
 * of java language constructs (class, package, method,...) which have
 * comments and have been processed by this run of javadoc.  All Doc items
 * are unique, that is, they are == comparable.
 *
 * @since JDK1.2
 * @author Robert Field
 */
public interface Doc extends java.io.Serializable, Comparable {

    /**
     * Return the text of the comment for this doc item.
     * Tags have been removed.
     */
    String commentText();

    /**
     * Return all tags in this Doc item.
     *
     * @return an array of Tag containing all tags on this Doc item.
     */
    Tag[] tags();

    /**
     * Return tags of the specified kind in this Doc item.
     *
     * @param tagname name of the tag kind to search for.
     * @return an array of Tag containing all tags whose 'kind()'
     * matches 'tagname'.
     */
    Tag[] tags(String tagname);

    /**
     * Return the see also tags in this Doc item.
     *
     * @return an array of SeeTag containing all &#64see tags.
     */
    SeeTag[] seeTags();

    /**
     * Return comment as tags. Includes inline tags 
     * (i.e. {&#64link <i>reference</i>} tags)  but not
     * regular tags.  
     * Each section of plain text is represented as a {@link Tag} 
     * of kind "Text".
     * Inline tags are represented as a {@link SeeTag} of kind "link".
     *
     * @return an array of {@link Tag}s representing the comment
     */
    Tag[] inlineTags();

    /**
     * Return the first sentence of the comment as tags. 
     * Includes inline tags 
     * (i.e. {&#64link <i>reference</i>} tags)  but not
     * regular tags.  
     * Each section of plain text is represented as a {@link Tag} 
     * of kind "Text".
     * Inline tags are represented as a {@link SeeTag} of kind "link".
     * First sentence is determined by 
     * {@link java.text.BreakIterator#getSentenceInstance(Locale)}.
     *
     * @return an array of {@link Tag}s representing the 
     * first sentence of the comment
     */
    Tag[] firstSentenceTags();

    /**
     * Return the full unprocessed text of the comment.  Tags
     * are included as text.  Used mainly for store and retrieve
     * operations like internalization.
     */
    String getRawCommentText();

    /**
     * Set the full unprocessed text of the comment.  Tags
     * are included as text.  Used mainly for store and retrieve
     * operations like internalization.
     */
    void setRawCommentText(String rawDocumentation);

    /**
     * Returns the name of this Doc item.
     *
     * @return  the name
     */
    String name();

    /**
     * Compares this Object with the specified Object for order.  Returns a
     * negative integer, zero, or a positive integer as this Object is less
     * than, equal to, or greater than the given Object.
     * <p>
     * Included so that Doc item are java.lang.Comparable.
     *
     * @param   o the <code>Object</code> to be compared.
     * @return  a negative integer, zero, or a positive integer as this Object
     *		is less than, equal to, or greater than the given Object.
     * @exception ClassCastException the specified Object's type prevents it
     *		  from being compared to this Object.
     */
    int compareTo(Object obj);

    /**
     * Is this Doc item a field?  False until overridden.
     *
     * @return true if it represents a field
     */
    boolean isField();

    /**
     * Is this Doc item a simple method (i.e. not a constructor)?  False
     * until overridden.
     *
     * @return true is it represents a method, false if it is anything
     * else, including constructor, field, class, ...
     */
    boolean isMethod();
    /**
     * Is this Doc item a constructor?  False until overridden.
     *
     * @return true is it represents a constructor
     */
    boolean isConstructor();

    /**
     * Is this Doc item a interface?  False until overridden.
     *
     * @return true is it represents a interface
     */
    boolean isInterface();

    /**
     * Is this Doc item a exception class?  False until overridden.
     *
     * @return true is it represents a exception
     */
    boolean isException();

    /**
     * Is this Doc item a error class?  False until overridden.
     *
     * @return true is it represents a error
     */
    boolean isError();

    /**
     * Is this Doc item a ordinary class (i.e. not an interface, exception,
     * or error)?  False until overridden.
     *
     * @return true is it represents a class
     */
    boolean isOrdinaryClass();

    /**
     * Is this Doc item a class.
     * Note: interfaces are not classes.
     * False until overridden.
     *
     * @return true is it represents a class
     */
    boolean isClass();

    /**
     * return true if this Doc is include in the active set.
     */
    boolean isIncluded();
}

