/*
 * @(#)ClassDocImpl.java	1.31 98/09/08
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.tools.javadoc;

import com.sun.javadoc.*;

import sun.tools.java.ClassDeclaration;
import sun.tools.java.ClassDefinition;
import sun.tools.java.ClassNotFound;
import sun.tools.java.Environment;
import sun.tools.java.Identifier;
import sun.tools.java.IdentifierToken;
import sun.tools.java.Imports;
import sun.tools.java.MemberDefinition;
import sun.tools.javac.SourceClass;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import java.text.CollationKey;
import java.text.Collator;

import java.lang.reflect.Modifier;


/**
 * Represents a java class and provides access to information
 * about the class, the class' comment and tags, and the 
 * members of the class.  A ClassDocImpl only exists if it was
 * processed in this run of javadoc.  References to classes 
 * which may or may not have been processed in this run are
 * referred to using Type (which can be converted to ClassDocImpl,
 * if possible).
 *
 * @see Type
 *
 * @since JDK1.2
 * @author Robert Field
 */
class ClassDocImpl extends ProgramElementDocImpl implements ClassDoc {

    /**
     *  Map class names to ClassDocImpl, so we can maintain uniqueness.
     */
    private static Map classMap = new HashMap();

    /**
     * all inner classes information.
     */

    private static final int INTERFACE = 1;
    private static final int ORDINARY_CLASS = 2;
    private static final int EXCEPTION = 3;
    private static final int ERROR = 4;

    private /*final*/ int kind;

    final String qualifier;
    final String name;

    final ClassDocImpl superclass;
    final PackageDocImpl containingPackage;
    final List fields = new ArrayList();
    final List methods = new ArrayList();
    final List constructors = new ArrayList();
    final List innerClasses = new ArrayList();
    ClassDocImpl[] interfaces = new ClassDocImpl[0];
    final List importedClasses = new ArrayList();
    final List importedPackages = new ArrayList();
    boolean /*final*/ isIncluded;
    private SerializedForm serializedForm;

    // bit masks for serializationKind.    
    private static final int NOT_SERIALIZABLE = 1;  
    private static final int SERIALIZABLE = 2;
    private static final int EXTERNALIZABLE = 4;
    private int serializationKind;     

    /* second phase constructor assist.
     * Cache declaration of java.io.Serializable & java.io.Externalizable
     * while determining the serializationKind of all ClassDocImpls.
     */
    private static ClassDeclaration serializableClassDecl; 
    private static ClassDeclaration externalizableClassDecl;

    /**
     * hold ClassDefinition and ClassDocImpl. For use in queue below.
     */
    private static class DefDoc {
        final ClassDefinition def; 
        final ClassDocImpl classdoc;

        DefDoc(ClassDefinition def, ClassDocImpl classdoc) {
            this.def = def;
            this.classdoc = classdoc;
        }
    }

    static LinkedList constructionCompletionQueue = new LinkedList();

    /**
     * common constructor - only called from other constructors.
     */
    private ClassDocImpl(Env env, ClassDeclaration decl, String documentation, 
                     int modifiers, 
                     ClassDefinition containing, ClassDocImpl superclass) {
        super(env, documentation, modifiers, containing);
        this.superclass = superclass;
	this.kind = kind;
        Identifier id = decl.getName();
        Identifier qid = id.getQualifier();
        this.qualifier = qid.toString();
	this.containingPackage = PackageDocImpl.getPackageDocImpl(env, qualifier);
        this.containingPackage.addClass(this);
        this.name = id.getFlatName().toString();
	classMap.put(qualifiedName(), this);
    }

    /**
     * Constructor - Normal (non-error) case
     */
    private ClassDocImpl(Env env, ClassDefinition def, ClassDeclaration decl) {
        this(env, decl, def.getDocumentation(), 
             def.getModifiers(), 
             def.getOuterClass(), superclass(env, def));
        this.kind = kind(def);
        this.isIncluded = containingPackage.isIncluded && shouldDocument(def);
        // queue this ClassDocImpl for construction completion
        constructionCompletionQueue.addLast(new DefDoc(def, this));
    }

    /**
     * Constructor - class not found case
     */
    private ClassDocImpl(Env env, ClassDeclaration decl) {
        this(env, decl, "", 
             0, 
             null, null);
        this.kind = ORDINARY_CLASS;
        this.isIncluded = false;
    }

    /**
     * Constructor assist
     *
     * Determine is it is an INTERFACE, ORDINARY_CLASS, EXCEPTION, ERROR.
     */
    private int kind(ClassDefinition def) {
        if (def.isInterface()) {
            return INTERFACE;
        } else if (superclass == null) {
	   return ORDINARY_CLASS;
	} else {
	  boolean isJavaLang = qualifier.equals("java.lang");
	  if (isJavaLang && name.equals("Exception")) {
	    return EXCEPTION;
	  } else if (isJavaLang && name.equals("Error")) {
	    return ERROR;
	  } else {
	    return superclass.kind;
	  }
	}
    }

    /**
     * Constructor assist
     *
     * Return the superclass of this class
     *
     * @return the ClassDocImpl for the superclass of this class, null if 
     * there is no superclass.
     */
    private static ClassDocImpl superclass(Env env, ClassDefinition def) {
        ClassDeclaration sup  = def.getSuperClass(env);
	// check for end of superclass chain, and for interfaces
	// extending Object.
        if (sup == null || def.isInterface()) {
            return null;
	} else {
	    return getClassDocImpl(env, sup);
	}
    }

    /** 
     * Do second phase construction of all ClassDocImpl until queue is empty.
     */
    static void secondPhaseConstruction(Env env) {
	serializableClassDecl =
	    env.getClassDeclaration(Identifier.lookup("java.io.Serializable"));
	externalizableClassDecl =
	    env.getClassDeclaration(Identifier.lookup("java.io.Externalizable"));
        while (constructionCompletionQueue.size() != 0) {
            DefDoc defdoc = (DefDoc)constructionCompletionQueue.removeFirst();
            defdoc.classdoc.secondPhaseConstructor(env, defdoc.def);
        }
        constructionCompletionQueue = null; // make sure it can be GC'ed
	serializableClassDecl = null;
	externalizableClassDecl = null;
    }


    /**
     * Second phase construtor assist.
     *
     * Calculate if class c implements Serializable and Externalizable.
     * The Externalizable interface marker overrides the Serializable
     * interface marker.
     *
     * PRECONDITIONS: static fields externalizableClassDecl & serializableClassDecl
     *                have been set appropriately.
     *
     * @returns a bit mask of NOT_SERIALIZABLE, SERIALIZABLE, EXTERNALIZABLE.
     */
    static private int implementsSerialExtern(Environment env, 
					      ClassDeclaration c, 
					      int implementsMask) 
	throws ClassNotFound 
    {
	if (c.equals(externalizableClassDecl)) {
	    // return since EXTERNALIZABLE overrides SERIALIZABLE anywhere
	    // in the class/interface hierarchy.
	    return EXTERNALIZABLE | SERIALIZABLE;
	}

	if (c.equals(serializableClassDecl)) {
	    implementsMask |= SERIALIZABLE;
	}

	for (; c != null ; c = c.getClassDefinition(env).getSuperClass()) {
	    ClassDeclaration intf[] = c.getClassDefinition(env).getInterfaces();
	    for (int i = 0 ; i < intf.length ; i++) {
		implementsMask |= 
		    implementsSerialExtern(env, intf[i], implementsMask);
		if ((implementsMask & EXTERNALIZABLE) != 0) {
	            // return since EXTERNALIZABLE overrides SERIALIZABLE anywhere
	    	    // in the class/interface hierarchy.
		    return EXTERNALIZABLE | SERIALIZABLE;
		}
	    }
	}
	return implementsMask;
    }

    /** 
     * Second phase constructor
     *
     * The following MUST be done after the constructor runs,
     * to avoid infinite recursion or duplicate entries.
     */
    private void secondPhaseConstructor(Env env, ClassDefinition def) {
	setImplementedInterfaces(env, def);
	setImports(env, def);
        fillInMembers(env, def);
	setSerializationKind(env, def);
	if (isSerializable()) {
	    serializedForm = new SerializedForm(env, def, this);
	}
    }

    /**
     * Second phase constructor assist
     *
     * Set interfaces which this class implements.
     *
     * Set an array of ClassDocImpl representing the interfaces
     * that this class implements, set an empty array if the
     * class implements no interfaces.
     */
    private void setImplementedInterfaces(Env env, ClassDefinition def) {   
        ClassDeclaration intf[] = def.getInterfaces();
        interfaces = new ClassDocImpl[intf.length];
        for(int i=0; i< intf.length; i++){
            interfaces[i] = getClassDocImpl(env, intf[i]);
        } 
    }

    /**
     * Second phase constructor assist
     *
     * Set classes and packages which this class imports.
     *
     * Set 'importedClasses' to be an list of ClassDocImpl representing the classes
     * explictly specified for import.
     *
     * Set 'importedPackages' to be an list of PackageDocImpl representing the packages
     * explictly specified for import.
     */
    private void setImports(Env env, ClassDefinition def) { 
        if (def instanceof SourceClass) {
            Imports imports = ((SourceClass)def).getImports();

            // set imported classes
            List cllist = imports.getImportedClasses();
            int clsize = cllist.size();
            for (int i = 0; i < clsize; ++i) {
                Identifier id = ((IdentifierToken)cllist.get(i)).getName();
                id = env.resolvePackageQualifiedName(id);
                try {
                    ClassDefinition imdef = env.getClassDefinition(id);
                    importedClasses.add(ClassDocImpl.getClassDocImpl(env, imdef));
                } catch (ClassNotFound exc) {
                    // YAK!!!
                    System.err.println("Internal error: *IMPORT*NOT*FOUND* " + id.toString() + " - ignoring");
                }
            }

            // set imported packages
            List pkglist = imports.getImportedPackages();
            int pkgsize = pkglist.size();
            for (int i = 0; i < pkgsize; ++i) {
                Identifier id = ((IdentifierToken)pkglist.get(i)).getName();
                importedPackages.add(PackageDocImpl.getPackageDocImpl(env, id.toString()));
            }
        }
    }

    private void setSerializationKind(Env env, ClassDefinition def) {
	try { 
	    serializationKind = 
		implementsSerialExtern(env, def.getClassDeclaration(), NOT_SERIALIZABLE);
	} catch (ClassNotFound e) {
	    Res.warning("doc.class_not_found", e.name.toString());
	}
    }

    /**
     * Constructor assist
     *
     * Fill in the member lists (fields, methods, constructors, and
     * innerclasses) of this class.
     */
    private void fillInMembers(Env env, ClassDefinition def) {
        MemberDefinition f;
        for (f = def.getFirstMember(); f != null; f = f.getNextMember()) {
            if (shouldDocument(f)) {
                if (f.isConstructor()) {
                    constructors.add(ConstructorDocImpl
				     .getConstructorDocImpl(env, f));
                } else if (f.isMethod()) {
                    methods.add(MethodDocImpl.getMethodDocImpl(env, f));
                } else if (f.isVariable()) {
                    fields.add(FieldDocImpl.getFieldDocImpl(env, f));
                } else if (f.isInnerClass()) {
                    innerClasses.add(getClassDocImpl(env, f.getInnerClass()));
                } // else ???
            }
        }
    }
 
    /**
     * Return true if this class implements <code>java.io.Serializable</code>.
     *
     * Since <code>java.io.Externalizable</code> extends 
     * <code>java.io.Serializable</code>, 
     * Externalizable objects are also Serializable.
     */
    public boolean isSerializable() {
	return (serializationKind & SERIALIZABLE) != 0;
    }
 
    /**
     * Return true if this class implements 
     * <code>java.io.Externalizable</code>.
     */
    public boolean isExternalizable() {
	return (serializationKind & EXTERNALIZABLE) != 0;
    }
    
    /**
     * Return the serialization methods for this class.
     *
     * @return an array of <code>MethodDocImpl</code> that represents
     * the serialization methods for this class.
     */
    public MethodDoc[] serializationMethods() {
	return (serializedForm == null) ? 
	    new MethodDoc[0] : serializedForm.methods();
    }

    /**
     * Return the Serializable fields of class.<p>
     *
     * Return either a list of default fields documented by 
     * <code>serial</code> tag<br>
     * or return a single <code>FieldDoc</code> for 
     * <code>serialPersistentField</code> member. 
     * There should be a <code>serialField</code> tag for
     * each Serializable field defined by an <code>ObjectStreamField</code>
     * array component of <code>serialPersistentField</code>.
     *
     * @returns an array of <code>FieldDoc</code> for the Serializable fields 
     * of this class.
     * 
     * @see #definesSerializableFields()
     * @see SerialFieldTagImpl
     */
    public FieldDoc[] serializableFields() {
	return (serializedForm == null) ?
	    new FieldDoc[0] : serializedForm.fields();
    }

    /** 
     *  Return true if Serializable fields are explicitly defined with
     *  the special class member <code>serialPersistentFields</code>.
     *
     * @see #serializableFields()
     * @see SerialFieldTagImpl
     */
    public boolean definesSerializableFields() {
	return serializedForm != null ? 
	    serializedForm.definesSerializableFields() : false;
    }

    /**
     * Factory method
     */
    static ClassDocImpl getClassDocImpl(Env env, ClassDefinition def) {
        // re-get the class definition to be sure that it 
        // has been basicCheck()ed.
        return getClassDocImpl(env, def.getClassDeclaration());
    }
  
    /**
     * Factory method
     */
    static ClassDocImpl getClassDocImpl(Env env, ClassDeclaration decl) {
        ClassDocImpl cd = lookup(decl);
        if (cd == null) {
            try { 
                ClassDefinition def = decl.getClassDefinition(env); 
		cd = new ClassDocImpl(env, def, decl);
            } catch (ClassNotFound exc) {
                Res.warning("doc.class_not_found", exc.name.toString());
		cd = new ClassDocImpl(env, decl);
            }
        }
        return cd;
    }
  
    /**
     * Lookup a ClassDocImpl by fully qualified classname.
     *
     * @param qualName a fully qualified name of a class
     * @return The ClassDocImpl having the name 'qualName',
     * null if this class was not processed by this run of javadoc..
     */
    static final ClassDocImpl lookup(String qualName) {
        return (ClassDocImpl)classMap.get(qualName);
    }
  
    /**
     * Lookup a ClassDocImpl by ClassDeclaration.
     * Return null if not found.
     */
    static final ClassDocImpl lookup(ClassDeclaration decl) {
        return lookup(qualifiedName(decl));
    }

    /** 
     * return true if this is a class, not an interface.
     */
    public boolean isClass() {
        return kind != INTERFACE;
    }

    /** 
     * Return true if this is a ordinary class,
     * not an exception, an error, or an interface.
     */
    public boolean isOrdinaryClass() {
        return kind == ORDINARY_CLASS;
    }

    /** 
     * Return true if this is an interface.
     */
    public boolean isInterface() {
        return kind == INTERFACE;
    }

    /** 
     * Return true if this is an exception class
     */
    public boolean isException() {
        return kind == EXCEPTION;
    }

    /** 
     * Return true if this is an error class
     */
    public boolean isError() {
        return kind == ERROR;
    }

    /** 
     * Return true if this class is abstract 
     */
    public boolean isAbstract() {
        return Modifier.isAbstract(modifiers);
    }

    /** 
     * return true if this class is include in the active set.
     */
    public boolean isIncluded() {
        return isIncluded;
    }

    /**
     * Return the package that this class is contained in.
     */
    public PackageDoc containingPackage() {
        return containingPackage;
    }

    /**
     * Return the class name without qualifier as a String.
     * <pre>
     * Example:
     *  for java.util.Hashtable
     *  return Hashtable
     * </pre>
     */
    public String name() {
        return name;
    }

    /**
     * Return unqualified name of type excluding any dimension information.
     * <p>
     * For example, a two dimensional array of String returns 'String'.
     */
    public String typeName() {
        return name;
    }

    /** 
     * Return the qualified class name as a String.
     * <pre>
     * Example:
     *  for java.util.Hashtable
     *  return java.util.Hashtable
     *  if no qualifier, just return flat name
     * </pre>
     */
    public String qualifiedName() {
        if (qualifier.length() == 0) {
            return name;
        } else {
            return (new StringBuffer(qualifier))
                          .append('.').append(name).toString();
        }
    }

    /**
     * Return qualified name of type excluding any dimension information.
     *<p>
    * For example, a two dimensional array of String 
     * returns 'java.lang.String'.
     */
    public String qualifiedTypeName() {
        return qualifiedName();
    }

    /** 
     * Return, as a string, the qualified class name of the specified 
     * class declaration..
     * <pre>
     * Example:
     *  for java.util.Hashtable
     *  return java.util.Hashtable
     *  if no qualifier, just return flat name
     * </pre>
     */
    static String qualifiedName(ClassDeclaration decl) {
        Identifier did = decl.getName();
        Identifier id = Identifier.lookup(did.getQualifier(), 
                                          did.getFlatName());
    
        return id.toString();
    }

    /**
     * Return the superclass of this class
     *
     * @return the ClassDocImpl for the superclass of this class, null if 
     * there is no superclass.
     */
    public ClassDoc superclass() {
        return superclass;
    }

    /**
     * Test whether this class is a subclass of the specified class.
     *
     * @param cd the candidate superclass.
     * @return true if cd is a superclass of this class.
     */
    public boolean subclassOf(ClassDoc cd) {
        if (cd == this) {
            return true;
        }
        if (superclass == null) {
            return false;
        }
        return superclass.subclassOf(cd);
    }

    /**
     * Return interfaces implemented by this class or interfaces extended
     * by this interface.
     *
     * @return An array of ClassDocImpl representing the interfaces.
     * Return an empty array if there are no interfaces.
     */
    public ClassDoc[] interfaces() {   
        // copy into new ClassDoc array.
        ClassDoc[] ta = new ClassDoc[interfaces.length];
        for(int i=0; i< interfaces.length; i++){
            ta[i] = interfaces[i];
        } 
        return ta;
    }

    /**
     * Return interfaces implemented by this class or interfaces extended
     * by this interface.
     *
     * @return An array of ClassDocImpl representing the interfaces.
     * Return an empty array if there are no interfaces.
     *
     * @deprecated Use {@link #interfaces()} - this method will go away by FCS.
     */
    public ClassDoc[] implementedInterfaces() {   
        return interfaces();
    }

    /**
     * Return fields in class.
     *
     * @return an array of FieldDocImpl for representing the visible
     * fields in this class.
     */
    public FieldDoc[] fields() {  
        return (FieldDoc[])fields.toArray(new FieldDoc[fields.size()]);
    }

    /**
     * Return methods in class.
     *
     * @return an array of MethodDocImpl for representing the visible
     * methods in this class.  Does not include constructors.
     */
    public MethodDoc[] methods() {  
        return (MethodDoc[])methods.toArray(new MethodDoc[methods.size()]);
    }

    /**
     * Return constructors in class.
     *
     * @return an array of ConstructorDocImpl for representing the visible
     * constructors in this class. 
     */
    public ConstructorDoc[] constructors() {  
        return (ConstructorDoc[])constructors.toArray(new ConstructorDoc[constructors.size()]);
    }

    /**
     * Return inner classes within this class.
     *
     * @return an array of ClassDocImpl for representing the visible
     * classes defined in this class. Anonymous and local classes
     * are not included.
     */
    public ClassDoc[] innerClasses() {
        return (ClassDoc[])innerClasses.toArray(new ClassDoc[innerClasses.size()]);
    }

    /**
     * Find a class within the context of this class.
     * Search order: qualified name, in this class (inner),
     * in this package, in the class imports, in the package
     * imports.
     * Return the ClassDocImpl if found, null if not found.
     */
    public ClassDoc findClass(String className) {
        ClassDoc cd;
        // check for fully qualified name
        cd = lookup(className);
        if (cd != null) {
            return cd;
        }
        // check for inner classes
        for (Iterator it = innerClasses.iterator(); it.hasNext(); ) {
            cd = (ClassDoc)it.next();
            if (cd.name().equals(className) || cd.name().endsWith(className)) {
                return cd;
            } else {
                ClassDoc innercd = cd.findClass(className);
                if (innercd != null) {
                    return innercd;
                }
            }
        }
        // check in this package
        cd = containingPackage.findClass(className);
        if (cd != null) {
            return cd;
        }
        // check class imports
        for (Iterator it = importedClasses.iterator(); it.hasNext(); ) {
            cd = (ClassDocImpl)it.next();
            if (cd.name().equals(className)) {
                return cd;
            }
        }
        // check in imported packages
        for (Iterator it = importedPackages.iterator(); it.hasNext(); ) {
            PackageDoc pd = (PackageDoc)it.next();
            cd = pd.findClass(className);
            if (cd != null) {
                return cd;
            }
        }
        return null; // not found
    }

    /**
     * Find a method in this class scope.
     * Search order: this class, outerclasses, interfaces, superclasses.
     * IMP: If see tag is defined in an inner class, which extends a
     * super class and if outerclass and the super class have a visible field
     * in common then Java compiler cribs about the ambiguity, but the
     * following code will serach in the above given search order.
     *
     * @param methodName the unqualified name to search for.
     * @param paramTypeArray the array of Strings for method parameter types. 
     * @return the first MethodDocImpl which matches, null if not found.
     */
    public MethodDoc findMethod(String methodName, String[] paramTypeArray) {
        MethodDoc method;
        for (int i = 0; i < methods.size(); i++) {
            method = (MethodDoc)methods.get(i);
            if (((ExecutableMemberDocImpl)method).
                           hasNameAndParameters(methodName, paramTypeArray)) {
                return method;
            }
        }
        if (containingClass != null) {
            method = containingClass.findMethod(methodName, paramTypeArray); 
            if (method != null) {
                return method;
            }
        }
        if (superclass != null) {
            method = superclass.findMethod(methodName, paramTypeArray); 
            if (method != null) {
                return method;
            }
        } 
        ClassDoc[] iin = interfaces();
        for (int i = 0; i < iin.length; i++) {
            method = ((ClassDocImpl)iin[i]).findMethod(methodName, 
                                                       paramTypeArray);
            if (method != null) { 
                return method;
            }
        }
        return null; 
    }

    /**
     * Find constructor in this class. 
     *
     * @param constrName the unqualified name to search for.
     * @param paramTypeArray the array of Strings for constructor parameters. 
     * @return the first ConstructorDocImpl which matches, null if not found.
     */
    public ConstructorDoc findConstructor(String constrName, 
                                          String[] paramTypeArray) {
        for (int i = 0; i < constructors.size(); i++) {
            ConstructorDoc constr = (ConstructorDoc)constructors.get(i);
            if (((ExecutableMemberDocImpl)constr).
                            hasNameAndParameters(constrName, paramTypeArray)) {
                return constr;
            }
        }
        return null;
    }

    /**
     * Find a field in this class scope.
     * Search order: this class, outerclasses, interfaces, superclasses.
     * IMP: If see tag is defined in an inner class, which extends a
     * super class and if outerclass and the super class have a visible field
     * in common then Java compiler cribs about the ambiguity, but the
     * following code will serach in the above given search order.
     *
     * @param fieldName the unqualified name to search for.
     * @return the first FieldDocImpl which matches, null if not found.
     */
    public FieldDoc findField(String fieldName) {
        FieldDoc field;
        for (int i = 0; i < fields.size(); i++) {
            FieldDocImpl fd = (FieldDocImpl)fields.get(i);
            if (fd.name().equals(fieldName)) {
                return fd;
            }
        }
        if (containingClass != null) {
            field = ((ClassDocImpl)containingClass).findField(fieldName); 
            if (field != null) {
                return field;
            }
        }
        if (superclass != null) {
            field = superclass.findField(fieldName); 
            if (field != null) {
                return field;
            }
        } 
        ClassDoc[] iin = interfaces();
        for (int i = 0; i < iin.length; i++) {
            field = ((ClassDocImpl)iin[i]).findField(fieldName);
            if (field != null) { 
                return field;
            }
        }
        return null;
    }

    /**
     * Get the list of classes declared as imported.
     * These are called "single-type-import declarations" in the JLS.
     *
     * @return an array of ClassDocImpl representing the imported classes.
     */
    public ClassDoc[] importedClasses() {
        return (ClassDoc[])importedClasses.toArray(new ClassDoc[importedClasses.size()]);
    }

    /**
     * Get the list of packages declared as imported.
     * These are called "type-import-on-demand declarations" in the JLS.
     *
     * @return an array of PackageDocImpl representing the imported packages.
     */
    public PackageDoc[] importedPackages() {
        return (PackageDoc[])importedPackages.toArray(new PackageDoc[importedPackages.size()]);
    }
        
    ClassDocImpl thisClassDocImpl() {
        return this;
    }

    /**
     * Return the type's dimension information, as a string.
     * <p>
     * For example, a two dimensional array of String returns '[][]'.
     *<p>
     * In ClassDocImpl so that it implements TypeDoc interface, always
     * returns empty string.
     */
    public String dimension() {
        return "";
    }

    /**
     * Return this type as a class.  Array dimensions are ignored.
     *<p>
     * In ClassDocImpl so that it implements TypeDoc interface, always
     * returns this.
     *
     * @return a ClassDocImpl if the type is a Class.
     * Return null if it is a primitive type..
     */
    public ClassDoc asClassDoc() {
        return this;
    }
    
    /**
     * Returns a string representation of the class.
     *
     * @return  the qualified class name
     */
    public String toString() {
        return qualifiedName();
    }
}


