/*
 * @(#)Comment.java	1.20 98/08/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.tools.javadoc;

import com.sun.javadoc.*;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;

import java.text.BreakIterator;
import java.text.CollationKey;

/**
 * Comment contains all information in comment part.
 *      It allows users to get first sentence of this comment, get
 *      comment for different tags...
 *
 * @author Kaiyang Liu (original)
 * @author Robert Field (rewrite)
 * @author Atul M Dambalkar
 */
class Comment {

    /**
     * raw text of the comment.
     */
    String commentString;

    /**
     * sorted comments with different tags.
     */
    List tagList;

    /**
     * text minus any tags.
     */
    String text;

    /**
     * Sentence instance from the BreakIterator.
     */
    static BreakIterator sentence;


    /**
     * constructor of Comment.
     */
    Comment(Doc holder, String commentString) {
	this.commentString = commentString;
        tagList = new ArrayList();
        if (sentence == null) {
            sentence = BreakIterator.getSentenceInstance(RootDocImpl.locale);
        }
        parseCommentStateMachine(holder);
    }

    private void warnIfEmpty(Doc holder, String tagName, String tx) {
        if (tx.length() == 0) {
            Res.warning("tag.tag_has_no_arguments", holder.toString(), tagName);
        }
    }

    /**
     * Private to the implementation of the constructor.
     * Separates the comment into the text part and zero to N tags.
     * Simple state machine is in one of three states:
     * <pre>
     * IN_TEXT: parsing the comment text or tag text.
     * TAG_NAME: parsing the name of a tag.
     * TAG_GAP: skipping through the gap between the tag name and
     * the tag text.
     * </pre>
     *
     */
    private void parseCommentStateMachine(Doc holder) {
        final int IN_TEXT = 1;
        final int TAG_GAP = 2;
        final int TAG_NAME = 3;
        int state = TAG_GAP;
        boolean newLine = true;
        String tagName = null;
        int tagStart = 0;
        int textStart = 0;
        int lastNonWhite = -1;
        int len = commentString.length();
        for (int inx = 0; inx < len; ++inx) {
            char ch = commentString.charAt(inx);
            boolean isWhite = Character.isWhitespace(ch);
            switch (state)  {
                case TAG_NAME:
                    if (isWhite) {
                        tagName = commentString.substring(tagStart, inx);
                        state = TAG_GAP;
                    }
                    break;
                case TAG_GAP:
                    if (isWhite) {
                        break;
                    }
                    textStart = inx;
                    state = IN_TEXT;
                    /* fall thru */
                case IN_TEXT:
                    if (newLine && ch == '@') {
                        parseCommentComponent(holder, tagName, textStart,
                                              lastNonWhite+1);
                        tagStart = inx;
                        state = TAG_NAME;
                    }
                    break;
            };
            if (ch == '\n') {
                newLine = true;
            } else if (!isWhite) {
                lastNonWhite = inx;
                newLine = false;
            }
        }
        // Finish what's currently being processed
        switch (state)  {
            case TAG_NAME:
                tagName = commentString.substring(tagStart, len);
                /* fall thru */
            case TAG_GAP:
                textStart = len;
                /* fall thru */
            case IN_TEXT:
                parseCommentComponent(holder, tagName, textStart, lastNonWhite+1);
                break;
        };
    }


    /**
     * Private to the implementation of the constructor.
     * Saves away the last parsed item.
     */
    private void parseCommentComponent(Doc holder, String tagName,
				       int from, int upto) {
        String tx = upto <= from ? "" : commentString.substring(from, upto);
        if (tagName == null) {
            text = tx;
        } else {
            TagImpl tag;
            if (tagName.equals("@exception") || tagName.equals("@throws")) {
	        warnIfEmpty(holder, tagName, tx);
                tag = new ThrowsTagImpl(holder, tagName, tx);
            } else if (tagName.equals("@param")) {
	        warnIfEmpty(holder, tagName, tx);
                tag = new ParamTagImpl(holder, tagName, tx);
            } else if (tagName.equals("@see")) {
	        warnIfEmpty(holder, tagName, tx);
                tag = new SeeTagImpl(holder, tagName, tx);
            } else if (tagName.equals("@serialField")) {
	        warnIfEmpty(holder, tagName, tx);
                tag = new SerialFieldTagImpl(holder, tagName, tx);
            } else if (tagName.equals("@return")) {
	        warnIfEmpty(holder, tagName, tx);
                tag = new TagImpl(holder, tagName, tx);
            } else if (tagName.equals("@author")) {
	        warnIfEmpty(holder, tagName, tx);
                tag = new TagImpl(holder, tagName, tx);
            } else if (tagName.equals("@version")) {
	        warnIfEmpty(holder, tagName, tx);
                tag = new TagImpl(holder, tagName, tx);
            } else {
                tag = new TagImpl(holder, tagName, tx);
            }
            tagList.add(tag);
        }
    }

    /**
     * Return the text of the comment.
     */
    String commentText() {
        return text;
    }

    /**
     * Return all tags in this comment.
     */
    Tag[] tags() {
        return (Tag[])tagList.toArray(new Tag[tagList.size()]);
    }

    /**
     * Return tags of the specified kind in this comment.
     */
    Tag[] tags(String tagname) {
	List found = new ArrayList();
        Iterator it = tagList.iterator();
	String target = tagname;
        if (target.charAt(0) != '@') {
            target = "@" + target;
        }
	while (it.hasNext()) {
	    TagImpl tag = (TagImpl)it.next();
	    if (tag.kind().equals(target)) {
		found.add(tag);
	    }
	}
        return (Tag[])found.toArray(new Tag[found.size()]);
    }

    /**
     * Return throws tags in this comment.
     */
    ThrowsTag[] throwsTags() {
	List found = new ArrayList();
        Iterator it = tagList.iterator();
	while (it.hasNext()) {
	    Object next = it.next();
	    if (next instanceof ThrowsTagImpl) {
		found.add(next);
	    }
	}
	return (ThrowsTag[])found.toArray(new ThrowsTag[found.size()]);
    }

    /**
     * Return param tags in this comment.
     */
    ParamTag[] paramTags() {
	List found = new ArrayList();
        Iterator it = tagList.iterator();
	while (it.hasNext()) {
	    Object next = it.next();
	    if (next instanceof ParamTagImpl) {
		found.add(next);
	    }
	}
	return (ParamTag[])found.toArray(new ParamTag[found.size()]);
    }

    /**
     * Return see also tags in this comment.
     */
    SeeTag[] seeTags() {
	List found = new ArrayList();
        Iterator it = tagList.iterator();
	while (it.hasNext()) {
	    Object next = it.next();
	    if (next instanceof SeeTagImpl) {
		found.add(next);
	    }
	}
	return (SeeTag[])found.toArray(new SeeTag[found.size()]);
    }

    /**
     * Return serialField tags in this comment.
     */
    SerialFieldTag[] serialFieldTags() {
	List found = new ArrayList();
        Iterator it = tagList.iterator();
	while (it.hasNext()) {
	    Object next = it.next();
	    if (next instanceof SerialFieldTagImpl) {
		found.add(next);
	    }
	}
	return (SerialFieldTag[])found.toArray(new SerialFieldTag[found.size()]);
    }


    static String localeSpecificFirstSentence(String s) {
        if (s == null || s.length() == 0) {
            return "";
        }
        if (RootDocImpl.locale.getLanguage().equals("en")) {
             return englishLanguageFirstSentence(s);
        }
        sentence.setText(s.replace('\n', ' '));
        int start = sentence.first();
        int end = sentence.next();
        return s.substring(start, end);

    }

    /**
     * Return the first sentence of a string, where a sentence ends
     * with a period followed be white space.
     */
    static String englishLanguageFirstSentence(String s) {
        if (s == null)
            return null;
        int len = s.length();
        boolean period = false;
        for (int i = 0 ; i < len ; i++) {
            switch (s.charAt(i)) {
                case '.':
                    period = true;
                    break;
                case ' ':
                case '\t':
                case '\n':
   	        case '\r':
  	        case '\f':
                    if (period) {
                        return s.substring(0, i);
                    }
                    break;
                default:
                    period = false;
            }
        }
        return s;
    }

    /**
     * Return array of tags with text and inline See Tags for a Doc comment.
     */
    static Tag[] getInlineTags(Doc holder, String inlinetext) {
        List taglist = new ArrayList();
        int delimend = 0, textstart = 0, len = inlinetext.length();
        if (len == 0) {
            return (Tag[])taglist.toArray(new Tag[taglist.size()]);
        }
        while (true) {
            int linkstart;
            if ((linkstart = inlineTagImplFound(inlinetext,
                                                   textstart)) == -1) {
                taglist.add(new TagImpl(holder, "Text",
                                    inlinetext.substring(textstart)));
                break;
            } else {
                int seetextstart = linkstart;
                if (linkstart >= 0) {
                    seetextstart += 6;
                    while (Character.isWhitespace(inlinetext.
                                                      charAt(seetextstart))) {
                        if (inlinetext.length() <= seetextstart) {
                            taglist.add(new TagImpl(holder, "Text",
                                inlinetext.substring(textstart, seetextstart)));
                            Res.warning("tag.Improper_Use_Of_Link_Tag",
                                                                   inlinetext);
                            return (Tag[])taglist.toArray(new Tag[taglist.size()]);
                        } else {
                            seetextstart++;
                        }
                    }
                    taglist.add(new TagImpl(holder, "Text",
                                inlinetext.substring(textstart, linkstart)));
                }
                textstart = seetextstart;   // this text is actually seetag
                if ((delimend = inlinetext.indexOf("}", textstart)) == -1) {
                    // store the text as it is with the {@link.
                    taglist.add(new TagImpl(holder, "Text",
                                      inlinetext.substring(textstart)));
                    Res.warning("tag.End_delimiter_missing_for_possible_SeeTag",
                                 inlinetext);
                    return (Tag[])taglist.toArray(new Tag[taglist.size()]);
                } else {
                    taglist.add(new SeeTagImpl(holder, "@link",
                                inlinetext.substring(textstart, delimend)));
                    textstart = delimend + 1;
                }
            }
            if (textstart == inlinetext.length()) {
                break;
            }
        }
	return (Tag[])taglist.toArray(new Tag[taglist.size()]);
    }

    /**
     * Recursively search for the string "{@link" or "{@LINK" followed by
     * white space,
     * if found
     *    return the index of the text following the white space.
     * else
     *    return -1.
     */
    static int inlineTagImplFound(String inlinetext, int start) {
        int linkstart;
        if (start == inlinetext.length()) {
            return -1;
        }
        if ((linkstart = inlinetext.indexOf("{@", start)) == -1) {
            return linkstart;
        } else {  // check for next 4 characters for link/LINK
            if (inlinetext.length() <= linkstart + 6) {
                return -1;
            }
            if (inlinetext.substring(linkstart + 2, linkstart + 6).
                toLowerCase().equals("link")) {
                // now check if next char is a whitespace, consume it.
                if (Character.isWhitespace(inlinetext.charAt(linkstart + 6))) {
                    return linkstart;  // found white space char.
                } else {    //
                    Res.warning("tag.Improper_Use_Of_Link_Tag", inlinetext);
                    return inlineTagImplFound(inlinetext, linkstart + 6);
                }
            } else {
                return -1;
            }
        }
    }

    /**
     * Return array of tags for the locale specific first sentence in the text.
     */
    static Tag[] firstSentenceTags(Doc holder, String text) {
        return getInlineTags(holder, localeSpecificFirstSentence(text));
    }

    /**
     * Return text for this Doc comment.
     */
    public String toString() {
	return text;
    }
}
