/*
 * @(#)DocImpl.java	1.12 98/06/23
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.tools.javadoc;

import com.sun.javadoc.*;

import java.text.CollationKey;

import sun.tools.java.ClassDefinition;
import sun.tools.java.MemberDefinition;

import sun.tools.util.ModifierFilter;

/**
 * abstract base class of all Doc classes.  Doc item's are representations
 * of java language constructs (class, package, method,...) which have
 * comments and have been processed by this run of javadoc.  All Doc items
 * are unique, that is, they are == comparable.
 *
 * @since JDK1.2
 * @author Robert Field
 * @author Atul M Dambalkar
 */
abstract class DocImpl implements Doc, Comparable {

    /**
     *  The complex comment object, lazily initialized.
     */
    private Comment comment;  

    static ModifierFilter showAccess;

    /**
     * The cached sort key, to take care of Natural Language Text sorting. 
     */
    CollationKey collationkey = null;

    /**
     *  Raw documentation string.
     */
    String documentation;

    /**
     * Constructor.
     */
    DocImpl(Env env, String documentation) {
        this.documentation = documentation;
    }

    /**
     * So subclasses have the option to do lazy initialization of 
     * "documentation" string.
     */
    String documentation() {
        if (documentation == null) {
            documentation = "";
        }
        return documentation;
    }

    /**
     * For lazy initialization of comment.
     */
    Comment comment() {
        if (comment == null) {
            comment = new Comment(this, documentation());
        }
        return comment;
    }

    /**
     *  Hook so we can allow unfiltered access, with a yet
     *  to be determined API addition.
     */
    boolean checkAccess() {
        return true;
    }

    /**
     * Return the text of the comment for this doc item.
     * TagImpls have been removed.
     */
    public String commentText() {
        return comment().commentText();
    }

    /**
     * Return all tags in this Doc item.
     *
     * @return an array of TagImpl containing all tags on this Doc item.
     */
    public Tag[] tags() {
        return comment().tags();
    }

    /**
     * Return tags of the specified kind in this Doc item.
     *
     * @param tagname name of the tag kind to search for.
     * @return an array of TagImpl containing all tags whose 'kind()'
     * matches 'tagname'.
     */
    public Tag[] tags(String tagname) {
        return comment().tags(tagname);
    }

    /**
     * Return the see also tags in this Doc item.
     * 
     * @return an array of SeeTag containing all &#64see tags.
     */
    public SeeTag[] seeTags() {
        return comment().seeTags();
    }

    public Tag[] inlineTags() {
        return Comment.getInlineTags(this, commentText());
    }

    public Tag[] firstSentenceTags() {
        return Comment.firstSentenceTags(this, commentText());
    }

    /**
     * Utility for subclasses which read HTML documentation files.
     */
    String readHTMLDocumentation(java.io.InputStream input, String filename) throws java.io.IOException {
        int filesize = input.available();
        byte[] filecontents = new byte[filesize];
        input.read(filecontents, 0, filesize);
        input.close();
        String rawDoc = new String(filecontents);

        String upper = null;
        int bodyIdx = rawDoc.indexOf("<body");
        if (bodyIdx == -1) {
            bodyIdx = rawDoc.indexOf("<BODY");
            if (bodyIdx == -1) {
                upper = rawDoc.toUpperCase();
                bodyIdx = upper.indexOf("<BODY");
                if (bodyIdx == -1) {
                    Res.error("javadoc.Body_missing_from_html_file", filename);
                    return "";
                }
            }
        }
        bodyIdx = rawDoc.indexOf('>', bodyIdx);
        if (bodyIdx == -1) {
            Res.error("javadoc.Body_missing_from_html_file", filename);
            return "";
        }
        ++bodyIdx;
        int endIdx = rawDoc.indexOf("</body", bodyIdx);
        if (endIdx == -1) {
            endIdx = rawDoc.indexOf("</BODY", bodyIdx);
            if (endIdx == -1) {
                if (upper == null) {
                    upper = rawDoc.toUpperCase();
                }
                endIdx = upper.indexOf("</BODY", bodyIdx);
                if (endIdx == -1) {
                    Res.error("javadoc.End_body_missing_from_html_file", filename);
                    return "";
                }
            }
        }
        return rawDoc.substring(bodyIdx, endIdx);
    }

    /**
     * Return the full unprocessed text of the comment.  Tags
     * are included as text.  Used mainly for store and retrieve
     * operations like internalization. 
     */
    public String getRawCommentText() {
        return documentation();
    }
    
    /**
     * Set the full unprocessed text of the comment.  Tags
     * are included as text.  Used mainly for store and retrieve
     * operations like internalization. 
     */
    public void setRawCommentText(String rawDocumentation) {
        documentation = rawDocumentation;
        comment = null;
    }
    
    /** check whether this member should be documented */
    boolean shouldDocument(MemberDefinition f){
        if (!checkAccess()) {
            return true;
        }
	if (f.isSynthetic()) {
	    return false;
	}

        // if it is an innerclass, should pass class documenting requirements
	ClassDefinition cdef = f.getInnerClass();
	if(cdef != null) {
	    if(!shouldDocument(cdef)) {
                return false;
            }
        }

	return showAccess.checkMember(f);
    }

    /** check whether this class should be documented */
    boolean shouldDocument(ClassDefinition cdef){
        if (!checkAccess()) {
            return true;
        }
	if (cdef.isSynthetic()) {
	    return false;
	}
	// we don't document anonymous or local classes
	if (cdef.isAnonymous() || cdef.isLocal()) {
	    return false;
        }

        //return showAccess.checkClass(cdef);
        return isVisible(cdef);
    }

    /**
     * Check the visibility if this is an inner class. 
     * if this is not an inner class, return true.
     * if this is an static visible inner class,
     *    return true.
     * if this is an visible inner class 
     *    if the outer class is visible return true.
     *    else return false.
     * IMPORTANT: This also allows, static inner classes to be defined in side
     * an inner class, which is not allowed by the compiler. So such an test
     * case will not reach upto this method itself, but if compiler allows it,
     * then that will go through.
     */
    boolean isVisible(ClassDefinition cdef) {
        if (!showAccess.checkClass(cdef)) {
            return false;
        }
        ClassDefinition outer = cdef.getOuterClass();
        return outer == null || cdef.isStatic() || isVisible(outer);
    }
       
    ClassDocImpl thisClassDocImpl() {
        return null;
    }

    /**
     * return a key for sorting.
     */
    CollationKey key() {
        if (collationkey == null) {
            collationkey = generateKey();
        }
        return collationkey;
    }

    /**
     * generate a key for sorting. <p>
     * default is toString();
     */
    CollationKey generateKey() {
        return RootDocImpl.collator.getCollationKey(name());
    }

    /**
     * Returns a string representation of this Doc item.
     *
     * @return  the name
     */
    public String toString() {
        return name();
    }

    /**
     * Returns the name of this Doc item.
     *
     * @return  the name
     */
    public abstract String name();

    /**
     * Compares this Object with the specified Object for order.  Returns a
     * negative integer, zero, or a positive integer as this Object is less
     * than, equal to, or greater than the given Object.
     * <p>
     * Included so that Doc item are java.lang.Comparable.
     * 
     * @param   o the <code>Object</code> to be compared.
     * @return  a negative integer, zero, or a positive integer as this Object
     *		is less than, equal to, or greater than the given Object.
     * @exception ClassCastException the specified Object's type prevents it
     *		  from being compared to this Object. 
     */
    public int compareTo(Object obj) {
        return key().compareTo(((DocImpl)obj).key());
    }

    /**
     * Is this Doc item a field?  False until overridden.
     *
     * @return true if it represents a field
     */
    public boolean isField() {
        return false;
    }

    /**
     * Is this Doc item a simple method (i.e. not a constructor)?  False 
     * until overridden.
     *
     * @return true is it represents a method, false if it is anything
     * else, including constructor, field, class, ...
     */
    public boolean isMethod() {
        return false;
    }

    /**
     * Is this Doc item a constructor?  False until overridden.
     *
     * @return true is it represents a constructor
     */
    public boolean isConstructor() {
        return false;
    }

    /**
     * Is this Doc item a interface?  False until overridden.
     *
     * @return true is it represents a interface
     */
    public boolean isInterface() {
        return false;
    }

    /**
     * Is this Doc item a exception class?  False until overridden.
     *
     * @return true is it represents a exception
     */
    public boolean isException() {
        return false;
    }

    /**
     * Is this Doc item a error class?  False until overridden.
     *
     * @return true is it represents a error
     */
    public boolean isError() {
        return false;
    }

    /**
     * Is this Doc item a ordinary class (i.e. not an interface, exception,
     * or error)?  False until overridden.
     *
     * @return true is it represents a class
     */
    public boolean isOrdinaryClass() {
        return false;
    }

    /**
     * Is this Doc item a class.
     * Note: interfaces are not classes.
     * False until overridden.
     *
     * @return true is it represents a class
     */
    public boolean isClass() {
        return false;
    }

    /** 
     * return true if this Doc is include in the active set.
     */
    public abstract boolean isIncluded();
}

