/*
 * @(#)Main.java	1.40 98/08/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.tools.javadoc;

import com.sun.javadoc.*;

import sun.tools.util.CommandLine;
import sun.tools.util.ModifierFilter;

import sun.tools.java.ClassDeclaration;
import sun.tools.java.ClassFile;
import sun.tools.java.Constants;
import sun.tools.java.Identifier;
import sun.tools.java.Package;	// Supercede java.lang.Package

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;

import java.io.IOException;
import java.io.File;
import java.io.FileNotFoundException;

/**
 * Main program of Javadoc.
 */
public final
class Main implements Constants {
    /**
     * Name of the program
     */
    static final String program = "javadoc";
    static final String standardDocletClassName =
                                     "com.sun.tools.doclets.standard.Standard";
    static final String oneOneDocletClassName =
                                     "com.sun.tools.doclets.oneone.OneOne";
    static List options = new ArrayList();
    static int nwarnings = 0;
    static int nerrors = 0;

    static ModifierFilter showAccess = null;
    static long defaultFilter = 
                     ModifierFilter.PUBLIC | ModifierFilter.PROTECTED;
    static DocletInvoker docletInvoker;

    /**
     * Exit, reporting errors and warnings.
     */
    static void exit() {
        int status = 0;

        if (nerrors > 0) {
            Res.notice( nerrors > 1? "main.errors" : "main.error",
                        "" + nerrors);
            status = 1;
        }
        if (nwarnings > 0) {
            Res.notice(nwarnings > 1?  "main.warnings" : "main.warning",
                       "" + nwarnings);
        }
        System.exit(status);
    }

    /**
     * Usage
     */
    static void usage() {
	Res.notice("main.usage");

        // let doclet print usage information (does nothing on error)
        if (docletInvoker != null) {
            docletInvoker.optionLength("-help");
        }
    }

    /**
     * Main program
     */
    public static void main(String argv[]) {
        String srcPathString = null;
        String sysClassPathString = null;
        String classPathString = null;
        String extDirsString = null;
        boolean verbosePath = false;
	long tm = System.currentTimeMillis();
	List javaNames = new ArrayList();
	int flags = F_WARNINGS;
	String encoding = null;
  
	// Preprocess @file arguments
	try {
	    argv = CommandLine.parse(argv);
	} catch (FileNotFoundException e) {
	    Res.error("main.cant.read", e.getMessage());
	    exit();
	} catch (IOException e) {
	    e.printStackTrace();
	    exit();
	}

        setDocletInvoker(argv);

	// Parse arguments
	for (int i = 0 ; i < argv.length ; i++) {
	    String arg = argv[i];
	    if (arg.equals("-verbose")){
		setOption(arg);
		flags |= F_VERBOSE;
	    } else if (arg.equals("-encoding")) {
		oneArg(argv, i++);
                encoding = argv[i];
	    } else if (arg.equals("-help")) {
                usage();
                exit();
	    } else if (arg.equals("-private")) {
		setOption(arg);
		setFilter(ModifierFilter.ALL_ACCESS);
	    } else if (arg.equals("-package")) {
		setOption(arg);
		setFilter(ModifierFilter.PUBLIC |
                          ModifierFilter.PROTECTED |
                          ModifierFilter.PACKAGE );
	    } else if (arg.equals("-protected")) {
		setOption(arg);
		setFilter(ModifierFilter.PUBLIC | ModifierFilter.PROTECTED );
	    } else if (arg.equals("-public")) {
		setOption(arg);
		setFilter(ModifierFilter.PUBLIC);
            } else if (arg.equals("-sourcepath")) {
		oneArg(argv, i++);
                if (srcPathString != null) {
                    usageError("main.option.already.seen", arg);
                }
                srcPathString = argv[i];
            } else if (arg.equals("-classpath")) {
		oneArg(argv, i++);
                if (classPathString != null) {
                    usageError("main.option.already.seen", arg);
                }
                classPathString = argv[i];
            } else if (arg.equals("-sysclasspath")) {
		oneArg(argv, i++);
                if (sysClassPathString != null) {
                    usageError("main.option.already.seen", arg);
                }
                sysClassPathString = argv[i];
            } else if (arg.equals("-bootclasspath")) {
		oneArg(argv, i++);
                if (sysClassPathString != null) {
                    usageError("main.option.already.seen", arg);
                }
                sysClassPathString = argv[i];
            } else if (arg.equals("-extdirs")) {
		oneArg(argv, i++);
                if (extDirsString != null) {
                    usageError("main.option.already.seen", arg);
                }
                extDirsString = argv[i];
            } else if (argv[i].equals("-Xverbosepath")) {
                verbosePath = true;
	    } else if (arg.equals("-overview")) {
		oneArg(argv, i++);
	    } else if (arg.equals("-doclet")) {
		i++;  // handled in setDocletInvoker
	    } else if (arg.equals("-docletpath")) {
		i++;  // handled in setDocletInvoker
	    } else if (arg.equals("-locale")) {
		oneArg(argv, i++);
                RootDocImpl.setLocale(argv[i]);
	    } else if (arg.equals("-1.1")) {
                setOption(arg);
            } 
            // call doclet for its options

	    // other arg starts with - is invalid
	    else if ( arg.startsWith("-") ) { 
                int optionLength;
                optionLength = docletInvoker.optionLength(arg);
                if (optionLength < 0) {
                    // error already displayed
                    exit();  
                } else if (optionLength == 0) {
                    // option not found
                    usageError("main.invalid_flag", arg);
                } else {
                    // doclet added option
                    if ((i + optionLength) > argv.length) {
                        usageError("main.requires_argument", arg);
                    }
                    List args = new ArrayList();
                    for (int j = 0; j < optionLength-1; ++j) {
                        args.add( argv[++i] );
                    }
                    setOption(arg, args);
                }
	    } else {
		javaNames.add(arg);
	    }
	}

	if (javaNames.isEmpty()) {
	    usageError("main.No_packages_or_classes_specified");
	}

        if (!docletInvoker.validOptions(RootDocImpl.options(options))) {
            // error message already displayed
            exit();
        }

        Env env = Env.create(srcPathString, classPathString, 
                             sysClassPathString, extDirsString,
                             flags, encoding);
        if (verbosePath) {
            env.printPaths();
        }

	RootDocImpl root = init(env, javaNames);

        // compiler stuff should be all gone now - purge it
        env = null; // let go of compiler stuff
        System.gc();

	// pass off control to the doclet
	docletInvoker.start(root);
	
	// We're done.
	if ((flags & F_VERBOSE) != 0) {
	    tm = System.currentTimeMillis() - tm;
	    Res.notice("main.done_in", Long.toString(tm));
	}
	
	exit();
    }

    static void setDocletInvoker(String[] argv) {
        String docletClassName = null;
        String docletPath = null;

	// Parse doclet specifying arguments
	for (int i = 0 ; i < argv.length ; i++) {
	    String arg = argv[i];
            if (arg.equals("-1.1")) {
                if (docletClassName != null) {
                    usageError("main.more_than_one_doclet_specified_0_and_1", 
                              docletClassName, oneOneDocletClassName);
		}                    
                docletClassName = oneOneDocletClassName;
	    } else if (arg.equals("-doclet")) {
		oneArg(argv, i++);
                if (docletClassName != null) {
                    usageError("main.more_than_one_doclet_specified_0_and_1", 
                              docletClassName, argv[i]);
                }                    
                docletClassName = argv[i];
	    } else if (arg.equals("-docletpath")) {
		oneArg(argv, i++);
                if (docletPath == null) {
                    docletPath = argv[i];
                } else {
                    docletPath += File.pathSeparator + argv[i];
                }                    
            }
        }

        if (docletClassName == null) {
            docletClassName = standardDocletClassName;
        }

        // attempt to find doclet
        docletInvoker = new DocletInvoker(docletClassName, docletPath);
    }

    static void setFilter(long filterBits) {
        if (showAccess != null) {
            Res.error("main.incompatible.access.flags");
	    usage();
	    exit();
        }
        showAccess = new ModifierFilter(filterBits);
    }
    
    static RootDocImpl init(Env env, List javaNames) {
        RootDocImpl root = null;

        try {
            root = initInner(env, javaNames);
        } catch (java.lang.OutOfMemoryError ee) {
	    Res.error("main.out.of.memory");
	    exit();
        } catch (Error ee) {
	    ee.printStackTrace();
	    Res.error("main.fatal.error");
	    exit();
	} catch (Exception ee) {
	    ee.printStackTrace();
	    Res.error("main.fatal.exception");
	    exit();
        }
        return root;
    }

    static RootDocImpl initInner(Env env, List javaNames) throws IOException {
        List userPkgs = new ArrayList();
        List userClasses = new ArrayList();
	try {
            Iterator iter = javaNames.iterator();
            while (iter.hasNext()) {
                String name = (String)iter.next();
                if (!parseSourceClass(env, name)) {
                    // not a class or source file, try as package
                    userPkgs.add(name);
                }
            }

            // Figure out which classes where encountered
            for (Enumeration e1 = env.getClasses() ; e1.hasMoreElements() ;) {
                ClassDeclaration decl = (ClassDeclaration)e1.nextElement();
                userClasses.add(decl);
            }

            iter = userPkgs.iterator();
            while (iter.hasNext()) {
                String name = (String)iter.next();
                parseSourcePackage(env, name);
            }
        } finally {
            // copy over error counts
            env.flushErrors();
            nerrors += env.nerrors;
            nwarnings += env.nwarnings;
        }

        if (nerrors > 0) {
            exit();
        }

        Res.notice("main.Building_tree");

        // clear out compilation temps
        System.gc();

        if (showAccess == null) {
            setFilter(defaultFilter);
        }

        // start up info to root
        return new RootDocImpl(env, userClasses, userPkgs, showAccess, options);
    }

    static void parseSourcePackage(Env env, String name) throws IOException {
        Identifier id = Identifier.lookup(name);
				// Check to see if the name refers to an entire
				// package.
        sun.tools.java.Package pkg = env.getPackage(id);
        boolean hasFiles = false;
        if (pkg.exists()) {
            Enumeration sourceEnum = pkg.getSourceFiles();
            while (sourceEnum.hasMoreElements()) {
                if (!hasFiles) {
                    Res.notice("main.Loading_source_files_for_package", name);
                    hasFiles = true;
                }
                env.parseFile((ClassFile)sourceEnum.nextElement());
            }
        }
        if (!hasFiles) {
            Res.error("main.source_not_found", name);
        }            
    }

    static boolean parseSourceClass(Env env, String name) throws IOException {
        Identifier id = Identifier.lookup(name);
				// Check to see if the name refers to an 
				// individual class.
        sun.tools.java.Package pkg = env.getPackage(id.getQualifier());
        if (pkg.exists()) {
            ClassFile sourceFile = pkg.getSourceFile(id.getName());
            if (sourceFile != null) {
				// The name refers to an individual class.
				//
				// TODO: insert later check to make sure that
				// <foo>.java contains a definition of <foo>,
				// as expected.
                env.parseFile(sourceFile);
                Res.notice("main.Loading_source_file_for_class", name);
                ClassDeclaration decl = env.getClassDeclaration(id);
                if (decl.getClassDefinition() == null ||
                    decl.getStatus() == CS_SOURCE) {
                    env.error(0, "wrong.source",
                              sourceFile.getPath(),
                              decl, pkg);
                    decl.setDefinition(null, CS_NOTFOUND);
                }
                return true; // Done
            }
        }
                                // Finally, try to find a file of that name
        File file = new File(name);
        try {
            env.parseFile(new ClassFile(file));
            Res.notice("main.Loading_source_file", file.toString());
            return true;
        } catch (IOException ee) {
            return false;
        }
    }

    /**
     * Set one arg option.
     * Error and exit if one argument is not provided.
     */
    static void oneArg(String[] args, int index) {
        if ((index + 1) < args.length) {
            setOption(args[index], args[index+1]);
        } else {
            usageError("main.requires_argument", args[index]);
        }
    }

    static void usageError(String key) {
        Res.error(key);
        usage();
        exit();
    }       

    static void usageError(String key, String a1) {
        Res.error(key, a1);
        usage();
        exit();
    }       

    static void usageError(String key, String a1, String a2) {
        Res.error(key, a1, a2);
        usage();
        exit();
    }       

    /**
     * indicate an option with no arguments was given.
     */
    static void setOption(String opt) {
        List args = new ArrayList();
        args.add(opt);
        options.add(args);
    }

    /**
     * indicate an option with one argument was given.
     */
    static void setOption(String opt, String argument) {
        List args = new ArrayList();
        args.add(opt);
        args.add(argument);
        options.add(args);
    }

    /**
     * indicate an option with the specified list of arguments was given.
     */
    static void setOption(String opt, List arguments) {
        List args = new ArrayList();
        args.add(opt);
        args.addAll(arguments);
        options.add(args);
    }
}


	
