/*
 * @(#)PackageDocImpl.java	1.18 98/05/14
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.tools.javadoc;

import com.sun.javadoc.*;

import sun.tools.java.ClassFile;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;

/**
 * Represents a java package.  Provides access to information
 * about the package, the package's comment and tags, and the 
 * classes in the package.
 *
 * @since JDK1.2
 * @author Kaiyang Liu (original)
 * @author Robert Field (rewrite)
 */
class PackageDocImpl extends DocImpl implements PackageDoc {

    /**
     *  Map package names to PackageDocImpl, so we can maintain uniqueness.
     */
    private static Map packageMap = new HashMap();

    private final String name;

    private String docPath;
 
    /**
     * all classes in this package
     */
    private final List classes;
    
    boolean isIncluded = false;

    /**
     * Constructor.
     * Get all information of this package, including all
     * classes in this package.
     */
    private PackageDocImpl(Env env, String name) {
        super(env, null);
        this.name = name;
        this.classes = new ArrayList();
        fetchDocumentation(env);
        packageMap.put(name, this);
    }

    /** 
     * Constructor assist.
     *
     * Using the ClassPath variable, search the directory for the package.doc
     * comment file, depending upon the package name. Read the file, and return
     * the unparsed String contents of the file.
     *
     * @return Unparsed Raw text in the file package.doc, for this package.
     */
    private void fetchDocumentation(Env env) {
        String packagepath = name.replace('.', File.separatorChar);
        String filename = packagepath + File.separatorChar + "package.html";

        try {
            ClassFile commentfile = env.getSourceClassFile(filename);
            if (commentfile == null) {
                // No comment file
                documentation = "";
            } else {
                // Comment file found
                if (commentfile.isZipped()) {
                    // Zipped - we need to process now
                    documentation = readHTMLDocumentation(commentfile.getInputStream(), filename);
                } else {
                    // regular file - do lazy init to save memory
                    docPath = commentfile.getPath();
                }
            }
        } catch (IOException exc) {            
            documentation = "";
            Res.error("javadoc.File_Read_Error", filename);
        }
    }  

    static PackageDocImpl getPackageDocImpl(Env env, String name) {
        PackageDocImpl pd = (PackageDocImpl)packageMap.get(name);
        if (pd == null) {
            pd = new PackageDocImpl(env, name);
        }
        return pd;
    }

    /**
     * lookup a PackageDocImpl by name.  Return null if not found.
     */
    static PackageDocImpl lookup(String packageName) {
        return (PackageDocImpl)packageMap.get(packageName);
    }

    void addClass(ClassDocImpl cd) {
        classes.add(cd);
    }

    /**
     * Get package name.
     */
    public String name() {
        return name;
    }

    /**
     * Do lazy initialization of "documentation" string.
     */
    String documentation() {
        if (documentation == null) {
            if (docPath == null) {
                // no doc file to be had
                documentation = "";
            } else {
                // read from file
                try {
                    documentation = readHTMLDocumentation(new FileInputStream(docPath), docPath);
                } catch (IOException exc) {            
                    documentation = "";
                    Res.error("javadoc.File_Read_Error", docPath);
                }
            }
        }
        return documentation;
    }
            
    /**
     * Add all included classes (including Exceptions and Errors) 
     * and interfaces.
     */
    void addAllClassesTo(List list) {
        for (Iterator it = classes.iterator(); it.hasNext(); ) {
            ClassDocImpl cls = (ClassDocImpl)it.next();
            if (cls.isIncluded) {
                list.add(cls);
            }
        }
    }
    
    /**
     * Get all classes (including Exceptions and Errors) 
     * and interfaces.
     *
     * @return all included classes and interfaces in this package.
     */
    public ClassDoc[] allClasses() {
        List ret = new ArrayList();
        for (Iterator it = classes.iterator(); it.hasNext(); ) {
            ClassDocImpl cls = (ClassDocImpl)it.next();
            if (cls.isIncluded) {
                ret.add(cls);
            }
        }
        return (ClassDoc[])ret.toArray(new ClassDoc[ret.size()]);
    }
    
    /**
     * Get ordinary classes (excluding Exceptions and Errors) 
     * in this package.
     *
     * @return included classes in this package.
     */
    public ClassDoc[] ordinaryClasses() {
        List ret = new ArrayList();
        for (Iterator it = classes.iterator(); it.hasNext(); ) {
            ClassDocImpl cls = (ClassDocImpl)it.next();
            if (cls.isIncluded && cls.isOrdinaryClass()) {
                ret.add(cls);
            }
        }
        return (ClassDoc[])ret.toArray(new ClassDoc[ret.size()]);
    }
    
    /**
     * Get Exception classes in this package.
     *
     * @return included Exceptions in this package.
     */
    public ClassDoc[] exceptions() {
        List ret = new ArrayList();
        for (Iterator it = classes.iterator(); it.hasNext(); ) {
            ClassDocImpl cls = (ClassDocImpl)it.next();
            if (cls.isIncluded && cls.isException()) {
                ret.add(cls);
            }
        }
        return (ClassDoc[])ret.toArray(new ClassDoc[ret.size()]);
    }

    /**
     * Get Error classes in this package.
     *
     * @return included Errors in this package.
     */
    public ClassDoc[] errors() {
        List ret = new ArrayList();
        for (Iterator it = classes.iterator(); it.hasNext(); ) {
            ClassDocImpl cls = (ClassDocImpl)it.next();
            if (cls.isIncluded && cls.isError()) {
                ret.add(cls);
            }
        }
        return (ClassDoc[])ret.toArray(new ClassDoc[ret.size()]);
    }
    
    /**
     * Get Interfaces in this package.
     *
     * @return included Errors in this package.
     */
    public ClassDoc[] interfaces() {
        List ret = new ArrayList();
        for (Iterator it = classes.iterator(); it.hasNext(); ) {
            ClassDocImpl cls = (ClassDocImpl)it.next();
            if (cls.isIncluded && cls.isInterface()) {
                ret.add(cls);
            }
        }
        return (ClassDoc[])ret.toArray(new ClassDoc[ret.size()]);
    }

    /**
     * Lookup for a class within this package.
     *
     * @return ClassDocImpl of found class, or null if not found.
     */
    public ClassDoc findClass(String className) {
        for (Iterator it = classes.iterator(); it.hasNext(); ) {
            ClassDoc cd = (ClassDoc)it.next();
            if (cd.name().equals(className)) {
                return cd;
            } 
        }
        return null;
    }
    
    /** 
     * return true if this package is include in the active set.
     */
    public boolean isIncluded() {
        return isIncluded;
    }

    /**
     * Returns a string representation of the package.
     *
     * @return the package name.
     */
    public String toString() {
        return name;
    }
}
