/*
 * @(#)RootDocImpl.java	1.24 98/06/30
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.tools.javadoc;

import com.sun.javadoc.*;

import sun.tools.java.Constants;
import sun.tools.java.ClassDefinition;
import sun.tools.java.ClassDeclaration;
import sun.tools.java.MemberDefinition;
import sun.tools.java.Identifier;

import sun.tools.util.ModifierFilter;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Enumeration;
import java.util.Locale;

import java.text.Collator;

import java.io.IOException;
import java.io.FileInputStream;


/**
 * This class holds the information from one run of javadoc.
 * Particularly the packages, classes and options specified
 * by the user..
 *
 * @since JDK1.2
 * @author Robert Field
 * @author Atul M Dambalkar
 */
class RootDocImpl extends DocImpl implements RootDoc, Constants {

    /**
     * list of classes specified on the command line.
     */
    private List cmdLineClasses;

    /**
     * list of packages specified on the command line.
     */
    private List cmdLinePackages;

    /**
     * a collection of all options.
     */
    List options;

    /**
     * The collator for this application. This is to take care of Locale 
     * Specific or Natural Language Text sorting.
     */
    static Collator collator = null;

    /**
     *  The locale to be used. If user doesen't provide this, then set it to
     *  default locale value.
     */
    static Locale locale = null;

    /**
     * The locale name will be set by Main, if option is provided on the
     * command line.
     */
    static String localeName = "";

    /**
     * Constructor
     *
     * initialize environment, classes, and packages information.
     * Classes and packages are those input from the command line.
     *
     * @param env the compilation environment
     * @param classes a list of ClassDeclaration
     * @param packages a list of package names (String)
     *
     */
    RootDocImpl(Env env, List classes, List packages, 
                ModifierFilter access, List options) {
        super(env, null);
        setLocale();
        this.options = options;
	showAccess = access;
        setClasses(env, classes);
        setPackages(env, packages);
        ClassDocImpl.secondPhaseConstruction(env);
        purgeMethodBodies(env);
    }


    /**
     * Set the locale for the application. Also set the collator to be used
     * depending upon this locale.
     */
    static void setLocale(String localename) {
        localeName = localename;
        setLocale();
    }

    static void setLocale() {
       if (locale == null) {
            locale = getLocale();
            if (locale == null) {  
                Main.exit();
            } else {
                Locale.setDefault(locale);
            }
        }
        if (collator == null) {
            collator = Collator.getInstance(locale);
        } 
    }
  
    /**
     * Get the locale if specified on the command line else return null and if
     * locale option is not used then return default locale.
     */
    static Locale getLocale() {
        Locale userlocale = null;
        if (localeName.length() > 0) {
            int firstuscore = localeName.indexOf('_'); 
            int seconduscore = -1;
            String language = null;
            String country = null;
            String variant = null;
            if (firstuscore == 2) {
                language = localeName.substring(0, firstuscore); 
                seconduscore = localeName.indexOf('_', firstuscore + 1); 
                if (seconduscore > 0) {
                    if (seconduscore != firstuscore + 3 || 
                           localeName.length() <= seconduscore + 1) {
                        Res.error("main.malformed_locale_name", localeName);
                        return null;
                    } 
                    country = localeName.substring(firstuscore + 1, 
                                                   seconduscore);
                    variant = localeName.substring(seconduscore + 1);
                } else if (localeName.length() == firstuscore + 3) {
                    country = localeName.substring(firstuscore + 1);
                } else {
                    Res.error("main.malformed_locale_name", localeName);
                    return null;
                } 
            } else if (firstuscore == -1 && localeName.length() == 2) {
                language = localeName; 
            } else {
                Res.error("main.malformed_locale_name", localeName);
                return null;
            } 
            userlocale = searchLocale(language, country, variant);
            if (userlocale == null) {
                Res.error("main.illegal_locale_name", localeName);
                return null;
            } else {
                return userlocale;
            }
        } else {
            return Locale.getDefault();
        }
    }

    /**
     * Search the locale for specified language, specified country and
     * specified variant.
     */
    static Locale searchLocale(String language, String country, 
                                                String variant) {
        Locale[] locales = Locale.getAvailableLocales();
        for (int i = 0; i < locales.length; i++) {
            if (locales[i].getLanguage().equals(language) &&
               (country == null || locales[i].getCountry().equals(country)) &&
               (variant == null || locales[i].getVariant().equals(variant))) {
                return locales[i];
            }
        }
        return null;
    }

    /**
     * Throw method bodies away. ("method" in this use
     * includes constructors).
     */
    private void purgeMethodBodies(Env env) {
        for (Enumeration e = env.getClasses() ; e.hasMoreElements() ;) {
            ClassDeclaration decl = (ClassDeclaration)e.nextElement();
            ClassDefinition def = decl.getClassDefinition();
            if (def != null) {
                MemberDefinition f;
                for (f = def.getFirstMember(); f != null; 
                            f = f.getNextMember()) { 
                    if (f.isMethod() && f.getValue() != null)
                        f.setValue(null);
                }
            }
        }
    }

    /**
     * Create a ClassDocImpl from a ClassDeclaration.
     *
     * @param env the compilation environment
     * @param decl a ClassDeclaration to convert
     * @return ClassDocImpl, if and only if defined already, is a source
     * class and is visible, otherwise return null.
     */
    private ClassDocImpl createClassDocImpl(Env env, ClassDeclaration decl) {
        ClassDocImpl cd = null;
        ClassDefinition def = decl.getClassDefinition();
        if (def != null && env.isSourceClass(def) && shouldDocument(def)) {
            cd = ClassDocImpl.getClassDocImpl(env, def);
        }
        return cd;
    }

    /**
     * initialize classes information. Those classes are input from
     * command line.
     *
     * @param env the compilation environment
     * @param classes a list of ClassDeclaration
     */
    private void setClasses(Env env, List classes) {
        cmdLineClasses = new ArrayList();
        for (Iterator iter = classes.iterator(); iter.hasNext();) {
            ClassDeclaration decl = (ClassDeclaration)iter.next();
            ClassDocImpl cd = createClassDocImpl(env, decl);

            if (cd != null) {
                cd.isIncluded = true;
                cmdLineClasses.add(cd);
            }
        }
    }

    /**
     * initialize packages information.
     *
     * @param env the compilation environment
     * @param packages a list of package names (String)
     */
    private void setPackages(Env env, List packages) {
        cmdLinePackages = new ArrayList();
        for (Iterator iter = packages.iterator() ; iter.hasNext() ;) {
            Identifier id = Identifier.lookup((String)iter.next());

            PackageDocImpl pd = PackageDocImpl.getPackageDocImpl(env, id.toString());
            pd.isIncluded = true;
            cmdLinePackages.add(pd);
        }

        for (Enumeration e = env.getClasses() ; e.hasMoreElements() ;) {
            ClassDeclaration decl = (ClassDeclaration)e.nextElement();
            createClassDocImpl(env, decl);
        }
    }

    /**
     *  Hook so we can allow unfiltered access, with a yet
     *  to be determined API change.
     */
    boolean checkAccess() {
        return true;
    }

    /**
     * Command line options. 
     *
     * <pre>
     * For example, given:
     *     javadoc -foo this that -bar other ...
     *
     * This method will return:
     *      options()[0][0] = "-foo"
     *      options()[0][1] = "this"
     *      options()[0][2] = "that"
     *      options()[1][0] = "-bar"
     *      options()[1][1] = "other"
     * </pre>
     *
     * @return an array of arrays of String.
     */
    public String[][] options() {
	return options(options);
    }

    static String[][] options(List optsList) {
        int cnt = optsList.size();
        String [][] opts = new String[cnt][];
        for (int i = 0; i < cnt; ++i) {
            List one = (List)optsList.get(i);
            opts[i] = (String[])one.toArray(new String[one.size()]);
        }
	return opts;
    }

    /**
     * packages specified on the command line.
     */
    public PackageDoc[] specifiedPackages() {
        return (PackageDoc[])cmdLinePackages
            .toArray(new PackageDoc[cmdLinePackages.size()]);
    }

    /**
     * classes and interfaces specified on the command line.
     */
    public ClassDoc[] specifiedClasses() {
        return (ClassDoc[])cmdLineClasses
            .toArray(new ClassDoc[cmdLineClasses.size()]);
    }

    /**
     * classes and interfaces to be documented.
     */
    public ClassDoc[] classes() {
        List classesToDocument = new ArrayList(cmdLineClasses);
        for (Iterator it = cmdLinePackages.iterator(); it.hasNext(); ) {
            ((PackageDocImpl)it.next()).addAllClassesTo(classesToDocument);
        }           
        return (ClassDoc[])classesToDocument
            .toArray(new ClassDoc[classesToDocument.size()]);
    }

    /**
     * Return the name of this Doc item.
     *
     * @return the string <code>"*RootDocImpl*"</code>.
     */
    public String name() {
        return "*RootDoc*";
    }

    /**
     * Do lazy initialization of "documentation" string.
     */
    String documentation() {
        if (documentation == null) {
            int cnt = options.size();
            String overviewPath = null;
            for (int i = 0; i < cnt; ++i) {
                List one = (List)options.get(i);
                if (((String)one.get(0)).equals("-overview")) {
                    overviewPath = (String)one.get(1);
                    break;
                }
            }
            if (overviewPath == null) {
                // no doc file to be had
                documentation = "";
            } else {
                // read from file
                try {
                    documentation = readHTMLDocumentation(new FileInputStream(overviewPath), overviewPath);
                } catch (IOException exc) {            
                    documentation = "";
                    Res.error("javadoc.File_Read_Error", overviewPath);
                }
            }
        }
        return documentation;
    }

    /**
     * Return a PackageDoc for the specified package name
     *
     * @param name package name
     *
     * @return a PackageDoc holding the specified package, null if 
     * this package is not referenced.
     */
    public PackageDoc packageNamed(String name) {
        return PackageDocImpl.lookup(name);
    }

    /**
     * Return a ClassDocImpl for the specified class/interface name
     *
     * @param qualifiedName qualified class name 
     *                        (i.e. includes package name).
     *
     * @return a ClassDocImpl holding the specified class, null if 
     * this class is not referenced.
     */
    public ClassDoc classNamed(String qualifiedName) {
        return ClassDocImpl.lookup(qualifiedName);
    }

    /** 
     * return true if this Doc is include in the active set.
     * RootDocImpl isn't even a program entity so it is always false.
     */
    public boolean isIncluded() {
        return false;
    }

    /**
     * Print error message, increment error count.
     *
     * @param msg message to print
     */
    public void printError(String msg) {
        Res.printError(msg);
    }    

    /**
     * Print warning message, increment warning count.
     *
     * @param msg message to print
     */
    public void printWarning(String msg) {
        Res.printWarning(msg);
    }    

    /**
     * Print a message.
     *
     * @param msg message to print
     */
    public void printNotice(String msg) {
        Res.printNotice(msg);
    }    
}     
