/*
 * @(#)SeeTagImpl.java	1.24 98/08/06
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.tools.javadoc;

import com.sun.javadoc.*;

import sun.tools.java.*;

import java.util.*;

/**
 * Represents a see also documentation tag.
 * The @see tag can be plain text, or reference a class or member.
 *
 * @version 06/10/97
 * @author Kaiyang Liu (original)
 * @author Robert Field (rewrite)
 * @author Atul M Dambalkar
 *
 */
class SeeTagImpl extends TagImpl implements SeeTag {

    /**
     * where of  where#what - i.e. the class name (may be empty)
     */
    String where;

    /**
     * what of  where#what - i.e. the member (may be null)
     */
    String what;

    PackageDoc referencedPackage;
    ClassDoc referencedClass;
    MemberDoc referencedMember;

    String label = "";

    SeeTagImpl(Doc holder, String name, String text) {
        super(holder, name, text);
        parseSeeString(holder);
        if (where != null) {
            ClassDocImpl container = null;
            if (holder instanceof MemberDocImpl) {
                container = ((MemberDocImpl)holder).thisClassDocImpl();
            } else if (holder instanceof ClassDocImpl) {
                container = (ClassDocImpl)holder;
            }
            findReferenced(holder, container);
        }
    }

    /**
     * get the class name part of @see, For instance,
     * if the comment is @see String#startsWith(java.lang.String) .
     *      This function returns String.
     * Returns null if format was not that of java reference.
     * Return empty string if class name was not specified..
     */
    public String referencedClassName() {
        return where;
    }
      
    /**
     * get the package referenced by  @see. For instance, 
     * if the comment is @see java.lang  
     *      This function returns a PackageDocImpl for java.lang
     * Returns null if no known package found.
     */
    public PackageDoc referencedPackage() {
	return referencedPackage;
    }

    /**
     * get the class referenced by the class name part of @see, For instance,
     * if the comment is @see String#startsWith(java.lang.String) .
     *      This function returns a ClassDocImpl for java.lang.String.
     * Returns null if class is not a class specified on the javadoc command line..
     */
    public ClassDoc referencedClass() {
	return referencedClass;
    }

    /**
     * get the name of the member referenced by the prototype part of @see,
     * For instance,
     * if the comment is @see String#startsWith(java.lang.String) .
     *      This function returns "startsWith(java.lang.String)"
     * Returns null if format was not that of java reference.
     * Return empty string if member name was not specified..
     */
    public String referencedMemberName() {
	return what;
    }

    /**
     * get the member referenced by the prototype part of @see,
     * For instance,
     * if the comment is @see String#startsWith(java.lang.String) .
     *      This function returns a MethodDocImpl for startsWith.
     * Returns null if member could not be determined.
     */
    public MemberDoc referencedMember() {
	return referencedMember;
    }


    /**
     * parse @see part of comment. Determine 'where' and 'what'
     */
    private void parseSeeString(Doc holder) {
        int len = text.length();
        if (len == 0) {
            return;
        }
        switch (text.charAt(0)) {
            case '<':
                if (text.charAt(len-1) != '>') {
                    Res.warning("tag.see.no_close_bracket_on_url",
                                holder.toString(), text);
                }
                return;
            case '"':
                if (len == 1 || text.charAt(len-1) != '"') {
                    Res.warning("tag.see.no_close_quote",
                                holder.toString(), text);
                } else {
//                    text = text.substring(1,len-1); // strip quotes
                }
                return;
        }

        // check that the text is one word, with possible parentheses
        // this part of code doesn't allow
        // @see <a href=.....>asfd</a>
        // comment it.

        // the code assumes that there is no initial white space.
        int parens = 0;
        int commentstart = 0;
        int start = 0;
        for (int i = start; i < len ; i++) {
            char ch = text.charAt(i);
            switch (ch) {
                case '(': parens++; break;
                case ')': parens--; break;
                case '[': case ']': case '.': case '#': break;
                case ',':
                    if (parens <= 0) {
                        Res.warning("tag.see.malformed_see_tag",
                                    holder.toString(), text);
                        return;
                    }
                    break;
                case ' ': case '\t': case '\n':
                    if (parens == 0) { //here onwards the comment starts.
                        commentstart = i;
                        i = len;
                    }
                    break;
                default:
                    if (!Character.isJavaIdentifierPart(ch)) {
                        Res.warning("tag.see.illegal_character",
                                    holder.toString(), ""+ch, text);
                        return;
                    }
                    break;
            }
        }
        if (parens != 0) {
            Res.warning("tag.see.malformed_see_tag",
                        holder.toString(), text);
            return;
        }

        String seetext = "";
        String labeltext = "";

        if (commentstart > 0) {
            seetext = text.substring(start, commentstart);
            labeltext = text.substring(commentstart + 1);
            // strip off the white space which can be between seetext and the
            // actual label.
            for (int i = 0; i < labeltext.length(); i++) {
                char ch = labeltext.charAt(i);
                if (!(ch == ' ' || ch == '\t' || ch == '\n')) {
                    label = labeltext.substring(i);
                    break;
                }
            }
        } else {
            seetext = text;
            label = "";
        }

        int sharp = seetext.indexOf('#');
        if (sharp >= 0) {
            // class#member
            where = seetext.substring(0, sharp);
	    what = seetext.substring(sharp + 1);
        } else {
            if (seetext.indexOf('(') >= 0) {
                Res.warning("tag.see.missing_sharp", holder.toString(), text);
                where = "";
		what = seetext;
            }
            else {
                // no member specified, text names class
                where = seetext;
		what = null;
            }
        }
    }

    /**
     * Find what is referenced by the see also.  If possible, sets
     * referencedClass and referencedMember.
     *
     * @param containingClass the class containing the comment containing
     * the tag. May be null, if, for example, it is a package comment.
     */
    private void findReferenced(Doc holder, ClassDocImpl containingClass) {
        if (where.length() > 0) {
            if (containingClass != null) {
                referencedClass = containingClass.findClass(where);
            } else {
                referencedClass = ClassDocImpl.lookup(where);
            }
            if (referencedClass == null) { /* may just not be in this run */
//                Res.warning("tag.see.class_not_found", holder.toString(), 
//                                                       where, text);
                // check if it's a package name 
                referencedPackage = PackageDocImpl.lookup(where);
                return;
            }
        } else {
            if (containingClass == null) {
                Res.warning("tag.see.class_not_specified",
                            holder.toString(), text);
                return;
            } else {
                referencedClass = containingClass;
            }
        }
        where = referencedClass.qualifiedName();

        if (what == null) {
            return;
        } else {
            int paren = what.indexOf('(');
            String memName = (paren >= 0 ? what.substring(0, paren) : what);
            String[] paramarr;
            if (paren > 0) {
                paramarr = new ParameterParseMachine(what.
                        substring(paren, what.length())).parseParameters();
                if (paramarr != null) {
                    referencedMember = findExecutableMember(memName, paramarr,
                                                            referencedClass);
                } else {
                    referencedMember = null;
                }
            } else {
                referencedMember = findExecutableMember(memName, null,
                                                        referencedClass);
                FieldDoc fd = ((ClassDocImpl)referencedClass).
                                                            findField(memName);
                if (referencedMember == null ||
                        (fd != null && fd.containingClass()
                            .subclassOf(referencedMember.containingClass()))) {
                    referencedMember = fd;
                }
            }
            if (referencedMember == null) {
                Res.warning("tag.see.can_not_find_member", holder.toString(),
                             what, where);
            }
        }
    }

    private MemberDoc findReferencedMethod(String memName, String[] paramarr,
                                           ClassDoc referencedClass) {
        MemberDoc meth = findExecutableMember(memName, paramarr, referencedClass);
        ClassDoc[] innerclasses = referencedClass.innerClasses();
        if (meth == null) {
   	    for (int i = 0; i < innerclasses.length; i++) {
                meth = findReferencedMethod(memName, paramarr, innerclasses[i]);
                if (meth != null) {
                    return meth;
                }
            }
        }
        return null;
    }
        
    private MemberDoc findExecutableMember(String memName, String[] paramarr,
                                           ClassDoc referencedClass) {
        if (memName.equals(referencedClass.name())) {
            return ((ClassDocImpl)referencedClass).findConstructor(memName, 
                                                                   paramarr);
        } else {   // it's a method.
            return ((ClassDocImpl)referencedClass).findMethod(memName, 
                                                              paramarr);
        }
    } 
 
    // separate "int, String" from "(int, String)"
    // (int i, String s) ==> [0] = "int",  [1] = String
    // (int[][], String[]) ==> [0] = "int[][]" // [1] = "String[]"
    class ParameterParseMachine {
        final int START = 0;
        final int TYPE = 1;
        final int NAME = 2;
        final int TNSPACE = 3;  // space between type and name
        final int ARRAYDECORATION = 4;
        final int ARRAYSPACE = 5;

        String parameters;

        StringBuffer typeId;

        List paramList;

        ParameterParseMachine(String parameters) {
            this.parameters = parameters;
            paramList = new ArrayList();
            typeId = new StringBuffer();
        }

        public String[] parseParameters() {
            if (parameters.equals("()")) {
                return new String[0];
            }   // now strip off '(' and ')'
            int state = START;
            int prevstate = START;
            parameters = parameters.substring(1, parameters.length() - 1);
            char[] chararr = parameters.toCharArray();
            for (int index = 0; index < chararr.length; index++) {
                char ch = chararr[index];
                switch (state) {
                    case START:
                        if (Character.isJavaIdentifierStart(ch)) {
                            typeId.append(ch);
                            state = TYPE;
                        }
                        prevstate = START;
                        break;
                    case TYPE:
                        if (Character.isJavaIdentifierPart(ch) || ch == '.') {
                            typeId.append(ch);
                        } else if (ch == '[') {
                            typeId.append(ch);
                            state = ARRAYDECORATION;
                        } else if (Character.isWhitespace(ch)) {
                            state = TNSPACE;
                        } else if (ch == ',') {  // no name, just type
                            addTypeToParamList();
                            state = START;
                        }
                        prevstate = TYPE;
                        break;
                    case TNSPACE:
                        if (Character.isJavaIdentifierStart(ch)) { // name
                            if (prevstate == ARRAYDECORATION) {
                                Res.warning("tag.missing_comma_space", 
                                             holder.toString(),
                                            "(" + parameters + ")");
                                return (String[])null;
                            }
                            addTypeToParamList();
                            state = NAME;
                        } else if (ch == '[') {
                            typeId.append(ch);
                            state = ARRAYDECORATION;
                        } else if (ch == ',') {   // just the type
                            addTypeToParamList();
                            state = START;
                        } // consume rest all
                        prevstate = TNSPACE;
                        break;
                    case ARRAYDECORATION:
                        if (ch == ']') {
                            typeId.append(ch);
                            state = TNSPACE;
                        } else if (!Character.isWhitespace(ch)) {
                            Res.warning("tag.illegal_char_in_arr_dim",
                                         holder.toString(),
                                        "(" + parameters + ")");
                            return (String[])null;
                        }
                        prevstate = ARRAYDECORATION;
                        break;
                    case NAME:
                        if (ch == ',') {  // just consume everything till ','
                            state = START;
                        }
                        prevstate = NAME;
                        break;
                }
            }
            if (state == ARRAYDECORATION ||
                (state == START && prevstate == TNSPACE)) {
                Res.warning("tag.illegal_see_tag", holder.toString(),
                            "(" + parameters + ")");
            }
            if (typeId.length() > 0) {
                paramList.add(typeId.toString());
            }
            return (String[])paramList.toArray(new String[0]);
        }

        void addTypeToParamList() {
            if (typeId.length() > 0) {
                paramList.add(typeId.toString());
                typeId.setLength(0);
            }
        }
    }

    /**
     * Return the kind of this tag.
     */
    public String kind() {
        return "@see";
    }

    /**
     * Return the label of the see tag.
     */
    public String label() {
        return label;
    }
}
