/*
 * @(#)SerialFieldTagImpl.java	1.7 98/06/29
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.tools.javadoc;

import com.sun.javadoc.*;

import java.util.*;

import sun.tools.java.*;

/**
 * Documents a Serializable field defined by an ObjectStreamField.
 * <pre>
 * The class parses and stores the three serialField tag parameters:
 *
 * - field name
 * - field type name
 *      (fully-qualified or visible from the current import context)
 * - description of the valid values for the field

 * </pre>
 * This tag is only allowed in the javadoc for the special member
 * serialPersistentFields.
 *
 * @author Joe Fialli
 *
 * @see java.io.ObjectStreamField
 */
class SerialFieldTagImpl extends TagImpl implements SerialFieldTag, Comparable {
    String   fieldName;     // Required Argument 1 of serialField
    String   fieldType;     // Required Argument 2 of serialField
    String   description;   // Optional Remaining Arguments of serialField.

    ClassDoc containingClass; // Class containing serialPersistentField member.
    ClassDoc fieldTypeDoc;  // ClassDocImpl of fieldType.
    FieldDocImpl matchingField; // FieldDocImpl with same name as fieldName.

   /* Constructor. */
   SerialFieldTagImpl(Doc holder, String name, String text) {
        super(holder, name, text);
        parseSerialFieldString(holder);

	if (holder instanceof MemberDocImpl) {
	    containingClass = ((MemberDocImpl)holder).containingClass();
	}
    }

    /*
     * The serialField tag is composed of three entities.
     *
     *   serialField  serializableFieldName serisliableFieldType
     *                 description of field.
     *
     * The fieldName and fieldType must be legal Java Identifiers.
     */
    private void parseSerialFieldString(Doc holder) {
        int len = text.length();

        // if no white space found
	/* Skip white space. */
	int inx = 0;
        for (; inx < len && Character.isWhitespace(text.charAt(inx)); inx++) ;

	/* find first word. */
	int first = inx;
	int last = inx;
	if (! Character.isJavaIdentifierStart(text.charAt(inx))) {
	    Res.warning("tag.serialField.illegal_character",
			holder.toString(), ""+ text.charAt(inx), text);
	    return;
	}

        for (; inx < len && Character.isJavaIdentifierPart(text.charAt(inx)); inx++);

	if (inx < len && ! Character.isWhitespace(text.charAt(inx))) {
	    Res.warning("tag.serialField.illegal_character",
			holder.toString(), ""+ text.charAt(inx), text);
	    return;
	}

        last = inx;
	fieldName = text.substring(first, last);

	/* Skip white space. */
        for (; inx < len && Character.isWhitespace(text.charAt(inx)); inx++) ;

	/* find second word. */
	first = inx;
	last = inx;

        for (; inx < len && ! Character.isWhitespace(text.charAt(inx)); inx++);
	if (inx < len && ! Character.isWhitespace(text.charAt(inx))) {
	    Res.warning("tag.serialField.illegal_character",
			holder.toString(), ""+ text.charAt(inx), text);
	    return;
	}
        last = inx;
	fieldType = text.substring(first, last);

	/* Skip leading white space. Rest of string is description for serialField.*/
        for (; inx < len && Character.isWhitespace(text.charAt(inx)); inx++) ;
	description = text.substring(inx);
    }

    /**
     * return a key for sorting.
     */
    String key() {
	return fieldName;
    }

    /*
     * Optional. Link this serialField tag to its corrsponding
     * field in the class. Note: there is no requirement that
     * there be a field in the class that matches serialField tag.
     */
    void mapToFieldDocImpl(FieldDocImpl fd) {
	matchingField = fd;
    }

    /**
     * Return the serialziable field name.
     */
    public String fieldName() {
        return fieldName;
    }

    /**
     * Return the field type string.
     */
    public String fieldType() {
        return fieldType;
    }

    /**
     * Return the ClassDocImpl for field type.
     *
     * @returns null if no ClassDocImpl for field type is visible from
     *          containingClass context.
     */
    public ClassDoc fieldTypeDoc() {
	if (fieldTypeDoc == null && containingClass != null) {
	    fieldTypeDoc = containingClass.findClass(fieldType);
	}
	return fieldTypeDoc;
    }

    /**
     * Return the corresponding FieldDocImpl for this SerialFieldTagImpl.
     *
     * @returns null if no matching FieldDocImpl.
     */
    FieldDocImpl getMatchingField() {
	return matchingField;
    }

    /**
     * Return the field comment. If there is no serialField comment, return
     * javadoc comment of corresponding FieldDocImpl.
     */
    public String description() {
	if (description.length() == 0 && matchingField != null) {

	    //check for javadoc comment of corresponding field.
	    Comment comment = matchingField.comment();
	    if (comment != null) {
		return comment.commentText();
	    }
	}
	return description;
    }

    /**
     * Return the kind of this tag.
     */
    public String kind() {
        return "@serialField";
    }

    /**
     * Convert this object to a string.
     */
    public String toString() {
        return name + ":" + text;
    }

    /**
     * Compares this Object with the specified Object for order.  Returns a
     * negative integer, zero, or a positive integer as this Object is less
     * than, equal to, or greater than the given Object.
     * <p>
     * Included to make SerialFieldTagImpl items java.lang.Comparable.
     *
     * @param   obj the <code>Object</code> to be compared.
     * @return  a negative integer, zero, or a positive integer as this Object
     *		is less than, equal to, or greater than the given Object.
     * @exception ClassCastException the specified Object's type prevents it
     *		  from being compared to this Object.
     * @since JDK1.2
     */
    public int compareTo(Object obj) {
	return key().compareTo(((SerialFieldTagImpl)obj).key());
    }
}
