/*
 * @(#)SerializedForm.java	1.9 98/07/13
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.tools.javadoc;

import com.sun.javadoc.*;

import sun.tools.java.ClassDefinition;
import sun.tools.java.Identifier;
import sun.tools.java.MemberDefinition;
import sun.tools.javac.SourceClass;

import java.util.ArrayList;
import java.util.List;

import java.lang.reflect.Modifier;


/**
 * The serialized form is the specification of a class' serialization
 * state. <p>
 *
 * It consists of the following information:<p>
 *
 * <pre>
 * 1. Whether class is Serializable or Externalizable.
 * 2. Javadoc for serialization methods.
 *    a. For Serializable, the optional readObject,writeObject.
 *       serialData tag describes, in prose, the sequence and type
 *       of optional data written by writeObject.
 *    b. For Externalizable, writeExternal and readExternal.
 *       serialData tag describes, in prose, the sequence and type
 *       of optional data written by writeExternal.
 * 3. Javadoc for serialization data layout.
 *    a. For Serializable, the name,type and description
 *       of each Serializable fields.
 *    b. For Externalizable, data layout is described by 2(b).
 * </pre>
 *
 * @since JDK1.2
 * @author Joe Fialli
 */
class SerializedForm {
    // List of MethodDocImpl - Serializable methods.
    private final List methods = new ArrayList();

    /* List of FieldDocImpl - Serializable fields.
     * Singleton list if class defines Serializable fields explicitly.
     * Otherwise, list of default serializable fields.
     * 0 length list for Externalizable.
     */
    private final List fields = new ArrayList();

    /* True if class specifies serializable fields explicitly.
     * using special static member, serialPersistentFields.
     */
    private boolean definesSerializableFields = false;

    // Specially treated field/method names defined by Serialization.
    private static final String SERIALIZABLE_FIELDS = "serialPersistentFields";
    private static final String READOBJECT  = "readObject";
    private static final String WRITEOBJECT = "writeObject";

    /**
     * Constructor.
     *
     * Catalog Serializable fields for Serializable class.
     * Catalog serialization methods for Serializable and
     * Externalizable classes.
     */
    SerializedForm(Env env, ClassDefinition def, ClassDocImpl cd) {
        if (cd.isExternalizable()) {
	    /* look up required public accessible methods,
	     *   writeExternal and readExternal.
	     */
            String[] readExternalParamArr = { "java.io.ObjectInput" };
            String[] writeExternalParamArr = { "java.io.ObjectOutput" };
	    MethodDoc md = cd.findMethod("readExternal", readExternalParamArr);
	    if (md != null) {
		methods.add(md);
            }
	    md = cd.findMethod("writeExternal", writeExternalParamArr);
	    if (md != null) {
		methods.add(md);
		Tag tag[] = md.tags("serialData");
	    }
	} else { // isSerializable()

	    MemberDefinition dsf = getDefinedSerializableFields(def);
	    if (dsf != null) {

		/* Define serializable fields with array of ObjectStreamField.
		 * Each ObjectStreamField should be documented by a
		 * serialField tag.
		 */
		definesSerializableFields = true;
		FieldDocImpl dsfDoc = FieldDocImpl.getFieldDocImpl(env, dsf);
		fields.add(dsfDoc);
		mapSerialFieldTagImplsToFieldDocImpls(dsfDoc, env, def);
            } else {

		/* Calculate default Serializable fields as all
		 * non-transient, non-static fields.
		 * Fields should be documented by serial tag.
		 */
		computeDefaultSerializableFields(env, def, cd);
	    }

	    /* Check for optional customized readObject & writeObject.
	     * writeObject and readObject can contain serialData tag.
	     */
	    addMethodIfExist(env, def, READOBJECT);
	    addMethodIfExist(env, def, WRITEOBJECT);
	}
    }

    /*
     * Check for explicit Serializable fields. 
     * Check for a private static array of ObjectStreamField with 
     * name SERIALIZABLE_FIELDS.
     */
    private MemberDefinition getDefinedSerializableFields(ClassDefinition def) {

	/* SERIALIZABLE_FIELDS can be private,
	 * so must lookup by ClassDefinition, not by ClassDocImpl.
	 */
	MemberDefinition f;
	for (f = def.getFirstMatch(Identifier.lookup(SERIALIZABLE_FIELDS));
	     f != null; f = f.getNextMember()) {
	    if (f != null && f.isStatic() && f.isVariable() && f.isPrivate()) {
		break;
	    }
	}
	return f;
    }

    /*
     * Compute default Serializable fields from all members of ClassDefinition.
     *
     * Since the fields of ClassDocImpl might not contain private or
     * package accessible fields, must walk over all members of ClassDefinition.
     */
    private void computeDefaultSerializableFields(Env env,
						  ClassDefinition def,
						  ClassDocImpl cd) {
	MemberDefinition f;
	for (f = def.getFirstMember(); f != null; f = f.getNextMember()) {
	    if (f.isVariable() && ! f.isTransient() && ! f.isStatic()) {
		FieldDocImpl fd = FieldDocImpl.getFieldDocImpl(env, f);
		fields.add(fd);
	    }
	}
    }

    /*
     * Catalog Serializable method if it exists in current ClassDefinition.
     * Do not look for method in superclasses.
     *
     * Serialization requires these methods to be private and
     * non-static.
     *
     * @param method should be an unqualified Serializable method
     *               name either READOBJECT or WRITEOBJECT.
     */
    private void addMethodIfExist(Env env, ClassDefinition def,
				  String methodName) {
	MemberDefinition md = null;
	for (md = def.getFirstMatch(Identifier.lookup(methodName));
	     md != null; md = md.getNextMember()) {
	    if (md.isMethod() && md.isPrivate() && ! md.isStatic()) {

		/*
		 * WARNING: not robust if unqualifiedMethodName is overloaded
		 *          method. Signature checking could make more robust.
		 * READOBJECT takes a single parameter, java.io.ObjectInputStream.
		 * WRITEOBJECT takes a single parameter, java.io.ObjectOutputStream.
		 */
		break;
	    }
	}
	if (md != null)
	    methods.add(MethodDocImpl.getMethodDocImpl(env, md));
    }

    /*
     * Associate serialField tag fieldName with FieldDocImpl member.
     * Note: A serialField tag does not have to map an existing field
     *       of a class.
     */
    private void mapSerialFieldTagImplsToFieldDocImpls(FieldDocImpl spfDoc,
					       Env env,
					       ClassDefinition def) {
	SerialFieldTag[] sfTag = spfDoc.serialFieldTags();
	for (int i = 0; i < sfTag.length; i++) {
	    Identifier fieldName = Identifier.lookup(sfTag[i].fieldName());

	    // Look for a FieldDocImpl that is documented by serialFieldTagImpl.
	    for (MemberDefinition md = def.getFirstMatch(fieldName);
		 md != null ; md = md.getNextMember()) {
		if (md.isVariable()) {
		    ((SerialFieldTagImpl)(sfTag[i])).mapToFieldDocImpl(FieldDocImpl.getFieldDocImpl(env, md));
		    break;
		}
	    }
	}
    }

    /**
     * Return serializable fields in class. <p>
     *
     * Returns either a list of default fields documented by serial tag comment or
     *         javadoc comment<p>
     * Or Returns a single FieldDocImpl for serialPersistentField. There is a
     *         serialField tag for each serializable field.<p>
     *
     * @return an array of FieldDocImpl for representing the visible
     *         fields in this class.
     */
    FieldDoc[] fields() {
        return (FieldDoc[])fields.toArray(new FieldDoc[fields.size()]);
    }

    /**
     * Return serialization methods in class.
     *
     * @return an array of MethodDocImpl for serialization methods in this class.
     */
    MethodDoc[] methods() {
        return (MethodDoc[])methods.toArray(new MethodDoc[methods.size()]);
    }

    /**
     * Returns true if Serializable fields are defined explicitly using
     * member, serialPersistentFields.
     *
     * @see #fields()
     */
    boolean definesSerializableFields() {
	return definesSerializableFields;
    }
}


