/*
 * @(#)TagImpl.java	1.8 98/07/09
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.tools.javadoc;

import com.sun.javadoc.*;

import java.util.*;

/**
 * Represents a documentation tag, e.g. @since, @author, @version.
 * Given a tag (e.g. "@since 1.2"), holds tag name (e.g. "@since") 
 * and tag text (e.g. "1.2").  TagImpls with structure or which require
 * special processing are handled by subclasses (ParamTagImpl, SeeTagImpl,
 * and ThrowsTagImpl
 * 
 * @author Robert Field
 * @author Atul M Dambalkar
 * @see SeeTagImpl
 * @see ParamTagImpl
 * @see ThrowsTagImpl
 * @see Doc#tags()
 *
 */
class TagImpl implements Tag {
    final String text;
    final String name;
    final Doc holder;

    /**
     *  Constructor
     */
    TagImpl(Doc holder, String name, String text) {
        this.holder = holder;
        this.name = name;
        this.text = text;
    }
    
    /**
     * Return the name of this tag.
     */
    public String name() {
        return name;
    }

    /**
     * Return the kind of this tag.
     */
    public String kind() {
        return name;
    }

    /**
     * Return the text of this tag, that is, portion beyond tag name.
     */
    public String text() {
        return text;
    }

    /**
     * for use by subclasses which have two part tag text.
     */
    String[] divideAtWhite() {
        String[] sa = new String[2];
        int len = text.length();
        // if no white space found
        sa[0] = text;
        sa[1] = "";
        for (int inx = 0; inx < len; ++inx) {
            char ch = text.charAt(inx);
            if (Character.isWhitespace(ch)) {
                sa[0] = text.substring(0, inx);
                for (; inx < len; ++inx) {
                    ch = text.charAt(inx);
                    if (!Character.isWhitespace(ch)) {
                        sa[1] = text.substring(inx, len);
                        break;
                    }
                }
                break;
            }
        }
        return sa;
    }            

    /**
     * convert this object to a string.
     */
    public String toString() {
        return name + ":" + text;
    }

    /**
     * For documentation comment with embedded @link tags, return the array of
     * TagImpls consisting of SeeTagImpl(s) and text containing TagImpl(s).
     * Within a comment string "This is an example of inline tags for a
     * documentaion comment {@link Doc {@link Doc commentlabel}}", 
     * where inside the inner braces, the first "Doc" carries exctly the same 
     * syntax as a SeeTagImpl and the second "commentlabel" is label for the Html 
     * Link, will return an array of TagImpl(s) with first element as TagImpl with 
     * comment text "This is an example of inline tags for a documentation 
     * comment" and second element as SeeTagImpl with referenced class as "Doc" 
     * and the label for the Html Link as "commentlabel".
     *
     * @return TagImpl[] Array of tags with inline SeeTagImpls.
     * @see ParamTagImpl
     * @see ThrowsTagImpl
     */
    public Tag[] inlineTags() {
        return Comment.getInlineTags(holder, text);
    }

    /**
     * Return array of tags for the first sentence in the doc comment text.
     */
    public Tag[] firstSentenceTags() {
        return Comment.firstSentenceTags(holder, text);
    }
}
