/*
 * @(#)TypeMaker.java	1.5 98/07/07
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.tools.javadoc;

import com.sun.javadoc.*;

import sun.tools.java.Constants;
import sun.tools.java.ClassDeclaration;

class ArrayType implements Type {

    /**
     * a dimension, like [][] = 2...
     */
    private final int dimension;

    /**
     * array of ...
     */
    private final Type type;

    ArrayType(Type type, int dimension) {
        this.type = type;
        this.dimension = dimension;
    }

    /**
     * Return the type's dimension information, as a string.
     * <p>
     * For example, a two dimensional array of String returns '[][]'.
     */
    public String dimension() {
        switch (dimension) {  // something I do all the time
            case 0: return "";
            case 1: return "[]";
            default: {
                StringBuffer sb = new StringBuffer();
                for (int cnt = 0; cnt < dimension; ++cnt) {
                    sb.append("[]");
                }
                return sb.toString();
            }
        }
    }

    /**
     * Return unqualified name of type excluding any dimension information.
     * <p>
     * For example, a two dimensional array of String returns 'String'.
     */
    public String typeName() {
        return type.typeName();
    }

    /**
     * Return qualified name of type excluding any dimension information.
     *<p>
     * For example, a two dimensional array of String
     * returns 'java.lang.String'.
     */
    public String qualifiedTypeName() {
        return type.qualifiedTypeName();
    }

    /**
     * Return this type as a class.  Array dimensions are ignored.
     *
     * @return a ClassDocImpl if the type is a Class.
     * Return null if it is a primitive type..
     */
    public ClassDoc asClassDoc() {
        return type.asClassDoc();
    }

    /**
     * Returns a hash code value for the object. This method is
     * supported for the benefit of hashtables such as those provided by
     * <code>java.util.HashMap</code> and <code>java.util.Hashtable</code>.
     *
     * @return  a hash code value for this object.
     * @see     #equals(java.lang.Object)
     * @see     java.util.HashMap
     * @see     java.util.Hashtable
     */
    public int hashCode() {
        return type.hashCode() + dimension;
    }

    /**
     * Compares two Objects for equality.
     * <p>
     * The <code>equals</code> method implements an equivalence relation.
     *
     * @param   obj   the reference object with which to compare.
     * @return  <code>true</code> if this object is the same as the obj
     *          argument; <code>false</code> otherwise.
     */
    public boolean equals(Object obj) {
        if (obj instanceof ArrayType) {
            ArrayType ot = (ArrayType)obj;
            return dimension == ot.dimension
                && type.equals(ot.type);
        } else {
            return false;
        }
    }

    /**
     * Returns a string representation of the type.
     *
     * Return name of type including any dimension information.
     * <p>
     * For example, a two dimensional array of String returns
     * <code>String[][]</code>.
     *
     * @return name of type including any dimension information.
     */
    public String toString() {
        return type.toString() + dimension();
    }
}

class PrimitiveType implements Type {

    private final String name;

    static final PrimitiveType voidType = new PrimitiveType("void");
    static final PrimitiveType booleanType = new PrimitiveType("boolean");
    static final PrimitiveType byteType = new PrimitiveType("byte");
    static final PrimitiveType charType = new PrimitiveType("char");
    static final PrimitiveType shortType = new PrimitiveType("short");
    static final PrimitiveType intType = new PrimitiveType("int");
    static final PrimitiveType longType = new PrimitiveType("long");
    static final PrimitiveType floatType = new PrimitiveType("float");
    static final PrimitiveType doubleType = new PrimitiveType("double");

    private PrimitiveType(String name) {
        this.name = name;
    }

    /**
     * Return unqualified name of type excluding any dimension information.
     * <p>
     * For example, a two dimensional array of String returns 'String'.
     */
    public String typeName() {
        return name;
    }

    /**
     * Return qualified name of type excluding any dimension information.
     *<p>
     * For example, a two dimensional array of String
     * returns 'java.lang.String'.
     */
    public String qualifiedTypeName() {
        return name;
    }

    /**
     * Return the type's dimension information, as a string.
     * <p>
     * For example, a two dimensional array of String returns '[][]'.
     */
    public String dimension() {
        return "";
    }

    /**
     * Return this type as a class.  Array dimensions are ignored.
     *
     * @return a ClassDocImpl if the type is a Class.
     * Return null if it is a primitive type..
     */
    public ClassDoc asClassDoc() {
        return null;
    }

    /**
     * Returns a string representation of the type.
     *
     * Return name of type including any dimension information.
     * <p>
     * For example, a two dimensional array of String returns
     * <code>String[][]</code>.
     *
     * @return name of type including any dimension information.
     */
    public String toString() {
        return typeName();
    }
}

/**
 * Constructs a class which represents a java type. Type can be
 * a class or primitive data type, like int, char...
 *
 * @since JDK1.2
 * @author Kaiyang Liu (original)
 * @author Robert Field (rewrite)
 */
class TypeMaker implements Constants {

    /**
     * Access for creating and returning types.
     */
    static Type getType(Env env, sun.tools.java.Type jt) {
        int dim = 0;
        while (jt.getTypeCode() == TC_ARRAY) {
            ++dim;
            jt = jt.getElementType();
        }
        Type typ = simpleType(env, jt);
        if (dim > 0) {
            return new ArrayType(typ, dim);
        } else {
            return typ;
        }
    }

    private static Type simpleType(Env env, sun.tools.java.Type jt) {
        switch (jt.getTypeCode()) {
            case TC_VOID:     return PrimitiveType.voidType;
            case TC_BOOLEAN:  return PrimitiveType.booleanType;
            case TC_BYTE:     return PrimitiveType.byteType;
            case TC_CHAR:     return PrimitiveType.charType;
            case TC_SHORT:    return PrimitiveType.shortType;
            case TC_INT:      return PrimitiveType.intType;
            case TC_LONG:     return PrimitiveType.longType;
            case TC_FLOAT:    return PrimitiveType.floatType;
            case TC_DOUBLE:   return PrimitiveType.doubleType;
            case TC_CLASS:
                ClassDeclaration decl = env.getClassDeclaration(jt);
                return ClassDocImpl.getClassDocImpl(env, decl);
        }
        return null;
    }
}
