/*
 * @(#)AudioAiffInputStream.java	1.21 98/07/22
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/**
 *
 * Used to manipulate streams from AIFF audio files.
 *
 * @version 1.21 98/07/22
 * @author Cania Lee, Kara Kytle, David Rivas
 */

package sun.audio;

import java.io.ByteArrayInputStream;

/**
 * read Aiff Header and set Audio Format.
**/

import java.io.InputStream;
import java.io.IOException;
import java.io.DataInputStream;
import java.io.SequenceInputStream;

class AudioAiffInputStream extends AudioContainerInputStream {

    private boolean DEBUG = false;
    
    /**
     * AudioAiffInputStream
     * Constructor for java.media.audio.AudioAiffInputStream.
     * Read data from any input stream and set Audio format.
     * Any bytes which have already been read from the stream 
     * are provided in an array of bytes.
     *
     * This should get called by ACIS.getStream().
     */
    AudioAiffInputStream(byte[] bytes, InputStream in)
	throws IOException, BadAudioHeaderException {

	super(bytes, in);
    }


    /**
     * readFormat()
     * read AIFF Header. 
     * @exception BadAudioHeaderException, IOException
     * @return AudioFormat instance
     */	
     protected AudioFormat readFormat(byte[] magicBytes)
	 throws BadAudioHeaderException, IOException {
	     
	 int sampleRate = -1; 
	 int channels = -1;
	 String encoding = null;
	 int sampleSize = -1;
	 int samplePerUnit = 1;
	 
	 int nread = 0;
	 int flags = 0;

	 boolean aifc = false;
	 byte bytes[] = new byte[4];
	 String s;
	 int chunksize, i ;
	 int bits = 0;

         // here we assemble the magicBytes + the input stream into
         // one InputStream

         DataInputStream dis;
	 if (magicBytes == null)  dis = new DataInputStream(in);
         else dis = new DataInputStream(new SequenceInputStream(new ByteArrayInputStream(magicBytes), in));

	 // Read the magic number
	 int magic = dis.readInt();
 	 nread += 4;

	 // Set Endian and SIGNED UNSSIGNED flags.
	 boolean isBigEndian = true;
	 
	 totalBytes = dis.readInt(); nread += 4;
	 dis.read(bytes, 0, 4); nread += 4;
	 s = new String( bytes );

	 // Is this an AIFC or just plain AIFF file.
	 if (s.startsWith("AIFC")) {
	     aifc = true;
	 }
	 if(DEBUG) {
	     System.err.println("This a " +
				((aifc) ? "AIFF-C" : "AIFF") +
				" stream.");
	 }

	 // Loop through the AIFF chunks until
	 // we get to the data.
	 while (true) {
	     
	     // Read the chunk name
	     dis.read(bytes, 0, 4 ); nread += 4;
	     s = new String( bytes );

	     if( DEBUG) {
		 System.err.println("Read a new AIFF chunk: \"" + s + "\"");
	     }
	     // Switch on the chunk name.
	     if (s.startsWith("FVER")) {
		 // Ignore format version for now.
		 dis.readLong(); nread += 8;
		 
	     } else if ( s.startsWith("COMM")) {
		 
		 chunksize = dis.readInt(); nread += 4;
		 
		 // AIFF vs. AIFFC
		 if (( chunksize != 18) && (chunksize != 34 )) {
		     throw new
			 BadAudioHeaderException("Invalid chunksize");
		 }

		 // Read header info.
		 channels = dis.readShort(); nread += 2;
		 dis.readInt(); nread += 4;
		 bits = dis.readShort(); nread += 2;
		 sampleRate = (int) read_ieee_extended (dis); nread += 10;

		 // If this is not AIFC then we assume it's
		 // a linearly encoded file.
		 encoding = AudioFormat.JAUDIO_LINEAR;
		 if (aifc) {
		     dis.read(bytes, 0, 4 ); nread += 4;
		     s = new String (bytes);
		     
		     if (s.startsWith("NONE")) 
			 encoding = AudioFormat.JAUDIO_LINEAR;
		     
		     else if ( s.startsWith("G722")) 
			 encoding = AudioFormat.JAUDIO_G722;
		     else if ( s.startsWith("ULAW"))
			 encoding = AudioFormat.JAUDIO_G711_ULAW;
		     else if ( s.startsWith("ALAW"))
			 encoding = AudioFormat.JAUDIO_G711_ALAW;
		     else 
			 throw new BadAudioHeaderException("Unknown encoding");
		 }
	     } else if ( s.startsWith("SSND")) {
		 // Data chunk.
                 // we are getting *weird* numbers for chunksize here sometimes;
                 // this really should be the size of the data chunk....
		 chunksize = dis.readInt(); nread += 4;

                 // Data offset
		 i = dis.readInt(); nread += 4;
		 // Block size
		 int blocksize = dis.readInt(); nread += 4;

                 // okay, now we are done reading the header.  we need to set the size
                 // of the data segment.  we know that sometimes the value we get for
                 // the chunksize is absurd.  this is the best i can think of:if the
                 // value seems okay, use it.  otherwise, we get our value of 
                 // totalBytes by assuming that everything left is the data segment;
                 // its length should be our original totalBytes (for all AIFF data chunks)
                 // minus what we've read so far.  
                 // $$kk: we should be able to get totalBytes for the data chunk right after
                 // we find "SSND."  however, some aiff files give *weird* numbers.  what
                 // is going on??

                 if (chunksize < totalBytes)  
                    totalBytes = chunksize;
                    else totalBytes -= nread;

		 break;
	     } else {
		 
		 // We're in no-mans land. Just read past
		 // the chunk.
		 chunksize = dis.readInt(); nread += 4;
                 nread += dis.skipBytes(chunksize);
		 //while (( -- chunksize ) >= 0 ) {
		 //    dis.readByte(); nread += 1;
		 //}
	     } // if
	 } // while


	 // Set the sample size & signed/unsigned.
	// $$kk: what's going on with the signed/unsigned bit here??
	 boolean isSigned = true;
	 switch (bits) {
	     case 8:
		 sampleSize = 1;
		 if(encoding == AudioFormat.JAUDIO_LINEAR) {
		     isSigned = false;
		 }
		 break;
	     case 16:
		 sampleSize = 2;
		 break;
	     default:
		 // $jdr: What should we do here?
	 }

	 /*
	  * $$jdr: This is totaly wrong. The data is NOT 16bit linear
	  * until the conversion. ONLY change the format there.
	  *
	  * We are going to convert the ULAW input into 16bit
	  * linear in setData(). so, set the sampleSize appropriately here.
	  *
	  * $$kk: 01.21.97: should not need this faked sampleSize anymore
	  */
//	 if (encoding == AudioFormat.JAUDIO_G711_ULAW) {
//	     sampleSize = 2;
//	 }
			 

	// set the headerSize; data should begin this number of bytes from the beginning of the file								
	headerSize = nread;

	 return (new AudioFormat(sampleRate, encoding,
				 sampleSize, samplePerUnit, channels,
				 isBigEndian, isSigned));
     }


    /**
     * read_ieee_extended
     * Extended precision IEEE floating-point conversion routine.
     * @argument DataInputStream
     * @return double
     * @exception IOException
     */
    private double read_ieee_extended(DataInputStream dis) throws IOException {

	double f = 0;
        int expon = 0;
        long hiMant = 0, loMant = 0;
        long t1, t2;
	double HUGE = ((double)3.40282346638528860e+38);


	expon = dis.readUnsignedShort();
	
	t1 = (long)dis.readUnsignedShort();
	t2 = (long)dis.readUnsignedShort();
	hiMant = t1 << 16 | t2; 
	
	t1 = (long)dis.readUnsignedShort();
	t2 = (long)dis.readUnsignedShort();
	loMant = t1 << 16 | t2; 

	if (expon == 0 && hiMant == 0 && loMant == 0) {
	    f = 0;
	} else {
	    if (expon == 0x7FFF)
		f = HUGE;
	    else {
		expon -= 16383;
		expon -= 31;
		f = (hiMant * Math.pow(2, expon));
		expon -= 32;
		f += (loMant * Math.pow(2, expon));
	    }
	}

        return f;
    }
}
