/*
 * @(#)AudioAuInputStream.java	1.25 98/07/22
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.audio;

/**
 * Au file format audio streams.
 *
 * @version 1.25 98/07/22
 * @author Cania Lee, Kara Kytle, David Rivas
 */

import java.io.InputStream;
import java.io.IOException;
import java.io.DataInputStream;
import java.io.ByteArrayInputStream;
import java.io.SequenceInputStream;

class AudioAuInputStream extends AudioContainerInputStream {

    static final int AU_ULAW_8       = 1;  /* 8-bit ISDN u-law */
    static final int AU_LINEAR_8     = 2;  /* 8-bit linear PCM */
    static final int AU_LINEAR_16    = 3;  /* 16-bit linear PCM */
    static final int AU_LINEAR_24    = 4;  /* 24-bit linear PCM */
    static final int AU_LINEAR_32    = 5;  /* 32-bit linear PCM */
    static final int AU_FLOAT        = 6;  /* 32-bit IEEE floating point */
    static final int AU_DOUBLE       = 7;  /* 64-bit IEEE floating point */
    static final int AU_ADPCM_G721   = 23; /* 4-bit CCITT g.721 ADPCM */
    static final int AU_ADPCM_G722   = 24; /* CCITT g.722 ADPCM */
    static final int AU_ADPCM_G723_3 = 25; /* CCITT g.723 3-bit ADPCM */
    static final int AU_ADPCM_G723_5 = 26; /* CCITT g.723 5-bit ADPCM */
    static final int AU_ALAW_8       = 27; /* 8-bit ISDN A-law */


    /**
     * AudioAuInputStream
     * Constructor for java.media.audio.AudioAuInputStream.
     * Read data from any input stream and set Audio format.
     * Any bytes which have already been read from the stream
     * are provided in an array of bytes.
     *
     * This should get called by ACIS.getStream().
     */
    AudioAuInputStream(byte[] bytes, InputStream in)
	throws IOException, BadAudioHeaderException {
		super(bytes, in);
    }


    /**
     * readFormat()
     * read Au Header.
     * @exception BadAudioHeaderException, IOException
     * @return AudioFormat instance
     */
    protected AudioFormat readFormat(byte[] bytes)
	throws BadAudioHeaderException, IOException {

        /* in is a protected InputStream variable
	 * defined in FilterInputStream.
	 * It is set by the constructor in FilterInputStream.
	 */

       int sampleRate = -1;
       int SamplePerUnit = 1;
       int channels = -1;
       String encoding = null;
       int sampleSize = -1;

       int nread = 0;
       int flags = 0;
       int encoding_local;

        // here we assemble the magicBytes + the input stream into
        // one InputStream

        DataInputStream dis;
        if (bytes == null)  dis = new DataInputStream(in);
        else dis = new DataInputStream(new SequenceInputStream(new ByteArrayInputStream(bytes), in));

        // Read the header.

        // Read the magic number (it's '.snd')
        int magic = dis.readInt();
        nread += 4;

       // What about the other ones?  $$kk: what??
       boolean isBigEndian = false;
       if ((magic == AU_SUN_MAGIC) ||
	   (magic == AU_DEC_MAGIC))
	   isBigEndian = true;

       // this is the dataLocation;
		int dataLocation = isBigEndian ? dis.readInt() : rllong(dis) ; nread += 4;

	   	
       // this is the dataSize in bytes, not including the header
       totalBytes = isBigEndian ? dis.readInt() : rllong(dis) ; nread += 4;

       // read the format type
       encoding_local = isBigEndian ? dis.readInt() : rllong(dis) ; nread += 4;

       // read the sample rate
       sampleRate = isBigEndian ? dis.readInt() : rllong(dis) ; nread += 4;

       // read the number of channels
       channels = isBigEndian ? dis.readInt() : rllong(dis) ; nread += 4;


       // Convert the AU encodings to JS ones.
       // and set signed, note that for linear8
       // data is unsigned.
       boolean isSigned = true;
       switch (encoding_local) {
	   case AU_ULAW_8:
	       encoding = AudioFormat.JAUDIO_G711_ULAW;
	       sampleSize = 1;
	       break;
	   case AU_ALAW_8:
	       encoding = AudioFormat.JAUDIO_G711_ALAW;
	       sampleSize = 1;
	       break;
	   case AU_LINEAR_8:
	       encoding = AudioFormat.JAUDIO_LINEAR;
	       sampleSize = 1;
	       isSigned = false;
	       break;
	   case AU_LINEAR_16:
	       encoding = AudioFormat.JAUDIO_LINEAR;
	       sampleSize = 2;
	       break;
	   case AU_LINEAR_24:
	       encoding = AudioFormat.JAUDIO_LINEAR;
	       sampleSize = 3;
	       break;
	   case AU_LINEAR_32:
	       encoding = AudioFormat.JAUDIO_LINEAR;
	       sampleSize = 4;
	       break;
	   case AU_FLOAT:
	       encoding = AudioFormat.JAUDIO_FLOAT;
	       sampleSize = 4;
	       break;
	   case AU_DOUBLE:
	       encoding = AudioFormat.JAUDIO_DOUBLE;
	       sampleSize = 8;
	       break;
	   case AU_ADPCM_G721:
	       encoding = AudioFormat.JAUDIO_G721_ADPCM;
	       sampleSize = 1;
	       /*
		* TOTAL HACK THAT NEEDS TO BE FIXED
		* We are GOING to be setting the size to 2,
		* when we decode.
		* If any body askes us for the format,
		* they had then better
		* use the GetData method to deal with us, otherwise they
		* will be getting different conversions.
		*/
	       sampleSize = 2;
	       /*		SamplePerUnit = 2; */
	       break;
	   case AU_ADPCM_G723_3:
	       encoding = AudioFormat.JAUDIO_G723_3;
	       sampleSize = 3;
	       SamplePerUnit = 8;
	       break;
	   case AU_ADPCM_G723_5:
	       encoding = AudioFormat.JAUDIO_G723_5;
	       sampleSize = 5;
	       SamplePerUnit = 8;
	       break;
	   default:
	       throw new
		   BadAudioHeaderException("Unsupported encoding");
       }

       // now we skip anything in the header that we haven't read.
       // we are not supporting the case of fragmented data here,
       // but we are going to use the dataLocation value to figure
       // out if we need to skip over anything before we get to the
       // actual data block.  we need to do this because we don't know
       // the size of the info block (at least 4 bytes).

       if (nread < dataLocation)
	        nread += dis.skipBytes(dataLocation - nread);

		// set the headerSize; data should begin this number of bytes from the beginning of the file
		// in the case of AU files, we should get this value from the dataLocation specified in the header.
		headerSize = dataLocation;

       return ( new AudioFormat(sampleRate, encoding, sampleSize,
				SamplePerUnit, channels,
				isBigEndian, isSigned));
    }
}
