/*
 * @(#)AudioClipData.java	1.3 98/06/29
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.audio;

/**
 * The AudioClipData holds an AudioContainerInputStream
 * used for data that can be used for an audio clip (i.e. a file).
 * The data may be cached or streamed.
 *
 * @version 1.3 98/06/29
 * @author Kara Kytle
 * @author David Rivas
 */


import java.net.URL;
import java.io.ByteArrayInputStream;
import java.io.IOException;


class AudioClipData {


	// a clip can be created from a URL or a byte array.
	// if it's created from a URL, this field is non-null
	URL url = null;

	// do we cache the data?  for an AudioClipData from a byte array, obviously yes.
	// can cache or not when creating from a URL.
	boolean isCached = false;

	// input stream we read audio data from.  we would actually like this class to
	// just subclass ACIS, but we can't satisfy the constructor requirement for a
	// FilterInputStream (must construct from an InputStream);
	AudioContainerInputStream acis = null;


    AudioClipData(byte[] data) throws BadAudioHeaderException, IOException {

		// get our ACIS of the right type.
		// after this, acis.stream is a BAIS
		// this can throw a BadAudioHeaderException or an IOException
		acis = AudioContainerInputStream.getStream(new ByteArrayInputStream(data));

		// make sure we got the data we need
		if (acis.getTotalFrames() == 0)
			throw new BadAudioHeaderException("incomplete header information for creating an AudioClipData: no data chunk size found.");

		isCached = true;
	}


    AudioClipData(URL url, boolean isCached) throws BadAudioHeaderException, IOException {

		// get our ACIS of the right type.
		// this can throw a BadAudioHeaderException or an IOException
		acis = AudioContainerInputStream.getStream(url.openStream());

		// make sure we got the data we need
		if (acis.getTotalFrames()  == 0)
			throw new BadAudioHeaderException("incomplete header information for creating an AudioClipData: no data chunk size found.");

		this.url = url;
		this.isCached = isCached;

		if (isCached)
			acis.setData();	// this can throw an IOException
	}


	void seek(int framePosition) throws IOException, HaeException {

		if (framePosition > acis.getTotalFrames())
			throw new HaeException("requested seek position " + framePosition + " exceeds total frame length " + acis.getTotalFrames());

		// frame size in bytes
		AudioFormat f = acis.getInputFormat();
		int frameSize = f.getSampleSize() * f.getChannels();

		int currentFrame = acis.getCurrentFrame();

		if (framePosition > currentFrame) {

			long bytesToSkip = ( framePosition - currentFrame) * frameSize;
			long bytesSkipped = skipBytes(bytesToSkip);

		} else {

			if (acis.stream instanceof ByteArrayInputStream) {

				// this may be a clip created from a byte array or a CACHED clip created
				// from a URL.

				acis.stream.reset();

				// skip over the header part at the beginning of the array, if it's there
				// this is sort of an awkward hack; setData() gets called on CACHED clips
				// created from a URL, so we have a ByteArrayInputStream that does NOT contain
				// the header.  if the clip was created from a ByteArrayInputStream, though, the
				// header is still in our byte array....

				if (url == null)
					skipBytes(acis.headerSize);

			} else {

				// this is an UNCACHED clip created from a URL

				acis.stream.close();
				acis.stream = url.openStream(); // can throw IOException
				skipBytes(acis.headerSize);
			}

			long bytesToSkip = framePosition * frameSize;
			long bytesSkipped = skipBytes(bytesToSkip);
		}
		acis.setCurrentFrame(framePosition);
	}


	long skipBytes(long bytesToSkip) throws IOException {

		long bytesSkipped = 0;
		while (bytesSkipped < bytesToSkip) {
			bytesSkipped += acis.stream.skip(bytesToSkip - bytesSkipped);
		}
		return bytesSkipped;
	}
}

