/*
 * @(#)AudioContainerInputStream.java	1.22 98/08/10
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.audio;

/**
 * AudioContainerInputStream
 * An AudioContainerInputStream is an abstract base class for
 * containers of audio data.
 * @see AudioWavInputStream
 * @see AudioAiffInputStream
 * @see AudioAuInputStream
 * @see AudioMidiInputStream
 * @see AudioRmfInputStream
 *
 * @version 1.22 98/08/10
 * @author Cania Lee
 * @author Kara Kytle
 * @author David Rivas
*/

/*
 * this class provides these things:
 * 1) the format(s) of the audio data
 * 2) the data itself, in the form of a byte array
 * 3) a way to read the data
 *		a) a way to read the header to get format information
 *		b) how to read the data from the original input stream
 *                 (setData)
*/


import java.io.InputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;


class AudioContainerInputStream extends MediaInputStream {

    protected boolean DEBUG = false  /*true*/ ;

    // Magic Numbers for Container Types.
    // If you're going to be supported, you'd better be on this list!
    // the magic number for each format is the value of the first
    // int read from a valid stream
    static final int MIDI_MAGIC         = 1297377380;
    static final int AIFF_MAGIC         = 1179603533;
    static final int RMF_MAGIC          = 1230128474;
    static final int RIFF_MAGIC         = 1380533830;
    static final int WAV_MAGIC          = 1463899717;
    static final int AU_SUN_MAGIC       = 779316836;
    static final int AU_DEC_MAGIC       = 779314176;
    static final int AU_SUN_INV_MAGIC   = 1684960046;
    static final int AU_DEC_INV_MAGIC   = 6583086;
    
    
    /** Used to convert ulaw format to linear format (hi-byte). */
    final static int ULAW_TABH[] = { 
	130, 134, 138, 142, 146, 150, 154, 158,
	162, 166, 170, 174, 178, 182, 186, 190,
	193, 195, 197, 199, 201, 203, 205, 207,
	209, 211, 213, 215, 217, 219, 221, 223,
	224, 225, 226, 227, 228, 229, 230, 231,
	232, 233, 234, 235, 236, 237, 238, 239,
	240, 241, 241, 242, 242, 243, 243, 244,
	244, 245, 245, 246, 246, 247, 247, 248,
	248, 248, 249, 249, 249, 249, 250, 250,
	250, 250, 251, 251, 251, 251, 252, 252,
	252, 252, 252, 252, 253, 253, 253, 253,
	253, 253, 253, 253, 254, 254, 254, 254,
	254, 254, 254, 254, 254, 254, 254, 254,
	255, 255, 255, 255, 255, 255, 255, 255,
	255, 255, 255, 255, 255, 255, 255, 255,
	255, 255, 255, 255, 255, 255, 255,   0,
	126, 121, 117, 113, 109, 105, 101,  97,
	 93,  89,  85,  81,  77,  73,  69,  65,
	 62,  60,  58,  56,  54,  52,  50,  48,
	 46,  44,  42,  40,  38,  36,  34,  32,
	 31,  30,  29,  28,  27,  26,  25,  24,
	 23,  22,  21,  20,  19,  18,  17,  16,
	 15,  14,  14,  13,  13,  12,  12,  11,
	 11,  10,  10,   9,   9,   8,   8,   7,
	  7,   7,   6,   6,   6,   6,   5,   5,
	  5,   5,   4,   4,   4,   4,   3,   3,
	  3,   3,   3,   3,   2,   2,   2,   2,
	  2,   2,   2,   2,   1,   1,   1,   1,
	  1,   1,   1,   1,   1,   1,   1,   1,
	  0,   0,   0,   0,   0,   0,   0,   0,
	  0,   0,   0,   0,   0,   0,   0,   0,
	  0,   0,   0,   0,   0,   0,   0,   0
    };

    /** Used to convert ulaw format to linear format (low-byte). */
    final static int ULAW_TABL[] = { 
	  0,   4,   8,  12,  17,  21,  25,  29,
	 33,  38,  42,  46,  50,  55,  59,  63,
	 66,  68,  71,  73,  75,  77,  79,  81,
	 83,  85,  88,  90,  92,  94,  96,  98,
	228, 229, 230, 231, 232, 233, 234, 235,
	236, 237, 238, 239, 240, 241, 243, 244,
	180,  53, 181,  54, 182,  55, 184,  56,
	185,  57, 186,  58, 187,  59, 188,  60,
	157, 221,  29,  94, 158, 222,  30,  95,
	159, 223,  31,  96, 160, 224,  32,  97,
	145, 177, 209, 241,  17,  50,  82, 114,
	146, 178, 210, 242,  18,  51,  83, 115,
	139, 155, 171, 187, 203, 219, 235, 251,
	 11,  28,  44,  60,  76,  92, 108, 124,
	136, 144, 152, 160, 168, 176, 184, 192,
	200, 208, 216, 224, 232, 240, 248,   0,
	  0, 252, 248, 244, 239, 235, 231, 227,
	223, 218, 214, 210, 206, 201, 197, 193,
	190, 188, 185, 183, 181, 179, 177, 175,
	173, 171, 168, 166, 164, 162, 160, 158,
	 28,  27,  26,  25,  24,  23,  22,  21,
	 20,  19,  18,  17,  16,  15,  13,  12,
	 76, 203,  75, 202,  74, 201,  72, 200,
	 71, 199,  70, 198,  69, 197,  68, 196,
	 99,  35, 227, 162,  98,  34, 226, 161,
	 97,  33, 225, 160,  96,  32, 224, 159,
	111,  79,  47,  15, 239, 206, 174, 142,
	110,  78,  46,  14, 238, 205, 173, 141,
	117, 101,  85,  69,  53,  37,  21,   5,
	245, 228, 212, 196, 180, 164, 148, 132,
	120, 112, 104,  96,  88,  80,  72,  64,
	 56,  48,  40,  32,  24,  16,   8,   0
    };


   /** Audio format object */
    AudioFormat f;

	/** 
	 * Audio format of output data.
	 * This is different than f, the format of the underlying data, if we are converting from ulaw
	 */
	 AudioFormat fout;

	 /** true if we're done reading the stream */
	 private boolean streamFinished = false;
	 

	/** Information on data conversions we'll need to do based on the format of the ACIS */
    boolean processBytes = false;
    boolean swapBytes = false;
    boolean convertFromUlaw = false;
    boolean convertUnsigned = false;
	boolean mixerBigEndian = false;

   /** Input stream based on the byte array */
    InputStream stream;

    static final int MAX_HEADER_SIZE = 12;     // max number of bytes we'll read
                                        // here before passing header-parsing
                                        // control to an ACIS subclass....

    /** bytes of audio data (not including the header */
	protected int totalBytes = 0; 

	/** frames of audio data */
    protected int totalFrames = 0; 

	/** current read position in the audio data */
	protected int currentFrame = 0;

	/** header size, in BYTES.  need this for resetting read position for a HaeAudioStreamClip */
	protected int headerSize = 0;
	

    /**
     * AudioContainerInputStream
     * Constructor for java.media.audio.AudioContainerInputStream.
     * Read data from any input stream and set Audio format.
     * Any bytes which have already been read from the stream 
     * are provided in an array of bytes.
     *
     * This should get called by one of the subclasses
     * in response to ACIS.getStream().
     */

    protected AudioContainerInputStream(byte[] bytes, InputStream in)
	throws IOException, BadAudioHeaderException {
	
		super(in);
		stream = in;

		// read the header and fill in the format var
		AudioFormat format = readFormat(bytes);
		setFormat(format);
		setupConversions();
    }


    /**
     * AudioContainerInputStream
     * Constructor for java.media.audio.AudioContainerInputStream.
     * Read data from any input stream and set Audio format.
     * The format of the data in the stream is specified in
     * the AudioFormat provided.
     *
     * This should get called by one of the subclasses in response
     * to ACIS.getStream().
     */
    AudioContainerInputStream(AudioFormat format, InputStream in) throws BadAudioHeaderException {

		super(in);
		stream = in;

		if (format == null || format.getEncoding() == null)
			throw new BadAudioHeaderException("invalid format object");

		setFormat(format);
		setupConversions();
    }

	
    protected void setFormat(Format theFormat) {
	
		if (theFormat instanceof AudioFormat) {

            // original format
			f = (AudioFormat)theFormat;			 
			totalFrames = getTotalBytes() / (f.getChannels() * f.getSampleSize());
		}

		else if (theFormat != null) /* may be null for MIDI and RMF */ /*throw new IOException("not a valid audio format!");*/
			System.out.println("not a valid audio format!");
    }


	/** Set up the conversion vars */
	protected void setupConversions() throws BadAudioHeaderException {

			// this can happen for midi and rmf
			if (f == null) return;


			// check the encoding and throw an exception for any we don't support
			if ( (f.getEncoding() != AudioFormat.JAUDIO_LINEAR) && (f.getEncoding() != AudioFormat.JAUDIO_G711_ULAW) ) {
				throw new BadAudioHeaderException("Unsupported audio encoding: " + f.getEncoding());
			}


			// swap bytes?

			try {
				HaeMixer mixer = HaeDevice.getSystemMixer();
				if (mixer != null)
					mixerBigEndian = mixer.isBigEndian();
			} catch (HaeException e) { 
				throw new BadAudioHeaderException("Mixer not initialized"); // $$kk: this is bad
			}

 			if( f.isBigEndian() == mixerBigEndian) {
				swapBytes = false;
			} else {
				swapBytes = (f.getSampleSize() == 1) ? false : true;
			}

			// sign conversions? 

			if( !f.isSigned() &&
				(f.getEncoding() == AudioFormat.JAUDIO_LINEAR)) {
				convertUnsigned = true;
			}
			
			// convert from ulaw?

			if( f.getEncoding().equals(AudioFormat.JAUDIO_G711_ULAW)) {

				// Can't deal with stereo ulaw.
				// $$jdr: O.k. we could, but not today.
				if( f.getChannels() != 1) {
				throw new BadAudioHeaderException ("Stereo ulaw");
				}
				convertFromUlaw = true;
			}

			if( swapBytes || convertUnsigned || convertFromUlaw) {
				processBytes = true;
				
				// setup the output format
				fout =	new AudioFormat(f.getSampleRate(), (convertFromUlaw ? AudioFormat.JAUDIO_LINEAR : f.getEncoding()), 
						(convertFromUlaw ? 2: f.getSampleSize()), f.getSamplePerUnit(),
						f.getChannels(), mixerBigEndian, (convertUnsigned ? false : f.isSigned()));

			} else fout = f;

			if (DEBUG) {
				System.out.println("format conversions: processBytes: " + processBytes + " swapBytes: " + swapBytes + "  convertUnsigned: " + convertUnsigned + " convertFromUlaw: " + convertFromUlaw);
				System.out.println("input format: " + f.toString());
				System.out.println("output format: " + fout.toString());
			}
	}


	
	/**
	 * read data from the stream int byte array.
	 * do format conversion if necessary.
	 * return the number of frames of valid data provided.
	 */
	int getConvertedData(byte[] data, int frames) throws IOException {

	// this can happen for midi and rmf
	if (f == null) 
		return 0;

	// this can happen if we're in getWavedata at the end of the stream, with 
	// no frames to read.
	if ( (data == null) || (frames <= 0) ) 
		return 0;

	// don't overrun the array.
	// $$kk: inputSampleFrameSize and outputSampleFrameSize had better not
	// be 0 or we divide by 0....
	int inputSampleFrameSize = f.getSampleSize() * f.getChannels();
	int outputSampleFrameSize = fout.getSampleSize() * fout.getChannels();

	if ((frames * outputSampleFrameSize) > data.length)
		frames = data.length / outputSampleFrameSize;

	// this is the number of bytes of data we can generate
	int audioSize = frames * outputSampleFrameSize;
	int bytesLeft = audioSize;
	int offset = 0;


	// if we're converting from ulaw, we read half the eventual number of bytes in to
	// the second half of the buffer; then we expand into the whole buffer during the
	// ulaw-to-linear conversion.
	if (convertFromUlaw == true) {
		offset = audioSize / 2;
		bytesLeft /= 2;
	}

	int bytesRead = 0;
	int totalBytesRead = 0;

	// this will read the audio data into the audioData array.  if we're converting from
	// ulaw, we read into the last half of the array.

	do {                
		bytesRead = stream.read(data, (int)offset, (int)bytesLeft);

		if (bytesRead <= 0) 
			break;

	    totalBytesRead += bytesRead;
		offset += bytesRead;
		bytesLeft -= bytesRead;
            
	} while( (bytesLeft != 0) && (stream.available() != 0) );

	// if we hit the end of the stream, mark it done
	if (bytesRead == -1)
		streamFinished = true;

	// if we got no data, return 
	if (totalBytesRead == 0)
		return 0;

	// now do any data conversions we need to do.
	if (processBytes) {

		int byteCount = 0;

		// If we need to swap them
		if(swapBytes) {

			while (byteCount < audioSize) {

			    byte firstByte = data[byteCount];
			    byte secondByte = data[byteCount+1];
			    
			    // Translate the bytes.
			    if( convertUnsigned) {
				firstByte = (firstByte >=0) ?
				    (byte)(0x80 | firstByte) :
				    (byte)(0x7F & firstByte);
				secondByte = (secondByte >=0) ?
				    (byte)(0x80 | secondByte) :
				    (byte)(0x7F & secondByte);
			    }
				        
			    // Do the swap.
			    data[byteCount] = secondByte;
			    data[byteCount + 1] = firstByte;
			            
			    // Next sample.
			    byteCount += 2;

			} // while
		} // swap bytes


		// If we need to convert from ulaw
		else if(convertFromUlaw) {

			// work through the second half of the byte array

			boolean bigEOut = mixerBigEndian;
			int lTab[] = AudioContainerInputStream.ULAW_TABL;
			int hTab[] = AudioContainerInputStream.ULAW_TABH;
			offset = audioSize / 2;

			while (offset < audioSize) {

			    data[byteCount] = (bigEOut) ?
				(byte)hTab[data[(int)offset] & 0xFF] :
				(byte)lTab[data[(int)offset] & 0xFF];
			    data[byteCount + 1] = (bigEOut) ?
				(byte)lTab[data[(int)offset] & 0xFF] :
				(byte)hTab[data[(int)offset] & 0xFF];
			            
			    // Next sample.
			    byteCount += 2;
				offset++;

			} // while
		} // convert from ulaw

		// If we need to convert unsigned
		else if(convertUnsigned) {

			while (byteCount < audioSize) {

				byte tempByte = data[byteCount];
				data[byteCount] = (tempByte >= 0) ?
				    (byte)(0x80 | tempByte) :
				    (byte)(0x7F & tempByte);

			    // Next sample.
			    byteCount++;

			} // while
		} // convert unsigned


	} // processBytes


	int framesRead = totalBytesRead / inputSampleFrameSize;
	updateCurrentFrame(framesRead);
	return framesRead;

	} // getConvertedData


    public Format[] getFormat() {

		Format[] format = { f };
		return format;
    }

    
    int getTotalBytes() {
			
		return totalBytes;
    }	

	int getTotalFrames() {

		return totalFrames;
	}

	int getCurrentFrame() {

		return currentFrame;
	}

	int getRemainingFrames() {

		return totalFrames - currentFrame;
	}


	// we should be either keeping track of the frames somewhere else or overriding the
	// input stream methods to do it; this is a hack to let us set the current frame
	// after skipping....
	void setCurrentFrame(int currentFrame) {

		this.currentFrame = currentFrame;
		
		if (currentFrame < getTotalFrames())
			streamFinished = false;
	}


	protected void updateCurrentFrame(int delta) {

		currentFrame += delta;
		int frames = getTotalFrames();


		if ( (frames > 0) && (currentFrame >= frames) ) {
			streamFinished = true;
		}
	}


    /**
     * Return the data as an array of bytes.
     * Returns null if this is not possible (i.e. any time the stream
     * is not a ByteArrayInputStream).
     */

     // $$kk: maybe should throw an exception if cannot return bytes??

    byte[] getBytes() {

        if (!(stream instanceof ByteArrayInputStream)) return null;

        ByteArrayInputStream bais = (ByteArrayInputStream)stream;	
        byte[] buffer = null;
        
        synchronized(stream) {

            int initAvail = bais.available();
            bais.reset();
            int totalAvail = bais.available();
            buffer = new byte[totalAvail];
            bais.read(buffer, 0, buffer.length);
            bais.reset();
            bais.skip(totalAvail - initAvail);            
        }
    
        return(buffer);
    }

    /**
     * Read the data from the input stream and into a byte array
     * This will not work on large files, but makes it easy to
     * support AudioData, and AudioDataStreams.
     * 
     * Note that the usual behaviour is to read all the data
     * (totalBytes' worth) and then close the stream.  We want
     * to allow for the case where we may not know the length
     * of the data: we should 1) know we failed setData and 
     * 2) *not* close the stream.  If (totalBytes ==0) =>
     * has not been set or is unknowable, we throw an IOException.
	 *
	 * $$kk: 07.22.98: many au files have -1 for totalLength.  i
	 * think that all the derivatives of this class can use the same
	 * setData() method, which reads totalBytes of data if the value
	 * is > 0; otherwise, it just reads until the end of the stream
	 * as it has always done for midi/rmf.  i am changing this method
	 * to combine the code from AudioMidiInputStream and all the
	 * sampled audio streams (which used the same code).
     */  
    boolean setData() throws IOException {

		// if we have a valid value for the totalBytes, read the data here.
		// otherwise, read data until the end of stream.

		if (totalBytes > 0) {

			return readTotalBytes();

		} else {

			return readAllBytes();	
		}
	}

	/**
	 * Read all bytes into a byte array until we read -1 from the stream.
	 */
	boolean readAllBytes() throws IOException {

	ByteArrayOutputStream ba  = new ByteArrayOutputStream();
	DataOutputStream      dos = new DataOutputStream(ba);
	byte[] buffer;
        int readCount = MAX_READ_LIMIT;
	byte tmp[]    = new byte[readCount];
	int num       = 0;
	int readByte  = 0;

        if (stream instanceof ByteArrayInputStream)
            return true;

	try {
	    
	    while (true) {
        	readByte = in.read(tmp, 0, readCount);
		if (readByte == -1) 
		    if (num == 0) {
			totalBytes = 0;
			return(false);
		    } else {
			break;
		    }
		
		dos.write(tmp, 0, readByte);
		num += readByte;
		Thread.currentThread().yield();
	    } // while
	    
	    in.close();
	    buffer = ba.toByteArray();

	} catch (IOException e) {
	    e.printStackTrace();
	    totalBytes = 0;
	    throw new IOException(e.getMessage());
	}
	
	totalBytes = num;
	stream = new ByteArrayInputStream(buffer);
	return(true);
    }


	/**
	 * Read totalBytes bytes into a byte array from the stream.
	 */
	boolean readTotalBytes() throws IOException {

		// Get a new buffer and copy it into the array.

		byte[] buffer = new byte[totalBytes];
		int gotbytes = readFully(buffer, 0, buffer.length);
		close();
		stream = new ByteArrayInputStream(buffer);
		return true;

	}


    /**
     * Position in the input stream of the beginning of the data segment
     */

    /* 
     * For Audio files, the data segment directly
     * follows the header information. After creating an
     * AudioContainerInputStream, the pointer points to 
     * the beginning of the data segment.
     */
    public long getDataPos() {
		return 0L;
    }

	boolean streamIsFinished() {

		return streamFinished;
	}

	

    /**
     * Read the audio header information, and then set AudioFormat fields.
     * Each subclass needs to have its own special way of accomplishing this
     * for a particular format.  The base class does nothing.
     */
     protected AudioFormat readFormat(byte[] bytes)
	 throws BadAudioHeaderException, IOException {

         /* 
         //   implement this in your derived class! 
         //   this used to be abstract; it should never be called
         //   directly; you should only instantiate an ACIS directly
         //   if you already have the format....
         */

         throw new BadAudioHeaderException("cannot read format in ACIS base class");
     }

    /*
     * Gets the appropriate subclass for this audio data stream.
     * The format is determined from header information read from
     * the beginning of the stream.
     *
     * Note that if this method is called with an object that contains
     * an ACIS (AudioStream, AudioDataStream, ContinuousAudioDataStream,
     * NativeAudioStream), that stream is extracted and returned.
     *
     * @throw	BadAudioHeaderException thrown if format data for
     * a supported audio format cannot be obtained from the stream.
     * @throw	IOException
     *
     */
    static AudioContainerInputStream getStream(InputStream in)
	throws IOException, BadAudioHeaderException {
	    
	// extract the ACIS if it exists already
	if (in instanceof AudioContainerInputStream) {
	    return (AudioContainerInputStream)in;
	}

	if (in instanceof AudioStream) {
	    return ((AudioStream) in).acis;
	}
	if (in instanceof AudioDataStream) {
	    return ((AudioDataStream) in).acis;
	}
	if (in instanceof NativeAudioStream) {
	    return ((NativeAudioStream) in).acis;
	}

	// okay, there is no ACIS to be found!  we will create one....
	//
	// here we get the container type tag.
	// it's just the first int from the stream
	//
	// we mark the stream because if we fail to get an
	// audio format we can use out of it, we want to put
	// the data we read back if possible
	// (for many streams, mark may not work)
	
	in.mark(MAX_HEADER_SIZE);
	byte[] bytes = new byte[4]; // size of int in bytes	

	// if we fail to recognize the format, we try to put any read bytes back; 
	// if we can't, we return them in the BadAudioHeaderException

	int bytesRead = in.read(bytes);

	// make bytes into an integer
	int firstInt = (bytes[0] << 24) + (bytes[1] << 16) +
	    (bytes[2] << 8) + (bytes[3] << 0);

	// when we create the subclass, we pass the array of bytes read
	// so that we don't throw the data away (need it at least for
	// au).  we're not going to play with reset() here; just have
	// one way of dealing with the read bytes.
	AudioContainerInputStream acis;
	switch(firstInt) {
		
	    case MIDI_MAGIC:
		acis = new AudioMidiInputStream(bytes, in);
		break;
		    
	    case AIFF_MAGIC:
		acis = new AudioAiffInputStream(bytes, in);
		break;

	    case RMF_MAGIC:
		acis = new AudioRmfInputStream(bytes, in);
		break;

	    case AU_SUN_MAGIC:
	    case AU_DEC_MAGIC:
	    case AU_SUN_INV_MAGIC:
	    case AU_DEC_INV_MAGIC:
		acis = new AudioAuInputStream(bytes, in);
		break;

	    case RIFF_MAGIC:
                byte[] moreBytes = new byte[MAX_HEADER_SIZE];
                System.arraycopy(bytes, 0, moreBytes, 0, bytes.length);
                bytesRead += in.read(moreBytes, bytesRead, 8);
                int nextInt = (moreBytes[8] << 24) + (moreBytes[9] << 16) +
	            (moreBytes[10] << 8) + (moreBytes[11] << 0);
                if (nextInt == WAV_MAGIC) {
		    acis = new AudioWavInputStream(moreBytes, in);
                    break;
                }
                else bytes = moreBytes;
                // now just fall through to the default case


	    default:	// it was something we couldn't handle
		try { 
		    in.reset();  
		    throw new
			BadAudioHeaderException("Unknown header");
		} catch (IOException e) {
		    throw new BadAudioHeaderException("Unknownheader",
							  bytes);
		}
	} // container type switch.

	return acis;	
    }

    /*
     * uLaw to linear conversion
     */
    byte[] ulawToLinear(byte[] ulaw) {

	byte linear[] = new byte[ulaw.length*2];
	for (int i = 0, j = 0; i < ulaw.length; i++, j += 2) {
	    linear[j]   = (byte) (ULAW_TABH[ulaw[i] & 0xFF]);
	    linear[j+1] = (byte) (ULAW_TABL[ulaw[i] & 0xFF]);
	}
	return linear;
    }



	/** return the input format from the data source, before conversions */
	AudioFormat getInputFormat() {

		return f;
	}

	/** return the output format to the engine, after conversions */
	AudioFormat getOutputFormat() {

		return fout;
	}

 
    /*
	 * $$kk: 07.22.98: moved rlshort and rllong from AudioWavInputStream here
	 * because we may need them for reading little-endian AU files as well.
	 */

    /** 
     * rllong 
     * Helper method to read 64 bits and changing the order of 
     * each bytes. 
     * @param DataInputStream 
     * @return 32 bits swapped value. 
     * @exception IOException
     */
    protected int rllong(DataInputStream dis) 
	throws IOException {

        int b1, b2, b3, b4 ;
        int i = 0;
 
       	i = dis.readInt();
 
       	b1 = ( i & 0xFF ) << 24 ;
       	b2 = ( i & 0xFF00 ) << 8;
       	b3 = ( i & 0xFF0000 ) >> 8;
       	b4 = ( i & 0xFF000000 ) >>> 24;
 
       	i = ( b1 | b2 | b3 | b4 );
 
        return i;
    }
	
	/**
     * rlshort
     * Helper method to read 16 bits value. Swap high with low byte.
     * @param DataInputStream
     * @return the swapped value.
     * @exception IOException
     */ 
    protected short rlshort(DataInputStream dis) 
	throws IOException {

	short s=0; 
	short high, low;

	s = dis.readShort();
	
	high = (short)(( s & 0xFF ) << 8) ;
	low = (short)(( s & 0xFF00 ) >>> 8);
	
	s = (short)( high | low );
	
	return s;
    }

    /**
     * Make the debug info print out.
     */
    void setVerbose(boolean v) {
	DEBUG = v;
    }
}
