/*
 * @(#)AudioDevice.java	1.34 98/08/10
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.audio;

import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import sun.audio.HaeAudioStream;
import sun.audio.HaeDevice;
import sun.audio.HaeMixer;
import sun.audio.HaeException;
import sun.audio.HaeMidiSong;
import sun.audio.HaeRmfSong;

/**
 * This class provides an interface to the Headspace Audio engine.
 *
 * This class emulates systems with multiple audio channels, mixing
 * multiple streams for the workstation's single-channel device.
 *
 * @see AudioData
 * @see AudioDataStream
 * @see AudioStream
 * @see AudioStreamSequence
 * @see ContinuousAudioDataStream
 * @author David Rivas
 * @author Kara Kytle
 * @version 1.34 98/08/10
 */

public class
AudioDevice {

    private boolean DEBUG = false  /*true*/ ;

    /** Path to the HeadSpace General MIDI soundbank file. */
    //private String soundBank = null;

    /** Hashtable of audio clips / input streams. */
    private Hashtable clipStreams;

    /** Are we currently playing audio? */
    private boolean playing = false;

    /** Handle to the HAE audio mixer. */
    private HaeMixer mixer = null;

    private int rate        = /* HaeMixer.SAMPLE_RATE_22K */
	                      HaeMixer.SAMPLE_RATE_44K
	                      ;
    private int interpMode  = /* HaeMixer.INTERPOLATION_LINEAR */
	                      HaeMixer.INTERPOLATION_2_POINT
	                      ;
    private int renderFlags = 0
	                      | HaeMixer.FLAGS_USE_STEREO
	                      | HaeMixer.FLAGS_USE_16
	                      ;


    /**
     * The default audio player. This audio player is initialized
     * automatically.
     */
    public static final AudioDevice device = new AudioDevice();

    /**
     * Create an AudioDevice instance.
     */
	private AudioDevice() {

		clipStreams = new Hashtable();
	}

    /**
     *  Open an audio channel.
     */
    public synchronized void openChannel(InputStream in) {

		// is this already playing?  if so, then just return
		if (clipStreams.get(in) != null) {
			if (DEBUG) {
				System.err.println("AudioDevice.openChannel(" +
					in + "): channel already open");
			}
		return;
		}


		// We use, exclusively AudioContainerInputStreams
		AudioContainerInputStream hin;

		// Make sure we have one.
		try {
			hin = AudioContainerInputStream.getStream(in);
		} catch (IOException e) {
			if (DEBUG) {
				System.err.println("AudioDevice.openChannel(" +
					in + "): could not create stream");
			}
			return;
		}


		// RMF & Midi Streams get turned into
		// songs. Otherwise we assume it's one of the rawdata streams
		// and create an AudioStream for it.
		try {

			// Songs and streams are both playables.
			HaePlayable playStream;

			// Is it a song or not?
			if( hin instanceof AudioRmfInputStream ||
			hin instanceof AudioMidiInputStream) {

				// Create the right kind of song ...
				HaeMidiSong song;
				song = (hin instanceof AudioRmfInputStream) ?
					new HaeRmfSong(mixer) : new HaeMidiSong(mixer);

		/*		// $jdr: Should put in a seperate
				// hack for setting up the soundbank.
				// Should only do this the first time.

				mixer.setSoundbank(soundBank);
		*/

				// ... open the song and set up the state ....
				song.open(hin.getBytes());


						//song.setLoopFlag(hin.loop);
				song.setLoopFlag(in instanceof ContinuousAudioDataStream);



				if( DEBUG) {
					String ls = (in instanceof ContinuousAudioDataStream) ?"Will" : "Will not";
					System.err.println(ls + " loop this song.");
				}

				playStream = song;

				// Set the format to render with if this
				// is the only clip (first clip).
				/* $$jdr: This isn't working properly in the
				 * HAE code.
				 */
				if(clipStreams.size() == 0) {
					rate = HaeMixer.SAMPLE_RATE_22K;
				}

			} else {
				// It's not a song, so it must be a stream.
				try {
					playStream = new HaeAudioStream(hin, in.toString());
				} catch (InvalidAudioFormatException e) {
					if (DEBUG) {
						System.err.println("AudioDevice.openChannel(" +
							in + "): could not create stream");
					}
					return;
				}

			// If this is the frist clip then set the
			// rate to whatever the rate of the stream is.
			/* $$jdr: This isn't working properly in the HAE code.
			if(clipStreams.size() == 0) {

				// Get the stream format and set the rate.
				int  streamRate = rate;
				Format f[] = (Format[])hin.getFormat();
				AudioFormat streamFormat = (AudioFormat)f[0];
				if( streamFormat != null) {
					streamRate = streamFormat.getSampleRate();
				}

				if ( streamRate <= 11025) {
					rate = HaeMixer.SAMPLE_RATE_11K;
				} else if(streamRate <= 22050) {
					rate = HaeMixer.SAMPLE_RATE_22K;
				} else {
					rate = HaeMixer.SAMPLE_RATE_44K;
				}
			}
			*/
		}

	    // ... Start the thing up, and keep it for later reference.

	    // $$kk: 02.03.98: only if we can get the audio device!
		// adding the clipStream element will cause open() to be
		// called again if play() is running, but i think that's
		// innocuous.  we need to make sure we're not adding and
		// adding streams when there's no device or we end up with
		// an OutOfMemory error (bug #4093603)!
		open();
		if (playing) {
			playStream.start();
			clipStreams.put(in, playStream);
		}


	} catch (HaeException e) {
	    if (DEBUG) {
		System.err.println("AudioDevice: " +
				   "Problem opening an new stream: "  + e);
	    }
	}


	// don't forget adjust for a new stream.
	notify();
    }


    /**
     *  Close an audio channel.
     */
    public synchronized void closeChannel(InputStream in) {

	if (in == null) return;		// can't go anywhere here!

	HaeNoise clip;

	// look for the stream in the hash table ....
	if ((clip = (HaeNoise) clipStreams.remove(in)) != null) {
	    if (DEBUG) {
		System.out.println("AudioDevice.closeChannel(): " +
				   "retrieved and closing stream:\n\t" + in);
	    }

	    // ... shut down the sound ...
	    clip.close();

	} else if (DEBUG) {
	    System.out.println("AudioDevice.closeChannel(): " +
			       "could not retrieve stream(" + in + ")");
	}

	// Tell the harvester we've gotten rid of one.
	notify();
    }


    /**
     * Open the device (done automatically)
     */
    public synchronized void open() {

		// $$kk: 02.03.98: we just try once to open the device.
		// fail quietly unless we're in debug mode.

		if (DEBUG) {
			System.out.println("AudioDevice.open(), playing = " + playing);
		}

		// we're going to try to open the device regardless of whether we think
		// playing is true, because other users of the device (jmf) can screw us
		// over.

		try {

			// if we don't have a mixer, we get one
			if (mixer == null) mixer = HaeDevice.getSystemMixer();

			// now we need to explicitly engage the device;
			// this will throw an HaeException if it fails.
			mixer.reengageAudio();

			// We got it.
			playing = true;

		} catch (HaeException hae) {

			if (DEBUG) {
				System.out.println("Failed to open audio device.");
			}

			if (DEBUG) {
				System.out.println("HaeException opening device: " + hae);
				hae.printStackTrace();
			}

			// Shut everything down if we couldn't get the device.
			closeStreams();
			playing = false;
		}


/*
		if (playing == false) {

			try {

				// if we don't have a mixer, we get one
				if (mixer == null) mixer = HaeDevice.getSystemMixer();

				// now we need to explicitly engage the device;
				// this will throw an HaeException if it fails.
				mixer.reengageAudio();

				// We got it.
				playing = true;

			} catch (HaeException hae) {

				if (DEBUG) {
					System.out.println("Failed to open audio device.");
				}

				if (DEBUG) {
					System.out.println("HaeException opening device: " + hae);
					hae.printStackTrace();
				}

				// Shut everything down if we couldn't get the device.
				closeStreams();
				return;
			}

		}
*/
	}


    /**
     * Close the device (done automatically)
     */
	public synchronized void close() {

		if (DEBUG) {
			System.err.println("Closing device.");
		}

		if (playing) {
			mixer.disengageAudio();
		}

		playing = false;
		notify();
    }


    /**
     * Wait for data
     */
    private synchronized boolean waitForData() throws InterruptedException {

		if (clipStreams.size() == 0) {

			// Shut down the device.
			close();

			// Wait for something to play
			do {

				if( DEBUG ) {
					System.err.println("Waiting for updates.");
				}

				wait();

			} while( clipStreams.size() == 0);

			// Reopen the device.
			open();
			return true;
		} else {
			wait();
		}
		return false;
	}


    /**
     * Play open audio stream(s)
     */
    public void play() {

		// Holdover from the old architecture.
		while (playing) {
			try {
				waitForData();
			} catch (InterruptedException e) {
				// the thread got interrupted, exit
			}
		}

		// if we exit play(), audio from the AudioPlayer is done.  there's no way to
		// get it back using AudioPlayer

		if (DEBUG) {
			System.out.println("exiting play()");
		}
	}

    /**
     * Close streams
     */
    public synchronized void closeStreams() {
	Enumeration e, k;

	if (DEBUG) {
	    System.err.println("AudioDevice: Closing all streams.");
	}

	// go throw the table of open streams
	// and close them down.
        for (e = clipStreams.elements(), k = clipStreams.keys();
	     e.hasMoreElements();) {
            ((HaeNoise) e.nextElement()).close();
	}

	if (DEBUG) {
	    System.err.println("Audio Device: Streams all closed.");
	}
	// Empty the hash table.
	clipStreams = new Hashtable();
    }

    /**
     * Number of channels currently open.
     */
    public int openChannels() {
	return clipStreams.size();
    }

    /**
     * Make the debug info print out.
     */
    void setVerbose(boolean v) {
	DEBUG = v;
    }
}
