/*
 * @(#)AudioPlayer.java	1.48 98/07/14
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.audio;

import java.util.Vector;
import java.util.Enumeration;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.FileOutputStream;

/**
 * This class provides an interface to play audio streams.
 *
 * To play an audio stream use:
 * <pre>
 *	AudioPlayer.player.start(audiostream);
 * </pre>
 * To stop playing an audio stream use:
 * <pre>
 *	AudioPlayer.player.stop(audiostream);
 * </pre>
 * To play an audio stream from a URL use:
 * <pre>
 *	AudioStream audiostream = new AudioStream(url.openStream());
 *	AudioPlayer.player.start(audiostream);
 * </pre>
 * To play a continuous sound you first have to
 * create an AudioData instance and use it to construct a
 * ContinuousAudioDataStream.
 * For example:
 * <pre>
 *	AudioData data = new AudioStream(url.openStream()).getData();
 *	ContinuousAudioDataStream audiostream = new ContinuousAudioDataStream(data);
 *	AudioPlayer.player.start(audiostream);
 * </pre>
 *
 * @see AudioData
 * @see AudioDataStream
 * @see AudioDevice
 * @see AudioStream
 * @author Arthur van Hoff, Thomas Ball
 * @version 	1.31, 05/22/97
 */

public
class AudioPlayer extends Thread {

    private AudioDevice devAudio;
    
    /**
     * The default audio player. This audio player is initialized
     * automatically.
     */
    public static final AudioPlayer player = getAudioPlayer();

    private static ThreadGroup getAudioThreadGroup() {

	ThreadGroup g = currentThread().getThreadGroup();
	while ((g.getParent() != null) &&
	       (g.getParent().getParent() != null)) {
	    g = g.getParent();
	}
	return g;
    }

    /**
     * Create an AudioPlayer thread in a privileged block.
     */

    private static AudioPlayer getAudioPlayer() {

		return (AudioPlayer)
		AudioSecurity.doPrivileged(
			new AudioSecurityAction() {
			public Object run() {
			return new AudioPlayer();
			}
		}, "UniversalThreadGroupAccess");
    }

    /**
     * Construct an AudioPlayer.
     */
    private AudioPlayer() {
	
	super(getAudioThreadGroup(), "Audio Player");
	devAudio = AudioDevice.device;

	try {
		AudioSecurity.doPrivileged(
			    new AudioSecurityAction() {
		    public Object run() {
			setPriority(MAX_PRIORITY);
			setDaemon(true);
			return null;
		    }
		}, "UniversalThreadAccess");		

	} catch (IllegalArgumentException e) {  
	    //System.out.println("exception creating AudioPlayer: " + e);
	} catch (SecurityException e2) {
	    //System.out.println("exception creating AudioPlayer: " + e);
        }

	devAudio.open();
	start();
    }

    /**
     * Start playing a stream. The stream will continue to play
     * until the stream runs out of data, or it is stopped.
     * @see AudioPlayer#stop
     */
    public synchronized void start(InputStream in) {
	
	devAudio.openChannel(in);
	notify();
    }

    /**
     * Stop playing a stream. The stream will stop playing,
     * nothing happens if the stream wasn't playing in the
     * first place.
     * @see AudioPlayer#start
     */
    public synchronized void stop(InputStream in) {
	
	devAudio.closeChannel(in);
    }

    /**
     * Main mixing loop. This is called automatically when the AudioPlayer
     * is created.
     */
    public void run() {

	devAudio.play();
	System.out.println("AudioPlayer exited; audio device may not be accessible.");

	// $$kk: 02.03.98: should not need this; AudioDevice closes the device itself
	// devAudio.close();
    }
}
