/*
 * @(#)AudioWavInputStream.java	1.25 98/08/10
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.audio;

/**
 * Process a .wav stream.
 *
 * @version 1.25 98/08/10
 * @author Cania Lee, Kara Kytle, David Rivas
 */

import java.io.InputStream;
import java.io.IOException;
import java.io.DataInputStream;
import java.io.ByteArrayInputStream;
import java.io.SequenceInputStream;

class AudioWavInputStream extends AudioContainerInputStream {

    static final int WAVE_FORMAT_UNKNOWN   = 0x0000;
    static final int WAVE_FORMAT_PCM       = 0x0001;
    static final int WAVE_FORMAT_ADPCM     = 0x0002;
    static final int WAVE_FORMAT_ALAW      = 0x0006;
    static final int WAVE_FORMAT_MULAW     = 0x0007;
    static final int WAVE_FORMAT_OKI_ADPCM = 0x0010;
    static final int WAVE_FORMAT_DIGISTD   = 0x0015;
    static final int WAVE_FORMAT_DIGIFIX   = 0x0016;
    static final int WAVE_IBM_FORMAT_MULAW = 0x0101;
    static final int WAVE_IBM_FORMAT_ALAW  = 0x0102;
    static final int WAVE_IBM_FORMAT_ADPCM = 0x0103;


    /**
     * AudioWavInputStream
     * Constructor for java.media.audio.AudioWavInputStream.
     * Read data from any input stream and set Audio format.
     * Any bytes which have already been read from the stream 
     * are provided in an array of bytes.
     *
     * This should get called by ACIS.getStream().
     */
    AudioWavInputStream(byte[] bytes, InputStream in)
	throws IOException, BadAudioHeaderException {
	    
        super(bytes, in);
    }


    /**
     * readFormat()
     * read WAV Header. 
     * @exception BadAudioHeaderException, IOException
     * @return AudioFormat instance
     */
    protected AudioFormat readFormat(byte[] bytes)
	throws BadAudioHeaderException, IOException {

        /* In is a protected InputStream variable
	 * defined in FilterInputStream.
	 * It is set by the constructor in FilterInputStream.
	 */

	short nbits;
	int length;
	int wav_type = 0;

	// $jdr: WHAT IS THIS. Who does the byte swapping here?

        boolean isBigEndian = false;    // $$kk: right??  this fixes everything; no more magic read-an-extra-byte!
	boolean isSigned = true;
	
	int sampleRate = -1; 
	int channels = -1;
	String encoding = null;
	int sampleSize = -1;
	int SamplePerUnit = 1;
	
	int nread = 0;
	byte hdr[] = new byte[4];
	String s;
       
        // here we assemble the magicBytes + the input stream into
        // one InputStream

        DataInputStream dis;
        if (bytes == null)  dis = new DataInputStream(in);
        else dis = new DataInputStream(new SequenceInputStream(new ByteArrayInputStream(bytes), in));

         // $$kk: this should be "RIFF" which we used in getStream() to get here.
         // we *really* need to revamp the way we determine formats....
         // the current mechanism (read the first 4 bytes to determine the
         // basic format type) is clearly broken here.  however, it works
         // for the cases i've seen -- i've never seen a MIDI file that 
         // starts with RIFF.

         // $$kk: we should maybe also handle "RIFX" files which are the same as RIFF
         // but with the reverse byte order....

        dis.read(hdr, 0, 4);  
		nread += 4;

	if (DEBUG) {	
            s = new String (hdr);
	    System.out.println("this should be RIFF: " + s);
        }

        // note: this is the total size of the "outside" RIFF block:
        // it's the length of all the format stuff etc. as well as the
        // data block.
        
        // we're setting totalBytes to the length of the data segment 
        // later on

        // Read the next 4 bytes in little-endian
        length = rllong(dis);
		nread += 4;

	if (DEBUG) {
	    System.out.println("total audio size including header = " + length);
	}

        // Read another magic number

        // $$kk:THIS IS ACTUALLY "WAVE", which is the info we *really* want; 
        // RIFF does not actually specify WAVE format...
	
        dis.read(hdr, 0, 4); 
		nread += 4;

        s = new String (hdr);

	if (DEBUG)	
	    System.out.println("this should be WAVE: " + s);

        // we should be providing all the bytes we read here... but we're not
        if (!(s.startsWith("WAVE"))) throw new BadAudioHeaderException("header specifies a non-WAVE format: " + s);

        // ok, now we have no more guarantees about the order of chunks or their
        // sizes, so we need to run a loop and look for things we recognize
        // (fmt and data chunks).  this code assumes that fmt comes before data;
        // i think it needs to.  

        int bytesRead;

        // find and read the "fmt" chunk
        // we break out of this loop either by hitting EOF or finding "fmt "
        
       while(true) {

            bytesRead = dis.read(hdr, 0, 4); 
            nread += 4;
            if (bytesRead == -1) throw new BadAudioHeaderException("reached EOF before finding fmt chunk");
            s = new String (hdr);
            if (s.startsWith("fmt ")) break;    // found the "fmt" chunk
            length = rllong(dis);
            nread += 4;
            if (length % 2 > 0) length++;
            nread += dis.skipBytes(length);
        }

	if (DEBUG)	
	    System.out.println("found the fmt chunk");

      	// Read the format chunk size.
        length = rllong(dis);
        nread += 4;

		// This is the nread position at the end of the format chunk
		int endLength = nread + length;


	if (DEBUG)
	    System.out.println("length of the format chunk is " + length );

	// Read the wave format data out of the format chunk.

	// encoding.
        wav_type = rlshort(dis); nread += 2;
	if (DEBUG)
	    System.out.println("wav_type is " + wav_type );

	if (wav_type == WAVE_FORMAT_PCM)
	    encoding = AudioFormat.JAUDIO_LINEAR;
	else if ( wav_type == WAVE_FORMAT_ALAW )
	    encoding = AudioFormat.JAUDIO_G711_ALAW;
	else if ( wav_type == WAVE_FORMAT_MULAW )
	    encoding = AudioFormat.JAUDIO_G711_ULAW;
	else 
	    throw new BadAudioHeaderException("Unsupport Encoding");

	if (DEBUG) 
	    System.out.println("Encoding: " + encoding );

	// channels
	channels = rlshort(dis); nread += 2;
	if (DEBUG) 
	    System.out.println("Channels: " + channels );

	// sample rate.
	sampleRate = rllong(dis); nread += 4;
	if (DEBUG) 
	    System.out.println("Sample Rate: "+ sampleRate);
 
        // this is the avgBytesPerSec
	rllong(dis); nread += 4;

        // this is blockAlign value
	rlshort(dis); nread += 2;
        
        // this is the PCM-specific value bitsPerSample
	nbits = rlshort(dis); nread += 2;
	if (DEBUG)
	    System.out.println("nbits : " + nbits );
 
	sampleSize = nbits / 8 ;

        // skip any difference between the length of the format chunk
        // and what we read

        // if the length of the chunk is odd, there's an extra pad byte
        // at the end.  i've never seen this in the format chunk, but i 
        // guess we should be sure; it would suck if we got one byte off....

        if (length % 2 != 0) length += 1;

        if (endLength > nread)
            nread += dis.skipBytes(endLength - nread);

	/*
	 * $$jdr: FIX THIS IT'S WRONG.
	 * We are going to convert the ULAW input into 16bit linear in
	 * setData(). So, set the sampleSize appropriately here.
	 * 
	 * $$kk: 01.21.97: should not need this faked sampleSize anymore
	 */
	if (wav_type == WAVE_FORMAT_MULAW ) {
//	    sampleSize = 2;
            isBigEndian = true; 
	}

	if (DEBUG) 
	    System.out.println("Sample Size: " + sampleSize );
 

        // find the "data" chunk
        // we break out of this loop either by hitting EOF or finding "data"
        // $$kk: if "data" chunk precedes "fmt" chunk we are hosed -- can this legally happen?
        
       while(true) {

            bytesRead = dis.read(hdr, 0, 4); nread += 4;
            if (bytesRead == -1) throw new BadAudioHeaderException("reached EOF before finding data chunk");
             s = new String (hdr);
            if (s.startsWith("data")) break;    // found the "fmt" chunk
            length = rllong(dis); nread += 4;
            if (length % 2 > 0) length++;
            nread += dis.skipBytes(length);
        }

	if (DEBUG)	
	    System.out.println("found the data chunk");

        // this is the length of the data chunk
        totalBytes = rllong(dis); nread += 4;

	if (DEBUG)	
	    System.out.println("length of the data chunk is: " + totalBytes);

	// set the headerSize; data should begin this number of bytes from the beginning of the file								
	headerSize = nread;

	return (new AudioFormat(sampleRate, encoding, sampleSize,
				SamplePerUnit, channels,
				isBigEndian, isSigned ));
	
    }
}
