/*
 * @(#)HaeAudioStream.java	1.39 98/08/10
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.audio;

import java.io.IOException;
import java.io.InputStream;
import java.io.PrintStream;
import sun.audio.AudioContainerInputStream;
import sun.audio.AudioFormat;
import sun.audio.Format;

class HaeAudioStream extends HaeWaveStream {

    String streamName;
    AudioContainerInputStream inputStream;
    AudioFormat inputStreamFormat;
    AudioFormat outputStreamFormat;
    
    boolean processBytes = false;
    boolean swapBytes = false;
    boolean convertFromUlaw = false;
    boolean convertUnsigned = false;

    int transfer;
    boolean verbose = false /*true*/ ;

    /**
     * Trace output stream.
     * Used by verbose mode.
     */
    protected final PrintStream trace; {
	trace = System.err;
    }
    
    /**
     * Creates a new HaeWaveStream from an AudioContainerInputStream.
     *
     * @param inputStream with audio data.
     * @param streamName name of the input stream
     *
     * @throws InvalidAudioFormatException if we can't open the stream on the mixer.
     */
    public HaeAudioStream(AudioContainerInputStream inputStream,
			  String streamName)
	                  throws HaeException, InvalidAudioFormatException {
			      
	super(HaeDevice.getSystemMixer());  // can throw HaeException

	this.inputStream = inputStream;
	this.streamName = streamName;

	// Get the format of the raw input and the output to the engine (after conversions)

	outputStreamFormat = inputStream.getOutputFormat();

	Format f[] = (Format[])inputStream.getFormat();
	if (f.length == 0) {
	    throw new InvalidAudioFormatException("Can't find a format on the input stream");
	}
	inputStreamFormat = (AudioFormat)f[0];
	if( inputStreamFormat == null) {
	    throw new InvalidAudioFormatException("Can't find a format on the input stream");
	} else {
	    if(verbose) {
		trace.println("Input stream is formatted: " + inputStreamFormat);
	    }
	}					     

	int sampleSizeInBytes = inputStreamFormat.getSampleSize();
	String encoding = inputStreamFormat.getEncoding();
	int channels = inputStreamFormat.getChannels();
	
	/*
	 * Compare endinans and see if samples are more than 1 byte
	 * long to decide if we should swap on output.
	 * $$jdr: This whole formatFlags thing needs to be changed.
	 */
	if(( inputStreamFormat.isBigEndian() == getMixer().isBigEndian())) {
	    swapBytes = false;
	} else {
	    swapBytes = (sampleSizeInBytes == 1) ? false : true;
	}


	// Do sign conversions if we have to.
	if( !inputStreamFormat.isSigned() &&
	    (encoding == AudioFormat.JAUDIO_LINEAR)) {
	    convertUnsigned = true;
	}
	
	// ... check the encoding ....
	if( encoding.equals(AudioFormat.JAUDIO_G711_ULAW)) {

	    // Can't deal with stereo ulaw.
	    // $$jdr: O.k. we could, but not today.
	    if( channels != 1) {
		throw new InvalidAudioFormatException ("Stereo ulaw");
	    }
	    convertFromUlaw = true;
	    
	    // Open up the channel for 16 bit values, which
	    // we will be converting to in the stream method.
	    sampleSizeInBytes = 2;
	    
	} else {
	    convertFromUlaw = false;
	}

	// For below.
	if( swapBytes || convertUnsigned || convertFromUlaw) {
	    processBytes = true;
	}


//
//	// ... and open it.
//	open(getTransferSize(), outputSampleSizeInBytes * 8, outputChannels, 
//	     (double) outputRate);
    }

	


	void open() throws HaeException {

		// ... and specifics for frame conversion to open the stream ...
		int rate = inputStreamFormat.getSampleRate();
		int sampleSizeInBytes = inputStreamFormat.getSampleSize();
		int channels = inputStreamFormat.getChannels();
		String encoding = inputStreamFormat.getEncoding();

		int outputRate = outputStreamFormat.getSampleRate();
		int outputSampleSizeInBytes = outputStreamFormat.getSampleSize(); 
		int outputChannels = outputStreamFormat.getChannels();
		int outputSampleFrameSize = outputSampleSizeInBytes * outputChannels;

		// Talk to me.
		if(verbose) {
			trace.println("HaeAudioStream: Opening up a: "
							  + channels + " channel, "
				  + sampleSizeInBytes * 8 + " bit, "
				  + rate + " stream.  The output to the engine will be: " 
				  + outputChannels + " channel, " 
				  + outputSampleSizeInBytes * 8 + " bit, "
				  + outputRate + " stream.");
		} 

		// ... and open it.
		open(getTransferSize(), outputSampleSizeInBytes * 8, outputChannels, (double) outputRate);
	}	


	/**
	 * Override start so that we call open on the stream here.  used to call open() in the
	 * constructor for HaeAudioStream, but that does not work for HaeAudioStreamClip.
	 * Clips need their own constructor called before open(), and need to have open() called
	 * _after_ seeking.
	 */
	// NEED TO GET RID OF OFFSET
	public void start(double volume, double pan, boolean reverb) throws HaeException {

		close();

		// re-establish state
		m_samplesSubmitted = 0;
		m_done = false;

		open();
		super.start(volume, pan, reverb);
	}


	
    /**
     * The callback from the HaeEngine.
     * Data to be played should be returned in the buffer that
     * is provided on the command line argument.
     * The buffer.set command can be used to return the
     * buffer and the size of the buffer.
     *
     * @param b buffer for data.
     * @param maxFrames frames wavedata can handle.
     */
    protected boolean getWavedata(HaeWaveStreamBuffer b, int maxFrames) {

        boolean endOfStream = false;    // true if we have reached the end of the stream

		try {
			
			// Figure out how many frames we want to read
			int framesToRead = maxFrames;
			int framesRead = 0;

			// do this calculation in samples, not bytes, because there's not a 1:1 correlation for ulaw data
			if (inputStream.getTotalFrames() != 0) {
//				int framesLeft = (b.getFrameCount(inputStream.getTotalBytes()) - (int)m_samplesSubmitted);
				int framesLeft = (inputStream.getTotalFrames() - (int)m_samplesSubmitted);
				framesToRead = Math.min(maxFrames, framesLeft);
			} 
			
			// Get a buffer from the engine.
			byte[] data = b.allocateByte(framesToRead);
			
			// Get valid data, with any conversions, from the stream
			framesRead = inputStream.getConvertedData(data, framesToRead);
			

			// Using streamIsFinished() lets us avoid a last callback after the stream is done.

			if (inputStream.streamIsFinished()) {
				processEndOfMedia();
				endOfStream = true;
			} 

			b.reduceFrameCount(framesRead);
			
			// update our total count of samples fed to engine
			m_samplesSubmitted += framesRead;

			// stop if we're done reading valid data from a fixed-length file
			// do this calculation in samples, not bytes, because there's not a 1:1 correlation for ulaw data
			if (inputStream.getTotalFrames() > 0) {
				if(m_samplesSubmitted >= inputStream.getTotalFrames()) {
					processEndOfMedia();
					endOfStream = true;
				}
			}
			
		} catch (Exception e) {
			System.err.println(getClass().getName() +
					   "(" + getStreamName() +
					   ") Exception during getWavedata callback. " +
 					   e);
		} catch (Error e) {
			System.err.println(getClass().getName() +
					   "(" + getStreamName() +
					   ") Error during getWavedata callback. " +
					   e);
		}

		// Keep playing if it wasn't end of media.
		if (verbose) {
				  trace.println("getWavedata returning: " +  (!endOfStream));
		}

		return (!endOfStream);
    }
	

    /**
     * Called when EndOfMedia on the stream happens.
     *
     * Usefull for waking people up.
     * Right now we don't really do anything.
     */
    protected void processEndOfMedia() {

	if (verbose) {
	    trace.println("AudioStream (verbose)"
                          + getStreamName() + " EOM");
	}
    }


    /**
     * Report on whats going on.
     */
    public void setVerbose(boolean v) {
	verbose = v;
    }


    public boolean getVerbose() {
	return verbose;
    }
    

    /**
     * Obtain the name of this stream. This name has
     * no meaning and is only used for identification purposes.
     *
     */
    public String getStreamName() {
	return streamName;
    }
}

    
