/*
 * @(#)HaeAudioStreamClip.java	1.4 98/07/13
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.audio;

import java.io.IOException;

class HaeAudioStreamClip extends HaeAudioStream {

    AudioClipData clipData;

	public static final int CLIP_START = 0;
	public static final int CLIP_END = -1;
	public static final int CONTINUOUS_LOOP = -1;


	// looping parameters
	int startFrame = 0;
	int endFrame = 0;
	int numLoops = 0;

	// current state
	int currentLoop = 0;
	boolean looping = true; // this should just be false if we seek past the end of the loop range


    /**
     * Creates a new HaeAudioStreamClip from an AudioClipData.
     *
     * @param clipData with audio data for a file or other fixed audio data source.
     * @param streamName name of the input stream
     *
     * @throws InvalidAudioFormatException if we can't open the stream on the mixer.
     */
    public HaeAudioStreamClip(AudioClipData clipData, String streamName) throws HaeException, InvalidAudioFormatException {

		super(clipData.acis, streamName);

		// we'll use the clip stream in this class
		this.clipData = clipData;

		// set the end frame to the end of the clip for starters
		endFrame = clipData.acis.getTotalFrames();
	}



	/**
	 * override start so that we reset looping.
	 */
	public void start(double volume, double pan, boolean reverb) throws HaeException {

		currentLoop = 0;
		super.start(volume, pan, reverb);
	}




	/**
	 * Seeks to the specified position in the audio data, in samples.
	 * Playback will start from this position.
	 * @throws IOException if there is an exception reading the data
	 * @throws HaeException if the request cannot be fulfilled for any other reason,
	 * such as that the position requested is greater than the length of the audio data.
	 * (Note: we would really want a specific exception with the relevant
	 * information like what went wrong and what the current position is,
	 * but this interface is just a hack for now....)
	 */

	public void seek(int framePosition) throws IOException, HaeException {

		// this can throw an IOException or a HaeException
		clipData.seek(framePosition);

		// if we seek to a position past the looping region, we should not
		// loop back
		if ( (numLoops > 0) && (framePosition > endFrame) )
			looping = false;
		if ( (numLoops > 0) && (framePosition < endFrame) )
			looping = true;
	}


	/*
	 * Sets up the looping parameters.
	 * @param startFrame: first frame included in the loop
	 * @param endFrame: first framae not included in the loop
	 * @param numLoops: number of times to loop.  1 means that
	 * the region will be looped over once, or played twice.
	 * To stop looping, call setLoop with all numLoops == 0.
	 * To loop continuously, use CONTINUOUS_LOOP for numLoops.
	 * To loop over the entire clip, use CLIP_START and CLIP_END for
	 * the start and end frames.
	 * @throws HaeException if the requested parameters are invalid.
	 * startFrame must be less than endFrame which must be less than
	 * the frame length of the clip.  numLoops must be greater than
	 * or equal to zero.
	 */

	 /**
	  * $$kk: should we throw an exception if they try to setLoop while we're looping?
	  * else must handle having looping changed anytime....  (this is better)
	  */

	public void setLoop(int startFrame, int endFrame, int numLoops) throws HaeException {

		// turn looping off
		if (numLoops == 0) {
			this.numLoops = 0;
			this.startFrame = 0;
			this.endFrame = clipData.acis.getTotalFrames();
			return;
		}

		if (endFrame == CLIP_END)
			endFrame = clipData.acis.getTotalFrames();

		if (endFrame <= startFrame)
			throw new HaeException("endFrame must be greater than startFrame");

		if (endFrame > clipData.acis.getTotalFrames())
			throw new HaeException("requested loop end position past end of clip");

		if ( (numLoops < 0) && (numLoops == CONTINUOUS_LOOP) )
			throw new HaeException("negative number of loops requested: " + numLoops);

		this.startFrame = startFrame;
		this.endFrame = endFrame;
		this.numLoops = numLoops;

		// make sure we set looping to false if our current position is outside the
		// loop range
		if (clipData.acis.getCurrentFrame() > endFrame)
			looping = false;
	}



    /**
     * The callback from the HaeEngine.
     * Data to be played should be returned in the buffer that
     * is provided on the command line argument.
     * The buffer.set command can be used to return the
     * buffer and the size of the buffer.
     *
     * @param b buffer for data.
     * @param maxFrames frames wavedata can handle.
     */
    protected boolean getWavedata(HaeWaveStreamBuffer b, int maxFrames) {

        boolean endOfStream = false;    // true if we have reached the end of the stream

		try {

			// Figure out how many frames we want to read
			// for a clip, framesToRead means the number of
			// frames we can read sequentially from the stream without seeking.
			int framesToRead = maxFrames;
			int framesRead = 0; // frames read from stream in this call

			// do this calculation in samples, not bytes, because there's not a 1:1 correlation for ulaw data

			// how far can we read without hitting the end or a loop point?
			int framesLeft;
			if ( (currentLoop < numLoops) && (looping == true) ) {
				// we are working within the loop section
				framesLeft = endFrame - clipData.acis.getCurrentFrame();
			} else {
				// we are working with the whole clip
				framesLeft = clipData.acis.getRemainingFrames();
			}

			// if this is 0 or a negative value, we may need to reset
			if ( (framesLeft <= 0) && ( (currentLoop < numLoops) || (numLoops == CONTINUOUS_LOOP) ) && (looping == true) ) {
				clipData.seek(startFrame);
				currentLoop++;
				framesLeft = endFrame - clipData.acis.getCurrentFrame();
			}

			// protect against negative values
			if (framesLeft < 0)
				framesLeft = 0;

			framesToRead = (int)Math.min(maxFrames, framesLeft);

			// Get a buffer from the engine.
			byte[] data = b.allocateByte(framesToRead);

			// Get valid data, with any conversions, from the stream
			// this also updates the current frame
			framesRead = (int)clipData.acis.getConvertedData(data, framesToRead);

			// update our total count of frames submitted to the engine.  this needs to change, but we're
			// living with it until we make the streams live in the engine until they're done playing....
			// also, we can only return samples played for this ENGINE stream; we may use several engine
			// streams in this java object instance.
			m_samplesSubmitted += framesRead;

			// Using streamIsFinished() lets us avoid a last callback after the stream is done.

			if ( (clipData.acis.streamIsFinished()) && ( (currentLoop >= numLoops) || (looping == false) ) ) {
				processEndOfMedia();
				endOfStream = true;
			}

			b.reduceFrameCount(framesRead);

		} catch (Exception e) {
			System.err.println(getClass().getName() +
					   "(" + getStreamName() +
					   ") Exception during getWavedata callback. " +
 					   e);
		} catch (Error e) {
			System.err.println(getClass().getName() +
					   "(" + getStreamName() +
					   ") Error during getWavedata callback. " +
					   e);
		}

		// Keep playing if it wasn't end of media.
		if (verbose) {
				  trace.println("getWavedata returning: " +  (!endOfStream));
		}

		return (!endOfStream);
    }
}


