/*
 * @(#)HaeMixer.java	1.34 98/08/10
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.audio;

import java.util.StringTokenizer;   // for parsing classpath
import java.io.File;                // for File.separator
import java.lang.reflect.InvocationTargetException;
import java.security.AccessController;
import java.security.PrivilegedAction;
/**
 * HaeMixer:
 * Encapsulates that raw audio output of HAE.  For now, only a single 
 * HaeMixer may be created.
 * Create one HaeMixer, set its parameters, then pass it to HaeNoise-derived 
 * objects.
 */

/* $$jdr we have an issue here for testing. How to test these
 * classes, and not make the interfaces public?
*/

class HaeMixer {

// $$kk    static final int ENVIORNMENT_UNKNOWN	= 0;
// $$kk    static final int ENVIORNMENT_JDK12	= 1;
// $$kk    static final int ENVIORNMENT_NETSCAPE	= 2;

// $$kk    static int environment = ENVIORNMENT_UNKNOWN;

    static {

	// Get the name of the audio library we will load.
	// This is a gross hack added to allow the name
	// of the library to be changed at compile time.
	// The AudioLibraryName class is created at build time.
        final String libraryName = AudioEnvironment.libraryName;

	if( AudioEnvironment.trace) {
	    System.err.println("Java Sound Engine Version: " +
			       AudioEnvironment.version);
	    System.err.println("Using library: " + libraryName);
	}

		AudioSecurity.doPrivileged(
			    new AudioSecurityAction() {
		    public Object run() {
			System.loadLibrary(libraryName);
			return null;
		    }
		}, "UniversalLinkAccess");	

	} // static	


	


// PUBLIC CONSTRUCTION:

    // this constructor will throw an exception if it cannot set the
    // soundbank

    public  HaeMixer(String soundbankPath) throws HaeException {   

        // construct the mixer
        construct();

	// establish the endianism of the hardware
	setHardwareEndianism();

        // set the soundbank
        if (!smartSetSoundbank(soundbankPath))
            throw new HaeException("Invalid sound bank: " + soundbankPath);
    }


    // this constructor will not throw an exception if it cannot set
    // the soundbank.

    public  HaeMixer() throws HaeException {

        // construct the mixer
        construct();

	// establish the endianism of the hardware
	setHardwareEndianism();

        // set the soundbank
        if (!smartSetSoundbank(null))
            System.out.println("Sound bank not set.");
    }

    public void construct() throws HaeException {

        if (m_gmIsOpen) {
            throw new HaeException("only one HaeMixer{} allowed");
        } else {
            m_streamCount = 0;

            m_quality = SAMPLE_RATE_22K;
            m_interpolation = INTERPOLATION_LINEAR;

	    // $$kk: 02.08.98: added FLAGS_ENABLE_REVERB;
	    // else reverb is *disabled*.
            m_modifiers =
		FLAGS_USE_16 |
		FLAGS_USE_STEREO |
		FLAGS_ENABLE_REVERB;

            m_reverbType = REVERB_TYPE_4;
            m_midiVoiceCount = 32;
            m_soundVoiceCount = 32;
            m_volumeDivisor = 6;

	    open();
            m_gmIsOpen = true;

	    if(AudioEnvironment.trace) {
		System.err.println("Mixer construted.");
	    }
        }
    }

    private void setHardwareEndianism() {
            
	String arch = null;

		arch = (String)AudioSecurity.doPrivileged(
			    new AudioSecurityAction() {
		    public Object run() {
			String value = System.getProperty("os.arch");
			return value;
		    }
		}, "UniversalPropertyRead");		


	if (arch != null) {
		if (arch.equals("x86")) {
		    //System.out.println("found x86 hardware");
		    m_bigEndian = false;
		} else if (arch.equals("sparc")) {
		    //System.out.println("found sparc hardware");
		    m_bigEndian = true;
		} else {
		    m_bigEndian = false;
		}
	}

	if( AudioEnvironment.trace) {
	    System.err.println("Rendering to a " +
			      ((m_bigEndian) ? "big endian" : "little endian")+
			      " machine.");
	}
    }


    public void finalize() {

        if (m_streamCount != 0) {
            System.err.println("Mixer finalizing, non-zero stream_count: " +
			       m_streamCount);
        }

        close();
        m_gmIsOpen = false;
    }

    // here is how we try to locate the soundbank:
    // 1) obtain a string:
    //    a) string passed in, if not null
    //    b) java.soundpath property, if not null
    //    c) HaeMixer.defaultSoundbank
    // 2) assume the string represents a full path and try to set the soundbank
    // 3) if this fails, assume the string represents a filename only
    // 4) look relative to java.home
    // 5) iterate through the classpath looking for it

    private boolean smartSetSoundbank(String str) {
            
        boolean set = false;

        try {
            
            String sb = null;

            // get the string we're going to work on.  it can be either a 
            // full path or a filename

            sb = str;

            // if a path was not provided, get it from the system properties
	    if (sb == null) { 


		sb = (String)AudioSecurity.doPrivileged(
			    new AudioSecurityAction() {
		    public Object run() {
			String value = System.getProperty("java.soundpath");
			return value;
		    }
		}, "UniversalPropertyRead");		


            } // if (sb == null)


            if (sb == null) { 
                sb = HaeMixer.defaultSoundbank;
            }

            // see if it is a full path
            try {
                setSoundbank(sb);
                soundbankPath = sb;
                set = true; // succeeded!
            } catch (HaeException e) {
	    }

	    
            String classpath = null, dir = null, fullPath;

            if (!set) {
            
                // it must be a filename.  look relative to java.home.


			dir = (String)AudioSecurity.doPrivileged(
					new AudioSecurityAction() {
				public Object run() {
				String value = System.getProperty("java.home");
				return value;
				}
			}, "UniversalPropertyRead");		

	    dir = dir + File.separatorChar +
		"lib" + File.separatorChar + "audio";
	    fullPath = dir + File.separatorChar + sb; 
	    
	    try { 
		setSoundbank(fullPath);
		soundbankPath = fullPath;
		set = true; // succeeded!
	    } catch (HaeException e) {  }

            } // if (!set)



            if (!set) {
            
                // okay, we didn't find it there.
		// now iterate through the classpath looking for it.

		classpath = (String)AudioSecurity.doPrivileged(
				new AudioSecurityAction() {
			public Object run() {
			String value = System.getProperty("java.class.path");
			return value;
			}
		}, "UniversalPropertyRead");		

	    StringTokenizer st = new StringTokenizer(classpath,
						     File.pathSeparator);
	    while (st.hasMoreTokens() && (!set)) {

		// this is the next element in the classpath
		dir = st.nextToken();
		
		// try appending the soundbank filename to the directory
		try { 
		    fullPath = dir + File.separatorChar + sb; 
		    setSoundbank(fullPath);
		    soundbankPath = fullPath;
		    set = true; // succeeded!
		} catch (HaeException e) {  
		    // if the dir ends in "lib" try
		    // looking in the "audio" subdir
		    if (dir.endsWith("lib")) {
			try {
			    fullPath = dir + File.separatorChar + "audio" +
				File.separatorChar + sb;
			    setSoundbank(fullPath);
			    soundbankPath = fullPath;
			    set = true; // succeeded!
			} catch (HaeException e2) {  }	// e2
		    }
		    
		    if (!set) {  
			// maybe the classpath element is a filename,
			// such as a .zip or .jar file.
			// try pulling off the last part of the dir
			// after the slash and looking in the directory
			if (dir.lastIndexOf(File.separatorChar) != -1) {
			    try {
				dir = dir.substring(
				    0, dir.lastIndexOf(File.separatorChar));
				try {
				    fullPath = dir + File.separatorChar + sb;
				    setSoundbank(fullPath);
				    soundbankPath = fullPath;
				    set = true; // succeeded!
				} catch (HaeException e3) {  
				    
				    // finally, if the dir ends in "lib"
				    // try looking in the "audio" subdir
				    if (dir.endsWith("lib")) {
					try {
					    fullPath = dir +
						File.separatorChar +
						"audio" +
						File.separatorChar +
						sb;
					    setSoundbank(fullPath);
					    soundbankPath = fullPath;
					    set = true; // succeeded!
					} catch (HaeException e4) {
					} // e4
				    } // "lib"
				    
				} // catch e3	
				
			    } catch (Exception e4) {
				// $jdr: THIS MUST STOP!
				// IF YOU ARE WORRYING ABOUT AN EXCEPTION
				// SOMETHING CATCH IT EXPLICITLY!
				//
			    	// in case we get an ArrayOutOfBoundsException
				// or something on the substring()....
			    }
				
			}// if File.separatorChar
		    } // !set
		} // catch e
		
	    } // while

	    } // !set (iterate through classpath)
	    
	} catch (Exception e) {
	    // jdr: ARRRRGGGGGH!!!!!!
	}  // main try/catch

        return set;
    }

    public AudioFormat getOutputFormat() {

        int sampleRate; 
        String encoding;
        int sampleSize;
        int samplesPerUnit;
        int noOfChannels;
        boolean bigEndian;
        boolean signed;

        // sample rate
        switch(m_quality) {

            case SAMPLE_RATE_11K:
                sampleRate = 11025;
                break;
            case SAMPLE_RATE_22K:
                sampleRate = 22050;
                break;
            case SAMPLE_RATE_44K:
                sampleRate = 44100;
                break;
            default:
                sampleRate = 22050;
                break;
        }

        encoding = AudioFormat.JAUDIO_LINEAR;
        sampleSize = ((m_modifiers & FLAGS_USE_16) > 0) ? 16 : 8;
        noOfChannels = ((m_modifiers & FLAGS_USE_STEREO) > 0) ? 2 : 1;
        samplesPerUnit = sampleSize * noOfChannels;
        bigEndian = m_bigEndian;
        signed = true; //???
        return new AudioFormat(sampleRate, encoding, sampleSize,
		       samplesPerUnit, noOfChannels,
		       bigEndian, signed);
    }



    public void setOutputFormat(AudioFormat f) throws HaeException {

        int quality;
        int modifiers;
        int interpolationMode = getInterpolation();

        switch(f.getSampleRate()) {

            case 11025:
                quality = SAMPLE_RATE_11K;
                break;
            case 22050:
                quality = SAMPLE_RATE_22K;
                break;
            case 44100:
                quality = SAMPLE_RATE_44K;
                break;
            default:
                quality = SAMPLE_RATE_22K;
                break;

        }

        modifiers = 0;
        if (f.getSampleSize() == 16)
            modifiers += FLAGS_USE_16;

        if (f.getChannels() == 2)
            modifiers += FLAGS_USE_STEREO;

        // can throw HaeException 
        setAudioModes(quality, interpolationMode, modifiers);
    }

    public void reengageAudio() throws HaeException {

	try {

		AudioSecurity.doPrivileged(
			new AudioSecurityExceptionAction() {
				public Object run() throws HaeException {
				reengageAudio0();
				return null;
				}
			});
	} catch (AudioSecurityActionException e) {
		throw (HaeException)e.getException();
	}

	if( AudioEnvironment.trace) {
		System.err.println("Engine Engaged.");
	}
    }



// PUBLIC OPERATIONS:
    public native void  setSoundbank(String soundbankPath) throws HaeException;
    public native void  setAudioModes(int quality, int interpolationMode, int modifiers) throws HaeException;
    public native void  setReverbType(int type) throws HaeException;
    public native void  setVoiceCounts(int maxMidiVoiceCount, int maxSoundVoiceCount, int volumeDivisor) throws HaeException;
    public native void  setMasterVolume(double theVolume);

    public native void  close();
    public synchronized native void disengageAudio();	
    public synchronized native void reengageAudio0() throws HaeException;

// PUBLIC DATA:
    public int  getQuality()            { return m_quality; }
    public int  getInterpolation()      { return m_interpolation; }
    public int  getModifiers()          { return m_modifiers; }
    public int  getReverbType()         { return m_reverbType; }
    public int  getMidiVoiceCount()     { return m_midiVoiceCount; }
    public int  getSoundVoiceCount()    { return m_soundVoiceCount; }
    public int  getVolumeDivisor()      { return m_volumeDivisor; }
    public  boolean isBigEndian()       { return m_bigEndian; }

    public native double        getMasterVolume();

    // get current tick in microseconds.
    // this is the number of microseconds the device has been open
    public native double        getTick();	       
	
    // number of samples that have been played through the device
    public native long          getSamplesPlayed();	

    public native int           getAudioLatency();

//  public native void          getInfo(HaeMixerInfo target);
//  public native void          getFrame(HaeMixerFrame target);

// PUBLIC CONSTANTS:
    // total number of voices. This is shared amongst streams and songs.
    public static final int	MAX_VOICES = 32;	

    public static final int	SAMPLE_RATE_11K = 11;			// output at 11 kHz
    public static final int	SAMPLE_RATE_22K = 22;			// 22 kHz
    public static final int	SAMPLE_RATE_44K = 44;			// 44 kHz

    public static final int	INTERPOLATION_DROP_SAMPLE = 0;
    public static final int	INTERPOLATION_2_POINT = 1;
    public static final int	INTERPOLATION_LINEAR = 2;

    public static final int	FLAGS_NONE = 0;
    public static final int	FLAGS_USE_16 = 1<<0;		// use 16 bit output
    public static final int	FLAGS_USE_STEREO = 1<<1;	// use stereo output
    public static final int	FLAGS_ENABLE_REVERB = 1<<2;	// $$kk: 02.08.98: enable reverb; must set this flag to enable reverb
														// (this is wrong: disable reverb)
    public static final int	FLAGS_STEREO_FILTER = 1<<3;	// if stereo is enabled, use a stereo filter

    public static final int	REVERB_NONE = 1;
    public static final int	REVERB_TYPE_1 = 1;		// None
    public static final int	REVERB_TYPE_2 = 2;		// Igor's Closet
    public static final int	REVERB_TYPE_3 = 3;		// Igor's Garage
    public static final int	REVERB_TYPE_4 = 4;		// Igor's Acoustic Lab
    public static final int	REVERB_TYPE_5 = 5;		// Igor's Cavern
    public static final int	REVERB_TYPE_6 = 6;		// Igor's Dungeon
    public static final int	REVERB_TYPE_7 = 7;
    public static final int	REVERB_TYPE_8 = 8;


// PRIVATE OPERATIONS:
    private native void	open() throws HaeException;

// PRIVATE DATA:
    private int                 m_streamCount;

    private int                 m_quality;
    private int                 m_interpolation;
    private int                 m_reverbType;
    private int                 m_modifiers;
    private int                 m_midiVoiceCount;
    private int                 m_soundVoiceCount;
    private int                 m_volumeDivisor;

    private static boolean      m_gmIsOpen          = false;
	private static boolean		m_bigEndian			= false;	// false for x86 hardware, true for sparc
    static private String       soundbankPath       = null;
    static private String       defaultSoundbank    = "soundbank.gm";

};


