/*
 * @(#)HaeWaveStream.java	1.15 98/04/16
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.audio;

import java.lang.Math;


/** HaeWaveStream{}
 *	Plays a sound whose waveform data is provided by an abstract method.
 *	getWavedata() must be defined in your subclass.  It will be called when
 *	waveform data is needed.  Please see HaeWaveStreamBuffer.java for
 *	more information.
 */

abstract class HaeWaveStream extends HaeWaveNoise
{
// PUBLIC CONSTRUCTION:
	public	HaeWaveStream(HaeMixer mixer)
	{
		super(mixer);
		m_buffer = null;

                m_transferSize = DEFAULT_TRANSFER_SIZE;
                m_samplesSubmitted = 0;
                m_done = false;
	}
	public	HaeWaveStream(HaeMixer mixer,
							int suggestedBufferBytes, 
							int bitsPerSample,	// 8 or 16 bit data
							int channelsPerFrame,	// 1 or 2 channels
							double sampleRate)
				throws HaeException
	{
		super(mixer);
		m_buffer = null;
                m_samplesSubmitted = 0;
                m_done = false;
		open(suggestedBufferBytes, bitsPerSample, channelsPerFrame, sampleRate);
	}

// PUBLIC OPERATIONS:
	public void				open(int suggestedBufferBytes, 
									int bitsPerSample,	// 8 or 16 bit data
									int channelsPerFrame,	// 1 or 2 channels
									double sampleRate)
								throws HaeException
	{
	int			bufferBytes;

		close();

		bufferBytes = 512;
		while (bufferBytes < suggestedBufferBytes)
		{
			bufferBytes *= 2;
		}
		if (suggestedBufferBytes - bufferBytes / 2 <
			bufferBytes - suggestedBufferBytes)
		{
			bufferBytes /= 2;
		}
		m_bufferBytes = bufferBytes;	// must be a power of two for HAE to work well

		setSampleInfo(sampleRate, bitsPerSample, channelsPerFrame);

		m_buffer = new HaeWaveStreamBuffer(getBytesPerFrame());
		nativeOpen(getSampleRate(),
					bitsPerSample, channelsPerFrame,
					getBytesPerFrame(), bufferBytes);

	}
	public void				close()
	{
		stop(false);
		//m_buffer = null;	// (OLD: let the garbage collector have it)
                                        // kk: can't set m_buffer to null here because
                                        // getWavedata may still be called after close()
                                        // during the stream shutdown process in the engine
	}

	public native void		throwOnLastError()
        								throws HaeException;



        /** 
        * Gets the requested transfer size (buffer size) for this stream.
        * This is a suggested value only, and the implementation may
        * choose to use a different value.
        * $$kk: ideally, we would return the *actual* transfer size....
        */
        public int getTransferSize() {

            return m_transferSize;
        }

        /** 
        * Sets the requested transfer size (buffer size) for this stream.
        * This is a suggested value only, and the implementation may
        * choose to use a different value.  If the transfer size is not
        * explicitly set, it will be assigned a default value.
        * $$kk: this will not do anything once the stream has been opened.
        *       how to handle this???
        */
        public void setTransferSize(int bytes) {
	    
            m_transferSize = bytes;
        }

        /** 
        * Gets the current position in the stream as the number of samples
        * played.
        */
        public long getSamplesPlayed() {

            long samplesPlayed;

            // if the stream exists in the engine, just ask it its position, which
            // is tracked internally.  

            if (!m_done) {
                samplesPlayed = privateGetSamplesPlayed(); 
                return samplesPlayed;
            }

            // ok, the stream has shut down in the engine.  
            // we need to calculate the exact position.

            // calculate delta in stream samples since stream end.
            // it's the change in mixer samples played since the stream end in the engine, times the conversion between mixer and stream sample rates
         
            long streamDelta = (long)((getMixer().getSamplesPlayed() - m_engineSamplesPlayedAtStreamEnd) * ((float) getSampleRate() / getMixer().getOutputFormat().getSampleRate()));
            samplesPlayed = Math.min(m_samplesSubmitted, (m_streamSamplesPlayedAtStreamEnd + streamDelta));

            return samplesPlayed;
        } 


        /**
        * Blocks until all unplayed buffers have been played.
        */
/*
// now implemented natively
//    
//	    public void drain() {
//  
//            double samplesSubmitted = m_samplesSubmitted;
//
//            while(getSamplesPlayed() < samplesSubmitted)
//                try { Thread.sleep(10); } catch (InterruptedException e) {  }
//        }
*/


        

/* TRY THIS LATER WHEN NATIVE THREADS ARE MORE UNDERSTOOD
    public void	play() throws HaeException
    {
		if (!isPlaying()) start();

		try
		{
	    	while(isPlaying())
	    	{
	    		wait();
	    	}
	    	throwOnLastError();
		}
		catch (InterruptedException e)
		{
		    System.err.println("Interrupted in stream play.");
		}
    }
*/
	public synchronized native void		stop(boolean withFade);	// frees memory

	public native void		setVolume(double volume);
	public native void		setPan(double position);// left=-1.0;center=0.0;right=1.0


        /**
        * Blocks until all unplayed buffers have been played.
        */
		public native void				drain();

        /** 
        * Discards any unplayed buffers for this stream.
        */
        public native void              flush();

        /** 
        * Pause the stream
        */
        public native void              pause();

        /**
        * Resume paused stream
        */
        public native void              resume();


// PUBLIC DATA:

	public native boolean	isPlaying();
	public native double	getVolume();
	public native double	getPan(); // left=-1.0;center=0.0;right=1.0
	public int				getBufferBytes()
		{ return m_bufferBytes; }

// PROTECTED OPERATIONS:

	// $$kk: 01.14.98: added offset param
	protected synchronized native void	privateStart(double volume,
											double position,
											boolean useReverb);
	protected native void	privateSetSampleRate(double sampleRate);
	protected native void	privateSetReverb(boolean useReverb);

	protected abstract boolean	getWavedata(HaeWaveStreamBuffer buffer,
											int maxFrameCount);
	protected void			streamingFinished()
	{
            // done in engine.  record the current mixer and stream samples played
            // so that we can still calculate the exact sample playback position.

            m_streamSamplesPlayedAtStreamEnd = getSamplesPlayed();
            m_engineSamplesPlayedAtStreamEnd = getMixer().getSamplesPlayed();
            m_done = true;

//          notifyAll();	//?? FOR LATER WHEN WE GET BETTER THREADING GOING
	}

        protected native long privateGetSamplesPlayed();


// PRIVATE OPERATIONS:
	protected synchronized native void	nativeOpen(double sampleRate,
										int bitsPerSample,
										int channelsPerFrame,
										int bytesPerFrame,
										int bufferBytes)
								throws HaeException;
	protected boolean		privateGetWavedata(int maxFrameCount)
	{
		return getWavedata(m_buffer, maxFrameCount);
	}


// PROTECTED DATA:

        protected static final int DEFAULT_TRANSFER_SIZE = 16384;

        protected int                   m_transferSize;
        protected long                  m_samplesSubmitted;


// PRIVATE DATA:
	private	int		        m_bufferBytes;
	private HaeWaveStreamBuffer	m_buffer;
    protected boolean					m_done;             // true if the stream
                                                            // is done streaming at the engine level; java object
                                                            // may exist longer

        // use these vars to hack an accurate sample count after the HAE stream is gone
        private long                    m_streamSamplesPlayedAtStreamEnd; 
        private long                    m_engineSamplesPlayedAtStreamEnd;                
};


