/*
 * @(#)AffineTransform.java	1.9 98/03/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

import java.awt.Shape;
import java.awt.geom.Point2D;
import java.awt.geom.NoninvertibleTransformException;

/**
 * This class represents a 2D affine transform which performs a linear
 * mapping from 2D coordinates to other 2D coordinates in a manner which
 * preserves the "straightness" and "parallelness" of lines.  Affine
 * transformations can be constructed using sequences of translations,
 * scales, flips, rotations, and shears.
 * <p>
 * Such a coordinate transformation can be represented by a 3 row by
 * 3 column matrix with a implied last row of [ 0 0 1 ] which is
 * applied to the coordinates by considering
 * them to be a column vector and multiplying the coordinate vector
 * by the matrix according to the following process.
 * <pre>
 *		[  m00  m01  m02  ] [ x ]   [ m00x + m01y + m02 ]
 *		[  m10  m11  m12  ] [ y ] = [ m10x + m11y + m12 ]
 *		[   0    0    1   ] [ 1 ]   [         1         ]
 * </pre>
 * @see TransformChain
 *
 * @version 10 Feb 1997
 * @author Jim Graham
 * @author Modified by Tony Tseung, Taligent
 */

public class AffineTransform {

    protected double m00, m10, m01, m11, m02, m12;

    /**
     * Constructs a new AffineTransform representing the Identity
     * transformation.
     */
    public AffineTransform() {
        m00 = m11 = 1.0;
    }


    public AffineTransform(java.awt.geom.AffineTransform at) {
	this(extractMatrix(at));
    }

    static double[] extractMatrix(java.awt.geom.AffineTransform at) {
	double[] ret = new double[6];
	at.getMatrix(ret);
	return ret;
    }

    /**
     * Constructs a new AffineTransform from 6 floating point values
     * representing the 6 specifiable entries of the 3x3 transformation matrix.
     */
    public AffineTransform(float m00, float m10,
        		   float m01, float m11,
        		   float m02, float m12) {
        this.m00 = m00;
        this.m10 = m10;
        this.m01 = m01;
        this.m11 = m11;
        this.m02 = m02;
        this.m12 = m12;
        fAccelerator = GetAccelerator(); // should be const in GetAccelerator
        if (fAccelerator != TYPE_IDENTITY || fAccelerator != TYPE_TRANSLATE)
            fRotateAccelerator = ROTATE_SLOW;
    }

    /**
     * Constructs a new AffineTransform from an array of floating point
     * values representing either the 4 non-translation or the 6 specifiable
     * entries of
     * the 3x3 transformation matrix.  The values are retrieved from the
     * array as { m00 m10 m01 m11 [m02 m12] }.
     * @param flatmatrix The float array containing the values to be set
     * in the new AffineTransform object. The length of the array is 
     * assumed to be at least 4. If the length of the array is 
     * less than 6, only the first 4 values are taken. If the length of
     * the array is greater than 6, the first 6 values are taken.
     */
    public AffineTransform(float[] flatmatrix) {
        m00 = flatmatrix[0];
        m10 = flatmatrix[1];
        m01 = flatmatrix[2];
        m11 = flatmatrix[3];
        if (flatmatrix.length > 4) {
            m02 = flatmatrix[4];
            m12 = flatmatrix[5];
        }
        fAccelerator = GetAccelerator(); // should be const in GetAccelerator
        if (fAccelerator != TYPE_IDENTITY || fAccelerator != TYPE_TRANSLATE)
            fRotateAccelerator = ROTATE_SLOW;
    }

    /**
     * Constructs a new AffineTransform from a 2-dimensional array of
     * floating point values representing either the 4 non-translation or the
     * 6 specifiable entries of the 3x3 transformation matrix.  The values are
     * retrieved from the 2-dimensional array using the row index as
     * the first subscript and the column index as the second.
     * @param matrix The matrix used to construct the new AffineTransform
     * object. Matrix is assumed to be at least 2x2. If the matrix
     * is smaller than 2x3, only the left upper 2x2 matrix is taken. If the
     * matrix is larger than 2x3, the left upper 2x3 matrix is taken.
     */
    public AffineTransform(float[][] matrix) {
        m00 = matrix[0][0];
        m10 = matrix[1][0];
        m01 = matrix[0][1];
        m11 = matrix[1][1];
        if (matrix.length > 2) {
            m02 = matrix[0][2];
            m12 = matrix[1][2];
        }
        fAccelerator = GetAccelerator(); // should be const in GetAccelerator
        if (fAccelerator != TYPE_IDENTITY || fAccelerator != TYPE_TRANSLATE)
            fRotateAccelerator = ROTATE_SLOW;
    }

    /**
     * Constructs a new AffineTransform from 6 double precision values
     * representing the 6 specifiable entries of the 3x3 transformation matrix.
     */
    public AffineTransform(double m00, double m10,
        		   double m01, double m11,
        		   double m02, double m12) {
        this.m00 = m00;
        this.m10 = m10;
        this.m01 = m01;
        this.m11 = m11;
        this.m02 = m02;
        this.m12 = m12;
        fAccelerator = GetAccelerator(); // should be const in GetAccelerator
        if (fAccelerator != TYPE_IDENTITY || fAccelerator != TYPE_TRANSLATE)
            fRotateAccelerator = ROTATE_SLOW;
    }

    /**
     * Constructs a new AffineTransform from an array of double precision
     * values representing either the 4 non-translation or the 6 specifiable
     * entries of
     * the 3x3 transformation matrix.  The values are retrieved from the
     * array as { m00 m10 m01 m11 [m02 m12] }.
     * @param flatmatrix The double array containing the values to be set
     * in the new AffineTransform object. The length of the array is 
     * assumed to be at least 4. If the length of the array is 
     * less than 6, only the first 4 values are taken. If the length of
     * the array is greater than 6, the first 6 values are taken.
     */
    public AffineTransform(double[] flatmatrix) {
        m00 = flatmatrix[0];
        m10 = flatmatrix[1];
        m01 = flatmatrix[2];
        m11 = flatmatrix[3];
        if (flatmatrix.length > 4) {
            m02 = flatmatrix[4];
            m12 = flatmatrix[5];
        }
        fAccelerator = GetAccelerator(); // should be const in GetAccelerator
        if (fAccelerator != TYPE_IDENTITY || fAccelerator != TYPE_TRANSLATE)
            fRotateAccelerator = ROTATE_SLOW;
    }

    /**
     * Constructs a new AffineTransform from a 2-dimensional array of
     * double precision values representing either the 4 non-translation or the
     * 6 specifiable entries of the 3x3 transformation matrix.  The values are
     * retrieved from the 2-dimensional array using the row index as
     * the first subscript and the column index as the second.
     * @param matrix The matrix used to construct the new AffineTransform
     * object. Matrix is assumed to be at least 2x2. If the matrix
     * is smaller than 2x3, only the left upper 2x2 matrix is taken. If the
     * matrix is larger than 2x3, the left upper 2x3 matrix is taken.
     */
    public AffineTransform(double[][] matrix) {
        m00 = matrix[0][0];
        m10 = matrix[1][0];
        m01 = matrix[0][1];
        m11 = matrix[1][1];
        if (matrix.length > 2) {
            m02 = matrix[0][2];
            m12 = matrix[1][2];
        }
        fAccelerator = GetAccelerator(); // should be const in GetAccelerator
        if (fAccelerator != TYPE_IDENTITY || fAccelerator != TYPE_TRANSLATE)
            fRotateAccelerator = ROTATE_SLOW;
    }

    /**
     * Resets this transform to the Identity transform.
     */
    public void setToIdentity() {
        InternalSetToIdentity();
        fAccelerator = TYPE_IDENTITY;
    }

    /**
     * Retrieves the 6 specifiable values in the 3x3 affine transformation
     * matrix into
     * an array of double precisions values.  The values are stored into
     * the array as { m00 m10 m01 m11 m02 m12 }.
     * @param flatmatrix The double array used to store the returned values.
     * The length of the array is assumed to be at least 6.
     */
    public void getMatrix(double[] flatmatrix) {
        flatmatrix[0] = m00;
        flatmatrix[1] = m10;
        if (fAccelerator==TYPE_ROTATE) {
            flatmatrix[0] = m00;
            flatmatrix[1] = m10;
            flatmatrix[2] = -m10;
            flatmatrix[3] = m00;
            if (flatmatrix.length > 4) {
                flatmatrix[4] = m02 == 0 ? 0 : -m02 * (m00 - 1.) + m12 * m10;
                flatmatrix[5] = m12 == 0 ? 0 : -m02 * m10 - m12 * (m00 - 1.);
            }
        }
        else {       
            flatmatrix[2] = m01;
            flatmatrix[3] = m11;
            if (flatmatrix.length > 4) {
                flatmatrix[4] = m02;
                flatmatrix[5] = m12;
            }
        }
    }

    /**
     * Retrieves the 6 specifiable values in the 3x3 affine transformation
     * matrix into
     * a 2-dimensional array of double precision values.  The values are
     * stored into the 2-dimensional array using the row index as the
     * first subscript and the column index as the second.
     * @param matrix The 2-dimensional double array to store the returned
     * values.  The array is assumed to be at least 2x3.
     */
    public void getMatrix(double[][] matrix) {
        matrix[0][0] = m00;
        matrix[1][0] = m10;
        if (fAccelerator == TYPE_ROTATE) {
            matrix[0][1] = -m10;
            matrix[1][1] = m00;
            matrix[0][2] = m02 == 0. ? 0. : -m02 * (m00 - 1.0) + m12 * m10;
            matrix[1][2] = m12 == 0. ? 0. : -m02 * m10 - m12 * (m00 - 1.0);
        }
        else {       
            matrix[0][1] = m01;
            matrix[1][1] = m11;
            matrix[0][2] = m02;
            matrix[1][2] = m12;
        }
    }

    /**
     * Concatenates this transform with a translation transformation.
     * This is equivalent to calling concatenate(T), where T is an
     * AffineTransform represented by the following matrix:
     * <pre>
     *		[   1    0    tx  ]
     *		[   0    1    ty  ]
     *		[   0    0    1   ]
     * </pre>
     */
    public void translate(double tx, double ty) {
/*
        m02 += tx * m00 + ty * m01;
        m12 += tx * m10 + ty * m11;
*/

        // this is really pre-translate
        if (tx==0.0 && ty == 0.0) return;

        switch (fAccelerator) {
        case TYPE_IDENTITY:
            m02 = tx;
            m12 = ty;
            fAccelerator = TYPE_TRANSLATE;
            break;
        case TYPE_TRANSLATE:
            m02 += tx;
            m12 += ty;
            break;
        case TYPE_ROTATE:
            if (m02 == 0. && m12 == 0.) {
                m02 = -tx;
                m12 = -ty;
                break;
            }
            RotateToAffine(fRotateAccelerator);
            fAccelerator = TYPE_AFFINE;
            //fall through
        case TYPE_AFFINE:
            m02 += ty * m01;
            m12 += tx * m10;
            //fall through
        case TYPE_SCALE:
            m02 += tx * m00;
            m12 += ty * m11;
            break;
        }
    }

    /**
     * Concatenates this transform with a rotation transformation.
     * This is equivalent to calling concatenate(R), where R is an
     * AffineTransform represented by the following matrix:
     * <pre>
     *		[   cos(theta)    -sin(theta)    0   ]
     *		[   sin(theta)     cos(theta)    0   ]
     *		[       0              0         1   ]
     * </pre>
     * Rotating with a positive angle theta rotates points on the positive
     * x axis toward the positive y axis.
     * @param theta The angle of rotation in radians.
     */
    public void rotate(double theta) {
/*
        double sin = Math.sin(theta);
        double cos = Math.cos(theta);
        double m0, m1;
        m0 = m00;
        m1 = m01;
        m00 =  cos * m0 + sin * m1;
        m01 = -sin * m0 + cos * m1;
        m0 = m10;
        m1 = m11;
        m10 =  cos * m0 + sin * m1;
        m11 = -sin * m0 + cos * m1;
*/
        rotate(theta, 0., 0.);
    }
    // this enables acceleration more effectively
    protected void rotate(double theta, double cx, double cy) {
        // this really pre-rotate:
        theta = FirstCycle(theta);
        if (theta == 0.0 || theta == k360DegreeInRadians)

        switch (fAccelerator) {
        case TYPE_IDENTITY:
            fAccelerator = TYPE_ROTATE;
            m01 = FirstCycle(theta);
            m02 = cx;
            m12 = cy;
            fRotateAccelerator = GetFastRotate(m01);
            ComputeCosAndSin();
            return;
        case TYPE_ROTATE:
            if ((m02 == cx) && (m12 == cy)) {
                m01 = FirstCycle(m01 + theta);
                fRotateAccelerator = GetFastRotate(m01);
                if (fRotateAccelerator != ROTATE_0) {
                    ComputeCosAndSin();
                } else {
                    fAccelerator = TYPE_IDENTITY;
                }
                return;
            }
            RotateToAffine(fRotateAccelerator);

        case TYPE_TRANSLATE:
        case TYPE_SCALE:
        case TYPE_AFFINE:
            double nCos = Math.cos(theta);
            double nSin = Math.sin(theta);
            double tx;
            double ty;
            if (cx != 0. || cy != 0.) {
                tx = -cx*(nCos-1.0) + cy*nSin ;
                ty = -cx*nSin - cy*(nCos-1);
            }
            else {
                tx = 0.0;
                ty = 0.0;
            }
            PreMatrixMult(nCos, nSin, -nSin, nCos, tx, ty);
            fAccelerator = GetAccelerator();
            return;
        }
    }

    /**
     * Concatenates this transform with a scaling transformation.
     * This is equivalent to calling concatenate(S), where S is an
     * AffineTransform represented by the following matrix:
     * <pre>
     *		[   sx   0    0   ]
     *		[   0    sy   0   ]
     *		[   0    0    1   ]
     * </pre>
     */
    public void scale(double sx, double sy) {
/*
        m00 *= sx;
        m10 *= sx;
        m01 *= sy;
        m11 *= sy;
*/
        scale(sx, sy, 0., 0.);
    }
    // this enables acceleration more effectively
    protected void scale(double sx, double sy, double cx, double cy) {
        // this is really pre-scale:
        double tx = cx;
        if (tx != 0.) tx *= 1.0 - sx;
        double ty = cy;
        if (ty != 0.) ty *= 1.0 - sy;
        switch (fAccelerator) {
        case TYPE_TRANSLATE:
            tx += m02;
            ty += m12;
        case TYPE_IDENTITY:
            fAccelerator = TYPE_SCALE;
            InternalSetToIdentity();
            m00 = sx;
            m11 = sy;
            m02 = tx;
            m12 = ty;
            return;
        case TYPE_SCALE:
            m02 += tx*m00;
            m12 += ty*m11;
            m00 *= sx;
            m11 *= sy;
            return;
        case TYPE_ROTATE:
            RotateToAffine(fRotateAccelerator);
            fAccelerator = TYPE_AFFINE;
            // fall through
        case TYPE_AFFINE:
            m02 += tx*m00 + ty*m01;
            m12 += ty*m11 + ty*m10;
            m00 *= sx;
            m01 *= sy;
            m10 *= sx;
            m11 *= sy;
            break;
        }
    }

    /**
     * Sets this transform to a translation transformation.
     * The matrix representing this transform becomes:
     * <pre>
     *		[   1    0    tx  ]
     *		[   0    1    ty  ]
     *		[   0    0    1   ]
     * </pre>
     */
    public void setToTranslation(double tx, double ty) {
/*
        m00 = 1.0;
        m10 = 0.0;
        m01 = 0.0;
        m11 = 1.0;
        m02 = tx;
        m12 = ty;
*/
        InternalSetToIdentity();
        if (tx != 0.0 || ty != 0.0) {
            fAccelerator = TYPE_TRANSLATE;
            m02 = tx;
            m12 = ty;
        }
        else
            fAccelerator = TYPE_IDENTITY;
    }

    /**
     * Sets this transform to a rotation transformation.
     * The matrix representing this transform becomes:
     * <pre>
     *		[   cos(theta)    -sin(theta)    0   ]
     *		[   sin(theta)     cos(theta)    0   ]
     *		[       0              0         1   ]
     * </pre>
     * Rotating with a positive angle theta rotates points on the positive
     * x axis toward the positive y axis.
     * @param theta The angle of rotation in radians.
     */
    public void setToRotation(double theta) {
/*
        double sin = Math.sin(theta);
        double cos = Math.cos(theta);
        m00 = cos;
        m10 = sin;
        m01 = -sin;
        m11 = cos;
        m02 = 0.0;
        m12 = 0.0;
*/
        setToRotation(theta, 0., 0.);
    }
    // this enables acceleration more effectively
    protected void setToRotation(double theta, double cx, double cy) {
        InternalSetToIdentity();
        fAccelerator = TYPE_ROTATE;
        theta = FirstCycle(theta);
        fRotateAccelerator = GetFastRotate(theta);
        if (fRotateAccelerator != ROTATE_0) {
            m01 = theta;
            m02 = cx;
            m12 = cy;
            ComputeCosAndSin();
        } else {
            fAccelerator = TYPE_IDENTITY;
        }
    }

    /**
     * Sets this transform to a scaling transformation.
     * The matrix representing this transform becomes:
     * <pre>
     *		[   sx   0    0   ]
     *		[   0    sy   0   ]
     *		[   0    0    1   ]
     * </pre>
     */
    public void setToScale(double sx, double sy) {
/*
        m00 = sx;
        m10 = 0.0;
        m01 = 0.0;
        m11 = sy;
        m02 = 0.0;
        m12 = 0.0;
*/
        setToScale(sx, sy, 0., 0.);
    }
    // this enables acceleration more effectively
    protected void setToScale(double sx, double sy, double cx, double cy) {
        InternalSetToIdentity();
        if (sx != 1.0 || sy != 1.0) {
            fAccelerator = TYPE_SCALE;
            m00 = sx;
            m11 = sy;
            m02 = cx*(1.0 - sx);
            m12 = cy*(1.0 - sy);
        }
        else {
            fAccelerator = TYPE_IDENTITY;
        }
    }

    /**
     * Prepends a Transform object Tx to this Transform object Cx.
     * Cx is updated to perform the combined transformation.
     * Transforming a point p by the updated transform Cx' is
     * equivalent to first transforming p by Tx and then transforming
     * the result by the original transform Cx. In other words,
     * Cx'(p) = Cx(Tx(p)).  In matrix notation, if this transform Cx
     * is represented by the matrix [this] and Tx is represented by
     * the matrix [Tx], then this method does the following:
     * <pre>
        	[this] = [this] x [Tx]
     * </pre>
     * Tx and Cx must be compatible Transform
     * objects whose combination can be mathematically reduced
     * to an AffineTransform.  Otherwise an exception is thrown.
     * @param Tx The Transform object to be prepended to this Transform object.
     */
    public void prepend(AffineTransform Tx) {
        this.preConcatenate(Tx);
    }

    /**
     * Appends a Transform object Tx to this Transform object Cx.
     * Cx is updated to perform the combined transformation.
     * Transforming a point p by the updated transform Cx' is
     * equivalent to first transforming p by the original transform
     * Cx and then transforming the result by Tx.  In other words,
     * Cx'(p) = Tx(Cx(p)).  In matrix notation, if this transform Cx
     * is represented by the matrix [this] and Tx is represented by
     * the matrix [Tx], then this method does the following:
     * <pre>
                [this] = [Tx] x [this]
     * </pre>
     * Tx and Cx must be compatible Transform
     * objects whose combination can be mathematically reduced
     * to an AffineTransform.  Otherwise an exception is thrown.
     * @param Tx The Transform object to be appended to this Transform object.
     */
    public void append(AffineTransform Tx) {
        this.concatenate(Tx);
    }

    /**
     * Concatenates this transformation with the specified AffineTransform
     * object.  In matrix notation, if this transform is represented by
     * the matrix [this] and AT is represented by the matrix [AT], then
     * this method does the following:
     * <pre>
        	[this] = [this] x [AT]
     * </pre>
     * This method does exactly what the method prepend does
     * but does not throw an exception for incompatible transform, because
     * the specified argument is an AffineTransform.
     * @param AT the AffineTransform to be concatenated.
     */
    public void concatenate(AffineTransform AT) {
/*
        double m0, m1;
        m0 = m00;
        m1 = m01;
        m00  = AT.m00 * m0 + AT.m10 * m1;
        m01  = AT.m01 * m0 + AT.m11 * m1;
        m02 += AT.m02 * m0 + AT.m12 * m1;
        m0 = m10;
        m1 = m11;
        m10  = AT.m00 * m0 + AT.m10 * m1;
        m11  = AT.m01 * m0 + AT.m11 * m1;
        m12 += AT.m02 * m0 + AT.m12 * m1;
*/
        /*
         * this really does pre-concatenate...
         */
        if (AT.fAccelerator == TYPE_IDENTITY)
            // OPERATOR IS IDENTITY
            return;
            //---- DONE ---

        if (fAccelerator == TYPE_IDENTITY) {
            // OPERAND IS IDENTITY
            fAccelerator = AT.fAccelerator;
            if (fAccelerator != TYPE_TRANSLATE) {
                // just copy everything over ...
                m00 = AT.m00;
                m10 = AT.m10;
                m01 = AT.m01;
                m11 = AT.m11;
            }
            m02 = AT.m02;
            m12 = AT.m12;
            fRotateAccelerator = AT.fRotateAccelerator;
            return;
            //---- DONE ---
        }
        if (fAccelerator == TYPE_TRANSLATE &&
            AT.fAccelerator == TYPE_TRANSLATE) {
            m02 += AT.m02;
            m12 += AT.m12;
        }

        if (AT.fAccelerator == TYPE_ROTATE) {
            // OPERATOR IS ROTATE
            if (fAccelerator == TYPE_ROTATE) {
                // both the same type: try to rotate!!
                if ((m02 == AT.m02) && (m12 == AT.m12)) {
                    // Success!! do the rotate
                    m01 = FirstCycle(m01 + AT.m01);
                    fRotateAccelerator = GetFastRotate(m01);
                    if (fRotateAccelerator != ROTATE_0) {
                        ComputeCosAndSin();
                    } else {
                        fAccelerator = TYPE_IDENTITY;
                    }
                    return;
                }
                // failure: convert 'this' to an TYPE_AFFINE matrix
                RotateToAffine(fRotateAccelerator);
            }

            // The operator is still an accelerated rotate matrix;
            // convert it to an TYPE_AFFINE matrix before multiplying.

            double tmp[] = new double[6];
            AT.getMatrix(tmp);
            PreMatrixMult(tmp[0], tmp[1], tmp[2], tmp[3], tmp[4], tmp[5]);
            fAccelerator = GetAccelerator();
            return;
            //---- DONE ---
        }

        if (AT.fAccelerator == TYPE_TRANSLATE) {
            translate(AT.m02, AT.m12);
            return;
            //---- DONE ---
        }

        if (fAccelerator == TYPE_ROTATE) {
            // THIS IS ROTATE, OPERATOR IS NOT
            RotateToAffine(fRotateAccelerator);
            //---- DONE ---
        }

        // Wiped out the special cases: brute force matrix multiply.
        double tmp[] = new double[6];
        AT.getMatrix(tmp);
        PreMatrixMult(tmp[0], tmp[1], tmp[2], tmp[3], tmp[4], tmp[5]);
        fAccelerator = GetAccelerator();
    }

    /*
     * The following four functions are added in addition to the original
     * AffineTransform to add acceleration to symetrical functions to
     * cancatenate, translate, scale, and rotate:
     */

    // Pre Concatenation:
    protected void preConcatenate(AffineTransform AT) {
        /*
         * this really does post-concatenate...
         */

        if (AT.fAccelerator == TYPE_IDENTITY) // OPERATOR IS Identity
            return; //---- DONE ---

        if (fAccelerator == TYPE_IDENTITY) {
            // OPERAND IS Identity
            fAccelerator = AT.fAccelerator;
            if (fAccelerator == TYPE_TRANSLATE) {
                m02 = AT.m02;
                m12 = AT.m12;
            }
            else {
                m00 = AT.m00;
                m10 = AT.m10;
                m01 = AT.m01;
                m11 = AT.m11;
                m02 = AT.m02;
                m12 = AT.m12;
            }
            fRotateAccelerator = AT.fRotateAccelerator;
            return;
            //---- DONE ---
        }
        if (fAccelerator == TYPE_TRANSLATE &&
            AT.fAccelerator == TYPE_TRANSLATE) {
            m02 += AT.m02;
            m12 += AT.m12;
            return;
        }

        if (AT.fAccelerator == TYPE_ROTATE) {
            // OPERATOR IS ROTATE
            if (fAccelerator == TYPE_ROTATE) {
                // both the same type: try to rotate!!
                if ((m02 == AT.m02) && (m12 == AT.m12)) {
                    // Success!! do the rotate
                    m01 += AT.m01;
                    m01 = FirstCycle(m01);
                    fRotateAccelerator = GetFastRotate(m01);
                    if (fRotateAccelerator != ROTATE_0) {
                        ComputeCosAndSin();
                    } else {
                        fAccelerator = TYPE_IDENTITY;
                    }
                    return;
                }
                // failure: convert 'this' to an TYPE_AFFINE matrix
                RotateToAffine(fRotateAccelerator);
            }

            // The operator is still an accelerated rotate matrix;
            // convert it to an TYPE_AFFINE matrix before multiplying.

            double tmp[] = new double[6];
            AT.getMatrix(tmp);
            PostMatrixMult(tmp[0], tmp[1], tmp[2], tmp[3], tmp[4], tmp[5]);
            fAccelerator = GetAccelerator();
            return;
            //---- DONE ---
        }

        if (fAccelerator == TYPE_ROTATE) {
            // THIS IS ROTATE, OPERATOR IS NOT
            RotateToAffine(fRotateAccelerator);
            if (AT.fAccelerator == TYPE_TRANSLATE) {
                m02 += AT.m02;
                m12 += AT.m12;
                fAccelerator = TYPE_AFFINE;
            }
            else {
                double tmp[] = new double[6];
                AT.getMatrix(tmp);
                PostMatrixMult(tmp[0], tmp[1], tmp[2], tmp[3], tmp[4], tmp[5]);
                fAccelerator = GetAccelerator();
            }
            return;
            //---- DONE ---
        }

        if (AT.fAccelerator == TYPE_TRANSLATE) {
            // OPERATOR IS TRANSLATE
            m02 += AT.m02;
            m12 += AT.m12;
            if (fAccelerator != TYPE_TRANSLATE) {
                fAccelerator = GetAccelerator();
            }
            return;
            //---- DONE ---
        }

        // Wiped out the special cases: brute force matrix multiplication.
        PostMatrixMult(AT.m00, AT.m10, AT.m01, AT.m11, AT.m02, AT.m12);
        fAccelerator = GetAccelerator();
    }

    protected void preTranslate(double tx, double ty) {
        // this is really post translate
        if (tx != 0.0 || ty != 0.0) {
            if (fAccelerator == TYPE_IDENTITY) {
                m02 = tx;
                m12 = ty;
                fAccelerator = TYPE_TRANSLATE;
                return;
            }
            else if (fAccelerator == TYPE_ROTATE) {
                RotateToAffine(fRotateAccelerator);
                fAccelerator = TYPE_AFFINE;
            }
            m02 += tx;
            m12 += ty;
        }
    }
    protected void preScale(double sx, double sy, double cx, double cy) {
        double tx = cx;
        if (tx != 0.) tx *= 1.0 - sx;
        double ty = cy;
        if (ty != 0.) ty *= 1.0 - sy;
        switch (fAccelerator) {
        case TYPE_TRANSLATE:
            tx = sx * m02;
            ty = sy * m12;
        case TYPE_IDENTITY:
            InternalSetToIdentity();
            fAccelerator = TYPE_SCALE;
            m00 = sx;
            m11 = sy;
            m02 = tx;
            m12 = ty;
            return;
        case TYPE_ROTATE:
            RotateToAffine(fRotateAccelerator);
            fAccelerator = TYPE_AFFINE;
            break;
        default:
            m02 *= sx;
            m12 *= sy;
        }
        m00 *= sx;
        m11 *= sy;
        if (fAccelerator==TYPE_SCALE) return;
        m01 *= sx;
        m10 *= sy;
    }

    protected void preRotate(double theta, double cx, double cy) {
        theta = FirstCycle(theta);
        if (theta == 0.0 || theta == k360DegreeInRadians)
            return;

        switch (fAccelerator) {
        case TYPE_IDENTITY: {
            fAccelerator = TYPE_ROTATE;
            m01 = theta;
            m02 = cx;
            m12 = cy;
            fRotateAccelerator = GetFastRotate(theta);
            ComputeCosAndSin();
            return;
            }
        case TYPE_ROTATE:
            if ((m02 == cx) && (m12 == cy)) {
                // no translate after last rotate
                m01 = FirstCycle(m01 + theta);
                fRotateAccelerator = GetFastRotate(m01);
                if (fRotateAccelerator != ROTATE_0) {
                    ComputeCosAndSin();
                } else {
                    fAccelerator = TYPE_IDENTITY;
                }
                return;
            }
            else {
                RotateToAffine(fRotateAccelerator);
                fAccelerator = GetAccelerator();
            }
        case TYPE_TRANSLATE:
        case TYPE_SCALE:
        case TYPE_AFFINE:
            double nCos = Math.cos(theta);
            double nSin = Math.sin(theta);
            double tx;
            double ty;
            if (cx != 0. || cy != 0.) {
                tx = -cx*(nCos-1.0) + cy*nSin ;
                ty = -cx*nSin - cy*(nCos-1);
            }
            else {
                tx = 0.0;
                ty = 0.0;
            }
            PostMatrixMult(nCos, nSin, -nSin, nCos, tx, ty);
            fAccelerator = GetAccelerator();
            return;
        }
    }

    /**
     * Returns a Transform object representing the inverse transformation.
     * @exception NoninvertibleTransformException  if the matrix cannot be
     *                                         inverted.
     */
    public AffineTransform createInverse() throws
    NoninvertibleTransformException {
        AffineTransform inverse =
            new AffineTransform(m00, m10, m01, m11, m02, m12);
        inverse.invert();
        return inverse;
    }

    /**
     * Transforms the specified ptSrc and stores the result in ptDst.
     * If ptDst is null, a new Point2D object will be allocated before
     * storing. In either case, ptDst containing the transformed point
     * is returned for convenience.
     * Note that ptSrc and ptDst can the same. In this case, the input
     * point will be overwritten with the transformed point.
     */
    public Point2D transform(Point2D ptSrc, Point2D ptDst) {
        if (ptDst == null) {
            ptDst = new Point2D.Float();
        }

        double x = ptSrc.getX();
        double y = ptSrc.getY();

/*
        // Copy source coords into local variables in case src == dst
        ptDst.setLocation(m00 * x + m01 * y + m02, m10 * x + m11 * y + m12);
*/

        switch (fAccelerator) {
        case TYPE_IDENTITY:
            ptDst.setLocation(x, y);
            break;
        case TYPE_TRANSLATE:
            ptDst.setLocation(x + m02, y + m12);
            break;
        case TYPE_ROTATE:
            if (m02 == 0.0 && m12 == 0.0) {
                switch (fRotateAccelerator) {
                case ROTATE_0:
                    ptDst.setLocation(x, y);
                    break;
                case ROTATE_90:
                    ptDst.setLocation(-y, x);
                    break;
                case ROTATE_180:
                    ptDst.setLocation(-x, -y);
                    break;
                case ROTATE_270:
                    ptDst.setLocation(y, -x);
                    break;
                case ROTATE_SLOW:
                    // == Arbitrary Rotate Points ===
                    ptDst.setLocation(m11*x - m10*y, m10*x + m11*y);
                    break;
                }
            }
            else {
                switch (fRotateAccelerator) {
                case ROTATE_0:
                    ptDst.setLocation(x, y);
                    break;
                case ROTATE_90:
                    ptDst.setLocation(-y + m02 + m12, x - m02 + m12);
                    break;
                case ROTATE_180:
                    ptDst.setLocation(-x + 2*m02, -y + 2*m12);
                    break;
                case ROTATE_270:
                    ptDst.setLocation(y + m02 - m12, -x + m02 + m12);
                    break;
                case ROTATE_SLOW:
                    // == Arbitrary Rotate Points ===
                    ptDst.setLocation(
                        m11*x - m10*y + (1.0 - m11)*m02 + m10*m12,
                        m10*x + m11*y + (1.0 - m11)*m12 - m10*m02);
                    break;
                }
            }
            break;
        case TYPE_SCALE:
            ptDst.setLocation(x*m00 + m02, y*m11 + m12);
            break;
        case TYPE_AFFINE:
            ptDst.setLocation(m00*x + m01*y + m02, m10*x + m11*y + m12);
            break;
        }
        return ptDst;
    }

    /**
     * Transforms an array of point objects by this transform.
     * @param ptSrc The array containing the source point objects.
     * @param ptDst The array where the transform point objects are returned.
     * @param srcOff The offset to the first point object to be transformed
     * in the source array.
     * @param dstOff The offset to the location where the first transformed
     * point object is stored in the destination array.
     * @param numPts The number of point objects to be transformed.
     */
    public void transform(Point2D[] ptSrc, int srcOff,
        		  Point2D[] ptDst, int dstOff,
        		  int numPts) {
/*
        while (--numPts >= 0) {
            // Copy source coords into local variables in case src == dst
            double x = ptSrc[srcOff].getX();
            double y = ptSrc[srcOff++].getY();
            ptDst[dstOff++].setLocation(m00 * x + m01 * y + m02,
        				m10 * x + m11 * y + m12);
        }
*/
        int count = numPts;

        if (fAccelerator == TYPE_IDENTITY ||
            (fAccelerator == TYPE_ROTATE && fRotateAccelerator == ROTATE_0))
        {
            if (ptSrc != ptDst || srcOff != dstOff) {
                for ( ; --count >= 0; ++srcOff, ++dstOff)
                    ptDst[dstOff].setLocation(ptSrc[srcOff]);
            }
            return;
        }

        switch (fAccelerator) {
        case TYPE_TRANSLATE:
            for ( ; --count >= 0; ++srcOff, ++dstOff) {
                ptDst[dstOff].setLocation(
                    ptSrc[srcOff].getX() + m02,
                    ptSrc[srcOff].getY() + m12);
            }
            break;
        case TYPE_ROTATE:
            if (m02==0.0 && m12==0.0) {
                switch (fRotateAccelerator) {
                case ROTATE_90:
                    for ( ; --count >= 0; ++srcOff, ++dstOff)
                        ptDst[dstOff].setLocation(
                            -ptSrc[srcOff].getY(), ptSrc[srcOff].getX());
                    break;
                case ROTATE_180:
                    for ( ; --count >= 0; ++srcOff, ++dstOff)
                        ptDst[dstOff].setLocation(
                            -ptSrc[srcOff].getX(), -ptSrc[srcOff].getY());
                    break;
                case ROTATE_270:
                    for ( ; --count >= 0; ++srcOff, ++dstOff)
                        ptDst[dstOff].setLocation(
                            ptSrc[srcOff].getY(), -ptSrc[srcOff].getX());
                    break;
                case ROTATE_SLOW:
                    // == Arbitrary Rotate Points ===
                    double cos = m11;
                    double sin = m10;
                    for ( ; --count >= 0; ++srcOff, ++dstOff) {
                        Point2D src = ptSrc[srcOff];
                        double sx = src.getX();
                        double sy = src.getY();
                        ptDst[dstOff].setLocation(
                            cos*sx - sin*sy, sin*sx + cos*sy);
                    }
                    break;
                }
            }
            else {
                double translateX;
                double translateY;
                switch (fRotateAccelerator) {
                case ROTATE_90:
                    translateX = m02 + m12;
                    translateY = -m02 + m12;
                    for ( ; --count >= 0; ++srcOff, ++dstOff)
                        ptDst[dstOff].setLocation(
                            -ptSrc[srcOff].getY() + translateX,
                            ptSrc[srcOff].getX() + translateY);
                    break;
                case ROTATE_180:
                    translateX = 2*m02;
                    translateY = 2*m12;
                    for ( ; --count >= 0; ++srcOff, ++dstOff)
                        ptDst[dstOff].setLocation(
                            -ptSrc[srcOff].getX() + translateX,
                            -ptSrc[srcOff].getY() + translateY);
                    break;
                case ROTATE_270:
                    translateX = m02 - m12;
                    translateY = m02 + m12;
                    for ( ; --count >= 0; ++srcOff, ++dstOff)
                        ptDst[dstOff].setLocation(
                            ptSrc[srcOff].getY() + translateX,
                            -ptSrc[srcOff].getX() + translateY);
                    break;
                case ROTATE_SLOW:
                    // == Arbitrary Rotate Points ===
                    double cos = m11;
                    double sin = m10;
                    double oneMinus = 1.0 - cos;
                    translateX = oneMinus*m02 + sin*m12;
                    translateY = oneMinus*m12 - sin*m02;
                    for ( ; --count >= 0; ++srcOff, ++dstOff) {
                        Point2D src = ptSrc[srcOff];
                        double sx = src.getX();
                        double sy = src.getY();
                        ptDst[dstOff].setLocation(
                            cos*sx - sin*sy + translateX,
                            sin*sx + cos*sy + translateY);
                    }
                    break;
                }
            }
            break;
        case TYPE_SCALE:
            for ( ; --count >= 0; ++srcOff, ++dstOff)
                ptDst[dstOff].setLocation(
                    ptSrc[srcOff].getX() * m00 + m02,
                    ptSrc[srcOff].getY() * m11 + m12);
            break;
        case TYPE_AFFINE:
            for ( ; --count >= 0; ++srcOff, ++dstOff) {
                Point2D src = ptSrc[srcOff];
                double sx = src.getX();
                double sy = src.getY();

                ptDst[dstOff].setLocation(
                    m00*sx + m01*sy + m02, m10*sx + m11*sy + m12);
            }
            break;
        }
    }

    /**
     * Transforms an array of floating point coordinates by this transform.
     * @param ptSrc The array containing the source point coordinates.
     * Each point is stored as a pair of x,y coordinates.
     * @param ptDst The array where the transformed point coordinates are
     * returned.  Each point is stored as a pair of x,y coordinates.
     * @param srcOff The offset to the first point to be transformed
     * in the source array.
     * @param dstOff The offset to the location where the first transformed
     * point is stored in the destination array.
     * @param numPts The number of points to be transformed.
     */
    public void transform(float[] srcPts, int srcOff,
        		  float[] dstPts, int dstOff,
        		  int numPts) {
/*
        while (--numPts >= 0) {
            double x = srcPts[srcOff++];
            double y = srcPts[srcOff++];
            dstPts[dstOff++] = (float) (m00 * x + m01 * y + m02);
            dstPts[dstOff++] = (float) (m10 * x + m11 * y + m12);
        }
*/
        int count = numPts;

        if (fAccelerator == TYPE_IDENTITY ||
            (fAccelerator == TYPE_ROTATE && fRotateAccelerator == ROTATE_0))
        {
            if (srcPts != dstPts || srcOff != dstOff) {	// if not overlapping
                while (--count >= 0) {
                    dstPts[dstOff++] = srcPts[srcOff++];  // x
                    dstPts[dstOff++] = srcPts[srcOff++];  // y
                }
            }
            return;
        }

        float tmp;

        switch (fAccelerator) {
        case TYPE_TRANSLATE:
            while (--count >= 0) {
                dstPts[dstOff++] = srcPts[srcOff++] + (float)m02;  // x
                dstPts[dstOff++] = srcPts[srcOff++] + (float)m12;  // y
            }
            break;
        case TYPE_ROTATE:
            if (m02==0.0 && m12==0.0) {
                switch (fRotateAccelerator) {
                case ROTATE_90:
                    while (--count >= 0) {
                        tmp = srcPts[srcOff++]; // x
                        dstPts[dstOff++] = -srcPts[srcOff++]; // y
                        dstPts[dstOff++] = tmp;
                    }
                    break;
                case ROTATE_180:
                    while (--count >= 0) {
                        dstPts[dstOff++] = -srcPts[srcOff++]; // x
                        dstPts[dstOff++] = -srcPts[srcOff++]; // y
                    }
                    break;
                case ROTATE_270:
                    while (--count >= 0) {
                        tmp = srcPts[srcOff++]; // x
                        dstPts[dstOff++] = srcPts[srcOff++]; // y
                        dstPts[dstOff++] = tmp;
                    }
                    break;
                case ROTATE_SLOW:
                    // == Arbitrary Rotate Points ===
                    double cos = m11;
                    double sin = m10;
                    while (--count >= 0) {
                        double sx = srcPts[srcOff++]; // x
                        double sy = srcPts[srcOff++]; // y
                        dstPts[dstOff++] = (float)(cos*sx - sin*sy);
                        dstPts[dstOff++] = (float)(sin*sx + cos*sy);
                    }
                    break;
                }
            }
            else {
                float translateX;
                float translateY;
                switch (fRotateAccelerator) {
                case ROTATE_90:
                    translateX = (float)(m02 + m12);
                    translateY = (float)(-m02 + m12);
                    while (--count >= 0) {
                        tmp = srcPts[srcOff++]; // x
                        dstPts[dstOff++] = -srcPts[srcOff++] + translateX; // y
                        dstPts[dstOff++] = tmp + translateY;
                    }
                    break;
                case ROTATE_180:
                    translateX = (float)(2*m02);
                    translateY = (float)(2*m12);
                    while (--count >= 0) {
                        dstPts[dstOff++] = -srcPts[srcOff++] + translateX; // x
                        dstPts[dstOff++] = -srcPts[srcOff++] + translateY; // y
                    }
                    break;
                case ROTATE_270:
                    translateX = (float)(m02 - m12);
                    translateY = (float)(m02 + m12);
                    while (--count >= 0) {
                        tmp = srcPts[srcOff++]; // x
                        dstPts[dstOff++] = srcPts[srcOff++] + translateX; // y
                        dstPts[dstOff++] = tmp + translateY;
                    }
                    break;
                case ROTATE_SLOW:
                    // == Arbitrary Rotate Points ===
                    double cos = m11;
                    double sin = m10;
                    double oneMinus = 1.0 - cos;
                    translateX = (float)(oneMinus*m02 + sin*m12);
                    translateY = (float)(oneMinus*m12 - sin*m02);
                    while (--count >= 0) {
                        double sx = srcPts[srcOff++]; // x
                        double sy = srcPts[srcOff++]; // y
                        dstPts[dstOff++] =
                            (float)(cos*sx - sin*sy) + translateX;
                        dstPts[dstOff++] =
                            (float)(sin*sx + cos*sy) + translateY;
                    }
                    break;
                }
            }
            break;
        case TYPE_SCALE:
            while (--count >= 0) {
                dstPts[dstOff++] =
                    (float)((double)srcPts[srcOff++] * m00 + m02); // x
                dstPts[dstOff++] =
                    (float)((double)srcPts[srcOff++] * m11 + m12); // y
            }
            break;
        case TYPE_AFFINE:
            while (--count >= 0) {
                double sx = srcPts[srcOff++];  // x
                double sy = srcPts[srcOff++];  // y
                dstPts[dstOff++] = (float)(m00*sx + m01*sy + m02);
                dstPts[dstOff++] = (float)(m10*sx + m11*sy + m12);
            }
            break;
        }
    }

    /**
     * Transforms an array of double precision coordinates by this transform.
     * @param srcPts The array containing the source point coordinates.
     * Each point is stored as a pair of x,y coordinates.
     * @param dstPts The array where the transformed point coordinates are
     * returned.  Each point is stored as a pair of x,y coordinates.
     * @param srcOff The offset to the first point to be transformed
     * in the source array.
     * @param dstOff The offset to the location where the first transformed
     * point is stored in the destination array.
     * @param numPts The number of point objects to be transformed.
     */
    public void transform(double[] srcPts, int srcOff,
        		  double[] dstPts, int dstOff,
        		  int numPts) {
/*
        while (--numPts >= 0) {
            double x = srcPts[srcOff++];
            double y = srcPts[srcOff++];
            dstPts[dstOff++] = m00 * x + m01 * y + m02;
            dstPts[dstOff++] = m10 * x + m11 * y + m12;
        }
*/

        int count = numPts;

        if (fAccelerator == TYPE_IDENTITY ||
            (fAccelerator == TYPE_ROTATE && fRotateAccelerator == ROTATE_0))
        {
            if (srcPts != dstPts || srcOff != dstOff) {	// if not overlapping
                while (--count >= 0) {
                    dstPts[dstOff++] = srcPts[srcOff++];  // x
                    dstPts[dstOff++] = srcPts[srcOff++];  // y
                }
            }
            return;
        }

        double tmp;

        switch (fAccelerator) {
        case TYPE_TRANSLATE:
            while (--count >= 0) {
                dstPts[dstOff++] = srcPts[srcOff++] + m02;  // x
                dstPts[dstOff++] = srcPts[srcOff++] + m12;  // y
            }
            break;
        case TYPE_ROTATE:
            if (m02==0.0 && m12==0.0) {
                switch (fRotateAccelerator) {
                case ROTATE_90:
                    while (--count >= 0) {
                        tmp = srcPts[srcOff++]; // x
                        dstPts[dstOff++] = -srcPts[srcOff++]; // y
                        dstPts[dstOff++] = tmp;
                    }
                    break;
                case ROTATE_180:
                    while (--count >= 0) {
                        dstPts[dstOff++] = -srcPts[srcOff++]; // x
                        dstPts[dstOff++] = -srcPts[srcOff++]; // y
                    }
                    break;
                case ROTATE_270:
                    while (--count >= 0) {
                        tmp = srcPts[srcOff++]; // x
                        dstPts[dstOff++] = srcPts[srcOff++]; // y
                        dstPts[dstOff++] = tmp;
                    }
                    break;
                case ROTATE_SLOW:
                    // == Arbitrary Rotate Points ===
                    double cos = m11;
                    double sin = m10;
                    while (--count >= 0) {
                        double sx = srcPts[srcOff++]; // x
                        double sy = srcPts[srcOff++]; // y
                        dstPts[dstOff++] = cos*sx - sin*sy;
                        dstPts[dstOff++] = sin*sx + cos*sy;
                    }
                    break;
                }
            }
            else {
                double translateX;
                double translateY;
                switch (fRotateAccelerator) {
                case ROTATE_90:
                    translateX = m02 + m12;
                    translateY = -m02 + m12;
                    while (--count >= 0) {
                        tmp = srcPts[srcOff++]; // x
                        dstPts[dstOff++] = -srcPts[srcOff++] + translateX; // y
                        dstPts[dstOff++] = tmp + translateY;
                    }
                    break;
                case ROTATE_180:
                    translateX = 2*m02;
                    translateY = 2*m12;
                    while (--count >= 0) {
                        dstPts[dstOff++] = -srcPts[srcOff++] + translateX; // x
                        dstPts[dstOff++] = -srcPts[srcOff++] + translateY; // y
                    }
                    break;
                case ROTATE_270:
                    translateX = m02 - m12;
                    translateY = m02 + m12;
                    while (--count >= 0) {
                        tmp = srcPts[srcOff++]; // x
                        dstPts[dstOff++] = srcPts[srcOff++] + translateX; // y
                        dstPts[dstOff++] = tmp + translateY;
                    }
                    break;
                case ROTATE_SLOW:
                    // == Arbitrary Rotate Points ===
                    double cos = m11;
                    double sin = m10;
                    double oneMinus = 1.0 - cos;
                    translateX = oneMinus*m02 + sin*m12;
                    translateY = oneMinus*m12 - sin*m02;
                    while (--count >= 0) {
                        double sx = srcPts[srcOff++]; // x
                        double sy = srcPts[srcOff++]; // y
                        dstPts[dstOff++] = cos*sx - sin*sy + translateX;
                        dstPts[dstOff++] = sin*sx + cos*sy + translateY;
                    }
                    break;
                }
            }
            break;
        case TYPE_SCALE:
            while (--count >= 0) {
                dstPts[dstOff++] = srcPts[srcOff++] * m00 + m02; // x
                dstPts[dstOff++] = srcPts[srcOff++] * m11 + m12; // y
            }
            break;
        case TYPE_AFFINE:
            while (--count >= 0) {
                double sx = srcPts[srcOff++];  // x
                double sy = srcPts[srcOff++];  // y
                dstPts[dstOff++] = m00*sx + m01*sy + m02;
                dstPts[dstOff++] = m10*sx + m11*sy + m12;
            }
            break;
        }
    }

    /**
     * Transforms an array of floating point coordinates by this transform,
     * storing the results into an array of doubles.
     * @param srcPts The array containing the source point coordinates.
     * Each point is stored as a pair of x,y coordinates.
     * @param dstPts The array where the transformed point coordinates are
     * returned.  Each point is stored as a pair of x,y coordinates.
     * @param srcOff The offset to the first point to be transformed
     * in the source array.
     * @param dstOff The offset to the location where the first transformed
     * point is stored in the destination array.
     * @param numPts The number of points to be transformed.
     */
    public void transform(float[] srcPts, int srcOff,
        		  double[] dstPts, int dstOff,
        		  int numPts) {
/*
        while (--numPts >= 0) {
            double x = srcPts[srcOff++];
            double y = srcPts[srcOff++];
            dstPts[dstOff++] = m00 * x + m01 * y + m02;
            dstPts[dstOff++] = m10 * x + m11 * y + m12;
        }
*/

        int count = numPts;

        if (fAccelerator == TYPE_IDENTITY ||
            (fAccelerator == TYPE_ROTATE && fRotateAccelerator == ROTATE_0))
        {
            while (--count >= 0) {
                dstPts[dstOff++] = srcPts[srcOff++];  // x
                dstPts[dstOff++] = srcPts[srcOff++];  // y
            }
            return;
        }

        double tmp;

        switch (fAccelerator) {
        case TYPE_TRANSLATE:
            while (--count >= 0) {
                dstPts[dstOff++] = (double)srcPts[srcOff++] + m02;  // x
                dstPts[dstOff++] = (double)srcPts[srcOff++] + m12;  // y
            }
            break;
        case TYPE_ROTATE:
            if (m02==0.0 && m12==0.0) {
                switch (fRotateAccelerator) {
                case ROTATE_90:
                    while (--count >= 0) {
                        tmp = srcPts[srcOff++]; // x
                        dstPts[dstOff++] = -srcPts[srcOff++]; // y
                        dstPts[dstOff++] = tmp;
                    }
                    break;
                case ROTATE_180:
                    while (--count >= 0) {
                        dstPts[dstOff++] = -srcPts[srcOff++]; // x
                        dstPts[dstOff++] = -srcPts[srcOff++]; // y
                    }
                    break;
                case ROTATE_270:
                    while (--count >= 0) {
                        tmp = srcPts[srcOff++]; // x
                        dstPts[dstOff++] = srcPts[srcOff++]; // y
                        dstPts[dstOff++] = tmp;
                    }
                    break;
                case ROTATE_SLOW:
                    // == Arbitrary Rotate Points ===
                    double cos = m11;
                    double sin = m10;
                    while (--count >= 0) {
                        double sx = srcPts[srcOff++]; // x
                        double sy = srcPts[srcOff++]; // y
                        dstPts[dstOff++] = cos*sx - sin*sy;
                        dstPts[dstOff++] = sin*sx + cos*sy;
                    }
                    break;
                }
            }
            else {
                double translateX;
                double translateY;
                switch (fRotateAccelerator) {
                case ROTATE_90:
                    translateX = m02 + m12;
                    translateY = -m02 + m12;
                    while (--count >= 0) {
                        tmp = srcPts[srcOff++]; // x
                        dstPts[dstOff++] = -srcPts[srcOff++] + translateX; // y
                        dstPts[dstOff++] = tmp + translateY;
                    }
                    break;
                case ROTATE_180:
                    translateX = 2*m02;
                    translateY = 2*m12;
                    while (--count >= 0) {
                        dstPts[dstOff++] = -srcPts[srcOff++] + translateX; // x
                        dstPts[dstOff++] = -srcPts[srcOff++] + translateY; // y
                    }
                    break;
                case ROTATE_270:
                    translateX = m02 - m12;
                    translateY = m02 + m12;
                    while (--count >= 0) {
                        tmp = srcPts[srcOff++]; // x
                        dstPts[dstOff++] = srcPts[srcOff++] + translateX; // y
                        dstPts[dstOff++] = tmp + translateY;
                    }
                    break;
                case ROTATE_SLOW:
                    // == Arbitrary Rotate Points ===
                    double cos = m11;
                    double sin = m10;
                    double oneMinus = 1.0 - cos;
                    translateX = oneMinus*m02 + sin*m12;
                    translateY = oneMinus*m12 - sin*m02;
                    while (--count >= 0) {
                        double sx = srcPts[srcOff++]; // x
                        double sy = srcPts[srcOff++]; // y
                        dstPts[dstOff++] = cos*sx - sin*sy + translateX;
                        dstPts[dstOff++] = sin*sx + cos*sy + translateY;
                    }
                    break;
                }
            }
            break;
        case TYPE_SCALE:
            while (--count >= 0) {
                dstPts[dstOff++] = srcPts[srcOff++] * m00 + m02; // x
                dstPts[dstOff++] = srcPts[srcOff++] * m11 + m12; // y
            }
            break;
        case TYPE_AFFINE:
            while (--count >= 0) {
                double sx = srcPts[srcOff++];  // x
                double sy = srcPts[srcOff++];  // y
                dstPts[dstOff++] = m00*sx + m01*sy + m02;
                dstPts[dstOff++] = m10*sx + m11*sy + m12;
            }
            break;
        }
    }

    /**
     * Transforms an array of double precision coordinates by this transform,
     * storing the results into an array of floats.
     * @param srcPts The array containing the source point coordinates.
     * Each point is stored as a pair of x,y coordinates.
     * @param dstPts The array where the transformed point coordinates are
     * returned.  Each point is stored as a pair of x,y coordinates.
     * @param srcOff The offset to the first point to be transformed
     * in the source array.
     * @param dstOff The offset to the location where the first transformed
     * point is stored in the destination array.
     * @param numPts The number of point objects to be transformed.
     */
    public void transform(double[] srcPts, int srcOff,
        		  float[] dstPts, int dstOff,
        		  int numPts) {
/*
        while (--numPts >= 0) {
            double x = srcPts[srcOff++];
            double y = srcPts[srcOff++];
            dstPts[dstOff++] = (float) (m00 * x + m01 * y + m02);
            dstPts[dstOff++] = (float) (m10 * x + m11 * y + m12);
        }
*/
        int count = numPts;

        if (fAccelerator == TYPE_IDENTITY ||
            (fAccelerator == TYPE_ROTATE && fRotateAccelerator == ROTATE_0))
        {
            while (--count >= 0) {
                dstPts[dstOff++] = (float)srcPts[srcOff++];  // x
                dstPts[dstOff++] = (float)srcPts[srcOff++];  // y
            }
            return;
        }

        double tmp;

        switch (fAccelerator) {
        case TYPE_TRANSLATE:
            while (--count >= 0) {
                dstPts[dstOff++] = (float)(srcPts[srcOff++] + m02);  // x
                dstPts[dstOff++] = (float)(srcPts[srcOff++] + m12);  // y
            }
            break;
        case TYPE_ROTATE:
            if (m02==0.0 && m12==0.0) {
                switch (fRotateAccelerator) {
                case ROTATE_90:
                    while (--count >= 0) {
                        tmp = srcPts[srcOff++]; // x
                        dstPts[dstOff++] = (float)(-srcPts[srcOff++]); // y
                        dstPts[dstOff++] = (float)tmp;
                    }
                    break;
                case ROTATE_180:
                    while (--count >= 0) {
                        dstPts[dstOff++] = (float)(-srcPts[srcOff++]); // x
                        dstPts[dstOff++] = (float)(-srcPts[srcOff++]); // y
                    }
                    break;
                case ROTATE_270:
                    while (--count >= 0) {
                        tmp = srcPts[srcOff++]; // x
                        dstPts[dstOff++] = (float)srcPts[srcOff++]; // y
                        dstPts[dstOff++] = (float)tmp;
                    }
                    break;
                case ROTATE_SLOW:
                    // == Arbitrary Rotate Points ===
                    double cos = m11;
                    double sin = m10;
                    while (--count >= 0) {
                        double sx = srcPts[srcOff++]; // x
                        double sy = srcPts[srcOff++]; // y
                        dstPts[dstOff++] = (float)(cos*sx - sin*sy);
                        dstPts[dstOff++] = (float)(sin*sx + cos*sy);
                    }
                    break;
                }
            }
            else {
                double translateX;
                double translateY;
                switch (fRotateAccelerator) {
                case ROTATE_90:
                    translateX = m02 + m12;
                    translateY = -m02 + m12;
                    while (--count >= 0) {
                        tmp = srcPts[srcOff++]; // x
                        dstPts[dstOff++] =
                            (float)(-srcPts[srcOff++] + translateX); // y
                        dstPts[dstOff++] =
                            (float)(tmp + translateY);
                    }
                    break;
                case ROTATE_180:
                    translateX = 2*m02;
                    translateY = 2*m12;
                    while (--count >= 0) {
                        dstPts[dstOff++] =
                            (float)(-srcPts[srcOff++] + translateX);  // x
                        dstPts[dstOff++] =
                            (float)(-srcPts[srcOff++] + translateY);  // y
                    }
                    break;
                case ROTATE_270:
                    translateX = m02 - m12;
                    translateY = m02 + m12;
                    while (--count >= 0) {
                        tmp = srcPts[srcOff++]; // x
                        dstPts[dstOff++] =
                            (float)(srcPts[srcOff++] + translateX); // y
                        dstPts[dstOff++] =
                            (float)(tmp + translateY);
                    }
                    break;
                case ROTATE_SLOW:
                    // == Arbitrary Rotate Points ===
                    double cos = m11;
                    double sin = m10;
                    double oneMinus = 1.0 - cos;
                    translateX = oneMinus*m02 + sin*m12;
                    translateY = oneMinus*m12 - sin*m02;
                    while (--count >= 0) {
                        double sx = srcPts[srcOff++]; // x
                        double sy = srcPts[srcOff++]; // y
                        dstPts[dstOff++] =
                            (float)(cos*sx - sin*sy + translateX);
                        dstPts[dstOff++] =
                            (float)(sin*sx + cos*sy + translateY);
                    }
                    break;
                }
            }
            break;
        case TYPE_SCALE:
            while (--count >= 0) {
                dstPts[dstOff++] = (float)(srcPts[srcOff++] * m00 + m02); // x
                dstPts[dstOff++] = (float)(srcPts[srcOff++] * m11 + m12); // y
            }
            break;
        case TYPE_AFFINE:
            while (--count >= 0) {
                double sx = srcPts[srcOff++];  // x
                double sy = srcPts[srcOff++];  // y
                dstPts[dstOff++] = (float)(m00*sx + m01*sy + m02);
                dstPts[dstOff++] = (float)(m10*sx + m11*sy + m12);
            }
            break;
        }
    }

    /**
     * Inverse transforms the specified ptSrc and stores the result in ptDst.
     * If ptDst is null, a new Point2D object will be allocated before
     * storing. In either case, ptDst containing the transformed point
     * is returned for convenience.
     * Note that ptSrc and ptDst can the same. In this case, the input
     * point will be overwritten with the transformed point.
     * @param ptSrc The point to be inverse transformed.
     * @param ptDst The resulting transformed point.
     * @exception NoninvertibleTransformException  if the matrix cannot be
     *                                         inverted.
     */
    public Point2D inverseTransform(Point2D ptSrc, Point2D ptDst)
        throws NoninvertibleTransformException {
/*
        try {
            createInverse().transform(ptSrc, ptDst);
        } catch (NoninvertibleTransformException e) {
            // REMIND:  Do something
        }
*/
        if (ptDst == null) {
            ptDst = new Point2D.Float();
        }

        double x = ptSrc.getX();
        double y = ptSrc.getY();

        switch (fAccelerator) {
        case TYPE_IDENTITY:
            ptDst.setLocation(x, y);
            break;
        case TYPE_TRANSLATE:
            ptDst.setLocation(x - m02, y - m12);
            break;
        case TYPE_SCALE:
            if (m00 != 0.0 || m11 != 0.0)
                ptDst.setLocation((x - m02)/m00, (y - m12)/m11);
            else
                throw
                    new NoninvertibleTransformException(
                        "inverseTransform(): Singular Scale Transform");
            break;
        case TYPE_ROTATE:
            switch (fRotateAccelerator) {
            case ROTATE_0:
                ptDst.setLocation(x, y);
                break;
            case ROTATE_90:
                ptDst.setLocation(y + m02 - m12, -x + m02 + m12);
                break;
            case ROTATE_180:
                ptDst.setLocation(-x + 2.0*m02, -y + 2.0*m12);
                break;
            case ROTATE_270:
                ptDst.setLocation(-y + m02 + m12, x - m02 + m12);
                break;
            case ROTATE_SLOW:
                // == Arbitrary Unrotate Point ===
                double cos = m11;
                double sin = m10;
                double oneMinus = 1.0 - cos;
                double oldX = x;
                ptDst.setLocation(
                    cos*oldX + sin*y + m02*oneMinus - m12*sin,
                    -sin*oldX + cos*y + m12*oneMinus + m02*sin);
                break;
            }
            break;
        case TYPE_AFFINE:
            double det = m00*m11 - m01*m10;
            if (det != 0.0)
                ptDst.setLocation(
                    (m11*(x-m02) - m01*(y-m12))/det,
                    (-m10*(x-m02) + m00*(y-m12))/det);
            else
                throw
                    new NoninvertibleTransformException(
                        "inverseTransform(): Singular Affine Transform");
            break;
        }
        return ptDst;
    }

    /**
     * Inverse transforms an array of double precision coordinates by
     * this transform.  
     * @param srcPts The array containing the source point coordinates.
     * Each point is stored as a pair of x,y coordinates.
     * @param dstPts The array where the transformed point coordinates are
     * returned.  Each point is stored as a pair of x,y coordinates.
     * @param srcOff The offset to the first point to be transformed
     * in the source array.
     * @param dstOff The offset to the location where the first transformed
     * point is stored in the destination array.
     * @param numPts The number of point objects to be transformed.
     * @exception NoninvertibleTransformException  if the matrix cannot be
     *                                         inverted.
     */
    public void inverseTransform(double[] srcPts, int srcOff,
                                 double[] dstPts, int dstOff,
                                 int numPts)
    throws NoninvertibleTransformException {
/*
        AffineTransform at;
        try {
            at = (AffineTransform) createInverse();
        } catch (NoninvertibleTransformException e) {
            throw e;
        }
        at.transform(srcPts, srcOff, dstPts, dstOff, numPts);
*/ 
        int count = numPts;
        if (fAccelerator == TYPE_IDENTITY ||
            (fAccelerator == TYPE_ROTATE && fRotateAccelerator == ROTATE_0))
        {
            if (srcPts != dstPts || srcOff != dstOff) {	// if not overlapping...
                while (--count >= 0) {
                    dstPts[dstOff++] = srcPts[srcOff++];  // x
                    dstPts[dstOff++] = srcPts[srcOff++];  // y
                }
            }
            return;
        }

        switch (fAccelerator) {
        case TYPE_TRANSLATE:
            while (--count >= 0) {
                dstPts[dstOff++] = srcPts[srcOff++] - m02;  // x
                dstPts[dstOff++] = srcPts[srcOff++] - m12;  // y
            }
            break;
        case TYPE_SCALE:
            if (m00 != 0.0 || m11 != 0.0)
                while (--count >= 0) {
                    dstPts[dstOff++] = (srcPts[srcOff++] - m02)/m00;  // x
                    dstPts[dstOff++] = (srcPts[srcOff++] - m12)/m11;  // y
                }
            else
                throw
                    new NoninvertibleTransformException(
                        "inverseTransform(): Singular Scale Transform");
            break;
        case TYPE_ROTATE:
            switch (fRotateAccelerator) {
            case ROTATE_90:
                while (--count >= 0) {
                    double tmp = srcPts[srcOff++];  // x
                    dstPts[dstOff++] = srcPts[srcOff++] + m02 - m12;  // y 
                    dstPts[dstOff++] = -tmp + m02 + m12;
                }
                break;
            case ROTATE_180:
                while (--count >= 0) {
                    dstPts[dstOff++] = -srcPts[srcOff++] + 2.0*m02;  // x
                    dstPts[dstOff++] = -srcPts[srcOff++] + 2.0*m12;  // y
                }
                break;
            case ROTATE_270:
                while (--count >= 0) {
                    double tmp = srcPts[srcOff++];  // x
                    dstPts[dstOff++] = -srcPts[srcOff++] + m02 + m12;  // y 
                    dstPts[dstOff++] = tmp - m02 + m12;
                }
                break;
            case ROTATE_SLOW:
                // == Arbitrary Unrotate Point ===
                double cos = m11;
                double sin = m10;
                double oneMinus = 1.0 - cos;
                while (--count >= 0) {
                    double sx = srcPts[srcOff++];  // x
                    double sy = srcPts[srcOff++];  // y
                    dstPts[dstOff++] =
                        cos*sx + sin*sy + m02*oneMinus - m12*sin;
                    dstPts[dstOff++] =
                        -sin*sx + cos*sy + m12*oneMinus + m02*sin;
                }
                break;
            }
            break;
        case TYPE_AFFINE:
            double det = m00*m11 - m01*m10;
            if (det != 0.0)
                while (--count >= 0) {
                    double sx = srcPts[srcOff++];  // x
                    double sy = srcPts[srcOff++];  // y
                    dstPts[dstOff++] = (m11*(sx-m02) - m01*(sy-m12))/det;
                    dstPts[dstOff++] = (-m10*(sx-m02) + m00*(sy-m12))/det;
                }
            else
                throw
                   new NoninvertibleTransformException(
                       "inverseTransform(): Singular Affine Transform");
            break;
        }
    }

    /**
     * Transforms a shape object by this transform.
     * @see Transform#createTransformedShape
     */
    public Shape createTransformedShape(Shape pSrc) {
        throw new NoSuchMethodError ("Not implemented yet");
    }

    // Round values to sane precision for printing
    // Note that Math.sin(Math.PI) has an error of about 10^-16
    private static double _matround(double matval) {
        return Math.rint(matval * 1E15) / 1E15;
    }

    /**
     * Returns a String that represents the value of this Object.
     */
    public String toString() {
        return ("AffineTransform[["
        	+ _matround(m00) + ", "
        	+ _matround(m01) + ", "
        	+ _matround(m02) + "], ["
        	+ _matround(m10) + ", "
        	+ _matround(m11) + ", "
        	+ _matround(m12) + "]]");
    }

    /**
     * Returns a copy of this AffineTransform object.
     */
    public Object clone() {
        return (new AffineTransform(m00, m10, m01, m11, m02, m12));
    }

    /**
     * Supporting acceleration:
     */

    private void InternalSetToIdentity() {
        m00 = m11 = 1.0;
        m10 = m01 = m02 = m12 = 0.0;
    }
    private void PreMatrixMult(
        double m00, double m10,
        double m01, double m11,
        double m02, double m12) {

/* using my notation, this is how the matrix mutiplication is done...
 | m00, m10, 0 |   | this.m00, this.m10, 0 |
 | m01, m11, 0 | * | this.m01, this.m11, 0 |
 | m02, m12, 1 |   | this.m02, this.m12, 1 |
   | m00*this.m00 + m10*this.m01,       m00*this.m10 + m10*this.m11,       0 |
 = | m01*this.m00 + m11*this.m01,       m01*this.m10 + m11*this.m11,       0 |
   | m02*this.m00 + m12*this.m01 + m12, m02*this.m10 + m12*this.m11 + m12, 1 |
 */
        double tmp = this.m01;
        this.m02 += m02 * this.m00 + m12 * tmp;
        this.m01  = m01 * this.m00 + m11 * tmp;
        this.m00  = m00 * this.m00 + m10 * tmp;
        tmp = this.m11;
        this.m12 += m02 * this.m10 + m12 * tmp;
        this.m11  = m01 * this.m10 + m11 * tmp;
        this.m10  = m00 * this.m10 + m10 * tmp;
    }
    private void PostMatrixMult(
        double m00, double m10,
        double m01, double m11,
        double m02, double m12) {

/* using my notation, this is how the matrix mutiplication is done...
 | this.m00, this.m10, 0 |   | m00, m10, 0 |
 | this.m01, this.m11, 0 | * | m01, m11, 0 |
 | this.m02, this.m12, 1 |   | m02, m12, 1 |
   | this.m00*m00 + this.m10*m01,       this.m00*m10 + this.m10*m11,       0 |
 = | this.m01*m00 + this.m11*m01,       this.m01*m10 + this.m11*m11,       0 |
   | this.m02*m00 + this.m12*m01 + m02, this.m02*m10 + this.m12*m11 + m12, 1 |
 */
        double tmp = this.m00;
        this.m00 = tmp * m00 + this.m10 * m01;
        this.m10 = tmp * m10 + this.m10 * m11;
        tmp = this.m01;
        this.m01 = tmp * m00 + this.m11 * m01;
        this.m11 = tmp * m10 + this.m11 * m11;
        tmp = this.m02;
        this.m02 = tmp * m00 + this.m12 * m01 + m02;
        this.m12 = tmp * m10 + this.m12 * m11 + m12;
    }

    // Data members
    private int fAccelerator = TYPE_IDENTITY;
    private int fRotateAccelerator = ROTATE_0;

    private int GetAccelerator() {
        // ----- shear -----
        if (m01 != 0.0 || m10 != 0.0)
            return (TYPE_AFFINE);
        // ----- scale -----
        if (m00 != 1.0 || m11 != 1.0)
            return (TYPE_SCALE);
        // ----- translate -----
        if (m02 != 0.0 || m12 != 0.0)
            return (TYPE_TRANSLATE);
        // ----- identity -----
        return (TYPE_IDENTITY);
    }

    private static final double k90DegreeInRadians = Math.PI / 2.;
    private static final double k180DegreeInRadians = Math.PI;
    private static final double k270DegreeInRadians =
        k90DegreeInRadians + k180DegreeInRadians;
    private static final double k360DegreeInRadians =
        k180DegreeInRadians * 2.;

    private int GetFastRotate(double radians) {
        if (radians == 0.0) return (ROTATE_0);
        if (radians == k90DegreeInRadians) return (ROTATE_90);
        if (radians == k180DegreeInRadians) return (ROTATE_180);
        if (radians == k270DegreeInRadians) return (ROTATE_270);
        return (ROTATE_SLOW);
    }
    private void RotateToAffine(int rotateAccelerator) {
        /*  RotateToAffine:
            rotate = | cos  sin  0 |
                     | -sin cos  0 |
                     | Tx   Ty   1 |
            where Tx = -Cx*(cos-1.0) + Cy*(sin)
              and Ty = -Cx*(sin) - Cy*(cos-1.0)
        */
        double centerX = m02; // centre X
        double centerY = m12; // centre Y
        double nCos = 1.0;
        double nSin = 0.0;
        switch (rotateAccelerator) {
        case ROTATE_0:
            break;
        case ROTATE_90:
            nCos = 0.0;
            nSin = 1.0;
            break;
        case ROTATE_180:
            nCos = -1.0;
            nSin = 0.0;
            break;
        case ROTATE_270:
            nCos = 0.0;
            nSin = -1.0;
            break;
        case ROTATE_SLOW:
            nCos = m11;
            nSin = m10;
        }
        m00 = nCos;
        m10 = nSin;
        m01 = -nSin;
        m11 = nCos;
        m02 = -centerX*(nCos-1.0) + centerY*nSin;
        m12 = -centerX*nSin - centerY*(nCos-1.0);
    }
    private double FirstCycle(double radians) {
        radians = Math.IEEEremainder(radians, k360DegreeInRadians);
        if (radians < 0.0)
            radians += k360DegreeInRadians;
        return radians;
    }
    private void ComputeCosAndSin() {
        double angle = m01;	// presume the rotation angle is already set...
        double c = 0., s = 0.;

        // Special-case rectilinear rotations because Cos and Sin don't return
        // the exact values (0, -1, or 1) for these cases
        if (angle == 0.) c = 1;
        else if (angle == k90DegreeInRadians) s = 1;
        else if (angle == k180DegreeInRadians) c = -1;
        else if (angle == k270DegreeInRadians) s = -1;
        else { // general case
            c = Math.cos(angle);
            s = Math.sin(angle);
        }
        m11 = c;	// the cosine get stored at scaleY position
        m10 = s;	// the sine get stored at ELEMENT_SHEAR_Y position
    }

    /*
     * methods needed by Areas 
     */

    /* enum EPrMatrixType */
    protected static final int TYPE_IDENTITY = 0;
    protected static final int TYPE_ROTATE = 1;
    protected static final int TYPE_TRANSLATE = 2;
    protected static final int TYPE_SCALE = 3;
    protected static final int TYPE_AFFINE = 4;

    /* enum ERotateType */
    protected static final int ROTATE_0 = 0;
    protected static final int ROTATE_90 = 1;
    protected static final int ROTATE_180 = 2;
    protected static final int ROTATE_270 = 3;
    protected static final int ROTATE_SLOW = 4;

    /* enum EMatrixIndex */
    public static final int ELEMENT_SCALE_X = 0;     // m00
    public static final int ELEMENT_SHEAR_Y = 1;     // m10
    public static final int ELEMENT_SHEAR_X = 2;     // m01
    public static final int ELEMENT_SCALE_Y = 3;     // m11
    public static final int ELEMENT_TRANSLATE_X = 4; // m02
    public static final int ELEMENT_TRANSLATE_Y = 5; // m12

    public double getElement(int index) {

        if (fAccelerator==TYPE_ROTATE) {
            switch (index) {
            case ELEMENT_SCALE_X:
                return m11;
            case ELEMENT_SHEAR_Y:
                return m10;
            case ELEMENT_SHEAR_X:
                return -m10;
            case ELEMENT_SCALE_Y:
                return m11;
            case ELEMENT_TRANSLATE_X:
                return (m02==0.0 ? 0.0 : -m02*(m11-1.0) + m12*m10);
            case ELEMENT_TRANSLATE_Y:
                return (m12==0.0 ? 0.0 : -m02*m10 - m12*(m11-1.0));
            }
        }
        else
            switch (index) {
            case ELEMENT_SCALE_X:
                return m00;
            case ELEMENT_SHEAR_Y:
                return m10;
            case ELEMENT_SHEAR_X:
                return m01;
            case ELEMENT_SCALE_Y:
                return m11;
            case ELEMENT_TRANSLATE_X:
                return m02;
            case ELEMENT_TRANSLATE_Y:
                return m12;
            }
        throw
            new IndexOutOfBoundsException(
                "AffineTransform getElement index out of range");
    }

    protected static final int kAngle = 0;   // m01
    protected static final int kSin = 1;     // m10
    protected static final int kCos = 2;     // m11
    protected static final int kCenterX = 3; // m02
    protected static final int kCenterY = 4; // m12

    protected double getRotationElement(int index) {

        if (fAccelerator==TYPE_ROTATE) {
            switch (index) {
            case kAngle: 
                return m01;
            case kSin: 
                return m10;
            case kCos: 
                return m11;
            case kCenterX: 
                return m02;
            case kCenterY: 
                return m12;
            }
        }
        throw
            new IndexOutOfBoundsException(
                "AffineTransform getRotationElement index out of range");
    }

    // copy ctor ?
    protected AffineTransform(AffineTransform t) {
        this(t.m00, t.m10, t.m01, t.m11, t.m02, t.m12);
    }
    protected int getTransformType() {
        return fAccelerator;
    }
    protected int getRotateType() {
        return fRotateAccelerator;
    }
    protected void invert() throws NoninvertibleTransformException {
        switch (fAccelerator) {
        case TYPE_IDENTITY:
            break;
        case TYPE_TRANSLATE:
            m02 = -m02;
            m12 = -m12;
            break;
        case TYPE_SCALE:
            if ((m00 == 0.0) || (m11 == 0.0))
                throw
                    new NoninvertibleTransformException(
                        "Singular Scale Transform");

            m02 /= -m00;
            m12 /= -m11;
            m00 = 1.0/m00;
            m11 = 1.0/m11;
            break;
        case TYPE_ROTATE:
            m01 = k360DegreeInRadians - m01;
            m10 = -m10;
            if (fRotateAccelerator == ROTATE_90) // Change to ROTATE_270
                fRotateAccelerator = ROTATE_270;
            else if (fRotateAccelerator == ROTATE_270) // Change to ROTATE_90
                fRotateAccelerator = ROTATE_90;
            break;
        case TYPE_AFFINE:
            double ad = m00 * m11;
            double bc = m01 * m10;
            if (ad == bc)
                throw
                    new NoninvertibleTransformException(
                        "Singular Affine Transform");
            double determinant =
                ((Math.abs(ad) > Math.abs(bc))
                ? ad*(1.0 - bc/ad)
                : bc*(ad/bc - 1.0));
            // affine simple case: inverse = classical adjoint/ determinant
            double temp = m00/determinant;
            m00 = m11/determinant;
            m11 = temp;
            m01 /= -determinant;
            m10 /= -determinant;
            double tx = m02;
            double ty = m12;
            m02 = -tx*m00-ty*m01;
            m12 = -tx*m10-ty*m11;
            break;
        }
    }

    public boolean equals(AffineTransform matrix) {
        if (fAccelerator == matrix.fAccelerator) {
            switch (fAccelerator) {
            case TYPE_IDENTITY:
                return true;
            case TYPE_TRANSLATE:
                return (m02 == matrix.m02) && (m12 == matrix.m12);
            case TYPE_ROTATE:
                return
                    (m01 == matrix.m01) &&
                    (m02 == matrix.m02) &&
                    (m12 == matrix.m12);
            }
        }
        return
            (m00 == matrix.m00) && (m10 == matrix.m10) &&
            (m01 == matrix.m01) && (m11 == matrix.m11) &&
            (m02 == matrix.m02) && (m12 == matrix.m12);
    }

    public boolean isIdentity() {
        return ((fAccelerator == TYPE_IDENTITY) ||
                ((fAccelerator == TYPE_ROTATE) &&
                 (getRotateType() == ROTATE_0)));
    }
    public boolean isTranslate() {
        return (isIdentity() || fAccelerator == TYPE_TRANSLATE);
    }
    public boolean isTranslateOrScale() {
        return (fAccelerator == TYPE_TRANSLATE || fAccelerator == TYPE_SCALE);
    }
    public boolean isRectilinear() {
        if ((fAccelerator == TYPE_IDENTITY) ||
            (fAccelerator == TYPE_TRANSLATE) ||
            (fAccelerator == TYPE_SCALE) ||
            ((fAccelerator == TYPE_ROTATE) &&
             (getRotateType() != ROTATE_SLOW)))
            return true;
        return
            (fAccelerator == TYPE_AFFINE) &&
            ((m01 == 0.0 && m10 == 0.0) || (m00 == 0.0 && m11 == 0.0));
    }
}
