/*
 * @(#)AreaPathIterator.java	1.7 98/07/09
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

import java.awt.geom.PathIterator;

public class AreaPathIterator implements PathIterator {

    TGLoop fGeometry = null;
    int fEndCurve = 0;
    int fCurrentIndex = 0;
    int fNextBreak = 0;

    public AreaPathIterator(TGLoop geometry) {
        fCurrentIndex = 0;
        fNextBreak = 0;
        if (geometry == null || geometry.isEmpty()) {
            fEndCurve = -1; // -1 to prevent the extra SEG_CLOSE being issued
            fGeometry = null;
        }
        else {
            // make a copy of the passed in geometry
            fGeometry = new TGLoop(geometry);

            // higher order curves are tessolated down to cubics...
            if (fGeometry.getOrder() > 4) {
                //fGeometry.approximateLowerOrder(4, 0.5);
                throw new IllegalArgumentException(
                    "AreaPathIterator: Order of curve is higher than cubic!");
            }

            if (!fGeometry.isBezier()) {
                fGeometry.refineToBeziers();
            }
            fEndCurve = fGeometry.getNumberOfPoints();
//System.out.println("AreaPathIterator() endC:" + fEndCurve);
        }
    }

    public int getWindingRule() {
        if (fGeometry == null || fGeometry.getEOFill()) 
            return WIND_EVEN_ODD;
        else
            return WIND_NON_ZERO;
    }

    public boolean isDone() {
//System.out.print((fCurrentIndex >= fEndCurve) ? ".\n" : "> ");
        return (fCurrentIndex >= fEndCurve);
    }

    public void next() {
//System.out.print("N");
        if (fGeometry != null) {
            if (fCurrentIndex < fNextBreak) {
                int offset = fGeometry.getOrder() - 1;
                if (offset > 0)
                    fCurrentIndex += offset;
            }
            else {
                // update the next c0 index
                fNextBreak = fGeometry.getNextDiscontinuity(fNextBreak);
                fCurrentIndex += 1;
            }
        }
    }

    public int currentSegment(float[] coords) {
        double dcoord[] = new double[6];
        int op = currentSegment(dcoord);
        int theSize = 0; // number of points to copy over...
        switch (op) {
        case SEG_CUBICTO:
            ++theSize;
        case SEG_QUADTO:
            ++theSize;
        case SEG_MOVETO:
        case SEG_LINETO:
            ++theSize;
        case SEG_CLOSE:
        default:
            break;
        }
        theSize *= 2; // pair per point...
        for (int i = 0; i < theSize; ++i)
            coords[i] = (float)dcoord[i];
        return op;
    }

    private final void extractPoints(int numberOfPoints, double[] coords) {
        TGPoint t;
        int j = 0;
        int p = fCurrentIndex;
        for (int i = 0; i < numberOfPoints; ++i) {
            t = fGeometry.getPoint(p++).divW();
            coords[j++] = t.x;
            coords[j++] = t.y;
        }
    }
    public int currentSegment(double[] coords) {

        if (fGeometry == null || fCurrentIndex == fNextBreak) {
            // at c0 discontinuity, there is a break... move to
            extractPoints(1, coords);
//System.out.println("MOVETO[" + fCurrentIndex + "(B:" + fNextBreak + ")] (" + coords[0] + ", " + coords[1] + ")");
            return SEG_MOVETO;
        }
        else { // c1 discontinuity...
            switch (fGeometry.getOrder()) {
            default:
//System.out.print("? - ");
                // should never get here...
            case 2:
                // line
                extractPoints(1, coords);
//System.out.println("LINETO[" + fCurrentIndex + "(B:" + fNextBreak + ")] (" + coords[0] + ", " + coords[1] + ")");
                return SEG_LINETO;
            case 3:
                // quadratic
                extractPoints(2, coords);
//System.out.println("QUADTO[" + fCurrentIndex + "(B:" + fNextBreak + ")] (" + coords[0] + ", " + coords[1] + "), (" + coords[2] + ", " + coords[3] + ")");
                return SEG_QUADTO;
            case 4:
                // cubic
                extractPoints(3, coords);
//System.out.println("CUB_TO[" + fCurrentIndex + "(B:" + fNextBreak + ")] (" + coords[0] + ", " + coords[1] + "), (" + coords[2] + ", " + coords[3] + "), (" + coords[4] + ", " + coords[5] + ")");
                return SEG_CUBICTO;
            }
        }
    }

}
