/*
 * @(#)TBezierList.java	1.9 98/09/10
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

/**
 * TBezierList stores a continuous sequence of TBezier curve's.
 * 
 * The control points are stored as a sequence of double numbers in an
 * array. x0, y0, x1, y1, ...
 *
 * The first eight elements (index 0 to 7) of the array represent the four
 * points that make up the first Bezier curve. Indices 6 to 13 make up the
 * second Bezier curve, and so on. All adjacent Bezier curves are joined
 * end-to-end. So, for example, elements 6 and 7 represent the last points
 * of the first Bezier curve and also the first point of the second curve.
 * 
 * One TBezierList constructor takes TGLoop as a parameter. Since TGLoop
 * has a winding rule associated with it, we need to preserve that
 * information. The winding rule determines whether a point lies inside or
 * outside the loop. To preserve the winding rule we define a private
 * data member fEOFill, which has the same default value as that of TGLoop.
 * The default value is true which indicates the Even-Odd winding rule.
 *
 * @version     
 * @author      Taligent, Inc.
 * @see TBezier
 * @since       JDK1.2
 */

class TBezierList {

    // Data members
    private double[] fPointSequence = null;
    private boolean  fEOFill = true; // needed to preserve the winding rule
                                     // of any TGLoop's passed.
    public final static double parameterEpsilon = 1.0E-4;

    public TBezierList() {
        fPointSequence = null;
    }

    public TBezierList(TGLoop loop) {
        TGLoop localLoop = new TGLoop(loop);    // make a local copy
        localLoop.closeLoop();
        fEOFill = loop.getEOFill();
        int presentOrder = loop.getOrder();
        if (presentOrder != 2) {
            if (presentOrder > 4) {
                //localLoop.approximateLowerOrder(4, 0.2);
                throw new IllegalArgumentException("TBezierList: Cannot deal with curve of higher order than cubic!");
            }
            else if ((presentOrder < 4) && (presentOrder > 2)) {
                localLoop.raiseOrder(4);
            }
            // else we know the order is equal to 4 already or 0

            // We'll just copy over the points now
            int numPts = localLoop.getNumberOfPoints();
            fPointSequence = new double[numPts*2];        // need twice as many double numbers

            for (int i=0; i<numPts; i++) {
                TGRPoint currentRationalPoint = localLoop.getPoint(i);
                TGPoint currentPoint = currentRationalPoint.divW();
                fPointSequence[2*i] = currentPoint.x;
                fPointSequence[2*i+1] = currentPoint.y;
            }
        }
        else {
            int numPts = localLoop.getNumberOfPoints();
            fPointSequence = new double[numPts * 6 - 4];

            for (int i=0; i<numPts; i++) {  // copy and double up the points...
                TGRPoint currentRationalPoint = localLoop.getPoint(i);
                TGPoint currentPoint = currentRationalPoint.divW();
                int index = 6*i;
                if (i > 0) {
                    fPointSequence[index-2] = currentPoint.x;
                    fPointSequence[index-1] = currentPoint.y;
                }
                fPointSequence[index] = currentPoint.x;
                fPointSequence[index+1] = currentPoint.y;
                if (i < (numPts - 1)) {
                    fPointSequence[index+2] = currentPoint.x;
                    fPointSequence[index+3] = currentPoint.y;
                }
            }
        }
    }

/*
    public TBezierList(TBezier bezCurve) {
        fPointSequence = new double[8];
        int i = 0;

        fPointSequence[i] = bezCurve.fPoint0_fX; i++;
        fPointSequence[i] = bezCurve.fPoint0_fY; i++;
        fPointSequence[i] = bezCurve.fPoint1_fX; i++;
        fPointSequence[i] = bezCurve.fPoint1_fY; i++;
        fPointSequence[i] = bezCurve.fPoint2_fX; i++;
        fPointSequence[i] = bezCurve.fPoint2_fY; i++;
        fPointSequence[i] = bezCurve.fPoint3_fX; i++;
        fPointSequence[i] = bezCurve.fPoint3_fY;
    }
*/

    //  Copy constructor
    public TBezierList(TBezierList src) {
        fEOFill = src.fEOFill;
        if (src.fPointSequence != null) {
            int len = src.fPointSequence.length;
            fPointSequence = new double[len];
            System.arraycopy(src.fPointSequence, 0, fPointSequence, 0, len);
        }
        else
            fPointSequence = null;
    }

    public void appendBezierSegment(TBezier bezCurve) {

        int arrayLen = 0;

        if (fPointSequence == null) {
            fPointSequence = new double[8];        // three additional points needed

            fPointSequence[arrayLen] = bezCurve.fPoint0_fX; arrayLen++;
            fPointSequence[arrayLen] = bezCurve.fPoint0_fY; arrayLen++;
        }
        else {
            double[] oldValues = fPointSequence;
            arrayLen = fPointSequence.length;

            // check if the Bezier curves can be joined or not
            if ((fPointSequence[arrayLen-2] != bezCurve.fPoint0_fX) ||
                 (fPointSequence[arrayLen-1] != bezCurve.fPoint0_fY))
                throw new IndexOutOfBoundsException("Unable to join two disjoint Bezier curves");

            fPointSequence = new double[arrayLen + 6];        // three additional points needed
            System.arraycopy(oldValues, 0, fPointSequence, 0, arrayLen);
        }

        fPointSequence[arrayLen] = bezCurve.fPoint1_fX; arrayLen++;
        fPointSequence[arrayLen] = bezCurve.fPoint1_fY; arrayLen++;
        fPointSequence[arrayLen] = bezCurve.fPoint2_fX; arrayLen++;
        fPointSequence[arrayLen] = bezCurve.fPoint2_fY; arrayLen++;
        fPointSequence[arrayLen] = bezCurve.fPoint3_fX; arrayLen++;
        fPointSequence[arrayLen] = bezCurve.fPoint3_fY;
    }

    // Appends a bezier list at the end of the current list. If the first point of
    // the new bezier list does not match the last point of this bezier list, a straight
    // line segment is added between the two curves to bridge the gap.
    public void appendBezierList(TBezierList bezList) {

        if (!bezList.isEmpty()) {

            if (fPointSequence == null) {
                copyFrom(bezList);
                return;
            }

            double[] oldValues = fPointSequence;
            int presentArrayLen = fPointSequence.length;
            int additionalArrayLen = bezList.fPointSequence.length;

            // check if the Bezier lists can be joined or not
            if ((fPointSequence[presentArrayLen-2] != bezList.fPointSequence[0]) ||
                 (fPointSequence[presentArrayLen-1] != bezList.fPointSequence[1])) {

                throw new IllegalArgumentException("Unable to append two disconnected Bezier lists.");
                // need to add a simple straight line Bezier segment
                //TGPoint p0 = new TGPoint(fPointSequence[presentArrayLen-2], fPointSequence[presentArrayLen-1]);
                //TGPoint p3 = new TGPoint(bezList.fPointSequence[0], bezList.fPointSequence[1]);
                //TGPoint p1 = (TGPoint)p0.clone(); // new TGPoint(p0.x + 0.33*(p3.x - p0.x), p0.y + 0.33*(p3.y - p0.y));
                //TGPoint p2 = (TGPoint)p3.clone(); // new TGPoint(p0.x + 0.66*(p3.x - p0.x), p0.y + 0.66*(p3.y - p0.y));
                //TBezier bezCurv = new TBezier(p0, p1, p2, p3);
                //appendBezierSegment(bezCurv);
                //oldValues = fPointSequence;
                //presentArrayLen = fPointSequence.length;
            }

            fPointSequence = new double[presentArrayLen + additionalArrayLen - 2];
            System.arraycopy(oldValues, 0, fPointSequence, 0, presentArrayLen);
            System.arraycopy(bezList.fPointSequence, 2, fPointSequence, presentArrayLen, additionalArrayLen - 2);
        }
    }

    // Joins a bezier list to the current bezier list. The parameter 'direction'
    // decides if the direction of the incoming bezier list is to be reversed or not.
    // If none of the two ends of the bezier list match, an exception is thrown.
    public void joinBezierList(TBezierList bezList, boolean direction) {

        if (!bezList.isEmpty()) {
            int additionalArrayLen = bezList.fPointSequence.length;

            // First make a copy of the input since we may need it change it
            TBezierList the = bezList;

            if (fPointSequence == null) {
                if ((additionalArrayLen > 8) && (direction == false)) {
                    the = new TBezierList(bezList);
                    the.reverseDirection();
                }
                copyFrom(the);
            }
            else {
                double[] oldValues = fPointSequence;
                int presentArrayLen = fPointSequence.length;

                if (((additionalArrayLen > 8) && (direction == false)) ||
                    ((additionalArrayLen == 8) &&
                     (fPointSequence[presentArrayLen-2] == bezList.fPointSequence[additionalArrayLen-2]) &&
                     (fPointSequence[presentArrayLen-1] == bezList.fPointSequence[additionalArrayLen-1]))) {
                    the = new TBezierList(bezList);
                    the.reverseDirection();
                }
                // check if the Bezier lists can be joined or not
                if ((fPointSequence[presentArrayLen-2] == the.fPointSequence[0]) &&
                     (fPointSequence[presentArrayLen-1] == the.fPointSequence[1])) {
                    fPointSequence = new double[presentArrayLen + additionalArrayLen - 2];
                    System.arraycopy(oldValues, 0, fPointSequence, 0, presentArrayLen);
                    System.arraycopy(the.fPointSequence, 2, fPointSequence, presentArrayLen, additionalArrayLen - 2);
                }
                else
                    throw new IllegalArgumentException("Unable to find a common end-point between two Bezier lists.");
            }
        }
    }

    // Joins the two ends of the Bezier list to convert the Bezier List into a loop
    // Adds a straight line segment, if necessary, to join its ends
    public void joinEnds() {

            if (isEmpty()) return;
            int len = fPointSequence.length;
            if ((fPointSequence[0] == fPointSequence[len-2]) &&
                 (fPointSequence[1] == fPointSequence[len-1]))     // the loops are already joined
                return;

            double[] oldValues = fPointSequence;
            fPointSequence = new double[len + 6];
            System.arraycopy(oldValues, 0, fPointSequence, 0, len);

            // need to add a simple straight line Bezier segment
            fPointSequence[len+4] = oldValues[0];
            fPointSequence[len+5] = oldValues[1];
            fPointSequence[len  ] = fPointSequence[len-2]; // fPointSequence[len-2] + 0.33*(fPointSequence[len+4] - fPointSequence[len-2]);
            fPointSequence[len+1] = fPointSequence[len-1]; // fPointSequence[len-1] + 0.33*(fPointSequence[len+5] - fPointSequence[len-1]);
            fPointSequence[len+2] = fPointSequence[len+4]; // fPointSequence[len-2] + 0.66*(fPointSequence[len+4] - fPointSequence[len-2]);
            fPointSequence[len+3] = fPointSequence[len+5]; // fPointSequence[len-1] + 0.66*(fPointSequence[len+5] - fPointSequence[len-1]);
    }

    public void transformBy(TGrafMatrix matrix) {
        int numPoints = fPointSequence.length / 2;

        for (int i=0; i<numPoints; i++) {
            TGPoint currentPoint = new TGPoint(fPointSequence[2*i], fPointSequence[2*i+1]);
            TGPoint transformedPoint = matrix.transformPoint(currentPoint);
            fPointSequence[2*i] = transformedPoint.x;
            fPointSequence[2*i+1] = transformedPoint.y;
        }
    }

    // Assignment operator:
    public final TBezierList copyFrom(TBezierList src) {
        fEOFill = src.fEOFill;
        if (src.fPointSequence != null) {
            int len = src.fPointSequence.length;
            fPointSequence = new double[len];
            System.arraycopy(src.fPointSequence, 0, fPointSequence, 0, len);
        }
        else
            fPointSequence = null;
        return this;
    }

    private double[] mergeArray(double[] array1, double[] array2) {
        // appends the contents of array2 at the end of array1
        // changes array1 but does not change array2
        int len1 = array1.length;
        int len2 = array2.length;
        double[] newArray = new double[len1 + len2];

        System.arraycopy(array1, 0, newArray, 0, len1);
        System.arraycopy(array2, 0, newArray, len1, len2);

        return newArray;
    }

    private void printArray(String header, double[] array) {
        int len = array.length;
        int index;
        System.out.print(header + "-> Size : " + len + ". Elements : ");
        for (index=0; index<len; index++)
            System.out.print("[" + array[index] + "] ");
        System.out.println();
    }

    private double[] RemoveZerosAndOnes(double[] array) {
        if ((array == null) || (array.length == 0)) return null;

        int oldlen = array.length;
        int index;
        int newlen = 0;

        for (index=0; index<oldlen; index++) {
            if ((array[index] >= parameterEpsilon) && (array[index] <= (1.0-parameterEpsilon)))
                newlen++;
        }
        if (newlen == 0) return null;

        double[] returnValue = new double[newlen];
        int newCounter = 0;

        for (index=0; index<oldlen; index++) {
            if ((array[index] >= parameterEpsilon) && (array[index] <= (1.0-parameterEpsilon)))
                returnValue[newCounter++] = array[index];
        }
        return returnValue;
    }

    // Returns a parameter value that is good enough for breaking a Bezier segment
    // Since breaking up a Bezier segment at parameter values near zero and one does
    // not yeild the desired result, we break it more towards the center.
    private double nearestGoodParameter(double parameter) {
        if (parameter <= 0.1) return 0.1;
        if (parameter >= 0.9) return 0.9;
        return parameter;
    }

    // The following function splits up both curve lists A and B at points where they intersect
    public void SplitAtIntersectionsWith(TBezierList otherCurve) {

        int numCurvsThis = getNumberOfCurves();
        int numCurvsOther = otherCurve.getNumberOfCurves();
        double[][] splitParamsThis = new double[numCurvsThis][0];
        double[][] splitParamsOther = new double[numCurvsOther][0];

        int cIndex, dIndex;
        for (cIndex = 0; cIndex<numCurvsThis; cIndex++) {
            for (dIndex = 0; dIndex<numCurvsOther; dIndex++) {

                TBezier curveA = getCurve(cIndex);
                TBezier curveB = otherCurve.getCurve(dIndex);

                double[][] t = TBezier.findIntersections(curveA, curveB);
                if (t != null) {
                    if ((t[0] != null) && (t[0].length > 0))
                        splitParamsThis[cIndex] = mergeArray(splitParamsThis[cIndex], t[0]);

                    if ((t[1] != null) && (t[1].length > 0))
                        splitParamsOther[dIndex] = mergeArray(splitParamsOther[dIndex], t[1]);
                }
            }
        }

        int thisSegmentsAdded = 0;
        int otherSegmentsAdded = 0;

        for (cIndex = 0; cIndex<numCurvsThis; cIndex++) {
            TBezier.Sort(splitParamsThis[cIndex]);
            int breakCount = BreakIntoCurveSegments(cIndex+thisSegmentsAdded, splitParamsThis[cIndex]);
            thisSegmentsAdded += breakCount;
        }

        for (dIndex = 0; dIndex<numCurvsOther; dIndex++) {
            TBezier.Sort(splitParamsOther[dIndex]);
            int breakCount = otherCurve.BreakIntoCurveSegments(dIndex+otherSegmentsAdded, splitParamsOther[dIndex]);
            otherSegmentsAdded += breakCount;
        }
    }

    // The following function checks if the straight line segments between each bezier edge
    // intersects or not, and if so breaks them up into two patches.
    public void SplitAtOverlapsWith(TBezierList otherCurve) {
        int cIndex, dIndex;

        boolean finished = false;

        while (!finished) {
            boolean curveBroken = false;

            int numCurvsThis = getNumberOfCurves();
            int numCurvsOther = otherCurve.getNumberOfCurves();

            for (cIndex = 0; cIndex<numCurvsThis && !curveBroken; cIndex++) {
                for (dIndex = 0; dIndex<numCurvsOther && !curveBroken; dIndex++) {

                    TBezier curveA = getCurve(cIndex);
                    TBezier curveB = otherCurve.getCurve(dIndex);

                    double[] t = TBezier.findOverlaps(curveA, curveB);

                    if ((t != null) &&
                        (! curveA.isDegenerate()) && (! curveB.isDegenerate()) &&
                        (! curveA.isClose (curveB))) {

                        if ((t[0] >= parameterEpsilon) && (t[0] <= (1.0-parameterEpsilon)) &&
                            (t[1] >= parameterEpsilon) && (t[1] <= (1.0-parameterEpsilon))) {

                            SplitCurveAtParameter(cIndex, nearestGoodParameter(t[0]));
                            otherCurve.SplitCurveAtParameter(dIndex, nearestGoodParameter(t[1]));

                            // Once a curve is broken we cannot continue further because
                            // the ends have changed and intersections have to be calculated afresh.
                            curveBroken = true;
                        }
                    }
                }
            }
            if (!curveBroken)
                finished = true;
        }
    }

    public void SplitAtSelfIntersections() {
        int cIndex, dIndex;

        int numCurvsThis = getNumberOfCurves();
        double[][] splitParamsThis = new double[numCurvsThis][0];

        for (cIndex = 0; cIndex<numCurvsThis; cIndex++) {
            for (dIndex = cIndex+1; dIndex<numCurvsThis; dIndex++) {

                TBezier curveA = getCurve(cIndex);
                TBezier curveB = getCurve(dIndex);

                double[][] t = TBezier.findIntersections(curveA, curveB);
                if (t != null) {
                    if ((t[0] != null) && (t[0].length > 0))
                        splitParamsThis[cIndex] = mergeArray(splitParamsThis[cIndex], t[0]);

                    if ((t[1] != null) && (t[1].length > 0))
                        splitParamsThis[dIndex] = mergeArray(splitParamsThis[dIndex], t[1]);
                }
            }
        }

        int thisSegmentsAdded = 0;

        for (cIndex = 0; cIndex<numCurvsThis; cIndex++) {
            TBezier.Sort(splitParamsThis[cIndex]);
            int breakCount = BreakIntoCurveSegments(cIndex+thisSegmentsAdded, splitParamsThis[cIndex]);
            thisSegmentsAdded += breakCount;
        }
    }

    // The following function splits up the BezierList at points where the derivatives are zero
    public void SplitAtZeroDerivatives() {
        int cIndex;
        int cSegmentsAdded = 0;

        int numCurvsThis = getNumberOfCurves();

        for (cIndex = 0; cIndex<numCurvsThis; cIndex++) {
            TBezier curveA = getCurve(cIndex + cSegmentsAdded);

            double[] t = curveA.findZeroDerivatives();

            if (t != null) {
                int breakCount = BreakIntoCurveSegments(cIndex+cSegmentsAdded, t);
                cSegmentsAdded += breakCount;
            }
        }
    }

    // The following function splits up the BezierList at points where the derivatives are infinity
    public void SplitAtInfiniteDerivatives() {
        int cIndex;
        int cSegmentsAdded = 0;

        int numCurvsThis = getNumberOfCurves();

        for (cIndex = 0; cIndex<numCurvsThis; cIndex++) {
            TBezier curveA = getCurve(cIndex + cSegmentsAdded);

            double[] t = curveA.findInfiniteDerivatives();

            if (t != null) {
                int breakCount = BreakIntoCurveSegments(cIndex+cSegmentsAdded, t);
                cSegmentsAdded += breakCount;
            }
        }
    }

    public void SplitCurveAtParameter(int curveIndex, double splitParameter) {
        int totalCurves = getNumberOfCurves();
        if (curveIndex >= totalCurves)
            throw new IndexOutOfBoundsException("Requested curve doesn't exist in the list");

        double[] oldValues = fPointSequence;
        int arrayLen = fPointSequence.length;

        fPointSequence = new double[arrayLen + 6];        // three additional points needed
        int startIndex = curveIndex * 6;                // index where new curve starts
        int remainingStartIndex = startIndex + 6;        // index where remaining (unchanged) curve starts
        int newRemainingStartIndex = remainingStartIndex + 6; // index where the unchanged curve appears finally

        System.arraycopy(oldValues, 0, fPointSequence, 0, startIndex + 2);
        System.arraycopy(oldValues, remainingStartIndex, fPointSequence, newRemainingStartIndex, arrayLen - remainingStartIndex);

        // left.fPoint0 = (TGPoint)fPoint0.clone();

        // left.fPoint1 = TGPoint.add(fPoint0, TGPoint.multiply(t, TGPoint.subtract(fPoint1, fPoint0)));
        fPointSequence[startIndex+2] = oldValues[startIndex] + splitParameter*(oldValues[startIndex+2] - oldValues[startIndex]);
        fPointSequence[startIndex+3] = oldValues[startIndex+1] + splitParameter*(oldValues[startIndex+3] - oldValues[startIndex+1]);

        // left.fPoint2 = TGPoint.add(fPoint1, TGPoint.multiply(t, TGPoint.subtract(fPoint2, fPoint1))); // temporary holding spot
        double tempXValue, tempYValue;
        tempXValue = oldValues[startIndex+2] + splitParameter*(oldValues[startIndex+4] - oldValues[startIndex+2]);
        tempYValue = oldValues[startIndex+3] + splitParameter*(oldValues[startIndex+5] - oldValues[startIndex+3]);

        // fPoint2 = TGPoint.add(fPoint2, TGPoint.multiply(t, TGPoint.subtract(fPoint3, fPoint2)));
        fPointSequence[startIndex+10] = oldValues[startIndex+4] + splitParameter*(oldValues[startIndex+6] - oldValues[startIndex+4]);
        fPointSequence[startIndex+11] = oldValues[startIndex+5] + splitParameter*(oldValues[startIndex+7] - oldValues[startIndex+5]);

        // fPoint1 = TGPoint.add(left.fPoint2, TGPoint.multiply(t, TGPoint.subtract(fPoint2, left.fPoint2)));
        fPointSequence[startIndex+8] = tempXValue + splitParameter*(fPointSequence[startIndex+10] - tempXValue);
        fPointSequence[startIndex+9] = tempYValue + splitParameter*(fPointSequence[startIndex+11] - tempYValue);

        // left.fPoint2 = TGPoint.add(left.fPoint1, TGPoint.multiply(t, TGPoint.subtract(left.fPoint2, left.fPoint1)));
        fPointSequence[startIndex+4] = fPointSequence[startIndex+2] + splitParameter*(tempXValue - fPointSequence[startIndex+2]);
        fPointSequence[startIndex+5] = fPointSequence[startIndex+3] + splitParameter*(tempYValue - fPointSequence[startIndex+3]);

        // fPoint0 = TGPoint.add(left.fPoint2, TGPoint.multiply(t, TGPoint.subtract(fPoint1, left.fPoint2)));
        fPointSequence[startIndex+6] = fPointSequence[startIndex+4] + splitParameter*(fPointSequence[startIndex+8] - fPointSequence[startIndex+4]);
        fPointSequence[startIndex+7] = fPointSequence[startIndex+5] + splitParameter*(fPointSequence[startIndex+9] - fPointSequence[startIndex+5]);

        // left.fPoint3 = (TGPoint)fPoint0.clone();
    }

    // returns the number of breaks that were actually introduced into the curve segment
    int BreakIntoCurveSegments(int curveIndex, double[] parameterValues) {

        parameterValues = RemoveZerosAndOnes(parameterValues); // Eliminate values that are 0.0 and 1.0

        if (parameterValues == null) return 0;
        int arrayLen = parameterValues.length;

        int index = 0;

        SplitCurveAtParameter(curveIndex, parameterValues[0]);
        index++;
        while (index < arrayLen) {
            double splitT = (parameterValues[index] - parameterValues[index-1])/(1.0 - parameterValues[index-1]);
            SplitCurveAtParameter(curveIndex+index, splitT);
            index++;
        }
        return arrayLen;
    }

    public TBezier getCurve(int curveIndex) {
        int totalCurves = getNumberOfCurves();
        if (curveIndex >= totalCurves)
            throw new IndexOutOfBoundsException("Requested curve doesn't exist in the list");

        int currentPointLocation = curveIndex * 6;        // initial value in the array
        TBezier tempBez = new TBezier();

        tempBez.fPoint0_fX = fPointSequence[currentPointLocation];
        tempBez.fPoint0_fY = fPointSequence[currentPointLocation+1];

        tempBez.fPoint1_fX = fPointSequence[currentPointLocation+2];
        tempBez.fPoint1_fY = fPointSequence[currentPointLocation+3];

        tempBez.fPoint2_fX = fPointSequence[currentPointLocation+4];
        tempBez.fPoint2_fY = fPointSequence[currentPointLocation+5];

        tempBez.fPoint3_fX = fPointSequence[currentPointLocation+6];
        tempBez.fPoint3_fY = fPointSequence[currentPointLocation+7];

        return tempBez;
    }

    public void reverseDirection() {
        int arrayLen = fPointSequence.length;
        double oldValues[] = fPointSequence;
        fPointSequence = new double[arrayLen];

        int numPoints = arrayLen/2;
        for (int i=0; i<numPoints; i++) {
            fPointSequence[2*i] = oldValues[2*(numPoints-i-1)];
            fPointSequence[2*i+1] = oldValues[2*(numPoints-i-1)+1];
        }
    }

    public int getNumberOfDiscontinuities() {
        return getNumberOfCurves() + 1;
    }

    public TGPoint getDiscontinuity(int index) {
        int totalEndPoints = getNumberOfCurves() + 1;
        if (index >= totalEndPoints)
            throw new IndexOutOfBoundsException("Requested discontinuity doesn't exist in the list");

        int currentPointLocation = index * 6;        // initial value in the array

        return new TGPoint(fPointSequence[currentPointLocation], fPointSequence[currentPointLocation+1]);
    }

    public int getNumberOfCurves() {
        if (fPointSequence == null)
            return 0;
        int arrayLen = fPointSequence.length;
        if (arrayLen > 0)      // we know it must be at least 8
            return (arrayLen - 2)/6;
        else
            return 0;
    }

    boolean isEmpty() {
        return (fPointSequence.length < 8);        // need minimum 8 values for at least one Bezier curve
    }

    public void deleteAll() {
        fPointSequence = null;
    }

    public String toString() {
        String returnValue = "TBezierList : ";

        if (fPointSequence == null)
            returnValue += fPointSequence;
        else {
            int len = fPointSequence.length/2;
            for (int i=0; i<len; i++)
                returnValue += "(" + fPointSequence[2*i] + "," + fPointSequence[2*i+1] + ") ";
        }

        return returnValue;
    }

    // This returns the bounds of the curve's end-points after breaking at all the tangents of 0 and infinity.
    public TGRect getBounds() {
        if (isEmpty()) {
            return new TGRect(    // Why? Look at TGStatics.java
                              TGStatics.kZeroRectTopLeftX,
                              TGStatics.kZeroRectTopLeftY,
                              TGStatics.kZeroRectBottomRightX,
                              TGStatics.kZeroRectBottomRightY);
        }

        TBezierList localCopy = new TBezierList(this);
        localCopy.SplitAtInfiniteDerivatives();
        localCopy.SplitAtZeroDerivatives();

        int numPts = localCopy.fPointSequence.length/2;
        TGPoint pt = new TGPoint(localCopy.fPointSequence[0], localCopy.fPointSequence[1]);
        TGRect bounds = new TGRect(pt, pt);

        // We need not traverse through each and every point in the list
        // Traverse through only the end points of each Bezier segment
        for (int i = 3; i < numPts; i+=3)
            bounds.extendTo(new TGPoint(localCopy.fPointSequence[2*i], localCopy.fPointSequence[2*i+1]));

        return bounds;
    }

    // The following function tells whether the curve intersects the rectangle or not
    public boolean intersects(TGRect r) {

        int numCurvsThis = getNumberOfCurves();
        TBezierList rectCurve = new TBezierList(new TGLoop(r));    // create an TBezierList out of the rectangle

        int numCurvsOther = rectCurve.getNumberOfCurves();

        boolean curvesIntersect = false;

        int cIndex, dIndex;
        for (cIndex = 0; cIndex<numCurvsThis; cIndex++) {
            for (dIndex = 0; dIndex<numCurvsOther; dIndex++) {

                TBezier curveA = getCurve(cIndex);
                TBezier curveB = rectCurve.getCurve(dIndex);

                double[][] t = TBezier.findIntersections(curveA, curveB);
                t[0] = RemoveZerosAndOnes(t[0]); // anupam: fixing boundary value problem
                if ((t[0] != null) && (t[0].length > 0)) {
                    curvesIntersect = true;
                    break;
                }
                t[1] = RemoveZerosAndOnes(t[1]); // anupam: fixing boundary value problem
                if ((t[1] != null) && (t[1].length > 0)) {
                    curvesIntersect = true;
                    break;
                }
            }
            if (curvesIntersect)
                break;
        }

        if (curvesIntersect == false) {
            // Since the curves do not intersect, there can only be two possibilities
            // Case 1: This curve is entirely within the rectangle
            // Case 2: This curve is entirely outside the rectangle

            // Check for Case 1 now: Is the curve entirely within the rectangle
            // Check for just one point of the curve
            if (r.contains(new TGPoint(fPointSequence[0], fPointSequence[1])))
                curvesIntersect = true;
                // We are now sure that the whole curve is inside the rectangle. Return true.
        }

        return curvesIntersect;
    }

    // new functions Anupam made-up for the contains function TGLoop:
    // Tells whether a point is contained within the loop formed by the BezierList.

    public final boolean contains(TGPoint point) {
        // We have to discretize the BezierList to a polygon first and then run
        // the test on the polygon. Doing the discretization by walking along the
        // parameter values is the most inefficient implementation. Instead we do
        // a discretization by doing preliminary checks on the control points.

        // First make a local copy of the BezierList since we need to change it.
        TBezierList localCopy = new TBezierList(this);
        localCopy.joinEnds();

        // Split the curves at points where curvature changes.
        localCopy.SplitAtInflexionPoints();
        localCopy.SplitAtInfiniteDerivatives();
        localCopy.SplitAtZeroDerivatives();
        // Now all segments of the Bezier list are 'simple' Bezier segments
        // i.e. the bezier curve lies within the convex hull of its control points.

        TGPolygon discretizedPolygon = new TGPolygon();
        discretizedPolygon.setEOFill (fEOFill); // preserve the winding rule of the
                                                // original loop.

        // Walk through each Bezier segment of the BezierList
        int numSegments = getNumberOfCurves();

        for (int i=0; i<numSegments; i++) {
            TBezier currentSegment = getCurve(i);
            if (currentSegment.isDegenerate())
                continue;

            if (i==0) // the first bezier segment
                discretizedPolygon.append(currentSegment.getPoint(0));    // the first point

            // Check if the test point lies within the convex hull of its control points
            if (currentSegment.containsWithinConvexHull(point)) {
                // We need to find the parameter value closest to the test point
		TGPoint pointHolder = new TGPoint();
                double t = currentSegment.findNearestParametric(point, pointHolder);
                // Add the nearest point to the discretized polygon
                // discretizedPolygon.append(currentSegment.evaluate(t));
                discretizedPolygon.append(pointHolder);
            }

            // Append the last point of this segment
            discretizedPolygon.append(currentSegment.getPoint(3));
        }

        return discretizedPolygon.contains(point);
    }


    // The following function splits up the BezierList if any segment has an inflexion point
    public final void SplitAtInflexionPoints() {
        int cIndex;
        int cSegmentsAdded = 0;
        double[] t = null;

        int numCurvsThis = getNumberOfCurves();

        for (cIndex = 0; cIndex<numCurvsThis; cIndex++) {
            if (getCurve(cIndex + cSegmentsAdded).hasInflexion()) {
                if (t == null) {
                    t = new double[1];    // allocate only if necessary, and only once
                    t[0] = 0.5;
                }

                int breakCount = BreakIntoCurveSegments(cIndex+cSegmentsAdded, t);
                cSegmentsAdded += breakCount;
            }
        }
    }

} // TBezierList
