/*
 * @(#)TEdgeComparator.java	1.4 98/03/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

class TEdgeComparator {
    TEdgeComparator(long numberOfEdges) {}

    // These two functions are used to determine the left/right order of two edges.
    final void SetPoint(TGPoint point) {
        fPoint.copyFrom(point);
    }
    final boolean LessThan(TPolyEdge edge0, TPolyEdge edge1) {
        boolean result = false;
        if (edge0.fRelevant && edge1.fRelevant) {    // check for trivial case where edges are relevant to the y-event
            if (edge0.fSlope != edge1.fSlope)
                result = (edge0.fSlope < edge1.fSlope);
            else
                result = (edge0.fIndex < edge1.fIndex);    // arbitrary, but consistent
        }
        else {
            TGPoint point0 = new TGPoint();
            TGPoint point1 = new TGPoint();
            int comparison = FindIntersection2(edge0, edge1, point0, point1);
            boolean edge0Less = (comparison & kEdge0Less) != 0;
            comparison &= kEdgeComparison;
            if (comparison == kIntersecting) {
                if (fPoint.y == point0.y)
                    result = (fPoint.x < point0.x);
                else
                    result = (fPoint.y < point0.y) ^ edge0Less;
            }
            else if (comparison == kCoincident) {
                    result = (edge0.fIndex < edge1.fIndex);    // arbitrary, but consistent
            }
            else if (comparison == kDisjoint) {
                    result = edge0Less;
            }
        }
        return result;
    }

    // This function finds the intersection point(s)of two edges.  It returns the number of
    // intersections.  If the edges cross, the intersection point is returned in the first point
    // parameter and 1 is returned.  If the edges are coincident, the two inclusive endpoints are
    // returned as intersection points, and the function returns 2.  Otherwise, 0 is returned.
    final int FindIntersection(
        TPolyEdge edge0,
        TPolyEdge edge1,
        TGPoint point0,
        TGPoint point1)
    {
        int comparison = FindIntersection2(edge0, edge1, point0, point1) & kEdgeComparison;
        return (comparison == kIntersecting) ? 1 : ((comparison == kCoincident) ? 2 : 0);
    }


    private TGPoint fPoint = new TGPoint();

    /* enum EEdgeComparison */
    public static final int kIntersecting = 0;
    public static final int kCoincident = 1;
    public static final int kDisjoint = 2;

    // mask for the EEdgeComparison:
    public static final int kEdgeComparison = 0x3;

    // edge0-less-than-edge1 flag:
    public static final int kEdge0Less = 0x8;

    // This function is used to compute the information necessary to make comparisons of two edges.
    // Three possible values are returned by the function:
    //   kIntersecting  The edges intersect, and the intersection point is returned in point0.  If edge0 has
    //                  a smaller slope than edge1, true is returned in edge0Less.
    //   kCoincident    The edges are coincident.  The two intersection points are returned in point0 and point1.
    //   kDisjoint      The edges do not intersect.  If edge0 is to the left of edge1, true is returned in edge0Less.

    private final int FindIntersection2(
        TPolyEdge edge0,
        TPolyEdge edge1,
        TGPoint point0,
        TGPoint point1)
    {
        int result;

        TGPoint top0 = edge0.fTop;
        TGPoint bottom0 = edge0.fBottom;
        TGPoint top1 = edge1.fTop;
        TGPoint bottom1 = edge1.fBottom;

/* this should not be the case, because the vertex engine should have filtered out the redundent points
        if (top0.equals(bottom0) || top1.equals(bottom1)) {
            // reject degenerate edges:
            return kDisjoint;
        }
*/

        TGPoint temp = new TGPoint();

        // Determine whether or not the line segments intersect.  The distance-from-line algorithm
        // is used because it yields consistent results for all the lines segments as a whole.
        ComputeDistances(edge0, top1, bottom1, temp);
        double d00 = temp.x;
        double d01 = temp.y;
        if ((d00 >= 0) != (d01 >= 0)) {    // does edge1 cross edge0?
            ComputeDistances(edge1, top0, bottom0, temp);
            double d10 = temp.x;
            double d11 = temp.y;
            if ((d10 >= 0) != (d11 >= 0)) {  // does edge0 cross edge1?
                if (d00 == 0)
                    point0.copyFrom(top1);
                else if (d01 == 0)
                    point0.copyFrom(bottom1);
                else if (d10 == 0)
                    point0.copyFrom(top0);
                else if (d11 == 0)
                    point0.copyFrom(bottom0);
                else {
                    double x;
                    double y = ComputeY(edge0, edge1);

                    // The calculation may be unstable, so ensure that the y value is within range of both lines segments.
                    y = Math.max(Math.max(y, top0.y), top1.y);
                    y = Math.min(Math.min(y, bottom0.y), bottom1.y);

                    if (Math.abs(edge0.fSlope) < Math.abs(edge1.fSlope))        // use the smaller slope for stability
                        x = ComputeX(y, edge0);
                    else if (edge0.fSlope != edge1.fSlope)
                        x = ComputeX(y, edge1);
                    else // slopes are equal (it can happen with finite precision math), so use the index for consistency
                        x = ComputeX(y, (edge0.fIndex < edge1.fIndex) ? edge0 : edge1);

                    // The calculation may be unstable, so ensure that the x value is within range of both lines segments.
                    if (top0.x < bottom0.x)
                        x = Math.min(Math.max(x, top0.x), bottom0.x);
                    else
                        x = Math.max(Math.min(x, top0.x), bottom0.x);

                    if (top1.x < bottom1.x)
                        x = Math.min(Math.max(x, top1.x), bottom1.x);
                    else
                        x = Math.max(Math.min(x, top1.x), bottom1.x);

                    point0.x = x;
                    point0.y = y;
                }

                result = kIntersecting;
                if ((d00 < d10) || (d01 > d11))
                    result |= kEdge0Less;
            }
            else {
                result = kDisjoint;
                if (d10 < 0)
                    result |= kEdge0Less;    // edge0 is to the left of edge1
            }
        }
        else if ((d00 == 0) && (d01 == 0)) {    // are the edges looking like they are coincident?
            ComputeDistances(edge1, top0, bottom0, temp);
            double d10 = temp.x;
            double d11 = temp.y;
            double topE0 = top0.y;
            double topE1 = top1.y;
            double bottomE0 = bottom0.y;
            double bottomE1 = bottom1.y;
            double leftE0;
            double rightE0;
            if (top0.x > bottom0.x) {
                leftE0 = bottom0.x;
                rightE0 = top0.x;
            }
            else {
                leftE0 = top0.x;
                rightE0 = bottom0.x;
            }
            double leftE1;
            double rightE1;
            if (top1.x > bottom1.x) {
                leftE1 = bottom1.x;
                rightE1 = top1.x;
            }
            else {
                leftE1 = top1.x;
                rightE1 = bottom1.x;
            }
            if (((leftE0 < rightE1) && (rightE0 > leftE1) && (topE0 < bottomE1) && (bottomE0 > topE1)) ||
                ((leftE0 < rightE1) && (rightE0 > leftE1) && (topE0 == bottomE1) && (bottomE0 == topE1)) ||
                ((leftE0 == rightE1) && (rightE0 == leftE1) && (topE0 < bottomE1) && (bottomE0 > topE1)))
            {
                // Check for coincidence in the other direction
                if ((d10 == 0) == (d11 == 0)) {    // are the edges definitely coincident?
                    // Return both inclusive endpoints as intersection points
                    if (Math.abs(edge0.fSlope) < 1)    {    // use y values for comparison?
                        if (top0.y > top1.y)
                            point0.copyFrom(top0);
                        else
                            point0.copyFrom(top1);

                        if (bottom0.y < bottom1.y)
                            point1.copyFrom(bottom0);
                        else
                            point1.copyFrom(bottom1);
                    }
                    else {    // use x values for comparison
                        TGPoint left0 = top0;
                        TGPoint right0 = bottom0;
                        TGPoint left1 = top1;
                        TGPoint right1 = bottom1;
                        if (top0.x > bottom0.x) {
                            left0 = bottom0;
                            right0 = top0;
                        }
                        if (top1.x > bottom1.x) {
                            left1 = bottom1;
                            right1 = top1;
                        }

                        if (left0.x > left1.x)
                            point0.copyFrom(left0);
                        else
                            point0.copyFrom(left1);

                        if (right0.x < right1.x)
                            point1.copyFrom(right0);
                        else
                            point1.copyFrom(right1);
                    }

                    result = kCoincident;
                }
                else {
                    result = kDisjoint;
                    if (d10 < 0)
                        result |= kEdge0Less;    // edge0 is to the left of edge1
                }
            }
            else {
                result = kDisjoint;
                if (d10 < 0)
                    result |= kEdge0Less;    // edge0 is to the left of edge1
            }
        }
        else {
            result = kDisjoint;
            if (d00 >= 0)
                result |= kEdge0Less;    // edge1 is to the right of edge0
        }

        return result;
    }

    static void ComputeDistances(
        TPolyEdge edge,
        TGPoint p0,
        TGPoint p1,
        TGPoint result)     // use the TGPoint to return the two distance values
    {
        TGPoint top = edge.fTop;
        TGPoint bottom = edge.fBottom;
        double dx = bottom.x - top.x;
        double dy = bottom.y - top.y;
        double d0 = ((p0.x - top.x) * dy) - ((p0.y - top.y) * dx);
        double d1 = ((p1.x - top.x) * dy) - ((p1.y - top.y) * dx);

        double minimum = 1e-10;    // anything as close as this is treated as smack on the edge
        if (Math.abs(d0) < minimum)
            d0 = 0;
        if (Math.abs(d1) < minimum)
            d1 = 0;
        result.x = d0;
        result.y = d1;
    }
    static double ComputeY(TPolyEdge edge0, TPolyEdge edge1) {
        // computes y value of intersection point
        double y;
        if (edge0.fTop.y == edge0.fBottom.y) // infinite slope?
            y = edge0.fTop.y;
        else if (edge1.fTop.y == edge1.fBottom.y) // infinite slope?
            y = edge1.fTop.y;
        else {
            TPolyEdge e0 = edge0;
            TPolyEdge e1 = edge1;
            // for consistency (the formula below looks symmetric, but finite precision means it is not!)
            if (edge0.fIndex < edge1.fIndex) {
                e0 = edge1;
                e1 = edge0;
            }
            y = (e1.fSlope * e1.fTop.y) - (e0.fSlope * e0.fTop.y);
            y -= (e1.fTop.x - e0.fTop.x);
            y /= (e1.fSlope - e0.fSlope);
        }

        return y;
    }
    static double ComputeX(double y, TPolyEdge edge) {
        // computes x at y
        return (edge.fSlope * (y - edge.fTop.y)) + edge.fTop.x;
    }
}
