/*
 * @(#)TGCurve.java	1.6 98/07/06
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

/*
    TGCurve specifies curved geometry for Pink.  This ranges from simple
    Bezier curve to sophisticated shapes.  Curves are represented with
    Non-Uniform Rational B-Splines (NURBS).  This representation gives exact
    conic curves, direct control over the continuity of the curve, and no
    limitations on the number control points.  TGCurves are able to
    represent the shape of all other geometric types in Pink.

    A number of tools are provided for defining and manipulating curves.
    Constructors are available to make exact arcs, simple beziers (like in
    MacDraw or Illustrator) or curves from other Pink geometry.  Other
    methods refine, evaluate and modify the curve.

    In the current implementation TGCurve should not be subclassed, except
    for trivial subclasses to add constructor methods (it doesn't use
    virtual methods).
*/

public class TGCurve extends TBaseCurve {

    protected TConicInfo fConicInfo;

    // Empty constructor for MCollectible stuff. Will not render (0
    // points, order 0) The only operators allowed on empty curves are
    // assignment and stream in.

    public TGCurve() {
        super();
        fConicInfo = null;
    }

    // Constructor for a three point quadratic Bezier
    public TGCurve(TGPoint p0, TGPoint p1, TGPoint p2) {
        super(p0, p1, p2);
        fConicInfo = null;
    }

    // Constructor for a simple four point cubic Bezier
    public TGCurve(TGPoint p0, TGPoint p1, TGPoint p2, TGPoint p3) {
        super(p0, p1, p2, p3);
        fConicInfo = null;
    }

    // Constructor for general NURB curve.  If knots is not specified
    // a "pinned" knot vector is created.  If controlPoints (and
    // knots) are specified, they are copied in.  Assertions fail if
    // number of points is less than the order, or if the order is
    // less than two.  If the points are not given, the number
    // specified in numberPoints are created, and all are set to the
    // origin (w=1).

    public TGCurve(int order, int numberPoints) {
        super(order, numberPoints);
        fConicInfo = null;
    }

/*
    public TGCurve(int order, int numberPoints, TGParametricArray knots) {
        super(order, numberPoints, knots);
        fConicInfo = null;
    }

    public TGCurve(int order, TGRPointArray points) {
        super(order, points);
        fConicInfo = null;
    }
*/

    public TGCurve(int order, TGRPointArray points, TGParametricArray knots) {
        super(order, points, knots);
        fConicInfo = null;
    }

    // Copying constructor
    public TGCurve(TGCurve curve) {
        super((TBaseCurve)curve);
        if (curve.fConicInfo != null) {
            if (fConicInfo != null)
                fConicInfo.copyFrom(curve.fConicInfo);
            else
                fConicInfo = new TConicInfo(curve.fConicInfo);
        }
    }

    public Object clone() {
        return new TGCurve(this);
    }

    public TGCurve copyFrom(TGCurve src) {
        super.copyFrom((TBaseCurve)src);

        if (src.fConicInfo != null) {
            if (fConicInfo != null)
                fConicInfo.copyFrom(src.fConicInfo);
            else
                fConicInfo = new TConicInfo(src.fConicInfo);
        }

        return this;
    }

    public boolean equals(TGCurve src) {
        return super.equals(src) && fConicInfo.equals(src.fConicInfo);
    }

    // set curve to an elliptical arc from angle1 to angle2.  Now, the Arc
    // always goes "counterclockwise" (because that's how sin's and cos's go),
    // BUT, because the coordinate system in Albert is by default "upsidedown"
    // from the mathematical one, the arcs go clockwise.  Got it?

    TGCurve(TGEllipse e, double angle1, double angle2) {
        super(3, 3);

        // If the angles are the same, just make a line with both
        // endpoints at the point indicated by the parameters.
        if (angle1 == angle2) {
            TGPoint v = new TGPoint();
            v.setPolarCoordinates(angle1);
            v.multiplyBy(e.getSize());
            v.multiplyBy(0.5);
            v.addBy(e.getCenter());
            for (int i = 0; i < 3; i++)
                fPoints.setPoint(i, v);
            setKnotScheme(kBezier);

            fConicInfo = null;  // not a conic but a single point...
        }
        // If the angles describe exactly one trip around, make an
        // ellipse and figure out which direction it should go.
        else if (prStatic.Remainder(angle1, 360.0) == prStatic.Remainder(angle2, 360.0)) { // prNumeric
            makeEllipticCurve(e);
            if (angle2 < angle1)
                reverseDirection();
        }
        else {
            TGRPointArray pts =
                new TGRPointArray(12); // Storage for max points we may get back

            int numPoints = prStatic.ArcAngleAnglePts(e, angle1, angle2, 0, pts); // prConic
            makeArc(numPoints, pts);
            // conic info has to be inserted after calling makeArc, because makeArc calls setKnotScheme which will nuke the conicInfo...
            fConicInfo = new TConicInfo(TConicInfo.arc_t, e.getBounds(), false);
        }
    }

    //
    // These constructors make TGCurves out of the other geometric objects.
    //
/* replace with below ctor to get rid of TGLine:
    public TGCurve(TGLine line) {
        super(2, 2);
        setPoint(0, new TGRPoint(line.getStartPoint()));
        setPoint(1, new TGRPoint(line.getEndPoint()));

        fConicInfo = null;
    }
*/
    public TGCurve(TGPoint startPoint, TGPoint endPoint) {
        super(2, 2);
        setPoint(0, new TGRPoint(startPoint));
        setPoint(1, new TGRPoint(endPoint));

        fConicInfo = null;
    }

/*
    public TGCurve(TGPolyline polyline) {
        super(2, polyline.getNumberOfPoints());

        TGPointArray pts = new TGPointArray();
        polyline.getPoints(pts);
        int numPts = pts.getNumberOfPoints();
        TGPoint p[] = pts.getArray();
        for (int i = 0; i < numPts; i++)
            fPoints.setPoint(i, p[i]);

        fConicInfo = null;
    }
*/

/*
    // Special 2D evaluate with the gaussian curvature.
    public TGPoint evaluate(
        double u, TGPoint tangent, TGPoint deriv2, DoubleType curvature)
    {
        curvature.fValue =
            tangent.x*deriv2.y - deriv2.x*tangent.y / Math.pow(tangent.vectorLength(), 3.0);

        return super.evaluate(u, tangent, deriv2);
    }
*/
    // In order for overloading not to break, the others must be exported:
    public TGPoint evaluate(double u) {
        return super.evaluate(u);
    }
    public TGRPoint evaluateW(double u) {
        return super.evaluateW(u);
    }
/*
    public TGPoint evaluate(double u, TGPoint tangent) {
        return super.evaluate(u, tangent);
    }
    public TGPoint evaluate(double u, TGPoint tangent, TGPoint deriv2) {
        return super.evaluate(u, tangent, deriv2);
    }

    public void controlPolyline(TGPolyline pline) {
        int numPts = super.getNumberOfPoints();
        pline.resizePointArray(numPts);
        TGRPoint pts[] = fPoints.getArray();

        for (int i = 0; i < numPts; i++) {
            TGRPoint p = pts[i];
            pline.setPoint(i, p.fX, p.fY);
        }
    }
*/

/*================================================================
// Modelling tools
//
*/

/*
    // This returns the bounds of the curve's control points.  (Due to
    // the convex hull property, the curve is always contained within
    // these bounds).  Note this bounds is not cached, but recomputed
    // with every call.

    public TGRect getBounds() {
        if (isEmpty())
            return new TGRect (0., 0., 0., 0.); // Why? Look at TGStatics.java

        int numPts = getNumberOfPoints();
        TGRPoint pts[] = fPoints.getArray();
        TGRect bounds = new TGRect();
        bounds.setToPoint(pts[0].divW());
        for (int i = 1; i < numPts; i++)
            bounds.extendTo(pts[i].divW());

        return bounds;
    }

    // Returns true if the curve intersects the rectangle.  NOTE:  This is
    // currently implemented by discretizing the curve in something resembling
    // screen space.  In the future, the implementation will be changed to
    // a truly device independant one.

    public boolean intersects(TGRect r) {
        TGCurve curve = new TGCurve(this);
        double length = curve.arcLength();
        double min = curve.getMinParameter();
        double max = curve.getMaxParameter();
        double delta = (max - min)/length;
        int knotIndex = curve.getNextDiscontinuity(0, 0);
        double knots[] = curve.fKnots.getArray();
        double disc = knots[knotIndex];

        while (min < max) {
            if (knots[knotIndex] == max) disc = max;
            // Changing to foll for suppressing "informational" TGPolyline intersectPoly(length);
            TGPolyline intersectPoly = new TGPolyline((int)length);

            int j = 0;
            for (double t = min; t < disc; t+=delta) {
                intersectPoly.setPoint(j++, curve.evaluate(t));
            }
            if (disc != max) {
                intersectPoly.setPoint(j++, curve.evaluate(disc - .00000000001));
            }
            else {
                intersectPoly.setPoint(j++, curve.evaluate(disc));
            }
            if (intersectPoly.intersects(r))
                return true;
            min = knots[knotIndex+1];
            knotIndex = curve.getNextDiscontinuity(knotIndex, 0);
            disc = knots[knotIndex];
        }
        return false;
    }
*/

    // Transform all of the curve's control points by the specified matrix
    public void transformBy(TGrafMatrix mat) {
        mat.transformPoints(fPoints);
    }

    // Overloading to add carry conic info:
    public void concatenate(TBaseCurve p) {
        concatenate(p, true);
    }
    public void concatenate(TBaseCurve p, boolean glue) {
//System.out.println("TGCurve.concatenate(\n glue = " + glue + "\n" + (TGCurve)p);
        super.concatenate(p, glue);
        fConicInfo = null;
    }
/*
    public void dragPosition(
        double u,
        TGPoint ToPoint,
        double segmentMinimum,
        double segmentMaximum)
    {
        super.dragPosition(u, ToPoint, segmentMinimum, segmentMaximum);
        fConicInfo = null;
    }
    public void dragTangent(
        double u,
        TGPoint toTangent,
        double segmentMinimum,
        double segmentMaximum)
    {
        super.dragTangent(u, toTangent, segmentMinimum, segmentMaximum);
        fConicInfo = null;
    }
    public void setOrder(int Order) {
        super.setOrder(Order);
        fConicInfo = null;
    }
*/
    public void setPoint(int i, TGRPoint p) {
        super.setPoint(i, p);
        fConicInfo = null;
    }
/*
    public void setPoints(TGRPointArray srcPoints) {
        super.setPoints(srcPoints);
        fConicInfo = null;
    }
*/
    public void setKnotScheme(int theType) {
        super.setKnotScheme(theType);
        fConicInfo = null;
    }
/*
    public void setKnots(TGParametricArray knots) {
        super.setKnots(knots);
        fConicInfo = null;
    }
    public void moveKnot(int index, double u) {
        super.moveKnot(index, u);
        fConicInfo = null;
    }
    public void insertKnot(double u) {
        super.insertKnot(u);
        fConicInfo = null;
    }
    public void makeDiscontinuity(double u, int cont) {
        super.makeDiscontinuity(u, cont);
        fConicInfo = null;
    }
*/
    public void refine(TGParametricArray newKnots) {
        super.refine(newKnots);
        fConicInfo = null;
    }
    protected void incrementOrder() {
        super.incrementOrder();
        fConicInfo = null;
    }
    public void prAppendCurve(TBaseCurve crv, boolean glue) {
        super.prAppendCurve(crv, glue);
        fConicInfo = null;
    }

/*
    void setBasisSpline(double B[], int step) {
        super.setBasisSpline(B, step);
        fConicInfo = null;
    }

    // Export other overload from base class
    public static void makeCompatible(TGCurve curve1, TGCurve curve2) {
        TBaseCurve.makeCompatible(curve1, curve2);
    }
*/

    //
    // Take the raw control points generated by the prStatic routines // prConic
    // and make a Nurb out of them. The input array pts is trashed!!!
    //

    protected void makeArc(int numPoints, TGRPointArray pts) {
        // InternalAssert(((numPoints - 1) % 2 == 0), "makeArc didn't get valid Bezier curve");
        if ((numPoints - 1) % 2 != 0)
            throw new IllegalArgumentException("makeArc didn't get valid Bezier curve");

        // Stomp the contents of the curve
        fOrder = 3;
        fNumPoints = numPoints;
        pts.resize(numPoints);
        fPoints.swap(pts);
        fKnots.resize(numPoints + fOrder);

        setKnotScheme(kBezier);
    }

    // Morph the ellipse into a curve.
    // It use to be in TGLoop constructor. I moved it here to avoid copying...
    protected void makeEllipticCurve(TGEllipse ellipse) {

        double kR2o2 = Math.sqrt(2.0) / 2.0;
        int i;

        //  first make a unit circle as a rational quadratic.
        //               6
        //           5 ----- 7
        //             |/ \| 8
        //           4 |\ /| 0
        //           3 ----- 1
        //               2

        fPoints.setPoint(0, 1.0, 0.0, 1.0);
        fPoints.setPoint(1, kR2o2, kR2o2, kR2o2);
        fPoints.setPoint(2, 0.0, 1.0, 1.0);
        fPoints.setPoint(3, -kR2o2, kR2o2, kR2o2);
        fPoints.setPoint(4, -1.0, 0.0, 1.0);
        fPoints.setPoint(5, -kR2o2, -kR2o2, kR2o2);
        fPoints.setPoint(6, 0.0, -1.0, 1.0);
        fPoints.setPoint(7, kR2o2, -kR2o2, kR2o2);
        fPoints.setPoint(8, 1.0, 0.0, 1.0);

        TGPoint radius = ellipse.getSize().divideBy(2.0);
        TGPoint center = ellipse.getCenter();

        // scale and translate unit circle to match ellipse description
        TGRPoint pts[] = fPoints.getArray();
        for (i = 0; i < fNumPoints; i++) {
            pts[i].fX = pts[i].fX * radius.x + pts[i].fW * center.x;
            pts[i].fY = pts[i].fY * radius.y + pts[i].fW * center.y;
        }

        fOrder = 3;

        setKnotScheme(kBezier);

        fConicInfo =
            new TConicInfo(TConicInfo.ellipse_t, ellipse.getBounds(), false);
    }

    public String toString() {
        String points = new String();
        String knots = new String();
        int nump = fNumPoints;
        for (int i = 0; i < nump; i++)
             points = points + fPoints.getPoint(i) + "\n";
        nump += fOrder;
        for (int j = 0; j < nump; j++)
             knots = knots + fKnots.getValue(j) + "\n";
        return (
            "{\norder = " + fOrder + "\n" +
            "num points = " + fNumPoints + "\n" +
            "points: " + "\n" + points +
            "knots: " + "\n" + knots +
            "\n}\n");
    }
}
