/*
 * @(#)TGLoop.java	1.9 98/09/10
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

/*
    TGLoop is a subclass of TGCurve that draws a filled shape.  If no
    knot vector is given, a "floating closed" loop is created.  This
    uses a uniform knot vector that wraps around in a continous fasion.
    If a pinned knot vector is used, then the loop behaves like a
    TGCurve that is closed with a straight line.

    For continously "floating closed" loops, TGCurve methods that deal
    with the parameter range, evaluation, etc. are modified to take into
    account the continously closed nature of the loop.  If the loop is
    modified so that it has a pinned knot vector (or the ends of the
    knot vector range don't match) then the spell is broken, and the
    loop again behaves as a TGCurve, closed with a straight line at
    render time.

    For some applications, a closed loop where the start and end control points
    are coincedent may be preferable; calling "closeLoop()" on a loop (floating
    or pinned) will create this configuration without changing the shape.

    If c0 discontinuities are encountered in the curve defining the
    loop, they are closed separately when the loop is rendered (internal
    segments are closed with straight lines).

    The direction of each piece is used to determine whether a point is inside
    or outside the loop, along with the setting of the EOFill flag.;
 */

public class TGLoop extends TGCurve {

    // Empty constructor for MCollectible stuff. Will not render (0 points,
    // order 0) The only operators allowed on empty loops are
    // assignment and stream in.

    public TGLoop() {
        super();
        fEOFill = true;
    }

    public TGLoop(TGLoop other) {
        super((TGCurve)other);
        fEOFill = other.fEOFill;
    }

/*
    public TGLoop(TGCurve c) {
        this(c, true);
    }
*/
    public TGLoop(TGCurve c, boolean EOFill)
    {
        super(c);
        fEOFill = EOFill;
    }

/*
    public TGLoop(int order, int numberPoints) {
        this(order, numberPoints, true);
    }
    public TGLoop(int order, int numberPoints, boolean EOFill)
    {
        super(order, numberPoints);
        setKnotScheme(kFloating);
        fEOFill = EOFill;
    }

    public TGLoop(int order, int numberPoints, TGParametricArray knots) {
        this(order, numberPoints, knots, true);
    }
    public TGLoop(int order, int numberPoints, TGParametricArray knots, boolean EOFill)
    {
        super(order, numberPoints, knots);
        fEOFill = EOFill;
    }

    public TGLoop(int order, TGRPointArray points) {
        this(order, points, true);
    }
    public TGLoop(int order, TGRPointArray points, boolean EOFill)
    {
        super(order, points);
        setKnotScheme(kFloating);
        fEOFill = EOFill;
    }
*/

    public TGLoop(int order, TGRPointArray points, TGParametricArray knots) {
        this(order, points, knots, true);
    }
    public TGLoop(
        int order, TGRPointArray points, TGParametricArray knots, boolean EOFill)
    {
        super(order, points, knots);
        fEOFill = EOFill;
    }

    //
    // These constructors make TGLoops out of the other geometric objects.
    //

    // The "seam" is at the top left corner
    public TGLoop(TGRect rect) {
        super(2, 4);
        fEOFill = true;

        fPoints.setPoint(0, rect.getTopLeft());
        fPoints.setPoint(1, rect.getTopRight());
        fPoints.setPoint(2, rect.getBottomRight());
        fPoints.setPoint(3, rect.getBottomLeft());

        fOrder = 2; // Linear

        prStatic.SetBezierKnots(fOrder, fNumPoints, fKnots); // prKnots
    }

    // The "seam" is at rightmost edge of the ellipse
    public TGLoop(TGEllipse ellipse) {
        super(3, 9);
        fEOFill = true;

        makeEllipticCurve(ellipse);
    }

/*
    // Rounded Rectangle.  The "seam" is at the bottom of the top right corner.
    public TGLoop(TGRect roundRect, TGPoint aspect) {
        super(3, 17);
        fEOFill = true;

        double r2o2 = 1/Math.sqrt(2.0);

        // Disgorge data out of RoundRect
        TGPoint topLeft = roundRect.getTopLeft();
        TGPoint bottomRight = roundRect.getBottomRight();
        TGPoint center = roundRect.getCenter();

        double xRadLeft = topLeft.x + aspect.x;
        double xRadRight = bottomRight.x - aspect.x;
        double yRadTop = topLeft.y + aspect.y;
        double yRadBottom = bottomRight.y - aspect.y;

        // I'm not sure if the following behavior is "correct"
        if (xRadLeft>xRadRight) {
            xRadLeft = center.x;
            xRadRight = xRadLeft;
        }
        if (yRadTop>yRadBottom) {
            yRadTop = center.y;
            yRadBottom = yRadTop;
        }

        //  Make the round rect as a rational quadratic.
        //              12 13  14
        //          11 ---------- 15
        //          10 |/      \| 0, 16
        //             |        |
        //           9 |        | 1
        //             |        |
        //           8 |\      /| 2
        //           7 ---------- 3
        //              6  5  4

        fPoints.setPoint(0, bottomRight.x, yRadTop, 1.0);
        fPoints.setPoint(1, bottomRight.x, center.y, 1.0);
        fPoints.setPoint(2, bottomRight.x, yRadBottom, 1.0);
        fPoints.setPoint(3, r2o2*bottomRight.x, r2o2*bottomRight.y, r2o2);
        fPoints.setPoint(4, xRadRight, bottomRight.y, 1.0);
        fPoints.setPoint(5, center.x, bottomRight.y, 1.0);
        fPoints.setPoint(6, xRadLeft, bottomRight.y, 1.0);
        fPoints.setPoint(7, r2o2*topLeft.x, r2o2*bottomRight.y, r2o2);
        fPoints.setPoint(8, topLeft.x, yRadBottom, 1.0);
        fPoints.setPoint(9, topLeft.x, center.y, 1.0);
        fPoints.setPoint(10, topLeft.x, yRadTop, 1.0);
        fPoints.setPoint(11, r2o2*topLeft.x, r2o2*topLeft.y, r2o2);
        fPoints.setPoint(12, xRadLeft, topLeft.y, 1.0);
        fPoints.setPoint(13, center.x, topLeft.y, 1.0);
        fPoints.setPoint(14, xRadRight, topLeft.y, 1.0);
        fPoints.setPoint(15, r2o2*bottomRight.x, r2o2*topLeft.y, r2o2);
        fPoints.setPoint(16, bottomRight.x, yRadTop, 1.0);

        fOrder = 3;

        setKnotScheme(kBezier);
    }
*/

    public TGLoop(TGPolygon polygon) {
        super(2, polygon.getNumberOfPoints());
        fEOFill = polygon.getEOFill();

        fOrder = 2;

        TGRPoint pts[] = fPoints.getArray();
        for (int i = 0; i < fNumPoints; i++)
            pts[i].copyFrom(polygon.getPoint(i));

        setKnotScheme(kBezier);
    }

    // Data access
    public boolean getEOFill() {
        return fEOFill;
    }

    // Modifiers
    public void setEOFill(boolean EOFill) {
        fEOFill = EOFill;
    }

    // Forces a loop to be closed, so the endpoints of the control
    // polygon actually meet.  If the loop was defined with a
    // floating knot vector, then points are added to retain the
    // original (continous) shape.  If a loop is aready closed
    // (last point equals first, pinned knot vector) nothing
    // changes.  If a loop has a pinned (or other) knot vector,
    // a straight line segment is added to close it.
    public void closeLoop() {
        int i;

        if (isCloseable()) {
            extend();
            super.refineToPinned();
            TGRPoint pts[] = fPoints.getArray();
            pts[fNumPoints - 1].copyFrom(pts[0]);   // Avoid round-off errors
        }
        else {
            if (!isPinned())
                super.refineToPinned();

            TGRPoint pts[] = fPoints.getArray();
            int cur = 0, nextJump = 0;
            int degree = fOrder - 1;

            // For each discontinous segment, first check to see if
            // it's closed, and if not, add a closing segment
            do {
                nextJump = getNextDiscontinuity(cur, 0);
                // check to see if segment is explictly closed
                if (pts[cur].equals(pts[nextJump - 1]))
                    cur = nextJump;
                else {
                    fPoints.resize(fNumPoints + degree);
                    pts = fPoints.getArray();  // it may be reallocated...
                    for (i = fNumPoints - 1; i >= nextJump; i--)
                        pts[i + degree].copyFrom(pts[i]);

                    fKnots.resize(fNumPoints + 2 * fOrder - 1);
                    double knots[] = fKnots.getArray();
                    for (i = fNumPoints + degree; i >= nextJump; i--)
                        knots[i + degree] = knots[i];

                    for (i = 0; i < degree - 1; i++) {
                        double u = (double)(i + 1) / (double)(degree);
                        TGRPoint p = pts[nextJump + i];
                        p.copyFrom(pts[nextJump - 1]);
                        p.multiplyBy(1.0 - u);
                        p.addBy(TGRPoint.multiply(u, pts[cur]));
                    }
                    pts[nextJump + degree - 1].copyFrom(pts[cur]);

                    double midKnot = (knots[nextJump] + knots[nextJump - 1]) / 2.0;
                    for (i = 0; i < degree; i++)
                        knots[nextJump + i] = midKnot;

                    fNumPoints += degree;
                    cur = nextJump + degree;    // add degree to account for points added
                }
            } while (cur < fNumPoints);
        }
    }

    //
    // Hit methods Implementation
    //

    private static final double  kParametricDelta = 1.0e-6;

    public boolean intersects(TGRect g) {
        //
        // A loop may contain c0 discontinuities. So we have to
        // check for intesection of rectangle boundary with all the
        // segments of the loop.
        //

        if (isEmpty()) return false;
        
        int startPar = 0, endPar;
        TGLoop loopSection = new TGLoop();
        boolean intersectResult = false;

        int lastValue = getNumberOfKnots() - 1;
        while ((endPar = getNextDiscontinuity(startPar)) != lastValue) {
            getSectionOfCurve(getKnot(startPar),
                              getKnot(endPar) - kParametricDelta, 
                              loopSection);
            TBezierList loopSegment = new TBezierList(loopSection);
            intersectResult = loopSegment.intersects(g);
            if (intersectResult == true) break;
            startPar = endPar;
        }
        return intersectResult;
    }


    public boolean contains(TGPoint pt) {
        // TGPolygon tester = makePoly();
        // return tester != null && tester.contains(pt);
        // return (new TBezierList(this)).contains(pt);

        //
        // A loop may contain c0 discontinuities. So we have to
        // check whether the point lies within any segment of the loop.
        //
        // We have to be careful about retaining the winding rule for
        // each of the disjoint segments. The winding rule directly
        // impacts whether a point lies inside or outside the loop.
        //

        if (isEmpty()) return false;

        int startPar = 0, endPar;
        TGLoop loopSection = new TGLoop();
        loopSection.setEOFill (fEOFill);  // carry forward the winding rule.
        boolean containResult = false;
        
        int lastValue = getNumberOfKnots() - 1;
        while ((endPar = getNextDiscontinuity(startPar)) != lastValue) {
            // getSectionOfCurve method does not affect the fEOFill member of
            // loopSection. So we don't have to restore the winding rule value
            // for loopSection inside this while loop.
            getSectionOfCurve(getKnot(startPar),
                              getKnot(endPar) - kParametricDelta, 
                              loopSection);
            TBezierList loopSegment = new TBezierList(loopSection);
            containResult = loopSegment.contains(pt);
            if (containResult == true) break;
            startPar = endPar;
        }
        return containResult;
    }


    //    A rectangle is contained in the loop if any point of the
    //    rectangle (say topLeft), is inside the loop AND the boundaries
    //    of the loop DO NOT intersect with that of the rectangle.
    //
    //    This property should hold for every continuous segment of the loop.

    public boolean contains(TGRect rect) {
        if (isEmpty()) return false;

        int startPar = 0, endPar;
        TGPoint pt = rect.getTopLeft();
        TGLoop loopSection = new TGLoop();
        loopSection.setEOFill (fEOFill);  // carry forward the winding rule.
        boolean containResult = false;
        
        int lastValue = getNumberOfKnots() - 1;
        while ((endPar = getNextDiscontinuity(startPar)) != lastValue) {
            getSectionOfCurve(getKnot(startPar),
                              getKnot(endPar) - kParametricDelta, 
                              loopSection);
            TBezierList loopSegment = new TBezierList(loopSection);
            containResult = ( loopSegment.contains(pt) && (! loopSegment.intersects (rect)) );
            if (containResult == true) break;
            startPar = endPar;
        }
        return containResult;
    }




    public TGRect getBounds() {
        if (isEmpty()) {
            return new TGRect(    // Why? Look at TGStatics.java
                              TGStatics.kZeroRectTopLeftX,
                              TGStatics.kZeroRectTopLeftY,
                              TGStatics.kZeroRectBottomRightX,
                              TGStatics.kZeroRectBottomRightY);
        }

/*
        TGLoop loop = new TGLoop(this);
        loop.closeLoop();
        double length = loop.arcLength();
        double min = loop.getMinParameter();
        double max = loop.getMaxParameter();
        double delta = (max - min)/length;
        TGRect b = new TGRect();
        b.setToPoint(loop.evaluate(min));

        max = min + (delta*length);  // better test condition than "max"
        for (double t = min; t < max; t += delta) {
            b.extendTo(loop.evaluate(t));
        }
        return b;
*/
//System.out.println("TGLoop::getBounds() called:");
        return (new TBezierList(this)).getBounds();
    }

    // operators
    public TGLoop copyFrom(TGLoop loop) {
        if (loop != this) {
            super.copyFrom(loop);
            fEOFill = loop.fEOFill;
        }
        return this;
    }

    public boolean equals(TGLoop loop) {
        return ((fEOFill == loop.fEOFill) && (super.equals(loop)));
    }

    public Object clone() {
        return new TGLoop(this);
    }


    // The following are duplicated from the TGCurve interface to provide
    // expected behavior for floating closed loops.

    public TGPoint evaluate(double u) {
        if ((u > fKnots.getValue(fNumPoints)) && isCloseable()) {
            extend();
            TGPoint p = super.evaluate(u);
            retract();
            return p;
        }
        else
            return super.evaluate(u);
    }

    public TGRPoint evaluateW(double u) {
        if ((u > fKnots.getValue(fNumPoints)) && isCloseable()) {
            extend();
            TGRPoint p = super.evaluateW(u);
            retract();
            return p;
        }
        else
            return super.evaluateW(u);
    }

/*
    public TGPoint evaluate(double u, TGPoint tangent) {
        if ((u > fKnots.getValue(fNumPoints)) && isCloseable()) {
            extend();
            TGPoint p = super.evaluate(u, tangent);
            retract();
            return p;
        }
        else
            return super.evaluate(u, tangent);
    }

    public TGPoint evaluate(double u, TGPoint tangent, TGPoint deriv2) {
        if ((u > fKnots.getValue(fNumPoints)) && isCloseable()) {
            extend();
            TGPoint p = super.evaluate(u, tangent, deriv2);
            retract();
            return p;
        }
        else
            return super.evaluate(u, tangent, deriv2);
    }

    public TGPoint evaluate(double u, TGPoint tangent, TGPoint deriv2, double curvature) {
        if ((u > fKnots.getValue(fNumPoints)) && isCloseable()) {
            extend();
            TGPoint p = super.evaluate(u, tangent, deriv2, curvature);
            retract();
            return p;
        }
        else
            return super.evaluate(u, tangent, deriv2, curvature);
    }

    public double arcLength(double uFrom, double uTo) {
        if (isCloseable()) {
            extend();
            double length = super.arcLength(uFrom, uTo);
            retract();
            return length;
        }
        else
            return super.arcLength(uFrom, uTo);
    }

    public double arcLength() {
        if (isCloseable()) {
            extend();
            double length = super.arcLength();
            retract();
            return length;
        }
        else
            return super.arcLength();
    }

    public double approximateParameterFromArcLength(double length, double tolerance) {
        if (isCloseable()) {
            extend();
            double param = super.approximateParameterFromArcLength(length, tolerance);
            retract();
            return param;
        }
        else
            return super.approximateParameterFromArcLength(length, tolerance);
    }
*/

/*
    public double getMaxParameter() {
        if (isCloseable())
            return fKnots.getValue(fNumPoints + fOrder - 1);
        else
            return super.getMaxParameter();
    }

    public double nearestParametric(TGPoint test) {
        if (isCloseable()) {
            extend();
            double nearest = super.nearestParametric(test);
            retract();
            return nearest;
        }
        else
            return super.nearestParametric(test);
    }

    public void insertKnot(double u) {
        if ((u > fKnots.getValue(fNumPoints)) && isCloseable())
            closeLoop();
        super.insertKnot(u);
    }

    public void makeDiscontinuity(double u, int cont) {
        if ((u > fKnots.getValue(fNumPoints)) && isCloseable())
            closeLoop();
        super.makeDiscontinuity(u, cont);
    }
*/

    public void refineToBeziers() {
        closeLoop();
        super.refineToBeziers();
    }

    public void refineToPinned() {
        closeLoop();
    }

/*
    public static void makeCompatible(TGLoop curve1, TGLoop curve2) {
        curve1.closeLoop();
        curve2.closeLoop();
        TGCurve.makeCompatible(curve1, curve2);
    }

    // Caution - if the section crosses the seam, you'll get the rest of
    // the loop.
    public void getSectionOfCurve(double maxValue, double minValue, TGCurve outCurve) {
        if (((minValue > fKnots.getValue(fNumPoints)) || (maxValue > fKnots.getValue(fNumPoints))) && isCloseable()) {
            extend();
            super.getSectionOfCurve(minValue, maxValue, outCurve);
            retract();
        }
        else
            super.getSectionOfCurve(minValue, maxValue, outCurve);
    }

    public void dragPosition(double u,  TGPoint toPoint, double segMinimum, double segMaximum) {
        if ((u >= fKnots.getValue(fNumPoints)) && isCloseable()) {
            int i;

            extend();
            super.dragPosition(u, toPoint, segMinimum, segMaximum);
            int addedPoints = fNumPoints - fOrder + 1;
            for (i = addedPoints; i < fNumPoints; i++)
            if (fPoints.getValue(i).equals(fPoints.getValue(i - addedPoints)))
                fPoints.setValue(i - addedPoints, fPoints.getValue(i));
            retract();
        }
        else
            super.dragPosition(u, toPoint, segMinimum, segMaximum);
    }

    public void dragTangent(double u,  TGPoint toPoint, double segMinimum, double segMaximum) {
        if ((u >= fKnots.getValue(fNumPoints)) && isCloseable()) {
            extend();
            super.dragTangent(u, toPoint, segMinimum, segMaximum);
            int i;
            int addedPoints = fNumPoints - fOrder + 1;
            for (i = addedPoints; i < fNumPoints; i++)
            if (fPoints.getValue(i).equals(fPoints.getValue(i - addedPoints)))
                fPoints.setValue(i - addedPoints, fPoints.getValue(i));
            retract();
        }
        else
            super.dragTangent(u, toPoint, segMinimum, segMaximum);
    }
*/

    public void raiseOrder(int newOrder) {
        if (isCloseable())
            closeLoop();
        super.raiseOrder(newOrder);
    }

/*
    public void approximateLowerOrder(int newOrder, double tolerance) {
        if (isCloseable())
            closeLoop();
        super.approximateLowerOrder(newOrder, tolerance);
    }
*/

    boolean isCloseable() {
        // For a loop to be closed in a continous way, the order - 2 intervals
        // at the end of the curve must equal the intervals at the beginning.
        // We lop off a couple FP bits with Compare in case the knots were
        // derived from other computations

        double knots[] = fKnots.getArray();
        int lastKnot = fOrder + fNumPoints-1;
        double tolerance = (knots[lastKnot] - knots[0]) * prStatic.kDoubleEpsilon * 4; // prNumeric
        double delta = knots[1] - knots[0];
        int i;

        if (delta == 0.0)
            return false;   // Pinned curve; can't be Ck continous

        boolean closeable = true;
        for (i = 1; (i < fOrder - 2) && closeable; i++)
            closeable = prStatic.Compare(knots[i + 1] - knots[i], delta, tolerance); // prSplineRoutines

        for (i = fNumPoints + 1; (i < lastKnot) && closeable; i++)
            closeable = prStatic.Compare(knots[i + 1] - knots[i], delta, tolerance); // prSplineRoutines

        return closeable;
    }

    private void extend() {
        double knots[] = fKnots.getArray();
        TGRPoint pts[] = fPoints.getArray();
        double knotDelta = knots[1] - knots[0];
        int kSize = 2*fOrder + fNumPoints - 1;
        int pSize = fNumPoints + fOrder - 1;
        int i;

        fKnots.resize(kSize);
        fPoints.resize(pSize);

        for (i = fOrder + fNumPoints; i < kSize; i++)
            knots[i] = knots[i - 1] + knotDelta;

        for (i = fNumPoints; i < pSize; i++)
            pts[i].copyFrom(pts[i - fNumPoints]);

        fNumPoints += fOrder - 1;
    }

    private void retract() {
        // first check to see if the loop still closes
        if (isCloseable()) {
            // Now see if the spacing on the "real" end knots
            // matches the current start end...
            double knots[] = fKnots.getArray();
            double knotDelta = knots[1] - knots[0];
            double tolerance =
                (knots[fOrder + fNumPoints-1] - knots[0]) * prStatic.kDoubleEpsilon * 4; // prNumeric
            boolean retractable = true;
            int i;

            int endk = fNumPoints + fOrder - 2;
            for (i = fNumPoints - 1; (i < endk) && retractable; i++)
                retractable =
                    prStatic.Compare( // prSplineRoutines
                        knots[i + 1] - knots[i],
                        knotDelta,
                        tolerance);
            if (retractable) {
                fNumPoints -= fOrder - 1;
                fPoints.resize(fNumPoints);
                fKnots.resize(fNumPoints + fOrder);
            }
        }
    }

    private boolean fEOFill;

/*
    private TGPolygon makePoly() {
        if (isEmpty()) return null;

        // this is good old do it for every pixel method, should be replaced
        // with the new cubic bezier method Anapum implemented:
        TGLoop tmpLoop = new TGLoop(this);
        tmpLoop.closeLoop();
        double length = tmpLoop.arcLength();
        double min = tmpLoop.getMinParameter();
        double max = tmpLoop.getMaxParameter();
        double delta = (max - min)/length;
        TGPolygon containPoly = new TGPolygon((int)length, tmpLoop.fEOFill);
        int j = 0;
        max = min + (delta*length);  // better test condition than "max"
        for (double t = min; t < max; t += delta) {
            containPoly.setPoint(j++, tmpLoop.evaluate(t));
        }

        return containPoly;
    }
*/

    public String toString() {
        String points = new String();
        String knots = new String();
        int nump = fNumPoints;
        for (int i = 0; i < nump; i++)
             points = points + fPoints.getPoint(i) + "\n";
        nump += fOrder;
        for (int j = 0; j < nump; j++)
             knots = knots + fKnots.getValue(j) + "\n";
        return (
            "{\nEOFill = " + fEOFill + "\n" +
            "order = " + fOrder + "\n" +
            "num points = " + fNumPoints + "\n" +
            "points: " + "\n" + points +
            "knots: " + "\n" + knots +
            "\n}\n");
    }
}
