/*
 * @(#)TGRPoint.java	1.4 98/03/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

/*================
||
||  TGRPoint
||
||      for homogeneous 2D coordinates
||
*/

/*
    TRGPoints are basic two-dimensional points with rational (homogeneous)
    coordinates.  These are used primarily by TGCurve (to provide an exact
    representation of conic curves with splines) and for matrix operations
    requiring a homogeneous coordinate.
    
    A TGPoint may be converted directly to a TGRPoint via the constructor,
    a "w" value of 1.0 is assumed.  However, the conversion in the other
    direction must be done explicitly with either a "divW" or "dropW"
    operation.  Note arithmetic operations are deliberatly -not- provided
    for TGRPoints, because their correct interpretation would be too
    application specific. ;
*/

final class TGRPoint implements Cloneable {

    public double fX;
    public double fY;
    public double fW;

    // Does NOT initialize data members
    public TGRPoint() {};

    public TGRPoint(double x, double y, double w) {
        fX = x;
        fY = y;
        fW = w;
    }

    public TGRPoint(double x, double y) {
        this(x, y, 1.0);
    }

    public TGRPoint(TGRPoint src) {
        this(src.fX, src.fY, src.fW);
    }

    // This sets fW to 1.0
    public TGRPoint(TGPoint p) {
        this(p.x, p.y);
    }

    public Object clone() {
        return new TGRPoint(this);
    }

    public String toString() {
        return ("(" + fX + "," + fY + "," + fW + ")");
    }

    // Assignment operator:
    public final TGRPoint copyFrom(TGRPoint src) {
        fX = src.fX;
        fY = src.fY;
        fW = src.fW;
        return this;
    }
    public final TGRPoint copyFrom(TGPoint src) {
        fX = src.x;
        fY = src.y;
        fW = 1.0;
        return this;
    }

    public boolean equals(TGRPoint src) {
        return ((fX == src.fX) && (fY == src.fY) && (fW==src.fW));
    }

    // Divide the fX and fY values by fW.
    public TGPoint divW() {
        return new TGPoint(fX / fW, fY / fW);
    }

    // throw away the fW.
    public TGPoint dropW() {
        return new TGPoint(fX, fY);
    }

/*
    // Allows XYW indexed as 012
    public double element(int index) {
        switch (index) {
            case 0: return fX;
            case 1: return fY;
            case 2: return fW;
            default:
                throw new IndexOutOfBoundsException("TGRPoint.element(): index");
        }
    }
*/

    //
    // Arithmatic operators on rational points are not defined as part
    // of the Pink API because they don't make clear sense geometrically
    // and algebraicly.  There are numerous (and ambigous!) definitions
    // for what they might do.
    //
    // However, for the internal implementation of the curve methods,
    // their definition as straight vector operations is quite useful,
    // and indeed necessary to implement similar 2D and 3D routines
    // with the same code using templates.
    //
    // If the call overhead is a problem (and on a box with fast FP it might
    // well be) these functions can easily be made inline.
    //

/*
    TGRPoint multiplyBy(TGRPoint src)
    {
        fX *= src.fX;
        fY *= src.fY;
        fW *= src.fW;
        return this;
    }
*/

    TGRPoint multiplyBy(double src)
    {
        fX *= src;
        fY *= src;
        fW *= src;
        return this;
    }

/*
    TGRPoint divideBy(TGRPoint src)
    {
        fX /= src.fX;
        fY /= src.fY;
        fW /= src.fW;
        return this;
    }
*/

    TGRPoint divideBy(double src)
    {
        fX /= src;
        fY /= src;
        fW /= src;
        return this;
    }

    TGRPoint addBy(TGRPoint src)
    {
        fX += src.fX;
        fY += src.fY;
        fW += src.fW;
        return this;
    }

/*
    TGRPoint subtractBy(TGRPoint src)
    {
        fX -= src.fX;
        fY -= src.fY;
        fW -= src.fW;
        return this;
    }

    static TGRPoint add(TGRPoint left, TGRPoint right)
    {
        return new TGRPoint(left.fX + right.fX, left.fY + right.fY, left.fW + right.fW);
    }

    static TGRPoint subtract(TGRPoint left, TGRPoint right)
    {
        return new TGRPoint(left.fX - right.fX, left.fY - right.fY, left.fW - right.fW);
    }

    static TGRPoint multiply(TGRPoint left, TGRPoint right)
    {
        return new TGRPoint(left.fX * right.fX, left.fY * right.fY, left.fW * right.fW);
    }

    static TGRPoint divide(TGRPoint left, TGRPoint right)
    {
        return new TGRPoint(left.fX / right.fX, left.fY / right.fY, left.fW / right.fW);
    }
*/

    static TGRPoint multiply(TGRPoint left, double right)
    {
        return new TGRPoint(left.fX * right, left.fY * right, left.fW * right);
    }

    static TGRPoint multiply(double right, TGRPoint left)
    {
        return new TGRPoint(left.fX * right, left.fY * right, left.fW * right);
    }

    static TGRPoint divide(TGRPoint left, double right)
    {
        return new TGRPoint(left.fX / right, left.fY / right, left.fW / right);
    }

    static void SetToZero(TGRPoint p)
    {
        p.fX = 0.0;
        p.fY = 0.0;
        p.fW = 0.0;
    }
}
