/*
 * @(#)TGrafMatrix.java	1.9 98/09/10
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

import java.awt.geom.NoninvertibleTransformException;

/*
    Albert 2D Transforms

TGrafMatrix can be instantiated directly.
All of the public methods of TGrafMatrix can be called directly.

transformPoint maps one point to another (TGPoint) or one rational point
    to another (TGRPoint). transformPoints transforms arrays of points.
    Subclasses must define transformPoints for both points and rational points.
    By default, transformPoint (singular) calls transformPoints (array).
    Subclasses can override transformPoint to improve performance.

untransformPoint passes a point backwards through the transformation.  If the
    transform is not invertible,  it raises an kSingularTransform ITransformException.
    By default, untransformPoint Clones and Inverts the transform.
    Subclasses can override untransformPoint to improve performance.  The matrix
    is not changed.

transformBounds transforms the corners of a rectangle, and finds the bounding box
    of these corners.  Subclasses can override this to improve performance.

transformVector interprets TGPoint as a vector.  It will only transform the magnitude
    and direction of the vector. By default, transformVector calls transformPoint.
    Subclasses can override transformVector to improve performance.

transformPoint(s) and Untransform(s) are fully implemented for TGrafMatrix.

Concatenation: By default concatination is a post-concatination.  All pre concatinations
    start with "Pre".  in addition to general pre and post concatenation, we've also provided
    multiplication by standard linear matrices.  translateBy and preTranslateBy multiply
    the matrix by delta translation. scaleBy and preScaleBy multiply the matrix with a scale
    matrix. CenterOfScale is the point around which everything scales.  rotateBy and preRotateBy
    multiply the matrix by a rotate matrix. The angle is measured in degrees from the positive x
    axis to the positive y axis. CenterOfRotate is the point around which everything rotates.

SetTo operations reset the entire matrix to a known state.

    setToIdentity resets to a matrix that does nothing to a point:
        (x,y,w)*identity = (x,y,w)

    setToTranslate resets to a matrix that adds delta to a point:
        (x,y,w)*translate = (x+deltaX, y+deltaY, w)

    setToScale resets to a matrix that scales around a point.
        If the centerOfScale is the origin:
        (x,y,w)*scale = (x*scaleX, y*scaleY, w)
        Otherwise:
        (x,y,w)*scale =
          ((x-centerX)*scaleX+centerX, (y-centerY)*scaleY+centerY, w)

    setToRotate resets to a matrix that rotatates around a point.

Math Operations:

    determinant gets the determinant.  If the determinant is 0, a matrix is not
        invertible.  The inverse of a matrix is the "Classical Adjoint" of a matrix
        divided by its determinant. You can invert a matrix yourself (the hard way):
            matrix.determinant (determinant);
            matrix.makeAdjoint ();
            matrix.normalize ();

    makeAdjoint converts the matrix to its "Classical Adjoint."

    transpose swaps the rows and columns of a matrix.

    normalize divides the matrix by the homogeneous value.
        normalize returns false iff h is 0.0.

    operator*=(newMatrix) is equivalent to matrix.concatWith(newMatrix).

Any Operation that causes the transform to change must also cause the transform's
    time stamp t change.  timeStamp gets the current time stamp.  updateTimeStamp
    allows subclasses to bump the time stamp when the transform's state changes.

************************************************************************************/


//=============//
// TGrafMatrix //
//=============//

public class TGrafMatrix extends AffineTransform {

    private static final TGRect fkZeroRect =  // Why?  Look at TGStatics.java
        new TGRect(
                   TGStatics.kZeroRectTopLeftX,
                   TGStatics.kZeroRectTopLeftY,
                   TGStatics.kZeroRectBottomRightX,
                   TGStatics.kZeroRectBottomRightY);

    private static final TGRect fkInfiniteRect =  // Why?  Look at TGStatics.java
        new TGRect(
                   TGStatics.kInfRectTopLeftX,
                   TGStatics.kInfRectTopLeftY,
                   TGStatics.kInfRectBottomRightX,
                   TGStatics.kInfRectBottomRightY);

    // Get Identity returns a static identity for matrices
    //                  Why is it commented? Look at TGStatics.java 
    // private static final TGrafMatrix kgIdentity = new TGrafMatrix();

    public static TGrafMatrix getIdentity() {
        // return kgIdentity;
        return new TGrafMatrix();   // Why? Look at TGStatics.java
    }

    // ====== Constructors and destructors ======
    public TGrafMatrix() {
        super();
    }

    public TGrafMatrix(java.awt.geom.AffineTransform t) {
        super(t);
    }

    public TGrafMatrix(TGrafMatrix matrix) {
        super(matrix);
    }

/*
    // translate:
    public TGrafMatrix(TGPoint delta) {
        super.setToTranslation(delta.x, delta.y);
    }
    // scale:
    public TGrafMatrix(TGPoint scale, TGPoint centerOfScale) {
        super.setToScale(scale.x, scale.y, centerOfScale.x, centerOfScale.y);
    }
    // rotate:
    public TGrafMatrix(double angle, TGPoint centerOfRotate) {
        super.setToRotation(angle * Math.PI / 180., centerOfRotate.x, centerOfRotate.y);
    }
*/

    // ====== Transform and Untransform ======

    public TGRPoint transformPoint(TGRPoint source) {

        TGRPoint destination = new TGRPoint();

        switch (getTransformType()) {
        case TYPE_IDENTITY:
            // == Identity Transform Rational Points ===
            destination.fX = source.fX;
            destination.fY = source.fY;
            break;

        case TYPE_TRANSLATE:
            // == Translate Rational Points ===
            destination.fX = source.fX + getElement(ELEMENT_TRANSLATE_X)*source.fW;
            destination.fY = source.fY + getElement(ELEMENT_TRANSLATE_Y)*source.fW;
            break;

        case TYPE_ROTATE:
            // == Rotate Rational Points ===
            double cx, cy;
            switch (getRotateType()) {
            case ROTATE_0:
                destination.fX = source.fX;
                destination.fY = source.fY;
                break;

            case ROTATE_90:
                cx = getRotationElement(kCenterX);
                cy = getRotationElement(kCenterY);
                destination.fX = -source.fY + source.fW*(cx + cy);
                destination.fY = source.fX + source.fW*(-cx + cy);
                break;

            case ROTATE_180:
                destination.fX = -source.fX + 2.0*source.fW*getRotationElement(kCenterX);
                destination.fY = -source.fY + 2.0*source.fW*getRotationElement(kCenterY);
                break;

            case ROTATE_270:
                cx = getRotationElement(kCenterX);
                cy = getRotationElement(kCenterY);
                destination.fX = source.fY + source.fW*(cx - cy);
                destination.fY = -source.fX + source.fW*(cx + cy);
                break;

            case ROTATE_SLOW:
                // == Arbitrary Rotate Rational Points ===
                double nSin = getRotationElement(kSin);
                double nCos = getRotationElement(kCos);
                double oneMinus = 1.0 - nCos;
                cx = getRotationElement(kCenterX);
                cy = getRotationElement(kCenterY);
                destination.fX =
                    nCos*source.fX - nSin*source.fY + source.fW*(cx*oneMinus + cy*nSin);
                destination.fY =
                    nSin*source.fX + nCos*source.fY + source.fW*(cy*oneMinus - cx*nSin);
                break;
            }
            break;

        case 3 /* TYPE_SCALE */:
            // == scale Rational Points ===
            destination.fX =
                source.fX*getElement(ELEMENT_SCALE_X) + source.fW*getElement(ELEMENT_TRANSLATE_X);
            destination.fY =
                source.fY*getElement(ELEMENT_SCALE_Y) + source.fW*getElement(ELEMENT_TRANSLATE_Y);
            break;

        case 4 /* TYPE_AFFINE */: {
            // == Affine Transform Rational Points ===
            destination.fX = getElement(ELEMENT_SCALE_X)*source.fX;
            destination.fY = getElement(ELEMENT_SHEAR_Y)*source.fX;
            destination.fX += getElement(ELEMENT_SHEAR_X)*source.fY;
            destination.fY += getElement(ELEMENT_SCALE_Y)*source.fY;
            destination.fX += getElement(ELEMENT_TRANSLATE_X)*source.fW;
            destination.fY += getElement(ELEMENT_TRANSLATE_Y)*source.fW;
            }
            break;
        }
        destination.fW = source.fW;

        return destination;
    }
/*
    public TGRPoint untransformPoint(TGRPoint source) {
        TGRPoint destination = new TGRPoint();

        switch (getTransformType()) {
        case TYPE_IDENTITY:
            // == Identity UnTransform Rational Point ===
            destination.fX = source.fX;
            destination.fY = source.fY;
            break;
        case TYPE_TRANSLATE:
            // == UnTranslate Rational Point ===
            destination.fX = source.fX - source.fW*getElement(ELEMENT_TRANSLATE_X);
            destination.fY = source.fY - source.fW*getElement(ELEMENT_TRANSLATE_Y);
            break;
        case TYPE_SCALE:
            // == UnScale Rational Point ===
            {
            double sx = getElement(ELEMENT_SCALE_X);
            double sy = getElement(ELEMENT_SCALE_Y);

            if (sx != 0.0 || sy != 0.0) {
                destination.fX = (source.fX - source.fW*getElement(ELEMENT_TRANSLATE_X))/sx;
                destination.fY = (source.fY - source.fW*getElement(ELEMENT_TRANSLATE_Y))/sy;
            }
            else
                throw new ArithmeticException("TGrafMatrix.untransformPoint(TGRPoint): Singular Transform");
            }
            break;
        case TYPE_ROTATE:
            // == Unrotate Rational Point ===
            switch (getRotateType()) {
            case ROTATE_0:
                {
                double cx = getRotationElement(kCenterX);
                double cy = getRotationElement(kCenterY);
                destination.fX = source.fX;
                destination.fY = source.fY;
                }
                break;
            case ROTATE_90:
                {
                double cx = getRotationElement(kCenterX);
                double cy = getRotationElement(kCenterY);
                destination.fX =  source.fY + source.fW*(cx - cy);
                destination.fY = -source.fX + source.fW*(cx + cy);
                }
                break;
            case ROTATE_180:
                destination.fX =
                    -source.fX + 2.0*destination.fW*getRotationElement(kCenterX);
                destination.fY =
                    -source.fY + 2.0*destination.fW*getRotationElement(kCenterY);
                break;
            case ROTATE_270:
                {
                double cx = getRotationElement(kCenterX);
                double cy = getRotationElement(kCenterY);
                destination.fX = -source.fY + destination.fW*(cx + cy);
                destination.fY = source.fX + destination.fW*(-cx + cy);
                }
                break;
            case ROTATE_SLOW:
                // == Arbitrary Unrotate Rational Point ===
                {
                double cos = getRotationElement(kCos);
                double sin = getRotationElement(kSin);
                double oneMinus = 1.0 - cos;
                double cx = getRotationElement(kCenterX);
                double cy = getRotationElement(kCenterY);
                destination.fX =
                    cos*source.fX + sin*source.fY + source.fW*(cx*oneMinus - cy*sin);
                destination.fY =
                    -sin*source.fX + cos*source.fY + source.fW*(cy*oneMinus + cx*sin);
                }
                break;
            }
            break;
        case TYPE_AFFINE:
            // == Affine UnTransform Rational Point ===
            {
            double sx = getElement(ELEMENT_SCALE_X);
            double sy = getElement(ELEMENT_SCALE_Y);
            double ix = getElement(ELEMENT_SHEAR_X);
            double iy = getElement(ELEMENT_SHEAR_Y);
            double tx = getElement(ELEMENT_TRANSLATE_X);
            double ty = getElement(ELEMENT_TRANSLATE_Y);
            double det = sx*sy - ix*iy;
            if (det != 0.0) {
                double x = source.fX;
                destination.fX = (x*sy - source.fY*ix + source.fW*(-tx*sy + ty*ix))/det;
                destination.fY = (-x*iy + source.fY*sx + source.fW*(tx*iy - ty*sx))/det;
            }
            else
                throw new ArithmeticException("TGrafMatrix.untransformPoint(TGRPoint): Singular Transform");
            }
            break;
        }
        destination.fW = source.fW;
        return destination;
    }
*/

    public TGPoint transformPoint(TGPoint source) {
        TGPoint destination = new TGPoint();
        transform(source, destination);
        return destination;
    }

    public TGPoint untransformPoint(TGPoint source) {
        TGPoint destination = new TGPoint();
        try {
            inverseTransform(source, destination);
        }
        catch (NoninvertibleTransformException e) {
            throw new ArithmeticException("TGrafMatrix.untransformPoint(TGPoint): Singular Transform");
        }
        return destination;
    }

/*
    // This may go away...
    public void transformPoints(TGPoint points[], TGPoint transformedPoints[]) {
        transform(points[0], transformedPoints[0]);
    }

    // This may go away...
    public void transformPoints(TGPoint source[], TGPoint destination[], int count) {
        transform(source, 0, destination, 0, count);
    }
*/

    public void transformPoints(TGPointArray points) {
        int count = points.getNumberOfPoints();
        TGPoint pts[] = points.getArray();
        
        transform(pts, 0, pts, 0, count);
    }

/*
    // This may go away...it's used by the rendering code
    public void transformPoints(TGRPoint points[], TGRPoint transformPoints[]) {
        transformPoints(points, transformPoints, 1);
    }
*/

    // This may go away...it's used by the rendering code
    public void transformPoints(TGRPoint source[], TGRPoint destination[], int count) {
        TGRPoint dest, src;

        int acc = getTransformType();
        boolean overlap = (source == destination);
        if (acc == TYPE_IDENTITY ||
            (acc == TYPE_ROTATE && acc == ROTATE_0)) {

            if (!overlap) {
                while (--count >= 0) {
                    dest = destination[count];
                    src = source[count];
                    dest.fX = src.fX;
                    dest.fY = src.fY;
                    dest.fW = src.fW;
                }
            }
        }

        switch (acc) {
        case TYPE_TRANSLATE:
            // == Translate Rational Points ===
            while (--count >= 0) {
                dest = destination[count];
                src = source[count];
                dest.fX = src.fX + getElement(ELEMENT_TRANSLATE_X)*src.fW;
                dest.fY = src.fY + getElement(ELEMENT_TRANSLATE_Y)*src.fW;
                dest.fW = src.fW;
            }
            break;

        case TYPE_ROTATE:
            // == Rotate Rational Points ===
            switch (getRotateType()) {
            case ROTATE_90:
                {
                double cx = getRotationElement(kCenterX);
                double cy = getRotationElement(kCenterY);
                double tx = cx + cy;
                double ty = -cx + cy;
                if (overlap) {
                    while (--count >= 0) {
                        dest = destination[count];
                        src = source[count];
                        double oldX = src.fX;
                        dest.fX = -src.fY + src.fW*tx;
                        dest.fY = oldX + src.fW*ty;
                    }
                } else {
                    while (--count >= 0) {
                        dest = destination[count];
                        src = source[count];
                        dest.fX = -src.fY + src.fW*tx;
                        dest.fY = src.fX + src.fW*ty;
                        dest.fW = src.fW;
                    }
                }
                }
                break;

            case ROTATE_180:
                {
                double cx = getRotationElement(kCenterX);
                double cy = getRotationElement(kCenterY);
                while (--count >= 0) {
                    dest = destination[count];
                    src = source[count];
                    dest.fX = -src.fX + 2.0*src.fW*cx;
                    dest.fY = -src.fY + 2.0*src.fW*cy;
                    dest.fW = src.fW;
                }
                }
                break;

            case ROTATE_270:
                {
                double cx = getRotationElement(kCenterX);
                double cy = getRotationElement(kCenterY);
                double tx = cx - cy;
                double ty = cx + cy;
                if (overlap) {
                    while (--count >= 0) {
                        dest = destination[count];
                        src = source[count];
                        double oldX = src.fX;
                        dest.fX = src.fY + src.fW*tx;
                        dest.fY = -oldX + src.fW*ty;
                    }
                } else {
                    while (--count >= 0) {
                        dest = destination[count];
                        src = source[count];
                        dest.fX = src.fY + src.fW*tx;
                        dest.fY = -src.fX + src.fW*ty;
                        dest.fW = src.fW;
                    }
                }
                }
                break;

            case ROTATE_SLOW:
                // == Arbitrary Rotate Rational Points ===
                {
                double cos = getRotationElement(kCos);
                double sin = getRotationElement(kSin);
                double oneMinus = 1.0 - cos;
                double cx = getRotationElement(kCenterX);
                double cy = getRotationElement(kCenterY);
                double tx = cx*oneMinus + cy*sin;
                double ty = cy*oneMinus - cx*sin;
                if (overlap) {
                    while (--count >= 0) {
                        dest = destination[count];
                        src = source[count];
                        double oldX = src.fX;
                        dest.fX = cos*oldX - sin*src.fY + src.fW*tx;
                        dest.fY = sin*oldX + cos*src.fY + src.fW*ty;
                    }
                } else {
                    while (--count >= 0) {
                        dest = destination[count];
                        src = source[count];
                        dest.fX = cos*src.fX - sin*src.fY + src.fW*tx;
                        dest.fY = sin*src.fX + cos*src.fY + src.fW*ty;
                        dest.fW = src.fW;
                    }
                }
                }
                break;
            }
            break;

        case TYPE_SCALE:
            // == scale Rational Points ===
            {
            double sx = getElement(ELEMENT_SCALE_X);
            double sy = getElement(ELEMENT_SCALE_Y);
            double tx = getElement(ELEMENT_TRANSLATE_X);
            double ty = getElement(ELEMENT_TRANSLATE_Y);
            while (--count >= 0) {
                dest = destination[count];
                src = source[count];
                dest.fX = src.fX*sx + src.fW*tx;
                dest.fY = src.fY*sy + src.fW*ty;
                dest.fW = src.fW;
            }
            }
            break;

        case TYPE_AFFINE:
            // == Affine Transform Rational Points ===
            {
            double sx = getElement(ELEMENT_SCALE_X);
            double sy = getElement(ELEMENT_SCALE_Y);
            double ix = getElement(ELEMENT_SHEAR_X);
            double iy = getElement(ELEMENT_SHEAR_Y);
            double tx = getElement(ELEMENT_TRANSLATE_X);
            double ty = getElement(ELEMENT_TRANSLATE_Y);
            if (overlap) {
                while (--count >= 0) {
                    dest = destination[count];
                    src = source[count];
                    double oldX = src.fX;
                    double oldY = src.fY;
                    dest.fX = sx*oldX;
                    dest.fY = iy*oldX;
                    dest.fX += ix*oldY;
                    dest.fY += sy*oldY;
                    dest.fX += tx*src.fW;
                    dest.fY += ty*src.fW;
                }
            } else {
                while (--count >= 0) {
                    dest = destination[count];
                    src = source[count];
                    dest.fX = sx*src.fX;
                    dest.fY = iy*src.fX;
                    dest.fX += ix*src.fY;
                    dest.fY += sy*src.fY;
                    dest.fX += tx*src.fW;
                    dest.fY += ty*src.fW;
                    dest.fW = src.fW;
                }
            }
            }
            break;
        }
    }

    public void transformPoints(TGRPointArray points) {
        TGRPoint src, dest;
        double translateX, translateY;

        int count = points.getNumberOfPoints();
        TGRPoint pts[] = points.getArray();

        transformPoints(pts, pts, count);
    }

    public TGRect transformBounds(TGRect r) {
        int acc = getTransformType();
        if (acc == TYPE_IDENTITY)
            return new TGRect(r);
        else if (r.equals(fkZeroRect) || r.equals(fkInfiniteRect))
            return new TGRect(r);
        else {
            TGRect newBounds = null;
            switch (acc) {
            case TYPE_TRANSLATE:
                {
                double tx = getElement(ELEMENT_TRANSLATE_X);
                double ty = getElement(ELEMENT_TRANSLATE_Y);
                newBounds =
                    new TGRect(r.x + tx, r.y + ty, r.getMaxX() + tx, r.getMaxY() + ty);
                }
                break;
            case TYPE_SCALE:
                {
                double sx = getElement(ELEMENT_SCALE_X);
                double sy = getElement(ELEMENT_SCALE_Y);
                double tx = getElement(ELEMENT_TRANSLATE_X);
                double ty = getElement(ELEMENT_TRANSLATE_Y);
                newBounds =
                    new TGRect(
                        r.x * sx + tx,
                        r.y * sy + ty,
                        r.getMaxX() * sx + tx,
                        r.getMaxY() * sy + ty);
                }
                break;
            case TYPE_ROTATE:
            case TYPE_AFFINE:
                newBounds = new TGRect();
                newBounds.setToPoint(transformPoint(r.getTopLeft()));
                newBounds.extendTo(transformPoint(r.getTopRight()));
                newBounds.extendTo(transformPoint(r.getBottomLeft()));
                newBounds.extendTo(transformPoint(r.getBottomRight()));
                break;
            }

            return newBounds;
        }
    }

    public TGRect untransformBounds(TGRect r) {

        int acc = getTransformType();
        if (acc == TYPE_IDENTITY)
            return new TGRect(r);
        if (r == fkZeroRect || r == fkInfiniteRect)
            return new TGRect(r);
        else {
            TGRect newBounds = null;
            switch (acc) {
            case TYPE_TRANSLATE:
                {
                double tx = getElement(ELEMENT_TRANSLATE_X);
                double ty = getElement(ELEMENT_TRANSLATE_Y);
                newBounds =
                    new TGRect(r.x - tx, r.y - ty, r.getMaxX() - tx, r.getMaxY() - ty);
                }
                break;
            case TYPE_SCALE:
                {
                double sx = getElement(ELEMENT_SCALE_X);
                double sy = getElement(ELEMENT_SCALE_Y);
                double tx = getElement(ELEMENT_TRANSLATE_X);
                double ty = getElement(ELEMENT_TRANSLATE_Y);
                newBounds =
                    new TGRect(
                        (r.x - tx)/sx,
                        (r.y - ty)/sy,
                        (r.getMaxX() - tx)/sx,
                        (r.getMaxY() - ty)/sy);
                }
                break;
            case TYPE_ROTATE:
            case TYPE_AFFINE:
                newBounds = new TGRect();
                newBounds.setToPoint(untransformPoint(r.getTopLeft()));
                newBounds.extendTo(untransformPoint(r.getTopRight()));
                newBounds.extendTo(untransformPoint(r.getBottomLeft()));
                newBounds.extendTo(untransformPoint(r.getBottomRight()));
                break;
            }
            return newBounds;
        }
    }

    // ====== Concatenation ======
    // Post Concatenation:
/*
    public void concatWith(TGrafMatrix matrix) {
        super.preConcatenate(matrix);
    }
    public void translateBy(TGPoint delta) {
        super.preTranslate(delta.x, delta.y);
    }
    public void scaleBy(TGPoint scale) {
        scaleBy(scale, TGPoint.kOrigin);
    }
    public void scaleBy(TGPoint scale, TGPoint centerOfScale) {
        super.preScale(scale.x, scale.y, centerOfScale.x, centerOfScale.y);
    }
    public void rotateBy(double degrees) {
        rotateBy(degrees, TGPoint.kOrigin);
    }
    public void rotateBy(double degrees, TGPoint centerOfRotate) {
        super.preRotate(degrees * Math.PI / 180., centerOfRotate.x, centerOfRotate.y);
    }
*/
    // Pre Concatenation:
    public void preConcatWith(TGrafMatrix matrix) {
        super.concatenate(matrix);
    }
/*
    public void preTranslateBy(TGPoint delta) {
        super.translate(delta.x, delta.y);
    }
    public void preScaleBy(TGPoint scale) {
        preScaleBy(scale, TGPoint.kOrigin);
    }
    public void preScaleBy(TGPoint scale, TGPoint centerOfScale) {
        super.scale(scale.x, scale.y, centerOfScale.x, centerOfScale.y);
    }
    public void preRotateBy(double degrees) {
        preRotateBy(degrees, TGPoint.kOrigin);
    }
    public void preRotateBy(double degrees, TGPoint centerOfRotate) {
        super.rotate(degrees * Math.PI / 180., centerOfRotate.x, centerOfRotate.y);
    }
*/

    // === SetTo ===
/*
    public void setToTranslate(TGPoint delta) {
        super.setToTranslation(delta.x, delta.y);
    }
    public void setToScale(TGPoint scale) {
        setToScale(scale, TGPoint.kOrigin);
    }
    public void setToScale(TGPoint scale, TGPoint centerOfScale) {
        super.setToScale(scale.x, scale.y, centerOfScale.x, centerOfScale.y);
    }
    public void setToRotate(double degrees, TGPoint centerOfRotate) {
        super.setToRotation(degrees * Math.PI / 180., centerOfRotate.x, centerOfRotate.y);
    }
*/

    // === Queries ===
    public int getType() {
        return getTransformType();
    }

    public boolean isAffine() {
        return true;
    }

    // === invert ===
    // exception is used
/*
    public void invert() {
        try {
            super.invert();
        }
        catch (NoninvertibleTransformException e) {
            throw new ArithmeticException("TGrafMatrix.invert(): Singular Transform");
        }
    }
*/

    // === Expert ===
    public double getElement(int index) {
        return super.getElement(index);
    }

    public Object clone() {
        return new TGrafMatrix(this);
    }
}
