/*
 * @(#)TNurbDiscretiser.java	1.7 98/07/06
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

/*
    The old vertex engine and discretiser stuff ...
    When all else don't do what you want, write your own discretiser... - tt.93
*/
abstract class TNurbDiscretiser {

    private static final TGRect fkInfiniteRect =  // Why?  Look at TGStatics.java
        new TGRect(
                   TGStatics.kInfRectTopLeftX,
                   TGStatics.kInfRectTopLeftY,
                   TGStatics.kInfRectBottomRightX,
                   TGStatics.kInfRectBottomRightY);

    protected TGRect fClipBounds;
    protected double fResolution;

    TNurbDiscretiser(TGRect clipBounds) {
        this(clipBounds, 0.1);
    }
    TNurbDiscretiser(TGRect clipBounds, double resolution) {
        fClipBounds = clipBounds;
        fResolution = resolution;
    }

    abstract void begin(boolean closed);    // initiliase the environment, if any
    abstract void endPath();                // end of path, ie. at c0 discontinuity or the end
    abstract void addPoint(TGPoint point);

    protected void DiscretizeOrderTwo(
        TGParametricArray knotVector,
        TGRPointArray pointArray,
        boolean closed)
    {
        int numPts = pointArray.getNumberOfPoints();
        double knots[] = knotVector.getArray();
        TGRPoint points[] = pointArray.getArray();

        begin(closed);

        for (int i = 0; i < numPts; i++) {
            addPoint(points[i].divW());
            // check for c0 discontinuity...
            if (knots[i + 1] == knots[i + 2]) {
                endPath();
            }
        }
        endPath();
    }

    //
    // NURB curve stuff is out of the Killer B's: Beatty, Bartels, Barsky,
    // "An Introduction to the use of Curves & Surfaces in C.G. and
    // Geometric Design".
    //

    protected void DiscretizeNurb(
        int uorder,
        TGParametricArray knotsVector,
        TGRPointArray controlPoints,
        boolean closed)
    {
        int order = uorder;
        int numPts = controlPoints.getNumberOfPoints();
        int i, j, steps = 1, lastBrkPoint;
        TGRect segmentBounds;
        boolean c0disCon = false;
        boolean skippedSegment = false;
        TGRPoint delta[] = (new TGRPointArray(order + 1)).getArray();
        TGRPoint t[] = (new TGRPointArray(order + 1)).getArray();

        // Check to see if at least one rational point exists.  If not,
        // divides can be avoided.  Note: Could speed rendering up by
        // looking for rationals on a span by span basis
        boolean RatFlag = false;
        double knots[] = knotsVector.getArray();
        TGRPoint controlPtr[] = controlPoints.getArray();
        for (i = 0; i < numPts; i++) {
            if (controlPtr[i].fW != 1.0) {
                RatFlag = true;
                break;
            }
        }

        double u, u1, first_u, last_u, tmp_u;

        // Find the interval for the first span
        first_u = knots[order-1];
        u = first_u;
        last_u = knots[numPts];
        int brkPoint = prStatic.FindInterval(u, knots, numPts, order); // prKnots
        u1 = knots[brkPoint + 1];

        begin(closed);

        // Step through the spans of the curve, forward differencing each span.
        while (u < last_u) {
            // Compute the bounds of the current segment
            segmentBounds =
                prStatic.GetSegmentBounds( // prSplineRoutines
                    brkPoint, order, controlPoints, RatFlag);

            // Is this segment visible?
            skippedSegment = !NonEmptyIntersects(fClipBounds, segmentBounds);
            if (!skippedSegment) {
                steps = GetCurveSteps(brkPoint, order, controlPtr, fResolution);

                // So we have at least order initial samples to prime the
                // forward diff
                steps = (steps < order)    ? order : steps;

                if (steps > 5000)    // If the curve is too huge,
                {                    // just draw the control mesh
                    for (j = brkPoint - order + 1; j < brkPoint; j++)
                        this.addPoint(controlPtr[j].divW());
                    skippedSegment = true;    // So no attempt is made to use FD info
                }
            }                        // If the segment was skipped, and we're drawing
            else                    // a loop, add the ctl points so the loop
            {                        // topology remains correct
                if (closed)
                    for (j = brkPoint - order + 1; j < brkPoint; j++)
                        addPoint(controlPtr[j].divW());
            }

            if (!skippedSegment)    // Segment is visible and of reasonable size
            {
                tmp_u = (u1 - u) / (double)steps;
                for (j = 0; j < order; j++) {
                    prStatic.NurbEval( // prSplineRoutines
                        controlPtr, (int)brkPoint, u + ((double)j * tmp_u), knots,
                        (int)order, RatFlag, t[j]);
                }

                // Note first point on the curve or a c0 discontinuity
                delta[0].copyFrom(t[0]);

                // store the first point if not a closed loop
                if ((u == first_u) || c0disCon) {
                    // Pin it to the bounds in case of round-off error
                    TGPoint vertex = RatFlag ? delta[0].divW() : delta[0].dropW();
                    segmentBounds.pin(vertex);
                    addPoint(vertex);
                }

                // Build the forward difference table from the initial samples on
                // the curve.
                int k;
                if (RatFlag)
                    for (i = 1; i < order; i++) {
                        k = 1;
                        for (j = 0; j < order - i; j++, k++) {
                            t[j].fX = t[k].fX - t[j].fX;
                            t[j].fY = t[k].fY - t[j].fY;
                            t[j].fW = t[k].fW - t[j].fW;
                        }
                        delta[i].copyFrom(t[0]);
                    }
                else
                    for (i = 1; i < order; i++) {
                        k = 1;
                        for (j = 0; j < order - i; j++, k++) {
                            t[j].fX = t[k].fX - t[j].fX;
                            t[j].fY = t[k].fY - t[j].fY;
                        }
                        delta[i].copyFrom(t[0]);
                    }

                // Step through the curve span
                for (i = 0; i < steps; i++) {
                    // Update the forward differences
                    k = 1;
                    if (RatFlag) {
                        for (j = 0; j < order-1; j++, k++) {
                            delta[j].fX += delta[k].fX;
                            delta[j].fY += delta[k].fY;
                            delta[j].fW += delta[k].fW;
                        }
                    }
                    else {
                        for (j = 0; j < order-1; j++, k++) {
                            delta[j].fX += delta[k].fX;
                            delta[j].fY += delta[k].fY;
                        }
                    }

                    // Don't emit the last point in the segment here - if it's
                    // a c0 discon, we want to emit the actual control point to
                    // eliminate the accumulated forward difference error.
                    if (i < steps - 1) {
                        // Pin it to the bounds in case of round-off error
                        TGPoint vertex =
                            RatFlag ? delta[0].divW() : delta[0].dropW();
                        segmentBounds.pin(vertex);
                        addPoint(vertex);
                    }
                }
            }

            // Move to the next interval on the curve.
            u = u1;
            lastBrkPoint = brkPoint;
            brkPoint = prStatic.FindNextInterval(brkPoint, knots, numPts); // prKnots
            u1 = knots[brkPoint + 1];

            if (closed || (!skippedSegment)) {
                // If we're on the last segment, then FindNextInterval
                // just returns brkPoint+1, so check for a c0 there.
                // Otherwise, brkPoint really is at the start of a new
                // segment, so look backwards for a possible c0 or c1 discon.
                c0disCon = true;
                boolean sameKnot, c1disCon = true;
                if (u == last_u) {
                    for (j = brkPoint; j < brkPoint + order; j++) {
                        sameKnot = (u == knots[j]);
                        c0disCon = c0disCon && (u == knots[j]);
                        if (j < (brkPoint + order - 1))
                            c1disCon = c1disCon && sameKnot;
                    }
                }
                else {
                    for (j = brkPoint; j >= brkPoint - (order - 1); j--) {
                        sameKnot = (u == knots[j]);
                        c0disCon = c0disCon && (u == knots[j]);
                        if (j >= brkPoint - order + 2)
                            c1disCon = c1disCon && sameKnot;
                    }
                }

                // If the next span starts with a c0 or c1 discontinuity, then the
                // span we just finished ended with one (duh).  If this is
                // the case, then the endpoint of the curve is the same as the
                // last control point in the segment, so emit the last control
                // point instead of the one computed by the forward difference,
                // to avoid the accumulated round-off in the forward difference.
                if (c0disCon || c1disCon) {
                    addPoint(controlPtr[lastBrkPoint].divW());
                    if (c0disCon)
                        endPath();
                }
                else {
                    // If the segment was skipped, then use ctl mesh because
                    // delta wasn't generated
                    if (skippedSegment)
                        addPoint(controlPtr[lastBrkPoint].divW());
                    else {
                        // Pin it to the bounds in case of round-off error
                        TGPoint vertex = RatFlag ? delta[0].divW() : delta[0].dropW();
                        segmentBounds.pin(vertex);
                        addPoint(vertex);
                    }
                }
            }
            else
                // The segment was skipped and it wasn't closed, so we can
                // just jump to the next segment!
                endPath();

        }

        endPath();        // Dan sez it's OK if End is called twice.
    }

    // Version of Intersects that ignores whether or not a rect is empty
    // (needed so perfectly horizontal or verticle rects still work)
    private static boolean NonEmptyIntersects(TGRect a, TGRect b) {
        return ((a.equals(fkInfiniteRect)) ||
		((a.x < b.getMaxX()) && (a.getMaxX() > b.x) &&
		 (a.y < b.getMaxY()) && (a.getMaxY() > b.y)));
    }

    //
    // After many dead ends, we wind up with this brute(!) force stepsize
    // estimator.  "All the world's a quadratic" method, just calculates the
    // maximum for a hypothetical quadratic segment based on the control mesh.
    //
    private static int GetCurveSteps(
        int brkPoint,    // Must be of type long so comparisons
        int order,        // with negative numbers works
        TGRPoint V[],
        double tolerance)
    {
        int numsteps = 0;
        TGPoint tmp;
        double xstep, ystep;

        // Should never see a linear curve (polyline) here, but if we do there's
        // an obvious answer
        if (order == 2)
            return 1;

        // Find the second derivative for a the hypothetical quadratic segment
        // formed by each subsequent set of three control points in the segment.
        // Use the maximum of this as the stepsize for the over-all segment.
        for (int i = brkPoint - order + 1; i < brkPoint - 1; i++)
        {
            tmp = V[i].divW();
            tmp.subtractBy(TGPoint.multiply(2, V[i+1].divW()));
            tmp.addBy(V[i+2].divW());
            xstep = Math.abs(2 * tmp.x) / tolerance;
            ystep = Math.abs(2 * tmp.y) / tolerance;
            numsteps = (int)Math.max(numsteps, Math.sqrt(Math.max(xstep, ystep)));
        }

        return numsteps;
    }
}
