/*
 * @(#)AppContext.java	1.15 98/09/20
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt;

import java.awt.Frame;
import java.awt.Toolkit;
import java.awt.event.InvocationEvent;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.Hashtable;
import java.util.Enumeration;

/**
 * The AppContext is a table referenced by ThreadGroup which stores
 * application service instances.  (If you are not writing an application
 * service, or don't know what one is, please do not use this class.)
 * The AppContext allows applet access to what would otherwise be
 * potentially dangerous services, such as the ability to peek at
 * EventQueues or change the look-and-feel of a Swing application.<p>
 *
 * Most application services use a singleton object to provide their
 * services, either as a default (such as getSystemEventQueue or 
 * getDefaultToolkit) or as static methods with class data (System).
 * The AppContext works with the former method by extending the concept
 * of "default" to be ThreadGroup-specific.  Application services
 * lookup their singleton in the AppContext.<p>
 *
 * For example, here we have a Foo service, with its pre-AppContext
 * code:<p>
 * <code><pre>
 *    public class Foo {
 *        private static Foo defaultFoo = new Foo();
 *
 *        public static Foo getDefaultFoo() {
 *            return defaultFoo;
 *        }
 *
 *    ... Foo service methods
 *    }</pre></code><p>
 *
 * The problem with the above is that the Foo service is global in scope,
 * so that applets and other untrusted code can execute methods on the
 * single, shared Foo instance.  The Foo service therefore either needs
 * to block its use by untrusted code using a SecurityManager test, or
 * restrict its capabilities so that it doesn't matter if untrusted code
 * executes it.<p>
 *
 * Here's the Foo class written to use the AppContext:<p>
 * <code><pre>
 *    public class Foo {
 *        public static Foo getDefaultFoo() {
 *            Foo foo = (Foo)AppContext.getAppContext().get(Foo.class);
 *            if (foo == null) {
 *                foo = new Foo();
 *                getAppContext().put(Foo.class, foo);
 *            }
 *            return foo;
 *        }
 *
 *    ... Foo service methods
 *    }</pre></code><p>
 *
 * Since a separate AppContext exists for each ThreadGroup, trusted
 * and untrusted code have access to different Foo instances.  This allows
 * untrusted code access to "system-wide" services -- the service remains
 * within the AppContext "sandbox".  For example, say a malicious applet 
 * wants to peek all of the key events on the EventQueue to listen for
 * passwords; if separate EventQueues are used for each ThreadGroup
 * using AppContexts, the only key events that applet will be able to
 * listen to are its own.  A more reasonable applet request would be to
 * change the Swing default look-and-feel; with that default stored in
 * an AppContext, the applet's look-and-feel will change without 
 * disrupting other applets or potentially the browser itself.<p>
 *
 * Because the AppContext is a facility for safely extending application
 * service support to applets, none of its methods may be blocked by a
 * a SecurityManager check in a valid Java implementation.  Applets may
 * therefore safely invoke any of its methods without worry of being 
 * blocked.
 *
 * @author  Thomas Ball
 * @author  Fred Ecks
 * @version 1.15 09/20/98
 */
public final class AppContext {

    /* Since the contents of an AppContext are unique to each Java 
     * session, this class should never be serialized. */

    /* The key to put()/get() the Java EventQueue into/from the AppContext.
     */
    public static final Object EVENT_QUEUE_KEY = new StringBuffer("EventQueue");

    /* A map of AppContexts, referenced by ThreadGroup.
     */
    private static Hashtable threadGroup2appContext = null;

    /*
     * The hashtable associated with this AppContext.  A private delegate
     * is used instead of subclassing Hashtable so as to avoid all of
     * Hashtable's potentially risky methods, such as clear(), elements(),
     * putAll(), etc.  (It probably doesn't need to be final since the
     * class is, but I don't trust the compiler to be that smart.)
     */
    private final Hashtable table;

    private final ThreadGroup threadGroup;

    private boolean isDisposed = false;	// true if AppContext is disposed

    static {
	// On the main Thread, we get the ThreadGroup, make a corresponding
	// AppContext, and instantiate the Java EventQueue.  This way, legacy
	// code is unaffected by the move to multiple AppContext ability.
	AccessController.doPrivileged(new PrivilegedAction() {
	  public Object run() {
	    ThreadGroup currentThreadGroup =
				Thread.currentThread().getThreadGroup();
	    ThreadGroup parentThreadGroup = currentThreadGroup.getParent();
	    while (parentThreadGroup != null) {
		// Find the root ThreadGroup to construct our main AppContext
		currentThreadGroup = parentThreadGroup;
		parentThreadGroup = currentThreadGroup.getParent();
	    }
	    AppContext appContext = new AppContext(currentThreadGroup);
	    numAppContexts = 1;
	    return appContext;
	  }
	});
    }

    /*
     * The total number of AppContexts, system-wide.  This number is
     * incremented at the beginning of the constructor, and decremented
     * at the end of dispose().  getAppContext() checks to see if this
     * number is 1.  If so, it returns the sole AppContext without
     * checking Thread.currentThread().
     */
    private static int numAppContexts;

    /**
     * Constructor for AppContext.  This method is <i>not</i> public,
     * nor should it ever be used as such.  The proper way to construct
     * an AppContext is through the use of SunToolkit.createNewAppContext.
     * A ThreadGroup is created for the new AppContext, a Thread is
     * created within that ThreadGroup, and that Thread calls
     * SunToolkit.createNewAppContext before calling anything else.
     * That creates both the new AppContext and its EventQueue.
     * 
     * @param   threadGroup     The ThreadGroup for the new AppContext
     * @see	sun.awt.SunToolkit
     * @since   JDK1.2
     */
    AppContext(ThreadGroup threadGroup) {
	numAppContexts++;

        table = new Hashtable(2);
        if (threadGroup2appContext == null) {
            threadGroup2appContext = new Hashtable(2, 0.2f);
        }
	this.threadGroup = threadGroup;
        threadGroup2appContext.put(threadGroup, this);
    }

    private static MostRecentThreadAppContext mostRecentThreadAppContext = null;

    /**
     * Returns the appropriate AppContext for the caller, 
     * as determined by its ThreadGroup.
     *
     * @return  the AppContext for the caller.
     * @see     java.lang.ThreadGroup
     * @since   JDK1.2
     */
    public final static AppContext getAppContext() {
	if ((numAppContexts == 1) && (mostRecentThreadAppContext != null))
	    return mostRecentThreadAppContext.appContext;

	final Thread currentThread = Thread.currentThread();

	// Note: this most recent Thread/AppContext caching is thread-hot.
	// A simple test using SwingSet found that 96.8% of lookups
	// were matched using the most recent Thread/AppContext.  By
	// instantiating a simple MostRecentThreadAppContext object on
	// cache misses, the cache hits can be processed without
	// synchronization.

	MostRecentThreadAppContext recent = mostRecentThreadAppContext;
	if ((recent != null) && (recent.thread == currentThread))
	    return recent.appContext;

	return (AppContext)AccessController.doPrivileged(new PrivilegedAction() {
	  public Object run() {
	    // Get the current ThreadGroup, and look for it and its
	    // parents in the hash from ThreadGroup to AppContext --
	    // it should be found, because we use createNewContext()
	    // when new AppContext objects are created.
	    ThreadGroup currentThreadGroup = currentThread.getThreadGroup();
	    ThreadGroup threadGroup = currentThreadGroup;
	    AppContext appContext =
		    (AppContext)threadGroup2appContext.get(threadGroup);
	    while (appContext == null) {
		threadGroup = threadGroup.getParent();
		if (threadGroup == null) {
		    // If we get here, we're running under a ThreadGroup that
		    // has no AppContext associated with it.  This should never
		    // happen, because createNewContext() should be used by the
		    // toolkit to create the ThreadGroup that everything runs
		    // under.
		    throw new RuntimeException("Invalid ThreadGroup");
		}
		appContext = (AppContext)threadGroup2appContext.get(threadGroup);
	    }
	    // In case we did anything in the above while loop, we add
	    // all the intermediate ThreadGroups to threadGroup2appContext
	    // so we won't spin again.
	    for (ThreadGroup tg = currentThreadGroup; tg != threadGroup; tg = tg.getParent()) {
	        threadGroup2appContext.put(tg, appContext);
	    }
	    // Now we're done, so we cache the latest key/value
	    mostRecentThreadAppContext =
		new MostRecentThreadAppContext(currentThread, appContext);
	    return appContext;
	  }
	});
    }

    private long DISPOSAL_TIMEOUT = 5000;  // Default to 5-second timeout
					   // for disposal of all Frames
					   // (we wait for this time twice,
					   // once for dispose(), and once
					   // to clear the EventQueue).

    private long THREAD_INTERRUPT_TIMEOUT = 1000;
			    // Default to 1-second timeout for all
			    // interrupted Threads to exit, and another
			    // 1 second for all stopped Threads to die.

    /**
     * Disposes of this AppContext, all of its top-level Frames, and
     * all Threads and ThreadGroups contained within it.
     * 
     * This method must be called from a Thread which is not contained
     * within this AppContext.
     *
     * @exception  IllegalThreadStateException  if the current thread is
     *                                    contained within this AppContext
     * @since	JDK1.2
     */
    public void dispose() throws IllegalThreadStateException {
	// Check to be sure that the current Thread isn't in this AppContext
	if (this.threadGroup.parentOf(Thread.currentThread().getThreadGroup())) {
	    throw new IllegalThreadStateException(
		"Current Thread is contained within AppContext to be disposed."
	      );
	}

	synchronized(this) {
	    if (this.isDisposed) {
		return; // If already disposed, bail.
	    }
	    this.isDisposed = true;
	}

	// First, we post an InvocationEvent to be run on the
	// EventDispatchThread which disposes of all top-level Frames

	final Object notificationLock = new Object();

	Runnable runnable = new Runnable() { public void run() {
	    Frame [] frames = Frame.getFrames();
	    for (int i = frames.length - 1; i >= 0; i--) {
		frames[i].dispose();	// Dispose of all top-level Frames
	    }
	    synchronized(notificationLock) {
		notificationLock.notifyAll(); // Notify caller that we're done
	    }
	} };
	synchronized(notificationLock) {
	    SunToolkit.postEvent(this,
		new InvocationEvent(Toolkit.getDefaultToolkit(), runnable));
	    try {
		notificationLock.wait(DISPOSAL_TIMEOUT);
	    } catch (InterruptedException e) { }
	}

	// Next, we post another InvocationEvent to the end of the
	// EventQueue.  When it's executed, we know we've executed all
	// events in the queue.

	runnable = new Runnable() { public void run() {
	    synchronized(notificationLock) {
		notificationLock.notifyAll(); // Notify caller that we're done
	    }
	} };
	synchronized(notificationLock) {
	    SunToolkit.postEvent(this,
		new InvocationEvent(Toolkit.getDefaultToolkit(), runnable));
	    try {
		notificationLock.wait(DISPOSAL_TIMEOUT);
	    } catch (InterruptedException e) { }
	}

	// Next, we interrupt all Threads in the ThreadGroup
	this.threadGroup.interrupt();
	    // Note, the EventDispatchThread we've interrupted may dump an
	    // InterruptedException to the console here.  This needs to be
	    // fixed in the EventDispatchThread, not here.

	// Next, we sleep 10ms at a time, waiting for all of the active
	// Threads in the ThreadGroup to exit.

	long startTime = System.currentTimeMillis();
	long endTime = startTime + (long)THREAD_INTERRUPT_TIMEOUT;
	while ((this.threadGroup.activeCount() > 0) &&
	       (System.currentTimeMillis() < endTime)) {
	    try {
		Thread.sleep(10);
	    } catch (InterruptedException e) { }
	}

	// Then, we stop any remaining Threads
	this.threadGroup.stop();

	// Next, we sleep 10ms at a time, waiting for all of the active
	// Threads in the ThreadGroup to die.

	startTime = System.currentTimeMillis();
	endTime = startTime + (long)THREAD_INTERRUPT_TIMEOUT;
	while ((this.threadGroup.activeCount() > 0) &&
	       (System.currentTimeMillis() < endTime)) {
	    try {
		Thread.sleep(10);
	    } catch (InterruptedException e) { }
	}

	// Next, we remove this and all subThreadGroups from threadGroup2appContext
	int numSubGroups = this.threadGroup.activeGroupCount();
	if (numSubGroups > 0) {
	    ThreadGroup [] subGroups = new ThreadGroup[numSubGroups];
	    numSubGroups = this.threadGroup.enumerate(subGroups);
	    for (int subGroup = 0; subGroup < numSubGroups; subGroup++) {
		threadGroup2appContext.remove(subGroups[subGroup]);
	    }
	}

	MostRecentThreadAppContext recent = mostRecentThreadAppContext;
	if ((recent != null) && (recent.appContext == this))
	    mostRecentThreadAppContext = null;
		// If the "most recent" points to this, clear it for GC

	// Finally, we destroy the ThreadGroup entirely.
	try {
	    this.threadGroup.destroy();
	} catch (IllegalThreadStateException e) {
	    // Fired if not all the Threads died, ignore it and proceed
	}

	this.table.clear(); // Clear out the Hashtable to ease garbage collection

	numAppContexts--;
    }

    private MostRecentKeyValue mostRecentKeyValue = null;

    /**
     * Returns the value to which the specified key is mapped in this context.
     *
     * @param   key   a key in the AppContext.
     * @return  the value to which the key is mapped in this AppContext;
     *          <code>null</code> if the key is not mapped to any value.
     * @see     #put(Object, Object)
     * @since   JDK1.2
     */
    public Object get(Object key) {
	// Note: this most recent key/value caching is thread-hot.
	// A simple test using SwingSet found that 72% of lookups
	// were matched using the most recent key/value.  By instantiating
	// a simple MostRecentKeyValue object on cache misses, the
	// cache hits can be processed without synchronization.

	MostRecentKeyValue recent = mostRecentKeyValue;
	if ((recent != null) && (recent.key == key))
	    return recent.value;

	Object value = table.get(key);
	mostRecentKeyValue = new MostRecentKeyValue(key, value);
	return value;
    }

    /**
     * Maps the specified <code>key</code> to the specified 
     * <code>value</code> in this AppContext.  Neither the key nor the 
     * value can be <code>null</code>.
     * <p>
     * The value can be retrieved by calling the <code>get</code> method 
     * with a key that is equal to the original key. 
     *
     * @param      key     the AppContext key.
     * @param      value   the value.
     * @return     the previous value of the specified key in this 
     *             AppContext, or <code>null</code> if it did not have one.
     * @exception  NullPointerException  if the key or value is
     *               <code>null</code>.
     * @see     #get(Object)
     * @since   JDK1.2
     */
    public Object put(Object key, Object value) {
	MostRecentKeyValue recent = mostRecentKeyValue;
	if ((recent != null) && (recent.key == key))
	    recent.value = value;
	return table.put(key, value);
    }

    /**
     * Removes the key (and its corresponding value) from this 
     * AppContext. This method does nothing if the key is not in the
     * AppContext.
     *
     * @param   key   the key that needs to be removed.
     * @return  the value to which the key had been mapped in this AppContext,
     *          or <code>null</code> if the key did not have a mapping.
     * @since   JDK1.2
     */
    public Object remove(Object key) {
	MostRecentKeyValue recent = mostRecentKeyValue;
	if ((recent != null) && (recent.key == key))
	    recent.value = null;
        return table.remove(key);
    }

    /**
     * Returns the root ThreadGroup for all Threads contained within
     * this AppContext.
     * @since	JDK1.2
     */
    public ThreadGroup getThreadGroup() {
	return threadGroup;
    }

    /**
     * Returns a string representation of this AppContext.
     * @since   JDK1.2
     */
    public String toString() {
	return getClass().getName() + "[threadGroup=" + threadGroup.getName() + "]";
    }
}

final class MostRecentThreadAppContext {
    final Thread thread;
    final AppContext appContext;
    MostRecentThreadAppContext(Thread key, AppContext value) {
	thread = key;
	appContext = value;
    }
}

final class MostRecentKeyValue {
    final Object key;
    Object value;
    MostRecentKeyValue(Object k, Object v) {
	key = k;
	value = v;
    }
}
