/*
 * @(#)SunToolkit.java	1.28 98/09/16
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt;

import java.awt.*;
import java.awt.image.*;
import java.awt.geom.AffineTransform;
import java.io.*;
import java.net.URL;
import java.net.JarURLConnection;
import java.util.*;
import sun.misc.SoftCache;
import sun.awt.font.FontDesignMetrics;
import sun.awt.image.*;
import sun.awt.im.InputMethod;
import sun.awt.im.InputMethodHighlightMapping;
import sun.awt.font.NativeFontWrapper;

public abstract class SunToolkit extends Toolkit {

    private static boolean hotjavaUrlCache = false; // REMIND: UGH!!!!!

    /* The key to put()/get() the PostEventQueue into/from the AppContext.
     */
    private static final String POST_EVENT_QUEUE_KEY = "PostEventQueue";

    public SunToolkit() {
	EventQueue eventQueue;
        String eqName = Toolkit.getProperty("AWT.EventQueueClass",
                                            "java.awt.EventQueue");
        try {
            eventQueue = (EventQueue)Class.forName(eqName).newInstance();
        } catch (Exception e) {
            System.err.println("Failed loading " + eqName + ": " + e);
            eventQueue = new EventQueue();
        }
	AppContext appContext = AppContext.getAppContext();
	appContext.put(AppContext.EVENT_QUEUE_KEY, eventQueue);

	PostEventQueue postEventQueue = new PostEventQueue(eventQueue);
	appContext.put(POST_EVENT_QUEUE_KEY, postEventQueue);
    }

    /*
     * Create a new AppContext, along with its EventQueue, for a
     * new ThreadGroup.  Browser code, for example, would use this
     * method to create an AppContext & EventQueue for an Applet.
     */
    public static AppContext createNewAppContext() {
	ThreadGroup threadGroup = Thread.currentThread().getThreadGroup();
	EventQueue eventQueue;
        String eqName = Toolkit.getProperty("AWT.EventQueueClass",
                                            "java.awt.EventQueue");
        try {
            eventQueue = (EventQueue)Class.forName(eqName).newInstance();
        } catch (Exception e) {
            System.err.println("Failed loading " + eqName + ": " + e);
            eventQueue = new EventQueue();
        }
	AppContext appContext = new AppContext(threadGroup);
	appContext.put(AppContext.EVENT_QUEUE_KEY, eventQueue);

	PostEventQueue postEventQueue = new PostEventQueue(eventQueue);
	appContext.put(POST_EVENT_QUEUE_KEY, postEventQueue);

	return appContext;
    }

    // mapping of components to peers, Hashtable<Component,Peer>
    protected static final Hashtable peerMap = new Hashtable();  

    /*
     * Fetch the peer associated with the given target (as specified
     * in the peer creation method).  This can be used to determine 
     * things like what the parent peer is.  If the target is null
     * or the target can't be found (either because the a peer was
     * never created for it or the peer was disposed), a null will
     * be returned.
     */
    protected static Object targetToPeer(Object target) {
	if (target != null) {
	    return peerMap.get(target);
	}
	return null;
    }

    protected static void targetDisposedPeer(Object target, Object peer) {
	if (target != null && peer != null) {
	    if (peerMap.get(target) == peer) {
		peerMap.remove(target);
	    }
	}
    }

    // mapping of Components to AppContexts, WeakHashMap<Component,AppContext>
    private static final Map appContextMap = 
        Collections.synchronizedMap(new WeakHashMap());

    /*
     * Fetch the AppContext associated with the given target.
     * This can be used to determine things like which EventQueue
     * to use for posting events to a Component.  If the target is
     * null or the target can't be found, a null with be returned.
     */
    public static AppContext targetToAppContext(Object target) {
	if (target != null) {
	    return (AppContext)appContextMap.get(target);
	}
	return null;
    }

    /*
     * Insert a mapping from target to AppContext, for later retrieval
     * via targetToAppContext() above.
     */
    public static void insertTargetMapping(Object target, AppContext appContext) {
	appContextMap.put(target, appContext);
    }

    /*
     * Post an AWTEvent to the Java EventQueue, using the PostEventQueue
     * to avoid possibly calling client code (EventQueueSubclass.postEvent())
     * on the toolkit (AWT-Windows/AWT-Motif) thread.
     */
    public static void postEvent(AppContext appContext, AWTEvent event) {
	PostEventQueue postEventQueue =
	    (PostEventQueue)appContext.get(POST_EVENT_QUEUE_KEY);
        postEventQueue.postEvent(event);
    }

    /*
     * Execute a chunk of code on the Java event handler thread for the
     * given target.  Does not wait for the execution to occur before
     * returning to the caller.
     */
    public static void executeOnEventHandlerThread(Object target, Runnable runnable) {
	AppContext appContext = targetToAppContext(target);
        postEvent(targetToAppContext(target),
            new MagicEvent(getDefaultToolkit(), runnable,
			   MagicEvent.PRIORITY_EVENT));
    }

    public Dimension getScreenSize() {
	return new Dimension(getScreenWidth(), getScreenHeight());
    }
    protected abstract int getScreenWidth();
    protected abstract int getScreenHeight();

    public FontMetrics getFontMetrics(Font font) {
        float ptSize = font.getSize2D();
        AffineTransform tx = (AffineTransform)font.getTransform();
        tx.scale(ptSize, ptSize);
        double[] matrix = new double[4];
        tx.getMatrix(matrix);
        float [] metrics = new float[4];
        NativeFontWrapper.getFontMetrics(
                    font, matrix, false, false, metrics);
	return new FontDesignMetrics(font,  metrics[0], metrics[1], metrics[2],
                                            metrics[3]);
    }
    
    public String[] getFontList() {
	String[] hardwiredFontList = {
	    "Dialog", "SansSerif", "Serif", "Monospaced", "DialogInput"

	    // -- Obsolete font names from 1.0.2.  It was decided that
	    // -- getFontList should not return these old names:
	    // , "Helvetica", "TimesRoman", "Courier", "ZapfDingbats"
	};
	return hardwiredFontList;
    }

    static SoftCache imgCache = new SoftCache();

    static synchronized java.awt.Image getImageFromHash(Toolkit tk, URL url) {
	SecurityManager sm = System.getSecurityManager();
	if (sm != null) {
	    try {
		java.security.Permission perm = 
		    url.openConnection().getPermission();
		if (perm != null) {
		    try {
			sm.checkPermission(perm);
		    } catch (SecurityException se) {
			// fallback to checkRead/checkConnect for pre 1.2
			// security managers
			if ((perm instanceof java.io.FilePermission) &&
			    perm.getActions().indexOf("read") != -1) {
			    sm.checkRead(perm.getName());
			} else if ((perm instanceof 
			    java.net.SocketPermission) &&
			    perm.getActions().indexOf("connect") != -1) {
			    sm.checkConnect(url.getHost(), url.getPort());
			} else {
			    throw se;
			}
		    }
		}
	    } catch (java.io.IOException ioe) {
		    sm.checkConnect(url.getHost(), url.getPort());
	    }
	}
	java.awt.Image img = (java.awt.Image)imgCache.get(url);
	if (img == null) {
	    try {	
		img = tk.createImage(new URLImageSource(url));
		imgCache.put(url, img);
	    } catch (Exception e) {
	    }
	}
	return img;
    }

    static synchronized java.awt.Image getImageFromHash(Toolkit tk,
                                                        String filename) {
	SecurityManager security = System.getSecurityManager();
	if (security != null) {
	    security.checkRead(filename);
	}
	java.awt.Image img = (java.awt.Image)imgCache.get(filename);
	if (img == null) {
	    try {	
		img = tk.createImage(new FileImageSource(filename));
		imgCache.put(filename, img);
	    } catch (Exception e) {
	    }
	}
	return img;
    }

    public java.awt.Image getImage(String filename) {
	return getImageFromHash(this, filename);
    }

    public java.awt.Image getImage(URL url) {
	return getImageFromHash(this, url);
    }

    public java.awt.Image createImage(String filename) {
	SecurityManager security = System.getSecurityManager();
	if (security != null) {
	    security.checkRead(filename);
	}
	return createImage(new FileImageSource(filename));
    }

    public java.awt.Image createImage(URL url) {
	SecurityManager sm = System.getSecurityManager();
	if (sm != null) {
	    try {
		java.security.Permission perm = 
		    url.openConnection().getPermission();
		if (perm != null) {
		    try {
			sm.checkPermission(perm);
		    } catch (SecurityException se) {
			// fallback to checkRead/checkConnect for pre 1.2
			// security managers
			if ((perm instanceof java.io.FilePermission) &&
			    perm.getActions().indexOf("read") != -1) {
			    sm.checkRead(perm.getName());
			} else if ((perm instanceof 
			    java.net.SocketPermission) &&
			    perm.getActions().indexOf("connect") != -1) {
			    sm.checkConnect(url.getHost(), url.getPort());
			} else {
			    throw se;
			}
		    }
		}
	    } catch (java.io.IOException ioe) {
		    sm.checkConnect(url.getHost(), url.getPort());
	    }
	}
	return createImage(new URLImageSource(url));
    }

    public java.awt.Image createImage(byte[] data, int offset, int length) {
	return createImage(new ByteArrayImageSource(data, offset, length));
    }

    protected EventQueue getSystemEventQueueImpl() {
	AppContext appContext = AppContext.getAppContext();
	EventQueue theEventQueue =
		(EventQueue)appContext.get(AppContext.EVENT_QUEUE_KEY);
        return theEventQueue;
    }

    /**
     * Give native peers the ability to query the native container 
     * given a native component (eg the direct parent may be lightweight).
     */
    public static Container getNativeContainer(Component c) {
	return Toolkit.getNativeContainer(c);
    }

    /**
     * Returns a new input method adapter for native input methods.
     */
    public abstract InputMethod getInputMethodAdapter() throws AWTException;

    /**
     * Returns whether enableInputMethods should be set to true for peered
     * TextComponent instances on this platform. False by default.
     */
    public boolean enableInputMethodsForTextComponent() {
        return false;
    }

    /**
     * Returns an input method highlight mapping interface that is
     * used to map abstract attributes to concrete ones.
     */
    public abstract InputMethodHighlightMapping getInputMethodHighlightMapping();
}



/*
 * PostEventQueue is a Thread that runs in the same AppContext as the
 * Java EventQueue.  It is a queue of AWTEvents to be posted to the
 * Java EventQueue.  The toolkit Thread (AWT-Windows/AWT-Motif) posts
 * events to this queue, which then calls EventQueue.postEvent().
 *
 * We do this because EventQueue.postEvent() may be overridden by client
 * code, and we mustn't ever call client code from the toolkit thread.
 */
class PostEventQueue extends Thread {
    static private int threadNum = 0;
    private EventQueueItem queueHead = null;
    private EventQueueItem queueTail = null;
    private boolean keepGoing = true;
    private final EventQueue eventQueue;

    PostEventQueue(EventQueue eq) {
	super("SunToolkit.PostEventQueue-"+threadNum);
	synchronized (PostEventQueue.class) { threadNum++; }
	eventQueue = eq;
	start();
    }

    /*
     * Continually post pending AWTEvents to the Java EventQueue.
     */
    public void run() {
	while (keepGoing && !isInterrupted()) {
	    try {
		EventQueueItem item;
		synchronized(this) {
		    while (keepGoing && (queueHead == null)) {
			wait();
		    }
		    if (!keepGoing)
			break;
		    item = queueHead;
		    queueHead = item.next;
		    if (queueHead == null)
			queueTail = null;
		}
		eventQueue.postEvent(item.event);
	    } catch (InterruptedException e) {
		keepGoing = false; // Exit gracefully when interrupted
	    }
	}
    }

    /*
     * Enqueue an AWTEvent to be posted to the Java EventQueue.
     */
    synchronized void postEvent(AWTEvent event) {
	EventQueueItem item = new EventQueueItem(event);

	if (queueHead == null) {
	    queueHead = queueTail = item;
	    notifyAll();
	} else {
	    queueTail.next = item;
	    queueTail = item;
	}
    }

    /*
     * Notifies this PostEventQueue to stop running.
     */
    synchronized void quitRunning() {
	keepGoing = false;
	notifyAll();
    }
} // class PostEventQueue

class EventQueueItem {
    AWTEvent event;
    EventQueueItem next;

    EventQueueItem(AWTEvent evt) {
        event = evt;
    }
} // class EventQueueItem
