/*
 * @(#)CMMImageLayout.java	1.9 98/04/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*********************************************************************/
/*
    Contains:    image layout description class

            Created by gbp, October 16, 1997

    Copyright:    (C) 1997 by Eastman Kodak Company, all rights reserved.

    File: CMMImageLayout.java  @(#)CMMImageLayout.java	1.2    11/06/97
*/
/*********************************************************************/


/**********************************************************************
 **********************************************************************
 **********************************************************************
 *** COPYRIGHT (c) Eastman Kodak Company, 1997                 ***
 *** As  an unpublished  work pursuant to Title 17 of the United    ***
 *** States Code.  All rights reserved.                             ***
 **********************************************************************
 **********************************************************************
 **********************************************************************/


package sun.awt.color;

import java.awt.image.BufferedImage;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import sun.awt.image.ByteComponentRaster;
import sun.awt.image.IntegerComponentRaster;


/* standard image description for communication with the CMM */
class CMMImageLayout {

    /* define image layout base so that BufferedImage.TYPE_INT_ARGB, etc.
       can be used directly */

    private static final int typeBase = 256;

    static public final int typeComponentUByte = typeBase + 0;
    /* component organization, each is 8 bits */

    static public final int typeComponentUShort12 = typeBase + 1;
    /* component organization, each is the low 12 bits of UShort */

    static public final int typeComponentUShort = typeBase + 2;
    /* component organization, each is 16 bits */

    static public final int typePixelUByte = typeBase + 3;
    /* pixel organization, each component is 8 bits */

    static public final int typePixelUShort12 = typeBase + 4;
    /* pixel organization, each component is the low 12 bits of UShort */

    static public final int typePixelUShort = typeBase + 5;
    /* pixel organization, each component is 16 bits */

    static public final int typeShort555 = typeBase + 6;
    /* red is 14-10, green is 9-5, blue is 4-0 */

    static public final int typeShort565 = typeBase + 7;
    /* red is 15-11, green is 10-5, blue is 4-0 */

    static public final int typeInt101010 = typeBase + 8;
    /* red is 29-20, green is 19-10, blue is 9-0 */

    public int    Type;            /* component format */
    public int    NumCols;         /* # of columns */
    public int    NumRows;         /* # of rows */
    public int    OffsetColumn;    /* byte offset to next column */
    public int    OffsetRow;       /* byte offset to next row */
    public int    NumChannels;     /* # of discrete components */
    public Object[]    chanData;   /* component arrays */
    public int[]  DataOffsets;     /* offset(s) in number of array elements
                                      from the beginning of the array(s) to
                                      the element in which the first sample
                                      of the first scanline is stored */

    public CMMImageLayout (int nChans) {
        chanData = new Object [nChans];
    }

    /* convert a BufferedImage into a common form which is easier for
       the native cmm to use */
    CMMImageLayout (int nPels, int nComps, int compSize)
    {
        switch (compSize) {
        case 1:
            Type = typePixelUByte;
            break;

        case 2:
            Type = typePixelUShort;
            break;

        default:
            throw new IllegalArgumentException (
                "more than 2 bytes per component");
        }

        chanData = new Object [1];             /* just a single address */

        NumCols = nPels;                       /* # of columns */
        NumRows = 1;                           /* # of rows */
        OffsetColumn = nComps * compSize;      /* bytes to next column */
        OffsetRow = NumCols * OffsetColumn;    /* bytes to next row */
        NumChannels = nComps;                  /* # of discrete channels */
        DataOffsets = new int[NumChannels];    /* unused for now, but needs
                                                  to be there */
    }



    /* convert a BufferedImage into a common form which is easier for
       the native cmm to use */
    CMMImageLayout (BufferedImage theImage)
    {
    int            i1;
    IntegerComponentRaster    intRaster;
    ByteComponentRaster    byteRaster;

        Type = theImage.getType();

        /* System.out.println ("Type = " + Type); */

        switch (Type) {
        case BufferedImage.TYPE_INT_RGB:
        case BufferedImage.TYPE_INT_ARGB:
        case BufferedImage.TYPE_INT_ARGB_PRE:
        
            NumChannels = 1;        /* rgb data packed into a 32-bit integer */
            NumCols = theImage.getWidth();    /* # of columns */
            NumRows = theImage.getHeight();    /* # of rows */

            chanData = new Object [NumChannels];    /* just a single address */
            DataOffsets = new int[NumChannels];

            OffsetColumn = 4;        /* bytes to next column */

            intRaster = (IntegerComponentRaster) theImage.getRaster();
            OffsetRow = intRaster.getScanlineStride() * 4; /* bytes to next
                                                              row */
            for (i1 = 0; i1 < NumChannels; i1++) {
                chanData[i1] = (Object) intRaster.getDataStorage();
                DataOffsets[i1] = intRaster.getDataOffset(i1);
            }

            break;

        case BufferedImage.TYPE_3BYTE_BGR:
        case BufferedImage.TYPE_4BYTE_ABGR:
        case BufferedImage.TYPE_4BYTE_ABGR_PRE:

            NumChannels = 1;        /* rgb data in separate bytes */
            NumCols = theImage.getWidth();    /* # of columns */
            NumRows = theImage.getHeight();    /* # of rows */

            chanData = new Object [NumChannels];    /* just a single address */
            DataOffsets = new int[NumChannels];

            if (Type == BufferedImage.TYPE_3BYTE_BGR) {
                OffsetColumn = 3;        /* bytes to next column */
            }
            else {
                OffsetColumn = 4;        /* bytes to next column */
            }

            byteRaster = (ByteComponentRaster) theImage.getRaster();
            OffsetRow = byteRaster.getScanlineStride(); /* bytes to next row */
            for (i1 = 0; i1 < NumChannels; i1++) {
                chanData[i1] = (Object) byteRaster.getDataStorage();
                DataOffsets[i1] = byteRaster.getDataOffset(i1);
            }

            break;

        case BufferedImage.TYPE_CUSTOM:

            break;

        case BufferedImage.TYPE_BYTE_BINARY:

            break;

        case BufferedImage.TYPE_BYTE_INDEXED:

            break;

            // REMIND: Missing other BufferedImage types
        default:
            System.out.println ("Unknown type:" + Type);
        }
    }
}
