/*
 * @(#)ICC_Transform.java	1.11 98/09/15
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*********************************************************************/
/*
    Contains:    ICC transform class

            Created by gbp, October 1, 1997

    Copyright:    (C) 1997 by Eastman Kodak Company, all rights reserved.

    @(#)ICC_Transform.java	1.2    11/06/97
*/
/*********************************************************************/


/**********************************************************************
 **********************************************************************
 **********************************************************************
 *** COPYRIGHT (c) Eastman Kodak Company, 1997                      ***
 *** As  an unpublished  work pursuant to Title 17 of the United    ***
 *** States Code.  All rights reserved.                             ***
 **********************************************************************
 **********************************************************************
 **********************************************************************/


package sun.awt.color;

import java.awt.color.ICC_Profile;
import java.awt.color.ProfileDataException;
import java.awt.color.CMMException;
import java.awt.image.BufferedImage;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import java.awt.image.ColorModel;
import java.awt.image.SampleModel;


public class ICC_Transform {
    long    ID;

    public static final int Any = -1;/* any rendering type, whichever is
                                        available */
                                     /* search order is icPerceptual,
                                        icRelativeColorimetric, icSaturation */

    /* Transform types */
    public static final int In = 1;
    public static final int Out = 2;
    public static final int Gamut = 3;
    public static final int Simulation = 4;

    /* the class initializer */
    static {
        if (ProfileDeferralMgr.deferring) {
            ProfileDeferralMgr.activateProfiles();
        }
    }

    /**
     * Constructs an empty ICC_Transform object.
     */
    public ICC_Transform ()
    {
    }


   /**
     * Constructs an ICC_Transform object corresponding to an ICC profile transform.
     */
    public ICC_Transform (    ICC_Profile profile,
                int renderType,
                int transformType)
    {
	if (profile == null) {
	    CMM.checkStatus(CMM.cmmStatBadProfile);
        }
        CMM.checkStatus(CMM.cmmGetTransform (profile, renderType, transformType, this));
    }



    /**
     * Constructs an ICC_Transform object from a list of ICC_Transform objects.
     */
    public ICC_Transform (    ICC_Transform[]    transforms)
    {
    int        cmmStatus;
    long[]    transformIDs;
    int        nTransforms, i1;

        nTransforms = transforms.length;

        transformIDs = new long [nTransforms];

        /* put transform IDs into an array */
        for (i1 = 0; i1 < nTransforms; i1++) {
            transformIDs [i1] = transforms[i1].ID;
        }

        cmmStatus = CMM.cmmCombineTransforms (transformIDs, this);

        if ((cmmStatus != CMM.cmmStatSuccess) || (ID == 0)) {
            throw new ProfileDataException ("Invalid profile sequence");
        }
    }



    /**
     * returns the ID of an ICC_Transform object.
     */
    long getID ()
    {
        return ID;
    }


    /**
     * Frees the resources associated with an ICC_Transform object.
     */
    protected void finalize ()
    {
        CMM.checkStatus (CMM.cmmFreeTransform (ID));
    }


    /* get the number of input components of a transform */
    public int
        getNumInComponents ()
    {
    int[]    nComps;
    
        nComps = new int [2];

        CMM.checkStatus (CMM.cmmGetNumComponents (ID, nComps));

        return nComps [0];
    }



    /* get the number of output components of a transform */
    public int
        getNumOutComponents ()
    {
    int[]    nComps;
    
        nComps = new int [2];

        CMM.checkStatus (CMM.cmmGetNumComponents (ID, nComps));

        return nComps [1];
    }



    /* color convert a BufferedImage */
    public void colorConvert (BufferedImage src, BufferedImage dest) {
        boolean srcOK, destOK;
        boolean srcPremult, destPremult;
        boolean srcHasAlpha, destHasAlpha;

        switch(src.getType()) {
        case BufferedImage.TYPE_INT_RGB:
        case BufferedImage.TYPE_INT_ARGB:
        case BufferedImage.TYPE_3BYTE_BGR:
            srcHasAlpha = false;
            srcPremult = false;
            srcOK = true;
            break;
        case BufferedImage.TYPE_4BYTE_ABGR:
            srcHasAlpha = true;
            srcPremult = false;
            srcOK = true;
            break;
        case BufferedImage.TYPE_INT_ARGB_PRE:
        case BufferedImage.TYPE_4BYTE_ABGR_PRE:
            srcHasAlpha = true;
            srcPremult = true;
            srcOK = true;
            break;
        default:
            srcHasAlpha = false;
            srcPremult = false;
            srcOK = false;
        }

        switch(dest.getType()) {
        case BufferedImage.TYPE_INT_RGB:
        case BufferedImage.TYPE_INT_ARGB:
        case BufferedImage.TYPE_3BYTE_BGR:
            destHasAlpha = false;
            destPremult = false;
            destOK = true;
            break;
        case BufferedImage.TYPE_4BYTE_ABGR:
            destHasAlpha = true;
            destPremult = false;
            destOK = true;
            break;
        case BufferedImage.TYPE_INT_ARGB_PRE:
        case BufferedImage.TYPE_4BYTE_ABGR_PRE:
            destHasAlpha = true;
            destPremult = true;
            destOK = true;
            break;
        default:
            destHasAlpha = false;
            destPremult = false;
            destOK = false;
        }

        if (srcOK && destOK) {
            int    cmmStatus;
            CMMImageLayout srcIL, destIL;
            ColorModel srcCMnp = null, destCMnp = null;

            /* UNPREMULT */
            if (srcHasAlpha && srcPremult) {
                srcCMnp = src.getColorModel().coerceData(src.getRaster(),
                                                         false);
            }
            if (srcHasAlpha && destPremult && (src != dest)) {
                /* REMIND - GROSS HACK to get a non-premultiplied dest CM */
                destCMnp = dest.getColorModel().coerceData(dest.getRaster(),
                                                           false);
            }

            srcIL = new CMMImageLayout (src);
            destIL = new CMMImageLayout (dest);
            int status = CMM.cmmColorConvert (ID, srcIL, destIL);

            /* REPREMULT */
            if (srcHasAlpha && srcPremult) {
                srcCMnp.coerceData(src.getRaster(), true);
            }
            CMM.checkStatus(status);
            if (destHasAlpha) {
                fixDestAlpha(src, dest, srcHasAlpha, destPremult, destCMnp);
            }
        }
        else {
            Raster srcRaster = src.getRaster();
            WritableRaster destRaster = dest.getRaster();
            ColorModel srcCM = src.getColorModel();
            ColorModel destCM = dest.getColorModel();
            int srcBands = srcRaster.getNumBands();
            int destBands = destRaster.getNumBands();
            ColorModel srcCMnp = null, destCMnp = null;

            srcHasAlpha = srcCM.hasAlpha();
            srcPremult = srcCM.isAlphaPremultiplied();
            destHasAlpha = destCM.hasAlpha();
            destPremult = destCM.isAlphaPremultiplied();

            if (srcHasAlpha) {
                srcBands -= 1;
            }

            if (destHasAlpha) {
                destBands -= 1;
            }

            if ((srcBands != srcCM.getNumColorComponents()) ||
                (destBands != destCM.getNumColorComponents())) {
                throw new IllegalArgumentException(
                    "Source or Destination ColorModel not supported");
            }

            Raster srcRaster1 = srcRaster;
            if (srcHasAlpha) {
                int[] bandList = new int[srcBands];
                for (int i = 0; i < srcBands; i++) {
                    bandList[i] = i;
                }
                srcRaster1 = srcRaster.createChild(srcRaster.getMinX(),
                    srcRaster.getMinY(), srcRaster.getWidth(),
                    srcRaster.getHeight(), srcRaster.getMinX(),
                    srcRaster.getMinY(), bandList);
            }

            WritableRaster destRaster1 = destRaster;
            if (destHasAlpha) {
                int[] bandList = new int[destBands];
                for (int i = 0; i < destBands; i++) {
                    bandList[i] = i;
                }
                destRaster1 = destRaster.createWritableChild(
                    destRaster.getMinX(), destRaster.getMinY(),
                    destRaster.getWidth(), destRaster.getHeight(),
                    destRaster.getMinX(), destRaster.getMinY(),
                    bandList);
            }


            /* UNPREMULT */
            if (srcHasAlpha && srcPremult) {
                srcCMnp = src.getColorModel().coerceData(src.getRaster(),
                                                         false);
            }
            if (srcHasAlpha && destPremult && (src != dest)) {
                /* REMIND - GROSS HACK to get a non-premultiplied dest CM */
                destCMnp = dest.getColorModel().coerceData(dest.getRaster(),
                                                           false);
            }

            colorConvert(srcRaster1, srcCM, destRaster1, destCM);

            /* REPREMULT */
            if (srcHasAlpha && srcPremult) {
                srcCMnp.coerceData(src.getRaster(), true);
            }
            if (destHasAlpha) {
                fixDestAlpha(src, dest, srcHasAlpha, destPremult, destCMnp);
            }
        }

    }



    private void fixDestAlpha(BufferedImage src,
                              BufferedImage dest,
                              boolean srcHasAlpha,
                              boolean destPremult,
                              ColorModel destCMnp) {
        if (srcHasAlpha && (src != dest)) {
            /* COPY SRC ALPHA TO DEST */
            /* src and dest should be same size */
            Raster srcraster = src.getRaster();
            WritableRaster destraster = dest.getRaster();
            int salphaband = srcraster.getNumBands() - 1;
            int dalphaband = destraster.getNumBands() - 1;
            int xs, ys;
            int xd = destraster.getMinX();
            int yd = destraster.getMinY();
            int xstart = srcraster.getMinX();
            int ystart = srcraster.getMinY();
            int xlimit = xstart + srcraster.getWidth();
            int ylimit = ystart + srcraster.getHeight();
            int alpha;
            int sbits =
                src.getColorModel().getComponentSize(salphaband);
            int dbits =
                dest.getColorModel().getComponentSize(dalphaband);
            int lshift = dbits - sbits;
            int rshift = -lshift;

            for (ys = ystart; ys < ylimit; ys++,yd++) {
                if (lshift > 0) {
                    for (xs = xstart; xs < xlimit; xs++,xd++) {
                        alpha = srcraster.getSample(xs, ys, salphaband);
                        alpha <<= lshift;
                        destraster.setSample(xd, yd, dalphaband, alpha);
                    }
                } else if (lshift == 0) {
                    for (xs = xstart; xs < xlimit; xs++,xd++) {
                        alpha = srcraster.getSample(xs, ys, salphaband);
                        destraster.setSample(xd, yd, dalphaband, alpha);
                    }
                } else {
                    for (xs = xstart; xs < xlimit; xs++,xd++) {
                        alpha = srcraster.getSample(xs, ys, salphaband);
                        alpha >>>= rshift;
                        destraster.setSample(xd, yd, dalphaband, alpha);
                    }
                }
            }

            if (destPremult) {
                destCMnp.coerceData(destraster, true);
            }
        } else if (!srcHasAlpha) {
            /* FILL DST ALPHA with 1.0 */
            WritableRaster destraster = dest.getRaster();
            int alphaband = destraster.getNumBands() - 1;
            int x, y;
            int xstart = destraster.getMinX();
            int ystart = destraster.getMinY();
            int xlimit = xstart + destraster.getWidth();
            int ylimit = ystart + destraster.getHeight();
            int alpha =
             (1 << dest.getColorModel().getComponentSize(alphaband))
                 - 1;

            for (y = ystart; y < ylimit; y++) {
                for (x = xstart; x < xlimit; x++) {
                    destraster.setSample(x, y, alphaband, alpha);
                }
            }
        }
    }


    /* color convert a Raster */
    public void colorConvert (Raster src, WritableRaster dest) {
        colorConvert(src, (ColorModel) null, dest, (ColorModel) null);
    }


    /* color convert a Raster using ColorModel data if available */
    private void colorConvert (Raster src, ColorModel srcCM,
                               WritableRaster dest, ColorModel destCM) {
        int sbands = src.getNumBands();
        int dbands = dest.getNumBands();
        int nlines = src.getHeight();    /* src and dest should be same size */
        int ncols = src.getWidth();
        byte[] input = new byte[ncols * sbands];
        byte[] output = new byte[ncols * dbands];
        int ys = src.getMinY();
        int yd = dest.getMinY();
        int xs, xd;
        int i, j, k, ndx;
        int sample;
        int[] sshift;
        int[] dshift;

        if (srcCM != null) {
            sshift = srcCM.getComponentSize();
        } else {
            sshift = src.getSampleModel().getSampleSize();
        }
        if (destCM != null) {
            dshift = destCM.getComponentSize();
        } else {
            dshift = dest.getSampleModel().getSampleSize();
        }

        for (i = 0; i < sbands; i++) {
            sshift[i] -= 8;
        }
        for (i = 0; i < dbands; i++) {
            dshift[i] -= 8;
        }

        for (i = 0; i < nlines; i++, ys++, yd++) {
            /* REMIND - what about signed/unsigned samples? */

            /* get src scanline */
            xs = src.getMinX();
            ndx = 0;
            for (j = 0; j < ncols; j++, xs++) {
                for (k = 0; k < sbands; k++, ndx++) {
                    sample = src.getSample(xs, ys, k);
                    int shift = sshift[k];
                    if (shift > 0) {
                        sample >>>= shift;
                    } else if (shift < 0) {
                        sample <<= (-shift);
                    }
                    input[ndx] = (byte) sample;
                }
            }

            /* convert scanline */
            colorConvert(input, output);

            /* store scanline to dest */
            xd = dest.getMinX();
            ndx = 0;
            for (j = 0; j < ncols; j++, xd++) {
                for (k = 0; k < dbands; k++, ndx++) {
                    sample = output[ndx] & 0xFF;
                    int shift = dshift[k];
                    if (shift > 0) {
                        sample <<= shift;
                    } else if (shift < 0) {
                        sample >>>= (-shift);
                    }
                    dest.setSample(xd, yd, k, sample);
                }
            }
        }
    }


    /* convert an array of colors in float format */
    /* each color is a contiguous set of array elements */
    /* the number of colors is */
    /* (size of the array) / (number of input/output components) */
    public float[] colorConvert(int ncol, float[] src, float[] dest) {
        int        i1;
        pelArrayInfo    thePelArrayInfo;
        short[]        srcShort, destShort;
        float[]        result;

        /* transform I/O correspond to arrays? */
        thePelArrayInfo = new pelArrayInfo (this, ncol, src, dest); 
        
        if (dest != null) {            /* check dest, make new if needed */
            result = dest;
        }
        else {
            result = new float [thePelArrayInfo.destSize];
        }

        srcShort = new short [thePelArrayInfo.srcSize];
        
            /* convert source into short array */
        for (i1 = 0; i1 < thePelArrayInfo.srcSize; i1++) {
            srcShort[i1] = (short) (((int) (src[i1] * 65535.0f)) & 0xffff);
        }

        destShort = colorConvert (srcShort, null);
    
            /* convert destination from short array */
        for (i1 = 0; i1 < thePelArrayInfo.destSize; i1++) {
            result[i1] = ((float) (destShort[i1] & 0xffff)) / 65535.0f;
        }
        
        return result;
    }


    /* convert an array of colors in short format */
    /* each color is a contiguous set of array elements */
    /* the number of colors is (size of the array) / (number of input/output components */
    public short[]
        colorConvert (    short[]        src,
                        short[]        dest)
    {
    pelArrayInfo    thePelArrayInfo;
    CMMImageLayout    srcIL, destIL;
    short[]        result;
    int        cmmStatus;

            /* transform I/O correspond to arrays? */
        thePelArrayInfo = new pelArrayInfo (this, src, dest); 
        
        if (dest != null) {            /* check dest, make new if needed */
            result = dest;
        }
        else {
            result = new short [thePelArrayInfo.destSize];
        }

            /* make the source image layout */
        srcIL = new CMMImageLayout (thePelArrayInfo.nPels, thePelArrayInfo.nSrc, 2);
        srcIL.chanData[0] = src;

            /* make the destination image layout */
        destIL = new CMMImageLayout (thePelArrayInfo.nPels, thePelArrayInfo.nDest, 2);
        destIL.chanData[0] = result;

        CMM.checkStatus (CMM.cmmColorConvert (ID, srcIL, destIL));

        return result;
    }



    /* convert an array of colors in byte format */
    /* each color is a contiguous set of array elements */
    /* the number of colors is (size of the array) / (number of input/output components */
    public byte[]
        colorConvert (    byte[]        src,
                        byte[]        dest)
    {
    pelArrayInfo    thePelArrayInfo;
    CMMImageLayout    srcIL, destIL;
    byte[]        result;
    int        cmmStatus;

            /* transform I/O correspond to arrays? */
        thePelArrayInfo = new pelArrayInfo (this, src, dest); 
        
        if (dest != null) {            /* check dest, make new if needed */
            result = dest;
        }
        else {
            result = new byte [thePelArrayInfo.destSize];
        }

            /* make the source image layout */
        srcIL = new CMMImageLayout (thePelArrayInfo.nPels, thePelArrayInfo.nSrc, 1);
        srcIL.chanData[0] = src;

            /* make the destination image layout */
        destIL = new CMMImageLayout (thePelArrayInfo.nPels, thePelArrayInfo.nDest, 1);
        destIL.chanData[0] = result;

        CMM.checkStatus (CMM.cmmColorConvert (ID, srcIL, destIL));

        return result;
    }
}



/* validate input and output arrays which contain pixels */
/* return component information */
class pelArrayInfo {
    int    nPels;
    int    nSrc;
    int    srcSize;
    int    nDest;
    int    destSize;
    
    pelArrayInfo (    ICC_Transform    transform,
            int            ncol,
            float[]        src,
            float[]        dest)
    {
        nSrc = transform.getNumInComponents ();
        nDest = transform.getNumOutComponents ();

        nPels = ncol;
        srcSize = nPels * nSrc;
        destSize = nPels * nDest;

        if (srcSize > src.length) {
            throw new IllegalArgumentException ("Inconsistent pel structure");
        }
        
        if (dest != null) {            /* check dest, make new if needed */
            checkDest (dest.length);
        }
    }

    
    pelArrayInfo (    ICC_Transform    transform,
            short[]        src,
            short[]        dest)
    {
        srcSize = src.length;

        initInfo (transform);
        destSize = nPels * nDest;
        
        if (dest != null) {            /* check dest, make new if needed */
            checkDest (dest.length);
        }
    }

    
    pelArrayInfo (    ICC_Transform    transform,
            byte[]        src,
            byte[]        dest)
    {
        srcSize = src.length;

        initInfo (transform);
        destSize = nPels * nDest;
        
        if (dest != null) {            /* check dest, make new if needed */
            checkDest (dest.length);
        }
    }

    
    void    initInfo (    ICC_Transform    transform)
    {
        nSrc = transform.getNumInComponents ();    /* get # of in and out components */
        nDest = transform.getNumOutComponents ();

        nPels = srcSize / nSrc;            /* calculate # of pels */

        if ((nPels * nSrc) != srcSize) {    /* only complete pels allowed */
            throw new IllegalArgumentException ("Inconsistent pel structure");
        }
    }

    
    void    checkDest (int length)
    {
        if (destSize > length) {
            throw new IllegalArgumentException ("Inconsistent pel structure");
        }
    }
    
}
