/*
 * @(#)FontDesignMetrics2D.java	1.8 98/06/29
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.font;

import java.awt.Font;
import java.awt.font.FontRenderContext;
import java.awt.geom.AffineTransform;

///*
// * This class provides a summary of the glyph measurements  for a Font
// * and a set of hints that guide their display.  It provides more metrics
// * information for the Font than the java.awt.FontMetrics class. There
// * is also some redundancy with that class.
// * <p>
// * The design metrics for a Font are obtained from Font.getDesignMetrics().
// * The FontDesignMetrics object returned will be independent of the
// * point size of the Font.
// * Most users are familiar with the idea of using <i>point size</i> to
// * specify the size of glyphs in a font. This point size defines a
// * measurement between the baseline of one line to the baseline of the
// * following line in a single spaced text document. The point size is
// * based on <i>typographic points</i>, approximately 1/72 of an inch.
// * <p>
// * The Java2D API adopts the convention that one point is equivalent
// * to one unit in user coordinates.  When using a normalized transform
// * for converting user space coordinates to device space coordinates (see
// * GraphicsConfiguration.getDefaultTransform() and
// * GraphicsConfiguration.getNormalizingTransform()), 72 user space units
// * equal 1 inch in device space.  In this case one point is 1/72 of an inch.
// * <p>
// * The FontDesignMetrics class expresses font metrics in terms of arbitrary
// * <i>typographic units</i> (not points) chosen by the font supplier
// * and used in the underlying platform font representations.  These units are
// * defined by dividing the em-square into a grid.  The em-sqaure is the
// * theoretical square whose dimensions are the full body height of the
// * font.  A typographic unit is the smallest measurable unit in the
// * em-square.  The number of units-per-em is determined by the font
// * designer.  The greater the units-per-em, the greater the precision
// * in metrics.  For example, Type 1 fonts divide the em-square into a
// * 1000 x 1000 grid, while TrueType fonts typically use a 2048 x 2048
// * grid.  The scale of these units can be obtained by calling
// * getUnitsPerEm().
// * <p>
// * Typographic units are relative -- their absolute size changes as the
// * size of the of the em-square changes.  An em-square is 9 points high
// * in a 9-point font.  Because typographic units are relative to the
// * em-square, a given location on a glyph will have the same coordinates
// * in typographic units regardless of the point size.
// * <p>
// * Converting typographic units to pixels requires computing pixels-per-em
// * (ppem).  This can be computed as:
// * <pre>
//         ppem = device_resolution * (inches-per-point) * pointSize
// * </pre>
// * where device resolution could be measured in pixels/inch and the point
// * size of a font is effectively points/em.  Using a normalized transform
// * from user space to device space (see above), results in 1/72 inch/point.
// * In this case, ppem is equal to the point size on a 72 dpi monitor, so
// * that an N point font displays N pixels high.  In general,
// * <pre>
//        pixel_units = typographic_units * (ppem / units_per_em)
// * </pre>
// * @see java.awt.Font
// * @see java.awt.GraphicsConfiguration#getDefaultTransform
// * @see java.awt.GraphicsConfiguration#getNormalizingTransform
// */

public class FontDesignMetrics2D {
    private Font  font;
    private FontRenderContext frc;
    private float ascent;
    private float descent;
    private float leading;
    private float maxAdvance;
	private float[] matrix;

  /*
   * Constructs a new FontDesignMetrics2D object for the given Font.
   * @param font a Font object.
   */

  public FontDesignMetrics2D(Font font, FontRenderContext frc,
                            float ascent, float descent,
                            float leading, float maxAdvance) {
      this.font = font;
      this.frc = frc;
      this.ascent = ascent;
      this.descent = descent;
      this.leading = leading;
      this.maxAdvance = maxAdvance;

      float ptSize = font.getSize2D();
      AffineTransform tx = frc.getTransform();
      tx.scale(ptSize, ptSize);
      matrix = new float[4];
	  matrix[0] = (float)tx.getScaleX();
	  matrix[1] = 0;
	  matrix[2] = 0;
	  matrix[3] = (float)tx.getScaleY();
  }

    public float getAdvance(char ch) {
        return NativeFontWrapper.getAdvance(font, ch, matrix,
                    frc.isAntiAliased(), frc.usesFractionalMetrics());
    }

    public float getAdvance(String str) {
        // REMIND: bogus implementation
        if (str.length() > 0) {
			int length = str.length();
            float width = 0;
            for (int i=0; i < length; i++) {
                width += getAdvance(str.charAt(i));
            }
            return width;
        } else {
            return 0;
        }
    }

    public float getAdvance(char data[], int off, int len) {
        return getAdvance(new String(data, off, len));
    }

    public float getMaxAdvance() {
	return maxAdvance;
    }

  /*
   * Returns the typographic ascent of the font. This is the maximum distance
   * glyphs in this font extend above the base line (measured in typographic
   * units).
   */
    public float getAscent() {
        return ascent;
    }

  /*
   * Returns the typographic descent of the font. This is the maximum distance
   * glyphs in this font extend below the base line.
   */
    public float getDescent() {
        return descent;
    }

    public float getLeading() {
        return leading;
    }

   /*
   * Returns the typographic ascent of the font. This is the maximum distance
   * glyphs in this font extend above the base line (measured in typographic
   * units).
   */
    public float getAscent(char c) {
        return ascent;
    }

  /*
   * Returns the typographic descent of the font. This is the maximum distance
   * glyphs in this font extend below the base line.
   */
    public float getDescent(char c) {
        return descent;
    }

    public float getLeading(char c) {
        return leading;
    }
    public float getHeight(char c) {
        return leading;
    }
    public float getHeight() {
        return leading;
    }
}







