/*
 * %W% %E%
 *
 * (C) Copyright IBM Corp. 1998, All Rights Reserved
 *
 * Portions copyright (c) 1998 Sun Microsystems, Inc.
 * All Rights Reserved.
 */

package sun.awt.font;

import java.awt.font.LineMetrics;
import java.awt.font.GraphicAttribute;

import java.awt.geom.Rectangle2D;

import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.Shape;

import java.util.Map;

public final class GraphicComponent extends TextLineComponent {

    public static final float GRAPHIC_LEADING = 2;

    private GraphicAttribute graphic;
    private int graphicCount;
    private int[] charsLtoV;  // possibly null

    // evaluated in computeVisualBounds
    private Rectangle2D.Float visualBounds = null;

    // used everywhere so we'll cache it
    private float graphicAdvance;

    // static and self-contained since it is needed before
    // GraphicComponent is fully constructed
    private static final class GraphicLineMetrics extends LineMetrics {

        private GraphicAttribute graphic;
        private int graphicCount;

        GraphicLineMetrics(GraphicAttribute graphic, int graphicCount) {

            this.graphicCount = graphicCount;
            this.graphic = graphic;
        }

        public int getNumChars() {
            return graphicCount;
        }

        public float getAscent() {
            return graphic.getAscent();
        }

        public float getDescent() {
            return graphic.getDescent();
        }

        public float getLeading() {
            return GRAPHIC_LEADING;
        }

        public float getHeight() {
            return getAscent() + getDescent() + getLeading();
        }

        public int getBaselineIndex() {
            int alignment = graphic.getAlignment();
            if (alignment < 0) {
                return Font.ROMAN_BASELINE;
            }
            else {
                return alignment;
            }
        }

        public float[] getBaselineOffsets() {
            return new float[] { 0, -7, -14 };  // or something...
        }

        public float getStrikethroughOffset() {
            return -(getAscent() / 2);
        }

        public float getStrikethroughThickness() {
            return getAscent() / 12;
        }

        public float getUnderlineOffset() {
            return getDescent() / 3;
        }

        public float getUnderlineThickness() {
            return getAscent() / 12;
        }
    }

    public GraphicComponent(GraphicAttribute graphic,
                            Map styles,
                            int[] charsLtoV, 
                            byte[] levels,
                            int start, 
                            int limit) {

        super(styles, new GraphicLineMetrics(graphic, limit-start));
        this.graphic = graphic;
        this.graphicCount = limit - start;
        if (charsLtoV != null) {
            this.charsLtoV = Bidi.getNormalizedOrder(charsLtoV, levels, start, limit);
        }
        else {
            this.charsLtoV = null;
        }
        this.graphicAdvance = graphic.getAdvance();
    }

    private void computeVisualBounds() {

        Rectangle2D bounds = graphic.getBounds();

        float width = (float) bounds.getWidth() +
                                 graphicAdvance * (graphicCount-1);

        visualBounds = new Rectangle2D.Float((float) bounds.getX(), 
                                             (float) bounds.getY(), 
                                             width, 
                                             (float) bounds.getHeight());
    }

    public float getItalicAngle() {
        
        return 0;
    }

    protected Rectangle2D handleGetVisualBounds() {

        if (visualBounds == null) {
            computeVisualBounds();
        }

        Rectangle2D rval = new Rectangle2D.Float();
        rval.setRect(visualBounds);
        return rval;
    }

    protected Shape handleGetOutline(float x, float y) {

        // should add API to GraphicAttribute for this
        return getVisualBounds();
    }

    protected void handleDraw(Graphics2D g2d, float x, float y) {

        for (int i=0; i < graphicCount; i++) {

            graphic.draw(g2d, x, y);
            x += graphicAdvance;
        }
    }

    public int getNumCharacters() {

        return graphicCount;
    }

    public float getCharX(int index) {

        int visIndex = charsLtoV==null? index : charsLtoV[index];
        return graphicAdvance * visIndex;
    }

    public float getCharY(int index) {

        return 0;
    }

    public float getCharAdvance(int index) {

        return graphicAdvance;
    }

    protected Rectangle2D handleGetCharVisualBounds(int index) {

        Rectangle2D bounds = graphic.getBounds();
        // don't modify their rectangle, just in case they don't copy
    
        Rectangle2D.Float charBounds = new Rectangle2D.Float();
        charBounds.setRect(bounds);
        charBounds.x += graphicAdvance * index;

        return charBounds;
    }

    // measures characters in context, in logical order
    public int getLineBreakIndex(int start, float width) {
        
        int index = (int) (width / graphicAdvance);
        if (index > graphicCount - start) {
            index = graphicCount - start;
        }
        return index;
    }

    // measures characters in context, in logical order
    public float getAdvanceBetween(int start, int limit) {

        return graphicAdvance * (limit - start);
    }

    public Rectangle2D getLogicalBounds() {

        float advance = graphicAdvance * graphicCount;

        if (visualBounds == null) {
            computeVisualBounds();
        }

        float left = visualBounds.x;
        float top = visualBounds.y;
        float width = advance; // ??? ignore visualBounds.width?
        float height = visualBounds.height;

	return new Rectangle2D.Float(left, top, width, height);
    }

    public String toString() {

        return "[Graphics:count=" + getNumCharacters() + "]";
    }

}
