/*
 * @(#)Shaping.java	1.5 98/09/23
 *
 * Copyright 1998-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
/*
 * @(#)Shaping.java	1.5 98/09/23
 *
 * (C) Copyright IBM Corp. 1998 - All Rights Reserved
 */

package sun.awt.font;

import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.Shape;

import java.awt.font.FontRenderContext;
import java.awt.font.GlyphMetrics;
import java.awt.font.GlyphVector;
import java.awt.font.LineMetrics;

import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;

/**
 * Utility routines for TextSourceLabel and ExtendedTextSourceLabel
 */

public class Shaping {

  // set c0 controls to FFFF
  public static void markSegment(Segment seg, boolean canMutate) {
    char[] chars = seg.text;
    int b = seg.start;
    int s = b;
    int e = s + seg.length;

    if (!canMutate) {
      while (s < e) {
	char c = chars[s];
	if (isFormatMark(c)) {
	  char[] temp = new char[seg.length];
	  System.arraycopy(chars, seg.start, temp, 0, seg.length);
	  chars = temp;
	  s -= seg.start;
	  e -= seg.start;

          // Increment s since we don't need to check this char again
	  chars[s++] = '\uffff';
	  seg.text = chars;
	  seg.start = 0;
	  break;
	}
	++s;
      }
      if (s == e) { // didn't change anything else
	return;
      }
    }

    while (s < e) {
      char c = chars[s];
      if (isFormatMark(c)) {
	chars[s] = '\uffff';
      }
      ++s;
    }
  }

  private static boolean isFormatMark(char c) {
    // this includes not only the format marks, but also tab, cr, lf, ps, and ls.
    // the compiler may not help us out here, this might be worth optimizing, if
    // everything else gets cleaned up.

    switch (c) {
    case '\u0009':
    case '\012': // aka u+000a, but compiler complains about premature eol
    case '\015': // aka u+000d, same problem
    case '\u200c':
    case '\u200d':
    case '\u200e':
    case '\u200f':
    case '\u2028':
    case '\u2029':
    case '\u202a':
    case '\u202b':
    case '\u202c':
    case '\u202d':
    case '\u202e':
    case '\u206a':
    case '\u206b':
    case '\u206c':
    case '\u206d':
    case '\u206e':
    case '\u206f':
    case '\ufeff':
      return true;

    default:
      return false;
    }
  }

  // remove 'format marks'
  // don't mutate the segment char array, if mutation is required, create a new char array
  // expend a fair amount of effort to avoid creating new array

  public static void compressSegment(Segment seg, boolean[] dropmap) {
    char[] chars = seg.text;
    int b = seg.start;
    int s = b;
    int e = s + seg.length;

    char[] nchars = null;
    int srcp = s; // will copy from srcp in chars
    int dstp = 0; // to dstp in nchars when copying is required

    while (s < e) {
      char c = chars[s];
      if (c == '\uffff' || isFormatMark(c)) {
	int len = s - srcp;
	if (dropmap != null) {
	  dropmap[s - b] = true;
	}
	while (++s < e) {
	  c = chars[s];
	  if (c == '\uffff' || isFormatMark(c)) {
	    if (dropmap != null) {
	      dropmap[s - b] = true;
	    }
	  } else {
	    break;
	  }
	}

	if (nchars == null) {
	  if (s == e) {
	    seg.length = len;
	    return;
	  } else if (len == 0) {
	    seg.length -= s - seg.start;
	    seg.start = s;
	    srcp = s++;
	    continue;
	  } else {
	    nchars = new char[seg.length];
	  }
	}

	System.arraycopy(chars, srcp, nchars, dstp, len);
	dstp += len;
	srcp = s;
	if (s == e) {
	  break;
	}
      }
      ++s;
    }

    if (nchars != null) {
      int len = s - srcp;
      if (len > 0) {
	System.arraycopy(chars, srcp, nchars, dstp, len);
	dstp += len;
      }

      seg.text = nchars;
      seg.start = 0;
      seg.length = dstp;
    }
  }

  public static String getShapedString(TextSource source) {
    Segment seg = new Segment();
    getShapedSegment(source, seg);
    compressSegment(seg, null);
    return new String(seg.text, seg.start, seg.length);
  }

  public static void getShapedSegment(TextSource source, Segment seg) {
    char[] context = source.getChars();
    int start = source.getStart();
    int length = source.getLength();

    if (source.dataIsLTR()) {
      // !!! assume no shaping or ligatures for now
      seg.set(context, start, length, false);
      markSegment(seg, false);
      return;
    }

    int contextStart = source.getContextStart();
    int contextLength = source.getContextLength();

    // get shape type of char on visual right
    // !!! assume no indic for now
    int rightType = NewArabicShaping.VALUE_NOSHAPE_NONE;
    for (int i = start - 1; i >= contextStart; --i) {
      rightType = NewArabicShaping.getShapeType(context[i]);
      if (rightType != NewArabicShaping.VALUE_TRANSPARENT) {
	break;
      }
    }

    // get shape type of char on visual left
    int leftType = NewArabicShaping.VALUE_NOSHAPE_NONE;
    for (int i = start + length, e = contextStart + contextLength; i < e; ++i) {
      leftType = NewArabicShaping.getShapeType(context[i]);
      if (leftType != NewArabicShaping.VALUE_TRANSPARENT) {
	break;
      }
    }

    char[] glyphs = new char[length];
    for (int i = length, j = start - 1; i > 0;) {
      glyphs[--i] = context[++j];
    }

    NewArabicShaping.shape(glyphs, leftType, rightType);

    ArabicLigaturizer.getLamAlefInstance().ligaturize(glyphs, 0, glyphs.length);

    // hack mirroring
    for (int i = 0; i < length; ++i) {
      char g = glyphs[i];
      if (isFormatMark(g)) {
	glyphs[i] = '\uffff';
      } else {
	glyphs[i] = getMirroredChar(g);
      }
    }

    seg.set(glyphs, 0, length, true);
  }

  // Not the same as mirroring, which is a glyph feature.  The best we can do is swap
  // with a paired character when there is one.  This won't position the character
  // properly, but it is a reasonable approximation.
  private static char[] mirrorPairs = {
    '\u0028', '\u0029', // ascii paired punctuation
    '\u003c', '\u003e',
    '\u005b', '\u005d',
    '\u007b', '\u007d',
    '\u2045', '\u2046', // math symbols (not complete)
    '\u207d', '\u207e',
    '\u208d', '\u208e',
    '\u2264', '\u2265',
    '\u3008', '\u3009', // chinese paired punctuation
    '\u300a', '\u300b',
    '\u300c', '\u300d',
    '\u300e', '\u300f',
    '\u3010', '\u3011',
    '\u3014', '\u3015',
    '\u3016', '\u3017',
    '\u3018', '\u3019',
    '\u301a', '\u301b',
  };

  public static char getMirroredChar(char c) {
    if (c <= '\u007d' || (c >= '\u2045' && c <= '\u301b')) {
      for (int i = 0; i < mirrorPairs.length; i++) {
	char mc = mirrorPairs[i];
	if (mc == c) {
	  return mirrorPairs[i + (((i & 0x1) == 0) ? 1 : -1)];
	} else if (mc > c) {
	  break;
	}
      }
    }

    return c;
  }

  static public void main(String[] args) {
    String[] tests = {
      ".,",
      ".\uffff,",
      ".\u0000,",
      ".\u001f,",
      ". ,",
      ".\u0000A,",
      ".\u0000\u0000A,",
      ".A,",
      ".A\u0000,",
      ".A\u0000\u0000,",
      ".\u0000A\u0000B\u0000C,",
      ".A\u0000B\u0000C\u0000,"
    };
    Segment seg = new Segment();

    for (int j = 0; j < tests.length; j++) {
      System.out.println("----");

      seg.set(tests[j].toCharArray(), 1, tests[j].length() - 2, true);

      boolean[] ligmap = new boolean[seg.length];
      System.out.print("in text: " + seg.text + " start: " + seg.start + " length: " + seg.length + " ");
      for (int i = 0; i < seg.length; i++) {
	System.out.print(Integer.toHexString(seg.text[seg.start + i]) + " ");
      }
      System.out.println();

      compressSegment(seg, ligmap);

      System.out.print("out text: " + seg.text + " start: " + seg.start + " length: " + seg.length + " ");
      for (int i = 0; i < seg.length; i++) {
	System.out.print(Integer.toHexString(seg.text[seg.start + i]) + " ");
      }
      System.out.println();
      for (int i = 0; i < ligmap.length; i++) {
	System.out.print(ligmap[i] + " ");
      }
      System.out.println();
    }
  }
}
