/*
 * @(#)StandardGlyphVector.java	1.17 98/09/09
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author Charlton Innovations, Inc.
 */

package sun.awt.font;

import java.awt.Graphics2D;
import java.awt.Font;
import java.awt.Polygon;        // remind - need a floating point version
import java.awt.geom.GeneralPath;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.geom.AffineTransform;
import java.text.CharacterIterator;
import java.awt.font.FontRenderContext;
import java.awt.Shape;
import java.awt.font.GlyphMetrics;
import java.awt.font.GlyphVector;
import java.awt.font.GlyphJustificationInfo;

/**
*   GlyphVector is a collection of Glyphs containing geometric information
*   for the placement the GlyphVector and of each Glyph in a transformed
*   coordinate space which should correspond to the device on which the
*   GlyphVector is ultimately to be displayed.  The GlyphVector does not
*   attempt any interpretation of the sequence of Glyphs it contains.
*   Relationships between adjacent Glyphs in sequence are solely used
*   in determination of placement of the Glyphs in the visual coordinate
*   space.
*   <p>
*   Instances of GlyphVector are created by a Font.
*   <p>
*   In a text processing application which can cache intermediate
*   representations of text, creation and subsequent caching of a GlyphVector
*   for use during rendering is the fastest method to present the visual
*   representation of characters to a user.
*   <p>
*   A GlyphVector is associated with exactly one Font, and can provide data
*   useful only in relation to the Font specified for the GlyphVector.  In
*   addition, metrics obtained from a GlyphVector are not generally
*   geometrically scaleable since the pixelization and spacing are
*   dependent on grid-fitting algorithms within a Font.  To facilitate
*   accurate measurement of a GlyphVector and its component Glyphs, you must
*   specify a scaling transform, anti-alias mode, and fractional metrics
*   mode at the time the GlyphVector is created.  (These characteristics can
*   be derived from the destination device.)
*   <p>
*   For each Glyph in the GlyphVector, you may obtain:
*       1) the position of the Glyph
*       2) the transform associated with the Glyph
*       3) the metrics of the Glyph in the context of the GlyphVector
*          (the metrics of the Glyph may be different under different
*           transforms, application specified rendering hints, and
*           the specific instance of the Glyph within the GlyphVector)
*   <p>
*   GlyphVectors are immutable.  There are no means provided to alter the
*   state of the GlyphVector.  Altering the data used to create the GlyphVector
*   will not alter the state of the GlyphVector.
*   <p>
*   Methods are provided to create new GlyphVectors which are the result of
*   editing editing operations on the GlyphVector such as Glyph insertion and
*   deletion.  These methods are most appropriate to applications which
*   are forming combinations such as ligatures from existing Glyphs or are
*   breaking such combinations into their component parts for visual
*   presentation.
*   <p>
*   Methods are provided to create new GlyphVectors which are the result of
*   specifying new positions for the Glyphs within the GlyphVector.  These
*   methods are most appropriate to applications which are performing
*   justification operations for the presentation of the Glyphs.
*   <p>
*   Methods are provided to return both the visual and logical bounds
*   of the entire GlyphVector or of individual Glyphs within the GlyphVector.
*   <p>
*   Methods are provided to return a Shape for the GlyphVector, and for
*   individual Glyphs within the GlyphVector.
*
*   @see Font
*   @see GlyphMetrics
*   @see TextLayout
*   @version 30 Mar 1998
*   @author Charlton Innovations, Inc.
*/

public class StandardGlyphVector extends GlyphVector {

    private Font font;
    private int [] glyphs;          // name used by native code
    private float [] positions;     // name used by native code
    private FontRenderContext frc;

    // transform matrices for:
    // fontSize * fontTX * frcTX
    private double [] glyphTX;

    public StandardGlyphVector(Font theFont, String unicodeStr,
                                                FontRenderContext frc)
    {
        this.font = theFont;
        this.glyphs = null;
        this.positions = null;
        this.frc = frc;
        double [] tx = getGlyphTX();
        NativeFontWrapper.populateGlyphVector(
            font, unicodeStr.toCharArray(),
            tx,
            frc.isAntiAliased(),
            frc.usesFractionalMetrics(),
            this);
    }
    public StandardGlyphVector(Font theFont, char[] unicodes,
                                                FontRenderContext frc)
    {
        this.font = theFont;
        this.glyphs = null;
        this.positions = null;
        this.frc = frc;
        double [] tx = getGlyphTX();
        NativeFontWrapper.populateGlyphVector(
            font, unicodes,
            tx,
            frc.isAntiAliased(),
            frc.usesFractionalMetrics(),
            this);
    }
    public StandardGlyphVector(Font theFont, CharacterIterator ciUnicode,
                                                FontRenderContext frc)
    {
        char [] unicodes = new char [ciUnicode.getEndIndex()
                                            - ciUnicode.getBeginIndex()];
        for(char c = ciUnicode.first();
                c != CharacterIterator.DONE;
                c = ciUnicode.next()) {
            unicodes[ciUnicode.getIndex() - ciUnicode.getBeginIndex()] = c;
        }
        this.font = theFont;
        this.glyphs = null;
        this.positions = null;
        this.frc = frc;
        double [] tx = getGlyphTX();
        NativeFontWrapper.populateGlyphVector(
            font, unicodes,
            tx,
            frc.isAntiAliased(),
            frc.usesFractionalMetrics(),
            this);
    }

    public StandardGlyphVector(Font theFont, int [] glyphCodes,
                                                FontRenderContext frc)
    {
        this.font = theFont;
        this.glyphs = glyphCodes;
        this.positions = null;
        this.frc = frc;
    }

    private double[] getGlyphTX()
    {
        if (null != glyphTX) {
            return glyphTX;
        }
        glyphTX = new double[6];
        float ptSize = font.getSize2D();
        AffineTransform tx = (AffineTransform)frc.getTransform().clone();
        tx.scale(ptSize, ptSize);
        tx.preConcatenate(font.getTransform());
    //  dest2D.draw and dest2D.fill will add the following transform...
    //    tx.preConcatenate(dest2D.transform);
        tx.getMatrix(glyphTX);
        return glyphTX;
    }

    public void performDefaultLayout()
    {
        double [] tx = getGlyphTX();
        NativeFontWrapper.layoutGlyphVector(
            font, tx,
            frc.isAntiAliased(),
            frc.usesFractionalMetrics(),
            (float)tx[4], (float)tx[5],
            this);
    }

    // methods associated with creation-time state
    /**
    *   Return the Font associated with this GlyphVector.
    *   @return Font used to create this GlyphVector
    *   @see Font
    */
    public Font getFont()
    {
        return this.font;
    }

    public FontRenderContext getFontRenderContext()
    {
        return this.frc;
    }

    // methods associated with the GlyphVector as a whole
    /**
    *   Return the number of Glyphs in this GlyphVector, you should
    *   Use this information to create arrays which are to be
    *   filled with results of other information retreival
    *   operations.
    *   @return number of Glyphs in this GlyphVector
    */
    public int getNumGlyphs()
    {
        return glyphs.length;
    }

    /**
    *   Return the glyphcode of the specified Glyph.
    *   This return value is meaningless to anything other
    *   than a Font and can be used to ask the Font
    *   about the existing of ligatures and other context
    *   sensitive information.
    *   @see Font#getLigature
    */
    public int getGlyphCode(int glyphIndex)
    {
        return glyphs[glyphIndex];
    }

    /**
    *   Return and array of glyphcodes for the the specified Glyphs.
    *   The contents of this return value are meaningless to anything other
    *   than a Font and can be used to ask the Font about the existence of
    *   ligatures and other context sensitive information.  Use this method
    *   for convenience and performance in processing of glyphcodes.
    *   If no array is passed in, one will be created for you.
    *   @see Font#getLigature
    */
    public int[] getGlyphCodes(int beginGlyphIndex, int numEntries,
                                int[] codeReturn)
    {
        int [] retval = codeReturn;
        if (retval == null) {
            retval = new int[numEntries];
        }
        for (int i=0; i < numEntries; i++) {
            retval[i] = glyphs[i + beginGlyphIndex];
        }
        return retval;
    }

    // methods associated with the GlyphVector as a whole,
    // after layout.
    /**
    *   Return the logical bounds of this GlyphVector.
    *   Use when positioning this GlyphVector in relation
    *   to visually adjacent GlyphVectors.
    */
    public Rectangle2D getLogicalBounds()
    {
        Shape theShape = getOutline();
		// REMIND: add left/right side bearing...
        return (Rectangle2D)theShape.getBounds();
    }

    /**
    *   Return the visual bounds of this GlyphVector
    *   The visual bounds is tightest Rectangle
    *   enclosing all non-background pixels in the
    *   rendered representation of the GlyphVector.
    */
    public Rectangle2D getVisualBounds()
    {
        Shape theShape = getOutline();
        return (Rectangle2D)theShape.getBounds();
    }

    /**
    *   Return a Shape whose interior corresponds to the
    *   visual representation of the GlyphVector.
    */
    public Shape getOutline() {
        return getOutline(0, 0);
    }

    /**
    *   Return a Shape whose interior corresponds to the
    *   visual representation of the GlyphVector. Offset
    *   to x, y
    */
    public Shape getOutline(float x, float y)
    {
        GeneralPath dstShape = new GeneralPath(GeneralPath.WIND_EVEN_ODD);
        double tx[] = getGlyphTX();
        dstShape = (GeneralPath)NativeFontWrapper.getGlyphVectorOutline(
                                        this, font, tx,
                                        frc.isAntiAliased(),
                                        frc.usesFractionalMetrics(),
                                        (float)tx[4]+x, (float)tx[5]+y,
                                        dstShape);
        return dstShape;
    }


    /**
    *   Return a Shape whose interior corresponds to the
    *   visual representation of the specified glyph
    *   within the GlyphVector.
    */
    public Shape getGlyphOutline(int glyphIndex)
    {
        GeneralPath dstShape = new GeneralPath(GeneralPath.WIND_EVEN_ODD);
        dstShape = (GeneralPath)NativeFontWrapper.getGlyphOutline(
                                        this, font, getGlyphTX(),
                                        frc.isAntiAliased(),
                                        frc.usesFractionalMetrics(),
                                        glyphIndex, dstShape);
        return dstShape;
    }

    /**
    *   Return the position of the specified glyph within the GlyphVector.
    *   This position corresponds to the leading edge of the baseline for
    *   the glyph.
    */
    public Point2D getGlyphPosition(int glyphIndex)
    {
        initPositions();
        return new Point2D.Float(   positions[glyphIndex * 2],
                                    positions[glyphIndex * 2 + 1]);
    }

    /**
    *   Set the position of the specified glyph within the GlyphVector.
    *   This position corresponds to the leading edge of the baseline for
    *   the glyph.
    */
    public void setGlyphPosition(int glyphIndex, Point2D position)
    {
        initPositions();
        positions[glyphIndex * 2] =     (float)position.getX();
        positions[glyphIndex * 2 + 1] = (float)position.getY();
    }

    /**
    *   Ensure that positions vector is initialized
    */
    private void initPositions()
    {
        if (positions == null) {
            performDefaultLayout();
        }
    }

    /**
    *   Return and array of glyph positions for the the specified Glyphs.
    *   The position of each glyph corresponds to the leading edge of the
    *   baseline for that glyph.
    *   Use this method for convenience and performance in processing of glyph
    *   positions.
    *   If no array is passed in, one will be created for you.
    *   Even numbered array entries beginning with position zero are the X
    *   coordinates of the glyph numbered beginGlyphIndex + position/2
    *   Odd numbered array entries beginning with position one are the Y
    *   coordinates of the glyph numbered beginGlyphIndex + (position-1)/2
    */
    public float[] getGlyphPositions(int beginGlyphIndex, int numEntries,
                                              float[] positionReturn)
    {
        initPositions();
        float [] retval = positionReturn;
        if (retval == null) {
            retval = new float[numEntries * 2];
        }
        for (int i=0; i < numEntries; i++) {
            retval[i*2] = positions[ (i + beginGlyphIndex) * 2];
            retval[i*2+1] = positions[ (i + beginGlyphIndex) * 2 + 1];
        }
        return retval;
    }

    /**
    *   Return the logical bounds of the specified Glyph within the GlyphVector.
    *   These logical bounds have a total of four edges, with two edges
    *   parallel to the baseline under the glyph's transform and the other two
    *   edges are shared with adjacent glyphs if they are present. Useful for
    *   hit-testing of the specified glyph, positioning of a caret at the
    *   leading or trailing edge of a glyph, and for drawing a highlight
    *   region around the specified glyph.
    *   @see #getGlyphVisualBounds
    */
    public Shape getGlyphLogicalBounds(int glyphIndex)
    {
        initPositions();
        Shape theShape = getGlyphOutline(glyphIndex);
        return (Shape)theShape.getBounds();
    }

    /**
    *   Return the visual bounds of the specified Glyph within the GlyphVector.
    *   These visual bounds have a total of four edges, representing the
    *   tightest polygon enclosing non-background pixels in the rendered
    *   representation of the Glyph whose edges are parallel to the edges
    *   of the logical bounds.  Useful for hit-testing of the specified glyph.
    *   @see #getGlyphLogicalBounds
    */
    public Shape getGlyphVisualBounds(int glyphIndex)
    {
        initPositions();
        Shape theShape = getGlyphOutline(glyphIndex);
        return (Shape)theShape.getBounds();
    }

    public GlyphMetrics getGlyphMetrics(int glyphIndex)
    {
	GlyphMetrics gm = new GlyphMetrics(
                                    0,
                                    new Rectangle2D.Float(),
                                    GlyphMetrics.STANDARD);
	NativeFontWrapper.getGlyphMetrics(
                            font,
                            glyphs[glyphIndex], getGlyphTX(),
                            frc.isAntiAliased(),
                            frc.usesFractionalMetrics(),
                            gm);
	return gm;
    }

    /**
    *   Get the transform of the specified glyph within the GlyphVector.
    */
    public AffineTransform getGlyphTransform(int glyphIndex) {
        return new AffineTransform();
    }

    /**
    *   Set the transform of the specified glyph within the GlyphVector.
    *   This can be used to rotate, mirror, translate and scale the
    *   glyph.  Adding a transform can result in signifant performance changes
    */
    public void setGlyphTransform(int glyphIndex, AffineTransform newTX) {
        // REMIND: finish me
    }


    public GlyphJustificationInfo getGlyphJustificationInfo(int glyphIndex) {
        // REMIND: finish me
        return null;
    }
    // general utility methods
    /**
    *   Test for complete equality.
    */
    public boolean equals(GlyphVector set)
    {
        // REMIND: finish me
        return false;
    }
}
