/*
 * @(#)StandardTextSource.java	1.2 98/09/21
 *
 * Copyright 1998-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
/*
 * @(#)StandardTextSource.java	1.2 98/09/21
 *
 * (C) Copyright IBM Corp. 1998 - All Rights Reserved
 */

package sun.awt.font;

import java.awt.Font;
import java.awt.font.FontRenderContext;
import java.awt.font.LineMetrics;

public class StandardTextSource extends TextSource {
  char[] chars;
  int start;
  int len;
  int cstart;
  int clen;
  int level; // assumed all uniform
  int linedir;
  Font font;
  FontRenderContext frc;
  LineMetrics lm;

  /**
   * Create a simple implementation of a TextSource.
   *
   * Chars is an array containing clen chars in the context, in logical order,
   * contiguously starting at cstart.  Start and len represent that portion
   * of the context representing the true source; start, like cstart, is
   * relative to the start of the character array.
   *
   * Level is the bidi level (0-15) for the entire context. Linedir is 0 if the
   * line on which the text is displayed is an ltr line, 1 if rtl. Font is the
   * font, frc is the render context, and lm is the line metrics for the entire
   * source text, but not necessarily the context.
   */
  public StandardTextSource(char[] chars, 
			    int start,
			    int len,
			    int cstart, 
			    int clen, 
			    int level, 
			    int linedir,
			    Font font, 
			    FontRenderContext frc,
			    LineMetrics lm) {
    if (chars == null) {
      throw new IllegalArgumentException("bad chars: null");
    }
    if (cstart < 0) {
      throw new IllegalArgumentException("bad cstart: " + cstart);
    }
    if (start < cstart) {
      throw new IllegalArgumentException("bad start: " + start + " for cstart: " + cstart);
    }
    if (clen < 0) {
      throw new IllegalArgumentException("bad clen: " + clen);
    }
    if (cstart + clen > chars.length) {
      throw new IllegalArgumentException("bad clen: " + clen + " cstart: " + cstart + " for array len: " + chars.length);
    }
    if (len < 0) {
      throw new IllegalArgumentException("bad len: " + len);
    }
    if ((start + len) > (cstart + clen)) {
      throw new IllegalArgumentException("bad len: " + len + " start: " + start + " for cstart: " + cstart + " clen: " + clen);
    }
    if (font == null) {
      throw new IllegalArgumentException("bad font: null");
    }
    if (frc == null) {
      throw new IllegalArgumentException("bad frc: null");
    }

    this.chars = chars;
    this.start = start;
    this.len = len;
    this.cstart = cstart;
    this.clen = clen;
    this.level = level;
    this.linedir = linedir;
    this.font = font;
    this.frc = frc;
    this.lm = lm != null ? lm : font.getLineMetrics(chars, cstart, clen, frc);
  }

  /** Create a StandardTextSource whose context is coextensive with the source. */
  public StandardTextSource(char[] chars, 
			    int start,
			    int len,
			    int level, 
			    int linedir,
			    Font font, 
			    FontRenderContext frc,
			    LineMetrics lm) {
    this(chars, start, len, start, len, level, linedir, font, frc, lm);
  }

  /** Create a StandardTextSource whose context and source are coextensive with the entire char array. */
  public StandardTextSource(char[] chars,
			    int level,
			    int linedir,
			    Font font,
			    FontRenderContext frc) {
    this(chars, 0, chars.length, 0, chars.length, level, linedir, font, frc, null);
  }

  /** Create a StandardTextSource whose context and source are all the text in the String. */
  public StandardTextSource(String str,
			    int level,
			    int linedir,
			    Font font,
			    FontRenderContext frc) {
    this(str.toCharArray(), 0, str.length(), 0, str.length(), level, linedir, font, frc, null);
  }

  // TextSource API

  public char[] getChars() {
    return chars;
  }

  public int getStart() {
    return start;
  }

  public int getLength() {
    return len;
  }

  public int getContextStart() {
    return cstart;
  }

  public int getContextLength() {
    return clen;
  }

  public boolean dataIsLTR() {
    return (level & 0x1) == 0;
  }

  public boolean lineIsLTR() {
    return (linedir & 0x1) == 0;
  }

  public int getBidiLevel() {
    return level;
  }

  public Font getFont() {
    return font;
  }

  public FontRenderContext getFRC() {
    return frc;
  }

  public LineMetrics getLineMetrics() {
    return lm;
  }

  public TextSource getSubSource(int start, int length) {
    if (start < 0 || length < 0 || (start + length) > len) {
      throw new IllegalArgumentException("bad start (" + start + ") or length (" + length + ")");
    }
    return new StandardTextSource(chars, this.start + start, length, cstart, clen, level, linedir, font, frc, lm);
  }

  public String toString() {
    StringBuffer buf = new StringBuffer(super.toString());
    buf.append("[start:");
    buf.append(start);
    buf.append(", len:" );
    buf.append(len);
    buf.append(", cstart:");
    buf.append(cstart);
    buf.append(", clen:" );
    buf.append(clen);
    buf.append(", chars:\"");
    for (int i = cstart, e = cstart + clen; i < e; ++i) {
      if (i > cstart) {
	buf.append(" ");
      }
      buf.append(Integer.toHexString(chars[i]));
    }
    buf.append("\"");
    buf.append(", level:");
    buf.append(level);
    buf.append(", linedir:");
    buf.append(linedir);
    buf.append(", font:");
    buf.append(font);
    buf.append(", frc:");
    buf.append(frc);
    buf.append(", lm:");
    buf.append(lm);
    buf.append("]");

    return buf.toString();
  }
}
