/*
 * %W% %E%
 *
 * (C) Copyright IBM Corp. 1998, All Rights Reserved
 *
 * Portions copyright (c) 1998 Sun Microsystems, Inc.
 * All Rights Reserved.
 */

package sun.awt.font;

import java.awt.font.LineMetrics;
import java.awt.font.TextAttribute;

import java.awt.geom.Area;
import java.awt.geom.Line2D;
import java.awt.geom.Rectangle2D;
import java.awt.geom.GeneralPath;

import java.awt.BasicStroke;
import java.awt.Graphics2D;
import java.awt.Color;
import java.awt.Paint;
import java.awt.Stroke;
import java.awt.Shape;

import java.util.Map;

import sun.awt.im.InputMethodHighlightMapping.InputMethodHighlightAttribute;

/**
 * This class is used by TextLine to represent runs of 
 * text or graphics with a single set of attributes.  It
 * also implements underlining, strikethrough and foreground
 * and background color.
 */
public abstract class TextLineComponent {

    private LineMetrics lineMetrics;

    // style stuff:
    private Paint fgPaint = null;
    private Paint bgPaint = null;
    private boolean swapColors = false;
    private boolean strikethrough = false;
    private Underline stdUnderline = null; // underline from TextAttribute.UNDERLINE_ON
    private Underline imUnderline = null; // input method underline

    protected TextLineComponent(Map styles, LineMetrics lm) {

        lineMetrics = lm;

        if (styles != null) {

            Object value = styles.get(TextAttribute.FOREGROUND);
            if (value != null) {
                fgPaint = (Paint) value;
            }

            value = styles.get(TextAttribute.BACKGROUND);
            if (value != null) {
                bgPaint = (Paint) value;
            }

            value = styles.get(TextAttribute.SWAP_COLORS);
            swapColors = TextAttribute.SWAP_COLORS_ON.equals(value);

            value = styles.get(TextAttribute.STRIKETHROUGH);
            strikethrough = TextAttribute.STRIKETHROUGH_ON.equals(value);

            value = styles.get(TextAttribute.UNDERLINE);
            stdUnderline = Underline.getUnderline(value);

            value = styles.get(InputMethodHighlightAttribute.INPUT_METHOD_UNDERLINE);
            imUnderline = Underline.getUnderline(value);
        }
    }

    public final LineMetrics getLineMetrics() {

        return lineMetrics;
    }

    // subclasses should draw using the current state of the Graphics2D
    protected abstract void handleDraw(Graphics2D g2d, float x, float y);

    private void drawTextAndEmbellishments(Graphics2D g2d, float x, float y) {

        handleDraw(g2d, x, y);

        if (!strikethrough && stdUnderline == null && imUnderline == null) {
            return;
        }

        Rectangle2D lb = getLogicalBounds();

        float x1 = x;
        float x2 = x1 + (float)lb.getWidth();

        if (strikethrough) {

            Stroke saveStroke = g2d.getStroke();
            float strikeY = y + lineMetrics.getStrikethroughOffset();
            g2d.setStroke(new BasicStroke(lineMetrics.getStrikethroughThickness()));
            g2d.draw(new Line2D.Float(x1, strikeY, x2, strikeY));
            g2d.setStroke(saveStroke);
        }
        
        float ulOffset = lineMetrics.getUnderlineOffset();
        float ulThickness = lineMetrics.getUnderlineThickness();

        if (stdUnderline != null) {
            stdUnderline.drawUnderline(g2d, ulThickness, x1, x2, y + ulOffset);
        }

        if (imUnderline != null) {
            imUnderline.drawUnderline(g2d, ulThickness, x1, x2, y + ulOffset);
        }
    }
    
    // might need more context for uniform underline, etc.
    public final void draw(Graphics2D g2d, float x, float y) {

        if (fgPaint == null && bgPaint == null && swapColors == false) {
            drawTextAndEmbellishments(g2d, x, y);
        }
        else {
            Paint savedPaint = g2d.getPaint();
            Paint foreground, background;

            if (swapColors) {
                foreground = bgPaint==null? Color.white : bgPaint;
                background = fgPaint==null? savedPaint : fgPaint;
            }
            else {
                foreground = fgPaint==null? savedPaint : fgPaint;
                background = bgPaint;
            }

            if (background != null) {

	        Rectangle2D bgArea = getLogicalBounds();
//		System.out.println("bgArea: " + bgArea + " x: " + x + " y: " + y);
		bgArea = new Rectangle2D.Float(x + (float)bgArea.getX(), 
					       y + (float)bgArea.getY(),
					       (float)bgArea.getWidth(),
					       (float)bgArea.getHeight());

                g2d.setPaint(background);
                g2d.fill(bgArea);
            }

            g2d.setPaint(foreground);
            drawTextAndEmbellishments(g2d, x, y);
            g2d.setPaint(savedPaint);
        }
    }

    protected abstract Rectangle2D handleGetCharVisualBounds(int index);

    /** 
     * Return the bottom of the Rectangle which encloses pixels
     * drawn by underlines.
     */
    private float getUnderlineMaxY() {

        float maxY = 0;
        if (stdUnderline != null) {

            float ulBottom = lineMetrics.getUnderlineOffset();
            ulBottom += stdUnderline.getLowerDrawLimit(lineMetrics.getUnderlineThickness());
            maxY = Math.max(maxY, ulBottom);
        }

        if (imUnderline != null) {

            float ulBottom = lineMetrics.getUnderlineOffset();
            ulBottom += imUnderline.getLowerDrawLimit(lineMetrics.getUnderlineThickness());
            maxY = Math.max(maxY, ulBottom);
        }

        return maxY;
    }

    public final Rectangle2D getCharVisualBounds(int index) {

        Rectangle2D visBounds = handleGetCharVisualBounds(index);

        if (swapColors || bgPaint != null 
                    || stdUnderline != null || imUnderline != null) {

            float minX = getCharX(index);

            float minY = 0, maxY = 0;

            if (swapColors || bgPaint != null) {
                
                minY = -lineMetrics.getAscent();
                maxY = minY + lineMetrics.getHeight();
            }

            maxY = Math.max(maxY, getUnderlineMaxY());

            visBounds.add(new Rectangle2D.Float(minX, minY, getCharAdvance(index), maxY-minY));
        }

        return visBounds;            
    }

    protected abstract Rectangle2D handleGetVisualBounds();

    public final Rectangle2D getVisualBounds() {

        Rectangle2D visBounds = handleGetVisualBounds();

        if (swapColors || bgPaint != null 
                    || stdUnderline != null || imUnderline != null) {

            float minX = 0;
	    Rectangle2D lb = getLogicalBounds();

            float minY = 0, maxY = 0;

            if (swapColors || bgPaint != null) {
                
                minY = (float)lb.getY();
                maxY = minY + (float)lb.getHeight();
            }

            maxY = Math.max(maxY, getUnderlineMaxY());
            
            visBounds.add(new Rectangle2D.Float(minX, minY, (float)lb.getWidth(), maxY-minY));
        }

        return visBounds;            
    }

    public abstract float getItalicAngle();

    protected abstract Shape handleGetOutline(float x, float y);

    public final Shape getOutline(float x, float y) {

        if (stdUnderline == null && imUnderline == null) {
            return handleGetOutline(x, y);
        }

        // NOTE:  The performace of the following code may
        // be very poor.

        float ulThickness = lineMetrics.getUnderlineThickness();

        Rectangle2D lb = getLogicalBounds();
        float x1 = x;
        float x2 = x1 + (float)lb.getWidth();

        Area area = null;

        if (stdUnderline != null) {
            Shape ul = stdUnderline.getUnderlineShape(ulThickness, x1, x2, y);
            area = new Area(ul);
        }

        if (imUnderline != null) {
            Shape ul = imUnderline.getUnderlineShape(ulThickness, x1, x2, y);
            Area ulArea = new Area(ul);
            if (area == null) {
                area = ulArea;
            }
            else {
                area.add(ulArea);
            }
        }

        // area won't be null here, since at least one underline exists.
        area.add(new Area(handleGetOutline(x, y)));

        return new GeneralPath(area);
    }

    public abstract int getNumCharacters();

    public abstract float getCharX(int index);
    public abstract float getCharY(int index);
    public abstract float getCharAdvance(int index);

    // measures characters in context, in logical order
    public abstract int getLineBreakIndex(int start, float width);

    // measures characters in context, in logical order
    public abstract float getAdvanceBetween(int start, int limit);

    public abstract Rectangle2D getLogicalBounds();
}
