/*
 * @(#)CompositionArea.java	1.18 98/09/03
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.im;

import java.awt.AWTEvent;
import java.awt.Canvas;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.FontMetrics;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.WindowEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.InputMethodEvent;
import java.awt.event.InputMethodListener;
import java.awt.font.FontRenderContext;
import java.awt.font.TextAttribute;
import java.awt.font.TextHitInfo;
import java.awt.font.TextLayout;
import java.awt.im.InputMethodRequests;
import java.text.AttributedCharacterIterator;
import java.text.AttributedCharacterIterator.Attribute;
import java.text.AttributedString;
import java.text.CharacterIterator;

/**
 * A composition area is used to display text that's being composed
 * using an input method in its own user interface environment,
 * typically in a root window.
 *
 * @version 1.18 09/03/98
 * @author JavaSoft International
 */

class CompositionArea extends Canvas implements  InputMethodListener,
						 InputMethodRequests {

    private AttributedCharacterIterator composedText;
    private TextLayout composedTextLayout;
    private TextHitInfo caret = null;
    private Frame compositionWindow;
    private Component target = null;
    private InputMethodContext inputMethodContext;
    private final static int TEXT_ORIGIN_X = 5;
    private final static int TEXT_ORIGIN_Y = 15;

    CompositionArea(InputMethodContext context) {

        compositionWindow = new Frame();

	// Set localized input window title. (Fall back to "Input Window")
	InputMethodManager inputMethodManager = InputMethodManager.getInstance();
	String windowTitle = inputMethodManager.getLocalizedString(InputMethodManager.INPUT_WINDOW_TITLE);
	if (windowTitle == null)
	    windowTitle = "Input Window";
        compositionWindow.setTitle(windowTitle);

        setSize(480, 18);
        setForeground(Color.black);
        setBackground(Color.white);
        
        // if we get the focus, we still want to let the client's
        // input context handle the event
        enableInputMethods(true);
        enableEvents(AWTEvent.KEY_EVENT_MASK);

        compositionWindow.add(this);
        compositionWindow.addWindowListener(new FrameWindowAdapter());
        addInputMethodListener(this);
        inputMethodContext = context;
	compositionWindow.enableInputMethods(false);
        compositionWindow.pack();
        Dimension windowSize = compositionWindow.getSize();
	Dimension screenSize = (getToolkit()).getScreenSize();
	compositionWindow.setLocation(screenSize.width - windowSize.width-20,
				    screenSize.height - windowSize.height-100);
	compositionWindow.setVisible(false);

    }
    
    // returns a 0-width rectangle
    private Rectangle getCaretRectangle(TextHitInfo caret) {
        int caretLocation = Math.round(composedTextLayout.getCaretInfo(caret)[0]);
        FontMetrics metrics = getGraphics().getFontMetrics();
        return new Rectangle(TEXT_ORIGIN_X + caretLocation,
                             TEXT_ORIGIN_Y - metrics.getAscent(),
                             0, metrics.getAscent() + metrics.getDescent());
    }

    public void paint(Graphics g) {
        super.paint(g);
	g.setColor(getForeground());
	composedTextLayout.draw((Graphics2D) g, TEXT_ORIGIN_X, TEXT_ORIGIN_Y);
	if (caret != null) {
	    Rectangle rectangle = getCaretRectangle(caret);
	    g.setXORMode(getBackground());
	    g.fillRect(rectangle.x, rectangle.y, 1, rectangle.height);
	    g.setPaintMode();
	}
    }
    
    public InputMethodRequests getInputMethodRequests() {
        return this;
    }
    
    public java.awt.im.InputContext getInputContext() {
        // return the client's input context so that events still get to
        // the same input method
        return inputMethodContext;
    }

    void setTarget(Component value) {
        target = value;
    }

    // workaround for the Solaris focus lost problem
    class FrameWindowAdapter extends WindowAdapter {
        public void windowActivated(WindowEvent e) {
	    requestFocus();
        }
    }

    //
    // InputMethodListener methods
    //

    private static final Attribute[] IM_ATTRIBUTES =
            { TextAttribute.INPUT_METHOD_HIGHLIGHT };

    public void inputMethodTextChanged(InputMethodEvent event) {
        AttributedCharacterIterator text = event.getText();
        int committedCharacterCount = event.getCommittedCharacterCount();

        // extract composed text and prepare it for display
        composedText = null;
        composedTextLayout = null;
        if (text == null
                || committedCharacterCount == text.getEndIndex() - text.getBeginIndex()) {
	    // there's no composed text to display, so hide the window
            compositionWindow.setVisible(false);
            caret = null;
        } else {
            // copy the composed text
            AttributedString composedTextString;
            composedTextString = new AttributedString(text,
                    text.getBeginIndex() + committedCharacterCount, // skip over committed text
                    text.getEndIndex(), IM_ATTRIBUTES);
            composedTextString.addAttribute(TextAttribute.FONT, getFont());
            composedText = composedTextString.getIterator();
            FontRenderContext context = ((Graphics2D) getGraphics()).getFontRenderContext();
            composedTextLayout = new TextLayout(composedText, context);
	    caret = event.getCaret();

	    // since we have composed text, show the window
            compositionWindow.setVisible(true);
	    paint(getGraphics());
	}

        // send any committed text to the text component
        if (committedCharacterCount > 0) {
            inputMethodContext.dispatchCommittedText(((Component) event.getSource()),
						     text, committedCharacterCount);
        }
    }

    public void caretPositionChanged(InputMethodEvent event) {
        caret = event.getCaret();
        if (compositionWindow.isVisible()) {
            paint(getGraphics());
        }
    }

    // InputMethodRequests methods
    public Rectangle getTextLocation(TextHitInfo offset) {
        Rectangle rectangle = getCaretRectangle(offset);
        Point location = getLocationOnScreen();
        rectangle.translate(location.x, location.y);
	return rectangle;
    }

    public TextHitInfo getLocationOffset(int x, int y) {
        if (composedTextLayout == null) {
	    return null;
	} else {
	    Point location = getLocationOnScreen();
	    x -= location.x + TEXT_ORIGIN_X;
	    y -= location.y + TEXT_ORIGIN_Y;
	    if (composedTextLayout.getBounds().contains(x, y)) {
                return composedTextLayout.hitTestChar(x, y);
            } else {
                return null;
            }
	}
    }

    public int getInsertPositionOffset() {
        // we don't have access to the client component's text.
	return 0;
    }

    private static final AttributedCharacterIterator EMPTY_TEXT =
            (new AttributedString("")).getIterator();

    public AttributedCharacterIterator getCommittedText(int beginIndex,
						       int endIndex,
						       Attribute[] attributes) {
        // we don't have access to the client component's text.
	return EMPTY_TEXT;
    }

    public int getCommittedTextLength() {
        // we don't have access to the client component's text.
	return 0;
    }


    public AttributedCharacterIterator cancelLatestCommittedText(Attribute[] attributes) {
        // we don't have access to the client component's text.
        return null;
    }

    public AttributedCharacterIterator getSelectedText(Attribute[] attributes) {
        // we don't have access to the client component's text.
        return EMPTY_TEXT;
    }

}
