/*
 * @(#)InputContext.java	1.18 98/08/14
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.im;

import java.awt.Component;
import java.util.Locale;
import java.awt.AWTEvent;
import java.awt.Toolkit;
import java.awt.event.FocusEvent;
import java.awt.event.InputMethodEvent;
import sun.awt.SunToolkit;
import java.lang.Character.Subset;

/**
 * This InputContext class contains parts of the implementation of
 * java.text.im.InputContext. These parts have been moved
 * here to avoid exposing protected members that are needed by the
 * subclass InputMethodContext. We'll merge the two classes back
 * together when we publish the input method engine API.
 *
 * @see java.awt.im.InputContext
 * @version 1.18 08/14/98
 * @author JavaSoft Asia/Pacific
 */

public class InputContext extends java.awt.im.InputContext {

    private InputMethod inputMethod;
    private boolean inputMethodCreationFailed;

    private Component previousClientComponent;
    private Component currentClientComponent;
    private Component compositionAreaComponent;
    private Subset[]  characterSubsets = null;
    private InputMethodDescriptor inputMethodDescriptor;

    /**
     * Constructs an InputContext.
     */
    protected InputContext() {
    }

    /**
     * @see java.awt.im.InputContext#selectInputMethod
     */
    public boolean selectInputMethod(Locale locale) {
	if (locale == null) {
	    throw new NullPointerException();
	}
        InputMethod inputMethod = getInputMethod();
        if (inputMethod == null) {
            return false;
        }
        return inputMethod.setLocale(locale);
    }

    /**
     * @see java.awt.im.InputContext#setCharacterSubsets
     */
    public void setCharacterSubsets(Subset[] subsets) {
        if (subsets == null)
	    characterSubsets = null;
	else {
	    characterSubsets = new Subset[subsets.length];
	    System.arraycopy(subsets, 0,
			     characterSubsets, 0, characterSubsets.length);
	}
	if (inputMethod != null) {
	    inputMethod.setCharacterSubsets(subsets);
	}
    }

    /**
     * @see java.awt.im.InputContext#dispatchEvent
     */
    public void dispatchEvent(AWTEvent event) {

        if (event instanceof InputMethodEvent)
            return;

        InputMethod inputMethod = getInputMethod();
        int id = event.getID();
        Component source;

	/*
	 * NOTE: When a Container is removing its Component which
	 * invokes this.removeNotify(), the Container has the global
	 * Component lock. It is possible to happen that an
	 * application thread is calling this.removeNotify() while an
	 * AWT event queue thread is dispatching a focus event via
	 * this.dispatchEvent(). If an input method uses AWT
	 * components (e.g., IIIMP status window), it causes deadlock,
	 * for example, Component.show()/hide() in this situation
	 * because hide/show tried to obtain the lock.  Therefore,
	 * it's necessary to obtain the global Component lock before
	 * activating or deactivating an input method.
	 */
        switch (id) {
	case FocusEvent.FOCUS_GAINED:
            source = (Component)event.getSource();
	    synchronized (source.getTreeLock()) {
		if (source instanceof CompositionArea) {
		    synchronized (this) {
			if (inputMethod != null) {
			    if (inputMethod instanceof InputMethodAdapter) {
				((InputMethodAdapter) inputMethod).
					setCompositionArea(source);
				((InputMethodAdapter) inputMethod).
					setClientComponent(previousClientComponent);
			    }
			    inputMethod.activate();
			}
			compositionAreaComponent = source;
		    }
		    break;
		}
		activate(source);
	    }
	    if(inputMethod != null)
		inputMethod.dispatchEvent(event);
	    break;

	case FocusEvent.FOCUS_LOST:
	    source = (Component)event.getSource();
	    if (source instanceof CompositionArea) {
		if (!((FocusEvent)event).isTemporary())
		    compositionAreaComponent = null;
		break;
	    }
	    if (inputMethod != null)
		inputMethod.dispatchEvent(event);

	    synchronized (source.getTreeLock()) {
		deactivate(source, ((FocusEvent)event).isTemporary());
	    }
	    break;

	default:
	    if(inputMethod != null)
		inputMethod.dispatchEvent(event);
        }
    }

    /**
     * Activates input context for the given component. If the input
     * context previously was active for a different component and there
     * is uncommitted text for that component, the text is committed.
     * Called by AWT when the keyboard focus moves to the component.
     *
     * @param client The client component into which text is entered.
     */
    synchronized void activate(Component client) {
        // this routine has to take care of two oddities:
        // - it's possible that bringing up the input window
        //   generates focus events that we'd rather not get,
        //   and makes the input window active, which we don't
        //   really want
        // - when switching focus between components that share
        //   an input context, we don't want to end composition
        //   until we know we're switching to a different component

        if (inputMethod != null) {
            if (previousClientComponent != null
                && previousClientComponent != client) {
	      if (inputMethod instanceof InputMethodAdapter) {
                    ((InputMethodAdapter) inputMethod).setClientComponent(previousClientComponent);
                }
                inputMethod.activate();
                endComposition();
                inputMethod.deactivate(false);
            }
        }

	currentClientComponent = client;
        previousClientComponent = null;

        if (inputMethod != null) {
	    if (inputMethod instanceof InputMethodAdapter) {
	        ((InputMethodAdapter) inputMethod).setCompositionArea(compositionAreaComponent);
                ((InputMethodAdapter) inputMethod).setClientComponent(client);
            }
            inputMethod.activate();
        }
    }
    
    /**
     * Deactivates input context for the given component. Called by AWT
     * when the keyboard focus moves away from the component.
     *
     * @param target The component into which text was entered.
     * @param isTemporary If the deactivate is caused by a temporary
     *        focus out.
     */
    synchronized void deactivate(Component client, boolean isTemporary) {

	// deactivate may be invoked after removeNotify call.
	if (currentClientComponent == null)
	    return;

        if (inputMethod != null) {
            inputMethod.deactivate(isTemporary);
        }
        previousClientComponent = client;
	currentClientComponent = null;
    }

    /**
     * Handles input method change requests from InputMethodManager
     */
    public synchronized void changeInputMethod(InputMethodDescriptor imDescriptor) {

	// If the same input method is specified, just keep it.
	if (inputMethodDescriptor != null) {
	    if (inputMethodDescriptor.equals(imDescriptor))
		return;
	}

	// Before switching input methods, clean it up.
	if (inputMethod != null) {
	    if (currentClientComponent != null) {
		endComposition();
		inputMethod.deactivate(false);
	    }

	    // TODO: There is a requirement that all input method
	    // instances be kept in the input context so that it can
	    // avoid any overhead involved with switching back. It's
	    // always destroyed and a new one is always created for
	    // JDK1.2.
	    inputMethod.dispose();
	    inputMethod = null;
	    inputMethodCreationFailed = false;
	}

	inputMethodDescriptor = imDescriptor;
	inputMethod = getInputMethodInstance(inputMethodDescriptor);

	// activate the new input method if there is the current client
	if (currentClientComponent != null) {
	    activate(currentClientComponent);
	}
    }

    /**
     * Returns the client component.
     */
    Component getClientComponent() {
        if (currentClientComponent != null)
	    return currentClientComponent;
        else 
	    return previousClientComponent;
    }

    /**
     * @see java.awt.im.InputContext#removeNotify
     */
    public synchronized void removeNotify(Component client) {
	if (client == null) {
	    throw new NullPointerException();
	}

	if (inputMethod == null)
	    return;

	// If the composition area is active for the client, send an
	// input method event with null text that will hide the active
	// input window and reset the composition area state.
	if ((compositionAreaComponent != null)
	    && ((client == previousClientComponent) || (client == currentClientComponent))) {
	    InputMethodEvent event = new InputMethodEvent(client,
					      InputMethodEvent.INPUT_METHOD_TEXT_CHANGED,
					      null,
					      0,
					      null,
					      null);
	    ((CompositionArea)compositionAreaComponent).inputMethodTextChanged(event);
	    compositionAreaComponent = null;
	}

	if (client == previousClientComponent) {
	    inputMethod.removeNotify();
	    previousClientComponent = null;
	    return;
	}

	if (client == currentClientComponent) {
	    inputMethod.removeNotify();
	    currentClientComponent = null;
	}
    }

    /**
     * @see java.awt.im.InputContext#endComposition
     */
    public synchronized void endComposition() {
        if (inputMethod != null) {
            inputMethod.endComposition();
        }
    }

    /**
     * @see java.awt.im.InputContext#dispose
     */
    public void dispose() {
        if (inputMethod != null) {
            inputMethod.dispose();
            inputMethod = null;
        }
	InputMethodManager.getInstance().remove(this);
    }

    /**
     * @see java.awt.im.InputContext#getInputMethodControlObject
     */
    public synchronized Object getInputMethodControlObject() {
        if (inputMethod != null) {
            return inputMethod.getControlObject();
        } else {
            return null;
        }
    }

    synchronized InputMethod getInputMethod() {
 
        if (inputMethod != null) {
            return inputMethod;
        }
        
        if (inputMethodCreationFailed) {
            return null;
        }

	inputMethod = getInputMethodInstance();
	return inputMethod;
    }

    /**
     * Gets an instance of the current input method given by
     * InputMethodManager as the current selection. The
     * inputMethodDescriptor field is set in this
     * method.
     *
     * @return the inputMethod value
     */
    InputMethod getInputMethodInstance() {
	InputMethodManager imm = InputMethodManager.getInstance();
	inputMethodDescriptor = imm.getCurrentInputMethod();
	return getInputMethodInstance(inputMethodDescriptor);
    }

    /**
     * Instantiates an input method described by
     * InputMethodDescriptor. setLocale() and setInputContext() of
     * InputMethod are invoked.
     *
     * The inputMethodCreationFailed field is set to true if the
     * instantiation failed.
     *
     * @return an InputMethod instance
     * @see InputMethod#setInputContext
     */
    InputMethod getInputMethodInstance(InputMethodDescriptor imDescriptor) {
	InputMethod inputMethodInstance = null;

        try {
	    switch (imDescriptor.getType()) {
	    case InputMethodDescriptor.HOST_ADAPTER:
		Toolkit toolkit = Toolkit.getDefaultToolkit();
		inputMethodInstance = ((SunToolkit) toolkit).getInputMethodAdapter();
		break;

	    case InputMethodDescriptor.IIIMP_ADAPTER:
	    case InputMethodDescriptor.JAVA_INPUT_METHOD:
		String className = imDescriptor.getClassName();
		ClassLoader cl = ClassLoader.getSystemClassLoader();
		if (cl != null) {
		    inputMethodInstance = (InputMethod) cl.loadClass(className).newInstance();
		} else {
		    inputMethodInstance = (InputMethod) Class.forName(className).newInstance();
		}

		break;
	    }
	    inputMethodInstance.setLocale(imDescriptor.getLocale());
            inputMethodInstance.setInputContext((InputMethodContext) this);
	    inputMethodInstance.setCharacterSubsets(characterSubsets);

        } catch (Exception e) {
            // there are a number of bad things that can happen while creating
            // the input method. In any case, we just continue without an
            // input method.
            inputMethodCreationFailed = true;

	    // if the instance has been created, then it means either
	    // setLocale() or setInputContext() failed.
	    if (inputMethodInstance != null)
		inputMethodInstance = null;
        }
        
        return inputMethodInstance;
    }
}
