/*
 * @(#)InputMethod.java	1.11 98/06/15
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.im;

import java.util.Locale;
import java.awt.AWTEvent;
import java.lang.Character.Subset;


/**
 * Defines the interface for an input method that supports complex text input.
 * Input methods traditionally support text input for languages that have
 * more characters than can be represented on a standard-size keyboard,
 * such as Chinese, Japanese, and Korean. However, they may also be used to
 * support phonetic text input for English or character reordering for Thai.
 * <p>
 * Subclasses of InputMethod can be loaded by the input method framework; they
 * can then be selected either through the API
 * (java.awt.im.InputContext.selectInputMethod) or the user interface (the
 * input method selection menu).
 * <p>
 * <b>Warning:</b> This class will be moved to a java.* package as part of making the
 * Input Method Engine SPI public; it may get modified at the same time.
 * Subclasses of sun.awt.im.InputMethod will not be compatible with Java
 * platform implementations that have a public Input Method Engine SPI.
 * Therefore, they can only be distributed as integrated parts of specific
 * implementations of the Java platform.
 *
 * @version 1.11 06/15/98
 * @author JavaSoft International
 */

public abstract class InputMethod {

    /**
     * The input method context, which is used to dispatch input method
     * events to the client component and to request information from
     * the client component.
     */
    protected InputMethodContext inputContext;

    /**
     * Constructs a new input method.
     */
    protected InputMethod() {
    }

    /**
     * Sets the input method context, which is used to dispatch input method
     * events to the client component and to request information from
     * the client component. This method is called once immediately
     * after instantiating an input method.
     * @param context the input method context for this input method
     */
    public void setInputContext(InputMethodContext context) {
        inputContext = context;
    }

    /**
     * Attempts to set the input locale. If the input method supports the
     * desired locale, it changes its behavior to support input for the locale and
     * returns true.
     * Otherwise, it returns false and does not change its behavior.
     * Called by {@link java.awt.im.InputContext#selectInputMethod}.
     *
     * @param locale locale to input
     * @return whether the specified locale is supported
     */
    public abstract boolean setLocale(Locale locale);

    /**
     * Sets the subsets of the Unicode character set that this input method
     * is allowed to input. Null may be passed in to indicate that all
     * characters are allowed. The initial value is null.
     * Called by {@link java.awt.im.InputContext#setCharacterSubsets}.
     *
     * @param subsets The subsets of the Unicode character set from which
     * characters may be input
     */
    public abstract void setCharacterSubsets(Subset[] subsets);

    /**
     * Dispatches the event to the input method. If input method support is
     * enabled for the focussed component, incoming events of certain types
     * are dispatched to the current input method for this component before
     * they are dispatched to the component's methods or event listeners.
     * The input method decides whether it needs to handle the event. If it
     * does, it also calls the event's <code>consume</code> method; this
     * causes the event to not get dispatched to the component's event
     * processing methods or event listeners.
     * <p>
     * Events are dispatched if they are instances of ComponentEvent or its
     * subclasses, except for instances of PaintEvent, FocusEvent or their
     * subclasses.
     * In JDK 1.2, this includes instances of the following AWT classes:
     * ComponentEvent, ContainerEvent, InputEvent, KeyEvent, MouseEvent,
     * WindowEvent; as well as a number of Swing event classes.
     * <p>
     * Called by {@link java.awt.im.InputContext#dispatchEvent}.
     *
     * @param event the event being dispatched to the input method
     */
    public abstract void dispatchEvent(AWTEvent event);

    /**
     * Activates the input method for immediate input processing. This
     * method is called when a client component receives a FOCUS_GAINED
     * event.
     * Called by {@link java.awt.im.InputContext#dispatchEvent}.
     */
    public abstract void activate();

    /**
     * Deactivates the input method. This method is called when a client
     * component receives a FOCUS_LOST event. The isTemporary argument
     * has the same meaning as in {@link java.awt.event.FocusEvent#isTemporary}.
     * Called by {@link java.awt.im.InputContext#dispatchEvent}.
     *
     * @param isTemporary whether the focus change is temporary
     */
    public abstract void deactivate(boolean isTemporary);
  
    /**
     * Notifies the input context that a client component has been
     * removed from its containment hierarchy, or that input method
     * support has been disabled for the component.
     * Called by {@link java.awt.im.InputContext#removeNotify}.
     */
    public abstract void removeNotify();

    /**
     * Ends any input composition that may currently be going on in this
     * context. Depending on the platform and possibly user preferences,
     * this may commit or delete uncommitted text. Any changes to the text
     * are communicated to the active component using an input method event.
     *
     * <p>
     * A text editing component may call this in a variety of situations,
     * for example, when the user moves the insertion point within the text
     * (but outside the composed text), or when the component's text is
     * saved to a file or copied to the clipboard.
     * Called by {@link java.awt.im.InputContext#endComposition}.
     */
    public abstract void endComposition();

    /**
     * Disposes of the input method and releases the resources used by it.
     * Called by {@link java.awt.im.InputContext#dispose}.
     */
    public abstract void dispose();

    /**
     * Returns a control object from this input method, or null. A
     * control object provides methods that control the behavior of the
     * input method or obtain information from the input method. The type
     * of the object is an input method specific class. Clients have to
     * compare the result against known input method control object
     * classes and cast to the appropriate class to invoke the methods
     * provided.
     * Called by {@link java.awt.im.InputContext#getInputMethodControlObject}.
     *
     * @return a control object from this input method, or null
     */
    public abstract Object getControlObject();

}
