/*
 * @(#)InputMethodContext.java	1.22 98/06/15
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.im;

import java.awt.Component;
import java.awt.AWTEvent;
import java.awt.Dimension;
import java.awt.Rectangle;
import java.awt.Window;
import java.awt.event.KeyEvent;
import java.awt.event.InputMethodEvent;
import java.awt.font.TextHitInfo;
import java.awt.im.InputMethodRequests;
import java.text.AttributedCharacterIterator;
import java.text.AttributedCharacterIterator.Attribute;
import java.text.CharacterIterator;
import sun.awt.im.CompositionArea;

/**
 * The InputMethodContext class handles communication between Component and
 * InputMethod. This class provides interfaces for use by input methods.
 * It is a subclass of InputContext, which provides interfaces for use by
 * components.
 * <p>
 * <b>Warning:</b> This class will be moved a java.* package as part of making the
 * Input Method Engine SPI public; it may get modified at the same time.
 * Clients of sun.awt.im.InputMethodContext will not be compatible with Java
 * platform implementations that have a public Input Method Engine SPI.
 * Therefore, they can only be distributed as integrated parts of specific
 * implementations of the Java platform.
 *
 * @version 1.22 06/15/98
 * @author JavaSoft International
 */

public class InputMethodContext
       extends sun.awt.im.InputContext
       implements InputMethodRequests {

    private boolean dispatchingCommittedText;

    /**
     * Constructs an InputMethodContext.
     */
    // Should be made protected in public API.
    // Clients should use InputContext.getInputContext to get instance.
    public InputMethodContext() {
        super();
	InputMethodManager.getInstance().add(this);
    }

    private boolean haveActiveClient() {
        Component client = getClientComponent();
        return client != null
               && client.getInputMethodRequests() != null;
    }

    /**
     * Creates an input method event from the arguments given
     * and dispatches it to the client component. For arguments,
     * see {@link java.awt.event.InputMethodEvent#InputMethodEvent}.
     */
    public void dispatchInputMethodEvent(int id,
                AttributedCharacterIterator text, int committedCharacterCount,
                TextHitInfo caret, TextHitInfo visiblePosition) {
        // We need to record the client component as the source so
        // that we have correct information if we later have to break up this
        // event into key events.
        // If AWTEvent gets a setSource method, change this back to a
        // simple dispatchEvent(AWTEvent) method and let the input method
        // create the event.
        Component source;
        Component handler;

        source = getClientComponent();
        if (haveActiveClient()) {
            handler = source;
        } else {
	    handler = getInputArea();
	}
        InputMethodEvent event = new InputMethodEvent(source,
                id, text, committedCharacterCount, caret, visiblePosition);
        handler.dispatchEvent(event);
    }

    /**
     * Dispatches committed text to a client component.
     * Called by composition window.
     *
     * @param client The component that the text should get dispatched to.
     * @param text The iterator providing access to the committed
     *        (and possible composed) text.
     * @param committedCharacterCount The number of committed characters in the text.
     */
    synchronized void dispatchCommittedText(Component client,
		 AttributedCharacterIterator text,
                 int committedCharacterCount) {
	// note that the client is not always the current client component -
	// some host input method adapters may dispatch input method events
	// through the Java event queue, and we may have switched clients while
        // the event was in the queue.
        if (committedCharacterCount == 0
                || text.getEndIndex() <= text.getBeginIndex()) {
            return;
        }
        long time = System.currentTimeMillis();
        dispatchingCommittedText = true;
        try {
            char keyChar = text.first();
            while (committedCharacterCount-- > 0 && keyChar != CharacterIterator.DONE) {
                KeyEvent keyEvent = new KeyEvent(client, KeyEvent.KEY_TYPED,
                                                 time, 0, KeyEvent.VK_UNDEFINED, keyChar);
                client.dispatchEvent(keyEvent);
                keyChar = text.next();
            }
        } finally {
            dispatchingCommittedText = false;
        }
    }

    // there was talk about a @dontDocument tag for javadoc. This would be a good place to use it.
    public void dispatchEvent(AWTEvent event) {
        // some host input method adapters may dispatch input method events
        // through the Java event queue. If the component that the event is
        // intended for isn't an active client, we need to dispatch this event
        // to the input window. Note that that component is not necessarily the
        // current client component, since we may have switched clients while
        // the event was in the queue.
	if (event instanceof InputMethodEvent) {
	    if (((Component) event.getSource()).getInputMethodRequests() == null) {
		getInputArea().dispatchEvent(event);
	    }
	} else {
	    // make sure we don't dispatch our own key events back to the input method
	    if (!dispatchingCommittedText) {
		super.dispatchEvent(event);
	    }
	}
    }

    private CompositionArea inputArea = null;

    private synchronized Component getInputArea() {
	if (inputArea == null) {
	    inputArea = new CompositionArea(this);
	}
        inputArea.setTarget(getClientComponent());

        return inputArea;
    }

    /**
     * Calls the current client component's implementation of getTextLocation.
     */
    public Rectangle getTextLocation(TextHitInfo offset) {
	return getReq().getTextLocation(offset);
    }

    /**
     * Calls the current client component's implementation of getLocationOffset.
     */
    public TextHitInfo getLocationOffset(int x, int y) {
	return getReq().getLocationOffset(x, y);
    }

    /**
     * Calls the current client component's implementation of getInsertPositionOffset.
     */
    public int getInsertPositionOffset() {
	return getReq().getInsertPositionOffset();
    }

    /**
     * Calls the current client component's implementation of getCommittedText.
     */
    public AttributedCharacterIterator getCommittedText(int beginIndex,
						       int endIndex,
						       Attribute[] attributes) {
	return getReq().getCommittedText(beginIndex, endIndex, attributes);
    }

    /**
     * Calls the current client component's implementation of getCommittedTextLength.
     */
    public int getCommittedTextLength() {
	return getReq().getCommittedTextLength();
    }


    /**
     * Calls the current client component's implementation of cancelLatestCommittedText.
     */
    public AttributedCharacterIterator cancelLatestCommittedText(Attribute[] attributes) {
	return getReq().cancelLatestCommittedText(attributes);
    }

    /**
     * Calls the current client component's implementation of getSelectedText.
     */
    public AttributedCharacterIterator getSelectedText(Attribute[] attributes) {
	return getReq().getSelectedText(attributes);
    }

    private InputMethodRequests getReq() {
        // use parent's isVisible since that's the one that we toggle
        // when showing or hiding the input window
	if (inputArea != null && inputArea.getParent().isVisible())
	    return inputArea.getInputMethodRequests();
	return getClientComponent().getInputMethodRequests();
    }
    
}
