/*
 * @(#)InputMethodManager.java	1.15 98/08/14
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.im;

import java.awt.CheckboxMenuItem;
import java.awt.Component;
import java.awt.Dialog;
import java.awt.Frame;
import java.awt.PopupMenu;
import java.awt.Menu;
import java.awt.MenuItem;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.security.AccessController;
import java.security.PrivilegedExceptionAction;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Locale;
import java.util.ResourceBundle;
import java.util.StringTokenizer;
import java.util.Vector;
import sun.awt.im.resources.InputMethodManagerResources;

/**
 * <code>InputMethodManager</code> is an abstract class that manages the input
 * method environment of JVM. There is only one <code>InputMethodManager</code>
 * instance in JVM that is executed under a separate daemon thread.
 * <code>InputMethodManager</code> performs the following:
 * <UL>
 * <LI>
 * Manages the list of input contexts registered by calling the <code>add</code>
 * method.</LI>
 * 
 * <LI>
 * Provides a user interface to switch input methods and broadcasts input
 * method switch requests to input contexts.</LI>
 * 
 * <LI>
 * Provides information on the current input method selection to an input
 * context for setting up an input method (initialization).</LI>
 * </UL>
 * The mechanism for supporting input method switch is as follows. (Note that
 * this may change in future releases.)
 * <UL>
 * <LI>
 * It uses platform-dependent window manager's menu (known as the <I>Window
 * menu </I>in Motif and the <I>System menu</I> or <I>Control menu</I> in
 * Win32) on each window which is popped up by clicking the left top box of
 * a window (known as <I>Window menu button</I> in Motif and <I>System menu
 * button</I> in Win32). This happens to be common in both Motif and Win32.</LI>
 * 
 * <LI>
 * When there is the IIIMP adapter presented in the environment, a menu item
 * is added to the window (manager) menu. This item label is obtained invoking
 * <code>getTriggerMenuString()</code>. If null is returned by this method, it
 * means that there is no IIIMP adapter in the environment. Frame and Dialog
 * invokes this method.</LI>
 * 
 * <LI>
 * This menu item means a trigger switch to the user to pop up a selection
 * menu.</LI>
 * 
 * <LI>
 * When the menu item of the window (manager) menu has been selected by the
 * user, Frame/Dialog invokes <code>notifyChangeRequest()</code> to notify
 * <code>InputMethodManager</code> that the user wants to switch input methods.</LI>
 * 
 * <LI>
 * <code>InputMethodManager</code> displays a pop-up menu to choose an input method.</LI>
 * 
 * <LI>
 * <code>InputMethodManager</code> broadcasts this input method change to all
 * <code>InputContext</code>'s registered in <code>InputMethodManager</code>.</LI>
 * </UL>
 * 
 * @version 1.15 08/14/98
 * @see sun.awt.im.InputContext
 * @see sun.awt.im.InputMethod
 * @see sun.awt.im.InputMethodAdapter
 * @author JavaSoft International
 */

public abstract class InputMethodManager {

    /**
     * Input Window title key for localization
     */
    final static String INPUT_WINDOW_TITLE = "CompositionArea.inputWindowTitle";

    /**
     * InputMethodManager thread name
     */
    private static final String threadName = "AWT-InputMethodManager";

    /**
     * Object for global locking
     */
    private static final Object LOCK = new Object();

    /**
     * The InputMethodManager instance
     */
    private static InputMethodManager inputMethodManager;

    /**
     * Returns the instance of InputMethodManager. This method creates
     * the instance that is unique in the Java VM if it has not been
     * created yet.
     *
     * @return the InputMethodManager instance
     */
    public static final InputMethodManager getInstance() {
	synchronized(LOCK) {
	    if (inputMethodManager == null) {
		ExecutableInputMethodManager imm = new ExecutableInputMethodManager();

		// Initialize the input method manager and start a
		// daemon thread if the user seems to use
		// IIIMP. Otherwise, just keep the instance.
		if (ExecutableInputMethodManager.canBeIIIMPUser()) {
		    imm.initialize();
		    Thread immThread = new Thread(imm, threadName);
		    immThread.setDaemon(true);
		    immThread.start();
		}
		inputMethodManager = imm;
	    }
	}
	return inputMethodManager;
    }

    /**
     * Gets a string for the trigger menu item that should be added to
     * the window manager menu. If no need to display the trigger menu
     * item, null is returned.
     */
    public abstract String getTriggerMenuString();

    /**
     * Notifies InputMethodManager that input method change has been
     * requested by the user. This notification triggers a popup menu
     * for user selection.
     *
     * @param comp Component that has accepted the change
     * request. This component has to be a Frame or Dialog.
     */
    public abstract void notifyChangeRequest(Component comp);

    /**
     * Gets the current selection of the input method. For JDK1.2, the
     * host adapter is the initial value.
     */
    public abstract InputMethodDescriptor getCurrentInputMethod();

    /**
     * Returns a localized string for the given key string. It returns
     * null if there is no localized string found.
     *
     * @param key Key to get localized string
     */
    abstract String getLocalizedString(String key);

    /**
     * Adds the specified input context to the list to be under
     * InputMethodManager's control. All input contexts managed by
     * InputMethodManager will be notified of input method changes.
     */
    abstract void add(InputContext inputContext);

    /**
     * Removes the specified input context from InputMethodManager so
     * that it will no longer be notified of input method changes.
     */
    abstract void remove(InputContext inputContext);

}

/**
 * <code>ExecutableInputMethodManager</code> is the implementation of the
 * <code>InputMethodManager</code> class. It is runnable as a separate
 * thread in the AWT environment.&nbsp;
 * <code>InputMethodManager.getInstance()</code> creates an instance of
 * <code>ExecutableInputMethodManager</code> and executes it as a deamon
 * thread.
 *
 * @see InputMethodManager
 */
class ExecutableInputMethodManager extends InputMethodManager
				   implements Runnable, ActionListener
{
    // File and class names for test input method
    private static final String TEST_INPUT_METHOD_FILE_NAME = "testinputmethod.jar";
    private static final String TEST_INPUT_METHOD_CLASS_NAME = "TestInputMethod";

    // Method name for getting supported locales
    private static String METHOD_NAME_getAvailableLocales = "getAvailableLocales";

    // IIIMP adapter class name (Fixed name for JDK1.2)
    static final String CLASS_NAME_IIIMP_ADAPTER = "sun.awt.im.iiimp.IIIMPAdapter";

    // IIIMP adapter class object if the class is available
    private Class iiimpAdapterClass;

    // IIIMP supported locale list
    private Locale[] iiimpLocaleList;

    // List of input context objects
    private Vector inputContextList = new Vector();

    // Menu item string for the trigger menu.
    private String triggerMenuString;

    // Popup menu for selecting an input method
    private PopupMenu selectionMenu;
    // Menu title and key
    private static String  switchIMMenu = "Select Input Method";
    private final static String  SWITCH_IM_MENU = "InputMethodManager.switchIMMenu";

    // current input method name
    private String currentSelection;

    // descriptor of current input method
    private InputMethodDescriptor currentInputMethod;

    // Host adapter name and key
    private static String hostAdapterName = "System Input Method";
    private final static String HOST_ADAPTER_NAME = "InputMethodManager.hostAdapterName";

    // IIIMP input method system name and key
    private static String IIIMPName = "Network Input Method";
    public final static String IIIMP_NAME = "InputMethodManager.IIIMPName";

    // descriptor of host adapter
    private InputMethodDescriptor hostAdapterDescriptor;

    // component that is requesting input method switch
    // must be Frame or Dialog
    private Component requestComponent;

    // basename for the resource bundle
    private InputMethodManagerResources resources;

    ExecutableInputMethodManager() {
	// Set up test input method if there is one
	String extensionsDirectories = (String)
	java.security.AccessController.doPrivileged
		(new sun.security.action.GetPropertyAction
		("java.ext.dirs", null));
	boolean haveTestInputMethodFile = false;
	if (extensionsDirectories != null) {
	    StringTokenizer tokenizer = 
		new StringTokenizer(extensionsDirectories, File.pathSeparator);
	    while (tokenizer.hasMoreTokens()) {
		String directory = tokenizer.nextToken();
	        final File file = new File
		    (directory + File.separator + TEST_INPUT_METHOD_FILE_NAME);
		Boolean canRead = (Boolean)
		    java.security.AccessController.doPrivileged
		    (new java.security.PrivilegedAction() {
			public Object run() {
			    return new Boolean(file.canRead());
			}
		    });
		if (canRead.booleanValue()) {
	            haveTestInputMethodFile = true;
	            break;
	        }
	    }
	}
        if (haveTestInputMethodFile) {
	    currentInputMethod = new InputMethodDescriptor(
						InputMethodDescriptor.JAVA_INPUT_METHOD,
						Locale.getDefault(),
						TEST_INPUT_METHOD_CLASS_NAME);
	    return;
	}

	// Set up host adapter
	// TODO: We should get the host adapter class name from the
	// toolkit so that all input methods can be instantiated in a
	// consistent way.
	hostAdapterDescriptor = new InputMethodDescriptor(InputMethodDescriptor.HOST_ADAPTER);

	// For JDK1.2, host adapter is always the initial setting for
	// the current input method
	currentInputMethod = hostAdapterDescriptor;
    }

    synchronized void initialize() {
	try {
	    resources = getResources();

	    // get translated strings
	    hostAdapterName = resources.getString(HOST_ADAPTER_NAME);
	    IIIMPName = resources.getString(IIIMP_NAME);
	    switchIMMenu = resources.getString(SWITCH_IM_MENU);
	} catch (Exception e) {
	    // resources should always have been set by getBundle().
	}

	currentSelection = hostAdapterName;

	try {
	    iiimpAdapterClass = (Class) AccessController.doPrivileged(
						     new PrivilegedExceptionAction() {
		public Object run() throws ClassNotFoundException {
		    ClassLoader cl = ClassLoader.getSystemClassLoader();
		    if (cl == null)
			return Class.forName(CLASS_NAME_IIIMP_ADAPTER);
		    return cl.loadClass(CLASS_NAME_IIIMP_ADAPTER);
		}
	    });
	    iiimpLocaleList = getSupportedLocales(iiimpAdapterClass);
	    triggerMenuString = switchIMMenu;
	} catch (Exception e) {
	    iiimpAdapterClass = null;
	}
    }

    public void run() {
	// If there is no IIIMP adapter, wait forever.
	while (iiimpAdapterClass == null) {
	    try {
		synchronized (this) {
		    wait();
		}
	    } catch (InterruptedException e) {
	    }
	}

	// initialize pop-up menu
	selectionMenu = new PopupMenu(triggerMenuString);

	// Loop for processing input method change requests
	while (true) {
	    waitForChangeRequest();
	    try {
		iiimpLocaleList = getSupportedLocales(iiimpAdapterClass);
	    } catch (IOException e) {
		iiimpLocaleList = null;
	    }
	    getInputMethodSelected();
	}
    }

    public synchronized InputMethodDescriptor getCurrentInputMethod() {
	return currentInputMethod;
    }

    void add(InputContext inputContext) {
	if (inputContextList.contains(inputContext))
	    return;
	inputContextList.addElement(inputContext);
    }

    void remove(InputContext inputContext) {
	inputContextList.removeElement(inputContext);
    }

    String getLocalizedString(String key) {
	String localizedString = null;
	try {
	    synchronized (this) {
		if (resources == null)
		    resources = getResources();
	    }
	    localizedString = resources.getString(key);
	} catch (Exception e) {
	}
	return localizedString;
    }

    public synchronized void notifyChangeRequest(Component comp) {
	if (!(comp instanceof Frame || comp instanceof Dialog))
	    return;

	// if busy with the current request, ignore this request.
	if (requestComponent != null)
	    return;

	requestComponent = comp;
	notify();
    }

    public String getTriggerMenuString() {
	return triggerMenuString;
    }

    // ActionListener implementation 
    public synchronized void actionPerformed(ActionEvent event) {
	String choice = event.getActionCommand();
	broadcastInputMethodChange(choice);
    }

    /*
     * Returns true if the environment indicates it's a IIIMP user.
     */
    static boolean canBeIIIMPUser() {

	Boolean result = (Boolean)
	    java.security.AccessController.doPrivileged(
					new java.security.PrivilegedAction() {
		public Object run() {
		    File f = null;
		    String userHome = System.getProperty("user.home");
		    String javaHome = System.getProperty("java.home");

		    // First, check if lib/iiimp.properties exists.
		    f = new File(javaHome + File.separator +
				 "lib" + File.separator + "iiimp.properties");
		    if (f != null && f.canRead())
			return new Boolean(true);

		    // Check if ~/.iiimp exists.
		    f = new File(userHome + File.separator + ".iiimp");
		    return new Boolean(f != null && f.canRead());
		}
	    });
	return result.booleanValue();
    }

    /**
     * Initializes resource bundle.
     */
    private InputMethodManagerResources getResources() throws Exception {
	InputMethodManagerResources resources;
	String rbBaseName = getClass().getPackage().getName()
			    + ".resources.InputMethodManagerResources";

	try {
	    resources = (InputMethodManagerResources) ResourceBundle.getBundle(rbBaseName);
	} catch (Exception e) {
	    // resources should always have been set by getBundle().
	    throw e;
	}
	return resources;
    }

    private synchronized void waitForChangeRequest() {
	try {
	    while (requestComponent == null) {
		wait();
	    }
	} catch (InterruptedException e) {
	}
    }

    private Locale[] getSupportedLocales(Class classObj) throws IOException {
	Method method;
	Locale[] locales;
	Hashtable localeList;

	try {
	    method = classObj.getMethod(METHOD_NAME_getAvailableLocales, null);
	} catch (Exception e) {
	    // in case we can't find the method, throw an IOException.
	    throw new IOException(METHOD_NAME_getAvailableLocales
				  + resources.getString("InputMethodManager.error.NotFound"));
	}

	try {
	    locales = (Locale[]) method.invoke(null, new Object[0]);
	} catch (Exception e) {
	    if (e instanceof InvocationTargetException) {
		Throwable realException = ((InvocationTargetException)e).getTargetException();
		if (realException instanceof IOException)
		    throw (IOException) realException;
	    }
	    // for other exceptions, throw an IOException.
	    throw new IOException(METHOD_NAME_getAvailableLocales
				  + resources.getString("InputMethodManager.error.InvocationError"));
	}

	if (locales == null)
	    locales = new Locale[0];
	return locales;
    }

    private void getInputMethodSelected() {
	MenuItem menuItem;

	if (iiimpLocaleList == null) {
	    requestComponent = null;
	    return;
	}

	// unavailable locales.
	selectionMenu.removeAll();
	// TODO: remove this workaround when 4141034 has been fixed.
        selectionMenu.removeNotify();

	// Add menu item for host adapter
	if (currentSelection.equals(hostAdapterName))
	    menuItem = new CheckboxMenuItem(hostAdapterName, true);
	else
	    menuItem = new MenuItem(hostAdapterName);
	menuItem.setActionCommand(hostAdapterName);
	menuItem.addActionListener(this);
	selectionMenu.add(menuItem);

	selectionMenu.addSeparator();

	// If there is no locale in the list, this means that the
	// IIIMP server connection has been disconnected some time
	// after once the connection was established.
	if (iiimpLocaleList.length == 0) {
	    menuItem = new MenuItem(IIIMPName);
	    menuItem.setEnabled(false);
	    selectionMenu.add(menuItem);
	} else {
	    // Create a cascaded menu for IIIMP input methods if there
	    // is more than one locales. Otherwise, add a single menu
	    // item.
	    Menu menu = selectionMenu;
	    if (iiimpLocaleList.length > 1)
		menu = new Menu(IIIMPName);
	    for (int i = 0; i < iiimpLocaleList.length; i++) {
		Locale locale = iiimpLocaleList[i];
		String localeString = locale.toString();
		String localeName = resources.getLocaleName(locale);
		String menuItemString;

		if (iiimpLocaleList.length == 1)
		    menuItemString = IIIMPName + " - " + localeName;
		else
		    menuItemString = localeName;

		if (currentSelection.equals(localeString))
		    menuItem = new CheckboxMenuItem(menuItemString, true);
		else
		    menuItem = new MenuItem(menuItemString);

		menuItem.setActionCommand(localeString);
		menuItem.addActionListener(this);
		menu.add(menuItem);
	    }
	    if (menu != selectionMenu)
		selectionMenu.add(menu);
	}

	synchronized (this) {
	    requestComponent.add(selectionMenu);
	    selectionMenu.show(requestComponent, 60, 80); // TODO: get proper x, y...
	    requestComponent = null;
	}
    }

    private void broadcastInputMethodChange(String choice) {
	InputMethodDescriptor desc = null;

	if (hostAdapterName.equals(choice)) {
	    desc = hostAdapterDescriptor;
	} else {
	    for (int i = 0; i < iiimpLocaleList.length; i++) {
		if (iiimpLocaleList[i].toString().equals(choice)) {
		    desc = new InputMethodDescriptor(InputMethodDescriptor.IIIMP_ADAPTER,
						     iiimpLocaleList[i],
						     CLASS_NAME_IIIMP_ADAPTER);
		    break;
		}
	    }
	}

	if (desc == null)
	    return;

	currentSelection = choice;
	currentInputMethod = desc;

	// broadcast input method change to all input contexts
	synchronized(inputContextList) {
	    Enumeration e = inputContextList.elements();
	    while (e.hasMoreElements()) {
		InputContext inputContext = (InputContext) e.nextElement();
		inputContext.changeInputMethod(desc);
	    }
	}
    }
}
