/*
 * @(#)BytePackedRaster.java	1.26 98/08/12
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.image;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import java.awt.image.RasterFormatException;
import java.awt.image.SampleModel;
import java.awt.image.MultiPixelPackedSampleModel;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferByte;
import java.awt.Rectangle;
import java.awt.Point;

/**
 * This class is useful for describing 1, 2 and 4 bit image data elements.
 * This raster has one band whose pixels are less than 8-bits and are
 * packed together into individual bytes in a single byte array.  This type of
 * raster can be used with an IndexColorModel. This raster uses a
 * MultiPixelPackedSampleModel.
 *
 * @version 10 Feb 1997
 */
public class BytePackedRaster extends WritableRaster {

    /** The data bit offset for each pixel. */
    int           dataBitOffset;

    /** Scanline stride of the image data contained in this Raster. */
    int           scanlineStride;

    /** The bit stride of a pixel. */
    int           pixelBitStride;

    /** The total number of bits required to store a pixel. */
    int           numberOfBits;

    /** The bit mask for extracting the pixel. */
    int           bitMask;

    /** The image data array. */
    byte[]        data;

    int type;

    static private native void initIDs();
    static {
        /* ensure that the necessary native libraries are loaded */
	NativeLibLoader.loadLibraries();
        initIDs();
    }

    /**
     * Constructs a BytePackedRaster with the given SampleModel.
     * The Raster's upper left corner is origin and it is the same
     * size as the SampleModel.  A DataBuffer large enough to describe the
     * Raster is automatically created.  SampleModel must be of type
     * MultiPixelPackedSampleModel.
     * @param sampleModel     The SampleModel that specifies the layout.
     * @param origin          The Point that specified the origin.
     */
    public BytePackedRaster(SampleModel sampleModel,
			       Point origin) {
        this(sampleModel,
             sampleModel.createDataBuffer(),
             new Rectangle(origin.x,
                           origin.y,
                           sampleModel.getWidth(),
                           sampleModel.getHeight()),
             origin,
             null);
    }

    /**
     * Constructs a BytePackedRaster with the given SampleModel
     * and DataBuffer.  The Raster's upper left corner is origin and
     * it is the same size as the SampleModel.  The DataBuffer is not
     * initialized and must be a DataBufferByte compatible with SampleModel.
     * SampleModel must be of type MultiPixelPackedSampleModel.
     * @param sampleModel     The SampleModel that specifies the layout.
     * @param dataBuffer      The DataBufferShort that contains the image data.
     * @param origin          The Point that specifies the origin.
     */
    public BytePackedRaster(SampleModel sampleModel,
                               DataBuffer dataBuffer,
                               Point origin) {
        this(sampleModel,
             dataBuffer,
             new Rectangle(origin.x,
                           origin.y,
                           sampleModel.getWidth(),
                           sampleModel.getHeight()),
             origin,
             null);
    }

    /**
     * Constructs a BytePackedRaster with the given SampleModel,
     * DataBuffer, and parent.  DataBuffer must be a DataBufferByte and
     * SampleModel must be of type MultiPixelPackedSampleModel.
     * When translated into the base Raster's
     * coordinate system, aRegion must be contained by the base Raster.
     * Origin is the coordinate in the new Raster's coordinate system of
     * the origin of the base Raster.  (The base Raster is the Raster's
     * ancestor which has no parent.)
     *
     * Note that this constructor should generally be called by other
     * constructors or create methods, it should not be used directly.
     * @param sampleModel     The SampleModel that specifies the layout.
     * @param dataBuffer      The DataBufferShort that contains the image data.
     * @param aRegion         The Rectangle that specifies the image area.
     * @param origin          The Point that specifies the origin.
     * @param parent          The parent (if any) of this raster.
     */
    public BytePackedRaster(SampleModel sampleModel,
                               DataBuffer dataBuffer,
                               Rectangle aRegion,
                               Point origin,
                               BytePackedRaster parent){
        super(sampleModel,dataBuffer,aRegion,origin, parent);
        if (!(dataBuffer instanceof DataBufferByte)) {
           throw new RasterFormatException("BytePackedRasters must have" +
                "byte DataBuffers");
        }
        DataBufferByte dbi = (DataBufferByte)dataBuffer;
        this.data = dbi.getData();

        if (sampleModel instanceof MultiPixelPackedSampleModel) {
            MultiPixelPackedSampleModel mppsm =
                (MultiPixelPackedSampleModel)sampleModel;
            this.type = IntegerComponentRaster.TYPE_BYTE_BINARY_SAMPLES;
            scanlineStride = mppsm.getScanlineStride();
            pixelBitStride = mppsm.getPixelBitStride();
            numberOfBits = mppsm.getPixelBitStride();
            dataBitOffset = mppsm.getDataBitOffset();
            int xOffset = aRegion.x - origin.x;
            int yOffset = aRegion.y - origin.y;
            dataBitOffset += xOffset*pixelBitStride +
                yOffset*scanlineStride*8;
            bitMask = (1 << numberOfBits) -1;
        } else {
            throw new RasterFormatException("BytePackedRasters must have"+
                "MultiPixelPackedSampleModel");
        }
        verify(false);
    }

    /** Returns the number of bits occupied by one pixel. */
    public int getNumberOfBits() {
        return numberOfBits;
    }

    /**
     * Returns the data bit offset for the Raster.  The data
     * bit offset is the bit index into the data array element
     * corresponding to the first sample of the first scanline.
     */
    public int getDataBitOffset() {
        return dataBitOffset;
    }

    /**
     * Returns the scanline stride -- the number of data array elements between
     * a given sample and the sample in the same column
     * of the next row.
     */
    public int getScanlineStride() {
        return scanlineStride;
    }

    /**
     * Returns pixel bit stride -- the number of bits between two
     * samples on the same scanline.
     */
    public int getPixelBitStride() {
        return pixelBitStride;
    }

    /**
     * Returns a reference to the entire data array.
     */
    public byte[] getDataStorage() {
        return data;
    }

    /**
     * Returns the data element at the specified
     * location.  There will be no explicit bounds checking on the parameters.
     * An ArrayOutOfBounds exception will be thrown at runtime
     * if data outside of the array is accessed.
     * A ClassCastException will be thrown if the input object is non null
     * and references anything other than an array of transferType.
     * @param x        The X coordinate of the pixel location.
     * @param y        The Y coordinate of the pixel location.
     * @param outData  An object reference to an array of type defined by
     *                 getTransferType() and length getNumDataElements().
     *                 If null an array of appropriate type and size will be
     *                 allocated.
     * @return         An object reference to an array of type defined by
     *                 getTransferType() with the request pixel data.
     */
    public Object getDataElements(int x, int y, Object obj) {
        byte outData[];
        if (obj == null) {
            outData = new byte[numDataElements];
        } else {
            outData = (byte[])obj;
        }
	int bitnum = dataBitOffset + (x-sampleModelTranslateX) * pixelBitStride;
	int element = data[(y-sampleModelTranslateY) * scanlineStride + (bitnum / 8)];
	outData[0] = (byte)((element >> (8 - (bitnum & 7))) & bitMask);
	return outData;
    }

    /**
     * Returns an array  of data elements from the specified rectangular
     * region.
     * There will be no explicit bounds checking on the parameters.
     * An ArrayOutOfBounds exception will be thrown at runtime
     * if data outside of the array is accessed.
     * A ClassCastException will be thrown if the input object is non null
     * and references anything other than an array of transferType.
     * <pre>
     *       byte[] bandData = (byte[])raster.getPixelData(x, y, w, h, null);
     *       int pixel;
     *       // To find a data element at location (x2, y2)
     *       pixel = bandData[((y2-y)*w + (x2-x))];
     * </pre>
     * @param x        The X coordinate of the upper left pixel location.
     * @param y        The Y coordinate of the upper left pixel location.
     * @param width    Width of the pixel rectangle.
     * @param height   Height of the pixel rectangle.
     * @param outData  An object reference to an array of type defined by
     *                 getTransferType() and length w*h*getNumDataElements().
     *                 If null an array of appropriate type and size will be
     *                 allocated.
     * @return         An object reference to an array of type defined by
     *                 getTransferType() with the request pixel data.
     /
    public Object getPixelData(int x, int y, int w, int h, Object obj) {
        byte outData[];
        if (obj == null) {
            outData = new byte[numDataElements*w*h];
        } else {
            outData = (byte[])obj;
        }
	int pixbits = pixelBitStride;
	int scanbit = dataBitOffset + (x-sampleModelTranslateX) * pixbits;
	int index = (y-sampleModelTranslateY) * scanlineStride;
	int outindex = 0;
	byte data[] = this.data;
	for (int j = 0; j < h; j++) {
	    int bitnum = scanbit;
	    for (int i = 0; i < w; i++) {
		outData[outindex++] =
                    (byte)(bitMask &
                          (data[index + (bitnum / 8)] >> (8 - (bitnum & 7))));
		bitnum += pixbits;
	    }
	    index += scanlineStride;
	}
	return outData;
    }

    /**
     * Returns a byte array containing the specified data elements
     * from the data array.  The band index will be ignored.
     * There will be no explicit bounds checking on the parameters.
     * An ArrayOutOfBounds exception will be thrown at runtime
     * if data outside of the array is accessed.
     * <pre>
     *       byte[] byteData = getByteData(x, y, band, w, h, null);
     *       // To find a data element at location (x2, y2)
     *       byte element = byteData[(y2-y)*w + (x2-x)];
     * </pre>
     * @param x        The X coordinate of the upper left pixel location.
     * @param y        The Y coordinate of the upper left pixel location.
     * @param width    Width of the pixel rectangle.
     * @param height   Height of the pixel rectangle.
     * @param band     The band to return, is ignored.
     * @param outData  If non-null, data elements
     *                 at the specified locations are returned in this array.
     * @return         Byte array with data elements.
     */
    public byte[] getByteData(int x, int y, int w, int h,
                              int band, byte[] outData) {
	return getByteData(x, y, w, h, outData);
    }

    /**
     * Returns a byte array containing the specified data elements
     * from the data array.
     * There will be no explicit bounds checking on the parameters.
     * An ArrayOutOfBounds exception will be thrown at runtime
     * if data outside of the array is accessed.
     * <pre>
     *       byte[] byteData = raster.getByteData(x, y, w, h, null);
     *       byte pixel;
     *       // To find a data element at location (x2, y2)
     *       pixel = byteData[((y2-y)*w + (x2-x))];
     * </pre>
     * @param x        The X coordinate of the upper left pixel location.
     * @param y        The Y coordinate of the upper left pixel location.
     * @param width    Width of the pixel rectangle.
     * @param height   Height of the pixel rectangle.
     * @param outData  If non-null, data elements
     *                 at the specified locations are returned in this array.
     * @return         Byte array with data elements.
     */
    public byte[] getByteData(int x, int y, int w, int h, byte[] outData) {
	if (outData == null) {
	    outData = new byte[w * h];
	}
	int pixbits = pixelBitStride;
	int scanbit = dataBitOffset + (x-sampleModelTranslateX) * pixbits;
	int index = (y-sampleModelTranslateY) * scanlineStride;
	int outindex = 0;
	byte data[] = this.data;
	for (int j = 0; j < h; j++) {
	    int bitnum = scanbit;
	    for (int i = 0; i < w; i++) {
		outData[outindex++] = (byte) (bitMask &
					      (data[index + (bitnum / 8)]
					       >> (8 - (bitnum & 7))));
		bitnum += pixbits;
	    }
	    index += scanlineStride;
	}
	return outData;
    }

    /**
     * Stores the data elements at the specified location.
     * There will be no explicit bounds checking on the parameters.
     * An ArrayOutOfBounds exception will be thrown at runtime
     * if data outside of the array is accessed.
     * A ClassCastException will be thrown if the input object is non null
     * and references anything other than an array of transferType.
     * @param x        The X coordinate of the pixel location.
     * @param y        The Y coordinate of the pixel location.
     * @param inData   An object reference to an array of type defined by
     *                 getTransferType() and length getNumDataElements()
     *                 containing the pixel data to place at x,y.
     */
    public void setDataElements(int x, int y, Object obj) {
        byte inData[] = (byte[])obj;
	int bitnum = dataBitOffset + (x-sampleModelTranslateX) * pixelBitStride;
	int index = (y-sampleModelTranslateY) * scanlineStride + (bitnum / 8);
	int shift = 7 - (bitnum & 7);
	byte element = data[index];
	element &= ~(bitMask << shift);
	element |= (inData[0] & bitMask) << shift;
	data[index] = element;
//         System.out.println("bitnum="+bitnum+" index="+index+" shift="+shift+" inData="+inData[0]+" bitMask="+bitMask);
    }

    /**
     * Stores the Raster data at the specified location.
     * There will be no explicit bounds checking on the parameters.
     * An ArrayOutOfBounds exception will be thrown at runtime
     * if data outside of the array is accessed.
     * @param x          The X coordinate of the pixel location.
     * @param y          The Y coordinate of the pixel location.
     * @param inRaster   Raster of data to place at x,y location.
     */
    public void setDataElements(int x, int y, Raster inRaster) {
	int w = inRaster.getWidth();
	int h = inRaster.getHeight();
	if (w <= 0 || h <= 0) {
	    return;
	}
	if (pixelBitStride == numberOfBits &&
	    inRaster instanceof BytePackedRaster)
	{
	    // Source is bytepacked and bits are contiguous
	    BytePackedRaster inByteRaster = (BytePackedRaster) inRaster;
	    if (inByteRaster.pixelBitStride == this.pixelBitStride &&
		inByteRaster.numberOfBits == this.numberOfBits)
	    {
		// Source pixel layout is equivalent to our layout
		int outscan = this.scanlineStride;
		int outbit = (this.dataBitOffset +
			      8 * (y-sampleModelTranslateY) * outscan +
			      (x-sampleModelTranslateX) * this.pixelBitStride);

                int baseSubRasterOffsetX = inByteRaster.getMinX() -
                    inByteRaster.getSampleModelTranslateX();
                int baseSubRasterOffsetY = inByteRaster.getMinY() -
                    inByteRaster.getSampleModelTranslateY();

		int inbit = (inByteRaster.dataBitOffset +
                              (baseSubRasterOffsetY*
                               inByteRaster.scanlineStride)+
                              (baseSubRasterOffsetX*
                               inByteRaster.pixelBitStride));

		if ((inbit & 7) == (outbit & 7)) {
		    // copy is bit aligned
		    byte[] inData = inByteRaster.data;
		    byte[] outData = this.data;
		    int inscan = inByteRaster.scanlineStride;
		    int copybits = w * numberOfBits;
		    int bits = inbit & 7;
		    if (bits != 0) {
			// Copy partial bytes on left
			int inbyte = inbit / 8;
			int outbyte = outbit / 8;
			int mask = 0xff >> bits;
			bits = 8 - bits;
			if (copybits < bits) {
			    mask = mask & (mask << (8 - copybits));
			    bits = copybits;
			}
			for (int j = 0; j < h; j++) {
			    int element = outData[outbyte];
			    element &= ~mask;
			    element |= (inData[inbyte] & mask);
			    outData[outbyte] = (byte) element;
			    inbyte += inscan;
			    outbyte += outscan;
			}
			inbit += bits;
			outbit += bits;
			copybits -= bits;
		    }
		    if (copybits >= 8) {
			// Copy whole bytes
			int inbyte = inbit / 8;
			int outbyte = outbit / 8;
			int copybytes = copybits / 8;
			if (copybytes == inscan && inscan == outscan) {
			    System.arraycopy(inData, inbyte,
					     outData, outbyte,
					     inscan * inByteRaster.height);
			} else {
			    for (int j = 0; j < h; j++) {
				System.arraycopy(inData, inbyte,
						 outData, outbyte,
						 copybytes);
				inbyte += inscan;
				outbyte += outscan;
			    }
			}
			bits = copybytes * 8;
			inbit += bits;
			outbit += bits;
			copybits -= bits;
		    }
		    if (copybits > 0) {
			// Copy partial bytes on right
			int inbyte = inbit / 8;
			int outbyte = outbit / 8;
			int mask = (0xff00 >> copybits) & 0xff;
			for (int j = 0; j < h; j++) {
			    int element = outData[outbyte];
			    element &= ~mask;
			    element |= (inData[inbyte] & mask);
			    outData[outbyte] = (byte) element;
			    inbyte += inscan;
			    outbyte += outscan;
			}
		    }
		    return;
		}
		// REMIND: we could do an else{} here and copy unaligned
		// contiguous data by shifting and merging pairs of bytes...
	    }
	}
	super.setDataElements(x, y, inRaster);
    }

    /**
     * Stores an array of data elements into the specified rectangular
     * region.
     * There will be no explicit bounds checking on the parameters.
     * An ArrayOutOfBounds exception will be thrown at runtime
     * if data outside of the array is accessed.
     * A ClassCastException will be thrown if the input object is non null
     * and references anything other than an array of transferType.
     * The data elements in the
     * data array are assumed to be packed.  That is, a data element
     * at location (x2, y2) would be found at:
     * <pre>
     *      inData[((y2-y)*w + (x2-x))]
     * </pre>
     * @param x        The X coordinate of the upper left pixel location.
     * @param y        The Y coordinate of the upper left pixel location.
     * @param w        Width of the pixel rectangle.
     * @param h        Height of the pixel rectangle.
     * @param inData   An object reference to an array of type defined by
     *                 getTransferType() and length w*h*getNumDataElements()
     *                 containing the pixel data to place between x,y and
     *                 x+h, y+h.
     */
    public void setDataElements(int x, int y, int w, int h, Object obj) {
        byte inData[] = (byte[])obj;
	int pixbits = pixelBitStride;
	int scanbit = dataBitOffset + x * pixbits;
	int index = y * scanlineStride;
	int outindex = 0;
	byte data[] = this.data;
	for (int j = 0; j < h; j++) {
	    int bitnum = scanbit;
	    for (int i = 0; i < w; i++) {
		int shift = 7 - (bitnum & 7);
		byte element = data[index + (bitnum / 8)];
		element &= ~(bitMask << shift);
		element |= (inData[outindex++] & bitMask) << shift;
		data[index + (bitnum / 8)] = element;
	    }
	    index += scanlineStride;
	}
    }

    /**
     * Stores a byte array of data elements into the specified rectangular
     * region.  The band index will be ignored.
     * There will be no explicit bounds checking on the parameters.
     * An ArrayOutOfBounds exception will be thrown at runtime
     * if data outside of the array is accessed.
     * The data elements in the
     * data array are assumed to be packed.  That is, a data element
     * at location (x2, y2) would be found at:
     * <pre>
     *      inData[((y2-y)*w + (x2-x))]
     * </pre>
     * @param x        The X coordinate of the upper left pixel location.
     * @param y        The Y coordinate of the upper left pixel location.
     * @param w        Width of the pixel rectangle.
     * @param h        Height of the pixel rectangle.
     * @param band     The band to set, is ignored.
     * @param inData   The data elements to be stored.
     */
    public void putByteData(int x, int y, int w, int h,
                            int band, byte[] inData) {
	putByteData(x, y, w, h, inData);
    }

    /**
     * Stores a byte array of data elements into the specified rectangular
     * region.
     * There will be no explicit bounds checking on the parameters.
     * An ArrayOutOfBounds exception will be thrown at runtime
     * if data outside of the array is accessed.
     * The data elements in the
     * data array are assumed to be packed.  That is, a data element
     * at location (x2, y2) would be found at:
     * <pre>
     *      inData[((y2-y)*w + (x2-x))]
     * </pre>
     * @param x        The X coordinate of the upper left pixel location.
     * @param y        The Y coordinate of the upper left pixel location.
     * @param w        Width of the pixel rectangle.
     * @param h        Height of the pixel rectangle.
     * @param inData   The data elements to be stored.
     */
    public void putByteData(int x, int y, int w, int h, byte[] inData) {
	int pixbits = pixelBitStride;
	int scanbit = dataBitOffset + x * pixbits;
	int index = y * scanlineStride;
	int outindex = 0;
	byte data[] = this.data;
	for (int j = 0; j < h; j++) {
	    int bitnum = scanbit;
	    for (int i = 0; i < w; i++) {
		int shift = 7 - (bitnum & 7);
		byte element = data[index + (bitnum / 8)];
		element &= ~(bitMask << shift);
		element |= (inData[outindex++] & bitMask) << shift;
		data[index + (bitnum / 8)] = element;
	    }
	    index += scanlineStride;
	}
    }

    /**
     * Creates a subraster given a region of the raster.  The x and y
     * coordinates specify the horizontal and vertical offsets
     * from the upper-left corner of this raster to the upper-left corner
     * of the subraster.  Note that the subraster will reference the same
     * DataBuffer as the parent raster, but using different offsets. The
     * bandList is ignored.
     * @param x               X offset.
     * @param y               Y offset.
     * @param width           Width (in pixels) of the subraster.
     * @param height          Height (in pixels) of the subraster.
     * @param x0              Translated X origin of the subraster.
     * @param y0              Translated Y origin of the subraster.
     * @param bandList        Array of band indices.
     * @exception RasterFormatException
     *            if the specified bounding box is outside of the parent raster.
     */
    public Raster createChild(int x, int y,
                              int width, int height,
                              int x0, int y0, int[] bandList) {
	WritableRaster newRaster = createWritableChild(x, y,
                                                       width, height,
                                                       x0, y0,
                                                       bandList);
	return (Raster) newRaster;
    }

    /**
     * Creates a Writable subRaster given a region of the Raster. The x and y
     * coordinates specify the horizontal and vertical offsets
     * from the upper-left corner of this Raster to the upper-left corner
     * of the subRaster.  The bandList is ignored.
     * A translation to the subRaster may also be specified.
     * Note that the subRaster will reference the same
     * DataBuffer as the parent Raster, but using different offsets.
     * @param x               X offset.
     * @param y               Y offset.
     * @param width           Width (in pixels) of the subraster.
     * @param height          Height (in pixels) of the subraster.
     * @param x0              Translated X origin of the subraster.
     * @param y0              Translated Y origin of the subraster.
     * @param bandList        Array of band indices.
     * @exception RasterFormatException
     *            if the specified bounding box is outside of the parent Raster.
     */
    public WritableRaster createWritableChild(int x, int y,
                                              int width, int height,
                                              int x0, int y0,
                                              int[] bandList) {
	if (x < this.minX) {
	    throw new RasterFormatException("x lies outside the raster");
	}
	if (y < this.minY) {
	    throw new RasterFormatException("y lies outside the raster");
	}
        if (x+width > this.minX + this.width) {
	    throw new RasterFormatException("(x + width) is outside of Raster");
        }
        if (y+height > this.minY + this.height) {
	    throw new RasterFormatException("(y + height) is outside of Raster");
        }

	SampleModel sm;

	if (bandList != null) {
            sm = sampleModel.createSubsetSampleModel(bandList);
        }
	else {
	    sm = sampleModel;
        }

        int deltaX = x0 - x;
        int deltaY = y0 - y;

	return new BytePackedRaster(sm,
				    dataBuffer,
				    new Rectangle(x0, y0, width, height),
				    new Point(sampleModelTranslateX+deltaX,
					      sampleModelTranslateY+deltaY),
				    this);
    }

    /**
     * Creates a raster with the same layout but using a different
     * width and height, and with new zeroed data arrays.
     */
    public WritableRaster createCompatibleWritableRaster(int w, int h) {
        if (w <= 0 || h <=0) {
            throw new RasterFormatException("negative "+
                                          ((w <= 0) ? "width" : "height"));
        }

        SampleModel sm = sampleModel.createCompatibleSampleModel(w,h);

        return new BytePackedRaster(sm, new Point(0,0));
    }

    /**
     * Creates a raster with the same layout and the same
     * width and height, and with new zeroed data arrays.
     */
    public WritableRaster createCompatibleWritableRaster () {
        return createCompatibleWritableRaster(width,height);
    }

    /**
     * Verify that the layout parameters are consistent with
     * the data.  If strictCheck
     * is false, this method will check for ArrayOutOfBounds conditions.  If
     * strictCheck is true, this method will check for additional error
     * conditions such as line wraparound (width of a line greater than
     * the scanline stride).
     * @return   String   Error string, if the layout is incompatible with
     *                    the data.  Otherwise returns null.
     */
    private void verify (boolean strictCheck) {
	int lastbit = (dataBitOffset
		       + (height-1) * scanlineStride * 8
		       + (width-1) * pixelBitStride
		       + numberOfBits - 1);
	if (lastbit / 8 >= data.length) {
	    throw new RasterFormatException("raster dimensions overflow " +
					    "array bounds");
	}
	if (strictCheck) {
	    if (height > 1) {
		lastbit = (width - 1) * pixelBitStride + numberOfBits - 1;
		if (lastbit / 8 >= scanlineStride) {
		    throw new RasterFormatException("data for adjacent" +
						    " scanlines overlaps");
		}
	    }
	}
    }

    public String toString() {
        return new String ("BytePackedRaster: width = "+width+" height = "+height
                           +" #channels "+numBands
                           +" xOff = "+sampleModelTranslateX
                           +" yOff = "+sampleModelTranslateY);
    }
}
