/*
 * @(#)RasterPrinterJob.java	1.7 98/09/14
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.java2d;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.geom.AffineTransform;
import java.awt.geom.Area;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.awt.print.Book;
import java.awt.print.Pageable;
import java.awt.print.PageFormat;
import java.awt.print.Paper;
import java.awt.print.Printable;
import java.awt.print.PrinterAbortException;
import java.awt.print.PrinterException;
import java.awt.print.PrinterJob;
import java.util.Enumeration;
import sun.awt.image.IntegerComponentRaster;

/** 
 * A class which rasterizes a printer job.
 *
 * @version 1.0 26 Jan 1998
 * @author Richard Blanchard
 */
public abstract class RasterPrinterJob extends PrinterJob {

 /* Class Constants */

    /**
     * The band size into which we allocate the page
     * is not larger than this size. In fact it is
     * very close to this size but is a multiple of
     * the line width.
     */
    private static final int MAX_BAND_SIZE = (1024 * 512);

/* Instance Variables */

    /**
     * The number of book copies to be printed.
     */
    private int mNumCopies = 1;
    
    /**
     * Collation effects the order of the pages printed
     * when multiple copies are requested. For two copies
     * of a three page document the page order is:
     *	mCollate true: 1, 2, 3, 1, 2, 3
     *	mCollate false: 1, 1, 2, 2, 3, 3
     */
    private boolean mCollate = false;

    /**
     * The zero based indices of the first and last
     * pages to be printed. If 'mFirstPage' is 
     * UNDEFINED_PAGE_NUM then the first page to
     * be printed is page 0. If 'mLastPage' is
     * UNDEFINED_PAGE_NUM then the last page to
     * be printed is the last one in the book.
     */
    private int mFirstPage = Pageable.UNKNOWN_NUMBER_OF_PAGES;
    private int mLastPage = Pageable.UNKNOWN_NUMBER_OF_PAGES;

    /**
     * The document to be printed. It is initialized to an
     * empty (zero pages) book.
     */
    private Pageable mDocument = new Book();

    /**
     * The name of the job being printed.
     */
    private String mDocName = new String("");

 /* Constructors */

    public RasterPrinterJob()
    {
    }

/* Abstract Methods */

    /**
     * Returns the resolution in dots per inch across the width
     * of the page.
     */
    abstract protected double getXRes();

    /**
     * Returns the resolution in dots per inch down the height
     * of the page.
     */
    abstract protected double getYRes();

    /**
     * Begin a new page. This call's Window's
     * StartPage routine.
     */
    abstract protected void startPage(PageFormat format, Printable painter,
                                      int index)
	throws PrinterException;

    /**
     * End a page.
     */
    abstract protected void endPage(PageFormat format, Printable painter,
                                    int index)
	throws PrinterException;

    /**
     * Prints the contents of the array of ints, 'data'
     * to the current page. The band is placed at the
     * location (x, y) in device coordinates on the
     * page. The width and height of the band is
     * specified by the caller.
     */
    abstract protected void printBand(int[] data, int x, int y,
				      int width, int height)
	throws PrinterException;

/* Instance Methods */

    /**
     * The pages in the document to be printed by this PrinterJob
     * are drawn by the Printable object 'painter'. The PageFormat
     * for each page is the default page format.
     * @param Printable Called to render each page of the document.
     */
    public void setPrintable(Printable painter) {
	setPageable(new OpenBook(defaultPage(new PageFormat()), painter));
    }

    /**
     * The pages in the document to be printed by this PrinterJob
     * are drawn by the Printable object 'painter'. The PageFormat
     * of each page is 'format'.
     * @param Printable Called to render each page of the document.
     * @param PageFormat The size and orientation of each page to
     *                   be printed.
     */
    public void setPrintable(Printable painter, PageFormat format) {
	setPageable(new OpenBook(format, painter));
    }

    /**
     * The pages in the document to be printed are held by the
     * Pageable instance 'document'. 'document' will be queried
     * for the number of pages as well as the PageFormat and
     * Printable for each page.
     * @param Pageable The document to be printed. It may not be null.
     * @exception NullPointerException the Pageable passed in was null.
     * @see PageFormat
     * @see Printable
     */
    public void setPageable(Pageable document) throws NullPointerException {
	if (document != null) {
	    mDocument = document;

	} else {
	    throw new NullPointerException();
	}
    }

    /**
     * Prints a set of pages.
     * @exception java.awt.print.PrinterException an error in the print system
     *                                          caused the job to be aborted
     * @see java.awt.print.Book
     * @see java.awt.print.Pageable
     * @see java.awt.print.Printable
     */
    public void print() throws PrinterException {
	int numCollatedCopies = getCollatedCopies();
	int numNonCollatedCopies = getNoncollatedCopies();

	/* Get the range of pages we are to print. If the
	 * last page to print is unknown, then we print to
	 * the end of the document. Note that firstPage
	 * and lastPage are 0 based page indices.
	 */
	int numPages = mDocument.getNumberOfPages();

	int firstPage = getFirstPage();
	int lastPage = getLastPage();
	if(lastPage == Pageable.UNKNOWN_NUMBER_OF_PAGES){
	    int totalPages = mDocument.getNumberOfPages();
	    if (totalPages != Pageable.UNKNOWN_NUMBER_OF_PAGES) {
		lastPage = mDocument.getNumberOfPages() - 1;
	    }
	}

        try {
	    startDoc();
            if (isCancelled()) throw new PrinterAbortException();

	    /* Three nested loops iterate over the document. The outer loop
	     * counts the number of collated copies while the inner loop
	     * counts the number of nonCollated copies. Normally, one of
	     * these two loops will only execute once; that is we will
	     * either print collated copies or noncollated copies. The
	     * middle loop iterates over the pages.
	     */
	    for(int collated = 0; collated < numCollatedCopies; collated++) {
		for(int i = firstPage, pageResult = Printable.PAGE_EXISTS;
		    (i <= lastPage ||
		     lastPage == Pageable.UNKNOWN_NUMBER_OF_PAGES)
		    && pageResult == Printable.PAGE_EXISTS;
		    i++)
		{
		    for(int nonCollated = 0;
			nonCollated < numNonCollatedCopies
			&& pageResult == Printable.PAGE_EXISTS;
			nonCollated++)
		    {
                        if (isCancelled()) {
			    throw new PrinterAbortException();
			}
			pageResult = printPage(mDocument, i);

		    }
		}
	    }

            if (isCancelled()) throw new PrinterAbortException();
        } finally {
	    endDoc();
        }
    }

    /**
     * The passed in PageFormat will be altered to describe
     * the default page size and orientation of the PrinterJob's
     * current printer.
     *  01/12/98 Note: Needs implementation.
     */
    public PageFormat defaultPage(PageFormat page) {
	return page;
    }

    /**
     * The passed in PageFormat is altered to be usable on
     * the PrinterJob's current printer.
     *  01/12/98 Note: Needs implementation.
     */
    public PageFormat validatePage(PageFormat page) {
	return page;
    }

    /**
     * Set the number of copies to be printed.
     */
    public void setCopies(int copies) {
    	mNumCopies = copies;
    }

    /**
     * Get the number of copies to be printed.
     */
    public int getCopies() {
	return mNumCopies;
    }

    /**
     * Get the name of the printing user.
     */
    public String getUserName() {
	return System.getProperty("user.name");
    }

    /**
     * Set the name of the document to be printed.
     * The document name can not be null.
     */
    public void setJobName(String jobName) {
        if (jobName != null) {
            mDocName = jobName;
        } else {
            throw new NullPointerException();
        }
    }

    /**
     * Get the name of the document to be printed.
     */
    public String getJobName() {
        return mDocName;
    }

    /**
     * Set the range of pages from a Book to be printed.
     * Both 'firstPage' and 'lastPage' are zero based
     * page indices. If either parameter is less than
     * zero then the page range is set to be from the
     * first page to the last.
     */
    protected void setPageRange(int firstPage, int lastPage) {
	if(firstPage >= 0 && lastPage >= 0) {
	    mFirstPage = firstPage;
	    mLastPage = lastPage;
	    if(mLastPage < mFirstPage) mLastPage = mFirstPage;
	} else {
	    mFirstPage = Pageable.UNKNOWN_NUMBER_OF_PAGES;
	    mLastPage = Pageable.UNKNOWN_NUMBER_OF_PAGES;
	}
    }

    /**
     * Return the zero based index of the first page to
     * be printed in this job.
     */
    protected int getFirstPage() {
	return mFirstPage == Book.UNKNOWN_NUMBER_OF_PAGES ? 0 : mFirstPage;
    }

    /**
     * Return the zero based index of the last page to
     * be printed in this job.
     */
    protected int getLastPage() {
	return mLastPage;
    }

    /**
     * Set whether copies should be collated or not.
     * Two collated copies of a three page document
     * print in this order: 1, 2, 3, 1, 2, 3 while
     * uncollated copies print in this order:
     * 1, 1, 2, 2, 3, 3.
     */
    protected void setCollated(boolean collate) {
	mCollate = collate;
    }

    /**
     * Return true if collated copies will be printed.
     */
    protected boolean isCollated() {
	return mCollate;
    }

    /**
     * Called by the print() method at the start of
     * a print job.
     */
    protected abstract void startDoc() throws PrinterException;

    /**
     * Called by the print() method at the end of
     * a print job.
     */
    protected abstract void endDoc() throws PrinterException;


    /**
     * Returns how many times the entire book should
     * be printed by the PrintJob. If the printer
     * itself supports collation then this method
     * should return 1 indicating that the entire
     * book need only be printed once and the copies
     * will be collated and made in the printer.
     */
    protected int getCollatedCopies() {
    	return isCollated() ? getCopies() : 1;
    }

    /**
     * Returns how many times each page in the book
     * should be consecutively printed by PrintJob.
     * If the printer makes copies itself then this
     * method should return 1.
     */
    protected int getNoncollatedCopies() {
	return isCollated() ? 1 : getCopies();
    }

    /**
     * Print a page from the provided document.
     * @return int Printable.PAGE_EXISTS if the page existed and was drawn and
     *             Printable.NO_SUCH_PAGE if the page did not exist.
     * @see java.awt.print.Printable
     */
    protected int printPage(Pageable document, int pageIndex) 
	throws PrinterException
    {
	PageFormat page = document.getPageFormat(pageIndex);
	Printable painter = document.getPrintable(pageIndex);

	/* Get the imageable area from Paper instead of PageFormat
	 * because we do not want it adjusted by the page orientation.
	 */
	Paper paper = page.getPaper();
    	double xScale = getXRes() / 72.0;
	double yScale = getYRes() / 72.0;

	/* The deviceArea is the imageable area in the printer's
	 * resolution.
    	 */
    	Rectangle2D deviceArea =
	    new Rectangle2D.Double(0.0, 0.0,
				   paper.getImageableWidth() * xScale,
				   paper.getImageableHeight() * yScale);

	/* Build and hold on to a uniform transform so that
	 * we can get back to device space at the beginning
	 * of each band.
	 */
	AffineTransform uniformTransform = new AffineTransform();

	/* The scale transform is used to switch from the
	 * device space to the user's 72 dpi space.
	 */
	AffineTransform scaleTransform = new AffineTransform();
	scaleTransform.scale(xScale, yScale);

	int bandWidth = (int) deviceArea.getWidth();

	/* Figure out the number of lines that will fit into
	 * our maximum band size. The hard coded 4 reflects the
	 * fact that we can only create 32 bit per pixel
	 * BufferedImages. FIX.
	 */
	int deviceAreaHeight = (int)deviceArea.getHeight();
	int bandHeight = (int)(MAX_BAND_SIZE / deviceAreaHeight / 4);

	int deviceLeft = (int)(paper.getImageableX() * xScale);
	int deviceTop  = (int)(paper.getImageableY() * yScale);

	/* The device transform is used to move the band down
	 * the page using translates. Normally this is all it
	 * would do, but since, when printing, the Window's
	 * DIB format wants the last line to be first (lowest) in
	 * memory, the deviceTransform moves the origin to the
	 * bottom of the band and flips the origin. This way the
	 * app prints upside down into the band which is the DIB
	 * format.
	 */
	AffineTransform deviceTransform = new AffineTransform();
	deviceTransform.translate(-deviceLeft, deviceTop);
	deviceTransform.translate(0, bandHeight);
	deviceTransform.scale(1, -1);

	/* Create a BufferedImage to hold the band. We set the clip
	 * of the band to be tight around the bits so that the
	 * application can use it to figure what part of the
	 * page needs to be drawn. The clip is never altered in
	 * this method, but we do translate the band's coordinate
	 * system so that the app will see the clip moving down the
	 * page though it s always around the same set of pixels.
	 */
	BufferedImage band = new BufferedImage(bandWidth, bandHeight,
					       BufferedImage.TYPE_INT_RGB);


	/* Have the app draw into a PeekGraphics object so we can
	 * learn something about the needs of the print job.
	 */
	//PeekGraphics peekGraphics = new PeekGraphics(band.createGraphics(),
	//					     this);

	PeekGraphics peekGraphics = createPeekGraphics(band.createGraphics(),
						       this);

        Rectangle2D.Double pageFormatArea =
	    new Rectangle2D.Double(page.getImageableX(),
				   page.getImageableY(),
				   page.getImageableWidth(),
				   page.getImageableHeight());
	peekGraphics.setClip(pageFormatArea);
	int pageResult = painter.print(peekGraphics, page, pageIndex);
	if (pageResult == Printable.PAGE_EXISTS) {
	    
	    startPage(page, painter, pageIndex);

	    Graphics2D bandGraphics = band.createGraphics();

	    Graphics2D pathGraphics = createPathGraphics(peekGraphics, this);

	    /* If we can convert the page directly to the
	     * underlying graphics system then we do not
	     * need to rasterize. We also may not need to
	     * create the 'band' if all the pages can take
	     * this path.
	     */
	    if (pathGraphics != null) {
		pathGraphics.transform(scaleTransform);
		pathGraphics.setClip(pageFormatArea);
		painter.print(pathGraphics, page, pageIndex);

	    /* This is the banded-raster printing loop.
	     * It should be moved into its own method.
	     */
	    } else {
		ProxyGraphics2D painterGraphics =
		    new ProxyGraphics2D(bandGraphics, this);

		bandGraphics.setClip(0, 0, bandWidth, bandHeight);
		bandGraphics.setColor(Color.black);

		Graphics2D clearGraphics = band.createGraphics();
		clearGraphics.setColor(Color.white);

		/* We need the actual bits of the BufferedImage to send to
		 * the native Window's code. 'data' points to the actual
		 * pixels. Right now these are in ARGB format with 8 bits
		 * per component. We need to use a monochrome BufferedImage
		 * for monochrome printers when this is supported by
		 * BufferedImage. FIX
		 */
		IntegerComponentRaster tile = (IntegerComponentRaster)band.getRaster();
		int[] data = tile.getDataStorage();

		/* Loop over the page moving our band down the page,
		 * calling the app to render the band, and then send the band
		 * to the printer.
		 */
		int deviceBottom = deviceTop + deviceAreaHeight;

		for (int bandTop = 0; bandTop <= deviceAreaHeight;
		     bandTop += bandHeight)
		{
				    
		    /* Put the band back into device space and 
		     * erase the contents of the band.
		     */
		    clearGraphics.fillRect(0, 0, bandWidth, bandHeight);

		    /* Put the band into the correct location on the
		     * page. Once the band is moved we translate the
		     * device transform so that the band will move down
		     * the page on the next iteration of the loop.
		     */
		    bandGraphics.setTransform(uniformTransform);
		    bandGraphics.transform(deviceTransform);
		    deviceTransform.translate(0, -bandHeight);

		    /* Switch the band from device space to user,
		     * 72 dpi, space.
		     */
		    bandGraphics.transform(scaleTransform);
		    bandGraphics.transform(new AffineTransform(page.getMatrix()));

		    Rectangle clip = bandGraphics.getClipBounds();

		    if (peekGraphics.hitsDrawingArea(clip)) {

			/* Have the app's painter image into the band
			 * and then send the band to the printer.
			 */
			painterGraphics.setDelegate((Graphics2D) bandGraphics.create());

			painter.print(painterGraphics, page, pageIndex);
			painterGraphics.dispose();
			printBand(data, deviceLeft, deviceTop + bandTop,
				  bandWidth, bandHeight);
		    }
		}

		clearGraphics.dispose();
	    }
	    bandGraphics.dispose();
	    endPage(page, painter, pageIndex);
	}

	return pageResult;
    }

    /**
     * If a print job is in progress, print() has been
     * called but has not returned, then this signals
     * that the job should be cancelled and the next
     * chance. If there is no print job in progress then
     * this call does nothing.
     */
    public void cancel() {
        // Not yet implemented
    }

    /**
     * Returns true is a print job is ongoing but will
     * be cancelled and the next opportunity. false is
     * returned otherwise.
     */
    public boolean isCancelled() {
        //Not yet implemented
        return false;
    }

    /**
     * Return the Pageable describing the pages to be printed.
     */
    protected Pageable getPageable() {
        return mDocument;
    }

    /**
     * Examine the metrics captured by the
     * <code>PeekGraphics</code> instance and
     * if capable of directly converting this
     * print job to the printer's control language
     * or the native OS's graphics primitives, then
     * return a <code>PathGraphics</code> to perform
     * that conversion. If there is not an object
     * capable of the conversion then return
     * <code>null</code>. Returning <code>null</code>
     * causes the print job to be rasterized.
     */
    protected Graphics2D createPathGraphics(PeekGraphics graphics,
					      PrinterJob printerJob) {
	
	return null;
    }

    /**
     * Create and return an object that will
     * gather and hold metrics about the print
     * job. This method is passed a <code>Graphics2D</code>
     * object that can be used as a proxy for the
     * object gathering the print job matrics. The
     * method is also supplied with the instance
     * controlling the print job, <code>printerJob</code>.
     */
    protected PeekGraphics createPeekGraphics(Graphics2D graphics,
					      PrinterJob printerJob) {
	
	return new PeekGraphics(graphics, printerJob);
    }

}
