/*
 * @(#)SunOutputManager.java	1.30 98/09/08
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.java2d;

import java.awt.AlphaComposite;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.Paint;
import java.awt.CompositeContext;
import java.awt.GraphicsConfiguration;
import java.awt.geom.AffineTransform;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.IndexColorModel;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import sun.java2d.pipe.AlphaColorPipe;
import sun.java2d.pipe.AlphaPaintPipe;
import sun.java2d.pipe.CompositePipe;
import sun.java2d.pipe.GeneralCompositePipe;
import sun.java2d.pipe.PixelDrawPipe;
import sun.java2d.pipe.PixelFillPipe;
import sun.java2d.pipe.SimpleRenderPipe;
import sun.java2d.pipe.TranslateablePipe;
import sun.java2d.pipe.LineStylePipe;
import sun.java2d.pipe.ShapeDrawPipe;
import sun.java2d.pipe.DuctusClipRenderer;
import sun.java2d.pipe.SpanClipRenderer;
import sun.java2d.pipe.DuctusShapeRenderer;
import sun.java2d.pipe.PixelToShapeConverter;
import sun.java2d.pipe.ShapeToPolyConverter;
import sun.java2d.pipe.DrawToFillConverter;

/**
 * This class provides methods to manage rendering pipelines and
 * related information about a particular screen configuration.
 * It is the superclass for all of the OutputManager objects in
 * the JavaSoft implementation of Java 2D.
 * This class provides a repository of generalized pipeline
 * construction utilities that can be used by other OutputManager
 * objects with common needs to construct their pipelines.
 */
public abstract class SunOutputManager implements OutputManager {
    /**
     * Construct the necessary PixelDrawPipe, PixelFillPipe, and
     * ShapeDrawPipe objects necessary for the given SunGraphics2D
     * object to respond to the various rendering methods.
     * This default implementation of the method constructs the
     * most generic forms of these pipelines and is meant to be
     * overridden by the subclasses with implementations that use
     * more specific rendering capabilities available depending
     * on the particular output format.
     */
    public void validatePipe(SunGraphics2D sg) {
	constructAlphaPipe(sg);
	constructGenericPipe(sg);
    }

    /**
     * Render the indicated image with the specified modifications.
     * The current transform, composite, and clip of the SunGraphics2D
     * are also applied.
     */
    public Object renderImage(SunGraphics2D sg, BufferedImage bImg, Color bg,
			    int dx, int dy, int w, int h, 
			    byte[] clipAlpha, int rasterScan,
                            Object cachedState) {
        new Exception().printStackTrace();
        return null;
    }

     public void drawString(SunGraphics2D sg, int encoding,
                           String str, float x, float y,
                           float alignment,
                           float[] matrix, Font font,
                            boolean doAntialias, boolean doFractEnable){
         throw new NoSuchMethodError("must be implemented in child class");
     }

    private static CompositePipe generalpipe;
    private static CompositePipe paintpipe;

    static {
	generalpipe = new GeneralCompositePipe();
	paintpipe = new AlphaPaintPipe();
    }

    /**
     * This method constructs a CompositePipe object for rendering
     * alpha Rasters to the destination.
     */
    public static void constructAlphaPipe(SunGraphics2D sg) {
	CompositePipe alphapipe;

	if (sg.compositeState == sg.COMP_CUSTOM) {
	    alphapipe = generalpipe;
	} else if (sg.paintState == sg.PAINT_TILE) {
	    alphapipe = paintpipe;
	} else {
	    alphapipe = sg.getOutputManager().getColorPipe();
	    if (alphapipe == null) {
		alphapipe = paintpipe;
	    }
	}

	sg.alphapipe = alphapipe;
    }
    
    /**
     * This method constructs a generic pipeline that uses the Ductus
     * renderer for all rendering.
     * The only support object this method depends on is a CompositePipe
     * which can be constructed from a SunGraphics2D object.
     */
    public static void constructGenericPipe(SunGraphics2D sg) {
	CompositePipe alphapipe = sg.alphapipe;
	PixelDrawPipe drawpipe;
	PixelFillPipe fillpipe;
	ShapeDrawPipe shapepipe;

	if (sg.clipState == SunGraphics2D.CLIP_SHAPE) {
	    // REMIND: Need a better way to validate the clip
	    sg.getCompBounds();
	    alphapipe = new SpanClipRenderer(alphapipe, sg.clipRegion);
	}
	shapepipe = new DuctusShapeRenderer(alphapipe);
	PixelToShapeConverter cvt = new PixelToShapeConverter(shapepipe);
	drawpipe = cvt;
	fillpipe = cvt;

	sg.drawpipe = drawpipe;
	sg.fillpipe = fillpipe;
	sg.shapepipe = shapepipe;
    }

    static final int NON_UNIFORM_SCALE_MASK =
	(AffineTransform.TYPE_GENERAL_TRANSFORM |
	 AffineTransform.TYPE_GENERAL_SCALE);

    /**
     * This method constructs an aliased rendering pipeline that uses
     * a SimpleRenderPipe object to perform all rendering.
     * If the SimpleRenderPipe implements the TranslateablePipe or the
     * LineStylePipe then more of the raw graphics primitives will be
     * filtered through it.
     * If the SimpleRenderPipe implements PixelDrawPipe or PixelFillPipe
     * and no geometry conversion is needed, then it will be used to
     * handle those methods directly.
     */
    public static void constructAliasedPipe(SunGraphics2D sg,
					    SimpleRenderPipe srp) {
	PixelDrawPipe drawpipe;
	PixelFillPipe fillpipe;
	ShapeDrawPipe shapepipe;

	boolean needTx;
	if (srp instanceof TranslateablePipe) {
	    needTx = (sg.transformState == SunGraphics2D.TRANSFORM_GENERIC);
	    int tx, ty;
	    if (sg.transformState == SunGraphics2D.TRANSFORM_TRANSLATEONLY) {
                // REMIND: this is bad if translation is not an integer
                // since anti-alias may be desired on shape edges with
                // non-integer transforms
		tx = (int) sg.transX;
		ty = (int) sg.transY;
	    } else {
		tx = ty = 0;
	    }
	    ((TranslateablePipe) srp).setOrigin(tx, ty);
	} else {
	    needTx = (sg.transformState != SunGraphics2D.TRANSFORM_ISIDENT);
	}
	if (srp instanceof ShapeDrawPipe) {
	    shapepipe = (ShapeDrawPipe) srp;
	} else {
	    shapepipe = new ShapeToPolyConverter(srp, needTx);
	}
	if (!needTx && srp instanceof PixelFillPipe) {
	    fillpipe = (PixelFillPipe) srp;
	} else {
	    fillpipe = new PixelToShapeConverter(shapepipe);
	}
	boolean converttoshapes;
	boolean converttofill;
	if (!(sg.stroke instanceof BasicStroke)) {
	    converttoshapes = converttofill = true;
	} else {
	    boolean thinlines;
	    if ((sg.transform.getType() & NON_UNIFORM_SCALE_MASK) != 0) {
		thinlines = false;
	    } else {
		BasicStroke bs = (BasicStroke) sg.stroke;
		float scale = (float)
		    Math.sqrt(Math.abs(sg.transform.getDeterminant()));
		float width = bs.getLineWidth();
		float dashes[] = bs.getDashArray();
		if (srp instanceof LineStylePipe) {
		    thinlines = true;
		    ((LineStylePipe) srp).setLineAttrs(scale, width,
						       dashes,
						       bs.getDashPhase(),
						       bs.getEndCap(),
						       bs.getLineJoin());
		} else {
		    thinlines = ((scale * width <= 1.0f) &&
				 (dashes == null));
		}
	    }
	    converttoshapes = !thinlines || needTx;
	    // ShapeDrawPipe's do a good job of wide lines and
	    //     a bad job of thin lines
	    // All other pipes do a good job of thin lines and
	    //     can't do wide lines at all...
	    if (srp instanceof ShapeDrawPipe) {
		converttofill = false;
		if (thinlines) {
		    shapepipe = new ShapeToPolyConverter(srp, shapepipe,
							 needTx);
		}
	    } else {
		converttofill = !thinlines;
	    }
	}
	if (converttofill) {
	    shapepipe = new DrawToFillConverter(srp, shapepipe, true);
	}
	if (!converttoshapes && srp instanceof PixelDrawPipe) {
	    drawpipe = (PixelDrawPipe) srp;
	} else {
	    drawpipe = new PixelToShapeConverter(shapepipe);
	}

	sg.drawpipe = drawpipe;
	sg.fillpipe = fillpipe;
	sg.shapepipe = shapepipe;
    }
}
