/*
 * @(#)GeneralRenderer.java	1.13 98/07/21
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author Charlton Innovations, Inc.
 */

package sun.java2d.loops;

import sun.java2d.loops.GraphicsPrimitives;
import sun.java2d.pipe.SpanIterator;
import java.awt.image.WritableRaster;
import sun.java2d.SunGraphics2D;
import java.util.Comparator;
import java.awt.Color;

/**
 * GeneralRenderer collection
 * Basically, a collection of components which permit basic
 * rendering to occur on rasters of any format
 */

public final class GeneralRenderer implements GraphicsPrimitives {

    //
    // int arrays used for the surface types:
    //
    static int[] ST_ANY_UNARY = { GraphicsPrimitive.TYPE_ANY_UNARY };

    public final GraphicsPrimitive[] getPrimitives() {
        GraphicsPrimitive[] primitives = {
            new  GraphicsPrimitiveProxy(this, "FillRectANY",
					ST_ANY_UNARY,
					FillRect.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "FillSpansANY",
					ST_ANY_UNARY,
					FillSpans.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "DrawLineANY",
					ST_ANY_UNARY,
					DrawLine.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "FillPolygonsANY",
					ST_ANY_UNARY,
					FillPolygons.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "DrawPolygonsANY",
					ST_ANY_UNARY,
					DrawPolygons.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "DrawRectANY",
					ST_ANY_UNARY,
					DrawRect.getMethodSignature()),
        };
        return primitives;
    }




    static void drawPoly(ImageData dstImage,
            Object dstPix,
            int xPoints[], int yPoints[], int off, int nPoints,
	    boolean close)
    {
	int mx, my, x1, y1;
	mx = x1 = xPoints[off];
	my = y1 = yPoints[off];
	while (--nPoints > 0) {
	    ++off;
	    int x2 = xPoints[off];
	    int y2 = yPoints[off];
	    GeneralRenderer.doDrawLine(dstImage, dstPix, x1, y1, x2, y2);
	    x1 = x2;
	    y1 = y2;
	}
	if (close && (x1 != mx || y1 != my)) {
	    GeneralRenderer.doDrawLine(dstImage, dstPix, x1, y1, mx, my);
	}
    }

    static void doDrawLine(ImageData dstImage, Object dstPix,
                            int origx1, int origy1, int origx2, int origy2) {
        int x1 = origx1;
        int y1 = origy1;
        int x2 = origx2;
        int y2 = origy2;

	WritableRaster dstRast = (WritableRaster) dstImage.getRaster();
        if (null == dstRast) {
            return;
        }

        int [] boundPts = { x1, y1, x2, y2 };
        if (!adjustLine(dstImage, boundPts)) {
            return;
        }
        x1 = boundPts[0];
        y1 = boundPts[1];
        x2 = boundPts[2];
        y2 = boundPts[3];

        /* this could be made smaller, more elegant, more traditional. */
        if (x1 == x2) {
            if (y1 > y2) {
                do {
                    dstRast.setDataElements(x1, y1, dstPix);
                    y1--;
                } while (y1 >= y2);
                return;
            } else {
                do {
                    dstRast.setDataElements(x1, y1, dstPix);
                    y1++;
                } while (y1 <= y2);
                return;
            }
        } else if (y1 == y2) {
            if (x1 > x2) {
                do {
                    dstRast.setDataElements(x1, y1, dstPix);
                    x1--;
                } while (x1 >= x2);
            } else {
                do {
                    dstRast.setDataElements(x1, y1, dstPix);
                    x1++;
                } while (x1 <= x2);
            }
        } else {
            int dx = origx2 - origx1;
            int dy = origy2 - origy1;
            int ax = (dx<0)? -dx: dx;
            int ay = (dy<0)? -dy: dy;
            int steps;
            int bumpmajor;
            int bumpminor;
            int errminor;
            int errmajor;
            int error;
            boolean xmajor;

            if (ax >= ay) {
                /* x is dominant */
                xmajor = true;
                errmajor = ay * 2;
                errminor = ax * 2;
                bumpmajor = (dx < 0) ? -1 : 1;
                bumpminor = (dy < 0) ? -1 : 1;
                error = - (errminor / 2);
                ax = -ax; /* For clipping adjustment below */
                steps = x2 - x1;
            } else {
                /* y is dominant */
                xmajor = false;
                errmajor = ax * 2;
                errminor = ay * 2;
                bumpmajor = (dy < 0) ? -1 : 1;
                bumpminor = (dx < 0) ? -1 : 1;
                ay = -ay; /* For clipping adjustment below */
                steps = y2 - y1;
            }
            error = - (errminor / 2);
            if (y1 != origy1) {
                int ysteps = y1 - origy1;
                if (ysteps < 0) {
                    ysteps = -ysteps;
                }
                error += ysteps * ax * 2;
            }
            if (x1 != origx1) {
                int xsteps = x1 - origx1;
                if (xsteps < 0) {
                    xsteps = -xsteps;
                }
                error += xsteps * ay * 2;
            }
            if (steps < 0) {
                steps = -steps;
            }
            if (xmajor) {
                do {
                    dstRast.setDataElements(x1, y1, dstPix);
                    x1 += bumpmajor;
                    error += errmajor;
                    if (error >= 0) {
                        y1 += bumpminor;
                        error -= errminor;
                    }
                } while (--steps >= 0);
            } else {
                do {
                    dstRast.setDataElements(x1, y1, dstPix);
                    y1 += bumpmajor;
                    error += errmajor;
                    if (error >= 0) {
                        x1 += bumpminor;
                        error -= errminor;
                    }
                } while (--steps >= 0);
            }
        }
    }

    static int OUTCODE_TOP      =1;
    static int OUTCODE_BOTTOM   =2;
    static int OUTCODE_LEFT     =4;
    static int OUTCODE_RIGHT    =8;

    static int outcode(int x, int y,
            int xmin, int ymin, int xmax, int ymax) {
        int code;
        if (y < ymin) {
            code = OUTCODE_TOP;
        } else if (y > ymax) {
            code = OUTCODE_BOTTOM;
        } else {
            code = 0;
        }
        if (x < xmin) {
            code |= OUTCODE_LEFT;
        } else if (x > xmax) {
            code |= OUTCODE_RIGHT;
        }
        return code;
    }

    static boolean adjustLine(
            ImageData dstImage, int [] boundPts) {
        int x1 = boundPts[0];
        int y1 = boundPts[1];
        int x2 = boundPts[2];
        int y2 = boundPts[3];
        int cxmin = dstImage.getXOutputArea();
        int cymin = dstImage.getYOutputArea();
        int cxmax = cxmin + dstImage.getDXOutputArea() - 1;
        int cymax = cymin + dstImage.getDYOutputArea() - 1;

        if ((cxmax < cxmin) || (cymax < cymin)) {
            return false;
        }

        if (x1 == x2) {
            if (x1 < cxmin || x1 > cxmax) {
                return false;
            }
            if (y1 > y2) {
                int t = y1;
                y1 = y2;
                y2 = t;
            }
            if (y1 < cymin) {
                y1 = cymin;
            }
            if (y2 > cymax) {
                y2 = cymax;
            }
            if (y1 > y2) {
                return false;
            }
            boundPts[1] = y1;
            boundPts[3] = y2;
        } else if (y1 == y2) {
            if (y1 < cymin || y1 > cymax) {
                return false;
            }
            if (x1 > x2) {
                int t = x1;
                x1 = x2;
                x2 = t;
            }
            if (x1 < cxmin) {
                x1 = cxmin;
            }
            if (x2 > cxmax) {
                x2 = cxmax;
            }
            if (x1 > x2) {
                return false;
            }
            boundPts[0] = x1;
            boundPts[2] = x2;
        } else {
            /* REMIND: This could overflow... */
            int outcode1, outcode2;
            int dx = x2 - x1;
            int dy = y2 - y1;
            int ax = (dx < 0) ? -dx : dx;
            int ay = (dy < 0) ? -dy : dy;
            boolean xmajor = (ax >= ay);

            outcode1 = outcode(x1, y1, cxmin, cymin, cxmax, cymax);
            outcode2 = outcode(x2, y2, cxmin, cymin, cxmax, cymax);
            while ((outcode1 | outcode2) != 0) {
                int xsteps, ysteps;
                if ((outcode1 & outcode2) != 0) {
                    return false;
                }
                if (outcode1 != 0) {
                    if (0 != (outcode1 & (OUTCODE_TOP | OUTCODE_BOTTOM))) {
                        if (0 != (outcode1 & OUTCODE_TOP)) {
                            y1 = cymin;
                        } else {
                            y1 = cymax;
                        }
                        ysteps = y1 - boundPts[1];
                        if (ysteps < 0) {
                            ysteps = -ysteps;
                        }
                        xsteps = 2 * ysteps * ax + ay;
                        if (xmajor) {
                            xsteps += ay - ax - 1;
                        }
                        xsteps = xsteps / (2 * ay);
                        if (dx < 0) {
                            xsteps = -xsteps;
                        }
                        x1 = boundPts[0] + xsteps;
                    } else if (0 !=
                            (outcode1 & (OUTCODE_LEFT | OUTCODE_RIGHT))) {
                        if (0 != (outcode1 & OUTCODE_LEFT)) {
                            x1 = cxmin;
                        } else {
                            x1 = cxmax;
                        }
                        xsteps = x1 - boundPts[0];
                        if (xsteps < 0) {
                            xsteps = -xsteps;
                        }
                        ysteps = 2 * xsteps * ay + ax;
                        if (!xmajor) {
                            ysteps += ax - ay - 1;
                        }
                        ysteps = ysteps / (2 * ax);
                        if (dy < 0) {
                            ysteps = -ysteps;
                        }
                        y1 = boundPts[1] + ysteps;
                    }
		    outcode1 = outcode(x1, y1, cxmin, cymin, cxmax, cymax);
                } else {
                    if (0 != (outcode2 & (OUTCODE_TOP | OUTCODE_BOTTOM))) {
                        if (0 != (outcode2 & OUTCODE_TOP)) {
                            y2 = cymin;
                        } else {
                            y2 = cymax;
                        }
                        ysteps = y2 - boundPts[3];
                        if (ysteps < 0) {
                            ysteps = -ysteps;
                        }
                        xsteps = 2 * ysteps * ax + ay;
                        if (xmajor) {
                            xsteps += ay - ax;
                        } else {
                            xsteps -= 1;
                        }
                        xsteps = xsteps / (2 * ay);
                        if (dx > 0) {
                            xsteps = -xsteps;
                        }
                        x2 = boundPts[2] + xsteps;
                    } else if (0 !=
                            (outcode2 & (OUTCODE_LEFT | OUTCODE_RIGHT))) {
                        if (0 != (outcode2 & OUTCODE_LEFT)) {
                            x2 = cxmin;
                        } else {
                            x2 = cxmax;
                        }
                        xsteps = x2 - boundPts[2];
                        if (xsteps < 0) {
                            xsteps = -xsteps;
                        }
                        ysteps = 2 * xsteps * ay + ax;
                        if (xmajor) {
                            ysteps -= 1;
                        } else {
                            ysteps += ax - ay;
                        }
                        ysteps = ysteps / (2 * ax);
                        if (dy > 0) {
                            ysteps = -ysteps;
                        }
                        y2 = boundPts[3] + ysteps;
                    }
		    outcode2 = outcode(x2, y2, cxmin, cymin, cxmax, cymax);
                }
            }
            boundPts[0] = x1;
            boundPts[1] = y1;
            boundPts[2] = x2;
            boundPts[3] = y2;
        }
        return true;
    }

}
class FillRectANY extends FillRect {
    FillRectANY() {
	super(GeneralRenderer.ST_ANY_UNARY);
    }
    public void FillRect(
	ImageData dstImage,
	Color color,
	int x, int y, int w, int h
    ) {
	Object dstPix = null;
	dstPix = dstImage.getColorModel().getDataElements(color.getRGB(),
							  dstPix);
	int lastX = x + w - 1;
	for (int i=0; i < h; i++) {
	    GeneralRenderer.doDrawLine(dstImage, dstPix,
					x, y+i, lastX, y+i);
	}
    }
}

class FillSpansANY extends FillSpans {
    FillSpansANY() {
	super(GeneralRenderer.ST_ANY_UNARY);
    }
    public void FillSpans(
	ImageData dstImage,
	Color color,
	SpanIterator si
    ) {
	int x1, x2, y1, y2;
	Object dstPix = null;
	dstPix = dstImage.getColorModel().getDataElements(color.getRGB(),
							  dstPix);
	int span[] = new int[4];
	WritableRaster dstRast = (WritableRaster) dstImage.getRaster();
	while (si.nextSpan(span)) {
	    x1 = span[0];
	    y1 = span[1];
	    x2 = span[2];
	    y2 = span[3];
	    do {
		int xsave = x1;
                do {
                    dstRast.setDataElements(x1, y1, dstPix);
                } while (++x1 < x2);
		x1 = xsave;
	    } while (++y1 < y2);
	}
    }
}

class DrawLineANY extends DrawLine {
    DrawLineANY() {
	super(GeneralRenderer.ST_ANY_UNARY);
    }
    public void DrawLine(
	ImageData dstImage,
	Color color,
	int x1, int y1, int x2, int y2
    ) {
	Object dstPix = null;
	dstPix = dstImage.getColorModel().getDataElements(color.getRGB(),
							  dstPix);
	if (y1 >= y2) {
	    GeneralRenderer.doDrawLine(dstImage, dstPix, x2, y2, x1, y1);
	} else {
	    GeneralRenderer.doDrawLine(dstImage, dstPix, x1, y1, x2, y2);
	}
    }
}

class FillPolygonsANY extends FillPolygons {
    FillPolygonsANY() {
	super(GeneralRenderer.ST_ANY_UNARY);
    }
    public void FillPolygons(
	ImageData dstImage,
	Color color,
	int xPoints[], int yPoints[], int nPoints[],
	int numPolys, boolean evenOdd
    ) {
	Object dstPix = null;
	dstPix = dstImage.getColorModel().getDataElements(color.getRGB(),
							  dstPix);
	EdgeCollection edges = new EdgeCollection(xPoints, yPoints,
						  nPoints, numPolys);
	if (edges.countUnfinished() != 0) {
	    edges.sortByHighestVertex();
	    edges.removeDuplicateEdges();
	    while (edges.countUnfinished() != 0) {
		edges.sortByLeftMostVertex();
		edges.drawActiveEdges(dstImage, dstPix, evenOdd);
		edges.advanceActiveEdges();
		edges.activateNewEdges();
		/* eliminates bottom-most pixel-row of each edge */
		edges.purgeUsedEdges();
	    }
	}
    }
}

class DrawPolygonsANY extends DrawPolygons {
    DrawPolygonsANY() {
	super(GeneralRenderer.ST_ANY_UNARY);
    }
    public void DrawPolygons(
	ImageData dstImage,
	Color color,
	int xPoints[], int yPoints[], int nPoints[],
	int numPolys, boolean close
    ) {
	Object dstPix = null;
	dstPix = dstImage.getColorModel().getDataElements(color.getRGB(),
							  dstPix);
	int off = 0;
	for (int i = 0; i < numPolys; i++) {
	    int numpts = nPoints[i];
	    GeneralRenderer.drawPoly(dstImage, dstPix,
				     xPoints, yPoints, off, numpts, close);
	    off += numpts;
	}
    }
}

class DrawRectANY extends DrawRect {
    DrawRectANY() {
	super(GeneralRenderer.ST_ANY_UNARY);
    }
    public void DrawRect(
	ImageData dstImage,
	Color color,
	int x, int y, int w, int h
    ) {
	Object dstPix = null;
	dstPix = dstImage.getColorModel().getDataElements(color.getRGB(),
							  dstPix);
	GeneralRenderer.doDrawLine(dstImage, dstPix, x, y, x + w - 1, y);
	GeneralRenderer.doDrawLine(dstImage, dstPix, x, y, x, y + h - 1);
	GeneralRenderer.doDrawLine(dstImage, dstPix,
				    x + w - 1, y, x + w - 1, y + h - 1);
	GeneralRenderer.doDrawLine(dstImage, dstPix,
				    x, y + h - 1, x + w - 1, y + h - 1);
    }
}

class EdgeCollection {

    private int remainingEdges; // number of edges awaiting completion
    private int activeEdges;    // number of edges active on current row
    private int firstActive;    // index in array of first active edge

    private Comparator leftComparator;
    private Comparator topLeftComparator;
    /**
    * edges array contains entries as:
    *   x x x x x x A A A A A A A U U U U U U U U U U U U
    *               ^              \--- activeEdges + firstActive
    *               \--- firstActive
    *               \ <------ remainingEdges ----------->
    * number of elements in state "x" = firstActive
    * number of elements in state "A" = activeEdges
    * number of elements in state "U" = remainingEdges - activeEdges
    */
    private EdgeEntry [] edges;
    
    EdgeCollection(int [] xPoints, int [] yPoints,
	    int [] nPoints, int numPolys) {
	remainingEdges = 0;
	activeEdges = 0;
	firstActive = 0;

	int totalPoints = numPolys;
	int pointIndex = 0;
	int edgeIndex = 0;
	for (int i=0; i<numPolys; i++) {
	    totalPoints += nPoints[i];
	}
	edges = new EdgeEntry[totalPoints];
	for (int i=0; i<numPolys; i++) {
	    int lastPoint = nPoints[i];
	    int x1 = xPoints[pointIndex];
	    int y1 = yPoints[pointIndex];
	    int x2 = xPoints[pointIndex + lastPoint - 1];
	    int y2 = yPoints[pointIndex + lastPoint - 1];
	    if ((x1 != x2) || (y1 != y2)) {
		/* add closing edge */
		edges[edgeIndex] = new EdgeEntry(x2, y2, x1, y1);
		if (edges[edgeIndex].initialized()) {
		    remainingEdges++;
		    edgeIndex++;
		}
	    }
	    while (lastPoint-- > 1) {
		x2 = xPoints[pointIndex + 1];
		y2 = yPoints[pointIndex + 1];
		edges[edgeIndex] = new EdgeEntry(x1, y1, x2, y2);
		if (edges[edgeIndex].initialized()) {
		    remainingEdges++;
		    edgeIndex++;
		}
		pointIndex++;
		x1 = x2;
		y1 = y2;
	    }
	    pointIndex++;
	}
	topLeftComparator = new TopLeftComparator();
	leftComparator = new LeftComparator();
    }

    int countUnfinished() {
	return remainingEdges;
    }

    void sortByHighestVertex() {
	if (remainingEdges > 0) {
	    sort(edges, firstActive, remainingEdges, topLeftComparator);
	}
    }

    void sortByLeftMostVertex() {
	if (activeEdges > 0) {
	    sort(edges, firstActive, activeEdges, leftComparator);
	}
    }

    void removeDuplicateEdges() {
	if (remainingEdges <= 0) {
	    return;
	}
	int srcIndex = firstActive;
	int lastIndex = firstActive + remainingEdges - 1;
	int dstIndex = firstActive;
	while (srcIndex < lastIndex) {
	    if (edges[srcIndex].mirrors(edges[srcIndex + 1])) {
		/* skip these two reversed edges */
		srcIndex += 2;
		remainingEdges -= 2;
		continue;
	    }
	    if (dstIndex != srcIndex) {
		edges[dstIndex] = edges[srcIndex];
	    }
	    dstIndex++;
	    srcIndex++;
	}
    }

    void drawEvenOddFromActiveEdges(ImageData dstImage, Object dstPix) {
	int edgesToDraw = activeEdges;
	int edgeIndex = firstActive;
	while (edgesToDraw >= 2) {
	    int y = edges[edgeIndex].getY1();
	    int x1 = edges[edgeIndex].getX1();
	    edgeIndex++;
	    int x2 = edges[edgeIndex].getX1();
	    edgeIndex++;
	    if (x2 > x1) {
		/* do not draw rightmost pixel of any segment */
		GeneralRenderer.doDrawLine(dstImage, dstPix, x1, y, x2-1, y);
	    }
	    edgesToDraw -= 2;
	}
    }

    void drawWindingFromActiveEdges(ImageData dstImage, Object dstPix) {
	int wound = 0;
	int x1 = -1;
	int y1 = -1;
	int edgesToDraw = activeEdges;
	int edgeIndex = firstActive;
	while (edgesToDraw > 0) {
	    EdgeEntry edge = edges[edgeIndex];
	    if (0 == wound) {
		x1 = edge.getX1();
		y1 = edge.getY1();
	    }
	    wound += edge.isFlipped()? -1: 1;
	    if (0 == wound) {
		int x2 = edge.getX1();
		if (x2 > x1) {
		    /* do not draw rightmost pixel of any segment */
		    GeneralRenderer.doDrawLine(dstImage, dstPix, x1, y1, x2-1, y1);
		}
	    }
	    edgesToDraw--;
	    edgeIndex++;
	}
    }

    void drawActiveEdges(ImageData dstImage, Object dstPix,
	    boolean evenOdd) {
	if (evenOdd) {
	    drawEvenOddFromActiveEdges(dstImage, dstPix);
	} else {
	    drawWindingFromActiveEdges(dstImage, dstPix);
	}
    }

    void advanceActiveEdges() {
	int cursor = firstActive;
	int firstInactive = firstActive + activeEdges;
	while (cursor < firstInactive) {
	    EdgeEntry edge = edges[cursor];
	    edge.bumpY1(1);
	    if (edge.getErrY() != 0) {
		edge.bumpError(edge.getErrX());
		while (edge.getError() > 0) {
		    edge.bumpX1(edge.getBumpX());
		    edge.bumpError(-edge.getErrY());
		}
	    }
	    cursor++;
	}
    }

    void activateNewEdges() {
	if (remainingEdges > 0) {
	    int y = edges[firstActive].getY1();
	    while (activeEdges < remainingEdges) {
		if (edges[firstActive + activeEdges].getY1() != y) {
		    break;
		}
		activeEdges++;
	    }
	}
    }

    void purgeUsedEdges() {
	if (activeEdges > 0) {
	    int edgesToCheck = activeEdges;
	    int cursor = firstActive;
	    while (edgesToCheck-- > 0) {
		if (edges[cursor].getY1() >= edges[cursor].getY2()) {
		    /* edge is entirely used up - nothing more to draw */
		    if (cursor != firstActive) {
			/* move first active into place of used edge */
			edges[cursor] = edges[firstActive];
		    }
		    firstActive++;
		    remainingEdges--;
		    activeEdges--;
		}
		cursor++;
	    }
	}
    }

    /**
     * adapted from java.util.Arrays, since the interface we need was
     * private there.
     *
     * Sorts the specified array according to the order induced by the
     * specified Comparator.  All elements in the array must be <i>mutually
     * comparable</i> by the specified comparator (that is,
     * comparator.compare(e1, e2) must not throw a typeMismatchException for
     * any elements e1 and e2 in the array).
     * <p>
     * The sorting algorithm is a tuned quicksort, adapted from Jon
     * L. Bentley and M. Douglas McIlroy's "Engineering a Sort Function",
     * Software-Practice and Experience, Vol. 23(11) P. 1249-1265 (November
     * 1993).  This algorithm offers n*log(n) performance on many data sets
     * that cause other quicksorts to degrade to quadratic performance.
     *
     */

     /**
     * Sorts the specified sub-array of EdgeEntries into ascending order
     * according to the specified Comparator interface.
     */
    private void sort(EdgeEntry x[], int off, int len, Comparator f) {
	// Insertion sort on smallest arrays
	if (len < 7) {
	    for (int i=off; i<len+off; i++)
		for (int j=i; j>off && f.compare(x[j-1], x[j]) > 0; j--)
		    swap(x, j, j-1);
	    return;
	}

	// Choose a partition element, v
	int m = off + len/2;       // Small arrays, middle element
	if (len > 7) {
	    int l = off;
	    int n = off + len - 1;
	    if (len > 40) {        // Big arrays, pseudomedian of 9
		int s = len/8;
		l = med3(x, l,     l+s, l+2*s, f);
		m = med3(x, m-s,   m,   m+s,   f);
		n = med3(x, n-2*s, n-s, n,     f);
	    }
	    m = med3(x, l, m, n, f); // Mid-size, med of 3
	}
	Object v = x[m];

	// Establish Invariant: v* (<v)* (>v)* v*
	int a = off, b = a, c = off + len - 1, d = c;
	while(true) {
	    int r;
	    while (b <= c && (r = f.compare(x[b], v)) <= 0) {
		if (r == 0)
		    swap(x, a++, b);
		b++;
	    }
	    while (c >= b && (r = f.compare(x[c], v)) >= 0) {
		if (x[c] == v)
		    swap(x, c, d--);
		c--;
	    }
	    if (b > c)
		break;
	    swap(x, b++, c--);
	}

	// Swap partition elements back to middle
	int s, n = off + len;
	s = Math.min(a-off, b-a  );  vecswap(x, off, b-s, s);
	s = Math.min(d-c,   n-d-1);  vecswap(x, b,   n-s, s);

	// Recursively sort non-partition-elements
	if ((s = b-a) > 1)
	    sort(x, off, s, f);
	if ((s = d-c) > 1)
	    sort(x, n-s, s, f);
    }

    /**
     * Returns the index of the median of the three indexed Objects,
     * according to the specified Comparator interface
     */
    private int med3(EdgeEntry x[], int a, int b, int c, Comparator f) {
	return (f.compare(x[a],x[b]) < 0
		    ? (f.compare(x[b],x[c])<0
			? b : f.compare(x[a],x[c])<0 ? c : a)
		    : (f.compare(x[b],x[c])>0
			? b : f.compare(x[a],x[c])>0 ? c : a));
    }

    /**
    * Swaps x[a] with x[b].
    */
    private void swap(EdgeEntry x[], int a, int b) {
	EdgeEntry t = x[a];
	x[a] = x[b];
	x[b] = t;
    }

    /**
    * Swaps x[a .. (a+n-1)] with x[b .. (b+n-1)].
    */
    private void vecswap(EdgeEntry x[], int a, int b, int n) {
	for (int i=0; i<n; i++, a++, b++) {
	    swap(x, a, b);
	}
    }


    private class EdgeEntry {
	private int x1,y1;
	private int x2,y2;
	/* bresenham coefficients for edge tracing */
	private int error;
	private int errx;
	private int erry;
	private int bumpx;
	private boolean flipped;
	private boolean initOK;

	EdgeEntry(int x1, int y1, int x2, int y2) {
	    this.initOK = false;
	    if ((y1 != y2)) {
		this.flipped = false;
		/* edge is not horizontal, or is a point */
		if (y1 > y2) {
		    /* 2nd point is higher, make it first */
		    this.flipped = true;
		    this.x1 = x2;
		    this.y1 = y2;
		    this.x2 = x1;
		    this.y2 = y1;
		} else {
		    /* 1st point is higher, keep it first */
		    this.x1 = x1;
		    this.y1 = y1;
		    this.x2 = x2;
		    this.y2 = y2;
		}
		int dx = this.x2 - this.x1;
		int dy = this.y2 - this.y1;
		int ax = (dx<0)? -dx: dx;
		int ay = (dy<0)? -dy: dy;
		if (dx < 0) {
		    this.bumpx = -1;
		    this.error = -ay;
		} else {
		    this.bumpx = 1;
		    this.error = 0;
		}
		this.errx = ax;
		this.erry = ay;
		initOK = true;
	    }
	}

	boolean initialized() { return initOK; }

	int getX1() { return x1; }
	int getY1() { return y1; }
	int getX2() { return x2; }
	int getY2() { return y2; }
	boolean isFlipped() { return flipped; }
	int getErrY() { return erry; }
	int getErrX() { return errx; }
	int getBumpX() { return bumpx; }
	int getError() { return error; }

	void bumpX1(int bump) { x1 += bump; }
	void bumpY1(int bump) { y1 += bump; }
	void bumpError(int bump) { error += bump; }

	/* returns true if edges have same endpoints, reversed */
	boolean mirrors(EdgeEntry e) {
	    return (e.x1 == x1) && (e.y1 == y1)
		    && (e.x2 == x2) && (e.y2 == y2)
		    && (e.flipped != flipped);
	}

	int isTopLeftOf(EdgeEntry e2) {
	    if (y1 < e2.y1) {
		return -1;
	    }
	    if (y1 > e2.y1) {
		return 1;
	    }
	    if (x1 < e2.x1) {
		return -1;
	    }
	    if (x1 > e2.x1) {
		return 1;
	    }
	    if (y2 < e2.y2) {
		return -1;
	    }
	    if (y2 > e2.y2) {
		return 1;
	    }
	    if (x2 < e2.x2) {
		return -1;
	    }
	    if (x2 > e2.x2) {
		return 1;
	    }
	    return 0;
	}

	int isLeftOf(EdgeEntry e2) {
	    if (x1 < e2.x1) {
		return -1;
	    }
	    if (x1 > e2.x1) {
		return 1;
	    }
	    return 0;
	}
    }

    private class TopLeftComparator implements Comparator {
	public int compare(Object o1, Object o2) {
	    return ((EdgeEntry)o1).isTopLeftOf((EdgeEntry)o2);
	}
    }

    private class LeftComparator implements Comparator {
	public int compare(Object o1, Object o2) {
	    return ((EdgeEntry)o1).isLeftOf((EdgeEntry)o2);
	}
    }
}
