/*
 * @(#)GraphicsPrimitive.java	1.10 98/08/10
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author Charlton Innovations, Inc.
 */

package sun.java2d.loops;

/**
* defines interface for primitives which can be placed into
* the graphic component manager framework
*/

public abstract class GraphicsPrimitive
{
    /**
    *   CONSTANTS USED BY ALL PRIMITIVES TO DESCRIBE THE SURFACES
    *   THEY CAN OPERATE ON
    **/

    /**
    * surface is unknown color model or sample model.
    */
    public static final int
        TYPE_CUSTOM         = java.awt.image.BufferedImage.TYPE_CUSTOM;

    /**
    * common surface formats defined in java.awt.image.BufferedImage
    */
    public static final int
        TYPE_INT_RGB        = java.awt.image.BufferedImage.TYPE_INT_RGB;
    public static final int
        TYPE_INT_ARGB       = java.awt.image.BufferedImage.TYPE_INT_ARGB;
    public static final int
        TYPE_INT_ARGB_PRE   = java.awt.image.BufferedImage.TYPE_INT_ARGB_PRE;
    public static final int
        TYPE_INT_BGR        = java.awt.image.BufferedImage.TYPE_INT_BGR;
    public static final int
        TYPE_3BYTE_BGR      = java.awt.image.BufferedImage.TYPE_3BYTE_BGR;
    public static final int
        TYPE_4BYTE_ABGR     = java.awt.image.BufferedImage.TYPE_4BYTE_ABGR;
    public static final int
        TYPE_4BYTE_ABGR_PRE =
                            java.awt.image.BufferedImage.TYPE_4BYTE_ABGR_PRE;
    public static final int
        TYPE_SHORT_565_RGB  = java.awt.image.BufferedImage.TYPE_USHORT_565_RGB;
    public static final int
        TYPE_SHORT_555_RGB  = java.awt.image.BufferedImage.TYPE_USHORT_555_RGB;
    public static final int
        TYPE_BYTE_GRAY      = java.awt.image.BufferedImage.TYPE_BYTE_GRAY;
    public static final int TYPE_USHORT_GRAY = java.awt.image.BufferedImage.TYPE_USHORT_GRAY;
    
    public static final int
        TYPE_BINARY         = java.awt.image.BufferedImage.TYPE_BYTE_BINARY;
    public static final int
        TYPE_BYTE_INDEXED   = java.awt.image.BufferedImage.TYPE_BYTE_INDEXED;

    /**
    * wildcard format which can be used by a GraphicsPrimitive
    * to indicate that it can operate on any data format
    * for a particular surface (generally means that GraphicsPrimitive can
    * operate on the surface with Get/SetRGB)
    */
    public static final int TYPE_ANY_UNARY      = -1;

    /**
    * wildcard format which indicate that the GraphicsPrimitive
    * can operate on any data formats for the source and destination surfaces
    * (generally means that GraphicsPrimitive can operate on the surfaces
    * with Get/SetRGB)
    */
    public static final int TYPE_ANY_BINARY     = -2;

    /**
    * wildcard format which indicate that the GraphicsPrimitive
    * is independent of the color model on an IntegerComponent sample model
    * surface
    */
    public static final int TYPE_INT_UNARY      = -3;

    /**
    * wildcard format which indicate that the GraphicsPrimitive
    * is independent of the color model on an IntegerComponent sample model
    * if the source and destination surfaces are the same color model
    */
    public static final int TYPE_INT_BINARY     = -4;

    /**
    * wildcard format which indicate that the GraphicsPrimitive
    * is independent of the color model on a ShortComponent sample model
    * surface
    */
    public static final int TYPE_SHORT_UNARY    = -5;

    /**
    * wildcard format which indicate that the GraphicsPrimitive
    * is independent of the color model on a ShortComponent sample model
    * if the source and destination surfaces are the same color model
    */
    public static final int TYPE_SHORT_BINARY   = -6;

    /**
    * wildcard format which indicate that the GraphicsPrimitive
    * is independent of the color model on a ShortComponent sample model
    * surface
    */
    public static final int TYPE_BYTE_UNARY     = -7;

    /**
    * wildcard format which indicate that the GraphicsPrimitive
    * is independent of the color model on a ShortComponent sample model
    * if the source and destination surfaces are the same color model
    */
    public static final int TYPE_BYTE_BINARY    = -8;

    /**
    * additional IntegerComponent types common on Windows
    */
    public static final int TYPE_INT_RGBx       = -10;
    public static final int TYPE_INT_BGRx       = -11;

    /**
    * additional 3 byte format common on Windows
    */
    public static final int TYPE_24BIT_RGB      = -12;

    /**
    * additional 4 byte format common on Windows
    * also useful for reading DIB from file.
    */
    public static final int TYPE_DIB_xRGB       = -13;

    /**
    * common formats for BITMASK transparency
    */
    public static final int TYPE_INT_ARGB_BM        = -14;
    public static final int TYPE_INT_ARGB_PRE_BM    = -15;
    public static final int TYPE_BYTE_INDEXED_BM    = -16;

    /**
     * Special Gray Scale types for rendering loops.  Really indexed
     * types, but colormap has all gray values.
     */
    public static final int TYPE_INDEX8_GRAY        = -17;
    public static final int TYPE_INDEX12_GRAY        = -18;

    /**
    *  INSTANCE DATA MEMBERS DESCRIBING CHARACTERISTICS OF THIS PRIMITIVE
    **/

    // Making these be instance data members (instead of virtual methods
    // overridden by subclasses) is actually cheaper, since each class
    // is a singleton.  As instance data members with final accessors,
    // accesses can be inlined.
    private int[] surfaceTypes;
    private int[] lockFlags;
    private String methodSignature;

    /**
     * Create a new GraphicsPrimitive with all the required descriptive
     * information
     */
    protected GraphicsPrimitive(int[] surfaceTypes, int[] lockFlags, 
				String methodSignature) {
	this.surfaceTypes = surfaceTypes;
	this.lockFlags = lockFlags;
	this.methodSignature = methodSignature;
    }

    /**
    *   METHODS TO DESCRIBE THE SURFACES PRIMITIVES
    *   CAN OPERATE ON AND THE FUNCTIONALITY THEY IMPLEMENT
    **/


    /**
    *   bit flags descriptive of locking needed by various
    *   GraphicsPrimitives on an image surface
    */
    public static final int LOCK_FOR_READ   = (1 << 0);
    public static final int LOCK_FOR_WRITE  = (1 << 1);

    /**
    *   locking flags for each surface this GraphicsPrimitive utilizes
    */
    public final int[] getLockFlags() {
	return lockFlags;
    }

    /**
     * Return true if this primitive can be used for the given signature
     * and surfaces.
     *
     * @param signature The signature of the given operation.  Must be
     *		== (not just .equals) the signature string given by the
     *		abstract class that declares the operation.
     * @param surfaces The surface types for the operation
     */
    final boolean satisfies(String signature, int[] surfaces) {
	if (signature != methodSignature) {
	    return false;
	}
	if (surfaces.length != surfaceTypes.length) {
	    return false;
	}
	for (int i = 0; i < surfaceTypes.length; i++) {
	    if (surfaces[i] != surfaceTypes[i]) {
		return false;
	    }
	}
	return true;
    }

    //
    // A version of satisfies used for regression testing
    //
    final boolean satisfiesSameAs(GraphicsPrimitive other) {
	return satisfies(other.methodSignature, other.surfaceTypes);
    }

    public static String typeStr(int type) {
        switch (type) {
        case TYPE_CUSTOM:           return "TYPE_CUSTOM";
        case TYPE_INT_RGB:          return "TYPE_INT_RGB";
        case TYPE_INT_ARGB:         return "TYPE_INT_ARGB";
        case TYPE_INT_ARGB_PRE:     return "TYPE_INT_ARGB_PRE";
        case TYPE_INT_BGR:          return "TYPE_INT_BGR";
        case TYPE_3BYTE_BGR:        return "TYPE_3BYTE_BGR";
        case TYPE_4BYTE_ABGR:       return "TYPE_4BYTE_ABGR";
        case TYPE_4BYTE_ABGR_PRE:   return "TYPE_4BYTE_ABGR_PRE";
        case TYPE_SHORT_565_RGB:    return "TYPE_SHORT_565_RGB";
        case TYPE_SHORT_555_RGB:    return "TYPE_SHORT_555_RGB";
        case TYPE_BYTE_GRAY:        return "TYPE_BYTE_GRAY";
        case TYPE_USHORT_GRAY:      return "TYPE_USHORT_GRAY";
        case TYPE_BINARY:           return "TYPE_BINARY";
        case TYPE_BYTE_INDEXED:     return "TYPE_BYTE_INDEXED";
        case TYPE_ANY_UNARY:        return "TYPE_ANY_UNARY";
        case TYPE_ANY_BINARY:       return "TYPE_ANY_BINARY";
        case TYPE_INT_UNARY:        return "TYPE_INT_UNARY";
        case TYPE_INT_BINARY:       return "TYPE_INT_BINARY";
        case TYPE_SHORT_UNARY:      return "TYPE_SHORT_UNARY";
        case TYPE_SHORT_BINARY:     return "TYPE_SHORT_BINARY";
        case TYPE_BYTE_UNARY:       return "TYPE_BYTE_UNARY";
        case TYPE_BYTE_BINARY:      return "TYPE_BYTE_BINARY";
        case TYPE_INT_RGBx:         return "TYPE_INT_RGBx";
        case TYPE_INT_BGRx:         return "TYPE_INT_BGRx";
        case TYPE_24BIT_RGB:        return "TYPE_24BIT_RGB";
        case TYPE_DIB_xRGB:         return "TYPE_DIB_xRGB";
        case TYPE_INT_ARGB_BM:      return "TYPE_INT_ARGB_BM";
        case TYPE_INT_ARGB_PRE_BM:  return "TYPE_INT_ARGB_PRE_BM";
        case TYPE_BYTE_INDEXED_BM:  return "TYPE_BYTE_INDEXED_BM";
        case TYPE_INDEX8_GRAY:      return "TYPE_INDEX8_GRAY";
        case TYPE_INDEX12_GRAY:      return "TYPE_INDEX12_GRAY";
        default:
            return "TYPE_UNKNOWN";
        }
    }
}
