/*
 * %W% %E%
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author Charlton Innovations, Inc.
 */

package sun.java2d.loops;

import sun.java2d.loops.GraphicsPrimitive;
import sun.java2d.loops.GraphicsPrimitives;

/**
*   GraphicsComponentMgr
*   provides services to 
*   1) register primitives for later use
*   2) locate an instance of a primitve based on characteristics
*/

public class GraphicsPrimitiveMgr {

    private static final boolean debugTrace = false;
    private static GraphicsPrimitive[][] deviceFunctions;
    private static GraphicsPrimitive[] collection;
    private static String deviceAtoms[];
    private static GraphicsPrimitiveMgr mgr = new GraphicsPrimitiveMgr();

    private GraphicsPrimitiveMgr() {
        deviceFunctions = new GraphicsPrimitive[1][];
    }

    /**
    *   add more instances of GraphicsPrimitives to the set under management
    */
    public synchronized static final void register(GraphicsPrimitives toAdd,
                                                            String deviceID) {
        if (deviceID != null) {
            int deviceIndex = devAtomIndex(deviceID);
            if (deviceIndex >= deviceFunctions.length) {
                GraphicsPrimitive[][] tempFunctions = new GraphicsPrimitive[deviceIndex + 1][];
                System.arraycopy(   deviceFunctions, 0, tempFunctions, 0,
                                    deviceFunctions.length);
                tempFunctions[deviceIndex] = null;
                deviceFunctions = tempFunctions;
            }
            deviceFunctions[deviceIndex] = register(toAdd, deviceIndex);
        } else {
            deviceFunctions[0] = register(toAdd, 0);
        }
    }

    private static final GraphicsPrimitive [] register(
                                        GraphicsPrimitives toAdd,
                                        int devIndex) {
        writeLog("Registering " + toAdd);
        GraphicsPrimitive[] newPrimitives = toAdd.getPrimitives();
        GraphicsPrimitive[] devCollection = deviceFunctions[devIndex];
        if (newPrimitives == null) {
            return devCollection;
        }
        int oldSize = 0;
        int newSize = newPrimitives.length;
        if (devCollection != null) {
            oldSize = devCollection.length;
        }
        GraphicsPrimitive[] temp = new GraphicsPrimitive[oldSize + newSize];
        if (devCollection != null) {
            System.arraycopy(devCollection, 0, temp, 0, oldSize);
        }
        System.arraycopy(newPrimitives, 0, temp, oldSize, newSize);
        return temp;
    }

    private static int devAtomIndex(String deviceID) {
        if (deviceAtoms == null) {
            deviceAtoms = new String[1];
            deviceAtoms[0] = deviceID;
            return 1;
        }
        // shortcut for performance in most common case ...
        if ((deviceAtoms.length == 1) && deviceAtoms[0].equals(deviceID)) {
            return 1;
        }
        // if more than one register device, slower search
        int numToSearch = deviceAtoms.length;
        int i;
        for (i = 0; i < numToSearch; i++) {
            if (deviceAtoms[i].equals(deviceID)) {
                return i+1;
            }
        }
        String [] tempAtoms = new String[numToSearch + 1];
        System.arraycopy(deviceAtoms, 0, tempAtoms, 0, numToSearch);
        tempAtoms[numToSearch] = deviceID;
        deviceAtoms = tempAtoms;
        return numToSearch + 1;
    }

    /**
    *   find an instance of a primitve base on specified characteristics
    *   1) signature:       the function to be performed.  This must == the
    *			    signature string of the class where the function
    *			    is declared.
    *   2) surfaceTypes:    data formats of surfaces to operate on
    *   3) deviceID:        if specified, tells which queue to
    *                       search first for destination surface
    *                       function implementations
    **/
    public synchronized static final GraphicsPrimitive
    locate(String signature, int [] surfaceTypes, String deviceID) {
        GraphicsPrimitive retval = null;
        if (deviceID != null) {
            int deviceIndex = devAtomIndex(deviceID);
            if (deviceIndex < deviceFunctions.length) {
                retval = locate(signature, surfaceTypes, deviceIndex);
            }
        }
        if (retval == null) {
            retval = locate(signature, surfaceTypes, 0);
        }
        return retval;
    }

    private static final GraphicsPrimitive
        locate( String signature, int [] surfaceTypes, int devIndex) {
        GraphicsPrimitive[] devCollection = deviceFunctions[devIndex];
        if (devCollection == null) {
            return null;
        }
        int numToSearch = devCollection.length;
        int i;
        for (i = numToSearch - 1; i >= 0; i--) {
	    GraphicsPrimitive prim = devCollection[i];
	    if (prim.satisfies(signature, surfaceTypes)) {
		if (prim instanceof GraphicsPrimitiveProxy) {
		    prim = ((GraphicsPrimitiveProxy) prim).instantiate();
		    devCollection[i] = prim;
		    if (debugTrace) {
			writeLog("Instantiated graphics primitive " + prim);
		    }
		}
                if (debugTrace) {
                    writeLog("Lookup found[" + i + "]["+ prim + "]");
                }
		promoteMatchingEntry(i, devIndex);
                return prim;
            }
        }
        writeLog("Lookup found nothing for:");
        writeLog(" " + signature);
//        if (deviceID != NULL) {
//            writeLock(" Device [" + deviceID + "]");
//        }
        for (i = 0; i < surfaceTypes.length; i++) {
            writeLog(" " + GraphicsPrimitive.typeStr(surfaceTypes[i]));
        }
        return null;
    }
    
    /**
    *   move MRU entry to first search position if not already "close"
    *   is is worth noting that this promotion does not change the
    *   LIFO search criteria which is used to prioritize multiple
    *   implementations of the same primitive
    */
    private static void promoteMatchingEntry(int i, int devIndex) {
        GraphicsPrimitive[] devCollection = deviceFunctions[devIndex];
        int length = devCollection.length;
        if (i < (length - 10)) {
            // "close" - heuristic chosen as a tradeoff between
            // search time and the time needed for reordering the queue
            // and is somewhat application dependent
            GraphicsPrimitive[] temp = new GraphicsPrimitive[length];
            System.arraycopy(devCollection, 0, temp, 0, i);
            System.arraycopy(devCollection, i + 1, temp, i, length - i - 1);
            temp[length-1] = devCollection[i];
            devCollection = temp;
            deviceFunctions[devIndex] = devCollection;
        }
    }

    private static void writeLog(String str) {
        if (debugTrace) {
            System.err.println(str);
        }
    }

    /**
     * Test that all of the GraphicsPrimitiveProxy objects actually
     * resolve to something.  Throws a RuntimeException if anything
     * is wrong, an has no effect if all is well.
     */
    // This is only really meant to be called from GraphicsPrimitiveProxyTest
    // in the regression tests directory, but it has to be here because
    // it needs access to a private data structure.  It's not very
    // big, though.
    public static void testPrimitiveInstantiation() {
	int count = 0;
	for (int i = 0; i < deviceFunctions.length; i++) {
	    GraphicsPrimitive[] prims = deviceFunctions[i];
	    for (int j = 0; j < prims.length; j++) {
		GraphicsPrimitive p = prims[j];
		if (p instanceof GraphicsPrimitiveProxy) {
		    GraphicsPrimitive r = ((GraphicsPrimitiveProxy) p).instantiate();
		    // instantiate checks that p.satisfiesSameAs(r)
		    count++;
		} else {
		    System.out.println(p + " has already been resolved.");
		}
	    }
	}
	System.out.println(count + " graphics primitives resolved correctly.");
    }
}

