/*
 * @(#)GraphicsPrimitiveProxy.java	1.3 98/07/21
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 *
 * CopyrightVersion 1.2
 *
 * @author Bill Foote
 */

package sun.java2d.loops;

/**
 *   GraphicsPrimitiveProxy
 *
 * Acts as a proxy for instances of GraphicsPrimitive, enabling lazy
 * classloading of these primitives.  This leads to a substantial
 * savings in start-up time and footprint.  In the typical case,
 * it has been found that a small number of GraphicsPrimitive instance
 * actually end up getting instantiated.
 * <p>
 * Because a GraphicsPrimitiveProxy is never returned from
 * GraphicsPrimitiveMgr.locate(), proxies don't need to return
 * valid lock flags.
 */


public class GraphicsPrimitiveProxy extends GraphicsPrimitive {

    private Object manager;
    private String relativeClassName;

    /**
     * Create a GraphicsPrimitiveProxy for a primitive with a no-argument
     * constructor.
     *
     * @param manager The manager class for this primitive.  The primitive
     *		must be in the same package as this manager.
     * @param relativeClassName  The name of the class this is a proxy for.
     *		This should not include the package.
     */
    public GraphicsPrimitiveProxy(Object manager, String relativeClassName,
				  int[] surfaceTypes, String methodSignature) {
	super(surfaceTypes, null, methodSignature);
	this.manager = manager;
	this.relativeClassName = relativeClassName;
    }

    //
    // Come up with the real instance.  Called from 
    // GraphicsPrimitiveMgr.locate()
    //
    GraphicsPrimitive instantiate() {
	String name = manager.getClass().getPackage().getName() + "."
			+ relativeClassName;
	try {
	    Class clazz = Class.forName(name);
	    GraphicsPrimitive p = (GraphicsPrimitive) clazz.newInstance();
	    if (!satisfiesSameAs(p)) {
		throw new RuntimeException("Primitive " + p 
					   + " incompatible with proxy for " 
					   + name);
	    }
	    return p;
	} catch (ClassNotFoundException ex) {
	    throw new RuntimeException(ex.toString());
	} catch (InstantiationException ex) {
	    throw new RuntimeException(ex.toString());
	} catch (IllegalAccessException ex) {
	    throw new RuntimeException(ex.toString());
	}
	// A RuntimeException should never happen in a deployed JDK, because 
	// the regression test GraphicsPrimitiveProxyTest will catch any
	// of these errors.
    }

}
