/*
 * @(#)LockableRaster.java	1.27 98/09/14
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author Charlton Innovations, Inc.
 */


package sun.java2d.loops;

import java.awt.Rectangle;
import java.awt.image.ColorModel;
import java.awt.image.Raster;
import java.awt.image.DataBuffer;
import java.awt.image.BufferedImage;
import java.awt.image.SampleModel;
import java.awt.image.PixelInterleavedSampleModel;
import java.awt.color.ColorSpace;
import java.awt.Transparency;
import sun.java2d.SunGraphics2D;
import sun.java2d.loops.ImageData;
import sun.java2d.loops.GraphicsPrimitive;

import sun.awt.image.BytePackedRaster;
import sun.awt.image.ByteComponentRaster;
import sun.awt.image.ByteInterleavedRaster;
import sun.awt.image.ShortComponentRaster;
import sun.awt.image.IntegerComponentRaster;
import java.awt.image.IndexColorModel;
import java.awt.image.ComponentColorModel;
import java.awt.image.DirectColorModel;
import java.awt.image.RasterFormatException;
import java.awt.image.WritableRaster;

public final class LockableRaster
{
    private Raster          parentRaster;
    private SunGraphics2D   parent2D;
    private BufferedImage   parentImage;
    private int             parentType;

    private static final int    PARENT_RASTER   = 1;
    private static final int    PARENT_2D       = 2;
    private static final int    PARENT_IMAGE    = 3;

    private static final int    metaFormat = GraphicsPrimitive.TYPE_INT_ARGB;

    private ColorModel  parentCM;

    private ImageData   cachedID;
    public ImageData    id;
    public int          type;
    public String       devID;

    static final int DCM_RED_MASK   = 0x00ff0000;
    static final int DCM_GREEN_MASK = 0x0000ff00;
    static final int DCM_BLUE_MASK  = 0x000000ff;
    static final int DCM_ALPHA_MASK = 0xff000000;
    static final int DCM_565_RED_MASK = 0xf800;
    static final int DCM_565_GRN_MASK = 0x07E0;
    static final int DCM_565_BLU_MASK = 0x001F;
    static final int DCM_555_RED_MASK = 0x7C00;
    static final int DCM_555_GRN_MASK = 0x03E0;
    static final int DCM_555_BLU_MASK = 0x001F;
    static final int DCM_BGR_RED_MASK = 0x0000ff;
    static final int DCM_BGR_GRN_MASK = 0x00ff00;
    static final int DCM_BGR_BLU_MASK = 0xff0000;

    public LockableRaster(Raster parentRaster, ColorModel parentCM) {
        this.parentRaster   = parentRaster;
        this.parent2D       = null;
        this.parentImage    = null;
        this.parentType     = PARENT_RASTER;
        this.id             = null;
        this.cachedID       = null;
        this.parentCM       = parentCM;
        this.type           = findRasterType(this.parentRaster,this.parentCM);
        this.devID          = null;
    }
    public LockableRaster(SunGraphics2D parent2D) {
        this.parentRaster   = null;
        this.parent2D       = parent2D;
        this.parentImage    = null;
        this.parentType     = PARENT_2D;
        Rectangle clip      = parent2D.getCompBounds();
        this.id             = parent2D.lock(clip.x, clip.y, 1, 1,
                                        GraphicsPrimitive.LOCK_FOR_WRITE);
        this.cachedID       = null;
        this.parentCM       = null;
        this.type           = GraphicsPrimitive.TYPE_CUSTOM;
        this.devID          = parent2D.getDevID();
        if (null != this.id) {
            this.type = this.id.getType();
            parent2D.unlock(this.id);
            this.id = null;
        }
    }
    public LockableRaster(BufferedImage parentImage) {
        this.parentRaster   = parentImage.getRaster();
        this.parent2D       = null;
        this.parentImage    = parentImage;
        this.parentType     = PARENT_IMAGE;
        this.id             = null;
        this.cachedID       = null;
        this.parentCM       = parentImage.getColorModel();
        this.type           = parentImage.getType();
        this.devID          = null;
        if (this.type == GraphicsPrimitive.TYPE_CUSTOM) {
            this.type = findRasterType(this.parentRaster, this.parentCM);
        } else {
            this.type = adjustCM(this.type, parentCM);
        }
    }
    public LockableRaster(int w, int h) {
        this.parent2D       = null;
        this.parentImage    = null;
        this.parentType     = PARENT_RASTER;
        this.id             = null;
        this.cachedID       = null;
        this.parentCM       = ColorModel.getRGBdefault();
        this.type           = metaFormat;
        this.devID          = null;
        this.parentRaster   =
            this.parentCM.createCompatibleWritableRaster(w, h);
    }

    public boolean isTranslucent() {
        if (parentCM == null) {
            return false;
        } else {
            return (parentCM.getTransparency() == Transparency.TRANSLUCENT);
        }
    }

    public boolean containsTransparentPixels() {
        if (parentCM == null) {
            return false;
        } else {
            return (Transparency.BITMASK == parentCM.getTransparency());
        }
    }

    public void lock(int lockFlags, int sX, int sY, int w, int h)
                                            throws RasterLockingException
    {
        switch (parentType) {
        case PARENT_RASTER:
        case PARENT_IMAGE:
            id = lockRaster(parentRaster, parentCM, sX, sY, w, h, lockFlags);
            break;
        case PARENT_2D:
            id = parent2D.lock(sX, sY, w, h, lockFlags);
            break;
        default:
            id = null;
            break;
        }
        if (null != id) {
            this.type = id.getType();
        } else {
            throw new RasterLockingException("Surface not available/bad request");
        }
    }

    public void unlock() {
        // REMIND: think about nested locking
        switch (parentType) {
        case PARENT_2D:
            parent2D.unlock(id);
            // fall through
        case PARENT_RASTER:
        case PARENT_IMAGE:
            break;
        default:
            break;
        }
        this.id = null;
    }

    private static int adjustCM(int type, ColorModel cm) {
        if (cm.getTransparency() == Transparency.BITMASK) {
            switch (type) {
            case GraphicsPrimitive.TYPE_BYTE_INDEXED:
                type = GraphicsPrimitive.TYPE_BYTE_INDEXED_BM;
                break;
            case GraphicsPrimitive.TYPE_INT_ARGB:
                type = GraphicsPrimitive.TYPE_INT_ARGB_BM;
                break;
            case GraphicsPrimitive.TYPE_INT_ARGB_PRE:
                type = GraphicsPrimitive.TYPE_INT_ARGB_PRE_BM;
                break;
            }
        } else if (cm instanceof IndexColorModel) {
            IndexColorModel icm = (IndexColorModel) cm;
            if ((type == GraphicsPrimitive.TYPE_BYTE_INDEXED) &&
                (ImageData.isAllOpaqueGrayICM(icm)) &&
                (icm.getPixelSize() == 8)) {
                type = GraphicsPrimitive.TYPE_INDEX8_GRAY;
            }
        }
        return type;
    }

    /**
    *   This method defines the type of the raster given a color model
    */
    private static int findRasterType(Raster raster, ColorModel cm) {
        int imageType   = GraphicsPrimitive.TYPE_CUSTOM;
        int numBands    = raster.getNumBands();

        if ((raster instanceof IntegerComponentRaster) &&
                (numBands == 3 || numBands == 4)) {
            IntegerComponentRaster iraster = (IntegerComponentRaster) raster;
            int pixSize = cm.getPixelSize();
            if (iraster.getPixelStride() == 1 &&
                    cm instanceof DirectColorModel  &&
                    (pixSize == 32 || pixSize == 24) || pixSize == 25) {

                DirectColorModel dcm = (DirectColorModel) cm;
                int redMask     = dcm.getRedMask();
                int greenMask   = dcm.getGreenMask();
                int blueMask    = dcm.getBlueMask();
                
                if ((redMask == DCM_RED_MASK)
                        && (greenMask == DCM_GREEN_MASK)
                        && (blueMask == DCM_BLUE_MASK)) {
                    boolean isAlphaPre = dcm.isAlphaPremultiplied();
                    if (dcm.getAlphaMask() == DCM_ALPHA_MASK) {
                        imageType = (isAlphaPre
                                     ? GraphicsPrimitive.TYPE_INT_ARGB_PRE
                                     : GraphicsPrimitive.TYPE_INT_ARGB);
                    } else if (pixSize == 24) { // No alpha
                        imageType = GraphicsPrimitive.TYPE_INT_RGB;
                    } else if (pixSize == 25) { // Bitmask alpha
                        imageType = GraphicsPrimitive.TYPE_INT_ARGB_BM;
                    }
                } else if ((pixSize == 24)
                        && (redMask == DCM_BGR_RED_MASK)
                        && (greenMask == DCM_BGR_GRN_MASK)
                        && (blueMask == DCM_BGR_BLU_MASK)) {
                    imageType = GraphicsPrimitive.TYPE_INT_BGR;
                }
            }
        } else if ((cm instanceof IndexColorModel) && (numBands == 1)) {
            IndexColorModel icm = (IndexColorModel) cm;
            int pixSize = icm.getPixelSize();
            if (raster instanceof BytePackedRaster) {
                imageType = GraphicsPrimitive.TYPE_BINARY;
            } else if (raster instanceof ByteComponentRaster) {
                /* Check for TYPE_BYTE_INDEXED here.  adjustCM will change
                   the type to TYPE_INDEX8_GRAY if necessary.  It's necessary
                   to check for TYPE_INDEX8_GRAY in adjustCM anyway, since
                   we may be changing a BuffferedImage of TYPE_BYTE_INDEXED
                   to TYPE_INDEX8_GRAY.
                 */
                ByteComponentRaster braster = (ByteComponentRaster) raster;
                if (braster.getPixelStride() == 1 && pixSize <= 8) {
                    
                    imageType = GraphicsPrimitive.TYPE_BYTE_INDEXED;
                    
                }
            } else if (raster instanceof ShortComponentRaster) {
                /* Check for TYPE_INDEX12_GRAY here, since we need the raster
                   and we will be changing from TYPE_CUSTOM, so findRasterType
                   will be called.
                 */
                ShortComponentRaster sraster = (ShortComponentRaster) raster;
                if (sraster.getPixelStride() == 1 && pixSize == 12 &&
                    ImageData.isAllOpaqueGrayICM(icm)) {
                    imageType = GraphicsPrimitive.TYPE_INDEX12_GRAY;
                }
            }
        } else if ((raster instanceof ShortComponentRaster)
                 && (cm instanceof DirectColorModel)
                 && (numBands == 3)) {
            
            DirectColorModel dcm = (DirectColorModel) cm;
            if (dcm.getRedMask() == DCM_565_RED_MASK) {
                if (dcm.getGreenMask() == DCM_565_GRN_MASK &&
                    dcm.getBlueMask()  == DCM_565_BLU_MASK) {
                    imageType = GraphicsPrimitive.TYPE_SHORT_565_RGB;
                }
            } else if (dcm.getRedMask() == DCM_555_RED_MASK) {
                if (dcm.getGreenMask() == DCM_555_GRN_MASK &&
                    dcm.getBlueMask() == DCM_555_BLU_MASK) {
                    
                    imageType = GraphicsPrimitive.TYPE_SHORT_555_RGB;
                }
            }
        } else if ((raster instanceof ShortComponentRaster)
                 && (cm instanceof ComponentColorModel)
                 && (cm.getColorSpace().getType() == ColorSpace.TYPE_GRAY)
                 && (numBands == 1)) {
            
            if (((ShortComponentRaster)raster).getPixelStride() == 1) {
                imageType = GraphicsPrimitive.TYPE_USHORT_GRAY;
            }
            
        } else if (raster instanceof ByteComponentRaster) {
            if ((cm instanceof ComponentColorModel) &&
                (raster.getSampleModel() instanceof
                 PixelInterleavedSampleModel) &&
                (numBands == 3 || numBands == 4)) {
                ComponentColorModel ccm = (ComponentColorModel) cm;
                ByteComponentRaster braster = (ByteComponentRaster) raster;
                int[] offs = braster.getDataOffsets();
                if (ccm.getNumComponents() != numBands) {
                    throw new RasterFormatException(
                                     "Number of components in ColorModel"
                                     + "does not match # in Raster");
                }
                if (braster.getPixelStride() == numBands) {
                    if (offs[0] == numBands-1 &&
                        offs[1] == numBands-2 &&
                        offs[2] == numBands-3) {
                        if (ccm.isAlphaPremultiplied() == false) {
                            if (numBands == 3) {
                                imageType = GraphicsPrimitive.TYPE_3BYTE_BGR;
                            } else if (offs[3] == 0) {
                                imageType = GraphicsPrimitive.TYPE_4BYTE_ABGR;
                            }
                        } else if (numBands == 4 && offs[3] == 0) {
                            imageType = GraphicsPrimitive.TYPE_4BYTE_ABGR_PRE;
                        }
                    } else if (numBands == 3 &&
                               offs[0] == 0 &&
                               offs[1] == 1 &&
                               offs[2] == 2) {
                        imageType = GraphicsPrimitive.TYPE_24BIT_RGB;
                    }
                }
            } else if (cm instanceof ComponentColorModel &&
                       cm.getColorSpace().getType() == ColorSpace.TYPE_GRAY &&
                       numBands == 1 &&
                       ((ByteComponentRaster)raster).getPixelStride() == 1) {
                
                imageType = GraphicsPrimitive.TYPE_BYTE_GRAY;
            }
                
        }

        imageType = adjustCM(imageType, cm);

        return imageType;
    }

    public static boolean prepareImageData(
                        ImageData idData,
                        Raster rRaster, ColorModel cCModel,
                        int type)
    {
        boolean retval = false;

        SampleModel sampleModel = rRaster.getSampleModel();
        int numBands = sampleModel.getNumBands();

        byte  [] baData     = null;
        int   [] naData     = null;
        short [] saData     = null;

        int scanlineStride = 0;
        int pixelStride    = 0;
        int pixPerDat      = 0;
        int bytePerChannel = 0;

        int offset          = 0;
        int offsetX         = 0;
        int offsetY         = 0;
        int rasterW = rRaster.getWidth();
        int rasterH = rRaster.getHeight();

        if (GraphicsPrimitive.TYPE_CUSTOM == type) {
            type = findRasterType(rRaster, cCModel);
        }

        switch(type) {
        case GraphicsPrimitive.TYPE_INT_RGB:
        case GraphicsPrimitive.TYPE_INT_ARGB:
        case GraphicsPrimitive.TYPE_INT_ARGB_BM:
        case GraphicsPrimitive.TYPE_INT_ARGB_PRE:
        case GraphicsPrimitive.TYPE_INT_ARGB_PRE_BM:
        case GraphicsPrimitive.TYPE_INT_BGR:
            IntegerComponentRaster icRaster = (IntegerComponentRaster)rRaster;
            if (null != (naData = icRaster.getDataStorage())) {
                idData.setIntDataArray(naData);

                scanlineStride  = icRaster.getScanlineStride();
                pixelStride     = icRaster.getPixelStride();
                pixPerDat       = 1;
                bytePerChannel  = 4;

                offset          = icRaster.getDataOffset(0);
                offsetX         = (offset % scanlineStride) / pixelStride;
                offsetY         = (offset - offsetX) / scanlineStride;

                retval          = true;
            }
            break;
        case GraphicsPrimitive.TYPE_3BYTE_BGR:
        case GraphicsPrimitive.TYPE_4BYTE_ABGR:
        case GraphicsPrimitive.TYPE_4BYTE_ABGR_PRE:
        case GraphicsPrimitive.TYPE_BYTE_GRAY:
        case GraphicsPrimitive.TYPE_BYTE_INDEXED:
        case GraphicsPrimitive.TYPE_BYTE_INDEXED_BM:
        case GraphicsPrimitive.TYPE_INDEX8_GRAY:
        case GraphicsPrimitive.TYPE_24BIT_RGB:
            ByteComponentRaster bcRaster = (ByteComponentRaster)rRaster;
            if (null != (baData = bcRaster.getDataStorage())) {
                idData.setByteDataArray(baData);

                scanlineStride  = bcRaster.getScanlineStride();
                pixelStride     = bcRaster.getPixelStride();
                pixPerDat       = 1;
                bytePerChannel  = 1;

                offset          = bcRaster.getDataOffset(0);
                offsetX         = (offset % scanlineStride) / pixelStride;
                offsetY         = (offset - offsetX) / scanlineStride;

                retval          = true;
            }
            break;
        case GraphicsPrimitive.TYPE_SHORT_555_RGB:
        case GraphicsPrimitive.TYPE_SHORT_565_RGB:
        case GraphicsPrimitive.TYPE_USHORT_GRAY:
        case GraphicsPrimitive.TYPE_INDEX12_GRAY:
            ShortComponentRaster scRaster = (ShortComponentRaster)rRaster;
            if (null != (saData = scRaster.getDataStorage())) {
                idData.setShortDataArray(saData);

                scanlineStride  = scRaster.getScanlineStride();
                pixelStride     = scRaster.getPixelStride();
                pixPerDat       = 1;
                bytePerChannel  = 2;

                offset          = scRaster.getDataOffset(0);
                offsetX         = (offset % scanlineStride) / pixelStride;
                offsetY         = (offset - offsetX) / scanlineStride;

                retval          = true;
            }
            break;
        case GraphicsPrimitive.TYPE_BINARY:
            BytePackedRaster bpRaster = (BytePackedRaster)rRaster;
            if (null != (baData = bpRaster.getDataStorage())) {
                idData.setByteDataArray(baData);

                scanlineStride  = bpRaster.getScanlineStride();
                pixelStride     = 1;
                pixPerDat       = 8 / bpRaster.getNumberOfBits();
                bytePerChannel  = 1;

                offset          = bpRaster.getDataBitOffset();
                offsetX         = ((offset % (scanlineStride<<3)) /
                                              bpRaster.getNumberOfBits());
                offsetY         = offset / (scanlineStride<<3);

                retval          = true;

            }
            break;
        case GraphicsPrimitive.TYPE_CUSTOM:
            pixPerDat = 1;
            retval = true;
            break;
        default:
            throw new InternalError("Unknown GraphicsPrimitive type");
        }
        if (retval) {
            idData.setType(type);
            if ((null != cCModel) && (cCModel instanceof IndexColorModel)) {

                int mapSize = ((IndexColorModel)cCModel).getMapSize();
                int[]           naLut    = new int[mapSize];
                ((IndexColorModel)cCModel).getRGBs(naLut);

                idData.setLutData(0);
                idData.setIntLutDataArray(naLut);
                idData.setLutDataLength(mapSize);
            }
            
	    idData.setTransparency(cCModel.getTransparency());
            idData.setDataPointer(0);
            idData.setPixelStride(pixelStride);
            idData.setScanlineStride(scanlineStride);
            idData.setNumChannels(numBands);
            idData.setPixelsPerDataUnit(pixPerDat);
            idData.setBytePerChannel(bytePerChannel);
	    idData.setColorModel(cCModel);
	    idData.setRaster(rRaster);
            idData.setXSize(rasterW);
            idData.setYSize(rasterH);
            idData.setOffsetX(offsetX);
            idData.setOffsetY(offsetY);
        }

        return retval;
    }

    // remind: this should move to ImageData setViewArea (default behavior)
    public static boolean prepareRasterBounds(ImageData idData)
    {
	return prepareRasterBounds(idData, 0, 0,
				   idData.getXSize(),
				   idData.getYSize());
    }

    // remind: this should move to ImageData setViewArea (default behavior)
    public static boolean prepareRasterBounds(ImageData idData,
					      int cx, int cy, int cw, int ch)
    {
        int viewX   = idData.getXViewArea();
        int viewY   = idData.getYViewArea();
        int viewDX  = idData.getDXViewArea();
        int viewDY  = idData.getDYViewArea();
        int right   = viewX + viewDX;
        int bottom  = viewY + viewDY;

        if (viewX < cx) {
            viewX = cx;
        }
        if (viewY < cy) {
            viewY = cy;
        }
        if (right > cx + cw) {
            right = cx + cw;
        }
        if (bottom > cy + ch) {
            bottom = cy + ch;
        }
        viewDX = right - viewX;
        viewDY = bottom - viewY;

        // compensate for sub-raster
        viewX += idData.getOffsetX();
        viewY += idData.getOffsetY();
        if ((viewDY > 0) && (viewDX > 0)) {
            idData.setDeviceArea(viewX, viewY, viewDX, viewDY);
            idData.setOutputArea(viewX, viewY, viewDX, viewDY);
            return true;
        }
        return false;
    }

    /**
     * The method prepares locks for a Raster to make it accessible
     * for direct operation on graphics content by GraphicsPrimitives
     */
    private ImageData lockRaster(Raster theRaster, ColorModel srcCM,
                                 int xView,int yView,int dxView, int dyView,
                                 int lockFlags)
    {
        if ((null == srcCM) || (null == theRaster)) {
            return null;
        }

        ImageData imgData = cachedID;
        if (null == imgData) {
            imgData = new ImageData();
            if (!prepareImageData(imgData, theRaster, srcCM, this.type)) {
                return null;
            }
            this.type = imgData.getType();
        }
        imgData.setViewArea(xView, yView, dxView, dyView);
        imgData.setLockMethod(lockFlags);
        if (prepareRasterBounds(imgData)) {
            cachedID = imgData;
            return imgData;
        }
        return null;
    }

    public String toString() {
        String str = "LockableRaster[";
        switch (parentType) {
        case PARENT_RASTER:
            str += "Raster[";
            str += parentRaster + "[";
            break;
        case PARENT_2D:
            str += "2D[";
            str += parent2D + "[";
            str += this.id + "[";
            str += GraphicsPrimitive.typeStr(this.type);
            break;
        case PARENT_IMAGE:
            str += "Image[";
            str += parentImage + "[";
            break;
        }
        if (id != null) {
            str += id;
        }
        return str;
    }
}
