/*
 * @(#)RasterOutputManager.java	1.106 98/10/13
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author Charlton Innovations, Inc.
 */

package sun.java2d.loops;

import sun.java2d.loops.GraphicsPrimitiveMgr;
import sun.awt.image.BufferedImageGraphics2D;
import sun.awt.image.IntegerComponentRaster;
import java.awt.image.ColorModel;
import java.awt.AlphaComposite;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import java.awt.Color;
import java.awt.BasicStroke;
import java.awt.Font;
import java.awt.font.GlyphVector;
import java.awt.image.DirectColorModel;
import java.awt.image.IndexColorModel;
import java.awt.image.BufferedImage;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.Transparency;
import java.awt.Shape;
import java.awt.font.FontRenderContext;
import java.awt.geom.AffineTransform;
import java.awt.geom.Arc2D;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Line2D;
import java.awt.geom.RoundRectangle2D;
import java.awt.geom.GeneralPath;
import java.awt.Composite;
import java.awt.CompositeContext;
import java.awt.RenderingHints;
import sun.java2d.pipe.AlphaColorPipe;
import sun.java2d.pipe.CompositePipe;
import sun.java2d.pipe.PixelDrawPipe;
import sun.java2d.pipe.PixelFillPipe;
import sun.java2d.OutputManager;
import sun.java2d.SunOutputManager;
import sun.java2d.SunGraphics2D;
import sun.java2d.DeviceClipping;
import sun.java2d.pipe.TranslateablePipe;
import sun.java2d.pipe.LineStylePipe;
import sun.java2d.pipe.SimpleRenderPipe;
import sun.java2d.pipe.ShapeToPolyConverter;
import sun.java2d.pipe.DrawToFillConverter;
import sun.java2d.pipe.SpanShapeRenderer;
import sun.java2d.pipe.SpanClipRenderer;
import sun.java2d.pipe.PixelToShapeConverter;
import sun.java2d.pipe.TextPipe;
import sun.java2d.pipe.SpanIterator;
import sun.java2d.HackWorkAround2DIndexed;
import sun.awt.SunHints;
import sun.awt.font.NativeFontWrapper;
import sun.security.action.GetPropertyAction;
import java.lang.ref.WeakReference;

/**
 * The RasterOutputManager is the default endpoint for many of the
 * rendering pipeline operations.  It implements the entry layer
 * of a multiple layered architecture that routes requests for primitive
 * graphic functions to appropriate handler implementations.
 *
 * Information obtained from the Destination Raster surface, the
 * Source Raster surface (in the case of a Binary operation), and the
 * arguments supplied by the caller are used to create a unique
 * component signature which is sent to the Graphic Component Manager
 * to locate the appropriate handler for the specified operation.
 *
 * All handlers implement one public method which receives arguments
 * specific to the caller's context.
 *
 * If no appropriate handler is found for a binary operation, the
 * RasterOutputManager converts the source pixels to a canonical
 * form by a BLT operation from the source to the common meta-format.  The
 * Binary operation is retried with the new source format, and the original
 * destination format.
 *
 * If no appropriate handler is found for a unary operation, or no appropriate
 * handler is found during the initial retry of a binary operation, the
 * RasterOutputManager converts the destination pixels to a canonical form
 * by a BLT operation from the destination raster to the common meta-format.
 * The operation is performed by the meta-format primitives (which are
 * defined to be available) and then the result is BLTed to the destination
 * raster.
 *
 * There are thread synchronization issues involved in the use of source
 * and destination rasters.  The RasterOutputManager utilizes a locking
 * protocol which locks the destination (and source, for binary operations)
 * raster(s) for the duration of the rendering operation.  For binary
 * operations, the correct order of locking the two surfaces is critical
 * to prevent possible deadlocks between threads which are competing for
 * the same surfaces.  The only way to insure consistent ordering of the
 * locks is to insure that all surfaces have a unique immutable
 * identification which is sortable.
 *
 * If the source and destination surfaces are the same raster, or sub-rasters
 * of the same ultimate parent raster, RasterOutputManager adjusts the
 * locking of the surface accordingly.
 *
 * Some caller requests can be reduced to a canonical form by the recognition
 * of the equivalance between certain operations. (Like SRC_OVER BLT from
 * an Opaque Source, with no blending, is the same as SRC operation)
 *
 */

public class RasterOutputManager
    extends SunOutputManager
    implements SimpleRenderPipe,
        PixelDrawPipe,
        PixelFillPipe,
        TextPipe
{
    private static final boolean debug = false;
    private static boolean debugTrace = false;
    private static boolean debugLog = false;
    private static boolean usePlatformFont = false;
    private static RasterOutputManager manager = new RasterOutputManager();

    static final int metaFormat = GraphicsPrimitive.TYPE_INT_ARGB;
    static final int defaultFormat = GraphicsPrimitive.TYPE_CUSTOM;
    static final String ditherID = OpaqueDitheredBlit.getMethodSignature();
    static final String copyID   = OpaqueBlit.getMethodSignature();

    public static final RasterOutputManager getManager()
    {
        return manager;
    }

    public static boolean usesPlatformFont()
    {
        return usePlatformFont;
    }

    static native boolean getPlatformFontVar();

    private RasterOutputManager()
    {
        GraphicsPrimitiveMgr.register(new CustomComponent(), null);
        GraphicsPrimitiveMgr.register(new GeneralRenderer(), null);
        GraphicsPrimitiveMgr.register(new DefaultComponent(), null);
        GraphicsPrimitiveMgr.register(new ShortCompositing(), null);
        GraphicsPrimitiveMgr.register(new IndexedCompositing(), null);
        GraphicsPrimitiveMgr.register(new GIFAcceleratorLoops(), null);
        GraphicsPrimitiveMgr.register(new XBGRLoops(), null);
        GraphicsPrimitiveMgr.register(new IntDiscreteRenderer(), null);
        GraphicsPrimitiveMgr.register(new ByteDiscreteRenderer(), null);
        GraphicsPrimitiveMgr.register(new ShortDiscreteRenderer(), null);
        GraphicsPrimitiveMgr.register(new ThreeByteRenderer(), null);
        GraphicsPrimitiveMgr.register(new Gray8Blters(), null);
        GraphicsPrimitiveMgr.register(new Gray16Blters(), null);
        GraphicsPrimitiveMgr.register(new TextRendering(), null);
        GraphicsPrimitiveMgr.register(new IndexGrayAccelerators(), null);
    }

    public static final ColorModel XRGBModel =
        new DirectColorModel(24,0x00ff0000,0x0000ff00,0x000000ff);
    public static final ColorModel XBGRModel =
        new DirectColorModel(24,0x000000ff,0x0000ff00,0x00ff0000);

    static {
	    // REMIND: Get rid of this when we decide to go with the
            // new architecture.  Since this is a temporary property
            // the property naming conventions were ignored for simplicity.
	    String prop = (String) java.security.AccessController.doPrivileged(
                    new GetPropertyAction("java2d.newarch.trace"));
            if ("true".equals(prop)) {
                System.out.println("Enabling new 2d architecture trace");
                debugTrace = true;
            } else if ("false".equals(prop)) {
                System.out.println("Disabling new 2d architecture trace");
                debugTrace = false;
            } else if (prop != null) {
                System.out.println("Unrecognized flag for new 2d architecture: "
+ prop);
            }
            // REMIND: Get rid of this when we decide to go with the
            // new architecture.  Since this is a temporary property
            // the property naming conventions were ignored for simplicity.
	    prop = (String) java.security.AccessController.doPrivileged(
                    new GetPropertyAction("java2d.newarch.log"));
            if ("true".equals(prop)) {
                System.out.println("Enabling new 2d architecture logging");
                debugLog = true;
            } else if ("false".equals(prop)) {
                System.out.println("Disabling new 2d architecture logging");
                debugLog = false;
            } else if (prop != null) {
                System.out.println("Unrecognized flag for new 2d architecture: "
+ prop);
            }
            // REMIND: Get rid of this when we decide to go with the
            // new architecture.  Since this is a temporary property
            // the property naming conventions were ignored for simplicity.
	    prop = (String) java.security.AccessController.doPrivileged(
                    new GetPropertyAction("java2d.font.usePlatformFont"));
            if (prop == null) {
                usePlatformFont = getPlatformFontVar();
            }
            if ("true".equals(prop) || usePlatformFont) {
                usePlatformFont = true;
            } else if ("false".equals(prop) || !usePlatformFont) {
                usePlatformFont = false;
            } else if (prop != null) {
                System.out.println("Unrecognized flag for platform fonts: "+
		                   prop);
            }
            if (prop != null || usePlatformFont) {
                if (usePlatformFont) {
                    System.out.println("Enabling platform fonts. NOTE: usePlatformFont is now deprecated and will be removed in the next build.");
                } else {
                    System.out.println("Disabling platform fonts");
                }
            }
    }

    /**
     * Construct the necessary PixelDrawPipe, PixelFillPipe, and
     * ShapeDrawPipe objects necessary for the given SunGraphics2D
     * object to respond to the various rendering methods.
     * This override constructs a pipeline which takes advantage
     * of the X11 rendering capabilities of an X11Graphics object
     * if it can, otherwise it lets the superclass define a generic
     * set of pipelines.
     */
    public void validatePipe(SunGraphics2D sg) {
        sg.simpleText = false;
	boolean simpleFill =
	    ((sg.paintState == SunGraphics2D.PAINT_SOLIDCOLOR) &&
	     ((sg.compositeState == SunGraphics2D.COMP_ISCOPY) ||
             (sg.compositeState == SunGraphics2D.COMP_XOR)) &&
             !((sg instanceof BufferedImageGraphics2D) &&
             (sg.compositeState == SunGraphics2D.COMP_XOR)));
	boolean simpleClip = (sg.clipState != SunGraphics2D.CLIP_SHAPE);
        
	if (sg.antialiasHint != SunHints.INTVAL_ANTIALIAS_ON ||
            sg.modeState == SunGraphics2D.XOR_MODE)
	{
            constructAlphaPipe(sg);
	    if (simpleFill) {
		SimpleRenderPipe srp;
		if (sg instanceof SimpleRenderPipe) {
		    srp = (SimpleRenderPipe) sg;
		} else {
		    srp = this;
		}
		if (simpleClip) {
                    if (sg.compositeState != SunGraphics2D.COMP_XOR) {
                        sg.simpleText = true;
                    }
		    constructAliasedPipe(sg, srp);
		} else {
		    if (srp instanceof TranslateablePipe) {
			((TranslateablePipe)sg).setOrigin(0,0);
		    }
		    if (srp instanceof LineStylePipe) {
			((LineStylePipe) srp).setLineAttrs
			    (0, 0, null, 0,
			     BasicStroke.CAP_BUTT,
			     BasicStroke.JOIN_BEVEL);
		    }
		    sg.shapepipe = new SpanShapeRenderer.Simple(srp);
		    PixelToShapeConverter cvt =
			new PixelToShapeConverter(sg.shapepipe);
		    sg.drawpipe = cvt;
		    sg.fillpipe = cvt;
		}
            } else {
		if (sg instanceof TranslateablePipe) {
		    ((TranslateablePipe)sg).setOrigin(0,0);
		}
		sg.shapepipe = new SpanShapeRenderer.Composite(sg.alphapipe);
		PixelToShapeConverter cvt =
		    new PixelToShapeConverter(sg.shapepipe);
		sg.drawpipe = cvt;
		sg.fillpipe = cvt;
            }
            if (sg instanceof DeviceClipping) {
                if (sg.clipState == SunGraphics2D.CLIP_DEVICE && sg.constrainClip == null) {
                    ((DeviceClipping)sg).removeClip();
                } else {
                    Rectangle r = sg.getCompBounds();
                    ((DeviceClipping)sg).changeClip(r.x, r.y,
                                                    r.width, r.height, true);
                }
            }
        } else {
            if (sg instanceof DeviceClipping) {
                ((DeviceClipping)sg).removeClip();
            }
            if (sg instanceof TranslateablePipe) {
		((TranslateablePipe)sg).setOrigin(0,0);
            }
	    super.validatePipe(sg);
        }
        if (simpleFill && simpleClip) {
            if (sg.compositeState != SunGraphics2D.COMP_XOR) {
                sg.simpleText = true;
            }
            if (sg instanceof TextPipe) {
                sg.textpipe = (TextPipe)sg;
                if (sg instanceof DeviceClipping
						&& !usePlatformFont)
                {
                    // disable JDK 1.1 code paths
                    sg.textpipe = (TextPipe)this;
                }
            } else {
                sg.textpipe = (TextPipe)this;
            }
        } else {
            sg.textpipe = (TextPipe)this;
        }

// 	System.out.println("paintState="+sg.paintState+
//                            " compositeState="+sg.compositeState+
//                            " BIG="+(sg instanceof BufferedImageGraphics2D));
//         System.out.println("simpleFill="+simpleFill+
//                            " simpleClip="+simpleClip+
//                            " simpleText="+ sg.simpleText);
//        System.out.println("draw    "+ sg.drawpipe);
//        System.out.println("alpha   "+ sg.alphapipe);
//        System.out.println("fill    "+ sg.fillpipe);
//        System.out.println("shape   "+ sg.shapepipe);
//        System.out.println("text    "+ sg.textpipe);
    }

    private class RenderImageCachedState {
        private SunGraphics2D   dest2D;
        private LockableRaster  lrDst;
        private Object          opState[];
		// REMIND: use WeakReference for following...
        private WeakReference   srcImage[];
        private WeakReference   lrSrc[];

        RenderImageCachedState(
                SunGraphics2D dest2D
        ) {
            this.dest2D = dest2D;
        }
		LockableRaster getDstLR() {
			if (null == this.lrDst) {
	            this.lrDst = new LockableRaster(dest2D);
			}
			return this.lrDst;
		}
		LockableRaster getSrcLR(BufferedImage si) {
			int index = getIndex(si);
			WeakReference wr = lrSrc[index];
			LockableRaster lr = null;
			if (null != wr) {
				lr = (LockableRaster)wr.get();
			}
			if (null == lr) {
				lr = new LockableRaster(si);
				lrSrc[index] = new WeakReference(lr);
			}
			return lr;
		}
		void setOpState(BufferedImage si, Object state) {
			int index = getIndex(si);
			opState[index] = state;
		}
		Object getOpState(BufferedImage si) {
			int index = getIndex(si);
			return opState[index];
		}
		boolean validate(SunGraphics2D sg2d) {
			return (sg2d == this.dest2D);
		}
		private synchronized int getIndex(BufferedImage si) {
			if (null == this.srcImage) {
				this.srcImage = new WeakReference[10];
				this.opState  = new Object[10];
				this.lrSrc = new WeakReference[10];
				this.srcImage[0] = new WeakReference(si);
				return 0;
			}
			for (int i=0; i < this.srcImage.length; i++) {
				WeakReference wr = this.srcImage[i];
				if (null == wr) {
					continue;
				}
				BufferedImage bi = (BufferedImage)wr.get();
				if (bi == si) {
					return i;
				}
			}
			for (int i=0; i < this.srcImage.length; i++) {
				WeakReference wr = this.srcImage[i];
				BufferedImage bi = null;
				if (null != wr) {
					bi = (BufferedImage)wr.get();
				}
				if (bi == null) {
					this.srcImage[i] = new WeakReference(si);
					return i;
				}
			}
			int oldSize = this.srcImage.length;
			int newSize = oldSize + 10;
			//
			Object state[] = new Object[newSize];
			System.arraycopy(this.opState, 0, state, 0,  oldSize);
			this.opState = state;
			//
			WeakReference wr[] = new WeakReference[newSize];
			System.arraycopy(this.lrSrc, 0, wr, 0,  oldSize);
			this.lrSrc = wr;
			//
			wr = new WeakReference[newSize];
			System.arraycopy(this.srcImage, 0, wr, 0,  oldSize);
			this.srcImage = wr;
			//
			this.srcImage[oldSize] = new WeakReference(si);
			return oldSize;
		}
    }

    /**
     * Render source image pixels from a BufferedImage onto destination
     * SunGraphics2D using:
     *  1) clip shape tile
     *  2) background color
     *  3) source pixels from a BufferedImage
     *
     *  may return an Object which contains cached state which can
     *  speed up further operations in the same context
     *
     *  uses:
     *  1) BinaryComposite
     *  2) BinaryCompositeBg
     *  3) OpaqueBlit
     *  4) OpaqueBlitBg
     *  5) OpaqueDitheredBlit
     *  6) OpaqueDitheredBlitBg
     */
    public Object renderImage(SunGraphics2D dest2D,
                              BufferedImage srcImage,
                              Color bkColor,
                              int dstX, int dstY,
                              int outWidth, int outHeight,
                              byte[] clipAlpha, int alphaScan,
                              Object cachedState) {
        RenderImageCachedState rics = null;
        if (null != cachedState) {
            rics = (RenderImageCachedState)cachedState;
			if (!rics.validate(dest2D)) {
				rics = new RenderImageCachedState(dest2D);
			}
		} else {
			rics = new RenderImageCachedState(dest2D);
		}
        LockableRaster dst = rics.getDstLR();
        LockableRaster src = rics.getSrcLR(srcImage);
	rics.setOpState(srcImage,
			compositeSrcDst(src, dst, dest2D.renderHint,
                                        dest2D.compositeState,
                                        dest2D.getComposite(),
                                        clipAlpha, 0, alphaScan, 0, 0,
                                        dstX, dstY, outWidth, outHeight,
                                        bkColor, rics.getOpState(srcImage)));
        return rics;
    }

    private class PaintLoopCachedState {
        SunGraphics2D   dest2D;
        Raster          srcRaster;
        ColorModel      srcCM;
        LockableRaster  lrSrc;
        LockableRaster  lrDst;
        Object          opState;
        PaintLoopCachedState(
                SunGraphics2D   dest2D,
                Raster srcRaster,
                ColorModel srcCM,
                LockableRaster lrSrc,
                LockableRaster lrDst
        ) {
            this.dest2D = dest2D;
            this.srcRaster = srcRaster;
            this.srcCM = srcCM;
            this.lrSrc = lrSrc;
            this.lrDst = lrDst;
            this.opState = null;
        }

        public boolean validate(
                SunGraphics2D   dest2D,
                Raster srcRaster,
                ColorModel srcCM
        ) {
            return (    (this.dest2D == dest2D)
                        && this.srcRaster.equals(srcRaster)
                        && this.srcCM.equals(srcCM));
        }
    }

    private class CopyAreaCachedState {
        SunGraphics2D   dest2D;
        LockableRaster  lrSrc;
        LockableRaster  lrDst;
        Object          opState;
        CopyAreaCachedState(SunGraphics2D dest2D, LockableRaster lrSrc,
                            LockableRaster lrDst) {
            this.dest2D = dest2D;
            this.lrSrc = lrSrc;
            this.lrDst = lrDst;
            this.opState = null;
        }

        public boolean validate(SunGraphics2D dest2D) {
            return (this.dest2D == dest2D);
        }
    }

    /**
     * Perform copyArea function
     */
    public Object copyArea(SunGraphics2D dest2D, byte[] clipAlpha,
                           int alphaScan, int alphaOffset,
                           int x, int y, int w, int h, int dx, int dy,
                           Object cachedState) {
        
        LockableRaster src = null;
        LockableRaster dst = null;;
        CopyAreaCachedState cacs = null;
        
        if (null != cachedState) {
            cacs = (CopyAreaCachedState)cachedState;
            if (cacs.validate(dest2D)) {
                src = cacs.lrSrc;
                dst = cacs.lrDst;
            } else {
                cacs = null;
            }
        }
        
        if (null == src) {
            src = new LockableRaster(dest2D);
        }
        if (null == dst) {
            dst = new LockableRaster(dest2D);
        }

        if (cacs == null) {
            cacs = new CopyAreaCachedState(dest2D, src, dst);
        }
        
        int srcX = x;
        int srcY = y;
        ColorModel srcCM = dest2D.getDeviceConfiguration().getColorModel();
        int renderMode = dest2D.renderHint;
        
        // If there is a forward overlap of the memory area, copy to an
        // intermediate buffer first
        if (
            (dest2D instanceof BufferedImageGraphics2D
                && ((dx >= x) && (dx <= (x+w)) && (dy >= y) && (dy <= (y+h))))
            || "WGraphics".equals(dest2D.getDevID())
        ) {
            BufferedImage interm = new BufferedImage(w, h, metaFormat);
            SunGraphics2D intermSrc2D = (SunGraphics2D)interm.createGraphics();
            LockableRaster intermSrc = new LockableRaster(intermSrc2D);
            compositeSrcDst(src, intermSrc, renderMode,
                            SunGraphics2D.COMP_ISCOPY,
                            AlphaComposite.Src, null, alphaOffset, alphaScan,
                            srcX, srcY, 0, 0, w, h, null, null);
            srcX = 0;
            srcY = 0;
            srcCM = interm.getColorModel();
            src = intermSrc;
        }
        
        int dstX = x+dx;
        int dstY = y+dy;
        int outWidth=w;
        int outHeight=h;
        cacs.opState = compositeSrcDst(src, dst, renderMode,
                                       dest2D.compositeState,
                                       dest2D.getComposite(),
                                       clipAlpha, alphaOffset, alphaScan,
                                       srcX, srcY, dstX, dstY, outWidth,
                                       outHeight, null, cacs.opState);
        return cacs;
    }

    private Object compositeSrcDst(LockableRaster src,
                                   LockableRaster dst,
                                   int renderMode,
                                   int compositeState,
                                   Composite composite,
                                   byte[] clipAlpha, int alphaOffset,
                                   int alphaScan,
                                   int srcX, int srcY,
                                   int dstX, int dstY,
                                   int outWidth, int outHeight,
                                   Color bkColor,
                                   Object cachedState) {

        boolean alphaComp = (composite instanceof AlphaComposite);
        boolean xorComp   = (composite instanceof XORComposite);
        
        if ((null == clipAlpha)
            && (SunGraphics2D.COMP_ISCOPY == compositeState)
            && !src.isTranslucent()) {  // REMIND: why bother testing?
            if (src.containsTransparentPixels()) {
                if (bkColor == null) {
                    if (true || SunHints.INTVAL_RENDER_SPEED == renderMode) {
                        return performTransparentBlit(src, dst,
                                                      srcX, srcY, dstX, dstY,
                                                      outWidth, outHeight,
                                                      cachedState);
                    } else {
                        return performTransparentDitheredBlit(src, dst,
                                                              srcX, srcY,
                                                              dstX, dstY,
                                                              outWidth,
                                                              outHeight,
                                                              cachedState);
                    }
                } else {
                    if (true || SunHints.INTVAL_RENDER_SPEED == renderMode) {
                        return performOpaqueBlitBg(src, dst, bkColor,
                                                   srcX, srcY, dstX, dstY,
                                                   outWidth, outHeight,
                                                   cachedState);
                    } else {
                        return performOpaqueDitheredBlitBg(src, dst,
                                                           bkColor,
                                                           srcX, srcY,
                                                           dstX, dstY,
                                                           outWidth, outHeight,
                                                           cachedState);
                    }
                }
            } else {
                if (true || SunHints.INTVAL_RENDER_SPEED == renderMode) {
                    return performOpaqueBlit(src, dst,
                                             srcX, srcY, dstX, dstY,
                                             outWidth, outHeight,
                                             cachedState);
                } else {
                    return performOpaqueDitheredBlit(src, dst,
                                                     srcX, srcY, dstX, dstY,
                                                     outWidth, outHeight,
                                                     cachedState);
                }
            }
        } else if (alphaComp || xorComp) {
            int alphaRule = 0;
            float extraAlpha = 1.0f;
            Color xorColor = Color.black;
            if (alphaComp) {
                AlphaComposite ac = (AlphaComposite)composite;
                alphaRule = ac.getRule();
                extraAlpha = ac.getAlpha();
            } else if (xorComp) {
                XORComposite xc = (XORComposite)composite;
                alphaRule = 0;
                xorColor = xc.getXorColor();
            } else {
                throw new InternalError("Unknown composite");
            }

            if (null != bkColor) {
                return performBinaryCompositeBg(src, dst,
                                                srcX, srcY, dstX, dstY,
                                                outWidth, outHeight,
                                                clipAlpha, alphaOffset,
						alphaScan,
                                                alphaRule, extraAlpha,
                                                bkColor, xorColor,
                                                cachedState);
            } else {
                return performBinaryComposite(src, dst,
                                              srcX, srcY, dstX, dstY,
                                              outWidth, outHeight,
                                              clipAlpha, alphaOffset,
					      alphaScan,
                                              alphaRule, extraAlpha,
                                              xorColor, cachedState);
            }
        } else { /* custom composite */

            BufferedImage srcImg = new BufferedImage(outWidth, outHeight,
                                                     metaFormat);
            LockableRaster srcLR = new LockableRaster(srcImg);
            compositeSrcDst(src, srcLR, renderMode,
                            SunGraphics2D.COMP_ISCOPY,
                            AlphaComposite.Src, null, alphaOffset,
                            alphaScan, srcX, srcY, 0, 0,
                            outWidth, outHeight, null, null);
            Raster srcRaster = srcImg.getRaster();

            BufferedImage devImg = new BufferedImage(outWidth, outHeight,
                                                     metaFormat);
            WritableRaster devRaster = devImg.getRaster();
            LockableRaster devLR = new LockableRaster(devImg);
            compositeSrcDst(dst, devLR, renderMode,
                            SunGraphics2D.COMP_ISCOPY,
                            AlphaComposite.Src, null, alphaOffset,
                            alphaScan, dstX, dstY, 0, 0,
                            outWidth, outHeight, null, null);
            srcX = 0;
            srcY = 0;

            RenderingHints hints =
                new RenderingHints(RenderingHints.KEY_RENDERING,
                                   (renderMode ==
				    SunHints.INTVAL_RENDER_SPEED) ?
                                   RenderingHints.VALUE_RENDER_SPEED :
                                   RenderingHints.VALUE_RENDER_QUALITY);
            
            CompositeContext compCtxt =
                composite.createContext(srcImg.getColorModel(),
                                        devImg.getColorModel(),
                                        hints);
            
            compCtxt.compose(srcRaster, devRaster, devRaster);

            return compositeSrcDst(devLR, dst, renderMode,
                                   SunGraphics2D.COMP_ISCOPY,
                                   AlphaComposite.Src, clipAlpha, alphaOffset,
                                   alphaScan, 0, 0, dstX, dstY,
                                   outWidth, outHeight, null, cachedState);
        }
    }

    /**
     * Render source image pixels from a Raster onto destination
     * SunGraphics2D using:
     *  1) clip shape tile
     *  2) specified alphaCompositeRule (instead of destination AC)
     *  3) source pixels from a Raster with ColorModel, premul flag
     *  4) xorColor
     *
     *  may return an Object which contains cached state which can
     *  speed up further operations in the same context
     *
     *  uses:
     *  1) BinaryComposite
     *  2) OpaqueBlit
     *  3) OpaqueDitheredBlit
     */
    public Object compositePaintLoop(SunGraphics2D dest2D, byte[] clipAlpha,
                                     int alphaOffset, int alphaScan,
                                     int dstX, int dstY, int outWidth,
                                     int outHeight, ColorModel srcCM,
                                     Raster srcRaster, Object cachedState) {

        LockableRaster src = null;
        LockableRaster dst = null;
        PaintLoopCachedState plcs = null;
        
        if (null != cachedState) {
            plcs = (PaintLoopCachedState)cachedState;
            if (plcs.validate(dest2D, srcRaster, srcCM)) {
                src = plcs.lrSrc;
                dst = plcs.lrDst;
            } else {
                plcs = null;
            }
        }
        
        if (null == src) {
            src = new LockableRaster(srcRaster, srcCM);
        }
        if (null == dst) {
            dst = new LockableRaster(dest2D);
        }
        if ((null == dst) || (null == src)) {
            throw new InternalError("Src="+src+" Dst="+dst);
        }

        if (plcs == null) {
            plcs = new PaintLoopCachedState(
                            dest2D, srcRaster, srcCM, src, dst);
        }
        
        plcs.opState =  compositeSrcDst(src, dst, dest2D.renderHint,
                                        dest2D.compositeState,
                                        dest2D.getComposite(),
                                        clipAlpha, alphaOffset, alphaScan,
                                        0, 0, dstX, dstY, outWidth, outHeight,
                                        null, plcs.opState);
        return plcs;
    }

    private class UnaryCachedState {
        SunGraphics2D   the2D;
        LockableRaster  theLR;
        PerformUnaryGraphicsPrimitive   ugp;
        UnaryCachedState(
                SunGraphics2D   the2D,
                LockableRaster  theLR,
                PerformUnaryGraphicsPrimitive   ugp) {
            this.the2D  = the2D;
            this.theLR  = theLR;
            this.ugp    = ugp;
        }
    }

    /**
    *   PerformBinaryGraphicsPrimitive
    *   encapsulates behavior of strategies for optimal performance of
    *   requested operation when either or both surfaces may need to
    *   be converted to an intermediate for the operation to succeed
    *
    *   Strategies in decreasing preference order:
    *   1) A op B
    *   2) A to A', A' op B
    *   3) B to B', A op B', B' to B
    *   4) A to A', B to B', A' op B', B' to B
    */
    private class PerformBinaryGraphicsPrimitive {
        GraphicsPrimitive   coreOp;
        OpaqueBlit          convertSrcOp;
        OpaqueBlit          readDstOp;
        OpaqueBlit          writeDstOp;

        String id;
        int srcType;
        int dstType;
        String deviceID;

        public PerformBinaryGraphicsPrimitive(
                                    String id, int srcType, int dstType,
                                    String deviceID)
        {
            this.id = id;
            this.srcType = srcType;
            this.dstType = dstType;
            this.deviceID = deviceID;

            // first strategy:  A op B
            int [] types = {srcType, dstType};
            coreOp = GraphicsPrimitiveMgr.locate(id, types, deviceID);
            if (null != coreOp) {
                if(debugLog) {
                    System.err.println("First " + srcType + "->" + dstType); 
                }
                return;
            }
            // second strategy: A to A', A' op B
            int [] types2 = {metaFormat, dstType};
            coreOp = GraphicsPrimitiveMgr.locate(id, types2, deviceID);
            if (null != coreOp) {
                convertSrcOp = createConverter(srcType, metaFormat);
                if (null != convertSrcOp) {
                    if (debugLog) {
                        System.err.println("Binary Strategy 2 " + srcType + "->" + metaFormat + "->" + dstType); 
                    }
                    return;
                }
            }
            // third strategy:  B to B', A op B', B' to B
            int [] types3 = {srcType, metaFormat};
            coreOp = GraphicsPrimitiveMgr.locate(id, types3, deviceID);
            if (null != coreOp) {
                readDstOp = createConverter(dstType, metaFormat);
                writeDstOp = createConverter(metaFormat, dstType);
                if ((null == readDstOp) || (null == writeDstOp)) {
                    reportError(id, srcType, dstType);
                    coreOp = null;
                    readDstOp = null;
                    writeDstOp = null;
                } else {
                    if (debugLog) {
                        System.err.println("Binary Strategy 3 " + srcType + "->" + metaFormat + + dstType + "->" + metaFormat); 
                    }
                    return;
                }
            }
            // fourth strategy: A to A', B to B', A' op B', B' to B
            int [] types4 = {metaFormat, metaFormat};
            coreOp = GraphicsPrimitiveMgr.locate(id, types4, deviceID);
            if (null != coreOp) {
                convertSrcOp = createConverter(srcType, metaFormat);
                readDstOp = createConverter(dstType, metaFormat);
                writeDstOp = createConverter(metaFormat, dstType);
                if ((null == convertSrcOp)
                        || (null == readDstOp)
                        || (null == writeDstOp)) {
                    reportError(id, srcType, dstType);
                    coreOp = null;
                    convertSrcOp = null;
                    readDstOp = null;
                    writeDstOp = null;
                } else {
                    if (debugLog) {
                        System.err.println("Binary Strategy 4 " + srcType + " " + metaFormat + " " + dstType);
                    }
                    return;
                }
            }
            reportError(id, srcType, dstType);
            return;
        }

        private OpaqueBlit createConverter(int srcType, int dstType) {
            GraphicsPrimitive converter;
            // REMIND: need to check RENDER_SPEED somehow before dithering
            int [] types = {srcType, dstType};
//         converter = GraphicsPrimitiveMgr.locate(ditherID, types, deviceID);
//            if (null != converter) {
//                return (OpaqueBlit)converter;
//            }
            converter = GraphicsPrimitiveMgr.locate(copyID, types, deviceID);
            if (null != converter) {
                return (OpaqueBlit)converter;
            }
            int [] types2 = {defaultFormat, defaultFormat};
            converter = GraphicsPrimitiveMgr.locate(copyID, types2, deviceID);
            return (OpaqueBlit)converter;
        }

        private void reportError(String id, int srcType, int dstType) {
            if (!debugTrace) {
                return;
            }
            System.err.println("Binary Graphics Operation failed:");
            System.err.println(" primitive: [" + id + "]");
            System.err.println(" surfaces: ["+srcType+"]["+dstType+"]");
            System.err.println(" surfaces: [" +
                                GraphicsPrimitive.typeStr(srcType) + "]["
                                + GraphicsPrimitive.typeStr(dstType) + "]");
        }
    }

    protected Object performOpaqueBlit(LockableRaster src, LockableRaster dst,
                    int srcX, int srcY, int dstX, int dstY, int w, int h,
                    Object cachedState)
    {
        PerformBinaryGraphicsPrimitive bgp = null;
        if (null != cachedState) {
            bgp = (PerformBinaryGraphicsPrimitive)cachedState;
            if (!bgp.id.equals(OpaqueBlit.getMethodSignature())
                    || (bgp.srcType != src.type)
                    || (bgp.dstType != dst.type)) {
                bgp = null;
            }
        }
        if (null == bgp) {
            bgp = new PerformBinaryGraphicsPrimitive(
                            OpaqueBlit.getMethodSignature(),
                            src.type, dst.type, dst.devID);
        }
        if ((null == bgp) || (null == bgp.coreOp)) {
            return null;
        }
        LockableRaster  coreSrc = src;
        LockableRaster  coreDst = dst;
        int sX = srcX;
        int sY = srcY;
        int dX = dstX;
        int dY = dstY;
        if (null != bgp.convertSrcOp) {
            coreSrc = convertFrom(bgp.convertSrcOp, src, srcX, srcY, w, h);
            sX = 0;
            sY = 0;
        }
        if (null != bgp.readDstOp) {
            coreDst = convertFrom(bgp.readDstOp, dst, dstX, dstY, w, h);
            dX = 0;
            dY = 0;
        }
        try {
            OpaqueBlit ob = (OpaqueBlit)bgp.coreOp;
            int [] lockflags = ob.getLockFlags();
            coreSrc.lock(lockflags[0], sX, sY, w, h);
            coreDst.lock(lockflags[1], dX, dY, w, h);
            ob.OpaqueBlit(coreSrc.id, coreDst.id, w, h);
        } catch (RasterLockingException e) {
        } finally {
            coreDst.unlock();
            coreSrc.unlock();
        }
        if (null != bgp.writeDstOp) {
            convertTo(bgp.writeDstOp, dst, dstX, dstY, w, h, coreDst);
        }
        return bgp;
    }

    protected Object performOpaqueDitheredBlit(
                    LockableRaster src, LockableRaster dst,
                    int srcX, int srcY, int dstX, int dstY, int w, int h,
                    Object cachedState)
    {
        PerformBinaryGraphicsPrimitive bgp = null;
        if (null != cachedState) {
            bgp = (PerformBinaryGraphicsPrimitive)cachedState;
            if (!bgp.id.equals(OpaqueDitheredBlit.getMethodSignature())
                    || (bgp.srcType != src.type)
                    || (bgp.dstType != dst.type)) {
                bgp = null;
            }
        }
        if (null == bgp) {
            bgp = new PerformBinaryGraphicsPrimitive(
                        OpaqueDitheredBlit.getMethodSignature(),
                        src.type, dst.type, dst.devID);
        }
        if ((null == bgp) || (null == bgp.coreOp)) {
            return performOpaqueBlit(src, dst,
                                        srcX, srcY, dstX, dstY, w, h,
                                        cachedState);
        }
        LockableRaster  coreSrc = src;
        LockableRaster  coreDst = dst;
        int sX = srcX;
        int sY = srcY;
        int dX = dstX;
        int dY = dstY;
        if (null != bgp.convertSrcOp) {
            coreSrc = convertFrom(bgp.convertSrcOp, src, srcX, srcY, w, h);
            sX = 0;
            sY = 0;
        }
        if (null != bgp.readDstOp) {
            coreDst = convertFrom(bgp.readDstOp, dst, dstX, dstY, w, h);
            dX = 0;
            dY = 0;
        }
        try {
            OpaqueDitheredBlit odb = (OpaqueDitheredBlit)bgp.coreOp;
            int [] lockflags = odb.getLockFlags();
            coreSrc.lock(lockflags[0], sX, sY, w, h);
            coreDst.lock(lockflags[1], dX, dY, w, h);
            odb.OpaqueDitheredBlit(coreSrc.id, coreDst.id, w, h);
        } catch (RasterLockingException e) {
        } finally {
            coreDst.unlock();
            coreSrc.unlock();
        }
        if (null != bgp.writeDstOp) {
            convertTo(bgp.writeDstOp, dst, dstX, dstY, w, h, coreDst);
        }
        return bgp;
    }

    protected Object performOpaqueBlitBg(
                    LockableRaster src, LockableRaster dst, Color bgColor,
                    int srcX, int srcY, int dstX, int dstY, int w, int h,
                    Object cachedState)
    {
        PerformBinaryGraphicsPrimitive bgp = null;
        if (null != cachedState) {
            bgp = (PerformBinaryGraphicsPrimitive)cachedState;
            if (!bgp.id.equals(OpaqueBlitBg.getMethodSignature())
                    || (bgp.srcType != src.type)
                    || (bgp.dstType != dst.type)) {
                bgp = null;
            }
        }
        if (null == bgp) {
            bgp = new PerformBinaryGraphicsPrimitive(
                            OpaqueBlitBg.getMethodSignature(),
                            src.type, dst.type, dst.devID);
        }
        if ((null == bgp) || (null == bgp.coreOp)) {
			if (debugLog) {
				System.err.println("---" + src);
				System.err.println("---" + dst);
			}
            return null;
        }
        LockableRaster  coreSrc = src;
        LockableRaster  coreDst = dst;
        int sX = srcX;
        int sY = srcY;
        int dX = dstX;
        int dY = dstY;
        if (null != bgp.convertSrcOp) {
            coreSrc = convertFrom(bgp.convertSrcOp, src, srcX, srcY, w, h);
            sX = 0;
            sY = 0;
        }
        if (null != bgp.readDstOp) {
            coreDst = convertFrom(bgp.readDstOp, dst, dstX, dstY, w, h);
            dX = 0;
            dY = 0;
        }
        try {
            OpaqueBlitBg ob = (OpaqueBlitBg)bgp.coreOp;
            int [] lockflags = ob.getLockFlags();
            coreSrc.lock(lockflags[0], sX, sY, w, h);
            coreDst.lock(lockflags[1], dX, dY, w, h);
            ob.OpaqueBlitBg(coreSrc.id, coreDst.id, w, h, bgColor);
        } catch (RasterLockingException e) {
        } finally {
            coreDst.unlock();
            coreSrc.unlock();
        }
        if (null != bgp.writeDstOp) {
            convertTo(bgp.writeDstOp, dst, dstX, dstY, w, h, coreDst);
        }
        return bgp;
    }

    protected Object performOpaqueDitheredBlitBg(
                    LockableRaster src, LockableRaster dst, Color bgColor,
                    int srcX, int srcY, int dstX, int dstY, int w, int h,
                    Object cachedState)
    {
        PerformBinaryGraphicsPrimitive bgp = null;
        if (null != cachedState) {
            bgp = (PerformBinaryGraphicsPrimitive)cachedState;
            if (!bgp.id.equals(OpaqueDitheredBlitBg.getMethodSignature())
                    || (bgp.srcType != src.type)
                    || (bgp.dstType != dst.type)) {
                bgp = null;
            }
        }
        if (null == bgp) {
            bgp = new PerformBinaryGraphicsPrimitive(
                        OpaqueDitheredBlitBg.getMethodSignature(),
                        src.type, dst.type, dst.devID);
        }
        if ((null == bgp) || (null == bgp.coreOp)) {
            return performOpaqueBlit(src, dst,
                                        srcX, srcY, dstX, dstY, w, h,
                                        cachedState);
        }
        LockableRaster  coreSrc = src;
        LockableRaster  coreDst = dst;
        int sX = srcX;
        int sY = srcY;
        int dX = dstX;
        int dY = dstY;
        if (null != bgp.convertSrcOp) {
            coreSrc = convertFrom(bgp.convertSrcOp, src, srcX, srcY, w, h);
            sX = 0;
            sY = 0;
        }
        if (null != bgp.readDstOp) {
            coreDst = convertFrom(bgp.readDstOp, dst, dstX, dstY, w, h);
            dX = 0;
            dY = 0;
        }
        try {
            OpaqueDitheredBlitBg odb = (OpaqueDitheredBlitBg)bgp.coreOp;
            int [] lockflags = odb.getLockFlags();
            coreSrc.lock(lockflags[0], sX, sY, w, h);
            coreDst.lock(lockflags[1], dX, dY, w, h);
            odb.OpaqueDitheredBlitBg(coreSrc.id, coreDst.id, w, h, bgColor);
        } catch (RasterLockingException e) {
        } finally {
            coreDst.unlock();
            coreSrc.unlock();
        }
        if (null != bgp.writeDstOp) {
            convertTo(bgp.writeDstOp, dst, dstX, dstY, w, h, coreDst);
        }
        return bgp;
    }

    protected Object performTransparentBlit(
                    LockableRaster src, LockableRaster dst,
                    int srcX, int srcY, int dstX, int dstY, int w, int h,
                    Object cachedState)
    {
        PerformBinaryGraphicsPrimitive bgp = null;
        if (null != cachedState) {
            bgp = (PerformBinaryGraphicsPrimitive)cachedState;
            if (!bgp.id.equals(TransparentBlit.getMethodSignature())
                    || (bgp.srcType != src.type)
                    || (bgp.dstType != dst.type)) {
                bgp = null;
            }
        }
        if (null == bgp) {
            bgp = new PerformBinaryGraphicsPrimitive(
                            TransparentBlit.getMethodSignature(),
                            src.type, dst.type, dst.devID);
        }
        if ((null == bgp) || (null == bgp.coreOp)) {
			if (debugLog) {
				System.err.println("---" + src);
				System.err.println("---" + dst);
			}
            return null;
        }
        LockableRaster  coreSrc = src;
        LockableRaster  coreDst = dst;
        int sX = srcX;
        int sY = srcY;
        int dX = dstX;
        int dY = dstY;
        if (null != bgp.convertSrcOp) {
            coreSrc = convertFrom(bgp.convertSrcOp, src, srcX, srcY, w, h);
            sX = 0;
            sY = 0;
        }
        if (null != bgp.readDstOp) {
            coreDst = convertFrom(bgp.readDstOp, dst, dstX, dstY, w, h);
            dX = 0;
            dY = 0;
        }
        try {
            TransparentBlit tb = (TransparentBlit)bgp.coreOp;
            int [] lockflags = tb.getLockFlags();
            coreSrc.lock(lockflags[0], sX, sY, w, h);
            coreDst.lock(lockflags[1], dX, dY, w, h);
            tb.TransparentBlit(coreSrc.id, coreDst.id, w, h);
        } catch (RasterLockingException e) {
        } finally {
            coreDst.unlock();
            coreSrc.unlock();
        }
        if (null != bgp.writeDstOp) {
            convertTo(bgp.writeDstOp, dst, dstX, dstY, w, h, coreDst);
        }
        return bgp;
    }

    protected Object performTransparentDitheredBlit(
                    LockableRaster src, LockableRaster dst,
                    int srcX, int srcY, int dstX, int dstY, int w, int h,
                    Object cachedState)
    {
        PerformBinaryGraphicsPrimitive bgp = null;
        if (null != cachedState) {
            bgp = (PerformBinaryGraphicsPrimitive)cachedState;
            if (!bgp.id.equals(TransparentDitheredBlit.getMethodSignature())
                    || (bgp.srcType != src.type)
                    || (bgp.dstType != dst.type)) {
                bgp = null;
            }
        }
        if (null == bgp) {
            bgp = new PerformBinaryGraphicsPrimitive(
                        TransparentDitheredBlit.getMethodSignature(),
                        src.type, dst.type, dst.devID);
        }
        if ((null == bgp) || (null == bgp.coreOp)) {
            return performOpaqueBlit(src, dst,
                                        srcX, srcY, dstX, dstY, w, h,
                                        cachedState);
        }
        LockableRaster  coreSrc = src;
        LockableRaster  coreDst = dst;
        int sX = srcX;
        int sY = srcY;
        int dX = dstX;
        int dY = dstY;
        if (null != bgp.convertSrcOp) {
            coreSrc = convertFrom(bgp.convertSrcOp, src, srcX, srcY, w, h);
            sX = 0;
            sY = 0;
        }
        if (null != bgp.readDstOp) {
            coreDst = convertFrom(bgp.readDstOp, dst, dstX, dstY, w, h);
            dX = 0;
            dY = 0;
        }
        try {
            TransparentDitheredBlit tdb = (TransparentDitheredBlit)bgp.coreOp;
            int [] lockflags = tdb.getLockFlags();
            coreSrc.lock(lockflags[0], sX, sY, w, h);
            coreDst.lock(lockflags[1], dX, dY, w, h);
            tdb.TransparentDitheredBlit(coreSrc.id, coreDst.id, w, h);
        } catch (RasterLockingException e) {
        } finally {
            coreDst.unlock();
            coreSrc.unlock();
        }
        if (null != bgp.writeDstOp) {
            convertTo(bgp.writeDstOp, dst, dstX, dstY, w, h, coreDst);
        }
        return bgp;
    }

    protected Object performBinaryComposite(
                    LockableRaster src, LockableRaster dst,
                    int srcX, int srcY, int dstX, int dstY, int w, int h,
                    byte [] alphaTile, int alphaOffset, int alphaScan,
                    int alphaRule, float extraAlpha,
                    Color xorColor, Object cachedState)
    {
        PerformBinaryGraphicsPrimitive bgp = null;
        if (null != cachedState) {
            bgp = (PerformBinaryGraphicsPrimitive)cachedState;
            if (!bgp.id.equals(BinaryComposite.getMethodSignature())
                    || (bgp.srcType != src.type)
                    || (bgp.dstType != dst.type)) {
                bgp = null;
            }
        }
        if (null == bgp) {
            bgp = new PerformBinaryGraphicsPrimitive(
                            BinaryComposite.getMethodSignature(),
                            src.type, dst.type, dst.devID);
        }
        if ((null == bgp) || (null == bgp.coreOp)) {
            return null;
        }
        LockableRaster  coreSrc = src;
        LockableRaster  coreDst = dst;
        int sX = srcX;
        int sY = srcY;
        int dX = dstX;
        int dY = dstY;
        if (null != bgp.convertSrcOp) {
            coreSrc = convertFrom(bgp.convertSrcOp, src, srcX, srcY, w, h);
            sX = 0;
            sY = 0;
        }
        if (null != bgp.readDstOp) {
            coreDst = convertFrom(bgp.readDstOp, dst, dstX, dstY, w, h);
            dX = 0;
            dY = 0;
        }
        try {
            BinaryComposite bc = (BinaryComposite)bgp.coreOp;
            int [] lockflags = bc.getLockFlags();
            coreSrc.lock(lockflags[0], sX, sY, w, h);
            coreDst.lock(lockflags[1], dX, dY, w, h);
            bc.BinaryComposite(coreSrc.id, coreDst.id, w, h,
                            alphaRule,
                            alphaTile, alphaOffset, alphaScan,
                            extraAlpha, xorColor);
        } catch (RasterLockingException e) {
        } finally {
            coreDst.unlock();
            coreSrc.unlock();
        }
        if (null != bgp.writeDstOp) {
            convertTo(bgp.writeDstOp, dst, dstX, dstY, w, h, coreDst);
        }
        return bgp;
    }

    protected Object performBinaryCompositeBg(
                    LockableRaster src, LockableRaster dst,
                    int srcX, int srcY, int dstX, int dstY, int w, int h,
                    byte [] alphaTile, int alphaOffset, int alphaScan,
                    int alphaRule, float extraAlpha,
                    Color bgColor, Color xorColor, Object cachedState)
    {
        PerformBinaryGraphicsPrimitive bgp = null;
        if (null != cachedState) {
            bgp = (PerformBinaryGraphicsPrimitive)cachedState;
            if (!bgp.id.equals(BinaryCompositeBg.getMethodSignature())
                    || (bgp.srcType != src.type)
                    || (bgp.dstType != dst.type)) {
                bgp = null;
            }
        }
        if (null == bgp) {
            bgp = new PerformBinaryGraphicsPrimitive(
                            BinaryCompositeBg.getMethodSignature(),
                            src.type, dst.type, dst.devID);
        }
        if ((null == bgp) || (null == bgp.coreOp)) {
            return null;
        }
        LockableRaster  coreSrc = src;
        LockableRaster  coreDst = dst;
        int sX = srcX;
        int sY = srcY;
        int dX = dstX;
        int dY = dstY;
        if (null != bgp.convertSrcOp) {
            coreSrc = convertFrom(bgp.convertSrcOp, src, srcX, srcY, w, h);
            sX = 0;
            sY = 0;
        }
        if (null != bgp.readDstOp) {
            coreDst = convertFrom(bgp.readDstOp, dst, dstX, dstY, w, h);
            dX = 0;
            dY = 0;
        }
        try {
            BinaryCompositeBg bc = (BinaryCompositeBg)bgp.coreOp;
            int [] lockflags = bc.getLockFlags();
            coreSrc.lock(lockflags[0], sX, sY, w, h);
            coreDst.lock(lockflags[1], dX, dY, w, h);
            bc.BinaryCompositeBg(coreSrc.id, coreDst.id, w, h,
                            alphaRule,
                            alphaTile, alphaOffset, alphaScan,
                            extraAlpha, bgColor, xorColor);
        } catch (RasterLockingException e) {
        } finally {
            coreDst.unlock();
            coreSrc.unlock();
        }
        if (null != bgp.writeDstOp) {
            convertTo(bgp.writeDstOp, dst, dstX, dstY, w, h, coreDst);
        }
        return bgp;
    }

    protected LockableRaster convertFrom(OpaqueBlit ob, LockableRaster src,
                                        int srcX, int srcY, int w, int h) {
        LockableRaster dst = new LockableRaster(w, h);
        try {
            int [] lockflags = ob.getLockFlags();
            src.lock(lockflags[0], srcX, srcY, w, h);
            dst.lock(lockflags[1], 0, 0, w, h);
            ob.OpaqueBlit(src.id, dst.id, w, h);
        } catch (RasterLockingException e) {
        } finally {
            dst.unlock();
            src.unlock();
        }
        return dst;
    }

    protected void convertTo(OpaqueBlit ob, LockableRaster dst,
                                int dstX, int dstY, int w, int h,
                                LockableRaster src) {
        try {
            int [] lockflags = ob.getLockFlags();
            src.lock(lockflags[0], 0, 0, w, h);
            dst.lock(lockflags[1], dstX, dstY, w, h);
            ob.OpaqueBlit(src.id, dst.id, w, h);
        } catch (RasterLockingException e) {
        } finally {
            dst.unlock();
            src.unlock();
        }
    }

    /**
    *   PerformUnaryGraphicsPrimitive
    *   encapsulates behavior of strategies for optimal performance of
    *   requested operation when destination surface may need to be
    *   converted to an intermediate for the operation to succeed
    *
    *   Strategies in decreasing preference order:
    *   1) op A
    *   2) op A, with a general purpose renderer
    *   3) A to A', op A', A' to A
    */
    private class PerformUnaryGraphicsPrimitive {
        GraphicsPrimitive   coreOp;
        OpaqueBlit          readDstOp;
        OpaqueBlit          writeDstOp;

        String id;
        int srcType;
        int dstType;
        String deviceID;

        public PerformUnaryGraphicsPrimitive(String id, int dstType,
                                                    String deviceID) {
            this.id = id;
            this.dstType = dstType;
            this.deviceID = deviceID;

            // first strategy:  op A
            int [] types = {dstType};
            coreOp = GraphicsPrimitiveMgr.locate(id, types, deviceID);
            if (null != coreOp) {
                return;
            }
            // second strategy:  op A, with a general purpose renderer
            if (dstType != GraphicsPrimitive.TYPE_BYTE_BINARY &&
                dstType != GraphicsPrimitive.TYPE_BINARY) {
                int [] types2 = {GraphicsPrimitive.TYPE_ANY_UNARY};
                coreOp = GraphicsPrimitiveMgr.locate(id, types2, deviceID);
                if (null != coreOp) {
                    if (debugLog) {
                        System.err.println("Unary Strategy 2 " + srcType + " " + dstType);
                    }
                    return;
                }
            }
            // third strategy: A to A', op A', A' to A
            int [] types3 = {metaFormat};
            coreOp = GraphicsPrimitiveMgr.locate(id, types3, deviceID);
            if (null != coreOp) {
                readDstOp = createConverter(dstType, metaFormat);
                writeDstOp = createConverter(metaFormat, dstType);
                if ((null == readDstOp) || (null == writeDstOp)) {
                    reportError(id, dstType);
                    coreOp = null;
                    readDstOp = null;
                    writeDstOp = null;
                } else {
                    if (debugLog) {
                        System.err.println("Unary Strategy 3 " + srcType + " " + dstType);
                    }
                    return;
                }
            }
            reportError(id, dstType);
            return;
        }

        private OpaqueBlit createConverter(int srcType, int dstType) {
            GraphicsPrimitive converter;
            // REMIND: need to check RENDER_SPEED somehow before dithering
            int [] types = {srcType, dstType};
//         converter = GraphicsPrimitiveMgr.locate(ditherID, types, deviceID);
//            if (null != converter) {
//                return (OpaqueBlit)converter;
//            }
            converter = GraphicsPrimitiveMgr.locate(copyID, types, deviceID);
            if (null != converter) {
                return (OpaqueBlit)converter;
            }
            int [] types2 = {defaultFormat, defaultFormat};
            converter = GraphicsPrimitiveMgr.locate(copyID, types2, deviceID);
            return (OpaqueBlit)converter;
        }

        private void reportError(String id, int dstType) {
            if (!debugTrace) {
                return;
            }
            System.err.println("Unary Graphics Operation failed:");
            System.err.println(" primitive: [" + id + "]");
            System.err.println(" surface: [" +
                                    GraphicsPrimitive.typeStr(dstType) + "]");
        }
    }

    /**
     * Render source color onto destination
     * SunGraphics2D using:
     *  1) shape tile
     *  2) specified alphaCompositeRule (instead of destination AC)
     *  3) foreground color
     *
     *  may return an Object which contains cached state which can
     *  speed up further operations in the same context
     *
     *  uses:
     *  1) ColorPaint
     */
    public Object compositeColorLoop( SunGraphics2D dest2D,
                        int alphaRule, float extraAlpha,
                        byte[] alphaMask, int alphaOffset, int alphaScan,
                        int dstX, int dstY, int w, int h,
                        Color foreColor, Object cachedState )
    {
//         Color foreColor = new Color(fgColor, true);
        LockableRaster dst = null;
        PerformUnaryGraphicsPrimitive ugp = null;
        UnaryCachedState ucs = null;
        if (null != cachedState) {
            ucs = (UnaryCachedState)cachedState;
            if (ucs.the2D == dest2D) {
                dst = ucs.theLR;
                ugp = ucs.ugp;
                if (!ugp.id.equals(ColorPaint.getMethodSignature())
                        || (ugp.dstType != dst.type)) {
                    ugp = null;
                }
            }
        }
        if (null == dst) {
            dst = new LockableRaster(dest2D);
        }
        if (null == ugp) {
            ugp = new PerformUnaryGraphicsPrimitive(
                            ColorPaint.getMethodSignature(),
                            dst.type, dst.devID);
        }
        if ((null == ugp)
                || (null == ugp.coreOp)) {
            // REMIND: this disables unary strategy 2
            new Exception().printStackTrace();
            // REMIND: should have exception thrown
            return null;
        }
        if (ucs == null) {
            ucs = new UnaryCachedState(dest2D, dst, ugp);
        }
        LockableRaster  coreDst = dst;
        int dX = dstX;
        int dY = dstY;
        if (null != ugp.readDstOp) {
            coreDst = convertFrom(ugp.readDstOp, dst, dstX, dstY, w, h);
            dX = 0;
            dY = 0;
        }
        try {
            ColorPaint cp = (ColorPaint)ugp.coreOp;
            coreDst.lock(cp.getLockFlags()[0], dX, dY, w, h);
            cp.ColorPaint(coreDst.id,
                            alphaRule,
                            alphaMask, alphaOffset, alphaScan,
                            extraAlpha, foreColor);
        } catch (RasterLockingException e) {
        } finally {
            coreDst.unlock();
        }
        if (null != ugp.writeDstOp) {
            convertTo(ugp.writeDstOp, dst, dstX, dstY, w, h, coreDst);
        }
        return ucs;
    }

    private Rectangle myIntersectRect(Rectangle r1, Rectangle r2) {
        int r1Left     = r1.x;
        int r1Right    = r1.x + r1.width;
        int r1Top      = r1.y;
        int r1Bottom   = r1.y + r1.height;

        int r2Left     = r2.x;
        int r2Right    = r2.x + r2.width;
        int r2Top      = r2.y;
        int r2Bottom   = r2.y + r2.height;

        r2Left     = (r1Left > r2Left)? r1Left: r2Left;
        r2Right    = (r1Right < r2Right)? r1Right: r2Right;
        r2Top      = (r1Top > r2Top)? r1Top: r2Top;
        r2Bottom   = (r1Bottom < r2Bottom)? r1Bottom: r2Bottom;

        r2Right = (r2Left > r2Right)? r2Left: r2Right;
        r2Bottom = (r2Top > r2Bottom)? r2Top: r2Bottom;


        return new Rectangle(r2Left,r2Top,r2Right-r2Left,r2Bottom-r2Top);
    }

    /**
    * SimpleRenderPipe methods
    */
    public void drawLine(SunGraphics2D dest2D,
            int x1, int y1, int x2, int y2) {
        LockableRaster dst = new LockableRaster(dest2D);
        PerformUnaryGraphicsPrimitive ugp = null;
        ugp = new PerformUnaryGraphicsPrimitive(
                        DrawLine.getMethodSignature(), dst.type, dst.devID);
        if ((null == ugp)
            || (null == ugp.coreOp)) {
            // REMIND: this disables unary strategy 2
            new Exception().printStackTrace();
            // REMIND: should have exception thrown
            return;
        }
        LockableRaster  coreDst = dst;
        int dX = (x1 > x2)? x2: x1;
        int dY = (y1 > y2)? y2: y1;
        int w = (x1 > x2)? (x1 - x2 + 1): (x2 - x1 + 1);
        int h = (y1 > y2)? (y1 - y2 + 1): (y2 - y1 + 1);
        int dstX = dX;
        int dstY = dY;
        if (null != ugp.readDstOp) {
            coreDst = convertFrom(ugp.readDstOp, dst, dstX, dstY, w, h);
            dX = 0;
            dY = 0;
        }
        try {
            DrawLine dl = (DrawLine)ugp.coreOp;
            coreDst.lock(dl.getLockFlags()[0], dX, dY, w, h);
            dl.DrawLine(coreDst.id, dest2D.getColor(),
                        x1, y1, x2, y2);
        } catch (RasterLockingException e) {
        } finally {
            coreDst.unlock();
        }
        if (null != ugp.writeDstOp) {
            convertTo(ugp.writeDstOp, dst, dstX, dstY, w, h, coreDst);
        }
    }

    public void fillRect(SunGraphics2D dest2D, int dX, int dY, int w, int h) {
        LockableRaster dst = new LockableRaster(dest2D);
        PerformUnaryGraphicsPrimitive ugp = null;
        ugp = new PerformUnaryGraphicsPrimitive(
                        FillRect.getMethodSignature(), dst.type, dst.devID);
        if ((null == ugp)
            || (null == ugp.coreOp)) {
            // REMIND: this disables unary strategy 2
            new Exception().printStackTrace();
            // REMIND: should have exception thrown
            return;
        }
        LockableRaster  coreDst = dst;
        int dstX = dX;
        int dstY = dY;
        if (null != ugp.readDstOp) {
            coreDst = convertFrom(ugp.readDstOp, dst, dstX, dstY, w, h);
            dX = 0;
            dY = 0;
        }
        try {
            FillRect fr = (FillRect)ugp.coreOp;
            coreDst.lock(fr.getLockFlags()[0], dX, dY, w, h);
            fr.FillRect(coreDst.id, dest2D.getColor(),
                        dX, dY, w, h);
        } catch (RasterLockingException e) {
        } finally {
            coreDst.unlock();
        }
        if (null != ugp.writeDstOp) {
            convertTo(ugp.writeDstOp, dst, dstX, dstY, w, h, coreDst);
        }
    }

    public void drawPolys(SunGraphics2D dest2D,
            int[] xPoints, int[] yPoints, int nPoints[],
            int numpolys, boolean close) {
        LockableRaster dst = new LockableRaster(dest2D);
        PerformUnaryGraphicsPrimitive ugp = null;
        ugp = new PerformUnaryGraphicsPrimitive(
                        DrawPolygons.getMethodSignature(), dst.type, dst.devID);
        if ((null == ugp)
            || (null == ugp.coreOp)) {
            // REMIND: this disables unary strategy 2
            new Exception().printStackTrace();
            // REMIND: should have exception thrown
            return;
        }
        LockableRaster  coreDst = dst;
        Rectangle clip = dest2D.getCompBounds();
        int dX = clip.x;
        int dY = clip.y;
        int w = clip.width;
        int h = clip.height;
        int dstX = dX;
        int dstY = dY;
        if (null != ugp.readDstOp) {
            coreDst = convertFrom(ugp.readDstOp, dst, dstX, dstY, w, h);
            dX = 0;
            dY = 0;
        }
        try {
            DrawPolygons dp = (DrawPolygons)ugp.coreOp;
            coreDst.lock(dp.getLockFlags()[0], dX, dY, w, h);
            dp.DrawPolygons(coreDst.id, dest2D.getColor(),
                            xPoints, yPoints, nPoints, numpolys, close);
        } catch (RasterLockingException e) {
        } finally {
            coreDst.unlock();
        }
        if (null != ugp.writeDstOp) {
            convertTo(ugp.writeDstOp, dst, dstX, dstY, w, h, coreDst);
        }
   }

    public void fillPolys(SunGraphics2D dest2D,
            int[] xPoints, int[] yPoints, int nPoints[],
            int numpolys, boolean evenodd) {
        LockableRaster dst = new LockableRaster(dest2D);
        PerformUnaryGraphicsPrimitive ugp = null;
        ugp = new PerformUnaryGraphicsPrimitive(
                        FillPolygons.getMethodSignature(), dst.type, dst.devID);
        if ((null == ugp)
            || (null == ugp.coreOp)) {
            // REMIND: this disables unary strategy 2
            new Exception().printStackTrace();
            // REMIND: should have exception thrown
            return;
        }
        LockableRaster  coreDst = dst;
        Rectangle clip = dest2D.getCompBounds();
        int dX = clip.x;
        int dY = clip.y;
        int w = clip.width;
        int h = clip.height;
        int dstX = dX;
        int dstY = dY;
        if (null != ugp.readDstOp) {
            coreDst = convertFrom(ugp.readDstOp, dst, dstX, dstY, w, h);
            dX = 0;
            dY = 0;
        }
        try {
            FillPolygons fp = (FillPolygons)ugp.coreOp;
            coreDst.lock(fp.getLockFlags()[0], dX, dY, w, h);
            fp.FillPolygons(coreDst.id, dest2D.getColor(),
                            xPoints, yPoints, nPoints, numpolys, evenodd);
        } catch (RasterLockingException e) {
        } finally {
            coreDst.unlock();
        }
        if (null != ugp.writeDstOp) {
            convertTo(ugp.writeDstOp, dst, dstX, dstY, w, h, coreDst);
        }
    }

    public void fillSpans(SunGraphics2D dest2D,
			  SpanIterator si) {
	int box[] = new int[4];
	si.getPathBox(box);
	int dX = box[0];
	int dY = box[1];
	int w = box[2] - dX;
	int h = box[3] - dY;

	LockableRaster dst = new LockableRaster(dest2D);
	int types[] = {dst.type};
	GraphicsPrimitive coreOp =
	    GraphicsPrimitiveMgr.locate(FillSpans.getMethodSignature(),
					types, dst.devID);
	if (coreOp != null) {
	    try {
		FillSpans fs = (FillSpans) coreOp;
		dst.lock(fs.getLockFlags()[0], dX, dY, w, h);
		fs.FillSpans(dst.id, dest2D.getColor(), si);
	    } catch (RasterLockingException e) {
	    } finally {
		dst.unlock();
	    }
	} else {
	    PerformUnaryGraphicsPrimitive ugp = null;
	    ugp = new PerformUnaryGraphicsPrimitive
		(FillRect.getMethodSignature(), dst.type, dst.devID);
	    if ((null == ugp)
                || (null == ugp.coreOp)
                || (null != ugp.readDstOp)
                || (null != ugp.writeDstOp)) {
		// REMIND: this disables unary strategy 2
		new Exception().printStackTrace();
		// REMIND: should have exception thrown
		return;
	    }
	    FillRect fr = (FillRect) ugp.coreOp;
	    Color c = dest2D.getColor();
	    try {
		dst.lock(fr.getLockFlags()[0], dX, dY, w, h);
		ImageData id = dst.id;
		while (si.nextSpan(box)) {
		    fr.FillRect(id, c, box[0], box[1],
				box[2] - box[0], box[3] - box[1]);
		}
	    } catch (RasterLockingException e) {
	    } finally {
		dst.unlock();
	    }
	}
    }

    /**
    * PixelDrawPipe methods
    */
    public void drawRect(SunGraphics2D dest2D,
			 int dX, int dY, int w, int h) {
        LockableRaster dst = new LockableRaster(dest2D);
        PerformUnaryGraphicsPrimitive ugp = null;
        ugp = new PerformUnaryGraphicsPrimitive(
                        DrawRect.getMethodSignature(), dst.type, dst.devID);
        if ((null == ugp)
            || (null == ugp.coreOp)) {
            // REMIND: this disables unary strategy 2
            new Exception().printStackTrace();
            // REMIND: should have exception thrown
            return;
        }
        LockableRaster  coreDst = dst;
        int dstX = dX;
        int dstY = dY;
        if (null != ugp.readDstOp) {
            coreDst = convertFrom(ugp.readDstOp, dst, dstX, dstY, w, h);
            dX = 0;
            dY = 0;
        }
        try {
            DrawRect dr = (DrawRect)ugp.coreOp;
            coreDst.lock(dr.getLockFlags()[0], dX, dY, w+1, h+1);
            dr.DrawRect(coreDst.id, dest2D.getColor(),
                        dX, dY, w, h);
        } catch (RasterLockingException e) {
        } finally {
            coreDst.unlock();
        }
        if (null != ugp.writeDstOp) {
            convertTo(ugp.writeDstOp, dst, dstX, dstY, w, h, coreDst);
        }
    }

    public void drawRoundRect(SunGraphics2D sg,
			      int x, int y, int w, int h,
			      int aW, int aH) {
	sg.shapepipe.draw(sg, new RoundRectangle2D.Float(x, y, w, h, aW, aH));
    }

    public void drawOval(SunGraphics2D sg,
			 int x, int y, int w, int h) {
	sg.shapepipe.draw(sg, new Ellipse2D.Float(x, y, w, h));
    }

    public void drawArc(SunGraphics2D sg,
			int x, int y, int w, int h,
			int start, int extent) {
	sg.shapepipe.draw(sg, new Arc2D.Float(x, y, w, h,
					      start, extent, Arc2D.OPEN));
    }

    public void drawPolyline(SunGraphics2D dest2D,
			     int xPoints[], int yPoints[],
			     int nPoints) {
        LockableRaster dst = new LockableRaster(dest2D);
        PerformUnaryGraphicsPrimitive ugp = null;
        ugp = new PerformUnaryGraphicsPrimitive(
                        DrawPolygons.getMethodSignature(), dst.type, dst.devID);
        if ((null == ugp)
            || (null == ugp.coreOp)) {
            // REMIND: this disables unary strategy 2
            new Exception().printStackTrace();
            // REMIND: should have exception thrown
            return;
        }
        LockableRaster  coreDst = dst;
        Rectangle clip = dest2D.getCompBounds();
        int dX = clip.x;
        int dY = clip.y;
        int w = clip.width;
        int h = clip.height;
        int dstX = dX;
        int dstY = dY;
        if (null != ugp.readDstOp) {
            coreDst = convertFrom(ugp.readDstOp, dst, dstX, dstY, w, h);
            dX = 0;
            dY = 0;
        }
        try {
            int nPointsArray[] = { nPoints };
            DrawPolygons dp = (DrawPolygons)ugp.coreOp;
            coreDst.lock(dp.getLockFlags()[0], dX, dY, w, h);
            dp.DrawPolygons(coreDst.id, dest2D.getColor(),
                            xPoints, yPoints, nPointsArray, 1, false);
        } catch (RasterLockingException e) {
        } finally {
            coreDst.unlock();
        }
        if (null != ugp.writeDstOp) {
            convertTo(ugp.writeDstOp, dst, dstX, dstY, w, h, coreDst);
        }
    }

    public void drawPolygon(SunGraphics2D dest2D,
			    int xPoints[], int yPoints[],
			    int nPoints) {
        LockableRaster dst = new LockableRaster(dest2D);
        PerformUnaryGraphicsPrimitive ugp = null;
        ugp = new PerformUnaryGraphicsPrimitive(
                        DrawPolygons.getMethodSignature(), dst.type, dst.devID);
        if ((null == ugp)
            || (null == ugp.coreOp)) {
            // REMIND: this disables unary strategy 2
            new Exception().printStackTrace();
            // REMIND: should have exception thrown
            return;
        }
        LockableRaster  coreDst = dst;
        Rectangle clip = dest2D.getCompBounds();
        int dX = clip.x;
        int dY = clip.y;
        int w = clip.width;
        int h = clip.height;
        int dstX = dX;
        int dstY = dY;
        if (null != ugp.readDstOp) {
            coreDst = convertFrom(ugp.readDstOp, dst, dstX, dstY, w, h);
            dX = 0;
            dY = 0;
        }
        try {
            int nPointsArray[] = { nPoints };
            DrawPolygons dp = (DrawPolygons)ugp.coreOp;
            coreDst.lock(dp.getLockFlags()[0], dX, dY, w, h);
            dp.DrawPolygons(coreDst.id, dest2D.getColor(),
                            xPoints, yPoints, nPointsArray, 1, true);
        } catch (RasterLockingException e) {
        } finally {
            coreDst.unlock();
        }
        if (null != ugp.writeDstOp) {
            convertTo(ugp.writeDstOp, dst, dstX, dstY, w, h, coreDst);
        }
    }

    /**
    * PixelFillPipe methods
    */
    public void fillRoundRect(SunGraphics2D sg,
			      int x, int y, int w, int h,
			      int aW, int aH) {
	sg.shapepipe.fill(sg, new RoundRectangle2D.Float(x, y, w, h, aW, aH));
    }

    public void fillOval(SunGraphics2D sg,
			 int x, int y, int w, int h) {
	sg.shapepipe.fill(sg, new Ellipse2D.Float(x, y, w, h));
    }

    public void fillArc(SunGraphics2D sg,
			int x, int y, int w, int h,
			int start, int extent) {
	sg.shapepipe.fill(sg, new Arc2D.Float(x, y, w, h,
					      start, extent, Arc2D.PIE));
    }

    public void fillPolygon(SunGraphics2D dest2D,
			    int xPoints[], int yPoints[],
			    int nPoints) {
        LockableRaster dst = new LockableRaster(dest2D);
        PerformUnaryGraphicsPrimitive ugp = null;
        ugp = new PerformUnaryGraphicsPrimitive(
                        FillPolygons.getMethodSignature(), dst.type, dst.devID);
        if ((null == ugp)
            || (null == ugp.coreOp)) {
            // REMIND: this disables unary strategy 2
            new Exception().printStackTrace();
            // REMIND: should have exception thrown
            return;
        }
        LockableRaster  coreDst = dst;
        Rectangle clip = dest2D.getCompBounds();
        int dX = clip.x;
        int dY = clip.y;
        int w = clip.width;
        int h = clip.height;
        int dstX = dX;
        int dstY = dY;
        if (null != ugp.readDstOp) {
            coreDst = convertFrom(ugp.readDstOp, dst, dstX, dstY, w, h);
            dX = 0;
            dY = 0;
        }
        try {
            int nPointsArray[] = {nPoints};
            FillPolygons fp = (FillPolygons)ugp.coreOp;
            coreDst.lock(fp.getLockFlags()[0], dX, dY, w, h);
            fp.FillPolygons(coreDst.id, dest2D.getColor(),
                            xPoints, yPoints, nPointsArray, 1, true);
        } catch (RasterLockingException e) {
        } finally {
            coreDst.unlock();
        }
        if (null != ugp.writeDstOp) {
            convertTo(ugp.writeDstOp, dst, dstX, dstY, w, h, coreDst);
        }
    }

    /**
    * SimpleRenderPipe helper methods
    */
    public Object drawLine(SunGraphics2D dest2D,
            int x1, int y1, int x2, int y2, Object cachedState) {
        LockableRaster dst = null;
        PerformUnaryGraphicsPrimitive ugp = null;
        UnaryCachedState ucs = null;
        if (null != cachedState) {
            ucs = (UnaryCachedState)cachedState;
            if (ucs.the2D == dest2D) {
                dst = ucs.theLR;
                ugp = ucs.ugp;
                if (!ugp.id.equals(DrawLine.getMethodSignature())
                        || (ugp.dstType != dst.type)) {
                    ugp = null;
                }
            }
        }
        if (null == dst) {
            dst = new LockableRaster(dest2D);
        }
        if (null == ugp) {
            ugp = new PerformUnaryGraphicsPrimitive(
                            DrawLine.getMethodSignature(),
                            dst.type, dst.devID);
        }
        if ((null == ugp)
            || (null == ugp.coreOp)) {
            // REMIND: this disables unary strategy 2
            new Exception().printStackTrace();
            // REMIND: should have exception thrown
            return null;
        }
        if (ucs == null) {
            ucs = new UnaryCachedState(dest2D, dst, ugp);
        }
        LockableRaster  coreDst = dst;
        int dX = (x1 > x2)? x2: x1;
        int dY = (y1 > y2)? y2: y1;
        int w = (x1 > x2)? (x1 - x2 + 1): (x2 - x1 + 1);
        int h = (y1 > y2)? (y1 - y2 + 1): (y2 - y1 + 1);
        int dstX = dX;
        int dstY = dY;
        if (null != ugp.readDstOp) {
            coreDst = convertFrom(ugp.readDstOp, dst, dstX, dstY, w, h);
            dX = 0;
            dY = 0;
        }
        try {
            DrawLine dl = (DrawLine)ugp.coreOp;
            coreDst.lock(dl.getLockFlags()[0], dX, dY, w, h);
            dl.DrawLine(coreDst.id, dest2D.getColor(),
                        x1, y1, x2, y2);
        } catch (RasterLockingException e) {
        } finally {
            coreDst.unlock();
        }
        if (null != ugp.writeDstOp) {
            convertTo(ugp.writeDstOp, dst, dstX, dstY, w, h, coreDst);
        }
        return ucs;
    }

    public Object fillRect(SunGraphics2D dest2D,
            int dX, int dY, int w, int h, Object cachedState) {
        LockableRaster dst = null;
        PerformUnaryGraphicsPrimitive ugp = null;
        UnaryCachedState ucs = null;
        if (null != cachedState) {
            ucs = (UnaryCachedState)cachedState;
            if (ucs.the2D == dest2D) {
                dst = ucs.theLR;
                ugp = ucs.ugp;
                if (!ugp.id.equals(FillRect.getMethodSignature())
                        || (ugp.dstType != dst.type)) {
                    ugp = null;
                }
            }
        }
        if (null == dst) {
            dst = new LockableRaster(dest2D);
        }
        if (null == ugp) {
            ugp = new PerformUnaryGraphicsPrimitive(
                            FillRect.getMethodSignature(),
                            dst.type, dst.devID);
        }
        if ((null == ugp)
            || (null == ugp.coreOp)) {
            // REMIND: this disables unary strategy 2
            new Exception().printStackTrace();
            // REMIND: should have exception thrown
            return null;
        }
        if (ucs == null) {
            ucs = new UnaryCachedState(dest2D, dst, ugp);
        }
        LockableRaster  coreDst = dst;
        int dstX = dX;
        int dstY = dY;
        if (null != ugp.readDstOp) {
            coreDst = convertFrom(ugp.readDstOp, dst, dstX, dstY, w, h);
            dX = 0;
            dY = 0;
        }
        try {
            FillRect fr = (FillRect)ugp.coreOp;
            coreDst.lock(fr.getLockFlags()[0], dX, dY, w, h);
            fr.FillRect(coreDst.id, dest2D.getColor(),
                        dX, dY, w, h);
        } catch (RasterLockingException e) {
        } finally {
            coreDst.unlock();
        }
        if (null != ugp.writeDstOp) {
            convertTo(ugp.writeDstOp, dst, dstX, dstY, w, h, coreDst);
        }
        return ucs;
    }

    public Object drawPolys(SunGraphics2D dest2D,
            int[] xPoints, int[] yPoints, int nPoints[],
            int numpolys, boolean close, Object cachedState) {
        LockableRaster dst = null;
        PerformUnaryGraphicsPrimitive ugp = null;
        UnaryCachedState ucs = null;
        if (null != cachedState) {
            ucs = (UnaryCachedState)cachedState;
            if (ucs.the2D == dest2D) {
                dst = ucs.theLR;
                ugp = ucs.ugp;
                if (!ugp.id.equals(DrawPolygons.getMethodSignature())
                        || (ugp.dstType != dst.type)) {
                    ugp = null;
                }
            }
        }
        if (null == dst) {
            dst = new LockableRaster(dest2D);
        }
        if (null == ugp) {
            ugp = new PerformUnaryGraphicsPrimitive(
                            DrawPolygons.getMethodSignature(),
                            dst.type, dst.devID);
        }
        if ((null == ugp)
            || (null == ugp.coreOp)) {
            // REMIND: this disables unary strategy 2
            new Exception().printStackTrace();
            // REMIND: should have exception thrown
            return null;
        }
        if (ucs == null) {
            ucs = new UnaryCachedState(dest2D, dst, ugp);
        }
        LockableRaster  coreDst = dst;
        Rectangle clip = dest2D.getCompBounds();
        int dX = clip.x;
        int dY = clip.y;
        int w = clip.width;
        int h = clip.height;
        int dstX = dX;
        int dstY = dY;
        if (null != ugp.readDstOp) {
            coreDst = convertFrom(ugp.readDstOp, dst, dstX, dstY, w, h);
            dX = 0;
            dY = 0;
        }
        try {
            DrawPolygons dp = (DrawPolygons)ugp.coreOp;
            coreDst.lock(dp.getLockFlags()[0], dX, dY, w, h);
            dp.DrawPolygons(coreDst.id, dest2D.getColor(),
                            xPoints, yPoints, nPoints, numpolys, close);
        } catch (RasterLockingException e) {
        } finally {
            coreDst.unlock();
        }
        if (null != ugp.writeDstOp) {
            convertTo(ugp.writeDstOp, dst, dstX, dstY, w, h, coreDst);
        }
        return ucs;
    }

    public Object fillPolys(SunGraphics2D dest2D,
            int[] xPoints, int[] yPoints, int nPoints[],
            int numpolys, boolean evenodd, Object cachedState) {
        LockableRaster dst = null;
        PerformUnaryGraphicsPrimitive ugp = null;
        UnaryCachedState ucs = null;
        if (null != cachedState) {
            ucs = (UnaryCachedState)cachedState;
            if (ucs.the2D == dest2D) {
                dst = ucs.theLR;
                ugp = ucs.ugp;
                if (!ugp.id.equals(FillPolygons.getMethodSignature())
                        || (ugp.dstType != dst.type)) {
                    ugp = null;
                }
            }
        }
        if (null == dst) {
            dst = new LockableRaster(dest2D);
        }
        if (null == ugp) {
            ugp = new PerformUnaryGraphicsPrimitive(
                            FillPolygons.getMethodSignature(),
                            dst.type, dst.devID);
        }
        if ((null == ugp)
            || (null == ugp.coreOp)) {
            // REMIND: this disables unary strategy 2
            new Exception().printStackTrace();
            // REMIND: should have exception thrown
            return null;
        }
        if (ucs == null) {
            ucs = new UnaryCachedState(dest2D, dst, ugp);
        }
        LockableRaster  coreDst = dst;
        Rectangle clip = dest2D.getCompBounds();
        int dX = clip.x;
        int dY = clip.y;
        int w = clip.width;
        int h = clip.height;
        int dstX = dX;
        int dstY = dY;
        if (null != ugp.readDstOp) {
            coreDst = convertFrom(ugp.readDstOp, dst, dstX, dstY, w, h);
            dX = 0;
            dY = 0;
        }
        try {
            FillPolygons fp = (FillPolygons)ugp.coreOp;
            coreDst.lock(fp.getLockFlags()[0], dX, dY, w, h);
            fp.FillPolygons(coreDst.id, dest2D.getColor(),
                            xPoints, yPoints, nPoints, numpolys, evenodd);
        } catch (RasterLockingException e) {
        } finally {
            coreDst.unlock();
        }
        if (null != ugp.writeDstOp) {
            convertTo(ugp.writeDstOp, dst, dstX, dstY, w, h, coreDst);
        }
        return ucs;
    }

    /**
    * PixelDrawPipe helper methods
    */
    public Object drawRect(SunGraphics2D dest2D,
			 int dX, int dY, int w, int h, Object cachedState) {
        LockableRaster dst = null;
        PerformUnaryGraphicsPrimitive ugp = null;
        UnaryCachedState ucs = null;
        if (null != cachedState) {
            ucs = (UnaryCachedState)cachedState;
            if (ucs.the2D == dest2D) {
                dst = ucs.theLR;
                ugp = ucs.ugp;
                if (!ugp.id.equals(DrawRect.getMethodSignature())
                        || (ugp.dstType != dst.type)) {
                    ugp = null;
                }
            }
        }
        if (null == dst) {
            dst = new LockableRaster(dest2D);
        }
        if (null == ugp) {
            ugp = new PerformUnaryGraphicsPrimitive(
                            DrawRect.getMethodSignature(),
                            dst.type, dst.devID);
        }
        if ((null == ugp)
            || (null == ugp.coreOp)) {
            // REMIND: this disables unary strategy 2
            new Exception().printStackTrace();
            // REMIND: should have exception thrown
            return null;
        }
        if (ucs == null) {
            ucs = new UnaryCachedState(dest2D, dst, ugp);
        }
        LockableRaster  coreDst = dst;
        int dstX = dX;
        int dstY = dY;
        if (null != ugp.readDstOp) {
            coreDst = convertFrom(ugp.readDstOp, dst, dstX, dstY, w, h);
            dX = 0;
            dY = 0;
        }
        try {
            DrawRect dr = (DrawRect)ugp.coreOp;
            coreDst.lock(dr.getLockFlags()[0], dX, dY, w+1, h+1);
            dr.DrawRect(coreDst.id, dest2D.getColor(),
                        dX, dY, w, h);
        } catch (RasterLockingException e) {
        } finally {
            coreDst.unlock();
        }
        if (null != ugp.writeDstOp) {
            convertTo(ugp.writeDstOp, dst, dstX, dstY, w, h, coreDst);
        }
        return ucs;
    }

//    public void drawRoundRect(SunGraphics2D sg,
//			      int x, int y, int w, int h,
//			      int aW, int aH) {
//	sg.shapepipe.draw(sg, new RoundRectangle2D.Float(x, y, w, h, aW, aH));
//    }
//
//    public void drawOval(SunGraphics2D sg,
//			 int x, int y, int w, int h) {
//	sg.shapepipe.draw(sg, new Ellipse2D.Float(x, y, w, h));
//    }
//
//    public void drawArc(SunGraphics2D sg,
//			int x, int y, int w, int h,
//			int start, int extent) {
//	sg.shapepipe.draw(sg, new Arc2D.Float(x, y, w, h,
//					      start, extent, Arc2D.OPEN));
//    }
//
    public Object drawPolyline(SunGraphics2D dest2D,
			     int xPoints[], int yPoints[],
			     int nPoints, Object cachedState) {
        LockableRaster dst = null;
        PerformUnaryGraphicsPrimitive ugp = null;
        UnaryCachedState ucs = null;
        if (null != cachedState) {
            ucs = (UnaryCachedState)cachedState;
            if (ucs.the2D == dest2D) {
                dst = ucs.theLR;
                ugp = ucs.ugp;
                if (!ugp.id.equals(DrawPolygons.getMethodSignature())
                        || (ugp.dstType != dst.type)) {
                    ugp = null;
                }
            }
        }
        if (null == dst) {
            dst = new LockableRaster(dest2D);
        }
        if (null == ugp) {
            ugp = new PerformUnaryGraphicsPrimitive(
                            DrawPolygons.getMethodSignature(),
                            dst.type, dst.devID);
        }
        if ((null == ugp)
            || (null == ugp.coreOp)) {
            // REMIND: this disables unary strategy 2
            new Exception().printStackTrace();
            // REMIND: should have exception thrown
            return null;
        }
        if (ucs == null) {
            ucs = new UnaryCachedState(dest2D, dst, ugp);
        }
        LockableRaster  coreDst = dst;
        Rectangle clip = dest2D.getCompBounds();
        int dX = clip.x;
        int dY = clip.y;
        int w = clip.width;
        int h = clip.height;
        int dstX = dX;
        int dstY = dY;
        if (null != ugp.readDstOp) {
            coreDst = convertFrom(ugp.readDstOp, dst, dstX, dstY, w, h);
            dX = 0;
            dY = 0;
        }
        try {
            int nPointsArray[] = { nPoints };
            DrawPolygons dp = (DrawPolygons)ugp.coreOp;
            coreDst.lock(dp.getLockFlags()[0], dX, dY, w, h);
            dp.DrawPolygons(coreDst.id, dest2D.getColor(),
                            xPoints, yPoints, nPointsArray, 1, false);
        } catch (RasterLockingException e) {
        } finally {
            coreDst.unlock();
        }
        if (null != ugp.writeDstOp) {
            convertTo(ugp.writeDstOp, dst, dstX, dstY, w, h, coreDst);
        }
        return ucs;
    }

    public Object drawPolygon(SunGraphics2D dest2D,
			    int xPoints[], int yPoints[],
			    int nPoints, Object cachedState) {
        LockableRaster dst = null;
        PerformUnaryGraphicsPrimitive ugp = null;
        UnaryCachedState ucs = null;
        if (null != cachedState) {
            ucs = (UnaryCachedState)cachedState;
            if (ucs.the2D == dest2D) {
                dst = ucs.theLR;
                ugp = ucs.ugp;
                if (!ugp.id.equals(DrawPolygons.getMethodSignature())
                        || (ugp.dstType != dst.type)) {
                    ugp = null;
                }
            }
        }
        if (null == dst) {
            dst = new LockableRaster(dest2D);
        }
        if (null == ugp) {
            ugp = new PerformUnaryGraphicsPrimitive(
                            DrawPolygons.getMethodSignature(),
                            dst.type, dst.devID);
        }
        if ((null == ugp)
            || (null == ugp.coreOp)) {
            // REMIND: this disables unary strategy 2
            new Exception().printStackTrace();
            // REMIND: should have exception thrown
            return null;
        }
        if (ucs == null) {
            ucs = new UnaryCachedState(dest2D, dst, ugp);
        }
        LockableRaster  coreDst = dst;
        Rectangle clip = dest2D.getCompBounds();
        int dX = clip.x;
        int dY = clip.y;
        int w = clip.width;
        int h = clip.height;
        int dstX = dX;
        int dstY = dY;
        if (null != ugp.readDstOp) {
            coreDst = convertFrom(ugp.readDstOp, dst, dstX, dstY, w, h);
            dX = 0;
            dY = 0;
        }
        try {
            int nPointsArray[] = { nPoints };
            DrawPolygons dp = (DrawPolygons)ugp.coreOp;
            coreDst.lock(dp.getLockFlags()[0], dX, dY, w, h);
            dp.DrawPolygons(coreDst.id, dest2D.getColor(),
                            xPoints, yPoints, nPointsArray, 1, true);
        } catch (RasterLockingException e) {
        } finally {
            coreDst.unlock();
        }
        if (null != ugp.writeDstOp) {
            convertTo(ugp.writeDstOp, dst, dstX, dstY, w, h, coreDst);
        }
        return ucs;
    }

    /**
    * PixelFillPipe helper methods
    */
//    public void fillRoundRect(SunGraphics2D sg,
//			      int x, int y, int w, int h,
//			      int aW, int aH) {
//	sg.shapepipe.fill(sg, new RoundRectangle2D.Float(x, y, w, h, aW, aH));
//    }
//
//    public void fillOval(SunGraphics2D sg,
//			 int x, int y, int w, int h) {
//	sg.shapepipe.fill(sg, new Ellipse2D.Float(x, y, w, h));
//    }
//
//    public void fillArc(SunGraphics2D sg,
//			int x, int y, int w, int h,
//			int start, int extent) {
//	sg.shapepipe.fill(sg, new Arc2D.Float(x, y, w, h,
//					      start, extent, Arc2D.PIE));
//    }
//
    public Object fillPolygon(SunGraphics2D dest2D,
			    int xPoints[], int yPoints[],
			    int nPoints, Object cachedState) {
        LockableRaster dst = null;
        PerformUnaryGraphicsPrimitive ugp = null;
        UnaryCachedState ucs = null;
        if (null != cachedState) {
            ucs = (UnaryCachedState)cachedState;
            if (ucs.the2D == dest2D) {
                dst = ucs.theLR;
                ugp = ucs.ugp;
                if (!ugp.id.equals(FillPolygons.getMethodSignature())
                        || (ugp.dstType != dst.type)) {
                    ugp = null;
                }
            }
        }
        if (null == dst) {
            dst = new LockableRaster(dest2D);
        }
        if (null == ugp) {
            ugp = new PerformUnaryGraphicsPrimitive(
                            FillPolygons.getMethodSignature(),
                            dst.type, dst.devID);
        }
        if ((null == ugp)
            || (null == ugp.coreOp)) {
            // REMIND: this disables unary strategy 2
            new Exception().printStackTrace();
            // REMIND: should have exception thrown
            return null;
        }
        if (ucs == null) {
            ucs = new UnaryCachedState(dest2D, dst, ugp);
        }
        LockableRaster  coreDst = dst;
        Rectangle clip = dest2D.getCompBounds();
        int dX = clip.x;
        int dY = clip.y;
        int w = clip.width;
        int h = clip.height;
        int dstX = dX;
        int dstY = dY;
        if (null != ugp.readDstOp) {
            coreDst = convertFrom(ugp.readDstOp, dst, dstX, dstY, w, h);
            dX = 0;
            dY = 0;
        }
        try {
            int nPointsArray[] = {nPoints};
            FillPolygons fp = (FillPolygons)ugp.coreOp;
            coreDst.lock(fp.getLockFlags()[0], dX, dY, w, h);
            fp.FillPolygons(coreDst.id, dest2D.getColor(),
                            xPoints, yPoints, nPointsArray, 1, true);
        } catch (RasterLockingException e) {
        } finally {
            coreDst.unlock();
        }
        if (null != ugp.writeDstOp) {
            convertTo(ugp.writeDstOp, dst, dstX, dstY, w, h, coreDst);
        }
        return ucs;
    }

private void drawTextOutline(SunGraphics2D dest2D, String s, float x, float y)
{
    Font font = dest2D.getFont();
    float ptSize = font.getSize2D();
    FontRenderContext frc = dest2D.getFontRenderContext();
    AffineTransform tx = (AffineTransform)frc.getTransform();
    tx.scale(ptSize, ptSize);
    tx.preConcatenate(font.getTransform());
//  dest2D.draw will add the following transform...
//    tx.preConcatenate(dest2D.transform);
    double matrix[] = { tx.getScaleX(), tx.getShearY(),
                        tx.getShearX(), tx.getScaleY()};
    x += (float)tx.getTranslateX();
    y += (float)tx.getTranslateY();
    GeneralPath dstShape = new GeneralPath(GeneralPath.WIND_EVEN_ODD);
    dstShape = (GeneralPath)NativeFontWrapper.drawStringOutline(
                                        s, x, y, font, matrix,
                                        frc.isAntiAliased(),
                                        frc.usesFractionalMetrics(),
                                        dstShape);
    dest2D.shapepipe.fill(dest2D, dstShape);
}

private void drawGlyphVectorOutline(SunGraphics2D dest2D,
                                    GlyphVector gv, float x, float y)
{
    Font font = gv.getFont();
    float ptSize = font.getSize2D();
    FontRenderContext frc = gv.getFontRenderContext();
    AffineTransform tx = (AffineTransform)frc.getTransform();
    tx.scale(ptSize, ptSize);
    tx.preConcatenate(font.getTransform());
//  dest2D.draw will add the following transform...
//    tx.preConcatenate(dest2D.transform);
    double matrix[] = { tx.getScaleX(), tx.getShearY(),
                        tx.getShearX(), tx.getScaleY()};
    x += (float)tx.getTranslateX();
    y += (float)tx.getTranslateY();
    GeneralPath dstShape = new GeneralPath(GeneralPath.WIND_EVEN_ODD);
    dstShape = (GeneralPath)NativeFontWrapper.getGlyphVectorOutline(
                                        gv, font, matrix,
                                        frc.isAntiAliased(),
                                        frc.usesFractionalMetrics(),
                                        x, y, dstShape);
    dest2D.shapepipe.fill(dest2D, dstShape);
}

    private class TextCachedState extends UnaryCachedState {
        double [] matrix;
        double [] devTX;
        boolean isAntiAliased;
        boolean usesFractionalMetrics;
        Font    theFont;
        Color   theColor;
        FontRenderContext frc;

        TextCachedState(
                SunGraphics2D   the2D,
                LockableRaster  theLR,
                PerformUnaryGraphicsPrimitive   ugp)
        {
            super(the2D, theLR, ugp);

            theFont = the2D.getFont();
            float ptSize = theFont.getSize2D();
            frc = the2D.getFontRenderContext();
            AffineTransform tx = (AffineTransform)frc.getTransform();
            tx.scale(ptSize, ptSize);
            tx.preConcatenate(theFont.getTransform());
            tx.preConcatenate(the2D.transform);
            matrix = new double[6];
            tx.getMatrix(matrix);
            devTX = new double[6];
            the2D.transform.getMatrix(devTX);
            theColor = the2D.getColor();
        }

        double [] getOrigin(float x, float y) {
            double origin[] = {x, y};
            the2D.transform.deltaTransform(origin, 0, origin, 0, 1);
            origin[0] += matrix[4];
            origin[1] += matrix[5];
            return origin;
        }
    }

// implementation of textpipe with state caching
    public Object drawString(SunGraphics2D g2d,
            String s, float x, float y, Object cachedState) {
        // REMIND pipe override for testing
        if (!g2d.simpleText) {
            drawTextOutline(g2d, s, x, y);
            return null;
        }
        LockableRaster dst = null;
        PerformUnaryGraphicsPrimitive ugp = null;
        TextCachedState tcs = null;
        if (null != cachedState) {
            tcs = (TextCachedState)cachedState;
            dst = tcs.theLR;
            ugp = tcs.ugp;
            if (!ugp.id.equals(DrawString.getMethodSignature())
                || (ugp.dstType != dst.type)) {
                ugp = null;
            }
        }
        if (null == dst) {
            dst = new LockableRaster(g2d);
        }
	if (dst.type == GraphicsPrimitive.TYPE_BYTE_INDEXED
                && g2d instanceof HackWorkAround2DIndexed)
        {
            drawTextOutline(g2d, s, x, y);
            return null;
        }
        if (null == ugp) {
            ugp = new PerformUnaryGraphicsPrimitive(
                            DrawString.getMethodSignature(),
                            dst.type, dst.devID);
        }
        if ((null == ugp)
                || (null == ugp.coreOp)
                || (null != ugp.readDstOp)
                || (null != ugp.writeDstOp)) {
            // default case ... draw text as a shape.
            drawTextOutline(g2d, s, x, y);
            return null;
        }
        if (tcs == null) {
            tcs = new TextCachedState(g2d, dst, ugp);
        }
        LockableRaster  coreDst = dst;
        Rectangle clip = g2d.getCompBounds();
        int dX = clip.x;
        int dY = clip.y;
        int w = clip.width;
        int h = clip.height;

	double origin[] = tcs.getOrigin(x, y);
        x = (float)origin[0];
        y = (float)origin[1];
        try {
            DrawString ds = (DrawString)ugp.coreOp;
            coreDst.lock(ds.getLockFlags()[0], dX, dY, w, h);
            ds.DrawString(coreDst.id, s, x, y, tcs.theColor,
                            tcs.theFont, tcs.matrix,
                            tcs.frc.isAntiAliased(),
                            tcs.frc.usesFractionalMetrics());
        } catch (RasterLockingException e) {
        } finally {
            coreDst.unlock();
        }
        return tcs;
    }

    public Object drawGlyphVector(SunGraphics2D g2d,
            GlyphVector gv, float x, float y, Object cachedState) {
        // REMIND pipe override for testing
        if (!g2d.simpleText) {
            drawGlyphVectorOutline(g2d, gv, x, y);
            return null;
        }
        LockableRaster dst = null;
        PerformUnaryGraphicsPrimitive ugp = null;
        TextCachedState tcs = null;
        if (null != cachedState) {
            tcs = (TextCachedState)cachedState;
            dst = tcs.theLR;
            ugp = tcs.ugp;
            if (!ugp.id.equals(DrawGlyphVector.getMethodSignature())
                || (ugp.dstType != dst.type)) {
                ugp = null;
            }
        }
        if (null == dst) {
            dst = new LockableRaster(g2d);
        }
	if (dst.type == GraphicsPrimitive.TYPE_BYTE_INDEXED
                && g2d instanceof HackWorkAround2DIndexed)
        {
            drawGlyphVectorOutline(g2d, gv, x, y);
            return null;
        }
        if (null == ugp) {
            ugp = new PerformUnaryGraphicsPrimitive(
                            DrawGlyphVector.getMethodSignature(),
                            dst.type, dst.devID);
        }
        if ((null == ugp)
                || (null == ugp.coreOp)
                || (null != ugp.readDstOp)
                || (null != ugp.writeDstOp)) {
            // default case ... draw text as a shape.
            drawGlyphVectorOutline(g2d, gv, x, y);
            return null;
        }
        if (tcs == null) {
            tcs = new TextCachedState(g2d, dst, ugp);
        }
        LockableRaster  coreDst = dst;
        Rectangle clip = g2d.getCompBounds();
        int dX = clip.x;
        int dY = clip.y;
        int w = clip.width;
        int h = clip.height;

        Font theFont = gv.getFont();
        FontRenderContext frc = gv.getFontRenderContext();
        float ptSize = theFont.getSize2D();
        AffineTransform tx = (AffineTransform)frc.getTransform();
        tx.scale(ptSize, ptSize);
        tx.preConcatenate(theFont.getTransform());
        tx.preConcatenate(g2d.transform);
        double [] matrix = new double[6];
        tx.getMatrix(matrix);

	double origin[] = tcs.getOrigin(x, y);
        x = (float)origin[0];
        y = (float)origin[1];

        try {
            DrawGlyphVector dgv = (DrawGlyphVector)ugp.coreOp;
	    gv.performDefaultLayout ();
            coreDst.lock(dgv.getLockFlags()[0], dX, dY, w, h);
            dgv.DrawGlyphVector(coreDst.id, gv, x, y,
                            tcs.theColor, theFont, matrix,
                            tcs.devTX,
                            frc.isAntiAliased(),
                            frc.usesFractionalMetrics());
        } catch (RasterLockingException e) {
        } finally {
            coreDst.unlock();
        }
        return tcs;
    }

    public Object drawChars(SunGraphics2D g2d,
            char data[], int offset, int length, int x, int y,
            Object cachedState) {
        // REMIND pipe override for testing
        if (!g2d.simpleText) {
            drawTextOutline(g2d, new String(data, offset, length), x, y);
            return null;
        }
        LockableRaster dst = null;
        PerformUnaryGraphicsPrimitive ugp = null;
        TextCachedState tcs = null;
        if (null != cachedState) {
            tcs = (TextCachedState)cachedState;
            dst = tcs.theLR;
            ugp = tcs.ugp;
            if (!ugp.id.equals(DrawChars.getMethodSignature())
                || (ugp.dstType != dst.type)) {
                ugp = null;
            }
        }
        if (null == dst) {
            dst = new LockableRaster(g2d);
        }
	if (dst.type == GraphicsPrimitive.TYPE_BYTE_INDEXED
                && g2d instanceof HackWorkAround2DIndexed)
        {
            drawTextOutline(g2d, new String(data, offset, length), x, y);
            return null;
        }
        if (null == ugp) {
            ugp = new PerformUnaryGraphicsPrimitive(
                            DrawChars.getMethodSignature(),
                            dst.type, dst.devID);
        }
        if ((null == ugp)
                || (null == ugp.coreOp)
                || (null != ugp.readDstOp)
                || (null != ugp.writeDstOp)) {
            // default case ... draw text as a shape.
            drawTextOutline(g2d, new String(data, offset, length), x, y);
            return null;
        }
        if (tcs == null) {
            tcs = new TextCachedState(g2d, dst, ugp);
        }
        LockableRaster  coreDst = dst;
        Rectangle clip = g2d.getCompBounds();
        int dX = clip.x;
        int dY = clip.y;
        int w = clip.width;
        int h = clip.height;

	double origin[] = tcs.getOrigin(x, y);
        int ux = (int)origin[0];
        int uy = (int)origin[1];
        try {
            DrawChars dc = (DrawChars)ugp.coreOp;
            coreDst.lock(dc.getLockFlags()[0], dX, dY, w, h);
            dc.DrawChars(coreDst.id, data, offset, length, ux, uy,
                            tcs.theColor, tcs.theFont, tcs.matrix,
                            tcs.frc.isAntiAliased(),
                            tcs.frc.usesFractionalMetrics());
        } catch (RasterLockingException e) {
        } finally {
            coreDst.unlock();
        }
        return tcs;
    }

    public Object drawBytes(SunGraphics2D g2d,
            byte data[], int offset, int length, int x, int y,
            Object cachedState) {
        // REMIND pipe override for testing
        if (!g2d.simpleText) {
            drawTextOutline(g2d,new String(data, 0, offset, length), x, y);
            return null;
        }
        LockableRaster dst = null;
        PerformUnaryGraphicsPrimitive ugp = null;
        TextCachedState tcs = null;
        if (null != cachedState) {
            tcs = (TextCachedState)cachedState;
            dst = tcs.theLR;
            ugp = tcs.ugp;
            if (!ugp.id.equals(DrawBytes.getMethodSignature())
                || (ugp.dstType != dst.type)) {
                ugp = null;
            }
        }
        if (null == dst) {
            dst = new LockableRaster(g2d);
        }
	if (dst.type == GraphicsPrimitive.TYPE_BYTE_INDEXED
                && g2d instanceof HackWorkAround2DIndexed)
        {
            drawTextOutline(g2d,new String(data, 0, offset, length), x, y);
            return null;
        }
        if (null == ugp) {
            ugp = new PerformUnaryGraphicsPrimitive(
                            DrawBytes.getMethodSignature(),
                            dst.type, dst.devID);
        }
        if ((null == ugp)
                || (null == ugp.coreOp)
                || (null != ugp.readDstOp)
                || (null != ugp.writeDstOp)) {
            // default case ... draw text as a shape.
            drawTextOutline(g2d,new String(data, 0, offset, length), x, y);
            return null;
        }
        if (tcs == null) {
            tcs = new TextCachedState(g2d, dst, ugp);
        }
        LockableRaster  coreDst = dst;
        Rectangle clip = g2d.getCompBounds();
        int dX = clip.x;
        int dY = clip.y;
        int w = clip.width;
        int h = clip.height;

	double origin[] = tcs.getOrigin(x, y);
        int ux = (int)origin[0];
        int uy = (int)origin[1];
        try {
            DrawBytes db = (DrawBytes)ugp.coreOp;
            coreDst.lock(db.getLockFlags()[0], dX, dY, w, h);
            db.DrawBytes(coreDst.id, data, offset, length, ux, uy,
                            tcs.theColor, tcs.theFont, tcs.matrix,
                            tcs.frc.isAntiAliased(),
                            tcs.frc.usesFractionalMetrics());
        } catch (RasterLockingException e) {
        } finally {
            coreDst.unlock();
        }
        return tcs;
    }
// end of implementation of textpipe

    /**
     * Returns the AlphaColorPipe object to perform composite operations
     * for solid colors.
     */

    public AlphaColorPipe getColorPipe() {
	return ARGBcolorpipe;
    }

    /**
     * Returns the preferred ColorModel to convert to when a source
     * image of the indicated srcModel needs to be composited.
     */
    public ColorModel getPreferredModel(ColorModel srcModel,
					Raster srcRaster,
					AlphaComposite ac,
                                        Color bg, boolean needsClip) {
        if (!needsClip && srcModel instanceof IndexColorModel &&
            srcModel.getColorSpace().isCS_sRGB()) {
            int rule = ac.getRule();
            float extraAlpha = ac.getAlpha();
            int transparency = srcModel.getTransparency();

            if (extraAlpha >= 1.0f) {
                if (rule == ac.SRC) {
                    return srcModel;
                }

                if (rule == ac.SRC_OVER && transparency == srcModel.OPAQUE) {
                    return srcModel;
                }
                if (rule == ac.SRC_OVER && transparency == srcModel.BITMASK) {
                    if (bg == null) {
                        return srcModel;
                    }
                    int trans = bg.getTransparency();
                    if (trans == srcModel.OPAQUE || trans == srcModel.BITMASK){
                        return srcModel;
                    }
                    // Fall through
                }
            }
            // Falls through to one of the default models
        }

        if (srcModel.getTransparency() == Transparency.OPAQUE) {
	    return XRGBModel;
	} else {
	    return ARGBModel;
	}
    }

/* REMIND: Remove these when FrameBuffer Architecture Completed */
    public static void
	ARGBpaintARGB(IntegerComponentRaster srcRaster, boolean srcPremul,
		      IntegerComponentRaster dstRaster,
                      int alphaRule, float extraAlpha,
		      byte[] clipAlpha, int alphaOffset, int alphaScan,
		      int srcX, int srcY, int dstX, int dstY, int w, int h,
                      int xorClr) {
        Color xorColor = new Color(xorClr, true);
        LockableRaster src = new LockableRaster(srcRaster, ARGBModel);
        LockableRaster dst = new LockableRaster(dstRaster, ARGBModel);
//        new Exception().printStackTrace();
        manager.performBinaryComposite(src, dst,
                                        srcX, srcY, dstX, dstY, w, h,
                                        clipAlpha, alphaOffset, alphaScan,
                                        alphaRule, extraAlpha,
                                        xorColor, null);
    }

// REMIND: legacy stuff which can be removed

    public static final ColorModel model = ColorModel.getRGBdefault();
    public static final ColorModel BitmaskARGBModel =
	new DirectColorModel(25, 0x00ff0000, 0x0000ff00, 0x000000ff,
                            0x01000000);
    public static final ColorModel ARGBModel = ColorModel.getRGBdefault();

    public static final ColorModel BitmaskABGRModel =
        new DirectColorModel(25, 0x000000ff, 0x0000ff00, 0x00ff0000,
                            0x01000000);
    public static final ColorModel ABGRModel =
        new DirectColorModel(32, 0x000000ff, 0x0000ff00, 0x00ff0000,
                            0xff000000);

    static final AlphaColorPipe ARGBcolorpipe = new AlphaColorPipe();
}
