/*
 * @(#)DuctusRenderer.java	1.7 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.java2d.pipe;

import java.awt.geom.AffineTransform;
import java.awt.BasicStroke;
import java.awt.Shape;
import java.awt.geom.PathIterator;
import sun.dc.path.PathConsumer;
import sun.dc.path.PathException;
import sun.dc.pr.Rasterizer;
import sun.dc.pr.PathStroker;
import sun.dc.pr.PathDasher;
import sun.dc.pr.PRException;

/**
 * This is the superclass of any class that needs to use the Ductus
 * Rasterizer.
 * This class provides utility routines to create, cache, initialize
 * and release Rasterizer objects.
 */
public class DuctusRenderer {
    static final int RasterizerCaps[] = {
	Rasterizer.BUTT, Rasterizer.ROUND, Rasterizer.SQUARE
    };

    static final int RasterizerCorners[] = {
	Rasterizer.MITER, Rasterizer.ROUND, Rasterizer.BEVEL
    };

    private static Rasterizer theRasterizer;

    public synchronized static Rasterizer getRasterizer() {
	Rasterizer r = theRasterizer;
	if (r == null) {
	    r = new Rasterizer();
	} else {
	    theRasterizer = null;
	}
	return r;
    }

    public synchronized static void dropRasterizer(Rasterizer r) {
	r.reset();
	theRasterizer = r;
    }

    /*
     * writeAlpha is not threadsafe (it uses a global table without
     * locking), so we must use this static synchronized accessor
     * method to serialize accesses to it.
     */
    public synchronized static void getAlpha(Rasterizer r, byte[] alpha,
					     int xstride, int ystride,
					     int offset)
	throws PRException
    {
	try {
	    r.writeAlpha(alpha, xstride, ystride, offset);
	} catch (InterruptedException e) {
	    Thread.currentThread().interrupt();
	}
    }

    /*
     * Returns a new PathConsumer that will take a path trajectory and
     * feed the stroked outline for that trajectory to the supplied
     * PathConsumer.
     */
    public static PathConsumer createStroker(PathConsumer consumer,
					     BasicStroke bs,
					     AffineTransform transform) {
	PathStroker stroker = new PathStroker(consumer);
	consumer = stroker;

	// REMIND: setPenFitting doesn't help if line width is exactly 0.
	// As a workaround we make sure line width is never less than .001
	stroker.setPenDiameter(Math.max(bs.getLineWidth(), 0.001f));
	float matrix[] = {1.0f, 0.0f, 0.0f, 1.0f, 0.5f, 0.5f};
	stroker.setOutputT6(matrix);
	if (transform != null) {
	    double dmatrix[] = new double[6];
	    transform.getMatrix(dmatrix);
	    for (int i = 0; i < 4; i++) {
		matrix[i] = (float) dmatrix[i];
	    }
	} else {
	    matrix = null;
	}
	stroker.setPenT4(matrix);
	stroker.setCaps(RasterizerCaps[bs.getEndCap()]);
	stroker.setCorners(RasterizerCorners[bs.getLineJoin()],
			   bs.getMiterLimit());
	float[] dashes = bs.getDashArray();
	if (dashes != null) {
	    PathDasher dasher = new PathDasher(stroker);
	    dasher.setDash(dashes, bs.getDashPhase());
	    dasher.setDashT4(matrix);
	    consumer = dasher;
	}
	stroker.setPenFitting(0.01f, 100);

	return consumer;
    }

    /*
     * Feed a path from a PathIterator to a Ductus PathConsumer.
     */
    public static void feedConsumer(PathIterator pi, PathConsumer consumer)
	throws PathException
    {
	consumer.beginPath();
	boolean pathClosed = false;
	float mx = 0.0f;
	float my = 0.0f;
	float point[]  = new float[6];

	while (!pi.isDone()) {
	    int type = pi.currentSegment(point);
	    if (pathClosed == true) {
		pathClosed = false;
		if (type != PathIterator.SEG_MOVETO) {
		    // Force current point back to last moveto point
		    consumer.beginSubpath(mx, my);
		}
	    }
	    switch (type) {
	    case PathIterator.SEG_MOVETO:
		mx = point[0];
		my = point[1];
		consumer.beginSubpath(point[0], point[1]);
		break;
	    case PathIterator.SEG_LINETO:
		consumer.appendLine(point[0], point[1]);
		break;
	    case PathIterator.SEG_QUADTO:
		// Quadratic curves take two points
		consumer.appendQuadratic(point[0], point[1],
					 point[2], point[3]);
		break;
	    case PathIterator.SEG_CUBICTO:
		// Cubic curves take three points
		consumer.appendCubic(point[0], point[1],
				     point[2], point[3],
				     point[4], point[5]);
		break;
	    case PathIterator.SEG_CLOSE:
		consumer.closedSubpath();
		pathClosed = true;
		break;
	    }
	    pi.next();
	}

	consumer.endPath();
    }

    /*
     * Returns a new Rasterizer that is ready to rasterize the given Shape.
     */
    public static Rasterizer createShapeRasterizer(PathIterator pi,
						   AffineTransform transform,
						   BasicStroke stroke) {
	Rasterizer r = getRasterizer();

        if (stroke != null) {
            r.setUsage(Rasterizer.STROKE);
	    // REMIND: setPenFitting doesn't help if line width is exactly 0.
	    // As a workaround we make sure line width is never less than .001
	    r.setPenDiameter(Math.max(stroke.getLineWidth(), .001f));
	    r.setPenDisplacement(0.5f, 0.5f);
	    float matrix[];
	    if (transform != null) {
		matrix = new float[4];
		double dmatrix[] = new double[6];
		transform.getMatrix(dmatrix);
		for (int i = 0; i < 4; i++) {
		    matrix[i] = (float) dmatrix[i];
		}
		r.setPenT4(matrix);
	    } else {
		matrix = null;
	    }
	    r.setCaps(RasterizerCaps[stroke.getEndCap()]);
	    r.setCorners(RasterizerCorners[stroke.getLineJoin()],
			 stroke.getMiterLimit());
	    float[] dashes = stroke.getDashArray();
	    if (dashes != null) {
		r.setDash(dashes, stroke.getDashPhase());
		r.setDashT4(matrix);
	    }
	    r.setPenFitting(0.01f, 15);
        } else {
            r.setUsage(pi.getWindingRule() == PathIterator.WIND_EVEN_ODD
                       ? Rasterizer.EOFILL
                       : Rasterizer.NZFILL);
        }

	r.beginPath();
	{
            boolean pathClosed = false;
            float mx = 0.0f;
            float my = 0.0f;
            float point[]  = new float[6];

	    while (!pi.isDone()) {
		int type = pi.currentSegment(point);
                if (pathClosed == true) {
                    pathClosed = false;
                    if (type != PathIterator.SEG_MOVETO) {
                        // Force current point back to last moveto point
                        r.beginSubpath(mx, my);
                    }
                }
		switch (type) {
		case PathIterator.SEG_MOVETO:
                    mx = point[0];
                    my = point[1];
		    r.beginSubpath(point[0], point[1]);
		    break;
		case PathIterator.SEG_LINETO:
		    r.appendLine(point[0], point[1]);
		    break;
		case PathIterator.SEG_QUADTO:
                    // Quadratic curves take two points
		    r.appendQuadratic(point[0], point[1],
				      point[2], point[3]);
		    break;
		case PathIterator.SEG_CUBICTO:
                    // Cubic curves take three points
		    r.appendCubic(point[0], point[1],
				  point[2], point[3],
				  point[4], point[5]);
		    break;
		case PathIterator.SEG_CLOSE:
		    r.closedSubpath();
                    pathClosed = true;
		    break;
		}
		pi.next();
	    }
	}

        try {
            r.endPath();
        } catch (PRException e) {
            e.printStackTrace();
        }

        return r;
    }
}
