/*
 * @(#)Region.java	1.3 98/07/17
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.java2d.pipe;

import java.awt.Rectangle;

/**
 * This class encapsulates a definition of a two dimensional region which
 * consists of a number of Y ranges each containing multiple X bands.
 */
public class Region {
    public static final int INIT_SIZE = 50;
    public static final int GROW_SIZE = 50;

    int[] bands = new int[INIT_SIZE];
    int endIndex;

    int prevYrow;
    int curYrow;

    int lox;
    int loy;
    int hix;
    int hiy;

    public Region() {
    }

    /*
     * Sets the rectangle of interest for storing and returning
     * region bands.
     */
    public void setOutputArea(Rectangle r) {
	setOutputArea(r.x, r.y, r.width, r.height);
    }

    /*
     * Sets the rectangle of interest for storing and returning
     * region bands.
     */
    public void setOutputArea(int x, int y, int w, int h) {
	lox = x;
	loy = y;
	hix = x + w;
	hiy = y + h;
    }

    /*
     * Appends the list of spans returned from the indicated
     * SpanIterator.  Each span must be at a higher starting
     * Y coordinate than the previous data or it must have a
     * Y range equal to the highest Y band in the region and a
     * higher X coordinate than any of the spans in that band.
     */
    public void appendSpans(SpanIterator si) {
	int[] box = new int[4];

	while (si.nextSpan(box)) {
	    appendSpan(box[0], box[1], box[2], box[3]);
	}

	endRow();
	calcBBox();
    }

    /*
     * Appends a single span defined by the 4 parameters
     * spanlox, spanloy, spanhix, spanhiy.
     * This span must be at a higher starting Y coordinate than
     * the previous data or it must have a Y range equal to the
     * highest Y band in the region and a higher X coordinate
     * than any of the spans in that band.
     */
    public void appendSpan(int spanlox, int spanloy,
			   int spanhix, int spanhiy) {
	if (spanlox < lox) spanlox = lox;
	if (spanloy < loy) spanloy = loy;
	if (spanhix > hix) spanhix = hix;
	if (spanhiy > hiy) spanhiy = hiy;
	if (spanhix <= spanlox || spanhiy <= spanloy) {
	    return;
	}

	if (endIndex == 0 || spanloy >= bands[curYrow + 1]) {
	    endRow();
	    needSpace(5);
	    bands[endIndex++] = spanloy;
	    bands[endIndex++] = spanhiy;
	    bands[endIndex++] = 0;
	} else if (spanloy == bands[curYrow] &&
		   spanhiy == bands[curYrow + 1] &&
		   spanlox >= bands[endIndex - 1]) {
	    if (spanlox == bands[endIndex - 1]) {
		bands[endIndex - 1] = spanhix;
		return;
	    }
	    needSpace(2);
	} else {
	    throw new InternalError("bad span");
	}
	bands[endIndex++] = spanlox;
	bands[endIndex++] = spanhix;
	bands[curYrow + 2]++;
    }

    void needSpace(int num) {
	if (endIndex + num >= bands.length) {
	    int[] newbands = new int[bands.length + GROW_SIZE];
	    System.arraycopy(bands, 0, newbands, 0, endIndex);
	    bands = newbands;
	}
    }

    void endRow() {
	int cur = curYrow;
	int prev = prevYrow;
	if (cur > prev) {
	    int[] bands = this.bands;
	    if (bands[prev + 1] == bands[cur] &&
		bands[prev + 2] == bands[cur + 2])
	    {
		int num = bands[cur + 2] * 2; 
		cur += 3;
		prev += 3;
		while (num > 0) {
		    if (bands[cur++] != bands[prev++]) {
			break;
		    }
		    num--;
		}
		if (num == 0) {
		    bands[prevYrow + 1] = bands[curYrow + 1];
		    endIndex = curYrow;
		    return;
		}
	    }
	}
	prevYrow = curYrow;
	curYrow = endIndex;
    }

    void calcBBox() {
	if (endIndex == 0) {
	    lox = loy = hix = hiy = 0;
	    return;
	}
	int lox = this.hix;
	int hix = this.lox;
	int hiyindex = 0;

	int i = 0;
	int[] bands = this.bands;
	while (i < endIndex) {
	    hiyindex = i;
	    int numbands = bands[i + 2];
	    i += 3;
	    if (lox > bands[i]) {
		lox = bands[i];
	    }
	    i += numbands * 2;
	    if (hix < bands[i - 1]) {
		hix = bands[i - 1];
	    }
	}

	this.lox = lox;
	this.loy = bands[0];
	this.hix = hix;
	this.hiy = bands[hiyindex + 1];
    }

    /*
     * Gets the bbox of the available path segments, clipped to the
     * OutputArea.
     */
    public void getBounds(int pathbox[]) {
	pathbox[0] = lox;
	pathbox[1] = loy;
	pathbox[2] = hix;
	pathbox[3] = hiy;
    }

    /*
     * Gets an iterator object to iterate over the spans in this region.
     */
    public RegionIterator getIterator() {
	return new RegionIterator(this);
    }
}
