/*
 * @(#)SpanShapeRenderer.java	1.8 98/09/22
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.java2d.pipe;

import sun.java2d.SunGraphics2D;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.awt.geom.Rectangle2D;

/**
 * This class is used to convert raw geometry into a span iterator
 * object using a simple flattening polygon scan converter.
 * The iterator can be passed on to special SpanFiller loops to
 * perform the actual rendering.
 */
public abstract class SpanShapeRenderer implements ShapeDrawPipe {
    public static class Composite extends SpanShapeRenderer {
	CompositePipe comppipe;

	public Composite(CompositePipe pipe) {
	    comppipe = pipe;
	}

	public Object startSequence(SunGraphics2D sg, Shape s,
				    Rectangle devR, int[] bbox) {
	    return comppipe.startSequence(sg, s, devR, bbox);
	}

	public void renderSpan(Object ctx, int y, int x1, int x2) {
	    comppipe.renderPathTile(ctx, null, 0, x2 - x1,
				    x1, y, x2 - x1, 1);
	}

	public void renderBox(Object ctx, int x, int y, int w, int h) {
	    comppipe.renderPathTile(ctx, null, 0, w, x, y, w, h);
	}

	public void endSequence(Object ctx) {
	    comppipe.endSequence(ctx);
	}
    }

    public static class Simple extends SpanShapeRenderer {
	SimpleRenderPipe srpipe;

	public Simple(SimpleRenderPipe pipe) {
	    srpipe = pipe;
	}

	public Object startSequence(SunGraphics2D sg, Shape s,
				    Rectangle devR, int[] bbox) {
	    return sg;
	}

	public void renderSpan(Object ctx, int y, int x1, int x2) {
	    srpipe.drawLine((SunGraphics2D) ctx, x1, y, x2 - 1, y);
	}

	public void renderBox(Object ctx, int x, int y, int w, int h) {
	    srpipe.fillRect((SunGraphics2D) ctx, x, y, w, h);
	}

	public void endSequence(Object ctx) {
	}
    }

    public void draw(SunGraphics2D sg, Shape s) {
	renderPath(sg, sg.stroke.createStrokedShape(s));
    }

    public static final int NON_RECTILINEAR_TRANSFORM_MASK =
	(AffineTransform.TYPE_GENERAL_TRANSFORM |
	 AffineTransform.TYPE_GENERAL_ROTATION);

    public void fill(SunGraphics2D sg, Shape s) {
	if (s instanceof Rectangle2D &&
	    (sg.transform.getType() & NON_RECTILINEAR_TRANSFORM_MASK) == 0)
	{
	    renderRect(sg, (Rectangle2D) s);
	    return;
	}
	renderPath(sg, s);
    }

    public abstract Object startSequence(SunGraphics2D sg, Shape s,
					 Rectangle devR, int[] bbox);

    public abstract void renderSpan(Object ctx, int y, int x1, int x2);

    public abstract void renderBox(Object ctx, int x, int y, int w, int h);

    public abstract void endSequence(Object ctx);

    public void renderRect(SunGraphics2D sg, Rectangle2D r) {
	double corners[] = {
	    r.getX(), r.getY(), r.getWidth(), r.getHeight(),
	};
	corners[2] += corners[0];
	corners[3] += corners[1];
	if (corners[2] <= corners[0] || corners[3] <= corners[1]) {
	    return;
	}
	sg.transform.transform(corners, 0, corners, 0, 2);
	if (corners[2] < corners[0]) {
	    double t = corners[2];
	    corners[2] = corners[0];
	    corners[0] = t;
	}
	if (corners[3] < corners[1]) {
	    double t = corners[3];
	    corners[3] = corners[1];
	    corners[1] = t;
	}
	int abox[] = {
	    (int) corners[0],
	    (int) corners[1],
	    (int) corners[2],
	    (int) corners[3],
	};
	Rectangle devR = new Rectangle(abox[0], abox[1],
				       abox[2] - abox[0],
				       abox[3] - abox[1]);
	Rectangle clipR = sg.getCompBounds();
	abox[0] = Math.max(abox[0], clipR.x);
	abox[1] = Math.max(abox[1], clipR.y);
	abox[2] = Math.min(abox[2], clipR.x + clipR.width);
	abox[3] = Math.min(abox[3], clipR.y + clipR.height);
	if (abox[0] >= abox[2] || abox[1] >= abox[3]) {
	    return;
	}
	Object context = startSequence(sg, r, devR, abox);
	if (sg.clipRegion == null) {
	    renderBox(context, abox[0], abox[1],
		      abox[2] - abox[0],
		      abox[3] - abox[1]);
	} else {
	    RegionIterator ri = sg.clipRegion.getIterator();
	    int lox = abox[0];
	    int loy = abox[1];
	    int hix = abox[2];
	    int hiy = abox[3];
	    while (ri.nextYRange(abox)) {
		if (abox[0] < loy) {
		    abox[0] = loy;
		}
		if (abox[1] > hiy) {
		    if (abox[0] >= hiy) {
			break;
		    }
		    abox[1] = hiy;
		}
		if (abox[0] >= abox[1]) {
		    continue;
		}
		while (ri.nextXBand(abox)) {
		    if (abox[2] < lox) {
			abox[2] = lox;
		    }
		    if (abox[3] > hix) {
			if (abox[2] >= hix) {
			    break;
			}
			abox[3] = hix;
		    }
		    if (abox[2] >= abox[3]) {
			continue;
		    }
		    renderBox(context, abox[2], abox[0],
			      abox[3] - abox[2],
			      abox[1] - abox[0]);
		}
	    }
	}
	endSequence(context);
    }

    public void renderPath(SunGraphics2D sg, Shape s) {
	Object context = null;

	Rectangle clipR = sg.getCompBounds();
	
	ShapeSpanIterator sr = new ShapeSpanIterator();
	try {
	    sr.setOutputArea(clipR);
	    sr.appendPath(s.getPathIterator(sg.transform));
	    int abox[] = new int[4];

	    try {
		sr.getPathBox(abox);
		Rectangle devR = new Rectangle(abox[0], abox[1],
					       abox[2] - abox[0],
					       abox[3] - abox[1]);
		abox[0] = Math.max(abox[0], clipR.x);
		abox[1] = Math.max(abox[1], clipR.y);
		abox[2] = Math.min(abox[2], clipR.x + clipR.width);
		abox[3] = Math.min(abox[3], clipR.y + clipR.height);
		if (abox[0] >= abox[2] || abox[1] >= abox[3]) {
		    return;
		}
		context = startSequence(sg, s, devR, abox);

		if (sg.clipRegion == null) {
		    spanLoop(context, sr, abox);
		} else {
		    spanClipLoop(context, sr, sg.clipRegion, abox);
		}
	    } finally {
		if (context != null) {
		    endSequence(context);
		}
	    }
	} finally {
	    sr.dispose();
	}
    }

    public void spanLoop(Object ctx, ShapeSpanIterator sr, int[] abox) {
	while (sr.nextSpan(abox)) {
	    renderSpan(ctx, abox[1], abox[0], abox[2]);
	}
    }

    public void spanClipLoop(Object ctx, ShapeSpanIterator sr,
			     Region r, int[] abox) {
	int[] rbox = new int[4];
	if (!sr.nextSpan(abox)) {
	    return;
	}
	int spany = abox[1];
	int spanlox = abox[0];
	int spanhix = abox[2];
	RegionIterator riy = r.getIterator();
	if (!riy.nextYRange(rbox)) {
	    return;
	}
	int regloy = rbox[0];
	int reghiy = rbox[1];
	RegionIterator rix = riy.createCopy();
	int reglox = Integer.MIN_VALUE;
	int reghix = Integer.MIN_VALUE;
	while (true) {
	    if (spany < regloy) {
		if (!sr.nextSpan(abox)) {
		    break;
		}
		if (spany != abox[1]) {
		    spany = abox[1];
		    rix.copyStateFrom(riy);
		    reglox = reghix = Integer.MIN_VALUE;
		}
		spanlox = abox[0];
		spanhix = abox[2];
		continue;
	    }
	    if (spany >= reghiy) {
		if (!riy.nextYRange(rbox)) {
		    break;
		}
		rix.copyStateFrom(riy);
		regloy = rbox[0];
		reghiy = rbox[1];
		reglox = reghix = Integer.MIN_VALUE;
		continue;
	    }
	    if (spanhix <= reglox) {
		if (!sr.nextSpan(abox)) {
		    break;
		}
		if (spany != abox[1]) {
		    spany = abox[1];
		    rix.copyStateFrom(riy);
		    reglox = reghix = Integer.MIN_VALUE;
		}
		spanlox = abox[0];
		spanhix = abox[2];
		continue;
	    }
	    if (spanlox >= reghix) {
		if (!rix.nextXBand(rbox)) {
		    rix.copyStateFrom(riy);
		    regloy = spany + 1;
		    reglox = reghix = Integer.MIN_VALUE;
		    continue;
		}
		reglox = rbox[2];
		reghix = rbox[3];
		continue;
	    }
	    // Assert: regloy <= spany < reghiy
	    // Assert: reglox < spanhix && reghix > spanlox
	    if (spanlox < reglox) {
		spanlox = reglox;
	    }
	    if (spanhix < reghix) {
		renderSpan(ctx, spany, spanlox, spanhix);
		reglox = spanhix;
	    } else {
		renderSpan(ctx, spany, spanlox, reghix);
		spanlox = reghix;
	    }
	}
    }
}
