/*
 * @(#)JdbcOdbcConnection.java	1.18 98/05/13
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
//----------------------------------------------------------------------------
//
// Module:      JdbcOdbcConnection.java
//
// Description: Impementation of the Connection interface class
//
// Product:     JDBCODBC (Java DataBase Connectivity using
//              Open DataBase Connectivity)
//
// Author:      Karl Moss
//
// Date:        March, 1996
//
// Copyright:   1990-1996 INTERSOLV, Inc.
//              This software contains confidential and proprietary
//              information of INTERSOLV, Inc.
//----------------------------------------------------------------------------

package sun.jdbc.odbc;

import java.util.Hashtable;
import java.sql.*;
import sun.io.*;

public class JdbcOdbcConnection
	extends		JdbcOdbcObject
	implements	JdbcOdbcConnectionInterface
{

	//====================================================================
	// Public methods
	//====================================================================

	//--------------------------------------------------------------------
	// Constructor
	// Perform any necessary initialization.
	//--------------------------------------------------------------------

	public JdbcOdbcConnection (
		JdbcOdbc odbcApi,
		int env,
		JdbcOdbcDriverInterface driver)
	{
		// Save the ODBC API object, and our environment handle

		OdbcApi = odbcApi;
		myDriver = driver;
		hEnv = env;

		// Init the connection handle, URL

		hDbc = 0;
		URL = null;

		// Init the warning list
		
		lastWarning = null;

		// Default the connection to closed
		
		closed = true;
	}

	//--------------------------------------------------------------------
	// finalize
	// Perform any cleanup when this object is garbage collected
	//--------------------------------------------------------------------

	protected void finalize ()
	{
		if (isTracing ()) {
			trace ("Connection.finalize " + this);
		}

		try {
			close ();
		}
		catch (SQLException ex) {
			// If an exception is thrown, ignore it
		}
	}

	//--------------------------------------------------------------------
	// initialize
	// Finish initializing the connection object.  This includes 
	// allocating a connection handle, and performing the connect.
	//--------------------------------------------------------------------

	public void initialize (
		String dsn, 
		java.util.Properties info,
		int timeout)
		throws SQLException
	{
		String	uid;
		String	pwd;
		String	connectString = "";

	
		// If we don't have a connection handle for this object,
		// allocate one

		if (closed) {
			hDbc = myDriver.allocConnection (hEnv);
		}

		// If a login timeout was supplied, set the new login
		// timeout value
		if (timeout > 0) {
			setLoginTimeout (timeout);
		}

		// Get the username and password from the Properties
		// table

		uid = info.getProperty ("user", "");
		pwd = info.getProperty ("password", "");
		OdbcApi.charSet = info.getProperty ("charSet", System.getProperty("file.encoding"));

		// Now build the connection string
		
		connectString = "DSN=" + dsn;

		if (uid != null) {
			connectString += ";UID=" + uid;
		}
		if (pwd != null) {
			connectString += ";PWD=" + pwd;
		}

		// Perform the connect.  If we get a SQLWarning, save it;
		// the application will have to poll for it

		try {
			OdbcApi.SQLDriverConnect (hDbc, connectString);
		}
		catch (SQLWarning ex) {
			lastWarning = ex;
		}
		catch (SQLException ex) {

			// If an exception was raised, gracefully close
			// down the connection.  This includes freeing the
			// connection handle

			myDriver.closeConnection (hDbc);

			// Re-throw the exception

			throw ex;
		}

		// We are now open!

		closed = false;

		// Create a new Hashtable to store each of our Statement objects

		statements = new Hashtable ();

		// If tracing, trace the driver name and version

		if (isTracing()) {
			DatabaseMetaData dma = getMetaData ();
	
			trace("Driver name:    " + dma.getDriverName ());
			trace("Driver version: " + dma.getDriverVersion ());
		}

		// Now that the connection has been established, get some default
		// information.  This includes building a Hashtable with all of the
		// SQL type information

		buildTypeInfo ();
	}

	//--------------------------------------------------------------------
	// createStatement
	// Returns an SQL Statement object
	//--------------------------------------------------------------------

	public Statement createStatement ()
		throws SQLException {

		if (isTracing ()) {
			trace ("*Connection.createStatement");
		}

		// Create a new Statement object

		JdbcOdbcStatement stmt = new JdbcOdbcStatement (this);

		// Initialize the object.  This will also allocate a
		// statement handle.

		stmt.initialize (OdbcApi, hDbc);

		// Keep a reference of this statement object

		registerStatement (stmt);

		return stmt;
	}

	//--------------------------------------------------------------------
	// prepareStatement
	// prepareStatement creates a precompiled SQL PreparedStatement object.
	//--------------------------------------------------------------------
	
	public PreparedStatement prepareStatement (
		String sql)
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*Connection.prepareStatement (" + sql + ")");
		}

		int hStmt;
		JdbcOdbcPreparedStatement ps = null;
		SQLWarning	warning = null;

		// First, allocate a statement handle for the operation

		hStmt = OdbcApi.SQLAllocStmt (hDbc);

		// Call SQLPrepare
		try {
			OdbcApi.SQLPrepare (hStmt, sql);
		}
		catch (SQLWarning ex) {

			// Save pointer to warning and save with
			// PreparedStatement object once it is created.

			warning = ex;
		}
		catch (SQLException ex) {

			// If we got an exception, we need to clean up
			// the statement handle we allocated, then
			// re-throw the exception

			OdbcApi.SQLFreeStmt (hStmt, OdbcDef.SQL_DROP);

			throw ex;
		}
				
		// Now, create a new PreparedStatement object, and initialize

		ps = new JdbcOdbcPreparedStatement (this);
		ps.initialize (OdbcApi, hDbc, hStmt, typeInfo);
		ps.setWarning (warning);

		// Keep a reference of this statement object

		registerStatement (ps);

		// Return the prepared statement

		return ps;
	}

	//--------------------------------------------------------------------
	// prepareCall
	// prepareCall create a pre-compiled SQL statement that is
	// a call on a stored procedure.
	//--------------------------------------------------------------------

	public CallableStatement prepareCall (
		String sql)
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*Connection.prepareCall (" + sql + ")");
		}

		int hStmt;
		JdbcOdbcCallableStatement cs = null;
		SQLWarning	warning = null;

		// First, allocate a statement handle for the operation

		hStmt = OdbcApi.SQLAllocStmt (hDbc);

		// Call SQLPrepare
		try {
			OdbcApi.SQLPrepare (hStmt, sql);
		}
		catch (SQLWarning ex) {

			// Save pointer to warning and save with
			// PreparedStatement object once it is created.

			warning = ex;
		}
		catch (SQLException ex) {

			// If we got an exception, we need to clean up
			// the statement handle we allocated, then
			// re-throw the exception

			OdbcApi.SQLFreeStmt (hStmt, OdbcDef.SQL_DROP);

			throw ex;
		}
				
		// Now, create a new CallableStatement object, and initialize

		cs = new JdbcOdbcCallableStatement (this);
		cs.initialize (OdbcApi, hDbc, hStmt, typeInfo);
		cs.setWarning (warning);

		// Keep a reference of this statement object

		registerStatement (cs);

		// Return the callable statement

		return cs;
	}
						
	//--------------------------------------------------------------------
	// nativeSQL
	// Convert the given generic SQL statement to the drivers native SQL.
	//--------------------------------------------------------------------

	public String nativeSQL (
		String query)
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*Connection.nativeSQL (" + query + ")");
		}

		String nativeQuery;

		try {
			nativeQuery = OdbcApi.SQLNativeSql (hDbc, query);
		}
		catch (SQLException ex) {

			// If an exception is thrown, simply return the
			// original string
			nativeQuery = query;
		}
		return nativeQuery;
	}

	//--------------------------------------------------------------------
	// setAutoCommit
	// If "autoCommit" is true, then all subsequent SQL statements will
	// be executed and committed as individual transactions.  Otherwise
	// (if "autoCommit" is false) then subsequent SQL statements will
	// all be part of the same transaction , which must be expicitly
	// committed with either a "commit" or "rollback" call.
  	// By default new connections are initialized with autoCommit "true".
	//--------------------------------------------------------------------
	
	public void setAutoCommit(
		boolean enableAutoCommit)
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*Connection.setAutoCommit (" + 
				enableAutoCommit + ")");
		}

		int	mode = OdbcDef.SQL_AUTOCOMMIT_ON;

		// Validate that the connection is valid

		validateConnection ();

		if (!enableAutoCommit) {
			mode = OdbcDef.SQL_AUTOCOMMIT_OFF;
		}

		OdbcApi.SQLSetConnectOption (hDbc, OdbcDef.SQL_AUTOCOMMIT,
				mode);
	}

	//--------------------------------------------------------------------
	// getAutoCommit
	// Test whether "autoCommit" is true.
	//--------------------------------------------------------------------

	public boolean getAutoCommit()
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*Connection.getAutoCommit");
		}

		int	mode;
		boolean rc = false;

		// Validate that the connection is valid

		validateConnection ();

		mode = OdbcApi.SQLGetConnectOption (hDbc,
				OdbcDef.SQL_AUTOCOMMIT);

		if (mode == OdbcDef.SQL_AUTOCOMMIT_ON) {
			rc = true;
		}
		return rc;
	}


	//--------------------------------------------------------------------
	// commit
	// You can use commit or rollback to commit or abort a transaction.
	// Note that by default all active Statement or ResultSet objects
	// associated with a Connection will be closed when it is committed
	// or aborted.
	//--------------------------------------------------------------------
	
	public void commit ()
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*Connection.commit");
		}
		OdbcApi.SQLTransact (hEnv, hDbc, OdbcDef.SQL_COMMIT);
	}

	//--------------------------------------------------------------------
	// rollback
	//--------------------------------------------------------------------

	public void rollback ()
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*Connection.rollback");
		}
		OdbcApi.SQLTransact (hEnv, hDbc, OdbcDef.SQL_ROLLBACK);
	}

	//--------------------------------------------------------------------
	// close
	// close frees up various state associated with the connection.
	// This includes freeing the connection handle, and disconnecting.
	//--------------------------------------------------------------------

	public void close () throws SQLException
	{
		if (isTracing ()) {
			trace ("*Connection.close");
		}

		// Close any statement objects that we still have a reference to

		closeAllStatements ();

		// If we still have a connection handle, close it

		if (!closed) {
			myDriver.disconnect (hDbc);
			myDriver.closeConnection (hDbc);
		}

		closed = true;
		URL = null;
	}

	//--------------------------------------------------------------------
	// isClosed
	// isClosed returns true if the connection is closed, which can
	// occur either due to an explicit call on "close" or due to
	// some fatal error on the connection.
	//--------------------------------------------------------------------

	public boolean isClosed() throws SQLException {
		return closed;
	}


	// You can obtain a DatabaseMetaData object to get information 
	// about the target database.
	public DatabaseMetaData getMetaData()
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*Connection.getMetaData");
		}

		JdbcOdbcDatabaseMetaData metaData;

		// Validate that the connection is valid

		validateConnection ();

		// Create the new MetaData object, and return it
		
		metaData = new JdbcOdbcDatabaseMetaData (OdbcApi, this);

		return metaData;
	}

	//--------------------------------------------------------------------
	// setReadOnly
	// You can put a connection in readonly mode as a hint to enable 
	// database optimizations.  Note that setReadOnly cannot be called
	// while in the middle of a transaction.
	//--------------------------------------------------------------------

	public void setReadOnly (
		boolean readOnly)
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*Connection.setReadOnly (" + readOnly + ")");
		}

		int	mode = OdbcDef.SQL_MODE_READ_WRITE;

		// Validate that the connection is valid

		validateConnection ();

		if (readOnly) {
			mode = OdbcDef.SQL_MODE_READ_ONLY;
		}

		// Since this is just a hint for ODBC, we'll catch any exceptions

		try {
			OdbcApi.SQLSetConnectOption (hDbc, OdbcDef.SQL_ACCESS_MODE,
							mode);
		}
		catch (SQLException ex) {
			if (isTracing()) {
				trace ("setReadOnly exception ignored");
			}
		}
	}

	//--------------------------------------------------------------------
	// isReadOnly
	// Returns TRUE if the connection is in read-only mode
	//--------------------------------------------------------------------

	public boolean isReadOnly()
		throws SQLException 
	{
		if (isTracing ()) {
			trace ("*Connection.isReadOnly");
		}

		int	mode;
		boolean rc = false;

		// Validate that the connection is valid

		validateConnection ();

		mode = OdbcApi.SQLGetConnectOption (hDbc,
				OdbcDef.SQL_ACCESS_MODE);

		if (mode == OdbcDef.SQL_MODE_READ_ONLY) {
			rc = true;
		}
		return rc;
	}

	//--------------------------------------------------------------------
	// The "catalog" selects a database.
	//--------------------------------------------------------------------

	public void setCatalog(
		String catalog)
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*Connection.setCatalog (" + catalog + ")");
		}

		// Validate that the connection is valid

		validateConnection ();

		OdbcApi.SQLSetConnectOption (hDbc,
				OdbcDef.SQL_CURRENT_QUALIFIER, catalog);
	}

	//--------------------------------------------------------------------
	// getCatalog
	// Gets the current catalog for the connection
	//--------------------------------------------------------------------

	public String getCatalog()
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*Connection.getCatalog");
		}

		// Validate that the connection is valid

		validateConnection ();

		return OdbcApi.SQLGetInfoString (hDbc, 
				OdbcDef.SQL_DATABASE_NAME);
	}

	//--------------------------------------------------------------------
	// setTransactionIsolation
	// You can call the following method to try to change the transaction
	// isolation level on a newly opened connection, using one of the 
	// TRANSACTION_* values.  Use the DatabaseMetaData class to find what
   	// isolation levels are supported by the current database.
 	// Note that setTransactionIsolation cannot be called while in the
	// middle of a transaction.
	//--------------------------------------------------------------------

	public void setTransactionIsolation (
		int level)
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*Connection.setTransactionIsolation (" +
				level + ")");
		}

		// Validate that the connection is valid

		validateConnection ();

		switch (level) {
		case TRANSACTION_NONE:

			// No transactions, turn on auto-commit

			setAutoCommit (true);
			break;
		case TRANSACTION_READ_UNCOMMITTED:
			setAutoCommit (false);
			OdbcApi.SQLSetConnectOption (hDbc, 
				OdbcDef.SQL_TXN_ISOLATION,
				OdbcDef.SQL_TXN_READ_UNCOMMITTED);
			break;
		case TRANSACTION_READ_COMMITTED:
			setAutoCommit (false);
			OdbcApi.SQLSetConnectOption (hDbc, 
				OdbcDef.SQL_TXN_ISOLATION,
				OdbcDef.SQL_TXN_READ_COMMITTED);
			break;
		case TRANSACTION_REPEATABLE_READ:
			setAutoCommit (false);
			OdbcApi.SQLSetConnectOption (hDbc, 
				OdbcDef.SQL_TXN_ISOLATION,
				OdbcDef.SQL_TXN_REPEATABLE_READ);
			break;
		case TRANSACTION_SERIALIZABLE:
			setAutoCommit (false);
			OdbcApi.SQLSetConnectOption (hDbc, 
				OdbcDef.SQL_TXN_ISOLATION,
				OdbcDef.SQL_TXN_SERIALIZABLE);
			break;
		default:

			// Unknown option.  Call the driver with the option,
			// if it is invalid the driver will let us know

			setAutoCommit (false);
			OdbcApi.SQLSetConnectOption (hDbc,
				OdbcDef.SQL_TXN_ISOLATION, level); 
			break;
		}
	}

	//--------------------------------------------------------------------
	// getTransactionIsolation
	//--------------------------------------------------------------------

	public int getTransactionIsolation ()
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*Connection.getTransactionIsolation");
		}

		int isolation = TRANSACTION_NONE;
		int SQLIsolation;

		// Validate that the connection is valid

		validateConnection ();

		SQLIsolation = OdbcApi.SQLGetConnectOption (hDbc,
				OdbcDef.SQL_TXN_ISOLATION);

		// SQLGetConnectOption with SQL_TXN_ISOLATION actually
		// returns a bitmask, but we'll check for equality

		switch (SQLIsolation) {
		case OdbcDef.SQL_TXN_READ_UNCOMMITTED:
			isolation = TRANSACTION_READ_UNCOMMITTED;
			break;
		case OdbcDef.SQL_TXN_READ_COMMITTED:
			isolation = TRANSACTION_READ_COMMITTED;
			break;
		case OdbcDef.SQL_TXN_REPEATABLE_READ:
			isolation = TRANSACTION_REPEATABLE_READ;
			break;
		case OdbcDef.SQL_TXN_SERIALIZABLE:
			isolation = TRANSACTION_SERIALIZABLE;
			break;
		default:
			isolation = SQLIsolation;
		}
		return isolation;
	}

	//--------------------------------------------------------------------
	// getWarnings
	// Return any warning information related to the current connection.
	// Note that SQLWarning may be a chain.  Returns null if no warnings
	// exist.
	//--------------------------------------------------------------------

	public SQLWarning getWarnings()
			throws SQLException
	{
		if (isTracing ()) {
			trace ("*Connection.getWarnings");
		}
		return lastWarning;
	}

	//--------------------------------------------------------------------
	// clearWarnings
	// Clears any warning information for the connection
	//--------------------------------------------------------------------

	public void clearWarnings()
			throws SQLException
	{
		lastWarning = null;
	}

	//--------------------------------------------------------------------
	// validateConnection
	// Validates that this connection is not closed.  If the connection
	// is closed, throw an exception
	//--------------------------------------------------------------------

	public void validateConnection ()
			throws SQLException
	{
		if (isClosed ()) {
			throw new SQLException ("Connection is closed");
		}
	}

	//--------------------------------------------------------------------
	// getHDBC
	// Returns the connection handle
	//--------------------------------------------------------------------
	public int getHDBC ()
	{
		return hDbc;
	}

	//--------------------------------------------------------------------
	// setURL
	// Sets the URL for the connection
	//--------------------------------------------------------------------
	public void setURL (
		String url)
	{
		URL = url;
	}

	//--------------------------------------------------------------------
	// getURL
	// Returns the URL for the connection
	//--------------------------------------------------------------------
	public String getURL ()
	{
		return URL;
	}

	//--------------------------------------------------------------------
	// setLoginTimeout
	// Sets the login timeout for the connection
	//--------------------------------------------------------------------

	protected void setLoginTimeout(
		int seconds)
		throws SQLException
	{
		OdbcApi.SQLSetConnectOption (hDbc, OdbcDef.SQL_LOGIN_TIMEOUT,
				seconds);
	}


	//--------------------------------------------------------------------
	// getODBCVer
	// Queries the data source and gets the current ODBC version number.
	// The major version is returned as an int.  Returns -1 if unknown.
	//--------------------------------------------------------------------

	public int getODBCVer ()
	{
		if (odbcVer == 0) {

			String	s;

			try {
				s =  OdbcApi.SQLGetInfoString (hDbc, 
					OdbcDef.SQL_ODBC_VER);
			}
			catch (SQLException ex) {
				s = "-1";
			}

			// Convert first two characters of version (major
			// version) to an integer value.

			Integer i = new Integer (s.substring (0, 2));

			// Cache the value

			odbcVer = i.intValue ();
		}

		return odbcVer;
	}

	//--------------------------------------------------------------------
	// buildTypeInfo
	// Builds the typeInfo Hashtable.  This table will be keyed by
	// the SQL type, and contain information about each type
	//--------------------------------------------------------------------

	protected void buildTypeInfo ()
		throws SQLException
	{
		typeInfo = new Hashtable ();

		// Get a result set containing all of the type information

		if (isTracing ()) {
			trace ("Caching SQL type information");
		}

		ResultSet rs = (getMetaData ()).getTypeInfo ();

		// Information for a single SQL type

		JdbcOdbcTypeInfo info;
		int sqlType;

		// Loop on the result set until we reach end of file

		boolean more = rs.next ();
		String name;
		int prec;

		while (more) {
			name = rs.getString (1);
			sqlType = rs.getInt (2);

			// Now that we have the type info, save it
			// in the Hashtable if we don't already have an
			// entry for this SQL type.

			if (typeInfo.get (new Integer (sqlType)) == null) {
				info = new JdbcOdbcTypeInfo ();
				info.setName (name);
				info.setPrec (rs.getInt (3));

				typeInfo.put (new Integer (sqlType), info);
			}
			more = rs.next ();
		}

		// Close the result set/statement.

		rs.close ();
	}

	//--------------------------------------------------------------------
	// registerStatement
	// Keep a reference of the given Statement object
	//--------------------------------------------------------------------

	protected void registerStatement (
		Statement stmt)
	{
		if (isTracing ()) {
			trace ("Registering Statement " + stmt);
		}

		// We use a Hashtable to keep the list.  The Statement object
		// is the key.  There is no need to store a meaningful value.

		statements.put (stmt, "");
	}

	//--------------------------------------------------------------------
	// deregisterStatement
	// Each Statement object has a reference to it in the owning Connection
	// object.  This method will remove that reference
	//--------------------------------------------------------------------

	public void deregisterStatement (
		Statement stmt)
	{
		if (statements.get (stmt) != null) {
			if (isTracing ()) {
				trace ("deregistering Statement " + stmt);
			}
			
			statements.remove (stmt);
		}
	}

	//--------------------------------------------------------------------
	// closeAllStatements
	// Close any statements that are still referenced.
	//--------------------------------------------------------------------

	public synchronized void closeAllStatements ()
		throws SQLException
	{
		if (isTracing ()) {
			trace ("" + statements.size () + " Statement(s) to close");
		}

		if (statements.size () == 0) {
			return;
		}

		java.util.Enumeration enum = statements.keys ();
		Statement stmt;

		// Loop while there are still Statements references

		while (enum.hasMoreElements ()) {

			// Get the statement reference

			stmt = (Statement) enum.nextElement ();

			// Close the Statement.  This will also call our
			// deregisterStatement method to remove it from the statement
			// list

			stmt.close ();
		}
	}

	//====================================================================
	// Data attributes
	//====================================================================

	protected JdbcOdbc OdbcApi;		// ODBC API interface object

	protected JdbcOdbcDriverInterface myDriver;
						// Pointer to the owning
						//  driver object

	protected int	hEnv;			// Environment handle

	protected int	hDbc;			// Database connection handle

	protected SQLWarning	lastWarning;	// Last SQLWarning generated by
						//  an operation

	protected boolean	closed;		// Flag indicating whether the
						//  connection has been closed.

	protected String URL;			// URL of connection

	protected int odbcVer;			// ODBC version number (i.e. 2)

	protected Hashtable typeInfo;	// Hashtable containing an entry
						//  for each row returned by
						//  DatabaseMetaData.getTypeInfo.

	protected Hashtable statements;	// Hashtable containing a list
						//  of all the Statement objects
						//  for this Connection



    //--------------------------JDBC 2.0-----------------------------

    /**
     * JDBC 2.0
     *
     * Same as createStatement() above, but allows the default result set
     * type and result set concurrency type to be overridden.
     *
     * @param resultSetType a result set type, see ResultSet.TYPE_XXX
     * @param resultSetConcurrency a concurrency type, see ResultSet.CONCUR_XXX
     * @return a new Statement object 
     * @exception SQLException if a database-access error occurs.
     */
    public Statement createStatement(int resultSetType, int resultSetConcurrency) 
      { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Same as prepareStatement() above, but allows the default result set
     * type and result set concurrency type to be overridden.
     *
     * @param resultSetType a result set type, see ResultSet.TYPE_XXX
     * @param resultSetConcurrency a concurrency type, see ResultSet.CONCUR_XXX
     * @return a new PreparedStatement object containing the
     * pre-compiled SQL statement 
     * @exception SQLException if a database-access error occurs.
     */
     public PreparedStatement prepareStatement(String sql, int resultSetType, 
					int resultSetConcurrency)
       { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Same as prepareCall() above, but allows the default result set
     * type and result set concurrency type to be overridden.
     *
     * @param resultSetType a result set type, see ResultSet.TYPE_XXX
     * @param resultSetConcurrency a concurrency type, see ResultSet.CONCUR_XXX
     * @return a new CallableStatement object containing the
     * pre-compiled SQL statement 
     * @exception SQLException if a database-access error occurs.
     */
    public CallableStatement prepareCall(String sql, int resultSetType, 
				 int resultSetConcurrency) 
	    { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Get the type-map object associated with this connection.
     * By default, the map returned is empty.
     */
    public java.util.Map getTypeMap() 
		{ throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Install a type-map object as the default type-map for
     * this connection.
     */
    public void setTypeMap(java.util.Map map) 
		{ throw new UnsupportedOperationException(); }
}




