/*
 * @(#)JdbcOdbcResultSet.java	1.20 98/05/08
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
//----------------------------------------------------------------------------
//
// Module:      JdbcOdbcResultSet.java
//
// Description: Impementation of the ResultSet interface class
//
// Product:     JDBCODBC (Java DataBase Connectivity using
//              Open DataBase Connectivity)
//
// Author:      Karl Moss
//
// Date:        March, 1996
//
// Copyright:   1990-1996 INTERSOLV, Inc.
//              This software contains confidential and proprietary
//              information of INTERSOLV, Inc.
//----------------------------------------------------------------------------

package sun.jdbc.odbc;

import java.math.BigDecimal;
import java.util.Calendar;
import java.sql.*;

public class JdbcOdbcResultSet
	extends		JdbcOdbcObject
	implements	JdbcOdbcResultSetInterface
{

	//====================================================================
	// Public methods
	//====================================================================

	//--------------------------------------------------------------------
	// Constructor
	// Perform any necessary initialization.
	//--------------------------------------------------------------------

	public JdbcOdbcResultSet ()
	{
		OdbcApi = null;
		hDbc    = OdbcDef.SQL_NULL_HDBC;
		hStmt   = OdbcDef.SQL_NULL_HSTMT;
		lastWarning = null;
		keepHSTMT = false;
		numResultCols = -1;
		lastColumnNull = false;
	}

	//--------------------------------------------------------------------
	// finalize
	// Perform any cleanup when this object is garbage collected
	//--------------------------------------------------------------------

	protected void finalize ()
	{
		if (isTracing ()) {
			trace ("ResultSet.finalize " + this);
		}

		try {
			close ();
		}
		catch (SQLException ex) {
			// If an exception is thrown, ignore it
		}
	}

	//--------------------------------------------------------------------
	// initialize
	// Initialize the result set object.  Give the ODBC API interface
	// object, the connection handle, and the statement handle.  The
	// keep flag is used to determine if the statement handle should
	// actually be dropped when closing the ResultSet
	//--------------------------------------------------------------------

	public void initialize (
		JdbcOdbc odbcApi,
		int hdbc,
		int hstmt,
		boolean keep,
		Statement ownerStmt)
		throws SQLException
	{
		OdbcApi = odbcApi;
		hDbc = hdbc;
		hStmt = hstmt;
		keepHSTMT = keep;

		// Now allocate an array of bound column objects.  Each
		// column in the result set will have a placeholder to store
		// information as needed (i.e. InputStreams)

		numberOfCols = getColumnCount ();

		boundCols = new JdbcOdbcBoundCol[numberOfCols];
		int colType;

		for (int i = 0; i < numberOfCols; i++) {
			boundCols[i] = new JdbcOdbcBoundCol ();
		}

		// Keep a reference of our owning statement object

		ownerStatement = ownerStmt;
	}

	//--------------------------------------------------------------------
	// wasNull
	// A column may have the value of SQL NULL; wasNull reports whether
	// the last column read had this special value.

	public boolean wasNull ()
		throws SQLException

	{
		return lastColumnNull;
	}

	//--------------------------------------------------------------------
	// getString
	//--------------------------------------------------------------------

	public String getString (
		int column)
		throws SQLException
	{
		// Reset last warning message

		clearWarnings ();
		lastColumnNull = false;

		// Re-map column if necessary

		column = mapColumn (column);

		// If a pseudo column exists, return a null

		if (getPseudoCol (column) != null) {
			lastColumnNull = true;
			return null;
		}

		int maxLen = getMaxCharLen (column);

		// If the column is null, return null

		if (maxLen == OdbcDef.SQL_NULL_DATA) {
			return null;
		}

		// Some data types should not have spaces trimmed

                boolean trimSpaces = false;
		int colType = getColumnType (column);
		
               /* switch (colType) {
		case Types.CHAR:
		case Types.LONGVARCHAR:
			trimSpaces = false;
			break;
		default:
			break;
                } */

		// Allow room for a null terminator

		maxLen++;

		// Get the column, trimming spaces if necessary

		String value = getDataString (column, maxLen, trimSpaces);

		if (value == null) {
			return value;
		}

		// If we are not trimming spaces from the String, we need to return
		// the proper length (i.e. no null terminator)

		int valueLen = value.length ();

		if ((valueLen == (maxLen - 1)) && (!trimSpaces)) {
			value = value.substring (0, maxLen - 1);
		}

		// If we read in our entire buffer for LONG data, there may be
		// more to read.

		if (((colType == Types.LONGVARCHAR) ||
			(colType == Types.LONGVARBINARY)) &&
			(valueLen == (maxLen - 1))) {

			String value2 = value;

			// Loop while we have filled the read buffer

			while (value2.length () == JdbcOdbcLimits.MAX_GET_DATA_LENGTH) {
				value2 = getDataString (column, maxLen, trimSpaces);
				if (value2 != null) {
					if (isTracing ()) {
						trace ("" + value2.length () + " byte(s) read");
					}

					// Get rid of the null terminator
                                                if (value2.length () == maxLen) {
						value2 = value2.substring (0, maxLen - 1);
					}
					value += value2;

					if (isTracing ()) {
						trace ("" + value.length () + " bytes total");
					}
				}
				else {
					break;
				}
			}
		}

		return value;
	}

	public String getString (
		String columnName)
		throws SQLException
	{
		return getString (findColumn (columnName));
	}

	//--------------------------------------------------------------------
	// getBoolean
	//--------------------------------------------------------------------

	public boolean getBoolean (
		int column)
		throws SQLException
	{
		boolean returnValue = false;

		// Reset last warning message

		clearWarnings ();
		lastColumnNull = false;

		// Re-map column if necessary

		column = mapColumn (column);

		// If a pseudo column does not exist, get the data from the
		// column

		if (getPseudoCol (column) == null) {

			// Get the value as an int and convert

			returnValue = (getInt (column) != 0);
		}
		else {
			lastColumnNull = true;
		}
		return returnValue;
	}

	public boolean getBoolean (
		String columnName)
		throws SQLException
	{
		return getBoolean (findColumn (columnName));
	}

	//--------------------------------------------------------------------
	// getByte
	//--------------------------------------------------------------------

	public byte getByte (
		int column)
		throws SQLException
	{
		byte returnValue = 0;

		// Reset last warning message

		clearWarnings ();
		lastColumnNull = false;

		// Re-map column if necessary

		column = mapColumn (column);

		// If a pseudo column does not exist, get the data from the
		// column

		if (getPseudoCol (column) == null) {

			// Get the value as an int and convert

			returnValue = (byte) (getInt (column));
		}
		else {
			lastColumnNull = true;
		}
		return returnValue;
	}

	public byte getByte (
		String columnName)
		throws SQLException
	{
		return getByte (findColumn (columnName));
	}

	//--------------------------------------------------------------------
	// getShort
	//--------------------------------------------------------------------

	public short getShort (
		int column)
		throws SQLException
	{
		short returnValue = 0;

		// Reset last warning message

		clearWarnings ();
		lastColumnNull = false;

		// Re-map column if necessary

		column = mapColumn (column);

		// If a pseudo column does not exist, get the data from the
		// column

		if (getPseudoCol (column) == null) {

			// Get the value as an int and convert

			returnValue = (short) (getInt (column));
		}
		else {
			lastColumnNull = true;
		}

		return returnValue;
	}

	public short getShort (
		String columnName)
		throws SQLException
	{
		return getShort (findColumn (columnName));
	}

	//--------------------------------------------------------------------
	// getInt
	//--------------------------------------------------------------------

	public int getInt (
		int column)
		throws SQLException
	{
		int returnValue = 0;

		// Reset last warning message

		clearWarnings ();
		lastColumnNull = false;

		// Re-map column if necessary

		column = mapColumn (column);

		// If a pseudo column does not exist, get the data from the
		// column

		if (getPseudoCol (column) == null) {
			Integer value = getDataInteger (column);

			// If the column was not null, get the value from the
			// Integer object
			if (value != null) {
				returnValue = value.intValue ();
			}
		}
		else {
			lastColumnNull = true;
		}
		return returnValue;
	}

	public int getInt (
		String columnName)
		throws SQLException
	{
		return getInt (findColumn (columnName));
	}

	//--------------------------------------------------------------------
	// getLong
	//--------------------------------------------------------------------

	public long getLong (
		int column)
		throws SQLException
	{
		long returnValue = 0;

		// Reset last warning message

		clearWarnings ();
		lastColumnNull = false;

		// Re-map column if necessary

		column = mapColumn (column);

		// If a pseudo column does not exist, get the data from the
		// column

		if (getPseudoCol (column) == null) {
			Double value = getDataDouble (column);

			// If the column was not null, get the value from the
			// Double object
			if (value != null) {
				returnValue = value.longValue ();
			}
		}
		else {
			lastColumnNull = true;
		}
		return returnValue;
	}

	public long getLong (
		String columnName)
		throws SQLException
	{
		return getLong (findColumn (columnName));
	}

	//--------------------------------------------------------------------
	// getFloat
	//--------------------------------------------------------------------

	public float getFloat (
		int column)
		throws SQLException
	{
		float returnValue = 0;

		// Reset last warning message

		clearWarnings ();
		lastColumnNull = false;

		// Re-map column if necessary

		column = mapColumn (column);

		// If a pseudo column does not exist, get the data from the
		// column

		if (getPseudoCol (column) == null) {
			Float value = getDataFloat (column);

			// If the column was not null, get the value from the
			// Float object
			if (value != null) {
				returnValue = value.floatValue ();
			}
		}
		else {
			lastColumnNull = true;
		}

		return returnValue;
	}

	public float getFloat (
		String columnName)
		throws SQLException
	{
		return getFloat (findColumn (columnName));
	}

	//--------------------------------------------------------------------
	// getDouble
	//--------------------------------------------------------------------

	public double getDouble (
		int column)
		throws SQLException
	{
		double returnValue = 0;

		// Reset last warning message

		clearWarnings ();
		lastColumnNull = false;

		// Re-map column if necessary

		column = mapColumn (column);

		// If a pseudo column does not exist, get the data from the
		// column

		if (getPseudoCol (column) == null) {
			Double value = getDataDouble (column);

			// If the column was not null, get the value from the
			// Double object
			if (value != null) {
				returnValue = value.doubleValue ();
			}
		}
		else {
			lastColumnNull = true;
		}
		return returnValue;
	}

	public double getDouble (
		String columnName)
		throws SQLException
	{
		return getDouble (findColumn (columnName));
	}

	//--------------------------------------------------------------------
	// getBigDecimal
	//--------------------------------------------------------------------

	public BigDecimal getBigDecimal ( 
		int column,
		int scale)
		throws SQLException
	{
		BigDecimal n = null;

		// Reset last warning message

		clearWarnings ();
		lastColumnNull = false;

		// Re-map column if necessary

		column = mapColumn (column);

		// If a pseudo column does not exist, get the data from the
		// column

		if (getPseudoCol (column) == null) {

			// Get the column as a String, so we don't lose any precision

			String value = getDataString (column, 300, true);

			// If the column was not null, get the value from the
			// String object

			if (value != null) {
				n = new BigDecimal(value);
				n = n.setScale(scale, BigDecimal.ROUND_HALF_EVEN);
			    }
		}
		else {
			lastColumnNull = true;
		}

		return n;
	}

	public BigDecimal getBigDecimal ( 
		String columnName,
		int scale)
		throws SQLException
	{
		return getBigDecimal (findColumn (columnName), scale);
	}

	//--------------------------------------------------------------------
	// getBytes
	//--------------------------------------------------------------------

	public synchronized byte[] getBytes (
		int column)
		throws SQLException
	{
		// Reset last warning message

		clearWarnings ();
		lastColumnNull = false;

		// Re-map column if necessary

		column = mapColumn (column);

		// If a pseudo column exists, return a null

		if (getPseudoCol (column) != null) {
			lastColumnNull = true;
			return null;
		}

		int maxLen = getMaxBinaryLen (column);

		// If the column is null, return null

		if (maxLen == OdbcDef.SQL_NULL_DATA) {
			return null;
		}

		// Get the column type

		int colType = getColumnType (column);

		// Create new byte array to read data into

		byte b[] = new byte[maxLen];

		// Read the binary data.  The number of bytes read will be
		// returned, -1 for NULL

		int numBytes;

		try {
			numBytes = OdbcApi.SQLGetDataBinary (hStmt, column, b);
		}
		catch (JdbcOdbcSQLWarning warn) {

			// Data truncation.  Ignore the error and get the length

			Integer value = (Integer) warn.value;
			numBytes = value.intValue ();
		}

		// If we attempted to read a null column, return a null

		if (numBytes == OdbcDef.SQL_NULL_DATA) {
			lastColumnNull = true;
			b = null;
		}

		// If we read less than the maximum number of bytes, create a
		// new byte array of the proper size.  Only do this if the column
		// is not a BINARY type, which is fixed length.

		if ((colType != Types.BINARY) &&
			(numBytes != maxLen)) {

			byte b2[] = new byte[numBytes];
			System.arraycopy (b, 0, b2, 0, numBytes);
			return b2;
		}
		else if (numBytes < JdbcOdbcLimits.MAX_GET_DATA_LENGTH) {
			
			// We read all of the requested buffer, and we knew
			// the actual size.  Return the buffer.

			return b;
		}

		// Now we have the possibility of having more data present.
		// Let's go ahead an attempt to read in all of the data.  We'll do
		// this by reading in a 'chunk' of data at a time, saving each 
		// chunk, then creating the final buffer once we know how
		// much data is actually present.  Note that we have already
		// read our first chunk.  We'll use a Hashtable to save
		// each chunk.

		java.util.Hashtable chunks = new java.util.Hashtable ();
		int chunkNo = 0;
		int total = 0;

		// Loop until there is no more data.  For very large columns this
		// can very easily result in an 'out of memory' error.  In the
		// case, the application should be using an InputStream to read
		// the column.

		while (numBytes != -1) {
			total += numBytes;
			chunks.put (new Integer (chunkNo++), b);

			try {
				numBytes = OdbcApi.SQLGetDataBinary (hStmt, column, b);
			}
			catch (JdbcOdbcSQLWarning warn) {

				// Data truncation.  Ignore the error and get the length

				Integer value = (Integer) warn.value;
				numBytes = value.intValue ();
			}
		}

		// Now we have read all of the data.  Create a new buffer 
		// to hold the entire column.  Again, this could cause an
		// 'out of memory' error for large data.

		if (isTracing ()) {
			trace ("Total bytes read: " + total + ".  Creating buffer");
		}

		byte b2[] = new byte[total];
		int b2Offset = 0;

		// Copy the data over

		for (int i = 0; i < chunkNo; i++) {
			b = (byte[]) chunks.get (new Integer (i));

			// Figure out how many bytes to copy

			numBytes = total;
			if (numBytes > JdbcOdbcLimits.MAX_GET_DATA_LENGTH) {
				numBytes = JdbcOdbcLimits.MAX_GET_DATA_LENGTH;
			}
			total -= numBytes;

			// Copy the data to our new buffer

			System.arraycopy (b, 0, b2, b2Offset, numBytes);
			b2Offset += numBytes;
		}
		return b2;
	}

	public byte[] getBytes (
		String columnName)
		throws SQLException
	{
		return getBytes (findColumn (columnName));
	}

	//--------------------------------------------------------------------
	// getDate
	// Get the date as a string, and convert to a Date object
	//--------------------------------------------------------------------

	public java.sql.Date getDate (
		int column)
		throws SQLException
	{
		// Reset last warning message

		clearWarnings ();
		lastColumnNull = false;

		// Re-map column if necessary

		column = mapColumn (column);

		// If a pseudo column exists, return a null

		if (getPseudoCol (column) != null) {
			lastColumnNull = true;
			return null;
		}

		String s = getDataStringDate (column);

		// If the column is null, return a null object

		if (s == null) {
			return null;
		}

		// Now convert to a Date object

		return Date.valueOf (s);
	}

	public java.sql.Date getDate (
		String columnName)
		throws SQLException
	{
		return getDate (findColumn (columnName));
	}

	//--------------------------------------------------------------------
	// getTime
	// Get the time as a string, and convert to a Time object
	//--------------------------------------------------------------------

	public java.sql.Time getTime (
		int column)
		throws SQLException
	{
		// Reset last warning message

		clearWarnings ();
		lastColumnNull = false;

		// Re-map column if necessary

		column = mapColumn (column);

		// If a pseudo column exists, return a null

		if (getPseudoCol (column) != null) {
			lastColumnNull = true;
			return null;
		}

		String s = getDataStringTime (column);

		// If the column is null, return a null object

		if (s == null) {
			return null;
		}

		// Now convert to a Time object

		return Time.valueOf (s);
	}

	public java.sql.Time getTime (
		String columnName)
		throws SQLException
	{
		return getTime (findColumn (columnName));
	}

	//--------------------------------------------------------------------
	// getTimestamp
	// Get the timestamp as a string, and convert to a Timestamp object
	//--------------------------------------------------------------------

	public java.sql.Timestamp getTimestamp (
		int column)
		throws SQLException
	{
		// Reset last warning message

		clearWarnings ();
		lastColumnNull = false;

		// Re-map column if necessary

		column = mapColumn (column);

		// If a pseudo column exists, return a null

		if (getPseudoCol (column) != null) {
			lastColumnNull = true;
			return null;
		}

		String s = getDataStringTimestamp (column);

		// If the column is null, return a null object

		if (s == null) {
			return null;
		}

		// If all we got was a date, put on default seconds

		if (s.length () == 10) {
			s += " 00:00:00";
		}

		// Now convert to a Timestamp object

		return Timestamp.valueOf (s);
	}

	public java.sql.Timestamp getTimestamp (
		String columnName)
		throws SQLException
	{
		return getTimestamp (findColumn (columnName));
	}

	//--------------------------------------------------------------------
	// The normal getString and getBytes methods are suitable for
	// reading normal sized data.  However occasionally it may be 
	// necessary to access LONGVARCHAR or LONGVARBINARY fields that
	// are multiple Megabytes in size.  To support this case we provide
	// getCharStream and getBinaryStream methods that return Java
	// IO streams from which data can be read in chunks.
	//--------------------------------------------------------------------

	public java.io.InputStream getAsciiStream (
		int column)
		throws SQLException
	{
		// Reset last warning message

		clearWarnings ();
		lastColumnNull = false;

		// Re-map column if necessary

		column = mapColumn (column);

		// We need to know the source column data type

		int colType = getColumnType (column);
		int dataType = Types.BINARY;
		
		switch (colType) {
		case Types.CHAR:
		case Types.VARCHAR:
		case Types.LONGVARCHAR:
			dataType = Types.CHAR;
			break;
		}

		JdbcOdbcInputStream inStream =  new JdbcOdbcInputStream (
			OdbcApi, hStmt, column,	JdbcOdbcInputStream.ASCII, dataType,
			ownerStatement);

		setInputStream (column, inStream);

		return inStream;
	}

	public java.io.InputStream getAsciiStream (
		String columnName)
		throws SQLException
	{
		return getAsciiStream (findColumn (columnName));
	}

	public java.io.InputStream getUnicodeStream (
		int column)
		throws SQLException
	{
		// Reset last warning message

		clearWarnings ();
		lastColumnNull = false;

		// Re-map column if necessary

		column = mapColumn (column);

		// We need to know the source column data type

		int colType = getColumnType (column);
		int dataType = Types.BINARY;
		
		switch (colType) {
		case Types.CHAR:
		case Types.VARCHAR:
		case Types.LONGVARCHAR:
			dataType = Types.CHAR;
			break;
		}

		JdbcOdbcInputStream inStream =  new JdbcOdbcInputStream (
			OdbcApi, hStmt, column,	JdbcOdbcInputStream.UNICODE, dataType,
			ownerStatement);
		setInputStream (column, inStream);

		return inStream;
	}

	public java.io.InputStream getUnicodeStream (
		String columnName)
		throws SQLException
	{
		return getUnicodeStream (findColumn (columnName));
	}

	public java.io.InputStream getBinaryStream (
		int column)
		throws SQLException
	{
		// Reset last warning message

		clearWarnings ();
		lastColumnNull = false;

		// Re-map column if necessary

		column = mapColumn (column);

		// We need to know the source column data type

		int colType = getColumnType (column);
		int dataType = Types.BINARY;
		
		switch (colType) {
		case Types.CHAR:
		case Types.VARCHAR:
		case Types.LONGVARCHAR:
			dataType = Types.CHAR;
			break;
		}

		JdbcOdbcInputStream inStream =  new JdbcOdbcInputStream (
			OdbcApi, hStmt, column,	JdbcOdbcInputStream.BINARY, dataType,
			ownerStatement);
		setInputStream (column, inStream);

		return inStream;
	}

	public java.io.InputStream getBinaryStream (
		String columnName)
		throws SQLException
	{
		return getBinaryStream (findColumn (columnName));
	}

	//--------------------------------------------------------------------
	// next
	// next moves us to the next row of the results.
	// It returns true if if has moved to a valid row, false if
	// all the rows have been processed.
	//--------------------------------------------------------------------

	public boolean next ()
		throws SQLException
	{
		boolean rc = true;
		lastColumnNull = false;

		// Close any/all input streams for the statement
		
		closeInputStreams ();

		// Reset last warning message

		clearWarnings ();

		try {
			rc = OdbcApi.SQLFetch (hStmt);
		}
		catch (SQLWarning ex) {

			// If we got a warning, set the last warning for
			// the result set.  We will assume that the fetch
			// positioned us to a vaild row.

			setWarning (ex);
		}
		return rc;
	}

	//--------------------------------------------------------------------
	// getRowNumber
	// Return the number of the current row.  The first row containing
	// actual data is row 1.
	//--------------------------------------------------------------------

	public int getRowNumber ()
		throws SQLException
	{
		int	rowNum = 0;

		// Reset last warning message

		clearWarnings ();

		try {
			rowNum = OdbcApi.SQLGetStmtOption (hStmt,
					OdbcDef.SQL_ROW_NUMBER);
		}
		catch (JdbcOdbcSQLWarning ex) {

			// If we got a warning, the return value was saved off
			// in the warning object.  Get it back and set the
			// return value.

		        BigDecimal n = (BigDecimal) ex.value;
			rowNum = n.intValue ();

			setWarning (JdbcOdbc.convertWarning (ex));
		}
		return rowNum;
	}


	//--------------------------------------------------------------------
	// getColumnCount
	// Return the number of columns in the ResultSet
	//--------------------------------------------------------------------

	public int getColumnCount ()
		throws SQLException
	{
		int	numCols = 0;

		// Reset last warning message

		clearWarnings ();

		// If pseudo columns exist, return the last pseudo
		// column

		if (lastPseudoCol > 0) {
			return lastPseudoCol;
		}

		// If column mappings exist, return the number of columns
		// that are being mapped

		if (colMappings != null) {
			return colMappings.length;
		}

		try {
			// If we haven't queried for the number of result
			// columns in the result set yet, do it now and
			// cache the result

			if (numResultCols == -1) {
				numResultCols = OdbcApi.SQLNumResultCols (
						hStmt);
			}
			numCols = numResultCols;
		}
		catch (JdbcOdbcSQLWarning ex) {

			// If we got a warning, the return value was saved off
			// in the warning object.  Get it back and set the
			// return value.

			BigDecimal n = (BigDecimal) ex.value;
			numCols = n.intValue ();

			setWarning (JdbcOdbc.convertWarning (ex));
		}
		return numCols;
	}

	//--------------------------------------------------------------------
	// getRowCount
	// The total number of rows returned by the query.  Note that on some
	// databases this method may be very expensive.
	//--------------------------------------------------------------------

	public int getRowCount ()
		throws SQLException
	{
		int	numRows = 0;

		// Reset last warning message

		clearWarnings ();

		try {
			numRows = OdbcApi.SQLRowCount (hStmt);
		}
		catch (JdbcOdbcSQLWarning ex) {

			// If we got a warning, the return value was saved off
			// in the warning object.  Get it back and set the
			// return value.

			BigDecimal n = (BigDecimal) ex.value;
			numRows = n.intValue ();

			setWarning (JdbcOdbc.convertWarning (ex));
		}
		return numRows;
	}

	//--------------------------------------------------------------------
	// close
	// close frees up internal state associated with the ResultSet
	// You should normally call close when you are done with a ResultSet
	// ResultSets are also closed automatically when their Statement is
	// closed.
	//--------------------------------------------------------------------

	public synchronized void close()
		throws SQLException
	{
		// Close any/all input streams for the statement
		
		closeInputStreams ();

		// Reset last warning message

		clearWarnings ();
		lastColumnNull = false;

		if ((OdbcApi != null) &&
		    (hStmt != OdbcDef.SQL_NULL_HSTMT)) {

			// If we are keeping the hStmt, simply close it

			if (keepHSTMT) {
				if (!freed) {
					OdbcApi.SQLFreeStmt (hStmt, OdbcDef.SQL_CLOSE);
					freed = true;
				}
			}
			else {
				// Not keeping the hStmt, drop it

				OdbcApi.SQLFreeStmt (hStmt, OdbcDef.SQL_DROP);
				hStmt = OdbcDef.SQL_NULL_HSTMT;
			}

			// Turn on flag indicating that the result set is now closed

			closed = true;
		}
	}

	//--------------------------------------------------------------------
	// getCursorName
	// Returns a cursor name that can be used to identify the
	// currnet position in the Resultset to a separate statement that is
	// executing a positioned update or positioned delete.  If the database
	// doesn't support positioned delete or update, it may return "" here.
	//--------------------------------------------------------------------

	public String getCursorName ()
		throws SQLException
	{
		String	value = "";

		// Reset last warning message

		clearWarnings ();

		try {
			value = OdbcApi.SQLGetCursorName (hStmt);
		}
		catch (JdbcOdbcSQLWarning ex) {

			// If we got a warning, the return value was saved off
			// in the warning object.  Get it back and set the
			// return value.

			value = (String) ex.value;
			setWarning (JdbcOdbc.convertWarning (ex));
		}
		return value.trim ();
	}

	//--------------------------------------------------------------------
	// getMetaData
	// You can obtain a ResultMetaData object to obtain information
	// about the number, types and proerties of the result columns.
	//--------------------------------------------------------------------

	public ResultSetMetaData getMetaData ()
		throws SQLException
	{
		if (isTracing()) {
			trace ("*ResultSet.getMetaData");
		}

		// If the result set has been closed, throw an exception

		if (closed) {
			throw new SQLException("ResultSet is closed");
		}

		return new JdbcOdbcResultSetMetaData (OdbcApi, this);
	}

	//--------------------------------------------------------------------
	// getObject
	// You can obtain a column result as a Java Object.
	// See the JDBC spec's "Dynamic Programming" chapter for details.
	// All of the getData methods above use the getObject interface.
	//--------------------------------------------------------------------

	public Object getObject (
		int column)
		throws SQLException
	{
		Object	value = null;

		// Get the SQL type of the column

		int sqlType = getColumnType (column);

		// Save the original column number in case it is remapped

		int origColumn = column;
		
		// Reset last warning message

		clearWarnings ();
		lastColumnNull = false;

		// Re-map column if necessary

		column = mapColumn (column);

		// If a pseudo column exists, return a null

		if (getPseudoCol (column) != null) {
			lastColumnNull = true;
			return null;
		}

		// For each SQL type, call the appropriate routine and
		// convert to the proper object type

		switch (sqlType) {

		case Types.CHAR:
		case Types.VARCHAR:
		case Types.LONGVARCHAR:
			value = (Object) getString (column);
			break;

		case Types.NUMERIC:
		case Types.DECIMAL:
			value = (Object) getBigDecimal (column, getScale (origColumn));
			break;

		case Types.BIT:
			value = (Object) new Boolean (getBoolean (column));
			break;

		case Types.TINYINT:
		case Types.SMALLINT:
		case Types.INTEGER:
			value = (Object) new Integer (getInt (column));
			break;

		case Types.BIGINT:
			value = (Object) new Long (getLong (column));
			break;

		case Types.REAL:
			value = (Object) new Float (getFloat (column));
			break;

		case Types.FLOAT:
		case Types.DOUBLE:
			value = (Object) new Double (getDouble (column));
			break;
			
		case Types.BINARY:
		case Types.VARBINARY:
		case Types.LONGVARBINARY:
			value = (Object) getBytes (column);
			break;

		case Types.DATE:
			value = (Object) getDate (column);
			break;

		case Types.TIME:
			value = (Object) getTime (column);
			break;

		case Types.TIMESTAMP:
			value = (Object) getTimestamp (column);
			break;

		}

		// If the column was null, always return a null object

		if (wasNull ()) {
			value = null;
		}


		return value;
	}

	public Object getObject (
		String columnName)
		throws SQLException
	{
		return getObject (findColumn (columnName));
	}

	//--------------------------------------------------------------------
	// getWarnings
	// Returns the first warning reported by calls on this ResultSet
	//--------------------------------------------------------------------
	public SQLWarning getWarnings ()
		throws SQLException
	{
		return lastWarning;
	}

	//--------------------------------------------------------------------
	// clearWarnings
	// Clears any warning information for the statement
	//--------------------------------------------------------------------

	public void clearWarnings()
			throws SQLException
	{
		lastWarning = null;
	}

	//--------------------------------------------------------------------
	// setWarning
	// Sets the warning 
	//--------------------------------------------------------------------

	public void setWarning (
			SQLWarning ex)
			throws SQLException
	{
		lastWarning = ex;
	}

	//--------------------------------------------------------------------
	// getHSTMT
	// Returns the statement handle for the result set
	//--------------------------------------------------------------------
	
	public int getHSTMT ()
	{
		return hStmt;
	}

	//--------------------------------------------------------------------
	// findColumn
	// Map a ResultSet column name to a ResultSet column index
	//--------------------------------------------------------------------
	public synchronized int findColumn (
		String columnName)
		throws SQLException
	{
		// Get the MetaData object if we don't already have one

		if (rsmd == null) {
			rsmd = getMetaData ();
			colNameToNum = new java.util.Hashtable ();
			colNumToName = new java.util.Hashtable ();
		}

		// Search for the mapping in the hash table
		
		Integer col = (Integer) colNameToNum.get (columnName);

		// If it was not found, try to get the name from the
		// MetaData.  If it is not found here, raise an exception

		if (col == null) {

			String name;

			for (int i = 1; i <= rsmd.getColumnCount(); i++) {

				// As an optimization, keep a separate hash
				// table mapping column numbers to names.  If
				// we don't find the column number in the hash
				// table, then we have to hit the data source
				// to get the name.  Otherwise, get the name
				// from the hash table

				name = (String) colNumToName.get (
						new Integer (i));

				if (name == null) {
					name = rsmd.getColumnName (i);
					// Add it to our hash tables
					colNameToNum.put (name,
						new Integer (i));
					colNumToName.put (new Integer (i),
						name);
				}

				// If the name matches, return the column
				// pointer.

				if (name.equalsIgnoreCase (columnName)) {
					return i;
				}
			}

			// The column name was not found

			throw new SQLException ("Column not found",
				"S0022");
		}
		return col.intValue ();
	}

	//====================================================================
	// Non-API methods
	//====================================================================

	//--------------------------------------------------------------------
	// getColAttribute
	// Given the column and attribute type, return the attribute value
	//--------------------------------------------------------------------

	public int getColAttribute (
		int column,
		int type)
		throws SQLException
	{
		int	value = 0;

		// Reset last warning message

		clearWarnings ();

		try {
			value = OdbcApi.SQLColAttributes (hStmt,
					column, type);
		}
		catch (JdbcOdbcSQLWarning ex) {

			// If we got a warning, the return value was saved off
			// in the warning object.  Get it back and set the
			// return value.

			BigDecimal n = (BigDecimal) ex.value;
			value = n.intValue ();

			setWarning (JdbcOdbc.convertWarning (ex));
		}
		return value;
	}

	//--------------------------------------------------------------------
	// getMaxCharLen
	// Returns the length of the column as needed for a character
	// representation
	//--------------------------------------------------------------------
	protected int getMaxCharLen (
			int column)
			throws SQLException
	{
		int colType = getColumnType (column);
		int maxLen = getColumnLength (column);

		if (maxLen != OdbcDef.SQL_NULL_DATA) {

			// Adjust for certain data types

			switch (colType) {
			case Types.BINARY:
			case Types.VARBINARY:
			case Types.LONGVARBINARY:

				// Adjust binary types being displayed
				// as character.  Each binary byte takes
				// two characters to display

				maxLen *= 2;
				break;
			case Types.DATE:
				maxLen = 10;
				break;
			case Types.TIME:
				maxLen = 8;
				break;
			case Types.TIMESTAMP:
				maxLen = 29;
				break;
			case Types.NUMERIC:
			case Types.DECIMAL:
				maxLen += 2;
				break;
			case Types.BIT:
				maxLen = 1;
				break;
			case Types.TINYINT:
				maxLen = 4;
				break;
			case Types.SMALLINT:
				maxLen = 6;
				break;
			case Types.INTEGER:
				maxLen = 11;
				break;
			case Types.BIGINT:
				maxLen = 20;
				break;
			case Types.REAL:
				maxLen = 13;
				break;
			case Types.FLOAT:
			case Types.DOUBLE:
				maxLen = 22;
				break;
			}


			// Sanity check the length.  If invalid, set to max

			if ((maxLen <= 0) ||
			(maxLen > JdbcOdbcLimits.MAX_GET_DATA_LENGTH)) {
				maxLen = JdbcOdbcLimits.MAX_GET_DATA_LENGTH;
			}
		}
		return maxLen;
	}

	//--------------------------------------------------------------------
	// getMaxBinaryLen
	// Returns the length of the column as needed for a binary
	// representation
	//--------------------------------------------------------------------
	protected int getMaxBinaryLen (
			int column)
			throws SQLException
	{
		int maxLen = getColumnLength (column);

		if (maxLen != OdbcDef.SQL_NULL_DATA) {

			// Sanity check the length.  If invalid, set to max

			if ((maxLen <= 0) ||
			(maxLen > JdbcOdbcLimits.MAX_GET_DATA_LENGTH)) {
				maxLen = JdbcOdbcLimits.MAX_GET_DATA_LENGTH;
			}
		}
		return maxLen;
	}

	//--------------------------------------------------------------------
	// getDataDouble
	// Returns the data for the given column as a Double object.  Returns
	// null if the value for the column was NULL.
	//--------------------------------------------------------------------

	public Double getDataDouble (
		int column)
		throws SQLException
	{
		Double n;
		lastColumnNull = false;
		try {
			n = OdbcApi.SQLGetDataDouble (hStmt, column);
		}
		catch (JdbcOdbcSQLWarning ex) {

			// If we got a warning, the return value was 
			// saved off in the warning object.  Get it
			// back and set the return value.

			n = (Double) ex.value;

			setWarning (JdbcOdbc.convertWarning (ex));
		}
		if (n == null) {
			lastColumnNull = true;
		}
		return n;
	}

	//--------------------------------------------------------------------
	// getDataFloat
	// Returns the data for the given column as a Float object.  Returns
	// null if the value for the column was NULL.
	//--------------------------------------------------------------------

	public Float getDataFloat (
		int column)
		throws SQLException
	{
		Float n;
		lastColumnNull = false;
		try {
			n = OdbcApi.SQLGetDataFloat (hStmt, column);
		}
		catch (JdbcOdbcSQLWarning ex) {

			// If we got a warning, the return value was 
			// saved off in the warning object.  Get it
			// back and set the return value.

			n = (Float) ex.value;

			setWarning (JdbcOdbc.convertWarning (ex));
		}
		if (n == null) {
			lastColumnNull = true;
		}
		return n;
	}

	//--------------------------------------------------------------------
	// getDataInteger
	// Returns the data for the given column as an Integer object.  Returns
	// null if the value for the column was NULL.
	//--------------------------------------------------------------------

	public Integer getDataInteger (
		int column)
		throws SQLException
	{
		Integer n;
		lastColumnNull = false;
		try {
			n = OdbcApi.SQLGetDataInteger (hStmt, column);
		}
		catch (JdbcOdbcSQLWarning ex) {

			// If we got a warning, the return value was 
			// saved off in the warning object.  Get it
			// back and set the return value.

			n = (Integer) ex.value;

			setWarning (JdbcOdbc.convertWarning (ex));
		}
		if (n == null) {
			lastColumnNull = true;
		}
		else if (column == sqlTypeColumn) {
		
			// If we are fetching data that is a catalog SQL type column,
			// we need to convert it to a JDBC SQL type.

			n = new Integer (OdbcDef.odbcTypeToJdbc (n.intValue()));
		}
		return n;
	}

	//--------------------------------------------------------------------
	// getDataLong
	// Returns the data for the given column as a Long object.  Returns
	// null if the value for the column was NULL.
	//--------------------------------------------------------------------

	public Long getDataLong (
		int column)
		throws SQLException
	{
		Long n = null;
		Double d = getDataDouble (column);
		
		if (d != null) {
			n = new Long (d.longValue ());
		}
		return n;
	}


	//--------------------------------------------------------------------
	// getDataString
	// Returns the data for the given column as an String object.  Returns
	// null if the value for the column was NULL.
	//--------------------------------------------------------------------

	public String getDataString (
		int column,
		int maxLen,
		boolean trimSpaces)
                throws SQLException
	{
		String s;
		lastColumnNull = false;
		try {
			s = OdbcApi.SQLGetDataString (hStmt, column, maxLen,
						trimSpaces);
		}
		catch (JdbcOdbcSQLWarning ex) {

			// If we got a warning, the return value was 
			// saved off in the warning object.  Get it
			// back and set the return value.

			s = (String) ex.value;
			setWarning (JdbcOdbc.convertWarning (ex));
		}
		if (s == null) {
			lastColumnNull = true;
		}
		else if (column == sqlTypeColumn) {
		
			// If we are fetching data that is a catalog SQL type column,
			// we need to convert it to a JDBC SQL type.  If an exception
			// is thrown during the conversion between a String a int,
			// ignore it and return the original string
			
			try {
				int colType = OdbcDef.odbcTypeToJdbc (
							(Integer.valueOf(s)).intValue());
				s = "" + colType;
			}
			catch (Exception ex) {
			}
		}
		return s;
	}

	//--------------------------------------------------------------------
	// getDataStringDate
	// Returns the data for the given column as an String Date object
	// (yyyy-mm-dd).  Returns null if the value for the column was NULL.
	//--------------------------------------------------------------------

	public String getDataStringDate (
		int column)
		throws SQLException
	{
		String s;
		lastColumnNull = false;
		try {
			s = OdbcApi.SQLGetDataStringDate (hStmt, column);
		}
		catch (JdbcOdbcSQLWarning ex) {

			// If we got a warning, the return value was 
			// saved off in the warning object.  Get it
			// back and set the return value.

			s = (String) ex.value;
			setWarning (JdbcOdbc.convertWarning (ex));
		}
		if (s == null) {
			lastColumnNull = true;
		}
		return s;
	}

	//--------------------------------------------------------------------
	// getDataStringTime
	// Returns the data for the given column as an String Time object
	// (hh:mm:ss).  Returns null if the value for the column was NULL.
	//--------------------------------------------------------------------

	public String getDataStringTime (
		int column)
		throws SQLException
	{
		String s;
		lastColumnNull = false;
		try {
			s = OdbcApi.SQLGetDataStringTime (hStmt, column);
		}
		catch (JdbcOdbcSQLWarning ex) {

			// If we got a warning, the return value was 
			// saved off in the warning object.  Get it
			// back and set the return value.

			s = (String) ex.value;
			setWarning (JdbcOdbc.convertWarning (ex));
		}
		if (s == null) {
			lastColumnNull = true;
		}
		return s;
	}

	//--------------------------------------------------------------------
	// getDataStringTimestamp
	// Returns the data for the given column as an String Timestamp object
	// (yyyy-mm-dd hh:mm:ss).  Returns null if the value for the column
	// was NULL.
	//--------------------------------------------------------------------

	public String getDataStringTimestamp (
		int column)
		throws SQLException
	{
		String s;
		lastColumnNull = false;
		try {
			s = OdbcApi.SQLGetDataStringTimestamp (hStmt, column);
		}
		catch (JdbcOdbcSQLWarning ex) {

			// If we got a warning, the return value was 
			// saved off in the warning object.  Get it
			// back and set the return value.

			s = (String) ex.value;
			setWarning (JdbcOdbc.convertWarning (ex));
		}
		if (s == null) {
			lastColumnNull = true;
		}
		return s;
	}

	//--------------------------------------------------------------------
	// getColumnLength
	// Returns the length of the given column number.
	//--------------------------------------------------------------------

	public int getColumnLength (
		int column)
		throws SQLException
	{
		int colLen = -1;

		// As an optimization, we'll cache the length into
		// the column array.

		if ((column > 0) &&
		    (column <= numberOfCols)) {
		    
			// Get the column length from the column array

			colLen = boundCols[column - 1].getLength ();

		}

		// If we don't know the length, ask the data source

		if (colLen == -1) {
			colLen = getColAttribute (column, 
				OdbcDef.SQL_COLUMN_LENGTH);
			
			// Cache the length into the column array

			if ((column > 0) &&
				(column <= numberOfCols)) {
				boundCols[column - 1].setLength (colLen);
			}
		}

		return colLen;
	}

	//--------------------------------------------------------------------
	// getScale
	// Number of digits to right of decimal
	//--------------------------------------------------------------------

	public int getScale (
		int column)
		throws SQLException
	{
		int value;

		// If a pseudo column exists for this column, return 0

		if (getPseudoCol (column) != null) {
			lastColumnNull = true;
			value = 0;
		}
		else {
			value = getColAttribute (column,
				OdbcDef.SQL_COLUMN_SCALE);
		}       	
		return value;

	}

	//--------------------------------------------------------------------
	// getColumnType
	// Returns the Java SQL type of the given column number.
	//--------------------------------------------------------------------

	public int getColumnType (
		int column)
		throws SQLException
	{
		int colType = OdbcDef.SQL_TYPE_UNKNOWN;

		// As an optimization, we'll cache the type information into
		// the column array.

		if ((column > 0) &&
		    (column <= numberOfCols)) {
		    
			// Get the column type from the column array

			colType = boundCols[column - 1].getType ();

		}

		// If we don't know the type, ask the data source

		if (colType == OdbcDef.SQL_TYPE_UNKNOWN) { 
			colType = getColAttribute (column, 
				OdbcDef.SQL_COLUMN_TYPE);

			// Convert it to the proper JDBC SQL type

			colType = OdbcDef.odbcTypeToJdbc (colType);

			// Cache the type into the column array
			if ((column > 0) &&
				(column <= numberOfCols)) {
				boundCols[column - 1].setType (colType);
			}
		}

		return colType;
	}

	//--------------------------------------------------------------------
	// setPseudoCols
	// Set the pseudo column information
	//--------------------------------------------------------------------

	public void setPseudoCols (
		int first,
		int last,
		JdbcOdbcPseudoCol pc[])
	{
		if (isTracing ()) {
			trace ("Setting pseudo columns, first=" +
				first + ", last=" + last);
		}
		firstPseudoCol = first;
		lastPseudoCol = last;
		pseudoCols = pc;
	}

	//--------------------------------------------------------------------
	// getPseudoCol
	// Given a column number, return the corresponding pseudo column,
	// null if one does not exist for this column
	//--------------------------------------------------------------------

	public JdbcOdbcPseudoCol getPseudoCol (
		int column)
	{
		JdbcOdbcPseudoCol pc = null;

		// If a pseudo column exists for this column, return
		// it from the array.  It is assumed that the pseudo
		// column array contains an element for each column between
		// (and including) the first and last pseudo column
		// specified.

		if ((column > 0) &&
		    (column >= firstPseudoCol) &&
		    (column <= lastPseudoCol)) {
			pc = pseudoCols[column - firstPseudoCol];
		}
		return pc;
	}

	//--------------------------------------------------------------------
	// setSQLTypeColumn
	// Some catalogs return SQL data types in ODBC format.  We need to
	// flag them so that they can be converted into Java SQL types.
	//--------------------------------------------------------------------

	public void setSQLTypeColumn (
		int column)
	{
		sqlTypeColumn = column;
	}
		
	//--------------------------------------------------------------------
	// setInputStream
	// Saves a reference to the newly allocated InputStream.  For each
	// InputStream that is created for the statement, a reference is
	// kept.  If the Statement is closed or the cursor is moved, the
	// InputStream is invalidated.  Once invalidated, if an application
	// attempts to read from the InputStream, an exception is raised.
	//--------------------------------------------------------------------

	protected void setInputStream (
		int column,
		JdbcOdbcInputStream inStream)
	{
		// Sanity check the column number 

		if ((column > 0) &&
		    (column <= numberOfCols)) {

			// Save the input stream in the bound column array

			boundCols[column - 1].setInputStream (inStream);
		}
	}

	//--------------------------------------------------------------------
	// closeInputStreams
	// Close (invalidate) any InputStreams that are currently in use
	// for the statement
	//--------------------------------------------------------------------

	protected void closeInputStreams ()
	{
		// Close all input streams

		for (int i = 0; i < numberOfCols; i++) {
			boundCols[i].closeInputStream ();
		}
	}

	//--------------------------------------------------------------------
	// setColumnMappings
	// In order to facilitate re-numbering/re-ordering result set columns,
	// an array of integers can be used.  If set, the length of the
	// integer array is the new number of columns in the result set, and
	// each of the elements re-maps the column number (i.e. element 0
	// of the array points to the first column number).  For example,
	// DatabaseMetaData.getTableTypes() needs to return only 1 column
	// in the result set.  The results from the ODBC call getTables returns
	// multiple columns, and the column needed is column number 4.  Thus,
	// once the result set is created:
	//
	//	int	map[] = new int[1];
	//	map[0] = 4;
	//	rs.setColumnMappings (map);
	//
	// This will set the number of result set columns to 1 (the size of
	// the mapping array), and column number 1 (the first element in the
	// array) will use column number 4 from the result set data.
	//--------------------------------------------------------------------

	public void setColumnMappings (
		int	map[])
	{
		colMappings = map;
	}

	//--------------------------------------------------------------------
	// mapColumn
	// Given a column number, map it to the corresponding column in
	// the result set.  If no column mappings exist, return the original
	// column number.  If the column number is out-of-range, return -1.
	//--------------------------------------------------------------------

	public int mapColumn (
		int	column)
	{
		int	map = column;

		if (colMappings != null) {

			// Validate column number

			if ((column > 0) &&
			    (column <= colMappings.length)) {
				map = colMappings[column - 1];
			}
			else {
				// Invalid column number
				map = -1;
			}
		}

		return map;
	}

	//====================================================================
	// Data attributes
	//====================================================================

	protected JdbcOdbc OdbcApi;		// ODBC API interface object

	protected int	hDbc;			// Database connection handle

	protected int	hStmt;			// Statement handle

	protected SQLWarning lastWarning;	// Last SQLWarning generated
						//  by an operation.

	protected boolean keepHSTMT;		// Flag indicating whether the
						//  statement handle should be
						//  kept upon close.  Another
						//  object may 'own' the 
						//  statement handle, and we
						//  don't want to destroy it.

	protected JdbcOdbcBoundCol boundCols[];	// An array of bound column
						//  objects.  See comments
						//  in initialize ()

	protected int	numberOfCols;		// Number of columns in the
						//  result set

	protected int   numResultCols;		// Cache for number of result
						//  columns

	protected int	firstPseudoCol;		// First pseudo column number

	protected int	lastPseudoCol;		// Last pseudo column number
	
	protected JdbcOdbcPseudoCol pseudoCols[];
						// An array of pseudo column
						//  objects.  See comments
						//  in JdbcOdbcPseudoCol

	protected int	colMappings[];		// An array used to map
						//  column numbers.  See 
						//  comments for
						//  setColumnMappings

	protected ResultSetMetaData rsmd;	// ResultSetMetaData object
						//  used for finding column
						//  names and mapping them to
						//  column numbers

	private java.util.Hashtable colNameToNum;
						// Hash table for mapping
						//  column names to numbers
	private java.util.Hashtable colNumToName;
						// Hash table for mapping
						//  column numbers to names

	private boolean lastColumnNull;		// true if the last column
						//  referenced by a getXXX
						//  function was null

	private boolean closed;				// true if the result set
						//  has been closed

	private int sqlTypeColumn;			// If a catalog result set
						//  has a SQL data type, we
						//  need to convert it to
						//  a JDBC SQL type

	private boolean freed;				// true if the statement
						//  has already been freed

	protected Statement ownerStatement;
						// Keep a reference of our
						//  owning statement object



    //--------------------------JDBC 2.0-----------------------------------

    //---------------------------------------------------------------------
    // Getter's and Setter's
    //---------------------------------------------------------------------

    /**
     * JDBC 2.0
     *
     * <p>Get the value of a column in the current row as a java.io.Reader.
     */
    public
    java.io.Reader getCharacterStream(int columnIndex) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * <p>Get the value of a column in the current row as a java.io.Reader.
     */
    public
    java.io.Reader getCharacterStream(String columnName) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Get the value of a column in the current row as a java.math.BigDecimal 
     * object.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @return the column value (full precision); if the value is SQL NULL, 
     * the result is null
     * @exception SQLException if a database-access error occurs.
     */
    public
    BigDecimal getBigDecimal(int columnIndex) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Get the value of a column in the current row as a java.math.BigDecimal 
     * object.
     *
     */
    public
    BigDecimal getBigDecimal(String columnName) { throw new UnsupportedOperationException(); }

    //---------------------------------------------------------------------
    // Traversal/Positioning
    //---------------------------------------------------------------------

    /**
     * JDBC 2.0
     *
     * <p>Determine if the cursor is before the first row in the result 
     * set.   
     *
     * @return true if before the first row, false otherwise. Returns
     * false when the result set contains no rows.
     * @exception SQLException if a database-access error occurs.
     */
    public
    boolean isBeforeFirst() { throw new UnsupportedOperationException(); }
      
    /**
     * JDBC 2.0
     *
     * <p>Determine if the cursor is after the last row in the result 
     * set.   
     *
     * @return true if after the last row, false otherwise.  Returns
     * false when the result set contains no rows.
     * @exception SQLException if a database-access error occurs.
     */
    public
    boolean isAfterLast() { throw new UnsupportedOperationException(); }
 
    /**
     * JDBC 2.0
     *
     * <p>Determine if the cursor is on the first row of the result set.   
     *
     * @return true if on the first row, false otherwise.   
     * @exception SQLException if a database-access error occurs.
     */
    public
    boolean isFirst() { throw new UnsupportedOperationException(); }
 
    /**
     * JDBC 2.0
     *
     * <p>Determine if the cursor is on the last row of the result set.   
     * Note: Calling isLast() may be expensive since the JDBC driver
     * might need to fetch ahead one row in order to determine 
     * whether the current row is the last row in the result set.
     *
     * @return true if on the last row, false otherwise. 
     * @exception SQLException if a database-access error occurs.
     */
    public
    boolean isLast() { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * <p>Moves to the front of the result set, just before the
     * first row. Has no effect if the result set contains no rows.
     *
     * @exception SQLException if a database-access error occurs, or
     * result set type is TYPE_FORWARD_ONLY
     */
    public
    void beforeFirst() { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * <p>Moves to the end of the result set, just after the last
     * row.  Has no effect if the result set contains no rows.
     *
     * @exception SQLException if a database-access error occurs, or
     * result set type is TYPE_FORWARD_ONLY.
     */
    public
    void afterLast() { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * <p>Moves to the first row in the result set.  
     *
     * @return true if on a valid row, false if no rows in the result set.
     * @exception SQLException if a database-access error occurs, or
     * result set type is TYPE_FORWARD_ONLY.
     */
    public
    boolean first() { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * <p>Moves to the last row in the result set.  
     *
     * @return true if on a valid row, false if no rows in the result set.
     * @exception SQLException if a database-access error occurs, or
     * result set type is TYPE_FORWARD_ONLY.
     */
    public
    boolean last() { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * <p>Determine the current row number.  The first row is number 1, the
     * second number 2, etc.  
     *
     * @return the current row number, else return 0 if there is no 
     * current row, or when on the insert row.
     * @exception SQLException if a database-access error occurs.
     */
    public
    int getRow() { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * <p>Move to an absolute row number in the result set.
     *
     * <p>If row is positive, moves to an absolute row with respect to the
     * beginning of the result set.  The first row is row 1, the second
     * is row 2, etc. 
     *
     * <p>If row is negative, moves to an absolute row position with respect to
     * the end of result set.  For example, calling absolute(-1) positions the 
     * cursor on the last row, absolute(-2) indicates the next-to-last
     * row, etc.
     *
     * <p>An attempt to position the cursor beyond the first/last row in
     * the result set, leaves the cursor before/after the first/last
     * row, respectively.
     *
     * <p>Note: Calling absolute(1) is the same as calling first().
     * Calling absolute(-1) is the same as calling last().
     *
     * @return true if on the result set, false if off.
     * @exception SQLException if a database-access error occurs, or 
     * row is 0, or result set type is TYPE_FORWARD_ONLY.
     */
    public
    boolean absolute( int row ) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * <p>Moves a relative number of rows, either positive or negative.
     * Attempting to move beyond the first/last row in the
     * result set positions the cursor before/after the
     * the first/last row. Calling relative(0) is valid, but does
     * not change the cursor position.
     *
     * <p>Note: Calling relative(1) is different than calling next()
     * since is makes sense to call next() when there is no current row,
     * for example, when the cursor is positioned before the first row
     * or after the last row of the result set.
     *
     * @return true if on a row, false otherwise.
     * @exception SQLException if a database-access error occurs, or there
     * is no current row, or result set type is TYPE_FORWARD_ONLY.
     */
    public
    boolean relative( int rows ) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * <p>Moves to the previous row in the result set.  
     *
     * <p>Note: previous() is not the same as relative(-1) since it
     * makes sense to call previous() when there is no current row.
     *
     * @return true if on a valid row, false if off the result set.
     * @exception SQLException if a database-access error occurs, or
     * result set type is TYPE_FORWAR_DONLY.
     */
    public
    boolean previous() { throw new UnsupportedOperationException(); }

    //---------------------------------------------------------------------
    // Properties
    //---------------------------------------------------------------------

    /**
     * JDBC 2.0
     *
     * The rows in a result set will be processed in a forward direction;
     * first-to-last.
     */
    public
    int FETCH_FORWARD = 1000;

    /**
     * JDBC 2.0
     *
     * The rows in a result set will be processed in a reverse direction;
     * last-to-first.
     */
    public
    int FETCH_REVERSE = 1001;

    /**
     * JDBC 2.0
     * 
     * The order in which rows in a result set will be processed is unknown.
     */
    public
    int FETCH_UNKNOWN = 1002;

    /**
     * JDBC 2.0
     *
     * Give a hint as to the direction in which the rows in this result set
     * will be processed.  The initial value is determined by the statement
     * that produced the result set.  The fetch direction may be changed
     * at any time.
     *
     * @exception SQLException if a database-access error occurs, or
     * the result set type is TYPE_FORWARD_ONLY and direction is not 
     * FETCH_FORWARD.
     */
    public
    void setFetchDirection(int direction) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Return the fetch direction for this result set.
     *
     * @exception SQLException if a database-access error occurs 
     */
    public
    int getFetchDirection() { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Give the JDBC driver a hint as to the number of rows that should 
     * be fetched from the database when more rows are needed for this result
     * set.  If the fetch size specified is zero, then the JDBC driver 
     * ignores the value, and is free to make its own best guess as to what
     * the fetch size should be.  The default value is set by the statement 
     * that creates the result set.  The fetch size may be changed at any 
     * time.
     *
     * @param rows the number of rows to fetch
     * @exception SQLException if a database-access error occurs, or the
     * condition 0 <= rows <= this.getMaxRows() is not satisfied.
     */
    public
    void setFetchSize(int rows) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Return the fetch size for this result set.
     *
     * @exception SQLException if a database-access error occurs 
     */
    public
    int getFetchSize() { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     */
    public
    int TYPE_FORWARD_ONLY = 1003;

    /**
     * JDBC 2.0
     */
    public
    int TYPE_SCROLL_INSENSITIVE = 1004;

    /**
     * JDBC 2.0
     */
    public
    int TYPE_SCROLL_SENSITIVE = 1005;

    /**
     * JDBC 2.0
     *
     * Return the type of this result set.  The type is determined based
     * on the statement that created the result set.
     *
     * @return TYPE_FORWARD_ONLY, TYPE_STATIC, TYPE_KEYSET, or 
     * TYPE_DYNAMIC.
     * @exception SQLException if a database-access error occurs
     */
    public
    int getType() { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     * 
     *  
     */
    public
    int CONCUR_READ_ONLY = 1007;

    /**
     * JDBC 2.0
     * 
     *  
     */
    public
    int CONCUR_UPDATABLE = 1008;

    /**
     * JDBC 2.0
     *
     * Return the concurrency of this result set.  The concurrency
     * used is determined by the statement that created the result set.
     *
     * @return the concurrency type, CONCUR_READ_ONLY, etc.
     * @exception SQLException if a database-access error occurs
     */
    public
    int getConcurrency() { throw new UnsupportedOperationException(); }

    //---------------------------------------------------------------------
    // Updates
    //---------------------------------------------------------------------

    /**
     * JDBC 2.0
     *
     * Determine if the current row has been updated. 
     *
     * The value returned depends on whether or not the result set
     * can detect changes, see DatabaseMetaData.rowUpdatesAreDetected().
     *
     * @return true if updated and changes are detected
     * @exception SQLException if a database-access error occurs
     * 
     * @see DatabaseMetaData#detectsRowChanges
     */
    public
    boolean rowUpdated() { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Determine if the current row has been inserted. 
     *
     * The value returned depends on whether or not the result set
     * can detect changes, see DatabaseMetaData.rowInsertsAreDetected().
     *
     * @return true if inserted and changes are detected
     * @exception SQLException if a database-access error occurs
     * 
     * @see DatabaseMetaData#detectsRowChanges
     */
    public
    boolean rowInserted() { throw new UnsupportedOperationException(); }
   
    /**
     * JDBC 2.0
     *
     * Determine if this row has been deleted.  A result set of type
     * keyset or static may contain rows that have been deleted.
     *
     * The value returned depends on whether or not the result set
     * can detect changes, see DatabaseMetaData.rowDeletesAreDetected().
     *
     * @return true if deleted and changes are detected
     * @exception SQLException if a database-access error occurs
     * 
     * @see DatabaseMetaData#detectsRowChanges
     */
    public
    boolean rowDeleted() { throw new UnsupportedOperationException(); }

    /** 
     * The updateXXX() methods are used to update column values in the
     * current row, or the insert row.  
     * The updateXXX() methods do not update the underlying
     * database, instead, they update an in-memory copy of the data read 
     * from the database.
     *  
     * The insert row is a special row associated with an updatable
     * result set.  It is essentially a buffer where a new row may
     * be constructed by calling the updateXXX() methods prior to 
     * inserting the row into the result set.
     *
     * To update the underlying database, updateRow(), insertRow, or
     * deleteRow may be called.
     */

    /**
     * JDBC 2.0
     */
    public
    void updateNull(int columnIndex) { throw new UnsupportedOperationException(); }  

    /**
     * JDBC 2.0
     */
    public
    void updateBoolean(int columnIndex, boolean x) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     */
    public
    void updateByte(int columnIndex, byte x) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     */
    public
    void updateShort(int columnIndex, short x) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     */
    public
    void updateInt(int columnIndex, int x) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     */
    public
    void updateLong(int columnIndex, long x) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     */
    public
    void updateFloat(int columnIndex, float x) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     */
    public
    void updateDouble(int columnIndex, double x) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     */
    public
    void updateBigDecimal(int columnIndex, BigDecimal x) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     */
    public
    void updateString(int columnIndex, String x) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     */
    public
    void updateBytes(int columnIndex, byte x[]) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     */
    public
    void updateDate(int columnIndex, java.sql.Date x) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     */
    public
    void updateTime(int columnIndex, java.sql.Time x) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     */
    public
    void updateTimestamp(int columnIndex, java.sql.Timestamp x)
      { throw new UnsupportedOperationException(); }

    /** 
     * JDBC 2.0
     */
    public
    void updateAsciiStream(int columnIndex, 
			   java.io.InputStream x, 
			   int length) { throw new UnsupportedOperationException(); }

    /** 
     * JDBC 2.0
     */
    public
    void updateBinaryStream(int columnIndex, 
			    java.io.InputStream x,
			    int length) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     */
    public
    void updateCharacterStream(int columnIndex,
			     java.io.Reader reader,
			     int length) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     */
    public
    void updateObject(int columnIndex, Object x, int scale)
      { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     */
    public
    void updateObject(int columnIndex, Object x) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     */
    public
    void updateNull(String columnName) { throw new UnsupportedOperationException(); }  

    /**
     * JDBC 2.0
     */
    public
    void updateBoolean(String columnName, boolean x) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     */
    public
    void updateByte(String columnName, byte x) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     */
    public
    void updateShort(String columnName, short x) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     */
    public
    void updateInt(String columnName, int x) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     */
    public
    void updateLong(String columnName, long x) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     */
    public
    void updateFloat(String columnName, float x) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     */
    public
    void updateDouble(String columnName, double x) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     */
    public
    void updateBigDecimal(String columnName, BigDecimal x) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     */
    public
    void updateString(String columnName, String x) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     */
    public
    void updateBytes(String columnName, byte x[]) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     */
    public
    void updateDate(String columnName, java.sql.Date x) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     */
    public
    void updateTime(String columnName, java.sql.Time x) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     */
    public
    void updateTimestamp(String columnName, java.sql.Timestamp x)
      { throw new UnsupportedOperationException(); }

    /** 
     * JDBC 2.0
     */
    public
    void updateAsciiStream(String columnName, 
			   java.io.InputStream x, 
			   int length) { throw new UnsupportedOperationException(); }

    /** 
     * JDBC 2.0
     */
    public
    void updateBinaryStream(String columnName, 
			    java.io.InputStream x,
			    int length) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     */
    public
    void updateCharacterStream(String columnName,
			     java.io.Reader reader,
			     int length) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     */
    public
    void updateObject(String columnName, Object x, int scale)
      { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     */
    public
    void updateObject(String columnName, Object x) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Insert the contents of the insert row into the result set and
     * the database.  Must be on the insert row when this method is called.
     *
     * @exception SQLException if a database-access error occurs
     */
    public
    void insertRow() { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Update the underlying database with the new contents of the
     * current row.  Cannot be called when on the insert row.
     */
    public
    void updateRow() { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Delete the current row from the result set and the underlying
     * database.  Cannot be called when on the insert row.
     *
     * @exception SQLException if a database-access error occurs
     */
    public
    void deleteRow() { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Refresh the value of the current row with the value in the database.
     * Cannot be called when on the insert row.
     *
     * @exception SQLException if a database-access error occurs
     */
    public
    void refreshRow() { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Move to the insert row.  The current row is remembered.
     *
     * Only the updateXXX() and insertRow() methods may be called 
     * when the result set is in insert mode.
     *
     * @exception SQLException if a database-access error occurs
     */
    public
    void moveToInsertRow() { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Move the cursor to the current row.  Has no effect unless the
     * cursor is on the insert row. 
     *
     * @exception SQLException if a database-access error occurs
     */
    public
    void moveToCurrentRow() { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Return the Statement that produced the ResultSet.
     */
    public
    Statement getStatement() { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Returns the value of column @i as a Java object.  Use the 
     * @map to determine the class from which to construct data of 
     * SQL structured and distinct types.
     *
     * @param i the first column is 1, the second is 2, ...
     * @param map the mapping from SQL type names to Java classes
     * @return an object representing the SQL value
     */
    public
    Object getObject(int i, java.util.Map map) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Get a REF(&lt;structured-type&gt;) column.
     *
     * @param i the first column is 1, the second is 2, ...
     * @return an object representing data of an SQL REF type
     */
    public
    Ref getRef(int i) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Get a BLOB column.
     *
     * @param i the first column is 1, the second is 2, ...
     * @return an object representing a BLOB
     */
    public
    Blob getBlob(int i) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Get a CLOB column.
     *
     * @param i the first column is 1, the second is 2, ...
     * @return an object representing a CLOB
     */
    public
    Clob getClob(int i) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Get an array column.
     *
     * @param i the first column is 1, the second is 2, ...
     * @return an object representing an SQL array
     */
    public
    Array getArray(int i) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Returns the value of column @i as a Java object.  Use the 
     * @map to determine the class from which to construct data of 
     * SQL structured and distinct types.
     *
     * @param colName the column name
     * @param map the mapping from SQL type names to Java classes
     * @return an object representing the SQL value
     */
    public
    Object getObject(String colName, java.util.Map map) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Get a REF(&lt;structured-type&gt;) column.
     *
     * @param colName the column name
     * @return an object representing data of an SQL REF type
     */
    public
    Ref getRef(String colName) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Get a BLOB column.
     *
     * @param colName the column name
     * @return an object representing a BLOB
     */
    public
    Blob getBlob(String colName) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Get a CLOB column.
     *
     * @param colName the column name
     * @return an object representing a CLOB
     */
    public
    Clob getClob(String colName) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Get an array column.
     *
     * @param colName the column name
     * @return an object representing an SQL array
     */
    public
    Array getArray(String colName) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Get the value of a column in the current row as a java.sql.Date object.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @return the column value; if the value is SQL NULL, the result is null
     * @exception SQLException if a database-access error occurs.
     */
    public java.sql.Date getDate(int columnIndex, Calendar cal) { throw new UnsupportedOperationException(); }

    /**
     * Get the value of a column in the current row as a java.sql.Date object.
     *
     * @param columnName is the SQL name of the column
     * @return the column value; if the value is SQL NULL, the result is null
     * @exception SQLException if a database-access error occurs.
     */
    public
    java.sql.Date getDate(String columnName, Calendar cal) { throw new UnsupportedOperationException(); }

    /**
     * Get the value of a column in the current row as a java.sql.Time object.
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @return the column value; if the value is SQL NULL, the result is null
     * @exception SQLException if a database-access error occurs.
     */
    public
    java.sql.Time getTime(int columnIndex, Calendar cal) { throw new UnsupportedOperationException(); }

    /**
     * Get the value of a column in the current row as a java.sql.Time object.
     *
     * @param columnName is the SQL name of the column
     * @return the column value; if the value is SQL NULL, the result is null
     * @exception SQLException if a database-access error occurs.
     */
    public
    java.sql.Time getTime(String columnName, Calendar cal) { throw new UnsupportedOperationException(); }

    /**
     * Get the value of a column in the current row as a java.sql.Timestamp 
     * object.    
     *
     * @param columnIndex the first column is 1, the second is 2, ...
     * @return the column value; if the value is SQL NULL, the result is null
     * @exception SQLException if a database-access error occurs.
     */
    public
    java.sql.Timestamp getTimestamp(int columnIndex, Calendar cal) 
      { throw new UnsupportedOperationException(); }

    /**
     * Get the value of a column in the current row as a java.sql.Timestamp 
     * object.
     *
     * @param columnName is the SQL name of the column
     * @return the column value; if the value is SQL NULL, the result is null
     * @exception SQLException if a database-access error occurs.
     */
    public
    java.sql.Timestamp getTimestamp(String columnName, Calendar cal)	
      { throw new UnsupportedOperationException(); }          

	public void cancelRowUpdates ()
	{ throw new UnsupportedOperationException(); }
}


