/*
 * @(#)JdbcOdbcStatement.java	1.19 98/05/13
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
//----------------------------------------------------------------------------
//
// Module:      JdbcOdbcStatement.java
//
// Description: Impementation of the Statement interface class
//
// Product:     JDBCODBC (Java DataBase Connectivity using
//              Open DataBase Connectivity)
//
// Author:      Karl Moss
//
// Date:        March, 1996
//
// Copyright:   1990-1996 INTERSOLV, Inc.
//              This software contains confidential and proprietary
//              information of INTERSOLV, Inc.
//----------------------------------------------------------------------------

package sun.jdbc.odbc;

import java.util.Hashtable;
import java.math.BigDecimal;
import java.sql.*;

public class JdbcOdbcStatement
	extends		JdbcOdbcObject
	implements	java.sql.Statement {

	//====================================================================
	// Public methods
	//====================================================================

	//--------------------------------------------------------------------
	// Constructor
	// Perform any necessary initialization.
	//--------------------------------------------------------------------

	public JdbcOdbcStatement (
		JdbcOdbcConnectionInterface con)
	{
		OdbcApi = null;
		hDbc    = OdbcDef.SQL_NULL_HDBC;
		hStmt   = OdbcDef.SQL_NULL_HSTMT;
		lastWarning = null;
		myConnection = con;
	}

	//--------------------------------------------------------------------
	// finalize
	// Perform any cleanup when this object is garbage collected
	//--------------------------------------------------------------------

	protected void finalize ()
	{
		if (isTracing ()) {
			trace ("Statement.finalize " + this);
		}

		try {
			close ();
		}
		catch (SQLException ex) {
			// If an exception is thrown, ignore it
		}
	}

	//--------------------------------------------------------------------
	// initialize
	// Initialize the Statement object.  Give the ODBC API interface
	// object, the connection handle, and optionally the statement
	// handle.  If no statement handle is given, one is created.
	//--------------------------------------------------------------------

	public void initialize (
		JdbcOdbc odbcApi,
		int hdbc)
		throws SQLException
	{
		// Allocate a new statement handle//

		int hstmt = odbcApi.SQLAllocStmt (hdbc);

		initialize (odbcApi, hdbc, hstmt, null);
	}

	public void initialize (
		JdbcOdbc odbcApi,
		int hdbc,
		int hstmt,
		Hashtable info)
		throws SQLException
	{
		OdbcApi = odbcApi;
		hDbc = hdbc;
		hStmt = hstmt;
		typeInfo = info;
	}

	//--------------------------------------------------------------------
	// executeQuery
	// executeQuery method executes the given SQL string which must
	// be a query that returns a single ResultSet.
	// (See also the more general "execute" below.)
	//--------------------------------------------------------------------

	public ResultSet executeQuery (
		String sql)
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*Statement.executeQuery (" + sql + ")");
		}
		ResultSet rs = null;

		// Execute the statement.  If execute returns true, a result
		// set exists.

		if (execute (sql)) {
			rs = getResultSet (false);
		}
		else {

			// No ResultSet was produced.  Raise an exception

			throw new SQLException ("No ResultSet was produced");
		}
		return rs;
	}

	//--------------------------------------------------------------------
	// executeUpdate
	// This method execute a SQL statement that is known to be a simple 
	// database update (such as a SQL UPDATE, INSERT, DELETE, ...)
	// The result is the number of rows that have been modified. 
	// (See also the more general "execute" below.)
	//--------------------------------------------------------------------
	
		public int executeUpdate (
			String sql)
			throws SQLException
		{
			if (isTracing ()) {
				trace ("*Statement.executeUpdate (" + sql + ")");
			}
			int numRows = -1;
	
			// Execute the statement.  If execute returns false, a
			// row count exists.
	
			if (!execute (sql)) {
				numRows = getUpdateCount ();
			}
			else {
	
				// No update count was produced (a ResultSet was).  Raise
				// an exception
	
				throw new SQLException ("No row count was produced");
			}
			return numRows;
		}
	
		//--------------------------------------------------------------------
		// execute
		// The "execute" method execute a SQL statement and returns "true"
		// if the first result is a ResultSet, false otherwise.  You can
		// then use getResultSet or getUpdateCount to retrieve the result,
		// and getMoreResults to move to any subseuqnet result(s).
		//--------------------------------------------------------------------
	
		public synchronized boolean execute (
			String sql)
			throws SQLException
		{
			if (isTracing ()) {
				trace ("*Statement.execute (" + sql + ")");
			}
			boolean hasResultSet = false;
			SQLWarning	warning = null;
	
			// Reset the statement handle and warning
	
			reset ();
	
			// Check for a 'FOR UPDATE' statement.  If present, change
			// the concurrency to lock
	
			lockIfNecessary (sql);
	
			// Call SQLExecDirect
	
			try {
				OdbcApi.SQLExecDirect (hStmt, sql);
			}
			catch (SQLWarning ex) {
	
				// Save pointer to warning and save with ResultSet
				// object once it is created.
	
				warning = ex;
			}
					
			// Now determine if there is a result set associated with
			// the SQL statement that was executed.  Get the column
			// count, and if it is not zero, there is a result set.
	
			if (getColumnCount () > 0) {
				hasResultSet = true;
			}
	
			return hasResultSet;
		}
	
		//--------------------------------------------------------------------
		// getResultSet
		// getResultSet returns the current result as a ResultSet.  It 
		// returns NULL if the current result is not a ResultSet.
		//--------------------------------------------------------------------
		public ResultSet getResultSet ()
			throws SQLException
		{
			if (isTracing ()) {
				trace ("*Statement.getResultSet");
			}
			return getResultSet (true);
		}
	
		//--------------------------------------------------------------------
		// getResultSet
		// getResultSet returns the current result as a ResultSet.  It 
		// returns NULL if the current result is not a ResultSet.
		//--------------------------------------------------------------------
		public ResultSet getResultSet (
			boolean checkCount)
			throws SQLException
		{
	
				 if (myResultSet != null)  // if resultset already retrieved,
					  {                // throw exception to avoid sequence error
					   throw new SQLException ("Invalid state for getResultSet");
					  }
	
			JdbcOdbcResultSet rs = null;
			int numCols = 1;
	
			// If we already know we have result columns, checkCount
			// is false.  This is an optimization to prevent unneeded
			// calls to getColumnCount
	
			if (checkCount) {
				numCols = getColumnCount ();
			}
	
			// Only return a result set if there are result columns
	
			if (numCols > 0) {
				rs = new JdbcOdbcResultSet ();
				rs.initialize (OdbcApi, hDbc, hStmt, true, this);
	
				// Save a copy of our last result set
	
				myResultSet = rs;
			}
			else {
				clearMyResultSet ();
			}
			return rs;
		}
	
		//--------------------------------------------------------------------
		// getUpdateCount
		// getUpdateCount returns the current result, which should be an update
		// count.   It returns -1 if there are no more results or if the
		// current result is a ResultSet.
		//--------------------------------------------------------------------
		public int getUpdateCount ()
			throws SQLException
		{
			if (isTracing ()) {
				trace ("*Statement.getUpdateCount");
			}
			int rowCount = -1;
	
			// Only return a row count for SQL statements that did not
			// return a result set.
	
			if (getColumnCount () == 0) {
				rowCount = getRowCount ();
			}
			return rowCount;
		}
	
		//--------------------------------------------------------------------
		// close
		// close frees up internal state associated with the statement.
		// You should normally call close when you are done with a statement.
		// N.B. any ResultSets associated with the Statement will become 
		// unusable when the Statement is closed.
		//--------------------------------------------------------------------
	
		public synchronized void close ()
			throws SQLException
		{
			if (isTracing ()) {
				trace ("*Statement.close");
			}
	
			// Close/clear our result set
	
			clearMyResultSet ();
	
			// Reset last warning message
	
			try {
				clearWarnings ();
				if (hStmt != OdbcDef.SQL_NULL_HSTMT) {
					OdbcApi.SQLFreeStmt (hStmt, OdbcDef.SQL_DROP);
					hStmt = OdbcDef.SQL_NULL_HSTMT;
				}
			}
			catch (SQLException ex) {
				// If we get an error, ignore
			}
	
			// Remove this Statement object from the Connection object's
			// list
	
			myConnection.deregisterStatement (this);
		}
	
		//--------------------------------------------------------------------
		// reset
		// Frees up internal state associated with the statement before re-using
		// the statement
		//--------------------------------------------------------------------
	
		protected void reset ()
			throws SQLException
		{
			clearWarnings ();
	
			if (myResultSet != null) {
				clearMyResultSet ();
			}
			else {
				OdbcApi.SQLFreeStmt (hStmt, OdbcDef.SQL_CLOSE);
			}
		}
	
	
		//--------------------------------------------------------------------
		// getMoreResults
		// getMoreResults moves to the next result.  It returns true if 
		// this result is a ResultSet.  getMoreResults also implicitly
		// closes the current ResultSet.
		//--------------------------------------------------------------------
	
		public boolean getMoreResults ()
			throws SQLException
		{
			if (isTracing ()) {
				trace ("*Statement.getMoreResults");
			}
			JdbcOdbcResultSet rs = null;
			SQLWarning	warning = null;
			boolean		hasResultSet = false;
	
			// clear previous warnings
	
			clearWarnings ();
	
			// Call SQLMoreResults
	
			try {
				hasResultSet = OdbcApi.SQLMoreResults (hStmt);
			}
			catch (SQLWarning ex) {
	
				// Save pointer to warning and save with ResultSet
				// object once it is created.
	
				warning = ex;
			}
					
			// There are more results (it may not be a result set, though)
	
			if (hasResultSet) {
	
				// Now determine if there is a result set associated
				// with the SQL statement that was executed.  Get the
				// column count, and if it is zero, there is not a
				// result set.
	
				if (getColumnCount () == 0) {
					hasResultSet = false;
				}
			}
	
			// Set the warning for the statement, if one was generated
	
			setWarning (warning);
	
			// Return the result set indicator
	
			return hasResultSet;
		}
	
	
		//--------------------------------------------------------------------
		// getMaxFieldSize
		// The maxFieldSize is a limit (in bytes) on how much data can be 
		// returned as part of a field.  If the limit is exceeded, the data
		// is truncated and a warning is added (see Statement.getWarnings).
		// You should normally not need to change the default limit.
		// Zero means no limit.
		//--------------------------------------------------------------------
	
		public int getMaxFieldSize ()
			throws SQLException
		{
			if (isTracing ()) {
				trace ("*Statement.getMaxFieldSize");
			}
			return getStmtOption (OdbcDef.SQL_MAX_LENGTH);
		}
	
		//--------------------------------------------------------------------
		// setMaxFieldSize
		//--------------------------------------------------------------------
	
		public void setMaxFieldSize (
			int max)
			throws SQLException
		{
			if (isTracing ()) {
				trace ("*Statement.setMaxFieldSize (" + max + ")");
			}
			OdbcApi.SQLSetStmtOption (hStmt, OdbcDef.SQL_MAX_LENGTH, max);
		}
	
	
		//--------------------------------------------------------------------
		// getMaxRows
		// The maxRows value is a limit on how many rows can be returned as
		// part of a ResutlSet.  You should normally not need to change
		// the default limit.  Zero means no limit.
		//--------------------------------------------------------------------
	
		public int getMaxRows ()
			throws SQLException
		{
			if (isTracing ()) {
				trace ("*Statement.getMaxRows");
			}
			return getStmtOption (OdbcDef.SQL_MAX_ROWS);
		}
	
		//--------------------------------------------------------------------
		// setMaxRows
		//--------------------------------------------------------------------
	
		public void setMaxRows (
			int max)
			throws SQLException
		{
			if (isTracing ()) {
				trace ("*Statement.setMaxRows (" + max + ")");
			}
			OdbcApi.SQLSetStmtOption (hStmt, OdbcDef.SQL_MAX_ROWS, max);
		}
	
	
		//--------------------------------------------------------------------
		// setEscapeProcessing
		// if escape scanning is on (the default) the driver will do escape
		// substitution before sending the SQL to the database.
		//--------------------------------------------------------------------
	
		public void setEscapeProcessing (
			boolean enable)
			throws SQLException
		{
			if (isTracing ()) {
				trace ("*Statement.setEscapeProcessing (" +
					enable + ")");
			}
			int	value = OdbcDef.SQL_NOSCAN_OFF;
			if (!enable) {
				value = OdbcDef.SQL_NOSCAN_ON;
			}
			OdbcApi.SQLSetStmtOption (hStmt, OdbcDef.SQL_NOSCAN, value);
		}
	
		//--------------------------------------------------------------------
		// getQueryTimeout
		// The queryTimeout is how many seconds the driver will allow for an
		// SQL statement to execute before giving up.  Zero means no limit.
		//--------------------------------------------------------------------
	
		public int getQueryTimeout ()
			throws SQLException
		{
			if (isTracing ()) {
				trace ("*Statement.getQueryTimeout");
			}
			return getStmtOption (OdbcDef.SQL_QUERY_TIMEOUT);
		}
	
		//--------------------------------------------------------------------
		// setQueryTimeout
		//--------------------------------------------------------------------
	
		public void setQueryTimeout (
			int seconds)
			throws SQLException
		{
			if (isTracing ()) {
				trace ("*Statement.setQueryTimeout (" +
					seconds + ")");
			}
			OdbcApi.SQLSetStmtOption (hStmt, OdbcDef.SQL_QUERY_TIMEOUT,
					seconds);
		}
	
	
		//--------------------------------------------------------------------
		// cancel
		// Cancel can be used by one thread to cancel a statement that
		// is being executed by another thread.
		//--------------------------------------------------------------------
	
		public void cancel ()
			throws SQLException
		{
			if (isTracing ()) {
				trace ("*Statement.cancel");
			}
	
			// Clear any warnings
			clearWarnings ();
	
			// Call SQLCancel
	
			try {
				OdbcApi.SQLCancel (hStmt);
			}
			catch (SQLWarning ex) {
	
				// Save last warning
	
				setWarning (ex);
			}
		}
	
		//--------------------------------------------------------------------
		// getWarnings will return any warning information related to
		// the current statement execution.  Note that SQLWarning may be
		// a chain.  If a statement is re-executed then warnings associated
		// with its previous execution will be automatically discarded.
		//--------------------------------------------------------------------
	
		public SQLWarning getWarnings ()
			throws SQLException
		{
			if (isTracing ()) {
				trace ("*Statement.getWarnings");
			}
			return lastWarning;
		}
	
		public void clearWarnings ()
			throws SQLException
		{
			lastWarning = null;
		}
	
		//--------------------------------------------------------------------
		// setWarning
		// Sets the warning 
		//--------------------------------------------------------------------
	
		public void setWarning (
			SQLWarning ex)
			throws SQLException
		{
			lastWarning = ex;
		}
	
		//--------------------------------------------------------------------
		// setCursorName
		// Defines the SQL cursor name that will be used by
		// subsequent Statement execute methods.
		//--------------------------------------------------------------------
	
		public void setCursorName (
			String name)
			throws SQLException
		{
			if (isTracing ()) {
				trace ("*Statement.setCursorName " + name + ")");
			}
			OdbcApi.SQLSetCursorName (hStmt, name);
		}
	
		//====================================================================
		// Protected methods
		//====================================================================
	
		//--------------------------------------------------------------------
		// getStmtOption
		// Invoke SQLGetStmtOption with the given option.
		//--------------------------------------------------------------------
	
		protected int getStmtOption (short fOption)
			throws SQLException
		{
			int	result = 0;
	
			// Reset last warning message
	
			clearWarnings ();
	
			try {
				result = OdbcApi.SQLGetStmtOption (hStmt, fOption);
			}
			catch (JdbcOdbcSQLWarning ex) {
	
				// If we got a warning, the return value was saved off
				// in the warning object.  Get it back and set the
				// return value.
	
				BigDecimal n = (BigDecimal) ex.value;
				result = n.intValue ();
	
				setWarning (JdbcOdbc.convertWarning (ex));
			}
			return result;
		}
	
		//--------------------------------------------------------------------
		// getColumnCount
		// Return the number of columns in the ResultSet
		//--------------------------------------------------------------------
	
		protected int getColumnCount ()
			throws SQLException
		{
			int	numCols = 0;
	
			try {
				numCols = OdbcApi.SQLNumResultCols (hStmt);
			}
			catch (JdbcOdbcSQLWarning ex) {
	
				// If we got a warning, the return value was saved off
				// in the warning object.  Get it back and set the
				// return value.
	
				BigDecimal n = (BigDecimal) ex.value;
				numCols = n.intValue ();
			}
			return numCols;
		}
	
		//--------------------------------------------------------------------
		// getRowCount
		// The total number of rows returned by the query.  Note that on some
		// databases this method may be very expensive.
		//--------------------------------------------------------------------
	
		protected int getRowCount ()
			throws SQLException
		{
			int	numRows = 0;
	
			try {
				numRows = OdbcApi.SQLRowCount (hStmt);
			}
			catch (JdbcOdbcSQLWarning ex) {
	
				// If we got a warning, the return value was saved off
				// in the warning object.  Get it back and set the
				// return value.
	
				BigDecimal n = (BigDecimal) ex.value;
				numRows = n.intValue ();
			}
			return numRows;
		}
	
		//--------------------------------------------------------------------
		// lockIfNecessary
		// If the given SQL statement contains a 'FOR UPDATE' clause, change
		// the concurrency to lock so that the row can then be updated.  Returns
		// true if the concurrency has been changed
		//--------------------------------------------------------------------
	
		protected boolean lockIfNecessary (
			String sql)
			throws SQLException
		{
			boolean	rc = false;
	
			// First, convert the statement to upper case
			
			String sqlStatement = sql.toUpperCase ();
	
			// Now, look for the FOR UPDATE keywords.  If there is any extra white
			// space between the FOR and UPDATE, this will fail.
	
			int index = sqlStatement.indexOf (" FOR UPDATE");
	
			// We found it.  Change our concurrency level to ensure that the
			// row can be updated.
	
			if (index > 0) {
				if (isTracing ()) {
					trace ("Setting concurrency for update");
				}
	
				try {
					OdbcApi.SQLSetStmtOption (hStmt, OdbcDef.SQL_CONCURRENCY,
										OdbcDef.SQL_CONCUR_LOCK);
				}
				catch (SQLWarning warn) {
	
					// Catch any warnings and place on the warning stack
					setWarning (warn);
				}
				rc = true;
			}
	
			return rc;
		}
	
		//--------------------------------------------------------------------
		// getPrecision
		// Given a SQL type, return the maximum precision for the column.
		// Returns -1 if not known
		//--------------------------------------------------------------------
	
		protected int getPrecision (	
			int sqlType)
		{
			int prec = -1;
			JdbcOdbcTypeInfo info;
	
			if (typeInfo != null) {
				info = (JdbcOdbcTypeInfo) typeInfo.get (
							new Integer (sqlType));
	
				if (info != null) {
					prec = info.getPrec ();
				}
			}
			
			return prec;
		}
	
		//--------------------------------------------------------------------
		// clearMyResultSet
		// If a ResultSet was created for this Statement, close it
		//--------------------------------------------------------------------
	
		protected synchronized void clearMyResultSet ()
			throws SQLException
		{
			if (myResultSet != null) {
				myResultSet.close ();
				myResultSet = null;
			}
		}
								   
		//====================================================================
		// Data attributes
		//====================================================================
	
		protected JdbcOdbc OdbcApi;		// ODBC API interface object
	
		protected int	hDbc;			// Database connection handle
	
		protected int	hStmt;			// Statement handle
	
		protected SQLWarning lastWarning;	// Last SQLWarning generated
							//  by an operation
	
		protected Hashtable typeInfo;	// Hashtable containing an entry
							//  for each row returned by
							//  DatabaseMetaData.getTypeInfo.
	
		protected ResultSet myResultSet;// The last ResultSet created
							//  for this Statement
	
		protected JdbcOdbcConnectionInterface myConnection;
							// The owning Connection object


		//--------------------------JDBC 2.0-----------------------------


		/**
		 * JDBC 2.0
		 *
		 * Give a hint as to the direction in which the rows in a result set
		 * will be processed. The hint applies only to result sets created 
		 * using this Statement object.  The default value is 
		 * ResultSet.FETCH_FORWARD.
		 *
		 * @param direction the initial direction for processing rows
		 * @exception SQLException if a database-access error occurs or direction
		 * is not one of ResultSet.FETCH_FORWARD, ResultSet.FETCH_REVERSE, or
		 * ResultSet.FETCH_UNKNOWN
		 */
		public void setFetchDirection(int direction) throws SQLException {
			throw new UnsupportedOperationException();
		}

		/**
		 * JDBC 2.0
		 *
		 * Determine the fetch direction.
		 *
		 * @return the default fetch direction
		 * @exception SQLException if a database-access error occurs
		 */
		public int getFetchDirection() throws SQLException {
			throw new UnsupportedOperationException();
		}

		/**
		 * JDBC 2.0
		 *
		 * Give the JDBC driver a hint as to the number of rows that should 
		 * be fetched from the database when more rows are needed.  The number 
		 * of rows specified only affects result sets created using this 
		 * statement. If the value specified is zero, then the hint is ignored.
		 * The default value is zero.
		 *
		 * @param rows the number of rows to fetch
		 * @exception SQLException if a database-access error occurs, or the
		 * condition 0 <= rows <= this.getMaxRows() is not satisfied.
		 */
		public void setFetchSize(int rows) throws SQLException{
			throw new UnsupportedOperationException();
		}
	  
		/**
		 * JDBC 2.0
		 *
		 * Determine the default fetch size.
		 */
		public int getFetchSize() throws SQLException{
			throw new UnsupportedOperationException();
		}

		/**
		 * JDBC 2.0
		 *
		 * Determine the default concurrency.
		 */
		public int getResultSetConcurrency() throws SQLException{
			throw new UnsupportedOperationException();
		}

		/**
		 * JDBC 2.0
		 *
		 * Determine the default result set type.
		 */
		public int getResultSetType() {
			throw new UnsupportedOperationException();
		}

		/**
		 * JDBC 2.0
		 *
		 * Adds a SQL command to the current batch of commmands for the statement.
		 * This method is optional.
		 *
		 * @param sql typically this is a static SQL INSERT or UPDATE statement
		 * @exception SQLException if a database-access error occurs, or the
		 * driver does not support batch statements
		 */
		public void addBatch( String sql ) throws SQLException{
			throw new UnsupportedOperationException();
		}

		/**
		 * JDBC 2.0
		 *
		 * Make the set of commands in the current batch empty.
		 * This method is optional.
		 *
		 * @exception SQLException if a database-access error occurs, or the
		 * driver does not support batch statements
		 */
		public void clearBatch() throws SQLException{
			throw new UnsupportedOperationException();
		}

		/**
		 * JDBC 2.0
		 * 
		 * Submit a batch of commands to the database for execution.
		 * This method is optional.
		 *
		 * @return an array of update counts containing one element for each
		 * command in the batch.  The array is ordered according 
		 * to the order in which commands were inserted into the batch
		 * @exception SQLException if a database-access error occurs, or the
		 * driver does not support batch statements
		 */
		public int[] executeBatch() throws SQLException {
			throw new UnsupportedOperationException();
		}

		/**
		 * JDBC 2.0
		 * 
		 * Return the Connection that produced the Statement.
		 */
		public Connection getConnection()  throws SQLException {
			throw new UnsupportedOperationException();
		}

	}
