/*
 * @(#)KeepAliveCache.java	1.19 98/06/29
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.net.www.http;

import java.io.IOException;
import java.io.NotSerializableException;
import java.util.Vector;
import java.util.Enumeration;
import java.util.Hashtable;
import java.net.URL;

/**
 * A class that implements a cache of idle Http connections for keep-alive
 *
 * @version 1.19, 06/29/98
 * @author Stephen R. Pietrowicz (NCSA)
 * @author Dave Brown
 */
public class KeepAliveCache extends Hashtable {

    /* maximum # keep-alive connections to maintain at once
     * This should be 2 by the HTTP spec, but we can hog bandwidth
     * by cranking it up.  It can be reset by setting system property
     * "http.maxConnections".
     */
    static int getMaxConnections() {
	int result;
	result = ((Integer)java.security.AccessController.doPrivileged(
	        new sun.security.action.GetIntegerAction("http.maxConnections",
							 2))).intValue();
        if (result <= 0) {
	    return 2;
	} else {
	    return result;
        }
    }

    static final int LIFETIME = 5000;
    // proxy timeouts can be longer
    static final int PROXY_LIFETIME = 10000;
    /**
     * Constructor
     */
    public KeepAliveCache() {}

    /**
     * Register this URL and HttpClient (that supports keep-alive) with the cache
     * @param url  The URL contains info about the host and port
     * @param http The HttpClient to be cached
     */
    public synchronized void put(final URL url, HttpClient http) {
	KeepAliveKey key = new KeepAliveKey(url);
	ClientVector v = (ClientVector)super.get(key);
	if (v == null) {

	    /* Unfortunately, we can't always believe the keep-alive timeout we got
	     * back from the server.  If I'm connected through a Netscape proxy
	     * to a server that sent me a keep-alive
	     * time of 15 sec, the proxy unilaterally terminates my connection
	     * The robustness to to get around this is in HttpClient.parseHTTP()
	     */
	    v = new ClientVector(this, key,
				 (http.usingProxy) ? PROXY_LIFETIME: LIFETIME);
	    v.put(http);
	    super.put(key, v);

	    final ClientVector finalV = v;

	    java.security.AccessController.doPrivileged(
		new java.security.PrivilegedAction() {
		public Object run() {
		    Thread t = null;
		    t = new Thread(finalV, "Keep-Alive-Timer:"+url.getHost());
		    t.setDaemon(true);
		    t.setPriority(Thread.MAX_PRIORITY - 2);
		    t.start();
		    return null;
		}
	    });
	} else {
	    v.put(http);
	}
    }

    /* remove an obsolete HttpClient from it's VectorCache */
    /*
    public synchronized void remove (HttpClient h) {
	KeepAliveKey key = new KeepAliveKey(h.url);
	ClientVector v = (ClientVector)super.get(key);
	if (v != null) {
	    v.remove(h);
	}
    }
    */
    /* called by a clientVector thread when all it's connections have timed out
     * and that vector of connections should be removed.
     */
    synchronized void removeVector(KeepAliveKey k) {
	super.remove(k);
    }

    /**
     * Check to see if this URL has a cached HttpClient
     */
    public Object get(URL url) {

	KeepAliveKey key = new KeepAliveKey(url);
	ClientVector v = (ClientVector)super.get(key);
	if (v == null) { // nothing in cache yet
	    return null;
	}
	return v.get();
    }

    /*
     * Do not serialize this class!
     */
    private void writeObject(java.io.ObjectOutputStream stream)
    throws IOException {
	throw new NotSerializableException();
    }

    private void readObject(java.io.ObjectInputStream stream)
    throws IOException, ClassNotFoundException {
	throw new NotSerializableException();
    }
}

/* FILO order for recycling HttpClients, should run in a thread
 * to time them out.  If > maxConns are in use, block.
 */


class ClientVector extends java.util.Stack implements Runnable {

    /* #presently checked out clients that we know about*/
    private int num = 0;
    // cache we're registered with
    private KeepAliveCache cache;
    private KeepAliveKey key;
    HttpClient[] inUse;

    // sleep time in milliseconds, before cache clear
    private int nap;

    ClientVector (KeepAliveCache cache, KeepAliveKey key, int nap) {
	this.cache=cache;
	this.key=key;
	this.nap = nap;
	inUse = new HttpClient[KeepAliveCache.getMaxConnections()];
    }

    synchronized HttpClient get() {
	if (empty()) {
	    return null;
	} else {
	    HttpClient hc = (HttpClient) pop();
	    if (inUse.length == num) {
		HttpClient[] tmp = new HttpClient[inUse.length * 2];
		System.arraycopy(inUse, 0, tmp, 0, inUse.length);
		inUse = tmp;
	    }
	    inUse[num++] = hc;
	    return hc;
	}
    }

    /* return a still valid, unused HttpClient */
    synchronized void put(HttpClient h) {
	if (num > 0) {
	    num--;
	}
	/* it's no longer in use */
	for (int i = 0; i < inUse.length; i++) {
	    if (inUse[i] == h) {
		inUse[i] = null;
		break;
	    }
	}
	push(h);
	notifyAll();
    }

    /* delete one no longer keeping alive */
    /*
    synchronized void remove(HttpClient h) {
	if (num > 0) {
	    num--;
	    for (int i = 0; i < inUse.length; i++) {
		if (h == inUse[i]) {
		    inUse[i] = null;
		    break;
		}
	    }
	}
	notifyAll();
    }
    */

    /* Sleeps for an alloted timeout, then checks for timed out connections.
     * Errs on the side of caution (leave connections idle for a relatively
     * short time).
     */
    public void run() {
	int total_cache;
	do {
	    try {
		Thread.sleep(nap);
	    } catch (InterruptedException e) {}
	    HttpClient h;
	    synchronized (cache) {
		synchronized (this) {
		    /* Remove all unused HttpClients.  Starting from the
		     * bottom of the stack (the least-recently used first).
		     * REMIND: It'd be nice to not remove *all* connections
		     * that aren't presently in use.  One could have been added
		     * a second ago that's still perfectly valid, and we're
		     * needlessly axing it.  But it's not clear how to do this
		     * cleanly, and doing it right may be more trouble than it's
		     * worth.
		     */
		    for (int i = 0; i < size(); i++) {
			h = (HttpClient) elementAt(i);
			h.closeServer();
		    }
		    removeAllElements();
		    if (num == 0) {
			cache.removeVector(key);
			return;
		    }
		}
	    }
	    /* don't do these callbacks under synchronization! */
	    for (int i = 0; i < inUse.length; i++) {
		if ((h = inUse[i]) != null) {
		    if (h.kas.hurry()) {
			put(h);
		    }
		}
	    }

	    synchronized (this) {
		total_cache = num + size();
	    }
	} while (total_cache > 0);
	cache.removeVector(key);
	return;
    }
}


class KeepAliveKey {
    private String	protocol = null;
    private String	host = null;
    private int		port = 0;

    /**
     * Constructor
     *
     * @param url the URL containing the protocol, host and port information
     */
    public KeepAliveKey(URL url) {
	this.protocol = url.getProtocol();
	this.host = url.getHost();
	this.port = url.getPort();
    }

    /**
     * Determine whether or not two objects of this type are equal
     */
    public boolean equals(Object obj) {
	if ((obj instanceof KeepAliveKey) == false)
	    return false;
	KeepAliveKey kae = (KeepAliveKey)obj;
	return host.equals(kae.host)
	    && (port == kae.port)
	    && protocol.equals(kae.protocol);
    }

    /**
     * The hashCode() for this object is the string hashCode() of
     * concatenation of the protocol, host name and port.
     */
    public int hashCode() {
	String str = protocol+host+port;
	return str.hashCode();
    }
}

