/*
 * @(#)Handler.java	1.38 98/07/22
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.net.www.protocol.file;

import java.net.InetAddress;
import java.net.URLConnection;
import java.net.URL;
import java.net.MalformedURLException;
import java.net.URLStreamHandler;
import java.io.InputStream;
import java.io.IOException;
import sun.net.www.ParseUtil;
import java.io.File;

/**
 * Open an file input stream given a URL.
 * @author	James Gosling
 * @author	Benjamin Renaud
 * @author	Michael McCloskey
 * @version 	1.38, 98/07/22
 */
public class Handler extends URLStreamHandler {

    static final int ABSOLUTE_SPEC = -1;
    static final int ABSOLUTE_PATH = 0;
    static final int RELATIVE_PATH = 1;

    private String getHost(URL url) {
	String host = url.getHost();
	if (host == null) {
	    host = "";
	}
	return host;
    }

    private int detectSpecType(String spec, int start) {
        int result = -1;
        if (spec.substring(0, start).equalsIgnoreCase("file:")) {
            result = ABSOLUTE_SPEC;
	} else if ((spec.charAt(start) == '/') || (spec.indexOf(":/") == 1)) {
            result = ABSOLUTE_PATH;
	}  else {
            result = RELATIVE_PATH;
	}
        return result;
    }

    private void parseRelativePath(URL url, String spec) {
	String file = url.getFile();
        file = normalize(file, 0, file.length(), ABSOLUTE_SPEC);
	if (!file.endsWith("/")) {
	    int lastSlash = file.lastIndexOf('/');
	    if (lastSlash != -1) {
		file = file.substring(0, lastSlash+1);
	    } else {
		file = "";
	    }
	}
        file = new String(file + spec);
        ParseUtil canonizer = new ParseUtil();
        file = canonizer.canonizeString(file);
	setURL(url, "file", getHost(url), -1, file, url.getRef());
    }

    private void parseAbsoluteSpec(URL url, String spec) {
	String host = "";
	String file = null;
	int len = spec.length();
	int startFile = 0;
        int nextSlash = 0;
	if ((len > 4) && spec.startsWith("//")) {
	    // we have a host
	    nextSlash = spec.indexOf('/', 2);
	    if (nextSlash == -1) {
		// this means the url looks like this: ftp://mymachine.eng
		// == to ftp://mymachine.eng/
		nextSlash = spec.length();
	    }
	    host = spec.substring(2, nextSlash);
	}
	// the URL constructor (which calls us) figured out if there
	// was a ref, and the url object has it set.
	// so we just need file
	file = spec.substring(nextSlash);
        ParseUtil canonizer = new ParseUtil();
        file = canonizer.canonizeString(file);
	setURL(url, "file", host, -1, file, url.getRef());
    }

    private void parseAbsolutePath(URL url, String spec) {
        ParseUtil canonizer = new ParseUtil();
        spec = canonizer.canonizeString(spec);
	setURL(url, "file", getHost(url), -1, spec, url.getRef());
    }

    /**
     * Turns all '\' and other File.separators to '/'
     * Removes all multiple occurence of '/', except the first
     * double-slash for an absolute spec.
     */
    private String normalize(String spec, int start, int limit, int specType) {
	spec = spec.substring(start, limit);
	StringBuffer subSpec = new StringBuffer(spec);
        int index = 0;
        if (File.separatorChar != '/') {
            index = spec.indexOf(File.separatorChar);
	    while(index != -1) {
                subSpec.setCharAt(index, '/');
                index = spec.indexOf(File.separatorChar, index+1);
            }
            spec = subSpec.toString();
	}

	index = spec.indexOf("//");
	if (index == 0) {
	    // special case: "///"
	    if (spec.length() > 2 && spec.charAt(2) == '/') {
		subSpec.delete(0, 2);
		spec = subSpec.toString();
		index = spec.indexOf("//");
	    } else {
		// host spec?
		if (specType == ABSOLUTE_SPEC)
		    index = spec.indexOf("//", 2);
            }
        }

        // remove multiple occurences of /
        // Slows down linearly with the number of surplus slashes
        while (index != -1) {
            subSpec.deleteCharAt(index);
            spec = subSpec.toString();
            index = spec.indexOf("//", index);
        }
        return spec;
    }

    protected void parseURL(URL u, String spec, int start, int limit) {
	int specType = detectSpecType(spec, start);
	String normalizedSpec = normalize(spec, start, limit, specType);

	switch (specType) {
	case ABSOLUTE_SPEC:
	    parseAbsoluteSpec(u, normalizedSpec);
	    break;
	case ABSOLUTE_PATH:
	    parseAbsolutePath(u, normalizedSpec);
	    break;
	case RELATIVE_PATH:
            parseRelativePath(u, normalizedSpec);
	    break;
	default:
	    throw new InternalError("invalid file type");
	}
    }

    public synchronized URLConnection openConnection(URL u)
           throws IOException {
	String host = u.getHost();
	if (host == null || host.equals("") || host.equals("~") ||
	    host.equals("localhost")) {
	    return new FileURLConnection(u);
	}

	/* If you reach here, it implies that you have a hostname
	   so attempt an ftp connection.
	 */
	URLConnection uc;
	URL ru;

	try {
	    ru = new URL("ftp", host, u.getFile() +
                             (u.getRef() == null ? "": "#" + u.getRef()));
	    uc = ru.openConnection();
	} catch (IOException e) {
	    uc = null;
	}
	if (uc == null) {
	    throw new IOException("Unable to connect to: " +
                                                       u.toExternalForm());
	}
	return uc;
    }
}
