/*
 * @(#)AuthenticationInfo.java	1.15 98/09/21
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.net.www.protocol.http;

import java.net.URL;
import java.util.Hashtable;

import sun.net.www.HeaderParser;


/**
 * AuthenticationInfo: Encapsulate the information needed to
 * authenticate a user to a server.
 *
 * @author Jon Payne
 * @author Herb Jellinek
 * @author Bill Foote
 */
// REMIND:  It would be nice if this class understood about partial matching.
//	If you're authorized for foo.com, chances are high you're also
//	authorized for baz.foo.com.
// NB:  When this gets implemented, be careful about the uncaching
//	policy in HttpURLConnection.  A failure on baz.foo.com shouldn't
//	uncache foo.com!
abstract
class AuthenticationInfo {

    // Constants saying what kind of authroization this is.  This determines
    // the namespace in the hash table lookup.
    static final char SERVER_AUTHENTICATION = 's';
    static final char PROXY_AUTHENTICATION = 'p';

    /**
     * Caches authentication info entered by user.  See cacheKey()
     */
    static private Hashtable cache = new Hashtable();
    // REMIND:  This cache just grows forever.  We should put in a bounded
    //		cache, or maybe something using WeakRef's.

    /** The type of authentication this is.  Used for key lookup */
    char type;

    /** The host we're authenticating against. */
    String host;

    /** The port on the host we're authenticating against. */
    int port;

    /** The realm we're authenticating against. */
    String realm;

    AuthenticationInfo(char type, String host, int port, String realm) {
	this.type = type;
	this.host = host;
	this.port = port;
	this.realm = realm;
    }

    /**
     * Returns info for the URL, for an HTTP server auth.  Used when we 
     * don't yet know the realm
     * (i.e. when we're preemptively setting the auth).
     */
    static AuthenticationInfo getServerAuth(URL url) {
	String key = SERVER_AUTHENTICATION + url.getHost() 
		     + ":" + url.getPort();
	AuthenticationInfo result = (AuthenticationInfo) cache.get(key);
	return result;
    }
    
    /**
     * Returns info for the URL, for an HTTP server auth.  Used when we 
     * do know the realm (i.e. when we're responding to a challenge)
     */
    static AuthenticationInfo getServerAuth(URL url, String realm) {
	String key = SERVER_AUTHENTICATION + url.getHost() 
		     + ":" + url.getPort() + ":" + realm;
	AuthenticationInfo result = (AuthenticationInfo) cache.get(key);
	return result;
    }

    /**
     * Returns a firewall authentication, for the given host/port.  Used
     * for preemptive header-setting.
     */
    static AuthenticationInfo getProxyAuth(String host, int port) {
	String key = PROXY_AUTHENTICATION + host + ":" + port;
	AuthenticationInfo result = (AuthenticationInfo) cache.get(key);
	return result;
    }

    /**
     * Returns a firewall authentication, for the given host/port and realm.
     * Used in response to a challenge.
     */
    static AuthenticationInfo getProxyAuth(String host, int port, String realm) {
	String key = PROXY_AUTHENTICATION + host + ":" + port + ":" + realm;
	AuthenticationInfo result = (AuthenticationInfo) cache.get(key);
	return result;
    }


    /**
     * Add this authentication to the cache
     */
    void addToCache() {
	cache.put(cacheKey(true), this);
	if (supportsPreemptiveAuthorization()) {
	    cache.put(cacheKey(false), this);
	}
    }

    /**
     * Remove this authentication from the cache
     */
    void removeFromCache() {
	cache.remove(cacheKey(true));
	if (supportsPreemptiveAuthorization()) {
	    cache.remove(cacheKey(false));
	}
    }

    /**
     * @return true if this authentication supports preemptive authorization
     */
    abstract boolean supportsPreemptiveAuthorization();

    /**
     * @return the name of the HTTP header this authentication wants set.
     *		This is used for preemptive authorization.
     */
    abstract String getHeaderName();

    /**
     * @return the value of the HTTP header this authentication wants set.
     *		Used for preemptive authorization.
     */
    abstract String getHeaderValue();

    /**
     * Set header(s) on the given connection.  Subclasses should override to
     * set more than one header.  This will only be called for
     * definitive (i.e. non-preemptive) authorization.
     * @param conn The connection to apply the header(s) to
     * @param p A source of header values for this connection, if needed.
     * @return true if all goes well, false if no headers were set.
     */
    boolean setHeaders(HttpURLConnection conn, HeaderParser p) {
	conn.setAuthenticationProperty(getHeaderName(), getHeaderValue());
	return true;
    }

    /** 
     * Give a key for hash table lookups.
     * @param includeRealm if you want the realm considered.  Preemptively
     *		setting an authorization is done before the realm is known.
     */
    String cacheKey(boolean includeRealm) {
     	// This must be kept in sync with the getXXXAuth() methods in this
	// class.
	if (includeRealm) {
	    return type + host + ":" + port + ":" + realm;
	} else {
	    return type + host + ":" + port;
	}
    }


}
