/*
 * @(#)DigestAuthentication.java	1.6 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.net.www.protocol.http;

import java.net.URL;
import java.net.PasswordAuthentication;
import java.util.Arrays;

import sun.net.www.HeaderParser;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;


/**
 * DigestAuthentication: Encapsulate an http server authentication using
 * the "Digest" scheme, as described in RFC2069.
 *
 * @author Bill Foote
 */

// REMIND:  This is a simple implementation of the basics of digest
//	authentication.  It's not as complete as it could be.  Here
//	are some features that it would be nice to add:
//	    *  Support for the opaque field (see /home/internet/rfc/rfc2069)
//	    *  Support for the nextnonce field, which would enable preemptive
//	       authorization

class DigestAuthentication extends AuthenticationInfo {

    private PasswordAuthentication pw;

    private URL url;

    private String authMethod;


    /**
     * Create a DigestAuthentication
     */
    public DigestAuthentication(URL url, String realm, String authMethod,
				PasswordAuthentication pw) {
	super(SERVER_AUTHENTICATION, url.getHost(), url.getPort(), realm);
	this.authMethod = authMethod;
	this.url = url;
	this.pw = pw;
    }


    /**
     * @return true if this authentication supports preemptive authorization
     */
    boolean supportsPreemptiveAuthorization() {
	return false;
    }

    /**
     * Not used for this method
     * @return the name of the HTTP header this authentication wants set
     */
    String getHeaderName() {
	return "";
    }

    /**
     * Not used for this method
     * @return the value of the HTTP header this authentication wants set
     */
    String getHeaderValue() {
	return "";
    }

    /**
     * Set header(s) on the given connection.
     * @param conn The connection to apply the header(s) to
     * @param p A source of header values for this connection, if needed.
     * @return true if all goes well, false if no headers were set.
     */
    boolean setHeaders(HttpURLConnection conn, HeaderParser p) {
	String nonce = p.findValue("nonce");
	String uri = url.getFile();
	if (nonce == null || authMethod == null || pw == null || realm == null) {
	    return false;
	}
	if (authMethod.length() >= 1) {
	    // Method seems to get converted to all lower case elsewhere.
	    // It really does need to start with an upper case letter
	    // here.
	    authMethod = Character.toUpperCase(authMethod.charAt(0))
			+ authMethod.substring(1).toLowerCase();
	}
	String algorithm = p.findValue("algorithm");
	if (algorithm == null || "".equals(algorithm)) {
	    algorithm = "MD5";	// The default, accoriding to rfc2069
	}

	String response;
	char[] passwd = pw.getPassword();
	try {
	    response = computeDigest(pw.getUserName(), passwd,
				     realm, conn.getMethod(), uri, nonce,
				     algorithm);
	    if (passwd != null) Arrays.fill(passwd, ' ');
	    passwd = null;
	} catch (NoSuchAlgorithmException ex) {
	    return false;
	} finally {
	    if (passwd != null) Arrays.fill(passwd, ' ');
	}

	String value = authMethod
			+ " username=\"" + pw.getUserName()
			+ "\", realm=\"" + realm
			+ "\", nonce=\"" + nonce
			+ "\", uri=\"" + uri
			+ "\", response=\"" + response
			+ "\"";

	conn.setAuthenticationProperty("Authorization", value);
	return true;
    }

    private String computeDigest(
			String userName,
			char[] password,
			String realm,
			String connMethod,
			String requestURI,
			String nonceString,
			String algorithm
		    ) throws NoSuchAlgorithmException
    {

	MessageDigest md = MessageDigest.getInstance(algorithm);

	String A1 = userName + ":" + realm + ":";
	String HashA1 = encode(A1, password, md);

	String A2 = connMethod + ":" + requestURI;
	String HashA2 = encode(A2, null, md);

	String combo = HashA1 + ":" +
	               nonceString + ":" +
		       HashA2;
	String finalHash = encode(combo, null, md);
	return finalHash;
    }

    private final static char charArray[] = {
        '0', '1', '2', '3', '4', '5', '6', '7',
        '8', '9', 'a', 'b', 'c', 'd', 'e', 'f'
    };

    private String encode(String src, char[] passwd, MessageDigest md) {
	md.update(src.getBytes());
	if (passwd != null) {
	    byte[] passwdBytes = new byte[passwd.length];
	    for (int i=0; i<passwd.length; i++)
		passwdBytes[i] = (byte)passwd[i];
	    md.update(passwdBytes);
	    Arrays.fill(passwdBytes, (byte)0x00);
	}
	byte[] digest = md.digest();

	StringBuffer res = new StringBuffer(digest.length * 2);
	for (int i = 0; i < digest.length; i++) {
	    int hashchar = ((digest[i] >>> 4) & 0xf);
	    res.append(charArray[hashchar]);
            hashchar = (digest[i] & 0xf);
	    res.append(charArray[hashchar]);
	}
	return res.toString();
    }

}
