/*
 * @(#)JarFileFactory.java	1.22 98/10/09
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.net.www.protocol.jar;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.jar.*;
import java.security.Permission;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.security.PrivilegedExceptionAction;
import java.security.PrivilegedActionException;

/* A factory for cached JAR file. This class is used to both retrieve
 * and cache Jar files.
 *
 * @author Benjamin Renaud
 * @since JDK1.2
 */
class JarFileFactory {

    /* the cache */
    private static Map cache = new Hashtable();

    private static int BUF_SIZE = 2048;

    /* the permissions for the JAR files retrieved. */
    private HashMap permissions = new HashMap();

    /* the connections for the JAR files retrieved. */
    private HashMap connections = new HashMap();

    private boolean isFileURL(URL url) {
	return url.getProtocol().equalsIgnoreCase("file");
    }

    public URLConnection getConnection(JarFile jarFile) {
	return (URLConnection)connections.get(jarFile);
    }

    public void cacheConnection(JarFile jarFile, URLConnection connection) {
	connections.put(jarFile, connection);
    }

    public Permission getPermission(JarFile jarFile) {
        return (Permission)permissions.get(jarFile);
    }

    public void cachePermission(JarFile jarFile, Permission permission) {
	permissions.put(jarFile, permission);
    }

    /* this is a mess, and a hack for beta 3. Clean up soon */
    private HashMap fileCache = new HashMap();
    
    public synchronized JarFile get(URL url) throws IOException {

	JarFile result = null;
	if (isFileURL(url)) {
	    /* this is a mess and a hack for beta 3. Clean up soon */
	    result = (JarFile)fileCache.get(url);
	    if (result == null) {
		result = new JarFile(url.getFile());
		fileCache.put(url, result);
	    }
	    URLConnection urlc = url.openConnection();
	    Permission perms = urlc.getPermission();
	    cachePermission(result, perms);
	    cacheConnection(result, urlc);
	} else {
	    result = getCachedJarFile(url);
	    if (result == null) {
		result = retrieve(url);
		cacheJarFile(url, result);
	    }  
	    if (result == null) {
		throw new FileNotFoundException(url.toString());
	    }
	}
	return result;
    }

    void cacheJarFile(URL url, JarFile file) {
	cache.put(url, file);
    }

    JarFile getCachedJarFile(URL url) {
	JarFile result = (JarFile)cache.get(url);

	/* if the JAR file is cached, the permission will always be there */
	if (result != null) {
	    Permission perm = getPermission(result);
	    if (perm != null) {
		SecurityManager sm = System.getSecurityManager();
		if (sm != null) {
		    try {
			sm.checkPermission(perm);
		    } catch (SecurityException se) {
			// fallback to checkRead/checkConnect for pre 1.2
			// security managers
			if ((perm instanceof java.io.FilePermission) &&
			    perm.getActions().indexOf("read") != -1) {
			    sm.checkRead(perm.getName());
			} else if ((perm instanceof 
			    java.net.SocketPermission) &&
			    perm.getActions().indexOf("connect") != -1) {
			    sm.checkConnect(url.getHost(), url.getPort());
			} else {
			    throw se;
			}
		    }
		}
	    }
	}
	return result;
    }

    /** 
     * Given a URL, retrieves a JAR file, caches it to disk, and creates a
     * cached JAR file object. 
     */
    public JarFile retrieve(URL url) throws IOException {
	JarFile result = null;
	final URLConnection connection =  url.openConnection();
	Permission permission = null;

	/* get the permission after the connection is established */
	permission = connection.getPermission();
	int jarLength = connection.getContentLength();

	try { 
	    result = (JarFile)
		AccessController.doPrivileged(new PrivilegedExceptionAction() {
		    public Object run() throws IOException {
			InputStream in = null;
			OutputStream out = null;
			try {
			    in = connection.getInputStream();
			    File tmpFile = File.createTempFile("jar_cache", null);
			    tmpFile.deleteOnExit();
			    out  = new FileOutputStream(tmpFile);
			    int read = 0;
			    byte[] buf = new byte[BUF_SIZE];
			    while ((read = in.read(buf)) != -1) {
				out.write(buf, 0, read);
			    }
			    return new JarFile(tmpFile);
			} finally {
			    if (in != null) {
				in.close();
			    }
			    if (out != null) {
				out.close();
			    }
			}
		    }
		});
	} catch (PrivilegedActionException pae) {
	    throw (IOException) pae.getException();
	}

	if (result != null) {
	    cachePermission(result, permission);
	    cacheConnection(result, connection);
	}

	return result;
    }
}
