/*
 * @(#)LogInputStream.java	1.4 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.rmi.log;

import java.io.*;

public
class LogInputStream extends InputStream {
    private InputStream in;
    private int length;

    /**
     * Creates a log input file with the specified system dependent
     * file descriptor.
     * @param fd the system dependent file descriptor
     * @param length the total number of bytes allowed to be read
     * @exception IOException If an I/O error has occurred.
     */
    public LogInputStream(InputStream in, int length) throws IOException {
	this.in = in;
	this.length = length;
    }

    /**
     * Check for read past end of log record.
     * Note that this no longer decrements `length':
     * doneRead() does that now.
     * @param numBytes number of bytes to be read
     * @exception LogException Attempt to read past end of log.
     */
    private void checkRead(int numBytes) throws LogException {
	if (length < numBytes)
	    throw new LogException("length=" +
				   String.valueOf(length-numBytes) +
				   ", numBytes=" +
				   String.valueOf(numBytes));
    }
    
    /**
     * Updates number of bytes available to be read.
     * @param numBytes number of bytes to be read
     * @exception IndexOutOfBoundsException More bytes were read than
     * were available to be read.
     */
    private void doneRead(int numBytes) throws IndexOutOfBoundsException {
	length -= numBytes;
	if (length < 0) {
	    throw new IndexOutOfBoundsException(
		    "LogInputStream internal error: length=" +
		    String.valueOf(length) +
		    ", numBytes=" +
		    String.valueOf(numBytes));
	}
    }
    
    /**
     * Reads a byte of data. This method will block if no input is 
     * available.
     * @return 	the byte read, or -1 if the end of the
     *		stream is reached.
     * @exception LogException Attempt to read past end of log.
     * @exception IOException If an I/O error has occurred.
     */
    public int read() throws IOException, LogException {
	checkRead(1);
	int c = in.read();
	if (c != -1)
	    doneRead(1);
	return c;
    }

    /**
     * Reads data into an array of bytes.
     * This method blocks until some input is available.
     * @param b	the buffer into which the data is read
     * @return  the actual number of bytes read, -1 is
     * 		returned when the end of the stream is reached.
     * @exception LogException Attempt to read past end of log.
     * @exception IOException If an I/O error has occurred.
     */
    public int read(byte b[]) throws IOException, LogException {
	checkRead(b.length);
	int n = in.read(b);
	if (n > 0)
	    doneRead(n);
	return n;
    }

    /**
     * Reads data into an array of bytes.
     * This method blocks until some input is available.
     * @param b	the buffer into which the data is read
     * @param off the start offset of the data
     * @param len the maximum number of bytes read
     * @return  the actual number of bytes read, -1 is
     * 		returned when the end of the stream is reached.
     * @exception LogException Attempt to read past end of log.
     * @exception IOException If an I/O error has occurred.
     */
    public int read(byte b[], int off, int len)
		throws IOException, LogException {
	checkRead(len);
	int n = in.read(b, off, len);
	if (n > 0)
	    doneRead(n);
	return n;
    }

    /**
     * Skips n bytes of input.
     * @param n the number of bytes to be skipped
     * @return	the actual number of bytes skipped.
     * @exception LogException Attempt to read past end of log.
     * @exception IOException If an I/O error has occurred.
     */
    public long skip(long n) throws IOException, LogException {
	if (n > Integer.MAX_VALUE)
	    throw new IOException("Too many bytes to skip - " + n);
	checkRead((int)n);
	long m = in.skip(n);
	if (m > 0)
	    doneRead((int)m);
	return m;
    }
    
    /**
     * Returns the number of bytes that can be read
     * without blocking.
     * @return the number of available bytes, which is initially
     *		equal to the file size.
     */
    public int available() {
	return length;
    }

    /**
     * Closes the input stream.  No further input can be read.
     * the stream.
     */
    public void close() {
	length = 0;
    }

    /**
     * Closes the stream when garbage is collected.
     */
    protected void finalize() throws IOException {
	close();
    }

}
