/*
 * @(#)Main.java	1.22 98/11/03
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.rmi.rmic;

import sun.tools.java.*;

import java.util.*;
import java.io.*;

import sun.tools.javac.SourceClass;

/**
 * Main "rmic" program.
 */

public class Main implements Constants {
    String classPathString;
    File destDir;
    int flags;
    long tm;
    Vector v;
    boolean nowrite;
    String props;
    boolean keepGenerated;
    int stubVersion;

    /**
     * Name of the program.
     */
    String program;

    /**
     * The stream where error message are printed.
     */
    OutputStream out;

    /**
     * Constructor.
     */
    public Main(OutputStream out, String program) {
	this.out = out;
	this.program = program;
    }

    /**
     * Output a message.
     */
    public void output(String msg) {
	PrintStream out =
	    this.out instanceof PrintStream ? (PrintStream)this.out
					    : new PrintStream(this.out, true);
	out.println(msg);
    }

    /**
     * Top level error message.  This method is called when the
     * environment could not be set up yet.
     */
    private void error(String msg) {
	output(getText(msg));
    }

    private void error(String msg, String arg1) {
	output(getText(msg, arg1));
    }

    private void error(String msg, String arg1, String arg2) {
	output(getText(msg, arg1, arg2));
    }

    /**
     * Usage
     */
    public void usage() {
	error("rmic.usage", program);
    }
    
    /**
     * Run the compiler
     */
    public synchronized boolean compile(String argv[]) {
	if (!parseArgs(argv)) {
	    return false;
	}

	if (v.size() == 0) {
	    usage();
	    return false;
	}

	return doCompile();
    }
    /*
     * Parse the arguments for compile.
     */
    public boolean parseArgs(String argv[]) {
	destDir = null;
	
	flags = F_WARNINGS;
	tm = System.currentTimeMillis();
	v = new Vector();
	nowrite = false;
	props = null;
	keepGenerated = false;
	stubVersion = STUB_VERSION_FAT;		// default stub version

	// Parse arguments
	for (int i = 0 ; i < argv.length ; i++) {
	    if (argv[i].equals("-g")) {
		flags &= ~F_OPT;
		flags |= F_DEBUG_LINES | F_DEBUG_VARS;
	    } else if (argv[i].equals("-O")) {
		flags &= ~F_DEBUG_LINES;
		flags &= ~F_DEBUG_VARS;
		flags |= F_OPT | F_DEPENDENCIES;
	    } else if (argv[i].equals("-nowarn")) {
		flags &= ~F_WARNINGS;
	    } else if (argv[i].equals("-debug")) {
		flags |= F_DUMP;
	    } else if (argv[i].equals("-depend")) {
		flags |= F_DEPENDENCIES;
	    } else if (argv[i].equals("-verbose")) {
		flags |= F_VERBOSE;
	    } else if (argv[i].equals("-nowrite")) {
		nowrite = true;
	    } else if (argv[i].equals("-keep") ||
		       argv[i].equals("-keepgenerated")) {
		keepGenerated = true;
	    } else if (argv[i].equals("-show")) {
		error("rmic.option.unsupported", "-show");
		usage();
		return false;
	    } else if (argv[i].equals("-classpath")) {
		if ((i + 1) < argv.length) {
		    if (classPathString != null) {
			error("rmic.option.already.seen", "-classpath");
			usage();
			return false;
		    }
		    classPathString = argv[++i];
		} else {
		    error("rmic.option.requires.argument", "-classpath");
		    usage();
		    return false;
		}
	    } else if (argv[i].equals("-d")) {
		if ((i + 1) < argv.length) {
		    if (destDir != null) {
			error("rmic.option.already.seen", "-d");
			usage();
			return false;
		    }
		    destDir = new File(argv[++i]);
		    if (!destDir.exists()) {
			error("rmic.no.such.directory", destDir.getPath());
			usage();
			return false;
		    }
		} else {
		    error("rmic.option.requires.argument", "-d");
		    usage();
		    return false;
		}
	    } else if (argv[i].equals("-v1.1")) {
		stubVersion = STUB_VERSION_1_1;
	    } else if (argv[i].equals("-vcompat")) {
		stubVersion = STUB_VERSION_FAT;
	    } else if (argv[i].equals("-v1.2")) {
		stubVersion = STUB_VERSION_1_2;
	    } else if (argv[i].startsWith("-")) {
		error("rmic.no.such.option", argv[i]);
		usage();
		return false;
	    } else {
		v.addElement(argv[i]);
	    }
	}
	return true;
    }

    /**
     * Do the compile with the switches and files already supplied
     */
    public boolean doCompile() {
	// Create batch environment
	ClassPath classPath;
	if (classPathString != null) {
	    classPath = new ClassPath(classPathString);
	} else {
	    classPath = new ClassPath();
	}
	BatchEnvironment env = new BatchEnvironment(out, classPath);
	env.flags |= flags;

	// Set the classfile version numbers
	// Compat and 1.1 stubs must retain the old version number.
	switch (stubVersion) {
	case STUB_VERSION_1_1:
	case STUB_VERSION_FAT:
	    // numbers as seen in JVM spec
	    // (actually, as seen in JDK1.1's sun.tools.java.BinaryClass)
	    env.majorVersion = 45;
	    env.minorVersion = 3;
	    break;
	default:
	    env.majorVersion = 45;
	    env.minorVersion = 3;
	}
	
	// Preload the "out of memory" error string just in case we run
	// out of memory during the compile.
	String noMemoryErrorString = getText("rmic.no.memory");
	String stackOverflowErrorString = getText("rmic.stack.overflow");

	try {
	    /** Load the classes on the command line
	     * Replace the entries in v with the ClassDefinition for the class
	     */
	    for (int i = v.size()-1; i >= 0; i-- ) {
		Identifier implClassName =
		    Identifier.lookup((String)v.elementAt(i));

		/*
		 * Fix bugid 4049354: support using '.' as an inner class
		 * qualifier on the command line (previously, only mangled
		 * inner class names were understood, like "pkg.Outer$Inner").
		 *
		 * The following method, also used by "javap", resolves the
		 * given unmangled inner class name to the appropriate
		 * internal identifier.  For example, it translates
		 * "pkg.Outer.Inner" to "pkg.Outer. Inner".
		 */
		implClassName = env.resolvePackageQualifiedName(implClassName);
		/*
		 * But if we use such an internal inner class name identifier
		 * to load the class definition, the Java compiler will notice
		 * if the impl class is a "private" inner class and then deny
		 * skeletons (needed unless "-v1.2" is used) the ability to
		 * cast to it.  To work around this problem, we mangle inner
		 * class name identifiers to their binary "outer" class name:
		 * "pkg.Outer. Inner" becomes "pkg.Outer$Inner".
		 */
		implClassName = Names.mangleClass(implClassName);

		ClassDeclaration decl = env.getClassDeclaration(implClassName);
		try {
		    ClassDefinition def = decl.getClassDefinition(env);
		    Generator.generate(env, def, destDir, stubVersion);
		} catch (ClassNotFound ex) {
		    env.error(0, "rmic.class.not.found", implClassName);
		}

	    }

	    // compile all classes that need compilation
	    ByteArrayOutputStream buf = new ByteArrayOutputStream(4096);
	    boolean done;

	    do {
		done = true;
		env.flushErrors();
		for (Enumeration e = env.getClasses() ; e.hasMoreElements() ; ) {
		    ClassDeclaration c = (ClassDeclaration)e.nextElement();
		    SourceClass src;

		    switch (c.getStatus()) {
		      case CS_UNDEFINED:
			if (!env.dependencies()) {
			    break;
			}
			// fall through

		      case CS_SOURCE:
			done = false;
			env.loadDefinition(c);
			if (c.getStatus() != CS_PARSED) {
			    break;
			}
			// fall through
			
		      case CS_PARSED:
			if (c.getClassDefinition().isInsideLocal()) {
			    // the enclosing block will check this one
			    continue;
			}
			done = false;

			src = (SourceClass)c.getClassDefinition(env);
			src.check(env);
			c.setDefinition(src, CS_CHECKED);
			// fall through

		      case CS_CHECKED:
			src = (SourceClass)c.getClassDefinition(env);
			// bail out if there were any errors
			if (src.getError()) {
			    c.setDefinition(src, CS_COMPILED);
			    break;
			}
			done = false;
			buf.reset();
			src.compile(buf);
			c.setDefinition(src, CS_COMPILED);
			src.cleanup(env);

			if (src.getError() || nowrite) {
			    continue;
			}

			String pkgName = c.getName().getQualifier().toString().replace('.', File.separatorChar);
			String className = c.getName().getFlatName().toString().replace('.', SIGC_INNERCLASS) + ".class";

			File file;
			if (destDir != null) {
			    if (pkgName.length() > 0) {
				file = new File(destDir, pkgName);
				if (!file.exists()) {
				    file.mkdirs();
				}
				file = new File(file, className);
			    } else {
				file = new File(destDir, className);
			    }
			} else {
			    ClassFile classfile = (ClassFile)src.getSource();
			    if (classfile.isZipped()) {
				env.error(0, "cant.write", classfile.getPath());
				continue;
			    }
			    file = new File(classfile.getPath());
			    file = new File(file.getParent(), className);
			}

			// Create the file
			try {
			    FileOutputStream out = new FileOutputStream(file.getPath());
			    buf.writeTo(out);
			    out.close();
			    if (env.verbose()) {
				output(getText("rmic.wrote", file.getPath()));
			    }
			} catch (IOException ee) {
			    env.error(0, "cant.write", file.getPath());
			}
		    }
		}
	    } while (!done);
	} catch (OutOfMemoryError ee) {
	    // The compiler has run out of memory.  Use the error string
	    // which we preloaded.
	    env.output(noMemoryErrorString);
	    return false;
	} catch (StackOverflowError ee) {
	    env.output(stackOverflowErrorString);
	    return false;
	} catch (Error ee) {
	    // We allow the compiler to take an exception silently if a program
	    // error has previously been detected.  Presumably, this makes the
	    // compiler more robust in the face of bad error recovery.
	    if (env.nerrors == 0 || env.dump()) {
		ee.printStackTrace();
		env.error(0, "fatal.error");
	    }
	} catch (Exception ee) {
	    if (env.nerrors == 0 || env.dump()) {
		ee.printStackTrace();
		env.error(0, "fatal.exception");
	    }
	}

	env.flushErrors();
	env.shutdown();

	boolean status = true;
	if (env.nerrors > 0) {
	    String msg = "";
	    if (env.nerrors > 1) {
		msg = getText("rmic.errors", env.nerrors);
	    } else {
		msg = getText("rmic.1error");
	    }
	    if (env.nwarnings > 0) {
		if (env.nwarnings > 1) {
		    msg += ", " + getText("rmic.warnings", env.nwarnings);
		} else {
		    msg += ", " + getText("rmic.1warning");
		}
	    }
	    output(msg);
	    status = false;
	} else {
	    if (env.nwarnings > 0) {
		if (env.nwarnings > 1) {
		    output(getText("rmic.warnings", env.nwarnings));
		} else {
		    output(getText("rmic.1warning"));
		}
	    }
	}

	// last step is to delete generated source files
	if (!keepGenerated) {
	    env.deleteGeneratedFiles();
	}

	// We're done
	if (env.verbose()) {
	    tm = System.currentTimeMillis() - tm;
	    output(getText("rmic.done_in", Long.toString(tm)));
	}
	
	return status;
    }

    /**
     * Main program
     */
    public static void main(String argv[]) {
	Main compiler = new Main(System.out, "rmic");
	System.exit(compiler.compile(argv) ? 0 : 1);
    }

    /**
     * Return the string value of a named resource in the rmic.properties
     * resource bundle.  If the resource is not found, null is returned.
     */
    public static String getString(String key) {
	if (!resourcesInitialized) {
	    initResources();
	}

	try {
	    return resources.getString(key);
	} catch (MissingResourceException ignore) {
	}
	return null;
    }

    private static boolean resourcesInitialized = false;
    private static ResourceBundle resources;

    private static void initResources() {
	try {
	    resources =
		ResourceBundle.getBundle("sun.rmi.rmic.resources.rmic");
	    resourcesInitialized = true;
	} catch (MissingResourceException e) {
	    throw new Error("fatal: missing resource bundle: " +
		e.getClassName());
	}
    }

    public static String getText(String key) {
	String message = getString(key);
	if (message == null) {
	    message = "no text found: \"" + key + "\"";
	}
	return message;
    }

    public static String getText(String key, int num) {
	return getText(key, Integer.toString(num), null, null);
    }

    public static String getText(String key, String arg0) {
	return getText(key, arg0, null, null);
    }

    public static String getText(String key, String arg0, String arg1) {
	return getText(key, arg0, arg1, null);
    }

    public static String getText(String key,
				 String arg0, String arg1, String arg2)
    {
	String format = getString(key);
	if (format == null) {
	    format = "no text found: " +
		"key = \"{0}\", arguments = \"{1}\", \"{2}\", \"{3}\"";
	}

	String[] args = new String[3];
	args[0] = (arg0 != null ? arg0.toString() : "null");
	args[1] = (arg1 != null ? arg1.toString() : "null");
	args[2] = (arg2 != null ? arg2.toString() : "null");

	return java.text.MessageFormat.format(format, args);
    }
}
