/*
 * @(#)LoaderHandler.java	1.21 98/07/31
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.rmi.server;

import java.lang.ref.*;
import java.io.*;
import java.net.*;
import java.security.*;
import java.util.*;
import java.rmi.server.LogStream;

import sun.security.action.GetPropertyAction;

/**
 * <code>LoaderHandler</code> provides the implementation of the static
 * methods of the <code>java.rmi.server.RMIClassLoader</code> class.
 *
 * @version 1.21, 07/31/98
 * @author  Ann Wollrath
 * @author  Peter Jones
 */
public final class LoaderHandler {

    /** RMI class loader log level */
    final static int logLevel = LogStream.parseLevel(
	(String) java.security.AccessController.doPrivileged(
            new GetPropertyAction("sun.rmi.loader.logLevel")));

    /**
     * value of "java.rmi.server.codebase" property, as cached at class
     * initialization time.  It may contain malformed URLs.
     */
    private static String codebase = null;
    static {
	String prop = (String) java.security.AccessController.doPrivileged(
            new GetPropertyAction("java.rmi.server.codebase"));
	if (prop != null && prop.trim().length() > 0) {
	    codebase = prop;
	}
    }

    /** list of URLs represented by the codebase property, if valid */
    private static URL[] codebaseURLs = null;

    /**
     * table mapping codebase URL path and context class loader pairs
     * to class loader instances.  Entries hold class loaders with weak
     * references, so this table does not prevent loaders from being
     * garbage collected.
     */
    private static HashMap loaderTable = new HashMap(5);

    /** references queue for cleared class loader entries */
    private static ReferenceQueue refQueue = new ReferenceQueue();

    /*
     * Disallow anyone from creating one of these.
     */
    private LoaderHandler() {}

    /**
     * Load a class from the codebase URL path specified by the
     * <code>java.rmi.server.codebase</code> property.
     */
    public static Class loadClass(String name)
	throws MalformedURLException, ClassNotFoundException
    {
	/*
	 * If it hasn't already been done, convert the codebase property
	 * into an array of URLs; this may throw a MalformedURLException.
	 */
	synchronized (LoaderHandler.class) {
	    if (codebaseURLs == null) {
		if (codebase != null) {
		    codebaseURLs = pathToURLs(codebase);
		} else {
		    codebaseURLs = new URL[0];
		}
	    }
	}

	return loadClass(codebaseURLs, name, getRMIContextClassLoader());
    }

    /**
     * Load a class from a URL.
     */
    public static Class loadClass(URL codebase, String name)
	throws MalformedURLException, ClassNotFoundException
    {
	if (codebase != null) {
	    return loadClass(new URL[] { codebase }, name);
	} else {
	    return loadClass(new URL[0], name);
	}
    }

    /**
     * Load a class from a network location (one or more URLs).
     */
    public static Class loadClass(String codebase, String name)
	throws MalformedURLException, ClassNotFoundException
    {
	if (codebase != null) {
	    return loadClass(pathToURLs(codebase), name);
	} else {
	    return loadClass(new URL[0], name);
	}
    }

    /**
     * Returns the class annotation (representing the location for
     * a class) that RMI will use to annotate the call stream when
     * marshalling objects of the given class.
     */
    public static String getClassAnnotation(Class cl) {
	String name = cl.getName();

	/*
	 * Class objects for arrays of primitive types never need an
	 * annotation, because they never need to be (or can be) downloaded.
	 *
	 * REMIND: should we (not) be annotating classes that are in
	 * "java.*" packages?
	 */
	if (name.length() >= 2 &&
	    name.charAt(0) == '[' && name.charAt(1) != 'L')
	{
	    return null;	// REMIND: are we allowed to return null?
	}

	/*
	 * Get the class's class loader.  If it is null, the system class
	 * loader, an ancestor of the base class loader (such as the loader
	 * for installed extensions), return the value of the
	 * "java.rmi.server.codebase" property.
	 */
	ClassLoader loader = cl.getClassLoader();
	if (loader == null) {
	    return codebase;
	}
	for (ClassLoader codebaseLoader = ClassLoader.getSystemClassLoader();
	     codebaseLoader != null;
	     codebaseLoader = codebaseLoader.getParent())
	{
	    if (loader == codebaseLoader) {
		return codebase;
	    }
	}

        /*
	 * Get the codebase URL path for the class loader, if it supports
	 * such a notion (i.e., if it is a URLClassLoader or subclass).
	 */
	String annotation = null;
	if (loader instanceof Loader) {
	    /*
	     * If the class loader is one of our RMI class loaders, we have
	     * already computed the class annotation string, and no
	     * permissions are required to know the URLs.
	     */
	    annotation = ((Loader) loader).getClassAnnotation();

	} else if (loader instanceof URLClassLoader) {
	    try {
		URL[] urls = ((URLClassLoader) loader).getURLs();
		if (urls != null) {
		    /*
		     * If the class loader is not one of our RMI class loaders,
		     * we must verify that the current access control context
		     * has permission to know all of these URLs.
		     */
		    SecurityManager sm = System.getSecurityManager();
		    if (sm != null) {
			for (int i = 0; i < urls.length; i++) {
			    Permission p =
				urls[i].openConnection().getPermission();
			    if (p != null) {
				sm.checkPermission(p);
			    }
			}
		    }

		    annotation = urlsToPath(urls);
		}
	    } catch (SecurityException e) {
		/*
		 * If access was denied to the knowledge of the class
		 * loader's URLs, fall back to the default behavior.
		 */
	    } catch (IOException e) {
		/*
		 * This shouldn't happen, although it is declared to be
		 * thrown by openConnection() and getPermission().  If it
		 * does happen, forget about this class loader's URLs and
		 * fall back to the default behavior.
		 */
	    }
	}

	if (annotation != null) {
	    return annotation;
	} else {
	    return codebase;		// REMIND: does this make sense??
	}
    }

    /**
     * Return the security context of the given class loader.
     */
    public static Object getSecurityContext(ClassLoader loader) {
	/*
	 * REMIND: This is a bogus JDK1.1-compatible implementation.
	 * This method should never be called by application code anyway
	 * (hence the deprecation), but should it do something different
	 * and perhaps more useful, like return a String or a URL[]?
	 */
	if (loader instanceof Loader) {
	    URL[] urls = ((Loader) loader).getURLs();
	    if (urls.length > 0) {
		return urls[0];
	    }
	}
	return null;
    }

    /**
     * Load a class from the RMI class loader corresponding to the given
     * codebase URL path in the current execution context.
     */
    private static Class loadClass(URL[] urls, String name)
	throws ClassNotFoundException
    {
	return loadClass(urls, name, getRMIContextClassLoader());
    }

    /**
     * Load a class from the RMI class loader corresponding to the given
     * codebase URL path and the given parent class loader.
     */
    private static Class loadClass(URL[] urls, String name,
				   ClassLoader parent)
	throws ClassNotFoundException
    {
	/*
	 * If no security manager is set, disable access to RMI class
	 * loaders and delegate request to the parent loader.
	 */
	SecurityManager sm = System.getSecurityManager();
	if (sm == null) {
	    try {
		return Class.forName(name, false, parent);
	    } catch (ClassNotFoundException e) {
		throw new ClassNotFoundException(e.getMessage() +
		    " (no security manager: RMI class loader disabled)",
		    e.getException());
	    }
	}

	ClassLoader loader;
	try {
	    loader = lookupLoader(urls, parent);
	} catch (SecurityException e) {
	    /*
	     * If the current access control context does not have permission
	     * to access all of the URLs in the codebase path, wrap the
	     * security exception in a ClassNotFoundException, so the caller
	     * can handle this outcome just like any other class loading
	     * failure (see bugid 4146529).
	     */
	    throw new ClassNotFoundException(
		"access denied to class loader", e);
	}

	if (loader != null) {
	    if (logLevel >= LogStream.VERBOSE)
		LogStream.log("loader").println(
		    "LoaderHandler.loadClass: loading class \"" + name + "\"" +
		    " from " + Arrays.asList(urls));
	}

	return Class.forName(name, false, loader);
    }

    /**
     * Convert a string containing a space-separated list of URLs into a
     * corresponding array of URL objects, throwing a MalformedURLException
     * if any of the URLs are invalid.
     */
    private static URL[] pathToURLs(String path)
	throws MalformedURLException
    {
	synchronized (pathToURLsCache) {
	    Object[] v = (Object[]) pathToURLsCache.get(path);
	    if (v != null) {
		return ((URL[])v[0]);
	    }
	}
	StringTokenizer st = new StringTokenizer(path);	// divide by spaces
	URL[] urls = new URL[st.countTokens()];
	for (int i = 0; st.hasMoreTokens(); i++) {
	    urls[i] = new URL(st.nextToken());
	}
	synchronized (pathToURLsCache) {
	    pathToURLsCache.put(path,
				new Object[] {urls, new SoftReference(path)});
	}
	return urls;
    }

    /** map from weak(key=string) to [URL[], soft(key)] */
    private static WeakHashMap pathToURLsCache = new WeakHashMap();

    /**
     * Convert an array of URL objects into a corresponding string
     * containing a space-separated list of URLs.
     *
     * Note that if the array has zero elements, the return value is
     * null, not the empty string.
     */
    private static String urlsToPath(URL[] urls) {
	if (urls.length == 0) {
	    return null;
	} else if (urls.length == 1) {
	    return urls[0].toExternalForm();
	} else {
	    StringBuffer path = new StringBuffer(urls[0].toExternalForm());
	    for (int i = 1; i < urls.length; i++) {
		path.append(' ');
		path.append(urls[i].toExternalForm());
	    }
	    return path.toString();
	}
    }

    /**
     * Return the class loader to be used as the parent for an RMI class
     * loader used in the current execution context.
     */
    private static ClassLoader getRMIContextClassLoader() {
	/*
	 * The current implementation simply uses the current thread's
	 * context class loader.
	 */
	return Thread.currentThread().getContextClassLoader();
    }

    /**
     * Look up the RMI class loader for the given codebase URL path
     * and the given parent class loader.  A new class loader instance
     * will be created and returned if no match is found.
     *
     * The current access control context is checked to verify that it
     * has all the permissions required to load classes from all the
     * URLs in the codebase path.
     */
    private static ClassLoader lookupLoader(final URL[] urls,
					    final ClassLoader parent)
    {
	/*
	 * If the requested codebase URL path is empty, the supplied
	 * parent class loader will be sufficient.
	 *
	 * REMIND: To be conservative, this optimization is commented out
	 * for now so that it does not open a security hole in the future
	 * by providing untrusted code with direct access to the public
	 * loadClass() method of a class loader instance that it cannot
	 * get a reference to.  (It's an unlikely optimization anyway.)
	 *
	 * if (urls.length == 0) {
	 *     return parent;
	 * }
	 */

	LoaderEntry entry;
	Loader loader;

	synchronized (LoaderHandler.class) {
	    /*
	     * Take this opportunity to remove from the table entries
	     * whose weak references have been cleared.
	     */
	    while ((entry = (LoaderEntry) refQueue.poll()) != null) {
		if (!entry.removed) {	// ignore entries removed below
		    loaderTable.remove(entry.key);
		}
	    }

	    /*
	     * Look up the codebase URL path and parent class loader pair
	     * in the table of RMI class loaders.
	     */
	    LoaderKey key = new LoaderKey(urls, parent);
	    entry = (LoaderEntry) loaderTable.get(key);

	    if (entry == null || (loader = (Loader) entry.get()) == null) {
		/*
		 * If entry was in table but it's weak reference was cleared,
		 * remove it from the table and mark it as explicitly cleared,
		 * so that new matching entry that we put in the table will
		 * not be erroneously removed when this entry is processed
		 * from the weak reference queue.
		 */
		if (entry != null) {
		    loaderTable.remove(key);
		    entry.removed = true;
		}

		/*
		 * A matching loader was not found, so create a new class
		 * loader instance for the requested codebase URL path and
		 * parent class loader.  The instance is created within an
		 * access control context retricted to the permissions
		 * necessary to load classes from its codebase URL path.
		 */
		AccessControlContext acc = getLoaderAccessControlContext(urls);
		loader = (Loader) java.security.AccessController.doPrivileged(
		    new java.security.PrivilegedAction() {
			public Object run() {
			    return new Loader(urls, parent);
			}
		    }, acc);

		/*
		 * Precompute to permissions required to access the loader;
		 * store loader and permissions in table for key.
		 */
		PermissionCollection perms = new Permissions();
		addPermissionsForURLs(urls, perms, false);
		entry = new LoaderEntry(key, loader, perms);
		loaderTable.put(key, entry);
	    }
	}

	/*
	 * Check that the current access control context has all of the
	 * permissions necessary to load classes from this loader.
	 */
	SecurityManager sm = System.getSecurityManager();
	if (sm != null) {
	    PermissionCollection perms = entry.permissions;
	    Enumeration enum = perms.elements();
	    while (enum.hasMoreElements()) {
		sm.checkPermission((Permission) enum.nextElement());
	    }
	}

	return loader;
    }

    /**
     * LoaderKey holds a codebase URL path and parent class loader pair
     * used to look up RMI class loader instances in its class loader cache.
     */
    private static class LoaderKey {

	private URL[] urls;

	private ClassLoader parent;

	private int hashValue;

	public LoaderKey(URL[] urls, ClassLoader parent) {
	    this.urls = urls;
	    this.parent = parent;

	    hashValue = parent.hashCode();
	    for (int i = 0; i < urls.length; i++) {
		hashValue ^= urls[i].hashCode();
	    }
	}

	public int hashCode() {
	    return hashValue;
	}

	public boolean equals(Object obj) {
	    if (obj instanceof LoaderKey) {
		LoaderKey other = (LoaderKey) obj;
		if (parent != other.parent) {
		    return false;
		}
		if (urls == other.urls) {
		    return true;
		}
		if (urls.length != other.urls.length) {
		    return false;
		}
		for (int i = 0; i < urls.length; i++) {
		    if (!urls[i].equals(other.urls[i])) {
			return false;
		    }
		}
		return true;
	    } else {
		return false;
	    }
	}
    }

    /**
     * LoaderEntry contains a weak reference to an RMIClassLoader.  The
     * weak reference is registered with the private static "refQueue"
     * queue.  The entry contains the codebase URL path and parent class
     * loader key for the loader so that the mapping can be removed from
     * the table efficiently when the weak reference is cleared, and it
     * holds a PermissionCollection object to store the precomputed set
     * of permissions required to access the loader.
     */
    private static class LoaderEntry extends WeakReference {

	public LoaderKey key;

	public PermissionCollection permissions;

	/**
	 * set to true if the entry has been removed from the table
	 * because it has been replaced, so it should not be attempted
	 * to be removed again
	 */
	public boolean removed = false;

	public LoaderEntry(LoaderKey key, Loader loader,
			   PermissionCollection permissions)
	{
	    super(loader, refQueue);
	    this.key = key;
	    this.permissions = permissions;
	}
    }

    /**
     * Return the access control context that a loader for the given
     * codebase URL path should execute with.
     */
    private static AccessControlContext getLoaderAccessControlContext(
	URL[] urls)
    {
	/*
	 * The approach used here is taken from the similar method
	 * getAccessControlContext() in the sun.applet.AppletPanel class.
	 */
	// begin with permissions granted to all code in current policy
	PermissionCollection perms = (PermissionCollection)
	    java.security.AccessController.doPrivileged(
		new java.security.PrivilegedAction() {
		public Object run() {
		    CodeSource codesource = new CodeSource(null, null);
		    Policy p = java.security.Policy.getPolicy();
		    if (p != null) {
			return p.getPermissions(codesource);
		    } else {
			return new Permissions();
		    }
		}
	    });

	// createClassLoader permission needed to create loader in context
	perms.add(new RuntimePermission("createClassLoader"));

	// add permissions to read any "java.*" property
	perms.add(new java.util.PropertyPermission("java.*","read"));

	// add permissions reuiqred to load from codebase URL path
	addPermissionsForURLs(urls, perms, true);

	/*
	 * Create an AccessControlContext that consists of a single
	 * protection domain with only the permissions calculated above.
	 */
	ProtectionDomain pd = new ProtectionDomain(
	    new CodeSource((urls.length > 0 ? urls[0] : null), null), perms);
	return new AccessControlContext(new ProtectionDomain[] { pd });
    }

    /**
     * Add the permissions necessary to load classes from a loader with
     * the given URLs for a codebase to the given permissions object.
     */
    private static void addPermissionsForURLs(URL[] urls,
					      PermissionCollection perms,
					      boolean forLoader)
    {
	/*
	 * The approach used here was derived from the method
	 * getAccessControlContext() in the sun.applet.AppletPanel class.
	 */
	for (int i = 0; i < urls.length; i++) {
	    URL url = urls[i];

	    try {
		/*
		 * Get the permission required to connect to the particular
		 * codebase URL, according to the URL's connection object.
		 */
		Permission p = url.openConnection().getPermission();

		if (p != null) {
		    if (p instanceof FilePermission) {
			/*
			 * If the codebase is a file, the permission required
			 * to actually read classes from the codebase URL is
			 * the permission to read all files beneath the last
			 * directory in the file path, either because JAR
			 * files can refer to other JAR files in the same
			 * directory, or because permission to read a
			 * directory is not implied by permission to read the
			 * contents of a directory, which all that might be
			 * granted.
			 */
			String path = p.getName();
			int endIndex = path.lastIndexOf(File.separatorChar);
			if (endIndex != -1) {
			    path = path.substring(0, endIndex+1);
			    if (path.endsWith(File.separator)) {
				path += "-";
			    }
			    perms.add(new FilePermission(path, "read"));
			} else {
			    /*
			     * No directory separator: use permission to
			     * read the file.
			     */
			    perms.add(p);
			}
		    } else {
			/*
			 * Otherwise, assume that the permission we have is
			 * the real permission required to download classes
			 * from beneath the codebase URL.
			 */
			perms.add(p);
			/*
			 * If the purpose of these permissions is to grant
			 * them to an instance of a URLClassLoader subclass,
			 * we must add permission to connect to and accept
			 * from the host of non-"file:" URLs, otherwise the
			 * getPermissions() method of URLClassLoader will
			 * throw a security exception.
			 */
			if (forLoader) {
			    String host = url.getHost();
			    if (host != null) {
				perms.add(new SocketPermission(host,
				    "connect, accept"));
			    }
			}
		    }
		}
	    } catch (IOException e) {
		/*
		 * This shouldn't happen, although it is declared to be
		 * thrown by openConnection() and getPermission().  If it
		 * does, don't bother granting or requiring any permissions
		 * for this URL.
		 */
	    }
	}
    }

    /**
     * Loader is the actual class of the RMI class loaders created
     * by the RMIClassLoader static methods.
     */
    private static class Loader extends URLClassLoader {

	private String annotation;

	private Loader(URL[] urls, ClassLoader parent) {
	    super(urls, parent);
	    /*
	     * Caching the value of class annotation string here assumes
	     * that the protected method addURL() is never called on this
	     * class loader.
	     */
	    annotation = urlsToPath(urls);
	}

	/**
	 * Return the string to be annotated with all classes loaded from
	 * this class loader.
	 */
	public String getClassAnnotation() {
	    return annotation;
	}

	/**
	 * Return the permissions to be granted to code loaded from the
	 * given code source.
	 */
	protected PermissionCollection getPermissions(CodeSource codesource) {
	    PermissionCollection perms = super.getPermissions(codesource);
	    /*
	     * Grant the same permissions that URLClassLoader would grant.
	     */
	    return perms;
	}
    }
}
