/*
 * @(#)DGCImpl.java	1.21 98/07/15
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.rmi.transport;

import java.util.*;
import java.rmi.*;
import java.rmi.dgc.DGC;
import java.rmi.dgc.Lease;
import java.rmi.dgc.VMID;
import java.rmi.server.LogStream;
import java.rmi.server.ObjID;
import java.rmi.server.RemoteServer;
import java.rmi.server.RemoteStub;
import java.rmi.server.ServerNotActiveException;
import sun.rmi.server.Dispatcher;
import sun.rmi.server.UnicastServerRef;

/**
 * This class implements the guts of the server-side distributed GC
 * algorithm
 *
 * @author Ann Wollrath
 */
final class DGCImpl implements DGC {

    /** "dgc" log level */
    static int logLevel = LogStream.parseLevel(getLogLevel());

    private static String getLogLevel() {
	return (String) java.security.AccessController.doPrivileged(
            new sun.security.action.GetPropertyAction("sun.rmi.dgc.logLevel"));
    }

    /*
     * this should be a *private* method since it is privileged.
     * this is identical to a method in DGCClient.  it is
     * necessary for accessing properties.
     */
    private static long getLong(String name, long def) {
	return ((Long) java.security.AccessController.doPrivileged(
                new sun.security.action.GetLongAction(name, def))).longValue();
    }

    /** Lease request time for DGC; defaults to 10 minutes */
    private final static long leaseValue =
	getLong("java.rmi.dgc.leaseValue", 600000);

    /** Lease check time for DGC; defaults to five minutes */
    private final static long leaseCheckInterval =
    	getLong("sun.rmi.dgc.checkInterval", 300000);

    /** well-known ID for the remote object collector */
    private static ObjID dgcID;
    /** remote implementation of DGC interface for this VM */
    private static DGCImpl dgc;
    /** table that maps VMID to LeaseInfo */
    private Hashtable leaseTable = new Hashtable();
    /** checks for lease expiration */
    private Thread checker = null;

    /**
     * Return the remote implementation of the DGC interface for
     * this VM.
     */
    static DGCImpl getDGCImpl()
    {
	return dgc;
    }

    /**
     * Construct a new server-side remote object collector at
     * a particular port. Disallow construction from outside.
     */
    private DGCImpl() {}

    /**
     * The dirty call adds the VMID "vmid" to the set of clients
     * that hold references to the object associated with the ObjID
     * id.  The long "sequenceNum" is used to detect late dirty calls.  If
     * the VMID "vmid" is null, a VMID will be generated on the
     * server (for use by the client in subsequent calls) and
     * returned.
     *
     * The client must call the "dirty" method to renew the lease
     * before the "lease" time expires or all references to remote
     * objects in this VM that the client holds are considered
     * "unreferenced".
     */
    public Lease dirty(ObjID[] ids, long sequenceNum, Lease lease)
    {
	VMID vmid = lease.getVMID();
	/*
	 * The server specifies the lease value; the client has
	 * no say in the matter.
	 */
	long duration = leaseValue;

	if (DGCImpl.logLevel >= LogStream.VERBOSE)
	    LogStream.log("dgc").println("DGCImpl.dirty: vmid = " + vmid);

	// create a VMID if one wasn't supplied
	if (vmid == null) {
	    vmid = new VMID();

	    if (DGCImpl.logLevel >= LogStream.BRIEF) {
		String clientHost;
		try {
		    clientHost = RemoteServer.getClientHost();
		} catch (ServerNotActiveException e) {
		    clientHost = "<unknown host>";
		}
		LogStream.log("dgc").println(
		    "DGCImpl.dirty: assigning vmid " + vmid +
		    " to client " + clientHost);
	    }
	}

	lease = new Lease(vmid, duration);
	// record lease information
	synchronized (leaseTable) {
	    LeaseInfo info = (LeaseInfo)leaseTable.get(vmid);
	    if (info == null) {
		leaseTable.put(vmid, new LeaseInfo(vmid, duration));
		if (checker == null) {
		    checker = (Thread)
			java.security.AccessController.doPrivileged(
			    new RMIThreadAction(new LeaseChecker(),
						"LeaseChecker", true));
		    checker.start();
		}
	    } else {
		info.renew(duration);
	    }
	}

	for (int i=0; i<ids.length; i++) {
	    if (DGCImpl.logLevel >= LogStream.VERBOSE)
		LogStream.log("dgc").println(
		    "DGCImpl.dirty: id = " + ids[i] +
		     ", vmid = " + vmid + ", duration = " + duration);

	    ObjectTable.referenced(ids[i], sequenceNum, vmid);
	}

	// return the VMID used
	return lease;
    }

    /**
     * The clean call removes the VMID from the set of clients
     * that hold references to the object associated with the LiveRef
     * ref.  The sequence number is used to detect late clean calls.  If the
     * argument "strong" is true, then the clean call is a result of a
     * failed "dirty" call, thus the sequence number for the VMID needs
     * to be remembered until the client goes away.
     */
    public void clean(ObjID[] ids, long sequenceNum, VMID vmid, boolean strong)
    {
	for (int i=0; i<ids.length; i++) {
	    if (DGCImpl.logLevel >= LogStream.VERBOSE)
		LogStream.log("dgc").println(
		    "DGCImpl.clean: id = " + ids[i] +
		    ", vmid = " + vmid + ", strong = " + strong);

	    ObjectTable.unreferenced(ids[i], sequenceNum, vmid, strong);
	}
    }

    /**
     * Register interest in receiving a callback when this VMID
     * becomes inaccessible.
     */
    void registerTarget(VMID vmid, Target target) {
	synchronized (leaseTable) {
	    LeaseInfo info = (LeaseInfo)leaseTable.get(vmid);
	    if (info == null) {
		target.vmidDead(vmid);
	    } else {
		synchronized (info.notifyList) {
		    if (!info.notifyList.contains(target)) {
			info.notifyList.addElement(target);
		    }
		}
	    }
	}
    }

    /**
     * Remove notification request.
     */
    void unregisterTarget(VMID vmid, Target target) {
	synchronized (leaseTable) {
	    LeaseInfo info = (LeaseInfo)leaseTable.get(vmid);
	    if (info != null) {
		synchronized (info.notifyList) {
		    info.notifyList.removeElement(target);
		}
	    }
	}
    }

    /**
     * Check if leases have expired.  If a lease has expired, remove
     * it from the table and notify all interested parties that the
     * VMID is essentially "dead".
     *
     * @return if true, there are leases outstanding; otherwise leases
     * no longer need to be checked
     */
    private boolean checkLeases() {
        long time = System.currentTimeMillis();
	Enumeration enum = null;
	LeaseInfo info = null;
	boolean leasesRemain = false;

	/* List of vmids that need to be removed from the leaseTable */
	Vector toUnregister = new Vector(5);

	/* Build a list of leaseInfo objects that need to have
	 * targets removed from their notifyList.  Remove expired
	 * leases from leaseTable.
	 */
	synchronized (leaseTable) {
	    enum = (Enumeration) leaseTable.elements();
	    while (enum.hasMoreElements()) {
		info = (LeaseInfo) enum.nextElement();
		if (info.expired(time)) {
		    toUnregister.add(info);
		    leaseTable.remove(info.vmid);			
		}
	    }
	    
	    if (leaseTable.isEmpty()) {
		checker = null;
		leasesRemain = false;
	    } else {
		leasesRemain = true;
	    }
	}
	
	/* Notify and unegister targets without holding the lock on
	 * the leaseTable so we avoid deadlock.
	 */
	Enumeration remove = toUnregister.elements();
	while (remove.hasMoreElements()) {
	    
	    info = (LeaseInfo) remove.nextElement();
	    
	    synchronized (info.notifyList) {
		enum = (Enumeration) info.notifyList.elements();
		while (enum.hasMoreElements()) {
		    Target target = (Target) enum.nextElement();
		    target.vmidDead(info.vmid);
		}
	    }
	}
	return leasesRemain;
    }

    static ObjID getID() {
	return dgcID;
    }

    static {
	/* Put remote collector object in table by hand to prevent
	 * listen on port. (Using RemoteServer.exportObject would
	 * cause transport to listen).
	 */
	try {
	    dgc = new DGCImpl();
	    dgcID = new ObjID(ObjID.DGC_ID);
	    LiveRef ref = new LiveRef(dgcID, 0);
	    UnicastServerRef disp = new UnicastServerRef(ref);
	    RemoteStub stub = disp.setSkeleton(dgc);
	    Target target = new Target(dgc, disp, stub, dgcID, true);
	    ObjectTable.putTarget(target);
	} catch (Exception e) {
	    System.out.println("couldn't init remote garbage collector");
	    System.out.println(e.getMessage());
	    e.printStackTrace();
	}
    }

    /**
     * The lease checker runs in a thread that checks if leases have expired.
     * This thread stops when there are no more leases to be checked.
     */
    private class LeaseChecker implements Runnable {
        public void run() {
            do {
	        try {
		    Thread.sleep(leaseCheckInterval);
	        } catch (InterruptedException e) {
	        }
	    } while (checkLeases());
        }
    }

    private static class LeaseInfo {
        VMID vmid;
        long expiration;
        Vector notifyList = new Vector();

        LeaseInfo(VMID vmid, long lease) {
	    this.vmid = vmid;
	    expiration = System.currentTimeMillis() + lease;
        }

        synchronized void renew(long lease) {
	    long newExpiration = System.currentTimeMillis() + lease;
	    if (newExpiration > expiration)
	        expiration = newExpiration;
        }

        boolean expired(long time) {
            if (expiration < time) {
                if (DGCImpl.logLevel >= LogStream.BRIEF)
                    LogStream.log("dgc").println(
                        "LeaseInfo.expired: " + vmid);
                return true;
            } else {
                return false;
            }
        }
    } /* end class LeaseInfo */


}  /* end class DGCImpl */


