/*
 * @(#)LiveRef.java	1.21 98/07/15
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.rmi.transport;

import java.lang.ref.*;
import java.io.*;
import java.util.*;
import java.rmi.*;
import java.rmi.server.*;
import sun.rmi.transport.tcp.TCPEndpoint;

public class LiveRef implements Cloneable {
    /** wire representation for the object*/
    private Endpoint ep;
    private ObjID id;

    /** cached connection service for the object */
    private transient Channel ch;

    /** flag to indicate whether this ref specifies a local server or
     * is a ref for a remote object (surrogate)
     */
    private boolean isLocal = false;

    /**
     * Construct a "well-known" live reference to a remote object
     * @param isLocalServer If true, indicates this ref specifies a local
     * server in this address space; if false, the ref is for a remote
     * object (hence a surrogate or proxy) in another address space.
     */
    public LiveRef(ObjID objID, Endpoint endpoint, boolean isLocalServer) {
	ep = endpoint;
	id = objID;
	isLocal = isLocalServer;
    }

    /**
     * Construct a new live reference for a server object in the local
     * address space.
     */
    public LiveRef(int port) {
	this((new ObjID()), port);
    }

    /**
     * Construct a new live reference for a server object in the local
     * address space, to use sockets of the specified type.
     */
    public LiveRef(int port,
		   RMIClientSocketFactory csf,
		   RMIServerSocketFactory ssf)
    {
	this((new ObjID()), port, csf, ssf);
    }

    /**
     * Construct a new live reference for a "well-known" server object
     * in the local address space.
     */
    public LiveRef(ObjID objID, int port) {
	this(objID, TCPEndpoint.getLocalEndpoint(port), true);
    }

    /**
     * Construct a new live reference for a "well-known" server object
     * in the local address space, to use sockets of the specified type.
     */
    public LiveRef(ObjID objID, int port, RMIClientSocketFactory csf,
		   RMIServerSocketFactory ssf)
    {
	this(objID, TCPEndpoint.getLocalEndpoint(port, csf, ssf), true);
    }

    /**
     * Return a shallow copy of this ref.
     */
    public Object clone() {
	try {
	    LiveRef newRef = (LiveRef) super.clone();
	    return newRef;
	} catch (CloneNotSupportedException e) {
	    throw new InternalError(e.toString());
	}
    }

    /**
     * Return the port number associated with this ref.
     */
    public int getPort() {
	return ((TCPEndpoint)ep).getPort();
    }
    
    /**
     * Return the client socket factory associated with this ref.
     */
    public RMIClientSocketFactory getClientSocketFactory() {
	return ((TCPEndpoint)ep).getClientSocketFactory();
    }
    
    /**
     * Return the server socket factory associated with this ref.
     */
    public RMIServerSocketFactory getServerSocketFactory() {
	return ((TCPEndpoint)ep).getServerSocketFactory();
    }
    
    /**
     * Export the object to accept incoming calls.
     */
    public void exportObject(Target target) throws RemoteException {
	ep.exportObject(target);
    }
    
    public Channel getChannel() throws RemoteException {
	if (ch == null) {
	    ch = ep.getChannel();
	}
	
	return ch;
    }

    public ObjID getObjID() {
	return id;
    }

    Endpoint getEndpoint() {
	return ep;
    }

    public String toString() {
	String type;
	
	if (isLocal)
	    type = "local";
	else
	    type = "remote";
	return "[endpoint:" + ep + "(" + type + ")," +
	    "objID:" + id + "]";
    }

    public int hashCode() {
	return id.hashCode();
    }

    public boolean equals(Object obj) {
	if (obj != null && obj instanceof LiveRef) {
	    LiveRef ref = (LiveRef)obj;
	    return (ep.equals(ref.ep) && id.equals(ref.id) &&
		    isLocal == ref.isLocal);
	} else {
	    return false;
	}
    }
    
    public boolean remoteEquals(Object obj) {
	if (obj != null && obj instanceof LiveRef) {
	    LiveRef ref = (LiveRef)obj;
	    return (ep.equals(ref.ep) && id.equals(ref.id));
	} else {
	    return false;
	}
    }
    
    public void write(ObjectOutput out, boolean useNewFormat)
	throws IOException
    {
	boolean isResultStream = false;

	if (out instanceof ConnectionOutputStream) {
	    ConnectionOutputStream stream = (ConnectionOutputStream)out;
	    if (isLocal) {
		// make sure impl is pinned before sending out reference
		Target target = ObjectTable.getTarget(id);
		if (target != null)
		    target.pinImpl();
		else
		    throw new IOException("attempt to marshal unexported " +
					  "object reference: " + this);
	    }
	    isResultStream = stream.isResultStream();
	    if (isResultStream) {
		// save liveRef until an ack is received.
		stream.saveRef(this);
	    }
	}
	// All together now write out the endpoint, id, and flag

	// (need to choose whether or not to use old JDK1.1 endpoint format)
	if (useNewFormat) {
	    ((TCPEndpoint) ep).write(out);
	} else {
	    ((TCPEndpoint) ep).writeHostPortFormat(out);
	}
	id.write(out);
	out.writeBoolean(isResultStream);
    }
    
    public static LiveRef read(ObjectInput in, boolean useNewFormat)
	throws IOException, ClassNotFoundException
    {
	Endpoint ep;
	ObjID id;

	// Now read in the endpoint, id, and result flag
	// (need to choose whether or not to read old JDK1.1 endpoint format)
	if (useNewFormat) {
	    ep = TCPEndpoint.read(in);
	} else {
	    ep = TCPEndpoint.readHostPortFormat(in);
	}
	id = ObjID.read(in);
	boolean isResultStream = in.readBoolean();

	LiveRef ref = new LiveRef(id, ep, false);
	
	if (in instanceof ConnectionInputStream) {
	    ConnectionInputStream stream = (ConnectionInputStream)in;
	    // save ref to send "dirty" call after all args/returns
	    // have been unmarshaled.
	    stream.saveRef(ref);
	    if (isResultStream) {
		// set flag in stream indicating that remote objects were
		// unmarshaled.  A DGC ack should be sent by the transport.
		stream.setAckNeeded();
	    }
	} else {
	    DGCClient.registerRefs(ep, Arrays.asList(new LiveRef[] { ref }));
	}

	return ref;
    }
}
