/*
 * @(#)ObjectTable.java	1.20 98/09/11
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.rmi.transport;

import java.lang.ref.*;
import java.rmi.Remote;
import java.rmi.NoSuchObjectException;
import java.rmi.dgc.VMID;
import java.rmi.server.ExportException;
import java.rmi.server.LogStream;
import java.rmi.server.ObjID;
import java.rmi.server.RemoteStub;
import java.security.PrivilegedAction;
import java.util.*;

import sun.misc.GC;
import sun.security.action.GetLongAction;

/**
 * Object table shared by all implementors of the Transport interface.
 * This table maps object ids to remote object targets in this address
 * space.
 *
 * @version 1.20, 09/11/98
 * @author  Ann Wollrath
 * @author  Peter Jones
 */
public final class ObjectTable {

    /** maximum interval between complete garbage collections of local heap */
    private final static long gcInterval =		// default 1 minute
	((Long) java.security.AccessController.doPrivileged(
	    new GetLongAction("sun.rmi.dgc.server.gcInterval",
			      60000))).longValue();

    /**
     * lock to enforce mutual exclusion on the following static fields:
     * objTable, implTable, lookupKey, keepAliveCount, reaper
     */
    private static Object lock = new Object();

    /** tables mapping to Target, keyed from ObjID and impl object */
    private static Map objTable = new HashMap();
    private static Map implTable = new HashMap();

    /** count of objects in the table, excluding permanent entries */
    private static int keepAliveCount = 0;

    /** thread to collect unreferenced objects from table */
    private static Thread reaper = null;

    /** queue notified when weak refs in the table are cleared */
    static ReferenceQueue reapQueue = new ReferenceQueue();

    /** handle for GC latency request (for future cancellation) */
    private static GC.LatencyRequest gcLatencyRequest = null;

    /*
     * Disallow anyone from creating one of these.
     */
    private ObjectTable() {}
    
    /**
     * Returns the target associated with the object id.
     */
    static Target getTarget(ObjID id) {
	synchronized (lock) {
	    return (Target)objTable.get(id);
	}
    }

    /**
     * Returns the target associated with the remote object
     */
    public static Target getTarget(Remote impl) {
	synchronized (lock) {
	    return (Target)implTable.get(new WeakRef(impl));
	}
    }

    /**
     * Returns the stub for the remote object <b>obj</b> passed
     * as a parameter. This operation is only valid <i>after</i>
     * the object has been exported.
     *
     * @return the stub for the remote object, <b>obj</b>.
     * @exception NoSuchObjectException if the stub for the
     * remote object could not be found.
     */
    public static RemoteStub getStub(Remote impl)
	throws NoSuchObjectException
    {
	Target target = getTarget(impl);
	if (target == null) {
	    throw new NoSuchObjectException("object not exported");
	} else {
	    return target.getStub();
	}
    }

   /**
    * Remove the remote object, obj, from the RMI runtime. If
    * successful, the object can no longer accept incoming RMI calls.
    * If the force parameter is true, the object is forcibly unexported
    * even if there are pending calls to the remote object or the
    * remote object still has calls in progress.  If the force
    * parameter is false, the object is only unexported if there are
    * no pending or in progress calls to the object.
    *
    * @param obj the remote object to be unexported
    * @param force if true, unexports the object even if there are
    * pending or in-progress calls; if false, only unexports the object
    * if there are no pending or in-progress calls
    * @return true if operation is successful, false otherwise
    * @exception NoSuchObjectException if the remote object is not
    * currently exported
    */
   public static boolean unexportObject(Remote obj, boolean force)
	throws java.rmi.NoSuchObjectException
    {
	synchronized (lock) {
	    Target target = getTarget(obj);
	    if (target == null) {
		throw new NoSuchObjectException("object not exported");
	    } else {
		if (target.unexport(force)) {
		    removeTarget(target);
		    return true;
		} else {
		    return false;
		}
	    }
	}
    }

    /**
     * Add target to object table.  If it is not a permanent entry, then
     * make sure that reaper thread is running to remove collected entries
     * and keep VM alive.
     */
    static void putTarget(Target target) throws ExportException {
	ObjID id = target.getObjID();
	WeakRef weakImpl = target.getWeakImpl();

	if (DGCImpl.logLevel >= LogStream.VERBOSE)
	    LogStream.log("dgc").println(
		"ObjectTable.putTarget: add object " + id);
	
	Remote impl = target.getImpl();
	if (impl == null) {
	    throw new ExportException(
		"internal error: attempt to export collected object");
	}

	synchronized (lock) {
	    if (objTable.containsKey(id)) {
		throw new ExportException(
		    "internal error: ObjID already in use");
	    } else if (implTable.containsKey(weakImpl)) {
		throw new ExportException("object already exported");
	    }

	    objTable.put(id, target);
	    implTable.put(weakImpl, target);

	    if (!target.isPermanent()) {
		++ keepAliveCount;

		if (reaper == null) {
		    reaper = (Thread) 
			java.security.AccessController.doPrivileged(
			    new RMIThreadAction(new Reaper(), "Reaper",
						false));
		    reaper.start();
		}

		/*
		 * While there are non-"permanent" objects in the object table,
		 * request a maximum latency for inspecting the entire heap
		 * from the local garbage collector, to place an upper bound
		 * on the time to discover remote objects that have become
		 * unreachable (and thus can be removed from the table).
		 */
		if (gcLatencyRequest == null) {
		    gcLatencyRequest = GC.requestLatency(gcInterval);
		}
	    }
	}
    }

    /**
     * Remove target from object table.  If it was the last non-permanent
     * entry, then kill the reaper thread so the VM can exit.
     *
     * NOTE: This method must only be invoked while synchronized on
     * the "lock" object, because it does not do so itself.
     */
    private static void removeTarget(Target target) {
	ObjID id = target.getObjID();
	WeakRef weakImpl = target.getWeakImpl();

	if (DGCImpl.logLevel >= LogStream.VERBOSE)
	    LogStream.log("dgc").println(
		"ObjetTable.removeTarget: remove object " + id);

	objTable.remove(id);
	implTable.remove(weakImpl);

	if (!target.isPermanent()) {
	    -- keepAliveCount;

	    if (keepAliveCount == 0) {
		java.security.AccessController.doPrivileged(
		    new PrivilegedAction() {
		    public Object run() {
			ObjectTable.reaper.interrupt();
			return null;
		    }
		});
		reaper = null;

		/*
		 * If there are no longer any non-permanent objects in the
		 * object table, we are no longer concerned with the latency
		 * of local garbage collection here.
		 */
		gcLatencyRequest.cancel();
		gcLatencyRequest = null;
	    }
	}
    }

    /**
     * Process client VM signalling reference for given ObjID: forward to 
     * correspoding Target entry.  If ObjID is not found in table,
     * no action is taken.
     */
    static void referenced(ObjID id, long sequenceNum, VMID vmid) {
	synchronized (lock) {
	    Target target = (Target) objTable.get(id);
	    if (target != null)
		target.referenced(sequenceNum, vmid);
	}
    }

    /**
     * Process client VM dropping reference for given ObjID: forward to 
     * correspoding Target entry.  If ObjID is not found in table,
     * no action is taken.
     */
    static void unreferenced(ObjID id, long sequenceNum, VMID vmid,
			     boolean strong)
    {
	synchronized (lock) {
	    Target target = (Target) objTable.get(id);
	    if (target != null)
		target.unreferenced(sequenceNum, vmid, strong);
	}
    }

    /**
     * The Reaper thread waits for notifications that weak references in the
     * object table have been cleared.  When it receives a notification, it
     * removes the corresponding entry from the table.
     *
     * Since the Reaper is created as a non-daemon thread, it also serves
     * to keep the VM from exiting while there are objects in the table
     * (other than permanent entries that should neither be reaped nor
     * keep the VM alive).
     */
    private static class Reaper implements Runnable {

	public void run() {
	    try {
		do {
		    // wait for next cleared weak reference
		    WeakRef weakImpl = (WeakRef) reapQueue.remove();

		    synchronized (lock) {
			Target target = (Target) implTable.get(weakImpl);
			if (target != null) {
			    if (!target.isEmpty()) {
				throw new Error(
				    "object with known references collected");
			    } else if (target.isPermanent()) {
				throw new Error("permanent object collected");
			    }
			    removeTarget(target);
			}
		    }
		} while (!Thread.interrupted());
	    } catch (InterruptedException e) {
		// pass away if interrupted
	    }
	}
    }
}
