/*
 * @(#)RMIThreadAction.java	1.3 98/09/21
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.rmi.transport;

import java.rmi.server.LogStream;

/*
 * RMIThreadAction is a utility class for creating threads using
 * AccessController.doPrivileged.
 *
 * The RMIThreadAction class centralizes the creation of the threads for use
 * by the RMI implementation runtime, so they are all placed in an appropriate
 * thread group.
 *
 * Note that since a thread in the root thread group is created via the
 * RMIThreadAction class, a public class, it does NOT acquire privileged
 * AccessController status, so it does not give capabilites that they
 * would not otherwise have.  Therefore, code in the RMI runtime that
 * calls RMIThreadAction should use the doPrivileged construct to
 * create threads.
 */
public class RMIThreadAction implements java.security.PrivilegedAction {

    private Runnable runnable;
    private String name;
    private boolean daemon;

    public RMIThreadAction(Runnable runnable, String name, boolean daemon) {
	this.runnable = runnable;
	this.name = name;
	this.daemon = daemon;
    }

    /*
     * Create a new thread for RMI runtime.
     */
    public Object run() {
	Thread t = new Thread(getThreadGroup(), runnable, "RMI " + name);
	t.setDaemon(daemon);
	return t;
    }

    /*
     * Return the thread group to use for RMI runtime threads.  If this
     * thread group has not yet been created, then create it as a child
     * of the system thread group.
     */
    private synchronized static ThreadGroup getThreadGroup() {
	if (group == null) {
	    /*
	     * Search up from current thread's thread group to find the
	     * root of the ThreadGroup tree (the system thread group).
	     */
	    ThreadGroup root = Thread.currentThread().getThreadGroup();
	    ThreadGroup parent;
	    while ((parent = root.getParent()) != null) {
		if (Transport.logLevel >= LogStream.VERBOSE) {
		    LogStream.log("transport").println(
			"RMIThread.getThreadGroup: " +
			"searching up from ThreadGroup: " + root);
		}
		root = parent;
	    }
	    if (Transport.logLevel >= LogStream.VERBOSE) {
		LogStream.log("transport").println(
		    "RMIThread.getThreadGroup: " +
		    "found root ThreadGroup: " + root);
	    }

	    /*
	     * As long as the implementations of the default SecurityManager's
	     * checkAccess() methods protect "system" threads by simply
	     * denying access (without permission) to threads in the topmost
	     * thread group (whose parent is null), we place RMI threads
	     * directly in the system thread group, so that they will be as
	     * protected as other system threads (see bugid 4166914).  We
	     * would prefer, however, to go back to putting RMI threads in a
	     * separate thread group, as coded below, to organize the threads
	     * better and not pollute the top-level thread group.
	     */
	    group = root;

//	    /*
//	     * Create thread group for RMI runtime under this system
//	     * thread group.  If this fails for security reasons,
//	     * at least create RMI thread group in current thread's
//	     * thread group.
//	     */
//	    try {
//		final ThreadGroup finalRoot = root;
//		java.security.AccessController.doPrivileged(
//		    new java.security.PrivilegedAction() {
//		    public Object run() {
//			group = new ThreadGroup(finalRoot, "RMI Runtime");
//			return null;
//		    }
//		});
//	    } catch (SecurityException e) {
//		if (Transport.logLevel >= LogStream.BRIEF) {
//		    LogStream ls = LogStream.log("transport");
//		    ls.println("RMIThread.getThreadGroup: " +
//			       "security exception: ");
//		    e.printStackTrace(ls);
//		}
//		group = new ThreadGroup("RMI runtime (applet)");
//	    }
//	    if (Transport.logLevel >= LogStream.VERBOSE) {
//		LogStream.log("transport").println(
//		    "RMIThread.getThreadGroup: " +
//		    "created RMI ThreadGroup:" + group);
//	    }
	}
	return group;
    }

    /*
     * ThreadGroup for RMI runtime implementation's threads;
     * accessed through getThreadGroup() method
     */
    private static ThreadGroup group;
}
