/*
 * @(#)StreamRemoteCall.java	1.12 98/08/19
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.rmi.transport;

import java.io.*;
import java.util.Vector;
import java.rmi.*;
import java.rmi.server.LogStream;
import java.rmi.server.ObjID;
import java.rmi.server.RemoteCall;
import java.rmi.RemoteException;
import java.rmi.server.ServerCloneException;
import java.rmi.activation.ActivationException;

/**
 * Stream-based implementation of the RemoteCall interface.
 *
 * @author Ann Wollrath
 */
public class StreamRemoteCall implements RemoteCall {
    private ConnectionInputStream in = null;
    private ConnectionOutputStream out = null;
    private Connection conn;
    private boolean resultStarted = false;
    
    public StreamRemoteCall(Connection c) {
	conn = c;
    }
    
    public StreamRemoteCall(Connection c, ObjID id, int op, long hash)
	throws RemoteException
    {
	try {
	    conn = c;
	    if (Transport.logLevel >= LogStream.VERBOSE)
		LogStream.log("transport").println(
		    "StreamRemoteCall.<init>: write remote call header...");
	    
	    // write out remote call header info...
	    // call header, part 1 (read by Transport)
	    conn.getOutputStream().write(TransportConstants.Call);
	    getOutputStream();		 // creates a MarshalOutputStream
	    id.write(out);		 // object id (target of call)
	    // call header, part 2 (read by Dispatcher)
	    out.writeInt(op);		 // method number (operation index)
	    out.writeLong(hash);         // stub/skeleton hash
	} catch (IOException e) {
	    throw new MarshalException("Error marshaling call header", e);
	}
    }

    /**
     * Return the connection associated with this call.
     */
    public Connection getConnection() {
	return conn;
    }
    
    /**
     * Return the output stream the stub/skeleton should put arguments/results
     * into.
     */
    public ObjectOutput getOutputStream() throws IOException {
	return getOutputStream(false);
    }
    
    private ObjectOutput getOutputStream(boolean resultStream)
	throws IOException 
    {
	if (out == null) {
	    if (Transport.logLevel >= LogStream.VERBOSE)
		LogStream.log("transport").println(
		    "StreamRemoteCall.getOutputStream");

	    out = new ConnectionOutputStream(conn, resultStream);
	}
	return out;
    }
    
    /**
     * Release the outputStream  Currently, will not complain if the
     * output stream is released more than once.
     */
    public void releaseOutputStream() throws IOException {
	try {
	    if (out != null) {
		out.flush(); 
	    }
	    conn.releaseOutputStream();
	}
	finally {
	    out = null;
	}
    }

    /**
     * Get the InputStream the stub/skeleton should get results/arguments
     * from.
     */
    public ObjectInput getInputStream() throws IOException {
	if (in == null) {
	    if (Transport.logLevel >= LogStream.VERBOSE)
		LogStream.log("transport").println(
		    "StreamRemoteCall.getInputStream");

	    in = new ConnectionInputStream(conn.getInputStream());
	}
	return in;
    }
    
    /**
     * Release the input stream, this would allow some transports to release
     * the channel early.
     */
    public void releaseInputStream() throws IOException {
	/* WARNING: Currently, the UnicastRef.java invoke methods rely
	 * upon this method not throwing an IOException.
	 */

	try {
	    if (in != null) {
		// add saved references to DGC table
		in.registerRefs();

		/* WARNING: The connection being passed to done may have
		 * already been freed.
		 */
		in.done(conn);
	    }
	    conn.releaseInputStream();
	} finally {
	    in = null;
	}
    }

    /**
     * Returns an output stream (may put out header information
     * relating to the success of the call).
     * @param success If true, indicates normal return, else indicates
     * exceptional return.
     * @exception StreamCorruptedException If result stream previously
     * acquired
     * @exception IOException For any other problem with I/O.
     */
    public ObjectOutput getResultStream(boolean success)
	throws IOException, StreamCorruptedException
    {
	/* make sure result code only marshaled once. */
	if (resultStarted)
	    throw new StreamCorruptedException("result already in progress");
	else
	    resultStarted = true;

	// write out return header
	// return header, part 1 (read by Transport)
	DataOutputStream wr = new DataOutputStream(conn.getOutputStream());
	wr.writeByte(TransportConstants.Return);// transport op
	getOutputStream(true);	// creates a MarshalOutputStream
	// return header, part 2 (read by client-side RemoteCall)
	if (success)		// 
	    out.writeByte(TransportConstants.NormalReturn);
	else
	    out.writeByte(TransportConstants.ExceptionalReturn);
	out.writeID();		// write id for gcAck
	return out;
    }

    /**
     * Do whatever it takes to execute the call.
     */
    public void executeCall() throws Exception {
	byte returnType;
	
	// read result header
	try {
	    releaseOutputStream();
	    DataInputStream rd = new DataInputStream(conn.getInputStream());
	    byte op = rd.readByte();
	    if (op != TransportConstants.Return) {
		if (Transport.logLevel >= LogStream.BRIEF)
		    LogStream.log("transport").println(
		    "StreamRemoteCall.executeCall: " +
		    "transport return code invalid: " + op);
		throw new UnmarshalException("Transport return code invalid");
	    }
	    getInputStream();
	    returnType = in.readByte();
	    in.readID();	// id for DGC acknowledgement
	} catch (UnmarshalException e) {
	    throw e;
	} catch (IOException e) {
	    throw new UnmarshalException("Error unmarshaling return header",
					 e);
	}

	// read return value
	switch (returnType) {
	case TransportConstants.NormalReturn:
	    break;
	    
	case TransportConstants.ExceptionalReturn:
	    Object ex;
	    try {
		ex = in.readObject();
	    } catch (Exception e) {
		throw new UnmarshalException("Error unmarshaling return", e);
	    }
	    
	    // An exception should have been received,
	    // if so throw it, else flag error
	    if (ex instanceof Exception) {
		exceptionReceivedFromServer((Exception) ex);
	    } else {
		throw new UnmarshalException("Return type not Exception");
	    }
	default:
	    if (Transport.logLevel >= LogStream.BRIEF)
		LogStream.log("transport").println(
		    "StreamRemoteCall.executeCall: return code invalid: " +
		    returnType);
	    throw new UnmarshalException("Return code invalid");
	}
    }

    /**
     * Routine that causes the stack traces of remote exceptions to be
     * filled in with the current stack trace on the client.  Detail
     * exceptions are filled in iteratively.
     */
    protected void exceptionReceivedFromServer(Exception ex) 
	throws Exception 
    {
	Throwable t = ex;
	while (t != null) {
	    t.fillInStackTrace();
	    if (t instanceof RemoteException) {
		t = ((RemoteException) t).detail;
	    } else if (t instanceof ServerCloneException) {
		t = ((ServerCloneException) t).detail;
	    } else if (t instanceof ActivationException) {
		t = ((ActivationException) t).detail;
	    } else {
		break;
	    }
	}
	throw ex;
    }
    
    public void done() throws IOException {
	/* WARNING: Currently, the UnicastRef.java invoke methods rely
	 * upon this method not throwing an IOException.
	 */

	releaseInputStream();
    }
}
