/*
 * @(#)HttpReceiveSocket.java	1.6 98/09/21
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.rmi.transport.proxy;

import java.io.*;
import java.net.Socket;
import java.net.InetAddress;

/**
 * The HttpReceiveSocket class extends the WrappedSocket class
 * by removing the HTTP protocol packaging from the input stream and
 * formatting the output stream as an HTTP response.
 *
 * NOTES:
 *
 * The output stream must be explicitly closed for the output to be
 * sent, since the HttpResponseOutputStream needs to buffer the entire
 * transmission to be able to fill in the content-length field of
 * the HTTP header.  Closing this socket will do this.
 *
 * The constructor blocks until the HTTP protocol header
 * is received.  This could be fixed, but I don't think it should be a
 * problem because this object would not be created unless the
 * HttpAwareServerSocket has detected the beginning of the header
 * anyway, so the rest should be there.
 *
 * This socket can only be used to process one POST and reply to it.
 * Another message would be received on a newly accepted socket anyway.
 */
class HttpReceiveSocket extends WrappedSocket implements RMISocketInfo {

    /** true if the HTTP header has pushed through the output stream yet */
    private boolean headerSent = false;

    /**
     * Layer on top of a pre-existing Socket object, and use specified
     * input and output streams.
     * @param socket the pre-existing socket to use
     * @param in the InputStream to use for this socket (can be null)
     * @param out the OutputStream to use for this socket (can be null)
     */
    public HttpReceiveSocket(Socket socket, InputStream in, OutputStream out)
        throws IOException
    {
	super(socket, in, out);

	this.in = new HttpInputStream(in != null ? in :
	                                           socket.getInputStream());
	this.out = (out != null ? out :
	            socket.getOutputStream());
    }

    /**
     * Indicate that this socket is not reusable.
     */
    public boolean isReusable()
    {
	return false;
    }

    /**
     * Get the address to which the socket is connected.  "null" is always
     * returned because since the HTTP request may have gone through a
     * proxy server, the originating host's IP address cannot be reliably
     * determined.
     */
    public InetAddress getInetAddress()
    {
	return null;
    }

    /**
     * Get the remote port to which the socket is connected.  Since the
     * HTTP request probably went through a proxy server, then this will
     * be the port on the gateway machine, which is probably not very
     * useful information.
     */
    public int getPort()
    {
	return socket.getPort();
    }

    /**
     * Get an OutputStream for this socket.
     */
    public OutputStream getOutputStream() throws IOException
    {
	if (!headerSent) { // could this be done in constructor??
	    DataOutputStream dos = new DataOutputStream(out);
	    dos.writeBytes("HTTP/1.0 200 OK\r\n");
	    dos.flush();
	    headerSent = true;
	    out = new HttpOutputStream(out);
	}
	return out;
    }

    /**
     * Close the socket.
     */
    public synchronized void close() throws IOException
    {
	getOutputStream().close(); // make sure response is sent
	socket.close();
    }

    /**
     * Return string representation of the socket.
     */
    public String toString()
    {
	return "HttpReceive" + socket.toString();
    }
}
