/*
 * @(#)TCPChannel.java	1.29 98/07/15
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.rmi.transport.tcp;

import java.lang.ref.SoftReference;
import java.io.*;
import java.net.Socket;
import java.util.*;
import java.rmi.*;
import java.rmi.server.ExportException;
import java.rmi.server.LogStream;
import java.rmi.server.RMISocketFactory;
import java.security.AccessController;
import java.security.AccessControlContext;
import sun.rmi.transport.Channel;
import sun.rmi.transport.Connection;
import sun.rmi.transport.Endpoint;
import sun.rmi.transport.Notifiable;
import sun.rmi.transport.RMIThreadAction;
import sun.rmi.transport.Transport;
import sun.rmi.transport.TransportConstants;
import sun.rmi.transport.Utils;

import sun.security.action.GetLongAction;

/**
 * TCPChannel is the socket-based implementation of the RMI Channel
 * abstraction.
 *
 * @author Ann Wollrath
 */
public class TCPChannel implements Channel, Runnable {
    /** endpoint for this channel */
    private TCPEndpoint ep;
    /** transport for this channel */
    private TCPTransport tr;
    /** objects to notify if the channel becomes dead */
    private Vector notifyList = new Vector();
    /** list of cached connections */
    private ArrayList freeList = new ArrayList();
    /** frees cached connections that have expired */
    private Reaper reaper = null;

    /** using multiplexer (for bi-directional applet communication */
    private boolean usingMultiplexer = false;
    /** connection multiplexer, if used */
    private ConnectionMultiplexer multiplexer = null;
    /** connection acceptor (should be in TCPTransport) */
    private ConnectionAcceptor acceptor;

    /** most recently authorized AccessControlContext */
    private AccessControlContext okContext;

    /** cache of authorized AccessControlContexts */
    private WeakHashMap authcache;

    /** the SecurityManager which authorized okContext and authcache */
    private SecurityManager cacheSecurityManager = null;
    
    private static long timeout;

    static {
	/*
	 * Set connection timeout (defaults to 30 seconds).
	 */
	Long tmp = (Long) java.security.AccessController.doPrivileged(
                    new GetLongAction("sun.rmi.transport.connectionTimeout", 
				      15000));
	timeout = tmp.longValue();
    }

    /**
     * Create channel for endpoint.
     */
    TCPChannel(TCPTransport tr, TCPEndpoint ep) {
	this.tr = tr;
	this.ep = ep;
    }

    /**
     * Return the endpoint for this channel.
     */
    public Endpoint getEndpoint() {
	return ep;
    }

    /**
     * Checks if the current caller has sufficient privilege to make
     * a connection to the remote endpoint.
     * @exception SecurityException if caller is not allowed to use this
     * Channel.
     */
    private void checkConnectPermission() throws SecurityException {
	SecurityManager security = System.getSecurityManager();
	if (security == null)
	    return;

	if (security != cacheSecurityManager) {
	    // The security manager changed: flush the cache
	    okContext = null;
	    authcache = new WeakHashMap();
	    cacheSecurityManager = security;
	}

	AccessControlContext ctx = AccessController.getContext();

	// If ctx is the same context as last time, or if it
	// appears in the cache, bypass the checkConnect.
	if (okContext == null ||
	       !(okContext.equals(ctx) || authcache.containsKey(ctx)))
	{
	    security.checkConnect(ep.getHost(), ep.getPort());
	    authcache.put(ctx, new SoftReference(ctx));
	    // A WeakHashMap is transformed into a SoftHashSet by making
	    // each value softly refer to its own key (Peter's idea).
	}
	okContext = ctx;
    }

    /**
     * Supplies a connection to the endpoint of the address space
     * for which this is a channel.  The returned connection may
     * be one retrieved from a cache of idle connections.
     */
    public Connection newConnection() throws RemoteException {
	TCPConnection conn;

	// loop until we find a free live connection (in which case
	// we return) or until we run out of freelist (in which case
	// the loop exits)
	do {
	    conn = null;
	    // try to get a free connection 
	    synchronized (freeList) {
		int elementPos = freeList.size()-1;

		if (elementPos >= 0) {
		    // If there is a security manager, make sure
		    // the caller is allowed to connect to the
		    // requested endpoint.
		    checkConnectPermission();
		    conn = (TCPConnection)freeList.get(elementPos);
		    freeList.remove(elementPos);
		}
	    }

	    // at this point, conn is null iff the freelist is empty,
	    // and nonnull if a free connection of uncertain vitality
	    // has been found.

	    if (conn != null) {
		// check to see if the connection has closed since last use
		if (!conn.isDead()) {
		    if (TCPTransport.logLevel >= LogStream.BRIEF)
			LogStream.log("tcp").println(
			  "TCPChannel.newConnection: reuse connection");

		    return conn;
		}

		// conn is dead, and cannot be reused (reuse => false)
		this.free(conn, false);
	    }
	} while (conn != null);

	// none free, so create a new connection
	return (createConnection());
    }

    /**
     * Create a new connection to the remote endpoint of this channel.
     * The returned connection is new.  The caller must already have
     * passed a security checkConnect or equivalent.
     */
    private Connection createConnection() throws RemoteException {
	Connection conn;

	if (TCPTransport.logLevel >= LogStream.BRIEF)
	    LogStream.log("tcp").println(
		"TCPChannel.newConnection: create connection");

	if (!usingMultiplexer) {
	    Socket sock = ep.newSocket();
	    conn = new TCPConnection(this, sock);

	    try {
		DataOutputStream out =
		    new DataOutputStream(conn.getOutputStream());
		writeTransportHeader(out);

		// choose protocol (single op if not reusable socket)
		if (!conn.isReusable()) {
		    out.writeByte(TransportConstants.SingleOpProtocol);
		} else {
		    out.writeByte(TransportConstants.StreamProtocol);
		    out.flush();
		    DataInputStream in =
			new DataInputStream(conn.getInputStream());
		    byte ack = in.readByte();
		    if (ack != TransportConstants.ProtocolAck)
			throw new MarshalException(
			"Transport protocol not supported by server");

		    String suggestedHost = in.readUTF();
		    int    suggestedPort = in.readInt();
		    if (TCPTransport.logLevel >= LogStream.VERBOSE)
			LogStream.log("tcp").println(
			    "TCPChannel.newConnection: server suggested " +
			    suggestedHost + ":" + suggestedPort);

		    // set local host name, if unknown
		    TCPEndpoint.setLocalHost(suggestedHost);
		    // do NOT set the default port, because we don't
		    // know if we can't listen YET...

		    // write out default endpoint to match protocol
		    // (but it serves no purpose)
		    TCPEndpoint localEp = TCPEndpoint.getLocalEndpoint(0,
			ep.getClientSocketFactory(),
                        ep.getServerSocketFactory());
		    out.writeUTF(localEp.getHost());
		    out.writeInt(localEp.getPort());
		    if (TCPTransport.logLevel >= LogStream.VERBOSE)
			LogStream.log("tcp").println(
			    "TCPChannel.newConnection: using " +
			    localEp.getHost() + ":" + localEp.getPort());

		    out.flush();
		}
	    } catch (IOException e) {
		if (e instanceof RemoteException)
		    throw (RemoteException) e;
		else
		    throw new MarshalException(
			"Error marshaling transport header", e);
	    }
	} else {
	    try {
		conn = multiplexer.openConnection();
	    } catch (IOException e) {
		synchronized (this) {
		    usingMultiplexer = false;
		    multiplexer = null;
		}
		throw new ConnectIOException(
		    "Error creating multiplexed connection", e);
	    }
	}
	return conn;
    }

    /**
     * Free the connection generated by this channel.
     * @param conn The connection
     * @param reuse If true, the connection is in a state in which it
     *        can be reused for another method call.
     */
    public void free(Connection conn, boolean reuse) {
	if (conn == null) return;

	if (reuse && conn.isReusable()) {
	    long lastuse = System.currentTimeMillis();

	    if (TCPTransport.logLevel >= LogStream.BRIEF)
		LogStream.log("tcp").println(
		    "TCPChannel.free: reuse connection");

	    /*
	     * Cache connection; if reaper for expired connection
	     * doesn't exist, then create one.
	     */
	    synchronized (freeList) {
		freeList.add(conn);
		if (reaper == null) {
		    if (TCPTransport.logLevel >= LogStream.BRIEF)
			LogStream.log("tcp").println(
		    		"TCPChannel.free: create reaper");
		    reaper = new Reaper();
		    Thread t = (Thread)
			java.security.AccessController.doPrivileged(
			    new RMIThreadAction(reaper,
						"ConnectionExpiration-" +
						ep.toString(), true));
		    t.start();
		}
	    }
		    
	    ((TCPConnection)conn).setLastUseTime(lastuse);
	    ((TCPConnection)conn).setExpiration(lastuse +
						timeout);
	} else { 
	    if (TCPTransport.logLevel >= LogStream.BRIEF)
		LogStream.log("tcp").println(
		    "TCPChannel.free: close connection");
	    try {
		conn.close();
	    } catch (IOException ignored) {
	    }
	}
    }

    /**
     * Send transport header over stream.
     */
    private void writeTransportHeader(DataOutputStream out)
	throws RemoteException
    {
	try {
	    // write out transport header
	    DataOutputStream dataOut =
		new DataOutputStream(out);
	    dataOut.writeInt(TransportConstants.Magic);
	    dataOut.writeShort(TransportConstants.Version);
	} catch (IOException e) {
	    throw new MarshalException("Error marshaling transport header", e);
	}
    }

    /**
     * Open multiplexed connection to server so that it can use it
     * to open connections back to here.
     */
    private ConnectionMultiplexer openMultiplexer() throws RemoteException {
	Socket sock = ep.newSocket();
	TCPConnection conn = new TCPConnection(this, sock);

	ConnectionMultiplexer mult;	
	try {
	    DataOutputStream out =
		new DataOutputStream(conn.getOutputStream());
	    writeTransportHeader(out);

	    // choose multiplexing protocol (fail if not reusable)
	    if (!conn.isReusable()) {
		throw new ConnectIOException(
		    "Cannot open multiplexed connection to " + ep);
	    } else {
		out.writeByte(TransportConstants.MultiplexProtocol);
		out.flush();
		DataInputStream in =
		    new DataInputStream(conn.getInputStream());
		byte ack = in.readByte();
		if (ack != TransportConstants.ProtocolAck) {
		    throw new MarshalException(
			"Transport protocol not supported by receiver");
		}

		String suggestedHost = in.readUTF();
		int    suggestedPort = in.readInt();
		if (TCPTransport.logLevel >= LogStream.VERBOSE)
		    LogStream.log("tcp").println(
			"TCPChannel.openMultiplexer: server suggested " +
			suggestedHost + ":" + suggestedPort);

		// set default endpoint info, if not already
		TCPEndpoint.setLocalHost(suggestedHost);
		TCPEndpoint.setDefaultPort(suggestedPort);

		TCPEndpoint localEp = TCPEndpoint.getLocalEndpoint(0,
		    ep.getClientSocketFactory(), ep.getServerSocketFactory());
		out.writeUTF(localEp.getHost());
		out.writeInt(localEp.getPort());
		if (TCPTransport.logLevel >= LogStream.VERBOSE)
		    LogStream.log("tcp").println(
			"TCPChannel.openMultiplexer: using " +
			localEp.getHost() + ":" + localEp.getPort());

		out.flush();

		mult = new ConnectionMultiplexer(this, conn.getInputStream(),
		    conn.getOutputStream(), true);
	    }
	} catch (IOException e) {
	    if (e instanceof RemoteException)
		throw (RemoteException) e;
	    else
		throw new MarshalException(
		    "Error marshaling transport header", e);
	}
	return mult;
    }

    /**
     * Execute thread that deals with multiplexed connections.
     */
    public void run() {
	try {
	    multiplexer.run();
	} catch (IOException e) {
	    if (TCPTransport.logLevel >= LogStream.VERBOSE) {
		LogStream ls = LogStream.log("tcp");
		synchronized (ls) {
		    ls.print("exception occurred in multiplexer: ");
		    e.printStackTrace(ls);
		}
	    }
	    multiplexer = null;

	    // try once to reopen multiplex connection??
	    try {
		haveMultiplexer();
	    } catch (RemoteException ex) {
	    }
	}
    }

    /**
     * Make sure that this endpoint has a multiplexed connection to its
     * endpoint to allow back communication.
     */
    void haveMultiplexer() throws RemoteException {
	if (multiplexer == null) {
	    if (TCPTransport.logLevel >= LogStream.VERBOSE)
		LogStream.log("tcp").println(
		    "TCPChannel.haveMultiplexer(): " +
		    "attempting to open multiplex connection");

	    multiplexer = openMultiplexer();
	    Thread t = (Thread) java.security.AccessController.doPrivileged(
		new RMIThreadAction(TCPChannel.this,
				    "Multiplexer-" + ep.getHost() +
				    ":" + ep.getPort(),
				    true));
	    t.start();
	}
    }

    /**
     * Use given connection multiplexer object to obtain new connections
     * through this channel.
     */
    synchronized void useMultiplexer(ConnectionMultiplexer newMultiplexer) {
	// for now, always just use the last one given
	multiplexer = newMultiplexer;

	usingMultiplexer = true;
    }

    /**
     * Accept a connection provided over a multiplexed channel.
     */
    void acceptMultiplexConnection(Connection conn) {
	if (acceptor == null) {
	    acceptor = new ConnectionAcceptor(tr);
	    acceptor.startNewAcceptor();
	}
	acceptor.accept(conn);
    }

    /**
     * Closes all the connections in the cache, whether timed out or not.
     */
    public void shedCache() {
	synchronized (freeList) {
	    ListIterator iter = freeList.listIterator();
	    while (iter.hasNext()) {
		Connection conn = (Connection)iter.next();
		try {
		    conn.close();
		} catch (java.io.IOException e) {
		    // eat exception
		}
	    }
	    freeList.clear();
	}
    }

    private boolean freeCachedConnections() {
	/*
	 * Remove each connection whose time out has expired.
	 */
	synchronized (freeList) {
	    int size = freeList.size();
	    
	    if (size > 0) {
		long time = System.currentTimeMillis();
		ListIterator iter = freeList.listIterator(size);

		while (iter.hasPrevious()) {
		    TCPConnection conn = (TCPConnection)iter.previous();
		    if (conn.expired(time)) {
			if (TCPTransport.logLevel >= LogStream.VERBOSE)
			    LogStream.log("tcp").println(
 				"TCPChannel.freeCachedConnections: " +
				"connection timeout expired");
			try {
			    conn.close();
			} catch (java.io.IOException e) {
			    // eat exception
			}
			iter.remove();
		    }
		}
	    }
	    
	    if (freeList.isEmpty()) {
		reaper = null;
		return false;
	    } else {
		return true;
	    }
	}
    }

    /**
     * Reaps expired cached connections; thread exits when freelist
     * is empty.
     */
    private class Reaper implements Runnable {
	public void run()
	{
	    do {
		try {
		    Thread.sleep(timeout);
		} catch (InterruptedException e) {
		}
		if (TCPTransport.logLevel >= LogStream.VERBOSE)
		    LogStream.log("tcp").println(
		      "TCPChannel.Reaper: " +
		      "wake up");
	    } while (freeCachedConnections());
		if (TCPTransport.logLevel >= LogStream.VERBOSE)
		    LogStream.log("tcp").println(
		      "TCPChannel.Reaper: " +
		      "exit");
	}
    }

    /**
     * Register interest in receiving a callback when this channel
     * becomes inaccessible.
     */
    public void addNotifiable(Endpoint ep, Notifiable obj) {
	synchronized (notifyList) {
	    if (!notifyList.contains(obj))
		notifyList.addElement(obj);
	}
    }

    /**
     * Remove notification request.
     */
    public void removeNotifiable(Endpoint ep, Notifiable obj) {
	notifyList.removeElement(obj);
    }
}

/**
 * ConnectionAcceptor manages accepting new connections and giving them
 * to TCPTransport's message handler on new threads.
 *
 * Since this object only needs to know which transport to give new
 * connections to, it doesn't need to be per-channel as currently
 * implemented.
 */
class ConnectionAcceptor implements Runnable {

    /** transport that will handle message on accepted connections */
    private TCPTransport transport;

    /** queue of connections to be accepted */
    private Vector queue = new Vector(4);

    /** thread ID counter */
    private static int threadNum = 0;

    /**
     * Create a new ConnectionAcceptor that will give connections
     * to the specified transport on a new thread.
     */
    public ConnectionAcceptor(TCPTransport transport) {
	this.transport = transport;
    }

    /**
     * Start a new thread to accept connections.
     */
    public void startNewAcceptor() {
	Thread t = (Thread) java.security.AccessController.doPrivileged(
	    new RMIThreadAction(ConnectionAcceptor.this, 
				"Multiplex Accept-" + ++ threadNum,
				true));
	t.start();
    }

    /**
     * Add connection to queue of connections to be accepted.
     */
    public void accept(Connection conn) {
	synchronized (queue) {
	    queue.addElement(conn);
	    queue.notify();
	}
    }

    /**
     * Give transport next accepted conection, when available.
     */
    public void run() {
	Connection conn;

	synchronized (queue) {
	    while (queue.size() == 0) {
		try {
		    queue.wait();
		} catch (InterruptedException e) {
		}
	    }
	    startNewAcceptor();
	    conn = (Connection) queue.elementAt(0);
	    queue.removeElementAt(0);
	}

	transport.handleMessages(conn, true);
    }
}
