/*
 * @(#)TCPEndpoint.java	1.22 98/07/15
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.rmi.transport.tcp;

import java.io.*;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.*;
import java.rmi.*;
import java.rmi.server.*;
import sun.rmi.transport.*;

/**
 * TCPEndpoint represents some communication endpoint for an address
 * space (VM).
 *
 * @author Ann Wollrath
 */
public class TCPEndpoint implements Endpoint {
    /** IP address or host name */
    private String host;
    /** port number */
    private int port;
    /** custom client socket factory (null if not custom factory) */
    private RMIClientSocketFactory csf = null;
    /** custom server socket factory (null if not custom factory) */
    private RMIServerSocketFactory ssf = null;
    
    /** if local, the transport object associated with this endpoint */
    private TCPTransport thisTransport = null;

    /** the local host name */
    private static String localHost;
    /** true if real local host name is known yet */
    private static boolean localHostKnown;
    /** lock for concurrent access of local host name */
    private static Object localHostLock = new Object();

    // this should be a *private* method since it is privileged
    private static int getInt(String name, int def) {
	return ((Integer)java.security.AccessController.doPrivileged(
                new sun.security.action.GetIntegerAction(name,
							 def))).intValue();
    }

    // this should be a *private* method since it is privileged
    private static boolean getBoolean(String name) {
	return ((Boolean)java.security.AccessController.doPrivileged(
            new sun.security.action.GetBooleanAction(name))).booleanValue();
    }

    /**
     * Find host name of local machine.  Property "java.rmi.server.hostname"
     * is used if set, so server administrator can compensate for the possible
     * inablility to get fully qualified host name from VM.
     */
    static {
	localHostKnown = true;	
	localHost = (String) java.security.AccessController.doPrivileged(
                new sun.security.action.GetPropertyAction("java.rmi.server.hostname"));

	// could try querying CGI program here?
	if (localHost == null) {
    	    try {
		InetAddress localAddr;

		// debug option to simulate restricted applet behavior
		// property "sun.rmi.transport.tcp.simulateApplet":
		//	>= 1	can only get "localhost" for host name
		//	>= 2	listen on socket throws security exception
		if (getInt("sun.rmi.transport.tcp.simulateApplet", 0) >= 1) {
		    localAddr = InetAddress.getByName("");
		} else {
		    localAddr = InetAddress.getLocalHost();		    
		}

		byte[] raw = localAddr.getAddress();
		if ((raw[0] == 127) &&
		    (raw[1] ==   0) &&
		    (raw[2] ==   0) &&
		    (raw[3] ==   1)) {
		    localHostKnown = false;
		}

		/* if the user wishes to use a fully qualified domain
		 * name then attempt to find one.  
		 */
		if (getBoolean("java.rmi.server.useLocalHostName")) {
		    localHost = FQDN.attemptFQDN(localAddr);
		} else {
		    /* default to using ip addresses, names will 
		     * work across seperate domains. 
		     */
		    localHost = localAddr.getHostAddress();
		}
	    } catch (Exception e) {
		localHostKnown = false;
		localHost = null;
	    }
	}

	if (TCPTransport.logLevel >= LogStream.BRIEF)
	    LogStream.log("tcp").println(
		"TCPEndpoint.<clinit>: " +
		"localHostKnown = " + localHostKnown +
		", localHost = " + localHost);
    }

    /** maps an endpoint key containing custom socket factories to
     * their own unique endpoint */
    // TBD: should this be a weak hash table?
    private static Hashtable localEndpoints = new Hashtable(3);

    /**
     * Create an endpoint for a specified host and port.
     * This should not be used by external classes to create endpoints
     * for servers in this VM; use getLocalEndpoint instead.
     */
    public TCPEndpoint(String host, int port) {
	if (host == null)
	    host = "";
	this.host = host;
	this.port = port;
    }

    /**
     * Create a custom socket factory endpoint for a specified host and port.
     * This should not be used by external classes to create endpoints
     * for servers in this VM; use getLocalEndpoint instead.
     */
    public TCPEndpoint(String host, int port, RMIClientSocketFactory csf,
		       RMIServerSocketFactory ssf)
    {
	this(host, port);
	this.csf = csf;
	this.ssf = ssf;
    }

    /**
     * Get an endpoint for the local address space on specified port.
     * If port number is 0, it returns shared default endpoint object
     * whose host name and port may or may not have been determined.
     */
    public static TCPEndpoint getLocalEndpoint(int port) {
	return getLocalEndpoint(port, null, null);
    }

    public static TCPEndpoint getLocalEndpoint(int port,
					       RMIClientSocketFactory csf,
					       RMIServerSocketFactory ssf)
    {
	/*
	 * Find mapping for an endpoint key to the local unique
	 * endpoint for this client/server socket factory pair (perhaps
	 * null) for the specific port.
	 */
	TCPEndpoint ep = null;
	    
	synchronized (localEndpoints) {
	    TCPEndpoint endpointKey = new TCPEndpoint(null, port, csf, ssf);
	    ep = (TCPEndpoint) localEndpoints.get(endpointKey);
	    if (ep == null) {
		ep = new TCPEndpoint(getLocalHost(), port, csf, ssf);
		ep.thisTransport = new TCPTransport(ep);
		localEndpoints.put(endpointKey, ep);

		if (TCPTransport.logLevel >= LogStream.BRIEF)
		    LogStream.log("tcp").println(
			"TCPEndpoint.getLocalEndpoint: " +
			"created local endpoint for socket factory " + ssf +
			" on port " + port);
	    }
	}

	return ep;
    }

    /**
     * Get local host name, if known yet
     */
    static String getLocalHost() {
	synchronized (localHostLock) {
	    return localHost;
	}
    }

    /**
     * Set the local host name, if currently unknown.
     */
    static void setLocalHost(String host) {
	synchronized (localHostLock) {
	    if (!localHostKnown) {
		localHost = host;
		localHostKnown = true;

		if (TCPTransport.logLevel >= LogStream.BRIEF)
		    LogStream.log("tcp").println(
			"TCPEndpoint.setLocalHost: local host set to " + host);
	    } else {
		return;
	    }
	}


	// change the host field of ALL the local endpoints...
	// TBD: this could be eliminated by making TCPEndpoints point
	// to their host names through another level of indirection...
	synchronized (localEndpoints) {
	    Enumeration enum = localEndpoints.elements();
	    while (enum.hasMoreElements()) {
		TCPEndpoint ep = (TCPEndpoint) enum.nextElement();
		ep.host = host;
	    }
	}
    }

    /**
     * Set the port of the (shared) default endpoint object.
     * When first created, it contains port 0 because the transport
     * hasn't tried to listen to get assigned a port, or if listening
     * failed, a port hasn't been assigned from the server.
     */
    static void setDefaultPort(int port) {
	setDefaultPort(port, null, null);
    }

    static void setDefaultPort(int port, RMIClientSocketFactory csf,
			       RMIServerSocketFactory ssf)
    {
	TCPEndpoint ep = getLocalEndpoint(0, csf, ssf);

	synchronized (localEndpoints) {
	    // do nothing if it has already been set
	    if (ep.port != 0)
		return;

	    ep.port = port;
	    // TBD: what if there already is an entry here??
	    TCPEndpoint newEndpointKey = new TCPEndpoint(null, port, csf, ssf);
	    localEndpoints.put(newEndpointKey, ep);

	    if (TCPTransport.logLevel >= LogStream.BRIEF)
		LogStream.log("tcp").println(
		    "TCPEndpoint.setDefaultPort: " +
		    "default port for server socket factory " + ssf +
		    " and client socket factory " + csf +
		    " set to " + port);
	}
    }

    /**
     * Returns transport for remote endpoint.
     */
    public Transport getTransport() {
	return getTransport(false);
    }
    
    /**
     * Returns the transport for this endpoint. 
     * @param local If true, indicates endpoint is for local server,
     * otherwise the endpoint is for a remote server and the default
     * transport (at port 0) should be used.
     */
    private Transport getTransport(boolean local) {
	if (local) {
	    return thisTransport;
	} else {
	    TCPEndpoint localEndpoint = getLocalEndpoint(0, csf, ssf);
	    return localEndpoint.getTransport(true);
	}
    }

    /**
     * Export the object to accept incoming calls.
     */
    public void exportObject(Target target) throws RemoteException {
	Transport tr = getTransport(true);
	tr.exportObject(target);
    }

    /**
     * Returns a channel for this (remote) endpoint.
     */
    public Channel getChannel() {
	Transport tr = getTransport(false);
	return tr.getChannel(this);
    }

    /**
     * Returns address for endpoint
     */
    public String getHost() {
	return host;
    }

    /**
     * Returns the port for this endpoint.  Note that the port may be
     * zero which indicates the default port for the VM for which this
     * is an endpoint.
     */
    public int getPort() {
	return port;
    }

    /**
     * Get the client socket factory associated with this endpoint.
     */
    public RMIClientSocketFactory getClientSocketFactory() {
	return csf;
    }

    /**
     * Get the server socket factory associated with this endpoint.
     */
    public RMIServerSocketFactory getServerSocketFactory() {
	return ssf;
    }
    
    /**
     * Return string representation for endpoint.
     */
    public String toString() {
	return "[" + host + ":" + port +
	    (ssf != null ? "," + ssf : "") +
	    (csf != null ? "," + csf : "") +
	    "]";
    }

    public int hashCode() {
	return port;
    }

    public boolean equals(Object obj) {
	if ((obj != null) && (obj instanceof TCPEndpoint)) {
	    TCPEndpoint ep = (TCPEndpoint)obj;
	    if (port != ep.port || !host.equals(ep.host))
		return false;
	    if (((csf == null) ^ (ep.csf == null)) ||
		((ssf == null) ^ (ep.ssf == null)))
		return false;
	    if (csf != null && !(csf.equals(ep.csf)))
		return false;
	    if (ssf != null && !(ssf.equals(ep.ssf)))
		return false;
	    return true;
	} else {
	    return false;
	}
    }

    /* codes for the self-describing formats of wire representation */
    private static final int FORMAT_HOST_PORT 		= 0;
    private static final int FORMAT_HOST_PORT_FACTORY	= 1;

    /**
     * Write endpoint to output stream.
     */
    public void write(ObjectOutput out) throws IOException {
	if (csf == null) {
	    out.writeByte(FORMAT_HOST_PORT);
	    out.writeUTF(host);
	    out.writeInt(port);
	} else {
	    out.writeByte(FORMAT_HOST_PORT_FACTORY);
	    out.writeUTF(host);
	    out.writeInt(port);
	    out.writeObject(csf);
	}
    }
    
    /**
     * Get the endpoint from the input stream.
     * @param in the input stream
     * @exception IOException If id could not be read (due to stream failure)
     */
    public static TCPEndpoint read(ObjectInput in)
	throws IOException, ClassNotFoundException
    {
	String host;
	int port;
	RMIClientSocketFactory csf = null;

	byte format = in.readByte();
	switch (format) {
	  case FORMAT_HOST_PORT:
	    host = in.readUTF();
	    port = in.readInt();
	    break;

	  case FORMAT_HOST_PORT_FACTORY:
	    host = in.readUTF();
	    port = in.readInt();
	    csf = (RMIClientSocketFactory)in.readObject();
	  break;

	  default:
	    throw new IOException("invalid endpoint format");
	}
	return new TCPEndpoint(host, port, csf, null);
    }

    /**
     * Write endpoint to output stream in older format used by
     * UnicastRef for JDK1.1 compatibility.
     */
    public void writeHostPortFormat(DataOutput out) throws IOException {
	if (csf != null) {
	    throw new InternalError("TCPEndpoint.writeHostPortFormat: " +
		"called for endpoint with non-null socket factory");
	}
	out.writeUTF(host);
	out.writeInt(port);
    }

    /**
     * Create a new endpoint from input stream data.
     * @param in the input stream
     */
    public static TCPEndpoint readHostPortFormat(DataInput in)
	throws IOException
    {
	String host = in.readUTF();
	int port = in.readInt();
	return new TCPEndpoint(host, port);
    }

    private static RMISocketFactory chooseFactory() {
	RMISocketFactory sf = RMISocketFactory.getSocketFactory();
	if (sf == null) {
	    sf = TCPTransport.defaultSocketFactory;
	}
	return sf;
    }

    /**
     * Open and return new client socket connection to endpoint.
     */
    Socket newSocket() throws RemoteException {
	if (TCPTransport.logLevel >= LogStream.VERBOSE)
	    LogStream.log("tcp").println(
		"TCPEndpoint.newSocket: opening socket to " + this);

	Socket socket;
	
	try {
	    RMIClientSocketFactory clientFactory = csf;
	    if (clientFactory == null) {
		clientFactory = chooseFactory();
	    }
	    socket = clientFactory.createSocket(host, port);
	    
	} catch (java.net.UnknownHostException e) {
	    throw new UnknownHostException("Unknown host: " + host, e);
	} catch (java.net.ConnectException e) {
	    throw new ConnectException("Connection refused to host: " +
		host, e);
	} catch (IOException e) {
	    throw new ConnectIOException("Exception creating connection to: " +
		host, e);
	}

	// set socket to disable Nagle's algorithm (always send immediately)
	// TBD: should this be left up to socket factory instead?
	try {
	    socket.setTcpNoDelay(true);
	} catch (Exception e) {
	    // if we fail to set this, ignore and proceed anyway
	}

	return socket;
    }

    /**
     * Return new server socket to listen for connections on this endpoint.
     */
    ServerSocket newServerSocket() throws IOException {
	if (TCPTransport.logLevel >= LogStream.VERBOSE)
	    LogStream.log("tcp").println(
		"TCPEndpoint.newServerSocket: creating server socket on " +
		this);

	// debug option to simulate restricted applet behavior
	// property "sun.rmi.transport.tcp.simulateApplet":
	//	>= 1	can only get "localhost" for host name
	//	>= 2	listen on socket throws security exception
	if (getInt("sun.rmi.transport.tcp.simulateApplet", 0) >= 2)
	    throw new sun.applet.AppletSecurityException("socket.listen");

	RMIServerSocketFactory serverFactory = ssf;
	if (serverFactory == null) {
	    serverFactory = chooseFactory();
	}
	ServerSocket server = serverFactory.createServerSocket(port);

	// if we listened on an anonymous port, set the default port
	// (for this socket factory)
	if (port == 0)
	    setDefaultPort(server.getLocalPort(), csf, ssf);

	return server;
    }

    /**
     * The class FQDN encapsulates a routine that makes a best effort
     * attempt to retrieve the fully qualified domain name of the local
     * host.
     *
     * @author 	Laird Dornin
     */
    private static class FQDN implements Runnable {
	
	/**
	 * strings in which we can store discovered fqdn
	 */
	private String reverseLookup;

	private String hostAddress;

	private FQDN(String hostAddress) {
	    this.hostAddress = hostAddress;
	}

	/**
	 * Do our best to obtain a fully qualified hostname for the local
	 * host.  Perform the following steps to get a localhostname:
	 * 
	 * 1. InetAddress.getLocalHost().getHostName() - if contains
	 *    '.' use as FQDN
	 * 2. if no '.' query name service for FQDN in a thread
	 *    Note: We query the name service for an FQDN by creating 
	 *    an InetAddress via a stringified copy of the local ip 
	 *    address; this creates an InetAddress with a null hostname.
	 *    Asking for the hostname of this InetAddress causes a name
	 *    service lookup.
	 *
	 * 3. if name service takes too long to return, use ip address
	 * 4. if name service returns but response contains no '.' 
	 *    default to ipaddress.
	 */
	static String attemptFQDN(InetAddress localAddr) 
	    throws java.net.UnknownHostException
	{
	    
	    String hostName = localAddr.getHostName();
	    
	    if (hostName.indexOf('.') < 0 ) {
		
		String hostAddress = localAddr.getHostAddress();
		FQDN f = new FQDN(hostAddress);
		
		int nameServiceTimeOut = 
		    TCPEndpoint.getInt("sun.rmi.transport.tcp.localHostNameTimeOut",
				       10000);

		try {
		    synchronized(f) {
			f.getFQDN();

			/* wait to obtain an FQDN */
			f.wait(nameServiceTimeOut);
		    }
		} catch (InterruptedException e) {
		    /* propagate the exception to the caller */
		    Thread.currentThread().interrupt();
		}
		hostName = f.getHost();

		if ((hostName == null) || (hostName.equals("")) 
		    || (hostName.indexOf('.') < 0 )) {
		
		    hostName = hostAddress;
		}
	    }
	    return hostName;
	}
	
	/** 
	 * Method that that will start a thread to wait to retrieve a
	 * fully qualified domain name from a name service.  The spawned
	 * thread may never return but we have marked it as a daemon so the vm
	 * will terminate appropriately.  
	 */
	private void getFQDN() {

	    /* FQDN finder will run in RMI threadgroup. */	    
	    Thread t = (Thread)
		java.security.AccessController.doPrivileged(
		    new RMIThreadAction(FQDN.this, "FQDN Finder", true));
	    t.start();
	}
	
	private synchronized String getHost() {
	    return reverseLookup;
	}

	/**
	 * thread to query a name service for the fqdn of this host.
	 */
	public void run()  {

	    String name = null;

	    try {
		name = InetAddress.getByName(hostAddress).getHostName();
	    } catch (java.net.UnknownHostException e) {
	    } finally {
		synchronized(this) {
		    reverseLookup = name;
		    this.notify();
		}
	    }
	}
    }
}
