/*
 * @(#)DSAParameters.java	1.10 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.security.provider;

import java.util.*;
import java.io.*;
import java.math.BigInteger;
import java.security.AlgorithmParametersSpi;
import java.security.spec.AlgorithmParameterSpec;
import java.security.spec.DSAParameterSpec;
import java.security.spec.InvalidParameterSpecException;

import sun.security.util.DerValue;
import sun.security.util.DerOutputStream;
import sun.security.util.BigInt;

/**
 * This class implements the parameter set used by the
 * Digital Signature Algorithm as specified in the FIPS 186
 * standard.
 *
 * @author Jan Luehe
 *
 * @version 1.10, 98/12/01
 *
 * @since JDK1.2
 */

public class DSAParameters extends AlgorithmParametersSpi {

    // the prime (p)
    protected BigInteger p;

    // the sub-prime (q)
    protected BigInteger q;

    // the base (g)
    protected BigInteger g;

    protected void engineInit(AlgorithmParameterSpec paramSpec) 
	throws InvalidParameterSpecException {
	    if (!(paramSpec instanceof DSAParameterSpec)) {
		throw new InvalidParameterSpecException
		    ("Inappropriate parameter specification");
	    }
	    this.p = ((DSAParameterSpec)paramSpec).getP();
	    this.q = ((DSAParameterSpec)paramSpec).getQ();
	    this.g = ((DSAParameterSpec)paramSpec).getG();
    }

    protected void engineInit(byte[] params) throws IOException {
	DerValue encodedParams = new DerValue(params);

	if (encodedParams.tag != DerValue.tag_Sequence) {
	    throw new IOException("DSA params parsing error");
	}

	encodedParams.data.reset();

	this.p = encodedParams.data.getInteger().toBigInteger();
	this.q = encodedParams.data.getInteger().toBigInteger();
	this.g = encodedParams.data.getInteger().toBigInteger();

	if (encodedParams.data.available() != 0) {
	    throw new IOException("encoded params have " +
				  encodedParams.data.available() +
				  " extra bytes");
	}
    }

    protected void engineInit(byte[] params, String decodingMethod)
	throws IOException {
	    engineInit(params);
    }

    protected AlgorithmParameterSpec engineGetParameterSpec(Class paramSpec)
	throws InvalidParameterSpecException {	    
	    try {
		Class dsaParamSpec = Class.forName
		    ("java.security.spec.DSAParameterSpec");
		if (dsaParamSpec.isAssignableFrom(paramSpec)) {
		    return new DSAParameterSpec(this.p, this.q, this.g);
		} else {
		    throw new InvalidParameterSpecException
			("Inappropriate parameter Specification");
		}
	    } catch (ClassNotFoundException e) {
		throw new InvalidParameterSpecException
		    ("Unsupported parameter specification: " + e.getMessage());
	    }
    }

    protected byte[] engineGetEncoded() throws IOException {
	DerOutputStream	out = new DerOutputStream();
	DerOutputStream bytes = new DerOutputStream();

	bytes.putInteger(new BigInt(p.toByteArray()));
	bytes.putInteger(new BigInt(q.toByteArray()));
	bytes.putInteger(new BigInt(g.toByteArray()));
	out.write(DerValue.tag_Sequence, bytes);
	return out.toByteArray();
    }

    protected byte[] engineGetEncoded(String encodingMethod)
	throws IOException {
	    return engineGetEncoded();
    }

    /*
     * Returns a formatted string describing the parameters.
     */
    protected String engineToString() {
	return "\n\tp: " + new BigInt(p).toString()
	    + "\n\tq: " + new BigInt(q).toString()
	    + "\n\tg: " + new BigInt(g).toString()
	    + "\n";
    }
}
