/*
 * @(#)DSAPublicKey.java	1.54 98/09/21
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.security.provider;

import java.util.*;
import java.io.*;
import java.math.BigInteger;
import java.security.InvalidKeyException;
import java.security.ProviderException;
import java.security.AlgorithmParameters;
import java.security.spec.DSAParameterSpec;
import java.security.spec.InvalidParameterSpecException;
import java.security.interfaces.DSAParams;

import sun.security.x509.X509Key;
import sun.security.x509.AlgIdDSA;
import sun.security.util.BigInt;
import sun.security.util.DerValue;
import sun.security.util.DerInputStream;
import sun.security.util.DerOutputStream;

/**
 * An X.509 public key for the Digital Signature Algorithm.
 *
 * @author Benjamin Renaud 
 *
 * @version 1.54, 98/12/01
 *
 * @see DSAPrivateKey
 * @see AlgIdDSA
 * @see DSA
 */

public final class DSAPublicKey extends X509Key 
implements java.security.interfaces.DSAPublicKey, Serializable {

    /** use serialVersionUID from JDK 1.1. for interoperability */
    private static final long serialVersionUID = -2994193307391104133L;

    /* the public key */
    private BigInteger y;

    /*
     * Keep this constructor for backwards compatibility with JDK1.1. 
     */
    public DSAPublicKey() {
    }

    /**
     * Make a DSA public key out of a public key and three parameters.
     */
    public DSAPublicKey(BigInteger y, BigInteger p, BigInteger q,
			BigInteger g) 
    throws InvalidKeyException {
	this.y = y;
	algid = new AlgIdDSA(p, q, g);

	try {
	    key = new DerValue(DerValue.tag_Integer, 
			       y.toByteArray()).toByteArray();
	    encode();
	} catch (IOException e) {
	    throw new InvalidKeyException("could not DER encode y: " +
					  e.getMessage());
	}
    }

    /**
     * Make a DSA public key from its DER encoding (X.509).
     */
    public DSAPublicKey(byte[] encoded) throws InvalidKeyException {
	decode(encoded);
    }

    /**
     * Returns the DSA parameters associated with this key, or null if the
     * parameters could not be parsed.
     */
    public DSAParams getParams() {
	try {
	    if (algid instanceof DSAParams) {
		return (DSAParams)algid;
	    } else {
		DSAParameterSpec paramSpec;
		AlgorithmParameters algParams = algid.getParameters();
		if (algParams == null) {
		    return null;
		}
		paramSpec = (DSAParameterSpec)algParams.getParameterSpec
		    (DSAParameterSpec.class);
		return (DSAParams)paramSpec;
	    }
	} catch (InvalidParameterSpecException e) {
	    return null;
	}
    }
                                                                   
    /**
     * Get the raw public value, y, without the parameters.
     *
     * @see getParameters
     */
    public BigInteger getY() {
	return y;
    }

    public String toString() {
        return "Sun DSA Public Key\n    Parameters:" + algid
            + "\n  y:\n" + (new BigInt(y)).toString() + "\n";
    }

    protected void parseKeyBits() throws InvalidKeyException {
	try {
	    DerInputStream in = new DerInputStream(key);
	    y = in.getInteger().toBigInteger();
	} catch (IOException e) {
	    throw new InvalidKeyException("Invalid key: y value\n" +
					  e.getMessage());
	}
    }
}
