/*
 * @(#)Sun.java	1.26 98/08/03
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.security.provider;

import java.io.*;
import java.util.*;
import java.security.*;

/**
 * The SUN Security Provider.
 *
 * @author Benjamin Renaud 
 *
 * @version 1.26, 98/12/01
 */

/**
 * Defines the SUN provider.
 *
 * Algorithms supported, and their names:
 *
 * - SHA is the message digest scheme described in FIPS 180-1. 
 *   Aliases for SHA are SHA-1 and SHA1.
 *
 * - SHA1withDSA is the signature scheme described in FIPS 186.
 *   (SHA used in DSA is SHA-1: FIPS 186 with Change No 1.)
 *   Aliases for SHA1withDSA are DSA, DSS, SHA/DSA, SHA-1/DSA, SHA1/DSA,
 *   SHAwithDSA, DSAWithSHA1, and the object
 *   identifier strings "OID.1.3.14.3.2.13", "OID.1.3.14.3.2.27" and
 *   "OID.1.2.840.10040.4.3".
 *
 * - DSA is the key generation scheme as described in FIPS 186.
 *   Aliases for DSA include the OID strings "OID.1.3.14.3.2.12"
 *   and "OID.1.2.840.10040.4.1".
 *
 * - MD5 is the message digest scheme described in RFC 1321.
 *   There are no aliases for MD5.
 */

public final class Sun extends Provider {

    private static final String INFO = "SUN " + 
    "(DSA key/parameter generation; DSA signing; " +
    "SHA-1, MD5 digests; SecureRandom; X.509 certificates; JKS keystore)";

    public Sun() {
	/* We are the SUN provider */
	super("SUN", 1.2, INFO);

	AccessController.doPrivileged(new java.security.PrivilegedAction() {
	    public Object run() {

	        /*
	         * Signature engines 
	         */
	        put("Signature.SHA1withDSA", "sun.security.provider.DSA");
	    
	        put("Alg.Alias.Signature.DSA", "SHA1withDSA");
	        put("Alg.Alias.Signature.DSS", "SHA1withDSA");
	        put("Alg.Alias.Signature.SHA/DSA", "SHA1withDSA");
	        put("Alg.Alias.Signature.SHA-1/DSA", "SHA1withDSA");
	        put("Alg.Alias.Signature.SHA1/DSA", "SHA1withDSA");
	        put("Alg.Alias.Signature.SHAwithDSA", "SHA1withDSA");
	        put("Alg.Alias.Signature.DSAWithSHA1", "SHA1withDSA");
	        put("Alg.Alias.Signature.OID.1.2.840.10040.4.3",
		    "SHA1withDSA");
	        put("Alg.Alias.Signature.1.2.840.10040.4.3", "SHA1withDSA");
	        put("Alg.Alias.Signature.1.3.14.3.2.13", "SHA1withDSA");
	        put("Alg.Alias.Signature.1.3.14.3.2.27", "SHA1withDSA");
	    
	        /*
	         *  Key Pair Generator engines 
	         */
                put("KeyPairGenerator.DSA", 
	            "sun.security.provider.DSAKeyPairGenerator");
                put("Alg.Alias.KeyPairGenerator.OID.1.2.840.10040.4.1", "DSA");
                put("Alg.Alias.KeyPairGenerator.1.2.840.10040.4.1", "DSA");
                put("Alg.Alias.KeyPairGenerator.1.3.14.3.2.12", "DSA");

	        /* 
	         * Digest engines 
	         */
	        put("MessageDigest.MD5", "sun.security.provider.MD5");
	        put("MessageDigest.SHA", "sun.security.provider.SHA");
	
	        put("Alg.Alias.MessageDigest.SHA-1", "SHA");
	        put("Alg.Alias.MessageDigest.SHA1", "SHA");

		/*
		 * Algorithm Parameter Generator engines
		 */
		put("AlgorithmParameterGenerator.DSA",
		    "sun.security.provider.DSAParameterGenerator");

		/*
		 * Algorithm Parameter engines
		 */
		put("AlgorithmParameters.DSA",
		    "sun.security.provider.DSAParameters");
		put("Alg.Alias.AlgorithmParameters.1.3.14.3.2.12", "DSA");
		put("Alg.Alias.AlgorithmParameters.1.2.840.10040.4.1", "DSA");

	        /*
	         * Key factories
	         */
	        put("KeyFactory.DSA", "sun.security.provider.DSAKeyFactory");
                put("Alg.Alias.KeyFactory.1.3.14.3.2.12", "DSA");
                put("Alg.Alias.KeyFactory.1.2.840.10040.4.1", "DSA");

	        /*
	         * SecureRandom
	         */
	         put("SecureRandom.SHA1PRNG",
		     "sun.security.provider.SecureRandom");

		/*
		 * Certificates
		 */
		put("CertificateFactory.X509",
		    "sun.security.provider.X509Factory");
		put("Alg.Alias.CertificateFactory.X.509", "X509");

		/*
		 * KeyStore
		 */
		put("KeyStore.JKS", "sun.security.provider.JavaKeyStore");

		return null;
	    }
	});
    }
}
