/*
 * @(#)KeyTool.java	1.49 98/09/04
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.security.tools;

import java.io.*;
import java.security.GeneralSecurityException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.Key;
import java.security.PublicKey;
import java.security.PrivateKey;
import java.security.Security;
import java.security.Signature;
import java.security.SignatureException;
import java.security.UnrecoverableKeyException;
import java.security.Principal;
import java.security.Identity;
import java.security.Signer;
import java.security.cert.Certificate;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.security.cert.CertificateException;
import java.util.*;

import sun.misc.BASE64Decoder;
import sun.misc.BASE64Encoder;
import sun.security.x509.AlgorithmId;
import sun.security.x509.X509CertImpl;
import sun.security.x509.X509CertInfo;
import sun.security.x509.X500Name;
import sun.security.x509.CertAndKeyGen;
import sun.security.x509.CertificateSubjectName;
import sun.security.x509.CertificateIssuerName;
import sun.security.x509.CertificateValidity;
import sun.security.x509.CertificateSerialNumber;
import sun.security.x509.CertificateAlgorithmId;
import sun.security.x509.X509Key;
import sun.security.x509.X500Signer;
import sun.security.pkcs.PKCS10;
import sun.security.provider.IdentityDatabase;
import sun.security.provider.SystemSigner;
import sun.security.provider.SystemIdentity;
import sun.security.provider.X509Factory;
import sun.security.util.DerOutputStream;

/**
 * This tool manages keystores.
 *
 * @author Jan Luehe
 *
 * @version 1.49, 09/04/98
 *
 * @see java.security.KeyStore
 * @see sun.security.provider.KeyProtector
 * @see sun.security.provider.JavaKeyStore
 *
 * @since JDK1.2
 */

public final class KeyTool {

    private boolean debug = false;
    private String command = null;
    private String sigAlgName = null;
    private String keyAlgName = "DSA";
    private boolean verbose = false;
    private int	keysize = 1024;
    private boolean rfc = false;
    private int	validity = 90;
    private String alias = null;
    private String dname = null;
    private String keyAlias = "mykey";
    private String dest = null;
    private String filename = null;
    private String storetype = null;
    private char[] storePass = null;
    private char[] storePassNew = null;
    private char[] keyPass = null;
    private char[] keyPassNew = null;
    private char[] oldPass = null;
    private char[] newPass = null;
    private String ksfname = null;
    private File ksfile = null;
    private InputStream ksStream = null; // keystore stream
    private InputStream inStream = null;
    private KeyStore keyStore = null;
    private boolean kssave = false;
    private boolean noprompt = false;
    private boolean trustcacerts = false;
    private CertificateFactory cf = null;

    private KeyTool() { }

    public static void main(String[] args) {
	KeyTool kt = new KeyTool();
	kt.run(args, System.out);
    }

    public void run(String[] args, PrintStream out) {
	try {
	    parseArgs(args);
	    doCommands(out);
	} catch (Exception e) {
	    System.out.println("keytool error: " + e.getMessage());
	    if (debug) {
		e.printStackTrace();
	    }
	    System.exit(1);
	} finally {
	    // zero-out passwords
	    if (storePass != null) {
		Arrays.fill(storePass, ' ');
		storePass = null;
	    }
	    if (storePassNew != null) {
		Arrays.fill(storePassNew , ' ');
		storePassNew = null;
	    }
	    if (keyPass != null) {
		Arrays.fill(keyPass, ' ');
		keyPass = null;
	    }
	    if (keyPassNew != null) {
		Arrays.fill(keyPassNew, ' ');
		keyPassNew = null;
	    }
	    if (oldPass != null) {
		Arrays.fill(oldPass, ' ');
		oldPass = null;
	    }
	    if (newPass != null) {
		Arrays.fill(newPass, ' ');
		newPass = null;
	    }
	}
    }

    /**
     * Parse command line arguments.
     */
    void parseArgs(String[] args) {

	if (args.length == 0) usage();

	int i=0;

	for (i=0; (i < args.length) && args[i].startsWith("-"); i++) {

	    String flags = args[i];
	    /*
	     * command modes
	     */
	    if (flags.equalsIgnoreCase("-certreq")) {
		command = "certreq";
	    } else if (flags.equalsIgnoreCase("-delete")) {
		command = "delete";
	    } else if (flags.equalsIgnoreCase("-export")) {
		command = "export";
	    } else if (flags.equalsIgnoreCase("-genkey")) {
		command = "genkey";
	    } else if (flags.equalsIgnoreCase("-help")) {
		usage();
		return;
	    } else if (flags.equalsIgnoreCase("-identitydb")) {
		command = "identitydb";
	    } else if (flags.equalsIgnoreCase("-import")) {
		command = "import";
	    } else if (flags.equalsIgnoreCase("-keyclone")) {
		command = "keyclone";
	    } else if (flags.equalsIgnoreCase("-keypasswd")) {
		command = "keypasswd";
	    } else if (flags.equalsIgnoreCase("-list")) {
		command = "list";
	    } else if (flags.equalsIgnoreCase("-printcert")) {
		command = "printcert";
	    } else if (flags.equalsIgnoreCase("-selfcert")) {
		command = "selfcert";
	    } else if (flags.equalsIgnoreCase("-storepasswd")) {
		command = "storepasswd";
	    }

	    /*
	     * specifiers
	     */
	    else if (flags.equalsIgnoreCase("-keystore")) {
		if (++i == args.length) usage();
		ksfname = args[i];
	    } else if (flags.equalsIgnoreCase("-storepass")) {
		if (++i == args.length) usage();
		storePass = args[i].toCharArray();
	    } else if (flags.equalsIgnoreCase("-storetype")) {
		if (++i == args.length) usage();
		storetype = args[i];
	    } else if (flags.equalsIgnoreCase("-keypass")) {
		if (++i == args.length) usage();
		keyPass = args[i].toCharArray();
	    } else if (flags.equalsIgnoreCase("-new")) {
		if (++i == args.length) usage();
		newPass = args[i].toCharArray();
	    } else if (flags.equalsIgnoreCase("-alias")) {
		if (++i == args.length) usage();
		alias = args[i];
	    } else if (flags.equalsIgnoreCase("-dest")) {
		if (++i == args.length) usage();
		dest = args[i];
	    } else if (flags.equalsIgnoreCase("-dname")) {
		if (++i == args.length) usage();
		dname = args[i];
	    } else if (flags.equalsIgnoreCase("-keysize")) {
		if (++i == args.length) usage();
		keysize = Integer.parseInt(args[i]);
	    } else if (flags.equalsIgnoreCase("-keyalg")) {
		if (++i == args.length) usage();
		keyAlgName = args[i];
	    } else if (flags.equalsIgnoreCase("-sigalg")) {
		if (++i == args.length) usage();
		sigAlgName = args[i];
	    } else if (flags.equalsIgnoreCase("-validity")) {
		if (++i == args.length) usage();
		validity = Integer.parseInt(args[i]);
	    } else if (flags.equalsIgnoreCase("-file")) {
		if (++i == args.length) usage();
		filename = args[i];
	    }

	    /*
	     * options
	     */
	    else if (flags.equalsIgnoreCase("-v")) {
		verbose = true;
	    } else if (flags.equalsIgnoreCase("-debug")) {
		debug = true;
	    } else if (flags.equalsIgnoreCase("-rfc")) {
		rfc = true;
	    } else if (flags.equalsIgnoreCase("-noprompt")) {
		noprompt = true;
	    } else if (flags.equalsIgnoreCase("-trustcacerts")) {
		trustcacerts = true;
	    } else  {
		System.err.println("Illegal option:  " + flags);
		usage();
	    }
	}

	if (i<args.length || command==null) usage();
    }

    /**
     * Execute the commands.
     */
    void doCommands(PrintStream out) throws Exception {

	if (command.equals("list") && verbose && rfc) {
	    System.err.println("Must not specify both -v and -rfc "
			       + "with 'list' command");
	    usage();
	}

	if (command.equals("keyclone") && dest == null) {
	    dest = getAlias("destination");
	    if (dest.equals("")) {
		throw new Exception("Must specify destination alias");
	    }
	}

	if (command.equals("delete") && alias == null) {
	    alias = getAlias(null);
	    if (alias.equals("")) {
		throw new Exception("Must specify alias");
	    }
	}

	// Make sure provided passwords are at least 6 characters long
	if (storePass != null && storePass.length < 6) {
	    throw new Exception("Keystore password must be at least 6 "
				+ "characters");
	}
	if (keyPass != null && keyPass.length < 6) {
	    throw new Exception("Key password must be at least 6 characters");
	}
	if (newPass != null && newPass.length < 6) {
	    throw new Exception("New password must be at least 6 characters");
	}
	
	// Check if keystore exists.
	// If no keystore has been specified at the command line, try to use
	// the default, which is located in $HOME/.keystore.
	// If the command is "genkey", "identitydb", "import", or "printcert",
	// it is OK not to have a keystore. 
	if (!command.equals("printcert")) {
	    if (ksfname == null) {
		ksfname = System.getProperty("user.home") + File.separator
		    + ".keystore";
	    }
	    try {
		ksfile = new File(ksfname);
		// Check if keystore file is empty
		if (ksfile.exists() && ksfile.length() == 0) {
		    throw new Exception("Keystore file exists, but is "
					+ "empty: " + ksfname);
		}
		ksStream = new FileInputStream(ksfile);
	    } catch (FileNotFoundException e) {
		if (!command.equals("genkey") && !command.equals("identitydb")
		    && !command.equals("import")) {
		    throw new Exception("Keystore file does not exist: " +
					ksfname);
		}
	    }
	}

	// Create new keystore
	if (storetype != null)
	    keyStore = KeyStore.getInstance(storetype);
	else
	    keyStore = KeyStore.getInstance(KeyStore.getDefaultType());

	/*
	 * Load the keystore data.
	 *
	 * At this point, it's OK if no keystore password has been provided.
	 * We want to make sure that we can load the keystore data, i.e.,
	 * the keystore data has the right format. If we cannot load the
	 * keystore, why bother asking the user for his or her password?
	 * Only if we were able to load the keystore, and no keystore
	 * password has been provided, we will prompt the user for the
	 * keystore password to verify the keystore integrity.
	 * This means that the keystore is loaded twice: first load operation
	 * checks the keystore format, second load operation verifies the
	 * keystore integrity.
	 *
	 * If the keystore password has already been provided (at the
	 * command line), however, the keystore is loaded only once, and the
	 * keystore format and integrity are checked "at the same time".
	 */
	keyStore.load(ksStream, storePass);
	if (ksStream != null)
	    ksStream.close();

	// All commands that create or modify the keystore require a keystore
	// password.
	if (storePass == null) {

	    if (command.equals("certreq") || command.equals("delete")
		|| command.equals("genkey") || command.equals("import")
		|| command.equals("keyclone") || command.equals("selfcert")
		|| command.equals("storepasswd")
		|| command.equals("keypasswd")
		|| command.equals("identitydb")) {
		int count = 0;
		do {
		    System.err.print("Enter keystore password:  ");
		    System.err.flush();
		    storePass = readPasswd(System.in);

		    // If we are creating a new keystore, insist that the
		    // password be at least 6 characters
		    if ((ksStream == null) && (storePass.length < 6)) {
			System.err.println("Keystore password is too short - "
					   + "must be at least 6 characters");
			storePass = null;
		    }
		    count++;
		} while ((storePass == null) && count < 3);
		if (storePass == null) {
		    System.err.println("Too many failures - try later");
		    return;
		}
	    } else if (!(command.equals("printcert"))) {
		System.err.print("Enter keystore password:  ");
		System.err.flush();
		storePass = readPasswd(System.in);
	    }

	    // Now verify the integrity of the keystore
	    if (ksStream != null) {
		ksStream = new FileInputStream(ksfile);
		keyStore.load(ksStream, storePass);
		ksStream.close();
	    }

	}

	// Create a certificate factory
	if (command.equals("printcert") || command.equals("import")
	       || command.equals("identitydb")) {
	    cf = CertificateFactory.getInstance("X509");
	}

	// Perform the specified command
	if (command.equals("certreq")) {
	    if (filename != null) {
		PrintStream ps = new PrintStream(new FileOutputStream
						 (filename));
		out = ps;
	    }
	    doCertReq(alias, sigAlgName, out);
	    if (verbose && filename != null) {
		System.err.println("Certification request stored in file <" +
				   filename + ">");
		System.err.println("Submit this to your CA");
	    }
	} else if (command.equals("delete")) {
	    doDeleteEntry(alias);
	    kssave = true;
	} else if (command.equals("export")) {
	    if (filename != null) {
		PrintStream ps = new PrintStream(new FileOutputStream
						 (filename));
		out = ps;
	    }
	    doExportCert(alias, out);
	    if (filename != null) {
		System.err.println("Certificate stored in file <" + filename
				   + ">");
	    }
	} else if (command.equals("genkey")) {
	    doGenKeyPair(alias, dname, keyAlgName, keysize, sigAlgName);
	    kssave = true;
	} else if (command.equals("identitydb")) {
	    InputStream inStream = System.in;
	    if (filename != null) {
		inStream = new FileInputStream(filename);
	    }
	    doImportIdentityDatabase(inStream);
	} else if (command.equals("import")) {
	    InputStream inStream = System.in;
	    if (filename != null) {
		inStream = new FileInputStream(filename);
	    }
	    String importAlias = (alias!=null)?alias:keyAlias;
	    if (keyStore.isKeyEntry(importAlias) == true) { 
		kssave = installReply(importAlias, inStream);
		if (kssave)
		    System.err.println("Certificate reply was installed "
				       + "in keystore");
		else
		    System.err.println("Certificate reply was not installed "
				       + "in keystore");
	    } else {
		kssave = addTrustedCert(importAlias, inStream);
		if (kssave)
		    System.err.println("Certificate was added to keystore");
		else
		    System.err.println("Certificate was not added to "
				       + "keystore");
	    }
	} else if (command.equals("keyclone")) {
	    keyPassNew = newPass;
	    doCloneKey(alias, dest);
	    kssave = true;
	} else if (command.equals("keypasswd")) {
	    keyPassNew = newPass;
	    doChangeKeyPasswd(alias);
	    kssave = true;
	} else if (command.equals("list")) {
	    if (alias != null) {
		doPrintEntry(alias, out, true);
	    } else {
		doPrintEntries(out);
	    }
	} else if (command.equals("printcert")) {
	    InputStream inStream = System.in;
	    if (filename != null) {
		inStream = new FileInputStream(filename);
	    }
	    doPrintCert(inStream, out);
	} else if (command.equals("selfcert")) {
	    doSelfCert(alias, dname, sigAlgName);
	    kssave = true;
	} else if (command.equals("storepasswd")) {
	    storePassNew = newPass;
	    if (storePassNew == null) {
		storePassNew = getNewPasswd("keystore password", storePass);
	    }
	    kssave = true;
	}

	// If we need to save the keystore, do so.
	if (kssave) {
	    if (verbose) {
		System.err.println("[Saving " + ksfname + "]");
	    }
	    FileOutputStream fout = new FileOutputStream(ksfname);
	    keyStore.store(fout, (storePassNew!=null)?storePassNew:storePass);
	    fout.close();
	}
    }

    /**
     * Creates a PKCS#10 cert signing request, corresponding to the
     * keys (and name) associated with a given alias.
     */
    private void doCertReq(String alias, String sigAlgName, PrintStream out)
	throws Exception
    {
	if (alias == null) {
	    alias = keyAlias;
	}

	Object[] objs = recoverPrivateKey(alias, storePass, keyPass);
	PrivateKey privKey = (PrivateKey)objs[0];
	if (keyPass == null)
	    keyPass = (char[])objs[1];
	
	Certificate cert = keyStore.getCertificate(alias);
	if (cert == null) {
	    throw new Exception(alias + " has no public key (certificate)");
	}
	PKCS10 request = new PKCS10(cert.getPublicKey());

	// Construct an X500Signer object, so that we can sign the request
	if (sigAlgName == null) {
	    // If no signature algorithm was specified at the command line,
	    // we choose one that is compatible with the selected private key
	    String keyAlgName = privKey.getAlgorithm();
	    if (keyAlgName.equalsIgnoreCase("DSA")
		   || keyAlgName.equalsIgnoreCase("DSS")) {
		sigAlgName = "SHA1WithDSA";
	    } else if (keyAlgName.equalsIgnoreCase("RSA")) {
		sigAlgName = "MD5WithRSA";
	    } else {
		throw new Exception("Cannot derive signature algorithm");
	    }
	}

	Signature signature = Signature.getInstance(sigAlgName);
	signature.initSign(privKey);
	X500Name subject = (X500Name)((X509Certificate)cert).getSubjectDN();
        X500Signer signer = new X500Signer(signature, subject);

	// Sign the request and base-64 encode it
	request.encodeAndSign(signer);
	request.print(out);
    }

    /**
     * Deletes an entry from the keystore.
     */
    private void doDeleteEntry(String alias) throws Exception {
	if (keyStore.containsAlias(alias) == false) {
	    throw new Exception("Alias <" + alias + "> does not exist");
	}
	keyStore.deleteEntry(alias);
    }

    /**
     * Exports a certificate from the keystore.
     */
    private void doExportCert(String alias, PrintStream out)
	throws Exception
    {
	if (storePass == null) {
	    printWarning();
	}
	if (alias == null) {
	    alias = keyAlias;
	}
	if (keyStore.containsAlias(alias) == false) {
	    throw new Exception("Alias <" + alias + "> does not exist");
	}

	X509Certificate cert = (X509Certificate)keyStore.getCertificate(alias);
	if (cert == null) {
	    throw new Exception("Alias <" + alias + "> has no certificate");
	}
	dumpCert(cert, out);
    }

    /**
     * Creates a new key pair and self-signed certificate.
     */
    private void doGenKeyPair(String alias, String dname, String keyAlgName,
			      int keysize, String sigAlgName)
	throws Exception
    {	
	if (alias == null) {
	    alias = keyAlias;
	}

	if (keyStore.containsAlias(alias) == true) {
	    throw new Exception("Key pair not generated, "
				+ "alias <" + alias + "> already exists");
	}

	if (sigAlgName == null) {
	    if (keyAlgName.equalsIgnoreCase("DSA")) {
		sigAlgName = "SHA1WithDSA";
	    } else if (keyAlgName.equalsIgnoreCase("RSA")) {
		sigAlgName = "MD5WithRSA";
	    } else {
		throw new Exception("Cannot derive signature algorithm");
	    }
	}
	CertAndKeyGen keypair = new CertAndKeyGen(keyAlgName, sigAlgName);

	// If DN is provided, parse it. Otherwise, prompt the user for it.
	X500Name x500Name;
	if (dname == null) {
	    x500Name = getX500Name();
	} else {
	    x500Name = new X500Name(dname);
	}

	if (verbose) {
	    System.err.println("Generating " + keysize
			     + " bit " + keyAlgName + " key pair and "
			     + "self-signed certificate (" + sigAlgName
			     + ")");
	    System.err.println("\tfor: " + x500Name);
	}

	keypair.generate(keysize);
	PrivateKey privKey = keypair.getPrivateKey();
	X509Certificate[] chain = new X509Certificate[1];
	chain[0] = keypair.getSelfCertificate(x500Name, validity*24*60*60);

	if (keyPass == null) {
	    // Prompt for key password
	    int count;
	    for (count = 0; (count) < 3 && (keyPass == null); count++) {
		System.err.println("Enter key password for <" + alias + ">");
		System.err.print("\t(RETURN if same as keystore password):  ");
		System.err.flush();
		keyPass = readPasswd(System.in);
		if (keyPass == null) {
		    keyPass = storePass;
		} else if (keyPass.length < 6) {
		    System.err.println("Key password is too short - "
				       + "must be at least 6 characters");
		    keyPass = null;
		}
	    }
	    if (count == 3) {
		throw new Exception("Too many failures - "
				    + "key not added to keystore");
	    }
	}

	keyStore.setKeyEntry(alias, privKey, keyPass, chain);
    }

    /**
     * Clones a key entry (a private key and its associated certificate chain)
     */
    private void doCloneKey(String orig, String dest)
	throws Exception
    {
	if (orig == null) {
	    orig = keyAlias;
	}

 	if (keyStore.containsAlias(dest) == true) {
	    throw new Exception("Destination alias <" + dest +
				"> already exists");
	}

	Object[] objs = recoverPrivateKey(orig, storePass, keyPass);
	PrivateKey privKey = (PrivateKey)objs[0];
	if (keyPass == null)
	    keyPass = (char[])objs[1];

	// User may choose to encrypt the private key in the cloned entry
	// using a different password
	if (keyPassNew == null) {
	    int count = 0;
	    do {
		keyPassNew = getKeyPasswd(dest, orig, keyPass);
		if (keyPassNew.length < 6) {
		    System.err.println("Password is too short " +
				       "- must be at least 6 characters");
		    keyPassNew = null;
		}
		count++;
	    } while ((keyPassNew == null) && count < 3);
	    if (keyPassNew == null) {
		throw new Exception("Too many failures. Key entry not cloned");
	    }
	}
	keyStore.setKeyEntry(dest, privKey, keyPassNew,
			     keyStore.getCertificateChain(orig));
    }

    /**
     * Changes a key password.
     */
    private void doChangeKeyPasswd(String alias) throws Exception
    {
	if (alias == null) {
	    alias = keyAlias;
	}
	Object[] objs = recoverPrivateKey(alias, storePass, keyPass);
	PrivateKey privKey = (PrivateKey)objs[0];
	if (keyPass == null)
	    keyPass = (char[])objs[1];

	if (keyPassNew == null) {
	    keyPassNew = getNewPasswd("key password for <" + alias + ">",
				      keyPass);
	}
	keyStore.setKeyEntry(alias, privKey, keyPassNew,
			     keyStore.getCertificateChain(alias));
    }

    /**
     * Imports a JDK 1.1-style identity database. We can only store one
     * certificate per identity, because we use the identity's name as the
     * alias (which references a keystore entry), and aliases must be unique.
     */
    private void doImportIdentityDatabase(InputStream in)
	throws Exception
    {
	byte[] encoded;
	ByteArrayInputStream bais;
	java.security.cert.X509Certificate newCert;
	java.security.cert.Certificate[] chain = null;
	PrivateKey privKey;
	boolean modified = false;

	IdentityDatabase idb = IdentityDatabase.fromStream(in);
	for (Enumeration enum = idb.identities(); enum.hasMoreElements();) {
	    Identity id = (Identity)enum.nextElement();
	    newCert = null;
	    // only store trusted identities in keystore
	    if ((id instanceof SystemSigner && ((SystemSigner)id).isTrusted())
		|| (id instanceof SystemIdentity
		    && ((SystemIdentity)id).isTrusted())) {
		// ignore if keystore entry with same alias name already exists
		if (keyStore.containsAlias(id.getName()) == true) {
		    System.err.println("Keystore entry for <"+id.getName()+
				       "> already exists");
		    continue;
		}
		java.security.Certificate[] certs = id.certificates();
		if (certs!=null && certs.length>0) {
		    // we can only store one user cert per identity.
		    // convert old-style to new-style cert via the encoding
		    DerOutputStream dos = new DerOutputStream();
		    certs[0].encode(dos);
		    encoded = dos.toByteArray();
		    bais = new ByteArrayInputStream(encoded);
		    newCert = (X509Certificate)cf.generateCertificate(bais);
		    bais.close();

		    // if certificate is self-signed, make sure it verifies
		    if (isSelfSigned(newCert)) {
			PublicKey pubKey = newCert.getPublicKey();
			try {
			    newCert.verify(pubKey);
			} catch (Exception e) {
			    // ignore this cert
			    continue;
			}
		    }

		    if (id instanceof SystemSigner) {
			System.err.println("Creating keystore entry for <" +
					   id.getName() + "> ...");
			if (chain==null) {
			    chain = new java.security.cert.Certificate[1];
			}
			chain[0] = newCert;
			privKey = ((SystemSigner)id).getPrivateKey();
			keyStore.setKeyEntry(id.getName(), privKey, storePass,
					     chain);
		    } else {
			keyStore.setCertificateEntry(id.getName(), newCert);
		    }
		    kssave = true;
		}
	    }
	}
	if (!kssave) {
	    System.err.println("No entries from identity database added");
	}
    }

    /**
     * Prints a single keystore entry.
     */
    private void doPrintEntry(String alias, PrintStream out,
			      boolean printWarning)
	throws Exception
    {
	if (storePass == null && printWarning == true) {
	    printWarning();
	}

	if (keyStore.containsAlias(alias) == false) {
	    throw new Exception("Alias <" + alias + "> does not exist");
	}

	if (verbose || rfc || debug) {
	    out.println("Alias name: " + alias);
	    out.println("Creation date: "
			+ (keyStore.getCreationDate(alias)).toString());
	} else {
	    out.print(alias + ", "
		      + (keyStore.getCreationDate(alias)).toString()
		      + ", ");
	}

	if (keyStore.isKeyEntry(alias) == true) {
	    // We have a key entry
	    if (verbose || rfc || debug) {
		out.println("Entry type: keyEntry");
	    } else {
		out.println("keyEntry,");
	    }

	    // Get the chain
	    Certificate[] chain = keyStore.getCertificateChain(alias);
	    if (chain != null) {
		if (verbose || rfc || debug) {
		    out.println("Certificate chain length: " + chain.length);
		    for (int i = 0; i < chain.length; i ++) {
			out.println("Certificate[" + (i + 1) + "]:");
			if (verbose && (chain[i] instanceof X509Certificate)) {
			    printX509Cert((X509Certificate)(chain[i]), out);
			} else if (debug) {
			    out.println(chain[i].toString());
			} else {
			    dumpCert(chain[i], out);
			}
		    }
		} else {
		    // Print the digest of the user cert only
		    out.println("Certificate fingerprint (MD5): " +
				getCertFingerPrint("MD5", chain[0]));
		}
	    }
	} else {
	    // We have a trusted certificate entry
	    Certificate cert = keyStore.getCertificate(alias);
	    if (verbose && (cert instanceof X509Certificate)) {
		out.println("Entry type: trustedCertEntry\n");
		printX509Cert((X509Certificate)cert, out);
	    } else if (rfc) {
		out.println("Entry type: trustedCertEntry\n");
		dumpCert(cert, out);
	    } else if (debug) {
		out.println(cert.toString());
	    } else {
		out.println("trustedCertEntry,");
		out.println("Certificate fingerprint (MD5): "
			    + getCertFingerPrint("MD5", cert));
	    }
	}
    }

    /**
     * Prints all keystore entries.
     */
    private void doPrintEntries(PrintStream out)
	throws Exception
    {
	if (storePass == null)
	    printWarning();
	else
	    out.println();

	out.println("Keystore type: " + keyStore.getType());
	out.println("Keystore provider: " + keyStore.getProvider().getName());
	out.println();

	out.println("Your keystore contains " + keyStore.size() + " entr"
		    + ((keyStore.size() == 1) ? "y:" : "ies:"));
	out.println();

	for (Enumeration e = keyStore.aliases(); e.hasMoreElements(); ) {
	    String alias = (String)e.nextElement();
	    doPrintEntry(alias, out, false);
	    if (verbose || rfc) {
		out.println("\n");
		out.println("*******************************************");
		out.println("*******************************************\n\n");
	    }
	}
    }

    /**
     * Reads a certificate (or certificate chain) and prints its contents in
     * a human readbable format.
     */
    private void doPrintCert(InputStream in, PrintStream out)
	throws Exception
    {
	Collection c = null;
	try {
	    c = cf.generateCertificates(in);
	} catch (CertificateException ce) {
	    throw new Exception("Failed to parse input");
	}
	if (c.isEmpty()) {
	    throw new Exception("Empty input");
	}
	Certificate[] certs = (Certificate[])c.toArray();
	for (int i=0; i<certs.length; i++) {
	    X509Certificate x509Cert = null;
	    try {
		x509Cert = (X509Certificate)certs[i];
	    } catch (ClassCastException cce) {
		throw new Exception("Not X.509 certificate");
	    }
	    if (certs.length > 1)
		out.println("Certificate[" + (i+1) + "]:");
	    printX509Cert(x509Cert, out);
	    if (i < (certs.length-1))
		out.println();
	}
    }

    /**
     * Creates a self-signed certificate, and stores it as a single-element
     * certificate chain.
     */
    private void doSelfCert(String alias, String dname, String sigAlgName)
	throws Exception
    {
	if (alias == null) {
	    alias = keyAlias;
	}

	Object[] objs = recoverPrivateKey(alias, storePass, keyPass);
	PrivateKey privKey = (PrivateKey)objs[0];
	if (keyPass == null)
	    keyPass = (char[])objs[1];

	// Determine the signature algorithm
	if (sigAlgName == null) {
	    // If no signature algorithm was specified at the command line,
	    // we choose one that is compatible with the selected private key
	    String keyAlgName = privKey.getAlgorithm();
	    if (keyAlgName.equalsIgnoreCase("DSA")
		   || keyAlgName.equalsIgnoreCase("DSS")) {
		sigAlgName = "SHA1WithDSA";
	    } else if (keyAlgName.equalsIgnoreCase("RSA")) {
		sigAlgName = "MD5WithRSA";
	    } else {
		throw new Exception("Cannot derive signature algorithm");
	    }
	}

	// Get the old certificate
	Certificate oldCert = keyStore.getCertificate(alias);
	if (oldCert == null) {
	    throw new Exception(alias + " has no public key");
	}
	if (!(oldCert instanceof X509Certificate)) {
	    throw new Exception(alias + " has no X.509 certificate");
	}

	// convert to X509CertImpl, so that we can modify selected fields
	// (no public APIs available yet)
	byte[] encoded = oldCert.getEncoded();
	X509CertImpl certImpl = new X509CertImpl(encoded);
	X509CertInfo certInfo = (X509CertInfo)certImpl.get(X509CertImpl.NAME
							   + "." +
							   X509CertImpl.INFO);

	// Extend its validity
	Date firstDate = new Date();
	Date lastDate = new Date();
	lastDate.setTime(firstDate.getTime () + validity*1000*24*60*60L);
	CertificateValidity interval = new CertificateValidity(firstDate,
							       lastDate);
	certInfo.set(X509CertInfo.VALIDITY, interval);

	// Make new serial number
	certInfo.set(X509CertInfo.SERIAL_NUMBER, new CertificateSerialNumber
		     ((int)(firstDate.getTime()/1000)));

	// Set owner and issuer fields
	X500Name owner;
	if (dname == null) {
	    // Get the owner name from the certificate
	    owner = (X500Name)certInfo.get(X509CertInfo.SUBJECT + "." +
					   CertificateSubjectName.DN_NAME);
	} else {
	    // Use the owner name specified at the command line
	    owner = new X500Name(dname);
	    certInfo.set(X509CertInfo.SUBJECT + "." +
			 CertificateSubjectName.DN_NAME, owner);
	}
	// Make issuer same as owner (self-signed!)
	certInfo.set(X509CertInfo.ISSUER + "." +
		     CertificateIssuerName.DN_NAME, owner);

	// The inner and outer signature algorithms have to match.
	// The way we achieve that is really ugly, but there seems to be no
	// other solution: We first sign the cert, then retrieve the
	// outer sigalg and use it to set the inner sigalg
	X509CertImpl newCert = new X509CertImpl(certInfo);
	newCert.sign(privKey, sigAlgName);
	AlgorithmId sigAlgid = (AlgorithmId)newCert.get(X509CertImpl.SIG_ALG);
	certInfo.set(CertificateAlgorithmId.NAME + "." +
		     CertificateAlgorithmId.ALGORITHM, sigAlgid);

	// Sign the new certificate
	newCert = new X509CertImpl(certInfo);
	newCert.sign(privKey, sigAlgName);

	// Store the new certificate as a single-element certificate chain
	keyStore.setKeyEntry(alias, privKey,
			     (keyPass != null) ? keyPass : storePass,
			     new Certificate[] { newCert } );

	if (verbose) {
	    System.err.println("New certificate (self-signed):");
	    System.err.print(newCert.toString());
	    System.err.println();
	}
    }

    /**
     * Processes a certificate reply from a certificate authority.
     *
     * <p>Builds a certificate chain on top of the certificate reply,
     * using trusted certificates from the keystore. The chain is complete
     * after a self-signed certificate has been encountered. The self-signed
     * certificate is considered a root certificate authority, and is stored
     * at the end of the chain.
     *
     * <p>The newly generated chain replaces the old chain associated with the
     * key entry.
     *
     * @return true if the certificate reply was installed, otherwise false.
     */
    private boolean installReply(String alias, InputStream in)
	throws Exception
    {
	if (alias == null) {
	    alias = keyAlias;
	}

	Object[] objs = recoverPrivateKey(alias, storePass, keyPass);
	PrivateKey privKey = (PrivateKey)objs[0];
	if (keyPass == null)
	    keyPass = (char[])objs[1];

	Certificate userCert = keyStore.getCertificate(alias);
	if (userCert == null) {
	    throw new Exception(alias + " has no public key (certificate)");
	}

	// Read the certificates in the reply
	Collection c = cf.generateCertificates(in);
	if (c.isEmpty()) {
	    throw new Exception("Reply has no certificates");
	}
	Certificate[] replyCerts = (Certificate[])c.toArray();
	Certificate[] newChain;
	if (replyCerts.length == 1) {
	    // single-cert reply
	    newChain = establishCertChain(userCert, replyCerts[0]);
	} else {
	    // cert-chain reply (e.g., PKCS#7)
	    newChain = validateReply(alias, userCert, replyCerts);
	}

	// Now store the newly established chain in the keystore. The new
	// chain replaces the old one.
	if (newChain != null) {
	    keyStore.setKeyEntry(alias, privKey,
				 (keyPass != null) ? keyPass : storePass,
				 newChain);
	    return true;
	} else {
	    return false;	    
	}
    }

    /**
     * Imports a certificate and adds it to the list of trusted certificates.
     *
     * @return true if the certificate was added, otherwise false.
     */
    private boolean addTrustedCert(String alias, InputStream in)
	throws Exception
    {
	if (alias == null) {
	    throw new Exception("Must specify alias");
	}
	if (keyStore.containsAlias(alias) == true) {
	    throw new Exception("Certificate not imported, "
				+ "alias <" + alias + "> already exists");
	}

	// Read the certificate
	X509Certificate cert = null;
	try {
	    cert = (X509Certificate)cf.generateCertificate(in);
	} catch (ClassCastException cce) {
	    throw new Exception("Input not an X.509 certificate");
	} catch (CertificateException ce) {
	    throw new Exception("Input not an X.509 certificate");
	}

	// if certificate is self-signed, make sure it verifies
	boolean selfSigned = false;
	if (isSelfSigned(cert)) {
	    cert.verify(cert.getPublicKey());
	    selfSigned = true;
	}

	if (noprompt) {
	    keyStore.setCertificateEntry(alias, cert);
	    return true;
	}

	// check if cert already exists in keystore
	String reply = null;
	String trustalias = keyStore.getCertificateAlias(cert);
	if (trustalias != null) {
	    System.err.println("Certificate already exists in keystore "
			       + "under alias <" + trustalias + ">");
	    reply = getYesNoReply("Do you still want to add it? [no]:  ");
	} else if (selfSigned) {
	    if (trustcacerts) {
		KeyStore caks = getCacertsKeyStore();
		if ((caks != null) &&
		    ((trustalias=caks.getCertificateAlias(cert)) != null)) {
		    System.err.println("Certificate already exists in " +
				       "system-wide CA keystore under " +
				       "alias <" + trustalias + ">");
		    reply = getYesNoReply("Do you still want to add it "
					  + "to your own keystore? [no]:  ");
		}
	    }
	    if (trustalias == null) {
		// Print the cert and ask user if they really want to add
		// it to their keystore
		printX509Cert(cert, System.out);
		reply = getYesNoReply("Trust this certificate? [no]:  ");
	    }
	}
	if (reply != null) {
	    if (reply.equalsIgnoreCase("YES")) {
		keyStore.setCertificateEntry(alias, cert);
		return true;
	    } else {
		return false;
	    }
	}

	// Try to establish trust chain
	try {
	    Certificate[] chain = establishCertChain(null, cert);
	    if (chain != null) {
		keyStore.setCertificateEntry(alias, cert);
		return true;
	    }
	} catch (Exception e) {
	    // Print the cert and ask user if they really want to add it to
	    // their keystore
	    printX509Cert(cert, System.out);
	    reply = getYesNoReply("Trust this certificate? [no]:  ");
	    if (reply.equalsIgnoreCase("YES")) {
		keyStore.setCertificateEntry(alias, cert);
		return true;
	    } else {
		return false;
	    }
	}

	return false;
    }

    /**
     * Prompts user for new password. New password must be different from
     * old one.
     *
     * @param prompt the message that gets prompted on the screen
     * @param oldPasswd the current (i.e., old) password
     */
    private char[] getNewPasswd(String prompt, char[] oldPasswd)
	throws Exception
    {
	char[] entered = null;
	char[] reentered = null;

	for (int count = 0; count < 3; count++) {
	    System.err.print("New " + prompt + ":  ");
	    entered = readPasswd(System.in);
	    if (entered.length < 6) {
		System.err.println("Password is too short " +
				   "- must be at least 6 characters");
	    } else if (Arrays.equals(entered, oldPasswd)) {
		System.err.println("Passwords must differ");
	    } else {
		System.err.print("Re-enter new " + prompt + ":  ");
		reentered = readPasswd(System.in);
		if (!Arrays.equals(entered, reentered)) {
		    System.err.println("They don't match; try again");
		} else {
		    Arrays.fill(reentered, ' ');
		    return entered;
		}
	    }
	    if (entered != null) {
		Arrays.fill(entered, ' ');
		entered = null;
	    }
	    if (reentered != null) {
		Arrays.fill(reentered, ' ');
		reentered = null;
	    }
	}
	throw new Exception("Too many failures - try later");
    }

    /**
     * Prompts user for alias name.
     */
    private String getAlias(String prompt) throws Exception {
	if (prompt != null) {
	    System.err.print("Enter " + prompt + " alias name:  ");
	} else {
	    System.err.print("Enter alias name:  ");
	}
	return (new BufferedReader(new InputStreamReader(
                                        System.in))).readLine();
    }

    /**
     * Prompts user for key password. User may select to choose the same
     * password (<code>otherKeyPass</code>) as for <code>otherAlias</code>.
     */
    private char[] getKeyPasswd(String alias, String otherAlias,
				char[] otherKeyPass)
	throws Exception
    {
	int count = 0;
	char[] keyPass = null;

	do {
	    if (otherKeyPass != null) {
		System.err.println("Enter key password for <" + alias + ">");
		System.err.print("\t(RETURN if same as for <" + otherAlias
				 + ">):  ");
	    } else {
		System.err.print("Enter key password for <" + alias + ">:  ");
	    }
	    System.err.flush();
	    keyPass = readPasswd(System.in);
	    if (keyPass == null) {
		keyPass = otherKeyPass;
	    }
	    count++;
	} while ((keyPass == null) && count < 3);

	if (keyPass == null) {
	    throw new Exception("Too many failures - try later");
	}

	return keyPass;
    }

    /**
     * Reads user password from given input stream.
     */
    private char[] readPasswd(InputStream in) throws IOException {
	char[] lineBuffer;
	char[] buf;
	int i;

	buf = lineBuffer = new char[128];

	int room = buf.length;
	int offset = 0;
	int c;

loop:	while (true) {
	    switch (c = in.read()) {
	      case -1: 
	      case '\n':
		break loop;

	      case '\r':
		int c2 = in.read();
		if ((c2 != '\n') && (c2 != -1)) {
		    if (!(in instanceof PushbackInputStream)) {
			in = new PushbackInputStream(in);
		    }
		    ((PushbackInputStream)in).unread(c2);
		} else 
		    break loop;

	      default:
		if (--room < 0) {
		    buf = new char[offset + 128];
		    room = buf.length - offset - 1;
		    System.arraycopy(lineBuffer, 0, buf, 0, offset);
		    Arrays.fill(lineBuffer, ' ');
		    lineBuffer = buf;
		}
		buf[offset++] = (char) c;
		break;
	    }
	}

	if (offset == 0) {
	    return null;
	}

	char[] ret = new char[offset];
	System.arraycopy(buf, 0, ret, 0, offset);
	Arrays.fill(buf, ' ');

	return ret;
    }

    /**
     * Prints a certificate in a human readable format.
     */
    private void printX509Cert(X509Certificate cert, PrintStream out)
	throws Exception
    {
	out.println("Owner: "
		    + (X500Name)(cert.getSubjectDN())
		    + "\n"
		    + "Issuer: "
		    + (X500Name)(cert.getIssuerDN())
		    + "\n"
		    + "Serial number: " + cert.getSerialNumber().toString(16)
		    + "\n"
		    + "Valid from: " + cert.getNotBefore().toString()
		    + " until: " + cert.getNotAfter().toString()
		    + "\n"
		    + "Certificate fingerprints:\n"
		    + "\t MD5:  " + getCertFingerPrint("MD5", cert)
		    + "\n"
		    + "\t SHA1: " + getCertFingerPrint("SHA1", cert));
    }

    /**
     * Returns true if the certificate is self-signed, false otherwise.
     */
    private boolean isSelfSigned(X509Certificate cert) {
	X500Name subject = (X500Name)cert.getSubjectDN();
	X500Name issuer = (X500Name)cert.getIssuerDN();
	return subject.equals(issuer);
    }

    /**
     * Returns true if the given certificate is trusted, false otherwise.
     */
    private boolean isTrusted(Certificate cert)
	throws Exception
    {
	if (keyStore.getCertificateAlias(cert) != null)
	    return true; // found in own keystore
	KeyStore caks;
	if (trustcacerts && ((caks = getCacertsKeyStore()) != null)) {
	    if (caks.getCertificateAlias(cert) != null)
		return true; // found in CA keystore
		
	}
	return false;
    }

    /**
     * Gets an X.500 name suitable for inclusion in a certification request.
     */
    private X500Name getX500Name() throws IOException {
	BufferedReader in;
	in = new BufferedReader(new InputStreamReader(System.in));
	String commonName = "Unknown";
	String organizationalUnit = "Unknown";
	String organization = "Unknown";
	String city = "Unknown";
	String state = "Unknown";
	String country = "Unknown";
	X500Name name;
	String userInput = null;

	do {
	    commonName = inputString(in,
		    "What is your first and last name?",
		    commonName);
	    organizationalUnit = inputString(in,
		    "What is the name of your organizational unit?",
		    organizationalUnit);
	    organization = inputString(in,
		    "What is the name of your organization?",
		    organization);
	    city = inputString(in,
		    "What is the name of your City or Locality?",
		    city);
	    state = inputString(in,
		    "What is the name of your State or Province?",
		    state);
	    country = inputString(in,
		    "What is the two-letter country code for this unit?",
		    country);
	    name = new X500Name(commonName, organizationalUnit, organization,
				city, state, country);
	    userInput = inputString(in, "Is <" + name + "> correct?", "no");
	} while (!userInput.equalsIgnoreCase("yes")
		    && !userInput.equalsIgnoreCase("y"));

	System.err.println();
	return name;
    }

    private String inputString(BufferedReader in, String prompt,
			       String defaultValue)
	throws IOException
    {
	System.err.println(prompt);
	System.err.print("  [" + defaultValue + "]:  ");
	System.err.flush();

	String value = in.readLine();
	if (value == null || value.equals(""))
	    value = defaultValue;
	return value;
    }

    /**
     * Writes an X.509 certificate in base64 or binary encoding to an output
     * stream.
     */
    private void dumpCert(Certificate cert, PrintStream out)
	throws IOException, CertificateException
    {
	if (rfc) {
	    BASE64Encoder encoder = new BASE64Encoder();
	    out.println(X509Factory.BEGIN_CERT);
	    encoder.encodeBuffer(cert.getEncoded(), out);
	    out.println(X509Factory.END_CERT);
	} else {
	    out.write(cert.getEncoded()); // binary
	}
    }

    /**
     * Converts a byte to hex digit and writes to the supplied buffer
     */
    private void byte2hex(byte b, StringBuffer buf) {
        char[] hexChars = { '0', '1', '2', '3', '4', '5', '6', '7', '8',
                            '9', 'A', 'B', 'C', 'D', 'E', 'F' };
        int high = ((b & 0xf0) >> 4);
        int low = (b & 0x0f);
        buf.append(hexChars[high]);
        buf.append(hexChars[low]);
    }

    /**
     * Converts a byte array to hex string
     */
    private String toHexString(byte[] block) {
        StringBuffer buf = new StringBuffer();
	int len = block.length;
        for (int i = 0; i < len; i++) {
             byte2hex(block[i], buf);
	     if (i < len-1) {
		 buf.append(":");
	     }
        } 
        return buf.toString();
    }

    /**
     * Recovers (private) key associated with given alias.
     *
     * @return an array of objects, where the 1st element in the array is the
     * recovered private key, and the 2nd element is the password used to
     * recover it.
     */
    private Object[] recoverPrivateKey(String alias, char[] storePass,
				       char[] keyPass)
	throws Exception
    {
	Key key = null;

	if (keyStore.containsAlias(alias) == false) {
	    throw new Exception("Alias <" + alias + "> does not exist");
	}
	if (keyStore.isKeyEntry(alias) == false) {
	    throw new Exception("Alias <" + alias + "> has no (private) key");
	}

	if (keyPass == null) {
	    // Try to recover the key using the keystore password
	    try {
		key = keyStore.getKey(alias, storePass);
		keyPass = storePass;
	    } catch (UnrecoverableKeyException e) {
		// Did not work out, so prompt user for key password
		keyPass = getKeyPasswd(alias, null, null);
		key = keyStore.getKey(alias, keyPass);
	    }
	} else {
	    key = keyStore.getKey(alias, keyPass);
	}
	if (!(key instanceof PrivateKey)) {
	    throw new Exception("Recovered key is not a private key");
	}
	return new Object[] {(PrivateKey)key, keyPass};
    }

    /**
     * Gets the requested finger print of the certificate.
     */
    private String getCertFingerPrint(String mdAlg, Certificate cert)
	throws Exception
    {
	byte[] encCertInfo = cert.getEncoded();
	MessageDigest md = MessageDigest.getInstance(mdAlg);
	byte[] digest = md.digest(encCertInfo);
        return toHexString(digest); 
    }

    /**
     * Prints warning about missing integrity check.
     */
    private void printWarning() {
	System.err.println();
	System.err.print("*****************  ");
	System.err.print("WARNING WARNING WARNING  ");
	System.err.println("*****************");                              
	System.err.print("* The integrity of the information stored in ");
	System.err.println("your keystore  *");
	System.err.print("* has NOT been verified!  ");
	System.err.println("In order to verify its integrity, *");
	System.err.print("* you must provide your keystore password.");
	System.err.println("                  *");
	System.err.print("*****************  ");
	System.err.print("WARNING WARNING WARNING  ");
	System.err.println("*****************");                              
	System.err.println();
    }

    /**
     * Validates chain in certification reply, and returns the ordered
     * elements of the chain (with user certificate first, and self-signed root
     * certificate last in the array).
     *
     * @param alias the alias name
     * @param userCert the user certificate of the alias
     * @param replyCerts the chain provided in the reply
     */
    private Certificate[] validateReply(String alias,
					Certificate userCert,
					Certificate[] replyCerts)
	throws Exception
    {
	// order the certs in the reply (we assume there is a self-signed
	// root certificate in the reply).
	// we know that all certs in the reply are of type X.509, because
	// we parsed them using an X.509 certificate factory
	X509Certificate[] orderedCerts
	    = new X509Certificate[replyCerts.length];
	Principal subject = null;
	int i;
	for (i=0; i<replyCerts.length; i++) {
	    if (isSelfSigned((X509Certificate)replyCerts[i])) {
		orderedCerts[orderedCerts.length-1]
		    = (X509Certificate)replyCerts[i];
		subject = ((X509Certificate)replyCerts[i]).getSubjectDN();
		replyCerts[i] = null;
		break;
	    }
	}
	if (i == replyCerts.length) {
	    throw new Exception("No self-signed certificate in reply");
	}

	// order the certs
	for (i=orderedCerts.length-2; i>=0; i--) {
	    int j;
	    for (j=0; j<replyCerts.length; j++) {
		if (replyCerts[j] == null)
		    continue;
		Principal issuer;
		issuer = ((X509Certificate)replyCerts[j]).getIssuerDN();
		if (subject.equals(issuer)) {
		    orderedCerts[i] = (X509Certificate)replyCerts[j];
		    subject = ((X509Certificate)replyCerts[j]).getSubjectDN();
		    replyCerts[j] = null;
		    break;
		}		
	    }
	    if (j == replyCerts.length) {
		throw new Exception("Incomplete certificate chain in reply");
	    }
	}

	// make sure that the user certificate in the reply matches the
	// original public key in the keystore
	PublicKey origPubKey = userCert.getPublicKey();
	PublicKey replyPubKey = orderedCerts[0].getPublicKey();
        if (!origPubKey.equals(replyPubKey)) {
	    String msg = "Public key in certificate reply different from "
		+ "public key for <" + alias + ">"; 
            throw new Exception(msg);
	}
	
	// now verify each cert in the ordered chain
	for (i=0; i<orderedCerts.length-1; i++) {
	    PublicKey pubKey = orderedCerts[i+1].getPublicKey();
	    try {
		orderedCerts[i].verify(pubKey);
	    } catch (Exception e) {
		throw new Exception("Certificate chain in reply does not "
				    + "verify: " + e.getMessage());
	    }
	}

	if (noprompt)
	    return orderedCerts;

	// do we trust the root cert at the top?
	if (!isTrusted(orderedCerts[orderedCerts.length-1])) {
	    System.err.println();
	    System.err.println("Root certificate in reply:\n");
	    printX509Cert(orderedCerts[orderedCerts.length-1], System.out);
	    System.err.println();
	    System.err.print("... is not trusted. ");
	    String reply = getYesNoReply("Install reply anyway? [no]:  "); 
	    if (reply.equalsIgnoreCase("YES")) {
		return orderedCerts;
	    } else {
		return null;
	    }
	}

	return orderedCerts;
    }

    /**
     * Establishes a certificate chain (using trusted certificates in the
     * keystore), starting with the user certificate
     * and ending at a self-signed certificate found in the keystore.
     *
     * @param userCert the user certificate of the alias
     * @param certToVerify the single certificate provided in the reply
     */
    private Certificate[] establishCertChain(Certificate userCert,
					     Certificate certToVerify)
	throws Exception
    {
	if (userCert != null) {
	    // Make sure that the public key of the certificate reply matches
	    // the original public key in the keystore
	    PublicKey origPubKey = userCert.getPublicKey();
	    PublicKey replyPubKey = certToVerify.getPublicKey();
	    if (!origPubKey.equals(replyPubKey)) {
		String msg = "Public keys in reply and keystore "
		    + "don't match"; 
		throw new Exception(msg);
	    }

	    // If the two certs are identical, we're done: no need to import
	    // anything
	    if (certToVerify.equals(userCert)) {
		String msg = "Certificate reply and certificate in keystore "
		    + "are identical";
		throw new Exception(msg);
	    }
	}

	// Build a hash table of all certificates in the keystore.
	// Use the subject distinguished name as the key into the hash table.
	// All certificates associated with the same subject distinguished
	// name are stored in the same hash table entry as a vector.
	Hashtable certs = null;
	if (keyStore.size() > 0) {
	    certs = new Hashtable(11);
	    keystorecerts2Hashtable(keyStore, certs);
	}
	if (trustcacerts) {
	    KeyStore caks = getCacertsKeyStore();
	    if (caks!=null && caks.size()>0) {
		if (certs == null)
		    certs = new Hashtable(11);
		keystorecerts2Hashtable(caks, certs);
	    }
	}

	// start building chain
	Vector chain = new Vector(2);
	if (buildChain((X509Certificate)certToVerify, chain, certs)) {
	    Certificate[] newChain = new Certificate[chain.size()];
	    // buildChain() returns chain with self-signed root-cert first and
	    // user-cert last, so we need to invert the chain before we store
	    // it
	    int j=0;
	    for (int i=chain.size()-1; i>=0; i--) {
		newChain[j] = (Certificate)chain.elementAt(i);
		j++;
	    }
	    return newChain;
	} else {
	    throw new Exception("Failed to establish chain from reply");
	}
    }

    /**
     * Recursively tries to establish chain from pool of trusted certs.
     *
     * @param certToVerify the cert that needs to be verified.
     * @param chain the chain that's being built.
     * @param certs the pool of trusted certs
     *
     * @return true if successful, false otherwise.
     */
    private boolean buildChain(X509Certificate certToVerify, Vector chain,
			       Hashtable certs) {
	Principal subject = certToVerify.getSubjectDN();
	Principal issuer = certToVerify.getIssuerDN();	
	if (subject.equals(issuer)) {
	    // reached self-signed root cert;
	    // no verification needed because it's trusted.
	    chain.addElement(certToVerify);
	    return true;
	}

	// Get the issuer's certificate(s)
	Vector vec = (Vector)certs.get(issuer);
	if (vec == null)
	    return false;
	
	// Try out each certificate in the vector, until we find one
	// whose public key verifies the signature of the certificate
	// in question.
	for (Enumeration issuerCerts = vec.elements();
	     issuerCerts.hasMoreElements(); ) {
	    X509Certificate issuerCert
		= (X509Certificate)issuerCerts.nextElement();
	    PublicKey issuerPubKey = issuerCert.getPublicKey();
	    try {
		certToVerify.verify(issuerPubKey);
	    } catch(Exception e) {
		continue;
	    }
	    if (buildChain(issuerCert, chain, certs)) {
		chain.addElement(certToVerify);
		return true;
	    }
	}
	return false;
    }

    /**
     * Prompts user for yes/no decision.
     *
     * @return the user's decision
     */
    private String getYesNoReply(String prompt)
	throws IOException
    {
	String reply = null;
	do {
	    System.err.print(prompt);
	    System.err.flush();
	    reply = (new BufferedReader(new InputStreamReader
					(System.in))).readLine();
	    if (reply.equals("") || reply.equalsIgnoreCase("n")
		|| reply.equalsIgnoreCase("no")) {
		reply = "NO";
	    } else if (reply.equalsIgnoreCase("y")
		       || reply.equalsIgnoreCase("yes")) {
		reply = "YES";
	    } else {
		System.err.println("Wrong answer, try again");
		reply = null;
	    }
	} while (reply == null);
	return reply;
    }

    /**
     * Returns the keystore with the configured CA certificates.
     */
    private KeyStore getCacertsKeyStore()
	throws Exception
    {
	String sep = File.separator;
	File file = new File(System.getProperty("java.home") + sep
			     + "lib" + sep + "security" + sep
			     + "cacerts");
	if (!file.exists())
	    return null;
	FileInputStream fis = new FileInputStream(file);
	KeyStore caks = KeyStore.getInstance("jks");
	caks.load(fis, null);
	fis.close();
	return caks;
    }

    /**
     * Stores the (leaf) certificates of a keystore in a hashtable.
     * All certs belonging to the same CA are stored in a vector that
     * in turn is stored in the hashtable, keyed by the CA's subject DN
     */
    private void keystorecerts2Hashtable(KeyStore ks, Hashtable hash)
	throws Exception
    {
	for (Enumeration aliases = ks.aliases(); aliases.hasMoreElements(); ) {
	    String alias = (String)aliases.nextElement();
	    Certificate cert = ks.getCertificate(alias);
	    if (cert != null) {
		Principal subjectDN = ((X509Certificate)cert).getSubjectDN();
		Vector vec = (Vector)hash.get(subjectDN);
		if (vec == null) {
		    vec = new Vector();
		    vec.addElement(cert);
		} else {
		    if (!vec.contains(cert))
			vec.addElement(cert);
		}
		hash.put(subjectDN, vec);
	    }
	}
    }

    /**
     * Prints the usage of this tool.
     */
    private void usage() {
	System.err.println("keytool usage:\n");

	System.err.print("-certreq     [-v] [-alias <alias>] ");
	System.err.println("[-sigalg <sigalg>]");
	System.err.println("\t     [-file <csr_file>] [-keypass <keypass>]");
	System.err.print("\t     [-keystore <keystore>] ");
	System.err.println("[-storepass <storepass>]");
	System.err.println("\t     [-storetype <storetype>]");
	System.err.println();

	System.err.println("-delete      [-v] -alias <alias>");
	System.err.print("\t     [-keystore <keystore>] ");
	System.err.println("[-storepass <storepass>]");
	System.err.println("\t     [-storetype <storetype>]");
	System.err.println();

	System.err.print("-export      [-v] [-rfc] [-alias <alias>] ");
	System.err.println("[-file <cert_file>]");
	System.err.print("\t     [-keystore <keystore>] ");
	System.err.println("[-storepass <storepass>]");
	System.err.println("\t     [-storetype <storetype>]");
	System.err.println();

	System.err.print("-genkey      [-v] [-alias <alias>] ");
	System.err.println("[-keyalg <keyalg>]");
	System.err.println("\t     [-keysize <keysize>] [-sigalg <sigalg>]");
	System.err.print("\t     [-dname <dname>] ");
	System.err.println("[-validity <valDays>]");
	System.err.print("\t     [-keypass <keypass>] ");
	System.err.println("[-keystore <keystore>]");
	System.err.print("\t     [-storepass <storepass>] ");
	System.err.println("[-storetype <storetype>]");
	System.err.println();

	System.err.println("-help");
	System.err.println();

	System.err.print("-identitydb  [-v] [-file <idb_file>] ");
	System.err.println("[-keystore <keystore>]");
	System.err.print("\t     [-storepass <storepass>] ");
	System.err.println("[-storetype <storetype>]");
	System.err.println();

	System.err.print("-import      [-v] [-noprompt] [-trustcacerts] ");
	System.err.println("[-alias <alias>]");
	System.err.print("\t     [-file <cert_file>] ");
	System.err.println("[-keypass <keypass>] ");
	System.err.print("\t     [-keystore <keystore>] ");
	System.err.println("[-storepass <storepass>]");
	System.err.println("\t     [-storetype <storetype>]");
	System.err.println();

	System.err.print("-keyclone    [-v] [-alias <alias>] ");
	System.err.println("-dest <dest_alias>");
	System.err.print("\t     [-keypass <keypass>] ");
	System.err.println("[-new <new_keypass>]");
	System.err.print("\t     [-keystore <keystore>] ");
	System.err.println("[-storepass <storepass>]");
	System.err.println("\t     [-storetype <storetype>]");
	System.err.println();

	System.err.println("-keypasswd   [-v] [-alias <alias>]");
	System.err.print("\t     [-keypass <old_keypass>] ");
	System.err.println("[-new <new_keypass>]");
	System.err.print("\t     [-keystore <keystore>] ");
	System.err.println("[-storepass <storepass>]");
	System.err.println("\t     [-storetype <storetype>]");
	System.err.println();

	System.err.println("-list        [-v | -rfc] [-alias <alias>]");
	System.err.print("\t     [-keystore <keystore>] ");
	System.err.println("[-storepass <storepass>]");
	System.err.println("\t     [-storetype <storetype>]");
	System.err.println();

	System.err.println("-printcert   [-v] [-file <cert_file>]");
	System.err.println();

	System.err.print("-selfcert    [-v] [-alias <alias>] ");
	System.err.println("[-sigalg <sigalg>]");
	System.err.print("\t     [-dname <dname>] ");
	System.err.println("[-validity <valDays>]");
	System.err.print("\t     [-keypass <keypass>] ");
	System.err.println("[-keystore <keystore>]");
	System.err.print("\t     [-storepass <storepass>] ");
	System.err.println("[-storetype <storetype>]");
	System.err.println();

	System.err.println("-storepasswd [-v] [-new <new_storepass>]");
	System.err.print("\t     [-keystore <keystore>] ");
	System.err.println("[-storepass <storepass>]");
	System.err.println("\t     [-storetype <storetype>]");
	System.err.println();

	System.exit(1);
    }
}
