/*
 * @(#)PolicyTool.java	1.30 98/09/30
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.security.tools;

import java.io.*;
import java.util.Vector;
import java.util.Enumeration;
import java.net.URL;
import java.net.MalformedURLException;
import java.lang.reflect.*;
import sun.misc.BASE64Decoder;
import sun.security.util.PropertyExpander;
import sun.security.util.PropertyExpander.ExpandException;
import java.awt.*;
import java.awt.event.*;
import java.security.cert.Certificate;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.security.cert.CertificateException;
import java.security.*;
import sun.security.provider.*;

/**
 * PolicyTool may be used by users and administrators to configure the
 * overall java security policy (currently stored in the policy file).
 * Using PolicyTool administators may add and remove policies from
 * the policy file. <p>
 *
 * @version 1.30, 09/30/98
 * @see java.security.Policy
 * @since   JDK1.2
 */

class PolicyTool {

    // anyone can add warnings
    Vector warnings;
    boolean newWarning = false;

    // set to true if policy modified.
    // this way upon exit we know if to ask the user to save changes
    boolean modified = false;

    private static final boolean testing = false;
    private static final Class[] PARAMS = { String.class, String.class};

    /*
     * All of the policy entries are read in from the
     * policy file and stored here.  Updates to the policy entries
     * using addEntry() and removeEntry() are made here.  To ultimately save
     * the policy entries back to the policy file, the SavePolicy button
     * must be clicked.
     **/
    private static String policyFileName = null;
    private Vector policyEntries = null;
    private PolicyParser parser = null;

    /* The public key alias information is stored here.  */
    private KeyStore keyStore = null;
    private String keyStoreName = " ";
    private String keyStoreType = " ";

    /**
     * default constructor
     */
    private PolicyTool() {
	policyEntries = new Vector();
	parser = new PolicyParser();
	warnings = new Vector();
    }

    /**
     * get the PolicyFileName
     */
    String getPolicyFileName() {
	return policyFileName;
    }

    /**
     * set the PolicyFileName
     */
    void setPolicyFileName(String policyFileName) {
	this.policyFileName = policyFileName;
    }

    /**
     * set the keyStore URL name
     */
    void setKeyStoreInfo(String keyStoreName, String keyStoreType) {
	this.keyStoreName = keyStoreName;
	this.keyStoreType = keyStoreType;
	modified = true;
    }

    /**
     * get the keyStore URL name
     */
    String getKeyStoreName() {
	return keyStoreName;
    }

    /**
     * get the keyStore Type
     */
    String getKeyStoreType() {
	return keyStoreType;
    }

    /**
     * Open and read a policy file
     */
    void openPolicy(String filename)
    throws MalformedURLException, NoSuchMethodException,
    ClassNotFoundException, InstantiationException, IllegalAccessException,
    InvocationTargetException, PolicyParser.ParsingException,
    IOException, FileNotFoundException, CertificateException, Exception {

	newWarning = false;

	// start fresh - blow away the current state
	policyEntries = new Vector();
	parser = new PolicyParser();
	warnings = new Vector();
	setPolicyFileName(null);
	setKeyStoreInfo(null, null);

	// see if user is opening a NEW policy file
	if (filename == null) {
	    modified = false;
	    return;
	}

	// Read in the policy entries from the file and
	// populate the parser vector table.  The parser vector
	// table only holds the entries as strings, so it only
	// guarantees that the policies are syntactically
	// correct.
	setPolicyFileName(filename);
	parser.read(new FileReader(filename));
	setKeyStoreInfo(parser.getKeyStoreUrl(), parser.getKeyStoreType());

	// Update the local vector with the same policy entries.
	// This guarantees that the policy entries are not only
	// syntactically correct, but semantically valid as well.
	Enumeration enum = parser.grantElements();
	while (enum.hasMoreElements()) {
	    PolicyParser.GrantEntry ge =
				(PolicyParser.GrantEntry)enum.nextElement();

	    // see if all the signers have public keys
	    if (ge.signedBy != null) {

		String signers[] = parseSigners(ge.signedBy);
		for (int i = 0; i < signers.length; i++) {
		    PublicKey pubKey = getPublicKeyAlias(signers[i]);
		    if (pubKey == null) {
			newWarning = true;
			warnings.addElement
			    ("Warning: A public key for alias '" + signers[i] +
			    "' does not exist.");
		    }
		}
	    }

	    // check to see if the Permissions are valid
	    Enumeration perms = ge.permissionElements();
	    while (perms.hasMoreElements()) {
		PolicyParser.PermissionEntry pe =
			(PolicyParser.PermissionEntry)perms.nextElement();
		try {
		    verifyPermission(pe.permission, pe.name, pe.action);
		} catch (ClassNotFoundException fnfe) {
		    newWarning = true;
		    warnings.addElement
			("Warning: Class not found: " + pe.permission + ".");
		}

		// see if all the permission signers have public keys
		if (pe.signedBy != null) {

		    String signers[] = parseSigners(pe.signedBy);

		    for (int i = 0; i < signers.length; i++) {
			PublicKey pubKey = getPublicKeyAlias(signers[i]);
			if (pubKey == null) {
			    newWarning = true;
			    warnings.addElement
			    ("Warning: A public key for alias '" + signers[i] +
			    "' does not exist.");
			}
		    }
		}
	    }
	    PolicyEntry pEntry = new PolicyEntry(this, ge);
	    policyEntries.addElement(pEntry);
	}
	if (newWarning == false)
	    warnings.addElement("Policy File opened successfully");

	// just read in the policy -- nothing has been modified yet
	modified = false;
    }

    /**
     * Save a policy to a file
     */
    void savePolicy(String filename)
    throws FileNotFoundException, IOException {
	// save the policy entries to a file
	parser.setKeyStoreUrl(keyStoreName);
	parser.setKeyStoreType(keyStoreType);
	parser.write(new FileWriter(filename));
	modified = false;
    }

    /**
     * Open the KeyStore
     */
    void openKeyStore(String keyStoreName, char[] password)
	throws KeyStoreException, NoSuchAlgorithmException,
	UnrecoverableKeyException, IOException,
	CertificateException, ExpandException {

	    if (keyStoreName == null || keyStoreName.length() == 0)
		throw new IOException("null Keystore name");
 
	    keyStoreName = PropertyExpander.expand
				(keyStoreName).replace
				(File.separatorChar, '/');
	    URL url = null;
	    try {
		url = new URL(keyStoreName);
	    } catch (java.net.MalformedURLException e) {
		File pfile = new File(policyFileName);
		URL policyURL = new URL("file:" + pfile.getCanonicalPath());
		url = new URL(policyURL , keyStoreName);
	    }
	    InputStream is = url.openStream();

	    // XXX We should add a command-line option that lets users
	    // specify the keystore type.
	    keyStore = KeyStore.getInstance(KeyStore.getDefaultType());
	    keyStore.load(is, password);
	    is.close();
    }
 
    /**
     * Add a Grant entry to the overall policy at the specified index.
     * A policy entry consists of a CodeSource.
     */
    boolean addEntry(PolicyEntry pe, int index) {

	if (index < 0) {
	    // new entry -- just add it to the end
	    policyEntries.addElement(pe);
	    parser.add(pe.getGrantEntry());
	} else {
	    // existing entry -- replace old one
	    PolicyEntry origPe = (PolicyEntry)policyEntries.elementAt(index);
	    parser.replace(origPe.getGrantEntry(), pe.getGrantEntry());
	    policyEntries.setElementAt(pe, index);
	}
	return true;
    }

    /**
     * Add a Permission entry to an existing PolicyEntry at the specified index.
     * A Permission entry consists of a permission, name, and actions.
     *
     * If the permission already exists, it is not added again.
     */
    boolean addPermEntry(PolicyEntry pe,
			PolicyParser.PermissionEntry newPerm,
			int index) {

	// first add the permission to the Policy Parser Vector
	PolicyParser.GrantEntry grantEntry = pe.getGrantEntry();
	if (grantEntry.contains(newPerm) == true)
	    return false;

	Vector permList = grantEntry.permissionEntries;
	if (index != -1)
	    permList.setElementAt(newPerm, index);
	else
	    permList.addElement(newPerm);

	modified = true;
	return true;
    }

    /**
     * Remove a Permission entry from an existing PolicyEntry.
     */
    boolean removePermEntry(PolicyEntry pe,
			PolicyParser.PermissionEntry perm) {

	// remove the Permission from the GrantEntry
	PolicyParser.GrantEntry ppge = pe.getGrantEntry();
	modified = ppge.remove(perm);
	return modified;
    }

    /**
     * remove an entry from the overall policy
     */
    boolean removeEntry(PolicyEntry pe) {

	parser.remove(pe.getGrantEntry());
	modified = true;
	return (policyEntries.removeElement(pe));

    }

    /**
     * retrieve all Policy Entries
     */
    PolicyEntry[] getEntry() {

	if (policyEntries.size() > 0) {
	    PolicyEntry entries[] = new PolicyEntry[policyEntries.size()];
	    for (int i = 0; i < policyEntries.size(); i++)
		entries[i] = (PolicyEntry)policyEntries.elementAt(i);
	    return entries;
	}
	return null;
    }

    /**
     * Retrieve the public key mapped to a particular name.
     * If the key has expired, a KeyException is thrown.
     */
    PublicKey getPublicKeyAlias(String name)
	throws CertificateException, NoSuchAlgorithmException,
	UnrecoverableKeyException, KeyStoreException, ExpandException {

	try {
	    openKeyStore(keyStoreName, null);
	    Certificate cert = keyStore.getCertificate(name);

	    if (cert == null)
		return null;
	    PublicKey pubKey = cert.getPublicKey();
	    return pubKey;
	} catch (IOException ioe) {
	    // no keystore
	    // XXX: should i display an error dialog to the user?
	    newWarning = true;
	    warnings.addElement
		("Warning: Unable to open Keystore: " + keyStoreName);
	    return null;
	}
    }

    /**
     * Retrieve all the alias names stored in the certificate database
     */
    String[] getPublicKeyAlias()
	throws CertificateException, NoSuchAlgorithmException,
	UnrecoverableKeyException, KeyStoreException, ExpandException {

	try {
	    int numAliases = 0;
	    String aliases[] = null;

	    openKeyStore(keyStoreName, null);
	    Enumeration enum = keyStore.aliases();
 
	    // first count the number of elements
	    while (enum.hasMoreElements()) {
		enum.nextElement();
		numAliases++;
	    }
 
	    if (numAliases > 0) {
		// now copy them into an array
		aliases = new String[numAliases];
		numAliases = 0;
		enum = keyStore.aliases();
		while (enum.hasMoreElements()) {
		    aliases[numAliases] =
				new String((String)enum.nextElement());
		    numAliases++;
		}
	    }
	    return aliases;
	} catch (IOException ioe) {
	    // no keystore
	    // XXX: should i display an error dialog to the user?
	    newWarning = true;
	    warnings.addElement
		("Warning: Unable to open Keystore: " + keyStoreName);
	    return null;
	}
    }

    /**
     * This method parses a single string of signers separated by commas
     * ("jordan, duke, pippen") into an array of individual strings.
     */
    String[] parseSigners(String signedBy) {

	String signers[] = null;
	int numSigners = 1;
	int signedByIndex = 0;
	int commaIndex = 0;
	int signerNum = 0;
			
	// first pass thru "signedBy" counts the number of signers
	while (commaIndex >= 0) {
	    commaIndex = signedBy.indexOf(',', signedByIndex);
	    if (commaIndex >= 0) {
		numSigners++;
		signedByIndex = commaIndex + 1;
	    }
	}
	signers = new String[numSigners];
			
	// second pass thru "signedBy" transfers signers to array
	commaIndex = 0;
	signedByIndex = 0;
	while (commaIndex >= 0) {
	    if ((commaIndex = signedBy.indexOf(',', signedByIndex)) >= 0) {
		// transfer signer and ignore trailing part of the string
		signers[signerNum] =
			signedBy.substring(signedByIndex, commaIndex).trim();
		signerNum++;
		signedByIndex = commaIndex + 1;
	    } else {
		// we are at the end of the string -- transfer signer
		signers[signerNum] = signedBy.substring(signedByIndex).trim();
	    }
	}
	return signers;
    }

    /**
     * Check to see if the Permission contents are OK
     */
    void verifyPermission(String type,
				    String name,
				    String actions)
	throws ClassNotFoundException,
	       InstantiationException,
	       IllegalAccessException,
	       NoSuchMethodException,
	       InvocationTargetException
    {

	//XXX we might want to keep a hash of created factories...
	Class pc = Class.forName(type);
	Constructor c = pc.getConstructor(PARAMS);
	Permission p =
		(Permission)c.newInstance(new Object[] { name, actions });
    }

    /*
     * Parse command line arguments.
     */
    static void parseArgs(String args[]) {
	/* parse flags */
	int n = 0;
 
	for (n=0; (n < args.length) && args[n].startsWith("-"); n++) {
 
	    String flags = args[n];
 
	    if (flags.equals("-file")) {
		if (++n == args.length) usage();
		policyFileName = args[n];
	    } else {
		System.err.println("Illegal option: " + flags);
		usage();
	    }
	}
    }

    static void usage() {
	System.out.println("Usage: policytool [options]");
	System.out.println();
	System.out.println("  [-file <file>]    policy file location");
	System.out.println();
 
	System.exit(1);
    }

    /**
     * run the PolicyTool
     */
    public static void main(String args[]) {
	parseArgs(args);
	ToolWindow tw = new ToolWindow(new PolicyTool());
	tw.displayToolWindow(args);
    }
}

/**
 * Each entry in the policy configuration file is represented by a
 * PolicyEntry object.
 *
 * A PolicyEntry is a (CodeSource,Permission) pair.  The
 * CodeSource contains the (URL, PublicKey) that together identify
 * where the Java bytecodes come from and who (if anyone) signed
 * them.  The URL could refer to localhost.  The URL could also be
 * null, meaning that this policy entry is given to all comers, as
 * long as they match the signer field.  The signer could be null,
 * meaning the code is not signed.
 *
 * The Permission contains the (Type, Name, Action) triplet.
 *
 */
class PolicyEntry {

    private CodeSource codesource;
    private PolicyTool tool;
    private PolicyParser.GrantEntry grantEntry;
    private boolean testing = false;

    /**
     * Create a PolicyEntry object from the information read in
     * from a policy file.
     */
    PolicyEntry(PolicyTool tool, PolicyParser.GrantEntry ge)
    throws MalformedURLException, NoSuchMethodException,
    ClassNotFoundException, InstantiationException, IllegalAccessException,
    InvocationTargetException, CertificateException,
    IOException, NoSuchAlgorithmException, UnrecoverableKeyException {

	this.tool = tool;

	URL location = null;
 
	// construct the CodeSource
	if (ge.codeBase != null)
	    location = new URL(ge.codeBase);
	this.codesource = new CodeSource(location, null);

	if (testing) {
	    System.out.println("Adding Policy Entry:");
	    System.out.println("    CodeBase = " + location);
	    System.out.println("    Signers = " + ge.signedBy);
	}

	this.grantEntry = ge;
    }

    /**
     * get the codesource associated with this PolicyEntry
     */
    CodeSource getCodeSource() {
	return codesource;
    }

    /**
     * get the GrantEntry associated with this PolicyEntry
     */
    PolicyParser.GrantEntry getGrantEntry() {
	return grantEntry;
    }

    /**
     * convert the Codebase/signer portion of this policy entry into a string
     */
    String codebaseToString() {

	String stringEntry = new String();

	if (grantEntry.codeBase != null &&
	    grantEntry.codeBase.equals("") == false)
	    stringEntry = stringEntry.concat
				("CodeBase \"" +
				grantEntry.codeBase +
				"\"");

	if (grantEntry.signedBy != null &&
	    grantEntry.signedBy.equals("") == false)
	    stringEntry = ((stringEntry.length() > 0) ?
		stringEntry.concat(", SignedBy \"" +
				grantEntry.signedBy +
				"\"") :
		stringEntry.concat("SignedBy \"" +
				grantEntry.signedBy +
				"\""));

	if (stringEntry.length() == 0)
	    return new String("CodeBase <ALL>");
	return stringEntry;
    }

    /**
     * convert this policy entry into a PolicyParser.PermissionEntry
     */
    PolicyParser.PermissionEntry toPermissionEntry(Permission perm) {

	String actions = null;

	// get the actions
	if (perm.getActions() != null &&
	    perm.getActions().trim() != "")
		actions = perm.getActions();

	PolicyParser.PermissionEntry pe = new PolicyParser.PermissionEntry
			(perm.getClass().getName(),
			perm.getName(),
			actions);
	return pe;
    }
}

/**
 * The main window for the PolicyTool
 */
class ToolWindow extends Frame {

    /* buttons and menus */
    public final String NEW_POLICY_FILE		= "New";
    public final String OPEN_POLICY_FILE	= "Open";
    public final String SAVE_POLICY_FILE	= "Save";
    public final String SAVE_AS_POLICY_FILE	= "Save As";
    public final String VIEW_WARNINGS		= "View Warning Log";
    public final String QUIT			= "Exit";
    public final String ADD_POLICY_ENTRY	= "Add Policy Entry";
    public final String EDIT_POLICY_ENTRY	= "Edit Policy Entry";
    public final String REMOVE_POLICY_ENTRY	= "Remove Policy Entry";
    public final String CHANGE_KEYSTORE		= "Change KeyStore";
    public final String ADD_PUBKEY_ALIAS	= "Add Public Key Alias";
    public final String REMOVE_PUBKEY_ALIAS	= "Remove Public Key Alias";

    /* gridbag index for components in the main window (MW) */
    public final int MW_EMPTY_LABEL1		= 0;
    public final int MW_FILENAME_LABEL		= 1;
    public final int MW_FILENAME_TEXTFIELD	= 2;
    public final int MW_KEYSTORE_LABEL		= 3;
    public final int MW_KEYSTORE_TEXTFIELD	= 4;
    public final int MW_EMPTY_LABEL2		= 5;
    public final int MW_PANEL			= 6;
    public final int MW_ADD_BUTTON		= 0;
    public final int MW_EMPTY_LABEL3		= 1;
    public final int MW_EDIT_BUTTON		= 2;
    public final int MW_EMPTY_LABEL4		= 3;
    public final int MW_REMOVE_BUTTON		= 4;
    public final int MW_EMPTY_LABEL5		= 7;
    public final int MW_POLICY_LIST		= 8;

    private PolicyTool tool;

    /**
     * Constructor
     */
    ToolWindow(PolicyTool tool) {
	this.tool = tool;
    }

    /**
     * Initialize the PolicyTool window with the necessary components
     */
    private void initWindow() {

	// create the top menu bar
	MenuBar menuBar = new MenuBar();

	// create a File menu
	Menu menu = new Menu("File");
	menu.add(NEW_POLICY_FILE);
	menu.add(OPEN_POLICY_FILE);
	menu.add(SAVE_POLICY_FILE);
	menu.add(SAVE_AS_POLICY_FILE);
	menu.add(VIEW_WARNINGS);
	menu.add(QUIT);
	menu.addActionListener(new FileMenuListener(tool, this));
	menuBar.add(menu);
	setMenuBar(menuBar);

	// create an Edit menu
	menu = new Menu("Edit");
	menu.add(ADD_POLICY_ENTRY);
	menu.add(EDIT_POLICY_ENTRY);
	menu.add(REMOVE_POLICY_ENTRY);

	/*
	 * XXX: the next two lines are commented out on purpose.
	 *	the code is left here just in case we decide
	 *	to restore this functionality later.
	 */
	// menu.add(ADD_PUBKEY_ALIAS);
	// menu.add(REMOVE_PUBKEY_ALIAS);

	menu.add(CHANGE_KEYSTORE);
	menu.addActionListener(new MainWindowListener(tool, this));
	menuBar.add(menu);
	setMenuBar(menuBar);

	// empty space
	Label label = new Label("");
	addNewComponent(this, label, MW_EMPTY_LABEL1,
			0, 0, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);

	// policy entry listing
	label = new Label("Policy File:");
	addNewComponent(this, label, MW_FILENAME_LABEL,
			0, 1, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);
	TextField tf = new TextField(50);
	tf.setEditable(false);
	addNewComponent(this, tf, MW_FILENAME_TEXTFIELD,
			1, 1, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);

	// KeyStore URL
	label = new Label("Keystore:");
	addNewComponent(this, label, MW_KEYSTORE_LABEL,
			0, 2, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);
	tf = new TextField(50);
	tf.setEditable(false);
	addNewComponent(this, tf, MW_KEYSTORE_TEXTFIELD,
			1, 2, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);

	// empty space
	label = new Label("");
	addNewComponent(this, label, MW_EMPTY_LABEL2,
			0, 3, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);

	// add ADD/REMOVE/EDIT buttons in a new panel
	Panel panel = new Panel();
	panel.setLayout(new GridBagLayout());

	Button button = new Button(ADD_POLICY_ENTRY);
	button.addActionListener(new MainWindowListener(tool, this));
	addNewComponent(panel, button, MW_ADD_BUTTON,
			0, 0, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);
	
	// empty space
	label = new Label("");
	addNewComponent(panel, label, MW_EMPTY_LABEL3,
			1, 0, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);

	button = new Button(EDIT_POLICY_ENTRY);
	button.addActionListener(new MainWindowListener(tool, this));
	addNewComponent(panel, button, MW_EDIT_BUTTON,
			2, 0, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);
	
	// empty space
	label = new Label("");
	addNewComponent(panel, label, MW_EMPTY_LABEL4,
			3, 0, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);

	button = new Button(REMOVE_POLICY_ENTRY);
	button.addActionListener(new MainWindowListener(tool, this));
	addNewComponent(panel, button, MW_REMOVE_BUTTON,
			4, 0, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);
	
	addNewComponent(this, panel, MW_PANEL,
			0, 4, 2, 1, 0.0, 0.0, GridBagConstraints.BOTH);

	// empty space
	label = new Label("");
	addNewComponent(this, label, MW_EMPTY_LABEL5,
			0, 5, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);

	String policyFile = tool.getPolicyFileName();
	if (policyFile == null) {
	    String userHome;
	    userHome = (String) java.security.AccessController.doPrivileged(
                    new sun.security.action.GetPropertyAction("user.home"));
	    policyFile = userHome + "/.java.policy";
	}

	try {
	    // open the policy file
	    tool.openPolicy(policyFile);

	    // display the policy entries via the policy list textarea
	    List list = new List(40, false);
	    list.addActionListener(new PolicyListListener(tool, this));
	    PolicyEntry entries[] = tool.getEntry();
	    if (entries != null) {
		for (int i = 0; i < entries.length; i++)
		    list.add(entries[i].codebaseToString());
	    }
	    TextField newFilename = (TextField)
				getComponent(MW_FILENAME_TEXTFIELD);
	    newFilename.setText(policyFile);
	    TextField ksName = (TextField)
				getComponent(MW_KEYSTORE_TEXTFIELD);
	    if (tool.getKeyStoreType() != null &&
		tool.getKeyStoreType().length() > 0)
		ksName.setText(tool.getKeyStoreName() +
				", " +
				tool.getKeyStoreType());
	    else
		ksName.setText(tool.getKeyStoreName());

	    initPolicyList(list);

	} catch (PolicyParser.ParsingException pppe) {
	    // add blank policy listing
	    List list = new List(40, false);
	    initPolicyList(list);
	    setVisible(true);

	    // display the error
	    displayErrorDialog(null,
		"Error parsing policy file " + policyFile + pppe.getMessage());

	    // exit
	    System.exit(1);
	} catch (FileNotFoundException fnfe) {

	    // add blank policy listing
	    List list = new List(40, false);
	    list.addActionListener(new PolicyListListener(tool, this));
	    initPolicyList(list);
	    tool.modified = false;
	    setVisible(true);

	    // display the error
	    displayErrorDialog(null,
		"Could not find Policy File: " + policyFile);

	} catch (Exception ee) {
	    // add blank policy listing
	    List list = new List(40, false);
	    list.addActionListener(new PolicyListListener(tool, this));
	    initPolicyList(list);
	    tool.modified = false;
	}
    }

    /**
     * Add a component to the PolicyTool window
     */
    void addNewComponent(Container container, Component component,
	int index, int gridx, int gridy, int gridwidth, int gridheight,
	double weightx, double weighty, int fill) {

	// add the component at the specified gridbag index
	container.add(component, index);

	// set the constraints
	GridBagLayout gbl = (GridBagLayout)container.getLayout();
	GridBagConstraints gbc = new GridBagConstraints();
	gbc.gridx = gridx;
	gbc.gridy = gridy;
	gbc.gridwidth = gridwidth;
	gbc.gridheight = gridheight;
	gbc.weightx = weightx;
	gbc.weighty = weighty;
	gbc.fill = fill;
	gbl.setConstraints(component, gbc);
    }

    /**
     * Init the policy_entry_list TEXTAREA component in the
     * PolicyTool window
     */
    void initPolicyList(List policyList) {

	// add the policy list
	add(policyList, MW_POLICY_LIST);
 
	// set the constraints
	GridBagLayout gbl = (GridBagLayout)getLayout();
	GridBagConstraints gbc = new GridBagConstraints();
	gbc.gridx = 0;
	gbc.gridy = 6;
	gbc.gridwidth = 2;
	gbc.gridheight = 1;
	gbc.weightx = 1.0;
	gbc.weighty = 1.0;
	gbc.fill = GridBagConstraints.BOTH;
	gbl.setConstraints(policyList, gbc);

    }

    /**
     * Replace the policy_entry_list TEXTAREA component in the
     * PolicyTool window with an updated one.
     */
    void replacePolicyList(List policyList) {

	// remove the original list of Policy Entries
	// and add the new list of entries
	List list = (List)getComponent(MW_POLICY_LIST);
	list.removeAll();
	String newItems[] = policyList.getItems();
	for (int i = 0; i < newItems.length; i++)
	    list.add(newItems[i]);
    }

    /**
     * display the main PolicyTool window
     */
    void displayToolWindow(String args[]) {

	setTitle("Policy Tool");
	setResizable(true);
	addWindowListener(new ToolWindowListener(this));
	setBounds(135, 80, 500, 500);
	setLayout(new GridBagLayout());

	initWindow();

	// display it
	setVisible(true);

	if (tool.newWarning == true) {
	    displayStatusDialog(this,
		"Errors have occurred while opening the policy configuration.  View the Warning Log for more information.");
	}
    }

    /**
     * displays a dialog box describing an error which occurred.
     */
    void displayErrorDialog(Window w, String error) {
	ToolDialog ed = new ToolDialog("Error", tool, this, true);

	// find where the PolicyTool gui is
	Point location = ((w == null) ?
		getLocationOnScreen() : w.getLocationOnScreen());
	ed.setBounds(location.x + 50, location.y + 50, 600, 100);
	ed.setLayout(new GridBagLayout());

	Label label = new Label(error);
	addNewComponent(ed, label, 0,
			0, 0, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);

	Button okButton = new Button("OK");
	okButton.addActionListener(new ErrorOKButtonListener(ed));
	addNewComponent(ed, okButton, 1,
			0, 1, 1, 1, 0.0, 0.0, GridBagConstraints.VERTICAL);

	ed.pack();
	ed.setVisible(true);
    }

    /**
     * displays a dialog box describing an error which occurred.
     */
    void displayErrorDialog(Window w, Exception e) {
	displayErrorDialog(w, e.toString());
    }

    /**
     * displays a dialog box describing the status of an event
     */
    void displayStatusDialog(Window w, String status) {
	ToolDialog sd = new ToolDialog("Status", tool, this, true);

	// find the location of the PolicyTool gui
	Point location = ((w == null) ?
		getLocationOnScreen() : w.getLocationOnScreen());
	sd.setBounds(location.x + 50, location.y + 50, 500, 100);
	sd.setLayout(new GridBagLayout());

	Label label = new Label(status);
	addNewComponent(sd, label, 0,
			0, 0, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);

	Button okButton = new Button("OK");
	okButton.addActionListener(new StatusOKButtonListener(sd));
	addNewComponent(sd, okButton, 1,
			0, 1, 1, 1, 0.0, 0.0, GridBagConstraints.VERTICAL);
	sd.pack();
	sd.setVisible(true);
    }

    /**
     * display the warning log
     */
    void displayWarningLog(Window w) {

	ToolDialog wd = new ToolDialog("Warning", tool, this, true);

	// find the location of the PolicyTool gui
	Point location = ((w == null) ?
		getLocationOnScreen() : w.getLocationOnScreen());
	wd.setBounds(location.x + 50, location.y + 50, 500, 100);
	wd.setLayout(new GridBagLayout());

	TextArea ta = new TextArea();
	ta.setEditable(false);
	for (int i = 0; i < tool.warnings.size(); i++) {
	    ta.append((String)tool.warnings.elementAt(i));
	    ta.append("\n");
	}
	addNewComponent(wd, ta, 0,
			0, 0, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);

	Label label = new Label("");
	addNewComponent(wd, label, 1,
			0, 1, 1, 1, 0.0, 0.0, GridBagConstraints.VERTICAL);

	Button okButton = new Button("OK");
	okButton.addActionListener(new CancelButtonListener(wd));
	addNewComponent(wd, okButton, 2,
			0, 2, 1, 1, 0.0, 0.0, GridBagConstraints.VERTICAL);
	wd.pack();
	wd.setVisible(true);
    }
}

/**
 * General dialog window
 */
class ToolDialog extends Dialog {

    /* necessary constants */
    public final int NOACTION		= 0;
    public final int QUIT		= 1;
    public final int NEW		= 2;
    public final int OPEN		= 3;

    public final String ALL_PERM_CLASS	= "java.security.AllPermission";
    public final String AWT_PERM_CLASS	= "java.awt.AWTPermission";
    public final String FILE_PERM_CLASS	= "java.io.FilePermission";
    public final String NET_PERM_CLASS	= "java.net.NetPermission";
    public final String PROP_PERM_CLASS	= "java.util.PropertyPermission";
    public final String REF_PERM_CLASS	= "java.lang.reflect.ReflectPermission";
    public final String RUN_PERM_CLASS	= "java.lang.RuntimePermission";
    public final String SEC_PERM_CLASS	= "java.security.SecurityPermission";
    public final String SER_PERM_CLASS	= "java.io.SerializablePermission";
    public final String SOCK_PERM_CLASS	= "java.net.SocketPermission";

    /* popup menus */
    public final String PERM		=
	"Permission:                                                       ";
    public final String ALL_PERM	= "AllPermission";
    public final String AWT_PERM	= "AWTPermission";
    public final String FILE_PERM	= "FilePermission";
    public final String NET_PERM	= "NetPermission";
    public final String PROPERTY_PERM	= "PropertyPermission";
    public final String REFLECT_PERM	= "ReflectPermission";
    public final String RUNTIME_PERM	= "RuntimePermission";
    public final String SECURITY_PERM	= "SecurityPermission";
    public final String SERIAL_PERM	= "SerializablePermission";
    public final String SOCKET_PERM	= "SocketPermission";

    /* more popu menus */
    public final String PERM_NAME	=
	"Target Name:                                                    ";
    public final String AWT_BANNER	= "showWindowWithoutWarningBanner";
    public final String AWT_CLIPBOARD	= "accessClipboard";
    public final String AWT_EVENT_QUEUE	= "accessEventQueue";
    public final String FILE_ALL_FILES	= "<<ALL FILES>>";
    public final String NET_PASSWD_AUTH	= "requestPasswordAuthentication";
    public final String NET_SET_AUTH	= "setDefaultAuthenticator";
    public final String REFLECT_CHECK	= "suppressAccessChecks";
    public final String R_ACCESS_CLASS	= "accessClassInPackage.<package name>";
    public final String R_DEFINE_CLASS	= "defineClassInPackage.<package name>";
    public final String R_CREATE_LOADER	= "createClassLoader";
    public final String R_GET_LOADER	= "getClassLoader";
    public final String R_EXIT_VM	= "exitVM";
    public final String R_GET_DOMAIN	= "getProtectionDomain";
    public final String R_SET_DOMAIN	= "setProtectionDomain";
    public final String R_SET_FACTORY	= "setFactory";
    public final String R_READ_FD	= "readFileDescriptor";
    public final String R_WRITE_FD	= "writeFileDescriptor";
    public final String R_SET_IO	= "setIO";
    public final String R_LOAD_LIB	= "loadLibrary.<library name>";
    public final String R_ACCESS_MEMBER	= "accessDeclaredMembers.<class name>";
    public final String R_PRINT		= "queuePrintJob";
    public final String R_THREAD	= "modifyThread";
    public final String R_THREAD_GROUP	= "modifyThreadGroup";
    public final String SEC_GET_POLICY	= "getPolicy";
    public final String SEC_SET_POLICY	= "setPolicy";
    public final String SEC_GET_PROP	= "getProperty.<property name>";
    public final String SEC_SET_PROP	= "setProperty.<property name>";
    public final String SEC_INSERT_PROV	= "insertProvider.<provider name>";
    public final String SEC_REM_PROV	= "removeProvider.<provider name>";
    public final String SEC_SET_SCOPE	= "setSystemScope";
    public final String SEC_SET_PUBKEY	= "setIdentityPublicKey";
    public final String SEC_SET_ID	= "setIdentityInfo";
    public final String SEC_ADD_CERT	= "addIdentityCertificate";
    public final String SEC_REMOVE_CERT	= "removeIdentityCertificate";
    public final String SEC_CLEAR_PROP	=
				"clearProviderProperties.<provider name>";
    public final String SEC_PUT_PROP	=
				"putProviderProperty.<provider name>";
    public final String SEC_REMOVE_PROP	=
				"removeProviderProperty.<provider name>";
    public final String SEC_GET_PRIVKEY	= "getSignerPrivateKey";
    public final String SEC_SET_KEYPAIR	= "setSignerKeyPair";
    public final String SER_SUBCLASS	= "enableSubclassImplementation";
    public final String SER_SUBSTITUTE	= "enableSubstitution";

    /* and more popup menus */
    public final String PERM_ACTIONS	=
	"Actions:                                                             ";
    public final String FILE_PERM_READ		= "read";
    public final String FILE_PERM_WRITE		= "write";
    public final String FILE_PERM_DELETE	= "delete";
    public final String FILE_PERM_EXECUTE	= "execute";
    public final String FILE_PERM_ALL	= "read, write, delete, execute";
    public final String PROP_PERM_READ		= "read";
    public final String PROP_PERM_WRITE		= "write";
    public final String PROP_PERM_ALL		= "read, write";
    public final String SOCKET_PERM_ACCEPT	= "accept";
    public final String SOCKET_PERM_CONNECT	= "connect";
    public final String SOCKET_PERM_LISTEN	= "listen";
    public final String SOCKET_PERM_RESOLVE	= "resolve";
    public final String SOCKET_PERM_ALL	= "accept, connect, listen, resolve";

    /* gridbag index for display OverWriteFile (OW) components */
    public final int OW_LABEL			= 0;
    public final int OW_EMPTY_LABEL1		= 1;
    public final int OW_OK_BUTTON		= 2;
    public final int OW_CANCEL_BUTTON		= 3;

    /* gridbag index for display PolicyEntry (PE) components */
    public final int PE_CODEBASE_LABEL		= 0;
    public final int PE_CODEBASE_TEXTFIELD	= 1;
    public final int PE_SIGNEDBY_LABEL		= 2;
    public final int PE_SIGNEDBY_TEXTFIELD	= 3;
    public final int PE_EMPTY_LABEL1		= 4;

    public final int PE_PANEL1			= 5;
    public final int PE_ADD_PERM_BUTTON		= 0;
    public final int PE_EMPTY_LABEL2		= 1;
    public final int PE_EDIT_PERM_BUTTON	= 2;
    public final int PE_EMPTY_LABEL3		= 3;
    public final int PE_REMOVE_PERM_BUTTON	= 4;

    public final int PE_EMPTY_LABEL4		= 6;
    public final int PE_PERM_LIST		= 7;
    public final int PE_EMPTY_LABEL5		= 8;

    public final int PE_PANEL2			= 9;
    public final int PE_CANCEL_BUTTON		= 0;
    public final int PE_EMPTY_LABEL6		= 1;
    public final int PE_DONE_BUTTON		= 2;

    /* the gridbag index for components in the Permission Dialog (PD) */
    public final int PD_DESC_LABEL		= 0;
    public final int PD_EMPTY_LABEL1		= 1;
    public final int PD_PERM_CHOICE		= 2;
    public final int PD_EMPTY_LABEL2		= 3;
    public final int PD_PERM_TEXTFIELD		= 4;
    public final int PD_NAME_CHOICE		= 5;
    public final int PD_EMPTY_LABEL3		= 6;
    public final int PD_NAME_TEXTFIELD		= 7;
    public final int PD_ACTIONS_CHOICE		= 8;
    public final int PD_EMPTY_LABEL4		= 9;
    public final int PD_ACTIONS_TEXTFIELD	= 10;
    public final int PD_SIGNEDBY_LABEL		= 11;
    public final int PD_EMPTY_LABEL5		= 12;
    public final int PD_SIGNEDBY_TEXTFIELD	= 13;
    public final int PD_EMPTY_LABEL6		= 14;
    public final int PD_CANCEL_BUTTON		= 15;
    public final int PD_OK_BUTTON		= 16;

    /* modes for KeyStore */
    public final int CHANGE_KEYSTORE		= 0;

    /* the gridbag index for components in the Change KeyStore Dialog (KSD) */
    public final int KSD_NAME_LABEL		= 0;
    public final int KSD_NAME_TEXTFIELD		= 1;
    public final int KSD_EMPTY_LABEL1		= 2;
    public final int KSD_TYPE_LABEL		= 3;
    public final int KSD_TYPE_TEXTFIELD		= 4;
    public final int KSD_EMPTY_LABEL2		= 5;
    public final int KSD_CANCEL_BUTTON		= 6;
    public final int KSD_OK_BUTTON		= 7;

    /* the gridbag index for components in the User Save Changes Dialog (USC) */
    public final int USC_LABEL			= 0;
    public final int USC_EMPTY_LABEL		= 1;
    public final int USC_PANEL			= 2;
    public final int USC_YES_BUTTON		= 0;
    public final int USC_EMPTY_LABEL2		= 1;
    public final int USC_NO_BUTTON		= 2;
    public final int USC_EMPTY_LABEL3		= 3;
    public final int USC_CANCEL_BUTTON		= 4;

    /* gridbag index for the ConfirmRemovePolicyEntryDialog (CRPE) */
    public final int CRPE_EMPTY_LABEL1		= 0;
    public final int CRPE_LABEL1		= 1;
    public final int CRPE_EMPTY_LABEL2		= 2;
    public final int CRPE_LABEL2		= 3;

    /* some private finals */
    private final int PERMISSION		= 0;
    private final int PERMISSION_NAME		= 1;
    private final int PERMISSION_ACTIONS	= 2;
    private final int PERMISSION_SIGNEDBY	= 3;

    PolicyTool tool;
    ToolWindow tw;

    ToolDialog(String title, PolicyTool tool, ToolWindow tw, boolean modal) {
	super(tw, modal);
	setTitle(title);
	this.tool = tool;
	this.tw = tw;
    }

    /**
     * ask user if they want to overwrite an existing file
     */
    void displayOverWriteFileDialog(String filename, int nextEvent) {

	// find where the PolicyTool gui is
	Point location = tw.getLocationOnScreen();
	setBounds(location.x + 75, location.y + 100, 400, 150);
	setLayout(new GridBagLayout());

	// ask the user if they want to over write the existing file
	Label label = new Label
		("OK to overwrite existing file " + filename + "?");
	tw.addNewComponent(this, label, OW_LABEL,
		0, 0, 2, 1, 0.0, 0.0, GridBagConstraints.BOTH);

	// empty space
	label = new Label("");
	tw.addNewComponent(this, label, OW_EMPTY_LABEL1,
		0, 1, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);

	// OK button
	Button button = new Button("OK");
	button.addActionListener(new OverWriteFileOKButtonListener
		(tool, tw, this, filename, nextEvent));
	tw.addNewComponent(this, button, OW_OK_BUTTON,
		0, 2, 1, 1, 0.0, 0.0, GridBagConstraints.VERTICAL);

	// Cancel button
	// -- if the user hits cancel, do NOT go on to the next event
	button = new Button("Cancel");
	button.addActionListener(new CancelButtonListener(this));
	tw.addNewComponent(this, button, OW_CANCEL_BUTTON,
		1, 2, 1, 1, 0.0, 0.0, GridBagConstraints.VERTICAL);

	setVisible(true);
    }

    /**
     * pop up a dialog so the user can enter info to add a new PolicyEntry
     * - if edit is TRUE, then the user is editing an existing entry
     *   and we should display the original info as well.
     *
     * - the other reason we need the 'edit' boolean is we need to know
     *   when we are adding a NEW policy entry.  in this case, we can
     *   not simply update the existing entry, because it doesn't exist.
     *   we ONLY update the GUI listing/info, and then when the user
     *   finally clicks 'OK' or 'DONE', then we can collect that info
     *   and add it to the policy.
     */
    void displayPolicyEntryDialog(boolean edit) {

	int listIndex = 0;
	PolicyEntry entries[] = null;
	List permList = new List(15, false);
	permList.addActionListener
		(new EditPermButtonListener(tool, tw, this, edit));

	// find where the PolicyTool gui is
	Point location = tw.getLocationOnScreen();
	setBounds(location.x + 75, location.y + 100, 600, 500);
	setLayout(new GridBagLayout());
	setResizable(false);

	if (edit) {
	    // get the selected item
	    entries = tool.getEntry();
	    List policyList = (List)tw.getComponent(tw.MW_POLICY_LIST);
	    listIndex = policyList.getSelectedIndex();

	    // get permission list
	    Vector permissions =
		entries[listIndex].getGrantEntry().permissionEntries;
	    for (int i = 0; i < permissions.size(); i++) {
		String permString = null;
		PolicyParser.PermissionEntry nextPerm =
			(PolicyParser.PermissionEntry)permissions.elementAt(i);
		StringWriter sw = new StringWriter();
		PrintWriter pw = new PrintWriter(sw);
		nextPerm.write(pw);
		if (File.separatorChar == '\\' &&
		    nextPerm.permission.equals(FILE_PERM_CLASS)) {
		    /*
		     * XXX  in windows, the double backslash originally in
		     *      the policy file gets automatically converted
		     *      into a single backslash every time we read it in.
		     *      so, we have to add the double backslash again.
		     *      sigh ...
		     */
		    nextPerm.name = addSingleBackSlash(nextPerm.name);
		    permString = addSingleBackSlash(sw.toString());
		} else {
		    permString = sw.toString();
		}
		permString = permString.substring
				(0, permString.indexOf(";") + 1);
		permList.add(permString);
	    }
	}

	// codebase label and textfield
	Label label = new Label("CodeBase:");
	tw.addNewComponent(this, label, PE_CODEBASE_LABEL,
		0, 0, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);
	TextField tf;
	tf = (edit ?
		new TextField(entries[listIndex].getGrantEntry().codeBase, 60) :
		new TextField(60));
	tw.addNewComponent(this, tf, PE_CODEBASE_TEXTFIELD,
		1, 0, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);

	// signedby label and textfield
	label = new Label("SignedBy:");
	tw.addNewComponent(this, label, PE_SIGNEDBY_LABEL,
		0, 1, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);
	tf = (edit ?
		new TextField(entries[listIndex].getGrantEntry().signedBy, 60) :
		new TextField(60));
	tw.addNewComponent(this, tf, PE_SIGNEDBY_TEXTFIELD,
		1, 1, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);

	// empty space
	label = new Label("");
	tw.addNewComponent(this, label, PE_EMPTY_LABEL1,
		0, 2, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);

	// panel for permission buttons
	Panel panel = new Panel();
	panel.setLayout(new GridBagLayout());
 
	Button button = new Button("Add Permission");
	button.addActionListener(new AddPermButtonListener
						(tool, tw, this, edit));
	tw.addNewComponent(panel, button, PE_ADD_PERM_BUTTON,
		0, 0, 1, 1, 0.0, 0.0, GridBagConstraints.VERTICAL);
 
	label = new Label("");
	tw.addNewComponent(panel, label, PE_EMPTY_LABEL2,
		1, 0, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);
 
	button = new Button("Edit Permission");
	button.addActionListener(new EditPermButtonListener
						(tool, tw, this, edit));
	tw.addNewComponent(panel, button, PE_EDIT_PERM_BUTTON,
		2, 0, 1, 1, 0.0, 0.0, GridBagConstraints.VERTICAL);
 
	label = new Label("");
	tw.addNewComponent(panel, label, PE_EMPTY_LABEL3,
		3, 0, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);
 
	button = new Button("Remove Permission");
	button.addActionListener(new RemovePermButtonListener
					(tool, tw, this, edit));
	tw.addNewComponent(panel, button, PE_REMOVE_PERM_BUTTON,
		4, 0, 1, 1, 0.0, 0.0, GridBagConstraints.VERTICAL);
 
	tw.addNewComponent(this, panel, PE_PANEL1,
		0, 3, 2, 1, 0.0, 0.0, GridBagConstraints.VERTICAL);

	// empty space
	label = new Label("");
	tw.addNewComponent(this, label, PE_EMPTY_LABEL4,
		0, 4, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);

	// permission list
	tw.addNewComponent(this, permList, PE_PERM_LIST,
		0, 5, 3, 1, 0.0, 0.0, GridBagConstraints.BOTH);
 
	// empty space
	label = new Label("");
	tw.addNewComponent(this, label, PE_EMPTY_LABEL5,
		0, 6, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);

	// panel for Done and Cancel buttons
	panel = new Panel();
	panel.setLayout(new GridBagLayout());

	// Cancel Button
	button = new Button("Cancel");
	button.addActionListener(new CancelButtonListener(this));
	tw.addNewComponent(panel, button, PE_CANCEL_BUTTON,
		2, 0, 1, 1, 0.0, 0.0, GridBagConstraints.VERTICAL);

	// empty label
	label = new Label("");
	tw.addNewComponent(panel, label, PE_EMPTY_LABEL6,
		1, 0, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);

	// Done Button
	button = new Button("Done");
	button.addActionListener
		(new AddEntryDoneButtonListener(tool, tw, this, edit));
	tw.addNewComponent(panel, button, PE_DONE_BUTTON,
		0, 0, 1, 1, 0.0, 0.0, GridBagConstraints.VERTICAL);

	// add the panel
	tw.addNewComponent(this, panel, PE_PANEL2,
		0, 7, 2, 1, 0.0, 0.0, GridBagConstraints.VERTICAL);

	setVisible(true);
    }

    /**
     * Read all the Policy information data in the dialog box
     * and construct a PolicyEntry object with it.
     */
    PolicyEntry getPolicyEntryFromDialog() 
    throws InvalidParameterException, MalformedURLException,
    NoSuchMethodException, ClassNotFoundException, InstantiationException,
    IllegalAccessException, InvocationTargetException,
    CertificateException, IOException, Exception {

	// get the Codebase
	TextField tf = (TextField)getComponent(PE_CODEBASE_TEXTFIELD);
	String codebase = null;
	if (tf.getText().trim().equals("") == false)
		codebase = new String(tf.getText().trim());
 
	// get the SignedBy
	tf = (TextField)getComponent(PE_SIGNEDBY_TEXTFIELD);
	String signedby = null;
	if (tf.getText().trim().equals("") == false)
		signedby = new String(tf.getText().trim());
 
	// construct a new GrantEntry
	PolicyParser.GrantEntry ge =
			new PolicyParser.GrantEntry(signedby, codebase);

	// get the new Permissions
	Vector perms = new Vector();
	List permList = (List)getComponent(PE_PERM_LIST);
	String permStrings[] = permList.getItems();
	for (int i = 0; i < permStrings.length; i++) {
	    PolicyParser.PermissionEntry pppe = new PolicyParser.PermissionEntry
		(parsePermissionString(permStrings[i], PERMISSION),
		parsePermissionString(permStrings[i], PERMISSION_NAME),
		parsePermissionString(permStrings[i], PERMISSION_ACTIONS));
	    pppe.signedBy =
		parsePermissionString(permStrings[i], PERMISSION_SIGNEDBY);
	    perms.addElement(pppe);
	}
	ge.permissionEntries = perms;

	// construct a new PolicyEntry object
	PolicyEntry entry = new PolicyEntry(tool, ge);

	return entry;
    }

    /**
     * display a dialog box for the user to enter KeyStore information
     */
    void keyStoreDialog(int mode) {

	// find where the PolicyTool gui is
	Point location = tw.getLocationOnScreen();
	setBounds(location.x + 25, location.y + 100, 500, 200);
	setLayout(new GridBagLayout());

	if (mode == CHANGE_KEYSTORE) {

	    // KeyStore label and textfield
	    Label label = new Label("New KeyStore URL:");
	    tw.addNewComponent(this, label, KSD_NAME_LABEL,
			0, 0, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);
	    TextField tf = new TextField(tool.getKeyStoreName(), 30);
	    tw.addNewComponent(this, tf, KSD_NAME_TEXTFIELD,
			1, 0, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);

	    // empty space
	    label = new Label("");
	    tw.addNewComponent(this, label, KSD_EMPTY_LABEL1,
			0, 2, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);

	    // KeyStore type and textfield
	    label = new Label("New KeyStore Type:");
	    tw.addNewComponent(this, label, KSD_TYPE_LABEL,
			0, 3, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);
	    tf = new TextField(tool.getKeyStoreType(), 30);
	    tw.addNewComponent(this, tf, KSD_TYPE_TEXTFIELD,
			1, 3, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);

	    // empty space
	    label = new Label("");
	    tw.addNewComponent(this, label, KSD_EMPTY_LABEL2,
			0, 4, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);

	    // cancel button
	    Button cancelButton = new Button("Cancel");
	    cancelButton.addActionListener(new CancelButtonListener(this));
	    tw.addNewComponent(this, cancelButton, KSD_CANCEL_BUTTON,
			1, 5, 1, 1, 0.0, 0.0, GridBagConstraints.VERTICAL);

	    // OK button
	    Button okButton = new Button("OK");
	    okButton.addActionListener
			(new ChangeKeyStoreOKButtonListener(tool, tw, this));
	    tw.addNewComponent(this, okButton, KSD_OK_BUTTON,
			0, 5, 1, 1, 0.0, 0.0, GridBagConstraints.VERTICAL);

	}
	setVisible(true);
    }

    /**
     * display a dialog box for the user to input Permission info
     *
     * if editPolicyEntry is false, then we are adding Permissions to
     * a new PolicyEntry, and we only update the GUI listing
     * with the new Permission.
     *
     * if edit is true, then we are editing an existing Permission entry.
     */
    void displayPermissionDialog(boolean editPolicyEntry, boolean edit) {

	PolicyParser.PermissionEntry editMe = null;

	// get the Permission selected from the Permission List
	List permList = (List)getComponent(PE_PERM_LIST);
	int permIndex = permList.getSelectedIndex();

	if (!editPolicyEntry && edit) {

	    // construct a PolicyParser.PermissionEntry from the
	    // selected item in the Permission List

	    String perm = parsePermissionString
			(permList.getItem(permIndex), PERMISSION);
	    String permName = parsePermissionString
			(permList.getItem(permIndex), PERMISSION_NAME);
	    String permActions = parsePermissionString
			(permList.getItem(permIndex), PERMISSION_ACTIONS);
	    editMe = new PolicyParser.PermissionEntry
				(perm, permName, permActions);
	    editMe.signedBy = parsePermissionString
			(permList.getItem(permIndex), PERMISSION_SIGNEDBY);

	} else if (editPolicyEntry && edit) {

	    // get the PolicyEntry selected from the main ToolWindow
	    List policyList = (List)tw.getComponent(tw.MW_POLICY_LIST);
	    int policyIndex = policyList.getSelectedIndex();

	    PolicyEntry entries[] = tool.getEntry();
	    Vector perms =
			entries[policyIndex].getGrantEntry().permissionEntries;
	    editMe = (PolicyParser.PermissionEntry)perms.elementAt(permIndex);
	}

	ToolDialog newTD = new ToolDialog("Permissions", tool, tw, true);

	// find where the PolicyTool gui is
	Point location = getLocationOnScreen();
	newTD.setBounds(location.x + 50, location.y + 100, 700, 250);
	newTD.setLayout(new GridBagLayout());
	newTD.setResizable(false);

	// description label
	Label label = (edit ?
			new Label("Edit Permission:") :
			new Label("Add New Permission:"));
	tw.addNewComponent(newTD, label, PD_DESC_LABEL,
			0, 0, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);

	// empty space
	label = new Label("");
	tw.addNewComponent(newTD, label, PD_EMPTY_LABEL1,
			0, 1, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);
 
	// permission choice
	Choice choice = new Choice();
	choice.add(PERM);
	choice.add(ALL_PERM);
	choice.add(AWT_PERM);
	choice.add(FILE_PERM);
	choice.add(NET_PERM);
	choice.add(PROPERTY_PERM);
	choice.add(REFLECT_PERM);
	choice.add(RUNTIME_PERM);
	choice.add(SECURITY_PERM);
	choice.add(SERIAL_PERM);
	choice.add(SOCKET_PERM);
	choice.addItemListener(new PermissionMenuListener(newTD));
	tw.addNewComponent(newTD, choice, PD_PERM_CHOICE,
			0, 2, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);

	// empty label
	tw.addNewComponent(newTD, new Label(""), PD_EMPTY_LABEL2,
			1, 2, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);
 
	// permission textfield
	TextField tf;
	tf = (edit ? new TextField(editMe.permission, 30) : new TextField(30));
	if (edit)
	    setPermissions(editMe.permission, choice);
	tw.addNewComponent(newTD, tf, PD_PERM_TEXTFIELD,
			2, 2, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);
 
	// name label and textfield
	choice = new Choice();
	choice.add(PERM_NAME);
	choice.addItemListener(new PermissionNameMenuListener(newTD));
	tf = (edit ? new TextField(editMe.name, 40) : new TextField(40));
	if (edit)
	    setPermissionNames(editMe.permission, choice, tf);
	tw.addNewComponent(newTD, choice, PD_NAME_CHOICE,
			0, 3, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);
	tw.addNewComponent(newTD, new Label(""), PD_EMPTY_LABEL3,
			1, 3, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);
	tw.addNewComponent(newTD, tf, PD_NAME_TEXTFIELD,
			2, 3, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);
 
	// actions label and textfield
	choice = new Choice();
	choice.add(PERM_ACTIONS);
	choice.addItemListener(new PermissionActionsMenuListener(newTD));
	tf = (edit ? new TextField(editMe.action, 40) : new TextField(40));
	if (edit)
	    setPermissionActions(editMe.permission, choice, tf);
	tw.addNewComponent(newTD, choice, PD_ACTIONS_CHOICE,
			0, 4, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);
	tw.addNewComponent(newTD, new Label(""), PD_EMPTY_LABEL4,
			1, 4, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);
	tw.addNewComponent(newTD, tf, PD_ACTIONS_TEXTFIELD,
			2, 4, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);

	// signedby label and textfield
	label = new Label("Signed By:");
	tw.addNewComponent(newTD, label, PD_SIGNEDBY_LABEL,
			0, 5, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);
	tw.addNewComponent(newTD, new Label(""), PD_EMPTY_LABEL5,
			1, 5, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);
	tf = (edit ? new TextField(editMe.signedBy, 40) : new TextField(40));
	tw.addNewComponent(newTD, tf, PD_SIGNEDBY_TEXTFIELD,
			2, 5, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);

	// empty space
	label = new Label("");
	tw.addNewComponent(newTD, label, PD_EMPTY_LABEL6,
			0, 6, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);
 
	// cancel button
	Button cancelButton = new Button("Cancel");
	cancelButton.addActionListener(new CancelButtonListener(newTD));
	tw.addNewComponent(newTD, cancelButton, PD_CANCEL_BUTTON,
			2, 7, 1, 1, 0.0, 0.0, GridBagConstraints.VERTICAL);

	// OK button
	Button okButton = new Button("OK");
	if (editPolicyEntry)
	    okButton.addActionListener(new EditPolicyEditPermOKButtonListener
					(tool, tw, this, newTD, edit));
	else
	    okButton.addActionListener(new NewPolicyEntryAddPermOKButtonListener
					(tool, tw, this, newTD, edit));
	tw.addNewComponent(newTD, okButton, PD_OK_BUTTON,
		0, 7, 1, 1, 0.0, 0.0, GridBagConstraints.VERTICAL);

	newTD.setVisible(true);
    }

    /**
     * construct a Permission object from the Permission Info Dialog Box
     */
    PolicyParser.PermissionEntry getPermFromDialog() {

	TextField tf = (TextField)getComponent(PD_PERM_TEXTFIELD);
	String permission = new String(tf.getText().trim());
	tf = (TextField)getComponent(PD_NAME_TEXTFIELD);
	String name = null;
	if (tf.getText().trim().equals("") == false)
	    name = new String(tf.getText().trim());
	if (permission.equals("") ||
	    (!permission.equals(ALL_PERM_CLASS) && name == null)) {
	    throw new InvalidParameterException
			("Permission and Target Name must have a value");
	}

	// get the Actions
	tf = (TextField)getComponent(PD_ACTIONS_TEXTFIELD);
	String actions = null;
	if (tf.getText().trim().equals("") == false)
	    actions = new String(tf.getText().trim());

	// get the Signed By
	tf = (TextField)getComponent(PD_SIGNEDBY_TEXTFIELD);
	String signedBy = null;
	if (tf.getText().trim().equals("") == false)
	    signedBy = new String(tf.getText().trim());

	PolicyParser.PermissionEntry pppe = new PolicyParser.PermissionEntry
				(permission, name, actions);
	pppe.signedBy = signedBy;
 
	// see if the signers have public keys
	if (signedBy != null) {
		String signers[] = tool.parseSigners(pppe.signedBy);
		for (int i = 0; i < signers.length; i++) {
		try {
		    PublicKey pubKey = tool.getPublicKeyAlias(signers[i]);
		    if (pubKey == null) {
			tool.warnings.addElement
			    ("Warning: A public key for alias '" + signers[i] +
			    "' does not exist.");
			tw.displayStatusDialog(this,
			    "Warning: A public key for alias '" + signers[i] +
			    "' does not exist.");
		    }
		} catch (Exception e) {
		    tw.displayErrorDialog(this, e);
		}
	    }
	}
	return pppe;
    }

    /**
     * confirm that the user REALLY wants to remove the Policy Entry
     */
    void displayConfirmRemovePolicyEntry() {

	// find the entry to be removed
	List list = (List)tw.getComponent(tw.MW_POLICY_LIST);
	int index = list.getSelectedIndex();
	PolicyEntry entries[] = tool.getEntry();

	// find where the PolicyTool gui is
	Point location = tw.getLocationOnScreen();
	setBounds(location.x + 25, location.y + 100, 600, 400);
	setLayout(new GridBagLayout());

	// empty space
	Label label = new Label("");
	tw.addNewComponent(this, label, CRPE_EMPTY_LABEL1,
			0, 0, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);

	// ask the user do they really want to do this?
	label = new Label("Remove this Policy Entry?");
	tw.addNewComponent(this, label, CRPE_LABEL1,
			1, 0, 2, 1, 0.0, 0.0, GridBagConstraints.BOTH);

	// empty space
	label = new Label("");
	tw.addNewComponent(this, label, CRPE_EMPTY_LABEL2,
		1, 1, 2, 1, 0.0, 0.0, GridBagConstraints.BOTH);

	// display the policy entry
	label = new Label("Grant " + entries[index].codebaseToString());
	tw.addNewComponent(this, label, CRPE_LABEL2,
			1, 2, 2, 1, 0.0, 0.0, GridBagConstraints.BOTH);
	Vector perms = entries[index].getGrantEntry().permissionEntries;
	for (int i = 0; i < perms.size(); i++) {
	    PolicyParser.PermissionEntry nextPerm =
			(PolicyParser.PermissionEntry)perms.elementAt(i);
	    StringWriter sw = new StringWriter();
	    PrintWriter pw = new PrintWriter(sw);
	    nextPerm.write(pw);
	    String permString = sw.toString();
	    label = new Label("    " +
		permString.substring(0, permString.indexOf(";") + 1));
	    tw.addNewComponent(this, label, CRPE_LABEL2 + 1 + i,
			1, 2 + 1 + i, 2, 1, 0.0, 0.0, GridBagConstraints.BOTH);
	}

	// empty space
	label = new Label("");
	tw.addNewComponent(this, label, CRPE_LABEL2 + 1 + perms.size(),
		1, 3 + perms.size(), 2, 1, 0.0, 0.0, GridBagConstraints.BOTH);

	// cancel button
	Button cancelButton = new Button("Cancel");
	cancelButton.addActionListener(new CancelButtonListener(this));
	tw.addNewComponent(this, cancelButton, CRPE_LABEL2 + 2 + perms.size(),
	    2, 4 + perms.size(), 1, 1, 0.0, 0.0, GridBagConstraints.VERTICAL);

	// OK button
	Button okButton = new Button("OK");
	okButton.addActionListener
		(new ConfirmRemovePolicyEntryOKButtonListener(tool, tw, this));
	tw.addNewComponent(this, okButton, CRPE_LABEL2 + 3 + perms.size(),
	    1, 4 + perms.size(), 1, 1, 0.0, 0.0, GridBagConstraints.VERTICAL);

	// empty space
	label = new Label("");
	tw.addNewComponent(this, label, CRPE_LABEL2 + 4 + perms.size(),
	    1, 5 + perms.size(), 2, 1, 0.0, 0.0, GridBagConstraints.BOTH);

	pack();
	setVisible(true);
    }

    /**
     * perform SAVE AS
     */
    void displaySaveAsDialog(int nextEvent) {

	// pop up a dialog box for the user to enter a filename.
	FileDialog fd = new FileDialog(tw, "Save As", FileDialog.SAVE);
	fd.setVisible(true);

	// see if the user hit cancel
	if (fd.getFile() == null ||
	    fd.getFile().equals(""))
	    return;
		
	// get the entered filename
	String filename = new String(fd.getDirectory() + fd.getFile());
	fd.dispose();

	// see if the file already exists
	File saveAsFile = new File(filename);
	if (saveAsFile.exists()) {
	    // display a dialog box for the user to enter policy info
	    ToolDialog td = new ToolDialog("Overwrite File", tool, tw, true);
	    td.displayOverWriteFileDialog(filename, nextEvent);
	} else {
	    try {
		// save the policy entries to a file
		tool.savePolicy(filename);

		// display status
		tw.displayStatusDialog
			(null, "Policy successfully written to " + filename);

		// display the new policy filename
		TextField newFilename = (TextField)tw.getComponent
				(tw.MW_FILENAME_TEXTFIELD);
		newFilename.setText(filename);
		tw.setVisible(true);

		// now continue with the originally requested command
		// (QUIT, NEW, or OPEN)
		userSaveContinue(tool, tw, this, nextEvent);

	    } catch (FileNotFoundException fnfe) {
		if (filename == null || filename.equals(""))
		    tw.displayErrorDialog(null, new FileNotFoundException
					("null filename"));
		else
		    tw.displayErrorDialog(null, new FileNotFoundException
					(filename + " not found"));
	    } catch (Exception ee) {
		tw.displayErrorDialog(null, ee);
	    }
	}
    }

    /**
     * ask user if they want to save changes
     */
    void displayUserSave(int select) {

	if (tool.modified == true) {

	    // find where the PolicyTool gui is
	    Point location = tw.getLocationOnScreen();
	    setBounds(location.x + 75, location.y + 100, 400, 150);
	    setLayout(new GridBagLayout());

	    Label label = new Label("     Save changes?");
	    tw.addNewComponent(this, label, USC_LABEL,
			0, 0, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);
	    Label empty = new Label("");
	    tw.addNewComponent(this, empty, USC_EMPTY_LABEL,
			0, 1, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);

	    Panel panel = new Panel();
	    panel.setLayout(new GridBagLayout());

	    Button yesButton = new Button("Yes");
	    yesButton.addActionListener
			(new UserSaveYesButtonListener(this, tool, tw, select));
	    tw.addNewComponent(panel, yesButton, USC_YES_BUTTON,
			0, 0, 1, 1, 0.0, 0.0, GridBagConstraints.VERTICAL);
	    empty = new Label("");
	    tw.addNewComponent(panel, empty, USC_EMPTY_LABEL2,
			1, 0, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);
	    Button noButton = new Button("No");
	    noButton.addActionListener
			(new UserSaveNoButtonListener(this, tool, tw, select));
	    tw.addNewComponent(panel, noButton, USC_NO_BUTTON,
			2, 0, 1, 1, 0.0, 0.0, GridBagConstraints.VERTICAL);
	    empty = new Label("");
	    tw.addNewComponent(panel, empty, USC_EMPTY_LABEL3,
			3, 0, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);
	    Button cancelButton = new Button("Cancel");
	    cancelButton.addActionListener
			(new UserSaveCancelButtonListener(this));
	    tw.addNewComponent(panel, cancelButton, USC_CANCEL_BUTTON,
			4, 0, 1, 1, 0.0, 0.0, GridBagConstraints.VERTICAL);

	    tw.addNewComponent(this, panel, USC_PANEL,
			0, 2, 1, 1, 0.0, 0.0, GridBagConstraints.BOTH);
	    
	    pack();
	    setVisible(true);
	} else {
	    // just do the original request (QUIT, NEW, or OPEN)
	    userSaveContinue(tool, tw, this, select);
	}
    }

    /**
     * when the user sees the 'YES', 'NO', 'CANCEL' buttons on the
     * displayUserSave dialog, and the click on one of them,
     * we need to continue the originally requested action
     * (either QUITting, opening NEW policy file, or OPENing an existing 
     * policy file.  do that now.
     */
    void userSaveContinue(PolicyTool tool, ToolWindow tw,
			ToolDialog us, int select) {

	// now either QUIT, open a NEW policy file, or OPEN an existing policy
	switch(select) {
	case us.QUIT:

	    tw.setVisible(false);
	    tw.dispose();
	    System.exit(0);

	case us.NEW:

	    try {
		tool.openPolicy(null);
	    } catch (Exception ee) {
		tool.modified = false;
		tw.displayErrorDialog(null, ee);
	    }
	
	    // display the policy entries via the policy list textarea
	    List list = new List(40, false);
	    list.addActionListener(new PolicyListListener(tool, tw));
	    tw.replacePolicyList(list);

	    // display null policy filename and keystore
	    TextField newFilename = (TextField)
				tw.getComponent(tw.MW_FILENAME_TEXTFIELD);
	    newFilename.setText("");
	    TextField ksName = (TextField)
				tw.getComponent(tw.MW_KEYSTORE_TEXTFIELD);
	    ksName.setText("");
	    tw.setVisible(true);
	    break;

	case us.OPEN:

	    // pop up a dialog box for the user to enter a filename.
	    FileDialog fd = new FileDialog(tw, "Open", FileDialog.LOAD);
	    fd.setVisible(true);
		
	    // see if the user hit 'cancel'
	    if (fd.getFile() == null ||
		fd.getFile().equals(""))
		return;

	    // get the entered filename
	    String filename = new String(fd.getDirectory() + fd.getFile());

	    try {
		// open and read the policy file
		tool.openPolicy(filename);

	    } catch (FileNotFoundException fnfe) {
		if (filename == null || filename.equals(""))
		    tw.displayErrorDialog(null, new FileNotFoundException
					("null filename"));
		else
		    tw.displayErrorDialog(null, new FileNotFoundException
					(filename + " not found"));
	    } catch (PolicyParser.ParsingException pppe) {
		tw.displayErrorDialog(null,
			"Error parsing policy file " +
			filename +
			": " +
			pppe.getMessage());
		tool.warnings.addElement
			("Error: Could not open policy file, " + filename +
			", because of parsing error: " + pppe.getMessage());
	    } catch (NoSuchMethodException nsme) {
		tw.displayErrorDialog(null, new Exception
		("Permission could not be mapped to an appropriate class"));
	    } catch (Exception ee) {
		tw.displayErrorDialog(null, ee);
	    } finally {
		tool.modified = false;
	    }

	    // display the policy entries via the policy list textarea
	    list = new List(40, false);
	    list.addActionListener(new PolicyListListener(tool, tw));
	    PolicyEntry entries[] = tool.getEntry();
	    if (entries != null) {
		for (int i = 0; i < entries.length; i++)
		    list.add(entries[i].codebaseToString());
	    }
	    tw.replacePolicyList(list);

	    // display the new policy filename and keystore
	    newFilename = (TextField)tw.getComponent(tw.MW_FILENAME_TEXTFIELD);
	    newFilename.setText(filename);
	    ksName = (TextField)tw.getComponent(tw.MW_KEYSTORE_TEXTFIELD);
	    if (tool.getKeyStoreType() != null &&
		tool.getKeyStoreType().length() > 0)
		ksName.setText(tool.getKeyStoreName() +
				", "
				+ tool.getKeyStoreType());
	    else
		ksName.setText(tool.getKeyStoreName());
	    tw.setVisible(true);

	    // inform user of warnings
	    if (tool.newWarning == true)
		tw.displayStatusDialog(null,
			"Errors have occurred while opening the policy configuration.  View the Warning Log for more information.");
	    break;
	}
    }

    /*
     * return a Menu list of actions for a given permission
     */
    void setPermissionActions(String permission, Choice actions,
				TextField field) {

	actions.removeAll();
	actions.add(PERM_ACTIONS);
	field.setEditable(true);

	if (permission.equals(ALL_PERM_CLASS)) {
	    field.setEditable(false);
	} else if (permission.equals(AWT_PERM_CLASS)) {
	    field.setEditable(false);
	} else if (permission.equals(FILE_PERM_CLASS)) {
	    actions.add(FILE_PERM_READ);
	    actions.add(FILE_PERM_WRITE);
	    actions.add(FILE_PERM_DELETE);
	    actions.add(FILE_PERM_EXECUTE);
	    actions.add(FILE_PERM_ALL);
	} else if (permission.equals(NET_PERM_CLASS)) {
	    field.setEditable(false);
	} else if (permission.equals(PROP_PERM_CLASS)) {
	    actions.add(PROP_PERM_READ);
	    actions.add(PROP_PERM_WRITE);
	    actions.add(PROP_PERM_ALL);
	} else if (permission.equals(REF_PERM_CLASS)) {
	    field.setEditable(false);
	} else if (permission.equals(RUN_PERM_CLASS)) {
	    field.setEditable(false);
	} else if (permission.equals(SEC_PERM_CLASS)) {
	    field.setEditable(false);
	} else if (permission.equals(SER_PERM_CLASS)) {
	    field.setEditable(false);
	} else if (permission.equals(SOCK_PERM_CLASS)) {
	    actions.add(SOCKET_PERM_LISTEN);
	    actions.add(SOCKET_PERM_CONNECT);
	    actions.add(SOCKET_PERM_ACCEPT);
	    actions.add(SOCKET_PERM_RESOLVE);
	    actions.add(SOCKET_PERM_ALL);
	}
    }

    /*
     * return a Menu list of names for a given permission
     */
    void setPermissionNames(String permission, Choice names, TextField field) {

	names.removeAll();
	names.add(PERM_NAME);
	field.setEditable(true);

	if (permission.equals(ALL_PERM_CLASS)) {
	    field.setEditable(false);
	} else if (permission.equals(AWT_PERM_CLASS)) {
	    names.add(AWT_CLIPBOARD);
	    names.add(AWT_EVENT_QUEUE);
	    names.add(AWT_BANNER);
	} else if (permission.equals(FILE_PERM_CLASS)) {
	    names.add(FILE_ALL_FILES);
	} else if (permission.equals(NET_PERM_CLASS)) {
	    names.add(NET_SET_AUTH);
	    names.add(NET_PASSWD_AUTH);
	} else if (permission.equals(PROP_PERM_CLASS)) {
	} else if (permission.equals(REF_PERM_CLASS)) {
	    names.add(REFLECT_CHECK);
	} else if (permission.equals(RUN_PERM_CLASS)) {
	    names.add(R_CREATE_LOADER);
	    names.add(R_GET_LOADER);
	    names.add(R_EXIT_VM);
	    names.add(R_SET_FACTORY);
	    names.add(R_SET_IO);
	    names.add(R_THREAD);
	    names.add(R_THREAD_GROUP);
	    names.add(R_GET_DOMAIN);
	    names.add(R_SET_DOMAIN);
	    names.add(R_READ_FD);
	    names.add(R_WRITE_FD);
	    names.add(R_LOAD_LIB);
	    names.add(R_ACCESS_CLASS);
	    names.add(R_DEFINE_CLASS);
	    names.add(R_ACCESS_MEMBER);
	    names.add(R_PRINT);
	} else if (permission.equals(SEC_PERM_CLASS)) {
	    names.add(SEC_GET_POLICY);
	    names.add(SEC_SET_POLICY);
	    names.add(SEC_GET_PROP);
	    names.add(SEC_SET_PROP);
	    names.add(SEC_INSERT_PROV);
	    names.add(SEC_REM_PROV);
	    names.add(SEC_SET_SCOPE);
	    names.add(SEC_SET_PUBKEY);
	    names.add(SEC_SET_ID);
	    names.add(SEC_ADD_CERT);
	    names.add(SEC_REMOVE_CERT);
	    names.add(SEC_CLEAR_PROP);
	    names.add(SEC_PUT_PROP);
	    names.add(SEC_REMOVE_PROP);
	    names.add(SEC_GET_PRIVKEY);
	    names.add(SEC_SET_KEYPAIR);
	} else if (permission.equals(SER_PERM_CLASS)) {
	    names.add(SER_SUBCLASS);
	    names.add(SER_SUBSTITUTE);
	} else if (permission.equals(SOCK_PERM_CLASS)) {
	}
    }

    /*
     * Initially set the Permission pulldown menu
     */
    void setPermissions(String permission, Choice choice) {
	if (permission.equals(ALL_PERM_CLASS)) {
	    choice.select(ALL_PERM);
	} else if (permission.equals(AWT_PERM_CLASS)) {
	    choice.select(AWT_PERM);
	} else if (permission.equals(FILE_PERM_CLASS)) {
	    choice.select(FILE_PERM);
	} else if (permission.equals(NET_PERM_CLASS)) {
	    choice.select(NET_PERM);
	} else if (permission.equals(PROP_PERM_CLASS)) {
	    choice.select(PROPERTY_PERM);
	} else if (permission.equals(REF_PERM_CLASS)) {
	    choice.select(REFLECT_PERM);
	} else if (permission.equals(RUN_PERM_CLASS)) {
	    choice.select(RUNTIME_PERM);
	} else if (permission.equals(SEC_PERM_CLASS)) {
	    choice.select(SECURITY_PERM);
	} else if (permission.equals(SER_PERM_CLASS)) {
	    choice.select(SERIAL_PERM);
	} else if (permission.equals(SOCK_PERM_CLASS)) {
	    choice.select(SOCKET_PERM);
	} 
    }

    /*
     * see if there are any single backslashes (check only on Windows)
     */
    String addSingleBackSlash(String filename) {

	String newFile = new String();
	int length = filename.length();
	int index = 0;
	char nextChar[] = new char[1];

	while (index < length) {
	    nextChar[0] = filename.charAt(index);
	    newFile = newFile.concat(new String(nextChar));
	    if (nextChar[0] == '\\') {
		newFile = newFile.concat(new String(nextChar));
		if (filename.charAt(index + 1) == '\\')
		    index++;
	    }
	    index++;
	} 
	return newFile;
    }

    /*
     * parse a permission string (in the Permission Listing)
     * into its permission, its name, and its actions
     *
     * XXX should be a StringTokenizer, huh...
     *     because of this implementation, we have to make a special
     *     case for the java.security.ALLPermission.
     */
    private String parsePermissionString(String perm, int select) {

	perm = perm.substring(perm.indexOf(" ") + 1);

        switch (select) {
        case PERMISSION:

		if (perm.indexOf(ALL_PERM_CLASS) != -1) {
		    return new String(ALL_PERM_CLASS);
		}

                // permission
                if (perm.indexOf("\"") == -1)
                    return perm.substring(0, perm.indexOf(";"));
                return perm.substring(0, perm.indexOf("\"") - 1);
        case PERMISSION_NAME:

		if (perm.indexOf(ALL_PERM_CLASS) != -1) {
		    return null;
		}

                // name
                if (perm.indexOf("\"") == -1)
                    return null;
                else
                    return perm.substring
                        (perm.indexOf("\"") + 1,
                        perm.indexOf("\"", perm.indexOf("\"") + 1));
        case PERMISSION_ACTIONS:

		if (perm.indexOf(ALL_PERM_CLASS) != -1) {
		    return null;
		}

                // actions
                if (perm.indexOf("\"") == -1 || perm.indexOf("\", \"") == -1) {
                    return null;
                } else {
		    String tmpString = perm.substring
                        (perm.indexOf("\", \"") + 4, perm.indexOf(";") - 1);
		    if (tmpString.indexOf("signedBy") != -1)
			return tmpString.substring
			    (0, tmpString.indexOf("signedBy") - 3);
		    else
			return tmpString;
		}
	case PERMISSION_SIGNEDBY:
		// signedBy
		if (perm.indexOf("signedBy") == -1)
		    return null;
		else
		    return perm.substring
			(perm.indexOf("signedBy") + 10, perm.indexOf(";") - 1);
        }
        return null;
    }
}

/**
 * Event handler for the PolicyTool window
 */
class ToolWindowListener implements WindowListener {

    private ToolWindow tw;

    ToolWindowListener(ToolWindow tw) {
	this.tw = tw;
    }

    public void windowOpened(WindowEvent we) {
    }

    public void windowClosing(WindowEvent we) {

	// XXX
	// should we ask user if they want to save changes?
	// (we do if they choose the Menu->Exit)
	// seems that if they kill the application by hand,
	// we don't have to ask.

	tw.setVisible(false);
	tw.dispose();
	System.exit(0);
    }

    public void windowClosed(WindowEvent we) {
	System.exit(0);
    }

    public void windowIconified(WindowEvent we) {
    }

    public void windowDeiconified(WindowEvent we) {
    }

    public void windowActivated(WindowEvent we) {
    }

    public void windowDeactivated(WindowEvent we) {
    }
}

/**
 * Event handler for the Policy List
 */
class PolicyListListener implements ActionListener {

    private PolicyTool tool;
    private ToolWindow tw;

    PolicyListListener(PolicyTool tool, ToolWindow tw) {
	this.tool = tool;
	this.tw = tw;

    }

    public void actionPerformed(ActionEvent e) {

	// display the permission list for a policy entry
	ToolDialog td = new ToolDialog("Policy Entry", tool, tw, true);
	td.displayPolicyEntryDialog(true);
    }
}

/**
 * Event handler for the File Menu
 */
class FileMenuListener implements ActionListener {

    private PolicyTool tool;
    private ToolWindow tw;

    FileMenuListener(PolicyTool tool, ToolWindow tw) {
	this.tool = tool;
	this.tw = tw;
    }

    public void actionPerformed(ActionEvent e) {

	if (e.getActionCommand().equals(tw.QUIT)) {

	    // ask user if they want to save changes
	    ToolDialog td = new ToolDialog("Save Changes", tool, tw, true);
	    td.displayUserSave(td.QUIT);

	    // the above method will perform the QUIT as long as the
	    // user does not CANCEL the request

	} else if (e.getActionCommand().equals(tw.NEW_POLICY_FILE)) {

	    // ask user if they want to save changes
	    ToolDialog td = new ToolDialog("Save Changes", tool, tw, true);
	    td.displayUserSave(td.NEW);
	
	    // the above method will perform the NEW as long as the
	    // user does not CANCEL the request

	} else if (e.getActionCommand().equals(tw.OPEN_POLICY_FILE)) {

	    // ask user if they want to save changes
	    ToolDialog td = new ToolDialog("Save Changes", tool, tw, true);
	    td.displayUserSave(td.OPEN);
	
	    // the above method will perform the OPEN as long as the
	    // user does not CANCEL the request

	} else if (e.getActionCommand().equals(tw.SAVE_POLICY_FILE)) {

	    // get the previously entered filename
	    String filename = ((TextField)
		    tw.getComponent(tw.MW_FILENAME_TEXTFIELD)).getText();

	    try {
		// save the policy entries to a file
		tool.savePolicy(filename);

		// display status
		tw.displayStatusDialog
		    (null, "Policy successfully written to " + filename);
	    } catch (FileNotFoundException fnfe) {
		if (filename == null || filename.equals(""))
		    tw.displayErrorDialog(null, new FileNotFoundException
					("null filename"));
		else
		    tw.displayErrorDialog(null, new FileNotFoundException
					(filename + " not found"));
	    } catch (Exception ee) {
		tw.displayErrorDialog(null, ee);
	    }

	} else if (e.getActionCommand().equals(tw.SAVE_AS_POLICY_FILE)) {

	    // user wants to SAVE AS
	    ToolDialog td = new ToolDialog("Save As", tool, tw, true);
	    td.displaySaveAsDialog(td.NOACTION);
	
	} else if (e.getActionCommand().equals(tw.VIEW_WARNINGS)) {
	    tw.displayWarningLog(null);
	}
    }
}

/**
 * Event handler for the main window buttons and Edit Menu
 */
class MainWindowListener implements ActionListener {

    private PolicyTool tool;
    private ToolWindow tw;

    MainWindowListener(PolicyTool tool, ToolWindow tw) {
	this.tool = tool;
	this.tw = tw;
    }

    public void actionPerformed(ActionEvent e) {

	if (e.getActionCommand().equals(tw.ADD_POLICY_ENTRY)) {

	    // display a dialog box for the user to enter policy info
	    ToolDialog td = new ToolDialog("Policy Entry", tool, tw, true);
	    td.displayPolicyEntryDialog(false);

	} else if (e.getActionCommand().equals(tw.REMOVE_POLICY_ENTRY)) {

	    // get the selected entry
	    List list = (List)tw.getComponent(tw.MW_POLICY_LIST);
	    int index = list.getSelectedIndex();
	    if (index < 0) {
		tw.displayErrorDialog
			(null, new Exception("No Policy Entry selected"));
		return;
	    }

	    // ask the user if they really want to remove the policy entry
	    ToolDialog td =
			new ToolDialog("Remove Policy Entry", tool, tw, true);
	    td.displayConfirmRemovePolicyEntry();

	} else if (e.getActionCommand().equals(tw.EDIT_POLICY_ENTRY)) {

	    // get the selected entry
	    List list = (List)tw.getComponent(tw.MW_POLICY_LIST);
	    int index = list.getSelectedIndex();
	    if (index < 0) {
		tw.displayErrorDialog
			(null, new Exception("No Policy Entry selected"));
		return;
	    }

	    // display the permission list for a policy entry
	    ToolDialog td = new ToolDialog("Policy Entry", tool, tw, true);
	    td.displayPolicyEntryDialog(true);

	} else if (e.getActionCommand().equals(tw.CHANGE_KEYSTORE)) {

	    // display a dialog box for the user to enter keystore info
	    ToolDialog td = new ToolDialog("Keystore", tool, tw, true);
	    td.keyStoreDialog(td.CHANGE_KEYSTORE);

	}
    }
}

/**
 * Event handler for OverWriteFileOKButton button
 */
class OverWriteFileOKButtonListener implements ActionListener {

    private PolicyTool tool;
    private ToolWindow tw;
    private ToolDialog td;
    private String filename;
    private int nextEvent;

    OverWriteFileOKButtonListener(PolicyTool tool, ToolWindow tw,
				ToolDialog td, String filename, int nextEvent) {
	this.tool = tool;
	this.tw = tw;
	this.td = td;
	this.filename = filename;
	this.nextEvent = nextEvent;
    }

    public void actionPerformed(ActionEvent e) {
	try {
	    // save the policy entries to a file
	    tool.savePolicy(filename);

	    // display status
	    tw.displayStatusDialog
			(null, "Policy successfully written to " + filename);

	    // display the new policy filename
	    TextField newFilename = (TextField)tw.getComponent
				(tw.MW_FILENAME_TEXTFIELD);
	    newFilename.setText(filename);
	    tw.setVisible(true);

	    // now continue with the originally requested command
	    // (QUIT, NEW, or OPEN)
	    td.setVisible(false);
	    td.dispose();
	    td.userSaveContinue(tool, tw, td, nextEvent);

	} catch (FileNotFoundException fnfe) {
	    if (filename == null || filename.equals(""))
		tw.displayErrorDialog(null, new FileNotFoundException
					("null filename"));
	    else
		tw.displayErrorDialog(null, new FileNotFoundException
					(filename + " not found"));
	    td.setVisible(false);
	    td.dispose();
	} catch (Exception ee) {
	    tw.displayErrorDialog(null, ee);
	    td.setVisible(false);
	    td.dispose();
	}
    }
}

/**
 * Event handler for AddEntryDoneButton button
 *
 * -- if edit is TRUE, then we are EDITing an existing PolicyEntry
 *    and we need to update both the policy and the GUI listing.
 *    if edit is FALSE, then we are ADDing a new PolicyEntry,
 *    so we only need to update the GUI listing.
 */
class AddEntryDoneButtonListener implements ActionListener {

    private PolicyTool tool;
    private ToolWindow tw;
    private ToolDialog td;
    private boolean edit;

    AddEntryDoneButtonListener(PolicyTool tool, ToolWindow tw,
				ToolDialog td, boolean edit) {
	this.tool = tool;
	this.tw = tw;
	this.td = td;
	this.edit = edit;
    }

    public void actionPerformed(ActionEvent e) {

	try {
	    // get a PolicyEntry object from the dialog policy info
	    PolicyEntry newEntry = td.getPolicyEntryFromDialog();
	    PolicyParser.GrantEntry newGe = newEntry.getGrantEntry();

	    // see if all the signers have public keys
	    if (newGe.signedBy != null) {
		String signers[] = tool.parseSigners(newGe.signedBy);
		for (int i = 0; i < signers.length; i++) {
		    PublicKey pubKey = tool.getPublicKeyAlias(signers[i]);
		    if (pubKey == null) {
			tool.warnings.addElement
			    ("Warning: A public key for alias '" + signers[i] +
			    "' does not exist.");
			tw.displayStatusDialog(td,
			    "Warning: A public key for alias '" + signers[i] +
			    "' does not exist.");
		    }
		}
	    }

	    // add the entry
	    List policyList = (List)tw.getComponent(tw.MW_POLICY_LIST);
	    if (edit) {
		int listIndex = policyList.getSelectedIndex();
		tool.addEntry(newEntry, listIndex);
		policyList.replaceItem(newEntry.codebaseToString(), listIndex);
	    } else {
		tool.addEntry(newEntry, -1);
		policyList.add(newEntry.codebaseToString());
	    }
	    td.setVisible(false);
	    td.dispose();

	} catch (Exception eee) {
	    tw.displayErrorDialog(td, eee);
	}
    }
}

/**
 * Event handler for ChangeKeyStoreOKButton button
 */
class ChangeKeyStoreOKButtonListener implements ActionListener {

    private PolicyTool tool;
    private ToolWindow tw;
    private ToolDialog td;

    ChangeKeyStoreOKButtonListener(PolicyTool tool, ToolWindow tw,
					ToolDialog td) {
	this.tool = tool;
	this.tw = tw;
	this.td = td;
    }

    public void actionPerformed(ActionEvent e) {

	String URLString = ((TextField)
		td.getComponent(td.KSD_NAME_TEXTFIELD)).getText().trim();
	String type = ((TextField)
		td.getComponent(td.KSD_TYPE_TEXTFIELD)).getText().trim();

	if ((URLString == null || URLString.equals("")) &&
	    type != null && type.equals("") == false) {
		tw.displayErrorDialog(td, new Exception
		    ("KeyStore URL must have a valid value"));
		return;
	}
	tool.setKeyStoreInfo(URLString, type);
	TextField ks = (TextField)tw.getComponent(tw.MW_KEYSTORE_TEXTFIELD);
	if (type != null && type.length() > 0)
	    ks.setText(URLString + ", " + type);
	else
	    ks.setText(URLString);
	td.dispose();
    }
}

/**
 * Event handler for AddPermButton button
 */
class AddPermButtonListener implements ActionListener {

    private PolicyTool tool;
    private ToolWindow tw;
    private ToolDialog td;
    private boolean editPolicyEntry;

    AddPermButtonListener(PolicyTool tool, ToolWindow tw,
				ToolDialog td, boolean editPolicyEntry) {
	this.tool = tool;
	this.tw = tw;
	this.td = td;
	this.editPolicyEntry = editPolicyEntry;
    }

    public void actionPerformed(ActionEvent e) {

	// display a dialog box for the user to enter permission info
	td.displayPermissionDialog(editPolicyEntry, false);
    }
}

/**
 * Event handler for AddPermOKButton button
 */
class NewPolicyEntryAddPermOKButtonListener implements ActionListener {

    private PolicyTool tool;
    private ToolWindow tw;
    private ToolDialog listDialog;
    private ToolDialog infoDialog;
    private boolean edit;

    NewPolicyEntryAddPermOKButtonListener(PolicyTool tool,
				ToolWindow tw,
				ToolDialog listDialog,
				ToolDialog infoDialog,
				boolean edit) {
	this.tool = tool;
	this.tw = tw;
	this.listDialog = listDialog;
	this.infoDialog = infoDialog;
	this.edit = edit;
    }

    public void actionPerformed(ActionEvent e) {

	try {
	    // read in the new permission info from Dialog Box
	    PolicyParser.PermissionEntry pppe =
			infoDialog.getPermFromDialog();

	    /*
	     * XXX	hate to add a special case check here ... but ...
	     *		we need to make sure windows users use
	     *		2 backs slashes in their filenames
	     */
	    if (pppe.permission.equals(listDialog.FILE_PERM_CLASS) &&
		File.separatorChar == '\\')
		pppe.name = listDialog.addSingleBackSlash(pppe.name);

	    try {
		tool.verifyPermission(pppe.permission, pppe.name, pppe.action);
	    } catch (ClassNotFoundException cnfe) {
		tool.warnings.addElement
			("Warning: Class not found: " + pppe.permission + ".");
		tw.displayStatusDialog(infoDialog,
			"Warning: Class not found: " + pppe.permission + ".");
	    }

	    // add the permission to the GUI permission list
	    StringWriter sw = new StringWriter();
	    PrintWriter pw = new PrintWriter(sw);
	    pppe.write(pw);
	    List permList =
		(List)listDialog.getComponent(listDialog.PE_PERM_LIST);

	    String permString = sw.toString();
	    if (edit) {
		// if editing, replace the original permission
		permList.replaceItem
			(permString.substring(0, permString.indexOf(";") + 1),
			permList.getSelectedIndex());
	    } else {
		// if adding, just add it to the end
		permList.add
			(permString.substring(0, permString.indexOf(";") + 1));
	    }
	    infoDialog.dispose();

	} catch (InvocationTargetException ite) {
	    tw.displayErrorDialog(infoDialog,
				new Exception("Invalid value for Actions"));
	} catch (Exception ee) {
	    tw.displayErrorDialog(infoDialog, ee);
	}
    }
}

/**
 * Event handler for EditPermOKButton button
 */
class EditPolicyEditPermOKButtonListener implements ActionListener {

    private PolicyTool tool;
    private ToolWindow tw;
    private ToolDialog listDialog;
    private ToolDialog infoDialog;
    private boolean edit;

    EditPolicyEditPermOKButtonListener(PolicyTool tool,
				ToolWindow tw,
				ToolDialog listDialog,
				ToolDialog infoDialog,
				boolean edit) {
	this.tool = tool;
	this.tw = tw;
	this.listDialog = listDialog;
	this.infoDialog = infoDialog;
	this.edit = edit;
    }

    public void actionPerformed(ActionEvent e) {
	try {
	    // read in the new permission info from Dialog Box
	    PolicyParser.PermissionEntry pppe =
			infoDialog.getPermFromDialog();
 
	    /*
	     * XXX      hate to add a special case check here ... but ...
	     *		we need to make sure windows users use
	     *		2 backs slashes in their filenames
	     */
	    if (pppe.permission.equals(listDialog.FILE_PERM_CLASS) &&
		File.separatorChar == '\\')
		pppe.name = listDialog.addSingleBackSlash(pppe.name);
 
	    try {
		tool.verifyPermission(pppe.permission, pppe.name, pppe.action);
	    } catch (ClassNotFoundException cnfe) {
		tool.warnings.addElement
			("Warning: Class not found: " + pppe.permission + ".");
		tw.displayStatusDialog(infoDialog,
			"Warning: Class not found: " + pppe.permission + ".");
	    }

	    // get the PolicyEntry selected from the main ToolWindow
	    List policyList = (List)tw.getComponent(tw.MW_POLICY_LIST);
	    int index = policyList.getSelectedIndex();
	    PolicyEntry entries[] = tool.getEntry();
	    PolicyEntry entry = entries[index];
 
	    // add the permission to the policy entry
	    List permList =
			(List)listDialog.getComponent(listDialog.PE_PERM_LIST);
	    boolean status = false;
	    if (edit)
		status = tool.addPermEntry
			(entry, pppe, permList.getSelectedIndex());
	    else
		status = tool.addPermEntry
			(entry, pppe, -1);
	    if (status == true) {

		// add the permission to the GUI permission list
		StringWriter sw = new StringWriter();
		PrintWriter pw = new PrintWriter(sw);
		pppe.write(pw);
		String permString = sw.toString();
		if (edit) {
		    permList.replaceItem
			(permString.substring(0, permString.indexOf(";") + 1),
			permList.getSelectedIndex());
		} else {
		    permList.add
			(permString.substring(0, permString.indexOf(";") + 1));
		}
	    }
	    infoDialog.dispose();

	} catch (InvocationTargetException ite) {
	    tw.displayErrorDialog(infoDialog,
				new Exception("Invalid value for Actions"));
	} catch (Exception ee) {
	    tw.displayErrorDialog(infoDialog, ee);
	}
    }
}

/**
 * Event handler for RemovePermButton button
 */
class RemovePermButtonListener implements ActionListener {

    private PolicyTool tool;
    private ToolWindow tw;
    private ToolDialog td;
    private boolean edit;

    RemovePermButtonListener(PolicyTool tool, ToolWindow tw,
				ToolDialog td, boolean edit) {
	this.tool = tool;
	this.tw = tw;
	this.td = td;
	this.edit = edit;
    }

    public void actionPerformed(ActionEvent e) {

	// get the Permission selected from the Permission List
	List permList = (List)td.getComponent(td.PE_PERM_LIST);
	int permIndex = permList.getSelectedIndex();

	if (permIndex < 0) {
	    tw.displayErrorDialog(td, new Exception("No permission selected"));
	    return;
	}

	// if editing an existing PolicyEntry, remove the original PolicyEntry.
	// otherwise, if adding a new PolicyEntry, simply update the GUI list.
	if (edit) {
	    // get the PolicyEntry selected from the main ToolWindow
	    List policyList = (List)tw.getComponent(tw.MW_POLICY_LIST);
	    int policyIndex = policyList.getSelectedIndex();

	    PolicyEntry entries[] = tool.getEntry();
	    Vector perms =
		entries[policyIndex].getGrantEntry().permissionEntries;
	    PolicyParser.PermissionEntry goner =
		(PolicyParser.PermissionEntry)perms.elementAt(permIndex);

	    // remove the Permission from the PolicyEntry
	    tool.removePermEntry(entries[policyIndex], goner);
	}
				
	// remove the permission from the display
	permList.remove(permIndex);

    }
}

/**
 * Event handler for Edit Permission button
 *
 * We need the editPolicyEntry boolean to tell us if the user is
 * adding a new PolicyEntry at this time, or editing an existing entry.
 * If the user is adding a new PolicyEntry, we ONLY update the
 * GUI listing.  If the user is editing an existing PolicyEntry, we
 * update both the GUI listing and the actual PolicyEntry.
 */
class EditPermButtonListener implements ActionListener {

    private PolicyTool tool;
    private ToolWindow tw;
    private ToolDialog td;
    private boolean editPolicyEntry;

    EditPermButtonListener(PolicyTool tool, ToolWindow tw,
				ToolDialog td, boolean editPolicyEntry) {
	this.tool = tool;
	this.tw = tw;
	this.td = td;
	this.editPolicyEntry = editPolicyEntry;
    }

    public void actionPerformed(ActionEvent e) {

	// get the Permission selected from the Permission List
	List list = (List)td.getComponent(td.PE_PERM_LIST);
	int permIndex = list.getSelectedIndex();

	if (permIndex < 0) {
	    tw.displayErrorDialog(td, new Exception("No permission selected"));
	    return;
	}
	td.displayPermissionDialog(editPolicyEntry, true);
    }
}

/**
 * Event handler for Permission Popup Menu
 */
class PermissionMenuListener implements ItemListener {

    private ToolDialog td;

    PermissionMenuListener(ToolDialog td) {
	this.td = td;
    }

    public void itemStateChanged(ItemEvent e) {

	Choice names = (Choice)td.getComponent(td.PD_NAME_CHOICE);
	Choice actions = (Choice)td.getComponent(td.PD_ACTIONS_CHOICE);
	TextField nameField =
			(TextField)td.getComponent(td.PD_NAME_TEXTFIELD);
	TextField actionsField =
			(TextField)td.getComponent(td.PD_ACTIONS_TEXTFIELD);
	TextField permField = (TextField)td.getComponent(td.PD_PERM_TEXTFIELD);
	TextField signedbyField =
			(TextField)td.getComponent(td.PD_SIGNEDBY_TEXTFIELD);

	// ignore if they choose the 'Permission:' item
	if (((String)e.getItem()).equals(td.PERM)) {
	    if (permField.getText() != null &&
		permField.getText().length() > 0) {

		Choice perms = (Choice)td.getComponent(td.PD_PERM_CHOICE);
		td.setPermissions(permField.getText(), perms);
	    }
	    return;
	}

	// if you change the permission, clear the name, actions, and signedBy
	if (permField.getText().indexOf((String)e.getItem()) == -1) {
	    nameField.setText("");
	    actionsField.setText("");
	    signedbyField.setText("");
	}

	// set the text in the textfield and also modify the
	// pull-down choice menus to reflect the correct possible
	// set of names and actions

	if (((String)e.getItem()).equals(td.AWT_PERM)) {
	    permField.setText(td.AWT_PERM_CLASS);
	    td.setPermissionNames(td.AWT_PERM_CLASS, names, nameField);
	    td.setPermissionActions(td.AWT_PERM_CLASS, actions, actionsField);
	} else if (((String)e.getItem()).equals(td.FILE_PERM)) {
	    permField.setText(td.FILE_PERM_CLASS);
	    td.setPermissionNames(td.FILE_PERM_CLASS, names, nameField);
	    td.setPermissionActions(td.FILE_PERM_CLASS, actions, actionsField);
	} else if (((String)e.getItem()).equals(td.NET_PERM)) {
	    permField.setText(td.NET_PERM_CLASS);
	    td.setPermissionNames(td.NET_PERM_CLASS, names, nameField);
	    td.setPermissionActions(td.NET_PERM_CLASS, actions, actionsField);
	} else if (((String)e.getItem()).equals(td.PROPERTY_PERM)) {
	    permField.setText(td.PROP_PERM_CLASS);
	    td.setPermissionNames(td.PROP_PERM_CLASS, names, nameField);
	    td.setPermissionActions(td.PROP_PERM_CLASS, actions, actionsField);
	} else if (((String)e.getItem()).equals(td.RUNTIME_PERM)) {
	    permField.setText(td.RUN_PERM_CLASS);
	    td.setPermissionNames(td.RUN_PERM_CLASS, names, nameField);
	    td.setPermissionActions(td.RUN_PERM_CLASS, actions, actionsField);
	} else if (((String)e.getItem()).equals(td.SOCKET_PERM)) {
	    permField.setText(td.SOCK_PERM_CLASS);
	    td.setPermissionNames(td.SOCK_PERM_CLASS, names, nameField);
	    td.setPermissionActions(td.SOCK_PERM_CLASS, actions, actionsField);
	} else if (((String)e.getItem()).equals(td.REFLECT_PERM)) {
	    permField.setText(td.REF_PERM_CLASS);
	    td.setPermissionNames(td.REF_PERM_CLASS, names, nameField);
	    td.setPermissionActions(td.REF_PERM_CLASS, actions, actionsField);
	} else if (((String)e.getItem()).equals(td.SECURITY_PERM)) {
	    permField.setText(td.SEC_PERM_CLASS);
	    td.setPermissionNames(td.SEC_PERM_CLASS, names, nameField);
	    td.setPermissionActions(td.SEC_PERM_CLASS, actions, actionsField);
	} else if (((String)e.getItem()).equals(td.SERIAL_PERM)) {
	    permField.setText(td.SER_PERM_CLASS);
	    td.setPermissionNames(td.SER_PERM_CLASS, names, nameField);
	    td.setPermissionActions(td.SER_PERM_CLASS, actions, actionsField);
	} else if (((String)e.getItem()).equals(td.ALL_PERM)) {
	    permField.setText(td.ALL_PERM_CLASS);
	    td.setPermissionNames(td.ALL_PERM_CLASS, names, nameField);
	    td.setPermissionActions(td.ALL_PERM_CLASS, actions, actionsField);
	}
    }
}

/**
 * Event handler for Permission Name Popup Menu
 */
class PermissionNameMenuListener implements ItemListener {

    private ToolDialog td;

    PermissionNameMenuListener(ToolDialog td) {
	this.td = td;
    }

    public void itemStateChanged(ItemEvent e) {

	if (((String)e.getItem()).indexOf(td.PERM_NAME) != -1)
	    return;

	TextField tf = (TextField)td.getComponent(td.PD_NAME_TEXTFIELD);
	tf.setText((String)e.getItem());
    }
}

/**
 * Event handler for Permission Actions Popup Menu
 */
class PermissionActionsMenuListener implements ItemListener {

    private ToolDialog td;

    PermissionActionsMenuListener(ToolDialog td) {
	this.td = td;
    }

    public void itemStateChanged(ItemEvent e) {

	if (((String)e.getItem()).indexOf(td.PERM_ACTIONS) != -1)
	    return;

	TextField tf = (TextField)td.getComponent(td.PD_ACTIONS_TEXTFIELD);
	if (tf.getText() == null || tf.getText().equals("")) {
	    tf.setText((String)e.getItem());
	} else {
	    if (tf.getText().indexOf((String)e.getItem()) == -1)
		tf.setText(tf.getText() + ", " + (String)e.getItem());
	}
    }
}

/**
 * Event handler for CancelButton button
 */
class CancelButtonListener implements ActionListener {

    private ToolDialog td;

    CancelButtonListener(ToolDialog td) {
	this.td = td;
    }

    public void actionPerformed(ActionEvent e) {
	td.setVisible(false);
	td.dispose();
    }
}

/**
 * Event handler for ErrorOKButton button
 */
class ErrorOKButtonListener implements ActionListener {

    private ToolDialog ed;

    ErrorOKButtonListener(ToolDialog ed) {
	this.ed = ed;
    }

    public void actionPerformed(ActionEvent e) {
	ed.setVisible(false);
	ed.dispose();
    }
}

/**
 * Event handler for StatusOKButton button
 */
class StatusOKButtonListener implements ActionListener {

    private ToolDialog sd;

    StatusOKButtonListener(ToolDialog sd) {
	this.sd = sd;
    }

    public void actionPerformed(ActionEvent e) {
	sd.setVisible(false);
	sd.dispose();
    }
}

/**
 * Event handler for UserSaveYes button
 */
class UserSaveYesButtonListener implements ActionListener {

    private ToolDialog us;
    private PolicyTool tool;
    private ToolWindow tw;
    private int select;

    UserSaveYesButtonListener(ToolDialog us, PolicyTool tool,
			ToolWindow tw, int select) {
	this.us = us;
	this.tool = tool;
	this.tw = tw;
	this.select = select;
    }

    public void actionPerformed(ActionEvent e) {

	// first get rid of the window
	us.setVisible(false);
	us.dispose();

	try {
	    String filename = ((TextField)
		    tw.getComponent(tw.MW_FILENAME_TEXTFIELD)).getText();
	    if (filename == null || filename.equals("")) {
		us.displaySaveAsDialog(select);

		// the above dialog will continue with the originally
		// requested command if necessary
	    } else {
		// save the policy entries to a file
		tool.savePolicy(filename);

		// display status
		tw.displayStatusDialog
			(null, "Policy successfully written to " + filename);

		// now continue with the originally requested command
		// (QUIT, NEW, or OPEN)
		us.userSaveContinue(tool, tw, us, select);
	    }
	} catch (Exception ee) {
	    // error -- just report it and bail
	    tw.displayErrorDialog(null, ee);
	}
    }
}

/**
 * Event handler for UserSaveNoButton
 */
class UserSaveNoButtonListener implements ActionListener {

    private PolicyTool tool;
    private ToolWindow tw;
    private ToolDialog us;
    private int select;

    UserSaveNoButtonListener(ToolDialog us, PolicyTool tool,
			ToolWindow tw, int select) {
	this.us = us;
	this.tool = tool;
	this.tw = tw;
	this.select = select;
    }

    public void actionPerformed(ActionEvent e) {
	us.setVisible(false);
	us.dispose();

	// now continue with the originally requested command
	// (QUIT, NEW, or OPEN)
	us.userSaveContinue(tool, tw, us, select);
    }
}

/**
 * Event handler for UserSaveCancelButton
 */
class UserSaveCancelButtonListener implements ActionListener {

    private ToolDialog us;

    UserSaveCancelButtonListener(ToolDialog us) {
	this.us = us;
    }

    public void actionPerformed(ActionEvent e) {
	us.setVisible(false);
	us.dispose();

	// do NOT continue with the originally requested command
	// (QUIT, NEW, or OPEN)
    }
}

/**
 * Event handler for ConfirmRemovePolicyEntryOKButtonListener
 */
class ConfirmRemovePolicyEntryOKButtonListener implements ActionListener {

    private PolicyTool tool;
    private ToolWindow tw;
    private ToolDialog us;

    ConfirmRemovePolicyEntryOKButtonListener(PolicyTool tool,
				ToolWindow tw, ToolDialog us) {
	this.tool = tool;
	this.tw = tw;
	this.us = us;
    }

    public void actionPerformed(ActionEvent e) {
	// remove the entry
	List list = (List)tw.getComponent(tw.MW_POLICY_LIST);
	int index = list.getSelectedIndex();
	PolicyEntry entries[] = tool.getEntry();
	tool.removeEntry(entries[index]);

	// redraw the window listing
	list = new List(40, false);
	list.addActionListener(new PolicyListListener(tool, tw));
	entries = tool.getEntry();
	if (entries != null) {
		for (int i = 0; i < entries.length; i++)
		    list.add(entries[i].codebaseToString());
	}
	tw.replacePolicyList(list);
	us.setVisible(false);
	us.dispose();
    }
}
