/*
 * @(#)AVA.java	1.18 98/07/10
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.security.x509;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import sun.security.util.*;
import sun.security.pkcs.PKCS9Attribute;

/**
 * X.500 Attribute-Value-Assertion (AVA):  an attribute, as identified by
 * some attribute ID, has some particular value.  Values are as a rule ASN.1
 * printable strings.  A conventional set of type IDs is recognized when
 * parsing (and generating) RFC 1779 syntax strings.
 *
 * <P>AVAs are components of X.500 relative names.  Think of them as being
 * individual fields of a database record.  The attribute ID is how you
 * identify the field, and the value is part of a particular record.
 *
 * @see X500Name
 * @see RDN
 *
 * @version 1.18
 *
 * @author David Brownell
 * @author Amit Kapoor
 * @author Hemma Prafullchandra
 */
// public ... when RDN is public and X.500Names can be
// constructed using RDNs, and all three classes are cleaner
final class AVA implements DerEncoder {
    ObjectIdentifier	oid;
    DerValue		value;

    /*
     * If the value has any of these characters in it, it must be quoted.
     * Backslash and quote characters must also be individually escaped.
     * Leading and trailing spaces, also multiple internal spaces, also
     * call for quoting the whole string.
     */
    private static final String specialChars = ",+=\n<>#;";

    /*
     * Values that aren't printable strings are emitted as BER-encoded
     * hex data.
     */
    private static final String digits = "0123456789ABCDEF";

    AVA(ObjectIdentifier type, DerValue val) {
	oid = type;
	value = val;
    }

    /*
     * Parse an RFC 1779 style AVA string:  CN=fee fie foe fum
     * or perhaps with quotes.  Not all defined AVA tags are supported;
     * of current note are X.400 related ones (PRMD, ADMD, etc).
     *
     * This terminates at unescaped AVA separators ("+") or RDN
     * separators (",", ";"), or DN terminators (">"), and removes
     * cosmetic whitespace at the end of values.
     */
    AVA(InputStream in) throws IOException {
	StringBuffer	temp = new StringBuffer();
	char		c;

	/*
	 * First get the keyword indicating the attribute's type,
	 * and map it to the appropriate OID.
	 */
	for (;;) {
	    c = (char) in.read();
	    if (c == '=')
		break;
            if (c == -1)
                throw new IOException("Incorrect AVA format");
	    temp.append(c);
	}

	String	keyword = temp.toString().toUpperCase().trim();

	if (keyword.equals("CN"))
	    oid = X500Name.commonName_oid;
	else if (keyword.equals("C"))
	    oid = X500Name.countryName_oid;
	else if (keyword.equals("L"))
	    oid = X500Name.localityName_oid;
	else if (keyword.equals("S") || keyword.equals("ST"))
	    oid = X500Name.stateName_oid;
	else if (keyword.equals("O"))
	    oid = X500Name.orgName_oid;

	else if (keyword.equals("OU"))
	    oid = X500Name.orgUnitName_oid;
	else if (keyword.equals("T"))
	    oid = X500Name.title_oid;
	else if (keyword.equals("IP"))
	    oid = X500Name.ipAddress_oid;
	else if (keyword.equals("STREET"))
	    oid = X500Name.streetAddress_oid;

	// else if (keyword.startsWith("OID.") { ... }

	else
	    throw new IOException("unsupported keyword " + keyword);

	/*
	 * Now parse the value.  "#hex", a quoted string, or a string
	 * terminated by "+", ",", ";", ">".  Whitespace before or after
	 * the value is stripped away.
	 */
	temp = new StringBuffer();
	for (c = (char) in.read();
		c == ' ' || c == '\n';
		c = (char) in.read())
	    continue;
        if (c == -1)
               throw new IOException("Incorrect AVA format");

	if (c == '#') {
	    throw new IOException("AVA parse, hex values NYI");

	} else if (c == '"') {
	    throw new IOException("AVA parse, quoted strings NYI");

	} else {

	    int r;
	    do {
		if (c == '\\')
		    c = (char) in.read();
		temp.append(c);
		r = in.read();
		c = (char)r;
	    } while (r != -1 && c != '+' && c != ',' && c != ';' && c != '>');

	    value = new DerValue(temp.toString().trim());
	}
    }

    AVA(DerInputStream in) throws IOException {
	DerValue	assertion = in.getDerValue();

	/*
	 * Individual attribute value assertions are SEQUENCE of two values.
	 * That'd be a "struct" outside of ASN.1.
	 */
	if (assertion.tag != DerValue.tag_Sequence)
	    throw new IOException("X500 AVA, not a sequence");

	oid = X500Name.intern(assertion.data.getOID());
	value = assertion.data.getDerValue();

	if (assertion.data.available() != 0)
	    throw new IOException("AVA, extra bytes = "
		+ assertion.data.available());
    }

    boolean equals(AVA other) {
	return oid.equals(other.oid) && value.equals(other.value);
    }

    public boolean equals(Object other) {
	if (other instanceof AVA)
	    return equals((AVA)other);
	else
	    return false;
    }

    /*
     * AVAs are encoded as a SEQUENCE of two elements.
     */
    public void encode(DerOutputStream out) throws IOException {
	derEncode(out);
    }

    /**
     * DER encode this object onto an output stream.
     * Implements the <code>DerEncoder</code> interface.
     *
     * @param out
     * the output stream on which to write the DER encoding.
     *
     * @exception IOException on encoding error.
     */
    public void derEncode(OutputStream out) throws IOException {
	DerOutputStream		tmp = new DerOutputStream();
	DerOutputStream		tmp2 = new DerOutputStream();

	tmp.putOID(oid);
	value.encode(tmp);
	tmp2.write(DerValue.tag_Sequence, tmp);
	out.write(tmp2.toByteArray());
    }

    /**
     * Returns a printable form of this attribute, using RFC 1779
     * syntax for individual attribute/value assertions.
     */
    public String toString() {
	String		keyword;

	if (oid.equals(X500Name.commonName_oid))
	    keyword = "CN";
	else if (oid.equals(X500Name.countryName_oid))
	    keyword = "C";
	else if (oid.equals(X500Name.localityName_oid))
	    keyword = "L";
	else if (oid.equals(X500Name.stateName_oid))
	    keyword = "ST";
	else if (oid.equals(X500Name.orgName_oid))
	    keyword = "O";
	else if (oid.equals(X500Name.orgUnitName_oid))
	     keyword = "OU";
	else if (oid.equals(X500Name.title_oid))
	    keyword = "T";
	else if (oid.equals(X500Name.ipAddress_oid))
	    keyword = "IP";
	else if (oid.equals(X500Name.streetAddress_oid))
	    keyword = "STREET";
	else if (oid.equals(PKCS9Attribute.EMAIL_ADDRESS_OID))
	    keyword = PKCS9Attribute.EMAIL_ADDRESS_STR;
	else
	    keyword = "OID." + oid.toString();

	/*
	 * Construct the value with as little copying and garbage
	 * production as practical.  First the keyword (mandatory),
	 * then the equals sign, finally the value.
	 */
	StringBuffer	retval = new StringBuffer(30);
	int		i;
	String		temp;

	retval.append(keyword);
	retval.append("=");

	try {
	    temp = value.getAsString();
	    if (temp == null) {
		byte	data [] = value.toByteArray();

		retval.append('#');
		for (i = 0; i < data.length; i++) {
		    retval.append(digits.charAt((data [i] >> 4) & 0x0f));
		    retval.append(digits.charAt(data [i] & 0x0f));
		}

	    } else {
		int		len;
		boolean	quoteNeeded = false;

		/*
		 * Special characters (e.g. AVA list separators) cause strings
		 * to need quoting, or at least escaping.  So do leading or
		 * trailing spaces, and multiple internal spaces.
		 */
		len = temp.length();

		for (i = 0; i < specialChars.length(); i++)
		    if (temp.indexOf(specialChars.charAt(i)) != -1) {
			quoteNeeded = true;
			break;
		    }

		if (len > 0 && !quoteNeeded) {
		    if (temp.charAt(0) == ' '
			    || temp.charAt(0) == '\n'
			    || temp.charAt(len - 1) == ' '
			    || temp.charAt(len - 1) == '\n')
			quoteNeeded = true;
		}
		if (len > 0 && !quoteNeeded) {
		    boolean		lastWasSpace = false;

		    for (i = 1; i < len - 2; i++) {
			if (temp.charAt(i) == ' ' || temp.charAt(i) == '\n') {
			    if (lastWasSpace) {
				quoteNeeded = true;
				break;
			    } else
				lastWasSpace = true;
			} else
			    lastWasSpace = false;
		    }
		}

		/*
		 * Emit the string ... quote it if needed, and be wary
		 * of the need to escape quotes and backslashes.
		 */
		if (quoteNeeded)
		    retval.append('"');

		for (i = 0; i < len; i++) {
		    char	c = temp.charAt(i);

		    if (c == '"' || c == '\\')
			retval.append('\\');
		    retval.append(c);
		}

		if (quoteNeeded)
		    retval.append('"');
	    }
	} catch (IOException e) {
	    throw new IllegalArgumentException("DER Value conversion");
	}

	return retval.toString();
    }
}
