/*
 * @(#)GeneralName.java	1.10 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.security.x509;

import java.io.IOException;

import sun.security.util.*;

/**
 * This class implements the ASN.1 GeneralName object class.
 * <p>
 * The ASN.1 syntax for this is:
 * <pre>
 * GeneralName ::= CHOICE {
 *    otherName                       [0]     OtherName,
 *    rfc822Name                      [1]     IA5String,
 *    dNSName                         [2]     IA5String,
 *    x400Address                     [3]     ORAddress,
 *    directoryName                   [4]     Name,
 *    ediPartyName                    [5]     EDIPartyName,
 *    uniformResourceIdentifier       [6]     IA5String,
 *    iPAddress                       [7]     OCTET STRING,
 *    registeredID                    [8]     OBJECT IDENTIFIER
 * }
 * </pre>
 * @author Amit Kapoor
 * @author Hemma Prafullchandra
 * @version 1.10
 */
public class GeneralName {

    // Private data members
    private GeneralNameInterface name = null;

    /**
     * Default constructor for the class.
     *
     * @param name the selected CHOICE from the list.
     */
    public GeneralName(GeneralNameInterface name) {
        this.name = name;
    }

    /**
     * Create the object from its DER encoded value.
     *
     * @param encName the DER encoded GeneralName.
     */
    public GeneralName(DerValue encName) throws IOException {
        short tag = (byte)(encName.tag & 0x1f);

        // NB. this is always encoded with the IMPLICIT tag
        // The checks only make sense if we assume implicit tagging,
        // with explicit tagging the form is always constructed.
        switch (tag) {
        case GeneralNameInterface.NAME_ANY:
            if (encName.isContextSpecific() && encName.isConstructed()) {
                encName.resetTag(DerValue.tag_Sequence);
                name = new OtherName(encName);
            } else
	        throw new IOException("Invalid encoding of Other-Name");
	    break;

        case GeneralNameInterface.NAME_RFC822:
            if (encName.isContextSpecific() && !encName.isConstructed()) {
                encName.resetTag(DerValue.tag_IA5String);
	        name = new RFC822Name(encName);
            } else
	        throw new IOException("Invalid encoding of RFC822 name");
	    break;

        case GeneralNameInterface.NAME_DNS:
            if (encName.isContextSpecific() && !encName.isConstructed()) {
                encName.resetTag(DerValue.tag_IA5String);
	        name = new DNSName(encName);
            } else
	        throw new IOException("Invalid encoding of DNS name");
	    break;

        case GeneralNameInterface.NAME_URI:
            if (encName.isContextSpecific() && !encName.isConstructed()) {
                encName.resetTag(DerValue.tag_IA5String);
	        name = new URIName(encName);
            } else
	        throw new IOException("Invalid encoding of URI");
	    break;

        case GeneralNameInterface.NAME_IP:
            if (encName.isContextSpecific() && !encName.isConstructed()) {
                encName.resetTag(DerValue.tag_OctetString);
	        name = new IPAddressName(encName);
            } else
	        throw new IOException("Invalid encoding of IP address");
	    break;

        case GeneralNameInterface.NAME_OID:
            if (encName.isContextSpecific() && !encName.isConstructed()) {
                encName.resetTag(DerValue.tag_ObjectId);
	        name = new OIDName(encName);
            } else
	        throw new IOException("Invalid encoding of OID name");
	    break;

        case GeneralNameInterface.NAME_DIRECTORY:
            if (encName.isContextSpecific() && encName.isConstructed()) {
                encName.resetTag(DerValue.tag_Sequence);
                name = new X500Name(encName);
            } else
	        throw new IOException("Invalid encoding of Directory name");
	    break;

        case GeneralNameInterface.NAME_EDI:
            if (encName.isContextSpecific() && encName.isConstructed()) {
                encName.resetTag(DerValue.tag_Sequence);
                name = new EDIPartyName(encName);
            } else
	        throw new IOException("Invalid encoding of EDI name");
	    break;

        default:
	    throw new IOException("Unrecognized GeneralName tag, ("
                                  + tag +")");
        }
    }

    /**
     * Return the type of the general name.
     */
    public int getType() {
        return (name.getType());
    }

    /**
     * Return the name as user readable string
     */
    public String toString() {
        return (name.toString());
    }

   /**
    * Encode the name to the specified DerOutputStream.
    *
    * @param out the DerOutputStream to encode the the GeneralName to.
    * @exception IOException on encoding errors.
    */
   public void encode(DerOutputStream out) throws IOException {
       DerOutputStream tmp = new DerOutputStream();
       name.encode(tmp);
       int nameType = name.getType();
       // constructed form
       if (nameType == GeneralNameInterface.NAME_ANY ||
           nameType == GeneralNameInterface.NAME_X400 ||
           nameType == GeneralNameInterface.NAME_DIRECTORY ||
           nameType == GeneralNameInterface.NAME_EDI) {

           out.writeImplicit(DerValue.createTag(DerValue.TAG_CONTEXT,
                             true, (byte)nameType), tmp);
       } else // primitive form
           out.writeImplicit(DerValue.createTag(DerValue.TAG_CONTEXT,
                             false, (byte)nameType), tmp);
   }
}
