/*
 * @(#)GeneralNames.java	1.9 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.security.x509;

import java.io.IOException;
import java.util.Enumeration;
import java.util.Vector;

import sun.security.util.*;

/**
 * This object class represents the GeneralNames type required in
 * X509 certificates.  
 * <p>The ASN.1 syntax for this is:
 * <pre>
 * GeneralNames ::= SEQUENCE SIZE (1..MAX) OF GeneralName
 * </pre>
 * @author Amit Kapoor
 * @author Hemma Prafullchandra
 * @version 1.9
 */
public class GeneralNames extends Vector {
    /**
     * Create the GeneralNames, decoding from the passed DerValue.
     *
     * @param derVal the DerValue to construct the GeneralNames from.
     * @exception GeneralNamesException on decoding error.
     * @exception IOException on error.
     */
    public GeneralNames(DerValue derVal)
    throws IOException, GeneralNamesException {
        if (derVal.tag != DerValue.tag_Sequence) {
            throw new IOException("Invalid encoding for GeneralNames.");
        }
        if (derVal.data.available() == 0) {
            throw new GeneralNamesException("No data available in "
                                      + "passed DER encoded value.");
        }
        // Decode all the GeneralName's
        while (derVal.data.available() != 0) {
            DerValue encName = derVal.data.getDerValue();

            GeneralName name = new GeneralName(encName);
            addElement(name);
        }
    }

    /**
     * The default constructor for this class.
     */
    public GeneralNames() {
        super(1, 1);
    }

    /**
     * Write the extension to the DerOutputStream.
     *
     * @param out the DerOutputStream to write the extension to.
     * @exception GeneralNamesException on encoding error.
     * @exception IOException on error.
     */
    public void encode(DerOutputStream out)
    throws IOException, GeneralNamesException {
        if (size() == 0) { return; }

        Enumeration names = elements();
        DerOutputStream temp = new DerOutputStream();

        while (names.hasMoreElements()) {
            Object obj = names.nextElement();
            if (!(obj instanceof GeneralNameInterface)) {
	        throw new GeneralNamesException("Element in GeneralNames "
                                         + "not of type GeneralName.");
            }
	    GeneralNameInterface intf = (GeneralNameInterface)obj;
	    DerOutputStream gname = new DerOutputStream();
            intf.encode(gname);
            int nameType = intf.getType();
            // constructed form
            if (nameType == GeneralNameInterface.NAME_ANY ||
                nameType == GeneralNameInterface.NAME_X400 ||
                nameType == GeneralNameInterface.NAME_DIRECTORY ||
                nameType == GeneralNameInterface.NAME_EDI) {

                temp.writeImplicit(DerValue.createTag(DerValue.TAG_CONTEXT,
                             true, (byte)nameType), gname);
            } else // primitive form
                temp.writeImplicit(DerValue.createTag(DerValue.TAG_CONTEXT,
                             false, (byte)nameType), gname);
        }
        out.write(DerValue.tag_Sequence, temp);
    }
}
