/*
 * @(#)OIDMap.java	1.18 98/07/10
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.security.x509;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.security.cert.CertificateException;
import java.security.cert.CertificateParsingException;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Properties;

import sun.security.util.*;

/**
 * This class defines the mapping from OID & name to classes and vice
 * versa.  Used by CertificateExtensions & PKCS10 to get the java
 * classes associated with a particular OID/name.
 *
 * Note that the classes have to be on the system classpath with
 * the new security model/class loader mechanism.
 *
 * @author Amit Kapoor
 * @author Hemma Prafullchandra
 * @version 1.18
 */
public class OIDMap {

    // user-friendly names
    private static final String ROOT = X509CertImpl.NAME + "." +
                                 X509CertInfo.NAME + "." +
                                 X509CertInfo.EXTENSIONS;
    private static final String AUTH_KEY_IDENTIFIER = ROOT + "." +
                                          AuthorityKeyIdentifierExtension.NAME;
    private static final String SUB_KEY_IDENTIFIER  = ROOT + "." +
                                          SubjectKeyIdentifierExtension.NAME;
    private static final String KEY_USAGE           = ROOT + "." +
                                          KeyUsageExtension.NAME;
    private static final String PRIVATE_KEY_USAGE   = ROOT + "." +
                                          PrivateKeyUsageExtension.NAME;
    private static final String POLICY_MAPPINGS     = ROOT + "." +
                                          PolicyMappingsExtension.NAME;
    private static final String SUB_ALT_NAME        = ROOT + "." +
                                          SubjectAlternativeNameExtension.NAME;
    private static final String ISSUER_ALT_NAME     = ROOT + "." +
                                          IssuerAlternativeNameExtension.NAME;
    private static final String BASIC_CONSTRAINTS   = ROOT + "." +
                                          BasicConstraintsExtension.NAME;
    private static final String NAME_CONSTRAINTS    = ROOT + "." +
                                          NameConstraintsExtension.NAME;
    private static final String POLICY_CONSTRAINTS  = ROOT + "." +
                                          PolicyConstraintsExtension.NAME;
    private static final String CRL_NUMBER  = ROOT + "." +
                                              CRLNumberExtension.NAME;
    private static final String CRL_REASON  = ROOT + "." +
                                              CRLReasonCodeExtension.NAME;
    private static final String NETSCAPE_CERT  = ROOT + "." +
                                              NetscapeCertTypeExtension.NAME;

    private static final Hashtable oid2Name = new Hashtable(15);
    private static final Hashtable name2OID = new Hashtable(15);
    private static final Hashtable name2Class = new Hashtable(15);

    // Initialize supported extensions
    static {
        loadNames();
        loadClasses();
    }

    // Load the name-to-oid map
    private static void loadNamesDefault(Properties props) {
        props.put(SUB_KEY_IDENTIFIER, "2.5.29.14");
        props.put(KEY_USAGE, "2.5.29.15");
        props.put(PRIVATE_KEY_USAGE, "2.5.29.16");
        props.put(SUB_ALT_NAME, "2.5.29.17");
        props.put(ISSUER_ALT_NAME, "2.5.29.18");
        props.put(BASIC_CONSTRAINTS, "2.5.29.19");
        props.put(CRL_NUMBER, "2.5.29.20");
        props.put(CRL_REASON, "2.5.29.21");
        props.put(NAME_CONSTRAINTS, "2.5.29.30");
        props.put(POLICY_MAPPINGS, "2.5.29.33");
        props.put(AUTH_KEY_IDENTIFIER, "2.5.29.35");
        props.put(POLICY_CONSTRAINTS, "2.5.29.36");
        props.put(NETSCAPE_CERT, "2.16.840.1.113730.1.1");
    }

    // Load the name-to-class map
    private static void loadClassDefault (Properties props) {
        props.put(AUTH_KEY_IDENTIFIER,
                   "sun.security.x509.AuthorityKeyIdentifierExtension");
        props.put(SUB_KEY_IDENTIFIER,
                  "sun.security.x509.SubjectKeyIdentifierExtension");
        props.put(KEY_USAGE,
                  "sun.security.x509.KeyUsageExtension");
        props.put(PRIVATE_KEY_USAGE,
                  "sun.security.x509.PrivateKeyUsageExtension");
        props.put(POLICY_MAPPINGS,
                  "sun.security.x509.PolicyMappingsExtension");
        props.put(SUB_ALT_NAME,
                  "sun.security.x509.SubjectAlternativeNameExtension");
        props.put(ISSUER_ALT_NAME,
                  "sun.security.x509.IssuerAlternativeNameExtension");
        props.put(BASIC_CONSTRAINTS,
                  "sun.security.x509.BasicConstraintsExtension");
        props.put(NAME_CONSTRAINTS,
                  "sun.security.x509.NameConstraintsExtension");
        props.put(POLICY_CONSTRAINTS,
                  "sun.security.x509.PolicyConstraintsExtension");
        props.put(CRL_NUMBER, "sun.security.x509.CRLNumberExtension");
        props.put(CRL_REASON, "sun.security.x509.CRLReasonCodeExtension");
        props.put(NETSCAPE_CERT, "sun.security.x509.NetscapeCertTypeExtension");
    }

    // Load the name-to-oid map
    private static void loadNames() {
        Properties props = new Properties();
	loadNamesDefault(props);

        Enumeration names = props.propertyNames();
        while (names.hasMoreElements()) {
            String name = (String)names.nextElement();
            String oidName = props.getProperty(name);
            ObjectIdentifier oid = new ObjectIdentifier(oidName);

            name2OID.put(name, oid);
            oid2Name.put(oid, name);
        }
    }

    // Load the name-to-class map
    private static void loadClasses() {
        Properties props = new Properties();
	loadClassDefault(props);

        Enumeration names = props.propertyNames();
        while (names.hasMoreElements()) {
            String name = (String)names.nextElement();
            String className = props.getProperty(name);

            name2Class.put(name, className);
        }
    }

    /**
     * Add a name to lookup table.
     *
     * @param className the name of the fully qualified class implementing
     *         the asn object.
     * @param oid the string representation of the object identifier for
     *         the class.
     * @param name the name of the attribute.
     * @exception CertificateException on errors.
     */
    public static void addAttribute(String className, String oid, String name)
    throws CertificateException {
        ObjectIdentifier objId = new ObjectIdentifier(oid);
	if (oid2Name.get(objId) != null) {
	    throw new CertificateException("Object identifier already exists");
	}
	if (name2OID.get(name) != null) {
	    throw new CertificateException("Name already exists");
	}
	if (name2Class.get(className) != null) {
	    throw new CertificateException("Class already exists");
	}
	oid2Name.put(objId, name);
	name2OID.put(name, objId);
	name2Class.put(name, className);
    }

    /**
     * Return user friendly name associated with the OID.
     *
     * @param oid the name of the object identifier to be returned.
     * @return the user friendly name or null if no name
     * is registered for this oid.
     */
    public static String getName(ObjectIdentifier oid) {
        return (String)oid2Name.get(oid);
    }

    /**
     * Return Object identifier for user friendly name.
     *
     * @param name the user friendly name.
     * @return the Object Identifier or null if no oid
     * is registered for this name.
     */
    public static ObjectIdentifier getOID(String name) {
        return (ObjectIdentifier)name2OID.get(name);
    }

    /**
     * Return the java class object associated with the user friendly name.
     *
     * @param name the user friendly name.
     * @exception CertificateException if class cannot be instantiated.
     */
    public static Class getClass(String name) throws CertificateException {
        String className = (String)name2Class.get(name);
        if (className == null)
            return null;
        try {
            Class extClass = Class.forName(className);
	    return (extClass);
	} catch (Exception e) {
            throw new CertificateException("Error instantiating class for "
                                + name + " " + e.toString());
	}
    }

    /**
     * Return the java class object associated with the object identifier..
     *
     * @param oid the name of the object identifier to be returned.
     * @exception CertificateException if class cannot be instatiated.
     */
    public static Class getClass(ObjectIdentifier oid)
    throws CertificateException {
        String name = getName(oid);
        if (name == null)
            return null;
        String className = (String)name2Class.get(name);
        if (className == null)
            return null;
        try {
            Class extClass = Class.forName(className);
	    return (extClass);
	} catch (Exception e) {
	    throw new CertificateException("Error instantiating class for "
                                   + name + " " + e.toString());
	}
    }
}
