/*
 * @(#)X500Name.java	1.38 98/06/29
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.security.x509;

import java.io.IOException;
import java.io.StringBufferInputStream;
import java.security.Principal;
import java.util.Enumeration;
import java.util.Vector;

import sun.security.util.*;
import sun.security.pkcs.PKCS9Attribute;

/**
 * X.500 names are used to identify entities, such as those which are
 * identified by X.509 certificates.  They are world-wide, hierarchical,
 * and descriptive.  Entities can be identified by attributes, and in
 * some systems can be searched for according to those attributes.
 *
 * @author David Brownell
 * @author Amit Kapoor
 * @author Hemma Prafullchandra
 * @version 1.38
 * @see GeneralName
 * @see GeneralNames
 * @see GeneralNameInterface
 */

public class X500Name implements Principal, GeneralNameInterface {
    /**
     * Constructs a name from a conventionally formatted string, such
     * as "CN=Dave, OU=JavaSoft, O=Sun Microsystems, C=US".
     * (RFC 1779 style).
     *
     * @param DN X.500 Distinguished Name
     */
    public X500Name(String dname) throws IOException {
	parseDN(dname);
    }

    /**
     * Constructs a name from fields common in enterprise application
     * environments.
     *
     * <P><EM><STRONG>NOTE:</STRONG>  The behaviour when any of
     * these strings contain characters outside the ASCII range
     * is unspecified in currently relevant standards.</EM>
     *
     * @param commonName common name of a person, e.g. "Vivette Davis"
     * @param organizationUnit small organization name, e.g. "Purchasing"
     * @param organizationName large organization name, e.g. "Onizuka, Inc."
     * @param country two letter country code, e.g. "CH"
     */
    public X500Name(String commonName, String organizationUnit,
                     String organizationName, String country)
    throws IOException {
	names = new RDN[4];
	/*
	 * NOTE:  it's only on output that little-endian
	 * ordering is used.
	 */
	names[3] = new RDN(1);
	names[3].assertion[0] = new AVA(commonName_oid,
		new DerValue(commonName));
	names[2] = new RDN(1);
	names[2].assertion[0] = new AVA(orgUnitName_oid,
		new DerValue(organizationUnit));
	names[1] = new RDN(1);
	names[1].assertion[0] = new AVA(orgName_oid,
		new DerValue(organizationName));
	names[0] = new RDN(1);
	names[0].assertion[0] = new AVA(countryName_oid,
		new DerValue(country));
    }

    /**
     * Constructs a name from fields common in Internet application
     * environments.
     *
     * <P><EM><STRONG>NOTE:</STRONG>  The behaviour when any of
     * these strings contain characters outside the ASCII range
     * is unspecified in currently relevant standards.</EM>
     *
     * @param commonName common name of a person, e.g. "Vivette Davis"
     * @param organizationUnit small organization name, e.g. "Purchasing"
     * @param organizationName large organization name, e.g. "Onizuka, Inc."
     * @param localityName locality (city) name, e.g. "Palo Alto"
     * @param stateName state name, e.g. "California"
     * @param country two letter country code, e.g. "CH"
     */
    public X500Name(String commonName, String organizationUnit,
                    String organizationName, String localityName,
                    String stateName, String country)
    throws IOException {
	names = new RDN[6];
	/*
	 * NOTE:  it's only on output that little-endian
	 * ordering is used.
	 */
	names[5] = new RDN(1);
	names[5].assertion[0] = new AVA(commonName_oid,
		new DerValue(commonName));
	names[4] = new RDN(1);
	names[4].assertion[0] = new AVA(orgUnitName_oid,
		new DerValue(organizationUnit));
	names[3] = new RDN(1);
	names[3].assertion[0] = new AVA(orgName_oid,
		new DerValue(organizationName));
	names[2] = new RDN(1);
	names[2].assertion[0] = new AVA(localityName_oid,
		new DerValue(localityName));
	names[1] = new RDN(1);
	names[1].assertion[0] = new AVA(stateName_oid,
		new DerValue(stateName));
	names[0] = new RDN(1);
	names[0].assertion[0] = new AVA(countryName_oid,
		new DerValue(country));
    }

    /**
     * Constructs a name from an ASN.1 encoded value.  The encoding
     * of the name in the stream uses DER (a BER/1 subset).
     *
     * @param value a DER-encoded value holding an X.500 name.
     */
    public X500Name(DerValue value) throws IOException {
	this(value.toDerInputStream());
    }

    /**
     * Constructs a name from an ASN.1 encoded input stream.  The encoding
     * of the name in the stream uses DER (a BER/1 subset).
     *
     * @param in DER-encoded data holding an X.500 name.
     */
    public X500Name(DerInputStream in) throws IOException {
	parseDER(in);
    }

    /**
     *  Constructs a name from an ASN.1 encoded byte array.
     *
     * @param name DER-encoded byte array holding an X.500 name.
     */
    public X500Name(byte[] name) throws IOException {
        DerInputStream in = new DerInputStream(name);
        parseDER(in);
    }

    /**
     * Compares this name with another, for equality.
     *
     * @return true iff the names are identical.
     */
    public boolean equals(Object other) {
	if (other instanceof X500Name)
	    return equals((X500Name)other);
	else
	    return false;
    }

    /**
     * Calculates a hash code value for the object.  Objects
     * which are equal will also have the same hashcode.
     */
    public int hashCode() {
	int retval = 0;
	int k = 1;

	for (int i=0; i < names.length; i++)
	    retval += names[i].hashCode() * k++;

	return retval;
    }

    /**
     * Compares this name with another, for equality.
     *
     * @return true iff the names are identical.
     */
    synchronized public boolean equals(X500Name other) {
	int i;

	if (this == other)
	    return true;

	if (names.length != other.names.length)
	    return false;
	for (i = 0; i < names.length; i++) {
	    if (!names[i].equals(other.names[i]))
		return false;
	}
	return true;
    }

    /*
     * Returns the name component as a Java string, regardless of its
     * encoding restrictions.
     */
    private String getString(DerValue attribute) throws IOException {
        if (attribute == null)
            return null;
	String	value = attribute.getAsString();

	if (value == null)
	    throw new IOException("not a DER string encoding, "
		    + attribute.tag);
	else
	    return value;
    }

    /**
     * Return type of GeneralName.
     */
    public int getType() {
        return (GeneralNameInterface.NAME_DIRECTORY);
    }

    /**
     * Returns a "Country" name component.  If more than one
     * such attribute exists, the topmost one is returned.
     *
     * @return "C=" component of the name, if any.
     */
    public String getCountry() throws IOException {
	DerValue attr = findAttribute(countryName_oid);

	return getString(attr);
    }


    /**
     * Returns an "Organization" name component.  If more than
     * one such attribute exists, the topmost one is returned.
     *
     * @return "O=" component of the name, if any.
     */
    public String getOrganization() throws IOException {
	DerValue attr = findAttribute(orgName_oid);

	return getString(attr);
    }


    /**
     * Returns an "Organizational Unit" name component.  If more
     * than one such attribute exists, the topmost one is returned.
     *
     * @return "OU=" component of the name, if any.
     */
    public String getOrganizationalUnit() throws IOException {
	DerValue attr = findAttribute(orgUnitName_oid);

	return getString(attr);
    }


    /**
     * Returns a "Common Name" component.  If more than one such
     * attribute exists, the topmost one is returned.
     *
     * @return "CN=" component of the name, if any.
     */
    public String getCommonName() throws IOException {
	DerValue attr = findAttribute(commonName_oid);

	return getString(attr);
    }


    /**
     * Returns a "Locality" name component.  If more than one
     * such component exists, the topmost one is returned.
     *
     * @return "L=" component of the name, if any.
     */
    public String getLocality() throws IOException {
	DerValue attr = findAttribute(localityName_oid);

	return getString(attr);
    }

    /**
     * Returns a "State" name component.  If more than one
     * such component exists, the topmost one is returned.
     *
     * @return "S=" component of the name, if any.
     */
    public String getState() throws IOException {
	DerValue attr = findAttribute(stateName_oid);

	return getString(attr);
    }

    /**
     * Returns a string form of the X.500 distinguished name.
     * The format of the string is from RFC 1779.
     */
    public String toString() {
	if (dn == null)
	    generateDN();
	return dn;
    }

    /**
     * Returns the value of toString().  This call is needed to
     * implement the java.security.Principal interface.
     */
    public String getName() { return toString(); }

    private String	dn = null; // RFC 1779 style DN, or null
    private RDN[]	names;	// RDNs

    /**
     * Find the first instance of this attribute in a "top down"
     * search of all the attributes in the name.
     */
    private DerValue findAttribute(ObjectIdentifier attribute) {
	int i;
	DerValue retval = null;

	for (i = 0; i < names.length; i++) {
	    retval = names[i].findAttribute(attribute);
	    if (retval != null)
		break;
	}
	return retval;
    }

    /****************************************************************/

    private void parseDER(DerInputStream in) throws IOException {
	//
	// X.500 names are a "SEQUENCE OF" RDNs, which means one or
	// more and order matters.  We scan them in order, which
	// conventionally is big-endian.
	//
	DerValue[] nameseq = in.getSequence(5);
	int i;

	names = new RDN[nameseq.length];
	for (i = 0; i < nameseq.length; i++)
	    names[i] = new RDN(nameseq[i]);
    }

    /**
     * Encodes the name in DER-encoded form.
     *
     * @deprecated Use encode() instead
     * @param out where to put the DER-encoded X.500 name
     */
    public void emit(DerOutputStream out) throws IOException {
	encode(out);
    }

    /**
     * Encodes the name in DER-encoded form.
     *
     * @param out where to put the DER-encoded X.500 name
     */
    public void encode(DerOutputStream out) throws IOException {
	DerOutputStream	tmp = new DerOutputStream();

        if (names != null) {
	    for (int i = 0; i < names.length; i++)
	        names[i].encode(tmp);
        }
	out.write(DerValue.tag_Sequence, tmp);
    }

    /**
     * Gets the name in DER-encoded form.
     *
     * @return the DER encoded byte array of this name.
     */
    public byte[] getEncoded() throws IOException {
        DerOutputStream	out = new DerOutputStream();
        DerOutputStream	tmp = new DerOutputStream();
        if (names != null) {
            for (int i = 0; i < names.length; i++)
                names[i].encode(tmp);
        }
        out.write(DerValue.tag_Sequence, tmp);
        return out.toByteArray();
    }

    /*
     * Parses a Distinguished Name (DN) in printable representation.
     * Only one AVA per RDN is supported.
     *
     * According to RFC 1779, RDNs in a DN are separated by comma.
     * The following examples show both methods of quoting a comma, so that it
     * is not considered a separator:
     *
     *     O="Sue, Grabbit and Runn" or
     *     O=Sue\, Grabbit and Runn
     */
    private void parseDN(String input) throws IOException {
        if (input == null || input.equals((String)"")) {
	    names = new RDN[0];
            return;
        }

	Vector dnVector = new Vector();
	int dnOffset = 0;
	int searchOffset;
	int rdnEnd;
	String rdnString;
	int quoteCount = 0;

	// Remove any leading and trailing whitespaces from the input
	String dnString = input.trim();

	for (searchOffset = 0;
	     (rdnEnd = dnString.indexOf(',', searchOffset)) >= 0;
	     searchOffset = rdnEnd + 1) {

	    quoteCount += countQuotes(dnString, searchOffset, rdnEnd);

	    /*
	     * We have encountered a comma.
	     * If the comma in the RDN under consideration is preceded by a
	     * backslash (escape), or by a double quote, the comma is part of
	     * the RDN.
	     * Otherwise, the comma is used as a separator, to delimit the
	     * RDN under consideration from any subsequent RDNs.
	     */
	    if (rdnEnd > 0 && dnString.charAt(rdnEnd - 1) != '\\'
		   && quoteCount != 1) {
		/*
		 * Comma is a separator
		 */
		rdnString = (dnString.substring(dnOffset, rdnEnd)).trim();

		// Parse RDN, and store it in vector
		parseRDN(rdnString, dnVector);

		// Increase the offset
		dnOffset = rdnEnd + 1;

		// Set quote counter back to zero
		quoteCount = 0;
	    }
	}

	// Parse last or only RDN, and store it in vector
	rdnString = (dnString.substring(dnOffset)).trim();
	parseRDN(rdnString, dnVector);

	/*
	 * Store the vector elements as an array of RDNs
	 * NOTE: It's only on output that little-endian ordering is used.
	 */
	int numElem = dnVector.size();
	names = new RDN[numElem];
	Enumeration e = dnVector.elements();
	int i = numElem - 1;

	while (e.hasMoreElements()) {
	    names[i--] = (RDN)e.nextElement();
	}
    }

    /*
     * Parses a Relative Distinguished Name (RDN) in printable representation.
     * Only one AVA per RDN is supported.
     */
    private void parseRDN(String rdnString, Vector dnVector)
    throws IOException {

	    RDN rdn = new RDN(1);
	    rdn.assertion[0] = new AVA(new StringBufferInputStream(rdnString));
	    dnVector.addElement(rdn);
    }

    /*
     * Counts double quotes in string.
     */
    private int countQuotes(String string, int from, int to) {
	int count = 0;

	for (int i = from; i < to; i++) {
	    if (string.charAt(i) == '"') {
		count++;
	    }
	}

	return count;
    }
    /*
     * Dump the printable form of a distinguished name.  Each relative
     * name is separated from the next by a ",", and assertions in the
     * relative names have "label=value" syntax.
     *
     * Uses RFC 1779 syntax (i.e. little-endian, comma separators)
     */
    private void generateDN() {
	int i;
	String fullname = null;

	for (i = names.length - 1; i >= 0; i--) {
	    if (fullname == null)
		fullname = names[i].toString();
	    else
		fullname += ", " + names[i].toString();
	}
	dn = fullname;
    }

    /****************************************************************/

    /*
     * Maybe return a preallocated OID, to reduce storage costs
     * and speed recognition of common X.500 attributes.
     *
     * N.B. "intern" in older systems just grew an internal table;
     * this one is fixed size.  It's probably fine this way.
     */
    static ObjectIdentifier intern(ObjectIdentifier oid) {
	if (oid.equals(commonName_oid))
	    return commonName_oid;
	if (oid.equals(countryName_oid))
	    return countryName_oid;
	if (oid.equals(localityName_oid))
	    return localityName_oid;
	if (oid.equals(stateName_oid))
	    return stateName_oid;

	if (oid.equals(streetAddress_oid))
	    return streetAddress_oid;
	if (oid.equals(orgName_oid))
	    return orgName_oid;
	if (oid.equals(orgUnitName_oid))
	    return orgUnitName_oid;
	if (oid.equals(title_oid))
	    return title_oid;
	if (oid.equals(ipAddress_oid))
	    return ipAddress_oid;
        if (oid.equals(PKCS9Attribute.EMAIL_ADDRESS_OID))
            return PKCS9Attribute.EMAIL_ADDRESS_OID;

	return oid;
    }

    /*
     * Selected OIDs from X.520
     */
    private static final int commonName_data[] = { 2, 5, 4, 3 };
    private static final int countryName_data[] = { 2, 5, 4, 6 };
    private static final int localityName_data[] = { 2, 5, 4, 7 };
    private static final int stateName_data[] = { 2, 5, 4, 8 };
    private static final int streetAddress_data[] = { 2, 5, 4, 9 };
    private static final int orgName_data[] = { 2, 5, 4, 10 };
    private static final int orgUnitName_data[] = { 2, 5, 4, 11 };
    private static final int title_data[] = { 2, 5, 4, 12 };

    /** OID for the "CN=" attribute, denoting a person's common name. */
    public static final ObjectIdentifier
	commonName_oid = new ObjectIdentifier(commonName_data);

    /** OID for the "C=" attribute, denoting a country. */
    public static final ObjectIdentifier
	countryName_oid = new ObjectIdentifier(countryName_data);

    /** OID for the "L=" attribute, denoting a locality (such as a city) */
    public static final ObjectIdentifier
	localityName_oid = new ObjectIdentifier(localityName_data);

    /** OID for the "O=" attribute, denoting an organization name */
    public static final ObjectIdentifier
	orgName_oid = new ObjectIdentifier(orgName_data);

    /** OID for the "OU=" attribute, denoting an organizational unit name */
    public static final ObjectIdentifier
	orgUnitName_oid = new ObjectIdentifier(orgUnitName_data);

    /** OID for the "S=" attribute, denoting a state (such as Delaware) */
    public static final ObjectIdentifier
	stateName_oid = new ObjectIdentifier(stateName_data);

    /** OID for the "STREET=" attribute, denoting a street address. */
    public static final ObjectIdentifier
	streetAddress_oid = new ObjectIdentifier(streetAddress_data);

    /** OID for the "T=" attribute, denoting a person's title. */
    public static final ObjectIdentifier
	title_oid = new ObjectIdentifier(title_data);

    /*
     * OIDs from other sources which show up in X.500 names we
     * expect to deal with often
     */
    private static final int ipAddress_data[] =		// SKIP
	{ 1, 3, 6, 1, 4, 1, 42, 2, 11, 2, 1 };

    /** OID for "IP=" IP address attributes, used with SKIP. */
    public static final ObjectIdentifier
	ipAddress_oid = new ObjectIdentifier(ipAddress_data);
}


/*
 * RDNs are a set of {attribute = value} assertions.  Some of those
 * attributes are "distinguished" (unique w/in context).  Order is
 * never relevant.
 *
 * Some X.500 names include only a single distinguished attribute
 * per RDN.  This style is currently common.
 *
 * Note that DER-encoded RDNs sort AVAs by assertion OID ... so that
 * when we parse this data we don't have to worry about canonicalizing
 * it, but we'll need to sort them when we expose the RDN class more.
 */
class RDN {
    protected AVA[]	assertion;

    RDN(DerValue set) throws IOException {
	if (set.tag != DerValue.tag_Set)
	    throw new IOException("X500 RDN");

	int j_max = 1;			// XXX j_max = f(data)!!
	int	j;

	assertion = new AVA[j_max];

	for (j = 0; j < j_max; j++)
	    assertion[j] = new AVA(set.data);
    }

    RDN(int i) { assertion = new AVA[i]; }

    public boolean equals(Object other) {
	if (other instanceof RDN)
	    return equals((RDN)other);
	else
	    return false;
    }

    boolean equals(RDN other) {
	int	i;

	if (other == this)
	    return true;
	if (assertion.length != other.assertion.length)
	    return false;

	for (i = 0; i < assertion.length; i++)
	    if (!assertion[i].equals(other.assertion[i]))
		return false;

	return true;
    }

    public int hashCode() {
        int retval = 0;
        int k = 1;

        for (int i=0; i < assertion.length; i++) {
            if (assertion[i].value != null)
                retval += assertion[i].value.length() * k++;
        }
        return retval;
    }


    DerValue findAttribute(ObjectIdentifier oid) {
	int i;

	for (i = 0; i < assertion.length; i++)
	    if (assertion[i].oid.equals(oid))
		return assertion[i].value;
	return null;
    }

    void encode(DerOutputStream out) throws IOException {
	out.putOrderedSetOf(DerValue.tag_Set, assertion);
    }

    /**
     * Returns a printable form of this RDN, using RFC 1779 style catenation
     * of attribute/value assertions.
     */
    public String toString() {
	int i;
	String relname = null;

	for (i = 0; i < assertion.length; i++) {
	    if (relname == null)
		relname = assertion[i].toString();
	    else
		relname += " + " + assertion[i];
	}
	return relname;
    }
}
