/*
 * @(#)CachedClass.java	1.11 98/06/09
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.tools.agent;

import java.util.Hashtable;
import java.util.Arrays;
import java.util.Set;
import java.util.Iterator;

/**
 *   The Cached* classes cache data retrieved from JVMDI  
 *   This class provides access to class information.
 *   @author Robert Field
 */

class CachedClass {
    final Class realClass;
    private Hashtable methodCache;        // keyed on methodID
    private Hashtable fieldCache;         // keyed on fieldID
    static private Hashtable classCache;  // keyed on realClass

    static private native String getSourceFileName(Class clazz)
                                    throws AbsentInformationException;
    static private native int[]  getLineNumbers(Class clazz)
                                    throws AbsentInformationException;
    static private native long[]  getMethodIDs(Class clazz)
                                    throws AbsentInformationException;
    static private native long[]  getFieldIDs(Class clazz)
                                    throws AbsentInformationException;

    private CachedClass(Class realClass) {
	this.realClass = realClass;
        this.methodCache = null;
        this.fieldCache = null;
    }

    Class getRealClass() {
        return realClass;
    }

    private Hashtable methodCache() {
        if (methodCache == null) {
            methodCache = new Hashtable();
        }
        return methodCache;
    }
        
    /**
     * for use only by CachedMethod.
     */
    CachedMethod findMethod(long methodID) {
        return (CachedMethod)methodCache().get(new Long(methodID));
    }
        
    /**
     * for use only by CachedMethod.
     */
    void cacheMethod(long methodID, CachedMethod method) {
        methodCache().put(new Long(methodID), method);
    }
        
    private Hashtable fieldCache() {
        if (fieldCache == null) {
            fieldCache = new Hashtable();
        }
        return fieldCache;
    }
        
    /**
     * for use only by CachedField.
     */
    CachedField findField(long fieldID) {
        return (CachedField)fieldCache().get(new Long(fieldID));
    }
        
    /**
     * for use only by CachedField.
     */
    void cacheField(long fieldID, CachedField field) {
        fieldCache().put(new Long(fieldID), field);
    }
        
    private static Hashtable classCache() {
        if (classCache == null) {
            classCache = new Hashtable();
        }
        return classCache;
    }
        
    String getName() {
        return realClass.getName();
    }

    String getSourceFileName() throws AbsentInformationException {
	return getSourceFileName(realClass);
    }

    Location lineToLocation(int line) throws AbsentInformationException {
        CachedMethod[] methods = getMethods();
        Location location = null;
        for (int i = 0; (i < methods.length) && (location == null); i++) {
            location = methods[i].lineToLocation(line);
        }
        return location;
    }

    int[] getLinenumbers() throws AbsentInformationException {
        CachedMethod[] methods = getMethods();
        int totalSize = 0;
        for (int i = 0; i < methods.length; i++) {
            totalSize += methods[i].getLinenumberCount();
        }

        int[] lineNumbers = new int[totalSize];
        if (totalSize == 0) {
            return lineNumbers;
        }
        int pos = 0;
        for (int i = 0; i < methods.length; i++) {
            int[] methodLines = methods[i].getLinenumbers();
            System.arraycopy(methodLines, 0, lineNumbers, 
                             pos, methodLines.length);
            pos += methodLines.length;
        }
        Arrays.sort(lineNumbers);
        // check for duplicates
        int prev = lineNumbers[0];
        for (int lead = 1; lead < totalSize; ++lead) {
            int curr = lineNumbers[lead];
            if (prev == curr) {
                // duplicate(s) found - remove them
                int follow = lead;
                for (++lead; lead < totalSize; ++lead) {
                    curr = lineNumbers[lead];
                    if (prev != curr) {
                        // compress duplicates in-place
                        lineNumbers[follow++] = curr;
                        prev = curr;
                    }
                }
                // shrink and return compressed line number array
                int[] newLineNumbers = new int[follow];
                System.arraycopy(lineNumbers, 0, newLineNumbers, 0, follow);
                return newLineNumbers;
            }
            prev = curr;
        }
        // no duplicates
                
        return lineNumbers;
    }

    CachedMethod[] getMethods() throws AbsentInformationException {
        long[] ids = getMethodIDs(realClass);
	int size =  ids.length;
	CachedMethod[] meths = new CachedMethod[size];
	for (int inx = 0; inx < size; ++inx) {
	    meths[inx] = getMethod(ids[inx]);
	}
	return meths;
    }

    CachedField[] getFields() throws AbsentInformationException {
        long[] ids = getFieldIDs(realClass);
	int size =  ids.length;
	CachedField[] flds = new CachedField[size];
	for (int inx = 0; inx < size; ++inx) {
	    flds[inx] = getField(ids[inx]);
	}
	return flds;
    }

    CachedClass getSuperclass() {
        Class superclass = realClass.getSuperclass();
        if (superclass == null) {
            return null;
        } else {
            return getCachedClass(superclass);
        }
    }

    static CachedClass getCachedClass(String className) throws ClassNotFoundException {
        return getCachedClass(Class.forName(className,
					    true, Agent.getAppLoader()));
    }

    static CachedClass getCachedClass(Class clazz) {
        CachedClass cls = (CachedClass)classCache().get(clazz);
        if (cls == null) {
            cls = new CachedClass(clazz);
            classCache.put(clazz, cls);
        }
        return cls;
    }

    static CachedClass findCachedClass(Class clazz) {
        return (CachedClass)classCache().get(clazz);
    }

    static void freeAllBut(Set keepSet) {
        classCache.clear();
        // The event object caches references to cached classes and methods
        // Since we've cleared many of them here, we must make sure they 
        // don't exist in the Handler's cache.
        Handler.the_event.clearCache();
        Iterator iter = keepSet.iterator();
        while (iter.hasNext()) {
            CachedClass clazz = (CachedClass)iter.next();
            classCache.put(clazz.getRealClass(), clazz);
        }
    }
            

    CachedMethod getMethod(long methodID) {
        return CachedMethod.getMethod(this, methodID);
    }

    CachedMethod getMethod(String name, String signature) {
        return CachedMethod.getMethod(this, name, signature);
    }

    CachedField getField(String name, String signature) {
        return CachedField.getField(this, name, signature);
    }

    CachedField getField(long fieldID) {
        return CachedField.getField(this, fieldID);
    }

    public String toString() {
        return getName();
    }
}
