/*
 * @(#)CachedMethod.java	1.7 98/05/27
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.tools.agent;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 *   The Cached* classes cache data retrieved from JVMDI.  
 *   This class provides access to method information.
 *   @author Robert Field
 */

class CachedMethod {
    final long methodID;
    final CachedClass cachedClass;
    final String name;
    final String signature;

    private byte[] codeBytes;

    private int endBCI;
    private LineEntry[] lineArray;
    private List variableTable;

    private int argumentCount = -1;

    static private class LineEntry {
        long startLocation;
        int line;

        LineEntry(long startLocation, int line) {
            this.startLocation = startLocation;
            this.line = line;
        }
    }

    static private class VariableEntry {
        long startLocation; 
        int scopeLength;
        String name;
        String signature;
        int slot;

        VariableEntry(long startLocation, int scopeLength, 
                      String name, String signature, int slot) {
            this.startLocation = startLocation;
            this.scopeLength = scopeLength;
            this.name = name;
            this.signature = signature;
            this.slot = slot;
        }
    }

    static protected native String getMethodName(Class clazz, long methodID);
    static protected native String getMethodSignature(Class clazz, 
                                                      long methodID);
    static protected native long getMethodID(Class clazz, 
                                   String name, String signature);

    static protected native long getStaticMethodID(Class clazz, 
                                         String name, String signature);

    static protected native byte[] getCode(Class clazz, long methodID);
    static protected native LineEntry[] getLineTable(Class clazz, 
                                                     long methodID)
                                         throws AbsentInformationException;
    static protected native long getEndLocation(Class clazz, long methodID);
    static protected native VariableEntry[] getVariableTable(Class clazz, 
                                                             long methodID)
                                         throws AbsentInformationException;
    static protected native Class getDeclaringClass(Class clazz, 
                                                    long methodID);
    static protected native int getModifiers(Class clazz, long methodID)
                                         throws AbsentInformationException;
    static protected native int getArgumentsSize(Class clazz, long methodID)
                                         throws AbsentInformationException;

    private CachedMethod(CachedClass cls, long methodID, 
                         String name, String signature) {
	this.methodID = methodID;
        this.cachedClass = cls;
        this.name = name;
        this.signature = signature;
    }

    private void fetchCode() {
        if (codeBytes == null) {
            codeBytes = getCode(getRealClass(), methodID);
        }
    }
    
        
    private void fetchLineTable() throws AbsentInformationException {
        if (lineArray == null) {
            lineArray = getLineTable(getRealClass(), methodID);
            endBCI = (int)getEndLocation(getRealClass(), methodID);
        }
    }

    private void fetchVariableTable() throws AbsentInformationException {
        if (variableTable == null) {
            VariableEntry[] variableArray =
                    getVariableTable(getRealClass(), methodID);

            variableTable = new ArrayList(variableArray.length);
            for (int i = 0; i < variableArray.length; i++) {
                VariableEntry entry = variableArray[i];
                CachedLocalVariable variable = new CachedLocalVariable(
                                                     this, entry.name,
                                                     entry.signature, 
                                                     (int)entry.startLocation,
                                                     entry.scopeLength,
                                                     entry.slot);
                variableTable.add(variable);
            }
        }
    }

    static CachedMethod getMethod(CachedClass cls, long methodID) {
        CachedMethod method = cls.findMethod(methodID);
        if (method == null) {
            String name = getMethodName(cls.getRealClass(), methodID);
            String signature = getMethodSignature(cls.getRealClass(), methodID);
            method = new CachedMethod(cls, methodID, name, signature);
            cls.cacheMethod(methodID, method);
        }
        return method;
    }

    protected static long getStaticOrInstanceMethodID(CachedClass cls, 
                                         String name, String signature) {
        long methodID;
        try {
            methodID = getMethodID(cls.getRealClass(), name, signature);
        } catch(NoSuchMethodError exc) {
            methodID = getStaticMethodID(cls.getRealClass(), name, signature);
        }
        return methodID;
    }

    static CachedMethod getMethod(CachedClass cls, 
                                  String name, String signature) {
        long methodID = getStaticOrInstanceMethodID(cls, name, signature);
        return getMethod(cls, methodID);
    }

    /**
     * Return the byte stored at a given offset within code bytes
     */
    public int byteAt(int bci) throws AbsentInformationException {
	fetchCode();
        return  codeBytes[bci] & 0xFF;
    }

    /**
     * Return the short stored at a given offset within code bytes
     */
    public int shortAt(int bci) throws AbsentInformationException {
	fetchCode();
        return ((codeBytes[bci] & 0xFF) << 8)
            | (codeBytes[bci+1] & 0xFF);
    }

    /**
     * Return the int stored at a given offset within code bytes
     */
    public int intAt(int bci) throws AbsentInformationException {
	fetchCode();
        return (codeBytes[bci] << 24)
            | ((codeBytes[bci+1] & 0xFF) << 16)
            | ((codeBytes[bci+2] & 0xFF) << 8)
            | ((codeBytes[bci+3] & 0xFF));
    }


    /**
     * Return the location of the specified line in this method.
     * Return null if no such line.
     */
    Location lineToLocation(int line) throws AbsentInformationException {
        fetchLineTable();
        
        Location location = null;

        for (int i = 0; i < lineArray.length; ++i) {
            LineEntry entry = lineArray[i];
            if (entry.line == line) {
                int end = endBCI;
                if (i+1 < lineArray.length) {
                    end = (int)lineArray[i+1].startLocation - 1;
                }
                location = new Location(this, line, 
                                        (int)entry.startLocation, end);
                break;
            }
        }        
        
	return location;
    }

    /**
     * Return the location of this byte code index.
     */
    Location bciToLocation(int bci) throws AbsentInformationException {
        fetchLineTable();

        if (bci > endBCI) {
            return null;
        }

        int line = -1;
        for (int i = lineArray.length - 1; i >= 0; i--) {
            LineEntry entry = lineArray[i];
            if (bci >= entry.startLocation) {
                line = entry.line;
                break;
            }
        }

        if (line == -1) {
            return null;
        } else {
            return new Location(this, line, bci, bci + 1);
        }
    }

    /**
     * Return the location of this method: first line, and 
     * byte code index range.
     */
    Location location() throws AbsentInformationException {
        fetchLineTable();

        LineEntry firstEntry = lineArray[0];
        int line = firstEntry.line;
        /* find lowest line number */
        for (int i = 1; i < lineArray.length; ++i) {
            int eline = lineArray[i].line;
            if (eline < line) {
                line = eline;
            }
        }
        return new Location(this, line, 
                            (int)firstEntry.startLocation, endBCI);
    }

    int getLinenumber() throws AbsentInformationException {
        return location().line;
    }

    int[] getLinenumbers() throws AbsentInformationException {
        fetchLineTable();

        int[] lineNumbers = new int[lineArray.length];
        for (int i = 0; i < lineArray.length; ++i) {
            lineNumbers[i] = lineArray[i].line;
        }

        return lineNumbers;
    }

    int getLinenumberCount() throws AbsentInformationException {
        fetchLineTable();
        return lineArray.length;
    }
        

    String getName() {
	return name;
    }

    Class getRealClass() {
        return cachedClass.getRealClass();
    }

    CachedClass getCachedClass() {
        return cachedClass;
    }

    public String getTypeSignature() {
	return signature;
    }

    CachedClass getDefiningClass() {
        Class clazz = getDeclaringClass(getRealClass(), methodID);
        return CachedClass.getCachedClass(clazz);
    }

    int getModifiers() throws AbsentInformationException {
	return getModifiers(getRealClass(), methodID);
    }

    List getVisibleLocalVariables(int bci) 
                                throws AbsentInformationException {
        fetchVariableTable();

        List visibleVariables = new ArrayList();
        Iterator iter = variableTable.iterator();
        while (iter.hasNext()) {
            CachedLocalVariable variable = (CachedLocalVariable)iter.next();
            if (variable.isVisible(bci)) {
                visibleVariables.add(variable);
            }
        }

        return visibleVariables;
    }

    /**
     * Return the number of arguments.
     */
    int getArgumentCount() throws AbsentInformationException {
        if (argumentCount == -1) {
            fetchVariableTable();
            int size = getArgumentsSize(getRealClass(), methodID);
            int count = 0;
            int sizeUsed = 0;
            Iterator iter = variableTable.iterator();
            while (iter.hasNext() && (sizeUsed < size)) {
                String signature = ((CachedLocalVariable)iter.next()).signature;
                count++;
                /*
                 * Use signature instead of slot number here because
                 * there's no guarantee that slot numbers will be zero-based
                 * word indices
                 */
                if (signature.equals("J") || signature.equals("D")) {
                    sizeUsed += 2;
                } else {
                    sizeUsed++;
                }
            }

            if (sizeUsed != size) {
                throw new InternalError("Invalid Arguments in Local Variable Table");
            }
        }

        return argumentCount;
    }

    boolean isArgument(CachedLocalVariable variable) 
                                    throws AbsentInformationException {
        fetchVariableTable();

        int count = getArgumentCount();
        for (int i = 0; i < count; i++) {
            CachedLocalVariable candidate = 
                              (CachedLocalVariable)variableTable.get(i);
            if (variable.name.equals(candidate.name)) {
                return true;
            }
        }

        return false;
    }
 
    public String toString() {
        return cachedClass.toString() + "." + getName();
    }
}
