/*
 * @(#)Event.java	1.20 98/06/29
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.tools.agent;

class Event extends Object implements Cloneable {
    int                 kind;
    Thread		thread;
    Class               clazz;
    long                methodID;
    int                 bci;
    Throwable           exception;
    Class               catch_clazz;
    long                catch_methodID;
    int                 catch_bci;

    /* Event kinds */
    static final   int SINGLE_STEP       = 1;
    static final   int BREAKPOINT        = 2;
    static final   int FRAME_POP         = 3;
    static final   int EXCEPTION         = 4;
    static final   int USER_DEFINED      = 5;
    static final   int THREAD_START      = 6;
    static final   int THREAD_END        = 7;
    static final   int VM_DEATH          = 99;

    private CachedMethod cachedMethod;
    private CachedMethod cachedCatchMethod;
    private long cachedMethodID = 0;
    private long cachedCatchMethodID = 0;

    /* state variable for event notifications:
     *   true if Handler is waiting for an event,
     *   false if it is processing
     */
    boolean             waiting_for_event;

    synchronized CachedMethod getMethod() {
        if (clazz == null) {
            return null;
        }
        if (cachedMethodID != methodID) {
            CachedClass cc = CachedClass.getCachedClass(clazz);
            cachedMethod = cc.getMethod(methodID);
            cachedMethodID = methodID;
        }
        return cachedMethod;
    }

    int getBCI() {
        return bci;
    }

    Throwable getException() {
        return exception;
    }

    synchronized CachedMethod getCatchMethod() {
        if (catch_clazz == null) {
            return null;
        }
        if (cachedCatchMethodID != catch_methodID) {
            CachedClass cc = CachedClass.getCachedClass(catch_clazz);
            cachedCatchMethod = cc.getMethod(catch_methodID);
            cachedCatchMethodID = catch_methodID;
        }
        return cachedCatchMethod;
    }

    synchronized void clearCache() {
        // Since method ids can be reused, we must have a way to clear the cached
        // methods whenever it's possible that any class known to the debugger might
        // be GC'ed
        cachedMethod = null;
        cachedCatchMethod = null;
        cachedMethodID = 0;
        cachedCatchMethodID = 0;
    }

    int getCatchBCI() {
        return catch_bci;
    }

    /* Step Handler needs to make a copy of the event data before it
     * gets overritten by other threads.
     * Courtesy of Gordon Hirsch of SAS
     */
    protected synchronized Object clone()
    {
        Event newQ;
	try {
	    newQ = (Event) super.clone();
	} catch (CloneNotSupportedException e) {
	    // this shouldn't happen, since we are Cloneable
	    throw new InternalError();
	}
        newQ.kind = kind;
        newQ.thread = thread;
        newQ.clazz = clazz;
        newQ.methodID = methodID;
        newQ.bci = bci;
        newQ.exception = exception;
        newQ.catch_clazz = catch_clazz;
        newQ.catch_methodID = catch_methodID;
        newQ.catch_bci = catch_bci;
        newQ.waiting_for_event = waiting_for_event;
        return newQ;
    }

    synchronized boolean nextEvent() {
	do {
	    try {
		wait();
	    } catch (InterruptedException ex) {
		// just drop through
	    }
	}
	while (waiting_for_event);
	return true;
    }

    synchronized void reset() {
	bci = -1;
    }

    public String toString() {
        if (kind == SINGLE_STEP) {
            return "event:SINGLE_STEP(" +
                thread + ", " + getMethod() + " @ " + bci +")";
        } else if (kind == BREAKPOINT) {
            return "event:BREAKPOINT(" +
                thread + ", " + getMethod() + " @ " + bci +")";
        } else if (kind == FRAME_POP) {
            return "event:FRAME_POP(" + thread + ")";
        } else if (kind == EXCEPTION) {
            return "event:EXCEPTION(" +
                thread + ", " + getMethod() + " @ " + bci +
                " - exc: " + exception + ")";
        } else if (kind == USER_DEFINED) {
            return "event:USER_DEFINED";
        } else if (kind == THREAD_START) {
            return "event:THREAD_START(" + thread + ")";
        } else if (kind == THREAD_END) {
            return "event:THREAD_END(" + thread + ")";
        } else if (kind == VM_DEATH) {
            return "event:VM_DEATH";
        } else  {
            return "event: invalid kind = " + kind;
        }
    }
}
