/*
 * @(#)Handler.java	1.35 98/05/27
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.tools.agent;

import java.util.Enumeration;
import java.util.Hashtable;

import java.io.IOException;

class Handler extends Thread implements AgentConstants {
    static Event	        the_event;

    private Agent		agent;
    private Hashtable		catchHash;

    static final boolean	debug = false;

    synchronized void catchExceptionClass(Class c) {
	if (catchHash.get(c) == null) {
	    catchHash.put(c, c);
	}
    }

    synchronized void ignoreExceptionClass(Class c) {
	catchHash.remove(c);
    }

    Class getCatchList()[] {
	Class list[] = new Class[catchHash.size()];
	Enumeration enum = catchHash.elements();
	for (int i = 0; i < list.length; i++) {
	    list[i] = (Class)enum.nextElement();
	}
	return list;
    }

    Handler(Agent ag) {
	super("Breakpoint handler");
	the_event = new Event();
        the_event.waiting_for_event = false;
	catchHash = new Hashtable();
	agent = ag;
    }

    private void handleException() throws IOException {
        if (the_event.exception instanceof ThreadDeath) {
            if (the_event.exception == MainThread.rethrownDeath) {
                Agent.message("rethrown ThreadDeath - ignoring");
            } else {
                Agent.message("Received ThreadDeath event: " + the_event);
                agent.setSingleStep(the_event.thread, false);
                // ThreadDeaths are never returned to a user because the current
                // architecture doesn't let the user do much once ThreadDeath is
                // thrown (for example, you cannot set hittable breakpoints)
                agent.stepHandler.notifyException(the_event, false);
            }
        } else {
            if (debug) {
                Agent.message("Received exception: " +
                              the_event.exception + the_event);
            }
            Class excClass = the_event.exception.getClass();
            // caught if it or any superclass is caught - from GVH of SAS
            boolean caught = false;                        
            while (!caught && (excClass != null)) {  
                caught = catchHash.containsKey(excClass);  
                excClass = excClass.getSuperclass();       
            } 
                                             
            // Exception is uncaught if it falls through to MainThread or if
            // no class catches it at all.
            boolean uncaught = (the_event.catch_clazz == agent.mainCallerClass) 
                               || (the_event.catch_clazz == null);
            boolean handled = agent.stepHandler.notifyException(the_event,
                                                                caught || uncaught); 
            if (!handled && (caught || uncaught)) {
                agent.setSingleStep(the_event.thread, false);
                agent.suspendAllThreads();
                agent.reportAppException(the_event);
            }
        }
    }

    private void handleSingleStep() throws IOException {
/***                    GVH SAS # break sync
  if (%%% some test for single-step not active on this thread %%%) {       
  // A single step was cancelled  between
  // the time it was hit and the time it was reported to us here.
  Agent.message("Ignoring step notification received after step cancelled");
  continue;
  }
***/
        if (debug) {
            Agent.message("Single step event: " + the_event);
        }
        agent.setSingleStep(the_event.thread, false);
        agent.suspendAllThreads();
        // check if StepHandler has handled - GVH SAS # 4046670
        boolean handled = agent.stepHandler.notifyBreak(the_event); 
        if (!handled) {                       
            agent.reportBreakpoint(the_event);
        }
    }

    private void handleBreakpoint() throws IOException {
        if (debug) {
            Agent.message("Breakpoint event: " + the_event);
        }
        Breakpoint bkpt = Breakpoint.get(the_event.getMethod(), 
                                         the_event.bci);
        if (bkpt == null) {
            // The breakpoint was removed between the time it 
            // was hit and the time it was reported to us here.
            Agent.message("Ignoring breakpoint notification: ");
            Agent.message("  received after breakpoint cancelled.");
            Agent.message("  Breakpoint list: ");
            Enumeration bps = Breakpoint.allBreakpoints();
            while (bps.hasMoreElements()) {
                Agent.message("    " + bps.nextElement());
            }
            return;
        }
        if (bkpt.thread != null && bkpt.thread != the_event.thread) {
            // Not our thread
            return;
        }
        agent.setSingleStep(the_event.thread, false);
        agent.suspendAllThreads();
        if (bkpt.type == Breakpoint.ONESHOT) {
            Breakpoint.deleteBreakpoint(bkpt.getMethod(), bkpt.bci);
        }
        // check if StepHandler has handled - GVH SAS # 4046670
        boolean handled = agent.stepHandler.notifyBreak(the_event); 
        if (!handled) {                       
            agent.reportBreakpoint(the_event);
        }
    }

    public void run() {
	try {
            while (true) {
                synchronized (the_event) {
                    // Go to WAIT_FOR_EVENT state to signal that
                    // we're ready to receive
                    // an event.
                    the_event.waiting_for_event = true;
                    the_event.notifyAll();

                    // Wait for next breakpoint event.
                    if (!the_event.nextEvent()) {
                        return;
                    }
                    // Assert: waiting_for_event is now false
                    // (set by C code)
                }

		if (agent.asyncOutputStream == null) {
		    continue;
		}
                if (agent.systemThread(the_event.thread) == true) {
                    continue;
                }

                // ARB SAS 4046722
		if (agent.debuggerTerminating) {	
                    continue;
		}
                Agent.message("Handler - " + the_event);
                switch (the_event.kind) {
                    case Event.EXCEPTION:
                        handleException();
                        break;
                    case Event.SINGLE_STEP:
                        handleSingleStep();
                        break;
                    case Event.BREAKPOINT:
                        handleBreakpoint();
                        break;
                    case Event.THREAD_START:
                        Agent.message(the_event.toString());
                        break;
                    case Event.THREAD_END:
                        agent.reportThreadEnd(the_event);
                        break;
                    case Event.FRAME_POP:
                        Agent.message(the_event.toString());
                        // -- for testing, can pretend it is a breakpoint --
                        // the_event.kind = Event.BREAKPOINT;
                        // agent.suspendAllThreads();
                        // agent.reportBreakpoint(the_event);
                        break;
                    default:
                        Agent.message("Received unhandled event: " + 
                                          the_event);
		}
	    }
	} catch (ThreadDeath td) {
            if (debug) {
                Agent.message("Handler: ThreadDeath received.");
            }
	    /* Clear all breakpoints before exiting. */
	    Breakpoint.clearAllBreakpoints();
            synchronized (the_event) {
                the_event.waiting_for_event = true;
                the_event.notifyAll();
            }
	    throw td;
	} catch (Exception ex) {
	    Agent.error(agent.exceptionStackTrace(ex));
	    Breakpoint.clearAllBreakpoints();
            synchronized (the_event) {
                the_event.waiting_for_event = true;
                the_event.notifyAll();
            }
	    return;
	}
    }
}
