/*
 * @(#)ObjectCache.java	1.5 98/03/26
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.tools.agent;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;

class ObjectCache {
    private int sequenceNumber;
    private Hashtable objectKey;    // Maps Object to Integer key
    private Object[] objects;       // Maps int key to Object
    private int reclaimIndex;       // Where to start reclaiming
    private final int initSize = 20;
    private final int firstSeqNum = 1;
    private final Object reclaimed = new Object();

    /**
     * Constructor
     */
    public ObjectCache() {
        sequenceNumber = firstSeqNum;
        objectKey = new Hashtable();
        objects = new Object[initSize];
        reclaimIndex = -1;         // Mark as nothing to reclaim
    }

    /**
     * Get the object with the specified ID.
     */
    public Object getObject(int objNum) {
        if (objNum == AgentConstants.NULL_OBJECT_ID) {
            return null;
        }
        if ((objNum > sequenceNumber) || (objNum < 0)) {
            throw new IllegalArgumentException("objNum " + objNum + 
                                  " sequenceNumber " + sequenceNumber);
        }
        return objects[objNum];
    }

    /**
     * Add an object, return the ID.  Must be private, otherwise
     * duplicates could be created.
     */
    private int addObject(Object obj) {
        int index = -1;
        while (reclaimIndex != -1) {
            if (reclaimIndex > sequenceNumber) {
                reclaimIndex = -1;
            } else if (objects[reclaimIndex] != reclaimed) {
                ++reclaimIndex;
            } else {
                index = reclaimIndex;
                break;
            }
        }
        if (index == -1) {
            ++sequenceNumber;
            if (sequenceNumber >= objects.length) {
                // Grow the objects array
                Object[] temp = new Object[2 * objects.length];
                for (int inx = firstSeqNum; inx < objects.length; ++inx) {
                    temp[inx] = objects[inx];
                }
                objects = temp;
            }
            index = sequenceNumber;
        }
        Integer key = new Integer(index);
        objectKey.put(obj, key);
        objects[index] = obj;
        return index;
    }

    /**
     * Get the object with the specified ID.
     */
    public int getId(Object obj) {
        if (obj == null) {
            return AgentConstants.NULL_OBJECT_ID;
        }
        Object key = objectKey.get(obj);
        return key == null ? 
                       addObject(obj) : 
                       ((Integer)key).intValue();
    }

    /**
     * Return an enumeration over all the objects.
     */
    public Enumeration elements() {
        return new Enumeration() {
            int inx = firstSeqNum;

            public boolean hasMoreElements() { 
                while ((inx <= sequenceNumber) && (objects[inx] == reclaimed)) {
                    ++inx;
                }
                return inx <= sequenceNumber; 
            }

            public Object nextElement() { 
                return objects[inx++]; 
            }
        };
    }

    /**
     * Free an object.
     */
    public void free(int objNum) {
        Object obj = getObject(objNum);
        objectKey.remove(obj);
        objects[objNum] = reclaimed;
        if ((reclaimIndex == -1) || (reclaimIndex > objNum)) {
            reclaimIndex = objNum;
        }
    }

    /**
     * Free all object in the cache except classes and those listed.
     */
    public void freeAllBut(int[] rootKeep) {
        // Extend kept objects with all objects that the RemoteObjects can reach.
        HashSet keepSet = new HashSet();
        for (int i = 0; i < rootKeep.length; ++i) {
            addReferences(keepSet, objects[rootKeep[i]]);
        }

        // We must also retain any class that has a breakpoint set, so that the 
        // event handling mechanism will work. (The class can't be freed anyway)
        Breakpoint[] breaks = Breakpoint.listBreakpoints();
        for (int i = 0; i < breaks.length; i++) {
            addReferences(keepSet, breaks[i].getRealClass());
        }

        // Clone the existing objects hashtable.
        Hashtable old_objectKey = objectKey;
        Object[] old_objects = objects;
        objectKey = new Hashtable();
        objects = new Object[old_objects.length];
        // mark all object slots as reclaimable.  These will be 
        // overwritten with the objects we are keeping 
        for (int index = firstSeqNum; index < old_objects.length; ++index) {
            objects[index] = reclaimed;
        }

        HashSet keptCachedClasses = new HashSet();   

        // keep the user specified objects and the objects they reference
        Iterator iter = keepSet.iterator();
        while (iter.hasNext()) {
            Object obj = iter.next();

            Integer key = null;

            // Class objects appear as CachedClass objects in this cache
            if (obj instanceof Class) {
                obj = CachedClass.findCachedClass((Class)obj);

                // Retain any CachedClass version of this class
                if (obj != null) {
                    keptCachedClasses.add(obj);
                    key = (Integer)old_objectKey.get(obj);
                }
            } else {
                key = (Integer)old_objectKey.get(obj);
            }

            // RemoteClass info is lazily retrieved by the debugger side, so
            // it is possible that something in the keep set never got sent 
            // (thus is not cached). Ignore all such objects.
            if (key != null) {
                Agent.message("sending back " + obj + " with key " + key);
                int index = key.intValue();
                objectKey.put(obj, key);
                objects[index] = obj;
            }
        }

        // Sync the CachedClass static cache with ours
        CachedClass.freeAllBut(keptCachedClasses);

        // Be very paranoid about GC lags by clearing unneeded collections. Class
        // objects can wind up back in the objectCache very quickly; this code gives
        // them every chance to be collected.
        old_objectKey.clear();
        for (int i = 0; i < old_objects.length; i++) {
            old_objects[i] = null;
        }
        keptCachedClasses.clear();
        keepSet.clear();
        for (int i = 0; i < breaks.length; i++) {
            breaks[i] = null;
        }

        old_objectKey = null;
        old_objects = null;
        keptCachedClasses = null;
        keepSet = null;
        breaks = null;
        System.gc();
        reclaimIndex = firstSeqNum;
    }

    protected void addReferences(Set keepSet, Object obj) {
        if ( (obj != null) && keepSet.add(obj)) {

            // .add call above added the object reference
	    Agent.message("addref: " + obj);

	    /* 
	     * Debugger-side RemoteObjects keep a RemoteClass reference,
	     * so the object's class must remain in the keep set.
	     */
	    addReferences(keepSet, obj.getClass());

            if (obj instanceof Class) {
                Class clazz = (Class)obj;

                /*
                 * The Debugger-side RemoteClass holds RemoteClass/RemoteObject
                 * values for superclass, interfaces, class loader. If the
                 * class is to remain in the cache, all those values must stay too.
                 */
                addReferences(keepSet, clazz.getSuperclass());

                Class[] interfaces = clazz.getInterfaces();
                for (int i = 0; i < interfaces.length; i++) {
                    addReferences(keepSet, interfaces[i]);
                }

                addReferences(keepSet, clazz.getClassLoader());
            }
	}
    }
}

