/*
 * @(#)StepHandler.java	1.9 98/06/24
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * This code courtesy of Gordon Hirsch of SAS Institute.
 *
 * Converted to JVMDI
 */

package sun.tools.agent;

import java.io.IOException;

/**
 * The Step Handler thread takes different step reqeusts from the Agent thread
 * and processes them asynchronously.
 */
class StepHandler extends Thread implements StepConstants, AgentConstants {  

    // Opcodes for quick instructions. The opc_* constants for other opcodes
    // are defined in RuntimeConstants
    static private final int opc_invokevirtual_quick = 214;    
    static private final int opc_invokenonvirtual_quick = 215;    
    static private final int opc_invokesuper_quick = 216;    
    static private final int opc_invokestatic_quick = 217;    
    static private final int opc_invokeinterface_quick = 218;    
    static private final int opc_invokevirtualobject_quick = 219;    
    static private final int opc_invokevirtual_quick_w = 226;    
    static private final int opc_ldc_quick = 203;    
    static private final int opc_ldc_w_quick = 204;    
    static private final int opc_getfield_quick = 206;    
    static private final int opc_putfield_quick = 207;    
    static private final int opc_getfield2_quick = 208;    
    static private final int opc_putfield2_quick = 209;    
    static private final int opc_putstatic_quick = 211;    
    static private final int opc_getstatic2_quick = 212;    
    static private final int opc_putstatic2_quick = 213;    
    static private final int opc_new_quick = 221;    
    static private final int opc_anewarray_quick = 222;    
    static private final int opc_multianewarray_quick = 223;    
    static private final int opc_checkcast_quick = 224;    
    static private final int opc_instanceof_quick = 225;    
    static private final int opc_getfield_quick_w = 227;    
    static private final int opc_putfield_quick_w = 228;    

    private final int MIN_QUICK_OPCODE = opc_ldc_quick;
    private final int MAX_QUICK_OPCODE = opc_putfield_quick_w;

    private Agent agent;

    // Note that Event holds information for
    // just a single program event (breakpoint 
    // or exception). It is overwritten on 
    // subsequent program events. For that reason, 
    // we keep a clone of the last event to 
    // wake up this thread here, so that the info 
    // does not go away.
    private Event   event;   

    // State of the current step operations
    private boolean waitingForBreak = false;
    private boolean exceptionOccurred = false;
    private boolean userMustSeeException = false;
    private boolean hitUserBreak = false;
    private boolean stepping = false;

    StepHandler(Agent a) {
        super("Step handler");
        agent = a;
    }

    /**
     * Handle step requests from the Agent (called on the Agent's thread)
     */
    synchronized void stepThread(Thread t, boolean stepLine) {
        requestStep(stepLine ? STEP_LINE : STEP_INSTRUCTION, STEP_INTO, t);
    }

    synchronized void stepNextThread(Thread t) {
        requestStep(STEP_LINE, STEP_OVER, t);
    }

    synchronized void stepOutThread(Thread t) {   
        requestStep(STEP_INSTRUCTION, STEP_OUT, t); 
    }                                             

    private synchronized void requestStep(int granularity, int depth, Thread thread)
    {
        // If we're already stepping, interrupt the operation by resetting the 
        // stepping flag and waking up the Step Handler thread
        if (stepping) {
            stepping = false;
            notify();
        }

        StepRequest.make(granularity, depth, thread);
    }

    /**
     * Handle breakpoint/exception notifications from the breakpoint handler 
     * (called on the Breakpoint handler's thread. Exceptions and breaks are 
     * handled here only if the step handler is waiting for them (ie it is 
     * processing a step command)
     */
    synchronized boolean notifyBreak(Event anEvent) {
        if (waitingForBreak) {
            event = (Event)anEvent.clone();   // Make a safe copy of the one-deep "queue"
            notify();
            return true;
        } else {
            return false;
        }
    }

    synchronized boolean notifyException(Event anEvent, boolean userMustSee) {
        if (waitingForBreak) {
            event = (Event)anEvent.clone();   // Make a safe copy of the one-deep "queue"
            exceptionOccurred = true;
            userMustSeeException = userMustSee;

            // For breakpoints, the Handler has already suspended all threads.
            // For exceptions, we need to do it ourselves right here.
            agent.suspendAllThreads();

            notify();
            return true;
        } else {
            return false;
        }
    }

    /**
     * Main loop for the step handler thread
     */
    public void run() {
        while (true) {
            try {
                // Wait for the next step request
                StepRequest request = waitForRequest();

                // Process the request
         if (request.depth == STEP_OUT)  {          
             stepOut(request.thread);        
         } else if (request.granularity == STEP_LINE) {
                    stepLine(request.depth, request.thread);
                } else {
		    // STEP_INSTRUCTION granularity, STEP_INTO depth
                    stepInstruction(request.thread);
		}
            } catch (InterruptedException e) {
                // If a new step request comes in while the current one is being
                // processed an InterruptedException is thrown, getting us back to 
                // here. All we have to do then is eat the InterruptedException and
                // return to the top of the loop.
                Agent.message("Step interrupted");
            }
        }
    }

    private StepRequest waitForRequest() {
        stepping = false;
        StepRequest request = StepRequest.get();
        stepping = true;
        return request;
    }

    /**
     * Wait for a notification from the Handler Thread that we've hit
     * breakpoint (or completed a single step, or run into an exception)
     * On completion, the exception occurred boolean flag is set if there was an
     * exception. If the Handler has determined that the exception must
     * be reported to the debug user, the userMustSeeException flag is also set.
     */
    synchronized void waitForBreak() throws InterruptedException {
        exceptionOccurred = false;
        event = null;
        waitingForBreak = true;
        Agent.message("waiting for break...");
        wait();
        waitingForBreak = false;
        
        if (stepping == false) {
            throw new InterruptedException();
	}

    }



    // table of instructions that might possibly involve an invoke
    static private boolean invokeMap[] = new boolean[256];

    static {
        for (int i = 0; i < invokeMap.length; i++) {
            invokeMap[i] = false;
	}

        // These are the obvious ones: direct invokes
        invokeMap[opc_invokevirtual] = true;
        invokeMap[opc_invokespecial] = true;    
        invokeMap[opc_invokestatic] = true;    
        invokeMap[opc_invokeinterface] = true; 

        // These quick invokes which are used by the VM to optimize the 
        // instructions above
        invokeMap[opc_invokevirtual_quick] = true;     
        invokeMap[opc_invokenonvirtual_quick] = true;     
        invokeMap[opc_invokesuper_quick] = true;     
        invokeMap[opc_invokestatic_quick] = true;     
        invokeMap[opc_invokeinterface_quick] = true;     
        invokeMap[opc_invokevirtualobject_quick] = true;     
        invokeMap[opc_invokevirtual_quick_w] = true;     

        // The ldc, and ldc_w instructions may cause a string constructor to be 
        // invoked. (Note this is not true about the ldc2_w instruction)
        invokeMap[opc_ldc] = true;  
        invokeMap[opc_ldc_w] = true;

        // Since ldc, and ldc_w should cause a string constructor to be invoked
        // only the first time they are encountered, it isn't clear whether it
        // is necessary to include the quick versions of those instructions. 
        // It depends on whether the string constructor comes before the translation
        // to quick instruction. Since this may differ from VM to VM, the
        // quick variants are considered potential invokers here.
        invokeMap[opc_ldc_quick] = true;     
        invokeMap[opc_ldc_w_quick] = true;     

        // There are a number of instructions which can cause
        // classes to be loaded, and the class loading process can cause
        // Constructors and initializers in one or more classes to be invoked.
        invokeMap[opc_new] = true;    
        invokeMap[opc_anewarray] = true;    
        invokeMap[opc_multianewarray] = true;    

        invokeMap[opc_checkcast] = true;    
        invokeMap[opc_instanceof] = true;   

        invokeMap[opc_getstatic] = true;    
        invokeMap[opc_putstatic] = true;    

        // These two seem unlikely to trigger a class load, but they do 
        // reference a class. Just in case, they are included here.
        invokeMap[opc_getfield] = true;    
        invokeMap[opc_putfield] = true;    

        // Quick versions of all the class loading instructions above
        invokeMap[opc_getfield_quick] = true;     
        invokeMap[opc_putfield_quick] = true;     
        invokeMap[opc_getfield2_quick] = true;     
        invokeMap[opc_putfield2_quick] = true;     
        invokeMap[opc_putstatic_quick] = true;     
        invokeMap[opc_getstatic2_quick] = true;     
        invokeMap[opc_putstatic2_quick] = true;     
        invokeMap[opc_new_quick] = true;     
        invokeMap[opc_anewarray_quick] = true;     
        invokeMap[opc_multianewarray_quick] = true;     
        invokeMap[opc_checkcast_quick] = true;     
        invokeMap[opc_instanceof_quick] = true;     
        invokeMap[opc_getfield_quick_w] = true;     
        invokeMap[opc_putfield_quick_w] = true;     

        // Note that jsr (jump subroutine) is not included here. While it acts like
        // an invoke, Java uses it for finally clauses. We don't want to step over
        // the finally code because from a user's perspective, the finally clause
        // is not actually a subroutine.
    }

    // table of instructions that might possibly involve an branch
    static private boolean branchMap[] = new boolean[256];

    static {
        for (int i = 0; i < branchMap.length; i++) {
            branchMap[i] = false;
	}

        // Unconditional branches
        branchMap[opc_goto] = true;
        branchMap[opc_goto_w] = true;

        // Conditional branches
        branchMap[opc_ifeq] = true;
        branchMap[opc_ifnull] = true;
        branchMap[opc_iflt] = true;
        branchMap[opc_ifle] = true;
        branchMap[opc_ifne] = true;
        branchMap[opc_ifnonnull] = true;
        branchMap[opc_ifgt] = true;
        branchMap[opc_ifge] = true;
        branchMap[opc_if_icmpeq] = true;
        branchMap[opc_if_icmpne] = true;
        branchMap[opc_if_icmplt] = true;
        branchMap[opc_if_icmpgt] = true;
        branchMap[opc_if_icmple] = true;
        branchMap[opc_if_icmpge] = true;

        // Branch to and from finally clause
        branchMap[opc_jsr] = true;
        branchMap[opc_jsr_w] = true;
        branchMap[opc_ret] = true;

        // Switches
        branchMap[opc_lookupswitch] = true;
        branchMap[opc_tableswitch] = true;

        // Exception throw
        branchMap[opc_athrow] = true;

        // Returns
        branchMap[opc_ireturn] = true;
        branchMap[opc_lreturn] = true;
        branchMap[opc_freturn] = true;
        branchMap[opc_dreturn] = true;
        branchMap[opc_areturn] = true;
        branchMap[opc_return] = true;
    }


    private boolean isInvokeOpcode(int opcode)
    {
        return invokeMap[opcode];
    }

    private boolean isBranchOpcode(int opcode) {
        return branchMap[opcode];
    }

    private int instructionLength(int opcode) {
        if (opcode < MIN_QUICK_OPCODE) {
            // Get normal opcodes the easy way
            return opcLengths[opcode];
        } else if (opcode <= MAX_QUICK_OPCODE) {
            // There's no handy array of quick opcode lengths
	    switch (opcode) {
		case opc_ldc_quick: {    
		    return 2;
		}

		case opc_invokeinterface_quick: { 
		    return 5;
		}

		case opc_multianewarray_quick: {
		    return 4;
		}

		// all other _quicks are three bytes long
		default: {
		    return 3;
		}
	    }
        } else {
            throw new IllegalArgumentException("Invalid opcode");
        }
    }

    private int getReturnBCI(StackFrame frame) {
        // Normally, we are returning to an instruction 
        // which follows an invoke of some sort.
        // But under some circumstances (exceptions, 
        // class loading, etc.) we are returning to
        // something other than an invoke. So, we get 
        // the opcode here and try to figure out the 
        // return address.
        int opcode;
        try {
            int bci = frame.getBCI();
            if (bci == -1) {
                return -1;
            } else {
                opcode = frame.getMethod().byteAt(bci);
            }
        } catch(AbsentInformationException exc) {
            return -1;  // indicate failure
        }

        // At least for now, we won't decode branch 
        // instructions to figure out a return address
        // Returning to such an instruction should be 
        // very rare anyway, and most branches have more
        // than one potential "next" instruction.
        if (isBranchOpcode(opcode)) {
            return -1;
        } else {
	    // Safe to assume the next instruction follows 
            // this one in memory
            return frame.getBCI() + instructionLength(opcode);
	}
    }

    /**
     * Return all the frames in the current stack trace. 
     * This is the equivalent of
     * RemoteThread.dumpStack, but in the debuggee process.
     */
    private StackFrame[] getCurrentStackTrace(Thread t) {
        return StackFrame.getStackFrames(t);
    }


    private boolean frameEquals(StackFrame f1, StackFrame f2, boolean compareBCI) {
        boolean posMatch;
        if (compareBCI) {
            posMatch = f1.getBCI() == f2.getBCI();
        } else {
            int ln1 = f1.getLinenumber();
            int ln2 = f2.getLinenumber();
            if (ln1 == -1 || ln2 == -1) { // on failure
                posMatch = f1.getBCI() == f2.getBCI();
            } else {
                posMatch = ln1 == ln2;
            }
        }
        return posMatch && (f1.getMethod() == f2.getMethod());
    }

    /**
     * Compare two stack traces. Determine whether the second 
     * indicates a program location that is part of the execution
     * of the source line in the first. 
     */
    private boolean isExecutingLine(StackFrame[] startStack,
                                 StackFrame[] currentStack) {
        int offset = currentStack.length - startStack.length;

        if (offset < 0) {
            return false;
	}

        boolean compareBCI = false;               // Initially, don't need exact bci match
        for (int i=0; i<startStack.length; i++) {
            if (!frameEquals(startStack[i], currentStack[i+offset], compareBCI)) {
                Agent.message("stacks don't match at level " + i);
                return false;
            }

            // After comparing the lowest pair of frames, 
            // the bci's must begin to match exactly
            compareBCI = true;
        }

        return true;
    }

    private boolean isWithinLine(StackFrame[] startStack,
                                 StackFrame[] currentStack) {
        if (currentStack.length != startStack.length) {
            return false;
        } else {
            return isExecutingLine(startStack, currentStack);
	}
    }

    /**
     * Set a temporary break at the target PC and
     * execute the debuggee until it stops
     * for the first time in the target thread (this 
     * may or may not be at the temporary break) or 
     * at a breakpoint or user-caught exception in 
     * another thread.
     * Upon completion, the boolean hitUserBreak 
     * instance field will be true if a user 
     * defined breakpoint was hit. exceptionOccurred 
     * and userMustSeeException will 
     * be set appropriately if an exception occurred 
     * (see waitForBreak for details)
     */
    private StackFrame[] runOnceTo(CachedMethod targetMethod, int targetBCI, Thread targetThread) 
        throws NoSuchMethodException, InterruptedException, IOException
    {
        boolean userBreakWasSet;
        StackFrame[] currentStack = null;
        CachedMethod currentMethod;
        int currentBCI;

        hitUserBreak = false;

        // There really should be a Handler method for this check, 
        // but in the interest of minimizing our changes, I won't add one.
        userBreakWasSet = Breakpoint.exists(targetMethod, targetBCI);

        try {
            if (!userBreakWasSet) {
                Breakpoint bkpt = Breakpoint.addBreakpoint(targetMethod, 
                                                           targetBCI, 
                                                           Breakpoint.ONESHOT);
                bkpt.setThread(targetThread);
            } else {
                Agent.message("run-to breakpoint already there, not setting");
            }

            // Continue to wait while exceptions are reported from other threads as long as the 
            // exceptions are not to be reported to the user.
            do {
                agent.resumeLastSuspendedThreads();  // JDK 1.1.2E
                waitForBreak();

                if (exceptionOccurred && !userMustSeeException && (event.thread != targetThread))
                    Agent.message("run-to ignoring exception in another thread");

            } while (exceptionOccurred && !userMustSeeException && (event.thread != targetThread));
            currentStack = getCurrentStackTrace(event.thread);
            currentMethod = currentStack[0].getMethod();
            currentBCI = currentStack[0].getBCI();
        } finally {
            // If we set the break and it's still there, remove it now.
            if (!userBreakWasSet &&
		    Breakpoint.exists(targetMethod, targetBCI)) {
                Agent.message("removing run-to breakpoint");
                Breakpoint.deleteBreakpoint(targetMethod, targetBCI);
            }
        }

        // If there has not been an exception, we've stopped due to a breakpoint. If we didn't
        //  set the breakpoint or if the debuggee stopped somewhere other than our breakpoint,
        // then it must have been a user defined breakpoint. Set the instance field accordingly.
        hitUserBreak = !exceptionOccurred && 
            (userBreakWasSet || (currentBCI != targetBCI) || (currentMethod != targetMethod));

        if (hitUserBreak) {
            Agent.message("Hit a real user breakpoint.");
	}

        return currentStack;
    }

    /**
     * Run the debugee until one of the following happens:
     * 1) we get to the target address in the current thread and the 
     * current call stack is smaller than maxStackLength,
     * 2) some user-defined breakpoint is hit in any thread,
     * 3) an exception caught by the debugger user is hit in any thread
     * The instance flags exceptionOccurred, userMustSeeException, and userBreakWasHit
     * are set as described in waitForBreak() and runOnceTo().
     */
    private StackFrame[] runTo(CachedMethod targetMethod, int targetBCI, 
                               Thread targetThread, int targetStackLength)
		    throws NoSuchMethodException, InterruptedException, 
                                   IOException {
        StackFrame[] startStack = getCurrentStackTrace(targetThread);
        StackFrame[] currentStack = startStack;
        CachedMethod currentMethod;
        int currentBCI;

        // Continue to run until we get to the target address, some other breakpoint,
        // an exception caught by the debugger user, or an error. 
        boolean done = false;
        while (!done) {
            currentStack = runOnceTo(targetMethod, targetBCI, targetThread);
            boolean gotRecursiveException;
            do { // Repeat the following while we get recursive exceptions
                gotRecursiveException = false;
                currentMethod = currentStack[0].getMethod();
                currentBCI = currentStack[0].getBCI();
                Thread currentThread = event.thread;
                
                // For an exception in the target thread that we 
                // don't have to show the user, attempt to  
                // run until the catch clause, and then determine
                // if we're at a new line. 
                if (exceptionOccurred && !userMustSeeException) {
                    CachedMethod catchMethod = event.getCatchMethod();
                    int catchBCI = event.getCatchBCI();
                    Agent.message("exception (" + event + 
                                  ") at " + currentMethod.getName() + " (" + currentBCI +
                                  ") while running over, going to catch at " + 
                                  catchMethod.getName() + " (" + catchBCI + ")");
    
                    currentStack = runOnceTo(catchMethod, catchBCI, targetThread);
                    currentMethod = currentStack[0].getMethod();
                    currentBCI = currentStack[0].getBCI();
    
                    if ((currentBCI != catchBCI) || (currentMethod != catchMethod)) {
                        // We hit something before we could get to the catch
                        // clause. It might be yet another exception in which
                        // case we give up on getting to the previous catch
                        // address, and repeat what we just did with the new
                        // catch address. If it's not an exception, we are 
                        // lost and stop here. (Better error recovery to come)
                        // (Recursive exception handling courtesy of Alan 
                        // Beale of SAS Institute)
                        if (exceptionOccurred) {
                            Agent.message("Recursive exception");
                            gotRecursiveException = true;
                        } else {
                            Agent.message("couldn't get to catch bci");
                            done = true;
                        }
                    } else if (!isExecutingLine(startStack, currentStack) ||
                             startStack.length == currentStack.length) {
                        // We've reached the catch-bci and that address is
                        // 1) somewhere outside the line we were stepping over, 
                        // *or* 
                        // 2) it is later on that very line (but likely someplace past
                        // breakAddress. We've had an exception, so we won't ever get to 
                        // breakAddress now).
                        //
                        // In either case, we want to stop the loop now. 
                        Agent.message("catch bci is not within line, stopping");
                        done = true;
                    }
                } else if (exceptionOccurred || hitUserBreak) {
                    // We've been interrupted by an exception that the user wants to see
                    // or a user breakpoint, so stop
                    done = true;
                } else {
    		// Hit our own break 
    
                    // We want to stop on our own break only if it's at the right level
                    // in the call stack. If the call stack is too large, 
                    // the debuggee is recursing and we've hit the breakpoint within 
                    // recursion.
                    if (currentStack.length <= targetStackLength) {
                        done = true;
    		} else {
                        Agent.message("debuggee is recursing. continuing...");
    		}
                }
    
            } while (gotRecursiveException);
        } 

        return currentStack;
    }

    /**
     * If enough information is available, run the application until it returns from the current
     * java method into the closest calling java method (native methods are skipped).
     */
    private boolean runToReturn(StackFrame[] stack, Thread t)  
        throws NoSuchMethodException, InterruptedException, IOException
    {
        boolean ran = false;

        if (stack.length > 1) {
            StackFrame returnFrame = stack[1];
            int returnBCI = getReturnBCI(returnFrame);
            if (returnBCI != -1) {
                for (;;) {
                    StackFrame[] currentStack = runTo(returnFrame.getMethod(), 
                                                      returnBCI, t, stack.length-1);
                    if (exceptionOccurred && userMustSeeException) {
                        break;  // interrupted by an important exception
                    } else if (hitUserBreak) {
                        break;  // interrupted by a user-set breakpoint
                    } else if (currentStack.length < stack.length) {
                        break;  // made back to the caller
                    }
                }
                ran = true;
            }
        }

        return ran;
    }

    /**
     * Notify the remote debugger that we have stopped. If we are on an 
     * exception about which the user must be informed, notify of that; 
     * otherwise notify that a breakpoint has been hit. 
     *
     * It is possible during instruction stepping that we will have a 
     * breakpoint (end of step) and an exception at the same time. Even 
     * in that case we notify of only one event depending solely on 
     * whether the userMustSee flag is set.
     */
    private void notifyRemoteDebugger() throws IOException
    {
        if (exceptionOccurred && userMustSeeException) {
            agent.reportAppException(event);
        } else {
	    // We ended on a regular breakpoint
            agent.reportBreakpoint(event);
        }
    }

    /**
     * Internal version of step instruction. Does 
     * the step, but it is left to the caller
     * to notify the remote debugger. Caller must
     * make sure we are not in a system thread.
     */
    synchronized private void internalStepInstruction(Thread t) 
        throws InterruptedException {
	    try {
		hitUserBreak = false;

		// Continue to resume the appl until the step 
                // concludes, a break is hit, or 
		// an exception that the user must see is hit. 
		do {
		    agent.setSingleStep(t, true);
		    agent.resumeLastSuspendedThreads();
		    waitForBreak();
		} while (exceptionOccurred && !userMustSeeException); 

		// It's possible that the step will conclude on a user breakpoint. 
		if (Breakpoint.exists(event.getMethod(), event.bci)) {
		    hitUserBreak = true;
		}

	    } catch (InterruptedException e) {
		// Cancel the current single step. Depending on the state of the 
		// interpreter, this may be ignored, but if so, when the step 
		// concludes, the fact that this flag was reset will tell us to
		// ignore the corresponding notification.
		agent.setSingleStep(t, false);
		throw e;
	    } catch (IllegalThreadStateException e) {
		Agent.error("step failed: " + e.getMessage());
	    }
    }

    /**
     * Step into or over a line
     */
    synchronized private void stepLine(int depth, Thread t) throws InterruptedException {
	if (!agent.systemThread(t)) {
	    Agent.message("next stepping " + t.getName());
	    try {
                StackFrame[] startStack = getCurrentStackTrace(t);

                if (startStack[0].getLinenumber() == -1) {
                    Agent.message("no line information available, single-stepping.");
                    internalStepInstruction(t);
		    notifyRemoteDebugger();
                    return;
                }

                // Loop until we reach another source line in the debuggee
		int opcode;
                boolean stepInterrupted;
                boolean stepComplete;
                StackFrame[] currentStack = startStack;
                CachedMethod currentMethod = currentStack[0].getMethod();
                int currentBCI = currentStack[0].getBCI();
                do {
                    try {
                        opcode = currentMethod.byteAt(currentBCI);
                    } catch (AbsentInformationException exc) {
                        Agent.message(
                            "byte-code information available, single-stepping.");
                        internalStepInstruction(t);
                        notifyRemoteDebugger();
                        return;
                    }
                       
                    // Decide whether to run past the current instruction or just step
                    // into it. If there's a chance that the instruction might invoke a 
                    // method, we choose the former (more complicated) option.
                    if ( isInvokeOpcode(opcode) && (depth == STEP_OVER) ) {
                        // Get the start of the next instruction so we can break on it
                        int breakBCI = currentBCI + instructionLength(opcode);
                        currentStack = runTo(currentMethod, breakBCI, t, startStack.length);
                        currentMethod = currentStack[0].getMethod();
                        currentBCI = currentStack[0].getBCI();
                    } else {
			// no danger of an invoke; step into the instruction
			internalStepInstruction(t);

			currentStack = getCurrentStackTrace(event.thread);
			currentMethod = currentStack[0].getMethod();
			currentBCI = currentStack[0].getBCI();

			// If we've accidentally stepped into a method, run until it returns.
			// This will happen if the instruction triggered an exception or if
			// there is an invoke implicit in the instruction that we don't know about.
			// (But we only do this if we're still in the same thread and if neither a 
			//  breakpoint nor user-caught exception have occurred.)
			if ( (depth == STEP_OVER) && 
			     (startStack.length < currentStack.length) &&
			     (t == event.thread) &&
			     !hitUserBreak && !(exceptionOccurred && userMustSeeException) )
			{
			    boolean returned = runToReturn(currentStack, t);
			    if (returned) {
				currentStack = getCurrentStackTrace(event.thread);
				currentMethod = currentStack[0].getMethod();
				currentBCI = currentStack[0].getBCI();
			    } else {
				// fall through. We can't find a reasonable address to run to. 
				// Since we could be here due to an exception or an unexpected invoke
				// in the normal processing of an instruction, we can't assume we will
				// be awakened by an exception if we just resume the application. So, 
				// we take the conservative approach and just stop here.
				Agent.message("unexpectedly stepped into a method. Can't find a return address"); 
				break;
			    }
                        }
                    }

                    // Determine loop conditions
                    if (depth == STEP_OVER) {
                        stepComplete = !isExecutingLine(startStack, currentStack);
		    }
                    else {
			// STEP_INSTRUCTION
                        stepComplete = !isWithinLine(startStack, currentStack);
		    }
                    stepInterrupted = (exceptionOccurred && userMustSeeException) || hitUserBreak;
                } while (!stepComplete && !stepInterrupted);

                notifyRemoteDebugger();

	    } catch (IllegalThreadStateException e) {
		Agent.error("next step failed: " + e.getMessage());
	    } catch (NoSuchMethodException e) {
		Agent.error("next step failed: " + e.getMessage());
	    } catch (IOException e) {
		Agent.error("next step failed: " + e.getMessage());
	    }
	}
    }

    /**
     * Step into an instruction
     */
    synchronized private void stepInstruction(Thread t) throws InterruptedException {
	if (!agent.systemThread(t)) {
	    Agent.message("stepping " + t.getName());
	    try {
		    internalStepInstruction(t);
		    notifyRemoteDebugger();
	    } catch (IOException e) {            
		    Agent.error("step failed: " + e.getMessage()); 
	    }
	}
    }

    /* Step out of a java method                                     
     * GVH SAS step out
     */
    synchronized private void stepOut(Thread t) throws InterruptedException {
	if (!agent.systemThread(t)) {                   
	    Agent.message("stepping out " + t.getName());             
	    try {                                                         
		StackFrame[] startStack = getCurrentStackTrace(t);    
		boolean ranToReturn = runToReturn(startStack, t);     
		if (! ranToReturn) {         
		    Agent.message("Unable to step out; single stepping");
		    hitUserBreak = false;                             
		    internalStepInstruction(t);                   
		}
		notifyRemoteDebugger();
	    } catch (IllegalThreadStateException e) { 
		    Agent.error("step out failed: " + e.getMessage());    
	    } catch (NoSuchMethodException e) {                        
		    Agent.error("step out failed: " + e.getMessage());    
	    } catch (IOException e) {              
		    Agent.error("step out failed: " + e.getMessage());    
	    }                                                         
	}                                                             
    }                                                             
}                                                                 

/**
 * Constants that describe the type of step being requested
 */
interface StepConstants {
    final int STEP_INSTRUCTION = 0;
    final int STEP_LINE        = 1;

    final int STEP_OVER        = 0;
    final int STEP_INTO        = 1;
    final int STEP_OUT         = 2;     
}

/**
 * StepRequest objects hold the data describing a requested step operation.
 * The StepRequest class is also used in synchronizing the sending and receiving
 * of step requests by the Agent and StepHandler threads respectively.
 *  
 * Note that step requests are not queued. New requests cancel previous ones. 
 * There is at most one referenced instance of StepRequest and it is held in
 * the static field called "instance". 
 */
class StepRequest implements StepConstants {
    static StepRequest instance = null;
    static Class thisClass = StepRequest.class;

    int granularity;       // Granularity: line or instruction
    int depth;             // Depth: step into or step over
    Thread thread;         // Target thread for step.

    /**
     *  Get the next step request, waiting if necessary
     */
    public static synchronized StepRequest get() {
        // If there's no request here, wait for one using the class object's monitor.
        while (instance == null) {
            try {
                Agent.message("Awaiting new step request");
                thisClass.wait();
            } catch (InterruptedException e) {
                Agent.message("Step Handler thread ignoring interrupt exception");
            }
        }

        StepRequest ret = instance;
        instance = null;
        return ret;
    }

    /**
     * Create a new step request, wiping out any previously requested step 
     * that has not yet been processed. Wake up the StepHandler if it is 
     * waiting for a request.
     */
    public static synchronized void make(int granularity, int depth, Thread thread) {
        Agent.message("Making new step request");
        instance = new StepRequest(granularity, depth, thread);
        thisClass.notify();
    }

    private StepRequest(int g, int d, Thread t) {
        granularity = g;
        depth = d;
        thread = t;
    }

}


    

