/*
 * @(#)RemoteAgent.java	1.67 98/05/27
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.tools.debug;

import java.io.*;
import java.net.*;
import java.util.*;
import sun.tools.agent.AgentConstants;


class RemoteAgent implements AgentConstants {
    /* The agent socket */
    private Socket socket;

    /* The input stream from the remote agent */
    private DataInputStream in;

    /* The output stream to the remote agent */
    private DataOutputStream out;

    /* Hashtable of active objects */
    private Hashtable objects = new Hashtable();

    /* The DebuggerCallback object. */
    private DebuggerCallback client;

    /* The client side of the Agent's stdout redirector */
    private AgentIn debugIn = null;
    private Thread debugInThread = null;

    private RemoteClass classObject;
    RemoteClass classClass;
    RemoteClass classString;
    boolean closeRemoteInterpreter = false;

    private RemoteThreadGroup systemThreadGroup;

    private boolean verbose = false;
    private boolean closing = false;
    private boolean closed = false;

    private int readPassword(String s) {
        int cookie = 0;
        int radix = PSWDCHARS.length();
        for (int i = 0; i < s.length(); i++) {
            int x = PSWDCHARS.indexOf(s.charAt(i));
            if (x == -1) {
                return 0;	// bad password
            }
            cookie = (cookie * radix) + x;
        }
        int port = 0;
        for (int i = 0; i < 8; i++) {
            int bits = (cookie & (6 << (i * 3)));
            port |= (bits >> (i + 1));
        }
        return port;
    }

    private void initSession(boolean verbose) throws Exception {

	/* Check that agent is the same version */
	int version = in.readInt();
	if (version != AGENT_VERSION) {
	    System.err.println(
		"Version mismatch between debugger and remote agent.");
	    System.exit(1);
	}

	/* Load current Agent classes */
	classObject = (RemoteClass)readValue(in);
	classClass = (RemoteClass)readValue(in);
	classString = (RemoteClass)readValue(in);
	message("loading classes ...");
	int nClasses = in.readInt();
	for (int i = 0 ; i < nClasses ; i++) {
	    RemoteClass c = (RemoteClass)readValue(in);
	}

	systemThreadGroup = (RemoteThreadGroup)readValue(in);

	setVerbose(verbose);
        closing = closed = false;
    }

    /**	
     *	Create a Thread that will retrieve and display any non-Java output.
     *	Needs to be high priority, else debugger may exit before
     *	it can be displayed.
     */
    private void displayNativeOutput(final BufferedReader in) {
	Thread thr = new Thread("output reader") { 
	    public void run() {
		try {
		    String line;
		    while ((line = in.readLine()) != null) {
                        try {
                            printToClientConsole(line + "\n");
                        } catch (Exception e) {
                            System.out.println(line);
                        }
		    }
		} catch (IOException ex) {
		    error("Failed reading output of child java interpreter.");
		}
	    }
	};
	thr.setPriority(Thread.MAX_PRIORITY-1);
	thr.start();
    }

    /**
     * Add a -classpath argument to the arguments passed to the exec'ed
     * VM with the contents of CLASSPATH environment variable,
     * if -classpath was not already specified.
     *
     * Perhaps this should be a protected instance method.
     *
     * @param javaArgs the arguments to the VM being exec'd that
     *                 potentially has a user specified -classpath argument.
     * @return a javaArgs whose -classpath option has been added
     */
    private static String processClasspath(String javaArgs) {
	
    String envcp = System.getProperty("env.class.path");
    if ((envcp != null) && (javaArgs.indexOf("-classpath ") == -1)) {
        StringBuffer munged = new StringBuffer(javaArgs);
        munged.append(" -classpath ");
        munged.append(envcp);
        return munged.toString();
    } else {
        return javaArgs;
    }
    }

    /**
     * We need to automatically add tools.jar to the exec'd VM bootclasspath
     * because the agent's classes are in tools.jar.  The agent classes 
     * must be in the boot class path to avoid frequent deadlocks 
     * during class loading of application classes. As long as there
     * is an application.home property defined and any existing 
     * -Xbootclasspath is not obviously in error, this method will 
     * return an argument string with a -Xbootclasspath option.
     *
     * If jdb was launched with a -Xbootclasspath argument, say:
     *       $ jdb -Xbootclasspath /path/to/classes ClassName
     * then the result will contain
     *       -Xbootclasspath jdk/lib/tools.jar:/path/to/classes
     * If jdb was launched with -Xbootclasspath/a ..., the "/a" will be
     * preserved.
     *
     * If jdb was not launched with -Xbootclasspath, 
     * then the result will contain
     *       -Xbootclasspath/a jdk/lib/tools.jar
     *
     * Perhaps this should be a protected instance method.
     *
     * @param javaArgs the arguments to the VM being exec'd that
     *                 potentially has a user specified -Xbootclasspath argument.
     * @return a javaArgs whose -Xbootclasspath option has been munged
     */
    private static String processBootClasspath(String javaArgs) {
        String appHome =  System.getProperty("application.home");
        if (appHome == null) {
            /* No application.home means we can't insert tools.jar */
            return javaArgs;
        }
                                         
        /* Where is tools.jar? */
        String toolsJar =  appHome + File.separator +
              "lib" + File.separator + "tools.jar";
        
        /* Extract any -Xbootclasspath from the java args */
        StringBuffer munged = new StringBuffer(javaArgs);
        String bootCpOption = new String("-Xbootclasspath/a");
        String cmdBootCp = null;
        int bcpStart = javaArgs.indexOf("-Xbootclasspath");
        if (bcpStart != -1) {
            int bcpEnd = javaArgs.indexOf(':', bcpStart);
            if (bcpEnd != -1) {
                int bcpArgStart = bcpEnd + 1;
                int bcpArgEnd = javaArgs.indexOf(' ', bcpArgStart);
                if (bcpArgEnd == -1)
                    bcpArgEnd = javaArgs.length();
                bootCpOption = javaArgs.substring(bcpStart, bcpEnd);
                cmdBootCp = javaArgs.substring(bcpArgStart, bcpArgEnd);
                munged = new StringBuffer(javaArgs.substring(0, bcpStart) +
                                          javaArgs.substring(bcpArgEnd));
            } else {
                /* malformed -Xbootclasspath. Don't mess with it, allow exec() to fail */
                return javaArgs;
            }
        }
    
        /* Append the new -Xbootclasspath option */
        munged.append(" ");
        munged.append(bootCpOption);
        munged.append(":");
        munged.append(toolsJar);
    
        if (cmdBootCp != null) {
            munged.append(File.pathSeparator + cmdBootCp);
        }
        munged.append(" ");
    
        return munged.toString();
    }

    /* Connect to a remote agent using a socket */
    RemoteAgent(String host, String password, String javaArgs,
                DebuggerCallback client, boolean verbose)  throws Exception {
	this.verbose = verbose;
        if (host.equals("127.0.0.1") && password == null) {
            // No remote (or local) interpreter to debug, so spawn one.
            String slash = File.separator;
            javaArgs = processClasspath(javaArgs);
            javaArgs = processBootClasspath(javaArgs);
            String cmd = new String(System.getProperty("java.home") +
                                    slash + "bin" + slash +
                                    "java -Xdebug -Djava.compiler=NONE " +
                                    javaArgs + " sun.tools.agent.EmptyApp");
            message("starting child: " + cmd);
            Process pr = Runtime.getRuntime().exec(cmd);
            BufferedReader in
	       = new BufferedReader(new InputStreamReader(pr.getInputStream()));
            BufferedReader err
	       = new BufferedReader(new InputStreamReader(pr.getErrorStream()));

            String agentStatus = in.readLine();
            if (agentStatus == null) {
                error("Failed to exec a child java interpreter with this command line: " + cmd);
                System.exit(1);
            }
            password = agentStatus.substring(agentStatus.indexOf("=") + 1);
            message("password returned: " + password);
            closeRemoteInterpreter = true;
            displayNativeOutput(in);
            displayNativeOutput(err);
        }

        this.client = client;
	socket = new Socket(host, readPassword(password));

        DataInputStream asyncIn = new DataInputStream(
            new BufferedInputStream(
                new Socket(host, getDebugPort()).getInputStream()));
	debugIn = new AgentIn(this, asyncIn, client, verbose);
	debugInThread = new Thread(debugIn, "Agent input");
	debugInThread.start();

        in = new DataInputStream(new
		BufferedInputStream(socket.getInputStream()));
	out = new DataOutputStream(new
		BufferedOutputStream(socket.getOutputStream()));
	message("connected");

        this.initSession(verbose);
    }

    /* Connect to the local agent thread, using pipes. */
    /**
     * This makes no sense unless agent is running in the same VM
     * as the remote agent.  We don't support this.  So this is
     * now commented out.
**************
    RemoteAgent(DebuggerCallback client, boolean verbose)  throws Exception {
        if (!Agent.isDebuggable()) {
            throw new IllegalAccessError();
        }
	this.verbose = verbose;
        this.client = client;

        PipedInputStream cmdInputPipe = new PipedInputStream();
        PipedOutputStream cmdOutputPipe = new PipedOutputStream();
        PipedInputStream asyncInputPipe = new PipedInputStream();
        Agent.connectToAgent(cmdOutputPipe, cmdInputPipe, asyncInputPipe);

        DataInputStream asyncIn = new DataInputStream(new 
                BufferedInputStream(asyncInputPipe));
	debugIn = new AgentIn(this, asyncIn, client, verbose);
	debugInThread = new Thread(debugIn, "Agent input");
	debugInThread.start();

	in = new DataInputStream(cmdInputPipe);
	out = new DataOutputStream(cmdOutputPipe);
	message("connected");

        Agent.addSystemThread(Thread.currentThread());
        initSession(verbose);
    }
************/

    /* Return the socket port being used. */
    int getDebugPort() {
	return socket.getPort();
    }

    /* Read in a remote source file. */
    InputStream getSourceFile(String clsName, String name) throws Exception {
        synchronized (out) {
            out.write(CMD_GET_SOURCE_FILE);
            out.writeUTF(name);
            out.writeUTF(clsName);
            try {
                getReply(CMD_GET_SOURCE_FILE);
            } catch (Exception e) {
                return null;
            }

            int length = in.readInt();
            if (length == -1) {
                return null;
            }
            message("getSourceFile: allocating " + length + " bytes.");
            byte buffer[] = new byte[length];
            try {
                in.readFully(buffer);
            } catch (IOException e) {
                error("unable to read " + name);
                return null;
            }
            return new ByteArrayInputStream(buffer);
        }
    }
    
    /* Ask a remote object for a string description of itself. */
    String objectToString(int id)  {
        synchronized (out) {
            try {
                out.write(CMD_OBJECT_TO_STRING);
                out.writeInt(id);
                getReply(CMD_OBJECT_TO_STRING);
                return in.readUTF();
            } catch (Exception ex) {
                return("<communication error>");
            }
        }
    }

    /* Close the connection */
    void close() {
	message("close(" + closeRemoteInterpreter + ")");
	closing = true;
        try {
            if (closeRemoteInterpreter) {
                debugIn.quit();   // Enable silent termination of agent input thread
                synchronized (out) {
                    out.write(CMD_QUIT);
                    getReply(CMD_QUIT);
                }
            } else {
                // If the other VM is not shutting down, make it let go of 
                //  its cached objects so they can be GC'ed.
                gc(new RemoteObject[0]);
                debugIn.quit();   // Enable silent termination of agent input thread
            }
            if (socket != null) {
                socket.close();
            }
        } catch (Exception e) {}
        closed = true;
    }

    /*
     * React to the closing of the communication channel from external
     * events (such as the abrupt termination of the remote VM) by
     * doing a minimal close.
     */
    void handleRemoteClose() {
	closing = true;
        try {
            /*
             * Unlike close, we don't kill the AgentIn thread because it
             * is repsonsible for calling this method and may need more cleanup
             */
            if (socket != null) {
                socket.close();
            }
        } catch (Exception e) {}
        closed = true;
    }

    /* Get an object from the cache */
    RemoteObject get(Integer h) {
	return (RemoteObject)objects.get(h);
    }

    /* Put an object in the cache */
    private void put(Integer h, RemoteObject obj) {
	objects.put(h, obj);
    }

    /* Read a value */
    RemoteValue readValue(DataInputStream in) throws Exception {
	while(true) {
	    int type = in.readInt();
	    switch (type) {
	      case TP_OBJECT: {
		  int id = in.readInt();
// message("read - obj" + id);
		  if (id == 0) {
		      return null;
		  }

		  Integer h = new Integer(id);
		  RemoteClass c = (RemoteClass)get(h);
		  if (c == null) {
		      put(h, c = new RemoteClass(this, id));
		  }

		  id = in.readInt();
		  h = new Integer(id);
		  RemoteObject obj = get(h);
		  if (obj == null) {
		      put(h, obj = new RemoteObject(this, id, c));
		  }
		  return obj;
	      }

	      case TP_CLASS: {
		  int id = in.readInt();
// message("read - class" + id);
		  Integer h = new Integer(id);
		  RemoteClass c = (RemoteClass)get(h);
		  if (c == null) {
		      put(h, c = new RemoteClass(this, id));
		  }
		  return c;
	      }

	      case TP_THREAD: {
		  int id = in.readInt();
		  Integer h = new Integer(id);
		  RemoteClass c = (RemoteClass)get(h);
		  if (c == null) {
		      put(h, c = new RemoteClass(this, id));
		  }

		  id = in.readInt();
		  h = new Integer(id);
		  RemoteThread t = (RemoteThread)get(h);
		  if (t == null) {
		      put(h, t = new RemoteThread(this, id, c));
		  }
		  return t;
	      }

	      case TP_THREADGROUP: {
		  int id = in.readInt();
		  Integer h = new Integer(id);
		  RemoteClass c = (RemoteClass)get(h);
		  if (c == null) {
		      put(h, c = new RemoteClass(this, id));
		  }

		  id = in.readInt();
		  h = new Integer(id);
		  RemoteThreadGroup tg = (RemoteThreadGroup)get(h);
		  if (tg == null) {
		      put(h, tg = new RemoteThreadGroup(this, id, c));
		  }
		  return tg;
	      }

	      case TP_STRING: {
		  int id = in.readInt();
		  Integer h = new Integer(id);
		  RemoteString s = (RemoteString)get(h);
		  if (s == null) {
		      put(h, s = new RemoteString(this, id));
		  }
		  return s;
	      }

	      case TC_ARRAY: {
		  int id = in.readInt();
		  if (id == 0) {
		      return null;
		  }
		  Integer h = new Integer(id);
		  RemoteClass c = (RemoteClass)get(h);
		  if (c == null) {
		      put(h, c = new RemoteClass(this, id));
		  }
		  id = in.readInt();
		  h = new Integer(id);
		  int size = in.readInt();
		  RemoteArray array = (RemoteArray)get(h);
		  if (array == null) {
		      put(h, array = new RemoteArray(this, id, c, size));
		  }
		  return array;
	      }

	      case TC_BOOLEAN:
		  return (RemoteValue)(new RemoteBoolean(in.readBoolean()));

	      case TC_BYTE:
		  return (RemoteValue)(new RemoteByte(in.readByte()));

	      case TC_CHAR:
		  return (RemoteValue)(new RemoteChar(in.readChar()));

	      case TC_SHORT:
		  return (RemoteValue)(new RemoteShort(in.readShort()));

	      case TC_INT:
		  return (RemoteValue)(new RemoteInt(in.readInt()));

	      case TC_LONG:
		  return (RemoteValue)(new RemoteLong(in.readLong()));

	      case TC_FLOAT:
		  return (RemoteValue)(new RemoteFloat(in.readFloat()));

	      case TC_DOUBLE:
		  return (RemoteValue)(new RemoteDouble(in.readDouble()));

	      default: {
		  error("invalid type code: " + type);
		  Thread.dumpStack();
		  return null;
	      }
	    }
	}
    }

    /* get a reply */
    private void getReply(int cmd) throws Exception {
        if (closed) {
            throw new NoSessionException();
        }
	out.flush();

	while (true) {
	    int c = 0;
	    try {
		c = in.readInt();
	    } catch (EOFException e) {
		if (!closing) {
		    error("unexpected eof");
		}
		return;
	    } catch (IOException ioe) {
		if (closing) {
                    return;
		}
                throw ioe;
	    }

	    if (c == CMD_EXCEPTION) {
		Exception e = new Exception(in.readUTF());
		throw e;
	    } else if (c != cmd) {
		error("unexpected reply: wanted " + cmd + ", got " + c);
	    } else {
		return;
	    }
	}
    }

    /* List the currently known classes. */
    RemoteClass[] listClasses() throws Exception  {
        synchronized (out) {
            out.write(CMD_LIST_CLASSES);
            getReply(CMD_LIST_CLASSES);
            int n = in.readInt();
            message("reading " + n + " classes...");
            RemoteClass list[] = new RemoteClass[n];

            for (int i = 0 ; i < n ; i++) {
                list[i] = (RemoteClass)readValue(in);
            }
            return list;
        }
    }

    /* Find a specified class. */
    RemoteClass findClass(String name) throws Exception {
	RemoteClass list[] = listClasses();
	for (int i = 0; i < list.length; i++) {
	    String className = list[i].getName();

	    /* Try full name (ie "java.lang.Object") */
	    if (name.equals(className)) {
		return list[i];
	    }

	    /* Next, try short name (ie "Object") */
	    int iDot = className.lastIndexOf(".");
	    if (iDot >= 0 && className.length() > iDot) {
		className = className.substring(iDot + 1);
	    }
	    if (name.equals(className)) {
		return list[i];
	    }
	}
	return getClassByName(name);
    }

    /* List threadgroups */
    RemoteThreadGroup[] listThreadGroups(RemoteThreadGroup tg) 
      throws Exception {
	RemoteThreadGroup list[] = null;
        synchronized (out) {
            out.write(CMD_LIST_THREADGROUPS);
            out.writeInt((tg == null) ? 0 : tg.getId());

            getReply(CMD_LIST_THREADGROUPS);
            int n = in.readInt();
            list = new RemoteThreadGroup[n];
            message("listThreadGroups: " + n + " groups");
            for (int i = 0 ; i < n ; i++) {
                list[i] = (RemoteThreadGroup)readValue(in);
            }
        }
	return list;
    }

    /* List a threadgroup's threads */
    RemoteThread[] listThreads(RemoteThreadGroup tg,
                               boolean recurse) throws Exception {
        synchronized (out) {
            out.write(CMD_LIST_THREADS);
            out.writeInt((tg == null) ? systemThreadGroup.getId() : tg.getId());
            out.writeBoolean(recurse);

            getReply(CMD_LIST_THREADS);
            int n = in.readInt();
            RemoteThread list[] = new RemoteThread[n];

            for (int i = 0 ; i < n ; i++) {
                list[i] = (RemoteThread)readValue(in);
            }
            return list;
        }
    }

    /* Suspend a specified thread. */
    void suspendThread(int id) throws Exception {
        synchronized (out) {
            out.write(CMD_SUSPEND_THREAD);
            out.writeInt(id);
            getReply(CMD_SUSPEND_THREAD);
        }
    }

    /* Resume a specified thread. */
    void resumeThread(int id) throws Exception {
        synchronized (out) {
            out.write(CMD_RESUME_THREAD);
            out.writeInt(id);
            getReply(CMD_RESUME_THREAD);
        }
    }

    /* Resume all of the last suspended threads. */
    void resumeLastSuspendedThreads() throws Exception {
        synchronized (out) {
            out.write(CMD_RESUME_LAST_SUSPENDED_THREADS);
            getReply(CMD_RESUME_LAST_SUSPENDED_THREADS);
        }
    }

    /* Suspend all of the non-system threads. */        // GVH SAS # 4046670
    void suspendAllThreads() throws Exception {         // GVH SAS # 4046670
        synchronized (out) {                            // GVH SAS # 4046670
            out.write(CMD_SUSPEND_ALL_THREADS);         // GVH SAS # 4046670
            getReply(CMD_SUSPEND_ALL_THREADS);          // GVH SAS # 4046670
        }                                               // GVH SAS # 4046670
    }                                                   // GVH SAS # 4046670
                                                        // GVH SAS # 4046670
    /* Dump a thread's stack. */
    RemoteStackFrame[] dumpStack(RemoteThread thread) throws Exception {
        synchronized (out) {
            out.write(CMD_GET_STACK_TRACE);
            out.writeInt(thread.getId());

            getReply(CMD_GET_STACK_TRACE);
            int n = in.readInt();
            RemoteStackFrame list[] = new RemoteStackFrame[n];
            for (int i = 0 ; i < n ; i++) {
                RemoteObject remoteFrame = (RemoteObject)readValue(in);
                list[i] = new RemoteStackFrame(remoteFrame, thread, i, this);
                list[i].className = in.readUTF();
                list[i].methodName = in.readUTF();
                list[i].methodSignature = in.readUTF();          // GVH SAS method signature
                list[i].lineno = in.readInt();
                list[i].bci = in.readInt();
                list[i].setRemoteClass((RemoteClass)readValue(in));

                int nVars = in.readInt();
                list[i].localVariables = new LocalVariable[nVars];
                for (int j = 0; j < nVars; j++) {
                    LocalVariable lvar = new LocalVariable();
                    lvar.slot = in.readInt();
                    lvar.name = in.readUTF();
                    lvar.signature = in.readUTF();
                    lvar.methodArgument = in.readBoolean();
                    message("lvar " + j + ": slot=" + lvar.slot +
                            ", name=" + lvar.name + ", sig=" + 
                            lvar.signature + ", arg=" + lvar.methodArgument);
                    list[i].localVariables[j] = lvar;
                }
            }
            return list;
        }
    }

    /* Stop the specified thread */
    void stopThread(int id) throws Exception {
        synchronized (out) {
            out.write(CMD_STOP_THREAD);
            out.writeInt(id);
            getReply(CMD_STOP_THREAD);
        }
    }

    /* Stop the specified threadgroup */
    void stopThreadGroup(int id) throws Exception {
        synchronized (out) {
            out.write(CMD_STOP_THREAD_GROUP);
            out.writeInt(id);
            getReply(CMD_STOP_THREAD_GROUP);
        }
    }

    /* Get the class name of a class */
    void getClassInfo(RemoteClass c) throws Exception {
        synchronized (out) {
            out.write(CMD_GET_CLASS_INFO);
            out.writeInt(c.id);

            getReply(CMD_GET_CLASS_INFO);
            c.name = in.readUTF();
            c.sourceName = in.readUTF();
            c.intf = in.readInt() != 0;
            c.superclass = (RemoteClass)readValue(in);
            c.loader = (RemoteObject)readValue(in);
            c.interfaces = new RemoteClass[in.readInt()];
            for (int i = 0 ; i < c.interfaces.length ; i++) {
                c.interfaces[i] = (RemoteClass)readValue(in);
            }
        }
    }

    /* Get a remote class by name */
    RemoteClass getClassByName(String name) throws Exception {
        synchronized (out) {
            out.write(CMD_GET_CLASS_BY_NAME);
            out.writeUTF(name);

            getReply(CMD_GET_CLASS_BY_NAME);
            return (RemoteClass)readValue(in);
        }
    }	

    /* Get threadgroup internal information */
    void getThreadGroupInfo(RemoteThreadGroup c) throws Exception {
        synchronized (out) {
            out.write(CMD_GET_THREADGROUP_INFO);
            out.writeInt(c.id);

            getReply(CMD_GET_THREADGROUP_INFO);
            c.parent = (RemoteThreadGroup)readValue(in);
            c.name = in.readUTF();
            c.maxPriority = in.readInt();
            c.daemon = in.readBoolean();
        }
    }

    /* Get the name of a thread */
    String getThreadName(int id) throws Exception {
        synchronized (out) {
            out.write(CMD_GET_THREAD_NAME);
            out.writeInt(id);

            getReply(CMD_GET_THREAD_NAME);
            return in.readUTF();
        }
    }

    /* Get the current status of a thread */
    int getThreadStatus(int id) throws Exception  {
        synchronized (out) {
            out.write(CMD_GET_THREAD_STATUS);
            out.writeInt(id);

            getReply(CMD_GET_THREAD_STATUS);
            return in.readInt();
        }
    }

    /* Free all unreferenced objects */
    void gc(RemoteObject[] objs) throws Exception  {
        synchronized (out) {
            out.write(CMD_MARK_OBJECTS);

            /* null objects in the list will be skipped. Count them now. */
            int nonNullCount = 0;
            for (int i = 0; i < objs.length; i++) {
                if (objs[i] != null)
                    nonNullCount++;
            }

            /* Send a list of all objects we want to keep. */
            out.writeInt(nonNullCount);
            for (int i = 0; i < objs.length; i++) {
                if (objs[i] != null) {
                    out.writeInt(objs[i].id);
                }
            }

            getReply(CMD_MARK_OBJECTS);
            out.write(CMD_FREE_OBJECTS);
            getReply(CMD_FREE_OBJECTS);

            // Start a new object cache
            Hashtable oldObjects = objects;
            objects = new Hashtable();

            /*
             * Read the set of saved ids from the agent (this will include
             * the marked objects). For each id preserve the corresponding
             * RemoteObject.
             */
            int id = in.readInt();
            while (id != NULL_OBJECT_ID) {
                Integer idObject = new Integer(id);
                RemoteObject obj = (RemoteObject)oldObjects.remove(idObject);
                objects.put(new Integer(id), obj);
                id = in.readInt();
            }
                
            /*
             * Invalidate all remaining RemoteObjects in the old cache.
             * Prevents a second free upon their finalization. 
             */
            Enumeration objectList = oldObjects.elements();
            while (objectList.hasMoreElements()) {
                RemoteObject obj = (RemoteObject) objectList.nextElement();
                obj.invalidate();
            }

        }
    }

    /* Turn on/off method call tracing. */
    void trace(boolean traceOn) throws Exception  {
        synchronized (out) {
            out.write(CMD_SYSTEM);
            out.writeInt(SYS_TRACE_METHOD_CALLS);
            out.writeInt(traceOn ? 1 : 0);
            getReply(CMD_SYSTEM);
        }
    }

    /* Turn on/off instruction tracing */
    void itrace(boolean traceOn) throws Exception  {
        synchronized (out) {
            out.write(CMD_SYSTEM);
            out.writeInt(SYS_TRACE_INSTRUCTIONS);
            out.writeInt(traceOn ? 1 : 0);
            getReply(CMD_SYSTEM);
        }
    }

    /* Report Agent's total memory usage.  */
    int totalMemory() throws Exception  {
        synchronized (out) {
            out.write(CMD_SYSTEM);
            out.writeInt(SYS_MEMORY_TOTAL);
            getReply(CMD_SYSTEM);
            return in.readInt();
        }
    }

    /* Report Agent's free memory.  */
    int freeMemory() throws Exception  {
        synchronized (out) {
            out.write(CMD_SYSTEM);
            out.writeInt(SYS_MEMORY_FREE);
            getReply(CMD_SYSTEM);
            return in.readInt();
        }
    }

    /* Run main program */
    RemoteThreadGroup run(int argc, String argv[]) throws Exception  {
        synchronized (out) {
            out.write(CMD_RUN);
            out.write(argc);
            for (int i = 0 ; i < argc ; i++) {
                out.writeUTF(argv[i]);
            }
            getReply(CMD_RUN);
            return (RemoteThreadGroup)readValue(in);
        }
    }

    /* Get remote copies of all the fields of an Object. */
    RemoteField[] getFields(int id) throws Exception {
        synchronized (out) {
            out.write(CMD_GET_FIELDS);
            out.writeInt(id);

            getReply(CMD_GET_FIELDS);
            int n = in.readInt();
            RemoteField list[] = new RemoteField[n];
            for (int i = 0 ; i < n ; i++) {
                int slot = in.readInt();
                String name = in.readUTF();
                String signature = in.readUTF();
                short access = in.readShort();
                RemoteClass clazz = (RemoteClass)readValue(in);
                list[i] = new RemoteField(this, slot, name, signature,
                                          access, clazz);
            }
            return list;
        }
    }

    /* Get all the elements of an array. */
    RemoteValue[] getElements(int id, int type,
                              int beginIndex, int endIndex) throws Exception 
    {
        synchronized (out) {
            out.write(CMD_GET_ELEMENTS);
            out.writeInt(id);
            out.writeInt(type);
            out.writeInt(beginIndex);
            out.writeInt(endIndex);

            getReply(CMD_GET_ELEMENTS);
            RemoteValue elements[] = new RemoteValue[in.readInt()];
            for (int i = 0 ; i < elements.length ; i++) {
                elements[i] = readValue(in);
            }
            return elements;
        }
    }

    /* Get remote copies of all the methods of an Object. */
    RemoteField[] getMethods(int id) throws Exception {
        synchronized (out) {
            out.write(CMD_GET_METHODS);
            out.writeInt(id);

            getReply(CMD_GET_METHODS);
            int n = in.readInt();
            message("getting (" + n + ") methods");
            RemoteField list[] = new RemoteField[n];
            for (int i = 0 ; i < n ; i++) {
                int slot = in.readInt();
                String name = in.readUTF();
                String signature = in.readUTF();
                short access = in.readShort();
message("Method1 " + name + " - " + signature + ", slot: " + slot);
                RemoteValue val = readValue(in);
                RemoteClass clazz = (RemoteClass)val;
                list[i] = new RemoteField(this, slot, name, signature,
                                          access, null);
message("Method2 " + name + " - " + signature + ", slot: " + slot + ", vars: " + list);
            }
            return list;
        }
    }

    /* Return the first line number of the specified method index, based
     * on the method array returned from getMethods(). */
    int getMethodLineNumber(int id, int index) throws Exception {
        synchronized (out) {
            out.write(CMD_GET_METHOD_LINENO);
            out.writeInt(id);
            out.writeInt(index);

            getReply(CMD_GET_METHOD_LINENO);
            return in.readInt();
        }
    }

    /* Get the type signature of a Class's field. */
/*** not used
    int getSlotSignature(int id, int n)[] throws Exception {
        synchronized (out) {
            out.write(CMD_GET_SLOT_SIGNATURE);
            out.writeInt(id);
            out.writeInt(n);

            getReply(CMD_GET_SLOT_SIGNATURE);
            int sig[] = new int[in.readInt()];
            for (int i = 0; i < sig.length; i++) {
                sig[i] = in.readInt();
            }
            return sig;
        }
    }
*****/

    /* Get the object in slot n of an Object. */
    RemoteValue getSlotValue(int id, int slot) throws Exception {
        synchronized (out) {
            out.write(CMD_GET_SLOT_VALUE);
            out.writeInt(id);
            out.writeInt(slot);

            getReply(CMD_GET_SLOT_VALUE);
            RemoteValue rv = readValue(in);
            if (rv != null && rv.isObject() && rv.getClass().getName().equals(
                "java.lang.ArrayIndexOutOfBoundsException")) {
                throw new 
                    ArrayIndexOutOfBoundsException("invalid slot index "+slot);
            }
            return rv;
        }
    }

    /* Get the object in slot n of a StackFrame. */
    RemoteValue getStackValue(int threadId, int frame, int slot,
                              char type) throws Exception {
        synchronized (out) {
            out.write(CMD_GET_STACK_VALUE);
            out.writeInt(threadId);
            out.writeInt(frame);
            out.writeInt(slot);
            out.writeChar(type);

            getReply(CMD_GET_STACK_VALUE);
            return readValue(in);
        }
    }

    /* Set slot n of an Object to an integer value. */
    void setSlotValue(int id, int slot, int value) throws Exception {
        synchronized (out) {
            out.write(CMD_SET_SLOT_VALUE);
            out.writeInt(id);
            out.writeInt(slot);
            out.writeInt(TC_INT);
            out.writeInt(value);
            getReply(CMD_SET_SLOT_VALUE);
        }
    }

    /* Set slot n of an Object to a boolean value. */
    void setSlotValue(int id, int slot, boolean value) throws Exception {
        synchronized (out) {
            out.write(CMD_SET_SLOT_VALUE);
            out.writeInt(id);
            out.writeInt(slot);
            out.writeInt(TC_BOOLEAN);
            out.writeBoolean(value);
            getReply(CMD_SET_SLOT_VALUE);
        }
    }

    /* Set slot n of an Object to a char value. */
    void setSlotValue(int id, int slot, char value) throws Exception {
        synchronized (out) {
            out.write(CMD_SET_SLOT_VALUE);
            out.writeInt(id);
            out.writeInt(slot);
            out.writeInt(TC_CHAR);
            out.writeChar(value);
            getReply(CMD_SET_SLOT_VALUE);
        }
    }

    /* Set slot n of an Object to a long value. */
    void setSlotValue(int id, int slot, long value) throws Exception {
        synchronized (out) {
            out.write(CMD_SET_SLOT_VALUE);
            out.writeInt(id);
            out.writeInt(slot);
            out.writeInt(TC_LONG);
            out.writeLong(value);
            getReply(CMD_SET_SLOT_VALUE);
        }
    }

    /* Set slot n of an Object to a float value. */
    void setSlotValue(int id, int slot, float value) throws Exception {
        synchronized (out) {
            out.write(CMD_SET_SLOT_VALUE);
            out.writeInt(id);
            out.writeInt(slot);
            out.writeInt(TC_FLOAT);
            out.writeFloat(value);
            getReply(CMD_SET_SLOT_VALUE);
        }
    }

    /* Set slot n of an Object to a double value. */
    void setSlotValue(int id, int slot, double value) throws Exception {
        synchronized (out) {
            out.write(CMD_SET_SLOT_VALUE);
            out.writeInt(id);
            out.writeInt(slot);
            out.writeInt(TC_DOUBLE);
            out.writeDouble(value);
            getReply(CMD_SET_SLOT_VALUE);
        }
    }

    /* Set slot n of a stackframe to an integer value. */
    void setStackValue(int threadId, int frame, int slot, int value) throws Exception {
        synchronized (out) {
            out.write(CMD_SET_STACK_VALUE);
            out.writeInt(threadId);
            out.writeInt(frame);
            out.writeInt(slot);
            out.writeInt(TC_INT);
            out.writeInt(value);
            getReply(CMD_SET_STACK_VALUE);
        }
    }

    /* Set slot n of a stackframe to a boolean value. */
    void setStackValue(int threadId, int frame, int slot, boolean value) throws Exception {
        synchronized (out) {
            out.write(CMD_SET_STACK_VALUE);
            out.writeInt(threadId);
            out.writeInt(frame);
            out.writeInt(slot);
            out.writeInt(TC_BOOLEAN);
            out.writeBoolean(value);
            getReply(CMD_SET_STACK_VALUE);
        }
    }

    /* Set slot n of a stackframe to a char value. */
    void setStackValue(int threadId, int frame, int slot, char value) throws Exception {
        synchronized (out) {
            out.write(CMD_SET_STACK_VALUE);
            out.writeInt(threadId);
            out.writeInt(frame);
            out.writeInt(slot);
            out.writeInt(TC_INT);
            out.writeInt((int)value);
            getReply(CMD_SET_STACK_VALUE);
        }
    }

    /* Set slot n of a stackframe to a long value. */
    void setStackValue(int threadId, int frame, int slot, long value) throws Exception {
        synchronized (out) {
            out.write(CMD_SET_STACK_VALUE);
            out.writeInt(threadId);
            out.writeInt(frame);
            out.writeInt(slot);
            out.writeInt(TC_LONG);
            out.writeLong(value);
            getReply(CMD_SET_STACK_VALUE);
        }
    }

    /* Set slot n of a stackframe to a float value. */
    void setStackValue(int threadId, int frame, int slot, float value) throws Exception {
        synchronized (out) {
            out.write(CMD_SET_STACK_VALUE);
            out.writeInt(threadId);
            out.writeInt(frame);
            out.writeInt(slot);
            out.writeInt(TC_DOUBLE);
            out.writeDouble((double)value);
            getReply(CMD_SET_STACK_VALUE);
        }
    }

    /* Set slot n of a stackframe to a double value. */
    void setStackValue(int threadId, int frame, int slot, double value) throws Exception {
        synchronized (out) {
            out.write(CMD_SET_STACK_VALUE);
            out.writeInt(threadId);
            out.writeInt(frame);
            out.writeInt(slot);
            out.writeInt(TC_DOUBLE);
            out.writeDouble(value);
            getReply(CMD_SET_STACK_VALUE);
        }
    }

    /* Set a breakpoint at a specified source line number in a class. */
    String setBreakpointLine(RemoteClass clazz, int lineno) throws Exception {
        synchronized (out) {
            out.write(CMD_SET_BRKPT_LINE);
            out.writeInt(clazz.getId());
            out.writeInt(lineno);

            getReply(CMD_SET_BRKPT_LINE);
            return in.readUTF();
        }
    }

    /* Set a breakpoint at the first line of a specified class method. */
    String setBreakpointMethod(RemoteClass clazz,
                               RemoteField method) throws Exception 
    {
        synchronized (out) {
            out.write(CMD_SET_BRKPT_METHOD);
            out.writeInt(clazz.getId());
            out.writeInt(method.slot);

            getReply(CMD_SET_BRKPT_METHOD);
            return in.readUTF();
        }
    }

    /* Return a list of breakpoints that are currently set. */
    String listBreakpoints()[] throws Exception {
        synchronized (out) {
            out.write(CMD_LIST_BREAKPOINTS);
            getReply(CMD_LIST_BREAKPOINTS);
            String list[] = new String[in.readInt()];
            for (int i = 0; i < list.length; i++) {
                list[i] = in.readUTF();
            }
            return list;
        }
    }

    /* Clear a breakpoint at a specific address. */
    String clearBreakpoint(RemoteClass clazz, int pc) throws Exception {
        synchronized (out) {
            message("clearing bkpt at " + clazz.getName() + ", pc " + pc);
            out.write(CMD_CLEAR_BRKPT);
            out.writeInt(clazz.getId());
            out.writeInt(pc);

            getReply(CMD_CLEAR_BRKPT);
            return in.readUTF();
        }
    }

    /* Clear a breakpoint at a specified line. */
    String clearBreakpointLine(RemoteClass clazz, int lineno) throws Exception {
        synchronized (out) {
            message("clearing bkpt at " + clazz.getName() + ":" + lineno);
            out.write(CMD_CLEAR_BRKPT_LINE);
            out.writeInt(clazz.getId());
            out.writeInt(lineno);

            getReply(CMD_CLEAR_BRKPT_LINE);
            return in.readUTF();
        }
    }

    /* Clear a breakpoint at the start of a specified method. */
    String clearBreakpointMethod(RemoteClass clazz,
                                 RemoteField method) throws Exception 
    {
        synchronized (out) {
            message("clearing bkpt at " + clazz.getName() + ":" + method.getName());
            out.write(CMD_CLEAR_BRKPT_METHOD);
            out.writeInt(clazz.getId());
            out.writeInt(method.slot);

            getReply(CMD_CLEAR_BRKPT_METHOD);
            return in.readUTF();
        }
    }

    /* Add an exception class to be "caught", i.e. enter the debugger when
     * thrown.
     */
    void catchExceptionClass(RemoteClass c) throws Exception {
        synchronized (out) {
            out.write(CMD_CATCH_EXCEPTION_CLS);
            out.writeInt(c.getId());
            getReply(CMD_CATCH_EXCEPTION_CLS);
        }
    }	

    /* Don't enter the debugger when the specified exception class is
     * thrown.
     */
    void ignoreExceptionClass(RemoteClass c) throws Exception {
        synchronized (out) {
            out.write(CMD_IGNORE_EXCEPTION_CLS);
            out.writeInt(c.getId());
            getReply(CMD_IGNORE_EXCEPTION_CLS);
        }
    }

    /* Return a list of the exceptions the debugger will stop on. */
    String[] getExceptionCatchList() throws Exception {
        synchronized (out) {
            out.write(CMD_GET_CATCH_LIST);
            getReply(CMD_GET_CATCH_LIST);
            int n = in.readInt();
            String list[] = new String[n];
            for (int i = 0; i < n; i++) {
                RemoteClass c = (RemoteClass)readValue(in);
                list[i] = c.getName();
            }
            return list;
        }
    }

    /* Get the source file path the Agent is currently using. */
    String getSourcePath() throws Exception {
        synchronized (out) {
            out.write(CMD_GET_SOURCEPATH);
            getReply(CMD_GET_SOURCEPATH);
            return in.readUTF();
        }
    }

    /* Set the Agent's source file path. */
    void setSourcePath(String pathList) throws Exception {
        synchronized (out) {
            out.write(CMD_SET_SOURCEPATH);
            out.writeUTF(pathList);
            getReply(CMD_SET_SOURCEPATH);
        }
    }

    /* Step through code an instruction or line. */
    void stepThread(int id, boolean stepLine) throws Exception {
        synchronized (out) {
            out.write(CMD_STEP_THREAD);
            out.writeInt(id);
            out.writeBoolean(stepLine);
            getReply(CMD_STEP_THREAD);
        }
    }

    /* Step through code an instruction or line. */
    void stepNextThread(int id) throws Exception {
        synchronized (out) {
            out.write(CMD_NEXT_STEP_THREAD);
            out.writeInt(id);
            getReply(CMD_NEXT_STEP_THREAD);
        }
    }

    /* Step to return address */                     // GVH SAS step out
    void stepOutThread(int id) throws Exception {    // GVH SAS step out
        synchronized (out) {                         // GVH SAS step out
            out.write(CMD_STEP_OUT_THREAD);          // GVH SAS step out
            out.writeInt(id);                        // GVH SAS step out
            getReply(CMD_STEP_OUT_THREAD);           // GVH SAS step out
        }                                            // GVH SAS step out
    }                                                // GVH SAS step out

    /* Set whether internal debugging messages are reported. */
    private void setVerbose(boolean b) throws Exception {
        synchronized (out) {
            out.write(CMD_SET_VERBOSE);
            out.writeBoolean(b);
            getReply(CMD_SET_VERBOSE);
        }
    }

    /* Notify agent that object reference has been finalized. */
    void objectFinalize(int id) throws Exception {
        synchronized (out) {
            out.write(CMD_OBJECT_FINALIZED);
            out.writeInt(id);
            getReply(CMD_OBJECT_FINALIZED);
        }
    }

    /* Return an array of all the line numbers which have code associated
     * with them. */
    int[] getLineNumbers(RemoteClass c) throws Exception {
        synchronized (out) {
            out.write(CMD_GET_LINENUMBERS);
            out.writeInt(c.getId());
            getReply(CMD_GET_LINENUMBERS);
            int n = in.readInt();
            int list[] = new int[ n ];
            for (int i = 0 ; i < n ; i++) {
                list[ i ] = in.readInt();
            }
            return list;
        }
    }

    /*
     * Funnel all client output through here to insure that client.printToConsole
     * is not called at the same time by two different threads. Otherwise we will
     * have a subtle difference in behavior from previous versions.
     */
    synchronized void printToClientConsole(String text) throws Exception {
        client.printToConsole(text);
    }
        

    /* Error message */
    void error(String msg) {
        try {
            printToClientConsole("[Internal debugger error: " + msg + "]\n");
        } catch (Exception e) {
            System.out.println("[Internal debugger error: " + msg + "]");
        }
        throw new InternalError();
    }

    /* Agent debugging message (only printed in verbose mode) */
    void message(String msg) {
	if (verbose) {
            try {
                printToClientConsole("[debugger: " + msg + "]\n");
            } catch (Exception e) {
                System.out.println("[debugger: " + msg + "]");
            }
	}
    }

    /* Remove the stack frame array reference from each cached RemoteThread. This   GVH SAS # 4027990
       method is called after each breakpoint and exception event. It ensures that  GVH SAS # 4027990
       fresh stack traces will be used by the RemoteThreads from this point forward GVH SAS # 4027990*/
    void resetStacks()                                      // GVH SAS # 4027990
    {                                                       // GVH SAS # 4027990
        Enumeration e = objects.elements();                 // GVH SAS # 4027990
        while (e.hasMoreElements())                         // GVH SAS # 4027990
        {                                                   // GVH SAS # 4027990
            Object obj = e.nextElement();                   // GVH SAS # 4027990
            if (obj instanceof RemoteThread)                // GVH SAS # 4027990
            {                                               // GVH SAS # 4027990
                ((RemoteThread)obj).resetStack();           // GVH SAS # 4027990
            }                                               // GVH SAS # 4027990
        }                                                   // GVH SAS # 4027990
    }                                                       // GVH SAS # 4027990
}
