/*
 * @(#)RemoteClass.java	1.24 98/03/18
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.tools.debug;

import java.lang.Exception;
import java.io.*;

/**
 * The RemoteClass class allows access to a class in a remote Java
 * interpreter.  
 *
 * @see RemoteDebugger
 * @author Thomas Ball
 */
public class RemoteClass extends RemoteObject {
    String name;
    String sourceName;
    boolean intf;
    RemoteClass superclass;
    RemoteObject loader;
    RemoteClass interfaces[] = null;
    private RemoteField methods[] = null;
    private RemoteField instanceFields[] = null;
    private RemoteField staticFields[] = null;

    RemoteClass(RemoteAgent agent, int id) {
	super(agent, TP_CLASS, id, agent.classClass);
    }

    private void getClassInfo() throws Exception {
	if (interfaces == null) {
	    agent.getClassInfo(this);
	}
    }

    private void loadFields() throws Exception {
	if (staticFields == null || instanceFields == null) {
	    RemoteField fields[] = agent.getFields(id);

	    int nStaticFields = 0;
	    for (int i = 0; i < fields.length; i++) {
		if (fields[i].isStatic() == true) {
		    nStaticFields++;
		}
	    }

	    staticFields = new RemoteField[nStaticFields];
	    instanceFields = new RemoteField[fields.length - nStaticFields];
	    int iInst = 0;
	    int iStatic = 0;
	    for (int i = 0; i < fields.length; i++) {
		if (fields[i].isStatic() == true) {
		    staticFields[iStatic++] = fields[i];
		} else {
		    instanceFields[iInst++] = fields[i];
		}
	    }
	}
    }
    
    /**
     * Returns the name of the class.
     *
     * @exception java.lang.Exception if a general exception occurs.
     */
    public String getName() throws Exception {
	getClassInfo();
	return name;
    }

    /**
     * Returns the name of the class as its type.
     *
     * @exception java.lang.Exception if a general exception occurs.
     */
    public String typeName() throws Exception {
	getClassInfo();
	return getName();
    }

    /**
     * Is this RemoteClass an interface?
     *
     * @exception java.lang.Exception if a general exception occurs.
     */
    public boolean isInterface() throws Exception {
	getClassInfo();
	return intf;
    }

    /**
     * Return the superclass for this class.
     *
     * @exception java.lang.Exception if a general exception occurs.
     */
    public RemoteClass getSuperclass() throws Exception {
	getClassInfo();
	return superclass;
    }

    /**
     * Return the classloader for this class.
     *
     * @exception java.lang.Exception if a general exception occurs.
     */
    public RemoteObject getClassLoader() throws Exception {
	getClassInfo();
	return loader;
    }

    /**
     * Return the interfaces for this class.
     *
     * @exception java.lang.Exception if a general exception occurs.
     */
    public RemoteClass getInterfaces()[] throws Exception {
	getClassInfo();
	return interfaces;
    }

    /**
     * Get the name of the source file referenced by this stackframe.
     *
     * @exception java.lang.Exception if a general exception occurs.
     */
    public String getSourceFileName() throws Exception {
	getClassInfo();
	return sourceName;
    }

    /**
     * Get the source file referenced by this stackframe.
     *
     * @exception java.lang.Exception if a general exception occurs.
     */
    public InputStream getSourceFile() throws Exception {
	getClassInfo();
	return agent.getSourceFile(name, sourceName);
    }

    /**
     * Return all the static fields for this class.
     *
     * @exception java.lang.Exception if a general exception occurs.
     */
    public RemoteField[] getFields() throws Exception {
	return getStaticFields();
    }

    /**
     * Return all the static fields for this class.
     *
     * @exception java.lang.Exception if a general exception occurs.
     */
    public RemoteField[] getStaticFields() throws Exception {
	loadFields();
	return staticFields;
    }

    /** Return all the instance fields for this class.
     *  Note:  because this is a RemoteClass method, only the name and
     *         type methods will be valid, not the data.
     *
     * @exception java.lang.Exception if a general exception occurs.
     */
    public RemoteField[] getInstanceFields() throws Exception {
	loadFields();
	return instanceFields;
    }

    /**
     * Return the static field, specified by index.
     *
     * @exception ArrayIndexOutOfBoundsException when the index is greater than the number of instance variables
     * @exception java.lang.Exception if a general exception occurs.
     */
    public RemoteField getField(int n) throws Exception {
	loadFields();
	if (n < 0 || n >= staticFields.length) {
	    throw new ArrayIndexOutOfBoundsException();
	}
	return staticFields[n];
    }

    /**
     * Return the static field, specified by name.
     *
     * @exception java.lang.Exception            if a general exception occurs.
     * @exception java.lang.NoSuchFieldException if a field with the specified
     *              name is not found.
     */
    public RemoteField getField(String name) throws 
      NoSuchFieldException, Exception {
	loadFields();

    /* Agent has laid out slots from Superclass to subclass, so start searching  GVH SAS 4065384
       at the end.                                                               GVH SAS 4065384 */
	for (int i = staticFields.length-1; i >= 0; i--) {                        /* GVH SAS 4065384 */
	    if (name.equals(staticFields[i].getName())) {
		return staticFields[i];
	    }
	}
	throw new NoSuchFieldException();
    }

    /**
     * Return the instance field, specified by its index.
     * Note:  because this is a RemoteClass method, only the name and
     *        type information is valid, not the data.
     *
     * @exception ArrayIndexOutOfBoundsException when the index is greater than the number of instance variables
     * @exception java.lang.Exception if a general exception occurs.
     */
    public RemoteField getInstanceField(int n) throws Exception {
	loadFields();
	if (n < 0 || n >= instanceFields.length) {
	    throw new ArrayIndexOutOfBoundsException();
	}
	return instanceFields[n];
    }

    /**
     * Return the value of a static field, specified by its index.
     *
     * @exception java.lang.Exception if a general exception occurs.
     */
    public RemoteValue getFieldValue(int n) throws Exception {
	return getField(n).getValue(id);
    }

    /**
     * Return the value of a static field, specified by name.
     *
     * @exception java.lang.Exception            if a general exception occurs.
     * @exception java.lang.NoSuchFieldException if a field with the specified
     *              name is not found.
     */
    public RemoteValue getFieldValue(String name) throws 
      NoSuchFieldException, Exception {
        return getField(name).getValue(id);      // GVH SAS # 4065384
    }

    /**
     * Return the method, specified by name. If the method name is overloaded,
     * any one of the methods with that name is returned.
     *
     * @exception java.lang.Exception             if a general exception occurs.
     * @exception java.lang.NoSuchMethodException if a matching method is
     *              not found.
     */
    public RemoteField getMethod(String name) 
      throws NoSuchMethodException, Exception {
	if (methods == null) {
	    methods = agent.getMethods(id);
	}
	for (int i = 0; i < methods.length; i++) {
	    if (name.equals(methods[i].getName())) {
		return methods[i];
	    }
	}
        throw new NoSuchMethodException();
    }

    /**
     * Return the class's methods.
     *
     * @exception java.lang.Exception if a general exception occurs.
     */
    public RemoteField[] getMethods() throws Exception {
	if (methods == null) {
	    methods = agent.getMethods(id);
	}
	return methods;
    }

    /**
     * Return the names of all methods supported by this class.
     *
     * @exception java.lang.Exception if a general exception occurs.
     */
    public String[] getMethodNames() throws Exception {
	if (methods == null) {
	    methods = agent.getMethods(id);
	}
	String ret[] = new String[methods.length];
	for (int i = 0; i < methods.length; i++) {
	    ret[i] = methods[i].getName();
	}
	return ret;
    }

    /** 
     * Return the first line number of the specified method, or throws
     * NoSuchLineNumber exception.  If the method name is overloaded,
     * the first line from any one of the methods with that name is returned.
     *
     * @param name	the name of the method
     * @return the method's beginning line number
     * @exception java.lang.Exception             if a general exception occurs.
     * @exception NoSuchLineNumberException       if no code is associated with the
     *              specified line number in a Java source file.
     * @exception java.lang.NoSuchMethodException if a matching method is
     *              not found.
     */
    public int getMethodLineNumber(String name) 
      throws NoSuchMethodException, NoSuchLineNumberException, Exception {
	if (methods == null) {
	    methods = agent.getMethods(id);
	}
	for (int i = 0; i < methods.length; i++) {
	    if (name.equals(methods[i].getName())) {
                int lineno = agent.getMethodLineNumber(id, i);
                if (lineno == -1) {
                    throw new NoSuchLineNumberException();
                }
                return lineno;
	    }
	}
        throw new NoSuchMethodException();
    }

    /** 
     * Return the first line number of the specified method, or throws
     * NoSuchLineNumber exception.
     *
     * @param index	the index of the method, as found in the array
     * returned from getMethods().
     * @return the method's beginning line number
     * @exception java.lang.Exception       if a general exception occurs.
     * @exception NoSuchLineNumberException if no code is associated with the
     *              specified line number in a Java source file.
     */
    public int getMethodLineNumber(int index) 
      throws IndexOutOfBoundsException, NoSuchLineNumberException, Exception {
	if (methods == null) {
	    methods = agent.getMethods(id);
	}
        if (index < 0 || index >= methods.length) {
            throw new IndexOutOfBoundsException();
        }
        int lineno = agent.getMethodLineNumber(id, index);
        if (lineno == -1) {
            throw new NoSuchLineNumberException();
        }
        return lineno;
    }

    /**
     * Set a breakpoint at a specified source line number in a class.
     *
     * @param lineno	the line number where the breakpoint is set
     * @return an empty string if successful, otherwise a description of the error.
     * @exception java.lang.Exception if a general exception occurs.
     */
    public String setBreakpointLine(int lineno) throws Exception {
	return agent.setBreakpointLine(this, lineno);
    }

    /**
     * Set a breakpoint at the first line of a class method.
     *
     * @param method	the method where the breakpoint is set
     * @return an empty string if successful, otherwise a description of the error.
     * @exception java.lang.Exception if a general exception occurs.
     */
    public String setBreakpointMethod(RemoteField method) throws Exception {
	return agent.setBreakpointMethod(this, method);
    }

    /**
     * Clear a breakpoint at a specific address in a class.
     *
     * @param pc	the address of the breakpoint to be cleared
     * @return an empty string if successful, otherwise a description of the error.
     * @exception java.lang.Exception if a general exception occurs.
     */
    public String clearBreakpoint(int pc) throws Exception {
	return agent.clearBreakpoint(this, pc);
    }

    /**
     * Clear a breakpoint at a specified line.
     *
     * @param lineno	the line number of the breakpoint to be cleared
     * @return an empty string if successful, otherwise a description of the error.
     * @exception java.lang.Exception if a general exception occurs.
     */
    public String clearBreakpointLine(int lineno) throws Exception {
	return agent.clearBreakpointLine(this, lineno);
    }

    /**
     * Clear a breakpoint at the start of a specified method.
     *
     * @param method	the method of the breakpoint to be cleared
     * @return an empty string if successful, otherwise a description of the error.
     * @exception java.lang.Exception if a general exception occurs.
     */
    public String clearBreakpointMethod(RemoteField method) throws Exception {
	return agent.clearBreakpointMethod(this, method);
    }

    private boolean isExceptionClass() throws Exception {
	if (getName().equals("java.lang.Exception") ||
	    getName().equals("java.lang.Error")) {
	    return true;
	}
	RemoteClass superClass = superclass;
	do {
	    agent.message("isExceptionClass: superClass=" +
			  superClass.getName());
	    if (superClass.getName().equals("java.lang.Exception") ||
	    	superClass.getName().equals("java.lang.Error")) {
		return true;
	    }
	    superClass = superClass.getSuperclass();
	} while (superClass.getName().equals("java.lang.Object") == false);
	return false;
    }

    /**
     * Enter the debugger when an instance of this class is thrown.
     *
     * @exception ClassCastException when this class isn't an exception class
     * @exception java.lang.Exception if a general exception occurs.
     */
    public void catchExceptions() throws Exception {
	if (isExceptionClass()) {
	    agent.catchExceptionClass(this);
	} else {
	    throw new ClassCastException();
	}
    }

    /**
     * Don't enter the debugger when an instance of this class is thrown.
     *
     * @exception ClassCastException when this class isn't an exception class
     * @exception java.lang.Exception if a general exception occurs.
     */
    public void ignoreExceptions() throws Exception {
	if (isExceptionClass()) {
	    agent.ignoreExceptionClass(this);
	} else {
	    throw new ClassCastException();
	}
    }

    /**
     * Return an array of all the class's source file line numbers which 
     * have code associated with them.  This array may be zero-length, as
     * the class may not have been compiled with line number information.
     *
     * @return  an array of integers of valid line numbers.
     * @exception java.lang.Exception if a general exception occurs.
     */
    public int[] getLineNumbers() throws Exception {
	getClassInfo();
        return agent.getLineNumbers(this);
    }

    /** Return a (somewhat verbose) description. */
    public String description() {
	return toString();
    }

    /** Return a (somewhat verbose) description. */
    public String toString() {
	try {
	    getClassInfo();
	    return toHex(id) + ":" + (intf ? "interface" : "class") +
	        "(" + getName() + ")";
	} catch (Exception ex) {
	    return ("<communications errors>");
	}
    }
}
