/*
 * @(#)RemoteDebugger.java	1.28 98/09/21
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.tools.debug;

import java.io.*;
import java.net.InetAddress;

/**
 * The RemoteDebugger class defines a client interface to the Java debugging
 * classes.  It is used to instantiate a connection with the Java interpreter
 * being debugged.  
 *
 * @version 1.0 26 June 1995
 * @author Thomas Ball
 */
public class RemoteDebugger {

    private RemoteAgent agent;

    /**
     * Create a remote debugger, connecting it with a running Java
     * interpreter.<P>To connect to a running interpreter, it must be 
     * started with the "-debug" option, whereupon it will print out 
     * the password for that debugging session.
     * @param host	the name of the system where a debuggable Java instance is running (default is localhost).
     * @param password	the password reported by the debuggable Java instance.  This should be null when starting a client interpreter.
     * @param client	the object to which notification messages are sent (it must support the DebuggerCallback interface)
     * @param verbose	turn on internal debugger message text
     * @exception java.lang.Exception if a general exception occurs.
     */
    public RemoteDebugger(String host, String password, 
			  DebuggerCallback client, 
                          boolean verbose) throws Exception {
	agent = new RemoteAgent(host, password, "", client, verbose);
    }

    /**
     * Create a remote debugger.<P>
     * @param javaArgs	optional java command-line parameters, such as -classpath
     * @param client	the object to which notification messages are sent (it must support the DebuggerCallback interface)
     * @param verbose	turn on internal debugger message text
     * @exception java.lang.Exception if a general exception occurs.
     */
    public RemoteDebugger(String javaArgs, DebuggerCallback client, 
                          boolean verbose) throws Exception {
// -- don't support local agent --
//        if (Agent.isDebuggable()) {
//            agent = new RemoteAgent(client, verbose);
//        } else {
	    /* Use the loopback IP address for socket IPC */
            agent = new RemoteAgent("127.0.0.1",
                                    null, javaArgs, client, verbose);
//        }
    }

    /**
     * Close the connection to the remote debugging agent.
     */
    public void close() {
	agent.close();
    }

    /**
     * Get an object from the remote object cache.
     *
     * @param id	the remote object's id
     * @return the specified RemoteObject, or null if not cached.
     */
    public RemoteObject get(Integer id) {
	return agent.get(id);
    }

    /**
     * List the currently known classes.
     *
     * @exception java.lang.Exception if a general exception occurs.
     */
    public RemoteClass[] listClasses() throws Exception {
	return agent.listClasses();
    }

    /**
     * Find a specified class.  If the class isn't already known by the
     * remote debugger, the lookup request will be passed to the Java
     * interpreter being debugged.
     *
     * NOTE:  Substrings, such as "String" for "java.lang.String" will
     * return with the first match, and will not be successfully found
     * if the request is passed to the remote interpreter.
     *
     * @param name	the name (or a substring of the name)of the class
     * @return the specified (Remote)Class, or null if not found.
     * @exception java.lang.Exception if a general exception occurs.
     */
    public RemoteClass findClass(String name) throws Exception {
	return agent.findClass(name);
    }

    /**
     * List threadgroups
     *
     * @param tg	the threadgroup which hold the groups to be listed, or null for all threadgroups
     * @exception java.lang.Exception if a general exception occurs.
     */
    public RemoteThreadGroup[] listThreadGroups(RemoteThreadGroup tg) throws Exception {
	return agent.listThreadGroups(tg);
    }

    /**
     * Continue an execution that has been suspended.  Restarts all threads 
     * suspended by virtue of having encountered a breakpoint or step.
     */
    public void cont() throws Exception {
        agent.resumeLastSuspendedThreads();
    }

    /**                                            GVH SAS # 4046670
     * Suspends all non-system threads in the debugged interpreter GVH SAS # 4046670
     */                                         // GVH SAS # 4046670
    public void suspend() throws Exception {    // GVH SAS # 4046670
        agent.suspendAllThreads();              // GVH SAS # 4046670
    }                                           // GVH SAS # 4046670
                                                // GVH SAS # 4046670

    /**
     * Free all objects referenced by the debugger.
     *
     * The remote debugger maintains a copy of each object it has examined,
     * so that references won't become invalidated by the garbage collector
     * of the Java interpreter being debugged.  The gc() method frees all
     * all of these references, except those specified to save.
     *
     * @param save_list	the list of objects to save.
     * @exception java.lang.Exception if a general exception occurs.
     */
    public void gc(RemoteObject[] save_list) throws Exception {
	agent.gc(save_list);
    }

    /**
     * Turn on/off method call tracing.
     *
     * When turned on, each method call is reported to the stdout of the
     * Java interpreter being debugged.  This output is not captured in
     * any way by the remote debugger.
     *
     * @param traceOn	turn tracing on or off
     * @exception java.lang.Exception if a general exception occurs.
     */
    public void trace(boolean traceOn) throws Exception {
	agent.trace(traceOn);
    }

    /**
     * Turn on/off instruction tracing.
     *
     * When turned on, each Java instruction is reported to the stdout of the
     * Java interpreter being debugged.  This output is not captured in
     * any way by the remote debugger.
     *
     * @param traceOn	turn tracing on or off
     * @exception java.lang.Exception if a general exception occurs.
     */
    public void itrace(boolean traceOn) throws Exception {
	agent.itrace(traceOn);
    }

    /**
     * Report the total memory usage of the Java interpreter being debugged.
     *
     * @exception java.lang.Exception if a general exception occurs.
     */
    public int totalMemory() throws Exception {
	return agent.totalMemory();
    }

    /**
     * Report the free memory available to the Java interpreter being debugged.
     *
     * @exception java.lang.Exception if a general exception occurs.
     */
    public int freeMemory() throws Exception {
	return agent.freeMemory();
    }

    /**
     * Load and run a runnable Java class, with any optional parameters.
     * The class is started inside a new threadgroup in the Java interpreter
     * being debugged.
     *
     * NOTE:  Although it is possible to run multiple runnable classes from
     * the same Java interpreter, there is no guarantee that all applets
     * will work cleanly with each other.  For example, two applets may want
     * exclusive access to the same shared resource,such as a specific port.
     *
     * @param argc	the number of parameters
     * @param argv	the array of parameters:  the class to be run is
     * first, followed by any optional parameters used by that class.
     * @return the new ThreadGroup the class is running in, or null on error
     * @exception java.lang.Exception if a general exception occurs.
     */
    public RemoteThreadGroup run(int argc, String argv[]) throws Exception {
	return agent.run(argc, argv);
    }

    /**
     * Return a list of the breakpoints which are currently set.
     *
     * @return an array of Strings of the form "class_name:line_number".
     * @exception java.lang.Exception if a general exception occurs.
     */
    public String[] listBreakpoints() throws Exception {
	return agent.listBreakpoints();
    }

    /**
     * Return the list of the exceptions the debugger will stop on.
     *
     * @return an array of exception class names, which may be zero-length.
     * @exception java.lang.Exception if a general exception occurs.
     */
    public String[] getExceptionCatchList() throws Exception {
	return agent.getExceptionCatchList();
    }

    /**
     * Return the source file path the Agent is currently using.
     *
     * @return a string consisting of a list of colon-delineated paths.
     * @exception java.lang.Exception if a general exception occurs.
     */
    public String getSourcePath() throws Exception {
	return agent.getSourcePath();
    }

    /**
     * Specify the list of paths to use when searching for a source file.
     *
     * @param pathList	a string consisting of a list of colon-delineated paths.
     * @exception java.lang.Exception if a general exception occurs.
     */
    public void setSourcePath(String pathList) throws Exception {
	agent.setSourcePath(pathList);
    }

    /**
     * This makes no sense unless agent is running in the same VM
     * as the remote agent.  We don't support this.  So this is
     * now a no-op.
     */
    public void addSystemThread() {
//      Agent.addSystemThread(Thread.currentThread());
    }

    /**
     * This makes no sense unless agent is running in the same VM
     * as the remote agent.  We don't support this.  So this is
     * now a no-op.
     */
    public void addSystemThread(Thread t) {
//      Agent.addSystemThread(t);
    }
}
