/*
 * @(#)RemoteObject.java	1.19 98/09/21
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.tools.debug;
import sun.tools.agent.AgentConstants;
import java.lang.Exception;

/**
 * The RemoteObject class allows access to an object in a remote
 * Java interpreter.  
 *
 * Remote objects are not created by the local debugger, but are returned
 * by the remote debugging agent when queried for the values of instance
 * or static variables of known objects (such as classes), or from local
 * (stack) variables.
 *
 * Each remote object has a reference cached by the remote Java interpreter,
 * so that the object will not be garbage-collected during examination.  The
 * RemoteDebugger's gc() operation frees references to objects that are no
 * longer being examined.
 *
 * @see RemoteDebugger
 * @see RemoteClass
 * @see RemoteString
 * @see RemoteThread
 * @see RemoteThreadGroup
 * @author Thomas Ball
 */
public class RemoteObject extends RemoteValue {
    RemoteAgent agent;
    RemoteClass clazz;
    int id;

    RemoteObject(RemoteAgent agent, int id, RemoteClass clazz) {
	this(agent, TP_OBJECT, id, clazz);
    }

    RemoteObject(RemoteAgent agent, int type, int id, RemoteClass clazz) {
	super(type);
	this.agent = agent;
	this.id = id;
	this.clazz = clazz;
    }

    /**
     * Returns the RemoteValue's type name ("Object").
     *
     * @exception java.lang.Exception if a general exception occurs.
     */
    public String typeName() throws Exception {
	return "Object";
    }

    /** Returns the id of the object. */
    public final int getId() {
	return id;
    }

    /** Returns the object's class. */
    public final RemoteClass getClazz() {
	return clazz;
    }

    /**
     * Returns the value of an object's instance variable.
     *
     * @param n	the slot number of the variable to be returned.
     * @exception java.lang.Exception if a general exception occurs.
     */
    public RemoteValue getFieldValue(int n) throws Exception {
	return clazz.getInstanceField(n).getValue(id);
    }

    /**
     * Returns the value of an object's instance variable.
     *
     * @param name	the name of the instance variable
     * @return the variable as a RemoteValue, or null if name not found.
     * @exception java.lang.Exception if a general exception occurs.
     */
    public RemoteValue getFieldValue(String name) throws Exception {
        RemoteField field = getField(name);      // GVH SAS # 4065384
        if (field == null) {                     // GVH SAS # 4065384
            return null;                         // GVH SAS # 4065384
        } else {                                 // GVH SAS # 4065384
            return field.getValue(id);           // GVH SAS # 4065384
        }                                        // GVH SAS # 4065384
    }                                            // GVH SAS # 4065384

    /**
     * Return the instance (non-static) fields of an object.
     *
     * @exception java.lang.Exception if a general exception occurs.
     */
    public RemoteField getFields()[] throws Exception {
	return clazz.getInstanceFields();
    }

    /**
     * Return an instance variable, specified by slot number.
     *
     * @param n	the slot number of the variable to be returned.
     * @exception java.lang.Exception if a general exception occurs.
     */
    public RemoteField getField(int n) throws Exception {
	return clazz.getInstanceField(n);
    }

    /**
     * Return an instance variable, specified by name.
     *
     * @param name	the name of the instance variable
     * @return the variable as a RemoteField, or null if name not found.
     * @exception java.lang.Exception if a general exception occurs.
     */
    public RemoteField getField(String name) throws Exception {
	RemoteField fields[] = clazz.getInstanceFields();
    /* Agent has laid out slots from Superclass to subclass, so start searching  GVH SAS 4065384
       at the end.                                                               GVH SAS 4065384 */
	for (int i = fields.length-1; i >= 0; i--) {                              /* GVH SAS 4065384 */
	    if (name.equals(fields[i].getName())) {
		return fields[i];
	    }
	}
	return null;
    }

    /**
     * Set a boolean instance variable, specified by name.  If the instance
     * variable doesn't exist, an IllegalAccessException is thrown.
     *
     * @param name	the name of the instance variable
     * @param value	the value to use
     * @exception java.lang.Exception if a general exception occurs.
     */
    public void setField(String name, boolean value) throws Exception {
	RemoteField rf = getField(name);
        if (rf == null) {
            throw new IllegalAccessException(
                name + " is not an instance variable of " + this);
        }
        rf.setValue(id, value);
    }

    /**
     * Set an int instance variable, specified by name.  If the instance
     * variable doesn't exist, an IllegalAccessException is thrown.
     *
     * @param name	the name of the instance variable
     * @param value	the value to use
     * @exception java.lang.Exception if a general exception occurs.
     */
    public void setField(String name, int value) throws Exception {
	RemoteField rf = getField(name);
        if (rf == null) {
            throw new IllegalAccessException(
                name + " is not an instance variable of " + this);
        }
        rf.setValue(id, value);
    }

    /**
     * Set a char instance variable, specified by name.  If the instance
     * variable doesn't exist, an IllegalAccessException is thrown.
     *
     * @param name	the name of the instance variable
     * @param value	the value to use
     * @exception java.lang.Exception if a general exception occurs.
     */
    public void setField(String name, char value) throws Exception {
	RemoteField rf = getField(name);
        if (rf == null) {
            throw new IllegalAccessException(
                name + " is not an instance variable of " + this);
        }
        rf.setValue(id, value);
    }

    /**
     * Set a long instance variable, specified by name.  If the instance
     * variable doesn't exist, an IllegalAccessException is thrown.
     *
     * @param name	the name of the instance variable
     * @param value	the value to use
     * @exception java.lang.Exception if a general exception occurs.
     */
    public void setField(String name, long value) throws Exception {
	RemoteField rf = getField(name);
        if (rf == null) {
            throw new IllegalAccessException(
                name + " is not an instance variable of " + this);
        }
        rf.setValue(id, value);
    }

    /**
     * Set a float instance variable, specified by name.  If the instance
     * variable doesn't exist, an IllegalAccessException is thrown.
     *
     * @param name	the name of the instance variable
     * @param value	the value to use
     * @exception java.lang.Exception if a general exception occurs.
     */
    public void setField(String name, float value) throws Exception {
	RemoteField rf = getField(name);
        if (rf == null) {
            throw new IllegalAccessException(
                name + " is not an instance variable of " + this);
        }
        rf.setValue(id, value);
    }

    /**
     * Set a double instance variable, specified by name.  If the instance
     * variable doesn't exist, an IllegalAccessException is thrown.
     *
     * @param name	the name of the instance variable
     * @param value	the value to use
     * @exception java.lang.Exception if a general exception occurs.
     */
    public void setField(String name, double value) throws Exception {
	RemoteField rf = getField(name);
        if (rf == null) {
            throw new IllegalAccessException(
                name + " is not an instance variable of " + this);
        }
        rf.setValue(id, value);
    }

    /**
     * Set an object instance variable, specified by name.  This isn't
     * currently supported, so an IllegalAccessException is always thrown.
     * (The API is defined so that this feature can be implemented in
     * future releases.)
     *
     * @param name	the name of the instance variable
     * @param object	the RemoteObject to use
     * @exception java.lang.Exception if a general exception occurs.
     */
    public void setField(String name, RemoteObject object) throws Exception {
        throw new IllegalAccessException();
    }

    /**
     * Set a boolean instance variable, specified by slot.  If the instance
     * variable doesn't exist, an IllegalAccessException is thrown.
     *
     * @param slot	the slot number of the instance variable
     * @param value	the value to use
     * @exception java.lang.Exception if a general exception occurs.
     */
    public void setField(int slot, boolean value) throws Exception {
        getField(slot).setValue(id, value);
    }

    /**
     * Set an int instance variable, specified by slot.  If the instance
     * variable doesn't exist, an IllegalAccessException is thrown.
     *
     * @param slot	the slot number of the instance variable
     * @param value	the value to use
     * @exception java.lang.Exception if a general exception occurs.
     */
    public void setField(int slot, int value) throws Exception {
        getField(slot).setValue(id, value);
    }

    /**
     * Set a char instance variable, specified by slot.  If the instance
     * variable doesn't exist, an IllegalAccessException is thrown.
     *
     * @param slot	the slot number of the instance variable
     * @param value	the value to use
     * @exception java.lang.Exception if a general exception occurs.
     */
    public void setField(int slot, char value) throws Exception {
        getField(slot).setValue(id, value);
    }

    /**
     * Set a long instance variable, specified by slot.  If the instance
     * variable doesn't exist, an IllegalAccessException is thrown.
     *
     * @param slot	the slot number of the instance variable
     * @param value	the value to use
     * @exception java.lang.Exception if a general exception occurs.
     */
    public void setField(int slot, long value) throws Exception {
        getField(slot).setValue(id, value);
    }

    /**
     * Set a float instance variable, specified by slot.  If the instance
     * variable doesn't exist, an IllegalAccessException is thrown.
     *
     * @param slot	the slot number of the instance variable
     * @param value	the value to use
     * @exception java.lang.Exception if a general exception occurs.
     */
    public void setField(int slot, float value) throws Exception {
        getField(slot).setValue(id, value);
    }

    /**
     * Set a double instance variable, specified by slot.  If the instance
     * variable doesn't exist, an IllegalAccessException is thrown.
     *
     * @param slot	the slot number of the instance variable
     * @param value	the value to use
     * @exception java.lang.Exception if a general exception occurs.
     */
    public void setField(int slot, double value) throws Exception {
        getField(slot).setValue(id, value);
    }

    /**
     * Set an object instance variable, specified by slot.  This isn't
     * currently supported, so an IllegalAccessException is always thrown.
     * (The API is defined so that this feature can be implemented in
     * future releases.)
     *
     * @param slot	the slot number of the instance variable
     * @param object	the RemoteObject to use
     * @exception java.lang.Exception if a general exception occurs.
     */
    public void setField(int slot, RemoteObject object) throws Exception {
        throw new IllegalAccessException();
    }

    /** Return a description of the object. */
    public String description() {
	try {
	    if (clazz == null) {
	        return toHex(id);
	    } else {
	        return "(" + getClazz().getName() + ")" + toHex(id);
	    }
	} catch (Exception ex) {
	    return ("<communications error>");
	}
    }

    /** Return object as a string. */
    public String toString() {
	String retStr = agent.objectToString(id);
	if (retStr.length() > 0) {
	    return retStr;
	} else {
	    return description();
	}
    }

    /**
     *
     * @exception java.lang.Exception if a general exception occurs.
     */
    protected void finalize () throws Exception {

        /* Notify agent only if not invalidated by previous user gc */
        if (id != AgentConstants.INVALID_OBJECT_ID) {
            agent.objectFinalize(id);
        }
    }

    /**
     * Prevents this object from being used after being garbage-collected.
     */
    protected void invalidate() {
        id = AgentConstants.INVALID_OBJECT_ID;
    }
}
