/*
 * @(#)RemoteStackFrame.java	1.20 98/09/21
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.tools.debug;

import java.lang.Exception;

/**
 * The RemoteStackFrame class provides access to a stackframe of a
 * suspended thread.  
 *
 * @see RemoteDebugger
 * @see RemoteThread
 * @author Thomas Ball
 */
public class RemoteStackFrame {
    private RemoteObject idFrame;
    private RemoteAgent agent;
    private RemoteThread thread;
    private RemoteClass clazz;
    private int iFrame;
    String className;
    String methodName;
    String methodSignature;
    int lineno;
    int bci;
    LocalVariable[] localVariables;

    RemoteStackFrame(RemoteObject id, RemoteThread t, int i,
			    RemoteAgent ag) {
	idFrame = id;
	thread = t;
	iFrame = i;
	agent = ag;
    }

    /** 
     * Return a specific (named) stack variable.  A slot number of -1
     * indicates that the variable is not currently in scope. 
     *
     * @exception java.lang.Exception if a general exception occurs.
     */
    public RemoteStackVariable getLocalVariable(String name) throws Exception {
	agent.message("getLocalVariable: name=" + name);
	agent.message("   stackframe=" + toString());
	agent.message("   " + localVariables.length + " local vars");
	for (int i = 0; i < localVariables.length; i++) {
	    agent.message("   trying " + localVariables[i].name);
	    if (localVariables[i].name.equals(name)) {
		RemoteValue v = agent.getStackValue(thread.getId(), iFrame,
		    localVariables[i].slot,
		    localVariables[i].signature.charAt(0));
		return new RemoteStackVariable(this,
                                               localVariables[i].slot, name,
					       localVariables[i].signature, 
                                               localVariables[i].methodArgument, v);
	    }
	}
	return null;
    }

    /**
     * Return an array of all valid local variables and method arguments
     * for this stack frame.
     *
     * @exception java.lang.Exception if a general exception occurs.
     */
    public RemoteStackVariable[] getLocalVariables() throws Exception {
	agent.message("getLocalVariables:");
	agent.message("   stackframe=" + toString());
	agent.message("   " + localVariables.length + " local vars");
	RemoteStackVariable ret[] =
	    new RemoteStackVariable[localVariables.length];
	for (int i = 0; i < ret.length; i++) {
	    RemoteValue v = agent.getStackValue(thread.getId(), iFrame,
		localVariables[i].slot,
		localVariables[i].signature.charAt(0));
	    ret[i] = new RemoteStackVariable(this,
                                             localVariables[i].slot,
					     localVariables[i].name,
					     localVariables[i].signature, 
                                             localVariables[i].methodArgument,
                                             v);
	}
	return ret;
    }

    /** Return the source file line number. */
    public int getLineNumber() {
	return lineno;
    }
    
    /** Get the method name referenced by this stackframe. */
    public String getMethodName() {
	return methodName;
    }

    /** Get the signature for the method referenced by this stackframe. */   // GVH SAS method signature
    public String getMethodSignature() {                                     // GVH SAS method signature
	return methodSignature;                                                  // GVH SAS method signature
    }                                                                        // GVH SAS method signature

    /** Get the program counter referenced by this stackframe. */
    public int getPC() {
        return bci;
    }

    /** Get the class this stackframe references. */
    public RemoteClass getRemoteClass() {
	return clazz;
    }

    /* Set the class this stackframe references. */
    void setRemoteClass(RemoteClass c) {
	clazz = c;
    }

    /* Sets an int stack variable value. */
    void setVariable(int slot, int value) throws Exception {
        agent.setStackValue(thread.getId(), iFrame, slot, value);
    }

    /* Sets a boolean stack variable value. */
    void setVariable(int slot, boolean value) throws Exception {
        agent.setStackValue(thread.getId(), iFrame, slot, value);
    }

    /* Sets a char stack variable value. */
    void setVariable(int slot, char value) throws Exception {
        agent.setStackValue(thread.getId(), iFrame, slot, value);
    }

    /* Sets a long stack variable value. */
    void setVariable(int slot, long value) throws Exception {
        agent.setStackValue(thread.getId(), iFrame, slot, value);
    }

    /* Sets a float stack variable value. */
    void setVariable(int slot, float value) throws Exception {
        agent.setStackValue(thread.getId(), iFrame, slot, value);
    }

    /* Sets a double stack variable value. */
    void setVariable(int slot, double value) throws Exception {
        agent.setStackValue(thread.getId(), iFrame, slot, value);
    }

    public String toString() {
	if (bci == -1) {
	    // Native method.
	    return className + "." + methodName + " (native method)";
	} else if (lineno == -1) {
	    // No line number information available, use pc.
	    return className + "." + methodName + " (pc " + bci + ")";
	} else {
	    int iPkg = className.lastIndexOf('.');
	    String baseName = (iPkg >= 0) ?
		className.substring(iPkg + 1) : className;
	    return className + "." + methodName +
		" (" + baseName + ":" + Integer.toString(lineno) + ")";
	}
    }
}

