/*
 * @(#)RemoteThread.java	1.19 98/09/21
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.tools.debug;

/**
 * The RemoteThread class allows access to a thread in a remote Java
 * interpreter.  
 *
 * @see RemoteDebugger
 * @see RemoteThreadGroup
 * @author Thomas Ball
 */
public class RemoteThread extends RemoteObject {

    private int currentFrame = 0;
    private RemoteStackFrame stack[];
    private boolean suspended = false;

    RemoteThread(RemoteAgent agent, int id, RemoteClass clazz) {
	super(agent, TP_THREAD, id, clazz);
    }

    /**
     * Return the name of the thread.
     *
     * @exception java.lang.Exception if a general exception occurs.
     */
    public String getName() throws Exception {
	return agent.getThreadName(id);
    }

    /** Return the current stackframe index */
    public int getCurrentFrameIndex() {
	return currentFrame;
    }

    /** Set the current stackframe index */
    public void setCurrentFrameIndex(int iFrame) {
	currentFrame = iFrame;
    }

    /** Reset the current stackframe */
    public void resetCurrentFrameIndex() {
	currentFrame = 0;
    }

    /**
     * Change the current stackframe to be one or more frames higher
     * (as in, away from the current program counter).
     *
     * @param nFrames	the number of stackframes
     * @exception IllegalAccessError when the thread isn't suspended or waiting at a breakpoint
     * @exception ArrayIndexOutOfBoundsException when the requested frame is beyond the stack boundary
     * @exception java.lang.Exception if a general exception occurs.
     */
    public void up(int nFrames) throws Exception {
	assureSuspended();
	if ((currentFrame + nFrames) >= getStack().length) {
	    throw new ArrayIndexOutOfBoundsException();
	}
	currentFrame += nFrames;
    }

    /**
     * Change the current stackframe to be one or more frames lower
     * (as in, toward the current program counter).
     *
     * @param nFrames	the number of stackframes
     * @exception IllegalAccessError when the thread isn't suspended or waiting at a breakpoint
     * @exception ArrayIndexOutOfBoundsException when the requested frame is beyond the stack boundary
     * @exception java.lang.Exception if a general exception occurs.
     */
    public void down(int nFrames) throws Exception {
	assureSuspended();
	if ((currentFrame - nFrames) < 0) {
	    throw new ArrayIndexOutOfBoundsException();
	}
	currentFrame -= nFrames;
    }

    /**
     * Return the thread status description.
     *
     * @exception java.lang.Exception if a general exception occurs.
     */
    public String getStatus() throws Exception {
	int status = agent.getThreadStatus(id);
	switch (status) {
	  case THR_STATUS_UNKNOWN:
	    return "";
	  case THR_STATUS_ZOMBIE:
	    return "zombie";
	  case THR_STATUS_RUNNING:
	    return "running";
	  case THR_STATUS_SLEEPING:
	    return "sleeping";
	  case THR_STATUS_MONWAIT:
	    return "waiting in a monitor";
	  case THR_STATUS_CONDWAIT:
	    return "cond. waiting";
	  case THR_STATUS_SUSPENDED:
	    return "suspended";
	  case THR_STATUS_BREAK:
	    return "at breakpoint";
	  default:
	    return "invalid status returned";
	}
    }

    /**
     * Dump the stack.
     *
     * @exception java.lang.Exception if a general exception occurs.
     */
    public RemoteStackFrame[] dumpStack() throws Exception {
	resetStack();                             // Force stack update GVH SAS # 4027990
	return getStack();                        // GVH SAS # 4027990
    }

    /**
     * Get the current stack frame.
     *
     * @exception IllegalAccessError when the thread isn't suspended or waiting at a breakpoint
     * @exception java.lang.Exception if a general exception occurs.
     */
    public RemoteStackFrame getCurrentFrame() throws Exception {
	assureSuspended();
	return getStack()[currentFrame];     // GVH SAS # 4027990
    }

    /**
     * Suspend execution of this thread.
     *
     * @exception java.lang.Exception if a general exception occurs.
     */
    public void suspend() throws Exception {
	if (suspended == false) {
	    agent.suspendThread(id);
        resetStack();                    // GVH SAS # 4027990, 4066751
	    suspended = true;
	}
    }

    /**
     * Resume execution of this thread.
     *
     * @exception java.lang.Exception if a general exception occurs.
     */
    public void resume() throws Exception {
	if (suspended == true) {
	    agent.resumeThread(id);
	    suspended = false;
	}
    }

    /**
     * Continue execution of this thread to the next instruction or line.
     *
     * @param	skipLine	true to execute to next source line, false to
     * next instruction.
     * @exception IllegalAccessError when the thread isn't suspended or waiting at a breakpoint
     * @exception java.lang.Exception if a general exception occurs.
     */
    public void step(boolean skipLine) throws Exception {
	assureSuspended();
	agent.stepThread(id, skipLine);
    }

    /**
     * Continue execution of this thread to the next line, but don't step
     * into a method call.  If no line information is available, next()
     * is equivalent to step().
     *
     * @exception IllegalAccessError when the thread isn't suspended or waiting at a breakpoint
     * @exception java.lang.Exception if a general exception occurs.
     */
    public void next() throws Exception {
	assureSuspended();
	agent.stepNextThread(id);
    }

    /**                                                                               
     * Continue execution of this thread until it returns. If a return address        
     * is not available, stepOut() is equivalent to step().                           
     * Courtesy of Gordon Hirsch of SAS.
     *                                                                                
     * @exception IllegalAccessError when the thread isn't suspended or waiting at a breakpoint   
     * @exception java.lang.Exception if a general exception occurs.                  
     */                                                                              
    public void stepOut() throws Exception {                               
	assureSuspended();
	agent.stepOutThread(id);                                                
    }                                                                      

    /* Throw IllegalAccessError if not suspended */
    private void assureSuspended() throws Exception {
	int status = agent.getThreadStatus(id);
	if (status != THR_STATUS_SUSPENDED && status != THR_STATUS_BREAK) {
	    throw new IllegalAccessError();
	}
    }

    /* Return whether this thread is suspended. 
     * New version courtesy of Alan Beale of SAS
     */
    public boolean isSuspended() {
	int status;
	try {
	    status = agent.getThreadStatus(id);
	} catch (Exception exc) {
	    throw new IllegalArgumentException("ThreadStatus failed with " + exc);
	}
	return status == THR_STATUS_SUSPENDED || status == THR_STATUS_BREAK;
    }

    /**
     * Resume this thread from a breakpoint, unless it previously suspended.
     *
     * @exception java.lang.Exception if a general exception occurs.
     */
    public void cont() throws Exception {
	if (suspended == false) {
	    agent.resumeThread(id);
	}
    }

    /**
     * Stop the remote thread.
     *
     * @exception java.lang.Exception if a general exception occurs.
     */
    public void stop() throws Exception {
	agent.stopThread(id);
    }

    /**
     * Return a stack variable from the current stackframe.
     *
     * @return the variable as a RemoteValue, or null if not found.
     * @exception java.lang.Exception if a general exception occurs.
     */
    public RemoteStackVariable getStackVariable(String name) throws Exception {
	agent.message("getStackVariable: thread=" + getName() +
		      ", currentFrame=" + currentFrame);
	assureSuspended();
	return getStack()[currentFrame].getLocalVariable(name);  // GVH SAS # 4027990
    }

    /**
     * Return the arguments and local variable from the current stackframe.
     *
     * @return an array of RemoteValues.
     * @exception java.lang.Exception if a general exception occurs.
     */
    public RemoteStackVariable[] getStackVariables() throws Exception {
	agent.message("getStackVariables: thread=" + getName() +
		      ", currentFrame=" + currentFrame);
	assureSuspended();
	return getStack()[currentFrame].getLocalVariables();   // GVH SAS # 4027990
    }

    private synchronized RemoteStackFrame[] getStack() throws Exception // GVH SAS # 4027990
    {                                                      // GVH SAS # 4027990
        if (stack == null)                                 // GVH SAS # 4027990
        {                                                  // GVH SAS # 4027990
            stack = agent.dumpStack(this);                 // GVH SAS # 4027990
        }                                                  // GVH SAS # 4027990
        return stack;                                      // GVH SAS # 4027990
    }                                                      // GVH SAS # 4027990

    synchronized void resetStack()                         // GVH SAS # 4027990
    {                                                      // GVH SAS # 4027990 
        stack = null;                                      // GVH SAS # 4027990
    }                                                      // GVH SAS # 4027990
}
