/*
 * @(#)Main.java	1.46 98/09/15
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.tools.jar;

import java.io.*;
import java.util.*;
import java.util.zip.*;
import java.util.jar.*;
import java.util.jar.Manifest;
import sun.tools.util.CommandLine;

/**
 * This class implements a simple utility for creating files in the JAR
 * (Java Archive) file format. The JAR format is based on the ZIP file
 * format, with optional meta-information stored in a MANIFEST entry.
 */
public
class Main {
    String program;
    PrintStream out, err;
    String fname, mname;
    String zname = "";
    String[] files;
    Hashtable filesTable = new Hashtable();
    Vector paths = new Vector();
    Vector v;
    CRC32 crc32 = new CRC32();
    /* cflag: create
     * uflag: update       
     * xflag: xtract
     * tflag: table
     * vflag: verbose
     * flag0: no zip compression (store only)
     * Mflag: DO NOT generate a manifest file (just ZIP)
     */
    boolean cflag, uflag, xflag, tflag, vflag, flag0, Mflag;

    static final String MANIFEST = JarFile.MANIFEST_NAME;
    static final String MANIFEST_DIR = "META-INF/";
    static final String VERSION = "1.0";
    static final char SEPARATOR = File.separatorChar;

    public Main(PrintStream out, PrintStream err, String program) {
	this.out = out;
	this.err = err;
	this.program = program;
    }

    /*
     * Starts main program with the specified arguments.
     */
    public synchronized boolean run(String args[]) {
	boolean ok = true;
        if (!parseArgs(args)) {
	    return false;
	}
	try {
	    if (cflag || uflag) {
                if (fname != null) {
                    // The name of the zip file as it would appear as it's own
                    // zip file entry. We use this to make sure that we don't
                    // add the zip file to itself.
                    zname = fname.replace(File.separatorChar, '/');
                    if (zname.startsWith("./")) {
                        zname = zname.substring(2);
                    }
		}
            }
            if (cflag) {
                Manifest manifest = null;
                InputStream in = null;
		if (!Mflag) {
		    if (mname != null) {
			in = new FileInputStream(mname);
			manifest = new Manifest(new BufferedInputStream(in));
		    } else {
			manifest = new Manifest();
                    }
                    addVersion(manifest);
		    addCreatedBy(manifest);
                }
                OutputStream out;
                if (fname != null) {
                    out = new FileOutputStream(fname);
                } else {
                    out = new FileOutputStream(FileDescriptor.out);
                    if (vflag) {
                        // Disable verbose output so that it does not appear
                        // on stdout along with file data
                        // error("Warning: -v option ignored");
                        vflag = false;
                    }
                }
                create(new BufferedOutputStream(out), expand(files), manifest);
                if (in != null) {
                    in.close();
                }
                out.close();
            } else if (uflag) {
                File inputFile = null, tmpFile = null;
                FileInputStream in;
                FileOutputStream out;
                if (fname != null) {
                    inputFile = new File(fname);
                    String path = inputFile.getParent();
                    tmpFile = File.createTempFile("tmp", null,
                              new File((path == null) ? "." : path));
                    in = new FileInputStream(inputFile);
                    out = new FileOutputStream(tmpFile);
                } else {
                    in = new FileInputStream(FileDescriptor.in);
                    out = new FileOutputStream(FileDescriptor.out);
                    vflag = false;
                }
                InputStream manifest = (!Mflag && (mname != null)) ?
                    (new FileInputStream(mname)) : null;
                expand(files);
                update(in, new BufferedOutputStream(out), manifest);
                in.close();
                out.close();
                if (manifest != null) {
                    manifest.close();
                }
                if (fname != null) {
                    // on Win32, we need this delete
                    inputFile.delete();
                    if(!tmpFile.renameTo(inputFile)) {
                        tmpFile.delete();
                        throw new IOException("Error in writing existing jar file");
                    }
                    tmpFile.delete();
                }
            } else if (xflag || tflag) {
		InputStream in;
		if (fname != null) {
		    in = new FileInputStream(fname);
		} else {
		    in = new FileInputStream(FileDescriptor.in);
		}
		if (xflag) {
		    extract(new BufferedInputStream(in), files);
		} else {
		    list(new BufferedInputStream(in), files);
		}
                in.close();
	    }
	} catch (IOException e) {
	    fatalError(e);
	    ok = false;
	} catch (Error ee) {
	    ee.printStackTrace();
	    ok = false;
	} catch (Throwable t) {
	    t.printStackTrace();
	    ok = false;
	}	
	out.flush();
	err.flush();
	return ok;
    }

    /*
     * Parse command line arguments.
     */
    boolean parseArgs(String args[]) {
	/* Preprocess and expand @file arguments */
	try {
	    args = CommandLine.parse(args);
	} catch (FileNotFoundException e) {
	    fatalError("can't open: " + e.getMessage());
	    return false;
	} catch (IOException e) {
	    fatalError(e);
	    return false;
	}
	/* parse flags */
	int count = 1;
	try {
	    String flags = args[0];
	    if (flags.startsWith("-")) {
		flags = flags.substring(1);
	    }
	    for (int i = 0; i < flags.length(); i++) {
		switch (flags.charAt(i)) {
		case 'c':
		    if (xflag || tflag || uflag) {
			usageError();
			return false;
		    }
		    cflag = true;
		    break;
                case 'u':
		    if (cflag || xflag || tflag) {
			usageError();
			return false;
		    }
                    uflag = true;
                    break;
		case 'x':
		    if (cflag || uflag || tflag) {
			usageError();
			return false;
		    }
		    xflag = true;
		    break;
		case 't':
		    if (cflag || uflag || xflag) {
			usageError();
			return false;
		    }
		    tflag = true;
		    break;
		case 'M':
		    Mflag = true;
		    break;
		case 'v':
		    vflag = true;
		    break;
		case 'f':
		    fname = args[count++];
		    break;
		case 'm':
		    mname = args[count++];
		    break;
		case '0':
		    flag0 = true;
		    break;
		default:
		    error("Illegal option: " + flags.charAt(i));
		    usageError();
		    return false;
		}
	    }
	} catch (ArrayIndexOutOfBoundsException e) {
	    usageError();
	    return false;
	}
	if (!cflag && !tflag && !xflag && !uflag) {
	    error("One of options -{ctxu} must be specified.");
	    usageError();
	    return false;
	}
	/* parse file arguments */
        int n = args.length - count;
	if (n > 0) {
            int k = 0;
	    String[] nameBuf = new String[n];
            try {
                for (int i = count; i < args.length; i++) {
                    if(args[i].equals("-C")) {
                        /* change the directory */
                        String dir = args[++i];
                        dir = (dir.endsWith(File.separator) ? 
                               dir : (dir + File.separator));
                        paths.addElement(dir.replace(File.separatorChar, '/'));
                        nameBuf[k++] = dir + args[++i];
                    } else {
                        nameBuf[k++] = args[i];
                    }
                }
            } catch (ArrayIndexOutOfBoundsException e) {
                usageError();
                return false;
            }
            files = new String[k];
            System.arraycopy(nameBuf, 0, files, 0, k);
        } else if (cflag) {
	    error("'c' flag requires that input files be specified!");
	    usageError();
	    return false;
	} else if (uflag) {
            if (mname != null) {
                /* just want to update the manifest */
                return true;
            } else {
                error("'u' flag requires manifest or input files be specified!");
                usageError();
                return false;
            }
	}
	return true;
    }

    /*
     * Expands list of files to process into full list of all files that
     * can be found by recursively descending directories.
     */
    String[] expand(String[] files) {
	v = new Vector();
	expand(null, files, v, filesTable);
	files = new String[v.size()];
	for (int i = 0; i < files.length; i++) {
	    files[i] = ((File)v.elementAt(i)).getPath();
	}
	return files;
    }

    void expand(File dir, String[] files, Vector v, Hashtable t) {
	if (files == null) {
	    return;
	}
	for (int i = 0; i < files.length; i++) {
	    File f;
	    if (dir == null) {
		f = new File(files[i]);
	    } else {
		f = new File(dir, files[i]);
	    }
	    if (f.isFile()) {
		if (!t.contains(f)) {
		    t.put(entryName(f.getPath()), f);
		    v.addElement(f);
		}
	    } else if (f.isDirectory()) {
                String dirPath = f.getPath();
                dirPath = (dirPath.endsWith(File.separator)) ? dirPath :
                    (dirPath + File.separator);
                t.put(entryName(dirPath), f);
		v.addElement(f);
		expand(f, f.list(), v, t);
	    } else {
		error(f + ": no such file or directory");
	    }
	}
    }

    /*
     * Creates a new JAR file.
     */
    void create(OutputStream out, String[] files, Manifest manifest)
	throws IOException
    {
	ZipOutputStream zos = new JarOutputStream(out);
	if (flag0) {
	    zos.setMethod(ZipOutputStream.STORED);
	}
	if (manifest != null) {
            if (vflag) {
                output("added manifest");
            }
            ZipEntry e = new ZipEntry(MANIFEST_DIR);
	    e.setTime(System.currentTimeMillis());
            e.setSize(0);
            e.setCrc(0);
            zos.putNextEntry(e);
            e = new ZipEntry(MANIFEST);
	    e.setTime(System.currentTimeMillis());
	    if (flag0) {
		crc32Manifest(e, manifest);
	    }
	    zos.putNextEntry(e);
	    manifest.write(zos);
	    zos.closeEntry();
	}
	for (int i = 0; i < files.length; i++) {
	    addFile(zos, new File(files[i]));
	}
	zos.close();
    }

    /*
     * update an existing jar file.
     */
    void update(InputStream in, OutputStream out,
                InputStream manifest) throws IOException 
    {
        Hashtable t = filesTable;
        Vector v = this.v;
        ZipInputStream zis = new ZipInputStream(in);
        ZipOutputStream zos = new JarOutputStream(out);
        ZipEntry e = null;
        boolean foundManifest = false;
        
        byte[] buf = new byte[1024];
        int n = 0;
        // put the old entries first, replace if necessary
        while((e = zis.getNextEntry()) != null) {
            String name = e.getName();
            if (Mflag && name.toUpperCase().equals(MANIFEST)) {
                continue;
            } else if (manifest != null 
                       && name.toUpperCase().equals(MANIFEST)) {
                // need to update the manifest
                foundManifest = true;
                Manifest old = new Manifest(zis);
                old.read(manifest);
                updateManifest(old, zos);
            } else {
                if(!t.containsKey(name)) { // copy the old stuff
                    // do our own compression
                    ZipEntry e2 = new ZipEntry(name);
                    e2.setMethod(e.getMethod());
                    e2.setTime(e.getTime());
                    e2.setComment(e.getComment());
                    e2.setExtra(e.getExtra());
                    if (e.getMethod() == ZipEntry.STORED) {
                        e2.setSize(e.getSize());
                        e2.setCrc(e.getCrc());
                    }
                    zos.putNextEntry(e2);
                    while((n = zis.read(buf, 0, buf.length)) != -1) {
                        zos.write(buf, 0, n);
                    }
                } else { // replace with the new files
                    addFile(zos, (File)(t.get(name)));
                    t.remove(name);
                }
            }
        }

        // add the remaining new files
        if (!t.isEmpty()) {
            for (int i = 0; i < v.size(); i++) {
                File f = (File)v.elementAt(i);
                if (t.containsValue(f)) {
                    addFile(zos, f);
                }
            }
        }
        
        if (!foundManifest && (manifest != null)) {
            updateManifest(new Manifest(manifest), zos);
        }
        
        zos.close();
    }
    
    private void updateManifest(Manifest m, ZipOutputStream zos) 
        throws IOException
    {
        addVersion(m);
	addCreatedBy(m);
        ZipEntry e = new ZipEntry(MANIFEST);
        e.setTime(System.currentTimeMillis());
        if (flag0) {
            e.setMethod(ZipEntry.STORED);
            crc32Manifest(e, m);
        }
        zos.putNextEntry(e);
        m.write(zos);
        if (vflag) {
            output("updated manifest");
        }
    }
    
    private String entryName(String name) {
        name = name.replace(File.separatorChar, '/');
        String matchPath = "";
        for(int i = 0; i < paths.size(); i++) {
            String path = (String)paths.elementAt(i);
            if (name.startsWith(path) && (path.length() > matchPath.length())) {
                matchPath = path;
            }
        }
        name = name.substring(matchPath.length());

	if (name.startsWith("/")) {
	    name = name.substring(1);
	} else if (name.startsWith("./")) {
	    name = name.substring(2);
	}
	return name;
    }

    private void addVersion(Manifest m) {
        Attributes global = m.getMainAttributes();
        if (global.getValue(Attributes.Name.MANIFEST_VERSION) == null) {
            global.put(Attributes.Name.MANIFEST_VERSION, VERSION);
        }
    }

    private void addCreatedBy(Manifest m) {
        Attributes global = m.getMainAttributes();
        if (global.getValue(new Attributes.Name("Created-By")) == null) {
	    String javaVendor = System.getProperty("java.vendor");
	    String jdkVersion = System.getProperty("java.version");
            global.put(new Attributes.Name("Created-By"), jdkVersion + " (" +
		       javaVendor + ")");
        }
    }

    /*
     * Adds a new file entry to the ZIP output stream.
     */
    void addFile(ZipOutputStream zos, File file) throws IOException {
        String name = file.getPath();
	boolean isDir = file.isDirectory();

        if (isDir) {
            name = name.endsWith(File.separator) ? name :
                (name + File.separator);
        }
        name = entryName(name);

	if (name.equals("") || name.equals(".") || name.equals(zname)) {
	    return;
        } else if ((name.equals(MANIFEST_DIR) || name.equals(MANIFEST))
                   && !Mflag) {
            if (vflag) {
                output("ignoring entry " + name);
            }
            return;
        }

	long size = isDir ? 0 : file.length();

	if (vflag) {
	    out.print("adding: " + name + " ");
	}
	ZipEntry e = new ZipEntry(name);
	e.setTime(file.lastModified());
	if (size == 0) {
	    e.setMethod(ZipEntry.STORED);
	    e.setSize(0);
	    e.setCrc(0);
	} else if (flag0) {
	    e.setSize(size);
            e.setMethod(ZipEntry.STORED);
	    crc32File(e, file);
	}
	zos.putNextEntry(e);
	if (!isDir) {
	    byte[] buf = new byte[1024];
	    int len;
	    InputStream is = new BufferedInputStream(new FileInputStream(file));
	    while ((len = is.read(buf, 0, buf.length)) != -1) {
		zos.write(buf, 0, len);
	    }
	    is.close();
	}
	zos.closeEntry();
	/* report how much compression occurred. */
	if (vflag) {
	    size = e.getSize();
	    long csize = e.getCompressedSize();
	    out.print("(in=" + size + ") (out=" + csize + ") ");
	    if (e.getMethod() == ZipEntry.DEFLATED) {
		long ratio = 0;
		if (size != 0) {
		    ratio = ((size - csize) * 100) / size;
		}
		output("(deflated " + ratio + "%)");
	    } else {
		output("(stored 0%)");
	    }
	}
    }

    /* compute the crc32 of a file.  This is necessary when the ZipOutputStream
     * is in STORED mode.
     */
    private void crc32Manifest(ZipEntry e, Manifest m) throws IOException {
	crc32.reset();
	CRC32OutputStream os = new CRC32OutputStream(crc32);
	m.write(os);
	e.setSize((long) os.n);
	e.setCrc(crc32.getValue());
    }
	

    /* compute the crc32 of a file.  This is necessary when the ZipOutputStream
     * is in STORED mode.
     */
    private void crc32File(ZipEntry e, File f) throws IOException {
	InputStream is = new BufferedInputStream(new FileInputStream(f));
	byte[] buf = new byte[1024];
	crc32.reset();
	int r = 0;
	int nread = 0;
	long len = f.length();
	while ((r = is.read(buf)) != -1) {
	    nread += r;
	    crc32.update(buf, 0, r);
	}
	is.close();
	if (nread != (int) len) {
	    throw new JarException("incorrect length while processing: " + f.getPath());
	}
	e.setCrc(crc32.getValue());
    }

    /*
     * Extracts specified entries from JAR file.
     */
    void extract(InputStream in, String files[]) throws IOException {
	ZipInputStream zis = new ZipInputStream(in);
	ZipEntry e;
	while ((e = zis.getNextEntry()) != null) {
	    if (files == null) {
		extractFile(zis, e);
	    } else {
		String name = e.getName();
		for (int i = 0; i < files.length; i++) {
                    String file = files[i].replace(File.separatorChar, '/');
		    if (name.startsWith(file)) {
			extractFile(zis, e);
			break;
		    }
		}
	    }
	}
    }

    /*
     * Extracts next entry from JAR file, creating directories as needed.
     */
    void extractFile(ZipInputStream zis, ZipEntry e) throws IOException {
        String name = e.getName();
	File f = new File(e.getName().replace('/', File.separatorChar));
	if (e.isDirectory()) {
	    if (!f.exists() && !f.mkdirs() || !f.isDirectory()) {
		throw new IOException(f + ": could not create directory");
	    }
	    if (vflag) {
		output("   created: " + name);
	    }
	} else {
	    if (f.getParent() != null) {
		File d = new File(f.getParent());
		if (!d.exists() && !d.mkdirs() || !d.isDirectory()) {
		    throw new IOException(d + ": could not create directory");
		}
	    }
	    OutputStream os = new FileOutputStream(f);
	    byte[] b = new byte[512];
	    int len;
	    while ((len = zis.read(b, 0, b.length)) != -1) {
		os.write(b, 0, len);
	    }
	    zis.closeEntry();
	    os.close();
	    if (vflag) {
		if (e.getMethod() == ZipEntry.DEFLATED) {
		    output(" extracted: " + name);
		} else {
		    output("  inflated: " + name);
		}
	    }
	}
    }

    /*
     * Lists contents of JAR file.
     */
    void list(InputStream in, String files[]) throws IOException {
	ZipInputStream zis = new ZipInputStream(in);
	ZipEntry e;
	while ((e = zis.getNextEntry()) != null) {
	    String name = e.getName();
	    /*
	     * In the case of a compressed (deflated) entry, the entry size
	     * is stored immediately following the entry data and cannot be
	     * determined until the entry is fully read. Therefore, we close
	     * the entry first before printing out its attributes.
	     */
	    zis.closeEntry();
	    if (files == null) {
		printEntry(e);
	    } else {
                for (int i = 0; i < files.length; i++) {
                    String file = files[i].replace(File.separatorChar, '/');
		    if (name.startsWith(file)) {
			printEntry(e);
			break;
		    }
		}
	    }
	}
    }

    /*
     * Prints entry information.
     */
    void printEntry(ZipEntry e) throws IOException {
	if (vflag) {
	    StringBuffer sb = new StringBuffer();
	    String s = Long.toString(e.getSize());
	    for (int i = 6 - s.length(); i > 0; --i) {
		sb.append(' ');
	    }
	    sb.append(s).append(' ').append(new Date(e.getTime()).toString());
	    sb.append(' ').append(e.getName());
	    output(sb.toString());
	} else {
	    output(e.getName());
	}
    }

    /*
     * Print usage message and die.
     */
    void usageError() {
	error("Usage: jar {ctxu}[vfm0M] [jar-file] [manifest-file] [-C dir] files ...");
	error("Options:");
	error("  -c  create new archive");
	error("  -t  list table of contents for archive"); 
	error("  -x  extract named (or all) files from archive");
	error("  -u  update existing archive");
	error("  -v  generate verbose output on standard output");
	error("  -f  specify archive file name");
	error("  -m  include manifest information from specified manifest file");
	error("  -0  store only; use no ZIP compression");
	error("  -M  Do not create a manifest file for the entries");
        error("  -C  change to the specified directory and include the following file");
	error("If any file is a directory then it is processed recursively.");
	error("The manifest file name and the archive file name needs to be specified");
        error("in the same order the 'm' and 'f' flags are specified.\n");
	error("Example 1: to archive two class files into an archive called classes.jar: ");
	error("     jar cvf classes.jar Foo.class Bar.class ");
	error("Example 2: use an existing manifest file 'mymanifest' and archive all the");
        error("         files in the foo/ directory into 'classes.jar': ");
	error("     jar cvfm classes.jar mymanifest -C foo/ .");
    }

    /*
     * A fatal exception has been caught.  No recovery possible
     */
    void fatalError(Exception e) {
	e.printStackTrace();
    }

    /*
     * A fatal condition has been detected; message is "s".  No recovery possible
     */
    void fatalError(String s) {
	error(program + ": " + s);
    }

    /**
     * Print an output message; like verbose output and the like
     */
    protected void output(String s) {
	out.println(s);
    }

    /**
     * Print an error mesage; like something is broken
     */
    protected void error(String s) {
	err.println(s);
    }

    /*
     * Main routine to start program.
     */
    public static void main(String args[]) {
	Main jartool = new Main(System.out, System.err, "jar");
	System.exit(jartool.run(args) ? 0 : 1);
    }
}

/* an OutputStream that doesn't send it's output anywhere, (but could).
 * It's here to find the CRC32 of a manifest, necessary for STORED only
 * mode in ZIP.
 */
final class CRC32OutputStream extends java.io.OutputStream {
	
    CRC32 crc;
    int n = 0;
    CRC32OutputStream(CRC32 crc) {
	this.crc = crc;
    }

    public void write(int r) throws IOException {
	crc.update(r);
	n++;
    }

    public void write(byte[] b) throws IOException {
	crc.update(b, 0, b.length);
	n += b.length;
    }

    public void write(byte[] b, int off, int len) throws IOException {
	crc.update(b, off, len);
	n += len - off;
    }
}
