/*
 * @(#)CompoundStatement.java	1.27 98/03/18
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.tools.tree;

import sun.tools.java.*;
import sun.tools.asm.Assembler;
import sun.tools.asm.Label;
import java.io.PrintStream;
import java.util.Hashtable;

public
class CompoundStatement extends Statement {
    Statement args[];

    /**
     * Constructor
     */
    public CompoundStatement(long where, Statement args[]) {
	super(STAT, where);
	this.args = args;
	// To avoid the need for subsequent null checks:
	for (int i = 0 ; i < args.length ; i++) {
	    if (args[i] == null) {
		args[i] = new CompoundStatement(where, new Statement[0]);
	    }
	}
    }

    /**
     * Insert a new statement at the front.
     * This is used to introduce an implicit super-class constructor call.
     */
    public void insertStatement(Statement s) {
	Statement newargs[] = new Statement[1+args.length];
	newargs[0] = s;
	for (int i = 0 ; i < args.length ; i++) {
	    newargs[i+1] = args[i];
	}
	this.args = newargs;
    }

    /**
     * Check statement
     */
    Vset check(Environment env, Context ctx, Vset vset, Hashtable exp) {
	checkLabel(env, ctx);
	if (args.length > 0) {
	    vset = reach(env, vset);
	    CheckContext newctx = new CheckContext(ctx, this);
	    // In this environment, 'resolveName' will look for local classes.
	    Environment newenv = Context.newEnvironment(env, newctx);
	    for (int i = 0 ; i < args.length ; i++) {
		vset = args[i].checkBlockStatement(newenv, newctx, vset, exp);
	    }
	    vset = vset.join(newctx.vsBreak);
	}
	return ctx.removeAdditionalVars(vset);
    }

    /**
     * Inline
     */
    public Statement inline(Environment env, Context ctx) {
	ctx = new Context(ctx, this);
	boolean expand = false;
	int count = 0;
	for (int i = 0 ; i < args.length ; i++) {
	    Statement s = args[i];
	    if (s != null) {
		if ((s = s.inline(env, ctx)) != null) {
		    if ((s.op == STAT) && (s.labels == null)) {
			count += ((CompoundStatement)s).args.length;
		    } else {
			count++;
		    }
		    expand = true;
		}
		args[i] = s;
	    }
	}
	switch (count) {
	  case 0:
	    return null;
	    
	  case 1:
	    for (int i = args.length ; i-- > 0 ;) {
		if (args[i] != null) {
		    return eliminate(env, args[i]);
		}
	    }
	    break;
	}
	if (expand || (count != args.length)) {
	    Statement newArgs[] = new Statement[count];
	    for (int i = args.length ; i-- > 0 ;) {
		Statement s = args[i];
		if (s != null) {
		    if ((s.op == STAT) && (s.labels == null)) {
			Statement a[] = ((CompoundStatement)s).args;
			for (int j = a.length ; j-- > 0 ; ) {
			    newArgs[--count] = a[j];
			}
		    } else {
			newArgs[--count] = s;
		    }
		}
	    }
	    args = newArgs;
	}
	return this;
    }

    /**
     * Create a copy of the statement for method inlining
     */
    public Statement copyInline(Context ctx, boolean valNeeded) {
	CompoundStatement s = (CompoundStatement)clone();
	s.args = new Statement[args.length];
	for (int i = 0 ; i < args.length ; i++) {
	    s.args[i] = args[i].copyInline(ctx, valNeeded);
	}
	return s;
    }

    /**
     * The cost of inlining this statement
     */
    public int costInline(int thresh, Environment env, Context ctx) {
	int cost = 0;
	for (int i = 0 ; (i < args.length) && (cost < thresh) ; i++) {
	    cost += args[i].costInline(thresh, env, ctx);
	}
	return cost;
    }

    /**
     * Code
     */
    public void code(Environment env, Context ctx, Assembler asm) {
	CodeContext newctx = new CodeContext(ctx, this);
	for (int i = 0 ; i < args.length ; i++) {
	    args[i].code(env, newctx, asm);
	}
	asm.add(newctx.breakLabel);
    }

    /**
     * Check if the first thing is a constructor invocation
     */
    public Expression firstConstructor() {
	return (args.length > 0) ? args[0].firstConstructor() : null;
    }

    /**
     * Print
     */
    public void print(PrintStream out, int indent) {
	super.print(out, indent);
	out.print("{\n");
	for (int i = 0 ; i < args.length ; i++) {
	    printIndent(out, indent+1);
	    if (args[i] != null) {
		args[i].print(out, indent + 1);
	    } else {
		out.print("<empty>");
	    }
	    out.print("\n");
	}
	printIndent(out, indent);
	out.print("}");
    }
}
