/*
 * @(#)bag.c	1.1 98/08/25
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/* General routines for manipulating a bag data structure */

#include <string.h>
#include "sys_api.h"
#include "bag.h"

struct bag {
    void *items;    /* hold items in bag, must align on itemSize */
    int used;       /* number of items in bag */
    int allocated;  /* space reserved */
    int itemSize;   /* size of each item, should init to sizeof item */
};

struct bag * 
bagCreateBag(int itemSize, int initialAllocation) {   
    struct bag *theBag = (struct bag *)sysMalloc(sizeof(struct bag));
    if (theBag == NULL) {
        return NULL;
    }
    itemSize = (itemSize + 7) & ~7;    /* fit 8 byte boundary */
    theBag->items = sysMalloc(initialAllocation * itemSize);
    if (theBag->items == NULL) {
        sysFree(theBag);
        return NULL;
    }
    theBag->used = 0;
    theBag->allocated = initialAllocation;
    theBag->itemSize = itemSize;
    return theBag;
}

void
bagDestroyBag(struct bag *theBag) 
{
    sysFree(theBag->items);
    sysFree(theBag);
}

void *
bagFind(struct bag *theBag, void *key) 
{
    char *items = theBag->items;
    int itemSize = theBag->itemSize;
    char *itemsEnd = items + (itemSize * theBag->used);
    
    for (; items < itemsEnd; items += itemSize) {
        if (*((void**)items) == key) {
            return items;
        }
    }
    return NULL;
}

void *
bagAdd(struct bag *theBag) 
{
    int allocated = theBag->allocated;
    int itemSize = theBag->itemSize;
    void *items = theBag->items;
    void *ret;

    /* if there are no unused slots reallocate */
    if (theBag->used >= allocated) {
        allocated *= 2;
        items = sysRealloc(items, allocated * itemSize);
        if (items == NULL) {
            return NULL;
        }
        theBag->allocated = allocated;
        theBag->items = items;
    }
    ret = ((char *)items) + (itemSize * (theBag->used)++);
    memset(ret, 0, itemSize);
    return ret;
}
    
void
bagDelete(struct bag *theBag, void *condemned) 
{
    int used = --(theBag->used);
    int itemSize = theBag->itemSize;
    void *items = theBag->items;
    void *tailItem = ((char *)items) + (used * itemSize);
    
    if (condemned != tailItem) {
        memcpy(condemned, tailItem, itemSize);
    }
}

void
bagDeleteAll(struct bag *theBag) 
{
    theBag->used = 0;
}

    
int
bagSize(struct bag *theBag) 
{
    return theBag->used;
}    

bool_t
bagEnumerateOver(struct bag *theBag, bagEnumerateFunction func, void *arg) 
{
    char *items = theBag->items;
    int itemSize = theBag->itemSize;
    char *itemsEnd = items + (itemSize * theBag->used);
    
    for (; items < itemsEnd; items += itemSize) {
        if (!(func)((void *)items, arg)) {
            return FALSE;
        }
    }
    return TRUE;
}
