/*
 * @(#)check_class.c	1.56 98/08/14
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*-
 *      code for verifying the date in a ClassClass structure for internal
 *      consistency.
 */

#include <ctype.h>

#include "oobj.h"
#include "interpreter.h"
#include "util.h"
#include "bool.h"
#include "tree.h"
#include "sys_api.h"
#include "common_exceptions.h"

static void CCerror (char *format, ...);

/*
 * Return address of the verifier function or 0 if you can't find it
 */
static verifier_fn_t getVerifier(void)
{
    static verifier_fn_t verifier_fn;
    if (!verifier_fn) {
	verifier_fn = (verifier_fn_t)FindBuiltinEntry("VerifyClassCodes");
    }
    return verifier_fn;
}

void
VerifyClass(ClassClass *cb)
{
    int i;
    if (CCIs(cb, Verified)) 
	return;
    /* Make sure we are not overriding any final methods or classes*/
    if (cbIsInterface(cb)) { 
	struct methodblock *mb;
	if ((cbSuperclass(cb) == NULL) ||
	    (cbSuperclass(cb) != classJavaLangObject)) { 
	    CCerror("Interface %s has bad superclass", cbName(cb));
	    return;
	}
	for (i = cbMethodsCount(cb), mb = cbMethods(cb); --i >= 0; mb++) {
	    if (mb->fb.access & ACC_STATIC) {
		if (mb->fb.name != utf8_literal_cls_init_name) { 
		    /* Only internal methods can be static */
		    CCerror("Illegal static method %s in interface %s",
			    mb->fb.name, cbName(cb));
		    return;
		}
	    }
	}
    } else if (cbSuperclass(cb)) { 
	/*EMPTY*/
        /* Ensuring that final classes are not subclassed, and that
	 * final methods are not overridden.
	 * DEFERRED into PrepareMethods
	 */
    } else if (cb != classJavaLangObject) {
	CCerror("Class %s does not have superclass", cbName(cb));
	return;
    }
	
    {
        bool_t result;
        char message[MSG_BUF_LEN];
	ExecEnv *ee = EE();
        JNIEnv *env = EE2JNIEnv(ee);
	verifier_fn_t verifier_fn = getVerifier();
	if (verifier_fn == 0) {
	    CCerror("Could not link verifier");
	    return;
 	}
	message[0] = 0;
        if ((*env)->PushLocalFrame(env, 100) < 0) {
	    return;	    /* out of memory */
	}
	result = (*verifier_fn)(env, MkRefLocal(env, cb), message,
				sizeof(message));
	(*env)->PopLocalFrame(env, 0);
	if (!result && !exceptionOccurred(ee)) {
	    ThrowVerifyError(0, message);
	    return;
	}
    }
    CCSet(cb, Verified);
    return;
}

static void 
CCerror (char *format, ...)
{
    char buf[MSG_BUF_LEN];
    va_list args;
    va_start(args, format);
    jio_vsnprintf(buf, sizeof(buf), format, args);        
    va_end(args);
    ThrowVerifyError(0, buf);
}
