/*
 * @(#)check_jni.c	1.3 98/08/18
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/* This file contains the additional JNI sanity checks enabled by
 * -Xcheck:jni
 */
#ifdef CHECK_JNI
#include "oobj.h"
#include "interpreter.h"
#include "tree.h"
#include "java_lang_String.h"

#define UNCHECKED() (&unchecked_jni_NativeInterface)

static const char *
jnienv_msg = "Using JNIEnv in the wrong thread";

static const char *
critical_msg = "Calling other JNI functions in the scope of "
    "Get/ReleasePrimitiveArrayCritical or Get/ReleaseStringCritical";

#define CHECK_ENV(env) \
    ((env != EE2JNIEnv(EE())) ? (UNCHECKED()->FatalError(env, jnienv_msg), 1) : 0)

#define CHECK_CRITICAL(env) \
    ((JNIEnv2EE(env)->critical_count) ? \
     (UNCHECKED()->FatalError(env, critical_msg), 1) : 0)

#define RECORD_STACK_BASE_CRITICAL(env, base) \
    void *_stack_base = JNIEnv2EE(env)->stack_base; \
    int _x = (_stack_base == NULL) ? \
                ((JNIEnv2EE(env)->stack_base = (void *)base), 0) : \
                1; \
    int _y = CHECK_ENV(env)

#define RECORD_STACK_BASE(env, base) \
    RECORD_STACK_BASE_CRITICAL(env, base); \
    int _z = CHECK_CRITICAL(env)

#define RECOVER_STACK_BASE(env) \
    JNIEnv2EE(env)->stack_base = _stack_base

static const char *
static_field_msg = "Wrong static field ID passed to JNI";

static const char *
instance_field_msg = "Wrong field ID passed to JNI";

static const char *
field_type_msg = "Type mismatch in JNI get/set field operations";

#define CHECK_STATIC_FIELD_ID(env, fb, cls) \
    if (1) { \
        ValidateClass(env, cls); \
        if (!(fb->access & ACC_STATIC) || \
            !is_subclass_of((ClassClass *)DeRef(env, cls),\
                            (fb)->clazz, JNIEnv2EE(env))) \
            UNCHECKED()->FatalError(env, static_field_msg); \
    } else ((void) 0)

#define CHECK_INSTANCE_FIELD_ID(env, fb, obj) \
    if (1) { \
        ValidateObject(env, obj); \
        if ((fb->access & ACC_STATIC) || \
            !is_subclass_of(obj_classblock(DeRef(env, obj)),\
                            (fb)->clazz, JNIEnv2EE(env))) \
            UNCHECKED()->FatalError(env, instance_field_msg); \
    } else ((void) 0)

#define CHECK_FIELD_TYPE(env, c) \
    if (!(c)) \
        UNCHECKED()->FatalError(env, field_type_msg)

static const char *str_msg1 = "JNI received a null string";
static const char *str_msg2 = "JNI string operation received a non-string";

#define CHECK_STRING(js) \
if (1) { \
    struct Hjava_lang_String *s; \
    ValidateObject(env, js); \
    s = (struct Hjava_lang_String*)DeRef(env, js); \
    if (s == NULL) { \
        UNCHECKED()->FatalError(env, str_msg1); \
    } \
    if (obj_classblock((JHandle *)s) != classJavaLangString) { \
        UNCHECKED()->FatalError(env, str_msg2); \
    } \
} else ((void) 0)

static const char *array_msg1 = "Null array passed to JNI array operations";
static const char *array_msg2 = "Array element type mismatch in JNI";

#define CHECK_ARRAY(jarray,elementType) \
if (1) { \
    HArrayOfByte *arr; \
    ValidateObject(env,jarray); \
    arr = (HArrayOfByte *)DeRef(env, jarray); \
    if (arr == NULL) \
        UNCHECKED()->FatalError(env, array_msg1); \
    if (elementType != -1 && obj_flags(arr) != elementType) \
        UNCHECKED()->FatalError(env, array_msg2); \
} else ((void) 0)

static void
ValidateObject(JNIEnv *env, jobject obj)
{
    if (obj && !isValidHandle(*(JHandle **)obj)) {
        UNCHECKED()->FatalError(env,
	    "Bad global or local ref passed to JNI");
    }
}

static void
ValidateClass(JNIEnv *env, jclass clazz)
{
    ClassClass *cb;

    ValidateObject(env, clazz);
    cb = (ClassClass *)DeRef(env, clazz);
    if (cb == NULL) {
        UNCHECKED()->FatalError(env, "JNI received a null class");
    }
    if (cb->methods && /* happens during bootstrap */
	(obj_classblock((JHandle *)cb) != classJavaLangClass)) {
        /* We do not expect jni_FatalError to return. */
        UNCHECKED()->FatalError(env,
			"JNI received a class argument that is not a class");
    }
}

static jint JNICALL
checked_jni_GetVersion(JNIEnv *env)
{
    RECORD_STACK_BASE(env, &env);
    jint result = UNCHECKED()->GetVersion(env);
    RECOVER_STACK_BASE(env);
    return result;
}

static jclass JNICALL
checked_jni_DefineClass(JNIEnv *env, const char *name, jobject loader,
		   const jbyte *buf, jsize len)
{
    RECORD_STACK_BASE(env, &env);
    jclass result;
    ValidateObject(env, loader);
    result = UNCHECKED()->DefineClass(env, name, loader, buf, len);
    RECOVER_STACK_BASE(env);
    return result;
}

static jclass JNICALL
checked_jni_FindClass(JNIEnv *env, const char *name)
{
    RECORD_STACK_BASE(env, &env);
    jclass result = UNCHECKED()->FindClass(env, name);
    RECOVER_STACK_BASE(env);
    return result;
}

static jmethodID JNICALL
checked_jni_FromReflectedMethod(JNIEnv *env, jobject method)
{
    RECORD_STACK_BASE(env, &env);
    jmethodID result;
    ValidateObject(env, method);
    result = UNCHECKED()->FromReflectedMethod(env, method);
    RECOVER_STACK_BASE(env);
    return result;
}

static jfieldID JNICALL
checked_jni_FromReflectedField(JNIEnv *env, jobject field)
{
    RECORD_STACK_BASE(env, &env);
    jfieldID result;
    ValidateObject(env, field);
    result = UNCHECKED()->FromReflectedField(env, field);
    RECOVER_STACK_BASE(env);
    return result;
}

static jobject JNICALL
checked_jni_ToReflectedMethod(JNIEnv *env, jclass cls, jmethodID methodID,
			      jboolean isStatic)
{
    RECORD_STACK_BASE(env, &env);
    jobject result;
    ValidateClass(env, cls);
    result = UNCHECKED()->ToReflectedMethod(env, cls, methodID, isStatic);
    RECOVER_STACK_BASE(env);
    return result;
}

static jclass JNICALL
checked_jni_GetSuperclass(JNIEnv *env, jclass sub)
{
    RECORD_STACK_BASE(env, &env);
    jclass result;
    ValidateClass(env, sub);
    result = UNCHECKED()->GetSuperclass(env, sub);
    RECOVER_STACK_BASE(env);
    return result;
}

static jboolean JNICALL
checked_jni_IsAssignableFrom(JNIEnv *env, jclass sub, jclass sup)
{
    RECORD_STACK_BASE(env, &env);
    jboolean result;
    ValidateClass(env, sub);
    ValidateClass(env, sup);
    result = UNCHECKED()->IsAssignableFrom(env, sub, sup);
    RECOVER_STACK_BASE(env);
    return result;
}

static jobject JNICALL
checked_jni_ToReflectedField(JNIEnv *env, jclass cls, jfieldID fieldID,
			     jboolean isStatic)
{
    RECORD_STACK_BASE(env, &env);
    jobject result;
    ValidateClass(env, cls);
    result = UNCHECKED()->ToReflectedField(env, cls, fieldID, isStatic);
    RECOVER_STACK_BASE(env);
    return result;
}

static jint JNICALL
checked_jni_Throw(JNIEnv *env, jobject obj)
{
    RECORD_STACK_BASE(env, &env);
    jint result;
    ValidateObject(env, obj);
    result = UNCHECKED()->Throw(env, obj);
    RECOVER_STACK_BASE(env);
    return result;
}

static jint JNICALL
checked_jni_ThrowNew(JNIEnv *env, jclass clazz, const char *msg)
{
    RECORD_STACK_BASE(env, &env);
    jint result;
    ValidateClass(env, clazz);
    result = UNCHECKED()->ThrowNew(env, clazz, msg);
    RECOVER_STACK_BASE(env);
    return result;
}

static jobject JNICALL
checked_jni_ExceptionOccurred(JNIEnv *env)
{
    RECORD_STACK_BASE(env, &env);
    jobject result;
    result = UNCHECKED()->ExceptionOccurred(env);
    RECOVER_STACK_BASE(env);
    return result;
}

static void JNICALL
checked_jni_ExceptionDescribe(JNIEnv *env)
{
    RECORD_STACK_BASE(env, &env);
    UNCHECKED()->ExceptionDescribe(env);
    RECOVER_STACK_BASE(env);
}

static void JNICALL
checked_jni_ExceptionClear(JNIEnv *env)
{
    RECORD_STACK_BASE(env, &env);
    UNCHECKED()->ExceptionClear(env);
    RECOVER_STACK_BASE(env);
}

static void JNICALL
checked_jni_FatalError(JNIEnv *env, const char *msg)
{
    RECORD_STACK_BASE(env, &env);
    UNCHECKED()->FatalError(env, msg);
    RECOVER_STACK_BASE(env);
}

static jint JNICALL
checked_jni_PushLocalFrame(JNIEnv *env, jint capacity)
{
    if (capacity < 0)
        UNCHECKED()->FatalError(env, "negative capacity");
    return UNCHECKED()->PushLocalFrame(env, capacity);
}

static jobject JNICALL
checked_jni_PopLocalFrame(JNIEnv *env, jobject result)
{
    return UNCHECKED()->PopLocalFrame(env, result);
}

static jobject JNICALL
checked_jni_NewGlobalRef(JNIEnv *env, jobject lobj)
{
    RECORD_STACK_BASE(env, &env);
    jobject result;
    ValidateObject(env, lobj);
    result = UNCHECKED()->NewGlobalRef(env,lobj);
    RECOVER_STACK_BASE(env);
    return result;
}

static void JNICALL
checked_jni_DeleteGlobalRef(JNIEnv *env, jobject gref)
{
    RECORD_STACK_BASE(env, &env);
    ValidateObject(env, gref);
    UNCHECKED()->DeleteGlobalRef(env,gref);
    RECOVER_STACK_BASE(env);
}

static void JNICALL
checked_jni_DeleteLocalRef(JNIEnv *env, jobject obj)
{
    RECORD_STACK_BASE(env, &env);
    ValidateObject(env, obj);
    UNCHECKED()->DeleteLocalRef(env, obj);
    RECOVER_STACK_BASE(env);
}

static jboolean JNICALL
checked_jni_IsSameObject(JNIEnv *env, jobject obj1, jobject obj2)
{
    RECORD_STACK_BASE(env, &env);
    jboolean result;
    ValidateObject(env, obj1);
    ValidateObject(env, obj2);
    result = UNCHECKED()->IsSameObject(env,obj1,obj2);
    RECOVER_STACK_BASE(env);
    return result;
}

static jobject JNICALL
checked_jni_NewLocalRef(JNIEnv *env, jobject ref)
{
    ValidateObject(env, ref);
    return UNCHECKED()->NewLocalRef(env, ref);
}

static jint JNICALL
checked_jni_EnsureLocalCapacity(JNIEnv *env, jint capacity)
{
    if (capacity < 0)
        UNCHECKED()->FatalError(env, "negative capacity");
    return UNCHECKED()->EnsureLocalCapacity(env, capacity);
}

static jobject JNICALL
checked_jni_AllocObject(JNIEnv *env, jclass clazz)
{
    RECORD_STACK_BASE(env, &env);
    jobject result;
    ValidateClass(env, clazz);
    result = UNCHECKED()->AllocObject(env,clazz);
    RECOVER_STACK_BASE(env);
    return result;
}

static jobject JNICALL
checked_jni_NewObject(JNIEnv *env, jclass clazz, jmethodID methodID, ...)
{
    RECORD_STACK_BASE(env, &env);
    va_list args;
    jobject result;
    ValidateClass(env, clazz);
    va_start(args, methodID);
    result = UNCHECKED()->NewObjectV(env,clazz,methodID,args);
    va_end(args);
    RECOVER_STACK_BASE(env);
    return result;
}

static jobject JNICALL
checked_jni_NewObjectV(JNIEnv *env, jclass clazz, jmethodID methodID, va_list args)
{
    RECORD_STACK_BASE(env, &env);
    jobject result;
    ValidateClass(env, clazz);
    result = UNCHECKED()->NewObjectV(env,clazz,methodID,args);
    RECOVER_STACK_BASE(env);
    return result;
}

static jobject JNICALL
checked_jni_NewObjectA(JNIEnv *env, jclass clazz, jmethodID methodID, jvalue *args)
{
    RECORD_STACK_BASE(env, &env);
    jobject result;
    ValidateClass(env, clazz);
    result = UNCHECKED()->NewObjectA(env,clazz,methodID,args);
    RECOVER_STACK_BASE(env);
    return result;
}

static jclass JNICALL
checked_jni_GetObjectClass(JNIEnv *env, jobject obj)
{
    RECORD_STACK_BASE(env, &env);
    jclass result;
    ValidateObject(env, obj);
    result = UNCHECKED()->GetObjectClass(env,obj);
    RECOVER_STACK_BASE(env);
    return result;
}

static jboolean JNICALL
checked_jni_IsInstanceOf(JNIEnv *env, jobject obj, jclass clazz)
{
    RECORD_STACK_BASE(env, &env);
    jboolean result;
    ValidateObject(env, obj);
    ValidateClass(env, clazz);
    result = UNCHECKED()->IsInstanceOf(env,obj,clazz);
    RECOVER_STACK_BASE(env);
    return result;
}

static jmethodID JNICALL
checked_jni_GetMethodID(JNIEnv *env, jclass clazz, const char *name,
		   const char *sig)
{
    RECORD_STACK_BASE(env, &env);
    jmethodID result;
    ValidateClass(env, clazz);
    result = UNCHECKED()->GetMethodID(env,clazz,name,sig);
    RECOVER_STACK_BASE(env);
    return result;
}

#define WRAPPER_CALLMETHOD(ResultType, Result) \
static ResultType JNICALL \
checked_jni_Call##Result##Method(JNIEnv *env, jobject obj, \
			    jmethodID methodID, ...) \
{ \
    RECORD_STACK_BASE(env, &env); \
    va_list args; \
    ResultType result; \
    ValidateObject(env, obj); \
    va_start(args,methodID); \
    result = UNCHECKED()->Call##Result##MethodV(env,obj,methodID,args); \
    va_end(args); \
    RECOVER_STACK_BASE(env); \
    return result; \
} \
static ResultType JNICALL \
checked_jni_Call##Result##MethodV(JNIEnv *env, jobject obj, jmethodID methodID, \
			     va_list args) \
{ \
    RECORD_STACK_BASE(env, &env); \
    ResultType result; \
    ValidateObject(env, obj); \
    result = UNCHECKED()->Call##Result##MethodV(env,obj,methodID,args); \
    RECOVER_STACK_BASE(env); \
    return result; \
} \
static ResultType JNICALL \
checked_jni_Call##Result##MethodA(JNIEnv *env, jobject obj, jmethodID methodID, \
			     jvalue * args) \
{ \
    RECORD_STACK_BASE(env, &env); \
    ResultType result; \
    ValidateObject(env, obj); \
    result = UNCHECKED()->Call##Result##MethodA(env,obj,methodID,args); \
    RECOVER_STACK_BASE(env); \
    return result; \
}

WRAPPER_CALLMETHOD(jobject,Object)
WRAPPER_CALLMETHOD(jboolean,Boolean)
WRAPPER_CALLMETHOD(jbyte,Byte)
WRAPPER_CALLMETHOD(jshort,Short)
WRAPPER_CALLMETHOD(jchar,Char)
WRAPPER_CALLMETHOD(jint,Int)
WRAPPER_CALLMETHOD(jlong,Long)
WRAPPER_CALLMETHOD(jfloat,Float)
WRAPPER_CALLMETHOD(jdouble,Double)

static void JNICALL
checked_jni_CallVoidMethod(JNIEnv *env, jobject obj, jmethodID methodID, ...)
{
    RECORD_STACK_BASE(env, &env);
    va_list args;
    ValidateObject(env, obj);
    va_start(args,methodID);
    UNCHECKED()->CallVoidMethodV(env,obj,methodID,args);
    va_end(args);
    RECOVER_STACK_BASE(env);
}

static void JNICALL
checked_jni_CallVoidMethodV(JNIEnv *env, jobject obj, jmethodID methodID,
		       va_list args)
{
    RECORD_STACK_BASE(env, &env);
    ValidateObject(env, obj);
    UNCHECKED()->CallVoidMethodV(env,obj,methodID,args);
    RECOVER_STACK_BASE(env);
}

static void JNICALL
checked_jni_CallVoidMethodA(JNIEnv *env, jobject obj, jmethodID methodID,
		       jvalue * args)
{
    RECORD_STACK_BASE(env, &env);
    ValidateObject(env, obj);
    UNCHECKED()->CallVoidMethodA(env,obj,methodID,args);
    RECOVER_STACK_BASE(env);
}

#define WRAPPER_CALLNONVIRTUALMETHOD(ResultType, Result) \
static ResultType JNICALL \
checked_jni_CallNonvirtual##Result##Method(JNIEnv *env, jobject obj, \
				      jclass clazz, jmethodID methodID, ...) \
{ \
    RECORD_STACK_BASE(env, &env); \
    va_list args; \
    ResultType result; \
    ValidateObject(env, obj); \
    va_start(args,methodID); \
    result = UNCHECKED()->CallNonvirtual##Result##MethodV(env,obj,clazz, \
						 methodID,args); \
    va_end(args); \
    RECOVER_STACK_BASE(env); \
    return result; \
} \
static ResultType JNICALL \
checked_jni_CallNonvirtual##Result##MethodV(JNIEnv *env, jobject obj, \
				       jclass clazz, jmethodID methodID, \
				       va_list args) \
{ \
    RECORD_STACK_BASE(env, &env); \
    ResultType result; \
    ValidateObject(env, obj); \
    result = UNCHECKED()->CallNonvirtual##Result##MethodV(env,obj,clazz, \
							    methodID,args); \
    RECOVER_STACK_BASE(env); \
    return result; \
} \
static ResultType JNICALL \
checked_jni_CallNonvirtual##Result##MethodA(JNIEnv *env, jobject obj, \
				       jclass clazz, jmethodID methodID, \
				       jvalue * args) \
{ \
    RECORD_STACK_BASE(env, &env); \
    ResultType result; \
    ValidateObject(env, obj); \
    result = UNCHECKED()->CallNonvirtual##Result##MethodA(env,obj,clazz, \
							    methodID,args); \
    RECOVER_STACK_BASE(env); \
    return result; \
}

WRAPPER_CALLNONVIRTUALMETHOD(jobject,Object)
WRAPPER_CALLNONVIRTUALMETHOD(jboolean,Boolean)
WRAPPER_CALLNONVIRTUALMETHOD(jbyte,Byte)
WRAPPER_CALLNONVIRTUALMETHOD(jshort,Short)
WRAPPER_CALLNONVIRTUALMETHOD(jchar,Char)
WRAPPER_CALLNONVIRTUALMETHOD(jint,Int)
WRAPPER_CALLNONVIRTUALMETHOD(jlong,Long)
WRAPPER_CALLNONVIRTUALMETHOD(jfloat,Float)
WRAPPER_CALLNONVIRTUALMETHOD(jdouble,Double)

static void JNICALL
checked_jni_CallNonvirtualVoidMethod(JNIEnv *env, jobject obj, jclass clazz,
				jmethodID methodID, ...)
{
    RECORD_STACK_BASE(env, &env);
    va_list args;
    ValidateObject(env, obj);
    va_start(args,methodID);
    UNCHECKED()->CallNonvirtualVoidMethodV(env,obj,clazz,methodID,args);
    va_end(args);
    RECOVER_STACK_BASE(env);
}

static void JNICALL
checked_jni_CallNonvirtualVoidMethodV(JNIEnv *env, jobject obj, jclass clazz,
				 jmethodID methodID,
				 va_list args)
{
    RECORD_STACK_BASE(env, &env);
    ValidateObject(env, obj);
    UNCHECKED()->CallNonvirtualVoidMethodV(env,obj,clazz,methodID,args);
    RECOVER_STACK_BASE(env);
}

static void JNICALL
checked_jni_CallNonvirtualVoidMethodA(JNIEnv *env, jobject obj, jclass clazz,
				 jmethodID methodID,
				 jvalue * args)
{
    RECORD_STACK_BASE(env, &env);
    ValidateObject(env, obj);
    UNCHECKED()->CallNonvirtualVoidMethodA(env,obj,clazz,methodID,args);
    RECOVER_STACK_BASE(env);
}

static jfieldID JNICALL
checked_jni_GetFieldID(JNIEnv *env, jclass clazz, const char *name,
		  const char *sig)
{
    RECORD_STACK_BASE(env, &env);
    jfieldID result;
    ValidateClass(env, clazz);
    result = UNCHECKED()->GetFieldID(env,clazz,name,sig);
    RECOVER_STACK_BASE(env);
    return result;
}

#define WRAPPER_GETFIELD(ReturnType,Result,ty) \
static ReturnType JNICALL \
checked_jni_Get##Result##Field(JNIEnv *env, jobject obj, jfieldID fieldID) \
{ \
    RECORD_STACK_BASE(env, &env); \
    struct fieldblock *fb = (struct fieldblock *)fieldID; \
    ReturnType result; \
    CHECK_INSTANCE_FIELD_ID(env, fb, obj); \
    CHECK_FIELD_TYPE(env, ty); \
    result = UNCHECKED()->Get##Result##Field(env,obj,fieldID); \
    RECOVER_STACK_BASE(env); \
    return result; \
}

WRAPPER_GETFIELD(jobject,Object,(fb->signature[0] == 'L' || fb->signature[0] == '['))
WRAPPER_GETFIELD(jboolean,Boolean,(fb->signature[0] == 'Z'))
WRAPPER_GETFIELD(jbyte,Byte,(fb->signature[0] == 'B'))
WRAPPER_GETFIELD(jshort,Short,(fb->signature[0] == 'S'))
WRAPPER_GETFIELD(jchar,Char,(fb->signature[0] == 'C'))
WRAPPER_GETFIELD(jint,Int,(fb->signature[0] == 'I'))
WRAPPER_GETFIELD(jlong,Long,(fb->signature[0] == 'J'))
WRAPPER_GETFIELD(jfloat,Float,(fb->signature[0] == 'F'))
WRAPPER_GETFIELD(jdouble,Double,(fb->signature[0] == 'D'))

#define WRAPPER_SETFIELD(ValueType,Result,ty) \
static void JNICALL \
checked_jni_Set##Result##Field(JNIEnv *env, jobject obj, jfieldID fieldID, \
			  ValueType val) \
{ \
    RECORD_STACK_BASE(env, &env); \
    struct fieldblock *fb = (struct fieldblock *)fieldID; \
    CHECK_INSTANCE_FIELD_ID(env, fb, obj); \
    CHECK_FIELD_TYPE(env, ty); \
    UNCHECKED()->Set##Result##Field(env,obj,fieldID,val); \
    RECOVER_STACK_BASE(env); \
}

WRAPPER_SETFIELD(jobject,Object,(fb->signature[0] == 'L' || fb->signature[0] == '['))
WRAPPER_SETFIELD(jboolean,Boolean,(fb->signature[0] == 'Z'))
WRAPPER_SETFIELD(jbyte,Byte,(fb->signature[0] == 'B'))
WRAPPER_SETFIELD(jshort,Short,(fb->signature[0] == 'S'))
WRAPPER_SETFIELD(jchar,Char,(fb->signature[0] == 'C'))
WRAPPER_SETFIELD(jint,Int,(fb->signature[0] == 'I'))
WRAPPER_SETFIELD(jlong,Long,(fb->signature[0] == 'J'))
WRAPPER_SETFIELD(jfloat,Float,(fb->signature[0] == 'F'))
WRAPPER_SETFIELD(jdouble,Double,(fb->signature[0] == 'D'))

static jmethodID JNICALL
checked_jni_GetStaticMethodID(JNIEnv *env, jclass clazz, const char *name,
			 const char *sig)
{
    RECORD_STACK_BASE(env, &env);
    jmethodID result;
    ValidateClass(env, clazz);
    result = UNCHECKED()->GetStaticMethodID(env,clazz,name,sig);
    RECOVER_STACK_BASE(env);
    return result;
}

#define WRAPPER_CALLSTATICMETHOD(ReturnType,Result) \
static ReturnType JNICALL \
checked_jni_CallStatic##Result##Method(JNIEnv *env, jclass clazz, \
				  jmethodID methodID, ...) \
{ \
    RECORD_STACK_BASE(env, &env); \
    va_list args; \
    ReturnType result; \
    ValidateClass(env, clazz); \
    va_start(args,methodID); \
    result = UNCHECKED()->CallStatic##Result##MethodV(env,clazz,methodID,args); \
    va_end(args); \
    RECOVER_STACK_BASE(env); \
    return result; \
} \
static ReturnType JNICALL \
checked_jni_CallStatic##Result##MethodV(JNIEnv *env, jclass clazz, \
				   jmethodID methodID, va_list args) \
{ \
    RECORD_STACK_BASE(env, &env); \
    ReturnType result; \
    ValidateClass(env, clazz); \
    result = \
        UNCHECKED()->CallStatic##Result##MethodV(env,clazz,methodID,args); \
    RECOVER_STACK_BASE(env); \
    return result; \
} \
static ReturnType JNICALL \
checked_jni_CallStatic##Result##MethodA(JNIEnv *env, jclass clazz, \
				   jmethodID methodID, jvalue *args) \
{ \
    RECORD_STACK_BASE(env, &env); \
    ReturnType result; \
    ValidateClass(env, clazz); \
    result = \
        UNCHECKED()->CallStatic##Result##MethodA(env,clazz,methodID,args); \
    RECOVER_STACK_BASE(env); \
    return result; \
}

WRAPPER_CALLSTATICMETHOD(jobject,Object)
WRAPPER_CALLSTATICMETHOD(jboolean,Boolean)
WRAPPER_CALLSTATICMETHOD(jbyte,Byte)
WRAPPER_CALLSTATICMETHOD(jshort,Short)
WRAPPER_CALLSTATICMETHOD(jchar,Char)
WRAPPER_CALLSTATICMETHOD(jint,Int)
WRAPPER_CALLSTATICMETHOD(jlong,Long)
WRAPPER_CALLSTATICMETHOD(jfloat,Float)
WRAPPER_CALLSTATICMETHOD(jdouble,Double)

static void JNICALL
checked_jni_CallStaticVoidMethod(JNIEnv *env, jclass cls, jmethodID methodID, ...)
{
    RECORD_STACK_BASE(env, &env);
    va_list args;
    ValidateClass(env, cls);
    va_start(args,methodID);
    UNCHECKED()->CallStaticVoidMethodV(env,cls,methodID,args);
    va_end(args);
    RECOVER_STACK_BASE(env);
}

static void JNICALL
checked_jni_CallStaticVoidMethodV(JNIEnv *env, jclass cls, jmethodID methodID,
			     va_list args)
{
    RECORD_STACK_BASE(env, &env);
    ValidateClass(env, cls);
    UNCHECKED()->CallStaticVoidMethodV(env,cls,methodID,args);
    RECOVER_STACK_BASE(env);
}

static void JNICALL
checked_jni_CallStaticVoidMethodA(JNIEnv *env, jclass cls, jmethodID methodID,
			     jvalue * args)
{
    RECORD_STACK_BASE(env, &env);
    ValidateClass(env, cls);
    UNCHECKED()->CallStaticVoidMethodA(env,cls,methodID,args);
    RECOVER_STACK_BASE(env);
}

static jfieldID JNICALL
checked_jni_GetStaticFieldID(JNIEnv *env, jclass clazz, const char *name,
			const char *sig)
{
    RECORD_STACK_BASE(env, &env);
    jfieldID result;
    ValidateClass(env, clazz);
    result = UNCHECKED()->GetStaticFieldID(env,clazz,name,sig);
    RECOVER_STACK_BASE(env);
    return result;
}

#define WRAPPER_GETSTATICFIELD(ReturnType,Result,ty) \
static ReturnType JNICALL \
checked_jni_GetStatic##Result##Field(JNIEnv *env, jclass clazz, jfieldID fieldID) \
{ \
    RECORD_STACK_BASE(env, &env); \
    struct fieldblock *fb = (struct fieldblock *)fieldID; \
    ReturnType result; \
    CHECK_STATIC_FIELD_ID(env, fb, clazz); \
    CHECK_FIELD_TYPE(env, ty); \
    result = UNCHECKED()->GetStatic##Result##Field(env,clazz,fieldID); \
    RECOVER_STACK_BASE(env); \
    return result; \
}

WRAPPER_GETSTATICFIELD(jobject,Object,(fb->signature[0] == 'L' || fb->signature[0] == '['))
WRAPPER_GETSTATICFIELD(jboolean,Boolean,(fb->signature[0] == 'Z'))
WRAPPER_GETSTATICFIELD(jbyte,Byte,(fb->signature[0] == 'B'))
WRAPPER_GETSTATICFIELD(jshort,Short,(fb->signature[0] == 'S'))
WRAPPER_GETSTATICFIELD(jchar,Char,(fb->signature[0] == 'C'))
WRAPPER_GETSTATICFIELD(jint,Int,(fb->signature[0] == 'I'))
WRAPPER_GETSTATICFIELD(jlong,Long,(fb->signature[0] == 'J'))
WRAPPER_GETSTATICFIELD(jfloat,Float,(fb->signature[0] == 'F'))
WRAPPER_GETSTATICFIELD(jdouble,Double,(fb->signature[0] == 'D'))

#define WRAPPER_SETSTATICFIELD(ValueType,Result,ty) \
static void JNICALL \
checked_jni_SetStatic##Result##Field(JNIEnv *env, jclass clazz, jfieldID fieldID, \
				ValueType value) \
{ \
    RECORD_STACK_BASE(env, &env); \
    struct fieldblock *fb = (struct fieldblock *)fieldID; \
    CHECK_STATIC_FIELD_ID(env, fb, clazz); \
    CHECK_FIELD_TYPE(env, ty); \
    UNCHECKED()->SetStatic##Result##Field(env,clazz,fieldID,value); \
    RECOVER_STACK_BASE(env); \
}

WRAPPER_SETSTATICFIELD(jobject,Object,(fb->signature[0] == 'L' || fb->signature[0] == '['))
WRAPPER_SETSTATICFIELD(jboolean,Boolean,(fb->signature[0] == 'Z'))
WRAPPER_SETSTATICFIELD(jbyte,Byte,(fb->signature[0] == 'B'))
WRAPPER_SETSTATICFIELD(jshort,Short,(fb->signature[0] == 'S'))
WRAPPER_SETSTATICFIELD(jchar,Char,(fb->signature[0] == 'C'))
WRAPPER_SETSTATICFIELD(jint,Int,(fb->signature[0] == 'I'))
WRAPPER_SETSTATICFIELD(jlong,Long,(fb->signature[0] == 'J'))
WRAPPER_SETSTATICFIELD(jfloat,Float,(fb->signature[0] == 'F'))
WRAPPER_SETSTATICFIELD(jdouble,Double,(fb->signature[0] == 'D'))

static jstring JNICALL
checked_jni_NewString(JNIEnv *env, const jchar *unicode, jsize len)
{
    RECORD_STACK_BASE(env, &env);
    jstring result;
    result = UNCHECKED()->NewString(env,unicode,len);
    RECOVER_STACK_BASE(env);
    return result;
}

static jsize JNICALL
checked_jni_GetStringLength(JNIEnv *env, jstring str)
{
    RECORD_STACK_BASE(env, &env);
    jsize result;
    CHECK_STRING(str);
    result = UNCHECKED()->GetStringLength(env,str);
    RECOVER_STACK_BASE(env);
    return result;
}

static const jchar * JNICALL
checked_jni_GetStringChars(JNIEnv *env, jstring str, jboolean *isCopy)
{
    RECORD_STACK_BASE(env, &env);
    const jchar *result;
    CHECK_STRING(str);
    result = UNCHECKED()->GetStringChars(env,str,isCopy);
    RECOVER_STACK_BASE(env);
    return result;
}

static void JNICALL
checked_jni_ReleaseStringChars(JNIEnv *env, jstring str, const jchar *chars)
{
    RECORD_STACK_BASE(env, &env);
    Classjava_lang_String *s;
    CHECK_STRING(str);
    s = unhand((HString *)DeRef(env, str));
    if (chars != unhand(s->value)->body + s->offset) {
	UNCHECKED()->FatalError(env,
            "Passed wrong C string to ReleaseStringCritical");
    }
    UNCHECKED()->ReleaseStringChars(env,str,chars);
    RECOVER_STACK_BASE(env);
}

static jstring JNICALL
checked_jni_NewStringUTF(JNIEnv *env, const char *utf)
{
    RECORD_STACK_BASE(env, &env);
    jstring result;
    result = UNCHECKED()->NewStringUTF(env,utf);
    RECOVER_STACK_BASE(env);
    return result;
}

static jsize JNICALL
checked_jni_GetStringUTFLength(JNIEnv *env, jstring str)
{
    RECORD_STACK_BASE(env, &env);
    jsize result;
    CHECK_STRING(str);
    result = UNCHECKED()->GetStringUTFLength(env,str);
    RECOVER_STACK_BASE(env);
    return result;
}

static const char * JNICALL
checked_jni_GetStringUTFChars(JNIEnv *env, jstring str, jboolean *isCopy)
{
    RECORD_STACK_BASE(env, &env);
    const char *result;
    CHECK_STRING(str);
    result = UNCHECKED()->GetStringUTFChars(env,str,isCopy);
    RECOVER_STACK_BASE(env);
    return result;
}

static void JNICALL
checked_jni_ReleaseStringUTFChars(JNIEnv *env, jstring str, const char* chars)
{
    RECORD_STACK_BASE(env, &env);
    CHECK_STRING(str);
    UNCHECKED()->ReleaseStringUTFChars(env,str,chars);
    RECOVER_STACK_BASE(env);
}

static jsize JNICALL
checked_jni_GetArrayLength(JNIEnv *env, jarray array)
{
    RECORD_STACK_BASE(env, &env);
    jsize result;
    CHECK_ARRAY(array, -1);
    result = UNCHECKED()->GetArrayLength(env,array);
    RECOVER_STACK_BASE(env);
    return result;
}

static jarray JNICALL
checked_jni_NewObjectArray(JNIEnv *env, jsize len, jclass clazz, jobject init)
{
    RECORD_STACK_BASE(env, &env);
    jarray result = UNCHECKED()->NewObjectArray(env,len,clazz,init);
    RECOVER_STACK_BASE(env);
    return result;
}

static jobject JNICALL
checked_jni_GetObjectArrayElement(JNIEnv *env, jarray array, jsize index)
{
    RECORD_STACK_BASE(env, &env);
    jobject result;
    CHECK_ARRAY(array, T_CLASS);
    result = UNCHECKED()->GetObjectArrayElement(env,array,index);
    RECOVER_STACK_BASE(env);
    return result;
}

static void JNICALL
checked_jni_SetObjectArrayElement(JNIEnv *env, jarray array, jsize index,
			     jobject val)
{
    RECORD_STACK_BASE(env, &env);
    CHECK_ARRAY(array, T_CLASS);
    UNCHECKED()->SetObjectArrayElement(env,array,index,val);
    RECOVER_STACK_BASE(env);
}

#define WRAPPER_NEWSCALARARRAY(Result) \
static jarray JNICALL \
checked_jni_New##Result##Array(JNIEnv *env, jsize len) \
{ \
    RECORD_STACK_BASE(env, &env); \
    jarray result = UNCHECKED()->New##Result##Array(env,len); \
    RECOVER_STACK_BASE(env); \
    return result; \
}

WRAPPER_NEWSCALARARRAY(Boolean)
WRAPPER_NEWSCALARARRAY(Byte)
WRAPPER_NEWSCALARARRAY(Short)
WRAPPER_NEWSCALARARRAY(Char)
WRAPPER_NEWSCALARARRAY(Int)
WRAPPER_NEWSCALARARRAY(Long)
WRAPPER_NEWSCALARARRAY(Float)
WRAPPER_NEWSCALARARRAY(Double)

#define WRAPPER_GETSCALARARRAYELEMENTS(ElementType,Result,ElementTag) \
static ElementType * JNICALL \
checked_jni_Get##Result##ArrayElements(JNIEnv *env, jarray array, jboolean *isCopy) \
{ \
    RECORD_STACK_BASE(env, &env); \
    ElementType *result; \
    CHECK_ARRAY(array, ElementTag); \
    result = UNCHECKED()->Get##Result##ArrayElements(env,array,isCopy); \
    RECOVER_STACK_BASE(env); \
    return result; \
}

WRAPPER_GETSCALARARRAYELEMENTS(jboolean,Boolean,T_BOOLEAN)
WRAPPER_GETSCALARARRAYELEMENTS(jbyte,Byte,T_BYTE)
WRAPPER_GETSCALARARRAYELEMENTS(jshort,Short,T_SHORT)
WRAPPER_GETSCALARARRAYELEMENTS(jchar,Char,T_CHAR)
WRAPPER_GETSCALARARRAYELEMENTS(jint,Int,T_INT)
WRAPPER_GETSCALARARRAYELEMENTS(jlong,Long,T_LONG)
WRAPPER_GETSCALARARRAYELEMENTS(jfloat,Float,T_FLOAT)
WRAPPER_GETSCALARARRAYELEMENTS(jdouble,Double,T_DOUBLE)

#define WRAPPER_RELEASESCALARARRAYELEMENTS(ElementType,Result,ElementTag) \
static void JNICALL \
checked_jni_Release##Result##ArrayElements(JNIEnv *env, jarray array, \
				      ElementType *elems, jint mode) \
{ \
    RECORD_STACK_BASE(env, &env); \
    CHECK_ARRAY(array, ElementTag); \
    if (elems != (ElementType *)unhand((HArrayOfByte *)DeRef(env, array))->body) { \
	UNCHECKED()->FatalError(env, \
            "Passed wrong C array to Release<type>ArrayElements"); \
    } \
    UNCHECKED()->Release##Result##ArrayElements(env,array,elems,mode); \
    RECOVER_STACK_BASE(env); \
}

WRAPPER_RELEASESCALARARRAYELEMENTS(jboolean,Boolean,T_BOOLEAN)
WRAPPER_RELEASESCALARARRAYELEMENTS(jbyte,Byte,T_BYTE)
WRAPPER_RELEASESCALARARRAYELEMENTS(jshort,Short,T_SHORT)
WRAPPER_RELEASESCALARARRAYELEMENTS(jchar,Char,T_CHAR)
WRAPPER_RELEASESCALARARRAYELEMENTS(jint,Int,T_INT)
WRAPPER_RELEASESCALARARRAYELEMENTS(jlong,Long,T_LONG)
WRAPPER_RELEASESCALARARRAYELEMENTS(jfloat,Float,T_FLOAT)
WRAPPER_RELEASESCALARARRAYELEMENTS(jdouble,Double,T_DOUBLE)

#define WRAPPER_GETSCALARARRAYREGION(ElementType,Result,ElementTag) \
static void JNICALL \
checked_jni_Get##Result##ArrayRegion(JNIEnv *env, jarray array, jsize start, \
				  jsize len, ElementType *buf) \
{ \
    RECORD_STACK_BASE(env, &env); \
    CHECK_ARRAY(array, ElementTag); \
    UNCHECKED()->Get##Result##ArrayRegion(env,array,start,len,buf); \
    RECOVER_STACK_BASE(env); \
}

WRAPPER_GETSCALARARRAYREGION(jboolean,Boolean,T_BOOLEAN)
WRAPPER_GETSCALARARRAYREGION(jbyte,Byte,T_BYTE)
WRAPPER_GETSCALARARRAYREGION(jshort,Short,T_SHORT)
WRAPPER_GETSCALARARRAYREGION(jchar,Char,T_CHAR)
WRAPPER_GETSCALARARRAYREGION(jint,Int,T_INT)
WRAPPER_GETSCALARARRAYREGION(jlong,Long,T_LONG)
WRAPPER_GETSCALARARRAYREGION(jfloat,Float,T_FLOAT)
WRAPPER_GETSCALARARRAYREGION(jdouble,Double,T_DOUBLE)

#define WRAPPER_SETSCALARARRAYREGION(ElementType,Result,ElementTag) \
static void JNICALL \
checked_jni_Set##Result##ArrayRegion(JNIEnv *env, jarray array, jsize start, \
				  jsize len, ElementType *buf) \
{ \
    RECORD_STACK_BASE(env, &env); \
    CHECK_ARRAY(array, ElementTag); \
    UNCHECKED()->Set##Result##ArrayRegion(env,array,start,len,buf); \
    RECOVER_STACK_BASE(env); \
}

WRAPPER_SETSCALARARRAYREGION(jboolean,Boolean,T_BOOLEAN)
WRAPPER_SETSCALARARRAYREGION(jbyte,Byte,T_BYTE)
WRAPPER_SETSCALARARRAYREGION(jshort,Short,T_SHORT)
WRAPPER_SETSCALARARRAYREGION(jchar,Char,T_CHAR)
WRAPPER_SETSCALARARRAYREGION(jint,Int,T_INT)
WRAPPER_SETSCALARARRAYREGION(jlong,Long,T_LONG)
WRAPPER_SETSCALARARRAYREGION(jfloat,Float,T_FLOAT)
WRAPPER_SETSCALARARRAYREGION(jdouble,Double,T_DOUBLE)

static jint JNICALL
checked_jni_RegisterNatives(JNIEnv *env, jclass clazz,
		       const JNINativeMethod *methods, jint nMethods)
{
    RECORD_STACK_BASE(env, &env);
    jint result = UNCHECKED()->RegisterNatives(env,clazz,methods,nMethods);
    RECOVER_STACK_BASE(env);
    return result;
}

static jint JNICALL
checked_jni_UnregisterNatives(JNIEnv *env, jclass clazz)
{
    RECORD_STACK_BASE(env, &env);
    jint result = UNCHECKED()->UnregisterNatives(env,clazz);
    RECOVER_STACK_BASE(env);
    return result;
}

static jint JNICALL
checked_jni_MonitorEnter(JNIEnv *env, jobject obj)
{
    RECORD_STACK_BASE(env, &env);
    jint result;
    if (DeRef(env, obj) == NULL) {
        UNCHECKED()->FatalError(env, "null object passed to MonitorEnter");
    }
    ValidateObject(env, obj);
    result = UNCHECKED()->MonitorEnter(env,obj);
    RECOVER_STACK_BASE(env);
    return result;
}

static jint JNICALL
checked_jni_MonitorExit(JNIEnv *env, jobject obj)
{
    RECORD_STACK_BASE(env, &env);
    jint result;
    if (DeRef(env, obj) == NULL) {
        UNCHECKED()->FatalError(env, "null object passed to MonitorExit");
    }
    ValidateObject(env, obj);
    result = UNCHECKED()->MonitorExit(env,obj);
    RECOVER_STACK_BASE(env);
    return result;
}

static jint JNICALL
checked_jni_GetJavaVM(JNIEnv *env, JavaVM **vm)
{
    RECORD_STACK_BASE(env, &env);
    jint result = UNCHECKED()->GetJavaVM(env,vm);
    RECOVER_STACK_BASE(env);
    return result;
}


static void JNICALL
checked_jni_GetStringRegion(JNIEnv *env, jstring str, jsize start, jsize len, jchar *buf)
{
    RECORD_STACK_BASE(env, &env);
    CHECK_STRING(str);
    UNCHECKED()->GetStringRegion(env, str, start, len, buf);
    RECOVER_STACK_BASE(env);
}

static void JNICALL
checked_jni_GetStringUTFRegion
  (JNIEnv *env, jstring str, jsize start, jsize len, char *buf)
{
    RECORD_STACK_BASE(env, &env);
    CHECK_STRING(str);
    UNCHECKED()->GetStringUTFRegion(env, str, start, len, buf);
    RECOVER_STACK_BASE(env);
}

static void * JNICALL
checked_jni_GetPrimitiveArrayCritical(JNIEnv *env, jarray array, jboolean *isCopy)
{
    RECORD_STACK_BASE_CRITICAL(env, &env);
    void *result;
    CHECK_ARRAY(array, -1);
    result = UNCHECKED()->GetPrimitiveArrayCritical(env, array, isCopy);
    RECOVER_STACK_BASE(env);
    return result;
}

static void JNICALL
checked_jni_ReleasePrimitiveArrayCritical
  (JNIEnv *env, jarray array, void *carray, jint mode)
{
    RECORD_STACK_BASE_CRITICAL(env, &env);
    CHECK_ARRAY(array, -1);
    if (carray != (void *)unhand((HArrayOfByte*)DeRef(env, array))->body) {
	UNCHECKED()->FatalError(env,
            "Passed wrong C array to ReleasePrimitiveArrayCritical");
    }
    UNCHECKED()->ReleasePrimitiveArrayCritical(env, array, carray, mode);
    RECOVER_STACK_BASE(env);
}

static const jchar* JNICALL
checked_jni_GetStringCritical(JNIEnv *env, jstring string, jboolean *isCopy)
{
    RECORD_STACK_BASE_CRITICAL(env, &env);
    const jchar * result;
    CHECK_STRING(string);
    result = UNCHECKED()->GetStringCritical(env, string, isCopy);
    RECOVER_STACK_BASE(env);
    return result;
}

static void JNICALL
checked_jni_ReleaseStringCritical(JNIEnv *env, jstring str, const jchar *chars)
{
    RECORD_STACK_BASE_CRITICAL(env, &env);
    Classjava_lang_String *s;
    CHECK_STRING(str);
    s = unhand((HString *)DeRef(env, str));
    if (chars != unhand(s->value)->body + s->offset) {
	UNCHECKED()->FatalError(env,
            "Passed wrong C string to ReleaseStringCritical");
    }
    UNCHECKED()->ReleaseStringCritical(env, str, chars);
    RECOVER_STACK_BASE(env);
}

static jweak JNICALL
checked_jni_NewWeakGlobalRef(JNIEnv *env, jobject obj)
{
    if (obj == NULL)
        UNCHECKED()->FatalError(env,
	    "Null object passed to NewWeakGlobalRef");
    return UNCHECKED()->NewWeakGlobalRef(env, obj);
}

static void JNICALL
checked_jni_DeleteWeakGlobalRef(JNIEnv *env, jweak ref)
{
    UNCHECKED()->DeleteWeakGlobalRef(env, ref);
}

static jboolean JNICALL
checked_jni_ExceptionCheck(JNIEnv *env)
{
    return exceptionOccurred(JNIEnv2EE(env));
}

const struct JNINativeInterface_ checked_jni_NativeInterface = {
    NULL,
    NULL,
    NULL,

    NULL,

    checked_jni_GetVersion,

    checked_jni_DefineClass,
    checked_jni_FindClass,

    checked_jni_FromReflectedMethod,
    checked_jni_FromReflectedField,

    checked_jni_ToReflectedMethod,

    checked_jni_GetSuperclass,
    checked_jni_IsAssignableFrom,

    checked_jni_ToReflectedField,

    checked_jni_Throw,
    checked_jni_ThrowNew,
    checked_jni_ExceptionOccurred,
    checked_jni_ExceptionDescribe,
    checked_jni_ExceptionClear,
    checked_jni_FatalError,

    checked_jni_PushLocalFrame,
    checked_jni_PopLocalFrame,

    checked_jni_NewGlobalRef,
    checked_jni_DeleteGlobalRef,
    checked_jni_DeleteLocalRef,
    checked_jni_IsSameObject,

    checked_jni_NewLocalRef,
    checked_jni_EnsureLocalCapacity,

    checked_jni_AllocObject,
    checked_jni_NewObject,
    checked_jni_NewObjectV,
    checked_jni_NewObjectA,

    checked_jni_GetObjectClass,
    checked_jni_IsInstanceOf,

    checked_jni_GetMethodID,

    checked_jni_CallObjectMethod,
    checked_jni_CallObjectMethodV,
    checked_jni_CallObjectMethodA,
    checked_jni_CallBooleanMethod,
    checked_jni_CallBooleanMethodV,
    checked_jni_CallBooleanMethodA,
    checked_jni_CallByteMethod,
    checked_jni_CallByteMethodV,
    checked_jni_CallByteMethodA,
    checked_jni_CallCharMethod,
    checked_jni_CallCharMethodV,
    checked_jni_CallCharMethodA,
    checked_jni_CallShortMethod,
    checked_jni_CallShortMethodV,
    checked_jni_CallShortMethodA,
    checked_jni_CallIntMethod,
    checked_jni_CallIntMethodV,
    checked_jni_CallIntMethodA,
    checked_jni_CallLongMethod,
    checked_jni_CallLongMethodV,
    checked_jni_CallLongMethodA,
    checked_jni_CallFloatMethod,
    checked_jni_CallFloatMethodV,
    checked_jni_CallFloatMethodA,
    checked_jni_CallDoubleMethod,
    checked_jni_CallDoubleMethodV,
    checked_jni_CallDoubleMethodA,
    checked_jni_CallVoidMethod,
    checked_jni_CallVoidMethodV,
    checked_jni_CallVoidMethodA,

    checked_jni_CallNonvirtualObjectMethod,
    checked_jni_CallNonvirtualObjectMethodV,
    checked_jni_CallNonvirtualObjectMethodA,
    checked_jni_CallNonvirtualBooleanMethod,
    checked_jni_CallNonvirtualBooleanMethodV,
    checked_jni_CallNonvirtualBooleanMethodA,
    checked_jni_CallNonvirtualByteMethod,
    checked_jni_CallNonvirtualByteMethodV,
    checked_jni_CallNonvirtualByteMethodA,
    checked_jni_CallNonvirtualCharMethod,
    checked_jni_CallNonvirtualCharMethodV,
    checked_jni_CallNonvirtualCharMethodA,
    checked_jni_CallNonvirtualShortMethod,
    checked_jni_CallNonvirtualShortMethodV,
    checked_jni_CallNonvirtualShortMethodA,
    checked_jni_CallNonvirtualIntMethod,
    checked_jni_CallNonvirtualIntMethodV,
    checked_jni_CallNonvirtualIntMethodA,
    checked_jni_CallNonvirtualLongMethod,
    checked_jni_CallNonvirtualLongMethodV,
    checked_jni_CallNonvirtualLongMethodA,
    checked_jni_CallNonvirtualFloatMethod,
    checked_jni_CallNonvirtualFloatMethodV,
    checked_jni_CallNonvirtualFloatMethodA,
    checked_jni_CallNonvirtualDoubleMethod,
    checked_jni_CallNonvirtualDoubleMethodV,
    checked_jni_CallNonvirtualDoubleMethodA,
    checked_jni_CallNonvirtualVoidMethod,
    checked_jni_CallNonvirtualVoidMethodV,
    checked_jni_CallNonvirtualVoidMethodA,

    checked_jni_GetFieldID,

    checked_jni_GetObjectField,
    checked_jni_GetBooleanField,
    checked_jni_GetByteField,
    checked_jni_GetCharField,
    checked_jni_GetShortField,
    checked_jni_GetIntField,
    checked_jni_GetLongField,
    checked_jni_GetFloatField,
    checked_jni_GetDoubleField,

    checked_jni_SetObjectField,
    checked_jni_SetBooleanField,
    checked_jni_SetByteField,
    checked_jni_SetCharField,
    checked_jni_SetShortField,
    checked_jni_SetIntField,
    checked_jni_SetLongField,
    checked_jni_SetFloatField,
    checked_jni_SetDoubleField,

    checked_jni_GetStaticMethodID,

    checked_jni_CallStaticObjectMethod,
    checked_jni_CallStaticObjectMethodV,
    checked_jni_CallStaticObjectMethodA,
    checked_jni_CallStaticBooleanMethod,
    checked_jni_CallStaticBooleanMethodV,
    checked_jni_CallStaticBooleanMethodA,
    checked_jni_CallStaticByteMethod,
    checked_jni_CallStaticByteMethodV,
    checked_jni_CallStaticByteMethodA,
    checked_jni_CallStaticCharMethod,
    checked_jni_CallStaticCharMethodV,
    checked_jni_CallStaticCharMethodA,
    checked_jni_CallStaticShortMethod,
    checked_jni_CallStaticShortMethodV,
    checked_jni_CallStaticShortMethodA,
    checked_jni_CallStaticIntMethod,
    checked_jni_CallStaticIntMethodV,
    checked_jni_CallStaticIntMethodA,
    checked_jni_CallStaticLongMethod,
    checked_jni_CallStaticLongMethodV,
    checked_jni_CallStaticLongMethodA,
    checked_jni_CallStaticFloatMethod,
    checked_jni_CallStaticFloatMethodV,
    checked_jni_CallStaticFloatMethodA,
    checked_jni_CallStaticDoubleMethod,
    checked_jni_CallStaticDoubleMethodV,
    checked_jni_CallStaticDoubleMethodA,
    checked_jni_CallStaticVoidMethod,
    checked_jni_CallStaticVoidMethodV,
    checked_jni_CallStaticVoidMethodA,

    checked_jni_GetStaticFieldID,

    checked_jni_GetStaticObjectField,
    checked_jni_GetStaticBooleanField,
    checked_jni_GetStaticByteField,
    checked_jni_GetStaticCharField,
    checked_jni_GetStaticShortField,
    checked_jni_GetStaticIntField,
    checked_jni_GetStaticLongField,
    checked_jni_GetStaticFloatField,
    checked_jni_GetStaticDoubleField,

    checked_jni_SetStaticObjectField,
    checked_jni_SetStaticBooleanField,
    checked_jni_SetStaticByteField,
    checked_jni_SetStaticCharField,
    checked_jni_SetStaticShortField,
    checked_jni_SetStaticIntField,
    checked_jni_SetStaticLongField,
    checked_jni_SetStaticFloatField,
    checked_jni_SetStaticDoubleField,

    checked_jni_NewString,
    checked_jni_GetStringLength,
    checked_jni_GetStringChars,
    checked_jni_ReleaseStringChars,

    checked_jni_NewStringUTF,
    checked_jni_GetStringUTFLength,
    checked_jni_GetStringUTFChars,
    checked_jni_ReleaseStringUTFChars,

    checked_jni_GetArrayLength,

    checked_jni_NewObjectArray,
    checked_jni_GetObjectArrayElement,
    checked_jni_SetObjectArrayElement,

    checked_jni_NewBooleanArray,
    checked_jni_NewByteArray,
    checked_jni_NewCharArray,
    checked_jni_NewShortArray,
    checked_jni_NewIntArray,
    checked_jni_NewLongArray,
    checked_jni_NewFloatArray,
    checked_jni_NewDoubleArray,

    checked_jni_GetBooleanArrayElements,
    checked_jni_GetByteArrayElements,
    checked_jni_GetCharArrayElements,
    checked_jni_GetShortArrayElements,
    checked_jni_GetIntArrayElements,
    checked_jni_GetLongArrayElements,
    checked_jni_GetFloatArrayElements,
    checked_jni_GetDoubleArrayElements,

    checked_jni_ReleaseBooleanArrayElements,
    checked_jni_ReleaseByteArrayElements,
    checked_jni_ReleaseCharArrayElements,
    checked_jni_ReleaseShortArrayElements,
    checked_jni_ReleaseIntArrayElements,
    checked_jni_ReleaseLongArrayElements,
    checked_jni_ReleaseFloatArrayElements,
    checked_jni_ReleaseDoubleArrayElements,

    checked_jni_GetBooleanArrayRegion,
    checked_jni_GetByteArrayRegion,
    checked_jni_GetCharArrayRegion,
    checked_jni_GetShortArrayRegion,
    checked_jni_GetIntArrayRegion,
    checked_jni_GetLongArrayRegion,
    checked_jni_GetFloatArrayRegion,
    checked_jni_GetDoubleArrayRegion,

    checked_jni_SetBooleanArrayRegion,
    checked_jni_SetByteArrayRegion,
    checked_jni_SetCharArrayRegion,
    checked_jni_SetShortArrayRegion,
    checked_jni_SetIntArrayRegion,
    checked_jni_SetLongArrayRegion,
    checked_jni_SetFloatArrayRegion,
    checked_jni_SetDoubleArrayRegion,

    checked_jni_RegisterNatives,
    checked_jni_UnregisterNatives,

    checked_jni_MonitorEnter,
    checked_jni_MonitorExit,

    checked_jni_GetJavaVM,

    checked_jni_GetStringRegion,
    checked_jni_GetStringUTFRegion,

    checked_jni_GetPrimitiveArrayCritical,
    checked_jni_ReleasePrimitiveArrayCritical,

    checked_jni_GetStringCritical,
    checked_jni_ReleaseStringCritical,

    checked_jni_NewWeakGlobalRef,
    checked_jni_DeleteWeakGlobalRef,

    checked_jni_ExceptionCheck,

};

#else
int dummy_check_jni_var = 0; /* shut off warnings */
#endif /* CHECK_JNI */
