/*
 * @(#)classinitialize.c	1.128 98/11/03
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*-
 *      Runtime routines to support classes.
 */

#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <stdlib.h>

#include "oobj.h"
#include "tree.h"
#include "util.h"
#include "interpreter.h"
#include "threads.h"
#include "monitor.h"
#include "exceptions.h"
#include "javaString.h"
#include "common_exceptions.h"
#include "utf.h"
#include "gc.h"

static struct fieldblock *
ResolveClassConstantField(unsigned type,
			  ClassClass *current_class,
			  ClassClass *target_class,
			  char *name,
			  char *signature,
			  struct execenv *ee);
static bool_t ResolveClassConstant0(ClassClass *Class, cp_item_type *, 
				    unsigned, struct execenv *, 
				    unsigned);

bool_t 
ResolveClassConstantFromClass(ClassClass *class, unsigned index, 
			      struct execenv *ee, unsigned mask)
{
    return ResolveClassConstantFromClass2(class, index, ee, mask, FALSE);
}


bool_t
ResolveClassConstant(cp_item_type *constant_pool, unsigned index, 
		     struct execenv *ee, unsigned mask)
{
    return ResolveClassConstant2(constant_pool, index, ee, mask, FALSE);
}

static bool_t
InitClassConstantClass(cp_item_type *constant_pool,
		       unsigned index, struct execenv *ee)
{
    ClassClass *class;
    unsigned char *type_table =
        constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
    sysAssert(CONSTANT_POOL_TYPE_TABLE_IS_RESOLVED(type_table, index));
    switch (CONSTANT_POOL_TYPE_TABLE_GET_TYPE(type_table, index)) {
    case CONSTANT_InterfaceMethodref: 
    case CONSTANT_Methodref: 
        class = constant_pool[index].mb->fb.clazz;
	break;
    case CONSTANT_Fieldref:
        class = constant_pool[index].fb->clazz;
	break;
    case CONSTANT_Class:
        class = constant_pool[index].clazz;
	break;
    default:
        class = 0;
    }
    if (class) {
        InitClass(class);
	return !exceptionOccurred(ee);
    }
    return TRUE;
}

static bool_t
ResolveClassConstantAndInit(ClassClass *class, cp_item_type *constant_pool,
			    unsigned index, 
			    struct execenv *ee, unsigned mask,
			    bool_t init)
{
    bool_t ret;

    ret = ResolveClassConstant0(class, constant_pool, index, ee, mask);
    if (!ret)
        return FALSE;

    if (init)
        ret = InitClassConstantClass(constant_pool, index, ee);

    return ret;
}

bool_t 
ResolveClassConstantFromClass2(ClassClass *class, unsigned index, 
			       struct execenv *ee, unsigned mask,
			       bool_t init)
{
    cp_item_type *constant_pool = cbConstantPool(class);
    return ResolveClassConstantAndInit(class, constant_pool, index, 
				       ee, mask, init);
}    


bool_t
ResolveClassConstant2(cp_item_type *constant_pool, unsigned index, 
		      struct execenv *ee, unsigned mask, bool_t init)
{
    ClassClass *class;
    if (ee && ee->current_frame && ee->current_frame->current_method) 
	class = fieldclass(&ee->current_frame->current_method->fb);
    else
        class = 0;
    return ResolveClassConstantAndInit(class, constant_pool, index, 
				       ee, mask, init);
}

/*
 * This routine resolves class constants like field names,
 * methodnames, and class names, to their runtime-specific values.
 * Field names are resolved to offsets from the beginning of an
 * instance, methodnames are resolved to pointers to code, and class
 * names are resolved to pointers to the class.
 *
 * This routine includes some tricky locking using the LINKCLASS lock. The 
 * idea is that when the LINKCLASS lock is entered, no code requiring any 
 * other Java or VM locks should be executed. The RESOLVE lock only ensures
 * the constant pool content is consistent w.r.t. constant pool type.
 * Multiple threads may end up resolving the same constant pool entry at 
 * the same time. This is harmless, because the constant pool will still
 * be consistent.
 */

static bool_t
ResolveClassConstant0(ClassClass *current_class, 
		      cp_item_type *constant_pool, 
		      unsigned index, 
		      struct execenv *ee, 
		      unsigned mask)
{
    unsigned char *type_table = 
	constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;

    ee = ee ? ee : EE();

    LINKCLASS_LOCK(EE2SysThread(ee));
    if (!CONSTANT_POOL_TYPE_TABLE_IS_RESOLVED(type_table, index)) {
	unsigned type = CONSTANT_POOL_TYPE_TABLE_GET_TYPE(type_table, index);
	sysAssert((mask & (1 << type)) != 0);
	switch(type) {
	  case CONSTANT_Class: {
	      unsigned name_index = constant_pool[index].i;
	      char *name = constant_pool[name_index].cp;
	      ClassClass *class = 0;

	      LINKCLASS_UNLOCK(EE2SysThread(ee));
	      class = FindClassFromClass(ee, name, FALSE, current_class);
	      if (class == 0) {
		  /* Exception already thrown */
		  return FALSE;
	      } else if (!VerifyClassAccess(current_class, class, TRUE)) {
		  char buf[MSG_BUF_LEN];
		  jio_snprintf(buf, sizeof(buf), 
			       "try to access class %s from class %s",
			       cbName(class), cbName(current_class));
		  ThrowIllegalAccessError(ee, buf);
		  return FALSE;
	      }

	      LINKCLASS_LOCK(EE2SysThread(ee));
	      /* In case the entry is resolve by another thread or when we
	       * call FindClassFromClass.
	       */
	      if (!CONSTANT_POOL_TYPE_TABLE_IS_RESOLVED(type_table, index)) {
		  constant_pool[index].clazz = class;
		  CONSTANT_POOL_TYPE_TABLE_SET_RESOLVED(type_table, index);
	      }
	      LINKCLASS_UNLOCK(EE2SysThread(ee));
	      return TRUE;
	  }

	  case CONSTANT_String:  {
	      unsigned value_index = constant_pool[index].i;
	      char *value = constant_pool[value_index].cp; /* string */
	      int length = utfstrlen(value);
	      unicode *body;
	      Hjava_lang_String *string;
	      int i;
	      /* Allocate a byte array.  Copy bytes in. */
	      HArrayOfChar *so;

	      LINKCLASS_UNLOCK(EE2SysThread(ee));

	      so = (HArrayOfChar *)allocArray(ee, T_CHAR, length);
	      if (so == 0) {
		  ThrowOutOfMemoryError(0, 0);
		  return FALSE;
	      }
	      body = unhand(so)->body;
	      utf2unicode(value, body, length, &i);
	      KEEP_POINTER_ALIVE(body);
	      string = (Hjava_lang_String *)
		  allocObject(ee, classJavaLangString);
	      if (string == 0) {
		  ThrowOutOfMemoryError(0, 0);
		  return FALSE;
	      }
	      unhand(string)->count = length;
	      unhand(string)->value = so;
	      
	      string = internString(string);
	      if (string == 0) {
		  ThrowOutOfMemoryError(0, 0);
		  return FALSE;
	      }
	      
	      LINKCLASS_LOCK(EE2SysThread(ee));
	      /* In case the entry is resolved by another thread. */
	      if (!CONSTANT_POOL_TYPE_TABLE_IS_RESOLVED(type_table, index)) {
		  constant_pool[index].str = string;
		  CONSTANT_POOL_TYPE_TABLE_SET_RESOLVED(type_table, index);
	      }
	      LINKCLASS_UNLOCK(EE2SysThread(ee));
	      return TRUE;
 	  } 
	   
	  case CONSTANT_NameAndType: {
	      /* nothing to do. */ 
	      CONSTANT_POOL_TYPE_TABLE_SET_RESOLVED(type_table, index);
	      LINKCLASS_UNLOCK(EE2SysThread(ee));
	      return TRUE;
	  }

	  case CONSTANT_InterfaceMethodref: 
	  case CONSTANT_Fieldref:
	  case CONSTANT_Methodref: {
	      unsigned key = constant_pool[index].i; /* constant pool entry */
	      unsigned class_index = key >> 16;      /* index of class */
	      unsigned name_type_index = key & 0xFFFF; /* index of name/type */
	      char *name, *signature;
	      struct fieldblock *fb = NULL;
	      ClassClass *target_class;

	      LINKCLASS_UNLOCK(EE2SysThread(ee));

	      /* First, resolve reference to the class */
	      if (!ResolveClassConstant0(current_class,
					 constant_pool,
					 class_index,
					 ee, 
					 1 << CONSTANT_Class)) {
		  return FALSE;
	      }
	      target_class = constant_pool[class_index].clazz;
	      LinkClass(target_class);
	      if (exceptionOccurred(ee)) {
		  return FALSE;
	      }

	      /* Resolving a name and type ref can't ever fail, so we don't
	       * bother to check the return value.
	       */
	      ResolveClassConstant0(current_class,
				    constant_pool,
				    name_type_index,
				    ee, 
				    1 << CONSTANT_NameAndType);

	      /* Resolve the field or method reference */
	      key = constant_pool[name_type_index].i;
	      name = constant_pool[key >> 16].cp;
	      signature = constant_pool[key & 0xFFFF].cp;

	      fb = ResolveClassConstantField(type,
					     current_class,
					     target_class,
					     name,
					     signature,
					     ee);
	      if (fb == NULL) {
		  return FALSE;
	      }

	      /* Check loader constraints */
	      if (!CheckSignatureLoaders(signature,
			      current_class ? cbLoader(current_class) : NULL,
					 cbLoader(fb->clazz))) {
		  return FALSE;
	      }

	      LINKCLASS_LOCK(EE2SysThread(ee));
	      /* In case the entry has already been resolved by another thread
	       * when we were resolving it as well */
	      if (!CONSTANT_POOL_TYPE_TABLE_IS_RESOLVED(type_table, index)) {
		  constant_pool[index].fb = fb;
		  CONSTANT_POOL_TYPE_TABLE_SET_RESOLVED(type_table, index);
	      }
	      LINKCLASS_UNLOCK(EE2SysThread(ee));

	      return TRUE;
	  }
	}
    }
    LINKCLASS_UNLOCK(EE2SysThread(ee));
    return TRUE;
} 

static struct fieldblock *
ResolveClassConstantField(unsigned type,
			  ClassClass *current_class,
			  ClassClass *target_class,
			  char *name,
			  char *signature,
			  struct execenv *ee)
{
    HashedNameAndType hashed;
    char buf[MSG_BUF_LEN];
    int len;
    ClassClass *cb;

    HashNameAndType(ee, name, signature, &hashed);

    switch (type) {
        case CONSTANT_Methodref: {
	    struct methodblock *mb;

	    if (cbIsInterface(target_class)) {
	        ThrowIncompatibleClassChangeError(ee, cbName(target_class));
		return NULL;
	    }
	    /* search in current and super classes */
	    for (cb = target_class; cb; cb = cbSuperclass(cb)) { 
		int n = cbMethodsCount(cb);
		mb = cbMethods(cb);
		for (; --n >= 0; mb++) {
		    if (NAMETYPE_MATCH(&hashed, &(mb->fb))) {
		        goto found;
		    }
		}
	    }

	    /* search for miranda methods */
	    {
	        struct methodblock **mt = cbMethodTable(target_class)->methods;
		int n;

		for (n = cbMethodTableSize(target_class) - 1; n > 0; n--) {
		    mb = mt[n];
		    if (NAMETYPE_MATCH(&hashed, &(mb->fb))) {
		        /* If the mb is in method table, but not declared,
			 * it must be a miranda method.
			 */
		        sysAssert(mb->fb.access & ACC_MIRANDA_METHOD);
			cb = mb->fb.clazz;
			goto found;
		    }
		}
	    }

	fail:
	    classname2string(cbName(target_class), buf, sizeof(buf));
	    len = strlen(buf);
	    /* On buffer overflow truncate string quietly */
	    (void) jio_snprintf(buf + len, sizeof(buf)-len,
		    ": method %s%s not found", name, signature);
	    ThrowNoSuchMethodError(ee, buf);
	    return NULL;

	found:
	    /* Check if erroneously "inheriting" a constructor */
	    if (target_class != cb)
	        if (mb->fb.name == utf8_literal_obj_init_name)
		    goto fail;

	    if (VerifyFieldAccess(current_class, cb, mb->fb.access, TRUE)) {
		return &mb->fb;
	    } else {
		jio_snprintf(buf,
			     sizeof(buf), 
			     "try to access method %s.%s%s from class %s",
			     cbName(cb),
			     name,
			     signature,
			     cbName(current_class));
		ThrowIllegalAccessError(ee, buf);
		return NULL;
	    }
	}
	
        case CONSTANT_Fieldref: {
	    for (cb = target_class; ; cb = cbSuperclass(cb)) { 
		int n = cbFieldsCount(cb);
		struct fieldblock *fb = cbFields(cb);
		for (; --n >= 0; fb++) {
		    if (NAMETYPE_MATCH(&hashed, fb)) {
			if (VerifyFieldAccess(current_class,
					      cb,
					      fb->access,
					      TRUE)) {
			    return fb;
			} else {
			    jio_snprintf(buf, sizeof(buf), 
			        "try to access field %s.%s from class %s",
			        cbName(cb), name, cbName(current_class));
			    ThrowIllegalAccessError(ee, buf);
			    return NULL;
			}
		    }
		}
		if (cbSuperclass(cb) == NULL) 
		    break;
	    }
	    classname2string(cbName(target_class), buf, sizeof(buf));
	    len = strlen(buf);
	    /* Reporting error, so truncate string quietly */
	    (void) jio_snprintf(buf + len, sizeof(buf)-len, 
				": field %s not found", name);
	    ThrowNoSuchFieldError(0, buf);
	    return NULL;
	}

        case CONSTANT_InterfaceMethodref: {
	    /* The imethod table of an interface contains this interface
	     * in the first slot, and all of the superinterfaces in the
	     * remaining slots
	     */
	    struct imethodtable *itable = cbIntfMethodTable(target_class);
	    int icount = itable->icount;
	    int i;
	    if (!cbIsInterface(target_class)) {
	        ThrowIncompatibleClassChangeError(ee, cbName(target_class));
		return NULL;
	    }
	    sysAssert(icount > 0);
	    sysAssert(itable->itable[0].classdescriptor == target_class);
	    for (i = 0; i < icount; i++) { 
		ClassClass *cb = itable->itable[i].classdescriptor;
		int n = cbMethodsCount(cb);
		struct methodblock *mb = cbMethods(cb);
		for (; --n >= 0; mb++) {
		    if (NAMETYPE_MATCH(&hashed, &(mb->fb))) {
		        return &mb->fb;
		    } 
		}
	    }
	    classname2string(cbName(target_class), buf, sizeof(buf));
	    len = strlen(buf);
	    /* On buffer overflow truncate string quietly */
	    (void) jio_snprintf(buf + len, sizeof(buf)-len,
		    ": method %s%s not found", name, signature);
	    ThrowNoSuchMethodError(ee, buf);
	    return NULL;
	}
	    
        default:
	    sysAssert(FALSE);
	    return NULL;	/* Should never be reached; keep lint happy */
    }
}


/* Verify that current_class can access new_class.  If the classloader_only
 * flag is set, we automatically allow any accesses in which current_class
 * doesn't have a classloader.
 */

bool_t
VerifyClassAccess(ClassClass *current_class, ClassClass *new_class, 
		  bool_t classloader_only) 
{
    if (current_class == NULL ||
	current_class == new_class ||
	IsPublic(cbAccess(new_class)) ||
	(oldjava && classloader_only && (cbLoader(current_class) == 0)) ||
	IsSameClassPackage(current_class, new_class)) {
        return TRUE;
    }
    return classloader_only &&
      !NEED_VERIFY(!IsTrustedClassLoader(cbLoader(current_class))) &&
      cbProtectionDomain(current_class) == cbProtectionDomain(new_class) &&
      cbLoader(current_class) == cbLoader(new_class);
}    

/* In certain cases, we only need a class's name.  We don't want to resolve
 * the class reference if it isn't, already. 
 */

char *GetClassConstantClassName(cp_item_type *constant_pool, int index)
{
    unsigned char *type_table = 
	constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
    ExecEnv *ee;
    char *name;

    if (type_table[index] & CONSTANT_POOL_ENTRY_RESOLVED) {
	return cbName(constant_pool[index].clazz);
    }

    ee = EE();
    LINKCLASS_LOCK(EE2SysThread(ee));
    switch(type_table[index]) {
        case CONSTANT_Class | CONSTANT_POOL_ENTRY_RESOLVED: {
	    ClassClass *cb = constant_pool[index].clazz;
	    name = cbName(cb);
	    break;
	}

        case CONSTANT_Class: {
	    int name_index = constant_pool[index].i;
	    name = constant_pool[name_index].cp;
	    break;
	}
	    
	default:
	    name = NULL;
    }
    LINKCLASS_UNLOCK(EE2SysThread(ee));
    return name;
}

/* Verify that current_class can a field of new_class, where that field's 
 * access bits are "access".  We assume that we've already verified that
 * class can access field_class.
 *
 * If the classloader_only flag is set, we automatically allow any accesses
 * in which current_class doesn't have a classloader.
 */


bool_t 
VerifyFieldAccess2(ClassClass *current_class, ClassClass *field_class, 
		   int access, bool_t classloader_only, bool_t protected_restriction)
{
    if (current_class == 0 ||
        current_class == field_class ||
	IsPublic(access) ||
	(oldjava && classloader_only && (cbLoader(current_class) == 0))) {
	return TRUE;
    }
    if (IsProtected(access)) {
	/* See if current_class is a subclass of field_class */
	ClassClass *cb;
	if (!protected_restriction) {
	    for (cb = cbSuperclass(current_class); cb; cb = cbSuperclass(cb)) {
	        if (cb == field_class) {
		    return TRUE;
		}
	    }
	}
    }
    if (!IsPrivate(access) && IsSameClassPackage(current_class, field_class)) {
        return TRUE;
    }
    return classloader_only &&
      !NEED_VERIFY(!IsTrustedClassLoader(cbLoader(current_class))) &&
      cbProtectionDomain(current_class) == cbProtectionDomain(field_class) &&
      cbLoader(current_class) == cbLoader(field_class);
}

bool_t
VerifyFieldAccess(ClassClass *current_class, ClassClass *field_class, 
		  int access, bool_t classloader_only)
{
    return VerifyFieldAccess2(current_class, field_class, access, classloader_only, 
			      FALSE);
}
