/*
 * @(#)classload.c	1.193 98/08/26
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*-
 *      Routines for loading and resolving class definitions.
 *
 */

#include <string.h>
#include <sys/types.h>
#include <stddef.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <setjmp.h>

#include "oobj.h"
#include "path.h"
#include "interpreter.h"
#include "util.h"
#include "monitor.h"
#include "tree.h"
#include "bool.h"
#include "signature.h"
#include "utf.h"
#include "gc.h"
#include "jlong.h"

#include "sys_api.h"
#include "byteorder.h"
#include "common_exceptions.h"

#ifdef BREAKPTS
#include "breakpoints.h"
#endif

#ifdef HPROF
#include "vmprofiler.h"
#endif /* HPROF */

#include "javaString.h"

/* Inner classes can be static, private or protected. */
#define RECOGNIZED_INNER_CLASS_MODIFIERS \
    (JVM_RECOGNIZED_CLASS_MODIFIERS | ACC_PRIVATE | ACC_PROTECTED | ACC_STATIC)

/*
 * Simple hash table used to keep track of packages loaded by the system
 * and the directory or JAR files that they were loaded from. Used in
 * java.lang.Package to maintain version information for loaded system
 * packages.
 */

#define HASHSIZE 31	    /* Number of buckets */

struct pkginfo {
    char *pkgname;	    /* Package name */
    char *filename;	    /* Directory or JAR file loaded from */
    struct pkginfo *next;   /* Next entry in hash table */
};

/* Hash table used to keep track of loaded packages */
static struct pkginfo *hashtab[HASHSIZE];
static int n_entries = 0;

/* Hash function */
static unsigned hash(const char *s, int n)
{
    unsigned val = 0;
    while (--n >= 0) {
	val = *s++ + 31 * val;
    }
    return val % HASHSIZE;
}

/*
 * Returns the package file name corresponding to the specified package
 * or class name, or null if not found.
 */
static char *
pkginfo_get(const char *name)
{
    char *cp = strrchr(name, '/');

    if (cp != 0) {
	struct pkginfo *pp;
	int n = cp - name + 1;
	for (pp = hashtab[hash(name,n)]; pp != 0; pp = pp->next) {
	    if (strncmp(name, pp->pkgname, n) == 0) {
		return pp->filename;
	    }
	}
    }
    return 0;
}

static int
pkginfo_names(char *names[], int size)
{
    int n = 0, i;

    if (n_entries > size) {
	return n_entries;
    }
    for (i = 0; i < HASHSIZE; i++) {
	struct pkginfo *pp = hashtab[i];
	while (pp != 0) {
	    names[n++] = pp->pkgname;
	    pp = pp->next;
	}
    }
    return n;
}

/*
 * Adds a new package entry for the specified class or package name and
 * corresponding directory or jar file name. If there was already an existing
 * package entry then it is replaced. Returns the previous entry or null if
 * none.
 */
static bool_t
pkginfo_put(const char *name, const char *filename)
{
    char *cp = strrchr(name, '/');

    if (cp != 0) {
	struct pkginfo *pp;
	int n = cp - name + 1;
	int i = hash(name,n);

	/* First check for previously loaded entry */
	for (pp = hashtab[i]; pp != 0; pp = pp->next) {
	    if (strncmp(name, pp->pkgname, n) == 0) {
		if ((pp->filename = sysStrdup(filename)) == 0) {
		    return FALSE;
		}
		return TRUE;
	    }
	}

	/* If not found then allocate a new entry and insert into table */
	if ((pp = sysMalloc(sizeof(struct pkginfo))) == 0) {
	    return FALSE;
	}
	if ((pp->pkgname = sysMalloc(n + 1)) == 0) {
	    sysFree(pp);
	    return FALSE;
	}
	memcpy(pp->pkgname, name, n);
	pp->pkgname[n] = '\0';
	if ((pp->filename = sysStrdup(filename)) == 0) {
	    sysFree(pp->pkgname);
	    sysFree(pp);
	    return FALSE;
	}
	pp->next = hashtab[i];
	hashtab[i] = pp;
	n_entries++;
    }
    return TRUE;
}

/*
 * If the specified package has been loaded by the system, then returns
 * the name of the directory or ZIP file that the package was loaded from.
 * Returns null if the package was not loaded.
 * Note: The specified name can either be the name of a class or package.
 * If a package name is specified, then it must be "/"-separator and also
 * end with a trailing "/".
 */
JNIEXPORT jstring JNICALL
JVM_GetSystemPackage(JNIEnv *env, jstring str)
{
    const char *name = (*env)->GetStringUTFChars(env, str, 0);
    if (name != 0) {
	sys_thread_t *self = EE2SysThread(JNIEnv2EE(env));
	char *fn;
	SYSLOADER_LOCK(self);
	fn = pkginfo_get(name);
	SYSLOADER_UNLOCK(self);
	(*env)->ReleaseStringUTFChars(env, str, name);
	if (fn != 0) {
	    return MkRefLocal(
		       env, makeJavaStringFromPlatformCString(fn, strlen(fn)));
	}
    }
    return 0;
}

/*
 * Returns an array of Java strings representing all of the currently
 * loaded system packages.
 * Note: The package names returned are "/"-separated and end with a
 * trailing "/".
 */
JNIEXPORT jobjectArray JNICALL
JVM_GetSystemPackages(JNIEnv *env)
{
    sys_thread_t *self = EE2SysThread(JNIEnv2EE(env));
    char **names;
    int size;

    SYSLOADER_LOCK(self);
    size = pkginfo_names(0, 0);
    names = sysMalloc(size * sizeof(char *));
    if (names != 0) {
	pkginfo_names(names, size);
    }
    SYSLOADER_UNLOCK(self);
    if (names != 0) {
	jobjectArray result = 0;
	jclass cls = (*env)->FindClass(env, "java/lang/String");
	if (cls != 0) {
	    result = (*env)->NewObjectArray(env, size, cls, 0);
	    while (size > 0 && !(*env)->ExceptionOccurred(env)) {
		jstring str = (*env)->NewStringUTF(env, names[--size]);
		if (str != 0) {
		    (*env)->SetObjectArrayElement(env, result, size, str);
		    (*env)->DeleteLocalRef(env, str);
		}
	    }
	}
	sysFree(names);
	return (*env)->ExceptionOccurred(env) ? 0 : result;
    }
    return 0;
}

/*
 * The second argument is:
 * 
 * 1. a real file name, when the function is invoked by local class loading;
 * 2. NULL, when the function is invoked by ClassLoader.defineClass or JNI
 *    DefineClass, or when an array class is loaded.
 */
static ClassClass *
AddBinClass(ClassClass * cb, char *filename)
{
    struct Hjava_lang_ClassLoader *loader = cbLoader(cb);
    sys_thread_t *self = sysThreadSelf();
    BINCLASS_LOCK(self);
    {
        char buf[MSG_BUF_LEN];
	char *ename = UpdateLoaderCache(cb, loader, buf, sizeof(buf));
	if (ename) {
	    BINCLASS_UNLOCK(self);
	    SignalError(0, ename, buf);
	    return NULL;
	}
    }
    if (nbinclasses >= sizebinclasses) {
        ClassClass **new_binclasses;
	if (binclasses == 0) {
	    new_binclasses = (ClassClass **)
	        sysMalloc(sizeof(ClassClass *) * (sizebinclasses = 50));
	} else {
	    new_binclasses = (ClassClass **)
	        sysRealloc(binclasses, sizeof(ClassClass *)
			   * (sizebinclasses = nbinclasses * 2));
	}
	if (new_binclasses == 0) {
	    BINCLASS_UNLOCK(self);
	    ThrowOutOfMemoryError(0, 0);
	    return NULL;
	}
	binclasses = new_binclasses;
    }
    binclasses[nbinclasses] = cb;
    nbinclasses++;
    if (verbose) {
        char clname[MSG_BUF_LEN];
	classname2string(cbName(cb), clname, sizeof(clname));
	if (filename) {
	    jio_fprintf(stderr, "[Loaded %s from %s]\n", clname, filename);
	} else {
	    jio_fprintf(stderr, "[Loaded %s]\n", clname);
	}
    }

    BINCLASS_UNLOCK(self);
#ifdef HPROF
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_CLASS_LOAD)) {
        jvmpi_load_class(cb);
    }
#endif /* HPROF */
#ifdef BREAKPTS
    if (debugging && cb) {
        notify_debugger_of_class_load(EE(), (JHandle *)cb);
    }
#endif 
    return cb;
}

void
MakeClassSticky(ClassClass *cb)
{
    CCSet(cb, Sticky);
}

/* Load a .class file normally from the local disk */
static ClassClass *
LoadClassFromFile(char *classname, char *fn, char *dir)
{
    ClassClass *cb = 0;
    int codefd;
    unsigned char *external_class;
    int size;
    jlong lsize;

    if ((codefd = sysOpen(fn, O_RDONLY, 0)) < 0)
	/* Open failed */
	return 0;

    if ((sysFileSizeFD(codefd, &lsize) < 0) ||
	jlong_ne(int_to_jlong((size = jlong_to_int(lsize))),lsize))
	/* Stat failed, or the file is just too big */
	return 0;

    /* Snarf the file into memory. */
    external_class = (unsigned char *)sysMalloc(size);
    if (external_class == 0)
        goto failed;

    if (sysRead(codefd, external_class, size) != size) {
	sysFree(external_class);
        goto failed;
    }

    sysClose(codefd);
    codefd = -1;

    cb = createInternalClass(external_class, external_class + size, 
			     NULL, classname, fn);
    sysFree(external_class);
    if (cb == 0) {
        /* exception already thrown */
	goto failed;
    }
    if (pkginfo_get(classname) == 0) {
	/* update table of loaded packages */
	if (!pkginfo_put(classname, dir)) {
	    goto failed;
	}
    }
    return cb;
failed:
    if (codefd >= 0)
	sysClose(codefd);
    return 0;
}

typedef void * * (JNICALL *ZipOpen_t)(const char *name, char **pmsg);
typedef void * * (JNICALL *FindEntry_t)(jzfile *zip, const char *name, 
					    jint *sizeP, jint *nameLen);
typedef jboolean (JNICALL *ReadEntry_t)
     (jzfile *zip, jzentry *entry, unsigned char *buf, char *namebuf);

static ZipOpen_t ZipOpen;
static FindEntry_t FindEntry;
static ReadEntry_t ReadEntry;

bool_t
LoadZipLibrary(void)
{
    char path[MAXPATHLEN];
    char load_error[STK_BUF_LEN];
    void *handle;
    sysBuildLibName(path, sizeof(path), java_dll_dir, "zip");
    handle = sysLoadLibrary(path, load_error, sizeof(load_error));
    if (handle == 0) {
        jio_fprintf(stderr, "Could not load library \"%s\", because %s\n",
		    path, load_error);
	return FALSE;
    }
    ZipOpen = (ZipOpen_t)sysFindLibraryEntry(handle, "ZIP_Open");
    FindEntry = (FindEntry_t)sysFindLibraryEntry(handle, "ZIP_FindEntry");
    ReadEntry = (ReadEntry_t)sysFindLibraryEntry(handle, "ZIP_ReadEntry");
    if (ZipOpen == 0 || FindEntry == 0 || ReadEntry == 0) {
        jio_fprintf(stderr, "Corrupted ZIP library '%s'\n", path);
	return FALSE;
    }
    return TRUE;
}

/* Load a .class file normally from a zip archive */
static ClassClass *
LoadClassFromZip(char *classname, char *fn, cpe_t *cpe)
{
    jzentry *entry;
    ClassClass *cb;
    unsigned char *clazz;
    jint filesize, name_len;
    char namebuf[MAXPATHLEN]; /* should be big enough for ZIP entries */
    char *filename = namebuf;

    entry = (*FindEntry)(cpe->zip, fn, &filesize, &name_len);
    if (entry == 0) {
	return 0;
    }
    if ((clazz = sysMalloc(filesize)) == 0) {
	return 0;
    }
    if (!(*ReadEntry)(cpe->zip, entry, clazz, filename)) {
	sysFree(clazz);
	return 0;
    }
    cb = createInternalClass(clazz, clazz + filesize, 0, classname, cpe->path);
    sysFree(clazz);
    if (cb == 0) {
        /* exception already thrown */
	return 0;
    }
    if (pkginfo_get(classname) == 0) {
	/* update table of loaded packages */
	if (!pkginfo_put(classname, cpe->path)) {
	    return 0;
	}
    }
    return cb;
}

cpe_t **
GetClassPath(void)
{
    static cpe_t **classpath;

    if (classpath == 0) {
	char *cps, *s;
	int ncpe = 1;
	cpe_t **cpp;
        int sep_len = strlen(PATH_SEPARATOR);

	if (init_sysclasspath == NULL) {
	    return 0;
	}
	if ((cps = sysStrdup(init_sysclasspath)) == 0) {
	    return 0;
	}
	for (s = cps; (s = strstr(s, PATH_SEPARATOR)) != 0; s += sep_len) {
            ncpe++;
	}
	cpp = classpath = sysMalloc((ncpe + 1) * sizeof(cpe_t *));
	if (cpp == 0) {
	    return 0;
	}
	while (cps && *cps) {
	    struct stat sbuf;
	    char *path = cps;
	    int type;
	    if ((cps = strstr(cps, PATH_SEPARATOR)) != 0) {
		*cps = '\0';
		cps += sep_len;
	    }
	    if (*path == '\0') {
		path = PATH_CURDIR;
	    }
	    if ((type = sysFileType(path)) < 0) {
		/* File or directory does not exist */
	        continue;
	    }
	    if (type == SYS_FILETYPE_REGULAR) {
		char *msg = 0;
		jzfile *zip;
		jlong t = sysTimeMillis();
		zip = (*ZipOpen)(path, &msg);
		if (verbose) {
		    jio_fprintf(stderr, "[Opened %s in %u ms]\n", path,
				jlong_to_uint(jlong_sub(sysTimeMillis(),t)));
		}
		if (zip != 0) {
		    cpe_t *cpe = sysMalloc(sizeof(cpe_t));
		    if (cpe == 0) {
			return 0;
		    }
		    cpe->type = CPE_ZIP;
		    cpe->zip = zip;
		    cpe->path = path;
		    *cpp++ = cpe;
		} else {
#ifdef DEBUG
		    jio_fprintf(stderr, "%s: %s\n", path,
					msg != 0 ? msg : strerror(errno));
#endif
		}
	    } else if (type == SYS_FILETYPE_DIRECTORY) {
		cpe_t *cpe = sysMalloc(sizeof(cpe_t));
		if (cpe == 0) {
		    return 0;
		}
		cpe->type = CPE_DIR;
		cpe->path = path;
		*cpp++ = cpe;
	    } else {
		/* Neither a file nor a directory, so just ignore it */
		continue;
	    }
	}
	*cpp = 0;
    }
    return classpath;
}

/*
 * Find a class file that's somewhere local, and not from a classloader.  
 * We still need to search for it using the classpath.
 */

ClassClass *LoadClassLocally(char *name)
{
    ClassClass *cb = 0;
    cpe_t **cpp;

    if (name[0] == DIR_SEPARATOR || name[0] == SIGNATURE_ARRAY) 
	return 0;

    for (cpp = GetClassPath(); cpp && *cpp != 0; cpp++) {
	cpe_t *cpe = *cpp;
	char path[MAXPATHLEN];
	if (cpe->type == CPE_DIR) {
	    if (jio_snprintf(path, sizeof(path), 
			     "%s%c%s." JAVAOBJEXT, cpe->path,
			     LOCAL_DIR_SEPARATOR, name) == -1) {
		return 0;
	    }
	    if ((cb = LoadClassFromFile(name, sysNativePath(path), cpe->path))) {
		return cb;
	    }
	} else if (cpe->type == CPE_ZIP) {
	    if (jio_snprintf(path, sizeof(path), "%s." JAVAOBJEXT, name) == -1){
		return 0;
	    }
	    if ((cb = LoadClassFromZip(name, path, cpe))) {
		return cb;
	    }
	}
    }
    return cb;
}

struct CICcontext {
    unsigned char *ptr;
    unsigned char *end_ptr;
    ClassClass *cb;
    jmp_buf jump_buffer;

    /*
     * We keep track of 2 regions of memory:
     *     the static initializer for the class (could free after it's run),
     *     and everything else.
     */
    int in_clinit;	/* whether we are loading the <clinit> method */

    unsigned char *clinit_space; /* used to hold the <clinit> method */
    unsigned char *main_space; /*  other classload space */

    /* in case we need the constant pool to clean up */
    cp_item_type *constant_pool;
    unsigned char *type_table;

    class_size_info offset;
    class_size_info size;
};

typedef struct CICcontext CICcontext;

static char *getAsciz(CICcontext *);


static unsigned long get1byte(CICcontext *);
static unsigned long get2bytes(CICcontext *);
static unsigned long get4bytes(CICcontext *);
static void getNbytes(CICcontext *, size_t count, char *buffer);

static char *getUTF8String(ExecEnv * ee, CICcontext *);
static void freeBuffers(CICcontext *);
static void freeUTF8EntriesFromContext(CICcontext* context);
static void freeUTF8EntriesFromClassClass(ClassClass* cb);
static void freeUTF8Entries(cp_item_type* constant_pool,
			    unsigned char* type_table,
			    int nconstants);
static void LoadConstantPool(CICcontext *);
static void ReadInCode(CICcontext *, struct methodblock *);
static uint32_t ReadLineTable(CICcontext *, struct methodblock *mb, uint32_t);
static void SortLineTable(struct methodblock *mb);
static void ReadExceptions(CICcontext *, struct methodblock *);
#ifdef JCOV
static void ReadCovTable(CICcontext *, struct methodblock *mb);
static void AddCovTableForNative(CICcontext *, ClassClass *cls);
static void CreateCovTable(CICcontext *, struct methodblock *mb);
#endif
static uint32_t ReadLocalVars(CICcontext *, struct methodblock *mb, uint32_t);

static void CLnomem(CICcontext *context);
static void allocClassSpace(CICcontext *context);
static void InitializeStaticVar(struct fieldblock *fb, CICcontext *context);

static ClassClass *createInternalClass0(unsigned char *ptr, 
					unsigned char *end_ptr, 
					struct Hjava_lang_ClassLoader *loader, 
					char *name,
					char *filename);


ClassClass *createInternalClass(unsigned char *ptr, 
				unsigned char *end_ptr, 
				struct Hjava_lang_ClassLoader *loader, 
				char *name,
				char *filename)
{
    ClassClass *cb;
#ifdef HPROF
    unsigned char *saved_ptr = ptr;

    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_CLASS_LOAD_HOOK)) {        
        jvmpi_load_class_hook(&ptr, &end_ptr, hpi_memory_interface->Malloc);
	if (ptr == NULL) {
	    ThrowOutOfMemoryError(0, 0);
	    return NULL;
	}
    }
#endif /* HPROF */

    cb = createInternalClass0(ptr, end_ptr, loader, name, filename);

#ifdef HPROF
    if (ptr != saved_ptr) {
        sysFree(ptr);
    }
#endif /* HPROF */

    return cb;
}

/* Create an internal class file from the indicated data.  It should be
 * in a buffer for which the first byte is at *ptr and the last byte is
 * just before *end_ptr. The class's classloader is indicated by the 
 * classloader argument.
 *
 * Since all storage needed by the class initialization method <clinit>
 * can be freed after the class is loaded, we store the <clinit> method
 * in a separate chunk of memory.
 */

static ClassClass *
createInternalClass0(unsigned char *ptr, 
		     unsigned char *end_ptr, 
		     struct Hjava_lang_ClassLoader *loader, 
		     char *name,
		     char *filename)
{
    ClassClass *cb;
    struct CICcontext context_block;
    struct CICcontext *context = &context_block;
    struct Classjava_lang_Class *ucb;
    int i, j;
    struct fieldblock *fb;
    struct methodblock *mb;
    unsigned int attribute_count;
    static check_format_fn_t check_format_fn;
    check_format_fn = 
        (check_format_fn_t)FindBuiltinEntry("VerifyClassFormat");
    if (check_format_fn == 0) {
        ThrowInternalError(0, "can't find class file format checker");
	return NULL;
    }

    memset(context, 0, sizeof(CICcontext));

    {
        char buf[200];
	jboolean measure_only =
	    NEED_VERIFY(loader != NULL) ? JNI_FALSE : JNI_TRUE;
	jboolean check_relaxed =
	    NEED_VERIFY(!IsTrustedClassLoader(loader)) ? JNI_FALSE : JNI_TRUE;
        jint res = (*check_format_fn)(name, ptr, end_ptr - ptr, 
				      &(context->size), 
				      buf, sizeof(buf),
				      measure_only,
				      check_relaxed);
	if (res == -1) { /* nomem */
	    ThrowOutOfMemoryError(0, 0);
	    return NULL;
	}
	if (res == -2) { /* format error */
	    ThrowClassFormatError(0, buf);
	    return NULL;
	}
	if (res == -3) { /* unsupported class version error */
	    ThrowUnsupportedClassVersionError(0, buf);
	    return NULL;
	}
	if (res == -4) { /* bad name */
	    ThrowNoClassDefFoundError(0, buf);
	    return NULL;
	}
    }

    allocClassSpace(context);

#ifdef HPROF
    /* Make sure we don't dump a partial class */
    if (jvmpi_event_flags) {
        jvmpi_dump_context_lock();
    }
#endif /* HPROF */

    if ((cb = allocClass(EE())) == NULL) {
        ThrowOutOfMemoryError(0, 0);
#ifdef HPROF
	if (jvmpi_event_flags) {
	    jvmpi_dump_context_unlock();
	}
#endif /* HPROF */
	return NULL;
    }

    cbLoader(cb) = loader;
    ucb = unhand(cb);

    /* Set up the context */
    context->ptr = ptr;
    context->end_ptr = end_ptr;
    context->cb = cb;

    if (setjmp(context->jump_buffer)) {
	/* We've gotten an error of some sort */
        /* We need to zero out these fields before freeing the memory
	 * buffers so that the garbage collector won't try to traverse
	 * freed memory.
         */
        cbConstantPool(cb) = NULL;
        cbFields(cb) = NULL;

	freeUTF8EntriesFromContext(context);
	freeBuffers(context);
#ifdef HPROF
	if (jvmpi_event_flags) {
	    jvmpi_dump_context_unlock();
	}
#endif /* HPROF */
	return NULL;
    }

    get4bytes(context); /* magic number */

    ucb->minor_version = get2bytes(context);
    ucb->major_version = get2bytes(context);
    ucb->loader = loader;

    LoadConstantPool(context);

    ucb->access = get2bytes(context) & JVM_RECOGNIZED_CLASS_MODIFIERS;

    /* This is a workaround for a javac bug. Interfaces are not
     * marked as ABSTRACT.
     */
    if (ucb->access & ACC_INTERFACE) {
        ucb->access |= ACC_ABSTRACT;
    }

    /* Get the name of the class */
    i = get2bytes(context);	/* index in constant pool of class */
    ucb->name = GetClassConstantClassName(context->constant_pool, i);
    context->constant_pool[i].clazz = cb;
    CONSTANT_POOL_TYPE_TABLE_SET_RESOLVED(context->type_table, i);

    /* Get the super class name. */
    i = get2bytes(context);	/* index in constant pool of class */
    if (i > 0) {
	ucb->super_name = 
	    GetClassConstantClassName(context->constant_pool, i);
    }

    i = ucb->implements_count = get2bytes(context);
    if (i > 0) {
	unsigned j;
	ucb->implements = (unsigned short *)
	    (context->main_space + context->offset.interfaces);
	for (j = 0; j < i; j++) {
	    ucb->implements[j] = get2bytes(context);
	}
    }

    if ((ucb->fields_count = get2bytes(context)) > 0) 
        ucb->fields = (struct fieldblock *)
	    (context->main_space + context->offset.fields);
    for (i = ucb->fields_count, fb = ucb->fields; --i >= 0; fb++) {
	fieldclass(fb) = cb;
	fb->access = get2bytes(context) & JVM_RECOGNIZED_FIELD_MODIFIERS;
	fb->name = getAsciz(context);
	fb->signature = getAsciz(context);
	attribute_count = get2bytes(context);
	for (j = 0; j < (int)attribute_count; j++) {
	    char *name = getAsciz(context);
	    unsigned long length = get4bytes(context);
	    if ((fb->access & ACC_STATIC) 
		 && name == utf8_literal_ConstantValue) {
		fb->access |= ACC_VALKNOWN;
		fb->u.offset = get2bytes(context);
	    } else {
		getNbytes(context, length, NULL);
	    }
	}
	if (fb->access & ACC_STATIC) {
	    InitializeStaticVar(fb, context);
	}
    }

    if ((ucb->methods_count = get2bytes(context)) > 0) 
	ucb->methods = (struct methodblock *)
	    (context->main_space + context->offset.methods);
    for (i = ucb->methods_count, mb = ucb->methods; --i >= 0; mb++) {
	fieldclass(&mb->fb) = cb;
	mb->fb.access = get2bytes(context) & JVM_RECOGNIZED_METHOD_MODIFIERS;
	mb->fb.name = getAsciz(context);
	mb->fb.signature = getAsciz(context);

	if (mb->fb.name == utf8_literal_cls_init_name &&
	    mb->fb.signature == utf8_literal_null_init_sig) {
	    /* The VM ignores the access flags of <clinit>. We reset the
	     * access flags to avoid future errors in the VM */
	    mb->fb.access = ACC_STATIC;
	    context->in_clinit = TRUE;
	}

	mb->args_size = Signature2ArgsSize(mb->fb.signature) 
	                + ((mb->fb.access & ACC_STATIC) ? 0 : 1);
	attribute_count = get2bytes(context);
	for (j = 0; j < attribute_count; j++) {
	    char *name = getAsciz(context);
	    if (CompiledCodeAttribute != NULL 
		     && strcmp(name, CompiledCodeAttribute) == 0
		     && cbLoader(cb) == 0) {
		unsigned long attribute_length = get4bytes(context);
		ReadInCompiledCode(context, mb, attribute_length, 
				   (unsigned long (*)(void *))get1byte,
				   (unsigned long (*)(void *))get2bytes,
				   (unsigned long (*)(void *))get4bytes,
				   (void (*)(void *, size_t, char *))getNbytes);
	    } else if ((name == utf8_literal_Code)
		       && !(mb->fb.access & (ACC_NATIVE | ACC_ABSTRACT))) {
		ReadInCode(context, mb);
	    } else if (name == utf8_literal_Exceptions) {
		ReadExceptions(context, mb);
	    } else {
		unsigned long length = get4bytes(context);
		getNbytes(context, length, NULL);
	    }
	}
	context->in_clinit = FALSE;
    }

    /* See if there are class attributes */
    attribute_count = get2bytes(context); 
    for (j = 0; j < attribute_count; j++) {
	char *name = getAsciz(context);
	unsigned long length = get4bytes(context);
	if (name == utf8_literal_SourceFile) {
	    ucb->source_name = getAsciz(context);
#ifdef JCOV
	} else if (name == utf8_literal_AbsoluteSourcePath) {
	    if (length == 2) {
	    	ucb->absolute_source_name = getAsciz(context);
	    } else
		getNbytes(context, length, NULL);
	} else if (name == utf8_literal_TimeStamp) {
	    unsigned high;
	    unsigned low;
	    int64_t  value;
	    jvalue    t1;

	    if (length == 8) {
		high = get4bytes(context);
		low  = get4bytes(context);
	    	value = ll_add(ll_shl(uint2ll(high), 32), uint2ll(low));
	    	SET_INT64(t1, &ucb->timestamp, value);
	    } else 
	    	getNbytes(context, length, NULL);
#endif
	} else if (name == utf8_literal_InnerClasses) {
	    int k = 0;
	    int number_of_classes = get2bytes(context);
	    ucb->innerclass_count = number_of_classes;
	    ucb->innerclasses = (struct innerclass_info *)
		(context->main_space + context->offset.innerclasses);
	    for (k = 0; k < number_of_classes; k++) {
	        unsigned short cb_access;
		ucb->innerclasses[k].inner_class_info_index =
		    get2bytes(context);
		ucb->innerclasses[k].outer_class_info_index =
		    get2bytes(context);
		ucb->innerclasses[k].inner_name_index =
		    get2bytes(context);
 		cb_access = get2bytes(context) &
		    RECOGNIZED_INNER_CLASS_MODIFIERS;
		/* This is a workaround for a javac bug. Interfaces are not
		 * marked as ABSTRACT.
		 */
		if (cb_access & ACC_INTERFACE) {
		    cb_access |= ACC_ABSTRACT;
		}
		ucb->innerclasses[k].inner_class_access_flags = cb_access;
	    }
	} else {
	    getNbytes(context, length, NULL);
	}
    }
#ifdef JCOV
    AddCovTableForNative(context, cb);
#endif

    /* The class is loaded and verified to be structually correct.
     * From this point on the responsibility of freeing malloc'ed
     * memory will be on FreeClass.
     *
     * If any class format error occurred before this, the control
     * flow would have been caught by setjmp and the function would
     * have returned NULL;
     */

    /* Save sysAlloced space for cleanup in FreeClass */
    cbClinitSpace(cb) = context->clinit_space;
    cbMainSpace(cb) = context->main_space;

    /* Valid class - let's put it in the class table. */
    {
        ClassClass *cb0 = AddBinClass(cb, filename);
	if (cb0 != cb) {
	    GCFreeClass(cb);
	}
#ifdef HPROF
	if (jvmpi_event_flags) {
	    jvmpi_dump_context_unlock();
	}
#endif /* HPROF */
	return cb0;
    }
}

/* Create a fake array class that has the specified fields */

ClassClass *
createFakeArrayClass(char *name,           /* name */
		     int base_type,        /* base_type */
		     int depth,            /* array dimension */
		     ClassClass *inner_cb, /* base type if T_CLASS */
		     struct Hjava_lang_ClassLoader *loader)
{ 
    ClassClass *cb;
    Classjava_lang_Class *ucb;
    int malloc_len;
    cp_item_type *constant_pool;
    unsigned char *type_table;
    ExecEnv * ee = EE();

    cb = allocClass(ee);
    if (cb == NULL) {
        ThrowOutOfMemoryError(ee, 0);
	return NULL;
    }

    ucb = unhand(cb);
    malloc_len = CONSTANT_POOL_ARRAY_LENGTH * 
        (sizeof(cp_item_type) + sizeof(unsigned char));
    constant_pool = sysCalloc(malloc_len, 1);
    if (constant_pool == 0) {
        ThrowOutOfMemoryError(ee, 0);
	return NULL;
    }
    ucb->classload_space.main = (unsigned char *)constant_pool;
    type_table = (unsigned char *)(constant_pool + CONSTANT_POOL_ARRAY_LENGTH);
    sysAssert(name[0] == SIGNATURE_ARRAY);
    ucb->major_version = JAVA_VERSION;
    ucb->minor_version = JAVA_MINOR_VERSION;
    ucb->name = AddUTF8(ee, name);
    if (ucb->name == NULL) {
	sysFree(constant_pool);
        ThrowOutOfMemoryError(0, 0);
	return NULL;
    }
    ucb->super_name = AddUTF8(ee, "java/lang/Object");
    if (ucb->name == NULL) {
        ReleaseUTF8(ee, ucb->name);
	sysFree(constant_pool);
        ThrowOutOfMemoryError(ee, 0);
	return NULL;
    }
    ucb->loader = loader;

    constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type = type_table;

    constant_pool[CONSTANT_POOL_ARRAY_DEPTH_INDEX].i = depth;     
    constant_pool[CONSTANT_POOL_ARRAY_TYPE_INDEX].i = base_type;
    type_table[CONSTANT_POOL_ARRAY_DEPTH_INDEX] = 
	CONSTANT_Integer | CONSTANT_POOL_ENTRY_RESOLVED;
    type_table[CONSTANT_POOL_ARRAY_TYPE_INDEX] = 
	CONSTANT_Integer | CONSTANT_POOL_ENTRY_RESOLVED;

    constant_pool[CONSTANT_POOL_ARRAY_CLASSNAME_INDEX].cp = ucb->name;
    constant_pool[CONSTANT_POOL_ARRAY_SUPERNAME_INDEX].cp = ucb->super_name;
    type_table[CONSTANT_POOL_ARRAY_CLASSNAME_INDEX] = 
	CONSTANT_Utf8 | CONSTANT_POOL_ENTRY_RESOLVED;
    type_table[CONSTANT_POOL_ARRAY_SUPERNAME_INDEX] = 
	CONSTANT_Utf8 | CONSTANT_POOL_ENTRY_RESOLVED;
	        
    if (base_type == T_CLASS) {
	/* Initialize the appropriate fields of the constant pool */
	constant_pool[CONSTANT_POOL_ARRAY_CLASS_INDEX].clazz = inner_cb;
	type_table[CONSTANT_POOL_ARRAY_CLASS_INDEX] = 
	    CONSTANT_Class | CONSTANT_POOL_ENTRY_RESOLVED;
	/* The class is public iff its base class is public */
	ucb->access = ACC_FINAL | ACC_ABSTRACT | (cbAccess(inner_cb) & ACC_PUBLIC);
    } else {
	/* Set the class field to something innocuous */
	type_table[CONSTANT_POOL_ARRAY_CLASS_INDEX] = 
	    CONSTANT_Integer | CONSTANT_POOL_ENTRY_RESOLVED;
	ucb->access = ACC_FINAL | ACC_ABSTRACT | ACC_PUBLIC;
    }
    /* set these up the last, so that GC sees a consistent state. */
    ucb->constantpool = constant_pool;
    ucb->constantpool_count = CONSTANT_POOL_ARRAY_LENGTH;

    ucb->innerclass_count = 0;     	/* Just to be very sure. */
    ucb->innerclasses = NULL;    	/* Just to be very sure. */

    {
        ClassClass *cb0 = AddBinClass(cb, NULL);
	if (cb0 != cb) {
	    GCFreeClass(cb);
	}
	if (cb0) {
	    return AddToLoadedClasses(EE(), cb0);
	}
	return cb0;
    }
}

/*
 * Create a classblock to represent a primitive type.
 * NOTE: We don't add it to the builtin class table, so this should be
 * called only once per primitive type.
 * See FindPrimitiveClass().
 */
ClassClass *
createPrimitiveClass(char *name, char sig, unsigned char typecode,
    unsigned char slotsize, unsigned char elementsize)
{
    ExecEnv * ee = EE();
    ClassClass *cb = allocClass(ee);
    Classjava_lang_Class *ucb = unhand(cb);
    cp_item_type *constant_pool; 
    unsigned char *type_table;

    int malloc_len = 3 * (sizeof(cp_item_type) + sizeof(unsigned char));

    constant_pool = sysCalloc(malloc_len, 1);
    if (constant_pool == 0) {
        ThrowOutOfMemoryError(0, 0);
	return NULL;
    }
    ucb->classload_space.main = (unsigned char *)constant_pool;
    type_table = (unsigned char*)(constant_pool + 3);

    constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type = type_table;
    ucb->loader = NULL;
    ucb->access = ACC_FINAL | ACC_ABSTRACT | ACC_PUBLIC;

    ucb->major_version = JAVA_VERSION;
    ucb->minor_version = JAVA_MINOR_VERSION;
    ucb->name = AddUTF8(ee, name);
    if (ucb->name == NULL) {
	sysFree(constant_pool);
        ThrowOutOfMemoryError(0, 0);
	return NULL;
    }
    ucb->super_name = AddUTF8(ee, "java/lang/Object");
    if (ucb->name == NULL) {
        ReleaseUTF8(ee, ucb->name);
	sysFree(constant_pool);
        ThrowOutOfMemoryError(0, 0);
	return NULL;
    }
    constant_pool[1].cp = ucb->name;
    constant_pool[2].cp = ucb->super_name;
    type_table[1] = CONSTANT_Utf8 | CONSTANT_POOL_ENTRY_RESOLVED;
    type_table[2] = CONSTANT_Utf8 | CONSTANT_POOL_ENTRY_RESOLVED;
    
    CCSet(cb, Primitive);
    cbTypeSig(cb) = sig;
    cbTypeCode(cb) = typecode;
    cbSlotSize(cb) = slotsize;
    cbElementSize(cb) = elementsize;
    MakeClassSticky(cb);

    ucb->constantpool = constant_pool;
    ucb->constantpool_count = 1;

    ucb->innerclass_count = 0;     	/* Just to be very sure. */
    ucb->innerclasses = NULL;    	/* Just to be very sure. */

#ifdef HPROF
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_CLASS_LOAD)) {
        jvmpi_load_class(cb);
    }
#endif /* HPROF */
    return cb;
}


/* ResolveClassConstantString is called by the classloader to resolve the
 * constant pool entry, before assigning a value to a final [i.e. constant]
 * String.
 */

static bool_t
ResolveClassConstantString(ClassClass *cb, unsigned index)
{
    /* This function is called before the verifier has ever had a chance
     * to look at the class.  So we should be extra careful.
     */
    union cp_item_type *constant_pool = cbConstantPool(cb);
    unsigned char *type_table = 
	constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
    if (type_table[index] == (CONSTANT_String | CONSTANT_POOL_ENTRY_RESOLVED))
	return TRUE; 
    if (type_table[index] == CONSTANT_String) { 
	return ResolveClassConstant(constant_pool, index, 0,
				    1 << CONSTANT_String);
    } 
    sysAssert(FALSE);
    return FALSE; /* never reached. */
}

static void InitializeStaticVar(struct fieldblock *fb, CICcontext *context) 
{
    jvalue t1, t2;
    char isig = fb->signature[0];
    if (fb->access & ACC_VALKNOWN) { 
	ClassClass *cb = context->cb;
	int index = fb->u.offset; /* set by "VALUE" attribute */
	union cp_item_type *constant_pool = cbConstantPool(cb);

	switch (isig) { 
	    case SIGNATURE_DOUBLE: 
		twoword_static_address(fb) = 
		    context->main_space + context->offset.fields2;
		context->offset.fields2 += sizeof(jvalue);
		sysAssert(context->offset.fields2 <= 
			  context->offset.main.code);
		SET_DOUBLE(t1, twoword_static_address(fb),
			   GET_DOUBLE(t2, &constant_pool[index]));
		break;

	    case SIGNATURE_LONG: 
		twoword_static_address(fb) = 
		    context->main_space + context->offset.fields2;
		context->offset.fields2 += sizeof(jvalue);
		sysAssert(context->offset.fields2 <= 
			  context->offset.main.code);
		SET_INT64(t1, twoword_static_address(fb),
			  GET_INT64(t2, &constant_pool[index]));
		break;

	    case SIGNATURE_CLASS:
		if (!ResolveClassConstantString(cb, index)) { 
		    /* exception already thrown */
		    longjmp(context->jump_buffer, 1);
		} else {
		    *(void **)normal_static_address(fb) = 
			   constant_pool[index].str;
		}
		break;

	    case SIGNATURE_BYTE: 
		*(int *)normal_static_address(fb) = 
		        (signed char)(constant_pool[index].i);
		break;

	    case SIGNATURE_CHAR:
		*(int *)normal_static_address(fb) = 
		        (unsigned short)(constant_pool[index].i);
		break;

	    case SIGNATURE_SHORT:
		*(int *)normal_static_address(fb) = 
		        (signed short)(constant_pool[index].i);
		break;
	    case SIGNATURE_BOOLEAN:
		*(int *)normal_static_address(fb) = 
		       (constant_pool[index].i != 0);
		break;

	    case SIGNATURE_INT:  
		*(int *)normal_static_address(fb) = constant_pool[index].i;
		break;
		
	    case SIGNATURE_FLOAT:
		*(float *)normal_static_address(fb) = constant_pool[index].f;
		break;
			
	    default: 
	        sysAssert(FALSE);
	}
    } else { 
	switch (isig) { 
	    default:
		*(int *)normal_static_address(fb) = 0;
		break;

	    case SIGNATURE_FLOAT:
		*(float *)normal_static_address(fb) = 0.0;
		break;

	    case SIGNATURE_LONG: 
		twoword_static_address(fb) = 
		    context->main_space + context->offset.fields2;
		context->offset.fields2 += sizeof(jvalue);
		sysAssert(context->offset.fields2 <=
			  context->offset.main.code);
		SET_INT64(t1, twoword_static_address(fb), ll_zero_const);
		break;

	    case SIGNATURE_DOUBLE: 
		twoword_static_address(fb) = 
		    context->main_space + context->offset.fields2;
		context->offset.fields2 += sizeof(jvalue);
		sysAssert(context->offset.fields2 <= 
			  context->offset.main.code);
		SET_DOUBLE(t1, twoword_static_address(fb), 0.0);
		break;

	    case SIGNATURE_CLASS:
		*normal_static_address(fb) = 0;
		break;
	}
    }
}

static void
allocClassSpace(CICcontext *context)
{
    unsigned int offset = 0;
    context->offset.constants = offset;
    offset += ALIGN_NATURAL(sizeof(cp_item_type) * context->size.constants) +
        ALIGN_NATURAL(sizeof(unsigned char) * context->size.constants);
    context->offset.fields = offset;
    offset += ALIGN_NATURAL(sizeof(struct fieldblock) * context->size.fields);
    context->offset.methods = offset;
    offset += ALIGN_NATURAL(sizeof(struct methodblock) * context->size.methods);
    context->offset.interfaces = offset;
    offset += ALIGN_NATURAL(sizeof(unsigned short) * context->size.interfaces);
    context->offset.fields2 = offset;
    offset += ALIGN_NATURAL(sizeof(int64_t) * context->size.fields2);
    context->offset.innerclasses = offset;
    offset += ALIGN_NATURAL(sizeof(struct innerclass_info) *
		     context->size.innerclasses);
    context->offset.main.code = offset;
    offset += ALIGN_NATURAL(context->size.main.code);
    context->offset.main.excs = offset;
    offset += ALIGN_NATURAL(sizeof(unsigned short) * context->size.main.excs);
    context->offset.main.etab = offset;
    offset += ALIGN_NATURAL(sizeof(struct CatchFrame) * context->size.main.etab);
    context->offset.main.lnum = offset;
    offset += ALIGN_NATURAL(sizeof(struct lineno) * context->size.main.lnum);
    context->offset.main.lvar = offset;
    offset += ALIGN_NATURAL(sizeof(struct localvar) * context->size.main.lvar);

    context->main_space = sysCalloc(offset, 1);
    if (context->main_space == 0)
        CLnomem(context);

    offset = 0;
    context->offset.clinit.code = offset;
    offset += ALIGN_NATURAL(context->size.clinit.code);
    context->offset.clinit.excs = offset;
    offset += ALIGN_NATURAL(sizeof(unsigned short) * context->size.clinit.excs);
    context->offset.clinit.etab = offset;
    offset += ALIGN_NATURAL(sizeof(struct CatchFrame) * context->size.clinit.etab);
    context->offset.clinit.lnum = offset;
    offset += ALIGN_NATURAL(sizeof(struct lineno) * context->size.clinit.lnum);
    context->offset.clinit.lvar = offset;
    offset += ALIGN_NATURAL(sizeof(struct localvar) * context->size.clinit.lvar);

    context->clinit_space = sysCalloc(offset, 1);
    if (context->clinit_space == 0)
        CLnomem(context);
}

static void LoadConstantPool(CICcontext *context) 
{
    ClassClass *cb = context->cb;
    int nconstants = get2bytes(context);
    unsigned char *type_table;
    cp_item_type *constant_pool;
    int i;
    jvalue t1;
    ExecEnv * ee = EE();

    sysAssert(nconstants == context->size.constants);

    constant_pool = (cp_item_type *) 
        (context->main_space + context->offset.constants);
    type_table = (unsigned char *)(constant_pool + nconstants);

    /* set up in case we have to free buffers on error */
    context->constant_pool = constant_pool;
    context->type_table = type_table;

    /* Read in the constant pool */
    for (i = CONSTANT_POOL_UNUSED_INDEX; i < nconstants; i++) {
	int type = get1byte(context);
	CONSTANT_POOL_TYPE_TABLE_PUT(type_table, i, type);
	switch (type) {
	  case CONSTANT_Utf8: {
	      char* result = getUTF8String(ee, context);
	      constant_pool[i].cp = result;
	      CONSTANT_POOL_TYPE_TABLE_SET_RESOLVED(type_table, i);
	      break;
	  }
	
	  case CONSTANT_Class:
	  case CONSTANT_String:
	    constant_pool[i].i = get2bytes(context);
	    break;

	  case CONSTANT_Fieldref:
	  case CONSTANT_Methodref:
	  case CONSTANT_InterfaceMethodref:
	    constant_pool[i].i = get4bytes(context);
	    break;

	  case CONSTANT_NameAndType:
	    constant_pool[i].i = get4bytes(context);
	    CONSTANT_POOL_TYPE_TABLE_SET_RESOLVED(type_table, i);
	    break;
	    
	  case CONSTANT_Integer:
	  case CONSTANT_Float:
	    constant_pool[i].i = get4bytes(context);
	    CONSTANT_POOL_TYPE_TABLE_SET_RESOLVED(type_table, i);
	    break;

	  case CONSTANT_Long:
	  case CONSTANT_Double: {
	      unsigned high = get4bytes(context);
	      unsigned low = get4bytes(context);
	      int64_t value;
	      value = ll_or(ll_shl(uint2ll(high), 32), uint2ll(low));
	      if (type == CONSTANT_Double) {
		  ll2double_bits(&value);
	      }
	      SET_INT64(t1, &constant_pool[i], value);
	      CONSTANT_POOL_TYPE_TABLE_SET_RESOLVED(type_table, i);
	      i++;		/* increment i for loop, too */
	      /* Indicate that the next object in the constant pool cannot
	       * be accessed independently.    */
	      CONSTANT_POOL_TYPE_TABLE_PUT(type_table, i, 0);
	      CONSTANT_POOL_TYPE_TABLE_SET_RESOLVED(type_table, i);
	      break;
	  }

	  default:
	      sysAssert(FALSE);
	}
    }

    constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type = type_table;

    /*
     * It is important to only set these after everything is setup,
     *  so that the GC sees a consistent state.
     */
    cbConstantPool(cb) = constant_pool;
    cbConstantPoolCount(cb) = nconstants;
}



static void ReadInCode(CICcontext *context, struct methodblock *mb)
{
    unsigned long attribute_length = get4bytes(context);
#ifdef DEBUG
    unsigned char *end_ptr = context->ptr + attribute_length;
#endif
    int attribute_count;
    uint32_t code_length;
    int i;

    if (cbMajorVersion(context->cb)==45 && cbMinorVersion(context->cb) <= 2) { 
	mb->maxstack = get1byte(context);
	mb->nlocals = get1byte(context);
	code_length = get2bytes(context);
    } else { 
	mb->maxstack = get2bytes(context);
	mb->nlocals = get2bytes(context);
	code_length = get4bytes(context);
    }
        
    mb->code_length = (unsigned short)code_length;

    if (context->in_clinit) {
        mb->code = context->clinit_space + context->offset.clinit.code;
	context->offset.clinit.code += ALIGN_UP(code_length, (sizeof(int)));
	sysAssert(context->offset.clinit.code <= context->offset.clinit.excs);
    } else {
        mb->code = context->main_space + context->offset.main.code;
	context->offset.main.code += ALIGN_UP(code_length,(sizeof(int)));
	sysAssert(context->offset.main.code <= context->offset.main.excs);
    }

    getNbytes(context, code_length, (char *)mb->code);
    if ((mb->exception_table_length = get2bytes(context)) > 0) {
	unsigned exception_table_size = mb->exception_table_length 
	                                  * sizeof(struct CatchFrame);
	if (context->in_clinit) {
	    mb->exception_table = (struct CatchFrame *)
	        (context->clinit_space + context->offset.clinit.etab);
	    context->offset.clinit.etab += exception_table_size;
	    sysAssert(context->offset.clinit.etab <= 
		      context->offset.clinit.lnum);
	} else {
	    mb->exception_table = (struct CatchFrame *)
	        (context->main_space + context->offset.main.etab);
	    context->offset.main.etab += exception_table_size;
	    sysAssert(context->offset.main.etab <= context->offset.main.lnum);
	}

	for (i = 0; i < mb->exception_table_length; i++) {
	    mb->exception_table[i].start_pc = get2bytes(context);
	    mb->exception_table[i].end_pc = get2bytes(context);		
	    mb->exception_table[i].handler_pc = get2bytes(context); 
	    mb->exception_table[i].catchType = get2bytes(context);
            mb->exception_table[i].compiled_CatchFrame = NULL;
	}
    }

    attribute_count = get2bytes(context);
    {
        uint32_t line_number_table_counter = 0;
	uint32_t localvar_table_counter = 0;
	if (context->in_clinit) {
	    mb->line_number_table = (struct lineno *)
	        (context->clinit_space + context->offset.clinit.lnum);
	    mb->localvar_table = (struct localvar *)
	        (context->clinit_space + context->offset.clinit.lvar);
	} else {
	    mb->line_number_table = (struct lineno *)
	        (context->main_space + context->offset.main.lnum);
	    mb->localvar_table = (struct localvar *)
	        (context->main_space + context->offset.main.lvar);
	}
	for (i = 0; i < attribute_count; i++) {
	    char *name = getAsciz(context);
	    if (name == utf8_literal_LineNumberTable) {
	        line_number_table_counter = 
		    ReadLineTable(context, mb, line_number_table_counter);
#ifdef JCOV
	    } else if (name == utf8_literal_CoverageTable && cov_file
		     && (jcov_flags & JCOV_MODE_DEFAULT)) {
	        ReadCovTable(context, mb); 
#endif
	    } else if (name == utf8_literal_LocalVariableTable) {
	        localvar_table_counter = 
		    ReadLocalVars(context, mb, localvar_table_counter);
	    } else {
	        unsigned long length = get4bytes(context);
		getNbytes(context, length, NULL);
	    }
	}
#ifdef JCOV
	if (cov_file && (jcov_flags & JCOV_MODE_METHODS))
	    CreateCovTable(context, mb); 
#endif
	mb->line_number_table_length = line_number_table_counter;
	mb->localvar_table_length = localvar_table_counter;
	/* The VM Spec doesn't seem to require line number tables are 
	 * ordered. The pc2lineno function in classruntime.c relies
	 * on this fact.
	 */
	SortLineTable(mb);
	if (context->in_clinit) {
	    context->offset.clinit.lnum += 
	        line_number_table_counter * sizeof(struct lineno);
	    context->offset.clinit.lvar +=
	        localvar_table_counter * sizeof(struct localvar);
	    sysAssert(context->offset.clinit.lnum <= 
		      context->offset.clinit.lvar);
	} else {
	    context->offset.main.lnum +=
	        line_number_table_counter * sizeof(struct lineno);
	    context->offset.main.lvar +=
	        localvar_table_counter * sizeof(struct localvar);
	    sysAssert(context->offset.main.lnum <= context->offset.main.lvar);
	}
    }

    sysAssert(context->ptr == end_ptr);
}

static int 
ln_compare(const void *p1, const void *p2)
{
    struct lineno *ln1 = (struct lineno *)p1;
    struct lineno *ln2 = (struct lineno *)p2;
    return ln1->pc - ln2->pc;
}

static void
SortLineTable(struct methodblock *mb)
{
    int length = mb->line_number_table_length;
    struct lineno *ln = mb->line_number_table;
    int i;
    int pc;

    if (length == 0)
        return;
    pc = 0;
    for (i = 0; i < length; i++) {
        int pc0 = ln[i].pc;
	/* Only call qsort unless we need to. */
        if (pc0 < pc) {
	    qsort(ln, length, sizeof(struct lineno), ln_compare);
	    return;
	}
	pc = pc0;
    }
}

static uint32_t 
ReadLineTable(CICcontext *context, struct methodblock *mb, uint32_t counter)
{
    unsigned long attribute_length = get4bytes(context);
#ifdef DEBUG
    unsigned char *end_ptr = context->ptr  + attribute_length;
#endif
    int i, table_length;
    if ((table_length = get2bytes(context)) > 0) {
	struct lineno *ln = mb->line_number_table + counter;
	for (i = table_length; --i >= 0; ln++) {
	    ln->pc = get2bytes(context);
	    ln->line_number = get2bytes(context);
	}
    }
    sysAssert(context->ptr == end_ptr);
    return counter + table_length;
}	
	
#ifdef JCOV
static void ReadCovTable(CICcontext *context, struct methodblock *mb)
{
    unsigned int attribute_length = get4bytes(context);
    unsigned char *end_ptr = context->ptr  + attribute_length;
    int bad_jcovTab = 0;
    int i;

    if ((mb->coverage_table_length = get2bytes(context)) > 0) {
	if ((2 + (mb->coverage_table_length * JCOVTAB_ELEMENT_LENGTH)) != attribute_length)
	{
	    getNbytes(context, (attribute_length - 2), NULL);
	    bad_jcovTab = 1;
	} else {
	    int sz = ALIGN_NATURAL
                       (mb->coverage_table_length * sizeof(struct covtable))
                         + ALIGN_NATURAL((mb->code_length + 1) * sizeof(int));
	    struct covtable *cv = sysCalloc(sz, 1);
    	    int *jcov_mode;
            int *hash;

	    mb->coverage_table = cv;
            jcov_mode = (int*)(mb->coverage_table + mb->coverage_table_length);
	    hash = jcov_mode + 1;
	    memset(jcov_mode, 0, (mb->code_length + 1) * sizeof(int));
    	    *jcov_mode |= JCOV_MODE_DEFAULT;		/* type of gathering */
	    for (i = 0; i < mb->coverage_table_length; cv++) {
	    	cv->pc = get2bytes(context);
	    	if (cv->pc > mb->code_length)
		    bad_jcovTab = 1;
	    	else
            	    *(hash + cv->pc) = i + 1;
	    	cv->type = get2bytes(context);
	    	cv->where_line = get4bytes(context);
	    	cv->where_pos  = get4bytes(context);
	    	cv->count = 0;
            	i++;
	    }
	}
    }
    if ((context->ptr != end_ptr) || (bad_jcovTab))
	mb->coverage_table = NULL;	/* Don't use a wrong coverage table */
}

static void CreateCovTable(CICcontext *context, struct methodblock *mb)
{
    int sz = ALIGN_NATURAL(sizeof(struct covtable) + sizeof(int));

    struct covtable *cv;
    int *jcov_mode;

    if (mb->coverage_table != NULL)
	return;
						/* Jcov table was loaded */
    cv = sysCalloc(sz, 1);
    mb->coverage_table = cv;
    mb->coverage_table_length = 1;		/* only one element - method itself */
    jcov_mode = (int *)(mb->coverage_table + mb->coverage_table_length);
    memset(jcov_mode, 0, sizeof(int));
    *jcov_mode |= JCOV_MODE_METHODS;		/* type of gathering */
    cv->pc = 0;
    cv->type = CT_METHOD;
    if (mb->line_number_table != NULL)		/* on the case of native methods */
        cv->where_line = mb->line_number_table->line_number;
    else
        cv->where_line = 0;
    cv->where_pos  = 0;
    cv->count = 0;
}	

static void 
AddCovTableForNative(CICcontext *context, ClassClass *cls)
{
    struct methodblock *mb;
    int j;

    if (((jcov_flags & JCOV_MODE_DEFAULT) && (cbTimestamp(cls) != 0) )	/* Compiled with jcov flag */
      || (jcov_flags & JCOV_MODE_METHODS) ) {
	for (mb = cbMethods(cls), j = cbMethodsCount(cls); --j >= 0; mb++) {
	    if (mb->fb.access & ACC_NATIVE) {
		CreateCovTable(context, mb);
	    }
	}
    }
}

#endif 

static uint32_t 
ReadLocalVars(CICcontext *context, struct methodblock *mb, uint32_t counter)
{
    unsigned long attribute_length = get4bytes(context);
#ifdef DEBUG
    unsigned char *end_ptr = context->ptr  + attribute_length;
#endif
    int i, table_length;
    if ((table_length = get2bytes(context)) > 0) {
	struct localvar *lv = mb->localvar_table + counter;
	for (i = table_length; --i >= 0; lv++) {
	    lv->pc0 = get2bytes(context);
	    lv->length = get2bytes(context);
	    lv->nameoff = get2bytes(context);
	    lv->sigoff = get2bytes(context);
	    lv->slot = get2bytes(context);
	}
    }
    sysAssert(context->ptr == end_ptr);
    return counter + table_length;
}	

static void
ReadExceptions(CICcontext *context, struct methodblock *mb)
{
    unsigned long attribute_length = get4bytes(context);
#ifdef DEBUG
    unsigned char *end_ptr = context->ptr + attribute_length;
#endif
    unsigned short nexceptions = get2bytes(context);

    if ((mb->nexceptions = nexceptions) > 0) {
	unsigned short *exceptions;
	unsigned int size = nexceptions * sizeof (unsigned short);
	if (context->in_clinit) {
	    exceptions = (unsigned short *)
	        (context->clinit_space + context->offset.clinit.excs);
	    context->offset.clinit.excs += size;
	    sysAssert(context->offset.clinit.excs <= 
		      context->offset.clinit.etab);
	} else {
	    exceptions = (unsigned short *)
	        (context->main_space + context->offset.main.excs);
	    context->offset.main.excs += size;
	    sysAssert(context->offset.main.excs <= context->offset.main.etab);
	}
	mb->exceptions = exceptions;
	while (nexceptions-- > 0) {
	    *exceptions++ = get2bytes(context);
	}
    }
    sysAssert(context->ptr == end_ptr);
}

unsigned Signature2ArgsSize(char *method_signature)
{
    char *p;
    unsigned int args_size = 0;
    for (p = method_signature; *p != SIGNATURE_ENDFUNC; p++) {
	switch (*p) {
	  case SIGNATURE_BOOLEAN:
	  case SIGNATURE_BYTE:
	  case SIGNATURE_CHAR:
	  case SIGNATURE_SHORT:
	  case SIGNATURE_INT:
	  case SIGNATURE_FLOAT:
	    args_size += 1;
	    break;
	  case SIGNATURE_CLASS:
	    args_size += 1;
	    while (*p != SIGNATURE_ENDCLASS) {
		p++;
	    }
	    break;
	  case SIGNATURE_ARRAY:
	    args_size += 1;
	    while ((*p == SIGNATURE_ARRAY)) p++;
	    /* If an array of classes, skip over class name, too. */
	    if (*p == SIGNATURE_CLASS) { 
	        while (*p != SIGNATURE_ENDCLASS) {
		    p++;
		}
	    } 
	    break;
	  case SIGNATURE_DOUBLE:
	  case SIGNATURE_LONG:
	    args_size += 2;
	    break;
	  case SIGNATURE_FUNC:	/* ignore initial (, if given */
	    break;
	  default:
	    /* impossible */
	    sysAssert(FALSE);
	}
    }
    return args_size;
}

void FreeClass(ClassClass *cb) 
{
    int i;
    ExecEnv * ee = EE();
    CompilerFreeClass(cb);	/* indicate that class is free */

    DeleteClassFromLoaderConstraints(cb);

    for (i=0; i < cbMethodsCount(cb); i++) {
        struct methodblock *mb = cbMethods(cb) + i;
	if (mb->terse_signature)
	    ReleaseUTF8(ee, mb->terse_signature);
#ifdef JCOV
	sysFree(mb->coverage_table);
#endif
    }

    {
        struct methodblock *mb = cbMirandaMethods(cb);
	for (i = cbMirandaMethodsCount(cb) - 1; i >= 0; i--, mb++) {
	    ReleaseUTF8(ee, mb->fb.name);
	}
    }

    freeUTF8EntriesFromClassClass(cb);

    sysFree(cbClinitSpace(cb));
    sysFree(cbMainSpace(cb));

    sysFree(cbObjectOffsets(cb));

    sysFree(cbMirandaMethods(cb));

    sysFree(cbMethodTableMem(cb));

    /* Interface method tables can be shared between child and super classes
     */
    if (cbImplementsCount(cb) != 0 || cbIsInterface(cb))
        sysFree(cbIntfMethodTable(cb));
    memset(unhand(cb), 0, sizeof(struct Classjava_lang_Class));
}
	
static unsigned long get1byte(CICcontext *context)
{
    unsigned char *ptr = context->ptr;
    sysAssert(context->end_ptr - ptr >= 1);
    {
	unsigned long value = ptr[0];
	(context->ptr) += 1;
	return value;
    }
}

static unsigned long get2bytes(CICcontext *context)
{
    unsigned char *ptr = context->ptr;
    sysAssert(context->end_ptr - ptr >= 2);
    {
	unsigned long value = (ptr[0] << 8) + ptr[1];
	(context->ptr) += 2;
	return value;
    }
}

static unsigned long get4bytes(CICcontext *context)
{
    unsigned char *ptr = context->ptr;
    sysAssert(context->end_ptr - ptr >= 4);
    {
	unsigned long value = (ptr[0] << 24) + (ptr[1] << 16) + 
	                             (ptr[2] << 8) + ptr[3];
	(context->ptr) += 4;
	return value;
    }
}

static void getNbytes(CICcontext *context, size_t count, char *buffer) 
{
    unsigned char *ptr = context->ptr;
    sysAssert(context->end_ptr - ptr >= count);
    if (buffer != NULL)
	memcpy(buffer, ptr, count);
    (context->ptr) += count;
}


static char *getAsciz(CICcontext *context) 
{
    ClassClass *cb = context->cb;
    union cp_item_type *constant_pool = cbConstantPool(cb);

    int value = get2bytes(context);
    return constant_pool[value].cp;
}

static char* getUTF8String(ExecEnv * ee, CICcontext* context)
{
    char buf[STK_BUF_LEN];
    char *chars;
    char *result;
    int length = get2bytes(context);
    if (length >= sizeof(buf)) {
        chars = sysMalloc(length + sizeof(char));
	if (chars == 0) {
	    CLnomem(context);
	}
    } else {
        chars = buf;
    }
    getNbytes(context, length, chars);
    chars[length] = '\0';
    result = AddUTF8(ee, chars);
    if (chars != buf) {
        sysFree(chars);
    }
    if (result == 0) {
        CLnomem(context);
    }
    return result;
}

static void freeBuffers(CICcontext * context)
{
    sysFree(context->clinit_space);
    sysFree(context->main_space);
}

static void freeUTF8EntriesFromContext(CICcontext* context)
{
    cp_item_type* constant_pool = context->constant_pool;
    unsigned char* type_table = context->type_table;
    int nconstants = context->size.constants;
    /* free the UTF8 entries. */
    freeUTF8Entries(constant_pool, type_table, nconstants);
    /* Null out the fields of the context. */
    context->constant_pool = NULL;
    context->type_table = NULL;
}

static void freeUTF8EntriesFromClassClass(ClassClass* cb)
{
    cp_item_type* constant_pool = cbConstantPool(cb);
    if (constant_pool != NULL) {
	unsigned char* type_table =
	    constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
	int nconstants = cbConstantPoolCount(cb);
	/* free the UTF8 entries. */
	freeUTF8Entries(constant_pool, type_table, nconstants);
    }
}
		    
static void freeUTF8Entries(cp_item_type* constant_pool,
			    unsigned char* type_table,
			    int nconstants)
{
    int i;
    if ((constant_pool != NULL) &&
	(type_table != NULL) &&
	(nconstants != 0)) {
	ExecEnv * ee = EE();
	for (i = CONSTANT_POOL_UNUSED_INDEX; i < nconstants; i++) {
	    if ((CONSTANT_POOL_TYPE_TABLE_GET_TYPE(type_table, i) == 
		 CONSTANT_Utf8) &&
		(CONSTANT_POOL_TYPE_TABLE_IS_RESOLVED(type_table, i))) {
		char* data = (char*) constant_pool[i].cp;
		if (data != NULL) {
		    ReleaseUTF8(ee, data);
		}
	    }
	}
    }
}

static void
CLnomem(CICcontext *context)
{
    ThrowOutOfMemoryError(0, "loading classes");
    longjmp(context->jump_buffer, 1);
}

