/*
 * @(#)classresolver.c	1.169 98/11/03
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*-
 *      Routines for loading and resolving class definitions.
 */

#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>

#include "util.h"
#include "oobj.h"
#include "util.h"
#include "interpreter.h"
#include "signature.h"
#include "tree.h"
#include "monitor.h"
#include "opcodes.h"
#include "sys_api.h"
#include "common_exceptions.h"

#ifdef BREAKPTS
#include "breakpoints.h"
#endif

#ifdef HPROF
#include "vmprofiler.h"
#endif /* HPROF */

ClassClass *classJavaLangObject;
ClassClass *classJavaLangClass = 0;
ClassClass *classJavaLangString;
ClassClass *classJavaLangThread;
ClassClass *classJavaLangClassLoader;
ClassClass *classJavaLangThrowable;
ClassClass *classJavaLangError;
ClassClass *classJavaLangException;
ClassClass *classJavaLangRuntimeException;
ClassClass *classJavaLangThreadDeath;

ClassClass *interfaceJavaLangCloneable;
ClassClass *interfaceJavaIoSerializable;

static void PrepareClass(ClassClass * cb);
static ClassClass *
FindArrayClassFromClassLoader(struct execenv *ee, char *name,
			      struct Hjava_lang_ClassLoader *loader,
			      bool_t throwError);

static bool_t InitPrimitiveClasses(void);

/* An Explanation of Class-related Locks
 *
 * There are three global locks related to class loading:
 *
 * BINCLASS_LOCK: ensures that only one thread is updating the global
 * class table (binclasses). This lock is also grabbed by the GC to ensure
 * that the GC always sees a valid global class table state.
 *
 * LINKCLASS_LOCK: ensures that only one thread updates per-class data
 * when resolving symbolic references. This lock is usually held only
 * for a short period of time just to protect updating the class
 * structures.
 *
 * SYSLOADER_LOCK: ensures that only one thread is loading a system
 * class (with null class loader) or an array class at a given time.
 */

sys_mon_t *_linkclass_lock;
sys_mon_t *_binclass_lock;
sys_mon_t *_sysloader_lock;

/*
 * For a complete description of class loader constraints and loaded
 * class cache, see the following paper:
 *
 *     "Dynamic Class Loading in the Java Virtual Machine"
 *
 * by Sheng Liang and Gilad Bracha. The paper will be published in the
 * proceedings of the OOPSLA'98 conference.
 *
 * The following structure stores a class name and a set of loaders.
 * The loaders are required to map the class name to the same class.
 *
 * The cb field exists to speed up checking. It is initially set to 
 * zero, and later set to the first class loaded by one of the 
 * loaders.
 *
 * Loader constraints form a "(class name, loader set) => class type"
 * mapping.
 */

typedef struct loader_constraint_t {
    struct loader_constraint_t *next;
    ClassClass *cb;				/* initially NULL */
    char *name;
    int num_loaders;
    int max_loaders;				/* capacity */
    struct Hjava_lang_ClassLoader *loaders[2];	/* initiating loaders */
} loader_constraint_t;

#define LOADER_CONSTRAINT_TABLE_SIZE 107
static loader_constraint_t *
loader_constraint_table[LOADER_CONSTRAINT_TABLE_SIZE];

/* Loader cache keeps a "(name, loader) => class" mapping. All the
 * classes in the cache have been verified to meet the class loader
 * constraints. The cache also speeds up class loading.
 *
 * Note that in the "(name, loader) => class" maping, "loader" is the
 * initiating loader, not necessary the loader that ultimately defines 
 * the class. A class loader may delegate class loading to another 
 * class loader. For example, when "java/lang/String" is loaded with 
 * an applet class loader, the resulting class in fact has NULL as its 
 * class loader.
 *
 * When a new class is about to be added to the cache, we make sure 
 * it doesn't break any of the existing constraints.
 *
 * During the course of execution, new constraints may be added as
 * the result of class linking. Before each new constraint is added,
 * we make sure classes loaded in the cache will satisfy the new 
 * constraint.
 */

typedef struct loader_cache_t {
    struct loader_cache_t *next;
    ClassClass *cb;
    struct Hjava_lang_ClassLoader *loader; /* initiating loader */
} loader_cache_t;

#define LOADER_CACHE_TABLE_SIZE 1009
static loader_cache_t *loader_cache[LOADER_CACHE_TABLE_SIZE] = {0};
#define HASH_INDEX(name, loader) \
    (((unsigned)name + (unsigned)loader) % LOADER_CACHE_TABLE_SIZE)

/* Add a new entry to the loader cache.
 */
static loader_cache_t *
AddLoaderCacheEntry(ClassClass *cb, struct Hjava_lang_ClassLoader *loader)
{
    char *cb_name = cbName(cb);
    loader_cache_t *entry = sysMalloc(sizeof(loader_cache_t));
    int index = HASH_INDEX(cb_name, loader);
    if (entry == NULL) {
        return NULL;
    } 

    sysAssert(BINCLASS_LOCKED(sysThreadSelf()));

#ifdef CL_DEBUG
    jio_fprintf(stderr,
		"added to loader cache: (%s, %x) => %x\n",
		cbName(cb), loader, cb);
#endif

    entry->next = loader_cache[index];
    entry->cb = cb;
    entry->loader = loader;
    loader_cache[index] = entry;

    return entry;
}

static ClassClass *
LookupLoaderCache(char *hashed_name, struct Hjava_lang_ClassLoader *loader)
{
    int index = HASH_INDEX(hashed_name, loader);
    loader_cache_t *entry = loader_cache[index];

    sysAssert(BINCLASS_LOCKED(sysThreadSelf()));
    while (entry) {
        if (cbName(entry->cb) == hashed_name && entry->loader == loader) {
	    return entry->cb;
	}
	entry = entry->next;
    }
    return NULL;
}

JNIEXPORT jclass JNICALL
JVM_FindLoadedClass(JNIEnv *env, jobject this, jstring name)
{
    struct Hjava_lang_ClassLoader *loader =
        (struct Hjava_lang_ClassLoader *)DeRef(env, this);
    ClassClass *cb;
    char slash_name_buf[STK_BUF_LEN];
    char *slash_name;
    int uni_len = (*env)->GetStringLength(env, name);
    int utf_len = (*env)->GetStringUTFLength(env, name);
    ExecEnv * ee = JNIEnv2EE(env);

    if (utf_len >= sizeof(slash_name_buf)) {
        slash_name = sysMalloc(utf_len + 1);
	if (slash_name == NULL) {
	    ThrowOutOfMemoryError(0, 0);
	    return NULL;
	}
    } else {
        slash_name = slash_name_buf;
    }
    
    (*env)->GetStringUTFRegion(env, name, 0, uni_len, slash_name);
    for (utf_len = 0; slash_name[utf_len]; utf_len++) {
        if (slash_name[utf_len] == '.') {
	    slash_name[utf_len] = '/';
	}
    }

    BINCLASS_LOCK(EE2SysThread(ee));
    {	
        char *hashed_name = HashUTF8(ee, slash_name);
	cb = hashed_name ? LookupLoaderCache(hashed_name, loader) : NULL;
    }
    BINCLASS_UNLOCK(EE2SysThread(ee));
    if (cb) {
        if (!CCIs(cb, Loaded)) {
	    LoadSuperclasses(cb);
	    if (!CCIs(cb, Loaded)) {
	        cb = 0;
	    }
	}
    }
    if (slash_name != slash_name_buf) {
        sysFree(slash_name);
    }
    return MkRefLocal(env, cb);
}

/* TODO: Change this to binary search for better performance. */
static loader_constraint_t **
FindLoaderConstraint(char *name, struct Hjava_lang_ClassLoader *loader)
{
    int index = (unsigned long)name % LOADER_CONSTRAINT_TABLE_SIZE;
    loader_constraint_t **pp = &loader_constraint_table[index];
    while (*pp) {
        loader_constraint_t *p = *pp;
	if (p->name == name) {
	    int i;
	    for (i = p->num_loaders - 1; i >= 0; i--) {
	        if (p->loaders[i] == loader) {
		    return pp;
		}
	    }
	}
	pp = &(p->next);
    }
    return pp;
}

static bool_t
EnsureLoaderConstraintCapacity(loader_constraint_t **pp, int nfree)
{
    loader_constraint_t *p = *pp;
    if (p->max_loaders - p->num_loaders < nfree) {
        int nh = offsetof(loader_constraint_t, loaders);
        int n = nfree + p->num_loaders;
	if (n < 2 * p->max_loaders) {
	    n = 2 * p->max_loaders;
	}
        p = sysMalloc(nh + sizeof(void *) * n);
	if (p == NULL) {
	    return FALSE;
	}
	memcpy(p, *pp, nh + sizeof(void *) * (*pp)->num_loaders);
	p->max_loaders = n;
	sysFree(*pp);
	*pp = p;
    }
    return TRUE;
}
 
static bool_t
ExtendLoaderConstraint(loader_constraint_t **pp, 
		      struct Hjava_lang_ClassLoader *loader)
{
    if (!EnsureLoaderConstraintCapacity(pp, 1)) {
	BINCLASS_UNLOCK(sysThreadSelf());
	ThrowOutOfMemoryError(0, 0);
        return FALSE;
    }
    (*pp)->loaders[(*pp)->num_loaders++] = loader;
    BINCLASS_UNLOCK(sysThreadSelf());
    return TRUE;
}

static bool_t
MergeLoaderConstraints(loader_constraint_t **pp1, loader_constraint_t **pp2)
{
    loader_constraint_t *p1, *p2;
    int i;

    /* make sure *pp1 has higher capacity */
    if ((*pp1)->max_loaders < (*pp2)->max_loaders) {
        loader_constraint_t **tmp = pp2;
	pp2 = pp1;
	pp1 = tmp;
    }

    p1 = *pp1;
    p2 = *pp2;
    
    if (!EnsureLoaderConstraintCapacity(pp1, p2->num_loaders)) {
	BINCLASS_UNLOCK(sysThreadSelf());
	ThrowOutOfMemoryError(0, 0);
        return FALSE;
    }
    for (i = 0; i < p2->num_loaders; i++) {
        p1->loaders[p1->num_loaders++] = p2->loaders[i];
    }

    /* release pp2 */
    *pp2 = p2->next;
    ReleaseUTF8(EE(), p2->name);
    sysFree(p2);

    BINCLASS_UNLOCK(sysThreadSelf());
    return TRUE;
}

/* Class cb has been initiated in loader. This function ensures that cb
 * meets the constraints stored in the hash table.
 */
char *
UpdateLoaderCache(ClassClass *cb,
		  struct Hjava_lang_ClassLoader *loader,
		  char *details,
		  int len)
{
    loader_constraint_t *p;

    sysAssert(BINCLASS_LOCKED(sysThreadSelf()));

    {
        ClassClass *cb0 = LookupLoaderCache(cbName(cb), loader);
	if (cb0) {
	    /* Not allowed to update an existing entry */
	    if (cb0 != cb) {
		jio_snprintf(details, len,
			     "trying to refine class %s (bad class loader?)",
			     cbName(cb));
		return "java/lang/LinkageError";
	    } else {
		return NULL;
	    }
	}
    }

    p = *(FindLoaderConstraint(cbName(cb), loader));

    if (p && p->cb != NULL && p->cb != cb) {
        jio_snprintf(details, len, "Class %s violates loader constraints",
		     cbName(cb));
	return "java/lang/LinkageError";
    }

    /* We can now add (cb, loader) into loader cache */
    if (!AddLoaderCacheEntry(cb, loader)) {
	if (len > 0)
	    details[0] = '\0';
	return "java/lang/OutOfMemoryError";
    }

    if (p && p->cb == NULL) {
        p->cb = cb;
    }
    return NULL;
}

/* Impose a new loader constraint: "name" must be loaded the same
 * way in loader1 and loader2.
 *
 * Note that the code is written so that even if there is any kind of
 * error, existing data structures will not be corruptted.
 */
static bool_t 
AddLoaderConstraint(char *name,
		    struct Hjava_lang_ClassLoader *loader1,
		    struct Hjava_lang_ClassLoader *loader2)
{
    sys_thread_t *self = sysThreadSelf();
    ExecEnv * ee = SysThread2EE(self);
    char *hashed_name = AddUTF8(ee, name);
    loader_constraint_t **pp1, **pp2;
    ClassClass *cb, *cb1, *cb2;

    if (hashed_name == 0) {
	ThrowOutOfMemoryError(0, 0);
	return FALSE;
    }

    BINCLASS_LOCK(self);

    cb1 = LookupLoaderCache(hashed_name, loader1);
    cb2 = LookupLoaderCache(hashed_name, loader2);
    
    if (cb1 != 0 && cb2 != 0 && cb1 != cb2) {
        char buf[MSG_BUF_LEN];
	BINCLASS_UNLOCK(self);
	jio_snprintf(buf, sizeof(buf), "Class %s violates loader constraints",
		     name);
	SignalError(0, "java/lang/LinkageError", buf);
        ReleaseUTF8(ee, hashed_name);
	return FALSE;
    } else {
        cb = cb1 ? cb1 : cb2;
    }

    pp1 = FindLoaderConstraint(hashed_name, loader1);
    pp2 = FindLoaderConstraint(hashed_name, loader2);

#ifdef CL_DEBUG
    if (*pp1 && *pp1 == *pp2) {
        jio_fprintf(stderr,
		    "creating new constraint: (%s, %x, %x) => %x\n",
		    hashed_name, loader1, loader2, cb);
    }
#endif

    if (*pp1 == NULL && *pp2 == NULL) {
        int index;
        loader_constraint_t *p = sysMalloc(sizeof(loader_constraint_t));
	if (p == NULL) {
	    BINCLASS_UNLOCK(self);
	    ThrowOutOfMemoryError(0, 0);
	    ReleaseUTF8(ee, hashed_name);
	    return FALSE;
	}

	p->name = hashed_name;
	p->loaders[0] = loader1;
	p->loaders[1] = loader2;
	p->num_loaders = 2;
	p->max_loaders = 2;
	p->cb = cb;

        index = (unsigned long)hashed_name % LOADER_CONSTRAINT_TABLE_SIZE;
	p->next = loader_constraint_table[index];
	loader_constraint_table[index] = p;
	BINCLASS_UNLOCK(self);
	return TRUE;
    } else if (*pp1 == *pp2) {
        /* constraint already imposed */
        sysAssert((*pp1)->cb == cb);
        BINCLASS_UNLOCK(self);
	ReleaseUTF8(ee, hashed_name);
	return TRUE;
    } else if (*pp1 == NULL) {
        ReleaseUTF8(ee, hashed_name);
        sysAssert((*pp2)->cb == cb);
        return ExtendLoaderConstraint(pp2, loader1);
    } else if (*pp2 == NULL) {
        ReleaseUTF8(ee, hashed_name);
        sysAssert((*pp1)->cb == cb);
        return ExtendLoaderConstraint(pp1, loader2);
    } else {
        ReleaseUTF8(ee, hashed_name);
        sysAssert((*pp1)->cb == cb);
        sysAssert((*pp2)->cb == cb);
	return MergeLoaderConstraints(pp1, pp2);
    }
}

static int 
ClassComponentLength(char *signature)
{
    int len = 0;
    if (*signature == SIGNATURE_ARRAY) {
        while (*signature == SIGNATURE_ARRAY) {
	    signature++;
	    len++;
	}
	if (*signature == SIGNATURE_CLASS) {
	    while (*signature != SIGNATURE_ENDCLASS) {
	        signature++;
		len++;
	    }
	}
    } else {
        while (*signature != SIGNATURE_ENDCLASS) {
	    signature++;
	    len++;
	}
    }
    return len + 1;
}

/* Make sure all class components (including arrays) in the given
 * signature will be resolved to the same class in both loaders.
 */
bool_t 
CheckSignatureLoaders(char *signature, 
		      struct Hjava_lang_ClassLoader *loader1,
		      struct Hjava_lang_ClassLoader *loader2)
{
    char classname[STK_BUF_LEN];
    char *p;
    char c;

    /* Nothing to do if loaders are the same. */
    if (loader1 == loader2)
       return TRUE;

    while (*signature) {
	switch (c = *signature) {
	case SIGNATURE_ARRAY:
	case SIGNATURE_CLASS: {
	    bool_t result;
	    int len = ClassComponentLength(signature);
	    if (len + 1 >= sizeof(classname)) {
		p = sysMalloc(len + 1);
		if (p == 0) {
		    ThrowOutOfMemoryError(0, 0);
		    return FALSE;
		}	
	    } else {
		p = classname;
	    }
	    if (c == SIGNATURE_ARRAY) {
	        strncpy(p, signature, len);
		p[len] = 0;
	    } else {
	        strncpy(p, signature + 1, len - 2);
		p[len - 2] = 0;
	    }
	    signature += len;
#ifdef CL_DEBUG
	    if (getenv("_FORCE_LOAD")) {
	        if (!FindClassFromClassLoader(0, p, FALSE, loader2, TRUE))
		    return FALSE;
	    }
#endif
	    result = AddLoaderConstraint(p, loader1, loader2);
	    
	    if (p != classname)
		sysFree(p);
	    if (!result)
		return FALSE;
	    break;
	}
	default:
	    signature++;
	}
    }
    return TRUE;
}

/* When a class is GC'ed, we need to delete the loader constrants and
 * loader cache entries that include the class or its loader.
 */
void 
DeleteClassFromLoaderConstraints(ClassClass *cb)
{
    struct Hjava_lang_ClassLoader *loader = cbLoader(cb);
    sys_thread_t *self = sysThreadSelf();
    int i;

    /* If the "Cached" flag is not set, the VM has not yet established
     * a link from the loader to the ClassClass object. In that case,
     * freeing the class does not necessarily mean the loader has been 
     * freed as well.
     */
    if (!CCIs(cb, Cached)) {
        loader = NULL;
    }

    BINCLASS_LOCK(self);
    for (i = 0; i < LOADER_CONSTRAINT_TABLE_SIZE; i++) {
        loader_constraint_t **cp = &(loader_constraint_table[i]);
	while (*cp) {
	    loader_constraint_t *p = *cp;
	    int j;
	    for (j = 0; j < p->num_loaders; j++) {
	        if (loader && (p->loaders[j] == loader)) {
#ifdef CL_DEBUG
		    jio_fprintf(stderr,
				"deleting loader constraint: (%s, %x)\n",
				p->name, loader);
#endif
		    p->num_loaders--;
		    while (j < p->num_loaders) {
		        p->loaders[j] = p->loaders[j + 1];
			j++;
		    }
		    break;
		}
	    }
	    if (p->num_loaders == 1) {
	        *cp = p->next;
		ReleaseUTF8(EE(), p->name);
		sysFree(p);
		continue;
	    }
	    
	    if (cb == p->cb) {
	        /* This can happen if the initiating loader outlives the class
		 */
	        p->cb = 0;
	    }
	    cp = &(p->next);
	}
    }
    for (i = 0; i < LOADER_CACHE_TABLE_SIZE; i++) {
        loader_cache_t **entryP = &(loader_cache[i]);
	while (*entryP) {
	    loader_cache_t *entry = *entryP;
	    if ((entry->cb == cb) || (loader && (entry->loader == loader))) {
	        *entryP = entry->next;
		sysFree(entry);
	    } else {
	        entryP = &(entry->next);
	    }
	}
    }
    BINCLASS_UNLOCK(self);
}

/* Make sure to clean up all allocated memory if anything goes wrong
 * in the middle.
 */
static void
PrepareFields(ClassClass *cb, unsigned slot)
{
    struct fieldblock *fb;
    int size;

    if (CCIs(cb, FieldPrepared))
        return;

    fb = cbFields(cb);
    for (size = 0; size < (int) cbFieldsCount(cb); size++, fb++) {
	char *signature = fieldsig(fb);
	int size = (((signature[0] == SIGNATURE_LONG || 
		      signature[0] == SIGNATURE_DOUBLE)) ? 2 : 1);
	if (fb->access & ACC_STATIC) {
	    /*EMPTY*/
	    /* initialized in classload.c */
	} else {
	    fb->u.offset = slot;
	    slot += size * sizeof(OBJECT);
	}
    }

    if (slot > 65535) {
        ThrowInternalError(0, "exceeding the 64K byte object size limit");
	return;
    }

    cbInstanceSize(cb) = (unsigned short)slot;
    /* To speed up the GC, we compute the offsets of all the objects
     * in the class. Since 0 is a legal offset, we add 1 to the offsets,
     * and terminate the array with 0.
     */
    {
        ClassClass *cb1 = cb;
        int n_object_slots = 0;
	unsigned short *object_offsets;
	while(cb1) {
	    fb = cbFields(cb1);
	    for (size = 0; size < cbFieldsCount(cb1); size++, fb++) {
	        if ((fieldIsArray(fb) || fieldIsClass(fb))
		    && !(fb->access & ACC_STATIC)) {
		    n_object_slots++;
		}
	    }
	    cb1 = cbSuperclass(cb1);
	}
	if (n_object_slots == 0) {
	    CCSet(cb, FieldPrepared);
	    return;
	}
	object_offsets = (unsigned short *)
	    sysMalloc(sizeof(unsigned short) * (n_object_slots + 1));
	if (object_offsets == 0) {
	    ThrowOutOfMemoryError(0, 0);
	    return;
	}
	object_offsets[n_object_slots] = 0;
	cb1 = cb;
	while(cb1) {
	    fb = cbFields(cb1);
	    for (size = 0; size < cbFieldsCount(cb1); size++, fb++) {
	        if ((fieldIsArray(fb) || fieldIsClass(fb))
		    && !(fb->access & ACC_STATIC)) {
		    /* add 1, since 0 is a legal offset. */
		    object_offsets[--n_object_slots] = fb->u.offset + 1;
		}
	    }
	    cb1 = cbSuperclass(cb1);
	}
	sysAssert(n_object_slots == 0);
	cbObjectOffsets(cb) = object_offsets;
    }
    CCSet(cb, FieldPrepared);
}

/* For arguments to native methods, any integral word types
 * (java boolean, byte, short, char, int) can be collapsed into
 * the same signature.  Thus as far as we're concerned
 * foo(int, boolean), bar(char, byte), baz(short, int) all
 * have the same signature.  This allows more hits on
 * custom invokers.  On sparc and x86 we could also collapse
 * float, but this wouldn't port everywhere.  Some kind
 * of conditional compilation here might make that possible.
 * Note that we cannot collapse return types at all, since
 * different shifts/masks must occur for all integral types.
 */
static void
computeTerseSignature(const char *sig, char* ret) 
{
    while (1) {
	switch (*sig++) {
	case 0:
	    *ret = 0;
	    return;
	case SIGNATURE_FUNC:
	    break;
	case SIGNATURE_VOID:
	    *ret++ = TERSE_SIG_VOID;
	    break;
	case SIGNATURE_BOOLEAN:
	    *ret++ = TERSE_SIG_BOOLEAN;
	    break;
	case SIGNATURE_BYTE:
	    *ret++ = TERSE_SIG_BYTE;
	    break;
	case SIGNATURE_SHORT:
	    *ret++ = TERSE_SIG_SHORT;
	    break;
	case SIGNATURE_CHAR:
	    *ret++ = TERSE_SIG_CHAR;
	    break;
	case SIGNATURE_INT:
	    *ret++ = TERSE_SIG_INT;
	    break;
	case SIGNATURE_FLOAT:
	    *ret++ = TERSE_SIG_FLOAT;
	    break;
	case SIGNATURE_LONG:
	    *ret++ = TERSE_SIG_LONG;
	    break;
	case SIGNATURE_DOUBLE:
	    *ret++ = TERSE_SIG_DOUBLE;
	    break;

	case SIGNATURE_ARRAY:
	    *ret++ = TERSE_SIG_OBJECT;
	    while (*sig == SIGNATURE_ARRAY)
	        sig++;
	    if (*sig++ == SIGNATURE_CLASS) {
		while (*sig ++ != SIGNATURE_ENDCLASS) {
		    ;
		}
	    }
	    break;
	case SIGNATURE_CLASS:
	    *ret++ = TERSE_SIG_OBJECT;
	    while (*sig++ != SIGNATURE_ENDCLASS) {
		;
	    }
	    break;
	case SIGNATURE_ENDFUNC:
	    *ret++ = TERSE_SIG_ENDFUNC;
	    break;
	default:
	    /* This should not ever happen.  If we get here there's
	     * either a bug in this loop or the classloader allowed
	     * a bogus signature.
	     */
	    sysAssert(FALSE);
	}
    }
}

static struct methodtable*
AllocMethodTable(ClassClass *cb, 
		 int mslot,
		 struct methodblock **super_methods,
		 int super_methods_count)
{
    struct methodtable *new_table;
    void *ptr;
    /*
     *	This should be the only place that method tables are
     *	allocated.  We allocate more than we need here and mask the
     *	resulting pointer because the methodtable pointer field in
     *	object handles is overloaded and sometimes hold array types
     *	and array lengths.  We must ensure that method table pointers
     *	are allocated on an appropriate boundary so we don't lose any
     *	address bits when we mask off the type portion of the pointer.
     */
    ptr = sysMalloc(sizeof(struct methodtable)
                       + (mslot - 1)* sizeof(struct methodblock *)
                       + METHODALIGNMENT - 1);

    if (ptr == NULL) {
        ThrowOutOfMemoryError(0, 0);
	return NULL;
    }
    cbMethodTableMem(cb) = ptr;
    new_table = (struct methodtable *)(ALIGN_UP(((long)ptr),METHODALIGNMENT));
    new_table->classdescriptor = cb;
    memset((char *)new_table->methods, 0, mslot * sizeof(struct methodblock *));
    if (super_methods) 
	memcpy((char *) new_table->methods,
	       (char *) super_methods,
	       super_methods_count * sizeof(struct methodblock *));
    return new_table;
}

static void
ReleaseTerseSignatures(ClassClass *cb, int n)
{
    ExecEnv * ee = EE();
    int i;
    for (i = 0; i < n; i++) {
        struct methodblock *mb = cbMethods(cb) + i;
	ReleaseUTF8(ee, mb->terse_signature);
	mb->terse_signature = 0;
    }
}

/* Make sure to clean up all allocated memory if anything goes wrong
 * in the middle.
 */
static void
PrepareMethods(ClassClass *cb)
{
    struct methodblock *mb;
    int size;
    struct methodtable *new_table;
    struct methodblock **super_methods;
    unsigned mslot, super_methods_count;
    ExecEnv * ee;

    if (CCIs(cb, MethodPrepared))
        return;

    ee = EE();
    mb = cbMethods(cb);
    for (size = 0; size < (int) cbMethodsCount(cb); size++, mb++) {
        char tsig[255 + 3];
	computeTerseSignature(mb->fb.signature, tsig);
	mb->terse_signature = AddUTF8(ee, tsig);
	if (mb->terse_signature == NULL) {
	    ThrowOutOfMemoryError(0, 0);
	    ReleaseTerseSignatures(cb, size);
	    return;
	}
	mb->fb.u.offset = 0;
	PrepareInvoker(mb);
    }

    if (cbIsInterface(cb)) { 
	/* We don't really need to built a method table for interfaces. */
	cbMethodTable(cb) = NULL;
	cbMethodTableSize(cb) = 0;
	mb = cbMethods(cb);
	/* Each method's offset is its index in the interface */
	for (size = 0; size < (int) cbMethodsCount(cb); size++, mb++) {
	    mb->fb.u.offset = size;
	}
	CCSet(cb, MethodPrepared);
	return;
    }

    if (cbSuperclass(cb) != NULL) { 
	ClassClass *super = cbSuperclass(cb);
	if (cbAccess(super) & ACC_FINAL) {
	    char details[MSG_BUF_LEN];
	    jio_snprintf(details, sizeof(details),
			 "Class %s is subclass of final class %s",
			 cbName(cb), cbName(super));
	    ThrowVerifyError(0, details);
	    ReleaseTerseSignatures(cb, cbMethodsCount(cb));
	    return;
	}
	mslot = cbMethodTableSize(super);
	super_methods = cbMethodTable(super)->methods;
        super_methods_count = cbMethodTableSize(super);
	/* Inherit one's parent's finalizer, if it has one */
	cbFinalizer(cb) = cbFinalizer(cbSuperclass(cb));
    } else { 
	mslot = 1;
	super_methods = NULL;
	super_methods_count = 0;
	cbFinalizer(cb) = NULL;
    }

    mb = cbMethods(cb);
    for (size = 0; size < (int) cbMethodsCount(cb); size++, mb++) {
	int count;

	if ((mb->fb.access & ACC_STATIC) || (mb->fb.access & ACC_PRIVATE))
	    continue;
	if (mb->fb.name == utf8_literal_obj_init_name)
	    continue;

	/* If this item has its own finalizer method, grab it */
	if (mb->fb.name == utf8_literal_finalizer_name &&
	    mb->fb.signature == utf8_literal_finalizer_sig &&
	    cb != classJavaLangObject) {
	    /* try to skip finalizers that are empty methods */
	    if (mb->code_length == 1 &&
		mb->code[0] == opc_return) {
		cbFinalizer(cb) = NULL;
	    } else {
		cbFinalizer(cb) = mb;
	    }
	}

	for (count = 0; count < super_methods_count; count++) {
	    struct methodblock *smb = super_methods[count];
	    if ((smb != NULL) && 
		(smb->fb.name == mb->fb.name) &&
		(smb->fb.signature == mb->fb.signature)) {
	        /* Private methods are not inherited outside of the class. */
	        if (smb->fb.access & ACC_PRIVATE)
		    continue;
	        /* Package-private methods are not inherited outside of the
		   package. */
	        if ((smb->fb.access & ACC_PROTECTED) ||
		    (smb->fb.access & ACC_PUBLIC) ||
		    IsSameClassPackage(smb->fb.clazz, cb)) {
		    if (smb->fb.access & ACC_FINAL) {
		        char details[MSG_BUF_LEN];
			jio_snprintf(details, 
				     sizeof(details),
				     "Class %s overrides final method %s.%s",
				     cbName(cb), 
				     mb->fb.name,
				     mb->fb.signature);
			ThrowVerifyError(0, details);
			ReleaseTerseSignatures(cb, cbMethodsCount(cb));
			return;
		    }
		    if (!CheckSignatureLoaders(smb->fb.signature,
					       cbLoader(cb),
					       cbLoader(smb->fb.clazz))) {
		        ReleaseTerseSignatures(cb, cbMethodsCount(cb));
			return;
		    }
		    mb->fb.u.offset = smb->fb.u.offset;
		    break;
		} 
	    }
	}

	if (mb->fb.u.offset == 0) { 
	    mb->fb.u.offset = mslot;
	    mslot++;
	}
    }

    if (mslot > 65535) {
        ThrowInternalError(0, "exceed 64K method table size limit");
	ReleaseTerseSignatures(cb, cbMethodsCount(cb));
	return;
    }

    new_table = AllocMethodTable(cb, mslot, super_methods, 
				 super_methods_count);
    if (new_table == NULL) {
        ReleaseTerseSignatures(cb, cbMethodsCount(cb));
        return;
    }

    mb = cbMethods(cb);
    for (size = 0; size < (int) cbMethodsCount(cb); size++, mb++) {
	int32_t offset = (int32_t)mb->fb.u.offset;
	if (offset > 0) { 
	    sysAssert(offset < mslot);
	    mt_slot(new_table, offset) = mb;
	}
    }
    cbMethodTable(cb) = new_table;
    cbMethodTableSize(cb) = (unsigned short)mslot;

    CCSet(cb, MethodPrepared);
    return;
}

#define ILLEGAL_ACCESS 0xFFFFFFFF

/* Make sure to clean up all allocated memory if anything goes wrong
 * in the middle.
 */
static void
PrepareInterfaces(ClassClass *cb)
{
    const size_t ITEM_SIZE =  sizeof(cbIntfMethodTable(cb)->itable[0]);
    bool_t isInterface = cbIsInterface(cb);

    if (CCIs(cb, InterfacePrepared))
        return;

    if (cbImplementsCount(cb) == 0 && !isInterface) {
	/* classes that don't implement their own interfaces can just inherit
	 * their parents imethodtable. */
	if (cb == classJavaLangObject) {
	    /* We can preinitialize the imethodtable of java.lang.Object */
	    static struct imethodtable t = { 0 };
	    cbIntfMethodTable(cb) = &t;
	} else { 
	    ClassClass *super = cbSuperclass(cb);
	    cbIntfMethodTable(cb) = cbIntfMethodTable(super);
	} 
	CCSet(cb, InterfacePrepared);
	return;
    } else { 
	cp_item_type *constant_pool = cbConstantPool(cb);
	ClassClass *super = cbSuperclass(cb);
	unsigned char *mallocSpace, *mallocSpaceEnd;
	ClassClass *icb;	/* temporary */
	struct imethodtable *super_itable = cbIntfMethodTable(super);
	struct imethodtable *this_itable = NULL;
	int super_itable_count = super_itable->icount;
	int i, j, k, icount, mcount;

	int n_miranda_methods = 0;
	struct methodblock *miranda_methods;
	struct methodtable *new_table;

	/* Resolve all the interfaces that this class implements, and 
	 * make sure that they all really are interfaces 
	 *
	 * icount will total all the interfaces that this class implements,
	 *    (include interfaces implemented by our superclass, and by 
	 *    interfaces that we implement.)
	 * mcount will total the total amount of space (in sizeof(long)) for 
	 *    which we'll have to allocate space for in the offsets table.
	 */
	icount = super_itable_count + (isInterface ? 1 : 0);
	mcount = 0;
	for (i = 0; i < (int)(cbImplementsCount(cb)); i++) {
	    int interface_index = cbImplements(cb)[i];
	    struct imethodtable *sub_itable;

	    icb = constant_pool[interface_index].clazz;
	    if (!cbIsInterface(icb)) {
	        ThrowIncompatibleClassChangeError(0, "Implementing class");
		return;
	    }
	        
	    sub_itable = cbIntfMethodTable(icb);
	    icount += sub_itable->icount;

	    if (!isInterface) 
		for (j = sub_itable->icount; --j >= 0; ) 
		    mcount += cbMethodsCount(sub_itable->itable[j].classdescriptor);
	}

	{ 
	    size_t this_itable_size = 
		offsetof(struct imethodtable, itable) + icount * ITEM_SIZE;
	    size_t offsets_size = mcount * sizeof(uint32_t);
	    mallocSpace = sysMalloc(this_itable_size + offsets_size);
	    if (mallocSpace == NULL) { 
		ThrowOutOfMemoryError(0, 0);
		return;
	    }
	    mallocSpaceEnd = mallocSpace + this_itable_size + offsets_size;
	    this_itable = (struct imethodtable *)mallocSpace;
	    mallocSpace += this_itable_size;
	    sysAssert(mallocSpace <= mallocSpaceEnd);
	}
	
	cbIntfMethodTable(cb) = this_itable;

	/* Start filling in the table. */
	icount = 0;
	if (isInterface) {
	    this_itable->itable[icount].classdescriptor = cb;
	    this_itable->itable[icount].offsets = NULL;	
	    icount++;
	}
	if (super_itable_count > 0) { 
	    /* We can copy our superclass's offset table.  The offsets
	       will stay the same.  */
	    memcpy(&this_itable->itable[icount], 
		   &super_itable->itable[0], 
		   super_itable_count * ITEM_SIZE);
	    icount += super_itable_count;
	}
	for (i = 0; i < (int)(cbImplementsCount(cb)); i++) {
	    /* Add the interfaces that we implement, either directly, or
	     * because those interfaces implement other interfaces.    */
	    int interface_index = cbImplements(cb)[i]; 
	    icb = constant_pool[interface_index].clazz;
	    memcpy(&this_itable->itable[icount], 
		   &cbIntfMethodTable(icb)->itable[0],
		   cbIntfMethodTable(icb)->icount * ITEM_SIZE);
	    icount += cbIntfMethodTable(icb)->icount;
	}
	sysAssert(!isInterface || super_itable_count == 0);
	/* Delete duplicates from the table.  This is very rare, so it can
	 * be quite inefficient.  */
	for (i = (isInterface ? 1 : super_itable_count); i < icount; i++) {
	    icb = this_itable->itable[i].classdescriptor;
	    for (j = 0; j < i; j++) { 
		if (icb == this_itable->itable[j].classdescriptor) { 
		    /* We have an overlap.  Item i is a duplicate.  Delete from
		     * the table */
		    for (k = i + 1; k < icount; k++) 
			this_itable->itable[k - 1] = this_itable->itable[k];
		    icount--;
		    i--; 	/* Reconsider this entry! */
		    break;
		}
	    }
	}
	this_itable->icount = icount;
	if (isInterface) { 
	    /* Nothing more to do for interfaces */
	    CCSet(cb, InterfacePrepared);
	    return;
	}

	/* For each additional interface . . . */
	for (i = super_itable_count; i < icount; i++) { 
	    /* The table length is the number of interface methods */
	    ClassClass *intfi = this_itable->itable[i].classdescriptor;
	    int intfi_count = cbMethodsCount(intfi);
	    unsigned long *offsets = (unsigned long *)mallocSpace;
	    mallocSpace += intfi_count * sizeof(unsigned long);
	    sysAssert(mallocSpace <= mallocSpaceEnd);
	    this_itable->itable[i].offsets = offsets;

	    /* Look at each interface method */
	    for (j = 0; j < intfi_count; j++) { 
		struct methodblock *imb = cbMethods(intfi) + j;
		if (!CheckSignatureLoaders(imb->fb.signature,
					   cbLoader(cb), 
					   cbLoader(imb->fb.clazz))) {
		    sysFree(cbIntfMethodTable(cb));
		    cbIntfMethodTable(cb) = NULL;
		    return;
		}
		if ((imb->fb.access & ACC_STATIC) != 0) { 
		    sysAssert(imb->fb.name == utf8_literal_cls_init_name);
		    offsets[j] = 0; 
		} else {
		    /* Find the class method that implements the interface 
		     * method */
		    struct methodblock **mbt = cbMethodTable(cb)->methods;
		    for (k = cbMethodTableSize(cb) - 1; k >= 0; --k) { 
		        struct methodblock *mb = mbt[k];
			if (mb != NULL && 
			    mb->fb.name == imb->fb.name &&
			    mb->fb.signature == imb->fb.signature) {

			    if (IsPublic(mb->fb.access)) {
			        offsets[j] = mb->fb.u.offset;
				break;
			    } else {
			        offsets[j] = ILLEGAL_ACCESS;
				n_miranda_methods++;
				break;
			    }
			}
		    }
		    if (k == -1) {
		        /* Make sure there is no private method trying to
			 * implement the interface method.
			 */
		        for (k = cbMethodsCount(cb) - 1; k >= 0; --k) {
			    struct methodblock *mb = cbMethods(cb) + k;
			    if (mb->fb.name == imb->fb.name &&
				mb->fb.signature == imb->fb.signature &&
				!(mb->fb.access & ACC_STATIC)) {
			        offsets[j] = ILLEGAL_ACCESS;
				break;
			    }
			}
			if (k < 0) {
			    offsets[j] = 0;
			}
			n_miranda_methods++;
		    }
		}
	    }
	}

	if (n_miranda_methods == 0) {
	    CCSet(cb, InterfacePrepared);
	    return;
	}

	miranda_methods = sysCalloc(n_miranda_methods, 
				    sizeof(struct methodblock));
	if (miranda_methods == NULL) {
	    ThrowOutOfMemoryError(0, 0);
	    sysFree(cbIntfMethodTable(cb));
	    cbIntfMethodTable(cb) = NULL;
	    return;
	}

	mcount = cbMethodTableSize(cb);
	{
	    /* Expand the method table */ 
	    void *ptr = cbMethodTableMem(cb);
	    new_table = AllocMethodTable(cb, 
					 mcount + n_miranda_methods, 
					 cbMethodTable(cb)->methods,
					 mcount);
	    if (new_table == 0) {
	        sysFree(cbIntfMethodTable(cb));
		cbIntfMethodTable(cb) = NULL;
		sysFree(miranda_methods);
	        return;
	    }
	    sysFree(ptr);
	}

	for (i = super_itable_count; i < icount; i++) { 
	    /* The table length is the number of interface methods */
	    ClassClass *intfi = this_itable->itable[i].classdescriptor;
	    int intfi_count = cbMethodsCount(intfi);
	    unsigned long *offsets = this_itable->itable[i].offsets;
	    int count = n_miranda_methods;

	    /* Look at each interface method */
	    for (j = 0; j < intfi_count; j++) { 
		struct methodblock *imb = cbMethods(intfi) + j;
		if ((imb->fb.access & ACC_STATIC) == 0 && 
		    (offsets[j] == 0 || offsets[j] == ILLEGAL_ACCESS)) {
		    struct methodblock *mb;
		    count--;
		    mb = miranda_methods + count;

		    mb->fb.clazz = cb;
		    if (offsets[j] == ILLEGAL_ACCESS) {
		        char buf[STK_BUF_LEN];
			/* create a *fake* name that begins with '+', not
			 * to be in conflict with other methods.
			 */
			jio_snprintf(buf, sizeof(buf), "+%s", imb->fb.name);
		        mb->fb.name = AddUTF8(EE(), buf);
			if (mb->fb.name == NULL) {
			    ThrowOutOfMemoryError(0, 0);
			    sysFree(cbIntfMethodTable(cb));
			    cbIntfMethodTable(cb) = NULL;
			    sysFree(miranda_methods);
			    sysFree(cbMethodTableMem(cb));
			    return;
			}
		    } else {
		        mb->fb.name = AddUTF8(EE(), imb->fb.name);
		    }
		    mb->fb.signature = imb->fb.signature;
		    mb->fb.access = imb->fb.access | ACC_MIRANDA_METHOD;
		    mb->fb.u.offset = mcount;
		    mb->terse_signature = imb->terse_signature;
		    mb->args_size = imb->args_size;

		    /* In reflection code (jvm.c), we want to obtain the
		     * interface method block from the miranda method
		     * created by the VM. Thus we store the interface 
		     * method block in mb->code
		     */
		    mb->code = (unsigned char *)imb;

		    PrepareInvoker(mb);
		    mt_slot(new_table, mcount) = mb;
		    offsets[j] = mcount;
		    mcount++;
		}
	    }
	    sysAssert(0 == count);
	}
	cbMethodTable(cb) = new_table;
	cbMethodTableSize(cb) = (unsigned short)mcount;
	cbMirandaMethods(cb) = miranda_methods;
	cbMirandaMethodsCount(cb) = n_miranda_methods;
    }
    CCSet(cb, InterfacePrepared);
}

/* Note that the LinkClass function is restartable. If anything goes
 * wrong during linking (due to, for example, out of memory), the 
 * function returns with an exception. However, no unrecoverable 
 * change will be made to the class. If LinkClass is later called
 * again, it should proceed normally to complete the unfinished work.
 */
void
LinkClass(ClassClass *cb) 
{
    ExecEnv *ee = EE();
    int i;

    if (CCIs(cb,Error)) {
        ThrowNoClassDefFoundError(0, cbName(cb));
	return;
    }

    if (CCIs(cb,Linked))
        return;

    sysAssert(CCIs(cb, Loaded));

    {
        /*
	 * Make sure the current class links with the system Throwable
	 * class. See bug 4157312.
	 */
        ClassClass *myThrowable = 
	    FindClassFromClass(ee, "java/lang/Throwable", FALSE, cb);
	if (myThrowable == NULL) {
	    /* exception thrown */
	    return;
	}
	if (classJavaLangThrowable && classJavaLangThrowable != myThrowable) {
	    char buf[MSG_BUF_LEN];
	    jio_snprintf(buf, sizeof(buf), 
			 "wrong Throwable class linked with %s", cbName(cb));
	    SignalError(0, "java/lang/LinkageError", buf);
	    return;
	}
    }

    /* link superclass */
    if (cbSuperclass(cb)) {
        if (cbIsInterface(cbSuperclass(cb))) {
	    char buf[256];
	    jio_snprintf(buf, sizeof(buf),
			 "Class %s has interface %s as super class",
			 cbName(cb),
			 cbName(cbSuperclass(cb)));
	    ThrowIncompatibleClassChangeError(0, buf);
	    return;
	}
	LinkClass(cbSuperclass(cb));
	if (exceptionOccurred(ee)) {
	    return;
	}
    }

    /* link interfaces */
    for (i = 0; i < (int)(cbImplementsCount(cb)); i++) {
        int interface_index = cbImplements(cb)[i];
	ClassClass *icb = cbConstantPool(cb)[interface_index].clazz;
	LinkClass(icb);
	if (exceptionOccurred(ee)) {
	    return;
	}
    }
    
    /* In case the class is linked in the process of linking its
       superclasses. */
    if (CCIs(cb,Linked)) {
        return;
    }
    /* It is OK to enter the class monitor here because no methods
     * in the class will be executing yet at this moment.
     */
    monitorEnter2(ee, obj_monitor(cb));
    /* In case the class is linked while waiting for the lock. */
    if (CCIs(cb,Linked)) {
        goto unlock;
    }

    if (NEED_VERIFY(cbLoader(cb) != NULL)) {
        VerifyClass(cb);
	if (exceptionOccurred(ee)) {
	    goto unlock;
	}
    }

    /* The verification process may have (recursively) prepared cb. */
    if (CCIs(cb, Linked))
        goto unlock;

    PrepareClass(cb);
    if (exceptionOccurred(ee)) {
	goto unlock;
    }

    CCSet(cb, Linked);

 unlock:
    monitorExit2(ee, obj_monitor(cb));
    return;
}

/*
 * Detect class circularities from LoadSuperclasses()
 */

static void
pushSeen(ExecEnv *ee, struct seenclass *seen)
{
    seen->next = ee->seenclasses;
    ee->seenclasses = seen;
}

static void
popSeen(ExecEnv *ee, struct seenclass *seen)
{
    if (seen != ee->seenclasses) /* paranoia */
	panic("popSeen: corrupt seen class stack");
    ee->seenclasses = ee->seenclasses->next;
}

static bool_t
checkSeen(ExecEnv *ee, ClassClass *cb)
{
    struct seenclass *seen = ee->seenclasses;
    while (seen) {
	if (cb == seen->cb)
	    return TRUE;
	seen = seen->next;
    }
    return FALSE;
}

/* Attempt to boot strap system classes. Note that this function
 * may be called recursively due to recursive class loading. However,
 * as soon as the function returns for the first time, system classes
 * should have been all set up.
 */
static bool_t
AttemptToBootstrapSystemClasses(ClassClass *cb)
{
    /* We need to set the classJavaLangClass and classJavaLangObject
     * global variables as soon as we've loaded the corresponding 
     * classes. Various parts of class loading relys on these variables
     * to be set. 
     */
    if (!cbLoader(cb)) {
        if (cbName(cb) == utf8_literal_java_lang_Class)
	    classJavaLangClass = cb;

	if (cbName(cb) == utf8_literal_java_lang_Object)
	    classJavaLangObject = cb;
    }

    if (classJavaLangClass == NULL) {
        int j;
	ClassClass **pcb;
	sys_thread_t *self = sysThreadSelf();

        if (!FindStickySystemClass(0, JAVAPKG "Class", TRUE)) {
	    /* exception already thrown */
	    return FALSE;
	}
	/* We need this for bootstrapping. java.lang.Class is
	 * already linked by FindStickySystemClass.
	 */
	BINCLASS_LOCK(self);
	for (j = nbinclasses, pcb = binclasses; --j >= 0; pcb++) {
	    cbHandle(*pcb)->methods = cbMethodTable(classJavaLangClass);
	}
	BINCLASS_UNLOCK(self);

	/* Load the primitive classes after the above backpatch.
	 * The primitive classes do not appear in the binclass
	 * table. */
	if (!InitPrimitiveClasses())
	    return FALSE;
    }

    /* Make sure we only try to load the system classes once. */
    if (classJavaLangClass != cb) {
        return TRUE;
    }

    classJavaLangObject = 
        FindStickySystemClass(NULL, JAVAPKG "Object", TRUE);
    if (classJavaLangObject == 0)
        return FALSE;

    classJavaLangString = 
        FindStickySystemClass(NULL, JAVAPKG "String", TRUE);
    if (classJavaLangString == 0)
        return FALSE;

    classJavaLangThread = 
        FindStickySystemClass(NULL, JAVAPKG "Thread", TRUE);
    if (classJavaLangThread == 0)
        return FALSE;

    /* Initialize this for the old native method interface. see oldnmi.c */
    Thread_classblock = classJavaLangThread;

    classJavaLangClassLoader = 
        FindStickySystemClass(NULL, JAVAPKG "ClassLoader", TRUE);
    if (classJavaLangClassLoader == 0)
        return FALSE;

    classJavaLangThreadDeath = 
        FindStickySystemClass(NULL, JAVAPKG "ThreadDeath", TRUE);
    if (classJavaLangThreadDeath == 0)
        return FALSE;

    classJavaLangThrowable =
        FindStickySystemClass(NULL, JAVAPKG "Throwable", TRUE);
    if (classJavaLangThrowable == 0)
        return FALSE;

    classJavaLangException =
        FindStickySystemClass(NULL, JAVAPKG "Exception", TRUE);
    if (classJavaLangException == 0)
        return FALSE;

    classJavaLangError = 
        FindStickySystemClass(NULL, JAVAPKG "Error", TRUE);
    if (classJavaLangError == 0)
        return FALSE;

    classJavaLangRuntimeException = 
        FindStickySystemClass(NULL, JAVAPKG "RuntimeException", TRUE);
    if (classJavaLangRuntimeException == 0)
        return FALSE;

    interfaceJavaLangCloneable =
        FindStickySystemClass(NULL, JAVAPKG "Cloneable", TRUE);
    if (interfaceJavaLangCloneable == 0)
        return FALSE;

    interfaceJavaIoSerializable =
        FindStickySystemClass(NULL, "java/io/Serializable", TRUE);
    if (interfaceJavaIoSerializable == 0)
        return FALSE;

    /* Make sure the java.lang.StackOverflowError is preloaded
     * so that we don't need to load it when stack overflow
     * actually occurs. Class loading may require stack space.
     */
    if (!FindStickySystemClass(NULL, JAVAPKG "StackOverflowError", TRUE))
        return FALSE;

    /* Preload IOException and FileNotFoundException so that we can print
       a sensible error message even if we run out of file descriptors */
    if (!FindStickySystemClass(NULL, JAVAIOPKG "IOException", TRUE))
        return FALSE;
    if (!FindStickySystemClass(NULL, JAVAIOPKG "FileNotFoundException", TRUE))
        return FALSE;

    /* Used to skip Method.invoke when in getCallerFrame. */
    {
	int i = 0;
	struct methodblock *mb = NULL;
	cb = FindStickySystemClass(NULL, JAVAPKG "reflect/Method", TRUE);
	if (cb == NULL)
	    return FALSE;
	mb = cbMethods(cb);
	for (i = cbMethodsCount(cb) - 1; i >= 0; i--, mb++) {
	    if (strcmp(mb->fb.name, "invoke") == 0) {
	        reflect_invoke_mb = mb;
		break;
	    }
	}
	if (reflect_invoke_mb == NULL)
	    return FALSE;
    }

    return TRUE;
}

static void
LoadSuperclasses0(ClassClass * cb, ExecEnv *ee) 
{ 
    ClassClass *super = 0;
    int i;

#ifndef TRIMMED
    if (verbose)
	jio_fprintf(stderr, "[Loading superclasses of %s]\n", cbName(cb));
#endif

    if (cbName(cb) == utf8_literal_java_lang_ref_Reference && 
        cbLoader(cb) == NULL) {
	/* During initialization the GC loads the Reference class before it
           loads any of the Reference subclasses, so there is no danger of
           loading one of these subclasses first and thereby setting the
           IsReference bit in its class block to the wrong value */
	CCSet(cb, Reference);
    }

    if (cbSuperclass(cb) == NULL) {
	if (cbSuperName(cb)) {
	    super = FindClassFromClass(ee, cbSuperName(cb), FALSE, cb);
	    if (super != NULL) { 
	        sysAssert(CCIs(super, Loaded));
		/* Don't allow a class to have a superclass it can't access */
		if (!VerifyClassAccess(cb, super, FALSE)) {
		    char buf[MSG_BUF_LEN];
		    jio_snprintf(buf, sizeof(buf), 
				 "cannot access superclass %s from class %s",
				 cbName(super), cbName(cb));
		    ThrowIllegalAccessError(ee, buf);
		    super = NULL;
		} 
	    }
	    if (super != NULL) {
		cbSuperclass(cb) = cbHandle(super);
		if (CCIs(super, Reference)) {
		    CCSet(cb, Reference);
		}
	    } else {
		cbSuperclass(cb) = NULL;
		/* exception already thrown */
		return;
	    } 
	} else if (cb == classJavaLangObject) {
	    cbSuperclass(cb) = 0;
	} else {
	    /* should not happen, every class other than Object should have 
	       a superclass */
	    sysAssert(FALSE); 
	}
    } 

    for (i = 0; i < (int)(cbImplementsCount(cb)); i++) {
        /* Be very careful, since not verified yet. . . 
	 */
        int interface_index = cbImplements(cb)[i];
	cp_item_type *constant_pool = cbConstantPool(cb);
	unsigned char *type_table = 
	    constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
	ClassClass *icb;
	char *interface_name;

	if (type_table[interface_index] == CONSTANT_Class) {
	    int name_index = constant_pool[interface_index].i;
	    interface_name = constant_pool[name_index].cp;
	    icb = 0;
	} else if (type_table[interface_index] == 
		   (CONSTANT_Class | CONSTANT_POOL_ENTRY_RESOLVED)) {
	    interface_name = 0;
	    icb = constant_pool[interface_index].clazz;
	} else {
	    /* should not happen. */
	    icb = 0;
	    interface_name = 0;
	    sysAssert(FALSE);
	}

	if (icb == 0) {
	    icb = FindClassFromClass(ee, interface_name, FALSE, cb);
	    if (icb) {
		/* Don't allow a class to have an interface it can't access */
		if (!VerifyClassAccess(cb, icb, FALSE)) {
		    char buf[MSG_BUF_LEN];
		    jio_snprintf(buf, sizeof(buf), 
				 "cannot access interface %s from class %s",
				 cbName(icb), cbName(cb));
		    ThrowIllegalAccessError(ee, buf);
		    return;
		} 
	        constant_pool[interface_index].clazz = icb;
		type_table[interface_index] |= CONSTANT_POOL_ENTRY_RESOLVED;
	    }
	} else {
	    if (checkSeen(ee, icb)) {
	        ThrowClassCircularityError(0, cbName(icb));
		return;
	    }
	}
	if (icb == 0) {
	    return;
	}
    }

    CCSet(cb, Loaded);

    return;
}

void
LoadSuperclasses(ClassClass *cb)
{
    static bool_t need_bootstrap = TRUE;
    struct seenclass this[1];
    ExecEnv *ee = EE();
    if (CCIs(cb, Loaded))
        return;
    if (need_bootstrap) {
        if (!AttemptToBootstrapSystemClasses(cb)) {
	    return;
	}
	need_bootstrap = FALSE;
    }
    if (checkSeen(ee, cb)) {
        ThrowClassCircularityError(0, cbName(cb));
	return;
    }    
    this->cb = cb;
    pushSeen(ee, this);
    monitorEnter2(ee, obj_monitor(cb));
    if (!CCIs(cb, Loaded)) {
        LoadSuperclasses0(cb, ee);
    }
    monitorExit2(ee, obj_monitor(cb));
    popSeen(ee, this);
}

static void
PrepareClass(ClassClass *cb)
{
    ExecEnv *ee = EE();
    ClassClass *super = cbSuperclass(cb);

#ifndef TRIMMED
    if (verbose)
        jio_fprintf(stderr, "[Preparing %s]\n", cbName(cb));
#endif

    PrepareFields(cb, super ? cbInstanceSize(super) : 0);
    if (exceptionOccurred(ee))
        return;
    PrepareMethods(cb);
    if (exceptionOccurred(ee))
        return;
    PrepareInterfaces(cb);
    if (exceptionOccurred(ee))
        return;

#ifdef BREAKPTS
    if (debugging) {
        notify_debugger_of_class_prepare(EE(), (JHandle *)cb);
    }
#endif 

}

/*
 * This is called by FindClassFromClass when a class name is being requested
 * by another class that was loaded via a classloader.
 */
static ClassClass *
ClassLoaderFindClass(ExecEnv *ee, struct Hjava_lang_ClassLoader *loader, 
		     char *name)
{
    static jmethodID loadClassID;
    jclass result;
    ClassClass *result_cb = 0;
    JNIEnv *env = EE2JNIEnv(ee);
    jstring stringName;
    char dot_name_buf[STK_BUF_LEN];
    char *dot_name;
    int i;

    if ((*env)->PushLocalFrame(env, 4) < 0) {
        return 0;
    }

    if (strlen(name) >= sizeof(dot_name_buf)) {
        dot_name = sysMalloc(strlen(name) + 1);
	if (dot_name == NULL) {
	    (*env)->PopLocalFrame(env, 0);
	    ThrowOutOfMemoryError(0, 0);
	    return 0;
	}
    } else {
        dot_name = dot_name_buf;
    }

    for (i = 0; name[i]; i++) {
        if (name[i] == '/') {
	    dot_name[i] = '.';
        } else {
	    dot_name[i] = name[i];
	}
    }
    dot_name[i] = 0;

    if (loadClassID == NULL) {
	jclass loaderCls = MkRefLocal(env, classJavaLangClassLoader);
	loadClassID = (*env)->GetMethodID(env, 
					  loaderCls, 
					  "loadClass",
				  "(Ljava/lang/String;)Ljava/lang/Class;");
	if (loadClassID == NULL) {
	    goto end;
	}
    }

    stringName = (*env)->NewStringUTF(env, dot_name);
    if (stringName == 0) {
        goto end;
    }
    result = (*env)->CallObjectMethod(env,
	    			      MkRefLocal(env, loader),
	    			      loadClassID,
	    			      stringName);
    result_cb = (ClassClass *)DeRef(env, result);

    if (!exceptionOccurred(ee) && result_cb == 0) {
	/* Loader returned NULL without raising an exception. */
	ThrowClassNotFoundException(ee, name);
	goto end;
    }
    
    /* We do not trust the ClassLoader.loadClass method actually returns
     * the right class.
     */
    if (!exceptionOccurred(ee) && strcmp(name, cbName(result_cb)) != 0) {
        char buf[MSG_BUF_LEN];
	jio_snprintf(buf, sizeof(buf), "Bad class name (expect: %s, get: %s)",
		     name, cbName(result_cb));
        ThrowNoClassDefFoundError(ee, buf);
	result_cb = 0;
    }

 end:
    (*env)->PopLocalFrame(env, 0);
    if (dot_name != dot_name_buf) {
        sysFree(dot_name);
    }
    return result_cb;
}

static void 
PrintVerboseClassDep(ExecEnv *ee, char *name)
{
    static long init_time;
    static long last_time;
    static num = 0;
    long cur_time = sysGetMilliTicks();
    JavaFrame frame_buf;	    
    JavaFrame *frame = ee->current_frame;
    while (frame) {
        struct methodblock *mb = frame->current_method;
	if (mb) {
	    if (strcmp(cbName(mb->fb.clazz), "java/lang/ClassLoader") == 0
		&& strcmp(mb->fb.name, "findSystemClass0") == 0) {
	        return;
	    }
	    if (mb->fb.clazz != classJavaLangClass
		|| 0 != strcmp(mb->fb.name, "forName")) {
		break;
	    }
	}
	frame = FRAME_PREV(frame, &frame_buf);
    }
    
    if (num == 0) {
        init_time = cur_time;
        last_time = cur_time;
    }
    {
        char buf[256];
        jio_fprintf(stderr, "[%3d %5ldms %3ld] ", num++, 
		    cur_time - init_time, cur_time - last_time);
	last_time = cur_time;
	if (frame &&
	    pc2string(frame->lastpc, frame, buf, buf + sizeof(buf)) &&
	    buf[0] != '\0') {
	    jio_fprintf(stderr, "%s", buf);
	} else {
	    jio_fprintf(stderr, "<system>");
	}
	jio_fprintf(stderr, " => %s\n", 
		    classname2string(name, buf, sizeof(buf)));
    }
}

/* Find the class with the given name.  
 * If "resolve" is true, then we initialize the class.
 */
ClassClass *
FindClass(struct execenv *ee, char *name, bool_t resolve)
{
    /* skip pseudo frames */
    ClassClass *from = ee ? getCallerClass(ee, 0) : 0;
    return FindClassFromClass(ee, name, resolve, from);
}

/* The last argument specifies whether this function throws 
 * ClassNotFoundException or NoClassDefFoundError when the class
 * file is not found.
 */
ClassClass *
FindClassFromClassLoader(struct execenv *ee, char *name, bool_t resolve,
			 struct Hjava_lang_ClassLoader *loader, 
			 bool_t throwError)
{
    ClassClass *cb = 0;
    sys_thread_t *self;

    ee = ee ? ee : EE();
    self = EE2SysThread(ee);
    
    if (loader) {
        /* Although class loaders are supposed to define loadClass as
	 * a synchronized method, the VM does not trust the class loader
	 * to do so. It enters the loader monitor just in case.
	 */
        monitorEnter2(ee, obj_monitor(loader));
    } else {
        SYSLOADER_LOCK(self);
    }

    BINCLASS_LOCK(self);
    {
        char *hashed_name = HashUTF8(ee, name);
	if (hashed_name)
	    cb = LookupLoaderCache(hashed_name, loader);
    }
    BINCLASS_UNLOCK(self);

    if (cb) {
        if (!CCIs(cb, Loaded)) {
	    LoadSuperclasses(cb);
	    if (!CCIs(cb, Loaded)) {
	        cb = 0;
	    }
	}
    } else {
        if (verboseclassdep) {
	    PrintVerboseClassDep(ee, name);
	}
	if (DisableAsyncEvents(ee) == 0) {
            if (name[0] == SIGNATURE_ARRAY) {
	        cb = FindArrayClassFromClassLoader(ee, name, loader,
						   throwError);
	    } else if (loader) {
	        cb = ClassLoaderFindClass(ee, loader, name);
		/* If the thrown exception is ClassNotFoundException, and
		 * throwError is TRUE, we rethrow a NoClassDefFoundError.
		 */
		if (exceptionOccurred(ee) && throwError) {
		    JHandle *exc = ee->exception.exc;
		    ClassClass *clsClassNotFoundException;
		    
		    exceptionClear(ee);
		    clsClassNotFoundException =
		        FindClass(ee, "java/lang/ClassNotFoundException",
				  TRUE);
		    if (clsClassNotFoundException) {
		        if (is_instance_of(exc, clsClassNotFoundException, 
					   ee)) {
			    ThrowNoClassDefFoundError(ee, name);
			} else {
			    exceptionThrow(ee, exc);
			}
		    }
		    cb = 0;
		}
	    } else {
	        cb = LoadClassLocally(name);
		if (cb) {
		    MakeClassSticky(cb);
		    LoadSuperclasses(cb);
		    if (!CCIs(cb, Loaded)) {
		        cb = 0;
		    }
		}
	    }

	    if (EnableAsyncEvents(ee)) {
	        cb = 0;
	    }
	}

	if (cb) {
	    char buf[MSG_BUF_LEN];
	    char *ename;

	    BINCLASS_LOCK(self);
	    ename = UpdateLoaderCache(cb, loader, buf, sizeof(buf));
	    BINCLASS_UNLOCK(self);

	    if (ename) {
		SignalError(ee, ename, buf);
	        cb = 0;
	    }
	}
    }

    if (loader) {
        monitorExit2(ee, obj_monitor(loader));
    } else {
        SYSLOADER_UNLOCK(self);
    }

    if (cb) {
	if (resolve) {
	    InitClass(cb);
	    if (exceptionOccurred(ee))
	        return 0;
	}
	return cb;
    } else {
        if (!exceptionOccurred(ee)) {
	    if (throwError) {
	        /* Do not try to recursively load NoClassDefFoundError */
	        if (loader == NULL &&
		    strcmp(name, "java/lang/NoClassDefFoundError") == 0) {
		    jio_fprintf(stderr, 
     "Can't find class java.lang.NoClassDefFoundError. (Wrong class path?)\n");
		    Exit(1);
		}
		ThrowNoClassDefFoundError(ee, name);
	    } else {
	        ThrowClassNotFoundException(ee, name);
	    }
	}
	return 0;
    }
}

ClassClass *
FindClassFromClass(struct execenv *ee, char *name, bool_t resolve,
		   ClassClass *from)
{
    return FindClassFromClassLoader(ee, name, resolve, 
			       (from) ? cbLoader(from) : NULL,
			       TRUE);
} 

/*
 * FindStickySystemClass is a variant of FindClassFromClass that makes the 
 * class sticky (immune to class GC) if the class can be found.  It is only
 * used on system classes, i.e. classes with no class loader, so it's 
 * equivalent to FindClassFromClass with "from" argument 0.  It returns 0
 * if the class can't be found, and like FindClass assumes that the caller
 * will deal with that.  Note that until the class has been made sticky it
 * must be kept somewhere where GC will find it.
 */
ClassClass *
FindStickySystemClass(struct execenv *ee, char *name, bool_t resolve)
{
    ClassClass *cb;
    if ((cb = FindClassFromClass(ee, name, resolve, 0)) != NULL) {
	MakeClassSticky(cb);
    }
    return cb;
}

/* Find the array class with the specified name.  */
static ClassClass *
FindArrayClassFromClassLoader(struct execenv *ee, char *name,
			      struct Hjava_lang_ClassLoader *from_loader,
			      bool_t throwError)
{ 
    char *name_p;
    int depth, base_type;
    ClassClass *cb = NULL;
    struct Hjava_lang_ClassLoader *loader;
    ClassClass *inner_cb;

    sysAssert(name[0] == SIGNATURE_ARRAY);

    /* Strip off all the initial SIGNATURE_ARRAY's to determine the depth,
     * and also what's left.  When we're done, name_p points at the first
     * non-character, and depth is the count of the array dpeth
     */
    for (name_p = name, depth = 0; *name_p == SIGNATURE_ARRAY; 
	 name_p++, depth++);

    /* Look at the character to determine what type of array we have. */
    switch(*name_p++) {
        case SIGNATURE_INT:    base_type = T_INT; break;
	case SIGNATURE_LONG:   base_type = T_LONG; break;
	case SIGNATURE_FLOAT:  base_type = T_FLOAT; break;
	case SIGNATURE_DOUBLE: base_type = T_DOUBLE; break;
	case SIGNATURE_BOOLEAN:base_type = T_BOOLEAN; break; 
	case SIGNATURE_BYTE:   base_type = T_BYTE; break;
	case SIGNATURE_CHAR:   base_type = T_CHAR; break;
	case SIGNATURE_SHORT:  base_type = T_SHORT; break;
	case SIGNATURE_CLASS:  base_type = T_CLASS; break;
        default:	       return NULL;	/* bad signature */
    }
    if (base_type == T_CLASS) { 
	char buffer_space[50];
	char *buffer = buffer_space;
	char *endChar = strchr(name_p, SIGNATURE_ENDCLASS);
	size_t nameLength = endChar - name_p;
	if (endChar == NULL || endChar[1] != '\0')
	    return NULL;
	if (nameLength >= sizeof(buffer_space)) { /* need bigger buffer */
	    buffer = sysMalloc(nameLength + 1);
	    if (buffer == 0) {
		ThrowOutOfMemoryError(0, 0);
		return NULL;
	    }
	}
	memcpy(buffer, name_p, nameLength);
	buffer[nameLength] = '\0';
	inner_cb = FindClassFromClassLoader(ee, buffer, FALSE, from_loader,
					    throwError);
	if (buffer != buffer_space) /* free buffer, if we malloc'ed it */
	    sysFree(buffer);
	if (inner_cb == NULL) 
	    return NULL;
	/* loader should either be fromLoader or NULL. */
	loader = cbLoader(inner_cb);
    } else { 
	if (*name_p != '\0')			/* bad signature */
	    return NULL;
	inner_cb = NULL;
	loader = NULL;
    }

    /* Is this really necessary? Should this have led to
       ClassCircularityError? */
    /* In case the array class has been loaded since our last check at the
     * beginning of the function.
     */
    BINCLASS_LOCK(sysThreadSelf());
    {
        char *hashed_name = HashUTF8(ee, name);
	if (hashed_name) {
	    cb = LookupLoaderCache(hashed_name, loader);
	}
    }
    BINCLASS_UNLOCK(sysThreadSelf());

    if (cb == NULL) { 
	/* Create the fake array class corresponding to this. */
	cb = createFakeArrayClass(name, base_type, depth, inner_cb, loader);
	if (cb) {
	    if (loader == NULL) {
	        MakeClassSticky(cb);
	    }
	    LoadSuperclasses(cb);
	    if (!CCIs(cb, Loaded)) {
	        cb = NULL;
	    }
	}
    }
    return cb;
}

/*
 * Pseudo-classes to represent primitive Java types.
 */

ClassClass *class_void;
ClassClass *class_boolean;
ClassClass *class_byte;
ClassClass *class_char;
ClassClass *class_short;
ClassClass *class_int;
ClassClass *class_long;
ClassClass *class_float;
ClassClass *class_double;

struct primtype {
    char		*name;
    char		typesig;
    unsigned char	typecode;
    unsigned char	slotsize;
    unsigned char	elementsize;
    ClassClass		**cellp;
} const primitive_types[] = {
    { "void",	SIGNATURE_VOID,	   T_VOID,	0, 0,	&class_void },
    { "boolean",SIGNATURE_BOOLEAN, T_BOOLEAN,	4, 1,	&class_boolean },
    { "byte",	SIGNATURE_BYTE,	   T_BYTE,	4, 1,	&class_byte },
    { "char",	SIGNATURE_CHAR,	   T_CHAR,	4, 2,	&class_char },
    { "short",	SIGNATURE_SHORT,   T_SHORT,	4, 2,	&class_short },
    { "int",	SIGNATURE_INT,	   T_INT,	4, 4,	&class_int },
    { "long",	SIGNATURE_LONG,	   T_LONG,	8, 8,	&class_long },
    { "float",	SIGNATURE_FLOAT,   T_FLOAT,	4, 4,	&class_float },
    { "double",	SIGNATURE_DOUBLE,  T_DOUBLE,	8, 8,	&class_double }
};

#define	PRIMITIVE_TYPE_COUNT	\
	(sizeof (primitive_types) / sizeof (primitive_types[0]))

#define	GET_PRIMITIVE_CLASS(nm) \
	(class_##nm ? class_##nm : FindPrimitiveClass(#nm))

ClassClass *
FindPrimitiveClass(char *name)
{
    int i;
    ClassClass *cb;
    const struct primtype *p;
    ExecEnv *ee = EE();
    
    for (i = 0; i < PRIMITIVE_TYPE_COUNT; i++) {
	p = &primitive_types[i];
	if (strcmp(name, p->name) == 0) {
	    cb = *p->cellp;
	    if (cb == NULL) {
		cb = createPrimitiveClass(p->name, p->typesig, p->typecode,
		    p->slotsize, p->elementsize);
		if (cb == NULL)
		    return NULL;
		LoadSuperclasses(cb);
		if (!CCIs(cb, Loaded)) {
		    return NULL;
		}
		InitClass(cb);
		if (exceptionOccurred(ee)) {
		    return NULL;
		}
		*p->cellp = cb;
	    }
	    return cb;
	}
    }
    return NULL;
}

static bool_t
InitPrimitiveClasses(void)
{
    int i;
    for (i = 0; i < PRIMITIVE_TYPE_COUNT; i++) {
        if (!FindPrimitiveClass(primitive_types[i].name)) {
	    return FALSE;
	}
    }
    return TRUE;
}

/*
 * Enumerate over all classes in the loader cache calling a function for
 * each one.  Expects the caller to lock BINCLASS_LOCK
 */
int
EnumerateOverLoaderCache(int (*func)(ClassClass *,
                                     struct Hjava_lang_ClassLoader *,
                                     void *), 
                         void *arg)
{
    int i;
    int ret = SYS_OK;

    sysAssert(BINCLASS_LOCKED(sysThreadSelf()));

    for (i = 0; i < LOADER_CACHE_TABLE_SIZE; i++) {
        loader_cache_t *entry = loader_cache[i];
        while (entry != NULL) {
            ret = func(entry->cb, entry->loader, arg);
            if (ret != SYS_OK) {
                break;
            }
            entry = entry->next;
        }
    }
    return ret;
}



#ifdef DEBUG
/* This is not used anywhere. It may be called from inside dbx. */
void
PrintLoadedClasses(void)
{
    int         j;
    ClassClass **pcb, *cb;
    BINCLASS_LOCK(sysThreadSelf());
    pcb = binclasses;
    for (j = nbinclasses; --j >= 0; pcb++) {
	cb = *pcb;
	jio_fprintf(stdout, "class=%s, 0x%p\n", cbName(cb), cbLoader(cb));
    }
    BINCLASS_UNLOCK(sysThreadSelf());
}
#endif
