/*
 * @(#)classruntime.c	1.160 98/09/14
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*-
 *      Runtime routines to support classes.
 */

#include <string.h>
#include <sys/types.h>
#include <stdlib.h>

#include "oobj.h"
#include "tree.h"
#include "interpreter.h"
#include "threads.h"
#include "monitor.h"
#include "exceptions.h"
#include "javaString.h"
#include "common_exceptions.h"
#include "sys_api.h"
#include "jlong.h"

#include "java_lang_Throwable.h"
#include "java_lang_Thread.h"

#ifdef BREAKPTS
#include "breakpoints.h"
#endif

#ifdef HPROF
#include "vmprofiler.h"
#endif /* HPROF */

#ifdef DEBUG

/*
 * This is a little debugging utility which can be useful for catch
 * corrupted or invalid arguments being passed from one method to
 * another.  All the invokers are set up to call this.  In the presence
 * of a JIT the only place this would end up being called was during calls
 * which actually go through an invoker, so it won't catch this problem
 * for calls from JIT'ed code to JIT'ed code.
 */
void
invokerValidateArgs(stack_item * optop, struct methodblock * mb)
{
    char * p = methodTerseSig(mb);

    if ((mb->fb.access & ACC_STATIC) == 0) {
	if (optop->h != NULL &&
	    (!isHandle(optop->h) || optop->h->methods == NULL))
	    panic("invalid handle on Java stack");
	optop += 1;
    }

    while (1) {
        switch (*p++) {
	case TERSE_SIG_LONG:
	case TERSE_SIG_DOUBLE:
	    optop += 2;
	    continue;

	case TERSE_SIG_BOOLEAN:
	case TERSE_SIG_SHORT:
	case TERSE_SIG_BYTE:
	case TERSE_SIG_CHAR:
	case TERSE_SIG_INT:
	case TERSE_SIG_FLOAT:
	    optop += 1;
            continue;

	case TERSE_SIG_ENDFUNC:
	    return;

        case TERSE_SIG_OBJECT: {
	    if (optop->h != NULL &&
		(!isHandle(optop->h) || optop->h->methods == NULL))
		panic("invalid handle on Java stack");
	    optop += 1;
	    continue;
	}
	default:
	    sysAssert(FALSE);
	    return;
        }
    }
    
}
#endif /* DEBUG */

/*
 * These are all flavors of method invocation depending on the type
 * of method (defined by its access flags) being invoked.  There
 * are 4 possibilities: java method, synchronized java method,
 * native (C) method, synchronized native method.  Each methodblock
 * has a pointer to the appropriate one for its access, and each
 * one of these procedures optimizes its particular case.
 *
 * They don't all use all the parameters, which is unfortunate.
 */
bool_t
invokeJavaMethod(JHandle *o, struct methodblock *mb, int args_size, ExecEnv *ee)
{
    int        nlocals = mb->nlocals; /* maximum number of locals */
    JavaFrame  *old_frame = ee->current_frame; /* previous frame */
    stack_item *optop = old_frame->optop; /* Points to just before args */
    JavaStack  *stack = old_frame->javastack; /* Current stack pointer */
    JavaFrame  *frame;

    /* Reserve at least two words for the return value from callee. */
    if (nlocals < 2) {
        nlocals = 2;
    }

    frame = (JavaFrame *)(optop + nlocals); /* tentative  */

#ifdef TRACING
    if (trace) {
	char where_am_i[100];
	pc2string(mb->code, old_frame, where_am_i,
		  where_am_i + sizeof where_am_i);
	jio_fprintf(stdout, "Entering %s\n", where_am_i);
	fflush(stdout);
    }
#endif /* TRACING */

    if (frame->ostack + mb->maxstack >= stack->end_data) {
      /* Without tmp_XXX variables, the C compiler gets confused
       * about pointer aliasing, and generates much worse code.
       */
        JavaStack *tmp_stack = stack;
	JavaFrame *tmp_frame = frame;
	stack_item *tmp_optop = optop;
	if (!ExpandJavaStack(ee, &tmp_stack, &tmp_frame, &tmp_optop, 
			     args_size, nlocals, mb->maxstack))
	    return FALSE;
	stack = tmp_stack;
	frame = tmp_frame;
	optop = tmp_optop;
    }

    frame->javastack = stack;
    frame->prev = old_frame;
    frame->vars = optop;
    frame->optop = frame->ostack;
    frame->current_method = mb; 
    frame->constant_pool = cbConstantPool(fieldclass(&mb->fb));
    frame->returnpc = mb->code;
    frame->monitor = 0;
#ifdef BREAKPTS
    frame->lastpc = 0;
#endif
    ee->current_frame = frame;

#ifdef HPROF
    if (jvmpi_event_flags & JVMPI_EVENT_METHOD_ENTRY_ON) {
        jvmpi_method_entry(ee, o);
    }
#endif /* HPROF */
    
#ifdef BREAKPTS
    if (debugging) {
        notify_debugger_of_frame_push(ee);
    }
#endif /* BREAKPTS */
    
#ifdef DEBUG
    invokerValidateArgs(optop, mb);
#endif /* DEBUG */
    return TRUE;
}


bool_t
invokeSynchronizedJavaMethod(JHandle *o, struct methodblock *mb, int args_size,
			    ExecEnv *ee)
{
    int        nlocals = mb->nlocals; /* maximum number of locals */
    JavaFrame  *old_frame = ee->current_frame; /* previous frame */
    stack_item *optop = old_frame->optop; /* Points to just before args */
    JavaStack  *stack = old_frame->javastack; /* Current stack pointer */
    JavaFrame  *frame;

    /* Reserve at least two words for the return value from callee. */
    if (nlocals < 2) {
        nlocals = 2;
    }

    frame = (JavaFrame *)(optop + nlocals); /* tentative  */

#ifdef TRACING
    if (trace) {
	char where_am_i[100];
	pc2string(mb->code, old_frame, where_am_i,
		  where_am_i + sizeof where_am_i);
	jio_fprintf(stdout, "Entering %s\n", where_am_i);
	fflush(stdout);
    }
#endif /* TRACING */

    if (frame->ostack + mb->maxstack >= stack->end_data) {
      /* Without tmp_XXX variables, the C compiler gets confused
       * about pointer aliasing, and generates much worse code.
       */
        JavaStack *tmp_stack = stack;
	JavaFrame *tmp_frame = frame;
	stack_item *tmp_optop = optop;
	if (!ExpandJavaStack(ee, &tmp_stack, &tmp_frame, &tmp_optop, 
			     args_size, nlocals, mb->maxstack))
	    return FALSE;
	stack = tmp_stack;
	frame = tmp_frame;
	optop = tmp_optop;
    }

    frame->javastack = stack;
    frame->prev = old_frame;
    frame->vars = optop;
    frame->optop = frame->ostack;
    frame->current_method = mb; 
    frame->constant_pool = cbConstantPool(fieldclass(&mb->fb));
    frame->returnpc = mb->code;

#ifndef FAST_MONITOR
    frame->monitor = monitorEnter2(ee, obj_monitor(o));
#else
    frame->monitor = (struct sys_mon *) o;
    monitorEnter2(ee, obj_monitor(o));
#endif
#ifdef DEBUG
    invokerValidateArgs(optop, mb);
#endif /* DEBUG */

#ifdef BREAKPTS
    frame->lastpc = 0;
#endif
    ee->current_frame = frame;

#ifdef HPROF
    if (jvmpi_event_flags & JVMPI_EVENT_METHOD_ENTRY_ON) {
        jvmpi_method_entry(ee, o);
    }
#endif /* HPROF */

#ifdef BREAKPTS
    if (debugging) {
        notify_debugger_of_frame_push(ee);
    }
#endif /* BREAKPTS */

#ifdef DEBUG
    invokerValidateArgs(optop, mb);
#endif /* DEBUG */
    return TRUE;
}

bool_t
invokeNativeMethod(JHandle *o, struct methodblock *mb, int args_size, 
		   ExecEnv *ee)
{
  /* We need to set up a frame even for normal native methods so that
   * we can use JNI functions in normal native methods.
   */
    JavaFrame *old_frame = ee->current_frame;
    stack_item *optop = old_frame->optop;
    JavaStack  *stack = old_frame->javastack; 
    JavaFrame  *frame = (JavaFrame *)(optop + args_size);

    if (frame->ostack + JNI_REF_INFO_SIZE + JNI_DEFAULT_LOCAL_CAPACITY >= 
	stack->end_data) {
      /* Without tmp_XXX variables, the C compiler gets confused
       * about pointer aliasing, and generates much worse code.
       */
        JavaStack *tmp_stack = stack;
	JavaFrame *tmp_frame = frame;
	if (!ExpandJavaStackForJNI(ee, &tmp_stack, &tmp_frame, 
			   JNI_REF_INFO_SIZE + JNI_DEFAULT_LOCAL_CAPACITY))
	    return FALSE;
	stack = tmp_stack;
	frame = tmp_frame;
    }

    frame->javastack = stack;
    frame->prev = old_frame;
    frame->optop = frame->ostack + JNI_REF_INFO_SIZE;
    frame->current_method = mb; 

    JNI_REFS_FREELIST(frame) = NULL;
    JNI_N_REFS_IN_USE(frame) = 0;
    JNI_REFS_CAPACITY(frame) = JNI_DEFAULT_LOCAL_CAPACITY;

    ee->current_frame = frame;
    
#ifdef DEBUG
    invokerValidateArgs(optop, mb);
#endif /* DEBUG */

#ifdef HPROF
    if (jvmpi_event_flags & JVMPI_EVENT_METHOD_ENTRY_ON) {
        jvmpi_method_entry(ee, o);
    }
#endif /* HPROF */

#ifdef BREAKPTS
    if (debugging) {
        notify_debugger_of_frame_push(ee);
    }
#endif /* BREAKPTS */
    
    optop = (*(stack_item * (*) (stack_item*, ExecEnv*)) mb->code)(optop, ee);

#ifdef HPROF
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_METHOD_EXIT)) {
        jvmpi_method_exit(ee);
    }
#endif /* HPROF */

#ifdef BREAKPTS
        if (debugging) {
            notify_debugger_of_frame_pop(ee);
        }
#endif /* BREAKPTS */

    old_frame->optop = optop;
    ee->current_frame = old_frame;

    TRACE_METHOD(ee, mb, args_size, TRACE_METHOD_NATIVE_RETURN);
    return !(exceptionOccurred(ee));
}

bool_t
invokeSynchronizedNativeMethod(JHandle *o, struct methodblock *mb, 
			       int args_size, ExecEnv *ee)
{
  /* We need to set up a frame even for normal native methods so that
   * we can use JNI functions in normal native methods.
   */
    JavaFrame *old_frame = ee->current_frame;
    stack_item *optop = old_frame->optop;
    JavaStack  *stack = old_frame->javastack; 
    JavaFrame  *frame = (JavaFrame *)(optop + args_size);
    struct sys_mon * monitor;

    if (frame->ostack + JNI_REF_INFO_SIZE + JNI_DEFAULT_LOCAL_CAPACITY >=
	stack->end_data) {
      /* Without tmp_XXX variables, the C compiler gets confused
       * about pointer aliasing, and generates much worse code.
       */
        JavaStack *tmp_stack = stack;
	JavaFrame *tmp_frame = frame;
	if (!ExpandJavaStackForJNI(ee, &tmp_stack, &tmp_frame,
			   JNI_REF_INFO_SIZE + JNI_DEFAULT_LOCAL_CAPACITY))
	    return FALSE;
	stack = tmp_stack;
	frame = tmp_frame;
    }

    frame->javastack = stack;
    frame->prev = old_frame;
    frame->optop = frame->ostack + JNI_REF_INFO_SIZE;
    frame->current_method = mb; 

    JNI_REFS_FREELIST(frame) = NULL;
    JNI_N_REFS_IN_USE(frame) = 0;
    JNI_REFS_CAPACITY(frame) = JNI_DEFAULT_LOCAL_CAPACITY;

    ee->current_frame = frame;

#ifndef FAST_MONITOR
    monitor = monitorEnter2(ee, obj_monitor(o));
#else /*FAST_MONITOR*/
    monitorEnter2(ee, obj_monitor(o));
#endif/*FAST_MONITOR*/

#ifdef DEBUG
    invokerValidateArgs(optop, mb);
#endif /* DEBUG */

#ifdef HPROF
    if (jvmpi_event_flags & JVMPI_EVENT_METHOD_ENTRY_ON) {
        jvmpi_method_entry(ee, o);
    }
#endif /* HPROF */
    
#ifdef BREAKPTS
    if (debugging) {
        notify_debugger_of_frame_push(ee);
    }
#endif /* BREAKPTS */
    
    optop = (*(stack_item * (*) (stack_item*, ExecEnv*)) mb->code)(optop, ee);

#ifdef HPROF    
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_METHOD_EXIT)) {
        jvmpi_method_exit(ee);
    }
#endif /* HPROF */

#ifdef BREAKPTS
        if (debugging) {
            notify_debugger_of_frame_pop(ee);
        }
#endif /* BREAKPTS */

#ifndef FAST_MONITOR
    sysMonitorExit(EE2SysThread(ee), monitor);
#else /*FAST_MONITOR*/
    monitorExit2(ee, obj_monitor(o));
#endif/*FAST_MONITOR*/

    

    old_frame->optop = optop;
    ee->current_frame = old_frame;

    TRACE_METHOD(ee, mb, args_size, TRACE_METHOD_NATIVE_RETURN);
    return !(exceptionOccurred(ee));
}

/* 
 * The invoke***JNINativeMethod functions are similar to
 * invoke***NativeMethod functions. The differences are here we call
 * sysInvokeNative to transfer the arguments onto the Java stack.
 * In addition, arguments that are Java objects or arrays are added
 * to the local registry.
 */

bool_t
invokeJNINativeMethod(JHandle *o, struct methodblock *mb, 
		      int args_size, ExecEnv *ee)
{
    JavaFrame  *old_frame = ee->current_frame; /* previous frame */
    stack_item *optop = old_frame->optop; /* Points to just before args */
    JavaStack  *stack = old_frame->javastack; 
    JavaFrame  *frame = (JavaFrame *)(optop + args_size);
    ClassClass **isStatic = 0;
      
    if (frame->ostack + JNI_REF_INFO_SIZE + JNI_DEFAULT_LOCAL_CAPACITY >= 
	stack->end_data) {
      /* Without tmp_XXX variables, the C compiler gets confused
       * about pointer aliasing, and generates much worse code.
       */
        JavaStack *tmp_stack = stack;
	JavaFrame *tmp_frame = frame;
	if (!ExpandJavaStackForJNI(ee, &tmp_stack, &tmp_frame,
			   JNI_REF_INFO_SIZE + JNI_DEFAULT_LOCAL_CAPACITY))
	    return FALSE;
	stack = tmp_stack;
	frame = tmp_frame;
    }

    frame->javastack = stack;
    frame->prev = old_frame;
    frame->optop = frame->ostack + JNI_REF_INFO_SIZE;
    frame->current_method = mb;

    JNI_REFS_FREELIST(frame) = NULL;
    JNI_N_REFS_IN_USE(frame) = 0;
    JNI_REFS_CAPACITY(frame) = JNI_DEFAULT_LOCAL_CAPACITY;

    ee->current_frame = frame;

    if (mb->fb.access & ACC_STATIC)
        isStatic = (ClassClass**)&(o);
    
#ifdef DEBUG
    invokerValidateArgs(optop, mb);
#endif /* DEBUG */

#ifdef HPROF
    if (jvmpi_event_flags & JVMPI_EVENT_METHOD_ENTRY_ON) {
        jvmpi_method_entry(ee, o);
    }
#endif /* HPROF */

#ifdef BREAKPTS
    if (debugging) {
        notify_debugger_of_frame_push(ee);
    }
#endif /* BREAKPTS */

    optop = (stack_item *)sysInvokeNative((void *)EE2JNIEnv(ee),
					  (void*)mb->code,
					  (long*)optop,
					  methodTerseSig(mb),
					  mb->args_size,
					  isStatic);

#ifdef HPROF    
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_METHOD_EXIT)) {
        jvmpi_method_exit(ee);
    }
#endif /* HPROF */

#ifdef BREAKPTS
        if (debugging) {
            notify_debugger_of_frame_pop(ee);
        }
#endif /* BREAKPTS */
    
    CHECK_CRITICAL_COUNT_ON_RETURN(ee);

    old_frame->optop = optop;
    ee->current_frame = old_frame;

    TRACE_METHOD(ee, mb, args_size, TRACE_METHOD_NATIVE_RETURN);
    return !(exceptionOccurred(ee));
}

bool_t
invokeJNISynchronizedNativeMethod(JHandle *o, struct methodblock *mb, 
				  int args_size, ExecEnv *ee)
{
    JavaFrame  *old_frame = ee->current_frame; /* previous frame */
    stack_item *optop = old_frame->optop; /* Points to just before args */
    JavaStack  *stack = old_frame->javastack; 
    JavaFrame  *frame = (JavaFrame *)(optop + args_size);
    ClassClass **isStatic = 0;
    struct sys_mon *monitor;
 
    if (frame->ostack + JNI_REF_INFO_SIZE + JNI_DEFAULT_LOCAL_CAPACITY >= 
	stack->end_data) {
      /* Without tmp_XXX variables, the C compiler gets confused
       * about pointer aliasing, and generates much worse code.
       */
        JavaStack *tmp_stack = stack;
	JavaFrame *tmp_frame = frame;
	if (!ExpandJavaStackForJNI(ee, &tmp_stack, &tmp_frame,
			   JNI_REF_INFO_SIZE + JNI_DEFAULT_LOCAL_CAPACITY))
	    return FALSE;
	stack = tmp_stack;
	frame = tmp_frame;
    }

    frame->javastack = stack;
    frame->prev = old_frame;
    frame->optop = frame->ostack + JNI_REF_INFO_SIZE;
    frame->current_method = mb; 

    JNI_REFS_FREELIST(frame) = NULL;
    JNI_N_REFS_IN_USE(frame) = 0;
    JNI_REFS_CAPACITY(frame) = JNI_DEFAULT_LOCAL_CAPACITY;

    ee->current_frame = frame;

    if (mb->fb.access & ACC_STATIC)
        isStatic = (ClassClass**)(&o);

#ifndef FAST_MONITOR
    monitor = monitorEnter2(ee, obj_monitor(o));
#else /*FAST_MONITOR*/
    monitorEnter2(ee, obj_monitor(o));
#endif/*FAST_MONITOR*/

#ifdef DEBUG
    invokerValidateArgs(optop, mb);
#endif /* DEBUG */

#ifdef HPROF
    if (jvmpi_event_flags & JVMPI_EVENT_METHOD_ENTRY_ON) {
        jvmpi_method_entry(ee, o);
    }
#endif /* HPROF */

#ifdef BREAKPTS
    if (debugging) {
        notify_debugger_of_frame_push(ee);
    }
#endif /* BREAKPTS */

    optop = (stack_item *)sysInvokeNative((void *)EE2JNIEnv(ee),
					  (void*)mb->code,
					  (long*)optop,
					  methodTerseSig(mb),
					  mb->args_size,
					  isStatic);

#ifdef HPROF
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_METHOD_EXIT)) {
        jvmpi_method_exit(ee);
    }
#endif /* HPROF */

#ifdef BREAKPTS
        if (debugging) {
            notify_debugger_of_frame_pop(ee);
        }
#endif /* BREAKPTS */

#ifndef FAST_MONITOR
    sysMonitorExit(EE2SysThread(ee), monitor);
#else /*FAST_MONITOR*/
    monitorExit2(ee, obj_monitor(o));
#endif/*FAST_MONITOR*/

    CHECK_CRITICAL_COUNT_ON_RETURN(ee);

    old_frame->optop = optop;
    ee->current_frame = old_frame;

    TRACE_METHOD(ee, mb, args_size, TRACE_METHOD_NATIVE_RETURN);
    return !(exceptionOccurred(ee));
}

/* defined in invokers.c, which is automatically generated by
 * InvokerGen.java.
 */
extern Invoker
getCustomInvoker(char *sig);
    
bool_t
invokeLazyNativeMethod(JHandle *o, struct methodblock *mb, int args_size, 
		       ExecEnv *ee)
{
    void *code;

    /* We lock the class before getting mb->code to make sure that
     * mb->code is consistent with mb->invoker. We don't rely on the
     * order of setting these two fields or that the CPU obeys TSO 
     * etc.
     */
    LINKCLASS_LOCK(EE2SysThread(ee));
    code = mb->code;
    LINKCLASS_UNLOCK(EE2SysThread(ee));

    if (code == NULL) {
        uint32_t lib_kind;
        if (verbose_jni) {
	    jio_fprintf(stderr, "[Dynamic-linking native method %s.%s ... ",
			cbName(mb->fb.clazz), mb->fb.name);
	}
    	code = dynoLink(mb, &lib_kind);
	if (code == NULL) {
	    if (verbose_jni) 
	        jio_fprintf(stderr, "not found]\n");
	} else {
	    LINKCLASS_LOCK(EE2SysThread(ee));
	    /* Check whether another thread already linked the native
	     * library while we are perofrming the linking.
	     */
	    if (mb->code == NULL) {
	        mb->code = code;
	        if (lib_kind != 0) {
		    if (verbose_jni) 
		        jio_fprintf(stderr, "JNI]\n");
		    /* lookup a custom invoker for this signature.  We only
		     * do this for non-synchronized JNI methods.  The 
		     * reasoning is that the cost of entering a monitor 
		     * diminishes the performance win of a custom invoker.
		     * This also simplifies and speeds up the custom invoker
		     * code.
		     */
		    if (mb->fb.access & ACC_SYNCHRONIZED) {
			mb->invoker = invokeJNISynchronizedNativeMethod;
		    } else {
		        Invoker inv = getCustomInvoker(methodTerseSig(mb));
			if (inv)
			    mb->invoker = inv;
			else
			    mb->invoker = invokeJNINativeMethod;
		    }
		} else {
		    if (verbose_jni) 
		        jio_fprintf(stderr, "old style]\n");
		    mb->invoker = (mb->fb.access & ACC_SYNCHRONIZED) 
		        ? invokeSynchronizedNativeMethod : invokeNativeMethod;
		}
	    }
	    LINKCLASS_UNLOCK(EE2SysThread(ee));
	}
    }
    if (mb->code) {
	return (mb->invoker)(o, mb, args_size, ee);
    } else {
	ThrowUnsatisfiedLinkError(ee, fieldname(&mb->fb));
	return FALSE;
    }
}

bool_t
invokeAbstractMethod(JHandle *o, struct methodblock *mb, int args_size, 
		     ExecEnv *ee)
{
    char buf[STK_BUF_LEN];
    /* We created fake method names in ResolveInterfaces() function
     * in classresolver.c.
     */
    if (mb->fb.name[0] == '+') {
        jio_snprintf(buf, sizeof(buf), 
		     "access non-public method %s.%s through an interface",
		     cbName(mb->fb.clazz),
		     mb->fb.name + 1);
	ThrowIllegalAccessError(ee, buf);
    } else {
        jio_snprintf(buf, sizeof(buf), 
		     "%s.%s",
		     cbName(mb->fb.clazz),
		     mb->fb.name);
	ThrowAbstractMethodError(ee, buf);
    }
    return FALSE;
}

unsigned char *
classLoaderLink(struct Hjava_lang_ClassLoader *loader, char *name,
		int args_size)
{
    ExecEnv *ee = EE();
    unsigned char *result = 0;
    JNIEnv *env = EE2JNIEnv(ee);
    jmethodID findNativeID = NULL;
    JavaFrame *current_frame = ee->current_frame;
    
    /*
     * adjust optop to make sure that the new local frame won't end up
     * on top of a previous frame.
     */
    current_frame->optop += args_size;

    if ((*env)->PushLocalFrame(env, 10) < 0) {
	/* remove optop offset we added */
	current_frame->optop -= args_size;
        return 0;
    }

    if (loader == 0) {
        if ((result = (unsigned char *)FindBuiltinEntry(name)))
	    goto done;
    }
        
    if (findNativeID == NULL) {
        findNativeID = (*env)->GetStaticMethodID(env,
			      MkRefLocal(env, classJavaLangClassLoader),
			      "findNative",
			      "(Ljava/lang/ClassLoader;Ljava/lang/String;)J");
	if (findNativeID == NULL) {
	    goto done;
	}
    }

    {
        jstring jname = (*env)->NewStringUTF(env, name);
	if (jname) {
	    result = jlong_to_ptr((*env)->CallStaticLongMethod(env, 
				    MkRefLocal(env, classJavaLangClassLoader),
				    findNativeID,
				    MkRefLocal(env, 
					       loader),
				    jname));
	    if ((*env)->ExceptionOccurred(env)) {
	        result = 0; 
	    }
	}
    }

 done:
    (*env)->PopLocalFrame(env, 0);

    /* remove optop offset we added */
    current_frame->optop -= args_size;
    return result;
}

/* Returns the native code pointer. Set *lib_info to 0 for non-JNI
 * native libraries, to >0 for JNI native libraries.
 */
void *
dynoLink(struct methodblock *mb, uint32_t *lib_info)
{
    int i;
    char sym_buf[STK_BUF_LEN];
    char *sym;
    ClassClass *cb = fieldclass(&mb->fb);
    void *result;
    int native_args_size = 
        (mb->fb.access & ACC_STATIC) ? mb->args_size + 2 : mb->args_size + 1;
    int len = maxMangledMethodNameLength(mb) + 10; /* Add 10 for trailing 0
						      and sysBuildFunName */
    
    if (len > STK_BUF_LEN) {
        sym = sysMalloc(len);
	if (sym == NULL) {
	    ThrowOutOfMemoryError(0, 0);
	    return NULL;
	}
    } else {
        sym = sym_buf;
    }
    
    for (i=0; ;i++) {
        mangleMethodName(mb, sym, len, MangleMethodName_JNI_SHORT);
        if (!sysBuildFunName(sym, len, native_args_size, i))
	    break;
        result = classLoaderLink(cbLoader(cb), sym, mb->args_size);
	if (result != 0) {
	    *lib_info = 0x00010001;
	    goto done;
	}
    }

    for (i=0; ;i++) {
        mangleMethodName(mb, sym, len, MangleMethodName_JNI_LONG);
	if (!sysBuildFunName(sym, len, native_args_size, i))
	    break;
        result = classLoaderLink(cbLoader(cb), sym, mb->args_size);
	if (result != 0) {
	    *lib_info = 0x00010001;
	    goto done;
	}
    }

    mangleMethodName(mb, sym, len, MangleMethodName_JDK_1);
    result = classLoaderLink(cbLoader(cb), sym, mb->args_size);
    *lib_info = 0;

 done:
    if (sym != sym_buf) {
        sysFree(sym);
    }
    return result;
}

typedef jint (JNICALL *JNI_OnLoad_t)(JavaVM *, void *);

bool_t
RunOnLoadHook(void *h)
{
    JNIEnv *env = EE2JNIEnv(EE());
    JNI_OnLoad_t JNI_OnLoad;
    const char *onLoadSymbols[] = JNI_ONLOAD_SYMBOLS;
    int i;

    for (i = 0; i < sizeof(onLoadSymbols) / sizeof(char *); i++) {
        JNI_OnLoad = (JNI_OnLoad_t)sysFindLibraryEntry(h, onLoadSymbols[i]);
	if (JNI_OnLoad) {
	    break;
	}
    }
    if (JNI_OnLoad) {
        jint ver;
	JavaVM *jvm;
	(*env)->GetJavaVM(env, &jvm);
	ver = (*JNI_OnLoad)(jvm, NULL);
	if (!JVM_IsSupportedJNIVersion(ver)) {
	    jio_fprintf(stderr, "unsupported JNI version: %lx", ver);
	    return FALSE;
	}
	return TRUE;
    } else {
        jio_fprintf(stderr, "corrupted library\n");
        return FALSE;
    }
}

static void *java_handle = 0;

bool_t 
LoadJavaLibrary(void)
{
    char buf[MAXPATHLEN];
    char load_error[STK_BUF_LEN*2];
    
    sysBuildLibName(buf, sizeof(buf), java_dll_dir, "java");
    java_handle = sysLoadLibrary(buf, load_error, sizeof(load_error));
    if (java_handle == 0) {
        jio_fprintf(stderr, "Can't load library \"%s\", because %s\n",
		    buf, load_error);
	return FALSE;
    }
    return RunOnLoadHook(java_handle);
}

void * FindBuiltinEntry(const char *name)
{
    return sysFindLibraryEntry(java_handle, name);
}


void
PrepareInvoker(struct methodblock *mb)
{
    unsigned int access = mb->fb.access;
    if (access & ACC_MACHINE_COMPILED) { 
        mb->invoker = invokeCompiledMethod;
    } else if ((access & (ACC_ABSTRACT | ACC_NATIVE)) == 0) {
        /* Normal java method */
        mb->invoker = (access & ACC_SYNCHRONIZED) 
	  ? invokeSynchronizedJavaMethod
	  : invokeJavaMethod;
    } else if (access & ACC_ABSTRACT) {
      /*        cbAccess(cb) |= ACC_ABSTRACT;*/
	mb->invoker = invokeAbstractMethod;
    } else { 
        mb->invoker = invokeLazyNativeMethod;
    }
}

static void
RunStaticInitializers(ClassClass *cb)
{
    struct methodblock *mb;
    int size;
    struct methodblock *initializer = NULL;
    struct execenv *ee = EE();

    mb = cbMethods(cb);
    for (size = 0; size < (int) cbMethodsCount(cb); size++, mb++) {
	sysAssert(mb->fb.u.offset < cbMethodTableSize(cb) || 
		  cbIsInterface(cb));
	if (mb->fb.name == utf8_literal_cls_init_name &&
	    mb->fb.signature == utf8_literal_null_init_sig) {
	    initializer = mb;
	    break;				/* only one initializer */
	}
    }
    
    /* The exact initialization procedure as documented in 2.16.5 of 
       the VM Spec. */

    monitorEnter2(ee, obj_monitor(cb));                        /* Step 1 */
    {
        void *init_thread = cbInitThread(cb);

	while (init_thread && init_thread != (void *)ee) {/* Step 2 */
	    monitorWait2(ee, obj_monitor(cb), TIMEOUT_INFINITY);
	    init_thread = cbInitThread(cb);
	}
	if ((init_thread && init_thread == (void *)ee) || /* Step 3 */
	    CCIs(cb, Initialized)) {                      /* Step 4 */
	    monitorExit2(ee, obj_monitor(cb));
	    return;
	}
	if (CCIs(cb, Error)) {                            /* Step 5 */
	    ThrowNoClassDefFoundError(ee, cbName(cb));
	    monitorExit2(ee, obj_monitor(cb));
	    return;
	}
	cbInitThread(cb) = ee;                            /* Step 6 */
    }
    monitorExit2(ee, obj_monitor(cb));

    InitializeForCompiler(cb);

    if (!cbIsInterface(cb)) {                             /* Step 7 */
        if (cbSuperclass(cb))
	    InitClass(cbSuperclass(cb));
    }

    if (ee && exceptionOccurred(ee)) {
        monitorEnter2(ee, obj_monitor(cb));
	cbInitThread(cb) = 0;
	CCSet(cb,Error);
	monitorNotifyAll2(ee, obj_monitor(cb));
	monitorExit2(ee, obj_monitor(cb));
	return;
    }

#ifndef TRIMMED
    if (verbose)
	jio_fprintf(stderr, "[Initializing %s]\n", cbName(cb));
#endif

                                                          /* Step 8 */
    if (initializer != NULL) {
#ifndef TRIMMED
        if (verbose)
	    jio_fprintf(stderr, "[Running static initializer for %s]\n", 
		    cbName(cb));
#endif
	sysAssert(initializer->code != NULL);
	do_execute_java_method(ee, cb, 0, 0, initializer, TRUE);
#ifdef BREAKPTS
        if (!debugging) {
#endif 
            /* Free the code for the initializer */
            sysFree(cbClinitSpace(cb));
            cbClinitSpace(cb) = NULL;
            /* Clean out other pointers for GC */
            initializer->code = NULL;
#ifdef BREAKPTS
        }
#endif 
	/* This isn't necessary.  But it makes sure mb->code 
	   isn't called */
	initializer->invoker = invokeAbstractMethod;
    }

    /* Now handle any exception */
    if (initializer == NULL || !ee || 
	!exceptionOccurred(ee)) {                         /* Step 9 */
        monitorEnter2(ee, obj_monitor(cb));
	cbInitThread(cb) = 0;
	CCSet(cb,Initialized);
	monitorNotifyAll2(ee, obj_monitor(cb));
	monitorExit2(ee, obj_monitor(cb));
    } else {                                              /* Step 10 */
        JHandle *oldError = ee->exception.exc;
	if (!is_instance_of(oldError, classJavaLangError, ee)) {
	    JHandle *newError;
	    exceptionClear(ee);
	    newError = execute_java_constructor(ee, 
					JAVAPKG "ExceptionInInitializerError",
						NULL, 
						"(L" JAVAPKG "Throwable;)",
						oldError);
	    if (!exceptionOccurred(ee)) { 
	        exceptionThrow(ee, newError);
	    } else { 
		/*EMPTY*/
	        /* don't do anything. Some serious error must have occurred
		 * when constructing the ExceptionInInitializerError 
		 * object.  Let that error stand.
		 */
	    }
	}
	monitorEnter2(ee, obj_monitor(cb));                    /* Step 11 */
	cbInitThread(cb) = 0;
	CCSet(cb,Error);
	monitorNotifyAll2(ee, obj_monitor(cb));
	monitorExit2(ee, obj_monitor(cb));
    }
}

/* Perform the class initialization as described in the VM spec (2.16.4).
 * Link the class if necessary.
 */
void
InitClass(ClassClass * cb)
{
    ExecEnv *ee;

    if (CCIs(cb,Initialized))
        return;

    LinkClass(cb);
    ee = EE();
    if (exceptionOccurred(ee))
        return;

    if ((cbName(cb)[0] != SIGNATURE_ARRAY) && !cbIsPrimitive(cb)) {
	/* Don't need to initialize or verify array or primitive classes */
        if (DisableAsyncEvents(ee)) {
	    /* async. exception pending */
	    return;
	}
	RunStaticInitializers(cb);
	EnableAsyncEvents(ee);
/*
 * This is commented out according to the updated VM spec.
    } else if (cbName(cb)[0] == SIGNATURE_ARRAY) {
	ClassClass *inner_cb = 
	    cbConstantPool(cb)[CONSTANT_POOL_ARRAY_CLASS_INDEX].clazz;
	if (inner_cb) {
	    InitClass(inner_cb);
	    if (exceptionOccurred(ee)) { 
	    } else {
	        CCSet(cb, Initialized);
	    }
	} else {
	    CCSet(cb, Initialized);
	}
        CCSet(cb, Initialized);
*/
    } else {
        CCSet(cb, Initialized);
    }
}

void
SignalError(struct execenv * ee, char *ename, char *DetailMessage)
{
    ClassClass          *cb;
    Hjava_lang_Throwable *ret;
    Classjava_lang_Throwable *new_obj;
    JavaFrame            *frame;
#ifndef TRIMMED
    char where[100];
#endif

    ee = ee ? ee :EE();

    if (0 && verbose) {
        jio_fprintf(stderr, "[Signaling in VM: %s, message: %s]\n", ename,
		    DetailMessage ? DetailMessage : "");

	printStackTrace(ee, 100, NULL);
    }

    if (exceptionOccurred(ee)) {
        if (verbose) {
	    jio_fprintf(stderr, "[Signaling Error: error already pending]\n");
	}
	return;
    }

    /*
     * Find the exception class (or the default).  If we can't find it
     * throw a preallocated object
     */
    if ((cb = FindClassFromClass(ee, ename, TRUE, NULL)) == 0) {
#ifndef TRIMMED
	if (ee->current_frame) {
	    pc2string(ee->current_frame->lastpc, 
		      ee->current_frame, where, where + sizeof where);
	    jio_fprintf(stderr, "Class missing for error: %s at %s\n",
			ename, where);
	} else {
	    jio_fprintf(stderr, "Class missing for error: %s\n", ename);
	}
#endif
	if (!exceptionOccurred(ee)) {
	    JHandle *exception = exceptionInternalObject(IEXC_NoClassDefinitionFound);
	    exceptionThrow(ee, exception);
        }
	return;
    }

    /*
     * Instantiate the object.  If we can't then throw a preallocated
     * object.
     */
    if ((ret = (Hjava_lang_Throwable *)allocObject(ee, cb)) == NULL) { 
        JHandle *exception = exceptionInternalObject(IEXC_OutOfMemory);
	exceptionThrow(ee, exception);
	return;
    }
    new_obj = unhand(ret);

    /* Backtrace through the interpreter frames. */
    frame = ee ? ee->current_frame : 0;
    if (frame == 0) {
#ifndef TRIMMED
	jio_fprintf(stderr, "Exception: %s (can't backtrace because of a missing "
		"context)\n", ename);
#else
	/*EMPTY*/
#endif
	/*	return;	*/
    }
#ifdef TRACING
    if (trace && frame) {
	jio_fprintf(stdout, "%6lX %8lX\tERROR %s\n", (long)threadSelf(), (long)frame->lastpc, ename);
	fflush(stdout);
    }
#endif /* TRACING */

    if (DetailMessage)
	new_obj->detailMessage = makeJavaStringUTF(DetailMessage);
    fillInStackTrace(ret, ee);

    /* This must be the last thing.
     */
    exceptionThrow(ee, (HObject *)ret);
}

static char *
addstr(char *s, char *buf, char *limit, char term)
{
    char c;
    while ((c = *s) && c != term && buf < limit) {
	*buf++ = c;
	s++;
    }
    return buf;
}


static char       *
adddec(long n, char *buf, char *limit)
{
    if (n < 0) {
	n = -n;
	if (buf < limit)
	    *buf++ = '-';
    }
    if (n >= 10)
	buf = adddec(n / 10, buf, limit);
    if (buf < limit)
	*buf++ = (char)((n % 10) + '0');
    return buf;
}


int
pc2lineno(struct methodblock *mb, unsigned short pc_offset)
{
    int length = mb->line_number_table_length;

    if (length > 0) {
	struct lineno *ln = mb->line_number_table;
	int l = 0;
	int u = length;
	/* Check if pc_offset < ln[l].pc */
	if (pc_offset < ln[l].pc) { 
	    return -1;
	}
	/* Check if pc_offset >= ln[u - 1].pc */
	if (pc_offset >= ln[u - 1].pc) {
	    return ln[u - 1].line_number;
	}
	/* Invariant:  ln[l].pc <= pc_offset < ln[u - 1].pc */ 
	while (l < u) {
	    int m = (l + u) >> 1;
	    if (pc_offset < ln[m].pc) {
	        u = m;
	    } else if (pc_offset >= ln[m + 1].pc) {
	        l = m;
	    } else {
	        /* ln[m].pc <= pc_offset < ln[m + 1].pc */ 
	        return ln[m].line_number;
	    }
	}
	/* l == u, pc is not found in any of the intervals.
	 * Invariant is broken.
	 */
	sysAssert(FALSE);
    }
    return -1;
}

static struct methodblock *pc2method(unsigned char *pc) {
    struct methodblock *mb;
    ClassClass *cb;
    int i, j;

    BINCLASS_LOCK(sysThreadSelf());
    for (i = nbinclasses; --i >= 0; ) {
	cb = binclasses[i];
	for (mb = cbMethods(cb), j = cbMethodsCount(cb); --j >= 0; mb++) {
	    unsigned char *code_start = mb->code;
	    if (mb->fb.access & ACC_NATIVE) {
	        if (mb->code == pc)
		    goto unlock;
	    } else if (pc >= code_start && pc < code_start + mb->code_length) {
		    goto unlock;
	    } else if (mb->fb.access & ACC_MACHINE_COMPILED) {
	        if (PCinCompiledCode(pc, mb))
		    goto unlock;
	    }
	}
    }
    mb = 0;

unlock:
    BINCLASS_UNLOCK(sysThreadSelf());
    return mb;
}

char *
pc2string(unsigned char *pc, JavaFrame * frame, char *buf, char *limit)
{
    struct methodblock *mb = NULL;

    if (buf >= limit) {
	return buf;
    }
    --limit;	/* Save room for terminating '\0' */
    if (frame)
	mb = frame->current_method;
    if (mb == 0) {
        mb = pc2method(pc);
    }
    if (mb != 0) {
	char cbuf[MSG_BUF_LEN];
	ClassClass *cb = fieldclass(&mb->fb);
	buf = addstr(classname2string(cbName(cb),
		     cbuf, sizeof(cbuf)), buf, limit, 0);
	buf = addstr(".", buf, limit, 0);
	buf = addstr(fieldname(&mb->fb), buf, limit, '(');
	if (cb != 0) {
	    if (mb->fb.access & ACC_NATIVE) {
	        buf = addstr("(Native Method)", buf, limit, 0);
	    } else if (mb->fb.access & ACC_MACHINE_COMPILED &&
		(frame == NULL || frame->constant_pool == NULL)) {
		buf = addstr("(Compiled Code)", buf, limit, 0);
	    } else {
		char        *fn;
		fn = cbSourceName(cb);
		if (fn) {
		    int         lineno;
		    /* Just want the short name */
		    fn = strrchr(fn, '/');
		    if (fn) {
			/* skip past last separator */
			fn += 1;
		    } else {
			fn = cbSourceName(cb);
		    }
		    buf = addstr("(", buf, limit, 0);
		    buf = addstr(fn, buf, limit, 0);
		    lineno = pc2lineno(mb, (unsigned short)(pc - mb->code));
		    if (lineno >= 0) {
			buf = addstr(":", buf, limit, 0);
			buf = adddec(lineno, buf, limit);
		    }
		    buf = addstr(")", buf, limit, 0);
		} else {
		    buf = addstr("(Unknown Source)", buf, limit, 0);
		}
	    }
	}
	*buf = 0;
    } else {
	buf[0] = '\0';
    }
    return buf;
}

ExecEnv *
EE(void)
{
    ExecEnv *ee;
    sys_thread_t *self = sysThreadSelf();

    if (self == 0) {
	ee = 0;
    } else {
	ee = SysThread2EE(self);
    }

    return ee;
}

void printStackTrace(ExecEnv *ee, int depth_limit, void (*pf)(char *, ...))
{
    JavaFrame *frame, frame_buf;

    /* Print the stack back trace. 
     * This code is adapted from fillInStackTrace and printStackTrace.
     */
    frame = ee->current_frame;
    if (IS_JIT_FRAME(frame))
	frame = CompiledFrameUpdate(frame);
    while (frame != NULL) {
    	if (frame->current_method != NULL) {
	    char where[MSG_BUF_LEN];
	    unsigned char *pc;
	    struct methodblock *mb = frame->current_method;

	    if (depth_limit-- <= 0) {
	        char *msg = "\t... (more frames not shown)\n";
	        if (pf) {
		    pf(msg);
		} else {
	            jio_fprintf(stderr, msg);
		}
	        break;
	    }

	    if (mb->fb.access & ACC_NATIVE) {
	        pc = mb->code;
	    } else if (IS_JIT_FRAME(frame)) {
		pc = CompiledCodePC(frame, mb);
	    } else {
	        pc = frame->lastpc;
            }

	    strncpy(where, "\tat ", 4);
	    pc2string(pc, frame, where + 4, where + sizeof(where));
	    if (pf) {
	        pf("%s\n", where);
	    } else {
	        jio_fprintf(stderr, "%s\n", where);
	    }
        }
	frame = FRAME_PREV(frame, &frame_buf);
    } 
}
