/*
 * @(#)debug.c	1.53 98/08/23
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Debugging help for users.
 */

#include <stdio.h>

#include "threads.h"
#include "debug.h"
#include "monitor.h"
#include "monitor_cache.h"
#include "exceptions.h"
#include "bool.h"
#include "gc.h"
#include "interpreter.h"
#include "vmprofiler.h"

#include "java_lang_Thread.h"

#ifdef LOGGING
int logging_level;
#endif

static char *printf_buffer;
static int printf_buffer_size = 1024 * 256; /* by default 256K print buffer */
static char *printf_buffer_ptr;

static void 
buffered_printf(char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    if (printf_buffer_ptr < printf_buffer + printf_buffer_size - 128) {
        printf_buffer_ptr += jio_vsnprintf(printf_buffer_ptr,
	    printf_buffer + printf_buffer_size - printf_buffer_ptr,
					   fmt,
					   args);
    }
    va_end(args);
}

static bool_t
allocPrintBuffer(void)
{

    char *buf_sz_str;
    sys_thread_t *self = sysThreadSelf();
    QUEUE_LOCK(self);
    CACHE_LOCK(self);
    REGISTRY_LOCK(self);
    buf_sz_str = getenv("_JVM_THREAD_DUMP_BUFFER_SIZE");
    if (buf_sz_str) {
        printf_buffer_size = atoi(buf_sz_str);
    }
    printf_buffer = printf_buffer_ptr = sysMalloc(printf_buffer_size);
    if (printf_buffer == NULL) {
        REGISTRY_UNLOCK(self);
	CACHE_UNLOCK(self);
	QUEUE_UNLOCK(self);
	return FALSE;
    }        
    sysThreadSingle();
    return TRUE;
}

static void
flushPrintBuffer(void)
{
    sys_thread_t *self = sysThreadSelf();
    sysThreadMulti();
    jio_fprintf(stderr, "%s\n", printf_buffer);
    sysFree(printf_buffer);
    REGISTRY_UNLOCK(self);
    CACHE_UNLOCK(self);
    QUEUE_UNLOCK(self);
}

static void 
dumpSysThreadInfo(sys_thread_t *tid)
{
    TID thr = (TID)(SysThread2EE(tid)->thread);
    if (thr) {
        buffered_printf("%s (0x%x)", 
			Object2CString((JHandle *) THREAD(thr)->name), tid);
    } else {
        buffered_printf("<unknown thread> (0x%x)", tid); 
    }
}

static void
threadDumpInfo(TID tid, bool_t verbose)
{
    buffered_printf("    %s", Object2CString((JHandle *) THREAD(tid)->name));
    if (verbose == TRUE) {
        ExecEnv *ee = (ExecEnv *)ll2ptr(THREAD(tid)->eetop);
	sys_thread_t *t = ee ? EE2SysThread(ee) : 0;

	buffered_printf(" (TID:0x%x", tid);

	if (THREAD(tid)->stillborn) {
	    buffered_printf(", stillborn");
	}

	if (t) {
	    void *nid = sysThreadNativeID(t);
	    int status = sysThreadGetStatus(t, NULL);
	    int interrupted = status & HPI_THREAD_INTERRUPTED;
	    status &= ~HPI_THREAD_INTERRUPTED;
	    buffered_printf(", sys_thread_t:0x%x, state:", t);
	    if (status & HPI_THREAD_SUSPENDED) {
	        buffered_printf("S");
	    } else if (status == HPI_THREAD_RUNNABLE) {
	        buffered_printf("R");
	    } else if (status == HPI_THREAD_MONITOR_WAIT) {
	        buffered_printf("MW");
	    } else if (status == HPI_THREAD_CONDVAR_WAIT) {
	        buffered_printf("CW");
	    } else {
	        buffered_printf("?");
	    }
	    if (interrupted) {
	        buffered_printf(", intr");
	    }
	    if (nid) {
	        buffered_printf(", native ID:0x%x", nid);
	    }
            buffered_printf(")");
	    buffered_printf(" prio=%d", THREAD(tid)->priority);
	    if (ee != NULL && exceptionOccurred(ee)) {
	        char buf[MSG_BUF_LEN];
	        exception_t exc = ee->exception.exc;
	        buffered_printf(": pending=%s",
				classname2string(cbName(obj_classblock(exc)),
						 buf,
						 sizeof(buf)));
	    }
	    buffered_printf("\n");
    	} else {
	    buffered_printf(") : <thread not active>\n");
	}
    } else {
        ExecEnv *ee = (ExecEnv *)ll2ptr(THREAD(tid)->eetop);
	sys_thread_t *t = ee ? EE2SysThread(ee) : 0;
	buffered_printf(" (0x%lx)\n", t);
    }
}

static int
DumpThreadsHelper(sys_thread_t *t, void *arg)
{
    struct execenv *ee = SysThread2EE(t);
    HThread *p = (HThread *)ee->thread;

    if (p) {
        threadDumpInfo(p, TRUE);
	printStackTrace(ee, 20, buffered_printf);
    }

    return SYS_OK;
}

static void
DumpThreads0(bool_t locked)
{
    static int alreadyHere = 0;

    if (alreadyHere) {
	return;
    }

    if (!locked) {
        if (!allocPrintBuffer()) {
	    return;
	}
    }

    buffered_printf("\nFull thread dump Classic VM (%s, %s):\n", FULL_VERSION,
		    sysGetSysInfo()->name);

    alreadyHere++;
    sysThreadEnumerateOver(DumpThreadsHelper, NULL);
    alreadyHere--;

    if (!locked) {
        flushPrintBuffer();
    }
}

void
DumpThreads(void)
{
    DumpThreads0(FALSE);
}

static void 
monitorDumpInfo(sys_mon_t *mid)
{
#define MAX_WAITERS 1000
    sys_thread_t *monitor_waiters[MAX_WAITERS];
    sys_thread_t *condvar_waiters[MAX_WAITERS];
    sys_mon_info info;
    int i;

    info.monitor_waiters = monitor_waiters;
    info.condvar_waiters = condvar_waiters;
    info.sz_monitor_waiters = MAX_WAITERS;
    info.sz_condvar_waiters = MAX_WAITERS;

    sysMonitorGetInfo(mid, &info);

    if (info.owner) {
        buffered_printf("owner ");
	dumpSysThreadInfo(info.owner);
	buffered_printf(" %ld %s\n",
			info.entry_count,
			info.entry_count == 1 ? "entry" : "entries");
    } else {
        buffered_printf("<unowned>\n");
    }

    if (info.n_monitor_waiters) {
        buffered_printf("\tWaiting to enter:\n");
	for (i = 0; i < info.n_monitor_waiters; i++) {
	    buffered_printf("\t    ");
	    if (i >= MAX_WAITERS) {
	        buffered_printf("... (%d more not shown)\n",
				i - MAX_WAITERS + 1);
		break;
	    }
	    dumpSysThreadInfo(info.monitor_waiters[i]);
	    buffered_printf("\n");
	}
    }

    if (info.n_condvar_waiters) {
        buffered_printf("\tWaiting to be notified:\n");
	for (i = 0; i < info.n_condvar_waiters; i++) {
	    buffered_printf("\t    ");
	    if (i >= MAX_WAITERS) {
	        buffered_printf("... (%d more not shown)\n",
				i - MAX_WAITERS + 1);
		break;
	    }
	    dumpSysThreadInfo(info.condvar_waiters[i]);
	    buffered_printf("\n");
	}
    }   
}

/*
 * The arg that is passed in here is a verbose flag that controls
 * whether information is to be printed about unused monitors.  Monitors 
 * that are passed in here are always in the monitor cache.
 */ 
static void
monitorDumpHelper(monitor_t *mid, void *arg)
{
    unsigned int key = mid->key;
    bool_t verbose = (bool_t) arg;

    if (verbose != FALSE || sysMonitorInUse(sysmon(mid))) {
	if (isHandle((void *)key)) {
	    buffered_printf("    %s: ", Object2CString((JHandle *) key));
	} else {
	    buffered_printf("    <unknown key> (0x%lx): ", (long)key);
	}
	monitorDumpInfo(sysmon(mid));
    }
}

/*
 * Access to monitor information
 */
static void
registeredDumpHelper(reg_mon_t *reg, void *ignored)
{
    buffered_printf("    %s: ", reg->name);
    monitorDumpInfo(reg->mid);
}

/*
 * If verbose == TRUE, we print information about unused cached monitors.
 * This is normally not done, but can be useful when debugging the cache
 * itself.
 */
static void
DumpMonitors0(bool_t locked, bool_t verbose)
{
    if (!locked) {
        if (!allocPrintBuffer()) {
	    return;
	}
    }
    /*
     * Dump the contents of the monitor cache
     */
    buffered_printf("Monitor Cache Dump:\n");
    monitorEnumerate(monitorDumpHelper, (void *) verbose);

    /*
     * Also dump other well-known monitors: what these will be is
     * system-dependent and does not belong in share code.
     */
    buffered_printf("Registered Monitor Dump:\n");
    registeredEnumerate(registeredDumpHelper, 0);

    if (!locked) {
        flushPrintBuffer();
    }
}

/*
 * Called from ^\, in which case we want to suppress information about
 * unused monitors in the cache.
 */
void
DumpMonitors(void)
{
    DumpMonitors0(FALSE, FALSE);
}

void
HandleSignalInVM(bool_t noncritical)
{
#ifdef HPROF
    if (jvmpi_event_flags && noncritical) {
	jvmpi_dump();
	return;
    }
#endif /* HPROF */
    if (allocPrintBuffer()) {
        DumpThreads0(TRUE);
	DumpMonitors0(TRUE, FALSE);
	flushPrintBuffer();
    }
}
