/*
 * @(#)executeJava.c	1.90 98/09/15
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Interpret a Java object file
 * This is the very heart of the interpreter. This can be
 * re-written in assembly language for speed on a specific platform,
 * but this file must remain as the reference, and for use with tracing
 * and debugging.
 */

#include <sys/types.h>
#include <string.h>
#include <stdarg.h>

#include "bool.h"
#include "opcodes.h"
#include "oobj.h"
#include "interpreter.h"
#include "signature.h"
#include "typecodes.h"
#include "tree.h"
#include "threads.h"
#include "monitor.h"
#include "exceptions.h"
#include "javaString.h"
#include "path.h"
#include "byteorder.h"
#include "jmath.h"
#include "common_exceptions.h"

#include "sys_api.h"

#include "java_lang_Thread.h"
#include "java_lang_Throwable.h"

#ifdef BREAKPTS
#include "breakpoints.h"
#endif


#ifdef HPROF
#include "vmprofiler.h"
#endif /* HPROF */


#ifdef TRACING
#  define TRACE(a) (trace ? jio_fprintf(stdout, "%6X %8X", (int)threadSelf(), (int)pc), \
	jio_printf a, fflush(stdout), 0 : 0)
#  define TRACEIF(o, a) (opcode == o ? TRACE(a) : 0)

#else
#  define TRACE(a)
#  define TRACEIF(a, b)
#endif

#ifdef STATISTICS
int instruction_count[256][256];
#endif

/*
 * This can be used to track handles
 */
static JHandle *
checkHandle(JHandle * h)
{
    sysAssert(h == NULL || (isHandle(h) && (unhand(h) != NULL)));
    return h;
}


/*
 * Java stack management
 */

/*
 * The following hack should be reevaluated using the current MSVC++.
 * It dates back to a version several releases old. [Tim, 10/24/96]
 */
#ifdef WIN32
static void HACK_NOP() {
    /* This is a hack to work around an optimizer bug in MSVC++ */
}
#else
#define HACK_NOP()
#endif

#ifdef BREAKPTS

#define WITH_OPCODE_CHANGE { \
	bool_t at_breakpoint = (*pc == opc_breakpoint);
#define END_WITH_OPCODE_CHANGE \
        if (at_breakpoint && (*pc != opc_breakpoint)) { \
        	opcode = *pc; \
                *pc = opc_breakpoint;  \
                set_breakpoint_opcode(ee, pc, opcode); \
                goto again; } }

#else  /* BREAKPTS */

#define WITH_OPCODE_CHANGE
#define END_WITH_OPCODE_CHANGE

#endif /* BREAKPTS */

/* Execute bytecodes at "pc" in the specified execution environment.
   Returns TRUE if completed OK, FALSE if finished because of some
   uncaught exception. */

bool_t
#ifdef BREAKPTS
ExecuteJava_C
#else
ExecuteJava
#endif /* BREAKPTS */
(unsigned char *initial_pc, ExecEnv *ee)
{
/* This is ugly.  Some machines (and compilers) produce better code from
 * positive offsets than from negative offsets.  So we cache ee->optop in
 * a local variable, but have it point OPTOP_OFFSET words further down the
 * stack.  That way, stack references will be a nonnegative index off offset
 */
#define OPTOP_OFFSET 4

#define S_INFO(offset)    (optop[offset + OPTOP_OFFSET])
#define S_INT(offset)     (S_INFO(offset).i)
#define S_FLOAT(offset)   (S_INFO(offset).f)
#define S_GET_HANDLE(offset)  (checkHandle(S_INFO(offset).h))
#define S_SET_HANDLE(offset, handle)  (S_INFO(offset).h =  checkHandle(handle))
#define S_OBJECT(offset)  (S_INFO(offset).o)
#define S_POINTER(offset) (S_INFO(offset).p)
#define S_OSTRING(offset) (Object2CString(S_GET_HANDLE(offset)))
#define S_LONGSTRING(offset) \
        (int642CString(S_LONG(offset), longstringbuf, sizeof(longstringbuf)))

#define S_DOUBLE(offset)  GET_DOUBLE(tdub,&(S_INFO(offset)))
#define S_DOUBLE2(offset)  GET_DOUBLE(tdub2,&(S_INFO(offset)))
#define S_LONG(offset)    GET_INT64(tdub,&(S_INFO(offset)))
#define S_LONG2(offset)    GET_INT64(tdub2,&(S_INFO(offset)))
#define SET_S_DOUBLE(offset, value) SET_DOUBLE(tdub,&(S_INFO(offset)), (value))
#define SET_S_LONG(offset, value)   SET_INT64(tdub,&(S_INFO(offset)), (value))

#define REAL_OPTOP()      (&S_INFO(0))

#define DECACHE_OPTOP()   (frame->optop = REAL_OPTOP())
#define CACHE_OPTOP()     (optop = frame->optop - OPTOP_OFFSET)

#ifdef JCOV

#define SIZE_AND_STACK(size, stack) pc += size; optop += stack; fl_next_optop = FALSE; continue
#define SIZE_AND_STACK_BUT(size, stack) pc += size; optop += stack; fl_next_optop = FALSE;
#define JUMP_AND_STACK(offset, stack) \
          if (1) { pc += offset; optop += stack; fl_next_optop = FALSE; \
                   if (offset > 0) continue; \
	           goto check_for_exception; } else ((void) 0)

#else
#define SIZE_AND_STACK(size, stack) pc += size; optop += stack; continue
#define SIZE_AND_STACK_BUT(size, stack) pc += size; optop += stack;
#define JUMP_AND_STACK(offset, stack) \
          if (1) { pc += offset; optop += stack; \
                   if (offset > 0) continue; \
	           goto check_for_exception; } else ((void) 0)


#endif

/* I can't use the do { } while(0) trick because the compiler complains about
 * the test not bein reachable.  So I'm forced to use an uglier trick.
 */
#define JAVA_ERROR(name, data) \
        if (1) { \
		frame->lastpc = pc;  \
	        SignalError(ee, JAVAPKG name, data); \
                goto handle_exception; \
	} else ((void) 0)


    register unsigned char  *pc = initial_pc;
    register JavaFrame       *frame = ee->current_frame;
    JavaFrame                *initial_frame = frame;
    register stack_item     *optop;
    register stack_item     *vars = frame->vars;

    unsigned char            opcode;
    cp_item_type            *constant_pool = frame->constant_pool;
    struct methodblock      *mb;
#ifdef TRACING
    char 		     longstringbuf[30];
#endif /* TRACING */

    JHandle                 *o;	/* object whose method we're calling  */
    int	    args_size;
    jvalue tdub, tdub2;
#ifdef JCOV
    int fl_next_optop = FALSE;
#endif
#ifdef STATISTICS
    int lastOpcode = opc_nop;
#endif

    CACHE_OPTOP();

    if (!threadCheckStack()) {
	/*
	 * Got a stack overflow: have to be careful to not do anything
	 * that writes to the stack.  "StackOverflowError" is a
	 * special kind of exception, because we can't call SignalError().
	 */
	frame->lastpc = pc;
        ThrowStackOverflowError(0, 0);
	return FALSE;
    }

#ifdef TRACING
    if (trace) {
	char where_am_i[100];
	pc2string(pc, frame, where_am_i, where_am_i + sizeof where_am_i);
	jio_fprintf(stdout, "Entering %s\n", where_am_i);
	fflush(stdout);
    }
#endif /* TRACING */

    /*CONSTCOND*/
    while (1) {
#ifdef BREAKPTS
	int stopped_at_single_step = FALSE;
#endif /* BREAKPTS */
	sysAssert(REAL_OPTOP() >= frame->ostack);
	sysAssert(!frame->current_method ||
		(frame->current_method->fb.access & ACC_NATIVE) ||
	       REAL_OPTOP() <= frame->ostack + frame->current_method->maxstack);

#ifdef BREAKPTS
	frame->lastpc = pc;
	if (single_stepping > 0 &&
            unhand((HThread *)(ee->thread))->single_step &&
            frame->current_method != NULL) {
	    DECACHE_OPTOP();
            if (debugging) {
                notify_debugger_of_single_step(ee, pc);
            }
	    stopped_at_single_step = TRUE;
	}
#endif
#ifdef JCOV
	if (frame->current_method && frame->current_method->coverage_table
            && cov_file && !fl_next_optop) {
		struct methodblock *mb = frame->current_method;
    		int *jcov_mode = (int*)(mb->coverage_table + mb->coverage_table_length);
		int *hash = jcov_mode + 1;
                int  addr = pc - mb->code;
		struct covtable *cv = mb->coverage_table;

		if ((*jcov_mode) & JCOV_MODE_METHODS) {
		   if (addr == 0)			/* for method only */
                       cv->count++;
		} else if (hash[addr] != 0) {
                   int nmcommand;

                   cv += (hash[addr] - 1);
		   nmcommand = cv->pc;
                   while ( nmcommand == cv->pc) {
                       if ( cv->type < CT_CASE) {
                          cv->count++;
                       }
                       if(cv == mb->coverage_table)
                          break;
                       cv--;
                   }
                }
                fl_next_optop = TRUE;
	}
#endif

	opcode = *pc;
again:
#ifdef STATISTICS
	instruction_count[lastOpcode][opcode]++;
	lastOpcode = opcode;
#endif

	switch (opcode) {

	    /* NOPs.  Don't do anything. */
	case opc_nop:
	    TRACE(("\tnop\n"));
	    SIZE_AND_STACK(1, 0);

	case opc_aload:
	case opc_iload:
        case opc_fload:
	    S_INFO(0) = vars[pc[1]];
	    TRACEIF(opc_iload, ("\tiload r%d => %d\n", pc[1], S_INT(0)));
	    TRACEIF(opc_aload, ("\taload r%d => %s\n", pc[1], S_OSTRING(0)));
	    TRACEIF(opc_fload, ("\tfload r%d => %g\n", pc[1], S_FLOAT(0)));
	    SIZE_AND_STACK(2, 1);

	case opc_lload:
	case opc_dload: {
	    int reg = pc[1];
	    S_INFO(0) = vars[reg];
	    S_INFO(1) = vars[reg + 1];
	    TRACEIF(opc_lload, ("\tlload r%d => %s\n", reg, S_LONGSTRING(0)));
	    TRACEIF(opc_dload, ("\tdload r%d => %g\n", reg, S_DOUBLE(0)));
	    SIZE_AND_STACK(2, 2);
	}

#define OPC_DO_LOAD_n(num)                                              \
	case opc_iload_##num:                                           \
	case opc_aload_##num:                                           \
	case opc_fload_##num:                                           \
	    S_INFO(0) = vars[num];                                      \
	    TRACEIF(opc_iload_##num,                                    \
		    ("\t%s => %d\n", opnames[opcode], S_INT(0)));       \
	    TRACEIF(opc_aload_##num,                                    \
		    ("\t%s => %s\n", opnames[opcode], S_OSTRING(0)));   \
	    TRACEIF(opc_fload_##num,                                    \
		    ("\t%s => %g\n", opnames[opcode], S_FLOAT(0)));     \
	    SIZE_AND_STACK(1, 1);					\
        case opc_lload_##num:                                         	\
	case opc_dload_##num:	                                        \
	    S_INFO(0) = vars[num];                                      \
	    S_INFO(1) = vars[num + 1];                                  \
	    TRACEIF(opc_lload_##num,                                    \
		    ("\t%s => %s\n", opnames[opcode], S_LONGSTRING(0))); \
	    TRACEIF(opc_dload_##num,                                    \
		    ("\t%s => %g\n", opnames[opcode], S_DOUBLE(0)));    \
	    SIZE_AND_STACK(1, 2);


	OPC_DO_LOAD_n(0)
	OPC_DO_LOAD_n(1)
	OPC_DO_LOAD_n(2)
	OPC_DO_LOAD_n(3)

	      /* store a local variable onto the stack. */
	case opc_istore:
	case opc_astore:
	case opc_fstore:
	    vars[pc[1]] = S_INFO(-1);
	    TRACEIF(opc_istore,("\tistore %d => r%d\n", S_INT(-1), pc[1]));
	    TRACEIF(opc_astore,("\tastore %s => r%d\n", S_OSTRING(-1),pc[1]));
	    TRACEIF(opc_fstore,("\tfstore %g => r%d\n", S_FLOAT(-1), pc[1]));
	    SIZE_AND_STACK(2, -1);

	case opc_lstore:
	case opc_dstore: {
	    int reg = pc[1];
	    vars[reg] = S_INFO(-2);
	    vars[reg + 1] = S_INFO(-1);
	    TRACEIF(opc_lstore, ("\tlstore %s => r%d\n", S_LONGSTRING(-2),reg));
	    TRACEIF(opc_dstore, ("\tdstore %g => r%d\n", S_DOUBLE(-2), reg));
	    SIZE_AND_STACK(2, -2);
	}

#define OPC_DO_STORE_n(num)                                             \
	case opc_istore_##num:				        	\
	case opc_astore_##num: 						\
	case opc_fstore_##num: 						\
	    vars[num] = S_INFO(-1);  					\
	    TRACEIF(opc_istore_##num, ("\t%s %d => r%d\n", 		\
		       opnames[opcode], S_INT(-1), num));  		\
	    TRACEIF(opc_astore_##num, ("\t%s %p => r%d\n", 		\
			      opnames[opcode], S_OSTRING(-1), num)); 	\
	    TRACEIF(opc_fstore_##num, ("\t%s %g => r%d\n", 		\
			      opnames[opcode], S_FLOAT(-1), num)); 	\
	    SIZE_AND_STACK(1, -1);                                      \
	case opc_lstore_##num: 						\
	case opc_dstore_##num: 						\
	    vars[num] = S_INFO(-2); 					\
	    vars[num + 1] = S_INFO(-1); 				\
	    TRACEIF(opc_lstore_##num, ("\t%s %s => r%d\n", 		\
			     opnames[opcode], S_LONGSTRING(-2), num)); 	\
	    TRACEIF(opc_dstore_##num, ("\t%s %g => r%d\n", 		\
			     opnames[opcode], S_DOUBLE(-2), num)); 	\
	    SIZE_AND_STACK(1, -2);

	OPC_DO_STORE_n(0)
	OPC_DO_STORE_n(1)
	OPC_DO_STORE_n(2)
	OPC_DO_STORE_n(3)


	    /* Return a value from a function */
	case opc_areturn:
	case opc_ireturn:
	case opc_freturn:
	    /* increment optop first to make sure GC scans the return object */
	    frame->prev->optop++;
	    frame->prev->optop[-1] = S_INFO(-1);
	    TRACEIF(opc_ireturn, ("\tireturn %d\n", S_INT(-1)));
	    TRACEIF(opc_areturn, ("\tareturn %s\n", S_OSTRING(-1)));
	    TRACEIF(opc_freturn, ("\tfreturn %g\n", S_FLOAT(-1)));
	    goto finish_return;

	case opc_lreturn:
	case opc_dreturn:
	    frame->prev->optop[0] = S_INFO(-2);
	    frame->prev->optop[1] = S_INFO(-1);
	    frame->prev->optop += 2;
	    TRACEIF(opc_lreturn, ("\tlreturn %s\n", S_LONGSTRING(-2)));
	    TRACEIF(opc_dreturn, ("\tdreturn %g\n", S_DOUBLE(-2)));
	    goto finish_return;

	    /* Return from a function with no value */
	case opc_return:
	    TRACE(("\treturn\n"));
	    goto finish_return;

	case opc_i2f:		/* Convert top of stack int to float */
	    S_FLOAT(-1) = (float)S_INT(-1);
	    TRACE(("\ti2f => %g\n", S_FLOAT(-1)));
	    SIZE_AND_STACK(1, 0);

	case opc_i2l:		/* Convert top of stack int to long */
	    SET_S_LONG(-1, int2ll(S_INT(-1)));
	    TRACE(("\ti2l => %s\n", S_LONGSTRING(-1)));
	    SIZE_AND_STACK(1, 1);

	case opc_i2d:
	    SET_S_DOUBLE(-1, S_INT(-1));
	    TRACE(("\ti2d => %g\n", S_DOUBLE(-1)));
	    SIZE_AND_STACK(1, 1);

	case opc_f2i:		/* Convert top of stack float to int */
	    S_INT(-1) = float2l(S_FLOAT(-1));
	    TRACE(("\tf2i => %d\n", S_INT(-1)));
	    SIZE_AND_STACK(1, 0);

	case opc_f2l:		/* convert top of stack float to long */
	    SET_S_LONG(-1, float2ll(S_FLOAT(-1)));
	    TRACE(("\tf2l => %s\n", S_LONGSTRING(-1)));
	    SIZE_AND_STACK(1, 1);

	case opc_f2d:		/* convert top of stack float to double */
	    SET_S_DOUBLE(-1, S_FLOAT(-1));
	    TRACE(("\tf2d => %g\n", S_DOUBLE(-1)));
	    SIZE_AND_STACK(1, 1);

	case opc_l2i:		/* convert top of stack long to int */
	    S_INT(-2) = ll2int(S_LONG(-2));
	    TRACE(("\tl2i => %d\n", S_INT(-2)));
	    SIZE_AND_STACK(1, -1);

	case opc_l2f:		/* convert top of stack long to float */
	    S_FLOAT(-2) = ll2float(S_LONG(-2));
	    HACK_NOP();
	    TRACE(("\tl2f => %g\n", S_FLOAT(-2)));
	    SIZE_AND_STACK(1, -1);

	case opc_l2d:		/* convert top of stack long to double */
	    SET_S_DOUBLE(-2, ll2double(S_LONG(-2)));
	    HACK_NOP();
	    TRACE(("\tl2d => %g\n", S_DOUBLE(-2)));
	    SIZE_AND_STACK(1, 0);

	case opc_d2i:
	    S_INT(-2) = double2l(S_DOUBLE(-2));
	    TRACE(("\td2i => %d\n", S_INT(-2)));
	    SIZE_AND_STACK(1, -1);

	case opc_d2f:
	    S_FLOAT(-2) = S_DOUBLE(-2);
	    HACK_NOP();
	    TRACE(("\td2f => %g\n", S_FLOAT(-2)));
	    SIZE_AND_STACK(1, -1);

	case opc_d2l:
	    SET_S_LONG(-2, double2ll(S_DOUBLE(-2)));
	    TRACE(("\td2l => %s\n", S_LONGSTRING(-2)));
	    SIZE_AND_STACK(1, 0);

	case opc_i2b:
	    S_INT(-1) = (signed char) S_INT(-1);
	    TRACE(("\ti2b => %d\n", S_INT(-1)));
	    SIZE_AND_STACK(1, 0);

	case opc_i2c:
	    S_INT(-1) &= 0xFFFF;
	    TRACE(("\ti2c => %d\n", S_INT(-1)));
	    SIZE_AND_STACK(1, 0);

	case opc_i2s:
	    S_INT(-1) = (signed short) S_INT(-1);
	    TRACE(("\ti2s => %d\n", S_INT(-1)));
	    SIZE_AND_STACK(1, 0);

        case opc_fcmpl:
        case opc_fcmpg: {
	    float l = S_FLOAT(-2);
	    float r = S_FLOAT(-1);
	    int value = NAN_CHECK(l,r,
			(l < r) ? -1 : (l > r) ? 1 : (l == r ) ?  0)
	                /* Beyond here, must be NaN. */
	                : (opcode == opc_fcmpl) ? -1
			: (opcode == opc_fcmpg) ? 1
			: 0;
	    S_INT(-2) = value;
	    TRACE(("\t%s => %d\n", opnames[opcode], S_INT(-2)));
	    SIZE_AND_STACK(1, -1);
	}

	case opc_lcmp: {
	    int64_t l = S_LONG(-4);
	    int64_t r = S_LONG(-2);
	    S_INT(-4) = (ll_lt(l, r)) ? -1 : (ll_gt(l, r)) ? 1 : 0;
	    TRACE(("\tlcmp => %d\n", S_INT(-4)));
	    SIZE_AND_STACK(1, -3);
	}

	case opc_dcmpl:
	case opc_dcmpg: {
	    double l = S_DOUBLE(-4);
	    double r = S_DOUBLE(-2);
	    int value = NAN_CHECK(l,r,
			(l < r) ? -1 : (l > r) ? 1 : (l == r ) ?  0)
	               /* Beyond here, must be NaN. */
			: (opcode == opc_dcmpl) ? -1
			: (opcode == opc_dcmpg) ? 1
			: 0;
	    S_INT(-4) = value;
	    TRACE(("\t%s => %d\n", opnames[opcode], S_INT(-4)));
	    SIZE_AND_STACK(1, -3);
	}

	    /* Push a 1-byte signed integer value onto the stack. */
	case opc_bipush:
	    S_INT(0) = (signed char)(pc[1]);
	    TRACE(("\tbipush %d\n", S_INT(0)));
	    SIZE_AND_STACK(2, 1);

	    /* Push a 2-byte signed integer constant onto the stack. */
	case opc_sipush:
	    S_INT(0) = PC2SIGNEDSHORT(pc);
	    TRACE(("\tsipush %d\n", S_INT(0)));
	    SIZE_AND_STACK(3, 1);

	case opc_ldc:
	    WITH_OPCODE_CHANGE
	        ResolveClassConstant(constant_pool, pc[1], ee,
				     (1 << CONSTANT_Integer) |
				     (1 << CONSTANT_Float) |
				     (1 << CONSTANT_String));
	        if (!exceptionOccurred(ee))
		    pc[0] = opc_ldc_quick;
	    END_WITH_OPCODE_CHANGE
	    goto check_for_exception;

	case opc_ldc_quick:
	    S_POINTER(0) = constant_pool[pc[1]].p;
	    TRACE(("\tldc1 #%d => 0x%x\n", pc[1], S_INT(0)));
	    SIZE_AND_STACK(2, 1);

	case opc_ldc_w:
	    WITH_OPCODE_CHANGE
	        ResolveClassConstant(constant_pool, GET_INDEX(pc + 1), ee,
				     (1 << CONSTANT_Integer) |
				     (1 << CONSTANT_Float) |
				     (1 << CONSTANT_String));
	        if (!exceptionOccurred(ee))
		    pc[0] = opc_ldc_w_quick;
	    END_WITH_OPCODE_CHANGE
	    goto check_for_exception;

	case opc_ldc_w_quick:
	    S_POINTER(0) = constant_pool[GET_INDEX(pc + 1)].p;
	    TRACE(("\tldc2 #%d => 0x%x\n", GET_INDEX(pc + 1), S_INT(0)));
	    SIZE_AND_STACK(3, 1);

	case opc_ldc2_w:	/* load a two word constant */
	    WITH_OPCODE_CHANGE
	        ResolveClassConstant(constant_pool, GET_INDEX(pc + 1), ee,
				 (1 << CONSTANT_Double) | (1 << CONSTANT_Long));
	        if (!exceptionOccurred(ee))
		    pc[0] = opc_ldc2_w_quick;
	    END_WITH_OPCODE_CHANGE
	    goto check_for_exception;

	case opc_ldc2_w_quick:
	    SET_S_LONG(0, GET_INT64(tdub2, &constant_pool[GET_INDEX(pc + 1)]));
	    TRACE(("\tldc2 #%d => long=%s double=%g\n", GET_INDEX(pc + 1),
		   S_LONGSTRING(0), S_DOUBLE2(0)));
	    SIZE_AND_STACK(3, 2);

#define CONSTANT_OP(opcode, this_type, value) 				\
	case opcode: 							\
	    S_INFO(0).this_type = value; 				\
	    TRACE(("\t%s\n", opnames[opcode])); 			\
	    SIZE_AND_STACK(1, 1);

	    /* Push miscellaneous constants onto the stack. */
	CONSTANT_OP(opc_aconst_null, p,        0)
	CONSTANT_OP(opc_iconst_m1,   i,       -1)
	CONSTANT_OP(opc_iconst_0,    i,        0)
	CONSTANT_OP(opc_iconst_1,    i,        1)
	CONSTANT_OP(opc_iconst_2,    i,        2)
	CONSTANT_OP(opc_iconst_3,    i,        3)
	CONSTANT_OP(opc_iconst_4,    i,        4)
	CONSTANT_OP(opc_iconst_5,    i,        5)
	CONSTANT_OP(opc_fconst_0,    f,      0.0)
	CONSTANT_OP(opc_fconst_1,    f,      1.0)
	CONSTANT_OP(opc_fconst_2,    f,      2.0)

	case opc_dconst_0:
	    SET_S_DOUBLE(0, 0.0);
	    TRACE(("\t%s\n", opnames[opcode]));
	    SIZE_AND_STACK(1, 2);

	case opc_dconst_1:
	    SET_S_DOUBLE(0, 1.0);
	    TRACE(("\t%s\n", opnames[opcode]));
	    SIZE_AND_STACK(1, 2);

	case opc_lconst_0:
	    SET_S_LONG(0, ll_zero_const);
	    TRACE(("\t%s\n", opnames[opcode]));
	    SIZE_AND_STACK(1, 2);

	case opc_lconst_1:
   	    SET_S_LONG(0, ll_one_const);
	    TRACE(("\t%s\n", opnames[opcode]));
	    SIZE_AND_STACK(1, 2);

#define BINARY_INT_OP(name, op, test)                                   \
	case opc_i##name:                                          	\
	    if (test && S_INT(-1) == 0)                              	\
               JAVA_ERROR("ArithmeticException", "/ by zero");          \
	    S_INT(-2) = INT_OP(S_INT(-2),op,S_INT(-1));               	\
	    TRACE(("\t%s => %d\n", opnames[opcode], S_INT(-2)));      	\
	    SIZE_AND_STACK(1, -1);                                    	\
	case opc_l##name:                                             	\
	    if (test && ll_eqz(S_LONG(-2)))                             \
               JAVA_ERROR("ArithmeticException", "/ by zero");          \
	    SET_S_LONG(-4, ll_##name(S_LONG(-4), S_LONG2(-2)));         \
	    TRACE(("\t%s => %s\n", opnames[opcode], S_LONGSTRING(-4))); \
	    SIZE_AND_STACK(1, -2);

	      /* Perform various binary integer operations */
	BINARY_INT_OP(add, +, 0)
	BINARY_INT_OP(sub, -, 0)
	BINARY_INT_OP(mul, *, 0)
	BINARY_INT_OP(and, &, 0)
	BINARY_INT_OP(or,  |, 0)
	BINARY_INT_OP(xor, ^, 0)
	BINARY_INT_OP(div, /, 1)
	BINARY_INT_OP(rem, %, 1)

#define BINARY_SHIFT_OP(name, op)                               	\
	case opc_i##name:                                          	\
	    S_INT(-2) = ((int)(S_INT(-2))) op (S_INT(-1) & 0x1F);       \
	    TRACE(("\t%s => %d\n", opnames[opcode], S_INT(-2)));      	\
	    SIZE_AND_STACK(1, -1);                                    	\
	case opc_l##name:                                             	\
	    SET_S_LONG(-3, ll_##name(S_LONG(-3), (S_INT(-1) & 0x3F)));	\
	    TRACE(("\t%s => %s\n", opnames[opcode], S_LONGSTRING(-3))); \
	    SIZE_AND_STACK(1, -1);                                      \

	BINARY_SHIFT_OP(shl, <<)
	BINARY_SHIFT_OP(shr, >>)

	case opc_iushr:
	    S_INT(-2) = ((unsigned int)(S_INT(-2))) >> (S_INT(-1) & 0x1F);
	    TRACE(("\t%s => %d\n", opnames[opcode], S_INT(-2)));
	    SIZE_AND_STACK(1, -1);
	case opc_lushr:
	    SET_S_LONG(-3, ll_ushr(S_LONG(-3), (S_INT(-1) & 0x3F)));
	    TRACE(("\t%s => %s\n", opnames[opcode], S_LONGSTRING(-3)));
	    SIZE_AND_STACK(1, -1);

	      /* Unary negation */
	case opc_ineg:
	    S_INT(-1) = -S_INT(-1);
	    TRACE(("\tineg => %d\n", S_INT(-1)));
	    SIZE_AND_STACK(1, 0);

	case opc_lneg:
            SET_S_LONG(-2, ll_neg(S_LONG(-2)));
	    TRACE(("\tlneg => %s\n", S_LONGSTRING(-2)));
	    SIZE_AND_STACK(1, 0);

	    /* Add a small constant to a local variable */
	case opc_iinc:
	    vars[pc[1]].i += ((signed char *) pc)[2];
	    TRACE(("\tiinc r%d+%d => %d\n",
		   pc[1], ((signed char *) pc)[2], vars[pc[1]].i));
	    SIZE_AND_STACK(3, 0);


#define BINARY_FLOAT_OP(name, op) 					\
	case opc_f##name: 						\
	    S_FLOAT(-2) = S_FLOAT(-2) op S_FLOAT(-1);                 	\
	    HACK_NOP();						        \
	    TRACE(("\t%s => %g\n", opnames[opcode], S_FLOAT(-2))); 	\
	    SIZE_AND_STACK(1, -1); 					\
	case opc_d##name: 						\
	    SET_S_DOUBLE(-4, S_DOUBLE(-4) op S_DOUBLE2(-2));   	        \
	    TRACE(("\t%s => %g\n", opnames[opcode], S_DOUBLE(-4)));     \
	    SIZE_AND_STACK(1, -2);

	BINARY_FLOAT_OP(add, +)
	BINARY_FLOAT_OP(sub, -)
	BINARY_FLOAT_OP(mul, *)
	BINARY_FLOAT_OP(div, /)

	case opc_frem:
	    S_FLOAT(-2) = DREM(S_FLOAT(-2), S_FLOAT(-1));
	    TRACE(("\t%s => %g\n", opnames[opcode], S_FLOAT(-2)));
	    SIZE_AND_STACK(1, -1);

	case opc_drem:
	    SET_S_DOUBLE(-4, DREM(S_DOUBLE(-4), S_DOUBLE2(-2)));
	    TRACE(("\t%s => %g\n", opnames[opcode], S_DOUBLE(-4)));
	    SIZE_AND_STACK(1, -2);

	case opc_fneg:
	    S_FLOAT(-1) = -S_FLOAT(-1);
	    TRACE(("\tfneg => %g\n", S_FLOAT(-1)));
	    SIZE_AND_STACK(1, 0);

	case opc_dneg:
	    SET_S_DOUBLE(-2, -S_DOUBLE(-2));
	    HACK_NOP();
	    TRACE(("\tdneg => %g\n", S_DOUBLE(-2)));
	    SIZE_AND_STACK(1, 0);

	    /* Unconditional goto.  Maybe save return address in register. */
	case opc_jsr:
	    S_INFO(0).addr = pc + 3;
	    optop++;
	    /*FALLTHROUGH*/

	case opc_goto: {
	    int skip = PC2SIGNEDSHORT(pc);
	    TRACE(("\t%s %lX\n", opnames[opcode], (long)(pc + skip)));
	    JUMP_AND_STACK(skip, 0);
	}

	case opc_jsr_w:
	    S_INFO(0).addr = pc + 5;
	    optop++;
	    /*FALLTHROUGH*/

	case opc_goto_w: {
	    int skip = PC2SIGNEDLONG(pc);
	    TRACE(("\t%s %lX\n", opnames[opcode], (long)(pc + skip)));
	    JUMP_AND_STACK(skip, 0);
	}

	    /* Return from subroutine.  Jump to address in register. */
	case opc_ret:
	    TRACE(("\tret %d (%lX)\n", pc[1], (long)(vars[pc[1]].addr)));
	    pc = vars[pc[1]].addr;
	    SIZE_AND_STACK_BUT(0, 0);
	    goto check_for_exception;


#ifdef JCOV
#define COVERAGE_IF(flag, frame, pc) coverage_if(flag, frame, pc)
#else
#define COVERAGE_IF(flag, frame, pc)
#endif

#define COMPARISON_OP(name, comparison)                                   \
	case opc_if_icmp##name: {					  \
            int skip = (S_INT(-2) comparison S_INT(-1))                   \
	           ? PC2SIGNEDSHORT(pc) : 3;                              \
            COVERAGE_IF(S_INT(-2) comparison S_INT(-1), frame, pc);         \
            TRACE(("\t%s goto %lX (%staken)\n",                           \
		   opnames[opcode],                                       \
		   (long)(pc + PC2SIGNEDSHORT(pc)),                       \
		   (S_INT(-2) comparison S_INT(-1)) ? "" : "not "));	  \
            JUMP_AND_STACK(skip, -2); 					  \
        }								  \
	case opc_if##name: {					     	  \
            int skip = (S_INT(-1) comparison 0) ? PC2SIGNEDSHORT(pc) : 3; \
            COVERAGE_IF(S_INT(-1) comparison 0, frame, pc);         \
            TRACE(("\t%s goto %lX (%staken)\n",                           \
		   opnames[opcode],                                       \
		   (long)(pc + PC2SIGNEDSHORT(pc)),                       \
		   (S_INT(-1) comparison 0) ? "" : "not "));	          \
            JUMP_AND_STACK(skip, -1); 					  \
        }                                                                 \

#define COMPARISON_OP2(name, nullname, comparison)                        \
        COMPARISON_OP(name, comparison)                                   \
	case opc_if_acmp##name: {					  \
            int skip = (S_POINTER(-2) comparison S_POINTER(-1))           \
	           ? PC2SIGNEDSHORT(pc) : 3;                              \
            COVERAGE_IF(S_POINTER(-2) comparison S_POINTER(-1), frame, pc); \
            TRACE(("\t%s goto %lX (%staken)\n",                           \
		   opnames[opcode],                                       \
		   (long)(pc + PC2SIGNEDSHORT(pc)),                       \
		   (S_POINTER(-2) comparison S_POINTER(-1)) ? "" : "not ")); \
            JUMP_AND_STACK(skip, -2); 					  \
        }								  \
	case opc_if##nullname: {				     	  \
            int skip = (S_POINTER(-1) comparison NULL) ? PC2SIGNEDSHORT(pc) : 3; \
            COVERAGE_IF(S_POINTER(-1) comparison NULL, frame, pc);         \
            TRACE(("\t%s goto %lX (%staken)\n",                          \
		   opnames[opcode],                                       \
		   (long)(pc + PC2SIGNEDSHORT(pc)),                       \
		   (S_INT(-1) comparison 0) ? "" : "not "));	          \
            JUMP_AND_STACK(skip, -1); 					  \
        }                                                                 \


	COMPARISON_OP2(eq, null, ==)	/* also generate acmp_eq and acmp_ne */
	COMPARISON_OP2(ne, nonnull, !=)
	COMPARISON_OP(lt, <)
	COMPARISON_OP(gt, >)
	COMPARISON_OP(le, <=)
	COMPARISON_OP(ge, >=)

	     /* pop stack, and error if it is null */
        case opc_nonnull_quick:
	    TRACE(("\tnonnull_quick\n"));
	    if (S_POINTER(-1) == NULL) {
		JAVA_ERROR("NullPointerException", 0);
	    }
	    SIZE_AND_STACK(1, -1);

	    /* Discard the top item on the stack */
	case opc_pop:
	    TRACE(("\tpop\n"));
	    SIZE_AND_STACK(1, -1);

	    /* Discard the top item on the stack */
	case opc_pop2:
	    TRACE(("\tpop2\n"));
	    SIZE_AND_STACK(1, -2);

	    /* Duplicate the top item on the stack */
	case opc_dup:
	    S_INFO(0) = S_INFO(-1);
	    TRACE(("\tdup\n"));
	    SIZE_AND_STACK(1, 1);

	case opc_dup2:		/* Duplicate the top two items on the stack */
	    S_INFO(0) = S_INFO(-2);
	    S_INFO(1) = S_INFO(-1);
	    TRACE(("\tdup2\n"));
	    SIZE_AND_STACK(1, 2);

	case opc_dup_x1:	/* insert S_INFO(-1) on stack past 1 */
	    S_INFO(0) = S_INFO(-1);
	    S_INFO(-1) = S_INFO(-2);
	    S_INFO(-2) = S_INFO(0);
	    TRACE(("\tdup_x1\n"));
	    SIZE_AND_STACK(1, 1);

	case opc_dup_x2:	/* insert S_INFO(-1) on stack past 2*/
	    S_INFO(0) = S_INFO(-1);
	    S_INFO(-1) = S_INFO(-2);
	    S_INFO(-2) = S_INFO(-3);
	    S_INFO(-3) = S_INFO(0);
	    TRACE(("\tdup_x2\n"));
	    SIZE_AND_STACK(1, 1);

	case opc_dup2_x1:	/* insert S_INFO(-1,-2) on stack past 1 */
	    S_INFO(1) = S_INFO(-1);
	    S_INFO(0) = S_INFO(-2);
	    S_INFO(-1) = S_INFO(-3);
	    S_INFO(-2) = S_INFO(1);
	    S_INFO(-3) = S_INFO(0);
	    TRACE(("\tdup2_x1\n"));
	    SIZE_AND_STACK(1, 2);

	case opc_dup2_x2:	/* insert S_INFO(-1,2) on stack past 2 */
	    S_INFO(1) = S_INFO(-1);
	    S_INFO(0) = S_INFO(-2);
	    S_INFO(-1) = S_INFO(-3);
	    S_INFO(-2) = S_INFO(-4);
	    S_INFO(-3) = S_INFO(1);
	    S_INFO(-4) = S_INFO(0);
	    TRACE(("\tdup2_x2\n"));
	    SIZE_AND_STACK(1, 2);

	case opc_swap: {        /* swap top two elements on the stack */
	    stack_item j = S_INFO(-1);
	    S_INFO(-1) = S_INFO(-2);
	    S_INFO(-2) = j;
	    TRACE(("\tswap\n"));
	    SIZE_AND_STACK(1, 0);
	}

	    /* Find the length of an array */
	case opc_arraylength: {
	    HArrayOfChar *arr = (HArrayOfChar *)S_GET_HANDLE(-1);
	    if (arr == 0)
		JAVA_ERROR("NullPointerException", 0);
	    S_INT(-1) = obj_length(arr);
	    TRACE(("\tarraylength %s => %d\n",
		   Object2CString((JHandle*)arr), S_INT(-1)));
	    SIZE_AND_STACK(1, 0);
	}

	    /* Goto one of a sequence of targets. */
	case opc_tableswitch: {
	    int *lpc = (int *) UCALIGN(pc + 1);
	    int key = S_INT(-1);
	    int low = ntohl(lpc[1]);
	    int high = ntohl(lpc[2]);
	    int skip;
	    TRACE(("\ttableswitch %d [%d-%d]\n", key, low, high));
	    key -= low;
#ifdef JCOV
	    coverage_switch(frame, pc, key, low, high);
#endif
	    skip = ((unsigned) key > high - low)
		        ? ntohl(lpc[0])
		        : ntohl(lpc[key + 3]);
	    JUMP_AND_STACK(skip, -1);
	}

	    /* Goto one of a sequence of targets */
	case opc_lookupswitch: {
	    int *lpc = (int *) UCALIGN(pc + 1);
	    int key = S_INT(-1);
	    int skip = ntohl(lpc[0]); /* default skip amount */
	    int npairs = ntohl(lpc[1]);
#ifdef JCOV
	    int nodef = TRUE;
#endif
	    TRACE(("\tlookupswitch %d\n", key));
	    while (--npairs >= 0) {
		lpc += 2;
		if (key == ntohl(lpc[0])) {
		    skip = ntohl(lpc[1]);
#ifdef JCOV
		    nodef = FALSE;
#endif
		    break;
		}
	    }
#ifdef JCOV
	    coverage_lookupswitch(frame, pc, npairs, nodef);
#endif
	    JUMP_AND_STACK(skip, -1);
	}


	    /* Throw to a given value */
	case opc_athrow: {
	    JHandle *o = S_GET_HANDLE(-1);
	    TRACE(("\tathrow => %s\n", Object2CString(o)));
	    if (o == 0) {
	        JAVA_ERROR("NullPointerException", 0);
	    } else {
		exceptionThrow(ee, o);
		goto handle_exception;
	    }

	}

#define DO_FIELD(isStatic) \
	    DECACHE_OPTOP();                                                  \
             WITH_OPCODE_CHANGE {                                             \
		struct fieldblock *fb =					      \
		    ResolveClassConstantFromPC(pc, opcode, constant_pool, ee, \
					       1 << CONSTANT_Fieldref);	      \
		if (!exceptionOccurred(ee)) {				      \
		    int rewrite;					      \
		    if (fb == NULL) {					      \
			SIZE_AND_STACK(0, 0);				      \
		    }							      \
		    rewrite = isStatic ? quickStaticAccess(opcode, pc, fb, ee)\
			               : quickFieldAccess(opcode, pc, fb, ee);\
		    if (rewrite < 0) {  				      \
			 goto handle_exception;				      \
		    } else if (rewrite > 0) { 				      \
			SIZE_AND_STACK(0, 0);				      \
		    } 							      \
		}							      \
	    } END_WITH_OPCODE_CHANGE					      \
	    goto check_for_exception					      \

	case opc_getfield:
	case opc_putfield:
	    DO_FIELD(FALSE);

	case opc_getstatic:
	case opc_putstatic:
	    DO_FIELD(TRUE);


#ifdef BREAKPTS
#define RESTORE_BREAKPOINT_OPCODE \
	    if (opcode == opc_breakpoint)				 \
		opcode = get_breakpoint_opcode(ee, pc, FALSE);		 
#else
#define RESTORE_BREAKPOINT_OPCODE
#endif

#define DO_INVOKER(isStatic) \
	    WITH_OPCODE_CHANGE {                                         \
		mb = ResolveClassConstantFromPC(pc, opcode,              \
						constant_pool, ee,       \
						1 << CONSTANT_Methodref);\
		if (!exceptionOccurred(ee)) {				 \
		    int result;                                          \
		    if (mb == NULL) { 					 \
			SIZE_AND_STACK(0, 0);				 \
		    } 							 \
		    args_size = mb->args_size;				 \
		    if (isStatic) {					 \
			o = (JHandle *)(cbHandle(fieldclass(&mb->fb)));	 \
		    } else { 						 \
			if ((o = S_GET_HANDLE(-args_size)) == NULL) 	 \
			    JAVA_ERROR("NullPointerException", 0);	 \
		    }							 \
                    if (!debugging) {                                    \
		        result = quickInvocation(opcode, pc, mb, ee);    \
		        if (result < 0) {                                \
			     goto handle_exception;                      \
		        } else if (result > 0) {                         \
			    SIZE_AND_STACK(0, 0);                        \
		        }                                                \
                    }                                                    \
		}							 \
	    } END_WITH_OPCODE_CHANGE					 \
	    opcode = *pc;						 \
            RESTORE_BREAKPOINT_OPCODE                                    \
	    if (exceptionOccurred(ee))					 \
		goto check_for_exception;				 \

	case opc_invokevirtual:
	    DO_INVOKER(FALSE);
	    if (opcode == opc_invokevirtual) {
		if (mb->fb.u.offset != 0)
		    mb = mt_slot(obj_array_methodtable(o), mb->fb.u.offset);
		frame->returnpc = 0;;
		optop -= args_size;
		goto callmethod;
	    } else {
		SIZE_AND_STACK(0, 0);
	    }

	case opc_invokespecial:
	    DO_INVOKER(FALSE);
	    if (opcode == opc_invokespecial) {
		mb = quickSelectSuperMethod(ee->current_frame, mb);
		frame->returnpc = 0;
		optop -= args_size;
		goto callmethod;
	    } else {
		SIZE_AND_STACK(0,0);
	    }

	case opc_invokestatic:
            DO_INVOKER(TRUE);
	    if (opcode == opc_invokestatic) {
		frame->returnpc = 0;
		optop -= args_size;
		goto callmethod;
	    } else {
		SIZE_AND_STACK(0, 0);
	    }

	case opc_invokeinterface:
	    WITH_OPCODE_CHANGE {
		unsigned InterfaceMethodref_index = GET_INDEX(pc + 1);
		ResolveClassConstant(constant_pool, InterfaceMethodref_index,
				     ee, 1 << CONSTANT_InterfaceMethodref);
		if (!exceptionOccurred(ee)) {
		    pc[0] = opc_invokeinterface_quick;
#ifdef TRACING
		    if (trace) {
			struct methodblock *imb = 
			    constant_pool[InterfaceMethodref_index].mb;
			ClassClass *intf = imb->fb.clazz;
			unsigned offset = imb->fb.u.offset;
			struct methodblock *mb = &cbMethods(intf)[offset];
			TRACE(("\t%s %s%s\n", opnames[opcode],
			       mb->fb.name, mb->fb.signature));
		    }
#endif
		}
	    } END_WITH_OPCODE_CHANGE
	    goto check_for_exception;

	case opc_putfield_quick: {
	    JHandle *o = S_GET_HANDLE(-2);
	    if (o == 0)
	        JAVA_ERROR("NullPointerException", 0);
	    TRACE(("\t%s %s.(#%d) <= %lX\n", opnames[opcode],
		   Object2CString(o), pc[1], (long)S_POINTER(-1)));
	    obj_setslot(o, pc[1], (long)S_POINTER(-1));
	    SIZE_AND_STACK(3, -2);
	}

	case opc_getfield_quick: {
	    JHandle *o = S_GET_HANDLE(-1);
	    if (o == 0)
	        JAVA_ERROR("NullPointerException", 0);
	    S_POINTER(-1) = (void *)obj_getslot(o, pc[1]);
	    TRACE(("\t%s %s.(#%d) <= %lX\n", opnames[opcode],
		   Object2CString(o), pc[1], (long)S_POINTER(-1)));
	    SIZE_AND_STACK(3, 0);
	}

	case opc_putfield2_quick: {
	    JHandle *o = S_GET_HANDLE(-3);
	    if (o == 0)
	        JAVA_ERROR("NullPointerException", 0);
	    TRACE(("\t%s %s.(#%d) <= long=%s double=%g\n", opnames[opcode],
		   Object2CString(o), pc[1], S_LONGSTRING(-2), S_DOUBLE2(-2)));
	    obj_setslot(o, pc[1], (long)S_POINTER(-2));
	    obj_setslot(o, pc[1] + 1, (long)S_POINTER(-1));
	    SIZE_AND_STACK(3, -3);
	}

	case opc_getfield2_quick: {
	    JHandle *o = S_GET_HANDLE(-1);
	    if (o == 0)
		JAVA_ERROR("NullPointerException", 0);
	    S_POINTER(-1) = (void *)obj_getslot(o, pc[1]);
	    S_POINTER(0) = (void *)obj_getslot(o, pc[1] + 1);
	    TRACE(("\t%s %s.(#%d) <= long=%s double=%g\n", opnames[opcode],
		   Object2CString(o), pc[1], S_LONGSTRING(-1), S_DOUBLE2(-1)));
	    SIZE_AND_STACK(3, 1);
	}


	case opc_putstatic_quick: {
	    struct fieldblock *fb = constant_pool[GET_INDEX(pc+1)].fb;
	    OBJECT *d = (OBJECT *)normal_static_address(fb);
#ifdef BREAKPTS
            if (watching_field_modification) {
                jvalue jval;
		char isig = fieldsig(fb)[0];
		if (isig == SIGNATURE_FLOAT) {
		    jval.f = S_FLOAT(-1);
		} else if (isig == SIGNATURE_CLASS || 
                           isig == SIGNATURE_ARRAY) {
		    jval.l = (jobject)S_OBJECT(-1); /* safe? */
		} else if (isig == SIGNATURE_BYTE) {
		    jval.b = S_INT(-1);
		} else if (isig == SIGNATURE_CHAR) {
		    jval.c = S_INT(-1);
		} else if (isig == SIGNATURE_SHORT) {
		    jval.s = S_INT(-1);
                } else if (isig == SIGNATURE_BOOLEAN) {
                    jval.z = S_INT(-1);
		} else {
		    jval.i = S_INT(-1);
                }
                notify_debugger_of_field_modification(ee, NULL, fb, jval);
            }
#endif /* BREAKPTS */
	    *d = S_OBJECT(-1);
#ifdef TRACING
	    if (trace) {
		char isig = fieldsig(fb)[0];
		if (isig == SIGNATURE_FLOAT) {
		    TRACE(("\t%s %s.%s <= %g\n", opnames[opcode],
			   cbName(fieldclass(fb)), fieldname(fb),
			   S_FLOAT(-1)));
		} else if (isig == SIGNATURE_CLASS || isig == SIGNATURE_ARRAY) {
		    TRACE(("\t%s %s.%s <= %s\n", opnames[opcode],
			   cbName(fieldclass(fb)), fieldname(fb),
			   S_OSTRING(-1)));
		} else {
		    TRACE(("\t%s %s.%s <= %d\n", opnames[opcode],
			   cbName(fieldclass(fb)), fieldname(fb),
			   S_INT(-1)));
		}
	    }
#endif /* TRACING */
	    SIZE_AND_STACK(3, -1);
	}

	case opc_getstatic_quick: {
	    struct fieldblock *fb = constant_pool[GET_INDEX(pc+1)].fb;
	    OBJECT *d = (OBJECT *) normal_static_address(fb);
#ifdef BREAKPTS
            if (watching_field_access) {
                notify_debugger_of_field_access(ee, NULL, fb);
            }
#endif /* BREAKPTS */
	    S_OBJECT(0) = *d;
#ifdef TRACING
	    if (trace) {
		char isig = fieldsig(fb)[0];
		if (isig == SIGNATURE_FLOAT) {
		    TRACE(("\t%s %s.%s <= %g\n", opnames[opcode],
			   cbName(fieldclass(fb)), fieldname(fb),
			   S_FLOAT(0)));
		} else if (isig == SIGNATURE_CLASS || isig == SIGNATURE_ARRAY) {
		    TRACE(("\t%s %s.%s <= %s\n", opnames[opcode],
			   cbName(fieldclass(fb)), fieldname(fb),
			   S_OSTRING(0)));
		} else {
		    TRACE(("\t%s %s.%s <= %d\n", opnames[opcode],
			   cbName(fieldclass(fb)), fieldname(fb),
			   S_INT(0)));
		}
	    }
#endif /* TRACING */
	    SIZE_AND_STACK(3, 1);
	}

	case opc_putstatic2_quick: {
	    struct fieldblock *fb = constant_pool[GET_INDEX(pc+1)].fb;
	    stack_item *location = (stack_item *)twoword_static_address(fb);
#ifdef BREAKPTS
            if (watching_field_modification) {
                jvalue jval;
		char isig = fieldsig(fb)[0];
		if (isig == SIGNATURE_LONG) {
		    jval.j = S_LONG(-2);
		} else {
		    jval.d = S_DOUBLE(-2);
                }
                notify_debugger_of_field_modification(ee, NULL, fb, jval);
            }
#endif /* BREAKPTS */
	    location[0] = S_INFO(-2);
	    location[1] = S_INFO(-1);
#ifdef TRACING
	    if (trace) {
		char *signature = fieldsig(fb);
		if (signature[0] == SIGNATURE_LONG) {
		    TRACE(("\t%s %s.%s <= %s\n", opnames[opcode],
			   cbName(fieldclass(fb)), fieldname(fb),
			   S_LONGSTRING(-2)));
		} else {
		    TRACE(("\t%s %s.%s <= %g\n", opnames[opcode],
			   cbName(fieldclass(fb)), fieldname(fb),
			   S_DOUBLE(-2)));
		}
	    }
#endif /* TRACING */
	    SIZE_AND_STACK(3, -2);
	}

	case opc_putfield_quick_w: {
	    struct fieldblock *fb = constant_pool[GET_INDEX(pc+1)].fb;
	    char isig = fieldsig(fb)[0];
	    unsigned int offset = fb->u.offset / sizeof (OBJECT);
	    if (isig == SIGNATURE_LONG || isig == SIGNATURE_DOUBLE) {
		JHandle *o = S_GET_HANDLE(-3);
		if (o == 0)
		    JAVA_ERROR("NullPointerException", 0);
#ifdef BREAKPTS
                if (watching_field_modification) {
                    jvalue jval;
                    if (isig == SIGNATURE_LONG) {
                        jval.j = S_LONG(-2);
                    } else {
                        jval.d = S_DOUBLE(-2);
                    }
                    notify_debugger_of_field_modification(ee, o, fb, jval);
                }
#endif /* BREAKPTS */
#ifdef TRACING
		if (trace) {
		    if (isig == SIGNATURE_LONG) {
			TRACE(("\t%s %s.%s <= %s\n", opnames[opcode],
			       Object2CString(o), fieldname(fb),
			       S_LONGSTRING(-2)));
		    } else {
			TRACE(("\t%s %s.%s <= %g\n", opnames[opcode],
			       Object2CString(o), fieldname(fb),
			       S_DOUBLE(-2)));
		    }
		}
#endif /* TRACING */
		obj_setslot(o, offset, (long)S_POINTER(-2));
		obj_setslot(o, offset + 1, (long)S_POINTER(-1));
		SIZE_AND_STACK(3, -3);
	    } else {
		JHandle *o = S_GET_HANDLE(-2);
		if (o == 0)
		    JAVA_ERROR("NullPointerException", 0);
#ifdef BREAKPTS
                if (watching_field_modification) {
                    jvalue jval;
                    if (isig == SIGNATURE_FLOAT) {
                        jval.f = S_FLOAT(-1);
                    } else if (isig == SIGNATURE_CLASS || 
                               isig == SIGNATURE_ARRAY) {
                        jval.l = (jobject)S_OBJECT(-1); /* safe? */
                    } else if (isig == SIGNATURE_BYTE) {
                        jval.b = S_INT(-1);
                    } else if (isig == SIGNATURE_CHAR) {
                        jval.c = S_INT(-1);
                    } else if (isig == SIGNATURE_SHORT) {
                        jval.s = S_INT(-1);
                    } else if (isig == SIGNATURE_BOOLEAN) {
                        jval.z = S_INT(-1);
                    } else {
                        jval.i = S_INT(-1);
                    }
                    notify_debugger_of_field_modification(ee, o, fb, jval);
                }
#endif /* BREAKPTS */
#ifdef TRACING
		if (trace) {
		    if (isig == SIGNATURE_FLOAT) {
			TRACE(("\t%s %s.%s <= %g\n", opnames[opcode],
			       Object2CString(o), fieldname(fb),
			       S_FLOAT(-1)));
		    } else if (isig == SIGNATURE_CLASS ||
			       isig == SIGNATURE_ARRAY) {
			TRACE(("\t%s %s.%s <= %s\n", opnames[opcode],
			       Object2CString(o), fieldname(fb),
			       S_OSTRING(-1)));
		    } else {
			TRACE(("\t%s %s.%s <= %d\n", opnames[opcode],
			       Object2CString(o), fieldname(fb),
			       S_INT(-1)));
		    }
		}
#endif /* TRACING */
		obj_setslot(o, offset, (long)S_POINTER(-1));
		SIZE_AND_STACK(3, -2);
	    }
	}

	case opc_getstatic2_quick: {
	    struct fieldblock *fb = constant_pool[GET_INDEX(pc+1)].fb;
	    stack_item *location = (stack_item *)twoword_static_address(fb);
#ifdef BREAKPTS
            if (watching_field_access) {
                notify_debugger_of_field_access(ee, NULL, fb);
            }
#endif /* BREAKPTS */
	    S_INFO(0) = location[0];
	    S_INFO(1) = location[1];
#ifdef TRACING
	    if (trace) {
		char *signature = fieldsig(fb);
		if (signature[0] == SIGNATURE_LONG) {
		    TRACE(("\t%s %s.%s <= %s\n", opnames[opcode],
			   cbName(fieldclass(fb)), fieldname(fb),
			   S_LONGSTRING(0)));
		} else {
		    TRACE(("\t%s %s.%s <= %g\n", opnames[opcode],
			   cbName(fieldclass(fb)), fieldname(fb),
			   S_DOUBLE(0)));
		}
	    }
#endif /* TRACING */
	    SIZE_AND_STACK(3, 2);
	}

	case opc_getfield_quick_w: {
	    struct fieldblock *fb = constant_pool[GET_INDEX(pc+1)].fb;
	    char isig = fieldsig(fb)[0];
	    unsigned int offset = fb->u.offset / sizeof (OBJECT);
	    JHandle *o = S_GET_HANDLE(-1);
	    if (o == 0)
		JAVA_ERROR("NullPointerException", 0);
#ifdef BREAKPTS
            if (watching_field_access) {
                notify_debugger_of_field_access(ee, o, fb);
            }
#endif /* BREAKPTS */
	    if (isig == SIGNATURE_LONG || isig == SIGNATURE_DOUBLE) {
		S_POINTER(-1) = (void *)obj_getslot(o, offset);
		S_POINTER(0) = (void *)obj_getslot(o, offset + 1);
#ifdef TRACING
		if (trace) {
		    if (isig == SIGNATURE_LONG) {
			TRACE(("\t%s %s.%s <= %s\n", opnames[opcode],
			       Object2CString(o), fieldname(fb),
			       S_LONGSTRING(-1)));
		    } else {
			TRACE(("\t%s %s.%s <= %g\n", opnames[opcode],
			       Object2CString(o), fieldname(fb),
			       S_DOUBLE(-1)));
		    }
		}
#endif /* TRACING */
		SIZE_AND_STACK(3, 1);
	    } else {
		S_POINTER(-1) = (void *)obj_getslot(o, offset);
#ifdef TRACING
		if (trace) {
		    if (isig == SIGNATURE_FLOAT) {
			TRACE(("\t%s %s.%s <= %g\n", opnames[opcode],
			       Object2CString(o), fieldname(fb),
			       S_FLOAT(-1)));
		    } else if (isig == SIGNATURE_CLASS ||
			       isig == SIGNATURE_ARRAY) {
			TRACE(("\t%s %s.%s <= %s\n", opnames[opcode],
			       Object2CString(o), fieldname(fb),
			       S_OSTRING(-1)));
		    } else {
			TRACE(("\t%s %s.%s <= %d\n", opnames[opcode],
			       Object2CString(o), fieldname(fb),
			       S_INT(-1)));
		    }
		}
#endif /* TRACING */
		SIZE_AND_STACK(3, 0);
	    }
	}

	case opc_invokenonvirtual_quick:
	    mb = constant_pool[GET_INDEX(pc + 1)].mb;
	    args_size = mb->args_size;
	    optop -= args_size;
	    frame->returnpc = pc + 3;
	    o = S_GET_HANDLE(0);
	    if (o == 0)
	        JAVA_ERROR("NullPointerException", 0);
	    goto callmethod;

	case opc_invokesuper_quick: {
	    ClassClass *currentClass = fieldclass(&frame->current_method->fb);
	    ClassClass *superClass = cbSuperclass(currentClass);
	    int offset = GET_INDEX(pc + 1);
	    mb = cbMethodTable(superClass)->methods[offset];
	    args_size = mb->args_size;
	    optop -= args_size;
	    frame->returnpc = pc + 3;
	    o = S_GET_HANDLE(0);
	    if (o == 0)
	        JAVA_ERROR("NullPointerException", 0);
	    goto callmethod;
	}

	case opc_invokevirtual_quick_w:
	    mb = constant_pool[GET_INDEX(pc + 1)].mb;
	    args_size = mb->args_size;
	    optop -= args_size;
	    frame->returnpc = pc + 3;
	    o = S_GET_HANDLE(0);
	    if (o == 0)
	        JAVA_ERROR("NullPointerException", 0);
	    mb = mt_slot(obj_array_methodtable(o), mb->fb.u.offset);
	    goto callmethod;

	case opc_invokestatic_quick:
	    mb = constant_pool[GET_INDEX(pc + 1)].mb;
	    args_size = mb->args_size;
	    optop -= args_size;
	    frame->returnpc = pc + 3;
	    /* o should be the handle to the method's class */
	    o = (JHandle *)(cbHandle(fieldclass(&mb->fb)));
	    goto callmethod;

	case opc_invokeinterface_quick: {
	    struct methodblock *imb = constant_pool[GET_INDEX(pc + 1)].mb;
	    ClassClass *intf = imb->fb.clazz;
	    unsigned offset = imb->fb.u.offset;
	    int guess = pc[4];
	    ClassClass *cb;
	    struct methodtable *mtable;
	    struct imethodtable *imtable;

	    args_size = pc[3];
	    optop -= args_size;
	    frame->returnpc = pc + 5;
	    o = S_GET_HANDLE(0);
	    if (o == 0)
	        JAVA_ERROR("NullPointerException", 0);

	    /* Currently, arrays don't implement any interfaces that have
	     * methods in them.  But this is cheap, and we may change our
	     * minds some time in the future. */
	    if (obj_flags(o) == T_NORMAL_OBJECT) {
		mtable = obj_methodtable(o);
		cb = mtable->classdescriptor;
	    } else {
		/* This never happens--arrays don't support interface calls! */
		cb = classJavaLangObject;
		mtable = cbMethodTable(cb);
	    }
	    imtable = cbIntfMethodTable(cb);
	    if (guess < 0 || guess >= imtable->icount
		   || imtable->itable[guess].classdescriptor != intf) {
		for (guess = imtable->icount - 1; ; guess--) {
		    if (guess < 0) {
			invokeInterfaceError(ee, pc, cb, intf);
			goto handle_exception;
		    }
		    if (imtable->itable[guess].classdescriptor == intf) {
			pc[4] = guess;
			break;
		    }
		}
	    }
	    mb = mt_slot(mtable, imtable->itable[guess].offsets[offset]);
	    goto callmethod;
	}

	case opc_invokevirtualobject_quick:
	    frame->returnpc = pc + 3;
	    args_size = pc[2];
	    optop -= args_size;
	    o = S_GET_HANDLE(0);
	    if (o == 0)
	        JAVA_ERROR("NullPointerException", 0);
	    mb = mt_slot(obj_array_methodtable(o), pc[1]);
	    goto callmethod;

	case opc_invokevirtual_quick:
	    frame->returnpc = pc + 3;
	    args_size = pc[2];
	    optop -= args_size;
	    o = S_GET_HANDLE(0);
	    if (o == 0)
	        JAVA_ERROR("NullPointerException", 0);
	    mb = mt_slot(obj_methodtable(o), pc[1]);

	callmethod:
	    sysAssert(o != 0);
	    TRACE(("\t%s %s.%s%s (%d)\n", opnames[opcode],
		   cbName(fieldclass(&mb->fb)), fieldname(&mb->fb),
		   fieldsig(&mb->fb), args_size));
	    TRACE_METHOD(ee, mb, args_size, TRACE_METHOD_ENTER);

#ifndef BREAKPTS		/* don't do it twice if BREAKPTS is defined */
	    frame->lastpc = pc;
#endif

	    DECACHE_OPTOP();
	    {
		bool_t result = mb->invoker(o, mb, args_size, ee);
#ifdef JCOV
		if (mb->fb.access & ACC_NATIVE)
    		    coverage_native_method(mb);
#endif

		frame = ee->current_frame;
		if ((pc = frame->returnpc) == 0) {
		    FixupQuickInvocation(ee, frame);
		    pc = frame->lastpc + 3;
		}
#ifdef BREAKPTS
		/* Record any cleared exception from witin a native
                 * method call.
                 */
		if (debugging && 
                    (mb->fb.access & ACC_NATIVE) &&
		    ! exceptionOccurred(ee)) {
		    notify_debugger_of_exception_catch(ee, pc, NULL);
		}
#endif
		if (!result) {
		    /* FALSE indicates an exception might have occurred */
		    pc = frame->lastpc;
		    goto check_for_exception;
		}
	    }
	    vars = frame->vars;
	    constant_pool = frame->constant_pool;
	    CACHE_OPTOP();
	    SIZE_AND_STACK(0, 0);

	case opc_instanceof:
	    WITH_OPCODE_CHANGE
	        ResolveClassConstant(constant_pool, GET_INDEX(pc + 1), ee,
				     1 << CONSTANT_Class);
	        if (!exceptionOccurred(ee))
		    pc[0] = opc_instanceof_quick;
	    END_WITH_OPCODE_CHANGE
	    goto check_for_exception;

	case opc_instanceof_quick: {
	    ClassClass *cb = constant_pool[GET_INDEX(pc + 1)].clazz;
	    JHandle *h = S_GET_HANDLE(-1);
	    S_INT(-1) = ((h != NULL) && is_instance_of(h, cb, ee));
	    TRACE(("\tinstanceof %s => %d\n", cbName(cb), S_INT(-1)));
	    SIZE_AND_STACK_BUT(3, 0); /* may error on implementations */
	    goto check_for_exception;
	}

	case opc_checkcast:
	    WITH_OPCODE_CHANGE
	        ResolveClassConstant(constant_pool, GET_INDEX(pc + 1), ee,
				     1 << CONSTANT_Class);
	        if (!exceptionOccurred(ee))
	            pc[0] = opc_checkcast_quick;
	    END_WITH_OPCODE_CHANGE
	    goto check_for_exception;

	case opc_checkcast_quick: {
	    JHandle *h = S_GET_HANDLE(-1);
	    ClassClass *cb = constant_pool[GET_INDEX(pc + 1)].clazz;
	    if (!is_instance_of(h, cb, ee)) {
	        /* Checkcast fails  */
		char buf[255];
		JAVA_ERROR("ClassCastException",
		      	   classname2string(cbName(obj_array_classblock(h)),
				            buf, sizeof(buf)));
	    }
	    TRACE(("\tcheckcast %s\n", cbName(cb)));
	    SIZE_AND_STACK(3, 0);
	}

	case opc_new:
	    WITH_OPCODE_CHANGE
	        if (ResolveClassConstant(constant_pool, GET_INDEX(pc + 1), ee,
					 1 << CONSTANT_Class)) {
		    ClassClass *cb = constant_pool[GET_INDEX(pc + 1)].clazz;
		    if (cbAccess(cb) & (ACC_INTERFACE | ACC_ABSTRACT))
			JAVA_ERROR("InstantiationError", cbName(cb));
		    InitClass(cb);
		}
	        if (!exceptionOccurred(ee))
	            pc[0] = opc_new_quick;
	    END_WITH_OPCODE_CHANGE
	    goto check_for_exception;

	case opc_new_quick: {
	    ClassClass *cb = constant_pool[GET_INDEX(pc + 1)].clazz;
	    HObject *ret;
#ifdef HPROF
	    /* So that the heap profiler can determine the line number of
	     * the most recent frame 
	     */
	    frame->lastpc = pc;
#endif /* HPROF */
	    ret = allocObject(ee, cb);
	    if (ret == 0) {
#ifndef HPROF
		frame->lastpc = pc;
#endif /* HPROF */
		JAVA_ERROR("OutOfMemoryError", 0);
	    }
	    S_SET_HANDLE(0, ret);
	    TRACE(("\tnew %s => %s\n", cbName(cb), S_OSTRING(0)));
	    SIZE_AND_STACK_BUT(3, 1);
	    goto check_for_exception;
	}

	case opc_anewarray:
	    WITH_OPCODE_CHANGE
	        ResolveClassConstant(constant_pool, GET_INDEX(pc + 1), ee,
				     1 << CONSTANT_Class);
	        if (!exceptionOccurred(ee))
		    pc[0] = opc_anewarray_quick;
	    END_WITH_OPCODE_CHANGE
	    goto check_for_exception;

	case opc_anewarray_quick: {
	    int size = S_INT(-1);
	    HArrayOfObject *ret;
	    if (size < 0)
		JAVA_ERROR("NegativeArraySizeException", 0);
#ifdef HPROF
	    /* So that the heap profiler can determine the line number of
	     * the most recent frame 
	     */
	    frame->lastpc = pc;
#endif /* HPROF */
	    ret = (HArrayOfObject *)allocArray(ee, T_CLASS, size);
	    if (ret == 0)
	        JAVA_ERROR("OutOfMemoryError", 0);
	    else {
		ClassClass *array_cb = constant_pool[GET_INDEX(pc + 1)].clazz;
		unhand(ret)->body[size] = (HObject*)array_cb;
		TRACE(("\tanewarray_quick %s[%d] => %s\n",
		       cbName(array_cb), size, Object2CString((JHandle*)ret)));
		S_SET_HANDLE(-1, (JHandle *)ret);
		SIZE_AND_STACK(3, 0);
	    }
	}

	case opc_multianewarray:
	    WITH_OPCODE_CHANGE
	       ResolveClassConstant(constant_pool, GET_INDEX(pc + 1), ee,
				    1 << CONSTANT_Class);
	       if (!exceptionOccurred(ee))
		   pc[0] = opc_multianewarray_quick;
	    END_WITH_OPCODE_CHANGE
	    goto check_for_exception;

	case opc_multianewarray_quick: {
	    ClassClass *array_cb = constant_pool[GET_INDEX(pc + 1)].clazz;
	    int dimensions = pc[3];
	    HObject *ret;
	    int i;

	    optop -= dimensions;
	    for (i = 0; i < dimensions; i++)
		if (S_INT(i) < 0)
		    JAVA_ERROR("NegativeArraySizeException", 0);
#ifdef HPROF
	    /* So that the heap profiler can determine the line number of
	     * the most recent frame 
	     */
	    frame->lastpc = pc;
#endif /* HPROF */
	    ret = MultiArrayAlloc(dimensions, array_cb, &S_INFO(0));
	    if (ret == 0)
		JAVA_ERROR("OutOfMemoryError", 0);
	    S_SET_HANDLE(0, ret);
	    TRACE(("\tmultianewarray_quick %s, dimensions %d => %s\n",
		       cbName(array_cb), dimensions,
		   Object2CString((JHandle*)ret)));
	    SIZE_AND_STACK(4, 1);
	}

	case opc_newarray: {
	    int size = S_INT(-1);
	    int type = pc[1];
	    JHandle *ret;
	    if (size < 0)
	        JAVA_ERROR("NegativeArraySizeException", 0);
#ifdef HPROF
	    /* So that the heap profiler can determine the line number of
	     * the most recent frame 
	     */
	    frame->lastpc = pc;
#endif /* HPROF */
	    ret = allocArray(ee, type, size);
	    if (ret == 0)
	        JAVA_ERROR("OutOfMemoryError", 0);
	    TRACE(("\tnewarray %d %d => %s\n",
		   size, sizearray(type, size), Object2CString(ret)));
	    S_SET_HANDLE(-1, ret);
	    SIZE_AND_STACK(2, 0);
	}

#define ARRAY_INTRO(type, indexOffset, arrayOffset)             \
	    int index = S_INT(indexOffset);                     \
	    H##type *arrh = (H##type *)(S_GET_HANDLE(arrayOffset)); \
	    type *arr;                                          \
	    if (arrh == 0)                                      \
	        JAVA_ERROR("NullPointerException", 0);          \
	    arr = unhand(arrh);                                 \
	    if (index < 0 || index >= obj_length(arrh))  {      \
                /* A 32-bit number is at most 10 digits */      \
		char buf[30];                                   \
		/* On (unlikely) overflow, quietly truncate */  \
		(void) jio_snprintf(buf, sizeof(buf), "%d", index); \
	        JAVA_ERROR("ArrayIndexOutOfBoundsException", buf);  \
	    }

#define ARRAY_LOAD_INTRO(type)   ARRAY_INTRO(type, -1, -2)

	case opc_iaload: {
	    ARRAY_LOAD_INTRO(ArrayOfInt);
	    S_INT(-2) = arr->body[index];
	    TRACE(("\tiaload ?[%d] => %d\n", index, S_INT(-2)));
	    SIZE_AND_STACK(1, -1);
	}

	case opc_laload: {
	    ARRAY_LOAD_INTRO(ArrayOfLong);
	    SET_S_LONG(-2, arr->body[index]);
	    TRACE(("\tlaload ?[%d] => %s\n", index, S_LONGSTRING(-2)));
	    SIZE_AND_STACK(1, 0);
	}

	case opc_faload: {
	    ARRAY_LOAD_INTRO(ArrayOfFloat);
	    S_FLOAT(-2) = arr->body[index];
	    TRACE(("\tfaload ?[%d] => %g\n", index, S_FLOAT(-2)));
	    SIZE_AND_STACK(1, -1);
	}

	case opc_daload: {
	    ARRAY_LOAD_INTRO(ArrayOfDouble);
	    SET_S_DOUBLE(-2, arr->body[index]);
	    TRACE(("\tdaload ?[%d] => %g\n", index, S_DOUBLE(-2)));
	    SIZE_AND_STACK(1, 0);
	}

	case opc_aaload: {
	    ARRAY_LOAD_INTRO(ArrayOfObject);
	    S_SET_HANDLE(-2, arr->body[index]);
	    TRACE(("\taaload ?[%d] => %s\n", index, S_OSTRING(-2)));
	    SIZE_AND_STACK(1, -1);
	}

	case opc_baload: {
	    ARRAY_LOAD_INTRO(ArrayOfByte);
	    S_INT(-2) = arr->body[index];
	    TRACE(("\tbaload ?[%d] => %d\n", index, S_INT(-2)));
	    SIZE_AND_STACK(1, -1);
	}

	case opc_caload: {
	    ARRAY_LOAD_INTRO(ArrayOfChar);
	    S_INT(-2) = arr->body[index];
	    TRACE(("\tcaload ?[%d] => %d\n", index, S_INT(-2)));
	    SIZE_AND_STACK(1, -1);
	}

	case opc_saload: {
	    ARRAY_LOAD_INTRO(ArrayOfShort);
	    S_INT(-2) = arr->body[index];
	    TRACE(("\tsaload ?[%d] => %d\n", index, S_INT(-2)));
	    SIZE_AND_STACK(1, -1);
	}

	case opc_iastore: {
	    ARRAY_INTRO(ArrayOfInt, -2, -3);
	    arr->body[index] = S_INT(-1);
	    TRACE(("\tiastore ?[%d] <= %d\n", index, S_INT(-1)));
	    SIZE_AND_STACK(1, -3);
	}

	case opc_lastore: {
	    ARRAY_INTRO(ArrayOfLong, -3, -4);
	    arr->body[index] = S_LONG(-2);
	    TRACE(("\tlastore ?[%d] <= %s\n", index, S_LONGSTRING(-2)));
	    SIZE_AND_STACK(1, -4);
	}

	case opc_fastore: {
	    ARRAY_INTRO(ArrayOfFloat, -2, -3);
	    arr->body[index] = S_FLOAT(-1);
	    TRACE(("\tfastore ?[%d] <= %f\n", index, S_FLOAT(-1)));
	    SIZE_AND_STACK(1, -3);
	}

	case opc_dastore: {
	    ARRAY_INTRO(ArrayOfDouble, -3, -4);
	    arr->body[index] = S_DOUBLE(-2);
	    TRACE(("\tdastore ?[%d] <= %g\n", index, S_DOUBLE(-2)));
	    SIZE_AND_STACK(1, -4);
	}

	case opc_bastore:{
	    ARRAY_INTRO(ArrayOfByte, -2, -3);
	    arr->body[index] = S_INT(-1);
	    TRACE(("\tbastore ?[%d] <= %d\n", index, S_INT(-1)));
	    SIZE_AND_STACK(1, -3);
	}

	case opc_castore:{
	    ARRAY_INTRO(ArrayOfChar, -2, -3);
	    arr->body[index] = S_INT(-1);
	    TRACE(("\tcastore ?[%d] <= %d\n", index, (unicode) S_INT(-1)));
	    SIZE_AND_STACK(1, -3);
	}

	case opc_sastore: {
	    ARRAY_INTRO(ArrayOfShort, -2, -3);
	    arr->body[index] = S_INT(-1);
	    TRACE(("\tsastore ?[%d] <= %d\n", index, (short) S_INT(-1)));
	    SIZE_AND_STACK(1, -3);
	}

	case opc_aastore: {
	    JHandle *value = S_GET_HANDLE(-1);
	    ClassClass *array_cb;
	    ARRAY_INTRO(ArrayOfObject, -2, -3);

	    array_cb = (ClassClass *)(arr->body[obj_length(arrh)]);
	    if (!is_instance_of(value, array_cb, ee)) {
		/* The value isn't null and doesn't match the type. */
		JAVA_ERROR("ArrayStoreException", 0);
	    }
	    arr->body[index] = value;
	    TRACE(("\taastore ?[%d] <= %s\n", index, Object2CString(value)));
	    SIZE_AND_STACK(1, -3);
	}

	case opc_monitorenter: {
	    JHandle *o = S_GET_HANDLE(-1);
	    TRACE(("%s(%s)\n", opnames[opcode], Object2CString(o)));
	    if (o == 0) {
	        JAVA_ERROR("NullPointerException", 0);
	    } else {
		monitorEnter2(ee, obj_monitor(o));
	    }
	    SIZE_AND_STACK(1, -1);
	}

	case opc_monitorexit: {
	    JHandle *o = S_GET_HANDLE(-1);
	    TRACE(("%s(%s)\n", opnames[opcode], Object2CString(o)));
	    if (o == 0) {
	        JAVA_ERROR("NullPointerException", 0);
	    } else {
		if (monitorExit2(ee, obj_monitor(o)) != 0)
		    /* IllegalMonitorStateException */ 
		    goto handle_exception;
	    }
	    SIZE_AND_STACK(1, -1);
	}

	case opc_breakpoint:
	    TRACE(("\tbreakpoint\n"));
#ifdef BREAKPTS
	    DECACHE_OPTOP();
	    opcode = get_breakpoint_opcode(ee, pc, !stopped_at_single_step);
	    goto again;
#else
	    SIZE_AND_STACK(1, 0);	/* treat as opc_nop */
#endif

	case opc_wide: {
	    int reg = GET_INDEX(pc + 2);
	    int offset;

	    switch(pc[1]) {
	        case opc_aload:
	        case opc_iload:
	        case opc_fload:
		    S_INFO(0) = vars[reg];
		    TRACEIF(opc_iload,
			      ("\tiload r%d => %d\n", reg, S_INT(0)));
		    TRACEIF(opc_aload,
			    ("\taload r%d => %s\n", reg, S_OSTRING(0)));
		    TRACEIF(opc_fload,
			    ("\tfload r%d => %g\n", reg, S_FLOAT(0)));
		    SIZE_AND_STACK(4, 1);
	        case opc_lload:
	        case opc_dload:
		    S_INFO(0) = vars[reg];
		    S_INFO(1) = vars[reg + 1];
		    TRACEIF(opc_lload,
			    ("\tlload r%d => %s\n", reg, S_LONGSTRING(0)));
		    TRACEIF(opc_dload,
			    ("\tdload r%d => %g\n", reg, S_DOUBLE(0)));
		    SIZE_AND_STACK(4, 2);
	        case opc_istore:
	        case opc_astore:
	        case opc_fstore:
		    vars[reg] = S_INFO(-1);
		    TRACEIF(opc_istore,
			    ("\tistore %d => r%d\n", S_INT(-1), reg));
		    TRACEIF(opc_astore,
			    ("\tastore %s => r%d\n", S_OSTRING(-1), reg));
		    TRACEIF(opc_fstore,
			    ("\tfstore %g => r%d\n", S_FLOAT(-1), reg));
		    SIZE_AND_STACK(4, -1);
	        case opc_lstore:
	        case opc_dstore:
		    vars[reg] = S_INFO(-2);
		    vars[reg + 1] = S_INFO(-1);
		    TRACEIF(opc_lstore,
			    ("\tlstore %s => r%d\n", S_LONGSTRING(-2), reg));
		    TRACEIF(opc_dstore,
			    ("\tdstore %g => r%d\n", S_DOUBLE(-2), reg));
		    SIZE_AND_STACK(4, -2);
	        case opc_iinc:
                    offset = PC2SIGNEDSHORT((unsigned char *)(pc+3));  
		    vars[reg].i += offset;
		    TRACE(("\tiinc r%d+%d => %d\n", reg, offset, vars[reg].i));
		    SIZE_AND_STACK(6, 0);
	        case opc_ret:
		    TRACE(("\tret %d (%lX)\n", reg, (long)vars[reg].addr));
		    pc = vars[reg].addr;
		    SIZE_AND_STACK_BUT(0, 0);
		    goto check_for_exception;
	        default:
		    JAVA_ERROR("InternalError", "undefined opcode");
	    }
	}

        case opc_invokeignored_quick: /* 1.1 optimization.  */
	    args_size = pc[1];
	    TRACE(("\tinvokeignored_quick\n"));
	    optop -= args_size;
	    o = S_GET_HANDLE(0);
	    if (pc[2] != 0 && o == 0)
	        JAVA_ERROR("NullPointerException", 0);
	    SIZE_AND_STACK(3, 0);

	case opc_software:
	case opc_hardware:
	default:
	    JAVA_ERROR("InternalError", "undefined opcode");

	} /* end of switch */

      check_for_exception:
	if (!exceptionOccurred(ee)) {
	   continue;
	}

      handle_exception: {
	  /* We've got an exception. */
	  JHandle *object = ee->exception.exc;
	  unsigned char *new_pc;

#ifdef BREAKPTS
          if (debugging) {
              notify_debugger_of_exception(ee, pc, ee->exception.exc);
          }
#endif
	  exceptionClear(ee);
	  new_pc = ProcedureFindThrowTag(ee, frame, object, pc);
          /* ProcedureFindThrowTag might have thrown an exception. */
	  if (!exceptionOccurred(ee)) {
	      exceptionThrow(ee, object);
	  }

	  DECACHE_OPTOP();
	  if (new_pc != 0) {
	      /* Found the exception handler. */
	      exceptionClear(ee);
	      vars = frame->vars;
	      constant_pool = frame->constant_pool;
	      frame->optop = frame->ostack; /* Reset to empty */
	      CACHE_OPTOP();
	      /* Push exception object onto the stack. */
	      S_SET_HANDLE(0, object);
	      optop++;
	      pc = new_pc;
#ifdef BREAKPTS
              if (debugging) {
                  notify_debugger_of_exception_catch(ee, pc, object);
              }
#endif
#ifdef TRACING
	      if (trace) {
		  char where_am_i[100];
		  pc2string(pc, frame,
			    where_am_i, where_am_i + sizeof where_am_i);
		  jio_fprintf(stdout, "Catch at %s\n", where_am_i);
		  fflush(stdout);
	      }
#endif /* TRACING */
#ifdef JCOV
              fl_next_optop = FALSE;
#endif
	      continue;		/* continue with the outer loop */
	  }
      }

        /* Fall through */
finish_return:
	/* Exit from a monitor, if necessary. */
        if (frame->monitor) {
#ifndef FAST_MONITOR
            sysMonitorExit(EE2SysThread(ee), frame->monitor);
#else
            monitorExit2(ee, (uintptr_t) frame->monitor);
#endif
        }
	/* Note that if frame == initial_frame, the frame is only used
	 * for callbacks. We don't need to obtain its profiling data.
	 */
	if (frame == initial_frame) {
	    return !exceptionOccurred(ee);
	}
	/* At this point we know it must not be a pseudo or native
	 * frame.
	 */
	sysAssert(frame->current_method &&
		  (frame->current_method->fb.access & ACC_NATIVE) == 0);
#ifdef HPROF
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_METHOD_EXIT)) {
        jvmpi_method_exit(ee);
    }
#endif /* HPROF */
#ifdef BREAKPTS
        if (debugging) {
            notify_debugger_of_frame_pop(ee);
        }
#endif /* BREAKPTS */

#ifdef TRACING
	if (trace) {
	    char where_am_i[100];
	    pc2string(pc, frame,
		      where_am_i, where_am_i + sizeof where_am_i);
	    jio_fprintf(stdout, "Leaving %s\n", where_am_i);
	    fflush(stdout);
	}
#endif /* TRACING */
	{
	    TRACE_METHOD(ee, frame->current_method, args_size, TRACE_METHOD_RETURN);
	    frame = ee->current_frame = frame->prev;
	    if (frame->returnpc == 0) {
		FixupQuickInvocation(ee, frame);
		frame->returnpc = frame->lastpc + 3;
	    }
	    if (exceptionOccurred(ee)) {
		/* Go back, and look for an exception handler. */
		pc = frame->lastpc;
	        goto handle_exception;
	    }
	    pc = frame->returnpc;
	    vars = frame->vars;
	    constant_pool = frame->constant_pool;
	    CACHE_OPTOP();
#ifdef JCOV
            fl_next_optop = FALSE;
#endif
	    continue;
	}
    } /* end of infinite loop */
}
