/*
 * @(#)inline.c	1.21 98/08/31
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Routines used to try to inline small, simple methods that only return 
 * constants or access fields. 
 */


#include "oobj.h"
#include "interpreter.h"
#include "bool.h"
#include "opcodes.h"
#include "tree.h"
#include "sys_api.h"

/*
 * Note that because the inliner and the verifier are now in different
 * DLLs, each DLL now has its own copy of the opcode_length table.
 */
#include "opcodes.length"

#define isXreturn(opcode) \
     ((opcode) >= opc_ireturn) && (((opcode) <= opc_areturn))

static unsigned long MethodInlining(struct methodblock *mb);

static unsigned long makeReturnResult(struct methodblock *mb, bool_t checkThis, 
				      long highReg, long opcodeArgCount);

static unsigned long makeOpcodeResult(struct methodblock *mb, bool_t checkThis, 
				      long highReg, long icount, long argCount,
				      unsigned char opcode, unsigned char op1, 
				      unsigned char op2);

static unsigned long makePoppingResult(struct methodblock *mb, bool_t checkThis,
				       long firstReg, long opcodeArgCount, 
				       long icount, unsigned char opcode, 
				       unsigned char op1, unsigned char op2);

#define NO_INLINE_FLAG          (0x1 << 24)
#define SAME_CLASS_FLAG          (0x1 << 25)
#define REDO_INLINING_FLAG (0x1 << 26)
#define NEEDS_TO_BE_QUICKED_FLAG (0x1 << 27)

#define INLINING_WORD1(simp) ((simp) & 0xFF)
#define INLINING_WORD2(simp) (((simp) >> 8) & 0xFF)
#define INLINING_WORD3(simp) (((simp) >> 16) & 0xFF)

#define MAKE_INLINING(op1, op2, op3) ((op1) << 0) + ((op2) << 8) + ((op3) << 16)



struct OpcodeInfoType { 
#ifdef DEBUG
    opcode_type opcode;
#else 
    unsigned char opcode;
#endif
    unsigned char inStack;
    unsigned char outStack;
#define CAN_ERROR 0x01		/* can give error in addition to NULL_CHECK */
#define NULL_CHECK 0x02		/* checks that first arg isn't null */
#define CONSTANT_POOL 0x04	/* uses the constant pool */
    unsigned char flags;
};
#define REGSIZE(ptr) (ptr->inStack)
#define REGNUM(ptr)  (ptr->outStack)

static struct OpcodeInfoType const OpcodeInfo[256] = 
{
    /*   { opc_pop  <number of words to pop from stack> } 
     *   { opc_iadd <words popped from stack> <words pushed to stack> }
     *   { opc_iload <words pushed to stack> } 
     *   { opc_iload_0 <words pushed to stack> <implicit register> } 
     *   { opc_return  <words returned> } 
     *     255 indicates opcode that we can't inline
     *   other values are special opcodes that must be handled specially
     */

    /* nop                  */   { opc_pop, 0 },
    /* aconst_null          */   { opc_iadd, 0, 1},
    /* iconst_m1            */   { opc_iadd, 0, 1},
    /* iconst_0             */   { opc_iadd, 0, 1},
    /* iconst_1             */   { opc_iadd, 0, 1},
    /* iconst_2             */   { opc_iadd, 0, 1},
    /* iconst_3             */   { opc_iadd, 0, 1},
    /* iconst_4             */   { opc_iadd, 0, 1},
    /* iconst_5             */   { opc_iadd, 0, 1},
    /* lconst_0             */   { opc_iadd, 0, 2},
    /* lconst_1             */   { opc_iadd, 0, 2},
    /* fconst_0             */   { opc_iadd, 0, 1},
    /* fconst_1             */   { opc_iadd, 0, 1},
    /* fconst_2             */   { opc_iadd, 0, 1},
    /* dconst_0             */   { opc_iadd, 0, 2},
    /* dconst_1             */   { opc_iadd, 0, 2},
    /* bipush               */   { opc_iadd, 0, 1},
    /* sipush               */   { opc_iadd, 0, 1},
    /* ldc                  */   { 255 },
    /* ldc_w                */   { 255 },
    /* ldc2_w               */   { 255 },
    /* iload                */   { opc_iload, 1},
    /* lload                */   { opc_iload, 2},
    /* fload                */   { opc_iload, 1},
    /* dload                */   { opc_iload, 2},
    /* aload                */   { opc_iload, 1},
    /* iload_0              */   { opc_iload_0, 1, 0},
    /* iload_1              */   { opc_iload_0, 1, 1},
    /* iload_2              */   { opc_iload_0, 1, 2},
    /* iload_3              */   { opc_iload_0, 1, 3},
    /* lload_0              */   { opc_iload_0, 2, 0},
    /* lload_1              */   { opc_iload_0, 2, 1},
    /* lload_2              */   { opc_iload_0, 2, 2},
    /* lload_3              */   { opc_iload_0, 2, 3},
    /* fload_0              */   { opc_iload_0, 1, 0},
    /* fload_1              */   { opc_iload_0, 1, 1},
    /* fload_2              */   { opc_iload_0, 1, 2},
    /* fload_3              */   { opc_iload_0, 1, 3},
    /* dload_0              */   { opc_iload_0, 2, 0},
    /* dload_1              */   { opc_iload_0, 2, 1},
    /* dload_2              */   { opc_iload_0, 2, 2},
    /* dload_3              */   { opc_iload_0, 2, 3},
    /* aload_0              */   { opc_iload_0, 1, 0},
    /* aload_1              */   { opc_iload_0, 1, 1},
    /* aload_2              */   { opc_iload_0, 1, 2},
    /* aload_3              */   { opc_iload_0, 1, 3},
    /* iaload               */   { opc_iadd, 2, 1, NULL_CHECK},
    /* laload               */   { opc_iadd, 2, 2, NULL_CHECK},
    /* faload               */   { opc_iadd, 2, 1, NULL_CHECK},
    /* daload               */   { opc_iadd, 2, 2, NULL_CHECK},
    /* aaload               */   { opc_iadd, 2, 1, NULL_CHECK},
    /* baload               */   { opc_iadd, 2, 1, NULL_CHECK},
    /* caload               */   { opc_iadd, 2, 1, NULL_CHECK},
    /* saload               */   { opc_iadd, 2, 1, NULL_CHECK},
    /* istore               */   { 255 },
    /* lstore               */   { 255 },
    /* fstore               */   { 255 },
    /* dstore               */   { 255 },
    /* astore               */   { 255 },
    /* istore_0             */   { 255 },
    /* istore_1             */   { 255 },
    /* istore_2             */   { 255 },
    /* istore_3             */   { 255 },
    /* lstore_0             */   { 255 },
    /* lstore_1             */   { 255 },
    /* lstore_2             */   { 255 },
    /* lstore_3             */   { 255 },
    /* fstore_0             */   { 255 },
    /* fstore_1             */   { 255 },
    /* fstore_2             */   { 255 },
    /* fstore_3             */   { 255 },
    /* dstore_0             */   { 255 },
    /* dstore_1             */   { 255 },
    /* dstore_2             */   { 255 },
    /* dstore_3             */   { 255 },
    /* astore_0             */   { 255 },
    /* astore_1             */   { 255 },
    /* astore_2             */   { 255 },
    /* astore_3             */   { 255 },
    /* iastore              */   { opc_iadd, 3, 0, NULL_CHECK},
    /* lastore              */   { opc_iadd, 4, 0, NULL_CHECK},
    /* fastore              */   { opc_iadd, 3, 0, NULL_CHECK},
    /* dastore              */   { opc_iadd, 4, 0, NULL_CHECK},
    /* aastore              */   { opc_iadd, 3, 0, NULL_CHECK},
    /* bastore              */   { opc_iadd, 3, 0, NULL_CHECK},
    /* castore              */   { opc_iadd, 3, 0, NULL_CHECK},
    /* sastore              */   { opc_iadd, 3, 0, NULL_CHECK},
    /* pop                  */   { opc_pop, 1},
    /* pop2                 */   { opc_pop, 2},
    /* dup                  */   { 255 },
    /* dup_x1               */   { 255 },
    /* dup_x2               */   { 255 },
    /* dup2                 */   { 255 },
    /* dup2_x1              */   { 255 },
    /* dup2_x2              */   { 255 },
    /* swap                 */   { 255 },
    /* iadd                 */   { opc_iadd, 2, 1},
    /* ladd                 */   { opc_iadd, 4, 2},
    /* fadd                 */   { opc_iadd, 2, 1},
    /* dadd                 */   { opc_iadd, 4, 2},
    /* isub                 */   { opc_iadd, 2, 1},
    /* lsub                 */   { opc_iadd, 4, 2},
    /* fsub                 */   { opc_iadd, 2, 1},
    /* dsub                 */   { opc_iadd, 4, 2},
    /* imul                 */   { opc_iadd, 2, 1},
    /* lmul                 */   { opc_iadd, 4, 2},
    /* fmul                 */   { opc_iadd, 2, 1},
    /* dmul                 */   { opc_iadd, 4, 2},
    /* idiv                 */   { opc_iadd, 2, 1, CAN_ERROR},
    /* ldiv                 */   { opc_iadd, 4, 2, CAN_ERROR},
    /* fdiv                 */   { opc_iadd, 2, 1},
    /* ddiv                 */   { opc_iadd, 4, 2},
    /* irem                 */   { opc_iadd, 2, 1, CAN_ERROR},
    /* lrem                 */   { opc_iadd, 4, 2, CAN_ERROR},
    /* frem                 */   { opc_iadd, 2, 1},
    /* drem                 */   { opc_iadd, 4, 2},
    /* ineg                 */   { opc_iadd, 1, 1},
    /* lneg                 */   { opc_iadd, 2, 2},
    /* fneg                 */   { opc_iadd, 1, 1},
    /* dneg                 */   { opc_iadd, 2, 2},
    /* ishl                 */   { opc_iadd, 2, 1},
    /* lshl                 */   { opc_iadd, 3, 2},
    /* ishr                 */   { opc_iadd, 2, 1},
    /* lshr                 */   { opc_iadd, 3, 2},
    /* iushr                */   { opc_iadd, 2, 1},
    /* lushr                */   { opc_iadd, 3, 2},
    /* iand                 */   { opc_iadd, 2, 1},
    /* land                 */   { opc_iadd, 4, 2},
    /* ior                  */   { opc_iadd, 2, 1},
    /* lor                  */   { opc_iadd, 4, 2},
    /* ixor                 */   { opc_iadd, 2, 1},
    /* lxor                 */   { opc_iadd, 4, 2},
    /* iinc                 */   { 255 },
    /* i2l                  */   { opc_iadd, 1, 2},
    /* i2f                  */   { opc_iadd, 1, 1},
    /* i2d                  */   { opc_iadd, 1, 2},
    /* l2i                  */   { opc_iadd, 2, 1},
    /* l2f                  */   { opc_iadd, 2, 1},
    /* l2d                  */   { opc_iadd, 2, 2},
    /* f2i                  */   { opc_iadd, 1, 1},
    /* f2l                  */   { opc_iadd, 1, 2},
    /* f2d                  */   { opc_iadd, 1, 2},
    /* d2i                  */   { opc_iadd, 2, 1},
    /* d2l                  */   { opc_iadd, 2, 2},
    /* d2f                  */   { opc_iadd, 2, 1},
    /* i2b                  */   { opc_iadd, 1, 1},
    /* i2c                  */   { opc_iadd, 1, 1},
    /* i2s                  */   { opc_iadd, 1, 1},
    /* lcmp                 */   { opc_iadd, 4, 1},
    /* fcmpl                */   { opc_iadd, 2, 1},
    /* fcmpg                */   { opc_iadd, 2, 1},
    /* dcmpl                */   { opc_iadd, 4, 1},
    /* dcmpg                */   { opc_iadd, 4, 1},
    /* ifeq                 */   { 255 },
    /* ifne                 */   { 255 },
    /* iflt                 */   { 255 },
    /* ifge                 */   { 255 },
    /* ifgt                 */   { 255 },
    /* ifle                 */   { 255 },
    /* if_icmpeq            */   { 255 },
    /* if_icmpne            */   { 255 },
    /* if_icmplt            */   { 255 },
    /* if_icmpge            */   { 255 },
    /* if_icmpgt            */   { 255 },
    /* if_icmple            */   { 255 },
    /* if_acmpeq            */   { 255 },
    /* if_acmpne            */   { 255 },
    /* goto                 */   { 255 },
    /* jsr                  */   { 255 },
    /* ret                  */   { 255 },
    /* tableswitch          */   { 255 },
    /* lookupswitch         */   { 255 },
    /* ireturn              */   { opc_return, 1},
    /* lreturn              */   { opc_return, 2},
    /* freturn              */   { opc_return, 1},
    /* dreturn              */   { opc_return, 2},
    /* areturn              */   { opc_return, 1},
    /* return               */   { opc_return, 0},
    /* getstatic            */   { 254 },
    /* putstatic            */   { 254 },
    /* getfield             */   { 254 },
    /* putfield             */   { 254 },
    /* invokevirtual        */   { 254 },
    /* invokespecial        */   { 254 },
    /* invokestatic         */   { 254 },
    /* invokeinterface      */   { 254 },
    /* xxxunusedxxx         */   { 255 },
    /* new                  */   { 255 },
    /* newarray             */   { opc_iadd, 1, 1, CAN_ERROR},
    /* anewarray            */   { 255 },
    /* arraylength          */   { opc_iadd, 1, 1, NULL_CHECK },
    /* athrow               */   { opc_iadd, 1, 0, NULL_CHECK | CAN_ERROR},
    /* checkcast            */   { 255 },
    /* instanceof           */   { 255 },
    /* monitorenter         */   { opc_iadd, 1, 0, NULL_CHECK | CAN_ERROR},
    /* monitorexit          */   { opc_iadd, 1, 0, NULL_CHECK | CAN_ERROR},
    /* wide                 */   { 255 },
    /* multianewarray       */   { 255 },
    /* ifnull               */   { 255 },
    /* ifnonnull            */   { 255 },
    /* goto_w               */   { 255 },
    /* jsr_w                */   { 255 },
    /* breakpoint           */   { 255 },
    /* ldc_quick            */   { opc_iadd, 0, 1, CONSTANT_POOL},
    /* ldc_w_quick          */   { opc_iadd, 0, 1, CONSTANT_POOL},
    /* ldc2_w_quick         */   { opc_iadd, 0, 2, CONSTANT_POOL},
    /* getfield_quick       */   { opc_iadd, 1, 1, NULL_CHECK},
    /* putfield_quick       */   { opc_iadd, 2, 0, NULL_CHECK},
    /* getfield2_quick      */   { opc_iadd, 1, 2, NULL_CHECK},
    /* putfield2_quick      */   { opc_iadd, 3, 0, NULL_CHECK},
    /* getstatic_quick      */   { opc_iadd, 0, 1, CONSTANT_POOL},
    /* putstatic_quick      */   { opc_iadd, 1, 0, CONSTANT_POOL},
    /* getstatic2_quick     */   { opc_iadd, 0, 2, CONSTANT_POOL},
    /* putstatic2_quick     */   { opc_iadd, 2, 0, CONSTANT_POOL},
    /* invokevirtual_quick  */   { 255 },
    /* i.kenonvirtual_quick */   { 255 },
    /* i.kesuper_quick      */   { 255 },
    /* invokestatic_quick   */   { 255 },
    /* invokeinterface_quick*/   { 255 },
    /* i.evirtualobj_quick  */   { 255 },
    /* invokeignored_quick  */   { opc_invokeignored_quick },
    /* new_quick            */   { opc_iadd, 0, 1, CAN_ERROR | CONSTANT_POOL},
    /* anewarray_quick      */   { opc_iadd, 1, 1, CAN_ERROR | CONSTANT_POOL},
    /* multianewarray_quick */   { 255 },
    /* checkcast_quick      */   { opc_iadd, 1, 1, CAN_ERROR | CONSTANT_POOL},
    /* instanceof_quick     */   { opc_iadd, 1, 1, CONSTANT_POOL},
    /* invokevirtual_quick_w*/   { 255 },
    /* getfield_quick_w     */   { opc_getfield_quick_w},
    /* putfield_quick_w     */   { opc_getfield_quick_w},
    /* nonnull_quick        */   { opc_nonnull_quick},
    /* bytecode 231         */   { 255 },
    /* bytecode 232         */   { 255 },
    /* bytecode 233         */   { 255 },
    /* bytecode 233         */   { 255 },
    /* bytecode 234         */   { 255 },
    /* bytecode 235         */   { 255 },
    /* bytecode 236         */   { 255 },
    /* bytecode 237         */   { 255 },
    /* bytecode 238         */   { 255 },
    /* bytecode 239         */   { 255 },
    /* bytecode 240         */   { 255 },
    /* bytecode 241         */   { 255 },
    /* bytecode 242         */   { 255 },
    /* bytecode 243         */   { 255 },
    /* bytecode 244         */   { 255 },
    /* bytecode 245         */   { 255 },
    /* bytecode 246         */   { 255 },
    /* bytecode 247         */   { 255 },
    /* bytecode 248         */   { 255 },
    /* bytecode 249         */   { 255 },
    /* bytecode 250         */   { 255 },
    /* bytecode 251         */   { 255 },
    /* bytecode 252         */   { 255 },
    /* bytecode 253         */   { 255 },
    /* bytecode 254         */   { 255 },
    /* bytecode 255         */   { 255 },
};

static struct OpcodeInfoType const OpcodeInfoSpecial[] = 
{
    /* getfield_quick       */   { opc_iadd, 1, 1, NULL_CHECK | CONSTANT_POOL},
    /* putfield_quick       */   { opc_iadd, 2, 0, NULL_CHECK | CONSTANT_POOL},
    /* getfield2_quick      */   { opc_iadd, 1, 2, NULL_CHECK | CONSTANT_POOL},
    /* putfield2_quick      */   { opc_iadd, 3, 0, NULL_CHECK | CONSTANT_POOL}
};

/* This method is called to determine whether the method "mb" called at
 * from instruction "pc" can be inlined.  If not, the value -1 is returned.
 * If so, the value 0 is returned, and the three bytes which should
 * overwrite the method invocation are stored in result[].  If the method
 * might be able to be inlined after it's been executed once, 1 is returned.
 */
int
MethodCallInline(unsigned char *pc, 
		 struct methodblock *currentMethod, 
		 struct methodblock *mb,
		 unsigned char *result) { 
    opcode_type opcode = pc[0];

    unsigned long inlining;

    if (opcode == opc_invokevirtual_quick 
	    || opcode == opc_invokevirtual_quick_w)
	/* This is a virtual method call.  No use even bother trying to
	 * inline the method, unless its final
	 */
	if (((mb->fb.access & ACC_FINAL) == 0) 
	       && ((cbAccess(mb->fb.clazz) & ACC_FINAL) == 0))
	    return -1;

    /* Inline only if the caller and callee are declared to be strictfp.
     * We don't want to inline when both the caller and the callee are widefp,
     * because the floating point proposal could change to require rounding on
     * method parameters.  That would be taken care of by the invoke 
     * instructions, but there is no way to express a forced rounding in the
     * JVM's instruction set in widefp mode. 
     */
    if (!((mb->fb.access & ACC_STRICT) &&
	  (currentMethod->fb.access & ACC_STRICT)))
        return -1;
    
    inlining = mb->inlining;
    if ((inlining & NO_INLINE_FLAG) != 0)
	return -1;
    /* If we haven't already determined if the target method has a
     * inlining, then do that determination */
    if (   (inlining == 0) 
	|| ((inlining & REDO_INLINING_FLAG) 
	    && !UseLosslessQuickOpcodes)) {
	if (   (mb->fb.access & (ACC_ABSTRACT | ACC_NATIVE | ACC_SYNCHRONIZED))
	    || (mb->exception_table_length > 0)) {  
	    /* Methods with exception handlers, or that are abstract, native,
	       etc. can't be inlined.
	     */
	    mb->inlining = NO_INLINE_FLAG;
	    return -1;
	} 
	inlining = mb->inlining = MethodInlining(mb);
	if ((inlining & NO_INLINE_FLAG) != 0) {
	    return -1;
	}
	if ((inlining & NEEDS_TO_BE_QUICKED_FLAG) != 0) {
	    /*
	     * The method needs to execute once so that bytecodes in it
	     * are rewritten into quicks.  We could conceivably do that
	     * here but it doesn't seem worth the effort.  Once the call
	     * completes we'll try to quick and inline.
	     */
	    mb->inlining = 0;
	    return 1;
	}
#ifdef DEBUG
	if (verbose) { 
	    int i = 0, j;
	    jio_fprintf(stdout, "%s.%s%s =>%s ",
		   cbName(mb->fb.clazz), mb->fb.name, mb->fb.signature, 
		   ((inlining & SAME_CLASS_FLAG) ? "*" : ""));
	    while (i < 3) { 
		int newOp = (inlining >> (i++ << 3)) & 0xFF;
		jio_fprintf(stdout, "%s ", opnames[newOp]);
		for (j = opcode_length[newOp]; j > 1; j--) 
		    jio_fprintf(stdout, "%d ", (inlining >> (i++ << 3)) & 0xFF);
	    }
	    jio_fprintf(stdout, "\n");
	} 
#endif
    }
    /* Does this inlining have a dependency on the constant pool, and so
     * can only be used on a method in the same class. */
    if ((inlining & SAME_CLASS_FLAG) != 0) { 
	if (fieldclass(&currentMethod->fb) != fieldclass(&mb->fb))
	    return -1;
    }
    /* There is a inlining.  Copy that value into "result" */
    result[0] = INLINING_WORD1(inlining);
    result[1] = INLINING_WORD2(inlining);
    result[2] = INLINING_WORD3(inlining);
    return 0;
}

/* Given a method, determine if it can be "inlined" into three or fewer
 * bytes.
 */

static unsigned long 
MethodInlining(struct methodblock *mb) { 
    unsigned char *code = mb->code;

    /* The checkThis flag indicates that the resulting code must 
     * throw a NullPointerException if the first argument is null
     */
    bool_t checkThis = ((mb->fb.access & ACC_STATIC) == 0);
    bool_t redoInlining = FALSE;
    unsigned char *pc;
    long stackSize, stackBase = 0;
    const struct OpcodeInfoType *opPtr;
	
    for (pc = code, stackSize = 0; ; pc++) { 
	/* At this point in our simulation of the execution of the method, 
	 * registers stackBase .. stackBase+stackSize - 1 are pushed onto the
	 * the stack.  pc points to the next instruction to look at.
	 */
	opcode_type opcode = *pc;
	opcode_type opcode2;
	long reg, regSize, nextReg;

	if (stackSize == 0)
	    stackBase = 0;
	nextReg = stackBase + stackSize;
	opPtr = &OpcodeInfo[opcode];

	switch(opPtr->opcode) { 

	    case opc_iload_0:	/* a register load.  regnum from opcode */
		reg = REGNUM(opPtr);
	    continueRegLoad: 
		regSize = REGSIZE(opPtr);
		if (stackSize == 0) /* stack is currently empty */
		    stackBase = reg;
		else if (nextReg != reg)
		    return NO_INLINE_FLAG;
		stackSize += regSize;
		continue;

	    case opc_iload:	/* register load.  regnum from pc[1] */
		reg = pc[1]; pc++;
		goto continueRegLoad;


	    case opc_pop:	/* pop stack, or nop */
		stackSize -= REGSIZE(opPtr);
		continue;

	    case opc_nonnull_quick: /* special instruction */
		if (nextReg == 1) {
		    /* We're checking register 0 to ensure that it isn't null */
		    stackSize = 0; checkThis = TRUE;
		    continue;
		} 
		return NO_INLINE_FLAG;

	    case opc_invokeignored_quick: /* special instruction */
	    {		
		int popCount = pc[1];
		if (pc[2] != 0) {  
		    /* We only know how to check register 0 for non-null ness */
		    if (nextReg != popCount)
			return NO_INLINE_FLAG;
		    checkThis = TRUE;
		} 
		stackSize -= popCount;
		pc += 2;
		continue;
	    }

	    case opc_return:	/* return void or value */
		return makeReturnResult(mb, checkThis, nextReg, REGSIZE(opPtr));

	    case opc_getfield_quick_w: /* or putfield_quick_w */
		opcode2 = pc[3];
		if (!((opcode == opc_getfield_quick_w) 
		        ? isXreturn(opcode2) : opcode2 == opc_return)) { 
		    return NO_INLINE_FLAG;
		} else { 
		    cp_item_type *cp = cbConstantPool(fieldclass(&mb->fb));

		    struct fieldblock *fb = cp[GET_INDEX(pc+1)].fb;
		    char isig = fieldsig(fb)[0];
		    bool_t big = (isig == SIGNATURE_LONG 
				       || isig == SIGNATURE_DOUBLE);
		    unsigned long offset = fb->u.offset / sizeof (OBJECT);
		    opPtr = OpcodeInfoSpecial
			    + (opcode == opc_getfield_quick_w ? 0 : 1)
			    + (big ? 2 : 0);
		    if (offset <= 255 && stackBase == 0) {
			if (UseLosslessQuickOpcodes) {
			    redoInlining = TRUE;
			} else { 
			    opcode2 = (opcode == opc_getfield_quick_w) 
				       ? (big ? opc_getfield2_quick 
					      : opc_getfield_quick)
				       : (big ? opc_putfield2_quick 
					      : opc_putfield_quick);
			    return makeOpcodeResult(mb, FALSE, nextReg, 
						    opPtr->inStack, 3, 
						    (unsigned char) opcode2,
						    (unsigned char) offset, 0);
			}
		    }
		}
		/* FALLTHROUGH */

	    case opc_iadd: {	/* any simple instruction */
		int ilength = opcode_length[opcode];
		unsigned long result;
		opcode2 = pc[ilength];

		if (!((opPtr->outStack > 0)
			 ? isXreturn(opcode2) 
			 : (opcode2 == opc_return || opcode == opc_athrow)))
		    return NO_INLINE_FLAG;

		if ((opPtr->flags & NULL_CHECK) && (stackBase == 0)) { 
		    /* We don't need to generate code to check for null, since
		     * the instruction already does it.
		     */
		    checkThis = FALSE;
		} 
		switch (ilength) { 
		case 1:
		    result = 
			makeOpcodeResult(mb, checkThis, nextReg, opPtr->inStack,
					 1, (unsigned char) opcode, 0, 0);
		    break;
		case 2:
		    result = 
			makeOpcodeResult(mb, checkThis, nextReg, opPtr->inStack,
					 2, (unsigned char) opcode, pc[1], 0);
		    break;
		case 3:
		    result = 
			makeOpcodeResult(mb, checkThis, nextReg, opPtr->inStack,
					 3, (unsigned char) opcode, 
					 pc[1], pc[2]);
		    break;
		default:
		    sysAssert(FALSE);
		    result = NO_INLINE_FLAG;
		    break;
		}
		if ((result & NO_INLINE_FLAG) == 0) { 
		    if (opPtr->flags & CONSTANT_POOL) 
			result |= SAME_CLASS_FLAG;
		    if (redoInlining) 
			result |= REDO_INLINING_FLAG;
		}
		return result;
	    }

	    case 254:		/* can't inline until quicked */
		return NEEDS_TO_BE_QUICKED_FLAG;

	    default:
		sysAssert(FALSE);
	    case 255:		/* random instruction */
		return NO_INLINE_FLAG;

	} /* of switch statement  */
    } /* end of for loop */
}
    

/* This method is called to create the code that is actually going to replace
 * the indicated method, when the method does nothing, or when it simply
 * returns one of its arguments.
 *
 * It takes the following arguments:
 *   mb:          Method we are examining
 *   checkThis:   If true,  We must specially check that the first argument
 *                "this" isn't null. 
 *   highReg,     One greater than the highest register on the stack when
 *                the return or Xreturn is called.
 *   returnSize   Size of the return (0 for return, 1 for ireturn, 
 *                 2 for lreturn, etc);
 *
 * We have to emulate the method call in 3 bytes.  At the time the method is
 * called, the arguments reg[0] . . . reg[mb->args_size - 1] are pushed on the
 * stack.  
 */


static unsigned long
makeReturnResult(struct methodblock *mb, bool_t checkThis, 
		 long highReg, long returnSize) 
{ 
    long argsSize = mb->args_size;
    static unsigned char poppers[] = { opc_nop, opc_pop, opc_pop2 };
    if (returnSize == 0) { 
	/* Return void */
	return MAKE_INLINING(opc_invokeignored_quick, 
			     (unsigned char) argsSize,
			     (checkThis ? 1 : 0));
    } else { 
	/* Return some value from the stack */
	long returnReg = highReg - returnSize;  
	long excessArgs =  argsSize - returnSize - returnReg;
	sysAssert(returnReg >= 0 && returnSize >= 0);
	if (returnReg == 0) { 
	    /* Returning reg0 or reg0/reg1 */
	    if (checkThis) { 
		/* Must be returning reg0, which is also checked. We require
		 * argsSize >= 2 (which is the same thing as excessArgs >= 1), 
		 * because otherwise the "dup" might overflow the stack.  More
		 * sophisticated inliners might see if there is space on
		 * the caller's stack.
		 */
		sysAssert(returnSize == 1);
		if (excessArgs > 2 || argsSize < 2) 
		    return NO_INLINE_FLAG;
		return MAKE_INLINING(poppers[excessArgs],
				     opc_dup,
				     opc_nonnull_quick);
	    } else { 
		/* We're returning reg0 or reg0/reg1 which isn't null checked. 
		 * We just pop extraneous stuff off the stack */
		return MAKE_INLINING(opc_invokeignored_quick,
				     (unsigned char) excessArgs, 
				     0);
	    }
	} else {
	    /* At this point, returnReg > 0.  We're returning something other
	     * than the bottom of the stack. 
	     */
	    if (returnSize == 1 && returnReg == 1 && excessArgs <= 2) { 
		return MAKE_INLINING(poppers[excessArgs], 
				     opc_swap, 
				     checkThis ? opc_nonnull_quick : opc_pop);
	    } 
	    return NO_INLINE_FLAG;
	}
    }
}


/* This method is called to create the code that is actually going to replace
 * the indicated method
 * 
 * makeOpcodeResult is used to create a inlining that can be used anywhere
 * It takes the following arguments:
 *
 *   mb:          Method we are examining
 *   checkThis:   If true,  We must specially check that the first argument
 *                "this" isn't null.  This condition is >>NOT<< tested by 
 *                the generated code.
 *   nextReg:     In the emulation, the highest register on the stack is
 *                reg[nextReg - 1].
 *   icount       The number of bytes of instructions that follow.
 *   opcode, op1, op2
 *                The bytes of instruction.
 *
 * We have to emulate the method call in 3 bytes.  At the time the method is
 * called, the arguments reg[0] . . . reg[mb->args_size - 1] are pushed on the
 * stack.  So in three bytes, we have to:
 *     Remove any excess arguments from the stack.
 *     Perform the operation on the indicated stack registers
 *     Remove any objects lower down on the stack (this is hard!)
 *     Make sure that reg[0] is checked for being non-null, if necessary.
 */


static unsigned long
makeOpcodeResult(struct methodblock *mb, bool_t checkThis, 
		 long nextReg, long opcodeArgCount, long icount, 
		 unsigned char opcode, unsigned char op1, unsigned char op2)
{ 
    long firstReg = (opcodeArgCount == 0) ? 0 : nextReg - opcodeArgCount;
    sysAssert(firstReg >= 0 && opcodeArgCount >= 0 && icount > 0);

    if (firstReg > 0) { 
	/* There are extra registers at the bottom of the stack */
	return makePoppingResult(mb, checkThis, firstReg, opcodeArgCount, 
				 icount, opcode, op1, op2);
    } else {
	/* No extra registers at bottom of stack */
	long argsSize = mb->args_size;
	long excessArgs = argsSize - opcodeArgCount; /* extra at top */
	long popSpace = 3 - icount; /* space to pop args at top */

	unsigned long result = 0;
	long i;

	if (checkThis) { 
	    /* Unless this is a constant method that ignores all of its 
	     * arguments, we don't really have any way of checking register 0
	     * if the instructions doesn't.  If it is a constant instruction,
	     * deduct one from both popSpace and from excessArgs, since we are
	     * popping that last argument when an opc_nonnull_quick;
	     */
	    if (opcodeArgCount > 0 || popSpace == 0)
		return NO_INLINE_FLAG;
	    popSpace--, excessArgs--; 
	    sysAssert(excessArgs >= 0);
	}
	if (excessArgs > 2 * popSpace) 
	    return NO_INLINE_FLAG;
	for (i = 0; i < popSpace; i++) { 
	    /* If excessArgs <= popSpace, the following generates excessArgs
	     * "pops" followed by nops.  Otherwise, it generates 
	     * excessArgs - popSpace pop2's followed by pop's.
	     */
	    unsigned char opcode = (excessArgs <= i) ? opc_nop 
		                    : (excessArgs  <= popSpace + i) ? opc_pop 
		                    : opc_pop2;
	    result |= (opcode << (i << 3));
	}
	if (checkThis) 
	    result |= opc_nonnull_quick << ((i++) << 3);
	sysAssert(i + icount == 3);
	switch (icount) { 
	   case 3: result |= op2 << ((i + 2) << 3);
	   case 2: result |= op1 << ((i + 1) << 3);
	   case 1: result |= opcode << ((i + 0) << 3);
	}
	return result;
    }
}


/* 
 * Called by makeOpcodeResult.  
 * Same arguments.  But there are extra arguments on the bottom to pop.
 */

static unsigned long 
makePoppingResult(struct methodblock *mb, bool_t checkThis, 
		  long firstReg, long opcodeArgCount, long icount, 
		  unsigned char opcode, 
		  unsigned char op1, unsigned char op2)
{ 
    long argsSize = mb->args_size;
    long excessArgs = argsSize - opcodeArgCount - firstReg; /* extra on top */

    if (icount > 1) 
	/* We're just not prepared to deal with this. */
	return NO_INLINE_FLAG;

    if (OpcodeInfo[opcode].outStack == 0) {
	unsigned long result = 0;
	/* Something like an array store, that leaves no value on the stack */
	int i = 0;
	/* We can't deal with checkThis, since it might reverse the order of
	 * an exception.  We have a total of two instructions to do all the 
	 * pre and post popping.
	 */
	if (checkThis || ((excessArgs + 1)/2 + (firstReg + 1)/2) > 2)
	    return NO_INLINE_FLAG;
	for (; excessArgs > 0; excessArgs -=2) /* pre popping */
	    result |= (excessArgs == 1 ? opc_pop : opc_pop2) << ((i++) << 3);
	result |= opcode << ((i++) << 3);
	for (; firstReg > 0; firstReg -=2) /* post popping */
	    result |= (firstReg == 1 ? opc_pop : opc_pop2) << ((i++) << 3);
	while (i < 3)
	    result |= opc_nop << ((i++) << 3);
	return result;
    }

    if (excessArgs > 0 || firstReg > 1)
	/* We can't both do useful work and remove more than this many 
	 * items from the stack. */
	return NO_INLINE_FLAG;

    if (opcodeArgCount == 1) { 
	return MAKE_INLINING(opc_swap, 
			     checkThis ? opc_nonnull_quick : opc_pop, 
			     opcode);
    } 
    if (   ((OpcodeInfo[opcode].flags & (NULL_CHECK | CAN_ERROR)) == 0)
	&& (OpcodeInfo[opcode].outStack == 1)) {
	/* The result creates one thing on the stack, and it can't error */
	return MAKE_INLINING(opcode, 
			     opc_swap,
			     checkThis ? opc_nonnull_quick : opc_pop);
    } 
    return NO_INLINE_FLAG;
}
