/*
 * @(#)interpreter.c	1.289 98/09/02
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Interpret a Java object file
 */

#include <sys/types.h>
#include <string.h>
#include <stdarg.h>

#include "opcodes.h"
#include "oobj.h"
#include "interpreter.h"
#include "signature.h"
#include "typecodes.h"
#include "tree.h"
#include "threads.h"
#include "monitor.h"
#include "exceptions.h"
#include "javaString.h"
#include "sys_api.h"
#include "common_exceptions.h"
#include "jlong.h"

#include "java_lang_Thread.h"
#include "java_lang_Throwable.h"

#ifdef BREAKPTS
#include "breakpoints.h"
#endif

#ifdef HPROF
#include "vmprofiler.h"
#endif /* HPROF */

struct arrayinfo {
    int index;
    char sig;      /* type signature. */
    char *name;
    int factor;
};

static struct arrayinfo const arrayinfo[] = {
    { 0,	     0,			 "N/A",		 0},
    { 0,	     0,			 "N/A",		 0},
    { T_CLASS,	    SIGNATURE_CLASS,	"class[]",	sizeof(OBJECT)},
    { 0,	     0,			 "N/A",		 0},
    { T_BOOLEAN,    SIGNATURE_BOOLEAN,	"bool[]",	sizeof(char)},
    { T_CHAR,	    SIGNATURE_CHAR,	"char[]",	sizeof(unicode)},
    { T_FLOAT,	    SIGNATURE_FLOAT,	"float[]",	sizeof(float)},
    { T_DOUBLE,	    SIGNATURE_DOUBLE,	"double[]",	sizeof(double)},
    { T_BYTE,	    SIGNATURE_BYTE,	"byte[]",	sizeof(char)},
    { T_SHORT,	    SIGNATURE_SHORT,	"short[]",	sizeof(short)},
    { T_INT,	    SIGNATURE_INT,	"int[]",	sizeof(int)},
    { T_LONG,	    SIGNATURE_LONG,	"long[]",	sizeof(int64_t)}
};

static void FreeJavaStackMemory(JavaStack *this_stack);

int min_javastack_chunk_size = 1000;

#ifdef TRACING
int trace;
int tracem;
#  define TRACE(a) (trace ? jio_fprintf(stdout, "%6X %8X", (int)threadSelf(), (int)pc), \
	jio_printf a, fflush(stdout), 0 : 0)
#  define TRACEIF(o, a) (opcode == o ? TRACE(a) : 0)

#else
#  define TRACE(a)
#  define TRACEIF(a, b)
#endif

#ifdef STATISTICS
extern int instruction_count[256][256];
#endif

/* Some JIT compilers don't want to use the versions of
 * opc_invokevirtual_quick and opc_getfield_quick that lose information
 */
bool_t UseLosslessQuickOpcodes;	

/*
 * Lock around rewriting method code for quick_ instruction optimization
 */
sys_mon_t *_code_lock;
#define CODE_LOCK_INIT()    monitorRegister(_code_lock, "Code rewrite lock")
#define CODE_LOCK(self)	    sysMonitorEnter(self, _code_lock)
#define CODE_UNLOCK(self)   sysMonitorExit(self, _code_lock)

void lock_code(ExecEnv *ee)
{
    CODE_LOCK(EE2SysThread(ee));
}

void unlock_code(ExecEnv *ee)
{
    CODE_UNLOCK(EE2SysThread(ee));
}

/*
 * Java stack management
 */


#define rewriteByte(ptr, offset, newvalue) \
    ((volatile unsigned char *) (ptr))[(offset)] = (newvalue);

static JavaStack *
InitializeJavaStack(ExecEnv *ee, JavaStack *previous_stack, 
		    JavaStack *stack, size_t size)
{
    stack->execenv = ee;
    stack->prev = previous_stack;
    stack->next = 0;
    if (previous_stack) {
	sysAssert(previous_stack->next == 0);
        previous_stack->next = stack;
	stack->stack_so_far = previous_stack->stack_so_far 
	          + size * sizeof(stack->data[0]);
    } else {
	stack->stack_so_far = size * sizeof(stack->data[0]);
    }
    stack->end_data = &stack->data[size];
    return stack;
}

JavaStack *
CreateNewJavaStack(ExecEnv *ee, JavaStack *previous_stack, size_t size)
{
    JavaStack *stack;

    if (size < min_javastack_chunk_size)
        size = min_javastack_chunk_size;

    stack = sysMalloc(sizeof(JavaStack) + (size - 1) * sizeof(stack_item));
    if (stack == 0) {
	return 0;
    }

    return InitializeJavaStack(ee, previous_stack, stack, size);
}

/*
 * Initialize the ExecEnv of a new thread.  This function is called by
 * the new thread on its startup. returns FALSE on error.
 */
bool_t
InitializeExecEnv(ExecEnv *ee, JHandle *thread, void *stack_base) 
{
    ee->stack_base = stack_base;
    ee->thread = thread;

    if ((ee->initial_stack = CreateNewJavaStack(ee, 0, 0)) == 0)
        return FALSE;
 
    exceptionClear(ee);		/* Don't want to assume this means 0'ed... */

    /* Stuff for the JNI: */
    InitializeJNIRootFrame(ee);
    return TRUE;
}

void DeleteExecEnv(ExecEnv *ee) {
    JavaStack * stack;

    /* Clean up allocation cache; takes HEAP_LOCK */
    allocCacheCleanup(&ee->alloc_cache);

    /*
     * release the java stack.  We want to make sure that the ee doesn't
     * contain a pointer to the stack while we are freeing it, so we clear
     * it first and put in a store barrier for machines with weak memory
     * ordering.
     */
    stack = ee->initial_stack;
    ee->initial_stack = 0;
    sysStoreBarrier();
    FreeJavaStackMemory(stack);

    /* JNI */
    DeleteJNIRootFrame(ee);

    if (ee->privileged)
	sysFree(ee->privileged);

    memset(ee, 0, offsetof(ExecEnv, sys_thr));
}

static void FreeJavaStackMemory(JavaStack *this_stack)
{
    while (this_stack) {
        JavaStack *next_stack = this_stack->next;
	sysFree(this_stack);
	this_stack = next_stack;
    }
}

int ExpandJavaStackForJNI(ExecEnv *ee,
			  JavaStack **stackP, 
			  JavaFrame **frameP,
			  int capacity)
{
    return ExpandJavaStack(ee, stackP, frameP, 0, 0, 0, capacity);
}

int ExpandJavaStack(ExecEnv *ee,
		    JavaStack **stackP, 
		    JavaFrame **frameP, 
		    stack_item **optopP,
		    int args_size,
		    int nlocals, 
		    int maxstack)
{
    JavaStack *stack = *stackP;
    JavaFrame *frame = *frameP;
    size_t required_size =
        frame->ostack - (stack_item *)frame + nlocals + maxstack;

    if (stack->next != 0) {
        if (required_size > stack->next->end_data - stack->next->data) {
	    JavaStack *this_stack = stack->next;
	    stack->next = 0;
	    FreeJavaStackMemory(this_stack);
	} 
    }

    if (stack->next != 0) 
        stack = stack->next;
    else {
        if (stack->stack_so_far +
	    JAVASTACK_CHUNK_SIZE(stack)*sizeof(stack->data[0]) > 
	    JavaStackSize) {
	    ThrowStackOverflowError(0, 0);
	    return 0;
	}
	stack = CreateNewJavaStack(ee, stack, required_size);
	if (stack == 0) {
	    ThrowOutOfMemoryError(0, 0);
	    return 0;
	} 
    }	
    frame = (JavaFrame *)(stack->data + nlocals);
    if (optopP) {
        stack_item *copied_optop;
	stack_item *optop = *optopP;
      	for (copied_optop = stack->data; --args_size >= 0 ; ) {
	    copied_optop[args_size] = optop[args_size];
	}
	*optopP = copied_optop;
    }
    *frameP = frame;
    *stackP = stack;
    return 1;
}

/* Call a Java method from C.  The java method may have any number of arguments
 * and may return a single 4 byte value which may be an integer, a pointer to
 * an object handle, whatever.
 */
long
execute_java_dynamic_method(ExecEnv *ee, HObject *obj, 
			   char *method_name, char *signature, ...)
{
    va_list args;
    long result;
    va_start(args, signature);
        result = do_execute_java_method_vararg(ee, obj, method_name, signature, 
					      0, FALSE, args, 0, FALSE);
    va_end(args);
    return result;
}


/* Call a static Java method from C.  The java method may have any number of
 * arguments and may return a single 4-byte value.
 */
long
execute_java_static_method(ExecEnv *ee, ClassClass *cb, 
			  char *method_name, char *signature, ...)
{
    va_list args;
    long result;
    va_start(args, signature);
        result = do_execute_java_method_vararg(ee, cb, method_name, signature,
					      0, TRUE, args, 0, FALSE);
    va_end(args);
    return result;
}

/* Create a new Java object from C.  For example:
 *
 *       obj = execute_java_constructor(0, "java/lang/Float", 0, "(F)", f);
 *
 * EE should be either 
 *     .) the current execution environment
 *     .) 0, in which case the current execution environment is calculated
 *     .) PRIVILEGED_EE, meaning explicitly allow this call to access 
 *        constructors for which the current EE doesn't have privileges.
 *        [i.e. you can access private constructors, etc.]
 *
 * Either classname or cb gives the class of the object to be constructed.  
 * Pass the name of the class in classname, or a pointer to the class object
 * in cb.  Whichever one you give, the other should be NULL.
 *
 * signature is the signature of the arguments to the constructor.  It's 
 * first and last character should be "(" and ")" respectively.  Following
 * are the arguments to the function.
 */

HObject *execute_java_constructor(struct execenv *ee,
				  char *classname,
				  ClassClass *cb,
				  char *signature, ...)
{
#ifdef TRACING
    intptr_t pc = 0;		/* for TRACE */
#endif

    HObject *obj;
    va_list args;
#ifndef MAX_SIGNATURE_LEN
#define MAX_SIGNATURE_LEN 256
#endif /* MAX_SIGNATURE_LEN */
    char real_signature[MAX_SIGNATURE_LEN];
    int len;
    ClassClass *current_class;
    struct methodblock *mb;
    bool_t security = TRUE;

    
    if (ee == PRIVILEGED_EE) {
	ee = EE();
	security = FALSE;
    } else if (ee == 0) {
        ee = EE();
    }

    if (ee)
	/* skip pseudo frames */
	current_class = getCallerClass(ee, 0);
    else
	current_class = 0;

    /* Find the class.  Make sure we can allocate an object of that class */
    if (cb == NULL) {
	cb = FindClassFromClass(ee, classname, TRUE, current_class);
	if (cb == NULL) {
	    if (!ee || !exceptionOccurred(ee)) {
		ThrowNoClassDefFoundError(ee, classname);
            }
	    return 0;
	}
    } else { 
	classname = cbName(cb);
    }
    if (cbAccess(cb) & (ACC_INTERFACE | ACC_ABSTRACT)) { 
	ThrowInstantiationException(0, cbName(cb));
	return 0;
    }
    if (security && !VerifyClassAccess(current_class, cb, FALSE)) {
	ThrowIllegalAccessException(0, cbName(cb));
	return 0;
    }
    
    /* Find the constructor */
    if (jio_snprintf(real_signature, sizeof(real_signature), "%sV", signature) == -1) {
	ThrowInternalError(0, "signature overflow");
	return 0;
    }
    for (len = cbMethodsCount(cb), mb = cbMethods(cb); --len >= 0; mb++) 
	if (mb->fb.name == utf8_literal_obj_init_name &&
	    !strcmp(mb->fb.signature, real_signature))
	    break;
    if (len < 0) { 
	ThrowNoSuchMethodError(0, 0);
	return 0;
    }
    if (security && !VerifyFieldAccess(current_class, fieldclass(&mb->fb), 
					mb->fb.access, FALSE)) {
	ThrowIllegalAccessException(0, 0);
	return 0;
    } 

    /* Allocate the object and call the constructor */
    if ((obj = allocObject(ee, cb)) == 0) { 
	ThrowOutOfMemoryError(0, 0);
	return 0;
    }
    TRACE(("execute_java_constructor new %s => %s\n", 
	   cbName(cb), Object2CString(obj)));
    va_start(args, signature);
        do_execute_java_method_vararg(ee, obj, 0, 0, mb, FALSE, args, 0, FALSE);
    va_end(args);
    return obj;
}


long
do_execute_java_method(ExecEnv *ee, void *obj, 
		      char *method_name, char *signature, 
		      struct methodblock *mb,
		      bool_t isStaticCall, ...) {
    va_list args;
    long result;
    va_start(args, isStaticCall);
        result = do_execute_java_method_vararg(ee, obj, 
					  method_name, signature, 
					  mb, isStaticCall, args, 0, FALSE);
    va_end(args);
    return result;
}


long
do_execute_java_method_vararg(ExecEnv *ee, void *obj, 
			      char *method_name, char *method_signature, 
			      struct methodblock *mb,
			      bool_t isStaticCall, va_list args, 
			      long *otherBits, bool_t shortFloats) {

    JavaFrame *frame;
    bool_t isInitializer;
    JNIEnv *env;

    int retval = 0;
    char *p;
    jvalue tdub;

    if (ee == 0)
        ee = EE();
    env = EE2JNIEnv(ee);

    if (otherBits) {
	*otherBits = 0;
    }

    if (mb) {
        method_name = fieldname(&mb->fb);
	method_signature = fieldsig(&mb->fb);
    }

    isInitializer = (method_name[0] == '<'  &&  /* fast test */
		     !isStaticCall && !strcmp(method_name, "<init>"));

    {
        int stack_size = 
	    mb ? mb->args_size : (Signature2ArgsSize(method_signature) + 1);
	if (stack_size < 2) {
	    stack_size = 2;
	}
	if ((*env)->PushLocalFrame(env, stack_size) < 0) {
	    return 0;
	}
    }

    frame = ee->current_frame;
    frame->monitor = 0;	/* not monitoring anything */
    
    /* Push the target object, if not a static call */
    if (!isStaticCall) 
        (frame->optop++)->p = obj;

    for (p = method_signature + 1; *p != SIGNATURE_ENDFUNC; p++) {
	switch (*p) {
	  case SIGNATURE_BOOLEAN:
	  case SIGNATURE_SHORT:
	  case SIGNATURE_BYTE:
	  case SIGNATURE_CHAR:
	  case SIGNATURE_INT:
	    (frame->optop++)->i = va_arg(args, long);
	    break;
	  case SIGNATURE_FLOAT:
	      if (shortFloats) 
		  (frame->optop++)->i = va_arg(args, long);
	      else 
		  (frame->optop++)->f = va_arg(args, double);
	      break;
	  case SIGNATURE_CLASS:
	    (frame->optop++)->h = va_arg(args, HObject *);
	    while (*p != SIGNATURE_ENDCLASS) p++;
	    break;
	  case SIGNATURE_ARRAY:
	    while ((*p == SIGNATURE_ARRAY)) p++;
	    if (*p == SIGNATURE_CLASS) 
	        { while (*p != SIGNATURE_ENDCLASS) p++; } 
	    (frame->optop++)->p = va_arg(args, void *);
	    break;
	  case SIGNATURE_LONG:
	    SET_INT64(tdub, frame->optop, va_arg(args, int64_t));
	    frame->optop += 2;
	    break;
	  case SIGNATURE_DOUBLE:
	    SET_DOUBLE(tdub, frame->optop, va_arg(args, double));
	    frame->optop += 2;
	    break;
	  default:
	    jio_fprintf(stderr, "Invalid method signature '%s'\n", method_name);
	    goto done;
	}
    }
    {
	unsigned char pc[4];
	cp_item_type  constant_pool[6];
	unsigned char cpt[6];
	
	pc[0] =  isStaticCall ? (mb ? opc_invokestatic_quick : opc_invokestatic)
	       : (mb != NULL) ? opc_invokenonvirtual_quick
	       : isInitializer ? opc_invokespecial : opc_invokevirtual;
	pc[1] = 0; pc[2] = 1;	/* constant pool entry #1 */
	pc[3] = opc_return;
	/* Type table */
	constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type = cpt;
	cpt[0] = CONSTANT_POOL_ENTRY_RESOLVED;

	if (mb) {
	    constant_pool[1].mb = mb;
	    cpt[1] = CONSTANT_POOL_ENTRY_RESOLVED | CONSTANT_Methodref;
	    cpt[3] = 0;
	} else {
	    ClassClass *cb = isStaticCall ? (ClassClass *)obj 
		                      : obj_array_classblock((JHandle *)obj);
	    /* Method entry */
	    constant_pool[1].i = (2 << 16) + 3; /* Class, NameAndType */
	    cpt[1] = CONSTANT_Methodref;
	    /* Class entry */
	    constant_pool[2].clazz = cb; 
	    cpt[2] = CONSTANT_POOL_ENTRY_RESOLVED | CONSTANT_Class;
	    /* Name and type entry */
	    constant_pool[3].i = (4 << 16) + 5;	/* Name, Type */
	    cpt[3] = CONSTANT_POOL_ENTRY_RESOLVED | CONSTANT_NameAndType;
	    /* Method name and type  */
	    method_name = AddUTF8(ee, method_name);
	    if (method_name == 0) {
	        ThrowOutOfMemoryError(0, 0);
		goto done;
	    }
	    method_signature = AddUTF8(ee, method_signature);
	    if (method_signature == 0) {
	        ReleaseUTF8(ee, method_name);
	        ThrowOutOfMemoryError(0, 0);
		goto done;
	    }
	    constant_pool[4].cp = method_name;
	    constant_pool[5].cp = method_signature;
	    cpt[4] = cpt[5] = CONSTANT_POOL_ENTRY_RESOLVED | CONSTANT_Utf8;
	}

	frame->constant_pool = constant_pool;

	/* Run the byte codes outside the lock and catch any exceptions. */
	exceptionClear(ee);
	if ((*pExecuteJava)(pc, ee) && (p[1] != SIGNATURE_VOID)) {
	    retval = frame->optop[-1].i;
	    if ((p[1] == SIGNATURE_DOUBLE || p[1] == SIGNATURE_LONG) 
		  && (otherBits != 0))
		*otherBits = frame->optop[-2].i;
	}
	if (cpt[3] == CONSTANT_NameAndType) {
	    ReleaseUTF8(ee, method_name);
	    ReleaseUTF8(ee, method_signature);
	}
    }

 done:
    /* Our caller can look at ee->exceptionKind and ee->exception. */
    (*env)->PopLocalFrame(env, 0);
    return retval;
}

char *
Object2CString(JHandle *oh)
{                               /* ARB to C string */
    ClassObject *o;
    ClassClass *cb;
    /* Using a static buffer is not MT-safe! */
    static char buf[100];
    int maxlen = sizeof(buf) - 10;

    if (oh == 0)
	return "NULL";
    if (!isHandle(oh)) { 
	(void) jio_snprintf(buf, sizeof(buf), "BOGUS-HANDLE[%x]", (intptr_t)oh);
	return buf;
    }
    o = unhand(oh);
    if (!isObject(o)) { 
	/* Silently truncate on unlikely overflow */
	(void) jio_snprintf(buf, sizeof(buf), "BOGUS-PTR[%d]", (int)o);
    } else {
	switch (obj_flags(oh)) {
	case T_NORMAL_OBJECT:
	    {
		char cbuf[MSG_BUF_LEN];
		cb = obj_classblock(oh);
		/* Quietly truncate on buffer overflow */
		(void) jio_snprintf(buf, sizeof(buf), "%s@%X%c%X",
			 classname2string(cbName(cb), cbuf, sizeof(cbuf)),
			 (int)oh, DIR_SEPARATOR, (int)o);
	    }
	    break;
	case T_CHAR: {
	    HArrayOfChar *str = (HArrayOfChar *) oh;
	    int len = obj_length(str);
	    unicode *s = unhand(str)->body;
	    unicode * body = s;
	    char *p = buf;
	    
	    len = (len > maxlen ? maxlen : len);

	    *p++ = '"';
	    while (--len >= 0)
		*p++ = *s++;
	    *p++ = '"';
	    *p++ = '\0';
	    KEEP_POINTER_ALIVE(body);
	    break;
	    }
	case T_CLASS:
#define BODYOF(h)   unhand(h)->body
	    {
		char cbuf[MSG_BUF_LEN];

		cb = ((ClassClass **) BODYOF((HArrayOfObject*)oh))[obj_length(oh)];
		/* Quietly truncate on buffer overflow */
		(void) jio_snprintf(buf, sizeof(buf), "%s[%d]",
			cb ? classname2string(cbName(cb), cbuf, sizeof(cbuf)) : "*MissingClassName*",
			obj_length(oh));
	    }
	    break;
	default:
	    /* Quietly truncate on buffer overflow */
	    (void) jio_snprintf(buf, sizeof(buf), "%s[%d]", 
		    arrayinfo[obj_flags(oh)].name, obj_length(oh));
	}
    }
    return buf;
}

bool_t
is_instance_of(JHandle * h, ClassClass *dcb, ExecEnv *ee)
{
    if (h == NULL) {		/* null can be cast to anything */
	return TRUE;
    } else if (obj_flags(h) == T_NORMAL_OBJECT) {
	/* h is an object.  Get its class and to the work */
	return is_subclass_of(obj_classblock(h), dcb, ee);
    } else {
	/* h is an array.  dcb must either be classJavaLangObject, or 
	   interfaceJavaLangCloneable, or an array class */
	return cbName(dcb)[0] == SIGNATURE_ARRAY
	           ? array_is_instance_of_array_type(h, dcb, ee)
		   : (dcb == classJavaLangObject 
		          || dcb == interfaceJavaLangCloneable
		          || dcb == interfaceJavaIoSerializable);
    }
}

bool_t
is_subclass_of(ClassClass *cb, ClassClass *dcb, ExecEnv *ee)
{
    if (cb == dcb) {
        return TRUE;
    }
    /* check the subclass of cache */
    if (dcb == cbLastSubclassOf(cb))
	return TRUE;

    if (cbAccess(dcb) & ACC_INTERFACE) {
        if (ImplementsInterface(cb, dcb, ee)) {
	    cbLastSubclassOf(cb) = dcb;
	    return TRUE;
	} else {
	    return FALSE;
	}
    } else {
	ClassClass * orig_cb = cb;
	for (cb = cbSuperclass(cb); cb; cb = cbSuperclass(cb)) {
	    if (cb == dcb) {
		cbLastSubclassOf(orig_cb) = dcb;
		return TRUE;
	    }
	}
	return FALSE;
    }
}

/* Return TRUE if cb implements the interface icb */

bool_t 
ImplementsInterface(ClassClass *cb, ClassClass *icb, ExecEnv *ee)
{
    int i;
    /* There is a chance cb is not yet linked, and imethodtable is not
     * yet set up. We do not want to link the class here, because we
     * don't want is_subclass_of to throw an exception.
     */
    if (CCIs(cb, Linked)) {
        /* This may be a little faster than looking at the constant pool */
        struct imethodtable *imtable = cbIntfMethodTable(cb);
	for (i = imtable->icount - 1; i >= 0; i--) {
	    if (imtable->itable[i].classdescriptor == icb) {
	        return TRUE;
	    }
	}
    } else {
	while (cb) {
	    cp_item_type *constant_pool = cbConstantPool(cb);
	    unsigned short *implements = cbImplements(cb);
	    for (i = cbImplementsCount(cb) - 1; i >= 0; i--) {
		int index = implements[i];
		ClassClass *interface_cb = constant_pool[index].clazz;
		if (is_subclass_of(interface_cb, icb, ee)) {
		    return TRUE;
		}
	    }
	    cb = cbSuperclass(cb);
	}
    }
    return FALSE;
}

/*  Return true if the array 'from' is assignable to variables of type
 *    'to_class'. The 'from' array is represented by 
 *  (a) the 'from_elt_class' i.e the class of the *elements* of the array,
 *  (b) the 'from_elt_type', if the array elements are primitive (and
 *                           'from_elt_class' is then NULL).
 */
bool_t
array_type_assignable_to_array_type(ClassClass *from_elt_class,
				    int from_elt_type,
				    ClassClass *to_class, ExecEnv *ee)
{
    /* 
     * type = 'kind' of object -indicates one of the primitives or a class.
     * array item     = the things stored in the array. 
     * array element  = the first level components of the array, which may 
     *                  be either (recursively) arrays, or items themselves.
     * Arrays item types are determined in one of 3 ways:
     * 1] If the items are primitive, the 'type' of the array is sufficient
     * 2] If the elements are classes, the 'type' of the array is T_CLASS
     *          and [length] points to the item class
     * 3] If the elements are arrays, the 'type' of the array is T_CLASS
     *        and [length] points to a class representing the elements.
     *        The "item" class may be obtained from the "element" class.
     * Thus, the 'to' items can be directly determined from the array class,
     * while the 'from' items must be inferrered from the 'from' elements
     * precondition: (cbName(to_class)[0] == SIGNATURE_ARRAY);
     */
    int to_depth = cbConstantPool(to_class)[CONSTANT_POOL_ARRAY_DEPTH_INDEX].i;
    int to_item_type =  
	cbConstantPool(to_class)[CONSTANT_POOL_ARRAY_TYPE_INDEX].i; 
    ClassClass *to_item_class = (to_item_type == T_CLASS)
	? cbConstantPool(to_class)[CONSTANT_POOL_ARRAY_CLASS_INDEX].clazz 
	: NULL;	
    /* Asssume the items are primitive, override if from_elt_class != NULL */
    int from_depth = 1;
    int from_item_type = from_elt_type;
    ClassClass *from_item_class = NULL;

    if (from_elt_class != NULL) { /* => Elts are reference objects */
	if (cbName(from_elt_class)[0] == SIGNATURE_ARRAY) { /*=>Arr of arrays*/
	    union cp_item_type *from_elt_cp = cbConstantPool(from_elt_class);
	    from_depth = from_elt_cp[CONSTANT_POOL_ARRAY_DEPTH_INDEX].i + 1;
	    from_item_type = from_elt_cp[CONSTANT_POOL_ARRAY_TYPE_INDEX].i;
	    from_item_class =
		from_elt_cp[CONSTANT_POOL_ARRAY_CLASS_INDEX].clazz;
	} else {   /* => Arr of classes */
	    from_depth = 1;
	    from_item_type = T_CLASS;
	    from_item_class = from_elt_class;
	}
    } else {
	/*  should be primitive */
	sysAssert(from_elt_type != T_CLASS);
    }
    if (from_depth > to_depth) { /*=> Match only if 'to' items are 'special'*/
	return (to_item_class == classJavaLangObject) || 
	    (to_item_class == interfaceJavaLangCloneable) ||
	    (to_item_class == interfaceJavaIoSerializable);
    } else if (from_depth == to_depth) { 
	/* If the elements are primitive, sufficient if the item type matches
	 * If not, if the from_item_class <= to_item_class */
	return  (to_item_type == from_item_type) 
	       && ((to_item_type != T_CLASS) 
		   || (to_item_class == from_item_class)
		   || is_subclass_of(from_item_class, to_item_class, ee));
    } else { 
	return FALSE;
    } 
}

bool_t array_is_instance_of_array_type(JHandle * h, ClassClass *cb,
				       ExecEnv *ee)
{
     ClassClass *from_class = NULL;
     int from_elt_type = obj_flags(h);
     
     if (from_elt_type == T_CLASS) {
	/* We have an array of classes.  Get the class type */
	ArrayOfObject *array = (ArrayOfObject *)unhand(h);
	size_t length = obj_length(h);
	from_class = (ClassClass *) array->body[length];
     }
     return array_type_assignable_to_array_type(from_class,
						from_elt_type, cb, ee);
}


/* An exception has been thrown.  Find the pc at which this exception
 * is handled.  Alternatively, rethrow the error if no appropriate pc
 * is found.
 */

unsigned char *
ProcedureFindThrowTag(ExecEnv *ee, JavaFrame *frame, 
		      JHandle *object, unsigned char *pc)
{
    if (frame->current_method == 0) {
	return 0;
    } else {
	/* Search current class for catch frame, based on pc value,
	   matching specified type. */
	struct methodblock *mb = frame->current_method;
	struct CatchFrame *cf = mb->exception_table;
	struct CatchFrame *cfEnd = cf + mb->exception_table_length;
	intptr_t pcTarget = pc - mb->code; /* offset from start of code */
	cp_item_type *constant_pool = frame->constant_pool;
	unsigned char *type_table = 
	    constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
	for (; cf < cfEnd; cf++) { 
	    if (cf->start_pc <= pcTarget && pcTarget < cf->end_pc) { 
		/* Is the class name a superclass of object's class. */
		if (cf->catchType == 0) {
		    /* Special hack put in for finally.  Always succeed. */
		    goto found;
		} else { 
		    /* The usual case */
		    int catchType = cf->catchType;
		    ClassClass *catchClass;

		    sysAssert(type_table[catchType] & CONSTANT_Class);

		    if (!CONSTANT_POOL_TYPE_TABLE_IS_RESOLVED(type_table, 
							      catchType)) {
		        if (!ResolveClassConstantFromClass(fieldclass(&mb->fb),
							   catchType,
							   ee, 
						   1 << CONSTANT_Class)) {
			    return 0;
			}
		    }
		    catchClass = constant_pool[catchType].clazz;

		    if (is_instance_of(object, catchClass, ee)) {
		        goto found;
		    }
		}
	    }
	}
	/* not found in this execution environment */
	return 0;
      found:
	return mb->code + cf->handler_pc;
    }
}


/* Called by opc_multianewarray.  Create a filled in multidimensional array
 *    dimensions:  The number of dimensions to fill in
 *    array_cb:    The type of the array
 *    sizes:       An array of the dimension sizes
 *
 * Return 0 if we've run out of space.  Otherwise, return the allocated array.
 */

static HObject *
MultiArrayAlloc2(ExecEnv *, int, ClassClass **, stack_item *, int);

HObject *
MultiArrayAlloc(int dimensions, ClassClass *array_cb, stack_item *sizes)
{
    /* Get the base type and depth of the array class we're constructing */
    cp_item_type *constant_pool = cbConstantPool(array_cb);
    int depth = constant_pool[CONSTANT_POOL_ARRAY_DEPTH_INDEX].i;
    int base_type = constant_pool[CONSTANT_POOL_ARRAY_TYPE_INDEX].i;
    ClassClass *sub_array_cb[MAXDIM + 1];
    char *class_name = cbName(array_cb);
    struct execenv *ee = EE();
    int i;

    sysAssert(class_name[0] == SIGNATURE_ARRAY);
    sysAssert(dimensions <= depth);

    /* Fill in sub_array_cb with the array types of each of the lower.
     * dimensions.  Put a 0 in the slot if we've reached a non-class base type
     */
    for (i = 1; i <= dimensions; i++) {
	if (i < depth) {
	    sub_array_cb[i - 1] = 
		FindClassFromClass(ee, class_name + i, FALSE, array_cb);
	} else if (i == depth && base_type == T_CLASS) {
	    sub_array_cb[i - 1] = 
		constant_pool[CONSTANT_POOL_ARRAY_CLASS_INDEX].clazz;
	} else {
	    sub_array_cb[i - 1] = 0;
	}
    }
    /* Call a recursive function that does all the work */
    return MultiArrayAlloc2(ee, dimensions, sub_array_cb, sizes, base_type);
}

/* This function does all the actual work for MultiArrayAlloc(), after that
 * function has done all the checking and initial work. 
 */
static HObject *
MultiArrayAlloc2(ExecEnv *ee, int dimensions, ClassClass **types,
		 stack_item *sizes, int base_type)
{
    int size = sizes[0].i;	/* size of this dimension */
    HObject *ret;
    int i;

    if (types[0] != 0) {
	ret = allocArray(ee, T_CLASS, size); /* generate the array */
	if (ret == 0) 
	    return 0;
	/* fill in the type  */
	unhand(((HArrayOfObject *)ret))->body[size] = (HObject*)types[0];
	if (dimensions > 1) {
	    /* fill in the subarrays */
	    for (i = 0; i < size; i++) {
		HObject *sub = MultiArrayAlloc2(ee, dimensions - 1, types + 1, 
						sizes + 1, base_type);
		if (sub == 0)	/* out of memory? */
		    return 0;
		unhand(((HArrayOfObject *)ret))->body[i] = sub;
	    } 
	} else {
	    /*EMPTY*/
	    /* this is the last dimension. Nothing to do */
	}
    } else {
	sysAssert(dimensions == 1);
	ret = allocArray(ee, base_type, size);
	if (ret == 0)		/* out of memory? */
	    return 0;
    }
    return ret;
}



#define JAVA_ERROR( name, data ) \
    if (1) { \
        ee->current_frame->lastpc = pc; \
	SignalError( ee, JAVAPKG name, data ); \
	return -1; \
    } else ((void) 0)

/* 
 * Should  a call to the specified methodblock be turned into
 * an invokesuper_quick instead of a invokenonvirtual_quick?
 *
 * The four conditions that have to be satisfied:
 *    The method isn't private
 *    The method isn't an <init> method
 *    The ACC_SUPER flag is set in the current class
 *    The method's class is a superclass (and not equal) to the current class.
 */ 
bool_t
isSpecialSuperCall(ClassClass *current_class, struct methodblock *mb) {
    /* Remove ifndef when code for opc_invokesuper_quick in the asm FI works. */
    if (((cbAccess(current_class) & ACC_SUPER) != 0)
	  && (!IsPrivate(mb->fb.access))
	  && (mb->fb.name != utf8_literal_obj_init_name)
	  && (current_class != classJavaLangObject)) {
	ClassClass *methodClass = fieldclass(&mb->fb);
	ClassClass *cb = cbSuperclass(current_class);
	for (; ; cb = cbSuperclass(cb)) {
	    if (cb == methodClass) 
		return TRUE;
	    if (cbSuperclass(cb) == 0)
		return FALSE;
	}
    }
    return FALSE;
}


int
quickFieldAccess(int opcode, unsigned char * pc, 
		 struct fieldblock *fb, ExecEnv *ee ){
    char *signature = fieldsig(fb);
    int result = 0;
    TRACE(("\t%s %s.%s\n", opnames[opcode], 
	   cbName(fieldclass(fb)), fieldname(fb)));
    if (fb->access & ACC_STATIC) {
	char buf[MSG_BUF_LEN];
	int len;
	classname2string(cbName(fieldclass(fb)), 
			 buf, sizeof(buf));
	len = strlen(buf);
	/* If buffer overflow, quietly truncate */
	(void) jio_snprintf(buf + len, sizeof(buf)-len,
		": field %s did not used to be static",
		fb->name);
	JAVA_ERROR("IncompatibleClassChangeError", buf);
    } else if ((fb->access & ACC_FINAL) 
	       && (opcode == opc_putfield) 
	       && ( (ee->current_frame->current_method == NULL)
		   || (fieldclass(fb) != 
		        fieldclass(&ee->current_frame->current_method->fb)))) { 
	char buf[MSG_BUF_LEN];
	int len;
	classname2string(cbName(fieldclass(fb)), buf, sizeof(buf));
	len = strlen(buf);
	/* If buffer overflow, quietly truncate */
	(void) jio_snprintf(buf + len, sizeof(buf)-len, 
			    ": field %s is final", fb->name);
	JAVA_ERROR("IllegalAccessError", buf);
    } else {
	unsigned int offset = fb->u.offset / sizeof (OBJECT);
	CODE_LOCK(EE2SysThread(ee));
	    if (pc[0] != opcode && pc[0] != opc_breakpoint) { 
		result = 1;
	    } else if (offset <= 0xFF && !UseLosslessQuickOpcodes) { 
		int delta = (signature[0] == SIGNATURE_LONG 
			     || signature[0] == SIGNATURE_DOUBLE)
		        ? (opc_getfield2_quick - opc_getfield)
		        : (opc_getfield_quick - opc_getfield);
		rewriteByte(pc, 1, offset);
		sysStoreBarrier();
		rewriteByte(pc, 0, opcode + delta);
	    } else { 
		int delta = (opc_getfield_quick_w - opc_getfield);
		rewriteByte(pc, 0, opcode + delta);
	    }
	CODE_UNLOCK(EE2SysThread(ee));
    }
    return result;
}


struct methodblock *
quickSelectSuperMethod(JavaFrame * frame, struct methodblock * mb)
{
    ClassClass *current_class = frame->current_method ? 
	fieldclass(&frame->current_method->fb) : NULL;
    struct methodblock *new_mb = mb;
    unsigned int offset = mb->fb.u.offset; 
    if ((current_class != NULL) && 
	isSpecialSuperCall(current_class, mb)) {
	ClassClass *superClass = cbSuperclass(current_class);
	new_mb = mt_slot(cbMethodTable(superClass), offset);
    }
    if (mb != new_mb) { 
	ClassClass *currentClass =
	    fieldclass(&frame->current_method->fb);
	ClassClass *superClass = cbSuperclass(currentClass);
	mb = cbMethodTable(superClass)->methods[mb->fb.u.offset];
    } 
    return mb;
}



int quickStaticAccess(int opcode, unsigned char * pc, 
		      struct fieldblock *fb, ExecEnv *ee ){
    TRACE(("\t%s %s.%s\n", opnames[opcode], 
	   cbName(fieldclass(fb)), fieldname(fb)));
    if ((fb->access & ACC_STATIC) == 0) {
	char buf[MSG_BUF_LEN];
	int len;
	classname2string(cbName(fieldclass(fb)), 
			 buf, sizeof(buf));
	len = strlen(buf);
	/* If buffer overflow, quietly truncate */
	(void) jio_snprintf(buf + len, sizeof(buf)-len,
		": field %s used to be static",
		fb->name);
	JAVA_ERROR("IncompatibleClassChangeError", buf);
    } else if ((fb->access & ACC_FINAL) 
	       && (opcode == opc_putstatic) 
	       && ( (ee->current_frame->current_method == NULL)
		   || (fieldclass(fb) != 
		        fieldclass(&ee->current_frame->current_method->fb)))) { 
	char buf[MSG_BUF_LEN];
	int len;
	classname2string(cbName(fieldclass(fb)), buf, sizeof(buf));
	len = strlen(buf);
	/* If buffer overflow, quietly truncate */
	(void) jio_snprintf(buf + len, sizeof(buf)-len, 
			    ": field %s is final", fb->name);
	JAVA_ERROR("IllegalAccessError", buf);
    } else {
	char *signature = fieldsig(fb);
	int delta = (signature[0] == SIGNATURE_LONG 
		       || signature[0] == SIGNATURE_DOUBLE)
		   ? (opc_getstatic2_quick - opc_getstatic)
		   : (opc_getstatic_quick - opc_getstatic);
	rewriteByte(pc, 0, opcode + delta);
    }
    return 0;
}

int quickInvocation(int opcode, unsigned char * pc, 
		    struct methodblock *mb, ExecEnv *ee ) {
    unsigned char quick_buffer[3];
    unsigned char inline_buffer[3];
    int result = 0;
    int inlineresult = 0;

    TRACE(("\t%s %s.%s%s\n", opnames[opcode], 
	   cbName(fieldclass(&mb->fb)), 
	   fieldname(&mb->fb), fieldsig(&mb->fb)));
    if ((opcode == opc_invokestatic) == ((mb->fb.access & ACC_STATIC) == 0)) { 
	char buf[MSG_BUF_LEN];
	int len;
	classname2string(cbName(fieldclass(&mb->fb)), buf, sizeof(buf));
	len = strlen(buf);
	/* If buffer overflow, quietly truncate */
	(void )jio_snprintf(buf + len, sizeof(buf)-len, 
			    ": method %s%s %sused to be static",
			    mb->fb.name, mb->fb.signature,
			    (opcode == opc_invokestatic ? "" : "did not "));
	JAVA_ERROR("IncompatibleClassChangeError", buf);
    }

    CODE_LOCK(EE2SysThread(ee));
    /* copy the invocation into a temporary buffer to do the rewrite */
    quick_buffer[0] = pc[0];
    quick_buffer[1] = pc[1];
    quick_buffer[2] = pc[2];
    
    if (quick_buffer[0] != opcode && quick_buffer[0] != opc_breakpoint) {
	/*
	 * the opcode changed since we entered the call so we should
	 * return and redispatch to the new opcode
	 */
	result = 1;
    } else if (opcode == opc_invokestatic) { 
	rewriteByte(quick_buffer, 0, opc_invokestatic_quick);
    } else if (opcode == opc_invokevirtual) {
	unsigned int offset = mb->fb.u.offset;
	if (mb->fb.access & ACC_PRIVATE) {
	    rewriteByte(quick_buffer, 0, opc_invokenonvirtual_quick);
	} else if (offset <= 0xFF && !UseLosslessQuickOpcodes) {
	    rewriteByte(quick_buffer, 1, mb->fb.u.offset); 
	    rewriteByte(quick_buffer, 2, mb->args_size); 
	    rewriteByte(quick_buffer, 0,
			fieldclass(&mb->fb) == classJavaLangObject 
			? opc_invokevirtualobject_quick
			: opc_invokevirtual_quick);
	} else { 
	    rewriteByte(quick_buffer, 0, opc_invokevirtual_quick_w);
	}
    } else if (opcode == opc_invokespecial) {
	struct methodblock *new_mb =
	    quickSelectSuperMethod(ee->current_frame, mb);
	if (mb == new_mb) { 
	    rewriteByte(quick_buffer, 0, opc_invokenonvirtual_quick);
	} else { 
	    int offset = mb->fb.u.offset;
	    rewriteByte(quick_buffer, 1, offset >> 8);
	    rewriteByte(quick_buffer, 2, offset & 0xFF);
	    rewriteByte(quick_buffer, 0, opc_invokesuper_quick);
	    mb = new_mb;
	} 
    } 
    
    if (result == 0) {
	/*
	 * The invocation has been quicked so try to do some inlining
	 * on the invocation.  If we can't inline then we just do the
	 * quicking.  If the inlining succeeds then we slap the inlining
	 * into place instead.  It possible that the inliner will decide
	 * that it can't inline until the method has been invoked once,
	 * because it has instructions that need to be quicked.  In that
	 * case we don't even do the quicking on this invocation since
	 * then we couldn't do the inlining later without add locking
	 * inside the invoke*_quick instructions, which we definitely don't
	 * want to do.  This will normally only happen when a class is first
	 * used since the inlining information is cached in the method block
	 * and only needs to be calculated once.
	 */
	inlineresult = MethodCallInline(quick_buffer,
					ee->current_frame->current_method,
					mb, inline_buffer);

	/*
	 * If the inliner wants us to execute this method once before
	 * it will decide if it can be inlined and the method has been
	 * JITed, we should just go ahead and do the quicking, since
	 * the bytecodes in the method will never be rewritten because
	 * the JIT code is being executed instead.
	 */
	if (inlineresult == 1 && mb->fb.access & ACC_MACHINE_COMPILED)
	    inlineresult = -1;

	if (inlineresult == 0) {
	    /* successfully inlined so copy from the inlining buffer */
	    rewriteByte(pc, 1, inline_buffer[1]);
	    sysStoreBarrier();
	    rewriteByte(pc, 2, inline_buffer[2]);
	    sysStoreBarrier();
	    rewriteByte(pc, 0, inline_buffer[0]);
	    result = 1;
	} else if (inlineresult == -1) {
	    /* it couldn't inline, so just copy the quicked version */
	    rewriteByte(pc, 1, quick_buffer[1]);
	    sysStoreBarrier();
	    rewriteByte(pc, 2, quick_buffer[2]);
	    sysStoreBarrier();
	    rewriteByte(pc, 0, quick_buffer[0]);
	    result = 1;
	} else {
	    /*
	     * The method needs to execute once before we can inline it
	     * so we need return 0 and do nothing this time.  Next time
	     * through we try again and it should succeed or fail.
	     */
	    result = 0;
	}
    }

    CODE_UNLOCK(EE2SysThread(ee));
    return result;
}


/*
 * This function is called when an earlier attempt to quick and inline
 * an invocation failed because the inliner couldn't attempt the inlining
 * until bytecode in the method to be inlined were quicked.  If the method
 * is still a non-quick version of the the bytecode then look up the method
 * that it would invoke and retry to quick and inline.  It should be able
 * determine if the method can be inlined at this point.
 */

void
FixupQuickInvocation(ExecEnv *ee, JavaFrame * frame)
{
    unsigned char * pc = frame->lastpc;
    int opc = pc[0];
    cp_item_type *constant_pool = frame->constant_pool;
    struct methodblock *mb;

    switch (opc) {
    case opc_invokestatic:
    case opc_invokespecial:
    case opc_invokevirtual:
	mb = ResolveClassConstantFromPC(pc, (unsigned char)opc, 
					constant_pool, ee,
					1 << CONSTANT_Methodref);
	if (!exceptionOccurred(ee) && mb != NULL) {
	    int result = quickInvocation(opc, pc, mb, ee);
	    /* at this point it should either quick or inline */
	    sysAssert(result != 0);
	}
    }
}

bool_t
invokeInterfaceError(ExecEnv *ee, unsigned char *pc, 
			  ClassClass *cb, ClassClass *intf)
{ 
    char buf[MSG_BUF_LEN];
    char *bufptr = buf, *endbuf = buf + MSG_BUF_LEN;

    jio_snprintf(bufptr, endbuf - bufptr, "class ");
       bufptr += strlen(bufptr);

    classname2string(cbName(cb), bufptr, endbuf - bufptr);
       bufptr += strlen(bufptr);

    jio_snprintf(bufptr, endbuf - bufptr, " does not implement interface ");
       bufptr += strlen(bufptr);

    classname2string(cbName(intf), bufptr, endbuf - bufptr);
    JAVA_ERROR("IncompatibleClassChangeError", buf);
}


/*
 * Note: Call InitializeInterpreter() BEFORE InitializeClassThread() to
 * ensure that the Java stack lock is initialized before the idle thread
 * has a chance to run!
 */
bool_t
InitializeInterpreter(void)
{
    InitializeObjectHash();

    _code_lock=(sys_mon_t *) sysMalloc(sysMonitorSizeof());
    if (_code_lock == NULL)
        return FALSE;
    CODE_LOCK_INIT();

    _sysloader_lock=(sys_mon_t *) sysMalloc(sysMonitorSizeof());
    if (_sysloader_lock == NULL)
        return FALSE;
    SYSLOADER_LOCK_INIT();

    _linkclass_lock=(sys_mon_t *) sysMalloc(sysMonitorSizeof());
    if (_linkclass_lock == NULL)
        return FALSE;
    LINKCLASS_LOCK_INIT();

    _binclass_lock=(sys_mon_t *) sysMalloc(sysMonitorSizeof());
    if (_binclass_lock == NULL)
        return FALSE;
    BINCLASS_LOCK_INIT();

    if (!InitializeJNI()) {
        return FALSE;
    }
#ifdef STATISTICS
    memset(instruction_count, 0, sizeof(instruction_count));
#endif
    return TRUE;
}

void *
ResolveClassConstantFromPC(unsigned char *pc, unsigned char opcode, 
			       cp_item_type *constant_pool, 
			       struct execenv *ee, 
			       unsigned mask) { 
    unsigned char opcode2;
    unsigned index;
    CODE_LOCK(EE2SysThread(ee));
	opcode2 = pc[0];
        index = GET_INDEX(pc + 1);
    CODE_UNLOCK(EE2SysThread(ee));
#ifdef BREAKPTS
    if (debugging) {
        if (opcode2 == opc_breakpoint) {	/* Find the real opcode */
	    opcode2 = get_breakpoint_opcode(ee, pc, FALSE);
	}
    }
#endif
    if (opcode2 != opcode) { 
	return NULL;
    } else { 
        if (!ResolveClassConstant(constant_pool, index, ee, mask)) {
	    return NULL;
	}
	/* Initialize class on static access. */
	switch (mask) {
	    case 1 << CONSTANT_Methodref: {
	        struct methodblock *mb = constant_pool[index].p;
		if (mb->fb.access & ACC_STATIC) {
		    InitClass(mb->fb.clazz);
		}
		break;
	    }
	    case 1 << CONSTANT_Fieldref: {
	        struct fieldblock *fb = constant_pool[index].p;
		if (fb->access & ACC_STATIC) {
		    InitClass(fb->clazz);
		}
		break;
	    }
	    default:
	        sysAssert(FALSE);
	}
	return (exceptionOccurred(ee)) ? NULL : constant_pool[index].p;
    } 
}

typedef void (*pProc)(void);

static struct exit_proc {
    pProc proc;
    struct exit_proc *next;
} *exit_procs = NULL;

int
PrepareToExit(void)
{
    struct exit_proc *pExit;

    pExit = exit_procs;
    while (pExit != NULL) {
        struct exit_proc *pNext = pExit->next;
        (*pExit->proc)();
        sysFree(pExit);
        pExit = pNext;
    }
    return sysShutdown();
}

void
Exit(int status)
{
    PrepareToExit();

    if (exit_hook)
        (*exit_hook)(status);
    else
        exit(status);
}

int
AtExit(void (*func)(void))
{
    struct exit_proc *pExit = sysMalloc(sizeof(struct exit_proc));
    if (pExit == NULL) {
        return 1;
    }
    pExit->proc = func;
    pExit->next = exit_procs;
    exit_procs = pExit;
    return 0;
}


void
Abort(void)
{
    static bool_t entered = FALSE;
    if (entered)
        return;
    entered = TRUE;

    if (abort_hook) 
        (*abort_hook)();
    else
        abort();
}

long ThreadCPUTimeMillis(void)
{
    return jlong_low(jlong_div(sysThreadCPUTime(), int_to_jlong(1000000)));
}

#ifdef STATISTICS
void
DumpStatistics ()
{
    long totals[256], total = 0;
    int percent;
    int i, j;
    for (i = 0; i < 256; i++) {
	long sum = 0;
	for (j = 0; j < 256; j++) sum += instruction_count[i][j];
	totals[i] = sum;
	total += sum;
    }
    for (i = 0; i < 256; i++) { 
	if (totals[i])
	    jio_fprintf(stdout, "%-23.23s %8d\n", opnames[i], totals[i]);
    }
    jio_fprintf(stdout, "%-23.23s %8d\n", "TOTAL", total);
    percent = total/100;
    for (i = 0; i < 256; i++) {
	for (j = 0; j < 256; j++) { 
	    if (instruction_count[i][j] >= percent) { 
		jio_fprintf(stdout, "%-23.23s => %-23.23s %8d %4.2f (%4.2f)\n", 
		       opnames[i], opnames[j], instruction_count[i][j], 
		       instruction_count[i][j] * 100.0 / total, 
		       instruction_count[i][j] * 100.0 / totals[i]);
	    }
	}
    }

}
#endif

#ifdef TRACING
void
trace_method(struct execenv* ee, struct methodblock* mb,
	     int args_size, int type)
{
    HArrayOfChar* threadName = threadSelf() ? getThreadName() : NULL;
    JavaFrame *frame;
    int depth;
    int i;
    for (frame = ee->current_frame, depth = 0;
	   frame != NULL;
	   frame = frame->prev)
	if (frame->current_method != 0) depth++;
    if (type == TRACE_METHOD_RETURN)
	depth--;
    jio_fprintf(stdout, "# ");
    if (threadName) {
	int threadNameLen = obj_length(threadName);
	unicode *threadNameStr = unhand(threadName)->body;
	for (i = 0; i < threadNameLen; i++)
	    putchar((char)(threadNameStr[i]));
    }
    jio_fprintf(stdout, " [%2d] ", depth);
    for (i = 0; i < depth; i++)
	jio_fprintf(stdout, "| ");
    jio_fprintf(stdout, "%c %s.%s%s ",
	   ((type == TRACE_METHOD_ENTER) ? '>' : '<'),
	   cbName(fieldclass(&mb->fb)),
	   fieldname(&mb->fb),
	   fieldsig(&mb->fb));
    if (type == TRACE_METHOD_ENTER)
	jio_fprintf(stdout, "(%d) entered\n", args_size);
    else if (exceptionOccurred(ee))
	jio_fprintf(stdout, "throwing %s\n", cbName(obj_classblock(ee->exception.exc)));
    else
	jio_fprintf(stdout, "returning\n");
    fflush(stdout);
}
#endif
