/*
 * @(#)jit.c	1.33 98/08/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "exceptions.h"
#include "common_exceptions.h"
#include "java_lang_String.h"
#include "java_lang_Object.h"
#include "jit.h"

/* This version number should be updated whenever there are changes to
 * the interface between the interpreter and JIT compilers. */
/*
 * JDK 1.1.3   5
 * JDK 1.2     6
 */
#define COMPILER_VERSION 6

/**
 * JavaVersion is a unique number which changes whenever the interpreter
 * or compiler interface changes.  Compiler writers should treat this value
 * as a fixed constant (ie, don't extract the numbers used to create it), 
 * as how this number is determined may change in future releases.
 */  
static int JavaVersion = (COMPILER_VERSION << 24) 
				| (45 << 16)
                                | 3;
/* The two lines above used to be:
 *
 *                       | (JAVA_VERSION << 16) | JAVA_MINOR_VERSION;
 *
 * However, the value of both version numbers has changed in 1.2.
 * Because the JIT interface did not change at the same time, the value
 * of '45.3' is hard-wired (for the moment).  This may be changed in
 * the future.
 */


/* Flag set when the compiler becomes initialized */
bool_t compilerInitialized = FALSE;

/* The name of the attribute that represents compiled code on this machine. */
char *CompiledCodeAttribute = NULL;

/*
 * Java profiler related symbols. No longer supported.
 */
bool_t java_profiler_isOn = FALSE;

void 
java_profiler_log(struct methodblock* caller,
		  struct methodblock* callee,
		  int32_t cost)
{
}

void
java_profiler_log_frame(JavaFrame *frame, int32_t cost)
{
}

static void (*p_InitializeForCompiler)(ClassClass *cb) = 0;
static void (*p_CompilerFreeClass)(ClassClass *cb) = 0;
static bool_t (*p_CompilerRegisterNatives)(ClassClass *cb) = 0;
static bool_t (*p_CompilerUnregisterNatives)(ClassClass *cb) = 0;
static bool_t (*p_invokeCompiledMethod)(JHandle *o, struct methodblock *mb, 
			       int args_size, ExecEnv *ee) = 0; 
static bool_t (*p_CompiledCodeSignalHandler)(int sig, void *info, void *uc) = 0;

static bool_t (*p_CompilerCompileClass)(ClassClass *cb) = 0;
static bool_t (*p_CompilerCompileClasses)(Hjava_lang_String *name) = 0;
static Hjava_lang_Object *(*p_CompilerCommand)(Hjava_lang_Object *any) = 0;
static void (*p_CompilerEnable)(void) = 0;
static void (*p_CompilerDisable)(void) = 0;
static bool_t (*p_PCinCompiledCode)(unsigned char *, struct methodblock *) = 0;
static unsigned char *(*p_CompiledCodePC)(JavaFrame *frame, struct methodblock *mb) = 0;
static JavaFrame *(*p_CompiledFramePrev)(JavaFrame *frame, JavaFrame *buf) = 0;
static void *(*p_CompiledFrameID)(JavaFrame *frame) = 0;
static JavaFrame * (*p_CompiledFrameUpdate)(JavaFrame *frame) = 0;

static void (*p_ReadInCompiledCode)(void *context, 
				    struct methodblock *mb, 
				    int attribute_length, 
				    unsigned long (*get1byte)(void *),
				    unsigned long (*get2bytes)(void *), 
				    unsigned long (*get4bytes)(void *), 
				    void (*getNbytes)(void *, size_t, char *));


void InitializeForCompiler(ClassClass *cb) { 
    if (p_InitializeForCompiler != NULL) 
	p_InitializeForCompiler(cb);
}

void CompilerFreeClass(ClassClass *cb) { 
    if (p_CompilerFreeClass != NULL) 
	p_CompilerFreeClass(cb);
}


bool_t CompilerRegisterNatives(ClassClass *cb) { 
    if (p_CompilerRegisterNatives != NULL) 
	return p_CompilerRegisterNatives(cb);
    else
	return TRUE;
}

bool_t CompilerUnregisterNatives(ClassClass *cb) { 
    if (p_CompilerUnregisterNatives != NULL) 
	return p_CompilerUnregisterNatives(cb);
    else
	return TRUE;
}

bool_t 
invokeCompiledMethod(JHandle *o, struct methodblock *mb, 
		     int args_size, ExecEnv *ee) 
{
    if (p_invokeCompiledMethod == NULL) { 
	ThrowInternalError(ee, "Error! Compiled methods not supported");
	return FALSE;
    } else { 
	return p_invokeCompiledMethod(o, mb, args_size, ee);
    }
}

bool_t CompiledCodeSignalHandler(int sig, void *info, void *uc)
{ 
    if (p_CompiledCodeSignalHandler != NULL) 
	return p_CompiledCodeSignalHandler(sig, info, uc);
    else
	return FALSE; /* we didn't handle the signal */
}

void ReadInCompiledCode(void *context, struct methodblock *mb, 
			int attribute_length, 
                        unsigned long (*get1byte)(void *),
                        unsigned long (*get2bytes)(void *),
                        unsigned long (*get4bytes)(void *),
                        void (*getNbytes)(void *, size_t, char *))
{
    if (p_ReadInCompiledCode != NULL) { 
	p_ReadInCompiledCode(context, mb, attribute_length,
			        get1byte, get2bytes, get4bytes, getNbytes);
    } else { 
	getNbytes(context, attribute_length, NULL);
    }
}

bool_t PCinCompiledCode(unsigned char *pc, struct methodblock *mb) { 
    return (p_PCinCompiledCode != NULL) && p_PCinCompiledCode(pc, mb);
}

unsigned char *CompiledCodePC(JavaFrame *frame, struct methodblock *mb) {
    return (p_CompiledCodePC == 0) ? NULL : p_CompiledCodePC(frame, mb);
}

JavaFrame *CompiledFramePrev(JavaFrame *frame, JavaFrame *buf) {
    return (p_CompiledFramePrev == 0) 
	? frame->prev 
	: p_CompiledFramePrev(frame, buf);
}

void *CompiledFrameID(JavaFrame *frame) {
    return (p_CompiledFrameID == 0) ? frame : p_CompiledFrameID(frame);
}

JavaFrame * CompiledFrameUpdate(JavaFrame *frame)
{
    if (p_CompiledFrameUpdate != 0)
	return p_CompiledFrameUpdate(frame);
    else
	return frame;
}

static JITInterface6 jitinterface;

static void
initJITInterface(void)
{
    memset(&jitinterface, 0, sizeof(jitinterface));
    jitinterface.base.JavaVersion = &JavaVersion;
    
    jitinterface.p_InitializeForCompiler = &p_InitializeForCompiler;
    jitinterface.p_invokeCompiledMethod = &p_invokeCompiledMethod;
    jitinterface.p_CompiledCodeSignalHandler = &p_CompiledCodeSignalHandler;
    jitinterface.p_CompilerFreeClass = &p_CompilerFreeClass;
    jitinterface.p_CompilerCompileClass = &p_CompilerCompileClass;
    jitinterface.p_CompilerCompileClasses = &p_CompilerCompileClasses;
    jitinterface.p_CompilerEnable = &p_CompilerEnable;
    jitinterface.p_CompilerDisable = &p_CompilerDisable;
    jitinterface.p_CompiledFramePrev = &p_CompiledFramePrev;
    jitinterface.p_CompiledFrameUpdate = &p_CompiledFrameUpdate;
    jitinterface.p_CompiledFrameID = &p_CompiledFrameID;
    jitinterface.p_ReadInCompiledCode = &p_ReadInCompiledCode;
    jitinterface.p_PCinCompiledCode = &p_PCinCompiledCode;
    jitinterface.p_CompiledCodePC = &p_CompiledCodePC;
    jitinterface.p_CompiledCodeAttribute = &CompiledCodeAttribute;
    jitinterface.p_CompilerRegisterNatives = &p_CompilerRegisterNatives;
    jitinterface.p_CompilerUnregisterNatives = &p_CompilerUnregisterNatives;

    jitinterface.UseLosslessQuickOpcodes = &UseLosslessQuickOpcodes;
    
    jitinterface.sysMalloc = hpi_memory_interface->Malloc;
    jitinterface.sysCalloc = hpi_memory_interface->Calloc;
    jitinterface.sysRealloc = hpi_memory_interface->Realloc;
    jitinterface.sysFree = hpi_memory_interface->Free;
    jitinterface.sysStrdup = hpi_memory_interface->Strdup;
    
    jitinterface.binclasses = &binclasses;
    jitinterface.nbinclasses = &nbinclasses;
    jitinterface.binclass_lock = &_binclass_lock;
    jitinterface.linkclass_lock = &_linkclass_lock;
    
    jitinterface.classJavaLangClass = &classJavaLangClass;
    jitinterface.classJavaLangObject = &classJavaLangObject;
    jitinterface.classJavaLangString = &classJavaLangString;
    jitinterface.classJavaLangThrowable = &classJavaLangThrowable;
    jitinterface.classJavaLangException = &classJavaLangException;
    jitinterface.classJavaLangRuntimeException =
	&classJavaLangRuntimeException;
    jitinterface.interfaceJavaLangCloneable = &interfaceJavaLangCloneable;
    
    jitinterface.EE = &EE;
    jitinterface.SignalError = &SignalError;
    jitinterface.exceptionInternalObject = &exceptionInternalObject;
    
    jitinterface.GetClassConstantClassName = &GetClassConstantClassName;
    jitinterface.ResolveClassConstant2 = &ResolveClassConstant2;
    jitinterface.ResolveClassConstantFromClass2 =
	&ResolveClassConstantFromClass2;
    jitinterface.VerifyClassAccess = &VerifyClassAccess;
    jitinterface.VerifyFieldAccess = &VerifyFieldAccess;
    jitinterface.FindClass = &FindClass;
    jitinterface.FindClassFromClass = &FindClassFromClass;
    jitinterface.InitClass = &InitClass;

    jitinterface.dynoLink = &dynoLink;

    jitinterface.do_execute_java_method_vararg = &do_execute_java_method_vararg;
    jitinterface.execute_java_static_method = &execute_java_static_method;

    jitinterface.is_subclass_of = &is_subclass_of;
    
    jitinterface.invokeJavaMethod = &invokeJavaMethod;
    jitinterface.invokeSynchronizedJavaMethod = &invokeSynchronizedJavaMethod;
    jitinterface.invokeAbstractMethod = &invokeAbstractMethod;
    jitinterface.invokeLazyNativeMethod = &invokeLazyNativeMethod;
    jitinterface.invokeSynchronizedNativeMethod =
	&invokeSynchronizedNativeMethod; 
    jitinterface.invokeCompiledMethod = &invokeCompiledMethod; 
    jitinterface.invokeNativeMethod = &invokeNativeMethod;
    jitinterface.invokeJNINativeMethod = &invokeJNINativeMethod;
    jitinterface.invokeJNISynchronizedNativeMethod =
	&invokeJNISynchronizedNativeMethod;
    
    jitinterface.monitorEnter2 = &monitorEnter2;
    jitinterface.monitorExit2 = &monitorExit2;
    jitinterface.monitorWait2 = &monitorWait2;
    jitinterface.monitorNotify2 = &monitorNotify2;
    jitinterface.monitorNotifyAll2 = &monitorNotifyAll2;
    jitinterface.monitorRegister = &monitorRegister;
    jitinterface.sysMonitorSizeof = hpi_thread_interface->MonitorSizeof;
    jitinterface.sysMonitorEnter = hpi_thread_interface->MonitorEnter;
    jitinterface.sysMonitorExit = hpi_thread_interface->MonitorExit;
    jitinterface.sysMonitorEntered = hpi_thread_interface->MonitorEntered;
    jitinterface.sysMonitorWait = hpi_thread_interface->MonitorWait;
    jitinterface.sysMonitorNotify = hpi_thread_interface->MonitorNotify;
    jitinterface.sysMonitorNotifyAll = hpi_thread_interface->MonitorNotifyAll;
    
    jitinterface.allocObject = &allocObject;
    jitinterface.allocArray = &allocArray;
    jitinterface.MultiArrayAlloc = &MultiArrayAlloc;
    jitinterface.sizearray = &sizearray;
    jitinterface.is_instance_of = &is_instance_of;
    jitinterface.classname2string = &classname2string;
    jitinterface.DumpThreads = &DumpThreads;
    
    jitinterface.ExpandJavaStackForJNI = &ExpandJavaStackForJNI;
    jitinterface.ExpandJavaStack = &ExpandJavaStack;

    jitinterface.ExecuteJava = pExecuteJava;
    
    jitinterface.now = ThreadCPUTimeMillis;

    /* Cast bool_t* to int* to shutup compiler. */
    jitinterface.java_profiler_isOn = (int *)&java_profiler_isOn;
    jitinterface.java_profiler_log = &java_profiler_log;
    jitinterface.JavaStackSize = &JavaStackSize;

    jitinterface.jio_snprintf = &jio_snprintf;
    jitinterface.javaString2CString = &javaString2CString;
    
    jitinterface.jni_mkRefLocal = &jni_mkRefLocal;
    jitinterface.cacheAlloc = &cacheAlloc;

    jitinterface.jvmpi_event_flags = &jvmpi_event_flags;
    jitinterface.getCustomInvoker = &getCustomInvoker;

    jitinterface.code_lock = &_code_lock;

    /* support for JVM<--->Profiler interface from the JIT */
    jitinterface.jvmpi_method_entry = &jvmpi_method_entry;
    jitinterface.jvmpi_method_exit = &jvmpi_method_exit;
    jitinterface.jvmpi_load_compiled_method = &jvmpi_load_compiled_method;
    jitinterface.jvmpi_unload_compiled_method = &jvmpi_unload_compiled_method;
    
}

JNIEXPORT
void JNICALL JVM_InitializeCompiler(JNIEnv *env, jclass compCls)
{
    void *address =
	(void *)classLoaderLink(NULL, "java_lang_Compiler_start", 0);
    if (address != 0) { 
	initJITInterface();
	(*(void (*) (JITInterface *)) address) ((JITInterface *)&jitinterface);
    }
    compilerInitialized = TRUE;
}

JNIEXPORT
jboolean JNICALL JVM_IsSilentCompiler(JNIEnv *env, jclass compCls)
{
  return TRUE;
}


JNIEXPORT
jboolean JNICALL JVM_CompileClass(JNIEnv *env, jclass compCls, jclass cls)
{
    Hjava_lang_Class *clazz = (Hjava_lang_Class *)DeRef(env, cls);

    if (clazz == NULL) { 
	ThrowNullPointerException(0, 0);
	return FALSE;
    } else if (p_CompilerCompileClass != NULL) { 
	return p_CompilerCompileClass(clazz);
    } else { 
	return FALSE;
    }
}

JNIEXPORT
jboolean JNICALL JVM_CompileClasses(JNIEnv *env, jclass compCls, jstring jname)
{
    Hjava_lang_String *name = (Hjava_lang_String *)DeRef(env, jname);

    if (name == NULL) { 
	ThrowNullPointerException(0, 0);
	return FALSE;
    } else if (p_CompilerCompileClasses != NULL) { 
	return p_CompilerCompileClasses(name);
    } else { 
	return FALSE;
    }
}

JNIEXPORT
jobject JNICALL JVM_CompilerCommand(JNIEnv *env, jclass compCls, jobject arg)
{
    Hjava_lang_Object *x = (Hjava_lang_Object *)DeRef(env, arg);

    if (x == NULL) { 
	ThrowNullPointerException(0, 0);
	return NULL;
    } else if (p_CompilerCommand != NULL) { 
	return MkRefLocal(env, p_CompilerCommand(x));
    } else { 
	return NULL;
    }
}


JNIEXPORT
void JNICALL JVM_EnableCompiler(JNIEnv *env, jclass compCls)
{
    if (p_CompilerEnable != NULL) 
	p_CompilerEnable();
}


JNIEXPORT
void JNICALL JVM_DisableCompiler(JNIEnv *env, jclass compCls)
{
    if (p_CompilerDisable != NULL) 
	p_CompilerDisable();
}
