/*
 * @(#)jni.c	1.132 98/11/03
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/******************************************************************************
 * Java Runtime Interface -- JDK Implementation
 * Copyright (c) 1996 Netscape Communications Corporation. All rights reserved.
 *****************************************************************************/

#include "oobj.h"
#include "gc.h"
#include "interpreter.h"
#include "threads.h"
#include "exceptions.h"
#include "opcodes.h"
#include "java_lang_Throwable.h"
#include "java_lang_ClassLoader.h"
#include "utf.h"
#include "reflect.h"
#include "common_exceptions.h"
#include "util.h"

#ifdef HPROF
#include "vmprofiler.h"
#include "jvmpi.h"
#endif /* HPROF */

#ifdef BREAKPTS
#include "breakpoints.h"
#include "jvmdi.h"
#endif /* BREAKPTS */

#include <sys/stat.h>
#include <fcntl.h>
#include <limits.h>

#define INVOKE_VIRTUAL    0x0100
#define INVOKE_NONVIRTUAL 0x0200
#define INVOKE_STATIC     0x0300

#define INVOKE_MASK       0xff00

/*
 * forward declarations
 */
static jint JNICALL 
jni_EnsureLocalCapacity(JNIEnv *env, jint capacity);

static void JNICALL
jni_DeleteLocalRef(JNIEnv *env, jobject obj);

static void JNICALL
jni_FatalError(JNIEnv *env, const char *msg);

static jmethodID JNICALL
jni_GetMethodID(JNIEnv *env, jclass clazz, const char *name, const char *sig);

static jobject JNICALL
jni_NewObject(JNIEnv *env, jclass clazz, jmethodID methodID, ...);

static jstring JNICALL
jni_NewStringUTF(JNIEnv *env, const char *bytes);

static jobject JNICALL 
jni_AllocObject(JNIEnv *env, jclass clazz);

static jint JNICALL
jni_GetJavaVM(JNIEnv *env, JavaVM **vm);

static void jni_delRef(JavaFrame *frame, JHandle **handle);
static JHandle **jni_addRef(JavaFrame *frame, JHandle *handle);

#define DeRef2(env, ref) (*(JHandle **)(ref))

void bad_critical_count_on_return(void)
{
    jni_FatalError(EE2JNIEnv(EE()),
		   "Leaving arrays locked on native method return");
}

/* 
 * With the Invocation API, JNI functions may be called anywhere in the 
 * native code, without know how the current stack pointer relate to the
 * recorded stack base.
 *
 * Every JNI function thus checks if the stack base needs to be set; and
 * if so, adjust the stack base to refect the current native stack setting.
 */

#define RECORD_STACK_BASE(env, base) \
    void *_stack_base = JNIEnv2EE(env)->stack_base; \
    int _x = (_stack_base == NULL) ? \
                ((JNIEnv2EE(env)->stack_base = (void *)base), 0) : \
                1

#define RECOVER_STACK_BASE(env) \
    JNIEnv2EE(env)->stack_base = _stack_base

const struct JNINativeInterface_ *jni_NativeInterfacePtr;

extern const struct JNIInvokeInterface_ jni_InvokeInterface;

/*
 * global variables 
 */

static int VM_created = 0;

/* 
 * The main EE is stored in the reserved0 field.
 */
static struct JavaVM_ main_vm = {&jni_InvokeInterface};

/*
 * The JNI Global References Lock:
 *
 * A thread must grab the lock before it can ADD to or DELETE from the global
 * reference table. The thread, however, does _not_ need to grab the lock 
 * just to READ a global reference table entry.
 */

JavaFrame *globalRefFrame = 0;

sys_mon_t *_globalref_lock = 0;

/* Multiple threads can pin the same object.
 * We must enter pinned objects into a global table, where a count
 * is kept.
 */

/* JNI pinning lock */
static sys_mon_t *_pinning_lock = 0;
#define JNI_PINNING_LOCK_INIT() \
     monitorRegister(_pinning_lock, "JNI pinning lock")
#define JNI_PINNING_LOCK(self)   sysMonitorEnter(self, _pinning_lock)
#define JNI_PINNING_LOCKED(self) sysMonitorEntered(self, _pinning_lock)
#define JNI_PINNING_UNLOCK(self) sysMonitorExit(self, _pinning_lock)

/* Currently, only JNI pins objects explicitly. Therefore we optimize
 * a little by only entering those objects that are pinned more than
 * once into the table.
 * REVIST THIS IF OTHER PARTS OF THE SYSTEM ALSO PINS OBJECTS!!!
 */

typedef struct PinnedObjBucket {
    int count;
    void *object;
    struct PinnedObjBucket *next;
} PinnedObjBucket;

#define PINNED_OBJ_TABLE_SIZE 151 /* a prime number */

static PinnedObjBucket *pinnedObjTable[PINNED_OBJ_TABLE_SIZE];

static PinnedObjBucket *alloced_buckets = NULL;
static int n_alloced_buckets = 0;
#define N_RESERVED_BUCKETS 10

static int pinObj(ExecEnv *ee, void *obj)
{
    unsigned int hash = (uintptr_t)obj % PINNED_OBJ_TABLE_SIZE;
    PinnedObjBucket *bucket;
    int result = TRUE;

    JNI_PINNING_LOCK(EE2SysThread(ee));

    bucket = pinnedObjTable[hash];

    while (bucket && bucket->object != obj) {
        bucket = bucket->next;
    }
#ifdef JNI_DEBUG
    jio_fprintf(stderr, "  pinObj: %08x, bucket = %08x, count = %d\n",
		obj, bucket, bucket ? bucket->count : 0);
#endif
    if (bucket == 0) {
        if (!pinned_object(obj)) {
	    pin_object(obj);
	} else {
	    if (alloced_buckets) {
	        bucket = alloced_buckets;
		alloced_buckets = alloced_buckets->next;
		n_alloced_buckets--;
	    } else {
	        bucket = (PinnedObjBucket *)sysMalloc(sizeof(PinnedObjBucket));
	    }
	    if (bucket == 0) {
	        result = FALSE;
	    } else {
	        bucket->next = pinnedObjTable[hash];
		bucket->count = 2;
		bucket->object = obj;
		pinnedObjTable[hash] = bucket;
	    }
	}
    } else
        bucket->count++;

    JNI_PINNING_UNLOCK(EE2SysThread(ee));
    return result;
}

static void unpinObj(ExecEnv *ee, void *obj)
{
    unsigned int hash = (uintptr_t)obj % PINNED_OBJ_TABLE_SIZE;
    PinnedObjBucket **last = &(pinnedObjTable[hash]);
    PinnedObjBucket *bucket;

    JNI_PINNING_LOCK(EE2SysThread(ee));

    bucket = pinnedObjTable[hash];

    while (bucket && bucket->object != obj) {
        last = &(bucket->next);
        bucket = bucket->next;
    }
#ifdef JNI_DEBUG
    jio_fprintf(stderr, "unpinObj: %08x, bucket = %08x, count = %d\n",
		obj, bucket, bucket ? bucket->count : 0);
#endif
    if (bucket == 0) {
        if (!pinned_object(obj))
	    jni_FatalError(EE2JNIEnv(ee), 
			   "Try to unpin an object that is not pinned");
	unpin_object(obj);
    } else
        if (--bucket->count == 1) {
	    *last = bucket->next;
	    if (n_alloced_buckets > N_RESERVED_BUCKETS) {
	        sysFree(bucket);
	    } else {
	        bucket->next = alloced_buckets;
	        alloced_buckets = bucket;
		n_alloced_buckets++;
	    }
	}

    JNI_PINNING_UNLOCK(EE2SysThread(ee));
}

/* Returns FALSE on failure */

int InitializeJNI(void)
{
    int i;
    JavaStack *globalStack;

    for(i=0;i<PINNED_OBJ_TABLE_SIZE;i++)
        pinnedObjTable[i] = 0;

    if (_globalref_lock == 0) {
        _globalref_lock=(sys_mon_t *) sysMalloc(sysMonitorSizeof());
	if (_globalref_lock == 0)
	    return FALSE;
	GLOBALREF_LOCK_INIT();
    }
    
    if (_pinning_lock == 0) {
        _pinning_lock=(sys_mon_t *) sysMalloc(sysMonitorSizeof());
	if (_pinning_lock == 0)
	    return FALSE;
	JNI_PINNING_LOCK_INIT();
    }

    globalStack = CreateNewJavaStack(0, 0, 0);
    if (globalStack == 0)
        return FALSE;

    globalRefFrame = (JavaFrame*)globalStack->data;
    memset(globalRefFrame, 0, sizeof(JavaFrame));
    globalRefFrame->optop = globalRefFrame->ostack + JNI_REF_INFO_SIZE;
    globalRefFrame->javastack = globalStack;

    JNI_REFS_FREELIST(globalRefFrame) = NULL;
    JNI_N_REFS_IN_USE(globalRefFrame) = 0;
    JNI_REFS_CAPACITY(globalRefFrame) = 0x7FFFFFFF; /* a huge upper bound */

    return TRUE;
}

void InitializeJNIRootFrame(ExecEnv *ee)
{
    JavaFrame *frame;
    struct JNIEnv_ *env_ = &ee->nativeInterface;

    env_->functions = jni_NativeInterfacePtr;

    frame = (JavaFrame *)(ee->initial_stack->data);
    memset(frame, 0, sizeof(JavaFrame));

    frame->javastack = ee->initial_stack;
    frame->optop = frame->ostack + JNI_REF_INFO_SIZE;

    JNI_REFS_FREELIST(frame) = NULL;
    JNI_N_REFS_IN_USE(frame) = 0;
    JNI_REFS_CAPACITY(frame) = JNI_DEFAULT_LOCAL_CAPACITY;

    ee->current_frame = frame;
}

void DeleteJNIRootFrame(ExecEnv *ee)
{
}

bool_t verbose_jni = FALSE;

static JHandle **jni_addRef(JavaFrame *frame, JHandle *handle)
{
    JHandle **ref;
    if (handle == 0)
        return 0;

    /* Make sure we are in a native frame or a pseudo frame. */
    sysAssert((frame->current_method == 0) ||
	      (frame->current_method->fb.access & ACC_NATIVE));
#ifdef JNI_DEBUG
    jio_fprintf(stderr, "Adding %s Ref #%d in frame %08x.\n",
		frame == globalRefFrame ? "global" : "local",
		JNI_N_REFS_IN_USE(frame), frame);
#endif

    if (verbose_jni) {
        if (++JNI_N_REFS_IN_USE(frame) > JNI_REFS_CAPACITY(frame)) {
	    jio_fprintf(stderr, 
			"***ALERT: JNI local ref creation exceeded capacity "
			"(creating: %d, limit: %d).\n",
			JNI_N_REFS_IN_USE(frame), 
			JNI_REFS_CAPACITY(frame));
       	    printStackTrace(EE(), 10, NULL);
        }
    }

    if (JNI_REFS_FREELIST(frame)) {
        ref = (JHandle**)(JNI_REFS_FREELIST(frame));
        *(JHandle**)(&JNI_REFS_FREELIST(frame)) = *ref;
    } else {
        JavaStack *javastack = frame->javastack;
        if (frame->optop >= javastack->end_data) {
	    JavaStack *tmp_stack = javastack;
	    JavaFrame *tmp_frame = frame;
	    ExecEnv *ee = EE();
	    if (!ExpandJavaStackForJNI(ee,
				       &tmp_stack,
				       &tmp_frame,
				       1)) {
	        if (frame != globalRefFrame) 
		    jni_FatalError(EE2JNIEnv(ee), 
				   "Out of memory when expanding local ref "
				   "table beyond capacity");
		else {
		    if (verbose_jni) {
		        jio_fprintf(stderr, 
			    "!!!JNI global ref creation exceeded capacity\n");
		    }
		    /* exception already thrown */
		    return 0;
		}
		    
            }
            frame->javastack = tmp_stack;
	    frame->optop = (stack_item *)tmp_frame;
        }
        ref = (JHandle**)(frame->optop);
        frame->optop++;
    }
    *ref = handle;
    return ref;
}

static void jni_delRef(JavaFrame *frame, JHandle **handle)
{
    if (handle == 0)
        return;

    /* Make sure we are in a native frame or a pseudo frame. */
    sysAssert((frame->current_method == 0) ||
              (frame->current_method->fb.access & ACC_NATIVE));
		
    if ((stack_item *)handle >= frame->ostack + JNI_REF_INFO_SIZE &&
        IN_JAVASTACK((stack_item *)frame, frame->javastack)) {
	/* Frequent case: the JNI frame does not span multiple
	 * stack segments.
	 */
	goto found;
    }

    {
        JavaStack *javastack = frame->javastack;
	struct methodblock *mb = frame->current_method;
	/* First, the extra stack segments. */
        while (!IN_JAVASTACK((stack_item*)frame, javastack)) {
	    if (IN_JAVASTACK((stack_item*)handle, javastack)) {
	        goto found;
	    }
	    javastack = javastack->prev;
	}
	/* Second, check the first stack segment. */
	if ((stack_item *)handle < javastack->end_data && 
	    (stack_item *)handle >= frame->ostack + JNI_REF_INFO_SIZE) {
	    goto found;
	}
	/* Third, check the arguments. */
	if (mb->fb.access & ACC_NATIVE) {
	    stack_item *prev_optop = frame->prev->optop;
	    if ((stack_item *)handle >= prev_optop &&
		(stack_item *)handle < prev_optop + mb->args_size) {
		goto found;
	    }
	}
    }

    jni_FatalError(EE2JNIEnv(EE()), 
		   "DeleteLocalRef: ref does not exist in the current frame");

 found:
    if (verbose_jni)
        --JNI_N_REFS_IN_USE(frame);
    *handle = (JHandle *)(JNI_REFS_FREELIST(frame));
    *(JHandle ***)(&JNI_REFS_FREELIST(frame)) = handle;
}

/* For JIT interface only. */
jobject jni_mkRefLocal(ExecEnv *ee, JHandle *jobj)
{
    return (jobject) jni_addRef(ee->current_frame, jobj);
}

/*
 * General Stuff
 */

#define DeRefClass(env, clazz) ((ClassClass *)DeRef((env), (clazz)))

static struct fieldblock *
jni_FindFieldBlock0(ClassClass *cb, HashedNameAndType *hashed, int searchSuper)
{
    struct fieldblock *fb;
    int i, nFields;
    while (cb) {
        /* search in this class (and superclasses)'s fields: */
        nFields = cbFieldsCount(cb);
        fb = cbFields(cb);
        for (i = 0; i < nFields; i++, fb++) {
	    if (NAMETYPE_MATCH(hashed, fb)) {
                return fb;
	    }
        }
	cb = searchSuper ? cbSuperclass(cb) : 0;
    }
    return 0;
}

static struct fieldblock*
jni_FindFieldBlock(JNIEnv *env, jclass cls, const char *name, 
		   const char *sig, int isStatic)
{
    RECORD_STACK_BASE(env, &env);
    ExecEnv *ee = JNIEnv2EE(env);
    struct fieldblock *fb;
    HashedNameAndType hashed;
    ClassClass *cb = DeRefClass(env, cls);

    HashNameAndType(ee, (char*)name, (char*)sig, &hashed);

    InitClass(cb);
    if (exceptionOccurred(ee)) {
        RECOVER_STACK_BASE(env);
        return NULL;
    }

    if ((fb = jni_FindFieldBlock0(cb, &hashed, TRUE)))
        goto done;

    /* Check interface members this class may have inherited. */
    if (isStatic) {
        struct imethodtable *imt = cbIntfMethodTable(cb);
	int j, icnt = imt->icount;
	for (j = 0; j < icnt; j++) {
            ClassClass * icb = imt->itable[j].classdescriptor;
	    if ((fb = jni_FindFieldBlock0(icb, &hashed, FALSE)))
	        goto done;
	}
    }

    fb = NULL;

done:
    if ((fb == NULL) || 
	(((fb->access & ACC_STATIC) == ACC_STATIC) != isStatic)) {
        ThrowNoSuchFieldError(ee, (char *)name);
        RECOVER_STACK_BASE(env);
        return NULL;
    }
    RECOVER_STACK_BASE(env);
    return fb;
}

static struct methodblock *
jni_FindMethodBlock0(ClassClass *cb, HashedNameAndType *hashed, 
		     int searchSuper)
{
    int i;
    struct methodblock *mb;
    if (cb == NULL) {
        return NULL;
    }
    for (i = cbMethodsCount(cb) - 1; i >= 0; i--) {
        mb = &(cbMethods(cb)[i]);
        if (NAMETYPE_MATCH(hashed, &(mb->fb))) {
	    return mb;
	}
    }
    if (!searchSuper) {
        return NULL;
    }

    mb = jni_FindMethodBlock0(cbSuperclass(cb), hashed, TRUE);
    if (mb) {
        return mb;
    }

    /* Look for methods inherited from super interfaces (including
     * miranda methods).
     */
    {
        struct imethodtable *imt = cbIntfMethodTable(cb);
	int j, icnt = imt->icount;
	for (j = 0; j < icnt; j++) {
            ClassClass * icb = imt->itable[j].classdescriptor;
	    if ((mb = jni_FindMethodBlock0(icb, hashed, FALSE))) {
	        return mb;
	    }
	}
    }
    return NULL;
}

static struct methodblock*
jni_FindMethodBlock(JNIEnv *env, jclass cls, const char *name,
		    const char *sig, int searchSuper)
{
    RECORD_STACK_BASE(env, &env);
    ExecEnv *ee = JNIEnv2EE(env);
    HashedNameAndType hashed;
    struct methodblock *mb = NULL;
    ClassClass *cb = DeRefClass(env, cls);

    HashNameAndType(ee, (char *)name, (char *)sig, &hashed);

    InitClass(cb);
    if (exceptionOccurred(ee)) {
        goto done;
    }

    /* never search superclasses for constructors */
    if (name[0] == '<') {
        searchSuper = FALSE;
    }
    mb = jni_FindMethodBlock0(cb, &hashed, searchSuper);
    if (mb == NULL) {
        ThrowNoSuchMethodError(ee, (char *)name);
    }

 done:
    RECOVER_STACK_BASE(env);
    return mb;
}

typedef char*
(*JNI_PushArguments_t)(JNIEnv *env, char *terse_signature,
                       JavaFrame *current_frame, void *args);

static char*
jni_PushArgumentsVararg(JNIEnv *env, char *terse_signature,
			JavaFrame *current_frame, void *a)
{
    char *p = terse_signature;
    jvalue tdub;
    va_list args;

    /* Assignment to va_list does not work on certain platforms:
     * args = *(va_list *)a;
     */
    memcpy(&args, a, sizeof(va_list));

    while (1) {
        switch (*p++) {
	case TERSE_SIG_BOOLEAN:
	case TERSE_SIG_SHORT:
	case TERSE_SIG_BYTE:
	case TERSE_SIG_CHAR:
	case TERSE_SIG_INT:
            (current_frame->optop++)->i = va_arg(args, long);
            continue;
	case TERSE_SIG_FLOAT:
	    (current_frame->optop++)->f = (float)va_arg(args, double);
            continue;
        case TERSE_SIG_OBJECT: {
	    jobject obj = va_arg(args, jobject);
	    (current_frame->optop++)->h = (JHandle *)DeRef(env, obj);
	    continue;
	}
	case TERSE_SIG_LONG:
            SET_INT64(tdub, current_frame->optop, va_arg(args, int64_t));
            current_frame->optop += 2;
            continue;
	case TERSE_SIG_DOUBLE:
	    SET_DOUBLE(tdub, current_frame->optop, va_arg(args, double));
            current_frame->optop += 2;
            continue;
	case TERSE_SIG_ENDFUNC:
	    return p;
	default:
	    sysAssert(FALSE);
	    return NULL;
        }
    }
}

static char*
jni_PushArgumentsArray(JNIEnv *env, char *terse_signature,
                       JavaFrame *current_frame, void *a)
{
    char *p = terse_signature;
    jvalue tdub;
    jvalue *args = *(jvalue **)a;

    /*CONSTCOND*/
    while (1) {
        switch (*p++) {
	case TERSE_SIG_BOOLEAN:
            (current_frame->optop++)->i = (*args++).z;
            continue;
	case TERSE_SIG_SHORT:
            (current_frame->optop++)->i = (*args++).s;
            continue;
	case TERSE_SIG_BYTE:
            (current_frame->optop++)->i = (*args++).b;
            continue;
	case TERSE_SIG_CHAR:
            (current_frame->optop++)->i = (*args++).c;
            continue;
	case TERSE_SIG_INT:
            (current_frame->optop++)->i = (*args++).i;
            continue;
	case TERSE_SIG_FLOAT:
	    (current_frame->optop++)->f = (*args++).f;
            continue;
	case TERSE_SIG_OBJECT: {
	    jobject obj = (*args++).l;
	    (current_frame->optop++)->h = (JHandle *)DeRef(env, obj);
	    continue;
	}
	case TERSE_SIG_LONG:
            SET_INT64(tdub, current_frame->optop, (*args++).j);
            current_frame->optop += 2;
            continue;
	case TERSE_SIG_DOUBLE:
            SET_DOUBLE(tdub, current_frame->optop, (*args++).d);
            current_frame->optop += 2;
            continue;
        case TERSE_SIG_ENDFUNC:
	    return p;
	default:
            sysAssert(FALSE);
	    return NULL;
        }
    }
}

static jvalue
jni_Invoke(JNIEnv *env, jobject self, jmethodID methodID,
           JNI_PushArguments_t pushArguments, void *args,
           int info)
{
    int invokeKind = info & INVOKE_MASK;
    char reqType = info & 0xff;
    RECORD_STACK_BASE(env, &env);
    ExecEnv *ee = JNIEnv2EE(env);
    jvalue result;
    HObject *obj = NULL;	/* Keep lint happy */
    ClassClass *cb;
    struct methodblock *mb = (struct methodblock*)methodID;
    JavaFrame *frame;
    char  *retType;

    unsigned char pc[6];
    cp_item_type  constant_pool[2];
    unsigned char cpt[2];
    
    constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].p = cpt;
    cpt[0] = CONSTANT_POOL_ENTRY_RESOLVED;

    /* Leave pc[0] and pc[3+] to be filled in later. */
    pc[1] = 0; pc[2] = 1;    /* constant pool entry #1 */

    constant_pool[1].mb = mb;
    cpt[1] = CONSTANT_POOL_ENTRY_RESOLVED | CONSTANT_Methodref;

    result.i = 0; /* so that lint won't complain */
    if (invokeKind == INVOKE_STATIC) {
        cb = DeRefClass(env, self);
	if (!(mb->fb.access & ACC_STATIC)) {
	    jni_FatalError(env, 
		 "Try to invoke an instance method as a static method");
	}
	if (!is_subclass_of(cb, mb->fb.clazz, ee)) {
	    jni_FatalError(env, "Wrong class passed to a static method");
        }
	pc[0] = opc_invokestatic_quick;
	pc[3] = opc_return;
    } else {
        obj = DeRef(env, self);
        if (obj == NULL) {
	    jni_FatalError(env, "Null object passed to a non-static method");
        }
        cb = obj_array_classblock(obj);

	if (mb->fb.access & ACC_STATIC) {
	    jni_FatalError(env, 
	        "Try to invoke a static method as a non-static method");
	}
	if (!is_subclass_of(cb, mb->fb.clazz, ee)) {
	    jni_FatalError(env,
			   "Wrong method ID used to invoke a Java method");
	}
        
	if (cbIsInterface(mb->fb.clazz)) {
	    /* interface method */
	    pc[0] = opc_invokeinterface_quick;
	    pc[3] = mb->args_size;
	    pc[4] = 0;
	    pc[5] = opc_return;
	} else {
	    if (mb->fb.access & ACC_PRIVATE) {
	        /* private */
		pc[0] = opc_invokenonvirtual_quick;
		pc[3] = opc_return;
	    } else if (invokeKind == INVOKE_NONVIRTUAL ||
		       mb->fb.name == utf8_literal_obj_init_name) {
	        /* nonvirtual or initializer */ 
		pc[0] = opc_invokenonvirtual_quick;
		pc[3] = opc_return;
	    } else {
                /* virtual method */
	        pc[0] = opc_invokevirtual_quick_w;
		pc[3] = opc_return;
            }
        }
    }
    
    if ((*env)->PushLocalFrame(env, mb->args_size < 2 ? 2 : mb->args_size)) {
        goto done;
    }
    frame = ee->current_frame;
    frame->monitor = NULL;
    frame->constant_pool = constant_pool;

    /* Push the target object, if not a static call */
    if (invokeKind != INVOKE_STATIC) 
        (frame->optop++)->p = obj;
    
    retType = pushArguments(env, mb->terse_signature, frame, args);

    if (*retType != reqType) {
         jni_FatalError(env,
	      "Native code expects wrong return type from Java callback");
    }

    /* Run the byte codes outside the lock and catch any exceptions. */
    exceptionClear(ee);
    if ((*pExecuteJava)(pc, ee)) {
        switch (reqType) {
	case TERSE_SIG_VOID:
	    (*env)->PopLocalFrame(env, NULL);
	    break;
	case TERSE_SIG_BOOLEAN:
	    result.z = (jboolean)frame->optop[-1].i;
	    (*env)->PopLocalFrame(env, NULL);
	    break;
	case TERSE_SIG_BYTE:
	    result.b = (jbyte)frame->optop[-1].i;
	    (*env)->PopLocalFrame(env, NULL);
	    break;
	case TERSE_SIG_CHAR:
	    result.c = (jchar)frame->optop[-1].i;
	    (*env)->PopLocalFrame(env, NULL);
	    break;
	case TERSE_SIG_SHORT:
	    result.s = (jshort)frame->optop[-1].i;
	    (*env)->PopLocalFrame(env, NULL);
	    break;
	case TERSE_SIG_INT:
	    result.i = (jint)frame->optop[-1].i;
	    (*env)->PopLocalFrame(env, NULL);
	    break;
	case TERSE_SIG_LONG: {
	    jvalue tmp;
	    result.j = GET_INT64(tmp, &frame->optop[-2].i);
	    (*env)->PopLocalFrame(env, NULL);
	    break;
	}
	case TERSE_SIG_FLOAT:
	    result.f = (jfloat)frame->optop[-1].f;
	    (*env)->PopLocalFrame(env, NULL);
	    break;
	case TERSE_SIG_DOUBLE: {
	    jvalue tmp;
	    result.d = GET_DOUBLE(tmp, &frame->optop[-2].i);
	    (*env)->PopLocalFrame(env, NULL);
	    break;
	}
	case TERSE_SIG_OBJECT: {
	    void *res = frame->optop[-1].p;
	    /* Call MkRefLocal *after* the old frame has been recovered. */
	    (*env)->PopLocalFrame(env, NULL);
	    result.l = MkRefLocal(env, res);
	    break;
	}
	}
    } else {
        (*env)->PopLocalFrame(env, NULL);
    }
 done:
    RECOVER_STACK_BASE(env);
    return result;
}

/* Version information */

static jint JNICALL
jni_GetVersion(JNIEnv *env)
{
    return JNI_VERSION_1_2;
}

/* Defining Classes */

ClassClass *
AddToLoadedClasses(ExecEnv *ee, ClassClass *cb)
{
    static struct methodblock *cached_mb = NULL;
    struct Hjava_lang_ClassLoader *loader = cb ? cbLoader(cb) : NULL;

    if (loader == NULL || cb == NULL) {
        return cb;
    }

    if (cached_mb == NULL) {
        HashedNameAndType hashed;
        HashNameAndType(ee, "addClass", "(Ljava/lang/Class;)V", &hashed);
        cached_mb = jni_FindMethodBlock0(classJavaLangClassLoader,
					 &hashed,
					 FALSE);
	sysAssert(cached_mb);
    }

    do_execute_java_method(ee, loader, NULL, NULL, cached_mb, FALSE, cb);

    if (exceptionOccurred(ee)) {
        return NULL;
    }
    return cb;
}
	    
static jclass JNICALL
jni_DefineClass(JNIEnv *env, const char *name, jobject loaderRef,
		const jbyte *buf, jsize bufLen)
{
    RECORD_STACK_BASE(env, &env);
    ClassClass *cb = 0;
    struct Hjava_lang_ClassLoader *loader = 
        (struct Hjava_lang_ClassLoader *)DeRef(env, loaderRef);
    jclass result;

    cb = createInternalClass((unsigned char*)buf,
			     (unsigned char*)buf + bufLen,
			     loader, 
			     (char *)name,
			     NULL);
    if (cb != NULL) {
        LoadSuperclasses(cb);
	if (!CCIs(cb, Loaded)) {
	    cb = NULL;
	}
        result = MkRefLocal(env, AddToLoadedClasses(JNIEnv2EE(env), cb));
	if (result) {
	    CCSet(cb, Cached);
	}
    } else {
        result = 0;
    }
    RECOVER_STACK_BASE(env);
    return result;
}

static jclass JNICALL
jni_FindClass(JNIEnv *env, const char *name)
{
    RECORD_STACK_BASE(env, &env);
    ExecEnv *ee = JNIEnv2EE(env);
    ClassClass *cb;
    jclass result;
    struct Hjava_lang_ClassLoader *loader;

    /* A temporary workaround to delay the loading of JIT to reduce
     * startup time. The JIT is loaded when the JNI FindClass function
     * is called.
     */
#ifdef DELAY_JIT_LOADING
    static bool_t compilerLoaded = FALSE;
#endif

    cb = getCallerClass(ee, 0);

    if (cb) {
        /* Special handling to make sure JNI_OnLoad and JNI_OnUnload are
	 * executed in the correct class context.
	 */
        if (cbLoader(cb) == NULL &&
	    strcmp(cbName(cb), "java/lang/ClassLoader$NativeLibrary") == 0) {
	    cb = (ClassClass *)execute_java_static_method(ee, cb, 
		                   "getFromClass", "()Ljava/lang/Class;");
	    if (exceptionOccurred(ee)) {
	        RECOVER_STACK_BASE(env);
		return NULL;
	    }
	}
	loader = cbLoader(cb);
    } else {
        /* We call ClassLoader.getSystemClassLoader to obtain the system
	 * class loader.
	 */
        static struct Hjava_lang_ClassLoader *sysLoader = NULL;
	if (sysLoader == NULL) {
	    sysLoader = (struct Hjava_lang_ClassLoader *)
	        execute_java_static_method(ee, classJavaLangClassLoader,
					   "getSystemClassLoader",
					   "()Ljava/lang/ClassLoader;");
	    if (exceptionOccurred(ee)) {
	        RECOVER_STACK_BASE(env);
		return NULL;
	    }
	}
	loader = sysLoader;
    }

    cb = FindClassFromClassLoader(ee, (char*)name, TRUE, loader, TRUE);

#ifdef DELAY_JIT_LOADING
    if (!compilerLoaded &&
	strcmp(name, "java/lang/System") != 0 &&
	strcmp(name, "java/lang/String") != 0 &&
	strcmp(name, "java/io/File") != 0 &&
	strcmp(name, "java/io/Win32FileSystem") != 0 &&
	strcmp(name, "java/lang/ClassLoader$NativeLibrary") != 0) {
	JHandle *exc = NULL;
	compilerLoaded = TRUE;
	/* Save and restore the exception */
	if (exceptionOccurred(ee)) {
	    exc = ee->exception.exc;
	    exceptionClear(ee);
	}
	FindClass(ee, JAVAPKG "Compiler", TRUE);
	if (!compilerInitialized) {
	    UseLosslessQuickOpcodes = FALSE;
#ifdef BREAKPTS
            if (debugging) {
                /* So that field watchpoints can work */
                UseLosslessQuickOpcodes = TRUE;
            }
#endif /* BREAKPTS */
	}
	exceptionClear(ee);
	if (exc) {
	    exceptionThrow(ee, exc);
	}
    }
#endif

    if (cb != NULL)
        result = MkRefLocal(env, (HObject *)cbHandle(cb));
    else
        result = NULL;
    RECOVER_STACK_BASE(env);
    return result;
}

static jclass JNICALL
jni_GetSuperclass(JNIEnv *env, jclass sub)
{
    RECORD_STACK_BASE(env, &env);
    ClassClass *cb = DeRefClass(env, sub);
    jclass result;

    /* Primitives and interfaces return NULL -- see JLS */
    if (cbIsPrimitive(cb) || cbIsInterface(cb))
        result = NULL;
    else
        result = MkRefLocal(env, (HObject *)(cbSuperclass(cb)));
    RECOVER_STACK_BASE(env);
    return result;
}

static jboolean JNICALL
jni_IsAssignableFrom(JNIEnv *env, jclass sub, jclass super)
{
    RECORD_STACK_BASE(env, &env);
    ExecEnv *ee = JNIEnv2EE(env);
    ClassClass *cb;
    ClassClass *scb;
    jboolean result;

    cb = DeRefClass(env, sub);
    scb = DeRefClass(env, super);

    result = reflect_is_assignable(cb, scb, ee);
    RECOVER_STACK_BASE(env);
    return result;
}


/* Working with Exceptions */

static jint JNICALL
jni_Throw(JNIEnv *env, jobject obj)
{
    RECORD_STACK_BASE(env, &env);
    ExecEnv *ee = JNIEnv2EE(env);
    Hjava_lang_Throwable *ret;

    ret = (Hjava_lang_Throwable*)DeRef2(env, obj);
    if (0 && verbose) {
        ClassClass *cls = obj_classblock(ret);
	Hjava_lang_String *detail = unhand(ret)->detailMessage;
	char *cstr = makeCString(detail);

	jio_fprintf(stderr, "[Throwing via JNI: %s, detail: %s]\n", 
		    cbName(cls), cstr); 

	printStackTrace(ee, 100, NULL);
    }
    if (!is_instance_of((HObject*)ret, classJavaLangThrowable, ee)) {
        jni_FatalError(env, 
	    "throw an object that is not an instance of java/lang/Throwable");
	/* Does not return. */
    }
    fillInStackTrace(ret, ee);
    exceptionThrow(ee, (HObject*)ret);
    RECOVER_STACK_BASE(env);
    return JNI_OK;
}

static jint JNICALL
jni_ThrowNew(JNIEnv *env, jclass clazz, const char *message)
{
    jmethodID mid = 0;
    jobject obj = 0;
    jstring msg = 0;
    jint result;

    if (jni_EnsureLocalCapacity(env, 2) < 0)
        return JNI_ENOMEM;

    mid = jni_GetMethodID(env, clazz, "<init>", "(Ljava/lang/String;)V");
    if (mid == 0) {
        result = JNI_ERR;
	goto done;
    }

    if (message) {
        msg = jni_NewStringUTF(env, message);
	if (msg == 0) {
            result = JNI_ERR;
	    goto done;
	}
    } else
        msg = 0;

    obj = jni_NewObject(env, clazz, mid, msg);
    if (obj == 0) {
        result = JNI_ERR;
	goto done;
    }

    result = jni_Throw(env, obj);

done:
    jni_DeleteLocalRef(env, msg);
    jni_DeleteLocalRef(env, obj);

    return result;
}

static jobject JNICALL
jni_ExceptionOccurred(JNIEnv *env)
{
    ExecEnv *ee = JNIEnv2EE(env);
    if (!exceptionOccurred(ee)) {
        /* frequent case */
        return NULL;
    } else {
        RECORD_STACK_BASE(env, &env);
	jobject result = MkRefLocal(env, ee->exception.exc);
	RECOVER_STACK_BASE(env);
	return result;
    }
}

/*
 * Describe to stderr
 */
static void JNICALL
jni_ExceptionDescribe(JNIEnv *env)
{
    RECORD_STACK_BASE(env, &env);
    ExecEnv *ee = JNIEnv2EE(env);
    exceptionDescribe(ee);
    RECOVER_STACK_BASE(env);
}

static void JNICALL
jni_ExceptionClear(JNIEnv *env)
{
    RECORD_STACK_BASE(env, &env);
    ExecEnv *ee = JNIEnv2EE(env);
    exceptionClear(ee);
    RECOVER_STACK_BASE(env);
}

static void JNICALL
jni_FatalError(JNIEnv *env, const char *msg)
{
    jio_fprintf(stderr,"FATAL ERROR in native method: %s\n", msg);
#ifdef JNI_PANIC_DEBUG
    *(int*)0 = 0;
#endif

    printStackTrace(JNIEnv2EE(env), 100, NULL);

    Abort();
}

/* Local Frame Management
 */

static jint JNICALL
jni_PushLocalFrame(JNIEnv *env, jint capacity)
{
    ExecEnv *ee = JNIEnv2EE(env);
    JavaFrame  *old_frame = ee->current_frame; /* previous frame */
    JavaStack  *stack = old_frame->javastack; 
    struct methodblock *mb = old_frame->current_method;
    JavaFrame  *frame = (JavaFrame *)
      ((mb && ((mb->fb.access & ACC_NATIVE) == 0)) ? 
            old_frame->ostack + mb->maxstack
          : old_frame->optop);

    if (frame->ostack + capacity + JNI_REF_INFO_SIZE >= stack->end_data) {
        JavaStack *tmp_stack = stack;
	JavaFrame *tmp_frame = frame;
	if (!ExpandJavaStackForJNI(ee, &tmp_stack, &tmp_frame,
				   capacity + JNI_REF_INFO_SIZE)) {
	    return JNI_ENOMEM;
	}
	stack = tmp_stack;
	frame = tmp_frame;
    }

    frame->javastack = stack;
    frame->prev = old_frame;
    frame->optop = frame->ostack + JNI_REF_INFO_SIZE;
    frame->current_method = 0; 

    JNI_REFS_FREELIST(frame) = NULL;
    JNI_N_REFS_IN_USE(frame) = 0;
    ee->current_frame = frame;
    JNI_REFS_CAPACITY(frame) = capacity;
    return JNI_OK;
}

static jobject JNICALL
jni_PopLocalFrame(JNIEnv *env, jobject result)
{
    RECORD_STACK_BASE(env, &env);
    ExecEnv *ee = JNIEnv2EE(env);
    JHandle *h = DeRef(env, result);
    ee->current_frame = ee->current_frame->prev;
    result = MkRefLocal(env, h);
    RECOVER_STACK_BASE(env);
    return result;
}

/*
 * Global References
 */

static jobject JNICALL
jni_NewGlobalRef(JNIEnv *env, jobject ref)
{
    RECORD_STACK_BASE(env, &env);
    jobject result;

    if (ref != 0) {
        GLOBALREF_LOCK(EE2SysThread(JNIEnv2EE(env)));
	result = (jobject)jni_addRef(globalRefFrame, DeRef(env, ref));
	GLOBALREF_UNLOCK(EE2SysThread(JNIEnv2EE(env)));
#ifdef HPROF
	if ((JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_JNI_GLOBALREF_ALLOC)) &&
	    result) {
	    jvmpi_new_globalref(JNIEnv2EE(env), result, DeRef(env, ref));
	}
#endif /* HPROF */	
	if (result == 0) {
	    ExecEnv *ee = JNIEnv2EE(env);
	    exceptionClear(ee);
	    ThrowOutOfMemoryError(ee, 
				  "out of memory allocating a JNI global ref");
	}
    } else
        result = 0;

    RECOVER_STACK_BASE(env);
    return result;
}

static void JNICALL
jni_DeleteGlobalRef(JNIEnv *env, jobject ref)
{
    RECORD_STACK_BASE(env, &env);
#ifdef HPROF
    if ((JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_JNI_GLOBALREF_FREE)) && ref) {
        jvmpi_delete_globalref(JNIEnv2EE(env), ref);
    }
#endif /* HPROF */	
    GLOBALREF_LOCK(EE2SysThread(JNIEnv2EE(env)));
    jni_delRef(globalRefFrame, (JHandle **)ref);
    GLOBALREF_UNLOCK(EE2SysThread(JNIEnv2EE(env)));
    RECOVER_STACK_BASE(env);
}

static void JNICALL
jni_DeleteLocalRef(JNIEnv *env, jobject obj)
{
    RECORD_STACK_BASE(env, &env);
    jni_delRef(JNIEnv2EE(env)->current_frame, (JHandle **)obj);
    RECOVER_STACK_BASE(env);
}

static jboolean JNICALL
jni_IsSameObject(JNIEnv *env, jobject r1, jobject r2)
{
    RECORD_STACK_BASE(env, &env);
    jboolean result = DeRef(env, r1) == DeRef(env, r2);
    RECOVER_STACK_BASE(env);
    return result;
}

static jobject JNICALL 
jni_NewLocalRef(JNIEnv *env, jobject ref)
{
    RECORD_STACK_BASE(env, &env);
    jobject result = MkRefLocal(env, DeRef(env, ref));
    RECOVER_STACK_BASE(env);
    return result;
}

static jint JNICALL 
jni_EnsureLocalCapacity(JNIEnv *env, jint capacity)
{
    ExecEnv *ee = JNIEnv2EE(env);
    JavaFrame  *frame = ee->current_frame;
    JavaStack  *stack = frame->javastack;

    if (JNI_REFS_CAPACITY(frame) - JNI_N_REFS_IN_USE(frame) >= capacity)
        return JNI_OK;
    if (stack->end_data - frame->optop < capacity) {
        /* infrequent case */
	capacity -= stack->end_data - frame->optop;
	if (!ExpandJavaStackForJNI(ee, &stack, &frame, capacity))
	    return JNI_ENOMEM;
    }
    JNI_REFS_CAPACITY(frame) = capacity + JNI_N_REFS_IN_USE(frame);
    return JNI_OK;
}

/*
 * Object Operations
 */

static jobject
jni_Construct(JNIEnv *env, jclass clazz, jmethodID methodID,
              JNI_PushArguments_t pushArguments, void *args)
{
    jobject result;
    struct methodblock *mb = (struct methodblock *)methodID;

    result = jni_AllocObject(env, clazz);

    if (result == NULL)
	return NULL;

    if (mb->fb.name != utf8_literal_obj_init_name)
        jni_FatalError(env, "a non-constructor passed to NewObject");
    if (mb->fb.clazz != (ClassClass *)DeRef(env, clazz))
	jni_FatalError(env, "wrong method ID passed to NewObject");

    jni_Invoke(env, result, methodID, pushArguments, args, 
	       INVOKE_VIRTUAL | TERSE_SIG_VOID);

    {
	ExecEnv *ee = JNIEnv2EE(env);
	if (exceptionOccurred(ee)) {
	    (*env)->DeleteLocalRef(env, result);
	    return NULL;
	}
    }
    return result;
}

static jobject JNICALL
jni_AllocObject(JNIEnv *env, jclass clazz)
{
    RECORD_STACK_BASE(env, &env);
    ExecEnv *ee = JNIEnv2EE(env);
    HObject *obj;
    ClassClass *cb;
    jobject result = 0;

    cb = DeRefClass(env, clazz);

    if (cbAccess(cb) & (ACC_INTERFACE | ACC_ABSTRACT)) {
        ThrowInstantiationException(ee, cbName(cb));
        goto done;
    }

    obj = allocObject(ee, cb);
    if (obj == NULL) { 
        ThrowOutOfMemoryError(ee, NULL);
        goto done;
    }
    result = MkRefLocal(env, obj);
done:
    RECOVER_STACK_BASE(env);
    return result;
}

static jobject JNICALL
jni_NewObjectA(JNIEnv *env, jclass clazz, jmethodID methodID, jvalue *args)
{
    return jni_Construct(env, clazz, methodID, jni_PushArgumentsArray, &args);
}

static jobject JNICALL
jni_NewObjectV(JNIEnv *env, jclass clazz, jmethodID methodID, va_list args)
{
    return jni_Construct(env, clazz, methodID, jni_PushArgumentsVararg, &args);
}

static jobject JNICALL
jni_NewObject(JNIEnv *env, jclass clazz, jmethodID methodID, ...)
{
    jobject result;
    va_list args;
    va_start(args, methodID);
    result = jni_Construct(env, clazz, methodID, jni_PushArgumentsVararg, 
			   &args);
    va_end(args);
    return result;
}

static jclass JNICALL
jni_GetObjectClass(JNIEnv *env, jobject obj)
{
    RECORD_STACK_BASE(env, &env);
    HObject *o = (HObject *)DeRef2(env, obj);
    ClassClass *cls = getClass(o);
    jclass result = MkRefLocal(env, (HObject*)cls);
    RECOVER_STACK_BASE(env);
    return result;
}

static jboolean JNICALL
jni_IsInstanceOf(JNIEnv *env, jobject obj, jclass clazz)
{
    RECORD_STACK_BASE(env, &env);
    ExecEnv *ee = JNIEnv2EE(env);
    ClassClass *cb = DeRefClass(env, clazz);
    jboolean result = is_instance_of(DeRef(env, obj), cb, ee);
    RECOVER_STACK_BASE(env);
    return result;
}

/*
 * Identifiers
 */

static jmethodID JNICALL
jni_GetMethodID(JNIEnv *env, jclass clazz, 
		const char *name, const char *sig)
{
    struct methodblock *mb;

    mb = jni_FindMethodBlock(env, clazz, name, sig, TRUE);

    if (mb == NULL)
        return NULL;

    if (mb->fb.access & ACC_STATIC) {
        ThrowNoSuchMethodError(0, (char *)name);
	return NULL;
    }

    return (jmethodID)mb;
}

static jmethodID JNICALL
jni_GetStaticMethodID(JNIEnv *env, jclass clazz, 
		      const char *name, const char *sig)
{
    struct methodblock *mb;
    
    mb = jni_FindMethodBlock(env, clazz, name, sig, TRUE);

    if (mb == NULL)
        return NULL;

    if (!(mb->fb.access & ACC_STATIC)) {
        ThrowNoSuchMethodError(0, (char *)name);
	return NULL;
    }

    return (jmethodID)mb;
}


static jfieldID JNICALL
jni_GetFieldID(JNIEnv *env, jclass clazz, 
	       const char *name, const char *sig)
{
    return (jfieldID)jni_FindFieldBlock(env, clazz, name, sig, 0);
}

#define obj_Getoffsetaddr(o, off) ((OBJECT*)((char *)unhand(o)+(off)))
#define obj_Getoffset(o, off)      (*obj_Getoffsetaddr(o, off))
#define obj_Setoffset(o, off, v)  (*obj_Getoffsetaddr(o, off) = (v))

static jobject JNICALL
jni_GetObjectField(JNIEnv *env, jobject obj, jfieldID fieldID)
{
    RECORD_STACK_BASE(env, &env);
    JHandle *o;
    JHandle *value;
    struct fieldblock *fb = (struct fieldblock *)fieldID;
    jobject result;

    o = DeRef2(env, obj);
    value = (JHandle*)obj_Getoffset(o, fb->u.offset);
    result = MkRefLocal(env, value);
    RECOVER_STACK_BASE(env);
    return result;
}

static jboolean JNICALL
jni_GetBooleanField(JNIEnv *env, jobject obj, jfieldID fieldID)
{
    RECORD_STACK_BASE(env, &env);
    JHandle *o;
    struct fieldblock *fb = (struct fieldblock *)fieldID;
    jboolean result;

    o = DeRef2(env, obj);
    result = (jboolean)obj_Getoffset(o, fb->u.offset);
    RECOVER_STACK_BASE(env);
    return result;
}

static jbyte JNICALL
jni_GetByteField(JNIEnv *env, jobject obj, jfieldID fieldID)
{
    RECORD_STACK_BASE(env, &env);
    JHandle *o;
    struct fieldblock *fb = (struct fieldblock *)fieldID;
    jbyte result;

    o = DeRef2(env, obj);
    result = (jbyte)obj_Getoffset(o, fb->u.offset);
    RECOVER_STACK_BASE(env);
    return result;
}

static jchar JNICALL
jni_GetCharField(JNIEnv *env, jobject obj, jfieldID fieldID)
{
    RECORD_STACK_BASE(env, &env);
    JHandle *o;
    struct fieldblock *fb = (struct fieldblock *)fieldID;
    jchar result;

    o = DeRef2(env, obj);
    result = (jchar)obj_Getoffset(o, fb->u.offset);
    RECOVER_STACK_BASE(env);
    return result;
}

static jshort JNICALL
jni_GetShortField(JNIEnv *env, jobject obj, jfieldID fieldID)
{
    RECORD_STACK_BASE(env, &env);
    JHandle *o;
    struct fieldblock *fb = (struct fieldblock *)fieldID;
    jshort result;

    o = DeRef2(env, obj);
    result = (jshort)obj_Getoffset(o, fb->u.offset);
    RECOVER_STACK_BASE(env);
    return result;
}

static jint JNICALL
jni_GetIntField(JNIEnv *env, jobject obj, jfieldID fieldID)
{
    RECORD_STACK_BASE(env, &env);
    JHandle *o;
    struct fieldblock *fb = (struct fieldblock *)fieldID;
    jint result;

    o = DeRef2(env, obj);
    result = (jint)obj_Getoffset(o, fb->u.offset);
    RECOVER_STACK_BASE(env);
    return result;
}

static jlong JNICALL
jni_GetLongField(JNIEnv *env, jobject obj, jfieldID fieldID)
{
    RECORD_STACK_BASE(env, &env);
    JHandle *o;
    struct fieldblock *fb = (struct fieldblock *)fieldID;
    OBJECT *slot;
    jvalue tmp;
    void *unhandO;
    jlong result;

    o = DeRef2(env, obj);
    unhandO = unhand(o);
    slot = obj_Getoffsetaddr(o, fb->u.offset);
    result = GET_INT64(tmp, slot);
    KEEP_POINTER_ALIVE(unhandO);
    RECOVER_STACK_BASE(env);
    return result;
}

static jfloat JNICALL
jni_GetFloatField(JNIEnv *env, jobject obj, jfieldID fieldID)
{
    RECORD_STACK_BASE(env, &env);
    JHandle *o;
    struct fieldblock *fb = (struct fieldblock *)fieldID;
    OBJECT *slot;
    void *unhandO;
    jfloat result;

    o = DeRef2(env, obj);
    unhandO = unhand(o);
    slot = obj_Getoffsetaddr(o, fb->u.offset);
    result = *(jfloat*)slot;
    KEEP_POINTER_ALIVE(unhandO);
    RECOVER_STACK_BASE(env);
    return result;
}

static jdouble JNICALL
jni_GetDoubleField(JNIEnv *env, jobject obj, jfieldID fieldID)
{
    RECORD_STACK_BASE(env, &env);
    struct fieldblock *fb = (struct fieldblock *)fieldID;
    JHandle *o;
    OBJECT *slot;
    jvalue tmp;
    void *unhandO;
    jdouble result;

    o = DeRef2(env, obj);
    unhandO = unhand(o);
    slot = obj_Getoffsetaddr(o, fb->u.offset);
    result = GET_DOUBLE(tmp, slot);
    KEEP_POINTER_ALIVE(unhandO);
    RECOVER_STACK_BASE(env);
    return result;
}

static void JNICALL
jni_SetObjectField(JNIEnv *env, jobject obj, jfieldID fieldID, jobject value)
{
    RECORD_STACK_BASE(env, &env);
    JHandle *o = DeRef2(env, obj);
    JHandle *v = DeRef(env, value);
    struct fieldblock *fb = (struct fieldblock *)fieldID;
    obj_Setoffset(o, fb->u.offset, (OBJECT)v);
    RECOVER_STACK_BASE(env);
}

static void JNICALL
jni_SetBooleanField(JNIEnv *env, jobject obj, jfieldID fieldID, jboolean value)
{
    RECORD_STACK_BASE(env, &env);
    JHandle *o;
    struct fieldblock *fb = (struct fieldblock *)fieldID;

    o = DeRef2(env, obj);
    obj_Setoffset(o, fb->u.offset, value);
    RECOVER_STACK_BASE(env);
}

static void JNICALL
jni_SetByteField(JNIEnv *env, jobject obj, jfieldID fieldID, jbyte value)
{
    RECORD_STACK_BASE(env, &env);
    JHandle *o;
    struct fieldblock *fb = (struct fieldblock *)fieldID;

    o = DeRef2(env, obj);
    obj_Setoffset(o, fb->u.offset, value);
    RECOVER_STACK_BASE(env);
}

static void JNICALL
jni_SetCharField(JNIEnv *env, jobject obj, jfieldID fieldID, jchar value)
{
    RECORD_STACK_BASE(env, &env);
    JHandle *o;
    struct fieldblock *fb = (struct fieldblock *)fieldID;

    o = DeRef2(env, obj);
    obj_Setoffset(o, fb->u.offset, value);
    RECOVER_STACK_BASE(env);
}

static void JNICALL
jni_SetShortField(JNIEnv *env, jobject obj, jfieldID fieldID, jshort value)
{
    RECORD_STACK_BASE(env, &env);
    JHandle *o;
    struct fieldblock *fb = (struct fieldblock *)fieldID;

    o = DeRef2(env, obj);
    obj_Setoffset(o, fb->u.offset, value);
    RECOVER_STACK_BASE(env);
}

static void JNICALL
jni_SetIntField(JNIEnv *env, jobject obj, jfieldID fieldID, jint value)
{
    RECORD_STACK_BASE(env, &env);
    JHandle *o;
    struct fieldblock *fb = (struct fieldblock *)fieldID;

    o = DeRef2(env, obj);
    obj_Setoffset(o, fb->u.offset, value);
    RECOVER_STACK_BASE(env);
}

static void JNICALL
jni_SetLongField(JNIEnv *env, jobject obj, jfieldID fieldID, jlong value)
{
    RECORD_STACK_BASE(env, &env);
    JHandle *o;
    struct fieldblock *fb = (struct fieldblock *)fieldID;
    OBJECT *slot;
    jvalue tmp;
    void *unhandO;

    o = DeRef2(env, obj);
    unhandO = unhand(o);
    slot = obj_Getoffsetaddr(o, fb->u.offset);
    SET_INT64(tmp, slot, value);
    KEEP_POINTER_ALIVE(unhandO);
    RECOVER_STACK_BASE(env);
}

static void JNICALL
jni_SetFloatField(JNIEnv *env, jobject obj, jfieldID fieldID,
		  jfloat value)
{
    RECORD_STACK_BASE(env, &env);
    JHandle *o;
    struct fieldblock *fb = (struct fieldblock *)fieldID;

    o = DeRef2(env, obj);
    obj_Setoffset(o, fb->u.offset, 
		  (*(int*)(&value)));
    RECOVER_STACK_BASE(env);
}

static void JNICALL
jni_SetDoubleField(JNIEnv *env, jobject obj, jfieldID fieldID,
		   jdouble value)
{
    RECORD_STACK_BASE(env, &env);
    JHandle *o;
    struct fieldblock *fb = (struct fieldblock *)fieldID;
    OBJECT *slot;
    jvalue tmp;
    void *unhandO;

    o = DeRef2(env, obj);
    unhandO = unhand(o);
    slot = obj_Getoffsetaddr(o, fb->u.offset);
    SET_DOUBLE(tmp, slot, value);
    KEEP_POINTER_ALIVE(unhandO);
    RECOVER_STACK_BASE(env);
}

/* Calling Methods */

#define DEFINE_CALLMETHOD(ResultType, Result, unionType, retType) \
static ResultType JNICALL \
jni_Call##Result##Method(JNIEnv *env, jobject obj, jmethodID methodID, ...) \
{ \
    ResultType result; \
    va_list args; \
 \
    va_start(args, methodID); \
    result = jni_Invoke(env, obj, methodID, \
                        jni_PushArgumentsVararg, &args, \
                        INVOKE_VIRTUAL | retType).unionType; \
    va_end(args); \
    return result; \
} \
static ResultType JNICALL \
jni_Call##Result##MethodV(JNIEnv *env, jobject obj, \
                          jmethodID methodID, va_list args) \
{ \
    ResultType result; \
\
    result = jni_Invoke(env, obj, methodID, \
                        jni_PushArgumentsVararg, &args, \
                        INVOKE_VIRTUAL | retType).unionType; \
    return result; \
} \
static ResultType JNICALL \
jni_Call##Result##MethodA(JNIEnv *env, jobject obj, \
			  jmethodID methodID, jvalue *args) \
{ \
    ResultType result; \
 \
    result = jni_Invoke(env, obj, methodID, \
                        jni_PushArgumentsArray, &args, \
                        INVOKE_VIRTUAL | retType).unionType; \
    return result; \
}

DEFINE_CALLMETHOD(jobject,Object,l,TERSE_SIG_OBJECT)
DEFINE_CALLMETHOD(jboolean,Boolean,z,TERSE_SIG_BOOLEAN)
DEFINE_CALLMETHOD(jbyte,Byte,b,TERSE_SIG_BYTE)
DEFINE_CALLMETHOD(jchar,Char,c,TERSE_SIG_CHAR)
DEFINE_CALLMETHOD(jshort,Short,s,TERSE_SIG_SHORT)
DEFINE_CALLMETHOD(jint,Int,i,TERSE_SIG_INT)
DEFINE_CALLMETHOD(jlong,Long,j,TERSE_SIG_LONG)
DEFINE_CALLMETHOD(jfloat,Float,f,TERSE_SIG_FLOAT)
DEFINE_CALLMETHOD(jdouble,Double,d,TERSE_SIG_DOUBLE)

static void JNICALL
jni_CallVoidMethod(JNIEnv *env, jobject obj, jmethodID methodID, ...) 
{                                
    va_list args;                

    va_start(args, methodID);
    jni_Invoke(env, obj, methodID,
	       jni_PushArgumentsVararg, &args,
	       INVOKE_VIRTUAL | TERSE_SIG_VOID);
    va_end(args);
}                                                  
                                                   
static void JNICALL
jni_CallVoidMethodV(JNIEnv *env, jobject obj,
		    jmethodID methodID, va_list args)
{
    jni_Invoke(env, obj, methodID,
	       jni_PushArgumentsVararg, &args,
	       INVOKE_VIRTUAL | TERSE_SIG_VOID);
}                                    
                                     
static void JNICALL
jni_CallVoidMethodA(JNIEnv *env, jobject obj,
		    jmethodID methodID, jvalue *args)
{
    jni_Invoke(env, obj, methodID,                
	       jni_PushArgumentsArray, &args,
	       INVOKE_VIRTUAL | TERSE_SIG_VOID);
}


#define DEFINE_CALLNONVIRTUALMETHOD(ResultType, Result, unionType, retType) \
static ResultType JNICALL \
jni_CallNonvirtual##Result##Method(JNIEnv *env, jobject obj, \
				   jclass clazz, jmethodID methodID, ...) \
{ \
    ResultType result; \
    va_list args; \
 \
    va_start(args, methodID); \
    result = jni_Invoke(env, obj, methodID, \
                        jni_PushArgumentsVararg, &args, \
                        INVOKE_NONVIRTUAL | retType).unionType; \
    va_end(args); \
    return result; \
} \
static ResultType JNICALL \
jni_CallNonvirtual##Result##MethodV(JNIEnv *env, jobject obj, \
				    jclass clazz, jmethodID methodID, \
				    va_list args) \
{ \
    ResultType result; \
 \
    result = jni_Invoke(env, obj, methodID, \
                        jni_PushArgumentsVararg, &args, \
                        INVOKE_NONVIRTUAL | retType).unionType; \
    return result; \
} \
static ResultType JNICALL \
jni_CallNonvirtual##Result##MethodA(JNIEnv *env, jobject obj, \
				    jclass clazz, jmethodID methodID, \
				    jvalue *args) \
{ \
    ResultType result; \
 \
    result = jni_Invoke(env, obj, methodID, \
                        jni_PushArgumentsArray, &args, \
                        INVOKE_NONVIRTUAL | retType).unionType; \
    return result; \
}

DEFINE_CALLNONVIRTUALMETHOD(jobject,Object,l,TERSE_SIG_OBJECT)
DEFINE_CALLNONVIRTUALMETHOD(jboolean,Boolean,z,TERSE_SIG_BOOLEAN)
DEFINE_CALLNONVIRTUALMETHOD(jbyte,Byte,b,TERSE_SIG_BYTE)
DEFINE_CALLNONVIRTUALMETHOD(jchar,Char,c,TERSE_SIG_CHAR)
DEFINE_CALLNONVIRTUALMETHOD(jshort,Short,s,TERSE_SIG_SHORT)
DEFINE_CALLNONVIRTUALMETHOD(jint,Int,i,TERSE_SIG_INT)
DEFINE_CALLNONVIRTUALMETHOD(jlong,Long,j,TERSE_SIG_LONG)
DEFINE_CALLNONVIRTUALMETHOD(jfloat,Float,f,TERSE_SIG_FLOAT)
DEFINE_CALLNONVIRTUALMETHOD(jdouble,Double,d,TERSE_SIG_DOUBLE)

static void JNICALL
jni_CallNonvirtualVoidMethod(JNIEnv *env, jobject obj, 
			     jclass clazz, jmethodID methodID, ...) 
{                                
    va_list args;

    va_start(args, methodID);
    jni_Invoke(env, obj, methodID,
	       jni_PushArgumentsVararg, &args,
	       INVOKE_NONVIRTUAL | TERSE_SIG_VOID);
    va_end(args);
}                                                  
                                                   
static void JNICALL
jni_CallNonvirtualVoidMethodV(JNIEnv *env, jobject obj,
			      jclass clazz, jmethodID methodID, va_list args)
{
    jni_Invoke(env, obj, methodID,
	       jni_PushArgumentsVararg, &args,
	       INVOKE_NONVIRTUAL | TERSE_SIG_VOID);
}                                    
                                     
static void JNICALL
jni_CallNonvirtualVoidMethodA(JNIEnv *env, jobject obj,
			      jclass clazz, jmethodID methodID, jvalue *args)
{
    jni_Invoke(env, obj, methodID,                
	       jni_PushArgumentsArray, &args,
	       INVOKE_NONVIRTUAL | TERSE_SIG_VOID);
}



#define DEFINE_CALLSTATICMETHOD(ResultType, Result, unionType, retType) \
static ResultType JNICALL \
jni_CallStatic##Result##Method(JNIEnv *env, jclass clazz, \
			       jmethodID methodID, ...)  \
{ \
    ResultType result; \
    va_list args; \
  \
    va_start(args, methodID); \
    result = jni_Invoke(env, clazz, methodID, \
                        jni_PushArgumentsVararg, &args, \
                        INVOKE_STATIC | retType).unionType; \
    va_end(args); \
    return result; \
} \
static ResultType JNICALL \
jni_CallStatic##Result##MethodV(JNIEnv *env, jclass clazz, \
				jmethodID methodID, va_list args) \
{ \
    ResultType result; \
  \
    result = jni_Invoke(env, clazz, methodID, \
                        jni_PushArgumentsVararg, &args, \
                        INVOKE_STATIC | retType).unionType; \
    return result; \
} \
static ResultType JNICALL \
jni_CallStatic##Result##MethodA(JNIEnv *env, jclass clazz, \
				jmethodID methodID, jvalue *args) \
{ \
    ResultType result; \
  \
    result = jni_Invoke(env, clazz, methodID, \
                        jni_PushArgumentsArray, &args, \
                        INVOKE_STATIC | retType).unionType; \
    return result; \
}

DEFINE_CALLSTATICMETHOD(jobject,Object,l,TERSE_SIG_OBJECT)
DEFINE_CALLSTATICMETHOD(jboolean,Boolean,z,TERSE_SIG_BOOLEAN)
DEFINE_CALLSTATICMETHOD(jbyte,Byte,b,TERSE_SIG_BYTE)
DEFINE_CALLSTATICMETHOD(jchar,Char,c,TERSE_SIG_CHAR)
DEFINE_CALLSTATICMETHOD(jshort,Short,s,TERSE_SIG_SHORT)
DEFINE_CALLSTATICMETHOD(jint,Int,i,TERSE_SIG_INT)
DEFINE_CALLSTATICMETHOD(jlong,Long,j,TERSE_SIG_LONG)
DEFINE_CALLSTATICMETHOD(jfloat,Float,f,TERSE_SIG_FLOAT)
DEFINE_CALLSTATICMETHOD(jdouble,Double,d,TERSE_SIG_DOUBLE)

static void JNICALL
jni_CallStaticVoidMethod(JNIEnv *env, jclass clazz, jmethodID methodID, ...) 
{                                
    va_list args;

    va_start(args, methodID);
    jni_Invoke(env, clazz, methodID,
	       jni_PushArgumentsVararg, &args,
	       INVOKE_STATIC | TERSE_SIG_VOID);
    va_end(args);
}                                                  
                                                   
static void JNICALL
jni_CallStaticVoidMethodV(JNIEnv *env, jclass clazz,
			  jmethodID methodID, va_list args)
{
    jni_Invoke(env, clazz, methodID,
	       jni_PushArgumentsVararg, &args,
	       INVOKE_STATIC | TERSE_SIG_VOID);
}                                    
                                     
static void JNICALL
jni_CallStaticVoidMethodA(JNIEnv *env, jclass clazz,
			  jmethodID methodID, jvalue *args)
{
    jni_Invoke(env, clazz, methodID,                
	       jni_PushArgumentsArray, &args,
	       INVOKE_STATIC | TERSE_SIG_VOID);
}

/* Accessing Static Fields */

static jfieldID JNICALL
jni_GetStaticFieldID(JNIEnv *env, jclass clazz, 
		     const char *name, const char *sig)
{
    return (jfieldID)jni_FindFieldBlock(env, clazz, name, sig, 1);
}

static jobject JNICALL
jni_GetStaticObjectField(JNIEnv *env, jclass clazz, jfieldID fieldID)
{
    RECORD_STACK_BASE(env, &env);
    struct fieldblock *fb = (struct fieldblock *)fieldID;
    OBJECT *slot;
    JHandle *value;
    jobject result;

    slot = (OBJECT*)(normal_static_address(fb));
    value = (JHandle*)*slot;
    result = MkRefLocal(env, value);
    RECOVER_STACK_BASE(env);
    return result;
}

static jboolean JNICALL
jni_GetStaticBooleanField(JNIEnv *env, jclass clazz, jfieldID fieldID)
{
    struct fieldblock *fb = (struct fieldblock *)fieldID;
    OBJECT *slot;

    slot = (OBJECT*)(normal_static_address(fb));
    return (jboolean)*slot;
}

static jbyte JNICALL
jni_GetStaticByteField(JNIEnv *env, jclass clazz, jfieldID fieldID)
{
    struct fieldblock *fb = (struct fieldblock *)fieldID;
    OBJECT *slot;

    slot = (OBJECT*)(normal_static_address(fb));
    return (jbyte)*slot;
}

static jchar JNICALL
jni_GetStaticCharField(JNIEnv *env, jclass clazz, jfieldID fieldID)
{
    struct fieldblock *fb = (struct fieldblock *)fieldID;
    OBJECT *slot;

    slot = (OBJECT*)(normal_static_address(fb));
    return (jchar)*slot;
}

static jshort JNICALL
jni_GetStaticShortField(JNIEnv *env, jclass clazz, jfieldID fieldID)
{
    struct fieldblock *fb = (struct fieldblock *)fieldID;
    OBJECT *slot;

    slot = (OBJECT*)(normal_static_address(fb));
    return (jshort)*slot;
}

static jint JNICALL
jni_GetStaticIntField(JNIEnv *env, jclass clazz, jfieldID fieldID)
{
    struct fieldblock *fb = (struct fieldblock *)fieldID;
    OBJECT *slot;

    slot = (OBJECT*)(normal_static_address(fb));
    return *slot;
}

static jlong JNICALL
jni_GetStaticLongField(JNIEnv *env, jclass clazz, jfieldID fieldID)
{
    struct fieldblock *fb = (struct fieldblock *)fieldID;
    OBJECT *slot;
    jvalue tmp;

    slot = (OBJECT*)(twoword_static_address(fb));
    return GET_INT64(tmp, slot);
}

static jfloat JNICALL
jni_GetStaticFloatField(JNIEnv *env, jclass clazz, jfieldID fieldID)
{
    struct fieldblock *fb = (struct fieldblock *)fieldID;
    OBJECT *slot;

    slot = (OBJECT*)(normal_static_address(fb));
    return *(float*)slot;
}

static jdouble JNICALL
jni_GetStaticDoubleField(JNIEnv *env, jclass clazz, jfieldID fieldID)
{
    struct fieldblock *fb = (struct fieldblock *)fieldID;
    OBJECT *slot;
    jvalue tmp;

    slot = (OBJECT*)(twoword_static_address(fb));
    return GET_DOUBLE(tmp, slot);
}

static void JNICALL
jni_SetStaticObjectField(JNIEnv *env, jclass clazz, jfieldID fieldID,
			 jobject value)
{
    RECORD_STACK_BASE(env, &env);
    struct fieldblock *fb = (struct fieldblock *)fieldID;
    OBJECT *slot;
    JHandle *v;

    v = DeRef(env, value);
    slot = (OBJECT*)(normal_static_address(fb));
    *slot = (OBJECT)v;
    RECOVER_STACK_BASE(env);
}

static void JNICALL
jni_SetStaticBooleanField(JNIEnv *env, jclass clazz, jfieldID fieldID,
			  jboolean value)
{
    struct fieldblock *fb = (struct fieldblock *)fieldID;
    OBJECT *slot;

    slot = (OBJECT*)(normal_static_address(fb));
    *slot = (OBJECT)value;
}

static void JNICALL
jni_SetStaticByteField(JNIEnv *env, jclass clazz, jfieldID fieldID,
		       jbyte value)
{
    struct fieldblock *fb = (struct fieldblock *)fieldID;
    OBJECT *slot;

    slot = (OBJECT*)(normal_static_address(fb));
    *slot = (OBJECT)value;
}

static void JNICALL
jni_SetStaticCharField(JNIEnv *env, jclass clazz, jfieldID fieldID,
		       jchar value)
{
    struct fieldblock *fb = (struct fieldblock *)fieldID;
    OBJECT *slot;

    slot = (OBJECT*)(normal_static_address(fb));
    *slot = (OBJECT)value;
}

static void JNICALL
jni_SetStaticShortField(JNIEnv *env, jclass clazz, jfieldID fieldID,
			jshort value)
{
    struct fieldblock *fb = (struct fieldblock *)fieldID;
    OBJECT *slot;

    slot = (OBJECT*)(normal_static_address(fb));
    *slot = (OBJECT)value;
}

static void JNICALL
jni_SetStaticIntField(JNIEnv *env, jclass clazz, jfieldID fieldID,
		      jint value)
{
    struct fieldblock *fb = (struct fieldblock *)fieldID;
    OBJECT *slot;

    slot = (OBJECT*)(normal_static_address(fb));
    *slot = (OBJECT)value;
}

static void JNICALL
jni_SetStaticLongField(JNIEnv *env, jclass clazz, jfieldID fieldID,
		       jlong value)
{
    struct fieldblock *fb = (struct fieldblock *)fieldID;
    OBJECT *slot;
    jvalue tmp;

    slot = (OBJECT*)(twoword_static_address(fb));
    SET_INT64(tmp, slot, value);
}

static void JNICALL
jni_SetStaticFloatField(JNIEnv *env, jclass clazz, jfieldID fieldID,
			jfloat value)
{
    struct fieldblock *fb = (struct fieldblock *)fieldID;
    float *slot;

    slot = (float*)(normal_static_address(fb));
    *slot = value;
}

static void JNICALL
jni_SetStaticDoubleField(JNIEnv *env, jclass clazz, jfieldID fieldID,
			 jdouble value)
{
    struct fieldblock *fb = (struct fieldblock *)fieldID;
    OBJECT *slot;
    jvalue tmp;

    slot = (OBJECT*)(twoword_static_address(fb));
    SET_DOUBLE(tmp, slot, value);
}

/*
 * String Operations
 */

#define GET_STRING(s, js) \
    s = (struct Hjava_lang_String*)DeRef(env, js)

#define GET_STRING_BODY(s, str_head, strBody) \
if (1) { \
    Classjava_lang_String *_str; \
    HArrayOfChar *_hac; \
 \
    _str = unhand(s); \
    _hac = (HArrayOfChar*)_str->value; \
    str_head = unhand(_hac)->body; \
    strBody = str_head + _str->offset; \
} else ((void) 0)

/* Unicode Interface */

static jstring JNICALL
jni_NewString(JNIEnv *env, const jchar *unicodeChars, jsize len)
{
    RECORD_STACK_BASE(env, &env);
    ExecEnv *ee = JNIEnv2EE(env);
    HArrayOfChar *val;
    unicode *p;
    struct Hjava_lang_String *str;
    jstring result;

    val = (HArrayOfChar *) allocArray(ee, T_CHAR, len);
    if (val == NULL) {
        ThrowOutOfMemoryError(ee, 0);
	result = 0;
    } else {
        p = unhand(val)->body;
	if (unicodeChars) memcpy(p, unicodeChars, (size_t) (len << 1L));
	KEEP_POINTER_ALIVE(p);
	str = (struct Hjava_lang_String *)allocObject(ee, classJavaLangString);
	if (str) {
	    unhand(str)->value = val;
	    unhand(str)->count = len;
	    result = MkRefLocal(env, (JHandle*)str);
	} else {
	    ThrowOutOfMemoryError(ee, 0);
	    result = 0;
	}
    }
    RECOVER_STACK_BASE(env);
    return result;
}

static jsize JNICALL
jni_GetStringLength(JNIEnv *env, jstring string)
{
    RECORD_STACK_BASE(env, &env);
    jsize result;
    struct Hjava_lang_String *s;

    GET_STRING(s, string);
    result = unhand(s)->count;
    RECOVER_STACK_BASE(env);
    return result;
}
   
static const jchar* JNICALL
jni_GetStringChars(JNIEnv *env, jstring string, jboolean *isCopy)
{
    RECORD_STACK_BASE(env, &env);
    unicode *str_head;
    const jchar *res;
    struct Hjava_lang_String *s;

    GET_STRING(s, string);
    GET_STRING_BODY(s, str_head, res);

    if (isCopy) 
        *isCopy = JNI_FALSE;
    if (!pinObj(JNIEnv2EE(env), str_head))
        res = 0;
    KEEP_POINTER_ALIVE(str_head);
    RECOVER_STACK_BASE(env);
    return res;
}

static void JNICALL
jni_ReleaseStringChars(JNIEnv *env, jstring str, const jchar *chars)
{
    RECORD_STACK_BASE(env, &env);
    struct Hjava_lang_String *s;
    unicode *str_head;
    const jchar *str_chars;

    GET_STRING(s, str);
    GET_STRING_BODY(s, str_head, str_chars);
    unpinObj(JNIEnv2EE(env), str_head);

    RECOVER_STACK_BASE(env);
}

/* UTF Interface */

static jstring JNICALL
jni_NewStringUTF(JNIEnv *env, const char *bytes)
{
    RECORD_STACK_BASE(env, &env);
    jstring result;
    struct Hjava_lang_String *jstr;
    HArrayOfChar *val;
    unicode *unicode_str;
    int unicode_length = utfstrlen((char *)bytes);

    val = (HArrayOfChar *) allocArray(JNIEnv2EE(env), T_CHAR, unicode_length);
    if (val == 0) {
        ThrowOutOfMemoryError(0, 0);
        result = 0;
    } else {
        unicode_str = unhand(val)->body;
	utf2unicode((char *)bytes, unicode_str, unicode_length, &unicode_length);
	KEEP_POINTER_ALIVE(unicode_str);
	jstr = (struct Hjava_lang_String *)allocObject(JNIEnv2EE(env), 
						       classJavaLangString);
	if (jstr) {
	    unhand(jstr)->value = val;
	    unhand(jstr)->count = unicode_length;
	    result = MkRefLocal(env, (JHandle*)jstr);
	} else {
	    ThrowOutOfMemoryError(0, 0);
	    result = 0;
	}
    }
    RECOVER_STACK_BASE(env);
    return result;
}

static jsize JNICALL
jni_GetStringUTFLength(JNIEnv *env, jstring string)
{
    RECORD_STACK_BASE(env, &env);
    struct Hjava_lang_String *s;
    jsize result;
    unicode *head, *body;

    GET_STRING(s, string);
    GET_STRING_BODY(s, head, body);

    /* The result of unicode2utfstrlen includes the room for tailing
       NULL! */
    result = unicode2utfstrlen(body, unhand(s)->count) - 1;
    KEEP_POINTER_ALIVE(head);
    RECOVER_STACK_BASE(env);
    return result;
}

static const char* JNICALL
jni_GetStringUTFChars(JNIEnv *env, jstring string, jboolean *isCopy)
{
    RECORD_STACK_BASE(env, &env);
    struct Hjava_lang_String *s;
    char *utf;
    unicode *head, *body;

    GET_STRING(s, string);
    GET_STRING_BODY(s, head, body);
    utf = unicode2utf(body, unhand(s)->count, 0, 0);
    if (utf == 0)
        ThrowOutOfMemoryError(0, 0);
    if (isCopy) 
        *isCopy = JNI_TRUE;
    KEEP_POINTER_ALIVE(head);
    RECOVER_STACK_BASE(env);
    return (const char*)utf;
}

static void JNICALL
jni_ReleaseStringUTFChars(JNIEnv *env, jstring str, const char *chars)
{
    sysFree((void*)chars);
}

/*
 * Scalar Array Operations
 */

#define GET_ARRAY(arr, jarray) \
    arr = (HArrayOfByte*)DeRef(env, array)

#define GET_ARRAY_BODY(arr, arrBody) \
    arrBody = (unhand(arr)->body)

static jsize JNICALL
jni_GetArrayLength(JNIEnv *env, jarray array)
{
    RECORD_STACK_BASE(env, &env);
    HArrayOfByte *arr;
    jsize result;
    GET_ARRAY(arr, array);
    result = obj_length(arr);
    RECOVER_STACK_BASE(env);
    return result;
}

/*
 * Object Array Operations
 */

static jarray JNICALL
jni_NewObjectArray(JNIEnv *env, jsize length, jclass elementClass,
                   jobject initialElement)
{
    RECORD_STACK_BASE(env, &env);
    ExecEnv *ee = JNIEnv2EE(env);
    HArrayOfObject *arr;
    ClassClass *cb;
    JHandle **body;
    jarray result = 0;

    cb = DeRefClass(env, elementClass);

    arr = (HArrayOfObject*)allocArray(ee, T_CLASS, length);
    if (arr == NULL) {
        ThrowOutOfMemoryError(ee, 0);
        goto done;
    }

    /* Give the array a "type" */
    body = unhand((HArrayOfObject*)arr)->body;
    body[length] = (JHandle*)cb;

    /* Initialize the array */
    if (initialElement != NULL) {
        JHandle *elt = DeRef(env, initialElement);
        jsize i;
        if (!is_instance_of(elt, cb, ee)) {
            ThrowArrayStoreException(ee, NULL);
            goto done;
        }
        for (i = 0; i < length; i++) {
            body[i] = elt;
        }
    }
    result = MkRefLocal(env, (JHandle*)arr);
done:
    RECOVER_STACK_BASE(env);
    return result;
}

static jobject JNICALL
jni_GetObjectArrayElement(JNIEnv *env, jarray array, jsize index)
{
    RECORD_STACK_BASE(env, &env);
    jsize len;
    ExecEnv *ee = JNIEnv2EE(env);
    jobject result;
    HArrayOfByte *arr;
    int8_t *body;
    
    GET_ARRAY(arr, array);
    GET_ARRAY_BODY(arr, body);
    len = obj_length(arr);

    if (index < 0 || index >= len) {
        ThrowArrayIndexOutOfBoundsException(ee, NULL);
	result = 0;
    } else
        result = MkRefLocal(env, ((HObject **)body)[index]);
    RECOVER_STACK_BASE(env);
    return result;
}

static void JNICALL
jni_SetObjectArrayElement(JNIEnv *env, jarray array, jsize index,
                          jobject value)
{
    RECORD_STACK_BASE(env, &env);
    ExecEnv *ee = JNIEnv2EE(env);
    jsize len;
    ClassClass* elementClass;
    HObject *val;
    HArrayOfByte *arr;
    int8_t *body;

    GET_ARRAY(arr, array);
    GET_ARRAY_BODY(arr, body);
    len = obj_length(arr);

    if (index < 0 || index >= len) {
        ThrowArrayIndexOutOfBoundsException(ee, NULL);
    } else {
        elementClass = ((ClassClass**)body)[len];
	val = DeRef(env, value);
	if (!is_instance_of(val, elementClass, ee)) {
	    ThrowArrayStoreException(ee, NULL);
	} else {
	    ((HObject **)body)[index] = val;
	}
    }
    RECOVER_STACK_BASE(env);
}

#define DEFINE_NEWPRIMITIVEARRAY(ElementTag,Result) \
static jarray JNICALL \
jni_New##Result##Array(JNIEnv *env, jsize len) \
{ \
    RECORD_STACK_BASE(env, &env); \
    jarray result; \
    JHandle *arr = (JHandle *)allocArray(JNIEnv2EE(env), ElementTag, len); \
    if (arr == NULL) { \
        ThrowOutOfMemoryError(0, 0); \
	result = 0; \
    } else \
        result = MkRefLocal(env, arr); \
    RECOVER_STACK_BASE(env); \
    return result; \
}

DEFINE_NEWPRIMITIVEARRAY(T_BOOLEAN,Boolean)
DEFINE_NEWPRIMITIVEARRAY(T_BYTE,Byte)
DEFINE_NEWPRIMITIVEARRAY(T_SHORT,Short)
DEFINE_NEWPRIMITIVEARRAY(T_CHAR,Char)
DEFINE_NEWPRIMITIVEARRAY(T_INT,Int)
DEFINE_NEWPRIMITIVEARRAY(T_LONG,Long)
DEFINE_NEWPRIMITIVEARRAY(T_FLOAT,Float)
DEFINE_NEWPRIMITIVEARRAY(T_DOUBLE,Double)

static void* JNICALL
jni_GetPrimitiveArrayElements(JNIEnv *env, jarray array, jboolean *isCopy)
{
    RECORD_STACK_BASE(env, &env);
    void *result;
    HArrayOfByte *arr;

    if (isCopy)
        *isCopy = JNI_FALSE;
    GET_ARRAY(arr, array);
    GET_ARRAY_BODY(arr, result);

    if (pinObj(JNIEnv2EE(env), result) == 0)
        result = NULL;

    RECOVER_STACK_BASE(env);
    return result;
}

static void JNICALL
jni_ReleasePrimitiveArrayElements(JNIEnv *env, jarray array, 
				  void *arrBody, jint mode)
{
    RECORD_STACK_BASE(env, &env);
    unpinObj(JNIEnv2EE(env), arrBody);
    RECOVER_STACK_BASE(env);
}

static void JNICALL
jni_GetPrimitiveArrayRegion(JNIEnv *env, jarray array, jsize start,
			    jsize len, void *buf)
{
    RECORD_STACK_BASE(env, &env);
    jsize arrLen;
    HArrayOfByte *arr;
    int8_t *body;
    int elementSize;

    GET_ARRAY(arr, array);
    GET_ARRAY_BODY(arr, body);
    arrLen = obj_length(arr);
    elementSize = 1 << T_SIZEFIELD(obj_flags(arr));

    if (start < 0 || len < 0 || start + len < 0 || start + len > arrLen) {
        ThrowArrayIndexOutOfBoundsException(0, 0);
    } else {
        memcpy(buf, body + start * elementSize, len * elementSize);
    }
    KEEP_POINTER_ALIVE(body);
    RECOVER_STACK_BASE(env);
}

static void JNICALL
jni_SetPrimitiveArrayRegion(JNIEnv *env, jarray array, jsize start,
			    jsize len, void *buf)
{
    RECORD_STACK_BASE(env, &env);
    jsize arrLen;
    HArrayOfByte *arr;
    int8_t *body;
    int elementSize;

    GET_ARRAY(arr, array);
    GET_ARRAY_BODY(arr, body);
    arrLen = obj_length(arr);
    elementSize = 1 << T_SIZEFIELD(obj_flags(arr));

    if (start < 0 || len < 0 || start + len < 0 || start + len > arrLen) {
        ThrowArrayIndexOutOfBoundsException(0, 0);
    } else {
        memcpy(body + start * elementSize, buf, len * elementSize);
    }
    KEEP_POINTER_ALIVE(body);
    RECOVER_STACK_BASE(env);
}

/*
 * Native Bootstrap
 */

static jint JNICALL
jni_RegisterNatives(JNIEnv *env, jclass clazz,
                    const JNINativeMethod *methods, jint nMethods)
{
    jint result;
    ClassClass *cb;
    int count;

    LINKCLASS_LOCK(EE2SysThread(JNIEnv2EE(env)));
    cb = DeRefClass(env, clazz);

    if (cbIsInterface(cb)) {
        jni_FatalError(env, "RegisterNatives called on an interface");
    }

    /* validate the methods */
    for (count = 0; count < nMethods; count++) {
        struct methodblock *mb;
        char *name = methods[count].name;
	char *sig = methods[count].signature;

        mb = jni_FindMethodBlock(env, clazz, name, sig, FALSE);
        if (mb == NULL || !(mb->fb.access & ACC_NATIVE)) {
	    LINKCLASS_UNLOCK(EE2SysThread(JNIEnv2EE(env)));
            ThrowNoSuchMethodError(0, name);
            return JNI_EINVAL;
        }
    }

    /* Register Natives */
    for (count = 0; count < nMethods; count++) {
        struct methodblock *mb;
        char *name = methods[count].name;
	char *sig = methods[count].signature;
        void *nativeProc = methods[count].fnPtr;

        mb = jni_FindMethodBlock(env, clazz, name, sig, FALSE);
	sysAssert(mb != NULL);
	if (verbose_jni) {
	    jio_fprintf(stderr, "[Registering JNI native method %s.%s]\n",
			cbName(mb->fb.clazz), mb->fb.name);
	}
	mb->invoker = (mb->fb.access & ACC_SYNCHRONIZED) ? 
                      invokeJNISynchronizedNativeMethod :
	              invokeJNINativeMethod;
        mb->code = nativeProc;
    }

    result = JNI_OK;
    if (CompilerRegisterNatives(cb) == FALSE)
	result = JNI_ERR;
    LINKCLASS_UNLOCK(EE2SysThread(JNIEnv2EE(env)));

    return result;
}

static jint JNICALL
jni_UnregisterNatives(JNIEnv *env, jclass clazz)
{
    RECORD_STACK_BASE(env, &env);
    jint result = JNI_OK;
    unsigned short nslots;
    struct methodblock *mb;
    ClassClass *cb;

    LINKCLASS_LOCK(EE2SysThread(JNIEnv2EE(env)));
    cb = DeRefClass(env, clazz);
    if (CompilerUnregisterNatives(cb) == FALSE)
	result = JNI_ERR;

    for (nslots = 0, mb = cbMethods(cb); 
	 nslots < cbMethodsCount(cb); nslots++, mb++)
        if (mb->fb.access & ACC_NATIVE) {
            mb->invoker = invokeLazyNativeMethod;
            mb->code = NULL;
        }

    LINKCLASS_UNLOCK(EE2SysThread(JNIEnv2EE(env)));

    RECOVER_STACK_BASE(env);
    return result;
}

/*
 * Monitors
 */

static jint JNICALL
jni_MonitorEnter(JNIEnv *env, jobject obj)
{
    RECORD_STACK_BASE(env, &env);
    monitorEnter2(JNIEnv2EE(env), (uintptr_t)DeRef2(env, obj));
    RECOVER_STACK_BASE(env);
    return JNI_OK;
}

static jint JNICALL
jni_MonitorExit(JNIEnv *env, jobject obj)
{
    RECORD_STACK_BASE(env, &env);
    monitorExit2(JNIEnv2EE(env), (uintptr_t)DeRef2(env, obj));
    RECOVER_STACK_BASE(env);
    return JNI_OK;
}

static void JNICALL
jni_GetStringRegion(JNIEnv *env, jstring str, jsize start, jsize len, jchar *buf)
{
    RECORD_STACK_BASE(env, &env);
    struct Hjava_lang_String *s;
    unicode *str_head;
    int str_len;
    const jchar *str_chars;
    
    GET_STRING(s, str);
    GET_STRING_BODY(s, str_head, str_chars);

    str_len = unhand(s)->count;
    if (start < 0 || len < 0 || start + len < 0 || start + len > str_len) {
        ThrowStringIndexOutOfBoundsException(0, NULL);
    } else {
        memcpy(buf, str_chars + start, len * sizeof(jchar));
    }
    KEEP_POINTER_ALIVE(str_head);
    RECOVER_STACK_BASE(env);
}

static void JNICALL
jni_GetStringUTFRegion
  (JNIEnv *env, jstring str, jsize start, jsize len, char *buf)
{
    RECORD_STACK_BASE(env, &env);
    struct Hjava_lang_String *s;
    unicode *str_head;
    int str_len;
    const jchar *str_chars;
    
    GET_STRING(s, str);
    GET_STRING_BODY(s, str_head, str_chars);

    str_len = unhand(s)->count;
    if (start < 0 || len < 0 || start + len < 0 || start + len > str_len) {
        ThrowStringIndexOutOfBoundsException(0, 0);
    } else {
        unicode2utf((jchar *)str_chars + start, len, buf, INT_MAX);
    }
    KEEP_POINTER_ALIVE(str_head);
    RECOVER_STACK_BASE(env);
}

static void * JNICALL
jni_GetPrimitiveArrayCritical(JNIEnv *env, jarray array, jboolean *isCopy)
{
    RECORD_STACK_BASE(env, &env);
    ExecEnv *ee = JNIEnv2EE(env);
    HArrayOfByte *arr;
    void *result;

    GET_ARRAY(arr, array);
    result = unhand(arr)->body;
    if (!pinObj(ee, result))
        result = 0;
    else
        ee->critical_count++;

    if (isCopy)
        *isCopy = JNI_FALSE;

    RECOVER_STACK_BASE(env);
    return result;
}

static void JNICALL
jni_ReleasePrimitiveArrayCritical
  (JNIEnv *env, jarray array, void *carray, jint mode)
{
    RECORD_STACK_BASE(env, &env);
    HArrayOfByte *arr;

    GET_ARRAY(arr, array);
    unpinObj(JNIEnv2EE(env), unhand(arr)->body);
    JNIEnv2EE(env)->critical_count--;
    RECOVER_STACK_BASE(env);
}

static const jchar* JNICALL
jni_GetStringCritical(JNIEnv *env, jstring string, jboolean *isCopy)
{
    RECORD_STACK_BASE(env, &env);
    unicode *str_head;
    const jchar *res;
    struct Hjava_lang_String *s;

    GET_STRING(s, string);
    GET_STRING_BODY(s, str_head, res);

    if (isCopy) 
        *isCopy = JNI_FALSE;
    if (!pinObj(JNIEnv2EE(env), str_head))
        res = 0;
    else
        JNIEnv2EE(env)->critical_count++;
    KEEP_POINTER_ALIVE(str_head);
    RECOVER_STACK_BASE(env);
    return res;
}

static void JNICALL
jni_ReleaseStringCritical(JNIEnv *env, jstring str, const jchar *chars)
{
    RECORD_STACK_BASE(env, &env);
    struct Hjava_lang_String *s;
    unicode *str_head;
    const jchar *str_chars;

    GET_STRING(s, str);
    GET_STRING_BODY(s, str_head, str_chars);
    unpinObj(JNIEnv2EE(env), str_head);
    JNIEnv2EE(env)->critical_count--;
    RECOVER_STACK_BASE(env);
}

static jmethodID JNICALL 
jni_FromReflectedMethod(JNIEnv *env, jobject method)
{
    RECORD_STACK_BASE(env, &env);
    jmethodID result = (jmethodID)reflect_get_methodblock(DeRef(env, method));
    RECOVER_STACK_BASE(env);
    return result;
}

static jfieldID JNICALL 
jni_FromReflectedField(JNIEnv *env, jobject field)
{
    RECORD_STACK_BASE(env, &env);
    jfieldID result = (jfieldID)reflect_get_fieldblock(DeRef(env, field));
    RECOVER_STACK_BASE(env);
    return result;
}

static jobject JNICALL
jni_ToReflectedMethod(JNIEnv *env, jclass cls, jmethodID methodID, jboolean isStatic)
{
    RECORD_STACK_BASE(env, &env);
    struct methodblock *mb = (struct methodblock *)methodID;
    jobject result = MkRefLocal(env, reflect_new_method(JNIEnv2EE(env), mb));
    RECOVER_STACK_BASE(env);
    return result;
}

static jobject JNICALL
jni_ToReflectedField(JNIEnv *env, jclass cls, jfieldID fieldID, jboolean isStatic)
{
    RECORD_STACK_BASE(env, &env);
    struct fieldblock *fb = (struct fieldblock *)fieldID;
    jobject result = MkRefLocal(env, reflect_new_field(JNIEnv2EE(env), fb));
    RECOVER_STACK_BASE(env);
    return result;
}

static jweak JNICALL
jni_NewWeakGlobalRef(JNIEnv *env, jobject obj)
{
    RECORD_STACK_BASE(env, &env);
    jweak result;
    result = (jweak)newJNIWeakRef(JNIEnv2EE(env), (JHandle *)DeRef(env, obj));
    if (result == 0)
        ThrowOutOfMemoryError(0, 0);
#ifdef HPROF
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_JNI_WEAK_GLOBALREF_ALLOC)) {
        jvmpi_new_weakref(JNIEnv2EE(env), result, (JHandle *)DeRef(env, obj));
    }
#endif /* HPROF */	
    RECOVER_STACK_BASE(env);
    return result;
}

static void JNICALL
jni_DeleteWeakGlobalRef(JNIEnv *env, jweak ref)
{
    RECORD_STACK_BASE(env, &env);
#ifdef HPROF
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_JNI_WEAK_GLOBALREF_FREE)) {
        jvmpi_delete_weakref(JNIEnv2EE(env), ref);
    }
#endif /* HPROF */	
    if (!deleteJNIWeakRef(JNIEnv2EE(env), (JHandle **)ref))
        jni_FatalError(env, "Illegal ref passed to DeleteWeakGlobalRef");
    RECOVER_STACK_BASE(env);    
}

static jboolean JNICALL
jni_ExceptionCheck(JNIEnv *env)
{
    return exceptionOccurred(JNIEnv2EE(env));
}

/*
 * Native Method Interface function table
 */

const struct JNINativeInterface_ unchecked_jni_NativeInterface = {
    NULL,
    NULL,
    NULL,

    NULL,

    jni_GetVersion,

    jni_DefineClass,
    jni_FindClass,

    jni_FromReflectedMethod,
    jni_FromReflectedField,

    jni_ToReflectedMethod,

    jni_GetSuperclass,
    jni_IsAssignableFrom,

    jni_ToReflectedField,

    jni_Throw,
    jni_ThrowNew,
    jni_ExceptionOccurred,
    jni_ExceptionDescribe,
    jni_ExceptionClear,
    jni_FatalError,

    jni_PushLocalFrame,
    jni_PopLocalFrame,

    jni_NewGlobalRef,
    jni_DeleteGlobalRef,
    jni_DeleteLocalRef,
    jni_IsSameObject,

    jni_NewLocalRef,
    jni_EnsureLocalCapacity,

    jni_AllocObject,
    jni_NewObject,
    jni_NewObjectV,
    jni_NewObjectA,

    jni_GetObjectClass,
    jni_IsInstanceOf,

    jni_GetMethodID,

    jni_CallObjectMethod,
    jni_CallObjectMethodV,
    jni_CallObjectMethodA,
    jni_CallBooleanMethod,
    jni_CallBooleanMethodV,
    jni_CallBooleanMethodA,
    jni_CallByteMethod,
    jni_CallByteMethodV,
    jni_CallByteMethodA,
    jni_CallCharMethod,
    jni_CallCharMethodV,
    jni_CallCharMethodA,
    jni_CallShortMethod,
    jni_CallShortMethodV,
    jni_CallShortMethodA,
    jni_CallIntMethod,
    jni_CallIntMethodV,
    jni_CallIntMethodA,
    jni_CallLongMethod,
    jni_CallLongMethodV,
    jni_CallLongMethodA,
    jni_CallFloatMethod,
    jni_CallFloatMethodV,
    jni_CallFloatMethodA,
    jni_CallDoubleMethod,
    jni_CallDoubleMethodV,
    jni_CallDoubleMethodA,
    jni_CallVoidMethod,
    jni_CallVoidMethodV,
    jni_CallVoidMethodA,

    jni_CallNonvirtualObjectMethod,
    jni_CallNonvirtualObjectMethodV,
    jni_CallNonvirtualObjectMethodA,
    jni_CallNonvirtualBooleanMethod,
    jni_CallNonvirtualBooleanMethodV,
    jni_CallNonvirtualBooleanMethodA,
    jni_CallNonvirtualByteMethod,
    jni_CallNonvirtualByteMethodV,
    jni_CallNonvirtualByteMethodA,
    jni_CallNonvirtualCharMethod,
    jni_CallNonvirtualCharMethodV,
    jni_CallNonvirtualCharMethodA,
    jni_CallNonvirtualShortMethod,
    jni_CallNonvirtualShortMethodV,
    jni_CallNonvirtualShortMethodA,
    jni_CallNonvirtualIntMethod,
    jni_CallNonvirtualIntMethodV,
    jni_CallNonvirtualIntMethodA,
    jni_CallNonvirtualLongMethod,
    jni_CallNonvirtualLongMethodV,
    jni_CallNonvirtualLongMethodA,
    jni_CallNonvirtualFloatMethod,
    jni_CallNonvirtualFloatMethodV,
    jni_CallNonvirtualFloatMethodA,
    jni_CallNonvirtualDoubleMethod,
    jni_CallNonvirtualDoubleMethodV,
    jni_CallNonvirtualDoubleMethodA,
    jni_CallNonvirtualVoidMethod,
    jni_CallNonvirtualVoidMethodV,
    jni_CallNonvirtualVoidMethodA,

    jni_GetFieldID,

    jni_GetObjectField,
    jni_GetBooleanField,
    jni_GetByteField,
    jni_GetCharField,
    jni_GetShortField,
    jni_GetIntField,
    jni_GetLongField,
    jni_GetFloatField,
    jni_GetDoubleField,

    jni_SetObjectField,
    jni_SetBooleanField,
    jni_SetByteField,
    jni_SetCharField,
    jni_SetShortField,
    jni_SetIntField,
    jni_SetLongField,
    jni_SetFloatField,
    jni_SetDoubleField,

    jni_GetStaticMethodID,

    jni_CallStaticObjectMethod,
    jni_CallStaticObjectMethodV,
    jni_CallStaticObjectMethodA,
    jni_CallStaticBooleanMethod,
    jni_CallStaticBooleanMethodV,
    jni_CallStaticBooleanMethodA,
    jni_CallStaticByteMethod,
    jni_CallStaticByteMethodV,
    jni_CallStaticByteMethodA,
    jni_CallStaticCharMethod,
    jni_CallStaticCharMethodV,
    jni_CallStaticCharMethodA,
    jni_CallStaticShortMethod,
    jni_CallStaticShortMethodV,
    jni_CallStaticShortMethodA,
    jni_CallStaticIntMethod,
    jni_CallStaticIntMethodV,
    jni_CallStaticIntMethodA,
    jni_CallStaticLongMethod,
    jni_CallStaticLongMethodV,
    jni_CallStaticLongMethodA,
    jni_CallStaticFloatMethod,
    jni_CallStaticFloatMethodV,
    jni_CallStaticFloatMethodA,
    jni_CallStaticDoubleMethod,
    jni_CallStaticDoubleMethodV,
    jni_CallStaticDoubleMethodA,
    jni_CallStaticVoidMethod,
    jni_CallStaticVoidMethodV,
    jni_CallStaticVoidMethodA,

    jni_GetStaticFieldID,

    jni_GetStaticObjectField,
    jni_GetStaticBooleanField,
    jni_GetStaticByteField,
    jni_GetStaticCharField,
    jni_GetStaticShortField,
    jni_GetStaticIntField,
    jni_GetStaticLongField,
    jni_GetStaticFloatField,
    jni_GetStaticDoubleField,

    jni_SetStaticObjectField,
    jni_SetStaticBooleanField,
    jni_SetStaticByteField,
    jni_SetStaticCharField,
    jni_SetStaticShortField,
    jni_SetStaticIntField,
    jni_SetStaticLongField,
    jni_SetStaticFloatField,
    jni_SetStaticDoubleField,

    jni_NewString,
    jni_GetStringLength,
    jni_GetStringChars,
    jni_ReleaseStringChars,

    jni_NewStringUTF,
    jni_GetStringUTFLength,
    jni_GetStringUTFChars,
    jni_ReleaseStringUTFChars,

    jni_GetArrayLength,
 
    jni_NewObjectArray,
    jni_GetObjectArrayElement,
    jni_SetObjectArrayElement,

    jni_NewBooleanArray,
    jni_NewByteArray,
    jni_NewCharArray,
    jni_NewShortArray,
    jni_NewIntArray,
    jni_NewLongArray,
    jni_NewFloatArray,
    jni_NewDoubleArray,

    (jboolean * (JNICALL *)
      (JNIEnv *env, jbooleanArray array, jboolean *isCopy))
    jni_GetPrimitiveArrayElements,
    (jbyte * (JNICALL *)
      (JNIEnv *env, jbyteArray array, jboolean *isCopy))
    jni_GetPrimitiveArrayElements,
    (jchar * (JNICALL *)
      (JNIEnv *env, jcharArray array, jboolean *isCopy))
    jni_GetPrimitiveArrayElements,
    (jshort * (JNICALL *)
      (JNIEnv *env, jshortArray array, jboolean *isCopy))
    jni_GetPrimitiveArrayElements,
    (jint * (JNICALL *)
      (JNIEnv *env, jintArray array, jboolean *isCopy))
    jni_GetPrimitiveArrayElements,
    (jlong * (JNICALL *)
      (JNIEnv *env, jlongArray array, jboolean *isCopy))
    jni_GetPrimitiveArrayElements,
    (jfloat * (JNICALL *)
      (JNIEnv *env, jfloatArray array, jboolean *isCopy))
    jni_GetPrimitiveArrayElements,
    (jdouble * (JNICALL *)
      (JNIEnv *env, jdoubleArray array, jboolean *isCopy))
    jni_GetPrimitiveArrayElements,

    (void (JNICALL *)
      (JNIEnv *env, jbooleanArray array, jboolean *elems, jint mode))
    jni_ReleasePrimitiveArrayElements,
    (void (JNICALL *)
      (JNIEnv *env, jbyteArray array, jbyte *elems, jint mode))
    jni_ReleasePrimitiveArrayElements,
    (void (JNICALL *)
      (JNIEnv *env, jcharArray array, jchar *elems, jint mode))
    jni_ReleasePrimitiveArrayElements,
    (void (JNICALL *)
      (JNIEnv *env, jshortArray array, jshort *elems, jint mode))
    jni_ReleasePrimitiveArrayElements,
    (void (JNICALL *)
      (JNIEnv *env, jintArray array, jint *elems, jint mode))
    jni_ReleasePrimitiveArrayElements,
    (void (JNICALL *)
      (JNIEnv *env, jlongArray array, jlong *elems, jint mode))
    jni_ReleasePrimitiveArrayElements,
    (void (JNICALL *)
      (JNIEnv *env, jfloatArray array, jfloat *elems, jint mode))
    jni_ReleasePrimitiveArrayElements,
    (void (JNICALL *)
      (JNIEnv *env, jdoubleArray array, jdouble *elems, jint mode))
    jni_ReleasePrimitiveArrayElements,

    (void (JNICALL *)
      (JNIEnv *env, jbooleanArray array, jsize start, jsize l, jboolean *buf))
    jni_GetPrimitiveArrayRegion,
    (void (JNICALL *)
      (JNIEnv *env, jbyteArray array, jsize start, jsize len, jbyte *buf))
    jni_GetPrimitiveArrayRegion,
    (void (JNICALL *)
      (JNIEnv *env, jcharArray array, jsize start, jsize len, jchar *buf))
    jni_GetPrimitiveArrayRegion,
    (void (JNICALL *)
      (JNIEnv *env, jshortArray array, jsize start, jsize len, jshort *buf))
    jni_GetPrimitiveArrayRegion,
    (void (JNICALL *)
      (JNIEnv *env, jintArray array, jsize start, jsize len, jint *buf))
    jni_GetPrimitiveArrayRegion,
    (void (JNICALL *)
      (JNIEnv *env, jlongArray array, jsize start, jsize len, jlong *buf))
    jni_GetPrimitiveArrayRegion,
    (void (JNICALL *)
      (JNIEnv *env, jfloatArray array, jsize start, jsize len, jfloat *buf))
    jni_GetPrimitiveArrayRegion,
    (void (JNICALL *)
      (JNIEnv *env, jdoubleArray array, jsize start, jsize len, jdouble *buf))
    jni_GetPrimitiveArrayRegion,

    (void (JNICALL *)
      (JNIEnv *env, jbooleanArray array, jsize start, jsize l, jboolean *buf))
    jni_SetPrimitiveArrayRegion,
    (void (JNICALL *)
      (JNIEnv *env, jbyteArray array, jsize start, jsize len, jbyte *buf))
    jni_SetPrimitiveArrayRegion,
    (void (JNICALL *)
      (JNIEnv *env, jcharArray array, jsize start, jsize len, jchar *buf))
    jni_SetPrimitiveArrayRegion,
    (void (JNICALL *)
      (JNIEnv *env, jshortArray array, jsize start, jsize len, jshort *buf))
    jni_SetPrimitiveArrayRegion,
    (void (JNICALL *)
      (JNIEnv *env, jintArray array, jsize start, jsize len, jint *buf))
    jni_SetPrimitiveArrayRegion,
    (void (JNICALL *)
      (JNIEnv *env, jlongArray array, jsize start, jsize len, jlong *buf))
    jni_SetPrimitiveArrayRegion,
    (void (JNICALL *)
      (JNIEnv *env, jfloatArray array, jsize start, jsize len, jfloat *buf))
    jni_SetPrimitiveArrayRegion,
    (void (JNICALL *)
      (JNIEnv *env, jdoubleArray array, jsize start, jsize len, jdouble *buf))
    jni_SetPrimitiveArrayRegion,

    jni_RegisterNatives,
    jni_UnregisterNatives,

    jni_MonitorEnter,
    jni_MonitorExit,

    jni_GetJavaVM,

    jni_GetStringRegion,
    jni_GetStringUTFRegion,

    jni_GetPrimitiveArrayCritical,
    jni_ReleasePrimitiveArrayCritical,

    jni_GetStringCritical,
    jni_ReleaseStringCritical,

    jni_NewWeakGlobalRef,
    jni_DeleteWeakGlobalRef,

    jni_ExceptionCheck,
};

JNIEXPORT jint JNICALL
JNI_GetDefaultJavaVMInitArgs(void *args_)
{
    JDK1_1InitArgs *args = (JDK1_1InitArgs *)args_;

    if (args->version == JNI_VERSION_1_2)
        return JNI_OK;

    args->version         = JNI_VERSION_1_1;
    args->properties      = 0;
    args->checkSource     = FALSE;
    args->javaStackSize   = JAVASTACKSIZE;
    args->verifyMode      = VERIFY_REMOTE;
    args->classpath       = NULL;
    args->vfprintf        = 0;
    args->exit            = 0;
    args->abort           = 0;
    args->enableClassGC   = TRUE;
    args->disableAsyncGC  = FALSE;
    args->enableVerboseGC = FALSE;
    args->verbose         = FALSE;
    args->debugging       = FALSE;
    args->debugPort       = 0;
    return GetDefaultJavaVMInitArgs(args);
}

JNIEXPORT jint JNICALL
JNI_CreateJavaVM(JavaVM **vm, void **penv, void *args)
{
    ExecEnv *ee;

    /* 
     * At the moment it's only possible to have one Java VM, 
     * since some of the runtime state is in global variables. 
     */
    if (VM_created)
        return JNI_EEXIST;

    if (InitializeJavaVM(args) == SYS_OK) {
        ExecEnv *ee = EE();
        *vm = (JavaVM *)(&main_vm);
	*(JNIEnv**)penv = EE2JNIEnv(ee);

#ifdef HPROF
	if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_THREAD_START))
	    jvmpi_thread_start(ee->thread);
#endif /* HPROF */

#ifdef BREAKPTS
        if (debugging) 
            notify_debugger_of_thread_start(ee, ee->thread);
#endif /* BREAKPTS */

	ee->stack_base = 0;

	VM_created = 1;
        return JNI_OK;
    } else {
        *vm = 0;
	*(JNIEnv**)penv = 0;
	return JNI_ERR;
    }
}

JNIEXPORT jint JNICALL
JNI_GetCreatedJavaVMs(JavaVM **vm_buf, jsize bufLen, jsize *numVMs)
{
    if (VM_created) {
        if (numVMs)
	    *numVMs = 1;
	if (bufLen > 0)
	    *vm_buf = (JavaVM *)(&main_vm);
    } else
        *numVMs = 0;
    return JNI_OK;
}

static jint JNICALL
jni_DestroyJavaVM(JavaVM *vm)
{
    JNIEnv *env;
    jint res;
#if defined(HPROF) || defined(BREAKPTS)
    ExecEnv *ee;
    JHandle *thr;
#endif /* HPROF || BREAKPTS */

    res = (*vm)->AttachCurrentThread(vm, (void **)&env, NULL);
    if (res < 0) {
        return res;
    }

#if defined(HPROF) || defined(BREAKPTS)
    ee = JNIEnv2EE(env);
    thr = ee->thread;
#endif /* HPROF || BREAKPTS */

    WaitForSingleThreaded();
    RunFinalizersOnExit();

#ifdef HPROF
    if ((JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_THREAD_END)) && thr)
        jvmpi_thread_end(thr);
#endif /* HPROF */

#ifdef BREAKPTS
    if (debugging) 
        notify_debugger_of_thread_end(ee, thr);
#endif /* BREAKPTS */

    PrepareToExit();

    return JNI_ERR;
}

static jint JNICALL
jni_GetJavaVM(JNIEnv *env, JavaVM **vm)
{
    *vm = (JavaVM *)(&main_vm);
    return JNI_OK;
}

static jint JNICALL
jni_AttachCurrentThread(JavaVM *vm, void **penv, void *_args)
{
    ExecEnv *ee;
    JavaVMAttachArgs *args = _args;

    /* Verify that the VM has been initialized */

    if (!VM_created) {
        panic("trying to attach a thread to an uninitialized VM");
    }

    /*
    if (args &&
        (args->version != JNI_VERSION_1_1 ||
	 args->version != JNI_VERSION_1_2)) {
	return JNI_EVERSION;
    }
    */

    /* If this thread is already a Java thread just return
     * it's JNIEnv.  
     *
     * Green threads can't really detached from the VM. So even when
     * ee is not zero, we go on and check ee->initial_stack to determine
     * whether the thread is really in the VM.
     */
    ee = EE();

    if (ee) {
        if (ee->initial_stack) {   /* already attached */
	    *(JNIEnv**)penv = EE2JNIEnv(ee);
	    return JNI_OK;
	}
    }

    {
	sys_thread_t *sys_thr;

	if (SYS_OK != sysThreadAlloc(&sys_thr)) {
	    return JNI_ERR;
	}

	ee = SysThread2EE(sys_thr);

	if (!InitializeExecEnv(ee, (JHandle *)NULL, (void *)&vm)) {
	    return JNI_ERR;
	}

	AdjustUserThreadCount(1);

	/* GC should start to work on the current thread after this,
	 * although we haven't set the ee->thread field yet.
	 */
	{
	    HThreadGroup *group = maingroup;
	    HString *name = NULL;
	    HThread *java_thr =
		(HThread *)allocObject(ee, classJavaLangThread);
	    if (java_thr == NULL) {
	        threadFree();
		return JNI_ERR;
	    }
	    ee->thread = (JHandle *)java_thr;
	    THREAD(java_thr)->eetop = ptr2ll(ee);
	    THREAD(java_thr)->priority = NormalPriority;
	    sysThreadSetPriority(sys_thr, NormalPriority);

	    if (args &&
		args->version == JNI_VERSION_1_2) {
	        if (args->group) {
		    group = (HThreadGroup *)DeRef(env, args->group);
		}
		if (args->name) {
		    name = makeJavaStringUTF(args->name);
		}
	    }

	    if (name) {
	        execute_java_dynamic_method(ee, (JHandle *)java_thr, "<init>",
	    "(Ljava/lang/ThreadGroup;Ljava/lang/Runnable;Ljava/lang/String;)V",
					    group, NULL, name);
	    } else {
	        execute_java_dynamic_method(ee, (JHandle *)java_thr, "<init>",
		    "(Ljava/lang/ThreadGroup;Ljava/lang/Runnable;)V",
					    group, NULL);
	    }

	    if (exceptionOccurred(ee)) {
	        exceptionClear(ee);
	        threadFree();
		return JNI_ERR;
	    }
	}

	*(JNIEnv **)penv = EE2JNIEnv(ee);

	threadInit();

	ee->stack_base = 0;

	return JNI_OK;
    }
}


static jint JNICALL
jni_DetachCurrentThread(JavaVM *vm)
{
    ExecEnv *ee = EE();
    if (ee == 0) {
        return JNI_EDETACHED; /* not attached */
    } else {
        /* The following RECORD_STACK_BASE must not have a matching
	 * RECOVER_STACK_BASE. sysThreadFree requires the conservative
	 * GC to be functional, but destroys the sys_thr structure.
	 */
        RECORD_STACK_BASE(EE2JNIEnv(ee), &vm);
	threadFree();
	sysThreadFree();
	return JNI_OK;
    }
}

static jint JNICALL
jni_GetEnv(JavaVM *vm, void **penv, jint version)
{
    sys_thread_t *self = sysThreadSelf();
    if (self) {
        if (version == JNI_VERSION_1_1 || version == JNI_VERSION_1_2) {
	    *penv = (void *)EE2JNIEnv(SysThread2EE(self));
	    return JNI_OK;
#ifdef HPROF
	} else if (version == JVMPI_VERSION_1) {
	    *penv = jvmpi_interface();
	    return JNI_OK;
#endif /* HPROF */

#ifdef BREAKPTS
	} else if (version == JVMDI_VERSION_1) {
            extern JVMDI_Interface_1 *
                jvmdi_GetInterface_1(JavaVM *interfaces_vm);
	    *penv = (void *)jvmdi_GetInterface_1(vm);
	    return JNI_OK;
#endif /* BREAKPTS */

	} else {
	    *penv = NULL;
	    return JNI_EVERSION;
	}
    } else {
        *penv = NULL;
        return JNI_EDETACHED;
    }
}

const struct JNIInvokeInterface_ jni_InvokeInterface = {
    NULL,
    NULL,
    NULL,

    jni_DestroyJavaVM,
    jni_AttachCurrentThread,
    jni_DetachCurrentThread,
    jni_GetEnv,
};
