/*
 * @(#)jvm.c	1.126 98/11/03
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/* Implement jvm.h functions here, unless they more suitably belong
 * to somewhere else.
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <limits.h>
#include <time.h>
#include <errno.h>

#include "jvm.h"
#include "jni.h"
#include "sys_api.h"
#include "oobj.h"
#include "interpreter.h"
#include "threads.h"
#include "utf.h"
#include "gc.h"
#include "tree.h"
#define _REFLECT_IMPL
#include "reflect.h"
#include "path.h"
#include "common_exceptions.h"
#include "sys_api.h"				/* sysAssert() */
#include "signature.h"
#include "typecodes.h"
#include "wrap.h"
#include "jlong.h"
#include "java_lang_ClassLoader.h"
#include "java_io_InputStream.h"
#include "java_lang_reflect_Method.h"
#include "java_lang_reflect_Constructor.h"
#include "java_lang_reflect_InvocationTargetException.h"
#include "java_lang_reflect_Field.h"
#include "java_lang_Throwable.h"
#include "java_lang_Thread.h"
#include "java_security_AccessControlContext.h"

#ifdef HPROF
#include "vmprofiler.h"
#endif

#include "opcodes.h"

struct methodblock *reflect_invoke_mb;

JNIEXPORT jint JNICALL
JVM_GetInterfaceVersion(void)
{
    return JVM_INTERFACE_VERSION;
}

/* Step back n frames, skip and pseudo frames in between.
 * This function is used in Class.forName, Class.newInstance, Method.Invoke, 
 * AccessController.doPrivileged.
 */

JavaFrame *
getCallerFrame(JavaFrame * frame, int n, JavaFrame * frame_buf)
{
    while (n-- > 0) {
        /* skip pseudo frames or Method.invoke() frames. */
        while (frame && (frame->current_method == 0 ||
			 frame->current_method == reflect_invoke_mb)) {
	    frame = frame->prev;
	}

	if (frame == 0) {
	    return 0;
	}

        /* skip a real frame */
	frame = FRAME_PREV(frame, frame_buf);
    }

    /* skip pseudo frames or Method.invoke() frames. */
    while (frame && (frame->current_method == 0 || 
		     frame->current_method == reflect_invoke_mb)) {
        frame = frame->prev;
    }

    return frame;
}    

ClassClass *
getCallerClass(ExecEnv * ee, int skip)
{
    JavaFrame frame_buf;
    JavaFrame * frame = getCallerFrame(ee->current_frame, skip, &frame_buf);
    if (frame) {
	return frame->current_method->fb.clazz;
    } else {
	return NULL;
    }
}

JNIEXPORT jint JNICALL
JVM_GetLastErrorString(char *buf, int len)
{
    return (jint)sysGetLastErrorString(buf, len);
}

JNIEXPORT char * JNICALL
JVM_NativePath(char *path)
{
    return sysNativePath(path);
}

/*
 * Find primitive classes
 */
JNIEXPORT jclass JNICALL
JVM_FindPrimitiveClass(JNIEnv *env, const char *utf)
{
    return MkRefLocal(env, FindPrimitiveClass((char *)utf));
}

/*
 * Link the class
 */
JNIEXPORT void JNICALL
JVM_ResolveClass(JNIEnv *env, jclass cls)
{
    ClassClass *cb = (ClassClass *)DeRef(env, cls);
    LinkClass(cb);
}

JNIEXPORT jclass JNICALL
JVM_FindClassFromClassLoader(JNIEnv *env, const char *name, jboolean init,
			     jobject loader, jboolean throwError)
{
    return MkRefLocal(env,
		      FindClassFromClassLoader(JNIEnv2EE(env),
			  (char *)name, init,
			  (struct Hjava_lang_ClassLoader *)DeRef(env, loader),
			  throwError));
}

/*
 * Reflection support functions
 */
JNIEXPORT jstring JNICALL
JVM_GetClassName(JNIEnv *env, jclass cls)
{
    ClassClass *cb = (ClassClass *)DeRef(env, cls);
    Hjava_lang_String *hstr;
    char *clname;
    jstring result;

    /* Check the cache */
    hstr = cbClassname(cb);
    if (hstr != NULL) {
	return MkRefLocal(env, hstr);
    }

    clname = sysMalloc(strlen(cbName(cb)) + 1);
    if (clname == 0) {
        return 0;
    }

    /* Benign race to fill cache */
    classname2string(cbName(cb), clname, strlen(cbName(cb)) + 1);

    result = MkRefLocal(env, cbClassname(cb) = makeJavaStringUTF(clname));

    sysFree(clname);
    return result;
}

JNIEXPORT jobjectArray JNICALL
JVM_GetClassInterfaces(JNIEnv *env, jclass cls)
{
    ClassClass *cb = (ClassClass *)DeRef(env, cls);
    HArrayOfObject *list;
    ExecEnv *ee = JNIEnv2EE(env);
    Hjava_lang_Class **listdata;
    int i;

    if (cbName(cb)[0] == SIGNATURE_ARRAY) {
	/* All arrays implement java.lang.Cloneable and java.io.Serializable */
	list = (HArrayOfObject *)allocArray(ee, T_CLASS, 2);
	if (list == 0) {
	    ThrowOutOfMemoryError(ee, 0);
	    return 0;
	}
	listdata = (Hjava_lang_Class **)unhand(list)->body;
	listdata[2] = classJavaLangClass;

	listdata[0] = interfaceJavaLangCloneable;
	listdata[1] = interfaceJavaIoSerializable;
    } else {
	list = (HArrayOfObject *)allocArray(ee, T_CLASS,
					    cbImplementsCount(cb));
	if (list == 0) {
	    ThrowOutOfMemoryError(ee, 0);
	    return 0;
	}
	listdata = (Hjava_lang_Class **)unhand(list)->body;
	listdata[cbImplementsCount(cb)] = classJavaLangClass;
	
	for (i = 0 ; i < (int)cbImplementsCount(cb) ; i++) {
	    if (!ResolveClassConstantFromClass(cb,  cbImplements(cb)[i], 
					       ee, 1 << CONSTANT_Class)) {
		/* An error occurred during class resolution; propagate it */
		return 0;
	    }
	    listdata[i] =
		cbHandle((cbConstantPool(cb)[cbImplements(cb)[i]].clazz));
	}
    }

    KEEP_POINTER_ALIVE(listdata);

    return MkRefLocal(env, list);
}

JNIEXPORT jobject JNICALL
JVM_GetClassLoader(JNIEnv *env, jclass cls)
{
    ClassClass *cb = (ClassClass *)DeRef(env, cls);
    return MkRefLocal(env, cbLoader(cb));
}

JNIEXPORT jboolean JNICALL
JVM_IsInterface(JNIEnv *env, jclass cls)
{
    ClassClass *cb = (ClassClass *)DeRef(env, cls);
    return cbIsInterface(cb);
}

/*
 * Get the class's signers, if any. Make sure to return a new array,
 * not the original. Returning the original is a security hole as
 * the caller could modify it.
 */
JNIEXPORT jobjectArray JNICALL
JVM_GetClassSigners(JNIEnv *env, jclass cls)
{
    Hjava_lang_Class *cb = (Hjava_lang_Class *)DeRef(env, cls);
    HArrayOfObject *copy, *orig;
    HObject **copy_data, **orig_data;
    int i;
    size_t n;

    orig = (HArrayOfObject *)  unhand(cb)->signers;
    if (orig == NULL) 
	return NULL;

    n = obj_length(orig);

    copy = (HArrayOfObject *)ArrayAlloc(T_CLASS, n);
    if (copy == NULL) {
	ThrowOutOfMemoryError(JNIEnv2EE(env), 0);
	return NULL;
    }

    copy_data = (HObject **) unhand(copy)->body;
    orig_data = (HObject **) unhand(orig)->body;

    /*XXX: Use i <= n, NOT i < n, to get copy_data[n] 
             which contains the type of objects in the array */

    for (i=0; i <= n; i++) {
	copy_data[i] = orig_data[i];
    }

    KEEP_POINTER_ALIVE(copy_data);
    KEEP_POINTER_ALIVE(orig_data);

    return MkRefLocal(env, copy);
}

JNIEXPORT void JNICALL
JVM_SetClassSigners(JNIEnv *env, jclass cls, jobjectArray signers)
{
    ClassClass *cb = (ClassClass *)DeRef(env, cls);
    unhand(cb)->signers = (HArrayOfObject *)DeRef(env, signers);
}

JNIEXPORT jobject JNICALL
JVM_GetProtectionDomain(JNIEnv *env, jclass cls)
{
    ClassClass *cb = (ClassClass *)DeRef(env, cls);

    if (cb == NULL) {
	ThrowNullPointerException(JNIEnv2EE(env), 0);
	return NULL;
    } else {
	return MkRefLocal(env, cbProtectionDomain(cb));
    }
}

JNIEXPORT void JNICALL
JVM_SetProtectionDomain(JNIEnv *env, jclass cls, jobject pd)
{
    ClassClass *cb = (ClassClass *)DeRef(env, cls);

    if (cb == NULL) {
	ThrowNullPointerException(0, 0);
    } else {
	/*
	 * if (cbProtectionDomain(cb) != NULL) {
	 *	   ThrowIllegalArgumentException(0, 
	 * "attempt to change the ProtectionDomain of a class");
	 * } else {
	 */
	cbProtectionDomain(cb) = (HObject *) DeRef(env, pd);
	/* } */
    }
}

/* function to abstract the use the address of a frame as a frame id */
static void *GetFrameID(JavaFrame *frame)
{
    if (IS_JIT_FRAME(frame))
	return CompiledFrameID(frame);
    else
	return frame;
}

JNIEXPORT jobject JNICALL
JVM_DoPrivileged(JNIEnv *env, 
		    jclass cls,
		    jobject action,
		    jobject context,
		    jboolean wrapException)
{
    ExecEnv *ee = JNIEnv2EE(env);
    JavaFrame frame_buf;
    JavaFrame *frame = getCallerFrame(ee->current_frame, 1, &frame_buf);
    privileged_item pi;
    struct methodblock *mb, *run_mb;
    ClassClass *cb;
    int i;
    HObject *result = NULL;
    HObject *actionh;
    static char * utf8_literal_run = NULL;
    static char * utf8_literal_voidobject = NULL;

    if (utf8_literal_run == NULL) {
	utf8_literal_run = AddUTF8(ee, "run");
	utf8_literal_voidobject = AddUTF8(ee, "()Ljava/lang/Object;");
    }

    /* make sure there's really a run method before we start */
    actionh = DeRef(ee, action);
    cb = obj_classblock(actionh);
    run_mb = NULL;
    for (i = 0, mb = cbMethods(cb); i < cbMethodsCount(cb); i++, mb++) {
	if (mbName(mb) == utf8_literal_run &&
	    mbSig(mb) == utf8_literal_voidobject) {
	    run_mb = mb;
	    break;
	}
    }

    if (run_mb == NULL) {
	SignalError(ee, JAVAPKG "InternalError", "No run method");
	return NULL;
    }

    /* Skip Method.invoke frames */
    while (frame && (frame->current_method == NULL || 
		     frame->current_method == reflect_invoke_mb)) {
        frame = FRAME_PREV(frame, &frame_buf);
    }

    if (frame) {
	/* doPrivileged frame */
	pi.frame_id = GetFrameID(ee->current_frame);
	/* class of the person that called doPrivileged */
	pi.clazz = frame->current_method->fb.clazz;
	/* context that was passed in */
	pi.context = context;

	/* link ourselves into the list */
	pi.next = ee->privileged;
	ee->privileged = &pi;
    }

    /* invoke the run method */
    result = (HObject *)do_execute_java_method_vararg(ee, actionh,
					       NULL, NULL, run_mb, FALSE,
					       NULL, NULL, FALSE);

    if (exceptionOccurred(ee)) {
	HObject * exc = ee->exception.exc;
	if (is_instance_of(exc, classJavaLangException, ee) &&
	    !is_instance_of(exc, classJavaLangRuntimeException, ee)) {
	    HObject * actionException;

	    /* we should wrap this exception so raise the wrapped exception */
	    exceptionClear(ee);
	    actionException = execute_java_constructor(ee, 
				   "java/security/PrivilegedActionException",
				   NULL,
				   "(Ljava/lang/Exception;)", exc);
	    exceptionThrow(ee, actionException);
	    
	}
    }

    /* done with action, remove ourselves from the list */
    /* should probably assert that ee->privileged == &pi */

    if (frame && (ee->privileged != NULL)) {
	ee->privileged = ee->privileged->next;
    }

    return MkRefLocal(env, result);
}

JNIEXPORT jobject JNICALL
JVM_GetStackAccessControlContext(JNIEnv *env, jclass cls)
{
    ExecEnv *ee = JNIEnv2EE(env);
    struct javaframe *frame, frame_buf;
    void *frame_id, *priv_frame;

    HArrayOfObject *ctx;
    HObject **ctxdata;
    HObject *previous_pd, *pd;
    int n,i;
#define MAX_LOCAL 12
    HObject  *local_array[MAX_LOCAL]; /* try to do it in a single pass */
    int la_len = 0;
    jboolean isPrivileged = JNI_FALSE;
    jobject privilegedContext = NULL;
    static jclass pd_class = NULL;
    static jclass acc_class = NULL;
    static jobject sys_acc = NULL;
    /* static jobject psys_acc = NULL; */
    jobject acc;

    Classjava_security_AccessControlContext *cacc;

    previous_pd = NULL; /* start with the system protection domain */
    priv_frame = NULL; /* set in case we have to make two passes */

    /* count the protection domains on the execution stack. We collapse
     * duplicate consecutive protection domains into a single one, as
     * well as stopping when we hit a privileged frame.
     */
    for (n = 0, frame = ee->current_frame ; frame != 0; ) {
	if (frame->current_method != 0) {

	    frame_id = GetFrameID(frame);
	    /* check the privileged frames to see if we have a match */
	    if (ee->privileged && (ee->privileged->frame_id == frame_id)) {
		/* this frame is privileged */
		isPrivileged = JNI_TRUE;
		privilegedContext = ee->privileged->context;
		priv_frame = frame_id;
		/* make sure pd is set to the caller of doPrivileged */
		pd = cbProtectionDomain(ee->privileged->clazz);
	    } else {
		pd = cbProtectionDomain(frame->current_method->fb.clazz);
	    }
	    /*printf("  <pd = %d, cls = %s>\n", pd,
	      cbName(frame->current_method->fb.clazz));*/
	    if ((pd != previous_pd) && (pd != NULL)) {
		if (la_len < MAX_LOCAL) {
		    local_array[la_len++] = pd;
		}
		n++;
		previous_pd = pd;
	    }
	    if (isPrivileged == JNI_TRUE)
		break;
        }
	frame = FRAME_PREV(frame, &frame_buf);
    }

    if (acc_class == NULL) {
       ClassClass *c = FindClass(ee, "java/security/AccessControlContext",
				 TRUE);
       if (c == NULL) {
	   return NULL;
       }
       acc_class = (*env)->NewGlobalRef(env, MkRefLocal(env, c));
       if (acc_class == NULL) {
	    return NULL;
       }
    }

    /* either all the domains on the stack were system domains, or
       we had a privileged system domain */
    if (n == 0) {
	if (isPrivileged && (privilegedContext == NULL)) {
	    return NULL;
	    /* uncomment this for a little slower version, that returns
	       an acc for privileged system code. We are returning null to 
	       speed up privileged system */
	    /*
	     * if (psys_acc == NULL) {
	     *    acc = (*env)->AllocObject(env, acc_class);
	     *    if (acc == 0) {
	     *      return NULL;
	     *    }
	     *    cacc = (Classjava_security_AccessControlContext *) 
	     *    unhand(DeRef(env, acc));
	     *    cacc->isPrivileged = isPrivileged;
	     *    psys_acc = (*env)->NewGlobalRef(env, acc);
	     * }
	     * return psys_acc;
	     */
	} else {
	    if ((privilegedContext != NULL) || (sys_acc == NULL)) {
	      acc = (*env)->AllocObject(env, acc_class);
	      if (acc == 0) {
		  /* exception already thrown */
		  return NULL;
	      }
	      if (privilegedContext != NULL) {
		  cacc = (Classjava_security_AccessControlContext *) 
		      unhand(DeRef(env, acc));
		  cacc->isPrivileged = isPrivileged;
		  cacc->privilegedContext = 
		      (struct Hjava_security_AccessControlContext *)
		                                           DeRef(env, privilegedContext);
		  return acc;
	      }
	      if (sys_acc) {
		  /* delete the cached global ref */
		  (*env)->DeleteGlobalRef(env, sys_acc);
	      }
	      sys_acc = (*env)->NewGlobalRef(env, acc);
	    }
	    return sys_acc;
	}
    }

    /* allocate the array */
    ctx = (HArrayOfObject *)allocArray(ee, T_CLASS, n);
    if (ctx == 0) {
	ThrowOutOfMemoryError(JNIEnv2EE(env), 0);
	return 0;
    }

    ctxdata = (HObject **)unhand(ctx)->body;

    if (pd_class == NULL) {
        jclass c = (*env)->FindClass(env, "java/security/ProtectionDomain");
	pd_class = (*env)->NewGlobalRef(env, c);
	(*env)->DeleteLocalRef(env, c);
    }

    ctxdata[n] = (HObject *) DeRef(env,pd_class);

    /* check and see if we can punt on a second pass */
    if (la_len == n) {
	for (i=0; i < n; i++) {
	    ctxdata[i] = local_array[i];
	}
    } else {
	/* nope, now that we have the count, do it again */
	previous_pd = NULL; /* start with the system protection domain */

	/* fill the array */
	for (i=0, frame = ee->current_frame ; frame != 0 && i < n;) {
	    if (frame->current_method != 0) {
		frame_id = GetFrameID(frame);
		if (frame_id == priv_frame) {
		    pd = cbProtectionDomain(ee->privileged->clazz);
		} else {
		    pd = cbProtectionDomain(frame->current_method->fb.clazz);
		}
		if ((pd != previous_pd) && (pd != NULL)) {
		    ctxdata[i++] = (HObject *) pd;
		    previous_pd = pd;
		}
		if (frame_id == priv_frame)
		    break;
	    }
	    frame = FRAME_PREV(frame, &frame_buf);
	}
    }

    KEEP_POINTER_ALIVE(ctxdata);

    acc = (*env)->AllocObject(env, acc_class);
    if (acc == 0) {
	/* exception already thrown */
	return NULL;
    }
    
    cacc = (Classjava_security_AccessControlContext *) unhand(DeRef(env, acc));
    cacc->context = ctx;
    cacc->isPrivileged = isPrivileged;
    cacc->privilegedContext = 
	(struct Hjava_security_AccessControlContext *)DeRef(env, privilegedContext);
    return acc;
}

JNIEXPORT jboolean JNICALL
JVM_IsArrayClass(JNIEnv *env, jclass cls)
{
    ClassClass *cb = (ClassClass *)DeRef(env, cls);
    return cbName(cb)[0] == SIGNATURE_ARRAY;
}

JNIEXPORT jboolean JNICALL
JVM_IsPrimitiveClass(JNIEnv *env, jclass cls)
{
    ClassClass *cb = (ClassClass *)DeRef(env, cls);
    return cbIsPrimitive(cb);
}

JNIEXPORT jclass JNICALL
JVM_GetComponentType(JNIEnv *env, jclass cls)
{
    ClassClass *ccb;
    ClassClass *cb = (ClassClass *)DeRef(env, cls);
    if (cbName(cb)[0] != SIGNATURE_ARRAY) {
	return NULL;
    }
    
    if ((ccb = reflect_find_class(&cbName(cb)[1], cb, NULL)) == NULL) {
	return NULL;				/* exception */
    }
    return MkRefLocal(env, ccb);
}

/* Remember that ACC_SUPER is added by the VM, and must be stripped. */
JNIEXPORT jint JNICALL
JVM_GetClassModifiers(JNIEnv *env, jclass cls)
{
    ClassClass *cb = (ClassClass *)DeRef(env, cls);
    int icount = cbInnerClassesCount(cb);
    int i;
    int access;
    union cp_item_type *cp = cbConstantPool(cb);

    /* Assume this is a toplevel class. */
    access = cbAccess(cb);

    /* But check if it happens to be member class. */
    if (icount != 0) {
	for (i = 0; i < icount; i++) {
	    int ioff = cbInnerClasses(cb)[i].inner_class_info_index;
	    
	    if (!ResolveClassConstantFromClass(cb, ioff, JNIEnv2EE(env),
					       1 << CONSTANT_Class))
		return 0;
	    
	    if (cb == cp[ioff].clazz) {
		/* Aha, this is really a member class. */
		access = cbInnerClasses(cb)[i].inner_class_access_flags;
		break;
	    }
	}
    }

    return (access & (~ACC_SUPER)) & ACC_WRITTEN_FLAGS;
}

JNIEXPORT jobjectArray JNICALL
JVM_GetClassFields(JNIEnv *env, jclass cls, jint which)
{
    ClassClass *cb = (ClassClass *)DeRef(env, cls);
    return MkRefLocal(env, reflect_fields(JNIEnv2EE(env), cb, which));
}

JNIEXPORT jobjectArray JNICALL
JVM_GetClassMethods(JNIEnv *env, jclass cls, jint which)
{
    ClassClass *cb = (ClassClass *)DeRef(env, cls);
    return MkRefLocal(env, reflect_methods(JNIEnv2EE(env), cb, which));
}

JNIEXPORT jobjectArray JNICALL
JVM_GetClassConstructors(JNIEnv *env, jclass cls, jint which)
{
    ClassClass *cb = (ClassClass *)DeRef(env, cls);
    return MkRefLocal(env, reflect_constructors(JNIEnv2EE(env), cb, which));
}

JNIEXPORT jobject JNICALL
JVM_GetClassField(JNIEnv *env, jclass cls, jstring name, jint which)
{
    ClassClass *cb = (ClassClass *)DeRef(env, cls);
    jobject result;
    const char *utfName;
    if (name == 0) {
        ThrowNoSuchFieldException(0, NULL);
	return 0;
    }
    utfName = (*env)->GetStringUTFChars(env, name, 0);
    if (utfName == 0) {
        return 0;
    }
    result = MkRefLocal(env,
	reflect_field(JNIEnv2EE(env), cb, (char *)utfName, which));
    (*env)->ReleaseStringUTFChars(env, name, utfName);
    return result;
}

JNIEXPORT jobject JNICALL
JVM_GetClassMethod(JNIEnv *env, jclass cls, jstring name, jobjectArray types,
		   jint which)
{
    ClassClass *cb = (ClassClass *)DeRef(env, cls);
    jobject result;
    const char *utfName;
    if (name == 0) {
        ThrowNoSuchFieldException(0, NULL);
	return 0;
    }
    utfName = (*env)->GetStringUTFChars(env, name, 0);
    if (utfName == 0) {
        return 0;
    }
    result =
	MkRefLocal(env, 
	    reflect_method(JNIEnv2EE(env),
		cb,
		(char *)utfName, 
		(HArrayOfObject *)DeRef(env, types),
		which));
    (*env)->ReleaseStringUTFChars(env, name, utfName);
    return result;
}

JNIEXPORT jobject JNICALL
JVM_GetClassConstructor(JNIEnv *env, jclass cls, jobjectArray types,
			jint which)
{
    ClassClass *cb = (ClassClass *)DeRef(env, cls);

    return MkRefLocal(env,
	reflect_constructor(JNIEnv2EE(env),
	    cb, 
	    (HArrayOfObject *)DeRef(env, types), 
	    which));
}


JNIEXPORT jobjectArray JNICALL
JVM_GetDeclaredClasses(JNIEnv *env, jclass ofClass)
{
    ExecEnv *ee = JNIEnv2EE(env);
    ClassClass *cb = (ClassClass *)DeRef(env, ofClass);
    union cp_item_type *cp = cbConstantPool(cb);
    ClassClass *icb;
    union cp_item_type *icp;
    ClassClass **member_cbs;
    int member_count = 0;
    HArrayOfObject *ret;
    ClassClass **uret;
    int icount = cbInnerClassesCount(cb);
    int i, j;
    
    if (icount == 0) {  /* neither an inner nor an outer class */
	ret = (HArrayOfObject *)allocArray(ee, T_CLASS, 0);
	if (ret == 0) {
	    ThrowOutOfMemoryError(ee, 0);
	    return 0;
	}
	return MkRefLocal(env, ret);
    }

    member_cbs = (ClassClass**)sysMalloc(icount * sizeof(ClassClass*));
    if (member_cbs == 0) {
	ThrowOutOfMemoryError(ee, 0);
	return 0;
    }
    
    /* find any "member" classes ... */
    for (i = 0; i < icount; i++) {
	int ooff = cbInnerClasses(cb)[i].outer_class_info_index;
	int ioff = cbInnerClasses(cb)[i].inner_class_info_index;
	bool_t agree = FALSE;
	
	if (ooff == 0) /* not a member */
	    continue;
	if (!ResolveClassConstantFromClass(cb, ooff, ee, 1 << CONSTANT_Class))
	    goto exception;
	if (cp[ooff].clazz != cb)
	    continue;
	
	/* "ofClass" is the outer class, so this is a member */
	if (!ResolveClassConstantFromClass(cb, ioff, ee, 1 << CONSTANT_Class))
	    goto exception;
	
	/* does member agree that "ofClass" is indeed its outer class? */
	icb = cp[ioff].clazz;
	icp = cbConstantPool(icb);
	for (j = 0; j < cbInnerClassesCount(icb); j++) {
	    int i_ioff = cbInnerClasses(icb)[j].inner_class_info_index;
	    int i_ooff = cbInnerClasses(icb)[j].outer_class_info_index;
	    if (!ResolveClassConstantFromClass(icb, i_ioff, ee,
					       1 << CONSTANT_Class))
		goto exception;
	    if (icp[i_ioff].clazz != icb)
		continue;
	    if (i_ooff == 0)
		break; /* member says it is not a member; agree is FALSE! */
	    if (!ResolveClassConstantFromClass(icb, i_ooff, ee,
					       1 << CONSTANT_Class))
		goto exception;
	    if (icp[i_ooff].clazz == cb) {
		member_cbs[member_count++] = cp[ioff].clazz;
		agree = TRUE;
		break;
	    }
	}
	if (!agree) {
	    char buf[MSG_BUF_LEN];
	    jio_snprintf(buf, sizeof(buf), 
			 "%s and %s disagree on InnerClasses attribute", 
			 cbName(cb), cbName(icb));
	    ThrowIncompatibleClassChangeError(ee, buf);
	    goto exception;
	}
    }

    ret = (HArrayOfObject *)allocArray(ee, T_CLASS, member_count);
    if (ret == 0) {
	ThrowOutOfMemoryError(ee, 0);
	goto exception;
    }
    uret = (ClassClass**)unhand(ret)->body;
    for (i = 0; i < member_count; i++)
	uret[i] = member_cbs[i];
    KEEP_POINTER_ALIVE(uret);
    sysFree(member_cbs);
    return MkRefLocal(env, ret);

exception:
    sysFree(member_cbs);
    return NULL;
}

JNIEXPORT jclass JNICALL
JVM_GetDeclaringClass(JNIEnv *env, jclass ofClass)
{
    ExecEnv *ee = JNIEnv2EE(env);
    ClassClass *cb = (ClassClass *)DeRef(env, ofClass);
    union cp_item_type *cp = cbConstantPool(cb);
    ClassClass *ocb;
    union cp_item_type *ocp;
    int icount = cbInnerClassesCount(cb);
    int i, j;
    
    if (icount == 0)  /* neither an inner nor an outer class */
	return NULL;
    
    for (i = 0; i < icount; i++) {
	int ioff =  cbInnerClasses(cb)[i].inner_class_info_index;
	int ooff = cbInnerClasses(cb)[i].outer_class_info_index;
	
	if (!ResolveClassConstantFromClass(cb, ioff, ee, 1 << CONSTANT_Class))
	    return NULL;
	if (cp[ioff].clazz != cb)
	    continue;
	
	/* "ofClass" is the inner index, so this must be the right record */
	if (ooff == 0) 
	    return NULL; /* not a member */
	if (!ResolveClassConstantFromClass(cb, ooff, ee, 1 << CONSTANT_Class))
	    return NULL;

	/* does outer class agree that "ofClass" is its member? */
	ocb = cp[ooff].clazz;
	ocp = cbConstantPool(ocb);
	for (j = 0; j < cbInnerClassesCount(ocb); j++) {
	    int o_ioff = cbInnerClasses(ocb)[j].inner_class_info_index;
	    int o_ooff = cbInnerClasses(ocb)[j].outer_class_info_index;
	    
	    if (o_ooff == 0)
		continue;
	    if (!ResolveClassConstantFromClass(ocb, o_ooff, ee,
					       1 << CONSTANT_Class))
		return NULL;
	    if (ocp[o_ooff].clazz != ocb)
		continue;
	    if (!ResolveClassConstantFromClass(ocb, o_ioff, ee,
					       1 << CONSTANT_Class))
		return NULL;
	    if (ocp[o_ioff].clazz == cb) {
		/* verified both parties are in agreement */
		return MkRefLocal(env, cp[ooff].clazz);
	    }
	}
	/* We found entry in "ofClass". But doesn't verify with outer class. */
	{
	    char buf[MSG_BUF_LEN];
	    jio_snprintf(buf, sizeof(buf), 
			 "%s and %s disagree on InnerClasses attribute", 
			 cbName(cb), cbName(ocb));
	    ThrowIncompatibleClassChangeError(ee, buf);
	    return NULL;
	}
    }
    return NULL;
}

JNIEXPORT jclass JNICALL
JVM_GetCallerClass(JNIEnv *env, int n)
{
    ExecEnv *ee = JNIEnv2EE(env);
    ClassClass *current_class = getCallerClass(ee, n);

    return MkRefLocal(env, current_class); 
}

JNIEXPORT jobject JNICALL
JVM_NewInstance(JNIEnv *env, jclass cls)
{
    ClassClass *cb = (ClassClass *)DeRef(env, cls);
    Hjava_lang_Object* obj;
    ExecEnv *ee = JNIEnv2EE(env);
    struct methodblock *mb;
    int len;
    ClassClass *mycb = getCallerClass(ee, 2);

    if (cbAccess(cb) & (ACC_INTERFACE | ACC_ABSTRACT)) { 
	ThrowInstantiationException(0, cbName(cb));
	return NULL;
    }

    ENSURE_INITIALIZED(ee, cb);

    for (len = cbMethodsCount(cb), mb = cbMethods(cb); --len >= 0; mb++) 
	if (mb->fb.name == utf8_literal_obj_init_name &&
	    mb->fb.signature == utf8_literal_null_init_sig)
	    break;
    if (len < 0) { 
	ThrowInstantiationException(0, cbName(cb));
	return NULL;
    }

    if (!VerifyClassAccess(mycb, cb, FALSE) ||
	!VerifyFieldAccess2(mycb, cb, mb->fb.access, FALSE, TRUE)) {
	ThrowIllegalAccessException(0, cbName(cb));
        return NULL;
    }

    /* Allocate the object  */
    if ((obj = allocObject(ee, cb)) == 0) { 
	ThrowOutOfMemoryError(ee, 0);
	return NULL;
    }

    /* Invoke the no-arg constructor */
    do_execute_java_method(ee, obj, 0, 0, mb, FALSE);
    
    return MkRefLocal(env, obj);
}

/*
 * Called from System.c
 */

JNIEXPORT jlong JNICALL
JVM_CurrentTimeMillis(JNIEnv *env, jclass ignored) {
    return sysTimeMillis();
}

JNIEXPORT void JNICALL
JVM_ArrayCopy(JNIEnv *env, jclass ignored,
	      jobject src, jint src_pos, 
	      jobject dst, jint dst_pos, jint length)
{
    HObject *srch = DeRef(env, src);
    HObject *dsth = DeRef(env, dst);
    OBJECT *srcptr, *dstptr;

    if ((srch == 0) || (dsth == 0)) {
	ThrowNullPointerException(JNIEnv2EE(env), 0);
	return;
    } else if ((obj_flags(srch) == T_NORMAL_OBJECT) || 
	       (obj_flags(dsth) == T_NORMAL_OBJECT) || 
	       (obj_flags(srch) != obj_flags(dsth))) {
	ThrowArrayStoreException(0, 0);
	return;
    } else if ((length < 0) || (src_pos < 0) || (dst_pos < 0) ||
	(length + src_pos > obj_length(srch)) ||
	(length + dst_pos > obj_length(dsth))) {
	ThrowArrayIndexOutOfBoundsException(0, 0);
	return;
    } 
    srcptr = unhand(srch);
    dstptr = unhand(dsth);
    switch(obj_flags(srch)) { 
         case T_BOOLEAN: case T_BYTE: 
	    memmove(&((ArrayOfByte *)dstptr)->body[dst_pos], 
		    &((ArrayOfByte *)srcptr)->body[src_pos],
		    length * sizeof(((ArrayOfByte *)dstptr)->body[0]));
	    break;

	case T_SHORT:
	    memmove(&((ArrayOfShort *)dstptr)->body[dst_pos], 
		    &((ArrayOfShort *)srcptr)->body[src_pos],
		    length * sizeof(((ArrayOfShort *)dstptr)->body[0]));
	    break;

	case T_CHAR:
	    memmove(&((ArrayOfChar *)dstptr)->body[dst_pos], 
		    &((ArrayOfChar *)srcptr)->body[src_pos],
		    length * sizeof(((ArrayOfChar *)dstptr)->body[0]));
	    break;

	case T_INT:
	    memmove(&((ArrayOfInt *)dstptr)->body[dst_pos], 
		    &((ArrayOfInt *)srcptr)->body[src_pos],
		    length * sizeof(((ArrayOfInt *)dstptr)->body[0]));
	    break;

	case T_LONG:
	    memmove(&((ArrayOfLong *)dstptr)->body[dst_pos], 
		    &((ArrayOfLong *)srcptr)->body[src_pos],
		    length * sizeof(((ArrayOfLong *)dstptr)->body[0]));
	    break;

	case T_FLOAT: 
	    memmove(&((ArrayOfFloat *)dstptr)->body[dst_pos], 
		    &((ArrayOfFloat *)srcptr)->body[src_pos],
		    length * sizeof(((ArrayOfFloat *)dstptr)->body[0]));
	    break;

	case T_DOUBLE: 
	    memmove(&((ArrayOfDouble *)dstptr)->body[dst_pos], 
		    &((ArrayOfDouble *)srcptr)->body[src_pos],
		    length * sizeof(((ArrayOfDouble *)dstptr)->body[0]));
	    break;

	case T_CLASS: {
	    size_t srclen = obj_length(srch);
	    size_t dstlen = obj_length(dsth);
	    ArrayOfObject *osrcptr = (ArrayOfObject *)srcptr;
	    ArrayOfObject *odstptr = (ArrayOfObject *)dstptr;
	    ClassClass *src_class = (ClassClass *)osrcptr->body[srclen];
	    ClassClass *dst_class = (ClassClass *)odstptr->body[dstlen];
	    int i;
	    if ((src_class != dst_class) && (dst_class != classJavaLangObject)){
		/* if the array types are not identical we need to check each 
		 * non-null element. Note that you can't do thecheck first and
		 * then use memmove later. This would create an opportunity
		 * for another thread to modify the array. */
		for (i = 0 ; i < length ; i++) {
		    HObject *elem = osrcptr->body[src_pos + i];
		    if (elem && !is_instance_of(elem, dst_class, 0)) {
			ThrowArrayStoreException(0, 0);
			return;
		    }
		    odstptr->body[dst_pos + i] = elem;
		}
	    } else {
		/*
                 * We can't use memmove or memcpy to move handles around since
                 * there's no guarantee that they will be moved atomically.
                 */
 		HObject **dstiter = &odstptr->body[dst_pos];
		HObject **srciter = &osrcptr->body[src_pos];
		if (odstptr != osrcptr || dstiter < srciter) {
		    /* copy from bottom to top. */
		    for (i = 0; i < length; i++) {
			*dstiter++ = *srciter++;
		    }
		} else {
		    /* copy from top to bottom. */
		    dstiter += length;
		    srciter += length;
		    for (i = 0; i < length; i++) {
			*--dstiter = *--srciter;
		    }
		}
	    }
	} 
    }
    /* Keep srcptr and dstptr from being moved */
    KEEP_POINTER_ALIVE(srcptr);
    KEEP_POINTER_ALIVE(dstptr);
}

/* Called from Runtime.c */

JNIEXPORT void JNICALL
JVM_Exit(jint code)
{
    if (code == 0) {
        RunFinalizersOnExit();
    }
    Exit((int)code);
}

JNIEXPORT void JNICALL
JVM_OnExit(void (*func)(void))
{
    AtExit(func);
}

JNIEXPORT void JNICALL
JVM_GC(void)
{
    gc(~0);  /* Force compaction by setting huge free space goal */
}

JNIEXPORT jlong JNICALL
JVM_MaxObjectInspectionAge()
{
    jlong d = (jlong)sysGetMilliTicks() - (jlong)timeOfLastGC;
    return (d >= 0) ? d : jlong_shl(jlong_one, 32);
}

JNIEXPORT void JNICALL
JVM_TraceInstructions(jboolean on)
{
#ifdef TRACING
    trace = on;
#endif
}

JNIEXPORT void JNICALL
JVM_TraceMethodCalls(jboolean on)
{
#ifdef TRACING
    tracem = on;
#endif
}

JNIEXPORT jlong JNICALL
JVM_TotalMemory(void)
{
    return ll_add(TotalObjectMemory(), TotalHandleMemory());
}

JNIEXPORT jlong JNICALL
JVM_FreeMemory(void)
{
    return ll_add(FreeObjectMemory(), FreeHandleMemory());
}

/* Called from Throwable.c */

#undef min
#define min(x, y) (((int)(x) < (int)(y)) ? (x) : (y))

/* NOTE: The second argument to printStackTrace0() is expected to be an object
 * with a void println(char[]) method, such as a PrintStream or a PrintWriter.
 */

JNIEXPORT void JNICALL
JVM_PrintStackTrace(JNIEnv *env, jobject throwable, jobject printable)
{
    Hjava_lang_Throwable *o = (Hjava_lang_Throwable *)DeRef(env, throwable);
    HObject *s = (HObject *)DeRef(env, printable);
    struct HArrayOfObject *backtrace =
	(HArrayOfObject *)(unhand(o)->backtrace);
    struct HArrayOfInt *pclistH = 0;

    if (backtrace != NULL)
	pclistH = (HArrayOfInt *)(unhand(backtrace)->body[0]);
    if (pclistH) {
	ExecEnv *ee = JNIEnv2EE(env);
	int32_t *dummy_ptr = unhand(pclistH)->body;
	unsigned char **pclist = (unsigned char **)(dummy_ptr);
	unsigned char **pclistend = 
	     (unsigned char **)(pclist + obj_length(pclistH));
	HObject **dummy_ptr2 = unhand(backtrace)->body;
	Hjava_lang_Object **objects = dummy_ptr2;
	objects++;
	for (; pclist < pclistend; pclist++, objects++) {
	    HString *jstr = NULL;
	    if (is_instance_of(*objects, classJavaLangString, ee)) {
	        jstr = (HString *)(*objects);
	    } else if (*pclist != 0) {
		char where[MSG_BUF_LEN];
		int i = 0;
		strncpy(where + i, "\tat ", 4);
		i += 4;
		pc2string(*pclist, 0, where + i, where + sizeof(where));
		jstr = makeJavaStringUTF(where);
	    }
	    if (exceptionOccurred(ee)) {
	        return;
	    }
	    if (jstr) {
		execute_java_dynamic_method(ee, s, "println",
					    "(Ljava/lang/String;)V", jstr);
	    }
	}
	KEEP_POINTER_ALIVE(dummy_ptr);
	KEEP_POINTER_ALIVE(dummy_ptr2);
    }
}

/*
 * The pc's are saved as an int[] (to make it easy for the GC).  The
 * following constant indicates how may array elements are needed per pc
 * It will be 1 on 32-bit machines and 2 on 64-bit machines.
 */
void 
fillInStackTrace(Hjava_lang_Throwable *o, ExecEnv *ee) {
    #define MULTIPLIER (sizeof(char *)/sizeof(int32_t))
    int size;
    Classjava_lang_Throwable *obj = unhand(o);
    JavaFrame *start_frame = ee->current_frame;
    struct HArrayOfInt *backtrace;
    struct HArrayOfObject *classlist = (HArrayOfObject *)obj->backtrace;
    JavaFrame *frame;
    int32_t *dummy_ptr;
    int32_t *dummy_ptr2;
    unsigned char **pclist, **pclistend;
    Hjava_lang_Object **objects;
    /* two buffers to avoid aliasing conflicts */
    JavaFrame start_frame_buf, frame_buf; 
    
    /* Go backwards, removing all frames that are just part of the <init> of
     * the throwable object "o"
     */

    if (start_frame != NULL) {
        struct methodblock *mb = start_frame->current_method;
    	if (mb != NULL) {
	    if ((mb->fb.clazz == classJavaLangThrowable) &&
		(strcmp(mb->fb.name, "fillInStackTrace") == 0) &&
		mb->fb.access & ACC_NATIVE)
	        start_frame = FRAME_PREV(start_frame, &start_frame_buf);
	}
    }
    while (start_frame != NULL) {
    	if (start_frame->current_method != NULL) {
	    if (fieldname(&start_frame->current_method->fb) != 
		    utf8_literal_obj_init_name ||
		(!IS_JIT_FRAME(start_frame) &&
		 start_frame->vars[0].h != (JHandle *)o))
		break;
        }
	start_frame = FRAME_PREV(start_frame, &start_frame_buf);
    }

    /* how many useful frames are there on the stack? */
    for (size = 0, frame = start_frame; frame != NULL; ) {
	if (frame->current_method != NULL) {
	    size++;
        }
	frame = FRAME_PREV(frame, &frame_buf);
    }

    /*
     * The classlist contains a pointer to an int[] and a list of
     * java.lang.Class.  The first item is the int[].  The size of
     * the array must be the length of the stack trace + 1.
     */
    if ((classlist == NULL) || obj_length(classlist) < size + 1) {
	HObject **classlist_body = NULL;
	classlist = (HArrayOfObject *)allocArray(ee, T_CLASS, size + 1);
	if (classlist == NULL) {
	    return;
	}
	classlist_body = unhand(classlist)->body;
	classlist_body[size + 1] = (HObject *)classJavaLangObject;
	KEEP_POINTER_ALIVE(classlist_body);
    }

    dummy_ptr2 = (int32_t *)unhand(classlist)->body;
    objects = (Hjava_lang_Object **)(dummy_ptr2);

    backtrace = (HArrayOfInt *)*objects;

    /* if backtrace is null or not long enough, increase it */
    if ((backtrace == NULL) || obj_length(backtrace) < size * MULTIPLIER) {
	backtrace = (HArrayOfInt *)allocArray(ee, T_INT, size * MULTIPLIER);
	if (backtrace == NULL) {
	    return;
	}
	*objects = (HObject *)backtrace;
    }

    objects++;

    /* Fill in the backtrace array with the pc's */
    dummy_ptr = unhand(backtrace)->body;
    pclist = (unsigned char **)(dummy_ptr);
    pclistend = (unsigned char **)(dummy_ptr + obj_length(backtrace));

    for (frame = start_frame; frame != NULL; ) {
	if (frame->current_method != NULL) {
	    struct methodblock *mb = frame->current_method;
	    if (mb->fb.access & ACC_NATIVE) {
	        *pclist = mb->code;
	    } else if (IS_JIT_FRAME(frame)) {
		*pclist = CompiledCodePC(frame, mb);
	    } else {
	        *pclist = frame->lastpc;
            }
	    /* We must eagerly compute the line number information for
	     * <clinit> because it may be freed before the class is GC'ed.
	     */
	    if (mb->fb.name == utf8_literal_cls_init_name) {
	        char where[MSG_BUF_LEN];
		HString *jstr;
		int i = 0;
		strncpy(where + i, "\tat ", 4);
		i += 4;
		pc2string(*pclist, frame, where + i, where + sizeof(where));
		jstr = makeJavaStringUTF(where);
		if (jstr == NULL) {
		    return;
		}
		*objects++ = (Hjava_lang_Object *)jstr;
	    } else {
	        *objects++ = (Hjava_lang_Object *)mb->fb.clazz;
	    }
	    pclist++;
        }
	frame = FRAME_PREV(frame, &frame_buf);
    } 
    /* Zero out any remaining space. */
    while (pclist < pclistend) {
	*pclist++ = 0;
	*objects++ = 0;
    }
    KEEP_POINTER_ALIVE(dummy_ptr);
    KEEP_POINTER_ALIVE(dummy_ptr2);

    /* Update the obj->backtrace field at the end to make sure
     * classlist has been built properly.
     */
    obj->backtrace = (HObject *)classlist;
}

JNIEXPORT void JNICALL
JVM_FillInStackTrace(JNIEnv *env, jobject throwable)
{
    ExecEnv *ee = JNIEnv2EE(env);
    Hjava_lang_Throwable *obj = (Hjava_lang_Throwable *)DeRef(env, throwable);
    fillInStackTrace(obj, ee);    
}

/*
 * rand1, rand2 are used to compute "Object Indentity Hash Values."  They are
 * used to mangle the object pointer (which was returned as the identity hash
 * code in releases prior to JDK 1.2).  This increases the difficulty of
 * causing real harm in the event that the type system is compromised (e.g.,
 * by a verifier bug). 
 *
 * rand1 and rand2 are initialized to pseudo-random values by a call to
 * InitializeObjectHash().  This must be called once, before the first call to
 * JVM_IHashCode, and never again.  FAILURE TO FOLLOW THIS PROTOCOL WILL CAUSE
 * OBJECT HASH VALUES TO CHANGE, WITH DIRE CONSEQUENCES!
 *
 * The job of InitializeObjectHash() is to set rand1 and rand2 to values
 * that are "reasonably random", subject to the constraint that they may
 * be computed quickly (so as not to slow down VM initialization).  The
 * exact computation is immaterial and may change from release to release.
 */

static uint32_t rand1 = 0, rand2;
void InitializeObjectHash(void)
{
    uint32_t int1, int2;
    rand1 = (uint32_t)&int1 + int2 + (uint32_t)clock();
    rand2 = (uint32_t)&int2 + int1 + (uint32_t)time(NULL);

    /*
     * Ensure that rand1 will be nonzero, which allows JVM_IHashCode
     * to assert that this initialization function has been called.
     */
    if (rand1==0)
	rand1 = 0xdeadbeef;
}

/* Called from object.c */
JNIEXPORT jint JNICALL
JVM_IHashCode(JNIEnv *env, jobject obj)
{
    uint32_t tmp;

    /* Effectively asserts that we've already called InitializeObjectHash. */
    sysAssert(rand1 != 0);

    /*
     * Return value is ((<obj ptr> ^ rand1) rot 17) ^ rand2.  This mangles
     * the pointer sufficiently that it should be pretty tough to
     * guess it without access to rand1 and rand2.  The mangling function
     * (due to John Rose) is clearly an automorphism.  Thus, it's extremely
     * unlikely to reduce the quality of the hash function (in terms of
     * "bucket dispersion").
     */
    tmp = (uint32_t)DeRef(env, obj) ^ rand1;
    return ((tmp >> 17) | (tmp << (8*sizeof(uint32_t)-17))) ^ rand2; 
}

JNIEXPORT void JNICALL
JVM_MonitorWait(JNIEnv *env, jobject obj, jlong millis)
{
    if (ll_ltz(millis)) {
	ThrowIllegalArgumentException(0, "timeout value is negative");
        return;
    }

#ifdef HPROF
    jvmpi_monitor_wait(DeRef(env, obj), millis);
#endif

    if (ll_eqz(millis)) {
        millis = TIMEOUT_INFINITY;
    }
    monitorWait2(JNIEnv2EE(env), obj_monitor(DeRef(env, obj)), millis);

#ifdef HPROF
    jvmpi_monitor_waited(DeRef(env, obj));
#endif
}

JNIEXPORT void JNICALL
JVM_MonitorNotify(JNIEnv *env, jobject obj)
{
    monitorNotify2(JNIEnv2EE(env), obj_monitor(DeRef(env, obj)));
}

JNIEXPORT void JNICALL
JVM_MonitorNotifyAll(JNIEnv *env, jobject obj)
{
    monitorNotifyAll2(JNIEnv2EE(env), obj_monitor(DeRef(env, obj)));
}

JNIEXPORT jobject JNICALL
JVM_Clone(JNIEnv *env, jobject obj)
{
    HObject *this = DeRef(env, obj);
    ExecEnv * ee = JNIEnv2EE(env);
    if (obj_flags(this) == T_NORMAL_OBJECT) {
	ClassClass *cb = obj_classblock(this);
	HObject *clone;
	if (!ImplementsInterface(cb, interfaceJavaLangCloneable, ee)) {
	    char clname[MSG_BUF_LEN];
	    classname2string(cbName(cb), clname, sizeof(clname));
	    ThrowCloneNotSupportedException(ee, clname);
	    return NULL;
	}
	clone = allocObject(ee, cb);
	if (clone == 0) {
	    char buf[MSG_BUF_LEN];
	    ThrowOutOfMemoryError(ee,
		classname2string(cbName(cb), buf, sizeof(buf)));
	    return NULL;
	}
	{
	    char *to = (char*)unhand(clone);
	    char *from = (char*)unhand(this);
	    memcpy(to, from, cbInstanceSize(cb));
	    KEEP_POINTER_ALIVE(to);
	    KEEP_POINTER_ALIVE(from);
	}
	return MkRefLocal(env, clone);
    } else { 
	int type = obj_flags(this);
	size_t size = obj_length(this);
	size_t bytes = 
	       sizearray(type, size) + (type == T_CLASS ? sizeof(OBJECT) : 0);
	HObject *clone = allocArray(ee, type, size);
	if (clone == 0) {
	    ThrowOutOfMemoryError(0, 0);
	    return NULL;
	}
	{
	    if (type == T_CLASS) {
		/* We can not memcpy array of handles because byte wise
		 * copying might make the garbage collector see bad handles.
		 */
		int i;
		ArrayOfObject *srcarray = (ArrayOfObject *)unhand(this);
		ArrayOfObject *dstarray = (ArrayOfObject *)unhand(clone);
		HObject **srcptr = &srcarray->body[0];
		HObject **dstptr = &dstarray->body[0];
		for (i = 0; i < size + 1; i++) {
		    *dstptr++ = *srcptr++;
		}
		KEEP_POINTER_ALIVE(srcarray);
		KEEP_POINTER_ALIVE(dstarray);
	    } else {
		char *to = (char*)unhand(clone);
		char *from = (char*)unhand(this);
		memcpy(to, from, bytes);
		KEEP_POINTER_ALIVE(to);
		KEEP_POINTER_ALIVE(from);
	    }
	}
	return MkRefLocal(env, clone);
    }
}

/* IO-related JVM calls */

/*
 * File I/O
 */

/* We should add JVM-specific error handling as needed */
JNIEXPORT jint JNICALL
JVM_Open(const char *fname, jint flags, jint mode) {
    int result = sysOpen(fname, flags, mode);
    if (result >= 0) {
	return result;
    } else {
	switch(errno) {
	case EEXIST:
	    return JVM_EEXIST;
	default:
	    return -1;
	}
    }
}

JNIEXPORT jint JNICALL
JVM_Close(jint fd) {
    return sysClose(fd);
}

JNIEXPORT jint JNICALL
JVM_Read(jint fd, char *buf, jint nbytes) {
    return sysRead(fd, buf, nbytes);
}

JNIEXPORT jint JNICALL
JVM_Write(jint fd, char *buf, jint nbytes) {
    return sysWrite(fd, buf, nbytes);
}

JNIEXPORT jint JNICALL
JVM_Available(jint fd, jlong *pbytes) {
    return sysAvailable(fd, (int64_t *)pbytes);
}

JNIEXPORT jlong JNICALL
JVM_Lseek(jint fd, jlong where, jint whence) {
    return sysSeek(fd, where, whence);
}

JNIEXPORT jint JNICALL
JVM_SetLength(jint fd, jlong length) {
    return sysSetLength(fd, length);
}

JNIEXPORT jint JNICALL
JVM_Sync(jint fd) {
    int ret = sysSync(fd);
    if (ret < 0) {
        SignalError(0, "java/io/SyncFailedException", 0);
    }
    return ret;
}


/*
 * 
 */

JNIEXPORT const char * JNICALL
JVM_GetClassNameUTF(JNIEnv *env, jclass c)
{
    ClassClass *cb = (ClassClass *)DeRef(env, c);
    return cbName(cb);
}

JNIEXPORT void JNICALL
JVM_GetClassCPTypes(JNIEnv *env, jclass c, unsigned char *res_types)
{
    ClassClass *cb = (ClassClass *)DeRef(env, c);
    int i;
    union cp_item_type *constant_pool = cbConstantPool(cb);
    unsigned char * types= constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
    for (i = cbConstantPoolCount(cb) - 1; i >= 0; i--) {
	res_types[i] = CONSTANT_POOL_TYPE_TABLE_GET_TYPE(types, i);
    }
}

JNIEXPORT jint JNICALL
JVM_GetClassCPEntriesCount(JNIEnv *env, jclass c)
{
    ClassClass *cb = (ClassClass *)DeRef(env, c);
    return cbConstantPoolCount(cb);
}

JNIEXPORT jint JNICALL
JVM_GetClassFieldsCount(JNIEnv *env, jclass c)
{
    ClassClass *cb = (ClassClass *)DeRef(env, c);
    return cbFieldsCount(cb);
}

JNIEXPORT jint JNICALL
JVM_GetClassMethodsCount(JNIEnv *env, jclass c)
{
    ClassClass *cb = (ClassClass *)DeRef(env, c);
    return cbMethodsCount(cb);
}

JNIEXPORT void JNICALL
JVM_GetMethodIxExceptionIndexes(JNIEnv *env, jclass c, jint method_index,
				unsigned short *exceptions)
{
    ClassClass *cb = (ClassClass *)DeRef(env, c);
    struct methodblock *mb = cbMethods(cb) + method_index;
    memcpy(exceptions, mb->exceptions, 
	   mb->nexceptions * sizeof(unsigned short));
}

JNIEXPORT jint JNICALL
JVM_GetMethodIxExceptionsCount(JNIEnv *env, jclass c, jint method_index)
{
    ClassClass *cb = (ClassClass *)DeRef(env, c);
    struct methodblock *mb = cbMethods(cb) + method_index;
    return mb->nexceptions;
}

JNIEXPORT void JNICALL
JVM_GetMethodIxByteCode(JNIEnv *env, jclass c, jint method_index, 
		        unsigned char *code)
{
    ClassClass *cb = (ClassClass *)DeRef(env, c);
    struct methodblock *mb = cbMethods(cb) + method_index;
    memcpy(code, mb->code, mb->code_length);
}

JNIEXPORT jint JNICALL
JVM_GetMethodIxByteCodeLength(JNIEnv *env, jclass c, jint method_index)
{
    ClassClass *cb = (ClassClass *)DeRef(env, c);
    struct methodblock *mb = cbMethods(cb) + method_index;
    return mb->code_length;
}

JNIEXPORT void JNICALL
JVM_GetMethodIxExceptionTableEntry(JNIEnv *env, jclass c, jint method_index,
				   jint entry_index,
				   JVM_ExceptionTableEntryType *entry)
{
    ClassClass *cb = (ClassClass *)DeRef(env, c);
    struct methodblock *mb = cbMethods(cb) + method_index;
    struct CatchFrame *exception_table = mb->exception_table + entry_index;

    sysAssert(method_index >= 0 && method_index < cbMethodsCount(cb));
    sysAssert(entry_index >= 0 && entry_index < mb->exception_table_length);

    if (method_index >= 0 && method_index < cbMethodsCount(cb) &&
        entry_index >= 0 && entry_index < mb->exception_table_length) {
    	entry->start_pc = exception_table->start_pc;
	entry->end_pc = exception_table->end_pc;
	entry->handler_pc = exception_table->handler_pc;
	entry->catchType = exception_table->catchType;
    }
}

JNIEXPORT jint JNICALL
JVM_GetFieldIxModifiers(JNIEnv *env, jclass c, int index)
{
    ClassClass *cb = (ClassClass *)DeRef(env, c);
    struct fieldblock *fb = cbFields(cb) + index;
    return fb->access & ACC_WRITTEN_FLAGS;
}

JNIEXPORT jint JNICALL
JVM_GetMethodIxModifiers(JNIEnv *env, jclass c, int index)
{
    ClassClass *cb = (ClassClass *)DeRef(env, c);
    struct methodblock *mb = cbMethods(cb) + index;
    return mb->fb.access & ACC_WRITTEN_FLAGS;
}

JNIEXPORT jint JNICALL
JVM_GetMethodIxExceptionTableLength(JNIEnv *env, jclass c, int index)
{
    ClassClass *cb = (ClassClass *)DeRef(env, c);
    struct methodblock *mb = cbMethods(cb) + index;
    return mb->exception_table_length;
}

JNIEXPORT jint JNICALL
JVM_GetMethodIxLocalsCount(JNIEnv *env, jclass c, int index)
{
    ClassClass *cb = (ClassClass *)DeRef(env, c);
    struct methodblock *mb = cbMethods(cb) + index;
    return mb->nlocals;
}

JNIEXPORT jint JNICALL
JVM_GetMethodIxArgsSize(JNIEnv *env, jclass c, int index)
{
    ClassClass *cb = (ClassClass *)DeRef(env, c);
    struct methodblock *mb = cbMethods(cb) + index;
    return mb->args_size;
}

JNIEXPORT jint JNICALL
JVM_GetMethodIxMaxStack(JNIEnv *env, jclass c, int index)
{
    ClassClass *cb = (ClassClass *)DeRef(env, c);
    struct methodblock *mb = cbMethods(cb) + index;
    return mb->maxstack;
}

JNIEXPORT jboolean JNICALL
JVM_IsConstructorIx(JNIEnv *env, jclass c, int index)
{
    ClassClass *cb = (ClassClass *)DeRef(env, c);
    struct methodblock *mb = cbMethods(cb) + index;
    return mb->fb.name == utf8_literal_obj_init_name;
}

JNIEXPORT const char * JNICALL
JVM_GetMethodIxNameUTF(JNIEnv *env, jclass c, jint index)
{
    ClassClass *cb = (ClassClass *)DeRef(env, c);
    struct methodblock *mb = cbMethods(cb) + index;
    return mb->fb.name;
}

JNIEXPORT const char * JNICALL
JVM_GetMethodIxSignatureUTF(JNIEnv *env, jclass c, jint index)
{
    ClassClass *cb = (ClassClass *)DeRef(env, c);
    struct methodblock *mb = cbMethods(cb) + index;
    return mb->fb.signature;
}

JNIEXPORT const char * JNICALL
JVM_GetCPFieldNameUTF(JNIEnv *env, jclass c, jint cp_index)
{
    ClassClass *cb = (ClassClass *)DeRef(env, c);
    union cp_item_type *cp = cbConstantPool(cb);
    unsigned char *type_table = cp[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
    switch(type_table[cp_index]) {
    case CONSTANT_Fieldref | CONSTANT_POOL_ENTRY_RESOLVED:
        return cp[cp_index].fb->name;
    case CONSTANT_Fieldref: {  
        int index = cp[cp_index].i; /* value of Fieldref field */
	int key2 = index & 0xFFFF;  /* index to NameAndType  */
	int name_index = cp[key2].i >> 16;
	return cp[name_index].cp;
    }
    default:
        (*env)->FatalError(env, "JVM_GetCPFieldNameUTF: illegal constant");
	return NULL;		/* Keep lint happy */
    }
}

JNIEXPORT const char * JNICALL
JVM_GetCPMethodNameUTF(JNIEnv *env, jclass c, jint cp_index)
{
    ClassClass *cb = (ClassClass *)DeRef(env, c);
    union cp_item_type *cp = cbConstantPool(cb);
    unsigned char *type_table = cp[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
    switch(type_table[cp_index]) {
    case CONSTANT_InterfaceMethodref | CONSTANT_POOL_ENTRY_RESOLVED:
    case CONSTANT_Methodref | CONSTANT_POOL_ENTRY_RESOLVED:
        return cp[cp_index].mb->fb.name;
    case CONSTANT_InterfaceMethodref:
    case CONSTANT_Methodref: {
        int index = cp[cp_index].i;
	int key2 = index & 0xFFFF;  /* index to NameAndType  */
	int name_index = cp[key2].i >> 16;
	return cp[name_index].cp;
    }
    default:
        (*env)->FatalError(env, "JVM_GetCPMethodNameUTF: illegal constant");
	return NULL;		/* Keep lint happy */
    }
}

JNIEXPORT const char * JNICALL
JVM_GetCPFieldSignatureUTF(JNIEnv *env, jclass c, jint cp_index)
{
    ClassClass *cb = (ClassClass *)DeRef(env, c);
    union cp_item_type *cp = cbConstantPool(cb);
    unsigned char *type_table = cp[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
    switch(type_table[cp_index]) {
    case CONSTANT_Fieldref | CONSTANT_POOL_ENTRY_RESOLVED:
        return cp[cp_index].fb->signature;
    case CONSTANT_Fieldref: {  
        int index = cp[cp_index].i; /* value of Fieldref field */
	int key2 = index & 0xFFFF;  /* index to NameAndType  */
	int signature_index = cp[key2].i & 0xFFFF;
	return cp[signature_index].cp;
    }
    default:
        (*env)->FatalError(env, "JVM_GetCPFieldSignatureUTF: illegal constant");
	return NULL;		/* Keep lint happy */
    }
}

JNIEXPORT const char * JNICALL
JVM_GetCPMethodSignatureUTF(JNIEnv *env, jclass c, jint cp_index)
{
    ClassClass *cb = (ClassClass *)DeRef(env, c);
    union cp_item_type *cp = cbConstantPool(cb);
    unsigned char *type_table = cp[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
    switch(type_table[cp_index]) {
    case CONSTANT_InterfaceMethodref | CONSTANT_POOL_ENTRY_RESOLVED:
    case CONSTANT_Methodref | CONSTANT_POOL_ENTRY_RESOLVED:
        return cp[cp_index].mb->fb.signature;
    case CONSTANT_InterfaceMethodref:
    case CONSTANT_Methodref: {
        int index = cp[cp_index].i;
	int key2 = index & 0xFFFF;  /* index to NameAndType  */
	int signature_index = cp[key2].i & 0xFFFF;
	return cp[signature_index].cp;
    }
    default:
        (*env)->FatalError(env, "JVM_GetCPMethodSignatureUTF: illegal constant");
	return NULL;		/* Keep lint happy */
    }
}

JNIEXPORT const char * JNICALL
JVM_GetCPClassNameUTF(JNIEnv *env, jclass c, jint cp_index)
{
    ClassClass *cb = (ClassClass *)DeRef(env, c);
    union cp_item_type *cp = cbConstantPool(cb);
    return GetClassConstantClassName(cp, cp_index);
}

JNIEXPORT const char * JNICALL
JVM_GetCPFieldClassNameUTF(JNIEnv *env, jclass c, jint cp_index)
{
    ClassClass *cb = (ClassClass *)DeRef(env, c);
    union cp_item_type *cp = cbConstantPool(cb);
    unsigned char *type_table = cp[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
    switch(type_table[cp_index]) {
    case CONSTANT_Fieldref | CONSTANT_POOL_ENTRY_RESOLVED:
        return cbName(cp[cp_index].fb->clazz);
    case CONSTANT_Fieldref: {  
        unsigned classkey = cp[cp_index].i >> 16;
	return GetClassConstantClassName(cp, classkey);
    }
    default:
        (*env)->FatalError(env, "JVM_GetCPFieldClassNameUTF: illegal constant");
	return NULL;		/* Keep lint happy */
    }
}

JNIEXPORT const char * JNICALL
JVM_GetCPMethodClassNameUTF(JNIEnv *env, jclass c, jint cp_index)
{
    ClassClass *cb = (ClassClass *)DeRef(env, c);
    union cp_item_type *cp = cbConstantPool(cb);
    unsigned char *type_table = cp[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
    switch(type_table[cp_index]) {
    case CONSTANT_InterfaceMethodref | CONSTANT_POOL_ENTRY_RESOLVED:
    case CONSTANT_Methodref | CONSTANT_POOL_ENTRY_RESOLVED:
        return cbName(cp[cp_index].mb->fb.clazz);
    case CONSTANT_InterfaceMethodref:
    case CONSTANT_Methodref: {
        unsigned classkey = cp[cp_index].i >> 16;
	return GetClassConstantClassName(cp, classkey);
    }
    default:
        (*env)->FatalError(env, "JVM_GetCPMethodClassNameUTF: illegal constant");
	return NULL;		/* Keep lint happy */
    }
}

JNIEXPORT jint JNICALL
JVM_GetCPFieldModifiers(JNIEnv *env, jclass c, int cp_index, 
			jclass calledClass)
{
    ClassClass *cb = (ClassClass *)DeRef(env, c);
    ClassClass *cbCalled = (ClassClass *)DeRef(env, calledClass);
    union cp_item_type *cp = cbConstantPool(cb);
    unsigned char *type_table = cp[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
    switch(type_table[cp_index]) {
    case CONSTANT_Fieldref | CONSTANT_POOL_ENTRY_RESOLVED:
        return cp[cp_index].fb->access & ACC_WRITTEN_FLAGS;
    case CONSTANT_Fieldref: {  
        int index = cp[cp_index].i; /* value of Fieldref field */
	int key2 = index & 0xFFFF;  /* index to NameAndType  */
	int signature_index = cp[key2].i & 0xFFFF;
	int name_index = cp[key2].i >> 16;
	struct fieldblock *fb;
	int n = cbFieldsCount(cbCalled);

	fb = cbFields(cbCalled);
	for (; --n >= 0; fb++) {
	    if ((fb->name == cp[name_index].cp) &&
		(fb->signature == cp[signature_index].cp)) {
	        return fb->access & ACC_WRITTEN_FLAGS;
	    }
	}
	return -1;
    }
    default:
        (*env)->FatalError(env, "JVM_GetCPFieldModifiers: illegal constant");
    }
    return 0; /* never reached */
}
    
JNIEXPORT jint JNICALL
JVM_GetCPMethodModifiers(JNIEnv *env, jclass c, int cp_index,
			 jclass calledClass)
{
    ClassClass *cb = (ClassClass *)DeRef(env, c);
    ClassClass *cbCalled = (ClassClass *)DeRef(env, calledClass);
    union cp_item_type *cp = cbConstantPool(cb);
    unsigned char *type_table = cp[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
    switch(type_table[cp_index]) {
    case CONSTANT_InterfaceMethodref | CONSTANT_POOL_ENTRY_RESOLVED:
    case CONSTANT_Methodref | CONSTANT_POOL_ENTRY_RESOLVED:
        return cp[cp_index].mb->fb.access & ACC_WRITTEN_FLAGS;
    case CONSTANT_InterfaceMethodref:
    case CONSTANT_Methodref: {
        int index = cp[cp_index].i; /* value of Fieldref field */
	int key2 = index & 0xFFFF;  /* index to NameAndType  */
	int signature_index = cp[key2].i & 0xFFFF;
	int name_index = cp[key2].i >> 16;
	struct methodblock *mb;
	int n = cbMethodsCount(cbCalled);
	mb = cbMethods(cbCalled);

	for (; --n >= 0; mb++) {
	    if ((mb->fb.name == cp[name_index].cp) &&
		(mb->fb.signature == cp[signature_index].cp)) {
	    	return mb->fb.access & ACC_WRITTEN_FLAGS;
	    }
	}
	return -1;
    }
    default:
        (*env)->FatalError(env, "JVM_GetCPMethodModifiers: illegal constant");
    }
    return 0; /* never reached */
}

JNIEXPORT void JNICALL
JVM_ReleaseUTF(const char *utf)
{
}

JNIEXPORT jboolean JNICALL
JVM_IsSameClassPackage(JNIEnv *env, jclass class1, jclass class2)
{
    ClassClass *cb1 = (ClassClass *)DeRef(env, class1);
    ClassClass *cb2 = (ClassClass *)DeRef(env, class2);
    return IsSameClassPackage(cb1, cb2);
}

/*
 * Thread class calls
 */

/* this is actually called by 
   AccessController.getInheritedAccessControlContext */

JNIEXPORT jobject JNICALL
JVM_GetInheritedAccessControlContext(JNIEnv *env, jclass cls)
{
    Hjava_lang_Thread *hp = (Hjava_lang_Thread *) JNIEnv2EE(env)->thread;
    return MkRefLocal(env, THREAD(hp)->inheritedAccessControlContext);
}


static void    /* first routine called in the new thread */
ThreadRT0(Hjava_lang_Thread *p)
{
    ExecEnv *ee = EE();

#ifdef TRACING
    if (trace) {
	jio_fprintf(stderr, "Started thread %p\n", p);
    }
#endif

    if (!unhand(p)->stillborn) {
	/* Otherwise this thread was asked to exit before it got to run */
        execute_java_dynamic_method(ee, (void *)p, "run", "()V");
    }
}

JNIEXPORT void JNICALL
JVM_StartThread(JNIEnv *env, jobject this)
{
    Hjava_lang_Thread *p = (Hjava_lang_Thread *)DeRef(env, this);
    struct Classjava_lang_Thread *tid;
    ExecEnv *ee;

    tid = THREAD(p);
    ee = (ExecEnv *)ll2ptr(tid->eetop);

    /* It is illegal to call Start on an active thread */
    if (ee) {
	ThrowIllegalThreadStateException(0, 0);
    } else {
	/*
	 * Note that a nonzero return value lumps together SYS_NOMEM and
	 * SYS_NORESOURCE and does not reflect any other error conditions.
	 */
	if (threadCreate(p, ProcStackSize, tid->priority,
			 (void (*)(void *)) ThreadRT0)) {
	    ThrowOutOfMemoryError(0, 0);
	}
    }
}

/*
 * We normally stop a thread by posting a ThreadDeath exception against
 * it.  When it handles the exception, the thread cooperatively exits.
 * However, if a thread hasn't started running yet, it is not prepared
 * to handle an exception.  That case motivated the use of a "stillborn"
 * flag that can be safely set before the thread starts running.  When
 * the thread does run, if it is marked to be stillborn, it immediately
 * exits.  However, the mechanism can also be used to fix an old bug
 * and handle other boundary cases.  If we *always* set tid->stillborn
 * to 1 when stop is called on a thread, this ensures that a stopped
 * thread can't be resurrected -- attempting to restart it will cause
 * it to exit as stillborn.  We only call threadPostException() if
 * tid->PrivateInfo is set just in case it ever assumes system-specific
 * thread structure is there.
 */
JNIEXPORT void JNICALL
JVM_StopThread(JNIEnv *env, jobject this, jobject exception)
{
    Hjava_lang_Thread *p = (Hjava_lang_Thread *)DeRef(env, this);
    struct Hjava_lang_Object *exc = (Hjava_lang_Object *)DeRef(env, exception);
    struct Classjava_lang_Thread *tid;
    ExecEnv *ee;

    if (exc == 0) {
	ThrowNullPointerException(JNIEnv2EE(env), 0);
	return;
    }

    tid = THREAD(p);

    QUEUE_LOCK(EE2SysThread(JNIEnv2EE(env)));

    ee = (ExecEnv *)ll2ptr(tid->eetop);
    if (ee && !tid->stillborn) {
        if (ee->async_disable_count) {
	    ee->pending_async_exc = exc;
	} else {
	    if (is_instance_of(exc, classJavaLangThreadDeath, ee)) {
	        tid->stillborn = 1;
	    }
	    threadPostException(p, exc);
	}
    }

    QUEUE_UNLOCK(EE2SysThread(JNIEnv2EE(env)));
}

/*
 * Note that we don't go to the native thread to determine whether the Java
 * thread is alive.  The status of the native thread doesn't matter, to the
 * first approximation at least, if the Java thread object isn't set up
 * properly.  However, we could also double-check the liveness of the native
 * thread.
 */
JNIEXPORT jboolean JNICALL
JVM_IsThreadAlive(JNIEnv *env, jobject this)
{
    struct Classjava_lang_Thread *tid;

    tid = THREAD(DeRef(env, this));
    return (jint)(ll_nez(tid->eetop) && !tid->stillborn);
}

JNIEXPORT void JNICALL
JVM_SuspendThread(JNIEnv *env, jobject this)
{
    Hjava_lang_Thread *p = (Hjava_lang_Thread *)DeRef(env, this);

    if (ll_nez(THREAD(p)->eetop)) {
	(void) threadSuspend(p);
    }
}

JNIEXPORT void JNICALL
JVM_ResumeThread(JNIEnv *env, jobject this)
{
    Hjava_lang_Thread *p = (Hjava_lang_Thread *)DeRef(env, this);

    if (ll_nez(THREAD(p)->eetop)) {
	(void) threadResume(p);
    }
}

/* We want to rely on priorities as set/got via the threadX functions.
 * However, it is legal to call setPriority() and getPriority() after
 * 'new' and before start() is called, in which case the thread's TID has
 * been allocated but its thread not yet created.  So we also maintain
 * tid->priority to bridge the gap.
 *
 * The priority should already have been check to fall within a
 * valid range by the software above.  All we do is pass it on to
 * the system level.  The JAVA level also set the priority field.
 */
JNIEXPORT void JNICALL
JVM_SetThreadPriority(JNIEnv *env, jobject this, jint prio)
{
    Hjava_lang_Thread *p = (Hjava_lang_Thread *)DeRef(env, this);

    if (ll_nez(THREAD(p)->eetop)) {
	threadSetPriority(p, prio);
    } 
}

JNIEXPORT void JNICALL
JVM_Yield(JNIEnv *env, jclass threadClass)
{
    threadYield();
}

JNIEXPORT void JNICALL
JVM_Sleep(JNIEnv *env, jclass threadClass, jlong millis)
{
    if (ll_ltz(millis)) {
	ThrowIllegalArgumentException(JNIEnv2EE(env),
	    "timeout value is negative");
	return;
    }

#ifdef HPROF
    jvmpi_monitor_wait(NULL, millis);
#endif

    threadSleep(millis);

#ifdef HPROF
    jvmpi_monitor_waited(NULL);
#endif

    return;
}

JNIEXPORT jobject JNICALL
JVM_CurrentThread(JNIEnv *env, jclass threadClass)
{
    ExecEnv *ee = JNIEnv2EE(env);

    return MkRefLocal(env, ee->thread);
}

/* This doesn't count methods implemented in C. */
JNIEXPORT jint JNICALL
JVM_CountStackFrames(JNIEnv *env, jobject this)
{
    Hjava_lang_Thread *hp = (Hjava_lang_Thread *)DeRef(env, this);

    if (hp == 0) {
	ThrowNullPointerException(0, 0);
	return -1;
    } else {
	Classjava_lang_Thread *p = unhand(hp);
	ExecEnv *this_ee = (ExecEnv *)ll2ptr(p->eetop);
	JavaFrame *frame, frame_buf;
	jint count;

	if (this_ee == 0) {
	    /* Thread hasn't any state yet. */
	    return 0;
	}
	frame = this_ee->current_frame;
	for (count = 0; frame != 0 ; ) {
            if ((frame->current_method != 0) &&
		!(frame->current_method->fb.access & ACC_NATIVE)) {
		count++;
	    }
	    frame = FRAME_PREV(frame, &frame_buf);
	}
	return count;
    }
}

JNIEXPORT void JNICALL
JVM_Interrupt(JNIEnv *env, jobject this)
{
    Hjava_lang_Thread *p = (Hjava_lang_Thread *)DeRef(env, this);
    ExecEnv *ee;

    QUEUE_LOCK(EE2SysThread(JNIEnv2EE(env)));

    ee = (ExecEnv *)ll2ptr(THREAD(p)->eetop);
    if (ee) {
        if (ee->async_disable_count) {
	    ee->interrupt_pending = TRUE;
	} else {
	    threadInterrupt(p);
	}
    }

    QUEUE_UNLOCK(EE2SysThread(JNIEnv2EE(env)));
}

JNIEXPORT jboolean JNICALL
JVM_IsInterrupted(JNIEnv *env, jobject this, jboolean clearInterrupted)
{
    Hjava_lang_Thread *p = (Hjava_lang_Thread *)DeRef(env, this);

    if (ll_nez(THREAD(p)->eetop)) {
	return (jboolean) threadIsInterrupted(p, clearInterrupted);
    } else {
	return JNI_FALSE;
    }
}

/*
 * SecurityManager support
 */

static struct Hjava_lang_ClassLoader *sys_loader = NULL;

bool_t InitializeSystemClassLoader(void)
{
    sys_loader = (struct Hjava_lang_ClassLoader *)
        execute_java_static_method(EE(), classJavaLangClassLoader,
				   "getSystemClassLoader",
				   "()Ljava/lang/ClassLoader;");
    return !exceptionOccurred(EE());
}

bool_t IsTrustedClassLoader(struct Hjava_lang_ClassLoader *loader)
{
    struct Hjava_lang_ClassLoader *cl;

    for (cl = sys_loader; cl; cl = unhand(cl)->parent) {
        if (cl == loader) {
	    return TRUE;
	}
    }
    return FALSE;
}

static bool_t
is_trusted_frame(JNIEnv *env, struct javaframe *frame)
{
    ExecEnv *ee = JNIEnv2EE(env);
    ClassClass *cb;
    if (ee->privileged && (ee->privileged->frame_id == GetFrameID(frame))) {
	cb = ee->privileged->clazz;
	if (cb && (!cbLoader(cb) || IsTrustedClassLoader(cbLoader(cb)))) {
	    return TRUE;
	}
    }
    return FALSE;
}

JNIEXPORT jclass JNICALL
JVM_CurrentLoadedClass(JNIEnv *env)
{
    struct javaframe *frame, frame_buf;
    ClassClass *cb;

    for (frame = EE()->current_frame ; frame != 0 ;) {

	/* if a method in a class in a trusted loader is in a doPrivileged,
	   return NULL */

	if (is_trusted_frame(env, frame))
	    return NULL;

	if ((frame->current_method != 0) &&
	    !(frame->current_method->fb.access & ACC_NATIVE)) {
	    cb = fieldclass(&frame->current_method->fb);
	    if (cb && cbLoader(cb) && !IsTrustedClassLoader(cbLoader(cb))) {
		return MkRefLocal(env, cbHandle(cb));
	    }
        }
	frame = FRAME_PREV(frame, &frame_buf);
    }
    return NULL;
}

JNIEXPORT jobject JNICALL
JVM_CurrentClassLoader(JNIEnv *env)
{
    struct javaframe *frame, frame_buf;
    ClassClass	*cb;

    for (frame = EE()->current_frame ; frame != 0 ;) {
	/* if a method in a class in a trusted loader is in a doPrivileged,
	   return NULL */

	if (is_trusted_frame(env, frame))
	    return NULL;

	if ((frame->current_method != 0) &&
	    !(frame->current_method->fb.access & ACC_NATIVE)) {
	    cb = fieldclass(&frame->current_method->fb);
	    if (cb && cbLoader(cb) && !IsTrustedClassLoader(cbLoader(cb))) {
	        return MkRefLocal(env, cbLoader(cb));
	    }
        }
	frame = FRAME_PREV(frame, &frame_buf);
    }
    return NULL;
}

JNIEXPORT jobjectArray JNICALL
JVM_GetClassContext(JNIEnv *env)
{
    struct javaframe *frame, frame_buf;
    HArrayOfObject *ctx;
    HObject **ctxdata;
    int n = 0;

    /* count the classes on the execution stack */
    for (frame = JNIEnv2EE(env)->current_frame ; frame != 0 ;) {
	if ((frame->current_method != 0) &&
	    !(frame->current_method->fb.access & ACC_NATIVE)) {
	    n++;
        }
	frame = FRAME_PREV(frame, &frame_buf);
    }

    /* allocate the array */
    ctx = (HArrayOfObject *)allocArray(JNIEnv2EE(env), T_CLASS, n);
    if (ctx == 0) {
	ThrowOutOfMemoryError(JNIEnv2EE(env), 0);
	return 0;
    }
    ctxdata = (HObject **)unhand(ctx)->body;
    ctxdata[n] = (HObject *)classJavaLangClass;

    /* fill the array */
    for (n = 0, frame = JNIEnv2EE(env)->current_frame ; frame != 0 ;) {
	if ((frame->current_method != 0) &&
	    !(frame->current_method->fb.access & ACC_NATIVE)) {
	    ctxdata[n++] = (HObject *)(fieldclass(&frame->current_method->fb));
        }
	frame = FRAME_PREV(frame, &frame_buf);
    }
    KEEP_POINTER_ALIVE(ctxdata);

    return MkRefLocal(env, ctx);
}

/* REMIND: The name string could probably be handled better... */
JNIEXPORT jint JNICALL
JVM_ClassDepth(JNIEnv *env, jstring name)
{
    struct javaframe *frame, frame_buf;
    ClassClass	*cb;
    char buf[STK_BUF_LEN], *p;
    jint depth = 0;

    javaString2CString((Hjava_lang_String *)DeRef(env, name), buf, sizeof(buf));
    for (p = buf ; *p ; p++) {
	if (*p == '.') {
	    *p = '/';
	}
    }

    for (frame = JNIEnv2EE(env)->current_frame ; frame != 0 ; ) {
	if ((frame->current_method != 0) &&
	    !(frame->current_method->fb.access & ACC_NATIVE)) {
	    cb = fieldclass(&frame->current_method->fb);
	    if (cb && !strcmp(cbName(cb), buf)) {
		return depth;
	    }
	    depth++;
	}
	frame = FRAME_PREV(frame, &frame_buf);
    }
    return -1;
}

JNIEXPORT jint JNICALL
JVM_ClassLoaderDepth(JNIEnv *env)
{
    struct javaframe *frame, frame_buf;
    ClassClass	*cb;
    jint depth = 0;

    for (frame = JNIEnv2EE(env)->current_frame ; frame != 0 ; ) {
	/* if a method in a class in a trusted loader is in a doPrivileged,
	   return -1 */

	if (is_trusted_frame(env, frame))
	    return -1;

	if ((frame->current_method != 0) && 
	    !(frame->current_method->fb.access & ACC_NATIVE)) {
	    cb = fieldclass(&frame->current_method->fb);
	    if (cb && cbLoader(cb) && !IsTrustedClassLoader(cbLoader(cb))) {
		return depth;
	    }
            depth++;
        }
	frame = FRAME_PREV(frame, &frame_buf);
    }
    return -1;
}

/*
 * The following functions from ObjectInputStream.c
 * cannot be done with JNI or reflection
 */

#define OBJECT_PACKAGE "java/io/"


/* Copied VerifyFieldAccess to avoid verifyclasses turinging it off.
 */
static bool_t
ForceVerifyFieldAccess(ClassClass *current_class,
		       ClassClass *field_class,
		       int access, bool_t classloader_only) {
    if (current_class == 0) {
	return TRUE;
    }
    if ((current_class == field_class) || IsPublic(access))
	return TRUE;

    if (IsProtected(access)) {
	/* See if current_class is a subclass of field_class */
	ClassClass *cb;
	for (cb = cbSuperclass(current_class); cb; cb = cbSuperclass(cb)) {
	    if (cb == field_class) 
		return TRUE;
	}
    }
    return (!IsPrivate(access)
	    && IsSameClassPackage(current_class, field_class));
}


/*
 * Allocate an initialized object of the specified class
 */
JNIEXPORT jobject JNICALL
JVM_AllocateNewObject(JNIEnv *env, jobject this, jclass currClass,
		      jclass initClass)
{
    struct Hjava_lang_Class *cb = 
		(Hjava_lang_Class *)DeRef(env, currClass);
    struct Hjava_lang_Class *initcb = 
		(Hjava_lang_Class *)DeRef(env, initClass);

    Hjava_lang_Object* obj;
    ExecEnv *ee = JNIEnv2EE(env);
    struct methodblock *mb;
    int len;

    if (cbAccess(cb) & (ACC_INTERFACE | ACC_ABSTRACT)) { 
	ThrowInstantiationException(0, cbName(cb));
	return NULL;
    }

    for (len = cbMethodsCount(initcb), mb = cbMethods(initcb); --len >= 0; mb++) 
	if (mb->fb.name == utf8_literal_obj_init_name &&
	    mb->fb.signature == utf8_literal_null_init_sig)
	    break;
    if (len < 0) { 
	ThrowNoSuchMethodError(ee, utf8_literal_obj_init_name);
	return NULL;
    }

    if (cb == initcb && !IsPublic(mb->fb.access)) {

      /* Calling the constructor for class 'cb'. 
       * Only allow calls to a public no-arg constructor.
       * This path corresponds to creating an Externalizable object.
       */
      ThrowIllegalAccessException(ee, mbName(mb));
      return NULL;
    }
   else if (!ForceVerifyFieldAccess(cb, initcb, mb->fb.access, FALSE)) {
      /* subclass 'cb' does not have access to no-arg constructor of 'initcb'.*/
      ThrowIllegalAccessException(ee, mbName(mb));
      return NULL;
    }

    /* Allocate the object  */
    if ((obj = allocObject(ee, cb)) == 0) { 
	ThrowOutOfMemoryError(ee, 0);	
        return NULL;
    }

    /* Invoke the no-arg constructor */
    do_execute_java_method(ee, obj, 0, 0, mb, FALSE, NULL);
    
    return MkRefLocal(env, obj);
}

/*
 * Allocate a new array of the specified length.
 * Do not initialize the cells.
 */
JNIEXPORT jobject JNICALL
JVM_AllocateNewArray(JNIEnv *env, jobject this, jclass currClass,
		     jint length)
{

    ClassClass* cb = (ClassClass *)DeRef(env, currClass);
    JHandle *obj = NULL;
    ExecEnv *ee = JNIEnv2EE(env);

    if (cbName(cb)[0] != '[') {
	SignalError(0, OBJECT_PACKAGE "InvalidClassException", cbName(cb));
	return NULL;
    }

    switch (cbName(cb)[1]) {
    case SIGNATURE_BYTE:
	obj = allocArray(ee, T_BYTE, length);
	break;

    case SIGNATURE_CHAR:
	obj = allocArray(ee, T_CHAR, length);
	break;

    case SIGNATURE_FLOAT:
	obj = allocArray(ee, T_FLOAT, length);
	break;

    case SIGNATURE_DOUBLE:
	obj = allocArray(ee, T_DOUBLE, length);
	break;

    case SIGNATURE_INT:
	obj = allocArray(ee, T_INT, length);
	break;

    case SIGNATURE_LONG:
	obj = allocArray(ee, T_LONG, length);
	break;

    case SIGNATURE_SHORT:
	obj = allocArray(ee, T_SHORT, length);
	break;


    case SIGNATURE_BOOLEAN:
	obj = allocArray(ee, T_BOOLEAN, length);
	break;

    case SIGNATURE_CLASS:
    {
	char buf[STK_BUF_LEN];
	char *classname;
	ClassClass *elemcb;
	int len = strlen(cbName(cb));

	if (len > STK_BUF_LEN) {
	    classname = sysMalloc(len);
	    if (classname == NULL) {
	        ThrowOutOfMemoryError(0, 0);
		return NULL;
	    }
	} else {
	    classname = buf;
	}
	  
	/* Copy  the classname omitting the leading "[L" and trailing ";" */
	strcpy(classname, &cbName(cb)[2]);
	classname[len - 3] = '\0';
	
	elemcb = FindClassFromClass(0, classname, TRUE, cb);
	if (classname != buf) {
	    sysFree(classname);
	}
	if (elemcb == NULL) {
	    /* exception already thrown */
	    return NULL;
	}
	obj = allocArray(ee, T_CLASS, length);
	if (obj != 0) {
	    unhand((HArrayOfObject *)obj)->body[length] = (JHandle*)elemcb;
	}
	break;
    }
    case SIGNATURE_ARRAY:
    {
	ClassClass *elemcb;
	elemcb = FindClassFromClass(0, &cbName(cb)[1], TRUE, cb);
	if (elemcb == NULL) {
	    /* exception already thrown */
	    return NULL;
	}
	obj = allocArray(ee, T_CLASS, length);
	if (obj != 0) {
	    unhand((HArrayOfObject *)obj)->body[length] = (JHandle*)elemcb;
	}
	break;
    }

    default:
	SignalError(0, OBJECT_PACKAGE "InvalidClassException", "Unknown signature");
	break;
    }
    if (obj == 0) {
	ThrowOutOfMemoryError(JNIEnv2EE(env), 0);
	return 0;
    }

    return MkRefLocal(env, obj);
}

/*
 * Load a class relative to the most recent class on the stack
 * with a non-null classloader.
 * It uses findClassFromClass which handles references to array
 * classes that come as "[[[[java.lang.Object". This functionality
 * is hard to achieve outside of native code. For example,
 * classloader.loadClass will not handle arrays.
 */
JNIEXPORT jclass JNICALL
JVM_LoadClass0(JNIEnv *env, jobject this, jclass currClass,
	       jstring currClassName)
{
    struct Hjava_lang_Class *class = (Hjava_lang_Class *)DeRef(env, currClass);
    struct javaframe *frame, frame_buf;
    ClassClass	*cb;
    char clname[STK_BUF_LEN], *p;
    int c;
    int unicode_len;
    jclass cls = 0;

    cb = class;			/* Default to use default system loader */
    if (cb == NULL) {
	for (frame = JNIEnv2EE(env)->current_frame ; frame != 0 ;) {
	    if ((frame->current_method != 0) &&
		!(frame->current_method->fb.access & ACC_NATIVE)){
		cb = fieldclass(&frame->current_method->fb);
		if (cb && cbLoader(cb)) {
		    /* Found a class with a non-zero class loader, use it */
		    break;
		}
	    }
	    frame = FRAME_PREV(frame, &frame_buf);
	}
    }

    /* Convert the string so we can call FindClassFromClass with a char * */
    unicode_len = (*env)->GetStringLength(env, currClassName);
    (*env)->GetStringUTFRegion(env, currClassName, 0, unicode_len, clname);
    p = clname;
    while (*p != '\0') {
	if (*p == '.') {
	    *p++ = '/';
	} else {
            next_utf2unicode(&p);
        }
    }

    /* passed FALSE to throw ClassNotFoundException on failure */
    if ((cb = FindClassFromClassLoader(JNIEnv2EE(env), clname, TRUE,
				       (cb) ? cbLoader(cb) : NULL, 
				       FALSE)) != 0) {
	cls = MkRefLocal(env, cbHandle(cb));
    }

    return cls;
}

/*
 * Code supporting reflection: used by Array.c
 */

/*
 * MACROS
 */

#define	CHECK_ARRAY(ee, harr, tag)				\
if (TRUE)							\
{								\
    if (harr == NULL) {						\
	ThrowNullPointerException(ee, NULL);			\
	goto tag;						\
    }								\
    if (obj_flags(harr) == T_NORMAL_OBJECT) {			\
	ThrowIllegalArgumentException(ee,			\
	    "argument is not an array");			\
	goto tag;						\
    }								\
} else ((void) 0)

#define	CHECK_ARRAY_AND_INDEX(ee, harr, index, tag)		\
if (TRUE)							\
{								\
    CHECK_ARRAY(ee, harr, tag);					\
    if (index < 0 || index >= obj_length(harr)) {		\
	ThrowArrayIndexOutOfBoundsException(ee, NULL);		\
	goto tag;						\
    }								\
} else ((void) 0)

#define	DECODE_ARRAY(ee, harr, ndx, p, pCode, tag, _p_)	        \
if (TRUE)							\
{								\
    CHECK_ARRAY_AND_INDEX(ee, harr, ndx, tag);			\
    _p_ = unhand(harr);						\
    switch(pCode = obj_flags(harr)) {				\
    case T_BOOLEAN:						\
    case T_BYTE:						\
	p = (char *) &((ArrayOfByte *)_p_)->body[ndx]; break;	\
    case T_SHORT:						\
	p = (char *) &((ArrayOfShort *)_p_)->body[ndx]; break;	\
    case T_CHAR:						\
	p = (char *) &((ArrayOfChar *)_p_)->body[ndx]; break;	\
    case T_INT:							\
	p = (char *) &((ArrayOfInt *)_p_)->body[ndx]; break;	\
    case T_LONG:						\
	p = (char *) &((ArrayOfLong *)_p_)->body[ndx]; break;	\
    case T_FLOAT:						\
	p = (char *) &((ArrayOfFloat *)_p_)->body[ndx]; break;	\
    case T_DOUBLE:						\
	p = (char *) &((ArrayOfDouble *)_p_)->body[ndx]; break;	\
    case T_CLASS:						\
	p = (char *) &((ArrayOfObject *)_p_)->body[ndx]; break;	\
    default:							\
	ThrowInternalError(ee, "illegal array type");		\
	goto tag;						\
    }								\
} else ((void) 0)

JNIEXPORT jint JNICALL
JVM_GetArrayLength(JNIEnv *env, jobject arr)
{
    struct Hjava_lang_Object *harr = (Hjava_lang_Object *)DeRef(env, arr);
    ExecEnv *ee = JNIEnv2EE(env);

    CHECK_ARRAY(ee, harr, out);
    return obj_length(harr);
  out:
    return 0;					/* exception */
}

JNIEXPORT jobject JNICALL
JVM_GetArrayElement(JNIEnv *env, jobject arr, jint index)
{
    struct Hjava_lang_Object *harr = (Hjava_lang_Object *)DeRef(env, arr);
    ExecEnv *ee = JNIEnv2EE(env);
    char *p;
    int pCode;
    jvalue v;
    OBJECT *obj_head;

    DECODE_ARRAY(ee, harr, index, p, pCode, bad, obj_head);

    REFLECT_GET(p, pCode, v, TRUE);
    
    KEEP_POINTER_ALIVE(obj_head);

    if (pCode == T_CLASS)
	return MkRefLocal(env, v.l);
    else
	return MkRefLocal(env, java_wrap(ee, v, (unsigned char)pCode));

  bad:
    return NULL;				/* exception */
}

JNIEXPORT jvalue JNICALL
JVM_GetPrimitiveArrayElement(JNIEnv *env, jobject arr, jint index, jint wCode)
{
    HObject *harr = (HObject *)DeRef(env, arr);
    ExecEnv *ee = JNIEnv2EE(env);
    char *p;
    int pCode;
    jvalue val;
    OBJECT *obj_head;

    DECODE_ARRAY(ee, harr, index, p, pCode, bad, obj_head);

    REFLECT_GET(p, pCode, val, TRUE);

    KEEP_POINTER_ALIVE(obj_head);

    if (pCode == T_CLASS)
	goto noconversion;

    if (pCode != wCode) 
	REFLECT_WIDEN(val, pCode, wCode, noconversion);

    return val;

  noconversion:
    ThrowIllegalArgumentException(ee, "array element type mismatch");

  bad:
    val.j = int2ll(0);
    return val;
}

JNIEXPORT void JNICALL
JVM_SetArrayElement(JNIEnv *env, jobject arr, jint index, jobject val)
{
    struct Hjava_lang_Object *harr = (Hjava_lang_Object *)DeRef(env, arr);
    struct Hjava_lang_Object *value = (Hjava_lang_Object *)DeRef(env, val);
    ExecEnv *ee = JNIEnv2EE(env);
    char *p;
    int pCode;
    OBJECT *obj_head;

    DECODE_ARRAY(ee, harr, index, p, pCode, bad, obj_head);

    if (pCode == T_CLASS) {
	long len = obj_length(harr);
	ArrayOfObject *optr = (ArrayOfObject *)unhand(harr);
	ClassClass *eltcb = (ClassClass *)optr->body[len];

	if (!is_instance_of(value, eltcb, ee))
	    goto noconversion;
	*(HObject **)p = value;

    } else {
	jvalue v;
	unsigned char vCode;

	if ((vCode = java_unwrap(value, &v)) == T_BAD)
	    return;				/* exception */
	if (vCode != pCode)
	    REFLECT_WIDEN(v, vCode, pCode, noconversion);
	REFLECT_SET(p, pCode, v, TRUE);
    }

    KEEP_POINTER_ALIVE(obj_head);

    return;

  noconversion:
    ThrowIllegalArgumentException(ee, "array element type mismatch");

  bad:
    return;
}

JNIEXPORT void JNICALL
JVM_SetPrimitiveArrayElement(JNIEnv *env, jobject arr, jint index, jvalue val,
			     unsigned char vCode)
{
    struct Hjava_lang_Object *harr =
	(Hjava_lang_Object *)DeRef(env, arr);
    ExecEnv *ee = JNIEnv2EE(env);
    char *p;
    int pCode;
    OBJECT *obj_head;

    DECODE_ARRAY(ee, harr, index, p, pCode, bad, obj_head);

    if (pCode == T_CLASS)
	goto noconversion;

    if (vCode != pCode)
	REFLECT_WIDEN(val, vCode, pCode, noconversion);
    REFLECT_SET(p, pCode, val, TRUE);

    KEEP_POINTER_ALIVE(obj_head);
    return;

  noconversion:
    ThrowIllegalArgumentException(ee, "array element type mismatch");

  bad:
    return;
}

JNIEXPORT jobject JNICALL
JVM_NewArray(JNIEnv *env, jclass eltClass, jint length)
{
    struct Hjava_lang_Class *eltType =
	(Hjava_lang_Class *)DeRef(env, eltClass);
    ExecEnv *ee = JNIEnv2EE(env);

    if (eltType == NULL) {
	ThrowNullPointerException(ee, NULL);
	return NULL;
    }
    if (length < 0) {
	ThrowNegativeArraySizeException(ee, NULL);
	return NULL;
    }
    if (cbName(eltType)[0] == SIGNATURE_ARRAY) {
	int eltDims =
	    cbConstantPool(eltType)[CONSTANT_POOL_ARRAY_DEPTH_INDEX].i;
	/* If eltDims is already 255, the we can't go on to create
	 * something of dims 256. */
	if (eltDims >= MAXDIM) {
	    sysAssert(eltDims == MAXDIM); /* Who allowed creation of a
					   * class block of 256 dims? */
	    ThrowIllegalArgumentException(ee, "too many dimensions");
	    return NULL;
	}
    }

    return MkRefLocal(env, reflect_new_array(ee, eltType, length));
}

JNIEXPORT jobject JNICALL
JVM_NewMultiArray(JNIEnv *env, jclass eltClass, jintArray dim)
{
    struct Hjava_lang_Class *eltType =
	(Hjava_lang_Class *)DeRef(env, eltClass);
    HArrayOfInt *hdim =
	(HArrayOfInt *)DeRef(env, dim);
    ExecEnv *ee = JNIEnv2EE(env);

    stack_item dims[MAXDIM];
    int i, len, l;
    int32_t *lp;
    char aname[MAXPATHLEN+8], *np = aname;
    ClassClass *eltcb, *acb;
    int eltDims = 0;

    if (eltType == NULL
	|| hdim == NULL) {
	ThrowNullPointerException(ee, NULL);
	return NULL;
    }

    len = obj_length(hdim);
    if (len <= 0) {
	sysAssert(len == 0); /* A -ve dimension array shouldn't have
			      * been created in the first place. */
	ThrowIllegalArgumentException(ee, "zero dimensions");
	return NULL;
    }

    if (cbName(eltType)[0] == SIGNATURE_ARRAY) {
	eltDims = cbConstantPool(eltType)[CONSTANT_POOL_ARRAY_DEPTH_INDEX].i;
    }

    if ((len + eltDims) > MAXDIM) {
	ThrowIllegalArgumentException(ee, "too many dimensions");
	return NULL;
    }

    lp = unhand(hdim)->body;

    for (i = 0; i < len; i++) {
	l = (int)lp[i];
	if (l < 0) {
	    ThrowNegativeArraySizeException(ee, NULL);
	    return NULL;
	}
	dims[i].i = l;
    }

    KEEP_POINTER_ALIVE(lp);

    eltcb = eltType;

    /*
     * Cook up a string naming the required array class
     */
    for (i = 0; i < len; i++) {
	*np++ = SIGNATURE_ARRAY;
    }
    if (cbIsPrimitive(eltcb)) {
	/* Primitive array */
	*np++ = cbTypeSig(eltcb);
    } else {
	char *eltname = cbName(eltcb);
	int eltnamelen = strlen(eltname);

        if (eltnamelen + len >= MAXPATHLEN) {
            ThrowInternalError(ee, "signature overflow");
            return NULL;
        }

	if (eltname[0] == SIGNATURE_ARRAY) {
	    /* Array of array type: handle it */
	    strcpy(np, eltname);
	    np += eltnamelen;
	} else {
	    /* Class or interface array */
	    *np++ = SIGNATURE_CLASS;
	    strcpy(np, eltname);
	    np += eltnamelen;
	    *np++ = SIGNATURE_ENDCLASS;
	}
    }
    *np = '\0';					/* NUL-terminate */
    
    if ((acb = FindClassFromClass(ee, aname, TRUE, eltcb)) == NULL) {
	/* XXX: Is this too severe? */
	ThrowNoClassDefFoundError(ee, aname);
	return NULL;
    }

    return MkRefLocal(env, MultiArrayAlloc(len, acb, dims));
}

/*
 * Code supporting reflection: used by Field.c
 */

/*
 * MACROS
 */

#define obj_getoffset(o, off)	(*(OBJECT *)((char *)unhand(o)+(off)))

/*
 *
 */

static ClassClass *
classJavaLangReflectField(void)
{
    static ClassClass *cb = NULL;
    if (cb) return (cb);
    cb = FindStickySystemClass(NULL, JAVAPKG "reflect/Field", TRUE);
    return cb;
}

/*
 * A reflected field's slot is an index into cbFields(cb) of its
 * declaring class.  Tries to be paranoid.
 */
#define	GET_CLASS_FIELDBLOCK(cb, slot)				\
    (((slot) < 0 || (slot) >= (cbFieldsCount(cb)))		\
	? NULL : (cbFields(cb) + (slot)))

/*
 * NOTE: Relies on the fact that fields are allocated contiguously in
 * cbFields(cb).  Else, we have to match on ID.
 */
#define	GET_FIELDBLOCK_SLOT(cb, fb) ((fb) - cbFields(cb))

static struct fieldblock *
get_field(
    ExecEnv *ee,
    struct Hjava_lang_reflect_Field * hf,
    struct Hjava_lang_Object * o,
    OBJECT **addrp,				/* out */
    ClassClass **typep)				/* out */
{
    ClassClass *cb, *tcb, *ocb = NULL;
    struct fieldblock *fb;
    OBJECT *p;
    int isStatic;

    /* Get class declaring member */
    cb = member_class(hf);
    /* Get type of field */
    tcb = member_type(hf);

    ENSURE_INITIALIZED(ee, cb);			/* new in 1.2 */

    /* Get declaring class fieldblock for slot */
    if ((fb = GET_CLASS_FIELDBLOCK(cb, member_slot(hf))) == NULL) {
	ThrowInternalError(0, "illegal Field");
	return NULL;				/* exception */
    }

    isStatic = fbIsStatic(fb);

    if (isStatic) {
	/* Don't check object */
	p = (cbIsPrimitive(tcb) && (cbSlotSize(tcb) == 8))
	    ? twoword_static_address(fb) : normal_static_address(fb);
	ocb = cb;

    } else {
	/* Check object is a non-null instance of declaring class */
	if (o == NULL) {
	    ThrowNullPointerException(ee, NULL);
	    return NULL;
	}
	ocb = obj_array_classblock(o);
	if (!(ocb == cb || is_instance_of(o, cb, ee))) {
	    ThrowIllegalArgumentException(ee,
		"object is not an instance of declaring class");
	    return NULL;
	}
	p = &obj_getoffset(o, fbOffset(fb));
    }

    /*
     * Access checking (unless overridden by Field)
     */
    if (!member_override(hf))
	if (!(IsPublic(cbAccess(cb)) && IsPublic(fbAccess(fb))))
	    if (!reflect_check_access(ee, cb, fbAccess(fb), ocb, FALSE))
		return NULL;			/* exception */

    *addrp = p;
    *typep = tcb;
    return (fb);
}

JNIEXPORT jobject JNICALL
JVM_GetField(JNIEnv *env, jobject field, jobject obj)
{
    struct Hjava_lang_reflect_Field *hf =
	(Hjava_lang_reflect_Field *)DeRef(env, field);
    struct Hjava_lang_Object *o =
	(Hjava_lang_Object *)DeRef(env, obj);
    ExecEnv *ee = JNIEnv2EE(env);
    ClassClass *cb;
    OBJECT *p;
    void *unhandO = o ? unhand(o) : 0;
    Hjava_lang_Object *result;

    if (get_field(ee, hf, o, &p, &cb) == NULL)
	return NULL;				/* exception */

    if (cbIsPrimitive(cb)) {
	jvalue v;
	unsigned char pCode = cbTypeCode(cb);

	REFLECT_GET(p, pCode, v, FALSE);
	result = java_wrap(ee, v, pCode);

    } else {
	result = *(HObject **)p;
    }
    KEEP_POINTER_ALIVE(unhandO);
    return MkRefLocal(env, result);
}

JNIEXPORT jvalue JNICALL
JVM_GetPrimitiveField(JNIEnv *env, jobject field, jobject obj, 
		      unsigned char wCode)
{
    struct Hjava_lang_reflect_Field *hf =
	(Hjava_lang_reflect_Field *)DeRef(env, field);
    struct Hjava_lang_Object *o =
	(Hjava_lang_Object *)DeRef(env, obj);
    ExecEnv *ee = JNIEnv2EE(env);
    ClassClass *cb;
    OBJECT *p;
    jvalue val;
    void *unhandO = o ? unhand(o) : 0;

    if (get_field(ee, hf, o, &p, &cb) != NULL
        && cbIsPrimitive(cb)) {
	unsigned char pCode = cbTypeCode(cb);

	REFLECT_GET(p, pCode, val, FALSE);
	if (pCode != wCode)
	    REFLECT_WIDEN(val, pCode, wCode, bad);

	KEEP_POINTER_ALIVE(unhandO);
	return val;
    }

bad: /* exception */
    ThrowIllegalArgumentException(ee, "field type mismatch");
    val.j = int2ll(0);		/* muffle compiler warning */
    return val;
}

JNIEXPORT void JNICALL
JVM_SetField(JNIEnv *env, jobject field, jobject obj, jobject val)
{
    struct Hjava_lang_reflect_Field *hf =
	(Hjava_lang_reflect_Field *)DeRef(env, field);
    struct Hjava_lang_Object *o =
	(Hjava_lang_Object *)DeRef(env, obj);
    struct Hjava_lang_Object *value =
	(Hjava_lang_Object *)DeRef(env, val);
    ExecEnv *ee = JNIEnv2EE(env);
    struct fieldblock *fb;
    ClassClass *cb;
    OBJECT *p;
    void *unhandO = o ? unhand(o) : 0;

    if ((fb = get_field(ee, hf, o, &p, &cb)) == NULL)
	return;					/* exception */

    if (fbIsFinal(fb) && (!member_override(hf))) {
	ThrowIllegalAccessException(ee, "field is final");
	return;
    }

    if (cbIsPrimitive(cb)) {
	jvalue v;
	unsigned char pCode = cbTypeCode(cb);
	unsigned char vCode;

	if ((vCode = java_unwrap(value, &v)) == T_BAD)
	    return;				/* exception */

	if (vCode != pCode)
	    REFLECT_WIDEN(v, vCode, pCode, bad);

	REFLECT_SET(p, pCode, v, FALSE);
	KEEP_POINTER_ALIVE(unhandO);
	return;

    } else if (is_instance_of(value, cb, ee)) {
	*(HObject **)p = value;
	KEEP_POINTER_ALIVE(unhandO);
	return;
    }

  bad:
    ThrowIllegalArgumentException(ee, "field type mismatch");
}

JNIEXPORT void JNICALL
JVM_SetPrimitiveField(JNIEnv *env, jobject field, jobject obj, jvalue val,
		      unsigned char vCode)
{
    struct Hjava_lang_reflect_Field *hf =
	(Hjava_lang_reflect_Field *)DeRef(env, field);
    struct Hjava_lang_Object *o =
	(Hjava_lang_Object *)DeRef(env, obj);
    ExecEnv *ee = JNIEnv2EE(env);
    struct fieldblock *fb;
    ClassClass *cb;
    OBJECT *p;
    void *unhandO = o ? unhand(o) : 0;

    if ((fb = get_field(ee, hf, o, &p, &cb)) == NULL)
	return;					/* exception */

    if (fbIsFinal(fb) && (!member_override(hf))) {
	ThrowIllegalAccessException(ee, "field is final");
	return;
    }

    if (cbIsPrimitive(cb)) {
	unsigned char pCode = cbTypeCode(cb);
	if (pCode != vCode)
	    REFLECT_WIDEN(val, vCode, pCode, bad);
	REFLECT_SET(p, pCode, val, FALSE);
	KEEP_POINTER_ALIVE(unhandO);
	return;
    }

  bad:
    ThrowIllegalArgumentException(ee, "field type mismatch");
}

static HObject *
new_field(ExecEnv *ee, struct fieldblock *fb)
{
    ClassClass *cb, *fcb;
    HString *hname;
    Hjava_lang_reflect_Field *hf;
    long slot;

    /* declaring class */
    cb = fbClass(fb);

    /* slot */
    slot = GET_FIELDBLOCK_SLOT(cb, fb);
    sysAssert((slot >= 0) && (slot < cbFieldsCount(cb)));
    sysAssert(fb == GET_CLASS_FIELDBLOCK(cb, slot));

    /* type */
    if ((fcb = reflect_find_class(fbSig(fb), cb, NULL)) == NULL)
	return NULL;				/* exception */

    /* name */
    if ((hname = makeJavaStringUTF(fbName(fb))) == NULL)
	return NULL;				/* exception */

    /* Field */
    if ((hf = (Hjava_lang_reflect_Field *)
	allocObject(ee, classJavaLangReflectField())) == NULL) {
	ThrowOutOfMemoryError(ee, NULL);
	return NULL;
    }

    /* initialize the Field */
    member_class(hf) = cbHandle(cb);
    member_slot(hf) = slot;
    member_name(hf) = hname;
    member_type(hf) = cbHandle(fcb);
    member_modifiers(hf) = fbAccess(fb) & ACC_WRITTEN_FLAGS;
    member_override(hf) = FALSE;

    return (HObject *)hf;
}

/*
 * NOTE: These macros always walk from subclass towards superclass
 */

#define	WALK_DECLARED_FIELDS(cb, fb, body)				\
if (TRUE)								\
{									\
    struct fieldblock *_list_ = cbFields(cb);				\
    int _i_, _list_count_ = cbFieldsCount(cb);				\
    /* Walk in reverse declared order */				\
    for (_i_ = _list_count_ - 1; _i_ >= 0; _i_--) {			\
	struct fieldblock *fb = &_list_[_i_];				\
	body;								\
    }									\
} else ((void) 0)

#define	WALK_INSTANCE_FIELDS(cb, fb, body)				\
if (TRUE)								\
{									\
    ClassClass *_cb_;   				                \
    for (_cb_ = cb; _cb_; _cb_ = cbSuperclass(_cb_)) {			\
        struct fieldblock *_fields_ = cbFields(_cb_);			\
	int _i_;                                                        \
        for (_i_ = cbFieldsCount(_cb_) - 1; _i_ >= 0; _i_--) {		\
	    struct fieldblock *fb = &(_fields_[_i_]);		       	\
	    body;						       	\
        }							       	\
    }									\
} else ((void) 0)

/* interfaces don't get a usable slottable */
#define	WALK_INTERFACE_FIELDS(cb, fb, body)				\
if (TRUE)								\
{									\
    struct imethodtable *_imt_ = cbIntfMethodTable(cb);			\
    int _j_, _icnt_ = _imt_->icount;					\
    /* Walk from subinterface towards superinterface */			\
    for (_j_ = 0; _j_ < _icnt_; _j_++) {				\
        ClassClass * _cb_ = _imt_->itable[_j_].classdescriptor;		\
        WALK_DECLARED_FIELDS(_cb_, fb, body);				\
    }									\
} else ((void) 0)

HArrayOfObject *
reflect_fields(
    ExecEnv *ee,
    ClassClass * cb,
    int which)
{
    HArrayOfObject *harr;
    HObject **arr;
    int count = 0, j;

    if (cbIsPrimitive(cb)
	|| (cbName(cb)[0] == SIGNATURE_ARRAY))
	return (HArrayOfObject *)
	    reflect_new_array(ee, classJavaLangReflectField(), 0);

    ENSURE_LINKED(ee, cb);			/* was ENSURE_INITIALIZED */

    switch (which) {

    case MEMBER_PUBLIC: {
	int isInterface = cbIsInterface(cb);

	if (!isInterface) {
	    /* Don't double-count interface fields */
	    WALK_INSTANCE_FIELDS(cb, fb, {
		/* count public class fields */
		if (fbIsPublic(fb))
		    count++;
	    });
	}

	/* all interface fields are public and accessible */
	WALK_INTERFACE_FIELDS(cb, fb, {
	    count++;
	});

	if ((harr = (HArrayOfObject *)
	    reflect_new_array(ee, classJavaLangReflectField(), count)) == NULL)
	    return NULL;			/* exception */
	arr = (HObject **) unhand(harr)->body;

	j = count;

	if (!isInterface) {
	    WALK_INSTANCE_FIELDS(cb, fb, {
		if (fbIsPublic(fb))
		    if ((arr[--j] = new_field(ee, fb)) == NULL)
			return NULL;		/* exception */
	    });
	}

	WALK_INTERFACE_FIELDS(cb, fb, {
	    if ((arr[--j] = new_field(ee, fb)) == NULL)
		return NULL;		/* exception */
	});

	sysAssert(j == 0);
	KEEP_POINTER_ALIVE(arr);
	return harr;
    }

    case MEMBER_DECLARED: {

	WALK_DECLARED_FIELDS(cb, fb, {
	    count++;
	});

	if ((harr = (HArrayOfObject *)
	    reflect_new_array(ee, classJavaLangReflectField(), count)) == NULL)
	    return NULL;			/* exception */
	arr = (HObject **) unhand(harr)->body;

	j = count;

	WALK_DECLARED_FIELDS(cb, fb, {
	    if ((arr[--j] = new_field(ee, fb)) == NULL)
		return NULL;		/* exception */
	});

	sysAssert(j == 0);
	KEEP_POINTER_ALIVE(arr);
	return harr;
    }

    default:
	ThrowInternalError(ee, "illegal member kind");
	return NULL;
    }
}


HObject *
reflect_field(
    ExecEnv *ee,
    ClassClass *cb,
    char *name,
    int which)
{
    char *hashed_name = HashUTF8(ee, name);

    if (cbIsPrimitive(cb)
	|| (cbName(cb)[0] == SIGNATURE_ARRAY)
	|| (hashed_name == NULL))
	goto nosuchfield;

    ENSURE_LINKED(ee, cb);			/* was ENSURE_INITIALIZED */

    switch (which) {

    case MEMBER_PUBLIC: {
	int isInterface = cbIsInterface(cb);

	if (!isInterface) {
	    WALK_INSTANCE_FIELDS(cb, fb, {
		if (fbIsPublic(fb)
		    && (fbName(fb) == hashed_name))
		    return new_field(ee, fb);
	    });
	}

	WALK_INTERFACE_FIELDS(cb, fb, {
	    if (fbName(fb) == hashed_name)
		return new_field(ee, fb);
	});

	break;
    }

    case MEMBER_DECLARED: {

	WALK_DECLARED_FIELDS(cb, fb, {
	    if (fbName(fb) == hashed_name)
		return new_field(ee, fb);
	});

	break;
    }

    default:
	break;
    }    

  nosuchfield:
    ThrowNoSuchFieldException(ee, name);
    return NULL;
}

HObject *
reflect_new_field(ExecEnv *ee, struct fieldblock *fb)
{
    return new_field(ee, fb);
}

struct fieldblock *
reflect_get_fieldblock(HObject *obj)
{
    struct Hjava_lang_reflect_Field *hf =
	(struct Hjava_lang_reflect_Field *)obj;
    ClassClass *cb = member_class(hf);
    return GET_CLASS_FIELDBLOCK(cb, member_slot(hf));
}

/*
 * Code supporting reflection: used by Method.c
 *
 * NOTE: Some of this should be replaced by direct JNI calls, as the
 * JNI already provides much of the necessary functionality.
 */


#define	array_length(h)	(h == NULL ? 0 : obj_length(h))

/*
 * Error utilities
 */

#define WrongNumberOfArguments(ee)				\
	ThrowIllegalArgumentException(ee, "wrong number of arguments")

#define	ArgumentTypeMismatch(ee)					\
	ThrowIllegalArgumentException(ee, "argument type mismatch")

static void
InvocationTargetException(ExecEnv * ee)
{
    struct Hjava_lang_Throwable *target;
    struct Hjava_lang_reflect_InvocationTargetException *wrapper;

    target = (struct Hjava_lang_Throwable *) ee->exception.exc;
    /* XXX So that SignalError finds the InvocationTargetException class */
    exceptionClear(ee);
    SignalError(ee, JAVAPKG "reflect/" "InvocationTargetException", NULL);
    wrapper = (struct Hjava_lang_reflect_InvocationTargetException *)
	ee->exception.exc;
    unhand(wrapper)->target = target;
}

/*
 * A reflected method's slot is an index into the methods list of its
 * declaring classblock.  This tries to be paranoid.
 */

#define	GET_CLASS_METHODBLOCK(cb, slot)				\
    (((slot) < 0 || (slot) >= (cbMethodsCount(cb)))		\
	? NULL : (cbMethods(cb) + (slot)))

/*
 * NOTE: Relies on the fact that methods are allocated contiguously in
 * cbMethods(cb).  Else, we have to match on ID.
 */
#define	GET_METHODBLOCK_SLOT(cb, mb) ((mb) - cbMethods(cb))

/*
 *
 */

static ClassClass *
classJavaLangReflectMethod(void)
{
    static ClassClass *cb = NULL;
    if (cb) return (cb);
    cb = FindStickySystemClass(NULL, JAVAPKG "reflect/Method", TRUE);
    return cb;
}

static ClassClass *
classJavaLangReflectConstructor(void)
{
    static ClassClass *cb = NULL;
    if (cb) return (cb);
    cb = FindStickySystemClass(NULL, JAVAPKG "reflect/Constructor", TRUE);
    return cb;
}

/*
 * Instance method resolution
 */
static struct methodblock *
get_object_methodblock(
    HObject *o,
    ClassClass *ocb,
    struct methodblock *mb)
{
    ClassClass *cb = mbClass(mb);		/* declaring class */
    struct methodblock *omb;

    sysAssert(!mbIsStatic(mb));
    sysAssert(CCIs(ocb, Linked));
    sysAssert(!cbIsInterface(ocb));

    if (cbIsInterface(cb)) {
	/*
	 * Interface method lookup
	 * An interface method's offset is its index in the interface
	 */
	struct imethodtable *imt = cbIntfMethodTable(ocb);
	int icount = imt->icount, i;
	for (i = 0; i < icount; i++) {
	    if (cb == imt->itable[i].classdescriptor) {
	        int offset = imt->itable[i].offsets[mbOffset(mb)];
		omb = mt_slot(obj_array_methodtable(o), offset);
		goto found;
	    }
	}
	goto nosuchmethod;

    } else if (mbAccess(mb) & ACC_PRIVATE) {
	/*
	 * Private methods are not in the dispatch table
	 */
	omb = mb;

    } else {
	/*
	 * Dynamic method lookup
	 * An instance method's offset is its index in the methodtable
	 */
	omb = mt_slot(obj_array_methodtable(o), mbOffset(mb));
    }

  found:
    sysAssert(mb->fb.name == omb->fb.name);
    sysAssert(mb->fb.signature == omb->fb.signature);
    return omb;

  nosuchmethod:
    ThrowNoSuchMethodException(0, mbName(mb));
    return NULL;
}

/*
 * Count parameters in sig
 */
static int
get_parameter_count(char *sig)
{
    char *p = sig + 1;
    int cnt = 0;

    for (; *p != SIGNATURE_ENDFUNC; p++) {
	switch (*p) {
	case SIGNATURE_CLASS:
	    while (*p != SIGNATURE_ENDCLASS) p++;
	    break;
	case SIGNATURE_ARRAY:
	    while (*p == SIGNATURE_ARRAY) p++;
	    if (*p == SIGNATURE_CLASS) while (*p != SIGNATURE_ENDCLASS) p++;
	    break;
	default:				/* primitive */
	    break;
	}
	cnt++;					/* 1 arg */
    }
    return cnt;
}

/*
 * Convert signature to argument types, resolving classes relative to cb.
 */
static HArrayOfObject *
get_parameter_types(
    ExecEnv	*ee,
    char	*sig,
    ClassClass	*cb,
    char	**cpp)
{
    char	*p;
    int		cnt, i = 0;
    HArrayOfObject *hclasses = NULL;
    Hjava_lang_Class	**classes = NULL;
    ClassClass	*pcb;

    sysAssert(sig[0] == SIGNATURE_FUNC);

    /*
     * Compute number of parameters
     */
    cnt = get_parameter_count(sig);

    /*
     * Allocate class array
     */
    if ((hclasses = reflect_new_class_array(ee, cnt)) == NULL)
	return NULL;				/* exception */
    classes = (Hjava_lang_Class **) unhand(hclasses)->body;

    /*
     * Resolve parameter types to classblocks
     */
    p = sig + 1;
    while (*p != SIGNATURE_ENDFUNC) {
	if ((pcb = reflect_find_class(p, cb, &p)) == NULL)
	    return NULL;			/* exception */
	classes[i++] = cbHandle(pcb);
    }
    KEEP_POINTER_ALIVE(classes);

    sysAssert(*p == SIGNATURE_ENDFUNC);
    sysAssert(i == cnt);

    if (cpp)
	*cpp = p + 1;

    return hclasses;

}

static HArrayOfObject *
get_exception_types(
    ExecEnv *ee,
    ClassClass *cb,
    struct methodblock *mb)
{
    unsigned short cnt;
    HArrayOfObject *exceptionTypes;

    cnt = mb->nexceptions;

    if ((exceptionTypes = reflect_new_class_array(ee, cnt)) == NULL)
	return NULL;				/* exception */

    if (cnt != 0) {
	ExecEnv *ee = EE();
	unsigned short *exceptions;
	Hjava_lang_Class **classes;
	unsigned short i;

	exceptions = mb->exceptions;
	classes = (Hjava_lang_Class **) unhand(exceptionTypes)->body;
	for (i = 0; i < cnt; i++) {
	    if (!ResolveClassConstantFromClass(cb, exceptions[i],
		ee, 1 << CONSTANT_Class)) {
		return NULL;			/* exception */
	    }
	    classes[i] = cbHandle(cbConstantPool(cb)[exceptions[i]].clazz);
	}
	KEEP_POINTER_ALIVE(classes);
    }

    return exceptionTypes;
}

/*
 * Push one argument onto stack, with typechecking.
 * Unwrap and widen primitives if necessary.
 */
static bool_t
push_arg(
    ExecEnv *ee,
    JavaFrame *current_frame,
    HObject *arg,
    ClassClass *argType)
{

    if (cbIsPrimitive(argType)) {
	jvalue v;
	unsigned char pCode = cbTypeCode(argType);
	unsigned char vCode;

	if ((vCode = java_unwrap(arg, &v)) == T_BAD)
	    return FALSE;			/* exception */

	if (vCode != pCode)
	    REFLECT_WIDEN(v, vCode, pCode, bad);

	REFLECT_SET(&(current_frame->optop)->i, pCode, v, FALSE);
	current_frame->optop += T_TYPE_WORDS(pCode);
	return TRUE;

    } else if (is_instance_of(arg, argType, ee)) {
	(current_frame->optop++)->h = arg;
	return TRUE;

    }

  bad:
    ArgumentTypeMismatch(ee);
    return FALSE;
}

/*
 * Pop result from stack after method invocation, with possible wrapping.
 */
static HObject *
pop_result(
    ExecEnv *ee,
    JavaFrame *current_frame,
    ClassClass *returnType)
{

    if (cbIsPrimitive(returnType)) {
	/*
	 * Wrap primitive value
	 */
	jvalue v;
	jvalue tmp;
	unsigned char pCode = cbTypeCode(returnType);

	switch (pCode) {
	case T_VOID:
	    return NULL;
        case T_BOOLEAN:
            v.z = (jboolean) (current_frame->optop)[-1].i;
            break;
        case T_BYTE:
            v.b = (jbyte) (current_frame->optop)[-1].i;
            break;
        case T_CHAR:
            v.c = (jchar) (current_frame->optop)[-1].i;
            break;
        case T_SHORT:
            v.s = (jshort) (current_frame->optop)[-1].i;
            break;
        case T_INT:
            v.i = (current_frame->optop)[-1].i;
            break;
        case T_FLOAT:
            v.f = (jfloat) (current_frame->optop)[-1].f;
            break;
	case T_LONG:
	    v.j = GET_INT64(tmp, &(current_frame->optop)[-2].i);
	    break;
	case T_DOUBLE:
	    v.d = GET_DOUBLE(tmp, &(current_frame->optop)[-2].i);
	    break;
	default: /* handle */
	    v.l = (jobject) (current_frame->optop)[-1].h;
	    break;
	}

	return java_wrap(ee, v, pCode);

    } else {
	/*
	 * Return object or array on stack
	 */
	return current_frame->optop[-1].h;
    }
}

/*
 * Invoke resolved method, with typechecking.
 *
 * Our caller has checked object type, has resolved method, and has
 * checked that the number of arguments matches that expected by
 * target method.  Primitive arguments are automatically unwrapped and
 * primitive return values are automatically wrapped.
 */
static HObject *
invoke(
    ExecEnv * ee,				/* required */
    HObject * obj,				/* NULL ok for static */
    struct methodblock * mb,			/* required */
    int nargs,					/* required */
    Hjava_lang_Class ** parameterTypes,		/* required */
    HObject ** args,				/* NULL ok for no args */
    Hjava_lang_Class * returnType,		/* required */
    bool_t isStaticCall)			/* required */

{

    JavaFrame *current_frame;
    int     i;
    unsigned char pc[6];
    cp_item_type constant_pool[4];
    unsigned char cpt[4];
    HObject *result = NULL;
    JNIEnv *env = EE2JNIEnv(ee);

    /*
     * Setup stack
     */
    {
        int stack_size = mbArgsSize(mb);
	if (stack_size < 2) {
	    stack_size = 2;
	}
	if ((*env)->PushLocalFrame(env, stack_size) < 0) {
	    return NULL;
	}
    }
    current_frame = ee->current_frame;
    current_frame->monitor = NULL;

    /*
     * NOTE: All returns from here onwards must goto out, to unwind stack
     */

    /*
     * Push target object if not static call
     */
    if (!isStaticCall)
	(current_frame->optop++)->p = obj;

    /*
     * Push each argument, with possible unwrapping and widening
     */
    for (i = 0; i < nargs; i++) {
	if (!push_arg(ee, current_frame, args[i], parameterTypes[i]))
	    goto out;				/* exception */
    }
    
    /*
     * Assemble glue code for invocation
     */
    constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type = cpt;
    cpt[0] = CONSTANT_POOL_ENTRY_RESOLVED;

    pc[0] = isStaticCall ? opc_invokestatic_quick : opc_invokenonvirtual_quick;
    pc[1] = 0; pc[2] = 1;			/* constant pool entry #1 */
    pc[3] = opc_return;

    constant_pool[1].mb = mb;
    cpt[1] = CONSTANT_POOL_ENTRY_RESOLVED | CONSTANT_Methodref;

    current_frame->constant_pool = constant_pool;

    /*
     * Run the byte codes outside the lock and catch any exceptions
     */
    exceptionClear(ee);
    if ((*pExecuteJava)(pc, ee)) {
	/*
	 * Success: wrap primitive value if needed
	 */
	result = pop_result(ee, current_frame, returnType);
    } else {
	/*
	 * Method threw an exception: wrap in an InvocationTargetException
	 */
	InvocationTargetException(ee);
    }

    /*
     * Done: caller looks at ee->exceptionKind and ee->exception
     */
  out:
    (*env)->PopLocalFrame(env, 0);
    return result;

}

JNIEXPORT jobject JNICALL
JVM_InvokeMethod(JNIEnv *env, jobject method, jobject obj, jobjectArray args0)
{
    struct Hjava_lang_reflect_Method *hm =
	(Hjava_lang_reflect_Method *)DeRef(env, method);
    struct Hjava_lang_Object *o =
	(Hjava_lang_Object *)DeRef(env, obj);
    HArrayOfObject *hargs =
	(HArrayOfObject *)DeRef(env, args0);
    ExecEnv *ee = JNIEnv2EE(env);

    ClassClass *cb, *icb;
    struct methodblock *mb, *imb;
    int     isStatic;
    long    cnt, acnt;
    HArrayOfObject *htypes;
    Hjava_lang_Class **types;
    HObject **args;
    struct Hjava_lang_Object *result;

    /* Get class declaring method */
    cb = member_class(hm);

    /* Get declaring class's methodblock for slot */
    if ((mb = GET_CLASS_METHODBLOCK(cb, member_slot(hm))) == NULL) {
	ThrowInternalError(ee, "invoke");
	return NULL;
    }

    ENSURE_INITIALIZED(ee, cb);			/* new in 1.2 */

    isStatic = mbIsStatic(mb);

    if (isStatic) {
	/* If static method, ignore object argument */
	icb = cb;
	imb = mb;
    } else {
	/* Check class of object against class declaring method */
	if (o == NULL) {
	    ThrowNullPointerException(ee, NULL);
	    return NULL;
	}
	icb = obj_array_classblock(o);
	if (!(icb == cb || is_instance_of(o, cb, ee))) {
	    ThrowIllegalArgumentException(ee,
		"object is not an instance of declaring class");
	    return NULL;
	}
	/* Resolve to instance method */
	if ((imb = get_object_methodblock(o, icb, mb)) == NULL)
	    return NULL;			/* exception */
    }

    /*
     * Access checking (unless overridden by Method)
     * XXX should be 'mb' or 'imb'?
     */
    if (!member_override(hm))
	if (!(IsPublic(cbAccess(cb)) && IsPublic(mbAccess(mb))))
	    if (!reflect_check_access(ee, cb, mbAccess(mb), icb, TRUE))
		return NULL;			/* exception */

    htypes = member_parameterTypes(hm);
    types = (Hjava_lang_Class **) unhand(htypes)->body;

    cnt = array_length(htypes);
    acnt = array_length(hargs);

    /* Check number of arguments */
    if (cnt != acnt) {
	WrongNumberOfArguments(ee);
	return NULL;
    }

    args = hargs ? unhand(hargs)->body : NULL;

    /* Invoke method, returning its result */
    result = invoke(ee, o, imb, cnt, types, args, 
		    member_returnType(hm), isStatic);
    KEEP_POINTER_ALIVE(args || types);
    return MkRefLocal(env, result);
}

JNIEXPORT jobject JNICALL
JVM_NewInstanceFromConstructor(JNIEnv *env, jobject c, jobjectArray args0)
{
    struct Hjava_lang_reflect_Constructor *hc =
	(Hjava_lang_reflect_Constructor *)DeRef(env, c);
    HArrayOfObject *hargs
	= (HArrayOfObject *)DeRef(env, args0);
    ExecEnv *ee = JNIEnv2EE(env);

    ClassClass *cb;
    struct methodblock *mb;
    long   cnt, acnt;
    HObject *o;
    HArrayOfObject *htypes;
    Hjava_lang_Class **types;
    HObject **args;

    /* Get class declaring constructor */
    cb = member_class(hc);

    if ((mb = GET_CLASS_METHODBLOCK(cb, member_slot(hc))) == NULL) {
	ThrowInternalError(ee, "newInstance");
	return NULL;
    }

    sysAssert(mbName(mb) == utf8_literal_obj_init_name);

    ENSURE_INITIALIZED(ee, cb);			/* new in 1.2 */

    /* Check instantiable class */
    if (cbAccess(cb) & (ACC_INTERFACE | ACC_ABSTRACT)) {
	ThrowInstantiationException(0, cbName(cb));
	return NULL;
    }

    /*
     * Access checking (unless overridden by Constructor)
     * XXX 'mb' or 'imb'?
     */
    if (!member_override(hc))
	if (!(IsPublic(cbAccess(cb)) && IsPublic(mbAccess(mb))))
	    if (!reflect_check_access(ee, cb, mbAccess(mb), cb, FALSE))
		return NULL;			/* exception */

    /* Allocate the object */
    if ((o = allocObject(ee, cb)) == NULL) {
	ThrowOutOfMemoryError(ee, 0);
	return NULL;
    }

    htypes = member_parameterTypes(hc);
    types = (Hjava_lang_Class **) unhand(htypes)->body;

    cnt = array_length(htypes);
    acnt = array_length(hargs);

    /* Check number of arguments */
    if (cnt != acnt) {
	WrongNumberOfArguments(ee);
	return NULL;
    }
    
    args = hargs ? unhand(hargs)->body : NULL;

    /* Invoke method: returns nothing */
    invoke(ee, o, mb, cnt, types, args, cbHandle(class_void), FALSE);

    KEEP_POINTER_ALIVE(args || types);

    /* Bail if exception */
    if (exceptionOccurred(ee))
	return NULL;

    return MkRefLocal(env, o);
}

/*
 * Support code for class Class native code
 */

static HObject *
new_method(ExecEnv *ee, struct methodblock *mb)
{
    ClassClass *cb;
    long slot;
    char *p;
    HArrayOfObject *parameterTypes, *exceptionTypes;
    ClassClass *returnType;
    HString *hname;
    Hjava_lang_reflect_Method *hm;

    /* declaring class */
    cb = mbClass(mb);

    /* slot */
    slot = GET_METHODBLOCK_SLOT(cb, mb);
    sysAssert((slot >= 0) && (slot < cbMethodsCount(cb)));

    /* parameter types and return type from signature */

    p = mbSig(mb);

    if ((parameterTypes = get_parameter_types(ee, p, cb, &p)) == NULL)
	return NULL;				/* exception */

    sysAssert(*(p-1) == SIGNATURE_ENDFUNC);

    if ((returnType = reflect_find_class(p, cb, NULL)) == NULL)
	return NULL;				/* exception */

    if ((exceptionTypes = get_exception_types(ee, cb, mb)) == NULL)
	return NULL;				/* exception */

    if ((hname = makeJavaStringUTF(mbName(mb))) == NULL)
	return NULL;				/* exception */

    if ((hm = (Hjava_lang_reflect_Method *)
	allocObject(ee, classJavaLangReflectMethod())) == NULL) {
	ThrowOutOfMemoryError(ee, 0);
	return NULL;
    }

    /* initialize the Method */
    member_class(hm) = cbHandle(cb);
    member_slot(hm) = slot;
    member_name(hm) = hname;
    member_returnType(hm) = cbHandle(returnType);
    member_parameterTypes(hm) = parameterTypes;
    member_exceptionTypes(hm) = exceptionTypes;
    member_modifiers(hm) = mbAccess(mb) & ACC_WRITTEN_FLAGS;
    member_override(hm) = FALSE;

    return (HObject *)hm;
}

/*
 * NOTE: These macros always walk from subclass towards superclass
 */

#define	WALK_DECLARED_METHODS(cb, mb, body)				\
if (TRUE)								\
{									\
    struct methodblock *_list_ = cbMethods(cb);				\
    int _i_, _list_count_ = cbMethodsCount(cb);				\
    /* Walk in reverse declared order */				\
    for (_i_ = _list_count_ - 1; _i_ >= 0; _i_--) {			\
	struct methodblock *mb = &_list_[_i_];				\
	body;								\
    }									\
} else ((void) 0)

#define	WALK_INSTANCE_METHODS(cb, mb, body)				\
if (TRUE)								\
{									\
    struct methodblock **_mt_ = cbMethodTable(cb)->methods;		\
    int _tcnt_ = cbMethodTableSize(cb), _i_;				\
    /* N.B. Slot 0 of methodtable is empty */				\
    /* Walk from subtype towards supertype */				\
    for (_i_ = _tcnt_ - 1; _i_ > 0; _i_--) {				\
        struct methodblock *mb = _mt_[_i_];				\
	if (mb->fb.access & ACC_MIRANDA_METHOD) {                       \
	    /* See ResolveInterfaces() in classresolver.c for details*/ \
	    /* Skip miranda methods generated for non-public methods */ \
            if (mb->fb.name[0] == '+') continue;                        \
	    /* Get interface method block. */                           \
	    mb = (struct methodblock *)mb->code;                        \
	}								\
	body;								\
    }									\
} else ((void) 0)

#define	WALK_SUPER_METHODS(cb, mb, body)				\
if (TRUE)								\
{									\
    ClassClass *_cb_;    						\
    for (_cb_ = (cb);  _cb_ != NULL; _cb_ = cbSuperclass(_cb_)) {       \
	WALK_DECLARED_METHODS(_cb_, mb, body);				\
    }									\
} else ((void) 0)

#define	WALK_INTERFACE_METHODS(cb, mb, body)				\
if (TRUE)								\
{									\
    struct imethodtable *_imt_ = cbIntfMethodTable(cb);			\
    int _icnt_ = _imt_->icount, _j_;					\
    /* Walk from subinterface towards superinterface */			\
    for (_j_ = 0; _j_ < _icnt_; _j_++) {				\
        ClassClass * _cb_ = _imt_->itable[_j_].classdescriptor;		\
        WALK_DECLARED_METHODS(_cb_, mb, body);				\
    }									\
} else ((void) 0)

HArrayOfObject *
reflect_methods(
    ExecEnv *ee,
    ClassClass * cb,
    int which)
{
    HArrayOfObject *harr;

    if (cbIsPrimitive(cb))
	return (HArrayOfObject *)
	    reflect_new_array(ee, classJavaLangReflectMethod(), 0);

    ENSURE_LINKED(ee, cb);			/* was ENSURE_INITIALIZED */

    switch (which) {

    case MEMBER_PUBLIC:
    {
	int isInterface = cbIsInterface(cb);
	int cnt = 0, j;
	HObject **arr;

	if (isInterface) {

	    /* All interface methods are public */
	    WALK_INTERFACE_METHODS(cb, mb, cnt++);

	} else {

	    /* Count public instance methods */
	    WALK_INSTANCE_METHODS(cb, mb, {
		if (mbIsPublic(mb)
		    && !mbIsSpecial(mb))
		    cnt++;
	    });

	    /* Count public static methods */
	    WALK_SUPER_METHODS(cb, mb, {
		if (mbIsPublic(mb)
		    && mbIsStatic(mb)
		    && !mbIsSpecial(mb))
		    cnt++;
	    });

	}

	/* Allocate array */
	if ((harr = (HArrayOfObject *)
	    reflect_new_array(ee, classJavaLangReflectMethod(), cnt)) == NULL)
	    return NULL;			/* exception */
	arr = (HObject **) unhand(harr)->body;

	/* Fill array in declaration order */
	j = cnt;

	if (isInterface) {

	    /* All interface methods are public */
	    WALK_INTERFACE_METHODS(cb, mb, {
		if ((arr[--j] = new_method(ee, mb)) == NULL)
		    return NULL;		/* exception */
	    });

	} else {

	    /* Gather public instance methods */
	    WALK_INSTANCE_METHODS(cb, mb, {
		if (mbIsPublic(mb)
		    && !mbIsSpecial(mb))
		    if ((arr[--j] = new_method(ee, mb)) == NULL)
			return NULL;		/* exception */
	    });

	    /* Gather public static methods */
	    WALK_SUPER_METHODS(cb, mb, {
		if (mbIsPublic(mb)
		    && mbIsStatic(mb)
		    && !mbIsSpecial(mb))
		    if ((arr[--j] = new_method(ee, mb)) == NULL)
			return NULL;		/* exception */
	    });

	}
	KEEP_POINTER_ALIVE(arr);

	sysAssert(j == 0);
	return harr;
    }

    case MEMBER_DECLARED:
    {
	int cnt = 0, j;
	HObject **arr;

	/* Filter out static initializers and constructors */
	WALK_DECLARED_METHODS(cb, mb, {
	    if (!mbIsSpecial(mb))
		cnt++;
	});

	/* Allocate array */
	if ((harr = (HArrayOfObject *)
	    reflect_new_array(ee, classJavaLangReflectMethod(), cnt)) == NULL)
	    return NULL;			/* exception */
	arr = (HObject **) unhand(harr)->body;

	/* Fill in declared order */
	j = cnt;

	/* Filter out static initializers and constructors */
	WALK_DECLARED_METHODS(cb, mb, {
	    if (!mbIsSpecial(mb))
		if ((arr[--j] = new_method(ee, mb)) == NULL)
		    return NULL;		/* exception */
	});
	KEEP_POINTER_ALIVE(arr);

	sysAssert(j == 0);
	return harr;
    }

    default:
	ThrowInternalError(ee, "reflect methods");
	return NULL;
    }
}

/*
 * Parameter counts have already been matched, and cnt > 0.
 */

static bool_t
match_parameter_types(
    struct methodblock *mb,
    HArrayOfObject *hclasses,
    int cnt)
{
    char *p = mbSig(mb) + 1;			/* skip SIGNATURE_FUNC */
    ClassClass *cb = mbClass(mb);		/* declaring class */
    Hjava_lang_Class **classes = (Hjava_lang_Class **)unhand(hclasses)->body;
    int i = 0;
    ClassClass *pcb;

    while (*p != SIGNATURE_ENDFUNC) {
	/*
	 * The match must be exact, not by name.  This can be slow...
	 */
	if ((pcb = reflect_find_class(p, cb, &p)) == NULL)
	    return FALSE;			/* exception */
	if (classes[i++] != cbHandle(pcb))	/* XXX match on handles */
	    return FALSE;			/* match failed */
    }
    KEEP_POINTER_ALIVE(classes);
    return TRUE;
}

HObject *
reflect_method(
    ExecEnv	*ee,
    ClassClass	*cb,
    char	*name,
    HArrayOfObject *types,
    int which)
{
    char *hashed_name = HashUTF8(ee, name);
    int tcnt = array_length(types);		/* handles NULL */
    int pcnt;

    if (cbIsPrimitive(cb)
	|| (hashed_name == NULL))
	goto nosuchmethod;

    ENSURE_LINKED(ee, cb);			/* was ENSURE_INITIALIZED */

    switch (which) {

    case MEMBER_PUBLIC: {

	if (cbIsInterface(cb)) {

	    WALK_INTERFACE_METHODS(cb, mb, {
		if ((hashed_name == mbName(mb))
		    && (tcnt == (pcnt = get_parameter_count(mbSig(mb))))
		    && (pcnt == 0 || match_parameter_types(mb, types, pcnt)))
		    return new_method(ee, mb);
	    });

	    goto nosuchmethod;
	}

	WALK_INSTANCE_METHODS(cb, mb, {
	    if (mbIsPublic(mb)
		&& !mbIsSpecial(mb)
		&& (hashed_name == mbName(mb))
		&& (tcnt == (pcnt = get_parameter_count(mbSig(mb))))
		&& (pcnt == 0 || match_parameter_types(mb, types, pcnt)))
		return new_method(ee, mb);
	});

	WALK_SUPER_METHODS(cb, mb, {
	    if (mbIsPublic(mb)
		&& mbIsStatic(mb)
		&& !mbIsSpecial(mb)
		&& (hashed_name == mbName(mb))
		&& (tcnt == (pcnt = get_parameter_count(mbSig(mb))))
		&& (pcnt == 0 || match_parameter_types(mb, types, pcnt)))
		return new_method(ee, mb);
	});

	break;
    }

    case MEMBER_DECLARED: {

	WALK_DECLARED_METHODS(cb, mb, {
	    if (!mbIsSpecial(mb)
		&& (hashed_name == mbName(mb))
		&& (tcnt == (pcnt = get_parameter_count(mbSig(mb))))
		&& (pcnt == 0 || match_parameter_types(mb, types, pcnt)))
		return new_method(ee, mb);
	});

	break;
    }

    default:
	break;

    }

  nosuchmethod:
    ThrowNoSuchMethodException(ee, name);
    return NULL;
}

/*
 *
 */

static HObject *
new_constructor(ExecEnv *ee, struct methodblock *mb)
{
    ClassClass *cb;
    long slot;
    char *p;
    HArrayOfObject *parameterTypes, *exceptionTypes;
    HString *hname;
    Hjava_lang_reflect_Constructor *hc;

    /* declaring class */
    cb = mbClass(mb);

    /* slot */
    slot = GET_METHODBLOCK_SLOT(cb, mb);
    sysAssert((slot >= 0) && (slot < cbMethodsCount(cb)));

    /* parameter types from signature */
    p = mbSig(mb);
    if ((parameterTypes = get_parameter_types(ee, p, cb, &p)) == NULL)
	return NULL;				/* exception */
    sysAssert(*(p-1) == SIGNATURE_ENDFUNC);

    if ((exceptionTypes = get_exception_types(ee, cb, mb)) == NULL)
	return NULL;				/* exception */

    if ((hname = makeJavaStringUTF(mbName(mb))) == NULL)
	return NULL;				/* exception */

    if ((hc = (Hjava_lang_reflect_Constructor *)
	allocObject(ee, classJavaLangReflectConstructor())) == NULL) {
	ThrowOutOfMemoryError(ee, 0);
	return NULL;
    }

    /* initialize the Constructor */
    member_class(hc) = cbHandle(cb);
    member_slot(hc) = slot;
    member_parameterTypes(hc) = parameterTypes;
    member_exceptionTypes(hc) = exceptionTypes;
    member_modifiers(hc) = mbAccess(mb) & ACC_WRITTEN_FLAGS;
    member_override(hc) = FALSE;

    return (HObject *)hc;
}

HArrayOfObject *
reflect_constructors(
    ExecEnv *ee,
    ClassClass * cb,
    int which)
{
    HArrayOfObject *harr;
    HObject **arr;
    int pub = (which == MEMBER_PUBLIC);
    int cnt = 0, j;

    /* Exclude primitive, interface and array types */
    if (cbIsPrimitive(cb)
	|| cbIsInterface(cb)
	|| cbName(cb)[0] == SIGNATURE_ARRAY)
	return (HArrayOfObject *)
	    reflect_new_array(ee, classJavaLangReflectConstructor(), 0);

    ENSURE_LINKED(ee, cb);			/* was ENSURE_INITIALIZED */
    
    WALK_DECLARED_METHODS(cb, mb, {
	if (mbIsSpecial(mb)
	    && (!pub || mbIsPublic(mb))
	    && (mbName(mb) == utf8_literal_obj_init_name))
	    cnt++;
    });

    /* Allocate array */
    if ((harr = (HArrayOfObject *)
	reflect_new_array(ee, classJavaLangReflectConstructor(), cnt)) == NULL)
	return NULL;				/* exception */
    arr = (HObject **) unhand(harr)->body;

    /* Fill in declared order */
    j = cnt;

    WALK_DECLARED_METHODS(cb, mb, {
	if (mbIsSpecial(mb)
	    && (!pub || mbIsPublic(mb))
	    && (mbName(mb) == utf8_literal_obj_init_name))
	    if ((arr[--j] = new_constructor(ee, mb)) == NULL)
		return NULL;			/* exception */
    });
    KEEP_POINTER_ALIVE(arr);

    sysAssert(j == 0);
    return harr;
}

HObject *
reflect_constructor(
    ExecEnv *ee,
    ClassClass *cb,
    HArrayOfObject *types,
    int which)
{
    int tcnt, pcnt;
    int pub = (which == MEMBER_PUBLIC);

    /* Exclude primitive, interface and array types */
    if (cbIsPrimitive(cb)
	|| cbIsInterface(cb)
	|| cbName(cb)[0] == SIGNATURE_ARRAY)
	goto nosuchmethod;

    ENSURE_LINKED(ee, cb);			/* was ENSURE_INITIALIZED */

    tcnt = array_length(types);

    WALK_DECLARED_METHODS(cb, mb, {
	if (mbIsSpecial(mb)
	    && (!pub || mbIsPublic(mb))
	    && (mbName(mb) == utf8_literal_obj_init_name)
	    && (tcnt ==	(pcnt = get_parameter_count(mbSig(mb))))
	    && (pcnt == 0 || match_parameter_types(mb, types, pcnt)))
	    return new_constructor(ee, mb);
    });


  nosuchmethod:
    ThrowNoSuchMethodException(ee, utf8_literal_obj_init_name);
    return NULL;
}

HObject * 
reflect_new_method(ExecEnv *ee, struct methodblock *mb)
{
    /* XXX what about <clinit>? */
    if (mbName(mb)[0] == '<')
        return new_constructor(ee, mb);
    else 
        return new_method(ee, mb);
}

struct methodblock *
reflect_get_methodblock(HObject *obj)
{
    struct Hjava_lang_reflect_Method *hm =
        (struct Hjava_lang_reflect_Method *)obj;
    ClassClass *cb = member_class(hm);
    return GET_CLASS_METHODBLOCK(cb, member_slot(hm));
}

/*
 * Networking library support
 */

extern GetInterfaceFunc GetHPI;

JNIEXPORT jint JNICALL
JVM_InitializeSocketLibrary(void)
{
    return GetHPI((void **)&hpi_socket_interface, "Socket", 1);
}

JNIEXPORT jint JNICALL
JVM_Socket(jint domain, jint type, jint protocol) {
    return sysSocket(domain, type, protocol);
}

JNIEXPORT jint JNICALL
JVM_SocketClose(jint fd) {
    return sysSocketClose(fd);
}

JNIEXPORT jint JNICALL
JVM_Recv(jint fd, char *buf, jint nBytes, jint flags) {
    return sysRecv(fd, buf, nBytes, flags);
}

JNIEXPORT jint JNICALL
JVM_Send(jint fd, char *buf, jint nBytes, jint flags) {
    return sysSend(fd, buf, nBytes, flags);
}

JNIEXPORT jint JNICALL
JVM_Timeout(int fd, long timeout) {
    return sysTimeout(fd, timeout);
}

JNIEXPORT jint JNICALL
JVM_Listen(jint fd, jint count) {
    return sysListen(fd, count);
}

JNIEXPORT jint JNICALL
JVM_Connect(jint fd, struct sockaddr *him, jint len) {
    return sysConnect(fd, him, len);
}

JNIEXPORT jint JNICALL
JVM_Accept(jint fd, struct sockaddr *him, jint *len) {
    return sysAccept(fd, him, (int *)len);
}

JNIEXPORT jint JNICALL
JVM_RecvFrom(jint fd, char *buf, int nBytes, 
		  int flags, struct sockaddr *from, int *fromlen) {
    return sysRecvFrom(fd, buf, nBytes, flags, from, fromlen);
}

/*
 * Routines to do datagram
 */
JNIEXPORT jint JNICALL
JVM_SendTo(jint fd, char *buf, int len, 
		int flags, struct sockaddr *to, int tolen) {
    return sysSendTo(fd, buf, len, flags, to, tolen);
}

JNIEXPORT jint JNICALL
JVM_SocketAvailable(jint fd, jint *result) {
    return sysSocketAvailable(fd, (long *)result);
}

/*
 * Note that this function expects a full path name.  If the file
 * doesn't exists, then it is going to throw an exception.  In case
 * you are managing LD_LIBRARY_PATH yourself, like
 * ClassLoader.NativeLibrary does, then you better deal with the
 * exception correctly.
 */
JNIEXPORT void * JNICALL 
JVM_LoadLibrary(const char *name)
{
    char load_error[STK_BUF_LEN*2];
    void *res = sysLoadLibrary(name, load_error, sizeof(load_error));
    if (res && verbose_jni) {
	jio_fprintf(stderr, "[Loaded native library: %s]\n", name);
    }
    if (res == NULL) {
	char msg[STK_BUF_LEN*2];
	jio_snprintf(msg, sizeof(msg), "%s: %s", name, load_error);
	ThrowUnsatisfiedLinkError(NULL, msg);
    }
    return res;
}

JNIEXPORT void JNICALL 
JVM_UnloadLibrary(void * handle)
{
   sysUnloadLibrary(handle);
}

JNIEXPORT void * JNICALL 
JVM_FindLibraryEntry(void * handle, const char *name)
{
    return sysFindLibraryEntry(handle, name);
}

JNIEXPORT jboolean JNICALL
JVM_IsNaN(jdouble d)
{
    return IS_NAN(d);
}

JNIEXPORT jboolean JNICALL
JVM_IsSupportedJNIVersion(jint version)
{
    return version == 0x00010001 || version == 0x00010002;
}

JNIEXPORT jstring JNICALL
JVM_InternString(JNIEnv *env, jstring this)
{
    Hjava_lang_String *str = 
        internString((Hjava_lang_String*)DeRef(env, this));
    if (str == 0) {
        ThrowOutOfMemoryError(0, 0);
	return NULL;
    }
    return MkRefLocal(env, str);
}

JNIEXPORT void * JNICALL
JVM_RawMonitorCreate(void)
{
    sys_mon_t *mon = sysMalloc(sysMonitorSizeof());
    if (mon == NULL) {
        return NULL;
    }
    sysMonitorInit(mon);
    return mon;
}

JNIEXPORT void JNICALL
JVM_RawMonitorDestroy(void *mon)
{
    sysMonitorDestroy(mon);
    sysFree(mon);
}

JNIEXPORT jint JNICALL
JVM_RawMonitorEnter(void *mon)
{
    return sysMonitorEnter(sysThreadSelf(), mon);
}

JNIEXPORT void JNICALL
JVM_RawMonitorExit(void *mon)
{
    sysMonitorExit(sysThreadSelf(), mon);
}

JNIEXPORT void * JNICALL
JVM_GetThreadInterruptEvent(void)
{
    return sysThreadInterruptEvent();
}
