/*
 * @(#)jvmdi.c	1.55 98/09/22
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Breakpoints, single-stepping and debugger event notification.
 * Debugger event notification/handling includes breakpoint, single-step
 * and exceptions.
 */

#include <sys/types.h>
#include <string.h>
#include <stdarg.h>

#include "bool.h"
#include "opcodes.h"
#include "oobj.h"
#include "sys_api.h"
#include "interpreter.h"
#include "common_exceptions.h"
#include "threads.h"
#include "gc.h"
#include "breakpoints.h"
#include "monitor_cache.h"
#include "jni.h"
#define NO_JVMDI_MACROS 1
#include "jvmdi.h"
#include "bag.h"
#include "jlong.h"

/* sole interface instance */
static JVMDI_Interface_1 interface;

/* this VM */
static JavaVM *vm;


/* Defined in inline.c thru an include of "opcodes.length". This is dangerous
   as is. There should be a header file with this extern shared by both .c files */
extern const short opcode_length[256];

/*
 * Set up a lock to protect JVMDI resources. The lock defined here must
 * be acquired by lock_for_debugger. For now at least, we just reuse an 
 * existing VM lock. Use of this lock will not interfere with VM 
 * operations in any important way. 
 */
#define DEBUGGER_LOCK(thread) BINCLASS_LOCK(thread)
#define DEBUGGER_UNLOCK(thread) BINCLASS_UNLOCK(thread)

#ifdef BREAKPTS

static JVMDI_EventHook eventHook = NULL;

static JVMDI_AllocHook allocHook = NULL;
static JVMDI_DeallocHook deallocHook = NULL;

#define INITIAL_BAG_SIZE 4

struct bkpt {
    unsigned char *pc;      /* key - must be first */
    unsigned char opcode;   /* opcode to restore */
    jobject classRef;       /* Prevents enclosing class from being gc'ed */
};

static struct bag *breakpoints;

struct fpop {
    JavaFrame *frame;       /* key - must be first */
    unsigned char *returnpc;
};

static struct bag *framePops;

struct fieldWatch {
    struct fieldblock *fb;  /* field to watch; key - must be first */
    jclass classRef;        /* Prevents enclosing class from being gc'ed */
};

static struct bag *watchedFieldModifications;
static struct bag *watchedFieldAccesses;

/* are one or more fields being watched?
 * these flags are accessed by the interpreter to determine if
 * jvmdi should be notified.
 */
bool_t watching_field_access = FALSE;  
bool_t watching_field_modification = FALSE;  

/* additional local frames to push to cover usage by client debugger */
#define LOCAL_FRAME_SLOP 10

typedef struct ThreadNode_ {
    JHandle *thread;
    jobject lastDetectedException;
    jboolean eventEnabled[JVMDI_MAX_EVENT_TYPE_VAL+1];
    JVMDI_StartFunction startFunction;  /* for debug threads only */
    void *startFunctionArg;             /* for debug threads only */
    struct ThreadNode_ *next;
} ThreadNode;
static volatile ThreadNode *threadList = NULL;

#define JVMDI_EVENT_GLOBAL_MASK 0xf0000000
#define JVMDI_EVENT_THREAD_MASK 0x7fffffff

static uint32_t eventEnable[JVMDI_MAX_EVENT_TYPE_VAL+1];

static void enableAllEvents(jint eventType, jboolean enabled) {
    if (enabled) {
        eventEnable[eventType] |=  JVMDI_EVENT_GLOBAL_MASK;
    } else {
        eventEnable[eventType] &=  ~JVMDI_EVENT_GLOBAL_MASK;
    }
}

static void enableThreadEvents(ThreadNode *node, 
                               jint eventType, jboolean enabled) {
    sys_thread_t *self = sysThreadSelf();
    uint32_t count;

    DEBUGGER_LOCK(self);

    /*
     * If state is changing, update the global eventEnable word
     * for this thread and the per-thread boolean flag.
     */
    if (node->eventEnabled[eventType] != enabled) {
        node->eventEnabled[eventType] = enabled;

        count = eventEnable[eventType] & JVMDI_EVENT_THREAD_MASK;
        if (enabled) {
            count++;
        } else {
            count--;
        }
        eventEnable[eventType] = 
            (JVMDI_EVENT_GLOBAL_MASK & eventEnable[eventType]) |
            (JVMDI_EVENT_THREAD_MASK & count);
    }

    DEBUGGER_UNLOCK(self);
}

static jboolean threadEnabled(jint eventType, ThreadNode *node) {
    return (node == NULL) ? JNI_FALSE : node->eventEnabled[eventType];
}

#define GLOBALLY_ENABLED(eventType) \
                 (eventEnable[eventType] & JVMDI_EVENT_GLOBAL_MASK)

/*
 * This macro is used in notify_debugger_* to determine whether a JVMDI 
 * event should be generated. Note that this takes an expression for the thread node
 * which we should avoid evaluating twice. That's why threadEnabled above
 * is a function, not a macro
 *
 * Also the first eventEnable[eventType] is redundant, but it serves as
 * a quick filter of events that are completely unreported.
 */
#define MUST_NOTIFY(eventType, threadNodeExpr) \
             (eventEnable[eventType] && \
              eventHook && \
              ( GLOBALLY_ENABLED(eventType) || \
                threadEnabled(eventType, threadNodeExpr)))

/* forward defs */
static jvmdiError JNICALL
jvmdi_Allocate(jlong size, jbyte** memPtr);
static jvmdiError JNICALL
jvmdi_Deallocate(jbyte* mem);

static jvmdiError
initializeJVMDI() 
{
    breakpoints = bagCreateBag(sizeof(struct bkpt), INITIAL_BAG_SIZE);
    framePops = bagCreateBag(sizeof(struct fpop), INITIAL_BAG_SIZE);
    watchedFieldModifications = 
        bagCreateBag(sizeof(struct fieldWatch), INITIAL_BAG_SIZE);
    watchedFieldAccesses = 
        bagCreateBag(sizeof(struct fieldWatch), INITIAL_BAG_SIZE);
    if (breakpoints == NULL || framePops == NULL || 
                     watchedFieldModifications == NULL || 
                     watchedFieldAccesses == NULL) {
        return JVMDI_ERROR_OUT_OF_MEMORY;
    }
    /*
     * Setup the events to be enabled/disabled on startup. All events
     * are enabled except single step, exception catch, method entry/exit.
     */
    enableAllEvents(JVMDI_EVENT_THREAD_START, JNI_TRUE);
    enableAllEvents(JVMDI_EVENT_THREAD_END, JNI_TRUE);
    enableAllEvents(JVMDI_EVENT_CLASS_LOAD, JNI_TRUE);
    enableAllEvents(JVMDI_EVENT_CLASS_PREPARE, JNI_TRUE);
    enableAllEvents(JVMDI_EVENT_CLASS_UNLOAD, JNI_TRUE);
    enableAllEvents(JVMDI_EVENT_FIELD_ACCESS, JNI_TRUE);
    enableAllEvents(JVMDI_EVENT_FIELD_MODIFICATION, JNI_TRUE);
    enableAllEvents(JVMDI_EVENT_BREAKPOINT, JNI_TRUE);
    enableAllEvents(JVMDI_EVENT_FRAME_POP, JNI_TRUE);
    enableAllEvents(JVMDI_EVENT_EXCEPTION, JNI_TRUE);
    enableAllEvents(JVMDI_EVENT_USER_DEFINED, JNI_TRUE);

    return JVMDI_ERROR_NONE;
}

/*
 * These functions maintain the linked list of currently running threads. 
 */
static ThreadNode *
findThread(JHandle *thread) {
    ThreadNode *node;
    sys_thread_t *self = sysThreadSelf();

    DEBUGGER_LOCK(self);

    node = (ThreadNode *)threadList;  /* cast away volatility */
    while (node != NULL) {
        if (node->thread == thread) {
            break;
        }
        node = node->next;
    }
    DEBUGGER_UNLOCK(self);

    return node;
}

static ThreadNode *
insertThread(JHandle *thread) {
    ThreadNode *node;
    sys_thread_t *self = sysThreadSelf();

    DEBUGGER_LOCK(self);

    node = (ThreadNode *)malloc(sizeof(*node));
    if (node != NULL) {
        node->thread = thread;
        node->lastDetectedException = NULL;
        memset(node->eventEnabled, 0, sizeof(node->eventEnabled));
        node->startFunction = NULL;
        node->startFunctionArg = NULL;
        node->next = (ThreadNode *)threadList; /* cast away volatility */
        threadList = node;
    } 
    DEBUGGER_UNLOCK(self);

    return node;
}

static jboolean 
removeThread(JHandle *thread) {
    ThreadNode *previous = NULL;
    ThreadNode *node; 
    sys_thread_t *self = sysThreadSelf();
    jboolean rc = JNI_FALSE;

    DEBUGGER_LOCK(self);

    node = (ThreadNode *)threadList;  /* cast away volatility */
    while (node != NULL) {
        if (node->thread == thread) {
            int i;
            if (previous == NULL) {
                threadList = node->next;
            } else {
                previous->next = node->next;
            }
            for (i = 0; i <= JVMDI_MAX_EVENT_TYPE_VAL; i++) {
                enableThreadEvents(node, i, JNI_FALSE);
            }
            free((void *)node);
            rc = JNI_TRUE;
            break;
        }
        previous = node;
        node = node->next;
    }
    DEBUGGER_UNLOCK(self);

    return rc;
}


static void
reportException(ExecEnv *ee, unsigned char *pc,
                HObject *object, struct javaframe *jframe) {
    JVMDI_Event event;
    JNIEnv *env = EE2JNIEnv(ee);
    struct methodblock *mb = jframe->current_method;

    if (mb == NULL) {
        return;
    }

    event.kind = JVMDI_EVENT_EXCEPTION;
    event.u.exception.thread = MkRefLocal(env, ee->thread);
    event.u.exception.clazz = MkRefLocal(env, mb->fb.clazz);
    event.u.exception.method = (jmethodID)mb;
    event.u.exception.location = ptrdiff_to_jlong(pc - mb->code);
    event.u.exception.exception = MkRefLocal(env, object);
    event.u.exception.catch_clazz = 0;
    event.u.exception.catch_method = 0;
    event.u.exception.catch_location = jlong_zero;

    /* walk up the stack to see if this exception is caught anywhere. */
    while (jframe != NULL) {
        struct methodblock *cmb = jframe->current_method;
        if ((cmb != NULL) && !(cmb->fb.access & ACC_NATIVE)) { /* skip C frames */
            unsigned char *cpc = 
                  ProcedureFindThrowTag(ee, jframe, object, jframe->lastpc);
            if (cpc != NULL) {
                event.u.exception.catch_clazz = 
                          (jclass)MkRefLocal(env, cmb->fb.clazz);
                event.u.exception.catch_method = (jmethodID)cmb;
                event.u.exception.catch_location =
			   ptrdiff_to_jlong(cpc - cmb->code);
                break;
            }
        }
        jframe = jframe->prev;
    }

    eventHook(env, &event);
}

void
notify_debugger_of_exception(ExecEnv *ee, unsigned char *pc, HObject *object)
{
    JNIEnv *env = EE2JNIEnv(ee);
    HObject *last;
    ThreadNode *threadNode;

    if (ee->thread == NULL) {
        return;
    }

    threadNode = findThread(ee->thread);
    if (threadNode == NULL) {
        /* ignore any exceptions before thread start */
        return;
    }

    if (threadNode->lastDetectedException != NULL) {
        last = (HObject *)DeRef(env, threadNode->lastDetectedException);
    } else {
        last = NULL;
    }

    if (object != last) {
        /* Get current frame before another one is pushed */
        struct javaframe *jframe = ee->current_frame;

        if ((*env)->PushLocalFrame(env, 5+LOCAL_FRAME_SLOP) < 0) {
            return;
        }
        
        if (last != NULL) {
            (*env)->DeleteGlobalRef(env, threadNode->lastDetectedException);
        }
        /*
         * This needs to be a global ref; otherwise, the detected
         * exception could be cleared and collected in a native method
         * causing later comparisons to be invalid.
         */
        threadNode->lastDetectedException = (*env)->NewGlobalRef(env,
                                                 MkRefLocal(env, object));

        if (MUST_NOTIFY(JVMDI_EVENT_EXCEPTION, threadNode)) {
            reportException(ee, pc, object, jframe);
        }

        (*env)->PopLocalFrame(env, 0);
    }
}

/*
 * This function is called by the interpreter whenever:
 * 1) The interpreter catches an exception, or
 * 2) The interpreter detects that a native method has returned
 *    without an exception set (i.e. the native method *may* have
 *    cleared an exception; there's no way to tell for sure)
 *
 * This function performs 2 tasks. It removes any exception recorded
 * as the last detected exception by notify_debugger_of_exception. It
 * also reports an event to the JVMDI client. (2) above implies that 
 * we need to make sure that an exception was actually cleared before
 * reporting the event. This 
 * can be done by checking whether there is currently a last detected
 * detected exception value saved for this thread.
 */
void
notify_debugger_of_exception_catch(ExecEnv *ee, unsigned char *pc, HObject *object)
{
    JNIEnv *env = EE2JNIEnv(ee);
    ThreadNode *threadNode;
    
    if (ee->thread == NULL) {
        return;
    }

    threadNode = findThread(ee->thread);
    if (threadNode == NULL) {
        /* ignore any exceptions before thread start */
        return;
    }

    if (MUST_NOTIFY(JVMDI_EVENT_EXCEPTION_CATCH, threadNode)) {
        JVMDI_Event event;
        JNIEnv *env = EE2JNIEnv(ee);
        struct javaframe *jframe = ee->current_frame;
        struct methodblock *mb = jframe->current_method;

        if (mb == NULL) {
            return;
        }

        /*
         * Report the caught exception if it is being caught in Java code
         * or if it was caught in native code after its throw was reported 
         * earlier.
         */
        if ((object != NULL) || 
            (threadNode->lastDetectedException != NULL)) {
            
            if ((*env)->PushLocalFrame(env, 3+LOCAL_FRAME_SLOP) < 0) {
                return;
            }
            
            event.kind = JVMDI_EVENT_EXCEPTION_CATCH;
            event.u.exception_catch.thread = MkRefLocal(env, ee->thread);
            event.u.exception_catch.clazz = MkRefLocal(env, mb->fb.clazz);
            event.u.exception_catch.method = (jmethodID)mb;
            event.u.exception_catch.location =
		ptrdiff_to_jlong(pc - mb->code);
            if (object != NULL) {
                event.u.exception_catch.exception = MkRefLocal(env, object);
            } else {
                event.u.exception_catch.exception = NULL;
            }
    
            eventHook(env, &event);
            (*env)->PopLocalFrame(env, 0);
        }
    }
    if (threadNode->lastDetectedException != NULL) {
        (*env)->DeleteGlobalRef(env, threadNode->lastDetectedException);
    }
    threadNode->lastDetectedException = NULL;
}

void
notify_debugger_of_single_step(ExecEnv *ee, unsigned char *pc)
{
    /*
     * The interpreter notifies us only for threads that have stepping enabled,
     * so we don't do any checks of the global or per-thread event
     * enable flags here.
     */

    if (eventHook) {
        JVMDI_Event event;
	struct javaframe *jframe = ee->current_frame;
	struct methodblock *mb = jframe->current_method;
	JNIEnv *env = EE2JNIEnv(ee);

	if (mb == NULL) {
	    return;
	}

	if ((*env)->PushLocalFrame(env, 2+LOCAL_FRAME_SLOP) < 0) {
	    return;
        }
	event.kind = JVMDI_EVENT_SINGLE_STEP;
	event.u.single_step.thread = MkRefLocal(env, ee->thread);
	event.u.single_step.clazz = MkRefLocal(env, mb->fb.clazz);
	event.u.single_step.method = (jmethodID)mb;
	event.u.single_step.location = ptrdiff_to_jlong(pc - mb->code);

	eventHook(env, &event);

	(*env)->PopLocalFrame(env, 0);
    }
}

void
notify_debugger_of_breakpoint(ExecEnv *ee, unsigned char *pc)
{
    if (MUST_NOTIFY(JVMDI_EVENT_BREAKPOINT, findThread(ee->thread))) {
        JVMDI_Event event;
	struct javaframe *jframe = ee->current_frame;
	struct methodblock *mb = jframe->current_method;
	JNIEnv *env = EE2JNIEnv(ee);

	if (mb == NULL) {
	    return;
	}
	if ((*env)->PushLocalFrame(env, 2+LOCAL_FRAME_SLOP) < 0) {
	    return;
        }
	event.kind = JVMDI_EVENT_BREAKPOINT;
	event.u.breakpoint.thread = MkRefLocal(env, ee->thread);
	event.u.breakpoint.clazz = MkRefLocal(env, mb->fb.clazz);
	event.u.breakpoint.method = (jmethodID)mb;
	event.u.breakpoint.location = ptrdiff_to_jlong(pc - mb->code);

	eventHook(env, &event);

	(*env)->PopLocalFrame(env, 0);
    }
}

void
notify_debugger_of_thread_start(ExecEnv *ee, JHandle *thread)
{
    JNIEnv *env = EE2JNIEnv(ee);

    /*
     * Look for existing thread info for this thread. If there is 
     * a ThreadNode already, it just means that it's a debugger thread
     * started by jvmdi_RunDebugThread; if not, we create the ThreadNode
     * here. 
     */
    ThreadNode *node = findThread(thread);
    if (node == NULL) {
        node = insertThread(thread);
        if (node == NULL) {
            (*env)->FatalError(env, "internal allocation error in JVMDI");
        }
    }

    if (eventHook && eventEnable[JVMDI_EVENT_THREAD_START]) {
        JVMDI_Event event;

        if ((*env)->PushLocalFrame(env, 1) < 0) {
            return;
        }
        event.kind = JVMDI_EVENT_THREAD_START;
        event.u.thread_change.thread = MkRefLocal(env, thread);

        eventHook(env, &event);

        (*env)->PopLocalFrame(env, 0);
    }
}

void
notify_debugger_of_thread_end(ExecEnv *ee, JHandle *thread)
{
    JNIEnv *env = EE2JNIEnv(ee);
    if (MUST_NOTIFY(JVMDI_EVENT_THREAD_END, findThread(thread))) {
        JVMDI_Event event;

        if ((*env)->PushLocalFrame(env, 1+LOCAL_FRAME_SLOP) < 0) {
            goto forgetThread;
        }
        event.kind = JVMDI_EVENT_THREAD_END;
        event.u.thread_change.thread = MkRefLocal(env, thread);

        eventHook(env, &event);

        (*env)->PopLocalFrame(env, 0);
    }

forgetThread:
    if (removeThread(thread) == JNI_FALSE) {
        (*env)->FatalError(env, "internal error in JVMDI (ending unstarted thread)");
    }
}

void
notify_debugger_of_field_access(ExecEnv *ee, JHandle *obj,
                                struct fieldblock *fb)
{
    sys_thread_t *self = sysThreadSelf();
    struct fieldWatch *fwfb;

    DEBUGGER_LOCK(self);
    fwfb = bagFind(watchedFieldAccesses, fb);
    DEBUGGER_UNLOCK(self);          

    if (fwfb != NULL && MUST_NOTIFY(JVMDI_EVENT_FIELD_ACCESS, 
                                    findThread(ee->thread))) {
        JVMDI_Event event;
        struct javaframe *jframe = ee->current_frame;
        struct javaframe frame_buf;
        struct methodblock *mb;
        jlocation location;
        JNIEnv *env = EE2JNIEnv(ee);

        /* We may have come through native code - remove any
         * empty frames.
         */
        jframe = getCallerFrame(jframe, 0, &frame_buf);
        if (jframe == NULL || (mb = jframe->current_method) == NULL) {
            return;
        }
        if (mb->fb.access & ACC_NATIVE) {
            location = (jlocation)(-1);
        } else {
            location = (jlocation)(jframe->lastpc - mb->code);
        }

        if ((*env)->PushLocalFrame(env, 4+LOCAL_FRAME_SLOP) < 0) {
            return;
        }
        event.kind = JVMDI_EVENT_FIELD_ACCESS;
        event.u.field_access.thread = MkRefLocal(env, ee->thread);
        event.u.field_access.clazz = MkRefLocal(env, mb->fb.clazz);
        event.u.field_access.method = (jmethodID)mb;
        event.u.field_access.location = location;
        event.u.field_access.field_clazz = MkRefLocal(env, fb->clazz);
        event.u.field_access.field = (jfieldID)fb;
        event.u.field_access.object =  obj == NULL? 
                                                NULL : 
                                                MkRefLocal(env, obj);

        eventHook(env, &event);

        (*env)->PopLocalFrame(env, 0);
    }
}

void
notify_debugger_of_field_modification(ExecEnv *ee, JHandle *obj,
                                      struct fieldblock *fb, jvalue jval)
{
    sys_thread_t *self = sysThreadSelf();
    struct fieldWatch *fwfb;

    DEBUGGER_LOCK(self);
    fwfb = bagFind(watchedFieldModifications, fb);
    DEBUGGER_UNLOCK(self);          

    if (fwfb != NULL && MUST_NOTIFY(JVMDI_EVENT_FIELD_MODIFICATION, 
                                    findThread(ee->thread))) {
        JVMDI_Event event;
        struct javaframe *jframe = ee->current_frame;
        struct javaframe frame_buf;
        struct methodblock *mb;
        jlocation location;
        JNIEnv *env = EE2JNIEnv(ee);
        char sig_type = fieldsig(fb)[0];

        /* We may have come through native code - remove any
         * empty frames.
         */
        jframe = getCallerFrame(jframe, 0, &frame_buf);
        if (jframe == NULL || (mb = jframe->current_method) == NULL) {
            return;
        }
        if (mb->fb.access & ACC_NATIVE) {
            location = (jlocation)(-1);
        } else {
            location = (jlocation)(jframe->lastpc - mb->code);
        }

        if ((*env)->PushLocalFrame(env, 5+LOCAL_FRAME_SLOP) < 0) {
            return;
        }
        if (sig_type == SIGNATURE_CLASS || sig_type == SIGNATURE_ARRAY) {
            jval.l = MkRefLocal(env, jval.l);
        }
        event.kind = JVMDI_EVENT_FIELD_MODIFICATION;
        event.u.field_modification.thread = MkRefLocal(env, ee->thread);
        event.u.field_modification.clazz = MkRefLocal(env, mb->fb.clazz);
        event.u.field_modification.method = (jmethodID)mb;
        event.u.field_modification.location = location;
        event.u.field_modification.field_clazz = MkRefLocal(env, fb->clazz);
        event.u.field_modification.field = (jfieldID)fb;
        event.u.field_modification.object =  obj == NULL? 
                                                      NULL : 
                                                      MkRefLocal(env, obj);
        event.u.field_modification.signature_type = sig_type;
        event.u.field_modification.new_value = jval;

        eventHook(env, &event);

        (*env)->PopLocalFrame(env, 0);
    }
}

void
notify_debugger_of_class_load(ExecEnv *ee, JHandle *clazz)
{
    if (MUST_NOTIFY(JVMDI_EVENT_CLASS_LOAD, findThread(ee->thread))) {
        JVMDI_Event event;
        JNIEnv *env = EE2JNIEnv(ee);

        if ((*env)->PushLocalFrame(env, 2+LOCAL_FRAME_SLOP) < 0) {
            return;
        }
        event.kind = JVMDI_EVENT_CLASS_LOAD;
        event.u.class_event.thread = MkRefLocal(env, ee->thread);
        event.u.class_event.clazz = MkRefLocal(env, clazz);

        eventHook(env, &event);

        (*env)->PopLocalFrame(env, 0);
    }
}

void
notify_debugger_of_class_prepare(ExecEnv *ee, JHandle *clazz)
{
    if (MUST_NOTIFY(JVMDI_EVENT_CLASS_PREPARE, findThread(ee->thread))) {
        JVMDI_Event event;
        JNIEnv *env = EE2JNIEnv(ee);

        if ((*env)->PushLocalFrame(env, 2+LOCAL_FRAME_SLOP) < 0) {
            return;
        }
        event.kind = JVMDI_EVENT_CLASS_PREPARE;
        event.u.class_event.thread = MkRefLocal(env, ee->thread);
        event.u.class_event.clazz = MkRefLocal(env, clazz);

        eventHook(env, &event);

        (*env)->PopLocalFrame(env, 0);
    }
}

void
notify_debugger_of_class_unload(ExecEnv *ee, JHandle *clazz)
{
    if (eventHook && eventEnable[JVMDI_EVENT_CLASS_UNLOAD]) {
        JVMDI_Event event;
        JNIEnv *env = EE2JNIEnv(ee);

        if ((*env)->PushLocalFrame(env, 2+LOCAL_FRAME_SLOP) < 0) {
            return;
        }
        event.kind = JVMDI_EVENT_CLASS_UNLOAD;
        event.u.class_event.thread = MkRefLocal(env, ee->thread);
        event.u.class_event.clazz = MkRefLocal(env, clazz);

        eventHook(env, &event);

        (*env)->PopLocalFrame(env, 0);
    }
}

static void
reportFrameEvent(ExecEnv *ee, jint kind) 
{
    JVMDI_Event event;
    JNIEnv *env = EE2JNIEnv(ee);
    JHandle *thread = ee->thread;
    JavaFrame *frame = ee->current_frame;
    struct methodblock *mb = frame->current_method;

    if (mb == NULL) {
        return;
    }

    if ((*env)->PushLocalFrame(env, 2+LOCAL_FRAME_SLOP) < 0) {
        return;
    }
    event.kind = kind;
    event.u.frame.thread = MkRefLocal(env, thread);
    event.u.frame.clazz = MkRefLocal(env, mb->fb.clazz);
    event.u.frame.method = (jmethodID)mb;
    event.u.frame.frame = (jframeID)frame;

    eventHook(env, &event);

    (*env)->PopLocalFrame(env, 0);
}

void
notify_debugger_of_frame_pop(ExecEnv *ee)
{
    JavaFrame *frame = ee->current_frame;

    /* Replace any frame pop sentinel */
    if (frame->prev->returnpc == FRAME_POP_SENTINEL) {
        JNIEnv *env = EE2JNIEnv(ee);
        sys_thread_t *self = sysThreadSelf();
        struct fpop *fp;
        
        DEBUGGER_LOCK(self);
        fp = bagFind(framePops, frame);
        if (fp == NULL) {            
            (*env)->FatalError(env, "internal error in JVMDI (phantom frame pop)");
        } else {
            frame->prev->returnpc = fp->returnpc;  /* restore real returnpc */
            bagDelete(framePops, fp);              /* fp now points to randomness */
        }
        DEBUGGER_UNLOCK(self);

        if (!exceptionOccurred(ee) &&
            MUST_NOTIFY(JVMDI_EVENT_FRAME_POP, findThread(ee->thread))) {
                reportFrameEvent(ee, JVMDI_EVENT_FRAME_POP);
        }
    }

    if (!exceptionOccurred(ee) && 
        MUST_NOTIFY(JVMDI_EVENT_METHOD_EXIT, findThread(ee->thread))) {
        reportFrameEvent(ee, JVMDI_EVENT_METHOD_EXIT);
    }
}

void
notify_debugger_of_frame_push(ExecEnv *ee)
{
    if (MUST_NOTIFY(JVMDI_EVENT_METHOD_ENTRY, findThread(ee->thread))) {
        reportFrameEvent(ee, JVMDI_EVENT_METHOD_ENTRY);
    }
}

/*
 * This function is called at the end of the VM initialization process.
 * It triggers the sending of an event to the JVMDI client. At this point
 * The JVMDI client is free to use all of JNI and JVMDI.
 */
void
notify_debugger_of_vm_init(ExecEnv *ee)
{    
    if (eventHook) {
        JVMDI_Event event;
	JNIEnv *env = EE2JNIEnv(ee);

	if ((*env)->PushLocalFrame(env, 0+LOCAL_FRAME_SLOP) < 0) {
	    return;
        }
	event.kind = JVMDI_EVENT_VM_INIT;
	eventHook(env, &event);

	(*env)->PopLocalFrame(env, 0);
    }
}

static void
handleExit(void)
{
    if (eventHook) {
        JVMDI_Event event;
	JNIEnv *env = EE2JNIEnv(EE());

	event.kind = JVMDI_EVENT_VM_DEATH;

	if ((*env)->PushLocalFrame(env, 0+LOCAL_FRAME_SLOP) < 0) {
	    return;
        }
	eventHook(env, &event);

	(*env)->PopLocalFrame(env, 0);
    }
}

/* single-step support */

bool_t single_stepping = 0;

void set_single_step_thread(TID tid, bool_t shouldStep) 
{
    bool_t wasStepping = THREAD(tid)->single_step;
    THREAD(tid)->single_step = shouldStep;
    if (wasStepping != shouldStep) {
        if (shouldStep) {
            single_stepping++;
        } else if (single_stepping > 0) {
            single_stepping--;
        }
    }
}

/**
 * Return the underlying opcode at the specified address.
 */
int
get_opcode(ExecEnv *ee, unsigned char *pc)
{
    struct bkpt *bp;
    int rv;
    sys_thread_t *self = sysThreadSelf();

    DEBUGGER_LOCK(self);
    bp = bagFind(breakpoints, pc);
    rv = (bp == NULL)? *pc : bp->opcode;
    DEBUGGER_UNLOCK(self);          
    return rv;
}

/**
 * Return the underlying opcode at the specified address.
 * Notify of the breakpoint, if it is still there and 'notify' is true.
 * For use outside jvmdi.
 */
int
get_breakpoint_opcode(ExecEnv *ee, unsigned char *pc, bool_t notify)
{
    struct bkpt *bp;
    int rv;
    sys_thread_t *self = sysThreadSelf();

    DEBUGGER_LOCK(self);
    bp = bagFind(breakpoints, pc);
    if (bp == NULL) {
        if (*pc == opc_breakpoint) {
            SignalError(ee, JAVAPKG "InternalError", "Phantom breakpoint");
            rv = -1;
        } else {
            /* in case it has since been removed - courtesy of SAS/GVH */
	    rv = *pc;  
        }
    } else {
        rv = bp->opcode;
    }
    DEBUGGER_UNLOCK(self);          

    if (notify && bp != NULL) {
        notify_debugger_of_breakpoint(ee, pc);
    }
    return rv;
}

/* The opcode at the breakpoint has changed, for example it
 * has been quicked, update the opcode stored in the breakpoint table.
 */
bool_t
set_breakpoint_opcode(ExecEnv *ee, unsigned char *pc, unsigned char opcode)
{
    struct bkpt *bp;
    sys_thread_t *self = sysThreadSelf();

    DEBUGGER_LOCK(self);
    bp = bagFind(breakpoints, pc);
    if (bp != NULL) {
        bp->opcode = opcode;
    }
    DEBUGGER_UNLOCK(self);        
  
    if (bp == NULL) {
        SignalError(ee, JAVAPKG "InternalError", "Phantom breakpoint");
        return FALSE;
    }
    return TRUE;
}

/* clean up breakpoint - doesn't actually remove it. 
 * lock must be held */
static bool_t
clear_bkpt(JNIEnv *env, struct bkpt *bp)
{
    sysAssert(*(bp->pc) == opc_breakpoint);
    *(bp->pc) = (unsigned char)(bp->opcode);
    
    /* 
     * De-reference the enclosing class so that it's GC 
     * is no longer prevented by this breakpoint. 
     */
    (*env)->DeleteGlobalRef(env, bp->classRef);
    return TRUE;
}

/* Exported JVMDI interface begins here */

#define NOT_NULL(ptr)   { if (ptr == NULL) return JVMDI_ERROR_NULL_POINTER; }
 
#define NOT_NULL2(ptr1,ptr2)   { if (ptr1 == NULL || ptr2 == NULL) \
                                      return JVMDI_ERROR_NULL_POINTER; }

#define THREAD_OK(sys_thread)  { if (sys_thread == NULL) \
				     return JVMDI_ERROR_UNATTACHED_THREAD; }

#define ALLOC(size, where) { jvmdiError allocErr = \
 jvmdi_Allocate(size, (jbyte**)where); \
 if (allocErr != JVMDI_ERROR_NONE) return allocErr; }

#define VALID_CLASS(cb) { if (cb==NULL || (obj_classblock((JHandle *)cb) \
          != classJavaLangClass)) return JVMDI_ERROR_INVALID_CLASS; }

#define DEBUG_ENABLED() { if (!debugging) return JVMDI_ERROR_ACCESS_DENIED; }

/* When we use JNI, we don't want the accesses recorded.
 * So we bypass the special debug version of JNI and go
 * straight to the normal "unchecked" JNI.
 */
#define UNCHECKED() (&unchecked_jni_NativeInterface)

  /*
   *  Threads
   */

static sys_thread_t *SYSTHREAD(HThread *tid)
{
    ExecEnv *ee = (ExecEnv *)ll2ptr(THREAD(tid)->eetop);
    return ee ? EE2SysThread(ee) : 0;
}

/* Used for counting enumerated threads
 */
static int
allThreadsCountHelper(sys_thread_t *t, void *arg)
{
    (*(int *)arg)++;

    return SYS_OK;
}

/* Used for filling in enumerated threads
 */
static int
allThreadsHelper(sys_thread_t *t, void *arg)
{
    struct execenv *ee = SysThread2EE(t);
    JNIEnv *env = EE2JNIEnv(EE());
    jthread **threadIter = (jthread **)arg;
    jobject thr;

    if ((*env)->PushLocalFrame(env, 1) < 0) {
        return SYS_NOMEM;
    }
    thr = MkRefLocal(env, ee->thread);
    **threadIter = (jthread)((*env)->NewGlobalRef(env, thr));
    (*threadIter)++;
    (*env)->PopLocalFrame(env, 0);

    return SYS_OK;
}

/* Return via "threadsPtr" all threads in this VM
 * ("threadsCountPtr" returns the number of such threads).
 * Memory for this table is allocated by the function specified 
 * in JVMDI_SetAllocationHooks().
 * Note: threads in the array are JNI global references and must
 * be explicitly managed.
 * Errors: JVMDI_ERROR_NULL_POINTER, JVMDI_ERROR_OUT_OF_MEMORY
 */
static jvmdiError JNICALL
jvmdi_GetAllThreads(jint *threadsCountPtr, jthread **threadsPtr)
{
    jvmdiError rc = JVMDI_ERROR_NONE;
    sys_thread_t *self = sysThreadSelf();

    DEBUG_ENABLED();
    NOT_NULL(threadsCountPtr);
    THREAD_OK(self);

    QUEUE_LOCK(self);
    *threadsCountPtr = 0;

    /* count the threads */
    sysThreadEnumerateOver(allThreadsCountHelper, (void *)threadsCountPtr);
    rc = jvmdi_Allocate(size_to_jlong(*threadsCountPtr * sizeof(jthread)),
                        (jbyte**)threadsPtr);
    if (rc == JVMDI_ERROR_NONE) {

        jthread *tempPtr = *threadsPtr;
        /* fill in the threads */
        int err = sysThreadEnumerateOver(allThreadsHelper, 
                                         (void *) &tempPtr);
        if (err == SYS_NOMEM) {
            rc = JVMDI_ERROR_OUT_OF_MEMORY;
        }
    }
    QUEUE_UNLOCK(self);

    return rc;
}

static jint
decodeThreadStatus(int sysState) {
    /* JVMDI ignores interrupt state for now. */
    sysState = sysState & (~SYS_THREAD_INTERRUPTED);

    if (sysState == SYS_THREAD_SUSPENDED) {
        /*
         * Suspended alone means that it was runnable before
         */
        return JVMDI_THREAD_STATUS_RUNNING;
    } else {
        /*
         * Mask any suspended bit, so we can decode pre-suspend status
         */
        sysState &= ~SYS_THREAD_SUSPENDED;
        switch (sysState) {
            case SYS_THREAD_RUNNABLE:
                return JVMDI_THREAD_STATUS_RUNNING;

            case SYS_THREAD_MONITOR_WAIT:
                /* threadSleep() uses the pointer to the SYSTHREAD as the monitor 
                 * registry_lock key, which is stored in the pointer to the 
                 * monitor's mon_wait structure.
                 */
                /* We can't get this info on Win32 yet.
                if ((int)SYSTHREAD(h)->mon_wait == (int)SYSTHREAD(h))
                    return JVMDI_THREAD_STATUS_SLEEPING;
                    */
                return JVMDI_THREAD_STATUS_MONITOR;
    
            case SYS_THREAD_CONDVAR_WAIT:
                /* see previous comment */
                /* We can't get this info on Win32 yet.
                if ((int)SYSTHREAD(h)->mon_wait == (int)SYSTHREAD(h))
                    return JVMDI_THREAD_STATUS_SLEEPING;
                    */
                return JVMDI_THREAD_STATUS_WAIT;
    
            default:
                return JVMDI_THREAD_STATUS_UNKNOWN;
        }
    }
}

static jint
decodeSuspendStatus(Hjava_lang_Thread *thread, int sysState) {
    jint status = 0;
    if (sysState & SYS_THREAD_SUSPENDED) {
        struct javaframe *currFrame;
        ExecEnv* thread_ee;
        status |= JVMDI_SUSPEND_STATUS_SUSPENDED;

        /* 
         * Find out if this thread is at a breakpoint
         */
        thread_ee = (ExecEnv *)ll2ptr(unhand(thread)->eetop);
        if ((thread_ee != 0) && 
            ((currFrame = thread_ee->current_frame) != 0) &&
            (currFrame->lastpc != 0) &&
            (currFrame->current_method != 0) &&
            ((currFrame->current_method->fb.access & ACC_NATIVE) == 0) &&
            (*currFrame->lastpc == opc_breakpoint)) {
            status |= JVMDI_SUSPEND_STATUS_BREAK;
        }
    }
    return status;
}

/* Return via "threadStatusPtr" the current status of the thread and 
 * via "suspendStatusPtr" information on suspension. If the thread
 * has been suspended, bits in the suspendStatus will be set, and the 
 * thread status indicates the pre-suspend status of the thread.
 * JVMDI_THREAD_STATUS_* flags are used to identify thread and 
 * suspend status.
 * Errors: JVMDI_ERROR_INVALID_THREAD, JVMDI_ERROR_NULL_POINTER
 */
static jvmdiError JNICALL
jvmdi_GetThreadStatus(jthread threadObj,
                      jint *threadStatusPtr, jint *suspendStatusPtr)
{
    Hjava_lang_Thread *thread;
    int sysState;

    DEBUG_ENABLED();
    NOT_NULL2(threadStatusPtr, suspendStatusPtr);

    thread  = (Hjava_lang_Thread *)DeRef(EE2JNIEnv(EE()), threadObj);

    if (thread == 0) {
        *threadStatusPtr = JVMDI_THREAD_STATUS_UNKNOWN;
        *suspendStatusPtr = 0;
    } else if (SYSTHREAD(thread) == 0) {
        *threadStatusPtr = JVMDI_THREAD_STATUS_ZOMBIE;
        *suspendStatusPtr = 0;
    } else {
        sysState = sysThreadGetStatus(SYSTHREAD(thread), NULL);
        *threadStatusPtr = decodeThreadStatus(sysState);
        *suspendStatusPtr = decodeSuspendStatus(thread, sysState);
    }

    return JVMDI_ERROR_NONE;
}

/* Suspend the specified thread.
 * Errors: JVMDI_ERROR_INVALID_THREAD, JVMDI_ERROR_THREAD_SUSPENDED,
 * JVMDI_ERROR_VM_DEAD 
 */
static jvmdiError JNICALL
jvmdi_SuspendThread(jthread threadObj)
{
    jvmdiError rc;
    int sysState;
    sys_thread_t *target;
    sys_thread_t *self = sysThreadSelf();
    ExecEnv *ee;
    JNIEnv *env;    

    DEBUG_ENABLED();
    THREAD_OK(self);

    ee = SysThread2EE(self);
    env = EE2JNIEnv(ee);
    target = SYSTHREAD((Hjava_lang_Thread *)DeRef(env, threadObj));

    if (target == self) {
        /*
         * Suspending self. Don't need to check for previous suspension.
         * Shouldn't grab any locks here.
         */
        JVM_SuspendThread(env, (jobject)threadObj);
        rc = JVMDI_ERROR_NONE;
    } else {
        /*
         * Suspending another thread. If it's already suspended 
         * report an error.
         */

        /*
         * Prevent the thread from holding critical VM locks while 
         * suspended. Also, making this call here prevents the thread
         * from changing state between the GetStatus and the Suspend
         */
        lock_for_debugger(ee);
    
        /*
         * Now that we're locked down, re-fetch the sys thead so that 
         * we can be sure it hasn't gone away.
         */
        target = SYSTHREAD((Hjava_lang_Thread *)DeRef(env, threadObj));
        if (target == 0) {
            /*
             * It has finished running and is a zombie
             */
            rc = JVMDI_ERROR_INVALID_THREAD;
        } else {
            sysState = sysThreadGetStatus(target, NULL);
            if (sysState & SYS_THREAD_SUSPENDED) {
                rc = JVMDI_ERROR_THREAD_SUSPENDED;
            } else {
                JVM_SuspendThread(env, (jobject)threadObj);
                rc = JVMDI_ERROR_NONE;
            }
        }
        unlock_for_debugger(ee);
    }
    return rc;
}

/* Resume the specified thread.
 * Errors: JVMDI_ERROR_INVALID_THREAD, JVMDI_ERROR_THREAD_NOT_SUSPENDED, 
 * JVMDI_ERROR_INVALID_TYPESTATE,
 * JVMDI_ERROR_VM_DEAD
 */
static jvmdiError JNICALL
jvmdi_ResumeThread(jthread threadObj)
{
    jvmdiError rc;
    int sysState;
    Hjava_lang_Thread *thread;
    sys_thread_t *self = sysThreadSelf();
    ExecEnv *ee;
    JNIEnv *env;    

    DEBUG_ENABLED();
    THREAD_OK(self);
    QUEUE_LOCK(self);

    ee = SysThread2EE(self);
    env = EE2JNIEnv(ee);

    thread = (Hjava_lang_Thread *)DeRef(env, threadObj);

    sysState = sysThreadGetStatus(SYSTHREAD(thread), NULL);
    if (sysState & SYS_THREAD_SUSPENDED) {
        JVM_ResumeThread(env, (jobject)threadObj);
        rc = JVMDI_ERROR_NONE;
    } else {
        rc = JVMDI_ERROR_THREAD_NOT_SUSPENDED;
    }
    QUEUE_UNLOCK(self);
    return rc;
}  

/* 
 * Kill the specified thread - generate the specified exception.
 * For the default stop() pass in an instance of ThreadDeath. 
 * JVMDI equivalent of Thread.stop().
 * Errors: JVMDI_ERROR_INVALID_THREAD
 */
static jvmdiError JNICALL
jvmdi_StopThread(jthread thread, jobject exception) {
    sys_thread_t *self = sysThreadSelf();
    ExecEnv *ee;
    JNIEnv *env;    

    DEBUG_ENABLED();
    THREAD_OK(self);

    ee = SysThread2EE(self);
    env = EE2JNIEnv(ee);

    JVM_StopThread(env, thread, exception);
    return JVMDI_ERROR_NONE;
}  

/* 
 * Interrupt the specified thread. 
 * JVMDI equivalent of Thread.interrupt().
 * Errors: JVMDI_ERROR_INVALID_THREAD
 */
static jvmdiError JNICALL
jvmdi_InterruptThread(jthread thread) {
    sys_thread_t *self = sysThreadSelf();
    ExecEnv *ee;
    JNIEnv *env;    

    DEBUG_ENABLED();
    THREAD_OK(self);

    ee = SysThread2EE(self);
    env = EE2JNIEnv(ee);

    JVM_Interrupt(env, thread);
    return JVMDI_ERROR_NONE;
}  

static jfieldID 
getFieldID(JNIEnv *env, jclass clazz, char *name, char *type)
{
    jfieldID id = (*env)->GetFieldID(env, clazz, name, type);
    jthrowable exc = (*env)->ExceptionOccurred(env);
    if (exc) {
            (*env)->ExceptionDescribe(env);
            (*env)->FatalError(env, "internal error in JVMDI");
    }
    return id;
}

/* count UTF length of Unicode chars */
static int 
lengthCharsUTF(jint uniLen, jchar *uniChars)  
{
    int i;
    int utfLen = 0;
    jchar *uniP = uniChars;

    for (i = 0 ; i < uniLen ; i++) {
        int c = *uniP++;
        if ((c >= 0x0001) && (c <= 0x007F)) {
            utfLen++;
        } else if (c > 0x07FF) {
            utfLen += 3;
        } else {
            utfLen += 2;
        }
    }
    return utfLen;
}

/* convert Unicode to UTF */
static void 
setBytesCharsUTF(jint uniLen, jchar *uniChars, char *utfBytes)  
{
    int i;
    jchar *uniP = uniChars;
    char *utfP = utfBytes;

    for (i = 0 ; i < uniLen ; i++) {
        int c = *uniP++;
        if ((c >= 0x0001) && (c <= 0x007F)) {
            *utfP++ = c;
        } else if (c > 0x07FF) {
            *utfP++ = (0xE0 | ((c >> 12) & 0x0F));
            *utfP++ = (0x80 | ((c >>  6) & 0x3F));
            *utfP++ = (0x80 | ((c >>  0) & 0x3F));
        } else {
            *utfP++ = (0xC0 | ((c >>  6) & 0x1F));
            *utfP++ = (0x80 | ((c >>  0) & 0x3F));
        }
    }
    *utfP++ = 0; /* terminating zero */
}


/* Return via "infoPtr" thread info.
 * Memory for the name string is allocated by the function specified 
 * in JVMDI_SetAllocationHooks().
 * Note: returned objects are allocated globally and must be explictly
 * freed with DeleteGlobalRef.
 * Errors: JVMDI_ERROR_INVALID_THREAD, JVMDI_ERROR_NULL_POINTER,
 * JVMDI_ERROR_OUT_OF_MEMORY
 */
static jvmdiError JNICALL
jvmdi_GetThreadInfo(jthread thread, JVMDI_thread_info *infoPtr) 
{
    static jfieldID nameID = 0;
    static jfieldID priorityID;
    static jfieldID daemonID;
    static jfieldID groupID;
    static jfieldID loaderID;
    jcharArray nmObj;
    jobject groupObj;
    jobject loaderObj;
    sys_thread_t *self = sysThreadSelf();
    ExecEnv *ee;
    JNIEnv *env;    
    
    DEBUG_ENABLED();
    NOT_NULL(infoPtr);
    THREAD_OK(self);

    ee = SysThread2EE(self);
    env = EE2JNIEnv(ee);
    if (nameID == 0) {
        jclass threadClass = (*env)->GetObjectClass(env, thread);
        nameID = getFieldID(env, threadClass, "name", "[C");
        priorityID = getFieldID(env, threadClass, "priority", "I");
        daemonID = getFieldID(env, threadClass, "daemon", "Z");
        groupID = getFieldID(env, threadClass, "group", 
                             "Ljava/lang/ThreadGroup;");
        loaderID = getFieldID(env, threadClass, "contextClassLoader", 
                              "Ljava/lang/ClassLoader;");
    }
    nmObj = (jcharArray)(UNCHECKED()->GetObjectField(env, thread, nameID));
    infoPtr->priority = UNCHECKED()->GetIntField(env, thread, priorityID);
    infoPtr->is_daemon = UNCHECKED()->GetBooleanField(env, thread, daemonID);
    groupObj = UNCHECKED()->GetObjectField(env, thread, groupID);
    infoPtr->thread_group = groupObj == NULL ? NULL :
                    (jthreadGroup)((*env)->NewGlobalRef(env, groupObj));
    loaderObj = UNCHECKED()->GetObjectField(env, thread, loaderID);
    infoPtr->context_class_loader = loaderObj == NULL ? NULL :
                    (*env)->NewGlobalRef(env, loaderObj);

    {   /* copy the thread name */
        jint uniLen = (*env)->GetArrayLength(env, nmObj);
        jchar *uniChars = (*env)->GetCharArrayElements(env, nmObj, NULL);
        jint utfLen = lengthCharsUTF(uniLen, uniChars);
        ALLOC(jint_to_jlong(utfLen+1), &(infoPtr->name));
        setBytesCharsUTF(uniLen, uniChars, infoPtr->name);
        (*env)->ReleaseCharArrayElements(env, nmObj, uniChars, JNI_ABORT);
    }
   
    return JVMDI_ERROR_NONE;
}
    
#ifndef FAST_MONITOR

typedef struct {
    JNIEnv *env;
    jvmdiError error;
    sys_thread_t *sysThread;
    int count;
    jobject *monIter;
} MonitorEnumArg;

static void 
deleteRefArray(JNIEnv *env, jobject *objects, int count) 
{
    int i;
    for (i = 0; i < count; i++) {
        (*env)->DeleteGlobalRef(env, objects[i]);
    }
}

static void 
ownedMonitorCountHelper(monitor_t *monitor, void *arg)
{
    MonitorEnumArg *monArg = arg;
    monArg->error = JVMDI_ERROR_NONE;

    /*
     * Count the monitor as owned if its owner thread is the
     * right one and if it is a Java monitor.
     */
    if ((sysMonitorOwner(sysmon(monitor)) == monArg->sysThread) &&
        (lookupJavaMonitor(sysmon(monitor)) != NULL)) {
        monArg->count++;
    }
}

static void 
ownedMonitorHelper(monitor_t *monitor, void *arg) 
{
    MonitorEnumArg *monArg = arg;
    monArg->error = JVMDI_ERROR_NONE;

    if (sysMonitorOwner(sysmon(monitor)) == monArg->sysThread) {
        JHandle *handle = lookupJavaMonitor(sysmon(monitor));
        if (handle) {
            jobject object;
            JNIEnv *env = monArg->env;
    
            if ((*env)->PushLocalFrame(env, 1) < 0) {
                monArg->error = JVMDI_ERROR_OUT_OF_MEMORY;
                return;
            }
            object = MkRefLocal(env, handle);
            *monArg->monIter = (jobject)((*env)->NewGlobalRef(env, object));
            (*env)->PopLocalFrame(env, 0);
    
            if (*monArg->monIter == 0) {
                monArg->error = JVMDI_ERROR_OUT_OF_MEMORY;
                return;
            }
            monArg->count++;
            monArg->monIter++;
        }
    }
}

#endif /* FAST_MONITOR */
  
/* Return via "infoPtr" monitor ownership info for the given thread.
 * The thread must be suspended before calling this function.
 * Memory for the owned monitor array is allocated by the function specified 
 * in JVMDI_SetAllocationHooks().
 * Note: returned objects are allocated globally and must be explictly
 * freed with DeleteGlobalRef.
 * Errors: JVMDI_ERROR_INVALID_THREAD, JVMDI_ERROR_NULL_POINTER,
 * JVMDI_ERROR_OUT_OF_MEMORY, JVMDI_ERROR_THREAD_NOT_SUSPENDED
 */
static jvmdiError JNICALL
jvmdi_GetOwnedMonitorInfo(jthread threadObj, JVMDI_owned_monitor_info *infoPtr) 
{
#ifdef FAST_MONITOR
    return JVMDI_ERROR_NOT_IMPLEMENTED;
#else
    jvmdiError rc = JVMDI_ERROR_NONE;
    int error;
    sys_thread_t *self = sysThreadSelf();
    MonitorEnumArg arg;
    ExecEnv *ee;
    JNIEnv *env;    
    Hjava_lang_Thread *thread;

    DEBUG_ENABLED();
    THREAD_OK(self);
    NOT_NULL(infoPtr);

    ee = SysThread2EE(self);
    env = EE2JNIEnv(ee);
    thread  = (Hjava_lang_Thread *)DeRef(env, threadObj);

    /* count the owned monitors */
    arg.env = env;
    arg.sysThread = SYSTHREAD(thread);
    arg.count = 0;

    CACHE_LOCK(self);
    monitorEnumerate(ownedMonitorCountHelper, &arg);
    CACHE_UNLOCK(self);

    infoPtr->owned_monitor_count = arg.count;
    if (arg.count == 0) {
        infoPtr->owned_monitors = NULL;
    } else {
        ALLOC(size_to_jlong(arg.count * sizeof(jobject)),
	      &infoPtr->owned_monitors);
    
        /* fill in the owned monitors */
        arg.count = 0;
        arg.monIter = infoPtr->owned_monitors;
        CACHE_LOCK(self);
        monitorEnumerate(ownedMonitorHelper, &arg);
        CACHE_UNLOCK(self);
    
        /*
         * If the counts don't match, something bad happened. Either
         * there was an error in the helper or threads are running and have
         * changed monitor states. In either case an error should be returned here.
         */
        if (infoPtr->owned_monitor_count != arg.count) {
            deleteRefArray(env, infoPtr->owned_monitors, arg.count);
            jvmdi_Deallocate((jbyte *)infoPtr->owned_monitors);
            if (arg.error != JVMDI_ERROR_NONE) {
                return JVMDI_ERROR_OUT_OF_MEMORY;
            } else {
                /* Monitors changed ==> threads were running */
                return JVMDI_ERROR_THREAD_NOT_SUSPENDED;
            }
        } 
    }

    return JVMDI_ERROR_NONE;
#endif /* FAST_MONITOR */
}

static jvmdiError JNICALL
jvmdi_GetCurrentContendedMonitor(jthread threadObj, jobject *monitorPtr) 
{
    Hjava_lang_Thread *thread;
    int sysState;
    sys_mon_t *sysMon;
    sys_thread_t *self = sysThreadSelf();
    JNIEnv *env;
    JHandle *handle;
    jobject monitor;

    DEBUG_ENABLED();
    NOT_NULL(monitorPtr);
    THREAD_OK(self);

    env = EE2JNIEnv(SysThread2EE(self));
    thread  = (Hjava_lang_Thread *)DeRef(EE2JNIEnv(EE()), threadObj);

    if (thread == 0) {
        return JVMDI_ERROR_INVALID_THREAD;
    } 
    
    if (SYSTHREAD(thread) == 0) {
        /* Zombies are not waiting on monitors */
        *monitorPtr = NULL;
        return JVMDI_ERROR_NONE;
    }
     
    QUEUE_LOCK(self);   /* Prevent anyone from resuming in the middle */
    sysState = sysThreadGetStatus(SYSTHREAD(thread), &sysMon);
    QUEUE_UNLOCK(self);
    if ((sysState & SYS_THREAD_SUSPENDED) == 0) {
        return JVMDI_ERROR_THREAD_NOT_SUSPENDED;
    }

    if (sysMon == NULL) {
        /* Not waiting on any monitor */
        *monitorPtr = NULL;
        return JVMDI_ERROR_NONE;
    }

    /*
     * Attempt to find the java object associated with the monitor 
     */
    handle = lookupJavaMonitor(sysMon);
    if (handle == NULL) {
        /* Waiting on an internal monitor of some kind */
        *monitorPtr = NULL;
        return JVMDI_ERROR_NONE;
    }

    if ((*env)->PushLocalFrame(env, 1) < 0) {
        return JVMDI_ERROR_OUT_OF_MEMORY;
    }
    monitor = MkRefLocal(env, handle);
    monitor = (*env)->NewGlobalRef(env, monitor);
    (*env)->PopLocalFrame(env, 0);
    if (monitor == NULL) {
        return JVMDI_ERROR_OUT_OF_MEMORY;
    }
    *monitorPtr = monitor;
    return JVMDI_ERROR_NONE;
}
  /*
   *  Thread Groups
   */

/* Return in 'groups' an array of top-level thread groups (parentless 
 * thread groups).
 * Note: returned objects are allocated globally and must be explictly
 * freed with DeleteGlobalRef.
 * Errors: JVMDI_ERROR_NULL_POINTER, JVMDI_ERROR_OUT_OF_MEMORY
 */
static jvmdiError JNICALL
jvmdi_GetTopThreadGroups(jint *groupCountPtr, 
                         jthreadGroup **groupsPtr)
{
    jobject sysgrp;
    sys_thread_t *self = sysThreadSelf();
    ExecEnv *ee;
    JNIEnv *env;    

    DEBUG_ENABLED();
    NOT_NULL2(groupCountPtr, groupsPtr);
    THREAD_OK(self);

    ee = SysThread2EE(self);
    env = EE2JNIEnv(ee);

    /* We only have the one top group */
    ALLOC(size_to_jlong(sizeof(jthreadGroup)), groupsPtr);
    *groupCountPtr = 1;

    if ((*env)->PushLocalFrame(env, 1) < 0) {
        return JVMDI_ERROR_OUT_OF_MEMORY;
    }
    sysgrp = MkRefLocal(env, systemgroup);
    **groupsPtr = (jthreadGroup)((*env)->NewGlobalRef(env, sysgrp));
    (*env)->PopLocalFrame(env, 0);
   
    return JVMDI_ERROR_NONE;
}

/* Return via "infoPtr" thread group info.
 * Memory for the name string is allocated by the function specified 
 * in JVMDI_SetAllocationHooks().
 * Note: returned objects are allocated globally and must be explictly
 * freed with DeleteGlobalRef.
 * Errors: JVMDI_ERROR_INVALID_THREAD_GROUP, JVMDI_ERROR_NULL_POINTER,
 * JVMDI_ERROR_OUT_OF_MEMORY
 */
static jvmdiError JNICALL
jvmdi_GetThreadGroupInfo(jthreadGroup group, 
                         JVMDI_thread_group_info *infoPtr) 
{
    static jfieldID parentID;
    static jfieldID nameID = 0;
    static jfieldID maxPriorityID;
    static jfieldID daemonID;
    jstring nmString;
    jobject parentObj;
    sys_thread_t *self = sysThreadSelf();
    ExecEnv *ee;
    JNIEnv *env;    
    
    DEBUG_ENABLED();
    NOT_NULL(infoPtr);
    THREAD_OK(self);

    ee = SysThread2EE(self);
    env = EE2JNIEnv(ee);

    if (nameID == 0) {
        jclass tgClass = (*env)->GetObjectClass(env, group);
        parentID = getFieldID(env, tgClass, "parent", 
                             "Ljava/lang/ThreadGroup;");
        nameID = getFieldID(env, tgClass, "name", 
                            "Ljava/lang/String;");
        maxPriorityID = getFieldID(env, tgClass, "maxPriority", "I");
        daemonID = getFieldID(env, tgClass, "daemon", "Z");
    }
    parentObj = UNCHECKED()->GetObjectField(env, group, parentID);
    infoPtr->parent = parentObj == NULL? NULL : 
                   (jthreadGroup)((*env)->NewGlobalRef(env, parentObj));
    nmString = (jstring)(UNCHECKED()->GetObjectField(env, group, nameID));
    infoPtr->max_priority = UNCHECKED()->GetIntField(env, group, 
                                                      maxPriorityID);
    infoPtr->is_daemon = UNCHECKED()->GetBooleanField(env, group, daemonID);

    {   /* copy the thread group name */
        jint nmLen = (*env)->GetStringUTFLength(env, nmString);
        const char *nmValue = (*env)->GetStringUTFChars(env, 
                                                         nmString, NULL);
        ALLOC(jint_to_jlong(nmLen+1), &(infoPtr->name));
        strcpy(infoPtr->name, nmValue);
        (*env)->ReleaseStringUTFChars(env, nmString, nmValue);
    }
   
    return JVMDI_ERROR_NONE;
}
    
static jvmdiError 
objectArrayToArrayOfObject(JNIEnv *env, jint cnt, jobject **destPtr,
                           jobjectArray arr) {
    jvmdiError rc;
                     
    /* allocate the destination array */
    rc = jvmdi_Allocate(size_to_jlong(cnt * sizeof(jobject)),
			(jbyte**)destPtr);

    /* fill-in the destination array */
    if (rc == JVMDI_ERROR_NONE) {
        int inx;
        jobject *rp = *destPtr;
        for (inx = 0; inx < cnt; inx++) {
            jobject obj = (*env)->NewGlobalRef(env, 
                          (*env)->GetObjectArrayElement(env, arr, inx));
            if (obj == NULL) {
                rc = JVMDI_ERROR_OUT_OF_MEMORY;
                break;
            }
            *rp++ = obj;
        }

        /* clean up partial array after any error */
        if (rc != JVMDI_ERROR_NONE) {
            for (rp--; rp >= *destPtr; rp--) {
                (*env)->DeleteGlobalRef(env, *rp);
            }
            jvmdi_Deallocate((jbyte *)*destPtr);
        }            
    }
    return rc;
}

/* Return via "threadCountPtr" the number of threads in this thread group.
 * Return via "threadsPtr" the threads in this thread group.
 * Return via "groupCountPtr" the number of thread groups in this thread group.
 * Return via "groupsPtr" the thread groups in this thread group.
 * Note: returned objects are allocated globally and must be explictly
 * freed with DeleteGlobalRef.
 * Errors: JVMDI_ERROR_INVALID_THREAD_GROUP, JVMDI_ERROR_NULL_POINTER,
 * JVMDI_ERROR_OUT_OF_MEMORY
 */
static jvmdiError JNICALL
jvmdi_GetThreadGroupChildren(jthreadGroup group, 
                             jint *threadCountPtr, jthread **threadsPtr,
                             jint *groupCountPtr, jthreadGroup **groupsPtr) {
    static jfieldID nthreadsID = 0;
    static jfieldID threadsID;
    static jfieldID ngroupsID;
    static jfieldID groupsID;
    jvmdiError rc;
    jint nthreads;
    jobject threads;
    jint ngroups;
    jobject groups;
    sys_thread_t *self = sysThreadSelf();
    ExecEnv *ee;
    JNIEnv *env;    
    
    DEBUG_ENABLED();
    NOT_NULL2(threadCountPtr, threadsPtr);
    NOT_NULL2(groupCountPtr, groupsPtr);
    THREAD_OK(self);

    ee = SysThread2EE(self);
    env = EE2JNIEnv(ee);

    if (nthreadsID == 0) {
        jclass tgClass = (*env)->GetObjectClass(env, group);
        nthreadsID = getFieldID(env, tgClass, "nthreads", "I");
        threadsID = getFieldID(env, tgClass, "threads", 
                             "[Ljava/lang/Thread;");
        ngroupsID = getFieldID(env, tgClass, "ngroups", "I");
        groupsID = getFieldID(env, tgClass, "groups", 
                             "[Ljava/lang/ThreadGroup;");
    }
    nthreads = UNCHECKED()->GetIntField(env, group, nthreadsID);
    threads = UNCHECKED()->GetObjectField(env, group, threadsID);
    ngroups = UNCHECKED()->GetIntField(env, group, ngroupsID);
    groups = UNCHECKED()->GetObjectField(env, group, groupsID);

    rc = objectArrayToArrayOfObject(env, nthreads, threadsPtr, threads);
    if (rc == JVMDI_ERROR_NONE) {
        rc = objectArrayToArrayOfObject(env, ngroups, groupsPtr, groups);

        /* deallocate all of threads list on error */
        if (rc != JVMDI_ERROR_NONE) {
            int inx;
            jthread *rp = *threadsPtr;
            for (inx = 0; inx < nthreads; inx++) {
                (*env)->DeleteGlobalRef(env, *rp++); 
            }
            jvmdi_Deallocate((jbyte *)*threadsPtr);
        } else {
            *threadCountPtr = nthreads;
            *groupCountPtr = ngroups;
        }
    }
    return rc;
}


  /*
   *  Stack access
   */

/* Return via "countPtr" the number of frames in the thread's call stack.
 * Errors: JVMDI_ERROR_INVALID_THREAD,
 * JVMDI_ERROR_NULL_POINTER
 */
static jvmdiError JNICALL
jvmdi_GetFrameCount(jthread thread, jint *countPtr) 
{
    ExecEnv *thread_ee;
    JavaFrame *frame;
    jint count = 0;

    DEBUG_ENABLED();

    thread_ee = (ExecEnv *)ll2ptr(THREAD(DeRef(EE2JNIEnv(EE()),thread))->eetop);
    if (thread_ee == 0) {
	return JVMDI_ERROR_INVALID_THREAD;
    }

    NOT_NULL(countPtr);

    frame = thread_ee->current_frame;
    while (frame != NULL) {
        /* skip pseudo frames */
        if (frame->current_method != 0) {
            count++;
	}
        frame = frame->prev;
    }

    *countPtr = count;

    return JVMDI_ERROR_NONE;
}    

/* Return via "framePtr" the frame ID for the current stack frame 
 * of this thread.  Thread must be suspended. 
 * Errors: JVMDI_ERROR_NO_MORE_FRAMES, JVMDI_ERROR_INVALID_THREAD,
 * JVMDI_ERROR_THREAD_NOT_SUSPENDED, JVMDI_ERROR_NULL_POINTER
 */
static jvmdiError JNICALL
jvmdi_GetCurrentFrame(jthread thread, jframeID *framePtr) 
{
    ExecEnv *thread_ee;
    JavaFrame *frame;

    DEBUG_ENABLED();

    thread_ee = (ExecEnv *)ll2ptr(THREAD(DeRef(EE2JNIEnv(EE()),thread))->eetop);

    if (thread_ee == 0) {
	/* Thread hasn't any state yet. */
	return JVMDI_ERROR_INVALID_THREAD;
    }
    NOT_NULL(framePtr);
    /* TO DO: check for not suspended */
    frame = thread_ee->current_frame;
    if (frame == NULL) {
        return JVMDI_ERROR_NO_MORE_FRAMES;
    }
    while (frame->current_method == 0) {
        frame = frame->prev;
        if (frame == NULL) {
            return JVMDI_ERROR_NO_MORE_FRAMES;
        }
    }
    *framePtr = (jframeID)frame;
    return JVMDI_ERROR_NONE;
}

/* Return via "framePtr" the frame ID for the stack frame that called 
 * the specified frame.
 * Errors: JVMDI_ERROR_NO_MORE_FRAMES, JVMDI_ERROR_INVALID_FRAMEID,
 * JVMDI_ERROR_NULL_POINTER
 */
static jvmdiError JNICALL
jvmdi_GetCallerFrame(jframeID called, jframeID *framePtr) 
{
    JavaFrame	*frame = (JavaFrame *)called;
    struct javaframe frame_buf;

    DEBUG_ENABLED();
    NOT_NULL(framePtr);
    if (frame == NULL) {
        return JVMDI_ERROR_INVALID_FRAMEID;
    }

    frame = getCallerFrame(frame, 1, &frame_buf);
    if (frame == NULL) {
	return JVMDI_ERROR_NO_MORE_FRAMES;
    }

    sysAssert(frame != & frame_buf);
    *framePtr = (jframeID)frame;
    return JVMDI_ERROR_NONE;
}

/* Return via "classPtr" and "methodPtr" the active method in the 
 * specified frame. Return via "locationPtr" the index of the
 * currently executing instruction within the specified frame,
 * return negative one if location not available.
 * Note: class returned via 'classPtr' is allocated locally and must
 * be explicitly managed.
 * Errors: JVMDI_ERROR_INVALID_FRAMEID, JVMDI_ERROR_NULL_POINTER
 * JVMDI_ERROR_OUT_OF_MEMORY
 */
static jvmdiError JNICALL
jvmdi_GetFrameLocation(jframeID frame, 
                       jclass *classPtr, jmethodID *methodPtr,
                       jlocation *locationPtr)
{
    JavaFrame	*jframe = (JavaFrame *)frame;
    struct methodblock *mb = jframe->current_method;
    jobject clazz;
    sys_thread_t *self = sysThreadSelf();
    ExecEnv *ee;
    JNIEnv *env;    

    DEBUG_ENABLED();
    NOT_NULL2(classPtr, methodPtr);
    THREAD_OK(self);

    ee = SysThread2EE(self);
    env = EE2JNIEnv(ee);

    if (mb == NULL) {
        return JVMDI_ERROR_INVALID_FRAMEID;
    }
    if ((*env)->PushLocalFrame(env, 1) < 0) {
        return JVMDI_ERROR_OUT_OF_MEMORY;
    }
    clazz = MkRefLocal(env, mb->fb.clazz);
    *classPtr = (jclass)((*env)->NewGlobalRef(env, clazz));
    (*env)->PopLocalFrame(env, 0);
    *methodPtr = (jmethodID)mb;

    if (mb->fb.access & ACC_NATIVE) {
        *locationPtr = jlong_minus_one;
    } else {
        *locationPtr = ptrdiff_to_jlong(jframe->lastpc - mb->code);
    }
    return JVMDI_ERROR_NONE;
}  			    

/* Send a JVMDI_EVENT_FRAME_POP event when the specified frame is
 * popped from the stack.  The event is sent after the pop has occurred.
 * Errors: JVMDI_ERROR_INVALID_FRAMEID, JVMDI_ERROR_NATIVE_FRAME
 */
static jvmdiError JNICALL
jvmdi_NotifyFramePop(jframeID frame)
{
    JavaFrame *jframe = (JavaFrame *)frame;
    struct methodblock *mb = jframe->current_method;
    sys_thread_t *self = sysThreadSelf();
    jvmdiError err;

    DEBUG_ENABLED();
    THREAD_OK(self);

    if (mb == NULL) {
        return JVMDI_ERROR_INVALID_FRAMEID;
    }
    if (mb->fb.access & ACC_NATIVE) {
        return JVMDI_ERROR_OPAQUE_FRAME;
    }
    
    /*
     * Frame pop notification operates on the calling frame, so 
     * make sure it exists.
     */
    if (jframe->prev == NULL) {
        /* First frame (no previous) must be opaque */
        err = JVMDI_ERROR_OPAQUE_FRAME;
    } else {
        DEBUGGER_LOCK(self);
        /* check if one already exists */
        if (bagFind(framePops, frame) != NULL) {    
            err = JVMDI_ERROR_DUPLICATE;
        } else {
            /* allocate space for it */
            struct fpop *fp = bagAdd(framePops);                     
            if (fp == NULL) {
                err = JVMDI_ERROR_OUT_OF_MEMORY;
            } else {
                fp->frame = jframe;
                /* save real returnpc */
                fp->returnpc = jframe->prev->returnpc;       

                /* munge stack to create frame pop event */
                jframe->prev->returnpc = FRAME_POP_SENTINEL; 

                err = JVMDI_ERROR_NONE;
            }
        }
        DEBUGGER_UNLOCK(self);          
    }

    return err;
}


  /*
   *  Local variables
   */

/* Return via "siPtr" the stack item at the specificied slot.
 * Errors: JVMDI_ERROR_INVALID_FRAMEID, JVMDI_ERROR_INVALID_SLOT,
 * JVMDI_ERROR_TYPE_MISMATCH, JVMDI_ERROR_NULL_POINTER,
 * JVMDI_ERROR_OPAQUE_FRAME
 */
static jvmdiError
getSlot(jframeID frame, jint slot, stack_item **siPtr)
{
    JavaFrame	*jframe = (JavaFrame *)frame;
    stack_item *vars;
    struct methodblock *mb;

    mb = jframe->current_method;
    vars = jframe->vars;
    if (mb == NULL || vars == NULL) {
        return JVMDI_ERROR_INVALID_FRAMEID;
    }
    if (mb->fb.access & ACC_NATIVE) {
        return JVMDI_ERROR_OPAQUE_FRAME;
    }
    if (slot >= mb->nlocals) {
        return JVMDI_ERROR_INVALID_SLOT;
    }
    *siPtr = vars + slot;
    return JVMDI_ERROR_NONE;
}

/* Return via "valuePtr" the value of the local variable at the
 * specificied slot.
 * Note: returned objects are allocated globally and must be explictly
 * freed with DeleteGlobalRef.
 * Errors: JVMDI_ERROR_INVALID_FRAMEID, JVMDI_ERROR_INVALID_SLOT,
 * JVMDI_ERROR_TYPE_MISMATCH, JVMDI_ERROR_NULL_POINTER,
 * JVMDI_ERROR_OPAQUE_FRAME, JVMDI_ERROR_OUT_OF_MEMORY
 */
static jvmdiError JNICALL
jvmdi_GetLocalObject(jframeID frame, jint slot, 
		     jobject *valuePtr)
{
    stack_item *si;
    jvmdiError err = getSlot(frame, slot, &si);
    sys_thread_t *self = sysThreadSelf();
    ExecEnv *ee;
    JNIEnv *env;    

    DEBUG_ENABLED();
    NOT_NULL(valuePtr);
    THREAD_OK(self);

    ee = SysThread2EE(self);
    env = EE2JNIEnv(ee);

    if (err == JVMDI_ERROR_NONE) {
        jobject obj;
        if ((*env)->PushLocalFrame(env, 1) < 0) {
            return JVMDI_ERROR_OUT_OF_MEMORY;
        }
	obj = MkRefLocal(env, si->o);
        *valuePtr = (jobject)((*env)->NewGlobalRef(env, obj));
        (*env)->PopLocalFrame(env, 0);
    }
    return err;
}

/* Return via "valuePtr" the value of the local variable at the
 * specificied slot.
 * Errors: JVMDI_ERROR_INVALID_FRAMEID, JVMDI_ERROR_INVALID_SLOT,
 * JVMDI_ERROR_TYPE_MISMATCH, JVMDI_ERROR_NULL_POINTER,
 * JVMDI_ERROR_OPAQUE_FRAME
 */
static jvmdiError JNICALL
jvmdi_GetLocalInt(jframeID frame, jint slot, 
		  jint *valuePtr)
{
    stack_item *si;
    jvmdiError err = getSlot(frame, slot, &si);

    DEBUG_ENABLED();
    NOT_NULL(valuePtr);
    if (err == JVMDI_ERROR_NONE) {
        *valuePtr = (jint)si->i;
    }
    return err;
}

/* Return via "valuePtr" the value of the local variable at the
 * specificied slot.
 * Errors: JVMDI_ERROR_INVALID_FRAMEID, JVMDI_ERROR_INVALID_SLOT,
 * JVMDI_ERROR_TYPE_MISMATCH, JVMDI_ERROR_NULL_POINTER,
 * JVMDI_ERROR_OPAQUE_FRAME
 */
static jvmdiError JNICALL
jvmdi_GetLocalLong(jframeID frame, jint slot, 
		   jlong *valuePtr)
{
    jvalue tmp;
    stack_item *si;
    jvmdiError err = getSlot(frame, slot, &si);

    DEBUG_ENABLED();
    NOT_NULL(valuePtr);
    if (err == JVMDI_ERROR_NONE) {
        *valuePtr = GET_INT64(tmp, si);
    }
    return err;
}

/* Return via "valuePtr" the value of the local variable at the
 * specificied slot.
 * Errors: JVMDI_ERROR_INVALID_FRAMEID, JVMDI_ERROR_INVALID_SLOT,
 * JVMDI_ERROR_TYPE_MISMATCH, JVMDI_ERROR_NULL_POINTER,
 * JVMDI_ERROR_OPAQUE_FRAME
 */
static jvmdiError JNICALL
jvmdi_GetLocalFloat(jframeID frame, jint slot, 
		    jfloat *valuePtr)
{
    stack_item *si;
    jvmdiError err = getSlot(frame, slot, &si);

    DEBUG_ENABLED();
    NOT_NULL(valuePtr);
    if (err == JVMDI_ERROR_NONE) {
        *valuePtr = (jfloat)si->f;
    }
    return err;
}

/* Return via "valuePtr" the value of the local variable at the
 * specificied slot.
 * Errors: JVMDI_ERROR_INVALID_FRAMEID, JVMDI_ERROR_INVALID_SLOT,
 * JVMDI_ERROR_TYPE_MISMATCH, JVMDI_ERROR_NULL_POINTER,
 * JVMDI_ERROR_OPAQUE_FRAME
 */
static jvmdiError JNICALL
jvmdi_GetLocalDouble(jframeID frame, jint slot, 
		     jdouble *valuePtr)
{
    jvalue tmp;
    stack_item *si;
    jvmdiError err = getSlot(frame, slot, &si);

    DEBUG_ENABLED();
    NOT_NULL(valuePtr);
    if (err == JVMDI_ERROR_NONE) {
        *valuePtr = (jdouble)GET_DOUBLE(tmp, si);
    }
    return err;
}

/* Set the value of the local variable at the specificied slot.
 * Errors: JVMDI_ERROR_INVALID_FRAMEID, JVMDI_ERROR_INVALID_SLOT,
 * JVMDI_ERROR_TYPE_MISMATCH, JVMDI_ERROR_OPAQUE_FRAME
 */
static jvmdiError JNICALL
jvmdi_SetLocalObject(jframeID frame, jint slot, jobject value)
{
    stack_item *si;
    jvmdiError err = getSlot(frame, slot, &si);
    
    DEBUG_ENABLED();

    if (err == JVMDI_ERROR_NONE) {
        si->o = (OBJECT)DeRef(EE2JNIEnv(EE()),value);
    }
    return err;
}

/* Set the value of the local variable at the specificied slot.
 * Errors: JVMDI_ERROR_INVALID_FRAMEID, JVMDI_ERROR_INVALID_SLOT,
 * JVMDI_ERROR_TYPE_MISMATCH, JVMDI_ERROR_OPAQUE_FRAME
 */
static jvmdiError JNICALL
jvmdi_SetLocalInt(jframeID frame, jint slot, jint value)
{
    stack_item *si;
    jvmdiError err = getSlot(frame, slot, &si);

    DEBUG_ENABLED();

    if (err == JVMDI_ERROR_NONE) {
        si->i = (int)value;
    }
    return err;
}

/* Set the value of the local variable at the specificied slot.
 * Errors: JVMDI_ERROR_INVALID_FRAMEID, JVMDI_ERROR_INVALID_SLOT,
 * JVMDI_ERROR_TYPE_MISMATCH, JVMDI_ERROR_OPAQUE_FRAME
 */
static jvmdiError JNICALL
jvmdi_SetLocalLong(jframeID frame, jint slot, jlong value)
{
    jvalue tmp;
    stack_item *si;
    jvmdiError err = getSlot(frame, slot, &si);

    DEBUG_ENABLED();

    if (err == JVMDI_ERROR_NONE) {
        SET_INT64(tmp, si, value);
    }
    return err;    
}

/* Set the value of the local variable at the specificied slot.
 * Errors: JVMDI_ERROR_INVALID_FRAMEID, JVMDI_ERROR_INVALID_SLOT,
 * JVMDI_ERROR_TYPE_MISMATCH, JVMDI_ERROR_OPAQUE_FRAME
 */
static jvmdiError JNICALL
jvmdi_SetLocalFloat(jframeID frame, jint slot, jfloat value)
{
    stack_item *si;
    jvmdiError err = getSlot(frame, slot, &si);

    DEBUG_ENABLED();

    if (err == JVMDI_ERROR_NONE) {
        si->f = value;
    }
    return err;
}

/* Set the value of the local variable at the specificied slot.
 * Errors: JVMDI_ERROR_INVALID_FRAMEID, JVMDI_ERROR_INVALID_SLOT,
 * JVMDI_ERROR_TYPE_MISMATCH, JVMDI_ERROR_OPAQUE_FRAME
 */
static jvmdiError JNICALL
jvmdi_SetLocalDouble(jframeID frame, jint slot, jdouble value)
{
    jvalue tmp;
    stack_item *si;
    jvmdiError err = getSlot(frame, slot, &si);

    DEBUG_ENABLED();

    if (err == JVMDI_ERROR_NONE) {
        SET_DOUBLE(tmp, si, value);
    }
    return err;    
}

 
  /*
   *  Breakpoints
   */

/* Set a breakpoint.
 */
static jvmdiError JNICALL
jvmdi_SetBreakpoint(jclass clazz, jmethodID method, 
                    jlocation bci)
{
    struct methodblock *mb = (struct methodblock *)method;
    unsigned char *code = mb->code;
    unsigned char *pc = code + jlong_to_ptrdiff(bci);
    jvmdiError err;
    sys_thread_t *self = sysThreadSelf();

    DEBUG_ENABLED();
    THREAD_OK(self);

    if (code == NULL) {
        return JVMDI_ERROR_INVALID_METHODID;
    }
    if (jlong_ltz(bci) || jlong_ge(bci,int_to_jlong(mb->code_length))) {
        return JVMDI_ERROR_INVALID_LOCATION;
    }

    DEBUGGER_LOCK(self);
    if (bagFind(breakpoints, pc) == NULL) {
        JNIEnv *env = EE2JNIEnv(SysThread2EE(self));
        jobject classRef = (*env)->NewGlobalRef(env, clazz);

        if (classRef == NULL) {
            err = JVMDI_ERROR_OUT_OF_MEMORY;
        } else {
            struct bkpt *bp = bagAdd(breakpoints);
            if (bp == NULL) {
                (*env)->DeleteGlobalRef(env, classRef);
                err = JVMDI_ERROR_OUT_OF_MEMORY;
            } else {
                bp->pc = pc;
                bp->opcode = *pc;
                /* Keep a reference to the class to prevent gc */
                bp->classRef = classRef;
                *pc = opc_breakpoint;
                err = JVMDI_ERROR_NONE;
            }
        }
    } else {
        err = JVMDI_ERROR_DUPLICATE;
    }
    DEBUGGER_UNLOCK(self);    

    return err;
}

/* Clear a breakpoint.
 */
static jvmdiError JNICALL
jvmdi_ClearBreakpoint(jclass clazz, jmethodID method, 
                      jlocation bci)
{
    struct bkpt *bp;
    struct methodblock *mb = (struct methodblock *)method;
    unsigned char *code = mb->code;
    unsigned char *pc = code + jlong_to_ptrdiff(bci);
    jvmdiError err;
    sys_thread_t *self = sysThreadSelf();
    ExecEnv *ee;
    JNIEnv *env;    

    DEBUG_ENABLED();
    THREAD_OK(self);

    ee = SysThread2EE(self);
    env = EE2JNIEnv(ee);

    if (code == NULL) {
        return JVMDI_ERROR_INVALID_METHODID;
    }
    if (jlong_ltz(bci) || jlong_ge(bci,int_to_jlong(mb->code_length))) {
        return JVMDI_ERROR_INVALID_LOCATION;
    }

    DEBUGGER_LOCK(self);
    bp = bagFind(breakpoints, pc);
    if (bp == NULL) {
        err = JVMDI_ERROR_NOT_FOUND;
    } else {
        clear_bkpt(env, bp);
        bagDelete(breakpoints, bp);
        err = JVMDI_ERROR_NONE;
    }
    DEBUGGER_UNLOCK(self);          

    return err;
}

static bool_t
clearAllBreakpointsHelper(void *item, void *arg)
{
    return clear_bkpt(arg, item);
}

/* Clear all breakpoints.
 * Errors: JVMDI_ERROR_VM_DEAD
 */
static jvmdiError JNICALL
jvmdi_ClearAllBreakpoints()
{
    sys_thread_t *self = sysThreadSelf();

    DEBUG_ENABLED();
    THREAD_OK(self);

    DEBUGGER_LOCK(self);
    {
        ExecEnv *ee = SysThread2EE(self);
        JNIEnv *env = EE2JNIEnv(ee);
        bagEnumerateOver(breakpoints, clearAllBreakpointsHelper, 
                         (void *)env);
        bagDeleteAll(breakpoints);
    }
    DEBUGGER_UNLOCK(self);          
        
    return JVMDI_ERROR_NONE;
}

  /*
   *  Field Watching
   */

/* Set a field watch.
 */
static jvmdiError
setFieldWatch(jclass clazz, jfieldID field, 
              struct bag *watchedFields, bool_t *ptrWatching)
{
    struct fieldblock *fb = (struct fieldblock *)field;
    jvmdiError err;
    sys_thread_t *self = sysThreadSelf();

    DEBUG_ENABLED();
    THREAD_OK(self);

    DEBUGGER_LOCK(self);
    if (bagFind(watchedFields, fb) == NULL) {
        JNIEnv *env = EE2JNIEnv(SysThread2EE(self));
        jobject classRef = (*env)->NewGlobalRef(env, clazz);

        if (classRef == NULL) {
            err = JVMDI_ERROR_OUT_OF_MEMORY;
        } else {
            struct fieldWatch *fw = bagAdd(watchedFields);
            if (fw == NULL) {
                (*env)->DeleteGlobalRef(env, classRef);
                err = JVMDI_ERROR_OUT_OF_MEMORY;
            } else {
                fw->fb = fb;
                /* Keep a reference to the class to prevent gc */
                fw->classRef = classRef;
                /* reset the global flag */
                *ptrWatching = JNI_TRUE;
                err = JVMDI_ERROR_NONE;
            }
	}
    } else {
        err = JVMDI_ERROR_DUPLICATE;
    }
    DEBUGGER_UNLOCK(self);    

    return err;
}

/* Clear a field watch.
 */
static jvmdiError
clearFieldWatch(jclass clazz, jfieldID field, 
                struct bag *watchedFields, bool_t *ptrWatching)
{
    struct fieldblock *fb = (struct fieldblock *)field;
    jvmdiError err;
    sys_thread_t *self = sysThreadSelf();
    struct fieldWatch *fw;

    DEBUG_ENABLED();
    THREAD_OK(self);

    DEBUGGER_LOCK(self);
    fw = bagFind(watchedFields, fb);
    if (fw != NULL) {
        JNIEnv *env = EE2JNIEnv(SysThread2EE(self));

        /* 
         * De-reference the enclosing class so that it's 
         * GC is no longer prevented by
         * this field watch. 
         */
        (*env)->DeleteGlobalRef(env, fw->classRef);
        /* delete it from list */
        bagDelete(watchedFields, fw);
        /* reset the global flag */
        *ptrWatching = bagSize(watchedFields) > 0;
        err = JVMDI_ERROR_NONE;
    } else {
        err = JVMDI_ERROR_NOT_FOUND;
    }
    DEBUGGER_UNLOCK(self);    

    return err;
}

/* Set a field modification watch.
 */
static jvmdiError JNICALL
jvmdi_SetFieldModificationWatch(jclass clazz, jfieldID field)
{
    return setFieldWatch(clazz, field, watchedFieldModifications, 
                         &watching_field_modification);
}

/* Clear a field modification watch.
 */
static jvmdiError JNICALL
jvmdi_ClearFieldModificationWatch(jclass clazz, jfieldID field)
{
    return clearFieldWatch(clazz, field, watchedFieldModifications, 
                           &watching_field_modification);
}

/* Set a field access watch.
 */
static jvmdiError JNICALL
jvmdi_SetFieldAccessWatch(jclass clazz, jfieldID field)
{
    return setFieldWatch(clazz, field, watchedFieldAccesses, 
                         &watching_field_access);
}

/* Clear a field access watch.
 */
static jvmdiError JNICALL
jvmdi_ClearFieldAccessWatch(jclass clazz, jfieldID field)
{
    return clearFieldWatch(clazz, field, watchedFieldAccesses, 
                           &watching_field_access);
}




  /*
   *  Events
   */

/* Set the routine which will handle in-coming events.
 * Passing NULL as hook unsets hook.
 */
static jvmdiError JNICALL
jvmdi_SetEventHook(JVMDI_EventHook hook)
{
    static int handleExitSet = FALSE;
    sys_thread_t *self = sysThreadSelf();
    jvmdiError initError;
    
    DEBUG_ENABLED();
    THREAD_OK(self);

    if (handleExitSet == FALSE) {
        initError = initializeJVMDI();
        if (initError != JVMDI_ERROR_NONE) {
            return initError;
        }
    
	AtExit(handleExit);
	handleExitSet = TRUE;
    }

    eventHook = hook;
    return JVMDI_ERROR_NONE;
}

static jvmdiError JNICALL
jvmdi_SetEventNotificationMode(jint mode, jint eventType, jthread thread, ...)
{
    jboolean isEnabled = (mode == JVMDI_ENABLE);
    jvmdiError rc = JVMDI_ERROR_NONE;
    va_list argPtr;

    /* Currently, we don't accept any additional arguments. */
    va_start(argPtr, thread);
    va_end(argPtr);

    DEBUG_ENABLED();
    
    if ((eventType < 0) || (eventType > JVMDI_MAX_EVENT_TYPE_VAL)) {
        return JVMDI_ERROR_INVALID_EVENT_TYPE;
    }

    if (thread != NULL) {
        /*
         * These events cannot be controlled on a per-thread basis
         */
        if ((eventType == JVMDI_EVENT_VM_INIT) ||
            (eventType == JVMDI_EVENT_VM_DEATH) ||
            (eventType == JVMDI_EVENT_THREAD_START) ||
            (eventType == JVMDI_EVENT_CLASS_UNLOAD)) {
            return JVMDI_ERROR_INVALID_THREAD;
        }
    }

    if (eventType == JVMDI_EVENT_SINGLE_STEP) {
        if (thread == NULL) {
            rc = JVMDI_ERROR_INVALID_THREAD;
        } else {
            set_single_step_thread((TID)DeRef(EE2JNIEnv(EE()), thread), 
                                   isEnabled);
        }
    } else if (thread == NULL) {
        enableAllEvents(eventType, isEnabled);
    } else {
        sys_thread_t *self = sysThreadSelf();
        ThreadNode *node;
        DEBUGGER_LOCK(self);

        node = findThread((JHandle *)DeRef(EE2JNIEnv(EE()), thread));
        if (node == NULL) {
            rc = JVMDI_ERROR_INVALID_THREAD;
        } else {
            enableThreadEvents(node, eventType, isEnabled);
        }

        DEBUGGER_UNLOCK(self);
    }

    return rc;
}


  /*
   *  Memory allocation
   */

/* Set the routines which will perform allocation and deallocation.
 * Passing NULL in both reverts to default allocator.
 */
static jvmdiError JNICALL
jvmdi_SetAllocationHooks(JVMDI_AllocHook ahook, JVMDI_DeallocHook dhook) 
{
    DEBUG_ENABLED();

    allocHook = ahook;
    deallocHook = dhook;
    return JVMDI_ERROR_NONE;
}


/* Allocate an area of memory of "size" bytes.  Return via "memPtr" 
 * a pointer to the beginning of this area. Use the allocator set
 * by JVMDI_SetAllocationHooks().
 * Errors: JVMDI_ERROR_NULL_POINTER, JVMDI_ERROR_OUT_OF_MEMORY
 */
static jvmdiError JNICALL
jvmdi_Allocate(jlong size, jbyte** memPtr) 
{
    jbyte *mem;
    
    DEBUG_ENABLED();
    NOT_NULL(memPtr);
    if (allocHook != NULL) {
      return allocHook(size, memPtr);
    }
    mem = (jbyte *)sysMalloc(jlong_to_size(size));
    if (mem == NULL) {
        return JVMDI_ERROR_OUT_OF_MEMORY;
    }
    *memPtr = mem;
    return JVMDI_ERROR_NONE;
}

/* Deallocate an area of memory of "mem" which was allocated by the
 * allocator set by JVMDI_SetAllocationHooks(). Use the deallocator set
 * by JVMDI_SetAllocationHooks().
 * Errors: JVMDI_ERROR_NULL_POINTER
 */
static jvmdiError JNICALL
jvmdi_Deallocate(jbyte* mem) 
{
    DEBUG_ENABLED();
    NOT_NULL(mem);
    if (deallocHook != NULL) {
      return deallocHook(mem);
    }
    sysFree(mem);
    return JVMDI_ERROR_NONE;
}


  /*
   *  Class access
   */

/* Return via "sigPtr" the class's signature (UTF8).  Form
 * is a JNI signature.  Examples:  "Ljava/io/File;"
 * "[Ljava/util/Date;" (an array of dates).
 * Memory for this string is allocated by the function specified 
 * in JVMDI_SetAllocationHooks().  
 * Errors: JVMDI_ERROR_INVALID_CLASS, JVMDI_ERROR_NULL_POINTER,
 * JVMDI_ERROR_OUT_OF_MEMORY
 */
static jvmdiError JNICALL
jvmdi_GetClassSignature(jclass clazz, char **sigPtr) 
{
    char *name;
    int len;
    ClassClass *cb;

    DEBUG_ENABLED();

    cb = (ClassClass *)DeRef(EE2JNIEnv(EE()), clazz);
    VALID_CLASS(cb);

    name = cbName(cb);
    len = strlen(name);
    if (name[0] == '[') { /* arrays are already in signature form */

        ALLOC(size_to_jlong(len+1), sigPtr);
        strcpy(*sigPtr, name);
    } else {
        char *sig;
        ALLOC(size_to_jlong(len+3), sigPtr);
        sig = *sigPtr;
        sig[0] = 'L';
        strcpy(sig+1, name);
        sig[len+1] = ';';
        sig[len+2] = 0;
    }        
    return JVMDI_ERROR_NONE;
}

/* Return via "statePtr" the current state bits of this class.
 * Errors: JVMDI_ERROR_INVALID_CLASS, JVMDI_ERROR_NULL_POINTER
 */
static jvmdiError JNICALL
jvmdi_GetClassStatus(jclass clazz, jint *statePtr) {
    jint state = 0;
    ClassClass *cb;

    DEBUG_ENABLED();
    NOT_NULL(statePtr);

    cb = (ClassClass *)DeRef(EE2JNIEnv(EE()), clazz);
    VALID_CLASS(cb);

    if (CCIs(cb, Verified)) {
        state |= JVMDI_CLASS_STATUS_VERIFIED;
    }

    if (CCIs(cb, FieldPrepared) &&
        CCIs(cb, MethodPrepared) &&
        CCIs(cb, InterfacePrepared)) {
        state |= JVMDI_CLASS_STATUS_PREPARED;
    }

    if (CCIs(cb, Initialized)) {
        state |= JVMDI_CLASS_STATUS_INITIALIZED;
    }

    if (CCIs(cb, Error)) {
        state |= JVMDI_CLASS_STATUS_ERROR;
    }

    *statePtr = state;
    return JVMDI_ERROR_NONE;
}

/* Return via "sourceNamePtr" the class's source path (UTF8).
 * Memory for this string is allocated by the function specified 
 * in JVMDI_SetAllocationHooks().
 * Errors: JVMDI_ERROR_INVALID_CLASS, JVMDI_ERROR_NULL_POINTER,
 * JVMDI_ERROR_OUT_OF_MEMORY, JVMDI_ERROR_ABSENT_INFORMATION
 */
static jvmdiError JNICALL
jvmdi_GetSourceFileName(jclass clazz, char **sourceNamePtr) 
{
    char *srcName;
    ClassClass *cb;

    DEBUG_ENABLED();

    cb = (ClassClass *)DeRef(EE2JNIEnv(EE()), clazz);
    VALID_CLASS(cb);
    srcName = cbSourceName(cb);
    if (srcName == NULL) {
        return JVMDI_ERROR_ABSENT_INFORMATION;
    }
    ALLOC(size_to_jlong(strlen(srcName)+1), sourceNamePtr);
    strcpy(*sourceNamePtr, srcName);
    return JVMDI_ERROR_NONE;
}

/* Return via "modifiersPtr" the modifiers of this class.
 * See JVM Spec "access_flags".
 * Errors: JVMDI_ERROR_INVALID_CLASS, JVMDI_ERROR_NULL_POINTER
 */
static jvmdiError JNICALL
jvmdi_GetClassModifiers(jclass clazz, jint *modifiersPtr) 
{
    sys_thread_t *self = sysThreadSelf();
    ExecEnv *ee;
    JNIEnv *env;    

    DEBUG_ENABLED();
    NOT_NULL(modifiersPtr);
    THREAD_OK(self);

    ee = SysThread2EE(self);
    env = EE2JNIEnv(ee);

    *modifiersPtr = JVM_GetClassModifiers(env, clazz);
    return JVMDI_ERROR_NONE;
}

/* Return via "methodsPtr" a list of the class's methods and 
 * constructors ("methodCountPtr" returns the number of methods and
 * constructors).
 * Memory for this list is allocated by the function specified 
 * in JVMDI_SetAllocationHooks().
 * Inherited methods are not included.
 * Errors: JVMDI_ERROR_INVALID_CLASS, JVMDI_ERROR_NULL_POINTER,
 * JVMDI_ERROR_OUT_OF_MEMORY
 */
static jvmdiError JNICALL
jvmdi_GetClassMethods(jclass clazz, 
                      jint *methodCountPtr, jmethodID **methodsPtr) 
{
    jint methods_count;
    struct methodblock *mbs;
    jmethodID *mids;
    int i;
    jint state;
    ClassClass *cb; 

    DEBUG_ENABLED();
    NOT_NULL2(methodCountPtr, methodsPtr);

    cb = (ClassClass *)DeRef(EE2JNIEnv(EE()), clazz);
    VALID_CLASS(cb);

    jvmdi_GetClassStatus(clazz, &state);
    if (!(state & JVMDI_CLASS_STATUS_PREPARED)) {
        return JVMDI_ERROR_CLASS_NOT_PREPARED;
    }

    methods_count = cbMethodsCount(cb);
    *methodCountPtr = methods_count;
    ALLOC(size_to_jlong(methods_count*sizeof(jmethodID)), methodsPtr);
    mbs = cbMethods(cb);
    mids = *methodsPtr;
    for (i = 0; i < methods_count; ++i) {
        mids[i] = (jmethodID)mbs++;
    }
    return JVMDI_ERROR_NONE;
}

/* Return via "fieldsPtr" a list of the class's fields ("fieldCountPtr" 
 * returns the number of fields).
 * Memory for this list is allocated by the function specified 
 * in JVMDI_SetAllocationHooks().
 * Inherited fields are not included.
 * Errors: JVMDI_ERROR_INVALID_CLASS, JVMDI_ERROR_NULL_POINTER,
 * JVMDI_ERROR_OUT_OF_MEMORY
 */
static jvmdiError JNICALL
jvmdi_GetClassFields(jclass clazz, 
                     jint *fieldCountPtr, jfieldID **fieldsPtr) 
{
    jint fields_count;
    struct fieldblock *fbs;
    jfieldID *fids;
    int i;
    jint state;
    ClassClass *cb;

    DEBUG_ENABLED();
    NOT_NULL2(fieldCountPtr, fieldsPtr);

    cb = (ClassClass *)DeRef(EE2JNIEnv(EE()), clazz);
    VALID_CLASS(cb);
    
    jvmdi_GetClassStatus(clazz, &state);
    if (!(state & JVMDI_CLASS_STATUS_PREPARED)) {
        return JVMDI_ERROR_CLASS_NOT_PREPARED;
    }

    fields_count = cbFieldsCount(cb);
    *fieldCountPtr = fields_count;
    ALLOC(size_to_jlong(fields_count*sizeof(jfieldID)), fieldsPtr);
    fbs = cbFields(cb);
    fids = *fieldsPtr;
    for (i = 0; i < fields_count; ++i) {
        fids[i] = (jfieldID)fbs++;
    }
    return JVMDI_ERROR_NONE;
}

static jvmdiError 
resolveConstantPoolIndex(ExecEnv *ee, JNIEnv *env, 
			 ClassClass *cb, unsigned short cpIdx, 
			 jclass *classPtr) 
{
    static unsigned mask = 1 << CONSTANT_Class;

    if (ResolveClassConstantFromClass2(cb, cpIdx, ee, mask, FALSE)) {
        ClassClass *cbRes = cbConstantPool(cb)[cpIdx].clazz;
        jobject lclazz;
        if ((*env)->PushLocalFrame(env, 1) < 0) {
            return JVMDI_ERROR_OUT_OF_MEMORY;
        }
        lclazz = MkRefLocal(env, cbRes);
        *classPtr = (jclass)((*env)->NewGlobalRef(env, lclazz));
        (*env)->PopLocalFrame(env, 0);
    } else {
        /* I don't really know what to do in the event of a resolve */
        /* failure */
        *classPtr = NULL;
    }
    return JVMDI_ERROR_NONE;
}

/* Return via "interfacesPtr" a list of the interfaces this class
 * declares ("interfaceCountPtr" returns the number of such interfaces).  
 * Memory for this list is allocated by the function specified 
 * in JVMDI_SetAllocationHooks().
 * Note: classes in the list are a JNI global reference and must
 * be explicitly managed.
 * Errors: JVMDI_ERROR_INVALID_CLASS, JVMDI_ERROR_NULL_POINTER,
 * JVMDI_ERROR_OUT_OF_MEMORY
 */
static jvmdiError JNICALL
jvmdi_GetImplementedInterfaces(jclass clazz,
                               jint *interfaceCountPtr, 
                               jclass **interfacesPtr) 
{
    jint interfaces_count;
    int i;
    unsigned short *cpIndices;
    jclass *imps;
    jint state;
    ClassClass *cb;   
    sys_thread_t *self = sysThreadSelf();
    ExecEnv *ee;
    JNIEnv *env;    

    DEBUG_ENABLED();
    NOT_NULL2(interfaceCountPtr, interfacesPtr);
    THREAD_OK(self);

    ee = SysThread2EE(self);
    env = EE2JNIEnv(ee);

    cb = (ClassClass *)DeRef(EE2JNIEnv(EE()), clazz);
    VALID_CLASS(cb);

    jvmdi_GetClassStatus(clazz, &state);
    if (!(state & JVMDI_CLASS_STATUS_PREPARED)) {
        return JVMDI_ERROR_CLASS_NOT_PREPARED;
    }

    interfaces_count = cbImplementsCount(cb);
    *interfaceCountPtr = interfaces_count;
    ALLOC(size_to_jlong(interfaces_count * sizeof(jclass)), interfacesPtr);
    cpIndices = cbImplements(cb);
    imps = *interfacesPtr;
    for (i = 0; i < interfaces_count; ++i) {
        jclass imp;
        jvmdiError err = resolveConstantPoolIndex(ee, env, cb, 
						  *cpIndices++, &imp);
        if (err != JVMDI_ERROR_NONE) {
            return err;
        }
        imps[i] = imp;
    }
    return JVMDI_ERROR_NONE;
}

/* Return via "isInterfacePtr" a boolean indicating whether this "clazz"
 * is an interface.
 * Errors: JVMDI_ERROR_INVALID_CLASS, JVMDI_ERROR_NULL_POINTER
 */
static jvmdiError JNICALL
jvmdi_IsInterface(jclass clazz, jboolean *isInterfacePtr) 
{
    sys_thread_t *self = sysThreadSelf();
    ExecEnv *ee;
    JNIEnv *env;    

    DEBUG_ENABLED();
    THREAD_OK(self);

    ee = SysThread2EE(self);
    env = EE2JNIEnv(ee);

    NOT_NULL(isInterfacePtr);
    *isInterfacePtr = JVM_IsInterface(env, clazz);
    return JVMDI_ERROR_NONE;
}


/* Return via "isArrayclassPtr" a boolean indicating whether this "clazz"
 * is an array class.
 * Errors: JVMDI_ERROR_INVALID_CLASS, JVMDI_ERROR_NULL_POINTER
 */
static jvmdiError JNICALL
jvmdi_IsArrayClass(jclass clazz, jboolean *isArrayClassPtr) 
{
    sys_thread_t *self = sysThreadSelf();
    ExecEnv *ee;
    JNIEnv *env;    

    DEBUG_ENABLED();
    THREAD_OK(self);

    ee = SysThread2EE(self);
    env = EE2JNIEnv(ee);

    NOT_NULL(isArrayClassPtr);
    *isArrayClassPtr = JVM_IsArrayClass(env, clazz);
    return JVMDI_ERROR_NONE;
}

/* Return via "classloaderPtr" the class loader that created this class
 * or interface, null if this class was not created by a class loader.
 * Note: object returned via 'classloaderPtr' is a JNI global 
 * reference and must be explicitly managed.
 * Errors: JVMDI_ERROR_INVALID_CLASS, JVMDI_ERROR_NULL_POINTER,
 * JVMDI_ERROR_OUT_OF_MEMORY
 */
static jvmdiError JNICALL
jvmdi_GetClassLoader(jclass clazz, jobject *classloaderPtr) 
{
    jobject loader;
    sys_thread_t *self = sysThreadSelf();
    ExecEnv *ee;
    JNIEnv *env;    

    DEBUG_ENABLED();
    NOT_NULL(classloaderPtr);
    THREAD_OK(self);

    ee = SysThread2EE(self);
    env = EE2JNIEnv(ee);

    if ((*env)->PushLocalFrame(env, 1) < 0) {
        return JVMDI_ERROR_OUT_OF_MEMORY;
    }
    loader = JVM_GetClassLoader(env, clazz);
    *classloaderPtr = (jobject)((*env)->NewGlobalRef(env, loader));
    (*env)->PopLocalFrame(env, 0);
    return JVMDI_ERROR_NONE;
}


  /*
   *  Object access
   */

/* Return via "hashCodePtr" a hash code that can be used in maintaining
 * hash table of object references. This function guarantees the same 
 * hash code value for a particular object throughout its life
 * Errors: JVMDI_ERROR_INVALID_OBJECT, JVMDI_ERROR_NULL_POINTER
 */
static jvmdiError JNICALL
jvmdi_GetObjectHashCode(jobject object, jint *hashCodePtr) 
{
    DEBUG_ENABLED();
    NOT_NULL(hashCodePtr);
    *hashCodePtr = JVM_IHashCode(EE2JNIEnv(EE()), object);
    return JVMDI_ERROR_NONE;
}

#define SysThread2JThread(env, systhread) \
  (*env)->NewGlobalRef(env, MkRefLocal(env, SysThread2EE(systhread)->thread));

static jvmdiError 
SysThreads2JThreads(JNIEnv *env, sys_thread_t **threads, int count) 
{
    sys_thread_t **thread;
    jvmdiError rc = JVMDI_ERROR_NONE;
    int i;

    if ((*env)->PushLocalFrame(env, count) < 0) {
        rc = JVMDI_ERROR_OUT_OF_MEMORY;
    } else {
        for (i = 0, thread = threads; i < count; i++, thread++) {
            *(jthread *)thread = SysThread2JThread(env, *thread);
            if (*thread == NULL) {
                rc = JVMDI_ERROR_OUT_OF_MEMORY;
                break;
            }
        }
        (*env)->PopLocalFrame(env, 0);
    }

    if (rc != JVMDI_ERROR_NONE) {
        deleteRefArray(env, (jthread *)threads, i);
    }

    return rc;
}

/* Return via "infoPtr" monitor info for the given object. 
 * All application threads should be suspended before calling this function.
 * Memory for the waiter arrays is allocated by the function specified 
 * in JVMDI_SetAllocationHooks().
 * Note: returned objects are allocated globally and must be explictly
 * freed with DeleteGlobalRef.
 * Errors: JVMDI_ERROR_INVALID_OBJECT, JVMDI_ERROR_THREAD_NOT_SUSPENDED,
 * JVMDI_ERROR_OUT_OF_MEMORY
 */


static jvmdiError JNICALL
jvmdi_GetMonitorInfo(jobject object, JVMDI_monitor_info *infoPtr) 
{
#ifdef FAST_MONITOR

    /*
     * Note: this ought to be implementable with fast monitors. All it needs
     * is an implementation of  lookupMonitor which is not currently there.
     */
    return JVMDI_ERROR_NOT_IMPLEMENTED;
#else
    sys_thread_t *self = sysThreadSelf();
    JNIEnv *env;
    uintptr_t monitorKey;
    monitor_t *monitor;
    jvmdiError rc;
    
    sysAssert(sizeof(jthread) == sizeof(sys_thread_t *));

    DEBUG_ENABLED();
    NOT_NULL(infoPtr);
    THREAD_OK(self);

    env = EE2JNIEnv(EE());

    monitorKey = obj_monitor(DeRef(env, object)); 
    monitor = lookupMonitor(self, monitorKey, FALSE);

    if (monitor == MID_NULL) {
        infoPtr->owner = NULL;
        infoPtr->entry_count = 0;
        infoPtr->waiter_count = 0;
        infoPtr->waiters = NULL;
    } else {
        sys_mon_t *sysMonitor = sysmon(monitor);
        sys_mon_info sysInfo;
        int i;
        jthread *thread;
        int totalWaiters;

        /*
         * Use the first call to count waiters
         */
        sysInfo.sz_monitor_waiters = 0;
        sysInfo.sz_condvar_waiters = 0;
        QUEUE_LOCK(self);
        sysMonitorGetInfo(sysMonitor, &sysInfo);
        QUEUE_UNLOCK(self);

        /*
         * Allocate buffer based on counts
         */
        totalWaiters = sysInfo.n_monitor_waiters + sysInfo.n_condvar_waiters;
        if (totalWaiters > 0) {
            ALLOC(size_to_jlong(totalWaiters * sizeof(sys_thread_t *)),
		  &infoPtr->waiters);
        } else {
            infoPtr->waiters = NULL;
        }

        /*
         * Set up waiter buffers and make the second call to get all the
         * info.
         */
        sysInfo.sz_monitor_waiters = sysInfo.n_monitor_waiters;
        sysInfo.sz_condvar_waiters = sysInfo.n_condvar_waiters;
        sysInfo.monitor_waiters = (sys_thread_t **)infoPtr->waiters;
        sysInfo.condvar_waiters = sysInfo.monitor_waiters + 
                                  sysInfo.n_monitor_waiters;

        QUEUE_LOCK(self);
        sysMonitorGetInfo(sysMonitor, &sysInfo);
        QUEUE_UNLOCK(self);

        if (sysInfo.owner != NULL) {
            if ((*env)->PushLocalFrame(env, 1) < 0) {
                jvmdi_Deallocate((jbyte *)infoPtr->waiters);
                return JVMDI_ERROR_OUT_OF_MEMORY;
            }
            infoPtr->owner = SysThread2JThread(env, sysInfo.owner);
            infoPtr->entry_count = sysInfo.entry_count;
            (*env)->PopLocalFrame(env, 0);
        } else {
            infoPtr->owner = NULL;
            infoPtr->entry_count = 0;
        }
        if ((sysInfo.sz_monitor_waiters != sysInfo.n_monitor_waiters) ||
            (sysInfo.sz_condvar_waiters != sysInfo.n_condvar_waiters)) {
            /* 
             * Something changed between calls. That means a significant
             * thread was running
             */
            jvmdi_Deallocate((jbyte *)infoPtr->waiters);
            return JVMDI_ERROR_THREAD_NOT_SUSPENDED;
        }
        infoPtr->waiter_count = totalWaiters;

        rc = SysThreads2JThreads(env, (sys_thread_t **)infoPtr->waiters,
                                 infoPtr->waiter_count);
        if (rc != JVMDI_ERROR_NONE) {
            jvmdi_Deallocate((jbyte *)infoPtr->waiters);
            return rc;
        }
    }
   
    return JVMDI_ERROR_NONE;
#endif /* FAST_MONITOR */    
}

  /*
   *  Field access
   */

/* Return via "namePtr" the field's name and via "signaturePtr" the
 * field's signature (UTF8).
 * Memory for these strings is allocated by the function specified 
 * in JVMDI_SetAllocationHooks().
 * Errors: JVMDI_ERROR_INVALID_FIELDID, JVMDI_ERROR_INVALID_CLASS,
 * JVMDI_ERROR_NULL_POINTER, JVMDI_ERROR_OUT_OF_MEMORY
 */
static jvmdiError JNICALL
jvmdi_GetFieldName(jclass clazz, jfieldID field, 
                   char **namePtr, char **signaturePtr)
{
    char *name;
    char *sig;
    struct fieldblock *fb = (struct fieldblock *)field;
    
    DEBUG_ENABLED();
    NOT_NULL2(namePtr, signaturePtr);
    if (fb == NULL) {
        return JVMDI_ERROR_INVALID_FIELDID;
    }

    name = fb->name;
    ALLOC(size_to_jlong(strlen(name)+1), namePtr);
    strcpy(*namePtr, name);

    sig = fb->signature;
    ALLOC(size_to_jlong(strlen(sig)+1), signaturePtr);
    strcpy(*signaturePtr, sig);

    return JVMDI_ERROR_NONE;
}

/* Return via "declaringClassPtr" the class in which this field is
 * defined.
 * Note: class returned via 'declaringClassPtr' is a JNI global 
 * reference and must be explicitly managed.
 * Errors: JVMDI_ERROR_INVALID_FIELDID, JVMDI_ERROR_INVALID_CLASS,
 * JVMDI_ERROR_NULL_POINTER, JVMDI_ERROR_OUT_OF_MEMORY
 */
static jvmdiError JNICALL
jvmdi_GetFieldDeclaringClass(jclass clazz, jfieldID field,
 			     jclass *declaringClassPtr)
{
    struct fieldblock *fb = (struct fieldblock *)field;
    jobject dclazz;
    sys_thread_t *self = sysThreadSelf();
    ExecEnv *ee;
    JNIEnv *env;    

    DEBUG_ENABLED();
    NOT_NULL(declaringClassPtr);
    THREAD_OK(self);

    ee = SysThread2EE(self);
    env = EE2JNIEnv(ee);

    if ((*env)->PushLocalFrame(env, 1) < 0) {
        return JVMDI_ERROR_OUT_OF_MEMORY;
    }
    dclazz = MkRefLocal(env, fb->clazz);
    *declaringClassPtr = (jobject)((*env)->NewGlobalRef(env, dclazz));
    (*env)->PopLocalFrame(env, 0);
    return JVMDI_ERROR_NONE;
}

/* Return via "modifiersPtr" the modifiers of this field.
 * See JVM Spec "access_flags".
 * Errors: JVMDI_ERROR_INVALID_FIELDID, JVMDI_ERROR_INVALID_CLASS,
 * JVMDI_ERROR_NULL_POINTER
 */
static jvmdiError JNICALL
jvmdi_GetFieldModifiers(jclass clazz, jfieldID field,
                        jint *modifiersPtr)
{
    struct fieldblock *fb = (struct fieldblock *)field;
    
    DEBUG_ENABLED();
    NOT_NULL(modifiersPtr);
    *modifiersPtr = fb->access;
    return JVMDI_ERROR_NONE;
}

/* Return via "isSyntheticPtr" a boolean indicating whether the field
 * is synthetic.
 * Errors: JVMDI_ERROR_INVALID_FIELDID, JVMDI_ERROR_INVALID_CLASS,
 * JVMDI_ERROR_NULL_POINTER
 */
static jvmdiError JNICALL
jvmdi_IsFieldSynthetic(jclass clazz, jfieldID field, 
		        jboolean *isSyntheticPtr)
{
    /* Synthetic attribute info is not yet available in this VM */
    return JVMDI_ERROR_NOT_IMPLEMENTED;
}


  /*
   *  Method access
   */

/* Return via "namePtr" the method's name and via "signaturePtr" the
 * method's signature (UTF8).
 * Memory for these strings is allocated by the function specified 
 * in JVMDI_SetAllocationHooks().
 * Errors: JVMDI_ERROR_INVALID_METHODID, JVMDI_ERROR_INVALID_CLASS,
 * JVMDI_ERROR_NULL_POINTER, JVMDI_ERROR_OUT_OF_MEMORY
 */
static jvmdiError JNICALL
jvmdi_GetMethodName(jclass clazz, jmethodID method, 
                    char **namePtr, char **signaturePtr)
{
    char *name;
    char *sig;
    struct methodblock *mb = (struct methodblock *)method;
    
    DEBUG_ENABLED();
    NOT_NULL2(namePtr, signaturePtr);
    if (mb == NULL) {
        return JVMDI_ERROR_INVALID_METHODID;
    }

    name = mb->fb.name;
    ALLOC(size_to_jlong(strlen(name)+1), namePtr);
    strcpy(*namePtr, name);

    sig = mb->fb.signature;
    ALLOC(size_to_jlong(strlen(sig)+1), signaturePtr);
    strcpy(*signaturePtr, sig);

    return JVMDI_ERROR_NONE;
}

/* Return via "declaringClassPtr" the class in which this method is
 * defined.
 * Note: class returned via 'declaringClassPtr' is a JNI global 
 * reference and must be explicitly managed.
 * Errors: JVMDI_ERROR_INVALID_METHODID, JVMDI_ERROR_INVALID_CLASS,
 * JVMDI_ERROR_NULL_POINTER, JVMDI_ERROR_OUT_OF_MEMORY
 */
static jvmdiError JNICALL
jvmdi_GetMethodDeclaringClass(jclass clazz, jmethodID method,
			      jclass *declaringClassPtr)
{
    struct methodblock *mb = (struct methodblock *)method;
    jobject dclazz;
    sys_thread_t *self = sysThreadSelf();
    ExecEnv *ee;
    JNIEnv *env;    

    DEBUG_ENABLED();
    NOT_NULL(declaringClassPtr);
    THREAD_OK(self);

    ee = SysThread2EE(self);
    env = EE2JNIEnv(ee);

    if ((*env)->PushLocalFrame(env, 1) < 0) {
        return JVMDI_ERROR_OUT_OF_MEMORY;
    }
    dclazz = MkRefLocal(env, mb->fb.clazz);
    *declaringClassPtr = (jclass)((*env)->NewGlobalRef(env, dclazz));
    (*env)->PopLocalFrame(env, 0);
    return JVMDI_ERROR_NONE;
}

/* Return via "modifiersPtr" the modifiers of this method.
 * See JVM Spec "access_flags".
 * Errors: JVMDI_ERROR_INVALID_METHODID, JVMDI_ERROR_INVALID_CLASS,
 * JVMDI_ERROR_NULL_POINTER
 */
static jvmdiError JNICALL
jvmdi_GetMethodModifiers(jclass clazz, jmethodID method,
                         jint *modifiersPtr)
{
    struct methodblock *mb = (struct methodblock *)method;
    
    DEBUG_ENABLED();
    NOT_NULL(modifiersPtr);
    *modifiersPtr = mb->fb.access;
    return JVMDI_ERROR_NONE;
}

/* Return via "maxPtr" the maximum number of words on the operand
 * stack at any point in during the execution of this method.
 * Errors: JVMDI_ERROR_INVALID_METHODID, JVMDI_ERROR_INVALID_CLASS,
 * JVMDI_ERROR_NULL_POINTER
 */
static jvmdiError JNICALL
jvmdi_GetMaxStack(jclass clazz, jmethodID method, jint *maxPtr)
{
    struct methodblock *mb = (struct methodblock *)method;
    
    DEBUG_ENABLED();
    NOT_NULL(maxPtr);
    *maxPtr = mb->maxstack;
    return JVMDI_ERROR_NONE;
}

/* Return via "maxPtr" the number of local variable slots used by 
 * this method. Note two-word locals use two slots.
 * Errors: JVMDI_ERROR_INVALID_METHODID, JVMDI_ERROR_INVALID_CLASS,
 * JVMDI_ERROR_NULL_POINTER
 */
static jvmdiError JNICALL
jvmdi_GetMaxLocals(jclass clazz, jmethodID method, jint *maxPtr)
{
    struct methodblock *mb = (struct methodblock *)method;
    
    DEBUG_ENABLED();
    NOT_NULL(maxPtr);
    *maxPtr = mb->nlocals;
    return JVMDI_ERROR_NONE;
}

/* Return via "sizePtr" the number of local variable slots used by 
 * arguments. Note two-word arguments use two slots.
 * Errors: JVMDI_ERROR_INVALID_METHODID, JVMDI_ERROR_INVALID_CLASS,
 * JVMDI_ERROR_NULL_POINTER
 */
static jvmdiError JNICALL
jvmdi_GetArgumentsSize(jclass clazz, jmethodID method, jint *sizePtr)
{
    struct methodblock *mb = (struct methodblock *)method;
    
    DEBUG_ENABLED();
    NOT_NULL(sizePtr);
    *sizePtr = mb->args_size;
    return JVMDI_ERROR_NONE;
}

/* Return via "tablePtr" the line number table ("entryCountPtr" 
 * returns the number of entries in the table).
 * Memory for this table is allocated by the function specified 
 * in JVMDI_SetAllocationHooks().
 * Errors: JVMDI_ERROR_INVALID_METHODID, JVMDI_ERROR_INVALID_CLASS,
 * JVMDI_ERROR_NULL_POINTER, JVMDI_ERROR_OUT_OF_MEMORY,
 * JVMDI_ERROR_ABSENT_INFORMATION
 */
static jvmdiError JNICALL
jvmdi_GetLineNumberTable(jclass clazz, jmethodID method,
 			 jint *entryCountPtr, 
                         JVMDI_line_number_entry **tablePtr)
{
    int i;
    struct methodblock *mb = (struct methodblock *)method;
    struct lineno *vmtbl = mb->line_number_table;
    uint32_t length = mb->line_number_table_length;
    JVMDI_line_number_entry *tbl;

    DEBUG_ENABLED();
    NOT_NULL2(entryCountPtr, tablePtr);

    if (vmtbl == NULL) {
        return JVMDI_ERROR_ABSENT_INFORMATION;
    }
    ALLOC(size_to_jlong(length * sizeof(JVMDI_line_number_entry)), tablePtr);
    *entryCountPtr = (jint)length;
    tbl = *tablePtr;
    for (i = 0; i < length; ++i) {
        tbl[i].start_location = int_to_jlong(vmtbl[i].pc);
        tbl[i].line_number = (jint)(vmtbl[i].line_number);
    }
    return JVMDI_ERROR_NONE;
}

/* Return via "startLocationPtr" the first location in the method.
 * Return via "endLocationPtr" the last location in the method.
 * If conventional byte code index locations are being used then
 * this returns zero and the number of byte codes minus one
 * (respectively).
 * If location information is not available return 
 * JVMDI_ERROR_ABSENT_INFORMATION and negative one through the pointers.
 * Errors: JVMDI_ERROR_INVALID_METHODID, JVMDI_ERROR_INVALID_CLASS,
 * JVMDI_ERROR_NULL_POINTER, JVMDI_ERROR_ABSENT_INFORMATION
 */
static jvmdiError JNICALL
jvmdi_GetMethodLocation(jclass clazz, jmethodID method,
                        jlocation *startLocationPtr, 
                        jlocation *endLocationPtr)
{
    struct methodblock *mb = (struct methodblock *)method;
    
    DEBUG_ENABLED();
    NOT_NULL2(startLocationPtr, endLocationPtr);

    *startLocationPtr = jlong_zero;
    *endLocationPtr = int_to_jlong(mb->code_length - 1);
    return JVMDI_ERROR_NONE;
}

/* Return via "tablePtr" the local variable table ("entryCountPtr" 
 * returns the number of entries in the table).
 * Memory for this table and for each "name" and "signature" 
 * string is allocated by the function specified in
 * JVMDI_SetAllocationHooks().
 * Errors: JVMDI_ERROR_INVALID_METHODID, JVMDI_ERROR_INVALID_CLASS,
 * JVMDI_ERROR_NULL_POINTER, JVMDI_ERROR_OUT_OF_MEMORY,
 * JVMDI_ERROR_ABSENT_INFORMATION
 */
static jvmdiError JNICALL
jvmdi_GetLocalVariableTable(jclass clazz, jmethodID method,
                            jint *entryCountPtr, 
                            JVMDI_local_variable_entry **tablePtr)
{
    int i;
    struct methodblock *mb = (struct methodblock *)method;
    struct localvar *vmtbl = mb->localvar_table;
    uint32_t length = mb->localvar_table_length;
    ClassClass *cb = mb->fb.clazz;
    union cp_item_type *constantpool = cbConstantPool(cb);
    JVMDI_local_variable_entry *tbl;
    
    DEBUG_ENABLED();
    NOT_NULL2(entryCountPtr, tablePtr)

    if (vmtbl == NULL) {
        return JVMDI_ERROR_ABSENT_INFORMATION;
    }
    ALLOC(size_to_jlong(length * sizeof(JVMDI_local_variable_entry)),
	  tablePtr);
    *entryCountPtr = (jint)length;
    tbl = *tablePtr;
    for (i = 0; i < length; ++i) {
        char *src;
        char *buf;

        tbl[i].start_location = int_to_jlong(vmtbl[i].pc0);

        tbl[i].length = (jint)(vmtbl[i].length);

        src = constantpool[vmtbl[i].nameoff].cp;
        ALLOC(size_to_jlong(strlen(src)+1), &buf);
        strcpy(buf, src);
        tbl[i].name = buf;

        src = constantpool[vmtbl[i].sigoff].cp;
        ALLOC(size_to_jlong(strlen(src)+1), &buf);
        strcpy(buf, src);
        tbl[i].signature = buf;

        tbl[i].slot = (jint)(vmtbl[i].slot);
    }
    return JVMDI_ERROR_NONE;
}

/* Return via "tablePtr" the exception handler table ("entryCountPtr" 
 * returns the number of entries in the table).
 * Memory for this table is allocated by the function specified 
 * in JVMDI_SetAllocationHooks().
 * Note: the classes returned in the "exception" field of
 * JVMDI_exception_handler_entry are JNI global references and 
 * must be explicitly managed.
 * Errors: JVMDI_ERROR_INVALID_METHODID, JVMDI_ERROR_INVALID_CLASS,
 * JVMDI_ERROR_NULL_POINTER
 */
static jvmdiError JNICALL
jvmdi_GetExceptionHandlerTable(jclass clazz, jmethodID method,
                               jint *entryCountPtr, 
                               JVMDI_exception_handler_entry **tablePtr)
{
    int i;
    struct methodblock *mb = (struct methodblock *)method;
    struct CatchFrame *vmtbl = mb->exception_table;
    uint32_t length = mb->exception_table_length;
    ClassClass *cb = mb->fb.clazz;
    JVMDI_exception_handler_entry *tbl;
    sys_thread_t *self = sysThreadSelf();
    ExecEnv *ee;
    JNIEnv *env;    

    DEBUG_ENABLED();
    NOT_NULL2(entryCountPtr, tablePtr)
    THREAD_OK(self);

    ee = SysThread2EE(self);
    env = EE2JNIEnv(ee);

    if (vmtbl == NULL) {
        return JVMDI_ERROR_ABSENT_INFORMATION;
    }
    ALLOC(size_to_jlong(length * sizeof(JVMDI_exception_handler_entry)),
	  tablePtr);
    *entryCountPtr = (jint)length;
    tbl = *tablePtr;
    for (i = 0; i < length; ++i) {
        jclass catchClass;
        jvmdiError err;

        tbl[i].start_location = int_to_jlong(vmtbl[i].start_pc);
        tbl[i].end_location = int_to_jlong(vmtbl[i].end_pc);
        tbl[i].handler_location = int_to_jlong(vmtbl[i].handler_pc);

        err = resolveConstantPoolIndex(ee, env, cb, vmtbl[i].catchType, 
				       &catchClass);
        if (err != JVMDI_ERROR_NONE) {
            return err;
        }
        tbl[i].exception = catchClass;
    }
    return JVMDI_ERROR_NONE;
}

/* Return via "exceptionsPtr" an array of the checked exceptions this method
 * may throw ("exceptionCountPtr" returns the number of such exceptions).
 * Memory for this array is allocated by the function specified 
 * in JVMDI_SetAllocationHooks().
 * Note: classes in the array are JNI global references and must
 * be explicitly managed.
 * Errors: JVMDI_ERROR_NULL_POINTER, JVMDI_ERROR_OUT_OF_MEMORY
 */
static jvmdiError JNICALL
jvmdi_GetThrownExceptions(jclass clazz, jmethodID method,
                          jint *exceptionCountPtr, jclass **exceptionsPtr)
{
    int i;
    struct methodblock *mb = (struct methodblock *)method;
    unsigned short *vmexcs = mb->exceptions;
    uint32_t length = mb->nexceptions;
    ClassClass *cb = mb->fb.clazz;
    jclass *excs;
    sys_thread_t *self = sysThreadSelf();
    ExecEnv *ee;
    JNIEnv *env;    

    DEBUG_ENABLED();
    NOT_NULL2(exceptionCountPtr, exceptionsPtr);
    THREAD_OK(self);

    ee = SysThread2EE(self);
    env = EE2JNIEnv(ee);


    ALLOC(size_to_jlong(length * sizeof(jclass)), exceptionsPtr);
    *exceptionCountPtr = (jint)length;
    excs = *exceptionsPtr;
    for (i = 0; i < length; ++i) {
        jclass exc;
        jvmdiError err = resolveConstantPoolIndex(ee, env, cb, vmexcs[i], 
						  &exc);
        if (err != JVMDI_ERROR_NONE) {
            return err;
        }
        excs[i] = exc;
    }
    return JVMDI_ERROR_NONE;
}

/* Return via "bytecodesPtr" the bytecodes implementing this method
 * ("bytecodeCountPtr" returns the number of byte codes).
 * Memory for this array is allocated by the function specified 
 * in JVMDI_SetAllocationHooks().
 * Errors: JVMDI_ERROR_NULL_POINTER, JVMDI_ERROR_INVALID_CLASS,
 * JVMDI_ERROR_INVALID_METHODID.
 */
static jvmdiError JNICALL
jvmdi_GetBytecodes(jclass clazz, jmethodID method,
		   jint *bytecodeCountPtr, jbyte **bytecodesPtr)
{
    struct methodblock *mb = (struct methodblock *)method;
    unsigned int length = mb->code_length;
    unsigned char *ptr;
    unsigned char *bufferStart;
    int i;
    sys_thread_t *self = sysThreadSelf();
    ExecEnv *ee;

    DEBUG_ENABLED();
    THREAD_OK(self);

    ee = SysThread2EE(self);

    ALLOC(uint_to_jlong(length), bytecodesPtr);
    *bytecodeCountPtr = (jint)length;
    for (i = 0; i < length; ++i) {
        (*bytecodesPtr)[i] = (mb->code)[i];
    }

    /* Re-insert real opcodes for each breakpoint currently set */
    bufferStart = (unsigned char *)*bytecodesPtr;
    ptr = bufferStart;
    while (ptr < bufferStart + length) {
        short opcodeLength;

        if (*ptr == opc_breakpoint) {
            *ptr = get_breakpoint_opcode(ee, mb->code + (ptr - bufferStart), FALSE);
            /* error will be caught below */
        }
        opcodeLength = opcode_length[*ptr];
        if (opcodeLength == -1) {
            jvmdi_Deallocate(*bytecodesPtr);
            return JVMDI_ERROR_INTERNAL;
        }
        ptr += opcodeLength;
    }
    return JVMDI_ERROR_NONE;
}

/* Return via "isNativePtr" a boolean indicating whether the method
 * is native.
 * Errors: JVMDI_ERROR_INVALID_METHODID, JVMDI_ERROR_INVALID_CLASS,
 * JVMDI_ERROR_NULL_POINTER
 */
static jvmdiError JNICALL
jvmdi_IsMethodNative(jclass clazz, jmethodID method, 
		    jboolean *isNativePtr)
{
    struct methodblock *mb = (struct methodblock *)method;
    
    DEBUG_ENABLED();
    NOT_NULL(isNativePtr);
    *isNativePtr = (mb->fb.access & ACC_NATIVE) != 0;
    return JVMDI_ERROR_NONE;
}

/* Return via "isSyntheticPtr" a boolean indicating whether the method
 * is synthetic.
 * Errors: JVMDI_ERROR_INVALID_METHODID, JVMDI_ERROR_INVALID_CLASS,
 * JVMDI_ERROR_NULL_POINTER
 */
static jvmdiError JNICALL
jvmdi_IsMethodSynthetic(jclass clazz, jmethodID method, 
		        jboolean *isSyntheticPtr)
{
    /* Synthetic attribute info is not yet available in this VM */
    return JVMDI_ERROR_NOT_IMPLEMENTED;
}

static void debugThreadWrapper(void *arg) {
    ThreadNode *node = findThread((JHandle *)arg);
    if (node == NULL) {
        return;
    }

    (node->startFunction)(node->startFunctionArg);
}

/* Start the execution of a debugger thread. The given thread must be
 * newly created and never previously run. Thread execution begins 
 * in the function "startFunc" which is passed the single argument, "arg".
 * The thread is set to be a daemon.
 * Errors: JVMDI_ERROR_NULL_POINTER, JVMDI_ERROR_INVALID_PRIORITY,
 * JVMDI_ERROR_OUT_OF_MEMORY, JVMDI_ERROR_INVALID_THREAD.
 */
static jvmdiError JNICALL
jvmdi_RunDebugThread(jthread thread, 
                     JVMDI_StartFunction proc, void *arg, int priority)
{
    TID tid;
    ThreadNode *node;

    DEBUG_ENABLED();
    NOT_NULL(proc);

    tid = (TID)DeRef(EE2JNIEnv(EE()), thread);

    /*
     * Make sure a thread hasn't been run for this thread object 
     * before.
     */
    if ((tid == 0) || (ll_nez(THREAD(tid)->eetop))) {
        return JVMDI_ERROR_INVALID_THREAD;
    }

    if ((priority < JVMDI_THREAD_MIN_PRIORITY) || (priority > JVMDI_THREAD_MAX_PRIORITY)) {
        return JVMDI_ERROR_INVALID_PRIORITY;
    }

    node = insertThread((JHandle *)tid);
    if (node == NULL) {
        return JVMDI_ERROR_OUT_OF_MEMORY;
    }

    node->startFunction = proc;
    node->startFunctionArg = arg;
    THREAD(tid)->daemon = 1;
    if (threadCreate(tid, 0, priority, debugThreadWrapper) != 0) {
        return JVMDI_ERROR_OUT_OF_MEMORY;
    }

    return JVMDI_ERROR_NONE;
}

  /*
   *  Debugger monitor support allows easy coordination among debugger 
   *  threads.
   */

/* Return via "monitorPtr" a newly created debugger monitor that can be 
 * used by debugger threads to coordinate. The semantics of the
 * monitor are the same as those described for Java language 
 * monitors in the Java Language Specification.
 * Errors: JVMDI_ERROR_NULL_POINTER, JVMDI_ERROR_OUT_OF_MEMORY
 */
static jvmdiError JNICALL
jvmdi_CreateRawMonitor(char *name, JVMDI_RawMonitor *monitorPtr)
{
    jthrowable currentException;
    sys_mon_t *sysMon;
    jvmdiError rc;
    sys_thread_t *self = sysThreadSelf();
    ExecEnv *ee;
    JNIEnv *env;    

    DEBUG_ENABLED();
    THREAD_OK(self);
    NOT_NULL2(name, monitorPtr);

    ee = SysThread2EE(self);
    env = EE2JNIEnv(ee);

    /*
     * Monitor registration can result in an out of memory exception, so
     * preserve and restore the current one here. 
     * Can't push a local frame here because it may itself throw an 
     * exception.
     */
    currentException = (*env)->ExceptionOccurred(env);

    ALLOC(size_to_jlong(sysMonitorSizeof()), &sysMon);
    monitorRegister(sysMon, name);
    if ((*env)->ExceptionOccurred(env)) {
        jvmdi_Deallocate((jbyte *)sysMon);
        rc = JVMDI_ERROR_OUT_OF_MEMORY;
    } else {
        *monitorPtr = sysMon;
        rc = JVMDI_ERROR_NONE;
    }

    /*
     * Restore exception state from before call
     */
    if (currentException != NULL) {
        (*env)->Throw(env, currentException);
    } else {
        (*env)->ExceptionClear(env);
    }

    return rc;
}

/* Destroy a debugger monitor created with JVMDI_CreateRawMonitor.
 * Errors: JVMDI_ERROR_NULL_POINTER
 */
static jvmdiError JNICALL
jvmdi_DestroyRawMonitor(JVMDI_RawMonitor monitor)
{
    DEBUG_ENABLED();
    NOT_NULL(monitor);

    sysMonitorDestroy((sys_mon_t *)monitor);
    return jvmdi_Deallocate((jbyte *)monitor);
}

/* Gain exclusive ownership of a debugger monitor.
 * Errors: JVMDI_ERROR_INVALID_MONITOR, JVMDI_ERROR_OUT_OF_MEMORY
 */
static jvmdiError JNICALL
jvmdi_RawMonitorEnter(JVMDI_RawMonitor monitor)
{
    int error;
    sys_thread_t *self = sysThreadSelf();

    DEBUG_ENABLED();
    THREAD_OK(self);
    NOT_NULL(monitor);

    error = sysMonitorEnter(self, (sys_mon_t *)monitor);
    return (error == SYS_OK) ? JVMDI_ERROR_NONE : JVMDI_ERROR_OUT_OF_MEMORY;
}

/* Release exclusive ownership of a debugger monitor.
 * Errors: JVMDI_ERROR_INVALID_MONITOR, JVMDI_ERROR_OUT_OF_MEMORY
 */
static jvmdiError JNICALL
jvmdi_RawMonitorExit(JVMDI_RawMonitor monitor)
{
    int error;
    sys_thread_t *self = sysThreadSelf();

    DEBUG_ENABLED();
    THREAD_OK(self);
    NOT_NULL(monitor);

    error = sysMonitorExit(self, (sys_mon_t *)monitor);
    return (error == SYS_OK) ? JVMDI_ERROR_NONE : JVMDI_ERROR_OUT_OF_MEMORY;
}

/* Wait for notification of the debugger monitor. The calling thread
 * must own the monitor. "millis" specifies the maximum time to wait, 
 * in milliseconds.
 * Errors: JVMDI_ERROR_INVALID_MONITOR, JVMDI_ERROR_NOT_MONITOR_OWNER,
 * JVMDI_ERROR_INTERRUPT
 */
static jvmdiError JNICALL
jvmdi_RawMonitorWait(JVMDI_RawMonitor monitor, jlong millis)
{
    int error;
    sys_thread_t *self = sysThreadSelf();

    DEBUG_ENABLED();
    THREAD_OK(self);
    NOT_NULL(monitor);

    error = sysMonitorWait(self, (sys_mon_t *)monitor, millis);
    switch (error) {
        case SYS_OK:
            return JVMDI_ERROR_NONE;
        case SYS_INTRPT:
            return JVMDI_ERROR_INTERRUPT;
        case SYS_ERR:
            return JVMDI_ERROR_NOT_MONITOR_OWNER;
        default:
            return JVMDI_ERROR_INTERNAL;
    }
}

/* Notify a single thread waiting on the debugger monitor. The calling 
 * thread must own the monitor.
 * Errors: JVMDI_ERROR_INVALID_MONITOR, JVMDI_ERROR_NOT_MONITOR_OWNER
 */
static jvmdiError JNICALL
jvmdi_RawMonitorNotify(JVMDI_RawMonitor monitor) {
    int error;
    sys_thread_t *self = sysThreadSelf();

    DEBUG_ENABLED();
    THREAD_OK(self);
    NOT_NULL(monitor);

    error = sysMonitorNotify(self, (sys_mon_t *)monitor);
    return (error == SYS_OK) ? JVMDI_ERROR_NONE : JVMDI_ERROR_NOT_MONITOR_OWNER;
}

/* Notify all threads waiting on the debugger monitor. The calling 
 * thread must own the monitor.
 * Errors: JVMDI_ERROR_INVALID_MONITOR, JVMDI_ERROR_NOT_MONITOR_OWNER
 */
static jvmdiError JNICALL
jvmdi_RawMonitorNotifyAll(JVMDI_RawMonitor monitor) {
    int error;
    sys_thread_t *self = sysThreadSelf();

    DEBUG_ENABLED();
    THREAD_OK(self);
    NOT_NULL(monitor);

    error = sysMonitorNotifyAll(self, (sys_mon_t *)monitor);
    return (error == SYS_OK) ? JVMDI_ERROR_NONE : JVMDI_ERROR_NOT_MONITOR_OWNER;
}

  
  /*
   *  Misc
   */

/* Return via "classesPtr" all classes currently loaded in the VM
 * ("classCountPtr" returns the number of such classes).
 * Memory for this table is allocated by the function specified 
 * in JVMDI_SetAllocationHooks().
 * Note: classes in the array are JNI global references and must
 * be explicitly managed.
 * Errors: JVMDI_ERROR_NULL_POINTER, JVMDI_ERROR_OUT_OF_MEMORY
 */
static jvmdiError JNICALL
jvmdi_GetLoadedClasses(jint *classCountPtr, jclass **classesPtr)
{
    jsize nClasses;
    jsize nLoadedClasses;
    ClassClass **pcb;
    jclass *jp;
    jsize inx;
    jvmdiError rc;
    sys_thread_t *self = sysThreadSelf();
    ExecEnv *ee;
    JNIEnv *env;    

    DEBUG_ENABLED();
    NOT_NULL2(classCountPtr, classesPtr);
    THREAD_OK(self);

    ee = SysThread2EE(self);
    env = EE2JNIEnv(ee);

    BINCLASS_LOCK(self);

    nClasses = get_nbinclasses();
    pcb = get_binclasses();

    nLoadedClasses = 0;
    for (inx = 0; inx < nClasses; inx++, pcb++) {
        ClassClass *cb = *pcb;
        if (!CCIs(cb, Error) && CCIs(cb, Loaded)) {
            nLoadedClasses++;
        }
    }
    
    rc = jvmdi_Allocate(size_to_jlong(nLoadedClasses * sizeof(jclass)),
			(jbyte**)classesPtr);
    if (rc == JVMDI_ERROR_NONE) {
        *classCountPtr = nLoadedClasses;
        jp = *classesPtr;
        pcb = get_binclasses();
        for (inx = 0; inx < nClasses; inx++, pcb++) {
            ClassClass *cb = *pcb;
            if (!CCIs(cb, Error) && 
                      (CCIs(cb, Loaded) || cbName(cb)[0] == SIGNATURE_ARRAY)) {
                jobject clazz;
                if ((*env)->PushLocalFrame(env, 1) < 0) {
                    rc = JVMDI_ERROR_OUT_OF_MEMORY;
                    break;
                }
                clazz = MkRefLocal(env, cb);
                *jp = (jclass)((*env)->NewGlobalRef(env, clazz));
                (*env)->PopLocalFrame(env, 0);
                if (*jp == NULL) {
                    rc = JVMDI_ERROR_OUT_OF_MEMORY;
                    break;
                }
                jp++;
            }
        }

        /* clean up partial array after any error */
        if (rc != JVMDI_ERROR_NONE) {
            for (jp--; jp >= *classesPtr; jp--) {
                (*env)->DeleteGlobalRef(env, *jp);
            }
            jvmdi_Deallocate((jbyte *)*classesPtr);
        }
            
    }

    BINCLASS_UNLOCK(self);
    return rc;
}

typedef struct {
    JNIEnv *env;
    struct Hjava_lang_ClassLoader *loader;
    jint count;
    jclass *classIter;
} EnumerateArg;

/* Used for counting the number of classes loaded through an initiating loader
 */
static int
loaderClassesCountHelper(ClassClass *cb, 
                         struct Hjava_lang_ClassLoader *loader, 
                         void *arg)
{
    EnumerateArg *enumArg = (EnumerateArg *)arg;

    if (loader == enumArg->loader) {
        enumArg->count++;
    }

    return SYS_OK;
}

/* Used for filling in classes loaded through an initiating loader
 */
static int
loaderClassesHelper(ClassClass *cb, 
                    struct Hjava_lang_ClassLoader *loader, 
                    void *arg)
{
    EnumerateArg *enumArg = (EnumerateArg *)arg;

    if (loader == enumArg->loader) {
        jclass clazz;
        JNIEnv *env = enumArg->env;
        if ((*env)->PushLocalFrame(env, 1) < 0) {
            return SYS_NOMEM;
        }

        clazz = MkRefLocal(env, cb);
        *enumArg->classIter = (jclass)((*env)->NewGlobalRef(env, clazz));
        (*env)->PopLocalFrame(env, 0);

        if (*enumArg->classIter == 0) {
            return SYS_NOMEM;
        }
        enumArg->count++;
        enumArg->classIter++;
    }

    return SYS_OK;
}

/* Return via "classesPtr" all classes loaded through the 
 * given initiating class loader.
 * ("classesCountPtr" returns the number of such threads).
 * Memory for this table is allocated by the function specified 
 * in JVMDI_SetAllocationHooks().
 * Note: classes in the array are JNI global references and must
 * be explicitly managed.
 * Errors: JVMDI_ERROR_NULL_POINTER, JVMDI_ERROR_OUT_OF_MEMORY
 */
static jvmdiError JNICALL
jvmdi_GetClassLoaderClasses(jobject initiatingLoader,
                            jint *classesCountPtr, jclass **classesPtr)
{
    jvmdiError rc = JVMDI_ERROR_NONE;
    int error;
    sys_thread_t *self = sysThreadSelf();
    EnumerateArg arg;
    ExecEnv *ee;
    JNIEnv *env;    

    DEBUG_ENABLED();
    THREAD_OK(self);
    NOT_NULL2(classesCountPtr, classesPtr);

    ee = SysThread2EE(self);
    env = EE2JNIEnv(ee);

    BINCLASS_LOCK(self);

    arg.loader =  (struct Hjava_lang_ClassLoader *)
                          DeRef(env, initiatingLoader);
    arg.count = 0;
    arg.env = env;

    /* count the classes */
    EnumerateOverLoaderCache(loaderClassesCountHelper, (void *)&arg);
    rc = jvmdi_Allocate(size_to_jlong(arg.count * sizeof(jclass)),
                        (jbyte**)classesPtr);
    *classesCountPtr = arg.count;
    if (rc == JVMDI_ERROR_NONE) {
        arg.count = 0;
        arg.classIter = *classesPtr;
        /* fill in the classes */
        error = EnumerateOverLoaderCache(loaderClassesHelper, 
                                           (void *)&arg);
        if (error == SYS_NOMEM) {
            int i;
            for (i = 0; i < arg.count; i++) {
                (*env)->DeleteGlobalRef(env, (*classesPtr)[i]);
            }
            jvmdi_Deallocate((jbyte *)*classesPtr);
            rc = JVMDI_ERROR_OUT_OF_MEMORY;
        } 
    }
    BINCLASS_UNLOCK(self);

    return rc;
}

/* Get the operand stack
* Errors: JVMDI_ERROR_NULL_POINTER, JVMDI_THREAD_NOT_SUSPENDED,
* JVMDI_ERROR_INVALID_FRAMEID,
* JVMDI_ERROR_NOT_IMPLEMENTED (if not canGetOperandStack).
*/
static jvmdiError JNICALL
jvmdi_GetOperandStack(jframeID frame,  jint *operandStackSizePtr, 
                JVMDI_operand_stack_element **operandStackPtr)
{
    return JVMDI_ERROR_NOT_IMPLEMENTED;
}

/* Pop the topmost stack frame of the stack of the thread argument
* Errors: JVMDI_ERROR_INVALID_THREAD, 
* JVMDI_ERROR_THREAD_NOT_SUSPENDED, JVMDI_ERROR_NO_MORE_FRAMES
* JVMDI_ERROR_NOT_IMPLEMENTED (if not canPopFrame).
*/
static jvmdiError JNICALL
jvmdi_PopFrame(jthread thread)
{
    return JVMDI_ERROR_NOT_IMPLEMENTED;
}

/* Set the BCI.
* Errors: JVMDI_ERROR_INVALID_FRAMEID,
* JVMDI_ERROR_INVALID_LOCATION,JVMDI_THREAD_NOT_SUSPENDED,
* JVMDI_ERROR_NOT_IMPLEMENTED (if not canSetOperandStack).
*/
static jvmdiError JNICALL
jvmdi_SetFrameLocation(jframeID frame, jlocation location)
{
    return JVMDI_ERROR_NOT_IMPLEMENTED;
}

/* Set the operand stack
* Note that here, the VM can interpret the operand stack types itself
* Errors: JVMDI_ERROR_INVALID_FRAMEID,
* JVMDI_THREAD_NOT_SUSPENDED,
* JVMDI_ERROR_ILLEGAL_ARGUMENT (if operandStackSize negative or
* operand types are out of range),
* JVMDI_ERROR_NOT_IMPLEMENTED (if not canSetOperandStack).
*
*/
static jvmdiError JNICALL
jvmdi_SetOperandStack(jframeID frame, jint operandStackSize, 
                      JVMDI_operand_stack_element *operandStack)
{
    return JVMDI_ERROR_NOT_IMPLEMENTED;
}

/* Return all instances of class clazz. 
* Errors: JVMDI_ERROR_INVALID_CLASS, 
* JVMDI_ERROR_NOT_IMPLEMENTED (if not canGetHeapInfo). 
*/  
static jvmdiError JNICALL
jvmdi_AllInstances(jclass clazz, jint *instanceCountPtr, 
                   jobject **instancesPtr)
{
    return JVMDI_ERROR_NOT_IMPLEMENTED;
}

/* All references to obj 
* Errors: JVMDI_ERROR_INVALID_OBJECT, 
* JVMDI_ERROR_NOT_IMPLEMENTED (if not canGetHeapInfo).
*/
static jvmdiError JNICALL
jvmdi_References(jobject obj, JVMDI_object_reference_info *refs)
{
    return JVMDI_ERROR_NOT_IMPLEMENTED;
}

/* Return via "classDefPtr" the definition of this class.
* Memory for this array is allocated by the function specified 
* in JVMDI_SetAllocationHooks().
* Errors: JVMDI_ERROR_INVALID_CLASS,
* JVMDI_ERROR_NOT_IMPLEMENTED (if not canGetClassDefinition)
*/
static jvmdiError JNICALL
jvmdi_GetClassDefinition(jclass clazz, 
                         JVMDI_class_definition *classDefPtr)
{
    return JVMDI_ERROR_NOT_IMPLEMENTED;
}

/* Redefine the classes defined by "classDefs". "classCount" 
* specifies number of classes to redefine.
* Errors: JVMDI_ERROR_INVALID_CLASS, JVMDI_ERROR_FAILS_VERIFICATION,
* JVMDI_ERROR_INVALID_CLASS_FORMAT, JVMDI_ERROR_CIRCULAR_CLASS_DEFINITION,
* JVMDI_ERROR_SCHEMA_CHANGE_NOT_IMPLEMENTED (if not canChangeSchema)
* JVMDI_ERROR_ADD_METHOD_NOT_IMPLEMENTED (if not canAddMethod)
* JVMDI_ERROR_NOT_IMPLEMENTED (if not canRedefineClasses)
*/
static jvmdiError JNICALL
jvmdi_RedefineClasses(jint classCount, 
                      JVMDI_class_definition *classDefs)
{
    return JVMDI_ERROR_NOT_IMPLEMENTED;
}

/* Return via "versionPtr" the JVMDI version number.
 * Errors: JVMDI_ERROR_NULL_POINTER
 */
static jvmdiError JNICALL
jvmdi_GetVersionNumber(jint *versionPtr)
{
    /* 
     * We allow version queries even when debugging is not enabled 
     */
    NOT_NULL(versionPtr);
    *versionPtr = JVMDI_VERSION_1;
    return JVMDI_ERROR_NONE;
}

/* Return via "capPtr" the capabilities of this JVMDI implementation.
 * Errors: JVMDI_ERROR_NULL_POINTER,
 */
static jvmdiError JNICALL
jvmdi_GetCapabilities(JVMDI_capabilities *capPtr) 
{
    DEBUG_ENABLED();
    NOT_NULL(capPtr);

    capPtr->can_watch_field_modification = JNI_TRUE;
    capPtr->can_watch_field_access = JNI_TRUE;
    capPtr->can_get_bytecodes = JNI_TRUE;
    capPtr->can_get_synthetic_attribute = JNI_FALSE;
    capPtr->can_get_current_contended_monitor = JNI_TRUE;
#ifdef FAST_MONITOR
    capPtr->can_get_owned_monitor_info = JNI_FALSE;
    capPtr->can_get_monitor_info = JNI_FALSE;
#else
    capPtr->can_get_owned_monitor_info = JNI_TRUE;
    capPtr->can_get_monitor_info = JNI_TRUE;
#endif /* FAST_MONITOR */
    capPtr->can_get_heap_info = JNI_FALSE;
    capPtr->can_get_operand_stack = JNI_FALSE;
    capPtr->can_set_operand_stack = JNI_FALSE;
    capPtr->can_pop_frame = JNI_FALSE;
    capPtr->can_get_class_definition = JNI_FALSE;
    capPtr->can_redefine_classes = JNI_FALSE; 
    capPtr->can_add_method = JNI_FALSE;
    capPtr->can_change_schema = JNI_FALSE;

    return JVMDI_ERROR_NONE;
}
    
/* Retrieve the JVMDI interface pointer.  This function private to VM.
 */
JVMDI_Interface_1 *
jvmdi_GetInterface_1(JavaVM *interfaces_vm)
{
  if (interface.SetEventHook == NULL) {
    vm = interfaces_vm;
    interface.SetEventHook = &jvmdi_SetEventHook;
    interface.SetEventNotificationMode = &jvmdi_SetEventNotificationMode;
    interface.GetThreadStatus = &jvmdi_GetThreadStatus;
    interface.GetAllThreads = &jvmdi_GetAllThreads;
    interface.SuspendThread = &jvmdi_SuspendThread;
    interface.ResumeThread = &jvmdi_ResumeThread;
    interface.StopThread = &jvmdi_StopThread;
    interface.InterruptThread = &jvmdi_InterruptThread;
    interface.GetThreadInfo = &jvmdi_GetThreadInfo;
    interface.GetOwnedMonitorInfo = &jvmdi_GetOwnedMonitorInfo;
    interface.GetCurrentContendedMonitor = &jvmdi_GetCurrentContendedMonitor;
    interface.RunDebugThread = &jvmdi_RunDebugThread;
    interface.CreateRawMonitor = &jvmdi_CreateRawMonitor;
    interface.DestroyRawMonitor = &jvmdi_DestroyRawMonitor;
    interface.RawMonitorEnter = &jvmdi_RawMonitorEnter;
    interface.RawMonitorExit = &jvmdi_RawMonitorExit;
    interface.RawMonitorWait = &jvmdi_RawMonitorWait;
    interface.RawMonitorNotify = &jvmdi_RawMonitorNotify;
    interface.RawMonitorNotifyAll = & jvmdi_RawMonitorNotifyAll;
    interface.GetTopThreadGroups = &jvmdi_GetTopThreadGroups;
    interface.GetThreadGroupInfo = &jvmdi_GetThreadGroupInfo;
    interface.GetThreadGroupChildren = &jvmdi_GetThreadGroupChildren;
    interface.GetFrameCount = &jvmdi_GetFrameCount;
    interface.GetCurrentFrame = &jvmdi_GetCurrentFrame;
    interface.GetCallerFrame = &jvmdi_GetCallerFrame;
    interface.GetFrameLocation = &jvmdi_GetFrameLocation;
    interface.NotifyFramePop = &jvmdi_NotifyFramePop;
    interface.GetLocalObject = &jvmdi_GetLocalObject;
    interface.GetLocalInt = &jvmdi_GetLocalInt;
    interface.GetLocalLong = &jvmdi_GetLocalLong;
    interface.GetLocalFloat = &jvmdi_GetLocalFloat;
    interface.GetLocalDouble = &jvmdi_GetLocalDouble;
    interface.SetLocalObject = &jvmdi_SetLocalObject;
    interface.SetLocalInt = &jvmdi_SetLocalInt;
    interface.SetLocalLong = &jvmdi_SetLocalLong;
    interface.SetLocalFloat = &jvmdi_SetLocalFloat;
    interface.SetLocalDouble = &jvmdi_SetLocalDouble;
    interface.SetBreakpoint = &jvmdi_SetBreakpoint;
    interface.ClearBreakpoint = &jvmdi_ClearBreakpoint;
    interface.ClearAllBreakpoints = &jvmdi_ClearAllBreakpoints;
    interface.SetFieldAccessWatch = &jvmdi_SetFieldAccessWatch;
    interface.ClearFieldAccessWatch = &jvmdi_ClearFieldAccessWatch;
    interface.SetFieldModificationWatch = &jvmdi_SetFieldModificationWatch;
    interface.ClearFieldModificationWatch = &jvmdi_ClearFieldModificationWatch;
    interface.SetAllocationHooks = &jvmdi_SetAllocationHooks;
    interface.Allocate = &jvmdi_Allocate;
    interface.Deallocate = &jvmdi_Deallocate;
    interface.GetClassSignature = &jvmdi_GetClassSignature;
    interface.GetClassStatus = &jvmdi_GetClassStatus;
    interface.GetSourceFileName = &jvmdi_GetSourceFileName;
    interface.GetClassModifiers = &jvmdi_GetClassModifiers;
    interface.GetClassMethods = &jvmdi_GetClassMethods;
    interface.GetClassFields = &jvmdi_GetClassFields;
    interface.GetImplementedInterfaces = &jvmdi_GetImplementedInterfaces;
    interface.IsInterface = &jvmdi_IsInterface;
    interface.IsArrayClass = &jvmdi_IsArrayClass;
    interface.GetClassLoader = &jvmdi_GetClassLoader;
    interface.GetObjectHashCode = &jvmdi_GetObjectHashCode;
    interface.GetMonitorInfo = &jvmdi_GetMonitorInfo;
    interface.GetFieldName = &jvmdi_GetFieldName;
    interface.GetFieldDeclaringClass = &jvmdi_GetFieldDeclaringClass;
    interface.GetFieldModifiers = &jvmdi_GetFieldModifiers;
    interface.IsFieldSynthetic = &jvmdi_IsFieldSynthetic;
    interface.GetMethodName = &jvmdi_GetMethodName;
    interface.GetMethodDeclaringClass = &jvmdi_GetMethodDeclaringClass;
    interface.GetMethodModifiers = &jvmdi_GetMethodModifiers;
    interface.GetMaxStack = &jvmdi_GetMaxStack;
    interface.GetMaxLocals = &jvmdi_GetMaxLocals;
    interface.GetArgumentsSize = &jvmdi_GetArgumentsSize;
    interface.GetLineNumberTable = &jvmdi_GetLineNumberTable;
    interface.GetMethodLocation = &jvmdi_GetMethodLocation;
    interface.GetLocalVariableTable = &jvmdi_GetLocalVariableTable;
    interface.GetExceptionHandlerTable = &jvmdi_GetExceptionHandlerTable;
    interface.GetThrownExceptions = &jvmdi_GetThrownExceptions;
    interface.GetBytecodes = &jvmdi_GetBytecodes;
    interface.IsMethodNative = &jvmdi_IsMethodNative;
    interface.IsMethodSynthetic = &jvmdi_IsMethodSynthetic;
    interface.GetLoadedClasses = &jvmdi_GetLoadedClasses;
    interface.GetClassLoaderClasses = &jvmdi_GetClassLoaderClasses;
    interface.PopFrame = &jvmdi_PopFrame;
    interface.SetFrameLocation = &jvmdi_SetFrameLocation;
    interface.GetOperandStack = &jvmdi_GetOperandStack;
    interface.SetOperandStack = &jvmdi_SetOperandStack;
    interface.AllInstances = &jvmdi_AllInstances;
    interface.References = &jvmdi_References;
    interface.GetClassDefinition = &jvmdi_GetClassDefinition;
    interface.RedefineClasses = &jvmdi_RedefineClasses;
    interface.GetVersionNumber = &jvmdi_GetVersionNumber;
    interface.GetCapabilities = &jvmdi_GetCapabilities;
  }
    return &interface;
}

#endif /* BREAKPTS */
