/*
 * @(#)jvmpi.c	1.21 98/09/02
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "oobj.h"
#include "interpreter.h"
#include "sys_api.h"
#include "alloc_cache.h"
#include "threads.h"
#include "byteorder.h"
#include "utf.h"
#include "monitor.h"
#include "monitor_cache.h"
#include "reflect.h"
#include "jlong.h"

unsigned int jvmpi_event_flags;                  /* whether hprof is enabled */

#ifdef HPROF

#include "jvmpi.h"
#include "vmprofiler.h"

/***************************************************************************
 * 
 * PART 1: Data structures used by the VM
 *
 **************************************************************************/

#define JVMPI_NAME_BUF_LEN 256

static int allocLocalSizeSaved = 0;

static void jvmpi_alloc_object2(ExecEnv *ee, JHandle *h, jint flag);
static void jvmpi_thread_start2(JHandle *tid, jint flag);
static void jvmpi_load_class2(ClassClass *cb, int flag);

static void jvmpi_cleanup(void);
static jint jvmpi_heap_dump(int heap_dump_level);
static jint jvmpi_monitor_dump(void);
static void jvmpi_GetCallTrace(JVMPI_CallTrace *trace, jint depth);
static jlong jvmpi_GetCurrentThreadCpuTime(void);
static JVMPI_RawMonitor jvmpi_RawMonitorCreate(char *lock_name);
static void jvmpi_RawMonitorEnter(JVMPI_RawMonitor lock_id);
static void jvmpi_RawMonitorExit(JVMPI_RawMonitor lock_id);
static void jvmpi_RawMonitorDestroy(JVMPI_RawMonitor lock_id);

static void jvmpi_RawMonitorWait(JVMPI_RawMonitor lock_id, jlong ms);
static void jvmpi_RawMonitorNotifyAll(JVMPI_RawMonitor lock_id);
static void jvmpi_SuspendThread(JNIEnv *env);
static void jvmpi_ResumeThread(JNIEnv *env);
static jint jvmpi_GetThreadStatus(JNIEnv *env);
static jboolean jvmpi_ThreadHasRun(JNIEnv *env);

static void jvmpi_RunGC(void);
static void jvmpi_ProfilerExit(jint exit_code);
static jint jvmpi_CreateSystemThread(char *name, jint priority, 
				       void (*f)(void *));

static jint jvmpi_EnableEvent(jint event_type, void *arg);
static jint jvmpi_DisableEvent(jint event_type, void *arg);
static jint jvmpi_RequestEvent(jint event_type, void *arg);

static void jvmpi_SetThreadLocalStorage(JNIEnv *env, void *ptr);
static void * jvmpi_GetThreadLocalStorage(JNIEnv *env);
static void jvmpi_DisableGC(void);
static void jvmpi_EnableGC(void);
static jobjectID jvmpi_GetThreadObject(JNIEnv *env);
static jobjectID jvmpi_GetMethodClass(jmethodID mid);

/* the interface through which we are talking to the profiler */
static JVMPI_Interface interface;
#define INTERFACE_CALL(f) (interface.f)

jvmpi_event_info_t jvmpi_event_info[JVMPI_MAX_EVENT_TYPE_VAL + 1] = {0};

#define JVMPI_UNHANDLE(o) ((o == NULL) ? (NULL) : (unhand(o)))

/* 
 * Data struture used by the heap dump routines to temporarily hold
 * trace information. For simplicity, we place an upper bound on the 
 * number of threads traced by GC.
 */
#define JVMPI_DUMP_MAX_TRACES 1024

/* The context information maintained during the process of collecting
 * a heap dump.
 */
typedef struct {
    char *ptr;              /* beginning of the allocated buffer */
    char *end_ptr;          /* end of the allocated buffer */

    int n_traces;           /* number of threads traced by GC */
    JVMPI_CallTrace traces[JVMPI_DUMP_MAX_TRACES];
    char *begin_dump;      /* where all the root records begin. */
    int dump_level;         /* level of information in the dump */
} jvmpi_dump_context_t;

static jvmpi_dump_context_t *jvmpi_dump_context;

static sys_mon_t *_jvmpi_dump_context_lock;

#define JVMPI_DUMP_CONTEXT_LOCK(ee) \
    sysMonitorEnter(EE2SysThread(ee), _jvmpi_dump_context_lock)
#define JVMPI_DUMP_CONTEXT_UNLOCK(ee) \
    sysMonitorExit(EE2SysThread(ee), _jvmpi_dump_context_lock)

/***************************************************************************
 * 
 * PART 2: Helper functions
 *
 **************************************************************************/

void jvmpi_dump_context_lock(void)
{
    JVMPI_DUMP_CONTEXT_LOCK(EE());
}

void jvmpi_dump_context_unlock(void)
{
    JVMPI_DUMP_CONTEXT_UNLOCK(EE());
}

static void *
jvmpi_calloc(unsigned int size)
{
    void *p = sysMalloc(size);
    if (p == NULL) {
        out_of_memory();
    }
    memset(p, 0, size);
    return p;
}

#define jvmpi_free sysFree

static int 
jvmpi_get_lineno(struct methodblock *mb, unsigned char *pc)
{
    if (mb->fb.access & ACC_MACHINE_COMPILED) {
        return -2;
    } else if (mb->fb.access & ACC_NATIVE) {
        return -3;
    } else {
        return pc ? pc2lineno(mb, (unsigned short)(pc - mb->code)) : -1;
    }
}

/* handler for enabling obj alloc notifications */
static jint
jvmpi_obj_alloc_on(void *_arg)
{
    /* disable thread-local allocation cache */
    allocLocalSizeSaved = allocLocalSize;
    allocLocalSize = 0;
    return JVMPI_SUCCESS;
}

/* handler for disabling obj alloc notifications */
static jint
jvmpi_obj_alloc_off(void *_arg)
{
    /* enable thread-local allocation cache */
    if (allocLocalSizeSaved > 0) {
        allocLocalSize = allocLocalSizeSaved;
    }
    return JVMPI_SUCCESS;
}

static void jvmpi_disable_all_notification(void)
{
    int i;
    for (i = 0; i <= JVMPI_MAX_EVENT_TYPE_VAL; i++) {
        if (jvmpi_event_info[i].flag != -1) {
	    jvmpi_event_info[i].flag = JVMPI_EVENT_DISABLED;
	}
    }
    jvmpi_event_flags = JVMPI_EVENT_PROFILING_OFF;
    jvmpi_obj_alloc_off(NULL);
}

/* setup */
static void jvmpi_setup(void)
{
    int i;
    
    /* fill in event info for enabling/disabling event notification */
    for (i = 0; i <= JVMPI_MAX_EVENT_TYPE_VAL; i++) {
        jvmpi_event_info[i].flag = -1;
	jvmpi_event_info[i].enable_handler = NULL;
	jvmpi_event_info[i].disable_handler = NULL;
    }
    
    jvmpi_event_info[JVMPI_EVENT_CLASS_LOAD].flag =
        JVMPI_EVENT_DISABLED;
    jvmpi_event_info[JVMPI_EVENT_CLASS_UNLOAD].flag =
        JVMPI_EVENT_DISABLED;
    jvmpi_event_info[JVMPI_EVENT_CLASS_LOAD_HOOK].flag =
        JVMPI_EVENT_DISABLED;
    jvmpi_event_info[JVMPI_EVENT_OBJ_ALLOC].flag =
        JVMPI_EVENT_DISABLED;
    jvmpi_event_info[JVMPI_EVENT_OBJ_ALLOC].enable_handler =
        jvmpi_obj_alloc_on;
    jvmpi_event_info[JVMPI_EVENT_OBJ_ALLOC].disable_handler =
        jvmpi_obj_alloc_off;
    jvmpi_event_info[JVMPI_EVENT_OBJ_FREE].flag =
        JVMPI_EVENT_DISABLED;
    jvmpi_event_info[JVMPI_EVENT_THREAD_START].flag =
        JVMPI_EVENT_DISABLED;
    jvmpi_event_info[JVMPI_EVENT_THREAD_END].flag =
        JVMPI_EVENT_DISABLED;
    jvmpi_event_info[JVMPI_EVENT_JNI_GLOBALREF_ALLOC].flag =
        JVMPI_EVENT_DISABLED;
    jvmpi_event_info[JVMPI_EVENT_JNI_GLOBALREF_FREE].flag = 
        JVMPI_EVENT_DISABLED;
    jvmpi_event_info[JVMPI_EVENT_JNI_WEAK_GLOBALREF_ALLOC].flag =
        JVMPI_EVENT_DISABLED;
    jvmpi_event_info[JVMPI_EVENT_JNI_WEAK_GLOBALREF_FREE].flag = 
        JVMPI_EVENT_DISABLED;
    jvmpi_event_info[JVMPI_EVENT_METHOD_ENTRY].flag =
        JVMPI_EVENT_DISABLED;
    jvmpi_event_info[JVMPI_EVENT_METHOD_ENTRY2].flag =
        JVMPI_EVENT_DISABLED;
    jvmpi_event_info[JVMPI_EVENT_METHOD_EXIT].flag =
        JVMPI_EVENT_DISABLED;
    jvmpi_event_info[JVMPI_EVENT_LOAD_COMPILED_METHOD].flag =
        JVMPI_EVENT_DISABLED;
    jvmpi_event_info[JVMPI_EVENT_UNLOAD_COMPILED_METHOD].flag =
        JVMPI_EVENT_DISABLED;
    jvmpi_event_info[JVMPI_EVENT_JVM_INIT_DONE].flag =
        JVMPI_EVENT_DISABLED;
    jvmpi_event_info[JVMPI_EVENT_JVM_SHUT_DOWN].flag =
        JVMPI_EVENT_DISABLED;
    jvmpi_event_info[JVMPI_EVENT_DUMP_DATA_REQUEST].flag =
        JVMPI_EVENT_DISABLED;
    jvmpi_event_info[JVMPI_EVENT_RESET_DATA_REQUEST].flag =
        JVMPI_EVENT_DISABLED;
    jvmpi_event_info[JVMPI_EVENT_OBJ_MOVE].flag =
        JVMPI_EVENT_DISABLED;
    jvmpi_event_info[JVMPI_EVENT_NEW_ARENA].flag =
        JVMPI_EVENT_DISABLED;
    jvmpi_event_info[JVMPI_EVENT_DELETE_ARENA].flag =
        JVMPI_EVENT_DISABLED;

    jvmpi_event_info[JVMPI_EVENT_RAW_MONITOR_CONTENDED_ENTER].flag =
        JVMPI_EVENT_DISABLED;
    jvmpi_event_info[JVMPI_EVENT_RAW_MONITOR_CONTENDED_ENTERED].flag =
        JVMPI_EVENT_DISABLED;
    jvmpi_event_info[JVMPI_EVENT_RAW_MONITOR_CONTENDED_EXIT].flag =
        JVMPI_EVENT_DISABLED;
    jvmpi_event_info[JVMPI_EVENT_MONITOR_CONTENDED_ENTER].flag =
        JVMPI_EVENT_DISABLED;
    jvmpi_event_info[JVMPI_EVENT_MONITOR_CONTENDED_ENTERED].flag =
        JVMPI_EVENT_DISABLED;
    jvmpi_event_info[JVMPI_EVENT_MONITOR_CONTENDED_EXIT].flag =
        JVMPI_EVENT_DISABLED;
    jvmpi_event_info[JVMPI_EVENT_MONITOR_WAIT].flag =
        JVMPI_EVENT_DISABLED;
    jvmpi_event_info[JVMPI_EVENT_MONITOR_WAITED].flag =
        JVMPI_EVENT_DISABLED;

    jvmpi_event_info[JVMPI_EVENT_GC_START].flag =
        JVMPI_EVENT_DISABLED;
    jvmpi_event_info[JVMPI_EVENT_GC_FINISH].flag =
        JVMPI_EVENT_DISABLED;

    /* set up the interface */
    interface.version = JVMPI_VERSION_1;
    interface.EnableEvent = jvmpi_EnableEvent;
    interface.DisableEvent = jvmpi_DisableEvent;
    interface.RequestEvent = jvmpi_RequestEvent;
    interface.GetCallTrace = jvmpi_GetCallTrace;
    interface.GetCurrentThreadCpuTime = jvmpi_GetCurrentThreadCpuTime;
    interface.ProfilerExit = jvmpi_ProfilerExit;
    interface.RawMonitorCreate = jvmpi_RawMonitorCreate;
    interface.RawMonitorEnter = jvmpi_RawMonitorEnter;
    interface.RawMonitorExit = jvmpi_RawMonitorExit;
    interface.RawMonitorWait = jvmpi_RawMonitorWait;
    interface.RawMonitorNotifyAll = jvmpi_RawMonitorNotifyAll;
    interface.RawMonitorDestroy = jvmpi_RawMonitorDestroy;
    interface.SuspendThread = jvmpi_SuspendThread;
    interface.ResumeThread = jvmpi_ResumeThread;
    interface.GetThreadStatus = jvmpi_GetThreadStatus;
    interface.ThreadHasRun = jvmpi_ThreadHasRun;
    interface.CreateSystemThread = jvmpi_CreateSystemThread;
    interface.SetThreadLocalStorage = jvmpi_SetThreadLocalStorage;
    interface.GetThreadLocalStorage = jvmpi_GetThreadLocalStorage;
    interface.DisableGC = jvmpi_DisableGC;
    interface.EnableGC = jvmpi_EnableGC;
    interface.RunGC = jvmpi_RunGC;
    interface.GetThreadObject = jvmpi_GetThreadObject;
    interface.GetMethodClass = jvmpi_GetMethodClass;
}

void * jvmpi_interface(void)
{
    AtExit(jvmpi_cleanup);
    jvmpi_event_flags = JVMPI_EVENT_PROFILING_ON;
    sysSetMonitoringOn(TRUE);
    _jvmpi_dump_context_lock = sysMalloc(sysMonitorSizeof());
    sysMonitorInit(_jvmpi_dump_context_lock);

    /* Set up event info and the interface */
    jvmpi_setup();
    return &interface;
}

/***************************************************************************
 * 
 * PART 3: JVM interface functions
 *
 **************************************************************************/

static void jvmpi_GetCallTrace(JVMPI_CallTrace *trace, jint depth)
{
    JavaFrame *frame = (JNIEnv2EE((JNIEnv *)(trace->env_id)))->current_frame;
    JavaFrame frame_buf;
    int n_frames = 0;
    
    if (frame && IS_JIT_FRAME(frame))
	frame = CompiledFrameUpdate(frame);
    while ((frame) && (n_frames < depth)) {
        struct methodblock *mb = frame->current_method;
	if (mb) {
	    JVMPI_CallFrame *jframe = &(trace->frames[n_frames]);
	    jframe->lineno = jvmpi_get_lineno(mb, frame->lastpc);
	    jframe->method_id = (jmethodID)mb;
	    n_frames++;
	}
	frame = FRAME_PREV(frame, &frame_buf);
    }
    trace->num_frames = n_frames;
}

static JVMPI_RawMonitor jvmpi_RawMonitorCreate(char *lock_name) {
    sys_mon_t *lock;
    lock = jvmpi_calloc(sysMonitorSizeof());
    monitorRegister(lock, lock_name);
    return (JVMPI_RawMonitor)lock;
}

static void jvmpi_RawMonitorEnter(JVMPI_RawMonitor lock_id)
{
    if (lock_id != NULL) {
        ExecEnv *ee = EE();
	sysMonitorEnter(EE2SysThread(ee), (sys_mon_t *)lock_id);
    }
}

static void jvmpi_RawMonitorExit(JVMPI_RawMonitor lock_id)
{
    if (lock_id != NULL) {
        ExecEnv *ee = EE();
	sysMonitorExit(EE2SysThread(ee), (sys_mon_t *)lock_id);
    }
}

static void jvmpi_RawMonitorWait(JVMPI_RawMonitor lock_id, jlong ms)
{
    sysMonitorWait(sysThreadSelf(), (sys_mon_t *)lock_id, ms);
}

static void jvmpi_RawMonitorNotifyAll(JVMPI_RawMonitor lock_id)
{
    sysMonitorNotifyAll(sysThreadSelf(), (sys_mon_t *)lock_id);
}

static void jvmpi_RawMonitorDestroy(JVMPI_RawMonitor lock_id)
{
    if (lock_id != NULL) {
        sysMonitorDestroy((sys_mon_t *)lock_id);
	jvmpi_free(lock_id);
    }
}

static void jvmpi_SuspendThread(JNIEnv *env)
{
    sysThreadProfSuspend(EE2SysThread(JNIEnv2EE(env)));
}

static void jvmpi_ResumeThread(JNIEnv *env)
{
    sysThreadProfResume(EE2SysThread(JNIEnv2EE(env)));
}

static jint jvmpi_GetThreadStatus(JNIEnv *env)
{
    return sysThreadGetStatus(EE2SysThread(JNIEnv2EE(env)), NULL);
}

static jboolean jvmpi_ThreadHasRun(JNIEnv *env)
{
    return sysThreadIsRunning(EE2SysThread(JNIEnv2EE(env)));
}

static jint 
jvmpi_CreateSystemThread(char *name, jint priority, void (*f)(void *))
{
    if (createSystemThread(name, priority, 0, f, NULL) < 0) {
        return JNI_ERR;
    } else {
        return JNI_OK;
    }
}

static jlong jvmpi_GetCurrentThreadCpuTime(void)
{
    return sysThreadCPUTime();
}

/* enable event notification */
static jint 
jvmpi_EnableEvent(jint event_type, void *arg)
{
    jvmpi_event_info_t *info;
    jint ret = JVMPI_SUCCESS;
    
    /* event type is greater than the max value for event types */
    if (event_type > JVMPI_MAX_EVENT_TYPE_VAL) {
       return JVMPI_NOT_AVAILABLE;
    }
    /* get event info */
    info = &(jvmpi_event_info[event_type]);
    /* no information on this event type */
    if (info->flag == -1) {
        return JVMPI_NOT_AVAILABLE;
    }
    if (info->enable_handler != NULL) {
        ret = (info->enable_handler)(arg);
    }
    if (event_type < 32) {
        jvmpi_event_flags |= 1 << event_type;
    }
    jvmpi_event_info[event_type].flag = JVMPI_EVENT_ENABLED;
    return ret;
}

/* disable event notification */
static jint 
jvmpi_DisableEvent(jint event_type, void *arg)
{
    jvmpi_event_info_t *info;
    jint ret = JVMPI_SUCCESS;
    
    /* event type is greater than the max value for event types */
    if (event_type > JVMPI_MAX_EVENT_TYPE_VAL) {
       return JVMPI_NOT_AVAILABLE;
    }
    /* get event info */
    info = &(jvmpi_event_info[event_type]);
    /* no information on this event type */
    if (info->flag == -1) {
        return JVMPI_NOT_AVAILABLE;
    }
    if (info->disable_handler != NULL) {
        ret = (info->disable_handler)(arg);
    }
    if (event_type < 32) {
        jvmpi_event_flags &= ~(1 << event_type);
    }
    jvmpi_event_info[event_type].flag = JVMPI_EVENT_DISABLED;
    return ret;
}

static void jvmpi_RunGC(void)
{
    gc(~0);
}

static void jvmpi_ProfilerExit(jint exit_code) 
{
    jvmpi_disable_all_notification();
    Exit(exit_code);
}

static void jvmpi_SetThreadLocalStorage(JNIEnv *env, void *ptr)
{
    (JNIEnv2EE(env))->profileData = ptr;
}

static void * jvmpi_GetThreadLocalStorage(JNIEnv *env)
{
    return (JNIEnv2EE(env))->profileData;
}
    
static void jvmpi_DisableGC(void)
{
    DisableGC(EE());
}

static void jvmpi_EnableGC(void)
{
    EnableGC(EE());
}

static void jvmpi_dump_object_event(JHandle *h)
{
    JVMPI_Event event;
    jvmpi_dump_context_t context_buf;
    ExecEnv *ee = EE();
    int obj_size, buf_size;
    int is_array;
    char *buf;
    ClassClass *cb;

    if (h == 0) {
        return;
    }

    cb = jvmpi_get_object_info(h, &is_array, &obj_size);

    if (!is_array && cb == classJavaLangClass) {
        cb = (ClassClass *)h;
        buf_size = 128 + sizeof(void *) *
	    (cbConstantPoolCount(cb) + cbFieldsCount(cb) + cbImplementsCount(cb));
    } else {
        buf_size = 32 + obj_size * 2;
    }

    buf = sysMalloc(buf_size);
    if (buf == NULL) {
        return;
    }
    
    JVMPI_DUMP_CONTEXT_LOCK(ee);
    memset(&context_buf, 0, sizeof(context_buf));
    jvmpi_dump_context = &context_buf;
    jvmpi_dump_context->dump_level = JVMPI_DUMP_LEVEL_2;
    jvmpi_dump_context->ptr = buf;
    jvmpi_dump_context->end_ptr = buf + buf_size;
    jvmpi_dump_object(h);
    event.env_id = EE2JNIEnv(ee);
    event.event_type = JVMPI_EVENT_OBJECT_DUMP | JVMPI_REQUESTED_EVENT;
    event.u.object_dump.data_len = jvmpi_dump_context->ptr - buf;
    event.u.object_dump.data = buf;
    INTERFACE_CALL(NotifyEvent)(&event);
    JVMPI_DUMP_CONTEXT_UNLOCK(ee);

    sysFree(buf);
}

static jint jvmpi_RequestEvent(jint event_type, void *arg)
{
    switch (event_type) {
    case JVMPI_EVENT_OBJ_ALLOC:
        jvmpi_alloc_object2(EE(),
			    jvmpi_obj_reverse_map(arg),
			    JVMPI_REQUESTED_EVENT);
	return JVMPI_SUCCESS;
    case JVMPI_EVENT_THREAD_START:
        jvmpi_thread_start2(jvmpi_obj_reverse_map(arg),
			    JVMPI_REQUESTED_EVENT);
	return JVMPI_SUCCESS;
    case JVMPI_EVENT_CLASS_LOAD:
        jvmpi_load_class2((ClassClass *)jvmpi_obj_reverse_map(arg),
			  JVMPI_REQUESTED_EVENT);
	return JVMPI_SUCCESS;
    case JVMPI_EVENT_OBJECT_DUMP:
        jvmpi_dump_object_event(jvmpi_obj_reverse_map(arg));
	return JVMPI_SUCCESS;
    case JVMPI_EVENT_HEAP_DUMP: {
        int heap_dump_level;

	if (arg == NULL) {
	    heap_dump_level = JVMPI_DUMP_LEVEL_2;
	} else {
	    heap_dump_level = ((JVMPI_HeapDumpArg *)arg)->heap_dump_level;
	}

        jvmpi_heap_dump(heap_dump_level);
	return JVMPI_SUCCESS;
    }
    case JVMPI_EVENT_MONITOR_DUMP:
        jvmpi_monitor_dump();
	return JVMPI_SUCCESS;
    default:
        return JVMPI_NOT_AVAILABLE;
    }
}

static jobjectID jvmpi_GetThreadObject(JNIEnv *env)
{
    return (jobjectID)JVMPI_UNHANDLE(JNIEnv2EE(env)->thread);
}

static jobjectID jvmpi_GetMethodClass(jmethodID mid)
{
    struct methodblock *mb = (struct methodblock *)mid;
    return (jobjectID)JVMPI_UNHANDLE(mb->fb.clazz);
}

/***************************************************************************
 * 
 * PART 4: Stub functions called by the other parts of the VM.
 *
 **************************************************************************/

/* this function is called from javai.c when it is safe to create threads */
jint jvmpi_jvm_init_done(void)
{
    /* notify the profiler that VM init is done and it can safely
       do things like thread creation */
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_JVM_INIT_DONE)) {
        JVMPI_Event event;
	event.env_id = EE2JNIEnv(EE());
	event.event_type = JVMPI_EVENT_JVM_INIT_DONE;
	INTERFACE_CALL(NotifyEvent)(&event);
    }
    return JNI_OK;
}

static void
jvmpi_cleanup(void)		/* called atexit */
{
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_JVM_SHUT_DOWN)) {
        JVMPI_Event event;
	event.env_id = EE2JNIEnv(EE());
	event.event_type = JVMPI_EVENT_JVM_SHUT_DOWN;
	INTERFACE_CALL(NotifyEvent)(&event);
    }
    jvmpi_disable_all_notification();
}

/* This function is called when the user asks for a heap
 * dump, by pressing Ctrl-\ on Solaris, for example).
 */
void jvmpi_dump(void)
{
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_DUMP_DATA_REQUEST)) {
        JVMPI_Event event;
	event.env_id = EE2JNIEnv(EE());
	event.event_type = JVMPI_EVENT_DUMP_DATA_REQUEST;
	INTERFACE_CALL(NotifyEvent)(&event);
    }
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_RESET_DATA_REQUEST)) {
        JVMPI_Event event;
	event.env_id = EE2JNIEnv(EE());
	event.event_type = JVMPI_EVENT_RESET_DATA_REQUEST;
	INTERFACE_CALL(NotifyEvent)(&event);
    }
}

/* Class loading/unloading */
void
jvmpi_load_class(ClassClass *cb)
{
    jvmpi_load_class2(cb, 0);
}

static void
jvmpi_load_class2(ClassClass *cb, int flag)
{
    ExecEnv *ee = EE();
    JVMPI_Event event;
    int num_methods = cbMethodsCount(cb);
    int num_fields = cbFieldsCount(cb);
    struct methodblock *mb;
    struct lineno *ln_tab;
    int ln_tab_len;
    struct fieldblock *fb;
    JVMPI_Method *methods;
    JVMPI_Field *statics;
    JVMPI_Field *instances;
    int num_static = 0;
    int num_instance = 0;
    int start = -1;
    int end = -1;
    int cur = -1;
    int i_ln = 0;
    int i_stat = 0;
    int i_inst = 0;
    int i;

    for (i = num_fields, fb = cbFields(cb); --i >= 0; fb++) {
        if (fb->access & ACC_STATIC) {
	    num_static++;
	} else {
	    num_instance++;
	}
    }
    statics = (JVMPI_Field *) jvmpi_calloc(num_static * sizeof(JVMPI_Field));
    instances = (JVMPI_Field *) jvmpi_calloc(num_instance * sizeof(JVMPI_Field));
    for (i = num_fields, fb = cbFields(cb); --i >= 0; fb++) {
        if ((fb->access & ACC_STATIC)) {
	    statics[i_stat].field_name = fb->name;
	    statics[i_stat].field_signature = fb->signature;
	    i_stat++;
	} else {
	    instances[i_inst].field_name = fb->name;
	    instances[i_inst].field_signature = fb->signature;
	    i_inst++;
	}
    }
    methods = (JVMPI_Method *) jvmpi_calloc(num_methods * sizeof(JVMPI_Method));
    for (i = 0, mb = cbMethods(cb); i < num_methods; i++, mb++) {
        ln_tab_len = mb->line_number_table_length;
	ln_tab = mb->line_number_table;
	start = -1;
	end = -1;
	
	if (ln_tab && (ln_tab_len > 0)) {
	    start = ln_tab[0].line_number;
	    end = ln_tab[ln_tab_len -1].line_number;
	    if (end < start) {
	        end = start;
	    }
	    for (i_ln = 1; i_ln < ln_tab_len; i_ln++) {
	        cur = ln_tab[i_ln].line_number;
		if (cur < start) {
		    start = cur;
		}
		if (cur > end) {
		    end = cur;
		}
	    }
	}
	        
	methods[i].method_name = mb->fb.name;
	methods[i].method_signature = mb->fb.signature;
	methods[i].method_id = (jmethodID)mb;
	methods[i].start_lineno = start;
	methods[i].end_lineno = end;
    }

    if (!flag) {
        DisableGC(ee);
    }

    /* fill event info and notify the profiler */
    event.event_type = JVMPI_EVENT_CLASS_LOAD | flag;
    event.env_id = EE2JNIEnv(ee);
    event.u.class_load.class_name = cbName(cb);
    event.u.class_load.source_name = cbSourceName(cb);
    event.u.class_load.num_interfaces = cbImplementsCount(cb);
    event.u.class_load.num_methods = num_methods;
    event.u.class_load.methods = methods;
    event.u.class_load.num_static_fields = num_static;
    event.u.class_load.statics = statics;
    event.u.class_load.num_instance_fields = num_instance;
    event.u.class_load.instances = instances;
    event.u.class_load.class_id = (jobjectID)(JVMPI_UNHANDLE(cb));

    INTERFACE_CALL(NotifyEvent)(&event);

    if (!flag) {
        EnableGC(ee);
    }

    jvmpi_free(statics);
    jvmpi_free(instances);
    jvmpi_free(methods);
}

void 
jvmpi_free_class(ClassClass *cb)
{
    ExecEnv *ee = EE();
    JVMPI_Event event;

    DisableGC(ee);
    /* fill event info and notify the profiler */
    event.event_type = JVMPI_EVENT_CLASS_UNLOAD;
    event.env_id = EE2JNIEnv(ee);
    event.u.class_unload.class_id = (jobjectID)(JVMPI_UNHANDLE(cb));

    INTERFACE_CALL(NotifyEvent)(&event);
    EnableGC(ee);
}

void
jvmpi_load_class_hook(unsigned char **ptrP, unsigned char **end_ptrP,
		      void * (*malloc_f)(unsigned int))
{
    ExecEnv *ee = EE();
    JVMPI_Event event;

    /* fill event info and notify the profiler */
    event.event_type = JVMPI_EVENT_CLASS_LOAD_HOOK;
    event.env_id = EE2JNIEnv(ee);
    event.u.class_load_hook.class_data = *ptrP;
    event.u.class_load_hook.class_data_len = *end_ptrP - *ptrP;
    event.u.class_load_hook.malloc_f = malloc_f;
    INTERFACE_CALL(NotifyEvent)(&event);
    
    *ptrP = event.u.class_load_hook.new_class_data;
    *end_ptrP = *ptrP + event.u.class_load_hook.new_class_data_len;
}

void
jvmpi_new_arena(void)
{
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_NEW_ARENA)) {
        JVMPI_Event event;
	event.env_id = EE2JNIEnv(EE());
	event.event_type = JVMPI_EVENT_NEW_ARENA;
	event.u.new_arena.arena_id = 1;
	event.u.new_arena.arena_name = "Java Heap";
	INTERFACE_CALL(NotifyEvent)(&event);
    }
}

/* Object Allocation/Free */
/* Called by GC routines after it allocates an object */
void 
jvmpi_alloc_object(ExecEnv *ee, JHandle *h)
{
    jvmpi_alloc_object2(ee, h, 0);
}

static void 
jvmpi_alloc_object2(ExecEnv *ee, JHandle *h, jint flag)
{
    JVMPI_Event event;
    int size;
    ClassClass *cb;
    int is_array;
     
    if (h == 0) 
        return;

    cb = jvmpi_get_object_info(h, &is_array, &size);
    
    /* RED and ORANGE reserves and cache blocks should not be reported 
       to the profiler. */
    if (cb != JVMPI_INVALID_CLASS) {

        if (!flag) {
	    DisableGC(ee);
	}

        /* fill event info and notify the profiler */
        event.event_type = JVMPI_EVENT_OBJ_ALLOC | flag;
	event.env_id = EE2JNIEnv(ee);
	event.u.obj_alloc.arena_id = 1;
	event.u.obj_alloc.class_id = (jobjectID)(JVMPI_UNHANDLE(cb));
	event.u.obj_alloc.is_array = is_array;
	event.u.obj_alloc.size = size;
	event.u.obj_alloc.obj_id = (jobjectID)(JVMPI_UNHANDLE(h));

	INTERFACE_CALL(NotifyEvent)(&event);

	if (!flag) {
	    EnableGC(ee);
	}
    }
}

/* Called by GC routines before it frees an object
 * BE CAREFUL!!! This function is called in single-threaded mode.
 */        
void 
jvmpi_free_object(JHandle *h)
{
    ExecEnv *ee = EE();
    JVMPI_Event event;

    DisableGC(ee);
    
    /* fill event info and notify the profiler */
    event.event_type = JVMPI_EVENT_OBJ_FREE;
    event.env_id = EE2JNIEnv(ee);
    event.u.obj_free.obj_id = (jobjectID)(JVMPI_UNHANDLE(h));
    
    INTERFACE_CALL(NotifyEvent)(&event);
    EnableGC(ee);
}

void 
jvmpi_move_object(void *old, void *new)
{
    ExecEnv *ee = EE();
    JVMPI_Event event;
    
    DisableGC(ee);

    /* fill event info and notify the profiler */
    event.event_type = JVMPI_EVENT_OBJ_MOVE;
    event.env_id = EE2JNIEnv(ee);
    event.u.obj_move.obj_id       = old;
    event.u.obj_move.arena_id     = 1;
    event.u.obj_move.new_obj_id   = new;
    event.u.obj_move.new_arena_id = 1;

    INTERFACE_CALL(NotifyEvent)(&event);
    EnableGC(ee);
}

void jvmpi_gc_start(void)
{
    ExecEnv *ee = EE();
    JVMPI_Event event;
    
    /* fill event info and notify the profiler */
    event.event_type = JVMPI_EVENT_GC_START;
    event.env_id = EE2JNIEnv(ee);

    INTERFACE_CALL(NotifyEvent)(&event);
}

void jvmpi_gc_finish(long used_objs, long used_obj_space, long total_obj_space)
{
    ExecEnv *ee = EE();
    JVMPI_Event event;
    
    /* fill event info and notify the profiler */
    event.event_type = JVMPI_EVENT_GC_FINISH;
    event.u.gc_info.used_objects = long_to_jlong(used_objs);
    event.u.gc_info.used_object_space = long_to_jlong(used_obj_space);
    event.u.gc_info.total_object_space = long_to_jlong(total_obj_space);
    event.env_id = EE2JNIEnv(ee);

    INTERFACE_CALL(NotifyEvent)(&event);
}

/* Thread Start/End */
void jvmpi_thread_start(JHandle *tid)
{
    jvmpi_thread_start2(tid, 0);
}

static void jvmpi_thread_start2(JHandle *tid, jint flag)
{
    ExecEnv *ee = EE();
    JVMPI_Event event;

    {
        Hjava_lang_Thread *thread = (Hjava_lang_Thread *)tid;
	HThreadGroup *group = THREAD(thread)->group;
	HThreadGroup *pgroup = group ? unhand(group)->parent : NULL;
	HArrayOfChar *jtname = THREAD(thread)->name;
	Hjava_lang_String *jgname = group ? unhand(group)->name : NULL;
	Hjava_lang_String *jpname = pgroup ? unhand(pgroup)->name : NULL;
	char thread_name[JVMPI_NAME_BUF_LEN];
	char parent_name[JVMPI_NAME_BUF_LEN];
	char group_name[JVMPI_NAME_BUF_LEN];
	
	/* fill event info and notify the profiler */
	event.event_type = JVMPI_EVENT_THREAD_START | flag;
	event.env_id = EE2JNIEnv(ee);
	
	if (jtname) {
	    unicode *body = unhand(jtname)->body;
	    unicode2utf(body, obj_length(jtname), 
			thread_name, sizeof(thread_name));
	    KEEP_POINTER_ALIVE(body);
	    event.u.thread_start.thread_name = thread_name;
	} else {
	    event.u.thread_start.thread_name = NULL;
	}

	if (jgname) {
	    javaString2UTF(jgname, group_name, sizeof(group_name));
	    event.u.thread_start.group_name = group_name;
	} else {
	    event.u.thread_start.group_name = NULL;
	}

	if (jpname) {
	    javaString2UTF(jpname, parent_name, sizeof(parent_name));
	    event.u.thread_start.parent_name = parent_name;
	} else {
	    event.u.thread_start.parent_name = NULL;
	}

	if (!flag) {
	    DisableGC(ee);
	}

	event.u.thread_start.thread_id = (jobjectID)(JVMPI_UNHANDLE(tid));
	event.u.thread_start.thread_env_id = ll2ptr(THREAD((TID)tid)->eetop);

	INTERFACE_CALL(NotifyEvent)(&event);
	if (!flag) {
	    EnableGC(ee);
	}
    }
}

void jvmpi_thread_end(JHandle *tid)
{
    ExecEnv *ee = EE();
    JVMPI_Event event;

    DisableGC(ee);

    /* fill event info and notify the profiler */
    event.event_type = JVMPI_EVENT_THREAD_END;
    event.env_id = EE2JNIEnv(ee);
    INTERFACE_CALL(NotifyEvent)(&event);
    EnableGC(ee);
}

void jvmpi_new_globalref(ExecEnv *ee, jobject ref, JHandle *h)
{
    JVMPI_Event event;
    
    DisableGC(ee);

    /* fill event info and notify the profiler */
    event.event_type = JVMPI_EVENT_JNI_GLOBALREF_ALLOC;
    event.env_id = EE2JNIEnv(ee);
    event.u.jni_globalref_alloc.obj_id = (jobjectID)(JVMPI_UNHANDLE(h));
    event.u.jni_globalref_alloc.ref_id = (jobject)ref;

    INTERFACE_CALL(NotifyEvent)(&event);
    EnableGC(ee);
}

void jvmpi_delete_globalref(ExecEnv *ee, jobject ref)
{
    JVMPI_Event event;
    
    DisableGC(ee);

    /* fill event info and notify the profiler */
    event.event_type = JVMPI_EVENT_JNI_GLOBALREF_FREE;
    event.env_id = EE2JNIEnv(ee);
    event.u.jni_globalref_free.ref_id = (jobject)ref;

    INTERFACE_CALL(NotifyEvent)(&event);
    EnableGC(ee);
}

void jvmpi_new_weakref(ExecEnv *ee, jobject ref, JHandle *h)
{
    JVMPI_Event event;
    
    DisableGC(ee);

    /* fill event info and notify the profiler */
    event.event_type = JVMPI_EVENT_JNI_WEAK_GLOBALREF_ALLOC;
    event.env_id = EE2JNIEnv(ee);
    event.u.jni_globalref_alloc.obj_id = (jobjectID)(JVMPI_UNHANDLE(h));
    event.u.jni_globalref_alloc.ref_id = (jobject)ref;

    INTERFACE_CALL(NotifyEvent)(&event);
    EnableGC(ee);
}

void jvmpi_delete_weakref(ExecEnv *ee, jobject ref)
{
    JVMPI_Event event;
    
    DisableGC(ee);

    /* fill event info and notify the profiler */
    event.event_type = JVMPI_EVENT_JNI_WEAK_GLOBALREF_FREE;
    event.env_id = EE2JNIEnv(ee);
    event.u.jni_globalref_free.ref_id = (jobject)ref;

    INTERFACE_CALL(NotifyEvent)(&event);
    EnableGC(ee);
}

/* Method entry/exit */
void jvmpi_method_entry(ExecEnv *ee, JHandle *h)
{
    JVMPI_Event event;
    struct methodblock *mb = ee->current_frame->current_method;

#ifdef DEBUG
    if (!(mb->fb.access & ACC_STATIC)) {
        sysAssert(reflect_is_assignable(getClass(h), mb->fb.clazz, ee));
    }
#endif /* DEBUG */

    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_METHOD_ENTRY)) {
        event.event_type = JVMPI_EVENT_METHOD_ENTRY;
	event.env_id = EE2JNIEnv(ee);
	event.u.method.method_id = (jmethodID)mb;

	INTERFACE_CALL(NotifyEvent)(&event);
    }
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_METHOD_ENTRY2)) {
	DisableGC(ee);

        event.event_type = JVMPI_EVENT_METHOD_ENTRY2;
	event.env_id = EE2JNIEnv(ee);
	event.u.method_entry2.method_id = (jmethodID)mb;
	event.u.method_entry2.obj_id = (jobjectID)
	    ((mb->fb.access & ACC_STATIC) ? 0 : JVMPI_UNHANDLE(h));

	INTERFACE_CALL(NotifyEvent)(&event);
	EnableGC(ee);
    }
}

void jvmpi_method_exit(ExecEnv *ee)
{
    JVMPI_Event event;
    
    /* fill event info and notify the profiler */
    event.event_type = JVMPI_EVENT_METHOD_EXIT;
    event.env_id = EE2JNIEnv(ee);
    event.u.method.method_id = 
        (jmethodID)ee->current_frame->current_method;
    INTERFACE_CALL(NotifyEvent)(&event);
}

/* compiled method load/unload from memory (called through the JIT interface) */
void jvmpi_load_compiled_method(compiled_method_t *compiled_method_info)
{
    JVMPI_Event event;
    ExecEnv *ee = compiled_method_info->ee;
    struct methodblock *mb = ee->current_frame->current_method;
    
    /* fill event info and notify the profiler */
    event.event_type = JVMPI_EVENT_LOAD_COMPILED_METHOD;
    event.env_id = EE2JNIEnv(compiled_method_info->ee);
    event.u.compiled_method_load.method_id = (jmethodID)mb;
    event.u.compiled_method_load.code_addr = compiled_method_info->code_addr;
    event.u.compiled_method_load.code_size = compiled_method_info->code_size;
    event.u.compiled_method_load.lineno_table_size = 
        compiled_method_info->lineno_table_len;
    event.u.compiled_method_load.lineno_table = 
        compiled_method_info->lineno_table;
    INTERFACE_CALL(NotifyEvent)(&event);
}

void jvmpi_unload_compiled_method(ExecEnv *ee)
{
    JVMPI_Event event;
    
    /* fill event info and notify the profiler */
    event.event_type = JVMPI_EVENT_UNLOAD_COMPILED_METHOD;
    event.env_id = EE2JNIEnv(ee);
    event.u.compiled_method_unload.method_id = 
        (jmethodID)ee->current_frame->current_method;
    INTERFACE_CALL(NotifyEvent)(&event);
}
        

/***************************************************************************
 * 
 * PART 6: Support for heap, monitor and object dumps
 *
 **************************************************************************/
/* 
 * Note: the functions called by the GC during a heap dump are run in single 
 * thread mode, so we cannot call functions like malloc.
 */

/* Allocate a block from the dump context. */
static void *jvmpi_dump_alloc(int size)
{
    void *result = jvmpi_dump_context->ptr;
    size = (size + 7) & (~7);
    jvmpi_dump_context->ptr += size;
    if (jvmpi_dump_context->ptr >= jvmpi_dump_context->end_ptr)
        return NULL;
    return result;
}

static void jvmpi_dump_write(void *buf, int size)
{
    jvmpi_dump_context->ptr += size;
    if (jvmpi_dump_context->ptr >= jvmpi_dump_context->end_ptr)
        return;
    memcpy(jvmpi_dump_context->ptr - size, buf, size);
}

static void jvmpi_dump_write_id(void *p)
{
    jvmpi_dump_write(&p, sizeof(void *));
}

static void jvmpi_dump_write_u4(unsigned int u4)
{
    u4 = htonl(u4);
    jvmpi_dump_write(&u4, sizeof(int32_t));
}

static void jvmpi_dump_write_u2(unsigned short u2)
{
    u2 = htons(u2);
    jvmpi_dump_write(&u2, sizeof(unsigned short));
}

static void jvmpi_dump_write_u1(unsigned char u1)
{
    jvmpi_dump_write(&u1, sizeof(unsigned char));
}

static void jvmpi_dump_write_64(void *p, bool_t is_double)
{
    jvalue t;
    int64_t i8 = GET_INT64(t, p);
    if (is_double) {
        ll2double_bits(&i8);
    }
    jvmpi_dump_write_u4(ll_high(i8));
    jvmpi_dump_write_u4(ll_low(i8));
}

static int jvmpi_get_dump_level()
{
    return jvmpi_dump_context->dump_level;
}

/* Add a trace of a root thread obj into jvmpi_dump_context */
void jvmpi_dump_add_trace(ExecEnv *ee)
{
    int n_frames = 0;
    JavaFrame *frame = ee->current_frame;
    JavaFrame frame_buf;
    JVMPI_CallTrace *trace;

    if (jvmpi_get_dump_level() == JVMPI_DUMP_LEVEL_0) {
        return;
    }

    trace = &jvmpi_dump_context->traces[jvmpi_dump_context->n_traces++];

    if (jvmpi_dump_context->n_traces >= JVMPI_DUMP_MAX_TRACES) {
        panic("too many threads");
    }

    if (frame && IS_JIT_FRAME(frame))
	frame = CompiledFrameUpdate(frame);
    while (frame) {
        struct methodblock *mb = frame->current_method;
	if (mb) {
	    n_frames++;
	}
	frame = FRAME_PREV(frame, &frame_buf);
    }

    trace->env_id = EE2JNIEnv(ee);
    trace->num_frames = n_frames;
    trace->frames = jvmpi_dump_alloc(n_frames * sizeof(JVMPI_CallFrame));
    if (trace->frames == NULL) {
        return;
    }

    jvmpi_GetCallTrace(trace, n_frames);
}

/* Called by GC to add a root into jvmpi_dump_context. */
void jvmpi_dump_add_root(JHandle *h,
			 int kind,
			 void *info,
			 void *extra_info)
{
    int size;
    ClassClass *cb;
    int is_array;

    if (unhand(h) == 0) 
        return;

    if (jvmpi_get_dump_level() == JVMPI_DUMP_LEVEL_0) {
        return;
    }

    cb = jvmpi_get_object_info(h, &is_array, &size);
    
    /* RED and ORANGE reserves and cache blocks should not be reported 
       to the profiler. */
    if (cb == JVMPI_INVALID_CLASS) {
        return;
    }

    if (jvmpi_dump_context->begin_dump == NULL) {
       jvmpi_dump_context->begin_dump = jvmpi_dump_context->ptr; 
    }

    switch (kind) {
        case GC_ROOT_UNKNOWN:
	    jvmpi_dump_write_u1((unsigned char)JVMPI_GC_ROOT_UNKNOWN);
	    jvmpi_dump_write_id(JVMPI_UNHANDLE(h));
	    return;
        case GC_ROOT_MONITOR_USED:
	    jvmpi_dump_write_u1((unsigned char)JVMPI_GC_ROOT_MONITOR_USED);
	    jvmpi_dump_write_id(JVMPI_UNHANDLE(h));
	    return;
        case GC_ROOT_STICKY_CLASS:
	    jvmpi_dump_write_u1((unsigned char)JVMPI_GC_ROOT_STICKY_CLASS);
	    jvmpi_dump_write_id(JVMPI_UNHANDLE(h));
	    return;
        case GC_ROOT_JNI_GLOBAL:
	    jvmpi_dump_write_u1((unsigned char)JVMPI_GC_ROOT_JNI_GLOBAL);
	    jvmpi_dump_write_id(JVMPI_UNHANDLE(h));
            jvmpi_dump_write_id(info);
	    return;
        case GC_ROOT_NATIVE_STACK:
	    jvmpi_dump_write_u1((unsigned char)JVMPI_GC_ROOT_NATIVE_STACK);
	    jvmpi_dump_write_id(JVMPI_UNHANDLE(h));
	    jvmpi_dump_write_id(extra_info);
	    return;
        case GC_ROOT_THREAD_BLOCK:
	    jvmpi_dump_write_u1((unsigned char)JVMPI_GC_ROOT_THREAD_BLOCK);
	    jvmpi_dump_write_id(JVMPI_UNHANDLE(h));
            jvmpi_dump_write_id(extra_info);
	    return;
        case GC_ROOT_JNI_LOCAL:
        case GC_ROOT_JAVA_FRAME: {
	    int i;
	    if (kind == GC_ROOT_JNI_LOCAL) {
	        jvmpi_dump_write_u1((unsigned char)JVMPI_GC_ROOT_JNI_LOCAL);
	    } else {
	        jvmpi_dump_write_u1((unsigned char)JVMPI_GC_ROOT_JAVA_FRAME);
	    }
	    jvmpi_dump_write_id(JVMPI_UNHANDLE(h));
	    for (i = 0; i < jvmpi_dump_context->n_traces; i++) {
	        JVMPI_CallTrace *trace = &jvmpi_dump_context->traces[i];
		jint trace_depth = trace->num_frames;
	        if (JNIEnv2EE(trace->env_id) == extra_info) {
		    jvmpi_dump_write_id((void *)EE2JNIEnv((ExecEnv *)extra_info));
		    jvmpi_dump_write_u4(trace_depth - (jint)info - 1);
		    return;
		}
	    }
	    sysAssert(FALSE);
	}
    }
}

/* dump an object instance */
static void jvmpi_dump_instance(JHandle *h, ClassClass *cb)
{
    OBJECT *obj = h->obj;
    ClassClass *cb0;
    int dump_level = jvmpi_get_dump_level();
    int size = 0;
    
    jvmpi_dump_write_u1(JVMPI_GC_INSTANCE_DUMP);
    jvmpi_dump_write_id(JVMPI_UNHANDLE(h));
    jvmpi_dump_write_id(JVMPI_UNHANDLE(cb));

    /* dump the field values */
    for (cb0 = cb; cb0; cb0 = cbSuperclass(cb0)) {
        int i;
	struct fieldblock *fb;

	for (i = cbFieldsCount(cb0), fb = cbFields(cb0);--i >= 0; fb++) {
	    char ty;
	    if (fb->access & ACC_STATIC) {
	        continue;
	    }
	    ty = fb->signature[0];
	    switch (ty) {
	    case 'L':
	    case '[':
	        size += sizeof(void *);
		break;
	    case 'J':
	    case 'D':
	        if (dump_level == JVMPI_DUMP_LEVEL_2) {
		    size += 8;
		}
		break;
	    case 'I':
	    case 'F':
	        if (dump_level == JVMPI_DUMP_LEVEL_2) {
		    size += 4;
		}
		break;
	    case 'S':
	    case 'C':
	        if (dump_level == JVMPI_DUMP_LEVEL_2) {
		    size += 2;
		}
		break;
	    case 'B':
	    case 'Z':
	        if (dump_level == JVMPI_DUMP_LEVEL_2) {
		    size += 1;
		}
		break;
	    }
	}
    }

    jvmpi_dump_write_u4(size);

    for (cb0 = cb; cb0; cb0 = cbSuperclass(cb0)) {
        int i;
	struct fieldblock *fb;

	for (i = cbFieldsCount(cb0), fb = cbFields(cb0); --i >= 0; fb++) {
	    int index;
	    char ty;
	    if (fb->access & ACC_STATIC) {
	        continue;
	    }
	    index = fb->u.offset / sizeof(OBJECT);
	    ty = fb->signature[0];
	    switch (ty) {
	    case 'L':
	    case '[':
	        jvmpi_dump_write_id(JVMPI_UNHANDLE((JHandle *)obj[index]));
		break;
	    case 'J':
	    case 'D':
	        if (dump_level == JVMPI_DUMP_LEVEL_2) {
		    jvmpi_dump_write_64(&(obj[index]), ty == 'D');
		}
		break;
	    case 'I':
	    case 'F':
	        if (dump_level == JVMPI_DUMP_LEVEL_2) {
		    jvmpi_dump_write_u4(obj[index]);
		}
		break;
	    case 'S':
	    case 'C':
	        if (dump_level == JVMPI_DUMP_LEVEL_2) {
		    jvmpi_dump_write_u2((unsigned short)obj[index]);
		}
		break;
	    case 'B':
	    case 'Z':
	        if (dump_level == JVMPI_DUMP_LEVEL_2) {
		    jvmpi_dump_write_u1((unsigned char)obj[index]);
		}
		break;
	    }
	}
    }    
} 
 
/* dump a class */   
static void jvmpi_dump_class(JHandle *h)
{
    ClassClass *cb = (ClassClass *)h;
    union cp_item_type *cp = cbConstantPool(cb);
    int dump_level = jvmpi_get_dump_level();
    int num_fields = cbFieldsCount(cb);
    struct fieldblock *fb = cbFields(cb);

    jvmpi_dump_write_u1(JVMPI_GC_CLASS_DUMP);
    jvmpi_dump_write_id(JVMPI_UNHANDLE(cb));
    jvmpi_dump_write_id(JVMPI_UNHANDLE(cbSuperclass(cb)));
    jvmpi_dump_write_id(JVMPI_UNHANDLE((JHandle *)cbLoader(cb)));
    jvmpi_dump_write_id(JVMPI_UNHANDLE(cbSigners(cb)));
    jvmpi_dump_write_id(JVMPI_UNHANDLE(cbProtectionDomain(cb)));
    jvmpi_dump_write_id(0);
    jvmpi_dump_write_id(0);
    jvmpi_dump_write_u4(cbInstanceSize(cb));

    /* dump the interfaces implemented by this class */
    {
        int i;
	unsigned char *cp_types = cp[0].type;
	for (i = 0; i < cbImplementsCount(cb); i++) {
	    int index = cbImplements(cb)[i];
	    switch (cp_types[index]) {
	    case CONSTANT_Class:
	        jvmpi_dump_write_id(NULL);
		break;
	    case CONSTANT_Class | CONSTANT_POOL_ENTRY_RESOLVED:
	        jvmpi_dump_write_id(JVMPI_UNHANDLE(cp[index].clazz));
		break;
	    }
	}
    }

    /* dump constant pool info */
    if (cp == 0 || cp[0].type == 0) {
        jvmpi_dump_write_u2(0);
    } else {
        int i;
	unsigned char *cp_types = cp[0].type;
	int n = 0;

	for (i = 1; i < cbConstantPoolCount(cb); i++) {
	    switch (cp_types[i]) {
	    case CONSTANT_String | CONSTANT_POOL_ENTRY_RESOLVED:
	    case CONSTANT_Class | CONSTANT_POOL_ENTRY_RESOLVED:
	      n++;
	      break;
	    }
	}

	jvmpi_dump_write_u2((unsigned short)n);

	for (i = 1; i < cbConstantPoolCount(cb); i++) {
	    switch (cp_types[i]) {
	    case CONSTANT_String | CONSTANT_POOL_ENTRY_RESOLVED:
	    case CONSTANT_Class | CONSTANT_POOL_ENTRY_RESOLVED:
	        jvmpi_dump_write_u2((unsigned short)i);
	        jvmpi_dump_write_u1(T_CLASS);
		jvmpi_dump_write_id(JVMPI_UNHANDLE((JHandle *)cp[i].p));
		break;
	    }
	}
    }

    /* dump static field values */
    if ((num_fields > 0) && (fb != NULL)) {
        int i;
	for (i = num_fields; --i >= 0; fb++) {
	    if (!(fb->access & ACC_STATIC)) {
	        continue;
	    }
	    switch (fb->signature[0]) {
	    case 'J':
	    case 'D': {
	        if (dump_level == JVMPI_DUMP_LEVEL_2) {
		    jvmpi_dump_write_64(twoword_static_address(fb),
					fb->signature[0] == 'D');
		}
		break;
	    }
	    case '[':
	    case 'L': {
	        void *p = *(void **)normal_static_address(fb);
	        jvmpi_dump_write_id(JVMPI_UNHANDLE((JHandle *)p));
		break;
	    }
	    case 'I':
	    case 'F': {
	        if (dump_level == JVMPI_DUMP_LEVEL_2) {
		    unsigned int nn = *(int*)normal_static_address(fb);
		    jvmpi_dump_write_u4(nn);
		}
		break;
	    }
	    case 'S':
	    case 'C': {
	        if (dump_level == JVMPI_DUMP_LEVEL_2) {
		    unsigned int nn = *(int*)normal_static_address(fb);
		    jvmpi_dump_write_u2((unsigned short)nn);
		}
		break;
	    }
	    case 'Z':
	    case 'B': {
	        if (dump_level == JVMPI_DUMP_LEVEL_2) {
		    unsigned int nn = *(int*)normal_static_address(fb);
		    jvmpi_dump_write_u1((unsigned char)nn);
		}
		break;
	    }
	    }
	}
    } 
}

/* dump an object array */
static void jvmpi_dump_obj_array(JHandle *h)
{
    int i;
    unsigned int size = obj_length(h);
    ClassClass *cb = ((ClassClass **)(h->obj))[size];
    HObject **body = (HObject **)(h->obj);
	    
    jvmpi_dump_write_u1(JVMPI_GC_OBJ_ARRAY_DUMP);
    jvmpi_dump_write_id(JVMPI_UNHANDLE(h));
    jvmpi_dump_write_u4(size);
    jvmpi_dump_write_id(JVMPI_UNHANDLE(cb));

    for (i = 0; i < size; i++) {
        jvmpi_dump_write_id(JVMPI_UNHANDLE(body[i]));
    }
}

/* dump a primitive array */
static void jvmpi_dump_prim_array(JHandle *h)
{
    int i;
    unsigned int size = obj_length(h);
        
    jvmpi_dump_write_u1(JVMPI_GC_PRIM_ARRAY_DUMP);
    jvmpi_dump_write_id(JVMPI_UNHANDLE(h));
    jvmpi_dump_write_u4(size);
    jvmpi_dump_write_u1((unsigned char)obj_flags(h));

    if (jvmpi_get_dump_level() == JVMPI_DUMP_LEVEL_2) {
        switch (obj_flags(h)) {
	case T_BOOLEAN:
	case T_BYTE: {  
	    char *body = (char *)(h->obj);
		    
	    for (i = 0; i < size; i++) {
	        jvmpi_dump_write_u1(body[i]);
	    }
	    break;
	}
	case T_SHORT: 
	case T_CHAR: {
	    unsigned short *body = (unsigned short *)(h->obj);
	    
	    for (i = 0; i < size; i++) {
	        jvmpi_dump_write_u2(body[i]);
	    }
	    break;
	}
	case T_INT: 
	case T_FLOAT: {
	    unsigned int *body = (unsigned int *)(h->obj);
	
	    for (i = 0; i < size; i++) {
	        jvmpi_dump_write_u4(body[i]);
	    }
	    break;
	}
	case T_LONG:
	case T_DOUBLE:{
	    jvalue *body = (jvalue *)(h->obj);
	
	    for (i = 0; i < size; i++) {
	        jvmpi_dump_write_64(&body[i], obj_flags(h) == T_DOUBLE);
	    }
	    break;
	}
	}
    }
}

/* Called by GC & jvmpi_RequestEvent to add an object into jvmpi_dump_context */
void jvmpi_dump_object(JHandle *h)
{
    if (jvmpi_dump_context->begin_dump == NULL) {
        jvmpi_dump_context->begin_dump = jvmpi_dump_context->ptr; 
    }

    if (jvmpi_get_dump_level() == JVMPI_DUMP_LEVEL_0) {
        jvmpi_dump_write_u1((unsigned char)obj_flags(h));
	jvmpi_dump_write_id(JVMPI_UNHANDLE(h));
    } else {
        switch (obj_flags(h)) {
	case T_NORMAL_OBJECT: {
	    ClassClass *cb = obj_classblock(h);
	    if (cb != classJavaLangClass) {
	        jvmpi_dump_instance(h, cb);
	    } else {
	        jvmpi_dump_class(h);
	    }
	    break;
	} 
	case T_CLASS: {
	    jvmpi_dump_obj_array(h);
	    break;
	}
	case T_BOOLEAN:
	case T_BYTE: 
	case T_SHORT: 
	case T_CHAR: 
	case T_INT: 
	case T_FLOAT:
	case T_LONG:
	case T_DOUBLE: {
	    jvmpi_dump_prim_array(h);
	    break;
	}
	}
    }
}



/* called when the profiler wants a heap dump */
static jint jvmpi_heap_dump(int heap_dump_level)
{
    jvmpi_dump_context_t context_buf;
    size_t buffer_size;
    ExecEnv *ee = EE();

    JVMPI_DUMP_CONTEXT_LOCK(ee);
    memset(&context_buf, 0, sizeof(context_buf));
    jvmpi_dump_context = &context_buf;
    jvmpi_dump_context->dump_level = heap_dump_level;
    
    /* get an estimate of heap dump size */
    buffer_size = jvmpi_get_heap_size() * 1.5;
    jvmpi_dump_context->end_ptr = jvmpi_dump_context->ptr = 0;
    while (jvmpi_dump_context->ptr >= jvmpi_dump_context->end_ptr) {
        /* allocate enough memory to collect the heap dump as we execute
	 * in single thread mode during heap dump collection and cannot
	 * afford to call malloc 
	 */
        char *ptr = sysMapMem(buffer_size, &buffer_size);
	if (ptr == NULL ||
	    sysCommitMem(ptr, buffer_size, &buffer_size) == NULL) {
	    out_of_memory();
	}
	jvmpi_dump_context->ptr = ptr;
	jvmpi_dump_context->end_ptr = ptr + buffer_size;
	/* notify the GC */
	heap_lock(ee);
	jvmpi_in_dump_heap(TRUE);
	gc(~0);
	jvmpi_in_dump_heap(FALSE);
	if (jvmpi_dump_context->ptr < jvmpi_dump_context->end_ptr) {
	    /* notify the profiler */
	    {
	        JVMPI_Event event;
		event.event_type = JVMPI_EVENT_HEAP_DUMP | JVMPI_REQUESTED_EVENT;
		event.u.heap_dump.dump_level = heap_dump_level;
		event.u.heap_dump.begin = jvmpi_dump_context->begin_dump;
		event.u.heap_dump.end = jvmpi_dump_context->ptr;
		event.u.heap_dump.num_traces = jvmpi_dump_context->n_traces;
		event.u.heap_dump.traces = jvmpi_dump_context->traces;

		DisableGC(ee);
		INTERFACE_CALL(NotifyEvent)(&event);
		EnableGC(ee);
	    }
	}
	heap_unlock(ee);
	/* increase buffer size in case we over shot the allocated size */
	sysUnmapMem(ptr, buffer_size, &buffer_size);
	buffer_size *= 1.5;
    }
    JVMPI_DUMP_CONTEXT_UNLOCK(ee);
    return JVMPI_SUCCESS;
}

static int
jvmpi_dump_monitor_threads_helper(sys_thread_t *t, void *arg)
{
    struct execenv *ee = SysThread2EE(t);
    HThread *p = (HThread *)ee->thread;

    if (p) {
        jvmpi_dump_add_trace(ee);
    }

    return SYS_OK;
}

static void jvmpi_dump_monitor(sys_mon_t *mid, int kind, void *arg)
{
    #define MAX_WAITERS 1000
    sys_thread_t *monitor_waiters[MAX_WAITERS];
    sys_thread_t *condvar_waiters[MAX_WAITERS];
    sys_mon_info info;
    int i;

    jvmpi_dump_write_u1((unsigned char)kind);
    jvmpi_dump_write_id(arg);
    if (kind == JVMPI_MONITOR_RAW) {
        jvmpi_dump_write_id(mid);
    }

    info.monitor_waiters = monitor_waiters;
    info.condvar_waiters = condvar_waiters;
    info.sz_monitor_waiters = MAX_WAITERS;
    info.sz_condvar_waiters = MAX_WAITERS;

    sysMonitorGetInfo(mid, &info);

    if (info.owner) {
        jvmpi_dump_write_id((void *)EE2JNIEnv(SysThread2EE(info.owner)));
	jvmpi_dump_write_u4(info.entry_count);
    } else {
        jvmpi_dump_write_id(NULL);
	jvmpi_dump_write_u4(0);
    }

    jvmpi_dump_write_u4(info.n_monitor_waiters);
    for (i = 0; i < info.n_monitor_waiters; i++) {
        jvmpi_dump_write_id((void *)EE2JNIEnv(SysThread2EE(info.monitor_waiters[i])));
    }

    jvmpi_dump_write_u4(info.n_condvar_waiters);
    for (i = 0; i < info.n_condvar_waiters; i++) {
        jvmpi_dump_write_id((void *)EE2JNIEnv(SysThread2EE(info.condvar_waiters[i])));
    }
}

static void jvmpi_dump_monitor_jmon_helper(monitor_t *mid, void *arg)
{
    extern bool_t isHandle(void *p); /* defined in gc.c */
    unsigned int key = mid->key;

    if (sysMonitorInUse(sysmon(mid)) && isHandle((void *)key)) {
        jvmpi_dump_monitor(sysmon(mid),
			   JVMPI_MONITOR_JAVA, 
			   JVMPI_UNHANDLE((JHandle*)key));
    }
}

static void jvmpi_dump_monitor_rmon_helper(reg_mon_t *reg, void *ignored)
{
    if (sysMonitorInUse(reg->mid)) {
        jvmpi_dump_monitor(reg->mid, JVMPI_MONITOR_RAW, reg->name);
    }
}

static void jvmpi_collect_monitor_info()
{
    sys_thread_t *self = sysThreadSelf();

    CACHE_LOCK(self);
    REGISTRY_LOCK(self);
    sysThreadSingle();

    sysThreadEnumerateOver(jvmpi_dump_monitor_threads_helper, (void *)0);
    if (jvmpi_dump_context->ptr + jvmpi_dump_context->n_traces * sizeof(jint)
	>= jvmpi_dump_context->end_ptr) {
        jvmpi_dump_context->ptr = jvmpi_dump_context->end_ptr;
    } else {
        int i;
	jint *threads_status = (jint *)jvmpi_dump_context->ptr;
        for (i = 0; i < jvmpi_dump_context->n_traces; i++) {
	    sys_thread_t *t = 
	       EE2SysThread(JNIEnv2EE(jvmpi_dump_context->traces[i].env_id));
            threads_status[i] = sysThreadGetStatus(t, NULL);
	}
        jvmpi_dump_context->ptr += jvmpi_dump_context->n_traces * sizeof(jint);
    }
    jvmpi_dump_context->begin_dump = jvmpi_dump_context->ptr;
    monitorEnumerate(jvmpi_dump_monitor_jmon_helper, (void *)0);
    registeredEnumerate(jvmpi_dump_monitor_rmon_helper, (void *)0);

    sysThreadMulti();
    REGISTRY_UNLOCK(self);
    CACHE_UNLOCK(self);
}

/* called when the profiler wants a monitor dump */
static jint jvmpi_monitor_dump(void)
{
    jvmpi_dump_context_t context_buf;
    unsigned int buffer_size;
    ExecEnv *ee = EE();

    JVMPI_DUMP_CONTEXT_LOCK(ee);
    memset(&context_buf, 0, sizeof(context_buf));
    jvmpi_dump_context = &context_buf;
    jvmpi_dump_context->dump_level = JVMPI_DUMP_LEVEL_2;

    /* get an estimate of monitor dump size */
    buffer_size = 1024 * 256; /* 256 K to start with */
    jvmpi_dump_context->end_ptr = jvmpi_dump_context->ptr = 0;
    while (jvmpi_dump_context->ptr >= jvmpi_dump_context->end_ptr) {
        /* allocate enough memory to collect the heap dump as we execute
	 * in single thread mode during heap dump collection and cannot
	 * afford to call malloc 
	 */
        char *ptr = sysMapMem(buffer_size, &buffer_size);
	if (ptr == NULL ||
	    sysCommitMem(ptr, buffer_size, &buffer_size) == NULL) {
	    out_of_memory();
	}
	jvmpi_dump_context->ptr = ptr;
	jvmpi_dump_context->end_ptr = ptr + buffer_size;
	

	/* Grab the QUEUE lock to make sure threads don't terminate */
	QUEUE_LOCK(EE2SysThread(ee));

	jvmpi_collect_monitor_info();

	if (jvmpi_dump_context->ptr < jvmpi_dump_context->end_ptr) {
	    int n_traces = jvmpi_dump_context->n_traces;
	    JVMPI_CallTrace *traces = jvmpi_dump_context->traces;

	    {
	        JVMPI_Event event;
		event.event_type = JVMPI_EVENT_MONITOR_DUMP | JVMPI_REQUESTED_EVENT;
		event.env_id = EE2JNIEnv(EE());
		event.u.monitor_dump.begin = jvmpi_dump_context->begin_dump;
		event.u.monitor_dump.end = jvmpi_dump_context->ptr;
		event.u.monitor_dump.num_traces = n_traces;
		event.u.monitor_dump.traces = traces;
		event.u.monitor_dump.threads_status =
		    (jint*)(jvmpi_dump_context->begin_dump) - n_traces;
		INTERFACE_CALL(NotifyEvent)(&event);
	    }
	}

	QUEUE_UNLOCK(EE2SysThread(ee));

	/* increase buffer size in case we over shot the allocated size */
	sysUnmapMem(ptr, buffer_size, &buffer_size);
	buffer_size *= 1.5;
    }
    JVMPI_DUMP_CONTEXT_UNLOCK(ee);
    return JVMPI_SUCCESS;
}

void jvmpi_monitor_contended_enter(sys_thread_t *self, sys_mon_t *mid)
{
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_RAW_MONITOR_CONTENDED_ENTER)) {
        char *name = lookupRegisteredMonitor(self, mid);
	if (name) {
	    JVMPI_Event event;
	    event.env_id = EE2JNIEnv(SysThread2EE(self));
	    event.event_type = JVMPI_EVENT_RAW_MONITOR_CONTENDED_ENTER;
	    event.u.raw_monitor.name = name;
	    event.u.raw_monitor.id = (JVMPI_RawMonitor)mid;
	    INTERFACE_CALL(NotifyEvent)(&event);
	    return;
	}
    }
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_MONITOR_CONTENDED_ENTER)) {
        JHandle *h = lookupJavaMonitor(mid);
	if (h) {
	    JVMPI_Event event;
	    DisableGC(SysThread2EE(self));
	    event.env_id = EE2JNIEnv(SysThread2EE(self));
	    event.event_type = JVMPI_EVENT_MONITOR_CONTENDED_ENTER;
	    event.u.monitor.object = (jobjectID)JVMPI_UNHANDLE(h);
	    INTERFACE_CALL(NotifyEvent)(&event);
	    EnableGC(SysThread2EE(self));
	    return;
	}
    }
}

void jvmpi_monitor_contended_entered(sys_thread_t *self, sys_mon_t *mid)
{
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_RAW_MONITOR_CONTENDED_ENTERED)) {
        char *name = lookupRegisteredMonitor(self, mid);
	if (name) {
	    JVMPI_Event event;
	    event.env_id = EE2JNIEnv(SysThread2EE(self));
	    event.event_type = JVMPI_EVENT_RAW_MONITOR_CONTENDED_ENTERED;
	    event.u.raw_monitor.name = name;
	    event.u.raw_monitor.id = (JVMPI_RawMonitor)mid;
	    INTERFACE_CALL(NotifyEvent)(&event);
	    return;
	}
    }
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_MONITOR_CONTENDED_ENTERED)) {
        JHandle *h = lookupJavaMonitor(mid);
	if (h) {
	    JVMPI_Event event;
	    DisableGC(SysThread2EE(self));
	    event.env_id = EE2JNIEnv(SysThread2EE(self));
	    event.event_type = JVMPI_EVENT_MONITOR_CONTENDED_ENTERED;
	    event.u.monitor.object = (jobjectID)JVMPI_UNHANDLE(h);
	    INTERFACE_CALL(NotifyEvent)(&event);
	    EnableGC(SysThread2EE(self));
	    return;
	}
    }
}

void jvmpi_monitor_contended_exit(sys_thread_t *self, sys_mon_t *mid)
{
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_RAW_MONITOR_CONTENDED_EXIT)) {
        char *name = lookupRegisteredMonitor(self, mid);
	if (name) {
	    JVMPI_Event event;
	    event.env_id = EE2JNIEnv(SysThread2EE(self));
	    event.event_type = JVMPI_EVENT_RAW_MONITOR_CONTENDED_EXIT;
	    event.u.raw_monitor.name = name;
	    event.u.raw_monitor.id = (JVMPI_RawMonitor)mid;
	    INTERFACE_CALL(NotifyEvent)(&event);
	    return;
	}
    }
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_MONITOR_CONTENDED_EXIT)) {
        JHandle *h = lookupJavaMonitor(mid);
	if (h) {
	    JVMPI_Event event;
	    DisableGC(SysThread2EE(self));
	    event.env_id = EE2JNIEnv(SysThread2EE(self));
	    event.event_type = JVMPI_EVENT_MONITOR_CONTENDED_EXIT;
	    event.u.monitor.object = (jobjectID)JVMPI_UNHANDLE(h);
	    INTERFACE_CALL(NotifyEvent)(&event);
	    EnableGC(SysThread2EE(self));
	    return;
	}
    }
}

void jvmpi_monitor_wait(JHandle *obj, jlong millis)
{
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_MONITOR_WAIT)) {
        ExecEnv *ee = EE();
        JVMPI_Event event;
	DisableGC(ee);
	event.env_id = EE2JNIEnv(ee);
	event.event_type = JVMPI_EVENT_MONITOR_WAIT;
	event.u.monitor_wait.object = (jobjectID)JVMPI_UNHANDLE(obj);
	event.u.monitor_wait.timeout = millis;
	INTERFACE_CALL(NotifyEvent)(&event);
	EnableGC(ee);
    }
}

void jvmpi_monitor_waited(JHandle *obj)
{
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_MONITOR_WAITED)) {
        ExecEnv *ee = EE();
        JVMPI_Event event;
	DisableGC(ee);
	event.env_id = EE2JNIEnv(ee);
	event.event_type = JVMPI_EVENT_MONITOR_WAITED;
	event.u.monitor_wait.object = (jobjectID)JVMPI_UNHANDLE(obj);
	INTERFACE_CALL(NotifyEvent)(&event);
	EnableGC(ee);
    }
}

jobject jobjectID2jobject(jobjectID obj)
{
    return MkRefLocal(EE2JNIEnv(EE()), jvmpi_obj_reverse_map(obj));
}

jobjectID jobject2jobjectID(jobject ref)
{
    return (jobjectID)JVMPI_UNHANDLE(DeRef(EE2JNIEnv(EE()), ref));
}

#endif /* HPROF */
