/*
 * @(#)monitor.c	1.61 98/08/14
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * System-independent, reentrant monitors.
 */

#include "typedefs.h"
#include "monitor.h"
#include "monitor_cache.h"
#include "threads.h"
#include "log.h"
#include "sys_api.h"
#include "common_exceptions.h"

#include "java_lang_Thread.h"


/*-
 * Registration of monitors not allocated from the monitor cache.
 *
 * All monitors except those in the monitor cache should be registered
 * via monitorRegister().
 *
 * One tricky point is that access to the monitor registry is
 * protected by a registered monitor.  That may require special
 * treatment when doing some operations, e.g. when a thread is dying
 * it leaves all held monitors, but uses the registry monitor to do
 * so.  The responsibility is currently put on the helper function
 * passed to registeredEnumerate() to take care of this special case.
 * registeredEnumerate() is intended to be the only access to the
 * monitor registry.
 *
 * Note also that monitorRegister() is used on the registry lock, and
 * also uses the lock.  This works because we don't actually need the
 * monitor to be registered to use it -- we only need it to be
 * initialized.
 */

sys_mon_t *_registry_lock = 0;
static reg_mon_t *MonitorRegistry;

void
monitorRegistryInit(void)
{
    _registry_lock = (sys_mon_t *) sysMalloc(sysMonitorSizeof());
    REGISTRY_LOCK_INIT();
}

void
monitorRegister(sys_mon_t *mid, char *name)
{
    reg_mon_t *reg;
    sys_thread_t *self = sysThreadSelf();

    sysMonitorInit(mid);

    if (self && _registry_lock) {
        REGISTRY_LOCK(self);
    }
    reg = MonitorRegistry;
    while (reg) {
        if (reg->mid == NULL) {
	    break;
	}
	reg = reg->next;
    }
    if (self && _registry_lock) {
        REGISTRY_UNLOCK(self);
    }

    if (!reg) {
        reg = (reg_mon_t *) sysMalloc(sizeof(reg_mon_t));
	if (!reg) {
	    ThrowOutOfMemoryError(0, 0);
	    return;
	}
    }
    reg->mid = mid;
    if (!(reg->name = sysMalloc(strlen(name) + 1))) {
	ThrowOutOfMemoryError(0, 0);
	return;
    }
    strcpy(reg->name, name);
    if (self && _registry_lock) {
        REGISTRY_LOCK(self);
    }
    reg->next = MonitorRegistry;
    MonitorRegistry = reg;

    if (self && _registry_lock) {
        REGISTRY_UNLOCK(self);
    }
}

/*-
 * If monitors are dynamically allocated and deallocated (none currently
 * are), it will be necessary that a registered monitor be unregistered
 * prior to deallocation to avoid dangling pointers in the monitor 
 * registry.
 */
void
monitorUnregister(sys_mon_t *mid)
{
    reg_mon_t *reg;

    REGISTRY_LOCK(sysThreadSelf());
    reg = MonitorRegistry;
    while (reg && reg->mid != mid) {
	reg = reg->next;
    }

    sysAssert(reg);	/* Else bad call or inconsistent state */

    reg->name = NULL;
    reg->mid = NULL;
    REGISTRY_UNLOCK(sysThreadSelf());
}

/*-
 * Call a function for each registered monitor.  Cf. monitorEnumerate().
 */

void
registeredEnumerate(void (*fcn)(reg_mon_t *, void *), void *cookie)
{
    reg_mon_t *reg;

    sysAssert(REGISTRY_LOCKED(sysThreadSelf()));

    reg = MonitorRegistry;
    while (reg) {
	/* fcn may do arbitrary things to reg, like free it */
	reg_mon_t * next = reg->next;
	fcn(reg, cookie);
	reg = next;
    }
}

/* This function is called when the HPI detects there is a contention
 * on the raw monitor. We do not acquire the Registry lock here to avoid
 * infinite recursion and deadlocks. It is safe to walk the registered
 * monitor list because we never remove any entries.
 */
char *
lookupRegisteredMonitor(sys_thread_t *self, sys_mon_t *mid)
{
    reg_mon_t *mreg;

    mreg = MonitorRegistry;
    while (mreg) {
        if (mreg->mid == mid) {
	    if (mreg->name[0] != '_') {
	        return mreg->name;
	    }
	    return NULL;
	}
        mreg = mreg->next;
    }
    return NULL;
}
