/*
 * @(#)monitor_cache.c	1.78 98/08/14
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#ifndef FAST_MONITOR

/*
 * Code to manage the monitor cache.
 *
 * The monitor cache is a hash table which contains monitors.  The
 * hash function hashes an arbitrary value (typically the address of
 * an object) to a single monitor structure.  The monitor cache and
 * its hash table grow as necessary, and are only limited by available
 * memory.  The hash table is grown when the total number of monitors
 * is HASHTABLE_MULTIPLE times the number of hash table buckets, i.e.
 * when the average hash chain length with all monitors in use would
 * be HASHTABLE_MULTIPLE.
 *
 * Once the cache has expanded, it is never contracted, although doing
 * so would be possible.
 */

#include "threads.h"
#include "monitor.h"
#include "monitor_cache.h"
#include "log.h"
#include "common_exceptions.h"

#include "sys_api.h"

sys_mon_t *_cache_lock;

/*
 * These macros are use to denature the "next" pointer in the monitor_t
 * structure to mark monitors which might be in use and shouldn't
 * be scavenged.  This is only used inside a small section of
 * monitorCacheScavenge and zapLocalCacheHelper.
 */

#define MARK_MON_LOCAL_REF(m) \
	(m) = (monitor_t *) (((int) m) | MON_LOCAL_CACHE_REF)
#define CLEAR_MON_LOCAL_REF(m) \
        (m) = (monitor_t *) (((int) m) & ~ MON_LOCAL_CACHE_REF)
#define CHECK_MON_LOCAL_REF(m) (((int) m) & MON_LOCAL_CACHE_REF)

/*
 * The following defines should preserve for initialization to work
 *	INIT_MONITOR_COUNT <= EXPANSION_QUANTUM * INIT_MONITOR_COUNT
 */
#define INIT_HASHTABLE_BUCKETS	128 /* Must be a power of two! */
#define INIT_MONITOR_COUNT	128 /* Initial number of monitors */
#define EXPANSION_QUANTUM	16 /* Number of monitors to expand by */
#define HASHTABLE_MULTIPLE	2  /* Length of avg hash chain before expand */
#define HASH_MASK		(monHashTableBuckets - 1)
#define HASH(k)			(((k >> 2) ^ (k >> 10)) & HASH_MASK)

/*
 * When the number of free monitors falls below FREE_MONITOR_THRESHOLD we
 * expand.  This is needed because some platforms need monitors in the
 * expansion code (e.g. Green threads on Solaris does a malloc which
 * may use at least two monitors).  Because the monitor cache is in
 * shared code, this number has to be conservative.
 */
#define FREE_MONITOR_THRESHOLD 5

/* The local monitor hash table must be a power of two words long */
#define LOCAL_HASH_MASK         (TLS_MONCACHE-1)
#define LOCAL_HASH(k)           ((k >> 3) & LOCAL_HASH_MASK)

#ifdef STATISTICS
#define COUNT_CACHE_HIT()	(localCacheHits += 1)
#define COUNT_CACHE_MISS() (localCacheMisses += 1)
#else
#define COUNT_CACHE_HIT() ((void) 0)
#define COUNT_CACHE_MISS() ((void) 0)
#endif

/*
 * This should probably just be defined as a function but it seems that
 * most of the x86 compilers aren't very agressive inliners and I'd
 * really like this to be inlined.
 */

#define checkCache(mon, ee, key) \
{ \
    monitor_t * _mon; \
    /* \
     * Stash the key currently being looked up; the monitor in the \
     * global cache for it, even if otherwise unused, is never collected. \
     */ \
    currentCacheKey(ee) = key; \
    if (systemIsMP) { \
	sysMemoryFlush(); \
    } \
    _mon = localMonCache(ee, LOCAL_HASH(key)); \
    if ((_mon != 0) && (_mon->key == key)) { \
	COUNT_CACHE_HIT(); \
        mon = _mon; \
    } else { \
	COUNT_CACHE_MISS(); \
	mon = NULL; \
    } \
}

/*
 * The hash table itself and its size.  We can't dynamically allocate the
 * initial hash table because we would need monitors to do so.  If the
 * table is eventually expanded, we just lose the initial memory.
 */
static monitor_t *initMonHashTable[INIT_HASHTABLE_BUCKETS] = {0};
monitor_t **monHashTable = &initMonHashTable[0];
int monHashTableBuckets = INIT_HASHTABLE_BUCKETS;

/*
 * The total number of monitors in the monitor cache
 */
int monCount = 0;

/*
 * The monitor cache free list and count of free monitors.  We use
 * the count to determine when we have fewer than FREE_MONITOR_THRESHOLD
 * monitors free, in which case we expand the cache.  We can't wait
 * for the free list to be empty before we expand, as some platforms
 * need monitors to expand (e.g. malloc on Solaris may use as many
 * as two monitors).  We need to overestimate here because this code
 * and FREE_MONITOR_THRESHOLD are shared across all platforms. 
 */
static monitor_t *monFreeList = 0;
static int monFreeCount = 0;
int systemIsMP = 0;

/*
 * This number represents the thread count when the
 * zapLocalCacheHelper happens. 
 * It does not need to be exactly correct.
 * It is only used as a reference for cache expansion.
 */
static int monThreadCount = 0;
#define THREAD_FACTOR (monThreadCount * 5)

typedef struct monitor_chunk_t {
  int size;
  struct monitor_chunk_t *next;
  monitor_t monitors[1];
} monitor_chunk_t;

static monitor_chunk_t *monitor_chunks;

#ifdef STATISTICS
static int monLookups = 0;
static int monHashFind = 0;
static int monCreate = 0;
static int localCacheHits = 0;
static int localCacheMisses = 0;

static void
dumpstats()
{
    jio_fprintf(stderr, "monLookups = %d\n", monLookups);
    jio_fprintf(stderr, "monHashFind = %d\n", monHashFind);
    jio_fprintf(stderr, "monCreate = %d\n", monCreate);
    jio_fprintf(stderr, "localCacheHits = %d\n", localCacheHits);
    jio_fprintf(stderr, "localCacheMisses = %d\n", localCacheMisses);
}
#endif

static monitor_t *findMonitor(ExecEnv *ee, uintptr_t, bool_t create);

/*
 * Must be called for each cached monitor prior to use.
 */
void 
monitorInit(monitor_t *mon) 
{
    sysAssert(mon != MID_NULL);
    sysMonitorInit(sysmon(mon));
}

/*
 * Must be called for each cached monitor once it will never again be
 * used.  Note that cached monitors are never monitorDestroy'd until
 * Exit, and doing so would be disasterous on some platforms.  DO NOT
 * CONFUSE THIS WITH monitorExit!  This cleans up resources on runtime
 * exit and is the converse of monitorInit.
 */
void 
monitorDestroy(monitor_t *mon) 
{
    sysAssert(mon != MID_NULL);
    sysMonitorDestroy(sysmon(mon));
}

/*
 * Scan for unused cached monitors and free them to the cache.  Return
 * the number freed.
 *
 * All monitors in the hash table should be considered, as they are
 * nominally in use.  A monitor is actually unused if its entry_count
 * is 0 and nobody is on any of its wait queues.  When these conditions
 * are true when a thread leaves a monitor (see sysMonitorExit())
 * the monitor's flags and monitor_owner field are also reset, but those
 * are not the determiners.
 *
 * We don't reclaim memory or other resources for monitors here.  The
 * monitor is simply reinitialized and put into a free list for reuse.
 * As we never allocate more monitors than we actually need (modulo some
 * padding), eagerly freeing monitors causes unnecessary overhead.  If we
 * grew that big once we are likely to do so again.
 */

static int
zapLocalCacheHelper(sys_thread_t *tid, void *ignored)
{
    uintptr_t key;
    int i;
    monitor_t ** ptr;
    ExecEnv *ee = SysThread2EE(tid);

    ++monThreadCount;

    /* For now, smash the whole cache; could preserve entry we're looking up */
    ptr = (monitor_t **) &localMonCache(ee, 0);
    for (i = TLS_MONCACHE; i > 0; i--) {
	*ptr++ = NULL;
    }
    if (systemIsMP) {
	sysMemoryFlush();
    }
    if ((key = currentCacheKey(ee)) != 0) {
	/*
	  * we already have cache_lock so just look it up.
	  */
	int hash = HASH(key);
	monitor_t *mon = monHashTable[hash];
	while (mon != 0) {
	    if (mon->key == key) {
		MARK_MON_LOCAL_REF(mon->next);
		return SYS_OK;
	    } else
		mon = mon->next;
		CLEAR_MON_LOCAL_REF(mon);
	}
	/*
	  * if we get here, then they are looking up a monitor which doesn't
	  * exist just yet., which is ok.
	  */
    }
    return SYS_OK;
}

static int
monitorCacheScavenge(void)
{
    int i;
    int count = 0;
    monitor_t **monp;
#ifdef DEBUG
    long start_time = sysGetMilliTicks();
    long end_time;
    int monCountLocal;
    int monFreeCountLocal;
#endif

    /*
     * If the _queue_lock hasn't been created yet, then threads
     * haven't been properly booted.  This means that the scheme
     * which is used to keep monitors alive until they are entered
     * won't work, so don't allow scavenging until threads are
     * booted.
     */
    if (_queue_lock == NULL)
	return 0;

    /*
     * Reset monThreadCount. After the following local monitor
     * cache scanvenge, it will record the current thread count.
     */
    monThreadCount = 0;

    /*
     * Clobber the thread-local monitor caches; we could be smarter
     * about doing this, but it would be more costly.  We do this
     * before cleaning up the global cache to avoid races from values
     * still in local caches.
     */
    sysThreadEnumerateOver(zapLocalCacheHelper, (void *) 0);

    for (i = monHashTableBuckets; --i >= 0; ) {
	for (monp = &monHashTable[i]; *monp != 0; ) {
	    monitor_t *mon = *monp;
	    if (!(CHECK_MON_LOCAL_REF(mon->next)) &&
		!sysMonitorInUse(sysmon(mon))) {
		sysAssert(sysMonitorOwner(sysmon(mon)) == NULL);

		/* Unlink the monitor from the hash chain */
		*monp = mon->next;
		/*
		 * Reinitialize monitor and put it on the free list.
		 * sys_mon_t reinitialization is done on sysMonitorExit.
		 */
		mon->key = 0;
		mon->next = monFreeList;
		monFreeList = mon;
		monFreeCount++;
		count++;
	    } else {			/* In use; skip over this one */
		if (CHECK_MON_LOCAL_REF(mon->next)) {
		    CLEAR_MON_LOCAL_REF(mon->next);
		}
		monp = &((*monp)->next);
	    }
	}
    }

#ifdef DEBUG
    monCountLocal = monCount;
    monFreeCountLocal = monFreeCount;
    end_time = sysGetMilliTicks();
    if (verbosegc) {
	jio_fprintf(stderr,
    "<Monitor cache GC freed %d of %d monitors in %ld ms (%d total free)>\n",
		count, monCountLocal, end_time-start_time, monFreeCountLocal);
    }
#endif

    return count;
}

/*
 * Add count monitors to the front of the monitor free list.  This
 * doesn't assume that the monFreeList is empty, and in fact it
 * probably isn't (see comment below when monitorCacheExpand() is
 * called from createMonitor()).  This is called with the monitor
 * cache *unlocked*.
 *
 * There is a crucial and nonobvious connection between the delayed
 * increment of monFreeCount in monitorCacheExpand() and the threshold
 * test in createMonitor().  Once we have started to expand, we want
 * to to get other threads wanting monitors into the createMonitor()
 * code that causes them to wait until expansion finishes.  This
 * avoids nasty problems like grabbing monitors in the unprotected
 * region of createMonitor() between adding new monitors to the
 * cache and expanding the hash table.  By postponing resetting
 * monFreeCount until after all expansion is done, we guarantee that
 * all threads creating monitors will be caught by the single, fast
 * threshold test.  Once inside, further tests decide what to do
 * with them, but the crucial normal case (no expansion) remains fast.
 */
void
monitorCacheExpand(int count)
{
    size_t nbytes = count * monitorSizeof();
    monitor_t *mon;
    monitor_chunk_t *new_chunk;
    int i;
    sys_thread_t *self = sysThreadSelf();

    while (count > 0) {
        new_chunk = sysMalloc(nbytes + offsetof(monitor_chunk_t, monitors));
	if (new_chunk) {
	    break;
	}
	/* we couldn't allocate count monitors, so let's try one less */
	count--;
	nbytes = count * monitorSizeof();
    }
    if (count == 0) {
	return;
    }

    Log2(2, "Expanding monitor cache by %d monitors to %d\n",
	 count, monCount+count);

    mon = new_chunk->monitors;
    memset((char *) mon, 0, nbytes);
    /* monFreeCount += count; Postponed!  See comment above */
    i = count;
    while (--i > 0) {
	monitor_t *next = (monitor_t *) ((char *)mon + (monitorSizeof()));
	monitorInit(mon);
	mon->next = next;
	mon = next;
    }
    monitorInit(mon);

    CACHE_LOCK(self);
    monCount += count;
    mon->next = monFreeList;
    monFreeList = new_chunk->monitors;

    new_chunk->size = count;
    new_chunk->next = monitor_chunks;
    monitor_chunks = new_chunk;

    /*
     * If after expansion the monCount is greater than or equal to 
     * HASHTABLE_MULTIPLE times the hash table size, expand the hash table
     * as well.  That is to say that we expand when the average hash chain
     * with all monitors in use is HASHTABLE_MULTIPLE long.
     */
    if (monCount < HASHTABLE_MULTIPLE * monHashTableBuckets) {
	monFreeCount += count; 		/* Increment when done expanding */
	CACHE_UNLOCK(self);
    } else {
	/* Expand the hash table: size must be a multiple of two */
	int hash, i;
	int oldMonHashTableBuckets = monHashTableBuckets;
	int newMonHashTableBuckets = monHashTableBuckets;
	monitor_t **oldMonHashTable = monHashTable;
	monitor_t **newMonHashTable;
	monitor_t **hashp;

	do {
	    newMonHashTableBuckets *= 2;
	} while (monCount >= HASHTABLE_MULTIPLE * newMonHashTableBuckets);

	Log2(2, "Expanding monitor cache hash table from %d to %d buckets\n",
	     monHashTableBuckets, newMonHashTableBuckets);
	nbytes = newMonHashTableBuckets * sizeof(monitor_t *);
	/*
	 * Remember that malloc may use a monitor, so needs to have
	 * a consistent monHashTable and monHashTableBuckets.  After
	 * we reset monHashTable and monHashTableBuckets we need to
	 * be sure not to enter any cache monitors before we've
	 * rehashed everything.
	 */
        CACHE_UNLOCK(self);
	if ((newMonHashTable = (monitor_t **) sysMalloc(nbytes)) == 0) {
	    /*
	     * We didn't get what we had hoped for, but we should be able
	     * to continue using the old table albeit with slower cache 
	     * access.  We'll try to expand the hash table again next time
	     * we allocate another block of monitors, if we haven't died
	     * by running out of malloc space yet.
	     */
	    CACHE_LOCK(self);
	    monFreeCount += count;	/* Increment when done expanding */
	    CACHE_UNLOCK(self);
	    return;
	}
	CACHE_LOCK(self);
	monHashTable = newMonHashTable;
	monHashTableBuckets = newMonHashTableBuckets;
	memset((char *) monHashTable, 0, nbytes);
	for (i = 0; i < oldMonHashTableBuckets; i++) {
	    mon = oldMonHashTable[i];
	    while (mon != 0) {
		monitor_t *next = mon->next;
		hash = HASH(mon->key);	/* must have new size set here */
		hashp = &monHashTable[hash];
		mon->next = *hashp;
		*hashp = mon;
		mon = next;
	    }
	}
	monFreeCount += count; 		/* Increment when done expanding */
	CACHE_UNLOCK(self);
        if (oldMonHashTable != &initMonHashTable[0]) {
            sysFree(oldMonHashTable);
        }
    }
}

/*
 * Expand the monitor cache if someone else isn't already doing so.
 * Return TRUE if the caller should immediately try to reallocate the
 * monitor, and FALSE otherwise.  The latter case can occur when
 * recursively expanding the cache, in which case we want to keep
 * working out of the FREE_MONITOR_THRESHOLD buffer of monitors.
 */
static sys_thread_t *expandingFreeList = 0;

static bool_t
monitorCacheMaybeExpand(ExecEnv * ee)
{
    bool_t tryAgain = FALSE;
    sys_thread_t *self = EE2SysThread(ee);

    sysAssert(CACHE_LOCKED(self));
    CACHE_UNLOCK(self);

    /*
     * During bootstrap we might not have a queue_lock yet.  In this
     * case we can't do any scavenging so we just need to expand the
     * cache.  This shouldn't happen in the normal case but it depends
     * on what the starting size of the monitor_cache is.
     */
    if (_queue_lock == NULL) {
	monitorCacheExpand(EXPANSION_QUANTUM);
	CACHE_LOCK(self);
	return TRUE;
    }


    /* CACHE_LOCK is now a regular lock, just like QUEUE_LOCK.
     */
    /* The lock usage here is a little bizarre because on Green
     * Threads the CACHE_LOCK is the scheduler lock.  You can't
     * acquire new locks when holding the scheduler lock and you can't
     * wait on it.  This means that we can't use the simple approach
     * of locking the CACHE_LOCK and simply doing whatever we need to
     * do.  Instead we're going to use QUEUE_LOCK for that.  We need
     * to hold the QUEUE_LOCK for the call to monitorCacheScavenge
     * since it calls sysThreadEnumerateOver which requires this lock
     * to be held.  We still need to lock CACHE_LOCK to protect the
     * monitor cache itself.  Other threads that want to expand the
     * monitor cache end up waiting on the QUEUE_LOCK.  Once an
     * expansion happens, the waiters are woken up with a notifyAll on
     * the QUEUE_LOCK.  This is less than ideal but is better than the
     * alternatives available right now.
     */

    QUEUE_LOCK(self);

    CACHE_LOCK(self);
    /* Redundant threshold test in case this thread blocked and some else GC */
    if (!expandingFreeList && monFreeCount < FREE_MONITOR_THRESHOLD) {
	monitorCacheScavenge();
	if (monFreeCount < FREE_MONITOR_THRESHOLD + THREAD_FACTOR) {
	    CACHE_UNLOCK(self);
	    expandingFreeList = sysThreadSelf();
	    QUEUE_UNLOCK(self);

	    /*
	     * We can't hold CACHE_LOCK when calling this because of Green
	     * threads.  monitorCacheExpand ends up calling malloc which
	     * will drop back into the monitor code to lock the heap.
	     */ 
	    monitorCacheExpand(EXPANSION_QUANTUM + THREAD_FACTOR);

	    /*
	     * We're done so notify any threads which are waiting for
	     * us to finish expanding the monitor cache.
	     */
	    QUEUE_LOCK(self);
	    expandingFreeList = 0;
	    sysMonitorNotifyAll(self, _queue_lock);
	    QUEUE_UNLOCK(self);
	    tryAgain = TRUE;
	} else {
	    /*
	     * We've scavenged enough monitors to keep on going so
	     * return, telling our caller to try again.
	     */
	    CACHE_UNLOCK(self);
	    QUEUE_UNLOCK(self);
	    tryAgain = TRUE;
	}
    } else if (expandingFreeList != sysThreadSelf()) {
	bool_t wasInterrupted = FALSE;

	CACHE_UNLOCK(self);
	/* Somebody is already expanding the free list, so wait */
	while (expandingFreeList) {
	    if (QUEUE_WAIT(self) == SYS_INTRPT) {
		/*
		 * if we get interrupted during this wait, we record
		 * the fact that we were interrupted and clear the
		 * interrupt.  Then once we've successfully waited, we
		 * restore the interrupt.
		 */
		if (sysThreadIsInterrupted(self, TRUE))
		    wasInterrupted = TRUE;
	    }
	}

	if (wasInterrupted)
	    sysThreadInterrupt(self);

	/*
	 * Someone might have tried to create a monitor but was
	 * forced to wait because someone else was expanding the
	 * cache.  During that time the expander might have created
	 * a monitor for the waiter's key, so when the waiter wakes
	 * up he needs to start the allocation from scratch.
	 */
	QUEUE_UNLOCK(self);
	tryAgain =  TRUE;
    } else {
	CACHE_UNLOCK(self);
	QUEUE_UNLOCK(self);
	/*
         * Else *I* am expanding the list, so let me continue using
         * monitors out of the FREE_MONITOR_THRESHOLD buffer.
         */
    }

    CACHE_LOCK(self);
    return tryAgain;
}


/* Initialize the monitor cache
 * 
 * Called during bootstrapping, so should be no chance of interruption.
 *
 * We need a sys_-level monitor initialization step.  We don't need
 * a sys_-level monitor destruction, as we never really destroy a
 * cache monitor -- they are just left in the table to be used again.
 */

bool_t  monitorsInitialized = FALSE;

void
monitorCacheInit(void)
{
    CACHE_LOCK_INIT();		/* The monitor cache lock is special */

    monitorCacheExpand(INIT_MONITOR_COUNT);
    monitorsInitialized = TRUE;
    systemIsMP = sysGetSysInfo()->isMP;

#ifdef STATISTICS
    AtExit(dumpstats);
#endif
}

/*
 * Destroy the current monitor cache.  This is only done at Exit, and
 * must not be done earlier; Java code cannot be run safely after the
 * monitor cache has been destroyed.  Doing this explicitly makes gives
 * tools like Purify a clearer view of what memory and system resources
 * are being leaked as opposed to being implicitly cleaned up on exit.
 * 
 * We currently do NOT free memory for the platform-independent portion
 * of the cache on Exit.
 */
void
monitorCacheDestroy(void)
{
    int i;
    monitor_t *mon;
    sys_thread_t *self = sysThreadSelf();

    CACHE_LOCK(self);	/* Just in case... */

    monitorsInitialized = FALSE;
    for (i = 0; i < monHashTableBuckets; i++) {
	mon = monHashTable[i];
	if (mon != 0) {
	    while (mon != 0) {
		monitorDestroy(mon);
	        mon = mon->next;
	    }
	}
    }

    CACHE_UNLOCK(self);
}

/*
 * Look up a monitor in the hash table.  Return MID_NULL if it is not
 * there.  Assume that the monitor cache lock is not held, except on green
 *
 * REMIND: Can there be a race here between local cache lookup and
 * cache GC?  This would be true if we ever did a lookupMonitor on a
 * monitor that isn't known to be in use.  That is not normally done
 * except by error, but it can happen at least in sysThreadExit where
 * the thread's monitor is fetched if it's in the cache.  Do we need
 * a non-cached lookup for cases like this, to avoid overhead in
 * cases where we will use the cache and don't want to worry?  Other-
 * wise, we would need to play a use_count game like on monitor creation.
 */

monitor_t *
lookupMonitor(sys_thread_t *self, uintptr_t key, bool_t create)
{
    monitor_t *mon;
    ExecEnv *ee = SysThread2EE(self);
    
    checkCache(mon, ee, key);
    if (mon)
	return mon;

    return findMonitor(ee, key, create);
}

/*
 * Get a monitor for the specified key.  If it already exists,
 * just return it.  If create is non-zero then create a new monitor.
 * Otherwise return MID_NULL.
 */
static monitor_t *
findMonitor(ExecEnv *ee, uintptr_t key, bool_t create)
{
    int	hash;
    monitor_t **hashp, *mon;
    sys_thread_t *self = sysThreadSelf();

    CACHE_LOCK(self);

#ifdef STATISTICS
    monLookups += 1;
#endif

RetryCreate:
    /* First look for the monitor, in case it's in use */
    hash = HASH(key);
    mon = monHashTable[hash];
    while (mon != 0) {
	if (mon->key == key) {
	    /* Should we consider this the normal case and inline unlock? */
	    goto unlock;
	}
	mon = mon->next;
    }

    if (!create) {
	CACHE_UNLOCK(self);
	return MID_NULL;
    }

    /*
     * If we get here, a monitor for key has not been allocated.  Now
     * we grab one from the free list, and add it to the hash table.  If
     * handing over the next monitor pushes us below FREE_MONITOR_THRESHOLD
     * free monitors, expand the monitor cache *first*.  On systems that
     * don't use native threads, like on Solaris with Green threads, 
     * sysMalloc() is probably made thread safe using a monitor from the
     * monitor cache, and in order to expand the free list, we need to
     * call sysMalloc().  So we make sure we have some small number of
     * monitors available when we expand.  We also have to play some
     * tricks to avoid deadlock between the malloc and monitor cache
     * monitors.
     */
    if (monFreeCount < FREE_MONITOR_THRESHOLD) {
        if (monitorCacheMaybeExpand(ee) == TRUE) {
    	    goto RetryCreate;
        } else {
	    if (monFreeCount == 0) {
		/*
		 * we were unable to allocate any new monitors and we
		 * have none left.
		 */
		mon = NULL;
		goto unlock;
	    }
	    /* Otherwise, allocate out of FREE_MONITOR_THRESHOLD */
	}
    }

    sysAssert(monFreeList != 0);
    /* We don't want to recreate anything that was created by expansion: */
    /* sysAssert(findMonitor(key) == MID_NULL); */

#ifdef STATISTICS
    monCreate += 1;
#endif
    mon = monFreeList;
    monFreeList = mon->next;
    monFreeCount--;
    sysAssert(!sysMonitorInUse(sysmon(mon)));

    mon->key = key;
    hashp = &monHashTable[hash];
    mon->next = *hashp;
    *hashp = mon;

unlock:
    localMonCache(ee, LOCAL_HASH(key)) = mon;
    CACHE_UNLOCK(self);
    return mon;
}

/*-
 * Look up a monitor in the monitor cache and enter it using
 * monitorEnter().  If there is no monitor cached for the given key
 * create one on the fly.
 */
sys_mon_t *
monitorEnter2(struct execenv *ee, uintptr_t key)
{
    monitor_t *mon;
    sys_thread_t * self = EE2SysThread(ee);

    checkCache(mon, ee, key);
    if (mon == MID_NULL)
	mon = findMonitor(ee, key, TRUE);
    if (mon != MID_NULL) {
	sysMonitorEnter(self, sysmon(mon));

	currentCacheKey(ee) = 0;
	return sysmon(mon);
    } else {
	/* this should only happen under low memory conditions */
	ThrowOutOfMemoryError(ee, "Couldn't allocate memory for monitor");
	return NULL;
    }
}

int
monitorExit2(struct execenv *ee, uintptr_t key)
{
    monitor_t * mon;
    sys_thread_t * self = EE2SysThread(ee);

    checkCache(mon, ee, key);
    if (mon == MID_NULL)
	mon = findMonitor(ee, key, FALSE);
    if (mon == MID_NULL || sysMonitorExit(self, sysmon(mon)) != SYS_OK) {
	ThrowIllegalMonitorStateException(ee, "current thread not owner");
	return -1;
    }

    currentCacheKey(ee) = 0;
    return 0;
}

/*-
 * Wake up the first thread waiting on the condition variable
 * associated with this monitor.
 */
void
monitorNotify2(struct execenv *ee, uintptr_t key)
{
    monitor_t * mon;
    sys_thread_t * self = EE2SysThread(ee);

    checkCache(mon, ee, key);
    if (mon == MID_NULL)
	mon = findMonitor(ee, key, FALSE);
    if (mon == MID_NULL || sysMonitorNotify(self, sysmon(mon)) != SYS_OK) {
	ThrowIllegalMonitorStateException(ee, "current thread not owner");
	return;
    }
    currentCacheKey(ee) = 0;
}

/*-
 * Wake up *all* the threads waiting on the condition variable
 * associated with this monitor.
 */
void
monitorNotifyAll2(struct execenv *ee, uintptr_t key)
{
    monitor_t * mon;
    sys_thread_t * self = EE2SysThread(ee);

    checkCache(mon, ee, key);
    if (mon == MID_NULL)
	mon = findMonitor(ee, key, FALSE);
    if (mon == MID_NULL || sysMonitorNotifyAll(self, sysmon(mon)) != SYS_OK) {
	ThrowIllegalMonitorStateException(ee, "current thread not owner");
	return;
    }
    currentCacheKey(ee) = 0;
}

void
monitorWait2(struct execenv *ee, uintptr_t key, int64_t millis)
{
    int ret;
    sys_thread_t * self = EE2SysThread(ee);
    monitor_t * mon;

    checkCache(mon, ee, key);
    if (mon == MID_NULL)
	mon = findMonitor(ee, key, FALSE);

    if (mon == MID_NULL) {
	ret = SYS_ERR;
    } else {
	ret = sysMonitorWait(self, sysmon(mon), millis);
	if (ret == SYS_TIMEOUT) {
	    ret = SYS_OK;
	}
    }
    currentCacheKey(ee) = 0;

    /* An asynchronous exception might have been thrown while we
     * wait on the monitor.
     * REMIND: 
     * This still does not fix the inherent race condition
     * where an async exception can be thrown after this check
     * is done. However, this race condition exists everywhere
     * else in the VM and should be fixed thoroughly in a future
     * release. 
     */
    if (exceptionOccurred(ee))
        return;

    if (ret == SYS_INTRPT) {
	ThrowInterruptedException(ee, "operation interrupted");
	return;
    }

    if (ret != SYS_OK) {
	ThrowIllegalMonitorStateException(ee, "current thread not owner");
	return;
    }
}


/*
 * monitorEnumerate() should be called with the cache lock held.
 */
void
monitorEnumerate(void (*fcn)(monitor_t *, void *), void *cookie)
{
    monitor_t **hashtable = monHashTable;
    monitor_t *mon, *next;
    int	i;

    sysAssert(CACHE_LOCKED(sysThreadSelf()));
    for (i = monHashTableBuckets; --i >= 0; ) {
	mon = hashtable[i];
	while (mon != 0) {
	    next = mon->next;
	    fcn(mon, cookie);
	    mon = next;
	}
    }
}

/*
 * Do not try to acquire the CACHE lock because that would lead to
 * infinite recursion. It is safe to traverse the monitor chunks 
 * outside of the CACHE lock, because we currently do not free 
 * monitor chunks.
 */
void *
lookupJavaMonitor(sys_mon_t *mid)
{
    extern bool_t isHandle(void *p); /* defined in gc.c */
    monitor_chunk_t *chunk = monitor_chunks;
    while (chunk) {
	monitor_t *mon = chunk->monitors;
	if ((char *)mid > (char *)mon && 
	    (char *)mid < (char *)mon + chunk->size * monitorSizeof()) {
	    mon = (monitor_t *)((char *)mid - sizeof(monitor_t));
	    if (isHandle((void *)mon->key)) {
	        return (void *)mon->key;
	    } else {
	        return NULL;
	    }
	}
	chunk = chunk->next;
    }
    return NULL;
}

#ifdef DEBUG

static void DumpMonitorCacheInfo()
{
    int i;
    monitor_t *mon;

    jio_fprintf(stderr, "Monitor cache info:\n");
    jio_fprintf(stderr, "  Initial monitor count: %d\n", INIT_MONITOR_COUNT);    
    jio_fprintf(stderr, "  Initial hash table buckets: %d\n",
	    INIT_HASHTABLE_BUCKETS);  
    jio_fprintf(stderr, "  Hash table expansion multiple: %d\n", HASHTABLE_MULTIPLE);
    jio_fprintf(stderr, "  Minimum number of free monitors before expansion: %d\n",
	    FREE_MONITOR_THRESHOLD);    
    jio_fprintf(stderr, "  Number of new monitors added per expansion: %d\n",
	    EXPANSION_QUANTUM);    
    jio_fprintf(stderr, "  Current total number of monitors: %d\n", monCount);
    jio_fprintf(stderr, "  Current number of free monitors: %d\n", monFreeCount);
    jio_fprintf(stderr, "  Current number of hash table buckets (power of 2!): %d\n",
	    monHashTableBuckets);

    jio_fprintf(stderr, "  Current hash table bucket dump:\n");
    for (i = 0; i < monHashTableBuckets; i++) {
	mon = monHashTable[i];
	if (mon != 0) {
	    jio_fprintf(stderr, "    Bucket %d: ", i);
	    while (mon != 0) {
	        jio_fprintf(stderr, "0x%p ", mon);
	        mon = mon->next;
	    }
	    jio_fprintf(stderr, "\n");
	}
    }
}

#endif

#endif /* #ifndef FAST_MONITOR */
