/*
 * @(#)reflect_util.c	1.39 98/09/23
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Support code for reflection API.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#define	_REFLECT_IMPL

#include "sys_api.h"				/* sysAssert() */
#include "signature.h"
#include "typecodes.h"
#include "wrap.h"
#include "reflect.h"
#include "common_exceptions.h"

/*
 * Find class object for signature -- including pseudo-class object
 * for primitive types.
 * Signals NoClassDefFoundError
 */
ClassClass *
reflect_find_class(
    char *sig,
    ClassClass *from,
    char **cpp)
{
    char buf[MAXPATHLEN+8], *p = sig, *q = buf, *r = (q + MAXPATHLEN), c;
    ClassClass *cb;

    switch (c = *p++) {
    case SIGNATURE_BOOLEAN:
	cb = class_boolean;	break;
    case SIGNATURE_BYTE:
	cb = class_byte;	break;
    case SIGNATURE_CHAR:
	cb = class_char;	break;
    case SIGNATURE_SHORT:
	cb = class_short;	break;
    case SIGNATURE_INT:
	cb = class_int;		break;
    case SIGNATURE_LONG:
	cb = class_long;	break;
    case SIGNATURE_FLOAT:
	cb = class_float;	break;
    case SIGNATURE_DOUBLE:
	cb = class_double;	break;
    case SIGNATURE_VOID:
	cb = class_void;	break;

    case SIGNATURE_CLASS:
	/*
	 * To convert to classname: exclude 'L', slashify, exclude ';'
	 * For example, "java/lang/Thread"
	 */
	while ((c = *p++) != SIGNATURE_ENDCLASS) {
	    *q++ = (c == '.') ? '/' : c;	/* slashify */
	    if (q >= r)
		goto toolong;
	}
	*q = '\0';
	cb = FindClassFromClass(0, buf, FALSE, from);	/* don't resolve */
	break;

    case SIGNATURE_ARRAY:
	/*
	 * To convert to classname: include '[', include 'L',
	 * slashify, include ';'
	 * For example, "[[Ljava/lang/Thread;" "[J"
	 */
	*q++ = c;				/* don't miss first '[' */
	while ((c = *p++) == SIGNATURE_ARRAY) {
	    *q++ = c;
	    if (q >= r)
		goto toolong;
	}
	*q++ = c;				/* component */
	if (c == SIGNATURE_CLASS) {		/* include 'L' */
	    while ((c = *p++) != SIGNATURE_ENDCLASS) {
		*q++ = (c == '.') ? '/' : c;	/* slashify */
		if (q >= r)
		    goto toolong;
	    }
	    *q++ = c;				/* include ';' */
	}
	*q = '\0';
	cb = FindClassFromClass(0, buf, FALSE, from);	/* don't resolve */
	break;

    default:
	ThrowNoClassDefFoundError(0, sig);	/* XXX severe error */
	cb = NULL;
    }

    if (cpp)					/* caller wants to skip */
	*cpp = p;
    return cb;

  toolong:
    ThrowInternalError(0, "signature overflow");
    return NULL;

}

/*
 * Check if field or method is accessible to client.  Throw an
 * IllegalAccessException and return FALSE if not.
 */
bool_t
reflect_check_access(
    ExecEnv *ee,
    ClassClass *field_class,			/* declaring class */
    int acc,					/* declared field access */
    ClassClass *target_class,			/* for protected */
    bool_t for_invoke)
{
    ClassClass *client_class = NULL;

    if (ee == NULL)
	ee = EE();

    /*
     * The "client" is the class associated with the nearest frame
     * that has a methodblock (i.e. is not a pseudo-frame).  In
     * particular, Method.invoke() frames must be pseudo-frames.
     * getCallerClass already skips Method.invoke frames, so pass 0
     * in that case.
     */
    client_class = getCallerClass(ee, (for_invoke) ? 0 : 1);

    if (client_class != field_class) {
	if (!VerifyClassAccess(client_class, field_class, FALSE))
	    goto bad;
	if (!VerifyFieldAccess(client_class, field_class, acc, FALSE))
	    goto bad;
    }

    /*
     * Additional test for protected members: JLS 6.6.2
     * This could go faster if we had package objects
     */
    if (IsProtected(acc))
	if (target_class != client_class)
	    if (!IsSameClassPackage(client_class, field_class))
		if (!is_subclass_of(target_class, client_class, ee))
		    goto bad;

    /* Passed all tests */
    return TRUE;

  bad:
    ThrowIllegalAccessException(ee, cbName(field_class)); /* exception */
    return FALSE;
}

/*
 * Array utilities
 */

HObject *			/* sigh */
reflect_new_array(
    ExecEnv *ee,
    ClassClass *cb,
    int cnt)
{
    HArrayOfObject *harr;
    int typecode;

    sysAssert(cb != NULL);
    sysAssert(cnt >= 0);

    typecode = cbIsPrimitive(cb) ? cbTypeCode(cb) : T_CLASS;

    if ((harr = (HArrayOfObject *) allocArray(ee, typecode, cnt)) == NULL) {
	ThrowOutOfMemoryError(ee, 0);
	return NULL;
    }
    if (typecode == T_CLASS)
	unhand(harr)->body[cnt] = (HObject *)cb;
    return (HObject *)harr;
}

HArrayOfObject *
reflect_new_class_array(ExecEnv *ee, int cnt)
{
    /* TODO: Special-case 0-length class arrays */
    return (HArrayOfObject *)reflect_new_array(ee, classJavaLangClass, cnt);
}


/*  A minor variant of array_type_assignable_to_array_type in
 *  interpreter.c. Since the type information associated with an
 *  array instance is different from that associated with the class.
 *  The difference arises in determining the array depth. Here,
 *  the depth is passed in explicitly. 
 *  Note. The elt/item distinction should be changed to the component/elt
 *  distinction in Ch 10 of the JLS.
 */
bool_t
reflect_array_type_assignable_to_array_type(ClassClass *from_item_class,
					    int from_depth,
					    int from_item_type,
					    ClassClass *to_class, ExecEnv *ee)
{
    int to_depth = cbConstantPool(to_class)[CONSTANT_POOL_ARRAY_DEPTH_INDEX].i;
    int to_item_type =  
	cbConstantPool(to_class)[CONSTANT_POOL_ARRAY_TYPE_INDEX].i; 
    ClassClass *to_item_class = (to_item_type == T_CLASS)
	? cbConstantPool(to_class)[CONSTANT_POOL_ARRAY_CLASS_INDEX].clazz 
	: NULL;	

    if (from_depth > to_depth) { /*=> Match only if 'to' items are 'special'*/
	return (to_item_class == classJavaLangObject) || 
	    (to_item_class == interfaceJavaLangCloneable) ||
	    (to_item_class == interfaceJavaIoSerializable);
    } else if (from_depth == to_depth) { 
	/* If the elements are primitive, sufficient if the item type matches
	 * If not, if the from_item_class <= to_item_class */
	return  (to_item_type == from_item_type) 
	       && ((to_item_type != T_CLASS) 
		   || (to_item_class == from_item_class)
		   || is_subclass_of(from_item_class, to_item_class, ee));
    } else { 
	return FALSE;
    } 
}


bool_t
reflect_is_assignable(ClassClass *fromcb, ClassClass *tocb, ExecEnv *ee)
{
    if (cbIsPrimitive(fromcb) || cbIsPrimitive(tocb))
	return fromcb == tocb;

    if (cbName(fromcb)[0] == SIGNATURE_ARRAY) {
	union cp_item_type *fromcb_pool = cbConstantPool(fromcb);
	int fromtype = fromcb_pool[CONSTANT_POOL_ARRAY_TYPE_INDEX].i;
	ClassClass* fromclass = (fromtype == T_CLASS)
	    ? fromcb_pool[CONSTANT_POOL_ARRAY_CLASS_INDEX].clazz
	    : NULL;
	int fromdepth = fromcb_pool[CONSTANT_POOL_ARRAY_DEPTH_INDEX].i;
	if (cbName(tocb)[0] == SIGNATURE_ARRAY) {
	    return (reflect_array_type_assignable_to_array_type(fromclass, 
								fromdepth,
								fromtype,
								tocb, ee));
	} else {
	    return ((tocb == classJavaLangObject)
		    ||(tocb == interfaceJavaLangCloneable)
		    ||(tocb == interfaceJavaIoSerializable));
	}
    }
    return is_subclass_of(fromcb, tocb, ee);
}

