/*
 * @(#)threads.c	1.89 98/09/02
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include <sys/types.h>
#include <string.h>

#include "threads.h"
#include "log.h"
#include "monitor.h"
#include "monitor_cache.h"
#include "sys_api.h"
#include "gc.h"
#include "common_exceptions.h"

#include "java_lang_Thread.h"
#include "java_lang_Throwable.h"

#ifdef HPROF
#include "vmprofiler.h"
#endif /* HPROF */

#ifdef BREAKPTS
#include "breakpoints.h"
#endif /* BREAKPTS */

static char * nameOfThread(TID tid, char * buf, int length);

sys_mon_t *_queue_lock;	/* Protects thread queue, thread count */

long JavaStackSize = JAVASTACKSIZE;

HThreadGroup *maingroup = NULL;
HThreadGroup *systemgroup = NULL;

HThread *
InitializeClassThread(ExecEnv *ee)
{
    HThread *main;
    ClassClass *cb;

    if (!classJavaLangThread) {
	return 0;
    }

    cb = classJavaLangThread;

    /*
     * We want to bootstrap the currently executing context into the
     * threads package.  Instantiate Thread, and hand to the thread's
     * package.
     */
    main = (HThread *) allocObject(ee, cb);
    if (main == 0) {
	return 0;
    }

    /* Find the ThreadGroup class */
    cb = FindStickySystemClass(ee, JAVAPKG "ThreadGroup", TRUE);
    if (!cb) {
	return 0;
    }

    /* Allocate the system thread group */
    systemgroup = 
	(HThreadGroup *)execute_java_constructor(PRIVILEGED_EE, 0, cb, "()");
    if (systemgroup == 0) {
	return 0;
    }

    /* Allocate the main thread group */
    maingroup =
        (HThreadGroup *)execute_java_constructor(0, JAVAPKG "ThreadGroup", 0,
                                "(Ljava/lang/ThreadGroup;Ljava/lang/String;)",
                                systemgroup, makeJavaString("main", 4));
    if (maingroup == 0) {
	return 0;
    }

    unhand(main)->eetop = ptr2ll(ee);
    ee->thread = (JHandle *)main;
    THREAD(main)->priority = NormalPriority;
    threadSetPriority(main, NormalPriority);

    /* Now call the constructor of the main thread.
     * This will add the current thread to the main group
     */
    execute_java_dynamic_method(0, (void *)main, "init",
        "(Ljava/lang/ThreadGroup;Ljava/lang/Runnable;Ljava/lang/String;)V",
                                maingroup, /* ThreadGroup */
                                NULL,      /* runnable */
                                makeJavaString("main", 4));

    return main;
}

void threadInit(void)
{
    ExecEnv *ee = EE();
    JHandle *p = ee->thread;

#ifdef HPROF
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_THREAD_START))
        jvmpi_thread_start(p);
#endif /* HPROF */

#ifdef BREAKPTS
    if (debugging) 
        notify_debugger_of_thread_start(ee, p);
#endif /* BREAKPTS */
}

/*
 * The following are defined as macros in threads.h:
 *
 * void threadYield(void)
 */

static void
saveStackBase(void *args)
{
    ExecEnv *ee = EE();
    void (*f)(void *) = (void (*)(void *))args;

    /* The parent did not allocate enough memory to execute the thread */
    if (ee->initial_stack == NULL) {
        return;
    }
    ee->stack_base = &args;
    threadInit();
    f(ee->thread);
    threadFree();
}

/*
 * threadCreate() creates a new thread, returning 0 if things went OK
 * and 1 if not.  A return value of 1 will currently be interpreted as
 * an out of memory situation although it actually covers additional 
 * resource-related problems.  See the comment below.
 */
int
threadCreate(TID tid, size_t stk_sz, int priority, void (*func)(void *))
{
    int err;
    sys_thread_t *t;

    sysAssert(THREAD(tid)->eetop == 0);

    /*
     * sysThreadCreate() returns SYS_OK if things are OK, and otherwise
     * may currently return either SYS_NOMEM, or SYS_NORESOURCE on some
     * platforms.  It's not clear what SYS_NORESOURCE really means, so
     * threadCreate effectively lumps this with SYS_NOMEM.  This could
     * be improved in the future.  We check below that an error return
     * really is one of those, so we don't lump more than we intended.
     */
    err = sysThreadCreate(&t, stk_sz, saveStackBase, (void *)func);
    if (err != SYS_OK) {
	sysAssert(err == SYS_NOMEM || err == SYS_NORESOURCE);
        return 1;		/* SYS_NOMEM or SYS_NORESOURCE */
    }
    if (!InitializeExecEnv(SysThread2EE(t), (JHandle *)tid, 0)) {
        /* start the child, which will immediately terminate */
        sysThreadResume(t);
        return 1;
    }

    if (!THREAD(tid)->daemon) {
        AdjustUserThreadCount(1);
    }

    THREAD(tid)->eetop = ptr2ll(SysThread2EE(t));
    sysThreadSetPriority(t, priority);
    sysThreadResume(t);
    return 0;
}

TID
threadSelf(void) /* Common: should be a macro? */
{
    sys_thread_t *t = sysThreadSelf();
    if (t) {
	return (TID) SysThread2EE(t)->thread;
    } else {
	return 0;
    }
}

static void freeMonitorCacheHelper(monitor_t *m, void *self)
{
    if (m && m->key) {
        sys_mon_t *mid = sysmon(m);
        while (sysMonitorEntered((sys_thread_t*) self, mid)) {
            sysMonitorExit((sys_thread_t*) self, mid);
        }
    }
}

void threadFree(void)
{
    sys_thread_t *self = sysThreadSelf();
    ExecEnv *ee = SysThread2EE(self);
    TID thread_handle = (TID) ee->thread;

    DisableAsyncEvents(ee);

    if (exceptionOccurred(ee)) {
	if (THREAD(thread_handle)->group != NULL) {
	    void *t = (void *)ee->exception.exc;
	    exceptionClear(ee);
	    execute_java_dynamic_method(ee, 
					(void *)THREAD(thread_handle)->group, 
					"uncaughtException", 
		    "(Ljava/lang/Thread;Ljava/lang/Throwable;)V", 
					thread_handle,
					t);
	}
    }

    THREAD(thread_handle)->stillborn = 1;

    if (THREAD(thread_handle)->group != NULL) {
	exceptionClear(ee);	/* don't run exit with a pending exception */
	execute_java_dynamic_method(ee, (void *)thread_handle, "exit", "()V");
    }
    sysAssert(THREAD(thread_handle)->group == NULL);

    EnableAsyncEvents(ee);

#ifdef HPROF
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_THREAD_END))
        jvmpi_thread_end((JHandle *)thread_handle);
#endif /* HPROF */

#ifdef BREAKPTS
    if (debugging) 
        notify_debugger_of_thread_end(ee, (JHandle *)thread_handle);
#endif /* BREAKPTS */

    /*
     * Smash the eetop field of the thread.  The thread
     * is still in the active threads queue at this point, so code that
     * works with an enumerated thread (GC's scanThread, Thread.dumpStack,
     * and Thread.countStackFrames must be able to deal with these
     * fields being zeroed.  Once the thread is taken off the active
     * queue, and before we lock the scheduler, the thread object might
     * be garbage collected, so we need to be done with it at that point.
     */  

    /* When allocObject(ee, classJavaLangThread) fails, JNI 
     * AttachCurrentThread may call threadFree with ee->thread set
     * to zero.
     */
    if (thread_handle) {
        QUEUE_LOCK(self);
	/* Users of eetop should check if they are zeroed */
	THREAD(thread_handle)->eetop = ll_zero_const;
	QUEUE_UNLOCK(self);

	/*
	 * This notifies anybody who is waiting for us to die.  Note that
	 * we notify *everybody* on the queue.
	 */
	monitorEnter2(ee, obj_monitor(thread_handle));
	monitorNotifyAll2(ee, obj_monitor(thread_handle));
	monitorExit2(ee, obj_monitor(thread_handle));
    }

#ifndef FAST_MONITOR
    /*
     * Unlock any java monitor the current thread still holds.
     * REMIND: It will not unlock any runtime system monitor;
     */
    CACHE_LOCK(self);
    monitorEnumerate(freeMonitorCacheHelper, self);
    CACHE_UNLOCK(self);
#else /* IS FAST_MONITOR */
    /*
     * For fast monitor scheme, it is very hard to release
     * unmatched monitor enter/exit. We do not handle it for
     * now, unless there is real need for that
     */
    if (ee->spareMonitor) {
        monitorCacheFree(ee->spareMonitor);
    }
#endif

    if (!thread_handle || !THREAD(thread_handle)->daemon) {
        AdjustUserThreadCount(-1);
    }

    DeleteExecEnv(ee);
    /*
     * GC won't scan this thread after this point.
     */
}

/*
 * Locking wrappers for sysThread* functions.  Each function
 * threadXXX enters the monitor for the thread object,
 * checks to see if the underlying sys_thread is still alive,
 * calls sysThreadXXX, then exits the monitor.
 */

#define NOVALUE_INIT(rtype, value)
#define VALUE_INIT(rtype, value)	\
    rtype ret = value;
#define CALL(func, N, arg1)		\
    func(ARGS ## N(t, arg1))
#define NOVALUE_CALL(func, N, arg1)	\
    CALL(func, N, arg1)
#define VALUE_CALL(func, N, arg1)	\
    ret = CALL(func, N, arg1)
#define NOVALUE_RETURN			\
    return
#define VALUE_RETURN			\
    return ret
#define ARGS0(arg0, arg1)		\
    arg0
#define ARGS1(arg0, arg1)		\
    arg0, arg1

#define LOCKING_THREAD_FUNC(ret, rtype, defval, func, N, argtype1)	\
rtype thread ## func(ARGS ## N(TID tid, argtype1 arg1))			\
{									\
    ret ## _INIT(rtype, defval)						\
    ExecEnv *ee;							\
    sys_thread_t *t;							\
    sys_thread_t *_self = sysThreadSelf();                              \
									\
    QUEUE_LOCK(_self);							\
    ee = (ExecEnv *)ll2ptr(THREAD(tid)->eetop);                         \
    if (ee) {					                        \
	t = EE2SysThread(ee);						\
	ret ## _CALL(hpi_thread_interface->Thread ## func, N, arg1);	\
    }									\
    QUEUE_UNLOCK(_self);						\
    ret ## _RETURN;							\
}

#define LOCKING_THREAD_FUNC_0(ret, rtype, defval, func)		\
    LOCKING_THREAD_FUNC(ret, rtype, defval, func, 0, X)
#define LOCKING_THREAD_FUNC_1(ret, rtype, defval, func, arg1)	\
    LOCKING_THREAD_FUNC(ret, rtype, defval, func, 1, arg1)

LOCKING_THREAD_FUNC_0(VALUE, int, SYS_OK, Resume)
LOCKING_THREAD_FUNC_0(VALUE, void *, NULL, StackPointer)
LOCKING_THREAD_FUNC_0(NOVALUE, void, NONE, Interrupt)
LOCKING_THREAD_FUNC_1(VALUE, int, SYS_OK, SetPriority, int /* pri */)
LOCKING_THREAD_FUNC_1(VALUE, int, SYS_OK, GetPriority, int * /* prip */)
LOCKING_THREAD_FUNC_1(VALUE, int, SYS_OK, IsInterrupted, long /* clear */)

int threadSuspend(TID tid)
{
    int ret = SYS_OK;
    ExecEnv *ee;
    ExecEnv *our_ee = EE();
    bool_t suspend_self = FALSE;

    if (our_ee->thread == (JHandle *) tid)
        suspend_self = TRUE;

    /*
     * In the case where we are suspending another thread, we need to
     * hold QUEUE_LOCK so that they don't exit while we are trying to
     * suspend them.  We also want to hold the system locks so that we
     * can be fairly sure that we aren't suspending a thread while it
     * holding a system lock.  We don't have to do anything when we are
     * suspending ourselves.
     */

    if (!suspend_self)
	lock_for_debugger(our_ee);
    ee = (ExecEnv *)ll2ptr(THREAD(tid)->eetop);
    if (ee) {
	ret = sysThreadSuspend(EE2SysThread(ee));
    }
    if (!suspend_self)
	unlock_for_debugger(our_ee);
    return ret;
}

void
threadSleep(int64_t millis)
{
    if (ll_eqz(millis)) {
	threadYield();
    } else {
        char mid_buf[128];
	sys_mon_t *mid = (sys_mon_t *)mid_buf;
	sys_thread_t *self = sysThreadSelf();
	int ret;

	sysAssert(sizeof(mid_buf) >= sysMonitorSizeof());
#ifdef LOGGING
	{
	    char namebuf[80];
	    char intbuf[32];
	    int642CString(millis, intbuf, sizeof(intbuf));
	    Log2(4, "threadSleep: TID %s millis %s\n",
		 nameOfThread(threadSelf(), namebuf, sizeof(namebuf)),
		 intbuf);
	}
#endif

	sysMonitorInit(mid);
	sysMonitorEnter(self, mid);
	ret = sysMonitorWait(self, mid, millis);
	sysMonitorExit(self, mid);
	sysMonitorDestroy(mid);

	if (SYS_INTRPT == ret) {
	    ThrowInterruptedException(0, "operation interrupted");
	}
    }
}

int
threadPostException(TID tid, void *vexc)
{
    ExecEnv *ee;
    int ret = FALSE;

    QUEUE_LOCK(sysThreadSelf());

    ee = (ExecEnv *)ll2ptr(THREAD(tid)->eetop);
    if (ee != NULL) {
	/*
	 * Remote exceptions don't currently fill in the backtrace; see
	 * bug 4010164.  Make sure we don't throw an old backtrace.
	 */
	JHandle *exc = (JHandle*) vexc;
	ExecEnv *ee = (ExecEnv *)ll2ptr(THREAD(tid)->eetop);
	sysAssert(ee != NULL);
	if (is_instance_of(exc, classJavaLangThrowable, ee)) {
	    Classjava_lang_Throwable *throwable =
		unhand((Hjava_lang_Throwable *)exc);
	    throwable->backtrace = NULL;
	}
	exceptionThrow(ee, exc);
	sysThreadPostException(EE2SysThread(ee), exc);

	ret = TRUE;
    }

    QUEUE_UNLOCK(sysThreadSelf());

    return ret;
}

bool_t DisableAsyncEvents(ExecEnv *ee)
{
    bool_t hasException;
    QUEUE_LOCK(EE2SysThread(ee));
    hasException = exceptionOccurred(ee);
    if (!hasException) {
        Hjava_lang_Thread *t = (Hjava_lang_Thread *)ee->thread;
        ee->async_disable_count++;
	if (t && threadIsInterrupted(t, TRUE)) {
	    ee->interrupt_pending = TRUE;
	}
    }
    QUEUE_UNLOCK(EE2SysThread(ee));
    return hasException;
}

bool_t EnableAsyncEvents(ExecEnv *ee)
{
    QUEUE_LOCK(EE2SysThread(ee));
    if (--ee->async_disable_count == 0) {
        Hjava_lang_Thread *t = (Hjava_lang_Thread *)ee->thread;
        if (ee->pending_async_exc) {
	    threadPostException(t, ee->pending_async_exc);
	}
	if (ee->interrupt_pending) {
	    threadInterrupt(t);
	}
	ee->pending_async_exc = NULL;
	ee->interrupt_pending = FALSE;
    }
    QUEUE_UNLOCK(EE2SysThread(ee));
    return exceptionOccurred(ee);
}

int createSystemThread(char *name, int priority,
		       long stack_size, void (*f)(void *), void *arg)
{
    TID tid = 
	(TID)execute_java_constructor(NULL, NULL, classJavaLangThread, 
	    "(Ljava/lang/ThreadGroup;Ljava/lang/String;)",
	    systemgroup, makeJavaString(name, strlen(name)));

    if (tid) {
        THREAD(tid)->daemon = 1;
    }
    if (tid == 0 ||
	threadCreate(tid, stack_size, priority, f) != 0) {
        ThrowOutOfMemoryError(0, 0);
        return SYS_ERR;
    }
    return SYS_OK;
}

static int UserThreadCount = 0;

void AdjustUserThreadCount(int n)
{
    sys_thread_t *self = sysThreadSelf();
    QUEUE_LOCK(self);
    UserThreadCount += n;
    if (UserThreadCount <= 1) {
        QUEUE_NOTIFYALL(self);
    }
    QUEUE_UNLOCK(self);
}

void WaitForSingleThreaded(void)
{
    sys_thread_t *self = sysThreadSelf();
    QUEUE_LOCK(self);
    while (UserThreadCount > 1) {
        QUEUE_WAIT(self);
    }
    QUEUE_UNLOCK(self);
}


/*
 * Access to thread information.  Quietly truncate name on jio_snprintf()
 * buffer overflow.
 */
static char *
nameOfThread(TID tid, char * buf, int length)
{
    if (!tid) {
	(void) jio_snprintf(buf, length, "name unknown (nil)");
    } else {
        (void) jio_snprintf(buf, length, "%s (0x%p)",
			    Object2CString((JHandle *) THREAD(tid)->name),
			    tid);
    } 
    return buf;
}


HArrayOfChar *
getThreadName(void)
{
    struct Classjava_lang_Thread *t;
    t = (struct Classjava_lang_Thread *) unhand(threadSelf());
    return (HArrayOfChar *)t->name;
}
