/*
 * @(#)util.c	1.76 98/09/02
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*-
	utility routines needed the interpreter.
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stddef.h>
#include <stdarg.h>

#include "oobj.h"
#include "interpreter.h"
#include "monitor.h"
#include "util.h"
#include "utf.h"
#include "typecodes.h"
#include "sys_api.h"
#include "common_exceptions.h"
#include "path.h"       /* DIR_SEPARATOR */

/* declared in interpreter.h */
PanicHook panic_hook = NULL;
OutOfMemoryHook out_of_memory_hook = NULL;

unicode *
str2unicode(char *str, unicode *ustr, long len)
{
    unicode *dst = ustr;

    memset((char *) dst, 0, len * sizeof(*dst));
    while (*str && --len >= 0)
	*ustr++ = 0xff & *str++;
    return dst;
}

void
unicode2str(unicode *src, char *dst, long len)
{
    while (--len >= 0)
	*dst++ = *src++;
    *dst = '\0';
}


int
jio_printf (const char *format, ...)
{
    int len;

    va_list args;
    va_start(args, format);
    len = jio_vfprintf(stdout, format, args);
    va_end(args);

    return len;
}

int
jio_fprintf (FILE *handle, const char *format, ...)
{
    int len;

    va_list args;
    va_start(args, format);
    len = jio_vfprintf(handle, format, args);
    va_end(args);

    return len;
}

int
jio_vfprintf (FILE *handle, const char *format, va_list args)
{
    /* vfprintf_hook may not have been installed in javah and javap */
    if (vfprintf_hook)
	return vfprintf_hook(handle, format, args);
    else
	return vfprintf(handle, format, args);
}

/*
 * WARNING: out_of_memory() aborts the runtime!  It should not be used
 * except in the case of out of memory situations that are clearly not
 * survivable, like running out of memory before the runtime is initialized.
 * If the runtime has finished initializing and you run out of memory, you
 * should throw an OutOfMemoryError instead.
 */
void
out_of_memory(void)
{
    if (out_of_memory_hook) {
	(*out_of_memory_hook) ();
    } else {
	jio_fprintf(stderr, "**Out of memory, exiting**\n");
    	Exit(1);
    }
}

void
panic(const char *format, ...)
{
    va_list ap;
    char buf[MSG_BUF_LEN];

    va_start(ap, format);
    
    /* If buffer overflow, quietly truncate */
    (void) jio_vsnprintf(buf, sizeof(buf), format, ap);

    if (panic_hook) {
    	(*panic_hook) (buf);
    } else {
    	jio_fprintf(stderr, "\n*** panic: %s\n", buf);
    }

    Abort();
}

char *
classname2string(char *src, char *dst, int size) {
    char *buf = dst;
    for (; (--size > 0) && (*src != '\0') ; src++, dst++) {
	if (*src == '/') {
	    *dst = '.';
	}  else {
	    *dst = *src;
	}
    }
    *dst = '\0';
    return buf;
}

/*
 * Warning: the buffer overflow checking here is pretty bogus although
 * should prevent overruns.
 */
void 
mangleMethodName(struct methodblock *mb, char *buffer, int buflen, 
		 int mangleType) 
{ 
    ClassClass *cb = fieldclass(&mb->fb);

    char *bufptr = buffer;
    char *bufend = buffer + buflen; 

    /* Copy over Java_ */
    (void) jio_snprintf(bufptr, bufend - bufptr, "Java_");
    bufptr += strlen(bufptr);

    if (mangleType == MangleMethodName_JDK_1) {
        /* JDK 1.0.2 convention */
        /* Mangle the class name */
        bufptr += mangleUTFString(cbName(cb), bufptr, bufend - bufptr, 
				  MangleUTF_Class);
        if (bufend - bufptr > 1) 
	    *bufptr++ = '_';
        bufptr += mangleUTFString(mb->fb.name, bufptr, bufend - bufptr, 
				  MangleUTF_Field);
	(void) jio_snprintf(bufptr, bufend - bufptr, "%s", "_stub");
	return;
    }

    /* New JNI convention */ 
    bufptr += mangleUTFString(cbName(cb), bufptr, bufend - bufptr, 
			      MangleUTF_JNI);
    if (bufend - bufptr > 1) 
        *bufptr++ = '_';
    bufptr += mangleUTFString(mb->fb.name, bufptr, bufend - bufptr, 
			      MangleUTF_JNI);

    /* JNI long name: includes a mangled signature. */
    if (mangleType == MangleMethodName_JNI_LONG) {
        if (bufend - bufptr > 2) {
	    *bufptr++ = '_';
	    *bufptr++ = '_';
        }
        bufptr += mangleUTFString2(mb->fb.signature + 1,
				   bufptr, 
				   bufend - bufptr,
				   MangleUTF_JNI,
				   SIGNATURE_ENDFUNC);
    }
}

int
maxMangledMethodNameLength(struct methodblock *mb)
{
    int maxlen = 12; /* Java_ prefix, _stub postfix, etc. */
    int ch, i;
    char *components[3];
    components[0] = cbName(mb->fb.clazz);
    components[1] = mb->fb.name;
    components[2] = mb->fb.signature;
    for (i = 0; i < 3; i++) {
        char *p = components[i];
	while ((ch = next_utf2unicode(&p)) != '\0') {
	    if (ch <= 0x7f && isalnum(ch)) {
	        maxlen++;
	    } else if (ch == '/') {
	        maxlen++;
	    } else if (ch == '_' || ch == '[' || ch == ';') {
	        maxlen += 2;
	    } else {
	        maxlen += 6;
	    }
	}
    }
    return maxlen;
}
    
static int
mangleUnicodeChar(int ch, char *bufptr, char *bufend)
{
    /* DANGER.  
     *   DO NOT USE jio_snprintf directly with format "%.5x"
     *   It doesn't understand it, and you will lose horribly.
     * REGNAD
     */
    char temp[10];
    sprintf(temp, "_%.5x", ch);	/* six characters always plus null */
    (void) jio_snprintf(bufptr, bufend - bufptr, "%s", temp);
    return strlen(bufptr);
}

int
mangleUTFString2(char *name, char *buffer, int buflen, int type, char endC) 
{ 
    char *ptr = name;
    char *bufptr = buffer;
    char *bufend = buffer + buflen - 1;	/* save space for NULL */
    int ch;
    while (((ch = next_utf2unicode(&ptr)) != endC) && (bufend - bufptr > 0)) {
	if (ch <= 0x7f && isalnum(ch)) {
	    *bufptr++ = (char)ch;
	} else if ((ch == '/' || ch == '$') && type == MangleUTF_Class) { 
	    *bufptr++ = '_';
	} else if (ch == '_' && type == MangleUTF_FieldStub) { 
	    *bufptr++ = '_';
        } else if (type == MangleUTF_JNI) {
	    char *esc = 0;
	    if (ch == '_')
	        esc = "_1";
	    else if (ch == '/')
	        esc = "_";
	    else if (ch == ';')
	        esc = "_2";
	    else if (ch == '[')
	        esc = "_3";
	    if (esc) {
	        jio_snprintf(bufptr, bufend - bufptr, esc);
	        bufptr += strlen(esc);
	    } else {
	        bufptr += mangleUnicodeChar(ch, bufptr, bufend);
	    }
        } else if (type == MangleUTF_Signature) {
	    if (isprint(ch)) {
	        *bufptr++ = ch;
	    } else {
	        bufptr += mangleUnicodeChar(ch, bufptr, bufend);
	    }
 	} else {
	    bufptr += mangleUnicodeChar(ch, bufptr, bufend);
	}
    }
    *bufptr = '\0';
    return bufptr - buffer;
}

int
mangleUTFString(char *name, char *buffer, int buflen, int type) 
{ 
    return mangleUTFString2(name, buffer, buflen, type, '\0');
}

/*
 * jio_snprintf(): sprintf with buffer limit.
 */

typedef struct InstanceData {
    char *buffer;
    char *end;
} InstanceData;

#define ERROR_RETVAL -1
#undef  SUCCESS
#define SUCCESS 0
#undef  CheckRet
#define CheckRet(x) { if ((x) == ERROR_RETVAL) return ERROR_RETVAL; }

static int 
put_char(InstanceData *this, int c)
{
    if (iscntrl(0xff & c) && c != '\n' && c != '\t') {
        c = '@' + (c & 0x1F);
        if (this->buffer >= this->end) {
            return ERROR_RETVAL;
        }
        *this->buffer++ = '^';
    }
    if (this->buffer >= this->end) {
        return ERROR_RETVAL;
    }
    *this->buffer++ = c;
    return SUCCESS;
}

static int
format_string(InstanceData *this, char *str, int left_justify, int min_width,
              int precision)
{
    int pad_length;
    char *p;

    if (str == 0) {
	return ERROR_RETVAL;
    }

    if ((int)strlen(str) < precision) {
        pad_length = min_width - strlen(str);
    } else {
        pad_length = min_width - precision;
    }
    if (pad_length < 0)
        pad_length = 0;
    if (left_justify) {
        while (pad_length > 0) {
            CheckRet(put_char(this, ' '));
            --pad_length;
        }
    }

    for (p = str; *p != '\0' && --precision >= 0; p++) {
        CheckRet(put_char(this, *p));
    }

    if (!left_justify) {
        while (pad_length > 0) {
            CheckRet(put_char(this, ' '));
            --pad_length;
        }
    }
    return SUCCESS;
}

#define MAX_DIGITS 32

static int
format_number(InstanceData *this, long value, int format_type, int left_justify,
              int min_width, int precision, bool_t zero_pad)
{
    int sign_value = 0;
    unsigned long uvalue;
    char convert[MAX_DIGITS+1];
    int place = 0;
    int pad_length = 0;
    static char digits[] = "0123456789abcdef";
    int base = 0;
    bool_t caps = FALSE;
    bool_t add_sign = FALSE;

    switch (format_type) {
      case 'o': case 'O':
          base = 8;
          break;
      case 'd': case 'D':
          add_sign = TRUE; /*FALLTHROUGH*/
      case 'u': case 'U':
          base = 10;
          break;
      case 'X':
          caps = TRUE; /*FALLTHROUGH*/
      case 'x':
          base = 16;
          break;
      case 'p':
	  caps = TRUE;  /*FALLTHROUGH*/
	  base = 16;
	  break;
    }
    sysAssert(base > 0 && base <= 16);

    uvalue = value;
    if (add_sign) {
        if (value < 0) {
            sign_value = '-';
            uvalue = -value;
        }
    }

    do {
        convert[place] = digits[uvalue % (unsigned)base];
        if (caps) {
            convert[place] = toupper(convert[place]);
        }
        place++;
        uvalue = (uvalue / (unsigned)base);
        if (place > MAX_DIGITS) {
            return ERROR_RETVAL;
        }
    } while(uvalue);
    convert[place] = 0;

    pad_length = min_width - place;
    if (pad_length < 0) {
        pad_length = 0;
    }
    if (left_justify) {
        if (zero_pad && pad_length > 0) {
            if (sign_value) {
                CheckRet(put_char(this, sign_value));
                --pad_length;
                sign_value = 0;
            }
            while (pad_length > 0) {
                CheckRet(put_char(this, '0'));
                --pad_length;
            }
        } else {
            while (pad_length > 0) {
                CheckRet(put_char(this, ' '));
                --pad_length;
            }
        }
    }
    if (sign_value) {
        CheckRet(put_char(this, sign_value));
    }

    while (place > 0 && --precision >= 0) {
        CheckRet(put_char(this, convert[--place]));
    }

    if (!left_justify) {
        while (pad_length > 0) {
            CheckRet(put_char(this, ' '));
            --pad_length;
        }
    }
    return SUCCESS;
}

int
jio_vsnprintf(char *str, size_t count, const char *fmt, va_list args)
{
    char *strvalue;
    long value;
    InstanceData this;
    bool_t left_justify, zero_pad, long_flag, fPrecision;
    int min_width, precision, ch;

    if (str == NULL) {
	return ERROR_RETVAL;
    }
    str[0] = '\0';

    this.buffer = str;
    this.end = str + count - 1;
    *this.end = '\0';		/* ensure null-termination in case of failure */

    while ((ch = *fmt++) != 0) {
        if (ch == '%') {
            zero_pad = long_flag = fPrecision = FALSE;
            left_justify = TRUE;
            min_width = 0;
            precision = this.end - this.buffer;
        next_char:
            ch = *fmt++;
            switch (ch) {
              case 0:
                  return ERROR_RETVAL;
              case '-':
                  left_justify = FALSE;
                  goto next_char;
              case '0': /* set zero padding if min_width not set */
                  if (min_width == 0)
                      zero_pad = TRUE;
		  /*FALLTHROUGH*/
              case '1': case '2': case '3':
              case '4': case '5': case '6':
              case '7': case '8': case '9':
                  if (fPrecision == TRUE) {
                      precision = precision * 10 + (ch - '0');
                  } else {
                      min_width = min_width * 10 + (ch - '0');
                  }
                  goto next_char;
              case '.':
                  fPrecision = TRUE;
                  precision = 0;
                  goto next_char;
              case 'l':
                  long_flag = TRUE;
                  goto next_char;
              case 's':
                  strvalue = va_arg(args, char *);
                  CheckRet(format_string(&this, strvalue, left_justify,
                                         min_width, precision));
                  break;
              case 'c':
                  ch = va_arg(args, int);
                  CheckRet(put_char(&this, ch));
                  break;
              case '%':
                  CheckRet(put_char(&this, '%'));
                  break;
              case 'd': case 'D':
              case 'u': case 'U':
              case 'o': case 'O':
              case 'x': case 'X':
                  value = long_flag ? va_arg(args, long) : va_arg(args, int);
                  CheckRet(format_number(&this, value, ch, left_justify,
                                         min_width, precision, zero_pad));
                  break;
	      case 'p':
		  value = (long) va_arg(args, char *);
		  CheckRet(format_number(&this, value, ch, left_justify,
		                         min_width, precision, zero_pad));
		  break;
              default:
                  return ERROR_RETVAL;
            }
        } else {
            CheckRet(put_char(&this, ch));
        }
    }
    *this.buffer = '\0';
    return strlen(str);
}

int 
jio_snprintf(char *str, size_t count, const char *fmt, ...)
{
    va_list args;
    int len;

    va_start(args, fmt);
    len = jio_vsnprintf(str, count, fmt, args);
    va_end(args);
    return len;
}

/* Return true if the two classes are in the same class package */

bool_t
IsSameClassPackage(ClassClass *class1, ClassClass *class2) 
{
    if (cbLoader(class1) != cbLoader(class2)) 
	return FALSE;
    else {
	char *name1 = cbName(class1);
	char *name2 = cbName(class2);
	char *last_slash1 = strrchr(name1, DIR_SEPARATOR);
	char *last_slash2 = strrchr(name2, DIR_SEPARATOR);
	if ((last_slash1 == NULL) || (last_slash2 == NULL)) {
	    /* One of the two doesn't have a package.  Only return true
	     * if the other one also doesn't have a package. */
	    return (last_slash1 == last_slash2); 
	} else {
	    size_t length1, length2;
	    if (*name1 == SIGNATURE_ARRAY) { 
		do name1++; while (*name1 == SIGNATURE_ARRAY);
		if (*name1 != SIGNATURE_CLASS) { 
		    /* Something is terribly wrong.  Shouldn't be here */
		    return FALSE;
		}
		name1++;
	    }
	    if (*name2 == SIGNATURE_ARRAY) { 
		do name2++; while (*name2 == SIGNATURE_ARRAY);
		if (*name2 != SIGNATURE_CLASS) { 
		    /* Something is terribly wrong.  Shouldn't be here */
		    return FALSE;
		}
		name2++;
	    }
	    length1 = last_slash1 - name1;
	    length2 = last_slash2 - name2;
	    return ((length1 == length2) 
		    && (strncmp(name1, name2, length1) == 0));
	}
    }
}

static sys_mon_t *_utf8_hash_lock = NULL;

#define UTF8_HASH_LOCK_INIT()    monitorRegister(_utf8_hash_lock, \
						"utf8 hash table")
#define UTF8_HASH_LOCK(self)	sysMonitorEnter(self, _utf8_hash_lock)
#define UTF8_HASH_LOCKED(self)	sysMonitorEntered(self, _utf8_hash_lock)
#define UTF8_HASH_UNLOCK(self)	sysMonitorExit(self, _utf8_hash_lock)

char *utf8_literal_obj_init_name;
char *utf8_literal_cls_init_name;
char *utf8_literal_null_init_sig;
char *utf8_literal_java_lang_ref_Reference;
char *utf8_literal_finalizer_name;
char *utf8_literal_finalizer_sig;
char *utf8_literal_java_lang_Class;
char *utf8_literal_java_lang_Object;
char *utf8_literal_LineNumberTable;
char *utf8_literal_CoverageTable;
char *utf8_literal_LocalVariableTable;
char *utf8_literal_ConstantValue;
char *utf8_literal_Code;
char *utf8_literal_Exceptions;
char *utf8_literal_SourceFile;
char *utf8_literal_AbsoluteSourcePath;
char *utf8_literal_TimeStamp;
char *utf8_literal_InnerClasses;

typedef struct utf8_bucket_t {
    struct utf8_bucket_t *next;
    unsigned int count;
    char utf8[1];
} utf8_bucket_t;

typedef struct utf8_table {
    int size;
    int count;
    utf8_bucket_t ** entries;
} utf8_table_t;

static utf8_table_t utf8_table;

static void
init_utf8_hash(void)
{
    ExecEnv * ee = EE();
    _utf8_hash_lock=(sys_mon_t *) sysMalloc(sysMonitorSizeof());
    if (_utf8_hash_lock == NULL)
        panic("can't allocate UTF8 hash lock");
    UTF8_HASH_LOCK_INIT();

    utf8_table.count = 0;
    utf8_table.size = 101;
    utf8_table.entries = (utf8_bucket_t **) sysCalloc(utf8_table.size,
						      sizeof(void *));

    /* Initialize the well-known utf8 literals */
    utf8_literal_obj_init_name = AddUTF8(ee, "<init>");
    utf8_literal_cls_init_name = AddUTF8(ee, "<clinit>");
    utf8_literal_null_init_sig = AddUTF8(ee, "()V");
    utf8_literal_java_lang_ref_Reference = AddUTF8(ee, "java/lang/ref/Reference");
    utf8_literal_finalizer_name = AddUTF8(ee, FINALIZER_METHOD_NAME);
    utf8_literal_finalizer_sig = AddUTF8(ee, FINALIZER_METHOD_SIGNATURE);
    utf8_literal_java_lang_Class = AddUTF8(ee, "java/lang/Class");
    utf8_literal_java_lang_Object = AddUTF8(ee, "java/lang/Object");
    utf8_literal_LineNumberTable = AddUTF8(ee, "LineNumberTable");
    utf8_literal_CoverageTable = AddUTF8(ee, "CoverageTable");
    utf8_literal_LocalVariableTable = AddUTF8(ee, "LocalVariableTable");
    utf8_literal_ConstantValue = AddUTF8(ee, "ConstantValue");
    utf8_literal_Code = AddUTF8(ee, "Code");
    utf8_literal_Exceptions = AddUTF8(ee, "Exceptions");
    utf8_literal_SourceFile = AddUTF8(ee, "SourceFile");
    utf8_literal_AbsoluteSourcePath = AddUTF8(ee, "AbsoluteSourcePath");
    utf8_literal_TimeStamp = AddUTF8(ee, "TimeStamp");
    utf8_literal_InnerClasses = AddUTF8(ee, "InnerClasses");
}

/* locks exported for lock_for_debugger in gc.c */

void lock_utf8_hash(ExecEnv *ee)
{
    if (_utf8_hash_lock == NULL)
        init_utf8_hash();
    UTF8_HASH_LOCK(EE2SysThread(ee));
}

void unlock_utf8_hash(ExecEnv *ee)
{
    UTF8_HASH_UNLOCK(EE2SysThread(ee));
}

static unsigned int
utf8_hash_fun(const char *s)
{
    unsigned raw_hash = 0;
    int c;
    while ((c = *s++)) {
      	raw_hash = raw_hash * 37 + c;
    }
    return raw_hash;
}

char *AddUTF8(ExecEnv * ee, const char *str)
{
    char *result;
    int index;
    utf8_bucket_t *bucket;

    lock_utf8_hash(ee);

    index = utf8_hash_fun(str) % utf8_table.size;
    for (bucket = utf8_table.entries[index]; bucket; bucket = bucket->next) {
        char *utf8 = bucket->utf8;
        if (utf8 == str || (utf8[0] == str[0] && strcmp(utf8, str) == 0)) {
	    bucket->count++;
	    result = utf8;
	    goto unlock;
	}
    }
    
    bucket = (utf8_bucket_t *)sysMalloc(offsetof(utf8_bucket_t, utf8) + 
					strlen(str) + 1);
    if (bucket == NULL) {
        result = NULL;
	goto unlock;
    }

    if (utf8_table.count + 1 > utf8_table.size * 4 / 3) {
	/* enlarge the table */
	int size = utf8_table.size;
	int new_size = size * 2 + 1;
	utf8_bucket_t ** new_entries;
	utf8_bucket_t ** old_entries = utf8_table.entries;

	new_entries = (utf8_bucket_t **) sysCalloc(new_size,
						   sizeof(void *));
	if (new_entries == NULL)
	    return NULL;
	
	utf8_table.size = new_size;
	utf8_table.entries = new_entries;
	while (size-- > 0) {
	    utf8_bucket_t * e = old_entries[size];
	    utf8_bucket_t * next;
	    while (e != NULL) {
		int hash;
		next = e->next;
		hash = utf8_hash_fun(e->utf8) % utf8_table.size;
		e->next = new_entries[hash];
		new_entries[hash] = e;
		e = next;
	    }
	}
	sysFree(old_entries);
	index = utf8_hash_fun(str) % utf8_table.size;
    }
    utf8_table.count++;


    strcpy(bucket->utf8, str);
    bucket->count = 1;
    bucket->next = utf8_table.entries[index];
    utf8_table.entries[index] = bucket;
    result = bucket->utf8;

 unlock:
    unlock_utf8_hash(ee);
    return result;
}


void ReleaseUTF8(ExecEnv * ee, const char *str)
{
    int index;
    utf8_bucket_t *bucket, **bucketP;

    lock_utf8_hash(ee);
    index = utf8_hash_fun(str) % utf8_table.size;

    for (bucketP = &(utf8_table.entries[index]); 
	 (bucket = *bucketP); 
	 bucketP = &(bucket->next)) {
        if (bucket->utf8 == str) {
	    if ((--bucket->count) == 0) {
	        *bucketP = bucket->next;
		sysFree(bucket);
		utf8_table.count--;
	    }
	    unlock_utf8_hash(ee);
	    return;
	}
    }
    panic("try to release a nonexistant UTF8 string");
}

char *HashUTF8(ExecEnv * ee, const char *str)
{
    char *result = NULL;
    int index;
    utf8_bucket_t *bucket;

    lock_utf8_hash(ee);
    index = utf8_hash_fun(str) % utf8_table.size;

    for (bucket = utf8_table.entries[index]; bucket; bucket = bucket->next) {
        char *utf8 = bucket->utf8;
        if (utf8 == str || (utf8[0] == str[0] && strcmp(utf8, str) == 0)) {
	    result = utf8;
	    goto unlock;
	}
    }

 unlock:
    unlock_utf8_hash(ee);
    return result;
}

void HashNameAndType(ExecEnv * ee,
		     const char *name,
		     const char *signature,
		     HashedNameAndType *hashed)
{
    hashed->name = HashUTF8(ee, name);
    hashed->signature = HashUTF8(ee, signature);
}

char *
int642CString(int64_t number, char *s, int len)
{
#define MAGIC int2ll(10000000)  /* slightly more than (2^64) ^ (1/3) */
    int group1, group2, group3;
    char *sign;

    if (ll_gez(number)) {
        sign = "";
        number = ll_neg(number);	/* this lets -2^63 still work! */
    } else {
	sign = "-";
    }

    group3 = - ll_low(ll_rem(number, MAGIC));	/* low 7 digits */
    number = ll_div(ll_add(number, int2ll(group3)), MAGIC);
    group2 = - ll_low(ll_rem(number, MAGIC));	/* next 7 digits */
    number = ll_div(ll_add(number, int2ll(group2)), MAGIC);
    group1 = - ll_low(number);	                /* high 7 digits */
    if (group1 != 0) {
        (void) jio_snprintf(s, len, "%s%d%07.7d%07.7d",
			    sign, group1, group2, group3);
    } else if (group2 != 0) {
        (void) jio_snprintf(s, len, "%s%d%07.7d", sign, group2, group3);
    } else {
        (void) jio_snprintf(s, len, "%s%d", sign, group3);      
    }
    return s;
}


/*
 * Common code shared between the JNI implementation and
 * exceptions.c.  The latter need to call using internal conventions,
 * not JNI.
 */
Hjava_lang_Class *
getClass(Hjava_lang_Object *p)
{
    if (obj_flags(p) == T_NORMAL_OBJECT) {
	return cbHandle(obj_classblock(p));
    } else {
	char buffer_space[STK_BUF_LEN];	/* usually, more than enough space */
	char *buffer = buffer_space; /* buffer != buffer_space means malloc'ed*/
	char *classname;
	ClassClass *result, *fromClass = NULL;
	switch (obj_flags(p)) {
	    case T_INT:      
		classname = SIGNATURE_ARRAY_STRING SIGNATURE_INT_STRING; 
		break;
	    case T_LONG:     
		classname = SIGNATURE_ARRAY_STRING SIGNATURE_LONG_STRING; 
		break;
	    case T_FLOAT:  
		classname = SIGNATURE_ARRAY_STRING SIGNATURE_FLOAT_STRING;
		break;
	    case T_DOUBLE:  
		classname = SIGNATURE_ARRAY_STRING SIGNATURE_DOUBLE_STRING;
		break;
	    case T_BOOLEAN: 
		classname = SIGNATURE_ARRAY_STRING SIGNATURE_BOOLEAN_STRING;
		break;
	    case T_BYTE:    
		classname = SIGNATURE_ARRAY_STRING SIGNATURE_BYTE_STRING;
		break;
	    case T_CHAR:    
		classname = SIGNATURE_ARRAY_STRING SIGNATURE_CHAR_STRING;
		break;
	    case T_SHORT:   
		classname = SIGNATURE_ARRAY_STRING SIGNATURE_SHORT_STRING;
		break;
	    case T_CLASS: {
		ArrayOfObject *array = (ArrayOfObject *)unhand(p);
		char *fromName;
		int fromNameLength;

		fromClass = (ClassClass *)(array->body[obj_length(p)]);
		fromName = cbName(fromClass);
		fromNameLength = strlen(fromName);
		/* Check if there is enough space in buffer.  Otherwise,
		 * make it bigger by allocating more space.
		 */
		if (fromNameLength + 5 >= sizeof(buffer_space)) {
		    buffer = sysMalloc(fromNameLength + 5);
		    if (buffer == 0) {
			ThrowOutOfMemoryError(0, 0);
			return 0;
		    }
		}
		classname = buffer;
		buffer[0] = SIGNATURE_ARRAY;
		if (cbName(fromClass)[0] == SIGNATURE_ARRAY) {
		    /* See above. This is safe. */
		    strcpy(buffer + 1, fromName);
		} else { 
		    /* See above. This is safe. We know buffer is big enough */
		    buffer[1] = SIGNATURE_CLASS;
		    strncpy(buffer + 2, fromName, fromNameLength);
		    buffer[2 + fromNameLength]  = SIGNATURE_ENDCLASS;
		    buffer[3 + fromNameLength]  = '\0';
		}
		break;
	    }
	    default:
		ThrowInternalError(0, "Illegal object type in getClass");
		return 0;
	}
	result = FindClassFromClass(0, classname, FALSE, fromClass);
	if (buffer != buffer_space) {
	    sysFree(buffer);
	}
	return cbHandle(result);
    }

}

