/*
 * @(#)BigInteger.c	1.12 98/09/21
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * plumbNum.c -- Source file for static methods comprising Java wrappers
 * for a subset of Colin Plumb's arbitrary precision positive integer
 * library.  These really are just wrappers; they don't add any functionality,
 * and they preserve all of Plumb's calls' quirks.  Look at Plumb's bn.doc
 * for descriptions of what each call does.
 */

#include "java_math_BigInteger.h"
#include "bn.h"
#include "prime.h"
#include <jni.h>
#include <jni_util.h>

typedef struct BigNum BigNum;

/* Forward refs for static functions */
static int byteArrayToBn(JNIEnv *env, jbyteArray byteArray, BigNum *bnP);
static jbyteArray bnToByteArray(JNIEnv *env, BigNum *bnP);
static int bnCall(JNIEnv *env, int call);


/*
 * Class:     java_math_BigInteger
 * Method:    plumbInit
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_java_math_BigInteger_plumbInit
  (JNIEnv *env, jclass unused)
{
    bnInit();
}

/*
 * Class:     java_math_BigInteger
 * Method:    plumbAdd
 * Signature: ([B[B)[B
 */
JNIEXPORT jbyteArray JNICALL Java_java_math_BigInteger_plumbAdd
  (JNIEnv *env, jclass unused, jbyteArray aA, jbyteArray bA)
{
    BigNum	a, b;
    jbyteArray  sum = NULL;

    bnBegin(&a); bnBegin(&b);

    if (byteArrayToBn(env, aA, &a) && byteArrayToBn(env, bA, &b)
	&& bnCall(env, bnAdd(&a,&b)))
	sum = bnToByteArray(env, &a);

    bnEnd(&a); bnEnd(&b);
    return sum;
}

/*
 * Class:     java_math_BigInteger
 * Method:    plumbSubtract
 * Signature: ([B[B)Ljava/math/BigInteger;
 */
JNIEXPORT jobject JNICALL Java_java_math_BigInteger_plumbSubtract
  (JNIEnv *env, jclass unused, jbyteArray aA, jbyteArray bA)
{
    BigNum	a, b;
    jbyteArray  diffMagnitude;
    jobject	diff = NULL;

    bnBegin(&a); bnBegin(&b);

    if (byteArrayToBn(env, aA, &a) && byteArrayToBn(env, bA, &b)) {
	int rc;
	bnCall(env, rc = bnSub(&a, &b));

	/* If subtraction succeeded, pack sign and magnitude into BigInteger */
	if (rc != -1 && ((diffMagnitude = bnToByteArray(env, &a)) != NULL)) {
	    long diffSignum = (rc == 1 ? -1 :
		    ((*env)->GetArrayLength(env, diffMagnitude) == 0 ? 0 : 1));
	    jclass    bigIntClass;
	    jmethodID consMethodID;
	    if ((bigIntClass = (*env)->FindClass(env,"java/math/BigInteger"))
		&& (consMethodID = (*env)->GetMethodID(env, bigIntClass,
						       "<init>", "([BI)V")))
		diff = (*env)->NewObject(env, bigIntClass, consMethodID,
					 diffMagnitude, diffSignum);
	}
    }

    bnEnd(&a); bnEnd(&b);
    return diff;
}

/*
 * Class:     java_math_BigInteger
 * Method:    plumbMultiply
 * Signature: ([B[B)[B
 */
JNIEXPORT jbyteArray JNICALL Java_java_math_BigInteger_plumbMultiply
  (JNIEnv *env, jclass unused, jbyteArray aA, jbyteArray bA)
{
    BigNum	a, b, dest;
    jbyteArray  product = NULL;

    bnBegin(&a); bnBegin(&b); bnBegin(&dest);

    if (byteArrayToBn(env, aA, &a) && byteArrayToBn(env, bA, &b)
	&& bnCall(env, bnMul(&dest, &a, &b)))
	product = bnToByteArray(env, &dest);

    bnEnd(&a); bnEnd(&b); bnEnd(&dest);
    return product;
}

/*
 * Class:     java_math_BigInteger
 * Method:    plumbDivide
 * Signature: ([B[B)[B
 */
JNIEXPORT jbyteArray JNICALL Java_java_math_BigInteger_plumbDivide
  (JNIEnv *env, jclass unused, jbyteArray aA, jbyteArray bA)
{
    BigNum	a, b, q, r;
    jbyteArray  quotient = NULL;

    bnBegin(&a); bnBegin(&b); bnBegin(&q); bnBegin(&r);

    if (byteArrayToBn(env, aA, &a) && byteArrayToBn(env, bA, &b)
	&& bnCall(env, bnDivMod(&q, &r, &a, &b)))
	quotient = bnToByteArray(env, &q);

    bnEnd(&a); bnEnd(&b); bnEnd(&q); bnEnd(&r);
    return quotient;
}

/*
 * Class:     java_math_BigInteger
 * Method:    plumbRemainder
 * Signature: ([B[B)[B
 */
JNIEXPORT jbyteArray JNICALL Java_java_math_BigInteger_plumbRemainder
  (JNIEnv *env, jclass unused, jbyteArray aA, jbyteArray bA)
{
    BigNum	a, b, r;
    jbyteArray  remainder = NULL;

    bnBegin(&a); bnBegin(&b); bnBegin(&r);

    if (byteArrayToBn(env, aA, &a) && byteArrayToBn(env, bA, &b)
	&& bnCall(env, bnMod(&r, &a, &b)))
	remainder = bnToByteArray(env, &r);

    bnEnd(&a); bnEnd(&b); bnEnd(&r);
    return remainder;
}

/*
 * Class:     java_math_BigInteger
 * Method:    plumbDivideAndRemainder
 * Signature: ([B[B)[[B
 */
JNIEXPORT jobjectArray JNICALL Java_java_math_BigInteger_plumbDivideAndRemainder
  (JNIEnv *env, jclass unused, jbyteArray aA, jbyteArray bA)
{
    BigNum  a, b, q, r;
    jarray  result = NULL;

    bnBegin(&a); bnBegin(&b); bnBegin(&q); bnBegin(&r);

    if (byteArrayToBn(env, aA, &a) && byteArrayToBn(env, bA, &b)
	&& bnCall(env, bnDivMod(&q, &r, &a, &b))) {
	/* Create a Byte array array to return results. */
	jclass byteArrayClass;
	if ((byteArrayClass  = (*env)->FindClass(env, "[B")) &&
	    (result = (*env)->NewObjectArray(env, 2, byteArrayClass, NULL))) {
	    (*env)->SetObjectArrayElement(env,result,0,bnToByteArray(env, &q));
	    (*env)->SetObjectArrayElement(env,result,1,bnToByteArray(env, &r));
	}
    }

    bnEnd(&a); bnEnd(&b); bnEnd(&q); bnEnd(&r);
    return result;
}

/*
 * Class:     java_math_BigInteger
 * Method:    plumbGcd
 * Signature: ([B[B)[B
 */
JNIEXPORT jbyteArray JNICALL Java_java_math_BigInteger_plumbGcd
  (JNIEnv *env, jclass unused, jbyteArray aA, jbyteArray bA)
{
    BigNum	a, b, dest;
    jbyteArray  gcd = NULL;

    bnBegin(&a); bnBegin(&b); bnBegin(&dest);

    if (byteArrayToBn(env, aA, &a) && byteArrayToBn(env, bA, &b)
	&& bnCall(env, bnGcd(&dest, &a, &b)))
	gcd = bnToByteArray(env, &dest);

    bnEnd(&a); bnEnd(&b); bnEnd(&dest);
    return gcd;
}

/*
 * Class:     java_math_BigInteger
 * Method:    plumbModPow
 * Signature: ([B[B[B)[B
 */
JNIEXPORT jbyteArray JNICALL Java_java_math_BigInteger_plumbModPow
  (JNIEnv *env, jclass unused, jbyteArray aA, jbyteArray bA, jbyteArray mA)
{
    BigNum	 a, b, m, dest;
    jbyteArray   result = NULL;

    bnBegin(&a); bnBegin(&b); bnBegin(&m); bnBegin(&dest);

    if (byteArrayToBn(env, aA, &a) && byteArrayToBn(env, bA, &b)
	&& byteArrayToBn(env, mA, &m)
	&& bnCall(env, bnExpMod(&dest, &a, &b, &m)))
	result = bnToByteArray(env, &dest);

    bnEnd(&a); bnEnd(&b); bnEnd(&m); bnEnd(&dest);
    return result;
}

/*
 * Class:     java_math_BigInteger
 * Method:    plumbModInverse
 * Signature: ([B[B)[B
 */
JNIEXPORT jbyteArray JNICALL Java_java_math_BigInteger_plumbModInverse
  (JNIEnv *env, jclass unused, jbyteArray aA, jbyteArray mA)
{
    BigNum	 a, m, dest;
    jbyteArray   modInv = NULL;

    bnBegin(&a); bnBegin(&m); bnBegin(&dest);

    if (byteArrayToBn(env, aA, &a) && byteArrayToBn(env, mA, &m)
	&& bnCall(env, bnInv(&dest, &a, &m)))
	modInv = bnToByteArray(env, &dest);

    bnEnd(&a); bnEnd(&m); bnEnd(&dest);
    return modInv;
}

/*
 * Class:     java_math_BigInteger
 * Method:    plumbSquare
 * Signature: ([B)[B
 */
JNIEXPORT jbyteArray JNICALL Java_java_math_BigInteger_plumbSquare
  (JNIEnv *env, jclass unused, jbyteArray aA)
{
    BigNum	 a, dest;
    jbyteArray   square = NULL;

    bnBegin(&a); bnBegin(&dest);

    if (byteArrayToBn(env, aA, &a) && bnCall(env, bnSquare(&dest, &a)))
	square = bnToByteArray(env, &dest);

    bnEnd(&a); bnEnd(&dest);
    return square;
}

/*
 * Class:     java_math_BigInteger
 * Method:    plumbGeneratePrime
 * Signature: ([B)[B
 */
JNIEXPORT jbyteArray JNICALL Java_java_math_BigInteger_plumbGeneratePrime
  (JNIEnv *env, jclass unused, jbyteArray aA)
{
    BigNum	a;
    jbyteArray  prime = NULL;

    bnBegin(&a);

    if (byteArrayToBn(env, aA, &a)
	&& bnCall(env, primeGen(&a, NULL, NULL, NULL, 0)))
	prime = bnToByteArray(env, &a);

    bnEnd(&a);
    return prime;
}


/*
 * byteArrayToBn - Converts a Java byte array into a bigNum.  Returns true
 * if the conversion succeeds, false if out of memory.  It is assumed that
 * the bigNum has already been "begun" (via bnBegin(bnP);).  If out of memory,
 * throws a Java exception.
 */
static int byteArrayToBn(JNIEnv *env, jbyteArray byteArray, BigNum *bnP)
{
    jint 	   len = (*env)->GetArrayLength(env, byteArray);
    unsigned char  *ptr = (*env)->GetPrimitiveArrayCritical(env, byteArray, 0);

    if (ptr) {
        int success = (bnInsertBigBytes(bnP, ptr, 0, len) != -1);
	(*env)->ReleasePrimitiveArrayCritical(env, byteArray, ptr, 0);
	if (!success)
	    JNU_ThrowOutOfMemoryError(env, "Allocating C bigint");
	return success;
    } else {
        return 0;
    }
}

/*
 * bnToByteArray - Converts a bigNum into a Java byte array.  If out of memory,
 * returns NULL and throws a Java exception.
 */
static jbyteArray bnToByteArray(JNIEnv *env, BigNum *bnP)
{
    int		len = (bnBits(bnP) + 7)/8;
    jbyteArray  result;

    if ((result = (*env)->NewByteArray(env, len))) {
	unsigned char  *ptr = 
	    (*env)->GetPrimitiveArrayCritical(env, result, NULL);
	if (ptr) {
	    bnExtractBigBytes(bnP, ptr, 0, len);
	    (*env)->ReleasePrimitiveArrayCritical(env, result, ptr, 0);
	} else {
	    return 0;
	}
    }
    return result;
}

/*
 * bnCall - Interpret the return code from an int-returning call in the bignum
 * package: if out of memory, throws a Java exception. Returns true if the
 * call succeeds.
 */

static int bnCall(JNIEnv *env, int rc)
{
    int success = (rc != -1);

    if (!success)
	JNU_ThrowOutOfMemoryError
	    (env, "Executing C bigint arithmetic operation");

    return success;
}
