/*
 * @(#)zip_util.c	1.36 98/08/17
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Support for reading ZIP/JAR files.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <limits.h>
#include <time.h>

#include "jni.h"
#include "jlong.h"
#include "jvm.h"
#include "zip_util.h"
#include "zlib.h"

#define MAXREFS 0xFFFF	/* max number of open zip file references */
#define MAXSIZE INT_MAX	/* max size of zip file or zip entry */

#define MCREATE()      JVM_RawMonitorCreate()
#define MLOCK(lock)    JVM_RawMonitorEnter(lock)
#define MUNLOCK(lock)  JVM_RawMonitorExit(lock)
#define MDESTROY(lock) JVM_RawMonitorDestroy(lock)

static jzfile *zfiles = 0;	/* currently open zip files */
static void *zfiles_lock = 0;

#define MAXPATHLEN 1024

/*
 * Initialize zip file support. Return 0 if successful otherwise -1
 * if could not be initialized.
 */
jint InitializeZip()
{
    extern void out_of_memory(void);
    static jboolean inited = JNI_FALSE;
    if (inited)
        return 0;
    zfiles_lock = MCREATE();
    if (zfiles_lock == 0) {
	return -1;
    }
    inited = JNI_TRUE;
    return 0;
}

/*
 * Reads len bytes of data into buf. Returns 0 if all bytes could be read,
 * otherwise returns -1.
 */
static jint readFully(jint fd, void *buf, jint len)
{
    unsigned char *bp = buf;
    while (len > 0) {
	jint n = JVM_Read(fd, (char *)bp, len);
	if (n <= 0) {
	    return -1;
	}
	bp += n;
	len -= n;
    }
    return 0;
}

/*
 * Allocates a new zip file object for the specified file name.
 * Returns the zip file object or NULL if not enough memory.
 */
jzfile *allocZip(const char *name)
{
    jzfile *zip = calloc(1, sizeof(jzfile));

    if (zip == 0) {
	return 0;
    }
    zip->name = strdup(name);
    if (zip->name == 0) {
	free(zip);
	return 0;
    }
    zip->lock = MCREATE();
    if (zip->lock == 0) {
	free(zip->name);
	free(zip);
	return 0;
    }
    return zip;
}

/*
 * Frees specified number of entries.
 */
static void freeEntries(jzentry *ep, jint count)
{
    jzentry * end = ep + count;

    for (; ep < end; ep++) {
	if (ep->name != 0) {
	    free(ep->name);
	}
	if (ep->extra != 0) {
	    free(ep->extra);
	}
	if (ep->comment != 0) {
	    free(ep->comment);
	}
    }
}

/*
 * Frees the specified zip file object.
 */
static void freeZip(jzfile *zip)
{
    if (zip->name != 0) {
	free(zip->name);
    }
    if (zip->lock != 0) {
	MDESTROY(zip->lock);
    }
    if (zip->comment != 0) {
	free(zip->comment);
    }
    if (zip->entries != 0) {
	freeEntries(zip->entries, zip->total);
	free(zip->entries);
    }
    if (zip->table != 0) {
	free(zip->table);
    }
    free(zip);
}

/*
 * Searches for end of central directory (END) header. The contents of
 * the END header will be read and placed in endbuf. Returns the file
 * position of the END header, otherwise returns 0 if the END header
 * was not found or -1 if an error occurred.
 */
static jint findEND(jzfile *zip, void *endbuf)
{
    unsigned char buf[ENDHDR * 2];
    jint len, pos;
    jint fd = zip->fd;

    /* Get the length of the zip file */
    len = pos = jlong_to_jint(JVM_Lseek(fd, 0, SEEK_END));
    if (len == -1) {
	return -1;
    }
    /*
     * Search backwards ENDHDR bytes at a time from end of file stopping
     * when the END header has been found. We need to make sure that we
     * handle the case where the signature may straddle a record boundary.
     * Also, the END header must be located within the last 64k bytes of
     * the file since that is the maximum comment length.
     */
    memset(buf, 0, sizeof(buf));
    while (len - pos < 0xFFFF) {
	unsigned char *bp;
	/* Number of bytes to check in next block */
	int count = 0xFFFF - (len - pos);
	if (count > ENDHDR) {
	    count = ENDHDR;
	}
	/* Shift previous block */
	memcpy(buf + count, buf, count);
	/* Update position and read next block */
	pos -= count;
	if (jlong_to_jint(JVM_Lseek(fd, pos, SEEK_SET)) == -1) {
	    return -1;
	}
	if (readFully(fd, buf, count) == -1) {
	    return -1;
	}
	/* Now scan the block for END header signature */
	for (bp = buf; bp < buf + count; bp++) {
	    if (GETSIG(bp) == ENDSIG) {
		/* Check for possible END header */
		jint endpos = pos + (bp - buf);
		jint clen = ENDCOM(bp);
		if (endpos + ENDHDR + clen == len) {
		    /* Found END header */
		    memcpy(endbuf, bp, ENDHDR);
		    if (jlong_to_jint(JVM_Lseek(fd, endpos + ENDHDR, SEEK_SET))
			== -1) {
			return -1;
		    }
		    if (clen > 0) {
			zip->comment = malloc(clen + 1);
			if (zip->comment == 0) {
			    return -1;
			}
			if (readFully(zip->fd, zip->comment, clen) == -1) {
			    free(zip->comment);
			    zip->comment = 0;
			    return -1;
			}
			zip->comment[clen] = '\0';
		    }
		    return endpos;
		}
	    }
	}
    }
    return 0; /* END header not found */
}

/*
 * Returns a hash code value for the specified string.
 */
static unsigned int hash(const char *s)
{
    int h = 0;
    while (*s != '\0') {
	h = 31*h + *s++;
    }
    return h;
}

/*
 * Reads zip file central directory. Returns the file position of first
 * CEN header, otherwise returns 0 if central directory not found or -1
 * if an error occurred. If zip->msg != NULL then the error was a zip
 * format error and zip->msg has the error text.
 */
static jint readCEN(jzfile *zip)
{
    jint endpos, locpos, cenpos, cenoff, cenlen;
    jint total, count, tablelen, i;
    unsigned char endbuf[ENDHDR], *cenbuf, *cp;
    jzentry *entries, **table;

    /* Clear previous zip error */
    zip->msg = 0;
    /* Get position of END header */
    endpos = findEND(zip, endbuf);
    if (endpos == 0) {
	return 0;  /* END header not found */
    }
    if (endpos == -1) {
	return -1; /* system error */
    }
    /* Get position and length of central directory */
    cenlen = ENDSIZ(endbuf);
    if (cenlen < 0 || cenlen > endpos) {
	zip->msg = "invalid END header (bad central directory size)";
	return -1;
    }
    cenpos = endpos - cenlen;
    /*
     * Get position of first local file (LOC) header, taking into
     * account that there maybe a stub prefixed to the zip file.
     */ 
    cenoff = ENDOFF(endbuf);
    if (cenoff < 0 || cenoff > cenpos) {
	zip->msg = "invalid END header (bad central directory offset)";
	return -1;
    }
    locpos = cenpos - cenoff;
    /* Get total number of central directory entries */
    total = zip->total = ENDTOT(endbuf);
    if (total < 0 || total * CENHDR > cenlen) {
	zip->msg = "invalid END header (bad entry count)";
	return -1;
    }
    /* Seek to first CEN header */
    if (jlong_to_jint(JVM_Lseek(zip->fd, cenpos, SEEK_SET)) == -1) {
	return -1;
    }
    /* Allocate temporary buffer for central directory bytes */
    cenbuf = malloc(cenlen);
    if (cenbuf == 0) {
	return -1;
    }
    /* Read central directory */
    if (readFully(zip->fd, cenbuf, cenlen) == -1) {
	free(cenbuf);
	return -1;
    }
    /* Allocate array for entries */
    entries = zip->entries = calloc(total, sizeof(jzentry));
    if (entries == 0) {
	free(cenbuf);
	return -1;
    }
    /* Allocate hash table */
    tablelen = zip->tablelen = total * 2;
    table = zip->table = calloc(tablelen, sizeof(jzentry *));
    if (table == 0) {
	free(cenbuf);
	free(entries);
	zip->entries = 0;
	return -1;
    }
    /* Now read the zip file entries */
    for (count = 0, cp = cenbuf; count < total; count++) {
	jzentry *ze = &entries[count];
	jint method, nlen, clen, elen;
	/* Check if enough bytes for CEN header */
	if ((cp - cenbuf) + CENHDR > cenlen) {
	    zip->msg = "invalid CEN header (bad header size)";
	    break;
	}
	/* Verify CEN header signature */
	if (GETSIG(cp) != CENSIG) {
	    zip->msg = "invalid CEN header (bad signature)";
	    break;
	}
	/* Check if entry is encrypted */
	if (CENVER(cp) & 1 == 1) {
	    zip->msg = "invalid CEN header (encrypted entry)";
	    break;
	}
	/* Get header field values */
	ze->time     = CENTIM(cp);
	ze->crc      = CENCRC(cp);
	ze->size     = CENLEN(cp);
	ze->csize    = CENSIZ(cp);
	nlen         = CENNAM(cp);
	elen         = CENEXT(cp);
	clen         = CENCOM(cp);
	method 	     = CENHOW(cp);
	ze->pos	     = CENOFF(cp) + locpos;
	cp += CENHDR;
	/* Verify header field values */
	if (ze->size < 0 || ze->csize < 0 || ze->csize > cenpos) {
	    zip->msg = "invalid CEN header (bad entry size)";
	    break;
	}
	if (ze->pos < 0 || ze->pos > cenpos - ze->csize) {
	    zip->msg = "invalid CEN header (bad entry offset)";
	    break;
	}
	if (method != STORED && method != DEFLATED) {
	    zip->msg = "invalid CEN header (bad compression method)";
	    break;
	}
	/* Negate position of entry LOC header */
	ze->pos = -ze->pos;
	/* Set compressed size to zero if entry uncompressed */
	if (method == STORED) {
	    ze->csize = 0;
	}
	/* Check if enough bytes for name, comment, and extra data */
	if ((cp - cenbuf) + nlen + clen + elen > cenlen) {
	    zip->msg = "invalid CEN header (bad header size)";
	    break;
	}
	/* Get the entry name */
	ze->name = malloc(nlen + 1);
	memcpy(ze->name, cp, nlen);
	ze->name[nlen] = '\0';
	cp += nlen;
	/* Get optional extra data */
	if (elen > 0) {
	    ze->extra = malloc(elen + 2);
	    if (ze->extra == 0) {
		free(ze->name);
		break;
	    }
	    /* Store the extra data size in the first two bytes */
	    ze->extra[0] = (char)elen;
	    ze->extra[1] = (char)(elen >> 8);
	    memcpy(&ze->extra[2], cp, elen);
	    cp += elen;
	}
	/* Get optional comment */
	if (clen > 0) {
	    ze->comment = malloc(clen + 1);
	    if (ze->comment == 0) {
		free(ze->name);
		free(ze->extra);
		break;
	    }
	    memcpy(ze->comment, cp, clen);
	    ze->comment[clen] = '\0';
	    cp += clen;
	}
	/* Now add the entry to the hash table */
	i = hash(ze->name) % tablelen;
	ze->next = table[i];
	table[i] = ze;
    }
    /* Free up temporary buffer */
    free(cenbuf);
    /* Check for error */
    if (count != total) {
	printf("count = %d, total = %d\n", count, total); /* DBG */
	/* Central directory was invalid, so free up entries and return */
	freeEntries(entries, count);
	free(entries);
	zip->entries = 0;
	free(table);
	zip->table = 0;
	return -1;
    }
    return cenpos;
}

/*
 * Opens a zip file for reading. Returns the jzfile object or NULL
 * if an error occurred. If a zip error occurred then *msg will be
 * set to the error message text if msg != 0. Otherwise, *msg will be
 * set to NULL.
 */
jzfile * JNICALL
ZIP_Open(const char *name, char **pmsg)
{
    char buf[MAXPATHLEN];
    jzfile *zip;

    if (InitializeZip()) {
        return NULL;
    }

    /* Clear zip error message */
    if (pmsg != 0) {
	*pmsg = 0;
    }

    if (strlen(name) >= MAXPATHLEN) {
        if (pmsg) {
            *pmsg = "zip file name too long";
        }
        return (jzfile *) 0;
    }
    strcpy(buf, name);
    JVM_NativePath(buf);
    name = buf;

    MLOCK(zfiles_lock);
    for (zip = zfiles; zip != 0; zip = zip->next) {
	if (strcmp(name, zip->name) == 0 && zip->refs < MAXREFS) {
	    zip->refs++;
	    break;
	}
    }
    MUNLOCK(zfiles_lock);
    if (zip == 0) {
	jlong len;
	/* If not found then allocate a new zip object */
	zip = allocZip(name);
	if (zip == 0) {
	    return 0;
	}
	zip->refs = 1;
	zip->fd = JVM_Open(name, O_RDONLY, 0);
	if (zip->fd == -1) {
	    freeZip(zip);
	    return 0;
	}
	len = jlong_to_jint(JVM_Lseek(zip->fd, 0, SEEK_END));
	if (len == -1) {
	    freeZip(zip);
	    return 0;
	}
	if (len > MAXSIZE) {
	    if (pmsg != 0) {
		*pmsg = "zip file too large";
	    }
	    freeZip(zip);
	    return 0;
	}
	if (readCEN(zip) <= 0) {
	    /* An error occurred while trying to read the zip file */
	    if (pmsg != 0) {
		/* Set the zip error message */
		*pmsg = zip->msg;
	    }
	    JVM_Close(zip->fd);
	    freeZip(zip);
	    return 0;
	}
	MLOCK(zfiles_lock);
	zip->next = zfiles;
	zfiles = zip;
	MUNLOCK(zfiles_lock);
    }
    return zip;
}

/*
 * Closes the specified zip file object.
 */
void ZIP_Close(jzfile *zip)
{
    MLOCK(zfiles_lock);
    if (--zip->refs > 0) {
	/* Still more references so just return */
	MUNLOCK(zfiles_lock);
	return;
    }
    /* No other references so close the file and remove from list */
    if (zfiles == zip) {
	zfiles = zfiles->next;
    } else {
	jzfile *zp;
	for (zp = zfiles; zp->next != 0; zp = zp->next) {
	    if (zp->next == zip) {
		zp->next = zip->next;
		break;
	    }
	}
    }
    MUNLOCK(zfiles_lock);
    JVM_Close(zip->fd);
    freeZip(zip);
    return;
}

/*
 * Returns the zip entry corresponding to the specified name, or
 * NULL if not found.
 */
jzentry * ZIP_GetEntry(jzfile *zip, const char *name)
{
    jzentry *entry = zip->table[hash(name) % zip->tablelen];
    while (entry != 0) {
	if (strcmp(name, entry->name) == 0) {
	    return entry;
	}
	entry = entry->next;
    }
    return 0;
}

/*
 * Returns the n'th (starting at zero) zip file entry, or NULL if the
 * specified index was out of range.
 */
jzentry * ZIP_GetNextEntry(jzfile *zip, jint n)
{
    if (n < 0 || n >= zip->total) {
	return 0;
    }
    return &zip->entries[n];
}

/*
 * Locks the specified zip file for reading.
 */
void ZIP_Lock(jzfile *zip)
{
    MLOCK(zip->lock);
}

/*
 * Unlocks the specified zip file.
 */
void ZIP_Unlock(jzfile *zip)
{
    MUNLOCK(zip->lock);
}

/* 
 * Reads local file (LOC) header for specified entry and sets position
 * of entry data.
 */
static jboolean readLOC(jzfile *zip, jzentry *entry)
{
    char locbuf[LOCHDR];
    jint pos = -entry->pos;

    /* Seek to beginning of LOC header */
    if (jlong_to_jint(JVM_Lseek(zip->fd, pos, SEEK_SET)) == -1) {
	return JNI_FALSE;
    }   
    /* Read LOC header */
    if (readFully(zip->fd, locbuf, LOCHDR) == -1) {
	return JNI_FALSE;
    }
    /* Verify signature */
    if (GETSIG(locbuf) != LOCSIG) {
	zip->msg = "invalid LOC header (bad signature)";
	return JNI_FALSE;
    }
    /* Set position of entry data */
    entry->pos = pos + LOCHDR + LOCNAM(locbuf) + LOCEXT(locbuf);
    return JNI_TRUE;
}
/*
 * Reads bytes from the specified zip entry. Assumes that the zip
 * file had been previously locked with ZIP_Lock(). Returns the
 * number of bytes read, or -1 if an error occurred. If err->msg != 0
 * then a zip error occurred and err->msg contains the error text.
 */
jint ZIP_Read(jzfile *zip, jzentry *entry, jint pos, void *buf, jint len)
{
    jint n, avail, size;

    /* Clear previous zip error */
    zip->msg = 0;
    /* Check specified position */
    size = entry->csize != 0 ? entry->csize : entry->size;
    if (pos < 0 || pos > size - 1) {
	zip->msg = "ZIP_Read: specified offset out of range";
	return -1;
    }
    /* Check specified length */
    if (len <= 0) {
	return 0;
    }
    avail = size - pos;
    if (len > avail) {
	len = avail;
    }
    /* Check if reading entry for the first time */
    if (entry->pos <= 0) {
	if (!readLOC(zip, entry)) {
	    return -1;
	}
    }
    /* Seek to beginning of entry data and read bytes */
    n = jlong_to_jint(JVM_Lseek(zip->fd, entry->pos + pos, SEEK_SET));
    if (n != -1) {
	n = JVM_Read(zip->fd, buf, len);
    }
    return n;
}

/*
 * Converts DOS (ZIP) time to UNIX time.
 */
jint ZIP_DosToUnixTime(jint dtime)
{
    struct tm tm;

    tm.tm_sec  = (dtime << 1) & 0x3E;
    tm.tm_min  = (dtime >> 5) & 0x3F;
    tm.tm_hour = (dtime >> 11) & 0x1F;
    tm.tm_mday = (dtime >> 16) & 0x1F;
    tm.tm_mon  = ((dtime >> 21) & 0x0F) - 1;
    tm.tm_year = ((dtime >> 25) & 0x7F) + 1980;

    return mktime(&tm);
}

/*
 * This function is used by the runtime system to load compressed entries
 * from ZIP/JAR files specified in the class path. It is defined here
 * so that it can be dynamically loaded by the runtime if the zip library
 * is found.
 */
jboolean
InflateFully(jzfile *zip, jzentry *entry, void *buf, char **msg)
{
    static z_stream *strm;
    char tmp[4096];
    jint pos = 0, count = entry->csize;
    jboolean status;

    *msg = 0; /* Reset error message */

    if (count == 0) {
	*msg = "inflateFully: entry not compressed";
	return JNI_FALSE;
    }

    if (strm == 0) {
	strm = calloc(1, sizeof(z_stream));
	if (inflateInit2(strm, -MAX_WBITS) != Z_OK) {
	    *msg = strm->msg;
	    free(strm);
	    return JNI_FALSE;
	}
    }

    strm->next_out = buf;
    strm->avail_out = entry->size;

    while (count > 0) {
	jint n = count > sizeof(tmp) ? sizeof(tmp) : count;
	ZIP_Lock(zip);
	n = ZIP_Read(zip, entry, pos, tmp, n);
	ZIP_Unlock(zip);
	if (n == 0) {
	    *msg = "inflateFully: Unexpected end of file";
	    inflateReset(strm);
	    return JNI_FALSE;
	}
	if (n < 0) {
	    inflateReset(strm);
	    return JNI_FALSE;
	}
	pos += n;
	count -= n;
	strm->next_in = (Bytef *)tmp;
	strm->avail_in = n;
	do {
	    switch (inflate(strm, Z_PARTIAL_FLUSH)) {
	    case Z_OK:
		break;
	    case Z_STREAM_END:
		if (count != 0 || strm->total_out != entry->size) {
		    *msg = "inflateFully: Unexpected end of stream";
		    inflateReset(strm);
		    return JNI_FALSE;
		}
		break;
	    default:
		break;
	    }
	} while (strm->avail_in > 0);
    }

    inflateReset(strm);
    return JNI_TRUE;
}

jzentry * JNICALL
ZIP_FindEntry(jzfile *zip, const char *name, jint *sizeP, jint *nameLenP)
{
    jzentry *entry = ZIP_GetEntry(zip, name);
    if (entry) {
        *sizeP = entry->size;
	*nameLenP = strlen(entry->name);
    }
    return entry;
}

/* Reads a zip file entry into the specified byte array */
jboolean JNICALL
ZIP_ReadEntry(jzfile *zip, jzentry *entry, unsigned char *buf, char *entryname)
{
    char *msg;
    strcpy(entryname, entry->name);
    if (entry->csize == 0) {
	/* Entry is stored */
	jint pos = 0, count = entry->size;
	while (count > 0) {
	    jint n;
	    ZIP_Lock(zip);
	    n = ZIP_Read(zip, entry, pos, buf, count);
	    msg = zip->msg;
	    ZIP_Unlock(zip);
	    if (n == -1) {
		jio_fprintf(stderr, "%s: %s\n", zip->name,
			    zip->msg != 0 ? zip->msg : strerror(errno));
		return JNI_FALSE;
	    }
	    buf += n;
	    pos += n;
	    count -= n;
	}
    } else {
	/* Entry is compressed */
	if (!InflateFully(zip, entry, buf, &msg)) {
	    if (*msg == 0) {
		msg = zip->msg;
	    }
	    jio_fprintf(stderr, "%s: %s\n", zip->name,
			zip->msg != 0 ? zip->msg : strerror(errno));
	    return JNI_FALSE;
	}
    }
    return JNI_TRUE;
}
