/*
 * @(#)GenAudioStreams.c	1.43 98/09/03
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/***************************************************************************/
/*
**	"GenAudioStreams.c"
**
**	This implements multi source audio streaming code.
**
**
**	History	-
**	10/5/96		Created
**	11/8/96		Added GM_AudioStreamGetStereoPosition
**	12/30/96	Changed copyright
**	1/24/97		Added fade API & GM_SetAudioStreamFadeRate
**				Changed GM_AudioStreamSetVolume to return 0
**				when stream is dead
**	3/25/97		Fixed a bug when servicing streams and it gets released
**				Removed hard coded constants for stream mode and replaced
**				with way spiffy defines
**	4/24/97		Fixed bug with STREAM_DESTROY that caused recursive callbacks
**	5/3/97		Fixed a few potential problems that the MOT compiler found
**	5/15/97		Reworked file streaming to support compressed streams
**	5/20/97		Fixed a memory leak with GM_AudioStreamFileStart
**	5/21/97		Renamed GM_AudioStreamStart to GM_AudioStreamSetup
**				Seperated load and starting of a stream into two seperate functions:
**				now you call GM_AudioStreamFileSetup OR GM_AudioStreamSetup then
**				call GM_AudioStreamStart
**	6/3/97		Added GM_AudioStreamGetReference
**				Modified GM_AudioStreamStop to not STREAM_DESTROY imeadiately. Instead
**				it calls it at the next GM_AudioServiceStream. This makes sure that
**				the right thread context is setup for the deallocation of memory
**	6/4/97		Changed GM_AudioStreamFileSetup & GM_AudioStreamSetup &
**				GM_AudioStreamGetData & GM_AudioStreamService to pass in a thread
**				context
**	6/16/97		Added GM_AudioStreamSetVolumeAll
**	6/25/97		Removed extra commented out code in GM_AudioStreamSetup
**				Redid setup and handling of small buffer sizes when getting the first and
**				second buffer. Fixed a bug in which the buffer would stop because the first
**				buffer get was smaller than the actual buffer allocated.
**	6/26/97		Modified PV_StartThisBufferPlaying to not shutdown the buffer if the oppisite
**				buffer length equals zero.
**				Fixed a bug with PV_FileStreamCallback that included extra data after
**				the sample end of file, so it caused playback noise.
**	7/16/97		Incorporated changes from David
**	7/22/97		Changed GM_AudioStreamSetVolumeAll to accecpt -1 as a volume change
**				to reset all volumes and recalculate all volumes
**	8/18/97		Added GM_AudioStreamFlush & GM_AudioStreamGetSampleOffset
**				Eliminated the 10k limit in GM_AudioStreamSetup to new streams buffer sizes.
**				This will have the effect of skipping if the buffer is too small or there is
**				not enough CPU bandwidth to process samples.
**	9/15/97		Added GM_AudioStreamGetReverb
**	9/25/97		Added some underflow condition code to stream engine
**	9/30/97		Eliminated lastAudioStreamErr, and changed GM_AudioStreamError to take
**				a stream reference to get the last error
**	10/13/97	Added new code to handle the underflow condition for the first two buffers
**				of GM_AudioStreamSetup. In that case its handle inside of GM_AudioStreamService.
**	10/16/97	Changed GM_AudioStreamError to return NO_ERR instead of NOT_SETUP if the stream
**				is invalid.
**				Cleaned out old commented out code from PV_AudioBufferCallback
**	11/10/97	Changed some preprocessor tests and flags to explicity test for flags rather
**				than assume
**	11/17/97	Added GM_AudioStreamGetSamplesPlayed & GM_AudioStreamUpdateSamplesPlayed from Kara
**	11/18/97	Fixed more underflow problems
**				Modified GM_AudioStreamPause & GM_AudioStreamResume to work with a particular
**				stream
**				Added GM_AudioStreamResumeAll & GM_AudioStreamPauseAll
**	11/19/97	Fixed a bug in GM_AudioStreamService::STREAM_MODE_DEAD that forgot to set the dataLength
**				Fixed bug with GM_AudioStreamSetup in which the dataLength (in frames) was setup as bytes
**				and fixed a potential bug in which bufferSize was less than round value and it rounded
**				into a negitive (large) number
**	11/24/97	Added GM_AudioStreamFlush
**	12/4/97		Modifed PV_StartThisBufferPlaying to set a flush flag to FALSE. Modified GM_AudioStream to
**				included a flush flag when GM_AudioStreamFlush is called. Modified
**				GM_AudioStreamService::STREAM_MODE_STOP_STREAM to detect the flush flag and free the stream
**				if the stream is already flushed.
**	12/8/97		Modified GM_AudioStreamStop to pass in a context
**				Changed GM_IsAudioStreamPlaying to check for the stream playing.
*/
/***************************************************************************/


//#define USE_DEBUG				2
//#define TEST_UNDERFLOW_CODE		1

#if USE_DEBUG
	//#include <stdio.h>
	//#define DEBUG_STR_NULL(x)	fprintf(stderr, "GenAudioStreams: %s\n", x)
	#define DEBUG_STR_NULL
#else
	#define DEBUG_STR_NULL
#endif

#include "X_API.h"
#include "GenSnd.h"
#include "GenPriv.h"

#if USE_STREAM_API == TRUE

#define STREAM_ID		FOUR_CHAR('L','I','V','E')	//	'LIVE' id for verification of valid stream

#define MAX_SAMPLE_OVERSAMPLE			4		// number of samples extra per buffer

// Stream modes
#define STREAM_MODE_INTERRUPT_ACTIVE	0x80	// this value is or'd onto the overall value
#define STREAM_MODE_DEAD				0x00
#define STREAM_MODE_START_BUFFER_1		0x01	// has to be 1
#define STREAM_MODE_START_BUFFER_2		0x02	// has to be 2
#define STREAM_MODE_STOP_STREAM			0x03
#define STREAM_MODE_FREE_STREAM			0x04


/* Used for file playback */
struct GM_AudioStreamFileInfo
{
	XFILENAME		playbackFile;
	XFILE			fileOpenRef;
	unsigned long	fileStartPosition;			// in bytes
	unsigned long	filePlaybackPosition;		// in bytes
	unsigned long	fileEndPosition;			// in bytes
	XBOOL			loopFile;

	AudioFileType	fileType;
	long			formatType;		// typed file compression mode

	XPTR			pBlockBuffer;	// used for decompression
	unsigned long	blockSize;		// used for decompression
};
typedef struct GM_AudioStreamFileInfo GM_AudioStreamFileInfo;


struct GM_AudioStream
{
	long					reference;
	long					streamID;
	long					playbackReference;

	OPErr					startupStatus;		// error return before startup
	short int				startupBufferFullCount;

	GM_StreamObjectProc		streamCallback;
	GM_StreamData			streamData;
	void					*pStreamBuffer;
	unsigned long			streamBufferLength;

	unsigned long			streamOrgLength1;
	unsigned long			streamOrgLength2;

	void					*pStreamData1;
	void					*pStreamData2;
	unsigned long			streamLength1;
	unsigned long			streamLength2;
	XBYTE					streamMode;					// Stream modes
	XBYTE					lastStreamBufferPlayed;
	short int				streamVolume;
	short int				streamStereoPosition;

	unsigned long			streamPlaybackPosition;		// in samples; samples in this stream processed by engine
    unsigned long			streamPlaybackOffset;		// in samples; total samples processed by engine when this stream starts
    unsigned long           samplesWritten;             // update in GM_AudioStreamService
    unsigned long           samplesPlayed;              // update in GM_AudioStreamUpdateSamplesPlayed
    unsigned long           residualSamples;			// if we're underflowing, then get more data and reset streamPlaybackOffset
														// before playing all samples, need to record that we can still play these
														// samples before reaching the new streamPlaybackOffset.

	XBOOL					streamActive;
	XBOOL					streamShuttingDown;
	XBOOL					streamPaused;
	XBOOL					streamFirstTime;			// first time active
	XBOOL					streamUnderflow;
	XBOOL					streamFlushed;				// only set to TRUE when flush is called. Reset to FALSE at start

	FIXED_VALUE				streamFadeRate;				// when non-zero fading is enabled
	FIXED_VALUE				streamFixedVolume;			// inital volume level that will be changed by streamFadeRate
	short int				streamFadeMaxVolume;		// max volume
	short int				streamFadeMinVolume;		// min volume
	XBOOL					streamEndAtFade;

	OPErr					streamErr;

	GM_AudioStreamFileInfo	*pFileStream;				// if not NULL, then streaming file
	struct GM_AudioStream	*pNext;
};
typedef struct GM_AudioStream GM_AudioStream;


static GM_AudioStream	*theStreams = NULL;

// verify reference is a valid audio stream structure
static GM_AudioStream * PV_AudioStreamGetFromReference(long reference)
{
	GM_AudioStream *pStream;
	GM_AudioStream * next;

	pStream = (GM_AudioStream *)reference;
	next = theStreams;
	while ( next != NULL )
	{
		if (next == pStream)
		{
			if (pStream->streamID == STREAM_ID)
			{
				return pStream;
			}
		}
		next = next->pNext;
	}
	return NULL;
}

static void PV_AddStream(GM_AudioStream *next)
{
	GM_AudioStream *last;

	if (next)
	{
		next->streamID = STREAM_ID;
		if (theStreams == NULL)
		{
			theStreams = next;
		}
		else
		{
			last = theStreams;
			while (last->pNext)
			{
				last = last->pNext;
			}
			last->pNext = next;
		}
		next->pNext = NULL;
	}
}


static void PV_FreeStream(GM_AudioStream *found)
{
	GM_AudioStream *next, *last;

	found = PV_AudioStreamGetFromReference((long)found);		// verify as valid
	if (found)
	{
		if (found->streamID == STREAM_ID)
		{
			last = next = theStreams;
			while (next)
			{
				if (next == found)						// found object in list?
				{
					if (next == theStreams)				// is object the top object
					{
						theStreams = next->pNext;		// yes, change to next object
					}
					else
					{
						if (last)						// no, change last to point beyond next
						{
							last->pNext = next->pNext;
						}
					}
					if (next->pFileStream)
					{
						XDisposePtr(next->pFileStream->pBlockBuffer);
						XDisposePtr(next->pFileStream);
					}
					XDisposePtr(next);					// clean up
					break;
				}
				last = next;
				next = next->pNext;
			}
		}
	}
}

// Get an empty GM_AudioStream. Will return 0 if can't allocate stream
static long PV_GetEmptyAudioStream(void)
{
	GM_AudioStream	*pStream;
	long			ref;

	ref = 0;
	pStream = (GM_AudioStream *)XNewPtr((long)sizeof(GM_AudioStream));
	if (pStream)
	{
// $$kk: i think this is dangerous because now the stream is in the list, but
// its properties may not be initialized.  we saw crashes in GM_AudioStreamGetFileSamplePosition
// when dividing by pStream->streamData.channelSize, which may be 0.
//		PV_AddStream(pStream);
		pStream->reference = -1;
		ref = (long)pStream;
	}
	return ref;
}


static unsigned long PV_GetSampleSizeInBytes(GM_StreamData * pAS)
{
	return pAS->channelSize * (pAS->dataBitSize / 8);
}


static void PV_FillBufferEndWithSilence(char *pDest, GM_StreamData * pAS)
{
	unsigned long	bufferSize, blockSize;
	unsigned long	count;
	short int		*pWData;

	if (pDest)
	{
		blockSize = MAX_SAMPLE_OVERSAMPLE * PV_GetSampleSizeInBytes(pAS);
		bufferSize = (pAS->dataLength * PV_GetSampleSizeInBytes(pAS));

		pDest += bufferSize;
		if (pAS->dataBitSize == 8)
		{
			for (count = 0; count < blockSize; count++)
			{
				*pDest++ = 0x80;
			}
		}
		else
		{
			pWData = (short int *)pDest;
			blockSize /= 2;
			for (count = 0; count < blockSize; count++)
			{
				*pWData++ = 0;
			}
		}
	}
}

static void PV_CopyLastSamplesToFirst(char *pSource, char *pDest, GM_StreamData * pAS)
{
	unsigned long	bufferSize, blockSize;

	if (pAS->dataLength && pSource && pDest)
	{
		blockSize = MAX_SAMPLE_OVERSAMPLE * PV_GetSampleSizeInBytes(pAS);
		bufferSize = (pAS->dataLength * PV_GetSampleSizeInBytes(pAS));
		if (bufferSize)
		{
			XBlockMove(pSource + bufferSize, pDest, blockSize);
		}
	}
}

static void PV_AudioSoundDoneCallback(INT32 reference)
{
	GM_AudioStream	*pStream;

	pStream = (GM_AudioStream *)reference;

	if (pStream)
	{
		pStream->playbackReference = -1;
	}
}

static void PV_AudioBufferCallback(INT32 reference, G_PTR pWhichBufferFinished, INT32 *pBufferSize_IN_OUT)
{
	GM_AudioStream	*pStream;

//#pragma unused (pWhichBufferFinished)
	pWhichBufferFinished = pWhichBufferFinished;

	pStream = (GM_AudioStream *)reference;

// This is the same thing as the above line, just faster.
//	pStream = PV_AudioStreamGetFromReference(reference);

	if (pStream)
	{
		#if USE_DEBUG && 0
		{
			char text[256];

			sprintf(text, "Playback Pos %ld Length %ld\r", pStream->streamPlaybackPosition, *pBufferSize_IN_OUT);
			DEBUG_STR(text);
		}
		#endif
		#if USE_DEBUG && 0
		{
			char text[256];

			sprintf(text, "buffer done %lx size %ld", pWhichBufferFinished, *pBufferSize_IN_OUT);
			DEBUG_STR(text);
		}
		#endif
		pStream->streamPlaybackPosition += *pBufferSize_IN_OUT;
		switch (pStream->streamMode & (~STREAM_MODE_INTERRUPT_ACTIVE))
		{
			default:
				DEBUG_STR_NULL("Bad case in PV_AudioBufferCallback");
				break;
			case STREAM_MODE_STOP_STREAM:
				DEBUG_STR_NULL("PV_AudioBufferCallback::STREAM_MODE_STOP_STREAM");
				pStream->streamMode = STREAM_MODE_INTERRUPT_ACTIVE | STREAM_MODE_FREE_STREAM;		// end
				pStream->streamShuttingDown = TRUE;
				*pBufferSize_IN_OUT = 0;
				break;
			// buffer 1 ends playback
			case STREAM_MODE_START_BUFFER_1:	// start buffer 2 playing
				DEBUG_STR_NULL("PV_AudioBufferCallback::STREAM_MODE_START_BUFFER_1");

				if (pStream->streamFirstTime)
				{
					// copy end of buffer 2 into the start of buffer 1
					// This only needs to happen once at the start because the buffers are different in the begining
					PV_CopyLastSamplesToFirst((char *)pStream->pStreamData2, (char *)pStream->pStreamData1, &pStream->streamData);
					pStream->streamFirstTime = FALSE;
				}

				*pBufferSize_IN_OUT = pStream->streamLength2;
				if (pStream->streamShuttingDown)
				{
					pStream->streamShuttingDown = TRUE;
					DEBUG_STR_NULL("    End of BUFFER_1");
					if (pStream->streamLength2)
					{
						pStream->streamMode = STREAM_MODE_INTERRUPT_ACTIVE | STREAM_MODE_STOP_STREAM;		// end
					}
					else
					{
						pStream->streamMode = STREAM_MODE_INTERRUPT_ACTIVE | STREAM_MODE_FREE_STREAM;		// end
					}
				}
				else
				{
					pStream->streamMode = STREAM_MODE_INTERRUPT_ACTIVE | STREAM_MODE_START_BUFFER_2;		// buffer1 read and playing2
				}
				break;
			// buffer 2 ends playback
			case STREAM_MODE_START_BUFFER_2:
				DEBUG_STR_NULL("PV_AudioBufferCallback::STREAM_MODE_START_BUFFER_2");
				*pBufferSize_IN_OUT = pStream->streamLength1;
				if (pStream->streamShuttingDown)
				{
					pStream->streamShuttingDown = TRUE;
					DEBUG_STR_NULL("    End of BUFFER_2");
					if (pStream->streamLength1)
					{
						pStream->streamMode = STREAM_MODE_INTERRUPT_ACTIVE | STREAM_MODE_STOP_STREAM;		// end
					}
					else
					{
						pStream->streamMode = STREAM_MODE_INTERRUPT_ACTIVE | STREAM_MODE_FREE_STREAM;		// end
					}
				}
				else
				{
					pStream->streamMode = STREAM_MODE_INTERRUPT_ACTIVE | STREAM_MODE_START_BUFFER_1;		// buffer2 read and playing1
				}
				break;
		}
	}
}



// Start playing buffer. Account for MAX_SAMPLE_OVERSAMPLE in length
static void PV_StartThisBufferPlaying(GM_AudioStream * pStream, XBYTE bufferNumber)
{
	XBYTE	mode;

	// we use (pStream->streamPlaybackOffset == -1) to indicate that the
	// value needs to be set.  this is the case 1) when the stream is just
	// starting up and 2) when the stream has shut down (probably due to
	// underflow).  this is now the centralized location for setting
	// pStream->streamPlaybackOffset to -1 during play.  (before, there were
	// potential timing problems, and we didn't necessarily count all the
	// committed samples....

	if (GM_IsSoundDone(pStream->playbackReference))
	{
		// if we were set to a valid offset before (i.e. we're not just starting)
		// record any residual samples that we should count.
		if (pStream->streamPlaybackOffset != -1)
		{
			pStream->residualSamples = pStream->samplesWritten - pStream->samplesPlayed;
		}
		pStream->streamPlaybackOffset = -1;
	}

	if (pStream->streamShuttingDown == FALSE)
	{
		switch (bufferNumber)
		{
			case 0:	// don't know which buffer to play, figure it out, eh?
				DEBUG_STR("PV_StartThisBufferPlaying:0");
				if (pStream->lastStreamBufferPlayed)
				{
					mode = pStream->streamMode &= (~STREAM_MODE_INTERRUPT_ACTIVE);
					if (mode != pStream->lastStreamBufferPlayed)
					{
						DEBUG_STR("DIFFERENT!!");
						mode = pStream->lastStreamBufferPlayed;
					}
					PV_StartThisBufferPlaying(pStream, mode);	// play last successfull buffer played
				}
				break;

			case STREAM_MODE_START_BUFFER_1:
				DEBUG_STR("PV_StartThisBufferPlaying:STREAM_MODE_START_BUFFER_1");
				if (pStream->streamLength1)
				{
					// if it is currently unset (-1 value), set the streamPlaybackOffset
					// to the current position of the engine in samples
					if (pStream->streamPlaybackOffset == -1)
					{
						pStream->streamPlaybackOffset = MusicGlobals->samplesWritten;
					}

					pStream->playbackReference =
						GM_BeginDoubleBuffer(	(G_PTR)pStream->pStreamData1,
												(G_PTR)pStream->pStreamData2,
												pStream->streamLength1,
												pStream->streamData.sampleRate,
												pStream->streamData.dataBitSize, pStream->streamData.channelSize,
												pStream->streamVolume,
												pStream->streamStereoPosition,
												(long)pStream,
												PV_AudioBufferCallback);
					if (pStream->playbackReference != -1)	// successfull?
					{
						// $$kk: 09.03.98: added GM_SetSampleDoneCallback callback to fix stream crossing the right way! 
						GM_SetSampleDoneCallback(pStream->playbackReference, PV_AudioSoundDoneCallback, (long)pStream);
						pStream->lastStreamBufferPlayed = STREAM_MODE_START_BUFFER_1;
					}
					pStream->streamFlushed = FALSE;
				}
				break;
			case STREAM_MODE_START_BUFFER_2:
				DEBUG_STR("PV_StartThisBufferPlaying:STREAM_MODE_START_BUFFER_2");
				if (pStream->streamLength2)
				{
	                // if it is currently unset (-1 value), set the streamPlaybackOffset
	                // to the current position of the engine in samples
	                if (pStream->streamPlaybackOffset == -1)
	                {
	                    pStream->streamPlaybackOffset = MusicGlobals->samplesWritten;
					}
					pStream->playbackReference =
						GM_BeginDoubleBuffer(	(G_PTR)pStream->pStreamData2,
												(G_PTR)pStream->pStreamData1,
												pStream->streamLength2,
												pStream->streamData.sampleRate,
												pStream->streamData.dataBitSize, pStream->streamData.channelSize,
												pStream->streamVolume,
												pStream->streamStereoPosition,
												(long)pStream,
												PV_AudioBufferCallback);
					if (pStream->playbackReference != -1)	// successfull?
					{
						// $$kk: 09.03.98: added GM_SetSampleDoneCallback callback to fix stream crossing the right way! 
						GM_SetSampleDoneCallback(pStream->playbackReference, PV_AudioSoundDoneCallback, (long)pStream);
						pStream->lastStreamBufferPlayed = STREAM_MODE_START_BUFFER_2;
					}
					pStream->streamFlushed = FALSE;
				}
				break;
		}
	}
}

//$$kk: 10.22.97
// i am taking out these pause/resume methods and replacing them with the
// last version; these for some reasons cause synch to be off ON WIN32 ONLY
// after pause/resume

// Pause this particular audio stream
void GM_AudioStreamPause(long reference)
{
	GM_AudioStream		*pStream;

	pStream = PV_AudioStreamGetFromReference(reference);
	if (pStream)
	{
		if (pStream->streamActive && (!pStream->streamPaused))
		{

			pStream->streamPaused = TRUE;
			GM_EndSample(pStream->playbackReference);

// $$kk: 09.03.98: should not need this anymore		
/*			
			//  $$kk: 02.13.98: set playbackReference = -1 whenever we call GM_EndSample to avoid jmf stream crossing
			long playbackReference = pStream->playbackReference;

			pStream->streamPaused = TRUE;

			pStream->playbackReference = -1;
			GM_EndSample(playbackReference);
*/
		}
	}
}

// Resume this particular audio stream
void GM_AudioStreamResume(long reference)
{
	GM_AudioStream		*pStream;

	pStream = PV_AudioStreamGetFromReference(reference);
	if (pStream)
	{
		if (pStream->streamActive && pStream->streamPaused)
		{
			pStream->streamPaused = FALSE;

			// $$kk: 02.20.98
			// if the stream is flushed and STREAM_MODE_DEAD, should not start playing here!
			// this change should fix the JMF stutter problem, bug #4098200
			if (!(pStream->streamFlushed))
				PV_StartThisBufferPlaying(pStream, (XBYTE)(pStream->streamMode & (~STREAM_MODE_INTERRUPT_ACTIVE)));
		}
	}
}



// Pause all audio streams
void GM_AudioStreamPauseAll(void)
{
	GM_AudioStream	*pStream;

	pStream = theStreams;
	while (pStream)
	{
		if (pStream->streamActive && (!pStream->streamPaused))
		{
			pStream->streamPaused = TRUE;
			GM_EndSample(pStream->playbackReference);

// $$kk: 09.03.98: should not need this anymore		
/*			
			//  $$kk: 02.13.98: set playbackReference = -1 whenever we call GM_EndSample to avoid jmf stream crossing
			long playbackReference = pStream->playbackReference;

			pStream->streamPaused = TRUE;

			pStream->playbackReference = -1;
			GM_EndSample(playbackReference);
*/
		}
		pStream = pStream->pNext;
	}
}

// Resume all audio streams
void GM_AudioStreamResumeAll(void)
{
	GM_AudioStream	*pStream;

	pStream = theStreams;
	while (pStream)
	{
		if (pStream->streamActive && pStream->streamPaused)
		{
			pStream->streamPaused = FALSE;
			PV_StartThisBufferPlaying(pStream, (XBYTE)(pStream->streamMode & (~STREAM_MODE_INTERRUPT_ACTIVE)));
		}
		pStream = pStream->pNext;
	}
}

OPErr GM_AudioStreamError(long reference)
{
	GM_AudioStream		*pStream;
	OPErr				theErr;

	theErr = NO_ERR;
	pStream = PV_AudioStreamGetFromReference(reference);
	if (pStream)
	{
		theErr = pStream->streamErr;
	}
	return theErr;
}

#if USE_HIGHLEVEL_FILE_API
// streaming file callback. Used for GM_AudioStreamFileStart to decode typed files.
static OPErr PV_FileStreamCallback(void *context, GM_StreamMessage message, GM_StreamData *pAS)
{
	OPErr					error;
	unsigned long			bufferSize, fileSize, outputBufferSize;
	OPErr					fileError;
	GM_AudioStreamFileInfo	*pASInfo;
	GM_AudioStream			*pStream;

#if TEST_UNDERFLOW_CODE
	static int				pv_count = 0;
	static int				pv_count2 = 0;
#endif

	context = context;
	error = NO_ERR;
	switch (message)
	{
		case STREAM_CREATE:
			DEBUG_STR_NULL("PV_FileStreamCallback::STREAM_CREATE\r");
			pStream = (GM_AudioStream *)pAS->streamReference;
			pASInfo = (GM_AudioStreamFileInfo *)pAS->userReference;
			pStream->pFileStream = pASInfo;

			pASInfo->fileOpenRef = XFileOpenForRead(&pASInfo->playbackFile);
			if (pASInfo->fileOpenRef)
			{
				// remember to take in account that the data length passed is always in audio frames not bytes
				bufferSize = pAS->dataLength * pAS->channelSize * (pAS->dataBitSize / 8);

				pAS->pData = XNewPtr(bufferSize);
				if (pAS->pData)
				{
					error = NO_ERR;
					XFileSetPosition(pASInfo->fileOpenRef, pASInfo->filePlaybackPosition);
				}
				else
				{
					error = MEMORY_ERR;
				}
			}
			else
			{
				error = GENERAL_BAD;
			}
			break;
		case STREAM_DESTROY:
			DEBUG_STR_NULL("PV_FileStreamCallback::STREAM_DESTROY\r");
			pASInfo = (GM_AudioStreamFileInfo *)pAS->userReference;
			if (pASInfo->fileOpenRef)
			{
				XFileClose(pASInfo->fileOpenRef);
			}
			XDisposePtr((XPTR)pAS->pData);
			pAS->pData = NULL;
			break;
		case STREAM_GET_SPECIFIC_DATA:
			error = NOT_SETUP;
			break;
		case STREAM_GET_DATA:
#if TEST_UNDERFLOW_CODE
			pv_count++;
			if (pv_count == 0)
			{
				pAS->dataLength = 0;
				DEBUG_STR("Fake underflow!");
				return NO_ERR;
			}
#endif
			pASInfo = (GM_AudioStreamFileInfo *)pAS->userReference;
			if (pAS->pData)
			{
				// get the desired length, and account for stereo and bit size
				fileError  = (OPErr)GM_ReadAndDecodeFileStream(	pASInfo->fileOpenRef,
															(AudioFileType)pASInfo->fileType,
															pASInfo->formatType,
															pASInfo->pBlockBuffer,
															pASInfo->blockSize,
															(XPTR)pAS->pData,
															pAS->dataLength,
															pAS->channelSize,
															pAS->dataBitSize,
															&outputBufferSize,
															&fileSize);

				#if USE_DEBUG && 1
				{
					char text[256];

					sprintf(text, "STREAM_GET_DATA::frames %ld outputBufferSize %ld fileSize %ld", pAS->dataLength, outputBufferSize, fileSize);
					DEBUG_STR(text);
				}
				#endif
				// NOTE: we are incrementing the file position by the buffer size rather
				// than the file size because our units are in audio bytes.
				pASInfo->filePlaybackPosition += outputBufferSize;

				// return actual length, if this is not done, then it will
				// continue to play the last buffer size
				// NOTE: we're returned the number of audio frames, not the actual byte size
				pAS->dataLength = outputBufferSize / pAS->channelSize / (pAS->dataBitSize / 8);

				if (pASInfo->filePlaybackPosition >= pASInfo->fileEndPosition)
				{
					// how much over are we? Our fileEndPosition is at the exact end of the file,
					// so if we're over then there's extra data after the sample data. We don't want to
					// include that in the playback because its noise.
					fileSize = pASInfo->filePlaybackPosition - pASInfo->fileEndPosition;
					pASInfo->filePlaybackPosition -= fileSize;
					pAS->dataLength -= (fileSize / pAS->channelSize / (pAS->dataBitSize / 8));

					pASInfo->filePlaybackPosition = pASInfo->fileStartPosition;

					XFileSetPosition(pASInfo->fileOpenRef, pASInfo->filePlaybackPosition);
					fileError = NO_ERR;

					if (pASInfo->loopFile)
					{
						// we've hit the end of the file, so reset to begining and continue
						// clear IMA transient data
						XSetMemory(pASInfo->pBlockBuffer, pASInfo->blockSize, 0);
						error = NO_ERR;
					}
					else
					{
						error = STREAM_STOP_PLAY;	// we've hit the end of the file, so stop
					}
				}
				else
				{
					error = NO_ERR;
				}
			}
			else
			{
				error = GENERAL_BAD;
			}
			break;
	}
	return error;
}
#endif	// USE_HIGHLEVEL_FILE_API

#if USE_HIGHLEVEL_FILE_API
// setup streaming a file and place it into pause mode. Don't start
long GM_AudioStreamFileSetup(void *context,
									XFILENAME *file, AudioFileType fileType,
									unsigned long bufferSize, GM_Waveform *pFileInfo,
									XBOOL loopFile)
{
	long					reference;
	GM_Waveform				*pWaveform;
	GM_AudioStreamFileInfo	*pStream;
	long					format, blockSize;

	reference = 0;
    pWaveform = GM_ReadFileInformation(file, fileType, &format, &blockSize);
	if (pWaveform)
	{
		pStream = (GM_AudioStreamFileInfo *)XNewPtr((long)sizeof(GM_AudioStreamFileInfo));
		if (pStream)
		{
			pStream->playbackFile = *file;
			pStream->loopFile = loopFile;
			pStream->formatType = format;
			pStream->fileType = fileType;
			pStream->blockSize = blockSize;
			if (blockSize)
			{
				pStream->pBlockBuffer = XNewPtr(blockSize);
			}
			// now the file is positioned right at the data block
			pStream->filePlaybackPosition = pWaveform->waveformID;
			pStream->fileStartPosition = pWaveform->waveformID;
			// we want the byte size
			pStream->fileEndPosition = pWaveform->waveSize + pStream->fileStartPosition;

			if (pFileInfo)
			{
				*pFileInfo = *pWaveform;
			}
			if (blockSize)
			{
				bufferSize = (bufferSize * blockSize) / blockSize;	// round down to blockSize
			}
			reference = GM_AudioStreamSetup(context, (long)pStream, PV_FileStreamCallback,
												bufferSize,
												pWaveform->sampledRate,
												pWaveform->bitSize,
												pWaveform->channels);
		}
		XDisposePtr(pWaveform);
	}
	return reference;
}
#endif // USE_HIGHLEVEL_FILE_API

// This will start a streaming audio object.
//
// INPUT:
//	userReference	This is a reference value that will be returned and should be passed along to all GM_AudioStream
//					functions.
//
//	pProc		is a GM_StreamObjectProc proc pointer. At startup of the streaming the proc will be called
//				with STREAM_CREATE, then followed by two STREAM_GET_DATA calls to get two buffers of data,
//				and finally STREAM_DESTROY when finished.
//
// OUTPUT:
//	long			This is an audio reference number. Will be non-zero for valid stream

long GM_AudioStreamSetup(void *context, long userReference, GM_StreamObjectProc pProc,
							unsigned long bufferSize,
							FIXED_VALUE sampleRate,	// Fixed 16.16 sample rate
							char dataBitSize,		// 8 or 16 bit data
							char channelSize)		// 1 or 2 channels of date
{
	long			reference;
	GM_AudioStream	*pStream;
	GM_StreamData	ssData;
	OPErr			theErr;
	unsigned long	byteLength;

	reference = 0;
	theErr = NO_ERR;
	DEBUG_STR_NULL("StreamSetup: Starting.\n");

	if (MusicGlobals->MaxEffects)
	{
//		if (bufferSize >= 10000L)
		{
			if ( (pProc) && ( (channelSize >= 1) || (channelSize <= 2) ) && ( (dataBitSize == 8) || (dataBitSize == 16) ) )
			{
				reference = PV_GetEmptyAudioStream();
				if (reference)
				{
					pStream = (GM_AudioStream *)reference;
					pStream->streamCallback = pProc;
					pStream->streamUnderflow = FALSE;
					pStream->reference = userReference;
					pStream->streamPlaybackPosition = 0;
					pStream->streamPlaybackOffset = -1;     // not known
					pStream->samplesWritten = 0;
					pStream->samplesPlayed = 0;
					pStream->residualSamples = 0;
					pStream->streamShuttingDown = FALSE;
					pStream->streamVolume = MAX_NOTE_VOLUME;
					pStream->streamStereoPosition = 0;
					ssData.pData = NULL;
					ssData.userReference = pStream->reference;
					ssData.streamReference = (long)pStream;
					ssData.sampleRate = sampleRate;
					ssData.dataBitSize = dataBitSize;
					ssData.channelSize = channelSize;
					if (bufferSize > 1000)
					{
						bufferSize -= (bufferSize % 1000L);		// round down to the nearest 100
					}
					// add over sample extra
					bufferSize += PV_GetSampleSizeInBytes(&ssData) * MAX_SAMPLE_OVERSAMPLE;
					pStream->streamBufferLength = bufferSize;

					// convert from bytes to frames
					ssData.dataLength = bufferSize / PV_GetSampleSizeInBytes(&ssData);

					theErr = (*pProc)(context, STREAM_CREATE, &ssData);

					if (theErr == NO_ERR)
					{
						pStream->streamData = ssData;
						pStream->pStreamBuffer = ssData.pData;
						pStream->pStreamData1 = pStream->streamData.pData;

						byteLength = ssData.dataLength * PV_GetSampleSizeInBytes(&ssData);
						pStream->pStreamData2 = (char *)pStream->streamData.pData + (byteLength / 2);
						pStream->streamLength1 = ssData.dataLength / 2;
						pStream->streamLength2 = ssData.dataLength / 2;

						pStream->streamOrgLength1 = pStream->streamLength1;
						pStream->streamOrgLength2 = pStream->streamLength2;
						pStream->streamMode = STREAM_MODE_DEAD;
						pStream->streamUnderflow = FALSE;
						pStream->startupBufferFullCount = 0;

						// ok, fill first buffer
						ssData.userReference = pStream->reference;
						ssData.streamReference = (long)pStream;
						ssData.pData = pStream->pStreamData1;
						// get the full amount this buffer only
						ssData.dataLength = pStream->streamLength1;
						ssData.dataBitSize = dataBitSize;
						ssData.channelSize = channelSize;
						ssData.sampleRate = sampleRate;
						DEBUG_STR_NULL("StreamSetup: Call callback with GET_DATA\n");

						theErr = (*pProc)(context, STREAM_GET_DATA, &ssData);

#if USE_DEBUG
						if( theErr)
						{
							DEBUG_STR_NULL("StreamSetup: Return from 1st callback with an error\n");
						}
						else
						{
							DEBUG_STR_NULL("StreamSetup: Return 1st from callback o.k.\n");
						}
#endif
						pStream->streamLength1 = ssData.dataLength;			// just in case it changes
						if (pStream->streamLength1 == 0)
						{	// underflow, get this buffer again
							pStream->streamUnderflow = TRUE;
						}
						else
						{
							pStream->startupBufferFullCount++;	// first buffer full
							if ( (theErr == NO_ERR) || (theErr == STREAM_STOP_PLAY) )
							{
								if (theErr == NO_ERR)
								{
									// update count of samples written
									pStream->samplesWritten += pStream->streamLength1;

									if (ssData.dataLength < MAX_SAMPLE_OVERSAMPLE)
									{
										ssData.dataLength += MAX_SAMPLE_OVERSAMPLE;		// going to click for sure
									}
									else
									{
										ssData.dataLength -= MAX_SAMPLE_OVERSAMPLE;
									}
									// copy end of buffer 1 into the start of buffer 2
									PV_CopyLastSamplesToFirst((char *)pStream->pStreamData1, (char *)pStream->pStreamData2, &ssData);

									// ok, now fill second buffer
									ssData.userReference = pStream->reference;
									ssData.streamReference = (long)pStream;

									// now, push second pointer out for oversampling, and get fewer bytes for this buffer
									ssData.pData = (char *)pStream->pStreamData2 + (PV_GetSampleSizeInBytes(&ssData) * MAX_SAMPLE_OVERSAMPLE);
									ssData.dataLength = pStream->streamLength2 - MAX_SAMPLE_OVERSAMPLE;

									theErr = (*pProc)(context, STREAM_GET_DATA, &ssData);

									pStream->streamLength2 = ssData.dataLength;			// just in case it changes

									// update count of samples written
									pStream->samplesWritten += pStream->streamLength2;

									if (pStream->streamLength2 == 0)
									{	// underflow, get this buffer again
										pStream->streamUnderflow = TRUE;
									}
									else
									{
										pStream->startupBufferFullCount++;	// second buffer full
									}
								}
								else
								{
									pStream->streamLength2 = 0;
								}

								pStream->startupStatus = theErr;
								theErr = NO_ERR;	// we don't want to return a stop fail here

								// Ok, start the sample playback
								pStream->streamData = ssData;

								// $$kk: now that everything is initialized to real values, add the stream to the list
								PV_AddStream(pStream);
							}
						}
						if ( (theErr != NO_ERR) && (theErr != STREAM_STOP_PLAY) )
						{	// we've got to dispose of the data now
							ssData.userReference = pStream->reference;
							ssData.streamReference = (long)pStream;
							ssData.pData = pStream->pStreamBuffer;
							ssData.dataLength = pStream->streamBufferLength;
							ssData.sampleRate = sampleRate;
							ssData.dataBitSize = dataBitSize;
							ssData.channelSize = channelSize;
							pStream->streamCallback = NULL;
							(*pProc)(context, STREAM_DESTROY, &ssData);
						}
					}
				}
			}
			else
			{
				theErr = PARAM_ERR;
			}
		}
	}
	else
	{
		theErr = NO_FREE_VOICES;
	}

#if USE_DEBUG
	if (theErr)
	{
		DEBUG_STR_NULL("Exiting stream create with an error.\n");
	}
	else
	{
		DEBUG_STR_NULL("Exiting stream create NO error.\n");
	}
#endif
	if (pStream)
	{
		pStream->streamErr = theErr;
	}

	return reference;
}


OPErr GM_AudioStreamStart(long reference)
{
	GM_AudioStream		*pStream;
	OPErr				theErr;

	theErr = NO_ERR;
	pStream = PV_AudioStreamGetFromReference(reference);
	if (pStream)
	{
		// Ok, start the sample playback
		pStream->streamActive = TRUE;
		pStream->streamPaused = FALSE;
		pStream->streamFirstTime = TRUE;
		if (pStream->startupBufferFullCount == 0)
		{	// underflow on first buffer
			pStream->streamMode = STREAM_MODE_DEAD;
		}
		else
		{	// everythings ok, or at least the first buffer has data
			pStream->streamMode = STREAM_MODE_START_BUFFER_1;
			PV_StartThisBufferPlaying(pStream, STREAM_MODE_START_BUFFER_1);
		}
		reference = (long)pStream;
		if (pStream->startupStatus == STREAM_STOP_PLAY)
		{
			pStream->streamShuttingDown = TRUE;
			if (pStream->startupBufferFullCount == 1)
			{
				pStream->streamLength2 = 0;
			}
		}
	}
	else
	{
		theErr = PARAM_ERR;
	}
	return theErr;
}

long GM_AudioStreamGetReference(long reference)
{
	GM_AudioStream		*pStream;
	long				userReference;

	pStream = PV_AudioStreamGetFromReference(reference);
	if (pStream)
	{
		userReference = pStream->reference;
	}
	return userReference;
}

void GM_AudioStreamStopAll(void *context)
{
	GM_AudioStream	*pStream;

	pStream = theStreams;
	while (pStream)
	{
		if (pStream->streamActive)
		{
			GM_AudioStreamStop(context, (long)pStream);
		}
		pStream = pStream->pNext;
	}
}

// This will stop a streaming audio object.
//
// INPUT:
//	This is the reference number returned from AudioStreamStart.
//
static OPErr PV_AudioStreamStopAndFreeNow(void *context, long reference)
{
	GM_AudioStream		*pStream;
	GM_StreamData		ssData;
	short int			theErr;
	GM_StreamObjectProc	pProc;

	pStream = PV_AudioStreamGetFromReference(reference);
	if (pStream)
	{
		if (pStream->streamActive)
		{
			pStream->streamLength1 = 0;
			pStream->streamLength2 = 0;		// don't play next buffer.
			GM_EndSample(pStream->playbackReference);

// $$kk: 09.03.98: should not need this anymore		
/*			
			//  $$kk: 02.13.98: set playbackReference = -1 whenever we call GM_EndSample to avoid jmf stream crossing
			long playbackReference = pStream->playbackReference;

			pStream->streamLength1 = 0;
			pStream->streamLength2 = 0;		// don't play next buffer.

			pStream->playbackReference = -1;
			GM_EndSample(playbackReference);
*/

			pStream->streamActive = FALSE;
			pStream->streamShuttingDown = FALSE;
		}
		if (pStream->streamCallback)
		{
			ssData = pStream->streamData;

			ssData.userReference = pStream->reference;
			ssData.streamReference = (long)pStream;
			ssData.pData = pStream->pStreamBuffer;
			ssData.dataLength = pStream->streamBufferLength;
			pProc = pStream->streamCallback;
			pStream->streamCallback = NULL;		// prevent recursive callbacks
			theErr = (*pProc)(context, STREAM_DESTROY, &ssData);
			pStream->reference = -1;
		}
		PV_FreeStream(pStream);
	}
	return NO_ERR;
}

// This will stop a streaming audio object, during the next GM_AudioServiceStream
//
// INPUT:
//	This is the reference number returned from AudioStreamStart.
//
OPErr GM_AudioStreamStop(void *context, long reference)
{
/*
	PV_AudioStreamStopAndFreeNow(context, reference);
*/
/*
	GM_AudioStream		*pStream;

	pStream = PV_AudioStreamGetFromReference(reference);
	if (pStream)
	{
		// Set various flags to iniate a shutdown of the stream from inside GM_AudioStreamService
		pStream->streamMode = STREAM_MODE_INTERRUPT_ACTIVE | STREAM_MODE_STOP_STREAM;		// end

		// if audio data is already dead, then
		if (GM_IsSoundDone(pStream->playbackReference))
		{	// we've already flushed the audio data, so just free it now
			PV_AudioStreamStopAndFreeNow(context, (long)pStream);
			pStream->streamMode = STREAM_MODE_DEAD;
		}
	}
*/

// $$kk: 01.06.98
// want stream to stay in engine until all samples played.
// we make it stop playing and mark it to be freed; it should not actually
// get freed until all samples have been played.
//	PV_AudioStreamStopAndFreeNow(context, reference);

	GM_AudioStream		*pStream;

	pStream = PV_AudioStreamGetFromReference(reference);

	if (pStream)
	{
		if (pStream->streamActive)
		{
			pStream->streamLength1 = 0;
			pStream->streamLength2 = 0;		// don't play next buffer.
			GM_EndSample(pStream->playbackReference);

// $$kk: 09.03.98: should not need this anymore		
/*			
			//  $$kk: 02.13.98: set playbackReference = -1 whenever we call GM_EndSample to avoid jmf stream crossing
			long playbackReference = pStream->playbackReference;

			// stop current voice
			pStream->playbackReference = -1;

			pStream->streamLength1 = 0;
			pStream->streamLength2 = 0;		// don't play next buffer.

			GM_EndSample(playbackReference);
*/
		}

		pStream->streamShuttingDown = TRUE;
		pStream->streamMode = STREAM_MODE_INTERRUPT_ACTIVE | STREAM_MODE_FREE_STREAM;		// end
	}
	return NO_ERR;
}

// Get the file position of a audio stream, in samples
unsigned long GM_AudioStreamGetFileSamplePosition(long reference)
{
	GM_AudioStreamFileInfo	*pInfo;
	GM_AudioStream			*pStream;
	unsigned long			samplePosition;
	short int				blockAlign;

	samplePosition = 0;
	pStream = PV_AudioStreamGetFromReference(reference);
	if (pStream)
	{
		blockAlign = (pStream->streamData.dataBitSize / 8) / pStream->streamData.channelSize;
		pInfo = pStream->pFileStream;
		if (pInfo)
		{
			samplePosition = pInfo->filePlaybackPosition / blockAlign;	// convert from bytes to samples
		}
		else
		{
			samplePosition = pStream->streamPlaybackPosition;	// no conversion because its in samples
		}
		// now that we have the stream position, add in the playback of the current chunk of stream
		// for an actual position
		samplePosition += GM_GetSamplePlaybackPosition(pStream->playbackReference);
	}
	return samplePosition;
}


OPErr GM_AudioStreamGetData(void *context, long reference, unsigned long startFrame, unsigned long stopFrame,
									XPTR pBuffer, unsigned long bufferLength)
{
	GM_AudioStream	*pStream;
	OPErr			theErr;
	GM_StreamData	ssData;

	theErr = NOT_SETUP;
	pStream = PV_AudioStreamGetFromReference(reference);
	if (pStream && pBuffer)
	{
		if (pStream->streamCallback)
		{
			ssData = pStream->streamData;
			ssData.dataLength = bufferLength / ssData.channelSize / (ssData.dataBitSize / 8);
			ssData.pData = (char *)pBuffer;
			ssData.userReference = pStream->reference;
			ssData.streamReference = (long)pStream;
			ssData.startSample = startFrame;
			ssData.endSample = stopFrame;
			theErr = (*pStream->streamCallback)(context, STREAM_GET_SPECIFIC_DATA, &ssData);
		}
	}
	return theErr;
}


// Set the stereo position of a audio stream
void GM_AudioStreamSetStereoPosition(long reference, short int stereoPosition)
{
	GM_AudioStream	*pStream;

	pStream = PV_AudioStreamGetFromReference(reference);
	if (pStream)
	{
		pStream->streamStereoPosition = stereoPosition;
		GM_ChangeSampleStereoPosition(pStream->playbackReference, stereoPosition);
	}
}

// Get the stereo position of a audio stream
short int GM_AudioStreamGetStereoPosition(long reference)
{
	GM_AudioStream	*pStream;
	short int		stereoPosition;

	stereoPosition = 0;
	pStream = PV_AudioStreamGetFromReference(reference);
	if (pStream)
	{
		stereoPosition = pStream->streamStereoPosition;
	}
	return stereoPosition;
}

// Get the playback offset in samples for the stream.
// This is the offset between the number of samples processed
// by the mixer and the number of samples processed from this stream.
unsigned long GM_AudioStreamGetSampleOffset(long reference)
{
	GM_AudioStream	*pStream;
    unsigned long offset = 0;


	pStream = PV_AudioStreamGetFromReference(reference);
	if (pStream)
	{
        offset = pStream->streamPlaybackOffset;
    }
    return offset;
}

// Get the engine's count of samples from this stream actually
// played through the device.
unsigned long GM_AudioStreamGetSamplesPlayed(long reference)
{
	GM_AudioStream	*pStream;
    unsigned long samplesPlayed = 0;


	pStream = PV_AudioStreamGetFromReference(reference);
	if (pStream)
	{
        samplesPlayed = pStream->samplesPlayed;
    }

    return samplesPlayed;
}



// Drain this stream
void GM_AudioStreamDrain(long reference)
{
	GM_AudioStream	*pStream;
	unsigned long	samplesWritten;

	// get the samples written.
	// we have to drain until samples played reaches this value.
	pStream = PV_AudioStreamGetFromReference(reference);
	if (pStream)
	{
		if (pStream->streamActive)
		{
			samplesWritten = pStream->samplesWritten;
		}
		else
		{
			return;
		}
	}
	else
	{
		return;
	}

	//fprintf(stderr, "samplesWritten = %d\n", samplesWritten);

	while(1)
	{
		pStream = PV_AudioStreamGetFromReference(reference);
		if (pStream)
		{
			if (pStream->streamActive)
			{
				unsigned newSamplesWritten;

				newSamplesWritten = pStream->samplesWritten;
				if (newSamplesWritten < samplesWritten)
				{
					samplesWritten =  newSamplesWritten;
				}

				if (pStream->samplesPlayed >= samplesWritten)
				{
					// done!
					//fprintf(stderr, "GM_AudioStreamDrain: played all samples\n");
					return;
				}
			}
			else
			{
				// stream no longer active
				//fprintf(stderr, "GM_AudioStreamDrain: stream no longer active\n");
				return;
			}
		}
		else
		{
			// stream no longer valid
			//fprintf(stderr, "GM_AudioStreamDrain: stream no longer valid\n");
			return;
		}

		// if we get here, the stream is still valid and active and has samples still to play.
		// wait 10 milliseconds
		XWaitMicroseocnds(10000);
	}
}



// Flush this stream
void GM_AudioStreamFlush(long reference)
{
	GM_AudioStream	*pStream;

	pStream = PV_AudioStreamGetFromReference(reference);
	if (pStream)
	{
		if (pStream->streamActive)
		{
			GM_EndSample(pStream->playbackReference);

// $$kk: 09.03.98: should not need this anymore		
/*			
			//  $$kk: 02.13.98: set playbackReference = -1 whenever we call GM_EndSample to avoid jmf stream crossing
			long playbackReference = pStream->playbackReference;

			// stop current voice
			pStream->playbackReference = -1;
			GM_EndSample(playbackReference);
*/
			// set to a dead voice, which will force a new complete buffer read before starting next buffer

			if (pStream->streamShuttingDown == TRUE)
			{
				pStream->streamMode = STREAM_MODE_STOP_STREAM;
			}

			else
			{
				pStream->streamMode = STREAM_MODE_DEAD;
			}

			pStream->streamUnderflow = TRUE;	// induce underflow
			pStream->streamFlushed = TRUE;

			// here we reduce the count of samples written to the number which we will
			// actually end up playing.  (the difference between the old and new values
			// is the amount of data flushed!)
			pStream->samplesWritten = GM_AudioStreamGetFileSamplePosition(reference);
        }
    }
}


// Set the volume level of a audio stream
void GM_AudioStreamSetVolume(long reference, short int newVolume)
{
	GM_AudioStream	*pStream;

	pStream = PV_AudioStreamGetFromReference(reference);
	if (pStream)
	{
		pStream->streamVolume = newVolume;
		GM_ChangeSampleVolume(pStream->playbackReference, newVolume);
	}
}

// Get the volume level of a audio stream
short int GM_AudioStreamGetVolume(long reference)
{
	GM_AudioStream	*pStream;
	short int		volume;

	volume = 0;
	pStream = PV_AudioStreamGetFromReference(reference);
	if (pStream)
	{
		volume = pStream->streamVolume;
	}
	return volume;
}

// set the volume level of all open streams. Scale is 0 to MAX_NOTE_VOLUME. If you pass
// -1, then it will reset all volumes. This is used to grab the master volume changes.
void GM_AudioStreamSetVolumeAll(short int newVolume)
{
	GM_AudioStream	*pStream;
	short int		thisVolume;

	pStream = theStreams;
	while (pStream)
	{
		if (newVolume == -1)
		{
			thisVolume = GM_GetSampleVolumeUnscaled(pStream->playbackReference);
		}
		else
		{
			thisVolume = newVolume;
		}
		pStream->streamVolume = thisVolume;
		GM_ChangeSampleVolume(pStream->playbackReference, thisVolume);
		pStream = pStream->pNext;
	}
}


// Set stream fade rate. Its a 16.16 fixed value
// Input:	reference	stream to affect
//			fadeRate	amount to change every 11 ms
//						example:	FLOAT_TO_FIXED(2.2) will decrease volume
//									FLOAT_TO_FIXED(2.2) * -1 will increase volume
//			minVolume	lowest volume level fade will go
//			maxVolume	highest volume level fade will go
void GM_SetAudioStreamFadeRate(long reference, FIXED_VALUE fadeRate,
							INT16 minVolume, INT16 maxVolume, XBOOL endStream)
{
	GM_AudioStream	*pStream;

	pStream = PV_AudioStreamGetFromReference(reference);
	if (pStream)
	{
		pStream->streamFixedVolume = LONG_TO_FIXED(pStream->streamVolume);
		pStream->streamFadeMaxVolume = maxVolume;
		pStream->streamFadeMinVolume = minVolume;
		pStream->streamEndAtFade = endStream;
		pStream->streamFadeRate = fadeRate;

	}
}

// Enable/Disable reverb on this particular audio stream
void GM_AudioStreamReverb(long reference, XBOOL useReverb)
{
	GM_AudioStream	*pStream;

	pStream = PV_AudioStreamGetFromReference(reference);
	if (pStream)
	{
		GM_ChangeSampleReverb(pStream->playbackReference, useReverb);
	}
}

XBOOL GM_AudioStreamGetReverb(long reference)
{
	GM_AudioStream	*pStream;
	XBOOL			verb;

	verb = FALSE;
	pStream = PV_AudioStreamGetFromReference(reference);
	if (pStream)
	{
		verb = GM_GetSampleReverb(pStream->playbackReference);
	}
	return verb;
}

// Set the sample rate of a audio stream
void GM_AudioStreamSetRate(long reference, unsigned long newRate)
{
	GM_AudioStream	*pStream;

	pStream = PV_AudioStreamGetFromReference(reference);
	if (pStream)
	{
		pStream->streamData.sampleRate = newRate;
		GM_ChangeSamplePitch(pStream->playbackReference, newRate);
	}
}


// Get the sample rate of a audio stream
unsigned long GM_AudioStreamGetRate(long reference)
{
	GM_AudioStream	*pStream;
	unsigned long	rate;

	rate = 0;
	pStream = PV_AudioStreamGetFromReference(reference);
	if (pStream)
	{
		rate = pStream->streamData.sampleRate;
	}
	return rate;
}

// Returns TRUE or FALSE if a given GM_AudioStream is still active
XBOOL GM_IsAudioStreamPlaying(long reference)
{
	GM_AudioStream	*pStream;
	XBOOL			active;


	active = FALSE;
	pStream = PV_AudioStreamGetFromReference(reference);
	if (pStream)
	{
		if (GM_IsSoundDone(pStream->playbackReference) == FALSE)
		{
			if (pStream->streamActive)
			{
				active = TRUE;
			}
		}
	}
	return active;
}

// Returns TRUE or FALSE if a given reference is still valid
XBOOL GM_IsAudioStreamValid(long reference)
{
	return (PV_AudioStreamGetFromReference(reference)) ? TRUE : FALSE;
}

// Process any fading streams
void PV_ServeStreamFades(void)
{
	GM_AudioStream	*pStream;
	long			value;

	pStream = theStreams;
	while (pStream)
	{
		if ((pStream->streamActive) && (pStream->streamPaused == FALSE))
		{
			if (pStream->streamFadeRate)
			{
				pStream->streamFixedVolume -= pStream->streamFadeRate;
				value = FIXED_TO_LONG(pStream->streamFixedVolume);
				if (value > pStream->streamFadeMaxVolume)
				{
					value = pStream->streamFadeMaxVolume;
					pStream->streamFadeRate = 0;
				}
				if (value < pStream->streamFadeMinVolume)
				{
					value = pStream->streamFadeMinVolume;
					pStream->streamFadeRate = 0;
				}
				pStream->streamVolume = value;

				GM_ChangeSampleVolume((INT32)pStream->playbackReference, (INT16)value);

				if ((pStream->streamFadeRate == 0) && pStream->streamEndAtFade)
				{
					GM_EndSample(pStream->playbackReference);

// $$kk: 09.03.98: should not need this anymore	
/*					
					//  $$kk: 02.13.98: set playbackReference = -1 whenever we call GM_EndSample to avoid jmf stream crossing
					long playbackReference = pStream->playbackReference;

					pStream->playbackReference = -1;
					GM_EndSample(playbackReference);
*/
					pStream->streamMode = STREAM_MODE_INTERRUPT_ACTIVE | STREAM_MODE_FREE_STREAM;		// end
				}
			}
		}
		pStream = pStream->pNext;
	}
}


// This is the streaming audio service routine. Call this as much as possible, but not during an
// interrupt. This is a very quick routine. A good place to call this is in your main event loop.
void GM_AudioStreamService(void *context)
{
	GM_AudioStream		*pStream, *pNext;
	GM_StreamData		ssData;
	GM_StreamObjectProc	theProc;
	XBOOL				done;
	OPErr				theErr;

	pStream = theStreams;
	while (pStream)
	{
		pNext = pStream->pNext;
		if ((pStream->streamActive) && (pStream->streamPaused == FALSE))
		{
			done = GM_IsSoundDone(pStream->playbackReference);
			// voice has shutdown, either from a voice setup change or some other problem, restart
			// the stream from the last place that we're aware of.
			if ( (done) && (pStream->streamShuttingDown == FALSE) )
			{
				pStream->streamMode = STREAM_MODE_DEAD;
			}
		}

		if ( ((pStream->streamActive) &&
			  (pStream->streamMode & STREAM_MODE_INTERRUPT_ACTIVE) &&
			  (pStream->streamPaused == FALSE)) ||
			   (pStream->streamUnderflow && (pStream->streamPaused == FALSE)))
		{
			theProc = pStream->streamCallback;
			if (theProc)
			{
				pStream->streamMode &= (~STREAM_MODE_INTERRUPT_ACTIVE);
				ssData = pStream->streamData;
				switch (pStream->streamMode)
				{
					default:
						DEBUG_STR_NULL("Bad case in GM_AudioStreamService");
						break;
					case STREAM_MODE_FREE_STREAM:
						DEBUG_STR_NULL("GM_AudioStreamService::STREAM_MODE_FREE_STREAM");

// $$kk: 02.06.98
// stream should stay in engine until all samples played.

						//fprintf(stderr, "GM_AudioStreamService::STREAM_MODE_FREE_STREAM %d\n", pStream);
						//fprintf(stderr, "pStream->samplesPlayed: %d, sample position: %d \n", pStream->samplesPlayed, GM_AudioStreamGetFileSamplePosition((long)pStream));

						// figure out whether all the samples have been played out through the device
						// before freeing the stream
						if (pStream->samplesPlayed >= GM_AudioStreamGetFileSamplePosition((long)pStream))
						{
							// This will free the stream
							//fprintf(stderr, "freeing stream %d\n", pStream);
							PV_AudioStreamStopAndFreeNow(context, (long)pStream);
						}
						else
						{
							// keep coming back through this loop to check again
							pStream->streamMode |= (STREAM_MODE_INTERRUPT_ACTIVE);
						}

//						// This will free the stream
//						PV_AudioStreamStopAndFreeNow(context, (long)pStream);
						break;
					case STREAM_MODE_STOP_STREAM:

// $$kk: 02.06.98
// stream should stay in engine until all samples played.

						//fprintf(stderr, "GM_AudioStreamService::STREAM_MODE_STOP_STREAM %d\n", pStream);
						GM_AudioStreamStop(context, (long)pStream);

/*
						if (pStream->streamFlushed)
						{	// we've already flushed the audio data, so just free it now
							PV_AudioStreamStopAndFreeNow(context, (long)pStream);
						}
						// otherwise we do nothing, because the audio callback will set the streamMode to STREAM_MODE_FREE_STREAM
						// when the samples are finished playing
*/
						break;
					case STREAM_MODE_DEAD:	// first buffer failed to fill
						DEBUG_STR_NULL("GM_AudioStreamService::STREAM_MODE_DEAD");
						ssData.dataLength = pStream->streamOrgLength1;
						ssData.pData = pStream->pStreamData1;
						ssData.userReference = pStream->reference;
						ssData.streamReference = (long)pStream;
						if ((*theProc)(context, STREAM_GET_DATA, &ssData) != NO_ERR)
						{
							DEBUG_STR_NULL("    STOP!");
							pStream->streamShuttingDown = TRUE;
							pStream->streamLength2 = 0;
							PV_FillBufferEndWithSilence((char *)ssData.pData, &ssData);
						}
						pStream->streamLength1 = ssData.dataLength;			// just in case it changes
						if (ssData.dataLength < MAX_SAMPLE_OVERSAMPLE)
						{
							ssData.dataLength += MAX_SAMPLE_OVERSAMPLE;		// going to click for sure
						}
						else
						{
							ssData.dataLength -= MAX_SAMPLE_OVERSAMPLE;
						}
						// copy end of buffer 1 into the start of buffer 2
						PV_CopyLastSamplesToFirst((char *)pStream->pStreamData1, (char *)pStream->pStreamData2, &ssData);

						if (pStream->streamLength1 == 0)
						{	// underflow, get this buffer again
							pStream->streamUnderflow = TRUE;
						}
						else
						{

							// update count of samples written
							pStream->samplesWritten += pStream->streamLength1;

							// ok, now fill second buffer
							ssData.userReference = pStream->reference;
							ssData.streamReference = (long)pStream;

							// now, push second pointer out for oversampling, and get fewer bytes for this buffer
							ssData.pData = (char *)pStream->pStreamData2 + (PV_GetSampleSizeInBytes(&ssData) * MAX_SAMPLE_OVERSAMPLE);
							ssData.dataLength = pStream->streamOrgLength2 - MAX_SAMPLE_OVERSAMPLE;

							theErr = (*theProc)(context, STREAM_GET_DATA, &ssData);
							pStream->streamLength2 = ssData.dataLength;			// just in case it changes
							if (pStream->streamLength2 == 0)
							{	// underflow, get this buffer again
								pStream->streamUnderflow = TRUE;
							}
							else
							{
								// update count of samples written
								pStream->samplesWritten += pStream->streamLength2;

								pStream->streamUnderflow = FALSE;
							}

							pStream->streamMode = STREAM_MODE_START_BUFFER_1;
							done = GM_IsSoundDone(pStream->playbackReference);
							if (done)
							{
								PV_StartThisBufferPlaying(pStream, STREAM_MODE_START_BUFFER_1);
							}
							pStream->streamData = ssData;
						}
						break;
					case STREAM_MODE_START_BUFFER_2:		// read buffer 1 into memory
						DEBUG_STR_NULL("GM_AudioStreamService::STREAM_MODE_START_BUFFER_2");
						if (pStream->streamShuttingDown == FALSE)
						{
							ssData.dataLength = pStream->streamOrgLength1 - MAX_SAMPLE_OVERSAMPLE;
							ssData.pData = (char *)pStream->pStreamData1 + (PV_GetSampleSizeInBytes(&ssData) * MAX_SAMPLE_OVERSAMPLE);
							ssData.userReference = pStream->reference;
							ssData.streamReference = (long)pStream;
							if ((*theProc)(context, STREAM_GET_DATA, &ssData) != NO_ERR)
							{
								DEBUG_STR_NULL("    STOP!");
								pStream->streamShuttingDown = TRUE;
								pStream->streamLength2 = 0;
								PV_FillBufferEndWithSilence((char *)ssData.pData, &ssData);
							}
							pStream->streamLength1 = ssData.dataLength;			// just in case it changes
							// copy end of buffer 1 into the start of buffer 2
							PV_CopyLastSamplesToFirst((char *)pStream->pStreamData1, (char *)pStream->pStreamData2, &ssData);

                            if (pStream->streamLength1 == 0)
							{	// underflow, get this buffer again
								pStream->streamUnderflow = TRUE;
							}
							else
							{
								// update count of samples written
								pStream->samplesWritten += pStream->streamLength1;

								// did our last buffer underflow?
								if (pStream->streamUnderflow)
								{
									done = GM_IsSoundDone(pStream->playbackReference);
									// voice has shutdown from running out of data, restart previous buffer.
									if (done)
									{
										PV_StartThisBufferPlaying(pStream, STREAM_MODE_START_BUFFER_1);

									}
								}
								pStream->streamUnderflow = FALSE;
							}
						}
						#if USE_DEBUG && 0
						{
							char text[256];

							sprintf(text, "B1-> %ld len %ld", pStream->streamPlaybackPosition, pStream->streamLength1);
							DEBUG_STR(text);
						}
						#endif
						break;
					case STREAM_MODE_START_BUFFER_1:		// read buffer 2 into memory
						DEBUG_STR_NULL("GM_AudioStreamService::STREAM_MODE_START_BUFFER_1");
						if (pStream->streamShuttingDown == FALSE)
						{
							ssData.dataLength = pStream->streamOrgLength2 - MAX_SAMPLE_OVERSAMPLE;
							ssData.pData = (char *)pStream->pStreamData2 + (PV_GetSampleSizeInBytes(&ssData) * MAX_SAMPLE_OVERSAMPLE);
							ssData.userReference = pStream->reference;
							ssData.streamReference = (long)pStream;
							if ((*theProc)(context, STREAM_GET_DATA, &ssData) != NO_ERR)
							{
								DEBUG_STR_NULL("    STOP!");
								pStream->streamShuttingDown = TRUE;
								pStream->streamLength1 = 0;
								PV_FillBufferEndWithSilence((char *)ssData.pData, &ssData);
							}
							pStream->streamLength2 = ssData.dataLength;
							// copy end of buffer 2 into the start of buffer 1
							PV_CopyLastSamplesToFirst((char *)pStream->pStreamData2, (char *)pStream->pStreamData1, &ssData);

							if (pStream->streamLength2 == 0)
							{	// underflow, get this buffer again
								pStream->streamUnderflow = TRUE;
							}
							else
							{
								// update count of samples written
								pStream->samplesWritten += pStream->streamLength2;

								// did our last buffer underflow?
								if (pStream->streamUnderflow)
								{
									done = GM_IsSoundDone(pStream->playbackReference);
									// voice has shutdown from running out of data, restart previous buffer.
									if (done)
									{
										PV_StartThisBufferPlaying(pStream, STREAM_MODE_START_BUFFER_2);
									}
								}
								pStream->streamUnderflow = FALSE;
							}
						}
						#if USE_DEBUG && 0
						{
							char text[256];

							sprintf(text, "B2-> %ld len %ld", pStream->streamPlaybackPosition, pStream->streamLength2);
							DEBUG_STR(text);
						}
						#endif
						break;
				}
			}
		}
		pStream = pNext;
	}
}

// update number of samples played for each stream
// delta is number of samples engine advanced, in its format
void GM_AudioStreamUpdateSamplesPlayed(long delta)
{
	GM_AudioStream		*pStream, *pNext;
    long                streamDelta;    // delta in stream-format samples
	long				samplesCommitted = 0;

	pStream = theStreams;

	while (pStream)
	{

		pNext = pStream->pNext;

        // convert from engine samples to stream samples
        // $$kk: HELP!! what is going on?  LONG_TO_FIXED(x) DOES *NOT* WORK HERE!
        streamDelta = (long)( ((float) ((long)(pStream->streamData.sampleRate / 65536.0 )) ) / ((float)(MusicGlobals->outputQuality == Q_22K ? 22050 : 44100))  * ((float)delta) );

		// we can't count higher than the number of samples processed by the engine for this stream
		samplesCommitted = GM_AudioStreamGetFileSamplePosition((long)pStream);

		// $$kk: 03.05.98: at the end of the stream, GM_GetSamplePlaybackPosition remains at its last (highest) value even though
		// pStream->streamPlaybackPosition already equals samplesWritten, so GM_GetSamplePlaybackPosition returns an impossibly high
		// value.  make sure we don't use it....
		if (samplesCommitted > pStream->samplesWritten)
				samplesCommitted = pStream->samplesWritten;

        // update number of samples played

        // do we know our offset, and are we in play range?
        if ((pStream->streamPlaybackOffset != -1) && (MusicGlobals->samplesPlayed > pStream->streamPlaybackOffset))
        {
            // we need to handle the case where the stream may just be starting up again
            if ((MusicGlobals->samplesPlayed - pStream->streamPlaybackOffset) < streamDelta)
            {
                streamDelta = MusicGlobals->samplesPlayed - pStream->streamPlaybackOffset;
            }

            // are we incrementing normally?
            if ((pStream->samplesPlayed + streamDelta) <= samplesCommitted)
            {
                pStream->samplesPlayed += streamDelta;
            }
            // no, actual underflow is occurring
            else
            {
                pStream->samplesPlayed = samplesCommitted;
			}
		}
		else
		{
			if (pStream->residualSamples > 0)
			{
				if (pStream->residualSamples < streamDelta)
				{
					streamDelta = pStream->residualSamples;
				}
				if (streamDelta > (samplesCommitted - pStream->samplesPlayed))
				{
					streamDelta = samplesCommitted - pStream->samplesPlayed;
				}
				pStream->samplesPlayed += streamDelta;
				pStream->residualSamples -= streamDelta;
			}
		}


// $$kk: 09.03.98: should not need this anymore	
/*					
		// 03.09.98: this is a *hack* to get around the stream crossing that can happen if a stream plays out and
		// its id becomes available to another stream, but the first one still holds a reference to it.

		if ( GM_IsSoundDone(pStream->playbackReference) )
		{
			pStream->playbackReference = -1;
		}
*/

		pStream = pNext;
	} // while
}


#endif	// USE_STREAM_API

// EOF of GenAudioStreams.c

