/*
 * @(#)GenSoundFiles.c	1.8 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
/*****************************************************************************/
/*
**	GenSoundFiles.c
**
**	Reads AIFF, WAVE, Sun AU sound files
**
**
**	History	-
**	10/1/96		Created
**	10/13/96	Added GM_ReadIntoMemoryWaveFile & GM_ReadIntoMemoryAIFFFile
**	11/5/96		Changed WaveformInfo to GM_Waveform
**				Added more error checking
**				Added GM_FreeWaveform
**	12/30/96	Changed copyright
**	1/20/97		Added support for Sun AU files
**	1/23/97		More memory fail checking code
**	1/29/97		Changed PV_ConvertFromIeeeExtended to return a FIXED_VALUE
**				instead of a double
**	2/4/97		Fixed problem with PV_ReadSunAUFile. Didn't define the procptr
**	2/5/97		Added GM_ReadFileIntoMemoryFromMemory with the ability to parse a
**				memory mapped file as a fileType
**	3/19/97		Fixed a positioning and platform word swap bug 
**				with PV_ReadIntoMemorySunAUFile
**	3/29/97		Removed 4 character constants and changed to use macro FOUR_CHAR
**	5/3/97		Fixed a few potential problems that the MOT compiler found
**	5/15/97		Put in code for ADPCM decoding for WAVE files & AIFF files
**	6/2/97		Finished up GM_ReadAndDecodeFileStream to support AU streaming files
**	6/9/97		Added support for X_WAVE_FORMAT_ALAW & X_WAVE_FORMAT_MULAW for wave files
**	7/8/97		Fixed divide by zero when reading WAVE files.
**	7/25/97		Added PV_ConvertToIeeeExtended
**	11/10/97	Changed some preprocessor tests and flags to explicity test for flags rather
**				than assume
*/
/*****************************************************************************/

//#define USE_DEBUG	2

#include "X_API.h"
#include "X_Formats.h"
#include "GenSnd.h"
#include <math.h>

#if USE_HIGHLEVEL_FILE_API == TRUE

#define ODD(x)			((long)(x) & 1L)

#if CPU_TYPE == kRISC
	#pragma options align=mac68k
#endif
#if ((CPU_TYPE == k80X86) || (CPU_TYPE == kSPARC))
	#pragma pack (1)
#endif

/**********************- WAVe Defines -**************************/
//
//  extended waveform format structure used for all non-PCM formats. this
//  structure is common to all non-PCM formats.
//
typedef struct
{
    XWORD		wFormatTag;			/* format type */
    XWORD		nChannels;			/* number of channels (i.e. mono, stereo...) */
    XDWORD		nSamplesPerSec;		/* sample rate */
    XDWORD		nAvgBytesPerSec;	/* for buffer estimation */
    XWORD		nBlockAlign;		/* block size of data */
    XWORD		wBitsPerSample;		/* number of bits per sample of mono data */
    XWORD		cbSize;				/* the count in bytes of the size of */
									/* extra information (after cbSize) */
} XWaveHeader;


//
//  IMA endorsed ADPCM structure definitions--note that this is exactly
//  the same format as Intel's DVI ADPCM.
//
//      for WAVE_FORMAT_IMA_ADPCM   (0x0011)
//
//
typedef struct 
{
        XWaveHeader		wfx;
        XWORD			wSamplesPerBlock;
} XWaveHeaderIMA;


// WAVE form wFormatTag IDs

// supported 
enum 
{
	X_WAVE_FORMAT_PCM					=	0x0001,
	X_WAVE_FORMAT_ALAW					=	0x0006,	/*  Microsoft Corporation  */
	X_WAVE_FORMAT_MULAW					=	0x0007,	/*  Microsoft Corporation  */
	X_WAVE_FORMAT_DVI_ADPCM				=	0x0011,	/*  Intel Corporation  */
	X_WAVE_FORMAT_IMA_ADPCM				=	0x0011	/*  Intel Corporation  */
};

// not supported yet.
enum 
{
	X_WAVE_FORMAT_UNKNOWN				=	0x0000,	/*  Microsoft Corporation  */
	X_WAVE_FORMAT_ADPCM					=	0x0002,	/*  Microsoft Corporation  */
	X_WAVE_FORMAT_IBM_CVSD				=	0x0005,	/*  IBM Corporation  */
	X_WAVE_FORMAT_OKI_ADPCM				=	0x0010,	/*  OKI  */
	X_WAVE_FORMAT_MEDIASPACE_ADPCM		=	0x0012,	/*  Videologic  */
	X_WAVE_FORMAT_SIERRA_ADPCM			=	0x0013,	/*  Sierra Semiconductor Corp  */
	X_WAVE_FORMAT_G723_ADPCM			=	0x0014,	/*  Antex Electronics Corporation  */
	X_WAVE_FORMAT_DIGISTD				=	0x0015,	/*  DSP Solutions, Inc.  */
	X_WAVE_FORMAT_DIGIFIX				=	0x0016,	/*  DSP Solutions, Inc.  */
	X_WAVE_FORMAT_DIALOGIC_OKI_ADPCM	=	0x0017,	/*  Dialogic Corporation  */
	X_WAVE_FORMAT_YAMAHA_ADPCM			=	0x0020,	/*  Yamaha Corporation of America  */
	X_WAVE_FORMAT_SONARC				=	0x0021,	/*  Speech Compression  */
	X_WAVE_FORMAT_DSPGROUP_TRUESPEECH	=	0x0022,	/*  DSP Group, Inc  */
	X_WAVE_FORMAT_ECHOSC1				=	0x0023,	/*  Echo Speech Corporation  */
	X_WAVE_FORMAT_AUDIOFILE_AF36		=	0x0024,	/*    */
	X_WAVE_FORMAT_APTX					=	0x0025,	/*  Audio Processing Technology  */
	X_WAVE_FORMAT_AUDIOFILE_AF10		=	0x0026,	/*    */
	X_WAVE_FORMAT_DOLBY_AC2				=	0x0030,	/*  Dolby Laboratories  */
	X_WAVE_FORMAT_GSM610				=	0x0031,	/*  Microsoft Corporation  */
	X_WAVE_FORMAT_ANTEX_ADPCME			=	0x0033,	/*  Antex Electronics Corporation  */
	X_WAVE_FORMAT_CONTROL_RES_VQLPC		=	0x0034,	/*  Control Resources Limited  */
	X_WAVE_FORMAT_DIGIREAL				=	0x0035,	/*  DSP Solutions, Inc.  */
	X_WAVE_FORMAT_DIGIADPCM				=	0x0036,	/*  DSP Solutions, Inc.  */
	X_WAVE_FORMAT_CONTROL_RES_CR10		=	0x0037,	/*  Control Resources Limited  */
	X_WAVE_FORMAT_NMS_VBXADPCM			=	0x0038,	/*  Natural MicroSystems  */
	X_WAVE_FORMAT_CS_IMAADPCM			=	0x0039,	/* Crystal Semiconductor IMA ADPCM */
	X_WAVE_FORMAT_G721_ADPCM			=	0x0040,	/*  Antex Electronics Corporation  */
	X_WAVE_FORMAT_MPEG					=	0x0050,	/*  Microsoft Corporation  */
	X_WAVE_FORMAT_CREATIVE_ADPCM		=	0x0200,	/*  Creative Labs, Inc  */
	X_WAVE_FORMAT_CREATIVE_FASTSPEECH8	=	0x0202,	/*  Creative Labs, Inc  */
	X_WAVE_FORMAT_CREATIVE_FASTSPEECH10	=	0x0203,	/*  Creative Labs, Inc  */
	X_WAVE_FORMAT_FM_TOWNS_SND			=	0x0300,	/*  Fujitsu Corp.  */
	X_WAVE_FORMAT_OLIGSM				=	0x1000,	/*  Ing C. Olivetti & C., S.p.A.  */
	X_WAVE_FORMAT_OLIADPCM				=	0x1001,	/*  Ing C. Olivetti & C., S.p.A.  */
	X_WAVE_FORMAT_OLICELP				=	0x1002,	/*  Ing C. Olivetti & C., S.p.A.  */
	X_WAVE_FORMAT_OLISBC				=	0x1003,	/*  Ing C. Olivetti & C., S.p.A.  */
	X_WAVE_FORMAT_OLIOPR				=	0x1004	/*  Ing C. Olivetti & C., S.p.A.  */
};

enum 
{
	X_WAVE			= FOUR_CHAR('W','A','V','E'),		//		'WAVE'
	X_RIFF			= FOUR_CHAR('R','I','F','F'),		//		'RIFF'
	X_DATA			= FOUR_CHAR('d','a','t','a'),		//		'data'
	X_FMT			= FOUR_CHAR('f','m','t',' ')		//		'fmt '
};

/**********************- AIFF Defines -**************************/

enum 
{
	X_AIFF			= FOUR_CHAR('A','I','F','F'),		//		'AIFF'
	X_AIFC			= FOUR_CHAR('A','I','F','C'),		//		'AIFC'
	X_Common		= FOUR_CHAR('C','O','M','M'),		//		'COMM'
	X_SoundData		= FOUR_CHAR('S','S','N','D'),		//		'SSND'
	X_Marker		= FOUR_CHAR('M','A','R','K'),		//		'MARK'
	X_Instrument	= FOUR_CHAR('I','N','S','T'),		//		'INST'
	X_FORM			= FOUR_CHAR('F','O','R','M'),		//		'FORM'
	X_CAT			= FOUR_CHAR('C','A','T',' '),		//		'CAT '
	X_LIST			= FOUR_CHAR('L','I','S','T'),		//		'LIST'
	X_BODY			= FOUR_CHAR('B','O','D','Y')		//		'BODY'
};

struct XIFFChunk
{
	long	ckID;	   /* ID */
	long	ckSize;    /* size */
	long	ckData;
};
typedef struct XIFFChunk XIFFChunk;


struct XAIFFHeader
{
	short int		numChannels;
	unsigned long	numSampleFrames;
	short int		sampleSize;
	unsigned char	sampleRate[10];
};
typedef struct XAIFFHeader XAIFFHeader;


struct XAIFFExtenedHeader
{
	short int		numChannels;
	unsigned long	numSampleFrames;
	short int		sampleSize;
	unsigned char	sampleRate[10];
	unsigned long	compressionType;
	char			compressionName[256];			/* variable length array, Pascal string */
};
typedef struct XAIFFExtenedHeader XAIFFExtenedHeader;

struct XInstrumentHeader
{
	unsigned char	baseFrequency;
	unsigned char	detune;
	unsigned char	lowFrequency;
	unsigned char	highFrequency;
	unsigned char	lowVelocity;
	unsigned char	highVelocity;
	short int		gain;

	short int		sustainLoop_playMode;
	short int		sustainLoop_beginLoop;
	short int		sustainLoop_endLoop;
	short int		releaseLoop_beginLoop;
	short int		releaseLoop_endLoop;
	short int		extra;
};
typedef struct XInstrumentHeader XInstrumentHeader;

struct XSoundData
{
	unsigned long	offset;
	unsigned long	blockSize;
};
typedef struct XSoundData XSoundData;

#if CPU_TYPE == kRISC
	#pragma options align=reset
#endif
#if ((CPU_TYPE == k80X86) || (CPU_TYPE == kSPARC))
	#pragma pack ()
#endif

/**********************- AU Defines -**************************/

#if CPU_TYPE == kRISC
	#pragma options align=mac68k
#endif
#if CPU_TYPE == k80X86
	#pragma pack (1)
#endif
	
#include "g72x.h"

typedef struct 
{
	unsigned long magic;          // magic number 
	unsigned long hdr_size;       // size of the whole header, including optional comment.
	unsigned long data_size;      // optional data size - usually unusalble. 
	unsigned long encoding;       // format of data contained in this file 
	unsigned long sample_rate;    // sample rate of data in this file 
	unsigned long channels;       // numbder of interleaved channels (usually 1 or 2) 
} SunAudioFileHeader;


// Note these are defined for big-endian architectures 
#define SUN_AUDIO_FILE_MAGIC_NUMBER (0x2E736E64L)

// Define the encoding fields 
#define	SUN_AUDIO_FILE_ENCODING_MULAW_8			(1)		// 8-bit ISDN u-law 
#define	SUN_AUDIO_FILE_ENCODING_LINEAR_8		(2)		// 8-bit linear PCM 
#define	SUN_AUDIO_FILE_ENCODING_LINEAR_16		(3)		// 16-bit linear PCM 
#define	SUN_AUDIO_FILE_ENCODING_LINEAR_24		(4)		// 24-bit linear PCM 
#define	SUN_AUDIO_FILE_ENCODING_LINEAR_32		(5)		// 32-bit linear PCM 
#define	SUN_AUDIO_FILE_ENCODING_FLOAT			(6)		// 32-bit IEEE floating point 
#define	SUN_AUDIO_FILE_ENCODING_DOUBLE			(7)		// 64-bit IEEE floating point 
#define	SUN_AUDIO_FILE_ENCODING_ADPCM_G721		(23)	// 4-bit CCITT g.721 ADPCM 
#define	SUN_AUDIO_FILE_ENCODING_ADPCM_G722		(24)	// CCITT g.722 ADPCM 
#define	SUN_AUDIO_FILE_ENCODING_ADPCM_G723_3	(25)	// CCITT g.723 3-bit ADPCM 
#define	SUN_AUDIO_FILE_ENCODING_ADPCM_G723_5	(26)	// CCITT g.723 5-bit ADPCM 
#define SUN_AUDIO_FILE_ENCODING_ALAW_8			(27) 	// 8-bit ISDN A-law 


#if CPU_TYPE == kRISC
	#pragma options align=reset
#endif
#if CPU_TYPE == k80X86
	#pragma pack ()
#endif

// internal structures
struct X_IFF
{
	long			formType;				//	either X_RIFF, or X_FORM;
	long			headerType;
	long			formPosition;			//	position in file of current FORM
	long			formLength;				//	length of FORM
	long			lastError;
	XFILE			fileReference;
};
typedef struct X_IFF X_IFF;



// IFF functions

static long IFF_Error(X_IFF *pIFF)
{
	if (pIFF)
	{
		return pIFF->lastError;
	}
	return NO_ERR;
}

static void IFF_SetFormType(X_IFF *pIFF, long formType)
{
	if (pIFF)
	{
		pIFF->formType = formType;
	}
}


/* -1 is error */
static long IFF_GetNextGroup(X_IFF *pIFF, XIFFChunk *pChunk)
{
	long err;

	err = XFileRead(pIFF->fileReference, pChunk, (long)sizeof(XIFFChunk) - sizeof(long));	// get chunk ID
	if (err != -1)
	{
		if (pIFF->formType == X_RIFF)
		{
			#if X_WORD_ORDER == FALSE	// motorola?
			pChunk->ckSize = XSwapLong(pChunk->ckSize);
			#endif
		}
		else
		{
			pChunk->ckSize = XGetLong(&pChunk->ckSize);
		}
		pChunk->ckData = 0L;
		pChunk->ckID = XGetLong(&pChunk->ckID);		// swap if not motorola
		switch(pChunk->ckID)
		{
			case X_FORM:
			case X_RIFF:
			case X_LIST:
			case X_CAT:
				pIFF->formPosition = XFileGetPosition(pIFF->fileReference);	 /* get current pos */
				pIFF->formLength = pChunk->ckSize;

				XFileRead(pIFF->fileReference, &pChunk->ckData, (long)sizeof(long));
				pChunk->ckData = XGetLong(&pChunk->ckData);
				break;
		}
	}
	pIFF->lastError = err;
	return err;
}


/******- Determine if a file is a IFF type -******************************/

static long IFF_FileType(X_IFF *pIFF)
{
	XIFFChunk type;

	if (pIFF)
	{
		XFileSetPosition(pIFF->fileReference, 0L);	/* set to begining of file */

		pIFF->formPosition = 0;
		pIFF->formLength = 0;
		XSetMemory(&type, (long)sizeof(XIFFChunk), 0);
		IFF_GetNextGroup(pIFF, &type);
		return( (type.ckID == pIFF->formType) ? type.ckData : -1L);
	}
	return -1L;
}

/*- scan past nested FORM's -*/
static long IFF_NextBlock(X_IFF *pIFF, long blockID)
{
	long saveFORM, saveFLEN;
	XIFFChunk type;
	long flag;

	flag = -1;
	while (XFileGetPosition(pIFF->fileReference) < (pIFF->formPosition + pIFF->formLength))
	{
		 saveFORM = pIFF->formPosition;
		 saveFLEN = pIFF->formLength;
		 if (IFF_GetNextGroup(pIFF, &type) == -1)	/* error? */
		 {
			break;
		 }
		 if (type.ckID == pIFF->formType)
		 {
			type.ckSize -= 4L;
		 }
		 pIFF->formPosition = saveFORM;
		 pIFF->formLength = saveFLEN;
		 if (type.ckID != blockID)
		 {
		 	if (XFileSetPositionRelative(pIFF->fileReference, type.ckSize + (type.ckSize&1)))
	 		{
	  			  flag = -1;  /* error */
	  			  break;
			}
		 }
		 else
		 {
	 		flag = 0;	/* ok, found block ID */
			break;
		 }
	}
	pIFF->lastError = flag;
	return flag;
}


/******- Scan a FORM for a 4 letter block -*******************************/
static long IFF_ScanToBlock(X_IFF *pIFF, long block)
{
	if (XFileSetPosition(pIFF->fileReference, pIFF->formPosition + 4) == 0)		// set to inside of FORM
 	{
		return(IFF_NextBlock(pIFF, block));
	}
	else
	{
		pIFF->lastError = -1;
		return -1;
	}
}

/******************- Return Chunk size -**********************************/
static long IFF_ChunkSize(X_IFF *pIFF, long block)
{
	long size;

	if (IFF_ScanToBlock(pIFF, block) == -1L)
	{
		return(-1); /* bad */
	}
	XFileSetPositionRelative(pIFF->fileReference, -4L);		// back-up and get size
	pIFF->lastError = XFileRead(pIFF->fileReference, &size, 4L);
	if (pIFF->formType == X_RIFF)
	{
		#if X_WORD_ORDER == FALSE	// motorola?
		size = XSwapLong(size);
		#endif
	}
	return(size);
}


/*- go inside a FORM that has been found -*/
static long IFF_NextForm(X_IFF *pIFF)
{
	XIFFChunk type;
	long err;

	XFileSetPosition(pIFF->fileReference, pIFF->formPosition);	// set to inside of FORM
	err = IFF_GetNextGroup(pIFF, &type);
	err |= (type.ckID == pIFF->formType) ? 0 : -1;
	pIFF->lastError =  err;
	return (err == -1) ? -1 : type.ckData;
}

static long IFF_CurrentForm(X_IFF *pIFF)
{
	pIFF->formPosition = XFileGetPosition(pIFF->fileReference);
	return IFF_NextForm(pIFF);
}

static long IFF_ReadBlock(X_IFF *pIFF, XPTR pData, long Length)
{
	pIFF->lastError = XFileRead(pIFF->fileReference, pData, Length);
	return(Length);
}


static long IFF_GetChunk(X_IFF *pIFF, long block, long size, XPTR p)
{
	if (IFF_ScanToBlock(pIFF, block) == -1L)
	{
		 return(-1); /* bad */
	}
	if (size == -1L)	/* size not known? */
	{
		XFileSetPositionRelative(pIFF->fileReference, -4L);		// back-up and get size
		pIFF->lastError = XFileRead(pIFF->fileReference, &size, (long)sizeof(long));
		if (pIFF->formType == X_RIFF)
		{
			#if X_WORD_ORDER == FALSE	// motorola?
			size = XSwapLong(size);
			#endif
		}
		else
		{
			size = XGetLong(&size);
		}
	}

	IFF_ReadBlock(pIFF, p, size);	/* read block */
	pIFF->lastError = 0;
	if (size&1) /* odd? */
	{
		XFileSetPositionRelative(pIFF->fileReference, 1L);		// skip one byte
	}
	return pIFF->lastError;
}


static long IFF_NextChunk(X_IFF *pIFF, long block, long size, XPTR p)
{
	if (IFF_NextBlock(pIFF, block) == -1)
	{
		 return(-1); /* bad */
	}
	if (size == -1L)	/* size not known? */
	{
		XFileSetPositionRelative(pIFF->fileReference, -4L);		// back-up and get size
		pIFF->lastError = XFileRead(pIFF->fileReference, &size, (long)sizeof(long));
		if (pIFF->formType == X_RIFF)
		{
			#if X_WORD_ORDER == FALSE	// motorola?
			size = XSwapLong(size);
			#endif
		}
		else
		{
			size = XGetLong(&size);
		}
	}
	IFF_ReadBlock(pIFF, p, size);	/* read block */
	if (size&1) /* odd? */
	{
		XFileSetPositionRelative(pIFF->fileReference, 1L);		// skip one byte
	}
	return pIFF->lastError;
}

/*
 * Unpack input codes and pass them back as bytes.
 * Returns 1 if there is residual input, returns -1 if eof, else returns 0.
 */
static XFILE		sunFileReference;
static unsigned int	in_buffer = 0;
static int			in_bits = 0;
static INLINE int PV_UnpackInput( unsigned char	*code, int bits)
{
	unsigned char		in_byte;

	if (in_bits < bits) 
	{
		if (XFileRead(sunFileReference, &in_byte, 1L) != 0)
		{
			*code = 0;
			return (-1);
		}
		in_buffer |= (in_byte << in_bits);
		in_bits += 8;
	}
	*code = in_buffer & ((1 << bits) - 1);
	in_buffer >>= bits;
	in_bits -= bits;
	return (in_bits > 0);
}

// Sun AU file code

static long PV_ReadSunAUFile(	long encoding,
								XFILE fileReference, 
								void *pSample,
								long sampleByteLength,
								unsigned long *pBufferLength
								)
{
	long				err;
	short				sample;
	unsigned char		code;
	struct g72x_state	state;
	int					out_size;
	int					(*dec_routine)(int i, int out_coding, struct g72x_state *state_ptr);
	int					dec_bits;
	short int			*pSample16;
	unsigned char		*pSample8;
	unsigned char		codeBlock[1026];
	unsigned long		writeLength;

	writeLength = 0;
	err = 0;
	dec_bits = 0;
	pSample8 = (unsigned char *)pSample;
	pSample16 = (short int *)pSample;
	switch (encoding)
	{
		case SUN_AUDIO_FILE_ENCODING_LINEAR_16:
			writeLength = sampleByteLength;
			XFileRead(fileReference, pSample, writeLength);
			break;
		case SUN_AUDIO_FILE_ENCODING_LINEAR_8:
			writeLength = sampleByteLength;
			XFileRead(fileReference, pSample, writeLength);
			break;
		case SUN_AUDIO_FILE_ENCODING_MULAW_8:
			while (sampleByteLength > 0)
			{
				XFileRead(fileReference, codeBlock, 1024L);
				for (out_size = 0; (out_size < 1024) && sampleByteLength; out_size++)
				{
					*pSample16++ = ulaw2linear(codeBlock[out_size]);
					sampleByteLength -= 2;
					writeLength += 2;
				}
			}
			break;
		case SUN_AUDIO_FILE_ENCODING_ALAW_8:
			while (sampleByteLength > 0)
			{
				XFileRead(fileReference, codeBlock, 1024L);
				for (out_size = 0; (out_size < 1024) && sampleByteLength; out_size++)
				{
					*pSample16++ = alaw2linear(codeBlock[out_size]);
					sampleByteLength -= 2;
					writeLength += 2;
				}
			}
			break;

		case SUN_AUDIO_FILE_ENCODING_ADPCM_G721:
			out_size = sizeof (short);
			dec_routine = g721_decoder;
			dec_bits = 4;
			goto decode_adpcm;
		case SUN_AUDIO_FILE_ENCODING_ADPCM_G723_3:
			out_size = sizeof (short);
			dec_routine = g723_24_decoder;
			dec_bits = 3;
			goto decode_adpcm;
		case SUN_AUDIO_FILE_ENCODING_ADPCM_G723_5:
			out_size = sizeof (short);
			dec_routine = g723_40_decoder;
			dec_bits = 5;

decode_adpcm:
			g72x_init_state(&state);
			sunFileReference = fileReference;
			in_bits = 0;
			in_buffer = 0;
			/* Read and unpack input codes and process them */
			while (PV_UnpackInput(&code, dec_bits) >= 0) 
			{
				sample = (*dec_routine)(code, AUDIO_ENCODING_LINEAR, &state);
				if (out_size == 2) 
				{
					*pSample16++ = sample;
					writeLength += 2;
				}
				else 
				{
					code = (unsigned char)sample;
					*pSample8++ = code;
					writeLength++;
				}
				sampleByteLength--;
				if (sampleByteLength <= 0)
				{
					break;
				}
			}
			break;
	}
	if (pBufferLength)
	{
		*pBufferLength = writeLength;
	}
	return err;
}






// WAVE functions

static long IFF_GetWAVFormatTag(X_IFF *pIFF)
{
	long		theErr;
	XWaveHeader	header;

	theErr = IFF_GetChunk(pIFF, X_FMT, (long)sizeof(XWaveHeader), (void *)&header);

	#if X_WORD_ORDER == FALSE	// motorola?
		header.wFormatTag = XSwapShort(header.wFormatTag);
	#endif
	return (long)header.wFormatTag;
}


static long IFF_GetWAVHeader(X_IFF *pIFF, XWaveHeader * pHeaderInfo)
{
	long	theErr;

	theErr = IFF_GetChunk(pIFF, X_FMT, (long)sizeof(XWaveHeader), (void *)pHeaderInfo);

	#if X_WORD_ORDER == FALSE	// motorola?
		pHeaderInfo->nSamplesPerSec = XSwapLong(pHeaderInfo->nSamplesPerSec);
		pHeaderInfo->nAvgBytesPerSec = XSwapLong(pHeaderInfo->nAvgBytesPerSec);
		pHeaderInfo->wFormatTag = XSwapShort(pHeaderInfo->wFormatTag);
		pHeaderInfo->nChannels = XSwapShort(pHeaderInfo->nChannels);
		pHeaderInfo->nBlockAlign = XSwapShort(pHeaderInfo->nBlockAlign);
		pHeaderInfo->wBitsPerSample = XSwapShort(pHeaderInfo->wBitsPerSample);
		pHeaderInfo->cbSize = XSwapShort(pHeaderInfo->cbSize);
	#endif
	return theErr;
}

static long IFF_GetWAVIMAHeader(X_IFF *pIFF, XWaveHeaderIMA * pHeaderInfo)
{
	long	theErr;

	theErr = IFF_GetChunk(pIFF, X_FMT, (long)sizeof(XWaveHeaderIMA), (void *)pHeaderInfo);

	#if X_WORD_ORDER == FALSE	// motorola?
		pHeaderInfo->wfx.nSamplesPerSec = XSwapLong(pHeaderInfo->wfx.nSamplesPerSec);
		pHeaderInfo->wfx.nAvgBytesPerSec = XSwapLong(pHeaderInfo->wfx.nAvgBytesPerSec);
		pHeaderInfo->wfx.wFormatTag = XSwapShort(pHeaderInfo->wfx.wFormatTag);
		pHeaderInfo->wfx.nChannels = XSwapShort(pHeaderInfo->wfx.nChannels);
		pHeaderInfo->wfx.nBlockAlign = XSwapShort(pHeaderInfo->wfx.nBlockAlign);
		pHeaderInfo->wfx.wBitsPerSample = XSwapShort(pHeaderInfo->wfx.wBitsPerSample);
		pHeaderInfo->wfx.cbSize = XSwapShort(pHeaderInfo->wfx.cbSize);
		pHeaderInfo->wSamplesPerBlock = XSwapShort(pHeaderInfo->wSamplesPerBlock);
	#endif
	return theErr;
}

// Get compressed and uncompressed size. Return 0 if successful, 1 if failure
static long IFF_GetWAVSampleSize(X_IFF *pIFF, unsigned long *pUncompressedSize, unsigned long *pCompressedSize)
{
	long			size, error;
	XWaveHeaderIMA	header;

	error = 0;
	size = 0;
	if ( (IFF_GetWAVIMAHeader(pIFF, &header) == 0) && pUncompressedSize && pCompressedSize)
	{
		if (IFF_ScanToBlock(pIFF, X_DATA) == 0)	/* skip to body */
		{
			XFileSetPositionRelative(pIFF->fileReference, -4L);		// back-up and get size
			pIFF->lastError = XFileRead(pIFF->fileReference, &size, (long)sizeof(long));
			if (pIFF->formType == X_RIFF)
			{
				#if X_WORD_ORDER == FALSE	// motorola?
					size = XSwapLong(size);
				#endif
			}
			*pCompressedSize = size;
			switch(header.wfx.wFormatTag)
			{
				default:
					size = 0;
					error = 1;
					break;
				case X_WAVE_FORMAT_ALAW:
				case X_WAVE_FORMAT_MULAW:
					size *= 2;	// double size
					break;
				case X_WAVE_FORMAT_PCM:
					// size will be valid
					break;
				// calculate the uncompressed byte size
				case X_WAVE_FORMAT_IMA_ADPCM:
					size *= 4;
/*
					{
					    unsigned long	cBlocks;
					    unsigned long	cbBytesPerBlock;

		                //
		                //  how many destination PCM bytes are needed to hold
		                //  the decoded ADPCM data
		                //
		                //  always round UP
		                //
		                cBlocks = size / header.wfx.nBlockAlign;
		                if (cBlocks)
		                {
			                cbBytesPerBlock = header.wSamplesPerBlock * 2; // dest block aline

			                if ( ! ((0xFFFFFFFFL / cbBytesPerBlock) < cBlocks) )
			                {
				                if ((size % header.wfx.nBlockAlign) == 0)
				                {
				                    size = cBlocks * cbBytesPerBlock;
				                }
				                else
				                {
				                    size = (cBlocks + 1) * cbBytesPerBlock;
			    	            }
							}
							else
							{
								size = 0;	// out of range
								error = 1;
							}
						}
						else
						{
							size = 0;	// out of range
							error = 1;
						}
					}
*/
					break;
			}
			*pUncompressedSize = size;
		}
	}
	else
	{
		error = 1;
	}
	return error;
}

static unsigned long PV_ReadWAVEAndDecompressIMA(XFILE fileReference, unsigned long sourceLength,
											char *pDestSample, unsigned long destLength,
											char outputBitSize, char channels,
											XPTR pBlockBuffer, unsigned long blockSize,
											unsigned long *pBufferLength)
{
	unsigned long			writeBufferLength, size, offset;
	unsigned long			error;
	XBOOL					customBlockBuffer;
	
	error = 1;
	writeBufferLength = 0;
	customBlockBuffer = FALSE;
	if (pBlockBuffer == NULL)
	{
		pBlockBuffer = XNewPtr(blockSize);
		customBlockBuffer = TRUE;
	}
	if (pBlockBuffer)
	{
		error = 0;
		while (sourceLength > 0)
		{
			sourceLength -= blockSize;
			if (sourceLength > blockSize)
			{
				size = XFileGetPosition(fileReference);	 /* get current pos */
				error = XFileRead(fileReference, pBlockBuffer, blockSize);
				size = XFileGetPosition(fileReference) - size;
			}
			else
			{
				// last block, just stop
				size = 0;
			}
			offset = 0;
			if (size)
			{
				offset = XDecodeIMA4Bit((unsigned char *)pBlockBuffer, size,
								(unsigned char *)pDestSample,
								blockSize,
								channels,
								outputBitSize);
			}
			if (offset == 0)
			{
				// we're done
				break;
			}
			destLength -= offset;
			if (destLength < 0)
			{
				break;
			}
			pDestSample += offset;
			writeBufferLength += offset;
		}
	}
	if (customBlockBuffer)
	{
		XDisposePtr(pBlockBuffer);
	}
	if (pBufferLength)
	{
		*pBufferLength = writeBufferLength;
	}
	return error;
}


// This will read into memory the entire wave file and return a GM_Waveform structure.
// When disposing make sure and dispose of both the GM_Waveform structure and the
// theWaveform inside of that structure with XDisposePtr
static GM_Waveform * PV_ReadIntoMemoryWaveFile(XFILENAME *file, XBOOL loadData, 
											XPTR pMemoryFile, unsigned long memoryFileSize,
											long *pFormat, long *pBlockSize)
{
	GM_Waveform		*pSample;
	X_IFF			*pIFF;
	XWaveHeaderIMA	waveHeader;
	unsigned long	size, sourceLength;

	pSample = NULL;
	if (pFormat)
	{
		*pFormat = 0;
	}
	pIFF = (X_IFF *)XNewPtr((long)sizeof(X_IFF));
	if (pIFF)
	{
		IFF_SetFormType(pIFF, X_RIFF);
		if (file)
		{
			pIFF->fileReference = XFileOpenForRead(file);
		}
		else
		{
			if (pMemoryFile)
			{
				pIFF->fileReference = XFileOpenForReadFromMemory(pMemoryFile, memoryFileSize);
			}
		}
		if (pIFF->fileReference)
		{
			pSample = (GM_Waveform *)XNewPtr((long)sizeof(GM_Waveform));
			if (pSample)
			{
				if (IFF_FileType(pIFF) == X_WAVE)
				{
					if (IFF_GetWAVIMAHeader(pIFF, &waveHeader) == 0)
					{
						pSample->channels = waveHeader.wfx.nChannels;
						pSample->bitSize = waveHeader.wfx.wBitsPerSample;
						pSample->sampledRate = waveHeader.wfx.nSamplesPerSec << 16L;
						pSample->baseMidiPitch = 60;
						// we want the byte size
						size = 0;
						if (IFF_GetWAVSampleSize(pIFF, &size, &sourceLength) == 0)
						{
							switch (waveHeader.wfx.wFormatTag)
							{
								case X_WAVE_FORMAT_IMA_ADPCM:
									pSample->bitSize = 16;
									break;
								case X_WAVE_FORMAT_ALAW:
								case X_WAVE_FORMAT_MULAW:
									pSample->bitSize = 16;
									break;
							}
						}
						else
						{
							pIFF->lastError = BAD_FILE_TYPE;
						}
						if (pFormat)
						{
							*pFormat = waveHeader.wfx.wFormatTag;
						}
						if (pBlockSize)
						{
							*pBlockSize = waveHeader.wfx.nBlockAlign;
						}
						// if no error, file will be positioned at the sample data
						pSample->waveSize = size;
						if (size)
						{
							pSample->waveFrames = (size / pSample->channels) / (pSample->bitSize / 8);
						}
						// now the file is positioned right at the data block. So let's allocate it and read
						// it into memory
						if (loadData && pSample->waveSize)
						{
							pSample->theWaveform = (SBYTE *)XNewPtr(pSample->waveSize);
							if (pSample->theWaveform)
							{
								switch(waveHeader.wfx.wFormatTag)
								{
									default:
										pIFF->lastError = BAD_FILE_TYPE;
										break;
									case X_WAVE_FORMAT_ALAW:
										pIFF->lastError = PV_ReadSunAUFile(SUN_AUDIO_FILE_ENCODING_ALAW_8,
															pIFF->fileReference, 
															pSample->theWaveform,
															pSample->waveSize,
															&size);
										break;
									case X_WAVE_FORMAT_MULAW:
										pIFF->lastError = PV_ReadSunAUFile(SUN_AUDIO_FILE_ENCODING_MULAW_8,
															pIFF->fileReference, 
															pSample->theWaveform,
															pSample->waveSize,
															&size);
										break;
									case X_WAVE_FORMAT_PCM:
										pIFF->lastError = XFileRead(pIFF->fileReference, 
																	pSample->theWaveform,
																	pSample->waveSize);
										if (pIFF->lastError == 0)
										{
										// now, if the file is 16 bit sample on a motorola ordered system, swap the bytes
										#if X_WORD_ORDER == FALSE	// motorola?
											if ((pIFF->lastError == 0) && (pSample->bitSize == 16) )
											{
												long			count, frames;
												short int		*pAudio;

												pAudio = (short int *)pSample->theWaveform;
												frames = pSample->waveFrames * pSample->channels;
												for (count = 0; count < frames; count++)
												{
													*pAudio = XSwapShort(*pAudio);
													pAudio++;
												}
											}
										#endif
										}
										break;
									case X_WAVE_FORMAT_IMA_ADPCM:
										pIFF->lastError = PV_ReadWAVEAndDecompressIMA(pIFF->fileReference, 
																sourceLength,
																(char *)pSample->theWaveform,
																pSample->waveSize,
																pSample->bitSize,
																pSample->channels,
																NULL,	// allocate block buffer
																waveHeader.wfx.nBlockAlign,
																NULL);
										break;
								}
							}
							else
							{
								pIFF->lastError = MEMORY_ERR;
							}
						}
						else
						{
							// now the file is positioned right at the data block
							pSample->waveformID = XFileGetPosition(pIFF->fileReference);
						}
					}
					else
					{
						pIFF->lastError = BAD_FILE_TYPE;
					}
				}
				else
				{
					pIFF->lastError = BAD_FILE_TYPE;
				}
				if (pIFF->lastError)
				{
					if (pSample)
					{
						XDisposePtr((XPTR)pSample->theWaveform);
					}
					XDisposePtr((XPTR)pSample);
					pSample = NULL;
				}
			}
			else
			{
				pIFF->lastError = MEMORY_ERR;
			}
			XFileClose(pIFF->fileReference);
		}
		XDisposePtr((XPTR)pIFF);
	}
	return pSample;
}


/*
 * C O N V E R T   F R O M   I E E E   E X T E N D E D  
 */

/* 
 * Copyright (C) 1988-1991 Apple Computer, Inc.
 * All rights reserved.
 *
 * Machine-independent I/O routines for IEEE floating-point numbers.
 *
 * NaN's and infinities are converted to HUGE_VAL or HUGE, which
 * happens to be infinity on IEEE machines.  Unfortunately, it is
 * impossible to preserve NaN's in a machine-independent way.
 * Infinities are, however, preserved on IEEE machines.
 *
 * These routines have been tested on the following machines:
 *    Apple Macintosh, MPW 3.1 C compiler
 *    Apple Macintosh, THINK C compiler
 *    Silicon Graphics IRIS, MIPS compiler
 *    Cray X/MP and Y/MP
 *    Digital Equipment VAX
 *
 *
 * Implemented by Malcolm Slaney and Ken Turkowski.
 *
 * Malcolm Slaney contributions during 1988-1990 include big- and little-
 * endian file I/O, conversion to and from Motorola's extended 80-bit
 * floating-point format, and conversions to and from IEEE single-
 * precision floating-point format.
 *
 * In 1991, Ken Turkowski implemented the conversions to and from
 * IEEE double-precision format, added more precision to the extended
 * conversions, and accommodated conversions involving +/- infinity,
 * NaN's, and denormalized numbers.
 */

#define UNSIGNED_TO_FLOAT(u)         (((double)((long)(u - 2147483647L - 1))) + 2147483648.0)

/****************************************************************
 * Extended precision IEEE floating-point conversion routine.
 ****************************************************************/

static FIXED_VALUE PV_ConvertFromIeeeExtended(unsigned char *bytes)
{
	double			f;
	int				expon;
	unsigned long	hiMant, loMant;
	FIXED_VALUE		ieeeRate;

	expon = ((bytes[0] & 0x7F) << 8L) | (bytes[1] & 0xFF);
	hiMant	=    ((unsigned long)(bytes[2] & 0xFF) << 24L)
			|    ((unsigned long)(bytes[3] & 0xFF) << 16L)
			|    ((unsigned long)(bytes[4] & 0xFF) << 8L)
			|    ((unsigned long)(bytes[5] & 0xFF));
	loMant	=    ((unsigned long)(bytes[6] & 0xFF) << 24L)
			|    ((unsigned long)(bytes[7] & 0xFF) << 16L)
			|    ((unsigned long)(bytes[8] & 0xFF) << 8L)
			|    ((unsigned long)(bytes[9] & 0xFF));

	if (expon == 0 && hiMant == 0 && loMant == 0) 
	{
		f = 0;
	}
	else 
	{
		if (expon == 0x7FFF) 
		{    /* Infinity or NaN */
//#ifndef HUGE_VAL
//	#define HUGE_VAL HUGE
//#endif
//			f = HUGE_VAL;
			f = 0.0;
		}
        else 
		{
			expon -= 16383;
			expon -= 31;
			f  = ldexp(UNSIGNED_TO_FLOAT(hiMant), expon);
			expon -= 32;
			f += ldexp(UNSIGNED_TO_FLOAT(loMant), expon);
		}
	}

	if (bytes[0] & 0x80)
	{
		ieeeRate = (FIXED_VALUE)(-f * 65536.0);
	}
	else
	{
		ieeeRate = (FIXED_VALUE)(f * 65536.0);
	}
	return ieeeRate;
}

/*
 * C O N V E R T   T O   I E E E   E X T E N D E D
 */

/* Copyright (C) 1988-1991 Apple Computer, Inc.
 * All rights reserved.
 *
 * Machine-independent I/O routines for IEEE floating-point numbers.
 *
 * NaN's and infinities are converted to HUGE_VAL or HUGE, which
 * happens to be infinity on IEEE machines.  Unfortunately, it is
 * impossible to preserve NaN's in a machine-independent way.
 * Infinities are, however, preserved on IEEE machines.
 *
 * These routines have been tested on the following machines:
 *    Apple Macintosh, MPW 3.1 C compiler
 *    Apple Macintosh, THINK C compiler
 *    Silicon Graphics IRIS, MIPS compiler
 *    Cray X/MP and Y/MP
 *    Digital Equipment VAX
 *
 *
 * Implemented by Malcolm Slaney and Ken Turkowski.
 *
 * Malcolm Slaney contributions during 1988-1990 include big- and little-
 * endian file I/O, conversion to and from Motorola's extended 80-bit
 * floating-point format, and conversions to and from IEEE single-
 * precision floating-point format.
 *
 * In 1991, Ken Turkowski implemented the conversions to and from
 * IEEE double-precision format, added more precision to the extended
 * conversions, and accommodated conversions involving +/- infinity,
 * NaN's, and denormalized numbers.
 */

#define FLOAT_TO_UNSIGNED(f)      ((unsigned long)(((long)(f - 2147483648.0)) + 2147483647L + 1))

static void PV_ConvertToIeeeExtended(FIXED_VALUE ieeeFixedRate, char *bytes)
{
	int				sign;
	int				expon;
	double			fMant, fsMant;
	unsigned long	hiMant, loMant;
	double			num;

	num = (double)ieeeFixedRate / 65536.0;
    if (num < 0) 
    {
        sign = 0x8000;
        num *= -1;
    } 
    else 
    {
        sign = 0;
    }

    if (num == 0) 
    {
        expon = 0; hiMant = 0; loMant = 0;
    }
    else 
    {
        fMant = frexp(num, &expon);
        if ((expon > 16384) || !(fMant < 1)) 
        {    /* Infinity or NaN */
            expon = sign|0x7FFF; 
            hiMant = 0; 
            loMant = 0; /* infinity */
        }
        else 
        {    /* Finite */
            expon += 16382;
            if (expon < 0) 
            {    /* denormalized */
                fMant = ldexp(fMant, expon);
                expon = 0;
            }
            expon |= sign;
            fMant = ldexp(fMant, 32);          
            fsMant = floor(fMant); 
            hiMant = FLOAT_TO_UNSIGNED(fsMant);
            fMant = ldexp(fMant - fsMant, 32); 
            fsMant = floor(fMant); 
            loMant = FLOAT_TO_UNSIGNED(fsMant);
        }
    }
    
    bytes[0] = expon >> 8;
    bytes[1] = expon;
    bytes[2] = hiMant >> 24;
    bytes[3] = hiMant >> 16;
    bytes[4] = hiMant >> 8;
    bytes[5] = hiMant;
    bytes[6] = loMant >> 24;
    bytes[7] = loMant >> 16;
    bytes[8] = loMant >> 8;
    bytes[9] = loMant;
}


static long IFF_GetAIFFHeader(X_IFF *pIFF, XAIFFHeader * pHeaderInfo)
{
	long	theErr;

	theErr = IFF_GetChunk(pIFF, X_Common, (long)sizeof(XAIFFHeader), (void *)pHeaderInfo);

	#if X_WORD_ORDER != FALSE	// intel?
		pHeaderInfo->numChannels = XSwapShort(pHeaderInfo->numChannels);
		pHeaderInfo->numSampleFrames = XSwapLong(pHeaderInfo->numSampleFrames);
		pHeaderInfo->sampleSize = XSwapShort(pHeaderInfo->sampleSize);
	#endif
	return theErr;
}

static long IFF_GetAIFFExtenedHeader(X_IFF *pIFF, XAIFFExtenedHeader * pHeaderInfo)
{
	long	theErr;
	char	size;

	theErr = IFF_GetChunk(pIFF, X_Common, (long)sizeof(XAIFFHeader) + sizeof(long), (void *)pHeaderInfo);

	#if X_WORD_ORDER != FALSE	// intel?
		pHeaderInfo->numChannels = XSwapShort(pHeaderInfo->numChannels);
		pHeaderInfo->numSampleFrames = XSwapLong(pHeaderInfo->numSampleFrames);
		pHeaderInfo->sampleSize = XSwapShort(pHeaderInfo->sampleSize);
		pHeaderInfo->compressionType = XSwapLong(pHeaderInfo->compressionType);
	#endif
	
	theErr = XFileRead(pIFF->fileReference, &size, 1L);
	pHeaderInfo->compressionName[0] = size;
	theErr = XFileRead(pIFF->fileReference, &pHeaderInfo->compressionName[1], size);
	return theErr;
}

static long IFF_GetAIFFInstrument(X_IFF *pIFF, XInstrumentHeader * pInstrumentInfo)
{
	long	theErr;

	theErr = IFF_GetChunk(pIFF, X_Instrument, (long)sizeof(XInstrumentHeader), (void *)pInstrumentInfo);

	#if X_WORD_ORDER != FALSE	// intel?
		pInstrumentInfo->gain = XSwapShort(pInstrumentInfo->gain);
		pInstrumentInfo->sustainLoop_playMode = XSwapShort(pInstrumentInfo->sustainLoop_playMode);
		pInstrumentInfo->sustainLoop_beginLoop = XSwapShort(pInstrumentInfo->sustainLoop_beginLoop);
		pInstrumentInfo->sustainLoop_endLoop = XSwapShort(pInstrumentInfo->sustainLoop_endLoop);
		pInstrumentInfo->releaseLoop_beginLoop = XSwapShort(pInstrumentInfo->releaseLoop_beginLoop);
		pInstrumentInfo->releaseLoop_endLoop = XSwapShort(pInstrumentInfo->releaseLoop_endLoop);
		pInstrumentInfo->extra = XSwapShort(pInstrumentInfo->extra);
	#endif
	return theErr;
}


/*
	unsigned short					numMarkers;			// 2 markers
	short							id
	unsigned long					position;			// 1
	pstring							name;				// begloop
	short							id
	unsigned long					position;			// 2
	pstring							name;				// endloop
*/
static long IFF_GetAIFFLoopPoints(X_IFF *pIFF, long *pLoopStart, long *pLoopEnd)
{
	char	loopMark[1024];
	long	theErr;
	char	*pData;
	short	len;

	*pLoopStart = 0;
	*pLoopEnd = 0;
	theErr = IFF_GetChunk(pIFF, X_Marker, 1023L, loopMark);
	if (theErr == 0)
	{
		pData = loopMark;
		len = XGetShort(pData);
		if (len >= 2)
		{
			pData += 2;				// skip marker count
			pData += 2;				// skip marker id
			*pLoopStart = XGetLong(pData);
			pData += 4;				// skip past start

			len = *pData;
			pData += len + 1;		// walk past first string
			pData += ODD(pData);

			pData += 2;				// skip marker id
			*pLoopEnd = XGetLong(pData);
		}
	}
	return theErr;
}

// Get compressed and uncompressed size. Return 0 if successful, 1 if failure
static long IFF_GetAIFFSampleSize(X_IFF *pIFF, long *pUncompressedSize, long *pCompressedSize)
{
	long				size, error;
	XAIFFExtenedHeader	header;

	size = 0L;
	error = 0;
	if (pUncompressedSize && pCompressedSize)
	{
		switch (pIFF->headerType)
		{
			case X_AIFF:
				IFF_GetAIFFHeader(pIFF, (XAIFFHeader *)&header);
				size = header.numSampleFrames * header.numChannels * (header.sampleSize / 8);
				break;
			case X_AIFC:
				IFF_GetAIFFExtenedHeader(pIFF, &header);
				switch (header.compressionType)
				{
					default:
						pIFF->lastError = BAD_FILE_TYPE;
						// fail because we don't know how to decompress
						break;
					case X_NONE:
						size = header.numSampleFrames * header.numChannels * (header.sampleSize / 8);
						break;
					case X_IMA4:
						// Sound Manager defines 64 samples per packet number of sample frames
						size = header.numSampleFrames * 64 * header.numChannels * (header.sampleSize / 8);
						break;
				}
				break;
		}
	}

	if (size)
	{
		// now position right to data block
		if (IFF_ScanToBlock(pIFF, X_SoundData))	// skip to body
		{
			error = 1;	// failed
		}
		else
		{
			XFileSetPositionRelative(pIFF->fileReference, -4L);		// back-up and get size
			pIFF->lastError = XFileRead(pIFF->fileReference, pCompressedSize, (long)sizeof(long));
			*pCompressedSize = XGetLong(pCompressedSize);
//			XFileSetPositionRelative(pIFF->fileReference, sizeof(long) * 2L);
		}
	}
	*pUncompressedSize = size;
	return error;
}

#define AIFF_IMA_BLOCK_SIZE		34L
#define AIFF_IMA_BUFFER_SIZE	AIFF_IMA_BLOCK_SIZE * 40

static long PV_ReadAIFFAndDecompressIMA(XFILE fileReference, long sourceLength,
											unsigned char *pDestSample, long destLength,
											char outputBitSize, char channels,
											unsigned long *pBufferLength,
											long *pIndexL, long *pIndexR)
{
	long					error;
	char					codeBlock[AIFF_IMA_BUFFER_SIZE];
	long					writeBufferLength, size, offset;

	error = 0;
	writeBufferLength = 0;
	sourceLength -= sourceLength % AIFF_IMA_BUFFER_SIZE;	// round to block size

	#if USE_DEBUG && 0
	{
		char text[256];
		
		sprintf(text, "sourceLength %ld, AIFF_IMA_BUFFER_SIZE %ld", sourceLength, AIFF_IMA_BUFFER_SIZE);
		DEBUG_STR(XCtoPstr(text));
	}
	#endif

	while (sourceLength > 0)
	{
		if (sourceLength > AIFF_IMA_BUFFER_SIZE)
		{
			size = XFileGetPosition(fileReference);	 /* get current pos */
			XFileRead(fileReference, codeBlock, AIFF_IMA_BUFFER_SIZE);
			size = XFileGetPosition(fileReference) - size;
		}
		else
		{
			// last block so just stop
			size = 0;
		}
		sourceLength -= AIFF_IMA_BUFFER_SIZE;
		offset = 0;
		if (size)
		{
			offset = XExpandIMAtoXBitLinearStreamed(codeBlock, size, 
												pDestSample, 
												AIFF_IMA_BLOCK_SIZE,
												channels, 
												outputBitSize,
												pIndexL, pIndexR);
			if (offset == 0)
			{
				// we're done
				break;
			}
			destLength -= offset;
			if (destLength < 0)
			{
				// time to quit, we've hit the end
				break;
			}
			else
			{
				pDestSample += offset;
				writeBufferLength += offset;
			}
		}
	}
	if (pBufferLength)
	{
		*pBufferLength = writeBufferLength;
	}
	return error;
}

// This will read into memory the entire AIFF file and return a GM_Waveform structure.
// When disposing make sure and dispose of both the GM_Waveform structure and the
// theWaveform inside of that structure with XDisposePtr
static GM_Waveform * PV_ReadIntoMemoryAIFFFile(XFILENAME *file, XBOOL loadData, 
							XPTR pMemoryFile, unsigned long memoryFileSize,
											long *pFormat, long *pBlockSize)
{
	GM_Waveform			*pSample;
	X_IFF				*pIFF;
	XAIFFExtenedHeader	aiffHeader;
	long				type, size, sourceLength;

	XSetMemory(&aiffHeader, (long)sizeof(XAIFFExtenedHeader), 0);
	pSample = NULL;
	if (pFormat)
	{
		*pFormat = 0;
	}
	if (pBlockSize)
	{
		*pBlockSize = 0;
	}
	pIFF = (X_IFF *)XNewPtr((long)sizeof(X_IFF));
	if (pIFF)
	{
		IFF_SetFormType(pIFF, X_FORM);
		if (file)
		{
			pIFF->fileReference = XFileOpenForRead(file);
		}
		else
		{
			if (pMemoryFile)
			{
				pIFF->fileReference = XFileOpenForReadFromMemory(pMemoryFile, memoryFileSize);
			}
		}
		if (pIFF->fileReference)
		{
			pSample = (GM_Waveform *)XNewPtr((long)sizeof(GM_Waveform));
			if (pSample)
			{
				type = IFF_FileType(pIFF);
				pIFF->headerType = type;
				if ( (type == X_AIFF) || (type == X_AIFC) )
				{
					if (type == X_AIFF)
					{
						IFF_GetAIFFHeader(pIFF, (XAIFFHeader *)&aiffHeader);
						aiffHeader.compressionType = X_NONE;
					}
					else
					{
						IFF_GetAIFFExtenedHeader(pIFF, &aiffHeader);
						if (pBlockSize)
						{
							*pBlockSize = sizeof(long) * 2;
						}
					}
					if (pFormat)
					{
						*pFormat = aiffHeader.compressionType;
					}
					pSample->channels = aiffHeader.numChannels;
					pSample->bitSize = aiffHeader.sampleSize;
					pSample->baseMidiPitch = 60;

					// get loop points, if there
					IFF_GetAIFFLoopPoints(pIFF, (long *)&pSample->startLoop, (long *)&pSample->endLoop);
					
					// Convert the ieee number into a 16.16 fixed value
					size = 0;
					pSample->sampledRate = PV_ConvertFromIeeeExtended(aiffHeader.sampleRate);
					if (pSample->sampledRate == 0)
					{
						pIFF->lastError = BAD_SAMPLE;
					}
					else
					{
						// we want the file byte size, uncompressed
						if (IFF_GetAIFFSampleSize(pIFF, &size, &sourceLength) == 0)
						{
							pSample->waveSize = size;
							pSample->waveFrames = pSample->waveSize / pSample->channels;
							pSample->waveFrames = pSample->waveFrames / ((pSample->bitSize == 8) ? 1 : 2);
						}
					}
					if (size == 0)
					{
						pIFF->lastError = BAD_SAMPLE;
					}
					// now the file is positioned right at the data block. So let's allocate it and read
					// it into memory
					XFileSetPositionRelative(pIFF->fileReference, sizeof(long) * 2L);
					if (loadData && size)
					{
						pSample->theWaveform = (SBYTE *)XNewPtr(pSample->waveSize);
						if (pSample->theWaveform)
						{
							switch (aiffHeader.compressionType)
							{
								default:
									pIFF->lastError = BAD_FILE_TYPE;
									break;
								case X_NONE:
									pIFF->lastError = XFileRead(pIFF->fileReference, 
																pSample->theWaveform,
																pSample->waveSize);
									#if X_WORD_ORDER != FALSE	// intel?
									if (pIFF->lastError == 0)
									{
										short int		*pAudio;
										long			count, frames;

										// now, if the file is 16 bit sample on a intel ordered system, swap the bytes
										if (pSample->bitSize == 16)
										{
											pAudio = (short int *)pSample->theWaveform;
											frames = pSample->waveFrames * pSample->channels;
											for (count = 0; count < frames; count++)
											{
												*pAudio = XSwapShort(*pAudio);
												pAudio++;
											}
										}
									}
									#endif
									break;

								case X_IMA4:
									{
										long	indexL, indexR;
									
										indexL = 0;
										indexR = 0;
										pIFF->lastError = PV_ReadAIFFAndDecompressIMA(pIFF->fileReference,
																	sourceLength,
																	(unsigned char *)pSample->theWaveform,
																	pSample->waveSize,
																	pSample->bitSize,
																	pSample->channels,
																	NULL,
																	&indexL, &indexR);
									}
									break;
							}

							if (pIFF->lastError == 0)
							{
								// now, if the file is 8 bit sample, change the sample phase
								if (pSample->bitSize == 8)
								{
									XPhase8BitWaveform((unsigned char *)pSample->theWaveform, pSample->waveSize);
								}
							}
						}
						else
						{
							pIFF->lastError = MEMORY_ERR;
						}
					}
					else
					{
						// now the file is positioned right at the data block
						pSample->waveformID = XFileGetPosition(pIFF->fileReference);
					}
				}
				else
				{
					pIFF->lastError = BAD_FILE_TYPE;
				}
				if (pIFF->lastError)
				{
					if (pSample)
					{
						XDisposePtr((XPTR)pSample->theWaveform);
					}
					XDisposePtr((XPTR)pSample);
					pSample = NULL;
				}
			}
			else
			{
				pIFF->lastError = MEMORY_ERR;
			}
			XFileClose(pIFF->fileReference);
		}
		XDisposePtr((XPTR)pIFF);
	}
	return pSample;
}

// This will read into memory the entire Sun AU file and return a GM_Waveform structure.
// When disposing make sure and dispose of both the GM_Waveform structure and the
// theWaveform inside of that structure with XDisposePtr
static GM_Waveform * PV_ReadIntoMemorySunAUFile(XFILENAME *file, XBOOL loadData, 
												XPTR pMemoryFile, unsigned long memoryFileSize,
													long *pFormat, long *pBlockSize)
{
	GM_Waveform			*pSample;
	SunAudioFileHeader	sunHeader;
	XFILE				fileReference;
	long				err;
	long				encoding, size;
	short int			bits;
	long				filePos, originalLength, waveLength;

	pSample = NULL;
	fileReference = 0;
	waveLength = 0;
	if (pBlockSize)
	{
		*pBlockSize = 0;
	}
	if (file)
	{
		fileReference = XFileOpenForRead(file);
	}
	else
	{
		if (pMemoryFile)
		{
			fileReference = XFileOpenForReadFromMemory(pMemoryFile, memoryFileSize);
		}
	}
	err = 0;
	if (fileReference)
	{
		err = XFileRead(fileReference, &sunHeader, (long)sizeof(SunAudioFileHeader));
		if (err == 0)
		{
			// now skip past any info string
			size = XGetLong(&sunHeader.hdr_size) - (long)sizeof(SunAudioFileHeader);
			filePos = XFileGetPosition(fileReference) + size;
			originalLength = XFileGetLength(fileReference) - size + sizeof(SunAudioFileHeader);

			XFileSetPosition(fileReference, filePos);	// go back to ending
		    // Make sure we've got a legitimate audio file
			encoding = XGetLong(&sunHeader.encoding);
		    if(XGetLong(&sunHeader.magic) == SUN_AUDIO_FILE_MAGIC_NUMBER)
		    {
				// determine what file types we support
				bits = -1;
				switch (encoding)
				{
					case SUN_AUDIO_FILE_ENCODING_LINEAR_8:
						waveLength = originalLength;
						bits = 8;
						break;
					case SUN_AUDIO_FILE_ENCODING_MULAW_8:
					case SUN_AUDIO_FILE_ENCODING_ALAW_8:
						waveLength = originalLength * 2;
						bits = 16;
						break;
					case SUN_AUDIO_FILE_ENCODING_LINEAR_16:
						waveLength = originalLength;
						bits = 16;
						break;
					case SUN_AUDIO_FILE_ENCODING_ADPCM_G721:
						waveLength = originalLength * 4;
						bits = 16;
						break;
					case SUN_AUDIO_FILE_ENCODING_ADPCM_G723_3:
					case SUN_AUDIO_FILE_ENCODING_ADPCM_G723_5:
						waveLength = originalLength * 4;
						bits = 16;
						break;
				}
				if (pFormat)
				{
					*pFormat = encoding;
				}
				if (bits != -1)
				{
					pSample = (GM_Waveform *)XNewPtr((long)sizeof(GM_Waveform));
					if (pSample)
					{
						pSample->channels = XGetLong(&sunHeader.channels);
						pSample->baseMidiPitch = 60;
						pSample->bitSize = bits;
						pSample->sampledRate = XGetLong(&sunHeader.sample_rate) << 16L;
						// we want the byte size
						pSample->waveSize = waveLength;
						// now the file is positioned right at the data block. So let's allocate it and read
						// it into memory
						if (loadData)
						{
							pSample->theWaveform = (SBYTE *)XNewPtr(pSample->waveSize);
							if (pSample->theWaveform)
							{
								pSample->waveFrames = pSample->waveSize / pSample->channels;
								pSample->waveFrames = pSample->waveFrames / ((pSample->bitSize == 8) ? 1 : 2);

								err = PV_ReadSunAUFile(encoding, fileReference,
														pSample->theWaveform, pSample->waveSize, NULL);
								if (err == 0)
								{
									// we don't need to byte swap these based upon platform because
									// we generate samples from runtime code, so the results are in
									// native format already

									// now, if the file is 8 bit sample, change the sample phase
									if (pSample->bitSize == 8)
									{
										XPhase8BitWaveform((unsigned char *)pSample->theWaveform, pSample->waveSize);
									}
								}
							}
							else
							{
								err = MEMORY_ERR;
							}
						}
						else
						{
							// now the file is positioned right at the data block
							pSample->waveformID = XFileGetPosition(fileReference);
						}
					}
					else
					{
						err = BAD_FILE_TYPE;
					}
					if (err)
					{
						if (pSample)
						{
							XDisposePtr((XPTR)pSample->theWaveform);
						}
						XDisposePtr((XPTR)pSample);
						pSample = NULL;
					}
				}
				else
				{
					err = MEMORY_ERR;
				}
			}
			else
			{
				err = BAD_FILE_TYPE;
			}
		}
		XFileClose(fileReference);
	}
	return pSample;
}


// Read into memory a file
GM_Waveform * GM_ReadFileIntoMemory(XFILENAME *file, AudioFileType fileType)
{
	GM_Waveform *pWave;

	pWave = NULL;
	switch (fileType)
	{
		case FILE_AIFF_TYPE:
			pWave = PV_ReadIntoMemoryAIFFFile(file, TRUE, NULL, 0L, NULL, NULL);		// load samples
			break;
		case FILE_WAVE_TYPE:
			pWave = PV_ReadIntoMemoryWaveFile(file, TRUE, NULL, 0L, NULL, NULL);		// load samples
			break;
		case FILE_AU_TYPE:
			pWave = PV_ReadIntoMemorySunAUFile(file, TRUE, NULL, 0L, NULL, NULL);		// load samples
			break;
	}
	return pWave;
}

// Read into memory a file
GM_Waveform * GM_ReadFileIntoMemoryFromMemory(void *pFileBlock, unsigned long fileBlockSize, AudioFileType fileType)
{
	GM_Waveform *pWave;

	pWave = NULL;
	switch (fileType)
	{
		case FILE_AIFF_TYPE:
			pWave = PV_ReadIntoMemoryAIFFFile(NULL, TRUE, pFileBlock, fileBlockSize, NULL, NULL);		// load samples
			break;
		case FILE_WAVE_TYPE:
			pWave = PV_ReadIntoMemoryWaveFile(NULL, TRUE, pFileBlock, fileBlockSize, NULL, NULL);		// load samples
			break;
		case FILE_AU_TYPE:
			pWave = PV_ReadIntoMemorySunAUFile(NULL, TRUE, pFileBlock, fileBlockSize, NULL, NULL);		// load samples
			break;
	}
	return pWave;
}

// Read file information from file, which is a fileType file. If pFormat is not NULL, then
// store format specific format type
GM_Waveform * GM_ReadFileInformation(XFILENAME *file, AudioFileType fileType, 
											long *pFormat, long *pBlockSize)
{
	GM_Waveform *pWave;

	pWave = NULL;
	if (pBlockSize)
	{
		*pBlockSize = 0;
	}
	switch (fileType)
	{
		case FILE_AIFF_TYPE:
			pWave = PV_ReadIntoMemoryAIFFFile(file, FALSE, NULL, 0L, 
												pFormat, pBlockSize);		// don't load samples
			*pBlockSize = sizeof(long) * 2;
			break;
		case FILE_WAVE_TYPE:
			pWave = PV_ReadIntoMemoryWaveFile(file, FALSE, NULL, 0L, 
												pFormat, pBlockSize);		// don't load samples
			break;
		case FILE_AU_TYPE:
			pWave = PV_ReadIntoMemorySunAUFile(file, FALSE, NULL, 0L,		// don't load samples
												pFormat, pBlockSize);		// don't load samples
			break;
	}
	return pWave;
}

// Read a block of data, based apon file type and format, decode and store into a buffer.
// Return length of buffer stored or 0 if error.
long GM_ReadAndDecodeFileStream(XFILE fileReference, 
										AudioFileType fileType, long format, 
										XPTR pBlockBuffer, unsigned long blockSize,
										XPTR pBuffer, unsigned long bufferFrames,
										short int channels, short int bitSize,
										unsigned long *pStoredBufferLength,
										unsigned long *pReadBufferLength)
{
	unsigned long			returnedLength, writeLength, bufferSize;
	long					fileError;
	unsigned short int		*pAudio;
	long					count, frames, filePosition;
	XBOOL					calculateFileSize;

	returnedLength = 0;
	writeLength = 0;
	calculateFileSize = TRUE;
	if (pBuffer && fileReference)
	{
		filePosition = XFileGetPosition(fileReference);
		bufferSize = bufferFrames * channels * (bitSize / 8);
		switch (fileType)
		{
			case FILE_AU_TYPE:
				fileError = PV_ReadSunAUFile(format, fileReference,
										pBuffer, bufferSize, &writeLength);
				
				calculateFileSize = FALSE;
				returnedLength = writeLength;
				// adjust for size actually read into the buffer
//				returnedLength = XFileGetPosition(fileReference) - filePosition;	// length of data read from the file
				break;
			case FILE_AIFF_TYPE:
				switch(format)
				{
					default:
						calculateFileSize = FALSE;
						break;
					case X_NONE:
						fileError = XFileRead(fileReference, 
													pBuffer,
													bufferSize);
						#if X_WORD_ORDER != FALSE	// intel?
						// now, if the file is 16 bit sample on a intel ordered system, swap the bytes
						if (bitSize == 16)
						{
							pAudio = (unsigned short int *)pBuffer;
							frames = bufferFrames * channels;	// convert to sample frames
							for (count = 0; count < frames; count++)
							{
								*pAudio = XSwapShort(*pAudio);
								pAudio++;
							}
						}
						#endif
						// now, if the file is 8 bit sample, change the sample phase
						if (bitSize == 8)
						{
							XPhase8BitWaveform((unsigned char *)pBuffer, bufferSize);
						}
						break;
					case X_IMA4:
						if (pBlockBuffer)
						{
							fileError = PV_ReadAIFFAndDecompressIMA(fileReference,
														bufferSize / 4,
														(unsigned char *)pBuffer,
														bufferSize,
														bitSize,
														channels,
														(unsigned long *)&writeLength,
														(long *)pBlockBuffer, 
														(long *)((char *)pBlockBuffer + sizeof(long)));

							returnedLength = bufferSize / 4;
						}
						else
						{
							writeLength = 0;
							returnedLength = 0;
						}
						calculateFileSize = FALSE;
						break;
				}
				break;
			case FILE_WAVE_TYPE:				
				switch(format)
				{
					default:
						calculateFileSize = FALSE;
						break;
					case X_WAVE_FORMAT_PCM:			// normal PCM data
						fileError = XFileRead(fileReference, 
													pBuffer,
													bufferSize);
						#if X_WORD_ORDER == FALSE	// motorola?
						// now, if the file is 16 bit sample on a intel ordered system, swap the bytes
						if (bitSize == 16)
						{
							pAudio = (unsigned short int *)pBuffer;
							frames = bufferFrames * channels;	// convert to sample frames
							for (count = 0; count < frames; count++)
							{
								*pAudio = XSwapShort(*pAudio);
								pAudio++;
							}
						}
						#endif
						break;

					case X_WAVE_FORMAT_ALAW:
						fileError = PV_ReadSunAUFile(SUN_AUDIO_FILE_ENCODING_ALAW_8,
											fileReference, 
											(char *)pBuffer,
											bufferSize,
											&writeLength);
						break;
					case X_WAVE_FORMAT_MULAW:
						fileError = PV_ReadSunAUFile(SUN_AUDIO_FILE_ENCODING_MULAW_8,
											fileReference, 
											(char *)pBuffer,
											bufferSize,
											&writeLength);
						break;
					case X_WAVE_FORMAT_IMA_ADPCM:	// IMA 4 to 1 compressed data
						if (pBlockBuffer)
						{
							fileError = PV_ReadWAVEAndDecompressIMA(fileReference, 
													bufferSize / 4,
													(char *)pBuffer,
													bufferSize,
													bitSize,
													channels,
													pBlockBuffer,
													blockSize,
													(unsigned long *)&writeLength);
							returnedLength = bufferSize / 4;
							// since we decode the samples at runtime, we don't have
							// to byte swap the words.
						}
						else
						{
							writeLength = 0;
							returnedLength = 0;
						}
						calculateFileSize = FALSE;
						break;
				}
				break;
		}
		if (calculateFileSize)
		{
			// adjust for size actually read into the buffer
			returnedLength = XFileGetPosition(fileReference) - filePosition;	// length of data read from the file
			writeLength = returnedLength;										// length of data created for audio buffer

			#if USE_DEBUG && 0
			{
				char text[256];
				
				sprintf(text, "%ld %ld", returnedLength, bufferLength);
				DEBUG_STR(XCtoPstr(text));
			}
			#endif
		}
		*pReadBufferLength = returnedLength;
		*pStoredBufferLength = writeLength;
	}
	return fileError;
}
#endif	// USE_HIGHLEVEL_FILE_API

// EOF of GenSoundFiles.c
