/*
 * @(#)HAE_API.h	1.5 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
/*****************************************************************************/
/*
**	HAE_API.h
**
**	This provides platform specfic functions for HAE
**
**	Written by Steve Hales
**
**	Confidential-- Internal use only
**
**	History	-
**	7/14/97		Created
**	7/22/97		Added HAE_GetAudioByteBufferSize
**	11/11/97	Added HAE_MaxDevices & HAE_SetDeviceID & HAE_GetDeviceID & HAE_GetDeviceName
*/
/*****************************************************************************/

#ifndef HAE_API
#define HAE_API

#ifdef __cplusplus
	extern "C" {
#endif

// This file contains API's that need to be defined in order to get HAE (IgorAudio)
// to link and compile.


// **** System setup and cleanup functions
// Setup function. Called before memory allocation, or anything serious. Can be used to
// load a DLL, library, etc.
// return 0 for ok, or -1 for failure
int HAE_Setup(void);

// Cleanup function. Called after all memory and the last buffers are deallocated, and
// audio hardware is released. Can be used to unload a DLL, library, etc.
// return 0 for ok, or -1 for failure
int HAE_Cleanup(void);

// **** Memory management
// allocate a block of locked, zeroed memory. Return a pointer
void * HAE_Allocate(unsigned long size);

// dispose of memory allocated with HAE_Allocate
void HAE_Deallocate(void * memoryBlock);

// Given a memory pointer and a size, validate of memory pointer is a valid memory address
// with at least size bytes of data avaiable from the pointer.
// This is used to determine if a memory pointer and size can be accessed without
// causing a memory protection fault.
// return 0 for valid, or 1 for bad pointer, or 2 for not supported.
int HAE_IsBadReadPointer(void *memoryBlock, unsigned long size);

// this will return the size of the memory pointer allocated with HAE_Allocate. Return
// 0 if you don't support this feature
unsigned long HAE_SizeOfPointer(void * memoryBlock);

// block move memory. This is basicly memcpy, but its exposed to take advantage of
// special block move speed ups various hardware might have available.
void HAE_BlockMove(void * source, void * dest, unsigned long size);

// **** Audio Card modifiers
// Return 1 if stereo hardware is supported, otherwise 0.
int HAE_IsStereoSupported(void);

// Return 1, if sound hardware support 16 bit output, otherwise 0.
int HAE_Is16BitSupported(void);

// Return 1, if sound hardware support 8 bit linear output, otherwise 0.
int HAE_Is8BitSupported(void);

// returned volume is in the range of 0 to 256. If you're hardware doesn't support this
// range, just scale it.
short int HAE_GetHardwareVolume(void);

// theVolume is in the range of 0 to 256. If you're hardware doesn't support this
// range, just scale it.
void HAE_SetHardwareVolume(short int theVolume);

// **** Timing services
// return microseconds, preferably quantized better than 1000 microseconds, but can live
// with it being as bad as 11000 microseconds.
unsigned long HAE_Microseconds(void);

// wait or sleep this thread for this many microseconds
void HAE_WaitMicroseocnds(unsigned long wait);

// **** File support
// NOTE: about the fileName parameters. For "C" string path name based OS's the name
// will be passed in from the C++ API. For MacOS, or structure based file systems, the
// native structure will be passed in. For example: on a MacOS a FSSpec pointer is passed
// from the C++ API to these various functions.

// Given the fileNameSource that comes from the call HAE_FileXXXX, copy the name
// in the native format to the pointer passed fileNameDest.
void HAE_CopyFileNameNative(void *fileNameSource, void *fileNameDest);

// Create a file, delete orignal if duplicate file name.
// Return -1 if error
long HAE_FileCreate(void *fileName);

// Open a file
// Return -1 if error, otherwise file handle
long HAE_FileOpenForRead(void *fileName);
long HAE_FileOpenForWrite(void *fileName);
long HAE_FileOpenForReadWrite(void *fileName);

// Close a file
void HAE_FileClose(long fileReference);

// Read a block of memory from a file
// Return -1 if error, otherwise length of data read.
long HAE_ReadFile(long fileReference, void *pBuffer, long bufferLength);

// Write a block of memory from a file
// Return -1 if error, otherwise length of data written.
long HAE_WriteFile(long fileReference, void *pBuffer, long bufferLength);

// set file position in absolute file byte position
long HAE_SetFilePosition(long fileReference, unsigned long filePosition);

// get file position in absolute file bytes
unsigned long HAE_GetFilePosition(long fileReference);

// get length of file
unsigned long HAE_GetFileLength(long fileReference);

// set the length of a file. Return 0, if ok, or -1 for error
int HAE_SetFileLength(long fileReference, unsigned long newSize);

// **** Audio card support
// Aquire and enabled audio card. sampleRate is 44100, 22050, or 11025; channels is 1 or 2;
// bits is 8 or 16.
// return 0 if ok, -1 if failed
int HAE_AquireAudioCard(void *context, long sampleRate, long channels, long bits);

// Release and free audio card.
// return 0 if ok, -1 if failed.
int HAE_ReleaseAudioCard(void *context);

unsigned long HAE_GetDeviceSamplesPlayedPosition(void);

// number of devices. ie different versions of the HAE connection. DirectSound and waveOut
// return number of devices. ie 1 is one device, 2 is two devices.
// NOTE: This function needs to function before any other calls may have happened.
long HAE_MaxDevices(void);

// set the current device. device is from 0 to HAE_MaxDevices()
// NOTE:	This function needs to function before any other calls may have happened.
//			Also you will need to call HAE_ReleaseAudioCard then HAE_AquireAudioCard
//			in order for the change to take place.
void HAE_SetDeviceID(long deviceID);

// return current device ID
// NOTE: This function needs to function before any other calls may have happened.
long HAE_GetDeviceID(void);

// get deviceID name
// NOTE:	This function needs to function before any other calls may have happened.
//			Format of string is a zero terminated comma delinated C string.
//			"platform,method,misc"
//	example	"MacOS,Sound Manager 3.0,SndPlayDoubleBuffer"
//			"WinOS,DirectSound,multi threaded"
//			"WinOS,waveOut,multi threaded"
//			"WinOS,VxD,low level hardware"
//			"WinOS,plugin,Director"
void HAE_GetDeviceName(long deviceID, char *cName, unsigned long cNameLength);

// Return the number of microseconds of real time that will be generated when calling
// HAE_BuildMixerSlice.
unsigned long HAE_GetSliceTimeInMicroseconds(void);

// Return the number of buffer blocks that are built at one time. The value returned
// from this function and HAE_GetSliceTimeInMicroseconds will give you the amount
// of lantancy
int HAE_GetAudioBufferCount(void);

// Return the number of bytes used for audio buffer for output to card
long HAE_GetAudioByteBufferSize(void);

// **** Audio Engine feedback functions. These functions are used to direct or get
//		information about the engine.
//
// NOTE:	These are external functions that your audio card code calls to process
//			a mixer buffer. You call this, don't define it.

// Based upon sample rate, channels, and bit size, build 11 milliseconds worth of mixer
// output, and store it in pAudioBuffer.
// bufferByteLength is assumed to be 256, 512, 1024, or 2048 only. The value is ignored
// at the moment. sampleFrames is how many sample frames. These two values should match
// ie. sampleFrames = bufferByteLength / channels / bitsize / 8
extern void HAE_BuildMixerSlice(void *context, void *pAudioBuffer,
					long bufferByteLength,
					long sampleFrames);

// Return the maximumn number of samples for 11 milliseconds worth of 44 khz data.
// Typically this is 512. Use this in your calculation of audio buffers
extern short HAE_GetMaxSamplePerSlice(void);


//CLS:  THREADING API:

// the type of function called by the frame thread
// "context" is a platform-specific value or (#if JAVA_SOUND) a JNIEnv* pointer
typedef void (*HAE_FrameThreadProc)(void* context);

// Create and start the frame thread
extern int	HAE_CreateFrameThread(void* context, HAE_FrameThreadProc proc);

// Stop and destroy the frame thread
extern int	HAE_DestroyFrameThread(void* context);

// Make the frame thread sleep for the given number of milliseconds
extern int	HAE_SleepFrameThread(void* context, long msec);


#ifdef __cplusplus
	}
#endif

#endif	// HAE_API

