/*
 * @(#)JncClassBase.cpp	1.6 98/03/18
 *
 * Copyright 1997 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


// JncClassBase.c


#include "JncEnv.h"
#include "JncClassBase.h"
#include "JncClassCache.h"
#include "JncID.h"
#include "JncClass.h"   // for GetClassName()


////////////////////////////////////////////////// EXISTENCE:

JncClassBase::JncClassBase(JncEnv& env, jclass clazz)
: m_env(env)
, m_jobj(NULL)
{
    if (!clazz)
    {
        ERROR0("ERROR: JncClassBase{} instantiated with NULL jclass\n");
    }
#if JNC_CACHE_IDS
    m_cache = JncClassCache::Get(env, clazz);
#else
    m_class = clazz;
    m_deleteClass = JNI_FALSE;
#endif
}
JncClassBase::JncClassBase(JncEnv& env, ConStr className)
: m_env(env)
, m_jobj(NULL)
{
    if (!className)
    {
        ERROR0("ERROR: JncClassBase{} instantiated with NULL class name\n");
    }
#if JNC_CACHE_IDS
    m_cache = JncClassCache::Get(env, GetClass(env, className));
#else
    m_class = GetClass(env, className);
    m_deleteClass = JNI_TRUE;
#endif
}
JncClassBase::JncClassBase(JncEnv& env, jobject obj)
: m_env(env)
, m_jobj(obj)
{
    if (!obj)
    {
        ERROR0("ERROR: JncClassBase{} instantiated with NULL jobject\n");
    }
#if JNC_CACHE_IDS
    m_cache = NULL;
#else
    m_class = NULL;
    m_deleteClass = JNI_TRUE;
#endif
}

VIR_I JncClassBase::~JncClassBase()
{
#if !JNC_CACHE_IDS
    if (m_class && m_deleteClass)
    {
        m_env->DeleteLocalRef(m_class);
    }
#endif
}


////////////////////////////////////////////////// OBJECT-CREATIION OPERATIONS:

jobject JncClassBase::NewObject(ConStr signature, ...)
{
va_list         args;
va_start(args, signature);

jmethodID       const methodID = GetConstructorID(signature);

    if (methodID)
    {
    jobject         obj;

        obj = m_env->functions->
                NewObjectV(m_env, GetClass(), methodID, args);
        m_env.CheckForPendingException();
        return obj;
    }
    return NULL;

va_end(args);
}


////////////////////////////////////////////////// DATA:

jclass JncClassBase::GetClass()
{
#if JNC_CACHE_IDS
    if (GetCache())
    {
        return m_cache->GetClass();
    }
    return NULL;
#else
    if (!m_class)
    {
        // we'll only reach here if this was JncClassBase(env, obj) was used
        m_class = GetClass(m_env, m_jobj);
    }
    return m_class;
#endif
}
#if JNC_CACHE_IDS
JncClassCache* JncClassBase::GetCache()
{
    if (!m_env.IsPendingException())
    {
        if (!m_cache)
        {
        jclass      const clazz = GetClass(m_env, m_jobj);

            // we'll only reach here if this was JncClassBase(env, obj) was used
            if (clazz)
            {
                m_cache = JncClassCache::Get(m_env, clazz);
                m_env.DeleteLocalRef(clazz);
            }
        }
        return m_cache;
    }
    return NULL;
}
#endif
jsize JncClassBase::GetClassName(char* buffer, jsize maxLength)
{
    return GetClassName(GetEnv(), GetClass(), buffer, maxLength);
}
jsize JncClassBase::GetClassSignature(char* buffer, jsize maxLength)
{
    return GetClassSignature(GetEnv(), GetClass(), buffer, maxLength);
}

STA_I jclass JncClassBase::GetClass(JncEnv& env, ConStr className)
{
    if (!env.IsPendingException() && className)
    {
    jclass      const clazz = env->FindClass(className);
    // THROWS: ClassFormatError, ClassCircularityError,
    //          NoClassDefFoundError, OutOfMemoryError

        if (!clazz)
        {
            ERROR1("ERROR: Couldn't get class object for class, \"%s\"\n", className);
            env.SetPendingException();
        }
        return clazz;
    }
    return NULL;
}
STA_I jclass JncClassBase::GetClass(JncEnv& env, jobject obj)
{
    if (!env.IsPendingException() && obj)
    {
    jclass      const clazz = env->GetObjectClass(obj);
    // THROWS: nothing

        if (!clazz)
        {
            ERROR0("ERROR: Couldn't get class object for an object\n");
        }
        return clazz;
    }
    return NULL;
}
STA_I jclass JncClassBase::GetSuperclass(JncEnv& env, jclass clazz)
{
    if (!env.IsPendingException() && clazz)
    {
    jclass      const superclass = env->GetSuperclass(clazz);
    // THROWS: nothing

        if (!superclass)
        {
            ERROR0("ERROR: Couldn't get superclass\n");
        }
        return superclass;
    }
    return NULL;
}
STA_I jsize JncClassBase::GetClassName(JncEnv& env, jclass clazz, char* buffer, jsize maxLength)
{
JncClass    jncClass(env, clazz);

    return jncClass.GetName(buffer, maxLength);
}
STA_I jsize JncClassBase::GetClassSignature(JncEnv& env, jclass clazz, char* buffer, jsize maxLength)
{
JncClass    jncClass(env, clazz);

    return jncClass.GetSignature(buffer, maxLength);
}

jfieldID JncClassBase::GetFieldID(ConStr name, ConStr signature)
{
    if (!m_env.IsPendingException())
    {
    jclass      clazz;

#if JNC_CACHE_IDS
        if (GetCache())
        {
        JncID const*    const id = m_cache->FindID(name, signature);

            if (id)
            {
                return id->GetFieldID();
            }
        }
#endif
        clazz = GetClass();
        if (clazz)
        {
        jfieldID    fieldID;

            fieldID = m_env->GetFieldID(clazz, name, signature);
            // THROWS: NoSuchFieldError, OutOfMemoryError,
            //          ExceptionOnInitializerError
            if (fieldID)
            {
#if JNC_CACHE_IDS
                if (GetCache())
                {
                    m_cache->SaveID(name, signature, fieldID);
                }
#endif
                return fieldID;
            }
            ERROR2("ERROR: Failure to get ID for non-static field, \"%s:%s\"\n", name, signature);
            m_env.SetPendingException();
        }
    }
    return NULL;
}
jfieldID JncClassBase::GetStaticFieldID(ConStr name, ConStr signature)
{
    if (!m_env.IsPendingException())
    {
    jclass      clazz;

#if JNC_CACHE_IDS
        if (GetCache())
        {
        JncID const*    const id = m_cache->FindID(name, signature);

            if (id)
            {
                return id->GetFieldID();
            }
        }
#endif
        clazz = GetClass();
        if (clazz)
        {
        jfieldID    fieldID;

            fieldID = m_env->GetStaticFieldID(clazz, name, signature);
            // THROWS: NoSuchFieldError, OutOfMemoryError,
            //          ExceptionOnInitializerError
            if (fieldID)
            {
#if JNC_CACHE_IDS
                if (GetCache())
                {
                    m_cache->SaveID(name, signature, fieldID);
                }
#endif
                return fieldID;
            }
            ERROR2("ERROR: Failure to get ID for static field, \"%s:%s\"\n", name, signature);
            m_env.SetPendingException();
        }
    }
    return NULL;
}

jmethodID JncClassBase::GetConstructorID(ConStr signature)
{
    return GetMethodID("<init>", signature);
}
jmethodID JncClassBase::GetMethodID(ConStr name, ConStr signature)
{
    if (!m_env.IsPendingException())
    {
    jclass      clazz;

#if JNC_CACHE_IDS
        if (GetCache())
        {
        JncID const*    const id = m_cache->FindID(name, signature);

            if (id)
            {
                return id->GetMethodID();
            }
        }
#endif
        clazz = GetClass();
        if (clazz)
        {
        jmethodID   methodID;

            methodID = m_env->GetMethodID(clazz, name, signature);
            // THROWS: NoSuchMethodError, OutOfMemoryError,
            //          ExceptionOnInitializerError
            if (methodID)
            {
#if JNC_CACHE_IDS
                if (GetCache())
                {
                    m_cache->SaveID(name, signature, methodID);
                }
#endif
                return methodID;
            }
            ERROR2("ERROR: Failure to get ID for non-static method, \"%s%s\"\n", name, signature);
            m_env.SetPendingException();
        }
    }
    return NULL;
}
jmethodID JncClassBase::GetStaticMethodID(ConStr name, ConStr signature)
{
    if (!m_env.IsPendingException())
    {
    jclass      clazz;

#if JNC_CACHE_IDS
        if (GetCache())
        {
        JncID const*    const id = m_cache->FindID(name, signature);

            if (id)
            {
                return id->GetMethodID();
            }
        }
#endif
        clazz = GetClass();
        if (clazz)
        {
        jmethodID   methodID;

            methodID = m_env->GetStaticMethodID(clazz, name, signature);
            // THROWS: NoSuchMethodError, OutOfMemoryError,
            //          ExceptionOnInitializerError
            if (methodID)
            {
#if JNC_CACHE_IDS
                if (GetCache())
                {
                    m_cache->SaveID(name, signature, methodID);
                }
#endif
                return methodID;
            }
            ERROR2("ERROR: Failure to get ID for static method, \"%s%s\"\n", name, signature);
            m_env.SetPendingException();
        }
    }
    return NULL;
}


////////////////////////////////////////////////// VARIABLES:

const char JncClassBase::m_sigBoolean[] = "Z";
const char JncClassBase::m_sigChar[] = "C";
const char JncClassBase::m_sigByte[] = "B";
const char JncClassBase::m_sigShort[] = "S";
const char JncClassBase::m_sigInt[] = "I";
const char JncClassBase::m_sigLong[] = "J";
const char JncClassBase::m_sigFloat[] = "F";
const char JncClassBase::m_sigDouble[] = "D";
const char JncClassBase::m_sigString[] = "Ljava/lang/String;";
const char JncClassBase::m_sigBooleanArray[] = "[Z";
const char JncClassBase::m_sigCharArray[] = "[C";
const char JncClassBase::m_sigByteArray[] = "[B";
const char JncClassBase::m_sigShortArray[] = "[S";
const char JncClassBase::m_sigIntArray[] = "[I";
const char JncClassBase::m_sigLongArray[] = "[J";
const char JncClassBase::m_sigFloatArray[] = "[F";
const char JncClassBase::m_sigDoubleArray[] = "[D";

const char JncClassBase::m_sigReturnsVoid[] = "()V";
const char JncClassBase::m_sigReturnsBoolean[] = "()Z";
const char JncClassBase::m_sigReturnsChar[] = "()C";
const char JncClassBase::m_sigReturnsByte[] = "()B";
const char JncClassBase::m_sigReturnsShort[] = "()S";
const char JncClassBase::m_sigReturnsInt[] = "()I";
const char JncClassBase::m_sigReturnsLong[] = "()J";
const char JncClassBase::m_sigReturnsFloat[] = "()F";
const char JncClassBase::m_sigReturnsDouble[] = "()D";
const char JncClassBase::m_sigReturnsString[] = "()Ljava/lang/String;";
const char JncClassBase::m_sigReturnsBooleanArray[] = "()[Z";
const char JncClassBase::m_sigReturnsCharArray[] = "()[C";
const char JncClassBase::m_sigReturnsByteArray[] = "()[B";
const char JncClassBase::m_sigReturnsShortArray[] = "()[S";
const char JncClassBase::m_sigReturnsIntArray[] = "()[I";
const char JncClassBase::m_sigReturnsLongArray[] = "()[J";
const char JncClassBase::m_sigReturnsFloatArray[] = "()[F";
const char JncClassBase::m_sigReturnsDoubleArray[] = "()[D";


