/*
 * @(#)SampleTools.c	1.9 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
/*****************************************************************************/
/*
**	SampleTools.c
**
**	Tools for manipulating 'snd' resources.
**
**
**	History	-
**	1/31/93		Created
**	6/23/95		Integrated into SoundMusicSys Library
**	11/22/95	Added MACE support for CreateSampleFromSnd
**	12/15/95	Changed memory allocation to X_API
**				Pulled out most Mac specific code
**	1/18/96		Spruced up for C++ extra error checking
**	2/3/96		Removed extra includes
**	2/11/96		No longer relies on MacOS Sound.h
**	2/12/96		Changed Boolean to SMSBoolean
**	2/17/96		Added more platform compile arounds
**	3/28/96		Added XPutShort & XPutLong
**	3/29/96		Tried to figure out ima 4 to 1 and failed
**	4/2/96		Added more snd tools to set loop points, rates, and basekey
**	4/9/96		Changed AIFFSampleRate from extended80 to char [10]
**	4/21/96		Removed CPU edian issues by use XGetShort & XGetLong
**				Removed register usage in parameters
**	4/25/96		Added PV_Swap16BitSamples to deal with 16 bit sample on non 68k ordered platforms
**	5/9/96		Fixed byte swap problem
**	6/30/96		Changed font and re tabbed
**	7/6/96		Fixed possible problem with XGetSamplePtrFromSnd by no zeroing pSampleData
**	7/14/96		Fixed structure alignment problem for Pentium compilers
**	9/22/96		Added the ability to select an IMA 16 bit decompression, or an 8 bit version
**	11/14/96	Removed dependancy on MacSpecificSMS.h
**	12/15/96	Removed last remants of MacOS code and moved it all in MacSpecificSMS.c
**				Moved all snd resource types and structs into X_Formats.h
**	12/30/96	Changed copyright
**	1/30/97		Changed all '' constants to defines
**	5/3/97		Fixed a few potential problems that the MOT compiler found
**	6/20/97		Added XGetSampleNameFromID
**	7/17/97		Wrapped X_WORD_ORDER macros around PV_Swap16BitSamples. Saved code size,
**				and fixed potential bug because its word order based, not platform based
**	7/24/97		Added XCreateSoundObjectFromData & XGetSoundEmbeddedStatus & 
**				XSetSoundEmbeddedStatus
**	8/6/97		Changed XGetSampleNameFromID to be smarter about how many types
**				to search through.
**				Added XCreateAudioObjectFromData
**	8/7/97		Added XTranslateFromWaveformToSampleData
**	9/3/97		Wrapped a few more functions around USE_CREATION_API == TRUE
**				
*/
/*****************************************************************************/
#include "X_API.h"
#include "X_Formats.h"


#if X_WORD_ORDER != FALSE
// This is used to swap samples that are ordered in a 68k way, to the native platform
static void PV_Swap16BitSamples(short int *pSamples, long length, short channels)
{
	length *= channels;		// walk through channels also
	while (length--)
	{
		*pSamples = XGetShort(pSamples);
		pSamples++;
	}
}
#endif

// This is used on 8 bit samples that are sign encoded. We want 0 to 255 not -128 to 127
void XPhase8BitWaveform(unsigned char * pByte, long size)
{
	while (size--)
	{
		*pByte++ -= 128;
	}
}

// Given a Mac snd pointer, this will return the encoding type, and a pointer to a SoundHeader structure
static void * PV_GetSoundHeaderPtr(XPTR pRes, short int *pEncode)
{
	XSoundHeader	*pSndBuffer;
	short int		soundFormat;
	short int		numSynths, numCmds;
	long			offset;
	char			*pSndFormat;

	numSynths = 0;
	numCmds = 0;
	*pEncode = -1;
	pSndBuffer = NULL;
	if (pRes)
	{
		pSndFormat = (char *)pRes;
		soundFormat = XGetShort(pSndFormat);
		switch (soundFormat)
		{
			case firstSoundFormat:
				// look inside the format 1 resource and decode offsets
				numSynths = XGetShort(pSndFormat + 2);
				numCmds = XGetShort(pSndFormat + 4 + numSynths * 6);		// get number of commands
				break;
			case secondSoundFormat:
				numSynths = 0;		// format 2 has none
				numCmds = XGetShort(pSndFormat + 4);
				break;
			default:
				soundFormat = -1;
				break;
		}

		if (soundFormat != -1)	/* did we get the right format? */
		{
			/* compute address of sound header. 
			*/
			offset = 6 + 6 * numSynths + 8 * numCmds;
			pSndBuffer = (XSoundHeader *) ((char *)pRes + offset);
			*pEncode = pSndBuffer->encode & 0xFF;
		}
	}
	return pSndBuffer;
}


// Given a Mac sample, and loop points, this will change the data associated with it
#if USE_CREATION_API == TRUE
void XSetSoundLoopPoints(XPTR pRes, long loopStart, long loopEnd)
{
	register XSoundHeader		*pSndBuffer;
	register XCmpSoundHeader 	*pCmpBuffer;
	register XExtSoundHeader	*pExtBuffer;
	short int					encode;

	pSndBuffer = (XSoundHeader *)PV_GetSoundHeaderPtr(pRes, &encode);
	if (pSndBuffer)	/* did we get the right format? */
	{
		switch (encode)
		{
			case stdSH:	// standard header
				XPutLong(&pSndBuffer->loopStart, loopStart);
				XPutLong(&pSndBuffer->loopEnd, loopEnd);
				break;

			case extSH:	// extended header
				pExtBuffer = (XExtSoundHeader *)pSndBuffer;
				XPutLong(&pExtBuffer->loopStart, loopStart);
				XPutLong(&pExtBuffer->loopEnd, loopEnd);
				break;
				
			case cmpSH:	// compressed header
				pCmpBuffer = (XCmpSoundHeader *)pSndBuffer;
				XPutLong(&pCmpBuffer->loopStart, loopStart);
				XPutLong(&pCmpBuffer->loopEnd, loopEnd);
				break;
		}
	}
}

void XSetSoundEmbeddedStatus(XPTR pRes, XBOOL soundEmbedded)
{
	register XSoundHeader		*pSndBuffer;
	register XCmpSoundHeader 	*pCmpBuffer;
	register XExtSoundHeader	*pExtBuffer;
	short int					encode;

	pSndBuffer = (XSoundHeader *)PV_GetSoundHeaderPtr(pRes, &encode);
	if (pSndBuffer)	/* did we get the right format? */
	{
		switch (encode)
		{
//			case stdSH:	// standard header
//				break;

			case extSH:	// extended header
				pExtBuffer = (XExtSoundHeader *)pSndBuffer;
				pExtBuffer->soundIsEmbedded = soundEmbedded;
				break;
				
			case cmpSH:	// compressed header
				pCmpBuffer = (XCmpSoundHeader *)pSndBuffer;
				pCmpBuffer->soundIsEmbedded = soundEmbedded;
				break;
		}
	}
}

XBOOL XGetSoundEmbeddedStatus(XPTR pRes)
{
	register XSoundHeader		*pSndBuffer;
	register XCmpSoundHeader 	*pCmpBuffer;
	register XExtSoundHeader	*pExtBuffer;
	short int					encode;
	XBOOL						soundEmbedded;
	
	soundEmbedded = FALSE;
	pSndBuffer = (XSoundHeader *)PV_GetSoundHeaderPtr(pRes, &encode);
	if (pSndBuffer)	/* did we get the right format? */
	{
		switch (encode)
		{
//			case stdSH:	// standard header
//				break;

			case extSH:	// extended header
				pExtBuffer = (XExtSoundHeader *)pSndBuffer;
				soundEmbedded = pExtBuffer->soundIsEmbedded;
				break;
				
			case cmpSH:	// compressed header
				pCmpBuffer = (XCmpSoundHeader *)pSndBuffer;
				soundEmbedded = pCmpBuffer->soundIsEmbedded;
				break;
		}
	}
	return soundEmbedded;
}

// Given a Mac sample, and a sample rate, this will change the data associated with it
void XSetSoundSampleRate(XPTR pRes, unsigned long sampleRate)
{
	register XSoundHeader		*pSndBuffer;
	register XCmpSoundHeader 	*pCmpBuffer;
	register XExtSoundHeader	*pExtBuffer;
	short int					encode;

	pSndBuffer = (XSoundHeader *)PV_GetSoundHeaderPtr(pRes, &encode);
	if (pSndBuffer)	/* did we get the right format? */
	{
		switch (encode)
		{
			case stdSH:	// standard header
				XPutLong(&pSndBuffer->sampleRate, sampleRate);
				break;

			case extSH:	// extended header
				pExtBuffer = (XExtSoundHeader *)pSndBuffer;
				XPutLong(&pExtBuffer->sampleRate, sampleRate);
				break;
				
			case cmpSH:	// compressed header
				pCmpBuffer = (XCmpSoundHeader *)pSndBuffer;
				XPutLong(&pCmpBuffer->sampleRate, sampleRate);
				break;
		}
	}
}

// Given a Mac sample, and a sample rate, this will change the data associated with it
void XSetSoundBaseKey(XPTR pRes, short int baseKey)
{
	register XSoundHeader		*pSndBuffer;
	register XCmpSoundHeader 	*pCmpBuffer;
	register XExtSoundHeader	*pExtBuffer;
	short int					encode;

	pSndBuffer = (XSoundHeader *)PV_GetSoundHeaderPtr(pRes, &encode);
	if (pSndBuffer)	/* did we get the right format? */
	{
		switch (encode)
		{
			case stdSH:	// standard header
				pSndBuffer->baseFrequency = baseKey;
				break;

			case extSH:	// extended header
				pExtBuffer = (XExtSoundHeader *)pSndBuffer;
				pExtBuffer->baseFrequency = baseKey;
				break;
				
			case cmpSH:	// compressed header
				pCmpBuffer = (XCmpSoundHeader *)pSndBuffer;
				pCmpBuffer->baseFrequency = baseKey;
				break;
		}
	}
}

short int XGetSoundBaseKey(XPTR pRes)
{
	register XSoundHeader		*pSndBuffer;
	register XCmpSoundHeader 	*pCmpBuffer;
	register XExtSoundHeader	*pExtBuffer;
	short int					encode;
	short int					baseKey;

	baseKey = 60;
	pSndBuffer = (XSoundHeader *)PV_GetSoundHeaderPtr(pRes, &encode);
	if (pSndBuffer)	/* did we get the right format? */
	{
		switch (encode)
		{
			case stdSH:	// standard header
				baseKey = pSndBuffer->baseFrequency;
				break;

			case extSH:	// extended header
				pExtBuffer = (XExtSoundHeader *)pSndBuffer;
				baseKey = pExtBuffer->baseFrequency;
				break;
				
			case cmpSH:	// compressed header
				pCmpBuffer = (XCmpSoundHeader *)pSndBuffer;
				baseKey = pCmpBuffer->baseFrequency;
				break;
		}
	}
	return baseKey;
}
#endif	// USE_CREATION_API == TRUE


// This will return a pointer into a snd2 pointer block or create a new block if MACE compressed
XPTR XGetSamplePtrFromSnd(XPTR pRes, SampleDataInfo *pInfo)
{
	register XPTR				hSnd, newSound, monoSound;
	register XSoundHeader		*pSndBuffer;
	register XCmpSoundHeader 	*pCmpBuffer;
	register XExtSoundHeader	*pExtBuffer;
	long						offset, count;
	char						*pSampleData;
	char						*pLeft, *pRight;
	short int					encode;

	pSampleData = NULL;
	hSnd = NULL;
	pInfo->size = 0;		// if left alone, then wrong type of resource
	pInfo->frames = 0;
	pInfo->rate = rate22khz;
	pInfo->loopStart = 0;
	pInfo->loopEnd = 0;
	pInfo->baseKey = kMiddleC;
	pInfo->bitSize = 8;
	pInfo->channels = 1;
	pInfo->compressionType = C_NONE;

	pSndBuffer = (XSoundHeader *)PV_GetSoundHeaderPtr(pRes, &encode);
	if (pSndBuffer)	/* did we get the right format? */
	{
		/* compute address of sound header. 
		*/
		switch (encode)
		{
			case stdSH:	// standard header
				pSampleData = (char *)&pSndBuffer->sampleArea[0];
				pInfo->size = XGetLong(&pSndBuffer->length);
				pInfo->frames = pInfo->size;
				pInfo->loopStart = XGetLong(&pSndBuffer->loopStart);
				pInfo->loopEnd = XGetLong(&pSndBuffer->loopEnd);
				pInfo->baseKey = pSndBuffer->baseFrequency;
				pInfo->rate = XGetLong(&pSndBuffer->sampleRate);
				pInfo->channels = 1;
				pInfo->bitSize = 8;			// defaults for standard headers
				pInfo->pMasterPtr = pRes;
				break;

			case extSH:	// extended header
				pExtBuffer = (XExtSoundHeader *)pSndBuffer;
				pSampleData = (char *)&pExtBuffer->sampleArea[0];
				pInfo->channels = XGetLong(&pExtBuffer->numChannels);
				pInfo->bitSize = XGetShort(&pExtBuffer->sampleSize);
				pInfo->frames = XGetLong(&pExtBuffer->numFrames);
				pInfo->size = pInfo->frames * (pInfo->channels) * (pInfo->bitSize / 8);
				pInfo->loopStart = XGetLong(&pExtBuffer->loopStart);
				pInfo->loopEnd = XGetLong(&pExtBuffer->loopEnd);

				#if X_WORD_ORDER != FALSE
				// if we're on a non 68k platform, swap sample data for 16 bit samples because its stored in a 68k way
				if (pInfo->bitSize == 16)
				{
					PV_Swap16BitSamples((short *)pSampleData, pInfo->frames, pInfo->channels);
				}
				#endif
				pInfo->baseKey = pExtBuffer->baseFrequency;
				pInfo->rate = XGetLong(&pExtBuffer->sampleRate);
				pInfo->pMasterPtr = pRes;
				break;
				
			case cmpSH:	// compressed header
				pCmpBuffer = (XCmpSoundHeader *)pSndBuffer;
				if ((pSampleData = pCmpBuffer->samplePtr) == NULL)	/* get ptr to sample data */
				{
					pSampleData = (char *) pCmpBuffer->sampleArea;
				}
				pInfo->channels = XGetLong(&pCmpBuffer->numChannels);
				pInfo->bitSize = XGetShort(&pCmpBuffer->sampleSize);
				pInfo->frames = XGetLong(&pCmpBuffer->numFrames);
				pInfo->loopStart = XGetLong(&pCmpBuffer->loopStart);
				pInfo->loopEnd = XGetLong(&pCmpBuffer->loopEnd);
				pInfo->baseKey = pCmpBuffer->baseFrequency;
				pInfo->rate = XGetLong(&pCmpBuffer->sampleRate);

				encode = XGetShort(&pCmpBuffer->compressionID);
				switch(encode)
				{
					default:
						DEBUG_STR("\nInvalid compression ID");
						break;
	
					case fixedCompression:
						offset = XGetLong(&pCmpBuffer->format);
						switch (offset)
						{
							case C_IMA4:	// IMA 4 : 1
								offset = pInfo->frames * 64;	// Sound Manager defines 64 samples per packet
								pInfo->frames = offset;			// number of sample frames
								if ((pCmpBuffer->futureUse2_0 & 0x80) == 0)
								{
								// do a 16 bit decompression. As we decompress the IMA we build a 16 bit sample
									pInfo->size = offset * pInfo->bitSize * pInfo->channels;
									pInfo->bitSize = 16;			// must change to final output size
									newSound = XNewPtr(pInfo->size);
									pInfo->pMasterPtr = newSound;
									if (newSound)
									{
										XExpandMacADPCMtoXBitLinear((char *)pSampleData, pInfo->frames, pInfo->channels,
																	newSound, 16);
									}
								}
								else
								{
								// do a 8 bit decompression. As we decompress the IMA we build a 8 bit sample
									pInfo->size = (offset * pInfo->bitSize * pInfo->channels) / 2;
									pInfo->bitSize = 8;			// must change to final output size
									newSound = XNewPtr(pInfo->size);
									pInfo->pMasterPtr = newSound;
									if (newSound)
									{
										XExpandMacADPCMtoXBitLinear((char *)pSampleData, pInfo->frames, pInfo->channels,
																	newSound, 8);
									}
								}
								pInfo->compressionType = C_IMA4;
								pSampleData = (char *)newSound;
								break;
							case C_ULAW:	// ulaw 2 : 1
								pInfo->bitSize = 16;								// must change, its stored as 8 bit
								pInfo->size = pInfo->frames * pInfo->channels * 2;	// always 16 bit
								newSound = XNewPtr(pInfo->size);
								pInfo->pMasterPtr = newSound;
								if (newSound)
								{
									XExpandULawto16BitLinear(	(unsigned char *)pSampleData, 
																(short int *)newSound, 
																pInfo->frames, pInfo->channels);
								}
								pInfo->compressionType = C_ULAW;
								pSampleData = (char *)newSound;
								break;
						}
						break;
					case threeToOne:
						pInfo->size = pInfo->frames * (pInfo->channels) * (pInfo->bitSize / 8);
						pInfo->size *= 6;	// 2 bytes at 3:1 is 6 bytes for a packet, 1 byte at 6:1 is 6 bytes too
						newSound = XNewPtr(pInfo->size);
						pInfo->pMasterPtr = newSound;
						if (newSound)
						{
							if (pInfo->channels == 1)
							{
								XExpandMace1to3(pSampleData, newSound, pInfo->frames, NULL, NULL, 1, 1);
								pInfo->frames *= 6;			// adjust the frame count to equal the real frames
							}
							else
							{
								monoSound = XNewPtr(pInfo->size / 2);
								if (monoSound)
								{
									XExpandMace1to3(pSampleData, newSound, pInfo->frames, NULL, NULL, pInfo->channels, 1);
									XExpandMace1to3(pSampleData, monoSound, pInfo->frames, NULL, NULL, pInfo->channels, 2);
									pLeft = (char *)newSound;
									pRight = (char *)monoSound;
									pInfo->frames *= 6;			// adjust the frame count to equal the real frames
									offset = pInfo->frames - 1;
									// copy the data into a stereo sample block, copy backwards so that we don't have to create
									// two blocks of data
									for (count = offset; count >= 0; count--)
									{
										pLeft[count*2+0] = pLeft[count];
										pLeft[count*2+1] = pRight[count];
									}
									XDisposePtr(monoSound);
								}
							}
						}
						pInfo->compressionType = C_MACE3;
						pSampleData = (char *)newSound;
						break;
					case sixToOne:
						pInfo->size = pInfo->frames * (pInfo->channels) * (pInfo->bitSize / 8);
						pInfo->size *= 6;	// 2 bytes at 3:1 is 6 bytes for a packet, 1 byte at 6:1 is 6 bytes too
						newSound = XNewPtr(pInfo->size);
						pInfo->pMasterPtr = newSound;
						if (newSound)
						{
							if (pInfo->channels == 1)
							{
								XExpandMace1to6(pSampleData, newSound, pInfo->frames, NULL, NULL, 1, 1);
								pInfo->frames *= 6;			// adjust the frame count to equal the real frames
							}
							else
							{
								monoSound = XNewPtr(pInfo->size / 2);
								if (monoSound)
								{
									XExpandMace1to6(pSampleData, newSound, pInfo->frames, NULL, NULL, pInfo->channels, 1);
									XExpandMace1to6(pSampleData, monoSound, pInfo->frames, NULL, NULL, pInfo->channels, 2);
									pLeft = (char *)newSound;
									pRight = (char *)monoSound;
									pInfo->frames *= 6;			// adjust the frame count to equal the real frames
									offset = pInfo->frames - 1;
									// copy the data into a stereo sample block, copy backwards so that we don't have to create
									// two blocks of data
									for (count = offset; count >= 0; count--)
									{
										pLeft[count*2+0] = pLeft[count];
										pLeft[count*2+1] = pRight[count];
									}
									XDisposePtr(monoSound);
								}
							}
						}
						pInfo->compressionType = C_MACE6;
						pSampleData = (char *)newSound;
						break;
				}
				break;
		}
	}
// verify loops
	{
		XBOOL	pass;
	
		pass = TRUE;
		if ((long)pInfo->loopStart < 0)
		{
			pass = FALSE;
		}
		if (pInfo->loopStart > pInfo->loopEnd)
		{
			pass = FALSE;
		}
		if (pInfo->loopEnd > pInfo->frames)
		{
			pass = FALSE;
		}
		if (pass == FALSE)
		{
			pInfo->loopStart = 0;
			pInfo->loopEnd = 0;
		}
	}
	return pSampleData;
}

// Given a sample ID, this will search through sample types and return a 'C' string
// of the resource name of the currently open resource files
#if USE_CREATION_API == TRUE
XBOOL XGetSampleNameFromID(short int sampleSoundID, char *cName)
{
	static long	sampleType[] = {ID_CSND, ID_ESND, ID_SND};
	short int	count;
	XBOOL		bad;

	bad = FALSE;
	if (cName)
	{
		for (count = 0; count < (sizeof(sampleType) / sizeof(long)); count++)
		{
			cName[0] = 0;
			XGetResourceName(sampleType[count], sampleSoundID, cName);
			if (cName[0])
			{
				break;
			}
		}
		if (cName[0] == 0)
		{
			bad = TRUE;
		}
	}
	return bad;
}
#endif

// Given a structure filled with no compression, create a type 1 sound object.
// Will return NULL if out of memory, failed to get the right info from the
// SampleDataInfo structure, and if pPCMData is NULL.
// pSampleInfo->pMasterPtr is ignored, pPCMData points to the PCM data formated
// the way pSampleInfo describes
#if USE_CREATION_API == TRUE
XPTR XCreateSoundObjectFromData(XPTR pPCMData, SampleDataInfo *pSampleInfo)
{
	XPTR			pNewSampleObject;
	XExtSndHeader1	*pSnd;

	pNewSampleObject = NULL;
	if (pSampleInfo)
	{
		if (pPCMData && pSampleInfo->size)
		{
			pNewSampleObject = XNewPtr((long)sizeof(XExtSndHeader1) + pSampleInfo->size);
			pSnd = (XExtSndHeader1 *)pNewSampleObject;
			if (pSnd)
			{
				XPutShort(&pSnd->sndHeader.type, firstSoundFormat);
				XPutShort(&pSnd->sndHeader.numModifiers, 1);
				XPutShort(&pSnd->sndHeader.modNumber, 5);
				XPutLong(&pSnd->sndHeader.modInit, 224);
				XPutShort(&pSnd->sndHeader.numCommands, 1);
				XPutShort(&pSnd->sndHeader.cmd, 0x8000 + bufferCmd);
				XPutShort(&pSnd->sndHeader.param1, 0);
				XPutLong(&pSnd->sndHeader.param2, sizeof(XSoundFormat1));

				XPutLong(&pSnd->sndBuffer.samplePtr, 0L);
				XPutLong(&pSnd->sndBuffer.numChannels, pSampleInfo->channels);
				XPutShort(&pSnd->sndBuffer.sampleSize, pSampleInfo->bitSize);
				XPutLong(&pSnd->sndBuffer.sampleRate, pSampleInfo->rate);
				XPutLong(&pSnd->sndBuffer.loopStart, pSampleInfo->loopStart);
				XPutLong(&pSnd->sndBuffer.loopEnd, pSampleInfo->loopEnd);
				pSnd->sndBuffer.encode = extSH;		// extended header
				pSnd->sndBuffer.baseFrequency = pSampleInfo->baseKey;
				XPutLong(&pSnd->sndBuffer.numFrames, pSampleInfo->frames);
				XBlockMove(pPCMData, &pSnd->sndBuffer.sampleArea, pSampleInfo->size);
			}
		}
	}
	return pNewSampleObject;
}
#endif	// USE_CREATION_API == TRUE


// Given a structure filled with no compression, create a AudioResource object.
// Will return NULL if out of memory, failed to get the right info from the
// SampleDataInfo structure, and if pPCMData is NULL.
// pSampleInfo->pMasterPtr is ignored, pPCMData points to the PCM data formated
// the way pSampleInfo describes
#if USE_CREATION_API == TRUE
static AudioResource *XCreateAudioObjectFromData(XPTR pPCMData, SampleDataInfo *pSampleInfo)
{

	AudioResource	*pNewSampleObject;

	pNewSampleObject = NULL;
	if (pSampleInfo)
	{
		if (pPCMData && pSampleInfo->size)
		{
			pNewSampleObject = (AudioResource *)XNewPtr((long)sizeof(AudioResource) + pSampleInfo->size);
			if (pNewSampleObject)
			{
				XPutLong(&pNewSampleObject->version, AUDIO_OBJECT_VERSION);
				XPutLong(&pNewSampleObject->dataLength, pSampleInfo->size);
				XPutLong(&pNewSampleObject->sampleFrames, pSampleInfo->frames);
				XPutLong(&pNewSampleObject->dataOffset, sizeof(AudioResource));
				XPutLong(&pNewSampleObject->audioType, AUDIO_SND);
				XPutLong(&pNewSampleObject->nameResourceType, ID_NULL);
				XPutLong(&pNewSampleObject->sampleRate, pSampleInfo->rate);
				XPutLong(&pNewSampleObject->loopStart, pSampleInfo->loopStart);
				XPutLong(&pNewSampleObject->loopEnd, pSampleInfo->loopEnd);
				XPutShort(&pNewSampleObject->baseMidiKey, pSampleInfo->baseKey);
				XPutShort(&pNewSampleObject->bitSize, pSampleInfo->channels);
				XPutShort(&pNewSampleObject->channels, pSampleInfo->bitSize);
				XBlockMove(pPCMData, (char *)(&pNewSampleObject->firstSampleFiller) + sizeof(long),
											pSampleInfo->size);
			}
		}
	}
	return pNewSampleObject;
}
#endif	// USE_CREATION_API == TRUE


#if USE_CREATION_API == TRUE
static XPTR XGetAudioObjectFromMemory(AudioResource *pAudioObject, SampleDataInfo *pInfo)
{
	XPTR				pSampleData;
	unsigned long		size;

	pSampleData = NULL;
	pInfo->size = 0;		// if left alone, then wrong type of resource
	pInfo->frames = 0;
	pInfo->rate = rate22khz;
	pInfo->loopStart = 0;
	pInfo->loopEnd = 0;
	pInfo->baseKey = kMiddleC;
	pInfo->bitSize = 8;
	pInfo->channels = 1;
	pInfo->compressionType = C_NONE;
	if (pAudioObject)
	{
		if (XGetLong(&pAudioObject->version) == AUDIO_OBJECT_VERSION)
		{
			switch (XGetLong(&pAudioObject->audioType))
			{
				case AUDIO_SND:
					size = XGetLong(&pAudioObject->dataLength);
					if (size)
					{
						pInfo->size = size;
						pInfo->frames = XGetLong(&pAudioObject->sampleFrames);
						pInfo->rate = XGetLong(&pAudioObject->sampleRate);
						pInfo->loopStart = XGetLong(&pAudioObject->loopStart);
						pInfo->loopEnd = XGetLong(&pAudioObject->loopEnd);
						pInfo->baseKey = XGetShort(&pAudioObject->baseMidiKey);
						pInfo->bitSize = XGetShort(&pAudioObject->bitSize);
						pInfo->channels = XGetShort(&pAudioObject->channels);
						pSampleData = XGetSamplePtrFromSnd((XPTR)((char *)(&pAudioObject->firstSampleFiller) + sizeof(long)), pInfo);
					}
					break;
			}
		}
	}
	return (XPTR)pSampleData;
}
#endif	// USE_CREATION_API == TRUE

// Translate a GM_Waveform structure into a SampleDataInfo structure
#if USE_CREATION_API == TRUE
void XTranslateFromWaveformToSampleData(GM_Waveform *pSource, SampleDataInfo *pDest)
{
	if (pSource && pDest)
	{
		pDest->rate = pSource->sampledRate;
		pDest->frames = pSource->waveFrames;
		pDest->size = pSource->waveSize;
		pDest->loopStart = pSource->startLoop;
		pDest->loopEnd = pSource->endLoop;
		pDest->bitSize = pSource->bitSize;
		pDest->channels = pSource->channels;
		pDest->baseKey = pSource->baseMidiPitch;
		pDest->theID = pSource->waveformID;
		pDest->compressionType = ID_NULL;
		pDest->pMasterPtr = pSource->theWaveform;
	}
}
#endif

/* EOF of SampleTools.c
*/

