/*
 * @(#)linlab.c	1.5 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * File: linlab.c	1.4 98/02/17
 *
 *	Contains:	This module contains functions to create
 *			the LinearLab FuT.
 *
 *			Created by msm, November 7, 1997
 *
 *********************************************************************
 *    COPYRIGHT (c) 1997-1998 Eastman Kodak Company
 *    As an unpublished work pursuant to Title 17 of the United
 *    States Code.  All rights reserved.
 *********************************************************************
 */


#include "kcptmgr.h"
#include "makefuts.h"

/*---------------------------------------------------------------------
 *  Defaults and other definitions
 *---------------------------------------------------------------------
 */
#define	NEUTRALBYTE	0.50196078431373	/* 128/255; a* = b* = 0.0 */

typedef struct fData_s {
	double		neutralgrid;
	KpInt32_t	chan;
} fData_t, FAR* fData_p;

/*---------------------------------------------------------------------
 *  xfun, yfun, zfun -- piecewise-linear input mappings, with gridpoint at
 *	neutral (a* = b* = 0; 8-bit encoding = 128)
 *---------------------------------------------------------------------
 */
static double xfunc (double x, KpGenericPtr_t dataP)
{
	if (dataP) {}

	return x;
}

static double
	yfunc (double y, KpGenericPtr_t dataP)
{
double	delta, neutralgrid;

	neutralgrid = ((fData_p) dataP)->neutralgrid;

	delta = y - NEUTRALBYTE;
	if (delta < 0.0) {
		y = neutralgrid * (y / NEUTRALBYTE);
	}
	else {
		y = 1.0 - (1.0 - neutralgrid) * ((1.0 - y) / (1.0 - NEUTRALBYTE));
	}

	return RESTRICT (y, 0.0, 1.0);
}

static double
	zfunc (double z, KpGenericPtr_t dataP)
{
double	delta, neutralgrid;

	neutralgrid = ((fData_p) dataP)->neutralgrid;

	delta = z - NEUTRALBYTE;
	if (delta < 0.0) {
		z = neutralgrid * (z / NEUTRALBYTE);
	}
	else {
		z = 1.0 - (1.0 - neutralgrid) * ((1.0 - z) / (1.0 - NEUTRALBYTE));
	}

	return RESTRICT (z, 0.0, 1.0);

}


/*------------------------------------------------------------------
 *  gfunc, ofunc -- quantization call-back functions for FuT library
 *
 * representing L* in [0, 100], a* and b* in [-128, 127]
 */
static fut_gtbldat_t
	gfunc (double *args, KpGenericPtr_t dataP)
{
double	p, l, a, b;

	l = args[0];
	a = args[1];
	b = args[2];

     /* Compute grid-table entry:  */
	switch (((fData_p) dataP)->chan)
	{
	   case 1:
			p = l;
			break;

	   case 2:
			p = a;
			break;

	   case 3:
			p = b;
			break;
	}

	return QUANT (p, FUT_GRD_MAXVAL);
}


/*---------------------------------------------------------------------
 *  outfun -- piecewise-linear output-table mappings; inverse of xfun, yfun, zfun
 *---------------------------------------------------------------------
 */
static fut_otbldat_t
	ofunc (fut_gtbldat_t q, KpGenericPtr_t dataP)
{
double	p, delta, neutralgrid;

	neutralgrid = ((fData_p) dataP)->neutralgrid;

	p = DEQUANT (q, FUT_GRD_MAXVAL);

	switch (((fData_p) dataP)->chan)
	{
	   case 1:   /* L* */
			break;

	   case 2:   /* a */
	   case 3:   /* b */
			delta = p - neutralgrid;
			if (delta < 0.0)
			   p = NEUTRALBYTE * (p / neutralgrid);
			else
			   p = 1.0 - (1.0 - NEUTRALBYTE) * ((1.0 - p) / (1.0 - neutralgrid));
			break;
	}

	return QUANT (p, FUT_MAX_PEL12);
}


/*----------------------------------------------------------------------
 *  get_linlab_fut --	construct an LinearLab FuT
 *						to establish a grid for composition

 		MUST be called from within a critical section!!!!!
 *----------------------------------------------------------------------
 */
fut_ptr_t
	get_linlab_fut (KpInt32_t size)
{
fut_ptr_t		futp;
fut_itbl_ptr_t	itblx, itbly, itblz;
fut_gtbl_ptr_t	gtbl [3];
fut_otbl_ptr_t	otbl [3];
KpInt32_t		ichan;
KpInt32_t		iomask, sizeArray[3];
KpInt32_t		xsize, ysize, zsize;
fData_t			fData;
double			neutralgrid;

/*  set_size -- set grid-table dimensions */
	neutralgrid = (double)(size / 2) / (double)(size - 1);

	/* assume all dimensions are the same */
	xsize = ysize = zsize = size;

/* Compute shared input tables:  */
	fData.neutralgrid = neutralgrid;

	itblx = fut_new_itblEx (xsize, xfunc, NULL);
	itbly = fut_new_itblEx (ysize, yfunc, &fData);
	itblz = fut_new_itblEx (zsize, zfunc, &fData);

/* Compute grid tables and output tables:  */
	iomask = FUT_IN(FUT_XYZ);

	sizeArray[0] = xsize;
	sizeArray[1] = ysize;
	sizeArray[2] = zsize;

	fData.neutralgrid = neutralgrid;

	for (ichan = 0; ichan < 3; ichan++) {
		fData.chan = ichan + 1;
		gtbl [ichan] = fut_new_gtblEx (iomask, gfunc, &fData, sizeArray);
		otbl [ichan] = fut_new_otblEx (ofunc, &fData);
	}

/* Assemble FuT:  */
	iomask = FUT_IN(FUT_XYZ) | FUT_OUT(FUT_XYZ);
	futp = fut_new (iomask, itblx, itbly, itblz,
							gtbl [0], otbl [0],
							gtbl [1], otbl [1],
							gtbl [2], otbl [2]);

	if ( futp->idstr != NULL ) {
		fut_free_idstr (futp->idstr);
		futp->idstr = NULL;
	}

	fut_free_tbls ((KpInt32_t FAR *)itblx, itbly, itblz,
							gtbl [0], gtbl [1], gtbl [2],
							otbl [0], otbl [1], otbl [2], NULL);

	return (futp);
}

