/*
 * @(#)ptiomem.c	1.4 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * File: ptiomem.c	2.46 98/02/17
 *
 *	Contains:	functions to read and write binary PT "memory" files.
 *
 *	The external database is responsible for loading and storing
 *	a block of memory which contains all of the PT information.
 *	It needs to know only the location and size of this block.
 *
 ****************************************************************
 *    COPYRIGHT (c) 1992-1998 Eastman Kodak Company.
 *    As an unpublished work pursuant to Title 17 of the United
 *    States Code.  All rights reserved.
 *****************************************************************
 */

#include <string.h>

#include "fut.h"
#include "fut_io.h"
#include "fut_util.h"
#include "kcms_sys.h"
#include "kcmptlib.h"
#include "kcptmgrd.h"
#include "attrib.h"
#include "kcptmgr.h"

/* prototypes */
static PTErr_t gridDimValid (PTType_t, PTRefNum_t, PTRefNum_p);

/* PTCheckIn reads the private data header and attributes of a PT in an
 * external memory block, then enters it into the checked in PT list.
 */

PTErr_t
	PTCheckIn (	PTRefNum_p	PTRefNumP,
				PTAddr_t	PTAddr)
{
PTErr_t		errnum;
KpFd_t		fd;
KpHandle_t	PTHdr = NULL, PTAttr = NULL;
#if !defined KCP_ICC_ONLY
fut_hdr_t	FAR *futp;
#endif
PTType_t	format;

/* Check for valid PTRefNumP */
	if (Kp_IsBadWritePtr(PTRefNumP, sizeof(*PTRefNumP))) {
		errnum = KCP_BAD_PTR;
		goto ErrOut4;
	}

/* initialize memory file manager to read the PT */
	if (KpOpen (NULL, "m", &fd, NULL, (KpGenericPtr_t)PTAddr, PTCHECKINSIZE) != KCMS_IO_SUCCESS) {
		errnum = KCP_SYSERR_1;
		goto ErrOut4;
	}

/* read the header info */
	errnum = TpReadHdr (&fd, &PTHdr, &format);
	if (errnum != KCP_SUCCESS) {
		goto ErrOut1;
	}

#if !defined KCP_ICC_ONLY
	switch (format) {
	case PTTYPE_FUTF:
	case PTTYPE_FTUF:						/* discard the attribute info */
		futp = lockBuffer (PTHdr);	/* get the attributes */
		errnum = readAttributes (&fd, futp->idstr_len, &PTAttr);
		unlockBuffer (PTHdr);
		if (errnum != KCP_SUCCESS) {
			goto ErrOut2;
		}

		if (PTMemTest () == 0) {		/* not enough memory to continue */
			errnum = KCP_NO_CHECKIN_MEM;
			goto ErrOut3;
		}

		break;

	default:
		break;
	}
#endif

/* enter PT into list */
	errnum = registerPT (PTHdr, PTAttr, PTRefNumP);
	if (errnum != KCP_SUCCESS) {
		goto ErrOut3;
	}

	switch (format) {
	case PTTYPE_MFT1:
	case PTTYPE_MFT2:						/* discard the attribute info */

		errnum = TpGenerateAttr (*PTRefNumP); /* generate constant attributes */
		break;

	default:
		break;
	}

	if (errnum != KCP_SUCCESS) {
ErrOut3:
		(void) freeAttributes(PTAttr);	/* free the attributes */

#if !defined KCP_ICC_ONLY
ErrOut2:
#endif
		(void) TpFreeHdr (PTHdr);		/* free the header */
	}

ErrOut1:
	(void) Kp_close (&fd);

ErrOut4:
	return (errnum);
}


/* PTActivate reads the PT data from an external memory block and
 * sets up the technology specific PT memory structures.  Before loading
 * the PT, check to make sure that it matches the "checked in" info.
 */
PTErr_t
	PTActivate(	PTRefNum_t	PTRefNum,
				KpInt32_t	mBlkSize,
				PTAddr_t	PTAddr)
{
PTErr_t	errnum;

	errnum = doActivate( PTRefNum, mBlkSize, PTAddr, 1L);

	return (errnum);
}

PTErr_t
	doActivate(	PTRefNum_t	PTRefNum,
				KpInt32_t	mBlkSize,
				PTAddr_t	PTAddr,
				KpInt32_t	crcMode)
{
PTErr_t		errnum;
KpFd_t		fd;
KpHandle_t	PTHdr, PTData, PTHdr2;
fut_hdr_t	FAR *futp;
#if !defined KCMS_NO_CRC
KpChar_t 	strCRCmade[KCM_MAX_ATTRIB_VALUE_LENGTH+1];
KpChar_t 	strCRCfound[KCM_MAX_ATTRIB_VALUE_LENGTH+1];
KpInt32_t	crcAttrSize, crc32;
#endif
KpInt32_t	attrSize;
PTType_t	format;

#if defined KCMS_NO_CRC
	if (crcMode) {}
#endif

	errnum = getPTStatus (PTRefNum);
	if (errnum != KCP_PT_INACTIVE) {
		goto ErrOut1;
	}

	/* initialize memory file manager */
	if (KpOpen (NULL, "m", &fd, NULL, (KpGenericPtr_t)PTAddr, mBlkSize) != KCMS_IO_SUCCESS) {
		errnum = KCP_SYSERR_1;
		goto ErrOut1;
	}

/* read in the encoded header, verify that its the same as the original, then discard */
	errnum = TpReadHdr (&fd, &PTHdr2, &format);
	if (errnum != KCP_SUCCESS ) {
		goto ErrOut2;
	}

/* get and save size of attributes */
    futp = lockBuffer (PTHdr2);
	attrSize = futp->idstr_len;
	unlockBuffer (PTHdr2);

	PTHdr = getPTHdr (PTRefNum);	/* get the original PT header */

/* make sure the PT header and checkin info match */
	errnum = TpCompareHdr (PTHdr, PTHdr2);

	(void) TpFreeHdr (PTHdr2);		/* free the header */

	if (errnum != KCP_SUCCESS) {	/* then check for an error in hdrVerify */
		goto ErrOut2;
	}

#if !defined KCP_ICC_ONLY
	switch (format) {
	case PTTYPE_FUTF:						/* discard the attribute info */
	case PTTYPE_FTUF:
		if (Kp_skip (&fd, attrSize) != KCMS_IO_SUCCESS){ /* may have been setAttribute after checkin */
			errnum = KCP_PTERR_3;
			goto ErrOut2;
		}

		break;

	default:
		break;
	}
#endif

	errnum = TpReadData (&fd, format, PTRefNum, PTHdr, &PTData);		/* get the PT data */
	if (errnum == KCP_SUCCESS) {
		if (PTMemTest () == 0) {			/* enough memory to continue operations? */
			errnum = KCP_NO_ACTIVATE_MEM;
			goto ErrOut3;
		}

	#if !defined KCMS_NO_CRC
		if (crcMode != 0) {
			errnum = TpCalCrc (PTHdr, PTData, &crc32);	/* calculate the CRC */
			if (errnum == KCP_SUCCESS) {
				KpItoa(crc32, strCRCmade);
				crcAttrSize = KCM_MAX_ATTRIB_VALUE_LENGTH;
				errnum = PTGetAttribute(PTRefNum, KCM_CRC, &crcAttrSize, strCRCfound);

				if (errnum == KCP_INVAL_PTA_TAG) {		/* if not present, just set it */
					PTSetAttribute(PTRefNum, KCM_CRC, strCRCmade);
					errnum = KCP_SUCCESS;
				}
				else {
					if ((errnum == KCP_SUCCESS)			/* if present, must match */
						&& (strcmp (strCRCmade, strCRCfound) != 0)) {
						errnum = KCP_INCON_PT;
						goto ErrOut3;
					}
				}
			}
		}
	#endif
	}

	if (errnum == KCP_SUCCESS) {	/* Everything's OK, now activate */
		makeActive (PTRefNum, PTData);
	}

ErrOut2:
	(void) Kp_close (&fd);
ErrOut1:
	return (errnum);


ErrOut3:
	(void) TpFreeData (PTData);	/* Release the PT memory */
	goto ErrOut2;
}


/* register a new PT using externally defined header and data */
PTErr_t
	PTNewPT (	PTAddr_p	PTHdr,
				PTAddr_p	PTData,
				PTRefNum_p	PTRefNumP)
{
PTErr_t errnum;

/* Check for valid PTHdr, PTData */
	if (Kp_IsBadReadPtr(PTHdr, sizeof(*PTHdr))) {
		errnum = KCP_BAD_PTR;
		goto ErrOut1;
	}

	if (Kp_IsBadReadPtr(PTData, sizeof(*PTData))) {
		errnum = KCP_BAD_PTR;
		goto ErrOut1;
	}

	if (Kp_IsBadWritePtr(PTRefNumP, sizeof(*PTRefNumP))) {
		errnum = KCP_BAD_PTR;
		goto ErrOut1;
	}

	errnum = registerPT (PTHdr, NULL, PTRefNumP);	/* enter PT into list */

	if ((errnum == KCP_SUCCESS) && (PTData != NULL)) {
		makeActive (*PTRefNumP, PTData);		/* activate the new PT */

		errnum = TpGenerateAttr (*PTRefNumP); /* generate constant attributes */
		if (errnum != KCP_SUCCESS) {
			(void) PTCheckOut(*PTRefNumP);	/* clean up if error */
		}
	}

ErrOut1:
	return (errnum);
}


/* deactivate a checked in PT */
PTErr_t PTDeActivate (PTRefNum_t PTRefNum)
{
PTErr_t errnum;

	errnum = getPTStatus (PTRefNum);		/* must currently be active */

	if ((errnum == KCP_PT_ACTIVE) || (errnum == KCP_SERIAL_PT)) {
		errnum = makeInActive (PTRefNum);	/* deactivate the PT */
	}

	return errnum;
}


/* check out a checked in PT. */
PTErr_t PTCheckOut (PTRefNum_t PTRefNum)
{
PTErr_t 	errnum;

	errnum = PTDeActivate(PTRefNum);				/* free the PT data */
	if ((errnum == KCP_SUCCESS) || (errnum == KCP_PT_INACTIVE)) {
		errnum = makeCheckedOut (PTRefNum);		/* release PT table */
	}

	return (errnum);
}


/* PTGetPTInfo returns the header, attribute, and data of a PT. */
PTErr_t
	PTGetPTInfo (	PTRefNum_t	PTRefNum,
					PTAddr_p*	PTHdr,
					PTAddr_p*	PTAttr,
					PTAddr_p*	PTData)
{
PTErr_t errnum;

	errnum = getPTStatus (PTRefNum);

	if ((errnum == KCP_PT_ACTIVE) || (errnum == KCP_PT_INACTIVE) || (errnum == KCP_SERIAL_PT)) {

		/* Check for valid ptrs */
		if ((PTHdr != NULL) && (Kp_IsBadWritePtr (PTHdr, sizeof(*PTHdr)))) {
			errnum = KCP_BAD_PTR;
			goto ErrOut1;
		}
		if ((PTAttr != NULL) && (Kp_IsBadWritePtr (PTAttr, sizeof(*PTAttr)))) {
			errnum = KCP_BAD_PTR;
			goto ErrOut1;
		}
		if ((PTData != NULL) && (Kp_IsBadWritePtr (PTData, sizeof(*PTData)))) {
			errnum = KCP_BAD_PTR;
			goto ErrOut1;
		}

		/* return the requested information */
		if (PTHdr != NULL) {
			*PTHdr = getPTHdr (PTRefNum);
		}
		if (PTAttr != NULL) {
			*PTAttr = getPTAttr (PTRefNum);
		}
		if ( ((errnum == KCP_PT_ACTIVE) || (errnum == KCP_SERIAL_PT)) && (PTData != NULL)) {
			*PTData = getPTData (PTRefNum);
		}
	}

ErrOut1:
	return (errnum);
}


/* PTGetSize calculates the size of a PT in FuT format.  */
PTErr_t
	PTGetSize (	PTRefNum_t	PTRefNum,
				KpInt32_p	mBlkSize)
{
PTErr_t		errnum;

	errnum = PTGetSizeF (PTRefNum, PTTYPE_FUTF, mBlkSize);

	return (errnum);
}


/* PTGetSizeF calculates the size of a PT in any format.  */
PTErr_t
	PTGetSizeF (PTRefNum_t	PTRefNum,
				PTType_t	format,
				KpInt32_p	mBlkSize)
{
PTErr_t		errnum;

	errnum = doGetSizeF (PTRefNum, format, mBlkSize, 1L);

	return (errnum);
}


PTErr_t
	doGetSizeF (PTRefNum_t	PTRefNum,
				PTType_t	format,
				KpInt32_p	mBlkSize,
				KpInt32_t	crcMode)
{
PTErr_t		errnum;
KpInt32_t	extSize, intSize;
KpHandle_t	PTHdr, PTData;
#if !defined KCP_ICC_ONLY
KpHandle_t	PTAttr;
#endif
#if !defined KCMS_NO_CRC
KpChar_t 	strCRCmade[KCM_MAX_ATTRIB_VALUE_LENGTH+1];
KpInt32_t	crc32;
#endif
PTRefNum_t	auxPTRefNum;

#if defined KCMS_NO_CRC
	if (crcMode) {}
#endif

	errnum = getPTStatus (PTRefNum);

	if ((errnum == KCP_PT_ACTIVE) || (errnum == KCP_PT_INACTIVE) || (errnum == KCP_SERIAL_PT)) {

		if (Kp_IsBadWritePtr(mBlkSize, sizeof(*mBlkSize)))
			return (KCP_BAD_PTR);

		switch (format) {
#if !defined KCP_ICC_ONLY
			case PTTYPE_FUTF:
			case PTTYPE_FTUF:
				extSize = sizeof(PT_t);		/* size of external header */
				break;
#endif

			case PTTYPE_MFT1:
			case PTTYPE_MFT2:
				extSize = (2 * sizeof (KpInt32_t)) + (4 * sizeof (u_int8))
					+ (MF_MATRIX_DIM * MF_MATRIX_DIM * sizeof (KpInt32_t));
				break;

			default:
				return (KCP_INVAL_PTTYPE);
		}

		if ((errnum == KCP_PT_ACTIVE) || (errnum == KCP_SERIAL_PT)) {

		/* when active, add size of external PT data block */
			PTHdr = getPTHdr (PTRefNum);
			PTData = getPTData (PTRefNum);
			intSize = TpGetDataSize (PTHdr, PTData, format);
			if (intSize == 0) {

				/* 	TpGetDataSize will return 0 if the grid table
					dimensions are not valid for the format specified.
					PTGetPTF will attempt to resize the grid table of
					the PT.  Check if that resizing is possible				*/
				errnum = gridDimValid (format, PTRefNum, NULL);
  				if (errnum != KCP_SUCCESS) {
					return errnum;
				}

				/*	The PT will be resized, by precomposing it with the
					appropriate auxiliary PT.  The size of the aux PT will
					determine the size of the new PT.  Find the aux PT to
					be used.												*/
				errnum = getResizeAuxPT (PTRefNum, &auxPTRefNum);
				if (errnum != KCP_SUCCESS) {
					return errnum;
				}

				/*	Determine the size of that auxPT						*/
				PTHdr = getPTHdr (auxPTRefNum);
				PTData = getPTData (auxPTRefNum);
				intSize = TpGetDataSize (PTHdr, PTData, format);
				PTCheckOut (auxPTRefNum);
				if (intSize == 0) {
					return KCP_INCON_PT;
				}
			}
			extSize += intSize;	/* add size of data */

		#if !defined KCMS_NO_CRC
			switch (format) {
			case PTTYPE_FUTF:
				if (crcMode != 0) {
					errnum = TpCalCrc (PTHdr, PTData, &crc32);
					if (errnum == KCP_SUCCESS) {
						KpItoa(crc32, strCRCmade);
						PTSetAttribute(PTRefNum, KCM_CRC, strCRCmade);
					}
				}
				break;

			default:
				break;
			}
		#endif
		}

/* add size of attributes. Must be done after CRC calculation */
#if !defined KCP_ICC_ONLY
		switch (format) {
		case PTTYPE_FUTF:
			PTAttr = getPTAttr (PTRefNum);
			extSize += getAttrSize(PTAttr);		/* plus size of attributes */

			break;

		default:
			break;
		}
#endif

		*mBlkSize = extSize;	/* return external size of PT */
		errnum = KCP_SUCCESS;
	}

	return (errnum);
}


/* PTGetPT writes a PT to external memory in FuT format.  */
PTErr_t
	PTGetPT (	PTRefNum_t	PTRefNum,
				KpInt32_t	mBlkSize,
				PTAddr_t	PTAddr)
{
PTErr_t		errnum;

	errnum = PTGetPTF (PTRefNum, PTTYPE_FUTF, mBlkSize, PTAddr);

	return (errnum);
}


/* PTGetPTF writes a PT to external memory in any format.  */
PTErr_t
	PTGetPTF (	PTRefNum_t	PTRefNum,
				PTType_t	format,
				KpInt32_t	mBlkSize,
				PTAddr_t	PTAddr)
{
PTErr_t		errnum, PTstatus;
KpHandle_t	PTHdr, PTAttr, PTData;
KpFd_t		fd;
KpInt32_t	attrSize, resultSize, nBytes;
PTRefNum_t	resizePTRefNum, thePTRefNum;
KpChar_p	memData;


	errnum = getPTStatus (PTRefNum);

	if ((errnum == KCP_PT_ACTIVE) || (errnum == KCP_PT_INACTIVE) || (errnum == KCP_SERIAL_PT)) {
		PTstatus = errnum;

		/* 	verify the dimensions of the grid table are valid for the
			specified format.  If they are not the right size attempt to
			create a grid table of the appropriate size						*/
		errnum = gridDimValid (format, PTRefNum,
							   &resizePTRefNum);
		if (errnum != KCP_SUCCESS) {
			goto ErrOut1;
		}

		if (resizePTRefNum != 0) {
			thePTRefNum = resizePTRefNum;

			/*	determine the size of the resized PT.  It may be smaller
				than than the buffer size, but it can't be larger			*/
			errnum = PTGetSizeF (thePTRefNum, format, &resultSize);
			if (errnum != KCP_SUCCESS) {
				goto ErrOut1;
			}
			if (resultSize > mBlkSize) {
				errnum = KCP_PT_BLOCK_TOO_SMALL;
				goto ErrOut1;
			}
		}
		else {
			thePTRefNum = PTRefNum;
			resultSize = mBlkSize;
		}

		PTAttr = getPTAttr (thePTRefNum);
		PTHdr = getPTHdr (thePTRefNum);
		PTData = getPTData (thePTRefNum);

		/* initialize memory file manager to write the PT */
		if (KpOpen (NULL, "m", &fd, NULL, (KpGenericPtr_t)PTAddr, mBlkSize) != KCMS_IO_SUCCESS) {
			errnum = KCP_SYSERR_1;
			goto ErrOut1;
		}

		attrSize = getAttrSize (PTAttr);					/* get size of attributes */
		errnum = TpWriteHdr (&fd, format, PTHdr, attrSize);	/* write the header info */

		if (errnum != KCP_SUCCESS) {
			Kp_close (&fd);
			goto ErrOut1;
		}

#if !defined KCP_ICC_ONLY
		switch (format) {
		case PTTYPE_FUTF:
			errnum = writeAttributes (&fd, PTAttr);	/* write the attributes */
			if (errnum != KCP_SUCCESS) {
				break;
			}

		default:
			break;
		}
#endif

		/* if PT active, write data to external format */
		if (((PTstatus == KCP_PT_ACTIVE) || (PTstatus == KCP_SERIAL_PT)) && (errnum == KCP_SUCCESS)) {
			errnum = TpWriteData (&fd, format, PTHdr, PTData);
		}
		(void) Kp_close (&fd);

		/*	if the result PT size is smaller than the memory block size
			fill the end of the memory block with zeros						*/
		nBytes = mBlkSize - resultSize;
		if (nBytes > 0) {
			memData = (KpChar_p)PTAddr + resultSize;
			while (nBytes--) {
				*memData++ = 0;
			}
		}

	}

ErrOut1:
	if (resizePTRefNum != 0) {
		PTCheckOut (resizePTRefNum);
	}

	return (errnum);
}


/* PTMemTest checks whether there is a reasonable amount of memory still available */
KpInt32_t PTMemTest(void)
{
KpHandle_t theBuffer;

#if defined (KPMAC)
	theBuffer = NewHandle (MINPTMEMSIZE);	/* do not want temp mem in this case */
#else
	theBuffer = allocBufferHandle (MINPTMEMSIZE);
#endif

	if (theBuffer == NULL) {
		return (0);
	}
	else {
		freeBuffer (theBuffer);
		return (1);
	}
}


/*
 * gridDimValid determines whether the grid table dimensions are valid
 * for the format specified.  If the dimensions are not valid and
 * resizePTRefNumP is not NULL then the function attempts to create a
 * PT with the correct size grid tables.  If it is successful the
 * PTRefNum of the resized PT is returned in the location pointed to by
 * resizePTRefNumP.  If resizing is not required the value returned in
 * the location pointed to by resizePTRefNumP is 0.
 *
 * NOTE:  	If this function creates a resized PT, that PT is checked in.
 *			it is the responsibility of the calling function to check out
 *			that PT.
 */
static PTErr_t
	gridDimValid(	PTType_t	format,
					PTRefNum_t	PTRefNum,
					PTRefNum_p	resizePTRefNumP)
{
KpHandle_t	PTData;
KpInt32_t	inputChans, outputChans, LUTDimensions;
PTRefNum_t	auxPTRefNum;
fut_ptr_t 	fut;
PTErr_t		retVal, error = KCP_SUCCESS;

	/*	Assume no resizing */
	if (NULL != resizePTRefNumP) {
		*resizePTRefNumP = 0;
	}

	/*	Convert the PTRefNum to a fut */
	PTData = getPTData (PTRefNum);
	fut = fut_lock_fut (PTData);
	if (fut == FUT_NULL) {
		return KCP_PTERR_2;
	}

	if ( ! IS_FUT (fut) ) {	/* check for valid fut */
		retVal = KCP_NOT_FUT;
		goto GetOut;
	}

	switch (format ) {

#if !defined KCP_ICC_ONLY
	case PTTYPE_FUTF:
		/* 	may want to check if any of the grid dimensions exceed the max
			grid dimension, but for now accept any size						*/
		break;
#endif

	case PTTYPE_MFT1:
	case PTTYPE_MFT2:
		/*	The grid dimensions must all be the same.  If they are not then
			attempt to build a grid table where all the dimensions are the
			same.															*/
		retVal = (PTErr_t) fut_mfutInfo (fut, &LUTDimensions, &inputChans,
									   &outputChans);
		if (1 != retVal) {
			if (-2 == retVal) {

				/*	if resizePTRefNumP is NULL then return success.			*/
				if (NULL == resizePTRefNumP) {
					break;
				}

				/*	Find the appropriate aux PT to precompose with the
					original PT to get the correct size grid table.			*/
				retVal = getResizeAuxPT (PTRefNum,
									 	 &auxPTRefNum);
				if (KCP_SUCCESS != retVal) {
					goto GetOut;
				}

			/*	Do the pre composition	*/
/* unlock this fut until we start counting locks ... */
				fut_unlock_fut (fut);
				retVal = PTCombine (PT_COMBINE_STD, auxPTRefNum, PTRefNum,
									resizePTRefNumP);
				if (KCP_SUCCESS != retVal) {
					error = retVal;
					*resizePTRefNumP = 0;
				}
/* ... then lock it again!!! */
				fut = fut_lock_fut (PTData);

				/*	Free the aux PT */
				retVal = PTCheckOut(auxPTRefNum);
				if (retVal != KCP_SUCCESS) {
					*resizePTRefNumP = 0;
					goto GetOut;
				}
				if (KCP_SUCCESS != error) {
					retVal = error;
					goto GetOut;
				}
 			} /* if (retVal == -2) */
			else {
				retVal = KCP_INVAL_GRID_DIM;
				goto GetOut;
			}
		} /* if (retVal != 1) */
		break;

	default:
		retVal = KCP_INVAL_PTTYPE;
	}

	retVal = KCP_SUCCESS;

GetOut:
	fut_unlock_fut (fut);
	return retVal;

} /* gridDimValid */


