/*
 * @(#)GlyphVector.cpp	1.19 98/09/13
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author Charlton Innovations, Inc.
 */

#include "GlyphVector.h"
#include "Strike.h"

// REMIND: speedup if we write a "hsDesc" which can live on the stack

static jclass   g_gvClass       = 0;
static jfieldID g_gvPositions   = 0;
static jfieldID g_gvGlyphs      = 0;

static jclass    g_shapeClass       = 0;
static jmethodID g_shapeMoveTo      = 0;
static jmethodID g_shapeLineTo      = 0;
static jmethodID g_shapeQuadTo      = 0;
static jmethodID g_shapeClosePath   = 0;

static Boolean initGVIDs(JNIEnv *env, jobject gv) {
    if (!g_gvClass) {
        g_gvClass = env->GetObjectClass(gv);
        if (!g_gvClass) {
            JNU_ThrowClassNotFoundException(env, "No GlyphVector class");
            return false;
        }
    }
    if (g_gvPositions && g_gvGlyphs) {
        return true;
    }
    g_gvPositions   = env->GetFieldID(g_gvClass, "positions", "[F");
    g_gvGlyphs      = env->GetFieldID(g_gvClass, "glyphs", "[I");
    if (!g_gvPositions || !g_gvGlyphs) {
        JNU_ThrowNoSuchFieldException(env,
                                    "GlyphVector positions or glyphs");
        return false;
    }
    return true;
}

static Boolean initShapeIDs(JNIEnv *env, jobject shape) {
    if (!g_shapeClass) {
        g_shapeClass = env->GetObjectClass(shape);
        if (!g_shapeClass) {
            JNU_ThrowClassNotFoundException(env, "No Shape class");
            return false;
        }
    }
    if (g_shapeMoveTo && g_shapeLineTo && g_shapeQuadTo && g_shapeClosePath) {
        return true;
    }
    g_shapeMoveTo       = env->GetMethodID(g_shapeClass, "moveTo", "(FF)V");
    g_shapeLineTo       = env->GetMethodID(g_shapeClass, "lineTo", "(FF)V");
    g_shapeQuadTo       = env->GetMethodID(g_shapeClass, "quadTo", "(FFFF)V");
    g_shapeClosePath    = env->GetMethodID(g_shapeClass, "closePath", "()V");
    if (!g_shapeMoveTo || !g_shapeLineTo
            || !g_shapeQuadTo || !g_shapeClosePath)
    {
        JNU_ThrowNoSuchMethodException(env,
                                    "Shape methods missing");
        return false;
    }
    return true;
}

void GlyphVector::allocateGlyphs(void)
{
    if (fNumGlyphs <= eDefaultStorage) {
        fGlyphs = fBaseGlyphs;
    } else {
        fGlyphs = new UInt32[fNumGlyphs];
    }
}

void GlyphVector::allocatePositions(void)
{
    if (fNumGlyphs <= eDefaultStorage) {
        fPositions = fBasePositions;
    } else {
        fPositions = new float[fNumGlyphs * 2];
    }
}

void GlyphVector::allocateImageRefs(void)
{
    if (fNumGlyphs <= eDefaultStorage) {
        fImageRefs = fBaseImageRefs;
    } else {
        fImageRefs = new ImageRef[fNumGlyphs];
    }
}

static inline int inRange(int c, int begin, int end) {
    return (c>=begin) && (c<=end);
}

void GlyphVector::initGlyphs(const jchar *theChars, FontTransform& tx)
{
    fontObject *myFO = (fontObject *)fFont;
    allocateGlyphs();
    Strike *theStrike;
    theStrike = &myFO->getStrike(tx, fIsAntiAliased, fUsesFractionalMetrics);

    if (fGlyphs) {
        for (int i=0; i<fNumGlyphs; i++) {
            jchar theChar = theChars[i];
            fGlyphs[i] = theStrike->CharToGlyph(theChar);
            if ( !(theChar & 0xfe00)) {
                // if roman assume no shaping, BIDI
                continue;
            }
            if (inRange(theChar, 0x0590, 0x05ff)) {
                // if Hebrew, assume shaping for presentation forms
                fNeedShaping = true;
                continue;
            }
            if (inRange(theChar, 0x0600, 0x06ff)) {
                // if Arabic, assume shaping for presentation forms
                fNeedShaping = true;
                continue;
            }
            if (inRange(theChar, 0x202a, 0x202e)) {
                // if directional control, assume shaping needed
                fNeedShaping = true;
                continue;
            }
            if (inRange(theChar, 0x206a, 0x206f)) {
                // if shaping control, assume shaping needed
                fNeedShaping = true;
                continue;
            }
        }
    }
}

GlyphVector::GlyphVector(JNIEnv *env, jstring theString,
                jdoubleArray matrix,
                jboolean isAntiAliased, jboolean usesFractionalMetrics,
                fontObject *fo)
    : fNeedShaping(false), fEnv(env), fFont(fo),
      fGlyphs(NULL), fPositions(NULL),
      fImageRefs(NULL), fMatrix(matrix),
      fIsAntiAliased(isAntiAliased),
      fUsesFractionalMetrics(usesFractionalMetrics),
      fNumGlyphs(0)
{
    fDevTX[0] = 1;
    fDevTX[1] = 0;
    fDevTX[2] = 0;
    fDevTX[3] = 1;

    if (JNU_IsNull(fEnv, theString) ) {
        JNU_ThrowIllegalArgumentException(fEnv,"String pointer is NULL!");
        return;
    }

    fNumGlyphs = env->GetStringLength(theString);
    FontTransform tx(fEnv, fMatrix);
    const jchar *theChars = env->GetStringCritical(theString, NULL);
    if (theChars) {
        initGlyphs(theChars, tx);
        env->ReleaseStringCritical(theString, theChars);
    }
}

GlyphVector::GlyphVector(JNIEnv *env,
                         jcharArray theCharArray, jint offset, jint count,
                         jdoubleArray matrix,
                         jboolean isAntiAliased,
                         jboolean usesFractionalMetrics,
                         fontObject *fo)
    : fNeedShaping(false), fEnv(env), fFont(fo),
      fGlyphs(NULL), fPositions(NULL),
      fImageRefs(NULL), fMatrix(matrix),
      fIsAntiAliased(isAntiAliased),
      fUsesFractionalMetrics(usesFractionalMetrics),
      fNumGlyphs(0)
{
    fDevTX[0] = 1;
    fDevTX[1] = 0;
    fDevTX[2] = 0;
    fDevTX[3] = 1;

    if (JNU_IsNull(fEnv, theCharArray) ) {
        JNU_ThrowIllegalArgumentException(fEnv,"Char array is NULL!");
        return;
    }

    if (offset + count > env->GetArrayLength(theCharArray)) {
        JNU_ThrowArrayIndexOutOfBoundsException(fEnv,
                                        "chars [offset + count]");
        return;
    }
    fNumGlyphs = count;
    FontTransform tx(fEnv, fMatrix);
    const jchar *theChars =
        (const jchar *)env->GetPrimitiveArrayCritical(theCharArray, NULL);
    if (theChars) {
        initGlyphs( theChars + offset, tx);
        env->ReleasePrimitiveArrayCritical(
                    theCharArray, (void *)theChars, JNI_ABORT);
    }
}

GlyphVector::GlyphVector(JNIEnv *env,
                         jbyteArray theByteArray, jint offset, jint count,
                         jdoubleArray matrix,
                         jboolean isAntiAliased,
                         jboolean usesFractionalMetrics,
                         fontObject *fo)
    : fNeedShaping(false), fEnv(env), fFont(fo),
      fGlyphs(NULL), fPositions(NULL),
      fImageRefs(NULL), fMatrix(matrix),
      fIsAntiAliased(isAntiAliased),
      fUsesFractionalMetrics(usesFractionalMetrics),
      fNumGlyphs(0)
{
    fDevTX[0] = 1;
    fDevTX[1] = 0;
    fDevTX[2] = 0;
    fDevTX[3] = 1;

    if (JNU_IsNull(fEnv, theByteArray) ) {
        JNU_ThrowIllegalArgumentException(fEnv,"Byte array is NULL!");
        return;
    }

    if (offset + count > env->GetArrayLength(theByteArray)) {
        JNU_ThrowArrayIndexOutOfBoundsException(fEnv,
                                        "chars [offset + count]");
        return;
    }
    fNumGlyphs = count;
    Strike *theStrike;
    {
        FontTransform tx(fEnv, matrix);
        theStrike = &fo->getStrike(tx, isAntiAliased, usesFractionalMetrics);
    }
    const jbyte *theBytes =
        (const jbyte *)env->GetPrimitiveArrayCritical(theByteArray, NULL);
    if (theBytes) {
        allocateGlyphs();
        if (fGlyphs) {
            for (int i=0; i<count; i++) {
                fGlyphs[i] = theStrike->CharToGlyph(theBytes[i+offset]);
            }
        }
        env->ReleasePrimitiveArrayCritical(
                    theByteArray, (void *)theBytes, JNI_ABORT);
    }
}

GlyphVector::GlyphVector(JNIEnv *env, jintArray theGlyphArray,
                         jdoubleArray matrix,
                         jboolean isAntiAliased,
                         jboolean usesFractionalMetrics,
                         fontObject *fo)
    : fNeedShaping(false), fEnv(env), fFont(fo),
      fGlyphs(NULL), fPositions(NULL),
      fImageRefs(NULL), fMatrix(matrix),
      fIsAntiAliased(isAntiAliased),
      fUsesFractionalMetrics(usesFractionalMetrics),
      fNumGlyphs(0)
{
    fDevTX[0] = 1;
    fDevTX[1] = 0;
    fDevTX[2] = 0;
    fDevTX[3] = 1;

    if (JNU_IsNull(fEnv, theGlyphArray)) {
        JNU_ThrowIllegalArgumentException(fEnv,"Glyph array is NULL!");
        return;
    }

    fNumGlyphs = env->GetArrayLength(theGlyphArray); 
    Strike *theStrike;
    {
        FontTransform tx(fEnv, matrix);
        theStrike = &fo->getStrike(tx, isAntiAliased, usesFractionalMetrics);
    }
    const jint *theGlyphs =
        (const jint *)env->GetPrimitiveArrayCritical(theGlyphArray, NULL);
    if (theGlyphs) {
        allocateGlyphs();
        int missingGlyph = theStrike->getMissingGlyphCode();
        if (fGlyphs) {
            for (int i=0; i<fNumGlyphs; i++) {
                fGlyphs[i] = theGlyphs[i];
            }
        }
        env->ReleasePrimitiveArrayCritical(
                    theGlyphArray, (void *)theGlyphs, JNI_ABORT);
    }
}

GlyphVector::GlyphVector(JNIEnv *env,
                         jdoubleArray matrix,
                         jboolean isAntiAliased,
                         jboolean usesFractionalMetrics,
                         fontObject *fo)
    : fNeedShaping(false), fEnv(env), fFont(fo),
      fGlyphs(NULL), fPositions(NULL),
      fImageRefs(NULL), fMatrix(matrix),
      fIsAntiAliased(isAntiAliased),
      fUsesFractionalMetrics(usesFractionalMetrics),
      fNumGlyphs(0)
{
    fDevTX[0] = 1;
    fDevTX[1] = 0;
    fDevTX[2] = 0;
    fDevTX[3] = 1;
}

GlyphVector::~GlyphVector() {
    if (fGlyphs && (fGlyphs != fBaseGlyphs)) {
        delete [] fGlyphs;
    }
    fGlyphs = NULL;
    if (fPositions && (fPositions != fBasePositions)) {
        delete [] fPositions;
    }
    fPositions = NULL;
}

Boolean GlyphVector::needShaping()
{
    return fNeedShaping;
}

void GlyphVector::setDevTX(jdoubleArray devTX) {
    FontTransform tx(fEnv, devTX);
    tx.getMatrixInto(fDevTX, 4);
}

void GlyphVector::setPositions(jfloatArray newPos)
{
//	return; // REMIND: workaround for 4143573
    if (!JNU_IsNull(fEnv, newPos)) {
        JFloatBuffer positionBuffer(fEnv, newPos);
        UInt32 numPositions = positionBuffer.getNumElements();
        const jfloat *thePositions = positionBuffer.buffer();
        if (thePositions) {
            if (numPositions == (fNumGlyphs *2)) {
                allocatePositions();
            } else {
                JNU_ThrowIllegalArgumentException(fEnv,
                                                "Wrong size positions array");
            }
            if (fPositions) {
                for (int i=0; i<numPositions; i++) {
                    fPositions[i] = thePositions[i];
                }
            }
        }
    }
}

void GlyphVector::positionGlyphsWithImages(
        jfloat x, jfloat y,
        jdoubleArray matrix,
        Boolean isAntiAliased, Boolean usesFractionalMetrics)
{
    Boolean loadPositions = false;
    FontTransform tx(fEnv, matrix);
    Strike& theStrike = fFont->getStrike(tx, isAntiAliased, usesFractionalMetrics);
    // load positions if they are not already specified
    if (!fPositions) {
        allocatePositions();
        loadPositions = true;
    }
    if (!fImageRefs) {
        allocateImageRefs();
    }
    if (!fPositions || !fImageRefs) {
        JNU_ThrowInternalError(fEnv, "Insufficient Memory");
        return;
    }
    for (int i=0; i < fNumGlyphs; i++) {
        hsGGlyph glyphRef;
        hsFixedPoint2 advanceXY;
        theStrike.getMetrics(fGlyphs[i], glyphRef, advanceXY);
        fImageRefs[i].rowBytes  = glyphRef.fRowBytes;
        fImageRefs[i].width     = glyphRef.fWidth;
        fImageRefs[i].height    = glyphRef.fHeight;
        fImageRefs[i].pixels    = glyphRef.fImage;
        if (!glyphRef.fImage && glyphRef.fWidth && glyphRef.fHeight) {
            fImageRefs[i].pixels = theStrike.getImage(fGlyphs[i]);
        }
        if (loadPositions) {
            fPositions[i*2] = x + hsFixedToFloat(glyphRef.fTopLeft.fX);
            fPositions[i*2+1] = y + hsFixedToFloat(glyphRef.fTopLeft.fY);
            x += hsFixedToFloat(advanceXY.fX);
            y += hsFixedToFloat(advanceXY.fY);
        } else {
            // this modifies the internal state in a fashion which
            // should not be returned to JAVA land...
            float tempX = fPositions[i*2];
            float tempY = fPositions[i*2 + 1];
            fPositions[i*2]   = tempX * fDevTX[0] + tempY * -fDevTX[1]
                                    + x + hsFixedToFloat(glyphRef.fTopLeft.fX);
            fPositions[i*2+1] = tempX * -fDevTX[2] + tempY * fDevTX[3]
                                    + y + hsFixedToFloat(glyphRef.fTopLeft.fY);
        }
    }
}

void GlyphVector::positionGlyphs(
        jfloat x, jfloat y,
        jdoubleArray matrix,
        Boolean isAntiAliased, Boolean usesFractionalMetrics)
{
    FontTransform tx(fEnv, matrix);
    Strike& theStrike = fFont->getStrike(tx, isAntiAliased, usesFractionalMetrics);
    // load positions if they are not already specified
    if (!fPositions) {
        allocatePositions();
    }
    if (!fPositions) {
        JNU_ThrowInternalError(fEnv, "Insufficient Memory");
        return;
    }
    for (int i=0; i < fNumGlyphs; i++) {
        hsGGlyph glyphRef;
        hsFixedPoint2 advanceXY;
        theStrike.getMetrics(fGlyphs[i], glyphRef, advanceXY);
        fPositions[i*2] = x;
        fPositions[i*2+1] = y;
        x += hsFixedToFloat(advanceXY.fX);
        y += hsFixedToFloat(advanceXY.fY);
    }
}

void GlyphVector::getGlyphCodes(jobject gvTarget)
{
    if (initGVIDs(fEnv, gvTarget)) {
        jintArray glyphs =
            (jintArray)fEnv->GetObjectField(gvTarget, g_gvGlyphs);
        UInt32Buffer gb(fEnv, glyphs);
        fNumGlyphs = gb.getNumElements();
        allocateGlyphs();
        if (fGlyphs) {
            UInt32 *glyphCodes = gb.buffer();
            if (glyphCodes) {
                for (int i=0; i < fNumGlyphs; i++) {
                    fGlyphs[i] = glyphCodes[i];
                }
            }
        } else {
	  fNumGlyphs = 0;
	}
    }
}

void GlyphVector::getPositions(jobject gvTarget, float x, float y,
                        jdoubleArray matrix,
                        Boolean isAntiAliased, Boolean usesFractionalMetrics)
{
    if (initGVIDs(fEnv, gvTarget)) {
        jfloatArray positions =
            (jfloatArray)fEnv->GetObjectField(gvTarget, g_gvPositions);
        setPositions(positions);
        if (!fPositions) {
            // position in absolute coordinates
            positionGlyphs(x,y, matrix, isAntiAliased, usesFractionalMetrics);
        } else {
            // caller's coordinates are relative to x, y.
            for (int i=0; i < fNumGlyphs; i++) {
                fPositions[i*2] += x;
                fPositions[i*2+1] += y;
            }
        }
    }
}

void GlyphVector::writeGlyphCodes(jobject gvTarget)
{
    if (initGVIDs(fEnv, gvTarget)) {
        jintArray glyphsArray = fEnv->NewIntArray(fNumGlyphs);
        if (glyphsArray) {
            UInt32Buffer glyphBuffer(fEnv, glyphsArray);
            UInt32 *glyphs = glyphBuffer.buffer();
            if (glyphs) {
                if (fGlyphs) {
                    for (int i=0; i < fNumGlyphs; i++) {
                        glyphs[i] = fGlyphs[i];
                    }
                } else {
                    for (int i=0; i < fNumGlyphs; i++) {
                        glyphs[i] = 0;
                    }
                }
            }
        }
        if (glyphsArray) {
            fEnv->SetObjectField(gvTarget, g_gvGlyphs, glyphsArray);
        }
    }
}

void GlyphVector::writePositions(jobject gvTarget)
{
    if (initGVIDs(fEnv, gvTarget)) {
        jfloatArray positions = fEnv->NewFloatArray(fNumGlyphs * 2);
        if (positions) {
            JFloatBuffer floatBuffer(fEnv, positions);
            jfloat *floats = floatBuffer.buffer();
            if (floats) {
                if (fPositions) {
                    for (int i=0; i < fNumGlyphs*2; i++) {
                        floats[i] = fPositions[i];
                    }
                } else {
                    for (int i=0; i < fNumGlyphs*2; i++) {
                        floats[i] = 0;
                    }
                }
            }
        }
        if (positions) {
            fEnv->SetObjectField(gvTarget, g_gvPositions, positions);
        }
    }
}

void
GlyphVector::scanGlyph(jobject shapeTarget,
    Strike& theStrike,
    hsPathSpline& path,
    CurveWalker& walker,
    float xPos, float yPos,
    jint glyphIndex)
{
    path.Init(&path);
    theStrike.GetGlyphOutline(fGlyphs[glyphIndex], path);

    for (int j = 0; j < path.fContourCount; j++) {
        Boolean firstTime = true;
        walker.Init(&path.fContours[j], true);
        while (walker.NextQuad()) {
            if (firstTime) {
                fEnv->CallVoidMethod(shapeTarget, g_shapeMoveTo,
                                    hsScalarToFloat(walker.a.fX) + xPos,
                                    hsScalarToFloat(walker.a.fY) + yPos);
                firstTime = false;
            }
            if (walker.drawLine) {
                fEnv->CallVoidMethod(shapeTarget, g_shapeLineTo,
                                    hsScalarToFloat(walker.b.fX) + xPos,
                                    hsScalarToFloat(walker.b.fY) + yPos);
            } else {
                fEnv->CallVoidMethod(shapeTarget, g_shapeQuadTo,
                                    hsScalarToFloat(walker.b.fX) + xPos,
                                    hsScalarToFloat(walker.b.fY) + yPos,
                                    hsScalarToFloat(walker.c.fX) + xPos,
                                    hsScalarToFloat(walker.c.fY) + yPos);
            }
        }
        fEnv->CallVoidMethod(shapeTarget, g_shapeClosePath);
    }
    path.Delete(&path);
}

void GlyphVector::scanShapeInto(jobject shapeTarget, jdoubleArray matrix,
                    jboolean isAntiAliased, jboolean usesFractionalMetrics)
{
    if (!fGlyphs || !fPositions || !initShapeIDs(fEnv, shapeTarget)) {
        return;
    }
    FontTransform tx(fEnv, matrix);
    Strike& theStrike = fFont->getStrike(tx, isAntiAliased, usesFractionalMetrics);

    hsPathSpline path;
    CurveWalker walker;
    for (int i=0; i < fNumGlyphs; i++) {
        float xPos = fPositions[i*2];
        float yPos = fPositions[i*2+1];

        scanGlyph(shapeTarget, theStrike, path, walker, xPos, yPos, i);
    }
    return;
}

void GlyphVector::scanShapeInto(jobject shapeTarget, jint glyphIndex,
                    jdoubleArray matrix,
                    jboolean isAntiAliased, jboolean usesFractionalMetrics)
{
    if (!fGlyphs || !fPositions || !initShapeIDs(fEnv, shapeTarget)) {
        return;
    }
    FontTransform tx(fEnv, matrix);
    Strike& theStrike = fFont->getStrike(tx, isAntiAliased, usesFractionalMetrics);
    hsPathSpline path;
    CurveWalker walker;
    scanGlyph(shapeTarget, theStrike, path, walker, 0, 0, glyphIndex);
    return;
}

// REMIND: lame.  the C++ compiler/linker complain if there
// is no implementation
void GlyphVector::operator delete(void *ptr)
{
}
