/*
 * @(#)hsGStyle.h	1.4 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Copyright (C) 1996-1997 all rights reserved by HeadSpin Technology Inc. Chapel Hill, NC USA
 *
 * This software is the property of HeadSpin Technology Inc. and it is furnished
 * under a license and may be used and copied only in accordance with the
 * terms of such license and with the inclusion of the above copyright notice.
 * This software or any other copies thereof may not be provided or otherwise
 * made available to any other person or entity except as allowed under license.
 * No title to and ownership of the software or intellectual property
 * therewithin is hereby transferred.
 *
 * HEADSPIN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
 * OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * This information in this software is subject to change without notice
*/

#ifndef hsGStyleDefined
#define hsGStyleDefined

#include "HeadSpin.h"
#include "hsGFont.h"

#if HS_BUILD_FOR_WIN32
	#undef GetTextFace
#endif

struct hsMatrix33;
class hsGTextContext;

#define kNoBold_hsGTextFace				hsScalar(1)
#define kNoSkew_hsGTextFace				hsScalar(0)
#define kNoXScale_hsGTextFace			hsScalar(1)
#define kNoOutlineWidth_hsGTextFace		hsScalar(0)
#define kNoTracking_hsGTextFace			hsScalar(0)
#define kNoUnderlineThickness_hsGTextFace	hsScalar(0)
#define kNoUnderlineOffset_hsGTextFace		hsScalar(0)

struct hsGTextFace {
	hsScalar	fBoldness;
	hsScalar	fSkew;
	hsScalar	fXScale;
	hsScalar	fOutlineWidth;
	hsScalar	fTracking;
	hsScalar	fUnderlineThickness;
	hsScalar	fUnderlineOffset;

	void		Reset();
	void		Read(hsStream* stream);
	void		Write(hsStream* stream);

	friend int	operator==(const hsGTextFace& a, const hsGTextFace& b);
	friend int	operator!=(const hsGTextFace& a, const hsGTextFace& b)
			{
				return !(a == b);
			}
};

class hsGStyle : public hsRefCnt {
public:
	enum {
		kFrame		= 0x01,		// else FILL
		kWindingFill	= 0x02,		// else XOR fill
		kClosedFrame 	= 0x04,		// else open frame
		kKernText		= 0x08		// else no kerning
	};
	enum CapType {
		kButtCap,
		kRoundCap,
		kSquareCap
	};
	enum JoinType {
		kMiterJoin,
		kRoundJoin,
		kBluntJoin
	};
protected:
	UInt32		fFlags;

	hsScalar 		fWidth;
	hsScalar 		fMiterLimit;
	CapType	 	fCapType;
	JoinType	 	fJoinType;

	hsGFontID		fFontID;
	hsScalar		fTextSize;
	hsGTextFace*	fFace;

	hsGStyle* 	CopyInto(hsGStyle* dst) const;
public:
				hsGStyle();
	virtual		~hsGStyle();

	hsScalar		GetWidth() const { return fWidth; }
	hsScalar		GetMiterLimit() const { return fMiterLimit; }
	UInt32		GetFlags() const { return fFlags; }
	CapType		GetCapType() const { return fCapType; }
	JoinType		GetJoinType() const { return fJoinType; }
	hsGFontID		GetFontID() const { return fFontID; }
	hsScalar		GetTextSize() const { return fTextSize; }
	//	This guy returns nil if there is no textface
	hsGTextFace*	GetTextFace(hsGTextFace* face) const;

	virtual void	SetWidth(hsScalar width);
	virtual void	SetMiterLimit(hsScalar miterLimit);
	virtual void	SetFlags(UInt32 flags);
	virtual void	SetCapType(CapType capType);
	virtual void	SetJoinType(JoinType joinType);
	virtual void	SetFontID(hsGFontID fontID);
	virtual void	SetTextSize(hsScalar textSize);
	virtual void	SetTextFace(const hsGTextFace* face);

	UInt32		GetScanStrokerFlags() const;

	virtual hsGStyle*	Copy() const;

	virtual void	Read(hsStream* stream);
	virtual void	Write(hsStream* stream);

	virtual hsScalar		MeasureText(UInt32 length, const void* text, class hsGAttribute* attr,
									hsPoint2* ascent, hsPoint2* descent);
	//	This must be overridden by specific text systems
	virtual hsGTextContext*	GetTextContext(const hsMatrix33* matrix, class hsGAttribute* attr);
};

/*@
@page hsgstyle.html

@struct hsGTextFace
This struct holds the algorithmic parameters for text styling. Use GetTextFace() and SetTextFace() on a hsGStyle.
Fields marked with *** automatically scale with the textsize in the style, so the value for a 1-point font should
be stored.
@field	fBoldness				multiplier for the amount of emboldening. 1.0 is default
@field	fSkew				amount to skew text in X. 0 is default
@field	fXScale				multiplier for scaling text (image and metrics) in X. 1.0 is default
@field	fOutlineWidth			width of the glyphs outline. 0 is default (filled rather than stroked) ***
@field	fTracking				added between pairs of glyphs. 0 is default ***
@field	fUnderlineThickness		thickness of underline. 0 is default (no underline) ***
@field	fUnderlineOffset		offset below the baseline for the top of the underline ***
@endstruct

@class hsGStyle
The style object defines how the object is to be rendered.  An object can be either framed or filled and this is controlled by
style flags.

@flags StyleOptions
These flags control framing and filling options
@token 	kFrame  draw the shape as a framed object instead of filling it.
@token 	kWindingFill use winding rather than XOR fill rules when the shape is not framed
@token 	kClosedFrame if set this flag causes the frame to close a shape if necessary when framing.
@token	kKernText		if set the kerning values in the font are used when measuring and drawing text
@endflags

@enum CapType
This enumeration controls the kind of cap that will be placed at the end of any thick lines. <IMG SRC="caps.jpg" ALT="Cap Type Diagram">
@token  kButtCap creates a squared off end right at the end point
@token kRoundCap rounds the end using a half circle of radius LineWidth/2 centered on the end point
@token kSquareCap squares off the end LineWidth/2 units from the endpoint
@endenum

@enum JoinType
This enumeration controls the kind of join that will be created between line segments. <IMG SRC="joins.jpg" ALT="Join Type Diagram">
@token kMiterJoin creates a mitered "pointed" edge at the joint.
@token kRoundJoin creates a curved join
@token kBluntJoin creates a flattened join
@endenum

@method GetWidth Return the width of the stroke outline
@method SetWidth Set the width of the stroke outline
@method GetMiterLimit Return the miter limit amount for mitered corners
@method SetMiterLimit Set the miter limit amount for mitered corners
@method GetFlags Get the style flags
@method SetFlags Set the style flags
@method SetCapType Set the cap style for the stroke outline
@method GetCapType Get the cap style for the stroke outline
@method GetJoinType Set the join style for the stroke outline
@method SetJoinType Get the join style for the stroke outline
@method GetFontID		Return the current fontID. 0 means default font
@method SetFontID		Set the fontID. 0 means default font
@method GetTextSize	Return the current text size.
@method SetTextSize	Set the text size.
@method GetTextFace	Return the current textFace, or nil if there is no current face
@method SetTextFace	Set the textFace, or pass nil to remove it for plain, unstylized text
@method Copy Create a copy of this style object
@endclass

@methoddef hsGStyle SetCapType
This method sets the current cap type for the shape
@param hsGStyle::CapType capType the new cap type
@endmethod

@methoddef hsGStyle SetJoinType
This method sets the current join type for the shape
@param hsGStyle::JoinType joinType the new join type
@endmethod

@methoddef hsGStyle SetMiterLimit
This method sets the current miter limit for the shape. When a miter join is created at a very acute angle the
calculated corner point can get arbitrarily far from the join point. To avoid this the miter is blunted off when the distance
is greater than the maximum miter distance .  The value given to set miter limit is multiplied by LineWidth/2 to get the
maximum miter distance. This parameter only has an effect when hsGStyle::kFrame flag is set and the Join type
is ^hsGStyle::kMiterJoin .
@param hsScalar miterLimit the number to multiply with LineWidth/2 to get the new maximum miter distance.
@endmethod

@methoddef hsGStyle SetWidth
This method sets the linewidth for framed rendering. The linewidth is used only when hsGStyle::kFrame flag is set.
@param hsScalar lineWidth  the width of the line to be drawn
@endmethod

@methoddef hsGStyle SetFontID
This method sets the fontID for the style. Valid fontIDs are available by calling hsGFontList::Get and passing an index,
or from hsGFontList::Find and passing a name.
@param hsGFontID	fontID	The new fontID
@endmethod

@endpage
*/
#endif
