/*
 * @(#)hsVector3.h	1.3 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Copyright (C) 1996-1997 all rights reserved by HeadSpin Technology Inc. Chapel Hill, NC USA
 *
 * This software is the property of HeadSpin Technology Inc. and it is furnished
 * under a license and may be used and copied only in accordance with the
 * terms of such license and with the inclusion of the above copyright notice.
 * This software or any other copies thereof may not be provided or otherwise
 * made available to any other person or entity except as allowed under license.
 * No title to and ownership of the software or intellectual property
 * therewithin is hereby transferred.
 *
 * HEADSPIN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
 * OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * This information in this software is subject to change without notice
*/

#ifndef hsVector3Defined
#define hsVector3Defined

#include "hsMatrix33.h"
#include "hsFixedTypes.h"

/////////////////////////////////////////////////////////////////////////////////////////////

/*
	If value is already close to hsScalar1, then this is a good approx. of 1/sqrt(value)
*/
static inline hsScalar hsInvSqrt(hsScalar value)
{
	hsScalar	guess;
	hsScalar	threeOverTwo = hsScalar1 + hsScalarHalf;

	value = hsScalarDiv2(value);
	guess = threeOverTwo - value;		// with initial guess = 1.0

	// repeat this line for better approx
	guess = hsScalarMul(guess, threeOverTwo - hsScalarMul(hsScalarMul(value, guess), guess));
	guess = hsScalarMul(guess, threeOverTwo - hsScalarMul(hsScalarMul(value, guess), guess));

	return guess;
}

/////////////////////////////////////////////////////////////////////////////////////////////
struct hsPoint3 {	
	hsScalar	fCoord[3];

	// Access methods
	hsScalar &x(void) { return fCoord[0]; }
	hsScalar &y(void) { return fCoord[1]; }
	hsScalar &z(void) { return fCoord[2]; }

	hsPoint3*	Set(hsScalar x, hsScalar y, hsScalar z) { fCoord[0] = x; fCoord[1] = y; fCoord[2] = z; return this; }
	
	friend inline hsPoint3 operator+(const hsPoint3& s, const hsPoint3& t);
	friend inline hsPoint3 operator+(const hsPoint3& s, const hsVector3& t);
	friend inline hsPoint3 operator-(const hsPoint3& s, const hsPoint3& t);
	friend inline hsPoint3 operator-(const hsPoint3& s);
	friend inline hsPoint3 operator*(const hsScalar& s, const hsPoint3& t);
	friend inline hsPoint3 operator*(const hsPoint3& t, const hsScalar& s);
	friend inline hsPoint3 operator/(const hsPoint3& t, const hsScalar& s);

	void Read(hsStream *stream);
	void Write(hsStream *stream);
};

/////////////////////////////////////////////////////////////////////////////////////////////

struct hsVector3 {
	hsScalar		fCoord[3];

	// Access methods
	hsScalar &x(void) { return fCoord[0]; }
	hsScalar &y(void) { return fCoord[1]; }
	hsScalar &z(void) { return fCoord[2]; }

	hsVector3*	Set(hsScalar x, hsScalar y, hsScalar z) { fCoord[0] = x; fCoord[1] = y; fCoord[2] = z; return this; }
	hsVector3*	Set(const hsPoint3*p) { fCoord[0] = p->fCoord[0]; fCoord[1] = p->fCoord[1]; fCoord[2] = p->fCoord[2]; return this;}
	hsVector3*	Add(const hsVector3* vec, hsVector3* dst = nil);
	hsVector3*	Sub(const hsVector3* vec, hsVector3* dst = nil);
	hsVector3*	Scale(hsScalar scale, hsVector3* dst = nil);

#if HS_SCALAR_IS_FIXED
	 hsScalar		Magnitude() const;
#else
	 hsScalar		Magnitude() const
				{		return hsSquareRoot(MagnitudeSquared());	}
#endif

#if HS_SCALAR_IS_FIXED
	 hsScalar		MagnitudeSquared() const;
#else
	 hsScalar		MagnitudeSquared() const
				{
					return (fCoord[0] * fCoord[0] + fCoord[1] * fCoord[1] + fCoord[2] * fCoord[2]);
				}
#endif

	void			Normalize()
				{
					hsScalar	length = this->Magnitude();
					hsScalar	invMag = hsScalarInvert(length);
					
					fCoord[0] = hsScalarMul(fCoord[0], invMag);
					fCoord[1] = hsScalarMul(fCoord[1], invMag);
					fCoord[2] = hsScalarMul(fCoord[2], invMag);
				}
	inline void		Renormalize()		// if the vector is already close to unit length
				{
					hsScalar	mag2 = *this * *this;
					hsScalar	invMag = hsInvSqrt(mag2);
					
					fCoord[0] = hsScalarMul(fCoord[0], invMag);
					fCoord[1] = hsScalarMul(fCoord[1], invMag);
					fCoord[2] = hsScalarMul(fCoord[2], invMag);
				}

	hsVector3 &Sub(const hsPoint3& s, const hsPoint3& t)
		{	Set(s.fCoord[0] - t.fCoord[0], s.fCoord[1] - t.fCoord[1], s.fCoord[2] - t.fCoord[2]); 
			return *this; };
	friend inline hsVector3 operator+(const hsVector3& s, const hsVector3& t);
	friend inline hsVector3 operator-(const hsVector3& s, const hsVector3& t);
	friend inline hsVector3 operator-(const hsVector3& s);
	friend inline hsVector3 operator*(const hsScalar& s, const hsVector3& t);
	friend inline hsVector3 operator*(const hsVector3& t, const hsScalar& s);
	friend inline hsVector3 operator/(const hsVector3& t, const hsScalar& s);
	friend inline hsScalar operator*(const hsVector3& t, const hsVector3& s);
	friend  hsVector3 operator%(const hsVector3& t, const hsVector3& s);

	void Read(hsStream *stream);
	void Write(hsStream *stream);
};



class hsPlaneSolver
{
	hsScalar	fCoeffX0;
	hsScalar	fCoeffY0;
	hsScalar	fCoeffX1;
	hsScalar	fCoeffY1;
	hsScalar	fCoeffX2;
	hsScalar	fCoeffY2;
public:
	int		ComputeCoeffs(const struct hsPoint3 verts[], hsScalar scale);	// scale = (1.0/65536.0);
	void	GetPlane(hsScalar x0, hsScalar x1, hsScalar x2, hsFixedPlane *out);
	void	GetPlaneScaled(hsScalar x0, hsScalar x1, hsScalar x2, hsFixedPlane *out);
};

struct hsFixedPoint2;

struct hsMatrix23 {
	hsScalar m[2][3];
	
	hsMatrix23() {}
	hsMatrix23(hsScalar m00, hsScalar m01, hsScalar m02, hsScalar m10, hsScalar m11, hsScalar m12);

	hsMatrix23(hsScalar ang, const hsPoint2& pivot);
	hsMatrix23(hsScalar cosAngle, hsScalar sinAngle, const hsPoint2& pivot);

	hsMatrix23&		Reset(void);
	hsMatrix23&		Scale(hsScalar scaleX, hsScalar scaleY, hsScalar aboutX, hsScalar aboutY);
	hsMatrix23&		Translate(hsScalar x, hsScalar y);
	hsMatrix23&		Rotate(hsScalar degrees, hsScalar aboutX, hsScalar aboutY);
	hsMatrix23&		Rotate(hsScalar cosAngle, hsScalar sinAngle, hsScalar aboutX, hsScalar aboutY);
	hsScalar			Determinant() const;
	hsMatrix23		Inverse() const;
	friend inline hsPoint2	operator*(const hsMatrix23& m, const hsPoint2& p);
	friend inline hsVector3 operator*(const hsMatrix23& m, const hsVector3& p);
	friend hsMatrix23	operator*(const hsMatrix23& a, const hsMatrix23& b);
	friend hsMatrix23	operator*(hsScalar a, const hsMatrix23& b);
	friend hsMatrix23	operator*(const hsMatrix23& a, hsScalar b);

	hsPoint2*			MapPoints(UInt32 count, hsPoint2 points[]) const;
	hsFixedPoint2*		MapPoints(UInt32 count, hsFixedPoint2 points[]) const;
	
	Boolean			IsIdentity(void) const;

	friend int operator==(const hsMatrix23& s, const hsMatrix23& t);
	friend int operator!=(const hsMatrix23& s, const hsMatrix23& t);
};


//////////////////////////////////////////////////////////////////////////

const int HS_TRANSFORM_SCALE		=	0x1;
const int HS_TRANSFORM_ROTATE		=	0x2;
const int HS_TRANSFORM_TRANSLATE 	= 	0x4;
const int HS_TRANSFORM_GENERAL		=	0xffff;

class hsTransform2d : public hsStreamable {
protected:
	hsMatrix23	fXForm;		// this is saved by Write
	UInt32		fTMask;		// this is saved by Write

	hsMatrix23	fInverse;
	hsMatrix23 	fNormalMat;
	Boolean		fDirty;
	void			Update();
public:
					hsTransform2d();
	
	void				Reset();
	hsTransform2d*	Scale(hsScalar scale, const hsPoint2 &p);
	hsTransform2d*	Translate(hsScalar dx, hsScalar dy);
	hsTransform2d*	Rotate(hsScalar a, const hsPoint2& p);
	hsTransform2d*	SetMatrix(const hsMatrix23* matrix);

	UInt32			GetTMask() const { return fTMask; }
	const hsMatrix23*	GetMatrix();
	const hsMatrix23*	GetInverseMatrix();
	const hsMatrix23*	GetNormalMatrix();
	void				Concat(hsTransform2d*);
	void				InverseConcat(hsTransform2d *);
	hsPoint2*			Transform(const hsPoint2 * in, hsPoint2 *out);
	hsPoint2*			InverseTransform(const hsPoint2 * in, hsPoint2 *out);
	hsVector3*		Transform(const hsVector3 *in, hsVector3 *out);
	
	virtual void		Read(hsStream* stream);
	virtual void		Write(hsStream* stream);
	virtual UInt32		GetStreamSize();
};

////////////////////////////////////////////////////////////////////////////

inline hsVector3 operator+(const hsVector3& s, const hsVector3& t)
{
	hsVector3		result;
	
	return *result.Set(s.fCoord[0] + t.fCoord[0], s.fCoord[1] + t.fCoord[1], s.fCoord[2] + t.fCoord[2]);
}

inline hsVector3 operator-(const hsVector3& s, const hsVector3& t)
{
	hsVector3		result;
	
	return *result.Set(s.fCoord[0] - t.fCoord[0], s.fCoord[1] - t.fCoord[1], s.fCoord[2] - t.fCoord[2]);
}

// unary minus
inline hsVector3 operator-(const hsVector3& s)
{
	hsVector3		result;	
	return *result.Set(-s.fCoord[0], -s.fCoord[1], -s.fCoord[2]);
}

inline hsVector3 operator*(const hsVector3& s, const hsScalar& t)
{
	hsVector3		result;	
	return *result.Set(hsScalarMul(s.fCoord[0], t), hsScalarMul(s.fCoord[1], t), hsScalarMul(s.fCoord[2], t));
}

inline hsVector3 operator/(const hsVector3& s, const hsScalar& t)
{
	hsVector3		result;	
	return *result.Set(hsScalarDiv(s.fCoord[0], t), hsScalarDiv(s.fCoord[1], t), hsScalarDiv(s.fCoord[2], t));
}

inline hsVector3 operator*(const hsScalar& t, const hsVector3& s)
{
	hsVector3		result;
	
	return *result.Set(hsScalarMul(s.fCoord[0], t), hsScalarMul(s.fCoord[1], t), hsScalarMul(s.fCoord[2], t));
}

inline hsScalar operator*(const hsVector3& t, const hsVector3& s)
{
	return hsScalarMul(t.fCoord[0], s.fCoord[0]) + hsScalarMul(t.fCoord[1], s.fCoord[1]) + hsScalarMul(t.fCoord[2], s.fCoord[2]);
}

////////////////////////////////////////////////////////////////////////////

inline hsPoint2 operator*(const hsMatrix23& m, const hsPoint2& p)
{
	hsPoint2	result;

	return *result.Set(	hsScalarMul(p.fX, m.m[0][0]) + hsScalarMul(p.fY, m.m[0][1]) + m.m[0][2],
					hsScalarMul(p.fX, m.m[1][0]) + hsScalarMul(p.fY, m.m[1][1]) + m.m[1][2]);
		
}

inline hsVector3 operator*(const hsMatrix23& m, const hsVector3& p)
{
	hsVector3		result;

	return *result.Set(	hsScalarMul(p.fCoord[0], m.m[0][0]) + hsScalarMul(p.fCoord[1], m.m[0][1]),
					hsScalarMul(p.fCoord[0], m.m[1][0]) + hsScalarMul(p.fCoord[1], m.m[1][1]),
					p.fCoord[2]);		
}

////////////////////////////////////////////////////////////////////////////

inline hsPoint3 operator+(const hsPoint3& s, const hsPoint3& t)
{
	hsPoint3		result;
	
	return *result.Set(s.fCoord[0] + t.fCoord[0], s.fCoord[1] + t.fCoord[1], s.fCoord[2] + t.fCoord[2]);
}

inline hsPoint3 operator+(const hsPoint3& s, const hsVector3& t)
{
	hsPoint3		result;
	
	return *result.Set(s.fCoord[0] + t.fCoord[0], s.fCoord[1] + t.fCoord[1], s.fCoord[2] + t.fCoord[2]);
}

inline hsPoint3 operator-(const hsPoint3& s, const hsPoint3& t)
{
	hsPoint3		result;
	
	return *result.Set(s.fCoord[0] - t.fCoord[0], s.fCoord[1] - t.fCoord[1], s.fCoord[2] - t.fCoord[2]);
}

// unary -
inline hsPoint3 operator-(const hsPoint3& s)
{
	hsPoint3		result;
	return *result.Set(-s.fCoord[0], -s.fCoord[1], -s.fCoord[2]);
}

inline hsPoint3 operator-(const hsPoint3& s, const hsVector3& t)
{
	hsPoint3		result;
	
	return *result.Set(s.fCoord[0] - t.fCoord[0], s.fCoord[1] - t.fCoord[1], s.fCoord[2] - t.fCoord[2]);
}

inline hsPoint3 operator*(const hsPoint3& s, const hsScalar& t)
{
	hsPoint3		result;	
	return *result.Set(hsScalarMul(s.fCoord[0], t), hsScalarMul(s.fCoord[1], t), hsScalarMul(s.fCoord[2], t));
}

inline hsPoint3 operator/(const hsPoint3& s, const hsScalar& t)
{
	hsPoint3		result;	
	return *result.Set(hsScalarDiv(s.fCoord[0], t), hsScalarDiv(s.fCoord[1], t), hsScalarDiv(s.fCoord[2], t));
}

inline hsPoint3 operator*(const hsScalar& t, const hsPoint3& s)
{
	hsPoint3		result;
	
	return *result.Set(hsScalarMul(s.fCoord[0], t), hsScalarMul(s.fCoord[1], t), hsScalarMul(s.fCoord[2], t));
}

////////////////////////////////////////////////////////////////////////////

struct hsMatrix44 {
	hsScalar 			fMap[4][4];
	
	hsMatrix44&		MakeTranslateMat(const hsVector3 *trans);
	hsMatrix44&		MakeScaleMat(const hsVector3 *scale);
	void			GetTranslate(hsVector3 *pt);
	void			GetTranspose(hsMatrix44 *out);
	hsMatrix44&		Translate(const hsVector3 *);
	hsMatrix44&		Scale(const hsVector3 *);
	hsMatrix44& 	Rotate(int axis, hsScalar radians);
	hsMatrix44&		Reset();
	hsMatrix44&		MakeCamera(const hsPoint3* from, const hsPoint3* at,
						const hsVector3* up);
	hsScalar			Determinant() const;
	hsMatrix44*		Inverse(hsMatrix44* inverse) const;

	friend hsPoint3		operator*(const hsMatrix44& m, const hsPoint3& p)
					{	
						hsPoint3 rVal;

						rVal.fCoord[0] = hsScalarMul(p.fCoord[0], m.fMap[0][0]) + hsScalarMul(p.fCoord[1], m.fMap[0][1]) + hsScalarMul(p.fCoord[2], m.fMap[0][2]) + m.fMap[0][3];
						rVal.fCoord[1] = hsScalarMul(p.fCoord[0], m.fMap[1][0]) + hsScalarMul(p.fCoord[1], m.fMap[1][1]) + hsScalarMul(p.fCoord[2], m.fMap[1][2])  + m.fMap[1][3];
						rVal.fCoord[2] = hsScalarMul(p.fCoord[0], m.fMap[2][0]) + hsScalarMul(p.fCoord[1], m.fMap[2][1]) + hsScalarMul(p.fCoord[2], m.fMap[2][2])  + m.fMap[2][3];
						return rVal;
					}
	friend  hsVector3 operator*(const hsMatrix44& m, const hsVector3& p);
	friend hsMatrix44	operator*(const hsMatrix44& a, const hsMatrix44& b);
	
	hsPoint3*			MapPoints(long count, hsPoint3 points[]) const;
	
	Boolean			IsIdentity(void);

	friend int operator==(const hsMatrix44& s, const hsMatrix44& t);
	friend int operator!=(const hsMatrix44& s, const hsMatrix44& t);

	void Read(hsStream *stream);
	void Write(hsStream *stream);
};

#endif
