/*
 * @(#)mlib_ImageCreate.c	1.7 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#ifdef _MSC_VER
#pragma warning(disable:4068 4244 4101)
#endif /*_MSC_VER*/

#ifdef __SUNPRO_C
#pragma ident        "@(#)mlib_ImageCreate.c	1.16        98/07/27 SMI"
#endif /* __SUNPRO_C */

/*
 * FILENAME: mlib_ImageCreate.c 
 *
 * FUNCTION
 *      mlib_ImageCreateStruct   - create image data structure
 *      mlib_ImageCreate         - create image data structure and allocate
 *                                 memory for image data
 *      mlib_ImageDelete         - delete image
 *      mlib_ImageCreateSubimage - create sub-image
 *
 *      mlib_ImageCreateRowTable   - create row starts pointer table
 *      mlib_ImageDeleteRowTable   - delete row starts pointer table
 * SYNOPSIS
 *        mlib_image *mlib_ImageCreateStruct(mlib_type type, 
 *                                           mlib_s32  channels, 
 *                                           mlib_s32  width, 
 *                                           mlib_s32  height,
 *                                           mlib_s32  stride, 
 *                                           void      *data)
 *
 *        mlib_image *mlib_ImageCreate(mlib_type type, 
 *                                     mlib_s32  channels, 
 *                                     mlib_s32  width, 
 *                                     mlib_s32  height)
 *
 *        void mlib_ImageDelete(mlib_image *img)
 *
 *        mlib_image *mlib_ImageCreateSubimage(mlib_image *img,
 *                                             mlib_s32 x, mlib_s32 y,
 *                                             mlib_s32 w, mlib_s32 h)
 *
 *        void *mlib_ImageCreateRowTable(mlib_image *img)
 *
 *        void mlib_ImageDeleteRowTable(mlib_image *img)
 *
 * ARGUMENT
 *      img      pointer to image data structure
 *      type     image data type, one of MLIB_BIT, MLIB_BYTE, MLIB_SHORT,
 *               MLIB_INT, MLIB_FLOAT or MLIB_DOUBLE
 *      channels number of image channels
 *      width    image width in pixels
 *      height   image height in pixels
 *      stride   linebytes( bytes to next row) of the image
 *      data     pointer to image data allocated by user
 *
 * DESCRIPTION
 *        mlib_ImageCreateStruct() creates a mediaLib image data structure 
 *        using parameter supplied by user.
 *
 *        mlib_ImageCreate() creates a mediaLib image data structure and 
 *        allocates memory space for image data.
 *
 *        mlib_ImageDelete() deletes the mediaLib image data structure
 *        and frees the memory space of the image data if it is allocated
 *        through mlib_ImageCreate().
 *
 *        mlib_ImageCreateSubimage() creates a mediaLib image structure 
 *        for a sub-image based on a source image.
 *
 *        mlib_ImageCreateRowTable() creates row starts pointer table and
 *        puts it into mlib_image->state field.
 *         
 *        mlib_ImageDeleteRowTable() delete row starts pointer table from
 *        image and puts NULL into mlib_image->state field.
 */

#include <stdlib.h>
#include <mlib_image.h>

/***************************************************************/

void *mlib_ImageCreateRowTable(mlib_image *image) {
  mlib_u8  **rtable, *tline;
  mlib_s32 i, im_height, im_stride;
  
  if (image->state) return image->state;
  
  im_height = mlib_ImageGetHeight(image);
  im_stride = mlib_ImageGetStride(image);
  tline     = mlib_ImageGetData(image);
  rtable    = mlib_malloc((3+im_height)*sizeof(mlib_u8 *));
  rtable[1] = (mlib_u8*)((void **)rtable + 1);
  rtable[2+im_height] = (mlib_u8*)((void **)rtable + 1);
  for(i = 0; i < im_height; i++) {
    rtable[i+2] = tline;
    tline    += im_stride;
  }
  image->state = ((void **)rtable + 2);
  
  return image->state;
}

/***************************************************************/

void mlib_ImageDeleteRowTable(mlib_image *image) {
  void **state = image->state;
  
  if (!state) return;
  
  mlib_free(state-2);
  image->state = 0;
}

/***************************************************************/

mlib_image *mlib_ImageCreateStruct(mlib_type type, 
                                   mlib_s32  channels, 
                                   mlib_s32  width, 
                                   mlib_s32  height,
                                   mlib_s32  stride, 
                                   void      *data)
{
  mlib_image         *image;
  int                wb;                /* width in bytes */

  /* sanity check */
  if (width < 0 || height < 0 || stride < 0 || channels < 0) return NULL;

  if (data == NULL){
    return NULL;
  }

  image = (mlib_image *)mlib_malloc(sizeof(mlib_image));
  if (image == NULL) {
    return NULL;
  };

  image->flags    = ~MLIB_IMAGE_ATTRIBUTESET;

  image->type     = type;
  image->channels = channels;
  image->width    = width;
  image->flags   |= ((width & 0xf) << 8);          /* set width field  */
  image->height   = height;
  image->flags   |= ((height & 0xf) << 12);        /* set height field */
  image->stride   = stride;
  image->flags   |= ((stride & 0xf) << 16);        /* set stride field */
  image->data     = data;
  image->flags   |= (mlib_addr)data & 0xff;
  image->flags   |= MLIB_IMAGE_USERALLOCATED;      /* user allocated data */

  /* Check if stride == width
   * If it is then image can be treated as a 1-D vector
   */
  switch (type) {
    case MLIB_DOUBLE:
      wb = width * channels * 8;
      break;
    case MLIB_FLOAT:
    case MLIB_INT:
      wb = width * channels * 4;
      break;
    case MLIB_SHORT:
      wb = width * channels * 2;
      break;
    case MLIB_BYTE:
      wb = width * channels;
      break;
    case MLIB_BIT:
      wb = (width * channels + 7) / 8;
      break;
    default:
      mlib_free(image);
      return NULL;
  }
  if ((stride != wb) ||
      ((type == MLIB_BIT) && (stride * 8 != width * channels))) {
    image->flags |= MLIB_IMAGE_ONEDVECTOR;
  } 
 
  image->flags &= MLIB_IMAGE_ATTRIBUTESET;
  image->state  = NULL;
    
  return image;
} 

/***************************************************************/

mlib_image *mlib_ImageCreate(mlib_type type, 
                             mlib_s32  channels, 
                             mlib_s32  width, 
                             mlib_s32  height)
{
  mlib_image        *image;
  int                wb;                /* width in bytes */
  void                *data;

  /* sanity check */
  if (width < 0 || height < 0 || channels < 0) return NULL;

  image = (mlib_image *)mlib_malloc(sizeof(mlib_image));
  if (image == NULL) {
    return NULL;
  };

  image->flags    = ~MLIB_IMAGE_ATTRIBUTESET;

  image->type     = type;
  image->channels = channels;
  image->width    = width;
  image->flags   |= ((width & 0xf) << 8);        /* set width field  */
  image->height   = height;
  image->flags   |= ((height & 0xf) << 12);      /* set height field */

  switch (type) {
    case MLIB_DOUBLE:
      wb = width * channels * 8;
      break;
    case MLIB_FLOAT:
    case MLIB_INT:
      wb = width * channels * 4;
      break;
    case MLIB_SHORT:
      wb = width * channels * 2;
      break;
    case MLIB_BYTE:
      wb = width * channels;
      break;
    case MLIB_BIT:
      wb = (width * channels + 7) / 8;
      break;
    default:
      mlib_free(image);
      return NULL;
  }

  data = mlib_malloc(wb * height);
  if (data == NULL) {
    mlib_free(image);
    return NULL;
  }

  image->stride = wb;
  image->flags |= ((wb & 0xf) << 16);                /* set stride field */
  image->data   = data;
  image->flags |= (mlib_addr)data & 0xff;

  if ((type == MLIB_BIT) && (wb * 8 != width * channels)) {
    image->flags |= MLIB_IMAGE_ONEDVECTOR;           /* not 1-d vector */
  }

  image->flags &= MLIB_IMAGE_ATTRIBUTESET;
  image->state  = NULL;
    
  return image;
} 

/***************************************************************/

void mlib_ImageDelete(mlib_image *img)
{

  if ((img->flags & MLIB_IMAGE_USERALLOCATED) == 0) {
    mlib_free(img->data);
  }
  mlib_ImageDeleteRowTable(img);
  mlib_free(img);
} 

/***************************************************************/

mlib_image *mlib_ImageCreateSubimage(mlib_image *img,
                                     mlib_s32 x, mlib_s32 y,
                                     mlib_s32 w, mlib_s32 h)
{
  mlib_image         *subimage;
  mlib_type        type;
  mlib_s32        channels;
  mlib_s32        width;                        /* for parent image */
  mlib_s32        height;                        /* for parent image */
  mlib_s32        stride;
  void                *data;

  /* sanity check */
  if (w <= 0 || h <= 0 || img == NULL) return NULL;

  type     = img->type;
  channels = img->channels;
  width    = img->width;
  height   = img->height;
  stride   = img->stride;

  /* if bit image, force the new origin to fall on byte boundary */
  if (type == MLIB_BIT) {
    x &= ~7;
  } 

  /* clip the sum-image with respect to the parent image */
  if (((x + w) <= 0) || ((y + h) <= 0) ||
      (x >= width) || (y >= height)) {
    return NULL;
  }
  else {
    if (x < 0) {
      w += x;                                        /* x is negative */
      x = 0;
    }
    if (y < 0) {
      h += y;                                        /* y is negative */
      y = 0;
    }
    if ((x + w) > width) {
      w = width - x;
    }
    if ((y + h) > height) {
      h = height - y;
    }
  }

  /* compute sub-image origin */
  switch (type) {
    case MLIB_DOUBLE:
      data = (mlib_u8 *)(img->data) + y * stride + x * channels * 8;
      break;
    case MLIB_FLOAT:
    case MLIB_INT:
      data = (mlib_u8 *)(img->data) + y * stride + x * channels * 4;
      break;
    case MLIB_SHORT:
      data = (mlib_u8 *)(img->data) + y * stride + x * channels * 2;
      break;
    case MLIB_BYTE:
      data = (mlib_u8 *)(img->data) + y * stride + x * channels;
      break;
    case MLIB_BIT:
      data = (mlib_u8 *)(img->data) + y * stride + x * channels / 8;
      break;
    default:
      return NULL;
  }

  subimage = mlib_ImageCreateStruct(type, 
                                    channels, 
                                    w,
                                    h,
                                    stride,
                                    data);

  return subimage;
}

/***************************************************************/
