/*
 * @(#)mlib_c_ImageConv3x3.c	1.8 98/09/05
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#ifdef _MSC_VER
#pragma warning(disable:4068 4244 4101)
#endif /*_MSC_VER*/

/*
 * This fixes a compiler problem with Windows Visual C++ and Solaris x86 C
 * compilers.  With Solaris Sparc compilers (4.2?), if a double floating point
 * value is stored in an integer and is larger than the integer, the value
 * stored will be MAX_INT.  With Windows Visual C++ and Solaris x86 C
 * compiler, the value is truncated.
 */
#ifdef WIN32
#  define NEED_CLAMPING
#endif

#ifdef NEED_CLAMPING
#  ifndef i386
#    define i386
#  endif
#endif


#ifdef __SUNPRO_C
#pragma ident	"@(#)mlib_c_ImageConv3x3.c	1.16	98/09/02 SMI"
#endif /* __SUNPRO_C */

/*
 * FUNCTION
 *      mlib_ImageConv3x3 - image convolution with edge condition
 *
 * SYNOPSIS
 *      mlib_status mlib_ImageConv3x3(mlib_image *dst,
 *                                    mlib_image *src,
 *                                    mlib_s32   *kernel,
 *                                    mlib_s32   scale,
 *                                    mlib_s32   cmask,
 *                                    mlib_edge  edge)
 *
 * ARGUMENTS
 *      dst       Pointer to destination image.
 *      src       Pointer to source image.
 *      kernel    Pointer to convolution kernel.
 *      scale     The scaling factor to convert the input integer
 *                coefficients into floating-point coefficients:
 *                floating-point coefficient = integer coefficient * 2^(-scale)
 *      cmask     Channel mask to indicate the channels to be convolved.
 *                Each bit of which represents a channel in the image. The
 *                channels corresponded to 1 bits are those to be processed.
 *      edge      Type of edge condition.
 *
 * DESCRIPTION
 *      2-D convolution, 3x3 kernel.
 *
 *      The src and the dst must be the same type and have same number
 *      of channels (1, 2, 3, or 4). The unselected channels are not
 *      overwritten. If both src and dst have just one channel,
 *      cmask is ignored.
 *
 *      The edge condition can be one of the following:
 *              MLIB_EDGE_DST_NO_WRITE  (default)
 *              MLIB_EDGE_DST_FILL_ZERO
 *              MLIB_EDGE_DST_COPY_SRC
 *              MLIB_EDGE_SRC_EXTEND
 */

/***********************************************************************/

#include <stdlib.h>
#include <mlib_image.h>

mlib_status mlib_c_conv3x3_32( mlib_image *src, mlib_image *dst,
             mlib_s32 *kernel, mlib_s32 scalef_expon,
                               mlib_s32 cmask, mlib_s32 edge);
mlib_status mlib_c_conv3x3_16( mlib_image *src, mlib_image *dst,
             mlib_s32 *kernel, mlib_s32 scalef_expon,
                               mlib_s32 cmask, mlib_s32 edge);
mlib_status mlib_c_conv3x3_8( mlib_image *src, mlib_image *dst,
            mlib_s32 *kernel, mlib_s32 scalef_expon,
                              mlib_s32 cmask, mlib_s32 edge);

mlib_status mlib_c_ImageConvCopyEdge(mlib_image *dst, mlib_image *src, mlib_s32 dx_l,
                                     mlib_s32 dx_r, mlib_s32 dy_t, mlib_s32 dy_b, mlib_s32 cmask);
mlib_status mlib_c_ImageConvClearEdge(mlib_image *img, mlib_s32 dx_l, mlib_s32 dx_r,
                                      mlib_s32 dy_t, mlib_s32 dy_b, mlib_s32 *color, mlib_s32 cmask);

#ifdef i386
#define SAT_U8(x) (mlib_s32)(((x) < MLIB_U8_MIN) ? MLIB_U8_MIN : (((x) > MLIB_U8_MAX) ? MLIB_U8_MAX : (x)))
#define SAT_S16(x) (mlib_s32)(((x) < MLIB_S16_MIN) ? MLIB_S16_MIN : (((x) > MLIB_S16_MAX) ? MLIB_S16_MAX : (x)))
#define SAT_S32(x) (mlib_s32)(((x) < MLIB_S32_MIN) ? MLIB_S32_MIN : (((x) > MLIB_S32_MAX) ? MLIB_S32_MAX : (x)))
#else
#define SAT_U8(x) (mlib_s32)((x) - sat)
#define SAT_S16(x) (mlib_s32)(x)
#define SAT_S32(x) (mlib_s32)(x)
#endif


/***********************************************************************/

#ifdef MLIB_TEST

mlib_status mlib_c_ImageConv3x3( mlib_image *dst,
                                 mlib_image *src,
                                 mlib_s32   *kernel,
                                 mlib_s32   scalef_exponent,
                                 mlib_s32   cmask,
                                 mlib_edge  edge)

#else
mlib_status mlib_ImageConv3x3( mlib_image *dst,
                               mlib_image *src,
                               mlib_s32   *kernel,
                               mlib_s32   scalef_exponent,
                               mlib_s32   cmask,
                               mlib_edge  edge)

#endif

{
   mlib_status stat;
   mlib_s32    slb, dlb,      /* line bytes (y stride) */
               src_wid, src_hgt,
               dst_wid, dst_hgt,
               src_nch, dst_nch;
   mlib_type   src_dtype, dst_dtype;
   mlib_s32    zero[4] = {0, 0, 0, 0};
   void        *sa, *da;

   if (dst == NULL || src == NULL) return MLIB_FAILURE;

   src_dtype = mlib_ImageGetType(src);
   dst_dtype = mlib_ImageGetType(dst);
   src_nch = mlib_ImageGetChannels(src);
   dst_nch = mlib_ImageGetChannels(dst);
   src_wid = mlib_ImageGetWidth(src);
   src_hgt = mlib_ImageGetHeight(src);
   dst_wid = mlib_ImageGetWidth(dst);
   dst_hgt = mlib_ImageGetHeight(dst);
   slb  = mlib_ImageGetStride(src);
   dlb  = mlib_ImageGetStride(dst);
   sa  = mlib_ImageGetData(src);
   da  = mlib_ImageGetData(dst);

   if ( src_dtype != dst_dtype || src_nch != dst_nch ||
        src_wid != dst_wid || src_hgt != dst_hgt || src_wid < 3 || src_hgt < 3 ||
        sa == NULL || da == NULL || kernel == NULL || scalef_exponent < 0) {
      stat =  MLIB_FAILURE;
   }
   else {
     if (src_nch == 1) cmask = 1;
      switch (src_dtype) {
      case MLIB_BYTE:
        if ( slb < (src_wid * src_nch)  ||  dlb < (dst_wid * dst_nch) )
          stat = MLIB_FAILURE;
        else {
        switch (edge) {
          case MLIB_EDGE_DST_FILL_ZERO:
            stat = mlib_c_conv3x3_8( src, dst, kernel, scalef_exponent, cmask, 0);
            if (stat != MLIB_SUCCESS) break;
            stat = mlib_c_ImageConvClearEdge( dst, 1, 1, 1, 1, zero, cmask);
            break;
          case MLIB_EDGE_DST_COPY_SRC:
            stat = mlib_c_conv3x3_8( src, dst, kernel, scalef_exponent, cmask, 0);
            if (stat != MLIB_SUCCESS) break;
            stat = mlib_c_ImageConvCopyEdge( dst, src, 1, 1, 1, 1, cmask);
            break;
          case MLIB_EDGE_SRC_EXTEND:
            stat = mlib_c_conv3x3_8( src, dst, kernel, scalef_exponent, cmask, 1);
            break;
          case MLIB_EDGE_DST_NO_WRITE:
          default:
            stat = mlib_c_conv3x3_8( src, dst, kernel, scalef_exponent, cmask, 0);
            break;
          }
        }
        break;
      case MLIB_SHORT:  /* convert wid from pixels to bytes */
        if ( slb < ((src_wid << 1) * src_nch)  ||  dlb < ((dst_wid << 1) * dst_nch) )
          stat = MLIB_FAILURE;
        else {
          switch (edge) {
          case MLIB_EDGE_DST_FILL_ZERO:
            stat = mlib_c_conv3x3_16( src, dst, kernel, scalef_exponent, cmask, 0);
            if (stat != MLIB_SUCCESS) break;
            stat = mlib_c_ImageConvClearEdge( dst, 1, 1, 1, 1, zero, cmask);
            break;
          case MLIB_EDGE_DST_COPY_SRC:
            stat = mlib_c_conv3x3_16( src, dst, kernel, scalef_exponent, cmask, 0);
            if (stat != MLIB_SUCCESS) break;
            stat = mlib_c_ImageConvCopyEdge( dst, src, 1, 1, 1, 1, cmask);
            break;
          case MLIB_EDGE_SRC_EXTEND:
            stat = mlib_c_conv3x3_16( src, dst, kernel, scalef_exponent, cmask, 1);
            break;
          case MLIB_EDGE_DST_NO_WRITE:
          default:
            stat = mlib_c_conv3x3_16( src, dst, kernel, scalef_exponent, cmask, 0);
            break;
          }
        }
        break;
      case MLIB_INT:  /* convert wid from pixels to bytes */
        if ( slb < ((src_wid << 2) * src_nch)  ||  dlb < ((dst_wid << 2) * dst_nch) )
          stat = MLIB_FAILURE;
        else {
          switch (edge) {
          case MLIB_EDGE_DST_FILL_ZERO:
            stat = mlib_c_conv3x3_32( src, dst, kernel, scalef_exponent, cmask, 0);
            if (stat != MLIB_SUCCESS) break;
            stat = mlib_c_ImageConvClearEdge( dst, 1, 1, 1, 1, zero, cmask);
            break;
          case MLIB_EDGE_DST_COPY_SRC:
            stat = mlib_c_conv3x3_32( src, dst, kernel, scalef_exponent, cmask, 0);
            if (stat != MLIB_SUCCESS) break;
            stat = mlib_c_ImageConvCopyEdge( dst, src, 1, 1, 1, 1, cmask);
            break;
          case MLIB_EDGE_SRC_EXTEND:
            stat = mlib_c_conv3x3_32( src, dst, kernel, scalef_exponent, cmask, 1);
            break;
          case MLIB_EDGE_DST_NO_WRITE:
          default:
            stat = mlib_c_conv3x3_32( src, dst, kernel, scalef_exponent, cmask, 0);
            break;
          }
        }
        break;
      default:
        stat = MLIB_FAILURE;
        break;
      }
   }
   return ( stat );
}

/************************************************************************/

mlib_status mlib_c_conv3x3_32 ( mlib_image *src,        /* source image */
                                mlib_image *dst,        /* destination image */
                                mlib_s32 *kern,         /* convolution kernel */
                                mlib_s32 scalef_expon,  /* scalefactor=2**N */
                                mlib_s32 cmask,
                                mlib_s32 edge)
{
  mlib_s32 *sl, *adr_src;                       /* pointers for pixel and line of source */
  mlib_s32 *da, *dl, *adr_dst;                  /* pointers for pixel and line of destination */
  mlib_s32 row, col, block;                     /* indices */
  mlib_s32 *row0, *row1, *row2;
  double   scalef;                              /* divide coeffs by scalefactor=2**scalef_expon */
  double   sum1, sum2, sum3;                    /* accumulators */
  double   k0, k1, k2, k3, k4, k5, k6, k7, k8;  /* coeff's */
  double   p00, p01, p02, p03, p04,             /* row 0 pixel values */
           p10, p11, p12, p13, p14,             /* row 1 pixel values */
           p20, p21, p22, p23, p24;             /* row 2 pixel values */
  int      sll,                                 /* line length (in pixels) of source image */
           dll,                                 /* line length (in pixels) of destination image */
           wid,                                 /* src width in pixels */
           hgt;                                 /* src height in pixels */
  int      num_blocks;                          /* do output pts 3 at a time; loop count for this */
  int      leftover_cols;                       /* what can't be done in blocks */
  int      chan, testchan, nchannel;

  scalef = 1.0;
  while (scalef_expon > 30) {
    scalef /= (1 << 30);
    scalef_expon -= 30;
  }

  scalef /= (1 << scalef_expon);

  /* keep kernel in regs */
  k0 = scalef * kern[0];  k1 = scalef * kern[1];  k2 = scalef * kern[2];
  k3 = scalef * kern[3];  k4 = scalef * kern[4];  k5 = scalef * kern[5];
  k6 = scalef * kern[6];  k7 = scalef * kern[7];  k8 = scalef * kern[8];

  hgt = mlib_ImageGetHeight(src);
  wid = mlib_ImageGetWidth(src);
  nchannel = mlib_ImageGetChannels(src);
  sll = mlib_ImageGetStride(src) >> 2;           /* cnvrt from bytes to 32-bit pxls */
  dll = mlib_ImageGetStride(dst) >> 2;           /* cnvrt from bytes to 32-bit pxls */
  adr_src = (mlib_s32 *)mlib_ImageGetData(src);
  adr_dst = (mlib_s32 *)mlib_ImageGetData(dst);

  if (edge != 0)  wid += 1;

  num_blocks = (wid - 2)/3;                      /* (wid-2): no output on 1st, last columns */
  leftover_cols = (wid - 2) - (num_blocks * 3);
  if (edge == 0) hgt -= 2;                       /* boundaries: no output for first & last rows */

  testchan = 1;
  for ( chan = nchannel-1; chan >= 0; chan--) {
    if ((cmask & testchan) == 0) {
      testchan <<= 1;
      continue;
    }
    testchan <<= 1;
    sl = adr_src + chan;
    dl = da = adr_dst + chan;
    if (edge == 0) dl = da += dll + nchannel;

  /*
   * set up rowbufs 1 and 2;
   * rowbuf 0 is always reloaded inside the loop:
   */

    row1 = sl;
    if (edge == 0) sl += sll;
    row2 = sl;
    sl += sll;

    for ( row = 0; row < hgt; row++ ) {

      row0 = row1;
      row1 = row2;
      row2 = sl;

      col = 0;

      p00 = row0[col]; p10 = row1[col]; p20 = row2[col];
      if (edge == 0) col += nchannel;
      p01 = row0[col]; p11 = row1[col]; p21 = row2[col]; col += nchannel;

      /* Process along row, getting 3 output points per loop iteration: */

#ifdef __SUNPRO_C
#pragma pipeloop(0)
#endif /* __SUNPRO_C */
      for ( block = 0; block < num_blocks; block++ ) {

        p02 = row0[col]; p12 = row1[col]; p22 = row2[col]; col+=nchannel;
        p03 = row0[col]; p13 = row1[col]; p23 = row2[col]; col+=nchannel;
        p04 = row0[col]; p14 = row1[col]; p24 = row2[col]; col+=nchannel;

        sum1 = sum2 = sum3 = 0.;

        sum1 += p00 * k0; sum2 += p01 * k0; sum3 += p02 * k0;
        sum1 += p01 * k1; sum2 += p02 * k1; sum3 += p03 * k1;
        sum1 += p02 * k2; sum2 += p03 * k2; sum3 += p04 * k2;
        sum1 += p10 * k3; sum2 += p11 * k3; sum3 += p12 * k3;
        sum1 += p11 * k4; sum2 += p12 * k4; sum3 += p13 * k4;
        sum1 += p12 * k5; sum2 += p13 * k5; sum3 += p14 * k5;
        sum1 += p20 * k6; sum2 += p21 * k6; sum3 += p22 * k6;
        sum1 += p21 * k7; sum2 += p22 * k7; sum3 += p23 * k7;
        sum1 += p22 * k8; sum2 += p23 * k8; sum3 += p24 * k8;

        *da = SAT_S32(sum1);
        da += nchannel;
        *da = SAT_S32(sum2);
        da += nchannel;
        *da = SAT_S32(sum3);
        da += nchannel;

        p00 = p03; p01 = p04;                    /* shift last 2 row0 elements */
        p10 = p13; p11 = p14;                    /* shift last 2 row1 elements */
        p20 = p23; p21 = p24;                    /* shift last 2 row2 elements */
      }

      if (leftover_cols) {

        p02 = row0[col];
        p12 = row1[col];
        p22 = row2[col]; col+=nchannel;
        sum1 = 0.;
        sum1 += p00 * k0;                        /* sum of prods */
        sum1 += p01 * k1;
        sum1 += p02 * k2;
        sum1 += p10 * k3;
        sum1 += p11 * k4;
        sum1 += p12 * k5;
        sum1 += p20 * k6;
        sum1 += p21 * k7;
        sum1 += p22 * k8;
        *da = SAT_S32(sum1);
        da += nchannel;
        p00 = p01; p01 = p02;                    /* shift last 2 row0 elements */
        p10 = p11; p11 = p12;                    /* shift last 2 row1 elements */
        p20 = p21; p21 = p22;                    /* shift last 2 row2 elements */

        if (leftover_cols > 1) {

          p02 = row0[col];
          p12 = row1[col];
          p22 = row2[col];
          sum1 = 0.;
          sum1 += p00 * k0;                        /* sum of prods */
          sum1 += p01 * k1;
          sum1 += p02 * k2;
          sum1 += p10 * k3;
          sum1 += p11 * k4;
          sum1 += p12 * k5;
          sum1 += p20 * k6;
          sum1 += p21 * k7;
          sum1 += p22 * k8;
          *da = SAT_S32(sum1);
          da += nchannel;
          p00 = p01; p01 = p02;                    /* shift last 2 row0 elements */
          p10 = p11; p11 = p12;                    /* shift last 2 row1 elements */
          p20 = p21; p21 = p22;                    /* shift last 2 row2 elements */
        }
      }

      if (edge != 0) {
        sum1 = 0.;
        sum1 += p00 * k0;                        /* sum of prods */
        sum1 += p01 * k1;
        sum1 += p01 * k2;
        sum1 += p10 * k3;
        sum1 += p11 * k4;
        sum1 += p11 * k5;
        sum1 += p20 * k6;
        sum1 += p21 * k7;
        sum1 += p21 * k8;
        *da = SAT_S32(sum1);
      }

      sl = sl + sll;                             /* set src ptrs to next row */
      dl = da = dl + dll;                        /* set dst ptrs to next row */
      if ((edge != 0) && (row == (hgt - 2))) sl -= sll;
    }
  }

  return( MLIB_SUCCESS );
}

/************************************************************************/

mlib_status mlib_c_conv3x3_16 ( mlib_image *src,        /* source image */
                                mlib_image *dst,        /* destination image */
                                mlib_s32 *kern,         /* convolution kernel */
                                mlib_s32 scalef_expon,  /* scalefactor=2**N */
                                mlib_s32 cmask,
                                mlib_s32 edge)
{
  mlib_s16 *sa, *sl, *adr_src;                  /* pointers for pixel and line of source */
  mlib_s16 *da, *dl, *adr_dst;                  /* pointers for pixel and line of destination */
  int      row, col, block;                     /* indices */
  double   scalef;                              /* divide coeffs by scalefactor=2**scalef_expon */
  double   sum1, sum2, sum3;                    /* accumulators */
  double   k0, k1, k2, k3, k4, k5, k6, k7, k8;  /* coeff's */
  double   p00, p01, p02, p03, p04,             /* row 0 pixel values */
           p10, p11, p12, p13, p14,             /* row 1 pixel values */
           p20, p21, p22, p23, p24;             /* row 2 pixel values */
  int      sll,                                 /* line length (in pixels) of source image */
           dll,                                 /* line length (in pixels) of destination image */
           wid,                                 /* src width in pixels */
           hgt;                                 /* src height in pixels */
  int      isum;                                /* integer accum */
  int      num_blocks;                          /* do output pts 3 at a time; loop count for this */
  int      leftover_cols;                       /* what can't be done in blocks */
  /*
   * Keep 3 rows of data in 32-bit int buffers to save cost of repeatedly
   * casting 16-bit short int data to 32-bit float. (To caste short to float,
   * it reads short from mem into int reg, converts to 32-bit int, writes it
   * to mem, then reads it into float reg, converting to float once inside
   * the float reg. Thus the cost is in converting from short to int, not
   * int to float.)
   */
  int       *rowbuf[3];                          /* int row buf ptrs */
  int       *row0, *row1, *row2;                 /* dedicated ptrs to rowbufs (never incr'd) */
  int       *prow;                               /* general-purpose ptr into a row */
  /* Store results of each row to an integer type intermediate output buf;
   * then convert row buf data to short and store in final destination.
   * This is actually faster than casting each float result to short as
   * it is computed:
   */
  int      *outbuf, *pout;                       /* int output buf & ptr */
  int      chan, testchan, nchannel;

#ifdef i386
  scalef = 1.0;
#else
  scalef = 65536.0;
#endif
  while (scalef_expon > 30) {
    scalef /= (1 << 30);
    scalef_expon -= 30;
  }

  scalef /= (1 << scalef_expon);

  /* keep kernel in regs */
  k0 = scalef * kern[0];  k1 = scalef * kern[1];  k2 = scalef * kern[2];
  k3 = scalef * kern[3];  k4 = scalef * kern[4];  k5 = scalef * kern[5];
  k6 = scalef * kern[6];  k7 = scalef * kern[7];  k8 = scalef * kern[8];

  hgt = mlib_ImageGetHeight(src);
  wid = mlib_ImageGetWidth(src);
  nchannel = mlib_ImageGetChannels(src);
  sll = mlib_ImageGetStride(src) >> 1;           /* cnvrt from bytes to 16-bit pxls */
  dll = mlib_ImageGetStride(dst) >> 1;           /* cnvrt from bytes to 16-bit pxls */
  adr_src = (mlib_s16 *)mlib_ImageGetData(src);
  adr_dst = (mlib_s16 *)mlib_ImageGetData(dst);

  if (edge != 0)  wid += 2;

  /* allocate row bufs: */

  rowbuf[0] = (int *)mlib_malloc(4 * wid*sizeof(int));
  if (rowbuf[0] == NULL) {
    return( MLIB_FAILURE );
  }
  rowbuf[1] = rowbuf[0] + wid ;
  rowbuf[2] = rowbuf[1] + wid ;
  outbuf    = rowbuf[2] + wid ;

  num_blocks = (wid - 2)/3;                      /* (wid-2): no output on 1st, last columns */
  leftover_cols = (wid - 2) - (num_blocks * 3);
  if (edge == 0) hgt -= 2;                       /* boundaries: no output for first & last rows */

  testchan = 1;
  for ( chan = nchannel-1; chan >= 0; chan--) {
    if ((cmask & testchan) == 0) {
      testchan <<= 1;
      continue;
    }
    testchan <<= 1;
    sl = sa = adr_src + chan;
    dl = da = adr_dst + chan;
    if (edge == 0) dl = da += dll + nchannel;

  /*
   * set up rowbufs 1 and 2;
   * rowbuf 0 is always reloaded inside the loop:
   */
    if (edge == 0) {
      for ( row = 1; row <= 2; row++ ) {         /* set up 2 line buffers */
        prow = rowbuf[row];                      /* ptr to int dst */
        for ( col = 0; col < wid; col++ ) {      /* copy shorts -> int buffer */
          *prow++ = (int)*sa;
          sa += nchannel;
        }
        sl = sa = sl + sll;                      /* set src ptrs to next row */
      }
    } else {
      for ( row = 1; row <= 2; row++ ) {         /* set up 2 line buffers */
        prow = rowbuf[row] + 1;                  /* ptr to int dst */
        for ( col = 1; col < wid-1; col++ ) {    /* copy shorts -> int buffer */
          *prow++ = (int)*sa;
          sa += nchannel;
        }
        rowbuf[row][0] = rowbuf[row][1];
        rowbuf[row][ wid-1] = rowbuf[row] [wid-2];
        sa = sl;
      }
      sl = sa = sl + sll;
    }

    row0 = rowbuf[0];                            /* set up line buf ptrs */
    row1 = rowbuf[1];
    row2 = rowbuf[2];

    for ( row = 0; row < hgt; row++ ) {

      pout = outbuf;
      prow = row0;                               /* save ptr to oldest row (load new row in this buf) */
      row0 = row1;                               /* former row 1 is now row 0 */
      row1 = row2;                               /* former row 2 is now row 1 */
      row2 = prow;                               /* former row0 buf is new row 2; fill with new data */
      if (edge == 0) {
        for ( col = 0; col < wid; col++ ) {      /* copy next row to int buffer */
          *prow++ = (int)*sa;
          sa += nchannel;
        }
      } else {
        if (row != (hgt-1)) {
          prow++;
          for ( col = 1; col < wid-1; col++ ) {  /* copy next row to int buffer */
            *prow++ = (int)*sa;
            sa += nchannel;
          }
          row2[0] = row2[1]; row2[ wid-1] = row2[ wid-2];
        } else  row2 = row1;
      }

      p00 = row0[0]; p01 = row0[1];
      p10 = row1[0]; p11 = row1[1];
      p20 = row2[0]; p21 = row2[1];
      col = 2;                                   /* next col to fill */

      /* Process along row, getting 3 output points per loop iteration: */

#ifdef __SUNPRO_C
#pragma pipeloop(0)
#endif /* __SUNPRO_C */
      for ( block = 0; block < num_blocks; block++ ) {

        p02 = row0[col]; p12 = row1[col]; p22 = row2[col++];
        p03 = row0[col]; p13 = row1[col]; p23 = row2[col++];
        p04 = row0[col]; p14 = row1[col]; p24 = row2[col++];

        sum1 = sum2 = sum3 = 0.;

        sum1 += p00 * k0; sum2 += p01 * k0; sum3 += p02 * k0;
        sum1 += p01 * k1; sum2 += p02 * k1; sum3 += p03 * k1;
        sum1 += p02 * k2; sum2 += p03 * k2; sum3 += p04 * k2;
        sum1 += p10 * k3; sum2 += p11 * k3; sum3 += p12 * k3;
        sum1 += p11 * k4; sum2 += p12 * k4; sum3 += p13 * k4;
        sum1 += p12 * k5; sum2 += p13 * k5; sum3 += p14 * k5;
        sum1 += p20 * k6; sum2 += p21 * k6; sum3 += p22 * k6;
        sum1 += p21 * k7; sum2 += p22 * k7; sum3 += p23 * k7;
        sum1 += p22 * k8; sum2 += p23 * k8; sum3 += p24 * k8;

        *pout++ = SAT_S16(sum1);
        *pout++ = SAT_S16(sum2);
        *pout++ = SAT_S16(sum3);

        p00 = p03; p01 = p04;                    /* shift last 2 row0 elements */
        p10 = p13; p11 = p14;                    /* shift last 2 row1 elements */
        p20 = p23; p21 = p24;                    /* shift last 2 row2 elements */
      }

      if (leftover_cols) {

        p02 = row0[col];
        p12 = row1[col];
        p22 = row2[col++];
        sum1 = 0.;
        sum1 += p00 * k0;                        /* sum of prods */
        sum1 += p01 * k1;
        sum1 += p02 * k2;
        sum1 += p10 * k3;
        sum1 += p11 * k4;
        sum1 += p12 * k5;
        sum1 += p20 * k6;
        sum1 += p21 * k7;
        sum1 += p22 * k8;
        *pout++ = SAT_S16(sum1);

        if (leftover_cols > 1) {

          p03 = row0[col];                       /* get next col of data */
          p13 = row1[col];
          p23 = row2[col];                       /* last col in row, so don't incr */
          sum1 = 0.;
          sum1 += p01 * k0;                      /* sum of prods */
          sum1 += p02 * k1;
          sum1 += p03 * k2;
          sum1 += p11 * k3;
          sum1 += p12 * k4;
          sum1 += p13 * k5;
          sum1 += p21 * k6;
          sum1 += p22 * k7;
          sum1 += p23 * k8;
          *pout++ = SAT_S16(sum1);
        }
      }

      pout = outbuf;
      for ( col = 0; col < wid - 2; col++ ) {    /* copy int outbuf to short dest buf */
        isum = *pout++;
#ifdef i386
        *da = isum;
#else
        *da = (isum >> 16);
#endif
        da += nchannel;
      }
      sl = sa = sl + sll;                        /* set src ptrs to next row */
      dl = da = dl + dll;                        /* set dst ptrs to next row */
    }
  }

  mlib_free( rowbuf[0] );
  return( MLIB_SUCCESS );
}

/****************************************************************************/

mlib_status mlib_c_conv3x3_8 ( mlib_image *src,        /* source image */
                               mlib_image *dst,        /* destination image */
                               mlib_s32 *kern,         /* convolution kernel */
                               mlib_s32 scalef_expon,  /* scalefactor=2**N */
                               mlib_s32 cmask,
                               mlib_s32 edge)
{
  int     row, col, block;                      /* indices */
  mlib_u8 *sa, *sl, *adr_src;                   /* pointers for pixel and line of source */
  mlib_u8 *da, *dl, *adr_dst;                   /* pointers for pixel and line of destination */
  double  scalef;                               /* divide coeffs by scalefactor=2**scalef_expon */
  double  sum1, sum2, sum3;                     /* accumulators */
  double  k0, k1, k2, k3, k4, k5, k6, k7, k8;   /* coeff's */
  double  p00, p01, p02, p03, p04,              /* row 0 pixel values */
          p10, p11, p12, p13, p14,              /* row 1 pixel values */
          p20, p21, p22, p23, p24;              /* row 2 pixel values */
  int     sll,                                  /* line length (in pixels) of source image */
          dll,                                  /* line length (in pixels) of destination image */
          wid,                                  /* src width in pixels */
          hgt;                                  /* src height in pixels */
  int     isum;                                 /* integer accum */
  int     num_blocks;                           /* do output pts 3 at a time; loop count for this */
  int     leftover_cols;                        /* what can't be done in blocks */
  /*
   * Keep 3 rows of data in 32-bit int buffers to save cost of repeatedly
   * casting 16-bit short int data to 32-bit float. (To caste short to float,
   * it reads short from mem into int reg, converts to 32-bit int, writes it
   * to mem, then reads it into float reg, converting to float once inside
   * the float reg. Thus the cost is in converting from short to int, not
   * int to float.)
   */
  int     *rowbuf[3];                           /* int row buf ptrs */
  int     *row0, *row1, *row2;                  /* dedicated ptrs to rowbufs (never incr'd) */
  int     *prow;                                /* general-purpose ptr into a row */
  /* Store results of each row to an integer type intermediate output buf;
   * then convert row buf data to short and store in final destination.
   * This is actually faster than casting each float result to short as
   * it is computed:
   */
  int     *outbuf, *pout;                       /* int output buf & ptr */
#ifndef i386
  int     tmp = 0x7FFFFFFF;
  double  sat;
#endif
  int     chan, testchan, nchannel;

#ifdef i386
  scalef = 1.0;
#else
  scalef = 16777216.0;
#endif
  while (scalef_expon > 30) {
    scalef /= (1 << 30);
    scalef_expon -= 30;
  }

  scalef /= (1 << scalef_expon);
#ifndef i386
  sat = (double)tmp + 1.;
#endif

  /* keep kernel in regs */
  k0 = scalef * kern[0];  k1 = scalef * kern[1];  k2 = scalef * kern[2];
  k3 = scalef * kern[3];  k4 = scalef * kern[4];  k5 = scalef * kern[5];
  k6 = scalef * kern[6];  k7 = scalef * kern[7];  k8 = scalef * kern[8];

  hgt = mlib_ImageGetHeight(src);
  wid = mlib_ImageGetWidth(src);
  nchannel = mlib_ImageGetChannels(src);
  sll = mlib_ImageGetStride(src);
  dll = mlib_ImageGetStride(dst);
  adr_src = (mlib_u8 *)mlib_ImageGetData(src);
  adr_dst = (mlib_u8 *)mlib_ImageGetData(dst);

  if (edge != 0)  wid += 2;

  /* allocate row bufs: */

  rowbuf[0] = (int *)mlib_malloc(4 * wid*sizeof(int));
  if (rowbuf[0] == NULL) {
    return( MLIB_FAILURE );
  }
  rowbuf[1] = rowbuf[0] + wid ;
  rowbuf[2] = rowbuf[1] + wid ;
  outbuf    = rowbuf[2] + wid ;

  num_blocks = (wid - 2)/3;                      /* (wid-2): no output on 1st, last columns */
  leftover_cols = (wid - 2) - (num_blocks * 3);

  if (edge == 0) hgt -= 2;                       /* boundaries: no output for first & last rows */

  testchan = 1;
  for ( chan = nchannel - 1; chan >= 0; chan--) {
    if ((cmask & testchan) == 0) {
      testchan <<= 1;
      continue;
    }
    testchan <<= 1;
    sl = sa = adr_src + chan;
    dl = da = adr_dst + chan;
    if (edge == 0) dl = da += dll + nchannel;

  /*
   * set up rowbufs 1 and 2;
   * rowbuf 0 is always reloaded inside the loop:
   */
    if (edge == 0) {
      for ( row = 1; row <= 2; row++ ) {         /* set up 2 line buffers */
        prow = rowbuf[row];                      /* ptr to int dst */
        for ( col = 0; col < wid; col++ ) {      /* copy shorts -> int buffer */
          *prow++ = (int)*sa;
          sa += nchannel;
        }
        sl = sa = sl + sll;                      /* set src ptrs to next row */
      }
    } else {
      for ( row = 1; row <= 2; row++ ) {         /* set up 2 line buffers */
        prow = rowbuf[row] + 1;                  /* ptr to int dst */
        for ( col = 1; col < wid-1; col++ ) {    /* copy shorts -> int buffer */
          *prow++ = (int)*sa;
          sa += nchannel;
        }
        rowbuf[row][0] = rowbuf[row][1];
        rowbuf[row][ wid-1] = rowbuf[row][ wid-2];
        sa = sl;
      }
      sl = sa = sl + sll;
    }

    row0 = rowbuf[0];                            /* set up line buf ptrs */
    row1 = rowbuf[1];
    row2 = rowbuf[2];

    for ( row = 0; row < hgt; row++ ) {

      pout = outbuf;
      prow = row0;                               /* save ptr to oldest row (load new row in this buf) */
      row0 = row1;                               /* former row 1 is now row 0 */
      row1 = row2;                               /* former row 2 is now row 1 */
      row2 = prow;                               /* former row0 buf is new row 2; fill with new data */
      if (edge == 0) {
        for ( col = 0; col < wid; col++ ) {      /* copy next row to int buffer */
          *prow++ = (int)*sa;
          sa += nchannel;
        }
      } else {
        if (row != (hgt-1)) {
          prow++;
          for ( col = 1; col < wid-1; col++ ) {  /* copy next row to int buffer */
            *prow++ = (int)*sa;
            sa += nchannel;
          }
          row2[0] = row2[1]; row2[ wid-1] = row2[ wid-2];
        } else  row2 = row1;
      }

      p00 = row0[0]; p01 = row0[1];
      p10 = row1[0]; p11 = row1[1];
      p20 = row2[0]; p21 = row2[1];
      col = 2;                                   /* next col to fill */

      /* Process along row, getting 3 output points per loop iteration: */

#ifdef __SUNPRO_C
#pragma pipeloop(0)
#endif /* __SUNPRO_C */
      for ( block = 0; block < num_blocks; block++ ) {

        p02 = row0[col]; p12 = row1[col]; p22 = row2[col++];
        p03 = row0[col]; p13 = row1[col]; p23 = row2[col++];
        p04 = row0[col]; p14 = row1[col]; p24 = row2[col++];

#ifdef i386
        sum1 = sum2 = sum3 = 0.5;
#else
        sum1 = sum2 = sum3 = 8388608.;
#endif

        sum1 += p00 * k0; sum2 += p01 * k0; sum3 += p02 * k0;
        sum1 += p01 * k1; sum2 += p02 * k1; sum3 += p03 * k1;
        sum1 += p02 * k2; sum2 += p03 * k2; sum3 += p04 * k2;
        sum1 += p10 * k3; sum2 += p11 * k3; sum3 += p12 * k3;
        sum1 += p11 * k4; sum2 += p12 * k4; sum3 += p13 * k4;
        sum1 += p12 * k5; sum2 += p13 * k5; sum3 += p14 * k5;
        sum1 += p20 * k6; sum2 += p21 * k6; sum3 += p22 * k6;
        sum1 += p21 * k7; sum2 += p22 * k7; sum3 += p23 * k7;
        sum1 += p22 * k8; sum2 += p23 * k8; sum3 += p24 * k8;

        *pout++ = SAT_U8(sum1);
        *pout++ = SAT_U8(sum2);
        *pout++ = SAT_U8(sum3);

        p00 = p03; p01 = p04;                    /* shift last 2 row0 elements */
        p10 = p13; p11 = p14;                    /* shift last 2 row1 elements */
        p20 = p23; p21 = p24;                    /* shift last 2 row2 elements */
      }

      if (leftover_cols) {

        p02 = row0[col];
        p12 = row1[col];
        p22 = row2[col++];
#ifdef i386
        sum1 = 0.5;
#else
        sum1 = 8388608.;
#endif
        sum1 += p00 * k0;                        /* sum of prods */
        sum1 += p01 * k1;
        sum1 += p02 * k2;
        sum1 += p10 * k3;
        sum1 += p11 * k4;
        sum1 += p12 * k5;
        sum1 += p20 * k6;
        sum1 += p21 * k7;
        sum1 += p22 * k8;
        *pout++ = SAT_U8(sum1);

        if (leftover_cols > 1) {

          p03 = row0[col];                       /* get next col of data */
          p13 = row1[col];
          p23 = row2[col];                       /* last col in row, so don't incr */
#ifdef i386
          sum1 = 0.5;
#else
          sum1 = 8388608.;
#endif
          sum1 += p01 * k0;                      /* sum of prods */
          sum1 += p02 * k1;
          sum1 += p03 * k2;
          sum1 += p11 * k3;
          sum1 += p12 * k4;
          sum1 += p13 * k5;
          sum1 += p21 * k6;
          sum1 += p22 * k7;
          sum1 += p23 * k8;
          *pout++ = SAT_U8(sum1);
        }
      }

      pout = outbuf;
      for ( col = 0; col < wid - 2; col++ ) {    /* copy int outbuf to short dest buf */
        isum = *pout++;
#ifdef i386
        *da = isum;
#else
        *da = ((isum >> 24) ^ 0x80);
#endif
        da += nchannel;
      }
      sl = sa = sl + sll;                        /* set src ptrs to next row */
      dl = da = dl + dll;                        /* set dst ptrs to next row */
    }
  }

  mlib_free( rowbuf[0] );
  return( MLIB_SUCCESS );
}
