/*
 * @(#)BufferedImageConsumer.java	1.26 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */



import java.util.Hashtable;
import java.awt.Transparency;
import java.awt.image.ImageProducer;
import java.awt.image.ImageConsumer;
import java.awt.image.ImageObserver;
import java.awt.image.ColorModel;
import java.awt.image.IndexColorModel;
import java.awt.Image;
import java.awt.image.BufferedImage;
import java.awt.color.ColorSpace;
import java.awt.color.ICC_ColorSpace;
import java.awt.color.ICC_ProfileGray;
import java.awt.color.ICC_ProfileRGB;
import java.awt.image.ComponentColorModel;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferInt;
import java.awt.image.DataBufferByte;


/**
 * The BufferedImageConsumer class implements a ImageConsumer which can be
 * attached to a ImageProducer object to retrieve a BufferedImage
 *
 * @see ColorModel#getRGBdefault
 *
 * @version 	1.7, 23 Nov 1996
 * @author 	Jim Graham
 */
public class BufferedImageConsumer implements java.awt.image.ImageConsumer {
    ImageProducer producer;

    int dstX;
    int dstY;
    int dstW = -1;
    int dstH = -1;

    ColorModel imageModel;
    byte[] bytePixels;
    int[] intPixels;
    int dstOff;
    int dstScan;

    private boolean allowGrayImages = false;
    private boolean grabbing;
    private int flags;

    private final int GRABBEDBITS = (ImageObserver.FRAMEBITS
				     | ImageObserver.ALLBITS);
    private final int DONEBITS = (GRABBEDBITS
				  | ImageObserver.ERROR);

    /**
     * Create a BufferedImageConsumer object to grab the (x, y, w, h)
     * rectangular
     * section of pixels from the image produced by the specified
     * ImageProducer into the given array.
     * The pixels are stored into the array in the default RGB ColorModel.
     * The RGB data for pixel (i, j) where (i, j) is inside the rectangle
     * (x, y, w, h) is stored in the array at
     * <tt>pix[(j - y) * scansize + (i - x) + off]</tt>.
     * @see ColorModel#getRGBdefault
     * @param img the image to retrieve pixels from
     * @param x the x coordinate of the upper left corner of the rectangle
     * of pixels to retrieve from the image, relative to the default
     * (unscaled) size of the image
     * @param y the y coordinate of the upper left corner of the rectangle
     * of pixels to retrieve from the image
     * @param w the width of the rectangle of pixels to retrieve
     * @param h the height of the rectangle of pixels to retrieve
     * @param pix the array of integers which are to be used to hold the
     * RGB pixels retrieved from the image
     * @param off the offset into the array of where to store the first pixel
     * @param scansize the distance from one row of pixels to the next in
     * the array
     */
    public BufferedImageConsumer(ImageProducer ip) {
	producer = ip;
	imageModel = ColorModel.getRGBdefault();
    }

    public BufferedImageConsumer(ImageProducer ip, boolean allowGrayImages) {
        this.allowGrayImages = allowGrayImages;
	producer = ip;
	imageModel = ColorModel.getRGBdefault();
    }

    /**
     * Request the PixelGrabber to start fetching the pixels.
     */
    public synchronized void startGrabbing() {
	if ((flags & DONEBITS) != 0) {
	    return;
	}
	if (!grabbing) {
	    grabbing = true;
	    flags &= ~(ImageObserver.ABORT);
	    producer.startProduction(this);
	}
    }

    /**
     * Request the PixelGrabber to abort the image fetch.
     */
    public synchronized void abortGrabbing() {
	imageComplete(IMAGEABORTED);
    }

    /**
     * Request the Image or ImageProducer to start delivering pixels and
     * wait for all of the pixels in the rectangle of interest to be
     * delivered.
     * @return true if the pixels were successfully grabbed, false on
     * abort, error or timeout
     * @exception InterruptedException
     *            Another thread has interrupted this thread.
     */
    public boolean grabPixels() throws InterruptedException {
	return grabPixels(0);
    }

    /**
     * Request the Image or ImageProducer to start delivering pixels and
     * wait for all of the pixels in the rectangle of interest to be
     * delivered or until the specified timeout has elapsed.
     * @param ms the number of milliseconds to wait for the image pixels
     * to arrive before timing out
     * @return true if the pixels were successfully grabbed, false on
     * abort, error or timeout
     * @exception InterruptedException
     *            Another thread has interrupted this thread.
     */
    public synchronized boolean grabPixels(long ms)
	throws InterruptedException
    {
	if ((flags & DONEBITS) != 0) {
	    return (flags & GRABBEDBITS) != 0;
	}
	long end = ms + System.currentTimeMillis();
	if (!grabbing) {
	    grabbing = true;
	    flags &= ~(ImageObserver.ABORT);
	    producer.startProduction(this);
	}
	while (grabbing) {
	    long timeout;
	    if (ms == 0) {
		timeout = 0;
	    } else {
		timeout = end - System.currentTimeMillis();
		if (timeout <= 0) {
		    break;
		}
	    }
	    wait(timeout);
	}
	return (flags & GRABBEDBITS) != 0;
    }

    /**
     * Return the status of the pixels.  The ImageObserver flags
     * representing the available pixel information are returned.
     * @see ImageObserver
     * @return the bitwise OR of all relevant ImageObserver flags
     */
    public synchronized int getStatus() {
	return flags;
    }

    /**
     * Get the width of the pixel buffer (after adjusting for image width).
     * If no width was specified for the rectangle of pixels to grab then
     * then this information will only be available after the image has
     * delivered the dimensions.
     * @return the final width used for the pixel buffer or -1 if the width
     * is not yet known
     * @see #getStatus
     */
    public synchronized int getWidth() {
	return (dstW < 0) ? -1 : dstW;
    }

    /**
     * Get the height of the pixel buffer (after adjusting for image height).
     * If no width was specified for the rectangle of pixels to grab then
     * then this information will only be available after the image has
     * delivered the dimensions.
     * @return the final height used for the pixel buffer or -1 if the height
     * is not yet known
     * @see #getStatus
     */
    public synchronized int getHeight() {
	return (dstH < 0) ? -1 : dstH;
    }

    /**
     * Get the pixel buffer.  If the PixelGrabber was not constructed
     * with an explicit pixel buffer to hold the pixels then this method
     * will return null until the size and format of the image data is
     * known.
     * Since the PixelGrabber may fall back on accumulating the data
     * in the default RGB ColorModel at any time if the source image
     * uses more than one ColorModel to deliver the data, the array
     * object returned by this method may change over time until the
     * image grab is complete.
     * @return either a byte array or an int array
     * @see #getStatus
     */
    public synchronized Object getPixels() {
	return (bytePixels == null)
	    ? ((Object) intPixels)
	    : ((Object) bytePixels);
    }

    /**
     * Get the ColorModel for the pixels stored in the array.  If the
     * PixelGrabber was constructed with an explicit pixel buffer then
     * this method will always return the default RGB ColorModel,
     * otherwise it may return null until the ColorModel used by the
     * ImageProducer is known.
     * Since the PixelGrabber may fall back on accumulating the data
     * in the default RGB ColorModel at any time if the source image
     * uses more than one ColorModel to deliver the data, the ColorModel
     * object returned by this method may change over time until the
     * image grab is complete.
     * @return the ColorModel object used for storing the pixels
     * @see #getStatus
     * @see ColorModel#getRGBdefault
     */
    public synchronized ColorModel getColorModel() {
	return imageModel;
    }

    /**
     * The setDimensions method is part of the ImageConsumer API which
     * this class must implement to retrieve the pixels.
     */
    public void setDimensions(int width, int height) {
	if (dstW < 0) {
	    dstW = width - dstX;
	}
	if (dstH < 0) {
	    dstH = height - dstY;
	}
	if (dstW <= 0 || dstH <= 0) {
	    imageComplete(STATICIMAGEDONE);
	} else if (intPixels == null &&
		   imageModel == ColorModel.getRGBdefault()) {
            if (allowGrayImages == false) {
                intPixels = new int[dstW * dstH];
            }
	    dstScan = dstW;
	    dstOff = 0;
	}
	flags |= (ImageObserver.WIDTH | ImageObserver.HEIGHT);
    }

    /**
     * The setHints method is part of the ImageConsumer API which
     * this class must implement to retrieve the pixels.
     */
    public void setHints(int hints) {
	return;
    }

    /**
     * The setProperties method is part of the ImageConsumer API which
     * this class must implement to retrieve the pixels.
     */
    public void setProperties(Hashtable props) {
	return;
    }

    /**
     * The setColorModel method is part of the ImageConsumer API which
     * this class must implement to retrieve the pixels.
     */
    public void setColorModel(ColorModel model) {
	return;
    }

    private void convertToRGB() {
	int size = dstW * dstH;
	int newpixels[] = new int[size];
	if (bytePixels != null) {
	    for (int i = 0; i < size; i++) {
		newpixels[i] = imageModel.getRGB(bytePixels[i] & 0xff);
	    }

	} else if (intPixels != null) {
	    for (int i = 0; i < size; i++) {
		newpixels[i] = imageModel.getRGB(intPixels[i]);
	    }
	}
	bytePixels = null;
	intPixels = newpixels;
	dstScan = dstW;
	dstOff = 0;
	imageModel = ColorModel.getRGBdefault();
    }

    /**
     * The setPixels method is part of the ImageConsumer API which
     * this class must implement to retrieve the pixels.
     */
    public void setPixels(int srcX, int srcY, int srcW, int srcH,
			  ColorModel model,
			  byte pixels[], int srcOff, int srcScan) {
	if (srcY < dstY) {
	    int diff = dstY - srcY;
	    if (diff >= srcH) {
		return;
	    }
	    srcOff += srcScan * diff;
	    srcY += diff;
	    srcH -= diff;
	}
	if (srcY + srcH > dstY + dstH) {
	    srcH = (dstY + dstH) - srcY;
	    if (srcH <= 0) {
		return;
	    }
	}
	if (srcX < dstX) {
	    int diff = dstX - srcX;
	    if (diff >= srcW) {
		return;
	    }
	    srcOff += diff;
	    srcX += diff;
	    srcW -= diff;
	}
	if (srcX + srcW > dstX + dstW) {
	    srcW = (dstX + dstW) - srcX;
	    if (srcW <= 0) {
		return;
	    }
	}
	int dstPtr = dstOff + (srcY - dstY) * dstScan + (srcX - dstX);
	if (intPixels == null) {
	    if (bytePixels == null) {
		bytePixels = new byte[dstW * dstH];
		dstScan = dstW;
		dstOff = 0;
		imageModel = model;
            } else if (imageModel != model) {
		convertToRGB();
	    }
	    if (bytePixels != null) {
		for (int h = srcH; h > 0; h--) {
		    System.arraycopy(pixels, srcOff, bytePixels, dstPtr, srcW);
		    srcOff += srcScan;
		    dstPtr += dstScan;
		}
	    }
	}
	if (intPixels != null) {
	    int dstRem = dstScan - srcW;
	    int srcRem = srcScan - srcW;
	    for (int h = srcH; h > 0; h--) {
		for (int w = srcW; w > 0; w--) {
		    intPixels[dstPtr++] = model.getRGB(pixels[srcOff++]&0xff);
		}
		srcOff += srcRem;
		dstPtr += dstRem;
	    }
	}
	flags |= ImageObserver.SOMEBITS;
    }

    /**
     * The setPixels method is part of the ImageConsumer API which
     * this class must implement to retrieve the pixels.
     */
    public void setPixels(int srcX, int srcY, int srcW, int srcH,
			  ColorModel model,
			  int pixels[], int srcOff, int srcScan) {
	if (srcY < dstY) {
	    int diff = dstY - srcY;
	    if (diff >= srcH) {
		return;
	    }
	    srcOff += srcScan * diff;
	    srcY += diff;
	    srcH -= diff;
	}
	if (srcY + srcH > dstY + dstH) {
	    srcH = (dstY + dstH) - srcY;
	    if (srcH <= 0) {
		return;
	    }
	}
	if (srcX < dstX) {
	    int diff = dstX - srcX;
	    if (diff >= srcW) {
		return;
	    }
	    srcOff += diff;
	    srcX += diff;
	    srcW -= diff;
	}
	if (srcX + srcW > dstX + dstW) {
	    srcW = (dstX + dstW) - srcX;
	    if (srcW <= 0) {
		return;
	    }
	}
	if (intPixels == null) {
	    if (bytePixels == null) {
		intPixels = new int[dstW * dstH];
		dstScan = dstW;
		dstOff = 0;
		imageModel = model;
            } else {
		convertToRGB();
	    }
	}
	int dstPtr = dstOff + (srcY - dstY) * dstScan + (srcX - dstX);
	if (imageModel == model) {
	    for (int h = srcH; h > 0; h--) {
		System.arraycopy(pixels, srcOff, intPixels, dstPtr, srcW);
		srcOff += srcScan;
		dstPtr += dstScan;
	    }
	} else {
	    if (imageModel != ColorModel.getRGBdefault()) {
		convertToRGB();
	    }
	    int dstRem = dstScan - srcW;
	    int srcRem = srcScan - srcW;
	    for (int h = srcH; h > 0; h--) {
		for (int w = srcW; w > 0; w--) {
		    intPixels[dstPtr++] = model.getRGB(pixels[srcOff++]);
		}
		srcOff += srcRem;
		dstPtr += dstRem;
	    }
	}
	flags |= ImageObserver.SOMEBITS;
    }

    /**
     * The imageComplete method is part of the ImageConsumer API which
     * this class must implement to retrieve the pixels.
     */
    public synchronized void imageComplete(int status) {
	grabbing = false;
	switch (status) {
	default:
	case IMAGEERROR:
	    flags |= ImageObserver.ERROR | ImageObserver.ABORT;
	    break;
	case IMAGEABORTED:
	    flags |= ImageObserver.ABORT;
	    break;
	case STATICIMAGEDONE:
	    flags |= ImageObserver.ALLBITS;
	    break;
	case SINGLEFRAMEDONE:
	    flags |= ImageObserver.FRAMEBITS;
	    break;
	}
	producer.removeConsumer(this);
	notifyAll();
    }

    public BufferedImage getImage() {
        if ((flags & STATICIMAGEDONE) != 0) {
            if (intPixels == null) {
                convertToRGB();
                if (intPixels == null) {
                    return null;
                }
            }
            // Create a buffered image from this
                // REMIND: What to do about ColorModel??
                //Tile t = new IntegerComponentTile(0, 0, dstW, dstH, 0, dstW,
                //                                  intPixels);
                DataBufferInt dbi = new DataBufferInt(intPixels,
                                                      intPixels.length);
                int bandMasks[] = {0x00ff0000,0x0000ff00,0x000000ff,0xff000000};
                WritableRaster wr =
                     Raster.createPackedRaster(dbi, dstW, dstH,
                                               dstW, bandMasks, null);
                return new
                    BufferedImage (imageModel, wr,
                                   imageModel.isAlphaPremultiplied(), null);
        }
        else return null;
    }

    public boolean isCModelGray(byte[] gray) {
        if (! (imageModel instanceof IndexColorModel)) {
            return false;
        }
        IndexColorModel icm = (IndexColorModel) imageModel;
        // Need to grab the arrays and test
        byte[] r = gray;
        byte[] g = new byte[256];
        byte[] b = new byte[256];

        icm.getReds(r);
        icm.getGreens(g);
        icm.getBlues(b);
        for (int i=0; i < icm.getMapSize(); i++) {
            if (r[i] != g[i] || r[i] != b[i]) return false;
        }

        return true;
    }

    public BufferedImage getByteImage() {
        if ((flags & STATICIMAGEDONE) != 0) {
            if (allowGrayImages) {
                byte[] gray = new byte[256];
                if (isCModelGray(gray)) {
                    if (bytePixels != null) {
                        byte[] data = new byte[dstW*dstH];
                        for (int i=0; i < data.length; i++) {
                            data[i] = gray[bytePixels[i]&0xff];
                        }
                        DataBufferByte dbb = new DataBufferByte(data,
                                                                data.length);
                        int bandOffsets[] = {0};
                        WritableRaster wr =
                          Raster.createInterleavedRaster(dbb,
                                                       dstW,dstH,dstW,1,
                                                       bandOffsets,null);

                        ColorSpace cs =
                            ColorSpace.getInstance(ColorSpace.CS_GRAY);
                        int[] bits = {8};

                        return new
                            BufferedImage(new ComponentColorModel(cs, bits,
                                                   false, false,
                                                   Transparency.OPAQUE,
                                                   DataBuffer.TYPE_BYTE),
                                          wr, false, null);
                    }
                }
            }

            if (bytePixels != null) {
                convertToRGB();
            }

            // Create a buffered image from this
            if (intPixels != null) {
                int[] cOffs = {3,2,1,0};// REMIND: SHould be 3,2,1,0
                byte[] data = new byte[dstW*dstH*4];
                int[] nBits = {8,8,8,8};
//                Tile tile = new ByteComponentTile(dstW, dstH, 4, cOffs, 4,
//                                                  dstW*4, data);

                // Copy the data
                int off = 0;
                for (int i=0; i < intPixels.length; i++) {
                    data[off]   = (byte) ((intPixels[i]>>24)&0xff);
                    data[off+1] = (byte) ((intPixels[i])    &0xff);
                    data[off+2] = (byte) ((intPixels[i]>>8) &0xff);
                    data[off+3] = (byte) ((intPixels[i]>>16)&0xff);
                    off += 4;
                }

                DataBufferByte dbb = new DataBufferByte(data, data.length);

                WritableRaster wr =
                   Raster.createInterleavedRaster(dbb,
                                                dstW,dstH,
                                                dstW,4,cOffs,
                                                null);

                ColorSpace cs = ColorSpace.getInstance(ColorSpace.CS_sRGB);
                // REMIND: What to do about ColorModel??

                return new
                    BufferedImage(new ComponentColorModel(cs, nBits, true,
                                                          false,
                                                   Transparency.TRANSLUCENT,
                                                    DataBuffer.TYPE_BYTE),
                                  wr, false, null);
            }
        }

        return null;
    }
    /**
     * DEPRECATED:  Replaced by getStatus().
     */
    public synchronized int status() {
	return flags;
    }
}
