/*
 * @(#)alphatest.c	1.2 98/04/13
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Performance and correctness testing harness for various alpha
 * composition algorithms.
 */

#include <stdlib.h>
#include <stdio.h>
#include <dlfcn.h>
#include <time.h>
#include <ctype.h>

#include "alphatest.h"

/*
 * Declarations of specific data for program operation.
 */

typedef struct {
    double	errsum;
    double	numcases;
    float	error;
    int		SA;
    int		SC;
    int		SP;
    int		DA;
    int		DC;
    int		DP;
    int		PA;
    int		EA256;
    int		resultA;
    int		resultC;
    float	refA;
    float	refC;
    char	*funcName;
} TestResult;

typedef struct {
    TestResult	alphaResults;
    TestResult	colorResults;
    TestResult	premulResults;
} TestResults;

typedef struct {
    double diffsum;
    double diffnum;
    float worstdiff;
    int worstSA;
    int worstSC;
    int worstSP;
    int worstDA;
    int worstDC;
    int worstDP;
    int worstPA;
    int worstEA256;
    int worstA;
    int worstC;
    float refA;
    float refC;
    char *funcName;
} XTestResults;

typedef struct {
    clock_t time;
    char *funcName;
} BenchResults;

typedef struct {
    int totalDim;
    int premulDim;
    unsigned char *srcBuf;
    unsigned char *dstBuf;
    unsigned char *resBuf;
    unsigned char **pathBufs;
} Buffers;

static void setuparrays(TestValues *vals, Buffers *bufs);
static void test(CompositeSet *set, TestValues *vals, Buffers *bufs);
static void testeqn(char *eqn, CompositeSet *set, TestValues *vals);
static void bench(CompositeSet *set, TestValues *vals, Buffers *bufs);

int verbose;
char *progname;

void
usage()
{
    int i;
    fprintf(stderr, "usage: %s [...options - see below...]\n", progname);
    fprintf(stderr, "    [-v[erbose] (verbose level +1 for each char)]\n");
    fprintf(stderr, "    [-set=<ComposSet symbol name>]\n");
    for (i = 0; i < NumCompSets; i++) {
	fprintf(stderr, "\t'%s' - %s\n",
		CompSetList[i].name, CompSetList[i].description);
    }
    fprintf(stderr, "    [-vals=<TestValues symbol name>]\n");
    for (i = 0; i < NumTestVals; i++) {
	fprintf(stderr, "\t'%s' - %s\n",
		TestValList[i].name, TestValList[i].description);
    }
    fprintf(stderr, "    [-t[est] (perform an accuracy test run)]\n");
    fprintf(stderr, "    [-b[ench] (perform a benchmark run)]\n");
    fprintf(stderr, "    [-eqn=<equation> (test a specific case)]\n");
    exit(1);
}

int
strcmpnc(char *s1, char *s2)
{
    int c1, c2;
    do {
	c1 = tolower(*s1++);
	c2 = tolower(*s2++);
	if (c1 != c2) {
	    return (c1 - c2);
	}
    } while (c1);

    return 0;
}

int
main(int argc, char **argv)
{
    CompositeSet *testset = CompSetList[0].set;
    TestValues *testvals = TestValList[0].vals;
    char *eqn = 0;
    int doTest = True;
    Buffers bufs;
    int i;

    progname = argv[0];
    verbose = 0;

    while (argc > 1) {
	if (strncmp(argv[1], "-set=", 5) == 0) {
	    char *sym = argv[1] + 5;
	    if (*sym == 0) {
		usage();
	    }
	    for (i = 0; i < NumCompSets; i++) {
		if (strcmp(sym, CompSetList[i].name) == 0) {
		    break;
		}
	    }
	    if (i < NumCompSets) {
		testset = CompSetList[i].set;
	    } else {
		fprintf(stderr, "Cannot find symbol '%s'\n", sym);
		usage();
	    }
	} else if (strncmp(argv[1], "-vals=", 6) == 0) {
	    char *sym = argv[1] + 6;
	    if (*sym == 0) {
		usage();
	    }
	    for (i = 0; i < NumTestVals; i++) {
		if (strcmp(sym, TestValList[i].name) == 0) {
		    break;
		}
	    }
	    if (i < NumTestVals) {
		testvals = TestValList[i].vals;
	    } else {
		fprintf(stderr, "Cannot find symbol '%s'\n", sym);
		usage();
	    }
	} else if (strncmp(argv[1], "-eqn=", 5) == 0) {
	    eqn = argv[1]+5;
	} else if (strncmp(argv[1], "-t", 2) == 0) {
	    doTest = True;
	} else if (strncmp(argv[1], "-b", 2) == 0) {
	    doTest = False;
	} else if (strncmp(argv[1], "-v", 2) == 0) {
	    for (verbose = 1; verbose < 8; verbose++) {
		if (strncmp(argv[1], "-verbose", verbose + 2) != 0) {
		    break;
		}
	    }
	} else {
	    usage();
	}
	argv++;
	argc--;
    }

    if (eqn) {
	testeqn(eqn, testset, testvals);
	exit(0);
    }

    setuparrays(testvals, &bufs);
    if (doTest) {
	test(testset, testvals, &bufs);
    } else {
	bench(testset, testvals, &bufs);
    }
}

static void
setuparrays(TestValues *vals, Buffers *bufs)
{
    int numvals255 = vals->numvals255;
    int *vals255 = vals->vals255;
    int dim, samples;
    unsigned char *sampBuf;
    int lo, hi;
    int a, c, A, C;

    /*
     * Square regions of samples are created for all possible
     * combinations of alpha and color samples.  The dimension of
     * each axis of this buffer is the total number of "255-based"
     * samples squared (total alpha samples times total color samples).
     */
    dim = numvals255 * numvals255;

    /*
     * The total number of samples is dim x dim.
     */
    samples = dim * dim;

    bufs->dstBuf = malloc(samples * 2 * sizeof(unsigned char));
    bufs->srcBuf = malloc(samples * 2 * sizeof(unsigned char));
    bufs->resBuf = malloc(samples * 2 * sizeof(unsigned char));
    bufs->pathBufs = malloc(numvals255 * sizeof(unsigned char *));

    /*
     * Store samples that meet premul constraint at the beginning of
     * the samples array and samples that don't meet the constraints
     * at the end of the array.  This way, the data that can be used
     * by the premultiplied cases is packed into the upper left corner
     * of the square.
     */
    sampBuf = malloc(dim * 2);
    lo = 0;
    hi = dim * 2;
    for (a = 0; a < numvals255; a++) {
	A = vals255[a];
	for (c = 0; c < numvals255; c++) {
	    C = vals255[c];
	    if (C > A) {
		sampBuf[--hi] = C;
		sampBuf[--hi] = A;
	    } else {
		sampBuf[lo++] = A;
		sampBuf[lo++] = C;
	    }
	}
    }
    if (lo != hi) {
	fprintf(stderr,
		"Premul cutoff assertion fails (lo = %d, hi = %d)!\n",
		lo, hi);
    }

    bufs->totalDim = dim;
    bufs->premulDim = lo / 2;

    lo = 0;
    for (a = 0; a < dim * 2; a += 2) {
	for (c = 0; c < dim * 2; c += 2) {
	    bufs->srcBuf[lo] = sampBuf[a];
	    bufs->dstBuf[lo] = sampBuf[c];
	    lo++;
	    bufs->srcBuf[lo] = sampBuf[a + 1];
	    bufs->dstBuf[lo] = sampBuf[c + 1];
	    lo++;
	}
    }

    /*
     * Finally create a constant "path alpha" buffer for each potential
     * alpha value.
     */
    for (a = 0; a < numvals255; a++) {
	bufs->pathBufs[a] = malloc(dim);
	memset(bufs->pathBufs[a], vals255[a], dim);
    }
}

static void
combinetest(TestResult *parent, TestResult *child)
{
    parent->errsum += child->errsum;
    parent->numcases += child->numcases;
    if (parent->error < child->error) {
	parent->error		= child->error;
	parent->SA		= child->SA;
	parent->SC		= child->SC;
	parent->SP		= child->SP;
	parent->DA		= child->DA;
	parent->DC		= child->DC;
	parent->DP		= child->DP;
	parent->PA		= child->PA;
	parent->EA256		= child->EA256;
	parent->resultA		= child->resultA;
	parent->resultC		= child->resultC;
	parent->refA		= child->refA;
	parent->refC		= child->refC;
	parent->funcName	= child->funcName;
    }
}

static void
combinetests(TestResults *parent, TestResults *child)
{
    combinetest(&parent->alphaResults, &child->alphaResults);
    combinetest(&parent->colorResults, &child->colorResults);
    combinetest(&parent->premulResults, &child->premulResults);
}

static void
reporttest(FILE *out, TestResult *res, char *label, char *type, int showavg)
{
    fprintf(out, label, type, res->funcName);
    fprintf(out, ":\n    %s(%d, %d, %s, %d, %d, %s, %d, %d)\n",
	    res->funcName,
	    res->SA, res->SC, (res->SP ? "True" : "False"),
	    res->DA, res->DC, (res->DP ? "True" : "False"),
	    res->PA, res->EA256);
    fprintf(out,
	    "\tResult    (A = %4d   , C = %4d   , A*C = %7.2f) out of 255\n",
	    res->resultA, res->resultC,
	    (res->resultA * res->resultC) / 255.0f);
    fprintf(out,
	    "\tReference (A = %7.2f, C = %7.2f, A*C = %7.2f) out of 255\n",
	    res->refA, res->refC, res->refA * res->refC / 255.0f);
    fprintf(out,
	    "\tError     (A = %7.2f, C = %7.2f, A*C = %7.2f) out of 255\n",
	    res->resultA - res->refA, res->resultC - res->refC,
	    (res->resultA * res->resultC - res->refA * res->refC) / 255.0f);
    if (showavg) {
	fprintf(out, "\tMean %serror: (%.2f / %.0f) == %.4f out of 255\n",
		type, res->errsum, res->numcases,
		res->errsum / res->numcases);
    }
}

static void
reporttests(FILE *out, TestResults *res, char *label, int showavg)
{
    if (verbose > 1) {
	reporttest(out, &res->alphaResults, label, "alpha ", showavg);
	reporttest(out, &res->colorResults, label, "color ", showavg);
	reporttest(out, &res->premulResults, label, "premul ", showavg);
	fprintf(out, "\n");
    } else {
	TestResult combined;
	memset(&combined, 0, sizeof(combined));
	combined.error = -1.0f;
	combinetest(&combined, &res->alphaResults);
	combinetest(&combined, &res->premulResults);
	reporttest(out, &combined, label, "", showavg);
    }
}

static void
puttestresults(float error, TestResult *res,
	       int sA, int sC, int sP,
	       int dA, int dC, int dP,
	       int pA, int eA256,
	       int iResA, int iResC, float fResA, float fResC)
{
    res->errsum += error;
    res->numcases++;

    if (error > res->error) {
	res->error	= error;
	res->SA		= sA;
	res->SC		= sC;
	res->SP		= sP;
	res->DA		= dA;
	res->DC		= dC;
	res->DP		= dP;
	res->PA		= pA;
	res->EA256	= eA256;
	res->resultA	= iResA;
	res->resultC	= iResC;
	res->refA	= fResA;
	res->refC	= fResC;
	if (verbose > 2) {
	    reporttest(stderr, res, "New worst case", "", False);
	}
    }
}

static void
rectcopy(unsigned char *dst, unsigned char *src, int w, int h, int scan)
{
    int i;

    scan -= w;
    while (--h >= 0) {
	for (i = 0; i < w; i++) {
	    *dst++ = *src++;
	}
	dst += scan;
	src += scan;
    }
}

static void
testone(CompositeFunc *func, ReferenceFunc *reffunc,
	Buffers *bufs, TestResults *res,
	int sP, int dP, int pA, int eA256)
{
    int iResA;
    int iResC;
    float fResA;
    float fResC;
    float diffA;
    float diffC;
    float diffmax;
    int w, h, dim;
    unsigned char *srcBuf = bufs->srcBuf;
    unsigned char *dstBuf = bufs->dstBuf;
    unsigned char *resBuf = bufs->resBuf;
    unsigned char *pathBuf = bufs->pathBufs[pA];
    int i, j, off;

    w = dP ? bufs->premulDim : bufs->totalDim;
    h = sP ? bufs->premulDim : bufs->totalDim;
    dim = bufs->totalDim;

    /*
     * Copy the dest buf to the result buf in case the algorithm
     * optimizes out the memory copies in the "no change" cases.
     */
    rectcopy(bufs->resBuf, bufs->dstBuf, w*2, h, dim*2);

    func(srcBuf, dim * 2, sP,
	 dstBuf, dim * 2, dP,
	 pathBuf, 0, eA256,
	 resBuf, w, h);

    off = 0;
    for (i = 0; i < h; i++) {
	for (j = 0; j < w; j++) {
	    int sA, sC, dA, dC, pA;

	    sA = srcBuf[off];
	    dA = dstBuf[off];
	    iResA = resBuf[off];
	    off++;
	    sC = srcBuf[off];
	    dC = dstBuf[off];
	    iResC = resBuf[off];
	    off++;
	    pA = *pathBuf++;

	    if (dP && iResC > iResA && verbose > 1) {
		fprintf(stderr, "Premultiplied result violates (C < A)!\n");
		fprintf(stderr, "    %s(%d, %d, %s, %d, %d, %s, %d, %d)",
			res->alphaResults.funcName,
			sA, sC, (sP ? "True" : "False"),
			dA, dC, (dP ? "True" : "False"),
			pA, eA256);
		fprintf(stderr, " == (A=%d, C=%d)\n", iResA, iResC);
	    }

	    reffunc(sA, sC, sP, dA, dC, dP, pA, eA256, &fResA, &fResC);

	    fResA *= 255.0f;
	    fResC *= 255.0f;

	    diffA = iResA - fResA;
	    if (diffA < 0) diffA = -diffA;
	    puttestresults(diffA, &res->alphaResults,
			   sA, sC, sP, dA, dC, dP, pA, eA256,
			   iResA, iResC, fResA, fResC);

	    diffC = iResC - fResC;
	    if (diffC < 0) diffC = -diffC;
	    puttestresults(diffC, &res->colorResults,
			   sA, sC, sP, dA, dC, dP, pA, eA256,
			   iResA, iResC, fResA, fResC);

	    diffC = (iResC * iResA / 255.0f) - (fResA * fResC / 255.0f);
	    if (diffC < 0) diffC = -diffC;
	    puttestresults(diffC, &res->premulResults,
			   sA, sC, sP, dA, dC, dP, pA, eA256,
			   iResA, iResC, fResA, fResC);
	}
	off += dim * 2 - w * 2;
	pathBuf -= w;
    }
}

static void
testallvals(CompositeFunc *func, ReferenceFunc *reffunc,
	    int srcPremul, int dstPremul,
	    TestValues *vals, Buffers *bufs, TestResults *res)
{
    int eA256, pA, sA, sC;
    int *vals255 = vals->vals255;
    int numvals255 = vals->numvals255 - 1;
    int *vals256 = vals->vals256;
    int numvals256 = vals->numvals256 - 1;

    for (eA256 = numvals256; eA256 >= 0; eA256--) {
	for (pA = numvals255; pA >= 0; pA--) {
	    testone(func, reffunc, bufs, res,
		    srcPremul, dstPremul, pA, vals256[eA256]);
	}
    }
}

static void
testonecomp(CompositeFunc *func, ReferenceFunc *reffunc, char *funcname,
	    TestValues *vals, Buffers *bufs, TestResults *parentres)
{
    TestResults res;

    memset(&res, 0, sizeof(res));
    res.alphaResults.funcName = funcname;
    res.colorResults.funcName = funcname;
    res.premulResults.funcName = funcname;
    testallvals(func, reffunc, False, False, vals, bufs, &res);
    testallvals(func, reffunc, False,  True, vals, bufs, &res);
    testallvals(func, reffunc,  True, False, vals, bufs, &res);
    testallvals(func, reffunc,  True,  True, vals, bufs, &res);
    if (verbose) {
	reporttests(stdout, &res, "Worst %serror for %s", True);
    }
    combinetests(parentres, &res);
}

static void
badeqn(char *eqn, char *why)
{
    fprintf(stderr, "%s: %s for equation:\n", progname, why);
    fprintf(stderr, "\t%s\n", eqn);
    usage();
}

static char *FUNCPROTOSTR =
    "%10[A-Za-z](%d, %d, %10[A-Za-z0-9], %d, %d, %10[A-Za-z0-9], %d, %d)";

static void
testeqn(char *eqn, CompositeSet *set, TestValues *vals)
{
    char funcbuf[11], *funcname;
    int sA, sC, sP;
    int dA, dC, dP;
    int pA, eA256;
    char SPstr[11], DPstr[11];
    unsigned char buf[5];	/* sA sC dA dC pA */
    CompositeFunc *setfunc;
    ReferenceFunc *reffunc;
    int iResA, iResC;
    float fResA, fResC;
    float diffA, diffC;
    TestResults res;

    if (sscanf(eqn, FUNCPROTOSTR,
	       &funcbuf,
	       &sA, &sC, &SPstr,
	       &dA, &dC, &DPstr,
	       &pA, &eA256) != 9)
    {
	badeqn(eqn, "sscanf failed");
    }

    if (strcmpnc(SPstr, "True") == 0) {
	sP = True;
    } else if (strcmpnc(SPstr, "False") == 0) {
	sP = False;
    } else if (sscanf(SPstr, "%d", &sP) != 1) {	
	badeqn(eqn, "unrecognized srcPremul string");
    }

    if (strcmpnc(DPstr, "True") == 0) {
	dP = True;
    } else if (strcmpnc(DPstr, "False") == 0) {
	dP = False;
    } else if (sscanf(DPstr, "%d", &dP) != 1) {	
	badeqn(eqn, "unrecognized dstPremul string");
    }

    buf[0] = sA;
    buf[1] = sC;
    buf[2] = dA;
    buf[3] = dC;
    buf[4] = pA;

    if (strcmpnc(funcbuf, "Clear") == 0) {
	funcname = "Clear";
	setfunc = set->ClearFunc;
	reffunc = ClearReference;
    } else if (strcmpnc(funcbuf, "Src") == 0) {
	funcname = "Src";
	setfunc = set->SrcFunc;
	reffunc = SrcReference;
    } else if (strcmpnc(funcbuf, "SrcOver") == 0) {
	funcname = "SrcOver";
	setfunc = set->SrcOverFunc;
	reffunc = SrcOverReference;
    } else if (strcmpnc(funcbuf, "DstOver") == 0) {
	funcname = "DstOver";
	setfunc = set->DstOverFunc;
	reffunc = DstOverReference;
    } else if (strcmpnc(funcbuf, "SrcIn") == 0) {
	funcname = "SrcIn";
	setfunc = set->SrcInFunc;
	reffunc = SrcInReference;
    } else if (strcmpnc(funcbuf, "DstIn") == 0) {
	funcname = "DstIn";
	setfunc = set->DstInFunc;
	reffunc = DstInReference;
    } else if (strcmpnc(funcbuf, "SrcOut") == 0) {
	funcname = "SrcOut";
	setfunc = set->SrcOutFunc;
	reffunc = SrcOutReference;
    } else if (strcmpnc(funcbuf, "DstOut") == 0) {
	funcname = "DstOut";
	setfunc = set->DstOutFunc;
	reffunc = DstOutReference;
    } else {
	badeqn(eqn, "unrecognized function name");
    }

    memset(&res, 0, sizeof(res));
    res.alphaResults.funcName = funcname;
    res.alphaResults.error = -1.0f;
    res.colorResults.funcName = funcname;
    res.colorResults.error = -1.0f;
    res.premulResults.funcName = funcname;
    res.premulResults.error = -1.0f;

    setfunc(&buf[0], 2, sP,
	    &buf[2], 2, dP,
	    &buf[4], 1, eA256,
	    &buf[2], 1, 1);
    iResA = buf[2];
    iResC = buf[3];

    if (dP && iResC > iResA) {
	fprintf(stderr, "Premultiplied result violates (C < A)!\n");
	fprintf(stderr, "    %s(%d, %d, %s, %d, %d, %s, %d, %d)",
		funcname,
		sA, sC, (sP ? "True" : "False"),
		dA, dC, (dP ? "True" : "False"),
		pA, eA256);
	fprintf(stderr, " == (A=%d, C=%d)\n", iResA, iResC);
    }

    reffunc(sA, sC, sP, dA, dC, dP, pA, eA256, &fResA, &fResC);

    fResA *= 255.0f;
    fResC *= 255.0f;

    diffA = iResA - fResA;
    if (diffA < 0) diffA = -diffA;
    puttestresults(diffA, &res.alphaResults,
		   sA, sC, sP, dA, dC, dP, pA, eA256,
		   iResA, iResC, fResA, fResC);

    diffC = iResC - fResC;
    if (diffC < 0) diffC = -diffC;
    puttestresults(diffC, &res.colorResults,
		   sA, sC, sP, dA, dC, dP, pA, eA256,
		   iResA, iResC, fResA, fResC);

    diffC = (iResC * iResA / 255.0f) - (fResA * fResC / 255.0f);
    if (diffC < 0) diffC = -diffC;
    puttestresults(diffC, &res.premulResults,
		   sA, sC, sP, dA, dC, dP, pA, eA256,
		   iResA, iResC, fResA, fResC);

    reporttests(stdout, &res, "Resulting %serror", True);
}

static void
test(CompositeSet *set, TestValues *vals, Buffers *bufs)
{
    TestResults res;

    memset(&res, 0, sizeof(res));
    testonecomp(set->ClearFunc,   ClearReference,   "Clear",
		vals, bufs, &res);
    testonecomp(set->SrcFunc,     SrcReference,     "Src",
		vals, bufs, &res);
    testonecomp(set->SrcOverFunc, SrcOverReference, "SrcOver",
		vals, bufs, &res);
    testonecomp(set->DstOverFunc, DstOverReference, "DstOver",
		vals, bufs, &res);
    testonecomp(set->SrcInFunc,   SrcInReference,   "SrcIn",
		vals, bufs, &res);
    testonecomp(set->DstInFunc,   DstInReference,   "DstIn",
		vals, bufs, &res);
    testonecomp(set->SrcOutFunc,  SrcOutReference,  "SrcOut",
		vals, bufs, &res);
    testonecomp(set->DstOutFunc,  DstOutReference,  "DstOut",
		vals, bufs, &res);
    fprintf(stdout, "\n");
    reporttests(stdout, &res, "Overall worst %serror", True);
}

static void
reportbench(FILE *out, BenchResults *res)
{
    fprintf(out, "Results for %s: %f seconds\n",
	    res->funcName, res->time / ((double) CLOCKS_PER_SEC));
}

static void
combinebench(BenchResults *parent, BenchResults *child)
{
    parent->time += child->time;
}

static void
benchallvals(CompositeFunc *func, int srcPremul, int dstPremul,
	     TestValues *vals, Buffers *bufs, BenchResults *res)
{
    int eA256, pA, sA, sC;
    int numvals255 = vals->numvals255 - 1;
    int *vals256 = vals->vals256;
    int numvals256 = vals->numvals256 - 1;
    int w, h, dim;
    clock_t start, end;

    w = dstPremul ? bufs->premulDim : bufs->totalDim;
    h = srcPremul ? bufs->premulDim : bufs->totalDim;
    dim = bufs->totalDim;

    start = clock();
    for (eA256 = numvals256; eA256 >= 0; eA256--) {
	for (pA = numvals255; pA >= 0; pA--) {
	    func(bufs->srcBuf, dim * 2, srcPremul,
		 bufs->dstBuf, dim * 2, dstPremul,
		 bufs->pathBufs[pA], 0, vals256[eA256],
		 bufs->resBuf, w, h);
	}
    }
    end = clock();
    res->time += (end - start);
}

static void
benchonecomp(CompositeFunc *func, char *funcname,
	     TestValues *vals, Buffers *bufs, BenchResults *parentres)
{
    BenchResults res;

    memset(&res, 0, sizeof(res));
    res.funcName = funcname;
    benchallvals(func, False, False, vals, bufs, &res);
    benchallvals(func, False,  True, vals, bufs, &res);
    benchallvals(func,  True, False, vals, bufs, &res);
    benchallvals(func,  True,  True, vals, bufs, &res);
    if (verbose) {
	reportbench(stdout, &res);
    }
    combinebench(parentres, &res);
}

static void
bench(CompositeSet *set, TestValues *vals, Buffers *bufs)
{
    BenchResults res;

    memset(&res, 0, sizeof(res));
    res.funcName = "all operations";
    benchonecomp(set->ClearFunc,   "Clear",   vals, bufs, &res);
    benchonecomp(set->SrcFunc,     "Src",     vals, bufs, &res);
    benchonecomp(set->SrcOverFunc, "SrcOver", vals, bufs, &res);
    benchonecomp(set->DstOverFunc, "DstOver", vals, bufs, &res);
    benchonecomp(set->SrcInFunc,   "SrcIn",   vals, bufs, &res);
    benchonecomp(set->DstInFunc,   "DstIn",   vals, bufs, &res);
    benchonecomp(set->SrcOutFunc,  "SrcOut",  vals, bufs, &res);
    benchonecomp(set->DstOutFunc,  "DstOut",  vals, bufs, &res);
    fprintf(stdout, "\n");
    reportbench(stdout, &res);
}
