/*
 * @(#)intset2.c	1.1 98/04/10
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * A CompositeSet that calculates the results using 12-bit integer math
 * in an attempt to preserve greater accuracy than the 8-bit integer set.
 * No intensive optimizations are made here, the answer is calculated
 * simply and straightforwardly.
 */

#include "alphatest.h"

#define NORM(a)			(((a) << 4) + ((a) >> 4))
#define DENORM(a)		((a) >> 4)
#define _RENORM(a)		(((a) + (1 << 11)) >> 12)
#define COMBINE(a, b)		_RENORM((a) * (b))
#define UNCOMBINE(a, b)		((((a) << 12) - (a)) / (b))
#define COMBINE2(a, b, A, B)	_RENORM((a) * (b) + (A) * (B))

#define ALPHA1			((1 << 12) - 1)
#define ALPHAINV(a)		(ALPHA1 - (a))
#define ISOPAQUE(a)		((a) >=	(0xfe80 >> 4))

#define COMPOSELOOP(FA, FB) 						\
    do {								\
	int normEA = extraAlpha256 * ALPHAINV(0) / 256;			\
	pathScan -= width;						\
	srcScan -= width * 2;						\
	dstScan -= width * 2;						\
	while (--height >= 0) {						\
	    int w = width;						\
	    while (--w >= 0) {						\
		int pathAlpha = *pathData++;				\
		if (pathAlpha) {					\
		    int srcAlpha;					\
		    int srcColor;					\
		    int dstAlpha;					\
		    int dstColor;					\
		    int Fa, Fb;						\
		    srcAlpha = NORM(srcData[0]);			\
		    srcColor = NORM(srcData[1]);			\
		    if (!srcPremul && !ISOPAQUE(srcAlpha)) {		\
			srcColor = COMBINE(srcColor, srcAlpha);		\
		    }							\
		    if (!ISOPAQUE(normEA)) {				\
			srcAlpha = COMBINE(srcAlpha, normEA);		\
			srcColor = COMBINE(srcColor, normEA);		\
		    }							\
		    dstAlpha = NORM(dstData[0]);			\
		    dstColor = NORM(dstData[1]);			\
		    if (!dstPremul && !ISOPAQUE(dstAlpha)) {		\
			dstColor = COMBINE(dstColor, dstAlpha);		\
		    }							\
		    Fa = (FA);						\
		    Fb = (FB);						\
		    srcAlpha = COMBINE2(srcAlpha, Fa, dstAlpha, Fb);	\
		    srcColor = COMBINE2(srcColor, Fa, dstColor, Fb);	\
		    if (pathAlpha < 255) {				\
			int c;						\
			pathAlpha = NORM(pathAlpha);			\
			c = ALPHAINV(pathAlpha);			\
			srcAlpha = COMBINE2(srcAlpha, pathAlpha,	\
					    dstAlpha, c);		\
			srcColor = COMBINE2(srcColor, pathAlpha,	\
					    dstColor, c);		\
		    }							\
		    if (!dstPremul && srcAlpha > 0) {			\
			srcColor = UNCOMBINE(srcColor, srcAlpha);	\
		    }							\
		    srcAlpha = DENORM(srcAlpha);			\
		    srcColor = DENORM(srcColor);			\
		    resData[0] = srcAlpha;				\
		    resData[1] = srcColor;				\
		}							\
		srcData += 2;						\
		dstData += 2;						\
		resData += 2;						\
	    }								\
	    pathData += pathScan;					\
	    srcData += srcScan;						\
	    dstData += dstScan;						\
	    resData += dstScan;						\
	}								\
    } while (0)

static void
ClearInt2(unsigned char *srcData, int srcScan, int srcPremul,
	  unsigned char *dstData, int dstScan, int dstPremul,
	  unsigned char *pathData, int pathScan, int extraAlpha256,
	  unsigned char *resData, int width, int height)
{
    COMPOSELOOP(0, 0);
}

static void
SrcInt2(unsigned char *srcData, int srcScan, int srcPremul,
	unsigned char *dstData, int dstScan, int dstPremul,
	unsigned char *pathData, int pathScan, int extraAlpha256,
	unsigned char *resData, int width, int height)
{
    COMPOSELOOP(ALPHA1, 0);
}

static void
SrcOverInt2(unsigned char *srcData, int srcScan, int srcPremul,
	    unsigned char *dstData, int dstScan, int dstPremul,
	    unsigned char *pathData, int pathScan, int extraAlpha256,
	    unsigned char *resData, int width, int height)
{
    COMPOSELOOP(ALPHA1, ALPHAINV(srcAlpha));
}

static void
DstOverInt2(unsigned char *srcData, int srcScan, int srcPremul,
	    unsigned char *dstData, int dstScan, int dstPremul,
	    unsigned char *pathData, int pathScan, int extraAlpha256,
	    unsigned char *resData, int width, int height)
{
    COMPOSELOOP(ALPHAINV(dstAlpha), ALPHA1);
}

static void
SrcInInt2(unsigned char *srcData, int srcScan, int srcPremul,
	  unsigned char *dstData, int dstScan, int dstPremul,
	  unsigned char *pathData, int pathScan, int extraAlpha256,
	  unsigned char *resData, int width, int height)
{
    COMPOSELOOP(dstAlpha, 0);
}

static void
DstInInt2(unsigned char *srcData, int srcScan, int srcPremul,
	  unsigned char *dstData, int dstScan, int dstPremul,
	  unsigned char *pathData, int pathScan, int extraAlpha256,
	  unsigned char *resData, int width, int height)
{
    COMPOSELOOP(0, srcAlpha);
}

static void
SrcOutInt2(unsigned char *srcData, int srcScan, int srcPremul,
	   unsigned char *dstData, int dstScan, int dstPremul,
	   unsigned char *pathData, int pathScan, int extraAlpha256,
	   unsigned char *resData, int width, int height)
{
    COMPOSELOOP(ALPHAINV(dstAlpha), 0);
}

static void
DstOutInt2(unsigned char *srcData, int srcScan, int srcPremul,
	   unsigned char *dstData, int dstScan, int dstPremul,
	   unsigned char *pathData, int pathScan, int extraAlpha256,
	   unsigned char *resData, int width, int height)
{
    COMPOSELOOP(0, ALPHAINV(srcAlpha));
}

CompositeSet Int2Set = {
    ClearInt2,
    SrcInt2,
    SrcOverInt2,
    DstOverInt2,
    SrcInInt2,
    DstInInt2,
    SrcOutInt2,
    DstOutInt2,
};
