/*
 * @(#)reference.c	1.1 98/04/10
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Reference functions for testing the various compositing methods
 * for accuracy.
 */

#include <stdio.h>

#include "alphatest.h"

/*
 * Since all calculations are performed in alpha-premultiplied form
 * and since premultiplying a color value by a zero alpha "loses
 * information", we are then unable to restore the original color
 * value when we "unpremultiply" the alpha out (for producing a
 * non-premultiplied result).
 *
 * This problem can create a discontinuity in the results of the
 * raw color components of non-premultiplied destinations since an
 * alpha of 1/255 will generally retain enough color information to
 * get back to the original raw color value and an alpha of 0 will
 * always result in a raw color value of 0.
 *
 * This problem is not very severe since the "real error", i.e. the
 * value whose accuracy we are most concerned with, is the product
 * of the alpha and the color.  That "multiplied" value does not
 * tend to have a discontinuity.  Another way of looking at this
 * is that if the resulting alpha value is 0, then the resulting
 * raw color value is irrelevant.
 *
 * On the other hand, various algorithmic optimizations in the composite
 * functions may choose to leave the color value untouched if it is
 * the case that the alpha will go to zero (since the color value thus
 * becomes irrelevant - why bother storing anything in the color storage
 * locations?).  Other calculations may take the post-processing step of
 * dividing by the resulting alpha to get a non-premultiplied value and
 * cancel the division out of the equation.  In those calculations, since
 * the "multiply by alpha" was cancelled out by the later "divide by
 * alpha" there is no loss of information when alpha happens to be 0.
 * In testing against such algorithms it would be nice not to have to
 * force such algorithms to zero the color components just to avoid
 * lots of differences in raw color components.
 *
 * Thus a workaround is devised to try to preserve color information
 * when the alpha is zero.
 *
 * The workaround is simple and relies on the fact that we are using
 * floating point arithmetic in these references functions.  We simply
 * approximate a zero alpha with a value that is so small that it has
 * negligible effect on the rest of the calculations (i.e. the error
 * it causes will be swamped by rounding results to byte storage values).
 * This approximation will produce "multiplied" color values that are
 * near zero, but still can be restored to their original value by
 * dividing out the "close but not quite zero" alpha value.  If such
 * a "near zero" multiplied color is combined with any color that has
 * any opacity at all, then it will "lose out" to that other color and
 * have negligible impact on the result.
 *
 * It is important to realize, though, that the important values to
 * track for accuracy during testing are the raw alpha result and the
 * multiplied color result, not the raw color result.
 */
#define EPSILON		(1.0e-20f)

/*
 * Inputs:
 *	alpha in the range [0 => 256]
 *	color in the range [0 => 256] (possibly premultiplied)
 *	premul in the set {False, True}
 *	dst in the set {False, True}
 * Outputs:
 *	normalpha in the range [0.0 => 1.0]
 *	premulcolor in the range [0.0 => 1.0] (premultiplied with normalpha)
 * Returns:
 *	None
 */
static float
prepareColor(int alpha, int color,
	     float *normalpha, float *premulcolor,
	     int premul, int dst)
{
    float A, C;

    A = alpha / 255.0f;
    if (A < EPSILON) {
	A = EPSILON;
    }
    C = color / 255.0f;
    if (!premul) {
	C = C * A;
    } else if (C > A) {
	fprintf(stderr,
		"Supposedly premultiplied %s Color(A=%f, C=%f) wasn't!\n",
		dst ? "Dest" : "Src", alpha, color);
    }
    *normalpha = A;
    *premulcolor = C;
}

/*
 * Inputs:
 *	srcAlpha, srcColor, dstAlpha, dstColor in the range [0 => 255]
 *		(all possibly premultiplied)
 *	srcPremul, dstPremul in the set {False, True}
 *	extraAlpha256 in the range [0 => 256]
 * Outputs:
 *	Aa, Ca, Ab, Cb in the range [0.0 => 1.0]
 *		(all Cx premultiplied with Ax)
 * Returns:
 *	None
 */
static void
prepare(int srcAlpha, int srcColor, int srcPremul,
	int dstAlpha, int dstColor, int dstPremul,
	int extraAlpha256,
	float *pAa, float *pCa, float *pAb, float *pCb)
{
    float Ae;		/* adjusted extra A */

    prepareColor(srcAlpha, srcColor, pAa, pCa, srcPremul, 0);
    prepareColor(dstAlpha, dstColor, pAb, pCb, dstPremul, 1);

    Ae = extraAlpha256 / 256.0;
    if (Ae < EPSILON) {
	Ae = EPSILON;
    }
    *pAa *= Ae;
    *pCa *= Ae;
}

#define COMBINE(Fa, Ca, Fb, Cb)		(Fa * Ca + Fb * Cb)

/*
 * Inputs:
 *	Fa, Aa, Ca, Fb, Ab, Cb in the range [0.0 => 1.0]
 *		(all premultiplied)
 *	pathAlpha in the range [0 => 255]
 * Outputs:
 *	resC, resA in the range [0.0 => 1.0]
 * Returns:
 *	None
 */
static void
refcalc(float Fa, float Aa, float Ca,
	float Fb, float Ab, float Cb, int dstPremul,
	int pathAlpha, float *pResA, float *pResC)
{
    float Pa, Pb;	/* Adjusted path alpha coefficients */
    float At, Ct;	/* Temporary (intermediate result) A, C */
    float Ar, Cr;	/* Final result A, C */
    int valid;

    At = COMBINE(Fa, Aa, Fb, Ab);
    Ct = COMBINE(Fa, Ca, Fb, Cb);

    Pa = pathAlpha / 255.0f;
    Pb = (1.0f - Pa);
    Ar = COMBINE(Pa, At, Pb, Ab);
    Cr = COMBINE(Pa, Ct, Pb, Cb);

    if (!dstPremul && Ar > 0.0f) {
	Cr /= Ar;
    }

    /*
     * Since we don't want the math above to generate a NaN (in
     * particular, the division calculation), we want to detect
     * that case and report it.  Since any comparison involving
     * a NaN is always false, we thus want to use tests that
     * return true if the results are valid so that a NaN will
     * fail the tests and be flagged as "invalid".
     */
    valid = True;
    valid = valid && (Ar >= 0.0f && Ar <= 1.0f);
    valid = valid && (Cr >= 0.0f && Cr <= 1.0f);
    if (dstPremul) {
	valid = valid && (Cr <= Ar);
    }
    if (!valid) {
	fprintf(stderr, "Assertion violation in results: A=%f, C=%f, %s\n",
		Ar, Cr, (dstPremul ? "Premul" : "Not Premul"));
    }

    *pResA = Ar;
    *pResC = Cr;
}

#define COMPOSITE(FA, FB)				\
    do {						\
	float Aa, Ca, Cao, Ab, Cb, Cbo;			\
							\
	prepare(srcAlpha, srcColor, srcPremul,		\
		dstAlpha, dstColor, dstPremul,		\
		extraAlpha256,				\
		&Aa, &Ca, &Ab, &Cb);			\
							\
	refcalc((FA), Aa, Ca, (FB), Ab, Cb,		\
		dstPremul, pathAlpha,			\
		resultA, resultC);			\
    } while (0)

void
ClearReference(int srcAlpha, int srcColor, int srcPremul,
	       int dstAlpha, int dstColor, int dstPremul,
	       int pathAlpha, int extraAlpha256,
	       float *resultA, float *resultC)
{
    COMPOSITE(0.0f, 0.0f);
}

void
SrcReference(int srcAlpha, int srcColor, int srcPremul,
	     int dstAlpha, int dstColor, int dstPremul,
	     int pathAlpha, int extraAlpha256,
	     float *resultA, float *resultC)
{
    COMPOSITE(1.0f, 0.0f);
}

void
SrcOverReference(int srcAlpha, int srcColor, int srcPremul,
		 int dstAlpha, int dstColor, int dstPremul,
		 int pathAlpha, int extraAlpha256,
		 float *resultA, float *resultC)
{
    COMPOSITE(1.0f, 1.0f - Aa);
}

void
DstOverReference(int srcAlpha, int srcColor, int srcPremul,
		 int dstAlpha, int dstColor, int dstPremul,
		 int pathAlpha, int extraAlpha256,
		 float *resultA, float *resultC)
{
    COMPOSITE(1.0f - Ab, 1.0f);
}

void
SrcInReference(int srcAlpha, int srcColor, int srcPremul,
	       int dstAlpha, int dstColor, int dstPremul,
	       int pathAlpha, int extraAlpha256,
	       float *resultA, float *resultC)
{
    COMPOSITE(Ab, 0.0f);
}

void
DstInReference(int srcAlpha, int srcColor, int srcPremul,
	       int dstAlpha, int dstColor, int dstPremul,
	       int pathAlpha, int extraAlpha256,
	       float *resultA, float *resultC)
{
    COMPOSITE(0.0f, Aa);
}

void
SrcOutReference(int srcAlpha, int srcColor, int srcPremul,
		int dstAlpha, int dstColor, int dstPremul,
		int pathAlpha, int extraAlpha256,
		float *resultA, float *resultC)
{
    COMPOSITE(1.0f - Ab, 0.0f);
}

void
DstOutReference(int srcAlpha, int srcColor, int srcPremul,
		int dstAlpha, int dstColor, int dstPremul,
		int pathAlpha, int extraAlpha256,
		float *resultA, float *resultC)
{
    COMPOSITE(0.0f, 1.0f - Aa);
}
