/*
 * @(#)DataCollector.java	1.4 98/03/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package benchmarks;

import java.util.*;
import java.rmi.*;
import java.rmi.server.*;
import java.rmi.registry.*;
import java.io.*;
import java.net.*;

/**
 * The RMI Server ,implementing two sets of services -that of a server 
 * a set of test cases and that of a data Collector logging performance figures
 * from various clients.
 */

public class DataCollector 
	extends java.rmi.server.UnicastRemoteObject 
	implements CollInterface, TestInterface, Runnable {

	private String Name;

	//Controls Verbose or Silent mode for the data Collector(def. Silent)
	//Set by the command line argument -showProg
	private static boolean showProg = false;

	/**
	 * The properties for the server
	 */
	protected static Properties ServerDesc;

	/**
	 * A log for all client information hashed on a client num assigned by
	 * the data collector to each client when it logs the client desc.
	 * 
	 */
	protected Hashtable ClientLog = new Hashtable();


	/**
	 * The set of Clients currently holding a reference to the server. 
	 * A client registers itself in the client set when it logs in it's 
	 * system properties.
	 * The data Collector continually pings each of these clients to 
	 * test the reference. If the connection is broken or if the client
	 * has logged in it's performance data, it is removed from the client
	 * set.  
	 *
	 */

	protected Hashtable ClientSet = new Hashtable();

	/**
	 * The median RMIPerfMark for all the clients
	 */
	 protected float MedianMark ;

	/**
	 * A client identification number assigned to each client at the time of
	 * of logging in the client desc. so that it can later be uniquely 
	 * identified by the server at the time of logging in performance data
	 */
	 
	protected static int ClientNum = 0;	

	/**
	 * A flag indicating that this is the first client to log in
	 * to the server. The first client starts the Client Watcher which is
	 * a thread continually pinging the clients holding a reference to the
	 * data Collector. 
	 *
	 */

	protected static boolean FIRST_CLIENT = true;

	/**
	 * A flag indicating that the DataCollector is in the process of 
	 * generating a report. Any clients trying to log in at this time
	 * are not allowed to do so. The Client is returned a NOT_ASSIGNED 
	 * instead of the Client ID. 
	 */
	protected static boolean GENERATING_REPORT = false; 

	/**
	 * This is the same as the NOT_ASSIGNED flag on the client side.
	 * It indicates to the client that an ID could not be assigned to it 
	 * by the data Collector, possibly because of a broken connection or
	 * because the data Collector is presently compiling a performance 
	 * report. The Client on receiving this indication tries logging in
	 * again periodically for a fixed number of times 
	 *
	 * @see TestClient#NOT_ASSIGNED
	 */
	protected static int NOT_ASSIGNED = -1;

	private Vector v;
	private Object[] a;

	/*
	 * Generates a new client Id
	 * 
	 */
	private synchronized final int getNextClientNum() {
		return ++ClientNum;
	}	

	/*
	 * Write the text to a file specified by the ReportDest property .
	 * Called by generateReport() to write the HTML report .    
	 * 
	 */
 
	private void writeToFile (String text) throws CollectorException{
		
		BufferedOutputStream buff;
	
		String reportDest = new String(ServerDesc.getProperty(
				"ReportDest", "report.html"));

		//Copy the string to a byte array.
		byte[] filecontents = new byte[text.length()];
		filecontents = text.getBytes();

		try {
			buff= new BufferedOutputStream(
				new FileOutputStream(reportDest));
			buff.write(filecontents);
			buff.flush();
			buff.close();
		}catch (IOException e) {
			throw new CollectorException("Could not generate " +
					"report. Exception : " +e.getMessage());
		}
	}

	/*
	 * This method sets up the data in the ClientLog in the form of a 
	 * hashtable keyed by the a test identification string . This is only
	 * to setup the data in a form suitable for the HTML report
	 */

	private Hashtable setupData (int numClients) {
		 
		//Note that this method assumes that the number of iterations
		//made in the tests is the same across all clients. eg. if
		//Client 1 makes 100, 500 and 1000 null calls, all the clients
		//do the same.

		Hashtable report = new Hashtable();
		TabularData table; 
		int numLoops;
		Integer attrib;
		Vector data;
		int[] itns;
		float[][][] td, sertd;
		long[][]lo;
		ClientData currClientData;
		Hashtable perfData = null;    
		AttribData attribData;
		Vector figures;
		int[] sizeList;
		int i;
		PerfFigures perfFigure;
		boolean arraysUsed = false;

		//Set up the tables using data for Client 1. If there are any
		//Clients the one to log in first is guranteed to be assigned 
		//an ID of 1. Hence the following.

		Integer currClientNum = new Integer(1);
				
		if (ClientLog.containsKey(currClientNum)) {
                        if ((currClientData =
                           (ClientData)ClientLog.get(currClientNum))!=null){
				perfData = currClientData.getPerfFigures();
				if (perfData != null) {
					Enumeration e = perfData.keys();
					while (e.hasMoreElements()) {
					   attrib = (Integer)e.nextElement();
					   attribData = 
					    (AttribData)perfData.get(attrib);
					   arraysUsed =
					    attribData.areArraysUsed();
					   data = attribData.getFigures();
					   numLoops = data.size();
					   String testname= attribData.getMeasuredAttrib();

					   if (!arraysUsed){
					     table = new TabularData(testname, numLoops,
							numClients);
					   }else {
					     sizeList =
						attribData.getSizeList();
					     table = new TabularData(testname,numLoops,
							numClients, 
							sizeList.length);

					     //Since the list of array sizes
					     //is the same across a particular
					     //test, it can be set here itself.
	
					     table.setSizeList(sizeList);
					   }
					   report.put(attrib,table);
					}
				}
			}
		}

		//Fill the tables in report data with measured data


		//Get the list of measured attributes
		Enumeration e = perfData.keys(); 

		while (e.hasMoreElements()) {

		  attrib = (Integer)e.nextElement();
		  table = (TabularData)report.get(attrib);

		  itns = table.getIterations(); 
		  td = table.getTimeData();
		  sertd = table.getserTimeData();
		  if (table.areArraysUsed()){
			sizeList = table.getSizeList();
		  }
		  lo = table.getLoopOvrhead();

		  int clientCount = 1;
		  currClientNum = new Integer(clientCount);
		  while (ClientLog.containsKey(currClientNum)) {
			if ((currClientData =
			   (ClientData)ClientLog.get(currClientNum))!=null) {
				perfData = currClientData.getPerfFigures();
				if (perfData != null) {
					attribData = 
					    (AttribData)perfData.get(attrib);

					figures = attribData.getFigures();
				
					//Note: figures.size() is the same as
					//      numLoops for each attribute,
					


					for (i = 0; i < figures.size(); i++) {
					    perfFigure=
					      (PerfFigures)figures.elementAt(i);
					    itns[i]= perfFigure.getIterations();
					    float[] timings = 
						perfFigure.getTotalTime();
					    float[] sertimings = 
						perfFigure.getserTotalTime();
					    td[i][clientCount - 1]= timings;
					    sertd[i][clientCount - 1]= sertimings;
					    lo[i][clientCount - 1]=
						perfFigure.getOverhead();
	
					}
				}
			}
			clientCount++;
			currClientNum = new Integer(clientCount);
		   }
		   table.setIterations(itns);
		   table.setTimeData(td);
		   table.setserTimeData(sertd);
		   table.setLoopOvrhead(lo);
		}

		return report;
	}

	/**
	 * Generate the HTML report summarizing the performance data logged 
	 * in by the clients. Synchronized with logClientDesc() and 
	 * logPerfData() since they all access common data structures like the
	 * Client log.
	 * 
	 * @exception CollectorException if the report cannot be generated
	 */ 

	protected synchronized void generateReport () 
		throws CollectorException{

		//a performance mark generated for each Client
		//It's the sum of the times taken for each test.   

	        float RMIPerformanceMark;

		//A vector of RMIPerformance marks for all the clients      
		//used to calculate a RMI marks for the day. 
		float[] RMIPerfMarks ;


		Integer currClientNum;
		int itns, i,j, k,count = 0;
		ClientData currClientData;
		Hashtable perfData,report = null;
		TabularData table;
		Integer measuredAttrib; 
		Properties clientDesc;
		Vector figures;
		PerfFigures perfFigure;
		float[] totTime;
		Date today = new Date();
		Float tmpItns[];
		int[] iterations;
		float[][][] timeData, sertimeData;
		int[] sizeList = null;
		float timeTaken, sertimeTaken ;
		boolean arraysNotUsed;
		long[][] loopOvrhead;
		AttribData attribData;
		float[] ItnsPerSecond;

		StringBuffer html_report_1  = new StringBuffer(
			"<HTML>\n" +
			"<HEAD>\n" + 
			"<TITLE>RMI Performance Report </TITLE>\n" + 
			"<!--Machine generated. Do not edit-->\n" +
			"<!--Generated on " +  today.toString() + " -->\n" +
			"<BODY BGCOLOR=\"#FFFFFF\" TEXT=#000000 LINK=#0000ff" +
				" VLINK=#000077 ALINK=#ff0000>\n" +
			"<CENTER>\n" +
			"<P ALIGN=LEFT>\n" +
			"<IMG SRC=\"docs/images/smi.logo.gif\" " +
			"ALT=\"Sun Microsystems\" HEIGHT=43 WIDTH=120>\n" +
			"</P>\n" +
			"</CENTER>\n" +
			"<HR SIZE=3 ALIGN=CENTER NOSHADE>\n" + 
			"<CENTER>\n" +
			"	<H1>RMIPerformanceMark measurements " +
			"Version " + ServerDesc.getProperty("Version","1.0") +
			"</H1>\n" +
			"	<H2>Measured on : "+today.toString()+"</H2>\n"+ 
			"</CENTER>\n" + 
			"<BR>\n\n" +
			"<P>\n" +
			"The <A HREF=\"#Measurements\">Measurement Data</A> " +
			"indicates the tests performed and the client " +
			"performance on each of these tests. " +
			"The <A HREF=\"#RMIMarks\">RMIPerfMark</A> " +
			"is calculated as the sum of " +
			"the times taken for each of the tests. The Median " +
			"mark shown at the bottom is the median of all the " +
			"RMIPerfMarks for all the clients." +
			"</P>" +
			"<P>\n" +
			"The <A HREF=\"#ServerProperties\">Server System " +
			"Properties</A> and the <A HREF=\"#ClientProperties\">"+
			"Client System Properties</A> show the system specs." +
			" for the server and the client respectively." +
			"</P>\n");

		StringBuffer html_report_2 = new StringBuffer (
			"<BR>\n\n" +
			"<HR SIZE=3 ALIGN=CENTER NOSHADE>\n" + 
			"<IMG SRC=\"docs/images/JAVA.85.GIF\" HEIGHT=85 " +
			"WIDTH=49 ALIGN=RIGHT>" +
			"Questions, comments or Suggestions? Send us " +
			"<A HREF=\"mailto: nandu@kesavan.east.sun.com\">" +
			"feedback</A>" +
			"</BODY>\n" + 
			"</HTML>");



		//Index for RMIPerfMarks vector 
		count = 0; 

		//Add Server system properites to the HTML report

		html_report_1.append("<H2><A NAME=\"ServerProperties\">" + 
		   	"Server System Properties</A>:</H2>\n"); 
		html_report_1.append("<PRE>\n");
		html_report_1.append ("Server running JDK " +
			ServerDesc.getProperty("java.version") );
		html_report_1.append(" on " +
			ServerDesc.getProperty("os.name"));
		html_report_1.append(" ,version " + 
			ServerDesc.getProperty("os.version") + "\n");
		html_report_1.append("OS Arch.: " + 
			ServerDesc.getProperty("os.arch") + "\n");
		html_report_1.append("Java Vendor: " +
			ServerDesc.getProperty("java.vendor")) ;
		html_report_1.append("</PRE>\n");
		html_report_1.append("<BR>");
		html_report_1.append("<HR SIZE=2 ALIGN=LEFT " +
					"NOSHADE WIDTH=\"25%\">\n");

		html_report_1.append("<H2><A NAME=\"ClientProperties\">" +
		"Client System Properties</A>:</H2>\n");
		html_report_1.append("<PRE>\n");

		//Output the client system properties

		//Iterate through the clients starting from client 1
		int clientCount = 1;
 		currClientNum = new Integer(clientCount);

		while (ClientLog.containsKey(currClientNum)) {
			if ((currClientData =
                      	   (ClientData)ClientLog.get(currClientNum))!=null){
                        	clientDesc = currClientData.getDesc();
				html_report_1.append("Client " + 
					currClientNum.intValue() +
					" running JDK " +
       	                        	clientDesc.getProperty("java.version"));
       		                html_report_1.append(" on " +
                                	clientDesc.getProperty("os.name"));
                         	html_report_1.append(" ,version " +
                                	clientDesc.getProperty("os.version") +
					"\n");
                         	html_report_1.append("OS Arch.: " +
                                	clientDesc.getProperty("os.arch") +
					"\n");
                         	html_report_1.append("Java Vendor: " +
                                	clientDesc.getProperty("java.vendor") +
					"\n\n");
				html_report_1.append("<BR>\n");
			}
		   	clientCount++;
		   	currClientNum = new Integer(clientCount);
		}
		html_report_1.append("</PRE>\n");

		int numberOfClients = --clientCount;

		html_report_1.append("<HR SIZE=2 ALIGN=LEFT " +
					"NOSHADE WIDTH=\"25%\">\n");

		html_report_1.append("<BR>");
		html_report_1.append("<H2><A NAME=\"Measurements\">" +
			" Measurement Data</H2>\n");
		html_report_1.append( 
			"<P>\n" +
			"Lesser is better. The following numbers " + 
			"represent the time taken , in seconds. " + 
			"The number in brackets indicates the overhead in " +
			"taking the timestamps (<EM> in milliseconds </EM>)." +
			"<B>Note:</B> <EM>This is merely "+
			"for information. They have already been factored " +
			"into the calculations for the actual times." +  
			"For tests involving arrays or " +
			"vectors, the array or vector sizes tested is " +
			"indicated as a bulleted list. The corresponding " +
			"timings are also in the form of a bulleted list.</EM>"+
			"</P>\n");  

 		//Set up a hashtable keyed by measured attribute suitable for a 
		//tabular representation of the results.  
		report = setupData(numberOfClients);	

		//Print out a list of tables one for each measured 
		//attribute showing the number of iterations per loop and the
		//time taken for each iteration
		Enumeration e = report.keys();

		while (e.hasMoreElements()) {
			measuredAttrib = (Integer)e.nextElement();

			//Get the data for the test
			table = (TabularData)report.get(measuredAttrib);

			//Set up table
			html_report_1.append("<TABLE BORDER COLS=" +
				(numberOfClients+1) + " WIDTH=\"50%\">\n");
			html_report_1.append("<CAPTION>Measured : " +
				table.getName() + "</CAPTION>\n");
			
			//Generate header row
			html_report_1.append("<TR>\n");
			html_report_1.append("<TD>No. of iterations</TD>\n");
			for (i = 1; i <= numberOfClients; i++){
			   html_report_1.append("<TD>Client " + i + " (RMI)</TD>\n");
			   html_report_1.append("<TD>Client " + i + " (Ser)</TD>\n");
			}
			html_report_1.append("</TR>\n");

			//Print out the data
			iterations = table.getIterations();

			if (table.areArraysUsed()) {
				sizeList = table.getSizeList();
				arraysNotUsed = false;
			}else
				arraysNotUsed = true;

			timeData = table.getTimeData();
			sertimeData = table.getserTimeData();
			loopOvrhead = table.getLoopOvrhead();

			for (i = 0; i < iterations.length; i++) {
			  html_report_1.append("<TR>\n");

			  //If the test does not involve the use of arrays
			  //or vectors , insert just the number of iterations
			  //else also output a bulleted list of array/vector
			  //sizes.

			  if (arraysNotUsed) {
			  	html_report_1.append("<TD>"+
				  	iterations[i]+"</TD>\n");
			  }else {
				html_report_1.append("<TD>");
				html_report_1.append(iterations[i]);
				html_report_1.append("<UL>");

				//Insert a bulleted list of array sizes. 

				for (j=0; j< sizeList.length; j++){
				   html_report_1.append("<LI>" +
					sizeList[j] + "</LI>\n");
				}
				
				html_report_1.append("</UL>\n");
				html_report_1.append("</TD>\n");
			  }

			  for (j = 0; j < numberOfClients; j++){

				html_report_1.append("<TD>");

				if (arraysNotUsed) {
				     timeTaken = timeData[i][j][0];
				     html_report_1.append(timeTaken +
					"(" + loopOvrhead[i][j] + ")");
				}else {
				     //A kludge to trick HTML into leaving 
				     //a line before the list so that the
				     //data is properly aligned.

				     html_report_1.append(".");
				     html_report_1.append("\n<UL>\n");
				     for (k=0; k<sizeList.length;k++) {

					html_report_1.append("<LI>");
					timeTaken = timeData[i][j][k];
					html_report_1.append(timeTaken +
					  "(" + loopOvrhead[i][j] + ")");
					html_report_1.append("</LI>\n");
				     }
				     html_report_1.append("</UL>\n");
				}
				html_report_1.append("</TD>\n");
				     
				     
				html_report_1.append("<TD>");

				if (arraysNotUsed) {
				     sertimeTaken = sertimeData[i][j][0];
				     html_report_1.append(sertimeTaken +
					"(" + loopOvrhead[i][j] + ")");
				}else {
				     //A kludge to trick HTML into leaving 
				     //a line before the list so that the
				     //data is properly aligned.

				     html_report_1.append(".");
				     html_report_1.append("\n<UL>\n");
				     for (k=0; k<sizeList.length;k++) {

					html_report_1.append("<LI>");
					sertimeTaken = sertimeData[i][j][k];
					html_report_1.append(sertimeTaken +
					  "(" + loopOvrhead[i][j] + ")");
					html_report_1.append("</LI>\n");
				     }
				     html_report_1.append("</UL>\n");
				}
				html_report_1.append("</TD>\n");
				     
				     
			  }
			  html_report_1.append("</TR>\n");
			}

			html_report_1.append("</TABLE>");
			html_report_1.append("<BR>");

			if (arraysNotUsed) {
				sizeList = null;
			}
			ItnsPerSecond = calcItnsPerSecond(timeData,
				 iterations, sizeList, arraysNotUsed);

			if (arraysNotUsed){
			       html_report_1.append("Iterations per second = ");
			       html_report_1.append(ItnsPerSecond[0] +" .\n");
			}else {
				for ( i = 0; i < ItnsPerSecond.length; i++){
				   html_report_1.append("Average no. of " +
				      "iterations per second, for size " );
				   html_report_1.append(sizeList[i]);
				   html_report_1.append(", is " +
					ItnsPerSecond[i] + " .\n");
				   html_report_1.append("<BR>");
				}
			}
				
			html_report_1.append("<BR>");
		}
			
		html_report_1.append("<HR SIZE=2 ALIGN=LEFT " +
					"NOSHADE WIDTH=\"25%\">\n");

		html_report_1.append("<BR>");


		//Compute the RMIPerformance mark for each client

		RMIPerfMarks = new float[numberOfClients];

		//Iterate through the clients starting from client 1
		clientCount = 1;
 		currClientNum = new Integer(clientCount);

		while (ClientLog.containsKey(currClientNum)) {
		  RMIPerformanceMark = 0;

		  if ((currClientData = 
		      (ClientData)ClientLog.get(currClientNum))!=null){
		       	perfData = currClientData.getPerfFigures();
			
		       if (perfData != null) {


		       	 Enumeration e1 = perfData.keys();


			 //Cycle through the list of tests for each client
			 //Measured attribute indicates the test performed.

			 while (e1.hasMoreElements()) {

			    measuredAttrib = (Integer)e1.nextElement();

			    //Get the set of performance figures for each 
			    //measured attribute. Each set consists of one or 
		 	    //more PerfFigures object containing the time taken
			    //taken for a particular number of iterations. Thus
			    //for each test we could also measure performance
			    //deterioration as number of iterations increases
			    //all in one go!
			    //Now!!isn't that great!!!

			    attribData =
				 (AttribData)perfData.get(measuredAttrib); 

			    if (attribData.areArraysUsed()){
				arraysNotUsed = false;
				sizeList = attribData.getSizeList();
			    }else{
				arraysNotUsed = true;
				sizeList = new int[1];  //dummy
			    }

			    figures = attribData.getFigures();

			    //Extract the data for each set of iterations.
			    for (i = 0; i < figures.size(); i++) {
				perfFigure = (PerfFigures)figures.elementAt(i); 
		            	itns = perfFigure.getIterations();
				if (arraysNotUsed) {
					totTime = new float[1];
				}else{
					totTime = new float[sizeList.length];
				}

			    	totTime = perfFigure.getTotalTime();
				
				for (j = 0; j < sizeList.length; j++) {
					RMIPerformanceMark += totTime[j];
				}

		            }	 
			 }
			 RMIPerfMarks[clientCount - 1] = RMIPerformanceMark;
			 currClientData.setPerfMark(RMIPerformanceMark);
			 if (showProg) {
			 	System.out.println("RMIPerformanceMark : " +
			     		RMIPerformanceMark );
			 }
		       }
	           }		
		   clientCount++;
		   currClientNum = new Integer(clientCount);
		}

		//Generate a table for the RMIPerfMarks
		html_report_1.append("<H2><A NAME=\"RMIMarks\">RMI PerfMarks" +
			"</A></H2>\n"); 
		html_report_1.append("<TABLE BORDER COLS=" + 
			numberOfClients  + " WIDTH=\"50%\">\n");

		//Generate header row
		html_report_1.append("<TR>\n");
		for (i = 1; i <= numberOfClients; i++){
			html_report_1.append("<TD>Client " + i + "</TD>\n");
		}
		html_report_1.append("</TR>\n");

		//Print out the PerfMarks
		html_report_1.append("<TR>\n");
		for (i = 0; i < numberOfClients; i++){
			html_report_1.append("<TD> " + RMIPerfMarks[i] +
				"</TD>\n");
		}
		html_report_1.append("</TR>\n");
		html_report_1.append("</TABLE>\n");
		html_report_1.append("<BR>");
		

		MedianMark = calcMedian(RMIPerfMarks);
		html_report_1.append("<H2>Median " +
				"RMIPerfMark: " + MedianMark + "</H2>\n");
		html_report_1.append("<BR>");
	  	html_report_1.append(html_report_2.toString());
		
		writeToFile(html_report_1.toString());

		return;
	}

	/**
         * Adds a logFile to the appropriate directory. 
	 * 
	 * @param data    the LogData to be stored in the log file
	 * @param name    the log filename
	 * @param version the test suite version
	 * @exception CollectorException if the log file cannot be stored
	 */
	public void addLog (LogData data,String name, String version) 
		throws CollectorException {

		ObjectOutputStream os = null;
		
		String subDirName = new String ("logs" +
			 version.replace('.','_'));

		String dirName = ServerDesc.getProperty("LogDir",
			"benchmarkLogs");
				
		File logDir = new File(dirName);
		if (logDir.exists()) {
			if (!logDir.isDirectory()){
			  System.out.println(dirName + " is not a " +
				"directory...");
			  System.exit(-1);
			}
		}else {
		 	if (!(logDir.mkdirs())) {
			   System.out.println("Could not create " +
				"directory " + dirName);
			   System.exit(-1);
			}
		}		

		File subDir = new File(logDir, subDirName);
		
		if (!subDir.isDirectory()){
			if (!(subDir.mkdir())) {
				System.out.println("Could not create " +
				   "subdirectory " + subDir);
			}
		}		
		File logFile = new File(subDir,name);

		if (!logFile.exists()) {
		   try {
			os = new ObjectOutputStream(
			   new FileOutputStream(logFile));
			os.writeObject(data);
			os.flush();
			os.close();
		   }catch (IOException e) {
			System.out.println("Data Collector err: addLog(): " +
				e.getMessage());
			e.printStackTrace();
			throw new CollectorException(e.getMessage());
		   }		
		}else {
		   System.out.println("A log with this name already exists. "+
			"Delete it or specify a new name...");
		   throw new CollectorException("Log with the name " +
				logFile.getPath() + " already "+
				"exists. Delete or specify new name...");
		}
	}	

	/**
	 * Store the performance data alongwith the date, etc. in the log file
	 */
	protected synchronized void storeData() {
	
		LogData data = new LogData();

		try {
			data.setDate(new Date());
			data.setRMIPerfMark(MedianMark);
			data.setClientLog(ClientLog);
			data.setServerDesc(ServerDesc);
			
			String path = ServerDesc.getProperty("LogDest",
				"perfdata");
			String version =ServerDesc.getProperty("Version","1.0");
			path += version.replace('.','_');
			path += ".log";

			String hostName = new String(ServerDesc.getProperty(
				"CentralStore", "localHost"));

			if (hostName.equals("localHost")){
			  	addLog(data, path, version);
			}else {
			  	String hostPath = new String("//");
				hostPath += hostName;
				RepositoryInterface store = 
				  (RepositoryInterface)Naming.lookup(hostPath +
					"/Repository");
				try {
				      store.addLog(data, path, version); 
				}catch (RepositoryException e){
				      System.out.println("DataCollector Err: " +
					e.getMessage());
				      System.exit(-1);
				}
			}			

		}catch (Exception e) {
			System.out.println("DataCollector Err: " +
						e.getMessage());
			e.printStackTrace();
			System.exit(-1);
		}
	}

	/**
	 * Calculates the average no. of iterations per second for a particular
	 * test, across all clients
	 *
	 * @param timeData      a 3 dimensional table consisting of the 
	 *                      performance figures for all the clients on
	 *			that test.
	 * @param itns	 	the range of iterations for the test.
	 * @param sizeList	the range of values for the second parameter, 
	 *			if a second parameter is used, else null
	 * @param arraysNotUsed a boolean indicating whether this test involves
	 *			a second parameter other than the number of 
 	 * 			iterations.
	 * @return  the average number of iterations per second for tests not 
	 *          involving a second parameter, else, an array of such avgs.
	 *          for each value of the second parameter.
	 */
	 protected float[] calcItnsPerSecond(float[][][] timeData, int[] itns,
			int[] sizeList, boolean arraysNotUsed) {

		int i, j, k;
		int numIterations, numClients, numSizes = 0;
		float[][] ItnsPerSecond = null;
		float[][][] ItnsPerSecondPerSize = null;
		float[] avgItnsPerSecond = null;
		float[][] avgItnsPerSecondPerSize = null;
		float[] medianItnsPerSecond;

		numIterations = itns.length;

		//Assuming no. of clients are the same across iteration ranges
		numClients = timeData[0].length;

		if (!(arraysNotUsed)){
			numSizes = sizeList.length;
			medianItnsPerSecond = new float[numSizes];
			ItnsPerSecondPerSize = 
			  new float[numClients][numSizes][numIterations];
			avgItnsPerSecondPerSize = 
			  new float[numSizes][numClients];
		}else{
		  	medianItnsPerSecond = new float[1];
			ItnsPerSecond = new float[numClients][numIterations];
			avgItnsPerSecond = new float[numClients];
		}

		for (i = 0; i < numIterations; i++ ) {
			for (j = 0; j < numClients; j++) {
				if (arraysNotUsed){
				   ItnsPerSecond[j][i]= 
					itns[i]/timeData[i][j][0];
				}else{

				   //Normalize all times to time taken for a 
				   //vector size of 100

				   for (k = 0; k < numSizes; k++){
					ItnsPerSecondPerSize[j][k][i] =
						itns[i]/timeData[i][j][k];
				   }

				}
			}
		}
		
		//Calculate Average iterations/second for each client
		
		for (i= 0; i < numClients; i++) {
			if (arraysNotUsed){
			   avgItnsPerSecond[i] = calcAvg(ItnsPerSecond[i]);
			}else {
			   for (j = 0; j < numSizes; j++) {
			       avgItnsPerSecondPerSize[j][i]=
					calcAvg(ItnsPerSecondPerSize[i][j]);
			   }
			}
		}

		if (arraysNotUsed){
			medianItnsPerSecond[0]= calcMedian(avgItnsPerSecond);
		}else{
			for (i =0; i < numSizes; i++){
			    medianItnsPerSecond[i] = 
	 			calcMedian(avgItnsPerSecondPerSize[i]);
			}
		}

		return medianItnsPerSecond;
	}
				
	/**
	 * Does a bubble sort (descending)  for an array of floating point
	 * values.
	 *
	 * @param a the floating point array
	 */

	 public void sort (float a[]) {

		int i,j;
		float tmp;


		for (i = 0; i < (a.length - 1); i ++){
			for (j = i + 1; j < a.length; j++) {
				if (a[i] < a[j]){
					tmp = a[i];
					a[i] = a[j];
					a[j] = tmp;
				}
			}
		}

		
		return;
	
	 }

	/**
	 * Calculates the median of for an array of floating point values 
	 *
	 * @param dataArray the array of floating point values
	 * @return the median 
	 */
	
	 public float calcMedian(float[] dataArray) {
		
		int n;
		float median;

		sort(dataArray);
		n = dataArray.length;
		median =(dataArray[(n-1)/2] + dataArray[n/2])/2;

		return median;

	 }

	
	/**
	 * Calculates the average for an array of floating point values
	 *
	 * @param dataArray the array of floating point values
	 * @return the average
	 */
	
	 public float calcAvg(float[] dataArray) {

		int i;
		float sum = 0.0f, average;
		
		for (i = 0; i < dataArray.length; i++ ) {
			sum += dataArray[i];
		}
		
		average = sum/dataArray.length;
		
		return average;
	}

	public DataCollector(String s) throws RemoteException  {
		Name = s;
	}

	/**
	 * A null call doing nothing, accepting no parametrs and returning 
	 * void
	 *
	 * @exception RemoteException if the remote operation fails
	 */
        public void nullCall () throws RemoteException {
                return;
        }

	/**
	 * A remote call with one integer argument
	 * 
	 * @exception RemoteException if the remote operation fails
	 */
        public void intCall(int IntData) throws RemoteException {
                return;
        }

	/**
	 * A remote call with two arguments, an Object and an integer
	 *
	 * @exception RemoteException if the remote operation fails
	 */
        public void twoArgCall(Object obj, int IntData)
                throws RemoteException{
                return;
        }

	/**
	 * A remote call with an Object argument. Serves as a general
	 * method for calls with arrays or any such object arguments
	 *
	 * @exception RemoteException if the remote operation fails
 	 */
        public void objCall(Object obj) throws RemoteException {
                return;
        }

	/**
	 * A remote call returning a remote object
	 *
	 * @exception RemoteException if the remote operation fails
	 */
        public Object getRemObj() throws RemoteException {

                RemoteObj remRef = new RemoteObj();
                return (remRef);

        }

	/**
	 * Create a new Vector of the specified size
	 *
	 * @param size the size of the vector
	 * @exception RemoteException if the remote call fails 
	 */

        public void createVector(int size) throws RemoteException {

                int i;

                v = new Vector(size);

                for (i = 0; i < size; i++) {
                        v.addElement(new RemoteObj(i));
                }
 		
		a = new Object[size];
		v.copyInto(a);

                return;
        }

	/**
	 * Returns the entire vector
	 *
	 * @see #getVector(int, int)
	 * @exception RemoteException if the remote call fails
	 */
	 
	public Vector getVector () throws RemoteException {

		return v;

	}

	/**
	 * Returns a vector containing the elements from <i>offset</i>
	 * through <i>offset + length - 1</i>
	 *
	 * @param offset the offset from which the elements are to be 
	 *               retreived
	 * @param length the number of elements to be retreived
	 * @see #getVector()
	 * @exception RemoteException if the remote call fails
	 */ 

        public  Vector getVector(int offset, int length)
		throws RemoteException {

		int i;
		Object[] dst = new Object[length];
		Vector slice = new Vector(length);

		//Copy elements from a[offset] thru a[offset + length -1] 
		//to dst[0] thru dst[length -1]
		System.arraycopy(a,offset,dst,0,length);

		//Copy the array elements to a vector and return this vector
	 	for (i = 0; i < length; i++)
			slice.addElement(dst[i]);

                return slice;
        }

	/**
	 * Simply throws a Collector Exception. (To test RMI calls generating
	 * exceptions)
	 *
	 * @exception RemoteException    if the remote call fails
	 * @exception CollectorException always thrown.
	 */
	public void exceptionCall() 
		throws RemoteException, CollectorException {

		//Throw a Collector exception
		throw new CollectorException();

	}


	/**
	 * Add the client properties to the Client Log and assign a client
	 * identification number to the client. Simultaneously, the client 
	 * also registers with the data Collector, by adding itself to the
	 * the client set. Synchronized with generateReport() and 
	 * logPerfData()
	 *
	 * @param c_desc the Client properites including the system properties
	 *               at the client end.
	 * @param client the Client interface which is passed to the data
	 *               Collector so that it can be added to the Client Set.
	 * @return       a client num to identify the client to the data 
	 *               Collector.
	 * @exception RemoteException if the server cannot log the client data
	 */

	public synchronized int logClientDesc (Properties c_desc, 
						ClientInterface client) 
		throws RemoteException {

		int newClientNum = NOT_ASSIGNED;

		//If the data Collector is generating the performance report
		//the client is not assigned an id ,insted NOT_ASSIGNED is 
		//returned. The client then periodically tries to log back till
		//either the maximum number of allowed attempts have been made
		//or it has been assigned an ID.

		if (!GENERATING_REPORT) {

			//get a new Client ID
			newClientNum = getNextClientNum();
			Integer ClientID = new Integer(newClientNum);

			ClientData cd = new ClientData(newClientNum, c_desc);
	
			if (showProg) {	
				System.out.println("Adding Client " +
					newClientNum + " to the Client Log " +
					"and Client Set.");
			}
			ClientLog.put(ClientID,cd );       
			ClientSet.put(ClientID,client);
			if (FIRST_CLIENT ) {

			     //Start the ClientWatcher which shall 
		             //ping the clients detecting if all 
			     //connections are alive.

			     Thread clientWatcher =
				 new Thread(this,"clientWatcher");
			     clientWatcher.start();
			     FIRST_CLIENT = false;
			}
		}

		return newClientNum;
	}

	/**
	 * Logs the performance data into the Client log. Identifies the 
	 * client by the clientNum parameter. Synchronized with logClientDesc()
	 * and generateReport()
	 *
	 * @param data A hashtable keyed on the measured attribute containing
	 *             the performance measurements for the client
	 * @clientNum  the client identification number assigned by the 
	 *             data Collector when logging in the client Desc
	 * @see #logClientDesc
	 * @exception RemoteException if the remote call fails
	 */

	public synchronized void logPerfData(Hashtable data, int clientNum)
		throws RemoteException {

		Integer ClientID = new Integer(clientNum);
		
		ClientData clientData = (ClientData)ClientLog.get(ClientID);
		clientData.addPerfFigures(data);
		ClientSet.remove(ClientID);
		
		return;
	}


	/** 
	 * This method pings the clients to check the connections. If the
	 * connection is broken, the client reference is removed from the
	 * client set. When the client set becomes empty, the data Collector
	 * generates a report 
	 */
	public void run() {

		while (!(ClientSet.isEmpty())) {
			//Keep pinging the clients every 5 seconds.
			try {
				Thread.currentThread().sleep(5000);
			}catch (InterruptedException e){} 

			Enumeration e = ClientSet.keys();
			while (e.hasMoreElements()) {
				
				//Get a reference to the next client in the 
				//client set.	
				Integer ClientID = (Integer)e.nextElement();
				ClientInterface client = 
				    (ClientInterface)ClientSet.get(ClientID);

				//Make a callback to the client to check the 
				//connection
				try {
					if (showProg) {
						System.out.println("Pinging #" +
					    		ClientID.intValue());
					}
					client.ping();
				}catch (RemoteException ex) {
					//Can't reach client, remove from
					//ClientSet and the Client Log.
					if (showProg) {
					     System.out.println("Removing " +
				 	      "Client # "+ClientID.intValue());
					}
					ClientSet.remove(ClientID);
					ClientLog.remove(ClientID);
				}
			}
		}

		//If the Client Log is *empty* then generate the report and 
		//store the data. While this is being done, set the 
		//GENERATING_REPORT flag to true, so that no new client
		//descriptions are logged in

		if (!(ClientLog.isEmpty())) {	
			GENERATING_REPORT = true;
			try {
				generateReport();
				if (showProg) {
					System.out.println("Generated Report");
				}
			}catch (CollectorException e) {
				System.out.println("Collector Error: " +
					e.getMessage());
				e.printStackTrace();
				System.exit(-1);
			}
			storeData();
			ClientLog.clear();
		}

		//reset the flags

		ClientNum = 0;
		FIRST_CLIENT = true;
		GENERATING_REPORT = false;
		if (showProg) {
		  	System.out.println("Stored New data");
		}
		System.out.println("Done... Exiting Succesfully");
		System.exit(0);
	}	

	/**
	 * Print the usage for this program
	 */

	public static void usage() {

		System.out.println("usage: java benchmarks.DataCollector " +
			"[-showProg] [-vers=\"java_version\"]");
		System.out.println("       -showProg              : " +
			"verbose execution. (default false)");
		System.out.println("       -vers=\"java_version\" : " +
			"the java version.");
		System.out.println("       -help                  : " +
			"displays this help");
		System.out.println("");
		System.out.println("NOTE: Specify the version as a command " +
			"line arg. only if you wish to override the version "+
			"string generated by System.getProperty(). E.g. for "+
			"internal builds, where the version string is not " +
			"detailed enough."); 
		
		return;
	}


	public static void main(String[] args) {

		Registry registry;
 		File src;
		int i, start, end;
		String version;


		System.setSecurityManager(new RMISecurityManager());
		ServerDesc = new Properties(System.getProperties());

		try {

			//Check usage

			if (args.length > 3) {
				usage();   	//error
				System.exit(0);
			}

			for (i = 0 ; i <args.length; i++) {
				if (args[i].equals("-showProg")){
					showProg = true;
				}else if (args[i].startsWith("-vers")){

					//Extract the version string enclosed
					//within quotes

					start= args[i].indexOf('\"');
					end = args[i].indexOf('\"',start+1);
					version = args[i].substring(start + 1,
						end);
					ServerDesc.put("java.version",version);
				}else{
				   	usage();
					System.exit(-1); //error
				}
			}

						
			ServerDesc.load(new FileInputStream(
				"benchmarks.server.properties"));
			DataCollector dc = new DataCollector("DataCollector");

			//Path is formed by appending the logdest string with
			//the test version (replacing all . by _) and a .log
			//extension.

			String path = ServerDesc.getProperty("LogDest",
						"perfdata");
			path += ".log";

                        src = new File(path);
			if (src.exists()) {
				System.out.println("A log file with this " +
				   "name already exists. Delete or specify "+
				   "new name.");
				System.exit(-1);
			}else {
			    if (showProg) {
	     		  	System.out.println("\tCreating new log...");
			    }
			}

			LocateRegistry.createRegistry(Registry.REGISTRY_PORT);
           		registry = LocateRegistry.getRegistry(
					Registry.REGISTRY_PORT);
            		registry.rebind("DataCollector", dc);

			if (showProg ) {
				System.out.println("Server bound in registry");
			}
			 
		} catch (Exception e) {
			System.out.println("DataCollector err: " +
				e.getMessage());
			e.printStackTrace();
		} 
	}
}

/**
 * TabularData is a table of data containing the measurement data for 
 * a particular client for a single measured attribute. 
 */

class TabularData {

	private int[] iterations;

	private int[] sizeList;

	private boolean arraysUsed ;

	private float[][][] timeData;

	private float[][][] sertimeData;

	private long[][] loopOvrhead;

	private String tname;

	//Constructor for data not involving arrays 

	protected TabularData (String testname, int numLoops, int numClients) {

		tname=testname;
		iterations = new int[numLoops];
		arraysUsed = false;
		sizeList = null;
		timeData = new float[numLoops][numClients][1];
		sertimeData = new float[numLoops][numClients][1];
		loopOvrhead = new long[numLoops][numClients];

	}

	//Constructor for data involving arrays or vectors.

	protected TabularData (String testname, int numLoops, int numClients, int numSizes) {

		tname=testname;
		iterations = new int[numLoops];
		arraysUsed = true;
		sizeList = new int[numSizes];
		timeData = new float[numLoops][numClients][numSizes];
		sertimeData = new float[numLoops][numClients][numSizes];

		loopOvrhead = new long[numLoops][numClients];

	}

	protected void setIterations(int[] itns) {
	
		System.arraycopy(itns,0,iterations,0,itns.length);
		return; 

	}	

	protected void setSizeList(int[] listOfArraySizes) {

		System.arraycopy(listOfArraySizes,0,sizeList,0,
			listOfArraySizes.length);
		return;

	}

	protected void setTimeData (float[][][] td) {

		for (int i =0; i <td.length;i++) {
		   for (int j =0; j < td[i].length; j++){
			System.arraycopy(td[i][j],0, timeData[i][j],
				0, td[i][j].length);
		   }
		}

		return;

	}

	protected void setserTimeData (float[][][] sertd) {

		for (int i =0; i <sertd.length;i++) {
		   for (int j =0; j < sertd[i].length; j++){
			System.arraycopy(sertd[i][j],0, sertimeData[i][j],
				0, sertd[i][j].length);
		   }
		}

		return;

	}

	protected void setLoopOvrhead (long[][] lo) {

		for (int i=0; i < lo.length; i++) {
			System.arraycopy(lo[i],0,loopOvrhead[i],0,lo[i].length);
		}
		return;

	}

	protected String getName() {
		
		return tname;

	}

	protected int[] getIterations() {
		
		return iterations;

	}

	protected boolean areArraysUsed() {

		return arraysUsed;

	}

	protected int[] getSizeList() {

		return sizeList;

	}

	protected float[][][] getTimeData() {

		return timeData;

	}

	protected float[][][] getserTimeData() {

		return sertimeData;

	}
	protected long[][] getLoopOvrhead() {

		return loopOvrhead;

	}
}




