/*
 * @(#)DataServerImpl.java	1.3 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package visualBenchmarks;

import java.rmi.*;
import java.rmi.registry.*;
import benchmarks.*;
import java.util.*;
import java.io.*;
import java.text.DateFormat;

/**
 * This class implements the functionality required by the LogViewer. E.g. it
 * fetches data from the log file, in the format requested by the LogViewer
 * clients. 
 */

public class DataServerImpl 
	extends java.rmi.server.UnicastRemoteObject
	implements DataServer {

	private String name; 

	private static String logPath;

	/**
	 * Gets a list of the log files.
	 * @return the list of log filenames.
	 * @exception RemoteException if the remote operation fails
	 */
	public String[] getVersions() 
		throws RemoteException {

		File logDirectory = new File(logPath);
		
		String[] logFilenames = logDirectory.list(new LogFilter());

		if (logFilenames ==  null){
			System.out.println("No files in Directory " +
				logDirectory.getAbsolutePath());
		}

		return logFilenames;
	}

	/**
	 * Fetch a list of the tests carried out in a particular version
	 * of the test suite. (<em>This version is set in the properties 
	 * file</em>). The list of tests returned also includes information 
	 * on wether arrays are used.
	 *
	 * @see TestInfo
	 * @return the list of tests carried out
	 * @exception RemoteException if the remote operation fails
	 */

	public Vector getListOfTests() 
		throws RemoteException {

		ObjectInputStream in = null;
		Vector testList = new Vector();
		LogData logData;
		Hashtable clientLog = null, perfData = null;
		Integer clientID;
		ClientData cd = null;
		AttribData ad = null;
		String testName ;

		//Assuming that for all logs stored in the same log directory,
		//the list of tests is the same (i.e for the same benchmarks
		//version, the list of tests is the same), we read only the 
		//first log

		File logDirectory = new File(logPath);
		String[] logs = logDirectory.list(new LogFilter());
		
		if (logs == null) {
			return null;
		}

	       	try {
       			in = new ObjectInputStream(
	       		new FileInputStream(new File(logDirectory,logs[0])));

			logData = (LogData)in.readObject();
			clientLog = logData.getClientLog();	

			if (clientLog != null) {
				//Assuming only one client 	
				clientID = new Integer(1);
				if (clientLog.containsKey(clientID)){
					cd=(ClientData)clientLog.get(clientID);
					perfData = cd.getPerfFigures();

					//perfData is a hashtable keyed by 
					//the test name.
					//Extract the list of keys and copy 
					//them to the Vector

					Enumeration e = perfData.keys();
					while (e.hasMoreElements()){
					   testName = (String)e.nextElement();
					   ad =
					    (AttribData)perfData.get(testName);
					   testList.addElement(
					    	new TestInfo((String)testName,
					    	ad.areArraysUsed()));
					}
				}
			}
			in.close();
						    		
		}catch (Exception e) {
			System.out.println("DataServerErr: " +
			  "(getListOfTests()): " + e.getMessage());
			e.printStackTrace();
		}


		return testList;
	}

	/**
	 * Deletes a logFile given the filename
	 * 
	 * @param key the filename for the log
	 * @exception RemoteException     if the remote operation fails
	 * @exception DataServerException if the file cannot be deleted
	 */
	public void deleteLog (String key)
		throws RemoteException, DataServerException {

		File logFile = new File (logPath, key);
		
		if (!(logFile.delete())) {
			throw new DataServerException("Could not delete " +
			   "the file: " + logFile);
		}
	}


		
	/**
	 * Fetch the details for a particular test. Returns the test info. 
	 * including Server system properties and the client system properties
	 * This method assumes that the test is carried out with only one
	 * client per server . Only the system properties for the first 
	 * client is returned. 
	 *
	 * @params key the log filename.
	 * @return the test details
	 * @exception DataServerException if the log file does not exist.
	 * @exception RemoteException if the remote operation fails.
	 */

	public String getLogInfo (String key)
		throws RemoteException, DataServerException {

		DateFormat df;
		StringBuffer displayText = new StringBuffer();
		Properties serverProps, cp ;
		Hashtable clientLog = null;
		int clientCount;
		Integer clientID; 
		ClientData cd = null;
		LogData logData = null;
		File logFile;

		logFile = new File (logPath, key);

		if (!logFile.exists()) {
			throw new DataServerException("LogFile with the path "+
				logFile.getAbsolutePath() +", does not exist.");
		}
	
		try {
			ObjectInputStream in = new ObjectInputStream(
				new FileInputStream(logFile));

			logData = (LogData)in.readObject();

			//Get the server-side system properties

			serverProps = logData.getServerDesc();
		
	 		displayText.append("Test carried out on " +
				logData.getDate().toString() + ".\n");
			displayText.append("Test Version : " +
				serverProps.getProperty("Version") + ".\n");
			displayText.append("Server running JDK " + 
				serverProps.getProperty("java.version")) ;
			displayText.append(" on " + 
				serverProps.getProperty("os.name"));
			displayText.append(" ,version " + 
				serverProps.getProperty("os.version") + ".\n");
			displayText.append("OS Arch: " + 
				serverProps.getProperty("os.arch") + ".\n");
			displayText.append("Java Vendor: " +
				serverProps.getProperty("java.vendor") + ".\n");
			displayText.append("\n\n");
		
			//Note: Assuming a single client per test.
			//Get the client-side system properties
			clientLog = logData.getClientLog();

			clientCount = 1;	
			clientID = new Integer(clientCount);
			if (clientLog.containsKey(clientID)) {
			    if ((cd = 
			      (ClientData)clientLog.get(clientID))!=null){
				cp = cd.getDesc();
				displayText.append("Client " + 
					clientID.intValue() +
					" running JDK " +
					cp.getProperty("java.version"));
				displayText.append(" on " +
					cp.getProperty("os.name"));
				displayText.append(" ,version " +
					cp.getProperty("os.version") +
					".\n");
				displayText.append("OS Arch.: " +
					cp.getProperty("os.arch") +
					".\n");
				displayText.append("Java Vendor: " +
					cp.getProperty("java.vendor") +
					".\n\n");
			     }
			}

		}catch (Exception e) {
			System.out.println("DataServerErr : (getTestInfo()): "+
				e.getMessage());
			e.printStackTrace();
		}
		if (displayText != null) 
			return displayText.toString();
		else 
			return "No Information available";

	}

	/**
	 * Get the log data for a set of tests.
	 * 
	 * @param keys an array of log filenames
	 * @return A vector of log data for the set of tests
	 * @exception DataServerException if the logFile does not exist
	 * @exception RemoteException if the remote operation fails
	 */

	public LogData[] getPerfLogs (String[] keys)
		throws RemoteException, DataServerException {

		ObjectInputStream in = null;

		LogData[] dataBank = new LogData[keys.length];
		File logFile;

		
		for (int i = 0; i < keys.length; i++) {
		    logFile = new File(logPath, keys[i]);

		    if (!logFile.exists()) {
			throw new DataServerException("LogFile with the path "+
				logFile.getAbsolutePath() +", does not exist.");
		    }
		    try {
		        in =new ObjectInputStream(new FileInputStream(logFile));
		        dataBank[i] = (LogData)in.readObject();
			in.close();
		    }catch (Exception e){
			System.out.println("DataServer error: getPerfLogs(): "+
				e.getMessage());
			e.printStackTrace();
		    }
		}

		return dataBank;
	}

		
	public DataServerImpl(String ds) 
		throws RemoteException {

		name = ds;

	}

	public static void main(String[] args){

		Properties ServerProps = new Properties(System.getProperties());
		
		System.setSecurityManager(new RMISecurityManager());
		
		try {

			DataServerImpl ds = new DataServerImpl("DataServer");

			LocateRegistry.createRegistry(1234);
			Registry registry =
			   LocateRegistry.getRegistry(1234);
			registry.rebind("DataServer",ds);

			System.out.println("DataServer bound in registry.");

			ServerProps.load(new FileInputStream(
				"visualBenchmarks.server.properties"));

			//Form the logfile path by appending the LogDest string
			//with the test version (with all . replaced by _) 
			logPath = 
			  ServerProps.getProperty("LogDir","benchmarkLogs");
			logPath += File.separatorChar;
			logPath += "logs";
			logPath +=
			   ServerProps.getProperty("Version","1.0").replace('.',
				'_');

		} catch (Exception e) {
			System.out.println("DataServerErr: (main()) " +
				e.getMessage());
			e.printStackTrace();
		}
	}
}
		

class LogFilter implements FilenameFilter {

 	public boolean accept (File dir, String name) {
	
		return (name.endsWith(".log"));
	}
}
