/*
 * @(#)DrawnRectangle.java	1.3 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package visualBenchmarks;

import java.awt.*;

/** 
 * The DrawnRectangle class paints a rectangle in the specified
 * component. It extends the java.awt.Rectangle class.
 */

public class DrawnRectangle extends Rectangle {

	private int thick;
	private Color lineColor, fillColor;

  	/**
	 * The default thickness for the rectangle border
	 */
	protected static int defaultThickness = 2;

	/**
 	 * The component into which the rectangle is to be drawn.
   	 */
	protected Component drawInto;
	
	/**
	 * Construct a rectangle into the specified component, with a default
	 * thickness, at location (0,0) with a height and width of 0
	 * 
	 * @param drawInto the component into which the rectangle is to be
	 *	           drawn
	 */
	public DrawnRectangle (Component drawInto) {
		this(drawInto, defaultThickness,0,0,0,0);
	}

	/**
	 * Construct a rectangle into the specified component, with a 
	 * specified  thickness, at location (0,0) ,and 0 height and width
	 * 
	 * @param drawInto the component into which the rectangle is to be
	 *	           drawn
	 * @param thick    the thickness of the rectangle to be constructed
	 */
	 
	public DrawnRectangle (Component drawInto, int thick) {
		this (drawInto, thick, 0,0,0,0);
	}

	/**
	 * Construct a rectangle into the specified component,with <em>default
	 * thickness</em>, at a specified location (x,y) and with the given 
	 * height and width.
	 *
	 * @param drawInto the component into which the rectangle is to be
	 *	           drawn
	 * @param x,y	   the location of the top left corner of the 
	 *		   rectangle
	 * @param w,h	   the width and height of the rectangle respectively
	 */
	 
	public DrawnRectangle (Component drawInto, int x, int y, int w, int h){
		this (drawInto, defaultThickness, x, y, w, h);
	}

	/**
	 * Construct a rectangle into the specified component,with a given
	 * thickness, at a specified location (x,y) and with the given height
	 * and width.
	 *
	 * @param drawInto the component into which the rectangle is to be
	 *	           drawn
	 * @param thick    the thickness of the rectangle to be constructed
	 * @param x,y	   the location of the top left corner of the 
	 *		   rectangle
	 * @param w,h	   the width and height of the rectangle respectively
	 */
	 
	public DrawnRectangle (Component drawInto, int thick, int x, int y,
				int w, int h) {
		this.drawInto = drawInto;
		this.thick = thick;
		setBounds (x,y,w,h);
	}
	
	/**
	 * Get the component , the rectangle is drawn into
	 *
	 * @return the component drawn into.
	 */

	public Component component() {
		return drawInto;
	}

	public int getThickness() {
		return thick;
	}

	public void setThickness (int thick) {
		this.thick = thick;
	}

	public void setLineColor (Color lineColor){
		this.lineColor = lineColor;
	}

	public void setFillColor (Color fillColor) {
		this.fillColor = fillColor;
	}

	/**
	 * Get the line color (border color) for the rectangle. If the line
	 * color is not set, this method sets it to three shades darker than
	 * than the background color.
	 *
	 * @return the line color 
	 */

	public Color getLineColor () {

		//If lineColor has not been set, set it to three shades
		//darker than the background color for the component being
		//drawn into.

		if (lineColor == null) {
			lineColor =
			   drawInto.getBackground().darker().darker().darker();
		}
		return lineColor;
	}

	/**
	 * Get the fill color. If the fill color is not set, this method sets
	 * it to the background color.
	 *
	 * @return the fill color
	 */

	public Color getFillColor() {

		if (fillColor == null) {
			fillColor = drawInto.getBackground();
		}
		return fillColor;
	}

	/**
	 * Get the inner bounds of the rectangle (excluding the border)
	 * 
	 * @return a Rectangle object representing the inner bounds
	 */

	public Rectangle getInnerBounds() {
		return new Rectangle ( x +thick, y + thick,
			width - (thick*2), height - (thick*2));
	}

	/**
	 * Draw the rectangle border.<em>To fill the rectangle, you also need
	 * to fill() the rectangle</em>. Gets the graphics object for the
	 * container and uses th paintFlat() method to draw the borders.
	 *
	 * @see #fill
	 */

	public void paint() {
		Graphics g = drawInto.getGraphics();
		paintFlat(g, getLineColor());
	}

  
	public void paintFlat (Graphics g, Color color) {

		g.setColor(color);
		for (int i =0; i < thick; i++) {
			g.drawRect(x+i, y+i,  width - (2*i) -1, 
				height - (2*i) -1);
		}
	}

  	/**
	 * Clear the interior of the rectangle. Fills the interior with the 
	 * background color.
	 * 
	 * @see #clearExterior
	 * @see #clear
	 */
	public void clearInterior () {
		fill (drawInto.getBackground());
	}

	/**
	 * Clear the borders of the rectangle
	 *
	 * @see #clearInterior
	 * @see #clear
	 */

	public void clearExterior() {
		paintFlat(drawInto.getGraphics(), drawInto.getBackground());
	}

	/**
    	 * Clears the rectangle
	 * 
	 * @see #clearInterior
	 * @see #clearExterior
	 */
	public void clear() {
		clearInterior();
		clearExterior();
	}

	public void fill (Color color) {
		Graphics g = drawInto.getGraphics();

		if ( g != null) {
			Rectangle r = getInnerBounds();
			g.setColor(color);
			g.fillRect(r.x, r.y, r.width, r.height);
			setFillColor(color);
		}
	}

	public String toString() {
		return (super.toString() + "[" + paramString() + "]");
	}

	public String paramString() {
		return "LineColor=" + getLineColor() + ",thickness=" +
			thick + ",fillColor=" + getFillColor();
	}

	/**
	 * Returns a color four shades brighter than the specified color
	 *
	 * @param color the color for which the brighter shade is required.
	 * @return the brighter shade
	 */

	protected Color brighter(Color color) {

		return color.brighter().brighter().brighter().brighter();

	}
}
