/*
 * @(#)EtchedRectangle.java	1.3 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package visualBenchmarks;

import java.awt.*;

/**
 * An extension of the visualBenchmarks.DrawnRectangle class, this class 
 * draws an <em>etched</em> rectangle.
 *
 * @see DrawnRectangle
 */

public class EtchedRectangle extends DrawnRectangle {

  	/**
	 * Constant for specifying Etched in borders
	 */
	public static int ETCHED_IN = 1;

	/**
	 * Constant for specifying Etched out borders
	 */
	public static int ETCHED_OUT = 2;

	/**
	 * The default etching for etched rectangles
 	 */
	protected static int  defaultEtching = ETCHED_IN;
	private int etching;

	//The etching effect is obtained by drawing the border rectangles in
	//one color and then drawing them slightly shifted to ther right in
	//another color (darker/brighter than the first). For an ethced-in
	//effect, the second time, the rectangles are drawn in a color 
	//brighter than the first. For an etched-out effect, they are drawn
	//brighter the *first* time around.
  
	private void paintEtched(Graphics g,Color topLeft, Color bottomRight){
	
		int i;
	
		int thick = getThickness();

		int w = width - thick;
		int h = height - thick;

		g.setColor(topLeft);
		
		for (i=0; i < thick/2 ; i++) {
			g.drawRect(x + i, y + i, w, h);
		}

		g.setColor(bottomRight);

		for (i=0; i < thick/2; i++) {
			g.drawRect(x + (thick/2) + i, y + (thick/2) + i,
				w, h);
		}

	}

	
	/**
	 * Constructs an Etched rectangle in a specified component with the
	 * default thickness for <i>DrawnRectangle</i>s, at location (0,0) 
	 * and with 0 height and width. 
	 * 
 	 * @param drawInto the component ,the rectangle is to be drawn into.
	 */

	//_default_thickness is a protected member inherited from 
  	//DrawnRectangle

	public EtchedRectangle(Component drawInto) {
		this (drawInto, defaultEtching, 
			defaultThickness, 0,0,0,0);
	}

	/**
	 * Constructs an etched in rectangle in a specified component with the
	 * specified thickness at location (0,0) and with 0 height and width
	 * 
 	 * @param drawInto the component ,the rectangle is to be drawn into.
	 * @param thick    the thickness for the border
	 */

	public EtchedRectangle(Component drawInto, int thick) {
		this (drawInto, defaultEtching, thick, 0,0,0,0);
	}

	/**
	 * Constructs an etched in rectangle in a specified component with the
	 * default thickness at location (x,y) and with the specifed 
	 * height and width
	 * 
 	 * @param drawInto the component ,the rectangle is to be drawn into.
	 * @param x,y      the co-ordinates for the top left corner of the 
	 * 		   rectangle
	 * @param w,h      the height and width of th rectangle.
	 */

	public EtchedRectangle(Component drawInto, int x, int y, int w, int h) {
		this (drawInto, defaultEtching, defaultThickness,
			x, y, w, h);
	}

	/**
	 * Constructs an etched in rectangle in a specified component with the
	 * specified thickness at location (x,y) and with the specifed 
	 * height and width
	 * 
 	 * @param drawInto the component ,the rectangle is to be drawn into.
	 * @param thick    the thickness for the border
	 * @param x,y      the co-ordinates for the top left corner of the 
	 * 		   rectangle
	 * @param w,h      the height and width of th rectangle.
	 */

	public EtchedRectangle(Component drawInto, int thick, int x, int y, 
			int w, int h) {
		this (drawInto, defaultEtching, thick, x, y, w, h);
	}

	/**
	 * Constructs an etched rectangle in a specified component with the
	 * specified thickness at location (x,y) and with the specifed 
	 * height and width. <em>Note:</em> Etched as specified.
	 * 
 	 * @param drawInto the component ,the rectangle is to be drawn into.
	 * @param etching  ETCHED_IN or ETCHED_OUT
	 * @param thick    the thickness for the border
	 * @param x,y      the co-ordinates for the top left corner of the 
	 * 		   rectangle
	 * @param w,h      the height and width of th rectangle.
	 *
	 * @see #ETCHED_IN
	 * @see #ETCHED_OUT
	 */

	public EtchedRectangle(Component drawInto, int etching, int thick, 
			int x, int y, int w, int h) {
		super (drawInto, thick, x, y, w, h);
		this.etching = etching;
	}

	/**
	 * Draw the rectangle etched in.
	 */

	public void etchedIn() {
		etching = ETCHED_IN;
	}

	/**
	 * Draw the rectangle etched out.
	 */
	public void etchedOut() {
		etching = ETCHED_OUT;
	}

	public boolean isEtchedIn() {
		return (etching == ETCHED_IN);
	}

	public void paint() {

		if (etching == ETCHED_IN) 
			paintEtchedIn();
		else
			paintEtchedOut();
	}

	/**
	 * Paint the rectangle with etched in borders
	 */
	public void paintEtchedIn() {
		Graphics g = drawInto.getGraphics();
	
		if (g != null) {
			paintEtched(g,getLineColor(),brighter(getLineColor()));
		}

		etchedIn();
	}

	/**
	 * Paint the rectangle with etched out borders
	 */
	public void paintEtchedOut(){
		Graphics g = drawInto.getGraphics();

		if (g != null) {
			paintEtched (g,brighter(getLineColor()),getLineColor());
		}

		etchedOut();
	}

	public String paramString() {
		return super.paramString() + "," + etching;
	}

}




