/*
 * @(#)CRLTest.java	1.9 98/07/10
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

import java.io.*;
import java.math.BigInteger;
import java.util.Calendar;
import java.util.Date;
import java.util.Iterator;
import java.util.Set;
import java.util.TimeZone;
import java.security.cert.X509CRLEntry;
import java.security.cert.CRLException;
import java.security.cert.X509CRL;
import java.security.cert.CertificateFactory;
import java.security.KeyPairGenerator;
import java.security.SecureRandom;
import java.security.KeyPair;
import java.security.PrivateKey;
import java.security.MessageDigest;
import sun.security.x509.*;
import sun.security.util.*;
import sun.misc.HexDumpEncoder;

/**
 * Usage: CRLTest <key_algorithm> <signature_algorithm>
 * E.g. CRLTest DSA SHAwithDSA
 */
public class CRLTest {

    private static void printSet(Set set, Object obj)
    throws CRLException {
        for (Iterator i = set.iterator(); i.hasNext();) {
            String attr = (String)i.next();
            System.out.println(attr);
            byte[] extData = null;
            if (obj instanceof X509CRL)
                extData = ((X509CRL)obj).getExtensionValue(attr);
            else if (obj instanceof X509CRLEntry)
                extData = ((X509CRLEntry)obj).getExtensionValue(attr);
            if (extData != null) {
                HexDumpEncoder encoder = new HexDumpEncoder();
                System.out.print(encoder.encodeBuffer(extData));
            }
        }
    }

    public static void main(String[] args) {
        X500Name	issuer;
        KeyPairGenerator keyGen;
        String		sigAlg;
        SecureRandom	prng;
        X509Key		publicKey;
        PrivateKey	privateKey;
        KeyPair		pair;
        String          fileName = "testCRL";

        if (args.length != 2) {
            System.err.println("Usage: CRLTest <key_alg> <sig_alg>");
            System.exit(1);
        }
        try {
            issuer = new X500Name("Dr Duke", "JavaSoft Division, SMI",
                                  "Palo Alto", "US");
            keyGen = KeyPairGenerator.getInstance(args[0]);
            sigAlg = args[1];

            prng = new SecureRandom();
            keyGen.initialize(1024, prng);
            pair = keyGen.generateKeyPair();
            publicKey = (X509Key)pair.getPublic();
            privateKey = pair.getPrivate();

            MessageDigest md = MessageDigest.getInstance("SHA");
            byte[] keyId = md.digest(publicKey.getEncoded());

            Date thisDate = new Date();
            Calendar cal = Calendar.getInstance(TimeZone.getTimeZone("PST"));
            cal.set(1999, 11, 15, 12, 30, 30);
            Date nextDate = cal.getTime();

            BigInteger serialNum = (new BigInt(40)).toBigInteger();
            X509CRLEntryImpl rCert = new X509CRLEntryImpl(serialNum, thisDate);
            X509CRLEntry[] badCerts = new X509CRLEntry[3];
            badCerts[0] = (X509CRLEntry)rCert;

            CRLExtensions crlEntryExts = new CRLExtensions();
            CRLReasonCodeExtension crlCode = new CRLReasonCodeExtension(2);
            crlEntryExts.set(crlCode.NAME, crlCode);
            serialNum = (new BigInt(60)).toBigInteger();
            rCert = new X509CRLEntryImpl(serialNum, thisDate, crlEntryExts);
            badCerts[1] = (X509CRLEntry)rCert;

            crlEntryExts = new CRLExtensions();
            crlCode = new CRLReasonCodeExtension(4);
            crlEntryExts.set(crlCode.NAME, crlCode);
            serialNum = (new BigInt(80)).toBigInteger();
            rCert = new X509CRLEntryImpl(serialNum, thisDate, crlEntryExts);
            badCerts[2] = (X509CRLEntry)rCert;

            CRLExtensions crlExts = new CRLExtensions();
            URIName uri = new URIName("http://www.javasoft.com");
            byte[] address = new byte[4];
            address[0] = 127; address[1] = 0; address[2] = 0;
            address[3] = 1;
            IPAddressName ip = new IPAddressName(address);
            IssuerAlternativeNameExtension issuerName =
                                new IssuerAlternativeNameExtension();
            GeneralNames issuerNames = (GeneralNames)
            issuerName.get(IssuerAlternativeNameExtension.ISSUER_NAME);

            issuerNames.addElement(ip);
            issuerNames.addElement(uri);

            KeyIdentifier kid = new KeyIdentifier(keyId);
            SerialNumber sn = new SerialNumber(42);
            AuthorityKeyIdentifierExtension aki =
                new AuthorityKeyIdentifierExtension(kid, issuerNames, sn);

            crlExts.set(aki.NAME, aki);

            // CRL Info
            X509CRLImpl crl = new X509CRLImpl(issuer, thisDate, nextDate,
                                              badCerts, crlExts);

            crl.sign(privateKey, sigAlg);
            crl.verify(publicKey);

            File foo = new File(fileName);
            FileOutputStream fos = new FileOutputStream(foo);
            fos.write(crl.getEncoded());
            fos.close();

            System.out.println("*** CRL ***");
            System.out.println(crl);
            System.out.println("*** End CRL ***");

            // check concrete methods
            if (((X509CRLEntry)badCerts[0]).equals(
                 ((X509CRLEntry)badCerts[1])))
                throw new Exception("Equals broken");
            else
                System.out.println("Equals okay .....................");

            // check concrete methods
            if (((X509CRLEntry)badCerts[0]).equals(
                 ((X509CRLEntry)badCerts[0])))
                System.out.println("Equals okay .....................");
            else
                throw new Exception("Equals broken");

            int hash1 =  ((X509CRLEntry)badCerts[1]).hashCode();
            int hash2 =  ((X509CRLEntry)badCerts[2]).hashCode();
            if (hash1 == hash2)
                throw new Exception("HashCode broken");
            else
                System.out.println("HashCode check okay, hash1=" + hash1
                                   + "; hash2=" + hash2);

            hash2 =  ((X509CRLEntry)badCerts[1]).hashCode();
            if (hash1 == hash2)
                System.out.println("HashCode check okay, hash1=" + hash1
                                   + "; hash2=" + hash2);
            else
                throw new Exception("HashCode broken");

            // extensions
            ByteArrayOutputStream boas = new ByteArrayOutputStream();
            crlExts.encode(boas, true);
            CRLExtensions testExt = new CRLExtensions();
            ByteArrayInputStream bais = new ByteArrayInputStream(
                                        boas.toByteArray());
            testExt.decode(bais);
            if (! crlExts.equals(testExt))
                throw new Exception("Decoded extension does not match\nOrig: "
                    + crlExts.toString() + "\nDecoded: " + testExt.toString());
            System.out.println("Encoded and decoded extensions match");
            hash1 = crlExts.hashCode();
            hash2 = testExt.hashCode();
            if (hash1 != hash2)
                throw new Exception("HashCode broken");
            
            // ------------- verify parsing -------------------
            InputStream inStrm = new FileInputStream(fileName);
            CertificateFactory cf = CertificateFactory.getInstance("X509");
            X509CRL x2 = (X509CRL)cf.generateCRL(inStrm);
            inStrm.close();

            if (!x2.equals(crl))
                throw new Exception("*** CRL mismatch ***");

            System.out.println("*** CRL match ***");
            System.out.println("Issued by: " + x2.getIssuerDN());
            System.out.println("Issued on: " + x2.getThisUpdate());
            System.out.println("Next issue on: " + x2.getNextUpdate());

            Set revCerts = x2.getRevokedCertificates();
            if (revCerts != null && !revCerts.isEmpty()) {
                X509CRLEntry rc;
                int cnt=1;
                for (Iterator i = revCerts.iterator(); i.hasNext(); ) {
                    rc = (X509CRLEntry)i.next();
                    System.out.println("[" + cnt++ + "]:" + rc.toString()); 
                    if (rc.hasExtensions()) {
                        Set critSet = rc.getCriticalExtensionOIDs();
                        if (critSet != null && !critSet.isEmpty()) {
                            System.out.print("Critical Entry extensions: ");
                            printSet(critSet, rc);
                        }
                        Set nonCritSet = rc.getNonCriticalExtensionOIDs();
                        if (nonCritSet != null && !nonCritSet.isEmpty()) {
                            System.out.print("Non-critical Entry extensions: ");
                            printSet(nonCritSet, rc);
                        }
                    }
                }
            }
            Set critSet = x2.getCriticalExtensionOIDs();
            if (critSet != null && !critSet.isEmpty()) {
                System.out.print("Critical extensions: ");
                printSet(critSet, x2);
            }

            Set nonCritSet = x2.getNonCriticalExtensionOIDs();
            if (nonCritSet != null && !nonCritSet.isEmpty()) {
                System.out.print("Non-critical extensions: ");
                printSet(nonCritSet, x2);
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
    }
}
