/*
 * @(#)CertParse.java	1.18 98/06/29
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

import java.io.*;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.security.cert.CertificateException;
import java.security.cert.CertificateExpiredException;
import java.security.*;
import sun.misc.HexDumpEncoder;
import sun.security.x509.*;

/**
 * This is a class used to test X.509 cert parsing and validation.
 * It works on cert changes as well as individual certs.
 */
public class CertParse
{
    private static void error (boolean allowable)
    {
	if (!allowable) {
	    System.out.println (" !!  TEST FAILED !! ");
	    System.exit (1);
	}
    }

    /*
     * CertParse cert1 [cert2 cert3 ... ] pass
     *		The test should pass.  The last cert is self-signed,
     *		and it's the "root CA" in a cert chain starting at
     *		the first cert.
     *
     * CertParse cert1 [cert2 cert3 ... ] fail
     *		As above, but the test should fail.  Perhaps one
     *		of the certs in the chain is invalid, or the chain
     *		is incomplete.
     */
    public static void main (String argv []) {
	boolean		isNegative = false;
	int		i;
	PublicKey	key = null;
	X500Name	issuer = null;

	if (argv.length < 2) {
	    System.out.println (
		"usage:  CertParse cert [cert ...] pass|fail");
	    error (false);
	}
	if (argv [argv.length - 1].equalsIgnoreCase ("fail"))
	    isNegative = true;

	//
	// Process certs starting at root CA.
	//
	for (i = argv.length - 2; !(i < 0); i--) {
	    String certFile = argv [i];

	    byte	certdata[] = null;

	    try {
		//
		// Read the cert in from the file.
		//
		{
		    InputStream	is;
		    int		len;

		    is = new FileInputStream(certFile);
		    len = is.available();
		    certdata = new byte[len];

                    int firstLen = is.read(certdata);
		    if (firstLen != len) {
                        // data is read in 2048 byte chunks according
                        // to native code so lets try one more read ...
                        int left = len - firstLen;
                        if (left > 2048) {
			    System.out.println("** short read, sorry");
			    error (isNegative);
			    continue;
                        }
                        int scndLen = is.read(certdata, firstLen, left);

		        if ((scndLen+firstLen) != len) {
			    error (isNegative);
			    continue;
                       }
		    }
		    is.close();
		}

		ByteArrayInputStream bais = new ByteArrayInputStream(certdata);
		CertificateFactory cf = CertificateFactory.getInstance("X509");
		X509Certificate cert
		    = (X509Certificate)cf.generateCertificate(bais);
		bais.close();

		System.out.println ("-----------------------");
		System.out.println ("Cert File = " + certFile);
		System.out.println (cert);

		//
		// Bootstrap chain processing (topmost cert in chain).
		//

		// assume self-signed
		if (key == null) {
		    key = cert.getPublicKey();
		    issuer = (X500Name)cert.getSubjectDN();
		}

		//
                // Verify the cert
		//
		try {
                    X500Name tmpDN = (X500Name)cert.getIssuerDN();

		    if (!tmpDN.equals(issuer)) {
			System.out.println ("Cert Issuer Mismatch:  "
				+ tmpDN + " expected, got " + issuer);
			error (isNegative);
			continue;
		    }
		    cert.verify(key);
		    System.out.println ("  ++ Cert verified");

		    // Validate the cert, recognizing expired ones.
                    cert.checkValidity();
		    System.out.println ("  ++ Cert is current");

		    //
		    // Set up for next cert in chain
		    //
		    key = cert.getPublicKey();
		    issuer = (X500Name)cert.getSubjectDN();

		} catch (CertificateException e) {
		    if (e instanceof CertificateExpiredException)
			System.out.println("  -- Cert expired");
		    else {
			System.out.println("  -- Cert doesn't verify!");
			e.printStackTrace();
		    }
		    error(isNegative);

		} catch (Exception e) {
		    System.out.println ("  -- Error during verify!");
		    e.printStackTrace ();
		    error (false);
		}

	    } catch (Exception e) {
		System.out.println ("** Exception:  " + e);
		e.printStackTrace ();
                // hmmm lets dump the certdata we have to a file
                try {
                    File foo = new File("failedCert");
                    FileOutputStream fos = new FileOutputStream(foo);
                    fos.write(certdata);
                    fos.close();
                } catch (IOException ex) {
                    System.out.println ("** Exception:  " + ex);
                    ex.printStackTrace ();
                }

                // hmmm lets dump the certdata
		//HexDumpEncoder encoder = new HexDumpEncoder();
		//System.out.print(encoder.encodeBuffer(certdata));
		error (false);
	    }
	}

	System.out.println ("PASS !!!");
	System.out.println ("");
    }
}
