/*
 * @(#)SimpleSigner.java	1.3 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
import java.security.*;
import java.security.cert.*;
import sun.security.x509.*;
import sun.security.pkcs.*;
import sun.security.util.*;
import java.util.Date;

/** 
 *  A simple extension of sun.security.x509.X500Signer that adds a
 *  no-fuss signing algorithm.  Except X500Signer is final, so we have
 *  to extend it "manually."  
 */

public class SimpleSigner {

    private Signature	sig;
    private X500Name	agent;		// XXX should be X509CertChain
    private AlgorithmId	digestAlgId;
    private AlgorithmId	encryptionAlgId;
    private AlgorithmId	algId;          // signature algid; combines digest + encryptn 
    private X509Key publicKey;
    private PrivateKey privateKey;
    private X509Certificate cert;
    
    public SimpleSigner(String digestAlg,
			String encryptionAlg,
			KeyPair keyPair, 
			X500Name agent) throws Exception {

	if (agent == null) {
	     agent = new X500Name("Uncle Duke","Acme Corp.","Palo Alto","US");
	}
	if (digestAlg == null) {
	    digestAlg = "SHA";
	}
	if (encryptionAlg == null) {
	    encryptionAlg = "DSA";
	}
 	if (keyPair == null) {
	    SecureRandom prng = new SecureRandom();
	    KeyPairGenerator keyGen = KeyPairGenerator.getInstance(encryptionAlg);
	    keyGen.initialize(512, prng);
	    keyPair = keyGen.generateKeyPair();      
	}
	publicKey = (X509Key) keyPair.getPublic();      
	privateKey = keyPair.getPrivate ();
	
	// the next line should be as follows, maybe
	//	this.sig = Signature.getInstance(digestAlg + "/" + encryptionAlg);
	if (encryptionAlg == "DSA") {
	    this.sig = Signature.getInstance(encryptionAlg);
	} else { // RSA
	    this.sig = Signature.getInstance(digestAlg + "/" + encryptionAlg);
	}
	this.sig.initSign(privateKey);
	
	this.agent = agent;
	this.digestAlgId = AlgorithmId.get(digestAlg);
	this.encryptionAlgId = AlgorithmId.get(encryptionAlg);
	this.algId = AlgorithmId.get(this.sig.getAlgorithm());

	this.cert = getSelfCert(1000);
    }

    /** 
     * Take the data and sign it.
     *
     * @param buf buffer holding the next chunk of the data to be signed
     * @param offset starting point of to-be-signed data
     * @param len how many bytes of data are to be signed
     * @return the signature for the input data.
     * @exception SignatureException on errors.
     */
    public byte[] simpleSign(byte[] buf, int offset, int len) 
    throws SignatureException {
	sig.update(buf, offset, len);
	return sig.sign();
    }

    /**
     * Returns the digest algorithm used to sign.
     */
    public AlgorithmId 	getDigestAlgId() {
        return digestAlgId;
    }

    /**
     * Returns the encryption algorithm used to sign.
     */
    public AlgorithmId 	getEncryptionAlgId() {
        return encryptionAlgId;
    }

    /**
     * Returns the name of the signing agent.
     */
    public X500Name	getSigner() {
        return agent;
    }

    public X509Certificate getCert() {
	return cert;
    }

    private X509Certificate	getSelfCert (long validity)
    throws CertificateException, InvalidKeyException, SignatureException,
        NoSuchAlgorithmException
    {
	X509CertImpl	cert;
	Date		firstDate, lastDate;
	
	try {
	    firstDate = new Date ();
	    lastDate = new Date ();
	    lastDate.setTime (lastDate.getTime () + validity * 1000);

            CertificateValidity interval =
                                   new CertificateValidity(firstDate,lastDate);
 
            X509CertInfo info = new X509CertInfo();
            // Add all mandatory attributes
            info.set(X509CertInfo.VERSION,
                     new CertificateVersion(CertificateVersion.V1));
            info.set(X509CertInfo.SERIAL_NUMBER,
                 new CertificateSerialNumber((int)(firstDate.getTime()/1000)));
            info.set(X509CertInfo.ALGORITHM_ID,
                     new CertificateAlgorithmId(algId));
            info.set(X509CertInfo.SUBJECT, new CertificateSubjectName(agent));
            info.set(X509CertInfo.KEY, new CertificateX509Key(publicKey));
            info.set(X509CertInfo.VALIDITY, interval);
            info.set(X509CertInfo.ISSUER,
                     new CertificateIssuerName(agent));

	    cert = new X509CertImpl(info);
	    cert.sign(privateKey, algId.getName());

	    return (X509Certificate)cert;

	} catch (java.io.IOException e) {
	    throw new CertificateException(" getSelfCert");
	} catch (java.security.NoSuchProviderException e) {
	    throw new CertificateException("No such Provider");
	}
    }

    public SignerInfo genSignerInfo(byte[] data)
	throws SignatureException, CertificateException, NoSuchAlgorithmException, 
	    InvalidKeyException {
	    return new SignerInfo((X500Name) cert.getIssuerDN(), 
				  new BigInt(cert.getSerialNumber()), 
				  getDigestAlgId(), algId, 
				  simpleSign(data, 0, data.length));
    }
}
