/*
 * @(#)V1toV3Cert.java	1.9 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

import java.io.*;
import java.util.*;
import java.math.*;
import java.security.*;
import sun.security.x509.*;
import sun.security.util.*;
import sun.misc.*;

/**
 * Classic example of taking an existing v1 certificate and
 * generating a v3 certificate.
 * Usage: V1toV3Cert <certFile> <key_algorithm> <signature_algorithm>
 * E.g. V1toV3Cert dave.cer DSA "SHA/DSA"
 */
public class V1toV3Cert {
    public static void main(String[] args) {
        byte[]	issuerId = { 1, 2, 3, 4, 5 };
        byte[]	subjectId = { 6, 7, 8, 9, 10 };
        byte[]	subjectKey = { 12, 13, 14, 15, 16, 17 };
        byte[]	generalData = { 9, 18, 27, 36, 45, 54, 63, 72, 81 };
        X500Name	issuer;
        KeyPairGenerator keyGen;
        String		sigAlg;
        SecureRandom	prng;
        X509Key		publicKey;
        PrivateKey	privateKey;
        KeyPair		pair;
        X509CertImpl	orig, cert;

        if (args.length != 3) {
            System.out.println("Usage: V1toV3Cert <certificate>" +
                               " <keyAlg> <sigAlg>");
            System.exit(1);
        }
        try {
            InputStream inStrm = new FileInputStream(args[0]);
                                           // FQN
            orig = new X509CertImpl(inStrm);
            inStrm.close();
            System.out.println("*** Original Certificate ***");
            System.out.println(orig);
            System.out.println("*** End Certificate ***");

            X509CertInfo certInfo = new X509CertInfo(orig.getTBSCertificate());
            cert = new X509CertImpl(certInfo);

            issuer = new X500Name("Dr Duke","ACME Corp.","Palo Alto","US");
            keyGen = KeyPairGenerator.getInstance(args[1]);
            sigAlg = args[2];
      
            prng = new SecureRandom();
            keyGen.initialize(1024, prng);
            pair = keyGen.generateKeyPair();      
            publicKey = (X509Key)pair.getPublic();      
            privateKey = pair.getPrivate();

            Date firstDate = new Date();
            Date lastDate = new Date();
            lastDate.setTime(lastDate.getTime() * 2);
            CertificateValidity interval = new CertificateValidity(firstDate,
                                                                   lastDate);

            // Reset all mandatory attributes
            cert.set(CertificateVersion.IDENT,
                     new CertificateVersion(CertificateVersion.V3));
            cert.set(CertificateSerialNumber.IDENT,
		new CertificateSerialNumber((int)(firstDate.getTime()/1000)));
            cert.set(CertificateAlgorithmId.IDENT,
                     new CertificateAlgorithmId(AlgorithmId.get(sigAlg)));

	    // don't reset the subject name

            cert.set(CertificateX509Key.IDENT,
			      new CertificateX509Key(publicKey));
            cert.set(CertificateValidity.IDENT, interval);
            cert.set(CertificateIssuerName.IDENT,
			      new CertificateIssuerName(issuer));

            // Version 2 stuff
            cert.set(CertificateIssuerUniqueIdentity.IDENT,
	    new CertificateIssuerUniqueIdentity(new UniqueIdentity(issuerId)));
            cert.set(CertificateSubjectUniqueIdentity.IDENT,
          new CertificateSubjectUniqueIdentity(new UniqueIdentity(subjectId)));

        // Create Extensions
        CertificateExtensions exts = new CertificateExtensions();
        RFC822Name mail = new RFC822Name("duke@acme.com");

        SubjectAlternativeNameExtension subjectName =
                                new SubjectAlternativeNameExtension();
        GeneralNames subjectNames = 
	(GeneralNames)subjectName.get(SubjectAlternativeNameExtension.SUBJECT_NAME);
      subjectNames.addElement(mail);

      exts.set(SubjectAlternativeNameExtension.NAME, subjectName);
      cert.set(CertificateExtensions.IDENT, exts);

      cert.sign(privateKey, sigAlg);
      cert.verify(publicKey);

      System.out.println("*** Certificate ***");
      System.out.println(cert);
      System.out.println("*** End Certificate ***");

      File foo = new File("revisedCert");
      FileOutputStream fos = new FileOutputStream(foo);
      cert.encode((OutputStream) fos);
      fos.close();

    } catch (Exception e) {
      e.printStackTrace();
    }
  } 
}
