/*
 * @(#)V3Certificate.java	1.14 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

import java.io.*;
import java.util.Calendar;
import java.util.Date;
import java.util.TimeZone;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.security.*;
import sun.security.x509.*;
import sun.security.util.*;
import sun.misc.BASE64Encoder;

/**
 * Usage: V3Certificate <key_algorithm> <signature_algorithm>
 * E.g. V3Certificate DSA SHAwithDSA
 */
public class V3Certificate {
    public static void main(String[] args) {
        byte[]	issuerId = { 1, 2, 3, 4, 5 };
        byte[]	subjectId = { 6, 7, 8, 9, 10 };
        X500Name	subject, issuer;
        KeyPairGenerator keyGen;
        String		sigAlg;
        SecureRandom	prng;
        X509Key		publicKey;
        PrivateKey	privateKey;
        KeyPair		pair;

        if (args.length != 2) {
            System.err.println("Usage: V3Certificate <key_alg> <sig_alg>");
            System.exit(1);
        }
        try {
            subject = new X500Name("Dr Duke","JavaSoft Inc.","Palo Alto","US");
            issuer = subject;
            keyGen = KeyPairGenerator.getInstance(args[0]);
            sigAlg = args[1];

            prng = new SecureRandom();
            keyGen.initialize(1024, prng);
            pair = keyGen.generateKeyPair();
            publicKey = (X509Key)pair.getPublic();
            privateKey = pair.getPrivate();

            MessageDigest md = MessageDigest.getInstance("SHA");
            byte[] keyId = md.digest(publicKey.getEncoded());

            Signature	signature = Signature.getInstance(sigAlg);
            signature.initSign(privateKey);

            Date firstDate = new Date();
            Calendar cal = Calendar.getInstance(TimeZone.getTimeZone("PST"));
            cal.set(1999, 11, 15, 12, 30, 30);
            Date lastDate = cal.getTime();
            CertificateValidity interval = new CertificateValidity(firstDate,
                                                                   lastDate);

            // Certificate Info
            X509CertInfo cert = new X509CertInfo();

            // Add all mandatory attributes
            cert.set(X509CertInfo.VERSION,
                              new CertificateVersion(CertificateVersion.V3));
            cert.set(X509CertInfo.SERIAL_NUMBER,
		new CertificateSerialNumber((int)(firstDate.getTime()/1000)));
            cert.set(X509CertInfo.ALGORITHM_ID,
                     new CertificateAlgorithmId(AlgorithmId.get(sigAlg)));
            cert.set(X509CertInfo.SUBJECT,
			      new CertificateSubjectName(subject));
            cert.set(X509CertInfo.KEY,
			      new CertificateX509Key(publicKey));
            cert.set(X509CertInfo.VALIDITY, interval);
            cert.set(X509CertInfo.ISSUER,
			      new CertificateIssuerName(issuer));

            // Version 2 stuff
            cert.set(X509CertInfo.ISSUER_ID,
                     new CertificateIssuerUniqueIdentity(
                         new UniqueIdentity(
			     new BitArray(issuerId.length*8 - 2, issuerId))));
            cert.set(X509CertInfo.SUBJECT_ID,
                     new CertificateSubjectUniqueIdentity(
                         new UniqueIdentity(subjectId)));

            // Create Extensions
            CertificateExtensions exts = new CertificateExtensions();

            RFC822Name mail = new RFC822Name("duke@javasoft.com");
            DNSName dns = new DNSName("javasoft.com");
            URIName uri = new URIName("http://www.javasoft.com");

            byte[] address = new byte[4];
            address[0] = 127; address[1] = 0; address[2] = 0; address[3] = 1;
            IPAddressName ip = new IPAddressName(address);

            int[] oidData = new int[4];
            oidData[0] = 1; oidData[1] = 2; oidData[2] = 3; oidData[3] = 4;
            OIDName oid = new OIDName(new ObjectIdentifier(oidData));

            SubjectAlternativeNameExtension subjectName =
                                new SubjectAlternativeNameExtension();
            IssuerAlternativeNameExtension issuerName =
                                new IssuerAlternativeNameExtension();

            GeneralNames subjectNames = (GeneralNames)
            subjectName.get(SubjectAlternativeNameExtension.SUBJECT_NAME);

            GeneralNames issuerNames = (GeneralNames)
            issuerName.get(IssuerAlternativeNameExtension.ISSUER_NAME);

            subjectNames.addElement(mail);
            subjectNames.addElement(dns);
            subjectNames.addElement(uri);

            issuerNames.addElement(ip);
            issuerNames.addElement(oid);

            cal.set(2000, 11, 15, 12, 30, 30);
            lastDate = cal.getTime();
            PrivateKeyUsageExtension pkusage =
                new PrivateKeyUsageExtension(firstDate, lastDate);

            KeyUsageExtension usage = new KeyUsageExtension();
            usage.set(KeyUsageExtension.CRL_SIGN, new Boolean(true));
            usage.set(KeyUsageExtension.DIGITAL_SIGNATURE, new Boolean(true));
            usage.set(KeyUsageExtension.NON_REPUDIATION, new Boolean(true));

            KeyIdentifier kid = new KeyIdentifier(keyId);
            SerialNumber sn = new SerialNumber(42);
            AuthorityKeyIdentifierExtension aki = new
                AuthorityKeyIdentifierExtension(kid, subjectNames, sn);

            SubjectKeyIdentifierExtension ski = new
                SubjectKeyIdentifierExtension(keyId);

            BasicConstraintsExtension cons = new
                             BasicConstraintsExtension(true, 10);

            PolicyConstraintsExtension pce = new
                             PolicyConstraintsExtension(2, 4);

            exts.set(subjectName.NAME, subjectName);
            exts.set(issuerName.NAME, issuerName);
            exts.set(pkusage.NAME, pkusage);
            exts.set(usage.NAME, usage);
            exts.set(aki.NAME, aki);
            exts.set(ski.NAME, ski);
            exts.set(cons.NAME, cons);
            exts.set(pce.NAME, pce);
            cert.set(X509CertInfo.EXTENSIONS, exts);

            X509CertImpl crt = new X509CertImpl(cert);
            crt.sign(privateKey, sigAlg);
            crt.verify(publicKey);

            FileOutputStream fos = new FileOutputStream(new File("sunV3"));
            crt.encode((OutputStream) fos);
            fos.close();

            // Certificate boundaries
            String beginCert = "-----BEGIN CERTIFICATE-----";
            String endCert = "-----END CERTIFICATE-----";
            BASE64Encoder encoder = new BASE64Encoder();
            fos = new FileOutputStream(new File("sunV3.b64"));
            PrintWriter pw = new PrintWriter(fos);
            pw.println(beginCert);
            pw.flush();
            encoder.encodeBuffer(crt.getEncoded(), fos);
            fos.flush();
            pw.println(endCert);
            pw.close();

            System.out.println("*** Certificate ***");
            System.out.println(crt);
            System.out.println("*** End Certificate ***");

            try {
                InputStream inStrm = new FileInputStream((String) "sunV3");
		CertificateFactory cf = CertificateFactory.getInstance("X509");
                X509Certificate x2
		    = (X509Certificate)cf.generateCertificate(inStrm);
                inStrm.close();

                if (!x2.equals(crt)) {
	            System.out.println("*** Certificate mismatch ***");
                }
                System.out.println("*** Certificate match ***");
            } catch (Exception e) {
                e.printStackTrace();
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
    }
}
