/*
 * @(#)EchoArray.java	1.2 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
import java.net.*;
import java.io.*;
import java.util.Date;

/*
 * Simple class to echo array objects.  It creates a thread that accepts
 * new connections and echos all objects that it receives.
 * Command line arguments are parsed as classes to be echoed.
 */
class EchoArray implements Runnable {
    String host = null;
    int iterations = 1;
    static boolean verbose = false;
    static boolean compare = false;
    static ServerSocket serversocket = null;
    ObjectOutputStream os;
    ObjectInputStream is;

    Date start;
    Date end;

    /* The types and sizes of arrays to test */
    static char[] arraytypes = { 'B', 'S', 'I', 'L', 'Z', 'F', 'D', 'C' };
    static int[] arraysizes = { 0, 1, 2, 32,
				 63, 64, 65,
				 127, 128, 129,
				 255, 256, 257,
				 1023, 1024, 1025,
				 2048, 4096, 8192,
				 10, 100, 1000, 10000 };

    public static void main(String args[]) {

	EchoArray echo = new EchoArray();
	echo.start(args);
    }

    private void start(String[] args) {
	int nextarg;
	for (nextarg = 0; nextarg < args.length; nextarg++) {
	    if (args[nextarg].equals("-r")) {
		if ((nextarg + 1) < args.length) {
		    iterations = Integer.parseInt(args[++nextarg]);
		} else {
		    System.out.println("-r requires a repeat count");
		    usage();
		    return;
		}
	    } else if (args[nextarg].equals("-h")) {
		if ((nextarg + 1) < args.length) {
		    host = args[++nextarg];
		} else {
		    System.out.println("-h requires a hostname");
		    usage();
		    return;
		}
	    } else if (args[nextarg].equals("-v"))
		verbose = true;
	    else if (args[nextarg].equals("-c"))
		compare = true;
	    else if (args[nextarg].startsWith("-")) {
		System.out.println("invalid flag: " + args[nextarg]);
		usage();
		return;
	    } else 
		break;		// rest of args are classes to test
	}

	
	try {
	    if (host == null) {
		serversocket = new ServerSocket(5066);
		new Thread(new EchoArray()).start();
	    }

	    // Open a socket to the EchoObj server
	    Socket client = null;
	    if (host != null) {
		client = new Socket(host, 5066);
	    } else {
		client = new Socket("", 5066);
	    }

	    os = new ObjectOutputStream(client.getOutputStream());
	    os.flush();
	    is = new ObjectInputStream(client.getInputStream());

	    // For each classname on the command line
	    // create an instance, send it , receive it and print each.
	    if (nextarg <args.length) {
		for(;nextarg < args.length; nextarg++) {
		    Object o = makeArrayFromArg(args[nextarg]);
		    echoObj(o);
		    System.out.println(args[nextarg] + ":\t" + this);

		}
	    } else {
		for (int type = 0; type < arraytypes.length; type++) {
		    for (int size = 0; size < arraysizes.length; size++) {
			Object o = makeArray(arraytypes[type], arraysizes[size]);
			echoObj(o);
			System.out.println(arraytypes[type] + "[" + arraysizes[size] + "]:\t" + this);
		    }
		}
	    }
	} catch (Exception e) {
	    e.printStackTrace();
	}
    }

    private void echoObj(Object o) throws Exception {
	start = new Date();
	for (int i = 0; i < iterations; i++) {
	    if (verbose) System.out.println("Writing " + o);
	    os.writeObject(o);
	    os.flush();
	    Object robj = is.readObject();
	    if (compare)
		compare(o, robj, System.out);
	    if (verbose) System.out.println("Reading " + o);
	}
	end = new Date();
    }

    private Object makeArrayFromArg(String arg) throws Exception {
	Object o = null;
	char type = arg.charAt(0);
	int size = Integer.parseInt(arg.substring(1));
System.out.println("making array of type " + type + ", size = " + size);

	return makeArray(type,size);
    }

    /**
     * Wait for a connection, when one is received create a new
     * EchoArray to handle the next accept and start echoing objects.
     */
    public void run() {
	try {
	    Socket s = serversocket.accept();
	    if (verbose) System.out.println("Starting new accept thread");
	    new Thread(new EchoArray()).start();

	    ObjectOutputStream out = new ObjectOutputStream(s.getOutputStream());
	    out.flush();
	    ObjectInputStream in = new ObjectInputStream(s.getInputStream());
	    int count = 1;
	    while (true) {
		Object o = in.readObject();
		if (verbose) System.out.println("Echo " + count++ + ": " + o);
		out.writeObject(o);
		out.flush();
	    }
	} catch (EOFException e) {
	    return;
	} catch (IOException e) {
	    e.printStackTrace();
	} catch (ClassNotFoundException e2) {
	    e2.printStackTrace();
	}
    }

    private Object makeArray(char type, int size) {
	switch (type) {
	case 'B':
	    byte[] bytearray = new byte[size];
	    for (int i = 0; i < size; i++)
		bytearray[i] = (byte)i;
	    return bytearray;

	case 'S':
	    short[] shortarray = new short[size];
	    for (int i = 0; i < size; i++)
		shortarray[i] = (short)i;
	    return shortarray;

	case 'I':
	    int[] intarray = new int[size];
	    for (int i = 0; i < size; i++)
		intarray[i] = (int)i;
	    return intarray;

	case 'L':
	    long[] longarray = new long[size];
	    for (int i = 0; i < size; i++)
		longarray[i] = 0x07060504030201L;
	    return longarray;

	case 'F':
	    float[] floatarray = new float[size];
	    for (int i = 0; i < size; i++)
		floatarray[i] = 3.14159827f;
	    return floatarray;

	case 'D':
	    double[] doublearray = new double[size];
	    for (int i = 0; i < size; i++)
		doublearray[i] = 3.14159827 * 2.0;
	    return doublearray;

	case 'Z':
	    boolean[] booleanarray = new boolean[size];
	    for (int i = 0; i < size; i++)
		booleanarray[i] =  (i & 1) == 0;
	    return booleanarray;

	case 'C':
	    char[] chararray = new char[size];
	    for (int i = 0; i < size; i++)
		chararray[i] = (char)i;
	    return chararray;

	default:
	    throw new IllegalArgumentException();
	}
    }
    private static void usage() {
	System.out.println("arguments are: [-r <iterations>] [-h <hostname>] [-v] classnames");
	System.out.println("\t-r n\trepeat for n iterations");
	System.out.println("\t-h\thostname to echo classes on");
	System.out.println("\t-v\tverbose");
    }

    public String toString() {
	long duration = (start == null || end == null ) ?
	    0 :
	    end.getTime() - start.getTime();
	return duration + " ms,\t(" + (iterations * 1000.0)/duration + " per sec)";
    }

    /*
     * Methods to compare arrays for equality
     */
    public static boolean compare(int a[], int b[], PrintStream out) {
	boolean result = true;
	
	if (a.length != b.length) {
	    out.println("Int array lengths differ: " +
			       a.length + " != " + b.length);
	    result = false;
	}
	
	for (int i = 0; i < a.length; i++){
	    if (a[i] != b[i] ) {
		out.println("Int array contents differ at index " + i + ": " +
			    a[i] + " != " + b[i]);
		result = false;
	    }
	}
	return result;
    }
    
    public static boolean compare(byte a[], byte b[], PrintStream out) {
	boolean result = true;
	
	if (a.length != b.length) {
	    out.println("Byte array lengths differ: " +
			       a.length + " != " + b.length);
	    result = false;
	}
	
	for (int i = 0; i < a.length; i++){
	    if (a[i] != b[i] ) {
		out.println("Byte array contents differ at index " + i + ": " +
			    a[i] + " != " + b[i]);
		result = false;
	    }
	}
	return result;
    }

    public static boolean compare(char a[], char b[], PrintStream out) {
	boolean result = true;
	
	if (a.length != b.length) {
	    out.println("Char array lengths differ: " +
			a.length + " != " + b.length);
	    result = false;
	}
	
	for (int i = 0; i < a.length; i++){
	    if (a[i] != b[i] ) {
		out.println("Char array contents differ at index " + i + ": " +
			    a[i] + " != " + b[i]);
		result = false;
	    }
	}
	return result;
    }
    
    public static boolean compare(short a[], short b[], PrintStream out) {
	boolean result = true;
	
	if (a.length != b.length) {
	    out.println("Short array lengths differ: " +
			a.length + " != " + b.length);
	    result = false;
	}
	
	for (int i = 0; i < a.length; i++){
	    if (a[i] != b[i] ) {
		out.println("Short array contents differ at index " + i + ": " +
			    a[i] + " != " + b[i]);
		result = false;
	    }
	}
	return result;
    }
    
    public static boolean compare(boolean a[], boolean b[], PrintStream out) {
	boolean result = true;
	
	if (a.length != b.length) {
	    out.println("Boolean array lengths differ: " +
			a.length + " != " + b.length);
	    result = false;
	}
	
	for (int i = 0; i < a.length; i++){
	    if (a[i] != b[i] ) {
		out.println("Boolean array contents differ at index " + i + ": " +
			    a[i] + " != " + b[i]);
		result = false;
	    }
	}
	return result;
    }
    
    public static boolean compare(float a[], float b[], PrintStream out) {
	boolean result = true;
	
	if (a.length != b.length) {
	    out.println("Float array lengths differ: " +
			a.length + " != " + b.length);
	    result = false;
	}
	
	for (int i = 0; i < a.length; i++){
	    if (a[i] != b[i] ) {
	    out.println("Float array contents differ at index " + i + ": " +
			a[i] + " != " + b[i]);
	    result = false;
	    }
	}
	return result;
    }
    
    public static boolean compare(double a[], double b[], PrintStream out) {
	boolean result = true;
	
	if (a.length != b.length) {
	    out.println("Double array lengths differ: " +
			a.length + " != " + b.length);
	    result = false;
	}
	
	for (int i = 0; i < a.length; i++){
	    if (a[i] != b[i] ) {
		out.println("Double array contents differ at index " + i + ": " +
			    a[i] + " != " + b[i]);
		result = false;
	    }
	}
	return result;
    }
    
    public static boolean compare(long a[], long b[], PrintStream out) {
	boolean result = true;
	
	if (a.length != b.length) {
	    out.println("Long array lengths differ: " +
			a.length + " != " + b.length);
	    result = false;
	}
	
	for (int i = 0; i < a.length; i++){
	    if (a[i] != b[i] ) {
		out.println("Long array contents differ at index " + i + ": " +
			    a[i] + " != " + b[i]);
		result = false;
	    }
	}
	return result;
    }

   public static boolean compare(String a[], String b[], PrintStream out) {
	boolean result = true;
	
	if (a.length != b.length) {
	    out.println("String array lengths differ: " +
			a.length + " != " + b.length);
	    result = false;
	}
	
	for (int i = 0; i < a.length; i++){
	    if (!a[i].equals(b[i])) {
		out.println("String array contents differ at index " + i + ": " +
			    a[i] + " != " + b[i]);
		result = false;
	    }
	}
	return result;
    }

    public static boolean compare(Object o1, Object o2, PrintStream out) {
	if (o1 instanceof int[] && o2 instanceof int[]) {
	    return compare((int[])o1, (int[])o2, out);
	} else if (o1 instanceof long[] && o2 instanceof long[]) {
	    return compare((long[])o1, (long[])o2, out);
	} else if (o1 instanceof short[] && o2 instanceof short[]) {
	    return compare((short[])o1, (short[])o2, out);
	} else if (o1 instanceof byte[] && o2 instanceof byte[]) {
	    return compare((byte[])o1, (byte[])o2, out);
	} else if (o1 instanceof char[] && o2 instanceof char[]) {
	    return compare((char[])o1, (char[])o2, out);
	} else if (o1 instanceof float[] && o2 instanceof float[]) {
	    return compare((float[])o1, (float[])o2, out);
	} else if (o1 instanceof double[] && o2 instanceof double[]) {
	    return compare((double[])o1, (double[])o2, out);
	} else if (o1 instanceof boolean[] && o2 instanceof boolean[]) {
	    return compare((boolean[])o1, (boolean[])o2, out);
	} else {
	    out.println("Unknown types to compare " + o1.getClass() +
			" =?= " + o2.getClass());
	    return false;
	}
    }

}
