/*
 * @(#)X11GraphicsEnvironment.java	1.17 98/09/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt;

import java.awt.GraphicsDevice;
import java.awt.Toolkit;
import java.awt.print.PrinterJob;

import sun.awt.motif.PSPrinterJob;
import sun.java2d.SunGraphicsEnvironment;
import java.io.InputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.File;
import java.io.StreamTokenizer;
import java.util.StringTokenizer;
import sun.awt.font.NativeFontWrapper;
import sun.awt.font.FontNameAliases;
import java.util.Map;
import java.util.Hashtable;
import java.util.NoSuchElementException;
import java.util.Vector;
import java.util.Locale;

/**
 * This is an implementation of a GraphicsEnvironment object for the
 * default local GraphicsEnvironment used by the JavaSoft JDK in X11
 * environments.
 *
 * @see GraphicsDevice
 * @see GraphicsConfiguration
 * @version 10 Feb 1997
 */

public class X11GraphicsEnvironment extends SunGraphicsEnvironment {

    Map fontNameMap;
    Map xlfdMap;

    static {
	java.security.AccessController.doPrivileged(
			  new sun.security.action.LoadLibraryAction("awt"));
    /*
     * Note: The MToolkit object depends on the static initializer
     * of X11GraphicsEnvironment to initialize the connection to
     * the X11 server.
     */
	initDisplay();
    }

    private static native synchronized void initDisplay();

    public X11GraphicsEnvironment() {
    }

    protected native int getNumScreens();

    protected GraphicsDevice makeScreenDevice(int screennum) {
	return new X11GraphicsDevice(screennum);
    }

    /**
     * Gets a <code>PrinterJob</code> object suitable for the
     * the current platform.
     * @return    a <code>PrinterJob</code> object.
     * @see       java.awt.PrinterJob
     * @since     JDK1.2
     */
    public PrinterJob getPrinterJob() {
	SecurityManager security = System.getSecurityManager();
	if (security != null) {
	    security.checkPrintJobAccess();
	}

	return new PSPrinterJob();
    }

    // disables family name mapping
    protected boolean platformReady() {
        return true;
    }

    // can have platform-specific overrides - see X11GraphicsConfiguration
    protected String getFontPropertyFD(String name) {
        return name;
    }

    // platform specific override
    // takes family name property in the following format:
    // 1) "-linotype-helvetica-medium-r-normal-sans-*-%d-*-*-p-*-iso8859-1"
    // and returns the name of the corresponding physical font

    // can have platform-specific overrides - see X11GraphicsConfiguration
    protected String getFileNameFromPlatformName(String platName) {
        String fileName = null;
        String fontID = specificFontIDForName(platName);

        initFontNameMap(super.fontPath);
//		if (fontID == null) {
//            fontID = vagueFontIDForName(name);
//		}
        if ((fontNameMap != null) && (fontID != null)) {
            fileName = (String)fontNameMap.get(fontID);
            if (fileName == null) {
                fontID = switchFontIDForName(platName);
                fileName = (String)fontNameMap.get(fontID);
            }
        }
        return fileName;
    }

    // platform specific override
    // takes family name property in the following format:
    // 1) "-linotype-helvetica-medium-r-normal-sans-*-%d-*-*-p-*-iso8859-1"
    // and returns the name from the corresponding physical font

    protected String parseFamilyNameProperty(String name) {
        String retval = "default";
        String fontID = specificFontIDForName(name);

//	if (fontID == null) {
//          fontID = vagueFontIDForName(name);
//	}

        if ((fontNameMap != null) && (fontID != null)) {
            String fileName = (String)fontNameMap.get(fontID);
            if (fileName == null) {
                fontID = switchFontIDForName(name);
                fileName = (String)fontNameMap.get(fontID);
            }
            if (fileName != null) {
                String fontName =
                    NativeFontWrapper.getFullNameByFileName(fileName);
                if (fontName != null) {
                    retval = fontName;
                }
            }
        }
        return retval;
    }

    private String switchFontIDForName(String name) {
        String part1 = getFontField(name, 3);
        String part2 = getFontField(name, 4);
        String part3 = getFontField(name, 5);
        String part4 = getFontField(name, 6);
        String part5 = getFontField(name, 14);
        String part6 = getFontField(name, 15);
        if (part3.equals("i")) {
            part3 = "o";
        } else if (part3.equals("o")) {
            part3 = "i";
        }
        String retVal = part1 + "-" + part2 + "-" + part3 + "-" + part4 +
			"-" + part5 + "-" + part6;
	return (retVal.toLowerCase (Locale.ENGLISH));
    }

    private String specificFontIDForName(String name) {
        String part1 = getFontField(name, 3);
        String part2 = getFontField(name, 4);
        String part3 = getFontField(name, 5);
        String part4 = getFontField(name, 6);
        String part5 = getFontField(name, 14);
        String part6 = getFontField(name, 15);
        String retVal = part1 + "-" + part2 + "-" + part3 + "-" + part4 +
			"-" + part5 + "-" + part6;
	return (retVal.toLowerCase (Locale.ENGLISH));
    }

    private String vagueFontIDForName(String name) {
        String part1 = getFontField(name, 4);
        String part2 = getFontField(name, 5);
        String part3 = getFontField(name, 6);
        String part4 = getFontField(name, 7);
        String retVal = part1 + "-" + part2 + "-" + part3 + "-" + part4;
	return (retVal.toLowerCase (Locale.ENGLISH));
    }

    private String getFontField(String fontString, int fieldNum) {
        int beginIndex = 0;
        int endIndex = 0;
        while ((--fieldNum > 0) && (beginIndex >= 0)) {
            beginIndex = fontString.indexOf("-", beginIndex)+1;
        }
        if (beginIndex == -1) {
            return "";
        }
        endIndex = fontString.indexOf("-", beginIndex);
        if (endIndex == -1) {
            return fontString.substring(beginIndex);
        } else {
            return fontString.substring(beginIndex, endIndex);
        }
    }

    // can have platform specific override
    protected void registerFontPath(String path) {
        final File fontDir = new File(path + File.separator + "fonts.dir");
	java.security.AccessController.doPrivileged(
				    new java.security.PrivilegedAction() {
	    public Object run() {
		try {
		    parseFontDir(fontDir);
		} catch (Exception e) {
		    // no relevant errors for application
		}
		return null;
	    }
	});
    }

    private native String [] getNativeFonts ();

    protected String getPropString (String fontFileName) {
        if (xlfdMap == null) {
            return fontFileName;
        }
	String sVal = (String) xlfdMap.get (fontFileName);
	if (sVal == null) { 
	    return fontFileName;
	} else {
	    return sVal;
	}
    }

    protected void registerNative (String fontName) {
	Vector v = new Vector (5, 5); 
	v.addElement (fontName);
	registeredFonts.put (fontName, fontName);
	NativeFontWrapper.registerFonts (v, v.size (), v,
					 FontNameAliases.FONTFORMAT_NATIVE);
    }

    private Vector nativeFonts;

    private void initNativeFonts () {
	if (nativeFonts != null) {
	    return;
	}

	String [] xFonts = getNativeFonts ();
	if (xFonts != null) {
	    nativeFonts = new Vector ();
	    if (fontNameMap == null) {
                fontNameMap = new Hashtable ();
	    }
	    if (xlfdMap == null) {
                xlfdMap = new Hashtable ();
	    }
  	    for (int i = 0; i < xFonts.length; i++) {
		String fontID = specificFontIDForName (xFonts [i]);
                String sVal = (String) fontNameMap.get (fontID);
		if (sVal == null) {
		   nativeFonts.addElement (xFonts [i]);
		   fontNameMap.put (fontID,  xFonts [i]);
		   xlfdMap.put (xFonts [i], xFonts [i]);
		}
	    }
	}
    }
	
    protected boolean registerNativeFonts () {
	if (nativeFonts != null) {
	    Vector v = new Vector (20, 10);
  	    for (int i = 0; i < nativeFonts.size (); i++) {
		String xlfd = (String) nativeFonts.elementAt (i);
		if (registeredFonts.containsKey (xlfd)) {
		    continue;
		}
		v.addElement (xlfd);
		registeredFonts.put (xlfd, xlfd);
	    }
	    NativeFontWrapper.registerFonts (v, v.size (), v,
				 FontNameAliases.FONTFORMAT_NATIVE);
	    return true;
	}

	return false;
    }

    private void initFontNameMap(String pathName) {
        if (fontNameMap !=null) {
            return;
        }
        StringTokenizer parser = new StringTokenizer(pathName, 
                                                     File.pathSeparator);
        try {
            while (parser.hasMoreTokens()) {
                String newPath = parser.nextToken();
                registerFontPath(newPath);
            }
        } catch (NoSuchElementException e) {
            System.err.println(e);
        }
	initNativeFonts ();
    }

    private void parseFontDir(File fontDir)
            throws FileNotFoundException, IOException {
        // file format:
        //  47
        //  Arial.ttf -monotype-arial-regular-r-normal--0-0-0-0-p-0-iso8859-1
        //  Arial-Bold.ttf -monotype-arial-bold-r-normal--0-0-0-0-p-0-iso8859-1
        //  ...
	String path = fontDir.getParent ();
        if (fontDir.canRead()) {
            FileInputStream fs = new FileInputStream(fontDir);
            StreamTokenizer st = new StreamTokenizer((InputStream)fs);
            st.eolIsSignificant(true);
            int ttype = st.nextToken();
            if (ttype == StreamTokenizer.TT_NUMBER) {
                int numEntries = (int)st.nval;
                ttype = st.nextToken();
                if (ttype == StreamTokenizer.TT_EOL) {
		    if (fontNameMap == null) {
                        fontNameMap = new Hashtable();
		    }
		    if (xlfdMap == null) {
                        xlfdMap = new Hashtable();
		    }
                    st.resetSyntax();
                    st.wordChars(32, 127);
	            st.wordChars(128 + 32, 255);
	            st.whitespaceChars(0, 31);

                    for (int i=0; i < numEntries; i++) {
                        ttype = st.nextToken();
                        if (ttype == StreamTokenizer.TT_EOF) {
                            break;
                        }
                        if (ttype != StreamTokenizer.TT_WORD) {
                            //System.out.println("Invalid Syntax 1: font.dir");
                            break;
                        }
                        int breakPos = st.sval.indexOf(" ");
                        if (breakPos <= 0) {
                            //System.out.println("Invalid Syntax 2: font.dir");
                            break;
                        }
                        String fileName = st.sval.substring(0, breakPos);
			String fontPart = st.sval.substring(breakPos+1);

                        String fontID = specificFontIDForName(fontPart);
			String sVal = (String) fontNameMap.get (fontID);

			/* REMIND: Fix this! Here's the situation. A
			 *    font file may be associated with two different
			 *    xlfd strings, in locales such as ja, ko,
			 *    zh and zh_TW. So, if the font file already
			 *    has an xlfd string associated with it, do
			 *    not add it to the fontNameMap since creating
			 *    a second C++ fontObject (via
			 *    NativeFontWrapper.registerFonts) is useless
		  	 *    since it will never get used. The C++ 
			 *    fontObjects are typically retrieved by
			 *    matching the "font name" associated with
			 *    a fontObject. If you have two fontObjects
			 *    that reference the same file, and later
			 *    retrieve a fontObject based on name, the
			 *    first one in the list will always be
			 *    returned. So, if you've created a 2nd
			 *    fontObject with a different xlfd string,
			 *    this fontObject will never get used because
			 *    the first one will always be found first.
			 *    To get around this problem, do not add
			 *    an entry to the fontNameMap even if one
			 *    does not exist for this particular xlfd
			 *    string iff an entry in the xlfdMap already
			 *    exists for this font file. What will then
			 *    happen is that later, when the "native"
			 *    fonts are registered, when this xlfd string
			 *    is found, a "native" fontObject will be
			 *    created. So, the correct glyphs will be
			 *    displayed, via the native rasterizer code path.
			 *
			 *    I believe some changes may need to be
			 *    made to SunGraphicsEnvironment in order to
			 *    make this work "correctly". It would seem
			 *    that multiple fontObjects need to be created
			 *    per font file if that font file is used for
			 *    more than one X11 encoding. Additionally then,
			 *    the font name returned for the fontObject
			 *    needs to be distinct each time. In the native
			 *    code, I appended the encoding associated with
			 *    the particular fontObject instance (if the
			 *    X11 endcoding is not iso8859-x), such as:
			 *    Ryumin Light (jisx0201.1976-0). This makes
			 *    the names distinct and allows the code to
			 *    retrieve the correct one given the xlfd string.
			 */

		    	String fullPath = path + File.separator + fileName;
			String xVal = (String) xlfdMap.get (fullPath);
			if ((xVal == null)  && (sVal == null)) {
                            fontNameMap.put(fontID, fileName);
			    xlfdMap.put (fullPath, fontPart);
			}

//                      System.out.println("s["+fontID+"]["+fileName+"]");
//			fontID = vagueFontIDForName(fontPart);
//                      fontNameMap.put(fontID, fileName);
//			System.out.println("v["+fontID+"]["+fileName+"]");

                        ttype = st.nextToken();
                        if (ttype != StreamTokenizer.TT_EOL) {
                            //System.out.println("Invalid Syntax 3: font.dir");
                            break;
                        }
                    }
                }
            }
        }
    }

    // can have platform-specific overrides - see X11GraphicsConfiguration
    protected void initCompositeFonts() {
//        loadFonts();
        super.initCompositeFonts();
    }
}
