/*
 * @(#)MDropTargetContextPeer.java	1.31 98/09/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.motif;

import java.awt.AWTPermission;
import java.awt.Component;
import java.awt.Container;
import java.awt.Insets;
import java.awt.Point;
import java.awt.Toolkit;

import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.FlavorMap;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;

import java.awt.dnd.DnDConstants;

import java.awt.dnd.DropTarget;
import java.awt.dnd.DropTargetContext;
import java.awt.dnd.DropTargetListener;
import java.awt.dnd.DropTargetEvent;
import java.awt.dnd.DropTargetDragEvent;
import java.awt.dnd.DropTargetDropEvent;
import java.awt.dnd.InvalidDnDOperationException;

import java.awt.dnd.peer.DropTargetContextPeer;

import java.io.InputStream;
import java.io.ByteArrayInputStream;
import java.io.ObjectInputStream;

import java.security.AccessController;
import java.security.AccessControlContext;

import javax.swing.SwingUtilities;

import java.lang.reflect.InvocationTargetException;

import java.util.Hashtable;
import java.util.Map;
import java.util.Collection;

import java.io.IOException;

import java.rmi.MarshalledObject;
import java.rmi.UnmarshalException;

import sun.awt.motif.MDragSourceContextPeer;

/**
 * <p>
 * The MDropTargetContextPeer class is the class responsible for handling
 * the interaction between the Motif DnD system and Java.
 * </p>
 *
 * @version 1.31
 * @since JDK1.2
 *
 */

final class MDropTargetContextPeer implements DropTargetContextPeer, Transferable {

    private static final Object _globalLock = new Object();

    /*
     * a primitive mechanism for advertising intra-JVM Transferables
     */

    private static Transferable         currentJVMLocalSourceTransferable = null;
    private static AccessControlContext currentJVMLocalSourceACtxt        = null;

    static void setCurrentJVMLocalSourceTransferable(Transferable t) throws InvalidDnDOperationException {
	synchronized(_globalLock) {
	    if (t != null && currentJVMLocalSourceTransferable != null) {
	    	throw new InvalidDnDOperationException();
	    } else {
	        currentJVMLocalSourceTransferable = t;
		currentJVMLocalSourceACtxt        = t != null ? AccessController.getContext() : null;
	    }
	} 
    }

    /**
     * obtain the transferable iff the operation is in the same VM
     */

    private static Transferable getJVMLocalSourceTransferable() {
	Transferable t = null;

	synchronized(_globalLock) {

	    /*
	     * If the Transferable is local then we should compare the 
	     * AccessContext of the initiator of the Drag with
	     * that of the receipient of the Drop, if they are equal
	     * or if the receipient is prepared to accept cross ctxt
	     *
	     */

		if (currentJVMLocalSourceTransferable != null) {
		    if (!currentJVMLocalSourceACtxt.equals(AccessController.getContext()))
 		        AccessController.checkPermission(new AWTPermission("acceptDropBetweenAccessControlContexts"));
		}

	    t = currentJVMLocalSourceTransferable;
	}

	return t;
    }

    /*
     * constants used by dropAccept() or dropReject()
     */

    final static int STATUS_NONE   =  0; // none pending
    final static int STATUS_WAIT   =  1; // drop pending
    final static int STATUS_ACCEPT =  2;
    final static int STATUS_REJECT = -1;

    private static MDropTargetContextPeerDispatcher listenerDispatcher = null;

    /**
     * create the peer
     */

    static MDropTargetContextPeer createMDropTargetContextPeer() {
	synchronized(_globalLock) {
	    if (listenerDispatcher == null) listenerDispatcher = new MDropTargetContextPeerDispatcher();
	}

	return new MDropTargetContextPeer();
    }

    /**
     * create the peer
     */

    MDropTargetContextPeer() {
	super();
    }

    /**
     * @return the DropTarget associated with this peer
     */

    public DropTarget getDropTarget() { return currentDT; }

    /**
     * @param actions set the current actions
     */

    public void setTargetActions(int actions) {
	    currentA = actions & (DnDConstants.ACTION_COPY_OR_MOVE | DnDConstants.ACTION_LINK);
    }
  
    /**
     * @return the current target actions
     */

    public int getTargetActions() { return currentA; }
  
    /**
     * signal drop complete
     */

    public void dropComplete(boolean success) {
	synchronized(lock) {
	    if (dropStatus == STATUS_NONE) {
		throw new InvalidDnDOperationException("No Drop pending");
	    }

	    if (currentDTC != null) currentDTC.removeNotify();

	    currentC   = null;
	    currentDT  = null;
	    currentDTC = null;
	    currentT   = null;
	    currentDA  = DnDConstants.ACTION_NONE;
	    currentSA  = DnDConstants.ACTION_NONE;
	    currentA   = DnDConstants.ACTION_NONE;

	    synchronized(_globalLock) {
	        currentJVMLocalSourceTransferable = null;
	        currentJVMLocalSourceACtxt        = null;
	    }

	    dropDone(nativeDragContext, nativeDropTransfer, local != null, success);

	    local        = null;
	    dropComplete = true;
	}
    }

    /**
     * get the Transferable associated with the drop
     */

    public synchronized Transferable getTransferable() {
	return this;
    }

    /**
     * @return current DataFlavors available
     */
    // NOTE: This method may be called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    public DataFlavor[] getTransferDataFlavors() {

	if (local != null) return local.getTransferDataFlavors();

	DataFlavor[]  dfs;

	synchronized(lock) {
    	    if (currentT == null)
		dfs = new DataFlavor[0];
    	    else {
		Object[] values = currentDT.getFlavorMap().getFlavorsForNatives(currentT).values().toArray();

		// should replace this with toArray(new DataFlavor[1]);

		dfs = new DataFlavor[values.length];

		System.arraycopy(values, 0, dfs, 0, values.length);
	    }
	}

	return dfs;
    }

    /**
     * @return if the flavor is supported
     */

    public boolean isDataFlavorSupported(DataFlavor df) {

	if (local != null) return isDataFlavorSupported(df);

	DataFlavor[] dfs = getTransferDataFlavors();

	for (int i = 0; i < dfs.length; i++)
	    if (df.equals(dfs[i])) return true;

	return false;
    }

    /**
     * @return the data 
     */

    public Object getTransferData(DataFlavor df) throws UnsupportedFlavorException, IOException, InvalidDnDOperationException {
	
	if (local != null) return local.getTransferData(df);

        String atom = null;

	synchronized(lock) {
	    if (dropStatus != STATUS_ACCEPT || dropComplete) 
		throw new InvalidDnDOperationException("No drop current");


	    Map atoms = currentDT.getFlavorMap().getNativesForFlavors(new DataFlavor[] { df } );

	    if ((atom = (String)atoms.get(df)) == null)
		throw new UnsupportedFlavorException(df);

	    int i;

	    for (i = 0; i < currentT.length; i++)
		if (currentT[i].equals(atom)) break;

	    if (i == currentT.length) throw new UnsupportedFlavorException(df);
	
	    if (nativeDropTransfer == 0) {
		nativeDropTransfer = startTransfer(nativeDragContext, atom);
	    } else 
	        addTransfer(nativeDropTransfer, atom);

	    for (nativeDataAvailable = null;
		 !atom.equals(nativeDataAvailable);) {
		try {
		    lock.wait();
		} catch (InterruptedException ie) {
		    // do nothing
		}
	    }

	    if (nativeData == null) 
		throw new IOException("no native data was transferred");

	    if (nativeData instanceof java.util.List) return nativeData;

	    ByteArrayInputStream bais = new ByteArrayInputStream((byte[])nativeData);
	    if (df.isRepresentationClassInputStream()) {
		final Class dfrc = df.getRepresentationClass();

		if (java.io.InputStream.class.equals(dfrc))
		    return bais; // simple case.
		else { // its a class that implements InputStream
		    java.lang.reflect.Constructor[] constructors = null;

		    try {
			constructors = (java.lang.reflect.Constructor[])
			java.security.AccessController.doPrivileged(
				    new java.security.PrivilegedAction() {
			    public Object run() {
				return dfrc.getConstructors();
			    }
			});
		    } catch (SecurityException se) {
			se.printStackTrace();

			throw new IOException(se.getMessage());
		    }

		    java.lang.reflect.Constructor constructor = null;

		    for (int j = 0; j < constructors.length; j++) {
			if (!java.lang.reflect.Modifier.isPublic(constructors[j].getModifiers())) continue;

			Class[] ptypes = constructors[j].getParameterTypes();

			if (ptypes != null && ptypes.length == 1 && java.io.InputStream.class.equals(ptypes[0])) {
			    constructor = constructors[j];
			    break;
			}
		    }

		    if (constructor == null) 
			throw new InvalidDnDOperationException("cant find public <init>(Ljava/io/InputStream;)V for class: " + dfrc.getName());

		    try {
		        return constructor.newInstance(new Object[] { bais } );
		    } catch (Exception e) {
			e.printStackTrace();

			throw new InvalidDnDOperationException(e.getMessage());
		    }
		}
	    } else if (df.isRepresentationClassRemote()) {
		if (currentDA != DnDConstants.ACTION_LINK)
		    throw new InvalidDnDOperationException("only ACTION_LINK is permissable for transfer java.rmi.Remote objects");
		try {
		    return ((MarshalledObject)(new ObjectInputStream(bais)).readObject()).get();
		} catch (Exception e) {
		    e.printStackTrace();

		    throw new InvalidDnDOperationException(e.getMessage());
		}
	    } else if (df.isRepresentationClassSerializable()) {
		try {
		    ObjectInputStream ois = new ObjectInputStream(bais);

		    return ois.readObject();
		} catch (IOException ioe) {
		    ioe.printStackTrace();
		    throw new InvalidDnDOperationException(ioe.getMessage());
		} catch (ClassNotFoundException cnfe) {
		    cnfe.printStackTrace();
		    throw new InvalidDnDOperationException(cnfe.getMessage());
		}
	    } else if (java.lang.String.class.equals(df.getRepresentationClass())) {
		return new String((byte[])nativeData);
	    }

	    throw new IOException("getTransferData() failed");
	}
    }


    /**
     * @return if the transfer is a local one
     */
  
    public boolean isTransferableJVMLocal() {
	return local != null || getJVMLocalSourceTransferable() != null;
    }

    /**
     * upcall to handle enter messages
     */
  
    private synchronized int handleEnterMessage(Component component, int x, int y, int dropAction, int actions, String[] atoms, long nativeCtxt) {
	Component  c    = null;
	DropTarget dt   = null;
	boolean    hit;
	Point	   hots = null;

	nativeDragContext = nativeCtxt;
        currentT          = atoms;
	currentSA         = actions;
	currentDA         = dropAction;

 	local = getJVMLocalSourceTransferable();

	// we do hit detection and co-ord translation "locked" of course

	synchronized(component.getTreeLock()) {
	    try {
	        c = ((Container)component).findComponentAt(x, y);
	    } catch (ClassCastException cce) {
	        c = component.locate(x, y);
	    }

	    hit = (c != null    		    &&
		   c.isShowing()		    &&
		   (dt = c.getDropTarget()) != null &&
		   dt.isActive()
	          );

	    if (hit) {
	    	 hots = translate(component, x, y, c);
            }
	}

	if (currentDTC != null) { // some wreckage from last time
	    currentDTC.removeNotify();
	    currentDTC = null;
	}

	if (hit) { 
	    currentC   = c;
	    currentDT  = dt;
	    currentDTC = currentDT.getDropTargetContext();

	    currentDTC.addNotify(this);

	    currentA   = dt.getDefaultActions();

	    // if no drop action selected pick one ...

	    if (currentDA == DnDConstants.ACTION_NONE) 
	        currentDA = dropAction = mapOperation(DnDConstants.ACTION_COPY | DnDConstants.ACTION_MOVE | DnDConstants.ACTION_LINK, currentA & currentSA);

	    listenerDispatcher.dispatchEnter(this, (DropTargetListener)dt, new DropTargetDragEvent(currentDTC, hots, dropAction, actions));

	    dropAction = currentDA;

	} else {
	    currentDT  = null;
	    currentDTC = null;
	    currentDA  = DnDConstants.ACTION_NONE;
	    currentSA  = DnDConstants.ACTION_NONE;
	    currentA   = DnDConstants.ACTION_NONE;
	}

	return currentDA;
    }

    /**
     * upcall to handle exit messages
     */
  
    private synchronized void handleExitMessage(Component component, long nativeCtxt) {
	
	nativeDragContext = nativeCtxt;

	currentDA = DnDConstants.ACTION_NONE;
	currentSA = DnDConstants.ACTION_NONE;

	local = null;

	if (currentC != null) {
	    DropTarget t = currentC.getDropTarget();

	    if (t == null) {
		currentC  = null;
		currentDT = null;
		currentT  = null;
		currentA  = DnDConstants.ACTION_NONE;

		if (currentDTC != null) currentDTC.removeNotify();

		currentDTC = null;

		return;
	    } else if (t != currentDT) {
	    	if (currentDTC != null) currentDTC.removeNotify();

		currentDT  = t;
		currentDTC = t.getDropTargetContext();

		currentDTC.addNotify(this);
	    }

	    if (currentDT.isActive()) {
		listenerDispatcher.dispatchExit(this, (DropTargetListener)currentDT, new DropTargetEvent(currentDTC));
	    }

	    currentA  = DnDConstants.ACTION_NONE;

	    currentDTC.removeNotify();
	}

	local = null;
    }

    /**
     * upcall to handle motion, synchronized with AWT_LOCK()
     */
  
    private synchronized int handleMotionMessage(Component component, int x, int y, int dropAction, boolean dropActionChanged, int actions, String[] atoms, long nativeCtxt) {
	Component  c     = null;
	Point      hots  = null;
	Point      nhots = null;
	boolean    hit   = false;
	boolean    nhit  = false;
	DropTarget dt    = null;
	DropTarget ndt   = null;

	nativeDragContext = nativeCtxt;
	currentT	  = atoms;
	currentSA         = actions;
	currentDA         = dropAction;

	if (currentC != null) {
	    boolean same = false;

	    synchronized(component.getTreeLock()) {
		try {
	            c = ((Container)component).findComponentAt(x, y);
		} catch (ClassCastException cce) {
	            c = component.locate(x, y);
		}

		hots = translate(component, x, y, currentC);

		same = currentC.equals(c);

		if (!same && c != null) {
		    nhots = translate(component, x, y, c);
		    nhit  = c.isShowing()		      &&
			    (ndt = c.getDropTarget()) != null &&
			    ndt.isActive();
		}

	        if ((dt = currentC.getDropTarget()) == null) {
		    currentDTC.removeNotify();
		    currentC   = null;
		    currentDTC = null;
		    currentT   = null;
	        } else {
		    if (currentDT != dt) {
			if (currentDTC != null) currentDTC.removeNotify();

			currentDT  = dt;
			currentDTC = null;
		    }

		    DropTargetContext dtc = currentDT.getDropTargetContext();

		    if (dtc != currentDTC) {
			if (currentDTC != null) currentDTC.removeNotify();
			(currentDTC = dtc).addNotify(this);
		    }

	            hit = currentC.isShowing() && currentDT.isActive();
		}
	    }

	    if (hit) {
	        if (!same) {
		    listenerDispatcher.dispatchExit(this, (DropTargetListener)currentDT, new DropTargetEvent(currentDTC));

		    currentDTC.removeNotify();
		    currentDTC = null;
		    currentC   = null;
	    	    currentT   = null;
		    currentDT  = null;
		    currentA   = DnDConstants.ACTION_NONE;
	        } else {
		    currentA = currentDT.getDefaultActions();

		    if (dropAction == DnDConstants.ACTION_NONE)
	    		currentDA = dropAction = mapOperation(DnDConstants.ACTION_COPY | DnDConstants.ACTION_MOVE | DnDConstants.ACTION_LINK, currentA & currentSA);

		    DropTargetDragEvent dtde = new  DropTargetDragEvent(currentDTC, hots, dropAction, actions);

		    listenerDispatcher.dispatchOver(this, (DropTargetListener)currentDT, dtde, dropActionChanged);

		    return currentDA;
		}
	    } else {
		currentA  = DnDConstants.ACTION_NONE;
		currentDA = DnDConstants.ACTION_NONE;
	    }
	}

	if (c == null) synchronized(component.getTreeLock()) {
	    try {
	        c = ((Container)component).findComponentAt(x, y);
	    } catch (ClassCastException cce) {
	        c = component.locate(x, y);
	    }


	    nhit = c != null			     &&
		   c.isShowing()		     &&
		   (ndt = c.getDropTarget()) != null &&
		   ndt.isActive();

	    if (nhit) nhots = translate(component, x, y, c);
	}

	if (nhit) {
	    currentC   = c;
	    currentT   = atoms;
	    currentDT  = ndt;
	    currentDTC = ndt.getDropTargetContext();

	    currentDTC.addNotify(this);

	    currentA   = ndt.getDefaultActions();

	    if (dropAction == DnDConstants.ACTION_NONE)
	    	currentDA = dropAction = mapOperation(DnDConstants.ACTION_COPY | DnDConstants.ACTION_MOVE | DnDConstants.ACTION_LINK, currentA & currentSA);

	    listenerDispatcher.dispatchEnter(this, (DropTargetListener)currentDT, new DropTargetDragEvent(currentDTC, nhots, dropAction, actions));
        } else synchronized(lock) {
	    currentC   = null;
	    currentDT  = null;
	    currentDTC = null;

	    currentA  = DnDConstants.ACTION_NONE;
	    currentDA = DnDConstants.ACTION_NONE;
	}

	return currentDA;
    }

    /**
     * upcall to handle the Drop message
     */
  
    private int handleDropMessage(Component component, int x, int y, int dropAction, int actions, String[] atoms,long nativeCtxt) {
	Component  c    = null;
	Point      hots = null;
	boolean    hit  = false;
	DropTarget dt   = null;

	nativeDragContext = nativeCtxt;
        currentT	  = atoms;
	currentSA         = actions;
	currentDA         = dropAction;

	synchronized(component.getTreeLock()) {
	    try { 
	        c = ((Container)component).findComponentAt(x, y);
	    } catch (ClassCastException cce) {
	        c = component.locate(x, y);
	    }
	
	    hit = (c != null    		    &&
		   c.isShowing()		    &&
		   (dt = c.getDropTarget()) != null &&
		   dt.isActive()
	          );

	    if (hit) {
		hots = translate(component, x, y, c);
	    }
	}

	if (hit) {
	    currentC  =  c;
	    currentDT = dt;

	    DropTargetContext dtc = dt.getDropTargetContext();

	    if (currentDTC != null) currentDTC.removeNotify();

	    (currentDTC = dtc).addNotify(this);
		
	    dropStatus     = STATUS_WAIT; // drop pending ACK
	    dropDispatched = false;

	    currentA = currentDT.getDefaultActions();

	    if ((local = getJVMLocalSourceTransferable()) != null) {
	        setCurrentJVMLocalSourceTransferable(null);
	    }

	    if (dropAction == DnDConstants.ACTION_NONE)
	        currentDA = dropAction = mapOperation(DnDConstants.ACTION_COPY | DnDConstants.ACTION_MOVE | DnDConstants.ACTION_LINK, currentSA & currentA);

	    listenerDispatcher.dispatchDrop(this, (DropTargetListener)currentDT, new DropTargetDropEvent(currentDTC, hots, dropAction, actions, local != null), dropLock);

	    if (dropStatus == STATUS_REJECT) {
	        if (currentDTC != null) currentDTC.removeNotify();

	        currentC   = null;
	        currentDT  = null;
	        currentDTC = null;
	        currentT   = null;
	        currentA   = DnDConstants.ACTION_NONE;
	        currentDA  = DnDConstants.ACTION_NONE;
	    }
	} else {
	    if (currentDTC != null) currentDTC.removeNotify();

	    currentC   = null;
	    currentDT  = null;
	    currentDTC = null;
	    currentT   = null;
	    currentA   = DnDConstants.ACTION_NONE;
	    currentDA  = DnDConstants.ACTION_NONE;
	}

	currentSA  = DnDConstants.ACTION_NONE;

	return currentDA;
    }

    /**
     * acceptDrag
     */

    public void acceptDrag(int dragOperation) {
	synchronized(lock) {
	    if (currentDT == null) {
		throw new InvalidDnDOperationException("No Drag pending");
	    }

	    if (currentDA != dragOperation)
	        currentDA = mapOperation(currentA & currentSA, dragOperation);
	}
    }

    /**
     * rejectDrag
     */

    public void rejectDrag() {
	synchronized(lock) {
	    if (currentDT == null) {
		throw new InvalidDnDOperationException("No Drag pending");
	    }

	    currentDA = DnDConstants.ACTION_NONE;
	}
    }

    /**
     * acceptDrop
     */

    public void acceptDrop(int dropOperation) {
	synchronized (lock) {
	    if (dropStatus == STATUS_NONE) {
		throw new InvalidDnDOperationException("No Drop pending");
	    }

	    if (currentDA != dropOperation)
	        currentDA = mapOperation(currentA & currentSA, dropOperation);

	    if (currentDA != DnDConstants.ACTION_NONE) {
	        dropStatus   = STATUS_ACCEPT;
		dropComplete = false;
	    } else {
	        dropStatus   = STATUS_REJECT;
	        currentA     = DnDConstants.ACTION_NONE;
	        dropComplete = true;
	    }
	}

	listenerDispatcher.dispatched();

	// now wait until the event dispatch thread reclaims AWT_LOCK

	synchronized(dropLock) {
	    while (!dropDispatched) try {
		dropLock.wait();
	    } catch (InterruptedException ie) {
		// do nothing!
	    }
	}
    }

    /**
     * reject Drop
     */

    public void rejectDrop() {
	synchronized(lock) {
	    if (dropStatus == STATUS_NONE) {
		throw new InvalidDnDOperationException("No Drop pending");
	    }

	    dropStatus   = STATUS_REJECT;
	    currentDA    = DnDConstants.ACTION_NONE;
	    dropComplete = true;
	}

	listenerDispatcher.dispatched();

	// now wait until the event dispatch thread reclaims AWT_LOCK

	synchronized(dropLock) {
	    while (!dropDispatched) try {
		dropLock.wait();
	    } catch (InterruptedException ie) {
		// do nothing!
	    }
	}
    }

    /**
     * get the Status of the drop
     */

    int getDropStatus() { synchronized(lock) { return dropStatus; } }

    /**
     * mapOperation
     */

    private int mapOperation(int actions, int operation) {
	int[] operations = { 
		DnDConstants.ACTION_MOVE,
		DnDConstants.ACTION_COPY,
		DnDConstants.ACTION_LINK,
	};

	operation &= actions;

	for (int i = 0; i < operations.length; i++) {
	    if ((operation & operations[i]) == operations[i]) {
	   	return operations[i];
	    }
	}

	return DnDConstants.ACTION_NONE;
    }

    /**
     * notify transfer complete
     */

    private void newData(String atom, String type, byte[] data) {
	synchronized(lock) {
	    nativeDataAvailable = atom;
	    nativeData          = data;

	    heartbeat();

 	    lock.notify();
	}
    }

    /**
     * notify transfer failed
     */

    private void transferFailed(String atom) {
	synchronized(lock) {
	    nativeDataAvailable = atom;
	    nativeData          = null;

	    heartbeat();

 	    lock.notify();
	}
    }

    /**
     * keep the Motif transfer alive - called with lock held
     */

    private void heartbeat() {
	if (!dropComplete) keepAlive(nativeDropTransfer);
    }

    /**
     * upcall from native code.
     */

    private void dropDispatched() {
	synchronized(dropLock) {
	    dropDispatched = true;
	    dropLock.notify();
	}
    }

    /**
     * translate co-ordinates - called with Component Tree Lock held
     */

    private Point translate(Component anc, int x, int y, Component desc) {
	Component c  = desc;
	int 	  dx = 0;
	int	  dy = 0;

	while (!c.equals(anc)) {
	    Point locn = c.getLocation();
	    dx += locn.x;
	    dy += locn.y;
	    c = MComponentPeer.getParent_NoClientCode(c);
	}

	return new Point(x - dx, y - dy);
    }

    /**
     *  construct a list ...
     */

    private static java.util.List getList(Object[] objects) {
	return java.util.Arrays.asList(objects);
    }

    /**
     * schedule a native DnD transfer
     */

    private native long startTransfer(long nativeDragContext, String atom);

    /**
     * schedule a native DnD data transfer
     */

    private native void addTransfer(long nativeDropTransfer, String atom);

    /**
     * signal that drop is completed
     */

    private native void dropDone(long nativeDragContext, long nativeDropTransfer, boolean localTx, boolean success);

    /**
     * downcall to keep Motif connection going
     */

    private native void keepAlive(long nativeDropTransfer);

    /*
     * instance vars ...
     */

    private Component	      currentC;
    private DropTarget        currentDT;
    private DropTargetContext currentDTC;
    private String[]	      currentT;

   
    private int		      currentA;   // target actions
    private int		      currentSA;  // source actions
    private int		      currentDA;  // current drop

    private long      	      nativeDragContext;	  

    private Transferable      local;

    private long      	      nativeDropTransfer;	  

    int		              dropStatus   = STATUS_NONE;
    boolean		      dropComplete = false;

    String		      nativeDataAvailable = null;
    Object		      nativeData          = null;

    Object	      	      lock                = new Object();

    boolean		      dropDispatched	  = false;
    Object		      dropLock	  	  = new Object();
}

/*
 * private class to dispatch listener callbacks via AWT EventQueue
 */

final class MDropTargetContextPeerDispatcher implements Runnable {

   /*
    * DropTargetListener method dispatcher constants
    */

    private final static int NONE       = 0;
    private final static int ENTER      = 1;
    private final static int OVER       = 2;
    private final static int CHANGED    = 3;
    private final static int EXIT       = 4;
    private final static int DROP       = 5;
	
    /*
     * callback from EventQueue
     */

    public void run() {
        try {
	    switch (dispatchType) {
	        case ENTER:
		    dtListener.dragEnter((DropTargetDragEvent)dtEvent);
		break;

	        case OVER:
		    dtListener.dragOver((DropTargetDragEvent)dtEvent);
		break;

	        case CHANGED:
		    dtListener.dropActionChanged((DropTargetDragEvent)dtEvent);
		break;

	        case EXIT:
		    dtListener.dragExit((DropTargetEvent)dtEvent);
		break;

	        case DROP:
		    dtListener.drop((DropTargetDropEvent)dtEvent);
		break;
	    }
	 } catch (Exception e) {
	    	e.printStackTrace();
	 } finally {
	        if (dispatchType == DROP) {
		    if (dtPeer.getDropStatus() == MDropTargetContextPeer.STATUS_WAIT) {
		        dtPeer.rejectDrop();
			throw new InvalidDnDOperationException("Drop rejected; DropTargetListener drop() method must either acceptDrop() or rejectDrop()");
		    }
	    	} else dispatched();
	}
    }

    /*
     * clear dispatcher state
     */

    public synchronized void dispatched() {
	if (dispatchType != NONE) {
	    dispatchType = NONE;
	    notify();
	}
    }

    /*
     * dispatch dragEnter
     */

    public void dispatchEnter(MDropTargetContextPeer dtp, DropTargetListener dtl, DropTargetEvent dte) {
	dispatch(dtp, dtl, dte, ENTER);
    }


    /*
     * dispatch dragOver
     */

    public void dispatchOver(MDropTargetContextPeer dtp, DropTargetListener dtl, DropTargetEvent dte, boolean dropActionChanged) {
	dispatch(dtp, dtl, dte, dropActionChanged ? CHANGED : OVER);
    }

    /*
     * dispatch dragExit
     */

    public void dispatchExit(MDropTargetContextPeer dtp, DropTargetListener dtl, DropTargetEvent dte) {
	dispatch(dtp, dtl, dte, EXIT);
    }

    /*
     * dispatch drop
     */

    public void dispatchDrop(MDropTargetContextPeer dtp, DropTargetListener dtl, DropTargetEvent dte, Object lock) {
	synchronized(lock) {
	    synchronized(this) {
		dispatchType = DROP;
		dtPeer	     = dtp;
		dtListener   = dtl;
		dtEvent      = dte;

		SwingUtilities.invokeLater(this);

	        while (dispatchType != NONE) try {
		    wait();
	    	} catch (InterruptedException ie) {
		    // do nothing!
	    	}
	    }

	    lock.notify();
	}
    }

    /*
     * dispatch it ... 
     */

    private void dispatch(MDropTargetContextPeer dtp, DropTargetListener dtl, DropTargetEvent dte, int type) {
	boolean done = false;

	synchronized(this) {
	    dispatchType = type;
	    dtPeer	 = dtp;
	    dtListener   = dtl;
	    dtEvent      = dte;
	}

	while (!done) try {
	    SwingUtilities.invokeAndWait((Runnable)this);
	    done = true;
	} catch (InterruptedException ie) {
	    ie.printStackTrace();
	} catch (InvocationTargetException ite) {
	    ite.printStackTrace();
	    done = true;
	}
    }

    /*
     * fields
     */
    
    private MDropTargetContextPeer   dtPeer;
    private DropTargetListener       dtListener;
    private DropTargetEvent          dtEvent;
    private int 		     dispatchType = NONE;
}
