/*
 * @(#)PSGraphics.java	1.31 98/07/07
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.motif;

import java.awt.Graphics;
import java.awt.PrintGraphics;
import java.awt.PrintJob;
import java.awt.Color;
import java.awt.Font;
import java.awt.Dimension;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.Image;
import java.awt.FontMetrics;
import java.awt.Toolkit;
import java.awt.image.PixelGrabber;
import java.awt.image.ColorModel;
import java.awt.image.IndexColorModel;
import java.awt.image.ImageObserver;
import java.awt.font.TextAttribute;
import java.text.CharacterIterator;
import java.text.AttributedCharacterIterator;

/**
 * PSGraphics is an object that encapsulates a graphics context for
 * rendering to a shared PostScript output stream.  This object is
 * only concerned with maintaining the state of its private graphics
 * attributes.  It uses a PSPrintStream object to manage the graphics
 * state of the PostScript interpreter which is being shared among
 * all instantiated PSGraphics objects.
 *
 * @version 1.31 07/07/98
 * @author Jim Graham
 * @author Amy Fowler
 * @author Richard Berlin
 */
public class PSGraphics extends Graphics implements PrintGraphics {
    public static Dimension Letter = new Dimension((72 * 17)/2, 11 * 72);
    public static Dimension Legal = new Dimension((72 * 17)/2, 14 * 72);
    public static Dimension Executive = new Dimension((72 * 15)/2, 10 * 72);
    public static Dimension A4 = new Dimension((int) (72 * 210 / 25.4),
					       (int) (72 * 297 / 25.4));

    // NB: if you add an instance variable, be sure also to add it to the
    // "copy" version of the constructor.

    PSPrintStream	stream;

    int			pageNumber;

    Color		foreground;
    Font		font;
    Rectangle		clipRect;

    int			originX;	// REMIND: change to float
    int			originY;	// REMIND: change to float
    int			adjustX;
    int			adjustY;

    FontMetrics		curMetrics;

    private boolean disposed = false;

    /**
     * Creates a child copy of the specified graphics context.
     */
    public PSGraphics(PSPrintStream stream, int pageNumber) {
	this.stream = stream;
	this.pageNumber = pageNumber;
	newAdjust();
    }

    /**
     * Creates a child copy of the specified graphics context.
     */
    public PSGraphics(PSGraphics psg) {

	// The order of these exactly matches the order in which the
	// instance variables are declared.  Please keep it this way;
	// it makes it much easier to ensure that all variables are
	// being properly copied.

	stream       = psg.stream;

	pageNumber   = psg.pageNumber;

	foreground   = psg.foreground;
	font         = psg.font;
	clipRect     = psg.clipRect;

	originX      = psg.originX;
	originY      = psg.originY;
	adjustX      = psg.adjustX;
	adjustY      = psg.adjustY;

	curMetrics   = psg.curMetrics;
    }

    /**
     * Create a child copy of this graphics context.
     */
    public Graphics create() {
	return new PSGraphics(this);
    }

    /*
     * Subclasses should override disposeImpl() instead of dispose(). Client
     * code should always invoke dispose(), never disposeImpl().
     */
    protected void disposeImpl() {}
    public final void dispose() {
        boolean call_disposeImpl = false;

        if (!disposed) {
            synchronized (this) {
                if (!disposed) {
                    disposed = call_disposeImpl = true;
                }
            }
        }

        if (call_disposeImpl) {
            disposeImpl();
        }
    }

    private void newAdjust() {
	adjustX = originX;
	adjustY = stream.paperDim.height - originY;
    }

    /** Return the current font. */
    public Font getFont() {
	// Bugs 4039449, 4054657 -- A component being printed had no default
	// font and dies.  If a component doesn't need a font for display
	// (comes from Frame), a font should be provided for printing as well.
	if (font == null)
	    setFont(new Font("Dialog", Font.PLAIN, 12));
	return font;
    }

    /** Return the font metrics associated with specified font. */
    public FontMetrics getFontMetrics(Font f) {
	return Toolkit.getDefaultToolkit().getFontMetrics(f);
    }

    /**
     * Sets the font for all subsequent text-drawing operations.
     */
    public void setFont(Font f) {
	if (font != f) {
	    font = f;
	    curMetrics = null;
	}
    }

    /**
     * Return the current color.
     */
    public Color getColor() {
	return foreground;
    }

    /**
     * Sets the current color.
     */
    public void setColor(Color c) {
	foreground = c;
    }

    public void translate(int x, int y) {
	originX += x;
	originY += y;
	newAdjust();
    }

    public void setPaintMode() {
    }

    public void setXORMode(Color color) {
    }

    public Rectangle getClipBounds() {
	return (clipRect == null
		? null
		: new Rectangle(clipRect.x - originX,
				clipRect.y - originY,
				clipRect.width, clipRect.height));
    }

    /** Crops the clipping rectangle for this Graphics context. */
    public synchronized void clipRect(int x, int y, int w, int h) {
	if (w <= 0) {w = 0;}
	if (h <= 0) {h = 0;}
	Rectangle newclip = new Rectangle(x + originX, y + originY, w, h);
	if (clipRect != null) {
	    newclip = newclip.intersection(clipRect);
	}
	clipRect = newclip;
    }

    /** Sets the clipping rectangle for this Graphics context. */
    public synchronized void setClip(int x, int y, int w, int h) {
	if (w <= 0) {w = 0;}
	if (h <= 0) {h = 0;}
	clipRect = new Rectangle(x + originX, y + originY, w, h);
    }

    /** Returns a Shape object representing the clip. */
    public Shape getClip() {
	return getClipBounds();
    }

    /** Sets the clip to a Shape (only Rectangle allowed). */
    public void setClip(Shape clip) {
        if (clip == null) {
	    clipRect = null;
	}
	else if (clip instanceof Rectangle) {
	    Rectangle r = (Rectangle) clip;
	    setClip(r.x, r.y, r.width, r.height);
	} else {
	    throw new IllegalArgumentException("setClip(Shape) only supports Rectangle objects");
	}
    }

    /** Draws given rectangle with rounded corners. */
    public void fillRoundRect(int x, int y, int w, int h, int arcW, int arcH) {
	stream.sendRoundRect(this, x, y, w, h,
			     arcW/2.0f, arcH/2.0f, true);
    }

    /** Fills given rectangle with rounded corners with foreground color. */
    public void drawRoundRect(int x, int y, int w, int h, int arcW, int arcH) {
	stream.sendRoundRect(this, x, y, w, h,
			     arcW/2.0f, arcH/2.0f, false);
    }

    /** Draws given oval. */
    public void drawOval(int x, int y, int w, int h) {
	drawArc(x, y, w, h, 0, 360);
    }

    /** Fills given oval with foreground color. */
    public void fillOval(int x, int y, int w, int h) {
	fillArc(x, y, w, h, 0, 360);
    }

    /** Draws given arc. */
    public void drawArc(int x, int y, int w, int h,
			int startAngle, int arcAngle) {
	stream.sendArc(this,
		       x + w/2, y + h/2, w/2, h/2,
		       startAngle, startAngle + arcAngle,
		       false);
    }

    /** Fills pie-slice defined by given arc with foreground color. */
    public void fillArc(int x, int y, int w, int h,
			int startAngle, int arcAngle) {
	stream.sendArc(this,
		       x + w/2, y + h/2, w/2, h/2,
		       startAngle, startAngle + arcAngle,
		       true);
    }

    /** Draws polygon defined by x-point array and y-point array */
    public void drawPolygon(int xp[], int yp[], int npoints) {
	stream.sendPolygon(this, xp, yp, npoints, true, false);
    }

    /** Draws an unclosed figure defined by x-point array and y-point array */
    public void drawPolyline(int xp[], int yp[], int npoints) {
	stream.sendPolygon(this, xp, yp, npoints, false, false);
    }

    /** Fills polygon defined by x-point array and y-point array */
    public void fillPolygon(int xp[], int yp[], int npoints) {
	stream.sendPolygon(this, xp, yp, npoints, false, true);
    }

    private synchronized boolean blitImage(Image img,
					   int dx1, int dy1, int dx2, int dy2,
					   int sx1, int sy1, int sx2, int sy2,
					   Color bgcolor, boolean normalize)
    {
	if (!normalize) {
	    if (sx2 < sx1) {
		int t;
		t = dx1; dx1 = dx2; dx2 = t;
		t = sx1; sx1 = sx2; sx2 = t;
	    }
	    if (sy2 < sy1) {
		int t;
		t = dy1; dy1 = dy2; dy2 = t;
		t = sy1; sy1 = sy2; sy2 = t;
	    }
	}
	PixelGrabber ig = new PixelGrabber(img,
					   sx1, sy1, sx2 - sx1, sy2 - sy1,
					   false);
	try {
	    // Workaround for bug in image code which causes image production
	    // to get confused and never deliver the pixels; we time out and
	    // explicitly re-start production ourselves.
	    boolean gotPixels = false;
	    while (gotPixels == false) {
		gotPixels = ig.grabPixels(5000);
		if ((ig.getStatus() & (ImageObserver.ERROR |
				       ImageObserver.ABORT)) != 0)
		{
		    if (bgcolor != null) {
			stream.fillRect(this, dx1, dy1, dx2 - dx1, dy2 - dy1,
					bgcolor);
		    }
		    System.out.println("Image status: "+ig.getStatus());
		    return false;
		}
	   	if (gotPixels == false) {
		    img.getSource().startProduction(ig);
		}
	    }
	} catch (InterruptedException e) {
	    if (bgcolor != null) {
		stream.fillRect(this, dx1, dy1, dx2 - dx1, dy2 - dy1, bgcolor);
	    }
	    return false;
	}
	Object pixels = ig.getPixels();
	ColorModel model = ig.getColorModel();
	if (normalize) {
	    sx2 = sx1 + ig.getWidth();
	    sy2 = sy1 + ig.getHeight();
	    if (dx2 <= dx1) {
		dx2 = dx1 + ig.getWidth();
	    }
	    if (dy2 <= dy1) {
		dy2 = dy1 + ig.getHeight();
	    }
	}

	if (pixels instanceof byte[]) {
	    stream.sendPixels(this,
			      ((byte[]) pixels), ((IndexColorModel) model),
			      dx1, dy1, dx2, dy2,
			      sx1, sy1, sx2, sy2,
			      bgcolor);
	} else {
	    stream.sendPixels(this,
			      ((int[]) pixels), model,
			      dx1, dy1, dx2, dy2,
			      sx1, sy1, sx2, sy2,
			      bgcolor);
	}
	return true;
    }

    public boolean drawImage(Image img, int x, int y,
			     ImageObserver o) {
	return blitImage(img,
			 x, y, x, y,
			 0, 0, -1, -1,
			 null, true);
    }

    public boolean drawImage(Image img, int x, int y,
			     Color bgcolor, ImageObserver o) {
	return blitImage(img,
			 x, y, x, y,
			 0, 0, -1, -1,
			 bgcolor, true);
    }

    public boolean drawImage(Image img, int x, int y, int width, int height,
			     ImageObserver o) {
	if (width == 0 || height == 0) {
	    return true;
	}
	return blitImage(img,
			 x, y, x+width, y+height,
			 0, 0, -1, -1,
			 null, true);
    }

    public boolean drawImage(Image img, int x, int y, int width, int height,
			     Color bgcolor, ImageObserver o) {
	if (width == 0 || height == 0) {
	    return true;
	}
	return blitImage(img,
			 x, y, x+width, y+height,
			 0, 0, -1, -1,
			 bgcolor, true);
    }

    public boolean drawImage(Image img,
			     int dx1, int dy1, int dx2, int dy2,
			     int sx1, int sy1, int sx2, int sy2,
			     ImageObserver observer)
    {
	if (dx1 == dx2 || dy1 == dy2) {
	    return true;
	}
	if (sx1 == sx2 || sy1 == sy2) {
	    return true;
	}
	return blitImage(img,
			 dx1, dy1, dx2, dy2,
			 sx1, sy1, sx2, sy2,
			 null, false);
    }

    public boolean drawImage(Image img,
			     int dx1, int dy1, int dx2, int dy2,
			     int sx1, int sy1, int sx2, int sy2,
			     Color bgcolor,
			     ImageObserver observer)
    {
	if (dx1 == dx2 || dy1 == dy2) {
	    return true;
	}
	if (sx1 == sx2 || sy1 == sy2) {
	    stream.fillRect(this, dx1, dy1, dx2 - dx1, dy2 - dy1, bgcolor);
	    return true;
	}
	return blitImage(img,
			 dx1, dy1, dx2, dy2,
			 sx1, sy1, sx2, sy2,
			 bgcolor, false);
    }

    /** Clears the rectangle indicated by x,y,w,h. */
    public void clearRect(int x, int y, int w, int h) {
	stream.fillRect(this, x, y, w, h, stream.paperColor);
    }

    /** Fills the given rectangle with the foreground color. */
    public void fillRect(int x, int y, int w, int h) {
	stream.sendRect(this, x, y, w, h, true);
    }

    /** Draws the given rectangle. */
    public synchronized void drawRect(int x, int y, int w, int h) {
	stream.sendRect(this, x, y, w, h, false);
    }

    /** Draws the given string. */
    public synchronized void drawString(String str, int x, int y) {
	stream.sendString(this, str, x, y);
    }

    /** Draws the given text. */
    public synchronized void drawString(AttributedCharacterIterator iterator, int x, int y) {
        int current = iterator.getBeginIndex();
        int end = iterator.getEndIndex();
        if (current >= end) {
            return;
        }

        Font oldFont = font;
        StringBuffer buffer = new StringBuffer();
        char c = iterator.first();
        while (current < end) {
            Font font = (Font) iterator.getAttribute(TextAttribute.FONT);
            int limit = iterator.getRunLimit(TextAttribute.FONT);
            for ( ; current < limit; current++) {
                buffer.append(c);
                c = iterator.next();
            }
            setFont(font);
            stream.sendString(this, buffer.toString(), x, y);
            buffer.setLength(0);
        }

        setFont(oldFont);
    }

    /** Draws the given character array. */
    public void drawChars(char chars[], int offset, int length, int x, int y) {
	drawString(new String(chars, offset, length), x, y);
    }

    /** Draws the given byte array. */
    public void drawBytes(byte bytes[], int offset, int length, int x, int y) {
	drawString(new String(bytes, 0, offset, length), x, y);
    }

    /** Draws the given line. */
    public synchronized void drawLine(int x1, int y1, int x2, int y2) {
	stream.sendLine(this, x1, y1, x2, y2);
    }

    /**
     * Copies an area of the window that this graphics context paints to.
     * Note that this cannot be done in PostScript so we should throw an error.
     * @param X the x-coordinate of the source.
     * @param Y the y-coordinate of the source.
     * @param W the width.
     * @param H the height.
     * @param dx the x-coordinate of the destination.
     * @param dy the y-coordinate of the destination.
     */
    public void copyArea(int X, int Y, int width, int height, int dx, int dy) {
	// REMIND: Not allowed
	new Exception().printStackTrace();
    }

    public String toString() {
	return getClass().getName() + "[" + originX + "," + originY + "]";
    }

    public PrintJob getPrintJob() {
        return stream.job;
    }
}
