/*
 * @(#)PSPaperSize.java	1.5 98/09/21
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.motif;

import java.awt.Dimension;

/**
 * PSPaperSize is an object that encapsulates standard paper sizes
 * as descriped in the Adobe PostScript Printer Description specification
 *
 * @author Richard Berlin
 */
class PSPaperSize {
    String paperSizeName;
    Dimension paperSize;
    int tolerance;

    /**
     * Returns the name associated with this paper size
     */
    public String name() {
	return paperSizeName;
    }

    /**
     * Returns the width and height of this paper size, in points,
     * as a Dimension object
     */
    public Dimension size() {
	return paperSize;
    }

    /**
     * Returns the width of this paper size, in points
     */
    public int width() {
        return paperSize.width;
    }

    /**
     * Returns the height of this paper size, in points
     */
    public int height() {
	return paperSize.height;
    }

    /**
     * Returns a String containing PostScript code to select this
     * paper size on a printer.
     *
     * @param PSLevel (1 or 2) if Level 2, emits much less code, but will
     * not work correctly on a printer which does not support PostScript
     * Level 2
     * @param manualFeed if true, request manual feed paper.  On level 1
     * interpreters this generates a different paper size request.
     */
    public String toPostScriptCode(int PSLevel, boolean manualFeed) {
	if (PSLevel == 2) {
	    return 
		  "mark {\n"
		+ "%%BeginFeature: ManualFeed " + manualFeed + "\n"
		+ "<</ManualFeed " + manualFeed + ">> setpagedevice\n"
		+ "%%EndFeature\n"
		+ "%%BeginFeature: PageSize *" + paperSizeName + "\n"
		+ "<< /PageSize [" + width() + " " + height() + "]\n"
		+ "   /ImagingBBox null\n"
		+ "   /Policies <</PageSize 3>>\n"
		+ " >> setpagedevice\n"
		+ "%%EndFeature\n"
		+ "} stopped cleartomark";
	} else {
	    String trayCmd = paperSizeName.toLowerCase();
	    if (!manualFeed) trayCmd = trayCmd + "tray";

	    return
		  "mark {\n"
		+ "%%BeginFeature: ManualFeed " + manualFeed + "\n"
		+ "1 dict dup /ManualFeed " + manualFeed + " put setpagedevice\n"
		+ "%%EndFeature\n"
		+ "} stopped {\n"
		+ "%%BeginFeature: *ManualFeed " + manualFeed + "\n"
		+ "  statusdict /manualfeed " + manualFeed + " put\n"
		+ "%%EndFeature\n"
		+ "} if cleartomark mark {\n"
		+ "%%BeginFeature: PageSize *" + paperSizeName + "\n"
		+ "3 dict begin\n"
		+ "   /PageSize [" + width() + " " + height() + "] def\n"
		+ "   /ImagingBBox null def\n"
		+ "   /Policies 1 dict dup/PageSize 3 put def\n"
		+ " currentdict end setpagedevice\n"
		+ "%%EndFeature\n"
		+ "} stopped {\n"
	        + "  cleartomark statusdict begin mark {\n"
		+ "%%BeginFeature: *PageSize " + paperSizeName + "\n"
		+ "  " + trayCmd + "\n"
		+ "%%EndFeature\n"
		+ "  } stopped end\n"
		+ "} if cleartomark\n"
		+ "% now make sure we got what we requested!\n"
		+ "gsave initgraphics clippath pathbbox grestore\n"
		+ "5 dict begin\n"
		+ "  3 -1 roll /yoff exch def 3 -1 roll /xoff exch def\n"
		+ "  yoff sub /h exch def xoff sub /w exch def\n"
		+ "  w " + (paperSize.width-36) + " lt h "
		         + (paperSize.height-36) + " lt or {\n"
		+ "    w 2 div xoff add h 2 div yoff add translate\n"
		+ "    w xoff 2 mul add " + paperSize.width + " div\n"
		+ "    h yoff 2 mul add " + paperSize.height + " div\n"
		+ "    2 copy gt {exch} if pop dup scale\n"
		+ "    " + (-paperSize.width/2) + " " + (-paperSize.height/2)
		     + " translate\n"
		+ "  } if\n"
 		+ "end\n";
	}
    }

    /**
     * Returns a String containing PostScript code to select this
     * paper size on a printer.
     */
    public String toPostScriptCode() {
	return toPostScriptCode(1, false);
    }

    /**
     * Returns a String object representing the value of this PSPaperSize.
     */
    public String toString() {
	return "PSPaperSize[" + paperSizeName + ": "
	    + paperSize.width + "x" + paperSize.height + "]";
    }

    /**
     * Creates a new PSPaperSize with the given name, width and height.
     * @param name the name of this paper size
     * @param the width (in points) of this paper size
     * @param the height (in points) of this paper size
     */
    public PSPaperSize(String name, int width, int height) {
	this.paperSizeName = name;
	this.tolerance = 0;
	this.paperSize = new Dimension(width, height);
    }

    /**
     * Creates a new PSPaperSize with the given name, width and height.
     * Also specifies a matching tolerance; this is required for certain
     * "standard" paper sizes (e.g. "Executive") for which the standard
     * specification is approximate.
     *
     * @param name the name of this paper size
     * @param tolerance the matching tolerance for this paper size
     * @param width the width (in points) of this paper size
     * @param height the height (in points) of this paper size
     */
    public PSPaperSize(String name, int tolerance, int width, int height) {
	this.paperSizeName = name;
	this.tolerance = tolerance;
	this.paperSize = new Dimension(width, height);
    }

    static final PSPaperSize standardSizeTable[] = {
//        new PSPaperSize( "A0",        2384, 3370 ),
//        new PSPaperSize( "A1",        1684, 2384 ),
//        new PSPaperSize( "A2",        1191, 1684 ),
        new PSPaperSize( "A3",        842, 1191 ),
        new PSPaperSize( "A4",        595,  842 ),
//        new PSPaperSize( "A5",        420,  595 ),
//        new PSPaperSize( "A6",        297,  420 ),
//        new PSPaperSize( "A7",        210,  297 ),
//        new PSPaperSize( "A8",        148,  210 ),
//        new PSPaperSize( "A9",        105,  148 ),
//        new PSPaperSize( "A10",       73,  105 ),

//        new PSPaperSize( "B0",        2920, 4127 ),
//        new PSPaperSize( "B1",        2064, 2920 ),
//        new PSPaperSize( "B2",        1460, 2064 ),
//        new PSPaperSize( "B3",        1032, 1460 ),
        new PSPaperSize( "B4",        729, 1032 ),
//        new PSPaperSize( "B5",        516,  729 ),
//        new PSPaperSize( "B6",        363,  516 ),
//        new PSPaperSize( "B7",        258,  363 ),
//        new PSPaperSize( "B8",        181,  258 ),
//        new PSPaperSize( "B9",        127,  181 ),
//        new PSPaperSize( "B10",       91,  127 ),

//        new PSPaperSize( "ISOB0",     2835, 4008 ),
//        new PSPaperSize( "ISOB1",     2004, 2835 ),
//        new PSPaperSize( "ISOB2",     1417, 2004 ),
//        new PSPaperSize( "ISOB3",     1001, 1417 ),
        new PSPaperSize( "ISOB4",     709, 1001 ),
//        new PSPaperSize( "ISOB5",     499,  709 ),
//        new PSPaperSize( "ISOB6",     354,  499 ),
//        new PSPaperSize( "ISOB7",     249,  354 ),
//        new PSPaperSize( "ISOB8",     176,  249 ),
//        new PSPaperSize( "ISOB9",     125,  176 ),
//        new PSPaperSize( "ISOB10",    88,  125 ),

//        new PSPaperSize( "C0",        2599, 3676 ),
//        new PSPaperSize( "C1",        1837, 2599 ),
//        new PSPaperSize( "C2",        1298, 1837 ),
//        new PSPaperSize( "C3",        918, 1298 ),
//        new PSPaperSize( "C4",        649,  918 ),
//        new PSPaperSize( "C5",        459,  649 ),
//        new PSPaperSize( "C6",        323,  459 ),
//        new PSPaperSize( "C7",        230,  323 ),

//        new PSPaperSize( "DL",        312,  624 ),

        new PSPaperSize( "Legal13",   612,  936 ),

        new PSPaperSize( "Letter",    612,  792 ),
        new PSPaperSize( "Legal",     612, 1008 ),
//        new PSPaperSize( "Statement", 396,  612 ),
        new PSPaperSize( "Tabloid",   792, 1224 ),
//        new PSPaperSize( "Ledger",    1224, 792 ),
//        new PSPaperSize( "7x9",       504,  648 ),
//        new PSPaperSize( "9x11",      648,  792 ),
//        new PSPaperSize( "9x12",      648,  864 ),
//        new PSPaperSize( "10x13",     720,  936 ),
//        new PSPaperSize( "10x14",     720, 1008 ),

//        new PSPaperSize( "Comm10",    297,  684 ),
//        new PSPaperSize( "Monarch",   279,  540 ),
//        new PSPaperSize( "Folio",     595,  935 ),
//        new PSPaperSize( "Quarto",    610,  780 ),


        new PSPaperSize( "A4Small",      595,  842 ),
        new PSPaperSize( "LetterSmall",  612,  792 ),

	new PSPaperSize( "Executive",   36,  540,  720)
    };


    /**
     * Returns a standard paper size with the given name
     * @param name the name of the desired paper size
     */
    public static PSPaperSize find(String name) {
	if (name == null)
	    return null;

	for (int i = 0; i < standardSizeTable.length; ++i)
	    if (name.equalsIgnoreCase(standardSizeTable[i].paperSizeName))
		return(standardSizeTable[i]);

	return null;
    }

    /**
     * Returns a standard paper size whose width and height differ from the
     * specified width and height by no more than dMax in each dimension
     *
     * @param width the width of the desired paper size
     * @param height the height of the desired paper size
     * @param dMax a matching tolerance
     */
    public static PSPaperSize find(int width, int height, int dMax) {
	int dX = dMax + 1;
	int dY = dMax + 1;
	PSPaperSize closest = null;
    
	for (int i = 0; i < standardSizeTable.length; ++i) {
	    PSPaperSize p = standardSizeTable[i];

	    int dW = Math.abs(p.paperSize.width - width);
	    int dH = Math.abs(p.paperSize.height - height);

	    if ((dW <= p.tolerance) && (dH <= p.tolerance)) {
		return p;
	    }
	    else if ((dW <= dMax) && (dH <= dMax) && ((dW + dH) < (dX + dY))) {
		dX = dW;
		dY = dH;
		closest = p;
	    }
	}

	return(closest);
    }

    /**
     * Returns the standard paper size with the given width and height
     *
     * @param width the width of the desired paper size
     * @param height the height of the desired paper size
     */
    public static PSPaperSize find(int width, int height) {
	return find(width, height, 0);
    }
}

