/*
 * @(#)PSPathGraphics.java	1.1 98/09/23
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.motif;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Paint;

import java.awt.geom.PathIterator;
import java.awt.print.PrinterJob;

import sun.java2d.PathGraphics;

/**
 * This class converts paths into PostScript
 * by breaking all graphics into fills and
 * clips of paths.
 */

class PSPathGraphics extends PathGraphics {

    PSPathGraphics(Graphics2D graphics, PrinterJob printerJob) {
	super(graphics, printerJob);
    }

    /**
     * Creates a new <code>Graphics</code> object that is 
     * a copy of this <code>Graphics</code> object.
     * @return     a new graphics context that is a copy of 
     *                       this graphics context.
     * @since      JDK1.0
     */
    public Graphics create() {

	return new PSPathGraphics((Graphics2D) getDelegate().create(),
				  getPrinterJob());
    }

    /*
     * Fill the path defined by <code>pathIter</code>
     * with the specified color.
     * The path is provided in device coordinates.
     */
    protected void deviceFill(PathIterator pathIter, Color color) {

	PSPrinterJob psPrinterJob = (PSPrinterJob) getPrinterJob();

	convertToPSPath(pathIter);
	psPrinterJob.setColor(color);
	psPrinterJob.fillPath();
    }

    /*
     * Set the printer device's clip to be the
     * path defined by <code>pathIter</code>
     * The path is provided in device coordinates.
     */
    protected void deviceClip(PathIterator pathIter) {

	PSPrinterJob psPrinterJob = (PSPrinterJob) getPrinterJob();

	/* If a gsave is in effect then we want to
	 * pop it so that the previous clip has no
	 * effect on the clip we are about to set.
	 * Priot to setting our new clip do a gsave
	 * so that we can get rid of the clip later.
	 */
	psPrinterJob.grestore();
	psPrinterJob.gsave();

	convertToPSPath(pathIter);
	psPrinterJob.selectClipPath();
    }

    /**
     * Given a Java2D <code>PathIterator</code> instance,
     * this method translates that into a PostScript path..
     */
    private void convertToPSPath(PathIterator pathIter) {

	float[] segment = new float[6];
	int segmentType;

	PSPrinterJob psPrinterJob = (PSPrinterJob) getPrinterJob();

	/* Map the PathIterator's fill rule into the PostScript
	 * fill rule.
	 */
	int fillRule;
	if (pathIter.getWindingRule() == PathIterator.WIND_EVEN_ODD) {
	    fillRule = PSPrinterJob.FILL_EVEN_ODD;
	} else {
	    fillRule = PSPrinterJob.FILL_WINDING;
	}
	psPrinterJob.setFillMode(fillRule);

	psPrinterJob.beginPath();

	while (pathIter.isDone() == false) {
	    segmentType = pathIter.currentSegment(segment);

	    switch (segmentType) {
	     case PathIterator.SEG_MOVETO:
		psPrinterJob.moveTo(segment[0], segment[1]);
		break;
	     
	     case PathIterator.SEG_LINETO:
		psPrinterJob.lineTo(segment[0], segment[1]);
		break;

	    /* Convert the quad path to a bezier.
	     */
	     case PathIterator.SEG_QUADTO:
		float lastX = psPrinterJob.getPenX();
		float lastY = psPrinterJob.getPenY();
		float c1x = lastX + (segment[0] - lastX) * 2 / 3;
		float c1y = lastY + (segment[1] - lastY) * 2 / 3;
		float c2x = segment[2] - (segment[2] - segment[0]) * 2/ 3;
		float c2y = segment[3] - (segment[3] - segment[1]) * 2/ 3;
		psPrinterJob.bezierTo(c1x, c1y,
				      c2x, c2y,
				      segment[2], segment[3]);
		break;

	     case PathIterator.SEG_CUBICTO:
		psPrinterJob.bezierTo(segment[0], segment[1],
				      segment[2], segment[3],
				      segment[4], segment[5]);
		break;

	     case PathIterator.SEG_CLOSE:
		psPrinterJob.closeSubpath();
		break;
	    }


	    pathIter.next();
	}

    }

}
