/*
 * @(#)PSPrintControl.java	1.14 98/07/07
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.motif;

import java.awt.*;
import java.io.*;
import java.util.*;


/**
 * A class which presents a modal print dialog to the user and
 * returns an appropriately configured print job.
 *
 * @version 	1.14 07/07/98
 * @author 	Amy Fowler
 */
public class PSPrintControl implements PrintControl {
    public final static String DEST_PROP = "awt.print.destination";
    public final static String PRINTER = "printer";
    public final static String FILE = "file";

    public final static String PRINTER_PROP = "awt.print.printer";

    public final static String FILENAME_PROP = "awt.print.fileName";

    public final static String NUMCOPIES_PROP = "awt.print.numCopies";

    public final static String OPTIONS_PROP = "awt.print.options";

    public final static String ORIENT_PROP = "awt.print.orientation";
    public final static String PORTRAIT = "portrait";
    public final static String LANDSCAPE = "landscape";

    public final static String PAPERSIZE_PROP = "awt.print.paperSize";
    public final static String LETTER = "letter";
    public final static String LEGAL = "legal";
    public final static String EXECUTIVE = "executive";
    public final static String A4 = "a4";

    private static ResourceBundle messageRB;
    /**
     * Initialize ResourceBundle
     */
    static void initResource() {
	try {
	    messageRB =
		ResourceBundle.getBundle("sun.awt.motif.resources.printcontrol");
	} catch (MissingResourceException e) {
	    throw new Error("Fatal: Resource for PSPrintControl is missing.");
	}
    }

    /**
     * get message string from resource
     */
    static synchronized String getMsg(String key) {
        if (messageRB == null) {
	    initResource();
	}
	try {
	    return messageRB.getString(key);
	} catch (MissingResourceException e) {
	    throw new Error("Fatal: Resource for PSPrintControl is broken, There is no " + key + " key in resource");
	}
    }

    private UPrintDialog printDialog;
    private Frame parentFrame;

    private Hashtable procTable = new Hashtable();

    private String destStr;
    private String copiesStr;
    private String bannerStr;
    private String orientStr;
    private String paperStr;
    private String printerName;
    private String fileName = null;
    private String lpOptions = null;
    private int	destType = PrintControl.PRINTER;
    private int	orient = PrintControl.PORTRAIT;
    private int	paperSize = PrintControl.PAPER_LETTER;
    private int	numCopies = 1;

    /**
     * Creates a modal print dialog to be used to allow the user to
     * configure the attributes of the print job.
     */
    public PSPrintControl(final Frame frame) {

	// A security check has already been performed in the
        // MToolkit.getPrintJob method.
	// So by the time we get here, it is OK for the current thread
	// to print either to a file (froma  Dialog we control!) or
	// to a chosen printer.

	// We raise privilege while we create the dialog, to avoid
	// the "warning applet window" banner.
	printDialog = (UPrintDialog) java.security.AccessController.doPrivileged(
	    new java.security.PrivilegedAction() {
	    public Object run() {
		return new UPrintDialog(frame);
	    }
	});

        parentFrame = frame;
    }

    /**
     * Shows the print dialog to the user.  If the user chooses "print",
     * returns an appropriately configured print job.  If the user chooses
     * "cancel", returns null.  If props is not null, the defaults for the
     * dialog area extracted from props.  If saveProps is true, any changes
     * the user made to the settings in the dialog are saved back into props.
     */
    public PSPrintJob initJob(String jobtitle, Properties props, boolean saveProps) {
	PSPrintJob job = null;

	readProps(props);
	printDialog.setDocumentTitle(jobtitle);
	printDialog.setNumCopies(numCopies);
	printDialog.setDestType(destType);
	printDialog.setDestString(PrintControl.PRINTER, printerName != null? printerName : "lp");
	printDialog.setDestString(PrintControl.FILE, fileName != null? fileName : "");
	printDialog.setOptionsString(lpOptions != null? lpOptions : "");
	printDialog.setBannerString(jobtitle);
	printDialog.setOrientation(orient);
	printDialog.setPaperSize(paperSize);

	job = (PSPrintJob)printDialog.getPrintJob(this);

        if (job != null) {
	    if (saveProps)
		writeProps(props);
	}
	return job;
    }

    public PrintJob createPrintJob(int destType, String destString, String lpOptions,
                      String bannerStr, int paperSize, int orient,
                      int numCopies) {
	OutputStream os = null;
	Process process = null;
	PSPrintJob returnJob;
	String[] execCmd = new String[1];
        String errorStr = null;

        this.destType = destType;
        this.bannerStr = bannerStr;
        this.numCopies = numCopies;
        this.orient = orient;

	switch(destType) {
	  case PrintControl.PRINTER:
            printerName = destString;
            this.lpOptions = lpOptions;
            try {
		execCmd = printExecCmd(destString, lpOptions, bannerStr, numCopies);
	 	process = Runtime.getRuntime().exec(execCmd);
		os = process.getOutputStream();
            } catch (IOException e) {
                errorStr = getMsg("error_msg.exec_print") + execCmd;
            }
	    break;
	  case PrintControl.FILE:
            fileName = destString;
            final File destFile = new File(fileName);
            String canonPath = "";
            try {
		canonPath = (String)
		    java.security.AccessController.doPrivileged(
			  new java.security.PrivilegedExceptionAction() {
		    public Object run() throws IOException {
			return destFile.getCanonicalPath();
		    }
		});

                if (destFile.exists()) {
                    String cancelCmd = getMsg("warning_button.cancel");
                    String overwriteCmd = getMsg("warning_button.overwrite");
                    PrintStatusDialog d = new PrintStatusDialog(parentFrame,
                                                  getMsg("warning_dialog.title"),
                                                  getMsg("warning_msg.file_exists")+fileName,
                                                  overwriteCmd,
                                                  cancelCmd);
                    d.show();
                    if (d.getCommand().equals(cancelCmd)) {
                        return null;
                    }
                }
		os = new FileOutputStream(destFile);
            } catch (java.security.PrivilegedActionException pae) {
                errorStr = getMsg("error_msg.create_file")+ canonPath;
            } catch (IOException e) {
                errorStr = getMsg("error_msg.create_file")+ canonPath;
            }
	    break;
	  default:
	    return null;
	}
        if (errorStr != null) {
	    PrintStatusDialog d = new PrintStatusDialog(parentFrame,
                                                        getMsg("error_dialog.title"),
                                                        errorStr,
                                                        getMsg("error_button.ok"));
            d.show();
	    return null;
	}

	Dimension pageDim;
	switch(paperSize) {
	  case PrintControl.PAPER_LEGAL:
	    pageDim = PSGraphics.Legal;
	    break;
	  case PrintControl.PAPER_EXECUTIVE:
	    pageDim = PSGraphics.Executive;
	    break;
	  case PrintControl.PAPER_A4:
	    pageDim = PSGraphics.A4;
	    break;
	  case PrintControl.PAPER_LETTER:
	  default:
	    pageDim = PSGraphics.Letter;
	}
	if (orient == PrintControl.LANDSCAPE) {
	    pageDim = new Dimension(pageDim.height, pageDim.width);
	}
	returnJob = new PSPrintJob(new PrintStream(
			                 new BufferedOutputStream(os), false),
				   bannerStr, pageDim, 72, this);
	procTable.put(returnJob, (process == null) ? new NullProcess()
                                                   : process);
	return returnJob;
    }

    public void finishedPrinting(PSPrintJob job) {
        Process proc = (Process) (procTable.get(job));
	new Thread(new CheckExitCodeRunnable(proc, parentFrame)).start();
    }

    private void readProps(Properties props) {
	if (props == null)
	    return;

  	if ((destStr = props.getProperty(DEST_PROP)) != null)
	    destType = destStr.equals(FILE)? PrintControl.FILE : PrintControl.PRINTER;

	printerName = props.getProperty(PRINTER_PROP);

   	fileName = props.getProperty(FILENAME_PROP);

	lpOptions = props.getProperty(OPTIONS_PROP);

	if ((orientStr = props.getProperty(ORIENT_PROP)) != null)
	    orient = orientStr.equals(LANDSCAPE)? PrintControl.LANDSCAPE : PrintControl.PORTRAIT;

	if ((paperStr = props.getProperty(PAPERSIZE_PROP)) != null) {
	    if (paperStr.equals(LETTER))
		paperSize = PrintControl.PAPER_LETTER;
	    else if (paperStr.equals(LEGAL))
		paperSize = PrintControl.PAPER_LEGAL;
	    else if (paperStr.equals(EXECUTIVE))
		paperSize = PrintControl.PAPER_EXECUTIVE;
	    else if (paperStr.equals(A4))
		paperSize = PrintControl.PAPER_A4;
	}

  	if ((copiesStr = props.getProperty(NUMCOPIES_PROP)) != null) {
	    Integer copiesInt = new Integer(copiesStr);
	    numCopies = copiesInt.intValue();
	}
    }

    private void writeProps(Properties props) {
	if (props == null)
	    return;

	props.put(DEST_PROP, (destType == PrintControl.PRINTER? PRINTER : FILE));

	if (printerName != null && !printerName.equals("") && !printerName.equals("lp"))
	    props.put(PRINTER_PROP, printerName);

	if (fileName != null && !fileName.equals(""))
	    props.put(FILENAME_PROP, fileName);

	if (lpOptions != null && !lpOptions.equals(""))
	    props.put(OPTIONS_PROP, lpOptions);

	props.put(ORIENT_PROP, (orient == PrintControl.PORTRAIT? PORTRAIT : LANDSCAPE));

	switch(paperSize) {
	    case PrintControl.PAPER_LEGAL:
		paperStr = LEGAL;
		break;
	    case PrintControl.PAPER_EXECUTIVE:
		paperStr = EXECUTIVE;
	  	break;
	    case PrintControl.PAPER_A4:
		paperStr = A4;
		break;
	    case PrintControl.PAPER_LETTER:
	    default:
		paperStr = LETTER;
	}
	props.put(PAPERSIZE_PROP, paperStr);

	if (numCopies > 0)
	    props.put(NUMCOPIES_PROP, new Integer(numCopies).toString());
    }

    private String[] printExecCmd(String printer, String options, String banner, int copies) {
	int PRINTER = 0x1;
        int OPTIONS = 0x2;
	int BANNER  = 0x4;
	int COPIES  = 0x8;
	int pFlags = 0;
	String execCmd[];
	int ncomps = 1;
	int n = 0;

	if (printer != null && !printer.equals("") && !printer.equals("lp")) {
	    pFlags |= PRINTER;
	    ncomps+=1;
	}
	if (options != null && !options.equals("")) {
	    pFlags |= OPTIONS;
	    ncomps+=1;
	}
	if (banner != null && !banner.equals("")) {
	    pFlags |= BANNER;
	    ncomps+=1;
	}
	if (copies > 1) {
	    pFlags |= COPIES;
	    ncomps+=1;
	}
	execCmd = new String[ncomps];
	execCmd[n++] = "/usr/bin/lp";
	if ((pFlags & PRINTER) != 0) {
	    execCmd[n++] = new String("-d" + printer);
	}
	if ((pFlags & BANNER) != 0) {
	    execCmd[n++] = new String("-t" + escapeString(banner));
	}
	if ((pFlags & COPIES) != 0) {
	    execCmd[n++] = new String("-n" + new Integer(copies).toString());
	}
	if ((pFlags & OPTIONS) != 0) {
 	        // multiple options currently don't work because /usr/bin/lp
	        // (Solaris 2.6) doesn't handle them correctly
	    execCmd[n++] = new String("-o" + options);
	}
	return execCmd;
    }


    /**
     * Adjusts a string for the two special characters which confuse the
     * print command: '\\' and '"'. An empirical check showed that no other
     * non-alphanumeric characters posed a problem, although not all of them
     * printed correctly (those that didn't were coverted to '_').
     */
    private String escapeString(String s) {
        StringTokenizer tokens = new StringTokenizer(s, "\"\\", true);
 	if (tokens.countTokens() == 1) {
 	    return s;
 	}

 	StringBuffer buf = new StringBuffer(s.length() * 2);
 	String token;
 	char c;

 	while (tokens.hasMoreTokens()) {
 	    token = tokens.nextToken();
 	    if (token.length() == 1) {
 	        c = token.charAt(0);
 		if (c == '"' || c == '\\') {
 		    buf.append('\\');
 		}
 		buf.append(c);
 	    }
 	    else {
 	        buf.append(token);
 	    }
 	}

 	return new String(buf);
    }
}

class NullProcess extends Process {
    public OutputStream getOutputStream() {
        return null;
    }
    public InputStream getInputStream() {
        return null;
    }
    public InputStream getErrorStream() {
        return null;
    }
    public int waitFor() {
        return 0;
    }
    public int exitValue() {
        return 0;
    }
    public void destroy() {
    }
}

class CheckExitCodeRunnable implements Runnable {
    private Process process;
    private Frame parentFrame;

    public CheckExitCodeRunnable(Process process, Frame parentFrame) {
	this.process = process;
	this.parentFrame = parentFrame;
    }

    public void run() {
	int returnVal = -1;
	String errorStr = null;

	try {
	    returnVal = process.waitFor();
	}
	catch (InterruptedException e) {
	    errorStr = PSPrintControl.getMsg("error_msg.interrupted");
	}

	if (returnVal != 0 && errorStr == null) {
	    byte[] errorText = new byte[256];

	    try {
	        process.getErrorStream().read(errorText);
		errorStr = PSPrintControl.getMsg("error_msg.error_code") +
		    ' ' + returnVal + ": " + new String(errorText).trim();
	    }
	    catch (IOException e) {
	        errorStr = PSPrintControl.getMsg("error_msg.unknown");
	    }
	}

	if (errorStr != null) {
	    new PrintStatusDialog(parentFrame,
				  PSPrintControl.getMsg("error_dialog.title"),
				  errorStr,
				  PSPrintControl.getMsg("error_button.ok")).
	        show();
	}
    }
}
