/*
 * @(#)PSPrintStream.java	1.18 98/10/29
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.motif;

import java.awt.Color;
import java.awt.Font;
import java.awt.Dimension;
import java.awt.Rectangle;
import java.awt.FontMetrics;
import java.awt.PrintJob;
import java.awt.image.ColorModel;
import java.awt.image.IndexColorModel;
import java.awt.image.PixelGrabber;
import java.io.InputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.PrintStream;
import java.io.BufferedInputStream;
import sun.io.CharToByteConverter;
import java.io.CharConversionException;
import java.util.Properties;

import sun.awt.CharsetString;
import sun.awt.FontDescriptor;
import sun.awt.PlatformFont;

/**
 * PSPrintStream is an object that encapsulates the current state of a
 * PostScript output stream.  This object is shared among many PSGraphics
 * objects so that they can manage multiplexed access to the single
 * PostScript execution context.
 * 
 * @version 1.18 10/29/98
 * @author Jim Graham
 * @author Amy Fowler
 * @author Richard Berlin
 */
class PSPrintStream {
    PrintStream		ps;
    PrintJob		job;
    public String	title;
    public Dimension	paperDim;
    Properties		fontProps;

    int			curPage = -1;

    Rectangle		curClip;
    Color		curColor;
    Font		curFont;
    int[]		curColormap;

    boolean		prologSent = false;
    boolean		lastFirst = false;

    Color		paperColor = Color.white;

    static String SetColorName = "SC";
    static String SetFontName = "SF";
    static String RectClipName = "RC";
    static String InitClipName = "IC";
    static String RectFillName = "RF";
    static String RectStrokeName = "RS";
    static String RRectFillName = "RRF";
    static String RRectStrokeName = "RRS";
    static String ArcFillName = "AF";
    static String ArcStrokeName = "AS";
    static String PolygonFillName = "PYF";
    static String PolygonStrokeName = "PYS";
    static String DrawStringName = "DS";
    static String DrawLineName = "DL";
    static String ScaleName = "S";
    static String StartPageName = "SP";
    static String EndPageName = "EP";
    static String GsaveName = "GS";
    static String GrestoreName = "GR";
    static String SetColormapName = "CM";
    static String DrawImageName = "DI";
    static String DrawIndexImageName = "DII";

    static String testString =
	"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ 1234567890";

    PSPrintStream(PrintJob job, PrintStream ps, String title,
		  Dimension paper, boolean lastFirst) {
	this.job = job;
	this.ps = ps;
	this.title = title;
	this.paperDim = paper;
	this.lastFirst = lastFirst;
	//enable privs so initProps can access system properties,
	// open the property file, etc.
	java.security.AccessController.doPrivileged(
			    new java.security.PrivilegedAction() {
	    public Object run() {
		fontProps = initProps();
		return null;
	    }
	});
    }

    /**
     * Initialize PostScript font properties
     */
    private Properties initProps() {
	// search psfont.propeties for fonts
	// and create and initialize fontProps if it exist.

	String jhome = System.getProperty("java.home");
	if (jhome != null){
	    String ulocale = System.getProperty("user.language", "en");
	    try {
		File f = new File(jhome + File.separator +
				  "lib" + File.separator + 
				  "psfont.properties." + ulocale);
		if (!f.canRead()){
		    f = new File(jhome + File.separator + "lib" +
				 File.separator + "psfont.properties");
		    if (!f.canRead()){
			return (Properties)null;
		    }
		}
		// Load property file
		InputStream in =
		    new BufferedInputStream(new FileInputStream(f.getPath()));
		Properties props = new Properties();
		props.load(in);
		in.close();
		return props;
	    } catch (Exception e){
		return (Properties)null;
	    }
	}
	return (Properties)null;
    }

    void startPage(int pageNumber) {
	curClip = null;
	curColor = null;
	curFont = null;
	curColormap = null;
	curPage = pageNumber;
	ps.println("\n%%Page:  "+pageNumber+" "+pageNumber);
	ps.println(StartPageName);
    }

    void endPage() {
	ps.println(EndPageName);
	curPage = -1;
	ps.flush();
    }

    void close(int numPages) {
	ps.println("%%Trailer");
	ps.println("%%Pages: "+numPages);
	ps.println("%%EOF");
	ps.flush();
    }

    private void outPoint(PSGraphics psg, int x, int y) {
	ps.print((psg.adjustX + x) + " " +
		 (psg.adjustY - y) + " ");
    }

    private void outRect(PSGraphics psg, int x, int y, int w, int h) {
	outPoint(psg, x, y);
	ps.print(w + " " + (-h) + " ");
    }

    private void outClip(PSGraphics psg) {
	// Note: This assumes that clip Rectangle objects are unique...
	Rectangle clip = psg.clipRect;
	if (curClip != clip) {
	    ps.println(InitClipName);
	    if (clip != null) {
		// Note: The clip is stored in absolute coordinates so we
		// need to adjust it back relative to the current origin
		// so that the outPoint adjustments do the right thing.
		outRect(psg,
			clip.x - psg.originX, clip.y - psg.originY,
			clip.width, clip.height);
		ps.println(RectClipName);
	    }
	    curClip = clip;
	}
    }

    private void outColor(Color c) {
	if (curColor != c) {
	    ps.println(c.getRed() + " " +
		       c.getGreen() + " " +
		       c.getBlue() + " " +
		       SetColorName);
	    curColor = c;
	}
    }

    private void outFont(PSGraphics psg) {
	Font font = psg.font;
	// REMIND: all we check is the size and style for now...
	if (curFont != font) {
	    curFont = font;
	    FontMetrics curMetrics = psg.curMetrics;
	    if (psg.curMetrics == null) {
		curMetrics = psg.getFontMetrics(font);
		psg.curMetrics = curMetrics;
	    }
	    int desiredWidth = curMetrics.stringWidth(testString);
	    int fontIndex = font.getStyle();
	    if (font.getFamily().startsWith("Helvetica") ||
		font.getFamily().startsWith("SansSerif")) {
		fontIndex = 0;
	    } else if (font.getFamily().startsWith("Dialog")) {
		fontIndex = 0;
	    } else if (font.getFamily().startsWith("Times") ||
		       font.getFamily().startsWith("Serif")) {
		fontIndex = 4;
	    } else {
		fontIndex = 8;
	    }
	    if (font.isBold())
		fontIndex += 1;
	    if (font.isItalic())
	        fontIndex += 2;
	    ps.println("% Font["
		       +font.getFamily()+", "
		       +font.getStyle()+", "
		       +font.getSize()
		       +"]");
	    ps.println(desiredWidth+" "+font.getSize()+" "+fontIndex+" "
		       +SetFontName);
	}
    }

    private void outStr(String str) {
	StringBuffer outbuf = new StringBuffer();
	outbuf.append('(');
	for (int i = 0; i < str.length(); i++) {
	    char ch = str.charAt(i);
	    if (ch == '(' || ch == ')' || ch == '\\')
		outbuf.append('\\');
	    outbuf.append(ch);
	}
	outbuf.append(')');
	ps.print(outbuf.toString());
    }

    synchronized void sendLine(PSGraphics psg,
			       int x1, int y1, int x2, int y2) {
	if (psg.pageNumber != curPage) {return;}
	outClip(psg);
	outColor(psg.foreground);
	outPoint(psg, x1, y1);
	outPoint(psg, x2, y2);
	ps.println(DrawLineName);
    }

    synchronized void sendRect(PSGraphics psg,
			       int x, int y, int w, int h,
			       Color color, boolean fill) {
	if (psg.pageNumber != curPage) {return;}
	outClip(psg);
	outColor(color);
	outRect(psg, x, y, w, h);
	ps.println(fill ? RectFillName : RectStrokeName);
    }

    void sendRect(PSGraphics psg, int x, int y, int w, int h, boolean fill) {
	sendRect(psg, x, y, w, h, psg.foreground, fill);
    }

    void fillRect(PSGraphics psg, int x, int y, int w, int h, Color color) {
	sendRect(psg, x, y, w, h, color, true);
    }

    synchronized void sendRoundRect(PSGraphics psg,
				    int x, int y, int w, int h,
				    float xRad, float yRad,
				    boolean fill) {
	if (psg.pageNumber != curPage) {return;}
	outClip(psg);
	outColor(psg.foreground);
	outRect(psg, x, y, -w, -h); // width/height used opposite for roundrect
	ps.print(xRad + " " + yRad + " ");
	ps.println(fill ? RRectFillName : RRectStrokeName);
    }

    synchronized void sendArc(PSGraphics psg,
			      int x, int y, int xRad, int yRad,
			      int startA, int arcA,
			      boolean fill) {
	if (psg.pageNumber != curPage) {return;}
	outClip(psg);
	outColor(psg.foreground);
	outPoint(psg, x, y);
	ps.print(xRad + " " + yRad + " " +
		 startA + " " + arcA + " ");
	ps.println(fill ? ArcFillName : ArcStrokeName);
    }

    synchronized void sendPolygon(PSGraphics psg,
				  int xp[], int yp[], int npoints,
				  boolean close, boolean fill) {
	if (psg.pageNumber != curPage) {return;}
	outClip(psg);
	outColor(psg.foreground);
	ps.print(close+" ");
	for(int i = 0; i < npoints; i++) {
	    outPoint(psg, xp[i], yp[i]);
	}
	ps.print(npoints+" ");
	ps.println(fill ? PolygonFillName : PolygonStrokeName);
    }	

    synchronized void sendString(PSGraphics psg,
				 String str, int x, int y) {
	if (psg.pageNumber != curPage) {return;}
	outClip(psg);
	outColor(psg.foreground);
	Font font = psg.font;
	FontMetrics curMetrics = psg.curMetrics;
	if (curMetrics == null) {
	    curMetrics = psg.getFontMetrics(font);
	    psg.curMetrics = curMetrics;
	}

        if (fontProps == null) {
	    outFont(psg);
	    outStr(str);
	    ps.print(" " + curMetrics.stringWidth(str) + " ");
	    outPoint(psg, x, y);
	    ps.println(DrawStringName);
	    return;
	}

	CharsetString[] acs =
	    ((MFontPeer)
	     (font.getPeer())).makeMultiCharsetString(str);

	for (int i = 0; i < acs.length; i++){
	    // out Font

	    CharsetString cs = acs[i];
	    CharToByteConverter fontCS = cs.fontDescriptor.fontCharset;

	    //FontMetrics curMetrics = psg.getFontMetrics(font);
	    int desiredWidth = curMetrics.charsWidth(cs.charsetChars, cs.offset, cs.length);
	    String charset = makeCharsetName(fontCS.getCharacterEncoding());
	    String style = PlatformFont.styleStr(font.getStyle());
	    String name =
		fontProps.getProperty(font.getName().toLowerCase(),
				      "serif");
	    String fontName =
		fontProps.getProperty(name + "." + charset + "." + style,
				      "Courier ISOF");
	    int fontIndex;
	    try {
		fontIndex =
		    Integer.parseInt(fontProps.getProperty(fontName));
	    } catch(NumberFormatException e){
		fontIndex = 0;
	    }

	    StringBuffer nativeStr = new StringBuffer();
	    byte[] strSeg = new byte[cs.length * 2];
	    int len;
	    try {
		len = fontCS.convert(cs.charsetChars, cs.offset,
                                     cs.offset + cs.length,
				     strSeg, 0, strSeg.length);
	    } catch(CharConversionException e){
		continue;
	    }
	    nativeStr.append('<');
	    for (int j = 0; j < len; j++){
		byte b = strSeg[j];
		// to avoid encoding conversion with println()
		String hexS = Integer.toHexString(b);
		int length = hexS.length();
		if (length > 2) {
		    hexS = hexS.substring(length - 2, length);
		} else if (length == 1) {
		    hexS = "0" + hexS;
		} else if (length == 0) {
		    hexS = "00";
		}
		nativeStr.append(hexS);
	    }
	    nativeStr.append('>');
	    ps.println("/testString " + nativeStr.toString() + " D");
	    ps.println("% Font[" + font.getName() + ", " +
		       PlatformFont.styleStr(font.getStyle()) + ", "
		       + font.getSize() + "]");
	    ps.println(desiredWidth + " " + font.getSize() + " " +
		       fontIndex + " " + SetFontName);

	    // out String
	    ps.println(nativeStr.toString());

	    // draw string
	    int width = curMetrics.charsWidth(cs.charsetChars, cs.offset, cs.length);
	    ps.print(" " + width + " ");
	    outPoint(psg, x, y);
	    ps.println(DrawStringName);
	    x += width;
	}
    }

    /*
     * Currently CharToByteConverter.getCharacterEncoding() return values are
     * not fixed yet. These are used as the part of the key of
     * psfont.propeties. When those name are fixed this routine can
     * be erased.
     */
    private String makeCharsetName(String name){
	if (name.indexOf("8859_1") != -1) {
	    return "latin1";
	} else {
	    return name.toLowerCase();
	}
	
    }

    private final static int blend(int alpha, int fg, int bg) {
	return ((alpha * fg) + ((255 - alpha) * bg)) / 255;
    }

    static private byte hexDigits[] = {
	(byte)'0', (byte)'1', (byte)'2', (byte)'3',
	(byte)'4', (byte)'5', (byte)'6', (byte)'7',
	(byte)'8', (byte)'9', (byte)'A', (byte)'B',
	(byte)'C', (byte)'D', (byte)'E', (byte)'F'
    };

    private int sendColors(byte[] colors, int start, int len,
			   byte outbuf[], int linepos) {
	while (len-- > 0) {
	    if (linepos + 3 > outbuf.length) {
		outbuf[linepos++] = (byte) '\n';
		ps.write(outbuf, 0, linepos);
		linepos = 0;
	    }

	    byte pixel = colors[start++];
	    outbuf[linepos++] = hexDigits[(pixel >>> 4) & 0xf];
	    outbuf[linepos++] = hexDigits[(pixel >>> 0) & 0xf];
	}
	return linepos;
    }

    private int sendColors(int[] colors, int start, int len,
			   byte outbuf[], int linepos) {
	while (len-- > 0) {
	    if (linepos + 7 > outbuf.length) {
		outbuf[linepos++] = (byte) '\n';
		ps.write(outbuf, 0, linepos);
		linepos = 0;
	    }

	    int pixel = colors[start++];
	    outbuf[linepos++] = hexDigits[(pixel >>> 20) & 0xf];
	    outbuf[linepos++] = hexDigits[(pixel >>> 16) & 0xf];
	    outbuf[linepos++] = hexDigits[(pixel >>> 12) & 0xf];
	    outbuf[linepos++] = hexDigits[(pixel >>> 8) & 0xf];
	    outbuf[linepos++] = hexDigits[(pixel >>> 4) & 0xf];
	    outbuf[linepos++] = hexDigits[(pixel >>> 0) & 0xf];
	}
	return linepos;
    }

    private void sendSubPixels(PSGraphics psg,
			       byte[] pixels, int pixscan,
			       int x1, int y1, int x2, int y2,
			       int x, int y, float scalew, float scaleh) {
	int w = x2 - x1;
	int h = y2 - y1;
	int index = y1 * pixscan + x1;
	int linepos = 0;
	byte outbuf[] = new byte[73];

	ps.print(x1 + " " + y1 + " ");
	ps.print(w + " " + h + " ");
	ps.print(scalew*w + " " + scaleh*h + " ");
	outPoint(psg, x, y);
	ps.println(DrawIndexImageName);
	for (int j = y1; j < y2; j++) {
	    linepos = sendColors(pixels, index, w, outbuf, linepos);
	    index += pixscan;
	}
	if (linepos > 0) {
	    outbuf[linepos++] = (byte) '\n';
	    ps.write(outbuf, 0, linepos);
	}
    }

    private void sendSubPixels(PSGraphics psg,
			       int[] pixels, int pixscan,
			       int x1, int y1, int x2, int y2,
			       int x, int y, float scalew, float scaleh) {
	int w = x2 - x1;
	int h = y2 - y1;
	int index = y1 * pixscan + x1;
	int linepos = 0;
	byte outbuf[] = new byte[73];

	ps.print(x1 + " " + y1 + " ");
	ps.print(w + " " + h + " ");
	ps.print(scalew*w + " " + scaleh*h + " ");
	outPoint(psg, x, y);
	ps.println(DrawImageName);
	for (int j = y1; j < y2; j++) {
	    linepos = sendColors(pixels, index, w, outbuf, linepos);
	    index += pixscan;
	}
	if (linepos > 0) {
	    outbuf[linepos++] = (byte) '\n';
	    ps.write(outbuf, 0, linepos);
	}
    }

    private void sendColormap(PSGraphics psg, int cmap[], int size) {
	if (curColormap != null) {
	    boolean matches = true;
	    for (int i = 0; i < size; i++) {
		if (cmap[i] != curColormap[i]) {
		    matches = false;
		    break;
		}
	    }
	    if (matches) {
		return;
	    }
	}

	curColormap = cmap;

	byte outbuf[] = new byte[73];

	ps.println(size + " " + SetColormapName);
	int linepos = sendColors(cmap, 0, size, outbuf, 0);
	if (linepos > 0) {
	    outbuf[linepos++] = (byte) '\n';
	    ps.write(outbuf, 0, linepos);
	}
    }

    synchronized void sendPixels(PSGraphics psg,
				 byte pixels[], IndexColorModel model,
				 int dx1, int dy1, int dx2, int dy2,
				 int sx1, int sy1, int sx2, int sy2,
				 Color bgcolor)
    {
	if (psg.pageNumber != curPage) {return;}
	int cmap[] = new int[256];
	int index = model.getMapSize();
	boolean trans[] = new boolean[256];
	boolean transpixels = false;
	int bgred, bggreen, bgblue;
	if (bgcolor != null) {
	    bgred = bgcolor.getRed();
	    bggreen = bgcolor.getGreen();
	    bgblue = bgcolor.getBlue();
	} else {
	    bgred = bggreen = bgblue = 0;
	}
	while (--index >= 0) {
	    int rgb = model.getRGB(index);
	    int alpha = (rgb >>> 24);
	    if (alpha != 0xff) {
		if (bgcolor != null) {
		    int red = blend(alpha, (rgb >>> 16) & 0xff, bgred);
		    int green = blend(alpha, (rgb >>> 8) & 0xff, bggreen);
		    int blue = blend(alpha, (rgb >>> 0) & 0xff, bgblue);
		    rgb = (red << 16) | (green << 8) | blue;
		} else if (alpha < 128) {
		    trans[index] = true;
		    transpixels = true;
		}
	    }
	    cmap[index] = rgb;
	}
	outClip(psg);
	sendColormap(psg, cmap, model.getMapSize());

	int w = sx2 - sx1;
	int h = sy2 - sy1;
	float scalew = ((float) (dx2 - dx1)) / ((float) w);
	float scaleh = ((float) (dy2 - dy1)) / ((float) h);
	if (transpixels) {
	    index = 0;
	    for (int j = 0; j < h; j++) {
		int startx = -1;
	    	for (int i = 0; i < w; i++) {
		    if (trans[pixels[index++] & 0xff]) { 
		    	if (startx >= 0) {
			    sendSubPixels(psg, pixels, w,
					  startx, j, i, j + 1,
					  dx1, dy1, scalew, scaleh);
			    startx = -1;
		    	}
		    } else if (startx < 0) {
			startx = i;
		    }
	    	}
	    	if (startx >= 0) {
		    sendSubPixels(psg, pixels, w,
				  startx, j, w, j + 1,
				  dx1, dy1, scalew, scaleh);
	    	}
	    }
	} else {
	    sendSubPixels(psg, pixels, w,
			  0, 0, w, h,
			  dx1, dy1, scalew, scaleh);
	}
    }

    synchronized void sendPixels(PSGraphics psg,
				 int pixels[], ColorModel model,
				 int dx1, int dy1, int dx2, int dy2,
				 int sx1, int sy1, int sx2, int sy2,
				 Color bgcolor)
    {
	if (psg.pageNumber != curPage) {return;}
	int w = sx2 - sx1;
	int h = sy2 - sy1;
	float scalew = ((float) (dx2 - dx1)) / ((float) w);
	float scaleh = ((float) (dy2 - dy1)) / ((float) h);
	int bgred, bggreen, bgblue;
	if (bgcolor != null) {
	    bgred = bgcolor.getRed();
	    bggreen = bgcolor.getGreen();
	    bgblue = bgcolor.getBlue();
	} else {
	    bgred = bggreen = bgblue = 0;
	}

	outClip(psg);
	int index = 0;
	int starty = -1;
	for (int j = 0; j < h; j++) {
	    int startx = -1;
	    for (int i = 0; i < w; i++) {
		int rgb = model.getRGB(pixels[index]);
		int alpha = (rgb >>> 24);
		if (alpha != 0xff) { 
		    if (bgcolor != null) {
			int red = blend(alpha, (rgb >>> 16) & 0xff, bgred);
			int green = blend(alpha, (rgb >>> 8) & 0xff, bggreen);
			int blue = blend(alpha, (rgb >>> 0) & 0xff, bgblue);
			pixels[index] = (red << 16) | (green << 8) | blue;
			if (startx < 0) {
			    startx = i;
			}
		    } else if (startx >= 0) {
			if (starty >= 0) {
			    sendSubPixels(psg, pixels, w,
					  0, starty, w, j,
					  dx1, dy1, scalew, scaleh);
			    starty = -1;
			}
			sendSubPixels(psg, pixels, w,
				      startx, j, i, j + 1,
				      dx1, dy1, scalew, scaleh);
			startx = -1;
		    }
		} else {
		    pixels[index] = rgb;
		    if (startx < 0) {
			startx = i;
		    }
		}
		index++;
	    }
	    if (startx == 0) {
		if (starty < 0) {
		    starty = j;
		}
	    } else if (startx > 0) {
		sendSubPixels(psg, pixels, w,
			      startx, j, w, j + 1,
			      dx1, dy1, scalew, scaleh);
	    }
	}
	if (starty >= 0) {
	    sendSubPixels(psg, pixels, w,
			  0, starty, w, h,
			  dx1, dy1, scalew, scaleh);
	}
    }

    void sendProlog() {
	if (prologSent) {
	    return;
	}
	PSPaperSize paperSize
	    = PSPaperSize.find(paperDim.width, paperDim.height, 1); 
	if (paperSize == null) {
	    paperSize = PSPaperSize.find(paperDim.height, paperDim.width, 1); 
	}

	ps.println("%!PS-Adobe-3.0");
	ps.println("%%Title: " + title);
	ps.println("%%Pages: (atend)");
	ps.println("%%PageOrder: " + (lastFirst ? "Special" : "Ascend"));
	ps.println("%%DocumentFonts: Helvetica Helvetica-Bold");
	ps.println("%%+Helvetica-Oblique Helvetica-BoldOblique");
	ps.println("%%+Times-Roman Times-Bold Times-Italic Times-BoldItalic");
	ps.println("%%+Courier Courier-Bold Courier-Oblique Courier-BoldOblique");
	ps.println("%%EndComments");
	ps.println("%%BeginProlog");
	ps.println("/BD {bind def} bind def");
	ps.println("/D {def} BD");
	ps.println("/ISOF {");
	ps.println("	dup findfont dup length 1 add dict begin {");
	ps.println("		1 index /FID eq {pop pop} {D} ifelse");
	ps.println("	} forall /Encoding ISOLatin1Encoding D");
	ps.println("	currentdict end definefont");
	ps.println("} BD");
	ps.println("/F [");
	if (fontProps == null){
	    ps.println("	/Helvetica ISOF");
	    ps.println("	/Helvetica-Bold ISOF");
	    ps.println("	/Helvetica-Oblique ISOF");
	    ps.println("	/Helvetica-BoldOblique ISOF");
	    ps.println("	/Times-Roman ISOF");
	    ps.println("	/Times-Bold ISOF");
	    ps.println("	/Times-Italic ISOF");
	    ps.println("	/Times-BoldItalic ISOF");
	    ps.println("	/Courier ISOF");
	    ps.println("	/Courier-Bold ISOF");
	    ps.println("	/Courier-Oblique ISOF");
	    ps.println("	/Courier-BoldOblique ISOF");
	} else {
	    int count = Integer.parseInt
		(fontProps.getProperty("font.num", "9"));
	    for (int i = 0; i < count; i++){
		ps.println("    /" + fontProps.getProperty
			   ("font." + String.valueOf(i), "Courier ISOF"));
	    }
	}
	ps.println("] D");
	ps.println("/R {4 2 roll moveto 1 index 0 rlineto");
	ps.println("	0 exch rlineto neg 0 rlineto closepath} BD");
	ps.println("/Adict 8 dict D");
	ps.println("Adict /mtrx matrix put");
	ps.println("/A { Adict begin");
	ps.println("    /endangle exch D /startangle exch D");
	ps.println("    /yrad exch D /xrad exch D");
	ps.println("	/y exch D /x exch D");
	ps.println("    /savematrix mtrx currentmatrix D");
	ps.println("    x y translate xrad yrad scale");
	ps.println("    0 0 1 startangle endangle arc");
	ps.println("    savematrix setmatrix");
	ps.println("    end } BD");
	ps.println("/PY {");
	ps.println("    dup 1 ge {3 1 roll moveto} if");
	ps.println("    -1 2 {pop lineto} for");
	ps.println("    {closepath} if");
	ps.println("} BD");
	ps.println("/CL { dup 0 exch -1 1 arct} BD");
	ps.println("/RCL { dup 0 exch 1 1 arct} BD");
	ps.println("/RRdict 8 dict D");
 	ps.println("RRdict /mtrx matrix put");
	ps.println("/RR {RRdict begin");
	ps.println("    /yrad exch D /xrad exch D");
	ps.println("    /height exch D /width exch D");
	ps.println("    /y exch D /x exch D");
	ps.println("	/width width xrad div D /height height yrad div D");
	ps.println("	/savematrix mtrx currentmatrix D");
	ps.println(" 	x y translate xrad yrad scale -90 rotate");
	ps.println("	1 0 moveto height RCL height 0 translate -90 rotate");
	ps.println("    width CL width 0 translate -90 rotate");
	ps.println("	height RCL height 0 translate -90 rotate width CL closepath");
	ps.println("	savematrix setmatrix");
	ps.println("	end } BD");	 
	ps.println("/"+SetColorName
		   +" {3 {255 div 3 1 roll} repeat setrgbcolor} BD");
	ps.println("/"+SetFontName +" {");
	ps.println("	F exch get exch scalefont setfont");
	if (fontProps == null) {
	    ps.println("	("+testString+") stringwidth pop div");
	} else {
	    ps.println(" testString stringwidth pop div");
	}
	ps.println("	currentfont exch scalefont setfont} BD");
	ps.println("/"+RectClipName+" {R clip newpath} BD");
	ps.println("/"+InitClipName+" {initclip} BD");
	ps.println("/"+RectFillName+" {R fill} BD");
	ps.println("/"+RectStrokeName+" {R stroke} BD");
	ps.println("/"+RRectFillName+" {RR fill} BD");
	ps.println("/"+RRectStrokeName+" {RR stroke} BD");
	ps.println("/"+ArcStrokeName+" {A stroke} BD");
	ps.println("/"+ArcFillName+" {6 -2 roll 2 copy moveto 6 2 roll A fill} BD");
	ps.println("/"+PolygonStrokeName+" {PY stroke} BD");
	ps.println("/"+PolygonFillName+" {PY eofill} BD");
	ps.println("/NZ {dup 1 lt {pop 1} if} BD");
	ps.println("/"+DrawStringName +" {");
	ps.println("	moveto 1 index stringwidth pop NZ sub");
	ps.println("	1 index length 1 sub NZ div 0");
	ps.println("	3 2 roll ashow newpath} BD");
	ps.println("/"+DrawLineName+" {moveto lineto stroke} BD");
	ps.println("/"+ScaleName+" {scale} BD");
	ps.println("/"+StartPageName+" {/P save D} BD");
	ps.println("/"+EndPageName+" {showpage P restore} BD");
	ps.println("/"+GsaveName+" {gsave} BD");
	ps.println("/"+GrestoreName+" {grestore} BD");
	ps.println("/"+SetColormapName+" {");
	ps.println("	/cmapr 256 array D");
	ps.println("	/cmapg 256 array D");
	ps.println("	/cmapb 256 array D");
	ps.println("	/cmapgray null D");
	ps.println("	3 string exch 0 exch 1 exch 1 sub {");
	ps.println("		currentfile 2 index readhexstring pop");
	ps.println("		cmapr 2 index 2 index 0 get put");
	ps.println("		cmapg 2 index 2 index 1 get put");
	ps.println("		cmapb 2 index 2 index 2 get put");
	ps.println("	pop pop } for pop");
	ps.println("} BD");
	ps.println("/"+DrawIndexImageName+" {");
	ps.println("	GS");
	ps.println("	/imgsave save D");	// NeWSprint leaves files open!
	ps.println("	translate scale");
	ps.println("	/imgh exch D");
	ps.println("	/imgw exch D");
	ps.println("	/imgy exch neg D");
	ps.println("	/imgx exch neg D");
	ps.println("	/imgstr imgw string D");
	ps.println("	/colorimage where {");
	ps.println("		pop");
	ps.println("		/cimgstr imgw 3 mul string D");
	ps.println("		imgw imgh 8 [imgw 0 0 imgh neg imgx imgy] {");
	ps.println("			currentfile imgstr readhexstring pop");
	ps.println("			0 1 imgw 1 sub {");
	ps.println("				2 copy get");
	ps.println("				cmapr 1 index get");
	ps.println("				cimgstr exch 3 index");
	ps.println("				3 mul exch put");
	ps.println("				cmapg 1 index get");
	ps.println("				cimgstr exch 3 index");
	ps.println("				3 mul 1 add exch put");
	ps.println("				cmapb 1 index get");
	ps.println("				cimgstr exch 3 index");
	ps.println("				3 mul 2 add exch put");
	ps.println("			pop pop } for pop cimgstr");
	ps.println("		} false 3 colorimage");
	ps.println("	} {");
	ps.println("		cmapgray null eq {");
	ps.println("			/cmapgray cmapr length array D");
	ps.println("			0 1 cmapr length 1 sub {");
	ps.println("				cmapgray exch");
	ps.println("				cmapr 1 index get .299 mul");
	ps.println("				cmapg 2 index get .587 mul");
	ps.println("				cmapb 3 index get .114 mul");
	ps.println("				add add cvi put");
	ps.println("			} for");
	ps.println("		} if");
	ps.println("		imgw imgh 8 [imgw 0 0 imgh neg 0 0] {");
	ps.println("			currentfile imgstr readhexstring pop");
	ps.println("			0 1 imgw 1 sub {");
	ps.println("				imgstr exch 2 copy get");
	ps.println("				cmapgray exch get put");
	ps.println("			} for");
	ps.println("		} image");
	ps.println("	} ifelse");
	ps.println("	imgsave restore");
	ps.println("	GR");
	ps.println("} BD");
	ps.println("/"+DrawImageName+" {");
	ps.println("	GS");
	ps.println("	/imgsave save D");	// NeWSprint leaves files open!
	ps.println("	translate scale");
	ps.println("	/imgh exch D");
	ps.println("	/imgw exch D");
	ps.println("	/imgy exch neg D");
	ps.println("	/imgx exch neg D");
	ps.println("	/imgstr imgw 3 mul string D");
	ps.println("	imgw imgh 8 [imgw 0 0 imgh neg imgx imgy] {");
	ps.println("		currentfile imgstr readhexstring pop");
	ps.println("	} bind false 3 colorimage");
	ps.println("    imgsave restore");
	ps.println("    GR");		
	ps.println("} BD");
	ps.println("%%EndProlog");
	ps.println("%%BeginSetup");
	if (paperSize != null) {
	    ps.println(paperSize.toPostScriptCode(1, false));
	}
	if (paperDim.width > paperDim.height)
	    ps.println("90 rotate 0 "+-(paperDim.height)+" translate");
	ps.println("%%EndSetup");
	ps.flush();
	prologSent = true;
    }
}
