/*
 * @(#)UPrintDialog.java	1.16 98/07/30
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.motif;

import java.awt.*;
import java.util.*;

 /**
  * A class which implements a unix modal print dialog.
  *
  * @version     1.16 07/30/98
  * @author      Amy Fowler
  */
class UPrintDialog extends Dialog {

    static final long serialVersionUID = -599066638841164608L;

    /**
     * Waiting print status (user response pending).
     */
    public final static int WAITING = 0;

    /**
     * Print print status (user activated "print").
     */
    public final static int PRINT = 1;

    /**
     * Cancelled print status (user activated "Cancel");
     */
    public final static int CANCELLED = 2;

    /**
     * Print status used to determine result from modal dialog.
     */
    public int printStatus = WAITING;

    // The print control to invoke to create the print job from
    // dialog properties.
    private PrintControl printControl;

    // The print job created as a result of user selecting "print"
    private PrintJob printJob;

    // Print destination variable.
    private int destType;

    // Print orientation variable.
    private int orientation;

    // Print paper size variable.
    private int paperSize;

    // Print Target Title
    private Label docTitle;

    // Copies Field
    private NumericField copiesField;

    // Print Destination Radiobox and Fields
    private CheckboxGroup destCBG;
    private Checkbox destCB[] = new Checkbox[PrintControl.DEST_N];
    private TextField destField[] = new TextField[PrintControl.DEST_N];
    private static int FIELD_LEN = 28;

    // Print Options
    private Label bannerLbl;
    private TextField bannerField;
    private Label optionsLbl;
    private TextField optionsField;

    // Orientation Radiobox
    private CheckboxGroup orientCBG;
    private Checkbox orientCB[] = new Checkbox[PrintControl.ORIENT_N];

    // Paper Size Radiobox
    private CheckboxGroup paperCBG;
    private Checkbox paperCB[] = new Checkbox[PrintControl.PAPER_N];

    // Dialog buttons
    private Button printButton;
    private Button cancelButton;
    private final static String PRINT_LABEL = new String(getMsg("button.print"));
    private final static String CANCEL_LABEL = new String(getMsg("button.cancel"));

    private static ResourceBundle messageRB;

    /**
     * Initialize ResourceBundle
     */
    static void initResource() {
	java.security.AccessController.doPrivileged(
				    new java.security.PrivilegedAction() {
	    public Object run() {
		try {
		    messageRB =
			ResourceBundle.getBundle(
					 "sun.awt.motif.resources.updialog");
		    return null;
		} catch (MissingResourceException e) {
		    throw
		     new Error("Fatal: Resource for PrintDialog is missing.");
		}
	    }
	});
    }

    /**
     * get message string from resource
     */
    static String getMsg(String key) {
	if (messageRB == null) {
	    initResource();
	}
	try {
	    return messageRB.getString(key);
	} catch (MissingResourceException e) {
	    throw new Error("Fatal: Resource for PrintDialog is broken, There is no " + key + " key in resource");
	}
    }

    /**
     * method to add component to dialog with specified gridbag constraints.
     */
    private void addComponent(Component comp, Container cont, GridBagLayout gridbag,
		GridBagConstraints c, int gridx, int gridy, int gridw,
	   	int weightx, int weighty, int anchor, int fill, Insets insets) {
	c.gridx = gridx;
	c.gridy = gridy;
	c.gridwidth = gridw;
	c.anchor = anchor;
	c.fill = fill;
	c.insets = insets;
	c.weightx = weightx;
	c.weighty = weighty;
	gridbag.setConstraints(comp, c);
	cont.add(comp);
    }

    /**
     * Creates a modal unix print dialog.
     */
    public UPrintDialog(Frame parent) {
	super(parent, getMsg("dialog.title"), true);

	Font font, boldFont;
	Insets noInsets = new Insets(0, 0, 0, 0);
	Insets topInsets = new Insets(10, 0, 0, 0);
	Insets topRightInsets = new Insets(8, 6, 0, 10);
	Insets leftInsets = new Insets(0, 40, 0, 0);
	Insets topLeftInsets = new Insets(10, 6, 0, 0);
	Insets bottomInsets = new Insets(0, 6, 10, 10);
	Insets rightInsets = new Insets(0, 0, 0, 10);
	Label label;
	int i;

	GridBagLayout gridbag = new GridBagLayout();
	GridBagConstraints c = new GridBagConstraints();
	setLayout(gridbag);

	font = getFont();
	boldFont = new Font("Dialog", Font.BOLD, font.getSize());
	setFont(boldFont);

	//
	// Create Print Target Label
	//
	label = new Label(getMsg("label.print"), Label.RIGHT);
	addComponent((Component)label, (Container)this, gridbag, c, 0, 0, 1, 0, 1,
			GridBagConstraints.EAST, GridBagConstraints.NONE, topInsets);
	docTitle = new Label("");
	docTitle.setFont(font);
	addComponent((Component)docTitle, (Container)this, gridbag, c, 1, 0, 3, 1, 1,
			GridBagConstraints.WEST, GridBagConstraints.BOTH, topLeftInsets);

	//
	// Create Copies Field
	//
	label = new Label(getMsg("label.copies"), Label.RIGHT);
	addComponent((Component)label, (Container)this, gridbag, c, 0, 1, 1, 0, 1,
			GridBagConstraints.EAST, GridBagConstraints.NONE, topInsets);
	copiesField = new NumericField(3);
	copiesField.setFont(font);
	addComponent((Component)copiesField, (Container)this, gridbag, c, 1, 1, 1, 0, 1,
			GridBagConstraints.WEST, GridBagConstraints.NONE, topLeftInsets);
	setNumCopies(1);

	//
	// Create Print Destination RadioBox
	//
	label = new Label(getMsg("label.printto"), Label.RIGHT);
	addComponent((Component)label, (Container)this, gridbag, c, 0, 2, 1, 0, 1,
			GridBagConstraints.EAST, GridBagConstraints.NONE, topInsets);

	destCBG = new CheckboxGroup();

	destCB[PrintControl.PRINTER] = new Checkbox(getMsg("checkbox.printer"), destCBG, true);
	addComponent((Component)destCB[PrintControl.PRINTER], (Container)this, gridbag,
                     c, 0, 3, 2, 0, 1, GridBagConstraints.WEST,
                     GridBagConstraints.NONE, leftInsets);
	destField[PrintControl.PRINTER] = new TextField(FIELD_LEN);
	destField[PrintControl.PRINTER].setFont(font);
	addComponent((Component)destField[PrintControl.PRINTER], (Container)this,
                     gridbag, c, 2, 3, 2, 1, 1, GridBagConstraints.CENTER,
                     GridBagConstraints.HORIZONTAL, rightInsets);

	destCB[PrintControl.FILE] = new Checkbox(getMsg("checkbox.file"), destCBG, false);
	addComponent((Component)destCB[PrintControl.FILE], (Container)this, gridbag,
                     c, 0, 4, 2, 0, 1,GridBagConstraints.WEST,
                     GridBagConstraints.NONE, leftInsets);
	destField[PrintControl.FILE] = new TextField(FIELD_LEN);
	destField[PrintControl.FILE].setFont(font);
	addComponent((Component)destField[PrintControl.FILE], (Container)this, gridbag,
                     c, 2, 4, 2, 1, 1, GridBagConstraints.CENTER,
                     GridBagConstraints.HORIZONTAL, rightInsets);

	destType = PrintControl.PRINTER;
	destField[PrintControl.FILE].disable();

	//
	// Create Options Fields
	//
	bannerLbl = new Label(getMsg("label.banner"), Label.RIGHT);
	addComponent((Component)bannerLbl, (Container)this, gridbag,
                     c, 0, 5, 3, 0, 1, GridBagConstraints.EAST,
                     GridBagConstraints.NONE, topInsets);
	bannerField = new TextField(24);
	bannerField.setFont(font);
	addComponent((Component)bannerField, (Container)this, gridbag,
                     c, 3, 5, 1, 1, 1, GridBagConstraints.WEST,
                     GridBagConstraints.HORIZONTAL, topRightInsets);

	optionsLbl = new Label(getMsg("label.commandoption"), Label.RIGHT);
	addComponent((Component)optionsLbl, (Container)this, gridbag,
                     c, 0, 6, 3, 0, 1, GridBagConstraints.EAST,
                     GridBagConstraints.VERTICAL, noInsets);
	optionsField = new TextField(24);
	optionsField.setFont(font);
	addComponent((Component)optionsField, (Container)this, gridbag,
                     c, 3, 6, 1, 1, 1, GridBagConstraints.WEST,
                     GridBagConstraints.HORIZONTAL, bottomInsets);

	Separator sep = new Separator(100, 4, Separator.HORIZONTAL);
	addComponent((Component)sep, (Container)this, gridbag,
                     c, 0, 7, 4, 1, 1, GridBagConstraints.CENTER,
                     GridBagConstraints.HORIZONTAL, noInsets);


	Panel pPanel = new Panel();
	GridBagLayout pGridbag = new GridBagLayout();
	pPanel.setLayout(pGridbag);
	addComponent((Component)pPanel, (Container)this, gridbag,
                     c, 0, 8, 4, 1, 1, GridBagConstraints.WEST,
                     GridBagConstraints.HORIZONTAL, noInsets);

	//
	// Create Paper Size RadioBox
	//
	label = new Label(getMsg("label.papersize"), Label.RIGHT);
	addComponent((Component)label, (Container)pPanel, pGridbag,
                     c, 0, 0, 1, 1, 1, GridBagConstraints.WEST,
                     GridBagConstraints.NONE, topInsets);
	Panel panel = new Panel();
	panel.setLayout(new GridLayout(2,2));

	addComponent((Component)panel, (Container)pPanel, pGridbag,
                     c, 0, 1, 1, 1, 1, GridBagConstraints.CENTER,
                     GridBagConstraints.NONE, bottomInsets);
	paperCBG = new CheckboxGroup();
	paperCB[PrintControl.PAPER_LETTER] = new Checkbox(getMsg("checkbox.letter"), paperCBG, true);
	panel.add(paperCB[PrintControl.PAPER_LETTER]);
	paperCB[PrintControl.PAPER_EXECUTIVE] = new Checkbox(getMsg("checkbox.executive"), paperCBG, false);
	panel.add(paperCB[PrintControl.PAPER_EXECUTIVE]);
	paperCB[PrintControl.PAPER_LEGAL] = new Checkbox(getMsg("checkbox.legal"), paperCBG, false);
	panel.add(paperCB[PrintControl.PAPER_LEGAL]);
	paperCB[PrintControl.PAPER_A4] = new Checkbox(getMsg("checkbox.a4"), paperCBG, false);
	panel.add(paperCB[PrintControl.PAPER_A4]);
	paperSize = PrintControl.PAPER_LETTER;

	//
	// Create Orientation RadioBox
	//
	label = new Label(getMsg("label.orientation"), Label.RIGHT);
	addComponent((Component)label, (Container)pPanel, pGridbag,
                     c, 1, 0, 1, 1, 1, GridBagConstraints.WEST,
                     GridBagConstraints.NONE, topInsets);
	panel = new Panel();
	panel.setLayout(new GridLayout(2,1));
	addComponent((Component)panel, (Container)pPanel, pGridbag,
                     c, 1, 1, 1, 1, 1, GridBagConstraints.CENTER,
                     GridBagConstraints.NONE, bottomInsets);
	orientCBG = new CheckboxGroup();
	orientCB[PrintControl.PORTRAIT] = new Checkbox(getMsg("checkbox.portrait"), orientCBG, true);
	panel.add(orientCB[PrintControl.PORTRAIT]);
	orientCB[PrintControl.LANDSCAPE] = new Checkbox(getMsg("checkbox.landscape"), orientCBG, false);
	panel.add(orientCB[PrintControl.LANDSCAPE]);
	orientation = PrintControl.PORTRAIT;

	sep = new Separator(100, 4, Separator.HORIZONTAL);
	addComponent((Component)sep, (Container)this, gridbag,
                     c, 0, 9, 4, 1, 0, GridBagConstraints.WEST,
                     GridBagConstraints.HORIZONTAL, noInsets);

	//
	// Create Dialog Button Panel
	//
	panel = new Panel();
	panel.setLayout(new FlowLayout(FlowLayout.CENTER, 28, 10));
	addComponent((Component)panel, (Container)this, gridbag,
                     c, 0, 10, 4, 1, 0, GridBagConstraints.CENTER,
                     GridBagConstraints.NONE, noInsets);
	panel.add(printButton = new Button(PRINT_LABEL));
	panel.add(cancelButton = new Button(CANCEL_LABEL));

	pack();
    }

    private void setupDestType(int newType) {
	destField[destType].disable();
        destField[newType].enable();
	if (newType == PrintControl.PRINTER) {
	    bannerLbl.enable();
	    bannerField.enable();
	    optionsLbl.enable();
	    optionsField.enable();
	} else {
	    bannerLbl.disable();
	    bannerField.disable();
	    optionsLbl.disable();
	    optionsField.disable();
	}
	destType = newType;
    }

    public PrintJob getPrintJob(PrintControl printControl) {
        this.printControl = printControl;
        printJob = null;
        show();
        return printJob;
    }

    public boolean action(Event event, Object arg) {
    	int i;
	if (event.target instanceof Checkbox) {
	    for (i=0; i < PrintControl.DEST_N; i++) {
    	    	if (event.target == destCB[i]) {
		    if (i != destType) {
			setupDestType(i);
			destField[i].requestFocus();
		    }
		    return true;
		}
	    }
	    for (i=0; i < PrintControl.ORIENT_N; i++) {
		if (event.target == orientCB[i]) {
		    orientation = i;
		    return true;
		}
	    }
	    for (i=0; i < PrintControl.PAPER_N; i++) {
		if (event.target == paperCB[i]) {
		    paperSize = i;
		    return true;
		}
	    }
	}
	if (PRINT_LABEL.equals(arg)) {
	    printStatus = PRINT;
            printJob = printControl.createPrintJob(destType, getDestString(destType),
                                                   getOptionsString(), getBannerString(),
                                                   getPaperSize(), getOrientation(),
                                                   getNumCopies());
            if (printJob != null) {
	        hide();
            }
	    return true;

	} else if (CANCEL_LABEL.equals(arg)) {
	    printStatus = CANCELLED;
	    hide();
	    return true;
	}
	return super.action(event, arg);
    }

    public boolean handleEvent(Event evt) {
	if (evt.id == Event.WINDOW_DESTROY) {
	    printStatus = CANCELLED;
	    hide();
	    return true;
	}
	return super.handleEvent(evt);
    }

    /**
     * Return the current number of print copies.
     */
    public int getNumCopies() {
	int copies = 0;
  	try {
	    copies = copiesField.getValue();
	} catch (NumberFormatException e) {
	}
	return copies;
    }

    /**
     * Return the current print destination type.
     */
    public int getDestType() {
	return destType;
    }

    /**
     * Return the destination string for the specified destination type.
     */
    public String getDestString(int dest) {
	return destField[dest].getText();
    }

    /**
     * Return the banner page title string.
     */
    public String getBannerString() {
	return bannerField.getText();
    }

    /**
     * Return the print options string.
     */
    public String getOptionsString() {
	return optionsField.getText();
    }

    /**
     * Return the current print orientation.
     */
    public int getOrientation() {
	return orientation;
    }

    /**
     * Returns the current paper size.
     */
    public int getPaperSize() {
	return paperSize;
    }

    /**
     * Sets the title of the printed document.
     */
    public void setDocumentTitle(String title) {
	docTitle.setText(title);
    }

    /**
     * Sets the number of print copies.
     */
    public void setNumCopies(int numCopies) {
	copiesField.setValue(numCopies);
    }

    /**
     * Sets the current print destination to the specified type.
     */
    public void setDestType(int dest) {
	if (dest != destType) {
	    destCB[dest].setState(true);
	    setupDestType(dest);
	}
    }

    /**
     * Sets the destination string of the specified destination type.
     */
    public void setDestString(int dest, String destStr) {
	destField[dest].setText(destStr);
    }

    /**
     * Sets the banner page title.
     */
    public void setBannerString(String title) {
	bannerField.setText(title);
    }

    /**
     * Sets the print options string.
     */
    public void setOptionsString(String options) {
	optionsField.setText(options);
    }

    /**
     * Sets the current print orientation to the specified type.
     */
    public void setOrientation(int orient) {
	if (orient != orientation) {
	    orientation = orient;
	    orientCB[orient].setState(true);
	}
    }

    /**
     * Sets the current print paper size to the specified type.
     */
    public void setPaperSize(int paper) {
	if (paper != paperSize) {
	    paperSize = paper;
	    paperCB[paper].setState(true);
	}
    }
}
