/*
 * @(#)TinyComponentPeer.java	1.27 98/09/21
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.awt.tiny;

import java.awt.*;
import java.awt.peer.*;
import sun.awt.*;
import sun.awt.image.ImageRepresentation;
import sun.awt.image.OffScreenImage;
import java.awt.image.ImageProducer;
import java.awt.image.ImageObserver;
import java.awt.image.ColorModel;
import java.awt.image.DirectColorModel;

abstract 
class TinyComponentPeer extends TinyWindow implements ComponentPeer, UpdateClient {
    // Add this total amount of padding to the width and height of any
    // component that is sized according to a corresponding text string
    static final int TEXT_XPAD = 8;
    static final int TEXT_YPAD = 6;

    Component	target;
    Rectangle	repaintRect;
    boolean	enabled;
    boolean	ancestorsEnabled;

    /**
     * Initialize the component.
     */
    void init(Component target) {
	this.target = target;

	if (target.isVisible()) {
	    show();
	}

	Color c = target.getBackground();
	if (c != null) {
	    winBackground(c);
	}

	Rectangle r = target.getBounds();
	setBounds(r.x, r.y, r.width, r.height);
	setCursor(target.getCursor());
	enabled = target.isEnabled();
	
	ancestorsEnabled = true;
	Component p = target.getParent();
	while ((p != null) && ancestorsEnabled) {
	    ancestorsEnabled = p.isEnabled();
	    p = p.getParent();
	}
    }

    /*
     * Repaint
     */
    synchronized Rectangle getRepaintRect() {
	Rectangle r = repaintRect;
	repaintRect = null;
	return r;
    }
    public int updatePriority() {
	return Thread.NORM_PRIORITY;
    }
    public void updateClient(Object arg) {
	Rectangle r = getRepaintRect();
	if (r != null) {
	    Graphics g = new TinyGraphics(this);
	    try {
		g.setColor(target.getForeground());
		g.setFont(target.getFont());
		g.clipRect(r.x, r.y, r.width, r.height);
		update(g);
	    } finally {
		g.dispose();
	    }
	}
    }
    public synchronized void repaint(long tm, int x, int y, int width, int height) {
	if (repaintRect == null) {
	    repaintRect = new Rectangle(x, y, width, height);
	} else {
	    repaintRect = repaintRect.union(new Rectangle(x, y, width, height));
	}
	ScreenUpdater.updater.notify(this, tm);
    }
    void repaint() {
	repaint(0, 0, 0, width, height);
    }

    public Color getWinBackground() {
	return target.getBackground();
    }

    static final int BACKGROUND_COLOR = 0;
    static final int HIGHLIGHT_COLOR = 1;
    static final int SHADOW_COLOR = 2;
    static final int FOREGROUND_COLOR = 3;
    public Color[] getGUIcolors() {
	Color c[] = new Color[4];
	c[BACKGROUND_COLOR] = getWinBackground();
	if (c[BACKGROUND_COLOR] == null) {
	    c[BACKGROUND_COLOR] = Color.lightGray;
	}

	c[HIGHLIGHT_COLOR] = c[BACKGROUND_COLOR].brighter();
	c[SHADOW_COLOR] = c[BACKGROUND_COLOR].darker();
	c[FOREGROUND_COLOR] = target.getForeground();
	if (c[FOREGROUND_COLOR] == null) {
	    c[FOREGROUND_COLOR] = Color.black;
	}

	if (   (c[BACKGROUND_COLOR].equals(c[HIGHLIGHT_COLOR])) 
	    && (c[BACKGROUND_COLOR].equals(c[SHADOW_COLOR]))) {
	    c[SHADOW_COLOR] = new Color(c[BACKGROUND_COLOR].getRed()   + 75,
					c[BACKGROUND_COLOR].getGreen() + 75,
					c[BACKGROUND_COLOR].getBlue()  + 75);
	    c[HIGHLIGHT_COLOR] = c[SHADOW_COLOR].brighter();
	} else if (c[BACKGROUND_COLOR].equals(c[HIGHLIGHT_COLOR])) {
	    c[HIGHLIGHT_COLOR] = c[SHADOW_COLOR];
	    c[SHADOW_COLOR] = c[SHADOW_COLOR].darker();
	}

	if (! isEnabled()) {
	    // Reduce the contrast
	    // Calculate the NTSC gray (NB: REC709 L* might be better!)
	    // for foreground and background; then multiply the foreground
	    // by the average lightness

	    Color tc = c[BACKGROUND_COLOR];
	    int bg = tc.getRed() * 30 + tc.getGreen() * 59 + tc.getBlue() * 11;

	    tc = c[FOREGROUND_COLOR];
	    int fg = tc.getRed() * 30 + tc.getGreen() * 59 + tc.getBlue() * 11;

	    float ave = (float) ((fg + bg) / 51000.0);  // 255 * 100 * 2

	    Color newForeground = new Color((int) (tc.getRed()   * ave),
					    (int) (tc.getGreen() * ave),
					    (int) (tc.getBlue()  * ave));

	    if (newForeground.equals(c[FOREGROUND_COLOR])) {
		// This probably means the foreground color is black or white
		newForeground = new Color(ave, ave, ave);
	    }
	    c[FOREGROUND_COLOR] = newForeground;
	}					   
	return c;
    }

    /**
     * Draw a 3D oval.
     */
    public void draw3DOval(Graphics g, Color colors[],
		    int x, int y, int w, int h, boolean raised) {
	Color c = g.getColor();
	g.setColor(raised ? colors[HIGHLIGHT_COLOR] : colors[SHADOW_COLOR]);
	g.drawArc(x, y, w, h, 45, 180);
	g.setColor(raised ? colors[SHADOW_COLOR] : colors[HIGHLIGHT_COLOR]);
	g.drawArc(x, y, w, h, 225, 180);
	g.setColor(c);
    }

    public void draw3DRect(Graphics g, Color colors[],
			   int x, int y, int width, int height,
			   boolean raised) {
	Color c = g.getColor();
	g.setColor(raised ? colors[HIGHLIGHT_COLOR] : colors[SHADOW_COLOR]);
	g.drawLine(x, y, x, y + height);
	g.drawLine(x + 1, y, x + width - 1, y);
	g.setColor(raised ? colors[SHADOW_COLOR] : colors[HIGHLIGHT_COLOR]);
	g.drawLine(x + 1, y + height, x + width, y + height);
	g.drawLine(x + width, y, x + width, y + height - 1);
	g.setColor(c);
    }    

    void paintComponent(Graphics g) {
	g.setColor(target.getForeground());
	g.drawRect(0, 0, width-1, height-1);
	g.setColor(target.getBackground());
	g.fillRect(1, 1, width-2, height-2);
    }
    public void paint(Graphics g) {
	paintComponent(g);
	g.setColor(target.getForeground());
	g.setFont(target.getFont());
	target.paint(g);
    }
    public void update(Graphics g) {
	paint(g);
    }
    public void print(Graphics g) {
	paintComponent(g);
	g.setColor(target.getForeground());
	g.setFont(target.getFont());
	target.print(g);
    }

    public Graphics getGraphics() {
	Graphics g = new TinyGraphics(this);
	g.setColor(target.getForeground());
	g.setFont(target.getFont());
	return g;
    }


    public Point getLocationOnScreen() {
	throw new InternalError("not implemented yet");
    }

    /*
     * Changing parameters
     */

    public void setBounds(int x, int y, int w, int h) {
	reshape(x, y, w, h);
    }

    public void setVisible(boolean b) {
	if (b) {
    	    winShow();
    	} else {
    	    winHide();
    	}
    }

    public void setEnabled(boolean value) {
	boolean repaintNeeded = (enabled != value);
	enabled = value;
	if (target instanceof Container) {
	    Component list[] = ((Container)target).getComponents();
	    for (int i = 0; i < list.length; ++i) {
		TinyComponentPeer p = (TinyComponentPeer) (list[i].getPeer());
		p.setEnabled(value);
	    }
	}
	if (repaintNeeded) {
	    repaint();
	}
    }

    protected boolean isEnabled() {
	return enabled && ancestorsEnabled;
    }

    public void setFont(Font f) {
	if (! target.getFont().equals(f)) {
	    repaint();
	}
    }
    public void setBackground(Color c) {
	boolean repaintNeeded = ! getWinBackground().equals(c);
	winBackground(c);
	if (repaintNeeded) {
	    repaint();
	}
    }
    public void setForeground(Color c) {
	if (! target.getForeground().equals(c)) {
	    repaint();
	}
    }

    public void setCursor(Cursor c) {
	if (c == null) {
	    winSetCursor(Cursor.DEFAULT_CURSOR);
	} else {
	    winSetCursor(c.getType());
	}
    }

    public Dimension getMinimumSize() {
	return target.size();
    }

    public Dimension getPreferredSize() {
	return getMinimumSize();
    }

    public boolean isFocusTraversable() {
	return false;
    }

    public ColorModel getColorModel() {
	return TinyToolkit.getStaticColorModel();
    }

    public FontMetrics getFontMetrics(Font font) {
	return TinyFontMetrics.getFontMetrics(font);
    }

    public java.awt.Toolkit getToolkit() {
	return Toolkit.getDefaultToolkit();
    }

    public void dispose() {
	ScreenUpdater.updater.removeClient(this);
	winDispose();
	target = null;
    }

    public String toString() {
	return getClass().getName() + "[" + target + "]";
    }

    public boolean targetEvent(int id, Object arg) {
	return targetEvent(new Event(target, id, arg));
    }

    public boolean targetEvent(Event evt) {
	try {
	    evt.target = target;
	    return target.postEvent(evt);
	} catch (ThreadDeath e) {
	    throw e;
	} catch (Throwable e) {
	    e.printStackTrace();
	}
	return false;
    }

    void handleFocusEvent(Event evt) {
	handleWindowEvent(evt);
	targetEvent(evt);
    }

    boolean handleKeyEvent(Event evt) {
	// NEW_EVENT_MODEL: Allow target to handle event first
	if (targetEvent(evt)) {
	    return true;
	}
	evt.target = this;
	return handleWindowEvent(evt);
    }

    public boolean handleEvent(Event evt) {
	return false;
    }

    public void handleEvent(AWTEvent e) {
    }

    public Image createImage(ImageProducer producer) {
	return new TinyImage(producer);
    }
    public Image createImage(int width, int height) {
	return new OffScreenImage(target, width, height);
    }
    public boolean prepareImage(Image img, int w, int h, ImageObserver o) {
	return TinyToolkit.prepareScrImage(img, w, h, o);
    }
    public int checkImage(Image img, int w, int h, ImageObserver o) {
	return TinyToolkit.checkScrImage(img, w, h, o);
    }

    public void beginValidate() {}
    public void endValidate() {}

    /**
     * DEPRECATED:  Replaced by setVisible(boolean).
     */
    public void show() {
	setVisible(true);
    }

    /**
     * DEPRECATED:  Replaced by setVisible(boolean).
     */
    public void hide() {
	setVisible(false);
    }

    /**
     * DEPRECATED:  Replaced by setEnabled(boolean).
     */
    public void enable() {
	setEnabled(true);
    }

    /**
     * DEPRECATED:  Replaced by setEnabled(boolean).
     */
    public void disable() {
	setEnabled(false);
    }

    /**
     * DEPRECATED:  Replaced by setBounds(int, int, int, int).
     */
    public void reshape(int x, int y, int width, int height) {
	winReshape(x, y, width, height);
    }

    /**
     * DEPRECATED:  Replaced by getMinimumSize().
     */
    public Dimension minimumSize() {
	return getMinimumSize();
    }

    /**
     * DEPRECATED:  Replaced by getPreferredSize().
     */
    public Dimension preferredSize() {
	return getPreferredSize();
    }

}
