/*
 * @(#)TinyGraphics.java	1.24 98/09/21
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.awt.tiny;

import java.awt.*;
import java.io.*;
import java.util.*;
import java.awt.geom.AffineTransform;
import java.awt.image.BilinearAffineTransformOp;
import java.awt.image.BufferedImage;
import java.awt.image.BufferedImageOp;
import java.awt.image.ByteDiscreteChannel;
import java.awt.image.ColorModel;
import java.awt.image.IntDiscreteChannel;
import java.awt.image.Tile;
import java.awt.geom.Transform;
import java.awt.image.ImageObserver;
import sun.awt.image.ImageRepresentation;
import sun.awt.image.GraphicsObserver;
import sun.awt.image.ScaledGraphicsObserver;
import sun.awt.image.StretchedGraphicsObserver;

// LOOK HERE
// There are some unresolved issues in this class w.r.t. multithreading.
// For example, what happens when you interleave a translate() with
// any of the methods that "adjust for the origin and pass onto an
// implementation".
// We clearly need to move locking a little higher.  Perhaps
// just the global AWT_LOCK() style stuff


/**
 * Graphics is an object that encapsulates a graphics context for a
 * particular canvas. 
 * 
 * @version 1.24 09/21/98
 * @author Arthur van Hoff
 */

public class TinyGraphics extends Graphics {
    Color	color;
    Color	xorColor;
    Font	font;
    int		originX;
    int		originY;
    Rectangle 	clip;
    boolean	touched;
    int		pDrawable;

    private native void createFromGraphics(TinyGraphics g);
    private native void createFromWindow(TinyWindow win);

    TinyGraphics(TinyGraphics g) {
	createFromGraphics(g);
    }
    TinyGraphics(TinyWindow win) {
	createFromWindow(win);
    }

    private native void imageCreate(ImageRepresentation ir);
    sun.awt.image.Image image;

    /**
     * Create a new Graphics Object based on this one.
     */
    public Graphics create() {
	TinyGraphics g = new TinyGraphics(this);
	g.color = color;
	g.font = font;
	g.originX = originX;
	g.originY = originY;
	g.clip = clip;
	return g;
    }

    /**
     * Translate
     */
    public void translate(int x, int y) {
	originX += x;
	originY += y;
    }

    /**
     * Disposes of this Graphics context. 
     */
    public native void dispose();

    /**
     * Get the current font.
     */
    public Font getFont() {
	return font;
    }

    /**
     * Set the current font.
     */
    public void setFont(Font font) {
	if ((font != null) && (this.font != font)) {
	    touched = true;
	    this.font = font;
	}
    }

    /**
     * Gets font metrics for the given font.
     */
    public FontMetrics getFontMetrics(Font font) {
	return TinyFontMetrics.getFontMetrics(font);
    }

    /**
     * Get the current color.
     */
    public Color getColor() {
	return color;
    }

    /**
     * Sets the current color.
     */
    public void setColor(Color c) {
	if ((c != null) && (c != color)) {
	    touched = true;
	    color = c;
	}
    }

    /**
     * Sets the paint mode to overwrite the destination with the
     * current color. This is the default paint mode.
     */
    public void setPaintMode() {
	if (xorColor != null) {
	    touched = true;
	    xorColor = null;
	}
    }

    /**
     * Sets the paint mode to alternate between the current color
     * and the given color.
     */
    public void setXORMode(Color c) {
	if ((c != null) && (c != xorColor)) {
	    touched = true;
	    xorColor = c;
	}
    }

    /** 
     * Gets the current clipping area 
     */
    public Rectangle getClipBounds() {
	if (clip == null) {
	    return null;
	}
	return new Rectangle(clip.x - originX,
			     clip.y - originY,
			     clip.width, clip.height);
    }

    private void changeClip(int x, int y, int w, int h, boolean set) {
	touched = true;
	x = (int)(originX + x);
	y = (int)(originY + y);
	w = (int)(w);
	h = (int)(h);
	if (set || clip == null) {
	    clip = new Rectangle(x, y, w, h);
	} else {
	    clip = clip.intersection(new Rectangle(x, y, w, h));
	}
    }

    /**
     * Crops the clipping rectangle for this Graphics context. 
     */
    public void clipRect(int x, int y, int w, int h) {
	changeClip(x, y, w, h, false);
    }

    /**
     * Sets the clipping rectangle for this Graphics context. 
     */
    public void setClip(int x, int y, int w, int h) {
	changeClip(x, y, w, h, true);
    }

    /** Returns a Shape object representing the clip. */
    public Shape getClip() {
	return getClipBounds();
    }

    /** Sets the clip to a Shape (only Rectangle allowed). */
    public void setClip(Shape clip) {
	if (clip == null) {
	    touched = true;
	    this.clip = null;
	} else if (clip instanceof Rectangle) {
	    Rectangle r = (Rectangle) clip;
	    changeClip(r.x, r.y, r.width, r.height, true);
	} else {
	    throw new IllegalArgumentException("setClip(Shape) only supports Rectangle objects");
	}
    }

    /** Draws the given line. */
    public native void drawLine(int x1, int y1, int x2, int y2);

    /** Fills the given rectangle with the foreground color. */
    public native void fillRect(int x, int y, int w, int h);
    
    /** Clears the rectangle indicated by x,y,w,h. */
    public void clearRect(int x, int y, int w, int h) {
	if (clip != null) {
	    int cx = clip.x;
	    int cy = clip.y;
	    int cw = clip.width;
	    int ch = clip.height;
	    if (x < cx){
		cw -= (cx - x);
		x = cx;
	    }
	    if (y < cy){
		ch -= (cy - y);
		y = cy;
	    }
	    if (x + w > cx + cw) {
		w = cx + cw - x;
	    }
	    if (y + h > cy + ch) {
		h = cy + ch - y;
	    }
	}
	if (w <= 0 || h <= 0) {
	    return;
	}
	// no offscreen drawing
	if (image == null) {
	    clearArea(x, y, w, h);
	} else {
	    Color c = color;
	    setColor(Color.white);
	    fillRect(x, y, w, h);
	    setColor(c);
	}
    }
    
    /** clearArea in the window */
    native void clearArea(int x, int y, int w, int h);

    /** Draws the given string. */
    public native void drawString(String str, int x, int y);

    /** Draws character. */
    public native void drawChars(char data[], int offset, int length, int x, int y);

    /** Draws bytes. */
    public native void drawBytes(byte data[], int offset, int length, int x, int y);

    /**
     * Draws an image at x,y in nonblocking mode with a callback object.
     */
    public boolean drawImage(Image img, int x, int y, ImageObserver observer) {
        if (img instanceof BufferedImage) {
            drawImage((BufferedImage)img, null, x, y, observer);
            return true;
        }
                                     
	TinyImage i = (TinyImage)img;
	if (i.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img,
				     ImageObserver.ERROR|ImageObserver.ABORT,
				     -1, -1, -1, -1);
	    }
	    return false;
	}
	ImageRepresentation ir = i.getImageRep(); 
        GraphicsObserver obs = new GraphicsObserver(this.create(), img,
                                                    x, y, null,
                                                    observer);
        return ir.drawToBufImage(obs);
    }

    /**
     * Draws an image scaled to x,y,w,h in nonblocking mode with a
     * callback object.
     */
    public boolean drawImage(Image img, int x, int y, int width, int height,
			     ImageObserver observer) {
	if (width == 0 || height == 0) {
	    return true;
	}
        // Check if this is actually a buffered image
        if (img instanceof BufferedImage) {
            BufferedImage bimg = (BufferedImage) img;
            // Append the new transform
            AffineTransform atfm = new AffineTransform();
            atfm.scale(width/bimg.getWidth(), height/bimg.getHeight());

            // REMIND:  Which interpolation to use?
            drawImage(bimg, new BilinearAffineTransformOp(atfm),
                      x, y, observer);
            
            return true;
        }
        
	TinyImage i = (TinyImage)img;
	if (i.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img,
				     ImageObserver.ERROR|ImageObserver.ABORT,
				     -1, -1, -1, -1);
	    }
	    return false;
	}
	ImageRepresentation ir = i.getImageRep();
        GraphicsObserver obs = new ScaledGraphicsObserver(this.create(), img,
                                                          x, y, width, height,
                                                          null, observer);
        
	return ir.drawToBufImage(obs);
    }

    /**
     * Draws an image at x,y in nonblocking mode with a solid background
     * color and a callback object.
     */
    public boolean drawImage(Image img, int x, int y, Color bg,
			     ImageObserver observer) {
        if (img instanceof BufferedImage) {
            drawImage((BufferedImage)img, null, x, y, observer);
            return true;
        }
                                     
	TinyImage i = (TinyImage)img;
	if (i.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img,
				     ImageObserver.ERROR|ImageObserver.ABORT,
				     -1, -1, -1, -1);
	    }
	    return false;
	}
	ImageRepresentation ir = i.getImageRep();
        GraphicsObserver obs = new GraphicsObserver(this.create(), img,
                                                    x, y, bg, observer);
        return ir.drawToBufImage(obs);
    }

    /**
     * Draws an image scaled to x,y,w,h in nonblocking mode with a
     * solid background color and a callback object.
     */
    public boolean drawImage(Image img, int x, int y, int width, int height,
			     Color bg, ImageObserver observer) {
	if (width == 0 || height == 0) {
	    return true;
	}
        // Check if this is actually a buffered image
        if (img instanceof BufferedImage) {
            BufferedImage bimg = (BufferedImage) img;
            // Append the new transform
            AffineTransform atfm = new AffineTransform();
            atfm.scale(width/bimg.getWidth(), height/bimg.getHeight());

            // REMIND: Which interpolation to use?
            drawImage(bimg, new BilinearAffineTransformOp(atfm),
                      x, y, observer);
            return true;
        }
        
	TinyImage i = (TinyImage)img;
	if (i.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img,
				     ImageObserver.ERROR|ImageObserver.ABORT,
				     -1, -1, -1, -1);
	    }
	    return false;
	}
	ImageRepresentation ir = i.getImageRep();
        GraphicsObserver obs = new ScaledGraphicsObserver(this.create(), img,
                                                    x, y, width, height,
                                                    bg, observer);
        
	return ir.drawToBufImage(obs);
    }

    /**
     * Draws a subrectangle of an image scaled to a destination rectangle
     * in nonblocking mode with a callback object.
     */
    public boolean drawImage(Image img,
			     int dx1, int dy1, int dx2, int dy2,
			     int sx1, int sy1, int sx2, int sy2,
			     ImageObserver observer)
    {
	if (dx1 == dx2 || dy1 == dy2) {
	    return true;
	}
        // Check if this is actually a buffered image
        if (img instanceof BufferedImage) {
            BufferedImage orig = (BufferedImage) img;
            int width  = dx2-dx1+1;
            int height = dy2-dy1+1;
            Tile subTile = orig.getTile().createSubtile(dx1, dy1,
                                                        width, height);
            ColorModel cm = orig.getColorModel();
            BufferedImage bimage = new BufferedImage(cm, subTile,
                                                    cm.isAlphaPremultiplied());
                                                     
            // Append the new transform
            AffineTransform atfm = new AffineTransform();
            atfm.scale(width/(sx2-sx1+1), height/(sy2-sy1+1));
        
            drawImage((BufferedImage)img, new BilinearAffineTransformOp(atfm),
                      sx1, sy1, observer);
            return true;
        }
        
	TinyImage i = (TinyImage) img;
	if (i.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img,
				     ImageObserver.ERROR|ImageObserver.ABORT,
				     -1, -1, -1, -1);
	    }
	    return false;
	}
	ImageRepresentation ir = i.getImageRep();
        GraphicsObserver obs = new StretchedGraphicsObserver(this.create(),
                                   img,
				   dx1, dy1, dx2, dy2,
				   sx1, sy1, sx2, sy2,
				   null, observer);
                                                          
	return ir.drawToBufImage(obs);
    }

    /**
     * Draws a subrectangle of an image scaled to a destination rectangle in
     * nonblocking mode with a solid background color and a callback object.
     */
    public boolean drawImage(Image img,
			     int dx1, int dy1, int dx2, int dy2,
			     int sx1, int sy1, int sx2, int sy2,
			     Color bgcolor,
			     ImageObserver observer)
    {
	if (dx1 == dx2 || dy1 == dy2) {
	    return true;
	}
        // Check if this is actually a buffered image
        if (img instanceof BufferedImage) {
            BufferedImage orig = (BufferedImage) img;
            int width  = dx2-dx1+1;
            int height = dy2-dy1+1;
            Tile subTile = orig.getTile().createSubtile(dx1, dy1,
                                                        width, height);
            ColorModel cm = orig.getColorModel();
            BufferedImage bimage = new BufferedImage(cm, subTile,
                                                    cm.isAlphaPremultiplied());
                                                     
            // Append the new transform
            AffineTransform atfm = new AffineTransform();
            atfm.scale(width/(sx2-sx1+1), height/(sy2-sy1+1));
        
            drawImage((BufferedImage)img, new BilinearAffineTransformOp(atfm),
                      sx1, sy1, observer);
            return true;
        }
        
	TinyImage i = (TinyImage) img;
	if (i.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img,
				     ImageObserver.ERROR|ImageObserver.ABORT,
				     -1, -1, -1, -1);
	    }
	    return false;
	}
	ImageRepresentation ir = i.getImageRep();
        GraphicsObserver obs = new StretchedGraphicsObserver(this.create(),
                                   img,
				   dx1, dy1, dx2, dy2,
				   sx1, sy1, sx2, sy2,
                                   bgcolor, observer);
                                                          
	return ir.drawToBufImage(obs);
    }

    /**
     * Copies an area of the canvas that this graphics context paints to.
     * @param X the x-coordinate of the source.
     * @param Y the y-coordinate of the source.
     * @param W the width.
     * @param H the height.
     * @param dx the x-coordinate of the destination.
     * @param dy the y-coordinate of the destination.
     */
    public native void copyArea(int x, int y, int w, int h, int dx, int dy);


    /** Draws a rounded rectangle. */
    public void drawRoundRect(int x, int y, int w, int h, int arcWidth, int arcHeight) {

	if (w <= 0 || h <= 0) {
	    return;
	}

	int aw = arcWidth / 2;
	int ah = arcHeight / 2;
	int x2 = x + w;
	int y2 = y + h;

	drawArc(x, y, arcWidth, arcHeight, 90, 90);
	drawLine(x + aw + 1, y, x2 - aw - 1, y);
	drawArc(x2 - arcWidth, y, arcWidth, arcHeight, 0, 90);
	drawLine(x2, y + ah + 1, x2, y2 - ah - 1);
	drawArc(x2 - arcWidth, y2 - arcHeight, arcWidth, arcHeight, 270, 90);
	drawLine(x2 - aw - 1, y2, x + aw + 1, y2);
	drawArc(x, y2 - arcHeight, arcWidth, arcHeight, 180, 90);
	drawLine(x, y2 - ah - 1, x, y + ah + 1);
    }

    /** Draws a filled rounded rectangle. */
    public void fillRoundRect(int x, int y, int w, int h, int arcWidth, int arcHeight) {
	
	if (w <= 0 || h <= 0) {
	    return;
	}

	int aw = arcWidth / 2;
	int ah = arcHeight / 2;
	int x2 = x + w;
	int y2 = y + h;

	fillRect(x, y + ah, aw, h - arcHeight);
	fillRect(x + aw, y, w - arcWidth, h);
	fillRect(x2 - aw, y + ah, aw, h - arcHeight);
	fillArc(x, y, arcWidth, arcHeight, 90, 90);
	fillArc(x2 - arcWidth, y, arcWidth, arcHeight, 0, 90);
	fillArc(x2 - arcWidth, y2 - arcHeight, arcWidth, arcHeight, 270, 90);
	fillArc(x, y2 - arcHeight, arcWidth, arcHeight, 180, 90);
    }

    /** Draws a polygon defined by an array of x points and y points */
    public void drawPolygon(int xPoints[], int yPoints[], int nPoints) {
	int i;
	for (i = 1 ; i < nPoints ; i++) {
	    drawLine(xPoints[i-1], yPoints[i-1], xPoints[i], yPoints[i]);
	}
	if (xPoints[i-1] != xPoints[0] || yPoints[i-1] != yPoints[0]) {
	    drawLine(xPoints[i-1], yPoints[i-1], xPoints[0], yPoints[0]);
	}
    }

    /** Draws an open figure defined by an array of x points and y points */
    public void drawPolyline(int xPoints[], int yPoints[], int nPoints) {
	for (int i = 1 ; i < nPoints ; i++) {
	    drawLine(xPoints[i-1], yPoints[i-1], xPoints[i], yPoints[i]);
	}
    }

    /** Fills a polygon with the current fill mask */
    public native void fillPolygon(int xPoints[], int yPoints[], int nPoints);

    /** Draws an oval to fit in the given rectangle */
    public void drawOval(int x, int y, int w, int h) {
	drawArc(x, y, w, h, 0, 360);
    }

    /** Fills an oval to fit in the given rectangle */
    public void fillOval(int x, int y, int w, int h) {
	fillArc(x, y, w, h, 0, 360);
    }

    /**
     * Draws an arc bounded by the given rectangle from startAngle to
     * endAngle. 0 degrees is a vertical line straight up from the
     * center of the rectangle. Positive angles indicate clockwise
     * rotations, negative angle are counter-clockwise.
     */
    public native void drawArc(int x, int y, int w, int h, int startAngle, int endAngle);

    /** fills an arc. arguments are the same as drawArc. */
    public native void fillArc(int x, int y, int w, int h, int startAngle, int endAngle);

    public String toString() {	
	return getClass().getName() + "[" + originX + "," + originY + "]";
    }
}
