/*
 * @(#)TinyVerticalScrollbar.java	1.16 98/09/21
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.awt.tiny;

import java.awt.*;

/**
 * A simple vertical scroll bar.
 */
class TinyVerticalScrollbar {


    /**
     * The thread that asynchronously tells the scrollbar to scroll.
     * @see #startScrolling
     */
    private static TinyScrollRepeater scrollThread;

    /**
     * The lock that controls scroll repeating
     * @see #startScrolling
     * @see #handleWindowEvent
     * @see TinyScrollRepeater
     */
    static final Object SCROLL_LOCK = new Object();

    /**
     * A boolean that notes whether the scrollbar should be
     * scrolling or not. Observed by ScrollThread.
     * @see #startScrolling
     * @see #handleWindowEvent
     * @see TinyScrollRepeater
     */
    static boolean scrolling = false;
  

    TinyScrollbarClient sb;
    int val;
    int min;
    int max;
    int vis;
    int line;
    int page;
    int mode;
    int mv;
    int tpts_x[];
    int tpts_y[];

    void notifyValue(int v) {
	if (v < min) {
	    v = min;
	} else if (v > max) {
	    v = max;
	}
	if ((v != val) && (sb != null)) {
	    sb.notifyValue(this, mode, v);
	}
    }

    // These are done this way because the horizontal scrollbar
    // overrides them
    void drawLine(Graphics g, int x1, int y1, int x2, int y2) {
	g.drawLine(x1, y1, x2, y2);
    }

    void fillPolygon(Graphics g, int x[], int y[], int npts) {
	g.fillPolygon(x, y, npts);
    }

    void fillRect(Graphics g, int x, int y, int w, int h) {
	g.fillRect(x, y, w, h);
    }

    void paint(Graphics g, Color colors[], int width, int height) {

	double f = (double)(height - 2*(width-1)) / Math.max(1, ((max - min) + vis));
	int v1 = width + (int)(f * (val - min));
	int v2 = (int)(f * vis);
	int w2 = width-4;

	if (max - min <= 0) {
            // max the thumb size to show that the scrollbar is empty
            v1 = w2 + 2;
            v2 = height - w2*2 - 4;
        }

	if (height < 3*w2 ) {
	    v1 = v2 = 0;
	    if (height < 2*w2 + 2) {
		w2 = (height-2)/2;
	    }
	} else if (v2 < 7) {
	// enforce a minimum handle size
	    v1 = Math.max(0, v1 - ((7 - v2)>>1));
	    v2 = 7;
	}

	int ctr   = width/2;
	int sbmin = ctr - w2/2;
	int sbmax = ctr + w2/2;

	// These polygons describe the end arrows
	if ((tpts_x == null) || (tpts_y == null)) {
	    tpts_x = new int[3];
	    tpts_y = new int[3];
	}

	// paint the background slightly darker
	{
	    Color back = colors[TinyComponentPeer.BACKGROUND_COLOR];
	    Color c = new Color((int) (back.getRed()   * 0.85),
				(int) (back.getGreen() * 0.85),
				(int) (back.getBlue()  * 0.85));

	    g.setColor(c);
	    fillRect(g, 0, 0, width, height);
	}

	// paint the thumb and arrows in the normal background color
	g.setColor(colors[TinyComponentPeer.BACKGROUND_COLOR]);
	if (v1 > 0) {
	    fillRect(g, 2, v1, width-4, v2);
	}

	tpts_x[0] = ctr;	tpts_y[0] = 2;
	tpts_x[1] = sbmin;	tpts_y[1] = w2;
	tpts_x[2] = sbmax;	tpts_y[2] = w2;
	fillPolygon(g, tpts_x, tpts_y, 3);

	tpts_y[0] = height-2;
	tpts_y[1] = height-w2;
	tpts_y[2] = height-w2;
	fillPolygon(g, tpts_x, tpts_y, 3);


	// // // // draw the "highlighted" edges
	g.setColor(colors[TinyComponentPeer.HIGHLIGHT_COLOR]);

	// outline
	drawLine(g, width - 1, 1, width - 1, height - 1);
	drawLine(g, 1, height - 1, width - 1, height - 1);

	// arrows
	drawLine(g, ctr, 1, sbmin, w2);
	drawLine(g, sbmin, height - w2, sbmax, height - w2);
	drawLine(g, sbmin, height - w2, ctr, height - 2);

	// thumb
	if (v1 > 0) {
	    drawLine(g, 2, v1, 2, v1 + v2 - 1);
	    drawLine(g, 2, v1, width-3, v1);
	}

	// // // // draw the "shadowed" edges
	g.setColor(colors[TinyComponentPeer.SHADOW_COLOR]);

	// outline
	drawLine(g, 0, 0, width, 0);
	drawLine(g, 0, 0, 0, height - 1);

	// arrows
	drawLine(g, sbmin, w2, sbmax, w2);
	drawLine(g, sbmax, w2, ctr, 1);
	drawLine(g, ctr, height-2, sbmax, height-w2);

	// thumb
	if (v1 > 0) {
	    drawLine(g, 2, v1 + v2 - 1, width-3, v1 + v2 - 1);
	    drawLine(g, width-3, v1, width-3, v1 + v2 - 1);
	}
    }


   /**
    *  Tell the scrollThread to start scrolling.
    */
    void startScrolling() {
	// Make that we scroll at least once
	scroll();

	// wake up the scroll repeater
	if (scrollThread == null) {
	    // If there isn't a scrollThread, then create
	    // one and start it.
	    scrollThread = new TinyScrollRepeater(this);
	    scrollThread.start();
	} else {
	    scrollThread.setScrollbar(this);
	}

	synchronized (SCROLL_LOCK) {
	    scrolling = true;
	    SCROLL_LOCK.notifyAll();
	}

    }

   /**
    *  Scroll one unit.
    *  @see notifyValue
    */
    void scroll() {
      switch (mode) {
	  case Event.SCROLL_LINE_UP:
	    notifyValue(val - line);
	    return;
	    
	  case Event.SCROLL_LINE_DOWN:
	    notifyValue(val + line);
	    return;
	    
	  case Event.SCROLL_PAGE_UP:
	    notifyValue(val - page);
	    return;

	  case Event.SCROLL_PAGE_DOWN:
	    notifyValue(val + page);
	    return;
      }
      return;
    }

    boolean handleWindowEvent(Event evt, int width, int height) {
	switch (evt.id) {
	case Event.MOUSE_DOWN:
	    if (evt.y < width-1) {
		mode = Event.SCROLL_LINE_UP;
		startScrolling();
		return true;
	    }
	    if (evt.y > height - (width-2)) {
		mode = Event.SCROLL_LINE_DOWN;
		startScrolling();
		return true;
	    }

	    double f = (double)(height - 2*(width-1)) / ((max - min) + vis);
	    int v1 = width + (int)(f * (val - min));
	    int v2 = (int)(f * vis);

	    if (evt.y < v1) {
		mode = Event.SCROLL_PAGE_UP;
		startScrolling();
		return true;
	    }
	    if (evt.y > v1 + v2) {
		mode = Event.SCROLL_PAGE_DOWN;
		startScrolling();
		return true;
	    }
	    mode = Event.SCROLL_ABSOLUTE;
	    mv = evt.y - v1;
	    return true;

	case Event.MOUSE_UP:
	  synchronized (SCROLL_LOCK) {
	      scrolling = false;
	  }
	  return true;

	case Event.MOUSE_DRAG:
	    if (mode == Event.SCROLL_ABSOLUTE) {
		notifyValue((((evt.y - (mv + width)) * ((max - min) + vis) / (height - 2*width))) + min);
	    }
	    return true;

	}

	return false;
    }
}



/**
 * This thread manages continuous scrolling while the mouse is
 * pressed down on a scrollbar's page-up, page-down, line-up, or
 * line-down areas.
 */

class TinyScrollRepeater extends Thread {

  /**
   * Time to pause before the first scroll repeat.
   */
  static int beginPause = 500; // Reminder - make this a user definable property

  /**
   * Time to pause between each scroll repeat.
   */
  static int repeatPause = 100; // Reminder - make this a user definable property

  /**
   * The scrollbar that we sending scrolling.
   */
  TinyVerticalScrollbar sb; 

  /**
   * newScroll gets reset when a new scrollbar gets set.
   */
  boolean newScroll;

  /**
   * Creates a new scroll repeater.
   * @param sb the scrollbar that this thread will scroll
   */
  TinyScrollRepeater(TinyVerticalScrollbar sb) {
      super("TinyScrollRepeater thread");
      this.setScrollbar(sb);
      newScroll = true;
  }
  
  /**
   * Sets the scrollbar.
   * @param sb the scrollbar that this thread will scroll
   */
  public synchronized void setScrollbar(TinyVerticalScrollbar sb) {
      this.sb = sb;
      newScroll = true;
  }

  /**
   * Called by Thread.start(). Starts the scroll repeater thread.
   */
  public void run () {
      boolean shouldScroll;	// local variable for thread safety
      
      while (true) {
	  if (newScroll) {
	      try {
		  // Pause before repeating. This gives the user time to release
		  // the mouse button before continuous scrolling starts.
		  sleep(beginPause); 
	      } catch (java.lang.InterruptedException e) {
	      }
	      newScroll = false;
	  }
	  
	  synchronized (TinyVerticalScrollbar.SCROLL_LOCK) {
	      shouldScroll = sb.scrolling;
	  }
	  
	  if (shouldScroll) {
	      // if sb.scrolling is true, then tell the scrollbar to scroll
	      sb.scroll();
	      try {
		  sleep(repeatPause); 
	      } catch (InterruptedException e) {
	      }
	  } else {
	      // else if sb.scrolling is false, wait until notified
	      try {
		  synchronized (TinyVerticalScrollbar.SCROLL_LOCK) {
		      TinyVerticalScrollbar.SCROLL_LOCK.wait();
		  }
	      } catch (InterruptedException e) {
	      }
	  }
      }
  }
}
