/*
 * @(#)TinyWinManagerFrame.java	1.22 98/06/29
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// This class interfaces to a platform's native window manager.

package sun.awt.tiny;

import java.awt.*;
import java.util.Vector;
import java.awt.image.ImageObserver;
import sun.awt.image.ImageRepresentation;

class TinyWinManagerFrameWarning extends TinyWindow {

    String warning;
    final static Color warnBack = Color.black;
    final static Color warnFore = Color.red;

    final static Font warnFont = new Font("Dialog", Font.BOLD, 10);

    TinyWinManagerFrameWarning(TinyWindow parent, String warning) {
	this.warning = warning;
	winCreate(parent, NATIVEWINDOWMANAGER);
	winBackground(warnBack);
    }

    void paint(Graphics g) {
	FontMetrics fm = g.getFontMetrics();
	g.drawString(warning, 5,
		     (height + fm.getAscent() - fm.getDescent()) / 2);
    }
}

class TinyWinManagerFrame extends TinyFrame {

    TinyWindow menubar;
    TinyWindow focus;
    TinyWindow warning;

    boolean inreshape;
    boolean resizable;
    boolean selected;
    String title;
    int mappedOnce;

    // this function is privileged! do not change it to public!
    private static int getInset(final String name, final int def) {
	Integer tmp = (Integer) java.security.AccessController.doPrivileged(
	    new sun.security.action.GetIntegerAction(name, def));
	return tmp.intValue();
    }

    final static Vector allWindows = new Vector();
    final static int MENU = 27;
    final static int WARN = 20;
    final static int TOP = getInset("awt.frame.topInset", 25);
    final static int LEFT = getInset("awt.frame.leftInset", 5);
    final static int RIGHT = getInset("awt.frame.rightInset", 5);
    final static int BOTTOM = getInset("awt.frame.bottomInset",5);
    final static Color frameLight = Color.lightGray;

    /**
     * Create a window frame.
     */
    TinyWinManagerFrame(TinyWindowPeer content, String title, String warning, boolean resizable) {
	allWindows.addElement(this);
	this.content = content;
	this.resizable = resizable;
	this.title = title;
	winCreate(null, NATIVEWINDOWMANAGER);
	winBackground(frameLight);
	winSetTitle(title);

	content.winCreate(this, NATIVEWINDOWMANAGER);
	content.winShow();

	if (warning != null) {
	   this.warning = new TinyWinManagerFrameWarning(this, warning);
	}
    }

    /**
     * Dispose
     */
    void dispose() {
	allWindows.removeElement(this);
	if (menubar != null) {
	    menubar.winDispose();
	}
	if (warning != null) {
	    warning.winDispose();
	}
	content.winDispose();
	super.winDispose();
    }

    /**
     * Compute the insets.
     */
    public Insets insets() {
	Insets in = new Insets(0, 0, 0, 0);
	in.top = TOP;
	in.bottom = BOTTOM;
	in.left = LEFT;
	in.right = RIGHT;

	if (menubar != null) {
	    in.top += MENU;
	}

	if (warning != null) {
	    in.top += WARN;
	}
	return in;
    }

    /**
     * Reshape the window frame and its contents.
     */
    public void winReshape(int x, int y, int width, int height) {
	int w = width - (LEFT + RIGHT);
	int h = height - (TOP + BOTTOM);

	if (inreshape) {
	    return;
	}
	super.winReshape(x, y, w, h);
	content.winReshape(-LEFT, -TOP, width, height);
	this.x = x + LEFT;
	this.y = y + TOP;
	int top = TOP;

	if (menubar != null) {
	    menubar.winReshape(LEFT, top, w, MENU);
	    top += MENU;
	}

	if (warning != null) {
	    warning.winReshape(0, top, w, WARN);
	}
    }

    /**
     * callback from NotifyConfigure event when
     * windowManager has changed the size and/or the position
     * of the top-window
     */
    void handleReshape(int x, int y, int width, int height) {
	inreshape = true;
	int w = width + (LEFT + RIGHT);
	int h = height + (TOP + BOTTOM);
	content.winReshape(-LEFT, -TOP, w, h);
	content.target.setBounds(x - LEFT, y - TOP, w, h);
    	content.target.validate();
	content.target.repaint();
	this.width = width;
	this.height = height;
	this.x = x;
	this.y = y;
	int top = TOP;

	if (menubar != null) {
	    menubar.winReshape(LEFT, top, width, MENU);
	    top += MENU;
	}

	if (warning != null) {
	    warning.winReshape(0, top, width, WARN);
	}
	inreshape = false;
    }

    /**
     * send event that this window can be deleted
     */
    void handleQuitWindow() {
	content.target.postEvent(new Event(content.target, Event.WINDOW_DESTROY, null));
    }

    /**
     * send event when window is iconified
     */
    public void handleIconify() {
	content.target.postEvent(new Event(content.target, Event.WINDOW_ICONIFY, null));
    	mappedOnce = 0;
    }

    /**
     * send event when window id deiconified
     */
    public void handleDeIconify() {
	if (mappedOnce == 0) {
	    mappedOnce++;
	} else {
	    content.target.postEvent(new Event(content.target, Event.WINDOW_DEICONIFY, null));
    	}
    }

    /**
     * Set the cursor image.
     */
    public void setCursor(int cursorType) {
	winSetCursor(cursorType);
    }

    /**
     * Set the title.
     */
    void setTitle(String title) {
    	this.title = title;
	winSetTitle(title);
    }

    /**
     * Set resizable.
     */
    void setResizable(boolean resizable) {
	this.resizable = resizable;
    }

    /**
     * Set the menubar
     */
    void setMenuBar(TinyMenuBarPeer menubar) {
	if (this.menubar != null) {
	    this.menubar.winDispose();
	}
	this.menubar = menubar;
	if (menubar != null) {
	    menubar.init(content);
	}
	winReshape(x - LEFT, y - TOP, width + (LEFT + RIGHT), height + (TOP + BOTTOM));
    }


    /**
     * Set the icon image.
     */
    public void setIconImage(Image img) {
	ImageRepresentation ir = ((TinyImage)img).getImageRep();
	ir.reconstruct(ImageObserver.ALLBITS);
	winSetIconImage(ir);
    }

    /**
     * Assign the focus to a new window.
     */
    void requestFocus(TinyWindow newFocus) {
	if (newFocus != focus) {
	    if (focus != null && selected == true) {
		focus.handleFocusEvent(new Event(focus, Event.LOST_FOCUS, null));
	    }
	    focus = newFocus;
	    if (focus != null) {
		if (selected == true) {
		    focus.handleFocusEvent(new Event(focus, Event.GOT_FOCUS, null));
		} else {
		    winFocus();
		}
	    }
	}
    }

    /**
     * Check if this window has the focus.
     */
    boolean hasFocus(TinyWindow focus) {
	return focus == this.focus;
    }

    /**
     * The window now has the keyboard focus.
     */
    public void handleFocusIn() {
	selected = true;
	if (title != null) {
	    if (focus != null) {
		focus.handleFocusEvent(new Event(focus, Event.GOT_FOCUS, null));
	    }
	}
    }

    /**
     * The window no longer has the keyboard focus.
     */
    public void handleFocusOut() {
	selected = false;
	if (title != null) {
	    if (focus != null) {
		focus.handleFocusEvent(new Event(focus, Event.LOST_FOCUS, null));
	    }
	}
    }

    /**
     * Event handler.
     */
    boolean handleWindowEvent(Event evt) {
	switch (evt.id) {
	  case Event.KEY_PRESS:
	  case Event.KEY_RELEASE:
	  case Event.KEY_ACTION:
	  case Event.KEY_ACTION_RELEASE:
	    if (focus != null) {
		return focus.handleKeyEvent(evt);
	    }
	    return true;
	}
    	return false;
    }
}




