/*
 * @(#)TinyWindow.java	1.22 98/06/29
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.awt.tiny;

import java.awt.*;
import sun.awt.image.ImageRepresentation;

/**
 * This class encapsulates access to the underlying window system's
 * concept of a window.  You must call create to activate the window,
 * it is initially unmapped.
 *
 * @author Arthur van Hoff
 * modified Siebe Brouwer
 */
public class TinyWindow {

    public final static int NATIVEWINDOWMANAGER = 0;
    public final static int TINYWINDOWMANAGER = 1;
    private long xid;
    TinyWindowPeer content;
    TinyWindow parent;
    int x;
    int y;
    int width;
    int height;

    /* In some environments (such as X), if you do a window move or resize
     * when there are events in your event queue, the coordinates in
     * the existing events won't be updated to compensate for the move
     * or resize.  Therefore, we need to keep track of what the window
     * system thinks our position in to we can manually compensate.
     * When the window configuration makes its way through the event queue,
     * these should match x, y, width, and height
     */
    int actualX;
    int actualY;
    int actualWidth;
    int actualHeight;

    // double-click processing
    static final int MULTICLICK_TIME = 500; // 500 milliseconds
    static int clickCount;
    static long lastTime;
    static TinyWindow lastWindow;

    native void winCreate(TinyWindow parent, int or);
    native void winReshape(int x, int y, int width, int height);

    native void winDispose();
    native void winShow();
    native void winHide();
    native void winFront();
    native void winBack();
    native void winBackground(Color c);
    native void winFocus();
    native void winSetCursor(int cursorType);

    /**
     * setting the properties used by the windowManager
     */
    native void winSetIconImage(ImageRepresentation ir);
    native void winSetTitle(String l);

    /**
     * return the property useWindowManager in the properties file
     */
    boolean useTinyWindowManager() {
	Boolean tmp = (Boolean)
	    java.security.AccessController.doPrivileged(
	new sun.security.action.GetBooleanAction("useTinyWindowManager"));
	return tmp.booleanValue():
    }

    TinyWindow getParentSubwindow() {
	return this;
    }

    /*
     * create a XWindow
     */
    void winCreate(TinyWindow parent) {
	winCreate(parent != null? parent.getParentSubwindow() : parent,
			 TINYWINDOWMANAGER);
    }

    /**
     * Convert to global coordinates.
     */
    Point toGlobal(Point pt) {
	for (TinyWindow win = this ; win != null ; win = win.parent) {
	    pt.translate(win.x, win.y);
	}
	return pt;
    }
    Point toGlobal(int x, int y) {
	return toGlobal(new Point(x, y));
    }

    /**
     * Convert to local coordinates.
     */
    Point toLocal(Point pt) {
	for (TinyWindow win = this ; win != null ; win = win.parent) {
	    pt.translate(-win.x, -win.y);
	}
	return pt;
    }
    Point toLocal(int x, int y) {
	return toLocal(new Point(x, y));
    }

    /**
     * Paint the window.
     */
    void paint(Graphics g) {
	g.setColor(Color.black);
	g.drawRect(0, 0, width-1, height-1);
	g.setColor(Color.lightGray);
	g.fillRect(1, 1, width-2, height-2);
    }

    /**
     * Expose call back.
     */
    public void handleExpose(int x, int y, int w, int h) {
	Graphics g = new TinyGraphics(this);
	try {
	    g.clipRect(x, y, w, h);
//	    g.setColor(getWinBackground());
//	    g.fillRect(x, y, w, h);
	    paint(g);
	} finally {
	    g.dispose();
	}
    }

    /**
     * Post Events
     */
    void postEvent(long when, int id, int x, int y, int key, int flags, Object arg) {
	handleWindowEvent(new Event(this, when, id, x, y, key, flags, arg));
    }

    void postEvent(Event e) {
	handleWindowEvent(e);
    }

    /**
     * Event handler
     */
    boolean handleWindowEvent(Event evt) {
	return false;
    }

    boolean handleKeyEvent(Event evt) {
	return handleWindowEvent(evt);
    }

    void handleFocusEvent(Event evt) {
	handleWindowEvent(evt);
    }

    /*
     * Low level call backs
     */
    public void handleKeyPress(long when, int x, int y, int key, int flags) {
	if (parent != null) {
	    x += this.x;
	    y += this.y;
	    parent.handleKeyPress(when, x, y, key, flags);
	} else {
	    postEvent(when, Event.KEY_PRESS, x, y, key, flags, null);
	}
    }

    public void handleKeyRelease(long when, int x, int y, int key, int flags) {
	if (parent != null) {
	    x += this.x;
	    y += this.y;
	    parent.handleKeyRelease(when, x, y, key, flags);
	} else {
	    postEvent(when, Event.KEY_RELEASE, x, y, key, flags, null);
	}
    }

    public void handleActionKey(long when, int x, int y, int key, int flags) {
	if (parent != null) {
	    x += this.x;
	    y += this.y;
	    parent.handleActionKey(when, x, y, key, flags);
	} else {
	    postEvent(when, Event.KEY_ACTION, x, y, key, flags, null);
	}
    }

    public void handleActionKeyRelease(long when, int x, int y, int key,
				       int flags) {
	if (parent != null) {
	    x += this.x;
	    y += this.y;
	    parent.handleActionKeyRelease(when, x, y, key, flags);
	} else {
	    postEvent(when, Event.KEY_ACTION_RELEASE, x, y, key, flags, null);
	}
    }

    public void handleMouseEnter(long when, int x, int y) {
	postEvent(when, Event.MOUSE_ENTER, x, y, 0, 0, null);
    }

    public void handleMouseExit(long when, int x, int y) {
	postEvent(when, Event.MOUSE_EXIT, x, y, 0, 0, null);
    }

    public void handleMouseDown(long when, int x, int y, int flags) {
	Event e = new Event(this, when, Event.MOUSE_DOWN, x, y, 0, flags, null);

	if (lastWindow != this) {
	    clickCount = 1;
	} else if ((when - lastTime) < MULTICLICK_TIME) {
	    ++clickCount;
	} else {
	    clickCount = 1;
	}

	lastTime = when;
	lastWindow = this;

	e.clickCount = clickCount;
	postEvent(e);
    }

    public void handleMouseUp(long when, int x, int y, int flags) {
	postEvent(when, Event.MOUSE_UP, x, y, 0, flags, null);
    }

    public void handleMouseMove(long when, int x, int y, int flags) {
	postEvent(when, Event.MOUSE_MOVE, x, y, 0, flags, null);
    }

    public void handleMouseDrag(long when, int x, int y, int flags) {
	postEvent(when, Event.MOUSE_DRAG, x, y, 0, flags, null);
    }

    public void handleFocusIn() {
    }

    public void handleFocusOut() {
    }

    void handleReshape(int x, int y, int w, int h) {}

    void handleQuitWindow() {}

    void handleDeIconify() {}

    void handleIconify() {}

    /**
     * Request the character focus.
     */
    public void requestFocus() {
	requestFocus(this);
    }

    void requestFocus(TinyWindow focus) {
//System.out.println("Requesting focus:"+this+" FOR:"+focus);
	if (parent != null) {
	    parent.requestFocus(focus);
	}
    }

    boolean hasFocus() {
	return hasFocus(this);
    }

    boolean hasFocus(TinyWindow win) {
	return (parent != null) && parent.hasFocus(win);
    }

    protected void finalize() {
	winDispose();
    }

    /**
     * Convert to String.
     */
    public String toStr() {
    	return getClass().getName() + "[" + x + "," + y + "," + width + "x" + height + "]";
    }
}



