/*
 * @(#)TinyWindowFrame.java	1.23 98/09/21
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.awt.tiny;

import java.awt.*;
import java.util.Vector;

class TinyWindowFrameWarning extends TinyWindow {
    String warning;

    final static Color warnBack = Color.black;
    final static Color warnFore = Color.red;
    final static Font warnFont = new Font("Dialog", Font.BOLD, 10);

    TinyWindowFrameWarning(TinyWindow parent, String warning) {
	this.warning = warning;
	winCreate(parent);
	winShow();
	winBackground(warnBack);
    }

    void paint(Graphics g) {
	g.setFont(warnFont);
	g.setColor(warnFore);

	FontMetrics fm = g.getFontMetrics();
	g.drawString(warning, 5, (height + fm.getAscent()) / 2);
    }
}

class TinyWindowFrame extends TinyFrame {
    
    final static Vector allWindows = new Vector();

    final static int INDRAG = 0;
    final static int INRESIZE = 1;
    final static int ICONIZE = 2;
    final static int MAXIMIZE = 3;
    final static int SHOWMENU = 4;
    final static int ICONIZED = 0;
    final static int MAXIMIZED = 1;

    final static int TITLE = 19;
    final static int MENU = 27;
    final static int WARN = 20;
    final static int BORDER = 6;
    final static int BORDER1 = BORDER - 1;
    final static int BORDER2 = BORDER*2 - 1;
    final static int BORDER3 = TITLE - 1;

    final static Color frameBlack = Color.black;
    final static Color frameDark = Color.gray;
    final static Color frameMedium = new Color(160, 160, 160);
    final static Color frameLight = Color.lightGray;
    final static Color frameWhite = Color.white;
    final static Color frameSelect = frameMedium;

    final static Font titleFont = new Font("Dialog", Font.PLAIN, 14);

    TinyWindow 	border;
    TinyWindow 	menubar;
    TinyWindow 	focus;
    TinyWindow 	warning;
    TinyMenuWindow menu;

    boolean 	resizable;
    boolean 	selected;
    String 	title;

    long 	mtm;
    int 	mx, my;
    int 	mw, mh;
    int 	max, may;
    int 	maw, mah;
    int 	mode = -1;
    int 	state = -1;
    
    /**
     * Create a window frame.
     */
    TinyWindowFrame(TinyWindowPeer content, String title, String warning, boolean resizable) {
	allWindows.addElement(this);
	this.content = content;
	this.title = title;
	this.resizable = resizable;

	winCreate(null, TINYWINDOWMANAGER);
	winBackground(frameLight);

	border = new TinyWindow();
	border.winCreate(this, TINYWINDOWMANAGER);
	border.winShow();

	content.winCreate(border, TINYWINDOWMANAGER);
	content.winShow();

	if (warning != null) {
	    this.warning = new TinyWindowFrameWarning(this, warning);
	}
    }

    /**
     * send event that this window can be deleted
     */
    void handleDeleteWindow() {
	content.target.postEvent(new Event(content.target, Event.WINDOW_DESTROY, null));
    }
    
    /**
     * Dispose
     */
    void dispose() {
	allWindows.removeElement(this);
	if (menubar != null) {
	    menubar.winDispose();
	}
	menubar = null;
	if (warning != null) {
	    warning.winDispose();
	}
	warning = null;
	content.winDispose();
	content = null;
	super.winDispose();
    }

    /**
     * Compute the insets.
     */
    public Insets insets() {
	Insets in = new Insets(0, 0, 0, 0);

	if (title != null) {
	    in.top    += TITLE + BORDER;
	    in.bottom += BORDER;
	    in.left   += BORDER;
	    in.right  += BORDER;
	}

	if (menubar != null) {
	    in.top += MENU;
	}
	if (warning != null) {
	    in.top += WARN;
	}
	return in;
    }

    /**
     * Reshape the window frame and its contents.
     */
    public void winReshape(int x, int y, int width, int height) {
	if ((warning != null) && (title == null)) {
	    height += WARN;
	}

	Insets in = insets();
	super.winReshape(x, y, width, height);
	border.winReshape(in.left, in.top, 
			  width - (in.left + in.right), 
			  height - (in.top + in.bottom));
	if ((title == null) && (warning != null)) {
	    content.winReshape(0, 0, width, height);
	} else {
	    content.winReshape(-in.left, -in.top, width, height);
	}

	int top = (title != null) ? (BORDER + TITLE) : 0;

	if (menubar != null) {
	    menubar.winReshape(in.left, top, width - (in.left + in.right), MENU);
	    top += MENU;
	}

	if (warning != null) {
	    warning.winReshape(in.left, top, width - (in.left + in.right), WARN);
	}
    }

    /**
     * Set the cursor image.
     */
    public void setCursor(int cursorType) {
	winSetCursor(cursorType);
    }


    /**
     * Set the title.
     */
    void setTitle(String title) {
	this.title = title;

	Graphics g = new TinyGraphics(this);
	try {
	    paintTitle(g);
	} finally {
	    g.dispose();
	}
    }

    /**
     * Set resizable.
     */
    void setResizable(boolean resizable) {
	this.resizable = resizable;
	paint();
    }

    /**
     * Set the menubar.
     */
    void setMenuBar(TinyMenuBarPeer menubar) {
	if (this.menubar != null) {
	    this.menubar.winDispose();
	}
	this.menubar = menubar;
	if (menubar != null) {
	    menubar.init(this);
	}
	winReshape(x, y, width, height);
    }

    /**
     * Set the icon image.
     */
    public void setIconImage(Image im) {
	// REMIND
    }

    /**
     * Draw a 3D rect in the right colors.
     */
    void frameRect(Graphics g, int x, int y, int w, int h, boolean raised) {
	g.setColor(raised ? frameWhite : frameDark);
	g.drawLine(x, y, x + w, y);
	g.drawLine(x, y, x, y + h);
	g.setColor(raised ? frameDark : frameWhite);
	g.drawLine(x + w, y, x + w, y + h);
	g.drawLine(x, y + h, x + w, y + h);
    }

    /**
     * Paint the title
     */
    void paintTitle(Graphics g) {
	// title box
	g.setColor(selected ? frameSelect : frameLight);
	g.fillRect(BORDER + TITLE + 1, BORDER + 1, width - (BORDER*2 + TITLE*3 + 2), BORDER3 - 1);
	g.setColor(frameLight);
	if (mode == INDRAG) {
	    frameRect(g, BORDER + TITLE, BORDER, width - (BORDER*2 + TITLE*3 + 1), BORDER3, false);
	} else {
	    frameRect(g, BORDER + TITLE, BORDER, width - (BORDER*2 + TITLE*3 + 1), BORDER3, true);
	}
	int w = width - (BORDER2 + TITLE*3 + 2);
	g.clipRect(BORDER + TITLE + 1, BORDER, w - 1, TITLE);
	g.setColor(frameBlack);
	g.setFont(titleFont);
	FontMetrics fm = g.getFontMetrics();
	g.drawString(title, BORDER + TITLE + 5 + (w - fm.stringWidth(title)) / 2,
		     BORDER + (TITLE + fm.getAscent()) / 2);
    }

    
    /**
     * paint the icon box
     */
    void paintIconBox(Graphics g) {
    	if (mode == ICONIZE) {	
	    g.setColor(selected ? frameSelect : frameLight);
	    g.fillRect(width - (BORDER + TITLE*2), BORDER, BORDER3, BORDER);
	    g.setColor(frameLight);
	    frameRect(g, width - (BORDER + TITLE*2), BORDER, BORDER3, BORDER3, false);
	} else {
	    frameRect(g, width - (BORDER + TITLE*2), BORDER, BORDER3, BORDER3, true);
	}
	frameRect(g, width - (BORDER + TITLE*2 - 7), BORDER + 7, TITLE - 15, TITLE - 15, true);

    }
    /**
     * 
     */
    void paintMenuBox(Graphics g) {
	if (mode == SHOWMENU) {
	    g.setColor(selected ? frameSelect : frameLight);
	    g.fillRect(BORDER, BORDER, BORDER3, BORDER3);
	    g.setColor(frameLight);
	    frameRect(g, BORDER, BORDER, BORDER3, BORDER3, false);
	} else {
	    frameRect(g, BORDER, BORDER, BORDER3, BORDER3, true);
	}
	frameRect(g, BORDER + 4, BORDER + 7, TITLE - 9, TITLE - 16, true);	
    }

    /**
     * paint the growbox
     */
    void paintGrowBox(Graphics g) {
	if (mode == MAXIMIZE) {
	    g.setColor(selected ? frameSelect : frameLight);
	    g.fillRect(width - (BORDER + TITLE), BORDER, BORDER3, BORDER3);
	    g.setColor(frameLight);
	    frameRect(g, width - (BORDER + TITLE), BORDER, BORDER3, BORDER3, false);
	} else {
	    frameRect(g, width - (BORDER + TITLE), BORDER, BORDER3, BORDER3, true);
	}
	if (state == MAXIMIZED) {
	    frameRect(g, width - (BORDER + TITLE - 4), BORDER + 4, TITLE - 9, TITLE - 9, false);
	} else {
	    frameRect(g, width - (BORDER + TITLE - 4), BORDER + 4, TITLE - 9, TITLE - 9, true);
	}
    }	

    /**
     * Paint the window borders.
     */
    void paint(Graphics g) {
	if (title == null) {
	    return;
	}

	// background
	g.setColor(selected ? frameSelect : frameLight);
	g.fillRect(0, 0, width, height);
	g.setColor(frameLight);

	// borders
	frameRect(g, 0, 0, width-1, height-1, true);
	frameRect(g, BORDER1, BORDER1, width - BORDER2, height - BORDER2, false);

	paintIconBox(g);
	paintMenuBox(g);
	paintGrowBox(g);
	
	if (resizable) {
	    frameRect(g, 0, BORDER + TITLE - 1, BORDER1, 1, false);
	    frameRect(g, width-BORDER, BORDER + TITLE - 1, BORDER1, 1, false);
	    frameRect(g, 0, height - (BORDER + TITLE - 1), BORDER1, 1, false);
	    frameRect(g, width-BORDER, height - (BORDER + TITLE - 1), BORDER1, 1, false);
	    frameRect(g, BORDER + TITLE - 1, 0, 1, BORDER1, false);
	    frameRect(g, width - (BORDER + TITLE + 1), 0, 1, BORDER1, false);
	    frameRect(g, BORDER + TITLE - 1, height-BORDER, 1, BORDER1, false);
	    frameRect(g, width - (BORDER + TITLE + 1), height-BORDER, 1, BORDER1, false);
	}

	// paint the title
	paintTitle(g);
    }

    /**
     * Paint the window borders.
     */
    void paint() {
	Graphics g = new TinyGraphics(this);
	try {
	    paint(g);
	} finally {
	    g.dispose();
	}
    }

    /**
     * Assign the focus to a new window.
     */
    void requestFocus(TinyWindow newFocus) {
	if (newFocus != focus) {
	    if (focus != null && selected == true) {
		focus.handleFocusEvent(new Event(focus, Event.LOST_FOCUS, null));	
	    }
	    focus = newFocus;
	    if (focus != null) {
		if (selected == true) {
		    focus.handleFocusEvent(new Event(focus, Event.GOT_FOCUS, null));
		} else {
		    winFocus();
		}
	    }
	}
    }

    /**
     * Check if this window has the focus.
     */
    boolean hasFocus(TinyWindow focus) {
	return focus == this.focus;
    }

    /**
     * The window now has the keyboard focus.
     */
    public void handleFocusIn() {
	selected = true;
	if (title != null) {
	    paint();
	    if (focus != null) {
		focus.handleFocusEvent(new Event(focus, Event.GOT_FOCUS, null));	
	    }
	}
    }

    /**
     * The window no longer has the keyboard focus.
     */
    public void handleFocusOut() {
	selected = false;
	if (title != null) {
	    paint();
	    if (focus != null) {
		focus.handleFocusEvent(new Event(focus, Event.LOST_FOCUS, null));
	    }
	}
    }

    /**
     * inside the menubox
     */
    boolean insideMenuBox(int x, int y) {
	return (mx > BORDER) && 	
	       (mx < BORDER + BORDER3) && 
	       (my < BORDER + TITLE);
    }
    
    /**
     * inside the GrowBox
     */
    boolean insideGrowBox(int x, int y) {
	return (mx > width - ((2 * BORDER3) + BORDER)) && 
	       (my < BORDER + TITLE) && 
	       (mx < width - BORDER);
    }

    /**
     * inside the Iconbox
     */
    boolean insideIconBox(int x, int y) {
	return (mx > width - ((2 * BORDER3) + BORDER)) && 
	       (my < BORDER + TITLE) && 
	       (mx < width - (BORDER3 + BORDER));
    }

    /**
     * inside the titlebox
     */
    boolean insideTitleBox(int x, int y) {
	return (mx > BORDER + BORDER3) && 
	       (mx < width - ((2 * BORDER3) + BORDER)) && 
	       (my < BORDER + TITLE);
    }    
    
    /**
     * when the window needs to be resized to the width and the
     * height of the screen, use this method
     */
    void handleMaximize() {
	state = MAXIMIZED;
	mode = -1;
	Dimension d = content.target.getToolkit().getScreenSize();
	maw = this.width;
	mah = this.height;
	max = this.x;
	may = this.y;
	content.target.setBounds(0, 0, d.width, d.height);	
	content.target.validate();
	content.target.repaint();
	Graphics g = new TinyGraphics(this);	

	try {
	    paintGrowBox(g);
	} finally {
	    g.dispose();
	}
    }

   /**
     * when the window needs to be resized to it's original 
     * size, use this method
     */
    void handleDeMaximize() {

	if (state != MAXIMIZED) {
	    return;
	}
	state = -1;
	mode = -1;
	content.target.setBounds(max, may, maw, mah);	
	content.target.validate();
	content.target.repaint();		      
	Graphics g = new TinyGraphics(this);	

	try {
	    paintGrowBox(g);
	} finally {
	    g.dispose();
	}		
    }

    /**
     * Event handler.
     */
    boolean handleWindowEvent(Event evt) {
	switch (evt.id) {
	  case Event.KEY_PRESS:
	  case Event.KEY_RELEASE:
	  case Event.KEY_ACTION:
	  case Event.KEY_ACTION_RELEASE:
	    if (focus != null) {
		return focus.handleKeyEvent(evt);
	    }
	    return true;

	case Event.ACTION_EVENT:

	    if (evt.arg.equals("Close")) {
		content.target.postEvent(new Event(content.target, 
						   Event.WINDOW_DESTROY, null));
	    	return true;
	    } else if (evt.arg.equals("Minimize")) {
		//REMIND: not implemented yet
	    } else if (evt.arg.equals("Maximize") && (state != MAXIMIZED)) {
		handleMaximize();
		return true;
	    } else if (evt.arg.equals("Restore")) {
	
		if (state == MAXIMIZED) {
		    handleDeMaximize();
		    return true;
		} else if (state == ICONIZED) {}   
	    }
	    return false;
	  case Event.MOUSE_ENTER:
	    winFocus();
	    return true;

	  case Event.MOUSE_UP:

	      switch (mode) {
       	      case MAXIMIZE:
		  if (insideGrowBox(evt.x, evt.y)) {
		      
		      if (state != MAXIMIZED) {
			  handleMaximize();
			  break;
		      } else if (state == MAXIMIZED) {
			  handleDeMaximize();
			  break;
		      }
		  }              

	       case ICONIZE:
		  mode = -1;
		  if (insideIconBox(evt.x, evt.y)) {
		      state = ICONIZED;
		  }
		  Graphics g = new TinyGraphics(this);	
		  try {
		      paintIconBox(g);
		  } finally {
		      g.dispose();
		  }
		  break;	
	    
	       case SHOWMENU:
		   menu.popdown(evt);
		   menu = null;
		   mode = -1;
		   g = new TinyGraphics(this);	
		   try {
		      paintMenuBox(g);
		  } finally {
		      g.dispose();
		  }
		  break;	
	    
	        case INDRAG:
		    mode = -1;
		    g = new TinyGraphics(this);	
		    try {
			paintTitle(g);
		    } finally {
			g.dispose();
		    }
		    break;
	    }
	    return true;
            
	  case Event.MOUSE_DOWN:
	      winFront();	   
	      mx = evt.x;
	      my = evt.y;
	      
	      if (insideMenuBox(mx, my)) {
		  
		  if (evt.when - mtm < 200) {
		      content.target.postEvent(new Event(content.target, 
							 Event.WINDOW_DESTROY, null));
		      return true;
		  }
		  mtm = evt.when;
		  mode = SHOWMENU;
		  Graphics g = new TinyGraphics(this);	
		  try {
		      paintMenuBox(g);
		  } finally {
		      g.dispose();
		  }
		  menu = new TinyMenuWindow(this);
		  menu.popup(); 
		  break;
	      }	else if (insideIconBox(mx, my)) {
		  mode = ICONIZE;
		  
		  Graphics g = new TinyGraphics(this);	
		  try {
		      paintIconBox(g);
		  } finally {
		      g.dispose();
		  }
	      } else if (insideGrowBox(mx, my)) {
		  mode = MAXIMIZE;
		  Graphics g = new TinyGraphics(this);	
		  try {
		      paintGrowBox(g);
		  } finally {
		      g.dispose();
		  }
	      }else if (insideTitleBox(mx, my)) {
		  mode = INDRAG;
		  Graphics g = new TinyGraphics(this);	
		  try {
		      paintTitle(g);
		  } finally {
		      g.dispose();
		  }
		  winFront();
		  break;
	      } else {
		  mode = INRESIZE;
	      }
	      return true;

	  case Event.MOUSE_DRAG:
	      switch (mode) {
	      
	      case SHOWMENU:
		  Point pt = menu.toLocal(toGlobal(evt.x, evt.y));
		  evt.x = pt.x;
		  evt.y = pt.y;
		  return menu.handleWindowEvent(evt);
	      case INDRAG:
		  content.target.move(x + (evt.x - mx), y + (evt.y - my));
		  break;	
	      case INRESIZE:
		  state = -1;
		  content.target.resize(evt.x, evt.y);
		  content.target.validate();
		  content.target.repaint();
		  break;
	      }
	      return true;
	}
	return false;
    }
}

/**
 * this class provides a popup menu for the tinyWindowManager
 */
class TinyMenuWindow extends TinyWindow {
    
    final static int MARGIN = 3;
    final static int BORDER = 2;
    int selected;
    Vector items;
    TinyWindow parent;

    /**
     * constructor
     */
    TinyMenuWindow(TinyWindowFrame parent) {
 	this.parent = parent;   
    	items = new Vector();
	items.addElement("Restore");
	items.addElement("Minimize");
	items.addElement("Maximize");
	items.addElement("Close");
    }

    /**
     * compute the size of the popup menu
     */
    void computeSize(int x, int y) {
	FontMetrics fm = TinyFontMetrics.getFontMetrics(TinyMenuBarPeer.menuFont);
	int nitems = items.size();
	int h = fm.getHeight() - fm.getLeading();
	h += (2 * MARGIN);
	int w = 40;
	
	for (int i = 0 ; i < nitems ; i++) {
	    String str = ((String)items.elementAt(i));
	    w = Math.max(w, fm.stringWidth(str) + 40);
	}
	winReshape(x, y, w, (2 * BORDER) + (h * nitems));
    }

    /**
     * show this menu
     */
    void popup() {
	selected = -1;
	int b = TinyWindowFrame.BORDER;
	int t = TinyWindowFrame.TITLE;
	winCreate(null, TINYWINDOWMANAGER);
	computeSize(parent.x + b, parent.y + b + t);
	winShow();
	winFront();
    }

    /**
     * remove this menu
     */
    void popdown(Event evt) {
	winHide();
	winDispose();
	if (selected != -1) {
	    String str = ((String)items.elementAt(selected));
	    parent.postEvent(evt.when, Event.ACTION_EVENT, 0, 0, 0, evt.modifiers, str);
	}
    }
    
   /**
    * paint this menu
    */     
    void paint(Graphics g) {
	g.setColor(TinyWindowFrame.frameLight);
	g.fillRect(0, 0, width, height);
	g.draw3DRect(0, 0, width - 1, height - 1, true);
	g.draw3DRect(1, 1, width - 3, height - 3, true);
	paintItems(g);
    }

    /**
     * handle the events of this menu
     */
    boolean handleWindowEvent(Event evt) {
	switch (evt.id) {
	case Event.MOUSE_DRAG:
	    int i = y2index(evt.y);
	    select(i);
	}
	return false;
    }

    /** 
     * calculate the index value
     */
    int y2index(int y) {
	
	if (y < 0 || y > height) {
	    return -1;
	}
	
	FontMetrics fm = TinyFontMetrics.getFontMetrics(TinyMenuBarPeer.menuFont);
	int h = fm.getHeight() - fm.getLeading();
	h += (2 * MARGIN);
	int s = BORDER + h;
	
	for (int i = 0 ; i < items.size() ; i++) {
	    if (y < s) {
		return i;
	    }
	    s += h;
	}
	return -1;
    }

    /** 
     * return the y value with the index value given
     */
    int index2y(int i) {
	
	if (i == -1) {
	    return 0;
	}
	FontMetrics fm = TinyFontMetrics.getFontMetrics(TinyMenuBarPeer.menuFont);
	int h = fm.getHeight() - fm.getLeading();
	h += (2 * MARGIN);
	return BORDER + MARGIN + (i * h);
    }

    /**
     * paint one item to the screen
     */
    void paintItem(Graphics g, int i) {
	int y = index2y(i);
	int x = BORDER + MARGIN;
	int w = width - (2 * BORDER);
	
	if (i == -1) {
	    return;
	}
	FontMetrics fm = TinyFontMetrics.getFontMetrics(TinyMenuBarPeer.menuFont);
	g.setFont(TinyMenuBarPeer.menuFont);
	int h = fm.getHeight() - fm.getLeading();
	h += (2 * MARGIN);
	
	if (i == selected) {
	    g.setColor(TinyWindowFrame.frameSelect);
	    g.fillRect(x - MARGIN, y - MARGIN, w, h);
	    g.setColor(TinyWindowFrame.frameLight);
	    g.draw3DRect(x - MARGIN, y - MARGIN, w - 1, h - 1, false);
	} else {
	    g.setColor(TinyWindowFrame.frameLight);
	    g.fillRect(x - MARGIN, y - MARGIN, w, h);
	}

	g.setColor(TinyWindowFrame.frameBlack);
	g.drawString(((String)items.elementAt(i)), x, y + fm.getAscent());
    }

    /**
     * select an item in the menu
     */
    void select(int i) {
	
	if (i == selected) {
	    return;
	}

	int s = this.selected;
	selected = i;
	Graphics g = new TinyGraphics(this);
	
	 try {
	     if (s != -1) {
		 paintItem(g, s);
	     }
	     paintItem(g, selected);
	 } finally {
	     g.dispose();
	 }
    }

    /**
     * paint all items
     */
    void paintItems(Graphics g) {
	for (int i = 0 ; i < items.size() ; i++) {
	    paintItem(g, i);
	}
    }
}






