/*
 * @(#)interrupt_md.c	1.65 98/09/15
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Solaris 2.X dependent interrupt handling code.
 */


/*
 * Header files.
 */
#include <stdio.h>
#include <signal.h>
#include <sys/trap.h>
#include <stdlib.h>
#include <string.h>

#include "hpi_impl.h"

#include "interrupt.h"
#include "itimer_md.h"
#include "internal_md.h"
#include "threads_md.h"
#include "schedule.h"
#include "signals.h"

static int pending_signals[N_INTERRUPTS];

/*
 * intrLock() -- Disable all interrupts.
 */
void
intrLock()
{
    sys_thread_t *cth = greenThreadSelf();
    machdep_t *machdepP;
    sigset_t newMask;
    
    if (cth) {
        machdepP = MACHDEP(cth);

        /* Sanity check */
	sysAssert(machdepP->intrLockCount >= 0);
    
	if (machdepP->intrLockCount++ == 0) {
	    sigfillset(&newMask);
	    green_sigprocmask(SIG_BLOCK, &newMask, &machdepP->intrLockOldMask);
	}
    } else {
	sigfillset(&newMask);
	green_sigprocmask(SIG_BLOCK, &newMask, NULL);
    }
}

/*
 * intrUnlock() -- Enable all interrupts.
 */
void
intrUnlock()
{
    sys_thread_t *cth = greenThreadSelf();
    machdep_t *machdepP;
    sigset_t newMask;
    
    if (cth) {
        machdepP = MACHDEP(cth);

        /* Sanity check */
	sysAssert(machdepP->intrLockCount > 0);
    
	if (--machdepP->intrLockCount == 0) {
	    green_sigprocmask(SIG_SETMASK, &machdepP->intrLockOldMask, NULL);
	}
    } else {
	sigemptyset(&newMask);
	green_sigprocmask(SIG_SETMASK, &newMask, NULL);
    }
}

/*
 * intrInitMD() -- Target-specific initialization.
 */
extern void
intrInitMD()
{
    memset(pending_signals, 0, sizeof(pending_signals));
    sigignore(SIGPIPE);
}

/*-
 * intrDispatchMD() -- Turn our signal into an intrDispatch().
 */
void
intrDispatchMD(int sig, siginfo_t *info, void *uc)
{
    int	old_errno = errno;
    sys_thread_t *self = greenThreadSelf();
    ucontext_t *ucp = (ucontext_t *)uc;

    Log1(1, "signalHandlerDispatch(sig=%d)\n", sig);

    SCHED_LOCK();

    intrDispatch(sig, info, uc);

    errno = old_errno;

    /*
     * We are not preempting the current thread.  Just return
     * and let the signal trampoline restore the state of
     * the interrupted thread.
     */
    sysAssert(!(SCHED_LOCKED() > 1 && _needReschedule));
    if (_needReschedule) {

	_needReschedule = 0;
	sysAssert(SCHED_LOCKED() == 1);

	/* DEBUG check that saved mask is correct */
	sysAssert(self->sigmaskdirty ||
	    memcmp(&ucp->uc_sigmask, &MACHDEP(self)->signalMask,
	    sizeof ucp->uc_sigmask) == 0);

	/* Force reload of current signal mask */
	self->sigmaskdirty = 1;

	YIELD();
    }
    SCHED_UNLOCK();

    if (SCHED_LOCKED() == 0) {
	/*
	 * Restore cached sigmask to the saved value, so that it will
	 * be correct when we return from the signal handler.
	 *
	 * It is important that all signals be masked until we return,
	 * which is currently true.  If this changes, block all signals
	 * before setting signalMask.
	 */
	MACHDEP(self)->signalMask = ucp->uc_sigmask;
    }
}

bool_t intrInUse(int sig)
{
    /* Signals used in green threads implementation. */
    return sig == SIGPIPE ||
      sig == SIGIO ||
      sig == SIGCHLD ||
      sig == SIGALRM;
}

void sysSignalNotify(int sig)
{
    sys_mon_t *mon = asyncMon(SYS_ASYNC_MON_USRSIG);
    pending_signals[sig]++;
    if (asyncEventNotify(mon)) {
	_needReschedule = 1;
    }
}

static int lookupSignal(void)
{
    int i;
    for (i = 0; i < N_INTERRUPTS; i++) {
        if (pending_signals[i]) {
	    pending_signals[i]--;
	    return i;
	}
    }
    return -1;
}

int sysSignalWait()
{
    sys_mon_t *mon = asyncMon(SYS_ASYNC_MON_USRSIG);
    sys_thread_t *self = sysThreadSelf();
    int sig;
    sysMonitorEnter(self, mon);
    while ((sig = lookupSignal()) == -1) {
        sysMonitorWait(self, mon, SYS_TIMEOUT_INFINITY);
    }
    sysMonitorExit(self, mon);
    return sig;
}
