/*
 * @(#)iomgr.c	1.109 98/09/15
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * I/O management, including non-blocking io.
 *
 * Eventually this module needs to have a formal interface, and
 * be replacable (e.g. it needs to be a porting layer).
 */

#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <fcntl.h>
#include <sys/fcntl.h>
#include <sys/param.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <errno.h>
#include <sys/filio.h>
#include <sys/ioctl.h>
#include <sys/uio.h>
#include <limits_md.h>
#include <stdio.h>
#include <sys/resource.h>
#include <dlfcn.h>

#include <stropts.h>
#include <sys/conf.h>
#include <poll.h>
#include <string.h>

#include "hpi_impl.h"

#include "threads_md.h"
#include "monitor_md.h"
#include "itimer_md.h"
#include "clock.h"
#include "iomgr.h"
#include "signals.h"

/*
 * POSIX/SYS API NOTES
 *
 * Since we have our own thread library (green), we want to override
 * regular system calls (POSIX open, read, close, etc) to do async IO.
 *
 * We also want to provide the sys<call> API (sysRead, sysOpen, etc),
 * which is the portability layer of the VM. These calls have a
 * different API from the POSIX call (they have different return codes
 * and the sys API doesn't use errno, for example), but share a lot of
 * implementation. It used to be that that functionality was split
 * into two files: iomgr.c, for strict POSIX and io_md.c for sys
 * calls, duplicating a lot of functionality.
 *
 * All the functionality now resides in the implementation of the
 * POSIX functions, and resides in this file. That implementation,
 * however, supports both APIs: if a the interruptible_io flag is set
 * on the current thread (sysThreadSelf()), it implements the sys API,
 * and if it isn't, it implements the POSIX API.
 *
 * As a result parts the sys API (that which doesn't get mapped
 * straight to POSIX) is implemented as a set of wrapper functions in
 * io_td.c which set the interruptible_io bit, call the POSIX API and
 * restore the old value. That bit is retrieved using the
 * GET_INTERRUPT_IO macro, defined in iomgr.h.
 *
 * ASYNC I/O IMPLEMENTATION NOTES
 *
 * fd_flags is an array indexed by the file descriptor used
 * to keep information about the state of the fd i.e. if it has 
 * been made nonblocking or whether its marked to be closed.
 *
 * When fd is opened, the per fd monitors are allocated (i.e. if
 * they have not been preallocated at init time. On the first
 * read/write/read/writev call the fd is marked nonblocking.
 * On close, the state for fd is cleared. The monitors are
 * not freed as they can be reused for subsequesnt opens refering
 * to same fd.
 *
 * After converting to the nonblocking mode, the io calls which return
 * EAGAIN need a condition variable/monitor to wait on. We provide 
 * a monitor per fd for this purpose.
 *
 * There is also a default async IO monitor which serves as condition 
 * variable or monitor for calls like poll and select which deal
 * with fd sets.
 *
 * Also a per fd reference count is maintained as to who is currently 
 * in the io manager doing io or waiting for an io signal. Now on close
 * the state of the fd is marked closed and if ref count is zero i.e. no
 * one else is in middle of io, then only system_close is called on fd.
 * If ref count is greater than zero, everyone waiting on the fd monitor
 * is woken up and the last thread to leave the io manager closes the fd.
 *
 * Thus per fd monitor guards two structures: fd_flags and fd_ref 
 *
 * FURTHER ASYNC I/O NOTES
 *
 * Note on interaction with (Java, not O/S) interrupts: because the
 * async I/O implemented in this file uses sysMonitorWait(), it needs
 * to work sensibly when sysMonitorWaits() return prematurely because
 * the waiting thread was interrupted.  The approach taken is based on
 * the fact that the code here is not really meant to be called from
 * the Java runtime, but rather from user native methods in replacement
 * of the routines normally supplied by native libraries.  The APIs
 * implemented should be that of the replaced libraries.  Given that
 * those don't know anything about Java-style interrupts, and that
 * the EINTR errno values that some of them set isn't really the same
 * concept, on interrupt we clear the interrupted state, record that
 * the interrupt happened, then continue with the call as though
 * nothing happened.  Once the call completes normally, if we had
 * suppressed an interrupt, the interrupted state is reset to propagate
 * to the caller.
 */

char *fd_flags;
fd_reference *fd_ref;

/*
 * io_lock guards the state of the io control data structures:
 * fdmon: per fd monitor
 * fdMap: structure taht maps fd->index into poll table
 * PollTable: used in poll call in the SIGIO signal handler 
 *            to decide whihc fd the io occured on.
 * 
 * For more details look at comments in signals.c
 */
sys_mon_t _io_lock;
sys_mon_t **fdmon;

/*
 * This lock protects the dl wrappers, assuring that two threads aren't
 * in libdl at the same time.
 */
sys_mon_t _dl_lock;

/*
 * The variable which records maximum number of files that
 * can be open at a time. Its initialized at init time and
 * is also the size of structures like fdmon_, fdMap and 
 * fd_flags.
 */
static int max_files = 0;
#define PREALLOCATE_FDS 16

#if defined(GPROF) && !defined(NATIVE)
extern int mainAlreadyEntered;
#endif

#define PRESERVE_ERRNO \
    int saved_errno = errno
#define RESTORE_ERRNO(ret) \
    if (ret >= 0) {		\
	errno = saved_errno;	\
    } else ((void) 0)

#define SIMPLE_INITFD(newfd, funcname) \
if (1) {							\
    if (newfd >= 0) {					\
	Log1(1, #funcname " fd: %d \n", newfd);		\
	sysAssert(fd_flags[newfd] == 0);		\
	fd_flags[newfd] = 0;				\
	if (initialize_monitors(newfd) == B_FALSE) {	/* Out of memory */ \
	    (*systable[SYS_CLOSE].addr)(newfd);		/* Close new fd */  \
	    errno = ENOMEM;				\
	    newfd = -1;					\
	}						\
    } else {						\
	Log1(1, #funcname " error: %d\n", errno);	\
    }							\
} else							\
    ((void) 0)

#define IO_OPTIMIZE /* REMIND: should this be here? */

syscall_t systable[] = {
	{ "write", 0 },
	{ "writev", 0 },
	{ "read", 0 },
	{ "readv", 0 },
	{ "getmsg", 0 },
	{ "putmsg", 0 },
	{ "poll", 0 },
	{ "open", 0 },
	{ "close", 0 },
	{ "pipe", 0 },
	{ "fcntl", 0},
	{ "dup", 0 },
	{ "creat", 0 },
	{ "accept", 0 },
	{ "recv", 0 },
	{ "send", 0 },
	{ "recvfrom", 0 },
	{ "sendto", 0 },
	{ "socket", 0 },
	{ "connect", 0 },
	{ "select", 0 },
 	{ "ioctl", 0 },
	{ "dlopen", 0 },
	{ "dlclose", 0 },
	{ "dlerror", 0 },
	{ "dlsym", 0 },
	{ "dladdr", 0 },
	{ "__filbuf", 0 }, /* unused */
	{ (char *) 0, 0 }
};

/* macro to copy fd_set. fd_set is an array of fd_mask */
/* NFDBITS is defined in select.h to be bits per fd_mask */
#define COPY_SET(_to, _from, _width)              \
        memcpy(_to, _from,                        \
               ((_width + NFDBITS-1) / (NFDBITS)) \
                * sizeof(fd_mask)                 \
               );

/*
 * Have we bootstrapped the primordial thread yet?
 */
bool_t threadBootstrappedP = FALSE;
static bool_t wrappersInited = FALSE;

/*
 * Call this to let us know that the VM has successfully bootstrapped
 * the primordial thread and that we can now make redirect all system
 * calls.
 */
void PrimordialThreadBootstrapped(bool_t c)
{
    threadBootstrappedP = c;
}

/************************************************************************
 * initialization and utility functions
 */

/*
 * Warning: initialize_monitors() does not itself deal with running
 * out of malloc memory, besides returning B_FALSE when this happens.
 * It is up to the caller to deal with the situation.
 */
static boolean_t
initialize_monitors(int desc)
{
    if (fdmon[desc] == 0) {
    	if ((fdmon[desc] =
		(sys_mon_t *) calloc(1, sizeof(sys_mon_t))) == NULL) {
	    sysAssert(fdmon[desc] != NULL);	/* Out of memory */
	    return B_FALSE;
	}
	sysMonitorInit(fdmon[desc]);
    }

    return B_TRUE;
}

static bool_t
initializeTTY(int fd, int mode)
{
    char * name;
    int flags;
    int newfd = -1;
    int result;

    /*
     *  If it's not a tty then we can safely make it non blocking
     */
    if (!isatty(fd))
	return TRUE;
    
    /*
     * Get the ttyname, so we can reopen it.
     */
    name = ttyname(fd);
    sysAssert(name != NULL);
    if (name == NULL)
	return FALSE;
    
    /*
     * we reopen the terminal we're attached to so that when we
     * set it nonblocking, it won't be inherited by our parent.
     */

    newfd = open(name, mode);
    if (newfd == -1)
	return FALSE;

    /* get the flag settings from the fd we inherited */
    while ((flags = fcntl(fd, F_GETFL, NULL))  == -1 && errno == EINTR)
	;
    if (flags == -1)
	goto failed;

    /* mark the new fd with our old flags */
    while ((result = fcntl(newfd, F_SETFL, flags))  == -1 && errno == EINTR)
	;
    if (result == -1)
	goto failed;

    /* replace the original fd with our new fd */
    while ((result = dup2(newfd, fd))  == -1 && errno == EINTR)
	;
    if (result == -1)
	goto failed;

    close(newfd);
    return TRUE;

  failed:
    close(newfd);
    return FALSE;
}


static void
initializeWrappers(void)
{
    syscall_t *tp = systable;
    struct rlimit nbr_files;
    int i;

    if (wrappersInited)
	return;

    /* Take whatever fd limit the shell has given us: */
    getrlimit(RLIMIT_NOFILE, &nbr_files);

    /* We are running out of file descriptors in fonts.
     * As a temporary fix, bump up the number of open
     * file descriptors to OPEN_MAX.
     */
    nbr_files.rlim_cur = nbr_files.rlim_max;
    setrlimit(RLIMIT_NOFILE, &nbr_files);

    max_files = nbr_files.rlim_cur;

    fdmon = (sys_mon_t **) calloc(max_files, sizeof(sys_mon_t *));
    if (fdmon == 0) {
	VM_CALL(panic)("out of memory");	/* Fatal startup error */
    }

    for (i = 0; i < PREALLOCATE_FDS; ++i) {
	if (initialize_monitors(i) == B_FALSE) {
	    VM_CALL(panic)("out of memory");	/* Fatal startup error */
	}
    }

    fd_flags = (char *) calloc(max_files, sizeof(char));
    fd_ref = (fd_reference *) calloc(max_files, sizeof(fd_reference));
    if ((fd_flags == 0) || (fd_ref == 0)) {
	VM_CALL(panic)("out of memory");	/* Fatal startup error */
    }

    while (tp->sym != (char *) 0) {
	tp->addr = (sysfunc_t) dlsym(RTLD_NEXT, tp->sym);
	tp++;
    }

    wrappersInited = TRUE;
}

/*
 * External routine:
 * DO all the necessary nonblocking io initialization.
 * Called at startup time from javai main routine.
 */
void InitializeAsyncIO(void)
{
    initializeWrappers();

    InitializeAsyncMonitors(max_files);
    IO_LOCK_INIT();  
    VM_CALL(monitorRegister)(&_dl_lock, "Dynamic loading lock");

    /*
     * If the fd is stdin, stderr or stdout we do not want to
     * make the fd nonblocking otherwise if you exit java
     * without making it blocking again, then your parent ends with a
     * a nonblocking fd, which it isn't likely to be able to deal with.
     * instead we try to find the terminal which the fd corresponds to
     * and open it directly.  That way we are talking to the right thing
     * but we can screw with it to our heart'ss content without breaking
     * other programs.  If it's not a TTY then we can just treat it
     * normally.  See bug 1237893.
     */
    if (initializeTTY(0, O_RDONLY) == FALSE)
	fd_flags[0] = FD_NBINIT; 
    /*
     * We could do the same for standard in and error but then we end up
     * disassociated from the controlling terminal so when we try to write
     * to it the process stops with SIGTOU.  See bugs 4114449 and 1237893
     */
    fd_flags[1] = FD_NBINIT; 
    fd_flags[2] = FD_NBINIT;

}

/*
 * If the second argument is true the fd is put into non-blocking
 * mode.  That is, a read when there is no data will not stop
 * the entire heavyweight process.
 *
 * Blocking io is currently being used at least one place - makeTTYSane
 * routine in javai.c.
 */

int nonblock_io(int desc, int onoff) {
    long flgs;
    sys_thread_t *self = sysThreadSelf();
 
    if (desc < 0 || desc > max_files || fdmon[desc] == NULL) {
	errno = EBADF;
	return -1;
    }

    sysAssert(desc < max_files);
    sysAssert(fdmon[desc] != NULL);

    sysMonitorEnter(self, fdmon[desc]); 
    IO_LOCK(self); /* Activate/Deactivate routine need this lock held */

    /* 
     * If you do not want to block and fd is nonblocking - do nothing
     * otherwise make fd nonbloking.
     */
    if ((onoff == IO_DONTBLOCK) && !(fd_flags[desc] & FD_NBINIT)) {
	/* if it's already non-blocking, do nothing... */
	if ((flgs = fcntl(desc, F_GETFL, 0)) == -1) {
	    IO_UNLOCK(self);
	    sysMonitorExit(self, fdmon[desc]);
	    return -1;
	}

	if (flgs & O_NONBLOCK) {
	    /*
	     * the fd started out non-blocking so record that the fd
	     * should have non-blocking semantics inside the wrappers
	     */
	    fd_flags[desc] |= FD_NONBLOCKING;
	} else {
	    /*
	     * Make the fd non-blocking for use by the wrappers
	     */
	    systable[SYS_FCNTL].addr(desc, F_SETFL, flgs | O_NONBLOCK);
	}

	ioctl(desc, I_SETSIG, S_INPUT | S_OUTPUT | S_HANGUP | S_HIPRI);

	asyncIOActivateFD(desc);
	fd_flags[desc] |= FD_NBINIT;

	Log1(1, "Made %d nonblocking\n", desc);

    } else if ((onoff == IO_BLOCK) && (fd_flags[desc] & FD_NBINIT)) {
 	/*
	 * If it's already blocking, do nothing... 
	 * otherwise make fd blocking.
	 */
	if ((flgs = fcntl(desc, F_GETFL, 0)) == -1) {
	    IO_UNLOCK(self);
	    sysMonitorExit(self, fdmon[desc]);
	    return -1;
	}

	fcntl(desc, F_SETFL, flgs & ~O_NONBLOCK);
	ioctl(desc, I_SETSIG, 0);

	fd_flags[desc] &= ~FD_NBINIT;

	Log1(1, "Made %d blocking\n", desc);
    }

    IO_UNLOCK(self);
    sysMonitorExit(self, fdmon[desc]);
    return 0;
}



/************************************************************************
 * open
 */

/*
 * We need to override every call through which we can get a new file
 * descriptor and make sure we are initializing the monitors for the new
 * fds before we return control. Known calls that are already taken care
 * of are: open, fcntl, pipe dup and creat.
 */

/*
 * Although we set the SA_RESTART flag on all our signal handlers we
 * still see open() occasionally return with errno == EINTR, at
 * least on SunOS 2.3.  This routine is an attempt to get aorund
 * this bug by making sure open does not return until it either
 * succeeds or errno != EINTR.
 */
int
open(const char *fname, int flags, ...)
{
    PRESERVE_ERRNO;
    int newfd = -1;
    va_list ap;
    unsigned int mode;
    sys_thread_t *self = sysThreadSelf();

    va_start(ap, flags);
    if (flags & O_CREAT) {
	mode = va_arg(ap, unsigned int);
    } else {
	mode = 0;
    }
    va_end(ap);

    if (threadBootstrappedP == FALSE) {
	if (!wrappersInited) initializeWrappers();
	newfd = (*systable[SYS_OPEN].addr)(fname, flags, mode);
	SIMPLE_INITFD(newfd, open);
	return newfd;
    }

    IO_LOCK(self);

    while(1) {
	newfd = (*systable[SYS_OPEN].addr)(fname, flags, mode);
	if ((newfd != -1) ||
	    ((errno != EAGAIN) && (errno != EINTR))) {
	    break;
	}
    }
 
    SIMPLE_INITFD(newfd, open);

    IO_UNLOCK(self);
    RESTORE_ERRNO(newfd);
    return newfd;
}


/************************************************************************
 * close
 */

/* Hold the per fd lock for the duration of system close */
int
system_close(int fd)
{
    int ret = -1;
    sys_thread_t *self = sysThreadSelf();

    IO_LOCK(self);
    if (fd_flags[fd] & FD_NBINIT) { 
        fd_flags[fd] = 0; 
        asyncIODeactivateFD(fd);
    }

    ret = (*systable[SYS_CLOSE].addr)(fd);
    Log1(1, "Closed fd: %d\n", fd);

    IO_UNLOCK(self);
    return ret;
}

int
close(int fd)
{
    PRESERVE_ERRNO;
    int ret = -1;
    sys_thread_t *self = sysThreadSelf();

    if (threadBootstrappedP == FALSE) {
	if (!wrappersInited) initializeWrappers();
	return (*systable[SYS_CLOSE].addr)(fd);
    }

    /*
     * We seem to call close on fd multiple times for
     * some reason - to be fixed in upper layer.
     */
    if (fd < 0 || fd >= max_files || fdmon[fd] == NULL) {
        Log1(1, "Closing bad fd: %d?\n", fd);
	errno = EBADF;
        return ret;
    }
    sysAssert(fdmon[fd] != NULL);
    sysMonitorEnter(self, fdmon[fd]);

    if (FD_INACTIVE(fd)) {
	if (!closed_fd(fd)) {
	    ret = system_close(fd);
	}
    } else {
	if (!closed_fd(fd)) {
	    fd_flags[fd] |= FD_CLOSED;
	    sysMonitorNotifyAll(self, fdmon[fd]);
	    /* we must enter the asyncMon to notify it */
	    sysMonitorEnter(self, asyncMon(SYS_ASYNC_MON_IO));
	    sysMonitorNotifyAll(self, asyncMon(SYS_ASYNC_MON_IO));
	    sysMonitorExit(self, asyncMon(SYS_ASYNC_MON_IO));
	    Log1(1, "Marked Closed fd: %d\n", fd);
	}
	ret = 0;
    }

    sysMonitorExit(self, fdmon[fd]);

    RESTORE_ERRNO(ret);
    return ret;
}


/************************************************************************
 * read
 */

ssize_t
read(int fd, void *buf, unsigned int nbytes)
{
    PRESERVE_ERRNO;
    sys_mon_t *mon;
    int nread = -1;
    sys_thread_t *self = sysThreadSelf();

    bool_t wasInterrupted = FALSE;
    int interrupt_io = GET_INTERRUPT_IO();

    if (threadBootstrappedP == FALSE) {
	if (!wrappersInited) initializeWrappers();
	return (*systable[SYS_READ].addr)(fd, buf, nbytes);
    }

    if (fd < 0 || fd >= max_files || fdmon[fd] == NULL) {
        Log1(1, "read on bad fd: %d?\n", fd);
	errno = EBADF;
        return -1;
    }

    check_single_fd(fd);

#ifdef IO_OPTIMIZE
    nread = (*systable[SYS_READ].addr)(fd, buf, nbytes);
    if ((nread != -1) ||
	(errno != EAGAIN && errno != EINTR) || NONBLOCKING_FD(fd)) {
	Log2(1, "Read for %d bytes finished (did not block) on fd: %d\n", 
	     nread, fd);
	return nread;
    }
#endif

    mon = fdmon[fd];
    sysAssert(mon != NULL);

    sysMonitorEnter(self, mon);
    FD_ADD_READER(fd);

    while(!closed_fd(fd)) {
	nread = (*systable[SYS_READ].addr)(fd, buf, nbytes);
	if ((nread != -1) ||
	    ((errno != EAGAIN) && (errno != EINTR)) || NONBLOCKING_FD(fd)) {
	    break;
	}
	if (errno == EAGAIN) {
	    if (sysMonitorWait(self, mon,
			       SYS_TIMEOUT_INFINITY) == SYS_INTRPT) {
		/* We've cleared interrupted state; remember and reraise */
		if (interrupt_io) {
		    nread = SYS_INTRPT; 	      /* INTERRUPTED */
		    break;
		} else {
		    wasInterrupted = TRUE;
		}
	    }
	}
    }

    if (wasInterrupted == TRUE) {
	sysThreadInterrupt(sysThreadSelf());	/* Reraise interrupt */
    }

    FD_REMOVE_READER(fd);
    /* 
     * If I am the last person to leave and fd has been closed
     * call system close on the fd.
     */
    if (FD_INACTIVE(fd) && closed_fd(fd)) {
	system_close(fd); 
	/* set EBADF ? */
    }

    Log2(1, "Read for %d bytes finished on fd: %d\n", nread, fd);
    sysMonitorExit(self, mon);
    RESTORE_ERRNO(nread);
    return nread;
}


/* TBD: optimize like read routine - pavani */
/* signature changed on Solaris 2.6! -brown */

#if defined(_ILP32) || defined(_LP64)
int
readv(int fd, const struct iovec *iov, int iovcnt)
#else
int
readv(int fd, struct iovec *iov, int iovcnt)
#endif
{
    PRESERVE_ERRNO;
    sys_mon_t *mon;
    int nread = -1;
    bool_t wasInterrupted = FALSE;
    sys_thread_t *self = sysThreadSelf();

    if (threadBootstrappedP == FALSE) {
	if (!wrappersInited) initializeWrappers();
	return (*systable[SYS_READV].addr)(fd, iov, iovcnt);
    }

    if (fd < 0 || fd >= max_files || fdmon[fd] == NULL) {
        Log1(1, "readv on bad fd: %d?\n", fd);
	errno = EBADF;
        return -1;
    }

    mon = fdmon[fd];
    sysAssert(mon != NULL);

    check_single_fd(fd);

    sysMonitorEnter(self, mon);
    FD_ADD_READER(fd);

    while(!closed_fd(fd)) {
	nread = (*systable[SYS_READV].addr)(fd, iov, iovcnt);
	if ((nread != -1) ||
	    ((errno != EAGAIN) && (errno != EINTR)) || NONBLOCKING_FD(fd)) {
	    break;
	}
	if (errno == EAGAIN) {
	    if (sysMonitorWait(self, mon,
			       SYS_TIMEOUT_INFINITY) == SYS_INTRPT) {
		/* We've cleared interrupted state; remember and reraise */
		wasInterrupted = TRUE;
	    }
	}
    }

    if (wasInterrupted == TRUE) {
	sysThreadInterrupt(sysThreadSelf());    /* Reraise interrupt */
    }

    FD_REMOVE_READER(fd);
    /* 
     * If I am the last person to leave and fd has been closed
     * call system close on the fd.
     */
    if (FD_INACTIVE(fd) && closed_fd(fd)) {
	system_close(fd); 
	/* set EBADF ? */
    }

    Log2(1, "Readv for %d bytes finished on fd: %d\n", nread, fd);
    sysMonitorExit(self, mon);
    RESTORE_ERRNO(nread);
    return nread;
}


/************************************************************************
 * write
 */

ssize_t
write(int fd, const void *buf, unsigned int nbytes)
{
    PRESERVE_ERRNO;
    sys_mon_t *mon;
    int nwrote;
    caddr_t next = (caddr_t)buf;
    int total_written = 0;
    bool_t wasInterrupted = FALSE;
    sys_thread_t *self = sysThreadSelf();

    if (threadBootstrappedP == FALSE) {
	if (!wrappersInited) initializeWrappers();
	return (*systable[SYS_WRITE].addr)(fd, buf, nbytes);
    }

    if (fd < 0 || fd >= max_files || fdmon[fd] == NULL) {
        Log1(1, "write on bad fd: %d?\n", fd);
	errno = EBADF;
        return -1;
    }
    check_single_fd(fd);

#ifdef IO_OPTIMIZE
    while (total_written < nbytes) {
	nwrote = (*systable[SYS_WRITE].addr)(fd, next + total_written,
			 		     nbytes - total_written);
	if (nwrote == -1) {
	    break;
	}
	total_written += nwrote;
    }	

    /* 
     * If done - return; basically whole write succeeded immediately.
     * If error was other than EINTR and EAGAIN - return immediately.
     * If there was a pending exception return immediately.
     */
    if ((nwrote != -1 && (total_written == nbytes)) ||
        ((nwrote == -1) && (errno != EAGAIN && errno != EINTR)) ||
	NONBLOCKING_FD(fd)) {
	Log2(1, "Write for %d bytes finished (did not block) on fd: %d\n",
	     nwrote, fd);
	return ((nwrote != -1) ? total_written : -1);
    }
    /* else take the slow path */
#endif

    mon = fdmon[fd];
    sysAssert(mon != NULL);

    sysMonitorEnter(self, mon); 
    FD_ADD_WRITER(fd);

    while(!closed_fd(fd)) {
	nwrote = (*systable[SYS_WRITE].addr)(fd, next + total_written,
			 		     nbytes - total_written);
	if ((nwrote == -1) && (errno != EAGAIN && errno != EINTR)) {
            total_written = -1;
	    break;
        }
	if (nwrote == -1) { 
	    /* with errno == EAGAIN or EINTR */
	    if (errno == EAGAIN) {
		if (sysMonitorWait(self, mon,
				   SYS_TIMEOUT_INFINITY) == SYS_INTRPT) {
		    /* We've cleared interrupted state; remember and reraise */
		    wasInterrupted = TRUE;
		}
	    }
	    continue;
	}
        total_written += nwrote;
        if (total_written >= nbytes) {
	    break;
	}
    } /* end of while */

    if (wasInterrupted == TRUE) {
	sysThreadInterrupt(sysThreadSelf());    /* Reraise interrupt */
    }

    /* 
     * If there was an error, make sure we return -1.  This covers the
     * case when the code in the while loop above is never executed.
     */
    if (closed_fd(fd)) {
	total_written = -1;
    }

    FD_REMOVE_WRITER(fd);
    /* 
     * If I am the last person to leave and fd has been closed
     * call system close on the fd.
     */
    if (FD_INACTIVE(fd) && closed_fd(fd)) {
	system_close(fd); 
	/* set EBADF ? */
    }

    Log2(1, "Write for %d bytes finished on fd: %d\n", nwrote, fd);
    sysMonitorExit(self, mon);
    RESTORE_ERRNO(total_written);
    return total_written;
}

int
writev(int fd, const struct iovec *iovec, int iovcnt)
{
    PRESERVE_ERRNO;
    sys_mon_t *mon;
    int total_written = 0;
    int nwrote, ndone = 0;
    struct iovec *iov = (struct iovec *)iovec;
    bool_t wasInterrupted = FALSE;
    sys_thread_t *self = sysThreadSelf();

    if (threadBootstrappedP == FALSE) {
	if (!wrappersInited) initializeWrappers();
	return (*systable[SYS_WRITEV].addr)(fd, iovec, iovcnt);
    }

    if (fd < 0 || fd >= max_files || fdmon[fd] == NULL) {
        Log1(1, "writev on bad fd: %d?\n", fd);
	errno = EBADF;
        return -1;
    }

    mon = fdmon[fd];
    sysAssert(mon != NULL);

    check_single_fd(fd);

    sysMonitorEnter(self, mon);
    FD_ADD_WRITER(fd);

    while (iovcnt > 0 && !closed_fd(fd)) {
	iov->iov_len -= ndone;
	iov->iov_base += ndone;
	nwrote = (*systable[SYS_WRITEV].addr)(fd, iov, iovcnt);

	while (nwrote == -1) {
	    if (errno == EAGAIN) {
		if (sysMonitorWait(self, mon, 
				   SYS_TIMEOUT_INFINITY) == SYS_INTRPT) {
		    /* We've cleared interrupted state; remember and reraise */
		    wasInterrupted = TRUE;
		}
	    } else if (errno != EINTR) {
		total_written = -1;
		goto unlock;
	    }
	    nwrote = (*systable[SYS_WRITEV].addr)(fd, iov, iovcnt);
	}
	total_written += nwrote;
	iov->iov_len += ndone;
	iov->iov_base -= ndone;
	ndone += nwrote;
	for (; (iovcnt > 0) && (ndone >= iov->iov_len) ; iovcnt--, iov++) {
	    ndone -= iov->iov_len;
	}
    }

unlock:
    if (wasInterrupted == TRUE) {
        sysThreadInterrupt(sysThreadSelf());    /* Reraise interrupt */
    }

    FD_REMOVE_WRITER(fd);
    /*
     * If I am the last person to leave and fd has been closed
     * call system close on the fd.
     */
    if (FD_INACTIVE(fd) && closed_fd(fd)) {
	system_close(fd); 
	/* set EBADF ? */
    }

    Log2(1, "Writev for %d bytes finished on fd: %d\n", total_written, fd);
    sysMonitorExit(self, mon);
    RESTORE_ERRNO(total_written);
    return total_written;
}


/************************************************************************
 * recv and recvfrom
 */

ssize_t
recv(int fd, char *buf, int nbytes, int flags)
{
    PRESERVE_ERRNO;
    sys_mon_t *mon;
    int nread = -1;
    sys_thread_t *self = sysThreadSelf();

    bool_t wasInterrupted = FALSE;
    int interrupt_io = GET_INTERRUPT_IO();

    if (threadBootstrappedP == FALSE) {
	if (!wrappersInited) initializeWrappers();
	return (*systable[SYS_RECV].addr)(fd, buf, nbytes, flags);
    }

    if (fd < 0 || fd >= max_files || fdmon[fd] == NULL) {
        Log1(1, "recv on bad fd: %d?\n", fd);
	errno = EBADF;
        return -1;
    }

    mon = fdmon[fd];

    sysAssert(mon != NULL);

    check_single_fd(fd);

    sysMonitorEnter(self, mon);
    FD_ADD_READER(fd);

    while(!closed_fd(fd)) {
	nread = (*systable[SYS_RECV].addr)(fd, buf, nbytes, flags);
	if ((nread != -1) ||
	    ((errno != EAGAIN) && (errno != EINTR)) || NONBLOCKING_FD(fd)) {
	    break;
	}
	if (errno == EAGAIN) {
	    if (sysMonitorWait(self, mon, 
			       SYS_TIMEOUT_INFINITY) == SYS_INTRPT) {
		/* We've cleared interrupted state; remember and reraise */
		if (interrupt_io) {
		    nread = SYS_INTRPT;
		    break;
		} else {
		    wasInterrupted = TRUE;
		}
	    }
	}
    }

    if (wasInterrupted == TRUE) {
	sysThreadInterrupt(sysThreadSelf());    /* Reraise interrupt */
    }

    FD_REMOVE_READER(fd);
    /* 
     * If I am the last person to leave and fd has been closed
     * call system close on the fd.
     */
    if (FD_INACTIVE(fd) && closed_fd(fd)) {
	system_close(fd); 
	/* set EBADF ? */
    }

    Log2(1, "recv for %d bytes finished on fd: %d\n", nread, fd);
    sysMonitorExit(self, mon);
    RESTORE_ERRNO(nread);
    return nread;
}

ssize_t
recvfrom(int fd, char *buf, int nbytes, int flags,
	 struct sockaddr *from, int *fromlen)
{
    PRESERVE_ERRNO;
    sys_mon_t *mon;
    int nread = -1;
    sys_thread_t *self = sysThreadSelf();

    bool_t wasInterrupted = FALSE;
    int interrupt_io = GET_INTERRUPT_IO();

    if (threadBootstrappedP == FALSE) {
	if (!wrappersInited) initializeWrappers();
	return (*systable[SYS_RECVFROM].addr)(fd, buf, nbytes,
					      flags, from, fromlen);
    }

    if (fd < 0 || fd >= max_files || fdmon[fd] == NULL) {
        Log1(1, "recvfrom on bad fd: %d?\n", fd);
	errno = EBADF;
        return -1;
    }
    mon = fdmon[fd];

    sysAssert(mon != NULL);

    check_single_fd(fd);

    sysMonitorEnter(self, mon);
    FD_ADD_READER(fd);

    while(!closed_fd(fd)) {
	nread = (*systable[SYS_RECVFROM].addr)
	    (fd, buf, nbytes, flags, from, fromlen);
	if ((nread != -1) || closed_fd(fd) ||
	    ((errno != EAGAIN) && (errno != EINTR)) || NONBLOCKING_FD(fd)) {
	    break;
	}
	if (errno == EAGAIN &&
	    sysMonitorWait(self, mon,
			   SYS_TIMEOUT_INFINITY) == SYS_INTRPT) {
	    /* We've cleared interrupted state; remember and reraise */
	    if (interrupt_io) {
		nread = SYS_INTRPT;
		break;
	    } else {
		wasInterrupted = TRUE;
	    }
	}
    }

    if (wasInterrupted == TRUE) {
	sysThreadInterrupt(sysThreadSelf());	/* Reraise interrupt */
    }

    FD_REMOVE_READER(fd);
    /* 
     * If I am the last person to leave and fd has been closed
     * call system close on the fd.
     */
    if (FD_INACTIVE(fd) && closed_fd(fd)) {
	system_close(fd);
	if (interrupt_io) {
	    errno = EBADF;
	}
    }
    Log2(1, "recvfrom for %d bytes finished on fd: %d\n", nread, fd);
    sysMonitorExit(self, mon);

    RESTORE_ERRNO(nread);
    return nread;
}


/************************************************************************
 * send
 */

ssize_t
send(int fd, const char *buf, int nbytes, int flags)
{
    PRESERVE_ERRNO;
    sys_mon_t *mon;
    int nwrote;
    caddr_t next = (caddr_t)buf;
    int total_written = 0;
    sys_thread_t *self = sysThreadSelf();

    bool_t wasInterrupted = FALSE;
    int interrupt_io = GET_INTERRUPT_IO();

    if (threadBootstrappedP == FALSE) {
	if (!wrappersInited) initializeWrappers();
	return (*systable[SYS_SEND].addr)(fd, buf, nbytes, flags);
    }

    if (fd < 0 || fd >= max_files) {
        Log1(1, "send on bad fd: %d?\n", fd);
	errno = EBADF;
        return -1;
    }

    mon = fdmon[fd];

    sysAssert(mon != NULL);

    check_single_fd(fd);
    sysMonitorEnter(self, mon); 
    FD_ADD_WRITER(fd);

    while(!closed_fd(fd)) {
	nwrote = (*systable[SYS_SEND].addr)(fd, next + total_written,
					    nbytes - total_written, flags);
	if ((nwrote == -1) && (errno != EAGAIN && errno != EINTR)) {
            total_written = -1;
	    break;
        }
	if (nwrote == -1) {
	    /* with errno == EAGAIN or EINTR */
	    if (errno == EAGAIN) {
		if (sysMonitorWait(self, mon, 
				   SYS_TIMEOUT_INFINITY) == SYS_INTRPT) {
		    /* We've cleared interrupted state; remember and reraise */
		    if (interrupt_io) {
			nwrote = SYS_INTRPT;
			break;
		    } else {
			wasInterrupted = TRUE;
		    }
		}
	    }
	    continue;
	}
        total_written += nwrote;
        if (total_written >= nbytes) {
	    break;
	}
    }

    if (wasInterrupted == TRUE) {
	sysThreadInterrupt(sysThreadSelf());    /* Reraise interrupt */
    }

    /* 
     * If there was an error, make sure we return -1.  This covers the
     * case when the code in the while loop above is never executed.
     */
    if (closed_fd(fd)) {
	total_written = -1;
    }

    FD_REMOVE_WRITER(fd);
    /* 
     * If I am the last person to leave and fd has been closed
     * call system close on the fd.
     */
    if (FD_INACTIVE(fd) && closed_fd(fd)) {
	system_close(fd);
	/* set EBADF ? */
    }

    Log2(1, "send for %d bytes finished on fd: %d\n", nwrote, fd);
    sysMonitorExit(self, mon);
    RESTORE_ERRNO(total_written);
    return total_written;
}

#ifdef __SunOS_5_4
ssize_t sendto(int fd, const char *buf, int len,
	       int flags, struct sockaddr *to, int tolen) {
#else
ssize_t sendto(int fd, const char *buf, int len,
	       int flags, const struct sockaddr *to, int tolen) {
#endif /* __SunOS_5_4 */
    PRESERVE_ERRNO;
    sys_mon_t *mon;
    int ret = 0;
    sys_thread_t *self = sysThreadSelf();
 
    bool_t wasInterrupted = FALSE;
    int interrupt_io = GET_INTERRUPT_IO();

    if (threadBootstrappedP == FALSE) {
	if (!wrappersInited) initializeWrappers();
	return (*systable[SYS_SENDTO].addr)(fd, buf, len, flags, to, tolen);
    }

    if (fd < 0 || fd >= max_files || fdmon[fd] == NULL) {
        Log1(1, "writev on bad fd: %d?\n", fd);
	errno = EBADF;
        return -1;
    }

    mon = fdmon[fd];
    sysAssert(mon != NULL);

    check_single_fd(fd);

    sysMonitorEnter(self, mon);
    if (fd >= 0) {
        FD_ADD_WRITER(fd);
	while(!closed_fd(fd)) {
	    ret = (*systable[SYS_SENDTO].addr)(fd, (char *)buf, len, flags, 
					       to, tolen);
	    if ((ret != -1) ||
	        ((errno != EAGAIN) && (errno != EINTR)) || NONBLOCKING_FD(fd)) {
		break;
	    }
	    if (errno == EAGAIN) {
		if (sysMonitorWait(self, mon,
				   SYS_TIMEOUT_INFINITY) == SYS_INTRPT) {
		    if (interrupt_io) {
			ret = SYS_INTRPT;	/* INTERRUPTED */
			break;
		    } else {
			wasInterrupted = TRUE;
		    }
		}
	    }
	}

	if (wasInterrupted == TRUE) {
	    sysThreadInterrupt(sysThreadSelf());    /* Reraise interrupt */
	}

	FD_REMOVE_WRITER(fd);
	/*
	 * If I am the last person to leave and fd has been closed
	 * call system close on the fd.
	 */
	if (FD_INACTIVE(fd) && closed_fd(fd)) {
	    system_close(fd); 
	    /* set EBADF ? */
	}
    }
    Log2(1, "sendto done on fd: %d, ret %d\n", fd, ret);
    sysMonitorExit(self, mon);
    RESTORE_ERRNO(ret);
    return ret;
}


/************************************************************************
 * accept
 */

int
accept(int fd, struct sockaddr *him, int *len) 
{
    PRESERVE_ERRNO;
    sys_mon_t *mon;
    int newfd = -1;
    sys_thread_t *self = sysThreadSelf();

    bool_t wasInterrupted = FALSE;
    int interrupt_io = GET_INTERRUPT_IO();

    if (threadBootstrappedP == FALSE) {
	if (!wrappersInited) initializeWrappers();
	newfd = (*systable[SYS_ACCEPT].addr)(fd, him, len);
	SIMPLE_INITFD(newfd, accept);
	return newfd;
    }

    if (fd < 0 || fd >= max_files || fdmon[fd] == NULL) {
        Log1(1, "readv on bad fd: %d?\n", fd);
	errno = EBADF;
        return -1;
    }

    mon = fdmon[fd];
    sysAssert(mon != NULL);

    check_single_fd(fd);    

    sysMonitorEnter(self, mon);
    FD_ADD_READER(fd);

    while(!closed_fd(fd)) {
	newfd = (*systable[SYS_ACCEPT].addr)(fd, him, len);
	if ((newfd != -1) ||
	    ((errno != EAGAIN) && (errno != EINTR)) || NONBLOCKING_FD(fd)) {
	    break;
	}
	if (errno == EAGAIN) {
	    if (sysMonitorWait(self, mon,
			       SYS_TIMEOUT_INFINITY) == SYS_INTRPT) {
		/* We've cleared interrupted state; remember and reraise */
		if (interrupt_io) {
		    newfd = SYS_INTRPT;
		    break;
		} else {
		    wasInterrupted = TRUE;
		}
	    }
	}
    }
    if (wasInterrupted == TRUE) {
	sysThreadInterrupt(sysThreadSelf());	/* Reraise interrupt */
    }

    FD_REMOVE_READER(fd);
    /* 
     * If I am the last person to leave and fd has been closed
     * call system close on the fd.
     */
    if (FD_INACTIVE(fd) && closed_fd(fd)) {
	system_close(fd); 
	/* set EBADF ? */
    }

    Log2(1, "Accepted socket on %d, new socket fd is %d\n", fd, newfd);
    sysMonitorExit(self, mon);

    SIMPLE_INITFD(newfd, accept);
    RESTORE_ERRNO(newfd);
    return newfd;
}


/************************************************************************
 * getmsg
 *
 * We do not initialize the fd to be nonblocking for getmsg and 
 * putmsg i.e. no calls to check_single_fd - due to a bug
 * in solaris 2.3 or so I am told by jpayne. I need to confirm
 * and revisit later - but for now this works as before - pavani. 
 *
 */

int
getmsg(int fd, struct strbuf *ctlptr, struct strbuf *dataptr, int *flags)
{
    PRESERVE_ERRNO;
    sys_mon_t *mon;
    int ret = -1;

    bool_t wasInterrupted = FALSE;
    int interrupt_io = GET_INTERRUPT_IO();
    sys_thread_t *self = sysThreadSelf();

    if (threadBootstrappedP == FALSE) {
	if (!wrappersInited) initializeWrappers();
	return (*systable[SYS_GETMSG].addr)(fd, ctlptr, dataptr, flags);
    }

    if (fd < 0 || fd >= max_files || fdmon[fd] == NULL) {
        Log1(1, "getmsg on bad fd: %d?\n", fd);
	errno = EBADF;
        return -1;
    }

    mon = fdmon[fd];
    sysAssert(mon != NULL);

    sysMonitorEnter(self, mon);
 
    FD_ADD_READER(fd);

    while(!closed_fd(fd)) {
	ret = (*systable[SYS_GETMSG].addr)(fd, ctlptr, dataptr, flags);
	if ((ret != -1) ||
	    ((errno != EAGAIN) && (errno != EINTR)) || NONBLOCKING_FD(fd)) {
	    break;
	}
	if (errno == EAGAIN) {
	    if (sysMonitorWait(self, mon,
			       SYS_TIMEOUT_INFINITY) == SYS_INTRPT) {
		/* We've cleared interrupted state; remember and reraise */
		if (interrupt_io) {
		    ret = SYS_INTRPT;
		    break;
		} else {
		    wasInterrupted = TRUE;
		}
	    }
	}
    }

    if (wasInterrupted == TRUE) {
	sysThreadInterrupt(sysThreadSelf());    /* Reraise interrupt */
    }

    FD_REMOVE_READER(fd);
    /*
     * If I am the last person to leave and fd has been closed
     * call system close on the fd.
     */
    if (FD_INACTIVE(fd) && closed_fd(fd)) {
	system_close(fd); 
	/* set EBADF ? */
    }

    Log2(1, "getmsg on fd: %d ret=%d\n", fd, ret);
    sysMonitorExit(self, mon);
    RESTORE_ERRNO(ret);
    return ret;
}


/************************************************************************
 * putmsg
 */
int
putmsg(int fd, const struct strbuf *ctlptr,
       const struct strbuf *dataptr, int flags)
{
    PRESERVE_ERRNO;
    sys_mon_t *mon;
    int ret = -1;

    int interrupt_io = GET_INTERRUPT_IO();
    bool_t wasInterrupted = FALSE;
    sys_thread_t *self = sysThreadSelf();

    if (threadBootstrappedP == FALSE) {
	if (!wrappersInited) initializeWrappers();
	return (*systable[SYS_PUTMSG].addr)(fd, ctlptr, dataptr, flags);
    }

    if (fd < 0 || fd >= max_files || fdmon[fd] == NULL) {
        Log1(1, "putmsg on bad fd: %d?\n", fd);
	errno = EBADF;
        return -1;
    }

    mon = fdmon[fd];
    sysAssert(mon != NULL);

    sysMonitorEnter(self, mon);
    FD_ADD_WRITER(fd);

    while(!closed_fd(fd)) {
	ret = (*systable[SYS_PUTMSG].addr)(fd, ctlptr, dataptr, flags);
	if ((ret != -1) ||
	    ((errno != EAGAIN) && (errno != EINTR)) || NONBLOCKING_FD(fd)) {
	    break;
	}
	if (errno == EAGAIN) {
	    if (sysMonitorWait(self, mon,
			       SYS_TIMEOUT_INFINITY) == SYS_INTRPT) {
		/* We've cleared interrupted state; remember and reraise */
		if (interrupt_io) {
		    ret = SYS_INTRPT;
		    break;
		} else {
		    wasInterrupted = TRUE;
		}
	    }
	}
    }

    if (wasInterrupted == TRUE) {
        sysThreadInterrupt(sysThreadSelf());    /* Reraise interrupt */
    }

    FD_REMOVE_WRITER(fd);
    /* 
     * If I am the last person to leave and fd has been closed
     * call system close on the fd.
     */
    if (FD_INACTIVE(fd) && closed_fd(fd)) {
	system_close(fd); 
	/* set EBADF ? */
    }

    Log1(1, "putmsg done on fd: %d\n", fd);
    sysMonitorExit(self, mon);
    RESTORE_ERRNO(ret);
    return ret;
}


/**************************************************************
 * poll
 */

int
poll(struct pollfd *fds, unsigned long nfds, int timeout)
{
    PRESERVE_ERRNO;
    int	ret = -1;
    timeval_t time_limit;
    int	poll_timeout = 0;
    int	i, fd, bad = 0;

    bool_t wasInterrupted = FALSE;
    int interrupt_io = GET_INTERRUPT_IO();
    sys_thread_t *self = sysThreadSelf();

    if (threadBootstrappedP == FALSE) {
	if (!wrappersInited) initializeWrappers();
	return (*systable[SYS_POLL].addr)(fds, nfds, timeout);
    }

    for (i = nfds; --i >= 0; ) {
	if ((fd = fds[i].fd) < max_files && fd >= 0 && fdmon[fd] != NULL) {
	    check_single_fd(fd);
	} else {
	    fds[i].revents = POLLNVAL;
	    bad++;
	}
    }

    if (bad) {
	return bad;
    }

    if (timeout > 0) {
	timeval_t tv;
	tv.tv_sec = timeout / 1000;		/* seconds */
	tv.tv_usec = (timeout % 1000) * 1000;	/* microseconds */
	time_limit = currentTime();
	timerAdd(&time_limit, (timeval_t *) &tv);
    }
    
    sysMonitorEnter(self, asyncMon(SYS_ASYNC_MON_IO));
    while (1) {
	timeval_t delay;
	/* 
	 * Check to see if any fd's have been marked closed.
	 * They haven't been closed as far as the system is
	 * concerned, but we want to mimic the correct behavior
	 * of the real poll in this case, which is:
	 * -set the revents for the closed fd's to POLLNVAL
	 * -don't touch the events
	 * -but don't set errno to anything
	 * -return the number of fd's that had something happen
	 * (e.g, the number of "closed" fd's we encounter)
	 */
	for (i = nfds; --i >= 0; ) {
	    if (closed_fd(fds[i].fd)) {
		fds[i].revents = POLLNVAL;
		bad++;
	    }
	}
	    
	if (bad) {
	    ret = bad;
	    break;
	}
	errno = 0;

	ret = (*systable[SYS_POLL].addr)(fds, nfds, poll_timeout);

	if (ret == -1) {
	    Log1(1, "poll => -1; errno = %d\n", errno);
	}

	/* If the call completed or the timeout is 0, return immediately */
	if (ret != 0 || timeout == 0) {
	    break;
	}

	if (timeout > 0) {
	    /* Would have blocked, see whether timer has expired */
	    timeval_t now = currentTime();
	    if (timerCmp(&now, &time_limit, >) || timerEq(&now, &time_limit)) {
		ret = 0;
		break;
	    }
	    delay = time_limit;
	    timerSub(&delay, &now);
	    timeout = (delay.tv_sec * 1000) + (delay.tv_usec / 1000);
	}

	/* The timeout is nonzero; wait forever if it's negative */
	if (sysMonitorWait(self, asyncMon(SYS_ASYNC_MON_IO),
		timeout < 0 ? SYS_TIMEOUT_INFINITY : timeout) == SYS_INTRPT) {
	    /* We've cleared interrupted state; remember and reraise */
	    if (interrupt_io) {
		ret = SYS_INTRPT;
	    } else {
		wasInterrupted = TRUE;
	    }
	}
    }

    if (wasInterrupted == TRUE) {
        sysThreadInterrupt(sysThreadSelf());    /* Reraise interrupt */
    }

    sysMonitorExit(self, asyncMon(SYS_ASYNC_MON_IO));
    RESTORE_ERRNO(ret);
    return ret;
}

/**************************************************************
 * select 
 *
 * Select system call in a multi-threaded environment.
 *
 * This works by affecting a poll.  If _select returns
 * non-negative, then we return that value.  Otherwise, we block
 * until *any* IO is ready, and then try again.  If we still come
 * up empty then we block for a smaller amount of time this time.
 * If we reach the time limit, then we just return 0, as select
 * normally would.
 */
int
select(int width, fd_set *readfds, fd_set *writefds,
       fd_set *exceptfds, struct timeval *tv)
{
    PRESERVE_ERRNO;
    int nfds = -1;
    struct timeval poll;
    timeval_t time_limit;
    fd_set reads, writes, excepts;
    int	i;

    bool_t wasInterrupted = FALSE;
    int interrupt_io = GET_INTERRUPT_IO();
    sys_thread_t *self = sysThreadSelf();

    if (threadBootstrappedP == FALSE) {
	if (!wrappersInited) initializeWrappers();
	return (*systable[SYS_SELECT].addr)(width,
					    readfds, writefds, exceptfds, tv);
    }

    for (i = 0; i < width; i++) {
	if (i >= OPEN_MAX) {
	    break;
	}
	if ((readfds != 0 && FD_ISSET(i, readfds))
	    || (writefds != 0 && FD_ISSET(i, writefds))
	    || (exceptfds != 0 && FD_ISSET(i, exceptfds))) {
	    if (i >= 0 && i < max_files && fdmon[i] != NULL) {
		check_single_fd(i);
	    }
	}
    }

    poll.tv_sec = poll.tv_usec = 0;
    if (tv != 0) {
	time_limit = currentTime();
	timerAdd(&time_limit, (timeval_t*)tv);
    }
    
    if (readfds != 0) {
	COPY_SET(&reads, readfds, width);
    }
    if (writefds != 0) {
	COPY_SET(&writes, writefds, width);
    }
    if (exceptfds != 0) {
	COPY_SET(&excepts, exceptfds, width);
    }

    sysMonitorEnter(self, asyncMon(SYS_ASYNC_MON_IO));
    while (1) {
	timeval_t delay;
	int timeout = SYS_TIMEOUT_INFINITY;

	nfds = (*systable[SYS_SELECT].addr)(width, readfds, writefds,
					    exceptfds, &poll);

	if (nfds != 0) {
	    break;
	}
	if (tv != 0) {
	    timeval_t now = currentTime();

	    if (timerCmp(&now, &time_limit, >) || timerEq(&now, &time_limit)) {
		nfds = 0;
		break;
	    }
	    delay = time_limit;
	    timerSub(&delay, &now);
	    timeout = (delay.tv_sec * 1000) + (delay.tv_usec / 1000);
	}
	if (sysMonitorWait(self, asyncMon(SYS_ASYNC_MON_IO), 
			   timeout) == SYS_INTRPT) {
	    /* We've cleared interrupted state; remember and reraise */
	    if (interrupt_io) {
		nfds = SYS_INTRPT;
	    } else {
		wasInterrupted = TRUE;
	    }
	}
	if (readfds != 0) {
	    COPY_SET(readfds, &reads, width);
	}
	if (writefds != 0) {
	    COPY_SET(writefds, &writes, width);
	}
	if (exceptfds != 0) {
	    COPY_SET(exceptfds, &excepts, width);
	}
    }

    if (wasInterrupted == TRUE) {
	sysThreadInterrupt(sysThreadSelf());    /* Reraise interrupt */
    }

    sysMonitorExit(self, asyncMon(SYS_ASYNC_MON_IO));
    RESTORE_ERRNO(nfds);
    return nfds;
}


/**************************************************************
 * pipe
 *
 * Although we set the SA_RESTART flag on all our signal handlers we
 * still see open() occasionally return with errno == EINTR, at
 * least on SunOS 2.3.  This routine is an attempt to get aorund
 * this bug by making sure open does not return until it either
 * succeeds or errno != EINTR.
 */
int
pipe(int fds[])
{
    PRESERVE_ERRNO;
    int sys_ret = 0;
    sys_thread_t *self = sysThreadSelf();

    if (threadBootstrappedP == FALSE) {
	if (!wrappersInited) initializeWrappers();
	sys_ret = (*systable[SYS_PIPE].addr)(fds);
	if (sys_ret != -1) {
	    int ret;
	    Log2(1, "Opened pipe fd0: %d fd1: %d\n", fds[0], fds[1]);
	    ret = initialize_monitors(fds[0]);
	    ret = (ret == B_FALSE ? ret : initialize_monitors(fds[1]));
	    if (ret == B_FALSE) {			/* Out of memory */
		(*systable[SYS_CLOSE].addr)(fds[0]);
		(*systable[SYS_CLOSE].addr)(fds[1]);
		errno = ENOMEM;
		sys_ret = -1;
	    }
	} else {
	    Log1(1, "pipe error: %d\n", errno);
	}
	return sys_ret;
    }

    IO_LOCK(self);

    while(1) {
	/*
	 * As unbelievable as it may seem, you cannot simply simulate
	 * the pipe() system call with a simple
	 * syscall(SYS_pipe,...).  SYS_pipe returns two values in
	 * %o0 and %o1, effectively returning a "long long", but the
	 * prototype for syscall only returns an int!
	 */
	sys_ret = (*systable[SYS_PIPE].addr)(fds);
	if ((sys_ret != -1) ||
	    ((errno != EAGAIN) && (errno != EINTR))) {
	    break;      
	}
    }
 
    if (sys_ret != -1) {
	int ret;
	Log2(1, "Opened pipe fd0: %d fd1: %d\n", fds[0], fds[1]);
	ret = initialize_monitors(fds[0]);
	ret = (ret == B_FALSE ? ret : initialize_monitors(fds[1]));
	if (ret == B_FALSE) {			/* Out of memory */
	    (*systable[SYS_CLOSE].addr)(fds[0]);
	    (*systable[SYS_CLOSE].addr)(fds[1]);
	    errno = ENOMEM;
	    sys_ret = -1;
	}
    } else {
	Log1(1, "pipe error: %d\n", errno);
    }

    IO_UNLOCK(self);
    RESTORE_ERRNO(sys_ret);
    return sys_ret;
}


/**************************************************************
 * fcntl 
 *
 * The following wrapper is necessary because nis+ library calls dup,
 * as a means of getting new fds - and we need to initialize monitors
 * for the newly opened fds before they can be used.
 * Note dup2 call uses fcntl internally - so the following wrapper
 * takes care of duplicating fds through fcntl or dup2
 */
int
fcntl(int fd, int cmd, ...)
{
    PRESERVE_ERRNO;
    int ret = -1;
    va_list ap;
    int arg;  /* optional third argument */
    sys_thread_t *self = sysThreadSelf();

    va_start(ap, cmd);
    arg = va_arg(ap, int);
    va_end(ap);

    if (threadBootstrappedP == FALSE) {
	if (!wrappersInited) initializeWrappers();
	ret = (*systable[SYS_FCNTL].addr)(fd, cmd, arg);
	if (cmd == F_DUPFD) {
	    SIMPLE_INITFD(ret, F_DUPFD);
	}
	return ret;
    }

    IO_LOCK(self);

    if (cmd == F_SETFL && fd_flags[fd] & FD_NBINIT) {
	/*
	 * We don't want the caller to turn this fd into a non-blocking
	 * fd so we preserve the fact that the fd is supposed
	 * to be non-blocking by adding O_NONBLOCK to the flags.
	 */
	arg |= O_NONBLOCK;
    } 

    while (1) {
	ret = (*systable[SYS_FCNTL].addr)(fd, cmd, arg);
	if ((ret != -1) ||  
	    ((errno != EAGAIN) && (errno != EINTR)) || NONBLOCKING_FD(fd)) {
	      break;
	}
    }

    if (cmd == F_DUPFD) {
	SIMPLE_INITFD(ret, F_DUPFD);
    } else if (cmd == F_SETFL) {
	/* update our notion of the semantics the caller expects */
	if (arg & O_NONBLOCK) {
	    /* we've become non-blocking */
	    fd_flags[fd] |= FD_NONBLOCKING;
	} else if (fd_flags[fd] & FD_NONBLOCKING) {
	    /* we used to be non-blocking but now we aren't */
	    fd_flags[fd] &= ~FD_NONBLOCKING;
	}
    } else if (cmd == F_GETFL) {
	/*
	 * we need to report back the user's notion of the non-blocking
	 * state and not ours.  
	 */
	if (!(fd_flags[fd] & FD_NONBLOCKING))
	    ret &= ~O_NONBLOCK;
    }

    IO_UNLOCK(self);
    RESTORE_ERRNO(ret);
    return ret;
}


/**************************************************************
 * dup
 */

int
dup(int fd)
{
    PRESERVE_ERRNO;
    int newfd = -1;
    sys_thread_t *self = sysThreadSelf();

    if (threadBootstrappedP == FALSE) {
	if (!wrappersInited) initializeWrappers();
	newfd = (*systable[SYS_DUP].addr)(fd);
	SIMPLE_INITFD(newfd, dup);
	return newfd;
    }

    IO_LOCK(self);

    while (1) {
	newfd = (*systable[SYS_DUP].addr)(fd);
	if ((newfd != -1) ||
	    ((errno != EAGAIN) && (errno != EINTR)) || NONBLOCKING_FD(fd)) {
	    break;      
	}
    }
    SIMPLE_INITFD(newfd, dup);

    IO_UNLOCK(self);
    RESTORE_ERRNO(newfd);
    return newfd;
}


/**************************************************************
 * creat
 */
int
creat(const char *path, mode_t mode)
{
    PRESERVE_ERRNO;
    int newfd = -1;
    sys_thread_t *self = sysThreadSelf();

    if (threadBootstrappedP == FALSE) {
	if (!wrappersInited) initializeWrappers();
	newfd = (*systable[SYS_CREAT].addr)(path, mode);
	SIMPLE_INITFD(newfd, creat);
	return newfd;
    }

    IO_LOCK(self);

    while (1) {
	newfd = (*systable[SYS_CREAT].addr)(path, mode);
	if ((newfd != -1) ||
	    ((errno != EAGAIN) && (errno != EINTR))) {
	    break;      
	}
    }
    if (newfd != -1) {
	Log1(1, "Creat fd: %d \n", newfd);
	if (initialize_monitors(newfd) == B_FALSE) {	/* Out of memory */
	    (*systable[SYS_CLOSE].addr)(newfd);		/* Close new fd */
	    errno = ENOMEM;
	    newfd = -1;
	}
    } else {
	Log1(1, "Creat error: %d\n", errno);
    }

    IO_UNLOCK(self);
    RESTORE_ERRNO(newfd);
    return newfd;
}

int
socket(int domain, int type, int protocol)
{
    PRESERVE_ERRNO;
    int newfd = -1;
    sys_thread_t *self = sysThreadSelf();

    if (threadBootstrappedP == FALSE) {
	if (!wrappersInited) initializeWrappers();
	newfd = (*systable[SYS_SOCKET].addr)(domain, type, protocol);
	SIMPLE_INITFD(newfd, socket);
	return newfd;
    }

    IO_LOCK(self);

    while (1) {
	newfd = (*systable[SYS_SOCKET].addr)(domain, type, protocol);
	if ((newfd != -1) ||
	    ((errno != EAGAIN) && (errno != EINTR))) {
	    break;      
	}
    }
    SIMPLE_INITFD(newfd, socket);

    IO_UNLOCK(self);
    RESTORE_ERRNO(newfd);
    return newfd;
}

int 
connect(int fd, struct sockaddr *him, int len) {
    PRESERVE_ERRNO;
    int result = -1;
    int delay = 50;	/* sleep 50 ms and try again */
    
#ifdef USE_SELECT
    fd_set write_fds;
    struct timeval  poll;
#else
    struct pollfd pfd;
#endif
    sys_mon_t *mon;
    sys_thread_t *self = sysThreadSelf();

    int interrupt_io = GET_INTERRUPT_IO();
    bool_t wasInterrupted = FALSE;

    if (fd < 0 || fd >= max_files || fdmon[fd] == NULL) {
        Log1(1, "connect operation on closed fd: %d?\n", fd);
	errno = EBADF;
        return result;
    }
    
    check_single_fd(fd);
    mon = fdmon[fd];
    sysAssert(mon != NULL);

    sysMonitorEnter(self, mon);
#ifdef USE_SELECT
    FD_ZERO(&write_fds);
    FD_SET(fd, &write_fds);
    poll.tv_sec = poll.tv_usec = 0;
#else
    pfd.fd = fd;
    pfd.events = POLLOUT;
#endif
    result = (*systable[SYS_CONNECT].addr)(fd, him, len);
    while (result < 0) {
	if (errno == EINPROGRESS) {
	    int cnt = 0;
	    /* we may want to allow for getting interrupted here */
#ifdef USE_SELECT
	    cnt = (*systable[SYS_SELECT].addr)(fd + 1, NULL, &write_fds,
					       NULL, &poll);
	    FD_SET(fd, &write_fds);
#else
	    cnt = poll(&pfd, 1, 0);
#endif
	    if (cnt == 1) {
		int bytes;

		if (ioctl(fd, FIONREAD, &bytes) == -1) {
		    if (errno == EPIPE) {
			errno = ECONNREFUSED;
		    }
		    /*
		     * If this fails with EPIPE, then we assume
		     * connection is refused - who knows what's
		     * right?  Not I.
		     */
		    sysAssert(result < 0);
		    break;
		}
		result = 0;
		break;
	    }
	    if (sysMonitorWait(self, mon, delay) == SYS_INTRPT) {
		if (interrupt_io) {
		    result = SYS_INTRPT;
		} else {
		    wasInterrupted = TRUE;
		}
		break;
	    }
	} else {
	    break;
	}
    } /* end of while */

    if (wasInterrupted == TRUE) {
	sysThreadInterrupt(sysThreadSelf());    /* Reraise interrupt */
    }
    sysMonitorExit(self, mon);
    RESTORE_ERRNO(result);

    return result;
}


/*
 * This is wrapper for _filbuf which is used in the implementation of
 * things like scanf and fread.  We have to wrap it on Solaris because of
 * the 2 level wrapping of library interface (i.e. read and _read).
 * Maybe we should be wrapping at the _read level instead?
 */

int
__filbuf(FILE * fp)
{
    PRESERVE_ERRNO;
    sys_mon_t *mon;
    int nread = -1;
    int fd = fp->_file;
    sys_thread_t *self = sysThreadSelf();
    int wasNonblock;

    bool_t wasInterrupted = FALSE;
    int interrupt_io = GET_INTERRUPT_IO();

    if (threadBootstrappedP == FALSE) {
	if (!wrappersInited) initializeWrappers();
	return (*systable[SYS_FILBUF].addr)(fp);
    }

    if (fd < 0 || fd >= max_files) {
	Log1(1, "read on bad fd: %d?\n", fd);
 	errno = EBADF;
	return -1;
    }

    /*
     * It's fairly likely that we haven't seen this fd before since we
     * don't wrap fopen and on Solaris fopen seems to go directly to
     * _open.
     */
    if (fdmon[fd] == NULL) {
	IO_LOCK(self);
	if (fdmon[fd] == NULL) {
	    if (initialize_monitors(fd) == B_FALSE) {
		VM_CALL(panic)("out of memory");
		return -1;
	    }
	}
	IO_UNLOCK(self);
    }

    mon = fdmon[fd];
    sysMonitorEnter(self, mon);
    Log1(1, "_filbuf on fd: %d\n", fd);
    wasNonblock = (fd_flags[fd] & FD_NBINIT);
    if (wasNonblock) {
	nonblock_io(fd, IO_BLOCK);
    }

    nread = (*systable[SYS_FILBUF].addr)(fp);
    sysAssert(nread != -1 || errno != EAGAIN);

    if (wasNonblock) {
	nonblock_io(fd, IO_DONTBLOCK);
    }

    sysMonitorExit(self, mon);
    RESTORE_ERRNO(nread);
    return nread;
}
