/*
 * @(#)monitor_md.c	1.72 98/08/20
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Monitor implementation for Green threads
 */

#include <limits.h>   /* for INT_MAX */
#include <string.h>

#include "hpi_impl.h"

#include "interrupt.h"
#include "itimer_md.h"
#include "schedule.h"
#include "queue.h"
#include "clock.h"

#include "threads_md.h"
#include "monitor_md.h"

static int monitorUndoInversion(sys_mon_t *, sys_thread_t *);
#ifdef DEBUG_PRIORITY
void checkInvariants(const char *mesg, int resched);
#endif /* DEBUG_PRIORITY */

/**
 ** Queue maniplation code
 **/

void
queueInsert(gt_queue_t *q, sys_thread_t *waiter)
{
    sys_thread_t *t;
    sys_thread_t *p;

    t = *q;
    p = 0;
    while (t && (t->priority >= waiter->priority)) {
	p = t;
	t = t->waitq;
    }

    if (p) {
	p->waitq = waiter;
    } else {
	*q = waiter;
    }

    waiter->waitq = t;
    return;
}

int
queueRemove(gt_queue_t *q, sys_thread_t *waiter)
{
    sys_thread_t *t;
    sys_thread_t *p;

    t = *q;
    p = 0;
    while (t && t != waiter) {
	p = t;
	t = t->waitq;
    }

    if (t) {
	if (p) {
	    p->waitq = t->waitq;
	} else {
	    *q = t->waitq;
	}
	return 1;
    } else {
	return 0;
    }
}


/*
 * This function signals a specific queue.  If the queue is the monitor's wait
 * queue, the thread is put on the runnable queue.  If the queue is not
 * the monitor's wait queue, it is put on the wait queue.
 */
int
queueSignal(sys_mon_t *mid, gt_queue_t *q)
{
    sys_thread_t *waiter;
    int need_to_reschedule = 0;
    sys_thread_t *self = greenThreadSelf();

    Log3(4, "queueSignal: mid 0x%x q 0x%x sys_thread_t 0x%x\n", mid, q, self);
    /*CONSTCOND*/
    while (1) {
	if ((waiter = queueRemoveHead(q, waiter)) != 0) {
	    if (q == &mid->monitor_waitq) {
		if (waiter->pending_suspend) {
		    waiter->state = MONITOR_SUSPENDED;
		    queueInsert(&mid->suspend_waitq, waiter);
		    waiter->pending_suspend = 0;
		    Log1(4, "queueSignal: making waiter 0x%x suspended\n",
			 waiter);
		    continue;
		} else {
		    Log1(4, "queueSignal: making waiter 0x%x runnable\n",
			 waiter);
		    mid->monitor_owner = waiter;
		    mid->entry_count = waiter->monitor_entry_count;
		    waiter->state = RUNNABLE;
		    waiter->mon_wait = 0;
		    waiter->monitor_entry_count = 0;
		    queueInsert(&runnable_queue, waiter);
		    if (mid->monitor_waitq != SYS_THREAD_NULL) {
			(void) monitorApplyInversion(mid);
		    }
		    if (runnable_queue->priority > self->priority) {
			need_to_reschedule = 1;
		    }
		}
	    } else {
		sys_thread_t *owner = mid->monitor_owner;
	        Log(4, "queueSignal: CONDVAR_WAIT -> MONITOR_WAIT\n");
		waiter->state = MONITOR_WAIT;
		queueInsert(&mid->monitor_waitq, waiter);
		
		/* if nothing else in the waitq then wake it up */
		if (owner == 0) {
		    q = &mid->monitor_waitq;
		    continue;
		} else {
		    (void) monitorApplyInversion(mid);
		}
	    }
	} else if (q == &mid->monitor_waitq) {
	
	    /* sysAssert below not valid given how monitor_cleanup works!
	    sysAssert(mid->monitor_owner == self
		    || mid->monitor_owner == SYS_THREAD_NULL); */
	    Log(4, "queueSignal: no waiter, set no monitor owner\n");
	    mid->monitor_owner = SYS_THREAD_NULL;
	    mid->entry_count = 0;
	}
	break;
    }
    return need_to_reschedule;
}

/*
 * signal all waiters on a queue.  We assume that the queue is the CV,
 * as that is the only queue this operation is permitted on.
 */
void
queueBroadcast(sys_mon_t *mid)
{
    sys_thread_t *waiter;

    while ((waiter = queueRemoveHead(&mid->condvar_waitq, waiter)) != 0) {
        Log1(4, "queueBroadcast: waiter 0x%x CONDVAR_WAIT -> MONITOR_WAIT\n",
	    waiter);
	waiter->state = MONITOR_WAIT;
	queueInsert(&mid->monitor_waitq, waiter);
    }
}


/*
 * Wait on a specific queue until signalled
 */
void
queueWait(sys_mon_t *mid, gt_queue_t *q)
{
    sys_thread_t *self = greenThreadSelf();

    Log3(4, "queueWait: sys_thread_t 0x%x q 0x%x mid 0x%x\n",
    	self, q, mid);
    sysAssert(SCHED_LOCKED());
    queueInsert(q, self);
    self->mon_wait = (sys_mon_t *) mid;

    if (q == &mid->monitor_waitq) {
        Log(4, "queueWait: MONITOR_WAIT\n");
	self->state = MONITOR_WAIT;
	(void) monitorApplyInversion(mid);
    } else {
        Log(4, "queueWait: CONDVAR_WAIT\n");
	self->state = CONDVAR_WAIT;
	self->monitor_entry_count = mid->entry_count;
	if (mid->flags & SYS_MON_INVERTED_PRIORITY) {
	    (void) monitorUndoInversion(mid, self);
	}
	(void) queueSignal(mid, &mid->monitor_waitq);
    }

#ifdef DEBUG_PRIORITY
    checkInvariants("queueWait", 1);
#endif /* DEBUG_PRIORITY */
    yieldContext(CONTEXT(self));
    return;
}

/*
 * Priority inversion
 */

void
monitorAddInversion(sys_mon_t *mid)
{
    sys_thread_t *monitor_owner = mid->monitor_owner;
    sysAssert(monitor_owner != SYS_THREAD_NULL);
    sysAssert(mid->monitor_waitq != SYS_THREAD_NULL);
    /* Temporarily removed: sysAssert(SCHED_LOCKED()); */
    sysAssert(monitor_owner != mid->monitor_waitq);
    sysAssert((mid->flags & SYS_MON_INVERTED_PRIORITY) == 0);

    /*
     * Save this monitor on our sorted list so that we can
     * undo the inversion correctly for multiple monitors.
     * Use the priority of the highest priority thread
     * waiting on this monitor.
     */
    mid->flags |= SYS_MON_INVERTED_PRIORITY;
    {
	sys_mon_t **mm = &monitor_owner->inversion_queue;
	sys_mon_t *m = *mm;
	int priority = mid->monitor_waitq->priority;

	while (m != NULL && priority < m->monitor_waitq->priority) {
	    mm = &m->next_inversion;
	    m = *mm;
	}
	mid->next_inversion = m;
	*mm = mid;
    }
}

int
monitorApplyInversion(sys_mon_t *mid)
{
    sys_thread_t *monitor_owner = mid->monitor_owner;
    int monitorOwnerPriority = monitor_owner->priority;
    int priority;
    /*
     * First inversion for this thread.  Save the real priority.
     */
    if (monitor_owner->inversion_queue == NULL) {
	monitor_owner->oldPriority = monitorOwnerPriority;
    }
    if ((mid->flags & SYS_MON_INVERTED_PRIORITY) != 0) {
	monitorRemoveInversion(mid, monitor_owner);
    }
    monitorAddInversion(mid);
    priority = INVERSION_PRIORITY(monitor_owner);
    Log4(3, "monitorApplyInversion: sys_thread_t = 0x%x"
	 "\n\tsys_mon_t 0x%x priority %d -> %d\n", monitor_owner,
	 mid, monitorOwnerPriority, priority);
    if (priority > monitorOwnerPriority) {
	return threadSetSchedulingPriority(monitor_owner, priority);
    } else {
	/* Should not lower priority */
	sysAssert(priority == monitorOwnerPriority);
	return 0;
    }
}

void
monitorRemoveInversion(sys_mon_t *mid, sys_thread_t *t)
{
    sysAssert((mid->flags & SYS_MON_INVERTED_PRIORITY) != 0);
    {
	sys_mon_t **mm = &t->inversion_queue;
	sys_mon_t *m;

	while ((m = *mm) != NULL) {
	    if (mid == m) {
		/*
		 * This is the monitor whose inversion is going away.
		 * Remove it from the queue.
		 */
		*mm = m->next_inversion;
		break;
	    } else {
		mm = &m->next_inversion;
	    }
	}
	sysAssert(m == mid);

	mid->flags &= ~SYS_MON_INVERTED_PRIORITY;
    }
}

static int
monitorUndoInversion(sys_mon_t *mid, sys_thread_t *tid)
{
    int oldPriority = tid->priority;
    int priority;

    sysAssert(mid->flags & SYS_MON_INVERTED_PRIORITY);
    monitorRemoveInversion(mid, tid);
    priority = (tid->inversion_queue != NULL) ? INVERSION_PRIORITY(tid)
        : tid->oldPriority;
    Log4(3, "monitorUndoInversion: sys_thread_t = 0x%x,\n\t"
         "sys_mon_t 0x%x " "priority %d -> %d\n", tid, mid,
         oldPriority, priority);
    if (priority < oldPriority) {
        return threadSetSchedulingPriority(tid, priority);
    } else {
        sysAssert(priority == oldPriority);
        return 0;
    }
}

/*
 * Operations on monitors
 */

/*
 * Return the size of the lib-dependent portion of monitors.  This
 * is done this way so that monitors can be all of one piece,
 * without paying the penalty of an extra level of indirection on
 * each sys_mon reference.  This is not how it is done for threads
 * and it might be a good idea to use a pointer the same way that
 * threads do.
 */
size_t
sysMonitorSizeof()
{
    return sizeof(struct sys_mon);
}

/*
 * Perform any system-dependent initialization of monitors.
 */
int
sysMonitorInit(sys_mon_t *mid)
{
    sysAssert(mid != SYS_MID_NULL);
    memset(mid, 0, sizeof(sys_mon_t));
    return SYS_OK;
}

/*
 * Free any system-dependent resources held by monitors.  There is
 * nothing to do for Green threads on Solaris.
 */
int
sysMonitorDestroy(sys_mon_t *mid)
{
    sysAssert(mid != SYS_MID_NULL);

    return SYS_OK;
}

/*
 * Enter a monitor
 *
 * A subtlety of the Solaris/Green threads package is the interaction
 * between monitor entry and scheduler locking for single-threaded execution.
 * If care is not taken in the low-level Java thread package support, race
 * conditions can unpredictably cause other threads to run at precisely the
 * times that you may expect to be single threaded.  This is often disasterous.
 *
 * The simple rule is that monitors must not be (newly) entered when the
 * scheduler is locked.  If that were to be attempted, there is some chance
 * that another thread might be holding the monitor.  In this case the running
 * thread waits for the monitor, which means yielding, the alternative being
 * deadlock.  So another thread may run.
 *
 * In the Solaris/Green threads, monitors are entered in two different
 * ways.  They are explicitly entered via monitorEnter(), sysMonitorEnter()
 * etc.  They are also implicitly entered when certain libc functions are
 * called (see the comment in synch.c).  If entering a monitor explicitly
 * while scheduling is locked can't be avoided in all contexts (e.g. when
 * enumerating threads or monitors), it is safe to acquire the necessary
 * monitor prior to locking the scheduler, so we will be guaranteed not to
 * wait once the scheduler is locked.  This doesn't really work for implicit
 * monitor use, where we simply have to avoid making certain libc calls when
 * scheduling is locked.
 *
 * To police monitor use, assertions may be placed in sysMonitorEnter()
 * and synch.c to abort if a monitor is entered when scheduling is locked.
 * Unfortunately, those in sysMonitorEnter() break the current implemen-
 * tation of ^\, so are commented out for now.
 */
int
sysMonitorEnter(sys_thread_t *self, sys_mon_t *mid)
{
    sysAssert(mid != SYS_MID_NULL);
    SCHED_LOCK();
    if (mid->monitor_owner) {
	if (mid->monitor_owner != self) {
	    /* sysAssert(_scheduling_lock == 1); */
	    if (profiler_on) {
	        goto contended;
	    } else {
                self->mon_enter = mid;
	        queueWait(mid, &mid->monitor_waitq);
                self->mon_enter = NULL;
	    }
	}
    } else {
	/* sysAssert(_scheduling_lock == 1); */
	mid->monitor_owner = self;
    }
    mid->entry_count++;
    SCHED_UNLOCK();
    return SYS_OK;

 contended:
    SCHED_UNLOCK();
    VM_CALL(monitorContendedEnter)(self, mid);
    SCHED_LOCK();
    if (mid->monitor_owner) {
	if (mid->monitor_owner != self) {
	    /* sysAssert(_scheduling_lock == 1); */
            self->mon_enter = mid;
	    queueWait(mid, &mid->monitor_waitq);
            self->mon_enter = NULL;
	}
    } else {
	/* sysAssert(_scheduling_lock == 1); */
	mid->monitor_owner = self;
    }
    mid->entry_count++;
    SCHED_UNLOCK();
    VM_CALL(monitorContendedEntered)(self, mid);
    return SYS_OK;
}

/*
 * Return true if we currently own this monitor (and threads have been
 * initialized.
 */
bool_t
sysMonitorEntered(sys_thread_t *self, sys_mon_t *mid)
{
    sysAssert(mid != SYS_MID_NULL);

    /* We can only have locked monitors if threads have been initialized */
    return mid->monitor_owner == self;
}

int
sysMonitorExit(sys_thread_t *self, sys_mon_t *mid)
{
    sysAssert(mid != SYS_MID_NULL);
    SCHED_LOCK();

    /*
     * Note: From Java we can't express an unmatched exit, only from C
     * code, so in principle we could turn this test into an assertion
     * and have this function return void.
     */
    if (mid->monitor_owner != self) {
#ifdef DEBUG_PRIORITY
	checkInvariants("sysMonitorExit", 0);
#endif /* DEBUG_PRIORITY */
	SCHED_UNLOCK();
	return SYS_ERR;
    } else {
        /* Exiting the monitor, so don't care about asynch notifications */
	mid->flags &= ~SYS_MON_STICKY_NOTIFICATION;
        --mid->entry_count;
	sysAssert(mid->entry_count >= 0);
        if (mid->entry_count == 0) {
	    if (mid->flags & SYS_MON_INVERTED_PRIORITY) {
        	(void) monitorUndoInversion(mid, self);
	    }
	    if (!mid->monitor_waitq && !mid->condvar_waitq &&
		!mid->suspend_waitq) {
		/* Nobody on queues, owner exiting, reinitialize sys_mon_t */
	        mid->monitor_owner = SYS_THREAD_NULL;
		/* Clear all flags except SYS_MON_HAS_HANDLER */
		mid->flags &= SYS_MON_HAS_HANDLER;
	    } else {
	        bool_t contended = (mid->monitor_waitq != NULL);
	        if (queueSignal(mid, &mid->monitor_waitq)) {
		    sysAssert(mid->monitor_owner != SYS_THREAD_NULL);
#ifdef DEBUG_PRIORITY
		    checkInvariants("sysMonitorExit queueSignal", 1);
#endif /* DEBUG_PRIORITY */
		    YIELD();
		    SCHED_UNLOCK();
		    if (contended && profiler_on) {
		        VM_CALL(monitorContendedExit)(self, mid);
		    }
		    return SYS_OK;
		}
	    }
	}
    }
    SCHED_UNLOCK();
    return SYS_OK;
}

int
sysMonitorNotify(sys_thread_t *self, sys_mon_t *mid)
{
    int ret = SYS_OK;

    sysAssert(mid != SYS_MID_NULL);
    SCHED_LOCK();
    if (self != mid->monitor_owner) {
	ret = SYS_ERR;
    } else {
	(void) queueSignal(mid, &mid->condvar_waitq);
    }
    
    /* don't need to signal the waitq since current thread has monitor */
#ifdef DEBUG_PRIORITY
    checkInvariants("sysMonitorNotify", 0);
#endif /* DEBUG_PRIORITY */
    SCHED_UNLOCK();
    return ret;
}

int
monitorBroadcast(sys_mon_t *mid)
{
    /* Now notify all of the waiters. */
    queueBroadcast(mid);

    /* If the monitor isn't locked, give it to the first waiter */
    return mid->monitor_waitq != NULL &&
	((mid->monitor_owner == NULL)
	? queueSignal(mid, &mid->monitor_waitq)
	: monitorApplyInversion(mid));
}

int
sysMonitorNotifyAll(sys_thread_t *self, sys_mon_t *mid)
{
    int ret = SYS_OK;

    sysAssert(mid != SYS_MID_NULL);
    SCHED_LOCK();
    if (self != mid->monitor_owner) {
	ret = SYS_ERR;
    } else {
        monitorBroadcast(mid);
    }

#ifdef DEBUG_PRIORITY
    checkInvariants("sysMonitorNotifyAll", 0);
#endif /* DEBUG_PRIORITY */
    SCHED_UNLOCK();
    return ret;
}

int
sysMonitorWait(sys_thread_t *self, sys_mon_t *mid, jlong millis)
{
    int ret = SYS_OK;
    timeval_t tm, timeout;
    int infinity;

    sysAssert(mid != SYS_MID_NULL);
    SCHED_LOCK();
    if (self != mid->monitor_owner) {
	ret = SYS_ERR;
	goto unlock;
    }
    if (self->interrupted) {
        self->interrupted = FALSE;
	ret = SYS_INTRPT;
	goto unlock;
    }

    sysAssert(self->state == RUNNABLE);
    sysAssert(self->mon_wait == 0);
    sysAssert(self->monitor_entry_count == 0);
    if ((mid->flags & SYS_MON_STICKY_NOTIFICATION) != 0) {
	mid->flags &= ~SYS_MON_STICKY_NOTIFICATION;
	goto unlock;
    }

    infinity = (millis == SYS_TIMEOUT_INFINITY) || (millis >= (jlong)INT_MAX);
               
    if (!infinity) {
        threadSetTimeout(&tm, millis/1000, (millis % 1000) * 1000);
	timeout = currentTime();
	timerAdd(&timeout, &tm);
	registerAlarm(self, timeout);
    }

    queueWait(mid, &mid->condvar_waitq);

    sysAssert(mid->monitor_owner == self);
    sysAssert(SCHED_LOCKED());

    if (!infinity) {
	cleanupPendingAlarm(self);
    }

    /* Did we get interrupted in mid-wait? */
    if (self->interrupted) {
        self->interrupted = FALSE;
        ret = SYS_INTRPT;
    }

unlock:
#ifdef DEBUG_PRIORITY
    checkInvariants("sysMonitorWait", 0);
#endif /* DEBUG_PRIORITY */
    SCHED_UNLOCK();

    return ret;
}

static int dumpQueue(sys_thread_t *t, sys_thread_t **waiters, int sz) {
    int n;
    for (n = 0; t != 0; t = t->waitq, n++, sz--) {
        if (sz > 0) {
	    waiters[n] = t;
	}
    }
    return n;
}

int
sysMonitorGetInfo(sys_mon_t *mid, sys_mon_info *info)
{
    info->owner = mid->monitor_owner;
    if (mid->monitor_owner) {
        info->entry_count = mid->entry_count;
    }
    info->n_monitor_waiters = dumpQueue(mid->monitor_waitq,
					info->monitor_waiters,
					info->sz_monitor_waiters);
    info->n_condvar_waiters = dumpQueue(mid->condvar_waitq,
					info->condvar_waiters,
					info->sz_condvar_waiters);
    return SYS_OK;
}


bool_t
sysMonitorInUse(sys_mon_t * mon)
{
    return mon->monitor_owner != NULL ||
	mon->monitor_waitq != NULL ||
	mon->suspend_waitq != NULL ||
        mon->condvar_waitq != NULL;
}

sys_thread_t *
sysMonitorOwner(sys_mon_t *mon)
{
    return mon->monitor_owner;
}

#ifdef DEBUG_PRIORITY
static void
checkMonitor(sys_mon_t *mon, const char *msg)
{
    bool_t problem = FALSE;
    sys_thread_t *t = mon->monitor_owner;

    if (t != NULL) {
	if (mon->monitor_waitq != SYS_THREAD_NULL) {
	    sys_mon_t *m = t->inversion_queue;
	    bool_t found = FALSE;
	    if (m != NULL) {
		int p = INVERSION_PRIORITY(t);
		sys_mon_t *next = m->next_inversion;
		if (m == mon) {
		    found = TRUE;
		}
		while (next != NULL) {
		    int pri0 = m->monitor_waitq->priority;
		    int pri1 = next->monitor_waitq->priority;
		    if (pri0 < pri1) {
			problem = TRUE;
			VM_CALL(jio_fprintf)(stdout, "%s t:0x%x m:0x%x: priority %d < %d\n", msg,
			    t, m, pri0, pri1);
		    }
		    if (next == mon) {
			found = TRUE;
		    }
		    m = next;
		    next = next->next_inversion;
		}
		if (p != t->priority) {
		    problem = TRUE;
		    VM_CALL(jio_fprintf)(stdout, "%s t:0x%x m:0x%x: scheduling priority %d != %d\n",
			msg, t, mon, t->priority, p);
		}
		if (found != TRUE) {
		    problem = TRUE;
		    VM_CALL(jio_fprintf)(stdout, "%s t:0x%x m:0x%x: monitor not found\n",
			msg, t, mon);
		}
	    } else {
		problem = TRUE;
		VM_CALL(jio_fprintf)(stdout, "%s t:0x%x m:0x%x: waitq without inversion queue\n",
		    msg, t, mon);
	    }
	}
    }
    if (problem == TRUE) {
	/* We certainly want something better than this! */
	VM_CALL(jio_fprintf)(stderr, "checkMonitor found a problem!\n");
    }
}

static void
registeredDumpHelper(reg_mon_t *reg, void *cookie)
{
    checkMonitor(reg->mid, (const char *)cookie);
}

/* [Sheng] For the sake of a clean HPI separation, I temporarily commented 
 * checkInvariants out. We should either get rid of this or fix it some
 * time in the future.
 */
#if 0 
static void
monitorDumpHelper(monitor_t *mid, void *cookie)
{
    checkMonitor(sysmon(mid), (const char *)cookie);
}

#include "monitor_cache.h"

void
checkInvariants(const char *msg, int resched)
{
    sysAssert(SCHED_LOCKED());
    monitorEnumerate_unlocked(monitorDumpHelper, (char *)msg);
    registeredEnumerate_unlocked(registeredDumpHelper, (char *)msg);
    if (resched != -1) {
	extern sys_thread_t *ThreadQueue;
	sys_thread_t *self = greenThreadSelf();
	int r = ThreadQueue != SYS_THREAD_NULL &&
	    self->state != RUNNABLE ||
	    runnable_queue != SYS_THREAD_NULL &&
	    runnable_queue->priority > self->priority;
	if (r != resched) {
	    VM_CALL(jio_fprintf)(stdout, "%s : resched %d != %d\n", msg, r, resched);
	}
    }
}
#else
void
checkInvariants(const char *msg, int resched)
{
}
#endif

#endif /* DEBUG_PRIORITY */

