/*
 * @(#)mutex_md.c	1.5 98/08/14
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Mutex HPI implementation for green threads.
 *
 * most of the implemetation is in macros in mutex_md.h
 */

#include <synch.h>
#include <errno.h>
#include <string.h>

#include "hpi_impl.h"

#include "schedule.h"
#include "mutex_md.h"
#include "threads_md.h"

/*
 * [sl] The earlier mutexLock/mutexUnlock implementation has a subtle
 * bug that I haven't been able to track down. It sometimes happens
 * that libc malloc calls mutexUnlock on a mutex not owned by the
 * current thread.
 *
 * To workaround this problem, I've put in a mutex cache. It relies on
 * the more robust sys monitor implementation for mutexes.
 */
#define MUTEX_CACHE

#ifdef MUTEX_CACHE

#define MUTEX_TABLE_SIZE 31

typedef struct mutex_bucket {
    struct mutex_bucket *next;
    mutex_t *mutex;
    sys_mon_t monitor;
} mutex_bucket_t;

static mutex_bucket_t *mutex_buckets[MUTEX_TABLE_SIZE] = {0};

/* We need to statically preallocate some buckets because calling malloc
 * requires a mutex lock. Without preallocating the malloc mutex, we run
 * into an infinite recursion.
 */
#define N_PREALLOCATED_BUCKETS 16
static mutex_bucket_t preallocated_buckets[N_PREALLOCATED_BUCKETS];
static int n_allocated_buckets = 0;

static mutex_bucket_t *
lookupMutex(mutex_t *m)
{
    mutex_bucket_t *p;
    int index = (unsigned int)m % MUTEX_TABLE_SIZE;
    SCHED_LOCK();
    for (p = mutex_buckets[index]; p; p = p->next) {
        if (p->mutex == m) {
	    break;
	}
    }
    if (p == NULL && n_allocated_buckets < N_PREALLOCATED_BUCKETS) {
        p = &(preallocated_buckets[n_allocated_buckets++]);
	p->mutex = m;
	sysMonitorInit(&(p->monitor));
	p->next = mutex_buckets[index];
	mutex_buckets[index] = p;
    }
    SCHED_UNLOCK();
    return p;
}

static mutex_bucket_t *
tryAddMutex(mutex_bucket_t *np)
{
    mutex_bucket_t *p;
    int index = (unsigned int)np->mutex % MUTEX_TABLE_SIZE;
    SCHED_LOCK();
    for (p = mutex_buckets[index]; p; p = p->next) {
        if (p->mutex == np->mutex) {
	    break;
	}
    }
    if (p == NULL) {
        n_allocated_buckets++;
        np->next = mutex_buckets[index];
	mutex_buckets[index] = np;
	p = np;
    }
    SCHED_UNLOCK();
    return p;
}

int
mutexLock(mutex_t * m)
{
    mutex_bucket_t *p;
    if (sysThreadSelf() == NULL) {
        return 0;
    }
    p = lookupMutex(m);
    if (p == NULL) {
        mutex_bucket_t *p1;
        p = sysMalloc(sizeof(mutex_bucket_t));
	if (p == NULL) {
	    return -1;
	}
	p->mutex = m;
	sysMonitorInit(&(p->monitor));
	p1 = tryAddMutex(p);
	if (p1 != p) {
	    sysMonitorDestroy(&(p->monitor));
	    sysFree(p);
	    p = p1;
	}
    }
    sysMonitorEnter(greenThreadSelf(), &(p->monitor));
    return 0;
}

int
mutexUnlock(mutex_t * m)
{
    mutex_bucket_t *p;
    if (sysThreadSelf() == NULL) {
        return 0;
    }
    p = lookupMutex(m);
    sysAssert(p != NULL);
    sysMonitorExit(greenThreadSelf(), &(p->monitor));
    return 0;
}

#else

static void
mutexWakeup(mutex_t * m)
{
    gt_queue_t * queue = (gt_queue_t *) &m->data;
    sys_thread_t * waiter = queueRemoveHead(queue, waiter);
    Log3(4, "0x%p 0x%p waking up 0x%p", m, mutexOwner(m), waiter);
    if (*queue) {
	Log1(4, " (0x%p still waiting)\n", *queue);
    } else {
	Log(4, "%s\n");
    }
    queueInsert(&runnable_queue, waiter);
}

static void
mutexSleep(mutex_t * m)
{
    sys_thread_t * self = greenThreadSelf();
    sysAssert(mutexOwner(m) != self);
    queueInsert((gt_queue_t *) &m->data, self);
    Log3(4, "0x%p 0x%p going to sleep waiting for 0x%p\n",
	 m, self, mutexOwner(m));
    yieldContext(CONTEXT(self));
    Log2(4, "0x%p 0x%p waking up\n", m, self);
}

int
mutexLock(mutex_t * m)
{
    if (greenThreadSelf() == 0)
        return 0;
    SCHED_LOCK();
    sysAssert(mutexOwner(m) != greenThreadSelf());
    if (mutexOwner(m) != 0) {
	mutexSleep((m));
    }
    mutexOwner(m) = greenThreadSelf();
    SCHED_UNLOCK();
    return 0;
}

int
mutexUnlock(mutex_t * m)
{
    if (greenThreadSelf() == 0)
        return 0;
    SCHED_LOCK();
    sysAssert(mutexOwner(m) == greenThreadSelf());
    if (*(gt_queue_t *)(&m->data) != NULL) {
	mutexWakeup((m));
    }
    mutexOwner(m) = NULL;
    SCHED_UNLOCK();
    return 0;
}
#endif /* MUTEX_CACHE */

/*
 * WARNING: It is crucial for the Solaris/Green threads implementation
 * of JAVA that we never try to acquire a monitor while scheduling is
 * locked.  If we did, and the monitor is already held by someone else,
 * then we would wait for it.  If we wait, we will switch contexts and
 * another thread will run at precisely the time when we don't want that.  
 * (Note that reacquiring a monitor that we already own isn't a problem:
 * we are always guaranteed to get it!)  We take pains throughout the
 * code to ensure that we don't enter a monitor when scheduling is locked.
 *
 * However, the Solaris implementation plays another trick with monitors.
 * libc calls (e.g. malloc()) must be made Green thread-safe.  libc is
 * normally made Solaris thread-safe by libthread providing definitions
 * for _mutex_lock() and _mutex_unlock().  Without libthread, libc's calls
 * to these functions are effectively noops.
 *
 * We make libc Green thread-safe by defining these functions to use our
 * monitors, taking them from the monitor cache using as the cache key
 * the mutex_t pointer that libc passes in.  Note, though, that this
 * leaves us with a new, hidden way to enter a monitor with scheduling
 * locked.  Because several threads may simultaneously want to call, say,
 * malloc(), passing _mutex_lock() the same pointer, we are subject
 * to the same race conditions that can defeat the intended behavior.
 *
 * This situation can only be managed by avoiding libc calls when the
 * scheduler is locked.  To ensure that this doesn't happen, assertions
 * have been placed in _mutex_lock() and _mutex_unlock() that will die
 * if those functions are called with scheduling locked.  (monitorExit()
 * acquires the monitor cache lock, so needs to be checked too.)  If
 * you find these assertions blowing up on you, look in the stack trace
 * for libc or other calls that are calling these mutex functions with
 * scheduling locked.
 */

int _mutex_lock(mutex_t * m);
int _mutex_unlock(mutex_t * m);

int
_mutex_lock(mutex_t * m)
{
    sysAssert(_scheduling_lock == 0);
    mutexLock(m);
    return 0;
}

int
_mutex_unlock(mutex_t * m)
{
    sysAssert(_scheduling_lock == 0);
    mutexUnlock(m);
    return 0;
}

#ifndef MUTEX_CACHE

int
_mutex_init(mutex_t * m, int type, void * arg)
{
    memset(m, 0, sizeof(*m));
    return 0;
}

int
mutex_init(mutex_t * m, int type, void * arg)
{
    return _mutex_init(m, type, arg);
}

#endif /* MUTEX_CACHE */
