/*
 * @(#)signals.c	1.61 98/10/27
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * A spl and interrupt implementation to run on top of Unix.
 */

/*
 * Header files.
 */
#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>

#include "hpi_impl.h"

#include "threads_md.h"
#include "monitor_md.h"
#include "interrupt.h"
#include "itimer_md.h"
#include "schedule.h"
#include "signals.h"
#include "iomgr.h"      	/* For io lock */

/*
 * Local variables.
 */
static sys_mon_t	*handlerMonitors[SYS_ASYNC_MON_MAX];

sys_mon_t *PendingNotifyQ;

/*
 * Local routine prototypes.
 */

/*
 * intrNotifyHandler() --
 *	Interrim routine that delivers signals in BKM's original fashion.
 */
static void intrNotifyHandler(int sig, void *siginfo, void *uc, void *arg);


#include <limits_md.h>
#include <poll.h>

/* The polling fd table.  This table is contains all the file
   descriptors which are currently attempting i/o but have gotten
   EAGAIN indicating the operation would block.  When a sigio comes
   in, this polling structure is used to figure out which file
   descriptor(s) is ready to do i/o.  This might seem like overkill,
   but for applications which have 50 or 100 or more open files, e.g.,
   network connections, the old way of waking up *all* the threads
   blocked in i/o is just too slow!  Imagine 100 threads attempting to
   reread their connection before the right one is tried.  Way too
   slow.

   This array is always packed.  That is, there are no holes in it.
   The goal is to pass the minimum amount of fd's to the kernel for
   examination, and to only pass valid fds.  Hopefully this isn't more
   work than it's worth. */

static struct pollfd *pollTable;

/* This is the number of fds to examine in fdsToPoll */
static int	fdCount = -1;

/*
 * pollOffset is indexed by file descriptor.  It maps from a file
 * descriptor to a position in the pollTable.  A pollOffset of -1
 * indicates fd is not allocated into pollTable.
 */
static short	* pollOffset;

/*
 * External routines.
 */

static void asyncIORegister(void);

void
InitializeAsyncMonitors(int max_files)
{
    int i;

    /* Async monitor numbering starts at 1 */
    for (i = 1; i < SYS_ASYNC_MON_MAX; i++) {
	handlerMonitors[i] = (sys_mon_t *) malloc(sysMonitorSizeof());
	if (handlerMonitors[i] == 0) {
	    VM_CALL(panic)("out of memory");		/* fatal error */
	}
	switch (i) {	/* Have to keep in sync with the enum... */
	    case SYS_ASYNC_MON_ALARM:
		VM_CALL(monitorRegister)(handlerMonitors[i], "Alarm monitor");
		break;
	    case SYS_ASYNC_MON_IO:
		VM_CALL(monitorRegister)(handlerMonitors[i], "I/O monitor");
		break;
	    case SYS_ASYNC_MON_CHILD:
		VM_CALL(monitorRegister)(handlerMonitors[i],
					 "Child death monitor");
		break;
	    case SYS_ASYNC_MON_USRSIG:
		VM_CALL(monitorRegister)(handlerMonitors[i],
					 "User signal monitor");
		break;
	    default:
	        VM_CALL(monitorRegister)(handlerMonitors[i],
					 "Unknown monitor");
		break;
	}
    }

    asyncIORegister();

    if (fdCount == -1) {
        short *offset;
        int slot;

	Log1(1, "Initializing pollOffset entries = %d\n", max_files);
	fdCount = 0;
	slot = max_files;
        pollOffset = (short *)calloc(max_files, sizeof(short));
	if (pollOffset == 0) {
	    VM_CALL(panic)("out of memory");		/* fatal error */
	}
	for (offset = pollOffset; --slot >= 0; offset += 1) {
	    *offset = -1;
	}
        pollTable = (struct pollfd *)calloc(max_files, sizeof(struct pollfd));
	if (pollTable == 0) {
	    VM_CALL(panic)("out of memory");
	}
    }
}

static int
interruptBroadcast(sys_mon_t *mid)
{
    /*
     * If a thread is currently inside the monitor, there is a potential
     * race condition.  It may have already checked its test condition,
     * and decided it wants to wait.  It will miss the notification,
     * assuming that none can arrive while it owns the monitor.
     *
     * Solve this race by setting the stick_notification boolean,
     * indicating that a CV notify was done while some thread owned
     * the monitor.  monitorExit() and monitorWait() are responsible
     * for clearing this boolean.  monitorWait() should also fail
     * to wait if the boolean is set.
     */
    if (mid->monitor_owner != SYS_THREAD_NULL) {
	mid->flags |= SYS_MON_STICKY_NOTIFICATION;
    }
    return monitorBroadcast(mid);
}

/*
 * Thread all outstanding CV notification.  Don't mess with the currently
 * running context, because we have no idea what state we are in.
 */
static int
processPendingNotification(void)
{
    sys_mon_t *mid;
    sys_mon_t *midq;
    int need_to_reschedule = 0;

    /*
     * Perform the notifyall's that couldn't be done because of locking.
     * Pull all of the pending CV notifies off of the queue, and process
     * them.  The notification is completely normal, except for the locking.
     * If a yield is necessary, it is done by the caller.
     */
    sysAssert(PendingNotifyQ);
    midq = PendingNotifyQ;
    while (midq) {
	mid = midq;
	midq = midq->pendingq;
	mid->pendingq = SYS_MID_NULL;
	mid->flags &= ~SYS_MON_PENDING_NOTIFICATION;

	/* Now notify all of the waiters */
	need_to_reschedule |= interruptBroadcast(mid);
    }
    PendingNotifyQ = SYS_MID_NULL;
    return need_to_reschedule;
}

/*
 * return a registered asynchronous notifier
 */
sys_mon_t *
asyncMon(async_mon_key_t key)
{
    if (key >= SYS_ASYNC_MON_MAX || key == 0) {
	return SYS_MID_NULL;
    } else {
	return handlerMonitors[key];
    }
}

static int asyncKeyToSig[SYS_ASYNC_MON_MAX] = {
        0,		/* Placeholder.  async_mon_key_t starts at 1 */
	SIGALRM,	/* SYS_ASYNC_MON_ALARM */
	SIGIO,		/* SYS_ASYNC_MON_IO */
	SIGCHLD,	/* SYS_ASYNC_MON_CHILD */
	0,              /* SYS_ASYNC_MON_USRSIG */
};

/*
 * This interface is used by client programs to register monitors that
 * they will recieve asynchronous notification on.  The async monitor key
 * is translated into a machine dependent Unix signal number.
 */
int
asyncNotifier(async_mon_key_t key)
{
    int sig;
    sys_mon_t *mid;

    if (key >= SYS_ASYNC_MON_MAX) {
	return -1;
    }

    sig = asyncKeyToSig[key];
    mid = handlerMonitors[key];

    mid->flags |= SYS_MON_HAS_HANDLER;
    /* Mustn't be a cache monitor: see monitor_md.h */

    /* Register our interrrupt handler */
    intrRegister(sig, intrNotifyHandler, handlerMonitors[key]);

    return 0;
}

void
asyncIOActivateFD(int fd)
{
    short *offsetp;
    struct pollfd *pollfd;
    sigset_t oldmask, newmask;

    /* Here we need to mask SIGIO signal so that the signal handler
       does not run & use the poll table while we are constructing it.
       This is necessary as poll table is shared data structure and we
       cannot use the locks in the signal handler.
       NOTE: The assumption here is sighold and sigrelse apply process
       wide - TBD: Confirm this is true! If NOT then we need to grab the
       SCHED_LOCK so that thread cannot context switch here -
       Well just to be on safe side we grab SCHED_LOCK as well - pavani*/

    SCHED_LOCK();
    /*
     * We used to use sighold(SIGIO) here and sigrelse(SIGIO) below, but
     * that is unsafe, as the sigrelse(SIGIO) always unblocks SIGIO,
     * even if it was already unblocked before the sighold(SIGIO).
     *
     * The SCHED_LOCK is still needed to keep signals other than SIGIO
     * from causing a reschedule.
     *
     * Dean Long 3/22/97
     */
    sigemptyset(&newmask);
    sigaddset(&newmask, SIGIO);
    green_sigprocmask(SIG_BLOCK, &newmask, &oldmask);

    offsetp = &pollOffset[fd];
    if (*offsetp == -1) {
	*offsetp = fdCount++;
    }

    pollfd = &pollTable[*offsetp];
    pollfd->fd = fd;		/* may be redundant */
    Log2(1, "Activating fd=%d onto slot=%d\n", fd, *offsetp);
    green_sigprocmask(SIG_SETMASK, &oldmask, NULL);
    SCHED_UNLOCK();
}

void
asyncIODeactivateFD(int fd)
{
    short *offsetp;
    sigset_t oldmask, newmask;
    int	slot;

    SCHED_LOCK();
    sigemptyset(&newmask);
    sigaddset(&newmask, SIGIO);
    green_sigprocmask(SIG_BLOCK, &newmask, &oldmask);

    offsetp = &pollOffset[fd];
    if ((slot = *offsetp) != -1) {
	/* Now copy the highest slot in the poll array into this slot,
	   and adjust the pollOffset to point at this new location.  But
	   only if we're not already in the last slot. */
	Log2(1, "Deactivating fd=%d from slot=%d\n", fd, slot);

	/* not reading or writing this
	   fd anymore */
	*offsetp = -1; 	/* this one's no longer occupied */
	if (slot != --fdCount) {	/* not last entry in pollTable */
	    int movefd = pollTable[fdCount].fd;

	    pollTable[slot] = pollTable[fdCount];
	    pollOffset[movefd] = slot;
	    Log3(1, "\t and moving fd=%d from slot %d to %d\n",
		 movefd, fdCount, slot);
	}
    }
    green_sigprocmask(SIG_SETMASK, &oldmask, NULL);
    SCHED_UNLOCK();
}

/*
 * Asynchronous notification of a condition variable
 *
 * NotifyAll the condition variable, and perform any scheduling updates
 * required.   Return TRUE if the caller needs to immediately
 * reschedule.
 *
 * ASSUMPTION: the current threads state is already saved, and we
 * are free to do anything we wish.
 */
int
asyncEventNotify(sys_mon_t *mid)
{
    int need_to_switch = 0;

    sysAssert(mid!=NULL);
    sysAssert(SCHED_LOCKED());
    if (SCHED_LOCKED() > 1) {
	intrLock();
	/*
	 * Queue this notification
	 */
	if ((mid->flags & SYS_MON_PENDING_NOTIFICATION) == 0) {
	    sysAssert(mid->pendingq == SYS_MID_NULL);
	    mid->pendingq = PendingNotifyQ;
	    PendingNotifyQ = mid;
	}
	mid->flags |= SYS_MON_PENDING_NOTIFICATION;
	intrUnlock();
    } else {
	/* Actually handle the notification */
	need_to_switch = interruptBroadcast(mid);
    }
    return need_to_switch;
}

static void
sigioNotifier(int interrupt, void *siginfo, void *context, void *arg)
{
    int resched = 0;
    int nfds, count;

    nfds = fdCount;
    /* Optimization: when there's only one fd to choose from it must have
       been the one that delivered the sigio, right? */
    if (nfds == 1) {
	Log1(1, "Notify single fd=%d\n", pollTable[0].fd);
	resched |= asyncEventNotify(fdmon[pollTable[0].fd]);
    } else if (nfds > 0) {
	int i = 0;

	/*
	 *  We look at the fds in the pollTable and set their event flags
	 * based on whether there are readers or writers on those fds.
	 * This helps us avoid waking up threads that are waiting to read
	 * when the fd is available for writing.  See bug 4056954.
	 */
	for (i = 0; i < nfds; i++) {
	    int fd = pollTable[i].fd;
	    if (FD_INACTIVE(fd)) {
		/*
		 * repack the pollTable so that the active fds are
		 * at the bottom.  fds which are currently don't have
		 * readers or writers get moved to the end so that we
		 * don't have to pass them into poll.  For very large
		 * sets of file descriptors, the overhead of passing
		 * this array into the kernel may be high.
		 */
		int newfd;

		/*
		 * begin searching down from the top of the table for
		 * an active fd to put in this place.
		 * This code is a litte weird since we are decrementing
		 * the bounds of the loop we are in.
		 */
		nfds -= 1;
		while (FD_INACTIVE(pollTable[nfds].fd) && nfds > i)
		    nfds -= 1;

		/*
		 * if we run out of fds, go back to the start of the loop
		 * so that we fall out the bottom.
		 */
		if (i >= nfds)
		    continue;

		newfd = pollTable[nfds].fd;

		/* move the active fd into this slot */
		pollTable[i].fd = newfd;
		pollOffset[newfd] = i;

		/* move this fd into the last slot */
		pollTable[nfds].fd = fd;
		pollOffset[fd] = nfds;
		fd = newfd;
	    }

	    /* set the poll flags for this fd */
	    pollTable[i].events = (FD_READERS(fd) > 0) ? POLLIN : 0 |
		(FD_WRITERS(fd) > 0) ? POLLOUT : 0;
	}

	sysAssert(nfds >= 0);
	Log2(1, "Looking at %d fds of %d fds\n", nfds, fdCount);
	/* don't even poll if we don't actually have an active fd */
	if (nfds > 0) {
	    count = (*systable[SYS_POLL].addr)((int)pollTable, nfds, 0);
	    Log1(1, "Poll returns %d fds to look at\n", count);
	    for (i = 0; (i < nfds) && (count>0); i++) {
		if ((pollTable[i].revents &
		     (POLLIN|POLLOUT|POLLERR|POLLHUP|POLLNVAL)) != 0) {
		    Log2(1, "Notify fd=%d on mon for revents %d \n",
			 pollTable[i].fd, pollTable[i].revents);
		    pollTable[i].events = (POLLIN|POLLOUT);
		    resched |= asyncEventNotify(fdmon[pollTable[i].fd]);
		     --count;
		}
	    }
	}
    }

    /* And for good measure... */
    /* This monitor is the default io monitor mainly used for
       thread manipulating the group fds to wait on i.e when
       a poll/select happens */
    resched |= asyncEventNotify(handlerMonitors[SYS_ASYNC_MON_IO]);

    /*
     * This function also handles SIGCLD.  In that case we notify
     * anyone waiting our implementation of waitid.
     */
    if (interrupt == SIGCLD) {
	resched |= asyncEventNotify(handlerMonitors[SYS_ASYNC_MON_CHILD]);
    }

    if (resched) {
	_needReschedule = 1;
    }
}

/* register the signal handler for asynchronous io */
static void
asyncIORegister()
{
    intrRegister(SIGIO, sigioNotifier, 0);

    /*
     * We also register SIGCLD to notify using sigioNotifier since
     * SIGCLD may mean the fds we have connected to our children to
     * are closed but this doesn't seem to generate a SIGIO on Solaris
     * like you'd expect.
     */
    intrRegister(SIGCLD, sigioNotifier, 0);
}

/*
 * intrNotifyHandler() --
 *	Handle a unix signal.  This is called from the interrupt
 *	routine, which saves all pending thread state.  If there is
 *	a registered monitor, notify it (and potentially yield).
 *
 *	There can be no signal which needs both async and sync
 *	notification.
 */
static void
intrNotifyHandler(int interrupt, void *siginfo, void *context, void *arg)
{
    sys_mon_t	*mon = (sys_mon_t *)arg;

    Log2(2, "Received signal %d interrupting sys_thread_t 0x%x\n", interrupt,
	 greenThreadSelf());

    if ((mon->flags & SYS_MON_HAS_HANDLER) != 0
	&& asyncEventNotify(mon)) {
	Log(2, " (Not Locked)\n");
	_needReschedule = 1;
    }
#ifdef DEBUG
    else {
	Log1(2, " (Locked: %d)\n", _scheduling_lock);
    }
#endif
}


/*
 * Lock the thread scheduler
 */
void
_sched_lock()
{
    ++_scheduling_lock;
}

/*
 * Thread all of the pending condition variable notifications that
 * couldn't be handled at interrupt time.
 *
 * ASSUMPTION: this is being called from a thread context.
 */
void
_sched_unlock()
{
    int need_to_reschedule;

    --_scheduling_lock;
    sysAssert(_scheduling_lock >= 0);
    if (!(_scheduling_lock == 0 && PendingNotifyQ)) {
	return;
    }

restart:
    intrLock();
    SCHED_LOCK();

    /*
     * Perform the notifyall's that couldn't be done because of locking.
     * The notification is completely normal, except for the locking.
     */
    if (PendingNotifyQ) {
	need_to_reschedule = processPendingNotification();

	intrUnlock();
	if (need_to_reschedule) {
	    YIELD();
	}

	/*
	 * We may have gotten an interrupt in between the last two
	 * statements (intrUnlock & YIELD).  Check, and if so,
	 * loop.  This also handles the case were we really do yield
	 * to somebody else, and they get an interrupt while yielding
	 * back to us.
	 *
	 * This is gross, but it keeps this thread from recursing in
	 * this routine.
	 */
	_scheduling_lock--;
	if (PendingNotifyQ) {
	    goto restart;
	}

    } else {
	_scheduling_lock--;
	intrUnlock();
    }

    return;
}

#define CATCH_SIGMASK_CHANGES

#ifdef CATCH_SIGMASK_CHANGES

#include <dlfcn.h>

/*
 * The following functions, interposed in front of sigprocmask and
 * _sigprocmask, are necessary to let us know when our saved
 * signal mask is no longer up to date.  The alternative is to
 * always look at the signal mask when we switch threads, but that
 * adds a system call for each reschdule.
 */

int
green_sigprocmask(int how, const sigset_t *set, sigset_t *oset)
{
    typedef int fn_t(int how, const sigset_t *set, sigset_t *oset);
    static fn_t *func = NULL;

    if (func == NULL) {
        func = (fn_t *)dlsym(RTLD_NEXT, "_sigprocmask");
    }

    return (*func)(how, set, oset);
}

int
_sigprocmask(int how, const sigset_t *set, sigset_t *oset)
{
    int ret;
    sys_thread_t *self = greenThreadSelf();

    sysAssert(!SCHED_LOCKED());
    SCHED_LOCK();
    ret = green_sigprocmask(how, set, oset);
    if (self != NULL) {
	self->sigmaskdirty = 1;
    }
    SCHED_UNLOCK();
    return ret;
}

int
sigprocmask(int how, const sigset_t *set, sigset_t *oset)
{
    return _sigprocmask(how, set, oset);
}
#endif
