/*
 * @(#)sys_api_td.c	1.24 98/08/24
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/filio.h>
#include <sys/socket.h>
#include <poll.h>
#include <string.h>

#include "hpi_impl.h"

#include "threads_md.h"
#include "monitor_md.h"
#include "itimer_md.h"
#include "iomgr.h"
#include "largefile.h"

/*
 * This macro is used as a wrapper around our POSIX overrides. It sets
 * the interruptible_io bit on the current sys_thread to true,
 * executes the POSIX call and restore the previous value of
 * interruptible_io.
 *
 * This macro should be used with care, and only for its stated
 * purpose.  
 */
#define INTERRUPT_IO(returnType, call) {                   \
    returnType result;                                     \
    sys_thread_t * thread = sysThreadSelf();               \
    int tmp;                                               \
    if (thread == 0) {                                     \
	result = call;                                     \
    } else {                                               \
	tmp = thread->interruptible_io;                    \
	thread->interruptible_io = TRUE;                   \
	result = call;                                     \
	thread->interruptible_io = tmp;                    \
    }                                                      \
    return result;                                         \
}

/*
 * I/O-related sys API
 */

int sysClose(int fd) {
    return close(fd);
}

size_t
sysRead(int fd, void *buf, unsigned int nBytes) {
    INTERRUPT_IO(size_t, read(fd, buf, nBytes))
}

size_t
sysWrite(int fd, const void *buf, unsigned int nBytes) {
    INTERRUPT_IO(size_t, write(fd, buf, nBytes))
}

/* 
 * Poll the fd for reading for timeout ms.  Returns 1 if something's
 * ready, 0 if it timed out, -1 on error (e.g., fd was closed or marked
 * closed).  Timeout in milliseconds.
 */
extern int _select (int width, fd_set *, fd_set *, fd_set *, struct timeval *);
#include "clock.h" /* timerCmp(), etc */

int
sysTimeout(int fd, long timeout) {
    sys_mon_t *mon; 
    int ret = -1;
    sys_thread_t *self = sysThreadSelf();

    if (fd < 0) {
	Log1(1, "Timeout on closed fd: %d?\n", fd);
	errno = EBADF;
	return -1;
    }
    
    check_single_fd(fd);

    mon = fdmon[fd];
    sysAssert(mon != NULL);
    
    sysMonitorEnter(self, mon);
    if (fd >= 0 && !closed_fd(fd)) {
	struct fd_set fds;
	struct timeval zt, limit, now;

	FD_ADD_READER(fd);
	zt.tv_sec = zt.tv_usec = 0;
	gettimeofday(&now, 0);
	limit.tv_sec = now.tv_sec + (timeout / 1000);
	limit.tv_usec = now.tv_usec + ((timeout % 1000) * 1000);
	FD_ZERO(&fds);
	FD_SET(fd, &fds);
	
	ret = _select(fd + 1, &fds, 0, 0, &zt);
	if (ret == 0) {
	    /* we have to wait */
	    while (1) {
		if (sysMonitorWait(self, mon, timeout) == SYS_INTRPT) {
		    ret = SYS_INTRPT;			/* INTERRUPTED */
		    break;
		}
		if (closed_fd(fd)) {
		    /* This fd closed by another thread: break, return error */
		    ret = -1;
		    errno = EBADF;
		    break;
		}
		FD_SET(fd, &fds);
		ret = _select(fd + 1, &fds, 0, 0, &zt);
		if (ret) {
		    break;
		}
		gettimeofday(&now, 0);
		if (timerCmp(&now, &limit, >) || timerEq(&now, &limit)) {
		    break;
		} else {
		    timeout = ((limit.tv_sec - now.tv_sec)*1000) +
			      ((limit.tv_usec - now.tv_usec) / 1000);
		}
	    }
	}
	FD_REMOVE_READER(fd);

	/* If this is the last thread to leave and fd has been closed
	 * call system close on the fd.
	 */
	if (FD_INACTIVE(fd) == 0 && closed_fd(fd)) {
	    system_close(fd);
	}
    } else {
	/* the fd has been closed */
	errno = EBADF;
    }
    sysMonitorExit(self, mon);
    return ret;
}

/* This seems to not work, as tested by src/share/net/UDPCloseTest the
   idea is that it should be faster, since select is implemented using
   poll.

jint 
sysTimeout(jint fd, jint timeout) {
    int ret;
    int revents = 0;
    struct pollfd pollTable[1];

    if (fd >= 0 && !closed_fd(fd)) {
	pollTable[0].fd = fd;
	pollTable[0].events =  POLLIN;
	ret = poll(pollTable, 1, timeout);
	revents = pollTable[0].revents;
      
	if (ret == 0) {
	    return ret;
	} else if (ret == -1 || (revents & POLLNVAL)) {
	    ret = -1;
	    errno = EBADF;
	}
    } else {
	errno = EBADF;
    }
    return ret;
} */

int
sysAvailable(int fd, jlong *pbytes) {
    sys_mon_t *mon;
    int ret = 0; 
    off64_t cur, end;
    int mode;
    sys_thread_t *self = sysThreadSelf();

    if (fd < 0) {
	Log1(1, "available operation on closed fd: %d?\n", fd);
	errno = EBADF;
	return ret;
    }

    mon = fdmon[fd];
    sysAssert(mon != NULL);

    sysMonitorEnter(self, mon); 

    if (fd >= 0) {
	if (sysFfileMode(fd, &mode) >= 0) {
	    if (S_ISCHR(mode) || S_ISFIFO(mode) || S_ISSOCK(mode)) {
		int n;
		if (ioctl(fd, FIONREAD, &n) >= 0) {
		    *pbytes = n;
		    ret = 1;
		}
	    } else if (((cur = lseek64_w(fd, 0L, SEEK_CUR)) != -1) &&
		       ((end = lseek64_w(fd, 0L, SEEK_END)) != -1) &&
		       (lseek64_w(fd, cur, SEEK_SET) != -1)) {
		*pbytes = end - cur;
		ret = 1;
	    }
	}
    } else {
	errno = EBADF;
    }

    sysMonitorExit(self, mon); 
    return ret;
}

jlong
sysSeek(int fd, jlong offset, int whence) {

    sys_mon_t *mon;
    jlong ret = -1;
    sys_thread_t *self = sysThreadSelf();

    if (fd < 0) {
	Log1(1, "lseek operation on closed fd: %d?\n", fd);
	errno = EBADF;
	return ret;
    }

    mon = fdmon[fd];
    sysAssert(mon != NULL);

    sysMonitorEnter(self, mon);  

    if (fd >= 0) {
	ret = lseek64_w(fd, offset, whence);
    } else {
	errno = EBADF;
    }

    sysMonitorExit(self, mon); 
    return ret;
}

int
sysSetLength(int fd, jlong length) {
    sys_mon_t *mon;
    int ret = -1;
    sys_thread_t *self = sysThreadSelf();

    if (fd < 0) {
	Log1(1, "setLength operation on closed fd: %d?\n", fd);
	errno = EBADF;
	return ret;
    }
    
    mon = fdmon[fd];
    sysAssert(mon != NULL);

    sysMonitorEnter(self, mon);  

    if (fd >= 0) {
	ret = ftruncate64_w(fd, length);
    } else {
	errno = EBADF;
    }

    sysMonitorExit(self, mon); 
    return ret;
}

int 
sysSync(int fd) {
    int result;
    sys_mon_t *mon;
    sys_thread_t *self = sysThreadSelf();

    if (fd < 0) {
	Log1(1, "sync operation on closed fd: %d?\n", fd);
	errno = EBADF;
	return SYS_ERR;
    }

    mon = fdmon[fd];
    sysAssert(mon != NULL);

    sysMonitorEnter(self, mon);
    if (fd < 0) {
	errno = EBADF;
    } else {
	if ((result = fsync(fd)) == -1) {
	    result = SYS_ERR;
	}
    }
    sysMonitorExit(self, mon); 

    return result;
}


/*
 * Net-related sys API
 */

int
sysRecv(int fd, char *buf, int nBytes, int flags) {
    INTERRUPT_IO(int, recv(fd, buf, nBytes, flags))
}

int 
sysSend(int fd, char *buf, int nBytes, int flags) {
    INTERRUPT_IO(int, send(fd, buf, nBytes, flags))
}

int 
sysConnect(int fd, struct sockaddr *him, int len) {
    INTERRUPT_IO(int, connect(fd, him, len))
}

int 
sysListen(int fd, long count) {
    INTERRUPT_IO(int, listen(fd, count))
}

int
sysAccept(int fd, struct sockaddr *him, int *len) {
    INTERRUPT_IO(int, accept(fd, him, len))
}

long
sysSocketAvailable(int fd, long *pbytes) {
    sys_mon_t *mon;
    int ret = 1;
    sys_thread_t *self = sysThreadSelf();

    if (fd < 0) {
	Log1(1, "socket available operation on closed fd: %d?\n", fd);
	return 0;
    }

    mon = fdmon[fd];
    sysAssert(mon != NULL);

    sysMonitorEnter(self, mon);     
    if (fd < 0 || ioctl(fd, FIONREAD, pbytes) < 0) {
	ret = 0;
    } 
    sysMonitorExit(self, mon); 
    return ret;
}

int
sysSocket(int domain, int type, int protocol) {
    return socket(domain, type, protocol);
}

int
sysSocketClose(int fd) {
    return sysClose(fd);
}

/*
 * Routines to do datagrams
 */

ssize_t
sysSendTo(int fd, char *buf, int len, 
	  int flags, struct sockaddr *to, int tolen) {
    INTERRUPT_IO(ssize_t, sendto(fd, buf, len, flags, to, tolen))
}

ssize_t
sysRecvFrom(int fd, char *buf, int nBytes, 
	    int flags, struct sockaddr *from, int *fromlen) {
    INTERRUPT_IO(ssize_t, recvfrom(fd, buf, nBytes, flags, from, fromlen))
}

