/*
 * @(#)threads_md.c	1.136 98/10/27
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Implementation of Java threads HPI on top of Green threads
 */

#include "hpi_impl.h"

#include <string.h>
#include <ucontext.h>
#include <sys/resource.h>
#include <errno.h>

#include "threads_md.h"
#include "internal_md.h"
#include "queue.h"
#include "schedule.h"
#include "itimer_md.h"
#include "clock.h"
#include "iomgr.h"

static void InitializeIdle(void);
static void greenThreadExit(void);

#ifdef DEBUG_PRIORITY
void checkInvariants(const char *mesg, int resched);
#endif /* DEBUG_PRIORITY */

static sys_thread_t *ThreadQueue;
static sys_mon_t *_sys_queue_lock;

static int ActiveThreadCount = 0;		/* All threads */
static sys_thread_t *idle_thread;

#define SYS_QUEUE_LOCK(self)    sysMonitorEnter(self, _sys_queue_lock)
#define SYS_QUEUE_LOCKED(self)  sysMonitorEntered(self, _sys_queue_lock)
#define SYS_QUEUE_UNLOCK(self)  sysMonitorExit(self, _sys_queue_lock)
#define SYS_QUEUE_NOTIFYALL(self)  sysMonitorNotifyAll(self, _sys_queue_lock)
#define SYS_QUEUE_WAIT(self) sysMonitorWait(self, _sys_queue_lock, \
					SYS_TIMEOUT_INFINITY)

/*
 * Thread C stack overflow check
 */

/*
 * sysThreadCheckStack() being a function call is unfortunate, as it
 * takes stack space to do, but that is weakly accounted for by the
 * previous stack redzone.  In general, where we can't predict stack
 * use by C, thread stack overflow checking doesn't really work.
 */ 

#define STACK_REDZONE 4096

int
sysThreadCheckStack() 
{
    sys_thread_t *tid = greenThreadSelf();

    /* Stacks grow toward lower addresses on Solaris... */
    if ((char *)(tid)->stack.base - (char *)&(tid) + STACK_REDZONE <
	    tid->stack.size) {
	return 1;
    } else {
	return 0;
    }
}

/* 
 * Allocate and initialize the sys_thread_t structure for an arbitary
 * native thread.
 */
int
sysThreadAlloc(sys_thread_t **tidP)
{
    *tidP = greenThreadSelf();
    return SYS_OK; /* always succeeds for green threads */
}

/*
 * threadBootstrapMD() bootstraps the UNIX process running from main()
 * into a first primordial thread.  This thread is distinguishable because
 * it uniquely has the primordial_thread flag on in its private data.
 * However, so that we have to special-case it as little as possible, we
 * set it up to look as closely as possible to a thread that we created.
 * One difference is that its stack is *not* known to us.
 */
int
threadBootstrapMD(sys_thread_t **tidP, sys_mon_t **lockP, int nb)
{
    sys_thread_t *tid;

    nReservedBytes = (nb + 7) & (~7);
    tid = allocThreadBlock();
    if (tid == NULL) {
        return SYS_NOMEM;
    }

    tid->primordial_thread = 1;
    tid->interrupted = FALSE;
    
    green_sigprocmask(SIG_SETMASK, NULL, &MACHDEP(tid)->signalMask);

    tid->state = RUNNABLE;
    setCurrentThread(tid);

    /* Initialize the queue lock monitor */
    _sys_queue_lock = (sys_mon_t *)sysMalloc(sysMonitorSizeof());
    if (_sys_queue_lock == NULL) {
        return SYS_ERR;
    }
    VM_CALL(monitorRegister)(_sys_queue_lock, "Thread queue lock");

    ActiveThreadCount++;        /* Count the main thread */

    tid->next = ThreadQueue;    /* Chain all threads */
    ThreadQueue = tid;

    {
	ucontext_t u;
	struct rlimit r;
	stackp_t base, limit;
	size_t stack_size;

	getcontext(&u);
	getrlimit(RLIMIT_STACK, &r);
	stack_size = (size_t)r.rlim_cur;
	/* On Solaris 2.6 (beta-refresh) ss_sp went from
	 * type char* to type void*.  The following won't
	 * compile on 2.6 w/o coercion to char*
	 */
	base = (char *)u.uc_stack.ss_sp + u.uc_stack.ss_size;
	limit = (stackp_t)((caddr_t)base - stack_size);

	tid->stack.base = base;
	tid->stack.size = (long)((caddr_t)base - (caddr_t)limit);
    }

    *lockP = _sys_queue_lock;
    *tidP = tid;

    setFPMode();

    return SYS_OK;
}

void
InitializeHelperThreads()
{
    /* Initialize the clock */
    InitializeClock();

    /* Create the idle thread */
    InitializeIdle();

    PrimordialThreadBootstrapped(TRUE); /* Start redirecting system calls */
}

/*
 * Access to the thread stack pointer of an arbitrary thread (for GC).
 */
void *
sysThreadStackPointer(sys_thread_t * tid)
{
    return currentStackPointer(tid);
}

long *
sysThreadRegs(sys_thread_t * tid, int *nregs)
{
    *nregs = 0;
    return 0; /* green threads do not need saved regs. */
}

static void removeFromActiveQ(sys_thread_t *p);

typedef struct {
    int arg;
    int ret;
    ucontext_t *oldContext;
    sys_thread_t *tid;
} start_args;

static void
start_func(void (*func)(int), start_args *args)
{
    int arg = args->arg;
    if (!saveContext(CONTEXT(args->tid), (stackp_t)&func)) {
	CONTEXT(args->tid)->unix_errno = 0;
	args->ret = 1;
	setcontext(args->oldContext);
    }
    SCHED_UNLOCK();
    (*func)(arg);
    sysThreadFree();
    greenThreadExit();
    /* never reached */
}

#if defined(i386)
/*
 * Workaround for BUG 4007713/4026661/1262991:
 * create a private copy (fix) for x86 green threads getcontext
 * called green_getcontext to avoid the libc getcontext bug.
 * Furthermore, the green_getcontext code differs from the
 * libc behavior since it does not call __cerror if the system
 * call fails. Reference to __cerror was removed since its no
 * longer a global in Solaris 2.6 libc. This change in API should
 * be benign to the VM (we don't check for getcontext return code).
 */

extern int green_getcontext(ucontext_t *);
#pragma unknown_control_flow(green_getcontext)
#endif

int
sysThreadCreate(sys_thread_t **tidP, long stack_size,
		void (*start)(void *), void *arg)
{ 
    sys_thread_t *tid = allocThreadBlock();
    if (tid == NULL) {
        return SYS_NOMEM;
    }
    
#ifdef DEBUG_PRIORITY
    SCHED_LOCK();
    checkInvariants("sysThreadCreate before", 0);
    SCHED_UNLOCK();
#endif /* DEBUG_PRIORITY */

    if (stack_size == 0) {
        stack_size = 128 * 1024;
    }

    /*
     * Allocate thread private memory and stack.
     */
    if (!allocateContextAndStack(tid, stack_size)) {
	/* Exception thrown by caller if necessary. */
	return SYS_NOMEM;	/* Only thing that can go wrong */
    }
    /*
     * All threads are created in a suspened (non-runnable) state.
     */
    tid->state = SUSPENDED;
    green_sigprocmask(SIG_SETMASK, NULL, &MACHDEP(tid)->signalMask);
    MACHDEP(tid)->intrLockCount = 0;
    {
	ucontext_t uc0, uc1;
	start_args args;
 
        args.arg = (int)arg;
        args.oldContext = &uc0;
        args.tid = tid;
        args.ret = 0;

	initContext(&uc1, &tid->stack, (unsigned int)start,
		    (void (*)(void))start_func,
		    (unsigned int)&args);

	SCHED_LOCK();
#if defined(i386)
	green_getcontext(&uc0);
	if (args.ret == 0) {
	    setcontext(&uc1);
	    VM_CALL(panic)("setcontext returned -1 errno %d", errno);
	}
#else
	if (swapcontext(&uc0, &uc1) == -1) {
	    VM_CALL(panic)("swapcontext returned -1 errno %d", errno);
	}
#endif
	sysAssert(args.ret == 1);
	sysAssert(args.tid == tid);
	SCHED_UNLOCK();
    }


    tid->interrupted = FALSE;

    SYS_QUEUE_LOCK(sysThreadSelf());
    ActiveThreadCount++;        /* Global thread count */
    tid->next = ThreadQueue;    /* Chain all threads */
    ThreadQueue = tid;
    /* Record whether this thread is system or user */
    SYS_QUEUE_UNLOCK(sysThreadSelf());

#ifdef DEBUG_PRIORITY
    SCHED_LOCK();
    checkInvariants("sysThreadCreate after", 0);
    SCHED_UNLOCK();
#endif /* DEBUG_PRIORITY */

    *tidP = tid;
    return SYS_OK;
}

/*
 * Destroy the specified thread.  If a thread is destroying itself,
 * this function will never return.
 *
 * NOTE: there is a big assumption in this code that the current thread
 * can destroy itself.  I.e. there are some strong order dependencies
 * in the code.  E.g. the stack can be nuked, and the thread can keep
 * running.  This works under Unix because freed memory is still valid
 * memory.  May need to revisit.
 *
 * This function gets called on initial thread creation, when the
 * the thread returns from its initial function we destroy it.  We don't
 * want to actually destroy a thread Object, but we do want to clean up
 * all of its private state.
 *
 * We also want to check the waiters queue to see if anybody is
 * waiting on the thread to die.
 */

#ifdef DEBUG
static void threadReallyDead(sys_thread_t *tid);
#endif /* DEBUG */

/*
 * Doesn't do anything. greenThreadFree does the real work.
 */
int
sysThreadFree()
{
    return SYS_OK;
}

static void 
greenThreadExit(void)
{
    sys_thread_t *tid = sysThreadSelf();

#ifdef DEBUG_PRIORITY
    SCHED_LOCK();
    checkInvariants("greenThreadExit before", 0);
    SCHED_UNLOCK();
#endif /* DEBUG_PRIORITY */

    SYS_QUEUE_LOCK(sysThreadSelf());
    removeFromActiveQ(tid);
    SYS_QUEUE_UNLOCK(sysThreadSelf());

    /*
     * WARNING: DO NOT ATTEMPT TO ENTER ANY MONITORS AFTER THIS POINT
     * (unless you are reentering something you are *guaranteed* to
     * already own).
     */
    SCHED_LOCK();

    /*
     * Free the thread stack and sys_thread_t. We continue to use the
     * "freed" stack, but we don't actually reclaim the stack until
     * later, when we are guaranteed to be done with it.
     */
    deleteContextAndStack(tid);

#ifdef DEBUG_PRIORITY
    checkInvariants("greenThreadExit during", -1);
#endif /* DEBUG_PRIORITY */

    /*
     * As threads only exit now, the exiting thread can't be waiting
     * on a monitor, so we don't have to clean up the mon_wait monitor.
     */
    sysAssert(tid->mon_wait == 0);

    /* Remove the dying thread from threadAlarmQ */
    cleanupPendingAlarm(tid);

#ifdef DEBUG
    threadReallyDead(tid);
#endif /* DEBUG */

#ifdef DEBUG_PRIORITY
    checkInvariants("greenThreadExit done", -1);
#endif /* DEBUG_PRIORITY */

    reschedule();	 /* Call with SCHED_LOCKED() */
    /* NOTREACHED */
    VM_CALL(panic)("Return from greenThreadExit(): a zombie is awake!\n");
}

/*
 * Current thread yield control
 *
 * Green threads originally supported forcing another thread to yield...
 */
void
sysThreadYield()
{
    sys_thread_t *self = greenThreadSelf();
 
    SCHED_LOCK();
    if (runnable_queue && runnable_queue->priority == self->priority) {
#ifdef DEBUG_PRIORITY
        checkInvariants("sysThreadYield 1", -1);
#endif /* DEBUG_PRIORITY */
        YIELD();
    }
#ifdef DEBUG_PRIORITY
    checkInvariants("sysThreadYield 2", 0);
#endif /* DEBUG_PRIORITY */
    SCHED_UNLOCK();
    return;
}

/*
 * Get the scheduling priority of a specified thread
 */
int
sysThreadGetPriority(sys_thread_t *tid, int *priority)
{
    SCHED_LOCK();
    *priority = (tid->vmsuspended == 1) ? tid->vmoldPriority : tid->priority; 
    SCHED_UNLOCK();
    return SYS_OK;
}

/*
 * Really set the scheduling priority of a specified thread (locked)
 */
int
threadSetSchedulingPriority(sys_thread_t *tid, int priority)
{
    sys_thread_t *self = greenThreadSelf();
    while (tid->priority != priority) {
	if (tid == self) {
	    tid->priority = priority;
	} else {
	    gt_queue_t *queue;
	    /*
	     * The priority is changing, which means we have to reorder the
	     * queue that the thread is queued upon.
	     */
	    tid->priority = priority;
	    switch(tid->state) {
		case RUNNABLE:
		    queue = &runnable_queue;
		    break;
		case MONITOR_WAIT:
		    queue = &tid->mon_wait->monitor_waitq;
		    break;
		case CONDVAR_WAIT:
		    queue = &tid->mon_wait->condvar_waitq;
		    break;
		case MONITOR_SUSPENDED:
		    queue = &tid->mon_wait->suspend_waitq;
		    break;
		case SUSPENDED:
		    /* The thread is not on any queue when it is suspended. */
		    /* so we don't have to do anything, right, or not? */
		    queue = NULL;
		    break;
		default:
		    /* REMIND: Should throw an exception! */
		    VM_CALL(jio_fprintf)(stderr, "Tried to set priority of a thread "
			"in unknown state %d\n", tid->state);
		    VM_CALL(jio_fprintf)(stderr, "See %s, line %d\n", __FILE__, __LINE__);
		    VM_CALL(panic)("unknown thread state");
	    }
	    if (queue != NULL) {
		queueRemove(queue, tid);
		queueInsert(queue, tid);
	    }
	}

	if (tid->state == MONITOR_WAIT) {
	    sys_mon_t *mid = tid->mon_wait;
	    tid = mid->monitor_owner;

	    monitorRemoveInversion(mid, tid);
	    monitorAddInversion(mid);
	    priority = INVERSION_PRIORITY(tid);
	} else {
	    break;
	}
    }
    return self->state != RUNNABLE ||
	runnable_queue && runnable_queue->priority > self->priority;
}

static int
threadSetTruePriority(sys_thread_t *tid, int priority)
{
    if (tid->inversion_queue != NULL) {
	tid->oldPriority = priority;
	priority = INVERSION_PRIORITY(tid);
    }
    return threadSetSchedulingPriority(tid, priority);
}

/*
 * Set the scheduling priority of a specified thread
 */
int
sysThreadSetPriority(sys_thread_t *tid, int priority)
{
    SCHED_LOCK();
    if (tid->vmsuspended == 1) {
	/* when vmsuspended, just set the previously saved real priority */
	tid->vmoldPriority = priority;
    } else {
	if (threadSetTruePriority(tid, priority)) {
#ifdef DEBUG_PRIORITY
	    checkInvariants("sysThreadSetPriority 1", 1);
#endif /* DEBUG_PRIORITY */
	    YIELD();
        }
    }    
#ifdef DEBUG_PRIORITY
    checkInvariants("sysThreadSetPriority 2", 0);
#endif /* DEBUG_PRIORITY */
    SCHED_UNLOCK();
    return SYS_OK;
}

/*
 * If we computed the scheduling priority simply as the maximum of a
 * list of priorities, including the normal priority, inversion priorities,
 * and suspension priority, we could implement suspension by removing
 * the normal priority from the list (and unsuspension by restoring it to
 * the list.)  I decided that was overkill for now, so we still have
 * different places where priorities are stored, depending on if the thread
 * is undergoing priority inversion, VM suspension, or both.
 * DREL 9/25/96
 */
static int
threadVMSuspend(sys_thread_t *t)
{
    t->vmsuspended = 1;
    t->vmoldPriority =
	(t->inversion_queue == NULL) ? t->priority : t->oldPriority;
    {
	int resched = threadSetTruePriority(t, -1);
#ifdef DEBUG_PRIORITY
	checkInvariants("threadVMSuspend", resched);
#endif /* DEBUG_PRIORITY */
	return resched;
    }
}

/* Resume execution of the specified thread */
int
sysThreadResume(sys_thread_t *tid)
{
    sys_mon_t *mid;

    SCHED_LOCK();
    tid->pending_suspend = 0;

    switch(tid->state) {
	case SUSPENDED:
	    tid->state = RUNNABLE;
	    queueInsert(&runnable_queue, tid);
	    if (tid->priority > greenThreadSelf()->priority) {
#ifdef DEBUG_PRIORITY
		checkInvariants("sysThreadResume 1", 1);
#endif /* DEBUG_PRIORITY */
		YIELD();
	    }
	    break;
	    
	case MONITOR_SUSPENDED:
	    mid = tid->mon_wait;
	    sysAssert(mid != NULL);
	    sysAssert(!(mid->monitor_owner == SYS_THREAD_NULL &&
		mid->monitor_waitq != SYS_THREAD_NULL));
	    threadWakeup(tid);
	    break;
    }

#ifdef DEBUG_PRIORITY
    checkInvariants("sysThreadResume 2", 0);
#endif /* DEBUG_PRIORITY */
    SCHED_UNLOCK();
    return SYS_OK;
}

/* Suspend execution of the specified thread */
int
sysThreadSuspend(sys_thread_t *tid)
{
    int err = 0;
    sys_thread_t *self = greenThreadSelf();

    SCHED_LOCK();

    if (tid == self) {
	self->state = SUSPENDED;
#ifdef DEBUG_PRIORITY
    checkInvariants("sysThreadSuspend 1", 1);
#endif /* DEBUG_PRIORITY */
	yieldContext(CONTEXT(self));
    } else {
	switch(tid->state) {
	    case RUNNABLE:
		queueRemove(&runnable_queue, tid);
		tid->state = SUSPENDED;
		break;

	    case MONITOR_WAIT:
	    case CONDVAR_WAIT:
		tid->pending_suspend = 1;
		break;

	    case SUSPENDED:
	    case MONITOR_SUSPENDED:
		err = -1;
		break;
	}
    }

#ifdef DEBUG_PRIORITY
    checkInvariants("sysThreadSuspend 2", 0);
#endif /* DEBUG_PRIORITY */
    SCHED_UNLOCK();
    return (err == 0 ? SYS_OK : SYS_ERR);
}

/*
 * The next set of global variables and the two routines that follow
 * are used by routines like the garbage collector to lock out other
 * threads for some period of time.  See the extensive comments in
 * interrupt_md.c for a complete explanation.
 */

/*
 * Stop all threads other than the current one.  The stopped threads
 * may be restarted with sysThreadMulti(); the operation of this
 * routine is atomic; it either stops all java threads or it stops
 * none of them.  Upon success (all threads stopped) this routine
 * returns SYS_OK, otherwise SYS_ERR.
 *
 * In general, sysThreadSingle() should take care of anything below
 * the HPI that needs to be done to safely run single-threaded.
 */
int
sysThreadSingle()
{
    SCHED_LOCK();

    return SYS_OK;
}

/*
 * Allow multi threaded execution to resume after a
 * sysThreadSingle() call.
 *
 * Note: When this routine is called the scheduler should already
 * have been locked by sysThreadSingle().  Each sysThreadSingle()
 * call should have a matching sysThreadMulti().
 */
void sysThreadMulti() 
{
    sysAssert(SCHED_LOCKED());

    SCHED_UNLOCK();
}

/* Set the current thread
 */
sys_thread_t *_CurrentThread;
 
/*
 * Enumerate over all threads, calling a function for each one.  A
 * paranoid helper function would be prepared to deal with threads
 * that have not been created by Java.
 */
int
sysThreadEnumerateOver(int (*func)(sys_thread_t *, void *), void *arg)
{
    sys_thread_t *tid;
    int err = SYS_OK;
    int i;

    sysAssert(SYS_QUEUE_LOCKED(sysThreadSelf()));

    tid = ThreadQueue;
    for (i = 0; i < ActiveThreadCount && tid != 0; i++) {
        /* skip idle thread */
        if (tid != idle_thread && tid != clock_thread && tid != slicer_thread) {
	    if ((err = (*func)(tid, arg)) != SYS_OK) {
	        break;
	    }
	}
        tid = tid->next;
    }

    return err;
}

void *
sysThreadNativeID(sys_thread_t *t)
{
    return NULL;
}

/*
 * Miscellaneous stuff in support of HPI for Green threads
 */
void
setCurrentThread(sys_thread_t *tid)
{
    _CurrentThread = tid;
}

/*
 * Remove this thread from the list of Active threads.
 */
static void
removeFromActiveQ(sys_thread_t * t) 
{
    sys_thread_t *prev;
    sys_thread_t *tid;

    sysAssert(SYS_QUEUE_LOCKED(sysThreadSelf()));

    ActiveThreadCount--;

    prev = 0;
    tid = ThreadQueue;
    while (tid) {
	if (tid == t) {
	    if (prev) {
		prev->next = tid->next;
	    } else {
		ThreadQueue = tid->next;
	    }
	    tid->next = 0;
	    break;
	}
	prev = tid;
	tid = tid->next;
    }
}

#ifdef DEBUG

/*
 * Diagnostic routines:
 *
 * If a thread is really dead should not appear in:
 *
 *  - the thread queue
 *  - the alarm queue
 *  - any monitor's waitq, condvar_waitq or suspend_waitq
 *  - the runnable_queue
 *
 * Note that this can only be tested between the time that a thread
 * is cleaned up after and the time that its stack is deleted!  This
 * makes it applicable in only a very small window in greenThreadExit(),
 * after which point the thread's sys_thread_t is deallocated.  Note
 * also that the thread may be checking itself, and so it is valid for
 * the thread being checked to be the current thread or to have certain
 * monitors at certain times, like _registry_lock when enumerating
 * registered monitors.
 */
static void
checkNotInQueue(sys_thread_t *tid, sys_thread_t *t)
{
    for (; t != 0; t = t->waitq) {
	sysAssert(t != tid);
    }
}

static void
threadNotOnThreadQueue(sys_thread_t *tid)
{
    sys_thread_t *t;

    /* Check the thread queue */
    t = ThreadQueue;
    while (t) {
	sysAssert(t != tid);
	t = t->next;
    }
}

static void
threadNotOnRunQueue(sys_thread_t *tid)
{
    checkNotInQueue(tid, runnable_queue);
}

static void
threadNotOnAlarmQueue(sys_thread_t *tid)
{
    checkNotInQueue(tid, threadAlarmQ);
}

static void
threadReallyDead(sys_thread_t *tid)
{
    sysAssert(SCHED_LOCKED());
    threadNotOnThreadQueue(tid);
    threadNotOnAlarmQueue(tid);
    threadNotOnRunQueue(tid);
}

#endif /* DEBUG */

/*
 * The mechanics of actually signalling an exception (in the future,
 * and Alarm or Interrupt) depend upon what thread implementation you
 * are using.
 */
void
sysThreadPostException(sys_thread_t *tid, void *exc)
{
    SCHED_LOCK();
    if (tid->state == CONDVAR_WAIT) {
	threadWakeup(tid);
    } else {
	/* XXX We may need to do something here XXX */
    }

#ifdef DEBUG_PRIORITY
    checkInvariants("sysThreadPostException", 0);
#endif /* DEBUG_PRIORITY */

    SCHED_UNLOCK();
}

void
threadWakeup(sys_thread_t *tid)
{
    sys_thread_t *self = greenThreadSelf();
    sys_mon_t *mid = tid->mon_wait;

    gt_queue_t *q = (tid->state == CONDVAR_WAIT)
        ? &mid->condvar_waitq : &mid->suspend_waitq;
    queueRemove(q, tid);
    queueInsert(&mid->monitor_waitq, tid);
    tid->state = MONITOR_WAIT;
    Log3(4, "0x%8x wakes up sys_mon_t 0x%x 0x%8x\n", self, mid, tid);
    if (mid->monitor_owner == SYS_THREAD_NULL) {
	Log3(4, "0x%8x signals sys_mon_t 0x%x 0x%8x\n", self, mid, tid);
	(void) queueSignal(mid, &mid->monitor_waitq);
    } else {
	(void) monitorApplyInversion(mid);
    }
}

/*
 * Support for (Java-level) interrupts.
 */
void
sysThreadInterrupt(sys_thread_t *tid)
{
    SCHED_LOCK();
    tid->interrupted = TRUE;
    if (tid->state == CONDVAR_WAIT) {
	threadWakeup(tid);
    } else {
	/* XXX We may need to do something here XXX */
    }

#ifdef DEBUG_PRIORITY
    checkInvariants("sysThreadInterrupt", 0);
#endif /* DEBUG_PRIORITY */

    SCHED_UNLOCK();
}

/*
 * Atomically read and clear the boolean flag depending on the
 * value of ClearInterrupted that's passed in.
 */
int
sysThreadIsInterrupted(sys_thread_t *tid, int ClearInterrupted)
{
    bool_t interrupted;

    SCHED_LOCK();
    interrupted = tid->interrupted;
    if (ClearInterrupted == 1) {
	tid->interrupted = FALSE;
    }   /* Otherwise leave it alone */
    SCHED_UNLOCK();
    return interrupted;
}

static jlong sliceStartTime = -1;

jlong
sysThreadCPUTime()
{
	sys_thread_t * t = greenThreadSelf();
	jlong time = gethrvtime();
	if (sliceStartTime == -1)
	    sliceStartTime = time;
	time += t->elapsedCPUTime - sliceStartTime;
	return time;
}

void
updateSliceStart()
{
	sys_thread_t * t = greenThreadSelf();
	jlong time = gethrvtime();
	if (sliceStartTime == -1)
	    sliceStartTime = time;
        t->elapsedCPUTime +=  time - sliceStartTime;
	sliceStartTime = time;
}

int
sysThreadGetStatus(sys_thread_t *tid, sys_mon_t **monitorPtr)
{
    int status;
    switch (tid->state) {
      case RUNNABLE:
	  status = SYS_THREAD_RUNNABLE;
          break;
      case SUSPENDED:
	  status = SYS_THREAD_SUSPENDED;
          break;
      case MONITOR_SUSPENDED:
	  status = SYS_THREAD_SUSPENDED | SYS_THREAD_MONITOR_WAIT;
          break;
      case MONITOR_WAIT:
	  if (tid->pending_suspend)
	      status = SYS_THREAD_SUSPENDED | SYS_THREAD_MONITOR_WAIT;
	  else
	      status = SYS_THREAD_MONITOR_WAIT;
          break;
      case CONDVAR_WAIT:
	  if (tid->pending_suspend)
	      status = SYS_THREAD_SUSPENDED | SYS_THREAD_CONDVAR_WAIT;
	  else
	      status = SYS_THREAD_CONDVAR_WAIT;
          break;
      default:
	return SYS_ERR;
    }
    if (monitorPtr) {
        if (status & SYS_THREAD_MONITOR_WAIT) {
            *monitorPtr = tid->mon_enter;
        } else if (status & SYS_THREAD_CONDVAR_WAIT) {
            *monitorPtr = tid->mon_wait;
        } else {
            *monitorPtr = NULL;
        }
    }
    return status;
}

sys_thread_t *sysThreadSelf()
{
    return greenThreadSelf();
}

HPI_SysInfo *
sysGetSysInfo()
{
    static HPI_SysInfo info = {"green threads", FALSE};
    return &info;
}

/*
 * Idle thread
 */

static void
idle_loop(void *p)
{
    while (1) {
        sysAssert(runnable_queue == NULL);
	pause();	/* Suspend idle thread */
    }
}

/*
 * Create the idle thread.
 */
static void
InitializeIdle(void)
{
    sysThreadCreate(&idle_thread, 0, idle_loop, 0);
    sysThreadSetPriority(idle_thread, MinimumPriority - 1);
    sysThreadResume(idle_thread);
}

/*
 * The time slicer thread is Solaris Green threads-only, and is currently
 * only run if the -ts flag is used when the runtime is started.  The time
 * slicer runs at priority MaximumPriority + 1, which is greater than any
 * user thread but less than the clock thread.  That way it can interrupt
 * any user thread but won't get in the way of the clock.  The time slicer
 * simply sets a timeout the size of the slice value ahead, then waits.
 * When it wakes up it is guaranteed to be scheduled, and in doing so will
 * cause any running user threads to yield and be rescheduled round-robin.
 */

static long TimeSlice = 0;	/* Millis per time slice; 0 means off */

static sys_mon_t *_slicer_lock;
#define SLICER_LOCK_INIT() \
     VM_CALL(monitorRegister)(_slicer_lock, "Time slicer lock")
#define SLICER_LOCK(self) sysMonitorEnter(self, _slicer_lock)
#define SLICER_WAIT(self, millis) sysMonitorWait(self, _slicer_lock, millis)

sys_thread_t *slicer_thread;

static void
timeSlicer(void *p)
{
    sys_thread_t *self = sysThreadSelf();
    _slicer_lock = (sys_mon_t *) sysMalloc(sysMonitorSizeof());
    SLICER_LOCK_INIT();
    /*
     * If the timeSlicer is run, then the TimeSlice value must have
     * been set to a positive value, the number of millis per slice.
     */
    sysAssert(TimeSlice != 0);

    Log1(3, "Time slicer started with slice value %d", TimeSlice);

    SLICER_LOCK(self);
    while (1) {
	/* Just wait and wake, wait and wake... */
	SLICER_WAIT(self, TimeSlice);
	Log1(3, "Time slicer woke up at %d\n", (int)(sysTimeMillis()));
    }

    /* If we ever got here: SLICER_UNLOCK(); */
}

static void
InitializeTimeSlicer(void)
{
    static bool_t timeSlicerCreated = FALSE;
    if (timeSlicerCreated) {
        return;
    }
    timeSlicerCreated = TRUE;
    sysThreadCreate(&slicer_thread, 0, timeSlicer, 0);
    sysThreadSetPriority(slicer_thread, MaximumPriority + 1);
    sysThreadResume(slicer_thread);
}

int sysAdjustTimeSlice(int newval)
{
    int oldTimeSlice = TimeSlice;
    if (newval) {
        TimeSlice = newval;
	InitializeTimeSlicer();
    }
    return oldTimeSlice;
}

/*
 * FIXME: Someone who knows more about green threads, please implement
 * a pair of real thread suspend/resume routines. These functions should
 * work independently with sysThreadSuspend/Resume, i.e., sysThreadProfResume
 * should not resume a thread suspended by sysThreadSuspend, and
 * vise-versa.
 */
void sysThreadProfSuspend(sys_thread_t *tid)
{
    SCHED_LOCK();
}

void sysThreadProfResume(sys_thread_t *tid)
{
    SCHED_UNLOCK();
}

bool_t sysThreadIsRunning(sys_thread_t *tid)
{
    unsigned int sum = 0;
    unsigned int *p;
    jmp_buf *context = &(CONTEXT(tid)->env);

    for (p = (unsigned int *)context;
	 p < (unsigned int *)((char *)context + sizeof(jmp_buf));
	 p++) {
        sum += *p;
    }

    if (sum == tid->last_sum) {
        return FALSE;
    }
    tid->last_sum = sum;
    return TRUE;
}

void * sysThreadInterruptEvent()
{
    return NULL;
}

