/*
 * @(#)condvar_md.c	1.25 98/09/16
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Condition variable HPI implementation for Solaris
 */

#include <errno.h>
#include <time.h>
#include <setjmp.h>
#include <signal.h>
#include <limits.h>

#include "hpi_impl.h"

#include "condvar_md.h"
#include "mutex_md.h"
#include "threads_md.h"

int
condvarInit(condvar_t *condvar)
{
    int err;

#ifdef USE_PTHREADS
    err = pthread_cond_init(&condvar->cond, NULL);
#else
    err = cond_init(&condvar->cond, USYNC_THREAD, 0 /* ignored */);
#endif
    condvar->counter = 0;
    return (err == 0 ? SYS_OK : SYS_ERR);
}

int
condvarDestroy(condvar_t *condvar)
{
    int err;

    err = cond_destroy((cond_t *) condvar);
    return (err == 0 ? SYS_OK : SYS_ERR);
}

int
condvarWait(condvar_t *condvar, mutex_t *mutex, thread_state_t wtype)
{
    sigjmp_buf jmpbuf;
    int err;

    sys_thread_t *self = sysThreadSelf();
    /*
     * There is no threads interface to get a thread's state. So, instead,
     * we use this hack so that the debugger agent can get at this thread's
     * state. Of course, this is not very reliable, but when a thread goes
     * to sleep, it *will* be reported as sleeping. During the transition
     * from running to sleep, it may be incorrectly reported, since the
     * setting of the state here is not atomic with the voluntary sleep.
     * The better fix is to extend the Solaris threads interface and have
     * the debugger agent call this interface OR to use libthread_db for
     * intra-process state reporting.
     *
     * Now, condition variables are used either for waiting to enter a
     * monitor (MONITOR_WAIT) or to execute a "wait()" method when already
     * holding a monitor (CONDVAR_WAIT). So, when condvarWait() is called
     * it could be to wait for a monitor or for a condition within a
     * monitor.  This is indicated by the "wtype" argument to condvarWait().
     * This type is set in the thread state before going to sleep.
     */
    self->state = wtype;

    thr_setspecific(sigusr1Jmpbufkey, &jmpbuf);
    if (sigsetjmp(jmpbuf, 1) == 0) {
	sigset_t osigset;

	thr_sigsetmask(SIG_UNBLOCK, &sigusr1Mask, &osigset);
again:
	err = cond_wait((cond_t *) condvar, (mutex_t *) mutex);
	switch(err) {
	case 0:
	    err = SYS_OK;
	    break;
	case EINTR: /* Signals other than USR1 were received. */
	    goto again;
	default:
	    err = SYS_ERR;
	}
	thr_sigsetmask(SIG_SETMASK, &osigset, NULL);
    } else {
	/*
	 * we've received a SIGUSR1 to interrupt our wait. We just return
	 * and something above use notices the change.
	 * clear the jump buf just to be paranoid.
	 */
	 thr_setspecific(sigusr1Jmpbufkey, NULL);
	 err = SYS_INTRPT;
    }
    /*
     * After having woken up, change the thread state to RUNNABLE, since
     * it is now runnable.
     */
    self->state = RUNNABLE;

    return err;
}


/*
 * Returns 0 if condition variable became true before timeout expired.
 * Returns 1 if timeout expired first.
 * Returns <0 if wait fails for any other reason.
 */
int
condvarTimedWait(condvar_t *condvar, mutex_t *mutex,
    jlong millis, thread_state_t wtype)
{
    sigjmp_buf jmpbuf;
    int err;
    struct timespec timeout;
    sys_thread_t *self;
    jlong end_time;
    
    if (millis < 0)
	return SYS_ERR;

    if (millis > (jlong)INT_MAX) {
	return condvarWait(condvar, mutex, wtype);
    }

    end_time = sysTimeMillis() + millis;

    self = sysThreadSelf();
    self->state = wtype;

    thr_setspecific(sigusr1Jmpbufkey, &jmpbuf);
    if (sigsetjmp(jmpbuf, 1) == 0) {
	sigset_t osigset;

	thr_sigsetmask(SIG_UNBLOCK, &sigusr1Mask, &osigset);

    again:
	timeout.tv_sec = end_time / 1000;
	timeout.tv_nsec = (end_time % 1000) * 1000000;
	err = cond_timedwait((cond_t *)condvar, (mutex_t *)mutex, &timeout);
	switch(err) {
	case 0:
	    err = SYS_OK;
	    break;
	case EINTR: /* Signals other than USR1 were received. */
	    if (sysTimeMillis() < end_time) {
	        goto again;
	    }
	    /*FALLTHRU*/
#ifdef USE_PTHREADS
	case ETIMEDOUT:
#else
	case ETIME:
#endif
	    err = SYS_TIMEOUT;
	    break;
	default:
	    err = SYS_ERR;
	}
	thr_sigsetmask(SIG_SETMASK, &osigset, NULL);
    } else {
	/*
	 * we've received a SIGUSR1 to interrupt our wait. We just return
	 * and something above use notices the change.
	 * clear the jump buf just to be paranoid.
	 */
	 thr_setspecific(sigusr1Jmpbufkey, NULL);
	 err = SYS_INTRPT;
    }
    self->state = RUNNABLE;
    return err;
}

int
condvarSignal(condvar_t *condvar)
{
    int err;

    err = cond_signal((cond_t *) condvar);
    condvar->counter++;
    return (err == 0 ? SYS_OK : SYS_ERR);
}

int
condvarBroadcast(condvar_t *condvar)
{
    int err;

    err = cond_broadcast((cond_t *) condvar);
    condvar->counter++;
    return (err == 0 ? SYS_OK : SYS_ERR);
}
