/*
 * @(#)system_md.c	1.65 98/08/26
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "hpi_impl.h"

#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h> /* timeval */
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <netdb.h>
#include <limits.h>
#include <errno.h>

#include <dlfcn.h>

#include "jni_md.h"
#include "mutex_md.h"

#ifndef NATIVE
extern void InitializeSbrk(void);
extern void InitializeAsyncIO(void);
extern void InitializeHelperThreads(void);
#endif /* NATIVE */
extern void InitializeMem(void);

#include "interrupt.h"
#include "threads_md.h"
#include "monitor_md.h"
#include "largefile.h"


#if !defined(_LFS_LARGEFILE) || !_LFS_LARGEFILE

#ifdef __GLIBC__
/* Doesn't matter what these are, there is no 64 bit support. */
typedef int u_longlong_t;
typedef int longlong_t;
typedef int timestruc_t;
#define _ST_FSTYPSZ 1
#endif /* __GLIBC__ */

/*
 * The stat64 structure must be provided on systems without large file
 * support (e.g., Solaris 2.5.1).  These definitions are from Solaris 2.6
 * sys/stat.h and sys/types.h.
 */

typedef u_longlong_t	ino64_t;	/* expanded inode type	*/
typedef longlong_t	blkcnt64_t;	/* count of file blocks */

struct	stat64 {
	dev_t	st_dev;
	long	st_pad1[3];
	ino64_t	st_ino;
	mode_t	st_mode;
	nlink_t st_nlink;
	uid_t 	st_uid;
	gid_t 	st_gid;
	dev_t	st_rdev;
	long	st_pad2[2];
	off64_t	st_size;
	timestruc_t st_atim;
	timestruc_t st_mtim;
	timestruc_t st_ctim;
	long	st_blksize;
	blkcnt64_t st_blocks;
	char	st_fstype[_ST_FSTYPSZ];
	long	st_pad4[8];
};

#define O_LARGEFILE     0x2000  /* Solaris 2.6 sys/fcntl.h */

#endif  /* !_LFS_LARGEFILE */


int sysThreadBootstrap(sys_thread_t **tidP, sys_mon_t **lockP, int nb)
{
    threadBootstrapMD(tidP, lockP, nb);

    intrInit();

#ifndef NATIVE
    /* Initialize the special case for sbrk on Solaris (see synch.c) */
    InitializeSbrk();
    /* Initialize the async io */
    InitializeAsyncIO();
    InitializeMem();
    /* Initialize Clock and Idle threads */
    InitializeHelperThreads();
#else /* if NATIVE */
    initializeContentionCountMutex();
    InitializeMem();
#endif /* NATIVE */

    return SYS_OK;
}

int sysShutdown()
{
    return SYS_OK;
}

long
sysGetMilliTicks()
{
    struct timeval tv;

    (void) gettimeofday(&tv, (void *) 0);
    return((tv.tv_sec * 1000) + (tv.tv_usec / 1000));
}

jlong
sysTimeMillis()
{
    struct timeval t;
    gettimeofday(&t, 0);
    return ((jlong)t.tv_sec) * 1000 + (jlong)(t.tv_usec/1000);
}

int
sysGetLastErrorString(char *buf, int len)
{
    if (errno == 0) {
	return 0;
    } else {
	const char *s = strerror(errno);
	int n = strlen(s);
	if (n >= len) n = len - 1;
	strncpy(buf, s, n);
	buf[n] = '\0';
	return n;
    }
}

/*
 * File system
 *
 * These are all the sys API which implement the straight POSIX
 * API. Those that do not are defined by thread-specific files
 * (i.e. io_md.c)
 */

int sysOpen(const char *path, int oflag, int mode)
{
    return open64_w(path, oflag, mode);
}

char *sysNativePath(char *path)
{
    return path;
}

/*
 * Support for 64 bit File I/O.
 */

/* 64 bit versions of low-level File I/O routines */

typedef int (*FSTAT64)(int fd, struct stat64 *);
typedef jlong (*LSEEK64)(int, jlong, int);
typedef int (*FTRUNCATE64)(int, jlong);
typedef int (*OPEN64)(const char *path, int oflag, mode_t mode);

static FSTAT64 fstat64_ptr = NULL;
static LSEEK64 lseek64_ptr = NULL;
static FTRUNCATE64 ftruncate64_ptr = NULL;
static OPEN64 open64_ptr = NULL;


/*
 * 64-bit IO initialization routine.  The #pragma makes this routine get called
 * when this code (shared library) is loaded.  Find the pointers to these libc
 * routines if they exist.
 */

static void init64IO(void) {
    fstat64_ptr = (FSTAT64) dlsym(RTLD_NEXT, "_fstat64");
    /* Look for others only if the first one succeeded */
    if (fstat64_ptr != NULL) {
	lseek64_ptr = (LSEEK64) dlsym(RTLD_NEXT, "_lseek64");
	ftruncate64_ptr = (FTRUNCATE64) dlsym(RTLD_NEXT, "_ftruncate64");
	open64_ptr = (OPEN64) dlsym(RTLD_NEXT, "_open64");
    }
}
#pragma init(init64IO)

int
sysFileSizeFD(int fd, jlong *size)
{
    int ret;

    if (fstat64_ptr != 0) {
        struct stat64 buf64;
	ret = (*fstat64_ptr)(fd, &buf64);
	*size = buf64.st_size;
    } else {
        struct stat buf;
	ret = fstat(fd, &buf);
	*size = buf.st_size;
    }
    return ret;
}

int
sysFfileMode(int fd, int *mode)
{
    int ret;
    
    if (fstat64_ptr != 0) {
        struct stat64 buf64;
	ret = (*fstat64_ptr)(fd, &buf64);
	(*mode) = buf64.st_mode;
    } else {
        struct stat buf;
	ret = fstat(fd, &buf);
	(*mode) = buf.st_mode;
    }
    return ret;
}

int
sysFileType(const char *path)
{
    int ret;
    struct stat buf;

    if ((ret = stat(path, &buf)) == 0) {
      mode_t mode = buf.st_mode & S_IFMT;
      if (mode == S_IFREG) return SYS_FILETYPE_REGULAR;
      if (mode == S_IFDIR) return SYS_FILETYPE_DIRECTORY;
      return SYS_FILETYPE_OTHER;
    }
    return ret;
}

/* 
 * Wrapper functions for low-level I/O routines - use the 64 bit
 * version if available, else revert to the 32 bit versions.
 */

off64_t 
lseek64_w(int fd, off64_t offset, int whence)
{
    if (lseek64_ptr != 0) {
        return (*lseek64_ptr)(fd, offset, whence);
    } else {
        if ((off_t) offset != offset) {
	    errno = EOVERFLOW;
	    return -1;
	}
	return lseek(fd, offset, whence);
    }
}

int 
ftruncate64_w(int fd, off64_t length) 
{
    if (ftruncate64_ptr != 0) {
        return (*ftruncate64_ptr)(fd, length);
    }
    if ((off_t)length != length) {
	errno = EOVERFLOW;
        return -1;
    }
    return ftruncate(fd, length);
}

int 
open64_w(const char *path, int oflag, int mode)
{
    int result;

    if (open64_ptr != 0) {
        /* 
	 * Using O_LARGEFILE with _open is the same as _open64.
	 * We use this because on green threads we have our own version
	 * of open, but no _open64.  So, rather than implement _open64
	 * we use O_LARGEFILE.
	 */
        result = open(path, oflag | O_LARGEFILE, mode);
    } else {
	result = open(path, oflag, mode);
    }
    if (result != -1) {
	/* If the open succeeded, the file might still be a directory */
	struct stat stat_buf;
	if (fstat(result, &stat_buf) != -1) {
	    if ((stat_buf.st_mode & S_IFMT) == S_IFDIR) {
		errno = EISDIR;
		close(result);
		return -1;
	    }
	} else {
	    close(result);
	    return -1;
	}
    } 
    return result;
}  

/*
 * Network APIs that are used without the green threads redirection
 * or native threads interruptible I/O. Use caution about where you
 * use these.
 */
struct hostent *sysGetHostByName(char *hostname)
{
    return gethostbyname(hostname);
}

void setFPMode(void)
{
#ifdef i386
    asm("	pushl $575");
    asm("	fldcw (%esp)");
    asm("	popl %eax");
#endif
}
