/*
 * @(#)signals_md.c	1.6 98/08/23
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Solaris 2.X dependent interrupt handling code.
 */


/*
 * Header files.
 */
#include <stdio.h>
#include <signal.h>
#include <ucontext.h>
#include <sys/trap.h>
#include <stdlib.h>
#include <string.h>

#include "sys_api.h"

#include "oobj.h"
#include "interpreter.h"
#include "threads.h"

static char *signalNames[] = {
    "no signal",
    "SIGHUP    1    hangup",
    "SIGINT    2    interrupt",
    "SIGQUIT   3*   quit",
    "SIGILL    4*   illegal instruction",
    "SIGTRAP   5*   trace trap",
    "SIGABRT   6*   abort (generated by abort(3) routine)",
    "SIGEMT    7*   emulator trap",
    "SIGFPE    8*   arithmetic exception",
    "SIGKILL   9    kill (cannot be caught, blocked, or ignored)",
    "SIGBUS    10*  bus error",
    "SIGSEGV   11*  segmentation violation",
    "SIGSYS    12*  bad argument to system call",
    "SIGPIPE   13   write on a pipe or other socket with no one to read it",
    "SIGALRM   14   alarm clock",
    "SIGTERM   15   software termination signal",
    "SIGUSR1   16   user defined signal 1",
    "SIGUSR2   17   user defined signal 2",
    "SIGCHLD   18@  child status change",
    "SIGPWR    19   power-fail restart",
    "SIGWINCH  20@  window changed (see termio(4) and win(4S))",
    "SIGURG    21@  urgent condition present on socket",
    "SIGPOLL   22   pollable event occurred (also SIGIO)",
    "SIGSTOP   23  stop (cannot be caught, blocked, or ignored)",
    "SIGTSTP   24+  stop signal generated from keyboard",
    "SIGCONT   25@  continue after stop",
    "SIGTTIN   26+  background read attempted from control terminal",
    "SIGTTOU   27+  background write attempted to control terminal",
    "SIGVTALRM 28   virtual time alarm (see getitimer(2))",
    "SIGPROF   29   profiling timer alarm (see getitimer(2))",
    "SIGXCPU   30   cpu time limit exceeded (see getrlimit(2))",
    "SIGXFSZ   31   file size limit exceeded (see getrlimit(2))",
    
    "SIGWAITING 32  process' lwps are blocked",
    "SIGLWP    33   private to thread library",
    "SIGFREEZE 34   private to checkpoint/resume",
    "SIGTHAW   35   private to checkpoint/resume",
    "SIGRTMIN  36   first (high-priority) realtime signal",
    "SIGRTMAX  43   last (lowest-priority) realtime signal"
};

/*
 * Decode a siginfo_t structure, and print out something interesting
 * about it.  Truncate and succeed if the buffer write overflows.  It's
 * not crucial that this does exactly the right thing.
 */
static char *
decode_si_code(int sig, siginfo_t *info)
{
    char *unknown = "unknown siginfo";
#define MSGMAXLEN 1024
    static char msg[MSGMAXLEN];
    int si_code = info->si_code;

    if (si_code == SI_NOINFO)
	return "no information";
    else if (SI_FROMUSER(info)) {
	char *si_user[] = {
	    "SI_USER", "SI_LWP", "SI_QUEUE",
	    "SI_TIMER", "SI_ASYNCIO", "SI_MESGQ"
	};
	if (si_code < SI_MESGQ || si_code > 0) {
	    return unknown;
	}
	si_code = abs(si_code);
	jio_snprintf(msg, sizeof(msg), "%s [pid: %ld, uid: %ld]",
		     si_user[si_code], (long)info->si_pid, (long)info->si_uid);
	return msg;
    } else {	/* SI_FROMKERNEL */
	switch(sig) {
	    case SIGILL: {
		char *si_code_name[] = {
		    "ILL_ILLOPC", "ILL_ILLOPN", "ILL_ILLADR",
		    "ILL_ILLTRP", "ILL_PRVOPC", "ILL_PRVREG",
		    "ILL_COPROC", "ILL_BADSTK"
		};
		if (si_code > NSIGILL || si_code == 0) {
		    return unknown;
		}
		jio_snprintf(msg, sizeof(msg), "%s [addr: 0x%p]\n",
			     si_code_name[si_code-1], info->si_addr);
		break;
	    }
	    case SIGFPE: {
		char *si_code_name[] = {
		    "FPE_INTDIV", "FPE_INTOVF", "FPE_FLTDIV",
		    "FPE_FLTOVF", "FPE_FLTUND", "FPE_FLTRES",
		    "FPE_FLTINV", "FPE_FLTSUB"
		};
		if (si_code > NSIGFPE || si_code == 0) {
		    return unknown;
		}
		jio_snprintf(msg, sizeof(msg), "%s [addr: 0x%p]\n",
			     si_code_name[si_code-1], info->si_addr);
		break;
	    }
	    case SIGSEGV: {
		char *si_code_name[] = {
		    "SEGV_MAPERR", "SEGV_ACCERR"
		};
		if (si_code > NSIGSEGV || si_code == 0) {
		    return unknown;
		}
		jio_snprintf(msg, sizeof(msg), "%s [addr: 0x%p]\n",
			     si_code_name[si_code-1], info->si_addr);
		break;
	    }
	    case SIGBUS: {
		char *si_code_name[] = {
		    "BUS_ADRALN", "BUS_ADRERR", "BUS_OBJERR"
		};
		if (si_code > NSIGBUS || si_code == 0) {
		    return unknown;
		}
		jio_snprintf(msg, sizeof(msg), "%s [addr: 0x%p]\n",
			     si_code_name[si_code-1], info->si_addr);
		break;
	    }
	    case SIGTRAP: {
		char *si_code_name[] = {
		    "TRAP_BRKPT", "TRAP_TRACE"
		};
		if (si_code > NSIGTRAP || si_code == 0) {
		    return unknown;
		}
		jio_snprintf(msg, sizeof(msg), "%p\n",
			     si_code_name[si_code-1]);
		break;
	    }
	    case SIGCHLD: {
		char *si_code_name[] = {
		    "CLD_EXITED", "CLD_KILLED", "CLD_DUMPED",
		    "CLD_TRAPPED", "CLD_STOPPED", "CLD_CONTINUED"
		};
		if (si_code > NSIGCLD || si_code ==  0) {
		    return unknown;
		}
		jio_snprintf(msg, sizeof(msg), "%s [pid: %ld, status: %ld]\n",
			     si_code_name[si_code-1], (long)info->si_pid,
			     (long)info->si_status);
		break;
	    }
	    case SIGPOLL: {
		char *si_code_name[] = {
		    "POLL_IN", "POLL_OUT", "POLL_MSG",
		    "POLL_ERR", "POLL_PRI", "POLL_HUP"
		};
		if (si_code > NSIGPOLL || si_code == 0) {
		    return unknown;
		}
		jio_snprintf(msg, sizeof(msg), "%s [band: %ld]\n",
			     si_code_name[si_code-1], (long)info->si_band);
		break;
	    }
	    default:
		return unknown;
	}
	return msg;
    }
}

/* There is a race condition here. We use a global variable to store the
 * thread ID that raised SIGQUIT. This is no big deal because SIGQUIT is
 * only used for informational purposes.
 */
static sys_thread_t *sigquit_thread = NULL;

static void UserHandler(int sig, void *siginfo, void *uc)
{
    if (sig == SIGQUIT) {
        sigquit_thread = sysThreadSelf();
    }

    sysSignalNotify(sig);

    /* We need to reinstate the signal handler each time... */
    sysSignal(sig, UserHandler);
}

static void
panicHandler(int sig, void *siginfo, void *uc)
{
    siginfo_t *info = siginfo;
    static int count = 0;

    if (count > 0) {
	Abort();
    }

#if defined(i386) && !defined(DEBUG)
    if (sig == SIGFPE && info->si_code == FPE_INTDIV) {
	/*
	 * instruction locations inside assembly
	 * interp for divide trapping
	 */
	extern int __idiv_inst5(void);
	extern int __idiv_trap5(void);
	extern int __idiv_inst6(void);
	extern int __idiv_trap6(void);
	extern int __irem_inst5(void);
	extern int __irem_trap5(void);
	extern int __irem_inst6(void);
	extern int __irem_trap6(void);
	int pc = ((ucontext_t*)uc)->uc_mcontext.gregs[PC];
	if (pc == (int) &__idiv_inst5) {
	    ((ucontext_t*)uc)->uc_mcontext.gregs[PC] = (int) &__idiv_trap5;
	    goto done;
	} else if (pc == (int) &__idiv_inst6) {
	    ((ucontext_t*)uc)->uc_mcontext.gregs[PC] = (int) &__idiv_trap6;
	    goto done;
	} else if (pc == (int) &__irem_inst5) {
	    ((ucontext_t*)uc)->uc_mcontext.gregs[PC] = (int) &__irem_trap5;
	    goto done;
	} else if (pc == (int) &__irem_inst6) {
	    ((ucontext_t*)uc)->uc_mcontext.gregs[PC] = (int) &__irem_trap6;
	    goto done;
	}
    }
#endif i386

    if (CompiledCodeSignalHandler(sig, info, uc) == TRUE) {
	goto done;
    }

    count += 1;

    if (sig > MAXSIG) {
	jio_fprintf(stderr, "Unknown signal: %d\n", sig);
    } else {
	jio_fprintf(stderr, "%s\n", signalNames[sig]);
    }

    if (info) {
	/* print out signal info */
	if (info->si_code != SI_NOINFO) {
	    jio_fprintf(stderr, "    si_signo [%d]: %s\n",
			info->si_signo, signalNames[info->si_signo]);
	    jio_fprintf(stderr, "    si_errno [%d]: %s\n",
			info->si_errno, strerror(info->si_errno));
	    jio_fprintf(stderr, "    si_code [%d]: %s\n",
			info->si_code, decode_si_code(sig, info));
	}

        if (sysThreadSelf()) {
            jio_fprintf(stderr, "\tstackpointer=%p\n",
              		sysThreadStackPointer(sysThreadSelf()));
        }
    }

    HandleSignalInVM(FALSE);

    Abort();

 done:
    /* We need to reinstate the signal handler each time... */
    sysSignal(sig, panicHandler);
}

JNIEXPORT void * JNICALL
JVM_RegisterSignal(jint sig, void *handler)
{
    signal_handler_t newHandler =
        handler == (void *)2 ? UserHandler : (signal_handler_t)handler;
    signal_handler_t oldHandler;

    switch (sig) {
    case SIGFPE:
    case SIGSEGV:
    case SIGILL:
        return (void *)-1; /* already used by VM */
	
    case SIGQUIT:
    case SIGABRT:
    case SIGBUS:
    case SIGEMT:
    case SIGSYS:
    case SIGTRAP:
    case SIGXCPU:
    case SIGXFSZ:
        if (!reduce_signal_usage)
	    return (void *)-1;
    }

    oldHandler = sysSignal(sig, newHandler);
    if (oldHandler == UserHandler) {
        return (void *)2;
    } else {
        return (void *)oldHandler;
    }
}

JNIEXPORT void JNICALL
JVM_RaiseSignal(jint sig)
{
    sysRaise(sig);
}

JNIEXPORT jint JNICALL
JVM_FindSignal(const char *name)
{
    if (!strcmp(name, "HUP"))  return SIGHUP;
    if (!strcmp(name, "INT"))  return SIGINT;
    if (!strcmp(name, "QUIT"))  return SIGQUIT;
    if (!strcmp(name, "ILL"))   return SIGILL;
    if (!strcmp(name, "TRAP"))  return SIGTRAP;
    if (!strcmp(name, "ABRT"))  return SIGABRT;
    if (!strcmp(name, "EMT"))   return SIGEMT;
    if (!strcmp(name, "FPE"))   return SIGFPE;
    if (!strcmp(name, "KILL"))  return SIGKILL;
    if (!strcmp(name, "BUS"))   return SIGBUS;
    if (!strcmp(name, "SEGV"))  return SIGSEGV;
    if (!strcmp(name, "SYS"))   return SIGSYS;
    if (!strcmp(name, "PIPE"))  return SIGPIPE;
    if (!strcmp(name, "ALRM"))  return SIGALRM;
    if (!strcmp(name, "TERM"))  return SIGTERM;
    if (!strcmp(name, "USR1"))  return SIGUSR1;
    if (!strcmp(name, "USR2"))  return SIGUSR2;
    if (!strcmp(name, "CHLD"))  return SIGCHLD;
    if (!strcmp(name, "PWR"))   return SIGPWR;
    if (!strcmp(name, "WINCH")) return SIGWINCH;
    if (!strcmp(name, "URG"))   return SIGURG;
    if (!strcmp(name, "POLL"))  return SIGPOLL;
    if (!strcmp(name, "IO"))    return SIGIO;
    if (!strcmp(name, "STOP"))  return SIGSTOP;
    if (!strcmp(name, "TSTP"))  return SIGTSTP;
    if (!strcmp(name, "CONT"))  return SIGCONT;
    if (!strcmp(name, "TTIN"))  return SIGTTIN;
    if (!strcmp(name, "TTOU"))  return SIGTTOU;
    if (!strcmp(name, "VTALRM")) return SIGVTALRM;
    if (!strcmp(name, "PROF"))  return SIGPROF;
    if (!strcmp(name, "XCPU"))  return SIGXCPU;
    if (!strcmp(name, "XFSZ"))  return SIGXFSZ;
    return -1;
}

static void sigthread(void *arg)
{
    while (1) {
        int sig = sysSignalWait();
	if (sig == SIGQUIT) {
	    jio_fprintf(stderr, "SIGQUIT\n");
	    
	    HandleSignalInVM(TRUE);
	    sigquit_thread = NULL;
	} else {
	    ExecEnv *ee = EE();
	    ClassClass *sigCls = FindStickySystemClass(ee, "sun/misc/Signal",
						       TRUE);
	    if (sigCls == NULL) {
	        exceptionClear(ee);
		continue;
	    }
	    execute_java_static_method(ee, sigCls, "dispatch", "(I)V", sig);
	    exceptionClear(ee);
	}
    }
}

void InitializeSignals()
{
    /* We have to install these signal handlers because Sun JIT relies on them
     * to be delivered.
     */
    sysSignal(SIGFPE, panicHandler);
    sysSignal(SIGSEGV, panicHandler);
    sysSignal(SIGILL, panicHandler);

    createSystemThread("Signal dispatcher", MaximumPriority, 0, 
		       sigthread, 0);

    if (!reduce_signal_usage) {
	sysSignal(SIGQUIT, UserHandler);

	sysSignal(SIGABRT, panicHandler);
	sysSignal(SIGBUS, panicHandler);
	sysSignal(SIGIOT, panicHandler);
	sysSignal(SIGEMT, panicHandler);
	sysSignal(SIGSYS, panicHandler);
	sysSignal(SIGTRAP, panicHandler);
	sysSignal(SIGXCPU, panicHandler);
	sysSignal(SIGXFSZ, panicHandler);
    }
}
