/*
 * @(#)UNIXProcess_md.c	1.44 98/08/14
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <wait.h>
#include <signal.h>

#include "jni.h"
#include "jvm.h"
#include "jni_util.h"
#include "java_lang_UNIXProcess.h"

/* java.io.FileDescriptor.fd */
static jfieldID fdID = (jfieldID) -1;

/* path in the environment */
static char **PATH = 0;
/* effective uid */
static uid_t uid;
/* effective group id */
static gid_t gid;

static void
parsePath () {
    char *path, *c, *len;
    int count = 0;
    int i;

    /* get uid, gid */
    uid = geteuid();
    gid = getegid();
    if ((path = getenv("PATH")) == 0) {
	return;
    }
    path = strdup(path);
    len = path + strlen(path);
    /* count path elements */
    for (c = path; c < len; c++) {
	if (*c == ':')
	    count++;
    }
    PATH = (char **)malloc((++count+1) * sizeof(char*));

    /* fill it in */
    PATH[0] = path;
    PATH[count] = 0;

    for (i = 1; i < count; i++) {
	c = strchr(path, ':');
	if (c == 0) { /* shouldn't happen */
	    /*	    jio_fprintf(stderr, "processmd.c: wrong count parsing path: i=%d, count=%d\n", i, count); */
	    break;
	}
	*c++ = 0;
	PATH[i] = path = c;
    }
}

/* return 0 if it is executable && readable by this process
 * -1 if no such file, -2 if it cannot be executed, and raise exception
 */
static int
statExecutable(char *exe, struct stat *b) {
    if (stat(exe, b)) { /* doesn't exist */
	return -1;
    }
    if (S_ISDIR(b->st_mode)) {
	/* cannot execute */
	return -2;
    }
    /* check for user permissions */
    if (b->st_uid == uid) {
	return (b->st_mode & S_IXUSR) ? 0 : -2;
    }
    /* check for group permissions */
    if (b->st_gid == gid) {
	return (b->st_mode & S_IXGRP) ? 0 : -2;
    }

    /* check for world permissions */
    return b->st_mode & S_IXOTH ? 0 : -2;
}

/* Find the command like a shell would.
 * signal an error for things not executable || not readable || not found.
 */

static char *fullPath(JNIEnv *env, char *part, char *full) {
    char **tmp;
    struct stat b;
    int ret;
    /*
     * If the filename we want to exec has any slashes in it then
     * we shouldn't do a path search, as in /foo ./foo or foo/bar.
     */
    if (strchr(part, '/') == NULL && PATH) {
	for (tmp = PATH; *tmp; tmp++) {
	    strcpy(full, *tmp);
	    /*
	     * empty path elements are like '.' so we don't want to append
	     * a slash to them.  Otherwise foo becomes /foo.
	     */
	    if (full[0] != '\0') {
		strcat(full, "/");
	    }
	    strcat(full, part);
	    ret = statExecutable(full, &b);
	    if (ret == -1) { /* doesn't exist */
		continue;
	    } else if (ret == -2) { /* can't execute */
		jio_snprintf(full, MAXPATHLEN, "%s: cannot execute", part);
		JNU_ThrowByName(env, "java/io/IOException", full);
		return 0;
	    } else {
		return full;
	    }
	}
    } else if (!(ret = statExecutable(part, &b))) {
	return part;
    } else if (ret == -2) { /* cannot execute */
	jio_snprintf(full, MAXPATHLEN, "%s: cannot execute", part);
	JNU_ThrowByName(env, "java/io/IOException", full);
	return 0;
    }

    /* not found if we got here */
    jio_snprintf(full, MAXPATHLEN, "%s: not found", part);
    JNU_ThrowByName(env, "java/io/IOException", full);
    return 0;
}

static jobject fork_wait_mon;
static int pforked = 0;

JNIEXPORT void JNICALL
Java_java_lang_UNIXProcess_initLockObject(JNIEnv *env, jclass cls)
{
    fork_wait_mon = JNU_NewObjectByName(env, "java/lang/Object", "()V");
    fork_wait_mon = (*env)->NewGlobalRef(env, fork_wait_mon);
}

/*
 * This routine is called once from the Process Reaper thread and
 * never returns.  Its job is to catch sigchld events and pass them
 * on to class UNIXProcess.
 */
JNIEXPORT void JNICALL
Java_java_lang_UNIXProcess_run(JNIEnv *env, jobject process)
{
    siginfo_t info;

    /* reaper thread loops for ever looking out for dead children */
    /* the loop below operates basically as a synchronized block */
    (*env)->MonitorEnter(env, fork_wait_mon);
    while (1) {
        while (pforked == 0) {
            JNU_MonitorWait(env, fork_wait_mon, 0);
        }
	pforked--;
	/* a little paranoia checking to make sure we don't underflow */
	if (pforked < 0)
	    pforked = 0;

        (*env)->MonitorExit(env, fork_wait_mon);

	/* block this reaper thread until the child process exits */
        if (waitid(P_ALL, 0, &info, WEXITED) != 0 || info.si_pid == 0) {
            continue;
        }

	(*env)->MonitorEnter(env, fork_wait_mon);

	JNU_CallStaticMethodByName(env,
				   NULL,
				   "java/lang/UNIXProcess",
				   "deadChild",
				   "(II)V",
                                   info.si_pid,
				   info.si_status);
	if ((*env)->ExceptionOccurred(env)) {
	    return;
	}
    }
}

JNIEXPORT jint JNICALL
Java_java_lang_UNIXProcess_forkAndExec(JNIEnv *env,
				       jobject process,
				       jobjectArray cmdarray,
				       jobjectArray envp,
				       jobject stdin_fd,
				       jobject stdout_fd,
				       jobject stderr_fd,
				       jobject sync_fd)
{
    /* let's do all the setup for the exec before we call fork1 */
    int fdin[2], fdout[2], fderr[2], fdsync[2];
    char **envstr = NULL;
    char fullpath[MAXPATHLEN+1];
    int i;
    char **argv;
    char * savedArgv0 = NULL;
    int argc;
    int pid = -1; /* this is the error return value */

    if (fdID == (jfieldID)-1) {
	jclass fdClass = (*env)->FindClass(env, "java/io/FileDescriptor");
	if (fdClass == NULL)
	    return 0;
	fdID = (*env)->GetFieldID(env, fdClass, "fd", "I");
	if (fdID == 0) {
	    JNU_ThrowInternalError(env, "Can't find field java.io.FileDescriptor.fd");
	}
    }

    /* On error_cleanup, we don't want to close() some random junk. */
    fdsync[0] = fdsync[1] =
	fdin[0] = fdin[1] =
	fdout[0] = fdout[1] =
	fderr[0] = fderr[1] = -1;

    argc = (*env)->GetArrayLength(env, cmdarray);
    if (argc == 0) {
	JNU_ThrowIllegalArgumentException(env, 0);
	return 0;
    }
    /* calloc null terminates for us */
    argv = calloc((argc + 1) * sizeof(char *), 1);
    if (argv == 0) {
	JNU_ThrowOutOfMemoryError(env, 0);
	goto error_cleanup;
    }

    for (i = 0; i < argc; i++) {
        jstring jstr = (*env)->GetObjectArrayElement(env, cmdarray, i);
	argv[i] = (char *)JNU_GetStringPlatformChars(env, jstr, NULL);
	if (argv[i] == NULL) {
	    goto error_cleanup;
	}
	(*env)->DeleteLocalRef(env, jstr);
    }

    if (PATH == 0) {
	parsePath();
    }

    savedArgv0 = argv[0];
    argv[0] = fullPath(env, *argv, fullpath);
    if (argv[0] == NULL) {
	/* fullPath has signalled an exception so we just return */
	goto error_cleanup;
    }

    if (envp != NULL) {
	int len = (*env)->GetArrayLength(env, envp);
	int i;

	/* calloc null terminates for us */
	envstr = calloc((len+1), sizeof(char *));
	if (envstr == 0) {
	    JNU_ThrowOutOfMemoryError(env, 0);
	    goto error_cleanup;
	}

	for (i = 0; i < len; ++i) {
	    jstring jstr = (*env)->GetObjectArrayElement(env, envp, i);
	    envstr[i] = (char *)JNU_GetStringPlatformChars(env, jstr, NULL);
	    if (envstr[i] == NULL) {
	        goto error_cleanup;
	    }
	    (*env)->DeleteLocalRef(env, jstr);
        }
    }

    if ((pipe(fdin) < 0) ||   /* stdin */
	(pipe(fdout) < 0) ||  /* stdout */
	(pipe(fderr) < 0) ||  /* stderr */
	(pipe(fdsync) < 0)) { /* fd for synchronization */
	/* make sure we clean up */
 	JNU_ThrowByName(env, "java/io/IOException", strerror(errno));
 	goto error_cleanup;
    }

    if ((pid = fork1()) == 0) {
	/* Child process */
	int i, max_fd;

	/* 0 open for reading, 1 open for writing */
	/* close as we go along to reduce the chance of running out of fds */
	close(fdsync[1]);
	dup2(fdin[0], 0);
	close(fdin[0]);
	dup2(fdout[1], 1);
	close(fdout[1]);
	dup2(fderr[1], 2);
	close(fderr[1]);

 	{
 	    int err;
 	    char c;
 	    err = read(fdsync[0], &c, 1);
 	    close(fdsync[0]);
 	}

	/* close everything */
	max_fd = sysconf(_SC_OPEN_MAX);
	for (i = 3; i < max_fd; i++) {
	    close(i);
	}


	if (envstr == NULL) {
	    execv(argv[0], argv);
	} else {
	    execve(argv[0], argv, envstr);
	}

	/*
	 * the exec failed, so there's not much we can do except exit.
	 * We use _exit so that we don't flush file descriptors that
	 * might be shared with our parent, even though we've probably
	 * already closed them at this point
	 */
	_exit(-1);
    } else {
  	if (pid < 0) {
	    /* make sure we clean up our side of the pipes */
  	    JNU_ThrowByName(env, "java/io/IOException", strerror(errno));
  	    goto error_cleanup;
	}

	/* clean up the child's side of the pipes */
	close(fdin[0]);
	close(fdout[1]);
	close(fderr[1]);
 	close(fdsync[0]);

	(*env)->MonitorEnter(env, fork_wait_mon);
	pforked++;
	JNU_Notify(env, fork_wait_mon);
	(*env)->MonitorExit(env, fork_wait_mon);
	if ((*env)->ExceptionOccurred(env)) {
	    goto error_cleanup;
	}

	(*env)->SetIntField(env, stdin_fd, fdID, fdin[1]);
	(*env)->SetIntField(env, stdout_fd, fdID, fdout[0]);
	(*env)->SetIntField(env, stderr_fd, fdID, fderr[0]);
	(*env)->SetIntField(env, sync_fd, fdID, fdsync[1]);
    }

 normal_cleanup:
    if (argv) {
	int len = (*env)->GetArrayLength(env, cmdarray);
	if (savedArgv0)
	    argv[0] = savedArgv0;
	for (i = 0; i < len; i++) {
	    jstring jstr = (*env)->GetObjectArrayElement(env, cmdarray, i);
	    JNU_ReleaseStringPlatformChars(env, jstr, argv[i]);
	    (*env)->DeleteLocalRef(env, jstr);
	}
	free(argv);
    }
    if (envstr) {
	int len = (*env)->GetArrayLength(env, envp);
	for (i = 0; i < len; i++) {
	    jstring jstr = (*env)->GetObjectArrayElement(env, envp, i);
	    JNU_ReleaseStringPlatformChars(env, jstr, envstr[i]);
	    (*env)->DeleteLocalRef(env, jstr);
	}
	free(envstr);
    }
    return pid;		/* pid will be -1 in the error case */

 error_cleanup:
    close(fdin[0]);
    close(fdin[1]);
    close(fdout[0]);
    close(fdout[1]);
    close(fderr[0]);
    close(fderr[1]);
    close(fdsync[0]);
    close(fdsync[1]);
    goto normal_cleanup;
}

JNIEXPORT void JNICALL
Java_java_lang_UNIXProcess_destroy(JNIEnv *env, jobject process)
{
    jboolean exc;
    int pid = JNU_GetFieldByName(env, &exc, process, "pid", "I").i;
    if (!exc) {
        kill(pid, SIGTERM);
    }
}

