/*
 * @(#)PlainDatagramSocketImpl.c	1.33 98/08/24
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include <errno.h>
#include <netinet/in.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>

#include "jvm.h"
#include "jni_util.h"
#include "net_util.h"

#include "java_net_SocketOptions.h"
#include "java_net_PlainDatagramSocketImpl.h"

/************************************************************************
 * PlainDatagramSocketImpl
 */

static jfieldID IO_fd_fdID;

static jfieldID pdsi_fdID;
static jfieldID pdsi_timeoutID;
static jfieldID pdsi_localPortID;

/*
 * Class:     java_net_PlainDatagramSocketImpl
 * Method:    init
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_java_net_PlainDatagramSocketImpl_init(JNIEnv *env, jclass cls) {

    pdsi_fdID = (*env)->GetFieldID(env, cls, "fd",
				   "Ljava/io/FileDescriptor;");
    pdsi_timeoutID = (*env)->GetFieldID(env, cls, "timeout", "I");
    pdsi_localPortID = (*env)->GetFieldID(env, cls, "localPort", "I");
    IO_fd_fdID = NET_GetFileDescriptorID(env);
}

/*
 * Class:     java_net_PlainDatagramSocketImpl
 * Method:    bind
 * Signature: (ILjava/net/InetAddress;)V
 */
JNIEXPORT void JNICALL
Java_java_net_PlainDatagramSocketImpl_bind(JNIEnv *env, jobject this,
					   jint port, jobject laddr) {

    jobject fdObj = (*env)->GetObjectField(env, this, pdsi_fdID);
    int fd;

    struct sockaddr_in lcladdr;
    int lcladdrlen= sizeof(lcladdr);
    int address;

    if (IS_NULL(laddr)) { /* shouldn't happen, but be safe */
	address = 0;
    } else {
	address = (*env)->GetIntField(env, laddr, ia_addressID);
    }
    if (IS_NULL(fdObj)) {
	JNU_ThrowNullPointerException(env, "fd object");
	return;
    } else {
	fd = (*env)->GetIntField(env, fdObj, IO_fd_fdID);
    }
    /* bind - pick a port number for local addr*/
    memset((char *) &lcladdr, 0, sizeof(lcladdr));
    lcladdr.sin_family      = AF_INET;
    lcladdr.sin_port        = htons((short)port);
    lcladdr.sin_addr.s_addr = (laddr ? htonl(address) : INADDR_ANY);

    if (bind(fd, (struct sockaddr *)&lcladdr, sizeof(lcladdr)) == -1) {
	if (errno == EADDRINUSE || errno == EADDRNOTAVAIL ||
	    errno == EPERM || errno == EACCES) {
	    JNU_ThrowByName(env, JNU_JAVANETPKG "BindException",
			    strerror(errno));
	} else {
	    JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			    strerror(errno));
	}
	return;
    }

    /* find what port system picked for us - obviously brain dead interface,
       should have told me in the bind call itself */
    if (getsockname(fd, (struct sockaddr *)&lcladdr, &lcladdrlen) == -1) {
        JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			strerror(errno));
	return;
    }
    (*env)->SetIntField(env, this, pdsi_localPortID,
			ntohs(lcladdr.sin_port));
}

/*
 * Class:     java_net_PlainDatagramSocketImpl
 * Method:    send
 * Signature: (Ljava/net/DatagramPacket;)V
 */
JNIEXPORT void JNICALL
Java_java_net_PlainDatagramSocketImpl_send(JNIEnv *env, jobject this,
					   jobject packet) {

    char BUF[MAX_BUFFER_LEN];
    char *fullPacket = NULL;
    int mallocedPacket = JNI_FALSE;
    /* The object's field */
    jobject fdObj = (*env)->GetObjectField(env, this, pdsi_fdID);

    jbyteArray packetBuffer;
    jobject packetAddress;
    jint packetBufferOffset, packetBufferLen, packetPort;

    /* The packetAddress address, family and port */
    jint address, family, port;
    /* The fdObj'fd */
    jint fd;

    ssize_t n = -1;
    struct sockaddr_in rmtaddr;

    if (IS_NULL(fdObj)) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			"Socket closed");
	return;
    }
    fd = (*env)->GetIntField(env, fdObj, IO_fd_fdID);

    if (IS_NULL(packet)) {
	JNU_ThrowNullPointerException(env, "packet");
	return;
    }

    packetBuffer = (*env)->GetObjectField(env, packet, dp_bufID);
    packetAddress = (*env)->GetObjectField(env, packet, dp_addressID);
    if (IS_NULL(packetBuffer) || IS_NULL(packetAddress)) {
	JNU_ThrowNullPointerException(env, "null buffer || null address");
	return;
    }

    packetPort = (*env)->GetIntField(env, packet, dp_portID);
    packetBufferOffset = (*env)->GetIntField(env, packet, dp_offsetID);
    packetBufferLen = (*env)->GetIntField(env, packet, dp_lengthID);

    address = (*env)->GetIntField(env, packetAddress, ia_addressID);
    family = (*env)->GetIntField(env, packetAddress, ia_familyID);

    rmtaddr.sin_port = htons((short)packetPort);
    rmtaddr.sin_addr.s_addr = htonl(address);
    rmtaddr.sin_family = family;

    if (packetBufferLen > MAX_BUFFER_LEN) {
	/* When JNI-ifying the JDK's IO routines, we turned
	 * read's and write's of byte arrays of size greater
	 * than 2048 bytes into several operations of size 2048.
	 * This saves a malloc()/memcpy()/free() for big
	 * buffers.  This is OK for file IO and TCP, but that
	 * strategy violates the semantics of a datagram protocol.
	 * (one big send) != (several smaller sends).  So here
	 * we *must* alloc the buffer.  Note it needn't be bigger
	 * than 65,536 (0xFFFF) the max size of an IP packet.
	 * Anything bigger should be truncated anyway.
	 *
	 * We may want to use a smarter allocation scheme at some
	 * point.
	 */
	if (packetBufferLen > MAX_PACKET_LEN) {
	    packetBufferLen = MAX_PACKET_LEN;
	}
	fullPacket = (char *)malloc(packetBufferLen);

	if (!fullPacket) {
	    JNU_ThrowOutOfMemoryError(env, "heap allocation failed");
	    return;
	} else {
	    mallocedPacket = JNI_TRUE;
	}
    } else {
	fullPacket = &(BUF[0]);
    }

    (*env)->GetByteArrayRegion(env, packetBuffer, packetBufferOffset, packetBufferLen,
			       (jbyte *)fullPacket);

    switch (JVM_SendTo(fd, fullPacket, packetBufferLen, 0,
			       (struct sockaddr *)&rmtaddr,
		       sizeof(rmtaddr))) {
        case JVM_IO_ERR:
	    JNU_ThrowIOException(env, strerror(errno));
	    break;
        case JVM_IO_INTR:
	    JNU_ThrowByName(env, "java/io/InterruptedIOException",
			    "operation interrupted");
    }
    if (mallocedPacket) {
	free(fullPacket);
    }
    return;
}

/*
 * Class:     java_net_PlainDatagramSocketImpl
 * Method:    peek
 * Signature: (Ljava/net/InetAddress;)I
 */
JNIEXPORT jint JNICALL
Java_java_net_PlainDatagramSocketImpl_peek(JNIEnv *env, jobject this,
					   jobject addressObj) {

    jobject fdObj = (*env)->GetObjectField(env, this, pdsi_fdID);
    jint timeout = (*env)->GetIntField(env, this, pdsi_timeoutID);
    jint fd;

    /* The address and family fields of addressObj */
    jint address, family;

    ssize_t n;
    struct sockaddr_in remote_addr;
    int remote_addrsize = sizeof (remote_addr);
    char buf[1];

    if (IS_NULL(fdObj)) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException", "Socket closed");
	return -1;
    } else {
	fd = (*env)->GetIntField(env, fdObj, IO_fd_fdID);
    }
    if (IS_NULL(addressObj)) {
	JNU_ThrowNullPointerException(env, "Null address in peek()");
    } else {
	address = (*env)->GetIntField(env, addressObj, ia_addressID);
	family = (*env)->GetIntField(env, addressObj, ia_familyID);
    }
    if (timeout) {
	int ret = JVM_Timeout(fd, timeout);
        if (ret == 0) {
            JNU_ThrowByName(env, JNU_JAVAIOPKG "InterruptedIOException",
			    "Peek timed out");
	    return ret;
	} else if (ret == JVM_IO_ERR) {
	    JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			    (errno == EBADF)?"Socket closed":strerror(errno));
	    return ret;
	} else if (ret == JVM_IO_INTR) {
	    JNU_ThrowByName(env, JNU_JAVAIOPKG "InterruptedIOException",
			    "operation interrupted");
	    return ret;	/* WARNING: SHOULD WE REALLY RETURN -2??? */
	}
    }
    n = JVM_RecvFrom(fd, buf, 1, MSG_PEEK,
		     (struct sockaddr *)&remote_addr, &remote_addrsize);

    if (n == JVM_IO_ERR) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			(errno == EBADF) ? "Socket closed" : strerror(errno));
        return 0;
    } else if (n == JVM_IO_INTR) {
	JNU_ThrowByName(env, "java/io/InterruptedIOException", 0);
        return 0;
    }
    (*env)->SetIntField(env, addressObj, ia_addressID,
			ntohl(remote_addr.sin_addr.s_addr));
    (*env)->SetIntField(env, addressObj, ia_familyID,
			remote_addr.sin_family);

    /* return port */
    return ntohs(remote_addr.sin_port);
}

/*
 * Class:     java_net_PlainDatagramSocketImpl
 * Method:    receive
 * Signature: (Ljava/net/DatagramPacket;)V
 */
JNIEXPORT void JNICALL
Java_java_net_PlainDatagramSocketImpl_receive(JNIEnv *env, jobject this,
					      jobject packet) {

    char BUF[MAX_BUFFER_LEN];
    char *fullPacket = NULL;
    int mallocedPacket = JNI_FALSE;
    jobject fdObj = (*env)->GetObjectField(env, this, pdsi_fdID);
    jint timeout = (*env)->GetIntField(env, this, pdsi_timeoutID);

    jbyteArray packetBuffer;
    jint packetBufferOffset, packetBufferLen;

    int fd;
    jbyteArray data;

    int datalen;
    int n;
    struct sockaddr_in remote_addr;
    int remote_addrsize = sizeof (remote_addr);

    if (IS_NULL(fdObj)) {
	JNU_ThrowNullPointerException(env, "fd object");
	return;
    }

    fd = (*env)->GetIntField(env, fdObj, IO_fd_fdID);

    if (IS_NULL(packet)) {
	JNU_ThrowNullPointerException(env, "packet");
	return;
    }

    packetBuffer = (*env)->GetObjectField(env, packet, dp_bufID);
    if (IS_NULL(packetBuffer)) {
	JNU_ThrowNullPointerException(env, "packet buffer");
        return;
    }
    packetBufferOffset = (*env)->GetIntField(env, packet, dp_offsetID);
    packetBufferLen = (*env)->GetIntField(env, packet, dp_lengthID);
    if (timeout) {
	int ret = JVM_Timeout(fd, timeout);
        if (ret == 0) {
            JNU_ThrowByName(env, JNU_JAVAIOPKG "InterruptedIOException",
			    "Receive timed out");
	    return;
	} else if (ret == JVM_IO_ERR) {
	    JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			    "Socket closed");
	    return;
	} else if (ret == JVM_IO_INTR) {
	    JNU_ThrowByName(env, JNU_JAVAIOPKG "InterruptedIOException",
			    "operation interrupted");
	    return;
	}
    }

    if (packetBufferLen > MAX_BUFFER_LEN) {

	/* When JNI-ifying the JDK's IO routines, we turned
	 * read's and write's of byte arrays of size greater
	 * than 2048 bytes into several operations of size 2048.
	 * This saves a malloc()/memcpy()/free() for big
	 * buffers.  This is OK for file IO and TCP, but that
	 * strategy violates the semantics of a datagram protocol.
	 * (one big send) != (several smaller sends).  So here
	 * we *must* alloc the buffer.  Note it needn't be bigger
	 * than 65,536 (0xFFFF) the max size of an IP packet.
	 * anything bigger is truncated anyway.
	 *
	 * We may want to use a smarter allocation scheme at some
	 * point.
	 */
	if (packetBufferLen > MAX_PACKET_LEN) {
	    packetBufferLen = MAX_PACKET_LEN;
	}
	fullPacket = (char *)malloc(packetBufferLen);

	if (!fullPacket) {
	    JNU_ThrowOutOfMemoryError(env, "heap allocation failed");
	    return;
	} else {
	    mallocedPacket = JNI_TRUE;
	}
    } else {
	fullPacket = &(BUF[0]);
    }
    n = JVM_RecvFrom(fd, fullPacket, packetBufferLen, 0,
		     (struct sockaddr *)&remote_addr, &remote_addrsize);
    /* truncate the data if the packet's length is too small */
    if (n > packetBufferLen) {
	n = packetBufferLen;
    }
    if (n == JVM_IO_ERR) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			"socket closed");
	(*env)->SetIntField(env, packet, dp_offsetID, 0);
	(*env)->SetIntField(env, packet, dp_lengthID, 0);
    } else if (n == JVM_IO_INTR) {
	JNU_ThrowByName(env, JNU_JAVAIOPKG "InterruptedIOException",
			"operation interrupted");
	(*env)->SetIntField(env, packet, dp_offsetID, 0);
	(*env)->SetIntField(env, packet, dp_lengthID, 0);
    } else {
	/*
	 * success - fill in received address...
	 *
	 * REMIND: Fill in an int on the packet, and create inetadd
	 * object in Java, as a performance improvement. Also
	 * construct the inetadd object lazily.
	 */
	jclass ia_clazz = (*env)->FindClass(env, "java/net/InetAddress");
	jmethodID ia_ctor = (*env)->GetMethodID(env, ia_clazz,
						"<init>", "()V");
	jobject packetAddress = (*env)->NewObject(env, ia_clazz, ia_ctor);
	(*env)->SetIntField(env, packetAddress, ia_addressID,
			    ntohl(remote_addr.sin_addr.s_addr));
	(*env)->SetIntField(env, packetAddress, ia_familyID,
			    remote_addr.sin_family);
	/* stuff the new Inetaddress in the packet */
	(*env)->SetObjectField(env, packet, dp_addressID, packetAddress);
	/* and fill in the data, remote address/port and such */
	(*env)->SetByteArrayRegion(env, packetBuffer, packetBufferOffset, n,
				   (jbyte *)fullPacket);
	(*env)->SetIntField(env, packet, dp_portID,
			    ntohs(remote_addr.sin_port));
	(*env)->SetIntField(env, packet, dp_lengthID, n);
	(*env)->SetIntField(env, packetAddress, ia_addressID,
			    ntohl(remote_addr.sin_addr.s_addr));
	(*env)->SetIntField(env, packetAddress, ia_familyID,
			    remote_addr.sin_family);
    }

    if (mallocedPacket) {
	free(fullPacket);
    }
}

/*
 * Class:     java_net_PlainDatagramSocketImpl
 * Method:    datagramSocketCreate
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_java_net_PlainDatagramSocketImpl_datagramSocketCreate(JNIEnv *env,
							   jobject this) {
    jobject fdObj = (*env)->GetObjectField(env, this, pdsi_fdID);
    int fd;

    int arg = -1;

    if (IS_NULL(fdObj)) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			"Socket closed");
	return;
    } else {
	fd =  JVM_Socket(AF_INET, SOCK_DGRAM, 0);
    }
    if (fd == JVM_IO_ERR) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			strerror(errno));
	return;
    }
    (*env)->SetIntField(env, fdObj, IO_fd_fdID, fd);
}

/*
 * Class:     java_net_PlainDatagramSocketImpl
 * Method:    datagramSocketClose
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_java_net_PlainDatagramSocketImpl_datagramSocketClose(JNIEnv *env,
							  jobject this) {
    /*
     * REMIND: PUT A LOCK AROUND THIS CODE
     */
    jobject fdObj = (*env)->GetObjectField(env, this, pdsi_fdID);
    int fd;

    if (IS_NULL(fdObj)) {
	JNU_ThrowNullPointerException(env, "fd object");
	return;
    } else {
        fd = (*env)->GetIntField(env, fdObj, IO_fd_fdID);
    }
    JVM_Close(fd);
    (*env)->SetIntField(env, fdObj, IO_fd_fdID, -1);
}

/*
 * Class:     java_net_PlainDatagramSocketImpl
 * Method:    socketSetOption
 * Signature: (ILjava/lang/Object;)V
 */
JNIEXPORT void JNICALL
Java_java_net_PlainDatagramSocketImpl_socketSetOption(JNIEnv *env,
						      jobject this,
						      jint opt,
						      jobject value) {

    jobject fdObj = (*env)->GetObjectField(env, this, pdsi_fdID);
    int fd;

    if (IS_NULL(fdObj)) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			"Socket closed");
	return;
    } else {
	fd = (*env)->GetIntField(env, fdObj, IO_fd_fdID);
    }
    if (IS_NULL(value)) {
	JNU_ThrowNullPointerException(env, "value argument");
	return;
    }
    if (fd < 0) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			"socket closed");
    } else if (opt == java_net_SocketOptions_IP_MULTICAST_IF) {
	/* REMIND: sysAssert type */
        int address = (*env)->GetIntField(env, value, ia_addressID);
	struct in_addr in;
	in.s_addr = htonl(address);
	errno = 0;
	if (setsockopt(fd, IPPROTO_IP, IP_MULTICAST_IF,
		       (const char*)&in, sizeof(in)) < 0) {
	    JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			    strerror(errno));
	}
    } else if (opt == java_net_SocketOptions_SO_REUSEADDR) {
	jclass iCls = (*env)->FindClass(env, "java/lang/Integer");
	jfieldID i_valueID = (*env)->GetFieldID(env, iCls, "value", "I");

	int tmp = (*env)->GetIntField(env, value, i_valueID);
	int arg = tmp ? -1: 0;
	if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR,
		       (char *)&arg, sizeof(arg)) < 0) {
	    JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			    strerror(errno));
	}
    } else if (opt == java_net_SocketOptions_SO_SNDBUF) {
	int arg;
	jclass iCls = (*env)->FindClass(env, "java/lang/Integer");
	jfieldID i_valueID = (*env)->GetFieldID(env, iCls, "value", "I");
	arg = (*env)->GetIntField(env, value, i_valueID);
	if (setsockopt(fd, SOL_SOCKET, SO_SNDBUF,
		       (char *)&arg, sizeof(arg)) < 0) {
	    NET_ThrowCurrent(env, "setsockopt() SO_SNDBUF");
	    return;

	}
    } else if (opt == java_net_SocketOptions_SO_RCVBUF) {
	int arg;
	jclass iCls = (*env)->FindClass(env, "java/lang/Integer");
	jfieldID i_valueID = (*env)->GetFieldID(env, iCls, "value", "I");
	arg = (*env)->GetIntField(env, value, i_valueID);
	if (setsockopt(fd, SOL_SOCKET, SO_RCVBUF,
		       (char *)&arg, sizeof(arg)) < 0) {
	    NET_ThrowCurrent(env, "setsockopt() SO_RCVBUF");
	    return;

	}
    } else {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			"invalid DatagramSocket option");
    }
}

/*
 * Returns relevant info as a jint.
 *
 * Class:     java_net_PlainDatagramSocketImpl
 * Method:    socketGetOption
 * Signature: (I)I
 */
JNIEXPORT jint JNICALL
Java_java_net_PlainDatagramSocketImpl_socketGetOption(JNIEnv *env, jobject this,
						      jint opt) {
    jobject fdObj = (*env)->GetObjectField(env, this, pdsi_fdID);
    int fd;

    if (IS_NULL(fdObj)) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			"Socket closed");
	return -1;
    } else {
	fd = (*env)->GetIntField(env, fdObj, IO_fd_fdID);
    }

    if (fd < 0) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			"socket closed");
    } else if (opt == java_net_SocketOptions_SO_BINDADDR) {
	/* find out local IP address, return as 32 bit value */
	struct sockaddr_in him;
	int len = sizeof(struct sockaddr_in);
	memset(&him, 0, len);
	if (getsockname(fd, (struct sockaddr *)&him, &len) == -1) {
	    JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			    strerror(errno));
	    return -1;
	}
	return ntohl(him.sin_addr.s_addr);
	/* == INADDR_ANY) ? -1: him.sin_addr.s_addr;*/
    } else if (opt == java_net_SocketOptions_IP_MULTICAST_IF) {
        struct in_addr in;
	int len = sizeof(struct in_addr);
	if (getsockopt(fd, IPPROTO_IP, IP_MULTICAST_IF,
		       (char *)&in, &len) < 0) {
	    JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			    strerror(errno));
	    return -1;
	}
	return ntohl(in.s_addr);
    } else if (opt == java_net_SocketOptions_SO_SNDBUF) {
	int result = -1;
	int len = sizeof(result);
	if (getsockopt(fd, SOL_SOCKET, SO_SNDBUF,
		       (char *)&result, &len) < 0) {
	    NET_ThrowCurrent(env, "getsockopt() SO_SNDBUF");
	    return -1;
	}
	return result;
    } else if (opt == java_net_SocketOptions_SO_RCVBUF) {
	int result = -1;
	int len = sizeof(result);
	if (getsockopt(fd, SOL_SOCKET, SO_RCVBUF,
		       (char *)&result, &len) < 0) {
	    NET_ThrowCurrent(env, "getsockopt() SO_RCVBUF");
	    return -1;
	}
	return result;
    } else {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			"invalid DatagramSocket option");
	return -1;
    }
}

/*
 * Multicast-related calls
 */

JNIEXPORT void JNICALL
Java_java_net_PlainDatagramSocketImpl_setTTL(JNIEnv *env, jobject this,
					     jbyte ttl) {
    Java_java_net_PlainDatagramSocketImpl_setTimeToLive(env, this, (jint)ttl);
}

/*
 * Class:     java_net_PlainDatagramSocketImpl
 * Method:    setTTL
 * Signature: (B)V
 */
JNIEXPORT void JNICALL
Java_java_net_PlainDatagramSocketImpl_setTimeToLive(JNIEnv *env, jobject this,
						    jint ttl) {

    jobject fdObj = (*env)->GetObjectField(env, this, pdsi_fdID);
    int fd;
    /* it is important to cast this to a char, otherwise setsockopt gets confused */
    char ittl = (char)ttl;

    if (IS_NULL(fdObj)) {
	JNU_ThrowNullPointerException(env, "fd object");
	return;
    } else {
	fd = (*env)->GetIntField(env, fdObj, IO_fd_fdID);
    }
    /* setsockopt to be correct ttl */
    if (setsockopt(fd, IPPROTO_IP, IP_MULTICAST_TTL, (char*)&ittl,
		   sizeof(ittl)) < 0) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			strerror(errno));
    }
}

/*
 * Class:     java_net_PlainDatagramSocketImpl
 * Method:    getTTL
 * Signature: ()B
 */
JNIEXPORT jbyte JNICALL
Java_java_net_PlainDatagramSocketImpl_getTTL(JNIEnv *env, jobject this) {
    return (jbyte)Java_java_net_PlainDatagramSocketImpl_getTimeToLive(env, this);
}


/*
 * Class:     java_net_PlainDatagramSocketImpl
 * Method:    getTTL
 * Signature: ()B
 */
JNIEXPORT jint JNICALL
Java_java_net_PlainDatagramSocketImpl_getTimeToLive(JNIEnv *env, jobject this) {

    jobject fdObj = (*env)->GetObjectField(env, this, pdsi_fdID);
    jint fd = -1;

    u_char ttl = 0;
    int len = sizeof(ttl);
    if (IS_NULL(fdObj)) {
	JNU_ThrowNullPointerException(env, "fd object");
	return -1;
    } else {
	fd = (*env)->GetIntField(env, fdObj, IO_fd_fdID);
    }
    /* getsockopt of ttl */
    if (getsockopt(fd, IPPROTO_IP, IP_MULTICAST_TTL,
		   (char*)&ttl, &len) < 0) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			strerror(errno));
	return -1;
    }
    return (jint)ttl;
}

/*
 * Class:     java_net_PlainDatagramSocketImpl
 * Method:    join
 * Signature: (Ljava/net/InetAddress;)V
 */
JNIEXPORT void JNICALL
Java_java_net_PlainDatagramSocketImpl_join(JNIEnv *env, jobject this,
					   jobject address) {

    jobject fdObj = (*env)->GetObjectField(env, this, pdsi_fdID);
    jint fd, addressAddress;

    struct sockaddr_in name;
    struct ip_mreq mname;

  if (IS_NULL(fdObj)) {
      JNU_ThrowNullPointerException(env, "fd object");
      return;
  } else {
      fd = (*env)->GetIntField(env, fdObj, IO_fd_fdID);
  }
  if (IS_NULL(address)) {
      JNU_ThrowNullPointerException(env, "address");
      return;
  } else {
      addressAddress = (*env)->GetIntField(env, address, ia_addressID);
  }

  /*
   * Set the multicast group address in the ip_mreq field
   * eventually this check should be done by the security manager
   */
  mname.imr_multiaddr.s_addr = htonl(addressAddress);
  if (!IN_MULTICAST(addressAddress)) {
      JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException", "not in multicast");
  }
  /* Set the interface to be default */
  mname.imr_interface.s_addr = INADDR_ANY;

  /* Join the multicast group */
  if (setsockopt(fd, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char *) &mname,
	   sizeof (mname)) < 0) {
      JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException","error setting options");
  }
  return;
}

/*
 * Class:     java_net_PlainDatagramSocketImpl
 * Method:    leave
 * Signature: (Ljava/net/InetAddress;)V
 */
JNIEXPORT void JNICALL
Java_java_net_PlainDatagramSocketImpl_leave(JNIEnv *env, jobject this,
					    jobject address) {

    jobject fdObj = (*env)->GetObjectField(env, this, pdsi_fdID);
    jint fd, addressAddress;

    struct sockaddr_in name;
    struct ip_mreq mname;

    if (IS_NULL(fdObj)) {
	JNU_ThrowNullPointerException(env, "fd object");
	return;
    } else {
	fd = (*env)->GetIntField(env, fdObj, IO_fd_fdID);
    }
    if (IS_NULL(address)) {
	JNU_ThrowNullPointerException(env, "address argument");
	return;
    } else {
	addressAddress = (*env)->GetIntField(env, address, ia_addressID);
    }
    /*
     * Set the multicast group address in the ip_mreq field
     * eventually this check should be done by the security manager
     */
    mname.imr_multiaddr.s_addr = htonl(addressAddress);
    if (!IN_MULTICAST(addressAddress)) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			strerror(errno));
    }
    /* Set the interface to be default */
    mname.imr_interface.s_addr = INADDR_ANY;

    /* Join the multicast group */
    if (setsockopt(fd, IPPROTO_IP, IP_DROP_MEMBERSHIP,(char *) &mname,
	   sizeof (mname)) < 0) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			strerror(errno));
    }
    return;
}


