/*
 * @(#)HAE_API_SolarisOS.c	1.7 98/05/08
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
/*****************************************************************************/
/*
**	HAE_API_SolarisOS.c
**
**	This provides platform specfic functions for Sun Solaris OS. This interface
**	for HAE uses the dev/audio device
**
**	Overview:
**		This works by creating another thread, aquires the sound device via
**		open("/dev/audio",O_WRONLY) call. Then allocating enough buffers, preps then,
**		and build buffers depending upon how much data has been written to the
**		audio device. The thread polls the current position in the audio stream
**		that has been written and when it falls below a buffer or two, more
**		are built in that thread.
**
**	Written by Steve Hales
**
**	Confidential-- Internal use only
**
**	History	-
**	12/1/97		Gleaned from the orignal GenSolarisTools.c file
**
**	JAVASOFT
**
**	05.06.98:	$$kk: changes to remain analagous with HAE_API_WinOS.c fix for bug #4135246.
**				made lastPos global and updated in HAE_AquireAudioCard.
*/
/*****************************************************************************/

#include <stdio.h>
#include <unistd.h>
#include <stdarg.h>
#include <stdlib.h>
#include <memory.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <errno.h>
#include <sys/audioio.h>
#include <string.h>

#include "X_API.h"
#include "HAE_API.h"

// How many audio frames to generate at one time 
#define HAE_SOLARIS_FRAMES_PER_BLOCK		8
#define HAE_SOLARIS_SOUND_PERIOD			10	// sleep time in milliseconds, between position checks

static void					*g_audioBufferBlock;
static long					g_audioByteBufferSize;			// size of audio buffers in bytes

static int					g_waveDevice = 0;
static long					g_shutDownDoubleBuffer;
static long					g_activeDoubleBuffer;

static short int			g_bitSize;
static short int			g_channels;

// $$kk: this is used in HAE_SetDeviceID and HAE_GetDeviceID but was not defined
long						g_currentDeviceID = 0;

 // number of samples per audio frame to generate
long						g_audioFramesToGenerate;

// How many audio frames to generate at one time 
static unsigned int			g_synthFramesPerBlock;	// setup upon runtime

static unsigned int			g_audioPeriodSleepTime;	// setup upon runtime

// $$kk: 03.17.98: these variables describe device capabilities and requirements
XBOOL						g_supports16			= 1;
XBOOL						g_supports8				= 1;
XBOOL						g_supportsStereo		= 1;
XBOOL						g_convertUnsigned		= 0;

// $$kk: 05.06.98: made lastPos a global variable
static long					lastPos;



// This file contains API's that need to be defined in order to get HAE (IgorAudio)
// to link and compile.

// **** System setup and cleanup functions
// Setup function. Called before memory allocation, or anything serious. Can be used to 
// load a DLL, library, etc.
// return 0 for ok, or -1 for failure
int HAE_Setup(void)
{
	// $$kk: 03.17.98: i am using this method to query device information for solaris.
	// here are the solaris drivers:
	// audiocs:		supports linear 16 bit data, mono or stereo
	//				does *not* support linear 8 bit data
	// sbpro:		sb16/AWE32 supports linear 8 and 16 bit data
	//				sbpro supports linear 8
	// dbri:		supports linear 16 bit data, mono or stereo
	//				does *not* support linear 8 bit data
	// audioamd:	supports no linear formats

	// we do not publically let people change the rendering format, so right now
	// we'll just do 16 if we can, otherwise 8.  this will have to be done more
	// carefully with the new api.

	int					pseudoDevice = 0;
	long				error;
    audio_device_t		deviceInfo;

	char sbproStr[] = "SUNW,sbpro";
	char sb16Str[] = "SUNW,sb16";
	char audiocsStr[] = "SUNW,CS4231";
	char dbriStr[] = "SUNW,dbri";
	char audioamdStr[] = "SUNW,am79c30";	

	pseudoDevice = open("/dev/audioctl", O_RDONLY);

	if (pseudoDevice == -1)
	{
		//fprintf(stderr, "could not get pseudoDevice\n");
		return 0;
	}

	error = ioctl(pseudoDevice, AUDIO_GETDEV, &deviceInfo);

	close(pseudoDevice);

		g_supports16		= 0;
		g_supports8			= 1;
		g_supportsStereo	= 1;   
		g_convertUnsigned	= 1;

	if (strcmp(sbproStr, deviceInfo.name) == 0)
	{
		// sbpro does not support 16 bit
		// it does support 8 bit with unsigned conversion
		// it does support stereo
		//fprintf(stderr, "found sbpro hardware \n");

		g_supports16		= 0;
		g_supports8			= 1;
		g_supportsStereo	= 1;   
		g_convertUnsigned	= 1;

		return 0;
	}

	if (strcmp(sb16Str, deviceInfo.name) == 0)
	{
		// sb16 does support 16 bit
		// it does support 8 bit with unsigned conversion
		// it does support stereo
		//fprintf(stderr, "found sb16 hardware \n");

		g_supports16		= 1;
		g_supports8			= 1;
		g_supportsStereo	= 1;   
		g_convertUnsigned	= 1;

		return 0;
	}

	if (strcmp(audiocsStr, deviceInfo.name) == 0)
	{
		// audiocs does support 16 bit
		// it does not support 8 bit linear
		// it does support stereo
		//fprintf(stderr, "found audiocs hardware \n");

		g_supports16		= 1;
		g_supports8			= 0;
		g_supportsStereo	= 1;   
		g_convertUnsigned	= 0;

		return 0;
	}

	if (strcmp(dbriStr, deviceInfo.name) == 0)
	{
		// dbri does support 16 bit
		// it does not support 8 bit linear
		// it does support stereo
		//fprintf(stderr, "found dbri hardware \n");

		g_supports16		= 1;
		g_supports8			= 0;
		g_supportsStereo	= 1;   
		g_convertUnsigned	= 0;

		return 0;
	}

	if (strcmp(audioamdStr, deviceInfo.name) == 0)
	{
		// audioamd does not support 16 bit
		// it does not support 8 bit linear
		// it does not support stereo
		//fprintf(stderr, "found audioamd hardware \n");

		g_supports16		= 0;
		g_supports8			= 0;
		g_supportsStereo	= 0;   
		g_convertUnsigned	= 0;

		return 0;
	}

	// hardware not identified.  just assume defaults apply....
	//fprintf(stderr, "hardware not identified\n");
	return 0;
}

// Cleanup function. Called after all memory and the last buffers are deallocated, and
// audio hardware is released. Can be used to unload a DLL, library, etc.
// return 0 for ok, or -1 for failure
int HAE_Cleanup(void)
{
	return 0;
}

// **** Memory management
// allocate a block of locked, zeroed memory. Return a pointer
void * HAE_Allocate(unsigned long size)
{
    // $$kk: 10.14.97
    // changed this line as per Liang He's performance recommendations
    //	data = (char *)malloc(size);

    // $$kk: 10.23.97
    // changed this line as per Liang He's performance revommendations.
    // buffer should be 8-byte aligned, not 4-byte aligned.
	//data = (char *)memalign(8, size);

	char *data;

	data = (char *)memalign(8, size + 8);
	if (data)
	{
		memset(data, 0, size);		// clear memory
	}
	return data;
}

// dispose of memory allocated with HAE_Allocate
void HAE_Deallocate(void * memoryBlock)
{
	if (memoryBlock)
	{
		free(memoryBlock);
	}
}

// Given a memory pointer and a size, validate of memory pointer is a valid memory address
// with at least size bytes of data avaiable from the pointer.
// This is used to determine if a memory pointer and size can be accessed without 
// causing a memory protection
// fault.
// return 0 for valid, or 1 for bad pointer, or 2 for not supported. 
int HAE_IsBadReadPointer(void *memoryBlock, unsigned long size)
{
	return 2;
}

// this will return the size of the memory pointer allocated with HAE_Allocate. Return
// 0 if you don't support this feature
unsigned long HAE_SizeOfPointer(void * memoryBlock)
{
	return 0;
}

// block move memory. This is basicly memcpy, but its exposed to take advantage of
// special block move speed ups, various hardware has available.
void HAE_BlockMove(void * source, void * dest, unsigned long size)
{
	if (source && dest && size)
	{
		memcpy(dest, source, size);
	}
}

// **** Audio Card modifiers
// Return 1 if stereo hardware is supported, otherwise 0.
int HAE_IsStereoSupported(void)
{
	return (int)g_supportsStereo;
}

// Return 1, if sound hardware support 16 bit output, otherwise 0.
int HAE_Is16BitSupported(void)
{
	return (int)g_supports16;
}


// Return 1, if sound hardware support 8 bit output, otherwise 0.
// some solaris drivers support 8bit u-law or a-law only, so they are not supported
int HAE_Is8BitSupported(void)
{
	return (int)g_supports8;
}


// returned volume is in the range of 0 to 256
short int HAE_GetHardwareVolume(void)
{
	return 256;
}

// theVolume is in the range of 0 to 256
void HAE_SetHardwareVolume(short int theVolume)
{
}

// **** Timing services
// return microseconds
unsigned long HAE_Microseconds(void)
{
	static hrtime_t solaristick = 0;

	if (solaristick == 0)
	{
		solaristick = gethrtime();
	}
	/* ghrtime is in nanoseconds, we want microseconds. */
	return (unsigned long) ((gethrtime() - solaristick) / 1000);
}

// wait or sleep this thread for this many microseconds
void HAE_WaitMicroseocnds(unsigned long waitAmount)
{
	unsigned long	ticks;

	ticks = HAE_Microseconds() + waitAmount;
	while (HAE_Microseconds() < ticks) 
	{
// kk: i don't think Sleep() exists on solaris
//		Sleep(0);	// Give up the rest of this time slice to other threads
	}
}

// **** File support
// Create a file, delete orignal if duplicate file name.
// Return -1 if error

// Given the fileNameSource that comes from the call HAE_FileXXXX, copy the name
// in the native format to the pointer passed fileNameDest.
void HAE_CopyFileNameNative(void *fileNameSource, void *fileNameDest)
{
	char *dest;
	char *src;

	if (fileNameSource && fileNameDest)
	{
		dest = (char *)fileNameDest;
		src = (char *)fileNameSource;
		if (src == NULL)
		{
			src = "";
		}
		if (dest)
		{
			while (*src)
			{
				*dest++ = *src++;
			}
			*dest = 0;
		}
	}
}

long HAE_FileCreate(void *fileName)
{
	int		file;

	file = open((char *)fileName, O_CREAT | O_RDWR | O_TRUNC);
	if (file != -1)
	{
		close(file);
	}
	return (file != -1) ? 0 : -1;
}


// Open a file
// Return -1 if error, otherwise file handle
long HAE_FileOpenForRead(void *fileName)
{
	if (fileName)
	{
		return open((char *)fileName, O_RDONLY);
	}
	return -1;
}

long HAE_FileOpenForWrite(void *fileName)
{
	if (fileName)
	{
		return open((char *)fileName, O_CREAT | O_RDWR | O_TRUNC);
	}
	return -1;
}

long HAE_FileOpenForReadWrite(void *fileName)
{
	if (fileName)
	{
		return open((char *)fileName, O_RDWR);
	}
	return -1;
}

// Close a file
void HAE_FileClose(long fileReference)
{
	close(fileReference);
}

// Read a block of memory from a file.
// Return -1 if error, otherwise length of data read.
long HAE_ReadFile(long fileReference, void *pBuffer, long bufferLength)
{
	if (pBuffer && bufferLength)
	{
		return read(fileReference, (char *)pBuffer, bufferLength);
	}
	return -1;
}

// Write a block of memory from a file
// Return -1 if error, otherwise length of data written.
long HAE_WriteFile(long fileReference, void *pBuffer, long bufferLength)
{
	if (pBuffer && bufferLength)
	{
		return write(fileReference, (char *)pBuffer, bufferLength);
	}
	return -1;
}

// set file position in absolute file byte position
// Return -1 if error, otherwise 0.
long HAE_SetFilePosition(long fileReference, unsigned long filePosition)
{
	return (lseek(fileReference, filePosition, SEEK_SET) == -1) ? -1 : 0;
}

// get file position in absolute file bytes
unsigned long HAE_GetFilePosition(long fileReference)
{
	return lseek(fileReference, 0, SEEK_CUR);
}

// get length of file
unsigned long HAE_GetFileLength(long fileReference)
{
	unsigned long pos;

	pos = lseek(fileReference, 0, SEEK_END);
	lseek(fileReference, 0, SEEK_SET);
	return pos;
}

// set the length of a file. Return 0, if ok, or -1 for error
int HAE_SetFileLength(long fileReference, unsigned long newSize)
{
//	return chsize(fileReference, newSize);
	return -1;
}

void PV_AudioWaveOutFrameThread(void* context)
{
    audio_info_t	sunAudioHeader;
    char			*pFillBuffer;

	//$$kk: 05.06.98: i am making lastPos global to fix P1 bug #4135246, in
	// which the jdk samples sometimes do not get audio.  depending
	// on the order of events when the audio device is opened and
	// closed, the frame thread may get stuck in the wait().
    long			count, currentPos, /*lastPos,*/ sampleFrameSize;
    
	unsigned long	startTime, stopTime, fillTime;
	int				i;

	ioctl(g_waveDevice, AUDIO_GETINFO, &sunAudioHeader);

	// calculate sample size for convertion of bytes to sample frames
	sampleFrameSize = 1;
		
	if (g_bitSize == 16)
	{
		sampleFrameSize *= 2;
	}

	if (g_channels == 2)
	{
		sampleFrameSize *= 2;
	}


	lastPos = sunAudioHeader.play.samples - ((g_audioByteBufferSize * HAE_SOLARIS_FRAMES_PER_BLOCK * 2) / sampleFrameSize);

	if (g_audioBufferBlock)
    {
		while ( (g_activeDoubleBuffer) && (g_shutDownDoubleBuffer == FALSE) )
		{
		    // Generate HAE_SOLARIS_FRAMES_PER_BLOCK frames of audio
		    pFillBuffer = (char *)g_audioBufferBlock;	
		    for (count = 0; count < HAE_SOLARIS_FRAMES_PER_BLOCK; count++)
			{
				// Generate one frame audio
				HAE_BuildMixerSlice(context, pFillBuffer, g_audioByteBufferSize,
															g_audioFramesToGenerate);
				pFillBuffer += g_audioByteBufferSize;

				if (g_shutDownDoubleBuffer)
				{
					break;	// time to quit
				}
		    }

			// for some solaris drivers, we must supply unsigned data when rendering 8 bit data
			if (g_convertUnsigned && (g_bitSize == 8))
			{
				pFillBuffer = (char *)g_audioBufferBlock;	
				for (i = 0; i < (g_audioByteBufferSize * HAE_SOLARIS_FRAMES_PER_BLOCK); i++)
				{
					*pFillBuffer = (*pFillBuffer >= 0) ? (0x80 | *pFillBuffer) : (0x7F & *pFillBuffer);
					pFillBuffer++;
				}
			}
	
			write(g_waveDevice, g_audioBufferBlock, g_audioByteBufferSize * HAE_SOLARIS_FRAMES_PER_BLOCK);

  	
			// $$kk: 05.06.98: i am incrementing lastPos right after the write operation
			// to lessen the chance of a device open/close messing us up in the wait loop.
			lastPos += (g_audioByteBufferSize * HAE_SOLARIS_FRAMES_PER_BLOCK) / sampleFrameSize;


		    // O.k. We're done for now.
		    // Let the rest of the system know we're done ....

			ioctl(g_waveDevice, AUDIO_GETINFO, &sunAudioHeader);
			currentPos = sunAudioHeader.play.samples;

			while ((currentPos < lastPos) && (g_shutDownDoubleBuffer == FALSE))
			{
				// $$kk: LocalThreadSleep does not exist
//				LocalThreadSleep(g_audioPeriodSleepTime);		// 10 ms
				HAE_SleepFrameThread(context, HAE_SOLARIS_SOUND_PERIOD);		// in ms
				
				ioctl(g_waveDevice, AUDIO_GETINFO, &sunAudioHeader);
				currentPos = sunAudioHeader.play.samples;
			}
			
			// $$kk: 05.06.98: moved this line up.
			//lastPos += (g_audioByteBufferSize * HAE_SOLARIS_FRAMES_PER_BLOCK) / sampleFrameSize;

//		    // ... and reschedule ourselves.
		}
		g_activeDoubleBuffer = FALSE;
    }
}


// Return the number of 11 ms buffer blocks that are built at one time.
int HAE_GetAudioBufferCount(void)
{
	return g_synthFramesPerBlock;
}

// Return the number of bytes used for audio buffer for output to card
long HAE_GetAudioByteBufferSize(void)
{
	return g_audioByteBufferSize;
}


// **** Audio card support
// Aquire and enabled audio card
// return 0 if ok, -1 if failed
int HAE_AquireAudioCard(void *context, long sampleRate, long channels, long bits)
{
	int				flag;
	short int		count;
	long			error;
    audio_info_t	sunAudioHeader;

	flag = 0;
	g_activeDoubleBuffer = FALSE;
	g_shutDownDoubleBuffer = TRUE;

	switch (sampleRate)
	{
		case 44100:
			g_audioFramesToGenerate = HAE_GetMaxSamplePerSlice();
			break;
		case 11025:
			sampleRate = 22050;
		case 22050:
			g_audioFramesToGenerate = HAE_GetMaxSamplePerSlice()/2;
			break;
	}

	// we're going to build this many buffers at a time
	g_synthFramesPerBlock = HAE_SOLARIS_FRAMES_PER_BLOCK;
	g_audioPeriodSleepTime = HAE_SOLARIS_SOUND_PERIOD;
	g_bitSize = bits;
	g_channels = channels;
	if (bits == 8)
	{
		g_audioByteBufferSize = (sizeof(char) * g_audioFramesToGenerate);
	}
	else
	{
		g_audioByteBufferSize = (sizeof(short int) * g_audioFramesToGenerate);
	}
	g_audioByteBufferSize *= channels;

	flag = 1;
	// allocate buffer blocks
	g_audioBufferBlock = HAE_Allocate(g_audioByteBufferSize * HAE_SOLARIS_FRAMES_PER_BLOCK);
	if (g_audioBufferBlock)
	{
		// try to open wave device
// $$kk: 12.17.97: need O_NONBLOCK flag to be compatible with windows
// change next 1 line
//	    g_waveDevice = open("/dev/audio",O_WRONLY);		
		g_waveDevice = open("/dev/audio",O_WRONLY|O_NONBLOCK);
	    if (g_waveDevice > 0)
	    {
			AUDIO_INITINFO(&sunAudioHeader);
// $$kk: 12.17.97: need AUDIO_GETINFO ioctl to get this to work on solaris x86 
// add next 1 line
			error = ioctl(g_waveDevice, AUDIO_GETINFO, &sunAudioHeader);

// $$kk: 03.16.98: not valid to call AUDIO_SETINFO ioctl with all the fields from AUDIO_GETINFO,
// so let's try init'ing again....
			AUDIO_INITINFO(&sunAudioHeader);

			// Set rendering format of the sun device.
			sunAudioHeader.play.sample_rate = sampleRate;
			sunAudioHeader.play.precision = bits;
			sunAudioHeader.play.channels = channels;
			sunAudioHeader.play.encoding = AUDIO_ENCODING_LINEAR;
		
			error = ioctl(g_waveDevice, AUDIO_SETINFO, &sunAudioHeader);

			if (error == 0)
			{
				g_shutDownDoubleBuffer = FALSE;
				g_activeDoubleBuffer = TRUE;	// must enable process, before thread begins


				// $$kk: 05.06.98: added this whole block.
				// we need to reset the lastPos each time the device gets acquired.
				// otherwise we may get stuck in the wait loop because we never count
				// up to the right sample position.
				{
				    audio_info_t	sunAudioHeader;
					long sampleFrameSize;

					ioctl(g_waveDevice, AUDIO_GETINFO, &sunAudioHeader);

					// calculate sample size for convertion of bytes to sample frames
					sampleFrameSize = 1;
					if (g_bitSize == 16)
					{
						sampleFrameSize *= 2;
					}
					if (g_channels == 2)
					{
						sampleFrameSize *= 2;
					}
					lastPos = sunAudioHeader.play.samples - ((g_audioByteBufferSize * HAE_SOLARIS_FRAMES_PER_BLOCK * 2) / sampleFrameSize);
				}

				
				/* Spin threads for device service and possibly
				 * stream service.
				 */
				// create thread to manage and render audio frames
				error = HAE_CreateFrameThread(context, PV_AudioWaveOutFrameThread);

				if (error == 0)
				{	// ok
					flag = 0;
				}
				else
				{
					flag = 1;
					g_activeDoubleBuffer = FALSE;
				}
			}
		}
	}

	if (flag)
	{	// something failed
		HAE_ReleaseAudioCard(context);
	}
	return flag;
}

// Release and free audio card.
// return 0 if ok, -1 if failed.
int HAE_ReleaseAudioCard(void *context)
{
	context = context;

	g_shutDownDoubleBuffer = TRUE;	// signal thread to stop
	HAE_DestroyFrameThread(context);

	if (g_waveDevice)
	{
		close(g_waveDevice);
		g_waveDevice = 0;
	}

	if (g_audioBufferBlock)
	{
		HAE_Deallocate(g_audioBufferBlock);
		g_audioBufferBlock = NULL;
	}
	return 0;
}

// return device position in samples
unsigned long HAE_GetDeviceSamplesPlayedPosition(void)
{
    audio_info_t	sunAudioHeader;
	unsigned long	pos;

	pos = 0;
	if (g_waveDevice)
	{
		ioctl(g_waveDevice, AUDIO_GETINFO, &sunAudioHeader);
		pos = sunAudioHeader.play.samples;
	}
	return pos;
}

// number of devices. ie different versions of the HAE connection. DirectSound and waveOut
// return number of devices. ie 1 is one device, 2 is two devices.
// NOTE: This function needs to function before any other calls may have happened.
long HAE_MaxDevices(void)
{
	return 1;	// only /dev/audio
}

// set the current device. device is from 0 to HAE_MaxDevices()
// NOTE:	This function needs to function before any other calls may have happened.
//			Also you will need to call HAE_ReleaseAudioCard then HAE_AquireAudioCard
//			in order for the change to take place.
void HAE_SetDeviceID(long deviceID)
{
	if (deviceID < HAE_MaxDevices())
	{
		g_currentDeviceID = deviceID;
	}
}

// return current device ID
// NOTE: This function needs to function before any other calls may have happened.
long HAE_GetDeviceID(void)
{
	return g_currentDeviceID;
}

// NOTE:	This function needs to function before any other calls may have happened.
//			Format of string is a zero terminated comma delinated C string.
//			"platform,method,misc"
//	example	"MacOS,Sound Manager 3.0,SndPlayDoubleBuffer"
//			"WinOS,DirectSound,multi threaded"
//			"WinOS,waveOut,multi threaded"
//			"WinOS,VxD,low level hardware"
//			"WinOS,plugin,Director"
void HAE_GetDeviceName(long deviceID, char *cName, unsigned long cNameLength)
{
	char		*data;
	static char *names[] =
			{	"SolarisOS,dev/audio,multi threaded",
			};
	if (cName && cNameLength)
	{
		if (deviceID < HAE_MaxDevices())
		{
			// can't do string functions here, it might be bad to do.
			data = names[deviceID];
			while (*data && (cNameLength > 0))
			{
				*cName = *data;
				cName++;
				data++;
				cNameLength--;
			}
			*cName = 0;
		}
	}
}


// EOF of HAE_API_SolarisOS.c
