/*
 * @(#)awt_InputMethod.c	1.32 98/06/28
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <Xm/Xm.h>
#include <Xm/RowColumn.h>
#include <Xm/MwmUtil.h>
#include <Xm/MenuShell.h>
#include <jni_util.h>
#include <jni.h>

#include "awt.h"
#include "awt_p.h"

#include <sun_awt_motif_MComponentPeer.h>
#include <sun_awt_motif_X11InputMethod.h>

#define THROW_OUT_OF_MEMORY_ERROR() \
	JNU_ThrowOutOfMemoryError((JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2), NULL)
#define SETARG(name, value)	XtSetArg(args[argc], name, value); argc++

#define MCOMPONENTPEER_CLASS_NAME	"sun/awt/motif/MComponentPeer"

struct X11InputMethodIDs {
  jfieldID pData;
} x11InputMethodIDs;

extern XVaNestedList awt_util_getXICStatusAreaList(Widget);
extern Widget awt_util_getXICStatusAreaWindow(Widget);

static void PreeditStartCallback(XIC, XPointer, XPointer);
static void PreeditDoneCallback(XIC, XPointer, XPointer);
static void PreeditDrawCallback(XIC, XPointer,
				XIMPreeditDrawCallbackStruct *);
static void PreeditCaretCallback(XIC, XPointer,
				 XIMPreeditCaretCallbackStruct *);

#define ROOT_WINDOW_STYLES	(XIMPreeditNothing | XIMStatusNothing)
#define NO_STYLES		(XIMPreeditNone | XIMStatusNone)

#define PreeditStartIndex	0
#define PreeditDoneIndex	1
#define PreeditDrawIndex	2
#define PreeditCaretIndex	3
#define NCALLBACKS		4

/*
 * Callback function pointers: the order has to match the *Index
 * values above.
 */
static XIMProc callback_funcs[NCALLBACKS] = {
    (XIMProc)PreeditStartCallback,
    (XIMProc)PreeditDoneCallback,
    (XIMProc)PreeditDrawCallback,
    (XIMProc)PreeditCaretCallback,
};

/*
 * X11InputMethodData keeps per X11InputMethod instance information. A pointer
 * to this data structure is kept in an X11InputMethod object (pData).
 */
typedef struct _X11InputMethodData {
    XIC		current_ic;	/* current X Input Context */
    XIC		ic_active;	/* X Input Context for active clients */
    XIC		ic_passive;	/* X Input Context for passive clients */
    XIMCallback	*callbacks;	/* callback parameters */
    jobject	peer;		/* MComponentPeer of client Window */
    jobject	x11inputmethod;	/* global ref to X11InputMethod instance */
				/* associated with the XIC */
    Widget	statusWidget;	/* IM status window widget */
} X11InputMethodData;

jobject currentX11InputMethodInstance = NULL; /* reference to the current
						 X11InputMethod instance */
Window  currentFocusWindow = NULL;  /* current window that has focus for input
				       method. (the best place to put this
				       information should be
				       currentX11InputMethodInstance's pData) */
static XIM X11im = NULL;
static jobject  mcompClass = NULL;
static jobject  awteventClass = NULL;

/*
 * Converts the wchar_t string to a multi-byte string calling wcstombs(). A
 * buffer is allocated by malloc() to store the multi-byte string. NULL is
 * returned if the given wchar_t string pointer is NULL or buffer allocation is
 * failed.
 */
static char *
wcstombsdmp(wchar_t *wcs, int len)
{
    size_t n;
    char *mbs;

    if (wcs == NULL)
	return NULL;

    n = len*MB_CUR_MAX + 1;

    mbs = (char *) malloc(n * sizeof(char));
    if (mbs == NULL) {
	THROW_OUT_OF_MEMORY_ERROR();
	return NULL;
    }

    /* TODO: check return values... Handle invalid characters properly...  */
    if (wcstombs(mbs, wcs, n) == (size_t)-1)
	return NULL;

    return mbs;
}

/*
 * Find a class for the given class name and return a global reference to the
 * class.
 */
static jobject
findClass(const char *className)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    jclass classClass;
    jobject objectClass;

    classClass = (*env)->FindClass(env, className);
    objectClass = (*env)->NewGlobalRef(env,classClass);

    if (JNU_IsNull(env, objectClass)) {
        JNU_ThrowClassNotFoundException(env, className);
    }
    return objectClass;
}

static void
destroyX11InputMethodData(JNIEnv *env, X11InputMethodData *pX11IMData)
{
    if (pX11IMData->callbacks)
	free((void *)pX11IMData->callbacks);
    /*
     * Destroy XICs
     */
    if (pX11IMData->ic_active != (XIC)0) {
#if 0
	/*
	 * This XmbResetIC call may be necessary to avoid a hang of
	 * XIM. However, at this point the client component may be
	 * gone, which causes another hang. This part disabled until
	 * the synchronization problem between the client component and
	 * the input method is resolved.
	 */
	char *committedString;

	if ((committedString = XmbResetIC(pX11IMData->ic_active)) != NULL)
	    XFree((void *)committedString);
#endif
	XDestroyIC(pX11IMData->ic_active);
	if (pX11IMData->ic_active != pX11IMData->ic_passive) {
#if 0
	    if ((committedString = XmbResetIC(pX11IMData->ic_passive)) != NULL)
		XFree((void *)committedString);
#endif
	    if (pX11IMData->current_ic != (XIC)0)
		XDestroyIC(pX11IMData->ic_passive);
	    pX11IMData->ic_passive = (XIC)0;
	    pX11IMData->current_ic = (XIC)0;
	}
    }
    if (env) {
	(*env)->DeleteGlobalRef(env, pX11IMData->peer);
	(*env)->DeleteGlobalRef(env, pX11IMData->x11inputmethod);
    }
    free((void *)pX11IMData);
}

/*
 * Sets or unsets the focus to the given XIC.
 */
static void
setXICFocus(XIC ic, unsigned short req)
{
    if (ic == NULL) {
        (void)fprintf(stderr, "Couldn't find X Input Context\n");
        return;
    }
    if (req == 1)
        XSetICFocus(ic);
    else
        XUnsetICFocus(ic);
}

/*
 * Sets the focus window to the given XIC.
 */
static void
setXICWindowFocus(XIC ic, Window w)
{
    if (ic == NULL) {
        (void)fprintf(stderr, "Couldn't find X Input Context\n");
        return;
    }
    (void) XSetICValues(ic, XNFocusWindow, w, NULL);
}

/*
 * Invokes XmbLookupString() to get something from the XIM. It invokes
 * X11InputMethod.dispatchCommittedText() if XmbLookupString() returns
 * committed text.  This function is called from handleKeyEvent in canvas.c and
 * it's under the Motif event loop thread context.
 *
 * Buffer usage: The function uses the local buffer first. If it's too small to
 * get the committed text, it allocates memory for the buffer. Note that
 * XmbLookupString() sometimes produces a non-null-terminated string.
 *
 * Returns True when there is a keysym value to be handled.
 */
#define LOOKUP_BUF_SIZE 64

Bool
awt_x11inputmethod_lookupString(XEvent *event, KeySym *keysymp)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    X11InputMethodData *pX11IMData;
    int buf_len = LOOKUP_BUF_SIZE;
    char mbbuf[LOOKUP_BUF_SIZE];
    char *buf;
    KeySym keysym = NoSymbol;
    Status status;
    int mblen;
    jstring javastr;
    XIC ic;
    Bool result = True;
    static Bool composing = False;

/*
      printf("lookupString: entering...\n");
     */
    pX11IMData = (X11InputMethodData *)
      (*env)->GetLongField(env, currentX11InputMethodInstance,
			   x11InputMethodIDs.pData);

    if (pX11IMData == NULL) {
        (void)fprintf(stderr, "Couldn't find X Input method Context\n");
        return result;
    }

    if ((ic = pX11IMData->current_ic) == (XIC)0)
	return result;

    buf = mbbuf;
    mblen = XmbLookupString(ic, (XKeyPressedEvent *)event, buf, buf_len - 1, &keysym, &status);

    /*
     * In case of overflow, a buffer is allocated and it retries
     * XmbLookupString().
     */
    if (status == XBufferOverflow) {
        buf_len = mblen + 1;
	buf = (char *)malloc(buf_len);
	if (buf == NULL) {
	    THROW_OUT_OF_MEMORY_ERROR();
	    return result;
	}
        mblen = XmbLookupString(ic, (XKeyPressedEvent *)event, buf, buf_len, &keysym, &status);
    }
    buf[mblen] = 0;

    switch (status) {
    case XLookupBoth:
	if (!composing) {
	    if (keysym < 128 || ((keysym & 0xff00) == 0xff00)) {
		*keysymp = keysym;
		result = False;
		break;
	    }
	}
	composing = False;
	/*FALLTHRU*/ 
    case XLookupChars:
/*
printf("lookupString: status=XLookupChars, type=%d, state=%x, keycode=%x, keysym=%x\n",
       event->type, event->xkey.state, event->xkey.keycode, keysym);
*/
        javastr = JNU_NewStringPlatform(env, (const char *)buf);
        if (javastr != NULL) {
            JNU_CallMethodByName(env, NULL,
				 currentX11InputMethodInstance,
				 "dispatchCommittedText",
				 "(Ljava/lang/String;)V",
				 javastr);
        }
	break;

    case XLookupKeySym:
/*
printf("lookupString: status=XLookupKeySym, type=%d, state=%x, keycode=%x, keysym=%x\n",
       event->type, event->xkey.state, event->xkey.keycode, keysym);
*/
 	if (keysym == XK_Multi_key)
	    composing = True;
 	if (! composing) {
 	    *keysymp = keysym;
 	    result = False;
 	}
	break;

    case XLookupNone:
/*
printf("lookupString: status=XLookupNone, type=%d, state=%x, keycode=%x, keysym=%x\n",
       event->type, event->xkey.state, event->xkey.keycode, keysym);
*/
	break;
    }

    if (buf != mbbuf)
	free((void *)buf);

    return result;
}


/*
 * Creates two XICs, one for active clients and the other for passive
 * clients. All information on those XICs are stored in the
 * X11InputMethodData given by the pX11IMData parameter.
 * 
 * For active clients: Try to use preedit callback to support
 * on-the-spot. If statusArea is true, the XIC to be created will
 * share the Status Area with Motif widgets (TextComponents). If the
 * preferable styles can't be used, fallback to root-window styles. If
 * root-window styles failed, fallback to None styles.
 * 
 * For passive clients: Try to use root-window styles. If failed,
 * fallback to None styles.
 */
static Bool
createXIC(Widget w, X11InputMethodData *pX11IMData,
	  jboolean statusArea, jobject peer)
{
    XIC active_ic, passive_ic;
    XVaNestedList preedit = NULL;
    XIMStyle on_the_spot_styles = XIMPreeditCallbacks,
	     active_styles = 0,
	     passive_styles = 0,
	     no_styles = 0;
    XIMCallback *callbacks;
    unsigned short i;
    XIMStyles *im_styles;

    if (X11im == NULL) {
        (void)fprintf(stderr, "Couldn't find X Input method\n");
        return False;
    }

    /*
     * If the parent window has one or more TextComponents, the status
     * area of Motif will be shared with the created XIC. Otherwise,
     * root-window style status is used.
     */
    on_the_spot_styles |= (statusArea == JNI_TRUE) ? XIMStatusArea : XIMStatusNothing;

    (void) XGetIMValues(X11im, XNQueryInputStyle, &im_styles, NULL);

    for (i = 0; i < im_styles->count_styles; i++) {
        active_styles |= im_styles->supported_styles[i] & on_the_spot_styles;
	passive_styles |= im_styles->supported_styles[i] & ROOT_WINDOW_STYLES;
	no_styles |= im_styles->supported_styles[i] & NO_STYLES;
    }

    XFree(im_styles);

    if (active_styles != on_the_spot_styles) {
	if (passive_styles == ROOT_WINDOW_STYLES)
	    active_styles = passive_styles;
	else {
	    if (no_styles == NO_STYLES)
		active_styles = passive_styles = NO_STYLES;
	    else
		active_styles = passive_styles = 0;
	}
    } else {
	if (passive_styles != ROOT_WINDOW_STYLES) {
	    if (no_styles == NO_STYLES)
		active_styles = passive_styles = NO_STYLES;
	    else
		active_styles = passive_styles = 0;
	}
    }

    if (active_styles == on_the_spot_styles) {
	callbacks = (XIMCallback *)malloc(sizeof(XIMCallback) * NCALLBACKS);
	if (callbacks == (XIMCallback *)NULL)
	    return False;
	pX11IMData->callbacks = callbacks;

	for (i = 0; i < NCALLBACKS; i++, callbacks++) {
	    callbacks->client_data = (XPointer) pX11IMData;
	    callbacks->callback = callback_funcs[i];
	}

	callbacks = pX11IMData->callbacks;
	preedit = (XVaNestedList)XVaCreateNestedList(0,
			XNPreeditStartCallback, &callbacks[PreeditStartIndex],
			XNPreeditDoneCallback,  &callbacks[PreeditDoneIndex],
			XNPreeditDrawCallback,  &callbacks[PreeditDrawIndex],
			XNPreeditCaretCallback, &callbacks[PreeditCaretIndex],
			NULL);
	if (preedit == (XVaNestedList)NULL)
	    goto err;

	if (on_the_spot_styles & XIMStatusArea) {
	    XVaNestedList status;
	    Widget parent;

	    status = awt_util_getXICStatusAreaList(w);
	    if (status == NULL)
		goto err;
	    pX11IMData->statusWidget = awt_util_getXICStatusAreaWindow(w);
	    pX11IMData->ic_active = XCreateIC(X11im,
					      XNClientWindow, XtWindow(pX11IMData->statusWidget),
					      XNFocusWindow, XtWindow(w), 
					      XNInputStyle, active_styles,
					      XNPreeditAttributes, preedit,
					      XNStatusAttributes, status,
					      NULL);
	    XFree((void *)status);
	} else {
	    pX11IMData->ic_active = XCreateIC(X11im,
					      XNClientWindow, XtWindow(w),
					      XNFocusWindow, XtWindow(w), 
					      XNInputStyle, active_styles,
					      XNPreeditAttributes, preedit,
					      NULL);
	}
	XFree((void *)preedit);
	pX11IMData->ic_passive = XCreateIC(X11im,
					   XNClientWindow, XtWindow(w),
					   XNFocusWindow, XtWindow(w), 
					   XNInputStyle, passive_styles,
					   NULL);
    } else {
	pX11IMData->ic_active = XCreateIC(X11im,
					  XNClientWindow, XtWindow(w),
					  XNFocusWindow, XtWindow(w),
					  XNInputStyle, active_styles,
					  NULL);
	pX11IMData->ic_passive = pX11IMData->ic_active;
    }

    if (pX11IMData->ic_active == (XIC)0
	|| pX11IMData->ic_passive == (XIC)0) {
	return False;
    }
    return True;

 err:
    if (preedit)
	XFree((void *)preedit);
    THROW_OUT_OF_MEMORY_ERROR();
    return False;
}

static void
PreeditStartCallback(XIC ic, XPointer client_data, XPointer call_data)
{
    /*ARGSUSED*/
    /* printf("Native: PreeditCaretCallback\n"); */
}

static void
PreeditDoneCallback(XIC ic, XPointer client_data, XPointer call_data)
{
    /*ARGSUSED*/
    /* printf("Native: StatusStartCallback\n"); */
}

/*
 * Translate the preedit draw callback items to Java values and invoke
 * X11InputMethod.dispatchComposedText().
 *
 * client_data: pointer to X11InputMethodData
 */
static void
PreeditDrawCallback(XIC ic, XPointer client_data,
		    XIMPreeditDrawCallbackStruct *pre_draw)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    X11InputMethodData *pX11IMData = (X11InputMethodData *) client_data;
    jmethodID x11imMethodID;

    XIMText *text;
    jstring javastr = NULL;
    jintArray style = NULL;

    /*    printf("Native: PreeditDrawCallback() \n"); */

    if (pre_draw == NULL)
        return;

    if ((text = pre_draw->text) != NULL) {
	if (text->string.multi_byte != NULL) {
	    if (pre_draw->text->encoding_is_wchar == False) {
		javastr = JNU_NewStringPlatform(env, (const char *)text->string.multi_byte);
	    } else {
		char *mbstr = wcstombsdmp(text->string.wide_char, text->length);
		if (mbstr == NULL)
		    return;
		javastr = JNU_NewStringPlatform(env, (const char *)mbstr);
		free(mbstr);
	    }
	}
	if (text->feedback != NULL) {
            int cnt;
            jint *tmpstyle;

            style = (*env)->NewIntArray(env, text->length);
            if (JNU_IsNull(env, style)) {
                THROW_OUT_OF_MEMORY_ERROR();
                return;
            }

	    if (sizeof(XIMFeedback) == sizeof(jint)) {
		/*
		 * Optimization to avoid copying the array
		 */
		(*env)->SetIntArrayRegion(env, style, 0,
					  text->length, (jint *)text->feedback);
	    } else {
		tmpstyle  = (jint *)malloc(sizeof(jint)*(text->length));
		if (tmpstyle == (jint *) NULL) {
		    THROW_OUT_OF_MEMORY_ERROR();
		    return;
		}
		for (cnt = 0; cnt < (int)text->length; cnt++)
			tmpstyle[cnt] = text->feedback[cnt];
		(*env)->SetIntArrayRegion(env, style, 0,
					  text->length, (jint *)tmpstyle);
	    }
        }
    }

    JNU_CallMethodByName(env, NULL, pX11IMData->x11inputmethod,
			 "dispatchComposedText",
			 "(Ljava/lang/String;[IIII)V",
			 javastr,
			 style,
			 (jint)pre_draw->chg_first,
			 (jint)pre_draw->chg_length,
			 (jint)pre_draw->caret);
}

static void
PreeditCaretCallback(XIC ic, XPointer client_data,
		     XIMPreeditCaretCallbackStruct *pre_caret)
{
    /*ARGSUSED*/
    /* printf("Native: PreeditCaretCallback\n"); */

}

/*
 * Class:     java_sun_awt_motif_X11InputMethod
 * Method:    initIDs
 * Signature: ()V
 */

/* This function gets called from the static initializer for
   X11InputMethod.java
   to initialize the fieldIDs for fields that may be accessed from C */
JNIEXPORT void JNICALL
Java_sun_awt_motif_X11InputMethod_initIDs(JNIEnv *env, jclass cls)
{
    x11InputMethodIDs.pData = (*env)->GetFieldID(env, cls, "pData", "J");
}


JNIEXPORT jboolean JNICALL
Java_sun_awt_motif_X11InputMethod_openXIM(JNIEnv *env,
					  jobject this)
{
    extern int xerror_handler();

    AWT_LOCK();

    X11im = XOpenIM(awt_display, NULL, NULL, NULL);

    /* Workaround for Solaris 2.6 bug 4097754. We're affected by this problem 
     * because Motif also calls XOpenIM for us. Re-registering the error handler
     * that MToolkit has registered already after calling XOpenIM avoids the
     * problem.
     */
    XSetErrorHandler(xerror_handler);

    AWT_UNLOCK();
    return (X11im != NULL) ? JNI_TRUE : JNI_FALSE;
}

JNIEXPORT jlong JNICALL
Java_sun_awt_motif_X11InputMethod_createXICNative(JNIEnv *env,
						  jobject this,
						  jobject comp,
						  jboolean statusArea)
{
    struct ComponentData *cdata;
    X11InputMethodData *pX11IMData;
    jobject globalRef;
    jfieldID mcompPDataID;
    XIC ic;

    AWT_LOCK();

    if (JNU_IsNull(env, comp)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return (jint)NULL;
    }

    pX11IMData = (X11InputMethodData *) calloc(1, sizeof(X11InputMethodData));
    if (pX11IMData == NULL) {
	THROW_OUT_OF_MEMORY_ERROR();
	AWT_UNLOCK();
	return (jint)NULL;
    }

    if (mcompClass == NULL)
	mcompClass = findClass(MCOMPONENTPEER_CLASS_NAME);

    mcompPDataID = (*env)->GetFieldID(env, mcompClass, "pData", "J");
    cdata = (struct ComponentData *) (*env)->GetLongField(env,comp,mcompPDataID);

    if (cdata == 0) {
	free((void *)pX11IMData);
        JNU_ThrowNullPointerException(env, "createXIC");
        AWT_UNLOCK();
        return (jint)NULL;
    }

    globalRef = (*env)->NewGlobalRef(env, this);
    pX11IMData->peer = (*env)->NewGlobalRef(env, comp);
    pX11IMData->x11inputmethod = globalRef;
    pX11IMData->statusWidget = (Widget) NULL;

    if (createXIC(cdata->widget, pX11IMData, statusArea, comp)
	== False) {
	destroyX11InputMethodData((JNIEnv *) NULL, pX11IMData);
	pX11IMData = (X11InputMethodData *) NULL;
    }

    AWT_UNLOCK();
    return (jint)pX11IMData;
}

JNIEXPORT jlong JNICALL
Java_sun_awt_motif_X11InputMethod_reconfigureXICNative(JNIEnv *env,
						       jobject this,
						       jobject comp,
						       jboolean statusArea,
						       jlong pData)
{
    X11InputMethodData *pX11IMData = (X11InputMethodData *) pData;

    AWT_LOCK();

    if (pX11IMData->current_ic == (XIC)0) {
	destroyX11InputMethodData(env, pX11IMData);
	pX11IMData = (X11InputMethodData *)NULL;
    } else {
	Bool active;
	struct ComponentData *cdata;
	jfieldID mcompPDataID;

	active = pX11IMData->current_ic == pX11IMData->ic_active;
	mcompPDataID = (*env)->GetFieldID(env, mcompClass, "pData", "J");
	cdata = (struct ComponentData *) (*env)->GetLongField(env,comp,mcompPDataID);
	if (cdata == 0) {
	    JNU_ThrowNullPointerException(env, "reconfigureXICNative");
	    destroyX11InputMethodData(env, pX11IMData);
	    pX11IMData = (X11InputMethodData *)NULL;
	}
	XDestroyIC(pX11IMData->ic_active);
	if (pX11IMData->ic_active != pX11IMData->ic_passive)
	    XDestroyIC(pX11IMData->ic_passive);
	pX11IMData->current_ic = (XIC)0;
	pX11IMData->ic_active = (XIC)0;
	pX11IMData->ic_passive = (XIC)0;
	if (createXIC(cdata->widget, pX11IMData, statusArea, comp)) {
	    pX11IMData->current_ic = active ?
			pX11IMData->ic_active : pX11IMData->ic_passive;
	    /*
	     * On Solaris2.6, setXICWindowFocus() has to be invoked
	     * before setting focus.
	     */
	    setXICWindowFocus(pX11IMData->current_ic, XtWindow(cdata->widget));
	    setXICFocus(pX11IMData->current_ic, True);
	} else {
	    destroyX11InputMethodData((JNIEnv *) NULL, pX11IMData);
	    pX11IMData = (X11InputMethodData *)NULL;
	}
    }

    AWT_UNLOCK();
    return (jint) pX11IMData;
}

JNIEXPORT void JNICALL
Java_sun_awt_motif_X11InputMethod_setXICFocus(JNIEnv *env,
					      jobject this,
					      jobject comp,
					      jboolean req,
					      jboolean active,
					      jlong data)
{
    X11InputMethodData *pX11IMData = (X11InputMethodData *) data;
    jfieldID mcompPDataID;
    struct ComponentData *cdata;

    AWT_LOCK();

    if (req) {
        struct ComponentData *cdata;
        jfieldID mcompPDataID;

	if (JNU_IsNull(env, comp)) {
	    AWT_UNLOCK();
	    return;
	}
	if (mcompClass == NULL)
	    mcompClass = findClass(MCOMPONENTPEER_CLASS_NAME);
        mcompPDataID = (*env)->GetFieldID(env, mcompClass, "pData", "J");
        cdata = (struct ComponentData *)(*env)->GetLongField(env,
							     comp,
							     mcompPDataID);
        if (cdata == 0) {
            JNU_ThrowNullPointerException(env, "setXICFocus pData");
            AWT_UNLOCK();
            return;
        }

	pX11IMData->current_ic = active ?
			pX11IMData->ic_active : pX11IMData->ic_passive;
	/*
	 * On Solaris2.6, setXICWindowFocus() has to be invoked
	 * before setting focus.
	 */
        setXICWindowFocus(pX11IMData->current_ic, XtWindow(cdata->widget));
        setXICFocus(pX11IMData->current_ic, req);
        currentX11InputMethodInstance = pX11IMData->x11inputmethod;
        currentFocusWindow =  XtWindow(cdata->widget);
    } else {
	currentX11InputMethodInstance = NULL;
        currentFocusWindow = NULL;
        setXICFocus(pX11IMData->current_ic, req);
	pX11IMData->current_ic = (XIC)0;
    }

    XSync(awt_display, False);
    AWT_UNLOCK();
}

JNIEXPORT void JNICALL
Java_sun_awt_motif_X11InputMethod_disposeXIC(JNIEnv *env,
					     jobject this,
					     jlong pdata)
{
    X11InputMethodData *pX11IMData = (X11InputMethodData *) pdata;

    AWT_LOCK();
    if (pX11IMData->x11inputmethod == currentX11InputMethodInstance) {
	currentX11InputMethodInstance = NULL;
	currentFocusWindow = NULL;
    }
    destroyX11InputMethodData(env, pX11IMData);
    pX11IMData = (X11InputMethodData *) NULL;
    AWT_UNLOCK();
}

JNIEXPORT void JNICALL
Java_sun_awt_motif_X11InputMethod_closeXIM(JNIEnv *env,
					   jobject this)
{
    AWT_LOCK();

    if (X11im != NULL) {
        XCloseIM(X11im);
        X11im = NULL;
    }
    AWT_UNLOCK();
}

JNIEXPORT jstring JNICALL
Java_sun_awt_motif_X11InputMethod_resetXIC(JNIEnv *env,
					   jobject this,
					   jlong data)
{
    X11InputMethodData *pX11IMData = (X11InputMethodData *) data;
    char *xText;
    jstring jText;
 
    AWT_LOCK();

    if (pX11IMData->current_ic)
	xText = XmbResetIC(pX11IMData->current_ic);
    else {
	/*
	 * If there is no reference to the current XIC, try to reset both XICs.
	 */
	xText = XmbResetIC(pX11IMData->ic_active);
	if (pX11IMData->ic_active != pX11IMData->ic_passive) {
	    char *tmpText = XmbResetIC(pX11IMData->ic_passive);
	    if (xText == (char *)NULL && tmpText)
		xText = tmpText;
	}
    }
    if (xText != NULL) {
	jText = JNU_NewStringPlatform(env, (const char *)xText);
	XFree((void *)xText);
    } else {
	jText = (jstring)0;
    }

    AWT_UNLOCK();
    return jText;
}

JNIEXPORT void JNICALL
Java_sun_awt_motif_X11InputMethod_configureStatusAreaNative(JNIEnv *env,
							    jobject this,
							    jlong data)
{
    X11InputMethodData *pX11IMData = (X11InputMethodData *) data;
    XVaNestedList status;

    if (pX11IMData->ic_active == (XIC)0)
	return;

    if (pX11IMData->statusWidget) {
        AWT_LOCK();
        status = awt_util_getXICStatusAreaList(pX11IMData->statusWidget);
        if (status != (XVaNestedList)NULL) {
            XSetICValues(pX11IMData->ic_active,
                         XNStatusAttributes, status,
                         NULL);
            XFree((void *)status);
        }
        AWT_UNLOCK();
    }
}
