/*
 * @(#)mlib_v_ImageConv3x3.c	1.4 98/09/06
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#pragma ident	"@(#)mlib_v_ImageConv3x3.c	1.14	98/09/02 SMI"

/*
 * FUNCTION
 *      mlib_ImageConv3x3 - image convolution with edge condition
 *
 * SYNOPSIS
 *      mlib_status mlib_ImageConv3x3(mlib_image *dst,
 *                                    mlib_image *src,
 *                                    mlib_s32   *kernel,
 *                                    mlib_s32   scale,
 *                                    mlib_s32   cmask,
 *                                    mlib_edge  edge)
 *
 * ARGUMENTS
 *      dst       Pointer to destination image.
 *      src       Pointer to source image.
 *      kernel    Pointer to convolution kernel.
 *      scale     The scaling factor to convert the input integer
 *                coefficients into floating-point coefficients:
 *                floating-point coefficient = integer coefficient * 2^(-scale)
 *      cmask     Channel mask to indicate the channels to be convolved.
 *                Each bit of which represents a channel in the image. The
 *                channels corresponded to 1 bits are those to be processed.
 *      edge      Type of edge condition.
 *
 * DESCRIPTION
 *      2-D convolution, 3x3 kernel.
 *
 *      The src and the dst must be the same type and have same number
 *      of channels (1, 2, 3, or 4). The unselected channels are not
 *      overwritten. If both src and dst have just one channel,
 *      cmask is ignored.
 *
 *      The edge condition can be one of the following:
 *              MLIB_EDGE_DST_NO_WRITE  (default)
 *              MLIB_EDGE_DST_FILL_ZERO
 *              MLIB_EDGE_DST_COPY_SRC
 *              MLIB_EDGE_SRC_EXTEND
 *
 * RESTRICTION
 *      The kernel coefficients used in function mlib_ImageConv3x3
 *      has to be preprocessed in a special way. 
 *      It is recommended to preprocess these coefficients 
 *      using the mlib_ImageConvKernelConvert function.
 */

/***********************************************************************/

#include <stdlib.h>
#include <mlib_image.h>

mlib_status mlib_v_conv3x3_32( mlib_image *src, mlib_image *dst,
             mlib_s32 *kernel, mlib_s32 scalef_expon,
                               mlib_s32 cmask, mlib_s32 edge);
mlib_status mlib_v_conv3x3_16nw( mlib_image *src, mlib_image *dst,
             mlib_s32 *kernel, mlib_s32 scalef_expon, mlib_s32 cmask);
mlib_status mlib_v_conv3x3_16ext( mlib_image *src, mlib_image *dst,
             mlib_s32 *kernel, mlib_s32 scalef_expon, mlib_s32 cmask);
mlib_status mlib_v_conv3x3_8nw( mlib_image *src, mlib_image *dst,
            mlib_s32 *kernel, mlib_s32 scalef_expon, mlib_s32 cmask);
mlib_status mlib_v_conv3x3_8ext( mlib_image *src, mlib_image *dst,
            mlib_s32 *kernel, mlib_s32 scalef_expon, mlib_s32 cmask);

mlib_status mlib_v_ImageConvCopyEdge(mlib_image *dst, mlib_image *src, mlib_s32 dx_l,
                                     mlib_s32 dx_r, mlib_s32 dy_t, mlib_s32 dy_b, mlib_s32 cmask);
mlib_status mlib_v_ImageConvClearEdge(mlib_image *img, mlib_s32 dx_l, mlib_s32 dx_r,
                                      mlib_s32 dy_t, mlib_s32 dy_b, mlib_s32 *color, mlib_s32 cmask);


/***********************************************************************/

#ifdef MLIB_TEST

mlib_status mlib_v_ImageConv3x3( mlib_image *dst,
                                 mlib_image *src,
                                 mlib_s32   *kernel,
                                 mlib_s32   scalef_exponent,
                                 mlib_s32   cmask,
                                 mlib_edge  edge)

#else
mlib_status mlib_ImageConv3x3( mlib_image *dst,
                               mlib_image *src,
                               mlib_s32   *kernel,
                               mlib_s32   scalef_exponent,
                               mlib_s32   cmask,
                               mlib_edge  edge)

#endif

{
   mlib_status stat;
   mlib_s32    slb, dlb,      /* line bytes (y stride) */
               src_wid, src_hgt,
               dst_wid, dst_hgt,
               src_nch, dst_nch;
   mlib_type   src_dtype, dst_dtype;
   mlib_s32    zero[4] = {0, 0, 0, 0};
   void        *sa, *da;

   if (dst == NULL || src == NULL) return MLIB_FAILURE;

   src_dtype = mlib_ImageGetType(src);
   dst_dtype = mlib_ImageGetType(dst);
   src_nch = mlib_ImageGetChannels(src);
   dst_nch = mlib_ImageGetChannels(dst);
   src_wid = mlib_ImageGetWidth(src);
   src_hgt = mlib_ImageGetHeight(src);
   dst_wid = mlib_ImageGetWidth(dst);
   dst_hgt = mlib_ImageGetHeight(dst);
   slb  = mlib_ImageGetStride(src);
   dlb  = mlib_ImageGetStride(dst);
   sa  = mlib_ImageGetData(src);
   da  = mlib_ImageGetData(dst);

   if ( src_dtype != dst_dtype || src_nch != dst_nch ||
        src_wid != dst_wid || src_hgt != dst_hgt || src_wid < 3 || src_hgt < 3 ||
        sa == NULL || da == NULL || kernel == NULL ) {
      stat =  MLIB_FAILURE;
   }
   else {
     if (src_nch == 1) cmask = 1;
      switch (src_dtype) {
      case MLIB_BYTE:
        if ( slb < (src_wid * src_nch)  ||  dlb < (dst_wid * dst_nch) ||
             scalef_exponent < 16 || scalef_exponent > 31)
          stat = MLIB_FAILURE;
        else {
        switch (edge) {
          case MLIB_EDGE_DST_FILL_ZERO:
            stat = mlib_v_conv3x3_8nw( dst, src, kernel, scalef_exponent, cmask);
            if (stat != MLIB_SUCCESS) break;
            stat = mlib_v_ImageConvClearEdge( dst, 1, 1, 1, 1, zero, cmask);
            break;
          case MLIB_EDGE_DST_COPY_SRC:
            stat = mlib_v_conv3x3_8nw( dst, src, kernel, scalef_exponent, cmask);
            if (stat != MLIB_SUCCESS) break;
            stat = mlib_v_ImageConvCopyEdge( dst, src, 1, 1, 1, 1, cmask);
            break;
          case MLIB_EDGE_SRC_EXTEND:
            stat = mlib_v_conv3x3_8ext( dst, src, kernel, scalef_exponent, cmask);
            break;
          case MLIB_EDGE_DST_NO_WRITE:
          default:
            stat = mlib_v_conv3x3_8nw( dst, src, kernel, scalef_exponent, cmask);
            break;
          }
        }
        break;
      case MLIB_SHORT:  /* convert wid from pixels to bytes */
        if ( slb < ((src_wid << 1) * src_nch)  ||  dlb < ((dst_wid << 1) * dst_nch) ||
             scalef_exponent < 17 || scalef_exponent > 32)
          stat = MLIB_FAILURE;
        else {
          switch (edge) {
          case MLIB_EDGE_DST_FILL_ZERO:
            stat = mlib_v_conv3x3_16nw( dst, src, kernel, scalef_exponent, cmask);
            if (stat != MLIB_SUCCESS) break;
            stat = mlib_v_ImageConvClearEdge( dst, 1, 1, 1, 1, zero, cmask);
            break;
          case MLIB_EDGE_DST_COPY_SRC:
            stat = mlib_v_conv3x3_16nw( dst, src, kernel, scalef_exponent, cmask);
            if (stat != MLIB_SUCCESS) break;
            stat = mlib_v_ImageConvCopyEdge( dst, src, 1, 1, 1, 1, cmask);
            break;
          case MLIB_EDGE_SRC_EXTEND:
            stat = mlib_v_conv3x3_16ext( dst, src, kernel, scalef_exponent, cmask);
            break;
          case MLIB_EDGE_DST_NO_WRITE:
          default:
            stat = mlib_v_conv3x3_16nw( dst, src, kernel, scalef_exponent, cmask);
            break;
          }
        }
        break;
      case MLIB_INT:  /* convert wid from pixels to bytes */
        if ( slb < ((src_wid << 2) * src_nch)  ||  dlb < ((dst_wid << 2) * dst_nch) ||
             scalef_exponent < 0)
          stat = MLIB_FAILURE;
        else {
          switch (edge) {
          case MLIB_EDGE_DST_FILL_ZERO:
            stat = mlib_v_conv3x3_32( src, dst, kernel, scalef_exponent, cmask, 0);
            if (stat != MLIB_SUCCESS) break;
            stat = mlib_v_ImageConvClearEdge( dst, 1, 1, 1, 1, zero, cmask);
            break;
          case MLIB_EDGE_DST_COPY_SRC:
            stat = mlib_v_conv3x3_32( src, dst, kernel, scalef_exponent, cmask, 0);
            if (stat != MLIB_SUCCESS) break;
            stat = mlib_v_ImageConvCopyEdge( dst, src, 1, 1, 1, 1, cmask);
            break;
          case MLIB_EDGE_SRC_EXTEND:
            stat = mlib_v_conv3x3_32( src, dst, kernel, scalef_exponent, cmask, 1);
            break;
          case MLIB_EDGE_DST_NO_WRITE:
          default:
            stat = mlib_v_conv3x3_32( src, dst, kernel, scalef_exponent, cmask, 0);
            break;
          }
        }
        break;
      default:
        stat = MLIB_FAILURE;
        break;
      }
   }
   return ( stat );
}
