/*
 * @(#)mlib_v_ImageConv3x3_16ext.c	1.5 98/09/06
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#pragma ident	"@(#)mlib_v_ImageConv3x3_16ext.c	1.7	98/09/02 SMI"

/*
 * FUNCTION
 *      mlib_conv3x3_16ext.c - convolve an 16-bit image with a 3x3 kernel
 *                       (edge = src extended)
 *
 *
 * SYNOPSIS
 *  mlib_status
 *  mlib_conv3x3_16ext( mlib_image *src,
 *                      mlib_image *dst,
 *                      mlib_s32 *kernel,
 *                      mlib_s32 scalef_expon
 *                      mlib_s32 cmask )
 * ARGUMENT
 *      src           Pointer to structure of source image
 *      dst           Pointer to structure of destination image
 *      kernel        Convolution kernel
 *      scalef_expons The scaling factor to convert the input integer
 *                    coefficients into floating-point coefficients:
 *                    floating-point coefficient = integer coefficient * 2^(-scale)
 *      cmask         Channel mask to indicate the channels to be convolved.
 *                    Each bit of which represents a channel in the image. The
 *                    channels corresponded to 1 bits are those to be processed.
 *
 * DESCRIPTION
 *      A 2-D convolution (3x3 kernel) for 16-bit images.
 */


#include <stdlib.h>
#include <mlib_image.h>
#include <vis_proto.h>

/************************************************************************/
#define GET_SRC_DST_PARAMETERS()                                         \
  adr_src = (mlib_u16*)mlib_ImageGetData(src);                           \
  adr_dst = (mlib_u16*)mlib_ImageGetData(dst);                           \
  slb = mlib_ImageGetStride(src) >> 1;                                   \
  dlb = mlib_ImageGetStride(dst) >> 1;                                   \
  dh = mlib_ImageGetHeight(dst);                                         \
  dw = mlib_ImageGetWidth(dst);
/************************************************************************/
#define LOAD_KERNEL_INTO_FLOAT()                                         \
  k1 = vis_to_float((kernel[0] & 0xFFFF0000) |                           \
                   ((kernel[0] >> 16) & 0xFFFF));                        \
  k2 = vis_to_float((kernel[1] & 0xFFFF0000) |                           \
                   ((kernel[1] >> 16) & 0xFFFF));                        \
  k3 = vis_to_float((kernel[2] & 0xFFFF0000) |                           \
                   ((kernel[2] >> 16) & 0xFFFF));                        \
  k4 = vis_to_float((kernel[3] & 0xFFFF0000) |                           \
                   ((kernel[3] >> 16) & 0xFFFF));                        \
  k5 = vis_to_float((kernel[4] & 0xFFFF0000) |                           \
                   ((kernel[4] >> 16) & 0xFFFF));                        \
  k6 = vis_to_float((kernel[5] & 0xFFFF0000) |                           \
                   ((kernel[5] >> 16) & 0xFFFF));                        \
  k7 = vis_to_float((kernel[6] & 0xFFFF0000) |                           \
                   ((kernel[6] >> 16) & 0xFFFF));                        \
  k8 = vis_to_float((kernel[7] & 0xFFFF0000) |                           \
                   ((kernel[7] >> 16) & 0xFFFF));                        \
  k9 = vis_to_float((kernel[8] & 0xFFFF0000) |                           \
                   ((kernel[8] >> 16) & 0xFFFF));
/************************************************************************/
#define PREPARE_INTERM_BUFFERS()                                         \
  buff_src = (mlib_d64*) mlib_malloc(4 * buf_slb * sizeof(mlib_d64));    \
  if (buff_src == NULL) return MLIB_FAILURE;                             \
                                                                         \
  sbuf1 = buff_src;                                                      \
  sbuf2 = sbuf1 + buf_slb;                                               \
  sbuf3 = sbuf2 + buf_slb;                                               \
  dbuf = sbuf3 + buf_slb;
/************************************************************************/
#define LOOP_INI()                                                       \
  ddst = (mlib_d64*)dbuf;                                                \
  da = d_a;                                                              \
  dend = da + (dw - 1);                                                  \
  prow = sbuf1;                                                          \
  sbuf1 = sbuf2; sbuf2 = sbuf3;                                          \
  sbuf3 = prow;                                                          \
  s1 = (mlib_d64*)sbuf1; s2 = (mlib_d64*)sbuf2;                          \
  s3 = (mlib_d64*)sbuf3;
/************************************************************************/
#define PREPARE_TO_COPY_INTERM_BUF_TO_DST()                              \
  ddst = dbuf;                                                           \
  /* prepare the destination addresses */                                \
  dp = (mlib_d64 *)((mlib_addr) da & (~7));                              \
  i = ((mlib_u16*) dp - (mlib_u16*) da);                                 \
  ddst = vis_alignaddr(ddst, 2*i);                                       \
  /* generate edge mask for the start point */                           \
  emask = vis_edge16(da, dend);                                          \
  sd1 = ddst[0];                                                         \
  if (emask != 0xf) {                                                    \
    sd0 = sd1;                                                           \
    sd1 = ddst[1];                                                       \
    sd0 = vis_faligndata(sd0, sd1);                                      \
    vis_pst_16(sd0, dp++, emask);                                        \
    ddst++;                                                              \
    i += 4;                                                              \
  }
/************************************************************************/
#define COPY_INTERM_BUF_TO_DST()                                         \
  for (; i <= (dw - 4); i += 4) {                                        \
    sd0 = sd1;                                                           \
    sd1 = ddst[1];                                                       \
    *dp++ = vis_faligndata(sd0, sd1);                                    \
    ddst++;                                                              \
  }
/************************************************************************/
#define COPY_TAIL()                                                      \
  if (i < dw)  {                                                         \
    sd0 = vis_faligndata(sd1, ddst[1]);                                  \
    emask = vis_edge16(dp, dend);                                        \
    vis_pst_16(sd0, dp, emask);                                          \
  }
/************************************************************************/
#define CONV_16_BEGIN(dsrc, ka)                                          \
  tmp0 = vis_fmuld8sux16(vis_read_hi(dsrc), ka);                         \
  tmp1 = vis_fmuld8ulx16(vis_read_hi(dsrc), ka);                         \
  tmp2 = vis_fmuld8sux16(vis_read_lo(dsrc), ka);                         \
  tmp3 = vis_fmuld8ulx16(vis_read_lo(dsrc), ka);                         \
  out0 = vis_fpadd32(tmp0, tmp1);                                        \
  out1 = vis_fpadd32(tmp2, tmp3);
/************************************************************************/
#define CONV_16(dsrc, ka)                                                \
  tmp0 = vis_fmuld8sux16(vis_read_hi(dsrc), ka);                         \
  tmp1 = vis_fmuld8ulx16(vis_read_hi(dsrc), ka);                         \
  tmp2 = vis_fmuld8sux16(vis_read_lo(dsrc), ka);                         \
  tmp3 = vis_fmuld8ulx16(vis_read_lo(dsrc), ka);                         \
  tmp0 = vis_fpadd32(tmp0, tmp1);                                        \
  out0 = vis_fpadd32(out0, tmp0);                                        \
  tmp2 = vis_fpadd32(tmp2, tmp3);                                        \
  out1 = vis_fpadd32(out1, tmp2);
/************************************************************************/
#define CONV_16_MIX(dsrc1, dsrc2, ka)                                    \
  tmp0 = vis_fmuld8sux16(vis_read_lo(dsrc1), ka);                        \
  tmp1 = vis_fmuld8ulx16(vis_read_lo(dsrc1), ka);                        \
  tmp2 = vis_fmuld8sux16(vis_read_hi(dsrc2), ka);                        \
  tmp3 = vis_fmuld8ulx16(vis_read_hi(dsrc2), ka);                        \
  tmp0 = vis_fpadd32(tmp0, tmp1);                                        \
  out0 = vis_fpadd32(out0, tmp0);                                        \
  tmp2 = vis_fpadd32(tmp2, tmp3);                                        \
  out1 = vis_fpadd32(out1, tmp2);
/************************************************************************/
#define CONV_16_LOAD_BEGIN(dsrc, ka, ld1, ld2)                           \
  tmp0 = vis_fmuld8sux16(vis_read_hi(dsrc), ka);                         \
  tmp1 = vis_fmuld8ulx16(vis_read_hi(dsrc), ka);                         \
  tmp2 = vis_fmuld8sux16(vis_read_lo(dsrc), ka);                         \
  tmp3 = vis_fmuld8ulx16(vis_read_lo(dsrc), ka);                         \
  ld1 = *sa_4;                                                           \
  sa_4 += nchannel;                                                      \
  out0 = vis_fpadd32(tmp0, tmp1);                                        \
  out1 = vis_fpadd32(tmp2, tmp3);                                        \
  ld2 = *d_buf++;
/************************************************************************/
#define CONV_16_LOAD(dsrc, ka, ld1, ld2)                                 \
  tmp0 = vis_fmuld8sux16(vis_read_hi(dsrc), ka);                         \
  tmp1 = vis_fmuld8ulx16(vis_read_hi(dsrc), ka);                         \
  tmp2 = vis_fmuld8sux16(vis_read_lo(dsrc), ka);                         \
  tmp3 = vis_fmuld8ulx16(vis_read_lo(dsrc), ka);                         \
  ld1 = *sa_4;                                                           \
  sa_4 += nchannel;                                                      \
  tmp0 = vis_fpadd32(tmp0, tmp1);                                        \
  out0 = vis_fpadd32(out0, tmp0);                                        \
  tmp2 = vis_fpadd32(tmp2, tmp3);                                        \
  out1 = vis_fpadd32(out1, tmp2);                                        \
  ld2 = *d_buf++;
/************************************************************************/
#define CONV_16_STORE(dsrc1, dsrc2, ka, st1, st2)                        \
  tmp0 = vis_fmuld8sux16(vis_read_lo(dsrc1), ka);                        \
  tmp1 = vis_fmuld8ulx16(vis_read_lo(dsrc1), ka);                        \
  tmp2 = vis_fmuld8sux16(vis_read_hi(dsrc2), ka);                        \
  tmp3 = vis_fmuld8ulx16(vis_read_hi(dsrc2), ka);                        \
  *s_buf1++ = st1;                                                       \
  tmp0 = vis_fpadd32(tmp0, tmp1);                                        \
  out0 = vis_fpadd32(out0, tmp0);                                        \
  tmp2 = vis_fpadd32(tmp2, tmp3);                                        \
  out1 = vis_fpadd32(out1, tmp2);                                        \
  *da = st2;                                                             \
  da +=  nchannel;

/************************************************************************/

mlib_status mlib_v_conv3x3_16ext_2( mlib_image *dst,         /*source  image */
                                    mlib_image *src,         /* destination image */
                                    mlib_s32   *kernel,      /* convolution kernel */
                                    mlib_s32   scalef_expon) /* scalefactor exponent */
{
  mlib_u16  *da, *d_a;                                          /* pointers to dst row */
  mlib_u16  *adr_dst,  *adr_src, *dend;                         /* pointers to src, dst data */
  mlib_u16  *sa, *sa2;                                          /* pointers to src rows */
  mlib_d64  *buff_src, *sbuf1, *sbuf2, *prow, *prow1;           /* pointers to rows in interm. src buf */
  mlib_d64  *sbuf3;                                             /* pointers to rows in interm. src buf */
  mlib_d64  *dbuf;                                              /* pointer to row in interm. dst buf */
  mlib_d64  *s1, *s2, *s3;                                      /* mlib_d64 pointers to rows in interm. src buf */
  mlib_d64  *ddst;                                              /* mlib_d64 pointer to row in interm. dst buf */
  mlib_d64  d1, d2, d_1, d_2;                                   /* data */
  mlib_d64  d3, d_3;                                            /* data */
  mlib_f32  k1, k2, k3, k4, k5, k6, k7, k8, k9;
  mlib_s32  dlb, slb,  buf_slb;                                 /* src, dst and interm. buf. strides */
  mlib_s32  dh, dw;
  mlib_d64  out0, out1, tmp0, tmp1, tmp2, tmp3;
  mlib_d64  *dsa, *dp;
  mlib_d64  sd0, sd1;
  mlib_s32  emask;
  int       gsr_scale, i, j;

  gsr_scale = 32 - scalef_expon;
  vis_write_gsr((gsr_scale << 3));

  GET_SRC_DST_PARAMETERS()
  LOAD_KERNEL_INTO_FLOAT()

  buf_slb = (4*dw+32)>>3;
  PREPARE_INTERM_BUFFERS()

  dw <<= 1;

  sa  = adr_src;
  sa2 = sa + slb;
  d_a = adr_dst;

  /* load interm. src buff */

  prow = sbuf2;
  prow1 = sbuf3;

  for (i = 0; i < 2; i++) {
    *((mlib_u16*)((mlib_u16*)prow + 2 + i)) =
    *((mlib_u16*)((mlib_u16*)prow1 + 2 + i)) =
    *((mlib_u16*)((mlib_u16*)sa + i));
  }

  prow++; prow1++;
  dsa = (mlib_d64 *) vis_alignaddr(sa, 4);
  sd1 = dsa[0]; dsa++;

#pragma pipeloop(0)
  for (i = 0; i < (dw - 2); i+= 4) {
    sd0 = sd1; sd1 = dsa[0];
    *prow++ = *prow1++ = vis_faligndata(sd0, sd1);
    dsa++;
  }
  *(mlib_f32*)sbuf2 = *(mlib_f32*)sbuf3 =
  *(mlib_f32*)((mlib_f32*)sbuf2 + 1);
  *(mlib_f32*)((mlib_u16*)sbuf2 + dw + 2) =
  *(mlib_f32*)((mlib_u16*)sbuf3 + dw + 2) =
  *(mlib_f32*)((mlib_u16*)sbuf2 + dw);

#pragma pipeloop(0)
  for (j = 0; j < dh; j++)  {
    LOOP_INI()

    prow = sbuf3;

    for (i = 0; i < 2; i++) {
      *((mlib_u16*)((mlib_u16*)prow + 2 + i)) =
      *((mlib_u16*)((mlib_u16*)sa2 + i));
    }

    prow++;
    dsa = (mlib_d64 *) vis_alignaddr(sa2, 4);
    sd1 = dsa[0]; dsa++;

#pragma pipeloop(0)
    for (i = 0; i < (dw - 2); i+= 4) {
      sd0 = sd1; sd1 = dsa[0];
      *prow++ = vis_faligndata(sd0, sd1);
      dsa++;
    }
    *(mlib_f32*)sbuf3 = *(mlib_f32*)((mlib_f32*)sbuf3 + 1);
    *(mlib_f32*)((mlib_u16*)sbuf3 + dw + 2) =
    *(mlib_f32*)((mlib_u16*)sbuf3 + dw);

    d1 = *s1; d2 = *s2; d3 = *s3;

#pragma pipeloop(0)
    for (i = 0; i < dw; i += 4) {
      d_1 = *(s1+1); d_2 = *(s2+1); d_3 = *(s3+1);
      CONV_16_BEGIN(d1, k1)
      CONV_16(d2, k4)
      CONV_16(d3, k7)
      CONV_16_MIX(d1, d_1, k2)
      CONV_16_MIX(d2, d_2, k5)
      CONV_16_MIX(d3, d_3, k8)
      CONV_16(d_1, k3)
      CONV_16(d_2, k6)
      CONV_16(d_3, k9)

      *ddst++ = vis_fpackfix_pair(out0, out1);

      d1 = d_1; d2 = d_2; d3 = d_3;
      s1++; s2++; s3++;
    }
    PREPARE_TO_COPY_INTERM_BUF_TO_DST()

#pragma pipeloop(0)
    COPY_INTERM_BUF_TO_DST()
    COPY_TAIL()

    if (j != dh - 2) sa2 = sa2 + slb;
    d_a += dlb;
  }
  mlib_free(buff_src);
  return( MLIB_SUCCESS );
}

/************************************************************************/

mlib_status mlib_v_conv3x3_16ext_3( mlib_image *dst,         /* source  image */
                                    mlib_image *src,         /* destination image */
                                    mlib_s32   *kernel,      /* convolution kernel */
                                    mlib_s32   scalef_expon) /* scalefactor exponent */
{
  mlib_u16  *da, *d_a;                                          /* pointers to dst row */
  mlib_u16  *adr_dst,  *adr_src, *dend;                         /* pointers to src, dst data */
  mlib_u16  *sa, *sa2;                                          /* pointers to src rows */
  mlib_d64  *buff_src, *sbuf1, *sbuf2, *prow, *prow1;           /* pointers to rows in interm. src buf */
  mlib_d64  *sbuf3;                                             /* pointers to rows in interm. src buf */
  mlib_d64  *dbuf;                                              /* pointer to row in interm. dst buf */
  mlib_d64  *s1, *s2, *s3;                                      /* mlib_d64 pointers to rows in interm. src buf */
  mlib_d64  *ddst;                                              /* mlib_d64 pointer to row in interm. dst buf */
  mlib_d64  d1, d2, d11, d12, d21, d22, d31, d32;               /* data */
  mlib_d64  d3, d13, d23, d33;                                  /* data */
  mlib_f32  k1, k2, k3, k4, k5, k6, k7, k8, k9;
  mlib_s32  dlb, slb,  buf_slb;                                 /* src, dst and interm. buf. strides */
  mlib_s32  dh, dw;
  mlib_d64  out0, out1, tmp0, tmp1, tmp2, tmp3;
  mlib_d64  *dsa, *dp;
  mlib_d64  sd0, sd1;
  mlib_s32  emask;
  int       gsr_scale, i, j;

  gsr_scale = 32 - scalef_expon;
  vis_write_gsr((gsr_scale << 3));

  GET_SRC_DST_PARAMETERS()
  LOAD_KERNEL_INTO_FLOAT()

  buf_slb = (6*dw+40)>>3;
  PREPARE_INTERM_BUFFERS()

  dw *= 3;

  sa  = adr_src;
  sa2 = sa + slb;
  d_a = adr_dst;

  /* load interm. src buff */

  prow = sbuf2;
  prow1 = sbuf3;
  for (i = 1; i < 4; i++) {
    for (j = 0; j < 3; j++) {
      *((mlib_u16*)((mlib_u16*)prow + 3*i + j)) =
      *((mlib_u16*)((mlib_u16*)prow1 + 3*i + j)) =
      *((mlib_u16*)((mlib_u16*)sa + 3*i - 3 + j));
    }
  }
  prow+=3; prow1+=3;
  dsa = (mlib_d64 *) vis_alignaddr(sa, 18);
  sd1 = dsa[0]; dsa++;

#pragma pipeloop(0)
  for (i = 0; i < (dw - 9); i+= 4) {
    sd0 = sd1; sd1 = dsa[0];
    *prow++ = *prow1++ = vis_faligndata(sd0, sd1);
    dsa++;
  }

  for (i = 0 ; i < 3; i++) {
    *(mlib_u16*)((mlib_u16*)sbuf2 + i) =
    *(mlib_u16*)((mlib_u16*)sbuf3 + i) =
    *(mlib_u16*)((mlib_u16*)sbuf2 + 3 + i);
    *(mlib_u16*)((mlib_u16*)sbuf2 + dw + 3 + i) =
    *(mlib_u16*)((mlib_u16*)sbuf3 + dw + 3 + i) =
    *(mlib_u16*)((mlib_u16*)sbuf2 + dw + i);
  }

#pragma pipeloop(0)
  for (j = 0; j < dh; j++)  {
    LOOP_INI()

    prow = sbuf3;
    for (i = 1; i < 4; i++) {
      *((mlib_u16*)((mlib_u16*)prow + 3*i)) =
      *((mlib_u16*)((mlib_u16*)sa2 + 3*i - 3));
      *((mlib_u16*)((mlib_u16*)prow + 3*i + 1)) =
      *((mlib_u16*)((mlib_u16*)sa2 + 3*i - 2));
      *((mlib_u16*)((mlib_u16*)prow + 3*i + 2)) =
      *((mlib_u16*)((mlib_u16*)sa2 + 3*i - 1));
    }
    prow+=3;
    dsa = (mlib_d64 *) vis_alignaddr(sa2, 18);
    sd1 = dsa[0]; dsa++;

#pragma pipeloop(0)
    for (i = 0; i < (dw - 9); i+= 4) {
      sd0 = sd1; sd1 = dsa[0];
      *prow++ = vis_faligndata(sd0, sd1);
      dsa++;
    }

    for (i = 0 ; i < 3; i++) {
      *(mlib_u16*)((mlib_u16*)sbuf3 + i) =
      *(mlib_u16*)((mlib_u16*)sbuf3 + 3 + i);
      *(mlib_u16*)((mlib_u16*)sbuf3 + dw + 3 + i) =
      *(mlib_u16*)((mlib_u16*)sbuf3 + dw + i);
    }

    vis_alignaddr(s1, 6);
    d1 = *s1; d2 = *s2; d3 = *s3;
    d11 = *(s1+1); d12 = *(s2+1); d13 = *(s3+1);

#pragma pipeloop(0)
    for (i = 0; i < dw; i += 4) {
      d21 = *(s1+2); d22 = *(s2+2); d23 = *(s3+2);
      CONV_16_BEGIN(d1, k1)
      CONV_16(d2, k4)
      d31 = vis_faligndata(d1, d11);
      CONV_16(d3, k7)
      CONV_16_MIX(d11, d21, k3)
      d32 = vis_faligndata(d2, d12);
      CONV_16_MIX(d12, d22, k6)
      d33 = vis_faligndata(d3, d13);
      CONV_16_MIX(d13, d23, k9)

      CONV_16(d31, k2)
      CONV_16(d32, k5)
      CONV_16(d33, k8)

      *ddst++ = vis_fpackfix_pair(out0, out1);

      d1 = d11; d2 = d12; d3 = d13;
      d11 = d21; d12 = d22; d13 = d23;
      s1++; s2++; s3++;
    }
    PREPARE_TO_COPY_INTERM_BUF_TO_DST()

#pragma pipeloop(0)
    COPY_INTERM_BUF_TO_DST()
    COPY_TAIL()

    if (j != dh - 2) sa2 = sa2 + slb;
    d_a += dlb;
  }
  mlib_free(buff_src);
  return( MLIB_SUCCESS );
}

/************************************************************************/

mlib_status mlib_v_conv3x3_16ext_4( mlib_image *dst,         /* source  image */
                                    mlib_image *src,         /* destination image */
                                    mlib_s32   *kernel,      /* convolution kernel */
                                    mlib_s32   scalef_expon) /* scalefactor exponent */
{
  mlib_u16  *da, *d_a;                                          /* pointers to dst row */
  mlib_u16  *adr_dst,  *adr_src, *dend;                         /* pointers to src, dst data */
  mlib_u16  *sa, *sa2;                                          /* pointers to src rows */
  mlib_d64  *buff_src, *sbuf1, *sbuf2, *prow, *prow1;           /* pointers to rows in interm. src buf */
  mlib_d64  *sbuf3;                                             /* pointers to rows in interm. src buf */
  mlib_d64  *dbuf;                                              /* pointer to row in interm. dst buf */
  mlib_d64  *s1, *s2, *s3;                                      /* mlib_d64 pointers to rows in interm. src buf */
  mlib_d64  *ddst;                                              /* mlib_d64 pointer to row in interm. dst buf */
  mlib_d64  d1, d2, d11, d12, d21, d22;                         /* data */
  mlib_d64  d3, d13, d23;                                       /* data */
  mlib_f32  k1, k2, k3, k4, k5, k6, k7, k8, k9;
  mlib_s32  dlb, slb,  buf_slb;                                 /* src, dst and interm. buf. strides */
  mlib_s32  dh, dw;
  mlib_d64  out0, out1, tmp0, tmp1, tmp2, tmp3;
  mlib_d64  *dsa, *dp;
  mlib_d64  sd0, sd1;
  mlib_s32  emask;
  int       gsr_scale, i, j;

  gsr_scale = 32 - scalef_expon;
  vis_write_gsr((gsr_scale << 3));

  GET_SRC_DST_PARAMETERS()
  LOAD_KERNEL_INTO_FLOAT()

  buf_slb = (8*dw+40)>>3;
  PREPARE_INTERM_BUFFERS()

  dw *= 4;

  sa  = adr_src;
  sa2 = sa + slb;
  d_a = adr_dst;

  /* load interm. src buff */

  prow = sbuf2 + 1;
  prow1 = sbuf3 + 1;

  dsa = (mlib_d64 *) vis_alignaddr(sa, 0);
  sd1 = dsa[0]; dsa++;

#pragma pipeloop(0)
  for (i = 0; i < dw; i+= 4) {
    sd0 = sd1; sd1 = dsa[0];
    *prow++ = *prow1++ = vis_faligndata(sd0, sd1);
    dsa++;
  }
  sbuf2[0] = sbuf3[0] = sbuf2[1];
  sbuf2[(dw >> 2) + 1] =
  sbuf3[(dw >> 2) + 1] =
  sbuf2[(dw >> 2)];

#pragma pipeloop(0)
  for (j = 0; j < dh; j++)  {
    LOOP_INI()

    prow = sbuf3 + 1;

    dsa = (mlib_d64 *) vis_alignaddr(sa2, 0);
    sd1 = dsa[0]; dsa++;

#pragma pipeloop(0)
    for (i = 0; i < dw; i+= 4) {
      sd0 = sd1; sd1 = dsa[0];
      *prow++ = vis_faligndata(sd0, sd1);
      dsa++;
    }
    sbuf3[0] = sbuf3[1];
    sbuf3[(dw >> 2) + 1] =  sbuf3[(dw >> 2)];

    d1 = *s1; d2 = *s2; d3 = *s3;
    d11 = *(s1+1); d12 = *(s2+1); d13 = *(s3+1);

#pragma pipeloop(0)
    for (i = 0; i < dw; i += 4) {
      d21 = *(s1+2); d22 = *(s2+2); d23 = *(s3+2);
      CONV_16_BEGIN(d1, k1)
      CONV_16(d2, k4)
      CONV_16(d3, k7)
      CONV_16(d11, k2)
      CONV_16(d12, k5)
      CONV_16(d13, k8)
      CONV_16(d21, k3)
      CONV_16(d22, k6)
      CONV_16(d23, k9)

      *ddst++ = vis_fpackfix_pair(out0, out1);

      d1 = d11; d2 = d12; d3 = d13;
      d11 = d21; d12 = d22; d13 = d23;
      s1++; s2++; s3++;
    }
    PREPARE_TO_COPY_INTERM_BUF_TO_DST()

#pragma pipeloop(0)
    COPY_INTERM_BUF_TO_DST()
    COPY_TAIL()

    if (j != dh - 2) sa2 = sa2 + slb;
    d_a += dlb;
  }
  mlib_free(buff_src);
  return( MLIB_SUCCESS );
}

/************************************************************************/

mlib_status mlib_v_conv3x3_16ext_mask( mlib_image *dst,         /* source  image */
                                       mlib_image *src,         /* destination image */
                                       mlib_s32   *kernel,      /* convolution kernel */
                                       mlib_s32   scalef_expon, /* scalefactor exponent */
                                       mlib_s32   cmask)        /* mask */
{
    mlib_u16  *da, *d_a;                                       /* pointers to dst row */
    mlib_u16  *adr_dst, *dend, *adr_src;                       /* pointers to src, dst data */
    mlib_u16  *sa, *sa2, *sa4, *sa_4, *prow;             /* pointers to src rows */
    mlib_u16  *buff_src, *sbuf1, *sbuf2, *sbuf3, *s_buf1;      /* pointers to rows in interm. src buf */
    mlib_u16  *dbuf, *d_buf;                                   /* pointers to row in interm. dst buf */
    mlib_d64  *s1, *s2, *s3;                                   /* mlib_d64 pointers to rows in interm. src buf */
    mlib_d64  *ddst;                                           /* mlib_d64 pointer to row in interm. dst buf */
    mlib_s32  dlb, slb, buf_slb;                               /* src, dst and interm. buf. strides */
    mlib_s32  dh, dw;
    mlib_f32  k1, k2, k3, k4, k5, k6, k7, k8, k9;              /* src, dst and interm. buf. strides */
    mlib_d64  out0, out1, tmp0, tmp1, tmp2, tmp3;
    mlib_d64  d1, d2, d3, d_1, d_2, d_3;                       /* data */
    mlib_d64  d21, d22, d23;                                   /* shifted data */
    int       gsr_scale, i, j, nchannel, chan, testchan;
    mlib_u16  t1, t2, t3, t4, tt1, tt2, tt3, tt4;              /* temp, used in load-store */

    adr_src = (mlib_u16*)mlib_ImageGetData(src);
    adr_dst = (mlib_u16*)mlib_ImageGetData(dst);
    nchannel = mlib_ImageGetChannels(src);
    slb = mlib_ImageGetStride(src) >> 1;
    dlb = mlib_ImageGetStride(dst) >> 1;
    dh = mlib_ImageGetHeight(dst);
    dw = mlib_ImageGetWidth(dst);

    buf_slb = (((dw+2)<<1)+24) & (~7);                                          /* buf_slb - 8_byte alligned */
    buff_src = (mlib_u16*) mlib_malloc(4 * buf_slb * sizeof(mlib_u8) + 8);  /* alloc. interm. src and dst buffer */
    if (buff_src == NULL) return MLIB_FAILURE;
                                                               /* edge - no write */

    buf_slb >>= 1;

    gsr_scale = 32 - scalef_expon;
    vis_write_gsr((gsr_scale << 3)+2);
    sbuf1 = (mlib_u16*) ((mlib_addr)(buff_src +8) & (~7));
    sbuf2 = sbuf1 + buf_slb;
    sbuf3 = sbuf2 + buf_slb;
    dbuf = sbuf3 +buf_slb;

    LOAD_KERNEL_INTO_FLOAT()

    testchan = 1;

    for ( chan = nchannel - 1; chan >= 0; chan--) {
      if ((cmask & testchan) == 0) {
        testchan <<= 1;
        continue;
      }
      testchan <<= 1;
      sa  = adr_src + chan;
      sa2 = sa + slb;
      sa_4 = sa4 = sa2 + slb;
      d_a = adr_dst + chan;

      /* load interm. src buff */
       for (i = 0, j = 1; j < (dw + 1); i+=nchannel, j++) {
        sbuf1[j] = sa2[i];  sbuf2[j] = sbuf3[j] = sa[i];
      }
      sbuf2[0] = sbuf2[1]; sbuf3[0] = sbuf3[1];
      sbuf2[dw+1] = sbuf2[dw]; sbuf3[dw+1] = sbuf3[dw];

      for (j = 0; j < dh - 1; j++)  {
        sbuf1[0] = sbuf1[1]; sbuf1[dw+1] = sbuf1[dw];
        ddst = (mlib_d64*)(dbuf);
        d_buf = dbuf;
        da = d_a;
        dend = da + (dw - 1)*nchannel;
        prow = sbuf1;
        sbuf1 = sbuf2; sbuf2 = sbuf3; sbuf3 = prow;
        s1 = (mlib_d64*)sbuf1; s2 = (mlib_d64*)sbuf2; s3 = (mlib_d64*)sbuf3;
        s_buf1 = sbuf1+1;
        d1 = *s1; d2 = *s2; d3 = *s3;

        /* first iteraction - no write data to dst (only in interm dst buff) */
        d_1 = *(s1+1); d_2 = *(s2+1); d_3 = *(s3+1);
        CONV_16_BEGIN(d1, k1)
        t1 = *sa_4; sa_4 += nchannel;
        CONV_16(d2, k4)
        t2 = *sa_4; sa_4 += nchannel;
        CONV_16(d3, k7)
        t3 = *sa_4; sa_4 += nchannel;
        d22 = vis_faligndata(d2, d_2);
        t4 = *sa_4; sa_4 += nchannel;
        d21 = vis_faligndata(d1, d_1);
        d23 = vis_faligndata(d3, d_3);
        CONV_16(d21, k2)
        CONV_16(d22, k5)
        *s_buf1++ = t1;
        CONV_16(d23, k8)
        *s_buf1++ = t2;
        CONV_16_MIX(d1, d_1, k3)
        *s_buf1++ = t3;
        CONV_16_MIX(d2, d_2, k6)
        *s_buf1++ = t4;
        CONV_16_MIX(d3, d_3, k9)

        *ddst++ = vis_freg_pair(vis_fpackfix(out0), vis_fpackfix(out1));
        d1 = d_1; d2 = d_2; d3 = d_3;
        s1++; s2++; s3++;

#pragma pipeloop(0)
        for (i = 4; i < dw; i += 4) {
         /* in each iteraction store result from prev. iteract.
           and load data for processing next row */
          d_1 = *(s1+1); d_2 = *(s2+1); d_3 = *(s3+1);
          CONV_16_LOAD_BEGIN(d1, k1, t1, tt1)
          CONV_16_LOAD(d2, k4, t2, tt2)
          CONV_16_LOAD(d3, k7, t3, tt3)
          d22 = vis_faligndata(d2, d_2);
          t4 = *sa_4; sa_4 += nchannel;
          d21 = vis_faligndata(d1, d_1);
          tt4 = *d_buf++;
          d23 = vis_faligndata(d3, d_3);
          CONV_16(d21, k2)
          *s_buf1++ = t1;
          CONV_16(d22, k5)
          *da = tt1; da += nchannel;
          CONV_16(d23, k8)
          CONV_16_STORE(d1, d_1, k3, t2, tt2)
          CONV_16_STORE(d2, d_2, k6, t3, tt3)
          CONV_16_STORE(d3, d_3, k9, t4, tt4)

          *ddst++ = vis_freg_pair(vis_fpackfix(out0), vis_fpackfix(out1));
          d1 = d_1; d2 = d_2; d3 = d_3;
          s1++; s2++; s3++;
        }

        *s_buf1++ = *sa_4; sa_4 += nchannel; *s_buf1++ = *sa_4; sa_4 += nchannel;
        *s_buf1++ = *sa_4; sa_4 += nchannel; *s_buf1++ = *sa_4;
        if ((mlib_addr) da <= (mlib_addr) dend) {
          *da = *d_buf++; da += nchannel;
        }
        if ((mlib_addr) da <= (mlib_addr) dend) {
          *da = *d_buf++; da += nchannel;
        }
        if ((mlib_addr) da <= (mlib_addr) dend) {
          *da = *d_buf++; da += nchannel;
        }
        if ((mlib_addr) da <= (mlib_addr) dend) {
          *da = *d_buf++;
        }
        if (j != dh - 3) {
          sa_4 = sa4 = sa4 + slb;
        } else sa_4 = sa4;
        d_a += dlb;
      }

      /* process last row - no need to load data */
      sbuf1[0] = sbuf1[1]; sbuf1[dw+1] = sbuf1[dw];
      ddst = (mlib_d64*)(dbuf);
      da = d_a;
      dend = da + (dw - 1)*nchannel;
      prow = sbuf1;
      sbuf1 = sbuf2; sbuf2 = sbuf3; sbuf3 = prow;
      s1 = (mlib_d64*)sbuf1; s2 = (mlib_d64*)sbuf2; s3 = (mlib_d64*)sbuf3;
      d_buf = dbuf;
      d1 = *s1; d2 = *s2; d3 = *s3 ;

      d_1 = *(s1+1); d_2 = *(s2+1); d_3 = *(s3+1);
      CONV_16_BEGIN(d1, k1)
      CONV_16(d2, k4)
      CONV_16(d3, k7)
      d22 = vis_faligndata(d2, d_2);
      d21 = vis_faligndata(d1, d_1);
      d23 = vis_faligndata(d3, d_3);
      CONV_16(d21, k2)
      CONV_16(d22, k5)
      CONV_16(d23, k8)
      CONV_16_MIX(d1, d_1, k3)
      CONV_16_MIX(d2, d_2, k6)
      CONV_16_MIX(d3, d_3, k9)

      *ddst++ = vis_freg_pair(vis_fpackfix(out0), vis_fpackfix(out1));
      d1 = d_1; d2 = d_2; d3 = d_3;
      s1++; s2++; s3++;

#pragma pipeloop(0)
      for (i = 4; i < dw; i += 4) {
        d_1 = *(s1+1); d_2 = *(s2+1); d_3 = *(s3+1);
        CONV_16_BEGIN(d1, k1)
        tt1 = *d_buf++;
        CONV_16(d2, k4)
        tt2 = *d_buf++;
        CONV_16(d3, k7)
        tt3 = *d_buf++;
        d22 = vis_faligndata(d2, d_2);
        tt4 = *d_buf++;
        d21 = vis_faligndata(d1, d_1);
        d23 = vis_faligndata(d3, d_3);
        CONV_16(d21, k2)
        CONV_16(d22, k5)
        *da = tt1; da += nchannel;
        CONV_16(d23, k8)
        *da = tt2; da += nchannel;
        CONV_16_MIX(d1, d_1, k3)
        *da = tt3; da += nchannel;
        CONV_16_MIX(d2, d_2, k6)
        *da = tt4; da += nchannel;
        CONV_16_MIX(d3, d_3, k9)

        *ddst++ = vis_freg_pair(vis_fpackfix(out0), vis_fpackfix(out1));
        d1 = d_1; d2 = d_2; d3 = d_3;
        s1++; s2++; s3++;
      }
      if ((mlib_addr) da <= (mlib_addr) dend) {
        *da = *d_buf++; da += nchannel;
      }
      if ((mlib_addr) da <= (mlib_addr) dend) {
        *da = *d_buf++; da += nchannel;
      }
      if ((mlib_addr) da <= (mlib_addr) dend) {
        *da = *d_buf++; da += nchannel;
      }
      if ((mlib_addr) da <= (mlib_addr) dend) {
        *da = *d_buf++;
      }
    }
    mlib_free(buff_src);
    return MLIB_SUCCESS ;
}

/************************************************************************/

mlib_status mlib_v_conv3x3_16ext( mlib_image *dst,         /* source  image */
                                  mlib_image *src,         /* destination image */
                                  mlib_s32 *kernel,        /* convolution kernel */
                                  mlib_s32   scalef_expon, /* scalefactor exponent */
                                  mlib_s32   cmask)        /* mask */

{
  mlib_s32 nchannel, chan, cmask1, i;
  mlib_status res;

  cmask1 = cmask;
  nchannel = mlib_ImageGetChannels(src);
  chan = 0;
  for(i = nchannel - 1; i >= 0; i--, cmask1 >>= 1)
    if ((cmask1 & 1) == 1) chan++;

  if ((chan == nchannel) && (nchannel != 1)) {
    switch (nchannel) {
      case 2:
        res = mlib_v_conv3x3_16ext_2(dst, src, kernel, scalef_expon);
        break;
      case 3:
        res = mlib_v_conv3x3_16ext_3(dst, src, kernel, scalef_expon);
        break;
      case 4:
        res = mlib_v_conv3x3_16ext_4(dst, src, kernel, scalef_expon);
        break;
      default: res = MLIB_FAILURE; break;
    }
  } else res = mlib_v_conv3x3_16ext_mask(dst, src, kernel, scalef_expon, cmask);
  return ( res );
}





