/*
 * @(#)mlib_v_ImageConv3x3_32.c	1.1 98/09/07
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#pragma ident	"@(#)mlib_v_ImageConv3x3_32.c	1.1	98/09/02 SMI"

/*
 * FUNCTION
 *      mlib_v_conv3x3_32.c - convolve an 32-bit image with a 3x3 kernel
 *
 *
 * SYNOPSIS
 *  mlib_status
 *  mlib_v_conv3x3_32( mlib_image *src,
 *                     mlib_image *dst,
 *                     mlib_s32 *kernel,
 *                     mlib_s32 scalef_expon
 *                     mlib_s32 cmask,
 *                     mlib_s32 edge )
 * ARGUMENT
 *      src           Pointer to structure of source image
 *      dst           Pointer to structure of destination image
 *      kernel        Convolution kernel
 *      scalef_expons The scaling factor to convert the input integer
 *                    coefficients into floating-point coefficients:
 *                    floating-point coefficient = integer coefficient * 2^(-scale)
 *      cmask         Channel mask to indicate the channels to be convolved.
 *                    Each bit of which represents a channel in the image. The
 *                    channels corresponded to 1 bits are those to be processed.
 *      edge          Edge == 0 - dst no write, else src extended.
 *
 * DESCRIPTION
 *      A 2-D convolution (3x3 kernel) for 32-bit images.
 */

#include <mlib_image.h>

/************************************************************************/

mlib_status mlib_v_conv3x3_32 ( mlib_image *src,        /* source image */
                                mlib_image *dst,        /* destination image */
                                mlib_s32 *kern,         /* convolution kernel */
                                mlib_s32 scalef_expon,  /* scalefactor=2**N */
                                mlib_s32 cmask,
                                mlib_s32 edge)
{
  mlib_s32 *sl, *adr_src;                       /* pointers for pixel and line of source */
  mlib_s32 *da, *dl, *adr_dst;                  /* pointers for pixel and line of destination */
  mlib_s32 row, col, block;                     /* indices */
  mlib_s32 *row0, *row1, *row2;
  double   scalef;                              /* divide coeffs by scalefactor=2**scalef_expon */
  double   sum1, sum2, sum3;                    /* accumulators */
  double   k0, k1, k2, k3, k4, k5, k6, k7, k8;  /* coeff's */
  double   p00, p01, p02, p03, p04,             /* row 0 pixel values */
           p10, p11, p12, p13, p14,             /* row 1 pixel values */
           p20, p21, p22, p23, p24;             /* row 2 pixel values */
  int      sll,                                 /* line length (in pixels) of source image */
           dll,                                 /* line length (in pixels) of destination image */
           wid,                                 /* src width in pixels */
           hgt;                                 /* src height in pixels */
  int      num_blocks;                          /* do output pts 3 at a time; loop count for this */
  int      leftover_cols;                       /* what can't be done in blocks */
  int      chan, testchan, nchannel;

  scalef = 1.0;
  while (scalef_expon > 30) {
    scalef /= (1 << 30);
    scalef_expon -= 30;
  }

  scalef /= (1 << scalef_expon);

  /* keep kernel in regs */
  k0 = scalef * kern[0];  k1 = scalef * kern[1];  k2 = scalef * kern[2];
  k3 = scalef * kern[3];  k4 = scalef * kern[4];  k5 = scalef * kern[5];
  k6 = scalef * kern[6];  k7 = scalef * kern[7];  k8 = scalef * kern[8];

  hgt = mlib_ImageGetHeight(src);
  wid = mlib_ImageGetWidth(src);
  nchannel = mlib_ImageGetChannels(src);
  sll = mlib_ImageGetStride(src) >> 2;           /* cnvrt from bytes to 32-bit pxls */
  dll = mlib_ImageGetStride(dst) >> 2;           /* cnvrt from bytes to 32-bit pxls */
  adr_src = (mlib_s32 *)mlib_ImageGetData(src);
  adr_dst = (mlib_s32 *)mlib_ImageGetData(dst);

  if (edge != 0)  wid += 1;

  num_blocks = (wid - 2)/3;                      /* (wid-2): no output on 1st, last columns */
  leftover_cols = (wid - 2) - (num_blocks * 3);
  if (edge == 0) hgt -= 2;                       /* boundaries: no output for first & last rows */

  testchan = 1;
  for ( chan = nchannel-1; chan >= 0; chan--) {
    if ((cmask & testchan) == 0) {
      testchan <<= 1;
      continue;
    }
    testchan <<= 1;
    sl = adr_src + chan;
    dl = da = adr_dst + chan;
    if (edge == 0) dl = da += dll + nchannel;

  /*
   * set up rowbufs 1 and 2;
   * rowbuf 0 is always reloaded inside the loop:
   */

    row1 = sl;
    if (edge == 0) sl += sll;
    row2 = sl;
    sl += sll;

    for ( row = 0; row < hgt; row++ ) {

      row0 = row1;
      row1 = row2;
      row2 = sl;

      col = 0;

      p00 = row0[col]; p10 = row1[col]; p20 = row2[col];
      if (edge == 0) col += nchannel;
      p01 = row0[col]; p11 = row1[col]; p21 = row2[col]; col += nchannel;

      /* Process along row, getting 3 output points per loop iteration: */

#pragma pipeloop(0)
      for ( block = 0; block < num_blocks; block++ ) {

        p02 = row0[col]; p12 = row1[col]; p22 = row2[col]; col+=nchannel;
        p03 = row0[col]; p13 = row1[col]; p23 = row2[col]; col+=nchannel;
        p04 = row0[col]; p14 = row1[col]; p24 = row2[col]; col+=nchannel;

        sum1 = sum2 = sum3 = 0.;

        sum1 += p00 * k0; sum2 += p01 * k0; sum3 += p02 * k0;
        sum1 += p01 * k1; sum2 += p02 * k1; sum3 += p03 * k1;
        sum1 += p02 * k2; sum2 += p03 * k2; sum3 += p04 * k2;
        sum1 += p10 * k3; sum2 += p11 * k3; sum3 += p12 * k3;
        sum1 += p11 * k4; sum2 += p12 * k4; sum3 += p13 * k4;
        sum1 += p12 * k5; sum2 += p13 * k5; sum3 += p14 * k5;
        sum1 += p20 * k6; sum2 += p21 * k6; sum3 += p22 * k6;
        sum1 += p21 * k7; sum2 += p22 * k7; sum3 += p23 * k7;
        sum1 += p22 * k8; sum2 += p23 * k8; sum3 += p24 * k8;

        *da = (int)sum1;
        da += nchannel;
        *da = (int)sum2;
        da += nchannel;
        *da = (int)sum3;
        da += nchannel;

        p00 = p03; p01 = p04;                    /* shift last 2 row0 elements */
        p10 = p13; p11 = p14;                    /* shift last 2 row1 elements */
        p20 = p23; p21 = p24;                    /* shift last 2 row2 elements */
      }

      if (leftover_cols) {

        p02 = row0[col];
        p12 = row1[col];
        p22 = row2[col]; col+=nchannel;
        sum1 = 0.;
        sum1 += p00 * k0;                        /* sum of prods */
        sum1 += p01 * k1;
        sum1 += p02 * k2;
        sum1 += p10 * k3;
        sum1 += p11 * k4;
        sum1 += p12 * k5;
        sum1 += p20 * k6;
        sum1 += p21 * k7;
        sum1 += p22 * k8;
        *da = (int)sum1;
        da += nchannel;
        p00 = p01; p01 = p02;                    /* shift last 2 row0 elements */
        p10 = p11; p11 = p12;                    /* shift last 2 row1 elements */
        p20 = p21; p21 = p22;                    /* shift last 2 row2 elements */

        if (leftover_cols > 1) {

          p02 = row0[col];
          p12 = row1[col];
          p22 = row2[col];
          sum1 = 0.;
          sum1 += p00 * k0;                        /* sum of prods */
          sum1 += p01 * k1;
          sum1 += p02 * k2;
          sum1 += p10 * k3;
          sum1 += p11 * k4;
          sum1 += p12 * k5;
          sum1 += p20 * k6;
          sum1 += p21 * k7;
          sum1 += p22 * k8;
          *da = (int)sum1;
          da += nchannel;
          p00 = p01; p01 = p02;                    /* shift last 2 row0 elements */
          p10 = p11; p11 = p12;                    /* shift last 2 row1 elements */
          p20 = p21; p21 = p22;                    /* shift last 2 row2 elements */
        }
      }

      if (edge != 0) {
        sum1 = 0.;
        sum1 += p00 * k0;                        /* sum of prods */
        sum1 += p01 * k1;
        sum1 += p01 * k2;
        sum1 += p10 * k3;
        sum1 += p11 * k4;
        sum1 += p11 * k5;
        sum1 += p20 * k6;
        sum1 += p21 * k7;
        sum1 += p21 * k8;
        *da = (int)sum1;
      }

      sl = sl + sll;                             /* set src ptrs to next row */
      dl = da = dl + dll;                        /* set dst ptrs to next row */
      if ((edge != 0) && (row == (hgt - 2))) sl -= sll;
    }
  }

  return( MLIB_SUCCESS );
}
