/*
 * @(#)dgalock.c	1.6 98/09/21
 *
 * Copyright 1998-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#if sparc

/* #define DGA_DEBUG */

#ifdef DGA_DEBUG
#define DEBUG_PRINT(x)	printf x
#else
#define DEBUG_PRINT(x)
#endif

#include <dga/dga.h>
#include <unistd.h>     /* ioctl */
#include <stdlib.h>
#include <sys/mman.h>   /* mmap */
#include <sys/visual_io.h>
#include <string.h>

/* X11 */
#include <X11/Xlib.h>

#include "jni.h"
#include "jdga.h"
#include "jdgadevice.h"

#include <dlfcn.h>

#define min(x, y)	((x) < (y) ? (x) : (y))
#define max(x, y)	((x) > (y) ? (x) : (y))

typedef struct _SolarisDgaLibInfo SolarisDgaLibInfo;

struct _SolarisDgaLibInfo {
    /* The external structure definition */
    JDgaLibInfo		ext;

    /* The general (non-device specific) information */
    Dga_drawable	dgaDraw;
    jboolean		needsSync;

    /* The device specific memory mapping information */
    SolarisJDgaDevInfo  devInfo;
    SolarisJDgaWinInfo  winInfo;
};

static int
    mmap_dgaDev(SolarisDgaLibInfo *pInfo, Dga_drawable dgadraw, int depth)
{
    void *handle = 0;
    struct vis_identifier visid;
    int fd;
    char libName[64];

    if (!pInfo->devInfo.function) {
        fd = dga_draw_devfd(dgadraw);
        if (ioctl(fd, VIS_GETIDENTIFIER, &visid) != 1) {
	    DEBUG_PRINT(("DGA Device is %s\n", visid.name));
            strcpy(libName, "libjdga");
            strcat(libName, visid.name);
            strcat(libName,".so");
	    DEBUG_PRINT(("calling %s\n", libName));
	    /* we use RTLD_NOW because of bug 4032715 */
            handle = dlopen(libName, RTLD_NOW);
            if (handle != 0) {
                JDgaStatus ret = JDGA_FAILED;
                void *sym = dlsym(handle, "SolarisJDgaDevOpen");
                if (sym != 0) {
                    pInfo->devInfo.majorVersion = JDGALIB_MAJOR_VERSION;
                    pInfo->devInfo.minorVersion = JDGALIB_MINOR_VERSION;
                    ret = (*(SolarisJDgaDevOpenFunc *)sym)(&(pInfo->devInfo));
                }
                if (ret != JDGA_SUCCESS) {
                    dlclose(handle);
                    handle = 0;
                }
            }
            if (handle == 0) {
                DEBUG_PRINT(("DGA Device %s is NOT supported.  "
			     "Using XIMAGE_METHOD.\n", visid.name));
                return JDGA_FAILED;
            }
        } else {
            DEBUG_PRINT(("Failed to identify DGA Device.\n"));
            return JDGA_FAILED;
        }
        /* REMIND: handle the version number here */
    }
    
    pInfo->winInfo.dgaDraw = dgadraw;
    pInfo->winInfo.devInfo = &(pInfo->devInfo);
    return (*pInfo->devInfo.function->winopen)(&(pInfo->winInfo));
}

static void
    unmap_dgaDev(SolarisDgaLibInfo *pInfo)
{
    DEBUG_PRINT(("winclose() called\n"));
   (*pInfo->devInfo.function->winclose)(&(pInfo->winInfo));
}

static JDgaLibInitFunc		Solaris_DGA_LibInit;
static JDgaGetLockFunc		Solaris_DGA_GetLock;
static JDgaReleaseLockFunc	Solaris_DGA_ReleaseLock;
static JDgaXRequestSentFunc	Solaris_DGA_XRequestSent;
static JDgaLibDisposeFunc	Solaris_DGA_LibDispose;

#pragma weak JDgaLibInit = Solaris_DGA_LibInit

static JDgaStatus
    Solaris_DGA_LibInit(JNIEnv *env, JDgaLibInfo **ppInfo, Display *d)
{
    SolarisDgaLibInfo *pInfo;

    /* Note: DGA_INIT can be called multiple times according to docs */
    DEBUG_PRINT(("DGA_INIT called\n"));
    DGA_INIT();

    /* REMIND: Allocate the pInfo structures outside of this library
       and associate each dgadrawable to each structure. */
    pInfo = (SolarisDgaLibInfo *) calloc(1, sizeof(SolarisDgaLibInfo));
    pInfo->ext.display = d;
    pInfo->ext.pGetLock = Solaris_DGA_GetLock;
    pInfo->ext.pReleaseLock = Solaris_DGA_ReleaseLock;
    pInfo->ext.pXRequestSent = Solaris_DGA_XRequestSent;
    pInfo->ext.pLibDispose = Solaris_DGA_LibDispose;

    *ppInfo = &pInfo->ext;
    return JDGA_SUCCESS;
}

static JDgaStatus
    Solaris_DGA_GetLock(JNIEnv *env, JDgaLibInfo *pJInfo,
			Drawable drawable, JDgaSurfaceInfo *pSurface,
			jint lox, jint loy, jint hix, jint hiy)
{
    SolarisDgaLibInfo *pInfo = (SolarisDgaLibInfo *) pJInfo;
    int depth, vis;
    int dlox, dloy, dhix, dhiy;
    
    /* REMIND: Should not ungrab when switching window */
    if (pInfo->dgaDraw != 0 && dga_draw_id(pInfo->dgaDraw) != drawable) {
        unmap_dgaDev(pInfo);
	XDgaUnGrabDrawable(pInfo->dgaDraw);
	pInfo->dgaDraw = 0;
	pInfo->needsSync = JNI_FALSE;
    }

    if (pInfo->dgaDraw == 0) {
	pInfo->dgaDraw = XDgaGrabDrawable(pInfo->ext.display, drawable);
	if (pInfo->dgaDraw == 0) {
	    DEBUG_PRINT(("DgaGrabDrawable failed for 0x%08x\n", drawable));
	    return JDGA_FAILED;
	}
        depth = dga_draw_depth(pInfo->dgaDraw);
        if (mmap_dgaDev(pInfo, pInfo->dgaDraw, depth) != JDGA_SUCCESS) {
            DEBUG_PRINT(("memory map failed for 0x%08x (depth = %d)\n",
                         drawable, depth));
            XDgaUnGrabDrawable(pInfo->dgaDraw);
            pInfo->dgaDraw = 0;
            return JDGA_FAILED;
        }
    }

    if (pInfo->needsSync) {
	XSync(pInfo->ext.display, False);
	pInfo->needsSync = JNI_FALSE;
    }

    DGA_DRAW_LOCK(pInfo->dgaDraw, -1);

    if (dga_draw_site(pInfo->dgaDraw) == DGA_SITE_NULL) {
	DEBUG_PRINT(("zombie drawable = 0x%08x\n", pInfo->dgaDraw));
	DGA_DRAW_UNLOCK(pInfo->dgaDraw);
        unmap_dgaDev(pInfo);
	XDgaUnGrabDrawable(pInfo->dgaDraw);
	pInfo->dgaDraw = 0;
	return JDGA_FAILED;
    }

    /* get the screen address of the drawable */
    dga_draw_bbox(pInfo->dgaDraw, &dlox, &dloy, &dhix, &dhiy);
    dhix += dlox;
    dhiy += dloy;
    DEBUG_PRINT(("window at (%d, %d) => (%d, %d)\n", dlox, dloy, dhix, dhiy));
    pSurface->window.lox = dlox;
    pSurface->window.loy = dloy;
    pSurface->window.hix = dhix;
    pSurface->window.hiy = dhiy;

            /* translate rendering coordinates relative to device bbox */
    lox += dlox;
    loy += dloy;
    hix += dlox;
    hiy += dloy;
    DEBUG_PRINT(("render at (%d, %d) => (%d, %d)\n", lox, loy, hix, hiy));

    vis = dga_draw_visibility(pInfo->dgaDraw);
    switch (vis) {
    case DGA_VIS_UNOBSCURED:
        pSurface->visible.lox = max(dlox, lox);
        pSurface->visible.loy = max(dloy, loy);
        pSurface->visible.hix = min(dhix, hix);
        pSurface->visible.hiy = min(dhiy, hiy);
        DEBUG_PRINT(("unobscured vis at (%d, %d) => (%d, %d)\n",
                     pSurface->visible.lox,
                     pSurface->visible.loy,
                     pSurface->visible.hix,
                     pSurface->visible.hiy));
        break;
    case DGA_VIS_PARTIALLY_OBSCURED: {
        short int x0, y0, x1, y1, *ptr;
        short int cliplox, cliploy, cliphix, cliphiy;

        /*
         * iterate to find out whether the clipped blit draws to a
         * single clipping rectangle
         */
        cliplox = cliphix = lox;
        cliploy = cliphiy = loy;
        ptr = dga_draw_clipinfo(pInfo->dgaDraw);
        while ((y0 = *ptr++) != DGA_Y_EOL) {
            y1 = *ptr++;
            DEBUG_PRINT(("DGA y range loy=%d hiy=%d\n", y0, y1));
            if (y0 < loy) {
                y0 = loy;
            }
            if (y1 > hiy) {
                y1 = hiy;
            }
            while ((x0 = *ptr++) != DGA_X_EOL) {
                x1 = *ptr++;
                DEBUG_PRINT(("  DGA x range lox=%d hix=%d\n", x0, x1));
                if (x0 < lox) {
                    x0 = lox;
                }
                if (x1 > hix) {
                    x1 = hix;
                }
                if (x0 < x1 && y0 < y1) {
                    if (cliploy == cliphiy) {
                                /* First rectangle intersection */
                        cliplox = x0;
                        cliploy = y0;
                        cliphix = x1;
                        cliphiy = y1;
                    } else {
                                /* Can we merge this rect with previous? */
                        if (cliplox == x0 && cliphix == x1 &&
                            cliploy <= y1 && cliphiy >= y0)
                            {
                                /* X ranges match, Y ranges touch */
                                /* => absorb the Y ranges together */
                                cliploy = min(cliploy, y0);
                                cliphiy = max(cliphiy, y1);
                            } else if (cliploy == y0 && cliphiy == y1 &&
                                       cliplox <= x1 && cliphix >= x0)
                                {
                                    /* Y ranges match, X ranges touch */
                                    /* => absorb the X ranges together */
                                    cliplox = min(cliplox, x0);
                                    cliphix = max(cliphix, x1);
                                } else {
                                    /* Assertion: any other combination */
                                    /* means non-rectangular intersect */
                                    DGA_DRAW_UNLOCK(pInfo->dgaDraw);
                                    return JDGA_FAILED;
                                }
                    }
                }
            }
        }
        DEBUG_PRINT(("DGA drawable fits\n"));
        pSurface->visible.lox = cliplox;
        pSurface->visible.loy = cliploy;
        pSurface->visible.hix = cliphix;
        pSurface->visible.hiy = cliphiy;
        break;
    }
    case DGA_VIS_FULLY_OBSCURED:
        pSurface->visible.lox =
            pSurface->visible.hix = lox;
        pSurface->visible.loy =
            pSurface->visible.hiy = loy;
        DEBUG_PRINT(("fully obscured vis\n"));
        break;
    default:
        DEBUG_PRINT(("unknown visibility = %d!\n", vis));
        DGA_DRAW_UNLOCK(pInfo->dgaDraw);
        return JDGA_FAILED;
    }

    pSurface->basePtr = pInfo->winInfo.mapAddr;
    pSurface->surfaceScan = pInfo->winInfo.mapLineStride;
    pSurface->surfaceWidth = pInfo->winInfo.mapWidth;
    pSurface->surfaceHeight = pInfo->winInfo.mapHeight;
    pSurface->surfaceDepth = pInfo->winInfo.mapDepth;

    return JDGA_SUCCESS;
}

static JDgaStatus
    Solaris_DGA_ReleaseLock(JNIEnv *env, JDgaLibInfo *pJInfo)
{
    SolarisDgaLibInfo *pInfo = (SolarisDgaLibInfo *) pJInfo;

    if (pInfo->dgaDraw != 0) {
	DGA_DRAW_UNLOCK(pInfo->dgaDraw);
    }

    return JDGA_SUCCESS;
}

static void
    Solaris_DGA_XRequestSent(JNIEnv *env, JDgaLibInfo *pJInfo,
			     Drawable drawable)
{
    SolarisDgaLibInfo *pInfo = (SolarisDgaLibInfo *) pJInfo;

    if (pInfo->dgaDraw != 0 && dga_draw_id(pInfo->dgaDraw) == drawable) {
	pInfo->needsSync = JNI_TRUE;
    }
}

static void
    Solaris_DGA_LibDispose(JNIEnv *env, JDgaLibInfo *pJInfo)
{
    SolarisDgaLibInfo *pInfo = (SolarisDgaLibInfo *) pJInfo;

    if (pInfo->winInfo.mapDepth != 0) {
	unmap_dgaDev(pInfo);
    }
    if (pInfo->dgaDraw != 0) {
	XDgaUnGrabDrawable(pInfo->dgaDraw);
	pInfo->dgaDraw = 0;
    }
    DEBUG_PRINT(("devclose() called\n"));
    pInfo->devInfo.function->devclose(&(pInfo->devInfo));
    free(pInfo);
}

#endif
