/*
 * @(#)WDropTargetContextPeer.java	1.18 98/09/29
 * 
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 * 
 * CopyrightVersion 1.2_alpha
 * 
 */

package sun.awt.windows;

import java.awt.AWTPermission;
import java.awt.Component;
import java.awt.Container;
import java.awt.Insets;
import java.awt.Point;
import java.awt.Toolkit;

import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.FlavorMap;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;

import java.awt.dnd.DnDConstants;

import java.awt.dnd.DropTarget;
import java.awt.dnd.DropTargetContext;
import java.awt.dnd.DropTargetListener;
import java.awt.dnd.DropTargetEvent;
import java.awt.dnd.DropTargetDragEvent;
import java.awt.dnd.DropTargetDropEvent;
import java.awt.dnd.InvalidDnDOperationException;

import java.awt.dnd.peer.DropTargetContextPeer;

import java.io.InputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.ObjectInputStream;

import java.security.AccessController;
import java.security.AccessControlContext;

import javax.swing.SwingUtilities;

import java.lang.reflect.InvocationTargetException;

import java.util.Iterator;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import java.rmi.MarshalledObject;
import java.rmi.UnmarshalException;

/**
 * <p>
 * The WDropTargetContextPeer class is the class responsible for handling
 * the interaction between the win32 DnD system and Java.
 * </p>
 *
 * @version 1.18
 *
 */

final class WDropTargetContextPeer implements DropTargetContextPeer, Transferable, Runnable {

    /*
     * dispatch constants
     */

    private final static int DISPATCH_NONE    = 0;
    private final static int DISPATCH_ENTER   = 1;
    private final static int DISPATCH_MOTION  = 2;
    private final static int DISPATCH_CHANGED = 3;
    private final static int DISPATCH_EXIT    = 4;
    private final static int DISPATCH_DROP    = 5;

    /*
     * global lock
     */

    private static final Object _globalLock = new Object();

    /*
     * a primitive mechanism for advertising intra-JVM Transferables
     */

    private static Transferable currentJVMLocalSourceTransferable  = null;
    private static AccessControlContext currentJVMLocalSourceACtxt = null;

    static void setCurrentJVMLocalSourceTransferable(Transferable t) throws InvalidDnDOperationException {
	synchronized(_globalLock) {
	    if (t != null && currentJVMLocalSourceTransferable != null) {
	    	throw new InvalidDnDOperationException();
	    } else {
	        currentJVMLocalSourceTransferable = t;
		currentJVMLocalSourceACtxt        = t != null ? AccessController.getContext() : null;
	    }
	} 
    }

    /**
     * obtain the transferable iff the operation is in the same VM
     */

    static Transferable getJVMLocalSourceTransferable() {
	Transferable t = null;

	synchronized(_globalLock) {

	    /*
	     * If the Transferable is local then we should compare the 
	     * AccessContext of the initiator of the Drag with
	     * that of the receipient of the Drop, if they are equal
	     * or if the receipient is prepared to accept cross ctxt
	     *
	     */

 	    if (currentJVMLocalSourceTransferable != null) {
		if (!currentJVMLocalSourceACtxt.equals(AccessController.getContext()))

		    AccessController.checkPermission(new AWTPermission("acceptDropBetweenAccessControllerContexts"));
	    }

	    t = currentJVMLocalSourceTransferable;
	}

	return t;
    }

    /*
     * constants used by dropAccept() or dropReject()
     */

    final static int STATUS_NONE   =  0; // none pending
    final static int STATUS_WAIT   =  1;
    final static int STATUS_ACCEPT =  2;
    final static int STATUS_REJECT = -1;

    private static WDropTargetContextPeer dropTargetContextPeer;

    /**
     * create the peer typically upcall from C++
     */

    static WDropTargetContextPeer getWDropTargetContextPeer() {
	synchronized(_globalLock) {
	    dropTargetContextPeer = new WDropTargetContextPeer();
	}

	return dropTargetContextPeer;
    }

    /**
     * upcall to encapsulate memory buffer transfer
     */

    private static ByteArrayInputStream getMemoryStream(byte[] data, String nativeType) {
	return new ByteArrayInputStream(data);
    }

    /**
     * upcall to encapsulate file transfer
     */

    private static FileInputStream getFileStream(String file, String nativeType, int stgmedium) throws IOException {
	return new WDropTargetContextPeerFileStream(file, nativeType, stgmedium);
    }

    /**
     * upcall to encapsulate IStream buffer transfer
     */

    private static Object getIStream(int istream, String nativeType) throws IOException {
	return new WDropTargetContextPeerIStream(istream, nativeType);
    }

    /**
     * upcall to encapsulate an array of objects as a List.
     */

    private static java.util.List getList(Object[] list) {
	return java.util.Arrays.asList(list);
    }

    /**
     * create the peer
     */

    WDropTargetContextPeer() {
	super();
    }

    /**
     * @return the DropTarget associated with this peer
     */

    public DropTarget getDropTarget() { synchronized(lock) { return currentDT; } }

    /**
     * @param actions set the current actions
     */

    public void setTargetActions(int actions) {
	synchronized(lock) {
	    currentA = actions & (DnDConstants.ACTION_COPY_OR_MOVE | DnDConstants.ACTION_LINK);
	}
    }
  
    /**
     * @return the current target actions
     */

    public int getTargetActions() { synchronized(lock) { return currentA; } }
  
    /**
     *  set the Autoscroll insets ... a noop
     */

    public synchronized void setAutoscrollInsets(Insets ins) {
	// noop
    }

    /**
     *  set the Autoscroll insets ... a noop
     */

    public synchronized void setAutoscrollDelay(int dly) {
	// noop
    }

    /**
     * does the platform support autoscrolling ... not likely
     */
 
    public synchronized boolean doesPlatformImplementAutoscrolling() { return false; }

    /**
     * get the Transferable associated with the drop
     */
 
    public synchronized Transferable getTransferable() {
        return this;
    }

    /**
     * @return current DataFlavors available
     */

    public DataFlavor[] getTransferDataFlavors() {
	DataFlavor[]  dfs;

	if (local != null) {  
	    dfs = local.getTransferDataFlavors();
	} else synchronized(lock) {
	    Object[] values = currentDT.getFlavorMap().getFlavorsForNatives(currentT).values().toArray();

	    dfs = new DataFlavor[values.length];

	    System.arraycopy(values, 0, dfs, 0, values.length);
	}

	return dfs;
    }

    /**
     * @return if the Flavor is supported
     */

    public boolean isDataFlavorSupported(DataFlavor df) {

	if (local != null) return local.isDataFlavorSupported(df);

  	DataFlavor[] dfs = getTransferDataFlavors();

	for (int i = 0; i < dfs.length; i++) {
	    if (dfs[i].equals(df)) return true;
	}
	
	return false;
    }

    /**
     * @return the data 
     */

    public Object getTransferData(DataFlavor df) throws UnsupportedFlavorException, IOException, InvalidDnDOperationException {
	
	if (local != null) return local.getTransferData(df);

        String atom = null;

	synchronized(lock) {
	    if (dropStatus != STATUS_ACCEPT || dropComplete) 
		throw new InvalidDnDOperationException("No drop current");


	    Map atoms = currentDT.getFlavorMap().getNativesForFlavors(new DataFlavor[] { df } );

	    if ((atom = (String)atoms.get(df)) == null)
		throw new UnsupportedFlavorException(df);

	    int i;

	    for (i = 0; i < currentT.length; i++) {
	        if (currentT[i].equals(atom)) break;
	    }

	    if (i == currentT.length)
		throw new UnsupportedFlavorException(df);

	}

	Object ret = getData(nativeDragContext, atom);

	if (ret == null) throw new IOException("no native data was transferred");

	if (ret instanceof java.util.List) return ret; // special case ...

	if (df.isRepresentationClassInputStream()) {
	    final Class dfrc = df.getRepresentationClass();

	    if (java.io.InputStream.class.equals(dfrc))
		return (InputStream)ret; // simple case
	    else {
	        java.lang.reflect.Constructor[] constructors = null;

		try {
		    constructors = (java.lang.reflect.Constructor[])
			java.security.AccessController.doPrivileged(
				    new java.security.PrivilegedAction() {
			    public Object run() {
				return dfrc.getConstructors();
			    }
			});
		} catch (SecurityException se) {
		    se.printStackTrace();
		    throw new IOException(se.getMessage());
		}

		java.lang.reflect.Constructor constructor = null;

		for (int j = 0; j < constructors.length; j++) {
		    if (!java.lang.reflect.Modifier.isPublic(constructors[j].getModifiers())) continue;

		    Class[] ptypes = constructors[j].getParameterTypes();

		    if (ptypes != null && ptypes.length == 1 && java.io.InputStream.class.equals(ptypes[0])) {
		        constructor = constructors[j];
		        break;
		    }
		}

		if (constructor == null) 
		     throw new InvalidDnDOperationException("cant find <init>(Ljava/io/InputStream;)V for class: " + dfrc.getName());


		try {
		    return constructor.newInstance(new Object[] { (InputStream)ret } );
		} catch (Exception e) {
		    e.printStackTrace();

		    throw new InvalidDnDOperationException(e.getMessage());
		}
	    }
	} else if (df.isRepresentationClassRemote()) {
	    if (currentDA != DnDConstants.ACTION_LINK)
		throw new InvalidDnDOperationException("only ACTION_LINK is permissable for transfer java.rmi.Remote objects");

	    try {
		InputStream is = (InputStream)ret;
	 	Object      o;

		o = ((MarshalledObject)(new ObjectInputStream(is)).readObject()).get();
		is.close();
		
		return o;
	    } catch (Exception e) {
		e.printStackTrace();

                throw new InvalidDnDOperationException(e.getMessage());
	    }
	} else if (df.isRepresentationClassSerializable()) {
	    try {
		InputStream is = (InputStream)ret;
		Object      o;

		o = (new ObjectInputStream(is)).readObject();

		is.close();

		return o;
	    } catch (Exception e) {
		e.printStackTrace();

                throw new InvalidDnDOperationException(e.getMessage());
	    }
	} else if (java.lang.String.class.equals(df.getRepresentationClass())) {
	    try { 
	        InputStream	      is   = (InputStream)ret;
	        ByteArrayOutputStream baos = new ByteArrayOutputStream();
		int		      len  = 0;
		byte[]		      buf  = new byte[64];

		while ((len = is.read(buf)) != -1) {
		    baos.write(buf, 0, len);
		}
	
	        return new String(baos.toByteArray());
	    } catch (Exception e) {
		e.printStackTrace();

                throw new InvalidDnDOperationException(e.getMessage());
	    }
	}

	throw new IOException("getTransferData() failed");
    }

    /**
     * @return if the transfer is a local one
     */
 
    public boolean isTransferableJVMLocal() {
        return local != null || getJVMLocalSourceTransferable() != null;
    }

    /**
     * called from EventDispatchThread via EventQueue
     */

    public void run() {
	int mp;

	synchronized(syncLock) { mp = messagePending; }

	switch (mp) {
	    case DISPATCH_ENTER:
		processEnterMessage();
		break;

	    case DISPATCH_CHANGED:
	    case DISPATCH_MOTION:
		processMotionMessage(mp);
		break;

	    case DISPATCH_EXIT:
		processExitMessage();
		break;

	    case DISPATCH_DROP:
		processDropMessage();
		return; // messagePending already cleared & synLock notified
	}
	
	synchronized(syncLock) {
	    if (messagePending != DISPATCH_NONE) {
	        messagePending = DISPATCH_NONE;
	        syncLock.notifyAll();
	    }	
	}
    }

    /**
     * upcall to handle enter messages
     */
  
    private int handleEnterMessage(Component component, int x, int y, int dropAction, int actions, String[] atoms, int nativeCtxt) {
	synchronized(syncLock) {
	    while (messagePending != DISPATCH_NONE) try {
                syncLock.wait();
            } catch (InterruptedException ie) {
                // do nothing
            }

	    // setup context

	    nativeDragContext = nativeCtxt;
	    currentNC	      = component;
	    currentT          = atoms;
	    currentX          = x;
	    currentY          = y;
	    currentSA         = actions;
	    currentDA         = dropAction;

	    // schedule callback 

	    messagePending = DISPATCH_ENTER; SwingUtilities.invokeLater(this);

	    while (messagePending != DISPATCH_NONE) try {
	        syncLock.wait();
	    } catch (InterruptedException ie) {
	        // do nothing
	    }

	    return currentA; // return target's response
	}
    }

    /**
     * actual processing on EventQueue Thread
     */

    private void processEnterMessage() {
	Component  c;
	DropTarget dt   = null;
	boolean    hit;
	Point	   hots = null;

	// we do hit detection and co-ord translation "locked" of course

	local = getJVMLocalSourceTransferable();

	synchronized(currentNC.getTreeLock()) {
	    try {
	        c = ((Container)currentNC).findComponentAt(currentX, currentY);
	    } catch (ClassCastException cce) {
		c = currentNC.locate(currentX, currentY);
	    }

	    hit = (c != null    		    &&
		   c.isShowing()		    &&
		   (dt = c.getDropTarget()) != null &&
		   dt.isActive()
	          );

	    if (hit) {
	    	 hots = translate(currentNC, currentX, currentY, c);
            }
	}

	synchronized(lock) {
	    if (currentDTC != null) { // some wreckage from last time
	        currentDTC.removeNotify();
	        currentDTC = null;
	    }
	}

	if (hit) { 
	    synchronized(lock) {
	        currentC   = c;
	        currentDT  = dt;
	        currentDTC = currentDT.getDropTargetContext();

	        currentDTC.addNotify(this);

	        currentA   = dt.getDefaultActions();
	    }

	    try {
	        ((DropTargetListener)dt).dragEnter(new DropTargetDragEvent(currentDTC, hots, currentDA, currentSA));
	    } catch (Exception e) {
		e.printStackTrace();
		synchronized(lock) { currentA = DnDConstants.ACTION_NONE; }
	    }
	} else synchronized(lock) {
	    currentDT  = null;
	    currentDTC = null;
	    currentA   = DnDConstants.ACTION_NONE;
	}
    }

    /**
     * upcall to handle exit messages, synchronized with AWT_LOCK()
     */

    private void handleExitMessage(Component component, int nativeCtxt) {
	synchronized(syncLock) {
	    while (messagePending != DISPATCH_NONE) try {
                syncLock.wait();
            } catch (InterruptedException ie) {
                // do nothing
            }

	    nativeDragContext = nativeCtxt;

	    messagePending = DISPATCH_EXIT; SwingUtilities.invokeLater(this);

	    while (messagePending != DISPATCH_NONE) try {
	        syncLock.wait();
	    } catch (InterruptedException ie) {
	        // do nothing!
	    }
	}
    }

    /**
     * 
     */ 

    private void processExitMessage() {
	DropTarget        dt  = null;
	DropTargetContext dtc = null;

	synchronized(lock) { 
	    if (currentC != null) {
	        DropTarget t = currentC.getDropTarget();

	    	if (t == null) {
		    currentC  = null;
		    currentDT = null;
		    currentT  = null;

		    if (currentDTC != null) currentDTC.removeNotify();
		    
		    currentDTC = null;

		    return;
	        }

		if (t != currentDT) {

		    if (currentDTC != null) currentDTC.removeNotify();

		    dt  = currentDT  = t;
		    dtc = currentDTC = t.getDropTargetContext();

		    currentDTC.addNotify(this);
	        } else {
		    dt  = currentDT;
		    dtc = currentDTC;
		}
	    }
	}

	if (dt != null && dt.isActive()) try {
		((DropTargetListener)dt).dragExit(new DropTargetEvent(dtc));
	} catch (Exception e) {
		e.printStackTrace();
	} finally {
	    synchronized(lock) {
		currentA  = DnDConstants.ACTION_NONE;
		currentSA = DnDConstants.ACTION_NONE;
		currentDA = DnDConstants.ACTION_NONE;
		currentC  = null;
		currentDT = null;
		currentT  = null;

		currentDTC.removeNotify();
		currentDTC = null;

		local = null;
	    }
	 }
    }

    /**
     * upcall to handle motion
     */
  
    private synchronized int handleMotionMessage(Component component, int x, int y, int dropAction, int actions, String[] atoms, int nativeCtxt) {
	int prevDA; // previous dropAction from last motion or Enter

	synchronized(syncLock) {
	    while (messagePending != DISPATCH_NONE) try {
                syncLock.wait();
            } catch (InterruptedException ie) {
                // do nothing
            }

	    // setup context

	    nativeDragContext = nativeCtxt;
	    currentNC	      = component;
	    currentT          = atoms;
	    currentX          = x;
	    currentY          = y;
	    currentSA         = actions;
	    prevDA	      = currentDA;
	    currentDA	      = dropAction;

	    // schedule callback 

	    messagePending = prevDA == currentDA ? DISPATCH_MOTION : DISPATCH_CHANGED;
	    SwingUtilities.invokeLater(this);

	    while (messagePending != DISPATCH_NONE) try {
	        syncLock.wait();
	    } catch (InterruptedException ie) {
	        // do nothing
	    }

	    return currentA; // return target's response
	}
    }

    /**
     *
     */

    private void processMotionMessage(int dispatch) {
	Component  	  c     = null;
	Point      	  hots  = null;
	Point      	  nhots = null;
	boolean    	  hit   = false;
	boolean    	  nhit  = false;
	DropTarget 	  dt    = null;
	DropTargetContext dtc   = null;
	DropTarget        ndt   = null;

	if (currentC != null) {
	    boolean same = false;

	    synchronized(currentNC.getTreeLock()) {
		try {
	            c = ((Container)currentNC).findComponentAt(currentX, currentY);
		} catch (ClassCastException cce) {
		    
	            c = currentNC.locate(currentX, currentY);
		}

		hots = translate(currentNC, currentX, currentY, currentC);

		same = currentC.equals(c);

		if (!same && c != null) {
		    nhots = translate(currentNC, currentX, currentY, c);
		    nhit  = c.isShowing()		      &&
			    (ndt = c.getDropTarget()) != null &&
			    ndt.isActive();
		}

		synchronized(lock) {
	            if ((dt = currentC.getDropTarget()) == null) {
		        currentDTC.removeNotify();
		        currentC   = null;
		        currentDTC = null;
		        currentT   = null;
	            } else {
		        if (currentDT != dt) {
			    if (currentDTC != null) currentDTC.removeNotify();
    
			    currentDT  = dt;
			    currentDTC = null;
		        }

			dtc = currentDT.getDropTargetContext();
		        if (dtc != currentDTC) {
			    if (currentDTC != null) currentDTC.removeNotify();
			    (currentDTC = dtc).addNotify(this);
		        }

	                hit = currentC.isShowing() && dt.isActive();
		    }
		}
	    }

	    if (hit) {
	        if (!same) {
		    try {
			((DropTargetListener)dt).dragExit(new DropTargetEvent(dtc));
		    } catch (Exception e) {
			e.printStackTrace();
		    } finally {
			synchronized(lock) {
		            currentDTC.removeNotify();
		            currentDTC = null;
		            currentC   = null;
		            currentT   = null;
			    currentA   = DnDConstants.ACTION_NONE;
			}
		    }
	        } else {
		    try {
			if (dispatch == DISPATCH_MOTION)
		            ((DropTargetListener)dt).dragOver(new DropTargetDragEvent(dtc, hots, currentDA, currentSA));
			else
		            ((DropTargetListener)dt).dropActionChanged(new DropTargetDragEvent(dtc, hots, currentDA, currentSA));
		    } catch (Exception e) {
			e.printStackTrace();
			synchronized (lock) { currentA = DnDConstants.ACTION_NONE; }
		    } finally {
	                return;
		    }
		}
	    } else synchronized(lock) { currentA = DnDConstants.ACTION_NONE; }
	}

	// lookup possibly new component

	if (c == null) synchronized(currentNC.getTreeLock()) {
	    try {
	        c = ((Container)currentNC).findComponentAt(currentX, currentY);
	    } catch (ClassCastException cce) {
	        c = currentNC.locate(currentX, currentY);
	    }

	    nhit = c != null			     &&
		   c.isShowing()		     &&
		   (ndt = c.getDropTarget()) != null &&
		   ndt.isActive();

	    if (nhit) nhots = translate(currentNC, currentX, currentY, c);
	}

	if (nhit) {
	    synchronized(lock) {
	        currentC   = c;

	        dt  = currentDT  = ndt;
	        dtc = currentDTC = ndt.getDropTargetContext();

	        currentDTC.addNotify(this);

	        currentA   = ndt.getDefaultActions();
	    }

	    try {
	        ((DropTargetListener)dt).dragEnter(new DropTargetDragEvent(dtc, nhots, currentDA, currentSA));
	    } catch (Exception e) {
		e.printStackTrace();
	        synchronized(lock) { currentA = DnDConstants.ACTION_NONE; }
	    }
        } else synchronized(lock) {
	    currentC   = null;
	    currentDT  = null;
	    currentDTC = null;

	    currentA = DnDConstants.ACTION_NONE;
	}
    }

    /**
     * upcall to handle the Drop message
     */
  
    private void handleDropMessage(Component component, int x, int y, int dropAction, int actions, String[] atoms,int nativeCtxt) {
	synchronized(syncLock) {
	    while (messagePending != DISPATCH_NONE) try {
                syncLock.wait();
            } catch (InterruptedException ie) {
                // do nothing
            }

	    // setup context

	    nativeDragContext = nativeCtxt;
	    currentNC	      = component;
	    currentT          = atoms;
	    currentX          = x;
	    currentY          = y;
	    currentSA         = actions;
	    currentDA	      = dropAction;

	    // schedule callback 

	    messagePending = DISPATCH_DROP; SwingUtilities.invokeLater(this);

	    while (messagePending != DISPATCH_NONE) try {
	        syncLock.wait();
	    } catch (InterruptedException ie) {
	        // do nothing
	    }
	}
    }

    /**
     * 
     */
  
    private void processDropMessage() {
	Component  c    = null;
	Point      hots = null;
	boolean    hit  = false;
	DropTarget dt   = null;

	synchronized(currentNC.getTreeLock()) {
	    try {
	        c = ((Container)currentNC).findComponentAt(currentX, currentY);
	    } catch (ClassCastException cce) {
	        c = currentNC.locate(currentX, currentY);
	    }
	
	    hit = (c != null    		    &&
		   c.isShowing()		    &&
		   (dt = c.getDropTarget()) != null &&
		   dt.isActive()
	          );

	    if (hit) {
		hots = translate(currentNC, currentX, currentY, c);
	    }
	}

	synchronized(lock) {
	    dropStatus   = STATUS_WAIT; // drop pending ACK
	    dropComplete = false;
	}

	if (hit) {
	    DropTargetContext dtc;
	    boolean	      exception = false;

	    synchronized(lock) {
	        currentC  = c;
	        currentDT = dt;

	        dtc = dt.getDropTargetContext();

	        if (currentDTC != null) currentDTC.removeNotify();

	        (currentDTC = dtc).addNotify(this);
	    }

	    synchronized(_globalLock) {
	        if ((local = getJVMLocalSourceTransferable()) != null)
	            setCurrentJVMLocalSourceTransferable(null);
	    }

	    try {

		/*
		 * for platform constistency we call dropExit before
		 * drop ...
		 */

	        ((DropTargetListener)dt).dragExit(new DropTargetEvent(dtc));

		// now process the drop.

	        ((DropTargetListener)dt).drop(new DropTargetDropEvent(dtc, hots, currentDA, currentSA, local != null));
	    } catch (Exception e) {
	 	e.printStackTrace();
		exception = true;
	    } finally {
	        synchronized(lock) {
		    if (dropStatus == STATUS_WAIT || exception) {
		        rejectDrop();
		    }
		}
	    }
	} else {
	    rejectDrop();
	}
    }

    /**
     * acceptDrag
     */

   public void acceptDrag(int dragOperation) {
	synchronized(lock) {
	    if (currentDT == null)
	        throw new IllegalArgumentException("no drag pending");

	    currentA = mapOperation(currentDT, dragOperation);
	}
   }

    /**
     * rejectDrag
     */

   public void rejectDrag() {
	synchronized(lock) {
	    if (currentDT == null)
	        throw new IllegalArgumentException("no drag pending");

	    currentA = DnDConstants.ACTION_NONE;
	}
    }

    /**
     * acceptDrop
     */

    public void acceptDrop(int dropOperation) {

	if (dropOperation == DnDConstants.ACTION_NONE)
	    throw new IllegalArgumentException("invalid acceptDrop() action");

	synchronized (lock) {
	    if (dropStatus != STATUS_WAIT) {
		throw new InvalidDnDOperationException("invalid acceptDrop()");
	    }

	    currentDA = currentA = mapOperation(currentDT, dropOperation & currentSA);

	    dropStatus   = STATUS_ACCEPT;
	    dropComplete = false;

	    dropDispatched();
	}
    }

    /**
     * reject Drop
     */

    public void rejectDrop() {
	synchronized(lock) {
	    if (dropStatus != STATUS_WAIT) {
		throw new InvalidDnDOperationException("invalid rejectDrop()");
	    }

	    dropStatus = STATUS_REJECT;
	}

	dropComplete(false);
    }

    /**
     * mapOperation
     */

    private int mapOperation(DropTarget dt, int operation) {
	int[] operations = { 
		DnDConstants.ACTION_MOVE,
		DnDConstants.ACTION_COPY,
		DnDConstants.ACTION_LINK,
	};
	int   ret = DnDConstants.ACTION_NONE;

	for (int i = 0; i < operations.length; i++) {
	    if ((operation & operations[i]) == operations[i]) {
		    ret = operations[i];
		    break;
	    }
	}

	return ret;
    }

    /**
     * signal drop complete
     */

    public void dropComplete(boolean success) {
	synchronized(lock) {
	    if (dropStatus == STATUS_NONE) {
		throw new InvalidDnDOperationException("No Drop pending");
	    }

	    if (currentDTC != null) currentDTC.removeNotify();

	    currentC   = null;
	    currentDT  = null;
	    currentDTC = null;
	    currentT   = null;
	    currentA   = DnDConstants.ACTION_NONE;

	    synchronized(_globalLock) {
	        currentJVMLocalSourceTransferable = null;
	    }

	    dropStatus   = STATUS_NONE;
	    dropComplete = true;
	}

	dropDispatched();
	dropDone(nativeDragContext, success ? currentDA : DnDConstants.ACTION_NONE);

	currentDA = DnDConstants.ACTION_NONE;

	nativeDragContext = 0;
    }

    /**
     * get the Status of the drop
     */

    int getDropStatus() { synchronized(lock) { return dropStatus; } }

    /**
     * translate co-ordinates - called with Component Tree Lock held
     */

    private Point translate(Component anc, int x, int y, Component desc) {
	Component c  = desc;
	int 	  dx = 0;
	int	  dy = 0;

	while (!c.equals(anc)) {
	    Point locn = c.getLocation();
	    dx += locn.x;
	    dy += locn.y;
	    c = c.getParent();
	}

	return new Point(x - dx, y - dy);
    }

    /**
     * called only by acceptDrop and dropComplete
     */

    private void dropDispatched() {
        synchronized(syncLock) {
	    if (messagePending == DISPATCH_DROP) {
		messagePending = DISPATCH_NONE;
		syncLock.notifyAll();
	    }
	}
    }

    /**
     * downcall to bind transfer data.
     */

     private native Object getData(int nativeContext, String atom);

    /**
     * downcall to notify that drop is complete
     */

     private native void dropDone(int nativeContext, int actions);

    /**
     * native code initialization
     */

    static native void Initialize();

    /*
     * instance vars ...
     */

    private Component	      currentC; 
    private DropTarget        currentDT;
    private DropTargetContext currentDTC;
    private String[]	      currentT;
    private int		      currentA;
    private int		      currentDA;


    private int	      	      nativeDragContext;

    private Transferable      local;

    // follow are used to pass context from platform upcall thread to
    // callback on EventQueue

    private int		      messagePending = DISPATCH_NONE;

    private Component	      currentNC;// current native Container
    private int		      currentX; // current X coord
    private int		      currentY; // current Y coord
    private int		      currentSA;// source actions

    // ...

    int		              dropStatus   = STATUS_NONE;
    boolean		      dropComplete = false;

    Object	      	      lock                = new Object();
    Object		      syncLock	  	  = new Object();
}

/**
 * package private class to handle file transfers
 */

class WDropTargetContextPeerFileStream extends FileInputStream {

    /**
     * construct file input stream
     */

    WDropTargetContextPeerFileStream(String name, String nativeType, int stgmedium) throws FileNotFoundException {
	super(name);

	this.nativeType = nativeType;
	this.stgmedium  = stgmedium;
    }

    /**
     * close
     */

    public void close() throws IOException {
	super.close();

	freeStgMedium(stgmedium);
    }

    /**
     * free underlying windows data structure
     */

    private native void freeStgMedium(int stgmedium);

    /*
     * fields
     */

    private String nativeType;
    private int    stgmedium;
}

/**
 * Package private class to access IStream objects
 */

class WDropTargetContextPeerIStream extends InputStream {

    /**
     * construct a WDropTargetContextPeerIStream wrapper
     */

    WDropTargetContextPeerIStream(int istream, String nativeType) throws IOException {
	super();

	if (istream == 0) throw new IOException("No IStream");

	this.istream    = istream;
	this.nativeType = nativeType;
    }

    /**
     * @return bytes available
     */

    public int available() throws IOException {
	if (istream == 0) throw new IOException("No IStream");
	
	return Available(istream);
    }

    private native int Available(int istream);

    /**
     * read
     */

    public int read() throws IOException {
	if (istream == 0) throw new IOException("No IStream");
	return Read(istream);
    }

    private native int Read(int istream) throws IOException;

    /**
     * read into buffer
     */

    public int read(byte[] b, int off, int len) throws IOException {
	if (istream == 0) throw new IOException("No IStream");
	return ReadBytes(istream, b, off, len);
    }

    private native int ReadBytes(int istream, byte[] b, int off, int len) throws IOException;

    /**
     * close
     */

    public void close() throws IOException {
	super.close();
	Close(istream);

	istream = 0;
    }

    private native void Close(int istream);

    /*
     * fields
     */

    private int	   istream;
    private String nativeType;
}
