/*
 * @(#)WInputMethod.java	1.29 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


package sun.awt.windows;

import java.awt.*;
import java.awt.peer.*;
import java.awt.event.*;
import java.awt.im.*;
import java.awt.font.*;
import java.text.*;
import java.text.AttributedCharacterIterator.Attribute;
import java.lang.Character.Subset;
import java.lang.Character.UnicodeBlock;
import java.lang.System;
import java.util.Locale;
import java.util.Map;
import java.util.HashMap;
import sun.awt.im.*;
import sun.awt.im.InputMethodHighlightMapping.InputMethodHighlightAttribute;

public class WInputMethod extends InputMethodAdapter
{
    private Component target;
    private int context;
    private boolean open; //default open status;
    private int cmode;    //default conversion mode;

    private boolean disposed = false;

    // attribute definition in Win32 (in IMM.H)
    public final static byte ATTR_INPUT			= 0x00;
    public final static byte ATTR_TARGET_CONVERTED	= 0x01;
    public final static byte ATTR_CONVERTED		= 0x02;
    public final static byte ATTR_TARGET_NOTCONVERTED	= 0x03;
    public final static byte ATTR_INPUT_ERROR		= 0x04;
    // cmode definition in Win32 (in IMM.H)
    public final static int  IME_CMODE_ALPHANUMERIC     = 0x0000;
    public final static int  IME_CMODE_NATIVE           = 0x0001;
    public final static int  IME_CMODE_KATAKANA         = 0x0002;  
    public final static int  IME_CMODE_LANGUAGE         = 0x0003;
    public final static int  IME_CMODE_FULLSHAPE        = 0x0008;
    public final static int  IME_CMODE_HANJACONVERT     = 0x0040;
    public final static int  IME_CMODE_ROMAN            = 0x0010;

    // Language ID and macro definition in Win32 
    public final static int  LANG_CHINESE               = 0x04;
    public final static int  LANG_JAPANESE              = 0x11;
    public final static int  LANG_KOREAN                = 0x12;
    public final static int  PRIMARYLANGID(int idLang){
        return (idLang & 0x3ff); 
    }
  
    // flag values for endCompositionNative() behavior
    private final static boolean COMMIT_INPUT           = true;
    private final static boolean DISCARD_INPUT          = false;

    private static WHighlightMapping highlightMapping = new WHighlightMapping();
    private static HashMap mappingTable;

    // Initialize highlight mapping table and its mapper
    static {
	HashMap map = new HashMap();
	HashMap value;

	value = new HashMap();
	value.put(InputMethodHighlightAttribute.INPUT_METHOD_UNDERLINE, InputMethodHighlightAttribute.UNDERLINE_LOW_DOTTED);
	map.put(InputMethodHighlight.UNSELECTED_RAW_TEXT_HIGHLIGHT, value);

	value = new HashMap();
	value.put(InputMethodHighlightAttribute.INPUT_METHOD_UNDERLINE, InputMethodHighlightAttribute.UNDERLINE_LOW_DOTTED);
	map.put(InputMethodHighlight.UNSELECTED_CONVERTED_TEXT_HIGHLIGHT, value);

	value = new HashMap();
	Color navyBlue = new Color(0, 0, 128);
	value.put(TextAttribute.FOREGROUND, navyBlue);
	value.put(TextAttribute.BACKGROUND, Color.white);
	value.put(TextAttribute.SWAP_COLORS, TextAttribute.SWAP_COLORS_ON);
	value.put(InputMethodHighlightAttribute.INPUT_METHOD_UNDERLINE, InputMethodHighlightAttribute.UNDERLINE_LOW_ONE_PIXEL);
	map.put(InputMethodHighlight.SELECTED_CONVERTED_TEXT_HIGHLIGHT, value);

	value = new HashMap();
	value.put(InputMethodHighlightAttribute.INPUT_METHOD_UNDERLINE, InputMethodHighlightAttribute.UNDERLINE_LOW_GRAY);
	map.put(InputMethodHighlight.SELECTED_RAW_TEXT_HIGHLIGHT, value);

	mappingTable = map;
    }

    public WInputMethod()
    {
	context = createNativeContext();
	cmode = getConversionStatus(context);
	open = getOpenStatus(context);
    }

    protected void finalize() throws Throwable
    {
	dispose();
	super.finalize();
    }

    /*
     * Subclasses should override disposeImpl() instead of dispose(). Client
     * code should always invoke dispose(), never disposeImpl().
     */
    protected void disposeImpl() {
	if (context!=0) {
	    destroyNativeContext(context);
	    context=0;
	}
    }
    public final void dispose() {
        boolean call_disposeImpl = false;

        if (!disposed) {
	    synchronized (this) {
	        if (!disposed) {
		    disposed = call_disposeImpl = true;
		}
	    }
	}

	if (call_disposeImpl) {
	    disposeImpl();
	}
    }

    /**
     * Returns null.
     *
     * @see sun.awt.im.InputMethod#getControlObject
     */
    public Object getControlObject() {
        return null;
    }

    public boolean setLocale(Locale lang) {
        // TODO: need to determine which locales are really supported
        return true;
    }

    /**
     * Implements InputMethod.setCharacterSubsets for Windows.
     *
     * @see sun.awt.im.InputMethod#setCharacterSubsets
     */
    public void setCharacterSubsets(Subset[] subsets) {
	if (subsets == null){
            setConversionStatus(context, cmode);
	    setOpenStatus(context, open);
	    return;
	}

        // Use first subset only. Other subsets in array is ignored.
        // This is restriction of Win32 implementation.
        Subset subset1 = subsets[0];

        int lang = PRIMARYLANGID(getNativeIMELanguage());

        if (lang == LANG_JAPANESE) {
            if (subset1 == UnicodeBlock.BASIC_LATIN || subset1 == InputSubset.LATIN_DIGITS) {
                setOpenStatus(context, false);
            } else {
                int newmode;
                if (subset1 == UnicodeBlock.CJK_UNIFIED_IDEOGRAPHS
                    || subset1 == InputSubset.KANJI 
                    || subset1 == UnicodeBlock.HIRAGANA) 
	            newmode = IME_CMODE_NATIVE | IME_CMODE_FULLSHAPE;
	        else if (subset1 == UnicodeBlock.KATAKANA)
	            newmode = IME_CMODE_NATIVE | IME_CMODE_KATAKANA| IME_CMODE_FULLSHAPE;
	        else if (subset1 == InputSubset.HALFWIDTH_KATAKANA)
	            newmode = IME_CMODE_NATIVE | IME_CMODE_KATAKANA;
                else 
                    return;
                setOpenStatus(context, true);
                newmode |= (getConversionStatus(context)&IME_CMODE_ROMAN);   // reserve ROMAN input mode
                setConversionStatus(context, newmode);
            }
        } else if (lang == LANG_KOREAN) {
	    if (subset1 == UnicodeBlock.BASIC_LATIN || subset1 == InputSubset.LATIN_DIGITS) {
                setOpenStatus(context, false);
            } else if (subset1 == UnicodeBlock.CJK_UNIFIED_IDEOGRAPHS 
                       || subset1 == InputSubset.HANJA
                       || subset1 == UnicodeBlock.HANGUL_SYLLABLES
                       || subset1 == UnicodeBlock.HANGUL_JAMO
                       || subset1 == UnicodeBlock.HANGUL_COMPATIBILITY_JAMO) {
                setOpenStatus(context, true);
                setConversionStatus(context, IME_CMODE_NATIVE);
            }
        } else if (lang == LANG_CHINESE) {
	    if (subset1 == UnicodeBlock.BASIC_LATIN || subset1 == InputSubset.LATIN_DIGITS) {
                setOpenStatus(context, false);
            } else if (subset1 == UnicodeBlock.CJK_UNIFIED_IDEOGRAPHS
		       || subset1 == InputSubset.TRADITIONAL_HANZI
		       || subset1 == InputSubset.SIMPLIFIED_HANZI) {
                setOpenStatus(context, true);
                setConversionStatus(context, IME_CMODE_NATIVE);
            }
        }
    }

    public void dispatchEvent(AWTEvent e)
    {
	if ( e instanceof ComponentEvent){
	    Component comp = ((ComponentEvent)e).getComponent();
	    if (comp==target){
		WComponentPeer peer = getNearestNativePeer(target);
		if (peer==null)		return;
		handleNativeIMEEvent(peer, e);
	    }
	}
    }

    public void activate()
    {
	target = getClientComponent();
	WComponentPeer peer = getNearestNativePeer(target);
	if (peer==null)		return;
	enableNativeIME(peer, context, !haveActiveClient() );
    }

    public void deactivate(boolean isTemporary)
    {
	WComponentPeer peer = getNearestNativePeer(target);
	target = null;
	if (peer==null)		return;
	disableNativeIME(peer);
    }

    /**
     * @see sun.awt.im.InputMethod#removeNotify
     */
    public void removeNotify() {
        endCompositionNative(context, DISCARD_INPUT);
    }

    /**
     * @see sun.awt.im.InputMethodHighlightMapping
     */
    public static InputMethodHighlightMapping getInputMethodHighlightMapping() {
	return (InputMethodHighlightMapping) highlightMapping;
    }

    public void endComposition()
    {
	//right now the native endCompositionNative() just cancel
        //the composition string, maybe a commtting is desired
        endCompositionNative(context, COMMIT_INPUT);
    }

    public void sendInputMethodEvent(int id, String text,
                                     int[] clauseBoundary, String[] clauseReading,
                                     int[] attributeBoundary, byte[] attributeValue,
				     int commitedTextLength, int caretPos, int visiblePos)
    {

	AttributedCharacterIterator iterator = null;

	if (text!=null) {

	    // construct AttributedString
	    AttributedString attrStr = new AttributedString(text);

	    // set Language Information
	    attrStr.addAttribute(Attribute.LANGUAGE,
					    Locale.getDefault(), 0, text.length());

	    // set Clause and Reading Information
	    if (clauseBoundary!=null && clauseReading!=null &&
		clauseReading.length!=0 && clauseBoundary.length==clauseReading.length+1 &&
		clauseBoundary[0]==0 && clauseBoundary[clauseReading.length]==text.length() )
	    {
		for (int i=0; i<clauseBoundary.length-1; i++) {
		    attrStr.addAttribute(Attribute.INPUT_METHOD_SEGMENT,
					    new Annotation(null), clauseBoundary[i], clauseBoundary[i+1]);
		    attrStr.addAttribute(Attribute.READING,
					    new Annotation(clauseReading[i]), clauseBoundary[i], clauseBoundary[i+1]);
		}
	    } else {
		// if (clauseBoundary != null)
		//    System.out.println("Invalid clause information!");

		attrStr.addAttribute(Attribute.INPUT_METHOD_SEGMENT,
					new Annotation(null), 0, text.length());
		attrStr.addAttribute(Attribute.READING,
					new Annotation(new String("")), 0, text.length());
	    }

	    // set Hilight Information
	    if (attributeBoundary!=null && attributeValue!=null &&
		attributeValue.length!=0 && attributeBoundary.length==attributeValue.length+1 &&
		attributeBoundary[0]==0 && attributeBoundary[attributeValue.length]==text.length() )
	    {
		for (int i=0; i<attributeBoundary.length-1; i++) {
		    InputMethodHighlight highlight;
		    switch (attributeValue[i]) {
			case ATTR_INPUT:
			case ATTR_INPUT_ERROR:
			default:
			    highlight = InputMethodHighlight.UNSELECTED_RAW_TEXT_HIGHLIGHT;
			    break;
			case ATTR_TARGET_CONVERTED:
			    highlight = InputMethodHighlight.SELECTED_CONVERTED_TEXT_HIGHLIGHT;
			    break;
			case ATTR_CONVERTED:
			    highlight = InputMethodHighlight.UNSELECTED_CONVERTED_TEXT_HIGHLIGHT;
			    break;
			case ATTR_TARGET_NOTCONVERTED:
			    highlight = InputMethodHighlight.SELECTED_RAW_TEXT_HIGHLIGHT;
			    break;
		    }
		    attrStr.addAttribute(TextAttribute.INPUT_METHOD_HIGHLIGHT,
					 highlight,
					 attributeBoundary[i], attributeBoundary[i+1]);
		}
	    } else {
		// if (attributeBoundary != null)
		//    System.out.println("Invalid attribute information!");

		attrStr.addAttribute(TextAttribute.INPUT_METHOD_HIGHLIGHT,
			     InputMethodHighlight.UNSELECTED_CONVERTED_TEXT_HIGHLIGHT,
			     0, text.length());
	    }

	    // get iterator
	    iterator = attrStr.getIterator();

	}

        Component source = getClientComponent();
        if (source == null)
            return;

        InputMethodEvent event = new InputMethodEvent(source,
                                                      id, 
                                                      iterator, 
                                                      commitedTextLength,
                                                      TextHitInfo.leading(caretPos),
                                                      TextHitInfo.leading(visiblePos));
        WToolkit.postEvent(WToolkit.targetToAppContext(source), event);
    }

    public Point inquireCandidatePosition()
    {
	Component client = getClientComponent();

	if (client == null) {
	    return new Point(0,0);
	}

	if (haveActiveClient()) {
	    Rectangle rc = inputContext.getTextLocation(TextHitInfo.leading(0));
	    return new Point(rc.x, rc.y + rc.height);
	} else {
	    Point pt = client.getLocationOnScreen();
	    Dimension size = client.getSize();
	    return new Point(pt.x, pt.y + size.height);
	}
    }

    // java.awt.Toolkit#getNativeContainer() is not available
    //	from this package
    private WComponentPeer getNearestNativePeer(Component comp)
    {
	if (comp==null)	    return null;

	ComponentPeer peer = comp.getPeer();
	if (peer==null)	    return null;

	while (peer instanceof java.awt.peer.LightweightPeer) {
	    comp = comp.getParent();
	    if (comp==null) return null;
	    peer = comp.getPeer();
	    if (peer==null) return null;
	}

	if (peer instanceof WComponentPeer)
	    return (WComponentPeer)peer;
	else
	    return null;

    }

    /**
     * Performs mapping from an abstract input method highlight
     * attribute to the XIM style visual attributes. This mapping has
     * to agree with convertVisualFeedbackToHighlight.
     * @see sun.awt.im.InputMethodHighlightMapping
     */
    private final static class WHighlightMapping implements InputMethodHighlightMapping {
	WHighlightMapping() {
	}
	/**
	 * @see sun.awt.im.InputMethodHighlightMapping#mapHighlight
	 */
	public Map mapHighlight(InputMethodHighlight highlight) {
	    HashMap attributes;

	    attributes = (HashMap) mappingTable.get(highlight);
	    return (Map) attributes.clone();
	}
    }

    private native int createNativeContext();
    private native void destroyNativeContext(int context);
    private native void enableNativeIME(WComponentPeer peer, int context, boolean useNativeCompWindow);
    private native void disableNativeIME(WComponentPeer peer);
    private native void handleNativeIMEEvent(WComponentPeer peer, AWTEvent e);
    private native void endCompositionNative(int context, boolean flag);
    private native void setConversionStatus(int context, int cmode);
    private native int  getConversionStatus(int context);
    private native void setOpenStatus(int context, boolean flag);
    private native boolean getOpenStatus(int context);
    private native int getNativeIMELanguage();
}
