/*
 * @(#)WMouseDragGestureRecognizer.java	1.2 98/06/29
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.windows;

import java.awt.Toolkit;
import java.awt.Component;

import java.awt.Point;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DragSource;
import java.awt.dnd.MouseDragGestureRecognizer;
import java.awt.dnd.DragGestureListener;

import java.awt.event.InputEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;

import java.lang.reflect.*;

/**
 * <p>
 * This subclass of MouseDragGestureRecognizer defines a DragGestureRecognizer
 * for Mouse based gestures on Win32.
 * </p>
 *
 * @author Laurence P. G. Cable
 * @version 1.2
 *
 * @see java.awt.dnd.DragGestureListener
 * @see java.awt.dnd.DragGestureEvent
 * @see java.awt.dnd.DragSource
 */

class WMouseDragGestureRecognizer extends MouseDragGestureRecognizer {

    /*
     * constant for number of pixels hysterisis before drag is determined
     * to have started
     */

    protected static final int motionThreshold = getMotionThreshold();

    private static int getMotionThreshold() {
	try {
	    return ((Integer)Toolkit.getDefaultToolkit().getDesktopProperty("DnD.gestureMotionThreshold")).intValue();
	} catch (Exception e) {
	    return 5; // defualt
	}
    }

    /*
     * button masks
     */

    protected static final int ButtonMask = InputEvent.BUTTON1_MASK;

    /*
     * modifier masks
     */

    protected static final int ModMask    = InputEvent.SHIFT_MASK | InputEvent.CTRL_MASK;

    /**
     * construct a new WMouseDragGestureRecognizer
     *
     * @param ds  The DragSource for the Component c
     * @param c   The Component to observe
     * @param act The actions permitted for this Drag
     * @param dgl The DragGestureRecognizer to notify when a gesture is detected
     *
     */

    protected WMouseDragGestureRecognizer(DragSource ds, Component c, int act, DragGestureListener dgl) {
	super(ds, c, act, dgl);
    }

    /**
     * construct a new WMouseDragGestureRecognizer
     *
     * @param ds  The DragSource for the Component c
     * @param c   The Component to observe
     * @param act The actions permitted for this Drag
     */

    protected WMouseDragGestureRecognizer(DragSource ds, Component c, int act) {
	this(ds, c, act, null);
    }

    /**
     * construct a new WMouseDragGestureRecognizer
     *
     * @param ds  The DragSource for the Component c
     * @param c   The Component to observe
     */

    protected WMouseDragGestureRecognizer(DragSource ds, Component c) {
	this(ds, c, DnDConstants.ACTION_NONE);
    }

    /**
     * construct a new WMouseDragGestureRecognizer
     *
     * @param ds  The DragSource for the Component c
     */

    protected WMouseDragGestureRecognizer(DragSource ds) {
	this(ds, null);
    }

    /**
     * determnine the Drag gesture from the event
     */

    protected int mapDragOperationFromModifiers(MouseEvent e) {
	int mods = e.getModifiers();

	if ((mods & ~(ButtonMask | ModMask)) != 0) {
	    return DnDConstants.ACTION_NONE;
	}

	if ((mods & InputEvent.BUTTON1_MASK) != InputEvent.BUTTON1_MASK) {
	    return DnDConstants.ACTION_NONE;
	}

	if ((mods &= ModMask) == 0) {
	    return DnDConstants.ACTION_MOVE;
	} else if (mods == ModMask) {
	    return DnDConstants.ACTION_LINK;
	} else if (mods == InputEvent.CTRL_MASK) {
	    return DnDConstants.ACTION_COPY;
	} else {
	    return DnDConstants.ACTION_NONE;
	}
    }

    /**
     * Invoked when the mouse has been clicked on a component.
     */

    public void mouseClicked(MouseEvent e) {
	// do nothing
    }

    /**
     * Invoked when a mouse button has been pressed on a component.
     */

    public void mousePressed(MouseEvent e) {
	events.clear();

	if (mapDragOperationFromModifiers(e) != DnDConstants.ACTION_NONE) {
	    appendEvent(e);
	}
    }

    /**
     * Invoked when a mouse button has been released on a component.
     */

    public void mouseReleased(MouseEvent e) {
	events.clear();
    }

    /**
     * Invoked when the mouse enters a component.
     */

    public void mouseEntered(MouseEvent e) {
	events.clear();
    }

    /**
     * Invoked when the mouse exits a component.
     */

    public void mouseExited(MouseEvent e) {

	if (!events.isEmpty()) { // gesture pending
	    int dragAction = mapDragOperationFromModifiers(e);

	    if (dragAction != DnDConstants.ACTION_NONE) {
	        appendEvent(e);

	        fireDragGestureRecognized(dragAction, ((MouseEvent)getTriggerEvent()).getPoint());
	    } else
		events.clear();
	}
    }

    /**
     * Invoked when a mouse button is pressed on a component.
     */

    public void mouseDragged(MouseEvent e) {
	if (!events.isEmpty()) { // gesture pending
	    int dop = mapDragOperationFromModifiers(e);

	    if (dop == DnDConstants.ACTION_NONE) {
		return;
	    }

	    MouseEvent trigger = (MouseEvent)events.get(0);


	    Point      origin  = trigger.getPoint();
	    Point      current = e.getPoint();

	    int        dx      = Math.abs(origin.x - current.x);
	    int        dy      = Math.abs(origin.y - current.y);

	    if (dx > motionThreshold || dy > motionThreshold) {
	        fireDragGestureRecognized(dop, ((MouseEvent)getTriggerEvent()).getPoint());
	    } else
		appendEvent(e);
	}
    }

    /**
     * Invoked when the mouse button has been moved on a component
     * (with no buttons no down).
     */

    public void mouseMoved(MouseEvent e) {
	// do nothing
    }
}
