/*
 * @(#)WPrinterJob.java	1.11 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.windows;

import java.awt.Color;
import java.awt.Graphics2D;

import java.awt.image.BufferedImage;

import java.awt.print.Pageable;
import java.awt.print.PageFormat;
import java.awt.print.Printable;
import java.awt.print.PrinterJob;

import java.io.IOException;

import sun.java2d.PeekGraphics;
import sun.java2d.PeekMetrics;

/**
 * A class which initiates and executes a Win32 printer job.
 *
 * @version 1.0 07 Nov 1997
 * @author Richard Blanchard
 */
public class WPrinterJob extends sun.java2d.RasterPrinterJob {

 /* Class Constants */


/* Instance Variables */

    /**
     * This is the Window's Polygon fill rule which
     * Selects alternate mode (fills the area between odd-numbered
     * and even-numbered polygon sides on each scan line).
     * It must match the value in <WINGDI.h> It can be passed
     * to setPolyFillMode().
     */ 
    protected static final int POLYFILL_ALTERNATE = 1;

    /**
     * This is the Window's Polygon fill rule which
     * Selects winding mode which fills any region
     * with a nonzero winding value). It must match
     * the value in <WINGDI.h> It can be passed
     * to setPolyFillMode().
     */ 
    protected static final int POLYFILL_WINDING = 2;

    /**
     * The maximum value for a Window's color component
     * as passed to selectSolidBrush.
     */
    private static final int MAX_WCOLOR = 255;

    /**
     * The Windows device context we will print into.
     * This variable is set after the Print dialog
     * is okayed by the user. If the user cancels
     * the print dialog, then this variable is 0.
     */
    private int mPrintDC;

    private int mPrintXRes;

    private int mPrintYRes;

    /**
     * The last color set into the print device context or
     * <code>null</code> if no color has been set.
     */
    private Color mLastColor;

 /* Static Initializations */

    static {
	java.security.AccessController.doPrivileged(
		    new sun.security.action.LoadLibraryAction("awt"));
    }

 /* Constructors */

    public WPrinterJob()
    {
    }

 /* Destructor */

    protected void finalize()
    {
	deleteDC();
    }

/* Instance Methods */

    /**
     * Display a dialog to the user allowing the modification of a
     * PageFormat instance.
     * The <code>page</code> argument is used to initialize controls
     * in the page setup dialog.
     * If the user cancels the dialog, then the method returns the
     * original <code>page</code> object unmodified.
     * If the user okays the dialog then the method returns a new
     * PageFormat object with the indicated changes.
     * In either case the original <code>page</code> object will
     * not be modified.
     * @param     page    the default PageFormat presented to the user
     *                    for modification
     * @return    the original <code>page</code> object if the dialog
     *            is cancelled, or a new PageFormat object containing
     *            the format indicated by the user if the dialog is
     *            acknowledged
     * @since     JDK1.2
     */
    public PageFormat pageDialog(PageFormat page) {
        PageFormat pageClone = (PageFormat) page.clone();
	boolean doIt = pageSetup(pageClone, null);
        return doIt ? pageClone : page;
    }

    /**
     * Presents the user a dialog for changing properties of the
     * print job interactively.
     * @returns false if the user cancels the dialog and
     *          true otherwise.
     */
    public boolean printDialog() {
	return jobSetup(getPageable());
    }

    /**
     * The passed in PageFormat will be altered to describe
     * the default page size and orientation of the PrinterJob's
     * current printer.
     *  01/12/98 Note: Needs implementation.
     */
    public PageFormat defaultPage(PageFormat page) {
	return page;
    }

    /**
     * The passed in PageFormat is altered to be usable on
     * the PrinterJob's current printer.
     *  01/12/98 Note: Needs implementation.
     */
    public PageFormat validatePage(PageFormat page) {
	return page;
    }

    /**
     * Examine the metrics captured by the
     * <code>PeekGraphics</code> instance and
     * if capable of directly converting this
     * print job to the printer's control language
     * or the native OS's graphics primitives, then
     * return a <code>PathGraphics</code> to perform
     * that conversion. If there is not an object
     * capable of the conversion then return
     * <code>null</code>. Returning <code>null</code>
     * causes the print job to be rasterized.
     */
    
    protected Graphics2D createPathGraphics(PeekGraphics peekGraphics,
					    PrinterJob printerJob) {

	WPathGraphics pathGraphics;
	PeekMetrics metrics = peekGraphics.getMetrics();

	/* If the application has drawn anything that
	 * out PathGraphics class can not handle then
	 * return a null PathGraphics.
	 */
	if (metrics.hasNonSolidColors()
			|| metrics.hasCompositing()
			|| metrics.hasImages()) {

	    pathGraphics = null;
	} else {

	    BufferedImage bufferedImage = new BufferedImage(8, 8,
					    BufferedImage.TYPE_INT_RGB);
	    Graphics2D bufferedGraphics = bufferedImage.createGraphics();

	    pathGraphics =  new WPathGraphics(bufferedGraphics, printerJob);
	}

	return pathGraphics;
    }
    

    protected double getXRes() {
	return mPrintXRes;
    }

    protected double getYRes() {
	return mPrintYRes;
    }

    /**
     * Returns how many times each page in the book
     * should be consecutively printed by PrinterJob.
     * Because the underlying Window's driver will
     * generate the copies, rather than having RasterPrinterJob
     * iterate over the number of copies, this method always returns
     * 1.
     */
    protected int getNoncollatedCopies() {
	return 1;
    }

    /**
     * Return the Window's device context that we are printing
     * into.
     */
    protected int getPrintDC() {
	return mPrintDC;
    }

    protected void beginPath() {
	beginPath(mPrintDC);
    }

    protected void endPath() {
	endPath(mPrintDC);
    }

    protected void closeFigure() {
	closeFigure(mPrintDC);
    }

    protected void fillPath() {
	fillPath(mPrintDC);
    }

    protected void moveTo(float x, float y) {
	moveTo(mPrintDC, x, y);
    }

    protected void lineTo(float x, float y) {
	lineTo(mPrintDC, x, y);
    }

    protected void polyBezierTo(float control1x, float control1y,
			        float control2x, float control2y,
				float endX, float endY) {

	polyBezierTo(mPrintDC, control1x, control1y,
			       control2x, control2y,
			       endX, endY);
    }

    /**
     * Set the current polgon fill rule into the printer device context.
     * The <code>fillRule</code> should
     * be one of the following Windows constants:
     * <code>ALTERNATE</code> or <code>WINDING</code>.
     */
    protected void setPolyFillMode(int fillRule) {
	setPolyFillMode(mPrintDC, fillRule);
    }

    /*
     * Create a Window's solid brush for the color specified
     * by <code>(red, green, blue)</code>. Once the brush
     * is created, select it in the current printing device
     * context and free the old brush.
     */
    protected void selectSolidBrush(Color color) {
	
	/* We only need to select a brush if the color has changed.
	*/
	if (color.equals(mLastColor) == false) {
	    mLastColor = color;
	    float[] rgb = color.getColorComponents(null);

	    selectSolidBrush(mPrintDC, (int) (rgb[0] * MAX_WCOLOR),
				       (int) (rgb[1] * MAX_WCOLOR),
				       (int) (rgb[2] * MAX_WCOLOR));
	}
    }

    /**
     * Return the x coordinate of the current pen
     * position in the print device context.
     */
    protected int getPenX() {

	return getPenX(mPrintDC);
    }


    /**
     * Return the y coordinate of the current pen
     * position in the print device context.
     */
    protected int getPenY() {

	return getPenY(mPrintDC);
    }

    /**
     * Set the current path in the printer device's
     * context to be clipping path.
     */
    protected void selectClipPath() {
	selectClipPath(mPrintDC);
    }

 /* Native Methods */

    /**
     * Displays the page setup dialog placing the user's
     * settings into 'page'.
     */
    public static native boolean pageSetup(PageFormat page,
					   Printable painter);

    /**
     * Displays the print dialog and records the user's settings
     * into this object. Return false if the user cancels the
     * dialog.
     */
    private native boolean jobSetup(Pageable doc);

    /**
     * Call Window's StartDoc routine to begin a
     * print job. The DC from the print dialog is
     * used. If the print dialog was not displayed
     * then a DC for the default pitner is created.
     */
    protected native void startDoc();

    /**
     * Call Window's EndDoc routine to end a
     * print job.
     */
    protected native void endDoc();

    /**
     * Call Window's deleteDC routine to end a
     * print job.
     */
    protected native void deleteDC();

    /**
     * Begin a new page. This call's Window's
     * StartPage routine.
     */
    protected native void startPage(PageFormat format, Printable painter,
                                    int index);
    /**
     * End a page. This call's Window's EndPage
     * routine.
     */
    protected native void endPage(PageFormat format, Printable painter,
                                  int index);

    /**
     * Prints the contents of the array of ints, 'data'
     * to the current page. The band is placed at the
     * location (x, y) in device coordinates on the
     * page. The width and height of the band is
     * specified by the caller.
     */
    protected native void printBand(int[] data, int x, int y,
				    int width, int height);

    /**
     * Begin a Window's rendering path in the device
     * context <code>printDC</code>.
     */
    protected native void beginPath(int printDC);

    /**
     * End a Window's rendering path in the device
     * context <code>printDC</code>.
     */
    protected native void endPath(int printDC);

    /**
     * Close a subpath in a Window's rendering path in the device
     * context <code>printDC</code>.
     */
    protected native void closeFigure(int printDC);

    /**
     * Fill a defined Window's rendering path in the device
     * context <code>printDC</code>.
     */
    protected native void fillPath(int printDC);

    /**
     * Move the Window's pen position to <code>(x,y)</code>
     * in the device context <code>printDC</code>.
     */
    protected native void moveTo(int printDC, float x, float y);

    /**
     * Draw a line from the current pen position to
     * <code>(x,y)</code> in the device context <code>printDC</code>.
     */
    protected native void lineTo(int printDC, float x, float y);
    
    protected native void polyBezierTo(int printDC,
				       float control1x, float control1y,
			               float control2x, float control2y,
				       float endX, float endY);

    /**
     * Set the current polgon fill rule into the device context
     * <code>printDC</code>. The <code>fillRule</code> should
     * be one of the following Windows constants:
     * <code>ALTERNATE</code> or <code>WINDING</code>.
     */
    protected native void setPolyFillMode(int printDC, int fillRule);

    /**
     * Create a Window's solid brush for the color specified
     * by <code>(red, green, blue)</code>. Once the brush
     * is created, select it in the device
     * context <code>printDC</code> and free the old brush.
     */
    protected native void selectSolidBrush(int printDC,
					   int red, int green, int blue);

    /**
     * Return the x coordinate of the current pen
     * position in the device context
     * <code>printDC</code>.
     */
    protected native int getPenX(int printDC);

    /**
     * Return the y coordinate of the current pen
     * position in the device context
     * <code>printDC</code>.
     */
    protected native int getPenY(int printDC);

    /**
     * Select the device context's current path
     * to be the clipping path.
     */
    protected native void selectClipPath(int printDC);

}
