/*
 * @(#)monitor_cache_md.c	1.12 98/08/21
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#ifdef FAST_MONITOR

/*
 * Code to manage the monitor cache(pool).
 *
 * The monitor pool uses a special link list data structure. Once the 
 * cache has expanded, it is never contracted, although doing
 * so would be possible.
 */

#include "threads.h"
#include "monitor.h"
#include "monitor_cache.h"
#include "common_exceptions.h"
#include "sys_api.h"

sys_mon_t *_cache_lock;

/* Exported symbol that signals fast monitor implementation */
JNIEXPORT void FastMonitor()
{
}

/*
 * NOTES: The follow must be an exact copy of monitor_md.h
 * The entire code assumes the knowledge of sys_mon_t implementation.
 */
typedef struct sys_mon {
    long            atomic_count;   /* Variable for atomic compare swap */
    void           *semaphore;      /* Semaphore used for the contention */
    sys_thread_t   *monitor_owner;  /* Current owner of this monitor */
    long            entry_count;    /* Recursion depth */
    sys_thread_t   *monitor_waiter; /* Monitor waiting queue head */
    long            waiter_count;   /* For debugging purpose */
} sys_mon_t;

/*
 * Internal macro used for convenience.
 * These macros reflect the monitor design. An extra word is allocated
 * in each object. The word has 3 states:
 * a) 0, no synchronization happens on this monitor;
 * b) ee | [0, 1, 2, 3], assume the ee is at least 8-byte aligned.
 *      the state indicates the *ee* is the current owner of the monitor,
 *      and the monitor has been entered 1, 2, 3 or 4 times.
 * c) mid | 4, assume the mid (the heavy weight monitor pointer) is
 *      8-byte aligned. This is a masked heavy weight monitor pointer.
 *      All synchronization operations will apply to the heavy weight monitor.
 */
#define monitorSize()   (sizeof(monitor_t) + sizeof(sys_mon_t))
#define key2mon(key)    ((sys_mon_t**) (((JHandle*) key)->obj))
#define same(mid, ee)   (((int32_t) mid & 0xfffffffc) == (int32_t) ee)
#define mask(mid)       ((sys_mon_t*) (((int32_t) mid) | 4))
#define unmask(mid)     ((sys_mon_t*) (((int32_t) mid) & 0xfffffffB))
#define next_mon(m)     ((monitor_t*) ((char *) m + monitorSize()))
#define sysmon2monitor(m) (((monitor_t*) m) - 1)
#define mon_owner(mid)  EE2SysThread((struct execenv *) (0xfffffffc & (int32_t) mid))
#define mon_entry(mid)  (1 + (3 & ((int32_t) mid)))
#define mon_align(p)    ((monitor_t*) ((7 + (long) p) & ~(8 - 1)))

#define DEFAULT_MONITOR_POOL_SIZE 128

static monitor_t *freeList = 0;
static monitor_t *blockList = 0;

static int (*MonitorEnter)(sys_thread_t *self, sys_mon_t *mid);
static int (*MonitorExit)(sys_thread_t *self, sys_mon_t *mid);

static int systemIsMP = 0;
static int monCount = 0;
static int useCount = 0;

/*
 * NOTES: Internal use only. No need to hold mutex. Declare it as
 * extern is just to keep the current header file. It can only be
 * called from monitorCacheInit() and monitorCacheAlloc().
 */

/*
 * NOTES: The monitor cache/pool uses a special data structure.
 * It has two levels of link list, block and node. Each block
 * has DEFAULT_MONITOR_POOL_SIZE's monitors. The first monitor
 * of each block will be used to link all blocks, which is the
 * blockList. The rest of monitors will be linked as the freeList.
 * It is more effecient for memory allocation.
 */
void monitorCacheExpand(int count)
{
    /* Allocate DEFAULT_MONITOR_POOL_SIZE number of sys_mon_t's */
    monitor_t *begin = sysMalloc(DEFAULT_MONITOR_POOL_SIZE * monitorSize() + 8);
    monitor_t *first = mon_align(begin);
    monitor_t *m = first;
    int i;

    if (begin == 0) {
        out_of_memory();
        return;
    }

    /* NOTES: Initialize the monitor_t list. See monitorInit(). */
    for (i = 0; i < DEFAULT_MONITOR_POOL_SIZE - 1; i++, m = next_mon(m)) {
        monitorInit(m);
    }

    /* Initialize the last monitor_t, and link it to the freeList. */
    monitorInit(m);
    m->next = freeList;

    /* Skip the first monitor structure */
    freeList = next_mon(first);
    monCount += DEFAULT_MONITOR_POOL_SIZE - 1;

    /* Link all blocks using the first sys_mon_t structure */
    first->next = blockList;
    blockList = first;

    /*
     * This is the hack to keep track the malloc pointer.
     * Be careful, the first monitor should never be considered
     * as a monitor. It is only used for maintained the data structure.
     */
    sysmon(first)->monitor_owner = (sys_thread_t *) begin;
}

/*
 * Assembly routine for atomic compare swap. It works only on Intel 486 or
 * later processor. Using __fastcall for performance reason. First
 * optimization need to consider inline this function.
 */
__declspec(naked) static void* __fastcall 
AtomicSwap(void** destination, void* exchange, void* comperand)
{
    __asm
    {
        mov eax, dword ptr [esp+4];
        lock cmpxchg dword ptr [ecx], edx;
        ret 4;
    }
}

/*
 * Initialize the global monitor cache.
 */
void monitorCacheInit(void)
{
    systemIsMP = sysGetSysInfo()->isMP ? 0xffffffff : 0;
    MonitorEnter = hpi_thread_interface->MonitorEnter;
    MonitorExit  = hpi_thread_interface->MonitorExit;

    CACHE_LOCK_INIT();
    monitorCacheExpand(0);
}

/*
 * Destroy the global monitor cache. It is currently not called.
 */
void monitorCacheDestroy(void)
{
    int i;
    sys_thread_t *self = sysThreadSelf();

    CACHE_LOCK(self);

    while (blockList != 0) {
        monitor_t * m = blockList;
        monitor_t * begin = m;
        blockList = blockList->next;

        for (i = 0; i < DEFAULT_MONITOR_POOL_SIZE; i++, m = next_mon(m)) {
            sysMonitorDestroy(sysmon(m));
        }
        sysFree(sysmon(begin)->monitor_owner);
    }

    freeList = 0;
    blockList = 0;

    CACHE_UNLOCK(self);
}

/*
 * Allocate a monitor from the cache/pool.
 */
sys_mon_t* monitorCacheAlloc(void)
{
    monitor_t *mon;
    sys_mon_t *mid;
    sys_thread_t *self = sysThreadSelf();

    CACHE_LOCK(self);

    if (!freeList) {
        monitorCacheExpand(0);
    }

    mon = freeList;
    freeList = freeList->next;
    useCount++;

    CACHE_UNLOCK(self);

    mon->next = 0;
    mon->key  = 0;

    mid = sysmon(mon);
    mid->atomic_count = 0;
    mid->monitor_waiter = 0;

    return mid;
}

/*
 * Return a monitor to the cache/pool. It is current only used
 * by thread exiting code to free its spare monitor.
 */
void monitorCacheFree(void * m)
{
    sys_mon_t *mid = (sys_mon_t*) m;
    monitor_t *mon = sysmon2monitor(m);
    sys_thread_t *self = sysThreadSelf();

    sysAssert(m != 0);

    mid->atomic_count = -1;
    mid->monitor_owner = 0;
    mid->entry_count = 0;
    mid->monitor_waiter = 0;
    mid->waiter_count = 0;
    mon->key = 0;

    CACHE_LOCK(self);

    mon->next = freeList;
    freeList = mon;
    useCount--;

    CACHE_UNLOCK(self);
}

/*
 * GC the monitor associated with object key. It is only called
 * by freeSweep routine from GC.
 */
void monitorCacheGC(uintptr_t key)
{
    JHandle *hp = (JHandle*) key;
    sys_mon_t *mid = (sys_mon_t*) *(hp->obj - 1);
    monitor_t *mon;

    sysAssert(is_monitor(mid));

    mid = unmask(mid);
    mon = sysmon2monitor(mid);

    sysAssert(mon->key == (uintptr_t) hp);

    /* If it is not being used, free it. */
    if (!sysMonitorInUse(mid)) {
        *(hp->obj - 1) = 0;

        mon->key = 0;
        mon->next = freeList;
        freeList = mon;

        useCount--;
    }
}

/**
 * In the following code and the code in monitor_md.c, we use
 * struct execenv *ee and sys_thread_t *self as thread id. It
 * can provide more debugging information etc. But when there
 * is real need for speed, and ee is not available, we can
 * change it to use mov eax, dword ptr FS:[18H] on WinNT, ?Win95.
 *
 * The following code will do the light weight monitor to heavy
 * weight monitor conversion using atomic swap. The thread that
 * do the conversion could be the owning thread (during the 
 * recursion) or another thread (during the contention). Anyway,
 * it must set the heavy weight monitor owner to the current owner
 * id before it swap it into the monitor word in the object.
 * The pmid is the current address of monitor word. It is used
 * to pin the object during the conversion.
 */
__declspec(dllexport) void __fastcall
monitorEnterImpl(struct execenv *ee, sys_mon_t** const pmid, JHandle *hp)
{
    /* Prepare a monitor for monitor conversion */
    sys_thread_t *self = EE2SysThread(ee);
    sys_mon_t * m = ee->spareMonitor ? ee->spareMonitor :
                    (ee->spareMonitor = monitorCacheAlloc());

    sysAssert(m->atomic_count == 0);
    sysAssert(m->monitor_waiter == 0);

    sysmon2monitor(m)->key = (uintptr_t) hp;

    for (;;) {
        /* monitor state may be changed, re-read the monitor word */
        sys_mon_t * volatile mid = *(pmid - 1); /* atomic point */

        if (mid == 0) {
            /*
             * There is no owner for this monitor (it must be the
             * owner thread of the monitor just exited. We simply
             * swap in a heavy weight monitor with current
             * thread is the owner, and set entry_count to 1.
             */
            m->monitor_owner = self;
            m->entry_count = 1;
            if (0 == AtomicSwap(pmid - 1, mask(m), 0)) {
                ee->spareMonitor = 0;
                return;
            }
        } else if (!is_monitor(mid)) {
            /*
             * This is a light weight monitor, and some thread
             * (it could be the current thread) owns this light
             * weight monitor. We prepare a heavy weight monitor
             * with the current owner and its recursion entry 
             * count, and atomically swap it into the monitor word.
             * When succeed, we do the normal heavy weight monitor
             * enter, otherwise we do the loop and try again.
             */
            m->monitor_owner = mon_owner(mid);
            m->entry_count = mon_entry(mid);
            if (mid == AtomicSwap(pmid - 1, mask(m), mid)) {
                ee->spareMonitor = 0;
                sysMonitorEnter(self, m);
                return;
            }
        } else /* is_monitor(mid) */ {
            /*
             * If there is a heavy weight monitor (it must be some
             * other thread just swap it in), we just use it.
             * The ee->spareMonitor can be used for next time.
             * So we don't need to free it for now.
             */
            sysMonitorEnter(self, unmask(mid));
            return;
        }
    } /* end for(;;) */
}


/*
 * This is the fastest monitor routine. With proper inlinine, the
 * actual code size could vary from 2 to 8 instructions, depending
 * the current register state. The basic idea is atomic compare 0
 * swap the monitor word with current thread id, (ee for now).
 */
__declspec(naked) sys_mon_t * __cdecl
monitorEnter2(struct execenv *ee, uintptr_t key)
{
    __asm
    {
        /* Load parameters into registers */
        mov  edx, dword ptr [esp+8];        // load handle/key;
        mov  ecx, dword ptr [esp+4];        // load ee
        mov  edx, dword ptr [edx];          // load handle->obj

        mov  eax, dword ptr [edx-4];        // load current monitor word

        /* At this point, mid in eax, ee in ecx, hp->obj in edx */

        test eax, eax;                      // test if it is zero
        jnz  SLOW_PATH;                     // zero means no owner yet

FAST_PATH:
        test edx, dword ptr [systemIsMP];   // treat MP differently
        jnz  MPM;

        cmpxchg dword ptr [edx-4], ecx;     // atomic swap
        jne  CONTENTION;                    // contention if not equal
        ret;

MPM:
        lock cmpxchg dword ptr [edx-4], ecx;// atomic swap
        jne  CONTENTION;                    // contention if not equal
        ret;

SLOW_PATH:
        test eax, 4;                        // test is_monitor(mid);
        jnz  HEAVY_PATH;                    // true if not zero

        /*
         * At this point, it is not a heavy weight monitor,
         * but a light weight monitor owned by some thread.
         * We need to detect if the current thread is owner
         * and recursion level is less than 4, i.e., the
         * ee <= mid && mid < ee + 3.
         */

        cmp  eax, ecx;                      // test mid < ee
        jb   CONTENTION;                    // contention if true

        add  ecx, 2;
        cmp  eax, ecx;                      // test mid > ee + 2
        ja   CONTENTION;                    // contention if true

        lea  ecx, byte ptr [eax+1];         // mid = mid + 1
        jmp  FAST_PATH;                     // reuse the same code

HEAVY_PATH:
        sub  eax, 4;                        // unmask(mid)
        add  ecx, 80H;                      // EE2SysThread(ee)
        push eax;                           // push mid
        push ecx;                           // push self
        call dword ptr [MonitorEnter];      // call sysMonitorEnter
        add  esp, 8;
        ret;

CONTENTION:
        mov  eax, dword ptr [esp+8];        // hp->obj in edx
        mov  ecx, dword ptr [esp+4];        // reload ee
        push eax;
        call monitorEnterImpl;              // hp on the stack
        ret;
    }
}

__declspec(dllexport) int __fastcall
monitorExitException(struct execenv *ee)
{
    ThrowIllegalMonitorStateException(ee, "current thread not owner");
    return SYS_ERR;
}

__declspec(naked) int __cdecl
monitorExit2(struct execenv *ee, uintptr_t key)
{
    __asm
    {
        /* Load parameters into registers */
        mov  ecx, dword ptr [esp+8];        // load JHandle/key;
        mov  ecx, dword ptr [ecx];          // load handle->obj
        mov  eax, dword ptr [ecx-4];        // load mid

        cmp  eax, dword ptr [esp+4];        // mid == ee
        jne  SLOW_PATH;

        /* Fast case: (mid, ee) in eax, hp->obj in ecx */
        xor  edx, edx;                      // zero edx

FAST_PATH:
        test ecx, dword ptr [systemIsMP];
        jnz  MPM;

        cmpxchg dword ptr [ecx-4], edx;     // atomic swap
        jne  SLOW_PATH;                     // goto slow case
        xor  eax, eax;                      // return SYS_OK;
        ret;

MPM:
        lock cmpxchg dword ptr [ecx-4], edx;// atomic swap
        jne  SLOW_PATH;                     // goto slow case
        xor  eax, eax;                      // return SYS_OK;
        ret;

SLOW_PATH:
        /* At this point, mid in eax, hp->obj in ecx */
        mov  edx, dword ptr [esp+4];        // load ee
        test eax, 4;                        // test is_monitor(mid)
        jnz  HEAVY_PATH;                    // heavy weight if true

        /* This is a light weight monitor owned by a thread. */

        cmp  eax, edx;                      // test mid < ee
        jb   ERROR_PATH;                    // error if true

        add  edx, 3;
        cmp  eax, edx;                      // test mid > ee + 3
        ja   ERROR_PATH;                    // error if true

        lea  edx, byte ptr [eax-1];         // mid = mid - 1
        jmp  FAST_PATH;                     // use the same code

HEAVY_PATH:
        sub  eax, 4;
        add  edx, 80H;

        push eax;                           // push mid
        push edx;                           // push self
        call dword ptr [MonitorExit];       // call sysMonitorExit
        add  esp, 8;

        test eax, eax;                      // test error
        jnz  ERROR_PATH;                    // throw exception when needed
        ret;

ERROR_PATH:
        mov  ecx, dword ptr [esp+4];        // load ee
        call monitorExitException;
        ret;
    }
}

/*
 * Thread waits (blocks) on this monitor until another thread 
 * notifies the monitor, or the thread is interrupted.
 */
void
monitorWait2(struct execenv *ee, uintptr_t key, int64_t millis)
{
    sys_mon_t** const pmid = key2mon(key);  /* pinning object */
    sys_thread_t *self = EE2SysThread(ee);  /* current thread */
    int ret;

    for (;;) {
        sys_mon_t * const mid = *(pmid - 1);/* atomic point */
        if (is_monitor(mid)) {
            *(int *) &pmid = 0;
            ret = sysMonitorWait(self, unmask(mid), millis);
            break;
        } else if (mid) {
            sys_mon_t* m = ee->spareMonitor ?  ee->spareMonitor
                : (ee->spareMonitor = monitorCacheAlloc());

            sysmon2monitor(m)->key = key;
            m->monitor_owner = mon_owner(mid);
            m->entry_count = mon_entry(mid);

            sysAssert(!is_monitor(mid));
            sysAssert(m->entry_count > 0);

            if (mid == AtomicSwap(pmid - 1, mask(m), mid)) {
                ee->spareMonitor = 0;
                *(int *) &pmid = 0;
                ret = sysMonitorWait(self, m, millis);
                break;
            }
        } else {
            ret = SYS_ERR;
            break;
        }
    }

    /* An asynchronous exception might have been thrown while we
     * wait on the monitor.
     * REMIND: 
     * This still does not fix the inherent race condition
     * where an async exception can be thrown after this check
     * is done. However, this race condition exists everywhere
     * else in the VM and should be fixed thoroughly in a future
     * release. 
     */
    if (exceptionOccurred(ee))
        return;

    if (ret == SYS_INTRPT) {
        ThrowInterruptedException(ee, "operation interrupted");
    } else if (ret != SYS_OK) {
        ThrowIllegalMonitorStateException(ee, "current thread not owner");
    }
}

void
monitorNotify2(struct execenv *ee, uintptr_t key)
{
    sys_mon_t *mid = *(key2mon(key) - 1);       /* atomic point */

    if (same(mid, ee)) {
        return;
    }
    
    if (is_monitor(mid)) {
        sys_thread_t *self = EE2SysThread(ee);
        mid = unmask(mid);
        /*
         * If the current thread does not own the monitor, the monitor
         * might be GCed in this time window and reallocated to other
         * object. We have to check its current key.
         */
        if (sysMonitorEntered(self, mid) &&
            sysmon2monitor(mid)->key == key) {
            sysMonitorNotify(self, mid);
            return;
        }
    }

    ThrowIllegalMonitorStateException(ee, "current thread not owner");
}

void
monitorNotifyAll2(struct execenv *ee, uintptr_t key)
{
    sys_mon_t *mid = *(key2mon(key) - 1);       /* atomic point */

    if (same(mid, ee)) {
        return;
    }
    
    if (is_monitor(mid)) {
        sys_thread_t *self = EE2SysThread(ee);
        mid = unmask(mid);                      /* unmask monitor */
        /*
         * If the current thread does not own the monitor, the monitor
         * might be GCed in this time window and reallocated to other
         * object. We have to check its current key.
         */
        if (sysMonitorEntered(self, mid) && 
            sysmon2monitor(mid)->key == key) {
            sysMonitorNotifyAll(self, mid);
            return;
        }
    }

    ThrowIllegalMonitorStateException(ee, "current thread not owner");
}

/*
 * The monitor enumeration may sometimes dump the spare monitor in the
 * *ee*. Since the spare monitor is important for the entire algorithm,
 * we will keep the situation.
 *
 * This function must be called while hold the monitor cache lock.
 * The enumeration is done block by block. See the above note regarding
 * the data structure being used here.
 */
void monitorEnumerate(void (*func)(monitor_t *, void *), void *p)
{
    monitor_t *block = blockList;

    while (block != 0) {
        monitor_t * m = next_mon(block);
        int i = 0;
        for (; i < DEFAULT_MONITOR_POOL_SIZE - 1; i++, m = next_mon(m)) {
            if (m->key != 0) {
                func(m, p);
            }
        }
        block = block->next;
    }
}

/*
 * Initialize the monitor_t. It has no key for now.
 * By default, the next pointer points to the monitor following it.
 */
void monitorInit(monitor_t * m)
{
    m->key = 0;
    m->next = next_mon(m);
    sysMonitorInit(sysmon(m));
}

void *
lookupJavaMonitor(sys_mon_t *mid)
{
    return NULL;
}

#endif
