/*
 * @(#)canonicalize_md.c	1.3 98/08/24
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Pathname canonicalization for Win32 file systems
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include <sys/stat.h>

#include <windows.h>
#include <winbase.h>
#include <errno.h>

#undef DEBUG_PATH		/* Define this to debug path code */

#define isfilesep(c) ((c) == '/' || (c) == '\\')
#define islb(c)      (IsDBCSLeadByte((BYTE)(c)))


/* Copy bytes to dst, not going past dend; return dst + number of bytes copied,
   or NULL if dend would have been exceeded.  If first != '\0', copy that byte
   before copying bytes from src to send - 1. */

static char *
cp(char *dst, char *dend, char first, char *src, char *send)
{
    char *p = src, *q = dst;
    if (first != '\0') {
	if (q < dend) {
	    *q++ = first;
	} else {
	    errno = ENAMETOOLONG;
	    return NULL;
	}
    }
    if (send - p > dend - q) {
	errno = ENAMETOOLONG;
	return NULL;
    }
    while (p < send) {
	*q++ = *p++;
    }
    return q;
}


/* Find first instance of '\\' at or following start.  Return the address of
   that byte or the address of the null terminator if '\\' is not found. */

static char *
nextsep(char *start)
{
    char *p = start;
    int c;
    while ((c = *p) && (c != '\\')) {
	p += ((islb(c) && p[1]) ? 2 : 1);
    }
    return p;
}


/* Tell whether the given string contains any wildcard characters */

static int
wild(char *start)
{
    char *p = start;
    int c;
    while (c = *p) {
	if ((c == '*') || (c == '?')) return 1;
	p += ((islb(c) && p[1]) ? 2 : 1);
    }
    return 0;
}


/* Convert a pathname to canonical form.  The input orig_path is assumed to
   have been converted to native form already, via JVM_NativePath().  This is
   necessary because _fullpath() rejects duplicate separator characters on
   Win95, though it accepts them on NT. */

int 
canonicalize(char *orig_path, char *result, int size)
{
    WIN32_FIND_DATA fd;
    HANDLE h;
    char path[1024];	/* Working copy of path */
    char *src, *dst, *dend;

    /* Reject paths that contain wildcards */
    if (wild(orig_path)) {
	errno = EINVAL;
	return -1;
    }

    /* Collapse instances of "foo\.." and ensure absoluteness.  Note that
       contrary to the documentation, the _fullpath procedure does not require
       the drive to be available.  It also does not reliably change all
       occurrences of '/' to '\\' on Win95, so now JVM_NativePath does that. */
    if(!_fullpath(path, orig_path, sizeof(path))) {
	return -1;
    }

    /* Correction for Win95: _fullpath may leave a trailing "\\"
       on a UNC pathname */
    if ((path[0] == '\\') && (path[1] == '\\')) {
	char *p = path + strlen(path);
	if ((p[-1] == '\\') && !islb(p[-2])) {
	    p[-1] = '\0';
	}
    }

    src = path;			/* Start scanning here */
    dst = result;		/* Place results here */
    dend = dst + size;		/* Don't go to or past here */

    /* Copy prefix, assuming path is absolute */
    if (isalpha(src[0]) && (src[1] == ':') && (src[2] == '\\')) {
	/* Drive specifier */
	*src = toupper(*src);	/* Canonicalize drive letter */
	if (!(dst = cp(dst, dend, '\0', src, src + 2))) {
	    return -1;
	}
	src += 2;
    } else if ((src[0] == '\\') && (src[1] == '\\')) {
	/* UNC pathname */
	char *p;
	p = nextsep(src + 2);	/* Skip past host name */
	if (!*p) {
	    /* A UNC pathname must begin with "\\\\host\\share",
	       so reject this path as invalid if there is no share name */
	    errno = EINVAL;
	    return -1;
	}
	p = nextsep(p + 1);	/* Skip past share name */
	if (!(dst = cp(dst, dend, '\0', src, p))) {
	    return -1;
	}
	src = p;
    } else {
	/* Invalid path */
	errno = EINVAL;
	return -1;
    }

    /* At this point we have copied either a drive specifier ("z:") or a UNC
       prefix ("\\\\host\\share") to the result buffer, and src points to the
       first byte of the remainder of the path.  We now scan through the rest
       of the path, looking up each prefix in order to find the true name of
       the last element of each prefix, thereby computing the full true name of
       the original path. */
    while (*src) {
	char *p = nextsep(src + 1);	/* Find next separator */
	char c = *p;
	assert(*src == '\\');		/* Invariant */
	*p = '\0';			/* Temporarily clear separator */
	h = FindFirstFile(path, &fd);	/* Look up prefix */
	*p = c;				/* Restore separator */
	if (h != INVALID_HANDLE_VALUE) {
	    /* Lookup succeeded; append true name to result and continue */
	    FindClose(h);
	    if (!(dst = cp(dst, dend, '\\',
			   fd.cFileName,
			   fd.cFileName + strlen(fd.cFileName)))) {
		return -1;
	    }
	    src = p;
	    continue;
	} else {
	    /* If the lookup of a particular prefix fails because the file does
	       not exist, because it is of the wrong type, or because access is
	       denied, then we copy the remainder of the input path as-is.
	       Other I/O errors cause an error return. */
	    DWORD errval = GetLastError();
	    if ((errval == ERROR_FILE_NOT_FOUND)
		|| (errval == ERROR_DIRECTORY)
		|| (errval == ERROR_PATH_NOT_FOUND)
		|| (errval == ERROR_BAD_NETPATH)
		|| (errval == ERROR_BAD_NET_NAME)
		|| (errval == ERROR_ACCESS_DENIED)
		|| (errval == ERROR_NETWORK_ACCESS_DENIED)) {
		if (!(dst = cp(dst, dend, '\0', src, src + strlen(src)))) {
		    return -1;
		}
		break;
	    } else {
#ifdef DEBUG_PATH
		jio_fprintf(stderr, "canonicalize: errval %d\n", errval);
#endif DEBUG_PATH
		return -1;
	    }
	}
    }

    if (dst >= dend) {
	errno = ENAMETOOLONG;
	return -1;
    }
    *dst = '\0';
    return 0;

}
