/*
 * @(#)HAE_API_WinOS.c	1.8 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*****************************************************************************/
/*
**	HAE_API_WinOS.c
**
**	This provides platform specfic functions for Windows 95/NT. This interface
**	for HAE is for Windows 95/NT and uses the waveOut API or DirectSound to send
**	buffer slices through the multimedia system.
**
**	Overview:
**		This works by creating another thread, aquires the sound card via
**		the WinOS waveOutOpen call. Then allocating enough buffers, preps then,
**		and build buffers depending upon how much data has been written to the
**		audio device. The thread polls the current position in the audio stream
**		that has been written and when it falls below a buffer or two, more
**		are built in that thread.
**
**	Written by Steve Hales
**
**	History	-
**	7/14/97		Created
**	7/18/97		Added Win32 file i/o functions rather than odd unix style ones
**	9/2/97		Fixed typo that referenced samples rather than bytes
**	9/3/97		Added support to detect stereo or 16 bit support
**	11/11/97	Added support for DirectSound.
**				Added HAE_MaxDevices & HAE_SetDeviceID & HAE_GetDeviceID & HAE_GetDeviceName
**	11/20/97	Modifed DirectSound implementation to handle differences for NT DirectSound
**	12/8/97		Added HAE_GetDeviceSamplesPlayedPosition
**
**	04.16.98	$$kk: added NO_DIRECT_SOUND def to javasound win32 makefile, and use it here
**				to def out direct sound stuff.  we *cannot* link with dsound.dll or we get an
**				unsatisfied link error on machines that don't have in on the system (#4110266).
**
**	05.06.98:	$$kk: made lastPos global and updated in HAE_AquireAudioCard to deal with
**				bug #4135246.
*/
/*****************************************************************************/

#ifndef WIN32_EXTRA_LEAN
	#define WIN32_EXTRA_LEAN
#endif
#ifndef WIN32_LEAN_AND_MEAN
	#define WIN32_LEAN_AND_MEAN
#endif

#include <windows.h>
#include <mmsystem.h>
#include <windowsx.h>
#include <dsound.h>

#include <stdio.h>
#include <fcntl.h>
#include <io.h>

#include "HAE_API.h"

#define USE_WIN32_FILE_IO				1

// current device ID. 0 is waveOut, 1 is directSound
static long					g_currentDeviceID = 0;

#define HAE_WAVEOUT_FRAMES_PER_BLOCK	5				// how much ms latancy can we handle (in ms)
#define HAE_WAVEOUT_SOUND_PERIOD		11				// sleep period between position checks (in ms)

static void					*g_audioBufferBlock[HAE_WAVEOUT_FRAMES_PER_BLOCK];
static long					g_audioByteBufferSize;			// size of audio buffers in bytes

static HWAVEOUT				g_waveDevice = 0;
static long					g_shutDownDoubleBuffer;
static long					g_activeDoubleBuffer;

// $$kk: 05.06.98: made lastPos a global variable
static long					lastPos;


#ifndef NO_DIRECT_SOUND
// How many audio frames to generate at one time
#define HAE_DIRECTSOUND_FRAMES_PER_BLOCK	8
#define HAE_DIRECTSOUND_FRAMES_PER_BLOCK_NT	16
// How long in milliseconds to sleep between polling DirectSound play position?
#define HAE_DIRECTSOUND_SOUND_PERIOD		6
#define HAE_DIRECTSOUND_SOUND_PERIOD_NT		3
#endif

// Structure to describe output sound buffer topology
struct DXSOUNDPOSDATA
{
	unsigned long	uiBufferSize;			// Size of DX sound play buffer
	unsigned long	uiBufferBlocks;			// Number of sub blocks in play buffer
	unsigned long	uiWriteBlock;			// Current block to fill with audio samples
	unsigned long	uiDestPos;				// current destination buffer offset
	unsigned long	uiBlockSize;			// size of data block to copy
	unsigned long	uiSynthFrameBytes;		// How many bytes per audio frame
	unsigned long	uiSynthFramesPerBlock;	// How many frames to audio process at once
	BOOL			bWaveEnd;				// Signal to stop processing audio
	BOOL			bDone;					// Has last generated frame been played
};
typedef struct DXSOUNDPOSDATA	DXSOUNDPOSDATA;

// Output buffer sub block info
struct SOUNDPOSBLOCK
{
	unsigned long	dwStart;				// Offset of block start
	unsigned long	dwEnd;					// Offset to block end
	BOOL			bEndingBlock;			// Does this block contain final samples?
};
typedef struct SOUNDPOSBLOCK	SOUNDPOSBLOCK;

// Format of DirectSound output buffer
// $$kk: 04.16.98: or the waveout output!
static WAVEFORMATEX			g_waveFormat;


#ifndef NO_DIRECT_SOUND
static LPDIRECTSOUND		g_pDirectSoundObject = NULL;
static LPDIRECTSOUNDBUFFER	g_pDirectSoundBuffer = NULL;
static LPDIRECTSOUNDBUFFER	g_pDirectSoundPrimaryBuffer = NULL;
static unsigned long		g_soundBufferByteSize;


// Thread used to poll DirectSound play position
HANDLE						g_hthreadSoundMon = NULL;

// Window handle of Client application
HWND						g_directSoundWindow = NULL;
// How is the output buffer divided for purposes of monitoring the play position
static long					g_nBufferSegments = 2;

 // Output sound buffer record
static DXSOUNDPOSDATA		g_SPD;
 // Are we currently generating audio frames?
static BOOL					g_bProcessingFrame = FALSE;
// Events to signal the start and end of a batch audio frame processing
static HANDLE				g_heventProcessFrameStart = NULL;
static HANDLE				g_heventProcessFrameFinish = NULL;
#endif

 // number of samples per audio frame to generate
long						g_audioFramesToGenerate;

// How many audio frames to generate at one time
static unsigned int			g_synthFramesPerBlock;	// setup upon runtime

static unsigned int			g_audioPeriodSleepTime;	// setup upon runtime

// This file contains API's that need to be defined in order to get HAE (IgorAudio)
// to link and compile.

// **** System setup and cleanup functions
// Setup function. Called before memory allocation, or anything serious. Can be used to
// load a DLL, library, etc.
// return 0 for ok, or -1 for failure
int HAE_Setup(void)
{
	return 0;
}

// Cleanup function. Called after all memory and the last buffers are deallocated, and
// audio hardware is released. Can be used to unload a DLL, library, etc.
// return 0 for ok, or -1 for failure
int HAE_Cleanup(void)
{
	return 0;
}

// **** Memory management
// allocate a block of locked, zeroed memory. Return a pointer
void * HAE_Allocate(unsigned long size)
{
	void	*data;

	data = 0;
	if (size)
	{
		// the GHND flag includes the GMEM_ZEROINIT flag
		data = (char *)GlobalAllocPtr(GHND, size);
	}
	return data;
}

// dispose of memory allocated with HAE_Allocate
void HAE_Deallocate(void * memoryBlock)
{
	if (memoryBlock)
	{
		GlobalFreePtr(memoryBlock);
	}
}

// Given a memory pointer and a size, validate of memory pointer is a valid memory address
// with at least size bytes of data avaiable from the pointer.
// This is used to determine if a memory pointer and size can be accessed without
// causing a memory protection
// fault.
// return 0 for valid, or 1 for bad pointer, or 2 for not supported.
int HAE_IsBadReadPointer(void *memoryBlock, unsigned long size)
{
	return (IsBadReadPtr(memoryBlock, size)) ? 1 : 0;
}

// this will return the size of the memory pointer allocated with HAE_Allocate. Return
// 0 if you don't support this feature
unsigned long HAE_SizeOfPointer(void * memoryBlock)
{
	unsigned long	size;
	HANDLE			hData;

	size = 0;
	if (memoryBlock)
	{
		hData = GlobalPtrHandle(memoryBlock);
		if (hData)
		{
			size = GlobalSize(hData);
		}
	}
	return size;
}

// block move memory. This is basicly memcpy, but its exposed to take advantage of
// special block move speed ups, various hardware has available.
void HAE_BlockMove(void * source, void * dest, unsigned long size)
{
	if (source && dest && size)
	{
		memcpy(dest, source, size);
	}
}

// **** Audio Card modifiers
// Return 1 if stereo hardware is supported, otherwise 0.
int HAE_IsStereoSupported(void)
{
	MMRESULT	theErr;
	int			status;
	WAVEOUTCAPS	info;

	status = 0;
	theErr = waveOutGetDevCaps(WAVE_MAPPER, &info, sizeof(WAVEOUTCAPS));
	if (theErr == MMSYSERR_NOERROR)
	{
		if (info.wChannels > 1)
		{
			status = 1;	// stereo supported
		}
	}
	return status;
}

// Return 1, if sound hardware support 16 bit output, otherwise 0.
int HAE_Is16BitSupported(void)
{
	MMRESULT	theErr;
	int			status, count;
	WAVEOUTCAPS	info;
	static int	caps[] = {	WAVE_FORMAT_1M16, WAVE_FORMAT_1S16, WAVE_FORMAT_2M16,
							WAVE_FORMAT_2S16, WAVE_FORMAT_4M16, WAVE_FORMAT_4S16};

	status = 0;
	theErr = waveOutGetDevCaps(WAVE_MAPPER, &info, sizeof(WAVEOUTCAPS));
	if (theErr == MMSYSERR_NOERROR)
	{
		for (count = 0; count < 6; count++)
		{
			if (info.dwFormats & caps[count])
			{
				status = 1;		// 16 bit supported
				break;
			}
		}
	}
	return status;
}

// $$kk: 03.17.98
// Return 1, if sound hardware support 8 bit linear output, otherwise 0.
// to my knowledge, all windows drivers do.  (should we check whether there
// is any audio hardware at all here??
int HAE_Is8BitSupported(void)
{
	return 1;
}

// returned volume is in the range of 0 to 256
short int HAE_GetHardwareVolume(void)
{
	MMRESULT	theErr;
	DWORD		volume;

	volume = 256;
	theErr = waveOutGetVolume((HWAVEOUT)WAVE_MAPPER, &volume);
	if (theErr == MMSYSERR_NOERROR)
	{
		volume = (volume & 0xFFFF) / 256;	// scale down to 0 to 256
	}
	return (short int)volume;
}

// theVolume is in the range of 0 to 256
void HAE_SetHardwareVolume(short int theVolume)
{
	MMRESULT	theErr;
	DWORD		volume;

	volume = theVolume * 256;	// scale up to 0 to 65536
	volume |= (volume << 16L);	// copy into high word
	theErr = waveOutSetVolume((HWAVEOUT)WAVE_MAPPER, volume);
}

// **** Timing services
// return microseconds
unsigned long HAE_Microseconds(void)
{
	static unsigned long starttick = 0;

	if (starttick == 0)
	{
		starttick = timeGetTime() * 1000L;
	}
	return ((timeGetTime() * 1000L) - starttick);
}

// wait or sleep this thread for this many microseconds
// CLS??: If this function is called from within the frame thread and
// JAVA_THREAD is non-zero, we'll probably crash.
void HAE_WaitMicroseocnds(unsigned long waitAmount)
{
	unsigned long	ticks;

	ticks = HAE_Microseconds() + waitAmount;
	while (HAE_Microseconds() < ticks)
	{
		Sleep(0);	// Give up the rest of this time slice to other threads
	}
}

// **** File support
// Create a file, delete orignal if duplicate file name.
// Return -1 if error

// Given the fileNameSource that comes from the call HAE_FileXXXX, copy the name
// in the native format to the pointer passed fileNameDest.
void HAE_CopyFileNameNative(void *fileNameSource, void *fileNameDest)
{
	char *dest;
	char *src;

	if (fileNameSource && fileNameDest)
	{
		dest = (char *)fileNameDest;
		src = (char *)fileNameSource;
		if (src == NULL)
		{
			src = "";
		}
		if (dest)
		{
			while (*src)
			{
				*dest++ = *src++;
			}
			*dest = 0;
		}
	}
}

long HAE_FileCreate(void *fileName)
{
#if USE_WIN32_FILE_IO == 0
	int		file;

	file = _creat((char *)fileName, _S_IREAD | _S_IWRITE | _O_RDWR | _O_TRUNC);
	if (file != -1)
	{
		_close(file);
	}
	return (file != -1) ? 0 : -1;
#else
	HANDLE	newFile;

	newFile = CreateFile((LPCTSTR)fileName, GENERIC_READ, FILE_SHARE_READ, NULL, CREATE_NEW | TRUNCATE_EXISTING,
									FILE_ATTRIBUTE_READONLY | FILE_FLAG_RANDOM_ACCESS,
									NULL);
	CloseHandle(newFile);
	return newFile ? 0 : -1;
#endif
}


// Open a file
// Return -1 if error, otherwise file handle
long HAE_FileOpenForRead(void *fileName)
{
	if (fileName)
	{
#if USE_WIN32_FILE_IO == 0
		return _open((char *)fileName, _O_RDONLY | _O_BINARY);
#else
	return (long)CreateFile((LPCTSTR)fileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,
									FILE_ATTRIBUTE_READONLY | FILE_FLAG_RANDOM_ACCESS,
									NULL);
#endif
	}
	return -1;
}

long HAE_FileOpenForWrite(void *fileName)
{
	if (fileName)
	{
#if USE_WIN32_FILE_IO == 0
		return _open((char *)fileName, _O_WRONLY | _O_CREAT | _O_TRUNC | _O_BINARY);
#else
	return (long)CreateFile((LPCTSTR)fileName, GENERIC_WRITE, 0, NULL, CREATE_NEW | TRUNCATE_EXISTING,
									FILE_FLAG_RANDOM_ACCESS,
									NULL);
#endif
	}
	return -1;
}

long HAE_FileOpenForReadWrite(void *fileName)
{
	if (fileName)
	{
#if USE_WIN32_FILE_IO == 0
		return _open((char *)fileName, _O_RDWR | _O_BINARY);
#else
	return (long)CreateFile((LPCTSTR)fileName, GENERIC_READ | GENERIC_WRITE, 0, NULL,
									OPEN_EXISTING,
									FILE_FLAG_RANDOM_ACCESS,
									NULL);
#endif
	}
	return -1;
}

// Close a file
void HAE_FileClose(long fileReference)
{
#if USE_WIN32_FILE_IO == 0
	_close(fileReference);
#else
	CloseHandle((HANDLE)fileReference);
#endif
}

// Read a block of memory from a file.
// Return -1 if error, otherwise length of data read.
long HAE_ReadFile(long fileReference, void *pBuffer, long bufferLength)
{
	if (pBuffer && bufferLength)
	{
#if USE_WIN32_FILE_IO == 0
		return _read(fileReference, (char *)pBuffer, bufferLength);
#else
		{
			DWORD	readFromBuffer;
			return ReadFile((HANDLE)fileReference, (LPVOID)pBuffer,
										bufferLength, &readFromBuffer,
										NULL) ? readFromBuffer : -1;
		}
#endif
	}
	return -1;
}

// Write a block of memory from a file
// Return -1 if error, otherwise length of data written.
long HAE_WriteFile(long fileReference, void *pBuffer, long bufferLength)
{
	if (pBuffer && bufferLength)
	{
#if USE_WIN32_FILE_IO == 0
		return _write(fileReference, (char *)pBuffer, bufferLength);
#else
		{
			DWORD	writtenFromBuffer;
			return WriteFile((HANDLE)fileReference, (LPVOID)pBuffer,
										bufferLength, &writtenFromBuffer,
										NULL) ? writtenFromBuffer : -1;
		}
#endif
	}
	return -1;
}

// set file position in absolute file byte position
// Return -1 if error, otherwise 0.
long HAE_SetFilePosition(long fileReference, unsigned long filePosition)
{
#if USE_WIN32_FILE_IO == 0
	return (_lseek(fileReference, filePosition, SEEK_SET) == -1) ? -1 : 0;
#else
	return (
			(SetFilePointer((HANDLE)fileReference, filePosition, NULL, FILE_BEGIN) == 0xFFFFFFFFL)
			 ? -1 : 0);
#endif
}

// get file position in absolute file bytes
unsigned long HAE_GetFilePosition(long fileReference)
{
#if USE_WIN32_FILE_IO == 0
	return _lseek(fileReference, 0, SEEK_CUR);
#else
	return SetFilePointer((HANDLE)fileReference, 0, NULL, FILE_CURRENT);
#endif
}

// get length of file
unsigned long HAE_GetFileLength(long fileReference)
{
	unsigned long pos;

#if USE_WIN32_FILE_IO == 0
	pos = _lseek(fileReference, 0, SEEK_END);
	_lseek(fileReference, 0, SEEK_SET);
#else
	pos = GetFileSize((HANDLE)fileReference, NULL);
	if (pos == 0xFFFFFFFFL)
	{
		pos = 0;
	}
#endif
	return pos;
}

// set the length of a file. Return 0, if ok, or -1 for error
int HAE_SetFileLength(long fileReference, unsigned long newSize)
{
#if USE_WIN32_FILE_IO == 0
	return _chsize(fileReference, newSize);
#else
	int error;

	error = -1;
	if (HAE_SetFilePosition(fileReference, newSize) == 0)
	{
		error = SetEndOfFile((HANDLE)fileReference) ? 0 : -1;
	}
	return error;
#endif
}

// Determine if we should use DirectSound or just fall back to the waveOut API.
// return 0 for waveOut, or 1 for DirectSound
static int PV_UseDirectSound(void)
{
/*
// this method does not seem to work
	WAVEOUTCAPS	info;
	int			status;
	MMRESULT	theErr;

	status = 0;
	theErr = waveOutGetDevCaps(WAVE_MAPPER, &info, sizeof(WAVEOUTCAPS));
	if (theErr == MMSYSERR_NOERROR)
	{
		if ((info.dwSupport & WAVECAPS_DIRECTSOUND) == WAVECAPS_DIRECTSOUND)
		{
			status = 1;
		}
	}
	return status;
*/
	return g_currentDeviceID;	// 0 is waveOut, 1 is DirectSound

}

static int PV_IsNT(void)
{
	static int	firstTime = TRUE;
	static int	underNT;

	if (firstTime)
	{
		underNT = (GetVersion() < 0x80000000) ? TRUE : FALSE;
	}
	return underNT;
}

#ifndef NO_DIRECT_SOUND
// When using DirectSound, we need a window handle (!). So this function looks
// around and waits around until it can find an active window handle.
// This has to be done for Windows only, because the message system needs
// a window handle to process messages. If we can't find one, bail with an
// error.
static HWND PV_GetMostActiveWindow(void)
{
	short int	count;
	HWND		theWindow;

	theWindow = NULL;
	// we're going to count to 100, and wait for a window to become active by
	// sleeping in this thread then checking to see if anything has changed.
	for (count = 0; count < 100; count++)
	{
		theWindow = GetForegroundWindow();
		if (theWindow == NULL)
		{
			theWindow = GetActiveWindow();
		}
		if (theWindow == NULL)
		{
			theWindow = GetFocus();
		}
		if (theWindow)
		{	// found something, so break out of here
			break;
		}
		Sleep(10);	// sleep 10 milliseconds
	}
	return theWindow;
}
#endif

static void PV_AudioWaveOutFrameThread(void* context)
{
	WAVEHDR			waveHeader[HAE_WAVEOUT_FRAMES_PER_BLOCK];
	long			waveHeaderCount;
	MMTIME			audioStatus;

	//$$kk: 05.06.98: i am making lastPos global to fix P1 bug #4135246, in
	// which the jdk samples sometimes do not get audio.  depending
	// on the order of events when the audio device is opened and
	// closed, the frame thread may get stuck in the wait().
    long			count, currentPos, /*lastPos,*/ error;

	char			*pFillBuffer;

	if (g_audioBufferBlock)
	{
		memset(&waveHeader, 0, sizeof(WAVEHDR) * HAE_WAVEOUT_FRAMES_PER_BLOCK);
		memset(&audioStatus, 0, (long)sizeof(MMTIME));
		audioStatus.wType = TIME_BYTES;	// get byte position

		error = waveOutGetPosition(g_waveDevice, &audioStatus, sizeof(MMTIME));
		lastPos = audioStatus.u.cb - (g_audioByteBufferSize * g_synthFramesPerBlock * 2);

		// now write out all of the data built
		for (count = 0; count < HAE_WAVEOUT_FRAMES_PER_BLOCK; count++)
		{
			waveHeader[count].lpData = (char *)g_audioBufferBlock[count];
			waveHeader[count].dwBufferLength = g_audioByteBufferSize * g_synthFramesPerBlock;
			waveHeader[count].dwFlags 		= 0;
			waveHeader[count].dwLoops 		= 0;
			error = waveOutPrepareHeader(g_waveDevice, &waveHeader[count], (long)sizeof(WAVEHDR));
		}

		waveHeaderCount = 0;
		while ( (g_activeDoubleBuffer) && (g_shutDownDoubleBuffer == FALSE) )
		{
			pFillBuffer = (char *)g_audioBufferBlock[waveHeaderCount];
			for (count = 0; count < HAE_WAVEOUT_FRAMES_PER_BLOCK; count++)
			{
				// Generate one frame audio
				HAE_BuildMixerSlice(context, pFillBuffer, g_audioByteBufferSize,
															g_audioFramesToGenerate);
				pFillBuffer += g_audioByteBufferSize;

				if (g_shutDownDoubleBuffer)
				{
					break;	// time to quit
				}
			}

			if (g_shutDownDoubleBuffer == FALSE)
			{
				if (PV_IsNT() == FALSE)		// under Win95?
				{
					waveHeader[waveHeaderCount].dwFlags &= WHDR_DONE;	// must do, or buffers stop on Win95
				}
				error = waveOutWrite(g_waveDevice, &waveHeader[waveHeaderCount], (long)sizeof(WAVEHDR));

				// $$kk: 05.06.98: i am incrementing lastPos right after the write operation
				// to lessen the chance of a device open/close messing us up in the wait loop.
				lastPos += (g_audioByteBufferSize * g_synthFramesPerBlock);


				waveHeaderCount++;
				if (waveHeaderCount == HAE_WAVEOUT_FRAMES_PER_BLOCK)
				{
					waveHeaderCount = 0;
				}
				error = waveOutGetPosition(g_waveDevice, &audioStatus, sizeof(MMTIME));
				currentPos = audioStatus.u.cb;
				while ((currentPos < lastPos) && (g_shutDownDoubleBuffer == FALSE))
				{
					HAE_SleepFrameThread(context, HAE_WAVEOUT_SOUND_PERIOD);		// in ms

					error = waveOutGetPosition(g_waveDevice, &audioStatus, sizeof(MMTIME));
					currentPos = audioStatus.u.cb;
				}

				// $$kk: 05.06.98: moved this line up.
				//lastPos += (g_audioByteBufferSize * g_synthFramesPerBlock);

			}
		}
	}
	waveOutReset(g_waveDevice);		// stop all audio before unpreparing headers

	for (count = 0; count < HAE_WAVEOUT_FRAMES_PER_BLOCK; count++)
	{
		error = waveOutUnprepareHeader(g_waveDevice, &waveHeader[count], (long)sizeof(WAVEHDR));
	}
	g_activeDoubleBuffer = FALSE;
}


// Return the number of 11 ms buffer blocks that are built at one time.
int HAE_GetAudioBufferCount(void)
{
	return g_synthFramesPerBlock;
}

// Return the number of bytes used for audio buffer for output to card
long HAE_GetAudioByteBufferSize(void)
{
	return g_audioByteBufferSize;
}


/*
static HRESULT PV_ConvertDSErrorCodeToSomethingReal(HRESULT hr)
{
	switch (hr)
	{
		default:
			hr = 1;
			break;
		case DS_OK:
			hr = 2;
			break;
		case DSERR_ALLOCATED:
			hr = 3;
			break;
		case DSERR_CONTROLUNAVAIL:
			hr = 4;
			break;
		case DSERR_INVALIDPARAM:
			hr = 5;
			break;
		case DSERR_INVALIDCALL:
			hr = 6;
			break;
		case DSERR_GENERIC:
			hr = 7;
			break;
		case DSERR_PRIOLEVELNEEDED:
			hr = 8;
			break;
		case DSERR_OUTOFMEMORY:
			hr = 9;
			break;
		case DSERR_BADFORMAT:
			hr = 10;
			break;
		case DSERR_UNSUPPORTED:
			hr = 11;
			break;
		case DSERR_NODRIVER:
			hr = 12;
			break;
		case DSERR_ALREADYINITIALIZED:
			hr = 13;
			break;
		case DSERR_NOAGGREGATION:
			hr = 14;
			break;
		case DSERR_BUFFERLOST:
			hr = 15;
			break;
		case DSERR_OTHERAPPHASPRIO:
			hr = 16;
			break;
		case DSERR_UNINITIALIZED:
			hr = 17;
			break;
	}
	return hr;
}
*/


#ifndef NO_DIRECT_SOUND

// DWORD PV_AudioDirectSoundFrameThread (LPVOID lpv)
//
// This is the thread function that fills the current block of the
// output buffer.
//
// This thread will sleep until the "g_heventProcessFrameStart" event
// is set by the PV_MonitorDirectSoundPosition thread.  After the current
// block has been filled, the "g_heventProcessFrameFinish" event is
// set and the thread goes back to sleep.
//
// The call to ProcessSampleFrame is what drives the synth engine.
// Each call to this function generates "g_SPD.uiSynthFrameBytes"
// bytes of audio data.  HAE_BuildMixerSlice will be called
// g_SPD.uiSynthFramesPerBlock times to fill the current block.
//
// If the flag "g_SPD.bWaveEnd" is set then PV_AudioDirectSoundFrameThread will fill
// the current block with silence instead of calling generating
// more audio samples.

static void PV_AudioDirectSoundFrameThread(void* context)
{
	HRESULT		error;
	LPVOID		lpWrite;
	DWORD		dwWriteLen;
	DWORD		count;
	char		*pFillBuffer;

	while ( (g_activeDoubleBuffer) && (g_shutDownDoubleBuffer == FALSE) )
	{
		g_bProcessingFrame = FALSE;
		// Wait for someone to wake us up.
		WaitForSingleObject(g_heventProcessFrameStart, INFINITE);
		g_bProcessingFrame = TRUE;

		// Lock the current block of the DirectSound buffer.
		error = IDirectSoundBuffer_Lock(g_pDirectSoundBuffer,
				g_SPD.uiDestPos, g_SPD.uiBlockSize,
				&lpWrite, &dwWriteLen, NULL, NULL, 0);

		if (error == DS_OK)
		{
			// If we're to generate more audio samples
			if (!g_SPD.bWaveEnd)
			{
				// For each frame of the current block
				for (count = 0; count < g_SPD.uiSynthFramesPerBlock; count++)
				{
					pFillBuffer = ((char *)lpWrite) + count * g_SPD.uiSynthFrameBytes;

					// Generate new audio samples, putting the directly
					// into the output buffer.
					HAE_BuildMixerSlice(context, pFillBuffer, g_audioByteBufferSize, g_audioFramesToGenerate);
					if (g_shutDownDoubleBuffer)
					{
						break;
					}
				}
			}
			else
			{
				// Fill the current block with appropriate silence value/
				memset(lpWrite, (g_waveFormat.wBitsPerSample == 16) ? 0 : 0x80, dwWriteLen);
			}
			// Unlock the block of SoundBuffer were we writing.
			error = IDirectSoundBuffer_Unlock(g_pDirectSoundBuffer,
													lpWrite, dwWriteLen, NULL, 0);

			// Advance to the next block in the output buffer
			g_SPD.uiWriteBlock++;
			// We're circular, so when we reach the end of the buffer,
			// start back at the beginning.
			if (g_SPD.uiWriteBlock > (g_SPD.uiBufferBlocks - 1))
			{
				g_SPD.uiWriteBlock = 0;
			}
			// Calculate the new offsets.
			g_SPD.uiDestPos = g_SPD.uiWriteBlock * g_SPD.uiBlockSize;
		}

		// Set the ProcessFrameFinish event before going back to sleep.
		SetEvent(g_heventProcessFrameFinish);
	}

	g_activeDoubleBuffer = FALSE;
}



// DWORD PV_MonitorDirectSoundPosition(LPVOID lpv)
//
// This is the main thread routine for our engine.
//
// Here's a synopsis of what's going on:
// - Divide the output buffer into g_nBufferSegments.
// - Poll the DirectSound play cursor.
// - When the play cursor crosses the transition
//   from iPrevBlock to iCurBlock, wake up the PV_AudioDirectSoundFrameThread thread
//   to fill the next block.
// - Wait for PV_AudioDirectSoundFrameThread to finish
// - Sleep so that we aren't hogging the CPU with our polling
// - Wake up and poll the play cursor again.
//
// When the g_SPD.bWaveEnd flag is set, indicating that we
// should stop:
// - Mark the block we were polling in when the flag was set.
// - Wake up the PV_AudioDirectSoundFrameThread thread to fill the last block.
// - Let the polling engine know that the last block has been
//   filled by setting the bWaitingForLastBlock flag.
// - Wait for the last block to be played by continuing to poll
//   until the play cursor has passed through the last block.
//
// Note:
//   The block indecies used in this polling thread are not the same
//   as the indecies used in the PV_AudioDirectSoundFrameThread thread.  They are
//   offset by one because the first block of the sound buffer is
//   filled with silence.  This allows the engine to generate the
//   first audio frames while this silent block is playing.
//
//   There is a problem with this design though.  If PV_AudioDirectSoundFrameThread takes
//   more time to process the current block than it takes to play the
//   previous block, the polling loop thread will have missed the next
//   block transition.  This will result in having to wait until the same
//   transition comes around again before more audio is generated.  The
//   blocks after the missed transition will not be filled with new data
//   therefore old audio blocks will be replayed.

static DWORD WINAPI PV_MonitorDirectSoundPosition(LPVOID lpv)
{
	BOOL			bWaitingForLastBlock;
	DWORD			dwPlayPos, dwWritePos;
	int				iPrevBlock;
	int				iCurBlock;
	SOUNDPOSBLOCK	*pSoundBlock;
	DWORD			dwCurPos;
	DWORD			count;

	lpv = lpv;
	dwCurPos = 0;
	iPrevBlock = 0;
	iCurBlock = 0;
	bWaitingForLastBlock = FALSE;
	pSoundBlock = (SOUNDPOSBLOCK *)HAE_Allocate(sizeof(SOUNDPOSBLOCK) * g_SPD.uiBufferBlocks);
	if (pSoundBlock)
	{
		// Calculate the boundaries for each of the blocks
		for (count = 0; count < g_SPD.uiBufferBlocks; count++)
		{
			pSoundBlock[count].dwStart = dwCurPos;
			pSoundBlock[count].dwEnd = dwCurPos + g_SPD.uiBlockSize;
			pSoundBlock[count].bEndingBlock = FALSE;
			dwCurPos += g_SPD.uiBlockSize;
		}

		// While the engine is not done
		while(!g_SPD.bDone)
		{
			// Give up some of our time slice
			Sleep(g_audioPeriodSleepTime);

			// Get the current play position
			IDirectSoundBuffer_GetCurrentPosition(g_pDirectSoundBuffer, &dwPlayPos, &dwWritePos);

			// If the play position has entered the current block and it hasn't
			// reached the end of the block yet
			if (dwPlayPos >= pSoundBlock[iCurBlock].dwStart &&
				dwPlayPos < pSoundBlock[iCurBlock].dwEnd)
			{
				// Wake up PV_AudioDirectSoundFrameThread to generate a block of audio
				SetEvent(g_heventProcessFrameStart);
				// Wait for PV_AudioDirectSoundFrameThread to finish
				WaitForSingleObject(g_heventProcessFrameFinish, INFINITE);

				// If the this is the last block to be generated
				// and we haven't marked it yet
				if (g_SPD.bWaveEnd && !bWaitingForLastBlock)
				{
					// Mark the current block as the last block
					pSoundBlock[iCurBlock].bEndingBlock = TRUE;
					bWaitingForLastBlock = TRUE;
				}

				// If we're waiting for the last block to play
				if (bWaitingForLastBlock)
				{
					iPrevBlock = iCurBlock - 1;
					if (iPrevBlock < 0)
					{
						iPrevBlock = g_SPD.uiBufferBlocks - 1;
					}
					// If the current block is the last block
					if (pSoundBlock[iPrevBlock].bEndingBlock)
					{
						// End the thread and stop playing sound
						g_SPD.bDone = TRUE;
						IDirectSoundBuffer_Stop(g_pDirectSoundBuffer);
					}
				}

				// Go to the next block.
				iCurBlock++;
				// If it's the last in the list wrap back to the first.
				if ((UINT)iCurBlock > g_SPD.uiBufferBlocks - 1)
				{
					iCurBlock = 0;
				}
			}
		}
	}

	// Get rid of our list of sound blocks
	HAE_Deallocate(pSoundBlock);
	return 0;
}

// void PV_ClearDirectSoundBuffer ()
//
// Files the DirectSound buffer with silence.

static void PV_ClearDirectSoundBuffer(void)
{
	HRESULT		error;
	LPVOID		lpWrite;
	DWORD		dwWriteLen;

	if (g_pDirectSoundBuffer)
	{
		error = IDirectSoundBuffer_Lock(g_pDirectSoundBuffer, 0,
										g_SPD.uiBlockSize * g_SPD.uiBufferBlocks,
										&lpWrite, &dwWriteLen, NULL, NULL,
										0);
		if (error == DS_OK)
		{
			memset(lpWrite, (g_waveFormat.wBitsPerSample == 16) ? 0 : 0x80, dwWriteLen);
			error = IDirectSoundBuffer_Unlock(g_pDirectSoundBuffer,
													lpWrite, dwWriteLen, NULL, 0);
		}
	}
}

// int PV_CreateDirectSoundThreads()
//
// Inititialize the SOUNDPOSDATA structure and create the threads
// and events for our engine
// returns -1 if failure, or 0 for success
static int PV_CreateDirectSoundThreads(void* context)
{
	int		status;
	DWORD	threadID;

	status = -1;
	g_heventProcessFrameStart = CreateEvent(NULL, FALSE, FALSE, NULL);
	if (g_heventProcessFrameStart)
	{
		g_heventProcessFrameFinish = CreateEvent(NULL, FALSE, FALSE, NULL);
		if (g_heventProcessFrameFinish)
		{
			// Create the monitor thread
			g_hthreadSoundMon = CreateThread(NULL, 0L,
												PV_MonitorDirectSoundPosition, 0L,
												0L, &threadID);
			if (g_hthreadSoundMon)
			{
				if (HAE_CreateFrameThread(context, PV_AudioDirectSoundFrameThread) >= 0)
				{
					// THREAD_PRIORITY_HIGHEST
					// THREAD_PRIORITY_NORMAL
					// THREAD_PRIORITY_TIME_CRITICAL
					SetThreadPriority(g_hthreadSoundMon, THREAD_PRIORITY_TIME_CRITICAL);
					status = 0;
				}
			}
		}
	}
	return status;
}


// setup direct sound, create a primary buffer and set format to match.
// Return 0 if success, or -1 if it fails
static int PV_SetupDirectSound(void* context, int sampleRate, int channels, int bitSize)
{
	int				status;
	HRESULT			error;
	DSBUFFERDESC	dsbd;

	status = -1;
	memset(&g_waveFormat, 0, (DWORD)sizeof(WAVEFORMATEX));

	g_directSoundWindow = PV_GetMostActiveWindow();
	if (g_directSoundWindow)
	{
		// Use standard PCM audio samples

		g_waveFormat.wFormatTag = WAVE_FORMAT_PCM;

		g_waveFormat.nSamplesPerSec = sampleRate;
		g_waveFormat.nChannels = channels;
		g_waveFormat.wBitsPerSample = bitSize;

		// Calculate size in bytes of an audio sample
		g_waveFormat.nBlockAlign =  g_waveFormat.nChannels * g_waveFormat.wBitsPerSample / 8;
		g_waveFormat.nAvgBytesPerSec = g_waveFormat.nSamplesPerSec * g_waveFormat.nBlockAlign;
		g_waveFormat.cbSize = 0;

		// Try and create a DirectSound object
		if (DirectSoundCreate(NULL, &g_pDirectSoundObject, NULL) == DS_OK)
		{
			// First thing that must be done after creating object
			// Set level to PRIORITY so that we can change format of primary buffer
			error = IDirectSound_SetCooperativeLevel(g_pDirectSoundObject,
												g_directSoundWindow, DSSCL_PRIORITY);
			if (error == DS_OK)
			{
				// Create a primary buffer and set its format to that of our synth engine
				dsbd.dwSize = sizeof(dsbd);
				dsbd.dwFlags = DSBCAPS_PRIMARYBUFFER;

				dsbd.dwBufferBytes = 0;
				dsbd.dwReserved = 0;
				dsbd.lpwfxFormat = NULL;
				error = IDirectSound_CreateSoundBuffer(g_pDirectSoundObject, &dsbd, &g_pDirectSoundPrimaryBuffer, NULL);

				if ( (error == DS_OK) && (g_pDirectSoundPrimaryBuffer) )
				{
					// By setting the primary buffer to have the same format as our generated
					// synth data, there will be no extra overhead when DirectSound mixes
					// our "Secondary" buffer into the "Primary" buffer.
					error = IDirectSoundBuffer_SetFormat(g_pDirectSoundPrimaryBuffer,
																			&g_waveFormat);
					if (error == DS_OK)
					{
						// Now we can create the secondary sound buffer we will use to output sound
						dsbd.dwSize = sizeof(dsbd);
						dsbd.dwFlags = DSBCAPS_STICKYFOCUS;

// $$kk: DSBCAPS_GETCURRENTPOSITION2 does not compile (DS vesion issue??)
//						dsbd.dwFlags |= DSBCAPS_GETCURRENTPOSITION2;

//						dsbd.dwFlags |= DSBCAPS_GLOBALFOCUS;
						dsbd.dwBufferBytes = g_waveFormat.nBlockAlign *
												g_audioFramesToGenerate *
												g_synthFramesPerBlock *
												g_nBufferSegments;
						dsbd.dwReserved = 0;
						dsbd.lpwfxFormat = &g_waveFormat;

						error = IDirectSound_CreateSoundBuffer(g_pDirectSoundObject, &dsbd, &g_pDirectSoundBuffer, NULL);
						if ( (error == DS_OK) && (g_pDirectSoundPrimaryBuffer) )
						{
							// Make sure we start at the beginning of the buffer
							error = IDirectSoundBuffer_SetCurrentPosition(g_pDirectSoundBuffer, 0);
							if (error == DS_OK)
							{
								// Start by writing an empty block
								PV_ClearDirectSoundBuffer();

								g_SPD.uiSynthFramesPerBlock = g_synthFramesPerBlock;
								// Number of bytes per audio frame
								g_SPD.uiSynthFrameBytes = g_audioFramesToGenerate * g_waveFormat.nBlockAlign;
								// Initial output buffer offset to start wrtting audio data
								g_SPD.uiDestPos = g_SPD.uiSynthFrameBytes * g_SPD.uiSynthFramesPerBlock;
								// Total number of bytes generated on each pass through PV_AudioDirectSoundFrameThread
								g_SPD.uiBlockSize = g_SPD.uiSynthFrameBytes * g_SPD.uiSynthFramesPerBlock;
								// Total number of blocks in output buffer
								g_SPD.uiBufferBlocks = g_nBufferSegments;
								// Total number of bytes in output buffer
								g_SPD.uiBufferSize = g_SPD.uiBlockSize * g_nBufferSegments;
								// Index of next block to we written
								g_SPD.uiWriteBlock = 1;
								g_SPD.bWaveEnd = FALSE;
								g_SPD.bDone = FALSE;

								// Create thread to poll sound play position
								status = PV_CreateDirectSoundThreads(context);
								if (status == 0)
								{
									// Start playing whatever is in the output buffer (silence initially)
									error = IDirectSoundBuffer_Play(g_pDirectSoundBuffer,
																			0,0,DSBPLAY_LOOPING);
									if (error == DS_OK)
									{
										status = 0;
									}
									else
									{
										status = -1;
									}
								}
							}
						}
					}
				}
			}
		}
	}
	return status;
}

static void PV_CleanupDirectSound(void)
{
	if (g_pDirectSoundBuffer)
	{
		IDirectSoundBuffer_Stop(g_pDirectSoundBuffer);
	}
	if (g_pDirectSoundPrimaryBuffer)
	{
		IDirectSoundBuffer_Release(g_pDirectSoundPrimaryBuffer);
		g_pDirectSoundPrimaryBuffer = NULL;
	}
	if (g_pDirectSoundBuffer)
	{
		IDirectSoundBuffer_Release(g_pDirectSoundBuffer);
		g_pDirectSoundBuffer = NULL;
	}
	if (g_pDirectSoundObject)
	{
		IDirectSoundBuffer_Release(g_pDirectSoundObject);
		g_pDirectSoundObject = NULL;
	}
	if (g_heventProcessFrameStart)
	{
		CloseHandle(g_heventProcessFrameStart);
		g_heventProcessFrameStart = NULL;
	}
	if (g_heventProcessFrameFinish)
	{
		CloseHandle(g_heventProcessFrameFinish);
		g_heventProcessFrameFinish = NULL;
	}
	g_directSoundWindow = NULL;
}
#endif


// **** Audio card support
// Aquire and enabled audio card
// return 0 if ok, -1 if failed
int HAE_AquireAudioCard(void *context, long sampleRate, long channels, long bits)
{
	int				flag;
	short int		count;
	long			bufferSize;
	long			error;
	WAVEFORMATEX	waveFormat;

	flag = 0;
	g_activeDoubleBuffer = FALSE;
	g_shutDownDoubleBuffer = TRUE;

	switch (sampleRate)
	{
		case 44100:
			g_audioFramesToGenerate = HAE_GetMaxSamplePerSlice();
			break;
		case 11025:
			sampleRate = 22050;
		case 22050:
			g_audioFramesToGenerate = HAE_GetMaxSamplePerSlice()/2;
			break;
	}

	// use DirectSound or waveOut API
	if (PV_UseDirectSound() == 0)
	{	// Use waveOut API
		// we're going to build this many buffers at a time
		g_synthFramesPerBlock = HAE_WAVEOUT_FRAMES_PER_BLOCK;

		waveFormat.wFormatTag = WAVE_FORMAT_PCM;

		waveFormat.nSamplesPerSec = sampleRate;
		waveFormat.nChannels = (unsigned short)channels;
		waveFormat.wBitsPerSample = (unsigned short)bits;

		// Calculate size in bytes of an audio sample
		waveFormat.nBlockAlign =  waveFormat.nChannels * waveFormat.wBitsPerSample / 8;
		waveFormat.nAvgBytesPerSec = waveFormat.nSamplesPerSec * waveFormat.nBlockAlign;
		waveFormat.cbSize = 0;
		g_waveFormat = waveFormat;

		if (waveFormat.wBitsPerSample == 8)
		{
			bufferSize = (sizeof(char) * g_audioFramesToGenerate);
		}
		else
		{
			bufferSize = (sizeof(short int) * g_audioFramesToGenerate);
		}
		bufferSize *= waveFormat.nChannels;
		g_audioByteBufferSize = bufferSize;


		// allocate buffer blocks
		flag = TRUE;
		for (count = 0; count < HAE_WAVEOUT_FRAMES_PER_BLOCK; count++)
		{
			g_audioBufferBlock[count] = HAE_Allocate(g_audioByteBufferSize * HAE_WAVEOUT_FRAMES_PER_BLOCK);
			if (g_audioBufferBlock[count] == NULL)
			{
				flag = FALSE;
				break;
			}
		}

		// try to open wave device
		if (flag)
		{
			error = waveOutOpen(&g_waveDevice,	WAVE_MAPPER, &waveFormat,
								0L, 0L,			CALLBACK_NULL
								//				| WAVE_FORMAT_DIRECT
								//				| WAVE_ALLOWSYNC
								//				| WAVE_MAPPED
								);
			if (error == 0)
			{
				g_shutDownDoubleBuffer = FALSE;
				g_activeDoubleBuffer = TRUE;	// must enable process, before thread begins


				// $$kk: 05.06.98: added this whole block.
				// we need to reset the lastPos each time the device gets acquired.
				// otherwise we may get stuck in the wait loop because we never count
				// up to the right sample position.
				{
					MMTIME			audioStatus;
					memset(&audioStatus, 0, (long)sizeof(MMTIME));
					audioStatus.wType = TIME_BYTES;	// get byte position

					error = waveOutGetPosition(g_waveDevice, &audioStatus, sizeof(MMTIME));

					lastPos = audioStatus.u.cb - (g_audioByteBufferSize * g_synthFramesPerBlock * 2);
				}


				// create thread that builds and sends buffers to the wave device
				if (HAE_CreateFrameThread(context, PV_AudioWaveOutFrameThread) >= 0)
				{
					flag = 0;
				}
				else
				{
					flag = 1;
					g_activeDoubleBuffer = FALSE;
				}
			}
			else
			{
				flag = 1;
			}
		}
	}

#ifndef NO_DIRECT_SOUND

	else
	{	// use DirectSound API
		// we're going to build this many buffers at a time
		if (PV_IsNT() == FALSE)		// under Win95?
		{
			g_synthFramesPerBlock = HAE_DIRECTSOUND_FRAMES_PER_BLOCK;
			g_audioPeriodSleepTime = HAE_DIRECTSOUND_SOUND_PERIOD;
		}
		else
		{
			g_synthFramesPerBlock = HAE_DIRECTSOUND_FRAMES_PER_BLOCK_NT;
			g_audioPeriodSleepTime = HAE_DIRECTSOUND_SOUND_PERIOD_NT;
		}
		g_shutDownDoubleBuffer = FALSE;
		g_activeDoubleBuffer = TRUE;	// must enable process, before thread begins
		g_audioByteBufferSize = g_audioFramesToGenerate * channels * (bits / 8);

		flag = PV_SetupDirectSound(context, sampleRate, channels, bits);
		if (flag != 0)
		{
			// something failed
			flag = 1;
			g_activeDoubleBuffer = FALSE;
		}
	}

#endif

	if (flag)
	{	// something failed
		HAE_ReleaseAudioCard(context);
	}
	return flag;
}

// Release and free audio card.
// return 0 if ok, -1 if failed.
int HAE_ReleaseAudioCard(void *context)
{
	long	count;
	int		status;

	g_shutDownDoubleBuffer = TRUE;	// signal the frame thread to stop
	status = PV_UseDirectSound();

#ifndef NO_DIRECT_SOUND

	if (status)
	{	// DirectSound
		// Did we create the SoundMon thread?
		if (g_hthreadSoundMon)
		{
			// Signal thread to stop generating audio
			g_SPD.bWaveEnd = TRUE;

			// wake up threads
			SetEvent(g_heventProcessFrameStart);
			SetEvent(g_heventProcessFrameFinish);

			// Start by writing an empty block
			PV_ClearDirectSoundBuffer();
			// wait for thread to end and only wait 500 ms
			count = WaitForSingleObject(g_hthreadSoundMon, 500);
			// If thread takes too long, kill it
			if (count == WAIT_TIMEOUT)
			{
				TerminateThread(g_hthreadSoundMon, 0);
				g_hthreadSoundMon = NULL;
			}
			CloseHandle(g_hthreadSoundMon);
			g_hthreadSoundMon = NULL;
		}
		PV_CleanupDirectSound();
	}

#endif

	HAE_DestroyFrameThread(context);

	// use DirectSound or waveOut API
	if (status == 0)
	{	// waveOut
		if (g_waveDevice)
		{
			waveOutReset(g_waveDevice);
			while (waveOutClose(g_waveDevice) == WAVERR_STILLPLAYING)
			{
				waveOutReset(g_waveDevice);
			}
			g_waveDevice = 0;
		}
		for (count = 0; count < HAE_WAVEOUT_FRAMES_PER_BLOCK; count++)
		{
			HAE_Deallocate(g_audioBufferBlock[count]);
			g_audioBufferBlock[count] = NULL;
		}
	}

	return 0;
}

// return device position in samples
unsigned long HAE_GetDeviceSamplesPlayedPosition(void)
{
	int				status, error, frameSize;
	MMTIME			audioStatus;
	unsigned long	pos;
	unsigned long	dwPlayPos, dwWritePos;

	pos = 0;
	frameSize = g_waveFormat.nChannels * g_waveFormat.wBitsPerSample / 8;
	status = PV_UseDirectSound();

#ifndef NO_DIRECT_SOUND


	if (status)
	{	// DirectSound
		IDirectSoundBuffer_GetCurrentPosition(g_pDirectSoundBuffer, &dwPlayPos, &dwWritePos);
		// return in samples
		pos = dwPlayPos / frameSize;
	}
	else

#endif

	{	// waveOut
		memset(&audioStatus, 0, (long)sizeof(MMTIME));
		audioStatus.wType = TIME_BYTES;	// get byte position

		error = waveOutGetPosition(g_waveDevice, &audioStatus, sizeof(MMTIME));
		// return in samples
		pos = audioStatus.u.cb / frameSize;
	}
	return pos;
}

// number of devices. ie different versions of the HAE connection. DirectSound and waveOut
// return number of devices. ie 1 is one device, 2 is two devices.
// NOTE: This function needs to function before any other calls may have happened.
long HAE_MaxDevices(void)
{
	return 2;	// waveOut and DirectSound
}

// set the current device. device is from 0 to HAE_MaxDevices()
// NOTE:	This function needs to function before any other calls may have happened.
//			Also you will need to call HAE_ReleaseAudioCard then HAE_AquireAudioCard
//			in order for the change to take place.
void HAE_SetDeviceID(long deviceID)
{
	if (deviceID < HAE_MaxDevices())
	{
		g_currentDeviceID = deviceID;
	}
}

// return current device ID
// NOTE: This function needs to function before any other calls may have happened.
long HAE_GetDeviceID(void)
{
	return g_currentDeviceID;
}

// NOTE:	This function needs to function before any other calls may have happened.
//			Format of string is a zero terminated comma delinated C string.
//			"platform,method,misc"
//	example	"MacOS,Sound Manager 3.0,SndPlayDoubleBuffer"
//			"WinOS,DirectSound,multi threaded"
//			"WinOS,waveOut,multi threaded"
//			"WinOS,VxD,low level hardware"
//			"WinOS,plugin,Director"
void HAE_GetDeviceName(long deviceID, char *cName, unsigned long cNameLength)
{
	char		*data;
	static char *names[] =
			{	"WinOS,waveOut,multi threaded",
				"WinOS,DirectSound,multi threaded"
			};
	if (cName && cNameLength)
	{
		if (deviceID < HAE_MaxDevices())
		{
			// can't do string functions here, it might be bad to do.
			data = names[deviceID];
			while (*data && (cNameLength > 0))
			{
				*cName = *data;
				cName++;
				data++;
				cNameLength--;
			}
			*cName = 0;
		}
	}
}


// EOF of HAE_API_WinOS.c
