/*
 * @(#)awt_Dialog.cpp	1.48 98/10/05
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "awt_Toolkit.h"
#include "awt_Dialog.h"

#include "java_awt_Dialog.h"

/* IMPORTANT! Read the README.JNI file for notes on JNI converted AWT code.
 */

/************************************************************************
 * AwtDialog fields
 */

jfieldID AwtDialog::titleID;

#if defined(DEBUG)
// counts how many nested modal dialogs are open, a sanity
// check to ensure the somewhat complicated disable/enable
// code is working properly
int AwtModalityNestCounter = 0;
#endif
// property name tagging windows disabled by modality
static const char * ModalDisableProp = "SunAwtModalDisableProp";

/************************************************************************
 * AwtDialog class methods
 */

AwtDialog::AwtDialog() {
    m_modalWnd = NULL;
}

AwtDialog::~AwtDialog() {
    if (m_modalWnd != NULL) {
	WmEndModal();
    }
}

const char* AwtDialog::GetClassName() {
  return "SunAwtDialog";
}

// Override RegisterClass so dialogs do not have icon by default
void AwtDialog::RegisterClass()
{
    const char* className = GetClassName();
    WNDCLASS  wc;

    if (!::GetClassInfo(::GetModuleHandle(NULL), className, &wc)) {
        wc.style         = CS_OWNDC | CS_SAVEBITS; 
	wc.lpfnWndProc   = (WNDPROC)::DefWindowProc;
	wc.cbClsExtra    = 0;
	wc.cbWndExtra    = DLGWINDOWEXTRA;
	wc.hInstance     = ::GetModuleHandle(NULL);
	wc.hIcon         = NULL; // Dialogs do not have an icon by default
	wc.hCursor       = NULL;
	wc.hbrBackground = NULL;
	wc.lpszMenuName  = NULL;
	wc.lpszClassName = className;

        ATOM ret = ::RegisterClass(&wc);
        ASSERT(ret != 0);
    }
}

/*
 * Create a new AwtDialog object and window.   
 */
AwtDialog* AwtDialog::Create(jobject peer, jobject parent) 
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    if (env->EnsureLocalCapacity(2) < 0) {
	return NULL;
    }
    int pData;
    jobject target = env->GetObjectField(peer, AwtObject::targetID);
    HWND hwndParent = NULL;
    AwtDialog* dialog;
    jobject background = 0;

    JNI_CHECK_NULL_GOTO(target, "null target", done);

    if (parent != NULL) {
	JNI_CHECK_PEER_GOTO(parent, done);
	{
	    AwtFrame* awtParent = (AwtFrame *)(JNI_GET_PDATA(parent));
	    hwndParent = awtParent->GetHWnd();
	}
    }
    dialog = new AwtDialog();
    JNI_CHECK_NULL_GOTO(dialog, "AwtDialog() failed", done);
    {
        int colorId = IS_WIN4X ? COLOR_3DFACE : COLOR_WINDOW;
	DWORD style = WS_POPUP | WS_CAPTION | WS_SYSMENU | WS_CLIPCHILDREN;
	style &= ~(WS_MINIMIZEBOX|WS_MAXIMIZEBOX);
	DWORD exStyle = IS_WIN4X ? WS_EX_WINDOWEDGE | WS_EX_DLGMODALFRAME : 0;

	if (GetRTL()) {
	    exStyle |= WS_EX_RIGHT | WS_EX_LEFTSCROLLBAR;
	    if (GetRTLReadingOrder())
	        exStyle |= WS_EX_RTLREADING;
	}

	jint x = env->GetIntField(target, AwtComponent::xID);
	jint y = env->GetIntField(target, AwtComponent::yID);
	jint width = env->GetIntField(target, AwtComponent::widthID);
	jint height = env->GetIntField(target, AwtComponent::heightID);

	dialog->CreateHWnd(env, "",
			   style, exStyle,
			   x, y, width, height,
			   hwndParent,
			   NULL,
			   ::GetSysColor(COLOR_WINDOWTEXT),
			   ::GetSysColor(colorId),
			   peer);

	dialog->RecalcNonClient();
	dialog->UpdateDialogIcon();
	dialog->UpdateSystemMenu();

	background = env->GetObjectField(target, AwtComponent::backgroundID);
	if (background == NULL) {
	    JNU_CallMethodByName(env, NULL, peer, "setDefaultColor", "(Z)V",
				 (IS_WIN4X != 0));
	}
    }
 done:
    env->DeleteLocalRef(background);
    env->DeleteLocalRef(target);

    return dialog;
}

//
// Disables top-level windows for a modal dialog
//	hWndDlg		Modal dialog 
//
void AwtDialog::ModalDisable(HWND hWndDlg)
{
#if defined(DEBUG)
    ASSERT(AwtModalityNestCounter >= 0 && AwtModalityNestCounter <= 1000); // sanity check
    AwtModalityNestCounter++;
#endif
    AwtDialog::ResetDisabledLevel(hWndDlg);
    ::EnumThreadWindows(AwtToolkit::MainThread(),
		    (WNDENUMPROC)DisableTopLevelsCallback,(LPARAM)hWndDlg) ;
}

// 
// Re-enables top-level windows for a modal dialog
//	hWndDlg		Modal dialog
//
void AwtDialog::ModalEnable(HWND hWndDlg)
{
#if defined(DEBUG)
    ASSERT(AwtModalityNestCounter > 0 && AwtModalityNestCounter <= 1000); // sanity check
    AwtModalityNestCounter--;
#endif
    int disabledLevel = AwtDialog::GetDisabledLevel(hWndDlg);
    ::EnumThreadWindows(AwtToolkit::MainThread(),
		    (WNDENUMPROC)EnableTopLevelsCallback, (LPARAM)disabledLevel);
    AwtDialog::ResetDisabledLevel(hWndDlg);
}

//
// Brings a window to foreground when modal dialog goes away
// (the Win32 Z-order is very screwy if another app is active
//  when the modal dialog goes away, so we explicitly activate 
//  the first enabled, visible, window we find)
//
//	hWndParent	Parent window of the dialog
//
void AwtDialog::ModalNextWindowToFront(HWND hWndParent)
{
    if ( ::IsWindowEnabled(hWndParent) && ::IsWindowVisible(hWndParent) ) {
    // always give parent first shot at coming to the front
	::SetForegroundWindow(hWndParent);
    } else {
    // otherwise, activate the first enabled, visible window we find
	::EnumThreadWindows(AwtToolkit::MainThread(),
		    (WNDENUMPROC)NextWindowToFrontCallback, 0L );
    }
}

//
// Disables top-level windows, incrementing a counter on each disabled
// window to indicate how many nested modal dialogs need the window disabled
//
BOOL CALLBACK AwtDialog::DisableTopLevelsCallback(HWND hWnd, LPARAM lParam)
{
    HWND	hWndDlg = (HWND)lParam;

    if (hWnd == hWndDlg || hWnd == AwtToolkit::GetInstance().GetHWnd()) {
	return TRUE; // ignore myself and toolkit when disabling
    }

    // the drop-down list portion of combo-boxes are actually top-level
    // windows with WS_CHILD, so ignore these guys...
    if ( (::GetWindowLong(hWnd, GWL_STYLE) & WS_CHILD) != 0) {
	return TRUE;
    }

    ASSERT(hWndDlg == NULL || IsWindow(hWndDlg));

    AwtDialog::IncrementDisabledLevel(hWnd, 1);
    return TRUE;
}

//
// Enables top-level window if no modal dialogs need it disabled
//
BOOL CALLBACK AwtDialog::EnableTopLevelsCallback(HWND hWnd, LPARAM lParam)
{
    HWND	hWndDlg = (HWND)lParam;
    int		disabledLevel = AwtDialog::GetDisabledLevel(hWnd);
    int		dlgDisabledLevel = (int)lParam;

    if (disabledLevel == 0 || disabledLevel  < dlgDisabledLevel) {
    // skip enabled windows or windows disabled after the 
    // given modal dialog became modal
	return TRUE;
    }

    AwtDialog::IncrementDisabledLevel(hWnd, -1);
    return TRUE;
}

//
// Brings the first enabled, visible window to the foreground
//
BOOL CALLBACK AwtDialog::NextWindowToFrontCallback(HWND hWnd, LPARAM lParam)
{
    if ( ::IsWindowVisible(hWnd) && ::IsWindowEnabled(hWnd) ) {
	::SetForegroundWindow(hWnd);
	return FALSE; // stop callbacks
    }
    return TRUE;
}

int AwtDialog::GetDisabledLevel(HWND hWnd) {
    int	disabledLevel = (int)::GetProp(hWnd, ModalDisableProp);
    ASSERT(disabledLevel >= 0 && disabledLevel <= 1000);
    return disabledLevel;
}

#if 0 // defined(DEBUG)
// debugging-only function; useful for tracking down disable/enable problems
static void DbgShowDisabledLevel(HWND hWnd, int disabledLevel)
{    
    // put disabled counter in title bar text
    char	szDisabled[256];
    wsprintf(szDisabled, "disabledLevel = %d", disabledLevel);
    ::SetWindowText(hWnd, szDisabled);
}
    #define DBG_SHOW_DISABLED_LEVEL( _win, _dl ) \
	    DbgShowDisabledLevel( (_win), (_dl) )
#else    
    #define DBG_SHOW_DISABLED_LEVEL( _win, _dl )
#endif

void AwtDialog::IncrementDisabledLevel(HWND hWnd, int increment) {
    int	disabledLevel = GetDisabledLevel(hWnd);
    ASSERT( increment == -1 || increment == 1); // only allow increment/decrement
    disabledLevel += increment;
    if (disabledLevel == 0 ) {
	// don't need the property any more...
	ResetDisabledLevel(hWnd);
    } else {
	::SetProp(hWnd, ModalDisableProp, (HANDLE)disabledLevel);
    }
    ::EnableWindow(hWnd, disabledLevel == 0);
    DBG_SHOW_DISABLED_LEVEL(hWnd, GetDisabledLevel(hWnd));
}

void AwtDialog::ResetDisabledLevel(HWND hWnd) {
    ::RemoveProp(hWnd, ModalDisableProp);
    DBG_SHOW_DISABLED_LEVEL(hWnd, GetDisabledLevel(hWnd));
}

MsgRouting AwtDialog::WmShowModal()
{
    ASSERT(::GetCurrentThreadId() == AwtToolkit::MainThread());

    // disable top-level windows
    AwtDialog::ModalDisable( GetHWnd() );
    // enable myself (could have been explicitly disabled via setEnabled)
    ::EnableWindow(GetHWnd(), TRUE); // is this the right behaviour? (rnk 6/4/98)
    // show window and set focus to it
    SendMessage(WM_AWT_COMPONENT_SHOW, SW_SHOW);
    // normally done by DefWindowProc(WM_ACTIVATE),
    // but we override that message so we do it explicitly
    ::SetFocus(GetHWnd());

    m_modalWnd = GetHWnd();

    return mrConsume;
}

MsgRouting AwtDialog::WmEndModal()
{
    ASSERT( ::GetCurrentThreadId() == AwtToolkit::MainThread() );
    ASSERT( ::IsWindowVisible( GetHWnd() ) );
    ASSERT( ::IsWindow(m_modalWnd) );

    // re-enable top-level windows
    AwtDialog::ModalEnable( GetHWnd() );
    
    HWND	hWndParent = ::GetParent( GetHWnd() );
    BOOL	isEnabled = ::IsWindowEnabled(hWndParent);

    // Need to temporarily enable the dialog's owner window.
    // If we don't enable the owner, Windows will activate
    // the last active top-level window, which could be
    // another application entirely (the dreaded bug 4065506).
    ::EnableWindow(hWndParent, TRUE);
    // hide the dialog
    SendMessage(WM_AWT_COMPONENT_SHOW, SW_HIDE);
    // restore the owner's true enabled state
    ::EnableWindow(hWndParent, isEnabled);

    // bring the next window in the stack to the front
    AwtDialog::ModalNextWindowToFront( hWndParent );
    m_modalWnd = NULL;

    return mrConsume;
}

void AwtDialog::SetResizable(BOOL isResizable)
{
    // call superclass
    AwtFrame::SetResizable(isResizable);

    long    style = GetStyle();
    long    xstyle = GetStyleEx();
    if (isResizable) {
    // remove modal frame
	xstyle &= ~WS_EX_DLGMODALFRAME;
    } else {
    // add modal frame
	xstyle |= WS_EX_DLGMODALFRAME;
    }
    // dialogs are never minimizable/maximizable, so remove those bits
    style &= ~(WS_MINIMIZEBOX|WS_MAXIMIZEBOX);
    SetStyle(style);
    SetStyleEx(xstyle);
    RedrawNonClient();
}

// Adjust system menu so that:
//  Non-resizable dialogs only have Move and Close items
//  Resizable dialogs also have Size
// Normally, Win32 dialog system menu handling is done via
// CreateDialog/DefDlgProc, but our dialogs are using DefWindowProc
// so we handle the system menu ourselves
// fixes 4057529 - robi.khan@eng
void AwtDialog::UpdateSystemMenu()
{
    HWND    hWndSelf = GetHWnd();
    BOOL    isResizable = ((GetStyle() & WS_THICKFRAME) != 0);
    BOOL    isMaximized = ::IsZoomed(hWndSelf);

    // before restoring the default menu, check if there is an
    // InputMethodManager menu item already.  Note that it assumes
    // that the length of the InputMethodManager menu item string
    // should not be longer than 256 bytes.
    MENUITEMINFO  mii;
    char          immItem[256];
    BOOL          hasImm;
    mii.cbSize = sizeof(MENUITEMINFO);
    mii.fMask = MIIM_TYPE;
    mii.cch = sizeof(immItem);
    mii.dwTypeData = immItem;
    hasImm = ::GetMenuItemInfo(GetSystemMenu(hWndSelf, FALSE),
                               SYSCOMMAND_IMM, FALSE, &mii);

    // restore the default menu
    ::GetSystemMenu(hWndSelf, TRUE);
    // now get a working copy of the menu
    HMENU   hMenuSys = GetSystemMenu(hWndSelf, FALSE);
    // remove inapplicable sizing commands
    ::DeleteMenu(hMenuSys, SC_MINIMIZE, MF_BYCOMMAND);
    ::DeleteMenu(hMenuSys, SC_RESTORE, MF_BYCOMMAND);
    ::DeleteMenu(hMenuSys, SC_MAXIMIZE, MF_BYCOMMAND);
    if (!isResizable) {
	::DeleteMenu(hMenuSys, SC_SIZE, MF_BYCOMMAND);
    }
    // remove separator if only 3 items left (Move, Separator, and Close)
    if (::GetMenuItemCount(hMenuSys) == 3) {
	MENUITEMINFO	mi;
	mi.cbSize = sizeof(MENUITEMINFO);
	mi.fMask = MIIM_TYPE;
	::GetMenuItemInfo(hMenuSys, 1, TRUE, &mi);
	if (mi.fType & MFT_SEPARATOR) {
	    ::DeleteMenu(hMenuSys, 1, MF_BYPOSITION);
	}
    }

    // if there was the InputMethodManager menu item, restore it.
    if (hasImm)
        ::AppendMenu(hMenuSys, MF_STRING, SYSCOMMAND_IMM, immItem);
}

// Call this when dialog isResizable property changes, to 
// hide or show the small icon in the upper left corner
void AwtDialog::UpdateDialogIcon()
{
    BOOL    isResizable = ((GetStyle() & WS_THICKFRAME) != 0);
    HWND    hWndOwner = GetWindow(GetHWnd(), GW_OWNER);
    
    if (isResizable) {
    // if we're resizable, use the icon from our owner window
	HICON	hIconOwnerBig;
	HICON	hIconOwnerSmall;
	HICON	hIconClassBig;
	HICON	hIconClassSmall;
	HICON	hIconDialog;

	// try getting owner's small icon
	hIconOwnerSmall = (HICON)::SendMessage(hWndOwner, WM_GETICON, ICON_SMALL, NULL);
        hIconOwnerBig = (HICON)::SendMessage(hWndOwner, WM_GETICON, ICON_BIG, NULL);
	hIconClassBig = (HICON)::GetClassLong(hWndOwner, GCL_HICON);
	hIconClassSmall= (HICON)::GetClassLong(hWndOwner, GCL_HICONSM);
	if (hIconOwnerSmall != NULL) {
	    hIconDialog = hIconOwnerSmall;
	} else if (hIconOwnerBig != NULL) {
	    hIconDialog = hIconOwnerBig;
	} else if (hIconClassSmall != NULL) {
	    hIconDialog = hIconClassSmall;
	} else {
	    hIconDialog = hIconClassBig;
	}
	SendMessage(WM_SETICON, ICON_SMALL, (LPARAM)hIconDialog);
	SendMessage(WM_SETICON, ICON_BIG, (LPARAM)hIconDialog);
    } else {
    // if we're not sizable remove the icon
	SendMessage(WM_SETICON, ICON_SMALL, NULL);
	SendMessage(WM_SETICON, ICON_BIG, NULL);
    }
}

// Override WmStyleChanged to adjust system menu for sizable/non-resizable dialogs
MsgRouting AwtDialog::WmStyleChanged(WORD wStyleType, LPSTYLESTRUCT lpss)
{
    UpdateSystemMenu();
    UpdateDialogIcon();
    return mrConsume;
}

MsgRouting AwtDialog::WmSize(int type, int w, int h)
{
    UpdateSystemMenu(); // adjust to reflect restored vs. maximized state
    return AwtFrame::WmSize(type, w, h);
}

LRESULT AwtDialog::WindowProc(UINT message, WPARAM wParam, LPARAM lParam)
{
    MsgRouting mr = mrDoDefault;
    LRESULT retValue = 0L;

    switch(message) {
	case WM_AWT_DLG_SHOWMODAL:
	    mr = WmShowModal();
	    break;
	case WM_AWT_DLG_ENDMODAL:
	    mr = WmEndModal();
	    break;
    }

    if (mr != mrConsume) {
        retValue = AwtFrame::WindowProc(message, wParam, lParam);
    }
    return retValue;
}

/************************************************************************
 * Dialog native methods
 */

extern "C" {

JNIEXPORT void JNICALL
Java_java_awt_Dialog_initIDs(JNIEnv *env, jclass cls)
{
    AwtDialog::titleID = env->GetFieldID(cls, "title", "Ljava/lang/String;");
}

} /* extern "C" */


/************************************************************************
 * DialogPeer native methods
 */

extern "C" {

/*
 * Class:     sun_awt_windows_WDialogPeer
 * Method:    create
 * Signature: (Lsun/awt/windows/WComponentPeer;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WDialogPeer_create(JNIEnv *env, jobject self,
					jobject parent) 
{
    int pData;
    AwtToolkit::CreateComponent(self, parent, 
				(AwtToolkit::ComponentFactory)
				AwtDialog::Create);
    JNI_CHECK_PEER_CREATION_RETURN(self);
}

/*
 * Class:     sun_awt_windows_WDialogPeer
 * Method:    _show
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WDialogPeer_showModal(JNIEnv *env, jobject self)
{
    int pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtDialog* d = (AwtDialog*)pData;
    d->SendMessage(WM_AWT_DLG_SHOWMODAL);
}

/*
 * Class:     sun_awt_windows_WDialogPeer
 * Method:    _hide
 * Signature: ()V
 */
JNIEXPORT void JNICALL 
Java_sun_awt_windows_WDialogPeer_endModal(JNIEnv *env, jobject self)
{
    int pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtDialog* d = (AwtDialog*)pData;
    d->SendMessage(WM_AWT_DLG_ENDMODAL);
}

/*
 * Class:     sun_awt_windows_WFramePeer
 * Method:    pSetIMMOption
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WDialogPeer_pSetIMMOption(JNIEnv *env, jobject self,
					       jstring option) 
{
    jint pData;
    char *coption;
    char *empty = "InputMethod";
    JNI_CHECK_PEER_RETURN(self);
    JNI_CHECK_NULL_RETURN(option, "IMMOption argument");

    AwtDialog* dialog = (AwtDialog *)pData;

    coption = (JNU_IsNull(env, option)) ? empty : (char *) JNU_GetStringPlatformChars(env, option, NULL);
    HMENU hSysMenu = ::GetSystemMenu(dialog->GetHWnd(), FALSE);
    ::AppendMenu(hSysMenu,  MF_STRING, SYSCOMMAND_IMM, coption);

    if (coption != empty)
        JNU_ReleaseStringPlatformChars(env, option, (const char *) coption);
}
} /* extern "C" */




