/*
 * @(#)awt_Font.cpp	1.40 98/10/02
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "jlong.h"
#include "awt_Font.h"
#include "awt_Toolkit.h"

#include "java_awt_Font.h"
#include "java_awt_FontMetrics.h"
#include "java_awt_Dimension.h"

#include "sun_awt_CharsetString.h"
#include "sun_awt_FontDescriptor.h"
#include "sun_awt_windows_WDefaultFontCharset.h"
#include "sun_awt_windows_WFontPeer.h"
#include "awt_Component.h"

/* IMPORTANT! Read the README.JNI file for notes on JNI converted AWT code.
 */

AwtFontCache fontCache;

extern jboolean IsMultiFont(JNIEnv *env, jobject obj) 
{
    if (obj == NULL) {
	return JNI_FALSE;
    }
    if (env->EnsureLocalCapacity(2))
        return JNI_FALSE;
    jobject peer = env->CallObjectMethod(obj, AwtFont::peerMID);
    if (peer == NULL) {
	return JNI_FALSE;
    }
    jobject props = env->GetObjectField(peer, AwtFont::propsID);
    jboolean result = props != NULL;
    env->DeleteLocalRef(peer);
    env->DeleteLocalRef(props);
    return result;
}

extern jstring GetFontProperty(JNIEnv *env, jobject font, jstring key) 
{
    /* should we be checking for nulls ? -br */
    jobject peer = env->CallObjectMethod(font, AwtFont::peerMID);
    ASSERT(peer != NULL);
    jobject props = env->GetObjectField(peer, AwtFont::propsID);
    ASSERT(props != NULL);
    jstring result = (jstring)JNU_CallMethodByName(
	env, NULL, props, "getProperty", 
	"(Ljava/lang/String;)Ljava/lang/String;", key).l;
    env->DeleteLocalRef(peer);
    env->DeleteLocalRef(props);
    return result;
}

/************************************************************************
 * AwtFont fields
 */

/* java.awt.Font class */
jclass AwtFont::jaFontCls;

/* sun.awt.windows.WFontMetrics fields */
jfieldID AwtFont::widthsID;
jfieldID AwtFont::ascentID;
jfieldID AwtFont::descentID;
jfieldID AwtFont::leadingID;
jfieldID AwtFont::heightID;
jfieldID AwtFont::maxAscentID;
jfieldID AwtFont::maxDescentID;
jfieldID AwtFont::maxHeightID;
jfieldID AwtFont::maxAdvanceID;

/* sun.awt.CharsetString fields */
jfieldID AwtFont::fontDescriptorID;
jfieldID AwtFont::charsetStringID;

/* java.awt.FontDescriptor fields */
jfieldID AwtFont::fontCharsetID;
jfieldID AwtFont::nativeNameID;

/* java.awt.Font fields */
jfieldID AwtFont::pDataID;
jfieldID AwtFont::nameID;
jfieldID AwtFont::sizeID;
jfieldID AwtFont::styleID;

/* java.awt.FontMetrics fields */
jfieldID AwtFont::fontID;

/* sun.awt.PlatformFont fields */
jfieldID AwtFont::propsID;
jfieldID AwtFont::componentFontsID;

/* sun.awt.windows.WDefaultFontCharset fields */
jfieldID AwtFont::fontNameID;

/* java.awt.Font methods */
jmethodID AwtFont::peerMID;

/* sun.awt.PlatformFont methods */
jmethodID AwtFont::makeMultiCharsetStringMID;
jmethodID AwtFont::makeConvertedMultiFontStringMID;

/* sun.awt.PlatformFont methods */
jmethodID AwtFont::getFontMID;

/* java.awt.FontMetrics methods */
jmethodID AwtFont::getHeightMID;


/************************************************************************
 * AwtFont methods 
 */

AwtFont::AwtFont(int num)
{
    if (num == 0) {  // not multi-font
        num = 1;
    }

    m_hFontNum = num;
    m_hFont = new HFONT[num];
    m_needConvert = new BOOL[num];
    for (int i = 0; i < num; i++) {
        m_hFont[i] = NULL;
        m_needConvert[i] = FALSE;
    }  

    m_textInput = -1;
    m_ascent = -1;
    m_overhang = 0;
}

AwtFont::~AwtFont()
{
    for (int i = 0; i < m_hFontNum; i++) {
        HFONT font = m_hFont[i];
        if (font != NULL && fontCache.Search(font)) {
            fontCache.Remove(font);
            /*  NOTE: delete of windows HFONT happens in FontCache::Remove
                      only when the final reference to the font is disposed */
        }
    }
    delete[] m_hFont;
    delete[] m_needConvert;
}

AwtFont* AwtFont::GetFont(JNIEnv *env, jobject font)
{
    jlong pData = env->GetLongField(font, AwtFont::pDataID);
    AwtFont* awtFont = (AwtFont*)jlong_to_ptr(pData);

    if (awtFont != NULL) {
	return awtFont;
    }

    awtFont = Create(env, font);

    env->SetLongField(font, AwtFont::pDataID, (long)awtFont);
    return awtFont;
}

AwtFont* AwtFont::Create(JNIEnv *env, jobject font)
{
    int fontSize = env->GetIntField(font, AwtFont::sizeID);
    int fontStyle = env->GetIntField(font, AwtFont::styleID);

    AwtFont* awtFont = NULL;
    jobjectArray compFont;
    int cfnum;

    if (env->EnsureLocalCapacity(3) < 0)
        return 0;

    if (IsMultiFont(env, font)) {
        compFont = GetComponentFonts(env, font);
        cfnum = env->GetArrayLength(compFont);
    } else {
        compFont = NULL;
        cfnum = 0;
    }


    WCHAR* wName;

    awtFont = new AwtFont(cfnum);
    if (cfnum > 0) {
        // Ask java.awt.Font for the input charset property.
	jstring key = env->NewStringUTF("inputtextcharset");
	jstring jtextin = GetFontProperty(env, font, key);
        WCHAR* textin = TO_WSTRING(jtextin);

	for (int i = 0; i < cfnum; i++) { 
            // nativeName is a pair of platform fontname and its charset
            // tied with a comma; "Times New Roman,ANSI_CHARSET".            
	    jobject fontDescriptor = env->GetObjectArrayElement(compFont, i);
	    jstring nativeName = 
		(jstring)env->GetObjectField(fontDescriptor, 
					     AwtFont::nativeNameID);
            wName = TO_WSTRING(nativeName);
	    ASSERT(wName);
            // Check to see if this font is suitable for input
	    // on AwtTextComponent?
            if ((textin != NULL) && (wcsstr(wName, textin) != NULL)) {
		awtFont->m_textInput = i;
	    }
	    HFONT hfonttmp = CreateHFont(wName, fontStyle, fontSize);
            awtFont->m_hFont[i] = hfonttmp;

            if (wcsstr(wName, L"NEED_CONVERTED"))
                awtFont->m_needConvert[i] = TRUE;
            else
                awtFont->m_needConvert[i] = FALSE;
	    env->DeleteLocalRef(fontDescriptor);
	    env->DeleteLocalRef(nativeName);
        }
	if (textin == NULL) {
	    // no entry "inputtextcharset" in font.properties or
	    // CHARSET for "inputtextcharset" is different from any CHARSET
	    // for Native font
	    awtFont->m_textInput = 0;
	}
    } else { 
        // Instantiation for English version.
	jstring fontName = (jstring)env->GetObjectField(font, AwtFont::nameID);
        wName = TO_WSTRING(fontName);

        WCHAR* wEName;
        if (!wcscmp(wName, L"Helvetica") || !wcscmp(wName, L"SansSerif")) {
	    wEName = L"Arial";
        } else if (!wcscmp(wName, L"TimesRoman") || !wcscmp(wName, L"Serif")) {
	    wEName = L"Times New Roman";
        } else if (!wcscmp(wName, L"Courier") || !wcscmp(wName, L"Monospaced")) {
	    wEName = L"Courier New";
        } else if (!wcscmp(wName, L"Dialog")) {
	    wEName = L"MS Sans Serif";
        } else if (!wcscmp(wName, L"DialogInput")) {
	    wEName = L"MS Sans Serif";
        } else if (!wcscmp(wName, L"ZapfDingbats")) {
	    wEName = L"WingDings";
        } else
            wEName = L"Arial";

	awtFont->m_textInput = 0;
        awtFont->m_hFont[0] = CreateHFont(wEName, fontStyle, fontSize);
	env->DeleteLocalRef(fontName);
    } 
    env->SetLongField(font, AwtFont::pDataID, (long)awtFont);
    env->DeleteLocalRef(compFont);
    return awtFont;
}


// Get suitable CHARSET from charset string in properties file.
static int GetNativeCharset(WCHAR* name)
{
    if (wcsstr(name, L"ANSI_CHARSET"))
        return ANSI_CHARSET;
    if (wcsstr(name, L"DEFAULT_CHARSET"))
        return DEFAULT_CHARSET;
    if (wcsstr(name, L"SYMBOL_CHARSET") || wcsstr(name, L"WingDings"))
        return SYMBOL_CHARSET;
    if (wcsstr(name, L"SHIFTJIS_CHARSET"))
        return SHIFTJIS_CHARSET;
    if (wcsstr(name, L"GB2312_CHARSET"))
        return GB2312_CHARSET;
    if (wcsstr(name, L"HANGEUL_CHARSET"))
        return HANGEUL_CHARSET;
    if (wcsstr(name, L"CHINESEBIG5_CHARSET"))
        return CHINESEBIG5_CHARSET;
    if (wcsstr(name, L"OEM_CHARSET"))
        return OEM_CHARSET;
    if (wcsstr(name, L"JOHAB_CHARSET"))
        return JOHAB_CHARSET;
    if (wcsstr(name, L"HEBREW_CHARSET"))
        return HEBREW_CHARSET;
    if (wcsstr(name, L"ARABIC_CHARSET"))
        return ARABIC_CHARSET;
    if (wcsstr(name, L"GREEK_CHARSET"))
        return GREEK_CHARSET;
    if (wcsstr(name, L"TURKISH_CHARSET"))
        return TURKISH_CHARSET;
    if (wcsstr(name, L"VIETNAMESE_CHARSET"))
        return VIETNAMESE_CHARSET;
    if (wcsstr(name, L"THAI_CHARSET"))
        return THAI_CHARSET;
    if (wcsstr(name, L"EASTEUROPE_CHARSET"))
        return EASTEUROPE_CHARSET;
    if (wcsstr(name, L"RUSSIAN_CHARSET"))
        return RUSSIAN_CHARSET;
    if (wcsstr(name, L"MAC_CHARSET"))
        return MAC_CHARSET;
    if (wcsstr(name, L"BALTIC_CHARSET"))
        return BALTIC_CHARSET;
    return ANSI_CHARSET;
}

static HFONT CreateHFont_sub(WCHAR* name, int style, int height)
{
    LOGFONTW logFont;

    logFont.lfWidth = 0;
    logFont.lfEscapement = 0;
    logFont.lfOrientation = 0;
    logFont.lfUnderline = FALSE;
    logFont.lfStrikeOut = FALSE;
    logFont.lfCharSet = GetNativeCharset(name);
    logFont.lfOutPrecision = OUT_DEFAULT_PRECIS;
    logFont.lfClipPrecision = CLIP_DEFAULT_PRECIS;
    logFont.lfQuality = DEFAULT_QUALITY;
    logFont.lfPitchAndFamily = DEFAULT_PITCH;

    // Set style
    logFont.lfWeight = (style & java_awt_Font_BOLD) ? FW_BOLD : FW_NORMAL;
    logFont.lfItalic = (style & java_awt_Font_ITALIC) != 0;
    logFont.lfUnderline = 0;//(style & java_awt_Font_UNDERLINE) != 0;

    // Get point size
    logFont.lfHeight = -height;

    // Set font name
    WCHAR tmpname[80];
    wcscpy(tmpname, name);
    WCHAR* delimit = wcschr(tmpname, L',');
    if (delimit != NULL)
        *delimit = L'\0';  // terminate the string after the font name.
    wcscpy(&(logFont.lfFaceName[0]), tmpname);
    HFONT hFont;
    // before wrapper function of CreateFontIndirect is made,
    // we have to distinguish the running environment
    if (IS_NT) {
        hFont = ::CreateFontIndirectW(&logFont);
        ASSERT(hFont != NULL);
    } else {
#ifdef WIN32
        ASSERT(IS_WIN95);
#endif

        LOGFONTA logFontA;
        logFontA.lfHeight           = logFont.lfHeight;
        logFontA.lfWidth            = logFont.lfWidth;
        logFontA.lfEscapement       = logFont.lfEscapement;
        logFontA.lfOrientation      = logFont.lfOrientation;
        logFontA.lfWeight           = logFont.lfWeight;
        logFontA.lfItalic           = logFont.lfItalic;
        logFontA.lfUnderline        = logFont.lfUnderline;
        logFontA.lfStrikeOut        = logFont.lfStrikeOut;
        logFontA.lfCharSet          = logFont.lfCharSet;
        logFontA.lfOutPrecision     = logFont.lfOutPrecision;
        logFontA.lfClipPrecision    = logFont.lfClipPrecision;
        logFontA.lfQuality          = logFont.lfQuality;
        logFontA.lfPitchAndFamily   = logFont.lfPitchAndFamily;
        
        // Some fontnames of Non-ASCII fonts like 'MS Minchou' are themselves
        // Non-ASCII.  They are assumed to be written in Unicode.
        // Hereby, they are converted into platform codeset. 
	LPCWSTR lpszTmp = logFont.lfFaceName;
        int tmplen = wcslen(lpszTmp);

        int mblen = WideCharToMultiByte(CP_ACP, 0,
	    lpszTmp, tmplen, NULL, 0, NULL, NULL);
        char* chartmp = new char[mblen+1];
        chartmp[mblen] = '\0';
        WideCharToMultiByte(CP_ACP, 0,
	    lpszTmp, tmplen, chartmp, mblen, NULL, NULL);
        strcpy(&logFontA.lfFaceName[0], chartmp);
        delete [] chartmp;
        
        // Use CreateFontIndirectA explicitly as CreateFontIndirectW is not
        // available on Windows 95.
        hFont = ::CreateFontIndirectA(&logFontA); 
        ASSERT(hFont != NULL);
    }

    return hFont;
}

HFONT AwtFont::CreateHFont(WCHAR* name, int style, int height)
{
    WCHAR longName[80];
    // 80 > (max face name(=30) + strlen("CHINESEBIG5_CHARSET")
    //       + strlen("NEED_CONVERTED") + and so)
    // longName doesn't have to be printable.  So, it is OK not to convert.
    if (IS_NT) {
        //wsprintfW only works on NT.  See bugid 4123362
        wsprintfW(longName, L"%ls-%d-%d", name, style, height);
    } else {
#ifdef WIN32
        ASSERT(IS_WIN95);
#endif
        swprintf(longName, L"%ls-%d-%d", name, style, height);
    }    

    HFONT hFont = fontCache.Lookup(longName);
    if (hFont != NULL) {
        fontCache.IncRefCount(hFont);
        return hFont;
    }

    hFont = CreateHFont_sub(name, style, height);
    fontCache.Add(longName, hFont);
    return hFont;
} 

void AwtFont::Cleanup()
{
    fontCache.Clear();
}

void AwtFont::SetupAscent(AwtFont* font)
{
    HDC hDC = ::GetDC(0);
    ASSERT(hDC != NULL);
    HGDIOBJ oldFont = ::SelectObject(hDC, font->GetHFont());

    TEXTMETRIC metrics;
    VERIFY(::GetTextMetrics(hDC, &metrics));
    font->SetAscent(metrics.tmAscent);

    ::SelectObject(hDC, oldFont); 
    VERIFY(::ReleaseDC(0, hDC));
}

void AwtFont::LoadMetrics(JNIEnv *env, jobject fontMetrics)
{
    if (env->EnsureLocalCapacity(3) < 0)
        return;
    jintArray widths = env->NewIntArray(256);
    if (widths == NULL) {
        /* no local refs to delete yet. */
	return;
    }
    jobject font = env->GetObjectField(fontMetrics, AwtFont::fontID);
    AwtFont* awtFont = AwtFont::GetFont(env, font);
    
    HDC hDC = ::GetDC(0);
    ASSERT(hDC != NULL);
    
    HGDIOBJ oldFont = ::SelectObject(hDC, awtFont->GetHFont());
    TEXTMETRIC metrics;
    VERIFY(::GetTextMetrics(hDC, &metrics));
    
    awtFont->m_ascent = metrics.tmAscent;
    
    int ascent = metrics.tmAscent;
    int descent = metrics.tmDescent;
    int leading = metrics.tmExternalLeading;
    env->SetIntField(fontMetrics, AwtFont::ascentID, ascent);
    env->SetIntField(fontMetrics, AwtFont::descentID, descent);
    env->SetIntField(fontMetrics, AwtFont::leadingID, leading);
    env->SetIntField(fontMetrics, AwtFont::heightID, metrics.tmAscent + 
		     metrics.tmDescent + leading);
    env->SetIntField(fontMetrics, AwtFont::maxAscentID, ascent); 
    env->SetIntField(fontMetrics, AwtFont::maxDescentID, descent);
    
    int maxHeight =  ascent + descent + leading;
    env->SetIntField(fontMetrics, AwtFont::maxHeightID, maxHeight);

    int maxAdvance = metrics.tmMaxCharWidth;
    env->SetIntField(fontMetrics, AwtFont::maxAdvanceID, maxAdvance);

    awtFont->m_overhang = metrics.tmOverhang;
	
    jint intWidths[256];
    memset(intWidths, 0, 256 * sizeof(int));
    VERIFY(::GetCharWidth(hDC, metrics.tmFirstChar, 
			  min(metrics.tmLastChar, 255), 
			  (int *)&intWidths[metrics.tmFirstChar]));
    env->SetIntArrayRegion(widths, 0, 256, intWidths);
    env->SetObjectField(fontMetrics, AwtFont::widthsID, widths);

    // Get font metrics on remaining fonts (if multifont).
    for (int j = 1; j < awtFont->GetHFontNum(); j++) {
	::SelectObject(hDC, awtFont->GetHFont(j));
	VERIFY(::GetTextMetrics(hDC, &metrics));
	env->SetIntField(fontMetrics, AwtFont::maxAscentID, 
			 max(ascent, metrics.tmAscent));
	env->SetIntField(fontMetrics, AwtFont::maxDescentID,
			 max(descent, metrics.tmDescent));
	env->SetIntField(fontMetrics, AwtFont::maxHeightID,
			 max(maxHeight, metrics.tmAscent + metrics.tmDescent 
			     + metrics.tmExternalLeading));
	env->SetIntField(fontMetrics, AwtFont::maxAdvanceID, 
			 max(maxAdvance, metrics.tmMaxCharWidth));
    }

    VERIFY(::SelectObject(hDC, oldFont));
    VERIFY(::ReleaseDC(0, hDC));
    env->DeleteLocalRef(font);
    env->DeleteLocalRef(widths);    
}

SIZE AwtFont::TextSize(AwtFont* font, int columns, int rows)
{
    HDC hDC = ::GetDC(0);
    ASSERT(hDC != NULL);
    HGDIOBJ oldFont = ::SelectObject(hDC, (font == NULL)
                                           ? ::GetStockObject(SYSTEM_FONT)
                                           : font->GetHFont());

    SIZE size;
    VERIFY(::GetTextExtentPoint(hDC, 
        "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ", 52, &size));

    VERIFY(::SelectObject(hDC, oldFont));
    VERIFY(::ReleaseDC(0, hDC));

    size.cx = size.cx * columns / 52;
    size.cy = size.cy * rows;
    return size;
}

char* AwtFont::javaString2multibyte(JNIEnv *env, jstring jstr, char buffer[], int length)
{
    if (length == 0) {
        buffer[0] = '\0';
    } else {
        const unsigned short *ustr = env->GetStringCritical(jstr, 0);
	if (ustr == 0)
	    return 0;
        int cchMultiByte = WideCharToMultiByte(
            CP_ACP, 0, ustr, length, buffer, length * 2 + 1, NULL, NULL);
	env->ReleaseStringCritical(jstr, ustr);
        ASSERT(cchMultiByte);
        buffer[cchMultiByte] = '\0';
    }

    return buffer;
}

jstring AwtFont::multibyte2javaString(JNIEnv *env, char * str)
{
    int str_len = strlen(str);
    if (str_len == 0) {
	return (jstring)env->NewStringUTF("");
    }
    jchar buffer[256];
    int jstr_len = MultiByteToWideChar(CP_ACP, 0, str, str_len, buffer, 0);
    jchar *jstr;
    int buf_size;
    if (jstr_len < 255) {
	ASSERT(jstr_len > 0);
	buf_size = sizeof(buffer);
        jstr = buffer;
    } else {
        buf_size = (jstr_len + 1) * sizeof(jchar);
        jstr = (jchar *)malloc(buf_size);
	if (jstr == 0) {
	    JNU_ThrowOutOfMemoryError(0,0);
	    return NULL;
	}
    }
    int n = MultiByteToWideChar(CP_ACP, 0, str, str_len, jstr, buf_size);
    ASSERT(n == jstr_len);
    jstring result = env->NewString(jstr, jstr_len);
    if (jstr != buffer)
        free(jstr);
    return result;
}

int AwtFont::getFontDescriptorNumber(JNIEnv *env, jobject font,
				     jobject fontDescriptor)
{
    int i, num;
    jobject refFontDescriptor;
    jobjectArray array;

    if (env->EnsureLocalCapacity(2) < 0)
        return 0;

    if (IsMultiFont(env, font)) {
	array = GetComponentFonts(env, font);
        num = env->GetArrayLength(array);
    } else {
        array = NULL;
        num = 0;
    }
    
    for (i = 0; i < num; i++){
	// Trying to identify the same FontDescriptor by comparing the
	// pointers.  
	refFontDescriptor = env->GetObjectArrayElement(array, i);
	if (env->IsSameObject(refFontDescriptor, fontDescriptor)) {
	    env->DeleteLocalRef(refFontDescriptor);
	    env->DeleteLocalRef(array);
	    return i;
	}
	env->DeleteLocalRef(refFontDescriptor);
    }
    env->DeleteLocalRef(array);
    return 0;	// Not found.  Use default.
}

/*
 * This is a faster version of the same function, which does most of 
 * the work in Java.
 */
SIZE  AwtFont::DrawStringSize_sub(jstring str, HDC hDC, 
				  jobject font, long x, long y, BOOL draw,
				  UINT codePage)
{
    SIZE size, temp;
    size.cx = size.cy = 0;

    if (str == NULL) {
	return size;
    }
  
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    if (env->EnsureLocalCapacity(3) < 0)
        return size;
    jobjectArray array = 0;

    int arrayLength = 0;

    if (env->GetStringLength(str) == 0) {
        return size;
    }
    if (IsMultiFont(env, font)) {
	jobject peer = env->CallObjectMethod(font, AwtFont::peerMID);
	array =  (jobjectArray)(env->CallObjectMethod(
	peer, AwtFont::makeConvertedMultiFontStringMID, str));
	ASSERT(!env->ExceptionOccurred());

	if (array != NULL) {
	    arrayLength = env->GetArrayLength(array);
	}
	env->DeleteLocalRef(peer);
    } else {
        array = NULL;
        arrayLength = 0;
    }
    AwtFont* awtFont = AwtFont::GetFont(env, font);
    HFONT oldFont = ::SelectObject(hDC, awtFont->GetHFont());

    if (arrayLength == 0) {
      	int length = env->GetStringLength(str);
        WCHAR* string = TO_WSTRING(str);
        VERIFY(::SelectObject(hDC, awtFont->GetHFont()));
        if (AwtComponent::GetRTLReadingOrder()){
	    if (IS_WIN95) {
	        // Start of conversion Code to fix arabic shaping problems
	        // with unicode support in win 95
	        LPSTR buffer =  (LPSTR) alloca((wcslen(string) + 1) * 2);
		int count = ::WideCharToMultiByte(codePage, 0, string, length,
						  buffer,
						  (wcslen(string) + 1) * 2,
						  NULL, NULL);
		VERIFY(!draw || ::ExtTextOutA(hDC, x, y,  ETO_RTLREADING, NULL,
					      buffer, count, NULL));
		// End Of Conversion Code 
	    } else {
	        VERIFY(!draw || ::ExtTextOutW(hDC, x, y, ETO_RTLREADING, NULL,
					      string, length, NULL));
	    }
	} else {
	    VERIFY(!draw || ::TextOutW(hDC, x, y, string, length));
	}
        VERIFY(::GetTextExtentPoint32W(hDC, string, length, &size));
    } else {
        for (int i = 0; i < arrayLength; i = i + 2) {
	    jobject fontDescriptor = env->GetObjectArrayElement(array, i);
	    if (fontDescriptor == NULL) {
	        break;
	    }

	    jbyteArray convertedBytes = 
	        (jbyteArray)env->GetObjectArrayElement(array, i + 1);
	    if (convertedBytes == NULL) {
	        env->DeleteLocalRef(fontDescriptor);
		break;
	    }

            int fdIndex = getFontDescriptorNumber(env, font, fontDescriptor);
            VERIFY(::SelectObject(hDC, awtFont->GetHFont(fdIndex)));

	    /* 
	     * The strange-looking code that follows this comment is
	     * the result of upstream optimizations. In the array of
	     * alternating font descriptor and buffers, the buffers
	     * contain their length in the first four bytes, a la
	     * Pascal arrays.
	     *
	     * Note: the buffer MUST be unsigned, or VC++ will sign
	     * extend buflen and bad things will happen.
	     */
	    unsigned char* buffer = 
		(unsigned char *)env->GetPrimitiveArrayCritical(convertedBytes,
								0);
	    int buflen = (buffer[0] << 24) | (buffer[1] << 16) | 
		(buffer[2] << 8) | buffer[3];

	    ASSERT(buflen >= 0);

	    /*
	     * the offsetBuffer, on the other hand, must be signed because
	     * TextOutA and GetTextExtentPoint32A expect it.
	     */
	    char* offsetBuffer = (char *)(buffer + 4);

	    VERIFY(!draw || ::TextOutA(hDC, x, y, offsetBuffer, buflen));
	    VERIFY(::GetTextExtentPoint32A(hDC, offsetBuffer, buflen, &temp));
	    env->ReleasePrimitiveArrayCritical(convertedBytes, buffer, 0);
	    
            x += temp.cx;
            size.cx += temp.cx;
            size.cy = (size.cy < temp.cy) ? temp.cy : size.cy;
	    env->DeleteLocalRef(fontDescriptor);
	    env->DeleteLocalRef(convertedBytes);
	}
    }
    env->DeleteLocalRef(array);

    VERIFY(::SelectObject(hDC, oldFont));
    return size;
}


/************************************************************************
 * WFontMetrics native methods
 */

extern "C" {

/*
 * Class:     sun_awt_windows_WFontMetrics
 * Method:    stringWidth
 * Signature: (Ljava/lang/String;)I
 */
JNIEXPORT jint JNICALL
Java_sun_awt_windows_WFontMetrics_stringWidth(JNIEnv *env, jobject self,
					      jstring str)
{
    if (str == NULL) {
	JNU_ThrowNullPointerException(env, "str argument");
	return NULL;
    }
    HDC hDC = ::GetDC(0);    ASSERT(hDC != NULL);

    jobject font = env->GetObjectField(self, AwtFont::fontID);
    
    long ret = AwtFont::getMFStringWidth(hDC, font, str);
    VERIFY(::ReleaseDC(0, hDC));
    return ret;
}

/*
 * Class:     sun_awt_windows_WFontMetrics
 * Method:    charsWidth
 * Signature: ([CII)I
 */
JNIEXPORT jint JNICALL
Java_sun_awt_windows_WFontMetrics_charsWidth(JNIEnv *env, jobject self,
					     jcharArray str, 
					     jint off, jint len) 
{
    if (str == NULL) {
	JNU_ThrowNullPointerException(env, "str argument");
	return NULL;
    }
    if ((len < 0) || (off < 0) || (len + off > (env->GetArrayLength(str)))) {
	JNU_ThrowArrayIndexOutOfBoundsException(env, "off/len argument");
	return NULL;
    }

    jchar *strp = new jchar[len];
    env->GetCharArrayRegion(str, off, len, strp);
    jstring jstr = env->NewString(strp, len);
    jint result = Java_sun_awt_windows_WFontMetrics_stringWidth(env, self,
								jstr);
    delete [] strp;
    return result;
}


/*
 * Class:     sun_awt_windows_WFontMetrics
 * Method:    bytesWidth
 * Signature: ([BII)I
 */
JNIEXPORT jint JNICALL
Java_sun_awt_windows_WFontMetrics_bytesWidth(JNIEnv *env, jobject self,
					     jbyteArray str, 
					     jint off, jint len)
{
    if (str == NULL) {
	JNU_ThrowNullPointerException(env, "bytes argument");
	return NULL;
    }
    if ((len < 0) || (off < 0) || (len + off > (env->GetArrayLength(str)))) {
	JNU_ThrowArrayIndexOutOfBoundsException(env, "off or len argument");
	return NULL;
    }
    char *pStrBody = (char *)env->GetPrimitiveArrayCritical(str, 0);
    char *pStr = pStrBody + off;
    
    jintArray array = (jintArray)env->GetObjectField(self, AwtFont::widthsID);
    jint *widths = (jint *)env->GetPrimitiveArrayCritical(array, 0);
    jint result = 0;

    for (; len; len--) {
    	result += widths[*pStr++];
    }

    env->ReleasePrimitiveArrayCritical(str, pStrBody, 0);
    env->ReleasePrimitiveArrayCritical(array, widths, 0);

    return result;
}    


/*
 * Class:     sun_awt_windows_WFontMetrics 
 * Method:    init
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WFontMetrics_init(JNIEnv *env, jobject self)
{
    jobject font = env->GetObjectField(self, AwtFont::fontID);
    if (font == NULL) {
	JNU_ThrowNullPointerException(env, "fontMetrics' font");
	return;
    }
    jlong pData = env->GetLongField(font, AwtFont::pDataID);

    AwtFont::LoadMetrics(env, self);
}


/*
 * Class:     sun_awt_windows_WFontMetrics 
 * Method:    initIDs
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WFontMetrics_initIDs(JNIEnv *env, jclass cls)
{
    AwtFont::widthsID = env->GetFieldID(cls, "widths", "[I");
    AwtFont::ascentID = env->GetFieldID(cls, "ascent", "I");
    AwtFont::descentID = env->GetFieldID(cls, "descent", "I");
    AwtFont::leadingID = env->GetFieldID(cls, "leading", "I");
    AwtFont::heightID = env->GetFieldID(cls, "height", "I");
    AwtFont::maxAscentID = env->GetFieldID(cls, "maxAscent", "I");
    AwtFont::maxDescentID = env->GetFieldID(cls, "maxDescent", "I");
    AwtFont::maxHeightID = env->GetFieldID(cls, "maxHeight", "I");
    AwtFont::maxAdvanceID = env->GetFieldID(cls, "maxAdvance", "I");

    ASSERT(AwtFont::widthsID != NULL);
    ASSERT(AwtFont::ascentID != NULL);
    ASSERT(AwtFont::descentID != NULL);
    ASSERT(AwtFont::leadingID != NULL);
    ASSERT(AwtFont::heightID != NULL);
    ASSERT(AwtFont::maxAscentID != NULL);
    ASSERT(AwtFont::maxDescentID != NULL);
    ASSERT(AwtFont::maxHeightID != NULL);
    ASSERT(AwtFont::maxAdvanceID != NULL);
}

} /* extern "C" */


/************************************************************************
 * java.awt.Font native methods
 */

extern "C" {

JNIEXPORT void JNICALL
Java_java_awt_Font_initIDs(JNIEnv *env, jclass cls) 
{
    AwtFont::peerMID = env->GetMethodID(cls, "getPeer",
                                        "()Ljava/awt/peer/FontPeer;");
    AwtFont::pDataID = env->GetFieldID(cls, "pData", "J");
    AwtFont::nameID = env->GetFieldID(cls, "name", "Ljava/lang/String;");
    AwtFont::sizeID = env->GetFieldID(cls, "size", "I");
    AwtFont::styleID = env->GetFieldID(cls, "style", "I");

    AwtFont::getFontMID = 
      env->GetStaticMethodID(cls, "getFont", 
			     "(Ljava/lang/String;)Ljava/awt/Font;");

    AwtFont::jaFontCls = cls;

    ASSERT(AwtFont::peerMID != NULL);
    ASSERT(AwtFont::pDataID != NULL);
    ASSERT(AwtFont::nameID != NULL);
    ASSERT(AwtFont::sizeID != NULL);
    ASSERT(AwtFont::styleID != NULL);

    ASSERT(AwtFont::getFontMID != NULL);
}

JNIEXPORT void JNICALL
Java_java_awt_Font_pDispose(JNIEnv *env, jobject self) 
{
    jlong pData = env->GetLongField(self, AwtFont::pDataID);
    AwtFont* awtFont = (AwtFont*)jlong_to_ptr(pData);

    if(NULL != awtFont){
      env->SetLongField(self, AwtFont::pDataID, (long)NULL);
      delete awtFont;
    }
}

} /* extern "C" */


/************************************************************************
 * java.awt.FontMetric native methods
 */

extern "C" {

JNIEXPORT void JNICALL
Java_java_awt_FontMetrics_initIDs(JNIEnv *env, jclass cls) 
{
    AwtFont::fontID = env->GetFieldID(cls, "font", "Ljava/awt/Font;");
    AwtFont::getHeightMID = env->GetMethodID(cls, "getHeight", "()I");

    ASSERT(AwtFont::fontID);
    ASSERT(AwtFont::getHeightMID);
}

} /* extern "C" */


/************************************************************************
 * sun.awt.CharsetString native methods
 */

extern "C" {

JNIEXPORT void JNICALL
Java_sun_awt_CharsetString_initIDs(JNIEnv *env, jclass cls) 
{
    AwtFont::fontDescriptorID = 
	env->GetFieldID(cls, "fontDescriptor", "Lsun/awt/FontDescriptor;");
    AwtFont::charsetStringID = 
	env->GetFieldID(cls, "charsetString", "Ljava/lang/String;");
}

} /* extern "C" */


/************************************************************************
 * sun.awt.FontDescriptor native methods
 */

extern "C" {

JNIEXPORT void JNICALL
Java_sun_awt_FontDescriptor_initIDs(JNIEnv *env, jclass cls) 
{
    AwtFont::fontCharsetID =
    	env->GetFieldID(cls, "fontCharset", "Lsun/io/CharToByteConverter;");
    AwtFont::nativeNameID = env->GetFieldID(cls, "nativeName",
					    "Ljava/lang/String;");

    ASSERT(AwtFont::fontCharsetID != NULL);
    ASSERT(AwtFont::nativeNameID != NULL);
}

} /* extern "C" */


/************************************************************************
 * sun.awt.PlatformFont native methods
 */

extern "C" {

JNIEXPORT void JNICALL
Java_sun_awt_PlatformFont_initIDs(JNIEnv *env, jclass cls) 
{
    AwtFont::propsID = env->GetFieldID(cls, "props", "Ljava/util/Properties;");
    AwtFont::componentFontsID = 
	env->GetFieldID(cls, "componentFonts", "[Lsun/awt/FontDescriptor;");
    AwtFont::makeMultiCharsetStringMID =
	env->GetMethodID(cls, "makeMultiCharsetString", 
			 "(Ljava/lang/String;)[Lsun/awt/CharsetString;");
    AwtFont::makeConvertedMultiFontStringMID =
	env->GetMethodID(cls, "makeConvertedMultiFontString", 
			 "(Ljava/lang/String;)[Ljava/lang/Object;"); 

    ASSERT(AwtFont::makeConvertedMultiFontStringMID != NULL);
    ASSERT(AwtFont::makeMultiCharsetStringMID != NULL);
    ASSERT(AwtFont::componentFontsID != NULL);
    ASSERT(AwtFont::propsID != NULL);


}

} /* extern "C" */


/************************************************************************
 * FontCache methods
 */

void AwtFontCache::Add(WCHAR* name, HFONT font)
{
    fontCache.m_head = new Item(name, font, fontCache.m_head);
}

HFONT AwtFontCache::Lookup(WCHAR* name)
{
    Item* item = fontCache.m_head;
    Item* lastItem = NULL;

    while (item != NULL) {
        if (wcscmp(item->name, name) == 0) {
	    return item->font;
	}
	lastItem = item;
	item = item->next;
    }
    return NULL;
}

BOOL AwtFontCache::Search(HFONT font)
{
    Item* item = fontCache.m_head;

    while (item != NULL) {
        if (item->font == font) {
	    return TRUE;
        }
        item = item->next;
    }
    return FALSE;
}

void AwtFontCache::Remove(HFONT font)
{
    Item* item = fontCache.m_head;
    Item* lastItem = NULL;

    while (item != NULL) {
        if (item->font == font) {
  	    if (DecRefCount(item) <= 0){
  	        if (lastItem == NULL) {
	            fontCache.m_head = item->next;
	        } else {
	        lastItem->next = item->next;
	        }
	     delete item;
	     }
   	     return;
        }
        lastItem = item;
        item = item->next;
    }
}

void AwtFontCache::Clear()
{
    Item* item = m_head;
    Item* next;

    while (item != NULL) {
        next = item->next;
	delete item;
	item = next;
    }

    m_head = NULL;
}

/* NOTE: In the interlock calls below the return value is different
         depending on which version of windows. However, all versions
	 return a 0 or less than value when the count gets there. Only
	 under NT 4.0 & 98 does the value actaully represent the new value. */

void AwtFontCache::IncRefCount(HFONT hFont){
    Item* item = fontCache.m_head;

    while (item != NULL){

        if (item->font == hFont){
            IncRefCount(item);
	    return;
        }
        item = item->next;
    }
}
         
LONG AwtFontCache::IncRefCount(Item* item){
    LONG    newVal;

    if(NULL != item){
        newVal = InterlockedIncrement((long*)&item->refCount);
    }
    return(newVal);
}

LONG AwtFontCache::DecRefCount(Item* item){
    LONG    newVal;

    if(NULL != item){
        newVal = InterlockedDecrement((long*)&item->refCount);
    }
    return(newVal);
}

AwtFontCache::Item::Item(const WCHAR* s, HFONT f, AwtFontCache::Item* n )
{
    name = wcsdup(s);
    font = f;
    next = n;
    refCount = 1;
}

AwtFontCache::Item::~Item() {
  VERIFY(::DeleteObject(font));
  free(name);
}

/////////////////////////////////////////////////////////////////////////////
// for canConvert native method of WDefaultFontCharset

class CSegTableComponent
{
public:
    CSegTableComponent();
    virtual ~CSegTableComponent();
    virtual void Create(LPWSTR name);
    virtual BOOL In(USHORT iChar) { ASSERT(FALSE); return FALSE; };
    LPWSTR GetFontName(){
	ASSERT(m_lpszFontName != NULL); return m_lpszFontName;
    };

private:
    LPWSTR m_lpszFontName;
};

CSegTableComponent::CSegTableComponent()
{
    m_lpszFontName = NULL;
}

CSegTableComponent::~CSegTableComponent()
{
    if (m_lpszFontName != NULL) {
	free(m_lpszFontName);
        m_lpszFontName = NULL;
    }
}

void CSegTableComponent::Create(LPWSTR name)
{
    if (m_lpszFontName != NULL) {
	free(m_lpszFontName);
        m_lpszFontName = NULL;
    }
    m_lpszFontName = wcsdup(name);
    ASSERT(m_lpszFontName);
}

#define CMAPHEX 0x70616d63 // = "cmap" (reversed)

// CSegTable: Segment table describing character coverage for a font
class CSegTable : public CSegTableComponent
{
public:
    CSegTable();
    virtual ~CSegTable();
    virtual BOOL In(USHORT iChar);
    BOOL HasCmap();
    virtual BOOL IsEUDC() { ASSERT(FALSE); return FALSE; };

protected:
    virtual void GetData(DWORD dwOffset, LPVOID lpData, DWORD cbData) {
	ASSERT(FALSE); };
    void MakeTable();
    static void SwapShort(USHORT& p);
    static void SwapULong(ULONG& p);

private:
    USHORT m_cSegCount;	    // number of segments
    PUSHORT m_piStart;	    // pointer to array of starting values
    PUSHORT m_piEnd;	    // pointer to array of ending values (inclusive)
    USHORT m_cSeg;	    // current segment (cache)
};

CSegTable::CSegTable()
{
    m_cSegCount = 0;
    m_piStart = NULL;
    m_piEnd = NULL;
    m_cSeg = 0;
}

CSegTable::~CSegTable()
{
    if (m_piStart != NULL)
	delete[] m_piStart;
    if (m_piEnd != NULL)
	delete[] m_piEnd;
}

#define OFFSETERROR 0

void CSegTable::MakeTable()
{
typedef struct tagTABLE{
    USHORT  platformID;
    USHORT  encodingID;
    ULONG   offset;
} TABLE, *PTABLE;

typedef struct tagSUBTABLE{
    USHORT  format;
    USHORT  length;
    USHORT  version;
    USHORT  segCountX2;
    USHORT  searchRange;
    USHORT  entrySelector;
    USHORT  rangeShift;
} SUBTABLE, *PSUBTABLE;

    USHORT aShort[2];
    (void) GetData(0, aShort, sizeof(aShort));
    USHORT nTables = aShort[1];
    SwapShort(nTables);

    // allocate buffer to hold encoding tables
    DWORD cbData = nTables * sizeof(TABLE);
    PTABLE pTables = new TABLE[nTables];
    PTABLE pTable = pTables;

    // get array of encoding tables.
    (void) GetData(4, (PBYTE) pTable, cbData);

    ULONG offsetFormat4 = OFFSETERROR;
    USHORT i;
    for (i = 0; i < nTables; i++) {
	SwapShort(pTable->encodingID);
	SwapShort(pTable->platformID);
	//for a Unicode font for Windows, platformID == 3, encodingID == 1
	if ((pTable->platformID == 3)&&(pTable->encodingID == 1)) {
	    offsetFormat4 = pTable->offset;
	    SwapULong(offsetFormat4);
	    break;
	}
	pTable++;
    }
    delete[] pTables;
    if (offsetFormat4 == OFFSETERROR) {
        return;
    }
//    ASSERT(offsetFormat4 != OFFSETERROR);

    SUBTABLE subTable;
    (void) GetData(offsetFormat4, &subTable, sizeof(SUBTABLE));
    SwapShort(subTable.format);
    SwapShort(subTable.segCountX2);
    ASSERT(subTable.format == 4);

    m_cSegCount = subTable.segCountX2/2;

    // read in the array of segment end values
    m_piEnd = new USHORT[m_cSegCount];
    ASSERT(m_piEnd);
    ULONG offset = offsetFormat4
	+ sizeof(SUBTABLE); //skip constant # bytes in subtable
    cbData = m_cSegCount * sizeof(USHORT);
    (void) GetData(offset, m_piEnd, cbData);
    for (i = 0; i < m_cSegCount; i++)
	SwapShort(m_piEnd[i]);
    ASSERT(m_piEnd[m_cSegCount-1] == 0xffff);

    // read in the array of segment start values
    m_piStart = new USHORT[m_cSegCount];
    ASSERT(m_piStart);
    offset += cbData	    //skip SegEnd array
	+ sizeof(USHORT);   //skip reservedPad
    (void) GetData(offset, m_piStart, cbData);
    for (i = 0; i < m_cSegCount; i++)
	SwapShort(m_piStart[i]);
    ASSERT(m_piStart[m_cSegCount-1] == 0xffff);
}

BOOL CSegTable::In(USHORT iChar)
{
    if (!HasCmap()) {
        return FALSE;
    }
//    ASSERT(m_piStart);
//    ASSERT(m_piEnd);

    if (iChar > m_piEnd[m_cSeg]) {
	for (; (m_cSeg < m_cSegCount)&&(iChar > m_piEnd[m_cSeg]); m_cSeg++);
    } else if (iChar < m_piStart[m_cSeg]) {
	for (; (m_cSeg > 0)&&(iChar < m_piStart[m_cSeg]); m_cSeg--);
    }

    if ((iChar <= m_piEnd[m_cSeg])&&(iChar >= m_piStart[m_cSeg])&&(iChar != 0xffff))
	return TRUE;
    else
	return FALSE;
}

inline BOOL CSegTable::HasCmap()
{
    return (((m_piEnd)&&(m_piStart)) ? TRUE : FALSE);
}

inline void CSegTable::SwapShort(USHORT& p)
{
    SHORT temp;

    temp = (SHORT)(HIBYTE(p) + (LOBYTE(p) << 8));
    p = temp;
}

inline void CSegTable::SwapULong(ULONG& p)
{
    ULONG temp;

    temp = (LONG) ((BYTE) p);
    temp <<= 8;
    p >>= 8;

    temp += (LONG) ((BYTE) p);
    temp <<= 8;
    p >>= 8;

    temp += (LONG) ((BYTE) p);
    temp <<= 8;
    p >>= 8;

    temp += (LONG) ((BYTE) p);
    p = temp;
}

class CStdSegTable : public CSegTable
{
public:
    CStdSegTable();
    virtual ~CStdSegTable();
    BOOL IsEUDC() { return FALSE; };
    virtual void Create(LPWSTR name);

protected:
    void GetData(DWORD dwOffset, LPVOID lpData, DWORD cbData);

private:
    HDC m_hTmpDC;
};

CStdSegTable::CStdSegTable()
{
    m_hTmpDC = NULL;
}

CStdSegTable::~CStdSegTable()
{
    ASSERT(m_hTmpDC == NULL);
}

inline void CStdSegTable::GetData(DWORD dwOffset,
				  LPVOID lpData, DWORD cbData)
{
    ASSERT(m_hTmpDC);
    DWORD nBytes =
	::GetFontData(m_hTmpDC, CMAPHEX, dwOffset, lpData, cbData);
    ASSERT(nBytes != GDI_ERROR);
}

void CStdSegTable::Create(LPWSTR name)
{
    CSegTableComponent::Create(name);

    HWND hWnd = ::GetDesktopWindow();
    ASSERT(hWnd);
    m_hTmpDC = ::GetWindowDC(hWnd);
    ASSERT(m_hTmpDC);

    HFONT hFont = CreateHFont_sub(name, 0, 20);

    HFONT hOldFont = ::SelectObject(m_hTmpDC, hFont);
    ASSERT(hOldFont);

    (void) MakeTable();

    VERIFY(::SelectObject(m_hTmpDC, hOldFont));
    VERIFY(::DeleteObject(hFont));
    VERIFY(::ReleaseDC(hWnd, m_hTmpDC) != 0);
    m_hTmpDC = NULL;
}

class CEUDCSegTable : public CSegTable
{
public:
    CEUDCSegTable();
    virtual ~CEUDCSegTable();
    BOOL IsEUDC() { return TRUE; };
    virtual void Create(LPWSTR name);

protected:
    void GetData(DWORD dwOffset, LPVOID lpData, DWORD cbData);

private:
    HANDLE m_hTmpFile;
    ULONG m_hTmpCMapOffset;
};

CEUDCSegTable::CEUDCSegTable()
{
    m_hTmpFile = NULL;
    m_hTmpCMapOffset = 0;
}

CEUDCSegTable::~CEUDCSegTable()
{
    ASSERT(m_hTmpFile == NULL);
    ASSERT(m_hTmpCMapOffset == 0);
}

inline void CEUDCSegTable::GetData(DWORD dwOffset,
				   LPVOID lpData, DWORD cbData)
{
    ASSERT(m_hTmpFile);
    ASSERT(m_hTmpCMapOffset);
    ::SetFilePointer(m_hTmpFile, m_hTmpCMapOffset + dwOffset,
	NULL, FILE_BEGIN);
    DWORD dwRead;
    VERIFY(::ReadFile(m_hTmpFile, lpData, cbData, &dwRead, NULL));
    ASSERT(dwRead == cbData);
}

void CEUDCSegTable::Create(LPWSTR name)
{
typedef struct tagHEAD{
    FIXED   sfnt_version;
    USHORT  numTables;
    USHORT  searchRange;
    USHORT  entrySelector;
    USHORT  rangeShift;
} HEAD, *PHEAD;

typedef struct tagENTRY{
    ULONG   tag;
    ULONG   checkSum;
    ULONG   offset;
    ULONG   length;
} ENTRY, *PENTRY;

    CSegTableComponent::Create(name);

    // create EUDC font file and make EUDCSegTable
    // after wrapper function for CreateFileW, we use only CreateFileW
    if (IS_NT) {
	m_hTmpFile = ::CreateFileW(name, GENERIC_READ,
	    FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    } else {
#ifdef WIN32
	ASSERT(IS_WIN95);
#endif
	char szFileName[_MAX_PATH];
	::WideCharToMultiByte(CP_ACP, 0, name, -1,
	    szFileName, sizeof(szFileName), NULL, NULL);
	m_hTmpFile = ::CreateFileA(szFileName, GENERIC_READ,
	    FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    }
    if (m_hTmpFile == INVALID_HANDLE_VALUE){
	m_hTmpFile = NULL;
	return;
    }

    HEAD head;
    DWORD dwRead;
    VERIFY(::ReadFile(m_hTmpFile, &head, sizeof(head), &dwRead, NULL));
    ASSERT(dwRead == sizeof(HEAD));
    SwapShort(head.numTables);
    ENTRY entry;
    for (int i = 0; i < head.numTables; i++){
	VERIFY(::ReadFile(m_hTmpFile, &entry, sizeof(entry), &dwRead, NULL));
	ASSERT(dwRead == sizeof(ENTRY));
	if (entry.tag == CMAPHEX)
	    break;
    }
    ASSERT(entry.tag == CMAPHEX);
    SwapULong(entry.offset);
    m_hTmpCMapOffset = entry.offset;

    (void) MakeTable();

    m_hTmpCMapOffset = 0;
    VERIFY(::CloseHandle(m_hTmpFile));
    m_hTmpFile = NULL;
}

class CSegTableManagerComponent
{
public:
    CSegTableManagerComponent();
    ~CSegTableManagerComponent();

protected:
    void MakeBiggerTable();
    CSegTableComponent **m_tables;
    int m_nTable;
    int m_nMaxTable;
};

#define TABLENUM 20

CSegTableManagerComponent::CSegTableManagerComponent()
{
    m_nTable = 0;
    m_nMaxTable = TABLENUM;
    m_tables = new CSegTableComponent*[m_nMaxTable];
}

CSegTableManagerComponent::~CSegTableManagerComponent()
{
    for (int i = 0; i < m_nTable; i++) {
	ASSERT(m_tables[i]);
	delete m_tables[i];
    }
    delete [] m_tables;
    m_tables = NULL;
}

void CSegTableManagerComponent::MakeBiggerTable()
{
    CSegTableComponent **tables =
	new CSegTableComponent*[m_nMaxTable + TABLENUM];
    ASSERT(tables);

    for (int i = 0; i < m_nMaxTable; i++)
	tables[i] = m_tables[i];

    delete[] m_tables;

    m_tables = tables;
    m_nMaxTable += TABLENUM;
}

class CSegTableManager : public CSegTableManagerComponent
{
public:
    CSegTable* GetTable(LPWSTR lpszFontName, BOOL fEUDC);
};

CSegTable* CSegTableManager::GetTable(LPWSTR lpszFontName, BOOL fEUDC)
{
    for (int i = 0; i < m_nTable; i++) {
	if ((((CSegTable*)m_tables[i])->IsEUDC() == fEUDC) &&
	    (wcscmp(m_tables[i]->GetFontName(),lpszFontName) == 0))
	    return (CSegTable*) m_tables[i];
    }

    if (m_nTable == m_nMaxTable) {
	(void) MakeBiggerTable();
    }
    ASSERT(m_nTable < m_nMaxTable);

    if (!fEUDC)
	m_tables[m_nTable] = new CStdSegTable;
    else
	m_tables[m_nTable] = new CEUDCSegTable;
    m_tables[m_nTable]->Create(lpszFontName);
    return (CSegTable*) m_tables[m_nTable++];
}

CSegTableManager g_segTableManager;

class CCombinedSegTable : public CSegTableComponent
{
public:
    CCombinedSegTable();
    void Create(LPWSTR name);
    BOOL In(USHORT iChar);

private:
    LPSTR GetCodePageSubkey();
    void GetEUDCFileName(LPWSTR lpszFileName, int cchFileName);
    static char m_szCodePageSubkey[16];
    static WCHAR m_szDefaultEUDCFile[_MAX_PATH];
    static BOOL m_fEUDCSubKeyExist;
    static BOOL m_fTTEUDCFileExist;
    CStdSegTable* m_pStdSegTable;
    CEUDCSegTable* m_pEUDCSegTable;
};

char CCombinedSegTable::m_szCodePageSubkey[16] = "";

WCHAR CCombinedSegTable::m_szDefaultEUDCFile[_MAX_PATH] = L"";

BOOL CCombinedSegTable::m_fEUDCSubKeyExist = TRUE;

BOOL CCombinedSegTable::m_fTTEUDCFileExist = TRUE;

CCombinedSegTable::CCombinedSegTable()
{
    m_pStdSegTable = NULL;
    m_pEUDCSegTable = NULL;
}

#include <locale.h>
LPSTR CCombinedSegTable::GetCodePageSubkey()
{
    if (strlen(m_szCodePageSubkey) > 0)
	return m_szCodePageSubkey;

    LPSTR lpszLocale = setlocale(LC_CTYPE, "");
    // cf lpszLocale = "Japanese_Japan.932"
    ASSERT(lpszLocale);
    LPSTR lpszCP = strchr(lpszLocale, (int) '.');
    if (lpszCP == NULL)
	return NULL;
    lpszCP++; // cf lpszCP = "932"
    char szSubKey[80];
    strcpy(szSubKey, "EUDC\\");
    strcpy(&(szSubKey[strlen(szSubKey)]), lpszCP);
    strcpy(m_szCodePageSubkey, szSubKey);
    return m_szCodePageSubkey;
}

void CCombinedSegTable::GetEUDCFileName(LPWSTR lpszFileName, int cchFileName)
{
    if (m_fEUDCSubKeyExist == FALSE)
	return;

    // get filename of typeface-specific TureType EUDC font
    LPSTR lpszSubKey = GetCodePageSubkey();
    if (lpszSubKey == NULL) {
	m_fEUDCSubKeyExist = FALSE;
	return; // can not get codepage information
    }
    HKEY hRootKey = HKEY_CURRENT_USER;
    HKEY hKey;
    LONG lRet = ::RegOpenKeyExA(hRootKey, lpszSubKey, 0, KEY_ALL_ACCESS, &hKey);
    if (lRet != ERROR_SUCCESS) {
	m_fEUDCSubKeyExist = FALSE;
	return; // no EUDC font
    }

    // get EUDC font file name
    WCHAR szFamilyName[80];
    wcscpy(szFamilyName, GetFontName());
    WCHAR* delimit = wcschr(szFamilyName, L',');
    if (delimit != NULL)
	*delimit = L'\0';
    DWORD dwType;
    UCHAR szFileName[_MAX_PATH];
    ::ZeroMemory(szFileName, sizeof(szFileName));
    DWORD dwBytes = sizeof(szFileName);
    // try Typeface-specific EUDC font
    char szTmpName[80];
    VERIFY(::WideCharToMultiByte(CP_ACP, 0, szFamilyName, -1,
	szTmpName, sizeof(szTmpName), NULL, NULL));
    LONG lStatus = ::RegQueryValueExA(hKey, (LPCSTR) szTmpName,
	NULL, &dwType, szFileName, &dwBytes);
    BOOL fUseDefault = FALSE;
    if (lStatus != ERROR_SUCCESS){ // try System default EUDC font
	if (m_fTTEUDCFileExist == FALSE)
	    return;
	if (wcslen(m_szDefaultEUDCFile) > 0) {
	    wcscpy(lpszFileName, m_szDefaultEUDCFile);
	    return;
	}
	char szDefault[] = "SystemDefaultEUDCFont";
	lStatus = ::RegQueryValueExA(hKey, (LPCSTR) szDefault,
	    NULL, &dwType, szFileName, &dwBytes);
	fUseDefault = TRUE;
	if (lStatus != ERROR_SUCCESS) {
	    m_fTTEUDCFileExist = FALSE;
	    // This font is associated with no EUDC font
	    // and there is no system default EUDC font
	    return;
	}
    }

    if (strcmp((LPCSTR) szFileName, "userfont.fon") == 0) {
	// This font is associated with no EUDC font
	// and the system default EUDC font is not TrueType
	m_fTTEUDCFileExist = FALSE;
	return;
    }

    ASSERT(strlen((LPCSTR)szFileName) > 0);
    VERIFY(::MultiByteToWideChar(CP_ACP, 0,
	(LPCSTR)szFileName, -1,	lpszFileName, cchFileName) != 0);
    if (fUseDefault)
	wcscpy(m_szDefaultEUDCFile, lpszFileName);
}

void CCombinedSegTable::Create(LPWSTR name)
{
    CSegTableComponent::Create(name);

    m_pStdSegTable =
	(CStdSegTable*) g_segTableManager.GetTable(name, FALSE/*not EUDC*/);
    WCHAR szEUDCFileName[_MAX_PATH];
    ::ZeroMemory(szEUDCFileName, sizeof(szEUDCFileName));
    (void) GetEUDCFileName(szEUDCFileName,
	sizeof(szEUDCFileName)/sizeof(WCHAR));
    if (wcslen(szEUDCFileName) > 0) {
	m_pEUDCSegTable = (CEUDCSegTable*) g_segTableManager.GetTable(
	    szEUDCFileName, TRUE/*EUDC*/);
	if (m_pEUDCSegTable->HasCmap() == FALSE)
	    m_pEUDCSegTable = NULL;
    }
}

BOOL CCombinedSegTable::In(USHORT iChar)
{
    ASSERT(m_pStdSegTable);
    if (m_pStdSegTable->In(iChar))
	return TRUE;

    if (m_pEUDCSegTable != NULL)
	return m_pEUDCSegTable->In(iChar);

    return FALSE;
}

class CCombinedSegTableManager : public CSegTableManagerComponent
{
public:
    CCombinedSegTable* GetTable(LPWSTR lpszFontName);
};

CCombinedSegTable* CCombinedSegTableManager::GetTable(LPWSTR lpszFontName)
{
    for (int i = 0; i < m_nTable; i++) {
	if (wcscmp(m_tables[i]->GetFontName(),lpszFontName) == 0)
	    return (CCombinedSegTable*) m_tables[i];
    }

    if (m_nTable == m_nMaxTable) {
	(void) MakeBiggerTable();
    }
    ASSERT(m_nTable < m_nMaxTable);

    m_tables[m_nTable] = new CCombinedSegTable;
    m_tables[m_nTable]->Create(lpszFontName);

    return (CCombinedSegTable*) m_tables[m_nTable++];
}


/************************************************************************
 * WDefaultFontCharset native methos
 */

extern "C" {

JNIEXPORT void JNICALL
Java_sun_awt_windows_WDefaultFontCharset_initIDs(JNIEnv *env, jclass cls) 
{
    AwtFont::fontNameID = env->GetFieldID(cls, "fontName", 
					  "Ljava/lang/String;");
    ASSERT(AwtFont::fontNameID != NULL);
}


/*
 * !!!!!!!!!!!!!!!!!!!! this does not work. I am not sure why, but
 * when active, this will reliably crash HJ, with no hope of debugging
 * for java.  It doesn't seem to crash the _g version.
 * !!!!!!!!!!!!!!!!!!!!!!!!!!!!
 *
 * I suspect may be running out of C stack: see alloca in
 * JNI_GET_STRING, the alloca in it.
 * 
 * (the method is prefixed with XXX so that the linker won't find it) */
JNIEXPORT jboolean JNICALL
Java_sun_awt_windows_WDefaultFontCharset_canConvert(JNIEnv *env, jobject self,
						    jchar ch)
{
    static CCombinedSegTableManager tableManager;

    jstring fontName = (jstring)env->GetObjectField(self, AwtFont::fontNameID);
    ASSERT(fontName != NULL);
    LPWSTR fontNameWStr = TO_WSTRING(fontName);
    CCombinedSegTable* pTable = tableManager.GetTable(fontNameWStr);
    return (pTable->In((USHORT) ch) ? JNI_TRUE : JNI_FALSE);
}

} /* extern "C" */
