/*
 * @(#)awt_Object.cpp	1.27 98/09/15
 *
 * Copyright 1996, 1997 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "awt_Object.h"
#include "ObjectList.h"
#ifdef DEBUG
#include "awt_Unicode.h"
#endif

#ifdef DEBUG
static BOOL reportEvents = FALSE;
#endif


/************************************************************************
 * AwtObject fields
 */

jfieldID AwtObject::pDataID;
jfieldID AwtObject::targetID;

jmethodID AwtObject::getPeerForTargetMID;


/************************************************************************
 * AwtObject methods
 */

AwtObject::AwtObject()
{
    theAwtObjectList.Add(this);
    m_peerObject = NULL;
    m_callbacksEnabled = TRUE;
}

AwtObject::~AwtObject()
{
    theAwtObjectList.Remove(this);
}

/*
 * Return the peer associated with some target.  This information is
 * maintained in a hashtable at the java level.  
 */
jobject AwtObject::GetPeerForTarget(JNIEnv *env, jobject target)
{
    jobject peer = GetPeer(env);
    jobject result = 
	env->CallObjectMethod(peer, AwtObject::getPeerForTargetMID,  target);

    ASSERT(!env->ExceptionOccurred());
    return result;
}

/* Execute a callback to the associated Java peer. */
void 
AwtObject::DoCallback(const char* methodName, const char* methodSig, ...)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    
    /* don't callback during the create & initialization process */
    if (m_peerObject != NULL && m_callbacksEnabled) {
        va_list args;
        va_start(args, methodSig);
#ifdef DEBUG
        if (reportEvents) {
	    jstring targetStr = 
		(jstring)JNU_CallMethodByName(env, NULL, GetTarget(env), 
					      "getName",
					      "()Ljava/lang/String;").l;
            ASSERT(!env->ExceptionOccurred());
            printf("Posting %s%s method to %S\n", methodName, methodSig,
		   TO_WSTRING(targetStr));
        }
#endif
	/* caching would do much good here */
	JNU_CallMethodByNameV(env, NULL, GetPeer(env), 
			      methodName, methodSig, args);
	{
	    jthrowable exc = env->ExceptionOccurred();
	    if (exc) {
	        env->DeleteLocalRef(exc);
	        env->ExceptionDescribe();
		env->ExceptionClear();
	    }
        }
	ASSERT(!env->ExceptionOccurred());
        va_end(args);
    }
}

void AwtObject::SendEvent(jobject event)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

#ifdef DEBUG
    if (reportEvents) {
	jstring eventStr = JNU_ToString(env, event);
        ASSERT(!env->ExceptionOccurred());
	jstring targetStr = 
	    (jstring)JNU_CallMethodByName(env, NULL, GetTarget(env),"getName",
					  "()Ljava/lang/String;").l;
        ASSERT(!env->ExceptionOccurred());
        printf("Posting %S to %S\n", TO_WSTRING(eventStr),
	       TO_WSTRING(targetStr));
    }
#endif
    /* Post event to the system EventQueue. */
    JNU_CallMethodByName(env, NULL, GetPeer(env), "postEvent", 
			 "(Ljava/awt/AWTEvent;)V", event);
    {
        jthrowable exc = env->ExceptionOccurred();
	if (exc) {
	    env->DeleteLocalRef(exc);
	    env->ExceptionDescribe();
	}
    }
    ASSERT(!env->ExceptionOccurred());
}

//
// (static)
// Switches to Windows thread via SendMessage and synchronously
// calls AwtObject::WinThreadExecProc with the given command id
// and parameters.
//
// Useful for writing code that needs to be synchronized with
// what's happening on the Windows thread.
//
LONG AwtObject::WinThreadExec(
    jobject			 	peerObject,
    UINT				cmdId,
    LPARAM 				param1,
    LPARAM 				param2,
    LPARAM 				param3,
    LPARAM 				param4 )
{
    ASSERT( IsWindow( AwtToolkit::GetInstance().GetHWnd() ) );
    ASSERT( peerObject != NULL);

    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    AwtObject* object = (AwtObject *)JNI_GET_PDATA(peerObject);
    ASSERT( !IsBadReadPtr(object, sizeof(AwtObject)) );

    ExecuteArgs		args;
    LONG		retVal;

    // setup arguments
    args.cmdId = cmdId;
    args.param1 = param1;
    args.param2 = param2;
    args.param3 = param3;
    args.param4 = param4;

    // call WinThreadExecProc on the toolkit thread
    retVal = ::SendMessage(AwtToolkit::GetInstance().GetHWnd(), WM_AWT_EXECUTE_SYNC, (WPARAM)object, (LPARAM)&args);

    return retVal;
}

LONG AwtObject::WinThreadExecProc(ExecuteArgs * args)
{
    ASSERT(FALSE); // no default handler
    return 0L;
}

/************************************************************************
 * WObjectPeer native methods
 */

extern "C" {

JNIEXPORT void JNICALL 
Java_sun_awt_windows_WObjectPeer_initIDs(JNIEnv *env, jclass cls) {
    AwtObject::pDataID = env->GetFieldID(cls, "pData", "I");
    AwtObject::targetID = env->GetFieldID(cls, "target", 
					      "Ljava/lang/Object;");

    AwtObject::getPeerForTargetMID = 
	env->GetMethodID(cls, "getPeerForTarget", 
			 "(Ljava/lang/Object;)Lsun/awt/windows/WObjectPeer;");

    ASSERT(AwtObject::pDataID != NULL);
    ASSERT(AwtObject::targetID != NULL);
    ASSERT(AwtObject::getPeerForTargetMID != NULL);
}

} /* extern "C" */
