/*
 * @(#)gc.c	1.256 98/10/27
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Java memory management and garbage collection
 *
 * Implementation notes:
 *
 * I [Tim] tried to split this file up into two: heap.c, covering heap
 * initialization and allocation routines, and gc.c, covering GC and heap
 * expansion.  Doing so changed the basic heap pointers from statics,
 * and the result (for the SunPro compiler, at least), was a 10% slowdown
 * of the garbage collector.  So now we've got a big file again.
 *
 * There are 2 interesting flags for debug print: verbosegc is good for
 * lightweight garbage collecting info, and is turned on by the menu
 * item "Verbose GC". tracegc is good for heavy-duty print of lots of
 * info that will take a long time.
 *
 * TO DO:
 *    Think about the policy issues more - how much memory to pre-allocate,
 *    when do we gc vs. when do we expand the heap (manageAllocFailure),
 *    when do we shrink the heap (evacuationChunk).  Also default handle
 *    space vs. object space split.
 *    Need to make this tunable for each implementation of the VM, and should
 *    allow user controls.
 */

#include "oobj.h"
#include "interpreter.h"
#include "tree.h"
#include "typecodes.h"
#include "monitor.h"
#include "util.h"
#include "threads.h"
#include "signature.h"
#include "monitor_cache.h"
#include "alloc_cache.h"
#include "jlong.h"

#include "sys_api.h"

#include "gc.h"
#include "java_lang_ref_Reference.h"
#include "java_lang_ref_SoftReference.h"

#ifdef BREAKPTS
#include "breakpoints.h"
#endif

#define VERBOSEGC

#undef max
#define max(a, b)	((a) < (b) ? (b) : (a))

float minHeapFreePercent = (float)0.35;
float maxHeapFreePercent = (float)0.60;
long minHeapExpansion = 1 * 1024 * 1024;
long maxHeapExpansion = 4 * 1024 * 1024;

jlong timeOfLastGC = 0;

static void scanMonitorCache(void);

#ifndef PAGED_HEAPS /************ CONTIGUOUS HEAPS: ********************/

/* 
 * Heap layout, pointers and pointer aliases set by SetLimits():
 *
 * |----------------------------mapped memory---------------------------|
 *	    |------------------committed memory----------------|
 * 
 *  -------------------------------------------------------------------
 * |   ... <|  handles	 |	    objects		       |> ...  |
 *  -------------------------------------------------------------------
 *  ^	     ^		^ ^				      ^		^
 *  |	     |		| |				      |		|
 *  |	    hpool	| hpoollimit		      opoollimit	|
 * heapbase  |		| opool				      |	    heaptop
 *	    hpmin   hpmax opmin				   opmax
 *
 * Constraints:
 *    hpool/hpmin and hpoollimit are HANDLEGRAIN aligned
 *    opool/opmin must be OBJECTGRAIN aligned *plus* sizeof(hdr)
 *    the committed memory must be contained by the mapped memory
 */

/* The extent of the object pool (must be committed memory) */
static unsigned char *opool, *opoollimit, *opoolhand;
/* The extent of the handle pool (must be committed memory) */
static unsigned char *hpool, *hpoollimit;
/* The free handle list (replaces hpoolhand, which was a hint) */
static unsigned char *hpoolfreelist;
/* The extent of the mapped memory */
static unsigned char *heapbase, *heaptop;

/* Pointer the end of the commited region of the mark bits */
static unsigned int * markend;
/* Pointer the end of the allocated region of the mark bits */
static unsigned int * markmax;

#else 		/************ PAGED HEAPS: ********************/

/* Pointer the end of the commited region of the mark bits */
static unsigned int * markend;

/*
 * Part of Java memory management and garbage collection.
 *
 * This supports a discontiguous gcable heap, which is useful for the
 * Mac OS, or other platforms without good memory mapping support.
 *
 * Reworked memory mgmt of gc to be more Mac friendly. Instead of allocating
 * all memory in a giant glob, and subdividing it into handle and object space,
 * we now allocate separate handle chunks and object chunks, and manage these
 * as a linked list of large chunks. The chunks still have to be locked down
 * in the Mac's system memory pool, but at least we allocate them more
 * rationally.
 * One more tidbit: due to requirement of 8B alignment, we alloc an extra
 * handle (or object word) so we can adjust startPtr to start at desired 
 * alignment.
 */ 

/* doubly-linked list of handle chunks, in address order: */
ChunkBlkP firstHanBlkP = NULL;	  
ChunkBlkP lastHanBlkP = NULL;
ChunkBlkP curHanBlkP = NULL;
/* doubly-linked list of object chunks, in address order: */
ChunkBlkP firstObjBlkP = NULL;
ChunkBlkP lastObjBlkP = NULL;
ChunkBlkP curObjBlkP = NULL;

/* PAGE MAP:
 * Every page of memory from the lowest chunk to the highest chunk given to
 * Java by the OS has an entry in the "page_map".  The page map records the
 * type of the page (handle/object/free [encoded in the chunk_size field]), 
 * the page offset to the beginning of the chunk, and a pointer to the markbits
 * for the page.  Since pages are aligned, it is simple to determinethe
 * PageMapEntry for any random pointer, and therefore "ValidHandle",
 * "isMarked", etc.
 */
typedef struct {
    short page_number;	      /* Which page is this in the chunk */
    short chunk_size;	      /* # of pages in this chunk, encodes chunk type */
    unsigned int*  mark_bits; /* ptr to the mark bits for this page */
} PageMapEntry;

PageMapEntry *page_map = NULL;	/* This can (and must) be re-allocated when 
				 * lower pages are allocated 
				 */
void* mem_base;		/* lowest page we have allocated */
void* mem_top;		/* ptr past highest page we have allocated */

static unsigned char *markbitsfreeptr;
static long marksize;           /* # bytes for global markbits pool */

/* number of entries in map: */
#define PageMapLength() (((char *)mem_top - (char *)mem_base) / PAGE_ALIGNMENT)
    
#define HANDLE_HEAP_TYPE    -1
#define FREE_HEAP_TYPE	     0
#define OBJECT_HEAP_TYPE     1

long gMaxMem = 0; /* DRW: Not used - maximum bytes of object memory to alloc */

#endif /************ END PAGED HEAPS ********************/

/* prototypes */
static int manageAllocFailure(size_t n, int overflow_type, int overflow_act);
static unsigned char *gc0(unsigned int, bool_t clear_soft_refs);
void validatePool(long quick);

static int deferredHandleTableOverflow;

#define MAX_GC_DEPTH 10
static void markChildren(JHandle *h, JHandle *limit, int depth,
			 bool_t traceRefs);
static void markChildrenFromTop(JHandle *h, JHandle *limit, bool_t traceRefs);
static void scanJavaFrame(JavaFrame *frame, stack_item *top_top_stack);
static HObject *realObjAlloc(ExecEnv *ee, struct methodtable *mptr, long n);
static HObject *AllocHandle(struct methodtable *mptr, ClassObject *p);
#ifdef DEBUG
void verifyObjectCtrs(void);
void verifyHandleCtrs(void);
void DumpHeapInfo(void);
#ifdef PAGED_HEAPS
void checkChunk(ChunkBlkP);
void verifyPageMap(void);
#endif /* PAGED_HEAPS */
#endif /* DEBUG */

#ifdef HPROF
#include "byteorder.h"
#include "vmprofiler.h"
static void jvmpi_gc_dump_heap(void);
static bool_t jvmpi_in_dump_heap_flag = FALSE;
#endif /* HPROF */

#define ROUND_SIZE(size) \
    ((size + sizeof(hdr) + (OBJECTGRAIN - 1)) & ~(OBJECTGRAIN - 1))

/*
 * Lock against heap modification.
 */
sys_mon_t *_heap_lock;

/*
 * Micro lock for disabling GC */
static sys_mon_t *_gc_disable_lock;
static int gc_disable_count = 0;

#define GC_DISABLE_LOCK(self) sysMonitorEnter(self, _gc_disable_lock)
#define GC_DISABLE_UNLOCK(self) sysMonitorExit(self, _gc_disable_lock)

static unsigned int *markbits;  /* global markbits pool */
#define BITSPERMARK (BITSPERCHAR * sizeof(*markbits))	/* 32 */

static long FreeObjectCtr;	/* # bytes of objects free */
static long TotalObjectCtr;	/* # bytes of objects space allocated so far */
static long MinTotalObjectCtr;	/* minimum size of object space in bytes */
static long FreeHandleCtr;	/* # bytes of free handles (not # of handles) */
static long TotalHandleCtr;	/* # bytes of handles allocated so far */
static long nfreed = 0;		/* # of blocks freed in last gc */
static long bytesfreed = 0;	/* # of bytes freed in last gc */
static long expanded;		/* # of bytes expanded/shrunk since last gc */

#ifdef DEBUG
static int markbitsInUse = FALSE; /* true during gc0_locked */
#endif
#ifdef TRACEGC
int tracegc = 0;
#endif /* TRACEGC */

/* Circular history of recent expansion/shrinkage activity */
#define EXPANSION_HIST_SIZE  (4)
static long expansionHist[EXPANSION_HIST_SIZE];
static int expansionHistNext = 0;	/* Place next sample at this index */

#ifdef VERBOSEGC

/*
 *  These are variables used by the verbosegc code to maintain various counts
 */

static long init_time;
static long scan_time;
static long sweep_time;
static long compact_time;
static long total_time;

static int reg_marked_object_count;
static int reg_marked_handle_count;
static int stack_marked_object_count;
static int stack_marked_handle_count;
#endif


/* List of unused classes found during GC */
static ClassClass *classFreeList = 0;


/* References and Finalizers */

static ClassClass *classJavaLangRefReference = 0;
static ClassClass *classJavaLangRefSoftReference;
static ClassClass *classJavaLangRefWeakReference;
static ClassClass *classJavaLangRefFinalReference;
static ClassClass *classJavaLangRefFinalizer;
static ClassClass *classJavaLangRefPhantomReference;

static Hjava_lang_ref_Reference
    *discoveredRefs;		/* Ref instances discovered by markChildren */

static Hjava_lang_ref_Reference
    *refEnd;			/* End marker for internal lists */

static HObject
    *refLock = 0;		/* java.lang.ref.Reference.lock */

static Hjava_lang_ref_Reference
    **pendingRefs = 0;		/* java.lang.ref.Reference.pending */

static int64_t
    *softRefClock;		/* java.lang.ref.SoftReference.clock */

static bool_t
    notifyRefLock = FALSE;	/* If true, notify the refLock after resuming
				   multi-threaded operation */

/* Age thresholds for soft references */
#define SOFT_MIN_EXPANDING_AGE	(1)	/* Age below which we don't clear
					   (unless desperate)  */
#define SOFT_MAX_EXPANDING_AGE	(8)	/* Maximum age we allow while
					   expanding */
#define SOFT_MAX_AGE		(32)	/* Maximum age ever allowed */

#ifdef VERBOSEGC
static int refEnqCounts[4];	/* Soft, weak, final, phantom */
static int softRefThreshold;
#endif


static void markPrimitiveClasses(void);

/* [Sheng] The "deferred handle" trick ensures that the recursive 
 * GC scanning routines don't overflow the native stack.
 *
 * - We place a hard upper bound on the depth of native stack by
 *   passing a depth argument to markChildren and markClassClass.
 *
 * - Once the depth bound is exceeded, we do not scan the handle,
 *   but enter it into the "deferred handle" table.
 *
 * - markChildrenFromTop iteratively (not recursively) scans all
 *   the handles in the deferred handle table until the table is
 *   empty. Each iteration may insert more entries into the deferred
 *   handle table. Note that every iteration is guaranteed to make 
 *   some progress, so the process always terminates.
 *
 * - If there is a handle table overflow, we iterate once more
 *   through all the handles in scanHandles() to reconcile the
 *   inconsistencies. Note that we are guaranteed that make some
 *   porgress in each pass, therefore the process always 
 *   terminates.
 *
 * It is rare to run into deep markChildren recursions.
 * (markChildren is already tail recursive, and the limit argument
 * effectively cuts off deep recursions.) It is even rarer to
 * overflow the deferred handle table.
 * 
 * I've run the system with a stack limit of 30 and handle table
 * size of 100. I haven't seen any cases of reiteration in 
 * scanHandles().
 *
 * I've stress tested the system with a stack depth limit of 4 and
 * handle table size of 3, and haven't seen any significant overhead.
 * (Although reiteration did occur in scanHandles().)
 */


struct deferred_handle_t {
  JHandle *handle;
  struct deferred_handle_t *next;
};

#define DEFERRED_HANDLE_TABLE_SIZE 100

static struct deferred_handle_t 
deferredHandleTable[DEFERRED_HANDLE_TABLE_SIZE];

static struct deferred_handle_t *free_deferred_handle_list;
static struct deferred_handle_t *used_deferred_handle_list;

static void initializeDeferredHandleTable(void)
{
    int i;
    used_deferred_handle_list = 0;
    free_deferred_handle_list = deferredHandleTable;
    for (i=0; i<DEFERRED_HANDLE_TABLE_SIZE - 1; i++)
        deferredHandleTable[i].next = deferredHandleTable + i + 1;
    deferredHandleTable[i].next = 0;
}

static struct deferred_handle_t *
putDeferredHandle(JHandle *h)
{
    struct deferred_handle_t *deferred_handle;
    if (free_deferred_handle_list == 0)
        return 0;
    deferred_handle = free_deferred_handle_list;
    free_deferred_handle_list = free_deferred_handle_list->next;

    deferred_handle->handle = h;
    deferred_handle->next = used_deferred_handle_list;
    used_deferred_handle_list = deferred_handle;
    return deferred_handle;
}

static JHandle *
getDeferredHandle(void)
{
    struct deferred_handle_t *deferred_handle;
    if (used_deferred_handle_list == 0)
        return 0;
    deferred_handle = used_deferred_handle_list;
    used_deferred_handle_list = used_deferred_handle_list->next;

    deferred_handle->next = free_deferred_handle_list;
    free_deferred_handle_list = deferred_handle;
    return deferred_handle->handle;
}

typedef struct JNIWeakRef {
    /* handle must be the first field. Code in jni.c assumes this! */ 
    JHandle *handle;
    struct JNIWeakRef *next;
    struct JNIWeakRef *prev;
    void *magic;
} JNIWeakRef;

/* A dummy cell to simplify pointer handling */
static JNIWeakRef weakRefDummy = {0,&weakRefDummy,&weakRefDummy,0};
/* The (doublely-linked) list of weak references created by JNI. */
static JNIWeakRef *weakRefList = &weakRefDummy;

JHandle **
newJNIWeakRef(ExecEnv *ee, JHandle *obj)
{
    JNIWeakRef * ref = (JNIWeakRef *)sysMalloc(sizeof(JNIWeakRef));
    if (ref == NULL)
        return NULL;
    ref->magic = (void *)ref;
    ref->handle = obj;
    HEAP_LOCK(EE2SysThread(ee));
    { 
        JNIWeakRef *prev = weakRefList->prev;
	prev->next = ref;
	ref->prev = prev;
	weakRefList->prev = ref;
	ref->next = weakRefList;
    }
    HEAP_UNLOCK(EE2SysThread(ee));
    return &ref->handle;
}

bool_t 
deleteJNIWeakRef(ExecEnv *ee, JHandle **handleP)
{
    JNIWeakRef * ref = (JNIWeakRef *)handleP;
    /* Sanity check that we are indeed deleting a weak ref. */
    if (ref->magic != (void *)ref)
        return FALSE;
    ref->magic = 0; /* Catch repeated deletes. */
    HEAP_LOCK(EE2SysThread(ee));
    ref->next->prev = ref->prev;
    ref->prev->next = ref->next;
    HEAP_UNLOCK(EE2SysThread(ee));
    sysFree(ref);
    return TRUE;
}

static void 
clearJNIWeakRefs(void)
{
    SetLimits();
    JNIWeakRef *ref = weakRefList->next;
    while (ref != weakRefList) {
        if (ref->handle && !IsMarked(ref->handle)) {
	    ref->handle = 0;
	}
	ref = ref->next;
    }
}

/* Hash table for interned strings.
 * TODO: reduce the number of hash table implementations in the
 * VM. Currently, there are three.
 */
/* HASH_TABLE_SIZE is a prime number, although it doesn't need to be. */
#define HASH_TABLE_SIZE 503

typedef struct string_bucket_type {
    struct string_bucket_type *next;
    Hjava_lang_String *string;
} string_bucket_type;

static string_bucket_type *string_hash_table[HASH_TABLE_SIZE];
static string_bucket_type *free_string_buckets = NULL;
static int n_free_string_buckets = 0;

static unsigned int
string_hash_fun(const unicode *s, int length)
{
    unsigned raw_hash = 0;
    int i;
    for (i = 0; i < length; i++) {
      	raw_hash = raw_hash * 37 + s[i];
    }
    return raw_hash;
}

static int
stringEqual(Hjava_lang_String *str, unicode *chars2, int length2)
{
    struct Classjava_lang_String *strObj1 = unhand(str);
    unicode *chars1;
    int length1 = strObj1->count;
    int i;

    if (length1 != length2)
        return FALSE;
    chars1 = unhand(strObj1->value)->body + strObj1->offset;
    for (i = 0; i < length1; i++) {
        if (chars1[i] != chars2[i])
	    return FALSE;
    }
    return TRUE;
}

Hjava_lang_String *
internString(Hjava_lang_String *str)
{
    Hjava_lang_String *result;
    struct Classjava_lang_String *strObj;
    int length;
    unicode *chars;
    int index;
    string_bucket_type *bucket;
    sys_thread_t *self = sysThreadSelf();

    HEAP_LOCK(self);
    strObj = unhand(str);
    chars = unhand(strObj->value)->body + strObj->offset;
    length = strObj->count;
    index = string_hash_fun(chars, length) % HASH_TABLE_SIZE;
    for (bucket = string_hash_table[index]; bucket; bucket = bucket->next) {
        if (bucket->string == str || 
	    stringEqual(bucket->string, chars, length)) {
	    result = bucket->string;
	    goto unlock;
	}
    }
    
    if (free_string_buckets) {
        bucket = free_string_buckets;
	free_string_buckets = free_string_buckets->next;
	n_free_string_buckets--;
    } else {
        bucket = (string_bucket_type *)sysMalloc(sizeof(string_bucket_type));
	if (bucket == NULL) {
	    result = NULL;
	    goto unlock;
	}
    }

    bucket->string = str;
    bucket->next = string_hash_table[index];
    string_hash_table[index] = bucket;
    result = bucket->string;

 unlock:
    HEAP_UNLOCK(self);
    return result;
}

static void
deleteUnusedInternedStrings(void)
{
    int i;
    SetLimits();

    for (i = 0; i < HASH_TABLE_SIZE; i++) {
        string_bucket_type **bucketP = &(string_hash_table[i]);
	while (*bucketP) {
	    string_bucket_type *bucket = *bucketP;
	    if (IsMarked(bucket->string)) {
	        bucketP = &(bucket->next);
	    } else {
	        *bucketP = bucket->next;
		n_free_string_buckets++;
		bucket->next = free_string_buckets;
		free_string_buckets = bucket;
	    }
	}
    }
}

/* This function must be called in multi-threaded mode. */
static void
freeInternedStringBuckets(void)
{
    /* We keep 10 buckets around. */
    while (n_free_string_buckets > 10) {
        string_bucket_type *bucket = free_string_buckets;
	free_string_buckets = bucket->next;
	n_free_string_buckets--;
	sysFree(bucket);
    }
}

/*
 * Power-of-two allocation size histogram
 */
#ifdef HEAPMETERS
int		enableHeapMeters;
static long	heapMeters[32];

static long	cacheCount;
static long	cacheFlushes;
static long	cacheAllocations;
static long	cacheBlockFills;
static long	cacheHandleFills;

static void
heapMeter(unsigned long i)
{
    register int h = 0;
    if (i & 0xffff0000) { h += 16; i >>= 16; }
    if (i & 0xff00)	{ h +=  8; i >>= 8; }
    if (i & 0xf0)	{ h +=  4; i >>= 4; }
    if (i & 0xc)	{ h +=  2; i >>= 2; }
    if (i & 0x2)	{ h +=  1; }
    heapMeters[h]++;
}

static int
getCacheMeters(sys_thread_t *t, void *arg)
{
    ExecEnv *ee = SysThread2EE(t);
    struct alloc_cache *cp = &ee->alloc_cache;
    cacheCount++;
    cacheAllocations += cp->cache_allocations;
    cacheBlockFills += cp->cache_block_fills;
    cacheHandleFills += cp->cache_handle_fills;
    return SYS_OK;
}

void 
dumpHeapMeters()
{
    int i;
    long n = 0, m = 0;
    sys_thread_t *self = sysThreadSelf();
#define	DUMPHEAPLOWER 3
#define	DUMPHEAPUPPER 24
    for (i = DUMPHEAPLOWER; i < DUMPHEAPUPPER; i++)
	n += heapMeters[i];
    jio_fprintf(stderr, "Heap object size histogram for %d allocations\n", n);
    for (i = DUMPHEAPLOWER; i < DUMPHEAPUPPER; i++) {
	long h = heapMeters[i];
	jio_fprintf(stderr, "%8d %-8d %8d %5.2f %6.2f%\n",
	    (1<<i), ((1<<i)<<1) - 1, h, h * 100.0 / (double)n, 
	    (m += h) * 100.0 / (double)n);
    }
#undef DUMPHEAPLOWER
#undef DUMPHEAPUPPER

    QUEUE_LOCK(self);
    (void) sysThreadEnumerateOver(getCacheMeters, NULL);
    QUEUE_UNLOCK(self);

    jio_fprintf(stderr, "Cache allocation statistics for %d caches\n",
	cacheCount);
    jio_fprintf(stderr, "    Block size = %d, handle count = %d,"
	" local alloc = %d\n",
	allocCacheSize, allocHandleCount, allocLocalSize);
    jio_fprintf(stderr, "    Local allocations = %d, block fills = %d,"
	" handle fills = %d, cache flushes = %d\n",
	cacheAllocations, cacheBlockFills, cacheHandleFills, cacheFlushes);
}
#endif /* HEAPMETERS */

/*
 * Thread-local allocation cache
 */

int allocCacheSize = ALLOC_CACHE_SIZE;
int allocLocalSize = ALLOC_LOCAL_SIZE;
int allocHandleCount = ALLOC_HANDLE_COUNT;

/* Dummy type for cache block */
#define	ALLOC_CACHE_BLOCK_TAG	mkatype(T_MKTYPE(T_UINTEGER, 2), 1)

static long
cacheFillBlock(ExecEnv *ee, struct alloc_cache *cp)
{
    JHandle *block;
    long size;
    sysAssert(cp->cache_busy);
    block = realObjAlloc(ee, ALLOC_CACHE_BLOCK_TAG,
			 ROUND_SIZE(allocCacheSize));
    if (block == NULL)
	return -1;
    cp->cache_tail = block;
    size = obj_geth((OBJECT)(block->obj) - sizeof(hdr));
    sysAssert(size == h_len(size));
#ifdef HEAPMETERS
    cp->cache_block_fills++;
#endif
    return (cp->cache_size = size);
}

static JHandle *
cacheFillHandles(struct execenv *ee, struct alloc_cache *cp)
{
    JHandle *hp, *ret = NULL;
    int i = 0;
    sysAssert(cp->cache_busy);
    sysAssert(cp->cache_handles == NULL);
    HEAP_LOCK(EE2SysThread(ee));
    do {
	POP_FREE_HANDLE(hp);
	if (hp == NULL)
	    goto out;
	sysAssert(hp->obj == 0);
	hp->methods = (struct methodtable *)cp->cache_handles;
	cp->cache_handles = ret = hp;
    } while (++i < allocHandleCount);
  out:
    FreeHandleCtr -= i * sizeof(JHandle);
    sysAssert(FreeHandleCtr >= 0);
    HEAP_UNLOCK(EE2SysThread(ee));
#ifdef HEAPMETERS
    cp->cache_handle_fills++;
#endif
    return ret;
}

/* This function is exported through the JIT interface. */
HObject *
cacheAlloc(ExecEnv *ee, struct methodtable *mptr, long size)
{
    long cache_size;
    OBJECT old_obj, new_obj;
    hdr old_hdr, new_hdr;
    JHandle *old_handle = NULL, *new_handle;
    struct alloc_cache *cp = &ee->alloc_cache;

    size = ROUND_SIZE(size);
    if (size < 0) {  /* overflow */
        goto return_null;
    }

    sysAssert(IS_ALIGNED(size, OBJECTGRAIN));

    /* It's very important to keep cp->cache_busy bit set while we
     * manipulate other cp entries, otherwise the GC thread may
     * clear out the entries that are in use. Most cp fields are
     * decleared to be volatile, but you should make sure your
     * compiler obeys the rules and generates a correct read/write
     * sequence.
     */
    cp->cache_busy = 1;

    if (size >= allocLocalSize) {
        goto cache_missed;
    }

    sysAssert(size < allocLocalSize);

    /* Use <= so we don't have a zero-size block leftover */
    if ((cache_size = cp->cache_size) <= size) {
        if ((cache_size = cacheFillBlock(ee, cp)) <= size) {
	    goto cache_missed;
	}
    }

    if ((new_handle = (JHandle *)(cp->cache_handles)) == NULL) {
        if ((new_handle = cacheFillHandles(ee, cp)) == NULL) {
	    goto cache_missed;
	}
    }

    cp->cache_handles = new_handle->methods;	/* pop */
    old_handle = cp->cache_tail;

    /*
     * [saraiya] The order of stores is important from here on.  We
     * don't want a situation where new_handle points into the middle
     * of the block pointed to by old_handle.  So we first split the
     * block, then point new_handle at the child, then update the type
     * (mptr) of old_handle.
     */

    old_obj = (OBJECT)(old_handle->obj);	/* &aligned */
    sysAssert((void *)old_obj != NULL);
    sysAssert(IS_ALIGNED(old_obj, OBJECTGRAIN));
    old_hdr = obj_geth(old_obj - sizeof(hdr));	/* ld(biased) */
    sysAssert(old_hdr == cache_size);

    new_obj = old_obj + size;			/* &aligned */
    sysAssert(IS_ALIGNED(new_obj, OBJECTGRAIN));
    new_hdr = cache_size - size;
    sysAssert(new_hdr != 0);			/*  */

    obj_seth(new_obj - sizeof(hdr), new_hdr);	/* st(biased) */
    obj_seth(old_obj - sizeof(hdr), size);	/* st(biased) */

    /* BE CAREFUL! From this point on, until new_handle->obj is set,
     * new_obj points to a used object that is not referred to by
     * any handles. The rest of the GC (in particular, the compaction
     * phase) must be able to deal with this.
     *
     * Both new_obj and old_obj are pinned at this point.
     */
    new_handle->methods = ALLOC_CACHE_BLOCK_TAG;
    new_handle->obj = (ClassObject *)new_obj;	/* &aligned */

    old_handle->methods = mptr;

    cp->cache_tail = new_handle;
    cp->cache_size = new_hdr;

#ifdef HEAPMETERS
    cp->cache_allocations++;
#endif

    cp->cache_busy = 0;

    /* keep both new_obj and old_obj pinned */
    if (new_obj) { /* Redundent check, keep new_obj pinned */
	return old_handle;
    }

    /* Never reached */
    KEEP_POINTER_ALIVE(old_obj); /* keep old_obj pinned */

 return_null:
    return NULL;

 cache_missed:
    cp->cache_busy = 0;
    return realObjAlloc(ee, mptr, size);
}

/* Must be called for current thread */
void
allocCacheCleanup(struct alloc_cache *cp)
{
    JHandle *hp1, *hp2;
    struct execenv *ee = EE();
    HEAP_LOCK(EE2SysThread(ee));				/* lock out GC */
    /* Release cache tail */
    sysAssert(!cp->cache_busy);
    cp->cache_tail = NULL;
    cp->cache_size = 0;
    /* Push handles onto global free list */
    hp1 = (JHandle *)(cp->cache_handles);
    cp->cache_handles = NULL;
    while (hp1) {
#ifdef PAGED_HEAPS
	ChunkBlkP MOH_chunk = ChunkBase(hp1);
#endif /* PAGED_HEAPS */
	hp2 = (JHandle *)hp1->methods;
	PUSH_FREE_HANDLE(hp1);
	FreeHandleCtr += sizeof(JHandle);
	hp1 = hp2;
    }
#ifdef HEAPMETERS
    cacheCount++;
    cacheAllocations += cp->cache_allocations;
    cacheBlockFills  += cp->cache_block_fills;
    cacheHandleFills += cp->cache_handle_fills;
#endif

    HEAP_UNLOCK(EE2SysThread(ee));
}

#ifdef DEBUG 
long allocBytes = 0;   /* sum of bytes requested of realObjAlloc */ 
long allocObjects = 0; /* sum of objects allocated */ 
int  allocBogus = 0;   /* true if overflow occurs */ 
#endif /* DEBUG */ 

int64_t
TotalObjectMemory(void)
{
    return int2ll(TotalObjectCtr);
}

int64_t
FreeObjectMemory(void)
{
    return int2ll(FreeObjectCtr);
}

int64_t
TotalHandleMemory(void)
{
    return int2ll(TotalHandleCtr);
}

int64_t
FreeHandleMemory(void)
{
    return int2ll(FreeHandleCtr);
}

static void
clearMarkBits(void)
{

    sysAssert(!markbitsInUse);
    
#ifdef PAGED_HEAPS
    {   /* clear the pinned flag of each object chunk: */
        ChunkBlkP blkP = firstObjBlkP;
	do {
#ifdef LDEBUG
	    checkChunk(blkP);
#endif /* LDEBUG */
	    blkP->chunkFlags = blkP->chunkFlags & ~CHUNK_PINNED;
	    blkP = blkP->nextPtr;
	} while (blkP != firstObjBlkP);
    }
#endif

    clearLocalMarkBits();
    /* clear global markbits pool: */
    if (markbits) {
	memset((char *)markbits, 0, (markend - markbits) * sizeof(*markbits));
    }
}

/* Search one chunk between p and limit for a free block of size "n".
 * If block found, sets block's header, decrements various counters, 
 * and returns ptr.  Otherwise returns NULL.
 */
static void *
searchChunk(register unsigned char *p,
	    register unsigned char *limit,
	    register long n)
{
    /* note that p points to headers, so is not OBJECTGRAIN aligned */
    while (p < limit) {	    		/* search for a free block */
	register unsigned char *next;
	register hdr h = obj_geth(p);
	register hdr h2;		/*excess free space in this block */
	sysAssert(h_len(h) >= sizeof(hdr));
	next = p + h_len(h);
	sysAssert(next <= limit + sizeof(ClassObject));
	if (h_free(h)) {
	    while (h_free(h2 = obj_geth(next))) { 
	      	/*coalesce next free run (why here?) */
		h_bumplen(h, h_len(h2));
		FILL_CHECK_WORD(next);
		next = p + h_len(h);
		obj_seth(p, h);
	    }
	    h2 = h_len(h) - n;
	    if (h2 >= 0) {  /* found block at least large enuf */
		if (h2 > 0) {
		    obj_setlf(p + n, h2, 1);
		    sysAssert(h2 >= OBJECTGRAIN);
		    FILL_CHECK_WORD(p + n);
		}
		obj_setlf(p, n, 0);
		FILL_CHECK_WORD(p);

                sysAssert(IS_ALIGNED(((long) p + sizeof(hdr)), OBJECTGRAIN));
		FreeObjectCtr -= n;
		sysAssert(FreeObjectCtr >= 0);
		return p;
	    }
	}
	p = next;
    } /* while (p < limit) */
    return NULL;
}

#ifdef PAGED_HEAPS /************ PAGED HEAPS: ********************/
/*
 * Allocate N pages, which are aligned and sized "alignment".
 * AllocChunk may use the first few bytes of the page for it's
 * own purposes (the CHUNK_BLK header), and the last chunk may be
 * slightly smaller than a page.
 *
 * Returns a ptr to the chunk block header, which is page aligned.
 * The usable space starts at "startPtr", and is aligned using inner_alignment.
 * The useful size (allocSize) is also aligned using inner_alignment.
 * The PAGE_ALIGNMENT must itself be inner_alignment aligned.
 *
 * Use FreeChunk to free.
 * This does NOT link the chunk into a free list.
 */
static ChunkBlkP
AllocChunk(size_t pages, size_t inner_alignment)
{
    void* chunkHandle;
    unsigned char* alignedPtr = NULL;
    ChunkBlkP chunk = NULL;
    size_t reqSize;
    
    reqSize = (PAGE_ALIGNMENT * pages) - OS_BLOCK_OVERHEAD;
    alignedPtr = sysAllocBlock(reqSize, &chunkHandle);
    
    if (alignedPtr != NULL) {
	size_t alignedHeaderSize = 
          ALIGN_UP(sizeof(ChunkBlk), inner_alignment);

	chunk = (ChunkBlkP)alignedPtr;
	/* if PAGE_ALIGNMENT is inner aligned, then alignedPtr will be also.
	 * if OS_BLOCK_OVERHEAD is inner aligned, 
	 * then reqSize will be aligned, and therefore allocSize and
	 * endPtr will be also.
	 */
	sysAssert((PAGE_ALIGNMENT % inner_alignment) == 0);
	sysAssert((OS_BLOCK_OVERHEAD % inner_alignment) == 0);
#ifdef DEBUG
	chunk->physEndPtr = alignedPtr + reqSize;
	/* save another (at least) 8 bytes for overwrite detection: */
        reqSize -= (inner_alignment > 2*sizeof(int)) ? inner_alignment : 
                                                            2*sizeof(int);
	/* use two words for overwrite protection - the first is more likely 
	 * to stomped by the gc, the second by the user */
        ((int *)(chunk->physEndPtr))[-2] = ALMOST_WORD;
        ((int *)(chunk->physEndPtr))[-1] = ULTIMATE_WORD;
#endif /* DEBUG */

	chunk->chunkHandle = chunkHandle;
	chunk->nextPtr = NULL;				/* caller must link */
	chunk->prevPtr = NULL;
	chunk->allocSize = reqSize - alignedHeaderSize; /* now aligned */
	chunk->startPtr = alignedPtr + alignedHeaderSize; /* now aligned */
	chunk->freePtr = chunk->startPtr;
	chunk->freeCnt = chunk->allocSize;
	chunk->endPtr = chunk->startPtr + chunk->allocSize; /* now aligned */

#ifdef WASTED_SPACE_IN_LEADER
	initWastedSpaceInChunk(chunk);
#ifdef TRACEGC
	if (tracegc) {
	    jio_fprintf(stderr, "<GC: allocated chunk, size %d, wasted %d>\n", 
		    chunk->allocSize, (int)chunk - (int)chunk->wasteStartPtr);
	}
#endif /* TRACEGC */
#endif /* END NOT WASTED_SPACE_IN_LEADER */

#ifdef DEBUG
	checkChunk(chunk);	
#endif /* DEBUG */
    }

    return chunk;
}

static void
FreeChunk(ChunkBlkP chunk)
{
#ifdef DEBUG
   checkChunk(chunk);
#endif /* DEBUG */
   sysFreeBlock(chunk->chunkHandle);  
}

/*
 * n is OBJECTGRAIN-aligned and includes header word
 */
static HObject *
realObjAlloc(ExecEnv *ee, struct methodtable *mptr, long n)
{
    unsigned char *p;
    HObject *ret;
    int overflow_type = OVERFLOW_NONE;
    int on_overflow = OVERFLOW_ACT_GC;
    ChunkBlkP blkP = curObjBlkP;      /* for looping check */

    sysAssert((n & (OBJECTGRAIN-1)) == 0);

    /* Global allocation */
    HEAP_LOCK(EE2SysThread(ee));
#ifdef HEAPMETERS
    if (enableHeapMeters)
	heapMeter((unsigned long)n);
#endif /* HEAPMETERS */

    while (1) {
	/* make sure current object chunk has free space, else look in others.
	 * If none, alloc another object chunk. */
	while (curObjBlkP->freeCnt < n) {
	    if ((curObjBlkP = curObjBlkP->nextPtr) == blkP) {
	    	/* we looped thru all chunks, but no free space */
		overflow_type = OVERFLOW_OBJECTS;   
		goto Overflow;
	    }
	}
	
	/* found a possible chunk */
	if ((p = searchChunk(curObjBlkP->freePtr, curObjBlkP->endPtr, n))
	     != NULL) {
	    goto GotIt;
	} else if ((curObjBlkP->freePtr != curObjBlkP->startPtr) &&
		    /* search other section of the chunk: */
		   ((p = searchChunk(curObjBlkP->startPtr,
		                     curObjBlkP->freePtr, n)) != NULL)) {
	    goto GotIt;
	} else {
	    if ((curObjBlkP = curObjBlkP->nextPtr) != blkP) {
		continue;   /* "while (1)" loop */
	    }
	}
#ifdef DEBUG
#ifdef VERBOSEGC
	if (verbosegc) {
	    jio_fprintf(stderr,
			"<GC: out of object space wanting %d bytes>\n", n);
	}
#endif
#endif /* DEBUG */
	overflow_type = OVERFLOW_OBJECTS;

Overflow:
	/* Allocation failed for lack of space: garbage collect or expand */
	on_overflow = manageAllocFailure(n, overflow_type, on_overflow);
	if (on_overflow == OVERFLOW_ACT_FAIL) {
	    /* Tried everything, couldn't satisfy request, so fail */
	    ret = 0;
	    goto GetOut;
	} else {
	    /* back from GC - try alloc again.
	     * reset anchor point for freespace search */
	    blkP = curObjBlkP;
	}
    } /* while (1) */

GotIt:				/* p points to new object */
#ifdef LDEBUG
    checkChunk(curObjBlkP);
#endif
    curObjBlkP->freeCnt -= n;
    curObjBlkP->freePtr = (unsigned char *) ((long) p + n);
    sysAssert(curObjBlkP->freeCnt >= 0);
    /*
     * Reset overflow act for handles.  If there was also object overflow,
     * we will have escalated the overflow_act.  If we got as far as
     * expanding the object space, overflow_act == OVERFLOW_ACT_DESPERATE,
     * meaning that we'll fail the handle allocation if a handle isn't
     * immediately available.  If this is the case we will have already
     * tried GCing, finalizing and freeing Refs, so it only makes sense to
     * back off the overflow act to OVERFLOW_ACT_EXPAND.  Otherwise we
     * can leave the overflow act as it is.
     */  
    if (overflow_type == OVERFLOW_OBJECTS &&
        on_overflow > OVERFLOW_ACT_EXPAND) {
        on_overflow = OVERFLOW_ACT_EXPAND;
    }
    ret = AllocHandle(mptr, (ClassObject *) (p + sizeof(hdr)));
    if (ret == 0) {
	obj_setfree(p);
	FreeObjectCtr += obj_len(p);
	curObjBlkP->freeCnt += obj_len(p);
	sysAssert(FreeObjectCtr <= TotalObjectCtr);
	/* on_overflow is OVERFLOW_ACT_GC */
	overflow_type = OVERFLOW_HANDLES;
	/* jump into the middle of the "while (1)" loop */
	goto Overflow;
    }

    memset(p + sizeof(hdr), 0, n - sizeof(hdr));

GetOut:
    HEAP_UNLOCK(EE2SysThread(ee));

#ifdef DEBUG
#ifdef VERBOSEGC
    if (verbosegc) {
	long oldBytes = allocBytes;
	allocObjects++;
	allocBytes += n;
	if ((allocBytes < oldBytes)) {
	    allocBogus = 1;
	}
    }
#endif
#endif

    return ret;
}

#else  /************ CONTIGUOUS HEAPS: ********************/

static HObject *
realObjAlloc(ExecEnv *ee, struct methodtable *mptr, long n)
{
    unsigned char *p;
    HObject *ret;
    int overflow_type = OVERFLOW_NONE;
    int on_overflow = OVERFLOW_ACT_GC;

    sysAssert((n & (OBJECTGRAIN-1)) == 0);

    /* Global allocation */
    HEAP_LOCK(EE2SysThread(ee));
#ifdef HEAPMETERS
    if (enableHeapMeters)
	heapMeter((unsigned long)n);
#endif /* HEAPMETERS */

    while (1) {
        /* Found a possible chunk */
        if ((p = searchChunk(opoolhand, opoollimit, n)) != NULL) {
	    goto GotIt;
	} else if ((opoolhand > opool) &&
		   /* Search other section of the chunk */
		   ((p = searchChunk(opool, opoolhand, n)) != NULL)) {
	    goto GotIt;
	} else {   /* We failed to satisfy the allocation request */
	    /*
	     * It is possible that due to coalescing of free memory, and the
	     * fact that we only reset opoolhand if we actually find the space
	     * we are looking for, that after searching the section of the
	     * chunk before opoolhand and still failing to find space we
	     * could leave opoolhand pointing into the interior of a free 
	     * block.  Normally we will just GC in this case, which resets
	     * opoolhand, but under some circumstances this might not happen.
	     * So we conservatively reset opoolhand to the base of the object
	     * space just in case.  Finer tuning is possible but inconvenient.
	     */
	    opoolhand = opool;
	}

#ifdef DEBUG
#ifdef VERBOSEGC
	if (verbosegc) {
	    jio_fprintf(stderr,
			"<GC: out of object space wanting %ld bytes>\n", n);
	}
#endif
#endif /* DEBUG */
	overflow_type = OVERFLOW_OBJECTS;

Overflow:
	/* Allocation failed for lack of space: garbage collect or expand */
	on_overflow = manageAllocFailure(n, overflow_type, on_overflow);
	if (on_overflow == OVERFLOW_ACT_FAIL) {
	    /* Tried everything, couldn't satisfy request, so fail */
	    ret = 0;
	    goto GetOut;
	} /* opoolhand has been reset by GC */
    } /* end while (1). */

GotIt:				/* p points to new object */
    opoolhand = (unsigned char *) ((long) p + n);
    /*
     * Reset overflow act for handles.  If there was also object overflow,
     * we will have escalated the overflow_act.  If we got as far as
     * expanding the object space, overflow_act == OVERFLOW_ACT_DESPERATE,
     * meaning that we'll fail the handle allocation if a handle isn't
     * immediately available.  If this is the case we will have already
     * tried GCing, finalizing and freeing Refs, so it only makes sense
     * to back off the overflow act to OVERFLOW_ACT_EXPAND.  Otherwise
     * we can leave the overflow act as it is.
     */  
    if (overflow_type == OVERFLOW_OBJECTS &&
        on_overflow > OVERFLOW_ACT_EXPAND) {
        on_overflow = OVERFLOW_ACT_EXPAND;
    }
    ret = AllocHandle(mptr, (ClassObject *) (p + sizeof(hdr)));
    if (ret == 0) {
        obj_setfree(p);
	FreeObjectCtr += obj_len(p);
	sysAssert(FreeObjectCtr <= TotalObjectCtr);
	/* on_overflow is OVERFLOW_ACT_GC */
	overflow_type = OVERFLOW_HANDLES;
	goto Overflow;
    }

#ifndef FAST_MONITOR
    memset(p + sizeof(hdr), 0, n - sizeof(hdr));
#else  /* FAST_MONITOR */
    memset(p + sizeof(ohdr), 0, n - sizeof(ohdr));
#endif /* FAST_MONITOR */

GetOut: 
#ifdef HPROF
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_OBJ_ALLOC)) {
        jvmpi_alloc_object(ee, ret);
    }
#endif /* HPROF */
    HEAP_UNLOCK(EE2SysThread(ee));

#ifdef DEBUG
#ifdef VERBOSEGC
    if (verbosegc) {
	long oldBytes = allocBytes;
	allocObjects++;
	allocBytes += n;
	if ((allocBytes < oldBytes)) {
	    allocBogus = 1;
	}
    }
#endif
#endif

    return ret;
}

#endif  /************ END CONTIGUOUS HEAPS ********************/

/*
 * Free handles are linked together by the "methods" field.  There is a
 * seperate list for each handle chunk, pointed to by the chunk's "freePtr"
 * field.  The contiguous heap uses hpoolfreeptr as the head of the free list.
 */
static HObject *
AllocHandle(struct methodtable *mptr, ClassObject *p)
{ 
    JHandle *hanP;
    SetLimits();
    
    if (p == 0) {
	return 0;
    }

#ifdef PAGED_HEAPS
    {
        ChunkBlkP blkP = curHanBlkP;
        /* make sure current handle chunk has free space, else look in others.
         * If none, fail. */
        while (curHanBlkP->freePtr == 0) {
  	    curHanBlkP = curHanBlkP->nextPtr;
	    if (curHanBlkP == blkP) {
	        goto NoLuck;
	    }
        }
    }
#endif /* PAGED_HEAPS */
   
    POP_FREE_HANDLE(hanP);
    if (hanP == NULL) {
        goto NoLuck;
    }

    sysAssert(ValidObject(p));

    /* should assert heap is locked */
    sysAssert(hanP->obj == 0);		    /* make sure it's really free */
    sysAssert(ValidHandle(hanP));

#ifdef PAGED_HEAPS
    curHanBlkP->freeCnt -= sizeof(JHandle); /* update counters...*/
    sysAssert(curHanBlkP->freeCnt >= 0);
    /* these should run out at the same time: */
    sysAssert((curHanBlkP->freeCnt == 0) == (curHanBlkP->freePtr == 0));
#ifdef LDEBUG
    checkChunk(curHanBlkP);
#endif
#endif /* PAGED_HEAPS */
    hanP->methods = mptr;		    /* setup new handle... */
    hanP->obj = (ClassObject *) p;
    FreeHandleCtr -= sizeof(JHandle);
    sysAssert(FreeHandleCtr >= 0);
    return (HObject *) hanP;

NoLuck:
#ifdef DEBUG
#ifdef VERBOSEGC
    if (verbosegc) {
	jio_fprintf(stderr, "<GC: out of handle space>\n");
    }
#endif
#endif /* DEBUG */
    return 0;
}


/*
 * Calculate the size in bytes of an array of type t and length l.
 */
int
sizearray(int t, int l)
{
    return ((t == T_CLASS) ? l * sizeof(JHandle*) : l << T_SIZEFIELD(t));
}

ClassClass *
allocClass(ExecEnv *ee)
{
    ClassClass *cb = (ClassClass *) 
        cacheAlloc(ee, 
		   classJavaLangClass ? cbMethodTable(classJavaLangClass) : 0,
		   sizeof(Classjava_lang_Class));
    if (cb != 0) {
	Classjava_lang_Class *ucb = unhand(cb);
	ucb->HandleToSelf = cb;
    }
    return cb;
}

/*
 * Allocate an array of type 't' with initial size 'l'.	 This routine
 * takes care of setting the method table pointer correctly.  It also
 * takes into consideration any special sizeing requirements, like 
 * those needed for arrays of classes.
 */
HObject *
allocArray(ExecEnv *ee, int t, int l)
{
    int size;

    sysAssert(t >= T_CLASS && t < T_MAXNUMERIC);

    /* Check if the array length can be stored in handles */
    if (l >> (sizeof(int) * 8 - METHOD_FLAG_BITS))
 	return 0;

    size = sizearray (t, l);

    if (t == T_CLASS)
	size += sizeof(OBJECT); /* for element type */

    return cacheAlloc(ee, mkatype(t, l), size);
}

/*
 * Allocate an object. This routine
 * should not be used to allocate arrays; use allocArray for that.
 */
HObject *
allocObject(ExecEnv *ee, ClassClass *cb)
{
    HObject *handle = cacheAlloc(ee, cbMethodTable(cb), cbInstanceSize(cb));

    /*
     * The class of the object we are allocating must either be
     * initialized or we must currently be executing inside the
     * static initializer for that class.
     */
    sysAssert(CCIs(cb, Initialized) || cbInitThread(cb) == ee);

    /*
     * If the class has a finalization method, create a finalizer for it.
     * Note that handle can be nil if we are out of memory!
     */

    if (handle) {
	sysAssert(obj_flags(handle) == T_NORMAL_OBJECT);
	if (cbFinalizer(cb) != NULL) {
	    execute_java_static_method(ee, classJavaLangRefFinalizer,
				       "register", "(Ljava/lang/Object;)V",
				       handle);
	}
    }

    return handle;
}

/* Called for expandHandleSpace - zero obj fields, and link methods fields 
 * into a free list. */
static void 
linkNewHandles(JHandle *start, JHandle *limit, unsigned char *oldFreeList)
{
    register JHandle *hp = start;
    register JHandle *hanLimit = limit;

    /*CONSTCOND*/
    while (1) {
	register JHandle *nexthp = hp + 1;
	hp->obj = 0;				/* set each handle free */
	if (nexthp == hanLimit) {
	   hp->methods = (struct methodtable *)oldFreeList;
	   break;
	} else {
	    /* link free handles via methods field */
	    hp->methods = (struct methodtable *)nexthp;	
	    hp = nexthp;
	}
    } /* while */
}

#ifdef PAGED_HEAPS /************ PAGED HEAPS: ********************/

/* When we get a new chunk outside of the range mapped by the page_map,
 * then we need to allocate a new page map, copy the old one into the
 * right place, set globals, etc. Returns true or false.
 */ 
static int 
reallocPageMap(ChunkBlkP blkP) 
{
    if (!page_map) { /* initial case */
	size_t map_len;
	
	mem_base = blkP;
	mem_top = (void *)
               ALIGN_UP((int)blkP->endPtr, PAGE_ALIGNMENT);
	
	map_len = PageMapLength() * sizeof(PageMapEntry);
	page_map = sysMalloc(map_len);
	if (!page_map) {
	    goto FAIL;
	}
	
#ifdef TRACEGC
	if (tracegc) {
	    jio_fprintf(stderr, "<GC: initialized page map to hold %d pages>\n",
		    PageMapLength());
	}
#endif

	/* clear new entries */
	memset(page_map, 0, map_len);
    } else {
	void* old_mem_base = mem_base;
	size_t old_map_len = PageMapLength();
	void* old_page_map = page_map;
	/* offset in pages of where old entries will start in new page map. */
	size_t page_offset;
	
	if ((void *)blkP < mem_base) {
	    mem_base = (unsigned char *)blkP;
	} else if ((void *)blkP >= mem_top) {
	    mem_top = (void *)
               ALIGN_UP((int)blkP->endPtr, PAGE_ALIGNMENT);
	}
    
	page_map = sysMalloc(PageMapLength() * sizeof(PageMapEntry));
	if (!page_map) {
	    goto FAIL;
	}
	
	/* note that old_mem_base must be >= mem_base */
	page_offset = ((int)old_mem_base - (int)mem_base) / PAGE_ALIGNMENT;

#ifdef TRACEGC
	if (tracegc) {
	    jio_fprintf(stderr, 
	    "<GC: new page map starts %d pages earlier than the old one>\n",
		    page_offset);
	}
#endif

	/* clear all entries */
	memset(page_map, 0, PageMapLength() * sizeof(PageMapEntry));
	/* copy old entries */
	memmove((unsigned char *)page_map +
	         (page_offset * sizeof(PageMapEntry)),
		old_page_map,
		old_map_len * sizeof(PageMapEntry));
	sysFree(old_page_map);
    }
    return 1;
FAIL:
#ifdef VERBOSEGC
    if (verbosegc) {
	jio_fprintf(stderr, "<GC: couldn't allocate page map>\n");
    }
#endif
    return 0;

}

/* Allocate mark bits for a page. Either allocate into
 * the chunk, or into our global markbits pool.
 */ 
static void* 
allocMarkBits(ChunkBlkP blkP) 
{
    void* result = NULL;

    result = allocMarkBitsLocally(blkP);
    if (result != NULL) {
#ifdef TRACEGC
	if (tracegc) {
	    jio_fprintf(stderr, "<GC: page's mark bits allocated within chunk>\n");
	}
#endif
        return result;
    }

#ifdef LDEBUG
    sysAssert(page_map);
    sysAssert(!markbitsInUse);
#endif
    if ((markbitsfreeptr + MARK_BITS_SIZE) > 
        ((unsigned char *)markbits + marksize)) {
	/* allocate new mark bits pool (this catches the intial case too ): */
	/* # of full pages used by chunk: */
	int numPages = (blkP->allocSize / PAGE_ALIGNMENT) + 1;
	void* oldPool = (char *)markbits;
	void* oldPoolEnd = markbitsfreeptr;
	int i;
	int limit = PageMapLength();

	/* mark bits only have to survive the extent of a gc, so toss old:*/
	if (markbits) {
	    sysFree((char *)markbits); 
	}

	/* DRW REMIND: Look at this */
	/* # of new pages we want to handle (leave growing room) */
	numPages = max(numPages, MIN_OBJ_PAGES) * 2; 
	marksize += (MARK_BITS_SIZE * numPages);
	markbits = (unsigned int*) sysMalloc(marksize);
	if (markbits == NULL) {
#ifdef VERBOSEGC
	    if (verbosegc) {
		jio_fprintf(stderr, "<GC: couldn't allocate new mark bits pool>\n");
	    }
#endif
	    return 0;
	}
#ifdef TRACEGC
	if (tracegc) {
	    jio_fprintf(stderr, "<GC: growing global mark bits pool to %d bytes>\n",
	            marksize);
	}
#endif		
	markbitsfreeptr = (unsigned char *)markbits;
    
	/* Walk over page_map, and update the mark_bits fields 
	 * if they used to come from the old pool. */
	if (oldPool) {
	    for (i = 0; i < limit; i++) {
		void* oldMarkBitsPtr = (void *) page_map[i].mark_bits;
		if ((oldMarkBitsPtr >= oldPool) &&
		    (oldMarkBitsPtr < oldPoolEnd)) {
		    page_map[i].mark_bits = (unsigned int*)markbitsfreeptr;
		    markbitsfreeptr += MARK_BITS_SIZE;
		}
	    }
	}
    }

#ifdef TRACEGC
    if (tracegc) {
	jio_fprintf(stderr, "<GC: using global mark bits pool>\n");
    }
#endif
    result = markbitsfreeptr;
    markbitsfreeptr += MARK_BITS_SIZE;
    
    sysAssert(markbitsfreeptr < 
              ((unsigned char *)markbits + marksize + MARK_BITS_SIZE));
    return result;  
}

/* Set up entries in the page map for new chunk, and allocate mark bits.
 * Returns true or false. If heapType == FREE_HEAP_TYPE, it will clear
 * map entries for the chunk.
 */
static int 
setPageMap(ChunkBlkP blkP, int heapType) 
{
    int numPages = PTR_2_PAGENUM(blkP->allocSize) + 1;
    int firstMapIndex;
    int i;
    
    if (((void *)blkP < mem_base) || ((void *)blkP >= mem_top)) {
	if (!reallocPageMap(blkP)) {
	    return 0;
	}
    }
    
    firstMapIndex = PTR_2_PAGENUM((int)blkP - (int)mem_base);
    for (i = 0; i < numPages; i++) {
	int mapIndex = firstMapIndex + i;
	
	if (heapType == FREE_HEAP_TYPE) {
	    page_map[mapIndex].page_number = 0;
	    page_map[mapIndex].mark_bits = 0;	
	    page_map[mapIndex].chunk_size = 0;
	} else {
	    void* markBits = allocMarkBits(blkP);
	    
	    if (markBits == NULL) {
		return 0;
	    }
	    
	    sysAssert(!page_map[mapIndex].mark_bits);
	    page_map[mapIndex].page_number = i;
	    page_map[mapIndex].mark_bits = markBits;	
	    if (heapType == HANDLE_HEAP_TYPE) {
		page_map[mapIndex].chunk_size = -1;
	    } else {
		page_map[mapIndex].chunk_size = numPages;
	    }
	}
    }
    return 1;
}

/* insert new chunk into a list in order (currently address order).
 * May modify chunkPtr pointers.
 */
static void
insertChunk(ChunkBlkP* first, ChunkBlkP* last, ChunkBlkP chunk)
{   
    if (!(*first)) {		/* first & last always defined together */
	(*first) = chunk;
	(*last) = chunk;
	chunk->nextPtr = chunk;	/* chain to itself */
	chunk->prevPtr = chunk;
    } else {	    
	ChunkBlkP chunksNextP = 0;
	/* when we find correct location, chunksNextP will will be set */
	sysAssert((*first)->prevPtr == (*last));
	sysAssert((*last)->nextPtr == (*first));
	for (chunksNextP = (*first); 
	     /* - */;
	     chunksNextP = chunksNextP->nextPtr) {
	    if (chunksNextP > chunk) {		/* found link */
		if (chunksNextP == (*first)) {
		    (*first) = chunk;
		}
		break;
	    } else if (chunksNextP == (*last)) { /* end of list */
		chunksNextP = chunksNextP->nextPtr; /* should eql (*first) */
		(*last) = chunk;
		break;
	    }
	}
	chunk->nextPtr = chunksNextP;		/* setup chunk */
	chunk->prevPtr = chunksNextP->prevPtr;
	chunksNextP->prevPtr = chunk;		/* link in chunk */
	chunk->prevPtr->nextPtr = chunk;
    }
    sysAssert((*first)->prevPtr == (*last));
    sysAssert((*last)->nextPtr == (*first));
#ifdef DEBUG
    verifyPageMap(); /* callers add to page map first, so now in synch */
#endif
}

static void
removeChunk(ChunkBlkP* first, ChunkBlkP* last, 
            ChunkBlkP* current, ChunkBlkP chunk)
{ 
#ifdef LDEBUG
    checkChunk(chunk);
    checkChunk((*first));
    checkChunk((*last));
    checkChunk((*current));
#endif

    /* if only one chunk, this doesn't change anything: */
    chunk->prevPtr->nextPtr = chunk->nextPtr;	
    chunk->nextPtr->prevPtr = chunk->prevPtr;	/* "" */
    
    chunk->prevPtr = NULL;
    chunk->nextPtr = NULL;

    if ((*first) == chunk) {		/* update globals */
	(*first) = (*last)->nextPtr;
    }
    if ((*last) == chunk) {
	(*last) = (*first)->prevPtr;
    }
    if ((*current) == chunk) {
	(*current) = (*first);		/* which one should we pick??? */
    }

#ifdef DEBUG
    if (*first) {
	sysAssert((*first)->prevPtr == (*last));
    }
    if ((*last)) {
	sysAssert((*last)->nextPtr == (*first));
    }
#endif
    sysAssert(((*first) == NULL) == ((*last) == NULL));
    sysAssert(((*first) == NULL) == ((*current) == NULL));
}

static void
shrinkObjectSpace(ChunkBlkP chunk)
{
    if (verbosegc) {
	jio_fprintf(stderr, "<GC: freeing object chunk 0x%lx>\n", chunk);
    }

#ifdef LDEBUG
    sysAssert(chunk->allocSize == chunk->freeCnt);
#endif /* LDEBUG */
    FreeObjectCtr -= chunk->allocSize;
    TotalObjectCtr -= chunk->allocSize;
    expanded += -chunk->allocSize;
    removeChunk(&firstObjBlkP, &lastObjBlkP, &curObjBlkP, chunk);
    setPageMap(chunk, FREE_HEAP_TYPE);
    FreeChunk(chunk);
}

/*
 * Expand the handle space.  The behavior of the handle space is simpler
 * than that of the object space: handles never move and the space can't
 * be fragmented, but that also means that contracting the handle space
 * is easy to block.
 * 
 * Be sure to zero all the new handles: a free handle is recognized by
 * having its obj field zeroed.	 Remember that the base of the handle
 * space must be HANDLEGRAIN aligned within the committed memory.
 *
 * Expand the handle space by allocating several pages in one chunk.
 * Allocate markbits as needed for the chunk.  Sets curHanBlkP on exit.
 *
 */
static int
expandHandleSpace(int needPages)
{
    ChunkBlkP blkP = NULL;
    int numPages;

    /* sysAssert(HEAP_LOCKED(ee)); ...except during InitializeAlloc */

    for (numPages = max(needPages, MIN_HANDLE_PAGES);
	 numPages >= needPages;
	 numPages--) {
        if (((numPages * PAGE_ALIGNMENT) <= 
	     (gMaxMem - (TotalObjectCtr + TotalHandleCtr))) &&
	    ((blkP = AllocChunk(numPages, HANDLEGRAIN)) !=
	     NULL)) {
	    break;
	}
    }

    if (blkP) {
	int szBlk = blkP->allocSize;
	
	if (!setPageMap(blkP, HANDLE_HEAP_TYPE)) {
	    return 0;	
	}

	if (verbosegc) {
	    jio_fprintf(stderr, 
	          "<GC: expanded handle space by 0x%lx bytes, chunk = 0x%lx>\n",
	           szBlk, blkP);
	}

	insertChunk(&firstHanBlkP, &lastHanBlkP, blkP);
	curHanBlkP = blkP;		      /* mark this block currrent */
	
	linkNewHandles((JHandle *)blkP->startPtr, (JHandle *)blkP->endPtr, 
		       blkP->freePtr);
	blkP->freePtr = blkP->startPtr;

	FreeHandleCtr += szBlk;
	TotalHandleCtr += szBlk;
	return 1;	    /* alloc worked */
    }
    if (verbosegc) {
        jio_fprintf(stderr, "<GC: couldn't allocate new handle chunk>\n");
    }

    return 0;		/* alloc failed */
}

/*
 * Expand the object space.  To go on we need to at least cover the
 * currently failing allocation, meaning we need a contiguous chunk of
 * enough pages to hold something of its size.
 * However, we normally also want to leave a user-controllable amount
 * of preallocated space free *after* we have satisfied the current
 * allocation.	*** DRW: Not Yet Implemented. ****
 *
 * The last_free parameter passed in here is ignored in the segemented version,
 * because it's almost impossible to glue two chunks together in the kind of
 * machines this segmented code will run on.
 *
 * Allocate markbits as needed for the chunk.  Sets curObjBlkP on exit.
 */
static int
expandObjectSpace(long sz, unsigned char *last_free)
{
	/* last_free is ignored */
	ChunkBlkP blkP = NULL;
	int numPages;
	int needPages;

#ifdef TRACEGC
	if (tracegc) {
	    jio_fprintf(stderr, "Going to expand because we need sz=0x%lx bytes\n",
		    sz);
	    if (firstObjBlkP) {
		DumpHeapInfo();
	    }
	}
#endif /* TRACEGC */

	/* sysAssert(HEAP_LOCKED(ee)); ...except during InitializeAlloc */

	/*
	 * AllocChunk may not give us exactly nPages*alignemnt bytes, 
	 * so ask for extra for now.
	 */
	needPages = PTR_2_PAGENUM(sz) + 2; 
	for (numPages = max(needPages, MIN_OBJ_PAGES);
	     numPages >= needPages;
	     numPages--) {
	    if (((numPages * PAGE_ALIGNMENT) <= 
		 (gMaxMem - (TotalObjectCtr + TotalHandleCtr))) &&
		(((blkP = AllocChunk(numPages, OBJECTGRAIN)) !=
		  NULL))) {
	        break;
	    }
	}

	if (blkP) {
	    int szBlk;
    	    /** Goofiness because we need the object ptr to be OBJECTGRAIN 
    	     ** aligned, and we store a header before that, so reset
    	     ** allocation ptrs to make it so. */
    	    /* startPtr is pointer to the HEADER! */
    	    blkP->startPtr = blkP->startPtr + OBJECTGRAIN - sizeof(hdr);
    	    blkP->freePtr = blkP->startPtr;
    	    /* keep end aligned, and make room for an "out of bounds" 
    	     * header, which also indicates end of chunk. */
    	    blkP->endPtr = blkP->endPtr - OBJECTGRAIN + sizeof(hdr) 
    	                   - OBJECTGRAIN;
    	    blkP->allocSize = (long)(blkP->endPtr - blkP->startPtr);
    	    blkP->freeCnt = blkP->allocSize;
    	
    	    szBlk = blkP->allocSize;
    
    	    if (!setPageMap(blkP, OBJECT_HEAP_TYPE)) {
    	        return 0;
	    }
    	    sysAssert(szBlk >= sz);
	    FILL_CHECK_WORD(blkP->startPtr + sizeof(hdr) + sizeof(int*));
#ifdef LDEBUG
    	    memset(blkP->startPtr, 0x55, szBlk);
#endif /* LDEBUG */
            if (verbosegc) {
    	        jio_fprintf(stderr, 
		 "<GC: expanded object space by 0x%lx bytes, chunk = 0x%lx>\n",
	           szBlk, blkP);
	    }

    	    insertChunk(&firstObjBlkP, &lastObjBlkP, blkP);
    	    curObjBlkP = blkP;			    /* make block currrent */
    	    obj_setlf(blkP->startPtr, szBlk, 1);    /* mark chunk free */
    	    /* zero-length block used to terminate some searches 
    	     * in realObjAlloc */
    	    obj_setlf(blkP->endPtr, 0, 0);	     
#ifdef DEBUG
    	    checkChunk(blkP);
#endif /* DEBUG */
    	    FreeObjectCtr += szBlk;
    	    TotalObjectCtr += szBlk;
	    expanded += szBlk;
    	    return 1;				/* alloc worked */
	}
	if (verbosegc) {
	    jio_fprintf(stderr, "<GC: couldn't allocate new object chunk>\n");
	}

	return 0;				/* alloc failed */
}

#else  /************ CONTIGUOUS HEAPS: ********************/

/*
 * Returns the required number of bytes to cover that active region
 * of the handle space and object space.
 */

static size_t
computeMarkSize(void)
{
    return (opoollimit - hpool + OBJECTGRAIN*BITSPERMARK - 1)/
	(OBJECTGRAIN*BITSPERMARK) * sizeof(*markbits);
}

/*
 * Expand the mark bits array.	We don't care whether the expansion that
 * led to this is of the handle or object space, just that the contiguous
 * memory that the mark bits must span has increased.
 *
 * There is no slack in expanding the mark bits: we have already alloca-
 * ted the additional memory that the increment represents, so we have
 * got to be able to allocate mark bits to cover it.  We don't need to
 * initialize the new mark bits: they're zeroed on the start of each GC.
 *
 * Note: Given that sysCommitMem() should be assumed to actually allocate
 * swap space, it is arguable that we should put off committing memory
 * for the mark bits until GC actually needs it.  But GC is pretty much
 * inevitable in Java, so it's not clear there's any real benefit.
 */
static int
expandMarkBits(void)
{
    unsigned char * commit_ptr;
    size_t commit_size;
    size_t actual_commit_size;
    size_t newmarksize;
    size_t marksize;

    sysAssert(!markbitsInUse);

    /* compute the current and required mark bits sizes */
    newmarksize = computeMarkSize();
    marksize = (markend - markbits) * sizeof(*markbits);
    if (marksize >= newmarksize) {
	/* we have enough left from the last time we expanded. */
	return 1;
    }

    commit_ptr = (unsigned char *)markend;
    commit_size = newmarksize - marksize;

#ifdef VERBOSEGC
    if (verbosegc) {
	jio_fprintf(stderr,
		    "<GC: need to expand mark bits to cover %d bytes>\n",
		    commit_size);
    }
#endif /* VERBOSEGC */

    commit_ptr = sysCommitMem(commit_ptr, commit_size, &actual_commit_size);
    if (!commit_ptr) {
#ifdef VERBOSEGC
	if (verbosegc) {
	    jio_fprintf(stderr, "<GC: expansion of mark bits failed>\n");
	}
#endif /* VERBOSEGC */
	return 0;
    }
    sysAssert(commit_ptr == (unsigned char *)markend);
    sysAssert(actual_commit_size >= commit_size);
    /*
     * compute the new end of the mark bits, which because of page rounding
     * in sysCommitMem may actually be greater than our needs
     */
    markend = markend + actual_commit_size / sizeof(*markbits);
    sysAssert(markend <= markmax);
    return 1;
}

static int
shrinkMarkBits(void)
{
    unsigned char * decommit_ptr;
    size_t decommit_size;
    size_t actual_decommit_size;
    size_t newmarksize;
    size_t marksize;

    sysAssert(!markbitsInUse);

    newmarksize = computeMarkSize();
    marksize = (markend - markbits) * sizeof(*markbits);
    sysAssert(marksize > newmarksize);

    decommit_size = marksize - newmarksize;
    decommit_ptr = (unsigned char *)markbits + newmarksize;

#ifdef VERBOSEGC
	if (verbosegc) {
	jio_fprintf(stderr,
		    "<GC: need to shrink mark bits to skip %d bytes>\n",
		    decommit_size);
	}
#endif /* VERBOSEGC */

    decommit_ptr = sysDecommitMem(decommit_ptr, decommit_size,
				  &actual_decommit_size);
    if (!decommit_ptr) {
#ifdef VERBOSEGC
	if (verbosegc) {
	    jio_fprintf(stderr, "<GC: shrinking of mark bits failed>\n");
    }
#endif /* VERBOSEGC */
	return 0;
    }
    sysAssert(decommit_ptr + actual_decommit_size == (unsigned char *)markend);
    sysAssert(actual_decommit_size <= decommit_size);
    markend = markend - actual_decommit_size / sizeof(*markbits);
    return 1;
}

/*
 * Expand the handle space.  The behavior of the handle space is simpler
 * than that of the object space: handles never move and the space can't
 * be fragmented, but that also means that contracting the handle space
 * is easy to block.
 *
 * If we were quicker to allocate new handles than GC to reclaim handles,
 * initially expanding by a fixed amount would be quick and easy.  But
 * given the attempt to stay small as much as possible, we are instead
 * using a strategy similar to the object store's, where we GC first,
 * and if we have to expand, we preallocate new space that scales with
 * the amount of existing space, subject to a minimum.
 * 
 * Be sure to zero all the new handles: a free handle is recognized by
 * having its obj field zeroed.	 Remember that the base of the handle
 * space must be HANDLEGRAIN aligned within the committed memory.
 *
 * If we are going to seriously try to contract the handle space we
 * should instead allocate handles from higher addresses toward lower
 * addresses.
 */

#define MIN_HANDLE_EXPANSION 4*1024

static int
expandHandleSpace(int numPages)
{   /* ignore numPages */
    size_t incr;
    size_t preallocIncr;
    unsigned char * oldhpool = hpool;

    /*
     * Expand by the maximum of a constant and the amount we need to
     * make FreeObjectCtr-n >= TotalObjectCtr*minHeapFreePercent.  The
     * constant 1 rounds up after the division.
     */
    preallocIncr = (size_t)(1 + ((TotalHandleCtr * minHeapFreePercent -
				  (FreeHandleCtr - (long)sizeof(JHandle))) /
				 (1 - minHeapFreePercent)));

    incr = max(MIN_HANDLE_EXPANSION, preallocIncr);

    /* Always scale incr up to a multiple of HANDLEGRAIN */
    incr = ALIGN_UP((incr),HANDLEGRAIN);

    if (hpool-incr < heapbase) {
        if (hpool <= heapbase) {
            if (!hpoolfreelist) {
                if (verbosegc) {
                    jio_fprintf(stderr,
                                "<GC: run out of handle space>\n");
                }
                return 0;
            }
            return 1;
        }
        incr = hpool - heapbase;
    }

    hpool = sysCommitMem(hpool - incr, incr, &incr);
    if (!hpool) {
        hpool = oldhpool;
        if (verbosegc) {
            jio_fprintf(stderr, "<GC: expansion of handle space failed>\n");
        }
        return (hpoolfreelist != 0);
    }
    /* Need to be HANDLEGRAIN aligned within committed memory */
    hpool = (unsigned char *)
        ALIGN_UP((intptr_t) hpool,(size_t)(HANDLEGRAIN));
    
    linkNewHandles((JHandle*)hpool, (JHandle*)oldhpool, hpoolfreelist);
    hpoolfreelist = hpool;
    FreeHandleCtr  += incr;
    TotalHandleCtr += incr;
#ifdef VERBOSEGC
    if (verbosegc) {
        jio_fprintf(stderr,
                    "<GC: expanded handle space by %d to %ld bytes, %ld%% free>\n",
                    incr, TotalHandleCtr,
		    100 * FreeHandleCtr / TotalHandleCtr);
    }
#endif
    return expandMarkBits();
}

/*
 * Expand the object space.  To go on we need to at least cover the
 * currently failing allocation, meaning we need a contiguous chunk of
 * its size minus the amount of space in the trailing free chunk, if any.
 * However, we normally also want to leave a user-controllable amount
 * of preallocated space free *after* we have satisfied the current
 * allocation.	This preallocated space is calculated as a percentage
 * of the total free space, so given possible fragmentation, that free
 * space does not guarantee large contiguous allocations.  But given
 * that most allocations are not that large, preallocation can generally
 * reduce the number of interruptions due to memory management operations.
 * The user controls the rate of preallocation by varying the value of
 * the minHeapFreePercent variable.
 *
 * [REMIND: Currently minFreeHeapPercent is global.  That is probably right
 * for a single application, but could it lead to applets fighting over
 * its value???]
 *
 * The last_free parameter passed in here is the last free chunk
 * of memory as determined by the most recent heap compaction in
 * *this* round of memory management.  If the last chunk isn't known,
 * last_free is 0.  There is no guarantee that the last chunk abuts
 * the end of the object space, but if it does, any new memory allocated
 * here is coalesced with the last chunk.
 *
 * Because we are expanding at the top of the space, we leave opoolhand
 * where it is.
 */
static int
expandObjectSpace(long n, unsigned char *last_free)
{
    size_t incr;
    long preallocIncr;
    unsigned char *incrptr;

    /*
     * Expand by the maximum of the minimum we need to satisfy the current
     * allocation (taking into account the last_free chunk) and the amount
     * we need such that:
     * FreeObjectCtr-n >= TotalObjectCtr*minHeapFreePercent.
     *
     * The constant 1 rounds up after the division.
     */
    preallocIncr = (long)(1 + ((TotalObjectCtr*minHeapFreePercent -
				(FreeObjectCtr-n)) /
			       (1 - minHeapFreePercent)));

    /*
     *  Clamp the allocation to the min and max expansion values
     */
    if (preallocIncr > maxHeapExpansion)
	preallocIncr = maxHeapExpansion;
    else if (preallocIncr < minHeapExpansion)
	preallocIncr = minHeapExpansion;
	
    /*
     * Clamp the max allocation to maximum free space in the heap.
     */
    if (preallocIncr > (heaptop - (opoollimit + sizeof(hdr))))
	preallocIncr = heaptop - (opoollimit + sizeof(hdr));

    if (last_free) {
        if (obj_len(last_free) >= n) {
            /* The last_free satisfy the request. */
            incr = 0;
        } else if (last_free + obj_len(last_free) == opoollimit) {
            /* There is a free block at the end of heap.
               We only need to expand the difference. */
	    incr = n - obj_len(last_free);
	} else {
            /* No free space at the end of heap.
               We have to expand the full size. */
	    incr = n;
	}
    } else {
        incr = n;
    }

    if (preallocIncr > 0) {	/* ## Temporary defensive measure */
	incr = max(incr, preallocIncr);
    }
    /* Always scale incr up to a multiple of OBJECTGRAIN */
    incr = ALIGN_UP((incr),(size_t)(OBJECTGRAIN));

    /* Expansion request should succeed when we don't need to expand. 
     * If we don't return here, sysCommitMem will return 0 and cause 
     * the memory expansion request to fail.
     */
    if (incr == 0)
        return 1;

    if (opoollimit + incr > heaptop) {
        if (verbosegc) {
	    jio_fprintf(stderr, "<GC: tried to expand object space over limit>\n");
	}
	return 0;
    } else {
        /* We could go on with anything down to n - obj_len(last_free) */
       incrptr = sysCommitMem(opoollimit + sizeof(hdr), incr, &incr);  
       if (!incrptr) {
	   if (verbosegc) {
	       jio_fprintf(stderr, "<GC: expansion of object space failed>\n");
	   }
	   return 0;
       }
       sysAssert(incrptr == opoollimit + sizeof(hdr));
       sysAssert(incrptr + incr <= heaptop);
       if (last_free && last_free + obj_len(last_free) == opoollimit) {
  	  /*
	   * The last free chunk abuts the end of the object space, so
	   * we can coalesce it with the new memory.
	   */
	   obj_bumplen(last_free, incr);
	   FILL_CHECK_WORD(last_free);
       } else {
	  /*
	   * The last free chunk either doesn't abut the end of the
	   * object space or we don't know where it is, so we don't
	   * try to coalesce the new space.
	   */
	   obj_setlf(opoollimit, incr, 1);
	   FILL_CHECK_WORD(opoollimit);
       }
       opoollimit = opoollimit + incr;
       obj_setlf(opoollimit, 0, 0);
       FreeObjectCtr  += incr;
       TotalObjectCtr += incr;
       expanded += incr;
       if (verbosegc) {
	 jio_fprintf(stderr,
		 "<GC: expanded object space by %d to %ld bytes, %ld%% free>\n",
		 incr, TotalObjectCtr, 100 * FreeObjectCtr / TotalObjectCtr);
       }
       return expandMarkBits();
    }
}

/*
 * Release the swap associated with some portion of the end of the object
 * space and move opoollimit down so that space is no longer used.
 */

static void
shrinkObjectSpace(size_t extra_space, unsigned char **plast_free)
{
    unsigned char *last_free = *plast_free;
    unsigned char * decommit_ptr;
    size_t actual_decommit_space;

    /* we can't return memory if the last free block isn't at the end */
    if (last_free == NULL ||
	last_free + obj_len(last_free) < opoollimit) {
	return;
    }

    if (extra_space > obj_len(last_free)) {
	/* we have don't have that much free space in the last object */
	extra_space = obj_len(last_free);
    }

    decommit_ptr = sysDecommitMem(opoollimit - extra_space + sizeof(hdr),
				  extra_space, &actual_decommit_space);
    if (decommit_ptr == NULL)
	return;

    sysAssert(decommit_ptr - (opoollimit - extra_space + sizeof(hdr)) ==
	      extra_space - actual_decommit_space);
    /* adjust all of the heap related pointers */
    sysAssert(obj_len(last_free) >= actual_decommit_space);
    /* if last_free is totally decommited, reset *plast_free to 0. */
    /* otherwise adjust the size of last_free, leave room for sentinel */
    if (last_free < decommit_ptr - sizeof(hdr)) {
        obj_setlf(last_free, decommit_ptr - last_free - sizeof(hdr), 1);
    } else {
        *plast_free = 0;
        last_free = 0;
    }
    opoollimit -= actual_decommit_space;
    obj_setlf(opoollimit, 0, 0);
    FreeObjectCtr  -= actual_decommit_space;
    TotalObjectCtr -= actual_decommit_space;
    expanded -= actual_decommit_space;

    /* adjust the mark bits for the new heap size */
    shrinkMarkBits();

    if (verbosegc) {
	jio_fprintf(stderr,
		    "<GC: shrank object space by %d to %ld bytes, %ld%% free>\n",
		    actual_decommit_space,
		    TotalObjectCtr, 100 * FreeObjectCtr / TotalObjectCtr);
    }
}
#endif  /************ END CONTIGUOUS HEAPS ********************/

/*
 * Deal with an out-of-memory situation caused by an overflow within
 * the Java heap, either of the handle area or the object area.	 This
 * thing is called with the heap lock, so doesn't have to worry about
 * someone else coming along and allocating or causing another GC.
 */
static int
manageAllocFailure(size_t n, int overflow_type, int overflow_act) {
    long freeObject;
    long freeHandle;
    unsigned char *last_free = 0;
    int minFree;

    sysAssert(HEAP_LOCKED(sysThreadSelf()));

    /*CONSTCOND*/
    while (1) {
	if (verbosegc) {
	    static jlong lastGC = jlong_zero_init;
	    jlong thisTime = sysTimeMillis();
	    jio_fprintf(stderr,
      "<GC: managing allocation failure: need %d bytes, type=%d, action=%d>\n",
			n, overflow_type, overflow_act);
	    jio_fprintf(stderr, "<GC: %d milliseconds since last GC>\n",
			jlong_gtz(lastGC)
			   ? jlong_to_int(jlong_sub(thisTime,lastGC))
			   : 0);
	    lastGC = thisTime;
	}

	/*
	 * Because of fragmentation, it isn't good enough only to know 
	 * how much free space is available.  Before we go try to 
	 * allocate again, we want to know whether the last memory 
	 * management operation we did made new progress.  So each 
	 * time around the while() we reset the progress indicators.
	 */
	if (overflow_type == OVERFLOW_OBJECTS) {
	    freeObject = FreeObjectCtr;
	    freeHandle = 0;
	} else {
	    freeObject = 0;
	    freeHandle = FreeHandleCtr;
	}

	switch (overflow_act) {

	case OVERFLOW_ACT_GC:
	    /*
	     * Try normal garbage collection; otherwise try expanding.
	     */
	    last_free = gc0(n, FALSE);
	    overflow_act = OVERFLOW_ACT_EXPAND;
	    break;

	case OVERFLOW_ACT_EXPAND:
	    /*
	     * Expand the overflowing area: Successful expansion should
	     * guarantee that the allocation will succeed, and should not
	     * over(pre)allocate.
	     */
	    overflow_act = OVERFLOW_ACT_REFS;
	    if (overflow_type == OVERFLOW_HANDLES) {
	        if (expandHandleSpace(0)) {
		    /* We got *some* memory */
		    return overflow_act;
		}
	    } else if (overflow_type == OVERFLOW_OBJECTS) {
	        if (expandObjectSpace(n, last_free)) {
		    /* We got *some* memory */
		    return overflow_act;
		}
	    }
	    break;

	case OVERFLOW_ACT_REFS:
	    /*
	     * Clear all known soft references
	     */
	    if (verbosegc) {
		jio_fprintf(stderr,
			    "<GC: clearing all remaining soft refs (%d)>\n",
			    n);
	    }
	    last_free = gc0(n, TRUE);  /* Once to find them */
	    last_free = gc0(n, FALSE); /* ... and again to reclaim the space */
	    overflow_act = OVERFLOW_ACT_DESPERATE;
	    break;

	case OVERFLOW_ACT_DESPERATE:
	    if (verbosegc) {
	        jio_fprintf(stderr, "<GC: totally out of heap space>\n");
	    }
	    /* Failed to make any progress: really out of heap */
	    return OVERFLOW_ACT_FAIL;

	default:
	    return OVERFLOW_ACT_FAIL;
	}

	/*
	 * Test whether the last memory management operation made
	 * sufficient progress to try allocating again.	 Otherwise,
	 * go directly to trying the next operation.  Subtlety: because
	 * of fragmentation, apparently adequate free space is not always
	 * enough, so we also require evidence of recent progress.  Even
	 * so, there's still no guarantee allocation will go through.
	 * Note that we subtract off the current allocation amount: the
	 * preallocated memory is the amount after this allocation.
	 * 
	 * DRW: The old pre-alloc factor required too much free mem at large
	 * values. PAGE_ALIGNMENT is probably too small, but it's better 
	 * than nothing. *** REMIND: Try other policies!
	 */
#ifdef PAGED_HEAPS
	minFree = PAGE_ALIGNMENT;
#else /* not PAGED_HEAPS: */
	if (overflow_type == OVERFLOW_OBJECTS) {
	    minFree = (int)(TotalObjectCtr*minHeapFreePercent);
	} else {
	    minFree = (int)(TotalHandleCtr*minHeapFreePercent);
	}
#endif /* not PAGED_HEAPS */

	if (overflow_type == OVERFLOW_OBJECTS) {
	    if (FreeObjectCtr > freeObject &&
		FreeObjectCtr > n &&
		FreeObjectCtr-n >= minFree) {
	        return overflow_act;	/* Try to allocate again */
	    }	/* else go back and switch on new overflow_act */
	} else {
	    if (FreeHandleCtr > freeHandle &&
		FreeHandleCtr-sizeof(JHandle) >= minFree) {
	        return overflow_act;    /* Try to allocate again */
	    }	/* else go back and switch on new overflow_act */
	}
    }
    /* NOTREACHED */
}

/* Mark the handles enclosed by the object "h" refers to, and recursively
 * call markChildren on those handles.
 *
 * The "limit" is used to prevent double-marking.  Since handles are scanned 
 * in address order, we only want to recursively mark children of handles that 
 * the handle scan has passed over (ie handles < limit), because scanHandles() 
 * will get to the later handles in a while anyway.
 *
 * To reduce the number of recursive calls, we store pending handles in
 * a stack until it is full.
 */

#define REC_MARK_STK_SIZE 100

#define PUSH_REC_MARK(h) \
    if (rec_mark_top < REC_MARK_STK_SIZE) { \
	sysAssert(ValidHandle(h));          \
	rec_mark_stk[rec_mark_top++] = h; \
    } else \
        markChildren(h, limit, depth + 1, traceRefs)

#ifdef TRACEMARK
#define REC_MARK(h) \
    if (!IsMarked(h)) { \
        if (tracegc > 1) { \
	    jio_fprintf(stderr, " mo %s\n", Object2CString(h)); \
        } \
        MarkPtr(h); \
	if (h < limit) { \
	    PUSH_REC_MARK(h); \
	} \
    } else ((void) 0)
#else
#define REC_MARK(h) \
    if (!IsMarked(h)) { \
        MarkPtr(h); \
	if (h < limit) { \
	    PUSH_REC_MARK(h); \
	} \
    } else ((void) 0)
#endif

static void
markChildren(register JHandle * h, register JHandle * limit, int depth,
	     bool_t traceRefs)
{
    register ClassClass *cb;
    register JHandle *sub;
    register ClassObject *p;
    JHandle *rec_mark_stk[REC_MARK_STK_SIZE];
    int rec_mark_top = 0;
    SetLimits();
    /* Use locals so that the compiler can cache values in registers */
    ClassClass *classJavaLangClass_local = classJavaLangClass;

    if (depth > MAX_GC_DEPTH) {
        if (putDeferredHandle(h) == 0)
	    deferredHandleTableOverflow = 1;
        return;
    }

ScanNext:
    sysAssert(ValidHandle(h));
    p = unhand(h);
    sysAssert(IsMarked(h));

#ifdef TRACEGC
    if (tracegc > 1) {
	jio_fprintf(stderr, "<%X/%X>\n", h, p);
	jio_fprintf(stderr, " <  %s\n", Object2CString(h));
    }
#endif /* TRACEGC */
    switch (obj_flags(h)) {
	case T_NORMAL_OBJECT:
	    cb = obj_classblock(h);
	    REC_MARK((JHandle *)cb);
	    if (cb != classJavaLangClass_local) {
	        unsigned short *object_offsets = cbObjectOffsets(cb);

		if (object_offsets) {
		    if (!traceRefs && CCIs(cb, Reference)) {
			Classjava_lang_ref_Reference *r
			  = (Classjava_lang_ref_Reference *)p;
			bool_t skipFirst = FALSE;
			if ((r->next == NULL)
			    && ValidHandle(r->referent)
			    && (!IsMarked(r->referent))) {
#ifdef TRACEGC
			    if (tracegc > 1) {
				jio_fprintf(stderr, " ref %s",
					    Object2CString(h));
				jio_fprintf(stderr, " -> %s\n",
					    Object2CString(r->referent));
			    }
#endif /* TRACEGC */
			    r->next = discoveredRefs;
			    discoveredRefs = (Hjava_lang_ref_Reference *)h;
			    skipFirst = TRUE;
			}
			{
			    int offset;
			    p = (ClassObject *)((char *)p - 1);
			    while ((offset = *object_offsets++)) {
				if (skipFirst && (offset == 1)) {
				    sub = *(JHandle **) ((char *) p + offset);
#ifdef TRACEGC
				    if (tracegc > 1) {
					jio_fprintf(stderr, " skip %s\n",
						    Object2CString(sub));
				    }
#endif /* TRACEGC */
				    continue;
				}
				sub = *(JHandle **) ((char *) p + offset);
				if (sub) {
				    REC_MARK(sub);
				}
			    }
			}
		    } else {
			int offset;
			p = (ClassObject *)((char *)p - 1);
			while ((offset = *object_offsets++)) {
			    sub = *(JHandle **) ((char *) p + offset);
			    if (sub) {
				REC_MARK(sub);
			    }
			}
		    }
		}

	    } else {
		int i;
		struct fieldblock *fb;
	        cb = (ClassClass *)h;

		/* We must be extra careful in scanning the internals of a class
		 * structure, because this routine may be called when a class
		 * is only partially loaded (in createInternalClass).
		 */
		if (cbConstantPool(cb) &&
		    cbConstantPool(cb)[CONSTANT_POOL_TYPE_TABLE_INDEX].type) {
		    union cp_item_type *constant_pool = cbConstantPool(cb);
		    union cp_item_type *cpp = 
		        constant_pool+ CONSTANT_POOL_UNUSED_INDEX;
		    union cp_item_type *end_cpp = 
		        &constant_pool[cbConstantPoolCount(cb)];
		    unsigned char *type_tab =
		        constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
		    unsigned char *this_type = 
		        &type_tab[CONSTANT_POOL_UNUSED_INDEX];

		    for ( ; cpp < end_cpp; cpp++, this_type++) {
		        switch (*this_type) {
			case CONSTANT_String|CONSTANT_POOL_ENTRY_RESOLVED: 
			case CONSTANT_Class|CONSTANT_POOL_ENTRY_RESOLVED: {
			    void *ptr = (*cpp).p;
			    if (ValidHandle(ptr) && ptr != cb) {
#ifdef TRACEGC
			        if (tracegc > 1 && !IsMarked(ptr)) {
				    if (((JHandle *)ptr)->obj &&
					((JHandle *)ptr)->methods) {
				        jio_fprintf(stderr, " ms %s\n", 
					      Object2CString((JHandle *) ptr));
				    } else {
				        jio_fprintf(stderr, " ms %8X\n", ptr);
				    }
				}
#endif /* TRACEGC */
				REC_MARK((JHandle *)ptr);
			    }
			    break;
			}
			}
			
		    } /* loop over constant pool*/
		}

		/* Scan class definitions looking for statics */
		if (cbFields(cb) && 
		    (cbFieldsCount(cb) > 0)) { /* defensive check */
		    for (i = cbFieldsCount(cb), fb = cbFields(cb); 
			 --i >= 0; fb++) {
		        if (fieldsig(fb) &&  /* Extra defensive */
			    (fieldIsArray(fb) || fieldIsClass(fb)) &&
			    (fb->access & ACC_STATIC)) {
			    JHandle *sub = 
			        *(JHandle **)normal_static_address(fb);
			    if (ValidHandle(sub)) {
			        REC_MARK(sub);
#ifdef TRACEMARK
				if (tracegc > 1) {
				    jio_fprintf(stderr, " mc1 %s\n", Object2CString(sub));
				}
#endif /* TRACEMARK */
			    }
			}
		    }
		}

		h = (JHandle *)cbSuperclass(cb);
		if (ValidHandle(h))
		    REC_MARK(h);

		h = (JHandle *)cbClassname(cb);
		if (ValidHandle(h))
		    REC_MARK(h);

		h = (JHandle *)cbLoader(cb);
		if (ValidHandle(h))
		    REC_MARK(h);

		h = (JHandle *)cbSigners(cb);
		if (ValidHandle(h))
		    REC_MARK(h);

		h = (JHandle *)cbProtectionDomain(cb);
		if (ValidHandle(h))
		    REC_MARK(h);
	    }
	    break;

	case T_CLASS: {     /* an array of classes */
	    register long n = obj_length(h);
	    ClassClass *cb = (ClassClass *) ((ArrayOfObject *) p)->body[n];
	    HObject **body;
	    /*
	     * If the caller to allocArray has not filled in the
	     * ClassClass pointer of the array yet, disallow unloading
	     * of classes.
	     */
	    if (cb != 0) 
	        REC_MARK((JHandle *)cb);
	    body = ((ArrayOfObject *) p)->body;
	    while (--n >= 0) {
		sub = body[n];
		if (sub) {
		    REC_MARK(sub);
		}
	    }
	}
    }

    if (rec_mark_top > 0) {
        h = rec_mark_stk[--rec_mark_top];
	goto ScanNext;
    }
#ifdef TRACEGC
    if (tracegc > 1) {
	jio_fprintf(stderr, " >  %s\n", Object2CString(h));
    }
#endif /* TRACEGC */
}

/* scanThread is able to deal with ee->thread being NULL. This happens
 * when a native thread is in the middle of attaching to the VM through JNI.
 * See jni.h for details.
 */
static int 
scanThread(sys_thread_t *t, void *arg)
{
    sys_thread_t *me = (sys_thread_t *)arg;
    ExecEnv *ee = SysThread2EE(t);
    JHandle *tobj = ee->thread;
    unsigned char **ssc, **limit;
    void *base;
    SetLimits();

    if (ee->initial_stack == NULL) {
        /* EE already destroyed. */
        return SYS_OK;
    }

#ifdef DEBUG
    if (tobj && !ValidHandle(tobj)) {
	Abort();
    }
#endif /* DEBUG */
#ifdef TRACEMARK
    if (tracegc) {
	jio_fprintf(stderr, "scanThread, t:%x, arg:%x, p:%x\n", t, arg, p);
    }
    if (tracegc > 1 && p && !IsMarked(p)) {
	if (((JHandle *)p)->obj && ((JHandle *)p)->methods) {
	    jio_fprintf(stderr, " ms1 %s\n", Object2CString((JHandle *)p));
	} else {
	    jio_fprintf(stderr, " ms1 %8X\n", p);
	}
    }
#endif /* TRACEMARK */

    /* Mark thread object */
    if (tobj) {
        MarkPtr(tobj);
    }

    {
	sys_thread_t *t = EE2SysThread(ee);
	long *regs;
	int nregs;

	/* Scan the saved registers */
	regs = sysThreadRegs(t, &nregs);
	for (nregs--; nregs >= 0; nregs--)
	    if (ValidHorO(regs[nregs])) {
	        MarkPtr(regs[nregs]);
#ifdef VERBOSEGC
		if (verbosegc) {
		    if (ValidHandle(regs[nregs]))
		        reg_marked_handle_count += 1;
		    else
		        reg_marked_object_count += 1;
		}
#endif /* VERBOSEGC */
	    }

	base = ee->stack_base;
	ssc  = sysThreadStackPointer(t);
    }

    if (EE2SysThread(ee) == me) {
	ssc = (unsigned char **) &me;
    }

    if (ssc == 0 || base == 0 || (ssc == base)) {
	/*
	 * If the stack does not have a top of stack pointer or a base
	 * pointer then it hasn't run yet and we don't need to scan 
	 * its stack.  When exactly each of these data becomes available
	 * may be system-dependent, but we need both to bother scanning.
	 */
	goto ScanJavaStack;
    }

    /* Align stack top, important on Windows 95. */
    if ((long)ssc % sizeof(void *)) {
        ssc = (unsigned char **)((long)(ssc) & ~(sizeof(void *) - 1));
    }

    limit = (unsigned char **) base;

#ifdef TRACEMARK
    if (tracegc) {
	jio_fprintf(stderr, "%8X [%10X,%10X] %d\n", p, ssc, limit, limit - ssc);
    }
#endif /* TRACEMARK */
    sysAssert(ssc != limit);

    /*
     * The code that scans the C stack is assuming that the current
     * stack pointer is at a lower address than the limit of the stack.
     * Obvioulsy, this is only true for downward growing stacks.  For
     * upward growing stack, we exchange ssc and limit before we start
     * to scan the stack.
     */

#if defined(STACK_GROWS_UP)
    {
        unsigned char  **tmp;

        tmp = limit;
        limit = ssc;
        ssc = tmp;
    }
#endif /* STACK_GROWS_UP */


    while (ssc < limit) {
	register unsigned char *ptr = *ssc;
	if (ValidHorO(ptr)) {
	    /* If ptr is (apparently) an object pointer, it is pinned */
#ifdef TRACEMARK
	    if (tracegc > 1 && !IsMarked(ptr) && ValidHandle(ptr)) {
		if (((JHandle *)ptr)->obj && ((JHandle *)ptr)->methods) {
		    jio_fprintf(stderr, " ms2 %s\n",
			    Object2CString((JHandle *) ptr));
		} else {
		    jio_fprintf(stderr, " ms2 0x%8X\n", ptr);
		}
	    }
#endif /* TRACEMARK */
	    MarkPtr(ptr);
#ifdef VERBOSEGC
	    if (ValidHandle(ptr))
		stack_marked_handle_count += 1;
	    else
		stack_marked_object_count += 1;
#endif /* VERBOSEGC */
	}
	ssc++;
    }

    /*
     * Whether or not we scan the thread stack, we decide independently
     * whether to scan the Java stack.  Doing so should be more robust
     * in the face of partially-initialized or partially-zeroed threads
     * during thread creation or exit, or changes to any of that code.
     */
ScanJavaStack:
    {
	JavaFrame *frame;

	/*
	 * Because of the Invocation API, the EE may not be on the C
	 * stack anymore.
	 */
	if (ValidHandle(ee->exception.exc))
	    MarkPtr(ee->exception.exc);

	if (ValidHandle(ee->pending_async_exc))
	    MarkPtr(ee->pending_async_exc);

	if ((frame = ee->current_frame) != 0) {
	    struct methodblock *prev_current_method = 0;
	    while (frame) {
	        struct methodblock *current_method = frame->current_method;
		/*
		 * If the previous frame was a transition frame from C back 
		 * to Java (indicated by prev_current_method == NULL), then 
		 * this new frame might not have set its optop.  We must be 
		 * conservative.   Otherwise, we can use the optop value.
		 *
		 * Also permit two consecutive frames with NULL current
		 * methods, in support of JITs.  See bug 4022856.
		 */
	        stack_item *top_top_stack = 
		  (prev_current_method == 0 && current_method != NULL &&
		   ((current_method->fb.access & ACC_NATIVE) == 0))
		    ? &frame->ostack[frame->current_method->maxstack] 
		    : frame->optop;
		scanJavaFrame(frame, top_top_stack);
		frame = frame->prev;
		prev_current_method = current_method;
	    }
	} else {
#ifdef TRACEMARK 
	    if (tracegc) {
		jio_fprintf(stderr, "ee=%x\n", ee);
		jio_fprintf(stderr, "frame=%x\n", ee->current_frame);
	    }
#else
	    /*EMPTY*/
#endif /* TRACEMARK */
	}

    }

    /* ScanAllocCache */ {
        struct alloc_cache *cp = &ee->alloc_cache;
	if (cp->cache_busy) {
	    /* Mark the cache tail */
	    JHandle *cache_tail = (JHandle *)(cp->cache_tail);
	    if (ValidHandle(cache_tail))
	        MarkPtr(cache_tail);
	    /* Local handles are untouched because their obj field
	       is NULL.  Blocks, if loaded, will be pinned from
	       the cacheAlloc frame. */
	} else {
	    JHandle *hp1, *hp2;
	    /* Release cache tail */
#ifdef	HEAPMETERS
	    if (cp->cache_tail) cacheFlushes++;
#endif
	    cp->cache_tail = NULL;
	    cp->cache_size = 0;
	    /* Push handles onto global free list */
	    hp1 = (JHandle *)(cp->cache_handles);
	    while (hp1) {
#ifdef PAGED_HEAPS
		ChunkBlkP MOH_chunk = ChunkBase(hp1);
#endif /* PAGED_HEAPS */
	        sysAssert(hp1->obj == NULL);
		hp2 = (JHandle *)hp1->methods;
		PUSH_FREE_HANDLE(hp1);
		FreeHandleCtr += sizeof(JHandle);
		hp1 = hp2;
	    }
	    cp->cache_handles = NULL;
	}
    }

    return SYS_OK;
}

static void 
scanThreads(void)
{
    sys_thread_t *self = sysThreadSelf();
    sysThreadEnumerateOver(scanThread, (void *)self);
    /* REMIND - should do something with the return value. */
}

static void scanJavaFrame(JavaFrame *frame, stack_item *top_top_stack)
{
    stack_item *ssc, *limit;
    JHandle *ptr;
    JavaStack *javastack;
    struct methodblock *mb = frame->current_method;
    SetLimits();

#ifdef TRACEMARK
    if (tracegc) {
        jio_fprintf(stderr, " scanning frame %8X\n", frame);
    }
#endif /* TRACEMARK */

    if (mb) {
        ClassClass *mbClass = fieldclass(&mb->fb);
	MarkPtr(cbHandle(mbClass));
    }

    limit = top_top_stack;
    javastack = frame->javastack;

    /* Scan the operand stack. */
    /*CONSTCOND*/
    while (1) {
        int is_first_chunk = IN_JAVASTACK((stack_item *)frame, javastack);
        for (ssc = is_first_chunk ? frame->ostack : javastack->data;
	     ssc < limit; ssc++) {
            ptr = ssc->h;
	    if (ValidHandle(ptr)) {	/* Never an object pointer */
#ifdef TRACEMARK
	        if (tracegc > 1 && !IsMarked(ptr)) {
		    jio_fprintf(stderr, " ms3 %8X\n", ptr);
		}
#endif /* TRACEMARK */
		MarkPtr(ptr);
	    }
	}
	if (is_first_chunk)
	    break;
	javastack = javastack->prev;
	limit = javastack->end_data;
    }

    /* Nothing more to do for pseudo and JIT frames. */
    if (mb == 0 || IS_JIT_FRAME(frame))
        return;

    if (mb->fb.access & ACC_NATIVE) {
        /* For native frames, we scan the arguments stored at the top 
	   of the previous frame. */
        JavaFrame *prev_frame = frame->prev;
	if (prev_frame == 0)
	    return;
	ssc = prev_frame->optop;
	limit = ssc + mb->args_size;
    } else {
        /* Scan local variables in Java frame */
        ssc = frame->vars;
	if (ssc == 0)
	    return;
	limit = (stack_item *)frame;
    }

    for (; ssc < limit; ssc++) {
        ptr = ssc->h;
	if (ValidHandle(ptr)) {	 /* Never an object pointer */
#ifdef TRACEMARK
	    if (tracegc > 1 && !IsMarked(ptr)) {
	        jio_fprintf(stderr, " ms4 %8X\n", ptr);
	    }
#endif /* TRACEMARK */
	    MarkPtr(ptr);
	}
    }
}

/* Mark the global references created by the Native Method interface */
static void scanJNIGlobalRefs(void)
{
    scanJavaFrame(globalRefFrame, globalRefFrame->optop);
}

/*
 * Scan the heap and chase pointers.  Scanning from pointers into the
 * middle of objects is normally disabled, apparently because it was
 * slow.  This should be revisited, especially if it reduces pinning.
 */
static void 
scanHeap(void)
{
#ifdef CHECK_INTERIOR_POINTERS
    SetLimits();
    unsigned char *o, *oo, *next;
    int i, size;

#ifdef TRACEGC
    if (tracegc) {
        jio_fprintf(stderr, "\nRecursive mark scan:\n");
    }
#endif /* TRACEGC */

    /*
     * Search the entire memory pool looking for references into the
     * the middle of objects.  If any are found, mark the object they
     * are part of.  Pointers into the middle of objects can happen
     * frequently, especially in the case of arrays.
     *
     * This piece of code is tricky because it needs to be fast.
     * It is basically walking through most of the mark bits looking
     * for set bits.  It does it in 32-bit increments for speed.
     */
    o = oo = opmin;
    next = o + obj_len(o);
    size = ((int) (opoollimit - opmin) / (OBJECTGRAIN * BITSPERMARK) + 1);
    for (i=0; i < size; i++) {
	sysAssert(((unsigned int)(oo - opmin) >> 7) == i);
	if (markbits[i]) {
	    unsigned int bits = markbits[i];
	    unsigned int shifts = OBJECTGRAIN * BITSPERMARK;
	    while (bits && oo < opmax) {
	        if (bits & 1) {
		    /* Mark the object, not the header */
		    MarkPtr((o + sizeof(hdr)));
		}
		bits >>= 1;
		shifts -= OBJECTGRAIN;
		oo += OBJECTGRAIN;
		if (oo == next) {
		  o = next;
		  next = o + obj_len(o);
		}
	    }
	    oo += shifts;
	} else {
	    oo += OBJECTGRAIN * BITSPERMARK;
	}
	
	if (oo >= opmax) {
	    break;
	}
	
	while (oo >= next) {
	  o = next;
	  next = o + obj_len(o);
	}
    }
#endif /* CHECK_INTERIOR_POINTERS */
}

static void
markChildrenFromTop(JHandle *h, JHandle *limit, bool_t traceRefs)
{
    do {
        markChildren(h, limit, 0, traceRefs);
    } while ((h = getDeferredHandle()));
}

/*
 * Walk the entire handle pool, and do a recursive search for
 * references.
 */
static void 
scanHandles(void)
{
    SetLimits();
    JHandle *hp;


    deferredHandleTableOverflow = 1;
    
    while (deferredHandleTableOverflow) {
        deferredHandleTableOverflow = 0;
        MAP_OVER_HANDLES_FROM_START(hp)
	    if (hp->obj) {
	        int flag;
		sysAssert(ValidObject(hp->obj));
	        if (IsMarked(hp->obj)) {
#ifdef PAGED_HEAPS
		    ChunkBlkP ObjChunk = ChunkBase(hp->obj);
		    /* note that this chunk is pinned */
		    ObjChunk->chunkFlags |= CHUNK_PINNED;
#endif /* PAGED_HEAPS */
		    MarkPtr((JHandle *) hp);
		}
		flag = obj_flags(hp);
		/* only objects and object arrays need to be recursively
		   marked. */
		if (flag == T_NORMAL_OBJECT || flag == T_CLASS)
		    if (IsMarked(hp)) {
		        markChildrenFromTop(hp, hp + 1, FALSE);
		    }
	    }
	END_MAP_OVER_HANDLES_FROM_START;
    }
}

/* Mark all classes that shouldn't be unloaded. */
static void
markStickyClasses(void)
{
    ClassClass **pcb = binclasses;
    int i;
    SetLimits();

    for (i = nbinclasses; --i >= 0; pcb++) {
	ClassClass *cb = *pcb;
	if (!classgc || (cbFlags(cb) & CCF_IsSticky))
	    MarkPtr(cbHandle(cb));
    }

    /* Mark handles to primitive classes */
    markPrimitiveClasses();
}

/*
 * Mark handles to the 9 primitive classes.  Currently, these are not
 * in binclasses, and they have no internal object pointers other than
 * cbClassname (which may be NULL).
 */
static void
markPrimitiveClasses(void)
{
    static ClassClass **primitive_classes[] = {
	&class_void, &class_boolean, &class_byte, &class_char, &class_short,
	&class_int, &class_long, &class_float, &class_double, NULL
    };
    ClassClass ***cbpp = primitive_classes;
    SetLimits();
    while (*cbpp) {
	ClassClass *cb = **cbpp;
	if (ValidHandle(cb)) {
	    MarkPtr(cb);
	    sysAssert(ValidObject(unhand(cb)));
	    if (ValidHandle(cbClassname(cb)))
		MarkPtr(cbClassname(cb));
	}
	cbpp++;
    }
}

static HObject **
getStaticField(ClassClass *cb, const char *name) /* ## Check signatures! */
{
    int i;
    struct fieldblock *fb;

    if (cbFields(cb) && (cbFieldsCount(cb) > 0)) {
	for (i = cbFieldsCount(cb), fb = cbFields(cb); --i >= 0; fb++) {
	    if (fieldsig(fb)
		&& (fieldIsArray(fb)
		    || fieldIsClass(fb)
		    || (fieldsig(fb)[0] == SIGNATURE_INT)
		    || (fieldsig(fb)[0] == SIGNATURE_LONG))
		&& (fb->access & ACC_STATIC)
		&& (fb->name[0] == name[0])
		&& (!strcmp(fb->name, name))) {
		if (fieldsig(fb)[0] == SIGNATURE_LONG) {
		    return (HObject **)twoword_static_address(fb);
		} else {
		    return (HObject **)normal_static_address(fb);
		}
	    }
	}
    }

    panic("GC: getStaticField failed: %s %s", cbName(cb), name);
    return 0;			/* Keep lint happy */
}

static ClassClass *
getStickySystemClass(ExecEnv *ee, char *name)
{
    ClassClass *c = FindStickySystemClass(ee, name, TRUE);
    if (!c) {
	panic("GC: getStickySystemClass failed: %s\n\tCLASSPATH may be incorrect", name);
    }
    return c;
}

void
InitializeRefs(void)
{
    if (!classJavaLangRefReference) {
	ExecEnv *ee = EE();
	classJavaLangRefReference
	  = getStickySystemClass(ee, "java/lang/ref/Reference");
	classJavaLangRefSoftReference
	  = getStickySystemClass(ee, "java/lang/ref/SoftReference");
	classJavaLangRefWeakReference
	  = getStickySystemClass(ee, "java/lang/ref/WeakReference");
	classJavaLangRefFinalReference
	  = getStickySystemClass(ee, "java/lang/ref/FinalReference");
	classJavaLangRefFinalizer
	    = getStickySystemClass(ee, "java/lang/ref/Finalizer");
	classJavaLangRefPhantomReference
	  = getStickySystemClass(ee, "java/lang/ref/PhantomReference");
	refLock = *getStaticField(classJavaLangRefReference, "lock");
	pendingRefs
	    = ((Hjava_lang_ref_Reference **)
	       getStaticField(classJavaLangRefReference, "pending"));
	softRefClock = (int64_t *)getStaticField(classJavaLangRefSoftReference,
						 "clock");

	/* Allocate a tiny dummy object for use as an end-of-list marker */
	refEnd = ((Hjava_lang_ref_Reference *)
		  realObjAlloc(ee, ALLOC_CACHE_BLOCK_TAG,
			       ROUND_SIZE(0)));
	if (!refEnd) {
	    panic("GC: Failed to allocate refEnd object");
	}
	discoveredRefs = refEnd;
    }
}

/* The clearing of soft references is governed by the heap expansion/shrinkage
   rate and by the time since each reference was created or last dereferenced.

   Time is measured in GC invocations.  The java.lang.ref.SoftReference class
   contains a private static long "clock" field that we increment each time
   processRefs is invoked.  When a soft ref is created or dereferenced, its
   internal timestamp is updated with the current value of the clock.

   The expansion/shrinkage rate is taken into account by using it to compute an
   age threshold before processing soft references.  Soft refs younger than
   this threshold are left alone, while older soft refs are examined and, if
   their referents are unmarked, cleared.

   The current heuristic for computing the threshold is to count how many of
   the previous EXPANSION_HIST_SIZE invocations of the GC expanded the heap,
   subtract from this the number of inocations that shrank the heap, and then
   divide this count by the number of samples.  If the resulting ratio is not
   positive, then the constant SOFT_MAX_AGE is used.  If positive, the ratio is
   inverted and used to select an age threshold between SOFT_MIN_EXPANDING_AGE
   and SOFT_MAX_EXPANDING_AGE.  This heuristic causes soft refs to be cleared
   more and more aggressively as the expansion rate increases but very slowly
   when the heap is in a steady state or is shrinking.
 */

static int
computeSoftRefThreshold(void)
{
    int i, t;
    long d;

    for (i = 0, d = 0; i < EXPANSION_HIST_SIZE; i++) {
	int h = expansionHist[i];
	d += (h > 0) ? +1 : ((h < 0) ? -1 : 0);
    }
    if (d <= 0) {
	t = SOFT_MAX_AGE;
    } else {
	float ratio;
	if (d > EXPANSION_HIST_SIZE) d = EXPANSION_HIST_SIZE;
	ratio = (float)d / (float)EXPANSION_HIST_SIZE;
	t = (int)((1.0 - ratio) *
		  (SOFT_MAX_EXPANDING_AGE - SOFT_MIN_EXPANDING_AGE)
		  + SOFT_MIN_EXPANDING_AGE);
    }
#ifdef VERBOSEGC
    softRefThreshold = t;
#endif
    return t;
}

static bool_t
preProcessSoftRef(Classjava_lang_ref_SoftReference *r,
		  int64_t softClock, int ageThreshold)
{
    int64_t t = r->timestamp;
    int a = ll2int(ll_sub(softClock,t));
    SetLimits();

    if (a < ageThreshold) {
	/* Don't clear this soft ref */
	HObject *o = r->referent;
	MarkPtr(o);
#ifdef PAGED_HEAPS
	markChildrenFromTop(o, (JHandle *) mem_top, TRUE);
#else
	markChildrenFromTop(o, (JHandle *) hpoollimit, TRUE);
#endif /* PAGED_HEAPS */
	return TRUE;
    } else {
	/* Clear this soft ref if its referent isn't otherwise marked */
	return FALSE;
    }
}

static void
processRefList (Hjava_lang_ref_Reference *first, bool_t clearReferent, int* rc)
{
    Hjava_lang_ref_Reference *h, *next, *keep = refEnd;
    Classjava_lang_ref_Reference *r;
    HObject *o;
#ifdef VERBOSEGC
    int nEnqueued = 0;
#endif
    SetLimits();

    for (h = first; h != refEnd; h = next) {
	r = (Classjava_lang_ref_Reference *)unhand(h);
	next = r->next;
	o = r->referent;

	/* Drop this Reference if its referent is marked */
	if (IsMarked(o)) {
#ifdef TRACEGC
	    if (tracegc) {
		jio_fprintf(stderr, "processRefList: Dropping %s\n",
			    Object2CString((HObject *)h));
	    }
#endif /* TRACEGC */
	    r->next = 0;
	    continue;
	}

	r->next = keep;
	keep = h;
    }

    for (h = keep; h != refEnd; h = next) {
	r = (Classjava_lang_ref_Reference *)unhand(h);
	next = r->next;

	if (clearReferent) {
	    r->referent = 0;
	} else {
	    o = r->referent;
	    MarkPtr(o);
#ifdef PAGED_HEAPS
	    markChildrenFromTop(o, (JHandle *) mem_top, TRUE);
#else
	    markChildrenFromTop(o, (JHandle *) hpoollimit, TRUE);
#endif /* PAGED_HEAPS */
	}

#ifdef TRACEGC
	if (tracegc) {
	    jio_fprintf(stderr, "processRefList: Enqueuing %s\n",
			Object2CString((HObject *)h));
	}
#endif /* TRACEGC */

	r->next = *pendingRefs;
	if (r->next == 0) r->next = h;
	*pendingRefs = h;
#ifdef VERBOSEGC
	nEnqueued++;
#endif
    }

#ifdef VERBOSEGC
    *rc = nEnqueued;
#endif
}

static void
processRefs(bool_t clearAllSoftRefs)
{
    Hjava_lang_ref_Reference
	*dSoft = refEnd,
        *dWeak = refEnd,
	*dFinal = refEnd,
	*dPhantom = refEnd;
    Hjava_lang_ref_Reference *h, *next;
    Hjava_lang_ref_Reference *oldPendingRefs = *pendingRefs;
    int st = computeSoftRefThreshold();
    int64_t sc = *softRefClock;
    *softRefClock = ll_add(sc,ll_one_const);

    /* Classify discovered Reference instances */
    for (h = discoveredRefs; h != refEnd; h = next) {
	ClassClass *cb = obj_classblock(h);
	Classjava_lang_ref_Reference *r = (Classjava_lang_ref_Reference *)unhand(h);
	next = r->next;
	if (is_subclass_of(cb, classJavaLangRefSoftReference, 0)) {
	    if (!clearAllSoftRefs
		&& preProcessSoftRef((Classjava_lang_ref_SoftReference *)r,
				     sc, st)) {
		r->next = 0;
	    } else {
		r->next = dSoft;
		dSoft = h;
	    }
	} else if (is_subclass_of(cb, classJavaLangRefWeakReference, 0)) {
	    r->next = dWeak;
	    dWeak = h;
	} else if (is_subclass_of(cb, classJavaLangRefFinalReference, 0)) {
	    r->next = dFinal;
	    dFinal = h;
	} else if (is_subclass_of(cb, classJavaLangRefPhantomReference, 0)) {
	    r->next = dPhantom;
	    dPhantom = h;
	} else {
	    panic("Reference of unknown type: %s", cbName(cb));
	}
    }
    discoveredRefs = refEnd;

    {
#ifdef VERBOSEGC
	int *rc = refEnqCounts;
#else
	int *rc = 0;
#endif
	processRefList(dSoft, TRUE, rc + 0);
	processRefList(dWeak, TRUE, rc + 1);
	processRefList(dFinal, FALSE, rc + 2);
	processRefList(dPhantom, FALSE, rc + 3);
    }

    /* In order to avoid potential deadlocks, delay notification until
       we're multi-threaded again */
    notifyRefLock = (*pendingRefs != oldPendingRefs);
}

static int freeHandle(JHandle *hp)
{
	    int freed;
#ifdef PAGED_HEAPS
	    ChunkBlkP MOH_chunk = ChunkBase(hp);
#endif /* PAGED_HEAPS */
#ifdef TRACEFREE
	    if (tracegc) {
		jio_fprintf(stderr, " fr %s (%8X/%8X)\n",
			Object2CString(hp), hp, hp->obj);
	    }
#endif /* TRACEFREE */
	    freed = obj_len((char *) hp->obj - sizeof(hdr));
#ifdef DEBUG
	    memset((char *) hp->obj, 0x55, freed - sizeof(hdr));
#endif /* DEBUG */
	    FreeObjectCtr += freed;
	    sysAssert(FreeObjectCtr <= TotalObjectCtr);
	    
#ifdef PAGED_HEAPS
	    {
		ChunkBlkP objP = ChunkBase(hp->obj);
#ifdef LDEBUG
		checkChunk(objP);
#endif /* LDEBUG */
		objP->freeCnt += freed;
	    }
#endif /* PAGED_HEAPS */

	    nfreed++;
	    bytesfreed += freed;

#ifdef DEBUG
	    /* When an object is free and pinned, it means that a JNI
	     * native method has forgot to call ReleaseStringChars,
	     * Release<...>ArrayElements or ReleasePrimitiveArrayCritical
	     * to release an unused array.
	     *
	     * We print a message and abort the VM. 
	     */
	    if (obj_pinned((char *) hp->obj - sizeof(hdr))) {
	        char *fname;
	        if (obj_flags(hp) == T_NORMAL_OBJECT)
		    fname = "ReleaseStringChars";
		else
		    fname = "Release<PrimitiveType>ArrayElements or "
		      "ReleasePrimitiveArrayCritical";
	        panic("A JNI native method forgot to call %s\n", fname);
	    }
#endif

#ifdef HPROF 
	    /* we should do the jvmpi notification before freeing h->obj 
	     * as we pass the actual obj pointer to the profiling agent. */
	    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_OBJ_FREE)) {
	        jvmpi_free_object(hp);
	    }
#endif /* HPROF */
	  
	    obj_setfree((char *) hp->obj - sizeof(hdr));
	    hp->obj = 0;

	    PUSH_FREE_HANDLE(hp);

#ifdef PAGED_HEAPS
	    MOH_chunk->freeCnt += sizeof(JHandle);
#endif /* PAGED_HEAPS */

	    FreeHandleCtr += sizeof(JHandle);
	    sysAssert(FreeHandleCtr <= TotalHandleCtr);
	    return freed;
}

/*
 * Scan for remaining unmarked objects and free them.
 */ 
static int
freeSweep(unsigned int free_space_goal)
{
    SetLimits();
    JHandle *hp;
    int do_compact_heap = 1;

#ifdef TRACEGC
    if (tracegc) {
	jio_fprintf(stderr, "freeSweep, free_space_goal:0x%x\n",
		    free_space_goal);
    }
#endif

    MAP_OVER_HANDLES_FROM_START(hp)
	sysAssert(hp->obj == NULL || ValidObject(hp->obj));
        if (hp->obj) {
#ifdef FAST_MONITOR
            if (is_monitor(key2mid(hp)) && !IsMarked(hp->obj)) {
                monitorCacheGC((uintptr_t) hp);
            }            
#endif /* FAST_MONITOR */
            if (!IsMarked(hp)) {
	        if (freeHandle(hp) >= free_space_goal) {
		    do_compact_heap = 0;
	        }
            }
	}
    END_MAP_OVER_HANDLES_FROM_START;

/* Reset allocation hands, so that freed memory gets used */	
#ifdef PAGED_HEAPS
    {
	ChunkBlkP chunk = firstObjBlkP;
	do {
	    chunk->freePtr = chunk->startPtr; /* reset object allocation ptr */
	    chunk = chunk->nextPtr;	      /* walk thru object chunks */
	} while (chunk != firstObjBlkP);
    }
    
    /* Prefer low memory over high memory. */
    curHanBlkP = firstHanBlkP;
    curObjBlkP = firstObjBlkP;
#else /* not PAGED_HEAPS: */
    opoolhand = opmin;
#endif /* PAGED_HEAPS */

    return do_compact_heap;
}

static void
unloadUnusedClasses(void)
{
    ClassClass **pcb;
    int i;
    SetLimits();

    pcb = binclasses + nbinclasses - 1;
    for (i = nbinclasses; --i >= 0; pcb--) {
	ClassClass *cb = *pcb;
	if (IsMarked(cbHandle(cb)))
	    continue;
	/* Mark the class block so that it won't be freed by freeSweep. */
	MarkPtr(cbHandle(cb));
	/* 
	 * Remove the class and add it to list of classes to be freed. 
	 * It is illegal to call free() during single-threading mode,
	 * so classes are freed immediately after resuming multi-
	 * threading mode.
	 */
	cbFinalizer(cb) = (struct methodblock *) classFreeList;
	classFreeList = cb;

	binclasses[i] = binclasses[--nbinclasses];
    }
}

static void
freeClasses(void)
{
    int classesFreed = 0;

    sysAssert(HEAP_LOCKED(sysThreadSelf()));
#ifdef HPROF
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_CLASS_UNLOAD)) {
        ClassClass *cb = classFreeList;
	while (cb) {
	    jvmpi_free_class(cb);
	    cb = (ClassClass *)cbFinalizer(cb);
	}
    }
#endif /* HPROF */

#ifdef BREAKPTS
    if (debugging) {
        ClassClass *cb = classFreeList;
        while (cb) {
            notify_debugger_of_class_unload(EE(), (JHandle *)cb);
            cb = (ClassClass *)cbFinalizer(cb);
        }
    }
#endif
    while (classFreeList != 0) {
	ClassClass *cb = classFreeList;
	classFreeList = (ClassClass *) cbFinalizer(cb);
	if (verbosegc) {
	    char cbuf[MSG_BUF_LEN];
	    jio_fprintf(stderr, "<GC: freeing class %s>\n", 
            	    classname2string(cbName(cb), cbuf, sizeof(cbuf)));
	}
	FreeClass(cb);
	freeHandle((JHandle *)cbHandle(cb));
	classesFreed++;
    }

    if (verbosegc && classesFreed > 0) {
	jio_fprintf(stderr, "<GC: unloaded and freed %d class%s>\n",
		classesFreed, classesFreed == 1 ? "" : "es");
    }
}

/*
 * COMPACT THE HEAP:
 * Concept: Assuming optimal packing, there are only two places that
 * holes, or bubbles in the heap can occur.  These are immediately before
 * each pinned range of memory, and at the tail of the heap (the last
 * range is always free).
 *
 * The goal of these schemes is to leave as many large free blocks available
 * as is possible.
 */

/*
 * This is the reverse of the above swap.  It is called once
 * on each allocated block and fixes the handle/pointer swap.
 */
#ifndef FAST_MONITOR
#define REVERSE_SWAP(p) { \
    JHandle *hp; \
    sysAssert(obj_swapped(p)); \
    obj_clearswapped(p); \
    hp = *(JHandle **) (p + sizeof(hdr)); \
    *(OBJECT *) (p + sizeof(hdr)) = (OBJECT) hp->obj; \
    hp->obj = (OBJECT *) (p + sizeof(hdr)); \
}
#else /* FAST_MONITOR */
#define REVERSE_SWAP(p) { \
    JHandle *hp; \
    sysAssert(obj_swapped(p)); \
    obj_clearswapped(p); \
    hp = *(JHandle **) (p + sizeof(ohdr)); \
    *(OBJECT *) (p + sizeof(ohdr)) = (OBJECT) hp->obj; \
    hp->obj = (OBJECT *) (p + sizeof(hdr)); \
}
#endif

#ifdef PAGED_HEAPS /************ PAGED HEAPS: ********************/

/* Picks a chunk to free:
 * To do this right, we'd like to know how the OS allocates chunks for us,
 * but I haven't noticed any real order.  For now, only try to free the last
 * unpinned chunk that is small enough to make things worthwhile. Note that we
 * need to have already processed the free lists of the possible destiniation
 * chunks.  If we chose to evacuate the first chunk, we wouldn't have any free
 * lists to copy things to.  Should change compact heap to process all
 * destination chunks before the evacuation chunks.
 */
static ChunkBlkP 
evacuationChunk()
{
    ChunkBlkP chunk = firstObjBlkP;
    ChunkBlkP highestChunk = NULL;	    /* highest unpinned chunk */
    
    /*
     * Only try to evacuate if there might be enough room in the other chunks,
     * "*2" to make sure it we won't immediatly ask for the chunk back again.
     */
    do {
	if (((chunk->chunkFlags & CHUNK_PINNED) == 0) &&
	    (FreeObjectCtr > (chunk->allocSize * 2)))
	    highestChunk = chunk;
	chunk = chunk->nextPtr;
    } while (chunk != firstObjBlkP);
    
    return highestChunk;
}

/*
 * This algorithm walks a chunk from beginning to end, performing
 * several operations:
 *  0. Pick a destination chunk for the chunk being compacted.	Usually we
 *     compact into the same chunk, but we can choose some chunks as candidates
 *     for evacuation if there are no pinned objects in the chunk.  Currently
 *     evacuates the last unpinned chunk, and only when async_call is true.
 *  1. Coalesce adjacent free blocks.
 *  2. Record free blocks in linked lists based off of object chunk's "freePtr".
 *     This list may be handy globally, but currently only created during
 *     compactHeap(). The list is sorted in address order. Be careful not to 
 *     swap free objects, because swapping and free links use the same word 
 *     in an object! Also note that the link is stored where we used to store
 *     a debugging check word (0x55555555), so the check word is now stored
 *     in the next word if the object is big enough.
 *  3. Copy unpinned, allocated blocks down into the lowest free
 *     block available.
 *  4. Free object chunk if no objects remain.
 *
 */

#define free_list_next(p)   \
        (*((unsigned char **)((unsigned char *)(p) + sizeof(hdr))))

/* Compact all data in this chunk.  If we want to evacuate a chunk, pass in
 *  a non-NULL value for destChunk.
 */
static int 
compactChunk(ChunkBlkP chunk, ChunkBlkP destChunk)
{
    unsigned char *p, *next, *prev;
    unsigned long plen;
    int blocks_moved = 0;
    /* use last_free as a pointer to the tail of the free list */
    unsigned char *last_free = 0;
    /* flags when we are evacuating to another chunk. */
    ChunkBlkP evacChunk = NULL;

#define RemoveLink(linkp, prevp, nextp) {	\
    if (prevp != NULL) {			\
	free_list_next(prevp) = (nextp);	\
    } else {					\
	destChunk->freePtr = (nextp);		\
    }						\
    if (last_free == (linkp)) {			\
	last_free = (nextp) ? (nextp) : (prevp);\
    }						\
}

#ifdef LDEBUG
    checkChunk(chunk);
    if (destChunk) {
	checkChunk(destChunk);
    }
#endif

    /* note that p points to headers, not aligned objects */
    p = (unsigned char *) chunk->startPtr;
    /* from now the freePtr field is a linked free list, so initialize list: */
    chunk->freePtr = NULL;
    prev = 0;
    if (destChunk) {
	evacChunk = chunk;
    } else {
	destChunk = chunk;		    /* copy into itself */
    }

#ifdef TRACEGC
    if (tracegc && evacChunk) {
	jio_fprintf(stderr, "Going to evacuate chunk 0x%lx to chunk 0x%lx\n",
		evacChunk, destChunk);
    }
#endif /* TRACEGC */

    /*
     *	- look at every object in this chunk
     */
    while (p < (unsigned char *) chunk->endPtr) {
	plen = obj_len(p);	  
	next = p + plen;

#ifdef LDEBUG
	sysAssert(ValidObject(p + sizeof(hdr)));
	sysAssert(ValidObject(next + sizeof(hdr)));
	sysAssert(plen > 0);
#endif

	/* There are three cases here (in order):
	 * p is free, p is used and not pinned, and p is used and pinned.
	 */
	if (obj_free(p)) {
	    int old_len = plen;
	    /* 1. Coalesce adjacent free blocks */
		
	    while (next && obj_free(next)) {
#ifdef LDEBUG
		sysAssert(obj_len(next) > 0);
#endif

		plen += obj_len(next);
		next += obj_len(next);
		blocks_moved++;
	    }
	    obj_setlf(p, plen, 1);
	    FILL_CHECK_WORD(p);

#ifdef TRACECOMPACT
	    if (tracegc) {
		if (plen > old_len) {
		    jio_fprintf(stderr,
		       "coaleced 0x%lx bytes into 0x%lx free bytes at 0x%lx\n", 
			    (plen - old_len), old_len, p);
		} else {
		    jio_fprintf(stderr, "found	0x%lx free bytes at 0x%lx\n", 
			    plen, p);
		}
	    }
#endif /* TRACECOMPACT */

	    /* 2. Record p in the free list in address order.  Since we are 
	     * walking the heap in address order, simply place p at the end 
	     * of the list.
	     */
	    if (last_free == NULL) {
		chunk->freePtr = p;
	    } else {
		free_list_next(last_free) = p;
	    }
	    free_list_next(p) = NULL;
	    last_free = p;
	    prev = p;
	} else if (!IsMarked(p + sizeof(hdr)) && !obj_pinned(p)) {
	    unsigned char *freep;		/* loop variable */
	    unsigned char *foundp = NULL;
	    unsigned char *free_before_foundp;	/* foundp's previous pointer */
    
	    /* 3. Copy p down into the lowest available free block */

	    /*
	     * Find lowest block that is big enough to hold the block.
	     * Free list is in address order, so just return the first one.
	     * Ignore 'prev' as we will handle it with a simpler case.
	     * If we are evacuationg this chunk, first search all other chunks
	     * for free space.	Failing that, compact into itself.
	     */
	    if (evacChunk) { /* first look other chunks for free space */
		ChunkBlkP someDestChunk = destChunk;
		do {
		    free_before_foundp = NULL;
		    if ((someDestChunk != chunk) &&
			(someDestChunk->freeCnt >= plen))
			for (freep = someDestChunk->freePtr; 
			     freep != NULL;
			     freep = free_list_next(freep)) {
			    if (obj_len(freep) >= plen) {
			    	/* test was GT, but that didn't seem right */
				foundp = freep;
				break;
			    }
			    free_before_foundp = freep;
			}
		    if (!foundp) {
			someDestChunk = someDestChunk->nextPtr;
#ifdef TRACEGC
			if (tracegc) {
			    jio_fprintf(stderr, 
			    "Try to evacuate rest of chunk 0x%lx to 0x%lx\n",
				    chunk, someDestChunk);
			}
#endif /* TRACEGC */
		    }
		} while (!foundp && (someDestChunk != destChunk) 
		         && (someDestChunk < chunk));
		
		if (foundp) {
		    destChunk = someDestChunk;
#ifdef LDEBUG
		    checkChunk(destChunk);
		    /* only chunks lower in memory have their free lists set. */
		    sysAssert(someDestChunk < chunk);	
		    sysAssert(ValidObject(foundp + sizeof(hdr)));
#endif
		} else {
    		    /* once we fail to evacaute a block, give up. */
		    evacChunk = NULL;	
		    destChunk = chunk;
		}
	    }
	    
	    if (!foundp) {
		free_before_foundp = NULL;
		for (freep = chunk->freePtr; 
		     freep != NULL; 
		     freep = free_list_next(freep)) {
		    if (obj_len(freep) >= plen) {
		        /* test was GT, but that didn't seem right */
			foundp = freep;
			break;
		    }
		    free_before_foundp = freep;
		}
	    }
	
	    /* did we find a block? */
	    if (foundp && foundp != prev) {
		unsigned char *oldp = p;
		/* remember this stuff because memmove will bash it: */
		unsigned long lowest_len = obj_len(foundp);
		unsigned char *next_freep = free_list_next(foundp);
		/*
		 * If a slot is found, move the block, and mark the old
		 * slot as free.  If the free space was bigger than we needed,
		 * break off a new chunk.
		 */
#ifdef TRACECOMPACT
		if (tracegc) {
		    jio_fprintf(stderr, "move1 0x%lx bytes from 0x%lx to 0x%lx\n",
		            plen, p, foundp);
		}
#endif /* TRACECOMPACT */
		sysAssert(obj_free(foundp));
		memmove(foundp, p, plen);
		blocks_moved++;
		obj_setfree(p);
		sysAssert(lowest_len >= plen);
		FILL_CHECK_WORD(p);
		p = foundp;

		/* it was bigger than we needed, break off a piece. */
		if (lowest_len > plen) {
		    unsigned char *fp = foundp;
		    fp += plen;
		    obj_setlf(fp, (lowest_len - plen), 1);
                    if (!IS_ALIGNED((lowest_len - plen),OBJECTGRAIN)) {
		        /* size still aligned? */
			sysAssert(IS_ALIGNED((lowest_len - plen),OBJECTGRAIN)); 
		    }
		    sysAssert(ValidObject(fp + sizeof(hdr)));
		    FILL_CHECK_WORD(fp);
		    /* link fp to the next free block, and replace foundp 
		     * in the free list with fp */
		    free_list_next(fp) = next_freep;
		    RemoveLink(foundp, free_before_foundp, fp);
		} else {
		    /* remove foundp from the free list */
		    RemoveLink(foundp, free_before_foundp, next_freep);
		}

		if (evacChunk) {
		    destChunk->freeCnt -= plen;
		    evacChunk->freeCnt += plen;
		}
		
		if (prev && obj_free(prev)) {
		    /* back up by two, to coalece adjacent free blocks */
		    next = prev; 
		    prev = 0;
		} else {
		    /* look at this block again as a free block */
		    next = oldp;  
		}
	    } else {
		/*
		 * no free slot found or it is the previous slot.  See if prev
		 * is free, and if so, just slide this range back.  Note that 
		 * the previous slot does not need to bigger than current 
		 * object, because we are just sliding.
		 */
		if (prev && obj_free(prev)) {
		    unsigned char *next_freep = free_list_next(prev);
#ifdef TRACECOMPACT
		    if (tracegc) {
			jio_fprintf(stderr, "move2 0x%lx bytes from 0x%lx to 0x%lx\n",
			        plen, p, prev);
		    }
#endif /* TRACECOMPACT */
		    memmove(prev, p, plen);
		    blocks_moved++;
		    next = prev + plen;
		    obj_setlf(next, (p - prev), 1);
		    sysAssert(ValidObject(prev + sizeof(hdr)));
		    sysAssert(next_freep == NULL);
		    sysAssert(last_free == prev);
		    FILL_CHECK_WORD(next);
		    p = prev;
		    /* remove prev from the free list.	prev must be the last 
		     * thing on the list */
		    if (foundp != prev) {
		    	/* we need to find the free block before prev */
			free_before_foundp = NULL;
			for (freep = chunk->freePtr; 
			     freep != prev; 
			     freep = free_list_next(freep)) {
			    free_before_foundp = freep;
			}
		    }
		    if (free_before_foundp) {
			free_list_next(free_before_foundp) = NULL;
		    } else {
			chunk->freePtr = next_freep; 
		    }
		    last_free = free_before_foundp; /* DRW double check this! */
		} else {
		    prev = p;
		}
	    }
	} else {
	    prev = p;
	}

	/* at this point, p needs to == the place the object was moved to, 
	 * so we can unswap it */
	if (next > p && !obj_free(p)) {
	    if (obj_swapped(p)) {
		REVERSE_SWAP(p);
	    }
	}
	p = next;
    }
#ifdef TRACEGC
    if (tracegc) {
	unsigned char* freep;
	for (freep = chunk->freePtr; 
	     freep != NULL;
	     freep = free_list_next(freep)) {
	    jio_fprintf(stderr, "free block 0x%lx  holds 0x%lx bytes\n", 
	            freep, obj_len(freep));
	}
    }
#endif /* TRACEGC */
    return blocks_moved;
}

static void
compactHeap(unsigned char **last_freep, unsigned int free_space_goal)
{
    register JHandle *hp;
    ChunkBlkP chunk;
    ChunkBlkP evacChunk = NULL;
    ChunkBlkP destChunk = NULL;
    int blocks_moved = 0;
    unsigned char *last_free = 0;
#ifdef DEBUG
    long start_time = sysGetMilliTicks();
#endif

    /*
     * Put a pointer to each object's handle in the first word of the object,
     * saving the first word of the object in the object pointer of the
     * handle. This way we can compact the heap looking only at the heap,
     * then go back and fix up the handles.  (we need to scan all objects in
     * heap order, not handle order to effectively compact everything).
     * if multiple handles point to one object, handles are linked together.
     * DRW: To Do - Shouldn't swap pinned objects!
     * Sheng: pinned objects are no longer swapped.
     */

    MAP_OVER_HANDLES_FROM_START(hp)
	if (ValidObject(hp->obj)) {
	    unsigned char *p = (unsigned char *)(hp->obj) - sizeof(hdr);
	    if (!IsMarked(p + sizeof(hdr)) && !obj_pinned(p)) {
		OBJECT T = *hp->obj;
	        *hp->obj = (OBJECT) hp;
		hp->obj = (OBJECT *) T;
		obj_setswapped(p);
	    }
	}
    END_MAP_OVER_HANDLES_FROM_START;
    
    /* There are many possible preconditions to freeing a chunk, but this is 
     * the most basic: Don't evacute if gc-ing due to memory request, or if 
     * there's only one chunk.
     */
    if ((firstObjBlkP != lastObjBlkP) &&
	(free_space_goal == 0xFFFFFFFF)) {
	evacChunk = evacuationChunk();
	if (evacChunk) {
	    /* becuase of the way we calc the free lists, the dest needs to be 
	     * before the evacChunk, so we can't evac the first chunk. */
	    if (evacChunk != firstObjBlkP) {
		destChunk = firstObjBlkP; /* not great, but good guess */
	    } else {
		evacChunk = NULL;
	    }
#ifdef TRACEGC
	    if (tracegc) {
		jio_fprintf(stderr, "Attempting to free block 0x%lx\n", evacChunk);
	    }
#endif /* TRACEGC */
	}
    }
    
    /*
     * The main loop - look at every object in every chunk
     */
    chunk = firstObjBlkP;		 /* iterates over all object chunks */
    do {
	int result = compactChunk(chunk, 
	                          (chunk == evacChunk) ? destChunk : NULL);
	if (result >= 0) {
	    blocks_moved += result;
	} else {	/* bail out due to async gc getting interupted */
	    *last_freep = 0;
	    break;	/* still need to do some clean up */
	}
	chunk = chunk->nextPtr;			/* advance to next chunk */
    } while (chunk != firstObjBlkP);
    
    if (evacChunk) {
	if (evacChunk->freeCnt == evacChunk->allocSize) {
	    shrinkObjectSpace(evacChunk);
	} else {
#ifdef TRACEGC
	    if (tracegc) {
		jio_fprintf(stderr, 
"Tried to free block 0x%lx, but it still is using 0x%lx bytes \n",
			evacChunk, evacChunk->allocSize - evacChunk->freeCnt);
	    }
#endif /* TRACEGC */
	}
    }

    /* turn free lists back into free hints the way searchChunk expects */
    chunk = firstObjBlkP;
    do {
	if (!chunk->freePtr) {	/* non-NULL free lists make good hints, */
	    chunk->freePtr = chunk->startPtr;	/* but NULL hints are bad. */
	}
	chunk = chunk->nextPtr;
    } while (chunk != firstObjBlkP);

    sysAssert(last_free == 0 || obj_free(last_free));
    *last_freep = last_free;

#ifdef DEBUG
    if (verbosegc) {
	long total_time = sysGetMilliTicks() - start_time;
	jio_fprintf(stderr, "<GC: compactHeap took %d ms, blocks_moved=0x%lx>\n",
		total_time, blocks_moved);
    }
#endif
}

#else  /************ CONTIGUOUS HEAPS: ********************/

/*
 * This is free list management code for the compacter.  It uses multiple
 * bins of power of 2 sizes.  Blocks within a bin are sorted by address and
 * linked together into a linked list.  The free_list_mask is used to keep
 * track of which bins actually have blocks in them so that searches which
 * would be likely to fail can be short circuited.
 */


#define NUM_FREE_LIST_BINS 8
static unsigned char *free_list_bin[NUM_FREE_LIST_BINS];
static unsigned int free_list_mask;

#define FREE_LIST_MARK_BIN(index) free_list_mask |= (1 << (index))
#define FREE_LIST_CLEAR_BIN(index) free_list_mask &= ~(1 << (index))
#define FREE_LIST_BIN_EMPTY(index) free_list_mask & (1 << (index))
#define FREE_LIST_NEXTP(ptr) (((unsigned char **) (ptr)) + 1)


/*
 * selectBin returns the lowest numbered bin which could possibly
 * accomodate the request.  This would be the point at which a search
 * would being or the bin into which a new free block would go.
 */

static int
selectBin(int len)
{
    int bin = 32;
    int i = 0;
    while (i < NUM_FREE_LIST_BINS) {
	if (len <= bin)
	    return i;
	i++;
	bin <<= 1;
    }
    return NUM_FREE_LIST_BINS - 1;
}


/*
 * Insert a free block in the free list.
 */

static void
addBlock(unsigned char * ptr)
{
    int ptrlen = obj_len(ptr);
    int bin = selectBin(ptrlen);
    unsigned char ** lastp;
    unsigned char * p;

    /*
     * Object must be at least large enough to hold a header word
     * and a pointer.  This is definitely true in the current sources.
     */
    sysAssert(obj_len(ptr) >= 2 * sizeof(void *));
    sysAssert(obj_free(ptr));

    /*
     * we walk over the singly linked list and keep a pointer to
     * where we got this pointer from.  This allows us to link
     * the new object into the list easily.
     */
    lastp = &free_list_bin[bin];
    p = *lastp;

    FREE_LIST_MARK_BIN(bin);
    while (p) {
	if (p > ptr)
	    break;
	lastp = FREE_LIST_NEXTP(p);
	p = *lastp;
    }
    (((unsigned char **) ptr)[1]) = *lastp;
    *lastp = ptr;
}


/*
 * Search for a block of at least size plen, unlink it from the free list
 * and return it.
 */

static unsigned char *
findBlock(int plen)
{
    int i;
    for (i = selectBin(plen);
	 ((free_list_mask & ~((1 << i) - 1)) != 0) &&
	     i < NUM_FREE_LIST_BINS; i++) {
	unsigned char ** lastp = &free_list_bin[i];
	unsigned char * p = *lastp;
	while (p) {
	    if (obj_len(p) >= plen) {
		unsigned char * nextp = *FREE_LIST_NEXTP(p);
		*lastp = nextp;
		sysAssert((free_list_mask & ~((1 << i) - 1)));
		if (nextp == NULL && free_list_bin + i == lastp)
		    FREE_LIST_CLEAR_BIN(i);
		return p;
	    }
	    lastp = FREE_LIST_NEXTP(p);
	    p = *lastp;
	}
    }
    return NULL;
}


/*
 * Returns the highest address block in the free list.
 */
static unsigned char *
lastFreeBlock(void)
{
    int i = NUM_FREE_LIST_BINS - 1;
    unsigned char ** lastp = &free_list_bin[i];
    while (i >= 0) {
	unsigned char * p = *lastp;
	while (p) {
	    lastp = FREE_LIST_NEXTP(p);
	    if (*lastp == NULL)
		return p;
	    p = *lastp;
	}
	i--;
	lastp--;
    }
    return NULL;
}


#ifdef DEBUG
static void
dumpFreeBlocks(bool_t search, char * msg)
{
    if (msg)
	jio_fprintf(stderr, "%s\n", msg);

    if (search) {
	SetLimits();
	unsigned char * p = opmin;
	unsigned char * next;
	int totalBytes = 0;
	int total = 0;
	int largest = 0;
	unsigned char * low_free_block = opmax;

	while (p < opmax) {
	    int plen = obj_len(p);
	    next = p + plen;
	    if (obj_free(p)) {
		int len = obj_len(p);
		jio_fprintf(stderr, "%p %ld (end %p)\n", p, obj_len(p),
			    p + obj_len(p));
		total++;
		totalBytes += len;
		sysAssert(next == opmax ||
			  obj_pinned(next) || IsMarked(next + sizeof(hdr)));

		if (p < low_free_block)
		    low_free_block = p;

		if (len > largest) largest = len;
	    }
	    p = next;
	}
	jio_fprintf(stderr, "%d free blocks %d free bytes %d average size %d"
		    "largest block 0x%p low free block\n",
		    total, totalBytes, totalBytes / total,
		    largest, low_free_block);
    } else {
	int i = 0;
	while (i < NUM_FREE_LIST_BINS) {
	    unsigned char * p = free_list_bin[i];
	    while (p) {
		jio_fprintf(stderr, "%p %ld\n", p, obj_len(p));
		p = *(((unsigned char **) p) + 1);
	    }
	    i++;
	}
    }
}


static void
validateHeap(unsigned char * end, unsigned char * prev_free)
{
    SetLimits();
    unsigned char * p = opmin;
    while (p < end) {
	int plen = obj_len(p);
	unsigned char * next = p + plen;
	if (obj_free(p)) {
	    sysAssert(next == end || 
		      obj_pinned(next) || IsMarked(next + sizeof(hdr)));
	}
	sysAssert(plen != 0);
	sysAssert(p + plen <= opmax);
	sysAssert(p + plen >= opmin);
	p = next;
    }
}
#endif /* DEBUG */

/*
 * Concept: Assuming optimal packing, there are only two places that
 * holes, or bubbles in the heap can occur.  These are immediately before
 * each pinned range of memory, and at the tail of the heap (the last
 * range is always free).
 *
 * This algorithm walks the heap from beginning to end, performing
 * several operations:
 *  1. Coalesce adjacent free blocks.
 *  2. Record bubbles in the free_list_bin[] array.  This is an array
 *     containing the free blocks found in the heap.
 *  3. Copy unpinned, allocated blocks down into the lowest free
 *     block available or slide the block down when a free block is
 *     right next to it.
 *
 * The goal of this scheme is to leave as many large free blocks available
 * as is possible.
 *
 * We swap objects but we don't normally mark them as swapped since we
 * don't really need the flag.  The DEBUG version does mark them just to
 * validate it's behaviour.
 */
static void
compactHeap(unsigned char **last_freep, unsigned int free_space_goal)
{
    SetLimits();
    JHandle *hp;
    unsigned char *p;
    unsigned char *prev_free;
#ifdef HPROF
    bool_t jvmpi_obj_move_on = JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_OBJ_MOVE);
#endif
    /*
     * p points at the current block we are operating on
     * prev_free points at the previous block if it's is free.
     */

#ifdef VERBOSEGC
    int blocks_moved = 0;
    long start_time, swap_time;
    int pinned_count = 0;
    int marked_count = 0;

    if (verbosegc)
	start_time = sysGetMilliTicks();
#endif /* VERBOSEGC */

    /*
     * First we have to swap a word in the handle with a word in the
     * object so that if we move the object, we can swap it back and
     * have the handle point at the object in the new location.
     */

    MAP_OVER_HANDLES_FROM_START(hp)
	if (hp->obj) {
	    unsigned char *p = (unsigned char *)(hp->obj) - sizeof(hdr);
	    /* Pinned objects are not swapped. */
	    if (!IsMarked(p + sizeof(hdr)) && !obj_pinned(p)) {
		obj_swap(hp);
#ifdef DEBUG
		obj_setswapped(p);
#endif
	    }
	}
    END_MAP_OVER_HANDLES_FROM_START;

#ifdef VERBOSEGC
    if (verbosegc) {
	swap_time = sysGetMilliTicks();
    }
#endif /* VERBOSEGC */

    /*
     * Reset the free list.
     */
    memset((char *)free_list_bin, 0, sizeof(free_list_bin));
    free_list_mask = 0;

    /*
     * The main loop - look at every object in the pool
     */
    p = opmin;
    prev_free = NULL;

    while (p < opmax) {
        size_t plen = obj_len(p);
	unsigned char * next = p + plen;
	sysAssert(next >= opmin && next <= opmax);
	sysAssert(next == opmax || obj_len(next) != 0);

	if (obj_free(p)) {

	    /* Coalesce adjacent free blocks */
	    while (next && obj_free(next)) {
		next += obj_len(next);
#ifdef VERBOSEGC
		blocks_moved++;
#endif /* VERBOSEGC */
	    }
	    plen = next - p;
	    /*
	     * If the previous free block is right next to this one
	     * then we can combine this with it into a larger free block.
	     * Otherwise we put the previous free block onto the free
	     * list and make this block the previous free block.
	     */
	    if (prev_free) {
		int prev_free_len = obj_len(prev_free);
		if (prev_free + prev_free_len == p) {
		    p = prev_free;
		    plen += prev_free_len;
		} else {
		    addBlock(prev_free);
		}
	    }

	    obj_setlf(p, plen, 1);
	    sysAssert(obj_len(p) != 0);

	    /*
	     * Save this block because we may just want to slide the
	     * objects after us down to here
	     */
	    prev_free = p;

	    sysAssert(next >=opmax || obj_len(next) != 0);
	} else if (!IsMarked(p + sizeof(hdr)) && !obj_pinned(p)) {
	    /*
	     * Copy p down into the lowest available free block.
	     * Find lowest block that is big enough to hold the block.
	     */
	    unsigned char *lowest_addr;

	    if (free_list_mask != 0) {
		lowest_addr = findBlock(plen);
	    } else
		lowest_addr = NULL;

	    /* did we find a slot? */
	    if (lowest_addr != NULL) {
		size_t lowest_len = obj_len(lowest_addr);

		/*
		 * If a slot is found, move the block, and mark the old
		 * slot as free.  If the free space was bigger than we needed,
		 * break off a new chunk.  Note that this object is somewhere
		 * below us and is guaranteed not to overlap with us so
		 * we can use memcpy instead of memmove, which may be faster.
		 */
		sysAssert(lowest_len >= plen);
		sysAssert(lowest_addr < p);
		sysAssert(obj_free(lowest_addr));
		sysAssert(obj_swapped(p));
		memcpy(lowest_addr, p, plen);
#ifdef HPROF
		if (jvmpi_obj_move_on) {
		    jvmpi_move_object(p + sizeof(hdr),
				      lowest_addr + sizeof(hdr));
		}
#endif /* HPROF */
#ifdef VERBOSEGC
		blocks_moved++;
#endif /* VERBOSEGC */
		obj_setfree(p);
		FILL_CHECK_WORD(p);

		if (prev_free) {
		    /*
		     * The previous block was free and we just freed this
		     * one, so we coalesce them into one large block. We
		     * may eventually be able to slide a bunch of objects
		     * into it's place.
		     */
		    obj_setlf(prev_free, obj_len(prev_free) + plen, 1);
		    sysAssert(obj_len(prev_free) != 0);
		} else {
		    prev_free = p;
		    sysAssert(obj_len(prev_free) != 0);
		}

		/* it was bigger than we needed, break off a piece. */
		if (lowest_len > plen) {
		    unsigned char *fp = lowest_addr;
		    fp += plen;
		    obj_setlf(fp, lowest_len - plen, 1);
		    FILL_CHECK_WORD(fp);
		    addBlock(fp);
		}

		sysAssert(!obj_free(lowest_addr) && obj_swapped(lowest_addr));
		REVERSE_SWAP(lowest_addr);
		sysAssert(next >= opmax || obj_len(next) != 0);
	    } else if (prev_free) {
		/*
		 * The block next to this one is free so we can slide
		 * some blocks down into it.
		 */
		unsigned char * ptr = next;
		int prev_len = obj_len(prev_free);
		int slidelen = plen;
		sysAssert(prev_len != 0);

		if (free_list_mask == 0) {
		    /*
		     * if there are no free blocks, we can move
		     * several objects safely.
		     */
		    while (slidelen < GC_SLIDE_LENGTH &&
			   ptr < opmax &&
			   !obj_free(ptr) && !obj_pinned(ptr) &&
			   !IsMarked(ptr + sizeof(hdr))) {
			int l = obj_len(ptr);
			ptr += l;
			slidelen += l;
		    }
		}
		
		next = ptr;
 		memmove(prev_free, p, slidelen);
#ifdef HPROF
		if (jvmpi_obj_move_on) {
		    int len = slidelen;
		    int dist = p - prev_free;
		    p = prev_free;
		    while (len > 0) {
		        int l = obj_len(p);
			sysAssert(l > 0);
			sysAssert(obj_swapped(p));
			sysAssert(!obj_free(p));
			jvmpi_move_object(p + sizeof(hdr) + dist,
					  p + sizeof(hdr));
			p += l;
			len -= l;
		    }
		}
#endif /* HPROF */

		/* the free block is now at the end of the block we moved */
		p = prev_free;
		sysAssert(obj_len(p) != 0);
		sysAssert(!obj_free(p));
		prev_free += slidelen;
		obj_setlf(prev_free, prev_len, 1);
		sysAssert(obj_len(prev_free) != 0);
		sysAssert(obj_free(prev_free));

		/* walk back over the block and swap the handles back */
		while (slidelen > 0) {
		    int l = obj_len(p);
		    sysAssert(l > 0);
		    sysAssert(obj_swapped(p));
		    sysAssert(!obj_free(p));
		    REVERSE_SWAP(p);
		    p += l;
		    slidelen -= l;
		}
	    } else {
		/* we couldn't move this block so just swap it back */
		sysAssert(obj_swapped(p));
		sysAssert(!obj_free(p));
		REVERSE_SWAP(p);
		if (prev_free) {
		    addBlock(prev_free);
		    prev_free = NULL;
		}
	    }
	} else {
	    /*
	     * This case corresponds to a marked or pinned object that we
	     * aren't allowed to move.
	     */

	    sysAssert(!obj_swapped(p));
	    if (prev_free) {
		addBlock(prev_free);
		prev_free = NULL;
	    }
#ifdef VERBOSEGC
	    if (verbosegc) {
		if (obj_pinned(p))
		    pinned_count += 1;
		else
		    marked_count += 1;
	    }
#endif /* VERBOSEGC */
	}
	sysAssert(next == opmax || obj_len(next) != 0);
	p = next;
	sysAssert(next >= p);
    }

#ifdef VERBOSEGC
    if (verbosegc) {
	long total_time = sysGetMilliTicks() - start_time;
	jio_fprintf(stderr, "<GC: compactHeap took %ld ms,"
		    " swap time = %ld ms, blocks_moved=%d>\n",
		    total_time, swap_time - start_time, blocks_moved);
	jio_fprintf(stderr, "<GC: %d explicitly pinned objects, "
		    "%d conservatively pinned objects>\n",
		    pinned_count, marked_count);
    }
#endif /* VERBOSEGC */

    if (prev_free)
	*last_freep = prev_free;
    else
	*last_freep = lastFreeBlock();

    sysAssert(*last_freep == 0 || obj_free(*last_freep));

#ifdef VERBOSEGC
    if (verbosegc && *last_freep) {
	jio_fprintf(stderr, "<GC: last free block at 0x%p of length %ld>\n",
		    *last_freep, obj_len(*last_freep));
	if (*last_freep + obj_len(*last_freep) < opmax)
	    jio_fprintf(stderr, "<GC: last free block is not at end>\n");
	else
	    jio_fprintf(stderr, "<GC: last free block is at end>\n");
	 
    }
#endif /* VERBOSEGC */
}

#endif  /************ END CONTIGUOUS HEAPS ********************/

/*
 * Garbage collect memory.  If 'async_call' is true, this function is being
 * called from the asych GC thread, and should immediately exit whenever
 * sysInterruptsPending() returns true.  The 'free_space_goal' is
 * the amount of memory in a single block which the caller is interested
 * in having free.  If such a block does not exist, compact the heap.
 *
 * gc0_locked() should only be called when holding the heap lock.  gc() or
 * realObjAlloc() acquires the heap lock, then calls gc0() which grabs the
 * finalization locks and calls gc_locked(), which does system-specific tasks
 * then calls gc0_locked().
 */

unsigned char *
gc0_locked(unsigned int free_space_goal, bool_t clear_soft_refs)
{
    /* SPARC only: "ta ST_FLUSH_WINDOWS" done by "gc" stub in machgc.s */
    int do_compact_heap;
    long start_time = 0;	/* Keep lint happy */

    unsigned char *last_free = 0;
    SetLimits();

#ifdef VERBOSEGC
    /* We need to initialize start_time if verbose OR if we're profiling */
    if (verbosegc) {
	start_time = sysGetMilliTicks();
	init_time = scan_time = sweep_time = compact_time = 0;
	reg_marked_object_count = stack_marked_object_count = 0;
	reg_marked_handle_count = stack_marked_handle_count = 0;
    }
#endif

    sysAssert(HEAP_LOCKED(sysThreadSelf()));
    sysAssert(QUEUE_LOCKED(sysThreadSelf()));
    sysAssert(BINCLASS_LOCKED(sysThreadSelf()));
    sysAssert(GLOBALREF_LOCKED(sysThreadSelf()));

    nfreed = 0;
    bytesfreed = 0; /* used by realObjAlloc to decide gc vs. allocate more */

    /* Update expansion history */
    expansionHist[expansionHistNext++] = expanded;
    if (expansionHistNext >= EXPANSION_HIST_SIZE) {
	expansionHistNext = 0;
    }
    expanded = 0;

    clearMarkBits();

#ifdef DEBUG
    markbitsInUse = TRUE;
#endif

    if (ValidHandle(refEnd)) {
        MarkPtr(refEnd);
    }

    scanMonitorCache();

    scanThreads();

    scanJNIGlobalRefs();

    scanHeap();
#ifdef VERBOSEGC
    if (verbosegc)
	init_time = sysGetMilliTicks() - start_time;
#endif

    markStickyClasses();

    scanHandles();
    processRefs(clear_soft_refs);
#ifdef VERBOSEGC
    if (verbosegc)
	scan_time = sysGetMilliTicks() - start_time - init_time;
#endif

#ifdef HPROF
    if (jvmpi_event_flags && jvmpi_in_dump_heap_flag) {
        jvmpi_gc_dump_heap();
    }
#endif /* HPROF */

    /* It is important to call clearJNIWeakRefs before unloadUnusedClasses
     * because unloadUnusedClasses will mark all the classes (See comment
     * in unloadUnusedClasses).
     */
    clearJNIWeakRefs();

    deleteUnusedInternedStrings();

    if (classgc) {
        unloadUnusedClasses();
    }

    /*
     * Since from now on unused classes will ultimately be freed in
     * freeClasses, we must make sure all unused objects are freed as
     * well. The system must never enter a situation where normal
     * objects point to freed objects.
     */

    do_compact_heap = freeSweep(free_space_goal);
#ifdef PDEBUG    /* always compact...*/
    do_compact_heap = 1;
#endif

#ifdef HPROF
    /* don't compact the heap during a dump, as we don't want the profiler
     * agent to get object_move events for objects already recorded in the
     * dump */
    if (jvmpi_event_flags && jvmpi_in_dump_heap_flag) {
        do_compact_heap = 0;
    }
#endif

#ifdef VERBOSEGC
    if (verbosegc) {
	sweep_time = sysGetMilliTicks() - start_time - init_time - scan_time;
    }
#endif

    if (do_compact_heap) {
	/* The heap is compacted in adress order, so there shouldn't be
	 * any benefit to doing it multiple times */
	compactHeap(&last_free, free_space_goal);
    }

#ifdef DEBUG
    markbitsInUse = FALSE;
#endif

    if (maxHeapFreePercent < 1) {
	size_t maxFree = (size_t)(TotalObjectCtr * maxHeapFreePercent);
	size_t extra_space = 0;
	if (FreeObjectCtr > maxFree) {
	    extra_space = FreeObjectCtr - maxFree;
	    /*
	     * we'd like to shrink by extra_space but we can't drop below
	     * our initial size, so calculate the minimum we could release
	     * without dropping below the initial size
	     */
	    if (TotalObjectCtr - extra_space < MinTotalObjectCtr) {
		extra_space = TotalObjectCtr - MinTotalObjectCtr;
	    }
	}
	if (extra_space > minHeapExpansion) {
	    if ((!do_compact_heap) &&
		(!(jvmpi_event_flags && jvmpi_in_dump_heap_flag))) { 
		/* we force a compaction so the free space is at the end */
	        compactHeap(&last_free, ~0);
	    }
	    shrinkObjectSpace(extra_space, &last_free);
	}
    }

#ifdef VERBOSEGC
    if (verbosegc) {
	compact_time = sysGetMilliTicks() - start_time - 
	    init_time - scan_time - sweep_time;

	total_time = sysGetMilliTicks() - start_time;
    }
#endif

    timeOfLastGC = sysGetMilliTicks();

    return last_free;
}

#ifdef VERBOSEGC
static void
printGCMessages(void)
{
    jio_fprintf
        (stderr,
	 "<GC: freed %ld objects, %ld bytes in %ld ms, %d%% free (%ld/%ld)>\n",
	 nfreed, bytesfreed, total_time,
	 (int)(100 * (double)FreeObjectCtr / (double)TotalObjectCtr),
	 FreeObjectCtr, TotalObjectCtr);
    jio_fprintf(stderr,
		"  <GC: init&scan: %ld ms, scan handles: %ld ms,"
		" sweep: %ld ms, compact: %ld ms>\n",
		init_time, scan_time, sweep_time, compact_time);
    jio_fprintf(stderr, "  <GC: %d register-marked objects, "
		"%d stack-marked objects>\n",
		reg_marked_object_count, stack_marked_object_count);
    jio_fprintf(stderr, "  <GC: %d register-marked handles, "
		"%d stack-marked handles>\n",
		reg_marked_handle_count, stack_marked_handle_count);
    jio_fprintf(stderr,
		"  <GC: refs: soft %d (age >= %d), weak %d, "
		"final %d, phantom %d>\n",
		refEnqCounts[0], softRefThreshold,
		refEnqCounts[1],
		refEnqCounts[2],
		refEnqCounts[3]);
}
#endif /* VERBOSEGC */

/*
 * Locking and unlocking special system monitors for GC
 */

void
lockGCLocks(struct execenv *ee)
{
    sysAssert(HEAP_LOCKED(EE2SysThread(ee)));

    /*
     * To ensure consistency of locked data structures, we should not 
     * be holding any of the following locks when GC occurs.  Special 
     * dispensation may be given to locks that protect code that is
     * proven safe at points of possible recursive entry.
     */
    sysAssert(!LINKCLASS_LOCKED(EE2SysThread(ee)));
    sysAssert(!QUEUE_LOCKED(EE2SysThread(ee)));
    sysAssert(!BINCLASS_LOCKED(EE2SysThread(ee)));
    sysAssert(!GLOBALREF_LOCKED(EE2SysThread(ee)));

    monitorEnter2(ee, obj_monitor(refLock));
    LINKCLASS_LOCK(EE2SysThread(ee));
    QUEUE_LOCK(EE2SysThread(ee));
    BINCLASS_LOCK(EE2SysThread(ee));
    GLOBALREF_LOCK(EE2SysThread(ee));
    CACHE_LOCK(EE2SysThread(ee));

    /* The GC_DISABLE lock has to be the leaf lock because sending JVMPI
     * events will acquire this lock, and it is unknown what other locks 
     * the thread may be holding at that point.
     */
    GC_DISABLE_LOCK(EE2SysThread(ee));
    while (gc_disable_count) {
        if (SYS_INTRPT == sysMonitorWait(EE2SysThread(ee),
					 _gc_disable_lock,
					 SYS_TIMEOUT_INFINITY)) {
	    sysThreadInterrupt(EE2SysThread(ee));
	}
    }
}

void
unlockGCLocks(struct execenv *ee)
{
    sysAssert(HEAP_LOCKED(EE2SysThread(ee)));

    /* Make sure we are holding the things we want to unlock */
    sysAssert(LINKCLASS_LOCKED(EE2SysThread(ee)));
    sysAssert(QUEUE_LOCKED(EE2SysThread(ee)));
    sysAssert(BINCLASS_LOCKED(EE2SysThread(ee)));
    sysAssert(GLOBALREF_LOCKED(EE2SysThread(ee)));

    GC_DISABLE_UNLOCK(EE2SysThread(ee));

    /* Remember that unlocking order must be the reverse of lock order! */
    CACHE_UNLOCK(EE2SysThread(ee));
    GLOBALREF_UNLOCK(EE2SysThread(ee));
    BINCLASS_UNLOCK(EE2SysThread(ee));
    QUEUE_UNLOCK(EE2SysThread(ee));
    LINKCLASS_UNLOCK(EE2SysThread(ee));
    monitorExit2(ee, obj_monitor(refLock));
}

/*
 * Interface to synchronous garbage collection from other memory manage-
 * ment code.  This is called call when memory allocation fails.  We
 * expect to have the heap lock when it is called. 
 */

unsigned char *gc_locked(unsigned int, bool_t); /* Call holding locks! */

static unsigned char *
gc0(unsigned int free_space_goal, bool_t clear_soft_refs)
{
    unsigned char *last_free = 0;
    struct execenv *ee = EE();

#ifdef TRACEGC
    if (tracegc) {
	jio_fprintf(stderr, "*** gc ***	 goal=%ld\n", free_space_goal);
	DumpHeapInfo();
    }
#ifdef LDEBUG
    verifyHandleCtrs();
    verifyObjectCtrs();
    validatePool(0);
#endif /* LDEBUG */
#endif /* TRACEGC */

    lockGCLocks(ee);

#ifdef HPROF
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_GC_START)) {
        jvmpi_gc_start();
    }
#endif /* HPROF */

    if (sysThreadSingle() == SYS_OK) {
	last_free = gc_locked(free_space_goal, clear_soft_refs);
    }

    sysThreadMulti();

    freeClasses();
    freeInternedStringBuckets();

#ifdef VERBOSEGC
    /* safe to print in multi-threaded mode */
    if (verbosegc)
        printGCMessages();
#endif

#ifdef HPROF
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_GC_FINISH)) {
        jvmpi_gc_finish(TotalHandleCtr - FreeHandleCtr,
			TotalObjectCtr - FreeObjectCtr,
			TotalObjectCtr);
    }
#endif /* HPROF */

    if (notifyRefLock) {
       monitorNotifyAll2(ee, obj_monitor(refLock));
    }
    unlockGCLocks(ee);

#ifdef TRACEGC
    if (tracegc) {
	jio_fprintf(stderr, "** gc done **\n");
	DumpHeapInfo();
    }
#ifdef LDEBUG
    verifyHandleCtrs();
    verifyObjectCtrs();
    validatePool(0);
#endif /* LDEBUG */
#endif /* TRACEGC */

    return last_free;
}

/*
 * User interface to synchronous garbage collection.  This is called
 * by an explicit call to GC. At the call we do not already have the
 * heap lock, so grab it.
 */
void 
gc(unsigned int free_space_goal)
{
    ExecEnv *ee = EE();
    HEAP_LOCK(EE2SysThread(ee));
    (void) gc0(free_space_goal, FALSE);
    HEAP_UNLOCK(EE2SysThread(ee));
}

/*
 * Initialize the Java heap.  Argument max_request is the maximum heap
 * memory we are prepared to devote to the heap, and min_request is the
 * minimum memory and amount allocated on startup.
 */ 

#ifdef PAGED_HEAPS /************ PAGED HEAPS: ********************/

static void
initializeAlloc0(long max_request, long min_request)
{
    /* set up global mark bits pool */
    markbits = NULL;
    markbitsfreeptr = NULL;
    marksize = 0;

    /* setup page map so it will get allocated */
    page_map = NULL;
    /* not really needed, but could flush out problems: */
    mem_base = (void *)0xffffffff; 
    mem_top = NULL;

    /* setup linked lists */
    /* doubly-linked list of handle chunks, in address order */
    firstHanBlkP = NULL;
    lastHanBlkP = NULL;
    curHanBlkP = NULL;
    /* doubly-linked list of object chunks, in address order */
    firstObjBlkP = NULL;
    lastObjBlkP = NULL;
    curObjBlkP = NULL;

    FreeObjectCtr = 0;
    TotalObjectCtr = FreeObjectCtr;
    expanded = 0;
    FreeHandleCtr = 0;
    TotalHandleCtr = FreeHandleCtr;

    gMaxMem = max_request;		/* save max memory to be used */

#ifdef LDEBUG
    min_request = min_request / 4;
#endif

    /* initial handle chunk - should be 20-25% object total space  */
    if (! expandHandleSpace(PTR_2_PAGENUM((int)(min_request * 0.20)) + 1)) {
	out_of_memory();
    }
    if (! expandObjectSpace(((int)(min_request * 0.80)), NULL)) {
	out_of_memory();
    }
    initializeDeferredHandleTable();
}

#else  /************ CONTIGUOUS HEAPS: ********************/

#define BACKOFF_FACTOR 0.75

static void
initializeAlloc0(int max_request, int min_request)
{
    size_t max, min;
    size_t marksize;

    /*
     * Allocate the initial heap memory.  In the abstract we first map
     * the maximum we are interested in, then commit to the minimum/initial
     * amount.	Note that as actually implemented on some platforms,
     * commitment might be a no-op, min might equal max etc.  
     *
     * Because we don't really need more than min_request to start up,
     * we can back off on the heap mapping if it initially fails.  There
     * is no guarantee that we will be allowed to expand to the max anyhow.
     * Related spaces (e.g. mark bits) scale off what we actually get.
     * We should back off in fairly large chunks so if we are successful
     * we'll be able to map the rest of what we'll need.
     */
    sysAssert(hpool == 0);
    heapbase = (unsigned char *) sysMapMem(max_request, &max);
    while (!heapbase) {
        max_request -= (int)(max_request * BACKOFF_FACTOR);
	if (max_request < min_request) {
	    out_of_memory();    /* Can't start up */
	}
	heapbase = (unsigned char *) sysMapMem(max_request, &max);
    }
    heaptop = heapbase + max;

    /*
     * We want same partition proportions (20%/80%) of the mapped 
     * memory as the committed memory.	The handle pool base must be
     * HANDLEGRAIN aligned within the committed memory.
     */
    hpool = heapbase + (intptr_t) ((heaptop - heapbase) * 0.20);
    hpool = hpool - (intptr_t) (min_request * 0.20);
    hpool = sysCommitMem(hpool, min_request, &min);
    if (!hpool) {
        out_of_memory();	/* Can't start up */
    }
    hpoollimit = hpool + min;
    hpool = (unsigned char *)
        ALIGN_UP((intptr_t) hpool,(uintptr_t)(HANDLEGRAIN));

    /*
     * Leave the bottom 20% of the pool for handles, the rest for objects.
     */
    opoollimit = hpoollimit - sizeof(hdr);
    hpoollimit = hpool + ALIGN_DOWN(((intptr_t) ((hpoollimit - hpool) * 0.20)),
                 (uintptr_t)HANDLEGRAIN);

    /*
     * The object pool base must be sizeof(hdr) past OBJECTGRAIN alignment
     */
    opool = hpoollimit;
    while (!IS_ALIGNED(((intptr_t) opool + sizeof(hdr)),OBJECTGRAIN)) { 
        opool += sizeof(hdr);
    }

    /*
     * Map the mark bits array, whose size is defined by the maximum
     * (mapped) memory, and commit to enough memory to map the committed
     * heap.
     */
    marksize = ((max/(OBJECTGRAIN*BITSPERMARK) + 1)) * sizeof(*markbits);
    markbits = (unsigned int *) sysMapMem(marksize, &marksize);
    if (!markbits) {
        out_of_memory();	/* Can't start up */
    }
    markmax = markbits + marksize / sizeof(*markbits);
    markend = markbits;

    /* Could postpone commit until GC; see comment at expandMarkBits() */
    expandMarkBits();

    /*
     * Make sure that all the handles are zero.	 We recognize a free handle
     * by the fact that its ->obj field is 0.
     */
    linkNewHandles((JHandle*)hpool, (JHandle*)hpoollimit, 0);
    hpoolfreelist = hpool;

    obj_setlf(opool, opoollimit - opool, 1);

#ifndef FAST_MONITOR
#ifdef DEBUG
    ((long *) opool)[1] = 0x55555555;
#endif /* DEBUG */
#endif /* FAST_MONITOR */

    obj_setlf(opoollimit, 0, 0);
    FreeObjectCtr = opoollimit - opool;
    MinTotalObjectCtr = TotalObjectCtr = FreeObjectCtr;
    expanded = 0;
    FreeHandleCtr = hpoollimit - hpool;
    TotalHandleCtr = FreeHandleCtr;

    opoolhand = opool;
    initializeDeferredHandleTable();
}

#endif  /************ END CONTIGUOUS HEAPS ********************/

/*
 * Initialize the Java heap.  Argument max_request is the maximum heap
 * memory we are prepared to devote to the heap, and min_request is the
 * minimum memory and amount allocated on startup.
 *
 * If preallocating the entire heap (#ifdef USE_MALLOC), then we always
 * allocate max_request on startup.
 *
 * The heap and heap reserve sizes, whether defaults or as overridden on
 * the command line, must satisfy the constraint
 *
 *     orange reserve + red reserve < initial heap <= maximum heap
 *
 * else the runtime will not start up.  In this case InitializeAlloc
 * returns FALSE to its caller, so the caller can do/say something
 * appropriate in the context or on the platform.
 */ 

bool_t
InitializeAlloc(long max_request, long min_request)
{
    _gc_disable_lock = sysMalloc(sysMonitorSizeof());
    sysMonitorInit(_gc_disable_lock);

#ifdef HPROF
    jvmpi_new_arena();
#endif /* HPROF */
	
    sysAssert(OBJECTGRAIN >= sizeof(hdr));

#ifdef USE_MALLOC
    /* If we are just malloc'ing, allocate max amount up front */
    min_request = max_request;
#endif /* USE_MALLOC */

    initializeAlloc0(max_request, min_request); /* max or min may be changed */

    /*
     * Allocate memory locks early to avoid order conflicts on early GC.
     * See bug 1259363, for instance.
     */
    _heap_lock = (sys_mon_t *) sysMalloc(sysMonitorSizeof());
    if (_heap_lock == NULL) {
        out_of_memory();
    }
    HEAP_LOCK_INIT();

#ifdef TRACEGC
    verifyHandleCtrs();
    verifyObjectCtrs();
    validatePool(0);
#endif

    timeOfLastGC = sysGetMilliTicks();
    return TRUE;
}

void EnableGC(ExecEnv *ee)
{
    GC_DISABLE_LOCK(EE2SysThread(ee));
    if (--gc_disable_count == 0) {
        sysMonitorNotifyAll(EE2SysThread(ee), _gc_disable_lock);
    }
    GC_DISABLE_UNLOCK(EE2SysThread(ee));
}

void DisableGC(ExecEnv *ee)
{
    GC_DISABLE_LOCK(EE2SysThread(ee));
    gc_disable_count++;
    GC_DISABLE_UNLOCK(EE2SysThread(ee));
}

bool_t GCEnabled(ExecEnv *ee)
{
    bool_t res;
    GC_DISABLE_LOCK(EE2SysThread(ee));
    res = (gc_disable_count == 0);
    GC_DISABLE_UNLOCK(EE2SysThread(ee));
    return res;
}

bool_t isHandle(void *p)
{
    SetLimits();
    return ValidHandle(p);
}

bool_t isObject(void *p)
{
    SetLimits();
    return ValidObject(p);
}

void heap_lock(ExecEnv *ee)
{
    HEAP_LOCK(EE2SysThread(ee));
}

void heap_unlock(ExecEnv *ee)
{
    HEAP_UNLOCK(EE2SysThread(ee));
}

void pin_object(void *obj)
{
    obj_pin((char *)obj - sizeof(hdr));
}

void unpin_object(void *obj)
{
    obj_unpin((char *)obj - sizeof(hdr));
}

int pinned_object(void *obj)
{
    return obj_pinned((char *)obj - sizeof(hdr));
}

void lock_for_debugger(struct execenv *ee)
{
    SYSLOADER_LOCK(EE2SysThread(ee));
    HEAP_LOCK(EE2SysThread(ee));

    /* This following is a copy of lockGCLocks except CACHE_LOCK, */
    /* because the cache lock must the last one to lock */
    monitorEnter2(ee, obj_monitor(refLock));
    LINKCLASS_LOCK(EE2SysThread(ee));
    QUEUE_LOCK(EE2SysThread(ee));
    BINCLASS_LOCK(EE2SysThread(ee));
    GLOBALREF_LOCK(EE2SysThread(ee));

    REGISTRY_LOCK(EE2SysThread(ee));
    lock_code(ee);
    lock_utf8_hash(ee);
    CACHE_LOCK(EE2SysThread(ee));
}

void unlock_for_debugger(struct execenv *ee)
{
    CACHE_UNLOCK(EE2SysThread(ee));
    unlock_utf8_hash(ee);
    unlock_code(ee);
    REGISTRY_UNLOCK(EE2SysThread(ee));

    /* This following is a copy of unlockGCLocks except CACHE_UNLOCK, */
    /* because the cache lock must the first one for unlocking */
    GLOBALREF_UNLOCK(EE2SysThread(ee));
    BINCLASS_UNLOCK(EE2SysThread(ee));
    QUEUE_UNLOCK(EE2SysThread(ee));
    LINKCLASS_UNLOCK(EE2SysThread(ee));
    monitorExit2(ee, obj_monitor(refLock));

    HEAP_UNLOCK(EE2SysThread(ee));
    SYSLOADER_UNLOCK(EE2SysThread(ee));
}

void GCFreeClass(ClassClass *cb) 
{
    ExecEnv *ee = EE();
    struct Classjava_lang_Class *ucb = unhand(cb);
    HEAP_LOCK(EE2SysThread(ee));
    FreeClass(cb);
    memset(ucb, 0, sizeof(struct Classjava_lang_Class));
    HEAP_UNLOCK(EE2SysThread(ee));
}


/* Finalizer support */

void
RunFinalizersOnExit(void)		/* Called by Exit */
{
    execute_java_static_method(0, classJavaLangRefFinalizer,
			       "runFinalizersOnExit", "()V");
}

bool_t isValidHandle(JHandle *h) 
{
    SetLimits();
    return ((ValidHandle(h) && h->obj != 0) || (h == NULL));
}

#ifdef HPROF

/* Get the type and size information of an object */
ClassClass * jvmpi_get_object_info(JHandle *h, int *is_array, int *size)
{
    SetLimits();

    *size = obj_len((char *)h->obj - sizeof(hdr)) - sizeof(hdr);
    if (h->methods) {
        switch(obj_flags(h)) {
	case T_NORMAL_OBJECT:
	    *is_array = 0;
	    return obj_classblock(h);

	case T_BOOLEAN:      
	case T_BYTE:      
	case T_SHORT:      
	case T_CHAR:      
	case T_INT:      
	case T_LONG:      
	case T_FLOAT:      
	case T_DOUBLE:
	    *is_array = obj_flags(h);
	    return NULL;

	case T_CLASS: {
	    HArrayOfObject *a = (HArrayOfObject *)h;
	    *is_array = T_CLASS;
	    return (ClassClass *)(unhand(a)->body[obj_length(a)]);
	}
	default:
	    return JVMPI_INVALID_CLASS; /* RED and ORANGE reserve,
					 * cache blocks. */
	}
    }
    return 0;
}

JHandle *jvmpi_obj_reverse_map(void *obj)
{
    SetLimits();
    JHandle *hp;

    MAP_OVER_HANDLES_FROM_START(hp)
        if ((void *)hp->obj == obj) {
	    return hp;
	}
    END_MAP_OVER_HANDLES_FROM_START;
    return NULL;
}

/* Set by jvmpi routines when we want a heap dump. */
void jvmpi_in_dump_heap(bool_t flag) 
{
    jvmpi_in_dump_heap_flag = flag;
}

unsigned long jvmpi_get_heap_size(void)
{
    return TotalObjectCtr;
}

/*
 * The following routines are cloned from the appropriate scan* 
 * GC funcitons. REMIND: keep them up to date!
 */
static void 
jvmpi_scan_java_frame(JavaFrame *frame,
		      stack_item *top_top_stack,
		      int jni_info,
		      ExecEnv *ee,
		      int dist_from_top)
{
    stack_item *ssc, *limit;
    JHandle *ptr;
    JavaStack *javastack;
    struct methodblock *mb = frame->current_method;
    SetLimits();
    bool_t is_native = mb ? mb->fb.access & ACC_NATIVE : TRUE;
    int root_info = IS_JIT_FRAME(frame) ? GC_ROOT_NATIVE_STACK :
                         (is_native ? jni_info : GC_ROOT_JAVA_FRAME);

    if (mb) {
        ClassClass *mbClass = fieldclass(&mb->fb);
	jvmpi_dump_add_root((JHandle *)mbClass, root_info, (void *)dist_from_top, ee);
    }

    limit = top_top_stack;
    javastack = frame->javastack;

    /* Scan the operand stack. */
    /*CONSTCOND*/
    while (1) {
        int is_first_chunk = IN_JAVASTACK((stack_item *)frame, javastack);
        for (ssc = is_first_chunk ? frame->ostack : javastack->data;
	     ssc < limit; ssc++) {
            ptr = ssc->h;
	    if (ValidHandle(ptr)) {	/* Never an object pointer */
 	        if (jni_info == GC_ROOT_JNI_GLOBAL) {
		    jvmpi_dump_add_root(ptr,
					GC_ROOT_JNI_GLOBAL,
					&(ssc->h),
					0);
		} else {
		    jvmpi_dump_add_root(ptr, root_info, (void *)dist_from_top, ee);
		}
	    }
	}
	if (is_first_chunk)
	    break;
	javastack = javastack->prev;
	limit = javastack->end_data;
    }

    /* Nothing more to do for pseudo and JIT frames. */
    if (mb == 0 || IS_JIT_FRAME(frame))
        return;

    if (mb->fb.access & ACC_NATIVE) {
        /* For native frames, we scan the arguments stored at the top 
	   of the previous frame. */
        JavaFrame *prev_frame = frame->prev;
	if (prev_frame == 0)
	    return;
	ssc = prev_frame->optop;
	limit = ssc + mb->args_size;
    } else {
        /* Scan local variables in Java frame */
        ssc = frame->vars;
	if (ssc == 0)
	    return;
	limit = (stack_item *)frame;
    }

    for (; ssc < limit; ssc++) {
        ptr = ssc->h;
	if (ValidHandle(ptr)) {	 /* Never an object pointer */
	    jvmpi_dump_add_root(ptr, root_info, (void *)dist_from_top, ee);
	}
    }
}

static JHandle *
jvmpi_find_handle_from_object(void *obj)
{
    JHandle *hp;
    SetLimits();

    MAP_OVER_HANDLES_FROM_START(hp)
	if (hp->obj == obj) {
	    return hp;
	}
    END_MAP_OVER_HANDLES_FROM_START;
    return NULL;
}

static int 
jvmpi_scan_thread_traces(sys_thread_t *t, void *arg)
{
    ExecEnv *ee = SysThread2EE(t);

    if (ee->initial_stack == NULL) {
        /* EE already destroyed. Thread is about to exit */
        return SYS_OK;
    }

    if (ee->current_frame) {
        jvmpi_dump_add_trace(ee);
    }
    return SYS_OK;
}

static int 
jvmpi_scan_thread_roots(sys_thread_t *t, void *arg)
{
    sys_thread_t *me = (sys_thread_t *) arg;
    ExecEnv *ee = SysThread2EE(t);
    unsigned char **ssc, **limit;
    void *base;
    SetLimits();

    if (ee->initial_stack == NULL) {
        /* EE already destroyed. Thread is about to exit */
        return SYS_OK;
    }

    {
	sys_thread_t *t = EE2SysThread(ee);
	long *regs;
	int nregs;

	/* Scan the saved registers */
	regs = sysThreadRegs(t, &nregs);
	for (nregs--; nregs >= 0; nregs--) {
	    if (ValidHandle(regs[nregs])) {
	        jvmpi_dump_add_root((JHandle *)regs[nregs],
				    GC_ROOT_NATIVE_STACK, 
				    NULL,
				    ee);
	    } else if (ValidObject(regs[nregs])) {
	        JHandle *h =
		    jvmpi_find_handle_from_object((void *)regs[nregs]);
		if (h) {
		    jvmpi_dump_add_root(h,
					GC_ROOT_NATIVE_STACK, 
					NULL,
					ee);
		}
	    }
	}

	base = ee->stack_base;
	ssc  = sysThreadStackPointer(t);
    }

    if (EE2SysThread(ee) == me) {
	ssc = (unsigned char **) &me;
    }

    if (ssc == 0 || base == 0 || (ssc == base)) {
	/*
	 * If the stack does not have a top of stack pointer or a base
	 * pointer then it hasn't run yet and we don't need to scan 
	 * its stack.  When exactly each of these data becomes available
	 * may be system-dependent, but we need both to bother scanning.
	 */
	goto ScanJavaStack;
    }

    /* Align stack top, important on Windows 95. */
    if ((long)ssc % sizeof(void *)) {
        ssc = (unsigned char **)((long)(ssc) & ~(sizeof(void *) - 1));
    }

    limit = (unsigned char **) base;

    sysAssert(ssc != limit);

    /*
     * The code that scans the C stack is assuming that the current
     * stack pointer is at a lower address than the limit of the stack.
     * Obvioulsy, this is only true for downward growing stacks.  For
     * upward growing stack, we exchange ssc and limit before we start
     * to scan the stack.
     */

#if defined(STACK_GROWS_UP)
    {
        unsigned char  **tmp;

        tmp = limit;
        limit = ssc;
        ssc = tmp;
    }
#endif /* STACK_GROWS_UP */


    while (ssc < limit) {
	register unsigned char *ptr = *ssc;
	if (ValidHandle(ptr)) {
	    /* If ptr is (apparently) an object pointer, it is pinned */
	    jvmpi_dump_add_root((JHandle *)ptr,
				GC_ROOT_NATIVE_STACK,
				NULL,
				ee);
	} else if (ValidObject(ptr)) {
	    JHandle *h = jvmpi_find_handle_from_object(ptr);
	    if (h) {
	        jvmpi_dump_add_root(h,
				    GC_ROOT_NATIVE_STACK, 
				    NULL,
				    ee);
	    }
	}
	ssc++;
    }

    /*
     * Whether or not we scan the thread stack, we decide independently
     * whether to scan the Java stack.  Doing so should be more robust
     * in the face of partially-initialized or partially-zeroed threads
     * during thread creation or exit, or changes to any of that code.
     */
ScanJavaStack:
    {
	JavaFrame *frame;

	/*
	 * Because of the Invocation API, the EE may not be on the C
	 * stack anymore.
	 */
	if (ValidHandle(ee->exception.exc))
	    jvmpi_dump_add_root(ee->exception.exc, 
				GC_ROOT_THREAD_BLOCK, 
				NULL,
				ee);

	if (ValidHandle(ee->pending_async_exc))
	    jvmpi_dump_add_root(ee->pending_async_exc,
				GC_ROOT_THREAD_BLOCK,
				NULL,
				ee);

	if ((frame = ee->current_frame) != 0) {
	    struct methodblock *prev_current_method = 0;
	    int dist_from_top = 0;
	    while (frame) {
	        struct methodblock *current_method = frame->current_method;
		/*
		 * If the previous frame was a transition frame from C back 
		 * to Java (indicated by prev_current_method == NULL), then 
		 * this new frame might not have set its optop.  We must be 
		 * conservative.   Otherwise, we can use the optop value.
		 *
		 * Also permit two consecutive frames with NULL current
		 * methods, in support of JITs.  See bug 4022856.
		 */
	        stack_item *top_top_stack = 
		  (prev_current_method == 0 && current_method != NULL &&
		   ((current_method->fb.access & ACC_NATIVE) == 0))
		    ? &frame->ostack[frame->current_method->maxstack] 
		    : frame->optop;
		jvmpi_scan_java_frame(frame,
				      top_top_stack,
				      GC_ROOT_JNI_LOCAL,
				      ee,
				      dist_from_top);
		{
		    JavaFrame *frame_ptr = frame;
		    JavaFrame frame_buf;
		    frame = frame->prev;
		    while (frame_ptr != frame) {
		        if (frame_ptr->current_method) {
			    dist_from_top++;
			}
		        frame_ptr = FRAME_PREV(frame_ptr, &frame_buf);
		    }
		}
		prev_current_method = current_method;
	    }
	} else {
	    /*EMPTY*/
	}

    }

    /* ScanAllocCache */ {
        struct alloc_cache *cp = &ee->alloc_cache;
	if (cp->cache_busy) {
	    /* Mark the cache tail */
	    JHandle *cache_tail = (JHandle *)(cp->cache_tail);
	    if (ValidHandle(cache_tail))
	        jvmpi_dump_add_root(cache_tail, 
				    GC_ROOT_NATIVE_STACK,
				    NULL,
				    ee);
	    /* Local handles are untouched because their obj field
	       is NULL.  Blocks, if loaded, will be pinned from
	       the cacheAlloc frame. */
	}
    }

    return SYS_OK;
}

static void 
jvmpi_scan_threads(void)
{
    sys_thread_t *self = sysThreadSelf();

    sysThreadEnumerateOver(jvmpi_scan_thread_traces, (void *) self);
    sysThreadEnumerateOver(jvmpi_scan_thread_roots, (void *) self);
}

static void
jvmpi_scan_globalrefs(void)
{
    jvmpi_scan_java_frame(globalRefFrame, globalRefFrame->optop,
			  GC_ROOT_JNI_GLOBAL, 0, 0);
}

static void
jvmpi_scan_primitive_classes(void)
{
    static ClassClass **primitive_classes[] = {
	&class_void, &class_boolean, &class_byte, &class_char, &class_short,
	&class_int, &class_long, &class_float, &class_double, NULL
    };
    ClassClass ***cbpp = primitive_classes;
    SetLimits();
    while (*cbpp) {
	ClassClass *cb = **cbpp;
	if (ValidHandle(cb)) {
	    jvmpi_dump_add_root((JHandle *)cb,
				GC_ROOT_STICKY_CLASS, 0, 0);
	}
	cbpp++;
    }
}

static void jvmpi_scan_sticky_classes(void)
{
    ClassClass **pcb = binclasses;
    int i;

    for (i = nbinclasses; --i >= 0; pcb++) {
	ClassClass *cb = *pcb;
	if (!classgc || (cbFlags(cb) & CCF_IsSticky))
	    jvmpi_dump_add_root((JHandle *)cb, 
				GC_ROOT_STICKY_CLASS, 0, 0);
    }

    /* Mark handles to primitive classes */
    jvmpi_scan_primitive_classes();
}

void jvmpi_scan_monitor_cache_helper(monitor_t *mid, void *cookie)
{
    SetLimits();
    JHandle *h = (JHandle*) mid->key;

    if (h && ValidHandle(h) && h->obj && sysMonitorInUse(sysmon(mid))) {
        jvmpi_dump_add_root(h, GC_ROOT_MONITOR_USED, 0, 0);
    }
}

static void jvmpi_scan_monitor_cache(void)
{
    monitorEnumerate(jvmpi_scan_monitor_cache_helper, 0);
}

static void jvmpi_gc_dump_heap(void)
{
    JHandle *hp;
    SetLimits();

    jvmpi_scan_threads();
    jvmpi_scan_globalrefs();
    jvmpi_scan_sticky_classes();
    jvmpi_scan_monitor_cache();

    MAP_OVER_HANDLES_FROM_START(hp)
	if (hp->obj && ValidObject(hp->obj) && IsMarked(hp)) {
	    jvmpi_dump_object(hp);
	}
    END_MAP_OVER_HANDLES_FROM_START;
}

#endif /* HPROF */

/*
 * Debugging facilities and useful stuff to call from dbx...
 */

#ifdef DEBUG

#ifdef PAGED_HEAPS /************ PAGED HEAPS: ********************/

void
checkChunk(ChunkBlkP chunk)
{
    sysAssert(chunk);
    sysAssert((long)(chunk->endPtr - chunk->startPtr) == chunk->allocSize);
    /* freePtr is either a head of a free list, or it's a hint, and 
     * must be in range: */
    sysAssert((chunk->freePtr == NULL) ||
	      ((chunk->freePtr >= chunk->startPtr) &&
	       (chunk->freePtr <= chunk->endPtr)));
    sysAssert((chunk->freeCnt >= 0) &&
	      (chunk->freeCnt <= chunk->allocSize));
    sysAssert(chunk->endPtr <= chunk->physEndPtr);
    sysAssert((*(int32_t *)(chunk->physEndPtr - 8)) == ALMOST_WORD);
    sysAssert((*(int32_t *)(chunk->physEndPtr - 4)) == ULTIMATE_WORD);
    sysCheckWastedSpace(chunk);
}

void 
verifyPageMap(void)
{
    ChunkBlkP blkP;
    int handlePagesByList = 0;
    int objectPagesByList = 0;
    int handlePagesByMap = 0;
    int objectPagesByMap = 0;
    int i;
    int limit = PageMapLength();
    
    /* may be called before heaps have been initialized, so be careful */
    sysAssert(page_map);
    blkP = firstHanBlkP;
    if (blkP) {
	do {
	    checkChunk(blkP);
	    handlePagesByList += PTR_2_PAGENUM(blkP->allocSize) + 1;
	    blkP = blkP->nextPtr;
	} while (blkP != firstHanBlkP);
    }

    blkP = firstObjBlkP;
    if (blkP) {	    
	do {
	    checkChunk(blkP);
	    objectPagesByList += PTR_2_PAGENUM(blkP->allocSize) + 1;
	    blkP = blkP->nextPtr;
	} while (blkP != firstObjBlkP);
    }

    for (i = 0; i < limit; i++) {
	if (page_map[i].chunk_size > 0) {
	    objectPagesByMap++;
	    sysAssert(page_map[i].mark_bits != NULL);
	} else if (page_map[i].chunk_size < 0) {
	    handlePagesByMap++;
	    sysAssert(page_map[i].mark_bits != NULL);
	} else {
	    sysAssert(page_map[i].mark_bits == NULL);
	}
    }
    
    sysAssert(handlePagesByList == handlePagesByMap);
    sysAssert(objectPagesByList == objectPagesByMap);
}

void 
DumpHeapInfo()
{
    ChunkBlkP hanP = firstHanBlkP;
    ChunkBlkP blkP = firstObjBlkP;

    jio_fprintf(stderr, "DumpHeapInfo\n");

    do {
	jio_fprintf(stderr, 
	 "handle chunk 0x%lx; freeCnt=0x%lx, freePtr=0x%lx, endPtr=0x%lx\n",
		hanP, hanP->freeCnt, hanP->freePtr, hanP->endPtr);
	checkChunk(hanP);
	hanP = hanP->nextPtr;
    } while (hanP != firstHanBlkP);

    do {
	jio_fprintf(stderr, 
"object chunk 0x%lx; freeCnt=0x%lx, freePtr=0x%lx, endPtr=0x%lx, pinned=%d\n",
	        blkP, blkP->freeCnt, blkP->freePtr, blkP->endPtr, 
		blkP->chunkFlags & CHUNK_PINNED);
	checkChunk(blkP);
	blkP = blkP->nextPtr;
    } while (blkP != firstObjBlkP);

    jio_fprintf(stderr, "\n");
}

#else  /************ CONTIGUOUS HEAPS: ********************/

void
DumpHeapInfo()
{
    ExecEnv *ee = EE();
    HEAP_LOCK(EE2SysThread(ee));

    jio_fprintf(stderr, "Heap layout info:\n");
    jio_fprintf(stderr, "  Mapped memory:\t[0x%p,0x%p), size: 0x%x\n",
	    heapbase, heaptop, heaptop-heapbase);
    jio_fprintf(stderr, "  Committed memory:\t[0x%p,0x%p), size: 0x%x\n",
	    hpool, opoollimit, opoollimit-hpool);
    jio_fprintf(stderr, "  Handle space: \t[0x%p,0x%p], size: 0x%x\n",
	    hpool, hpoollimit, opool-hpool);
    jio_fprintf(stderr, "  Object space: \t[0x%p,0x%p], size: 0x%x\n",
	    opool, opoollimit, opoollimit-opool+sizeof(hdr));

    HEAP_UNLOCK(EE2SysThread(ee));
}

#endif  /************ END CONTIGUOUS HEAPS ********************/

int
isMarked(char *p)
{
    SetLimits();
    return (IsMarked(p) ? 1 : 0);
}

int
isHorO(void *p)
{
    SetLimits();
    return (ValidHorO(p) ? 1 : 0);
}

void
printPool(char *s)
{
    SetLimits();
    register unsigned char *p;
    register col = 0;

    if (s)
	jio_fprintf(stderr, "\nprintPool: %s\n", s);

    MAP_OVER_OBJECTS_FROM_START(p)
	if (++col > 7) {
	    jio_fprintf(stderr, "\n");
	    col = 1;
	}
	if (obj_len(p) < 4 
#ifdef PAGED_HEAPS
	    || obj_len(p) > (MOO_chunk->allocSize)
#endif
	    ) {
	    jio_fprintf(stderr, "Bogus len 0x%lx in blk 0x%p\n", obj_len(p), p);
	    break;
	}
	jio_fprintf(stderr, "%lX%c%c%c\t", obj_len(p),
		p == MOO_start ? '*' : ' ',
		obj_free(p) ? 'F' : ' ',
		IsMarked(p + sizeof(hdr)) ? 'P' : ' ');
    END_MAP_OVER_OBJECTS_FROM_START;
}

void
printHandles(char *s)
{
    SetLimits();
    JHandle *hp;

    if (s)
	jio_fprintf(stderr, "\nprintHandles: %s", s);

    MAP_OVER_HANDLES_FROM_START(hp)
	if (hp->obj) {
	  jio_fprintf(stderr, "han=%p %c  obj=%p %c\n",
		  hp,
		  ValidHandle(hp) ? (IsMarked(hp) ? 'M' : ' ') : '*',
		  hp->obj,
		  ValidObject(hp->obj) ? (IsMarked(hp->obj) ? 'P' : ' ') : '*');
	}
    END_MAP_OVER_HANDLES_FROM_START;
    jio_fprintf(stderr, "\n");
}

void
verifyObjectCtrs(void)
{
    SetLimits();
    register unsigned char *p;
    int free = 0;
    int inuse = 0;
    int total = 0;
    int chunkFree = 0;

    MAP_OVER_OBJECTS_FROM_START(p)
	if (obj_free(p)) {
	    chunkFree += obj_len(p);
	} else {
	    inuse += obj_len(p);
	}
	total += obj_len(p);
	if ((p + obj_len(p)) >= MOO_limit) {
#ifdef PAGED_HEAPS
	    sysAssert(chunkFree == MOO_chunk->freeCnt);
#endif /* PAGED_HEAPS */
	    free += chunkFree;
	    chunkFree = 0;
	}
    END_MAP_OVER_OBJECTS_FROM_START;

    sysAssert(free == FreeObjectCtr);
    sysAssert(inuse == TotalObjectCtr - FreeObjectCtr);
    sysAssert(total == TotalObjectCtr);
    
/*  For more verbose output:

    jio_fprintf(stderr, "verifyObjectCtrs() detected:\n");
    jio_fprintf(stderr, "    %d bytes free versus %d bytes accumulated\n",
	    free, FreeObjectCtr);
    jio_fprintf(stderr, "    %d bytes in use versus %d bytes accumulated\n",
	    inuse, TotalObjectCtr - FreeObjectCtr);
    jio_fprintf(stderr, "    %d bytes total versus %d bytes accumulated\n",
	    total, TotalObjectCtr);
*/
}

static int cachedHandlesSize;
static int
countCachedHandles(sys_thread_t *t, void *arg)
{
    ExecEnv *ee = SysThread2EE(t);
    JHandle *hp = ee->alloc_cache.cache_handles;
    while (hp) {
	sysAssert(hp->obj == NULL);
	cachedHandlesSize += sizeof(JHandle);
	hp = (JHandle *)hp->methods;
    }
    return SYS_OK;
}

void
verifyHandleCtrs(void)
{
    SetLimits();
    register JHandle *hp;
    int free = 0;
    int inuse = 0;
    int total = 0;
    int chunkFree = 0;

    MAP_OVER_HANDLES_FROM_START(hp);
	if (hp->obj == 0) {
	    chunkFree += sizeof(JHandle);
	} else {
	    inuse += sizeof(JHandle);
	}
	total += sizeof(JHandle);
	if ((hp + 1) >= MOH_limit) {
#ifdef PAGED_HEAPS
	    sysAssert(chunkFree == MOH_chunk->freeCnt);
#endif /* PAGED_HEAPS */
	    free += chunkFree;
	    chunkFree = 0;
	}
    END_MAP_OVER_HANDLES_FROM_START;
    
    {
	/* 
	 * Account for handles held by thread local allocation cache.
	 */
	void * self = SysThread2EE(sysThreadSelf())->thread;
	int err;
	cachedHandlesSize = 0;
	err = sysThreadEnumerateOver(countCachedHandles, (void *)
				     self);
	free -= cachedHandlesSize;
	inuse += cachedHandlesSize;
    }

    {
	/* 
	 * Are all free handles on the free list? A rather slow, but
	 * good, test for unreturned handles.
	 */
	JHandle *iter = (JHandle *)hpoolfreelist;
	int freeListSize = 0;
	while (iter) {
	    freeListSize++;
	    iter = (JHandle *)iter->methods;
	}
	freeListSize *= sizeof(JHandle);

	if (freeListSize != free) {
	    /* painfully slowly find the leaked handle */
	    MAP_OVER_HANDLES_FROM_START(hp);
	        if (hp->obj == 0) { /* a free handle */
		    int found = 0;
		    register JHandle *fhp = (JHandle *)hpoolfreelist;
		    while (fhp) {
			if (fhp == hp) {
			    found = 1;
			    break;
			}
			fhp = (JHandle *)fhp->methods;
		    }
#ifdef VERBOSEGC
		    if (verbosegc && !found) {
			jio_fprintf(stderr, "<GC: handle leaked: 0x%p>\n", hp);
		    }
#endif
		}
	    END_MAP_OVER_HANDLES_FROM_START(hp);
	}
	/* now that we printed the leak, die. */
	sysAssert(freeListSize == free);
    }
    
    sysAssert(free == FreeHandleCtr);
    sysAssert(inuse == TotalHandleCtr - FreeHandleCtr);
    sysAssert(total == TotalHandleCtr);

/*  For more verbose output:

    jio_fprintf(stderr, "verifyHandleCtrs() detected:\n");
    jio_fprintf(stderr, "    %d bytes free versus %d bytes accumulated\n",
	    free, FreeHandleCtr);
    jio_fprintf(stderr, "    %d bytes in use versus %d bytes accumulated\n",
	    inuse, TotalHandleCtr - FreeHandleCtr);
    jio_fprintf(stderr, "    %d bytes total versus %d bytes accumulated\n",
	    total, TotalHandleCtr);
*/
}

void 
validatePool(long quick)
{
    SetLimits();
    register unsigned char *p;
    register JHandle *hp;
    int err = 0;

#ifdef PAGED_HEAPS
    verifyPageMap();
#endif /* PAGED_HEAPs */

    MAP_OVER_OBJECTS_FROM_START(p)
	sysAssert(obj_len(p) >= OBJECTGRAIN);  /* 4 didn't seem right */
	sysAssert((p + obj_len(p)) <= MOO_limit);
	sysAssert(ValidObject(p + sizeof(hdr)));
#ifdef PAGED_HEAPS  
	sysAssert(p >= MOO_chunk->startPtr);
#endif /* PAGED_HEAPs */
	if (obj_free(p)) {
	    FILL_CHECK_WORD(p);
	} else if (!quick) {
	    MAP_OVER_HANDLES_FROM_START(hp)
		if (hp->obj == (ClassObject *) (p + sizeof(hdr))) {
		    goto HasHandle;
		}
	    END_MAP_OVER_HANDLES_FROM_START;
	    jio_fprintf(stderr, "@%p -- allocated object with no handle\n", p);
	    err++;
	}
HasHandle:
    err = err;	/* the label needs something to attach itself to... */
    END_MAP_OVER_OBJECTS_FROM_START;
    

    MAP_OVER_HANDLES_FROM_START(hp)
	if (hp->obj) {
	    if (ValidObject(hp->obj) && 
	        obj_free((char *) hp->obj - sizeof(hdr))) {
		jio_fprintf(stderr,
		       "@%p/%p -- Handle points to free block\n", hp, hp->obj);
		err++;
	    } else {
		switch (obj_flags(hp)) {
	       	case T_NORMAL_OBJECT:
	       	    if (obj_methodtable(hp) == 0) {
	       		jio_fprintf(stderr, 
	       		"1 @%p/%p -- Invalid method table\n", hp, hp->methods);
	       		err++;
	       	    } else if (obj_methodtable(hp) != 
	       	               cbMethodTable(obj_classblock(hp))) {
	       		jio_fprintf(stderr, 
	       		  "2 @%p/%p -- Invalid method table\n", hp, hp->methods);
	       		err++;
	       	    }
	       	    break;
	           
	       	default:
	       	    if (obj_flags(hp) < T_CLASS ) {
			jio_fprintf(stderr, 
			 "3 @%p/%p -- Invalid method table\n", hp, hp->methods);
	       		err++;
		    }
		}
	    }
	}
    END_MAP_OVER_HANDLES_FROM_START;

#ifdef TRACEGC
    if (tracegc)
	jio_fprintf(stderr, err == 0 
	        ? "OK memory pool\n" 
	        : "***%d errors in the memory pool***\n", err);
#endif /* TRACEGC */
}

#endif /* DEBUG */


void scanMonitorCacheHelper(monitor_t *mid, void *cookie)
{
    SetLimits();
    JHandle *h = (JHandle*) mid->key;

    if (h && ValidHandle(h) && h->obj && sysMonitorInUse(sysmon(mid))) {
        MarkPtr(h);
    }
}

static void scanMonitorCache(void)
{
    monitorEnumerate(scanMonitorCacheHelper, 0);
}
